(* ::Package:: *)

(* ::Section:: *)
(* Public Declaration Part: executes in Theorema`Provers` *)


With[ {lang = "English"},

MessageName[ rewriteInteractiveProver, "usage", lang] = "Inference rules for rewriting and predicate logic";

MessageName[ failedInteractiveRule, "usage", lang] = "Interactive rule could not be applied";

MessageName[ RRPgoalInKB, "usage", lang] = "Knowledge base contains the proof goal";
MessageName[ RRPnotEquGoal, "usage", lang] = "Prove negated equalitiy/equivalence by contradiction";
MessageName[ RRPandKB, "usage", lang] = "Split conjunctions in the knowledge base";
MessageName[ RRPandGoal, "usage", lang] = "Split conjunction in the goal";
MessageName[ RRPorGoal, "usage", lang] = "Prove disjunction by proving last disjunct";
MessageName[ RRPorKB, "usage", lang] = "Case distinction based on a disjunction in the knowledge base";
MessageName[ RRPimplGoalDirect, "usage", lang] = "Prove implication directly";
MessageName[ RRPequivKB, "usage", lang] = "Transform equivalences in the knowledge base into implications";
MessageName[ RRPmodusPonensTollens, "usage", lang] = "Modus ponens and modus tollens";
MessageName[ RRPsimpImplication, "usage", lang] = "Simplify implications by splitting disjunctions in the premise";
MessageName[ RRPcontradiction, "usage", lang] = "Prove by contradiction";

MessageName[ RRPexpandNegationKB, "usage", lang] = "Simplify negated assumptions";
MessageName[ RRPexpandNegationGoal, "usage", lang] = "Simplify negated goal";
MessageName[ RRPexpandPiecewise, "usage", lang] = "Simplify/expand piecewise expressions";
MessageName[ RIexpandSuch, "usage", lang] = "Eliminate implicitly defined terms";

MessageName[ RRPforallGoal, "usage", lang] = "Prove universally quantified goal";
MessageName[ RRPsplitForallGoal, "usage", lang] = "Split universally quantified goal into individual quantifiers";
MessageName[ RRPexistsKB, "usage", lang] = "Eliminate existential quantifier in the knowledge base";
MessageName[ RIforallGoalSeqLeft, "usage", lang] = "Apply structural induction on universally quantified sequence variables in the goal";

MessageName[ RRPgenerateRewriteRules, "usage", lang] = "Generate rewrite rules from initial proof situation";
MessageName[ RRPgoalRewriting, "usage", lang] = "Goal rewriting based on formulae in the knowledge base";
MessageName[ RRPelementarySubstitution, "usage", lang] = "Substitution based on quantified equalities and equivalences in the knowledge base";
MessageName[ RRPgroundSubstitution, "usage", lang] = "Substitution based on ground equalities and equivalences in the knowledge base";
MessageName[ RRPexpandDef, "usage", lang] = "Expand explicit definitions";
MessageName[ RRPimplicitDef, "usage", lang] = "Expand implicit definitions";

MessageName[ RRPorGoalInteractive, "usage", lang] = "[I] Select disjunct to prove";
MessageName[ RRPexchangeGoalKBInteractive, "usage", lang] = "[I] Exchange the goal and a formula in the knowledge base";
MessageName[ RRPsplitForallGoalInteractive, "usage", lang] = "[I] Split universally quantified goal interactively";
MessageName[ RRPinstantiateInteractive, "usage", lang] = "[I] Instantiate quantified formula";
MessageName[ RRPforallKBInteractive, "usage", lang] = "Instantiate a universally quantified formula in the knowledge base";	(* only needed for text in proof tree *)
MessageName[ RRPselectImplicationInteractive, "usage", lang] = "[I] Select an implication in the knowledge base for proving its premise and assuming its conclusion";
MessageName[ RRPrewritePS, "usage", lang] = "[I] Rewrite the proof situation";
MessageName[ RRPintroduceFormula, "usage", lang] = "[I] Introduce an arbitrary formula";
MessageName[ RRPaddAbbreviation, "usage", lang] = "[I] Add an abbreviation for a term/formula";

MessageName[ RRPaddKnowledge, "usage", lang] = "[M] Add new knowledge to the proof situation";
MessageName[ RIhideKnowledge, "usage", lang] = "[M] Hide knowledge";
MessageName[ RIremoveKnowledge, "usage", lang] = "[M] Remove knowledge entirely";

MessageName[ RRPsplitConjunction, "usage", lang] = "Split conjunction in the goal into two parts";
MessageName[ RRPcheckConditions, "usage", lang] = "Check conditions of rewrite rules";

] (* With *)


(* ::Section::Closed:: *)
(* Private Implementation Part: executes in Theorema`Provers`Private` *)


Begin["`Private`"]

With[ {lang = "English"},

translate[ "Interactive Rules", lang] = "Interactive Rules";
translate[ "Management Rules", lang] = "Knowledge-Management Rules";
translate[ "Miscellaneous Rules", lang] = "Miscellaneous";
translate[ "Propositional Logic Rules", lang] = "Propositional Logic";
translate[ "RewriteInteractiveProver", lang] = "Inference Rules for Rewriting and Predicate Logic";
translate[ "Quantifier Rules", lang] = "Logical Quantifiers";
translate[ "Rules for Rewriting", lang] = "Rewriting";

translate[ "accept", lang] = "Accept";
translate[ "addAbbreviation", lang] = "Add Abbreviation";
translate[ "addGlob", lang] = "add globally";
translate[ "addLoc", lang] = "add locally";
translate[ "addKnowledge", lang] = "Add Knowledge";
translate[ "applyRepl", lang] = "apply replacement";

translate[ "cancel", lang] = "Cancel";
translate[ "checkConds", lang] = "Check conditions";
translate[ "currentGoal", lang] = "Current Goal";
translate[ "currentKnowledge", lang] = "Current Knowledge";

translate[ "defRules", lang] = "Definitions";
translate[ "distinguishCases", lang] = "distinguish cases";

translate[ "exchangeGoal", lang] = "Exchange Goal and Assumption";

translate[ "goal", lang] = "Goal";

translate[ "hideKnowledge", lang] = "Hide Knowledge";

translate[ "instantiateForall", lang] = "Instantiate Universally Quantified Formula";
translate[ "instantiateExists", lang] = "Instantiate Existentially Quantified Formula";
translate[ "introduceFormula", lang] = "Introduce Formula";

translate[ "keepOrig", lang] = "Keep originals";
translate[ "knowledge", lang] = "Knowledge";

translate[ "proofSituation", lang] = "Proof Situation";
translate[ "proveConds", lang] = "Prove conditions";
translate[ "proveDomCondsImmediately", lang] = "prove domain conditions immediately";
translate[ "proveFormula", lang] = "prove formula";
translate[ "provePremise", lang] = "prove premise";

translate[ "removeKnowledge", lang] = "Remove Knowledge";
translate[ "removeOrig", lang] = "remove originals";
translate[ "rewriteProofSit", lang] = "Rewrite Proof Situation";
translate[ "rewriteRules", lang] = "Rules";

translate[ "scanRulesFirst", lang] = "Replace first";
translate[ "scanRulesAll", lang] = "Replace all";
translate[ "scanRulesRepeated", lang] = "Replace repeatedly";
translate[ "scanFormulas", lang] = "Scan formulas, replace individually";
translate[ "selectFormula", lang] = "Select Quantified Formula";
translate[ "selectGoal", lang] = "Select goal to prove";
translate[ "selectImplication", lang] = "Select Implication";
translate[ "simplifyForall", lang] = "Simplify Conditions/Premises";
translate[ "substRules", lang] = "Substitutions";

translate[ "ttAddGlobLoc", lang] = "Specify whether to add the formulas locally (only current proof situation) or globally (all relevant proof situations)";
translate[ "ttApplyRepl", lang] = "Specify whether to immediately carry out replacements in the proof situation, if the formula is an equality/equivalence";
translate[ "ttCheckConds", lang] = "Specify whether to check conditions arising from conditional rewrite-rules immediately by backward rewriting";
translate[ "ttDistinguishProve", lang] = "Specify whether to distinguish two cases depending on whether the formula holds or not, or whether to just prove the formula";
translate[ "ttKeepRemoveOrig", lang] = "Specify whether to keep the original formulas in the knowledge base or replace them by the new ones";
translate[ "ttProveConds", lang] = "Specify whether to prove conditions arising from conditional rewrite-rules in separate sub-proofs and apply the rules anyway";
translate[ "ttProveDistinguish", lang] = "Specify whether the premise of the implication should be proved in a subproof, or whether two cases should be distinguished based on whether it holds or not";
translate[ "ttProveDomCondsImmediately", lang] = "Specify whether domain conditions on the instantiated variables shall be proved immediately in a subproof";
translate[ "ttRewritingControl", lang] = "Select the method that is used for rewriting (i.e. the \"control\")";
translate[ "ttSimplifyForall", lang] = "Specify whether conditions and premises of the universal quantifier shall be simplified or not";






(* ::Section::Closed:: *)
(* Propositional Logic *)


proofStepText[ RRPandGoal, lang, {{g_}}, _, ___] :=
	{textCell[ "For proving ", formulaReference[ g],
		" we prove the individual conjuncts one after the other, immediately adding the ones already proved to our knowledge base."]
	};
subProofHeader[ RRPandGoal, lang, _, {generated_List}, ___, pVal_, {p_}] := {textCell[ "Proof of ", formulaReference[ Part[ generated, p]], ":"],
	textCell[ "We need to prove"],
	goalCell[ Part[ generated, p], "."]
	};

proofStepText[ RRPandKB, lang, {used_List}, {generated_List}, ___] :=
	Module[ {suffix = If[ Length[ used] === 1, "", "s"]},
		{textCell[ "We split the conjunction" <> suffix <> " ", formulaReferenceSequence[ used, lang], " and add"],
			assumptionListCells[ generated, ",", ""],
			textCell[ "to the knowledge base."]
		}
	];
	
proofStepText[ RRPorKB, lang, {{g_}}, {generated_List}, ___] :=
	{textCell[ "Based on the assumption ", formulaReference[ g], " we distinguish " <>
		intToText[ lang, Length[ generated]] <> " cases:"]
	};
subProofHeader[ RRPorKB, lang, _, {generated_List}, ___, pVal_, {p_}] :=
	{textCell[ "Case ", ToString[p], ": We assume"], assumptionCell[ Part[ generated, p], "."]};

intToText[ lang, a_Integer] :=
	Switch[ a,
		2, "two",
		3, "three",
		4, "four",
		5, "five",
		6, "six",
		7, "seven",
		8, "eight",
		9, "nine",
		10, "ten",
		11, "eleven",
		12, "twelve",
		_, ToString[ a]
	];
	
proofStepText[ RRPequivKB, lang, {equiv_List}, gen:{{_, _}..}, ___] /; Length[ equiv] === Length[ gen] :=
	Module[ {},
		Flatten[ MapThread[(
				{
					textCell[ "From equivalence ", formulaReference[ #1], " we obtain the implications"],
					assumptionListCells[ #2, ",", "."]
				}
				)&,
				{equiv, gen}
			],
			1
		]
	];
	
(* We still need the old proof-step-text, because many proofs have already been generated with the old modus-ponens rule. *)
proofStepText[ RRPmodusPonensTollens, lang, u_List, {{g_FML$, other___FML$}}, rest___] :=
	proofStepText[ RRPmodusPonensTollens, lang, u, {{g}, {other}}, Sequence@@Replace[ {rest}, ("action" -> "simplification") -> ("action" -> "modus ponens"), {1}]];
subProofHeader[ RRPmodusPonensTollens, lang, u_, {{g_FML$, cond_FML$, goal_FML$}}, rest__] :=
	subProofHeader[ RRPmodusPonensTollens, lang, u, {{g}, {cond, goal}}, rest];
(* The following is the new proof-step-text for modus-ponens. *)
proofStepText[ RRPmodusPonensTollens, lang, {u:{__FML$}}, {g:{__FML$}, _List}, ___, "usedForms" -> refs_List, "action" -> act_String, ___] := 
	Prepend[
		Flatten[ MapThread[ {textCell[ "From formula ", formulaReference[ #1], " we can infer"], assumptionCell[ #2, "."]}&,
							{u, g}
				],
				1
		],
		If[ refs === {},
			textCell[ "We apply ", act, ":"],
			textCell[ "We apply ", act, ", making use of ", formulaReferenceSequence[ refs, lang], ":"]
		]
	];
subProofHeader[ RRPmodusPonensTollens, lang, _, {_, {cond_FML$, _}}, ___, {1}] :=
	{textCell[ "First, however, we have to prove"], goalCell[ cond, "."]};
subProofHeader[ RRPmodusPonensTollens, lang, _, {_, {cond_FML$, goal_FML$}}, ___, {2}] :=
	{textCell[ "Now we continue proving our original goal ", formulaReference[ goal], " under the additional assumption"],
		assumptionCell[ cond, "."]
	};
	
proofStepText[ RRPsimpImplication, lang, {used_List}, {generated_List}, ___] :=
	Module[ {suffix, poss},
		{suffix, poss} = If[ Length[ used] === 1, {"", "its"}, {"s", "their"}];
		{textCell[ "We simplify the implication" <> suffix <> " ",
				formulaReferenceSequence[ used, lang],
				" by splitting the disjunctions in " <> poss <> " premise" <> suffix <> ". Hence, we add"
			],
			assumptionListCells[ generated, ",", ""],
			textCell[ "to the knowledge base."]
		}
	];
	
	


(* ::Section::Closed:: *)
(* Miscellaneous *)


	
proofStepText[ RRPexpandNegationKB, lang, {{used_FML$}}, {{gen_FML$}}, ___] := 
	{textCell[ "We simplify the negated formula ", formulaReference[ used], " in our current knowledge base to"],
				assumptionCell[ gen, "."]
	};
				
proofStepText[ RRPexpandNegationGoal, lang, {{used_FML$}}, {{gen_FML$}}, ___] := 
	{textCell[ "We simplify the negated goal ", formulaReference[ used], " to"],
				goalCell[ gen, "."]
	};
	
proofStepText[ RRPexpandPiecewise, lang, {{f_}, refs_List}, {gen_List, _List}, rest___?OptionQ, _] :=
	Module[ {place, new, orig, simp, out},
		{place, orig, new, simp} = {"place", "orig", "new", "simp"} /. {rest};
		out = Switch[ place,
			"kb",
			{textCell[ "We expand the piecewise-expression ", inlineTheoremaExpression[ orig], " in assumption ", formulaReference[ f], "."]},
			"goal",
			{textCell[ "We expand the piecewise-expression ", inlineTheoremaExpression[ orig], " in the current goal ", formulaReference[ f], "."]},
			_,
			{textCell[ "We expand the piecewise-expression ", inlineTheoremaExpression[ orig], "."]}
		];
		Join[ out,
			Which[
				!TrueQ[ simp],
				{},
				refs === {},
				{textCell[ "Before, however, we simplify it to ", inlineTheoremaExpression[ new], "."]},
				True,
				{textCell[ "Before, however, we simplify it to ", inlineTheoremaExpression[ new], ", using ", formulaReferenceSequence[ refs, lang], "."]}
			],
			If[ Length[ gen] > 1,
				{textCell[ "We now distinguish " <> intToText[ lang, Length[ gen]] <> " cases:"]},
				Switch[ place,
					"kb",
					{textCell[ "Hence, we add"], assumptionCell[ First[ gen], ""], textCell[ "to our knowledge base."]},
					_,
					{textCell[ "Hence, we now prove"], goalCell[ First[ gen], "."]}
				]
			]
		]
	];
subProofHeader[ RRPexpandPiecewise, lang, _, {gen_List, assm_List}, ___, "place" -> place_, ___, {p_}] :=
	Join[
		{textCell[ "Case ", ToString[ p], ": We assume"], assumptionListCells[ assm[[p]], ",", ""]},
		Switch[ place,
			"kb",
			{textCell[ "and"], assumptionCell[ gen[[p]], "."]},
			_,
			{textCell[ "and prove"], goalCell[ gen[[p]], "."]}
		]
	];
	
proofStepText[ RIexpandSuch, lang, {{f_}}, _, ___, "data" -> {{p_, ___}, term_, _}, ___] :=
	Switch[ p,
		1,
		{textCell[ "We eliminate the implicit term ", inlineTheoremaExpression[ term], " in the current goal ", formulaReference[ f], "."]},
		2,
		{textCell[ "We eliminate the implicit term ", inlineTheoremaExpression[ term], " in assumption ", formulaReference[ f], "."]},
		_,
		{}
	];
subProofHeader[ RIexpandSuch, lang, _, {{cond_, __}}, ___, {1}] :=
	{textCell[ "Before, however, we need to make sure that a suitable term exists, i.e. we have to prove"], goalCell[ cond, "."]};
subProofHeader[ RIexpandSuch, lang, _, {{_, new_, r_, c_, f_}}, ___, "data" -> {{p_, ___}, term_, fix_}, ___, {2}] :=
	Join[
		{textCell[ "Now we replace ", inlineTheoremaExpression[ term], " by the fresh term ", inlineTheoremaExpression[ fix], ", assume"],
		assumptionListCells[ {r, c, f}, ",", ""]},
		Switch[ p,
			1,
			{textCell[ "and prove"], goalCell[ new, "."]},
			2,
			{textCell[ "and"], assumptionCell[ new, ""], textCell[ "and prove our original goal."]},
			_,
			{}
		]
	];
				
	


(* ::Section:: *)
(* Quantifiers *)


proofStepText[ RRPsplitForallGoal, lang, {{g_}}, {{newG_}}, ___] :=
	{
		textCell[ "We split the multi-range in our universally quantified goal ", formulaReference[ g],
		" into smaller ranges. Therefore, we now prove"],
		goalCell[ newG, "."]
	};
    
proofStepText[ RIforallGoalSeqLeft, lang, {{g_}}, {{newG_}}, ___, "var" -> v_, ___] :=
	{
		textCell[ "We perform structural induction on the sequence variable ", inlineTheoremaExpression[ v], " in our current goal ", formulaReference[ g],
		". Hence the goal is reduced to"],
		goalCell[ newG, "."]
	};


(* ::Section::Closed:: *)
(* Rewriting *)


proofStepText[ RRPgenerateRewriteRules, lang, {}, {}, pVal_] := {textCell[ "We generate rewrite rules and try to simplify the current goal and the current knowledge base by computation."]};

proofStepText[ goalRewriting, lang, {{g_, rewrittenBy__}}, {}, ___] :=
	{textCell[ "Our current goal ", formulaReference[ g], " follows readily from ", formulaReferenceSequence[ {rewrittenBy}, lang],
		", hence we are done with this branch of the proof."]};

proofStepText[ RRPimplicitDef, lang, {{def_FML$, ref_List}, used__}, g_List, rest1___, "introConstFor" -> {old_, new_}, rest2___] := 
	Module[ {conds},
		conds = "conditions" /. Cases[ {rest1, rest2}, (Rule|RuleDelayed)[ _String, _]] /. "conditions" -> False;
		If[ TrueQ[ conds],
			{textCell[ "We try to expand the implicit definition ", formulaReference[ def], ":"]},
			Prepend[ implicitDefText[ {used}, g, ref, old, new], textCell[ "We expand the implicit definition ", formulaReference[ def], ":"]]
		]
	];
subProofHeader[ RRPimplicitDef, lang, __, {1}] :=
	{};
subProofHeader[ RRPimplicitDef, lang, {{def_FML$, ref_List}, used__}, g_List, ___, "introConstFor" -> {old_, new_}, ___, {2}] :=
	Prepend[ implicitDefText[ {used}, g, ref, old, new],
		textCell[ "Now we are done with checking conditions and can proceed with expanding the implicit definition ", formulaReference[ def], ":"]
	];
	
implicitDefText[ u_List, g_List, ref_List, old_, new_] :=
	Module[ {stepText = {}, j},
		stepText = {textCell[ "We replace every occurrence of ", inlineTheoremaExpression[ old], " by the new constant ", 
			inlineTheoremaExpression[ new], " with"],
			assumptionListCells[ g[[1]], ",", "."]};
		If[ ref =!= {},
			AppendTo[ stepText, textCell[ "The applicability follows immediately from ", formulaReferenceSequence[ ref, lang], "."]]
		];
		(* g[[2]] is the new goal. {} if no rewrite happened in the goal *)
		If[ g[[2]] =!= {},
			stepText = Join[ stepText,
				{textCell[ "For proving ", formulaReference[ u[[1, 1]]], " it suffices now to prove"], goalCell[ g[[2, 1]], "."]}
			]
		];
		(* Each of the remaining is an expansion in the KB. Produce a line of text for each of them *)
		Do[
			stepText = Join[ stepText, 
							{textCell[ "The assumption ", formulaReference[ u[[j - 1, 1]]], " becomes"], 
							assumptionCell[ g[[j, 1]]]}
						],
			{j, 3, Length[ g]}
		];
		stepText
	];
	
proofStepText[ rl:(RRPelementarySubstitution|RRPgroundSubstitution|RRPexpandDef), lang, u_List, g_List, rest1___, "usedForms" -> subst:{{_, _List}..}, rest2___] :=
	Module[ {conds, stepText, methodText, formulaText},
		conds = "conditions" /. Cases[ {rest1, rest2}, (Rule|RuleDelayed)[ _String, _]] /. "conditions" -> False;
		{stepText, methodText, formulaText} =
			Switch[ rl,
				RRPexpandDef,
				{"expand explicit definitions:", "definition ", "definition "},
				_,
				{"apply substitutions:", "substitution ", "formula "}
			];
		If[ conds =!= False,
			{textCell[ "We try to " <> stepText]},
			Prepend[ rewritingProofText[ u, g, subst, methodText, formulaText], textCell[ "We " <> stepText]]
		]
	];
subProofHeader[ RRPelementarySubstitution|RRPgroundSubstitution|RRPexpandDef, lang, ___, "conditions" -> cond_FML$, ___, {1}] :=
	{textCell[ "Therefore, we first need to make sure that the condition"], goalCell[ cond], textCell[ "holds."]};
subProofHeader[ RRPelementarySubstitution|RRPgroundSubstitution|RRPexpandDef, lang, ___, {1}] :=
	{};
subProofHeader[ rl:(RRPelementarySubstitution|RRPgroundSubstitution|RRPexpandDef), lang, u_List, g_List, ___, "usedForms" -> subst:{{_, _List}..}, ___, {2}] :=
	Module[ {stepText, methodText, formulaText},
		{stepText, methodText, formulaText} =
			Switch[ rl,
				RRPexpandDef,
				{"expanding the definitions:", "definition ", "definition "},
				_,
				{"applying the substitutions:", "substitution ", "formula "}
			];
		Prepend[ rewritingProofText[ u, g, subst, methodText, formulaText],
			textCell[ "Now we are done with checking conditions and can proceed with " <> stepText]
		]
	];

rewritingProofText[ u_List, g_List, subst_List, methodText_String, formulaText_String] :=
	Module[ {stepText = {}, j, repl},
		If[ u[[1]] =!= g[[1]],
			repl = subst[[1]];
			stepText = {textCell[ "In order to prove ", formulaReference[ u[[1, 1]]], ", using " <> formulaText, 
						formulaReference[ repl[[1]]], ", we now have to show"], goalCell[ g[[1, 1]], "."]};
			If[ repl[[2]] =!= {},
				AppendTo[ stepText, textCell[ "The applicability of this " <> methodText <> "follows immediately from ", formulaReferenceSequence[ repl[[2]], lang], "."]]
			]
		];
		Do[
			repl = subst[[j]];
			stepText = Join[ stepText, 
							{textCell[ "From ", formulaReference[ u[[j, 1]]], " we know, by " <> formulaText, formulaReference[ repl[[1]]], ","], 
							assumptionCell[ g[[j, 1]]]},
							If[ repl[[2]] === {},
								{},
								{textCell[ "The applicability of this " <> methodText <> "follows immediately from ",
								formulaReferenceSequence[ repl[[2]], lang], "."]}
							]
						],
			{j, 2, Length[ g]}
		];
		stepText
	];
	
proofStepText[ RRPcheckConditions, lang, _, {{g_FML$}}, ___] :=
	{textCell[ "Therefore, we first need to make sure that the condition"], goalCell[ g], textCell[ "is satisfied."]};
proofStepText[ RRPcheckConditions, lang, __] :=
	{textCell[ "Therefore, we first need to make sure that all conditions are satisfied."]};
subProofHeader[ RRPcheckConditions, lang, _, {g_List}, ___, {i_}] :=
	{textCell[ "Condition " <> ToString[ i] <> ": We have to show"], goalCell[ g[[i]], "."]};
	
	


(* ::Section::Closed:: *)
(* Interactive *)


proofStepText[ RRPexchangeGoalKBInteractive, lang, {{g_FML$}}, {{neg_FML$}}, ___] := 
	{ textCell[ "We assume the negation of our current goal ", formulaReference[ g], ":"],
				assumptionCell[ neg], textCell[ "and try to derive a contradiction."]};
proofStepText[ RRPexchangeGoalKBInteractive, lang, {{g_FML$, fml_FML$}}, {{neg_FML$, negFml_FML$}}, ___] := 
	{ textCell[ "We exchange our current goal ", formulaReference[ g], " and formula ", formulaReference[ fml], ", meaning that we prove"],
				goalCell[ negFml], textCell[ "under the additional assumption"], assumptionCell[ neg, "."]};
				
proofStepText[ RRPforallKBInteractive, lang, {u:{__}}, {{g_}}, ___, "instantiation" -> inst_, ___] :=
	{
		textCell[ "Formula", If[ Length[ u] === 1, " ", "s "], formulaReferenceSequence[ u, lang], " hold", If[ Length[ u] === 1, "s ", " "], "in particular for ",
		inlineTheoremaExpressionSeq[ inst, lang], ", i.e. we know"], assumptionCell[ g, "."]
	};
proofStepText[ RRPforallKBInteractive, lang, {u:{__}}, {{_, _}}, ___, "instantiation" -> inst_, ___] :=
	{
		textCell[ "Formula", If[ Length[ u] === 1, " ", "s "], formulaReferenceSequence[ u, lang], " hold", If[ Length[ u] === 1, "s ", " "], "in particular for ",
		inlineTheoremaExpressionSeq[ inst, lang], "."]
	};
subProofHeader[ RRPforallKBInteractive, lang, _, {{_, g_}}, ___, {1}] :=
	{textCell[ "We immediately prove all domain-conditions imposed on the variables instantiated for, so we prove"],
	goalCell[ g, "."]};
subProofHeader[ RRPforallKBInteractive, lang, {u:{__}}, {{g1_, g2_}}, ___, {2}] :=
	With[ {suffix = If[ Length[ u] === 1, " ", "s "]},
		{
			textCell[ "We now continue with the main branch of the proof, additionally knowing"],
			assumptionCell[ g2, ""],
			textCell[ "and the instance", suffix, "of formula", suffix, formulaReferenceSequence[ u, lang], ", i.e."],
			assumptionCell[ g1, "."]
		}
	];
	
proofStepText[ RRPselectImplicationInteractive, lang, {{f_FML$, g_FML$}}, {{_}, {_, _}}, ___] := 
	{textCell[ "We prove the premise of implication ", formulaReference[ f], " as well as our current goal ", formulaReference[ g], " assuming its conclusion:"]};
subProofHeader[ RRPselectImplicationInteractive, lang, _, {{goal_FML$}, _}, ___, {1}] :=
	{textCell[ "First part: We prove the premise of the implication, i.e. we prove"], goalCell[ goal, "."]};
subProofHeader[ RRPselectImplicationInteractive, lang, {{_, g_FML$}}, {_, assm:{_, _}}, ___, {2}] :=
	{textCell[ "Second part: We prove our original goal ", formulaReference[ g], " under the additional assumptions"], assumptionListCells[ assm, ",", "."]};

proofStepText[ RRPselectImplicationInteractive, lang, {{f_FML$, g_FML$}}, {{_, _}, {_}}, ___] := 
	{textCell[ "We distinguish two cases, based upon whether the premise of implication ", formulaReference[ f], " holds or not:"]};
subProofHeader[ RRPselectImplicationInteractive, lang, _, {_, {assm_FML$}}, ___, {1}] :=
	{textCell[ "First case: We assume that the premise does not hold, i.e. we assume"], assumptionCell[ assm, "."]};
subProofHeader[ RRPselectImplicationInteractive, lang, {{_, g_FML$}}, {assm:{_, _}, _}, ___, {2}] :=
	{textCell[ "Second case: We assume that the premise holds, so we prove our original goal ", formulaReference[ g],
		" under the additional assumptions"], assumptionListCells[ assm, ",", "."]};
	
proofStepText[ RRPintroduceFormula, lang, _, {{f1_}}, ___] := 
	{textCell[ "We add formula ", inlineTheoremaExpression[ formula@f1], " to our knowledge base."]};
subProofHeader[ RRPintroduceFormula, lang, _, {{f1_}}, ___, {1}] := 
	{textCell[ "Before we are allowed to do this, we of course have to prove the formula. Therefore, we now prove"], goalCell[ f1, "."]};
subProofHeader[ RRPintroduceFormula, lang, _, {{f1_}}, ___, {2}] := 
	{textCell[ "We now continue with the main branch of our proof, having"], assumptionCell[ f1, ""], textCell[ "among our assumptions."]};
proofStepText[ RRPintroduceFormula, lang, _, {{f1_, _}}, ___] := 
	{textCell[ "We distinguish two cases, based upon whether formula ", inlineTheoremaExpression[ formula@f1], " holds or not."]};
subProofHeader[ RRPintroduceFormula, lang, _, {{_, f2_}}, ___, {1}] := 
	{textCell[ "First case: We assume the negation of the formula, i.e. we assume"], assumptionCell[ f2, "."]};
subProofHeader[ RRPintroduceFormula, lang, _, {{f1_, _}}, ___, {2}] := 
	{textCell[ "Second case: We assume that the formula holds, i.e. we assume"], assumptionCell[ f1, "."]};
	
proofStepText[ RRPaddAbbreviation, lang, _, {{f_}}, ___] := 
	{textCell[ "We introduce the abbreviation"], assumptionCell[ f, "."]};
	
proofStepText[ RRPaddKnowledge|RIremoveKnowledge|RIhideKnowledge, lang, ___] := {};

proofStepText[ failedInteractiveRule, lang, ___] := {};
	
] (* With *)

End[]
