(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[    408898,      11508]
NotebookOptionsPosition[    278134,       8047]
NotebookOutlinePosition[    375756,      10481]
CellTagsIndexPosition[    368743,      10287]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Reduction Fields", "Title",
 CellChangeTimes->{{3.6457582886658278`*^9, 3.64575830204416*^9}},
 CellID->1754384754],

Cell[CellGroupData[{

Cell["Preliminaries", "Section",
 CellChangeTimes->{{3.645758308145159*^9, 3.645758310460437*^9}},
 CellID->467603975],

Cell[CellGroupData[{

Cell["Description", "Subsection",
 CellChangeTimes->{{3.660905279393654*^9, 3.660905281827258*^9}},
 CellID->76201264],

Cell[TextData[{
 "This notebook defines the functor \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"ReductionField", "[", "K", "]"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] that extends the domain ",
 Cell[BoxData[
  FormBox["K", TraditionalForm]]],
 " with the operations necessary for being an algorithmic reduction ring. \
Indeed, if ",
 Cell[BoxData[
  FormBox["K", TraditionalForm]]],
 " is a field, the resulting domain ",
 StyleBox["is",
  FontSlant->"Italic"],
 " an algorithmic reduction ring; this comparatively simple fact is proved in \
section \[OpenCurlyDoubleQuote]Fields Constitute Reduction Rings\
\[CloseCurlyDoubleQuote]."
}], "Text",
 CellChangeTimes->{{3.6609052906256733`*^9, 3.660905303792097*^9}, {
  3.6609053828998356`*^9, 3.660905486047217*^9}},
 CellID->134032288],

Cell[TextData[{
 "However, one comment on fields and reduction rings still is in order: \
According to the original development of reduction ring theory in Buchberger\
\[CloseCurlyQuote]s paper from 1983, and also in the various papers by \
Stifter, fields actually do ",
 StyleBox["not",
  FontSlant->"Italic"],
 " constitute algorithmic reduction rings, because in an infinite field any \
two non-zero elements necessarily have infinitely many minimal non-trivial \
common reducibles (namely all non-zero field elements), and hence there \
cannot exist a function that returns the tuple of all of them.\nThe solution \
to this problem was to introduce an equivalence relation in reduction rings \
such that it suffices to consider only ",
 StyleBox["one",
  FontSlant->"Italic"],
 " representative of each equivalence class of minimal non-trivial common \
reducibles, both in Buchberger\[CloseCurlyQuote]s criterion and in Buchberger\
\[CloseCurlyQuote]s algorithm. The equivalence relation that was found to be \
suitable is admittedly very strong, such that apart from fields in any other \
known \[OpenCurlyDoubleQuote]elementary\[CloseCurlyDoubleQuote] reduction \
ring it collapses to the equality relation, but it solves the problem for \
fields, and this is all that was asked for. In a field, there are in total \
only two equivalence classes: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"{", "0", "}"}], TraditionalForm]]],
 " and the set of all non-zero elements.\nIn polynomial reduction rings, two \
monomials in the same power-product are equivalent iff their coefficients are \
equivalent in the underlying reduction ring \[Dash] this means that if the \
coefficient domain is a field, any two non-zero monomials in the same \
power-product are equivalent."
}], "Text",
 CellChangeTimes->{{3.6609054936600304`*^9, 3.6609060651020346`*^9}, {
  3.660906095802888*^9, 3.660906123602137*^9}},
 CellID->930305492]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Packages", "Subsection",
 CellChangeTimes->{{3.6554648625324345`*^9, 3.6554648653872395`*^9}},
 CellID->52797998],

Cell[BoxData[
 RowBox[{"$PackageDirectory", "=", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{
    "$RootDirectory", ",", "\"\<scratch\>\"", ",", "\"\<amaletzk\>\"", ",", 
     "\"\<Theorema Environment Data\>\""}], "}"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.666438783233775*^9, 3.666438821083324*^9}},
 CellID->498486718],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<LogicSets.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554648672904434`*^9, 3.6554648811744676`*^9}, {
   3.6622888598566933`*^9, 3.662288892694751*^9}, 3.666439692303625*^9},
 CellID->59508975],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<AlgebraicStructures.m\>\""}], 
    "}"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6554648672904434`*^9, 3.6554648811744676`*^9}, {
   3.6622888598566933`*^9, 3.6622888756907215`*^9}, 3.6664396933352737`*^9},
 CellID->75555280],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<ReductionRings.m\>\""}], "}"}], 
   "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6554648672904434`*^9, 3.6554648811744676`*^9}, {
   3.6622888598566933`*^9, 3.6622888756907215`*^9}, {3.6622950274141893`*^9, 
   3.6622950299569936`*^9}, 3.666439694239731*^9},
 CellID->634019458],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6554648672904434`*^9, 3.6554648811744676`*^9}},
 CellID->408174415],

Cell[BoxData[
 RowBox[{"MakeTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6554648744508557`*^9, 3.6554648845596733`*^9}},
 CellID->251583237]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Analysis", "Subsection",
 CellChangeTimes->{{3.640668830437483*^9, 3.640668834318727*^9}},
 CellID->1386333768],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"AnalyzeProofs", "[", 
   RowBox[{"{", 
    RowBox[{
    "\"\<Theory\>\"", ",", 
     "\"\</home/amaletzk/Theorema Notebooks/ElementaryTheories\>\""}], "}"}], 
   "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.6406688560104437`*^9, 3.640668867513785*^9}, 
   3.64067664060292*^9, 3.644222019629964*^9},
 CellID->1039312110],

Cell[CellGroupData[{

Cell[BoxData["\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/Reduction Ring Theory/Polynomials\\\" (1 of 3)\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQBgC8GYnL
  "],
 CellID->1754373566],

Cell[BoxData["\<\"Reading proof 1 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyC9U8PG0g1IAwC334k1
  "],
 CellID->1029482356],

Cell[BoxData["\<\"Reading proof 51 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyCta+Jn6QakAbQ/iMc=
  "],
 CellID->676061299],

Cell[BoxData["\<\"Reading proof 101 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyCtL5pk6QakAbPuiL4=
  "],
 CellID->1850448490],

Cell[BoxData["\<\"Reading proof 151 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyDtw1Fm6QakAbTziOI=
  "],
 CellID->70235299],

Cell[BoxData["\<\"Reading proof 201 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyBtNrHL0g1IAwC4eolp
  "],
 CellID->426922802],

Cell[BoxData["\<\"Reading proof 251 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyD9WXuhpRuQBgC8IonX
  "],
 CellID->1237071357],

Cell[BoxData["\<\"Reading proof 301 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyDtsH+zpRuQBgC7AonK
  "],
 CellID->1870636140],

Cell[BoxData["\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/Reduction Ring Theory/ReductionRings\\\" (2 of 3)\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyAddGObpRuQBgC8U4n4
  "],
 CellID->802449962],

Cell[BoxData["\<\"Reading proof 1 of 132 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB91nq7pRuQBgC75onX
  "],
 CellID->1864026098],

Cell[BoxData["\<\"Reading proof 51 of 132 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyC97/YFSzcgDQDAZIqB
  "],
 CellID->374700447],

Cell[BoxData["\<\"Reading proof 101 of 132 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9QfqdpRuQBgC8aInx
  "],
 CellID->808617339],

Cell[BoxData["\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories\\\" (3 of 3)\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9s+qfpRuQBgC/qYpp
  "],
 CellID->39084793],

Cell[BoxData["\<\"Reading proof 1 of 122 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyBteI/Ryg1IAwC3PIkp
  "],
 CellID->1889518183],

Cell[BoxData["\<\"Reading proof 51 of 122 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB95riQlRuQBgC72Ym+
  "],
 CellID->2026876725],

Cell[BoxData["\<\"Reading proof 101 of 122 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyD9YIm6lRuQBgC8AonE
  "],
 CellID->349652243],

Cell[BoxData["\<\"Creating forward rules ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyC9RsbAyg1IAwC24IkR
  "],
 CellID->1812907063],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyC94Z+PlRuQBgC91ooT
  "],
 CellID->1437928279]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"AnalyzeProofs", "[", "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.652511268886958*^9, 3.652511283482717*^9}},
 CellID->1990427608],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Fields\\\" (\"\>", "\[InvisibleSpace]", 
   "1", "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"1\"\>",
    "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Fields\" (", 1, " of ", "1", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.65251128551866*^9, 3.6551281230619783`*^9},
 CellID->74143475],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"36\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "36"],
  Editable->False]], "Print",
 CellChangeTimes->{3.65251128551866*^9, 3.6551281230619783`*^9},
 CellID->130980690],

Cell[BoxData["\<\"Creating forward rules:\"\>"], "Print",
 CellChangeTimes->{3.65251128551866*^9, 3.6551281266031847`*^9},
 CellID->631158878],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.65251128551866*^9, 3.655128126618785*^9},
 CellID->249417802],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "131"}],
  SequenceForm["Creating rule ", 1, " of ", 131],
  Editable->False]], "Print",
 CellChangeTimes->{3.65251128551866*^9, 3.655128126634385*^9},
 CellID->498183016],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "131"}],
  SequenceForm["Creating rule ", 51, " of ", 131],
  Editable->False]], "Print",
 CellChangeTimes->{3.65251128551866*^9, 3.655128126665585*^9},
 CellID->710555543],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "131"}],
  SequenceForm["Creating rule ", 101, " of ", 131],
  Editable->False]], "Print",
 CellChangeTimes->{3.65251128551866*^9, 3.655128126665585*^9},
 CellID->17027108],

Cell[BoxData["\<\"Creating thinned data ...\"\>"], "Print",
 CellChangeTimes->{3.65251128551866*^9, 3.655128126681185*^9},
 CellID->59531876],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->{3.65251128551866*^9, 3.655128126696785*^9},
 CellID->54622194]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"ComputeFormulaPositions", "[", "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.6406695270485573`*^9, 3.6406695415951242`*^9}, {
   3.64067664934828*^9, 3.640676663567005*^9}, 3.6412902845300865`*^9, 
   3.641290346822996*^9, {3.6442219079512787`*^9, 3.644221909273131*^9}},
 CellID->19428626],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CheckIntegrity", "[", 
  RowBox[{"\"\<CheckPositions\>\"", "\[Rule]", 
   RowBox[{"{", 
    RowBox[{"FileNameJoin", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"CurrentValue", "[", "\"\<NotebookDirectory\>\"", "]"}], ",", 
       RowBox[{"CurrentValue", "[", "\"\<NotebookFileName\>\"", "]"}]}], 
      "}"}], "]"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.64067003109768*^9, 3.6406700374311132`*^9}, {
   3.6406797903560762`*^9, 3.640679792950843*^9}, 3.6442220291527653`*^9},
 CellID->1909803754],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.640670040858255*^9, 3.640670099679185*^9, 3.640670397721546*^9, 
   3.640679793300535*^9, 3.6406798406436234`*^9, 3.640679891162312*^9, 
   3.640679971009297*^9, 3.64068009182163*^9, 3.640680173340745*^9, 
   3.641037395456129*^9, 3.641116403388618*^9, 3.641286881014108*^9, 
   3.642225466578094*^9, 3.642234672070362*^9, 3.642240151894621*^9, 
   3.6423063269273643`*^9, {3.642308832539488*^9, 3.642308857563932*^9}, 
   3.642321396697977*^9, 3.6423285030212173`*^9, 3.642493582768566*^9, 
   3.642763270724422*^9, 3.642834604308413*^9, 3.643428832901271*^9, 
   3.6434441292624083`*^9, 3.643532851309215*^9, 3.644918543299676*^9, 
   3.645518304732339*^9},
 CellID->636966437]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Count", "[", 
  RowBox[{
   RowBox[{"NotebookGet", "[", "]"}], ",", 
   RowBox[{"Cell", "[", 
    RowBox[{"_", ",", "\"\<FormalTextInputFormula\>\"", ",", "___"}], "]"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{3.652511315573008*^9},
 CellID->664825817],

Cell[BoxData["57"], "Output",
 CellChangeTimes->{3.652511317366551*^9},
 CellID->1831977478]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Auxiliary Functions", "Subsection",
 CellChangeTimes->{{3.640668579267561*^9, 3.640668583347287*^9}},
 CellID->1454182661],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "LoadPackage", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", "name_String", "]"}], ":=", 
   RowBox[{"LoadPackage", "[", 
    RowBox[{"{", "name", "}"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", 
    RowBox[{"l", ":", 
     RowBox[{"{", "___String", "}"}]}], "]"}], ":=", 
   RowBox[{"Scan", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{"FileNameJoin", "[", 
        RowBox[{"{", 
         RowBox[{"$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
          RowBox[{"#", "<>", "\"\<.m\>\""}]}], "}"}], "]"}], "]"}], "&"}], 
     ",", "l"}], "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640668603629292*^9, 3.640668727460284*^9}, 
   3.641286816086794*^9, 3.6434459961973133`*^9, {3.655128063515674*^9, 
   3.6551280791323013`*^9}, {3.666439700658169*^9, 3.666439708330182*^9}},
 CellID->2074904605],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "FindFormula", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"FindFormula", "[", "label_String", "]"}], ":=", 
   RowBox[{"NotebookLocate", "[", "label", "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640677756847624*^9, 3.6406777896243896`*^9}, 
   3.640677822869519*^9, {3.640677854902379*^9, 3.6406778579321823`*^9}},
 CellID->1980565193],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<ProofPalette\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.640668742006607*^9, 3.640668751417616*^9}},
 CellID->444074165],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<TheoryTools/TheoryTools\>\"", 
  "]"}]], "Input",
 CellChangeTimes->{{3.6551280894439197`*^9, 3.655128093515527*^9}, {
   3.6618523961991096`*^9, 3.6618523989447145`*^9}, 3.666439712694866*^9},
 CellID->438171343]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Definitions", "Section",
 CellChangeTimes->{{3.645758331159669*^9, 3.645758333072812*^9}},
 CellID->1664317174],

Cell[CellGroupData[{

Cell["Informal Explanation", "Subsection",
 CellChangeTimes->{{3.590834066367938*^9, 3.590834069395162*^9}},
 CellID->1100887197],

Cell[TextData[{
 "The \[OpenCurlyDoubleQuote]ReductionField\[CloseCurlyDoubleQuote] functor \
takes a domain ",
 Cell[BoxData[
  FormBox["K", TraditionalForm]]],
 ", which should be a field, as input and returns an extension of ",
 Cell[BoxData[
  FormBox["K", TraditionalForm]]],
 " where all operations that are required for being a reduction ring are \
defined. Turning a field into a reduction ring is, in fact, a rather trivial \
and straightforward task, so there is not very much to explain."
}], "Text",
 CellChangeTimes->{{3.590836862330344*^9, 3.590836990989859*^9}},
 CellID->899326618],

Cell[TextData[{
 StyleBox["Remark:",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 "\nWe cannot define \[Succeeds] as ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Succeeds]", "1", "\[Succeeds]", "0"}], TraditionalForm]]],
 " because of axiom (R7): If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["c", "1"], ",", 
    SubscriptBox["c", "2"], ",", 
    RowBox[{"a", "\[NotEqual]", "0"}]}], TraditionalForm]]],
 " are such that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     SubscriptBox["c", "1"], ",", 
     SubscriptBox["c", "2"], ",", "a", ",", "1", ",", "1"}], "]"}], 
   TraditionalForm]]],
 " holds, then there must exist ",
 Cell[BoxData[
  FormBox["m", TraditionalForm]]],
 " with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"1", "\[Succeeds]", 
     RowBox[{"1", "+", "c"}]}], "\[Implies]", 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"a", "+", 
      RowBox[{"m", " ", "c"}]}]}]}], TraditionalForm]]],
 ", because with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Succeeds]", "1", "\[Succeeds]", "0"}], TraditionalForm]]],
 " the only minimal non-trivial common reducible of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "2"], TraditionalForm]]],
 " is 1. However, an ",
 Cell[BoxData[
  FormBox["m", TraditionalForm]]],
 " with this property necessarily must be either ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "-", "1"}], TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"1", "\[Succeeds]", 
      RowBox[{"1", "+", "c"}]}], "\[DoubleLeftRightArrow]", "c"}], "=", 
    RowBox[{"-", "1"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"a", "-", "m"}]}], "\[DoubleLeftRightArrow]", 
    RowBox[{"(", 
     RowBox[{"m", "=", 
      RowBox[{
       RowBox[{"a", "\[Or]", "m"}], "=", 
       RowBox[{"a", "-", "1"}]}]}], ")"}]}], TraditionalForm]]],
 ", assuming that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "1"}], TraditionalForm]]],
 ". Still, ",
 Cell[BoxData[
  FormBox["m", TraditionalForm]]],
 " also needs to satisfy a second constraint: Whenever ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", 
    SubscriptBox["\[Rule]", "c"], "d"}], TraditionalForm]]],
 ", then ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m", " ", "b"}], 
    SubscriptBox["\[LeftRightArrow]", "c"], 
    RowBox[{"m", " ", "d"}]}], TraditionalForm]]],
 ". In particular, ",
 Cell[BoxData[
  FormBox["d", TraditionalForm]]],
 " could be 1, meaning that a necessary condition for the RHS to hold is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m", " ", "b"}], "=", 
    RowBox[{
     RowBox[{"1", "\[Or]", 
      RowBox[{"m", " ", "b"}]}], "=", 
     RowBox[{
      RowBox[{"0", "\[Or]", "m"}], "=", 
      RowBox[{
       RowBox[{"1", "\[Or]", "m"}], "=", "0"}]}]}]}], TraditionalForm]]],
 ". The only feasible choices ",
 StyleBox["independent",
  FontWeight->"Bold"],
 " of ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " apparently are ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "=", "0"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "=", "1"}], TraditionalForm]]],
 ", both violating the constraint ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "=", 
    RowBox[{
     RowBox[{"a", "\[Or]", "m"}], "=", 
     RowBox[{"a", "-", "1"}]}]}], TraditionalForm]]],
 " for all ",
 Cell[BoxData[
  FormBox[
   RowBox[{"1", "\[NotEqual]", "a", "\[NotEqual]", 
    RowBox[{"1", "+", "1"}]}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.650016266691139*^9, 3.650016457208703*^9}, {
  3.650016568458042*^9, 3.650016907507285*^9}, {3.650016967990028*^9, 
  3.650017166010673*^9}, {3.6500172084900837`*^9, 3.650017302376658*^9}, {
  3.650017366592079*^9, 3.6500173992629967`*^9}, {3.6500174531227713`*^9, 
  3.650017462199832*^9}, {3.6500175061566267`*^9, 3.650017518418559*^9}, {
  3.6519177030125313`*^9, 3.651917705482781*^9}},
 CellID->908816666]
}, Closed]],

Cell[CellGroupData[{

Cell["Definitions", "Subsection",
 CellChangeTimes->{{3.590834077485271*^9, 3.590834079203567*^9}},
 CellID->1612768911],

Cell["", "OpenEnvironment",
 CellID->1395386996],

Cell[CellGroupData[{

Cell["Definition: Reduction Field", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.590834086551167*^9, 3.590834097613612*^9}},
 CellID->846251564],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "K"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb", 
          991632779}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590834118789295*^9, 3.59083412060068*^9}, 
   3.590834192543613*^9},
 CellTags->{
  "ID:991632779", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->991632779],

Cell[BoxData[
 RowBox[{
  RowBox[{"ReductionField", "[", "K", "]"}], ":=", 
  UnderscriptBox["\[CapitalDelta]", 
   RowBox[{"N", "\[Superset]", "K"}]]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb", 
          13541845}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590834131298801*^9, 3.590834150894432*^9}, {
  3.590834195265446*^9, 3.590834198023637*^9}, {3.607338570126086*^9, 
  3.607338571042476*^9}},
 CellTags->{
  "ID:13541845", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->13541845],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"x", ",", "y", ",", "m", ",", "i", ",", "j"}]]], "GlobalDeclaration",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb", 
          1796175771}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590834155631195*^9, 3.590834162038236*^9}, {
   3.590835653716524*^9, 3.590835654761516*^9}, 3.590835736931481*^9, {
   3.590835857534343*^9, 3.590835857719519*^9}, {3.594035781653264*^9, 
   3.594035796088778*^9}, {3.595057237860121*^9, 3.595057239074228*^9}, {
   3.608549656805819*^9, 3.608549657027662*^9}},
 CellTags->{
  "ID:1796175771", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->1796175771],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "N"], 
   RowBox[{"[", "x", "]"}]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[Element]", "K"], 
   RowBox[{"[", "x", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field \[Element])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:204732646", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590834174903511*^9, 3.590834204167776*^9}, {
  3.594035769076688*^9, 3.594035802319771*^9}, {3.595057241576427*^9, 
  3.59505724279177*^9}, {3.607338573441447*^9, 3.607338574438855*^9}},
 CellTags->{
  "ID:204732646", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field \[Element]"},
 CellID->204732646],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["0", "N"], ":=", 
  UnderscriptBox["0", "K"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:125417415", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590834212291209*^9, 3.590834223184213*^9}, {
  3.60733857797979*^9, 3.607338578718776*^9}},
 CellTags->{
  "ID:125417415", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field 0"},
 CellID->125417415],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "N"], ":=", 
  UnderscriptBox["1", "K"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1874597366", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590834226564992*^9, 3.590834227719658*^9}, {
  3.590834301140429*^9, 3.590834306105637*^9}, {3.607338579605127*^9, 
  3.607338580344264*^9}},
 CellTags->{
  "ID:1874597366", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field 1"},
 CellID->1874597366],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["+", "N"], "y"}], ":=", 
  RowBox[{"x", 
   UnderscriptBox["+", "K"], "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field +)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:791499509", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590834230396772*^9, 3.590834231953101*^9}, {
  3.590834309451824*^9, 3.590834319179139*^9}, {3.607338581083604*^9, 
  3.607338582025319*^9}},
 CellTags->{
  "ID:791499509", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field +"},
 CellID->791499509],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["-", "N"], "x"}], ":=", 
  RowBox[{
   UnderscriptBox["-", "K"], "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field - unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1535890097", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590834236952189*^9, 3.590834237913265*^9}, {
  3.590834322747439*^9, 3.590834330094577*^9}, {3.607338582911779*^9, 
  3.607338583946119*^9}},
 CellTags->{
  "ID:1535890097", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field - unary"},
 CellID->1535890097],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["-", "N"], "y"}], ":=", 
  RowBox[{"x", 
   UnderscriptBox["-", "K"], "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field -)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:328879666", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59083424066084*^9, 3.590834241411703*^9}, {
  3.59083433417483*^9, 3.590834342143107*^9}, {3.607338586012732*^9, 
  3.607338586899378*^9}},
 CellTags->{
  "ID:328879666", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field -"},
 CellID->328879666],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["*", "N"], "y"}], ":=", 
  RowBox[{"x", 
   UnderscriptBox["*", "K"], "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1643707054", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590834244357064*^9, 3.590834245168309*^9}, {
  3.590834344423505*^9, 3.59083435255525*^9}, {3.607338587712107*^9, 
  3.607338588561945*^9}},
 CellTags->{
  "ID:1643707054", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field *"},
 CellID->1643707054],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["^", "N"], 
   RowBox[{"(", 
    RowBox[{"-", "1"}], ")"}]}], ":=", 
  RowBox[{"x", 
   UnderscriptBox["^", "K"], 
   RowBox[{"(", 
    RowBox[{"-", "1"}], ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field inverse)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1452171966", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590834251084514*^9, 3.5908342532792*^9}, {
  3.590834356765349*^9, 3.590834369007526*^9}, {3.593492306648311*^9, 
  3.593492316096326*^9}, {3.607338590776666*^9, 3.607338623494469*^9}},
 CellTags->{
  "ID:1452171966", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field inverse"},
 CellID->1452171966],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["/", "N"], "y"}], ":=", 
  RowBox[{"x", 
   UnderscriptBox["/", "K"], "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field /)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1982468810", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590834256284003*^9, 3.590834257314351*^9}, {
  3.59083437128784*^9, 3.590834379706184*^9}, {3.607338594615625*^9, 
  3.607338595649371*^9}},
 CellTags->{
  "ID:1982468810", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field /"},
 CellID->1982468810],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["\[Succeeds]", "N"], "y"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    RowBox[{
     RowBox[{"y", "\[Equal]", 
      UnderscriptBox["0", "N"]}], "\[And]", 
     RowBox[{"x", "\[NotEqual]", 
      UnderscriptBox["0", "N"]}]}],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field \[Succeeds])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1869336472", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590835167630392*^9, 3.590835303577233*^9}, 
   3.590915643714789*^9, {3.591944159719538*^9, 3.591944186027738*^9}, {
   3.607338632531077*^9, 3.607338674965283*^9}, {3.608006973949766*^9, 
   3.608006982966725*^9}},
 CellTags->{
  "ID:1869336472", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field \[Succeeds]"},
 CellID->1869336472],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["multN", "N"], "[", "x", "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {"0", "\[DoubleLeftArrow]", 
      RowBox[{"x", "\[Equal]", 
       UnderscriptBox["0", "N"]}]},
     {"1", "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field multN)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1352206528", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590835591175243*^9, 3.590835629628766*^9}, {
  3.607338687287591*^9, 3.607338693271454*^9}, {3.631855526610804*^9, 
  3.631855533838937*^9}, {3.6318565686764917`*^9, 3.6318565728879557`*^9}},
 CellTags->{
  "ID:1352206528", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field multN"},
 CellID->1352206528],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"m", ",", "x"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"m", ",", "x", ",", "1"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field mult)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:305610078", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590835801303161*^9, 3.5908358098978*^9}, {
  3.607338696112415*^9, 3.607338696998873*^9}},
 CellTags->{
  "ID:305610078", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field mult"},
 CellID->305610078],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"m", ",", "x", ",", "i"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", "\[NotEqual]", 
     UnderscriptBox["0", "N"]}], "\[And]", 
    RowBox[{"x", "\[NotEqual]", 
     UnderscriptBox["0", "N"]}], "\[And]", 
    RowBox[{"i", "\[Equal]", "1"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field mult 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1075639810", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590835813629157*^9, 3.590835827784371*^9}, {
  3.590835863338486*^9, 3.590835863671684*^9}, {3.607338698660938*^9, 
  3.607338744157862*^9}},
 CellTags->{
  "ID:1075639810", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field mult 1"},
 CellID->1075639810],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdm", "N"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      UnderscriptBox["0", "N"], "\[DoubleLeftArrow]", 
      RowBox[{"y", "\[Equal]", 
       UnderscriptBox["0", "N"]}]},
     {
      RowBox[{"x", 
       UnderscriptBox["/", "N"], "y"}], "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field rdm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1151861642", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590835919678329*^9, 3.590835981499283*^9}, {
  3.607338755191405*^9, 3.607338763910882*^9}},
 CellTags->{
  "ID:1151861642", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field rdm"},
 CellID->1151861642],

Cell[TextData[{
 "In fields we have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isMNTCR", "[", 
     RowBox[{"x", ",", "y", ",", "a", ",", "1", ",", "1"}], "]"}], 
    "\[DoubleLeftRightArrow]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"x", "\[NotEqual]", "0"}], "\[And]", 
      RowBox[{"y", "\[NotEqual]", "0"}], "\[And]", 
      RowBox[{"a", "\[NotEqual]", "0"}]}], ")"}]}], TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.650017866132967*^9, 3.650017939180152*^9}},
 CellID->1667267410],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", 
   RowBox[{"x", ",", "y", ",", "i", ",", "j"}], "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
      "\[DoubleLeftArrow]", 
      RowBox[{
       RowBox[{"x", "\[Equal]", 
        UnderscriptBox["0", "N"]}], "\[Or]", 
       RowBox[{"y", "\[Equal]", 
        UnderscriptBox["0", "N"]}], "\[Or]", 
       RowBox[{"x", "\[Equal]", "y"}]}]},
     {
      RowBox[{"\[LeftAngleBracket]", 
       UnderscriptBox["1", "N"], "\[RightAngleBracket]"}], 
      "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field lcrd)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2067272590", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590836002915568*^9, 3.590836028325472*^9}, {
   3.59083610249812*^9, 3.590836155969228*^9}, {3.593770688493574*^9, 
   3.593770707590914*^9}, {3.59445899725414*^9, 3.594459014168063*^9}, {
   3.607338767861363*^9, 3.607338814182342*^9}, {3.607338857153561*^9, 
   3.607338862313775*^9}, 3.608022550933746*^9, {3.608549661642466*^9, 
   3.608549689909762*^9}, {3.610097371719789*^9, 3.610097389949559*^9}, {
   3.651917732661746*^9, 3.6519177909181137`*^9}},
 CellTags->{
  "ID:2067272590", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field lcrd"},
 CellID->2067272590],

Cell[TextData[{
 "In fields we have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isMNTCR", "[", 
     RowBox[{"x", ",", "a"}], "]"}], "\[DoubleLeftRightArrow]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"x", "\[NotEqual]", "0"}], "\[And]", 
      RowBox[{"a", "\[NotEqual]", "0"}]}], ")"}]}], TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.650017918030673*^9, 3.6500179400979433`*^9}},
 CellID->2143937706],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", "x", "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
      "\[DoubleLeftArrow]", 
      RowBox[{"x", "\[Equal]", 
       UnderscriptBox["0", "N"]}]},
     {
      RowBox[{"\[LeftAngleBracket]", 
       UnderscriptBox["1", "N"], "\[RightAngleBracket]"}], 
      "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field lcrd unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1497531972", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594457500533925*^9, 3.594457509444688*^9}, {
  3.594457567269925*^9, 3.594457569518091*^9}, {3.607338827030346*^9, 
  3.607338828026788*^9}, {3.608022568933454*^9, 3.608022571780412*^9}, {
  3.610097409019419*^9, 3.610097414222321*^9}, {3.651917740120343*^9, 
  3.651917762864506*^9}, {3.651917793226944*^9, 3.6519177961295557`*^9}},
 CellTags->{
  "ID:1497531972", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field lcrd unary"},
 CellID->1497531972],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpm", "N"], "[", 
   RowBox[{"x", ",", "y", ",", "m", ",", "i", ",", "j"}], "]"}], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["/", "N"], "x"}], ",", 
    RowBox[{"m", 
     UnderscriptBox["/", "N"], "y"}]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(field cpm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:529122150", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594458848634239*^9, 3.594458873410174*^9}, {
  3.595057257894746*^9, 3.595057262115848*^9}, {3.60733882950469*^9, 
  3.607338830759937*^9}, {3.610097420566798*^9, 3.610097426360496*^9}, {
  3.651917765106505*^9, 3.65191776782506*^9}, {3.6519177988115387`*^9, 
  3.651917800936863*^9}},
 CellTags->{
  "ID:529122150", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field cpm"},
 CellID->529122150],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1031992887]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->879906266]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Tests", "Section",
 CellChangeTimes->{{3.6457585997377357`*^9, 3.6457586005547743`*^9}},
 CellID->1178317413],

Cell["", "OpenEnvironment",
 CellID->1922534331],

Cell[CellGroupData[{

Cell["\<\
Definition: Reduction Field of \[DoubleStruckCapitalQ]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645758617538411*^9, 3.645758623366171*^9}, {
  3.645779863236308*^9, 3.645779869581542*^9}},
 CellID->1594116140],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalQ]", ":=", 
  RowBox[{
  "ReductionField", "[", "\[DoubleStruckCapitalQ]", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalQ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1788558801", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645758629365995*^9, 3.64575865814565*^9}, {
  3.645779856209463*^9, 3.64577985761366*^9}},
 CellTags->{
  "ID:1788558801", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "\[ScriptCapitalQ]"},
 CellID->1788558801],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1329207975]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->14854175],

Cell[CellGroupData[{

Cell["rdm", "Subsection",
 CellChangeTimes->{{3.645758713081973*^9, 3.6457587133789263`*^9}},
 CellID->164644379],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalQ]"], "[", 
  RowBox[{
   FractionBox["7", "3"], ",", 
   RowBox[{"-", "5"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.645758885720047*^9, 3.6457589131514807`*^9}, 
   3.6457802116023293`*^9},
 CellID->1133492443],

Cell[BoxData[
 FormBox[
  RowBox[{"-", 
   FractionBox["7", "15"]}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{{3.645758914272398*^9, 3.645758923490251*^9}},
 CellID->868504151]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1133492443"]]]]}, {None, None}},
 CellChangeTimes->{3.645758923885021*^9},
 CellID->1196757825],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1133492443-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645758924002537*^9},
 CellID->1859826640]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalQ]"], "[", 
  RowBox[{"0", ",", "2"}], "]"}]], "Computation",
 CellChangeTimes->{{3.64575893705577*^9, 3.645758940364196*^9}, 
   3.645780213960229*^9},
 CellID->1608921888],

Cell[BoxData[
 FormBox["0", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.6457589412216883`*^9},
 CellID->140517660]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1608921888"]]]]}, {None, None}},
 CellChangeTimes->{3.64575894164361*^9},
 CellID->1855696697],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1608921888-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6457589417602377`*^9},
 CellID->2086722562]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalQ]"], "[", 
  RowBox[{"2", ",", "0"}], "]"}]], "Computation",
 CellChangeTimes->{{3.6457589452617598`*^9, 3.6457589462918577`*^9}, 
   3.6457802155075808`*^9},
 CellID->1883766832],

Cell[BoxData[
 FormBox["0", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645758947302628*^9},
 CellID->947393027]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1883766832"]]]]}, {None, None}},
 CellChangeTimes->{3.645758947732892*^9},
 CellID->1118382957],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1883766832-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645758947851665*^9},
 CellID->1301382415]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalQ]"], "[", 
  RowBox[{
   FractionBox["3", "2"], ",", 
   FractionBox["3", "2"]}], "]"}]], "Computation",
 CellChangeTimes->{{3.645759044302833*^9, 3.645759052033946*^9}, 
   3.6457802169100924`*^9},
 CellID->536728615],

Cell[BoxData[
 FormBox["1", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645759053115004*^9},
 CellID->2053160890]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c536728615"]]]]}, {None, None}},
 CellChangeTimes->{3.645759053752342*^9},
 CellID->1664406102],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c536728615-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645759053857319*^9},
 CellID->1128431625]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["lcrd", "Subsection",
 CellChangeTimes->{{3.645758717754362*^9, 3.645758718310535*^9}},
 CellID->1583099554],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalQ]"], "[", 
  RowBox[{
   FractionBox["7", "3"], ",", 
   RowBox[{"-", "5"}], ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.6457590629437513`*^9, 3.6457590789192057`*^9}, 
   3.645780220040036*^9},
 CellID->990169324],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "1", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{{3.64575906410443*^9, 3.645759079636038*^9}},
 CellID->1941861415]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c990169324"]]]]}, {None, None}},
 CellChangeTimes->{3.645759080033092*^9},
 CellID->485343236],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c990169324-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645759080154263*^9},
 CellID->1755171163]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalQ]"], "[", 
  RowBox[{"0", ",", "2", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.645759087568233*^9, 3.645759091219722*^9}, 
   3.645780221411625*^9},
 CellID->1956353311],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.6457590917485447`*^9},
 CellID->1461111709]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1956353311"]]]]}, {None, None}},
 CellChangeTimes->{3.645759092156543*^9},
 CellID->733820202],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1956353311-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645759092269422*^9},
 CellID->1067638783]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalQ]"], "[", 
  RowBox[{"2", ",", "0", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.645759101376976*^9, 3.645759101711063*^9}, 
   3.645780223005354*^9},
 CellID->1600026178],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645759102827608*^9},
 CellID->780050836]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1600026178"]]]]}, {None, None}},
 CellChangeTimes->{3.645759103292118*^9},
 CellID->1554263949],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1600026178-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645759103407403*^9},
 CellID->1663999394]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalQ]"], "[", 
  RowBox[{
   FractionBox["3", "2"], ",", 
   FractionBox["3", "2"], ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.6457591090667143`*^9, 3.6457591128677893`*^9}, 
   3.645780224443664*^9},
 CellID->1090208346],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645759113653558*^9},
 CellID->12333322]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1090208346"]]]]}, {None, None}},
 CellChangeTimes->{3.645759114082632*^9},
 CellID->128058836],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1090208346-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6457591142038937`*^9},
 CellID->1864007722]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalQ]"], "[", 
  FractionBox["7", "3"], "]"}]], "Computation",
 CellChangeTimes->{{3.6457591243098516`*^9, 3.645759127403528*^9}, 
   3.645780226469433*^9},
 CellID->1440655820],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "1", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645759127712327*^9},
 CellID->81088794]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1440655820"]]]]}, {None, None}},
 CellChangeTimes->{3.645759128139338*^9},
 CellID->246761544],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1440655820-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6457591282552557`*^9},
 CellID->192133376]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalQ]"], "[", "0", "]"}]], "Computation",\

 CellChangeTimes->{3.645759134297615*^9, 3.645780227760518*^9},
 CellID->82552198],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645759134752915*^9},
 CellID->1574628483]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c82552198"]]]]}, {None, None}},
 CellChangeTimes->{3.645759135181851*^9},
 CellID->995526489],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c82552198-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6457591352988787`*^9},
 CellID->764313443]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["cpm", "Subsection",
 CellChangeTimes->{{3.645758722465295*^9, 3.645758722872635*^9}},
 CellID->1292034652],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalQ]"], "[", 
  RowBox[{
   FractionBox["7", "3"], ",", 
   RowBox[{"-", "5"}], ",", "1", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.6457591834924192`*^9, 3.6457591862926483`*^9}, 
   3.6457802314426117`*^9},
 CellID->613726844],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    FractionBox["3", "7"], ",", 
    RowBox[{"-", 
     FractionBox["1", "5"]}]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645759187085305*^9},
 CellID->1299536985]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c613726844"]]]]}, {None, None}},
 CellChangeTimes->{3.645759187463315*^9},
 CellID->1784347710],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c613726844-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6457591875824213`*^9},
 CellID->728776616]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Theorem: Fields Constitute Reduction Rings (\[Checkmark])\
\>", "Section",
 CellChangeTimes->{{3.590836358167693*^9, 3.59083636134455*^9}, {
  3.645758349352075*^9, 3.645758360580639*^9}, {3.652511209482236*^9, 
  3.652511215617371*^9}},
 CellID->2031767909],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isField", "[", "K", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Fields.nb", 1470999011}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6519170864903173`*^9, 3.651917090260399*^9}},
 CellTags->{
  "ID:1470999011", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->1470999011],

Cell["", "Text",
 CellID->1259024642],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"\[ScriptCapitalK]", "=", 
   RowBox[{"ReductionField", "[", "K", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Fields.nb", 853918182}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913233633284*^9, 3.590913243354645*^9}, {
  3.651917113428156*^9, 3.651917114646727*^9}},
 CellTags->{
  "ID:853918182", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->853918182],

Cell[CellGroupData[{

Cell["Fields (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.651917943724175*^9, 3.651917944627728*^9}, {
  3.6519190265258837`*^9, 3.6519190301919928`*^9}},
 CellID->1912339895],

Cell["", "OpenEnvironment",
 CellID->1605366551],

Cell[CellGroupData[{

Cell["Theorem: Reduction Fields are Fields", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6519179519347067`*^9, 3.6519179594125147`*^9}},
 CellID->704319825],

Cell[BoxData[
 RowBox[{"isCommRing1", "[", "\[ScriptCapitalK]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCommRing1 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:82222568", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590912748560351*^9, 3.590912762686562*^9}, {
   3.590912818999735*^9, 3.59091282371304*^9}, {3.59091321630479*^9, 
   3.590913218325221*^9}, {3.590913251386957*^9, 3.590913251864804*^9}, 
   3.6519171231119947`*^9},
 CellTags->{
  "ID:82222568", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "isCommRing1 \[ScriptCapitalK]"},
 CellID->82222568],

Cell[BoxData[
 RowBox[{"isField", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isField \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1537851509", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.651917966536318*^9, 3.6519179691727247`*^9}},
 CellTags->{
  "ID:1537851509", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "isField \[ScriptCapitalK]"},
 CellID->1537851509],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2009629793]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1116315797],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCommRing1 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCommRing1 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isCommRing1", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isCommRing1", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:82222568"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p82222568-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p82222568-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:82222568", "Proof|ID:82222568-1"},
 CellID->83229303],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p82222568-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->229712744]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isField \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isField \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isField", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isField", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1537851509"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1537851509-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1537851509-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1537851509", "Proof|ID:1537851509-1"},
 CellID->710651429],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1537851509-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->746747984]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Reduction Rings (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.651917182959581*^9, 3.651917185605096*^9}, {
  3.6520820870875673`*^9, 3.652082091257348*^9}},
 CellID->1445748822],

Cell["", "OpenEnvironment",
 CellID->596016869],

Cell[CellGroupData[{

Cell["\<\
Lemma: Alternative Description of Notions in Reduction Fields\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6519216853519907`*^9, 3.651921698000444*^9}},
 CellID->1128420351],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalK]"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "m"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Fields.nb", 2102486692}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.651921703187111*^9, 3.651921726256713*^9}, {
  3.651921769877788*^9, 3.651921771564828*^9}, {3.6519221237063913`*^9, 
  3.651922123928803*^9}},
 CellTags->{
  "ID:2102486692", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->2102486692],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalK]"], "[", 
   RowBox[{"a", ",", "c", ",", "m"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"m", "\[Equal]", 
      RowBox[{"a", 
       UnderscriptBox["/", "\[ScriptCapitalK]"], "c"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:274992099", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.651921759807239*^9, 3.651921805114566*^9}},
 CellTags->{
  "ID:274992099", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "1"},
 CellID->274992099],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalK]"], "[", 
   RowBox[{"a", ",", "c"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:816763103", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.651921729436425*^9, 3.6519217790085497`*^9}},
 CellTags->{
  "ID:816763103", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "2"},
 CellID->816763103],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"m", ",", "c"}]], "\[ScriptCapitalK]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"m", "\[Equal]", 
      RowBox[{"a", 
       UnderscriptBox["/", "\[ScriptCapitalK]"], "c"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", "\[Equal]", 
      UnderscriptBox["0", "\[ScriptCapitalK]"]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:83877220", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.651922127364921*^9, 3.651922174872361*^9}},
 CellTags->{
  "ID:83877220", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "3"},
 CellID->83877220],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"{", "c", "}"}]], "\[ScriptCapitalK]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", "\[Equal]", 
      UnderscriptBox["0", "\[ScriptCapitalK]"]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1728540744", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6519221822481318`*^9, 3.651922187867708*^9}, {
  3.6618423041502624`*^9, 3.661842305008264*^9}},
 CellTags->{
  "ID:1728540744", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "4"},
 CellID->1728540744],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", "j"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Fields.nb", 1733996578}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6520090957222633`*^9, 3.652009097144668*^9}},
 CellTags->{
  "ID:1733996578", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->1733996578],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isCommonReducible", "\[ScriptCapitalK]"], "[", 
   RowBox[{"b", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"b", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{"b", "\[NotEqual]", "c"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:371484949", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.652009067170844*^9, 3.6520091384556017`*^9}},
 CellTags->{
  "ID:371484949", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "5"},
 CellID->371484949],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalK]"], "[", 
   RowBox[{"b", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"b", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{"b", "\[NotEqual]", "c"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:316306510", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.652009774866119*^9, 3.652009802770669*^9}},
 CellTags->{
  "ID:316306510", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "6"},
 CellID->316306510],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalK]"], "[", 
   RowBox[{"b", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalK]"], "[", 
   RowBox[{"b", ",", "c", ",", "a", ",", "i", ",", "j"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(7)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1250046151", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65201038536782*^9, 3.652010393384304*^9}},
 CellTags->{
  "ID:1250046151", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "7"},
 CellID->1250046151],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalK]"], "[", 
   RowBox[{"c", ",", "a"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(8)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:304323379", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.652072382006187*^9, 3.6520723973152103`*^9}},
 CellTags->{
  "ID:304323379", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "8"},
 CellID->304323379],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalK]"], "[", 
   RowBox[{"c", ",", "a"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalK]"], "[", 
   RowBox[{"c", ",", "a"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(9)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1618605896", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6520724048980093`*^9, 3.6520724118959827`*^9}},
 CellTags->{
  "ID:1618605896", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "9"},
 CellID->1618605896],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1503924104]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1855794389],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["/", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], 
                   StyleBox["c", "ExpressionVariable"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["/", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], 
                   StyleBox["c", "ExpressionVariable"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:274992099"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p274992099-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p274992099-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:274992099", "Proof|ID:274992099-1"},
 CellID->1519173141],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p274992099-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1937965720]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:816763103"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p816763103-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p816763103-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:816763103", "Proof|ID:816763103-1"},
 CellID->2071197478],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p816763103-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.652008641246398*^9},
 CellID->1092343128]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["/", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], 
                   StyleBox["c", "ExpressionVariable"]}]}], "\[And]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["/", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], 
                   StyleBox["c", "ExpressionVariable"]}]}], "\[And]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:83877220"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p83877220-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p83877220-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:83877220", "Proof|ID:83877220-1"},
 CellID->1705428850],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p83877220-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1125367110]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["c", "ExpressionVariable"]], 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["c", "ExpressionVariable"]], 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1728540744"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1728540744-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1728540744-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1728540744", "Proof|ID:1728540744-1"},
 CellID->115960158],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1728540744-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1053170671]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["c", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["c", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:371484949"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p371484949-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p371484949-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:371484949", "Proof|ID:371484949-1"},
 CellID->1532696513],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p371484949-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->623568866]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(6)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(6)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["c", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["c", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:316306510"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p316306510-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p316306510-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:316306510", "Proof|ID:316306510-1"},
 CellID->293200241],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p316306510-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.652072893492152*^9, 3.6520728955158033`*^9}, 
   3.652073620951782*^9},
 CellID->1149536317]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(7)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(7)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1250046151"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1250046151-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1250046151-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1250046151", "Proof|ID:1250046151-1"},
 CellID->143418677],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1250046151-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1204687680]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(8)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(8)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:304323379"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p304323379-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p304323379-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:304323379", "Proof|ID:304323379-1"},
 CellID->1825869525],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p304323379-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->403531515]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(9)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(9)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1618605896"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1618605896-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1618605896-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1618605896", "Proof|ID:1618605896-1"},
 CellID->175965684],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1618605896-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1012531353]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1342787779],

Cell[CellGroupData[{

Cell["Lemma: Reduction Fields satisfy Reduction Ring Axioms", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.590912528453717*^9, {3.590912673310073*^9, 3.590912688760049*^9}, {
   3.590912784952359*^9, 3.590912801288868*^9}, {3.651917430980213*^9, 
   3.651917431845427*^9}},
 CellID->109341845],

Cell[BoxData[
 RowBox[{"isPartIrreflOrder", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalK]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalK]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isPartIrreflOrder \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:514433530", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59091655640823*^9, 3.590916578077216*^9}, {
   3.594028456512166*^9, 3.594028465203914*^9}, 3.596191214823338*^9, {
   3.607773326647754*^9, 3.607773329734789*^9}, {3.651917124437532*^9, 
   3.651917125207509*^9}},
 CellTags->{
  "ID:514433530", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "isPartIrreflOrder \[ScriptCapitalK]"},
 CellID->514433530],

Cell[BoxData[
 RowBox[{"isNoetherian", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalK]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalK]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isNoetherian \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:31238831", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185243606122*^9, 3.610185248564473*^9}, {
  3.6519171259403343`*^9, 3.651917126714593*^9}},
 CellTags->{
  "ID:31238831", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "isNoetherian \[ScriptCapitalK]"},
 CellID->31238831],

Cell[BoxData[
 RowBox[{"Multipliers", "[", "\[ScriptCapitalK]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Multipliers \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1572875342", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913259751965*^9, 3.590913264560975*^9}, 
   3.6519171280382853`*^9},
 CellTags->{
  "ID:1572875342", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "Multipliers \[ScriptCapitalK]"},
 CellID->1572875342],

Cell[BoxData[
 RowBox[{"R0", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R0 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:907858661", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913286891715*^9, 3.590913289519525*^9}, 
   3.651917129986766*^9},
 CellTags->{
  "ID:907858661", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R0 \[ScriptCapitalK]"},
 CellID->907858661],

Cell[BoxData[
 RowBox[{"R1", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R1 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1598185495", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.5909133104027*^9, 3.590913322109912*^9}, 
   3.6519171306869097`*^9},
 CellTags->{
  "ID:1598185495", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R1 \[ScriptCapitalK]"},
 CellID->1598185495],

Cell[BoxData[
 RowBox[{"R2", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R2 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:325105305", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913311652075*^9, 3.590913323321193*^9}, 
   3.651917131387651*^9},
 CellTags->{
  "ID:325105305", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R2 \[ScriptCapitalK]"},
 CellID->325105305],

Cell[BoxData[
 RowBox[{"R3", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R3 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1281658915", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913312278519*^9, 3.590913324239264*^9}, 
   3.6519171321249437`*^9},
 CellTags->{
  "ID:1281658915", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R3 \[ScriptCapitalK]"},
 CellID->1281658915],

Cell[BoxData[
 RowBox[{"R4", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R4 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1991263829", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913312942302*^9, 3.590913325377272*^9}, 
   3.651917132825877*^9},
 CellTags->{
  "ID:1991263829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R4 \[ScriptCapitalK]"},
 CellID->1991263829],

Cell[BoxData[
 RowBox[{"R5", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R5 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:891013116", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913313568732*^9, 3.590913326222446*^9}, 
   3.651917133526257*^9},
 CellTags->{
  "ID:891013116", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R5 \[ScriptCapitalK]"},
 CellID->891013116],

Cell[BoxData[
 RowBox[{"R6", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R6 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:887229842", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913314159087*^9, 3.590913326959886*^9}, 
   3.651917134337335*^9},
 CellTags->{
  "ID:887229842", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R6 \[ScriptCapitalK]"},
 CellID->887229842],

Cell[BoxData[
 RowBox[{"R7", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R7 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:4782934", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913314749098*^9, 3.590913327947563*^9}, 
   3.651917135294532*^9},
 CellTags->{
  "ID:4782934", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R7 \[ScriptCapitalK]"},
 CellID->4782934],

Cell[BoxData[
 RowBox[{"R8", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R8 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1656004832", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913315347097*^9, 3.590913328976076*^9}, 
   3.651917136186964*^9},
 CellTags->{
  "ID:1656004832", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R8 \[ScriptCapitalK]"},
 CellID->1656004832],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1455483400]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1061646154],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isPartIrreflOrder \[ScriptCapitalK])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isPartIrreflOrder \[ScriptCapitalK])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isPartIrreflOrder", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], ",", 
             RowBox[{"ReductionField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isPartIrreflOrder", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], ",", 
             RowBox[{"ReductionField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:514433530"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p514433530-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p514433530-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:514433530", "Proof|ID:514433530-1"},
 CellID->1256466167],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p514433530-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->717504818]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isNoetherian \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isNoetherian \[ScriptCapitalK])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isNoetherian", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], ",", 
             RowBox[{"ReductionField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isNoetherian", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], ",", 
             RowBox[{"ReductionField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:31238831"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p31238831-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p31238831-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:31238831", "Proof|ID:31238831-1"},
 CellID->2034491483],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p31238831-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1669813662]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Multipliers \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Multipliers \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"Multipliers", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"Multipliers", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1572875342"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1572875342-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1572875342-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1572875342", "Proof|ID:1572875342-1"},
 CellID->1020709635],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1572875342-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1654379224]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R0 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R0 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R0", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R0", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:907858661"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p907858661-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p907858661-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:907858661", "Proof|ID:907858661-1"},
 CellID->422286296],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p907858661-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->143092594]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R1 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R1 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R1", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R1", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1598185495"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1598185495-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1598185495-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1598185495", "Proof|ID:1598185495-1"},
 CellID->1472209392],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1598185495-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1261820315]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R2 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R2 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R2", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R2", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:325105305"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p325105305-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p325105305-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:325105305", "Proof|ID:325105305-1"},
 CellID->332550437],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p325105305-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1985798367]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R3 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R3 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R3", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R3", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1281658915"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1281658915-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1281658915-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1281658915", "Proof|ID:1281658915-1"},
 CellID->1299588167],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1281658915-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1806825968]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R4 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R4 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R4", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R4", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1991263829"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1991263829-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1991263829-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1991263829", "Proof|ID:1991263829-1"},
 CellID->291588382],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1991263829-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2022277391]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R5 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R5 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R5", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R5", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:891013116"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p891013116-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p891013116-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:891013116", "Proof|ID:891013116-1"},
 CellID->1089921470],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p891013116-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->763023223]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R6 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R6 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R6", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R6", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:887229842"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p887229842-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p887229842-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:887229842", "Proof|ID:887229842-1"},
 CellID->595600766],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p887229842-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->153744125]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R7 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R7 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R7", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R7", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:4782934"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p4782934-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p4782934-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:4782934", "Proof|ID:4782934-1"},
 CellID->209477845],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p4782934-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1003838192]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R8 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R8 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R8", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R8", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1656004832"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1656004832-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1656004832-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1656004832", "Proof|ID:1656004832-1"},
 CellID->1100102311],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1656004832-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1570103634]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->2080814435],

Cell[CellGroupData[{

Cell["Theorem: Reduction Fields are Reduction Rings", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.590836382026204*^9, 3.590836405537914*^9}},
 CellID->1560907237],

Cell[BoxData[
 RowBox[{"isReductionRing", "[", "\[ScriptCapitalK]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReductionRing \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:89016907", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590836412406496*^9, 3.590836437491688*^9}, {
  3.590836531559107*^9, 3.59083654159638*^9}, {3.607338911658421*^9, 
  3.607338922877867*^9}, {3.6519170960300817`*^9, 3.6519171193165503`*^9}},
 CellTags->{
  "ID:89016907", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "isReductionRing \[ScriptCapitalK]"},
 CellID->89016907],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->58748202]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1288020955],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReductionRing \[ScriptCapitalK])\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(isReductionRing \[ScriptCapitalK])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isReductionRing", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isReductionRing", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:89016907"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p89016907-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p89016907-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:89016907", "Proof|ID:89016907-1"},
 CellID->91423861],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p89016907-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->160819088]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Algorithmic Reduction Rings (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.651917190780504*^9, 3.651917195724601*^9}, {
  3.6525111451592693`*^9, 3.652511201753603*^9}},
 CellID->1867114071],

Cell["", "OpenEnvironment",
 CellID->1232421138],

Cell[CellGroupData[{

Cell["\<\
Lemma: Alternative Characterization of \[Tilde] in \[ScriptCapitalK]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.6520847406127462`*^9, {3.652086523792762*^9, 3.652086537392749*^9}, {
   3.652177961073176*^9, 3.6521779735901623`*^9}},
 CellID->1198258968],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalK]"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "m"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Fields.nb", 406426141}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.651921703187111*^9, 3.651921726256713*^9}, {
  3.651921769877788*^9, 3.651921771564828*^9}, {3.6519221237063913`*^9, 
  3.651922123928803*^9}},
 CellTags->{
  "ID:406426141", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->406426141],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalK]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"m", 
       UnderscriptBox["*", "\[ScriptCapitalK]"], "a"}], 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalK]"], 
      RowBox[{"m", 
       UnderscriptBox["*", "\[ScriptCapitalK]"], "b"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(10)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:648650518", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.652086547581908*^9, 3.652086600466132*^9}},
 CellTags->{
  "ID:648650518", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "10"},
 CellID->648650518],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
    RowBox[{"m", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalK]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       UnderscriptBox["multN", "\[ScriptCapitalK]"], "[", "c", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalK]"], 
       RowBox[{"[", "n", "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        UnderscriptBox["mult", "\[ScriptCapitalK]"], "[", 
        RowBox[{
         RowBox[{"m", 
          UnderscriptBox["*", "\[ScriptCapitalK]"], "n"}], ",", "c", ",", 
         "i"}], "]"}], "\[DoubleLongLeftRightArrow]", 
       RowBox[{
        UnderscriptBox["mult", "\[ScriptCapitalK]"], "[", 
        RowBox[{"n", ",", "c", ",", "i"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(11)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1161287156", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.652086620549049*^9, 3.652086706209132*^9}},
 CellTags->{
  "ID:1161287156", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "11"},
 CellID->1161287156],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Tilde]", "m"], "\[ScriptCapitalK]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{"a", "\[NotEqual]", 
          UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
         RowBox[{"b", "\[NotEqual]", 
          UnderscriptBox["0", "\[ScriptCapitalK]"]}], "\[And]", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"m", "\[Equal]", 
           RowBox[{"b", 
            UnderscriptBox["/", "\[ScriptCapitalK]"], "a"}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]},
       {
        RowBox[{
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"a", "\[Equal]", 
           UnderscriptBox["0", "\[ScriptCapitalK]"]}], 
          TagBox[")",
           "AutoParentheses"]}], "\[And]", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"b", "\[Equal]", 
           UnderscriptBox["0", "\[ScriptCapitalK]"]}], 
          TagBox[")",
           "AutoParentheses"]}], "\[And]", 
         RowBox[{"m", "\[NotEqual]", 
          UnderscriptBox["0", "\[ScriptCapitalK]"]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(12)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:731335719", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.652084755873242*^9, 3.652084874849999*^9}},
 CellTags->{
  "ID:731335719", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "12"},
 CellID->731335719],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Tilde]", "\[ScriptCapitalK]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", "\[Equal]", 
      UnderscriptBox["0", "\[ScriptCapitalK]"]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", "\[Equal]", 
      UnderscriptBox["0", "\[ScriptCapitalK]"]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(13)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:424912135", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.652089023473731*^9, 3.65208904578428*^9}},
 CellTags->{
  "ID:424912135", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "13"},
 CellID->424912135],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->691543564]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1614400098],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(10)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(10)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], 
               StyleBox["b", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], 
                 StyleBox["a", "ExpressionVariable"]}], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], 
                 StyleBox[
                 "b", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], 
               StyleBox["b", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], 
                 StyleBox["a", "ExpressionVariable"]}], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], 
                 StyleBox[
                 "b", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:648650518"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p648650518-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p648650518-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:648650518", "Proof|ID:648650518-1"},
 CellID->1191457334],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p648650518-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->401303854]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(11)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(11)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["mult", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], 
                    StyleBox["n", "ExpressionVariable"]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox["mult", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["mult", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], 
                    StyleBox["n", "ExpressionVariable"]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   UnderscriptBox["mult", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1161287156"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1161287156-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1161287156-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1161287156", "Proof|ID:1161287156-1"},
 CellID->1090967611],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1161287156-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->929802187]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(12)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(12)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Tilde]", 
               StyleBox["m", "ExpressionVariable"]], 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], 
                    UnderscriptBox["/", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], 
                    StyleBox["a", "ExpressionVariable"]}]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Tilde]", 
               StyleBox["m", "ExpressionVariable"]], 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], 
                    UnderscriptBox["/", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], 
                    StyleBox["a", "ExpressionVariable"]}]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:731335719"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p731335719-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p731335719-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:731335719", "Proof|ID:731335719-1"},
 CellID->1170894549],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p731335719-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1475281485]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(13)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(13)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:424912135"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p424912135-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p424912135-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:424912135", "Proof|ID:424912135-1"},
 CellID->1080465504],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p424912135-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->802282293]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1420224099],

Cell[CellGroupData[{

Cell["Lemma: Properties of red in \[ScriptCapitalK]", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6521781251919518`*^9, 3.652178131739884*^9}},
 CellID->784653472],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "\[ScriptCapitalK]", "]"}]], 
    RowBox[{"[", "A", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["red", "\[ScriptCapitalK]"], "[", "A", "]"}], 
      "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"a", "\[Element]", "A"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", "\[Equal]", 
        UnderscriptBox["0", "\[ScriptCapitalK]"]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(16)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:504163934", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6521789863375387`*^9, 3.6521790344244013`*^9}},
 CellTags->{
  "ID:504163934", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "16"},
 CellID->504163934],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "\[ScriptCapitalK]", "]"}]], 
    RowBox[{"[", 
     RowBox[{"A", ",", "B"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", 
     UnderscriptBox["\[SquareSubset]", "\[ScriptCapitalK]"], "B"}], 
    "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["red", "\[ScriptCapitalK]"], "[", "A", "]"}], 
           "\[Equal]", 
           RowBox[{"{", "}"}]}]},
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["red", "\[ScriptCapitalK]"], "[", "B", "]"}], 
           "\[NotEqual]", 
           RowBox[{"{", "}"}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(17)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1091730754", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.652178702693363*^9, 3.652178779799025*^9}, {
  3.6521789274985847`*^9, 3.6521789727890577`*^9}},
 CellTags->{
  "ID:1091730754", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "17"},
 CellID->1091730754],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->805694964]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1983325194],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(16)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(16)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["red", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[Equal]", 
             "\[EmptySet]"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Element]", 
                   StyleBox["A", "ExpressionVariable"]}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], 
                  "]"}]]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["red", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[Equal]", 
             "\[EmptySet]"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Element]", 
                   StyleBox["A", "ExpressionVariable"]}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionField", "[", 
                  StyleBox["K", "ExpressionVariable"], 
                  "]"}]]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:504163934"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p504163934-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p504163934-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:504163934", "Proof|ID:504163934-1"},
 CellID->189456215],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p504163934-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->251515074]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(17)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(17)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["A", "ExpressionVariable"], 
             UnderscriptBox["\[SquareSubset]", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], 
             StyleBox["B", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], "\[Equal]", 
               "\[EmptySet]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               "\[EmptySet]"}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["A", "ExpressionVariable"], 
             UnderscriptBox["\[SquareSubset]", 
              RowBox[{"ReductionField", "[", 
                StyleBox["K", "ExpressionVariable"], "]"}]], 
             StyleBox["B", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], "\[Equal]", 
               "\[EmptySet]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  RowBox[{"ReductionField", "[", 
                    StyleBox["K", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               "\[EmptySet]"}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1091730754"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1091730754-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1091730754-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1091730754", "Proof|ID:1091730754-1"},
 CellID->726474309],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1091730754-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->19880576]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->13625358],

Cell[CellGroupData[{

Cell["\<\
Lemma: Reduction Fields satisfy Axioms of Algorithmic Reduction Rings\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.651917437724382*^9, 3.6519174477671127`*^9}},
 CellID->1399361628],

Cell[BoxData[
 RowBox[{"R9", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R9 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1818840710", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913317183609*^9, 3.590913330114065*^9}, 
   3.651917139199432*^9},
 CellTags->{
  "ID:1818840710", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R9 \[ScriptCapitalK]"},
 CellID->1818840710],

Cell[BoxData[
 RowBox[{"R10", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R10 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:718799609", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913317920502*^9, 3.590913332902015*^9}, 
   3.651917139974145*^9},
 CellTags->{
  "ID:718799609", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R10 \[ScriptCapitalK]"},
 CellID->718799609],

Cell[BoxData[
 RowBox[{"R11", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R11 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:563907237", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913318550186*^9, 3.59091333451809*^9}, 
   3.651917140711617*^9},
 CellTags->{
  "ID:563907237", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R11 \[ScriptCapitalK]"},
 CellID->563907237],

Cell[BoxData[
 RowBox[{"R12", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R12 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1162205858", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590913319540639*^9, 3.590913335627507*^9}, 
   3.65191714144341*^9},
 CellTags->{
  "ID:1162205858", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R12 \[ScriptCapitalK]"},
 CellID->1162205858],

Cell[BoxData[
 RowBox[{"R13", "[", "\[ScriptCapitalK]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R13 \[ScriptCapitalK])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:372596839", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.652089712366151*^9, 3.652089713765683*^9}},
 CellTags->{
  "ID:372596839", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R13 \[ScriptCapitalK]"},
 CellID->372596839],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->940321384]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->144901910],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R9 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R9 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R9", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R9", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1818840710"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1818840710-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1818840710-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1818840710", "Proof|ID:1818840710-1"},
 CellID->271819944],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1818840710-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1423115614]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R10 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R10 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R10", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R10", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:718799609"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p718799609-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p718799609-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:718799609", "Proof|ID:718799609-1"},
 CellID->300836608],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p718799609-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->191748015]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R11 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R11 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R11", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R11", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:563907237"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p563907237-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p563907237-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:563907237", "Proof|ID:563907237-1"},
 CellID->1260116027],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p563907237-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->625079081]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R12 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R12 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R12", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R12", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1162205858"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1162205858-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1162205858-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1162205858", "Proof|ID:1162205858-1"},
 CellID->1594052474],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1162205858-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->475932409]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R13 \[ScriptCapitalK])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R13 \[ScriptCapitalK])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R13", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R13", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:372596839"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p372596839-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p372596839-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:372596839", "Proof|ID:372596839-1"},
 CellID->488894158],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p372596839-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->850600494]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->228270346],

Cell[CellGroupData[{

Cell["Theorem: Reduction Fields are Algorithmic Reduction Rings", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.651917458493029*^9, 3.6519174632313538`*^9}},
 CellID->1906185036],

Cell[BoxData[
 RowBox[{"isAlgoReductionRing", "[", "\[ScriptCapitalK]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isAlgoReductionRing \[ScriptCapitalK])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:476788814", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6519174675316677`*^9, 3.6519174701838923`*^9}},
 CellTags->{
  "ID:476788814", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "isAlgoReductionRing \[ScriptCapitalK]"},
 CellID->476788814],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1565750311]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1552441335],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isAlgoReductionRing \[ScriptCapitalK])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isAlgoReductionRing \[ScriptCapitalK])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isAlgoReductionRing", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isField", "[", 
               StyleBox["K", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isAlgoReductionRing", "[", 
           RowBox[{"ReductionField", "[", 
             StyleBox["K", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:476788814"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p476788814-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p476788814-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:476788814", "Proof|ID:476788814-1"},
 CellID->1236650353],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p476788814-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->399055891]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
},
WindowSize->{1103, 761},
WindowMargins->{{1, Automatic}, {Automatic, 0}},
CounterAssignments->{{"TheoremaFormulaCounter" -> 7}},
Magnification->1.,
FrontEndVersion->"9.0 for Linux x86 (64-bit) (February 7, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Theorema"}, "TheoremaCore.nb"]]], 
   
   Cell["Modify the definitions below to change the default appearance of all \
cells in a given style. \nMake modifications to any definition using commands \
in the Format menu.", "Text"], 
   Cell[
    CellGroupData[{
      Cell["Notebook Options Settings", "Section"], 
      Cell[
      "The options defined for the style below will be used at the Notebook \
level.", "Text"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Notebook"], InputAliases -> {"and" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[And]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "or" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", "\[Or]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "not" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Not]", "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "impl" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Implies]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[DoubleLongLeftRightArrow]", 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                "\[SelectionPlaceholder]", "\[Equal]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox[
                  
                  RowBox[{
                   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                  "a:=b"], "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", ":=", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "andn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[And]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "orn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Or]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "equivn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[DoubleLeftRightArrow]", "\[Piecewise]", 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "cdist" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Piecewise]", 
                 
                 GridBox[{{
                   "\[SelectionPlaceholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}, {
                   "\[Placeholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "far" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exr" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "farc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exrc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "let" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["let", 
                  RowBox[{"\[Placeholder]", "=", "\[Placeholder]"}]], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "app" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293a", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "prep" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293b", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "join" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22c8", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "elemT" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22ff", Identity, SyntaxForm -> "a+b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "subop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "supop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SuperscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "subsupop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubsuperscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "oop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "uoop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderoverscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "osop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox[
                OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
                "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "op" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               TagBox["\[SelectionPlaceholder]", Identity], 
               TagBox[")", "AutoParentheses"]}], "[]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e114", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e115", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "()" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:fd3e", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:fd3f", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e8", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e9", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e117", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e118", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "[[]]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e6", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e7", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(())" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2e28", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2e29", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<<>>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27ea", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27eb", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{{}}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2983", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2984", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(" -> 
            TagBox["(", "AutoParentheses"], ")" -> 
            TagBox[")", "AutoParentheses"]}, Background -> 
          RGBColor[0.957, 0.96, 0.97]], 
         Cell[
          StyleData["Notebook", "Printout"], Background -> None]}, Open]]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell["Style Environment Names", "Section"], 
      Cell[
       StyleData[All, "Working"], DockedCells -> {
         Cell[
         "Theorema 2.0", "SBO", FontTracking -> 8, 
          CellMargins -> {{0, 0}, {0, 0}}, 
          CellFrameMargins -> {{0, 50}, {0, 0}}, FontSize -> 28, FontSlant -> 
          "Plain", FontColor -> RGBColor[0.29, 0.46, 0.6], TextAlignment -> 
          Right, Background -> RGBColor[0.36, 0.54, 0.69]]}], 
      Cell[
       StyleData[
       All, "Layout", StyleDefinitions -> StyleData[All, "Working"]], 
       MenuSortingValue -> 1200]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Styles for Title and Section Cells", "Section"], 
      Cell[
       CellGroupData[{
         Cell["Title", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], CellMargins -> {{27, 27}, {10, 30}}, 
             TextAlignment -> Center, MenuSortingValue -> 1100, 
             MenuCommandKey -> "1", FontFamily -> "Bitstream Charter", 
             FontSize -> 36, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Title", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{6, 10}, {5, 30}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], CellMargins -> {{27, 27}, {20, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1150, FontFamily -> 
             "Bitstream Charter", FontSize -> 24, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{6, 10}, {5, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], CellMargins -> {{27, 27}, {8, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], CellMargins -> {{27, 27}, {18, 20}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Sectioning", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], ShowGroupOpener -> True, MenuSortingValue -> 
             1250, MenuCommandKey -> "4", FontFamily -> "Bitstream Charter", 
             FontSize -> 20, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Section", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{2, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {8, 12}}, 
             MenuSortingValue -> 1300, MenuCommandKey -> "5", FontFamily -> 
             "Bitstream Charter", FontSize -> 14, FontWeight -> "Bold", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Subsection", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{9, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {2, 10}}, 
             MenuSortingValue -> 1350, MenuCommandKey -> "6", FontFamily -> 
             "Bitstream Charter", FontSize -> 12, FontWeight -> "Bold", 
             FontSlant -> "Italic", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsection", "Printout"], ShowGroupOpener -> False,
              CellMargins -> {{35, 0}, {7, 14}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsubsection"], CellDingbat -> None, 
             ShowGroupOpener -> True, 
             CellMargins -> {{27, Inherited}, {Inherited, Inherited}}, 
             FontFamily -> "Bitstream Charter", FontSize -> 11, FontWeight -> 
             "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsubsection", "Printout"], 
             CellMargins -> {{50, 0}, {7, 14}}]}, Closed]]}, Open]]}, Open]], 
   
   Cell[
    CellGroupData[{
      Cell["Styles for Publication\[Hyphen]ready Documents", "Section"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Author"], CellMargins -> {{27, 30}, {20, 5}}, 
          MenuSortingValue -> 1900, FontFamily -> "Helvetica", FontSize -> 14,
           FontSlant -> "Italic"], 
         Cell[
          StyleData["Author", "Printout"], CellMargins -> {{6, 10}, {0, 10}}, 
          FontSize -> 10]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Abstract"], CellFrame -> {{0, 0}, {0.5, 0.5}}, 
          CellMargins -> {{27, 140}, {20, 20}}, LineSpacing -> {0.9, 0}, 
          MenuSortingValue -> 2050, FontFamily -> "Times", FontSize -> 12], 
         Cell[
          StyleData["Abstract", "Printout"], Hyphenation -> True, 
          LineSpacing -> {1, 2}, FontSize -> 10]}, Closed]]}, Closed]], 
   Cell[
    CellGroupData[{
      Cell["Text", "Section"], 
      Cell[
       StyleData["Text"], CellMargins -> {{27, 27}, {Inherited, Inherited}}, 
       TextJustification -> 1., ParagraphIndent -> 15, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["EnvironmentText"], ShowCellBracket -> Automatic, 
       CellMargins -> {{82, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextJustification -> 1., ParagraphSpacing -> {0.3, 0.}, 
       MenuSortingValue -> 1480, MenuCommandKey -> "T", FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Item"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "Item", 
          CounterAssignments -> {{"Subitem", 0}, {"Subsubitem", 0}}, 
          MenuSortingValue -> 1600, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Item", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemParagraph"], CellMargins -> {{81, 10}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          DefaultNewCellStyle -> "Item", MenuSortingValue -> 1600, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemParagraph", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "Subitem", 
          CounterAssignments -> {{"Subsubitem", 0}}, MenuSortingValue -> 1610,
           FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subitem", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemParagraph"], CellMargins -> {{105, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subitem", 
          MenuSortingValue -> 1610, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemParagraph", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subsubitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1620, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subsubitem", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemParagraph"], 
          CellMargins -> {{129, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1625, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemParagraph", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "ItemNumbered", 
          CounterAssignments -> {{"SubitemNumbered", 0}, {
            "SubsubitemNumbered", 0}}, MenuSortingValue -> 1630, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemNumbered", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "SubitemNumbered", 
          CounterAssignments -> {{"SubsubitemNumbered", 0}}, MenuSortingValue -> 
          1640, FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemNumbered", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], ".", 
              CounterBox["SubsubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> 
          "SubitemNumbered", CounterIncrements -> "SubsubitemNumbered", 
          MenuSortingValue -> 1645, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemNumbered", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Theorema", "Section"], 
      Cell[
       StyleData["GlobalDeclaration"], CellFrame -> {{0, 2}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 3}, {2, 2}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       InitializationCell -> True, CellFrameMargins -> 10, CellFrameColor -> 
       RGBColor[0.73, 0.65, 0.61], CellFrameLabelMargins -> {{33, 7}, {1, 1}},
        ShowAutoStyles -> False, MenuSortingValue -> 1441, MenuCommandKey -> 
       "G", FontSize -> 14], 
      Cell[
       StyleData["OpenEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {-4, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       StyleData["EnvironmentHeader"], 
       CellFrame -> {{False, False}, {False, False}}, 
       CellMargins -> {{76, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 6, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       ShowAutoStyles -> False, MenuSortingValue -> 1450, FontSize -> 14, 
       FontWeight -> "Bold", FontVariations -> {"CapsType" -> "SmallCaps"}, 
       Background -> RGBColor[0.89, 0.8, 0.69]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FormalTextInputFormula"], ShowCellBracket -> Automatic, 
          CellMargins -> {{96, 10}, {2, 2}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
          InitializationCell -> True, CellFrameLabelMargins -> 53, 
          ShowAutoStyles -> False, LimitsPositioningTokens -> {}, 
          MenuSortingValue -> 1470, MenuCommandKey -> "F", FontWeight -> 
          "Bold", FontColor -> RGBColor[0.25, 0.19, 0.2]], 
         Cell[
          StyleData["FormalTextInputFormula", "Layout"], Editable -> False], 
         Cell[
          StyleData["FormalTextInputFormula", "Presentation"], Editable -> 
          False]}, Open]], 
      Cell[
       StyleData["EndEnvironmentMarker"], 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextAlignment -> Right, MenuSortingValue -> 1480], 
      Cell[
       StyleData["CloseEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {False, True}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {Inherited, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Computation"], CellFrame -> {{1.5, 0}, {0, 0}}, 
          ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 7}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
          CellFrameMargins -> {{8, 8}, {3, 4}}, CellFrameColor -> 
          RGBColor[0.486, 0.33, 0.32], ShowAutoStyles -> False, 
          MenuSortingValue -> 1510, FontWeight -> "Bold", FontColor -> 
          RGBColor[0.23, 0.25, 0.34], Background -> 
          RGBColor[0.88, 0.88, 0.92]], 
         Cell[
          StyleData["Computation", "Layout"], Editable -> False], 
         Cell[
          StyleData["Computation", "Presentation"], Editable -> False]}, 
        Open]], 
      Cell[
       StyleData["ComputationResult"], CellFrame -> {{1.5, 0}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
       CellFrameMargins -> {{8, 8}, {4, 3}}, CellFrameColor -> 
       RGBColor[0.486, 0.33, 0.32], StyleMenuListing -> None, FontColor -> 
       RGBColor[0.23, 0.25, 0.34], Background -> RGBColor[0.88, 0.88, 0.92]], 
      
      Cell[
       StyleData["ComputationInfo"], ShowCellBracket -> False, 
       CellMargins -> {{120, 10}, {10, 2}}, StyleMenuListing -> None, 
       FontFamily -> $CellContext`Arial], 
      Cell[
       StyleData["ComputationInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{144, 10}, {10, 2}}, StyleMenuListing -> None], 
      Cell[
       StyleData["ProofInfo"], ShowCellBracket -> False, 
       CellMargins -> {{76, 10}, {6, 6}}, CellFrameLabelMargins -> 15, 
       StyleMenuListing -> None, FontFamily -> "Arial"], 
      Cell[
       StyleData["ProofInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{100, 10}, {0, 0}}, StyleMenuListing -> None], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FrameLabel"], LanguageCategory -> None, StyleMenuListing -> 
          None, FontFamily -> "Times New Roman", FontSize -> 10, FontWeight -> 
          "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
          FontVariations -> {
           "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
            "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
          RGBColor[0.55, 0.44, 0.42], Background -> None], 
         Cell[
          StyleData["FrameLabel", "Printout"]]}, Open]], 
      Cell[
       StyleData["OpenArchive"], CellFrame -> {{False, False}, {False, True}},
        ShowCellBracket -> Automatic, CellMargins -> {{70, 70}, {0, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 20}, CellFrameMargins -> 3, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], CellFrameLabelMargins -> 1, 
       DefaultFormatType -> DefaultInputFormatType, ShowAutoStyles -> False, 
       MenuSortingValue -> 1210, FontWeight -> Bold, Background -> 
       RGBColor[0.73, 0.74, 0.84]], 
      Cell[
       StyleData["ArchiveInfo"], ShowCellBracket -> Automatic, 
       CellMargins -> {{70, 70}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 5, CellFrameLabelMargins -> 3, ShowCellTags -> 
       False, ShowAutoStyles -> False, MenuSortingValue -> 1220, FontWeight -> 
       "Bold", Background -> RGBColor[0.88, 0.88, 0.92]], 
      Cell[
       StyleData["CloseArchive"], 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> 
       Automatic, CellMargins -> {{0, 0}, {5, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 25}, CellFrameMargins -> 2, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], TextAlignment -> Center, 
       MenuSortingValue -> 1230, FontFamily -> "Arial Black", FontSize -> 24, 
       FontWeight -> "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       Background -> RGBColor[0.73, 0.65, 0.61]], 
      Cell[
       StyleData["IncludeArchive"], CellFrame -> {{4, 0}, {False, False}}, 
       ShowCellBracket -> Automatic, 
       CellMargins -> {{81, 100}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameColor -> RGBColor[0.55, 0.44, 0.42], 
       CellFrameLabels -> {{"\[RightSkeleton]", None}, {None, None}}, 
       CellFrameLabelMargins -> 25, MenuSortingValue -> 1445], 
      Cell[
       StyleData["ExpressionVariable"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionABF"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionMeta"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["AutoParentheses"], Editable -> False, SpanMaxSize -> 3, 
          FontOpacity -> 0.15], 
         Cell[
          StyleData["AutoParentheses", "Layout"], FontOpacity -> 0], 
         Cell[
          StyleData["AutoParentheses", "SlideShow"], FontOpacity -> 0]}, 
        Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Emphasized"], MenuCommandKey -> "E", FontSlant -> 
          "Italic"], 
         Cell[
          StyleData["Emphasized", "SlideShow"], FontSlant -> "Plain", 
          FontColor -> RGBColor[0.36, 0.54, 0.69]]}, Open]], 
      Cell[
       StyleData["GenericButton"], FontFamily -> "Bitstream Charter", 
       FontWeight -> "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
       FontVariations -> {
        "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
         "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
       RGBColor[0.23, 0.25, 0.34]], 
      Cell[
       StyleData["TheoremaForm"], "TwoByteSyntaxCharacterAutoReplacement" -> 
       True, LineSpacing -> {1.25, 0, 2}, SingleLetterItalics -> True, 
       StyleMenuListing -> None, FontFamily -> "Courier"], 
      Cell[
       StyleData["ComponentActive"], FontFamily -> "Times New Roman"], 
      Cell[
       StyleData["ComponentInactive"], FontFamily -> "Times New Roman", 
       FontOpacity -> 0.4], 
      Cell[
       StyleData["ComponentEmpty"], FontFamily -> "Times New Roman", 
       FontVariations -> {"StrikeThrough" -> True}], 
      Cell[
       StyleData["TabView"], FontFamily -> "Times New Roman", FontSize -> 10], 
      Cell[
       StyleData["TabViewLabel"], FontFamily -> "Times New Roman", FontSize -> 
       10, FontWeight -> "Bold"], 
      Cell[
       StyleData["FormReference", StyleDefinitions -> StyleData["Text"]], 
       FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> "Plain",
        FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["FormReferenceHover", StyleDefinitions -> StyleData["Text"]],
        FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> 
       "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> True}, 
       FontColor -> RGBColor[0.404, 0.43, 0.545]], 
      Cell[
       StyleData["DisplayFormula"], CellMargins -> {{36, 6}, {7, 7}}, 
       ConversionRules :> {
        "HTML" -> {{"", ""}, {"<p class=\"Equation\">", "</p>"}}}, 
       CellHorizontalScrolling -> True, DefaultFormatType -> TraditionalForm, 
       InputAutoReplacements -> {}, TextAlignment -> Center, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       ButtonBoxOptions -> {Appearance -> {Automatic, None}}]}, Open]]}, 
  Visible -> False, FrontEndVersion -> 
  "9.0 for Linux x86 (64-bit) (February 7, 2013)", StyleDefinitions -> 
  "StylesheetFormatting.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "ID:991632779"->{
  Cell[28454, 790, 641, 16, 30, "GlobalDeclaration",
   CellTags->{
    "ID:991632779", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->991632779]},
 "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"->{\

  Cell[28454, 790, 641, 16, 30, "GlobalDeclaration",
   CellTags->{
    "ID:991632779", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->991632779],
  Cell[29098, 808, 816, 20, 30, "GlobalDeclaration",
   CellTags->{
    "ID:13541845", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->13541845],
  Cell[29917, 830, 942, 21, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1796175771", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->1796175771],
  Cell[30862, 853, 1198, 34, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:204732646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field \[Element]"},
   CellID->204732646],
  Cell[32063, 889, 828, 23, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:125417415", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field 0"},
   CellID->125417415],
  Cell[32894, 914, 881, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1874597366", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field 1"},
   CellID->1874597366],
  Cell[33778, 940, 926, 26, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:791499509", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field +"},
   CellID->791499509],
  Cell[34707, 968, 931, 26, 28, "FormalTextInputFormula",
   CellTags->{
    "ID:1535890097", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field - unary"},
   CellID->1535890097],
  Cell[35641, 996, 924, 26, 28, "FormalTextInputFormula",
   CellTags->{
    "ID:328879666", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field -"},
   CellID->328879666],
  Cell[36568, 1024, 928, 26, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1643707054", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field \[RawStar]"},
   CellID->1643707054],
  Cell[37499, 1052, 1073, 30, 28, "FormalTextInputFormula",
   CellTags->{
    "ID:1452171966", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field inverse"},
   CellID->1452171966],
  Cell[38575, 1084, 928, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1982468810", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field /"},
   CellID->1982468810],
  Cell[39506, 1112, 1412, 43, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1869336472", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field \[Succeeds]"},
   CellID->1869336472],
  Cell[40921, 1157, 1138, 30, 46, "FormalTextInputFormula",
   CellTags->{
    "ID:1352206528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field multN"},
   CellID->1352206528],
  Cell[42062, 1189, 1108, 33, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:305610078", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field mult"},
   CellID->305610078],
  Cell[43173, 1224, 1379, 42, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1075639810", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field mult 1"},
   CellID->1075639810],
  Cell[44555, 1268, 1141, 33, 57, "FormalTextInputFormula",
   CellTags->{
    "ID:1151861642", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field rdm"},
   CellID->1151861642],
  Cell[46230, 1321, 1758, 44, 68, "FormalTextInputFormula",
   CellTags->{
    "ID:2067272590", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field lcrd"},
   CellID->2067272590],
  Cell[48437, 1383, 1455, 37, 63, "FormalTextInputFormula",
   CellTags->{
    "ID:1497531972", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field lcrd unary"},
   CellID->1497531972],
  Cell[49895, 1422, 1279, 34, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:529122150", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field cpm"},
   CellID->529122150],
  Cell[52014, 1492, 888, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1788558801", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "\[ScriptCapitalQ]"},
   CellID->1788558801],
  Cell[71073, 2150, 671, 16, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1470999011", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->1470999011],
  Cell[71787, 2171, 756, 18, 33, "GlobalDeclaration",
   CellTags->{
    "ID:853918182", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->853918182],
  Cell[73223, 2214, 978, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:82222568", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isCommRing1 \[ScriptCapitalK]"},
   CellID->82222568],
  Cell[74204, 2240, 801, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1537851509", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isField \[ScriptCapitalK]"},
   CellID->1537851509],
  Cell[81513, 2471, 877, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2102486692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->2102486692],
  Cell[82393, 2495, 1359, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:274992099", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "1"},
   CellID->274992099],
  Cell[83755, 2538, 1128, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:816763103", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "2"},
   CellID->816763103],
  Cell[84886, 2573, 1655, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:83877220", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "3"},
   CellID->83877220],
  Cell[86544, 2629, 1491, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1728540744", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "4"},
   CellID->1728540744],
  Cell[88038, 2678, 660, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1733996578", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->1733996578],
  Cell[88701, 2696, 1310, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371484949", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "5"},
   CellID->371484949],
  Cell[90014, 2735, 1297, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:316306510", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "6"},
   CellID->316306510],
  Cell[91314, 2774, 1011, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1250046151", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "7"},
   CellID->1250046151],
  Cell[92328, 2804, 1123, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:304323379", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "8"},
   CellID->304323379],
  Cell[93454, 2839, 949, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1618605896", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "9"},
   CellID->1618605896],
  Cell[152947, 4371, 1200, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:514433530", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isPartIrreflOrder \[ScriptCapitalK]"},
   CellID->514433530],
  Cell[154150, 4406, 1062, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:31238831", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isNoetherian \[ScriptCapitalK]"},
   CellID->31238831],
  Cell[155215, 4439, 840, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1572875342", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "Multipliers \[ScriptCapitalK]"},
   CellID->1572875342],
  Cell[156058, 4463, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:907858661", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R0 \[ScriptCapitalK]"},
   CellID->907858661],
  Cell[156866, 4486, 808, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1598185495", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R1 \[ScriptCapitalK]"},
   CellID->1598185495],
  Cell[157677, 4509, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:325105305", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R2 \[ScriptCapitalK]"},
   CellID->325105305],
  Cell[158485, 4532, 810, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1281658915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R3 \[ScriptCapitalK]"},
   CellID->1281658915],
  Cell[159298, 4555, 808, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1991263829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R4 \[ScriptCapitalK]"},
   CellID->1991263829],
  Cell[160109, 4578, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:891013116", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R5 \[ScriptCapitalK]"},
   CellID->891013116],
  Cell[160917, 4601, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:887229842", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R6 \[ScriptCapitalK]"},
   CellID->887229842],
  Cell[161725, 4624, 799, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:4782934", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R7 \[ScriptCapitalK]"},
   CellID->4782934],
  Cell[162527, 4647, 808, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1656004832", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R8 \[ScriptCapitalK]"},
   CellID->1656004832],
  Cell[198419, 5761, 965, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:89016907", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isReductionRing \[ScriptCapitalK]"},
   CellID->89016907],
  Cell[203204, 5912, 874, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:406426141", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->406426141],
  Cell[204081, 5936, 1457, 46, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:648650518", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "10"},
   CellID->648650518],
  Cell[205541, 5984, 1961, 61, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:1161287156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "11"},
   CellID->1161287156],
  Cell[207505, 6047, 2302, 71, 78, "FormalTextInputFormula",
   CellTags->{
    "ID:731335719", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "12"},
   CellID->731335719],
  Cell[209810, 6120, 1415, 47, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:424912135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "13"},
   CellID->424912135],
  Cell[241004, 6924, 1665, 56, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:504163934", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "16"},
   CellID->504163934],
  Cell[242672, 6982, 1864, 59, 61, "FormalTextInputFormula",
   CellTags->{
    "ID:1091730754", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "17"},
   CellID->1091730754],
  Cell[255582, 7349, 808, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1818840710", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R9 \[ScriptCapitalK]"},
   CellID->1818840710],
  Cell[256393, 7372, 808, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:718799609", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R10 \[ScriptCapitalK]"},
   CellID->718799609],
  Cell[257204, 7395, 807, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:563907237", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R11 \[ScriptCapitalK]"},
   CellID->563907237],
  Cell[258014, 7418, 810, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1162205858", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R12 \[ScriptCapitalK]"},
   CellID->1162205858],
  Cell[258827, 7441, 782, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:372596839", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R13 \[ScriptCapitalK]"},
   CellID->372596839],
  Cell[274237, 7924, 846, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:476788814", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isAlgoReductionRing \[ScriptCapitalK]"},
   CellID->476788814]},
 "ID:13541845"->{
  Cell[29098, 808, 816, 20, 30, "GlobalDeclaration",
   CellTags->{
    "ID:13541845", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->13541845]},
 "ID:1796175771"->{
  Cell[29917, 830, 942, 21, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1796175771", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->1796175771]},
 "ID:204732646"->{
  Cell[30862, 853, 1198, 34, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:204732646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field \[Element]"},
   CellID->204732646]},
 "field \[Element]"->{
  Cell[30862, 853, 1198, 34, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:204732646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field \[Element]"},
   CellID->204732646]},
 "ID:125417415"->{
  Cell[32063, 889, 828, 23, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:125417415", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field 0"},
   CellID->125417415]},
 "field 0"->{
  Cell[32063, 889, 828, 23, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:125417415", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field 0"},
   CellID->125417415]},
 "ID:1874597366"->{
  Cell[32894, 914, 881, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1874597366", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field 1"},
   CellID->1874597366]},
 "field 1"->{
  Cell[32894, 914, 881, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1874597366", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field 1"},
   CellID->1874597366]},
 "ID:791499509"->{
  Cell[33778, 940, 926, 26, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:791499509", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field +"},
   CellID->791499509]},
 "field +"->{
  Cell[33778, 940, 926, 26, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:791499509", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field +"},
   CellID->791499509]},
 "ID:1535890097"->{
  Cell[34707, 968, 931, 26, 28, "FormalTextInputFormula",
   CellTags->{
    "ID:1535890097", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field - unary"},
   CellID->1535890097]},
 "field - unary"->{
  Cell[34707, 968, 931, 26, 28, "FormalTextInputFormula",
   CellTags->{
    "ID:1535890097", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field - unary"},
   CellID->1535890097]},
 "ID:328879666"->{
  Cell[35641, 996, 924, 26, 28, "FormalTextInputFormula",
   CellTags->{
    "ID:328879666", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field -"},
   CellID->328879666]},
 "field -"->{
  Cell[35641, 996, 924, 26, 28, "FormalTextInputFormula",
   CellTags->{
    "ID:328879666", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field -"},
   CellID->328879666]},
 "ID:1643707054"->{
  Cell[36568, 1024, 928, 26, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1643707054", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field \[RawStar]"},
   CellID->1643707054]},
 "field \[RawStar]"->{
  Cell[36568, 1024, 928, 26, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1643707054", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field \[RawStar]"},
   CellID->1643707054]},
 "ID:1452171966"->{
  Cell[37499, 1052, 1073, 30, 28, "FormalTextInputFormula",
   CellTags->{
    "ID:1452171966", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field inverse"},
   CellID->1452171966]},
 "field inverse"->{
  Cell[37499, 1052, 1073, 30, 28, "FormalTextInputFormula",
   CellTags->{
    "ID:1452171966", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field inverse"},
   CellID->1452171966]},
 "ID:1982468810"->{
  Cell[38575, 1084, 928, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1982468810", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field /"},
   CellID->1982468810]},
 "field /"->{
  Cell[38575, 1084, 928, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1982468810", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field /"},
   CellID->1982468810]},
 "ID:1869336472"->{
  Cell[39506, 1112, 1412, 43, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1869336472", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field \[Succeeds]"},
   CellID->1869336472]},
 "field \[Succeeds]"->{
  Cell[39506, 1112, 1412, 43, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:1869336472", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field \[Succeeds]"},
   CellID->1869336472]},
 "ID:1352206528"->{
  Cell[40921, 1157, 1138, 30, 46, "FormalTextInputFormula",
   CellTags->{
    "ID:1352206528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field multN"},
   CellID->1352206528]},
 "field multN"->{
  Cell[40921, 1157, 1138, 30, 46, "FormalTextInputFormula",
   CellTags->{
    "ID:1352206528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field multN"},
   CellID->1352206528]},
 "ID:305610078"->{
  Cell[42062, 1189, 1108, 33, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:305610078", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field mult"},
   CellID->305610078]},
 "field mult"->{
  Cell[42062, 1189, 1108, 33, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:305610078", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field mult"},
   CellID->305610078]},
 "ID:1075639810"->{
  Cell[43173, 1224, 1379, 42, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1075639810", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field mult 1"},
   CellID->1075639810]},
 "field mult 1"->{
  Cell[43173, 1224, 1379, 42, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1075639810", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field mult 1"},
   CellID->1075639810]},
 "ID:1151861642"->{
  Cell[44555, 1268, 1141, 33, 57, "FormalTextInputFormula",
   CellTags->{
    "ID:1151861642", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field rdm"},
   CellID->1151861642]},
 "field rdm"->{
  Cell[44555, 1268, 1141, 33, 57, "FormalTextInputFormula",
   CellTags->{
    "ID:1151861642", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field rdm"},
   CellID->1151861642]},
 "ID:2067272590"->{
  Cell[46230, 1321, 1758, 44, 68, "FormalTextInputFormula",
   CellTags->{
    "ID:2067272590", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field lcrd"},
   CellID->2067272590]},
 "field lcrd"->{
  Cell[46230, 1321, 1758, 44, 68, "FormalTextInputFormula",
   CellTags->{
    "ID:2067272590", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field lcrd"},
   CellID->2067272590]},
 "ID:1497531972"->{
  Cell[48437, 1383, 1455, 37, 63, "FormalTextInputFormula",
   CellTags->{
    "ID:1497531972", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field lcrd unary"},
   CellID->1497531972]},
 "field lcrd unary"->{
  Cell[48437, 1383, 1455, 37, 63, "FormalTextInputFormula",
   CellTags->{
    "ID:1497531972", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field lcrd unary"},
   CellID->1497531972]},
 "ID:529122150"->{
  Cell[49895, 1422, 1279, 34, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:529122150", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field cpm"},
   CellID->529122150]},
 "field cpm"->{
  Cell[49895, 1422, 1279, 34, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:529122150", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "field cpm"},
   CellID->529122150]},
 "ID:1788558801"->{
  Cell[52014, 1492, 888, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1788558801", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "\[ScriptCapitalQ]"},
   CellID->1788558801]},
 "\[ScriptCapitalQ]"->{
  Cell[52014, 1492, 888, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1788558801", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "\[ScriptCapitalQ]"},
   CellID->1788558801]},
 "ID:1470999011"->{
  Cell[71073, 2150, 671, 16, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1470999011", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->1470999011]},
 "ID:853918182"->{
  Cell[71787, 2171, 756, 18, 33, "GlobalDeclaration",
   CellTags->{
    "ID:853918182", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->853918182]},
 "ID:82222568"->{
  Cell[73223, 2214, 978, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:82222568", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isCommRing1 \[ScriptCapitalK]"},
   CellID->82222568]},
 "isCommRing1 \[ScriptCapitalK]"->{
  Cell[73223, 2214, 978, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:82222568", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isCommRing1 \[ScriptCapitalK]"},
   CellID->82222568]},
 "ID:1537851509"->{
  Cell[74204, 2240, 801, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1537851509", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isField \[ScriptCapitalK]"},
   CellID->1537851509]},
 "isField \[ScriptCapitalK]"->{
  Cell[74204, 2240, 801, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1537851509", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isField \[ScriptCapitalK]"},
   CellID->1537851509]},
 "Proof|ID:82222568"->{
  Cell[75163, 2272, 2303, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:82222568", "Proof|ID:82222568-1"},
   CellID->83229303]},
 "Proof|ID:82222568-1"->{
  Cell[75163, 2272, 2303, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:82222568", "Proof|ID:82222568-1"},
   CellID->83229303]},
 "Proof|ID:1537851509"->{
  Cell[77985, 2359, 2302, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1537851509", "Proof|ID:1537851509-1"},
   CellID->710651429]},
 "Proof|ID:1537851509-1"->{
  Cell[77985, 2359, 2302, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1537851509", "Proof|ID:1537851509-1"},
   CellID->710651429]},
 "ID:2102486692"->{
  Cell[81513, 2471, 877, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2102486692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->2102486692]},
 "ID:274992099"->{
  Cell[82393, 2495, 1359, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:274992099", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "1"},
   CellID->274992099]},
 "1"->{
  Cell[82393, 2495, 1359, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:274992099", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "1"},
   CellID->274992099]},
 "ID:816763103"->{
  Cell[83755, 2538, 1128, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:816763103", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "2"},
   CellID->816763103]},
 "2"->{
  Cell[83755, 2538, 1128, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:816763103", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "2"},
   CellID->816763103]},
 "ID:83877220"->{
  Cell[84886, 2573, 1655, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:83877220", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "3"},
   CellID->83877220]},
 "3"->{
  Cell[84886, 2573, 1655, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:83877220", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "3"},
   CellID->83877220]},
 "ID:1728540744"->{
  Cell[86544, 2629, 1491, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1728540744", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "4"},
   CellID->1728540744]},
 "4"->{
  Cell[86544, 2629, 1491, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1728540744", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "4"},
   CellID->1728540744]},
 "ID:1733996578"->{
  Cell[88038, 2678, 660, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1733996578", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->1733996578]},
 "ID:371484949"->{
  Cell[88701, 2696, 1310, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371484949", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "5"},
   CellID->371484949]},
 "5"->{
  Cell[88701, 2696, 1310, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371484949", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "5"},
   CellID->371484949]},
 "ID:316306510"->{
  Cell[90014, 2735, 1297, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:316306510", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "6"},
   CellID->316306510]},
 "6"->{
  Cell[90014, 2735, 1297, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:316306510", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "6"},
   CellID->316306510]},
 "ID:1250046151"->{
  Cell[91314, 2774, 1011, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1250046151", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "7"},
   CellID->1250046151]},
 "7"->{
  Cell[91314, 2774, 1011, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1250046151", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "7"},
   CellID->1250046151]},
 "ID:304323379"->{
  Cell[92328, 2804, 1123, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:304323379", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "8"},
   CellID->304323379]},
 "8"->{
  Cell[92328, 2804, 1123, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:304323379", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "8"},
   CellID->304323379]},
 "ID:1618605896"->{
  Cell[93454, 2839, 949, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1618605896", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "9"},
   CellID->1618605896]},
 "9"->{
  Cell[93454, 2839, 949, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1618605896", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "9"},
   CellID->1618605896]},
 "Proof|ID:274992099"->{
  Cell[94561, 2876, 6252, 152, 70, "ProofInfo",
   CellTags->{"Proof|ID:274992099", "Proof|ID:274992099-1"},
   CellID->1519173141]},
 "Proof|ID:274992099-1"->{
  Cell[94561, 2876, 6252, 152, 70, "ProofInfo",
   CellTags->{"Proof|ID:274992099", "Proof|ID:274992099-1"},
   CellID->1519173141]},
 "Proof|ID:816763103"->{
  Cell[101334, 3047, 4838, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:816763103", "Proof|ID:816763103-1"},
   CellID->2071197478]},
 "Proof|ID:816763103-1"->{
  Cell[101334, 3047, 4838, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:816763103", "Proof|ID:816763103-1"},
   CellID->2071197478]},
 "Proof|ID:83877220"->{
  Cell[106735, 3191, 7487, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:83877220", "Proof|ID:83877220-1"},
   CellID->1705428850]},
 "Proof|ID:83877220-1"->{
  Cell[106735, 3191, 7487, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:83877220", "Proof|ID:83877220-1"},
   CellID->1705428850]},
 "Proof|ID:1728540744"->{
  Cell[114742, 3388, 6018, 149, 70, "ProofInfo",
   CellTags->{"Proof|ID:1728540744", "Proof|ID:1728540744-1"},
   CellID->115960158]},
 "Proof|ID:1728540744-1"->{
  Cell[114742, 3388, 6018, 149, 70, "ProofInfo",
   CellTags->{"Proof|ID:1728540744", "Proof|ID:1728540744-1"},
   CellID->115960158]},
 "Proof|ID:371484949"->{
  Cell[121282, 3556, 6744, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:371484949", "Proof|ID:371484949-1"},
   CellID->1532696513]},
 "Proof|ID:371484949-1"->{
  Cell[121282, 3556, 6744, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:371484949", "Proof|ID:371484949-1"},
   CellID->1532696513]},
 "Proof|ID:316306510"->{
  Cell[128546, 3735, 6721, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:316306510", "Proof|ID:316306510-1"},
   CellID->293200241]},
 "Proof|ID:316306510-1"->{
  Cell[128546, 3735, 6721, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:316306510", "Proof|ID:316306510-1"},
   CellID->293200241]},
 "Proof|ID:1250046151"->{
  Cell[135882, 3916, 5734, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:1250046151", "Proof|ID:1250046151-1"},
   CellID->143418677]},
 "Proof|ID:1250046151-1"->{
  Cell[135882, 3916, 5734, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:1250046151", "Proof|ID:1250046151-1"},
   CellID->143418677]},
 "Proof|ID:304323379"->{
  Cell[142138, 4076, 4828, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:304323379", "Proof|ID:304323379-1"},
   CellID->1825869525]},
 "Proof|ID:304323379-1"->{
  Cell[142138, 4076, 4828, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:304323379", "Proof|ID:304323379-1"},
   CellID->1825869525]},
 "Proof|ID:1618605896"->{
  Cell[147486, 4219, 4334, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1618605896", "Proof|ID:1618605896-1"},
   CellID->175965684]},
 "Proof|ID:1618605896-1"->{
  Cell[147486, 4219, 4334, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1618605896", "Proof|ID:1618605896-1"},
   CellID->175965684]},
 "ID:514433530"->{
  Cell[152947, 4371, 1200, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:514433530", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isPartIrreflOrder \[ScriptCapitalK]"},
   CellID->514433530]},
 "isPartIrreflOrder \[ScriptCapitalK]"->{
  Cell[152947, 4371, 1200, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:514433530", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isPartIrreflOrder \[ScriptCapitalK]"},
   CellID->514433530]},
 "ID:31238831"->{
  Cell[154150, 4406, 1062, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:31238831", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isNoetherian \[ScriptCapitalK]"},
   CellID->31238831]},
 "isNoetherian \[ScriptCapitalK]"->{
  Cell[154150, 4406, 1062, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:31238831", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isNoetherian \[ScriptCapitalK]"},
   CellID->31238831]},
 "ID:1572875342"->{
  Cell[155215, 4439, 840, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1572875342", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "Multipliers \[ScriptCapitalK]"},
   CellID->1572875342]},
 "Multipliers \[ScriptCapitalK]"->{
  Cell[155215, 4439, 840, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1572875342", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "Multipliers \[ScriptCapitalK]"},
   CellID->1572875342]},
 "ID:907858661"->{
  Cell[156058, 4463, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:907858661", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R0 \[ScriptCapitalK]"},
   CellID->907858661]},
 "R0 \[ScriptCapitalK]"->{
  Cell[156058, 4463, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:907858661", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R0 \[ScriptCapitalK]"},
   CellID->907858661]},
 "ID:1598185495"->{
  Cell[156866, 4486, 808, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1598185495", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R1 \[ScriptCapitalK]"},
   CellID->1598185495]},
 "R1 \[ScriptCapitalK]"->{
  Cell[156866, 4486, 808, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1598185495", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R1 \[ScriptCapitalK]"},
   CellID->1598185495]},
 "ID:325105305"->{
  Cell[157677, 4509, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:325105305", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R2 \[ScriptCapitalK]"},
   CellID->325105305]},
 "R2 \[ScriptCapitalK]"->{
  Cell[157677, 4509, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:325105305", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R2 \[ScriptCapitalK]"},
   CellID->325105305]},
 "ID:1281658915"->{
  Cell[158485, 4532, 810, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1281658915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R3 \[ScriptCapitalK]"},
   CellID->1281658915]},
 "R3 \[ScriptCapitalK]"->{
  Cell[158485, 4532, 810, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1281658915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R3 \[ScriptCapitalK]"},
   CellID->1281658915]},
 "ID:1991263829"->{
  Cell[159298, 4555, 808, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1991263829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R4 \[ScriptCapitalK]"},
   CellID->1991263829]},
 "R4 \[ScriptCapitalK]"->{
  Cell[159298, 4555, 808, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1991263829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R4 \[ScriptCapitalK]"},
   CellID->1991263829]},
 "ID:891013116"->{
  Cell[160109, 4578, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:891013116", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R5 \[ScriptCapitalK]"},
   CellID->891013116]},
 "R5 \[ScriptCapitalK]"->{
  Cell[160109, 4578, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:891013116", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R5 \[ScriptCapitalK]"},
   CellID->891013116]},
 "ID:887229842"->{
  Cell[160917, 4601, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:887229842", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R6 \[ScriptCapitalK]"},
   CellID->887229842]},
 "R6 \[ScriptCapitalK]"->{
  Cell[160917, 4601, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:887229842", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R6 \[ScriptCapitalK]"},
   CellID->887229842]},
 "ID:4782934"->{
  Cell[161725, 4624, 799, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:4782934", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R7 \[ScriptCapitalK]"},
   CellID->4782934]},
 "R7 \[ScriptCapitalK]"->{
  Cell[161725, 4624, 799, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:4782934", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R7 \[ScriptCapitalK]"},
   CellID->4782934]},
 "ID:1656004832"->{
  Cell[162527, 4647, 808, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1656004832", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R8 \[ScriptCapitalK]"},
   CellID->1656004832]},
 "R8 \[ScriptCapitalK]"->{
  Cell[162527, 4647, 808, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1656004832", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R8 \[ScriptCapitalK]"},
   CellID->1656004832]},
 "Proof|ID:514433530"->{
  Cell[163493, 4679, 2716, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:514433530", "Proof|ID:514433530-1"},
   CellID->1256466167]},
 "Proof|ID:514433530-1"->{
  Cell[163493, 4679, 2716, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:514433530", "Proof|ID:514433530-1"},
   CellID->1256466167]},
 "Proof|ID:31238831"->{
  Cell[166729, 4776, 2683, 77, 70, "ProofInfo",
   CellTags->{"Proof|ID:31238831", "Proof|ID:31238831-1"},
   CellID->2034491483]},
 "Proof|ID:31238831-1"->{
  Cell[166729, 4776, 2683, 77, 70, "ProofInfo",
   CellTags->{"Proof|ID:31238831", "Proof|ID:31238831-1"},
   CellID->2034491483]},
 "Proof|ID:1572875342"->{
  Cell[169932, 4872, 2319, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1572875342", "Proof|ID:1572875342-1"},
   CellID->1020709635]},
 "Proof|ID:1572875342-1"->{
  Cell[169932, 4872, 2319, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1572875342", "Proof|ID:1572875342-1"},
   CellID->1020709635]},
 "Proof|ID:907858661"->{
  Cell[172773, 4960, 2273, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:907858661", "Proof|ID:907858661-1"},
   CellID->422286296]},
 "Proof|ID:907858661-1"->{
  Cell[172773, 4960, 2273, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:907858661", "Proof|ID:907858661-1"},
   CellID->422286296]},
 "Proof|ID:1598185495"->{
  Cell[175566, 5047, 2283, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1598185495", "Proof|ID:1598185495-1"},
   CellID->1472209392]},
 "Proof|ID:1598185495-1"->{
  Cell[175566, 5047, 2283, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1598185495", "Proof|ID:1598185495-1"},
   CellID->1472209392]},
 "Proof|ID:325105305"->{
  Cell[178371, 5135, 2273, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:325105305", "Proof|ID:325105305-1"},
   CellID->332550437]},
 "Proof|ID:325105305-1"->{
  Cell[178371, 5135, 2273, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:325105305", "Proof|ID:325105305-1"},
   CellID->332550437]},
 "Proof|ID:1281658915"->{
  Cell[181165, 5222, 2283, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1281658915", "Proof|ID:1281658915-1"},
   CellID->1299588167]},
 "Proof|ID:1281658915-1"->{
  Cell[181165, 5222, 2283, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1281658915", "Proof|ID:1281658915-1"},
   CellID->1299588167]},
 "Proof|ID:1991263829"->{
  Cell[183970, 5310, 2282, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1991263829", "Proof|ID:1991263829-1"},
   CellID->291588382]},
 "Proof|ID:1991263829-1"->{
  Cell[183970, 5310, 2282, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1991263829", "Proof|ID:1991263829-1"},
   CellID->291588382]},
 "Proof|ID:891013116"->{
  Cell[186774, 5398, 2274, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:891013116", "Proof|ID:891013116-1"},
   CellID->1089921470]},
 "Proof|ID:891013116-1"->{
  Cell[186774, 5398, 2274, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:891013116", "Proof|ID:891013116-1"},
   CellID->1089921470]},
 "Proof|ID:887229842"->{
  Cell[189568, 5485, 2273, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:887229842", "Proof|ID:887229842-1"},
   CellID->595600766]},
 "Proof|ID:887229842-1"->{
  Cell[189568, 5485, 2273, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:887229842", "Proof|ID:887229842-1"},
   CellID->595600766]},
 "Proof|ID:4782934"->{
  Cell[192361, 5572, 2263, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:4782934", "Proof|ID:4782934-1"},
   CellID->209477845]},
 "Proof|ID:4782934-1"->{
  Cell[192361, 5572, 2263, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:4782934", "Proof|ID:4782934-1"},
   CellID->209477845]},
 "Proof|ID:1656004832"->{
  Cell[195143, 5659, 2283, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1656004832", "Proof|ID:1656004832-1"},
   CellID->1100102311]},
 "Proof|ID:1656004832-1"->{
  Cell[195143, 5659, 2283, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1656004832", "Proof|ID:1656004832-1"},
   CellID->1100102311]},
 "ID:89016907"->{
  Cell[198419, 5761, 965, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:89016907", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isReductionRing \[ScriptCapitalK]"},
   CellID->89016907]},
 "isReductionRing \[ScriptCapitalK]"->{
  Cell[198419, 5761, 965, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:89016907", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isReductionRing \[ScriptCapitalK]"},
   CellID->89016907]},
 "Proof|ID:89016907"->{
  Cell[199540, 5795, 2335, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:89016907", "Proof|ID:89016907-1"},
   CellID->91423861]},
 "Proof|ID:89016907-1"->{
  Cell[199540, 5795, 2335, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:89016907", "Proof|ID:89016907-1"},
   CellID->91423861]},
 "ID:406426141"->{
  Cell[203204, 5912, 874, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:406426141", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb"},
   CellID->406426141]},
 "ID:648650518"->{
  Cell[204081, 5936, 1457, 46, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:648650518", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "10"},
   CellID->648650518]},
 "10"->{
  Cell[204081, 5936, 1457, 46, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:648650518", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "10"},
   CellID->648650518]},
 "ID:1161287156"->{
  Cell[205541, 5984, 1961, 61, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:1161287156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "11"},
   CellID->1161287156]},
 "11"->{
  Cell[205541, 5984, 1961, 61, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:1161287156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "11"},
   CellID->1161287156]},
 "ID:731335719"->{
  Cell[207505, 6047, 2302, 71, 78, "FormalTextInputFormula",
   CellTags->{
    "ID:731335719", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "12"},
   CellID->731335719]},
 "12"->{
  Cell[207505, 6047, 2302, 71, 78, "FormalTextInputFormula",
   CellTags->{
    "ID:731335719", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "12"},
   CellID->731335719]},
 "ID:424912135"->{
  Cell[209810, 6120, 1415, 47, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:424912135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "13"},
   CellID->424912135]},
 "13"->{
  Cell[209810, 6120, 1415, 47, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:424912135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "13"},
   CellID->424912135]},
 "Proof|ID:648650518"->{
  Cell[211382, 6178, 6210, 154, 28, "ProofInfo",
   CellTags->{"Proof|ID:648650518", "Proof|ID:648650518-1"},
   CellID->1191457334]},
 "Proof|ID:648650518-1"->{
  Cell[211382, 6178, 6210, 154, 28, "ProofInfo",
   CellTags->{"Proof|ID:648650518", "Proof|ID:648650518-1"},
   CellID->1191457334]},
 "Proof|ID:1161287156"->{
  Cell[218112, 6351, 8164, 196, 26, "ProofInfo",
   CellTags->{"Proof|ID:1161287156", "Proof|ID:1161287156-1"},
   CellID->1090967611]},
 "Proof|ID:1161287156-1"->{
  Cell[218112, 6351, 8164, 196, 26, "ProofInfo",
   CellTags->{"Proof|ID:1161287156", "Proof|ID:1161287156-1"},
   CellID->1090967611]},
 "Proof|ID:731335719"->{
  Cell[226797, 6566, 7934, 184, 26, "ProofInfo",
   CellTags->{"Proof|ID:731335719", "Proof|ID:731335719-1"},
   CellID->1170894549]},
 "Proof|ID:731335719-1"->{
  Cell[226797, 6566, 7934, 184, 26, "ProofInfo",
   CellTags->{"Proof|ID:731335719", "Proof|ID:731335719-1"},
   CellID->1170894549]},
 "Proof|ID:424912135"->{
  Cell[235252, 6769, 4778, 122, 26, "ProofInfo",
   CellTags->{"Proof|ID:424912135", "Proof|ID:424912135-1"},
   CellID->1080465504]},
 "Proof|ID:424912135-1"->{
  Cell[235252, 6769, 4778, 122, 26, "ProofInfo",
   CellTags->{"Proof|ID:424912135", "Proof|ID:424912135-1"},
   CellID->1080465504]},
 "ID:504163934"->{
  Cell[241004, 6924, 1665, 56, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:504163934", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "16"},
   CellID->504163934]},
 "16"->{
  Cell[241004, 6924, 1665, 56, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:504163934", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "16"},
   CellID->504163934]},
 "ID:1091730754"->{
  Cell[242672, 6982, 1864, 59, 61, "FormalTextInputFormula",
   CellTags->{
    "ID:1091730754", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "17"},
   CellID->1091730754]},
 "17"->{
  Cell[242672, 6982, 1864, 59, 61, "FormalTextInputFormula",
   CellTags->{
    "ID:1091730754", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "17"},
   CellID->1091730754]},
 "Proof|ID:504163934"->{
  Cell[244693, 7052, 4245, 112, 28, "ProofInfo",
   CellTags->{"Proof|ID:504163934", "Proof|ID:504163934-1"},
   CellID->189456215]},
 "Proof|ID:504163934-1"->{
  Cell[244693, 7052, 4245, 112, 28, "ProofInfo",
   CellTags->{"Proof|ID:504163934", "Proof|ID:504163934-1"},
   CellID->189456215]},
 "Proof|ID:1091730754"->{
  Cell[249458, 7183, 5119, 131, 26, "ProofInfo",
   CellTags->{"Proof|ID:1091730754", "Proof|ID:1091730754-1"},
   CellID->726474309]},
 "Proof|ID:1091730754-1"->{
  Cell[249458, 7183, 5119, 131, 26, "ProofInfo",
   CellTags->{"Proof|ID:1091730754", "Proof|ID:1091730754-1"},
   CellID->726474309]},
 "ID:1818840710"->{
  Cell[255582, 7349, 808, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1818840710", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R9 \[ScriptCapitalK]"},
   CellID->1818840710]},
 "R9 \[ScriptCapitalK]"->{
  Cell[255582, 7349, 808, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1818840710", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R9 \[ScriptCapitalK]"},
   CellID->1818840710]},
 "ID:718799609"->{
  Cell[256393, 7372, 808, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:718799609", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R10 \[ScriptCapitalK]"},
   CellID->718799609]},
 "R10 \[ScriptCapitalK]"->{
  Cell[256393, 7372, 808, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:718799609", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R10 \[ScriptCapitalK]"},
   CellID->718799609]},
 "ID:563907237"->{
  Cell[257204, 7395, 807, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:563907237", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R11 \[ScriptCapitalK]"},
   CellID->563907237]},
 "R11 \[ScriptCapitalK]"->{
  Cell[257204, 7395, 807, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:563907237", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R11 \[ScriptCapitalK]"},
   CellID->563907237]},
 "ID:1162205858"->{
  Cell[258014, 7418, 810, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1162205858", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R12 \[ScriptCapitalK]"},
   CellID->1162205858]},
 "R12 \[ScriptCapitalK]"->{
  Cell[258014, 7418, 810, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1162205858", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R12 \[ScriptCapitalK]"},
   CellID->1162205858]},
 "ID:372596839"->{
  Cell[258827, 7441, 782, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:372596839", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R13 \[ScriptCapitalK]"},
   CellID->372596839]},
 "R13 \[ScriptCapitalK]"->{
  Cell[258827, 7441, 782, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:372596839", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "R13 \[ScriptCapitalK]"},
   CellID->372596839]},
 "Proof|ID:1818840710"->{
  Cell[259765, 7472, 2282, 69, 28, "ProofInfo",
   CellTags->{"Proof|ID:1818840710", "Proof|ID:1818840710-1"},
   CellID->271819944]},
 "Proof|ID:1818840710-1"->{
  Cell[259765, 7472, 2282, 69, 28, "ProofInfo",
   CellTags->{"Proof|ID:1818840710", "Proof|ID:1818840710-1"},
   CellID->271819944]},
 "Proof|ID:718799609"->{
  Cell[262569, 7560, 2277, 68, 26, "ProofInfo",
   CellTags->{"Proof|ID:718799609", "Proof|ID:718799609-1"},
   CellID->300836608]},
 "Proof|ID:718799609-1"->{
  Cell[262569, 7560, 2277, 68, 26, "ProofInfo",
   CellTags->{"Proof|ID:718799609", "Proof|ID:718799609-1"},
   CellID->300836608]},
 "Proof|ID:563907237"->{
  Cell[265366, 7647, 2278, 68, 26, "ProofInfo",
   CellTags->{"Proof|ID:563907237", "Proof|ID:563907237-1"},
   CellID->1260116027]},
 "Proof|ID:563907237-1"->{
  Cell[265366, 7647, 2278, 68, 26, "ProofInfo",
   CellTags->{"Proof|ID:563907237", "Proof|ID:563907237-1"},
   CellID->1260116027]},
 "Proof|ID:1162205858"->{
  Cell[268164, 7734, 2287, 69, 26, "ProofInfo",
   CellTags->{"Proof|ID:1162205858", "Proof|ID:1162205858-1"},
   CellID->1594052474]},
 "Proof|ID:1162205858-1"->{
  Cell[268164, 7734, 2287, 69, 26, "ProofInfo",
   CellTags->{"Proof|ID:1162205858", "Proof|ID:1162205858-1"},
   CellID->1594052474]},
 "Proof|ID:372596839"->{
  Cell[270972, 7822, 2277, 68, 26, "ProofInfo",
   CellTags->{"Proof|ID:372596839", "Proof|ID:372596839-1"},
   CellID->488894158]},
 "Proof|ID:372596839-1"->{
  Cell[270972, 7822, 2277, 68, 26, "ProofInfo",
   CellTags->{"Proof|ID:372596839", "Proof|ID:372596839-1"},
   CellID->488894158]},
 "ID:476788814"->{
  Cell[274237, 7924, 846, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:476788814", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isAlgoReductionRing \[ScriptCapitalK]"},
   CellID->476788814]},
 "isAlgoReductionRing \[ScriptCapitalK]"->{
  Cell[274237, 7924, 846, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:476788814", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields.nb", "isAlgoReductionRing \[ScriptCapitalK]"},
   CellID->476788814]},
 "Proof|ID:476788814"->{
  Cell[275241, 7957, 2358, 70, 28, "ProofInfo",
   CellTags->{"Proof|ID:476788814", "Proof|ID:476788814-1"},
   CellID->1236650353]},
 "Proof|ID:476788814-1"->{
  Cell[275241, 7957, 2358, 70, 28, "ProofInfo",
   CellTags->{"Proof|ID:476788814", "Proof|ID:476788814-1"},
   CellID->1236650353]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"ID:991632779", 315726, 8810},
 {"Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb", \
316002, 8817},
 {"ID:13541845", 329537, 9191},
 {"ID:1796175771", 329752, 9198},
 {"ID:204732646", 329970, 9205},
 {"field \[Element]", 330216, 9212},
 {"ID:125417415", 330458, 9219},
 {"field 0", 330685, 9226},
 {"ID:1874597366", 330918, 9233},
 {"field 1", 331147, 9240},
 {"ID:791499509", 331381, 9247},
 {"field +", 331608, 9254},
 {"ID:1535890097", 331841, 9261},
 {"field - unary", 332082, 9268},
 {"ID:328879666", 332322, 9275},
 {"field -", 332549, 9282},
 {"ID:1643707054", 332782, 9289},
 {"field \[RawStar]", 333030, 9296},
 {"ID:1452171966", 333275, 9303},
 {"field inverse", 333518, 9310},
 {"ID:1982468810", 333761, 9317},
 {"field /", 333991, 9324},
 {"ID:1869336472", 334227, 9331},
 {"field \[Succeeds]", 334478, 9338},
 {"ID:1352206528", 334725, 9345},
 {"field multN", 334964, 9352},
 {"ID:305610078", 335204, 9359},
 {"field mult", 335439, 9366},
 {"ID:1075639810", 335677, 9373},
 {"field mult 1", 335918, 9380},
 {"ID:1151861642", 336160, 9387},
 {"field rdm", 336395, 9394},
 {"ID:2067272590", 336634, 9401},
 {"field lcrd", 336871, 9408},
 {"ID:1497531972", 337111, 9415},
 {"field lcrd unary", 337360, 9422},
 {"ID:529122150", 337605, 9429},
 {"field cpm", 337838, 9436},
 {"ID:1788558801", 338075, 9443},
 {"\[ScriptCapitalQ]", 338325, 9450},
 {"ID:1470999011", 338571, 9457},
 {"ID:853918182", 338790, 9464},
 {"ID:82222568", 339006, 9471},
 {"isCommRing1 \[ScriptCapitalK]", 339276, 9478},
 {"ID:1537851509", 339530, 9485},
 {"isField \[ScriptCapitalK]", 339796, 9492},
 {"Proof|ID:82222568", 340054, 9499},
 {"Proof|ID:82222568-1", 340209, 9503},
 {"Proof|ID:1537851509", 340364, 9507},
 {"Proof|ID:1537851509-1", 340526, 9511},
 {"ID:2102486692", 340680, 9515},
 {"ID:274992099", 340899, 9522},
 {"1", 341116, 9529},
 {"ID:816763103", 341344, 9536},
 {"2", 341561, 9543},
 {"ID:83877220", 341788, 9550},
 {"3", 342003, 9557},
 {"ID:1728540744", 342230, 9564},
 {"4", 342449, 9571},
 {"ID:1733996578", 342680, 9578},
 {"ID:371484949", 342899, 9585},
 {"5", 343116, 9592},
 {"ID:316306510", 343344, 9599},
 {"6", 343561, 9606},
 {"ID:1250046151", 343790, 9613},
 {"7", 344009, 9620},
 {"ID:304323379", 344239, 9627},
 {"8", 344456, 9634},
 {"ID:1618605896", 344685, 9641},
 {"9", 344903, 9648},
 {"Proof|ID:274992099", 345138, 9655},
 {"Proof|ID:274992099-1", 345299, 9659},
 {"Proof|ID:816763103", 345458, 9663},
 {"Proof|ID:816763103-1", 345620, 9667},
 {"Proof|ID:83877220", 345779, 9671},
 {"Proof|ID:83877220-1", 345938, 9675},
 {"Proof|ID:1728540744", 346097, 9679},
 {"Proof|ID:1728540744-1", 346261, 9683},
 {"Proof|ID:371484949", 346422, 9687},
 {"Proof|ID:371484949-1", 346584, 9691},
 {"Proof|ID:316306510", 346744, 9695},
 {"Proof|ID:316306510-1", 346905, 9699},
 {"Proof|ID:1250046151", 347065, 9703},
 {"Proof|ID:1250046151-1", 347229, 9707},
 {"Proof|ID:304323379", 347390, 9711},
 {"Proof|ID:304323379-1", 347552, 9715},
 {"Proof|ID:1618605896", 347713, 9719},
 {"Proof|ID:1618605896-1", 347877, 9723},
 {"ID:514433530", 348032, 9727},
 {"isPartIrreflOrder \[ScriptCapitalK]", 348318, 9734},
 {"ID:31238831", 348580, 9741},
 {"isNoetherian \[ScriptCapitalK]", 348854, 9748},
 {"ID:1572875342", 349111, 9755},
 {"Multipliers \[ScriptCapitalK]", 349386, 9762},
 {"ID:907858661", 349644, 9769},
 {"R0 \[ScriptCapitalK]", 349899, 9776},
 {"ID:1598185495", 350147, 9783},
 {"R1 \[ScriptCapitalK]", 350404, 9790},
 {"ID:325105305", 350653, 9797},
 {"R2 \[ScriptCapitalK]", 350908, 9804},
 {"ID:1281658915", 351156, 9811},
 {"R3 \[ScriptCapitalK]", 351413, 9818},
 {"ID:1991263829", 351663, 9825},
 {"R4 \[ScriptCapitalK]", 351920, 9832},
 {"ID:891013116", 352169, 9839},
 {"R5 \[ScriptCapitalK]", 352424, 9846},
 {"ID:887229842", 352671, 9853},
 {"R6 \[ScriptCapitalK]", 352926, 9860},
 {"ID:4782934", 353171, 9867},
 {"R7 \[ScriptCapitalK]", 353422, 9874},
 {"ID:1656004832", 353666, 9881},
 {"R8 \[ScriptCapitalK]", 353923, 9888},
 {"Proof|ID:514433530", 354178, 9895},
 {"Proof|ID:514433530-1", 354339, 9899},
 {"Proof|ID:31238831", 354497, 9903},
 {"Proof|ID:31238831-1", 354655, 9907},
 {"Proof|ID:1572875342", 354813, 9911},
 {"Proof|ID:1572875342-1", 354977, 9915},
 {"Proof|ID:907858661", 355138, 9919},
 {"Proof|ID:907858661-1", 355298, 9923},
 {"Proof|ID:1598185495", 355457, 9927},
 {"Proof|ID:1598185495-1", 355621, 9931},
 {"Proof|ID:325105305", 355782, 9935},
 {"Proof|ID:325105305-1", 355942, 9939},
 {"Proof|ID:1281658915", 356101, 9943},
 {"Proof|ID:1281658915-1", 356265, 9947},
 {"Proof|ID:1991263829", 356427, 9951},
 {"Proof|ID:1991263829-1", 356590, 9955},
 {"Proof|ID:891013116", 356750, 9959},
 {"Proof|ID:891013116-1", 356911, 9963},
 {"Proof|ID:887229842", 357070, 9967},
 {"Proof|ID:887229842-1", 357230, 9971},
 {"Proof|ID:4782934", 357386, 9975},
 {"Proof|ID:4782934-1", 357540, 9979},
 {"Proof|ID:1656004832", 357695, 9983},
 {"Proof|ID:1656004832-1", 357859, 9987},
 {"ID:89016907", 358013, 9991},
 {"isReductionRing \[ScriptCapitalK]", 358292, 9998},
 {"Proof|ID:89016907", 358555, 10005},
 {"Proof|ID:89016907-1", 358711, 10009},
 {"ID:406426141", 358860, 10013},
 {"ID:648650518", 359078, 10020},
 {"10", 359298, 10027},
 {"ID:1161287156", 359529, 10034},
 {"11", 359751, 10041},
 {"ID:731335719", 359983, 10048},
 {"12", 360203, 10055},
 {"ID:424912135", 360433, 10062},
 {"13", 360653, 10069},
 {"Proof|ID:648650518", 360889, 10076},
 {"Proof|ID:648650518-1", 361051, 10080},
 {"Proof|ID:1161287156", 361212, 10084},
 {"Proof|ID:1161287156-1", 361377, 10088},
 {"Proof|ID:731335719", 361539, 10092},
 {"Proof|ID:731335719-1", 361701, 10096},
 {"Proof|ID:424912135", 361861, 10100},
 {"Proof|ID:424912135-1", 362023, 10104},
 {"ID:504163934", 362177, 10108},
 {"16", 362397, 10115},
 {"ID:1091730754", 362628, 10122},
 {"17", 362850, 10129},
 {"Proof|ID:504163934", 363088, 10136},
 {"Proof|ID:504163934-1", 363249, 10140},
 {"Proof|ID:1091730754", 363409, 10144},
 {"Proof|ID:1091730754-1", 363573, 10148},
 {"ID:1818840710", 363729, 10152},
 {"R9 \[ScriptCapitalK]", 363986, 10159},
 {"ID:718799609", 364235, 10166},
 {"R10 \[ScriptCapitalK]", 364492, 10173},
 {"ID:563907237", 364740, 10180},
 {"R11 \[ScriptCapitalK]", 364997, 10187},
 {"ID:1162205858", 365246, 10194},
 {"R12 \[ScriptCapitalK]", 365505, 10201},
 {"ID:372596839", 365755, 10208},
 {"R13 \[ScriptCapitalK]", 366012, 10215},
 {"Proof|ID:1818840710", 366267, 10222},
 {"Proof|ID:1818840710-1", 366430, 10226},
 {"Proof|ID:718799609", 366590, 10230},
 {"Proof|ID:718799609-1", 366750, 10234},
 {"Proof|ID:563907237", 366908, 10238},
 {"Proof|ID:563907237-1", 367069, 10242},
 {"Proof|ID:1162205858", 367229, 10246},
 {"Proof|ID:1162205858-1", 367393, 10250},
 {"Proof|ID:372596839", 367554, 10254},
 {"Proof|ID:372596839-1", 367714, 10258},
 {"ID:476788814", 367866, 10262},
 {"isAlgoReductionRing \[ScriptCapitalK]", 368155, 10269},
 {"Proof|ID:476788814", 368425, 10276},
 {"Proof|ID:476788814-1", 368586, 10280}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[579, 22, 121, 2, 84, "Title",
 CellID->1754384754],
Cell[CellGroupData[{
Cell[725, 28, 118, 2, 73, "Section",
 CellID->467603975],
Cell[CellGroupData[{
Cell[868, 34, 118, 2, 70, "Subsection",
 CellID->76201264],
Cell[989, 38, 816, 21, 70, "Text",
 CellID->134032288],
Cell[1808, 61, 1920, 34, 70, "Text",
 CellID->930305492]
}, Closed]],
Cell[CellGroupData[{
Cell[3765, 100, 126, 2, 70, "Subsection",
 CellID->52797998],
Cell[3894, 104, 344, 8, 70, "Input",
 CellID->498486718],
Cell[4241, 114, 358, 8, 70, "Input",
 CellID->59508975],
Cell[4602, 124, 374, 8, 70, "Input",
 CellID->75555280],
Cell[4979, 134, 421, 9, 70, "Input",
 CellID->634019458],
Cell[5403, 145, 155, 3, 70, "Input",
 CellID->408174415],
Cell[5561, 150, 155, 3, 70, "Input",
 CellID->251583237]
}, Closed]],
Cell[CellGroupData[{
Cell[5753, 158, 124, 2, 70, "Subsection",
 CellID->1386333768],
Cell[CellGroupData[{
Cell[5902, 164, 356, 10, 70, "Input",
 CellID->1039312110],
Cell[CellGroupData[{
Cell[6283, 178, 597, 11, 70, "Print",
 CellID->1754373566],
Cell[6883, 191, 519, 10, 70, "Print",
 CellID->1029482356],
Cell[7405, 203, 519, 10, 70, "Print",
 CellID->676061299],
Cell[7927, 215, 521, 10, 70, "Print",
 CellID->1850448490],
Cell[8451, 227, 519, 10, 70, "Print",
 CellID->70235299],
Cell[8973, 239, 520, 10, 70, "Print",
 CellID->426922802],
Cell[9496, 251, 521, 10, 70, "Print",
 CellID->1237071357],
Cell[10020, 263, 521, 10, 70, "Print",
 CellID->1870636140],
Cell[10544, 275, 599, 11, 70, "Print",
 CellID->802449962],
Cell[11146, 288, 519, 10, 70, "Print",
 CellID->1864026098],
Cell[11668, 300, 519, 10, 70, "Print",
 CellID->374700447],
Cell[12190, 312, 520, 10, 70, "Print",
 CellID->808617339],
Cell[12713, 324, 580, 11, 70, "Print",
 CellID->39084793],
Cell[13296, 337, 519, 10, 70, "Print",
 CellID->1889518183],
Cell[13818, 349, 520, 10, 70, "Print",
 CellID->2026876725],
Cell[14341, 361, 520, 10, 70, "Print",
 CellID->349652243],
Cell[14864, 373, 519, 10, 70, "Print",
 CellID->1812907063],
Cell[15386, 385, 497, 10, 70, "Print",
 CellID->1437928279]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[15932, 401, 166, 4, 70, "Input",
 CellID->1990427608],
Cell[CellGroupData[{
Cell[16123, 409, 602, 12, 70, "Print",
 CellID->74143475],
Cell[16728, 423, 339, 7, 70, "Print",
 CellID->130980690],
Cell[17070, 432, 142, 2, 70, "Print",
 CellID->631158878],
Cell[17215, 436, 130, 2, 70, "Print",
 CellID->249417802],
Cell[17348, 440, 329, 7, 70, "Print",
 CellID->498183016],
Cell[17680, 449, 331, 7, 70, "Print",
 CellID->710555543],
Cell[18014, 458, 332, 7, 70, "Print",
 CellID->17027108],
Cell[18349, 467, 141, 2, 70, "Print",
 CellID->59531876],
Cell[18493, 471, 120, 2, 70, "Print",
 CellID->54622194]
}, Open  ]]
}, Open  ]],
Cell[18640, 477, 325, 6, 70, "Input",
 CellID->19428626],
Cell[CellGroupData[{
Cell[18990, 487, 541, 12, 70, "Input",
 CellID->1909803754],
Cell[19534, 501, 735, 12, 70, "Output",
 CellID->636966437]
}, Open  ]],
Cell[CellGroupData[{
Cell[20306, 518, 345, 10, 70, "Input",
 CellID->664825817],
Cell[20654, 530, 92, 2, 70, "Output",
 CellID->1831977478]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[20795, 538, 128, 2, 70, "Subsection",
 CellID->1454182661],
Cell[20926, 542, 985, 25, 70, "Input",
 CellID->2074904605],
Cell[21914, 569, 416, 9, 70, "Input",
 CellID->1980565193],
Cell[22333, 580, 169, 3, 70, "Input",
 CellID->444074165],
Cell[22505, 585, 261, 5, 70, "Input",
 CellID->438171343]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[22815, 596, 117, 2, 43, "Section",
 CellID->1664317174],
Cell[CellGroupData[{
Cell[22957, 602, 129, 2, 38, "Subsection",
 CellID->1100887197],
Cell[23089, 606, 597, 13, 70, "Text",
 CellID->899326618],
Cell[23689, 621, 4131, 144, 70, "Text",
 CellID->908816666]
}, Closed]],
Cell[CellGroupData[{
Cell[27857, 770, 120, 2, 30, "Subsection",
 CellID->1612768911],
Cell[27980, 774, 48, 1, 12, "OpenEnvironment",
 CellID->1395386996],
Cell[CellGroupData[{
Cell[28053, 779, 398, 9, 43, "EnvironmentHeader",
 CellID->846251564],
Cell[28454, 790, 641, 16, 30, "GlobalDeclaration",
 CellTags->{
  "ID:991632779", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->991632779],
Cell[29098, 808, 816, 20, 30, "GlobalDeclaration",
 CellTags->{
  "ID:13541845", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->13541845],
Cell[29917, 830, 942, 21, 33, "GlobalDeclaration",
 CellTags->{
  "ID:1796175771", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->1796175771],
Cell[30862, 853, 1198, 34, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:204732646", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field \[Element]"},
 CellID->204732646],
Cell[32063, 889, 828, 23, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:125417415", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field 0"},
 CellID->125417415],
Cell[32894, 914, 881, 24, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1874597366", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field 1"},
 CellID->1874597366],
Cell[33778, 940, 926, 26, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:791499509", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field +"},
 CellID->791499509],
Cell[34707, 968, 931, 26, 28, "FormalTextInputFormula",
 CellTags->{
  "ID:1535890097", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field - unary"},
 CellID->1535890097],
Cell[35641, 996, 924, 26, 28, "FormalTextInputFormula",
 CellTags->{
  "ID:328879666", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field -"},
 CellID->328879666],
Cell[36568, 1024, 928, 26, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1643707054", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field \[RawStar]"},
 CellID->1643707054],
Cell[37499, 1052, 1073, 30, 28, "FormalTextInputFormula",
 CellTags->{
  "ID:1452171966", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field inverse"},
 CellID->1452171966],
Cell[38575, 1084, 928, 26, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:1982468810", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field /"},
 CellID->1982468810],
Cell[39506, 1112, 1412, 43, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:1869336472", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field \[Succeeds]"},
 CellID->1869336472],
Cell[40921, 1157, 1138, 30, 46, "FormalTextInputFormula",
 CellTags->{
  "ID:1352206528", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field multN"},
 CellID->1352206528],
Cell[42062, 1189, 1108, 33, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:305610078", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field mult"},
 CellID->305610078],
Cell[43173, 1224, 1379, 42, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:1075639810", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field mult 1"},
 CellID->1075639810],
Cell[44555, 1268, 1141, 33, 57, "FormalTextInputFormula",
 CellTags->{
  "ID:1151861642", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field rdm"},
 CellID->1151861642],
Cell[45699, 1303, 528, 16, 24, "EnvironmentText",
 CellID->1667267410],
Cell[46230, 1321, 1758, 44, 68, "FormalTextInputFormula",
 CellTags->{
  "ID:2067272590", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field lcrd"},
 CellID->2067272590],
Cell[47991, 1367, 443, 14, 24, "EnvironmentText",
 CellID->2143937706],
Cell[48437, 1383, 1455, 37, 63, "FormalTextInputFormula",
 CellTags->{
  "ID:1497531972", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field lcrd unary"},
 CellID->1497531972],
Cell[49895, 1422, 1279, 34, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:529122150", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "field cpm"},
 CellID->529122150],
Cell[51177, 1458, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1031992887]
}, Open  ]],
Cell[51258, 1462, 48, 1, 16, "CloseEnvironment",
 CellID->879906266]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[51355, 1469, 115, 2, 43, "Section",
 CellID->1178317413],
Cell[51473, 1473, 48, 1, 70, "OpenEnvironment",
 CellID->1922534331],
Cell[CellGroupData[{
Cell[51546, 1478, 465, 12, 70, "EnvironmentHeader",
 CellID->1594116140],
Cell[52014, 1492, 888, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1788558801", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "\[ScriptCapitalQ]"},
 CellID->1788558801],
Cell[52905, 1518, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1329207975]
}, Open  ]],
Cell[52986, 1522, 47, 1, 70, "CloseEnvironment",
 CellID->14854175],
Cell[CellGroupData[{
Cell[53058, 1527, 113, 2, 70, "Subsection",
 CellID->164644379],
Cell[CellGroupData[{
Cell[53196, 1533, 283, 8, 70, "Computation",
 CellID->1133492443],
Cell[53482, 1543, 206, 6, 70, "ComputationResult",
 CellID->868504151]
}, Open  ]],
Cell[CellGroupData[{
Cell[53725, 1554, 378, 10, 70, "ComputationInfo",
 CellID->1196757825],
Cell[54106, 1566, 651, 17, 70, "ComputationInfoBody",
 CellID->1859826640]
}, Closed]],
Cell[CellGroupData[{
Cell[54794, 1588, 237, 6, 70, "Computation",
 CellID->1608921888],
Cell[55034, 1596, 141, 3, 70, "ComputationResult",
 CellID->140517660]
}, Open  ]],
Cell[CellGroupData[{
Cell[55212, 1604, 377, 10, 70, "ComputationInfo",
 CellID->1855696697],
Cell[55592, 1616, 653, 17, 70, "ComputationInfoBody",
 CellID->2086722562]
}, Closed]],
Cell[CellGroupData[{
Cell[56282, 1638, 244, 6, 70, "Computation",
 CellID->1883766832],
Cell[56529, 1646, 139, 3, 70, "ComputationResult",
 CellID->947393027]
}, Open  ]],
Cell[CellGroupData[{
Cell[56705, 1654, 378, 10, 70, "ComputationInfo",
 CellID->1118382957],
Cell[57086, 1666, 651, 17, 70, "ComputationInfoBody",
 CellID->1301382415]
}, Closed]],
Cell[CellGroupData[{
Cell[57774, 1688, 283, 8, 70, "Computation",
 CellID->536728615],
Cell[58060, 1698, 140, 3, 70, "ComputationResult",
 CellID->2053160890]
}, Open  ]],
Cell[CellGroupData[{
Cell[58237, 1706, 377, 10, 70, "ComputationInfo",
 CellID->1664406102],
Cell[58617, 1718, 650, 17, 70, "ComputationInfoBody",
 CellID->1128431625]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[59316, 1741, 113, 2, 70, "Subsection",
 CellID->1583099554],
Cell[CellGroupData[{
Cell[59454, 1747, 303, 8, 70, "Computation",
 CellID->990169324],
Cell[59760, 1757, 232, 7, 70, "ComputationResult",
 CellID->1941861415]
}, Open  ]],
Cell[CellGroupData[{
Cell[60029, 1769, 376, 10, 70, "ComputationInfo",
 CellID->485343236],
Cell[60408, 1781, 650, 17, 70, "ComputationInfoBody",
 CellID->1755171163]
}, Closed]],
Cell[CellGroupData[{
Cell[61095, 1803, 259, 6, 70, "Computation",
 CellID->1956353311],
Cell[61357, 1811, 206, 7, 70, "ComputationResult",
 CellID->1461111709]
}, Open  ]],
Cell[CellGroupData[{
Cell[61600, 1823, 377, 10, 70, "ComputationInfo",
 CellID->733820202],
Cell[61980, 1835, 651, 17, 70, "ComputationInfoBody",
 CellID->1067638783]
}, Closed]],
Cell[CellGroupData[{
Cell[62668, 1857, 259, 6, 70, "Computation",
 CellID->1600026178],
Cell[62930, 1865, 203, 7, 70, "ComputationResult",
 CellID->780050836]
}, Open  ]],
Cell[CellGroupData[{
Cell[63170, 1877, 378, 10, 70, "ComputationInfo",
 CellID->1554263949],
Cell[63551, 1889, 651, 17, 70, "ComputationInfoBody",
 CellID->1663999394]
}, Closed]],
Cell[CellGroupData[{
Cell[64239, 1911, 307, 8, 70, "Computation",
 CellID->1090208346],
Cell[64549, 1921, 202, 7, 70, "ComputationResult",
 CellID->12333322]
}, Open  ]],
Cell[CellGroupData[{
Cell[64788, 1933, 377, 10, 70, "ComputationInfo",
 CellID->128058836],
Cell[65168, 1945, 653, 17, 70, "ComputationInfoBody",
 CellID->1864007722]
}, Closed]],
Cell[CellGroupData[{
Cell[65858, 1967, 239, 6, 70, "Computation",
 CellID->1440655820],
Cell[66100, 1975, 207, 7, 70, "ComputationResult",
 CellID->81088794]
}, Open  ]],
Cell[CellGroupData[{
Cell[66344, 1987, 377, 10, 70, "ComputationInfo",
 CellID->246761544],
Cell[66724, 1999, 652, 17, 70, "ComputationInfoBody",
 CellID->192133376]
}, Closed]],
Cell[CellGroupData[{
Cell[67413, 2021, 188, 5, 70, "Computation",
 CellID->82552198],
Cell[67604, 2028, 204, 7, 70, "ComputationResult",
 CellID->1574628483]
}, Open  ]],
Cell[CellGroupData[{
Cell[67845, 2040, 375, 10, 70, "ComputationInfo",
 CellID->995526489],
Cell[68223, 2052, 650, 17, 70, "ComputationInfoBody",
 CellID->764313443]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[68922, 2075, 112, 2, 70, "Subsection",
 CellID->1292034652],
Cell[CellGroupData[{
Cell[69059, 2081, 314, 8, 70, "Computation",
 CellID->613726844],
Cell[69376, 2091, 297, 10, 70, "ComputationResult",
 CellID->1299536985]
}, Open  ]],
Cell[CellGroupData[{
Cell[69710, 2106, 377, 10, 70, "ComputationInfo",
 CellID->1784347710],
Cell[70090, 2118, 651, 17, 70, "ComputationInfoBody",
 CellID->728776616]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[70802, 2142, 268, 6, 43, "Section",
 CellID->2031767909],
Cell[71073, 2150, 671, 16, 33, "GlobalDeclaration",
 CellTags->{
  "ID:1470999011", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->1470999011],
Cell[71747, 2168, 37, 1, 24, "Text",
 CellID->1259024642],
Cell[71787, 2171, 756, 18, 33, "GlobalDeclaration",
 CellTags->{
  "ID:853918182", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->853918182],
Cell[CellGroupData[{
Cell[72568, 2193, 183, 3, 38, "Subsection",
 CellID->1912339895],
Cell[72754, 2198, 48, 1, 70, "OpenEnvironment",
 CellID->1605366551],
Cell[CellGroupData[{
Cell[72827, 2203, 393, 9, 70, "EnvironmentHeader",
 CellID->704319825],
Cell[73223, 2214, 978, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:82222568", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "isCommRing1 \[ScriptCapitalK]"},
 CellID->82222568],
Cell[74204, 2240, 801, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1537851509", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "isField \[ScriptCapitalK]"},
 CellID->1537851509],
Cell[75008, 2263, 66, 1, 70, "EndEnvironmentMarker",
 CellID->2009629793]
}, Open  ]],
Cell[75089, 2267, 49, 1, 70, "CloseEnvironment",
 CellID->1116315797],
Cell[CellGroupData[{
Cell[75163, 2272, 2303, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:82222568", "Proof|ID:82222568-1"},
 CellID->83229303],
Cell[77469, 2342, 479, 12, 70, "ProofInfoBody",
 CellID->229712744]
}, Closed]],
Cell[CellGroupData[{
Cell[77985, 2359, 2302, 69, 70, "ProofInfo",
 CellTags->{"Proof|ID:1537851509", "Proof|ID:1537851509-1"},
 CellID->710651429],
Cell[80290, 2430, 481, 12, 70, "ProofInfoBody",
 CellID->746747984]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[80820, 2448, 190, 3, 30, "Subsection",
 CellID->1445748822],
Cell[81013, 2453, 47, 1, 70, "OpenEnvironment",
 CellID->596016869],
Cell[CellGroupData[{
Cell[81085, 2458, 425, 11, 70, "EnvironmentHeader",
 CellID->1128420351],
Cell[81513, 2471, 877, 22, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2102486692", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->2102486692],
Cell[82393, 2495, 1359, 41, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:274992099", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "1"},
 CellID->274992099],
Cell[83755, 2538, 1128, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:816763103", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "2"},
 CellID->816763103],
Cell[84886, 2573, 1655, 54, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:83877220", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "3"},
 CellID->83877220],
Cell[86544, 2629, 1491, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1728540744", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "4"},
 CellID->1728540744],
Cell[88038, 2678, 660, 16, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1733996578", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->1733996578],
Cell[88701, 2696, 1310, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:371484949", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "5"},
 CellID->371484949],
Cell[90014, 2735, 1297, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:316306510", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "6"},
 CellID->316306510],
Cell[91314, 2774, 1011, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1250046151", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "7"},
 CellID->1250046151],
Cell[92328, 2804, 1123, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:304323379", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "8"},
 CellID->304323379],
Cell[93454, 2839, 949, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1618605896", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "9"},
 CellID->1618605896],
Cell[94406, 2867, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1503924104]
}, Open  ]],
Cell[94487, 2871, 49, 1, 70, "CloseEnvironment",
 CellID->1855794389],
Cell[CellGroupData[{
Cell[94561, 2876, 6252, 152, 70, "ProofInfo",
 CellTags->{"Proof|ID:274992099", "Proof|ID:274992099-1"},
 CellID->1519173141],
Cell[100816, 3030, 481, 12, 70, "ProofInfoBody",
 CellID->1937965720]
}, Closed]],
Cell[CellGroupData[{
Cell[101334, 3047, 4838, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:816763103", "Proof|ID:816763103-1"},
 CellID->2071197478],
Cell[106175, 3173, 523, 13, 70, "ProofInfoBody",
 CellID->1092343128]
}, Closed]],
Cell[CellGroupData[{
Cell[106735, 3191, 7487, 178, 70, "ProofInfo",
 CellTags->{"Proof|ID:83877220", "Proof|ID:83877220-1"},
 CellID->1705428850],
Cell[114225, 3371, 480, 12, 70, "ProofInfoBody",
 CellID->1125367110]
}, Closed]],
Cell[CellGroupData[{
Cell[114742, 3388, 6018, 149, 70, "ProofInfo",
 CellTags->{"Proof|ID:1728540744", "Proof|ID:1728540744-1"},
 CellID->115960158],
Cell[120763, 3539, 482, 12, 70, "ProofInfoBody",
 CellID->1053170671]
}, Closed]],
Cell[CellGroupData[{
Cell[121282, 3556, 6744, 160, 70, "ProofInfo",
 CellTags->{"Proof|ID:371484949", "Proof|ID:371484949-1"},
 CellID->1532696513],
Cell[128029, 3718, 480, 12, 70, "ProofInfoBody",
 CellID->623568866]
}, Closed]],
Cell[CellGroupData[{
Cell[128546, 3735, 6721, 160, 70, "ProofInfo",
 CellTags->{"Proof|ID:316306510", "Proof|ID:316306510-1"},
 CellID->293200241],
Cell[135270, 3897, 575, 14, 70, "ProofInfoBody",
 CellID->1149536317]
}, Closed]],
Cell[CellGroupData[{
Cell[135882, 3916, 5734, 141, 70, "ProofInfo",
 CellTags->{"Proof|ID:1250046151", "Proof|ID:1250046151-1"},
 CellID->143418677],
Cell[141619, 4059, 482, 12, 70, "ProofInfoBody",
 CellID->1204687680]
}, Closed]],
Cell[CellGroupData[{
Cell[142138, 4076, 4828, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:304323379", "Proof|ID:304323379-1"},
 CellID->1825869525],
Cell[146969, 4202, 480, 12, 70, "ProofInfoBody",
 CellID->403531515]
}, Closed]],
Cell[CellGroupData[{
Cell[147486, 4219, 4334, 115, 70, "ProofInfo",
 CellTags->{"Proof|ID:1618605896", "Proof|ID:1618605896-1"},
 CellID->175965684],
Cell[151823, 4336, 482, 12, 70, "ProofInfoBody",
 CellID->1012531353]
}, Closed]],
Cell[152320, 4351, 48, 1, 70, "OpenEnvironment",
 CellID->1342787779],
Cell[CellGroupData[{
Cell[152393, 4356, 551, 13, 70, "EnvironmentHeader",
 CellID->109341845],
Cell[152947, 4371, 1200, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:514433530", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "isPartIrreflOrder \[ScriptCapitalK]"},
 CellID->514433530],
Cell[154150, 4406, 1062, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:31238831", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "isNoetherian \[ScriptCapitalK]"},
 CellID->31238831],
Cell[155215, 4439, 840, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1572875342", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "Multipliers \[ScriptCapitalK]"},
 CellID->1572875342],
Cell[156058, 4463, 805, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:907858661", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R0 \[ScriptCapitalK]"},
 CellID->907858661],
Cell[156866, 4486, 808, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1598185495", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R1 \[ScriptCapitalK]"},
 CellID->1598185495],
Cell[157677, 4509, 805, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:325105305", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R2 \[ScriptCapitalK]"},
 CellID->325105305],
Cell[158485, 4532, 810, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1281658915", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R3 \[ScriptCapitalK]"},
 CellID->1281658915],
Cell[159298, 4555, 808, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1991263829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R4 \[ScriptCapitalK]"},
 CellID->1991263829],
Cell[160109, 4578, 805, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:891013116", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R5 \[ScriptCapitalK]"},
 CellID->891013116],
Cell[160917, 4601, 805, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:887229842", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R6 \[ScriptCapitalK]"},
 CellID->887229842],
Cell[161725, 4624, 799, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:4782934", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R7 \[ScriptCapitalK]"},
 CellID->4782934],
Cell[162527, 4647, 808, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1656004832", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R8 \[ScriptCapitalK]"},
 CellID->1656004832],
Cell[163338, 4670, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1455483400]
}, Open  ]],
Cell[163419, 4674, 49, 1, 70, "CloseEnvironment",
 CellID->1061646154],
Cell[CellGroupData[{
Cell[163493, 4679, 2716, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:514433530", "Proof|ID:514433530-1"},
 CellID->1256466167],
Cell[166212, 4759, 480, 12, 70, "ProofInfoBody",
 CellID->717504818]
}, Closed]],
Cell[CellGroupData[{
Cell[166729, 4776, 2683, 77, 70, "ProofInfo",
 CellTags->{"Proof|ID:31238831", "Proof|ID:31238831-1"},
 CellID->2034491483],
Cell[169415, 4855, 480, 12, 70, "ProofInfoBody",
 CellID->1669813662]
}, Closed]],
Cell[CellGroupData[{
Cell[169932, 4872, 2319, 69, 70, "ProofInfo",
 CellTags->{"Proof|ID:1572875342", "Proof|ID:1572875342-1"},
 CellID->1020709635],
Cell[172254, 4943, 482, 12, 70, "ProofInfoBody",
 CellID->1654379224]
}, Closed]],
Cell[CellGroupData[{
Cell[172773, 4960, 2273, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:907858661", "Proof|ID:907858661-1"},
 CellID->422286296],
Cell[175049, 5030, 480, 12, 70, "ProofInfoBody",
 CellID->143092594]
}, Closed]],
Cell[CellGroupData[{
Cell[175566, 5047, 2283, 69, 70, "ProofInfo",
 CellTags->{"Proof|ID:1598185495", "Proof|ID:1598185495-1"},
 CellID->1472209392],
Cell[177852, 5118, 482, 12, 70, "ProofInfoBody",
 CellID->1261820315]
}, Closed]],
Cell[CellGroupData[{
Cell[178371, 5135, 2273, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:325105305", "Proof|ID:325105305-1"},
 CellID->332550437],
Cell[180647, 5205, 481, 12, 70, "ProofInfoBody",
 CellID->1985798367]
}, Closed]],
Cell[CellGroupData[{
Cell[181165, 5222, 2283, 69, 70, "ProofInfo",
 CellTags->{"Proof|ID:1281658915", "Proof|ID:1281658915-1"},
 CellID->1299588167],
Cell[183451, 5293, 482, 12, 70, "ProofInfoBody",
 CellID->1806825968]
}, Closed]],
Cell[CellGroupData[{
Cell[183970, 5310, 2282, 69, 70, "ProofInfo",
 CellTags->{"Proof|ID:1991263829", "Proof|ID:1991263829-1"},
 CellID->291588382],
Cell[186255, 5381, 482, 12, 70, "ProofInfoBody",
 CellID->2022277391]
}, Closed]],
Cell[CellGroupData[{
Cell[186774, 5398, 2274, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:891013116", "Proof|ID:891013116-1"},
 CellID->1089921470],
Cell[189051, 5468, 480, 12, 70, "ProofInfoBody",
 CellID->763023223]
}, Closed]],
Cell[CellGroupData[{
Cell[189568, 5485, 2273, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:887229842", "Proof|ID:887229842-1"},
 CellID->595600766],
Cell[191844, 5555, 480, 12, 70, "ProofInfoBody",
 CellID->153744125]
}, Closed]],
Cell[CellGroupData[{
Cell[192361, 5572, 2263, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:4782934", "Proof|ID:4782934-1"},
 CellID->209477845],
Cell[194627, 5642, 479, 12, 70, "ProofInfoBody",
 CellID->1003838192]
}, Closed]],
Cell[CellGroupData[{
Cell[195143, 5659, 2283, 69, 70, "ProofInfo",
 CellTags->{"Proof|ID:1656004832", "Proof|ID:1656004832-1"},
 CellID->1100102311],
Cell[197429, 5730, 482, 12, 70, "ProofInfoBody",
 CellID->1570103634]
}, Closed]],
Cell[197926, 5745, 48, 1, 70, "OpenEnvironment",
 CellID->2080814435],
Cell[CellGroupData[{
Cell[197999, 5750, 417, 9, 70, "EnvironmentHeader",
 CellID->1560907237],
Cell[198419, 5761, 965, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:89016907", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "isReductionRing \[ScriptCapitalK]"},
 CellID->89016907],
Cell[199387, 5786, 64, 1, 70, "EndEnvironmentMarker",
 CellID->58748202]
}, Open  ]],
Cell[199466, 5790, 49, 1, 70, "CloseEnvironment",
 CellID->1288020955],
Cell[CellGroupData[{
Cell[199540, 5795, 2335, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:89016907", "Proof|ID:89016907-1"},
 CellID->91423861],
Cell[201878, 5867, 479, 12, 70, "ProofInfoBody",
 CellID->160819088]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[202406, 5885, 210, 5, 30, "Subsection",
 CellID->1867114071],
Cell[202619, 5892, 48, 1, 12, "OpenEnvironment",
 CellID->1232421138],
Cell[CellGroupData[{
Cell[202692, 5897, 509, 13, 43, "EnvironmentHeader",
 CellID->1198258968],
Cell[203204, 5912, 874, 22, 42, "GlobalDeclaration",
 CellTags->{
  "ID:406426141", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb"},
 CellID->406426141],
Cell[204081, 5936, 1457, 46, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:648650518", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "10"},
 CellID->648650518],
Cell[205541, 5984, 1961, 61, 52, "FormalTextInputFormula",
 CellTags->{
  "ID:1161287156", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "11"},
 CellID->1161287156],
Cell[207505, 6047, 2302, 71, 78, "FormalTextInputFormula",
 CellTags->{
  "ID:731335719", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "12"},
 CellID->731335719],
Cell[209810, 6120, 1415, 47, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:424912135", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "13"},
 CellID->424912135],
Cell[211228, 6169, 65, 1, 26, "EndEnvironmentMarker",
 CellID->691543564]
}, Open  ]],
Cell[211308, 6173, 49, 1, 16, "CloseEnvironment",
 CellID->1614400098],
Cell[CellGroupData[{
Cell[211382, 6178, 6210, 154, 28, "ProofInfo",
 CellTags->{"Proof|ID:648650518", "Proof|ID:648650518-1"},
 CellID->1191457334],
Cell[217595, 6334, 480, 12, 70, "ProofInfoBody",
 CellID->401303854]
}, Closed]],
Cell[CellGroupData[{
Cell[218112, 6351, 8164, 196, 26, "ProofInfo",
 CellTags->{"Proof|ID:1161287156", "Proof|ID:1161287156-1"},
 CellID->1090967611],
Cell[226279, 6549, 481, 12, 70, "ProofInfoBody",
 CellID->929802187]
}, Closed]],
Cell[CellGroupData[{
Cell[226797, 6566, 7934, 184, 26, "ProofInfo",
 CellTags->{"Proof|ID:731335719", "Proof|ID:731335719-1"},
 CellID->1170894549],
Cell[234734, 6752, 481, 12, 70, "ProofInfoBody",
 CellID->1475281485]
}, Closed]],
Cell[CellGroupData[{
Cell[235252, 6769, 4778, 122, 26, "ProofInfo",
 CellTags->{"Proof|ID:424912135", "Proof|ID:424912135-1"},
 CellID->1080465504],
Cell[240033, 6893, 480, 12, 70, "ProofInfoBody",
 CellID->802282293]
}, Closed]],
Cell[240528, 6908, 48, 1, 12, "OpenEnvironment",
 CellID->1420224099],
Cell[CellGroupData[{
Cell[240601, 6913, 400, 9, 43, "EnvironmentHeader",
 CellID->784653472],
Cell[241004, 6924, 1665, 56, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:504163934", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "16"},
 CellID->504163934],
Cell[242672, 6982, 1864, 59, 61, "FormalTextInputFormula",
 CellTags->{
  "ID:1091730754", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "17"},
 CellID->1091730754],
Cell[244539, 7043, 65, 1, 26, "EndEnvironmentMarker",
 CellID->805694964]
}, Open  ]],
Cell[244619, 7047, 49, 1, 16, "CloseEnvironment",
 CellID->1983325194],
Cell[CellGroupData[{
Cell[244693, 7052, 4245, 112, 28, "ProofInfo",
 CellTags->{"Proof|ID:504163934", "Proof|ID:504163934-1"},
 CellID->189456215],
Cell[248941, 7166, 480, 12, 70, "ProofInfoBody",
 CellID->251515074]
}, Closed]],
Cell[CellGroupData[{
Cell[249458, 7183, 5119, 131, 26, "ProofInfo",
 CellTags->{"Proof|ID:1091730754", "Proof|ID:1091730754-1"},
 CellID->726474309],
Cell[254580, 7316, 480, 12, 70, "ProofInfoBody",
 CellID->19880576]
}, Closed]],
Cell[255075, 7331, 46, 1, 12, "OpenEnvironment",
 CellID->13625358],
Cell[CellGroupData[{
Cell[255146, 7336, 433, 11, 43, "EnvironmentHeader",
 CellID->1399361628],
Cell[255582, 7349, 808, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1818840710", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R9 \[ScriptCapitalK]"},
 CellID->1818840710],
Cell[256393, 7372, 808, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:718799609", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R10 \[ScriptCapitalK]"},
 CellID->718799609],
Cell[257204, 7395, 807, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:563907237", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R11 \[ScriptCapitalK]"},
 CellID->563907237],
Cell[258014, 7418, 810, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1162205858", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R12 \[ScriptCapitalK]"},
 CellID->1162205858],
Cell[258827, 7441, 782, 20, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:372596839", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "R13 \[ScriptCapitalK]"},
 CellID->372596839],
Cell[259612, 7463, 65, 1, 26, "EndEnvironmentMarker",
 CellID->940321384]
}, Open  ]],
Cell[259692, 7467, 48, 1, 16, "CloseEnvironment",
 CellID->144901910],
Cell[CellGroupData[{
Cell[259765, 7472, 2282, 69, 28, "ProofInfo",
 CellTags->{"Proof|ID:1818840710", "Proof|ID:1818840710-1"},
 CellID->271819944],
Cell[262050, 7543, 482, 12, 70, "ProofInfoBody",
 CellID->1423115614]
}, Closed]],
Cell[CellGroupData[{
Cell[262569, 7560, 2277, 68, 26, "ProofInfo",
 CellTags->{"Proof|ID:718799609", "Proof|ID:718799609-1"},
 CellID->300836608],
Cell[264849, 7630, 480, 12, 70, "ProofInfoBody",
 CellID->191748015]
}, Closed]],
Cell[CellGroupData[{
Cell[265366, 7647, 2278, 68, 26, "ProofInfo",
 CellTags->{"Proof|ID:563907237", "Proof|ID:563907237-1"},
 CellID->1260116027],
Cell[267647, 7717, 480, 12, 70, "ProofInfoBody",
 CellID->625079081]
}, Closed]],
Cell[CellGroupData[{
Cell[268164, 7734, 2287, 69, 26, "ProofInfo",
 CellTags->{"Proof|ID:1162205858", "Proof|ID:1162205858-1"},
 CellID->1594052474],
Cell[270454, 7805, 481, 12, 70, "ProofInfoBody",
 CellID->475932409]
}, Closed]],
Cell[CellGroupData[{
Cell[270972, 7822, 2277, 68, 26, "ProofInfo",
 CellTags->{"Proof|ID:372596839", "Proof|ID:372596839-1"},
 CellID->488894158],
Cell[273252, 7892, 480, 12, 70, "ProofInfoBody",
 CellID->850600494]
}, Closed]],
Cell[273747, 7907, 47, 1, 12, "OpenEnvironment",
 CellID->228270346],
Cell[CellGroupData[{
Cell[273819, 7912, 415, 10, 43, "EnvironmentHeader",
 CellID->1906185036],
Cell[274237, 7924, 846, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:476788814", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields.nb",
    "isAlgoReductionRing \[ScriptCapitalK]"},
 CellID->476788814],
Cell[275086, 7948, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1565750311]
}, Open  ]],
Cell[275167, 7952, 49, 1, 16, "CloseEnvironment",
 CellID->1552441335],
Cell[CellGroupData[{
Cell[275241, 7957, 2358, 70, 28, "ProofInfo",
 CellTags->{"Proof|ID:476788814", "Proof|ID:476788814-1"},
 CellID->1236650353],
Cell[277602, 8029, 480, 12, 70, "ProofInfoBody",
 CellID->399055891]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
