(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[    828713,      23327]
NotebookOptionsPosition[    592206,      17177]
NotebookOutlinePosition[    762949,      21357]
CellTagsIndexPosition[    746152,      20952]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Reduction Ring of Integers", "Title",
 CellChangeTimes->{{3.64552035821696*^9, 3.645520367119165*^9}, {
   3.657347305788457*^9, 3.657347311076866*^9}, 3.660906325279291*^9},
 CellID->2066005206],

Cell[CellGroupData[{

Cell["Preliminaries", "Section",
 CellChangeTimes->{{3.592043442336009*^9, 3.592043446608702*^9}, {
  3.645520429206111*^9, 3.645520432005458*^9}},
 CellID->1422121268],

Cell[CellGroupData[{

Cell["Description", "Subsection",
 CellChangeTimes->{{3.660906164131008*^9, 3.6609061661434116`*^9}},
 CellID->163575360],

Cell["\<\
This notebook contains the definition of the algorithmic reduction ring of \
integers, i.e. the domain that extends \[DoubleStruckCapitalZ] by the various \
operations necessary for being an algorithmic reduction ring. Section \
\[OpenCurlyDoubleQuote]ReductionIntegers Constitutes Reduction Ring\
\[CloseCurlyDoubleQuote] proves that the domain thus obtained really belongs \
to the class of algorithmic reduction rings.\
\>", "Text",
 CellChangeTimes->{{3.6609061733974247`*^9, 3.660906289258828*^9}},
 CellID->597779389]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Packages", "Subsection",
 CellChangeTimes->{{3.655465515143981*^9, 3.6554655180611863`*^9}},
 CellID->376409692],

Cell[BoxData[
 RowBox[{"$PackageDirectory", "=", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{
    "$RootDirectory", ",", "\"\<scratch\>\"", ",", "\"\<amaletzk\>\"", ",", 
     "\"\<Theorema Environment Data\>\""}], "}"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.666438783233775*^9, 3.666438821083324*^9}},
 CellID->498486718],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<LogicSets.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.656670589533224*^9, 3.6566705941196322`*^9}, {
   3.656767086897564*^9, 3.656767101545989*^9}, 3.666440168681233*^9},
 CellID->406742757],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<AlgebraicStructures.m\>\""}], 
    "}"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.656670589533224*^9, 3.6566705941196322`*^9}, {
   3.656767086897564*^9, 3.656767101545989*^9}, {3.656771002965444*^9, 
   3.6567710077546525`*^9}, 3.666440169697361*^9},
 CellID->432730220],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<NatInt.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.656670589533224*^9, 3.6566705941196322`*^9}, {
   3.656767086897564*^9, 3.656767101545989*^9}, {3.6572537957223663`*^9, 
   3.6572537969235687`*^9}, 3.666440170665226*^9},
 CellID->135231296],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Tuples.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.656670589533224*^9, 3.6566705941196322`*^9}, {
   3.656767086897564*^9, 3.656767101545989*^9}, {3.656771002965444*^9, 
   3.6567710077546525`*^9}, {3.657253798935972*^9, 3.6572537999187737`*^9}, 
   3.6664401717212963`*^9},
 CellID->45708515],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<ReductionRings.m\>\""}], "}"}], 
   "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.656670589533224*^9, 3.6566705941196322`*^9}, {
   3.656767086897564*^9, 3.656767101545989*^9}, {3.656771002965444*^9, 
   3.6567710077546525`*^9}, {3.657253798935972*^9, 3.6572537999187737`*^9}, {
   3.657347326442893*^9, 3.6573473290792975`*^9}, 3.666440172737281*^9},
 CellID->851969895],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Sequences.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.656670589533224*^9, 3.6566705941196322`*^9}, {
   3.656767086897564*^9, 3.656767101545989*^9}, {3.656771002965444*^9, 
   3.6567710077546525`*^9}, {3.657253798935972*^9, 3.6572537999187737`*^9}, {
   3.657347326442893*^9, 3.6573473290792975`*^9}, {3.6579560193401475`*^9, 
   3.6579560210873504`*^9}, 3.666440173841474*^9},
 CellID->559226997],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.655465521727193*^9, 3.655465525096799*^9}},
 CellID->524928492],

Cell[BoxData[
 RowBox[{"MakeTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6554655291060057`*^9, 3.6554655322416115`*^9}},
 CellID->31934633],

Cell[BoxData[
 RowBox[{"UpdateTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6554655291060057`*^9, 3.6554655322416115`*^9}, {
  3.661852486242468*^9, 3.6618524872096696`*^9}},
 CellID->113843512]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Analysis", "Subsection",
 CellChangeTimes->{{3.640668830437483*^9, 3.640668834318727*^9}},
 CellID->1386333768],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"AnalyzeProofs", "[", 
   RowBox[{"{", 
    RowBox[{
    "\"\<Theory\>\"", ",", 
     "\"\</home/amaletzk/Theorema Notebooks/ElementaryTheories\>\""}], "}"}], 
   "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.6406688560104437`*^9, 3.640668867513785*^9}, 
   3.64067664060292*^9, 3.644222019629964*^9},
 CellID->1039312110],

Cell[CellGroupData[{

Cell[BoxData["\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/Reduction Ring Theory/Polynomials\\\" (1 of 3)\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQBgC8GYnL
  "],
 CellID->1754373566],

Cell[BoxData["\<\"Reading proof 1 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyC9U8PG0g1IAwC334k1
  "],
 CellID->1029482356],

Cell[BoxData["\<\"Reading proof 51 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyCta+Jn6QakAbQ/iMc=
  "],
 CellID->676061299],

Cell[BoxData["\<\"Reading proof 101 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyCtL5pk6QakAbPuiL4=
  "],
 CellID->1850448490],

Cell[BoxData["\<\"Reading proof 151 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyDtw1Fm6QakAbTziOI=
  "],
 CellID->70235299],

Cell[BoxData["\<\"Reading proof 201 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyBtNrHL0g1IAwC4eolp
  "],
 CellID->426922802],

Cell[BoxData["\<\"Reading proof 251 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyD9WXuhpRuQBgC8IonX
  "],
 CellID->1237071357],

Cell[BoxData["\<\"Reading proof 301 of 304 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyDtsH+zpRuQBgC7AonK
  "],
 CellID->1870636140],

Cell[BoxData["\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/Reduction Ring Theory/ReductionRings\\\" (2 of 3)\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyAddGObpRuQBgC8U4n4
  "],
 CellID->802449962],

Cell[BoxData["\<\"Reading proof 1 of 132 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB91nq7pRuQBgC75onX
  "],
 CellID->1864026098],

Cell[BoxData["\<\"Reading proof 51 of 132 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyC97/YFSzcgDQDAZIqB
  "],
 CellID->374700447],

Cell[BoxData["\<\"Reading proof 101 of 132 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9QfqdpRuQBgC8aInx
  "],
 CellID->808617339],

Cell[BoxData["\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories\\\" (3 of 3)\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9s+qfpRuQBgC/qYpp
  "],
 CellID->39084793],

Cell[BoxData["\<\"Reading proof 1 of 122 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyBteI/Ryg1IAwC3PIkp
  "],
 CellID->1889518183],

Cell[BoxData["\<\"Reading proof 51 of 122 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB95riQlRuQBgC72Ym+
  "],
 CellID->2026876725],

Cell[BoxData["\<\"Reading proof 101 of 122 ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyD9YIm6lRuQBgC8AonE
  "],
 CellID->349652243],

Cell[BoxData["\<\"Creating forward rules ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyC9RsbAyg1IAwC24IkR
  "],
 CellID->1812907063],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQA+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyC94Z+PlRuQBgC91ooT
  "],
 CellID->1437928279]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"ComputeFormulaPositions", "[", "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.6406695270485573`*^9, 3.6406695415951242`*^9}, {
   3.64067664934828*^9, 3.640676663567005*^9}, 3.6412902845300865`*^9, 
   3.641290346822996*^9, {3.6442219079512787`*^9, 3.644221909273131*^9}},
 CellID->19428626],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CheckIntegrity", "[", 
  RowBox[{"\"\<CheckPositions\>\"", "\[Rule]", 
   RowBox[{"{", 
    RowBox[{"FileNameJoin", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"CurrentValue", "[", "\"\<NotebookDirectory\>\"", "]"}], ",", 
       RowBox[{"CurrentValue", "[", "\"\<NotebookFileName\>\"", "]"}]}], 
      "}"}], "]"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.64067003109768*^9, 3.6406700374311132`*^9}, {
   3.6406797903560762`*^9, 3.640679792950843*^9}, 3.6442220291527653`*^9},
 CellID->1909803754],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.640670040858255*^9, 3.640670099679185*^9, 3.640670397721546*^9, 
   3.640679793300535*^9, 3.6406798406436234`*^9, 3.640679891162312*^9, 
   3.640679971009297*^9, 3.64068009182163*^9, 3.640680173340745*^9, 
   3.641037395456129*^9, 3.641116403388618*^9, 3.641286881014108*^9, 
   3.642225466578094*^9, 3.642234672070362*^9, 3.642240151894621*^9, 
   3.6423063269273643`*^9, {3.642308832539488*^9, 3.642308857563932*^9}, 
   3.642321396697977*^9, 3.6423285030212173`*^9, 3.642493582768566*^9, 
   3.642763270724422*^9, 3.642834604308413*^9, 3.643428832901271*^9, 
   3.6434441292624083`*^9, 3.643532851309215*^9, 3.644918543299676*^9, 
   3.645518304732339*^9},
 CellID->636966437]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Auxiliary Functions", "Subsection",
 CellChangeTimes->{{3.640668579267561*^9, 3.640668583347287*^9}},
 CellID->1454182661],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "LoadPackage", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", "name_String", "]"}], ":=", 
   RowBox[{"LoadPackage", "[", 
    RowBox[{"{", "name", "}"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", 
    RowBox[{"l", ":", 
     RowBox[{"{", "___String", "}"}]}], "]"}], ":=", 
   RowBox[{"Scan", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{"FileNameJoin", "[", 
        RowBox[{"{", 
         RowBox[{"$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
          RowBox[{"#", "<>", "\"\<.m\>\""}]}], "}"}], "]"}], "]"}], "&"}], 
     ",", "l"}], "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640668603629292*^9, 3.640668727460284*^9}, 
   3.641286816086794*^9, 3.6434459961973133`*^9, {3.657253631526078*^9, 
   3.657253637922089*^9}, {3.66644018328414*^9, 3.666440189828239*^9}},
 CellID->2074904605],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "FindFormula", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"FindFormula", "[", "label_String", "]"}], ":=", 
   RowBox[{"NotebookLocate", "[", "label", "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640677756847624*^9, 3.6406777896243896`*^9}, 
   3.640677822869519*^9, {3.640677854902379*^9, 3.6406778579321823`*^9}},
 CellID->1980565193],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<ProofPalette\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.640668742006607*^9, 3.640668751417616*^9}},
 CellID->444074165],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<TheoryTools/TheoryTools\>\"", 
  "]"}]], "Input",
 CellChangeTimes->{{3.640668759133181*^9, 3.640668768089197*^9}, {
  3.657253652898115*^9, 3.6572536538965173`*^9}, {3.666440196848494*^9, 
  3.66644019923251*^9}},
 CellID->6535056]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Definitions", "Section",
 CellChangeTimes->{{3.592043452298074*^9, 3.592043454336059*^9}, 
   3.645520415673079*^9, 3.645520788276498*^9},
 CellID->2107536411],

Cell[CellGroupData[{

Cell["Informal Explanation", "Subsection",
 CellChangeTimes->{{3.6455207802308273`*^9, 3.645520783670279*^9}},
 CellID->640760558],

Cell[TextData[{
 "The ring of integers \[DoubleStruckCapitalZ] can be made a reduction ring \
using the ordering ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"0", "<", 
     RowBox[{"-", "1"}], "<", "1", "<", 
     RowBox[{"-", "2"}], "<", "2", "<", 
     RowBox[{"-", "3"}], "<", "3", "<"}], "..."}], TraditionalForm]]],
 " and setting the sets of multipliers ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["M", "c"], ":=", 
    RowBox[{"\[DoubleStruckCapitalZ]", "\\", 
     RowBox[{"{", "0", "}"}]}]}], TraditionalForm]]],
 " for all ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " (see [Buchberger1983])."
}], "Text",
 CellChangeTimes->{{3.594015911918421*^9, 3.594016022251897*^9}, {
  3.594016063398175*^9, 3.594016100490673*^9}, {3.594019549337946*^9, 
  3.594019559473611*^9}, {3.59401960302018*^9, 3.594019604976195*^9}, {
  3.596191191598151*^9, 3.596191191598265*^9}},
 CellID->36588863],

Cell["Note the following:", "Text",
 CellChangeTimes->{{3.594019562898003*^9, 3.594019567098725*^9}},
 CellID->1883788369],

Cell[CellGroupData[{

Cell[TextData[{
 "lr and lcrd are defined as in [Buchberger1983]; The only thing one has to \
take into account is that ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " might be equal, in which case there is no lcrd"
}], "Item",
 CellChangeTimes->{{3.594019574221288*^9, 3.594019609497523*^9}, {
  3.594019656232167*^9, 3.594019679947*^9}},
 CellID->440934229],

Cell["cp is defined in the obvious way", "Item",
 CellChangeTimes->{{3.594019574221288*^9, 3.594019653286078*^9}, {
  3.594019698208364*^9, 3.594019698220347*^9}},
 CellID->504979177],

Cell[TextData[{
 "rdm, as it is defined below, gives the most efficient reduction: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "-", 
    RowBox[{
     RowBox[{"rdm", "[", 
      RowBox[{"x", ",", "y"}], "]"}], "*", "y"}]}], TraditionalForm]]],
 " can certainly not be reduced modulo ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " any more. The justification that rdm is indeed a reduction multiplier \
(i.e. the result of one reduction step is less than the original element) can \
be shown by a tedious case analysis."
}], "Item",
 CellChangeTimes->{{3.594019574221288*^9, 3.594019653286078*^9}, {
  3.594019698208364*^9, 3.594019710240823*^9}, {3.594023167552541*^9, 
  3.594023220268651*^9}, {3.594039123461864*^9, 3.594039168860663*^9}},
 CellID->2079792210]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Auxiliary Functions", "Subsection",
 CellChangeTimes->{{3.6455207955565577`*^9, 3.6455207992516747`*^9}},
 CellID->470486761],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.6576097764660997`*^9, 3.657609778197703*^9}},
 CellID->248992438],

Cell["", "OpenEnvironment",
 CellID->740029918],

Cell[CellGroupData[{

Cell["Definition: Even", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.657609351288553*^9, 3.6576093542369585`*^9}},
 CellID->164698789],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "x"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"even", "[", "x", "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{"\[Wolf]EvenQ", "[", "x", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(even EvenQ)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1959508349", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594017456066879*^9, 3.59401749031068*^9}, {
  3.619249428413967*^9, 3.619249450382564*^9}, {3.657609380897405*^9, 
  3.6576093947190294`*^9}},
 CellTags->{
  "ID:1959508349", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even EvenQ"},
 CellID->1959508349],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->94497015]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->33655351],

Cell[TextData[{
 "We postulate that the abstract \[OpenCurlyDoubleQuote]quo\
\[CloseCurlyDoubleQuote] function defined in NatInt.nb agrees with the \
built-in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " \[OpenCurlyDoubleQuote]Quotient\[CloseCurlyDoubleQuote] function, and that \
an integer is even iff its remainder after division by 2 is 0. The former is \
only made use of in computations, for effectively computing quotients, \
whereas the latter is used only in proofs to rewrite the auxiliary predicate \
\[OpenCurlyDoubleQuote]even\[CloseCurlyDoubleQuote] into something there is \
theory about."
}], "Text",
 CellChangeTimes->{{3.6576092802772284`*^9, 3.6576093444401407`*^9}, {
  3.6576094978028107`*^9, 3.657609607939004*^9}, {3.65760963979426*^9, 
  3.6576096463462715`*^9}},
 CellID->938862428],

Cell["", "OpenEnvironment",
 CellID->1583138317],

Cell[CellGroupData[{

Cell["Postulate: even and quo", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.645520805796958*^9, {3.6455208732110558`*^9, 3.645520879423916*^9}, {
   3.657609271681613*^9, 3.657609273382016*^9}, {3.6576094095390553`*^9, 
   3.6576094116606593`*^9}},
 CellID->416750769],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"x", ",", 
   RowBox[{
   "y", "\[Element]", "\[DoubleStruckCapitalZ]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 1696678769}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594017441120542*^9, 3.594017448280679*^9}, {
  3.594020599385403*^9, 3.59402059953359*^9}, {3.594444642369181*^9, 
  3.594444642625345*^9}, {3.645520887410843*^9, 3.645520887742581*^9}, {
  3.6576094165122676`*^9, 3.6576094191018724`*^9}},
 CellTags->{
  "ID:1696678769", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->1696678769],

Cell[BoxData[
 RowBox[{
  RowBox[{"quo", "[", 
   RowBox[{"x", ",", "y"}], "]"}], "\[Equal]", 
  RowBox[{"\[Wolf]Quotient", "[", 
   RowBox[{"x", ",", "y"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(quo Quotient)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:227367582", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594020603902209*^9, 3.594020616682896*^9}, {
  3.619249435296054*^9, 3.61924945245189*^9}, {3.6576094221594777`*^9, 
  3.657609422861479*^9}},
 CellTags->{
  "ID:227367582", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "quo Quotient"},
 CellID->227367582],

Cell[BoxData[
 RowBox[{
  RowBox[{"even", "[", "x", "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"mod", "[", 
     RowBox[{"x", ",", "2"}], "]"}], "\[Equal]", "0"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(even mod)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:74202101", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6576094442335167`*^9, 3.657609460518745*^9}},
 CellTags->{
  "ID:74202101", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even mod"},
 CellID->74202101],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->96759121]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1750417022]
}, Closed]],

Cell[CellGroupData[{

Cell["Lemmas (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.657609784858915*^9, 3.6576098004745417`*^9}, {
  3.657620196020224*^9, 3.657620200123031*^9}},
 CellID->12017291],

Cell["", "OpenEnvironment",
 CellID->29963980],

Cell[CellGroupData[{

Cell["Lemma: Simple Properties of even", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.657609805622551*^9, 3.6576098106301603`*^9}},
 CellID->462348848],

Cell[BoxData[
 RowBox[{"even", "[", "0", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(even 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:226250890", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6576204611536922`*^9, 3.657620463322096*^9}},
 CellTags->{
  "ID:226250890", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even 0", "IR"},
 CellID->226250890],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{"even", "[", "1", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(even 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:175457915", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6576204679709044`*^9, 3.657620470607309*^9}},
 CellTags->{
  "ID:175457915", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even 1", "IR"},
 CellID->175457915],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "a", "\[Element]", "\[DoubleStruckCapitalZ]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 52392318}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657609815076168*^9, 3.6576098189293747`*^9}},
 CellTags->{
  "ID:52392318", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->52392318],

Cell[BoxData[
 RowBox[{"even", "[", 
  RowBox[{"a", "+", "a"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(even twice)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:782819098", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657867432302557*^9, 3.657867435079362*^9}},
 CellTags->{
  "ID:782819098", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even twice"},
 CellID->782819098],

Cell[BoxData[
 RowBox[{
  RowBox[{"even", "[", "a", "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"even", "[", 
     RowBox[{"a", "+", "1"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(even +1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:759529887", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657609827072589*^9, 3.6576098633582525`*^9}},
 CellTags->{
  "ID:759529887", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even +1"},
 CellID->759529887],

Cell[BoxData[
 RowBox[{
  RowBox[{"even", "[", "a", "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"even", "[", 
     RowBox[{"a", "-", "1"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(even -1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:243153755", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6579665288466635`*^9, 3.6579665311242676`*^9}},
 CellTags->{
  "ID:243153755", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even -1"},
 CellID->243153755],

Cell[BoxData[
 RowBox[{
  RowBox[{"even", "[", 
   RowBox[{"\[LeftBracketingBar]", "a", "\[RightBracketingBar]"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{"even", "[", "a", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(even abs)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:188962611", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657609878178279*^9, 3.6576098903775*^9}},
 CellTags->{
  "ID:188962611", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even abs"},
 CellID->188962611],

Cell[BoxData[
 RowBox[{
  RowBox[{"even", "[", "a", "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"quo", "[", 
      RowBox[{"a", ",", "2"}], "]"}], "+", 
     RowBox[{"quo", "[", 
      RowBox[{"a", ",", "2"}], "]"}]}], "\[Equal]", "a"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(even quo twice)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1148058", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6576099211407537`*^9, 3.6576099627148275`*^9}, {
  3.657610727818971*^9, 3.657610739596992*^9}},
 CellTags->{
  "ID:1148058", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even quo twice"},
 CellID->1148058],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"even", "[", "a", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"quo", "[", 
        RowBox[{"a", ",", "2"}], "]"}], "+", 
       RowBox[{"quo", "[", 
        RowBox[{"a", ",", "2"}], "]"}]}], ")"}], "+", "1"}], "\[Equal]", 
    "a"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(odd quo twice)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:441075310", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6579634096133842`*^9, 3.6579634278810167`*^9}, {
  3.657963481577311*^9, 3.657963483792515*^9}},
 CellTags->{
  "ID:441075310", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "odd quo twice"},
 CellID->441075310],

Cell[BoxData[
 RowBox[{
  RowBox[{"even", "[", "a", "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"quo", "[", 
     RowBox[{
      RowBox[{"a", "-", "1"}], ",", "2"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"quo", "[", 
      RowBox[{"a", ",", "2"}], "]"}], "-", "1"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(even quo -1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:256130067", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657966563119924*^9, 3.657966633474848*^9}},
 CellTags->{
  "ID:256130067", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even quo -1"},
 CellID->256130067],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"even", "[", "a", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"quo", "[", 
     RowBox[{
      RowBox[{"a", "-", "1"}], ",", "2"}], "]"}], "\[Equal]", 
    RowBox[{"quo", "[", 
     RowBox[{"a", ",", "2"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(odd quo -1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:739691344", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657966642429263*^9, 3.6579666491060753`*^9}},
 CellTags->{
  "ID:739691344", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "odd quo -1"},
 CellID->739691344],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "b", "\[Element]", "\[DoubleStruckCapitalZ]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 10064687}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658034101734062*^9, 3.6580341054156685`*^9}},
 CellTags->{
  "ID:10064687", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->10064687],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "+", "a"}], "\[GreaterEqual]", "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", "\[GreaterEqual]", 
    RowBox[{"quo", "[", 
     RowBox[{"b", ",", "2"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(twice \[GreaterEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:208882588", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658033272498006*^9, 3.658033290594037*^9}, {
  3.658033461336337*^9, 3.6580334727087574`*^9}, {3.658033518713238*^9, 
  3.658033550350094*^9}, {3.658033635418043*^9, 3.658033640612852*^9}},
 CellTags->{
  "ID:208882588", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice \[GreaterEqual]"},
 CellID->208882588],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "+", "a"}], ">", "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", ">", 
    RowBox[{"quo", "[", 
     RowBox[{"b", ",", "2"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(twice >)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:47095122", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6580336596604853`*^9, 3.658033684464529*^9}},
 CellTags->{
  "ID:47095122", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice >"},
 CellID->47095122],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->6197297]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->874473],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(even 0,[IR])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(even 0,[IR])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"even", "[", "0", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"even", "[", "0", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:226250890"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p226250890-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p226250890-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:226250890", "Proof|ID:226250890-1"},
 CellID->75108773],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p226250890-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->593942575]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(even 1,[IR])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(even 1,[IR])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"\[Not]", 
         RowBox[{
           TagBox["(", "AutoParentheses"], 
           RowBox[{"even", "[", "1", "]"}], 
           TagBox[")", "AutoParentheses"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"\[Not]", 
         RowBox[{
           TagBox["(", "AutoParentheses"], 
           RowBox[{"even", "[", "1", "]"}], 
           TagBox[")", "AutoParentheses"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:175457915"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p175457915-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p175457915-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:175457915", "Proof|ID:175457915-1"},
 CellID->36435349],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p175457915-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->33117230]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(even twice)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(even twice)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{"even", "[", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], "+", 
             StyleBox["a", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{"even", "[", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], "+", 
             StyleBox["a", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:782819098"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p782819098-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p782819098-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:782819098", "Proof|ID:782819098-1"},
 CellID->387405268],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p782819098-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->814506024]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(even +1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(even +1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"even", "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"even", "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "+", "1"}], "]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"even", "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"even", "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "+", "1"}], "]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:759529887"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p759529887-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p759529887-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:759529887", "Proof|ID:759529887-1"},
 CellID->198480204],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p759529887-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->401602925]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(even -1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(even -1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"even", "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"even", "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "-", "1"}], "]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"even", "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"even", "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "-", "1"}], "]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:243153755"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p243153755-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p243153755-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:243153755", "Proof|ID:243153755-1"},
 CellID->86433169],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p243153755-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->250829815]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(even abs)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(even abs)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"even", "[", 
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["a", "ExpressionVariable"], "\[RightBracketingBar]"}],
              "]"}], "\[DoubleLeftRightArrow]", 
           RowBox[{"even", "[", 
             StyleBox["a", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"even", "[", 
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["a", "ExpressionVariable"], "\[RightBracketingBar]"}],
              "]"}], "\[DoubleLeftRightArrow]", 
           RowBox[{"even", "[", 
             StyleBox["a", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:188962611"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p188962611-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p188962611-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:188962611", "Proof|ID:188962611-1"},
 CellID->608316900],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p188962611-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->665515500]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(even quo twice)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(even quo twice)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"even", "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "2"}], "]"}], 
               "+", 
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "2"}], "]"}]}], 
             "\[Equal]", 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"even", "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "2"}], "]"}], 
               "+", 
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "2"}], "]"}]}], 
             "\[Equal]", 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1148058"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1148058-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1148058-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1148058", "Proof|ID:1148058-1"},
 CellID->314369717],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1148058-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->5103723]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(odd quo twice)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(odd quo twice)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"even", "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{"quo", "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", "2"}], "]"}], 
                 "+", 
                 RowBox[{"quo", "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", "2"}], "]"}]}], 
               "+", "1"}], "\[Equal]", 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"even", "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{"quo", "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", "2"}], "]"}], 
                 "+", 
                 RowBox[{"quo", "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", "2"}], "]"}]}], 
               "+", "1"}], "\[Equal]", 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:441075310"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p441075310-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p441075310-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:441075310", "Proof|ID:441075310-1"},
 CellID->900195017],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p441075310-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->387398711]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(even quo -1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(even quo -1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"even", "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"quo", "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "-", "1"}], ",", 
                 "2"}], "]"}], "\[Equal]", 
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "2"}], "]"}], 
               "-", "1"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"even", "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"quo", "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "-", "1"}], ",", 
                 "2"}], "]"}], "\[Equal]", 
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "2"}], "]"}], 
               "-", "1"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:256130067"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p256130067-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p256130067-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:256130067", "Proof|ID:256130067-1"},
 CellID->748444859],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p256130067-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->312842733]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(odd quo -1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(odd quo -1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"even", "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"quo", "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "-", "1"}], ",", 
                 "2"}], "]"}], "\[Equal]", 
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "2"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"even", "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"quo", "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "-", "1"}], ",", 
                 "2"}], "]"}], "\[Equal]", 
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "2"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:739691344"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p739691344-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p739691344-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:739691344", "Proof|ID:739691344-1"},
 CellID->138102035],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p739691344-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->242339988]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(twice \[GreaterEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(twice \[GreaterEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}, {
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "+", 
               StyleBox["a", "ExpressionVariable"]}], "\[GreaterEqual]", 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], "\[GreaterEqual]", 
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", "2"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}, {
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "+", 
               StyleBox["a", "ExpressionVariable"]}], "\[GreaterEqual]", 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], "\[GreaterEqual]", 
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", "2"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:208882588"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p208882588-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p208882588-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:208882588", "Proof|ID:208882588-1"},
 CellID->148478097],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p208882588-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6580445294988117`*^9, 3.6580445312460146`*^9}},
 CellID->54070004]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(twice >)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(twice >)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}, {
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "+", 
               StyleBox["a", "ExpressionVariable"]}], ">", 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], ">", 
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", "2"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}, {
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "+", 
               StyleBox["a", "ExpressionVariable"]}], ">", 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], ">", 
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", "2"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:47095122"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p47095122-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p47095122-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:47095122", "Proof|ID:47095122-1"},
 CellID->275585894],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p47095122-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->454783722]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Reduction Integers", "Subsection",
 CellChangeTimes->{{3.645520758978251*^9, 3.6455207726852293`*^9}, {
  3.645778740755512*^9, 3.6457787445846157`*^9}},
 CellID->1312761442],

Cell["", "OpenEnvironment",
 CellID->104840774],

Cell[CellGroupData[{

Cell["Definition: Reduction Integers", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594016214792312*^9, 3.594016221869604*^9}, {
  3.6455212186496572`*^9, 3.645521222437627*^9}, {3.645778748455607*^9, 
  3.645778752016635*^9}},
 CellID->262829580],

Cell[BoxData[
 RowBox[{
  RowBox[{"ReductionIntegers", "[", "]"}], ":=", 
  UnderscriptBox["\[CapitalDelta]", "N"]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 1818128647}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594016233405784*^9, 3.594016270846026*^9}, {
   3.6457787603778057`*^9, 3.645778765166157*^9}, 3.657254309386469*^9},
 CellTags->{
  "ID:1818128647", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->1818128647],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"x", ",", "y", ",", "m", ",", "i", ",", "j"}]]], "GlobalDeclaration",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 489394336}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.594016289465345*^9, {3.594016394026027*^9, 3.59401644640808*^9}, {
   3.594016857932058*^9, 3.594016858044373*^9}, {3.594016895099517*^9, 
   3.594016895321997*^9}, {3.59402253343012*^9, 3.594022537272529*^9}, {
   3.594023141764181*^9, 3.594023143402718*^9}, {3.595057073683889*^9, 
   3.595057075338367*^9}, {3.608549457125707*^9, 3.608549457417302*^9}},
 CellTags->{
  "ID:489394336", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->489394336],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "N"], 
   RowBox[{"[", "x", "]"}]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", "\[Element]", "\[DoubleStruckCapitalZ]"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Element])", 
         "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1246011018", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594016429308848*^9, 3.594016457456615*^9}, {
  3.595057079721053*^9, 3.595057080860827*^9}, {3.6455206650043373`*^9, 
  3.6455206709220457`*^9}},
 CellTags->{
  "ID:1246011018", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Element]"},
 CellID->1246011018],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["0", "N"], ":=", "0"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:234727221", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59401660106928*^9, 3.594016606283608*^9}},
 CellTags->{
  "ID:234727221", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] 0"},
 CellID->234727221],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "N"], ":=", "1"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:849552024", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594016609365615*^9, 3.594016614110324*^9}},
 CellTags->{
  "ID:849552024", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] 1"},
 CellID->849552024],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["+", "N"], "y"}], ":=", 
  RowBox[{"x", "+", "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] +)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1301356286", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594016480755541*^9, 3.594016502600849*^9}},
 CellTags->{
  "ID:1301356286", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] +"},
 CellID->1301356286],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["-", "N"], "y"}], ":=", 
  RowBox[{"x", "-", "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] -)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:196003840", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59401650771786*^9, 3.594016517217531*^9}},
 CellTags->{
  "ID:196003840", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] -"},
 CellID->196003840],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["-", "N"], "x"}], ":=", 
  RowBox[{"-", "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] - unary)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1933917000", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594016524398826*^9, 3.594016532418878*^9}},
 CellTags->{
  "ID:1933917000", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] - unary"},
 CellID->1933917000],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["*", "N"], "y"}], ":=", 
  RowBox[{"x", "*", "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:963583550", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594016540341018*^9, 3.594016546548307*^9}},
 CellTags->{
  "ID:963583550", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] *"},
 CellID->963583550],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["\[Succeeds]", "N"], "y"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", "x", "\[RightBracketingBar]"}], ">", 
         
         RowBox[{"\[LeftBracketingBar]", "y", "\[RightBracketingBar]"}]}]},
       {
        RowBox[{
         RowBox[{
          RowBox[{"\[LeftBracketingBar]", "x", "\[RightBracketingBar]"}], 
          "\[Equal]", 
          RowBox[{"\[LeftBracketingBar]", "y", "\[RightBracketingBar]"}]}], 
         "\[And]", 
         RowBox[{"x", ">", "y"}]}]}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Succeeds])", 
         "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1947278342", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59401663678781*^9, 3.5940167691222258`*^9}, {
  3.607337931083676*^9, 3.607337934664544*^9}},
 CellTags->{
  "ID:1947278342", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Succeeds]"},
 CellID->1947278342],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["multN", "N"], "[", "x", "]"}], ":=", 
  "1"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] multN)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:90094145", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594016794284824*^9, 3.594016825674865*^9}, {
   3.607337937412462*^9, 3.607337940053592*^9}, 3.631856558979321*^9, 
   3.6573474643471355`*^9},
 CellTags->{
  "ID:90094145", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] multN"},
 CellID->90094145],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"m", ",", "x"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"m", ",", "x", ",", "1"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1432577745", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594016848268086*^9, 3.594016877675537*^9}},
 CellTags->{
  "ID:1432577745", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult"},
 CellID->1432577745],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"m", ",", "x", ",", "i"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"m", "\[NotEqual]", "0"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult 1)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1000779593", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594016882232373*^9, 3.594016991121423*^9}, {
   3.607337953960338*^9, 3.607337965304874*^9}, 3.6573471714068203`*^9, {
   3.6573475431272736`*^9, 3.6573475442972755`*^9}, {3.6573475772289333`*^9, 
   3.6573475904733567`*^9}},
 CellTags->{
  "ID:1000779593", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult 1"},
 CellID->1000779593],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lr", "N"], "[", "x", "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{"quo", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "x", "\[RightBracketingBar]"}], ",", 
        "2"}], "]"}], "\[DoubleLeftArrow]", 
      RowBox[{"even", "[", "x", "]"}]},
     {
      RowBox[{"-", 
       RowBox[{"quo", "[", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", "x", "\[RightBracketingBar]"}], 
           "+", "1"}], ")"}], ",", "2"}], "]"}]}], "\[DoubleLeftArrow]", 
      "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] lr)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:948580964", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594017291931881*^9, 3.594017299521196*^9}, {
   3.594017573882301*^9, 3.59401764961049*^9}, 3.594017692675063*^9, {
   3.594022618470815*^9, 3.594022636780717*^9}, {3.594457368378702*^9, 
   3.594457378095785*^9}, {3.6073379700028*^9, 3.607337970956162*^9}},
 CellTags->{
  "ID:948580964", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lr"},
 CellID->948580964],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdm", "N"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {"0", "\[DoubleLeftArrow]", 
      RowBox[{"y", "\[Equal]", "0"}]},
     {
      RowBox[{"quo", "[", 
       RowBox[{
        RowBox[{"x", "+", 
         RowBox[{"quo", "[", 
          RowBox[{"y", ",", "2"}], "]"}]}], ",", "y"}], "]"}], 
      "\[DoubleLeftArrow]", 
      RowBox[{
       RowBox[{"y", ">", "0"}], "\[And]", 
       RowBox[{"x", "\[GreaterEqual]", "0"}]}]},
     {
      RowBox[{"-", 
       RowBox[{"quo", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "x"}], "+", 
          RowBox[{"quo", "[", 
           RowBox[{
            RowBox[{"y", "-", "1"}], ",", "2"}], "]"}]}], ",", "y"}], "]"}]}],
       "\[DoubleLongLeftArrow]", 
      RowBox[{
       RowBox[{"y", ">", "0"}], "\[And]", 
       RowBox[{"x", "<", "0"}]}]},
     {
      RowBox[{"-", 
       RowBox[{
        UnderscriptBox["rdm", "N"], "[", 
        RowBox[{"x", ",", 
         RowBox[{"-", "y"}]}], "]"}]}], "\[DoubleLongLeftArrow]", 
      "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] rdm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1258513368", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594017010664371*^9, 3.594017016462996*^9}, {
   3.594017086378167*^9, 3.594017110085201*^9}, {3.594017145804906*^9, 
   3.59401717954676*^9}, {3.594019102964903*^9, 3.594019160323684*^9}, {
   3.59401921697438*^9, 3.594019302355194*^9}, {3.594019929879594*^9, 
   3.594019941840062*^9}, {3.594020086136639*^9, 3.594020094598402*^9}, {
   3.594020633678233*^9, 3.594020659169091*^9}, {3.594020869342512*^9, 
   3.594020869565602*^9}, {3.594021603522921*^9, 3.594021609893842*^9}, {
   3.59402206471816*^9, 3.594022072994627*^9}, {3.59402244102116*^9, 
   3.594022475336938*^9}, {3.594037680103941*^9, 3.594037692137421*^9}, {
   3.594038006689027*^9, 3.594038029926896*^9}, {3.594038677782*^9, 
   3.594038678781965*^9}, {3.594101962005004*^9, 3.594102108542291*^9}, {
   3.59410221632647*^9, 3.594102216695624*^9}, {3.594103741185761*^9, 
   3.594103940919456*^9}, {3.594104101470147*^9, 3.594104116930902*^9}, {
   3.594104463511523*^9, 3.594104524955774*^9}, {3.607337974205626*^9, 
   3.607337975054923*^9}, 3.614931570745865*^9, {3.657971775484291*^9, 
   3.657971777075494*^9}},
 CellTags->{
  "ID:1258513368", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] rdm"},
 CellID->1258513368],

Cell[TextData[{
 "In the definition below note that ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["min", 
    UnderscriptBox["\[Succeeds]", "N"]], TraditionalForm]]],
 " in fact gives the ",
 StyleBox["maximum",
  FontWeight->"Bold"],
 " of its arguments!"
}], "EnvironmentText",
 CellChangeTimes->{{3.610092298554093*^9, 3.610092336035374*^9}, {
  3.610094214803531*^9, 3.610094216759786*^9}, {3.6576100535537868`*^9, 
  3.6576100542089877`*^9}},
 CellID->1870060344],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", 
   RowBox[{"x", ",", "y", ",", "i", ",", "j"}], "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
      "\[DoubleLeftArrow]", 
      RowBox[{
       RowBox[{"x", "\[Equal]", "y"}], "\[Or]", 
       RowBox[{"x", "\[Equal]", "0"}], "\[Or]", 
       RowBox[{"y", "\[Equal]", "0"}]}]},
     {
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        SubscriptBox["min", 
         UnderscriptBox["\[Succeeds]", "N"]], "[", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{
           UnderscriptBox["lr", "N"], "[", "x", "]"}], ",", 
          RowBox[{
           UnderscriptBox["lr", "N"], "[", "y", "]"}]}], 
         "\[RightAngleBracket]"}], "]"}], "\[RightAngleBracket]"}], 
      "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1342406859", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594017723655617*^9, 3.594017837258601*^9}, {
  3.594018705613314*^9, 3.594018734608988*^9}, {3.607337982099866*^9, 
  3.607337990053351*^9}, {3.608549466947844*^9, 3.608549467803156*^9}, {
  3.610095884108262*^9, 3.610095921175071*^9}, {3.610096023345984*^9, 
  3.61009603445414*^9}, {3.610096801216476*^9, 3.610096812875121*^9}},
 CellTags->{
  "ID:1342406859", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd"},
 CellID->1342406859],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", "x", "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
      "\[DoubleLeftArrow]", 
      RowBox[{"x", "\[Equal]", "0"}]},
     {
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        UnderscriptBox["lr", "N"], "[", "x", "]"}], "\[RightAngleBracket]"}], 
      "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd unary)", 
         "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:456350999", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594457383345808*^9, 3.594457395732788*^9}, {
  3.595056643665148*^9, 3.595056682046463*^9}, {3.610096840193464*^9, 
  3.610096855974403*^9}},
 CellTags->{
  "ID:456350999", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd unary"},
 CellID->456350999],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpm", "N"], "[", 
   RowBox[{"x", ",", "y", ",", "m", ",", "i", ",", "j"}], "]"}], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["rdm", "N"], "[", 
     RowBox[{"m", ",", "x"}], "]"}], ",", 
    RowBox[{
     UnderscriptBox["rdm", "N"], "[", 
     RowBox[{"m", ",", "y"}], "]"}]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[ScriptCapitalR]\[DoubleStruckCapitalZ] cpm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1330653407", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594459156922512*^9, 3.59445919875409*^9}, {
  3.595056925543981*^9, 3.595056927944702*^9}, {3.610081806722775*^9, 
  3.610081807394762*^9}, {3.610096049875489*^9, 3.610096050617129*^9}},
 CellTags->{
  "ID:1330653407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] cpm"},
 CellID->1330653407],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->629448807]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1539081838]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Tests", "Section",
 CellChangeTimes->{{3.592043465736425*^9, 3.592043466408566*^9}},
 CellID->1210929308],

Cell["", "OpenEnvironment",
 CellID->81008174],

Cell[CellGroupData[{

Cell["Definition: Reduction Integers", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645778829037125*^9, 3.645778833480809*^9}, {
  3.645779901518318*^9, 3.645779909555133*^9}},
 CellID->1093450566],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalZ]", ":=", 
  RowBox[{"ReductionIntegers", "[", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2131968713", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645778836859263*^9, 3.645778848239134*^9}, {
  3.645779895700758*^9, 3.645779897031863*^9}},
 CellTags->{
  "ID:2131968713", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalZ]"},
 CellID->2131968713],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->121089228]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->910413054],

Cell[CellGroupData[{

Cell["rdm", "Subsection",
 CellChangeTimes->{{3.645521388834046*^9, 3.6455213891676807`*^9}},
 CellID->321798810],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"7", ",", "5"}], "]"}]], "Computation",
 CellChangeTimes->{{3.594036002335985*^9, 3.594036035629359*^9}, 
   3.6457801600745068`*^9},
 CellID->1429456864],

Cell[BoxData[
 FormBox["1", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.607338334617406*^9, 3.607338415917868*^9, {3.610096252381105*^9, 
   3.61009626002697*^9}, 3.61493175396686*^9, 3.615713723142683*^9, 
   3.615801217397859*^9, 3.6314471082074537`*^9, 3.631447151157099*^9, 
   3.631449073621139*^9, 3.6357577952221937`*^9, 3.645521362617606*^9},
 CellID->978138576]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1429456864"]]]]}, {None, None}},
 CellChangeTimes->{3.6455213629616833`*^9},
 CellID->1265921044],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1429456864-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6455213632012253`*^9},
 CellID->1885477897]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"7", ",", 
   RowBox[{"-", "5"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.594036280283913*^9, 3.594036283009744*^9}, 
   3.645780162022171*^9},
 CellID->157142288],

Cell[BoxData[
 FormBox[
  RowBox[{"-", "1"}], $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.60733841827236*^9, 3.610096264560988*^9, 
  3.614931756606223*^9, 3.615713725512819*^9, 3.615801219516182*^9, 
  3.631447154291107*^9, 3.631449076116292*^9, 3.635757797832473*^9, 
  3.645521364991166*^9},
 CellID->1465402715]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c157142288"]]]]}, {None, None}},
 CellChangeTimes->{3.645521365496704*^9},
 CellID->473424448],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c157142288-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521365621229*^9},
 CellID->786290236]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"10", ",", "5"}], "]"}]], "Computation",
 CellChangeTimes->{{3.59403628816842*^9, 3.594036294465135*^9}, 
   3.645780163239318*^9},
 CellID->1960775227],

Cell[BoxData[
 FormBox["2", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.607338420308015*^9, 3.610096266597832*^9, 
  3.614931758488602*^9, 3.615713727912413*^9, 3.61580122129382*^9, 
  3.6314471568430243`*^9, 3.63575779994202*^9, 3.645521367235977*^9},
 CellID->378014128]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1960775227"]]]]}, {None, None}},
 CellChangeTimes->{3.645521367609178*^9},
 CellID->1901130687],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1960775227-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521367735867*^9},
 CellID->944906225]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"197", ",", "23"}], "]"}]], "Computation",
 CellChangeTimes->{{3.594104613227373*^9, 3.59410462892233*^9}, 
   3.64578016556012*^9},
 CellID->1637496245],

Cell[BoxData[
 FormBox["9", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.60733842281637*^9, 3.610096269005639*^9, 
  3.614931760597201*^9, 3.61571373230719*^9, 3.615801223537896*^9, 
  3.631447159588129*^9, 3.635757802074977*^9, 3.6455213695961647`*^9},
 CellID->1717908025]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1637496245"]]]]}, {None, None}},
 CellChangeTimes->{3.645521370166006*^9},
 CellID->1813781623],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1637496245-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6455213702843723`*^9},
 CellID->890136672]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"1", ",", "8"}], "]"}]], "Computation",
 CellChangeTimes->{{3.610096275490749*^9, 3.610096291745301*^9}, 
   3.645780167585998*^9},
 CellID->2123737964],

Cell[BoxData[
 FormBox["0", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.610096292726091*^9, 3.614931762364984*^9, 3.615713734494789*^9, 
   3.615801225349659*^9, {3.631447161936618*^9, 3.631447171421142*^9}, 
   3.635757804434573*^9, 3.6455213718428593`*^9},
 CellID->1518450864]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c2123737964"]]]]}, {None, None}},
 CellChangeTimes->{3.645521372224245*^9},
 CellID->975643154],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c2123737964-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.64552137235476*^9},
 CellID->1403959631]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
  RowBox[{
   RowBox[{"-", "2"}], ",", "4"}], "]"}]], "Computation",
 CellChangeTimes->{{3.614931773227404*^9, 3.614931784983179*^9}, 
   3.645780168840116*^9},
 CellID->252382512],

Cell[BoxData[
 FormBox["0", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.614931785624665*^9, 3.615713737593803*^9, 3.615801227533114*^9, {
   3.631447164644587*^9, 3.631447174083405*^9}, {3.635757807380039*^9, 
   3.6357578129382153`*^9}, 3.645521374086513*^9},
 CellID->1505423261]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c252382512"]]]]}, {None, None}},
 CellChangeTimes->{3.6455213744672613`*^9},
 CellID->796291593],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c252382512-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521374596664*^9},
 CellID->1915561639]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
  RowBox[{
   RowBox[{"-", "4"}], ",", 
   RowBox[{"-", "7"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.594037951335516*^9, 3.594037959550343*^9}, 
   3.6457801700576887`*^9},
 CellID->266870911],

Cell[BoxData[
 FormBox["1", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.607338428390072*^9, 3.610096309693001*^9, 3.614931766722092*^9, 
   3.615713744263214*^9, {3.631447186020172*^9, 3.6314471889838037`*^9}, 
   3.635757818727903*^9, 3.6455213791355057`*^9},
 CellID->116083407]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c266870911"]]]]}, {None, None}},
 CellChangeTimes->{3.645521379658697*^9},
 CellID->2029286797],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c266870911-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521379788121*^9},
 CellID->1380713228]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["lcrd", "Subsection",
 CellChangeTimes->{{3.64552140532687*^9, 3.64552140588316*^9}},
 CellID->1370888350],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]"], "[", 
  RowBox[{
   RowBox[{"-", "7"}], ",", "5", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.594036310504385*^9, 3.594036313565996*^9}, {
   3.594037881298804*^9, 3.594037887225668*^9}, {3.610096304215646*^9, 
   3.610096305031246*^9}, 3.645780173849616*^9},
 CellID->1703330943],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"-", "4"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{
  3.607338425805935*^9, {3.610096296100505*^9, 3.61009630567989*^9}, 
   3.614931764726314*^9, 3.615713741675498*^9, {3.6314471780722723`*^9, 
   3.6314471809658127`*^9}, 3.6357578156657267`*^9, 3.645521376700822*^9},
 CellID->1536458168]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1703330943"]]]]}, {None, None}},
 CellChangeTimes->{3.6455213771129932`*^9},
 CellID->246257879],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1703330943-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521377237521*^9},
 CellID->1309963877]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"0", ",", "10", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.645521483396628*^9, 3.6455214864585247`*^9}, 
   3.6457801751490192`*^9},
 CellID->1150003754],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645521487430808*^9},
 CellID->2113262834]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1150003754"]]]]}, {None, None}},
 CellChangeTimes->{3.645521487866346*^9},
 CellID->1719993529],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1150003754-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521488001605*^9},
 CellID->1784884467]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.645521509516058*^9, 3.645521509777831*^9}, 
   3.645780176402931*^9},
 CellID->1557862139],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645521510581188*^9},
 CellID->1353116481]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1557862139"]]]]}, {None, None}},
 CellChangeTimes->{3.64552151102379*^9},
 CellID->1512462514],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1557862139-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521511150663*^9},
 CellID->642311510]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"5", ",", 
   RowBox[{"-", "5"}], ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.645521494375142*^9, 3.6455214999063997`*^9}, 
   3.64578017776798*^9},
 CellID->581379970],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"-", "3"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645521501423936*^9},
 CellID->1478552428]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c581379970"]]]]}, {None, None}},
 CellChangeTimes->{3.64552150185597*^9},
 CellID->1858231219],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c581379970-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521501995749*^9},
 CellID->382299358]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]"], "[", "0", "]"}]], "Computation",\

 CellChangeTimes->{{3.645521419789138*^9, 3.6455214427765837`*^9}, 
   3.64578017905767*^9},
 CellID->1243235648],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645521443334746*^9},
 CellID->1849233252]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1243235648"]]]]}, {None, None}},
 CellChangeTimes->{3.645521443785227*^9},
 CellID->505878642],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1243235648-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6455214439253807`*^9},
 CellID->1259921456]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"-", "7"}], "]"}]], "Computation",
 CellChangeTimes->{{3.6455214523422937`*^9, 3.6455214533358173`*^9}, 
   3.645780180495288*^9},
 CellID->1587692045],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"-", "4"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.6455214541257877`*^9},
 CellID->1456581472]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1587692045"]]]]}, {None, None}},
 CellChangeTimes->{3.645521454654262*^9},
 CellID->416142635],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1587692045-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521454791716*^9},
 CellID->417178908]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]"], "[", "6", "]"}]], "Computation",\

 CellChangeTimes->{{3.6455214581571493`*^9, 3.645521462427155*^9}, 
   3.645780181855035*^9},
 CellID->1242138448],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "3", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{{3.6455214586865177`*^9, 3.645521462917197*^9}},
 CellID->545447630]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1242138448"]]]]}, {None, None}},
 CellChangeTimes->{3.645521463323453*^9},
 CellID->850780133],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1242138448-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521463457343*^9},
 CellID->1266603761]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["cpm", "Subsection",
 CellChangeTimes->{{3.64552141103255*^9, 3.645521411586837*^9}},
 CellID->479188981],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalZ]"], "[", 
  RowBox[{
   RowBox[{"-", "7"}], ",", "5", ",", 
   RowBox[{"-", "4"}], ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.594037910929813*^9, 3.594037916041262*^9}, {
   3.610096314716894*^9, 3.610096336085331*^9}, 3.64578018560715*^9},
 CellID->111822425],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"1", ",", 
    RowBox[{"-", "1"}]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{{3.607338430162545*^9, 3.607338435788356*^9}, {
   3.610096315630022*^9, 3.610096336811402*^9}, 3.614931768716398*^9, 
   3.615713746410453*^9, {3.631447194929003*^9, 3.631447198101015*^9}, 
   3.635757821445896*^9, 3.645521381488871*^9},
 CellID->425746342]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c111822425"]]]]}, {None, None}},
 CellChangeTimes->{3.645521381863418*^9},
 CellID->1022918257],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c111822425-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521381993579*^9},
 CellID->105124728]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"7", ",", "5", ",", 
   RowBox[{"-", "4"}], ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{3.6455215347765408`*^9, 3.645780187116898*^9},
 CellID->1993907386],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"-", "1"}], ",", 
    RowBox[{"-", "1"}]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645521536298799*^9},
 CellID->930951114]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1993907386"]]]]}, {None, None}},
 CellChangeTimes->{3.6455215367088537`*^9},
 CellID->1993783205],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1993907386-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6455215368317738`*^9},
 CellID->1845365743]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"7", ",", 
   RowBox[{"-", "5"}], ",", 
   RowBox[{"-", "4"}], ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.645521541182469*^9, 3.645521542232008*^9}, 
   3.645780188481505*^9},
 CellID->1136958002],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"-", "1"}], ",", "1"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645521542761199*^9},
 CellID->526249381]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1136958002"]]]]}, {None, None}},
 CellChangeTimes->{3.6455215431991053`*^9},
 CellID->1383108897],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1136958002-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521543325828*^9},
 CellID->1634101449]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalZ]"], "[", 
  RowBox[{
   RowBox[{"-", "7"}], ",", 
   RowBox[{"-", "5"}], ",", 
   RowBox[{"-", "4"}], ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{3.64552154709403*^9, 3.645780189772009*^9},
 CellID->54162398],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"1", ",", "1"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645521547659539*^9},
 CellID->32754321]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c54162398"]]]]}, {None, None}},
 CellChangeTimes->{3.645521548028741*^9},
 CellID->1556773737],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c54162398-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521548156994*^9},
 CellID->1285835433]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"7", ",", "5", ",", "4", ",", "1", ",", "1"}], "]"}]], "Computation",\

 CellChangeTimes->{{3.645521554952517*^9, 3.6455215558362083`*^9}, 
   3.645780191173379*^9},
 CellID->1435882704],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"1", ",", "1"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645521556753302*^9},
 CellID->568066442]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1435882704"]]]]}, {None, None}},
 CellChangeTimes->{3.645521557185397*^9},
 CellID->1903147392],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1435882704-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645521557313476*^9},
 CellID->548348087]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Theorem: ReductionIntegers Constitutes Reduction Ring (\[Checkmark])\
\>", "Section",
 CellChangeTimes->{{3.592043458571405*^9, 3.592043460801741*^9}, {
  3.645520470258786*^9, 3.6455204875194683`*^9}, {3.6572549655578213`*^9, 
  3.6572549701598296`*^9}, {3.6580442761887565`*^9, 3.658044280650364*^9}},
 CellID->1769514847],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"\[ScriptCapitalZ]", "=", 
   RowBox[{"ReductionIntegers", "[", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 335127278}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657254453452722*^9, 3.657254461580336*^9}},
 CellTags->{
  "ID:335127278", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->335127278],

Cell[CellGroupData[{

Cell["Reduction Ring (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6572543520993443`*^9, 3.6572543547045484`*^9}, {
  3.657891247514464*^9, 3.657891251601671*^9}},
 CellID->44748028],

Cell[CellGroupData[{

Cell["\<\
Commutative Ring with Identity (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.657257059039304*^9, 3.657257077884137*^9}},
 CellID->415794492],

Cell["", "OpenEnvironment",
 CellID->1331628327],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[ScriptCapitalZ] is Unitary Commutative Ring and Integral Domain\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594023294093077*^9, 3.594023313754166*^9}, {
  3.594446445332117*^9, 3.594446449537635*^9}, {3.6572545100964212`*^9, 
  3.6572545123896255`*^9}, {3.6573485560244527`*^9, 3.657348561047662*^9}},
 CellID->1603429019],

Cell[BoxData[
 RowBox[{"isCommRing1", "[", "\[ScriptCapitalZ]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCommRing1 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2044589078", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594023347448846*^9, 3.594023357232757*^9}, {
  3.594446434012581*^9, 3.594446439591906*^9}, {3.657254466509945*^9, 
  3.657254467757947*^9}},
 CellTags->{
  "ID:2044589078", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommRing1 \[ScriptCapitalZ]"},
 CellID->2044589078],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"a", 
       UnderscriptBox["*", "\[ScriptCapitalZ]"], "b"}], "\[Equal]", 
      UnderscriptBox["0", "\[ScriptCapitalZ]"]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"a", "\[Equal]", 
       UnderscriptBox["0", "\[ScriptCapitalZ]"]}], "\[Or]", 
      RowBox[{"b", "\[Equal]", 
       UnderscriptBox["0", "\[ScriptCapitalZ]"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(integral domain \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:48028479", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6573484459506598`*^9, 3.65734850292196*^9}},
 CellTags->{
  "ID:48028479", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "integral domain \[ScriptCapitalZ]"},
 CellID->48028479],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2061832140]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1577743469],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCommRing1 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCommRing1 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isCommRing1", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isCommRing1", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2044589078"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2044589078-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2044589078-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2044589078", "Proof|ID:2044589078-1"},
 CellID->47164179],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2044589078-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->94304473]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(integral domain \[ScriptCapitalZ])\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(integral domain \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["b", "ExpressionVariable"]}], "\[Equal]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIntegers", "[", "]"}]]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[Or]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionIntegers", "[", "]"}]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["b", "ExpressionVariable"]}], "\[Equal]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIntegers", "[", "]"}]]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[Or]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionIntegers", "[", "]"}]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:48028479"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p48028479-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p48028479-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:48028479", "Proof|ID:48028479-1"},
 CellID->584664],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p48028479-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->563218315]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Order Relation (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6572570841397476`*^9, 3.6572570868385525`*^9}, {
  3.6573470569474196`*^9, 3.657347061097027*^9}},
 CellID->26895694],

Cell["", "OpenEnvironment",
 CellID->228008460],

Cell[CellGroupData[{

Cell["\<\
Lemma: Properties of the Order Relation on \[ScriptCapitalZ]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.657257092657363*^9, 3.657257101018977*^9}},
 CellID->3757693],

Cell[BoxData[
 RowBox[{"isPartIrreflOrder", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalZ]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isPartIrreflOrder \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1813588854", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028425273078*^9, 3.594028441312077*^9}, {
   3.594028474524201*^9, 3.594028489538728*^9}, 3.596191154686295*^9, {
   3.607773300007142*^9, 3.607773304214125*^9}, {3.6572564335212016`*^9, 
   3.6572564351436043`*^9}},
 CellTags->{
  "ID:1813588854", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isPartIrreflOrder \[ScriptCapitalZ]"},
 CellID->1813588854],

Cell[BoxData[
 RowBox[{"isTotalIrreflOrder", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalZ]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isTotalIrreflOrder \[ScriptCapitalZ])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:187039377", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6572572479924326`*^9, 3.657257251752039*^9}},
 CellTags->{
  "ID:187039377", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isTotalIrreflOrder \[ScriptCapitalZ]"},
 CellID->187039377],

Cell[BoxData[
 RowBox[{"isNoetherian", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalZ]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isNoetherian \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:159454607", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185133124382*^9, 3.610185144937083*^9}, {
  3.657256437608409*^9, 3.6572564390904117`*^9}},
 CellTags->{
  "ID:159454607", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isNoetherian \[ScriptCapitalZ]"},
 CellID->159454607],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->270718515]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->38854093],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isPartIrreflOrder \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isPartIrreflOrder \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isPartIrreflOrder", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", 
            RowBox[{"ReductionIntegers", "[", "]"}]], ",", 
           RowBox[{"ReductionIntegers", "[", "]"}]}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isPartIrreflOrder", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", 
            RowBox[{"ReductionIntegers", "[", "]"}]], ",", 
           RowBox[{"ReductionIntegers", "[", "]"}]}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1813588854"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1813588854-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1813588854-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1813588854", "Proof|ID:1813588854-1"},
 CellID->80460269],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1813588854-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->40342381]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isTotalIrreflOrder \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isTotalIrreflOrder \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isTotalIrreflOrder", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", 
            RowBox[{"ReductionIntegers", "[", "]"}]], ",", 
           RowBox[{"ReductionIntegers", "[", "]"}]}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isTotalIrreflOrder", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", 
            RowBox[{"ReductionIntegers", "[", "]"}]], ",", 
           RowBox[{"ReductionIntegers", "[", "]"}]}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:187039377"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p187039377-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p187039377-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:187039377", "Proof|ID:187039377-1"},
 CellID->106498233],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p187039377-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->257809711]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isNoetherian \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isNoetherian \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isNoetherian", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", 
            RowBox[{"ReductionIntegers", "[", "]"}]], ",", 
           RowBox[{"ReductionIntegers", "[", "]"}]}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isNoetherian", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", 
            RowBox[{"ReductionIntegers", "[", "]"}]], ",", 
           RowBox[{"ReductionIntegers", "[", "]"}]}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:159454607"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p159454607-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p159454607-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:159454607", "Proof|ID:159454607-1"},
 CellID->187116588],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p159454607-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->323553876]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Multipliers & R0-R4 (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6572571138266*^9, 3.6572571153398027`*^9}, {
  3.6573470809568615`*^9, 3.6573470901140776`*^9}, {3.6573492527024765`*^9, 
  3.6573492569144845`*^9}},
 CellID->217483992],

Cell["", "OpenEnvironment",
 CellID->368849717],

Cell[CellGroupData[{

Cell["\<\
Lemma: \[ScriptCapitalZ] Satisfies Multipliers and R0-R4\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.657347100520296*^9, 3.65734711980193*^9}, {
  3.657453560557459*^9, 3.65745356673507*^9}},
 CellID->808264381],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
    RowBox[{"[", 
     RowBox[{"c", ",", "m"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"m", ",", "c"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"m", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalZ]"]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult alternative \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:38031932", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6573479348127613`*^9, 3.6573479740000305`*^9}, {
  3.6573480189905095`*^9, 3.65734801927131*^9}},
 CellTags->{
  "ID:38031932", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "mult alternative \[ScriptCapitalZ]"},
 CellID->38031932],

Cell[BoxData[
 RowBox[{"Multipliers", "[", "\[ScriptCapitalZ]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Multipliers \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1890039364", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028500321373*^9, 3.594028503373398*^9}, 
   3.6572564407128143`*^9},
 CellTags->{
  "ID:1890039364", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "Multipliers \[ScriptCapitalZ]"},
 CellID->1890039364],

Cell[BoxData[
 RowBox[{"R0", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R0 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1152753107", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028507793889*^9, 3.594028509762232*^9}, 
   3.657256443115218*^9},
 CellTags->{
  "ID:1152753107", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R0 \[ScriptCapitalZ]"},
 CellID->1152753107],

Cell[BoxData[
 RowBox[{"R1", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R1 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:307366433", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028515432167*^9, 3.594028517002437*^9}, 
   3.657256444722021*^9},
 CellTags->{
  "ID:307366433", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R1 \[ScriptCapitalZ]"},
 CellID->307366433],

Cell[BoxData[
 RowBox[{"R2", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R2 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:963452595", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028527178973*^9, 3.594028530824697*^9}, 
   3.594028568213922*^9, 3.6572564460012236`*^9},
 CellTags->{
  "ID:963452595", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R2 \[ScriptCapitalZ]"},
 CellID->963452595],

Cell[BoxData[
 RowBox[{"R3", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R3 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1084156204", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.59402853396289*^9, 3.594028569392589*^9, 
  3.6572564470620255`*^9},
 CellTags->{
  "ID:1084156204", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R3 \[ScriptCapitalZ]"},
 CellID->1084156204],

Cell[BoxData[
 RowBox[{"R4", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R4 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2041570274", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.594028537396702*^9, 3.594028570130371*^9, 
  3.657256448372428*^9},
 CellTags->{
  "ID:2041570274", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R4 \[ScriptCapitalZ]"},
 CellID->2041570274],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->382946846]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->79265178],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult alternative \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult alternative \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              
              RowBox[{"ReductionIntegers", "[", 
                "]"}]]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              
              RowBox[{"ReductionIntegers", "[", 
                "]"}]]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:38031932"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p38031932-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p38031932-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:38031932", "Proof|ID:38031932-1"},
 CellID->279953348],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p38031932-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->17290125]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Multipliers \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Multipliers \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"Multipliers", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"Multipliers", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1890039364"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1890039364-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1890039364-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1890039364", "Proof|ID:1890039364-1"},
 CellID->845457494],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1890039364-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->223964381]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R0 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R0 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R0", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R0", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1152753107"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1152753107-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1152753107-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1152753107", "Proof|ID:1152753107-1"},
 CellID->229850786],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1152753107-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->41398535]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R1 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R1 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R1", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R1", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:307366433"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p307366433-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p307366433-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:307366433", "Proof|ID:307366433-1"},
 CellID->93056147],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p307366433-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->31010785]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R2 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R2 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R2", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R2", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:963452595"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p963452595-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p963452595-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:963452595", "Proof|ID:963452595-1"},
 CellID->8129192],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p963452595-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->201562422]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R3 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R3 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R3", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R3", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1084156204"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1084156204-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1084156204-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1084156204", "Proof|ID:1084156204-1"},
 CellID->477880951],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1084156204-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->360511373]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R4 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R4 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R4", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R4", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2041570274"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2041570274-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2041570274-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2041570274", "Proof|ID:2041570274-1"},
 CellID->299709042],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2041570274-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->377565118]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Reducibility in \[ScriptCapitalZ] (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.657349729688515*^9, 3.6573497372077284`*^9}, {
  3.6578848512112045`*^9, 3.6578848552360115`*^9}},
 CellID->11190904],

Cell["", "OpenEnvironment",
 CellID->624303376],

Cell[CellGroupData[{

Cell["Lemma: Order Relation in \[ScriptCapitalZ]", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.657867878945343*^9, 3.657867885029354*^9}},
 CellID->473325153],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 260068164}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6578679045917883`*^9, 3.6578679125322027`*^9}, {
  3.6578679896275377`*^9, 3.657867989736738*^9}},
 CellTags->{
  "ID:260068164", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->260068164],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
  RowBox[{"[", 
   RowBox[{"sgn", "[", "a", "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership sgn \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:211178897", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65735853652492*^9, 3.6573585466181374`*^9}},
 CellTags->{
  "ID:211178897", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership sgn \[ScriptCapitalZ]"},
 CellID->211178897],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalZ]"]}], "\[And]", 
    RowBox[{"b", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalZ]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{"sgn", "[", "a", "]"}], 
       UnderscriptBox["*", "\[ScriptCapitalZ]"], 
       RowBox[{"sgn", "[", "b", "]"}]}], "\[Equal]", 
      UnderscriptBox["1", "\[ScriptCapitalZ]"]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Or]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{"sgn", "[", "a", "]"}], 
       UnderscriptBox["*", "\[ScriptCapitalZ]"], 
       RowBox[{"sgn", "[", "b", "]"}]}], "\[Equal]", 
      RowBox[{
       UnderscriptBox["-", "\[ScriptCapitalZ]"], 
       UnderscriptBox["1", "\[ScriptCapitalZ]"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* sgn \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:118405539", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6578879604024663`*^9, 3.657888015381163*^9}},
 CellTags->{
  "ID:118405539", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "* sgn \[ScriptCapitalZ]"},
 CellID->118405539],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{
   UnderscriptBox["0", "\[ScriptCapitalZ]"], 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], 
   "a"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 minimal \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:88914628", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6573602416177173`*^9, 3.6573602559385424`*^9}, {
  3.6573602889326*^9, 3.6573602909450035`*^9}},
 CellTags->{
  "ID:88914628", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "0 minimal \[ScriptCapitalZ]"},
 CellID->88914628],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["+", "\[ScriptCapitalZ]"], "a"}], 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], 
    RowBox[{"b", 
     UnderscriptBox["+", "\[ScriptCapitalZ]"], "b"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(twice \[Succeeds] twice \[ScriptCapitalZ])", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:69480817", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6578647317906127`*^9, 3.6578648136449566`*^9}, {
  3.657864910599127*^9, 3.6578649216439466`*^9}},
 CellTags->{
  "ID:69480817", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice \[Succeeds] twice \[ScriptCapitalZ]"},
 CellID->69480817],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], 
    RowBox[{"a", 
     UnderscriptBox["+", "\[ScriptCapitalZ]"], "c"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{"a", ">", "0"}], "\[And]", 
         RowBox[{"0", ">", "c"}], "\[And]", 
         RowBox[{"c", "\[GreaterEqual]", 
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{"a", "+", "a"}], ")"}]}]}]}]},
       {
        RowBox[{
         RowBox[{"0", ">", "a"}], "\[And]", 
         RowBox[{"c", ">", "0"}], "\[And]", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{"a", "+", "a"}], ")"}]}], ">", "c"}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[Succeeds] + \[DoubleRightArrow] \[ScriptCapitalZ])", 
         "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:24170142", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6574283704971185`*^9, 3.657428379716735*^9}},
 CellTags->{
  "ID:24170142", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[DoubleRightArrow] \[ScriptCapitalZ]"},
 CellID->24170142],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{"a", ">", "0"}], "\[And]", 
         RowBox[{"0", ">", "c"}], "\[And]", 
         RowBox[{"c", "\[GreaterEqual]", 
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{"a", "+", "a"}], ")"}]}]}]}]},
       {
        RowBox[{
         RowBox[{"0", ">", "a"}], "\[And]", 
         RowBox[{"c", ">", "0"}], "\[And]", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{"a", "+", "a"}], ")"}]}], ">", "c"}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], 
    RowBox[{"a", 
     UnderscriptBox["+", "\[ScriptCapitalZ]"], "c"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[Succeeds] + \[DoubleLeftArrow] \[ScriptCapitalZ])", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:302704588", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6574314841319885`*^9, 3.6574314961752095`*^9}},
 CellTags->{
  "ID:302704588", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[DoubleLeftArrow] \[ScriptCapitalZ]"},
 CellID->302704588],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], 
    RowBox[{"a", 
     UnderscriptBox["+", "\[ScriptCapitalZ]"], "c"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{"a", ">", "0"}], "\[And]", 
         RowBox[{"0", ">", "c"}], "\[And]", 
         RowBox[{"c", "\[GreaterEqual]", 
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{"a", "+", "a"}], ")"}]}]}]}]},
       {
        RowBox[{
         RowBox[{"0", ">", "a"}], "\[And]", 
         RowBox[{"c", ">", "0"}], "\[And]", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{"a", "+", "a"}], ")"}]}], ">", "c"}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Succeeds] + \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:131195953", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65742604874564*^9, 3.657426133845789*^9}, {
  3.657428355053091*^9, 3.6574283577362957`*^9}},
 CellTags->{
  "ID:131195953", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[ScriptCapitalZ]"},
 CellID->131195953],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], 
    RowBox[{"a", 
     UnderscriptBox["-", "\[ScriptCapitalZ]"], "c"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{"a", ">", "0"}], "\[And]", 
         RowBox[{"c", ">", "0"}], "\[And]", 
         RowBox[{
          RowBox[{"a", "+", "a"}], "\[GreaterEqual]", "c"}]}]},
       {
        RowBox[{
         RowBox[{"0", ">", "a"}], "\[And]", 
         RowBox[{"0", ">", "c"}], "\[And]", 
         RowBox[{"c", ">", 
          RowBox[{"a", "+", "a"}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Succeeds] - \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:129207038", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6579615096852465`*^9, 3.65796155157132*^9}},
 CellTags->{
  "ID:129207038", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] - \[ScriptCapitalZ]"},
 CellID->129207038],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{"a", ">", "0"}], "\[And]", 
         RowBox[{"a", ">", "b"}], "\[And]", 
         RowBox[{"b", "\[GreaterEqual]", 
          RowBox[{"-", "a"}]}]}]},
       {
        RowBox[{
         RowBox[{"0", ">", "a"}], "\[And]", 
         RowBox[{"b", ">", "a"}], "\[And]", 
         RowBox[{
          RowBox[{"-", "a"}], ">", "b"}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Succeeds] alternative \[ScriptCapitalZ])", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:59300728", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657433263919915*^9, 3.6574333480664635`*^9}, {
  3.6574334143353796`*^9, 3.657433424849798*^9}, {3.657433463600266*^9, 
  3.6574335251921744`*^9}, {3.6574335798078704`*^9, 3.657433633128764*^9}},
 CellTags->{
  "ID:59300728", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] alternative \[ScriptCapitalZ]"},
 CellID->59300728],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], 
     RowBox[{"a", 
      UnderscriptBox["+", "\[ScriptCapitalZ]"], "c"}]}], "\[And]", 
    RowBox[{"b", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], "a"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"b", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], 
   RowBox[{"b", 
    UnderscriptBox["+", "\[ScriptCapitalZ]"], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"sgn", "[", "a", "]"}], 
       UnderscriptBox["*", "\[ScriptCapitalZ]"], 
       RowBox[{"sgn", "[", "b", "]"}]}], ")"}], 
     UnderscriptBox["*", "\[ScriptCapitalZ]"], 
     "c"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Succeeds] +' \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:698610132", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6573582363481927`*^9, 3.657358300916706*^9}, {
  3.657358336220568*^9, 3.6573583410097766`*^9}, {3.6573583777946415`*^9, 
  3.65735840573429*^9}, {3.657358474764412*^9, 3.657358477260416*^9}, {
  3.6573585579593573`*^9, 3.657358559924961*^9}},
 CellTags->{
  "ID:698610132", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] +' \[ScriptCapitalZ]"},
 CellID->698610132],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->577559729]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->17576423],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership sgn \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership sgn \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
           RowBox[{"sgn", "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
           RowBox[{"sgn", "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:211178897"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p211178897-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p211178897-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:211178897", "Proof|ID:211178897-1"},
 CellID->273631166],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p211178897-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->138417704]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* sgn \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* sgn \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{"sgn", "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 RowBox[{"sgn", "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}], "\[Equal]", 
               UnderscriptBox["1", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[Or]", 
             RowBox[{
               RowBox[{
                 RowBox[{"sgn", "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 RowBox[{"sgn", "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["Minus", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 UnderscriptBox["1", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{"sgn", "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 RowBox[{"sgn", "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}], "\[Equal]", 
               UnderscriptBox["1", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[Or]", 
             RowBox[{
               RowBox[{
                 RowBox[{"sgn", "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 RowBox[{"sgn", "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["Minus", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 UnderscriptBox["1", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:118405539"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p118405539-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p118405539-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:118405539", "Proof|ID:118405539-1"},
 CellID->55870784],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p118405539-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->10869643]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 minimal \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 minimal \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["a", "ExpressionVariable"]}], 
             TagBox[")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["a", "ExpressionVariable"]}], 
             TagBox[
             ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:88914628"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p88914628-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p88914628-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:88914628", "Proof|ID:88914628-1"},
 CellID->298080950],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p88914628-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->276114159]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(twice \[Succeeds] twice \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(twice \[Succeeds] twice \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["a", "ExpressionVariable"]}], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["a", "ExpressionVariable"]}], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:69480817"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p69480817-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p69480817-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:69480817", "Proof|ID:69480817-1"},
 CellID->8097116],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p69480817-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->109020929]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Succeeds] + \[DoubleRightArrow] \
\[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Succeeds] + \[DoubleRightArrow] \
\[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
               RowBox[{"0", ">", 
                 StyleBox["c", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[GreaterEqual]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                   TagBox[")", "AutoParentheses"]}]}]}], "\[Or]", 
             RowBox[{
               RowBox[{"0", ">", 
                 StyleBox["a", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ">", "0"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                   TagBox[")", "AutoParentheses"]}], ">", 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
               RowBox[{"0", ">", 
                 StyleBox["c", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[GreaterEqual]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                   TagBox[")", "AutoParentheses"]}]}]}], "\[Or]", 
             RowBox[{
               RowBox[{"0", ">", 
                 StyleBox["a", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ">", "0"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                   TagBox[")", "AutoParentheses"]}], ">", 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:24170142"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p24170142-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p24170142-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:24170142", "Proof|ID:24170142-1"},
 CellID->297064343],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p24170142-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->175955906]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Succeeds] + \[DoubleLeftArrow] \[ScriptCapitalZ])\"\>",
        "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Succeeds] + \[DoubleLeftArrow] \[ScriptCapitalZ])\"\>",
        "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
               RowBox[{"0", ">", 
                 StyleBox["c", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[GreaterEqual]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                   TagBox[")", "AutoParentheses"]}]}]}], "\[Or]", 
             RowBox[{
               RowBox[{"0", ">", 
                 StyleBox["a", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ">", "0"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                   TagBox[")", "AutoParentheses"]}], ">", 
                 StyleBox["c", "ExpressionVariable"]}]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
               RowBox[{"0", ">", 
                 StyleBox["c", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[GreaterEqual]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                   TagBox[")", "AutoParentheses"]}]}]}], "\[Or]", 
             RowBox[{
               RowBox[{"0", ">", 
                 StyleBox["a", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ">", "0"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                   TagBox[")", "AutoParentheses"]}], ">", 
                 StyleBox["c", "ExpressionVariable"]}]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:302704588"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p302704588-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p302704588-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:302704588", "Proof|ID:302704588-1"},
 CellID->172202783],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p302704588-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->54069213]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Succeeds] + \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Succeeds] + \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["c", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{"0", ">", 
                   StyleBox["c", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[GreaterEqual]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{"0", ">", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                    TagBox[")", "AutoParentheses"]}], ">", 
                   StyleBox["c", "ExpressionVariable"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["c", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{"0", ">", 
                   StyleBox["c", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[GreaterEqual]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{"0", ">", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                    TagBox[")", "AutoParentheses"]}], ">", 
                   StyleBox["c", "ExpressionVariable"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:131195953"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p131195953-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p131195953-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:131195953", "Proof|ID:131195953-1"},
 CellID->88300092],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p131195953-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->414777165]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Succeeds] - \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Succeeds] - \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["c", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], "\[GreaterEqual]", 
                   
                   StyleBox["c", "ExpressionVariable"]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{"0", ">", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{"0", ">", 
                   StyleBox["c", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ">", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["c", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], "\[GreaterEqual]", 
                   
                   StyleBox["c", "ExpressionVariable"]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{"0", ">", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{"0", ">", 
                   StyleBox["c", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ">", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:129207038"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p129207038-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p129207038-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:129207038", "Proof|ID:129207038-1"},
 CellID->33212788],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p129207038-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->378179640]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Succeeds] alternative \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Succeeds] alternative \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ">", 
                   StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[GreaterEqual]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["a", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{"0", ">", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ">", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["a", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}], ">", 
                   StyleBox["b", "ExpressionVariable"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ">", 
                   StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[GreaterEqual]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["a", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{"0", ">", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ">", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["a", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}], ">", 
                   StyleBox["b", "ExpressionVariable"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:59300728"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p59300728-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p59300728-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:59300728", "Proof|ID:59300728-1"},
 CellID->946864890],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p59300728-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->387614559]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Succeeds] +' \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Succeeds] +' \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 StyleBox["c", "ExpressionVariable"]}]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["a", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{"sgn", "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], 
                   RowBox[{"sgn", "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 StyleBox["c", "ExpressionVariable"]}]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["a", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{"sgn", "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], 
                   RowBox[{"sgn", "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:698610132"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p698610132-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p698610132-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:698610132", "Proof|ID:698610132-1"},
 CellID->357893767],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p698610132-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->550606355]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->122261486],

Cell[CellGroupData[{

Cell["Lemma: Reducibility in \[ScriptCapitalZ]", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6573564581918693`*^9, 3.6573564639482794`*^9}},
 CellID->279811452],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 310863676}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657357808485241*^9, 3.6573578174396567`*^9}},
 CellTags->{
  "ID:310863676", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->310863676],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"a", ",", 
      RowBox[{
       UnderscriptBox["-", "\[ScriptCapitalZ]"], "c"}], ",", "m"}], "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"a", ",", "c", ",", 
      RowBox[{
       UnderscriptBox["-", "\[ScriptCapitalZ]"], "m"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible - \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:471044386", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658041421056538*^9, 3.6580414410557733`*^9}, {
  3.658041626680499*^9, 3.658041628131302*^9}, {3.658041742573103*^9, 
  3.6580417499207153`*^9}},
 CellTags->{
  "ID:471044386", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible - \[ScriptCapitalZ]"},
 CellID->471044386],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"a", ",", "c", ",", "m"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"m", "\[NotEqual]", "0"}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[Or]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             RowBox[{"a", ">", "0"}], "\[And]", 
             RowBox[{
              RowBox[{"m", "*", "c"}], ">", "0"}], "\[And]", 
             RowBox[{
              RowBox[{"a", "+", "a"}], "\[GreaterEqual]", 
              RowBox[{"m", "*", "c"}]}]}]},
           {
            RowBox[{
             RowBox[{"0", ">", "a"}], "\[And]", 
             RowBox[{"0", ">", 
              RowBox[{"m", "*", "c"}]}], "\[And]", 
             RowBox[{
              RowBox[{"m", "*", "c"}], ">", 
              RowBox[{"a", "+", "a"}]}]}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible alternative \[ScriptCapitalZ])", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:379223163", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6574379894368315`*^9, 3.6574381172010555`*^9}, {
  3.657438498793326*^9, 3.6574384990273266`*^9}},
 CellTags->{
  "ID:379223163", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible alternative \[ScriptCapitalZ]"},
 CellID->379223163],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
       RowBox[{"a", ",", "c", ",", "m"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
       RowBox[{"b", ",", "c", ",", "m"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"sgn", "[", "a", "]"}], "\[Equal]", 
      RowBox[{"sgn", "[", "b", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible sgn \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:176657795", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6578799170043344`*^9, 3.6578799522135963`*^9}},
 CellTags->{
  "ID:176657795", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible sgn \[ScriptCapitalZ]"},
 CellID->176657795],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"a", ",", "c"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", "\[NotEqual]", "0"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Or]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{"a", ">", "0"}], "\[And]", 
           RowBox[{
            RowBox[{"a", "+", "a"}], "\[GreaterEqual]", 
            RowBox[{
            "\[LeftBracketingBar]", "c", "\[RightBracketingBar]"}]}]}]},
         {
          RowBox[{
           RowBox[{"0", ">", "a"}], "\[And]", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"(", 
              RowBox[{"a", "+", "a"}], ")"}]}], ">", 
            RowBox[{"\[LeftBracketingBar]", "c", "\[RightBracketingBar]"}]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(isReducible alternative' \[DoubleRightArrow] \[ScriptCapitalZ])", 
         "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:217289551", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6574427262882366`*^9, 3.657442735008652*^9}},
 CellTags->{
  "ID:217289551", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", 
   "isReducible alternative' \[DoubleRightArrow] \[ScriptCapitalZ]"},
 CellID->217289551],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", "\[NotEqual]", "0"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Or]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{"a", ">", "0"}], "\[And]", 
           RowBox[{
            RowBox[{"a", "+", "a"}], "\[GreaterEqual]", 
            RowBox[{
            "\[LeftBracketingBar]", "c", "\[RightBracketingBar]"}]}]}]},
         {
          RowBox[{
           RowBox[{"0", ">", "a"}], "\[And]", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"(", 
              RowBox[{"a", "+", "a"}], ")"}]}], ">", 
            RowBox[{"\[LeftBracketingBar]", "c", "\[RightBracketingBar]"}]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"a", ",", "c"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(isReducible alternative' \[DoubleLeftArrow] \[ScriptCapitalZ])", 
         "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:497414030", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6574427402814617`*^9, 3.6574427452734704`*^9}},
 CellTags->{
  "ID:497414030", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", 
   "isReducible alternative' \[DoubleLeftArrow] \[ScriptCapitalZ]"},
 CellID->497414030],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"a", ",", "c"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", "\[NotEqual]", "0"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Or]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{"a", ">", "0"}], "\[And]", 
           RowBox[{
            RowBox[{"a", "+", "a"}], "\[GreaterEqual]", 
            RowBox[{
            "\[LeftBracketingBar]", "c", "\[RightBracketingBar]"}]}]}]},
         {
          RowBox[{
           RowBox[{"0", ">", "a"}], "\[And]", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"(", 
              RowBox[{"a", "+", "a"}], ")"}]}], ">", 
            RowBox[{"\[LeftBracketingBar]", "c", "\[RightBracketingBar]"}]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(isReducible alternative' \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:464408938", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6574374799069347`*^9, 3.657437557846671*^9}, {
  3.6574375954739375`*^9, 3.6574376113235655`*^9}, {3.657442655651313*^9, 
  3.6574426615481234`*^9}},
 CellTags->{
  "ID:464408938", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible alternative' \[ScriptCapitalZ]"},
 CellID->464408938],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"a", ",", 
    RowBox[{
     UnderscriptBox["-", "\[ScriptCapitalZ]"], "c"}]}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"a", ",", "c"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible -' \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:304230885", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6580410820669427`*^9, 3.6580410969337687`*^9}, {
  3.6580420277572036`*^9, 3.6580420296760073`*^9}},
 CellTags->{
  "ID:304230885", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible -' \[ScriptCapitalZ]"},
 CellID->304230885],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"{", "c", "}"}]], "\[ScriptCapitalZ]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["-", "\[ScriptCapitalZ]"], "a"}], 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", 
      RowBox[{"{", "c", "}"}]], "\[ScriptCapitalZ]"], 
    RowBox[{
     UnderscriptBox["-", "\[ScriptCapitalZ]"], "b"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow] - \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:97311931", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6578854824361134`*^9, 3.6578855148685703`*^9}, {
  3.657885599826319*^9, 3.6578856061755304`*^9}, {3.661842805676345*^9, 
  3.661842810683954*^9}},
 CellTags->{
  "ID:97311931", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[RightArrow] - \[ScriptCapitalZ]"},
 CellID->97311931],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
       RowBox[{"a", ",", "c", ",", "m"}], "]"}], "\[And]", 
      RowBox[{"b", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], "a"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"b", ",", "c", ",", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"sgn", "[", "a", "]"}], 
         UnderscriptBox["*", "\[ScriptCapitalZ]"], 
         RowBox[{"sgn", "[", "b", "]"}]}], ")"}], 
       UnderscriptBox["*", "\[ScriptCapitalZ]"], "m"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible \[Succeeds] \[ScriptCapitalZ])", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:99245919", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6573578360192895`*^9, 3.6573578610261335`*^9}, {
  3.657358663477943*^9, 3.6573586826035767`*^9}},
 CellTags->{
  "ID:99245919", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds] \[ScriptCapitalZ]"},
 CellID->99245919],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"a", ",", "c"}], "]"}], "\[And]", 
    RowBox[{"b", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], "a"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"b", ",", "c"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(isReducible \[Succeeds]' \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:162605324", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657356467817086*^9, 3.6573565459576235`*^9}, 
   3.657356824199712*^9, {3.657357610829894*^9, 3.657357628379925*^9}, {
   3.6573578207624626`*^9, 3.657357821948065*^9}},
 CellTags->{
  "ID:162605324", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds]' \[ScriptCapitalZ]"},
 CellID->162605324],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"a", ",", "c"}], "]"}], "\[And]", 
    RowBox[{"c", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], "b"}], "\[And]", 
    RowBox[{"b", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalZ]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"a", ",", "b"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(isReducible \[Succeeds]'' \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:40987937", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6579499888107877`*^9, 3.6579500342546673`*^9}},
 CellTags->{
  "ID:40987937", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds]'' \[ScriptCapitalZ]"},
 CellID->40987937],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->51573105]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->608032019],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible - \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible - \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["Minus", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["Minus", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["Minus", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["Minus", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:471044386"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p471044386-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p471044386-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:471044386", "Proof|ID:471044386-1"},
 CellID->450891268],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p471044386-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->551822482]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible alternative \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible alternative \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[NotEqual]", "0"}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], " ", 
                    StyleBox["c", "ExpressionVariable"]}], ">", "0"}], 
                   "\[And]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], "\[GreaterEqual]", 
                    
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], " ", 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[Or]", 
                 RowBox[{
                   RowBox[{"0", ">", 
                    StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{"0", ">", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], " ", 
                    StyleBox["c", "ExpressionVariable"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], " ", 
                    StyleBox["c", "ExpressionVariable"]}], ">", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[NotEqual]", "0"}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], " ", 
                    StyleBox["c", "ExpressionVariable"]}], ">", "0"}], 
                   "\[And]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], "\[GreaterEqual]", 
                    
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], " ", 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[Or]", 
                 RowBox[{
                   RowBox[{"0", ">", 
                    StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{"0", ">", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], " ", 
                    StyleBox["c", "ExpressionVariable"]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], " ", 
                    StyleBox["c", "ExpressionVariable"]}], ">", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:379223163"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p379223163-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p379223163-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:379223163", "Proof|ID:379223163-1"},
 CellID->249907335],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p379223163-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->557242539]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible sgn \[ScriptCapitalZ])\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible sgn \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"sgn", "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{"sgn", "[", 
               StyleBox["b", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"sgn", "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{"sgn", "[", 
               StyleBox["b", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:176657795"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p176657795-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p176657795-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:176657795", "Proof|ID:176657795-1"},
 CellID->334897452],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p176657795-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->205455470]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible alternative' \[DoubleRightArrow] \
\[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible alternative' \[DoubleRightArrow] \
\[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", "0"}], 
             "\[And]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], "\[GreaterEqual]", 
                   
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{"0", ">", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                    TagBox[")", "AutoParentheses"]}], ">", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", "0"}], 
             "\[And]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], "\[GreaterEqual]", 
                   
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{"0", ">", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                    TagBox[")", "AutoParentheses"]}], ">", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:217289551"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p217289551-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p217289551-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:217289551", "Proof|ID:217289551-1"},
 CellID->880922821],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p217289551-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->276481536]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible alternative' \[DoubleLeftArrow] \
\[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible alternative' \[DoubleLeftArrow] \
\[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", "0"}], 
             "\[And]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], "\[GreaterEqual]", 
                   
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{"0", ">", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                    TagBox[")", "AutoParentheses"]}], ">", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", "0"}], 
             "\[And]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], "\[GreaterEqual]", 
                   
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{"0", ">", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                    TagBox[")", "AutoParentheses"]}], ">", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:497414030"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p497414030-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p497414030-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:497414030", "Proof|ID:497414030-1"},
 CellID->110073726],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p497414030-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->23417434]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible alternative' \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible alternative' \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", "0"}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], "\[GreaterEqual]", 
                    
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}], "\[Or]", 
                 RowBox[{
                   RowBox[{"0", ">", 
                    StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                    TagBox[")", "AutoParentheses"]}], ">", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", "0"}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ">", "0"}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], "\[GreaterEqual]", 
                    
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}], "\[Or]", 
                 RowBox[{
                   RowBox[{"0", ">", 
                    StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "+", 
                    StyleBox["a", "ExpressionVariable"]}], ")"}]}], 
                    TagBox[")", "AutoParentheses"]}], ">", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:464408938"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p464408938-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p464408938-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:464408938", "Proof|ID:464408938-1"},
 CellID->577762604],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p464408938-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->591630630]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible -' \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible -' \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["Minus", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["Minus", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:304230885"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p304230885-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p304230885-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:304230885", "Proof|ID:304230885-1"},
 CellID->115875544],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p304230885-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->93717885]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] - \[ScriptCapitalZ])\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] - \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["c", "ExpressionVariable"]], 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["Minus", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               StyleBox["c", "ExpressionVariable"]], 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               UnderscriptBox["Minus", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["c", "ExpressionVariable"]], 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["Minus", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               StyleBox["c", "ExpressionVariable"]], 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             RowBox[{
               UnderscriptBox["Minus", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:97311931"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p97311931-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p97311931-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:97311931", "Proof|ID:97311931-1"},
 CellID->616242356],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p97311931-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->16956372]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible \[Succeeds] \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible \[Succeeds] \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["a", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   RowBox[{"sgn", "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], 
                   RowBox[{"sgn", "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 StyleBox["m", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["a", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   RowBox[{"sgn", "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], 
                   RowBox[{"sgn", "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 StyleBox["m", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:99245919"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p99245919-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p99245919-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:99245919", "Proof|ID:99245919-1"},
 CellID->518473742],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p99245919-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->20241181]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible \[Succeeds]' \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible \[Succeeds]' \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["a", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["a", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:162605324"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p162605324-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p162605324-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:162605324", "Proof|ID:162605324-1"},
 CellID->291524205],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p162605324-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->569496899]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible \[Succeeds]'' \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible \[Succeeds]'' \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:40987937"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p40987937-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p40987937-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:40987937", "Proof|ID:40987937-1"},
 CellID->103343062],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p40987937-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->9396178]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->31873811],

Cell[CellGroupData[{

Cell["Lemma: Properties of lr", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.657349748392948*^9, 3.657349754024558*^9}, {
  3.657355840333784*^9, 3.657355843453789*^9}, {3.657356445743047*^9, 
  3.6573564509534564`*^9}},
 CellID->687299096],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
   RowBox[{"[", "c", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 11694492}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657355618485794*^9, 3.6573556454582415`*^9}, {
   3.6573556898247194`*^9, 3.6573556901055202`*^9}, 3.6573564173353977`*^9},
 CellTags->{
  "ID:11694492", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->11694492],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership lr \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:224215037", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6573557753596697`*^9, 3.657355785842888*^9}, 
   3.657356418755*^9},
 CellTags->{
  "ID:224215037", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership lr \[ScriptCapitalZ]"},
 CellID->224215037],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}], "\[Equal]", 
    "0"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"c", "\[Equal]", "0"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lr = 0 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:78177024", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657620387971964*^9, 3.657620404882393*^9}, {
  3.6576209948390303`*^9, 3.6576209951198306`*^9}},
 CellTags->{
  "ID:78177024", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr = 0 \[ScriptCapitalZ]"},
 CellID->78177024],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}], ">", "0"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"even", "[", "c", "]"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"c", "\[NotEqual]", "0"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lr > 0 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:634190310", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657611289506563*^9, 3.657611360642688*^9}, {
  3.657620996414633*^9, 3.6576209966954336`*^9}, {3.6578602605097437`*^9, 
  3.657860265253152*^9}, {3.6578603597123175`*^9, 3.65786036088232*^9}},
 CellTags->{
  "ID:634190310", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr > 0 \[ScriptCapitalZ]"},
 CellID->634190310],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"0", ">", 
    RowBox[{
     UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"even", "[", "c", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 > lr \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:588999973", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6576113718279076`*^9, 3.6576113858991327`*^9}, {
  3.657611715707312*^9, 3.65761172020012*^9}},
 CellTags->{
  "ID:588999973", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "0 > lr \[ScriptCapitalZ]"},
 CellID->588999973],

Cell[BoxData[
 RowBox[{
  RowBox[{"even", "[", "c", "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}], "+", 
     RowBox[{
      UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}]}], 
    "\[Equal]", 
    RowBox[{"\[LeftBracketingBar]", "c", "\[RightBracketingBar]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lr twice even \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:57477705", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657862830339863*^9, 3.6578628854163594`*^9}},
 CellTags->{
  "ID:57477705", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr twice even \[ScriptCapitalZ]"},
 CellID->57477705],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"even", "[", "c", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}], "+", 
     RowBox[{
      UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}]}], 
    "\[Equal]", 
    RowBox[{"-", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "c", "\[RightBracketingBar]"}], "+", 
       "1"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lr twice odd \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:206701743", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65786289135997*^9, 3.6578629203604207`*^9}},
 CellTags->{
  "ID:206701743", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr twice odd \[ScriptCapitalZ]"},
 CellID->206701743],

Cell[BoxData[
 RowBox[{
  RowBox[{"c", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalZ]"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}], ",", "c"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lr isReducible \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:34042360", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6573556139617863`*^9, 3.65735566176027*^9}, {
  3.657355692570324*^9, 3.657355703178343*^9}, {3.657356419878202*^9, 
  3.657356422296206*^9}},
 CellTags->{
  "ID:34042360", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isReducible \[ScriptCapitalZ]"},
 CellID->34042360],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
    RowBox[{"[", "a", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], "a"}], "\[Implies]", 
    
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
       RowBox[{"a", ",", "c"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lr minimal isReducible \[ScriptCapitalZ])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:692837602", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657355669965884*^9, 3.6573557418820105`*^9}, {
  3.6573564235754085`*^9, 3.6573564304550204`*^9}},
 CellTags->{
  "ID:692837602", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr minimal isReducible \[ScriptCapitalZ]"},
 CellID->692837602],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}], ",", "c", 
      ",", "m"}], "]"}], "\[Implies]", 
    RowBox[{
     RowBox[{"sgn", "[", 
      RowBox[{
       UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}], "]"}], 
     "\[NotEqual]", 
     RowBox[{"sgn", "[", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}], 
       UnderscriptBox["-", "\[ScriptCapitalZ]"], 
       RowBox[{"m", 
        UnderscriptBox["*", "\[ScriptCapitalZ]"], "c"}]}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lr isReducible sgn \[ScriptCapitalZ])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:44304045", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657875210841649*^9, 3.6578752411837025`*^9}, {
  3.6578752764709644`*^9, 3.6578753504462943`*^9}},
 CellTags->{
  "ID:44304045", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isReducible sgn \[ScriptCapitalZ]"},
 CellID->44304045],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->403862672]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->230715858],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership lr \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership lr \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
           RowBox[{
             UnderscriptBox["lr", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
           RowBox[{
             UnderscriptBox["lr", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:224215037"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p224215037-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p224215037-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:224215037", "Proof|ID:224215037-1"},
 CellID->519188695],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p224215037-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->271118078]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lr = 0 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lr = 0 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["lr", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}], "\[Equal]", "0"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[Equal]", 
             "0"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["lr", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}], "\[Equal]", "0"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[Equal]", 
             "0"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:78177024"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p78177024-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p78177024-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:78177024", "Proof|ID:78177024-1"},
 CellID->336735725],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p78177024-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->326983128]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lr > 0 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lr > 0 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["lr", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}], ">", "0"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{"even", "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", "0"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["lr", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}], ">", "0"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{"even", "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", "0"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:634190310"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p634190310-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p634190310-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:634190310", "Proof|ID:634190310-1"},
 CellID->77986082],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p634190310-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->508208461]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 > lr \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 > lr \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"0", ">", 
             RowBox[{
               UnderscriptBox["lr", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"even", "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"0", ">", 
             RowBox[{
               UnderscriptBox["lr", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"even", "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:588999973"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p588999973-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p588999973-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:588999973", "Proof|ID:588999973-1"},
 CellID->1365258],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p588999973-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->429365998]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lr twice even \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lr twice even \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"even", "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "+", 
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}], "\[Equal]", 
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["c", "ExpressionVariable"], 
               "\[RightBracketingBar]"}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"even", "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "+", 
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}], "\[Equal]", 
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["c", "ExpressionVariable"], 
               "\[RightBracketingBar]"}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:57477705"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p57477705-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p57477705-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:57477705", "Proof|ID:57477705-1"},
 CellID->253970289],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p57477705-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1669627]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lr twice odd \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lr twice odd \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"even", "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "+", 
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}], "\[Equal]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"-", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}], ")"}]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"even", "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "+", 
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}], "\[Equal]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"-", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["c", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}], ")"}]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:206701743"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p206701743-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p206701743-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:206701743", "Proof|ID:206701743-1"},
 CellID->43456446],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p206701743-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->37592898]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lr isReducible \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lr isReducible \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:34042360"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p34042360-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p34042360-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:34042360", "Proof|ID:34042360-1"},
 CellID->169464375],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p34042360-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->325628575]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lr minimal isReducible \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lr minimal isReducible \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["lr", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["lr", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:692837602"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p692837602-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p692837602-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:692837602", "Proof|ID:692837602-1"},
 CellID->371670326],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p692837602-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->251346555]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lr isReducible sgn \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lr isReducible sgn \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"sgn", "[", 
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "]"}], 
             "\[NotEqual]", 
             RowBox[{"sgn", "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["-", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], 
                   StyleBox["c", "ExpressionVariable"]}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"sgn", "[", 
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "]"}], 
             "\[NotEqual]", 
             RowBox[{"sgn", "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["-", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], 
                   StyleBox["c", "ExpressionVariable"]}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:44304045"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p44304045-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p44304045-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:44304045", "Proof|ID:44304045-1"},
 CellID->477556229],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p44304045-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->413433274]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->580411371],

Cell[CellGroupData[{

Cell["Lemma: Alternative Characterization of Auxiliary Notions", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6573562540967107`*^9, 3.657356272598343*^9}, {
  3.657356717245924*^9, 3.657356740021964*^9}, {3.657453925769701*^9, 
  3.657453935550918*^9}},
 CellID->285853213],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "c1", ",", "c2", ",", "m1", ",", "m2"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 16238117}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65735627657635*^9, 3.6573562846727643`*^9}, {
  3.657453698290101*^9, 3.657453727945753*^9}, {3.6574539151928825`*^9, 
  3.657453916300484*^9}},
 CellTags->{
  "ID:16238117", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->16238117],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "i"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"i", "\[Element]", 
      SubscriptBox["\[DoubleStruckCapitalZ]", 
       RowBox[{"1", ",", "\[Ellipsis]", ",", 
        RowBox[{
         UnderscriptBox["multN", "\[ScriptCapitalZ]"], "[", "c1", "]"}]}]]}], 
     
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"i", "\[Equal]", "1"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] multN \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:468309069", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65745407359556*^9, 3.657454102830012*^9}, 
   3.6574542721371093`*^9},
 CellTags->{
  "ID:468309069", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Element] multN \[ScriptCapitalZ]"},
 CellID->468309069],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
    UnderscriptBox["multN", "\[ScriptCapitalZ]"], "[", "c1", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 560244434}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657454115481634*^9, 3.6574541276652555`*^9}},
 CellTags->{
  "ID:560244434", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->560244434],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areIrrelative", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"c1", ",", "c2", ",", "m1", ",", "m2", ",", "i", ",", "j"}], "]"}],
   "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"c1", "\[NotEqual]", "c2"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(areIrrelative alternative \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:142903111", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657453905333665*^9, 3.65745398815421*^9}},
 CellTags->{
  "ID:142903111", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "areIrrelative alternative \[ScriptCapitalZ]"},
 CellID->142903111],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isCommonReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"c1", "\[NotEqual]", "c2"}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
         RowBox[{"a", ",", "c1"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
         RowBox[{"a", ",", "c2"}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(isCommonReducible alternative \[ScriptCapitalZ])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:36948313", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657453702080908*^9, 3.657453796180273*^9}, {
  3.657453951244546*^9, 3.657453952726548*^9}},
 CellTags->{
  "ID:36948313", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommonReducible alternative \[ScriptCapitalZ]"},
 CellID->36948313],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isCommonReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"c1", "\[NotEqual]", "c2"}]},
       {
        RowBox[{"c1", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalZ]"]}]},
       {
        RowBox[{"c2", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalZ]"]}]},
       {
        RowBox[{"\[Not]", 
         RowBox[{
          RowBox[{
           SubscriptBox["min", 
            UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"]], "[", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{
             RowBox[{
              UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c1", "]"}], 
             ",", 
             RowBox[{
              UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c2", "]"}]}], 
            "\[RightAngleBracket]"}], "]"}], 
          UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], "a"}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(isCommonReducible alternative' \[ScriptCapitalZ])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:46630982", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657880390301566*^9, 3.657880456398882*^9}, {
  3.6578819685629387`*^9, 3.6578819776421547`*^9}, {3.657882234184605*^9, 
  3.657882259269449*^9}, {3.6578824216969347`*^9, 3.6578824247389393`*^9}},
 CellTags->{
  "ID:46630982", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommonReducible alternative' \[ScriptCapitalZ]"},
 CellID->46630982],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->96563173]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1163307],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] multN \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] multN \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", 
                RowBox[{"1", ",", "\[Ellipsis]", ",", 
                  RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["i", "ExpressionVariable"], "\[Equal]", 
             "1"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", 
                RowBox[{"1", ",", "\[Ellipsis]", ",", 
                  RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["i", "ExpressionVariable"], "\[Equal]", 
             "1"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:468309069"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p468309069-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p468309069-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:468309069", "Proof|ID:468309069-1"},
 CellID->223234557],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p468309069-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2459748]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(areIrrelative alternative \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(areIrrelative alternative \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areIrrelative", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox[
             "c2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areIrrelative", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox[
             "c2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:142903111"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p142903111-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p142903111-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:142903111", "Proof|ID:142903111-1"},
 CellID->4630108],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p142903111-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->182063627]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCommonReducible alternative \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCommonReducible alternative \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:36948313"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p36948313-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p36948313-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:36948313", "Proof|ID:36948313-1"},
 CellID->239177438],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p36948313-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->14134654]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCommonReducible alternative' \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCommonReducible alternative' \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    RowBox[{
                    SubscriptBox["min", 
                    UnderscriptBox["\[Succeeds]", 
                    RowBox[{"ReductionIntegers", "[", "]"}]]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}], 
                    "\[RightAngleBracket]"}], "]"}], 
                    UnderscriptBox["\[Succeeds]", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], 
                    StyleBox["a", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
               RowBox[{
                 StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    RowBox[{
                    SubscriptBox["min", 
                    UnderscriptBox["\[Succeeds]", 
                    RowBox[{"ReductionIntegers", "[", "]"}]]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}], 
                    "\[RightAngleBracket]"}], "]"}], 
                    UnderscriptBox["\[Succeeds]", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], 
                    StyleBox["a", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:46630982"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p46630982-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p46630982-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:46630982", "Proof|ID:46630982-1"},
 CellID->250839951],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p46630982-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->35813888]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->113752866],

Cell[CellGroupData[{

Cell["Lemma: Minimal Non-Trivial Common Reducibles", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6578703541480913`*^9, 3.65787036450651*^9}},
 CellID->614730048],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
     RowBox[{"[", "c", "]"}]}]], 
   RowBox[{"c", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalZ]"]}]], 
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"c", ",", 
    RowBox[{
     UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c", "]"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lr isMNTCR unary \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:69669781", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657870541894421*^9, 3.657870574248878*^9}},
 CellTags->{
  "ID:69669781", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isMNTCR unary \[ScriptCapitalZ]"},
 CellID->69669781],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}]], 
  RowBox[{
   RowBox[{"c1", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalZ]"]}], "\[And]", 
   RowBox[{"c2", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalZ]"]}], "\[And]", 
   RowBox[{"c1", "\[NotEqual]", "c2"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 65735690}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.6578715508125935`*^9},
 CellTags->{
  "ID:65735690", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->65735690],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
    UnderscriptBox["multN", "\[ScriptCapitalZ]"], "[", "c1", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 431845789}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.6578715591742077`*^9},
 CellTags->{
  "ID:431845789", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->431845789],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["isNTCR", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"c1", ",", "c2", ",", 
   RowBox[{
    SubscriptBox["min", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"]], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c1", "]"}], ",", 
      RowBox[{
       UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c2", "]"}]}], 
     "\[RightAngleBracket]"}], "]"}], ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lr isNTCR \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:96649542", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6578715786742425`*^9, 3.6578715811390467`*^9}},
 CellTags->{
  "ID:96649542", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isNTCR \[ScriptCapitalZ]"},
 CellID->96649542],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["isMNTCR", "\[ScriptCapitalZ]"], "[", 
  RowBox[{"c1", ",", "c2", ",", 
   RowBox[{
    SubscriptBox["min", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"]], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c1", "]"}], ",", 
      RowBox[{
       UnderscriptBox["lr", "\[ScriptCapitalZ]"], "[", "c2", "]"}]}], 
     "\[RightAngleBracket]"}], "]"}], ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lr isMNTCR \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:132861224", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6578703877661505`*^9, 3.657870526918395*^9}, {
  3.657871061813134*^9, 3.6578710662279425`*^9}, {3.6578712560490756`*^9, 
  3.6578712594810815`*^9}, {3.6578715493617907`*^9, 3.6578715577858057`*^9}},
 CellTags->{
  "ID:132861224", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isMNTCR \[ScriptCapitalZ]"},
 CellID->132861224],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->65813332]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->76710532],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lr isMNTCR unary \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lr isMNTCR unary \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["lr", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:69669781"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p69669781-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p69669781-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:69669781", "Proof|ID:69669781-1"},
 CellID->22101709],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p69669781-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->279677689]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lr isNTCR \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lr isNTCR \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               RowBox[{
                 SubscriptBox["min", 
                  UnderscriptBox["\[Succeeds]", 
                   RowBox[{"ReductionIntegers", "[", "]"}]]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], "]"}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               RowBox[{
                 SubscriptBox["min", 
                  UnderscriptBox["\[Succeeds]", 
                   RowBox[{"ReductionIntegers", "[", "]"}]]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], "]"}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:96649542"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p96649542-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p96649542-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:96649542", "Proof|ID:96649542-1"},
 CellID->151366490],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p96649542-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->624894588]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lr isMNTCR \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lr isMNTCR \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               RowBox[{
                 SubscriptBox["min", 
                  UnderscriptBox["\[Succeeds]", 
                   RowBox[{"ReductionIntegers", "[", "]"}]]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], "]"}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIntegers", "[", "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               RowBox[{
                 SubscriptBox["min", 
                  UnderscriptBox["\[Succeeds]", 
                   RowBox[{"ReductionIntegers", "[", "]"}]]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lr", 
                    RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}], 
                   "\[RightAngleBracket]"}], "]"}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:132861224"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p132861224-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p132861224-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:132861224", "Proof|ID:132861224-1"},
 CellID->94818643],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p132861224-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->573039597]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R5-R8 (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.657347133966755*^9, 3.657347134699956*^9}, {
  3.6578848630516253`*^9, 3.657884866312031*^9}, {3.6578911126200275`*^9, 
  3.657891116691634*^9}},
 CellID->703472201],

Cell["", "OpenEnvironment",
 CellID->425914962],

Cell[CellGroupData[{

Cell["Lemma: \[ScriptCapitalZ] Satisfies R5-R8", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594023267892793*^9, 3.594023282533146*^9}, {
  3.594023678997537*^9, 3.594023689893055*^9}, {3.65725451491683*^9, 
  3.6572545418736773`*^9}, {3.6574535479214373`*^9, 3.6574535504642415`*^9}},
 CellID->258230160],

Cell[BoxData[
 RowBox[{"R5", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R5 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:595906550", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028542338926*^9, 3.59402857094131*^9}, 
   3.65725644969843*^9},
 CellTags->{
  "ID:595906550", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R5 \[ScriptCapitalZ]"},
 CellID->595906550],

Cell[BoxData[
 RowBox[{"R6", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R6 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:384714041", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028544815682*^9, 3.594028571789276*^9}, 
   3.657256450837232*^9},
 CellTags->{
  "ID:384714041", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R6 \[ScriptCapitalZ]"},
 CellID->384714041],

Cell[BoxData[
 RowBox[{"R7", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R7 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:739712574", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028547623214*^9, 3.594028572563634*^9}, 
   3.657256452054034*^9},
 CellTags->{
  "ID:739712574", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R7 \[ScriptCapitalZ]"},
 CellID->739712574],

Cell[BoxData[
 RowBox[{"R8", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R8 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:782958792", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028549629948*^9, 3.594028573264369*^9}, 
   3.6572564531928363`*^9},
 CellTags->{
  "ID:782958792", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R8 \[ScriptCapitalZ]"},
 CellID->782958792],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1686190119]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1985203424],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R5 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R5 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R5", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R5", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:595906550"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p595906550-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p595906550-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:595906550", "Proof|ID:595906550-1"},
 CellID->801014817],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p595906550-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->655992473]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R6 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R6 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R6", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R6", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:384714041"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p384714041-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p384714041-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:384714041", "Proof|ID:384714041-1"},
 CellID->4523962],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p384714041-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->640491936]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R7 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R7 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R7", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R7", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:739712574"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p739712574-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p739712574-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:739712574", "Proof|ID:739712574-1"},
 CellID->802064395],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p739712574-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {61., 67.}}]], "ProofInfoBody",
 CellID->218110605]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R8 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R8 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R8", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R8", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:782958792"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p782958792-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p782958792-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:782958792", "Proof|ID:782958792-1"},
 CellID->148939037],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p782958792-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->89453536]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Result (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6573497177232943`*^9, 3.657349720500099*^9}, {
  3.6578912399796505`*^9, 3.657891244082458*^9}},
 CellID->461632795],

Cell["", "OpenEnvironment",
 CellID->288863783],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalZ] is Reduction Ring", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594023698805514*^9, 3.594023716070514*^9}, {
  3.6572543442837305`*^9, 3.65725434429933*^9}, {3.6572545213752413`*^9, 
  3.657254523621645*^9}},
 CellID->1803439939],

Cell[BoxData[
 RowBox[{"isReductionRing", "[", "\[ScriptCapitalZ]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReductionRing \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:67773804", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028605825412*^9, 3.594028616196035*^9}, 
   3.6572564551272397`*^9},
 CellTags->{
  "ID:67773804", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReductionRing \[ScriptCapitalZ]"},
 CellID->67773804],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1852266050]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->346871786],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReductionRing \[ScriptCapitalZ])\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(isReductionRing \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isReductionRing", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isReductionRing", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:67773804"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p67773804-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p67773804-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:67773804", "Proof|ID:67773804-1"},
 CellID->766315013],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p67773804-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->91080121]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Algorithmic Reduction Ring (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.657254360133358*^9, 3.6572543654685674`*^9}, {
  3.6580442672967405`*^9, 3.658044271430748*^9}},
 CellID->811185245],

Cell[CellGroupData[{

Cell["R9 & R10 (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6579468875693316`*^9, 3.6579468955565453`*^9}},
 CellID->91867520],

Cell["", "OpenEnvironment",
 CellID->209634667],

Cell[CellGroupData[{

Cell["Lemma: \[ScriptCapitalZ] Satisfies R9 and R10", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.65725452630485*^9, 3.6572545551025004`*^9}, {
  3.65794693227901*^9, 3.657946934884215*^9}},
 CellID->1132188],

Cell[BoxData[
 RowBox[{"R9", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R9 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:156995908", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028551665708*^9, 3.594028574480219*^9}, 
   3.6572569406974955`*^9},
 CellTags->{
  "ID:156995908", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R9 \[ScriptCapitalZ]"},
 CellID->156995908],

Cell[BoxData[
 RowBox[{"R10", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R10 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1051911108", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028557586333*^9, 3.594028576800571*^9}, 
   3.657256942023498*^9},
 CellTags->{
  "ID:1051911108", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R10 \[ScriptCapitalZ]"},
 CellID->1051911108],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->474076086]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->377455191],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R9 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R9 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R9", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R9", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:156995908"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p156995908-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p156995908-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:156995908", "Proof|ID:156995908-1"},
 CellID->802545094],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p156995908-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->235298705]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R10 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R10 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R10", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R10", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1051911108"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1051911108-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1051911108-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1051911108", "Proof|ID:1051911108-1"},
 CellID->67102701],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1051911108-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->284680562]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R11 & R12 (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.65794690362176*^9, 3.6579469057589636`*^9}, {
  3.657948684915696*^9, 3.6579486886597033`*^9}, {3.6580442523831143`*^9, 
  3.658044256657522*^9}},
 CellID->47141123],

Cell["", "OpenEnvironment",
 CellID->99683],

Cell[CellGroupData[{

Cell["Lemma: rdm with Positive Reductor", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.657961291253663*^9, 3.657961317758109*^9}, {
  3.6579658842703314`*^9, 3.6579658879987383`*^9}, {3.6580436015771713`*^9, 
  3.6580436102975864`*^9}},
 CellID->525282118],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "y", "\[Element]", "\[DoubleStruckCapitalN]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 160940549}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657963102589245*^9, 3.657963106333252*^9}},
 CellTags->{
  "ID:160940549", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->160940549],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "x", "\[Element]", "\[DoubleStruckCapitalZ]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 46650045}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6579613421565523`*^9, 3.6579613566489773`*^9}, {
  3.657963093182429*^9, 3.657963096661235*^9}, {3.657966060331041*^9, 
  3.657966060518241*^9}},
 CellTags->{
  "ID:46650045", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->46650045],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "\[GreaterEqual]", "0"}], "\[Implies]", 
  RowBox[{
   RowBox[{"quo", "[", 
    RowBox[{
     RowBox[{"x", "+", 
      RowBox[{"quo", "[", 
       RowBox[{"y", ",", "2"}], "]"}]}], ",", "y"}], "]"}], "\[GreaterEqual]",
    "0"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(rdm pos \[GreaterEqual] 0 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:816204582", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6580309964184065`*^9, 3.6580310089618287`*^9}},
 CellTags->{
  "ID:816204582", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos \[GreaterEqual] 0 \[ScriptCapitalZ]"},
 CellID->816204582],

Cell[BoxData[
 RowBox[{
  RowBox[{"0", "\[GreaterEqual]", "x"}], "\[Implies]", 
  RowBox[{
   RowBox[{"quo", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "x"}], "+", 
      RowBox[{"quo", "[", 
       RowBox[{
        RowBox[{"y", "-", "1"}], ",", "2"}], "]"}]}], ",", "y"}], "]"}], 
   "\[GreaterEqual]", "0"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(rdm neg \[GreaterEqual] 0 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:171042210", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658031014234638*^9, 3.6580310237818546`*^9}},
 CellTags->{
  "ID:171042210", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg \[GreaterEqual] 0 \[ScriptCapitalZ]"},
 CellID->171042210],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "\[GreaterEqual]", "0"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"quo", "[", 
       RowBox[{
        RowBox[{"x", "+", 
         RowBox[{"quo", "[", 
          RowBox[{"y", ",", "2"}], "]"}]}], ",", "y"}], "]"}], "\[Equal]", 
      "0"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"y", ">", 
      RowBox[{"x", "+", "x"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm pos 0 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:160702195", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657961389409035*^9, 3.657961444399132*^9}, 
   3.657964429111376*^9, {3.6579660081033487`*^9, 3.6579660581314373`*^9}},
 CellTags->{
  "ID:160702195", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos 0 \[ScriptCapitalZ]"},
 CellID->160702195],

Cell[BoxData[
 RowBox[{
  RowBox[{"0", "\[GreaterEqual]", "x"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"quo", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "x"}], "+", 
         RowBox[{"quo", "[", 
          RowBox[{
           RowBox[{"y", "-", "1"}], ",", "2"}], "]"}]}], ",", "y"}], "]"}], 
      "\[Equal]", "0"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"y", "\[GreaterEqual]", 
      RowBox[{"-", 
       RowBox[{"(", 
        RowBox[{"x", "+", "x"}], ")"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm neg 0 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:478744619", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657966086445487*^9, 3.6579660914686956`*^9}, {
   3.6579661279259596`*^9, 3.6579661585956135`*^9}, {3.657966199342885*^9, 
   3.6579661995612855`*^9}, 3.657966229872139*^9},
 CellTags->{
  "ID:478744619", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg 0 \[ScriptCapitalZ]"},
 CellID->478744619],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[GreaterEqual]", "0"}], "\[And]", 
    RowBox[{
     RowBox[{"quo", "[", 
      RowBox[{
       RowBox[{"x", "+", 
        RowBox[{"quo", "[", 
         RowBox[{"y", ",", "2"}], "]"}]}], ",", "y"}], "]"}], "\[NotEqual]", 
     "0"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"x", ",", "y", ",", 
    RowBox[{"quo", "[", 
     RowBox[{
      RowBox[{"x", "+", 
       RowBox[{"quo", "[", 
        RowBox[{"y", ",", "2"}], "]"}]}], ",", "y"}], "]"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm pos isReducible \[ScriptCapitalZ])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:680200521", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6580307692000074`*^9, 3.658030813379285*^9}},
 CellTags->{
  "ID:680200521", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos isReducible \[ScriptCapitalZ]"},
 CellID->680200521],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"0", ">", "x"}], "\[And]", 
    RowBox[{
     RowBox[{"quo", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "x"}], "+", 
        RowBox[{"quo", "[", 
         RowBox[{
          RowBox[{"y", "-", "1"}], ",", "2"}], "]"}]}], ",", "y"}], "]"}], 
     "\[NotEqual]", "0"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"x", ",", "y", ",", 
    RowBox[{"-", 
     RowBox[{"quo", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "x"}], "+", 
        RowBox[{"quo", "[", 
         RowBox[{
          RowBox[{"y", "-", "1"}], ",", "2"}], "]"}]}], ",", "y"}], "]"}]}]}],
    "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm neg isReducible \[ScriptCapitalZ])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:122969425", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6580308420677357`*^9, 3.658030853502556*^9}, 
   3.658032412616495*^9, 3.6580337501406446`*^9},
 CellTags->{
  "ID:122969425", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg isReducible \[ScriptCapitalZ]"},
 CellID->122969425],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], "\[Element]", 
  "\[DoubleStruckCapitalZ]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership rdm' \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:509693672", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658039370561723*^9, 3.658039380530141*^9}},
 CellTags->{
  "ID:509693672", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership rdm' \[ScriptCapitalZ]"},
 CellID->509693672],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"x", ",", "y"}], "]"}], "\[Equal]", "0"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"x", ",", "y"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm = 0' \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:576030870", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657971337809122*^9, 3.657971397479227*^9}, {
  3.6579715106262255`*^9, 3.6579715541347017`*^9}, {3.6579720474093685`*^9, 
  3.657972048688571*^9}, {3.6579720944902515`*^9, 3.657972094786652*^9}},
 CellTags->{
  "ID:576030870", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm = 0' \[ScriptCapitalZ]"},
 CellID->576030870],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"x", ",", "y", ",", 
    RowBox[{
     UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"x", ",", "y"}], "]"}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm isReducible' \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:825630046", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6580383856313906`*^9, 3.658038399421815*^9}},
 CellTags->{
  "ID:825630046", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm isReducible' \[ScriptCapitalZ]"},
 CellID->825630046],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->22057905]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->607571403],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm pos \[GreaterEqual] 0 \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm pos \[GreaterEqual] 0 \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", "0"}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{"quo", "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], "+", 
                   RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", "2"}], "]"}]}], 
                 ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             "\[GreaterEqual]", "0"}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", "0"}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{"quo", "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], "+", 
                   RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", "2"}], "]"}]}], 
                 ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             "\[GreaterEqual]", "0"}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:816204582"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p816204582-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p816204582-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:816204582", "Proof|ID:816204582-1"},
 CellID->451910702],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p816204582-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->421911099]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm neg \[GreaterEqual] 0 \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm neg \[GreaterEqual] 0 \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"0", "\[GreaterEqual]", 
             StyleBox["x", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{"quo", "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}], "+", 
                   RowBox[{"quo", "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "-", "1"}], ",", 
                    "2"}], "]"}]}], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             "\[GreaterEqual]", "0"}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"0", "\[GreaterEqual]", 
             StyleBox["x", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{"quo", "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}], "+", 
                   RowBox[{"quo", "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "-", "1"}], ",", 
                    "2"}], "]"}]}], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             "\[GreaterEqual]", "0"}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:171042210"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p171042210-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p171042210-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:171042210", "Proof|ID:171042210-1"},
 CellID->29599316],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p171042210-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->364315036]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm pos 0 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm pos 0 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", "0"}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "+", 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", "2"}], "]"}]}], 
                   ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               "0"}], "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], ">", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "+", 
                 StyleBox[
                 "x", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", "0"}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "+", 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", "2"}], "]"}]}], 
                   ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               "0"}], "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], ">", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "+", 
                 StyleBox[
                 "x", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:160702195"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p160702195-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p160702195-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:160702195", "Proof|ID:160702195-1"},
 CellID->461788959],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p160702195-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6579693025631475`*^9, 3.657969305074752*^9}},
 CellID->700076]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm neg 0 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm neg 0 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"0", "\[GreaterEqual]", 
             StyleBox["x", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}], "+", 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "-", "1"}], ",", 
                    "2"}], "]"}]}], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               "0"}], "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[GreaterEqual]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"-", 
                   RowBox[{"(", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "+", 
                    StyleBox["x", "ExpressionVariable"]}], ")"}]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{"0", "\[GreaterEqual]", 
             StyleBox["x", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}], "+", 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "-", "1"}], ",", 
                    "2"}], "]"}]}], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               "0"}], "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[GreaterEqual]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"-", 
                   RowBox[{"(", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "+", 
                    StyleBox["x", "ExpressionVariable"]}], ")"}]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:478744619"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p478744619-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p478744619-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:478744619", "Proof|ID:478744619-1"},
 CellID->1001675670],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p478744619-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->75688501]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm pos isReducible \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm pos isReducible \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", "0"}], 
             "\[And]", 
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "+", 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", "2"}], "]"}]}], 
                   ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], 
               "\[NotEqual]", "0"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{"quo", "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "+", 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", "2"}], "]"}]}], 
                   ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", "0"}], 
             "\[And]", 
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "+", 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", "2"}], "]"}]}], 
                   ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], 
               "\[NotEqual]", "0"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{"quo", "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "+", 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", "2"}], "]"}]}], 
                   ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:680200521"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p680200521-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p680200521-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:680200521", "Proof|ID:680200521-1"},
 CellID->98328192],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p680200521-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6580363101115417`*^9, 3.6580363120927453`*^9}},
 CellID->484485766]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm neg isReducible \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm neg isReducible \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"0", ">", 
               StyleBox["x", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}], "+", 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "-", "1"}], ",", 
                    "2"}], "]"}]}], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], 
               "\[NotEqual]", "0"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"-", 
                   RowBox[{"quo", "[", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}], "+", 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "-", "1"}], ",", 
                    "2"}], "]"}]}], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
                 TagBox[")", "AutoParentheses"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"0", ">", 
               StyleBox["x", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}], "+", 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "-", "1"}], ",", 
                    "2"}], "]"}]}], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], 
               "\[NotEqual]", "0"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{"-", 
                   RowBox[{"quo", "[", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}], "+", 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "-", "1"}], ",", 
                    "2"}], "]"}]}], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
                 TagBox[")", "AutoParentheses"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:122969425"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p122969425-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p122969425-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:122969425", "Proof|ID:122969425-1"},
 CellID->526144461],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p122969425-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->296102259]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership rdm' \[ScriptCapitalZ])\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(membership rdm' \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["rdm", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Element]", 
           "\[DoubleStruckCapitalZ]"}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["rdm", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Element]", 
           "\[DoubleStruckCapitalZ]"}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:509693672"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p509693672-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p509693672-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:509693672", "Proof|ID:509693672-1"},
 CellID->220546910],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p509693672-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->112818037]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm = 0' \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm = 0' \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["rdm", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             "0"}], "\[DoubleLeftRightArrow]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["rdm", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             "0"}], "\[DoubleLeftRightArrow]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:576030870"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p576030870-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p576030870-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:576030870", "Proof|ID:576030870-1"},
 CellID->586909032],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p576030870-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->222576414]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm isReducible' \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm isReducible' \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["rdm", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["rdm", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:825630046"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p825630046-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p825630046-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:825630046", "Proof|ID:825630046-1"},
 CellID->36300196],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p825630046-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->319066684]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->472132368],

Cell[CellGroupData[{

Cell["Lemma: Properties of rdm", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.658039050460758*^9, 3.658039054251565*^9}},
 CellID->75812443],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Integers.nb", 47390677}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65803906278478*^9, 3.6580390703195934`*^9}},
 CellTags->{
  "ID:47390677", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->47390677],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalZ]"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
    RowBox[{"x", ",", "y"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership rdm \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:362000401", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658039072456797*^9, 3.6580390867464223`*^9}},
 CellTags->{
  "ID:362000401", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership rdm \[ScriptCapitalZ]"},
 CellID->362000401],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
   RowBox[{"x", ",", "y", ",", 
    RowBox[{
     UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"x", ",", "y"}], "]"}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm isReducible \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:98296722", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658039098898844*^9, 3.6580391205516815`*^9}},
 CellTags->{
  "ID:98296722", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm isReducible \[ScriptCapitalZ]"},
 CellID->98296722],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"x", ",", "y"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["rdm", "\[ScriptCapitalZ]"], "[", 
     RowBox[{"x", ",", "y"}], "]"}], "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalZ]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm = 0 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:357219590", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658039124623289*^9, 3.658039145340125*^9}},
 CellTags->{
  "ID:357219590", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm = 0 \[ScriptCapitalZ]"},
 CellID->357219590],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->134173173]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->130463965],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership rdm \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership rdm \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
           RowBox[{
             UnderscriptBox["rdm", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
           RowBox[{
             UnderscriptBox["rdm", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:362000401"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p362000401-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p362000401-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:362000401", "Proof|ID:362000401-1"},
 CellID->71841439],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p362000401-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->11485083]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm isReducible \[ScriptCapitalZ])\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(rdm isReducible \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["rdm", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["rdm", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:98296722"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p98296722-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p98296722-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:98296722", "Proof|ID:98296722-1"},
 CellID->414103899],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p98296722-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6580436254452133`*^9, 3.6580436272704163`*^9}},
 CellID->375149138]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm = 0 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm = 0 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["rdm", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox["0", 
              
              RowBox[{"ReductionIntegers", "[", 
                "]"}]]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["rdm", 
                RowBox[{"ReductionIntegers", "[", "]"}]], "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox["0", 
              
              RowBox[{"ReductionIntegers", "[", 
                "]"}]]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:357219590"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p357219590-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p357219590-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:357219590", "Proof|ID:357219590-1"},
 CellID->55429374],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p357219590-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->342077788]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->4706945],

Cell[CellGroupData[{

Cell["Lemma: \[ScriptCapitalZ] Satisfies R11 and R12", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6579469653822684`*^9, 3.657946970935878*^9}, {
  3.6579487133701468`*^9, 3.657948716692952*^9}},
 CellID->55456765],

Cell[BoxData[
 RowBox[{"R11", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R11 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1086665202", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028560064308*^9, 3.594028578457564*^9}, 
   3.6572569431623*^9},
 CellTags->{
  "ID:1086665202", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R11 \[ScriptCapitalZ]"},
 CellID->1086665202],

Cell[BoxData[
 RowBox[{"R12", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R12 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:963670239", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594028562065757*^9, 3.59402857941581*^9}, 
   3.6572569444415026`*^9},
 CellTags->{
  "ID:963670239", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R12 \[ScriptCapitalZ]"},
 CellID->963670239],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->275608555]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->202410225],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R11 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R11 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R11", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R11", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1086665202"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1086665202-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1086665202-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1086665202", "Proof|ID:1086665202-1"},
 CellID->726182844],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1086665202-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->66930221]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R12 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R12 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R12", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R12", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:963670239"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p963670239-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p963670239-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:963670239", "Proof|ID:963670239-1"},
 CellID->70242044],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p963670239-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->166946871]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R13 (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6579469119053745`*^9, 3.6579469147601795`*^9}, 
   3.657948706287734*^9, 3.657949525566574*^9, {3.6579612069199147`*^9, 
   3.6579612119119234`*^9}},
 CellID->702407332],

Cell["", "OpenEnvironment",
 CellID->23716938],

Cell[CellGroupData[{

Cell["Lemma: \[ScriptCapitalZ] Satisfies R13", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.657946942465828*^9, 3.657946959157857*^9}, 
   3.6579487202809587`*^9},
 CellID->117869787],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "\[ScriptCapitalZ]", "]"}]], 
    RowBox[{"[", 
     RowBox[{"A", ",", "B"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", 
     UnderscriptBox["\[SquareSubset]", "\[ScriptCapitalZ]"], "B"}], 
    "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[Exists]", 
        RowBox[{"b", "\[Element]", "B"}]], 
       RowBox[{"b", "\[NotEqual]", 
        UnderscriptBox["0", "\[ScriptCapitalZ]"]}]], 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[ForAll]", 
         RowBox[{"a", "\[Element]", "A"}]], 
        RowBox[{"a", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalZ]"]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]"], "b"}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[SquareSubset] \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:483352095", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657948748501408*^9, 3.6579488663908153`*^9}, {
  3.6579553152309103`*^9, 3.6579553375545497`*^9}},
 CellTags->{
  "ID:483352095", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[SquareSubset] \[ScriptCapitalZ]"},
 CellID->483352095],

Cell[BoxData[
 RowBox[{"R13", "[", "\[ScriptCapitalZ]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R13 \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:466277585", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6572544126118503`*^9, 3.6572544149362545`*^9}, {
  3.6572544747623596`*^9, 3.6572544759947615`*^9}},
 CellTags->{
  "ID:466277585", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R13 \[ScriptCapitalZ]"},
 CellID->466277585],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->160356015]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->138398448],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[SquareSubset] \[ScriptCapitalZ])\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(\[SquareSubset] \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  RowBox[{"ReductionIntegers", "[", "]"}], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  RowBox[{"ReductionIntegers", "[", "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["A", "ExpressionVariable"], 
             UnderscriptBox["\[SquareSubset]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             StyleBox["B", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
              RowBox[{
                StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 RowBox[{"ReductionIntegers", "[", "]"}]]}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   RowBox[{"ReductionIntegers", "[", "]"}]]}]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 StyleBox[
                 "b", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  RowBox[{"ReductionIntegers", "[", "]"}], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  RowBox[{"ReductionIntegers", "[", "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["A", "ExpressionVariable"], 
             UnderscriptBox["\[SquareSubset]", 
              RowBox[{"ReductionIntegers", "[", "]"}]], 
             StyleBox["B", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
              RowBox[{
                StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 RowBox[{"ReductionIntegers", "[", "]"}]]}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   RowBox[{"ReductionIntegers", "[", "]"}]]}]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  RowBox[{"ReductionIntegers", "[", "]"}]], 
                 StyleBox[
                 "b", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:483352095"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p483352095-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p483352095-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:483352095", "Proof|ID:483352095-1"},
 CellID->579238228],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p483352095-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->25986173]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R13 \[ScriptCapitalZ])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R13 \[ScriptCapitalZ])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"R13", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"R13", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:466277585"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p466277585-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p466277585-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:466277585", "Proof|ID:466277585-1"},
 CellID->54970627],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p466277585-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->293687751]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Result (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6579469193153877`*^9, 3.6579469215149913`*^9}, {
  3.658044259028726*^9, 3.6580442633343334`*^9}},
 CellID->283529052],

Cell["", "OpenEnvironment",
 CellID->16240552],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[ScriptCapitalZ] is Algorithmic Reduction Ring\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6572569986515975`*^9, 3.6572570080428143`*^9}},
 CellID->134430315],

Cell[BoxData[
 RowBox[{"isAlgoReductionRing", "[", "\[ScriptCapitalZ]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isAlgoReductionRing \[ScriptCapitalZ])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:25825837", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6572570106168184`*^9, 3.6572570161704283`*^9}},
 CellTags->{
  "ID:25825837", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isAlgoReductionRing \[ScriptCapitalZ]"},
 CellID->25825837],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->712858629]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->53324427],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isAlgoReductionRing \[ScriptCapitalZ])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isAlgoReductionRing \[ScriptCapitalZ])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isAlgoReductionRing", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isAlgoReductionRing", "[", 
         RowBox[{"ReductionIntegers", "[", "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:25825837"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p25825837-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p25825837-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:25825837", "Proof|ID:25825837-1"},
 CellID->84901906],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p25825837-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->749931683]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
},
WindowSize->{1106, 762},
WindowMargins->{{0, Automatic}, {0, Automatic}},
CounterAssignments->{{"TheoremaFormulaCounter" -> 1}},
Magnification->1.,
FrontEndVersion->"9.0 for Linux x86 (64-bit) (February 7, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Theorema"}, "TheoremaCore.nb"]]], 
   
   Cell["Modify the definitions below to change the default appearance of all \
cells in a given style. \nMake modifications to any definition using commands \
in the Format menu.", "Text"], 
   Cell[
    CellGroupData[{
      Cell["Notebook Options Settings", "Section"], 
      Cell[
      "The options defined for the style below will be used at the Notebook \
level.", "Text"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Notebook"], InputAliases -> {"and" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[And]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "or" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", "\[Or]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "not" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Not]", "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "impl" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Implies]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[DoubleLongLeftRightArrow]", 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                "\[SelectionPlaceholder]", "\[Equal]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox[
                  
                  RowBox[{
                   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                  "a:=b"], "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", ":=", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "andn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[And]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "orn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Or]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "equivn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[DoubleLeftRightArrow]", "\[Piecewise]", 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "cdist" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Piecewise]", 
                 
                 GridBox[{{
                   "\[SelectionPlaceholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}, {
                   "\[Placeholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "far" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exr" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "farc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exrc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "let" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["let", 
                  RowBox[{"\[Placeholder]", "=", "\[Placeholder]"}]], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "app" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293a", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "prep" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293b", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "join" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22c8", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "elemT" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22ff", Identity, SyntaxForm -> "a+b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "subop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "supop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SuperscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "subsupop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubsuperscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "oop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "uoop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderoverscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "osop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox[
                OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
                "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "op" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               TagBox["\[SelectionPlaceholder]", Identity], 
               TagBox[")", "AutoParentheses"]}], "[]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e114", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e115", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "()" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:fd3e", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:fd3f", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e8", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e9", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e117", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e118", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "[[]]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e6", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e7", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(())" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2e28", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2e29", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<<>>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27ea", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27eb", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{{}}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2983", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2984", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(" -> 
            TagBox["(", "AutoParentheses"], ")" -> 
            TagBox[")", "AutoParentheses"]}, Background -> 
          RGBColor[0.957, 0.96, 0.97]], 
         Cell[
          StyleData["Notebook", "Printout"], Background -> None]}, Open]]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell["Style Environment Names", "Section"], 
      Cell[
       StyleData[All, "Working"], DockedCells -> {
         Cell[
         "Theorema 2.0", "SBO", FontTracking -> 8, 
          CellMargins -> {{0, 0}, {0, 0}}, 
          CellFrameMargins -> {{0, 50}, {0, 0}}, FontSize -> 28, FontSlant -> 
          "Plain", FontColor -> RGBColor[0.29, 0.46, 0.6], TextAlignment -> 
          Right, Background -> RGBColor[0.36, 0.54, 0.69]]}], 
      Cell[
       StyleData[
       All, "Layout", StyleDefinitions -> StyleData[All, "Working"]], 
       MenuSortingValue -> 1200]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Styles for Title and Section Cells", "Section"], 
      Cell[
       CellGroupData[{
         Cell["Title", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], CellMargins -> {{27, 27}, {10, 30}}, 
             TextAlignment -> Center, MenuSortingValue -> 1100, 
             MenuCommandKey -> "1", FontFamily -> "Bitstream Charter", 
             FontSize -> 36, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Title", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{6, 10}, {5, 30}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], CellMargins -> {{27, 27}, {20, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1150, FontFamily -> 
             "Bitstream Charter", FontSize -> 24, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{6, 10}, {5, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], CellMargins -> {{27, 27}, {8, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], CellMargins -> {{27, 27}, {18, 20}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Sectioning", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], ShowGroupOpener -> True, MenuSortingValue -> 
             1250, MenuCommandKey -> "4", FontFamily -> "Bitstream Charter", 
             FontSize -> 20, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Section", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{2, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {8, 12}}, 
             MenuSortingValue -> 1300, MenuCommandKey -> "5", FontFamily -> 
             "Bitstream Charter", FontSize -> 14, FontWeight -> "Bold", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Subsection", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{9, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {2, 10}}, 
             MenuSortingValue -> 1350, MenuCommandKey -> "6", FontFamily -> 
             "Bitstream Charter", FontSize -> 12, FontWeight -> "Bold", 
             FontSlant -> "Italic", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsection", "Printout"], ShowGroupOpener -> False,
              CellMargins -> {{35, 0}, {7, 14}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsubsection"], CellDingbat -> None, 
             ShowGroupOpener -> True, 
             CellMargins -> {{27, Inherited}, {Inherited, Inherited}}, 
             FontFamily -> "Bitstream Charter", FontSize -> 11, FontWeight -> 
             "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsubsection", "Printout"], 
             CellMargins -> {{50, 0}, {7, 14}}]}, Closed]]}, Open]]}, Open]], 
   
   Cell[
    CellGroupData[{
      Cell["Styles for Publication\[Hyphen]ready Documents", "Section"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Author"], CellMargins -> {{27, 30}, {20, 5}}, 
          MenuSortingValue -> 1900, FontFamily -> "Helvetica", FontSize -> 14,
           FontSlant -> "Italic"], 
         Cell[
          StyleData["Author", "Printout"], CellMargins -> {{6, 10}, {0, 10}}, 
          FontSize -> 10]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Abstract"], CellFrame -> {{0, 0}, {0.5, 0.5}}, 
          CellMargins -> {{27, 140}, {20, 20}}, LineSpacing -> {0.9, 0}, 
          MenuSortingValue -> 2050, FontFamily -> "Times", FontSize -> 12], 
         Cell[
          StyleData["Abstract", "Printout"], Hyphenation -> True, 
          LineSpacing -> {1, 2}, FontSize -> 10]}, Closed]]}, Closed]], 
   Cell[
    CellGroupData[{
      Cell["Text", "Section"], 
      Cell[
       StyleData["Text"], CellMargins -> {{27, 27}, {Inherited, Inherited}}, 
       TextJustification -> 1., ParagraphIndent -> 15, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["EnvironmentText"], ShowCellBracket -> Automatic, 
       CellMargins -> {{82, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextJustification -> 1., ParagraphSpacing -> {0.3, 0.}, 
       MenuSortingValue -> 1480, MenuCommandKey -> "T", FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Item"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "Item", 
          CounterAssignments -> {{"Subitem", 0}, {"Subsubitem", 0}}, 
          MenuSortingValue -> 1600, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Item", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemParagraph"], CellMargins -> {{81, 10}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          DefaultNewCellStyle -> "Item", MenuSortingValue -> 1600, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemParagraph", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "Subitem", 
          CounterAssignments -> {{"Subsubitem", 0}}, MenuSortingValue -> 1610,
           FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subitem", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemParagraph"], CellMargins -> {{105, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subitem", 
          MenuSortingValue -> 1610, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemParagraph", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subsubitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1620, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subsubitem", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemParagraph"], 
          CellMargins -> {{129, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1625, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemParagraph", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "ItemNumbered", 
          CounterAssignments -> {{"SubitemNumbered", 0}, {
            "SubsubitemNumbered", 0}}, MenuSortingValue -> 1630, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemNumbered", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "SubitemNumbered", 
          CounterAssignments -> {{"SubsubitemNumbered", 0}}, MenuSortingValue -> 
          1640, FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemNumbered", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], ".", 
              CounterBox["SubsubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> 
          "SubitemNumbered", CounterIncrements -> "SubsubitemNumbered", 
          MenuSortingValue -> 1645, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemNumbered", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Theorema", "Section"], 
      Cell[
       StyleData["GlobalDeclaration"], CellFrame -> {{0, 2}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 3}, {2, 2}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       InitializationCell -> True, CellFrameMargins -> 10, CellFrameColor -> 
       RGBColor[0.73, 0.65, 0.61], CellFrameLabelMargins -> {{33, 7}, {1, 1}},
        ShowAutoStyles -> False, MenuSortingValue -> 1441, MenuCommandKey -> 
       "G", FontSize -> 14], 
      Cell[
       StyleData["OpenEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {-4, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       StyleData["EnvironmentHeader"], 
       CellFrame -> {{False, False}, {False, False}}, 
       CellMargins -> {{76, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 6, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       ShowAutoStyles -> False, MenuSortingValue -> 1450, FontSize -> 14, 
       FontWeight -> "Bold", FontVariations -> {"CapsType" -> "SmallCaps"}, 
       Background -> RGBColor[0.89, 0.8, 0.69]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FormalTextInputFormula"], ShowCellBracket -> Automatic, 
          CellMargins -> {{96, 10}, {2, 2}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
          InitializationCell -> True, CellFrameLabelMargins -> 53, 
          ShowAutoStyles -> False, LimitsPositioningTokens -> {}, 
          MenuSortingValue -> 1470, MenuCommandKey -> "F", FontWeight -> 
          "Bold", FontColor -> RGBColor[0.25, 0.19, 0.2]], 
         Cell[
          StyleData["FormalTextInputFormula", "Layout"], Editable -> False], 
         Cell[
          StyleData["FormalTextInputFormula", "Presentation"], Editable -> 
          False]}, Open]], 
      Cell[
       StyleData["EndEnvironmentMarker"], 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextAlignment -> Right, MenuSortingValue -> 1480], 
      Cell[
       StyleData["CloseEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {False, True}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {Inherited, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Computation"], CellFrame -> {{1.5, 0}, {0, 0}}, 
          ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 7}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
          CellFrameMargins -> {{8, 8}, {3, 4}}, CellFrameColor -> 
          RGBColor[0.486, 0.33, 0.32], ShowAutoStyles -> False, 
          MenuSortingValue -> 1510, FontWeight -> "Bold", FontColor -> 
          RGBColor[0.23, 0.25, 0.34], Background -> 
          RGBColor[0.88, 0.88, 0.92]], 
         Cell[
          StyleData["Computation", "Layout"], Editable -> False], 
         Cell[
          StyleData["Computation", "Presentation"], Editable -> False]}, 
        Open]], 
      Cell[
       StyleData["ComputationResult"], CellFrame -> {{1.5, 0}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
       CellFrameMargins -> {{8, 8}, {4, 3}}, CellFrameColor -> 
       RGBColor[0.486, 0.33, 0.32], StyleMenuListing -> None, FontColor -> 
       RGBColor[0.23, 0.25, 0.34], Background -> RGBColor[0.88, 0.88, 0.92]], 
      
      Cell[
       StyleData["ComputationInfo"], ShowCellBracket -> False, 
       CellMargins -> {{120, 10}, {10, 2}}, StyleMenuListing -> None, 
       FontFamily -> $CellContext`Arial], 
      Cell[
       StyleData["ComputationInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{144, 10}, {10, 2}}, StyleMenuListing -> None], 
      Cell[
       StyleData["ProofInfo"], ShowCellBracket -> False, 
       CellMargins -> {{76, 10}, {6, 6}}, CellFrameLabelMargins -> 15, 
       StyleMenuListing -> None, FontFamily -> "Arial"], 
      Cell[
       StyleData["ProofInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{100, 10}, {0, 0}}, StyleMenuListing -> None], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FrameLabel"], LanguageCategory -> None, StyleMenuListing -> 
          None, FontFamily -> "Times New Roman", FontSize -> 10, FontWeight -> 
          "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
          FontVariations -> {
           "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
            "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
          RGBColor[0.55, 0.44, 0.42], Background -> None], 
         Cell[
          StyleData["FrameLabel", "Printout"]]}, Open]], 
      Cell[
       StyleData["OpenArchive"], CellFrame -> {{False, False}, {False, True}},
        ShowCellBracket -> Automatic, CellMargins -> {{70, 70}, {0, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 20}, CellFrameMargins -> 3, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], CellFrameLabelMargins -> 1, 
       DefaultFormatType -> DefaultInputFormatType, ShowAutoStyles -> False, 
       MenuSortingValue -> 1210, FontWeight -> Bold, Background -> 
       RGBColor[0.73, 0.74, 0.84]], 
      Cell[
       StyleData["ArchiveInfo"], ShowCellBracket -> Automatic, 
       CellMargins -> {{70, 70}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 5, CellFrameLabelMargins -> 3, ShowCellTags -> 
       False, ShowAutoStyles -> False, MenuSortingValue -> 1220, FontWeight -> 
       "Bold", Background -> RGBColor[0.88, 0.88, 0.92]], 
      Cell[
       StyleData["CloseArchive"], 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> 
       Automatic, CellMargins -> {{0, 0}, {5, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 25}, CellFrameMargins -> 2, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], TextAlignment -> Center, 
       MenuSortingValue -> 1230, FontFamily -> "Arial Black", FontSize -> 24, 
       FontWeight -> "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       Background -> RGBColor[0.73, 0.65, 0.61]], 
      Cell[
       StyleData["IncludeArchive"], CellFrame -> {{4, 0}, {False, False}}, 
       ShowCellBracket -> Automatic, 
       CellMargins -> {{81, 100}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameColor -> RGBColor[0.55, 0.44, 0.42], 
       CellFrameLabels -> {{"\[RightSkeleton]", None}, {None, None}}, 
       CellFrameLabelMargins -> 25, MenuSortingValue -> 1445], 
      Cell[
       StyleData["ExpressionVariable"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionABF"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionMeta"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["AutoParentheses"], Editable -> False, SpanMaxSize -> 3, 
          FontOpacity -> 0.15], 
         Cell[
          StyleData["AutoParentheses", "Layout"], FontOpacity -> 0], 
         Cell[
          StyleData["AutoParentheses", "SlideShow"], FontOpacity -> 0]}, 
        Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Emphasized"], MenuCommandKey -> "E", FontSlant -> 
          "Italic"], 
         Cell[
          StyleData["Emphasized", "SlideShow"], FontSlant -> "Plain", 
          FontColor -> RGBColor[0.36, 0.54, 0.69]]}, Open]], 
      Cell[
       StyleData["GenericButton"], FontFamily -> "Bitstream Charter", 
       FontWeight -> "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
       FontVariations -> {
        "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
         "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
       RGBColor[0.23, 0.25, 0.34]], 
      Cell[
       StyleData["TheoremaForm"], "TwoByteSyntaxCharacterAutoReplacement" -> 
       True, LineSpacing -> {1.25, 0, 2}, SingleLetterItalics -> True, 
       StyleMenuListing -> None, FontFamily -> "Courier"], 
      Cell[
       StyleData["ComponentActive"], FontFamily -> "Times New Roman"], 
      Cell[
       StyleData["ComponentInactive"], FontFamily -> "Times New Roman", 
       FontOpacity -> 0.4], 
      Cell[
       StyleData["ComponentEmpty"], FontFamily -> "Times New Roman", 
       FontVariations -> {"StrikeThrough" -> True}], 
      Cell[
       StyleData["TabView"], FontFamily -> "Times New Roman", FontSize -> 10], 
      Cell[
       StyleData["TabViewLabel"], FontFamily -> "Times New Roman", FontSize -> 
       10, FontWeight -> "Bold"], 
      Cell[
       StyleData["FormReference", StyleDefinitions -> StyleData["Text"]], 
       FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> "Plain",
        FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["FormReferenceHover", StyleDefinitions -> StyleData["Text"]],
        FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> 
       "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> True}, 
       FontColor -> RGBColor[0.404, 0.43, 0.545]], 
      Cell[
       StyleData["DisplayFormula"], CellMargins -> {{36, 6}, {7, 7}}, 
       ConversionRules :> {
        "HTML" -> {{"", ""}, {"<p class=\"Equation\">", "</p>"}}}, 
       CellHorizontalScrolling -> True, DefaultFormatType -> TraditionalForm, 
       InputAutoReplacements -> {}, TextAlignment -> Center, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       ButtonBoxOptions -> {Appearance -> {Automatic, None}}]}, Open]]}, 
  Visible -> False, FrontEndVersion -> 
  "9.0 for Linux x86 (64-bit) (February 7, 2013)", StyleDefinitions -> 
  "StylesheetFormatting.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "ID:1959508349"->{
  Cell[22887, 612, 1225, 37, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1959508349", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even EvenQ"},
   CellID->1959508349]},
 "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Integers.nb"->{\

  Cell[22887, 612, 1225, 37, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1959508349", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even EvenQ"},
   CellID->1959508349],
  Cell[25647, 695, 919, 22, 32, "GlobalDeclaration",
   CellTags->{
    "ID:1696678769", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->1696678769],
  Cell[26569, 719, 969, 26, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:227367582", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "quo Quotient"},
   CellID->227367582],
  Cell[27541, 747, 973, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:74202101", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even mod"},
   CellID->74202101],
  Cell[29328, 809, 747, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:226250890", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even 0", "IR"},
   CellID->226250890],
  Cell[30078, 831, 770, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:175457915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even 1", "IR"},
   CellID->175457915],
  Cell[30851, 854, 694, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:52392318", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->52392318],
  Cell[31548, 874, 770, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:782819098", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even twice"},
   CellID->782819098],
  Cell[32321, 897, 968, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:759529887", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even +1"},
   CellID->759529887],
  Cell[33292, 928, 970, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:243153755", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even -1"},
   CellID->243153755],
  Cell[34265, 959, 887, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:188962611", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even abs"},
   CellID->188962611],
  Cell[35155, 985, 1103, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1148058", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even quo twice"},
   CellID->1148058],
  Cell[36261, 1020, 1292, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:441075310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "odd quo twice"},
   CellID->441075310],
  Cell[37556, 1064, 1075, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:256130067", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even quo -1"},
   CellID->256130067],
  Cell[38634, 1099, 1167, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739691344", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "odd quo -1"},
   CellID->739691344],
  Cell[39804, 1139, 694, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:10064687", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->10064687],
  Cell[40501, 1159, 1261, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208882588", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice \[GreaterEqual]"},
   CellID->208882588],
  Cell[41765, 1198, 1058, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47095122", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice >"},
   CellID->47095122],
  Cell[84434, 2508, 784, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1818128647", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->1818128647],
  Cell[85221, 2529, 1008, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:489394336", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->489394336],
  Cell[86232, 2554, 1323, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1246011018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Element]"},
   CellID->1246011018],
  Cell[87558, 2595, 836, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:234727221", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] 0"},
   CellID->234727221],
  Cell[88397, 2619, 837, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:849552024", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] 1"},
   CellID->849552024],
  Cell[89237, 2643, 887, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1301356286", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] +"},
   CellID->1301356286],
  Cell[90127, 2669, 883, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:196003840", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] -"},
   CellID->196003840],
  Cell[91013, 2695, 898, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1933917000", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] - unary"},
   CellID->1933917000],
  Cell[91914, 2722, 884, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:963583550", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[RawStar]"},
   CellID->963583550],
  Cell[92801, 2748, 1735, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1947278342", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Succeeds]"},
   CellID->1947278342],
  Cell[94539, 2802, 987, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:90094145", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] multN"},
   CellID->90094145],
  Cell[95529, 2831, 1154, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1432577745", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult"},
   CellID->1432577745],
  Cell[96686, 2867, 1387, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1000779593", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult 1"},
   CellID->1000779593],
  Cell[98076, 2909, 1631, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:948580964", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lr"},
   CellID->948580964],
  Cell[99710, 2955, 2942, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1258513368", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] rdm"},
   CellID->1258513368],
  Cell[103125, 3046, 1982, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1342406859", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd"},
   CellID->1342406859],
  Cell[105110, 3099, 1357, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:456350999", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd unary"},
   CellID->456350999],
  Cell[106470, 3138, 1330, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1330653407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] cpm"},
   CellID->1330653407],
  Cell[108602, 3207, 861, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2131968713", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalZ]"},
   CellID->2131968713],
  Cell[141922, 4313, 711, 18, 33, "GlobalDeclaration",
   CellTags->{
    "ID:335127278", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->335127278],
  Cell[143719, 4368, 912, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2044589078", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommRing1 \[ScriptCapitalZ]"},
   CellID->2044589078],
  Cell[144634, 4393, 1617, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:48028479", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "integral domain \[ScriptCapitalZ]"},
   CellID->48028479],
  Cell[154327, 4689, 1210, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1813588854", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isPartIrreflOrder \[ScriptCapitalZ]"},
   CellID->1813588854],
  Cell[155540, 4724, 1045, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:187039377", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isTotalIrreflOrder \[ScriptCapitalZ]"},
   CellID->187039377],
  Cell[156588, 4757, 1067, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:159454607", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isNoetherian \[ScriptCapitalZ]"},
   CellID->159454607],
  Cell[166552, 5074, 1423, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:38031932", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "mult alternative \[ScriptCapitalZ]"},
   CellID->38031932],
  Cell[167978, 5119, 842, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1890039364", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "Multipliers \[ScriptCapitalZ]"},
   CellID->1890039364],
  Cell[168823, 5143, 810, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1152753107", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R0 \[ScriptCapitalZ]"},
   CellID->1152753107],
  Cell[169636, 5166, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:307366433", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R1 \[ScriptCapitalZ]"},
   CellID->307366433],
  Cell[170446, 5189, 831, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:963452595", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R2 \[ScriptCapitalZ]"},
   CellID->963452595],
  Cell[171280, 5212, 808, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1084156204", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R3 \[ScriptCapitalZ]"},
   CellID->1084156204],
  Cell[172091, 5235, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2041570274", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R4 \[ScriptCapitalZ]"},
   CellID->2041570274],
  Cell[191942, 5863, 820, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:260068164", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->260068164],
  Cell[192765, 5886, 882, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:211178897", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership sgn \[ScriptCapitalZ]"},
   CellID->211178897],
  Cell[193650, 5911, 1887, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:118405539", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[RawStar] sgn \[ScriptCapitalZ]"},
   CellID->118405539],
  Cell[195540, 5973, 942, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:88914628", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "0 minimal \[ScriptCapitalZ]"},
   CellID->88914628],
  Cell[196485, 6000, 1345, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:69480817", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice \[Succeeds] twice \[ScriptCapitalZ]"},
   CellID->69480817],
  Cell[197833, 6042, 1883, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:24170142", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[DoubleRightArrow] \[ScriptCapitalZ]"},
   CellID->24170142],
  Cell[199719, 6102, 1885, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:302704588", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[DoubleLeftArrow] \[ScriptCapitalZ]"},
   CellID->302704588],
  Cell[201607, 6162, 1892, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:131195953", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[ScriptCapitalZ]"},
   CellID->131195953],
  Cell[203502, 6221, 1723, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:129207038", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] - \[ScriptCapitalZ]"},
   CellID->129207038],
  Cell[205228, 6275, 1822, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:59300728", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] alternative \[ScriptCapitalZ]"},
   CellID->59300728],
  Cell[207053, 6330, 1722, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:698610132", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] +' \[ScriptCapitalZ]"},
   CellID->698610132],
  Cell[267006, 7947, 767, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:310863676", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->310863676],
  Cell[267776, 7969, 1525, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:471044386", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible - \[ScriptCapitalZ]"},
   CellID->471044386],
  Cell[269304, 8014, 2346, 71, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:379223163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible alternative \[ScriptCapitalZ]"},
   CellID->379223163],
  Cell[271653, 8087, 1625, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:176657795", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible sgn \[ScriptCapitalZ]"},
   CellID->176657795],
  Cell[273281, 8139, 1988, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:217289551", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", 
     "isReducible alternative' \[DoubleRightArrow] \[ScriptCapitalZ]"},
   CellID->217289551],
  Cell[275272, 8200, 1988, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:497414030", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", 
     "isReducible alternative' \[DoubleLeftArrow] \[ScriptCapitalZ]"},
   CellID->497414030],
  Cell[277263, 8261, 2055, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:464408938", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible alternative' \[ScriptCapitalZ]"},
   CellID->464408938],
  Cell[279321, 8322, 1138, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:304230885", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible -' \[ScriptCapitalZ]"},
   CellID->304230885],
  Cell[280462, 8354, 1465, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:97311931", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[RightArrow] - \[ScriptCapitalZ]"},
   CellID->97311931],
  Cell[281930, 8400, 1820, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:99245919", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds] \[ScriptCapitalZ]"},
   CellID->99245919],
  Cell[283753, 8455, 1359, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:162605324", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds]' \[ScriptCapitalZ]"},
   CellID->162605324],
  Cell[285115, 8494, 1338, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40987937", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds]'' \[ScriptCapitalZ]"},
   CellID->40987937],
  Cell[353427, 10308, 807, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:11694492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->11694492],
  Cell[354237, 10330, 949, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:224215037", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership lr \[ScriptCapitalZ]"},
   CellID->224215037],
  Cell[355189, 10358, 1170, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:78177024", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr = 0 \[ScriptCapitalZ]"},
   CellID->78177024],
  Cell[356362, 10396, 1427, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:634190310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr > 0 \[ScriptCapitalZ]"},
   CellID->634190310],
  Cell[357792, 10441, 1186, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:588999973", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "0 > lr \[ScriptCapitalZ]"},
   CellID->588999973],
  Cell[358981, 10479, 1190, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:57477705", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr twice even \[ScriptCapitalZ]"},
   CellID->57477705],
  Cell[360174, 10515, 1393, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:206701743", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr twice odd \[ScriptCapitalZ]"},
   CellID->206701743],
  Cell[361570, 10561, 1143, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:34042360", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isReducible \[ScriptCapitalZ]"},
   CellID->34042360],
  Cell[362716, 10593, 1514, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:692837602", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr minimal isReducible \[ScriptCapitalZ]"},
   CellID->692837602],
  Cell[364233, 10641, 1715, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:44304045", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isReducible sgn \[ScriptCapitalZ]"},
   CellID->44304045],
  Cell[404898, 11810, 892, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:16238117", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->16238117],
  Cell[405793, 11835, 1455, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:468309069", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Element] multN \[ScriptCapitalZ]"},
   CellID->468309069],
  Cell[407251, 11883, 798, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:560244434", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->560244434],
  Cell[408052, 11906, 1129, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142903111", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "areIrrelative alternative \[ScriptCapitalZ]"},
   CellID->142903111],
  Cell[409184, 11940, 1671, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:36948313", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommonReducible alternative \[ScriptCapitalZ]"},
   CellID->36948313],
  Cell[410858, 11990, 2281, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:46630982", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommonReducible alternative' \[ScriptCapitalZ]"},
   CellID->46630982],
  Cell[441624, 12786, 1196, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:69669781", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isMNTCR unary \[ScriptCapitalZ]"},
   CellID->69669781],
  Cell[442823, 12822, 993, 27, 70, "GlobalDeclaration",
   CellTags->{
    "ID:65735690", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->65735690],
  Cell[443819, 12851, 774, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:431845789", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->431845789],
  Cell[444596, 12874, 1242, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:96649542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isNTCR \[ScriptCapitalZ]"},
   CellID->96649542],
  Cell[445841, 12909, 1400, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:132861224", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isMNTCR \[ScriptCapitalZ]"},
   CellID->132861224],
  Cell[468145, 13489, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:595906550", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R5 \[ScriptCapitalZ]"},
   CellID->595906550],
  Cell[468953, 13512, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:384714041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R6 \[ScriptCapitalZ]"},
   CellID->384714041],
  Cell[469763, 13535, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739712574", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R7 \[ScriptCapitalZ]"},
   CellID->739712574],
  Cell[470573, 13558, 809, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:782958792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R8 \[ScriptCapitalZ]"},
   CellID->782958792],
  Cell[481642, 13914, 848, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:67773804", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReductionRing \[ScriptCapitalZ]"},
   CellID->67773804],
  Cell[485946, 14054, 809, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:156995908", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R9 \[ScriptCapitalZ]"},
   CellID->156995908],
  Cell[486758, 14077, 813, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1051911108", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R10 \[ScriptCapitalZ]"},
   CellID->1051911108],
  Cell[493209, 14285, 695, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:160940549", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->160940549],
  Cell[493907, 14305, 794, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:46650045", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->46650045],
  Cell[494704, 14327, 1058, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:816204582", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos \[GreaterEqual] 0 \[ScriptCapitalZ]"},
   CellID->816204582],
  Cell[495765, 14359, 1107, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:171042210", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg \[GreaterEqual] 0 \[ScriptCapitalZ]"},
   CellID->171042210],
  Cell[496875, 14393, 1496, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160702195", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos 0 \[ScriptCapitalZ]"},
   CellID->160702195],
  Cell[498374, 14442, 1673, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:478744619", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg 0 \[ScriptCapitalZ]"},
   CellID->478744619],
  Cell[500050, 14496, 1435, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:680200521", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos isReducible \[ScriptCapitalZ]"},
   CellID->680200521],
  Cell[501488, 14542, 1606, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:122969425", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg isReducible \[ScriptCapitalZ]"},
   CellID->122969425],
  Cell[503097, 14594, 914, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:509693672", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership rdm' \[ScriptCapitalZ]"},
   CellID->509693672],
  Cell[504014, 14620, 1400, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:576030870", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm = 0' \[ScriptCapitalZ]"},
   CellID->576030870],
  Cell[505417, 14662, 1124, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:825630046", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm isReducible' \[ScriptCapitalZ]"},
   CellID->825630046],
  Cell[550814, 15949, 753, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:47390677", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->47390677],
  Cell[551570, 15971, 950, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:362000401", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership rdm \[ScriptCapitalZ]"},
   CellID->362000401],
  Cell[552523, 15998, 1110, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:98296722", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm isReducible \[ScriptCapitalZ]"},
   CellID->98296722],
  Cell[553636, 16029, 1271, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:357219590", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm = 0 \[ScriptCapitalZ]"},
   CellID->357219590],
  Cell[569267, 16475, 811, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1086665202", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R11 \[ScriptCapitalZ]"},
   CellID->1086665202],
  Cell[570081, 16498, 811, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:963670239", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R12 \[ScriptCapitalZ]"},
   CellID->963670239],
  Cell[576461, 16705, 1970, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:483352095", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[SquareSubset] \[ScriptCapitalZ]"},
   CellID->483352095],
  Cell[578434, 16767, 841, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:466277585", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R13 \[ScriptCapitalZ]"},
   CellID->466277585],
  Cell[588775, 17065, 845, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:25825837", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isAlgoReductionRing \[ScriptCapitalZ]"},
   CellID->25825837]},
 "even EvenQ"->{
  Cell[22887, 612, 1225, 37, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1959508349", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even EvenQ"},
   CellID->1959508349]},
 "ID:1696678769"->{
  Cell[25647, 695, 919, 22, 32, "GlobalDeclaration",
   CellTags->{
    "ID:1696678769", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->1696678769]},
 "ID:227367582"->{
  Cell[26569, 719, 969, 26, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:227367582", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "quo Quotient"},
   CellID->227367582]},
 "quo Quotient"->{
  Cell[26569, 719, 969, 26, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:227367582", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "quo Quotient"},
   CellID->227367582]},
 "ID:74202101"->{
  Cell[27541, 747, 973, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:74202101", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even mod"},
   CellID->74202101]},
 "even mod"->{
  Cell[27541, 747, 973, 29, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:74202101", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even mod"},
   CellID->74202101]},
 "ID:226250890"->{
  Cell[29328, 809, 747, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:226250890", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even 0", "IR"},
   CellID->226250890]},
 "even 0"->{
  Cell[29328, 809, 747, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:226250890", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even 0", "IR"},
   CellID->226250890]},
 "IR"->{
  Cell[29328, 809, 747, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:226250890", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even 0", "IR"},
   CellID->226250890],
  Cell[30078, 831, 770, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:175457915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even 1", "IR"},
   CellID->175457915]},
 "ID:175457915"->{
  Cell[30078, 831, 770, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:175457915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even 1", "IR"},
   CellID->175457915]},
 "even 1"->{
  Cell[30078, 831, 770, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:175457915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even 1", "IR"},
   CellID->175457915]},
 "ID:52392318"->{
  Cell[30851, 854, 694, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:52392318", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->52392318]},
 "ID:782819098"->{
  Cell[31548, 874, 770, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:782819098", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even twice"},
   CellID->782819098]},
 "even twice"->{
  Cell[31548, 874, 770, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:782819098", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even twice"},
   CellID->782819098]},
 "ID:759529887"->{
  Cell[32321, 897, 968, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:759529887", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even +1"},
   CellID->759529887]},
 "even +1"->{
  Cell[32321, 897, 968, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:759529887", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even +1"},
   CellID->759529887]},
 "ID:243153755"->{
  Cell[33292, 928, 970, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:243153755", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even -1"},
   CellID->243153755]},
 "even -1"->{
  Cell[33292, 928, 970, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:243153755", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even -1"},
   CellID->243153755]},
 "ID:188962611"->{
  Cell[34265, 959, 887, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:188962611", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even abs"},
   CellID->188962611]},
 "even abs"->{
  Cell[34265, 959, 887, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:188962611", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even abs"},
   CellID->188962611]},
 "ID:1148058"->{
  Cell[35155, 985, 1103, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1148058", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even quo twice"},
   CellID->1148058]},
 "even quo twice"->{
  Cell[35155, 985, 1103, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1148058", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even quo twice"},
   CellID->1148058]},
 "ID:441075310"->{
  Cell[36261, 1020, 1292, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:441075310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "odd quo twice"},
   CellID->441075310]},
 "odd quo twice"->{
  Cell[36261, 1020, 1292, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:441075310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "odd quo twice"},
   CellID->441075310]},
 "ID:256130067"->{
  Cell[37556, 1064, 1075, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:256130067", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even quo -1"},
   CellID->256130067]},
 "even quo -1"->{
  Cell[37556, 1064, 1075, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:256130067", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even quo -1"},
   CellID->256130067]},
 "ID:739691344"->{
  Cell[38634, 1099, 1167, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739691344", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "odd quo -1"},
   CellID->739691344]},
 "odd quo -1"->{
  Cell[38634, 1099, 1167, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739691344", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "odd quo -1"},
   CellID->739691344]},
 "ID:10064687"->{
  Cell[39804, 1139, 694, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:10064687", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->10064687]},
 "ID:208882588"->{
  Cell[40501, 1159, 1261, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208882588", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice \[GreaterEqual]"},
   CellID->208882588]},
 "twice \[GreaterEqual]"->{
  Cell[40501, 1159, 1261, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208882588", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice \[GreaterEqual]"},
   CellID->208882588]},
 "ID:47095122"->{
  Cell[41765, 1198, 1058, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47095122", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice >"},
   CellID->47095122]},
 "twice >"->{
  Cell[41765, 1198, 1058, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47095122", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice >"},
   CellID->47095122]},
 "Proof|ID:226250890"->{
  Cell[42974, 1244, 1693, 54, 70, "ProofInfo",
   CellTags->{"Proof|ID:226250890", "Proof|ID:226250890-1"},
   CellID->75108773]},
 "Proof|ID:226250890-1"->{
  Cell[42974, 1244, 1693, 54, 70, "ProofInfo",
   CellTags->{"Proof|ID:226250890", "Proof|ID:226250890-1"},
   CellID->75108773]},
 "Proof|ID:175457915"->{
  Cell[45187, 1317, 1973, 62, 70, "ProofInfo",
   CellTags->{"Proof|ID:175457915", "Proof|ID:175457915-1"},
   CellID->36435349]},
 "Proof|ID:175457915-1"->{
  Cell[45187, 1317, 1973, 62, 70, "ProofInfo",
   CellTags->{"Proof|ID:175457915", "Proof|ID:175457915-1"},
   CellID->36435349]},
 "Proof|ID:782819098"->{
  Cell[47679, 1398, 2385, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:782819098", "Proof|ID:782819098-1"},
   CellID->387405268]},
 "Proof|ID:782819098-1"->{
  Cell[47679, 1398, 2385, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:782819098", "Proof|ID:782819098-1"},
   CellID->387405268]},
 "Proof|ID:759529887"->{
  Cell[50584, 1489, 2948, 88, 70, "ProofInfo",
   CellTags->{"Proof|ID:759529887", "Proof|ID:759529887-1"},
   CellID->198480204]},
 "Proof|ID:759529887-1"->{
  Cell[50584, 1489, 2948, 88, 70, "ProofInfo",
   CellTags->{"Proof|ID:759529887", "Proof|ID:759529887-1"},
   CellID->198480204]},
 "Proof|ID:243153755"->{
  Cell[54052, 1596, 2947, 88, 70, "ProofInfo",
   CellTags->{"Proof|ID:243153755", "Proof|ID:243153755-1"},
   CellID->86433169]},
 "Proof|ID:243153755-1"->{
  Cell[54052, 1596, 2947, 88, 70, "ProofInfo",
   CellTags->{"Proof|ID:243153755", "Proof|ID:243153755-1"},
   CellID->86433169]},
 "Proof|ID:188962611"->{
  Cell[57519, 1703, 2687, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:188962611", "Proof|ID:188962611-1"},
   CellID->608316900]},
 "Proof|ID:188962611-1"->{
  Cell[57519, 1703, 2687, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:188962611", "Proof|ID:188962611-1"},
   CellID->608316900]},
 "Proof|ID:1148058"->{
  Cell[60726, 1800, 3156, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:1148058", "Proof|ID:1148058-1"},
   CellID->314369717]},
 "Proof|ID:1148058-1"->{
  Cell[60726, 1800, 3156, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:1148058", "Proof|ID:1148058-1"},
   CellID->314369717]},
 "Proof|ID:441075310"->{
  Cell[64398, 1913, 3584, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:441075310", "Proof|ID:441075310-1"},
   CellID->900195017]},
 "Proof|ID:441075310-1"->{
  Cell[64398, 1913, 3584, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:441075310", "Proof|ID:441075310-1"},
   CellID->900195017]},
 "Proof|ID:256130067"->{
  Cell[68502, 2036, 3103, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:256130067", "Proof|ID:256130067-1"},
   CellID->748444859]},
 "Proof|ID:256130067-1"->{
  Cell[68502, 2036, 3103, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:256130067", "Proof|ID:256130067-1"},
   CellID->748444859]},
 "Proof|ID:739691344"->{
  Cell[72125, 2145, 3341, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:739691344", "Proof|ID:739691344-1"},
   CellID->138102035]},
 "Proof|ID:739691344-1"->{
  Cell[72125, 2145, 3341, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:739691344", "Proof|ID:739691344-1"},
   CellID->138102035]},
 "Proof|ID:208882588"->{
  Cell[75986, 2260, 3323, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:208882588", "Proof|ID:208882588-1"},
   CellID->148478097]},
 "Proof|ID:208882588-1"->{
  Cell[75986, 2260, 3323, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:208882588", "Proof|ID:208882588-1"},
   CellID->148478097]},
 "Proof|ID:47095122"->{
  Cell[79898, 2372, 3234, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:47095122", "Proof|ID:47095122-1"},
   CellID->275585894]},
 "Proof|ID:47095122-1"->{
  Cell[79898, 2372, 3234, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:47095122", "Proof|ID:47095122-1"},
   CellID->275585894]},
 "ID:1818128647"->{
  Cell[84434, 2508, 784, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1818128647", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->1818128647]},
 "ID:489394336"->{
  Cell[85221, 2529, 1008, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:489394336", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->489394336]},
 "ID:1246011018"->{
  Cell[86232, 2554, 1323, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1246011018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Element]"},
   CellID->1246011018]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Element]"->{
  Cell[86232, 2554, 1323, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1246011018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Element]"},
   CellID->1246011018]},
 "ID:234727221"->{
  Cell[87558, 2595, 836, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:234727221", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] 0"},
   CellID->234727221]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] 0"->{
  Cell[87558, 2595, 836, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:234727221", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] 0"},
   CellID->234727221]},
 "ID:849552024"->{
  Cell[88397, 2619, 837, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:849552024", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] 1"},
   CellID->849552024]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] 1"->{
  Cell[88397, 2619, 837, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:849552024", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] 1"},
   CellID->849552024]},
 "ID:1301356286"->{
  Cell[89237, 2643, 887, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1301356286", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] +"},
   CellID->1301356286]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] +"->{
  Cell[89237, 2643, 887, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1301356286", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] +"},
   CellID->1301356286]},
 "ID:196003840"->{
  Cell[90127, 2669, 883, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:196003840", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] -"},
   CellID->196003840]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] -"->{
  Cell[90127, 2669, 883, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:196003840", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] -"},
   CellID->196003840]},
 "ID:1933917000"->{
  Cell[91013, 2695, 898, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1933917000", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] - unary"},
   CellID->1933917000]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] - unary"->{
  Cell[91013, 2695, 898, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1933917000", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] - unary"},
   CellID->1933917000]},
 "ID:963583550"->{
  Cell[91914, 2722, 884, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:963583550", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[RawStar]"},
   CellID->963583550]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[RawStar]"->{
  Cell[91914, 2722, 884, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:963583550", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[RawStar]"},
   CellID->963583550]},
 "ID:1947278342"->{
  Cell[92801, 2748, 1735, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1947278342", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Succeeds]"},
   CellID->1947278342]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Succeeds]"->{
  Cell[92801, 2748, 1735, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1947278342", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Succeeds]"},
   CellID->1947278342]},
 "ID:90094145"->{
  Cell[94539, 2802, 987, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:90094145", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] multN"},
   CellID->90094145]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] multN"->{
  Cell[94539, 2802, 987, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:90094145", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] multN"},
   CellID->90094145]},
 "ID:1432577745"->{
  Cell[95529, 2831, 1154, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1432577745", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult"},
   CellID->1432577745]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult"->{
  Cell[95529, 2831, 1154, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1432577745", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult"},
   CellID->1432577745]},
 "ID:1000779593"->{
  Cell[96686, 2867, 1387, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1000779593", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult 1"},
   CellID->1000779593]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult 1"->{
  Cell[96686, 2867, 1387, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1000779593", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult 1"},
   CellID->1000779593]},
 "ID:948580964"->{
  Cell[98076, 2909, 1631, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:948580964", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lr"},
   CellID->948580964]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lr"->{
  Cell[98076, 2909, 1631, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:948580964", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lr"},
   CellID->948580964]},
 "ID:1258513368"->{
  Cell[99710, 2955, 2942, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1258513368", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] rdm"},
   CellID->1258513368]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] rdm"->{
  Cell[99710, 2955, 2942, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1258513368", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] rdm"},
   CellID->1258513368]},
 "ID:1342406859"->{
  Cell[103125, 3046, 1982, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1342406859", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd"},
   CellID->1342406859]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd"->{
  Cell[103125, 3046, 1982, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1342406859", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd"},
   CellID->1342406859]},
 "ID:456350999"->{
  Cell[105110, 3099, 1357, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:456350999", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd unary"},
   CellID->456350999]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd unary"->{
  Cell[105110, 3099, 1357, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:456350999", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd unary"},
   CellID->456350999]},
 "ID:1330653407"->{
  Cell[106470, 3138, 1330, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1330653407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] cpm"},
   CellID->1330653407]},
 "\[ScriptCapitalR]\[DoubleStruckCapitalZ] cpm"->{
  Cell[106470, 3138, 1330, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1330653407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] cpm"},
   CellID->1330653407]},
 "ID:2131968713"->{
  Cell[108602, 3207, 861, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2131968713", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalZ]"},
   CellID->2131968713]},
 "\[ScriptCapitalZ]"->{
  Cell[108602, 3207, 861, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2131968713", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalZ]"},
   CellID->2131968713]},
 "ID:335127278"->{
  Cell[141922, 4313, 711, 18, 33, "GlobalDeclaration",
   CellTags->{
    "ID:335127278", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->335127278]},
 "ID:2044589078"->{
  Cell[143719, 4368, 912, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2044589078", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommRing1 \[ScriptCapitalZ]"},
   CellID->2044589078]},
 "isCommRing1 \[ScriptCapitalZ]"->{
  Cell[143719, 4368, 912, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2044589078", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommRing1 \[ScriptCapitalZ]"},
   CellID->2044589078]},
 "ID:48028479"->{
  Cell[144634, 4393, 1617, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:48028479", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "integral domain \[ScriptCapitalZ]"},
   CellID->48028479]},
 "integral domain \[ScriptCapitalZ]"->{
  Cell[144634, 4393, 1617, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:48028479", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "integral domain \[ScriptCapitalZ]"},
   CellID->48028479]},
 "Proof|ID:2044589078"->{
  Cell[146409, 4454, 1851, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:2044589078", "Proof|ID:2044589078-1"},
   CellID->47164179]},
 "Proof|ID:2044589078-1"->{
  Cell[146409, 4454, 1851, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:2044589078", "Proof|ID:2044589078-1"},
   CellID->47164179]},
 "Proof|ID:48028479"->{
  Cell[148780, 4530, 4325, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:48028479", "Proof|ID:48028479-1"},
   CellID->584664]},
 "Proof|ID:48028479-1"->{
  Cell[148780, 4530, 4325, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:48028479", "Proof|ID:48028479-1"},
   CellID->584664]},
 "ID:1813588854"->{
  Cell[154327, 4689, 1210, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1813588854", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isPartIrreflOrder \[ScriptCapitalZ]"},
   CellID->1813588854]},
 "isPartIrreflOrder \[ScriptCapitalZ]"->{
  Cell[154327, 4689, 1210, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1813588854", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isPartIrreflOrder \[ScriptCapitalZ]"},
   CellID->1813588854]},
 "ID:187039377"->{
  Cell[155540, 4724, 1045, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:187039377", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isTotalIrreflOrder \[ScriptCapitalZ]"},
   CellID->187039377]},
 "isTotalIrreflOrder \[ScriptCapitalZ]"->{
  Cell[155540, 4724, 1045, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:187039377", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isTotalIrreflOrder \[ScriptCapitalZ]"},
   CellID->187039377]},
 "ID:159454607"->{
  Cell[156588, 4757, 1067, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:159454607", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isNoetherian \[ScriptCapitalZ]"},
   CellID->159454607]},
 "isNoetherian \[ScriptCapitalZ]"->{
  Cell[156588, 4757, 1067, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:159454607", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isNoetherian \[ScriptCapitalZ]"},
   CellID->159454607]},
 "Proof|ID:1813588854"->{
  Cell[157810, 4799, 2139, 65, 70, "ProofInfo",
   CellTags->{"Proof|ID:1813588854", "Proof|ID:1813588854-1"},
   CellID->80460269]},
 "Proof|ID:1813588854-1"->{
  Cell[157810, 4799, 2139, 65, 70, "ProofInfo",
   CellTags->{"Proof|ID:1813588854", "Proof|ID:1813588854-1"},
   CellID->80460269]},
 "Proof|ID:187039377"->{
  Cell[160469, 4883, 2135, 64, 70, "ProofInfo",
   CellTags->{"Proof|ID:187039377", "Proof|ID:187039377-1"},
   CellID->106498233]},
 "Proof|ID:187039377-1"->{
  Cell[160469, 4883, 2135, 64, 70, "ProofInfo",
   CellTags->{"Proof|ID:187039377", "Proof|ID:187039377-1"},
   CellID->106498233]},
 "Proof|ID:159454607"->{
  Cell[163124, 4966, 2103, 63, 70, "ProofInfo",
   CellTags->{"Proof|ID:159454607", "Proof|ID:159454607-1"},
   CellID->187116588]},
 "Proof|ID:159454607-1"->{
  Cell[163124, 4966, 2103, 63, 70, "ProofInfo",
   CellTags->{"Proof|ID:159454607", "Proof|ID:159454607-1"},
   CellID->187116588]},
 "ID:38031932"->{
  Cell[166552, 5074, 1423, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:38031932", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "mult alternative \[ScriptCapitalZ]"},
   CellID->38031932]},
 "mult alternative \[ScriptCapitalZ]"->{
  Cell[166552, 5074, 1423, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:38031932", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "mult alternative \[ScriptCapitalZ]"},
   CellID->38031932]},
 "ID:1890039364"->{
  Cell[167978, 5119, 842, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1890039364", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "Multipliers \[ScriptCapitalZ]"},
   CellID->1890039364]},
 "Multipliers \[ScriptCapitalZ]"->{
  Cell[167978, 5119, 842, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1890039364", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "Multipliers \[ScriptCapitalZ]"},
   CellID->1890039364]},
 "ID:1152753107"->{
  Cell[168823, 5143, 810, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1152753107", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R0 \[ScriptCapitalZ]"},
   CellID->1152753107]},
 "R0 \[ScriptCapitalZ]"->{
  Cell[168823, 5143, 810, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1152753107", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R0 \[ScriptCapitalZ]"},
   CellID->1152753107]},
 "ID:307366433"->{
  Cell[169636, 5166, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:307366433", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R1 \[ScriptCapitalZ]"},
   CellID->307366433]},
 "R1 \[ScriptCapitalZ]"->{
  Cell[169636, 5166, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:307366433", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R1 \[ScriptCapitalZ]"},
   CellID->307366433]},
 "ID:963452595"->{
  Cell[170446, 5189, 831, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:963452595", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R2 \[ScriptCapitalZ]"},
   CellID->963452595]},
 "R2 \[ScriptCapitalZ]"->{
  Cell[170446, 5189, 831, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:963452595", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R2 \[ScriptCapitalZ]"},
   CellID->963452595]},
 "ID:1084156204"->{
  Cell[171280, 5212, 808, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1084156204", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R3 \[ScriptCapitalZ]"},
   CellID->1084156204]},
 "R3 \[ScriptCapitalZ]"->{
  Cell[171280, 5212, 808, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1084156204", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R3 \[ScriptCapitalZ]"},
   CellID->1084156204]},
 "ID:2041570274"->{
  Cell[172091, 5235, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2041570274", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R4 \[ScriptCapitalZ]"},
   CellID->2041570274]},
 "R4 \[ScriptCapitalZ]"->{
  Cell[172091, 5235, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2041570274", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R4 \[ScriptCapitalZ]"},
   CellID->2041570274]},
 "Proof|ID:38031932"->{
  Cell[173053, 5267, 3640, 101, 70, "ProofInfo",
   CellTags->{"Proof|ID:38031932", "Proof|ID:38031932-1"},
   CellID->279953348]},
 "Proof|ID:38031932-1"->{
  Cell[173053, 5267, 3640, 101, 70, "ProofInfo",
   CellTags->{"Proof|ID:38031932", "Proof|ID:38031932-1"},
   CellID->279953348]},
 "Proof|ID:1890039364"->{
  Cell[177211, 5387, 1852, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1890039364", "Proof|ID:1890039364-1"},
   CellID->845457494]},
 "Proof|ID:1890039364-1"->{
  Cell[177211, 5387, 1852, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1890039364", "Proof|ID:1890039364-1"},
   CellID->845457494]},
 "Proof|ID:1152753107"->{
  Cell[179584, 5463, 1816, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1152753107", "Proof|ID:1152753107-1"},
   CellID->229850786]},
 "Proof|ID:1152753107-1"->{
  Cell[179584, 5463, 1816, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1152753107", "Proof|ID:1152753107-1"},
   CellID->229850786]},
 "Proof|ID:307366433"->{
  Cell[181920, 5539, 1806, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:307366433", "Proof|ID:307366433-1"},
   CellID->93056147]},
 "Proof|ID:307366433-1"->{
  Cell[181920, 5539, 1806, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:307366433", "Proof|ID:307366433-1"},
   CellID->93056147]},
 "Proof|ID:963452595"->{
  Cell[184245, 5614, 1805, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:963452595", "Proof|ID:963452595-1"},
   CellID->8129192]},
 "Proof|ID:963452595-1"->{
  Cell[184245, 5614, 1805, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:963452595", "Proof|ID:963452595-1"},
   CellID->8129192]},
 "Proof|ID:1084156204"->{
  Cell[186570, 5689, 1816, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1084156204", "Proof|ID:1084156204-1"},
   CellID->477880951]},
 "Proof|ID:1084156204-1"->{
  Cell[186570, 5689, 1816, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1084156204", "Proof|ID:1084156204-1"},
   CellID->477880951]},
 "Proof|ID:2041570274"->{
  Cell[188907, 5765, 1816, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:2041570274", "Proof|ID:2041570274-1"},
   CellID->299709042]},
 "Proof|ID:2041570274-1"->{
  Cell[188907, 5765, 1816, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:2041570274", "Proof|ID:2041570274-1"},
   CellID->299709042]},
 "ID:260068164"->{
  Cell[191942, 5863, 820, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:260068164", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->260068164]},
 "ID:211178897"->{
  Cell[192765, 5886, 882, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:211178897", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership sgn \[ScriptCapitalZ]"},
   CellID->211178897]},
 "membership sgn \[ScriptCapitalZ]"->{
  Cell[192765, 5886, 882, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:211178897", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership sgn \[ScriptCapitalZ]"},
   CellID->211178897]},
 "ID:118405539"->{
  Cell[193650, 5911, 1887, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:118405539", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[RawStar] sgn \[ScriptCapitalZ]"},
   CellID->118405539]},
 "\[RawStar] sgn \[ScriptCapitalZ]"->{
  Cell[193650, 5911, 1887, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:118405539", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[RawStar] sgn \[ScriptCapitalZ]"},
   CellID->118405539]},
 "ID:88914628"->{
  Cell[195540, 5973, 942, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:88914628", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "0 minimal \[ScriptCapitalZ]"},
   CellID->88914628]},
 "0 minimal \[ScriptCapitalZ]"->{
  Cell[195540, 5973, 942, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:88914628", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "0 minimal \[ScriptCapitalZ]"},
   CellID->88914628]},
 "ID:69480817"->{
  Cell[196485, 6000, 1345, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:69480817", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice \[Succeeds] twice \[ScriptCapitalZ]"},
   CellID->69480817]},
 "twice \[Succeeds] twice \[ScriptCapitalZ]"->{
  Cell[196485, 6000, 1345, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:69480817", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice \[Succeeds] twice \[ScriptCapitalZ]"},
   CellID->69480817]},
 "ID:24170142"->{
  Cell[197833, 6042, 1883, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:24170142", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[DoubleRightArrow] \[ScriptCapitalZ]"},
   CellID->24170142]},
 "\[Succeeds] + \[DoubleRightArrow] \[ScriptCapitalZ]"->{
  Cell[197833, 6042, 1883, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:24170142", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[DoubleRightArrow] \[ScriptCapitalZ]"},
   CellID->24170142]},
 "ID:302704588"->{
  Cell[199719, 6102, 1885, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:302704588", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[DoubleLeftArrow] \[ScriptCapitalZ]"},
   CellID->302704588]},
 "\[Succeeds] + \[DoubleLeftArrow] \[ScriptCapitalZ]"->{
  Cell[199719, 6102, 1885, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:302704588", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[DoubleLeftArrow] \[ScriptCapitalZ]"},
   CellID->302704588]},
 "ID:131195953"->{
  Cell[201607, 6162, 1892, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:131195953", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[ScriptCapitalZ]"},
   CellID->131195953]},
 "\[Succeeds] + \[ScriptCapitalZ]"->{
  Cell[201607, 6162, 1892, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:131195953", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[ScriptCapitalZ]"},
   CellID->131195953]},
 "ID:129207038"->{
  Cell[203502, 6221, 1723, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:129207038", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] - \[ScriptCapitalZ]"},
   CellID->129207038]},
 "\[Succeeds] - \[ScriptCapitalZ]"->{
  Cell[203502, 6221, 1723, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:129207038", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] - \[ScriptCapitalZ]"},
   CellID->129207038]},
 "ID:59300728"->{
  Cell[205228, 6275, 1822, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:59300728", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] alternative \[ScriptCapitalZ]"},
   CellID->59300728]},
 "\[Succeeds] alternative \[ScriptCapitalZ]"->{
  Cell[205228, 6275, 1822, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:59300728", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] alternative \[ScriptCapitalZ]"},
   CellID->59300728]},
 "ID:698610132"->{
  Cell[207053, 6330, 1722, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:698610132", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] +' \[ScriptCapitalZ]"},
   CellID->698610132]},
 "\[Succeeds] +' \[ScriptCapitalZ]"->{
  Cell[207053, 6330, 1722, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:698610132", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] +' \[ScriptCapitalZ]"},
   CellID->698610132]},
 "Proof|ID:211178897"->{
  Cell[208930, 6388, 2649, 77, 70, "ProofInfo",
   CellTags->{"Proof|ID:211178897", "Proof|ID:211178897-1"},
   CellID->273631166]},
 "Proof|ID:211178897-1"->{
  Cell[208930, 6388, 2649, 77, 70, "ProofInfo",
   CellTags->{"Proof|ID:211178897", "Proof|ID:211178897-1"},
   CellID->273631166]},
 "Proof|ID:118405539"->{
  Cell[212099, 6484, 5656, 142, 70, "ProofInfo",
   CellTags->{"Proof|ID:118405539", "Proof|ID:118405539-1"},
   CellID->55870784]},
 "Proof|ID:118405539-1"->{
  Cell[212099, 6484, 5656, 142, 70, "ProofInfo",
   CellTags->{"Proof|ID:118405539", "Proof|ID:118405539-1"},
   CellID->55870784]},
 "Proof|ID:88914628"->{
  Cell[218274, 6645, 3039, 87, 70, "ProofInfo",
   CellTags->{"Proof|ID:88914628", "Proof|ID:88914628-1"},
   CellID->298080950]},
 "Proof|ID:88914628-1"->{
  Cell[218274, 6645, 3039, 87, 70, "ProofInfo",
   CellTags->{"Proof|ID:88914628", "Proof|ID:88914628-1"},
   CellID->298080950]},
 "Proof|ID:69480817"->{
  Cell[221832, 6751, 4303, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:69480817", "Proof|ID:69480817-1"},
   CellID->8097116]},
 "Proof|ID:69480817-1"->{
  Cell[221832, 6751, 4303, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:69480817", "Proof|ID:69480817-1"},
   CellID->8097116]},
 "Proof|ID:24170142"->{
  Cell[226654, 6886, 6292, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:24170142", "Proof|ID:24170142-1"},
   CellID->297064343]},
 "Proof|ID:24170142-1"->{
  Cell[226654, 6886, 6292, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:24170142", "Proof|ID:24170142-1"},
   CellID->297064343]},
 "Proof|ID:302704588"->{
  Cell[233465, 7065, 6308, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:302704588", "Proof|ID:302704588-1"},
   CellID->172202783]},
 "Proof|ID:302704588-1"->{
  Cell[233465, 7065, 6308, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:302704588", "Proof|ID:302704588-1"},
   CellID->172202783]},
 "Proof|ID:131195953"->{
  Cell[240292, 7244, 6514, 165, 70, "ProofInfo",
   CellTags->{"Proof|ID:131195953", "Proof|ID:131195953-1"},
   CellID->88300092]},
 "Proof|ID:131195953-1"->{
  Cell[240292, 7244, 6514, 165, 70, "ProofInfo",
   CellTags->{"Proof|ID:131195953", "Proof|ID:131195953-1"},
   CellID->88300092]},
 "Proof|ID:129207038"->{
  Cell[247326, 7428, 5698, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:129207038", "Proof|ID:129207038-1"},
   CellID->33212788]},
 "Proof|ID:129207038-1"->{
  Cell[247326, 7428, 5698, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:129207038", "Proof|ID:129207038-1"},
   CellID->33212788]},
 "Proof|ID:59300728"->{
  Cell[253544, 7594, 5798, 146, 70, "ProofInfo",
   CellTags->{"Proof|ID:59300728", "Proof|ID:59300728-1"},
   CellID->946864890]},
 "Proof|ID:59300728-1"->{
  Cell[253544, 7594, 5798, 146, 70, "ProofInfo",
   CellTags->{"Proof|ID:59300728", "Proof|ID:59300728-1"},
   CellID->946864890]},
 "Proof|ID:698610132"->{
  Cell[259861, 7759, 6175, 155, 70, "ProofInfo",
   CellTags->{"Proof|ID:698610132", "Proof|ID:698610132-1"},
   CellID->357893767]},
 "Proof|ID:698610132-1"->{
  Cell[259861, 7759, 6175, 155, 70, "ProofInfo",
   CellTags->{"Proof|ID:698610132", "Proof|ID:698610132-1"},
   CellID->357893767]},
 "ID:310863676"->{
  Cell[267006, 7947, 767, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:310863676", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->310863676]},
 "ID:471044386"->{
  Cell[267776, 7969, 1525, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:471044386", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible - \[ScriptCapitalZ]"},
   CellID->471044386]},
 "isReducible - \[ScriptCapitalZ]"->{
  Cell[267776, 7969, 1525, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:471044386", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible - \[ScriptCapitalZ]"},
   CellID->471044386]},
 "ID:379223163"->{
  Cell[269304, 8014, 2346, 71, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:379223163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible alternative \[ScriptCapitalZ]"},
   CellID->379223163]},
 "isReducible alternative \[ScriptCapitalZ]"->{
  Cell[269304, 8014, 2346, 71, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:379223163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible alternative \[ScriptCapitalZ]"},
   CellID->379223163]},
 "ID:176657795"->{
  Cell[271653, 8087, 1625, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:176657795", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible sgn \[ScriptCapitalZ]"},
   CellID->176657795]},
 "isReducible sgn \[ScriptCapitalZ]"->{
  Cell[271653, 8087, 1625, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:176657795", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible sgn \[ScriptCapitalZ]"},
   CellID->176657795]},
 "ID:217289551"->{
  Cell[273281, 8139, 1988, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:217289551", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", 
     "isReducible alternative' \[DoubleRightArrow] \[ScriptCapitalZ]"},
   CellID->217289551]},
 "isReducible alternative' \[DoubleRightArrow] \[ScriptCapitalZ]"->{
  Cell[273281, 8139, 1988, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:217289551", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", 
     "isReducible alternative' \[DoubleRightArrow] \[ScriptCapitalZ]"},
   CellID->217289551]},
 "ID:497414030"->{
  Cell[275272, 8200, 1988, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:497414030", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", 
     "isReducible alternative' \[DoubleLeftArrow] \[ScriptCapitalZ]"},
   CellID->497414030]},
 "isReducible alternative' \[DoubleLeftArrow] \[ScriptCapitalZ]"->{
  Cell[275272, 8200, 1988, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:497414030", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", 
     "isReducible alternative' \[DoubleLeftArrow] \[ScriptCapitalZ]"},
   CellID->497414030]},
 "ID:464408938"->{
  Cell[277263, 8261, 2055, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:464408938", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible alternative' \[ScriptCapitalZ]"},
   CellID->464408938]},
 "isReducible alternative' \[ScriptCapitalZ]"->{
  Cell[277263, 8261, 2055, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:464408938", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible alternative' \[ScriptCapitalZ]"},
   CellID->464408938]},
 "ID:304230885"->{
  Cell[279321, 8322, 1138, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:304230885", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible -' \[ScriptCapitalZ]"},
   CellID->304230885]},
 "isReducible -' \[ScriptCapitalZ]"->{
  Cell[279321, 8322, 1138, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:304230885", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible -' \[ScriptCapitalZ]"},
   CellID->304230885]},
 "ID:97311931"->{
  Cell[280462, 8354, 1465, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:97311931", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[RightArrow] - \[ScriptCapitalZ]"},
   CellID->97311931]},
 "\[RightArrow] - \[ScriptCapitalZ]"->{
  Cell[280462, 8354, 1465, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:97311931", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[RightArrow] - \[ScriptCapitalZ]"},
   CellID->97311931]},
 "ID:99245919"->{
  Cell[281930, 8400, 1820, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:99245919", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds] \[ScriptCapitalZ]"},
   CellID->99245919]},
 "isReducible \[Succeeds] \[ScriptCapitalZ]"->{
  Cell[281930, 8400, 1820, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:99245919", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds] \[ScriptCapitalZ]"},
   CellID->99245919]},
 "ID:162605324"->{
  Cell[283753, 8455, 1359, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:162605324", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds]' \[ScriptCapitalZ]"},
   CellID->162605324]},
 "isReducible \[Succeeds]' \[ScriptCapitalZ]"->{
  Cell[283753, 8455, 1359, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:162605324", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds]' \[ScriptCapitalZ]"},
   CellID->162605324]},
 "ID:40987937"->{
  Cell[285115, 8494, 1338, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40987937", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds]'' \[ScriptCapitalZ]"},
   CellID->40987937]},
 "isReducible \[Succeeds]'' \[ScriptCapitalZ]"->{
  Cell[285115, 8494, 1338, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40987937", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds]'' \[ScriptCapitalZ]"},
   CellID->40987937]},
 "Proof|ID:471044386"->{
  Cell[286608, 8543, 4977, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:471044386", "Proof|ID:471044386-1"},
   CellID->450891268]},
 "Proof|ID:471044386-1"->{
  Cell[286608, 8543, 4977, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:471044386", "Proof|ID:471044386-1"},
   CellID->450891268]},
 "Proof|ID:379223163"->{
  Cell[292105, 8689, 7043, 174, 70, "ProofInfo",
   CellTags->{"Proof|ID:379223163", "Proof|ID:379223163-1"},
   CellID->249907335]},
 "Proof|ID:379223163-1"->{
  Cell[292105, 8689, 7043, 174, 70, "ProofInfo",
   CellTags->{"Proof|ID:379223163", "Proof|ID:379223163-1"},
   CellID->249907335]},
 "Proof|ID:176657795"->{
  Cell[299668, 8882, 5329, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:176657795", "Proof|ID:176657795-1"},
   CellID->334897452]},
 "Proof|ID:176657795-1"->{
  Cell[299668, 8882, 5329, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:176657795", "Proof|ID:176657795-1"},
   CellID->334897452]},
 "Proof|ID:217289551"->{
  Cell[305517, 9035, 6019, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:217289551", "Proof|ID:217289551-1"},
   CellID->880922821]},
 "Proof|ID:217289551-1"->{
  Cell[305517, 9035, 6019, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:217289551", "Proof|ID:217289551-1"},
   CellID->880922821]},
 "Proof|ID:497414030"->{
  Cell[312056, 9208, 6041, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:497414030", "Proof|ID:497414030-1"},
   CellID->110073726]},
 "Proof|ID:497414030-1"->{
  Cell[312056, 9208, 6041, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:497414030", "Proof|ID:497414030-1"},
   CellID->110073726]},
 "Proof|ID:464408938"->{
  Cell[318616, 9381, 6151, 156, 70, "ProofInfo",
   CellTags->{"Proof|ID:464408938", "Proof|ID:464408938-1"},
   CellID->577762604]},
 "Proof|ID:464408938-1"->{
  Cell[318616, 9381, 6151, 156, 70, "ProofInfo",
   CellTags->{"Proof|ID:464408938", "Proof|ID:464408938-1"},
   CellID->577762604]},
 "Proof|ID:304230885"->{
  Cell[325287, 9556, 4079, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:304230885", "Proof|ID:304230885-1"},
   CellID->115875544]},
 "Proof|ID:304230885-1"->{
  Cell[325287, 9556, 4079, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:304230885", "Proof|ID:304230885-1"},
   CellID->115875544]},
 "Proof|ID:97311931"->{
  Cell[329885, 9684, 4842, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:97311931", "Proof|ID:97311931-1"},
   CellID->616242356]},
 "Proof|ID:97311931-1"->{
  Cell[329885, 9684, 4842, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:97311931", "Proof|ID:97311931-1"},
   CellID->616242356]},
 "Proof|ID:99245919"->{
  Cell[335245, 9829, 6270, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:99245919", "Proof|ID:99245919-1"},
   CellID->518473742]},
 "Proof|ID:99245919-1"->{
  Cell[335245, 9829, 6270, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:99245919", "Proof|ID:99245919-1"},
   CellID->518473742]},
 "Proof|ID:162605324"->{
  Cell[342033, 10002, 4719, 122, 70, "ProofInfo",
   CellTags->{"Proof|ID:162605324", "Proof|ID:162605324-1"},
   CellID->291524205]},
 "Proof|ID:162605324-1"->{
  Cell[342033, 10002, 4719, 122, 70, "ProofInfo",
   CellTags->{"Proof|ID:162605324", "Proof|ID:162605324-1"},
   CellID->291524205]},
 "Proof|ID:40987937"->{
  Cell[347272, 10143, 5110, 130, 70, "ProofInfo",
   CellTags->{"Proof|ID:40987937", "Proof|ID:40987937-1"},
   CellID->103343062]},
 "Proof|ID:40987937-1"->{
  Cell[347272, 10143, 5110, 130, 70, "ProofInfo",
   CellTags->{"Proof|ID:40987937", "Proof|ID:40987937-1"},
   CellID->103343062]},
 "ID:11694492"->{
  Cell[353427, 10308, 807, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:11694492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->11694492]},
 "ID:224215037"->{
  Cell[354237, 10330, 949, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:224215037", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership lr \[ScriptCapitalZ]"},
   CellID->224215037]},
 "membership lr \[ScriptCapitalZ]"->{
  Cell[354237, 10330, 949, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:224215037", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership lr \[ScriptCapitalZ]"},
   CellID->224215037]},
 "ID:78177024"->{
  Cell[355189, 10358, 1170, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:78177024", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr = 0 \[ScriptCapitalZ]"},
   CellID->78177024]},
 "lr = 0 \[ScriptCapitalZ]"->{
  Cell[355189, 10358, 1170, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:78177024", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr = 0 \[ScriptCapitalZ]"},
   CellID->78177024]},
 "ID:634190310"->{
  Cell[356362, 10396, 1427, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:634190310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr > 0 \[ScriptCapitalZ]"},
   CellID->634190310]},
 "lr > 0 \[ScriptCapitalZ]"->{
  Cell[356362, 10396, 1427, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:634190310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr > 0 \[ScriptCapitalZ]"},
   CellID->634190310]},
 "ID:588999973"->{
  Cell[357792, 10441, 1186, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:588999973", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "0 > lr \[ScriptCapitalZ]"},
   CellID->588999973]},
 "0 > lr \[ScriptCapitalZ]"->{
  Cell[357792, 10441, 1186, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:588999973", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "0 > lr \[ScriptCapitalZ]"},
   CellID->588999973]},
 "ID:57477705"->{
  Cell[358981, 10479, 1190, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:57477705", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr twice even \[ScriptCapitalZ]"},
   CellID->57477705]},
 "lr twice even \[ScriptCapitalZ]"->{
  Cell[358981, 10479, 1190, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:57477705", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr twice even \[ScriptCapitalZ]"},
   CellID->57477705]},
 "ID:206701743"->{
  Cell[360174, 10515, 1393, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:206701743", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr twice odd \[ScriptCapitalZ]"},
   CellID->206701743]},
 "lr twice odd \[ScriptCapitalZ]"->{
  Cell[360174, 10515, 1393, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:206701743", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr twice odd \[ScriptCapitalZ]"},
   CellID->206701743]},
 "ID:34042360"->{
  Cell[361570, 10561, 1143, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:34042360", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isReducible \[ScriptCapitalZ]"},
   CellID->34042360]},
 "lr isReducible \[ScriptCapitalZ]"->{
  Cell[361570, 10561, 1143, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:34042360", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isReducible \[ScriptCapitalZ]"},
   CellID->34042360]},
 "ID:692837602"->{
  Cell[362716, 10593, 1514, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:692837602", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr minimal isReducible \[ScriptCapitalZ]"},
   CellID->692837602]},
 "lr minimal isReducible \[ScriptCapitalZ]"->{
  Cell[362716, 10593, 1514, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:692837602", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr minimal isReducible \[ScriptCapitalZ]"},
   CellID->692837602]},
 "ID:44304045"->{
  Cell[364233, 10641, 1715, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:44304045", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isReducible sgn \[ScriptCapitalZ]"},
   CellID->44304045]},
 "lr isReducible sgn \[ScriptCapitalZ]"->{
  Cell[364233, 10641, 1715, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:44304045", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isReducible sgn \[ScriptCapitalZ]"},
   CellID->44304045]},
 "Proof|ID:224215037"->{
  Cell[366104, 10702, 2817, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:224215037", "Proof|ID:224215037-1"},
   CellID->519188695]},
 "Proof|ID:224215037-1"->{
  Cell[366104, 10702, 2817, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:224215037", "Proof|ID:224215037-1"},
   CellID->519188695]},
 "Proof|ID:78177024"->{
  Cell[369441, 10802, 2934, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:78177024", "Proof|ID:78177024-1"},
   CellID->336735725]},
 "Proof|ID:78177024-1"->{
  Cell[369441, 10802, 2934, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:78177024", "Proof|ID:78177024-1"},
   CellID->336735725]},
 "Proof|ID:634190310"->{
  Cell[372894, 10905, 3276, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:634190310", "Proof|ID:634190310-1"},
   CellID->77986082]},
 "Proof|ID:634190310-1"->{
  Cell[372894, 10905, 3276, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:634190310", "Proof|ID:634190310-1"},
   CellID->77986082]},
 "Proof|ID:588999973"->{
  Cell[376690, 11016, 3262, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:588999973", "Proof|ID:588999973-1"},
   CellID->1365258]},
 "Proof|ID:588999973-1"->{
  Cell[376690, 11016, 3262, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:588999973", "Proof|ID:588999973-1"},
   CellID->1365258]},
 "Proof|ID:57477705"->{
  Cell[380472, 11129, 3623, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:57477705", "Proof|ID:57477705-1"},
   CellID->253970289]},
 "Proof|ID:57477705-1"->{
  Cell[380472, 11129, 3623, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:57477705", "Proof|ID:57477705-1"},
   CellID->253970289]},
 "Proof|ID:206701743"->{
  Cell[384612, 11247, 4469, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:206701743", "Proof|ID:206701743-1"},
   CellID->43456446]},
 "Proof|ID:206701743-1"->{
  Cell[384612, 11247, 4469, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:206701743", "Proof|ID:206701743-1"},
   CellID->43456446]},
 "Proof|ID:34042360"->{
  Cell[389600, 11387, 3454, 95, 70, "ProofInfo",
   CellTags->{"Proof|ID:34042360", "Proof|ID:34042360-1"},
   CellID->169464375]},
 "Proof|ID:34042360-1"->{
  Cell[389600, 11387, 3454, 95, 70, "ProofInfo",
   CellTags->{"Proof|ID:34042360", "Proof|ID:34042360-1"},
   CellID->169464375]},
 "Proof|ID:692837602"->{
  Cell[393573, 11501, 4318, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:692837602", "Proof|ID:692837602-1"},
   CellID->371670326]},
 "Proof|ID:692837602-1"->{
  Cell[393573, 11501, 4318, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:692837602", "Proof|ID:692837602-1"},
   CellID->371670326]},
 "Proof|ID:44304045"->{
  Cell[398411, 11636, 5404, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:44304045", "Proof|ID:44304045-1"},
   CellID->477556229]},
 "Proof|ID:44304045-1"->{
  Cell[398411, 11636, 5404, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:44304045", "Proof|ID:44304045-1"},
   CellID->477556229]},
 "ID:16238117"->{
  Cell[404898, 11810, 892, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:16238117", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->16238117]},
 "ID:468309069"->{
  Cell[405793, 11835, 1455, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:468309069", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Element] multN \[ScriptCapitalZ]"},
   CellID->468309069]},
 "\[Element] multN \[ScriptCapitalZ]"->{
  Cell[405793, 11835, 1455, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:468309069", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Element] multN \[ScriptCapitalZ]"},
   CellID->468309069]},
 "ID:560244434"->{
  Cell[407251, 11883, 798, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:560244434", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->560244434]},
 "ID:142903111"->{
  Cell[408052, 11906, 1129, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142903111", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "areIrrelative alternative \[ScriptCapitalZ]"},
   CellID->142903111]},
 "areIrrelative alternative \[ScriptCapitalZ]"->{
  Cell[408052, 11906, 1129, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142903111", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "areIrrelative alternative \[ScriptCapitalZ]"},
   CellID->142903111]},
 "ID:36948313"->{
  Cell[409184, 11940, 1671, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:36948313", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommonReducible alternative \[ScriptCapitalZ]"},
   CellID->36948313]},
 "isCommonReducible alternative \[ScriptCapitalZ]"->{
  Cell[409184, 11940, 1671, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:36948313", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommonReducible alternative \[ScriptCapitalZ]"},
   CellID->36948313]},
 "ID:46630982"->{
  Cell[410858, 11990, 2281, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:46630982", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommonReducible alternative' \[ScriptCapitalZ]"},
   CellID->46630982]},
 "isCommonReducible alternative' \[ScriptCapitalZ]"->{
  Cell[410858, 11990, 2281, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:46630982", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommonReducible alternative' \[ScriptCapitalZ]"},
   CellID->46630982]},
 "Proof|ID:468309069"->{
  Cell[413292, 12063, 3517, 95, 70, "ProofInfo",
   CellTags->{"Proof|ID:468309069", "Proof|ID:468309069-1"},
   CellID->223234557]},
 "Proof|ID:468309069-1"->{
  Cell[413292, 12063, 3517, 95, 70, "ProofInfo",
   CellTags->{"Proof|ID:468309069", "Proof|ID:468309069-1"},
   CellID->223234557]},
 "Proof|ID:142903111"->{
  Cell[417327, 12177, 6252, 156, 70, "ProofInfo",
   CellTags->{"Proof|ID:142903111", "Proof|ID:142903111-1"},
   CellID->4630108]},
 "Proof|ID:142903111-1"->{
  Cell[417327, 12177, 6252, 156, 70, "ProofInfo",
   CellTags->{"Proof|ID:142903111", "Proof|ID:142903111-1"},
   CellID->4630108]},
 "Proof|ID:36948313"->{
  Cell[424099, 12352, 7090, 172, 70, "ProofInfo",
   CellTags->{"Proof|ID:36948313", "Proof|ID:36948313-1"},
   CellID->239177438]},
 "Proof|ID:36948313-1"->{
  Cell[424099, 12352, 7090, 172, 70, "ProofInfo",
   CellTags->{"Proof|ID:36948313", "Proof|ID:36948313-1"},
   CellID->239177438]},
 "Proof|ID:46630982"->{
  Cell[431707, 12543, 8948, 210, 70, "ProofInfo",
   CellTags->{"Proof|ID:46630982", "Proof|ID:46630982-1"},
   CellID->250839951]},
 "Proof|ID:46630982-1"->{
  Cell[431707, 12543, 8948, 210, 70, "ProofInfo",
   CellTags->{"Proof|ID:46630982", "Proof|ID:46630982-1"},
   CellID->250839951]},
 "ID:69669781"->{
  Cell[441624, 12786, 1196, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:69669781", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isMNTCR unary \[ScriptCapitalZ]"},
   CellID->69669781]},
 "lr isMNTCR unary \[ScriptCapitalZ]"->{
  Cell[441624, 12786, 1196, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:69669781", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isMNTCR unary \[ScriptCapitalZ]"},
   CellID->69669781]},
 "ID:65735690"->{
  Cell[442823, 12822, 993, 27, 70, "GlobalDeclaration",
   CellTags->{
    "ID:65735690", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->65735690]},
 "ID:431845789"->{
  Cell[443819, 12851, 774, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:431845789", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->431845789]},
 "ID:96649542"->{
  Cell[444596, 12874, 1242, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:96649542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isNTCR \[ScriptCapitalZ]"},
   CellID->96649542]},
 "lr isNTCR \[ScriptCapitalZ]"->{
  Cell[444596, 12874, 1242, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:96649542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isNTCR \[ScriptCapitalZ]"},
   CellID->96649542]},
 "ID:132861224"->{
  Cell[445841, 12909, 1400, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:132861224", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isMNTCR \[ScriptCapitalZ]"},
   CellID->132861224]},
 "lr isMNTCR \[ScriptCapitalZ]"->{
  Cell[445841, 12909, 1400, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:132861224", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isMNTCR \[ScriptCapitalZ]"},
   CellID->132861224]},
 "Proof|ID:69669781"->{
  Cell[447395, 12955, 3449, 95, 70, "ProofInfo",
   CellTags->{"Proof|ID:69669781", "Proof|ID:69669781-1"},
   CellID->22101709]},
 "Proof|ID:69669781-1"->{
  Cell[447395, 12955, 3449, 95, 70, "ProofInfo",
   CellTags->{"Proof|ID:69669781", "Proof|ID:69669781-1"},
   CellID->22101709]},
 "Proof|ID:96649542"->{
  Cell[451363, 13069, 7422, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:96649542", "Proof|ID:96649542-1"},
   CellID->151366490]},
 "Proof|ID:96649542-1"->{
  Cell[451363, 13069, 7422, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:96649542", "Proof|ID:96649542-1"},
   CellID->151366490]},
 "Proof|ID:132861224"->{
  Cell[459304, 13266, 7438, 179, 70, "ProofInfo",
   CellTags->{"Proof|ID:132861224", "Proof|ID:132861224-1"},
   CellID->94818643]},
 "Proof|ID:132861224-1"->{
  Cell[459304, 13266, 7438, 179, 70, "ProofInfo",
   CellTags->{"Proof|ID:132861224", "Proof|ID:132861224-1"},
   CellID->94818643]},
 "ID:595906550"->{
  Cell[468145, 13489, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:595906550", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R5 \[ScriptCapitalZ]"},
   CellID->595906550]},
 "R5 \[ScriptCapitalZ]"->{
  Cell[468145, 13489, 805, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:595906550", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R5 \[ScriptCapitalZ]"},
   CellID->595906550]},
 "ID:384714041"->{
  Cell[468953, 13512, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:384714041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R6 \[ScriptCapitalZ]"},
   CellID->384714041]},
 "R6 \[ScriptCapitalZ]"->{
  Cell[468953, 13512, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:384714041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R6 \[ScriptCapitalZ]"},
   CellID->384714041]},
 "ID:739712574"->{
  Cell[469763, 13535, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739712574", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R7 \[ScriptCapitalZ]"},
   CellID->739712574]},
 "R7 \[ScriptCapitalZ]"->{
  Cell[469763, 13535, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739712574", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R7 \[ScriptCapitalZ]"},
   CellID->739712574]},
 "ID:782958792"->{
  Cell[470573, 13558, 809, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:782958792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R8 \[ScriptCapitalZ]"},
   CellID->782958792]},
 "R8 \[ScriptCapitalZ]"->{
  Cell[470573, 13558, 809, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:782958792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R8 \[ScriptCapitalZ]"},
   CellID->782958792]},
 "Proof|ID:595906550"->{
  Cell[471540, 13590, 1807, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:595906550", "Proof|ID:595906550-1"},
   CellID->801014817]},
 "Proof|ID:595906550-1"->{
  Cell[471540, 13590, 1807, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:595906550", "Proof|ID:595906550-1"},
   CellID->801014817]},
 "Proof|ID:384714041"->{
  Cell[473867, 13665, 1805, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:384714041", "Proof|ID:384714041-1"},
   CellID->4523962]},
 "Proof|ID:384714041-1"->{
  Cell[473867, 13665, 1805, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:384714041", "Proof|ID:384714041-1"},
   CellID->4523962]},
 "Proof|ID:739712574"->{
  Cell[476192, 13740, 1807, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:739712574", "Proof|ID:739712574-1"},
   CellID->802064395]},
 "Proof|ID:739712574-1"->{
  Cell[476192, 13740, 1807, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:739712574", "Proof|ID:739712574-1"},
   CellID->802064395]},
 "Proof|ID:782958792"->{
  Cell[478519, 13815, 1807, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:782958792", "Proof|ID:782958792-1"},
   CellID->148939037]},
 "Proof|ID:782958792-1"->{
  Cell[478519, 13815, 1807, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:782958792", "Proof|ID:782958792-1"},
   CellID->148939037]},
 "ID:67773804"->{
  Cell[481642, 13914, 848, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:67773804", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReductionRing \[ScriptCapitalZ]"},
   CellID->67773804]},
 "isReductionRing \[ScriptCapitalZ]"->{
  Cell[481642, 13914, 848, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:67773804", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReductionRing \[ScriptCapitalZ]"},
   CellID->67773804]},
 "Proof|ID:67773804"->{
  Cell[482647, 13947, 1870, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:67773804", "Proof|ID:67773804-1"},
   CellID->766315013]},
 "Proof|ID:67773804-1"->{
  Cell[482647, 13947, 1870, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:67773804", "Proof|ID:67773804-1"},
   CellID->766315013]},
 "ID:156995908"->{
  Cell[485946, 14054, 809, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:156995908", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R9 \[ScriptCapitalZ]"},
   CellID->156995908]},
 "R9 \[ScriptCapitalZ]"->{
  Cell[485946, 14054, 809, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:156995908", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R9 \[ScriptCapitalZ]"},
   CellID->156995908]},
 "ID:1051911108"->{
  Cell[486758, 14077, 813, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1051911108", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R10 \[ScriptCapitalZ]"},
   CellID->1051911108]},
 "R10 \[ScriptCapitalZ]"->{
  Cell[486758, 14077, 813, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1051911108", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R10 \[ScriptCapitalZ]"},
   CellID->1051911108]},
 "Proof|ID:156995908"->{
  Cell[487727, 14109, 1807, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:156995908", "Proof|ID:156995908-1"},
   CellID->802545094]},
 "Proof|ID:156995908-1"->{
  Cell[487727, 14109, 1807, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:156995908", "Proof|ID:156995908-1"},
   CellID->802545094]},
 "Proof|ID:1051911108"->{
  Cell[490054, 14184, 1819, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1051911108", "Proof|ID:1051911108-1"},
   CellID->67102701]},
 "Proof|ID:1051911108-1"->{
  Cell[490054, 14184, 1819, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1051911108", "Proof|ID:1051911108-1"},
   CellID->67102701]},
 "ID:160940549"->{
  Cell[493209, 14285, 695, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:160940549", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->160940549]},
 "ID:46650045"->{
  Cell[493907, 14305, 794, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:46650045", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->46650045]},
 "ID:816204582"->{
  Cell[494704, 14327, 1058, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:816204582", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos \[GreaterEqual] 0 \[ScriptCapitalZ]"},
   CellID->816204582]},
 "rdm pos \[GreaterEqual] 0 \[ScriptCapitalZ]"->{
  Cell[494704, 14327, 1058, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:816204582", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos \[GreaterEqual] 0 \[ScriptCapitalZ]"},
   CellID->816204582]},
 "ID:171042210"->{
  Cell[495765, 14359, 1107, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:171042210", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg \[GreaterEqual] 0 \[ScriptCapitalZ]"},
   CellID->171042210]},
 "rdm neg \[GreaterEqual] 0 \[ScriptCapitalZ]"->{
  Cell[495765, 14359, 1107, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:171042210", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg \[GreaterEqual] 0 \[ScriptCapitalZ]"},
   CellID->171042210]},
 "ID:160702195"->{
  Cell[496875, 14393, 1496, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160702195", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos 0 \[ScriptCapitalZ]"},
   CellID->160702195]},
 "rdm pos 0 \[ScriptCapitalZ]"->{
  Cell[496875, 14393, 1496, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160702195", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos 0 \[ScriptCapitalZ]"},
   CellID->160702195]},
 "ID:478744619"->{
  Cell[498374, 14442, 1673, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:478744619", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg 0 \[ScriptCapitalZ]"},
   CellID->478744619]},
 "rdm neg 0 \[ScriptCapitalZ]"->{
  Cell[498374, 14442, 1673, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:478744619", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg 0 \[ScriptCapitalZ]"},
   CellID->478744619]},
 "ID:680200521"->{
  Cell[500050, 14496, 1435, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:680200521", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos isReducible \[ScriptCapitalZ]"},
   CellID->680200521]},
 "rdm pos isReducible \[ScriptCapitalZ]"->{
  Cell[500050, 14496, 1435, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:680200521", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos isReducible \[ScriptCapitalZ]"},
   CellID->680200521]},
 "ID:122969425"->{
  Cell[501488, 14542, 1606, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:122969425", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg isReducible \[ScriptCapitalZ]"},
   CellID->122969425]},
 "rdm neg isReducible \[ScriptCapitalZ]"->{
  Cell[501488, 14542, 1606, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:122969425", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg isReducible \[ScriptCapitalZ]"},
   CellID->122969425]},
 "ID:509693672"->{
  Cell[503097, 14594, 914, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:509693672", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership rdm' \[ScriptCapitalZ]"},
   CellID->509693672]},
 "membership rdm' \[ScriptCapitalZ]"->{
  Cell[503097, 14594, 914, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:509693672", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership rdm' \[ScriptCapitalZ]"},
   CellID->509693672]},
 "ID:576030870"->{
  Cell[504014, 14620, 1400, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:576030870", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm = 0' \[ScriptCapitalZ]"},
   CellID->576030870]},
 "rdm = 0' \[ScriptCapitalZ]"->{
  Cell[504014, 14620, 1400, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:576030870", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm = 0' \[ScriptCapitalZ]"},
   CellID->576030870]},
 "ID:825630046"->{
  Cell[505417, 14662, 1124, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:825630046", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm isReducible' \[ScriptCapitalZ]"},
   CellID->825630046]},
 "rdm isReducible' \[ScriptCapitalZ]"->{
  Cell[505417, 14662, 1124, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:825630046", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm isReducible' \[ScriptCapitalZ]"},
   CellID->825630046]},
 "Proof|ID:816204582"->{
  Cell[506696, 14703, 3562, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:816204582", "Proof|ID:816204582-1"},
   CellID->451910702]},
 "Proof|ID:816204582-1"->{
  Cell[506696, 14703, 3562, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:816204582", "Proof|ID:816204582-1"},
   CellID->451910702]},
 "Proof|ID:171042210"->{
  Cell[510778, 14824, 3971, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:171042210", "Proof|ID:171042210-1"},
   CellID->29599316]},
 "Proof|ID:171042210-1"->{
  Cell[510778, 14824, 3971, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:171042210", "Proof|ID:171042210-1"},
   CellID->29599316]},
 "Proof|ID:160702195"->{
  Cell[515269, 14953, 4103, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:160702195", "Proof|ID:160702195-1"},
   CellID->461788959]},
 "Proof|ID:160702195-1"->{
  Cell[515269, 14953, 4103, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:160702195", "Proof|ID:160702195-1"},
   CellID->461788959]},
 "Proof|ID:478744619"->{
  Cell[519957, 15087, 4958, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:478744619", "Proof|ID:478744619-1"},
   CellID->1001675670]},
 "Proof|ID:478744619-1"->{
  Cell[519957, 15087, 4958, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:478744619", "Proof|ID:478744619-1"},
   CellID->1001675670]},
 "Proof|ID:680200521"->{
  Cell[525434, 15238, 4996, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:680200521", "Proof|ID:680200521-1"},
   CellID->98328192]},
 "Proof|ID:680200521-1"->{
  Cell[525434, 15238, 4996, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:680200521", "Proof|ID:680200521-1"},
   CellID->98328192]},
 "Proof|ID:122969425"->{
  Cell[531020, 15392, 6137, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:122969425", "Proof|ID:122969425-1"},
   CellID->526144461]},
 "Proof|ID:122969425-1"->{
  Cell[531020, 15392, 6137, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:122969425", "Proof|ID:122969425-1"},
   CellID->526144461]},
 "Proof|ID:509693672"->{
  Cell[537677, 15571, 3040, 88, 70, "ProofInfo",
   CellTags->{"Proof|ID:509693672", "Proof|ID:509693672-1"},
   CellID->220546910]},
 "Proof|ID:509693672-1"->{
  Cell[537677, 15571, 3040, 88, 70, "ProofInfo",
   CellTags->{"Proof|ID:509693672", "Proof|ID:509693672-1"},
   CellID->220546910]},
 "Proof|ID:576030870"->{
  Cell[541237, 15678, 4008, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:576030870", "Proof|ID:576030870-1"},
   CellID->586909032]},
 "Proof|ID:576030870-1"->{
  Cell[541237, 15678, 4008, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:576030870", "Proof|ID:576030870-1"},
   CellID->586909032]},
 "Proof|ID:825630046"->{
  Cell[545765, 15807, 4100, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:825630046", "Proof|ID:825630046-1"},
   CellID->36300196]},
 "Proof|ID:825630046-1"->{
  Cell[545765, 15807, 4100, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:825630046", "Proof|ID:825630046-1"},
   CellID->36300196]},
 "ID:47390677"->{
  Cell[550814, 15949, 753, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:47390677", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
   CellID->47390677]},
 "ID:362000401"->{
  Cell[551570, 15971, 950, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:362000401", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership rdm \[ScriptCapitalZ]"},
   CellID->362000401]},
 "membership rdm \[ScriptCapitalZ]"->{
  Cell[551570, 15971, 950, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:362000401", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership rdm \[ScriptCapitalZ]"},
   CellID->362000401]},
 "ID:98296722"->{
  Cell[552523, 15998, 1110, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:98296722", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm isReducible \[ScriptCapitalZ]"},
   CellID->98296722]},
 "rdm isReducible \[ScriptCapitalZ]"->{
  Cell[552523, 15998, 1110, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:98296722", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm isReducible \[ScriptCapitalZ]"},
   CellID->98296722]},
 "ID:357219590"->{
  Cell[553636, 16029, 1271, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:357219590", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm = 0 \[ScriptCapitalZ]"},
   CellID->357219590]},
 "rdm = 0 \[ScriptCapitalZ]"->{
  Cell[553636, 16029, 1271, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:357219590", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm = 0 \[ScriptCapitalZ]"},
   CellID->357219590]},
 "Proof|ID:362000401"->{
  Cell[555063, 16079, 3374, 93, 70, "ProofInfo",
   CellTags->{"Proof|ID:362000401", "Proof|ID:362000401-1"},
   CellID->71841439]},
 "Proof|ID:362000401-1"->{
  Cell[555063, 16079, 3374, 93, 70, "ProofInfo",
   CellTags->{"Proof|ID:362000401", "Proof|ID:362000401-1"},
   CellID->71841439]},
 "Proof|ID:98296722"->{
  Cell[558956, 16191, 4330, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:98296722", "Proof|ID:98296722-1"},
   CellID->414103899]},
 "Proof|ID:98296722-1"->{
  Cell[558956, 16191, 4330, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:98296722", "Proof|ID:98296722-1"},
   CellID->414103899]},
 "Proof|ID:357219590"->{
  Cell[563875, 16325, 4370, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:357219590", "Proof|ID:357219590-1"},
   CellID->55429374]},
 "Proof|ID:357219590-1"->{
  Cell[563875, 16325, 4370, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:357219590", "Proof|ID:357219590-1"},
   CellID->55429374]},
 "ID:1086665202"->{
  Cell[569267, 16475, 811, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1086665202", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R11 \[ScriptCapitalZ]"},
   CellID->1086665202]},
 "R11 \[ScriptCapitalZ]"->{
  Cell[569267, 16475, 811, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1086665202", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R11 \[ScriptCapitalZ]"},
   CellID->1086665202]},
 "ID:963670239"->{
  Cell[570081, 16498, 811, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:963670239", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R12 \[ScriptCapitalZ]"},
   CellID->963670239]},
 "R12 \[ScriptCapitalZ]"->{
  Cell[570081, 16498, 811, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:963670239", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R12 \[ScriptCapitalZ]"},
   CellID->963670239]},
 "Proof|ID:1086665202"->{
  Cell[571048, 16530, 1820, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1086665202", "Proof|ID:1086665202-1"},
   CellID->726182844]},
 "Proof|ID:1086665202-1"->{
  Cell[571048, 16530, 1820, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1086665202", "Proof|ID:1086665202-1"},
   CellID->726182844]},
 "Proof|ID:963670239"->{
  Cell[573388, 16606, 1810, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:963670239", "Proof|ID:963670239-1"},
   CellID->70242044]},
 "Proof|ID:963670239-1"->{
  Cell[573388, 16606, 1810, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:963670239", "Proof|ID:963670239-1"},
   CellID->70242044]},
 "ID:483352095"->{
  Cell[576461, 16705, 1970, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:483352095", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[SquareSubset] \[ScriptCapitalZ]"},
   CellID->483352095]},
 "\[SquareSubset] \[ScriptCapitalZ]"->{
  Cell[576461, 16705, 1970, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:483352095", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[SquareSubset] \[ScriptCapitalZ]"},
   CellID->483352095]},
 "ID:466277585"->{
  Cell[578434, 16767, 841, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:466277585", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R13 \[ScriptCapitalZ]"},
   CellID->466277585]},
 "R13 \[ScriptCapitalZ]"->{
  Cell[578434, 16767, 841, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:466277585", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R13 \[ScriptCapitalZ]"},
   CellID->466277585]},
 "Proof|ID:483352095"->{
  Cell[579431, 16799, 5793, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:483352095", "Proof|ID:483352095-1"},
   CellID->579238228]},
 "Proof|ID:483352095-1"->{
  Cell[579431, 16799, 5793, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:483352095", "Proof|ID:483352095-1"},
   CellID->579238228]},
 "Proof|ID:466277585"->{
  Cell[585743, 16966, 1810, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:466277585", "Proof|ID:466277585-1"},
   CellID->54970627]},
 "Proof|ID:466277585-1"->{
  Cell[585743, 16966, 1810, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:466277585", "Proof|ID:466277585-1"},
   CellID->54970627]},
 "ID:25825837"->{
  Cell[588775, 17065, 845, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:25825837", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isAlgoReductionRing \[ScriptCapitalZ]"},
   CellID->25825837]},
 "isAlgoReductionRing \[ScriptCapitalZ]"->{
  Cell[588775, 17065, 845, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:25825837", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isAlgoReductionRing \[ScriptCapitalZ]"},
   CellID->25825837]},
 "Proof|ID:25825837"->{
  Cell[589775, 17098, 1885, 58, 28, "ProofInfo",
   CellTags->{"Proof|ID:25825837", "Proof|ID:25825837-1"},
   CellID->84901906]},
 "Proof|ID:25825837-1"->{
  Cell[589775, 17098, 1885, 58, 28, "ProofInfo",
   CellTags->{"Proof|ID:25825837", "Proof|ID:25825837-1"},
   CellID->84901906]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"ID:1959508349", 629799, 17940},
 {"Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", 630101, 17947},
 {"even EvenQ", 658612, 18677},
 {"ID:1696678769", 658853, 18684},
 {"ID:227367582", 659073, 18691},
 {"quo Quotient", 659312, 18698},
 {"ID:74202101", 659550, 18705},
 {"even mod", 659779, 18712},
 {"ID:226250890", 660012, 18719},
 {"even 0", 660245, 18726},
 {"IR", 660474, 18733},
 {"ID:175457915", 660932, 18746},
 {"even 1", 661165, 18753},
 {"ID:52392318", 661403, 18760},
 {"ID:782819098", 661619, 18767},
 {"even twice", 661854, 18774},
 {"ID:759529887", 662091, 18781},
 {"even +1", 662320, 18788},
 {"ID:243153755", 662554, 18795},
 {"even -1", 662783, 18802},
 {"ID:188962611", 663017, 18809},
 {"even abs", 663248, 18816},
 {"ID:1148058", 663481, 18823},
 {"even quo twice", 663721, 18830},
 {"ID:441075310", 663959, 18837},
 {"odd quo twice", 664202, 18844},
 {"ID:256130067", 664444, 18851},
 {"even quo -1", 664683, 18858},
 {"ID:739691344", 664923, 18865},
 {"odd quo -1", 665160, 18872},
 {"ID:10064687", 665398, 18879},
 {"ID:208882588", 665615, 18886},
 {"twice \[GreaterEqual]", 665874, 18893},
 {"ID:47095122", 666123, 18900},
 {"twice >", 666352, 18907},
 {"Proof|ID:226250890", 666592, 18914},
 {"Proof|ID:226250890-1", 666750, 18918},
 {"Proof|ID:175457915", 666906, 18922},
 {"Proof|ID:175457915-1", 667064, 18926},
 {"Proof|ID:782819098", 667220, 18930},
 {"Proof|ID:782819098-1", 667379, 18934},
 {"Proof|ID:759529887", 667536, 18938},
 {"Proof|ID:759529887-1", 667695, 18942},
 {"Proof|ID:243153755", 667852, 18946},
 {"Proof|ID:243153755-1", 668010, 18950},
 {"Proof|ID:188962611", 668166, 18954},
 {"Proof|ID:188962611-1", 668325, 18958},
 {"Proof|ID:1148058", 668480, 18962},
 {"Proof|ID:1148058-1", 668633, 18966},
 {"Proof|ID:441075310", 668786, 18970},
 {"Proof|ID:441075310-1", 668946, 18974},
 {"Proof|ID:256130067", 669104, 18978},
 {"Proof|ID:256130067-1", 669263, 18982},
 {"Proof|ID:739691344", 669420, 18986},
 {"Proof|ID:739691344-1", 669579, 18990},
 {"Proof|ID:208882588", 669736, 18994},
 {"Proof|ID:208882588-1", 669895, 18998},
 {"Proof|ID:47095122", 670051, 19002},
 {"Proof|ID:47095122-1", 670207, 19006},
 {"ID:1818128647", 670357, 19010},
 {"ID:489394336", 670578, 19017},
 {"ID:1246011018", 670799, 19024},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Element]", 671120, 19031},
 {"ID:234727221", 671402, 19038},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] 0", 671702, 19045},
 {"ID:849552024", 671972, 19052},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] 1", 672272, 19059},
 {"ID:1301356286", 672543, 19066},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] +", 672845, 19073},
 {"ID:196003840", 673117, 19080},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] -", 673417, 19087},
 {"ID:1933917000", 673688, 19094},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] - unary", 674002, 19101},
 {"ID:963583550", 674280, 19108},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[RawStar]", 674598, 19115},
 {"ID:1947278342", 674878, 19122},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Succeeds]", 675201, 19129},
 {"ID:90094145", 675483, 19136},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] multN", 675789, 19143},
 {"ID:1432577745", 676062, 19150},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult", 676371, 19157},
 {"ID:1000779593", 676648, 19164},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult 1", 676961, 19171},
 {"ID:948580964", 677239, 19178},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] lr", 677542, 19185},
 {"ID:1258513368", 677815, 19192},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] rdm", 678122, 19199},
 {"ID:1342406859", 678398, 19206},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd", 678708, 19213},
 {"ID:456350999", 678985, 19220},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd unary", 679305, 19227},
 {"ID:1330653407", 679587, 19234},
 {"\[ScriptCapitalR]\[DoubleStruckCapitalZ] cpm", 679895, 19241},
 {"ID:2131968713", 680172, 19248},
 {"\[ScriptCapitalZ]", 680425, 19255},
 {"ID:335127278", 680673, 19262},
 {"ID:2044589078", 680894, 19269},
 {"isCommRing1 \[ScriptCapitalZ]", 681171, 19276},
 {"ID:48028479", 681430, 19283},
 {"integral domain \[ScriptCapitalZ]", 681712, 19290},
 {"Proof|ID:2044589078", 681980, 19297},
 {"Proof|ID:2044589078-1", 682142, 19301},
 {"Proof|ID:48028479", 682300, 19305},
 {"Proof|ID:48028479-1", 682455, 19309},
 {"ID:1813588854", 682604, 19313},
 {"isPartIrreflOrder \[ScriptCapitalZ]", 682894, 19320},
 {"ID:187039377", 683161, 19327},
 {"isTotalIrreflOrder \[ScriptCapitalZ]", 683451, 19334},
 {"ID:159454607", 683717, 19341},
 {"isNoetherian \[ScriptCapitalZ]", 683995, 19348},
 {"Proof|ID:1813588854", 684262, 19355},
 {"Proof|ID:1813588854-1", 684424, 19359},
 {"Proof|ID:187039377", 684583, 19363},
 {"Proof|ID:187039377-1", 684743, 19367},
 {"Proof|ID:159454607", 684901, 19371},
 {"Proof|ID:159454607-1", 685061, 19375},
 {"ID:38031932", 685212, 19379},
 {"mult alternative \[ScriptCapitalZ]", 685496, 19386},
 {"ID:1890039364", 685759, 19393},
 {"Multipliers \[ScriptCapitalZ]", 686036, 19400},
 {"ID:1152753107", 686297, 19407},
 {"R0 \[ScriptCapitalZ]", 686556, 19414},
 {"ID:307366433", 686807, 19421},
 {"R1 \[ScriptCapitalZ]", 687064, 19428},
 {"ID:963452595", 687313, 19435},
 {"R2 \[ScriptCapitalZ]", 687570, 19442},
 {"ID:1084156204", 687820, 19449},
 {"R3 \[ScriptCapitalZ]", 688079, 19456},
 {"ID:2041570274", 688331, 19463},
 {"R4 \[ScriptCapitalZ]", 688590, 19470},
 {"Proof|ID:38031932", 688846, 19477},
 {"Proof|ID:38031932-1", 689004, 19481},
 {"Proof|ID:1890039364", 689162, 19485},
 {"Proof|ID:1890039364-1", 689325, 19489},
 {"Proof|ID:1152753107", 689486, 19493},
 {"Proof|ID:1152753107-1", 689649, 19497},
 {"Proof|ID:307366433", 689809, 19501},
 {"Proof|ID:307366433-1", 689968, 19505},
 {"Proof|ID:963452595", 690125, 19509},
 {"Proof|ID:963452595-1", 690283, 19513},
 {"Proof|ID:1084156204", 690440, 19517},
 {"Proof|ID:1084156204-1", 690603, 19521},
 {"Proof|ID:2041570274", 690764, 19525},
 {"Proof|ID:2041570274-1", 690927, 19529},
 {"ID:260068164", 691081, 19533},
 {"ID:211178897", 691301, 19540},
 {"membership sgn \[ScriptCapitalZ]", 691582, 19547},
 {"ID:118405539", 691843, 19554},
 {"\[RawStar] sgn \[ScriptCapitalZ]", 692125, 19561},
 {"ID:88914628", 692386, 19568},
 {"0 minimal \[ScriptCapitalZ]", 692655, 19575},
 {"ID:69480817", 692908, 19582},
 {"twice \[Succeeds] twice \[ScriptCapitalZ]", 693206, 19589},
 {"ID:24170142", 693474, 19596},
 {"\[Succeeds] + \[DoubleRightArrow] \[ScriptCapitalZ]", 693792, 19603},
 {"ID:302704588", 694071, 19610},
 {"\[Succeeds] + \[DoubleLeftArrow] \[ScriptCapitalZ]", 694389, 19617},
 {"ID:131195953", 694669, 19624},
 {"\[Succeeds] + \[ScriptCapitalZ]", 694949, 19631},
 {"ID:129207038", 695210, 19638},
 {"\[Succeeds] - \[ScriptCapitalZ]", 695490, 19645},
 {"ID:59300728", 695750, 19652},
 {"\[Succeeds] alternative \[ScriptCapitalZ]", 696048, 19659},
 {"ID:698610132", 696317, 19666},
 {"\[Succeeds] +' \[ScriptCapitalZ]", 696599, 19673},
 {"Proof|ID:211178897", 696867, 19680},
 {"Proof|ID:211178897-1", 697027, 19684},
 {"Proof|ID:118405539", 697185, 19688},
 {"Proof|ID:118405539-1", 697345, 19692},
 {"Proof|ID:88914628", 697502, 19696},
 {"Proof|ID:88914628-1", 697659, 19700},
 {"Proof|ID:69480817", 697814, 19704},
 {"Proof|ID:69480817-1", 697970, 19708},
 {"Proof|ID:24170142", 698124, 19712},
 {"Proof|ID:24170142-1", 698282, 19716},
 {"Proof|ID:302704588", 698439, 19720},
 {"Proof|ID:302704588-1", 698600, 19724},
 {"Proof|ID:131195953", 698759, 19728},
 {"Proof|ID:131195953-1", 698919, 19732},
 {"Proof|ID:129207038", 699077, 19736},
 {"Proof|ID:129207038-1", 699237, 19740},
 {"Proof|ID:59300728", 699394, 19744},
 {"Proof|ID:59300728-1", 699552, 19748},
 {"Proof|ID:698610132", 699709, 19752},
 {"Proof|ID:698610132-1", 699870, 19756},
 {"ID:310863676", 700023, 19760},
 {"ID:471044386", 700243, 19767},
 {"isReducible - \[ScriptCapitalZ]", 700523, 19774},
 {"ID:379223163", 700784, 19781},
 {"isReducible alternative \[ScriptCapitalZ]", 701084, 19788},
 {"ID:176657795", 701355, 19795},
 {"isReducible sgn \[ScriptCapitalZ]", 701639, 19802},
 {"ID:217289551", 701902, 19809},
 {"isReducible alternative' \[DoubleRightArrow] \[ScriptCapitalZ]", 702250, \
19817},
 {"ID:497414030", 702548, 19825},
 {"isReducible alternative' \[DoubleLeftArrow] \[ScriptCapitalZ]", 702894, \
19833},
 {"ID:464408938", 703191, 19841},
 {"isReducible alternative' \[ScriptCapitalZ]", 703493, 19848},
 {"ID:304230885", 703765, 19855},
 {"isReducible -' \[ScriptCapitalZ]", 704047, 19862},
 {"ID:97311931", 704308, 19869},
 {"\[RightArrow] - \[ScriptCapitalZ]", 704590, 19876},
 {"ID:99245919", 704850, 19883},
 {"isReducible \[Succeeds] \[ScriptCapitalZ]", 705148, 19890},
 {"ID:162605324", 705417, 19897},
 {"isReducible \[Succeeds]' \[ScriptCapitalZ]", 705719, 19904},
 {"ID:40987937", 705990, 19911},
 {"isReducible \[Succeeds]'' \[ScriptCapitalZ]", 706292, 19918},
 {"Proof|ID:471044386", 706569, 19925},
 {"Proof|ID:471044386-1", 706730, 19929},
 {"Proof|ID:379223163", 706889, 19933},
 {"Proof|ID:379223163-1", 707050, 19937},
 {"Proof|ID:176657795", 707209, 19941},
 {"Proof|ID:176657795-1", 707370, 19945},
 {"Proof|ID:217289551", 707529, 19949},
 {"Proof|ID:217289551-1", 707690, 19953},
 {"Proof|ID:497414030", 707849, 19957},
 {"Proof|ID:497414030-1", 708010, 19961},
 {"Proof|ID:464408938", 708169, 19965},
 {"Proof|ID:464408938-1", 708330, 19969},
 {"Proof|ID:304230885", 708489, 19973},
 {"Proof|ID:304230885-1", 708650, 19977},
 {"Proof|ID:97311931", 708808, 19981},
 {"Proof|ID:97311931-1", 708966, 19985},
 {"Proof|ID:99245919", 709122, 19989},
 {"Proof|ID:99245919-1", 709280, 19993},
 {"Proof|ID:162605324", 709437, 19997},
 {"Proof|ID:162605324-1", 709599, 20001},
 {"Proof|ID:40987937", 709758, 20005},
 {"Proof|ID:40987937-1", 709917, 20009},
 {"ID:11694492", 710068, 20013},
 {"ID:224215037", 710287, 20020},
 {"membership lr \[ScriptCapitalZ]", 710567, 20027},
 {"ID:78177024", 710827, 20034},
 {"lr = 0 \[ScriptCapitalZ]", 711092, 20041},
 {"ID:634190310", 711345, 20048},
 {"lr > 0 \[ScriptCapitalZ]", 711612, 20055},
 {"ID:588999973", 711867, 20062},
 {"0 > lr \[ScriptCapitalZ]", 712134, 20069},
 {"ID:57477705", 712388, 20076},
 {"lr twice even \[ScriptCapitalZ]", 712667, 20083},
 {"ID:206701743", 712927, 20090},
 {"lr twice odd \[ScriptCapitalZ]", 713206, 20097},
 {"ID:34042360", 713466, 20104},
 {"lr isReducible \[ScriptCapitalZ]", 713747, 20111},
 {"ID:692837602", 714008, 20118},
 {"lr minimal isReducible \[ScriptCapitalZ]", 714307, 20125},
 {"ID:44304045", 714577, 20132},
 {"lr isReducible sgn \[ScriptCapitalZ]", 714866, 20139},
 {"Proof|ID:224215037", 715137, 20146},
 {"Proof|ID:224215037-1", 715298, 20150},
 {"Proof|ID:78177024", 715456, 20154},
 {"Proof|ID:78177024-1", 715614, 20158},
 {"Proof|ID:634190310", 715771, 20162},
 {"Proof|ID:634190310-1", 715931, 20166},
 {"Proof|ID:588999973", 716089, 20170},
 {"Proof|ID:588999973-1", 716248, 20174},
 {"Proof|ID:57477705", 716404, 20178},
 {"Proof|ID:57477705-1", 716562, 20182},
 {"Proof|ID:206701743", 716719, 20186},
 {"Proof|ID:206701743-1", 716880, 20190},
 {"Proof|ID:34042360", 717038, 20194},
 {"Proof|ID:34042360-1", 717196, 20198},
 {"Proof|ID:692837602", 717353, 20202},
 {"Proof|ID:692837602-1", 717515, 20206},
 {"Proof|ID:44304045", 717674, 20210},
 {"Proof|ID:44304045-1", 717833, 20214},
 {"ID:16238117", 717984, 20218},
 {"ID:468309069", 718203, 20225},
 {"\[Element] multN \[ScriptCapitalZ]", 718490, 20232},
 {"ID:560244434", 718755, 20239},
 {"ID:142903111", 718976, 20246},
 {"areIrrelative alternative \[ScriptCapitalZ]", 719281, 20253},
 {"ID:36948313", 719554, 20260},
 {"isCommonReducible alternative \[ScriptCapitalZ]", 719865, 20267},
 {"ID:46630982", 720140, 20274},
 {"isCommonReducible alternative' \[ScriptCapitalZ]", 720453, 20281},
 {"Proof|ID:468309069", 720736, 20288},
 {"Proof|ID:468309069-1", 720897, 20292},
 {"Proof|ID:142903111", 721056, 20296},
 {"Proof|ID:142903111-1", 721216, 20300},
 {"Proof|ID:36948313", 721373, 20304},
 {"Proof|ID:36948313-1", 721532, 20308},
 {"Proof|ID:46630982", 721689, 20312},
 {"Proof|ID:46630982-1", 721848, 20316},
 {"ID:69669781", 721999, 20320},
 {"lr isMNTCR unary \[ScriptCapitalZ]", 722284, 20327},
 {"ID:65735690", 722546, 20334},
 {"ID:431845789", 722765, 20341},
 {"ID:96649542", 722985, 20348},
 {"lr isNTCR \[ScriptCapitalZ]", 723256, 20355},
 {"ID:132861224", 723512, 20362},
 {"lr isMNTCR \[ScriptCapitalZ]", 723787, 20369},
 {"Proof|ID:69669781", 724051, 20376},
 {"Proof|ID:69669781-1", 724208, 20380},
 {"Proof|ID:96649542", 724363, 20384},
 {"Proof|ID:96649542-1", 724522, 20388},
 {"Proof|ID:132861224", 724680, 20392},
 {"Proof|ID:132861224-1", 724841, 20396},
 {"ID:595906550", 724994, 20400},
 {"R5 \[ScriptCapitalZ]", 725252, 20407},
 {"ID:384714041", 725502, 20414},
 {"R6 \[ScriptCapitalZ]", 725760, 20421},
 {"ID:739712574", 726010, 20428},
 {"R7 \[ScriptCapitalZ]", 726268, 20435},
 {"ID:782958792", 726518, 20442},
 {"R8 \[ScriptCapitalZ]", 726776, 20449},
 {"Proof|ID:595906550", 727032, 20456},
 {"Proof|ID:595906550-1", 727193, 20460},
 {"Proof|ID:384714041", 727352, 20464},
 {"Proof|ID:384714041-1", 727511, 20468},
 {"Proof|ID:739712574", 727668, 20472},
 {"Proof|ID:739712574-1", 727829, 20476},
 {"Proof|ID:782958792", 727988, 20480},
 {"Proof|ID:782958792-1", 728149, 20484},
 {"ID:67773804", 728301, 20488},
 {"isReductionRing \[ScriptCapitalZ]", 728583, 20495},
 {"Proof|ID:67773804", 728849, 20502},
 {"Proof|ID:67773804-1", 729007, 20506},
 {"ID:156995908", 729158, 20510},
 {"R9 \[ScriptCapitalZ]", 729416, 20517},
 {"ID:1051911108", 729667, 20524},
 {"R10 \[ScriptCapitalZ]", 729929, 20531},
 {"Proof|ID:156995908", 730188, 20538},
 {"Proof|ID:156995908-1", 730349, 20542},
 {"Proof|ID:1051911108", 730509, 20546},
 {"Proof|ID:1051911108-1", 730672, 20550},
 {"ID:160940549", 730826, 20554},
 {"ID:46650045", 731046, 20561},
 {"ID:816204582", 731265, 20568},
 {"rdm pos \[GreaterEqual] 0 \[ScriptCapitalZ]", 731570, 20575},
 {"ID:171042210", 731844, 20582},
 {"rdm neg \[GreaterEqual] 0 \[ScriptCapitalZ]", 732149, 20589},
 {"ID:160702195", 732423, 20596},
 {"rdm pos 0 \[ScriptCapitalZ]", 732696, 20603},
 {"ID:478744619", 732954, 20610},
 {"rdm neg 0 \[ScriptCapitalZ]", 733227, 20617},
 {"ID:680200521", 733485, 20624},
 {"rdm pos isReducible \[ScriptCapitalZ]", 733778, 20631},
 {"ID:122969425", 734046, 20638},
 {"rdm neg isReducible \[ScriptCapitalZ]", 734339, 20645},
 {"ID:509693672", 734607, 20652},
 {"membership rdm' \[ScriptCapitalZ]", 734891, 20659},
 {"ID:576030870", 735154, 20666},
 {"rdm = 0' \[ScriptCapitalZ]", 735425, 20673},
 {"ID:825630046", 735682, 20680},
 {"rdm isReducible' \[ScriptCapitalZ]", 735969, 20687},
 {"Proof|ID:816204582", 736240, 20694},
 {"Proof|ID:816204582-1", 736402, 20698},
 {"Proof|ID:171042210", 736562, 20702},
 {"Proof|ID:171042210-1", 736723, 20706},
 {"Proof|ID:160702195", 736882, 20710},
 {"Proof|ID:160702195-1", 737044, 20714},
 {"Proof|ID:478744619", 737204, 20718},
 {"Proof|ID:478744619-1", 737367, 20722},
 {"Proof|ID:680200521", 737528, 20726},
 {"Proof|ID:680200521-1", 737689, 20730},
 {"Proof|ID:122969425", 737848, 20734},
 {"Proof|ID:122969425-1", 738010, 20738},
 {"Proof|ID:509693672", 738170, 20742},
 {"Proof|ID:509693672-1", 738331, 20746},
 {"Proof|ID:576030870", 738490, 20750},
 {"Proof|ID:576030870-1", 738652, 20754},
 {"Proof|ID:825630046", 738812, 20758},
 {"Proof|ID:825630046-1", 738973, 20762},
 {"ID:47390677", 739125, 20766},
 {"ID:362000401", 739344, 20773},
 {"membership rdm \[ScriptCapitalZ]", 739626, 20780},
 {"ID:98296722", 739887, 20787},
 {"rdm isReducible \[ScriptCapitalZ]", 740170, 20794},
 {"ID:357219590", 740432, 20801},
 {"rdm = 0 \[ScriptCapitalZ]", 740701, 20808},
 {"Proof|ID:362000401", 740963, 20815},
 {"Proof|ID:362000401-1", 741123, 20819},
 {"Proof|ID:98296722", 741280, 20823},
 {"Proof|ID:98296722-1", 741439, 20827},
 {"Proof|ID:357219590", 741597, 20831},
 {"Proof|ID:357219590-1", 741758, 20835},
 {"ID:1086665202", 741912, 20839},
 {"R11 \[ScriptCapitalZ]", 742174, 20846},
 {"ID:963670239", 742427, 20853},
 {"R12 \[ScriptCapitalZ]", 742687, 20860},
 {"Proof|ID:1086665202", 742945, 20867},
 {"Proof|ID:1086665202-1", 743109, 20871},
 {"Proof|ID:963670239", 743270, 20875},
 {"Proof|ID:963670239-1", 743430, 20879},
 {"ID:483352095", 743582, 20883},
 {"\[SquareSubset] \[ScriptCapitalZ]", 743867, 20890},
 {"ID:466277585", 744131, 20897},
 {"R13 \[ScriptCapitalZ]", 744391, 20904},
 {"Proof|ID:483352095", 744648, 20911},
 {"Proof|ID:483352095-1", 744810, 20915},
 {"Proof|ID:466277585", 744970, 20919},
 {"Proof|ID:466277585-1", 745130, 20923},
 {"ID:25825837", 745281, 20927},
 {"isAlgoReductionRing \[ScriptCapitalZ]", 745571, 20934},
 {"Proof|ID:25825837", 745841, 20941},
 {"Proof|ID:25825837-1", 745998, 20945}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[579, 22, 201, 3, 84, "Title",
 CellID->2066005206],
Cell[CellGroupData[{
Cell[805, 29, 168, 3, 73, "Section",
 CellID->1422121268],
Cell[CellGroupData[{
Cell[998, 36, 121, 2, 70, "Subsection",
 CellID->163575360],
Cell[1122, 40, 533, 9, 70, "Text",
 CellID->597779389]
}, Closed]],
Cell[CellGroupData[{
Cell[1692, 54, 125, 2, 70, "Subsection",
 CellID->376409692],
Cell[1820, 58, 344, 8, 70, "Input",
 CellID->498486718],
Cell[2167, 68, 355, 8, 70, "Input",
 CellID->406742757],
Cell[2525, 78, 419, 9, 70, "Input",
 CellID->432730220],
Cell[2947, 89, 406, 9, 70, "Input",
 CellID->135231296],
Cell[3356, 100, 457, 10, 70, "Input",
 CellID->45708515],
Cell[3816, 112, 513, 10, 70, "Input",
 CellID->851969895],
Cell[4332, 124, 561, 11, 70, "Input",
 CellID->559226997],
Cell[4896, 137, 151, 3, 70, "Input",
 CellID->524928492],
Cell[5050, 142, 154, 3, 70, "Input",
 CellID->31934633],
Cell[5207, 147, 208, 4, 70, "Input",
 CellID->113843512]
}, Closed]],
Cell[CellGroupData[{
Cell[5452, 156, 124, 2, 70, "Subsection",
 CellID->1386333768],
Cell[CellGroupData[{
Cell[5601, 162, 356, 10, 70, "Input",
 CellID->1039312110],
Cell[CellGroupData[{
Cell[5982, 176, 597, 11, 70, "Print",
 CellID->1754373566],
Cell[6582, 189, 519, 10, 70, "Print",
 CellID->1029482356],
Cell[7104, 201, 519, 10, 70, "Print",
 CellID->676061299],
Cell[7626, 213, 521, 10, 70, "Print",
 CellID->1850448490],
Cell[8150, 225, 519, 10, 70, "Print",
 CellID->70235299],
Cell[8672, 237, 520, 10, 70, "Print",
 CellID->426922802],
Cell[9195, 249, 521, 10, 70, "Print",
 CellID->1237071357],
Cell[9719, 261, 521, 10, 70, "Print",
 CellID->1870636140],
Cell[10243, 273, 599, 11, 70, "Print",
 CellID->802449962],
Cell[10845, 286, 519, 10, 70, "Print",
 CellID->1864026098],
Cell[11367, 298, 519, 10, 70, "Print",
 CellID->374700447],
Cell[11889, 310, 520, 10, 70, "Print",
 CellID->808617339],
Cell[12412, 322, 580, 11, 70, "Print",
 CellID->39084793],
Cell[12995, 335, 519, 10, 70, "Print",
 CellID->1889518183],
Cell[13517, 347, 520, 10, 70, "Print",
 CellID->2026876725],
Cell[14040, 359, 520, 10, 70, "Print",
 CellID->349652243],
Cell[14563, 371, 519, 10, 70, "Print",
 CellID->1812907063],
Cell[15085, 383, 497, 10, 70, "Print",
 CellID->1437928279]
}, Open  ]]
}, Open  ]],
Cell[15609, 397, 325, 6, 70, "Input",
 CellID->19428626],
Cell[CellGroupData[{
Cell[15959, 407, 541, 12, 70, "Input",
 CellID->1909803754],
Cell[16503, 421, 735, 12, 70, "Output",
 CellID->636966437]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[17287, 439, 128, 2, 70, "Subsection",
 CellID->1454182661],
Cell[17418, 443, 982, 25, 70, "Input",
 CellID->2074904605],
Cell[18403, 470, 416, 9, 70, "Input",
 CellID->1980565193],
Cell[18822, 481, 169, 3, 70, "Input",
 CellID->444074165],
Cell[18994, 486, 280, 6, 70, "Input",
 CellID->6535056]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[19323, 498, 165, 3, 43, "Section",
 CellID->2107536411],
Cell[CellGroupData[{
Cell[19513, 505, 130, 2, 38, "Subsection",
 CellID->640760558],
Cell[19646, 509, 925, 26, 70, "Text",
 CellID->36588863],
Cell[20574, 537, 122, 2, 70, "Text",
 CellID->1883788369],
Cell[CellGroupData[{
Cell[20721, 543, 427, 12, 70, "Item",
 CellID->440934229],
Cell[21151, 557, 183, 3, 70, "Item",
 CellID->504979177],
Cell[21337, 562, 772, 18, 70, "Item",
 CellID->2079792210]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[22158, 586, 131, 2, 30, "Subsection",
 CellID->470486761],
Cell[CellGroupData[{
Cell[22314, 592, 124, 2, 28, "Subsubsection",
 CellID->248992438],
Cell[22441, 596, 47, 1, 12, "OpenEnvironment",
 CellID->740029918],
Cell[CellGroupData[{
Cell[22513, 601, 371, 9, 43, "EnvironmentHeader",
 CellID->164698789],
Cell[22887, 612, 1225, 37, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1959508349", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even EvenQ"},
 CellID->1959508349],
Cell[24115, 651, 64, 1, 26, "EndEnvironmentMarker",
 CellID->94497015]
}, Open  ]],
Cell[24194, 655, 47, 1, 16, "CloseEnvironment",
 CellID->33655351],
Cell[24244, 658, 817, 16, 63, "Text",
 CellID->938862428],
Cell[25064, 676, 48, 1, 12, "OpenEnvironment",
 CellID->1583138317],
Cell[CellGroupData[{
Cell[25137, 681, 507, 12, 43, "EnvironmentHeader",
 CellID->416750769],
Cell[25647, 695, 919, 22, 32, "GlobalDeclaration",
 CellTags->{
  "ID:1696678769", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->1696678769],
Cell[26569, 719, 969, 26, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:227367582", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "quo Quotient"},
 CellID->227367582],
Cell[27541, 747, 973, 29, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:74202101", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even mod"},
 CellID->74202101],
Cell[28517, 778, 64, 1, 26, "EndEnvironmentMarker",
 CellID->96759121]
}, Open  ]],
Cell[28596, 782, 49, 1, 16, "CloseEnvironment",
 CellID->1750417022]
}, Closed]],
Cell[CellGroupData[{
Cell[28682, 788, 182, 3, 22, "Subsubsection",
 CellID->12017291],
Cell[28867, 793, 46, 1, 70, "OpenEnvironment",
 CellID->29963980],
Cell[CellGroupData[{
Cell[28938, 798, 387, 9, 70, "EnvironmentHeader",
 CellID->462348848],
Cell[29328, 809, 747, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:226250890", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even 0", "IR"},
 CellID->226250890],
Cell[30078, 831, 770, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:175457915", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even 1", "IR"},
 CellID->175457915],
Cell[30851, 854, 694, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:52392318", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->52392318],
Cell[31548, 874, 770, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:782819098", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even twice"},
 CellID->782819098],
Cell[32321, 897, 968, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:759529887", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even +1"},
 CellID->759529887],
Cell[33292, 928, 970, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:243153755", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even -1"},
 CellID->243153755],
Cell[34265, 959, 887, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:188962611", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even abs"},
 CellID->188962611],
Cell[35155, 985, 1103, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1148058", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even quo twice"},
 CellID->1148058],
Cell[36261, 1020, 1292, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:441075310", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "odd quo twice"},
 CellID->441075310],
Cell[37556, 1064, 1075, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:256130067", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "even quo -1"},
 CellID->256130067],
Cell[38634, 1099, 1167, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:739691344", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "odd quo -1"},
 CellID->739691344],
Cell[39804, 1139, 694, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:10064687", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->10064687],
Cell[40501, 1159, 1261, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:208882588", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice \[GreaterEqual]"},
 CellID->208882588],
Cell[41765, 1198, 1058, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:47095122", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice >"},
 CellID->47095122],
Cell[42826, 1235, 63, 1, 70, "EndEnvironmentMarker",
 CellID->6197297]
}, Open  ]],
Cell[42904, 1239, 45, 1, 70, "CloseEnvironment",
 CellID->874473],
Cell[CellGroupData[{
Cell[42974, 1244, 1693, 54, 70, "ProofInfo",
 CellTags->{"Proof|ID:226250890", "Proof|ID:226250890-1"},
 CellID->75108773],
Cell[44670, 1300, 480, 12, 70, "ProofInfoBody",
 CellID->593942575]
}, Closed]],
Cell[CellGroupData[{
Cell[45187, 1317, 1973, 62, 70, "ProofInfo",
 CellTags->{"Proof|ID:175457915", "Proof|ID:175457915-1"},
 CellID->36435349],
Cell[47163, 1381, 479, 12, 70, "ProofInfoBody",
 CellID->33117230]
}, Closed]],
Cell[CellGroupData[{
Cell[47679, 1398, 2385, 72, 70, "ProofInfo",
 CellTags->{"Proof|ID:782819098", "Proof|ID:782819098-1"},
 CellID->387405268],
Cell[50067, 1472, 480, 12, 70, "ProofInfoBody",
 CellID->814506024]
}, Closed]],
Cell[CellGroupData[{
Cell[50584, 1489, 2948, 88, 70, "ProofInfo",
 CellTags->{"Proof|ID:759529887", "Proof|ID:759529887-1"},
 CellID->198480204],
Cell[53535, 1579, 480, 12, 70, "ProofInfoBody",
 CellID->401602925]
}, Closed]],
Cell[CellGroupData[{
Cell[54052, 1596, 2947, 88, 70, "ProofInfo",
 CellTags->{"Proof|ID:243153755", "Proof|ID:243153755-1"},
 CellID->86433169],
Cell[57002, 1686, 480, 12, 70, "ProofInfoBody",
 CellID->250829815]
}, Closed]],
Cell[CellGroupData[{
Cell[57519, 1703, 2687, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:188962611", "Proof|ID:188962611-1"},
 CellID->608316900],
Cell[60209, 1783, 480, 12, 70, "ProofInfoBody",
 CellID->665515500]
}, Closed]],
Cell[CellGroupData[{
Cell[60726, 1800, 3156, 94, 70, "ProofInfo",
 CellTags->{"Proof|ID:1148058", "Proof|ID:1148058-1"},
 CellID->314369717],
Cell[63885, 1896, 476, 12, 70, "ProofInfoBody",
 CellID->5103723]
}, Closed]],
Cell[CellGroupData[{
Cell[64398, 1913, 3584, 104, 70, "ProofInfo",
 CellTags->{"Proof|ID:441075310", "Proof|ID:441075310-1"},
 CellID->900195017],
Cell[67985, 2019, 480, 12, 70, "ProofInfoBody",
 CellID->387398711]
}, Closed]],
Cell[CellGroupData[{
Cell[68502, 2036, 3103, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:256130067", "Proof|ID:256130067-1"},
 CellID->748444859],
Cell[71608, 2128, 480, 12, 70, "ProofInfoBody",
 CellID->312842733]
}, Closed]],
Cell[CellGroupData[{
Cell[72125, 2145, 3341, 96, 70, "ProofInfo",
 CellTags->{"Proof|ID:739691344", "Proof|ID:739691344-1"},
 CellID->138102035],
Cell[75469, 2243, 480, 12, 70, "ProofInfoBody",
 CellID->242339988]
}, Closed]],
Cell[CellGroupData[{
Cell[75986, 2260, 3323, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:208882588", "Proof|ID:208882588-1"},
 CellID->148478097],
Cell[79312, 2354, 549, 13, 70, "ProofInfoBody",
 CellID->54070004]
}, Closed]],
Cell[CellGroupData[{
Cell[79898, 2372, 3234, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:47095122", "Proof|ID:47095122-1"},
 CellID->275585894],
Cell[83135, 2466, 479, 12, 70, "ProofInfoBody",
 CellID->454783722]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[83675, 2485, 180, 3, 30, "Subsection",
 CellID->1312761442],
Cell[83858, 2490, 47, 1, 70, "OpenEnvironment",
 CellID->104840774],
Cell[CellGroupData[{
Cell[83930, 2495, 501, 11, 70, "EnvironmentHeader",
 CellID->262829580],
Cell[84434, 2508, 784, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1818128647", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->1818128647],
Cell[85221, 2529, 1008, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:489394336", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->489394336],
Cell[86232, 2554, 1323, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1246011018", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Element]"},
 CellID->1246011018],
Cell[87558, 2595, 836, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:234727221", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] 0"},
 CellID->234727221],
Cell[88397, 2619, 837, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:849552024", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] 1"},
 CellID->849552024],
Cell[89237, 2643, 887, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1301356286", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] +"},
 CellID->1301356286],
Cell[90127, 2669, 883, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:196003840", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] -"},
 CellID->196003840],
Cell[91013, 2695, 898, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1933917000", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] - unary"},
 CellID->1933917000],
Cell[91914, 2722, 884, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:963583550", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[RawStar]"},
 CellID->963583550],
Cell[92801, 2748, 1735, 52, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1947278342", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] \[Succeeds]"},
 CellID->1947278342],
Cell[94539, 2802, 987, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:90094145", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] multN"},
 CellID->90094145],
Cell[95529, 2831, 1154, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1432577745", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult"},
 CellID->1432577745],
Cell[96686, 2867, 1387, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1000779593", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] mult 1"},
 CellID->1000779593],
Cell[98076, 2909, 1631, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:948580964", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lr"},
 CellID->948580964],
Cell[99710, 2955, 2942, 73, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1258513368", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] rdm"},
 CellID->1258513368],
Cell[102655, 3030, 467, 14, 70, "EnvironmentText",
 CellID->1870060344],
Cell[103125, 3046, 1982, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1342406859", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd"},
 CellID->1342406859],
Cell[105110, 3099, 1357, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:456350999", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] lcrd unary"},
 CellID->456350999],
Cell[106470, 3138, 1330, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1330653407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalR]\[DoubleStruckCapitalZ] cpm"},
 CellID->1330653407],
Cell[107803, 3175, 65, 1, 70, "EndEnvironmentMarker",
 CellID->629448807]
}, Open  ]],
Cell[107883, 3179, 49, 1, 70, "CloseEnvironment",
 CellID->1539081838]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[107981, 3186, 111, 2, 43, "Section",
 CellID->1210929308],
Cell[108095, 3190, 46, 1, 70, "OpenEnvironment",
 CellID->81008174],
Cell[CellGroupData[{
Cell[108166, 3195, 433, 10, 70, "EnvironmentHeader",
 CellID->1093450566],
Cell[108602, 3207, 861, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2131968713", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[ScriptCapitalZ]"},
 CellID->2131968713],
Cell[109466, 3231, 65, 1, 70, "EndEnvironmentMarker",
 CellID->121089228]
}, Open  ]],
Cell[109546, 3235, 48, 1, 70, "CloseEnvironment",
 CellID->910413054],
Cell[CellGroupData[{
Cell[109619, 3240, 113, 2, 70, "Subsection",
 CellID->321798810],
Cell[CellGroupData[{
Cell[109757, 3246, 240, 6, 70, "Computation",
 CellID->1429456864],
Cell[110000, 3254, 400, 7, 70, "ComputationResult",
 CellID->978138576]
}, Open  ]],
Cell[CellGroupData[{
Cell[110437, 3266, 380, 10, 70, "ComputationInfo",
 CellID->1265921044],
Cell[110820, 3278, 650, 17, 70, "ComputationInfoBody",
 CellID->1885477897]
}, Closed]],
Cell[CellGroupData[{
Cell[111507, 3300, 256, 7, 70, "Computation",
 CellID->157142288],
Cell[111766, 3309, 342, 7, 70, "ComputationResult",
 CellID->1465402715]
}, Open  ]],
Cell[CellGroupData[{
Cell[112145, 3321, 376, 10, 70, "ComputationInfo",
 CellID->473424448],
Cell[112524, 3333, 646, 17, 70, "ComputationInfoBody",
 CellID->786290236]
}, Closed]],
Cell[CellGroupData[{
Cell[113207, 3355, 238, 6, 70, "Computation",
 CellID->1960775227],
Cell[113448, 3363, 299, 5, 70, "ComputationResult",
 CellID->378014128]
}, Open  ]],
Cell[CellGroupData[{
Cell[113784, 3373, 378, 10, 70, "ComputationInfo",
 CellID->1901130687],
Cell[114165, 3385, 647, 17, 70, "ComputationInfoBody",
 CellID->944906225]
}, Closed]],
Cell[CellGroupData[{
Cell[114849, 3407, 239, 6, 70, "Computation",
 CellID->1637496245],
Cell[115091, 3415, 300, 5, 70, "ComputationResult",
 CellID->1717908025]
}, Open  ]],
Cell[CellGroupData[{
Cell[115428, 3425, 378, 10, 70, "ComputationInfo",
 CellID->1813781623],
Cell[115809, 3437, 649, 17, 70, "ComputationInfoBody",
 CellID->890136672]
}, Closed]],
Cell[CellGroupData[{
Cell[116495, 3459, 238, 6, 70, "Computation",
 CellID->2123737964],
Cell[116736, 3467, 309, 6, 70, "ComputationResult",
 CellID->1518450864]
}, Open  ]],
Cell[CellGroupData[{
Cell[117082, 3478, 377, 10, 70, "ComputationInfo",
 CellID->975643154],
Cell[117462, 3490, 647, 17, 70, "ComputationInfoBody",
 CellID->1403959631]
}, Closed]],
Cell[CellGroupData[{
Cell[118146, 3512, 256, 7, 70, "Computation",
 CellID->252382512],
Cell[118405, 3521, 311, 6, 70, "ComputationResult",
 CellID->1505423261]
}, Open  ]],
Cell[CellGroupData[{
Cell[118753, 3532, 378, 10, 70, "ComputationInfo",
 CellID->796291593],
Cell[119134, 3544, 647, 17, 70, "ComputationInfoBody",
 CellID->1915561639]
}, Closed]],
Cell[CellGroupData[{
Cell[119818, 3566, 277, 8, 70, "Computation",
 CellID->266870911],
Cell[120098, 3576, 310, 6, 70, "ComputationResult",
 CellID->116083407]
}, Open  ]],
Cell[CellGroupData[{
Cell[120445, 3587, 377, 10, 70, "ComputationInfo",
 CellID->2029286797],
Cell[120825, 3599, 647, 17, 70, "ComputationInfoBody",
 CellID->1380713228]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[121521, 3622, 111, 2, 70, "Subsection",
 CellID->1370888350],
Cell[CellGroupData[{
Cell[121657, 3628, 374, 8, 70, "Computation",
 CellID->1703330943],
Cell[122034, 3638, 421, 10, 70, "ComputationResult",
 CellID->1536458168]
}, Open  ]],
Cell[CellGroupData[{
Cell[122492, 3653, 379, 10, 70, "ComputationInfo",
 CellID->246257879],
Cell[122874, 3665, 648, 17, 70, "ComputationInfoBody",
 CellID->1309963877]
}, Closed]],
Cell[CellGroupData[{
Cell[123559, 3687, 264, 6, 70, "Computation",
 CellID->1150003754],
Cell[123826, 3695, 204, 7, 70, "ComputationResult",
 CellID->2113262834]
}, Open  ]],
Cell[CellGroupData[{
Cell[124067, 3707, 378, 10, 70, "ComputationInfo",
 CellID->1719993529],
Cell[124448, 3719, 648, 17, 70, "ComputationInfoBody",
 CellID->1784884467]
}, Closed]],
Cell[CellGroupData[{
Cell[125133, 3741, 259, 6, 70, "Computation",
 CellID->1557862139],
Cell[125395, 3749, 204, 7, 70, "ComputationResult",
 CellID->1353116481]
}, Open  ]],
Cell[CellGroupData[{
Cell[125636, 3761, 377, 10, 70, "ComputationInfo",
 CellID->1512462514],
Cell[126016, 3773, 647, 17, 70, "ComputationInfoBody",
 CellID->642311510]
}, Closed]],
Cell[CellGroupData[{
Cell[126700, 3795, 278, 7, 70, "Computation",
 CellID->581379970],
Cell[126981, 3804, 225, 7, 70, "ComputationResult",
 CellID->1478552428]
}, Open  ]],
Cell[CellGroupData[{
Cell[127243, 3816, 376, 10, 70, "ComputationInfo",
 CellID->1858231219],
Cell[127622, 3828, 646, 17, 70, "ComputationInfoBody",
 CellID->382299358]
}, Closed]],
Cell[CellGroupData[{
Cell[128305, 3850, 219, 6, 70, "Computation",
 CellID->1243235648],
Cell[128527, 3858, 204, 7, 70, "ComputationResult",
 CellID->1849233252]
}, Open  ]],
Cell[CellGroupData[{
Cell[128768, 3870, 377, 10, 70, "ComputationInfo",
 CellID->505878642],
Cell[129148, 3882, 650, 17, 70, "ComputationInfoBody",
 CellID->1259921456]
}, Closed]],
Cell[CellGroupData[{
Cell[129835, 3904, 238, 6, 70, "Computation",
 CellID->1587692045],
Cell[130076, 3912, 227, 7, 70, "ComputationResult",
 CellID->1456581472]
}, Open  ]],
Cell[CellGroupData[{
Cell[130340, 3924, 377, 10, 70, "ComputationInfo",
 CellID->416142635],
Cell[130720, 3936, 647, 17, 70, "ComputationInfoBody",
 CellID->417178908]
}, Closed]],
Cell[CellGroupData[{
Cell[131404, 3958, 220, 6, 70, "Computation",
 CellID->1242138448],
Cell[131627, 3966, 234, 7, 70, "ComputationResult",
 CellID->545447630]
}, Open  ]],
Cell[CellGroupData[{
Cell[131898, 3978, 377, 10, 70, "ComputationInfo",
 CellID->850780133],
Cell[132278, 3990, 648, 17, 70, "ComputationInfoBody",
 CellID->1266603761]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[132975, 4013, 110, 2, 70, "Subsection",
 CellID->479188981],
Cell[CellGroupData[{
Cell[133110, 4019, 350, 8, 70, "Computation",
 CellID->111822425],
Cell[133463, 4029, 465, 11, 70, "ComputationResult",
 CellID->425746342]
}, Open  ]],
Cell[CellGroupData[{
Cell[133965, 4045, 377, 10, 70, "ComputationInfo",
 CellID->1022918257],
Cell[134345, 4057, 646, 17, 70, "ComputationInfoBody",
 CellID->105124728]
}, Closed]],
Cell[CellGroupData[{
Cell[135028, 4079, 261, 6, 70, "Computation",
 CellID->1993907386],
Cell[135292, 4087, 269, 9, 70, "ComputationResult",
 CellID->930951114]
}, Open  ]],
Cell[CellGroupData[{
Cell[135598, 4101, 380, 10, 70, "ComputationInfo",
 CellID->1993783205],
Cell[135981, 4113, 650, 17, 70, "ComputationInfoBody",
 CellID->1845365743]
}, Closed]],
Cell[CellGroupData[{
Cell[136668, 4135, 306, 8, 70, "Computation",
 CellID->1136958002],
Cell[136977, 4145, 249, 8, 70, "ComputationResult",
 CellID->526249381]
}, Open  ]],
Cell[CellGroupData[{
Cell[137263, 4158, 380, 10, 70, "ComputationInfo",
 CellID->1383108897],
Cell[137646, 4170, 648, 17, 70, "ComputationInfoBody",
 CellID->1634101449]
}, Closed]],
Cell[CellGroupData[{
Cell[138331, 4192, 294, 8, 70, "Computation",
 CellID->54162398],
Cell[138628, 4202, 228, 7, 70, "ComputationResult",
 CellID->32754321]
}, Open  ]],
Cell[CellGroupData[{
Cell[138893, 4214, 376, 10, 70, "ComputationInfo",
 CellID->1556773737],
Cell[139272, 4226, 646, 17, 70, "ComputationInfoBody",
 CellID->1285835433]
}, Closed]],
Cell[CellGroupData[{
Cell[139955, 4248, 272, 7, 70, "Computation",
 CellID->1435882704],
Cell[140230, 4257, 229, 7, 70, "ComputationResult",
 CellID->568066442]
}, Open  ]],
Cell[CellGroupData[{
Cell[140496, 4269, 378, 10, 70, "ComputationInfo",
 CellID->1903147392],
Cell[140877, 4281, 647, 17, 70, "ComputationInfoBody",
 CellID->548348087]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[141585, 4305, 334, 6, 43, "Section",
 CellID->1769514847],
Cell[141922, 4313, 711, 18, 33, "GlobalDeclaration",
 CellTags->{
  "ID:335127278", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->335127278],
Cell[CellGroupData[{
Cell[142658, 4335, 189, 3, 38, "Subsection",
 CellID->44748028],
Cell[CellGroupData[{
Cell[142872, 4342, 164, 4, 70, "Subsubsection",
 CellID->415794492],
Cell[143039, 4348, 48, 1, 70, "OpenEnvironment",
 CellID->1331628327],
Cell[CellGroupData[{
Cell[143112, 4353, 604, 13, 70, "EnvironmentHeader",
 CellID->1603429019],
Cell[143719, 4368, 912, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2044589078", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommRing1 \[ScriptCapitalZ]"},
 CellID->2044589078],
Cell[144634, 4393, 1617, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:48028479", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "integral domain \[ScriptCapitalZ]"},
 CellID->48028479],
Cell[146254, 4445, 66, 1, 70, "EndEnvironmentMarker",
 CellID->2061832140]
}, Open  ]],
Cell[146335, 4449, 49, 1, 70, "CloseEnvironment",
 CellID->1577743469],
Cell[CellGroupData[{
Cell[146409, 4454, 1851, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:2044589078", "Proof|ID:2044589078-1"},
 CellID->47164179],
Cell[148263, 4513, 480, 12, 70, "ProofInfoBody",
 CellID->94304473]
}, Closed]],
Cell[CellGroupData[{
Cell[148780, 4530, 4325, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:48028479", "Proof|ID:48028479-1"},
 CellID->584664],
Cell[153108, 4648, 479, 12, 70, "ProofInfoBody",
 CellID->563218315]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[153636, 4666, 194, 3, 70, "Subsubsection",
 CellID->26895694],
Cell[153833, 4671, 47, 1, 70, "OpenEnvironment",
 CellID->228008460],
Cell[CellGroupData[{
Cell[153905, 4676, 419, 11, 70, "EnvironmentHeader",
 CellID->3757693],
Cell[154327, 4689, 1210, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1813588854", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isPartIrreflOrder \[ScriptCapitalZ]"},
 CellID->1813588854],
Cell[155540, 4724, 1045, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:187039377", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isTotalIrreflOrder \[ScriptCapitalZ]"},
 CellID->187039377],
Cell[156588, 4757, 1067, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:159454607", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isNoetherian \[ScriptCapitalZ]"},
 CellID->159454607],
Cell[157658, 4790, 65, 1, 70, "EndEnvironmentMarker",
 CellID->270718515]
}, Open  ]],
Cell[157738, 4794, 47, 1, 70, "CloseEnvironment",
 CellID->38854093],
Cell[CellGroupData[{
Cell[157810, 4799, 2139, 65, 70, "ProofInfo",
 CellTags->{"Proof|ID:1813588854", "Proof|ID:1813588854-1"},
 CellID->80460269],
Cell[159952, 4866, 480, 12, 70, "ProofInfoBody",
 CellID->40342381]
}, Closed]],
Cell[CellGroupData[{
Cell[160469, 4883, 2135, 64, 70, "ProofInfo",
 CellTags->{"Proof|ID:187039377", "Proof|ID:187039377-1"},
 CellID->106498233],
Cell[162607, 4949, 480, 12, 70, "ProofInfoBody",
 CellID->257809711]
}, Closed]],
Cell[CellGroupData[{
Cell[163124, 4966, 2103, 63, 70, "ProofInfo",
 CellTags->{"Proof|ID:159454607", "Proof|ID:159454607-1"},
 CellID->187116588],
Cell[165230, 5031, 480, 12, 70, "ProofInfoBody",
 CellID->323553876]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[165759, 5049, 251, 4, 70, "Subsubsection",
 CellID->217483992],
Cell[166013, 5055, 47, 1, 70, "OpenEnvironment",
 CellID->368849717],
Cell[CellGroupData[{
Cell[166085, 5060, 464, 12, 70, "EnvironmentHeader",
 CellID->808264381],
Cell[166552, 5074, 1423, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:38031932", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "mult alternative \[ScriptCapitalZ]"},
 CellID->38031932],
Cell[167978, 5119, 842, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1890039364", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "Multipliers \[ScriptCapitalZ]"},
 CellID->1890039364],
Cell[168823, 5143, 810, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1152753107", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R0 \[ScriptCapitalZ]"},
 CellID->1152753107],
Cell[169636, 5166, 807, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:307366433", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R1 \[ScriptCapitalZ]"},
 CellID->307366433],
Cell[170446, 5189, 831, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:963452595", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R2 \[ScriptCapitalZ]"},
 CellID->963452595],
Cell[171280, 5212, 808, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1084156204", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R3 \[ScriptCapitalZ]"},
 CellID->1084156204],
Cell[172091, 5235, 807, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2041570274", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R4 \[ScriptCapitalZ]"},
 CellID->2041570274],
Cell[172901, 5258, 65, 1, 70, "EndEnvironmentMarker",
 CellID->382946846]
}, Open  ]],
Cell[172981, 5262, 47, 1, 70, "CloseEnvironment",
 CellID->79265178],
Cell[CellGroupData[{
Cell[173053, 5267, 3640, 101, 70, "ProofInfo",
 CellTags->{"Proof|ID:38031932", "Proof|ID:38031932-1"},
 CellID->279953348],
Cell[176696, 5370, 478, 12, 70, "ProofInfoBody",
 CellID->17290125]
}, Closed]],
Cell[CellGroupData[{
Cell[177211, 5387, 1852, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1890039364", "Proof|ID:1890039364-1"},
 CellID->845457494],
Cell[179066, 5446, 481, 12, 70, "ProofInfoBody",
 CellID->223964381]
}, Closed]],
Cell[CellGroupData[{
Cell[179584, 5463, 1816, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1152753107", "Proof|ID:1152753107-1"},
 CellID->229850786],
Cell[181403, 5522, 480, 12, 70, "ProofInfoBody",
 CellID->41398535]
}, Closed]],
Cell[CellGroupData[{
Cell[181920, 5539, 1806, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:307366433", "Proof|ID:307366433-1"},
 CellID->93056147],
Cell[183729, 5597, 479, 12, 70, "ProofInfoBody",
 CellID->31010785]
}, Closed]],
Cell[CellGroupData[{
Cell[184245, 5614, 1805, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:963452595", "Proof|ID:963452595-1"},
 CellID->8129192],
Cell[186053, 5672, 480, 12, 70, "ProofInfoBody",
 CellID->201562422]
}, Closed]],
Cell[CellGroupData[{
Cell[186570, 5689, 1816, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1084156204", "Proof|ID:1084156204-1"},
 CellID->477880951],
Cell[188389, 5748, 481, 12, 70, "ProofInfoBody",
 CellID->360511373]
}, Closed]],
Cell[CellGroupData[{
Cell[188907, 5765, 1816, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:2041570274", "Proof|ID:2041570274-1"},
 CellID->299709042],
Cell[190726, 5824, 481, 12, 70, "ProofInfoBody",
 CellID->377565118]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[191256, 5842, 213, 3, 70, "Subsubsection",
 CellID->11190904],
Cell[191472, 5847, 47, 1, 70, "OpenEnvironment",
 CellID->624303376],
Cell[CellGroupData[{
Cell[191544, 5852, 395, 9, 70, "EnvironmentHeader",
 CellID->473325153],
Cell[191942, 5863, 820, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:260068164", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->260068164],
Cell[192765, 5886, 882, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:211178897", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership sgn \[ScriptCapitalZ]"},
 CellID->211178897],
Cell[193650, 5911, 1887, 60, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:118405539", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[RawStar] sgn \[ScriptCapitalZ]"},
 CellID->118405539],
Cell[195540, 5973, 942, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:88914628", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "0 minimal \[ScriptCapitalZ]"},
 CellID->88914628],
Cell[196485, 6000, 1345, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:69480817", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "twice \[Succeeds] twice \[ScriptCapitalZ]"},
 CellID->69480817],
Cell[197833, 6042, 1883, 58, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:24170142", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[DoubleRightArrow] \[ScriptCapitalZ]"},
 CellID->24170142],
Cell[199719, 6102, 1885, 58, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:302704588", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[DoubleLeftArrow] \[ScriptCapitalZ]"},
 CellID->302704588],
Cell[201607, 6162, 1892, 57, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:131195953", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] + \[ScriptCapitalZ]"},
 CellID->131195953],
Cell[203502, 6221, 1723, 52, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:129207038", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] - \[ScriptCapitalZ]"},
 CellID->129207038],
Cell[205228, 6275, 1822, 53, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:59300728", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] alternative \[ScriptCapitalZ]"},
 CellID->59300728],
Cell[207053, 6330, 1722, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:698610132", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Succeeds] +' \[ScriptCapitalZ]"},
 CellID->698610132],
Cell[208778, 6379, 65, 1, 70, "EndEnvironmentMarker",
 CellID->577559729]
}, Open  ]],
Cell[208858, 6383, 47, 1, 70, "CloseEnvironment",
 CellID->17576423],
Cell[CellGroupData[{
Cell[208930, 6388, 2649, 77, 70, "ProofInfo",
 CellTags->{"Proof|ID:211178897", "Proof|ID:211178897-1"},
 CellID->273631166],
Cell[211582, 6467, 480, 12, 70, "ProofInfoBody",
 CellID->138417704]
}, Closed]],
Cell[CellGroupData[{
Cell[212099, 6484, 5656, 142, 70, "ProofInfo",
 CellTags->{"Proof|ID:118405539", "Proof|ID:118405539-1"},
 CellID->55870784],
Cell[217758, 6628, 479, 12, 70, "ProofInfoBody",
 CellID->10869643]
}, Closed]],
Cell[CellGroupData[{
Cell[218274, 6645, 3039, 87, 70, "ProofInfo",
 CellTags->{"Proof|ID:88914628", "Proof|ID:88914628-1"},
 CellID->298080950],
Cell[221316, 6734, 479, 12, 70, "ProofInfoBody",
 CellID->276114159]
}, Closed]],
Cell[CellGroupData[{
Cell[221832, 6751, 4303, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:69480817", "Proof|ID:69480817-1"},
 CellID->8097116],
Cell[226138, 6869, 479, 12, 70, "ProofInfoBody",
 CellID->109020929]
}, Closed]],
Cell[CellGroupData[{
Cell[226654, 6886, 6292, 160, 70, "ProofInfo",
 CellTags->{"Proof|ID:24170142", "Proof|ID:24170142-1"},
 CellID->297064343],
Cell[232949, 7048, 479, 12, 70, "ProofInfoBody",
 CellID->175955906]
}, Closed]],
Cell[CellGroupData[{
Cell[233465, 7065, 6308, 160, 70, "ProofInfo",
 CellTags->{"Proof|ID:302704588", "Proof|ID:302704588-1"},
 CellID->172202783],
Cell[239776, 7227, 479, 12, 70, "ProofInfoBody",
 CellID->54069213]
}, Closed]],
Cell[CellGroupData[{
Cell[240292, 7244, 6514, 165, 70, "ProofInfo",
 CellTags->{"Proof|ID:131195953", "Proof|ID:131195953-1"},
 CellID->88300092],
Cell[246809, 7411, 480, 12, 70, "ProofInfoBody",
 CellID->414777165]
}, Closed]],
Cell[CellGroupData[{
Cell[247326, 7428, 5698, 147, 70, "ProofInfo",
 CellTags->{"Proof|ID:129207038", "Proof|ID:129207038-1"},
 CellID->33212788],
Cell[253027, 7577, 480, 12, 70, "ProofInfoBody",
 CellID->378179640]
}, Closed]],
Cell[CellGroupData[{
Cell[253544, 7594, 5798, 146, 70, "ProofInfo",
 CellTags->{"Proof|ID:59300728", "Proof|ID:59300728-1"},
 CellID->946864890],
Cell[259345, 7742, 479, 12, 70, "ProofInfoBody",
 CellID->387614559]
}, Closed]],
Cell[CellGroupData[{
Cell[259861, 7759, 6175, 155, 70, "ProofInfo",
 CellTags->{"Proof|ID:698610132", "Proof|ID:698610132-1"},
 CellID->357893767],
Cell[266039, 7916, 480, 12, 70, "ProofInfoBody",
 CellID->550606355]
}, Closed]],
Cell[266534, 7931, 47, 1, 70, "OpenEnvironment",
 CellID->122261486],
Cell[CellGroupData[{
Cell[266606, 7936, 397, 9, 70, "EnvironmentHeader",
 CellID->279811452],
Cell[267006, 7947, 767, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:310863676", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->310863676],
Cell[267776, 7969, 1525, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:471044386", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible - \[ScriptCapitalZ]"},
 CellID->471044386],
Cell[269304, 8014, 2346, 71, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:379223163", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible alternative \[ScriptCapitalZ]"},
 CellID->379223163],
Cell[271653, 8087, 1625, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:176657795", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible sgn \[ScriptCapitalZ]"},
 CellID->176657795],
Cell[273281, 8139, 1988, 59, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:217289551", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", 
   "isReducible alternative' \[DoubleRightArrow] \[ScriptCapitalZ]"},
 CellID->217289551],
Cell[275272, 8200, 1988, 59, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:497414030", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", 
   "isReducible alternative' \[DoubleLeftArrow] \[ScriptCapitalZ]"},
 CellID->497414030],
Cell[277263, 8261, 2055, 59, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:464408938", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible alternative' \[ScriptCapitalZ]"},
 CellID->464408938],
Cell[279321, 8322, 1138, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:304230885", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible -' \[ScriptCapitalZ]"},
 CellID->304230885],
Cell[280462, 8354, 1465, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:97311931", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[RightArrow] - \[ScriptCapitalZ]"},
 CellID->97311931],
Cell[281930, 8400, 1820, 53, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:99245919", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds] \[ScriptCapitalZ]"},
 CellID->99245919],
Cell[283753, 8455, 1359, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:162605324", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds]' \[ScriptCapitalZ]"},
 CellID->162605324],
Cell[285115, 8494, 1338, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:40987937", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReducible \[Succeeds]'' \[ScriptCapitalZ]"},
 CellID->40987937],
Cell[286456, 8534, 64, 1, 70, "EndEnvironmentMarker",
 CellID->51573105]
}, Open  ]],
Cell[286535, 8538, 48, 1, 70, "CloseEnvironment",
 CellID->608032019],
Cell[CellGroupData[{
Cell[286608, 8543, 4977, 127, 70, "ProofInfo",
 CellTags->{"Proof|ID:471044386", "Proof|ID:471044386-1"},
 CellID->450891268],
Cell[291588, 8672, 480, 12, 70, "ProofInfoBody",
 CellID->551822482]
}, Closed]],
Cell[CellGroupData[{
Cell[292105, 8689, 7043, 174, 70, "ProofInfo",
 CellTags->{"Proof|ID:379223163", "Proof|ID:379223163-1"},
 CellID->249907335],
Cell[299151, 8865, 480, 12, 70, "ProofInfoBody",
 CellID->557242539]
}, Closed]],
Cell[CellGroupData[{
Cell[299668, 8882, 5329, 134, 70, "ProofInfo",
 CellTags->{"Proof|ID:176657795", "Proof|ID:176657795-1"},
 CellID->334897452],
Cell[305000, 9018, 480, 12, 70, "ProofInfoBody",
 CellID->205455470]
}, Closed]],
Cell[CellGroupData[{
Cell[305517, 9035, 6019, 154, 70, "ProofInfo",
 CellTags->{"Proof|ID:217289551", "Proof|ID:217289551-1"},
 CellID->880922821],
Cell[311539, 9191, 480, 12, 70, "ProofInfoBody",
 CellID->276481536]
}, Closed]],
Cell[CellGroupData[{
Cell[312056, 9208, 6041, 154, 70, "ProofInfo",
 CellTags->{"Proof|ID:497414030", "Proof|ID:497414030-1"},
 CellID->110073726],
Cell[318100, 9364, 479, 12, 70, "ProofInfoBody",
 CellID->23417434]
}, Closed]],
Cell[CellGroupData[{
Cell[318616, 9381, 6151, 156, 70, "ProofInfo",
 CellTags->{"Proof|ID:464408938", "Proof|ID:464408938-1"},
 CellID->577762604],
Cell[324770, 9539, 480, 12, 70, "ProofInfoBody",
 CellID->591630630]
}, Closed]],
Cell[CellGroupData[{
Cell[325287, 9556, 4079, 109, 70, "ProofInfo",
 CellTags->{"Proof|ID:304230885", "Proof|ID:304230885-1"},
 CellID->115875544],
Cell[329369, 9667, 479, 12, 70, "ProofInfoBody",
 CellID->93717885]
}, Closed]],
Cell[CellGroupData[{
Cell[329885, 9684, 4842, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:97311931", "Proof|ID:97311931-1"},
 CellID->616242356],
Cell[334730, 9812, 478, 12, 70, "ProofInfoBody",
 CellID->16956372]
}, Closed]],
Cell[CellGroupData[{
Cell[335245, 9829, 6270, 154, 70, "ProofInfo",
 CellTags->{"Proof|ID:99245919", "Proof|ID:99245919-1"},
 CellID->518473742],
Cell[341518, 9985, 478, 12, 70, "ProofInfoBody",
 CellID->20241181]
}, Closed]],
Cell[CellGroupData[{
Cell[342033, 10002, 4719, 122, 70, "ProofInfo",
 CellTags->{"Proof|ID:162605324", "Proof|ID:162605324-1"},
 CellID->291524205],
Cell[346755, 10126, 480, 12, 70, "ProofInfoBody",
 CellID->569496899]
}, Closed]],
Cell[CellGroupData[{
Cell[347272, 10143, 5110, 130, 70, "ProofInfo",
 CellTags->{"Proof|ID:40987937", "Proof|ID:40987937-1"},
 CellID->103343062],
Cell[352385, 10275, 477, 12, 70, "ProofInfoBody",
 CellID->9396178]
}, Closed]],
Cell[352877, 10290, 46, 1, 70, "OpenEnvironment",
 CellID->31873811],
Cell[CellGroupData[{
Cell[352948, 10295, 476, 11, 70, "EnvironmentHeader",
 CellID->687299096],
Cell[353427, 10308, 807, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:11694492", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->11694492],
Cell[354237, 10330, 949, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:224215037", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership lr \[ScriptCapitalZ]"},
 CellID->224215037],
Cell[355189, 10358, 1170, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:78177024", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr = 0 \[ScriptCapitalZ]"},
 CellID->78177024],
Cell[356362, 10396, 1427, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:634190310", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr > 0 \[ScriptCapitalZ]"},
 CellID->634190310],
Cell[357792, 10441, 1186, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:588999973", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "0 > lr \[ScriptCapitalZ]"},
 CellID->588999973],
Cell[358981, 10479, 1190, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:57477705", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr twice even \[ScriptCapitalZ]"},
 CellID->57477705],
Cell[360174, 10515, 1393, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:206701743", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr twice odd \[ScriptCapitalZ]"},
 CellID->206701743],
Cell[361570, 10561, 1143, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:34042360", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isReducible \[ScriptCapitalZ]"},
 CellID->34042360],
Cell[362716, 10593, 1514, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:692837602", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr minimal isReducible \[ScriptCapitalZ]"},
 CellID->692837602],
Cell[364233, 10641, 1715, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:44304045", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isReducible sgn \[ScriptCapitalZ]"},
 CellID->44304045],
Cell[365951, 10693, 65, 1, 70, "EndEnvironmentMarker",
 CellID->403862672]
}, Open  ]],
Cell[366031, 10697, 48, 1, 70, "CloseEnvironment",
 CellID->230715858],
Cell[CellGroupData[{
Cell[366104, 10702, 2817, 81, 70, "ProofInfo",
 CellTags->{"Proof|ID:224215037", "Proof|ID:224215037-1"},
 CellID->519188695],
Cell[368924, 10785, 480, 12, 70, "ProofInfoBody",
 CellID->271118078]
}, Closed]],
Cell[CellGroupData[{
Cell[369441, 10802, 2934, 84, 70, "ProofInfo",
 CellTags->{"Proof|ID:78177024", "Proof|ID:78177024-1"},
 CellID->336735725],
Cell[372378, 10888, 479, 12, 70, "ProofInfoBody",
 CellID->326983128]
}, Closed]],
Cell[CellGroupData[{
Cell[372894, 10905, 3276, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:634190310", "Proof|ID:634190310-1"},
 CellID->77986082],
Cell[376173, 10999, 480, 12, 70, "ProofInfoBody",
 CellID->508208461]
}, Closed]],
Cell[CellGroupData[{
Cell[376690, 11016, 3262, 94, 70, "ProofInfo",
 CellTags->{"Proof|ID:588999973", "Proof|ID:588999973-1"},
 CellID->1365258],
Cell[379955, 11112, 480, 12, 70, "ProofInfoBody",
 CellID->429365998]
}, Closed]],
Cell[CellGroupData[{
Cell[380472, 11129, 3623, 99, 70, "ProofInfo",
 CellTags->{"Proof|ID:57477705", "Proof|ID:57477705-1"},
 CellID->253970289],
Cell[384098, 11230, 477, 12, 70, "ProofInfoBody",
 CellID->1669627]
}, Closed]],
Cell[CellGroupData[{
Cell[384612, 11247, 4469, 121, 70, "ProofInfo",
 CellTags->{"Proof|ID:206701743", "Proof|ID:206701743-1"},
 CellID->43456446],
Cell[389084, 11370, 479, 12, 70, "ProofInfoBody",
 CellID->37592898]
}, Closed]],
Cell[CellGroupData[{
Cell[389600, 11387, 3454, 95, 70, "ProofInfo",
 CellTags->{"Proof|ID:34042360", "Proof|ID:34042360-1"},
 CellID->169464375],
Cell[393057, 11484, 479, 12, 70, "ProofInfoBody",
 CellID->325628575]
}, Closed]],
Cell[CellGroupData[{
Cell[393573, 11501, 4318, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:692837602", "Proof|ID:692837602-1"},
 CellID->371670326],
Cell[397894, 11619, 480, 12, 70, "ProofInfoBody",
 CellID->251346555]
}, Closed]],
Cell[CellGroupData[{
Cell[398411, 11636, 5404, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:44304045", "Proof|ID:44304045-1"},
 CellID->477556229],
Cell[403818, 11776, 479, 12, 70, "ProofInfoBody",
 CellID->413433274]
}, Closed]],
Cell[404312, 11791, 47, 1, 70, "OpenEnvironment",
 CellID->580411371],
Cell[CellGroupData[{
Cell[404384, 11796, 511, 12, 70, "EnvironmentHeader",
 CellID->285853213],
Cell[404898, 11810, 892, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:16238117", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->16238117],
Cell[405793, 11835, 1455, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:468309069", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[Element] multN \[ScriptCapitalZ]"},
 CellID->468309069],
Cell[407251, 11883, 798, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:560244434", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->560244434],
Cell[408052, 11906, 1129, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:142903111", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "areIrrelative alternative \[ScriptCapitalZ]"},
 CellID->142903111],
Cell[409184, 11940, 1671, 48, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:36948313", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommonReducible alternative \[ScriptCapitalZ]"},
 CellID->36948313],
Cell[410858, 11990, 2281, 62, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:46630982", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isCommonReducible alternative' \[ScriptCapitalZ]"},
 CellID->46630982],
Cell[413142, 12054, 64, 1, 70, "EndEnvironmentMarker",
 CellID->96563173]
}, Open  ]],
Cell[413221, 12058, 46, 1, 70, "CloseEnvironment",
 CellID->1163307],
Cell[CellGroupData[{
Cell[413292, 12063, 3517, 95, 70, "ProofInfo",
 CellTags->{"Proof|ID:468309069", "Proof|ID:468309069-1"},
 CellID->223234557],
Cell[416812, 12160, 478, 12, 70, "ProofInfoBody",
 CellID->2459748]
}, Closed]],
Cell[CellGroupData[{
Cell[417327, 12177, 6252, 156, 70, "ProofInfo",
 CellTags->{"Proof|ID:142903111", "Proof|ID:142903111-1"},
 CellID->4630108],
Cell[423582, 12335, 480, 12, 70, "ProofInfoBody",
 CellID->182063627]
}, Closed]],
Cell[CellGroupData[{
Cell[424099, 12352, 7090, 172, 70, "ProofInfo",
 CellTags->{"Proof|ID:36948313", "Proof|ID:36948313-1"},
 CellID->239177438],
Cell[431192, 12526, 478, 12, 70, "ProofInfoBody",
 CellID->14134654]
}, Closed]],
Cell[CellGroupData[{
Cell[431707, 12543, 8948, 210, 70, "ProofInfo",
 CellTags->{"Proof|ID:46630982", "Proof|ID:46630982-1"},
 CellID->250839951],
Cell[440658, 12755, 478, 12, 70, "ProofInfoBody",
 CellID->35813888]
}, Closed]],
Cell[441151, 12770, 47, 1, 70, "OpenEnvironment",
 CellID->113752866],
Cell[CellGroupData[{
Cell[441223, 12775, 398, 9, 70, "EnvironmentHeader",
 CellID->614730048],
Cell[441624, 12786, 1196, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:69669781", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isMNTCR unary \[ScriptCapitalZ]"},
 CellID->69669781],
Cell[442823, 12822, 993, 27, 70, "GlobalDeclaration",
 CellTags->{
  "ID:65735690", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->65735690],
Cell[443819, 12851, 774, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:431845789", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->431845789],
Cell[444596, 12874, 1242, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:96649542", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isNTCR \[ScriptCapitalZ]"},
 CellID->96649542],
Cell[445841, 12909, 1400, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:132861224", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "lr isMNTCR \[ScriptCapitalZ]"},
 CellID->132861224],
Cell[447244, 12946, 64, 1, 70, "EndEnvironmentMarker",
 CellID->65813332]
}, Open  ]],
Cell[447323, 12950, 47, 1, 70, "CloseEnvironment",
 CellID->76710532],
Cell[CellGroupData[{
Cell[447395, 12955, 3449, 95, 70, "ProofInfo",
 CellTags->{"Proof|ID:69669781", "Proof|ID:69669781-1"},
 CellID->22101709],
Cell[450847, 13052, 479, 12, 70, "ProofInfoBody",
 CellID->279677689]
}, Closed]],
Cell[CellGroupData[{
Cell[451363, 13069, 7422, 178, 70, "ProofInfo",
 CellTags->{"Proof|ID:96649542", "Proof|ID:96649542-1"},
 CellID->151366490],
Cell[458788, 13249, 479, 12, 70, "ProofInfoBody",
 CellID->624894588]
}, Closed]],
Cell[CellGroupData[{
Cell[459304, 13266, 7438, 179, 70, "ProofInfo",
 CellTags->{"Proof|ID:132861224", "Proof|ID:132861224-1"},
 CellID->94818643],
Cell[466745, 13447, 480, 12, 70, "ProofInfoBody",
 CellID->573039597]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[467274, 13465, 233, 4, 70, "Subsubsection",
 CellID->703472201],
Cell[467510, 13471, 47, 1, 70, "OpenEnvironment",
 CellID->425914962],
Cell[CellGroupData[{
Cell[467582, 13476, 560, 11, 70, "EnvironmentHeader",
 CellID->258230160],
Cell[468145, 13489, 805, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:595906550", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R5 \[ScriptCapitalZ]"},
 CellID->595906550],
Cell[468953, 13512, 807, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:384714041", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R6 \[ScriptCapitalZ]"},
 CellID->384714041],
Cell[469763, 13535, 807, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:739712574", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R7 \[ScriptCapitalZ]"},
 CellID->739712574],
Cell[470573, 13558, 809, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:782958792", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R8 \[ScriptCapitalZ]"},
 CellID->782958792],
Cell[471385, 13581, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1686190119]
}, Open  ]],
Cell[471466, 13585, 49, 1, 70, "CloseEnvironment",
 CellID->1985203424],
Cell[CellGroupData[{
Cell[471540, 13590, 1807, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:595906550", "Proof|ID:595906550-1"},
 CellID->801014817],
Cell[473350, 13648, 480, 12, 70, "ProofInfoBody",
 CellID->655992473]
}, Closed]],
Cell[CellGroupData[{
Cell[473867, 13665, 1805, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:384714041", "Proof|ID:384714041-1"},
 CellID->4523962],
Cell[475675, 13723, 480, 12, 70, "ProofInfoBody",
 CellID->640491936]
}, Closed]],
Cell[CellGroupData[{
Cell[476192, 13740, 1807, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:739712574", "Proof|ID:739712574-1"},
 CellID->802064395],
Cell[478002, 13798, 480, 12, 70, "ProofInfoBody",
 CellID->218110605]
}, Closed]],
Cell[CellGroupData[{
Cell[478519, 13815, 1807, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:782958792", "Proof|ID:782958792-1"},
 CellID->148939037],
Cell[480329, 13873, 479, 12, 70, "ProofInfoBody",
 CellID->89453536]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[480857, 13891, 190, 3, 70, "Subsubsection",
 CellID->461632795],
Cell[481050, 13896, 47, 1, 70, "OpenEnvironment",
 CellID->288863783],
Cell[CellGroupData[{
Cell[481122, 13901, 517, 11, 70, "EnvironmentHeader",
 CellID->1803439939],
Cell[481642, 13914, 848, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:67773804", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isReductionRing \[ScriptCapitalZ]"},
 CellID->67773804],
Cell[482493, 13938, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1852266050]
}, Open  ]],
Cell[482574, 13942, 48, 1, 70, "CloseEnvironment",
 CellID->346871786],
Cell[CellGroupData[{
Cell[482647, 13947, 1870, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:67773804", "Proof|ID:67773804-1"},
 CellID->766315013],
Cell[484520, 14007, 478, 12, 70, "ProofInfoBody",
 CellID->91080121]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[485059, 14026, 202, 3, 30, "Subsection",
 CellID->811185245],
Cell[CellGroupData[{
Cell[485286, 14033, 137, 2, 28, "Subsubsection",
 CellID->91867520],
Cell[485426, 14037, 47, 1, 70, "OpenEnvironment",
 CellID->209634667],
Cell[CellGroupData[{
Cell[485498, 14042, 445, 10, 70, "EnvironmentHeader",
 CellID->1132188],
Cell[485946, 14054, 809, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:156995908", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R9 \[ScriptCapitalZ]"},
 CellID->156995908],
Cell[486758, 14077, 813, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1051911108", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R10 \[ScriptCapitalZ]"},
 CellID->1051911108],
Cell[487574, 14100, 65, 1, 70, "EndEnvironmentMarker",
 CellID->474076086]
}, Open  ]],
Cell[487654, 14104, 48, 1, 70, "CloseEnvironment",
 CellID->377455191],
Cell[CellGroupData[{
Cell[487727, 14109, 1807, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:156995908", "Proof|ID:156995908-1"},
 CellID->802545094],
Cell[489537, 14167, 480, 12, 70, "ProofInfoBody",
 CellID->235298705]
}, Closed]],
Cell[CellGroupData[{
Cell[490054, 14184, 1819, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1051911108", "Proof|ID:1051911108-1"},
 CellID->67102701],
Cell[491876, 14243, 481, 12, 70, "ProofInfoBody",
 CellID->284680562]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[492406, 14261, 237, 4, 22, "Subsubsection",
 CellID->47141123],
Cell[492646, 14267, 43, 1, 70, "OpenEnvironment",
 CellID->99683],
Cell[CellGroupData[{
Cell[492714, 14272, 492, 11, 70, "EnvironmentHeader",
 CellID->525282118],
Cell[493209, 14285, 695, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:160940549", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->160940549],
Cell[493907, 14305, 794, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:46650045", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->46650045],
Cell[494704, 14327, 1058, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:816204582", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos \[GreaterEqual] 0 \[ScriptCapitalZ]"},
 CellID->816204582],
Cell[495765, 14359, 1107, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:171042210", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg \[GreaterEqual] 0 \[ScriptCapitalZ]"},
 CellID->171042210],
Cell[496875, 14393, 1496, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:160702195", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos 0 \[ScriptCapitalZ]"},
 CellID->160702195],
Cell[498374, 14442, 1673, 52, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:478744619", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg 0 \[ScriptCapitalZ]"},
 CellID->478744619],
Cell[500050, 14496, 1435, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:680200521", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm pos isReducible \[ScriptCapitalZ]"},
 CellID->680200521],
Cell[501488, 14542, 1606, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:122969425", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm neg isReducible \[ScriptCapitalZ]"},
 CellID->122969425],
Cell[503097, 14594, 914, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:509693672", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership rdm' \[ScriptCapitalZ]"},
 CellID->509693672],
Cell[504014, 14620, 1400, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:576030870", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm = 0' \[ScriptCapitalZ]"},
 CellID->576030870],
Cell[505417, 14662, 1124, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:825630046", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm isReducible' \[ScriptCapitalZ]"},
 CellID->825630046],
Cell[506544, 14694, 64, 1, 70, "EndEnvironmentMarker",
 CellID->22057905]
}, Open  ]],
Cell[506623, 14698, 48, 1, 70, "CloseEnvironment",
 CellID->607571403],
Cell[CellGroupData[{
Cell[506696, 14703, 3562, 102, 70, "ProofInfo",
 CellTags->{"Proof|ID:816204582", "Proof|ID:816204582-1"},
 CellID->451910702],
Cell[510261, 14807, 480, 12, 70, "ProofInfoBody",
 CellID->421911099]
}, Closed]],
Cell[CellGroupData[{
Cell[510778, 14824, 3971, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:171042210", "Proof|ID:171042210-1"},
 CellID->29599316],
Cell[514752, 14936, 480, 12, 70, "ProofInfoBody",
 CellID->364315036]
}, Closed]],
Cell[CellGroupData[{
Cell[515269, 14953, 4103, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:160702195", "Proof|ID:160702195-1"},
 CellID->461788959],
Cell[519375, 15069, 545, 13, 70, "ProofInfoBody",
 CellID->700076]
}, Closed]],
Cell[CellGroupData[{
Cell[519957, 15087, 4958, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:478744619", "Proof|ID:478744619-1"},
 CellID->1001675670],
Cell[524918, 15221, 479, 12, 70, "ProofInfoBody",
 CellID->75688501]
}, Closed]],
Cell[CellGroupData[{
Cell[525434, 15238, 4996, 134, 70, "ProofInfo",
 CellTags->{"Proof|ID:680200521", "Proof|ID:680200521-1"},
 CellID->98328192],
Cell[530433, 15374, 550, 13, 70, "ProofInfoBody",
 CellID->484485766]
}, Closed]],
Cell[CellGroupData[{
Cell[531020, 15392, 6137, 160, 70, "ProofInfo",
 CellTags->{"Proof|ID:122969425", "Proof|ID:122969425-1"},
 CellID->526144461],
Cell[537160, 15554, 480, 12, 70, "ProofInfoBody",
 CellID->296102259]
}, Closed]],
Cell[CellGroupData[{
Cell[537677, 15571, 3040, 88, 70, "ProofInfo",
 CellTags->{"Proof|ID:509693672", "Proof|ID:509693672-1"},
 CellID->220546910],
Cell[540720, 15661, 480, 12, 70, "ProofInfoBody",
 CellID->112818037]
}, Closed]],
Cell[CellGroupData[{
Cell[541237, 15678, 4008, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:576030870", "Proof|ID:576030870-1"},
 CellID->586909032],
Cell[545248, 15790, 480, 12, 70, "ProofInfoBody",
 CellID->222576414]
}, Closed]],
Cell[CellGroupData[{
Cell[545765, 15807, 4100, 109, 70, "ProofInfo",
 CellTags->{"Proof|ID:825630046", "Proof|ID:825630046-1"},
 CellID->36300196],
Cell[549868, 15918, 480, 12, 70, "ProofInfoBody",
 CellID->319066684]
}, Closed]],
Cell[550363, 15933, 47, 1, 70, "OpenEnvironment",
 CellID->472132368],
Cell[CellGroupData[{
Cell[550435, 15938, 376, 9, 70, "EnvironmentHeader",
 CellID->75812443],
Cell[550814, 15949, 753, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:47390677", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb"},
 CellID->47390677],
Cell[551570, 15971, 950, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:362000401", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "membership rdm \[ScriptCapitalZ]"},
 CellID->362000401],
Cell[552523, 15998, 1110, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:98296722", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm isReducible \[ScriptCapitalZ]"},
 CellID->98296722],
Cell[553636, 16029, 1271, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:357219590", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "rdm = 0 \[ScriptCapitalZ]"},
 CellID->357219590],
Cell[554910, 16070, 65, 1, 70, "EndEnvironmentMarker",
 CellID->134173173]
}, Open  ]],
Cell[554990, 16074, 48, 1, 70, "CloseEnvironment",
 CellID->130463965],
Cell[CellGroupData[{
Cell[555063, 16079, 3374, 93, 70, "ProofInfo",
 CellTags->{"Proof|ID:362000401", "Proof|ID:362000401-1"},
 CellID->71841439],
Cell[558440, 16174, 479, 12, 70, "ProofInfoBody",
 CellID->11485083]
}, Closed]],
Cell[CellGroupData[{
Cell[558956, 16191, 4330, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:98296722", "Proof|ID:98296722-1"},
 CellID->414103899],
Cell[563289, 16307, 549, 13, 70, "ProofInfoBody",
 CellID->375149138]
}, Closed]],
Cell[CellGroupData[{
Cell[563875, 16325, 4370, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:357219590", "Proof|ID:357219590-1"},
 CellID->55429374],
Cell[568248, 16443, 480, 12, 70, "ProofInfoBody",
 CellID->342077788]
}, Closed]],
Cell[568743, 16458, 45, 1, 70, "OpenEnvironment",
 CellID->4706945],
Cell[CellGroupData[{
Cell[568813, 16463, 451, 10, 70, "EnvironmentHeader",
 CellID->55456765],
Cell[569267, 16475, 811, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1086665202", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R11 \[ScriptCapitalZ]"},
 CellID->1086665202],
Cell[570081, 16498, 811, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:963670239", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R12 \[ScriptCapitalZ]"},
 CellID->963670239],
Cell[570895, 16521, 65, 1, 70, "EndEnvironmentMarker",
 CellID->275608555]
}, Open  ]],
Cell[570975, 16525, 48, 1, 70, "CloseEnvironment",
 CellID->202410225],
Cell[CellGroupData[{
Cell[571048, 16530, 1820, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1086665202", "Proof|ID:1086665202-1"},
 CellID->726182844],
Cell[572871, 16589, 480, 12, 70, "ProofInfoBody",
 CellID->66930221]
}, Closed]],
Cell[CellGroupData[{
Cell[573388, 16606, 1810, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:963670239", "Proof|ID:963670239-1"},
 CellID->70242044],
Cell[575201, 16664, 480, 12, 70, "ProofInfoBody",
 CellID->166946871]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[575730, 16682, 235, 4, 22, "Subsubsection",
 CellID->702407332],
Cell[575968, 16688, 46, 1, 70, "OpenEnvironment",
 CellID->23716938],
Cell[CellGroupData[{
Cell[576039, 16693, 419, 10, 70, "EnvironmentHeader",
 CellID->117869787],
Cell[576461, 16705, 1970, 60, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:483352095", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "\[SquareSubset] \[ScriptCapitalZ]"},
 CellID->483352095],
Cell[578434, 16767, 841, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:466277585", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "R13 \[ScriptCapitalZ]"},
 CellID->466277585],
Cell[579278, 16790, 65, 1, 70, "EndEnvironmentMarker",
 CellID->160356015]
}, Open  ]],
Cell[579358, 16794, 48, 1, 70, "CloseEnvironment",
 CellID->138398448],
Cell[CellGroupData[{
Cell[579431, 16799, 5793, 148, 70, "ProofInfo",
 CellTags->{"Proof|ID:483352095", "Proof|ID:483352095-1"},
 CellID->579238228],
Cell[585227, 16949, 479, 12, 70, "ProofInfoBody",
 CellID->25986173]
}, Closed]],
Cell[CellGroupData[{
Cell[585743, 16966, 1810, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:466277585", "Proof|ID:466277585-1"},
 CellID->54970627],
Cell[587556, 17024, 480, 12, 70, "ProofInfoBody",
 CellID->293687751]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[588085, 17042, 192, 3, 22, "Subsubsection",
 CellID->283529052],
Cell[588280, 17047, 46, 1, 12, "OpenEnvironment",
 CellID->16240552],
Cell[CellGroupData[{
Cell[588351, 17052, 421, 11, 43, "EnvironmentHeader",
 CellID->134430315],
Cell[588775, 17065, 845, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:25825837", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers.nb", "isAlgoReductionRing \[ScriptCapitalZ]"},
 CellID->25825837],
Cell[589623, 17089, 65, 1, 26, "EndEnvironmentMarker",
 CellID->712858629]
}, Open  ]],
Cell[589703, 17093, 47, 1, 16, "CloseEnvironment",
 CellID->53324427],
Cell[CellGroupData[{
Cell[589775, 17098, 1885, 58, 28, "ProofInfo",
 CellTags->{"Proof|ID:25825837", "Proof|ID:25825837-1"},
 CellID->84901906],
Cell[591663, 17158, 479, 12, 70, "ProofInfoBody",
 CellID->749931683]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
