(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[   4178193,     115997]
NotebookOptionsPosition[   3362361,      93640]
NotebookOutlinePosition[   3913616,     108200]
CellTagsIndexPosition[   3851160,     106611]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Reduction Polynomials", "Title",
 CellChangeTimes->{{3.643445838445785*^9, 3.64344584240105*^9}},
 CellID->1766055249],

Cell[CellGroupData[{

Cell["Preliminaries", "Section",
 CellChangeTimes->{{3.6434458667234993`*^9, 3.64344586970465*^9}},
 CellID->775912285],

Cell[CellGroupData[{

Cell["Description", "Subsection",
 CellChangeTimes->{{3.6404170077899017`*^9, 3.6404170090888367`*^9}, {
  3.660737814755328*^9, 3.6607378168624487`*^9}},
 CellID->600814793],

Cell[TextData[{
 "This notebook introduces the abstract classes of ",
 StyleBox["reduction polynomial domains",
  FontSlant->"Italic"],
 " and ",
 StyleBox["algorithmic reduction polynomial",
  FontSlant->"Italic"],
 " domains over arbitrary coefficient domain \[ScriptCapitalR] and \
power-product domain \[ScriptCapitalT]. Roughly, a domain \[ScriptCapitalP] \
in the class of reduction polynomials over \[ScriptCapitalR] and \
\[ScriptCapitalT] needs to define some functions (most importantly ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ", the coefficient function) that satisfy certain properties (e.g. ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " must respect addition). Algorithmic reduction polynomials, in addition, \
require \[ScriptCapitalP] to provide additional functions with a clearly \
defined behaviour that are needed to prove that algorithmic reduction \
polynomials form algorithmic reduction rings."
}], "Text",
 CellChangeTimes->{{3.660737821590719*^9, 3.6607381585199904`*^9}, {
  3.660738194566052*^9, 3.660738247452077*^9}, {3.660738992446688*^9, 
  3.66073899316873*^9}},
 CellID->510256476],

Cell["\<\
This, indeed, constitutes the main part of this notebook: Assuming that \
\[ScriptCapitalR] (i.e. the coefficient domain) is a (algorithmic) reduction \
ring, and that \[ScriptCapitalT] is a power-product domain, prove that \
(algorithmic) reduction polynomials over \[ScriptCapitalR] and \
\[ScriptCapitalT] are (algorithmic) reduction rings as well (reduction rings \
and algorithmic reduction rings are defined in \
\[OpenCurlyDoubleQuote]ReductionRings.nb\[CloseCurlyDoubleQuote]). This fact, \
known as \[OpenCurlyDoubleQuote]conservation theorem for polynomial rings\
\[CloseCurlyDoubleQuote] in reduction ring theory, is one of the central, \
most important results of the theory, as it allows us to build towers of \
reduction rings, starting from a given one and then repeatedly forming \
polynomial rings.\
\>", "Text",
 CellChangeTimes->{{3.660738252085342*^9, 3.660738440407113*^9}, {
  3.6607392077730045`*^9, 3.660739238737775*^9}},
 CellID->98348958],

Cell["\<\
It is important to note that everything in this notebook is on a rather high \
level of abstraction: no appeal is made to concrete domains or functors, but \
all results are stated and proved for generic domains belonging to the class \
of (algorithmic) reduction polynomials, i.e. no particular structure or \
representation of polynomials is presumed throughout the document. In order \
to prove that a concrete polynomial domain is a (algorithmic) reduction ring, \
it is then sufficient to prove that it is a (algorithmic) reduction \
polynomial domain \[Dash] and this is in general much easier.
An example of a concrete domain that was proved to be an algorithmic \
reduction polynomial domain can be found in \
\[OpenCurlyDoubleQuote]PolyTuples.nb\[CloseCurlyDoubleQuote]: there, \
polynomials are represented as ordered tuples of monomials, and monomials in \
turn are represented as pairs of coefficients and power-products.\
\>", "Text",
 CellChangeTimes->{{3.6607384651675296`*^9, 3.660738490264965*^9}, {
  3.6607385295132103`*^9, 3.6607389142162137`*^9}},
 CellID->173799993],

Cell["\<\
The structure of this notebook is as follows: Section \
\[OpenCurlyDoubleQuote]Reduction Polynomials\[CloseCurlyDoubleQuote] \
introduces most of the auxiliary notions needed, alongside the classes of \
reduction polynomial domains and algorithmic reduction polynomial domains. \
Then, in the three subsequent sections, elements of these classes are proved \
to constitute commutative rings with unit, reduction rings, and algorithmic \
reduction rings. Please note that the proofs are carried out under the \
weakest possible assumptions on the coefficient domain \[ScriptCapitalR], \
e.g. in the proof that reduction polynomials form commutative rings with \
unit, \[ScriptCapitalR] is only assumed to be a commutative ring with unit, \
but not necessarily a reduction ring.\
\>", "Text",
 CellChangeTimes->{{3.6607389223956814`*^9, 3.660739138405037*^9}},
 CellID->17970129]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Packages", "Subsection",
 CellChangeTimes->{{3.655465669242052*^9, 3.6554656725336576`*^9}},
 CellID->8118277],

Cell[BoxData[
 RowBox[{"$PackageDirectory", "=", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{
    "$RootDirectory", ",", "\"\<scratch\>\"", ",", "\"\<amaletzk\>\"", ",", 
     "\"\<Theorema Environment Data\>\""}], "}"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.666438783233775*^9, 3.666438821083324*^9}},
 CellID->498486718],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Tuples.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.656311306991623*^9, 3.6563113257896557`*^9}, 
   3.666440303641692*^9},
 CellID->59840707],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<ReductionRings.m\>\""}], "}"}], 
   "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.656311306991623*^9, 3.6563113345724716`*^9}, 
   3.6664403048659163`*^9},
 CellID->428817074],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<NatInt.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.656311306991623*^9, 3.6563113345724716`*^9}, {
   3.6563142375303874`*^9, 3.6563142392463903`*^9}, 3.666440306058515*^9},
 CellID->513028701],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<AlgebraicStructures.m\>\""}], 
    "}"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.656311306991623*^9, 3.6563113345724716`*^9}, {
   3.6563142375303874`*^9, 3.6563142392463903`*^9}, {3.656762149864876*^9, 
   3.656762170410112*^9}, 3.6664403071689157`*^9},
 CellID->239747824],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.655465678165268*^9, 3.655465684140078*^9}},
 CellID->590188203],

Cell[BoxData[
 RowBox[{"MakeTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6554656902240887`*^9, 3.6554656937964945`*^9}},
 CellID->414054049],

Cell[BoxData[
 RowBox[{"UpdateTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.656322374898106*^9, 3.6563223781273117`*^9}},
 CellID->74859277],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DeleteCases", "[", 
  RowBox[{
   RowBox[{"Map", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Quiet", "[", 
       RowBox[{"Check", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"With", "[", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{"s", "=", 
              RowBox[{"\"\<VAR$\>\"", "<>", "#", "<>", "\"\<$TM\>\""}]}], 
             "}"}], ",", 
            RowBox[{"Context", "[", "s", "]"}]}], "]"}], ";", "#"}], ",", 
         "Null"}], "]"}], "]"}], "&"}], ",", 
     RowBox[{"{", 
      RowBox[{
      "\"\<i\>\"", ",", "\"\<j\>\"", ",", "\"\<n\>\"", ",", "\"\<m\>\"", ",", 
       "\"\<P\>\"", ",", "\"\<x\>\"", ",", "\"\<y\>\"", ",", "\"\<z\>\"", 
       ",", "\"\<a\>\"", ",", "\"\<A\>\"", ",", "\"\<b\>\"", ",", "\"\<c\>\"",
        ",", "\"\<\[ScriptCapitalR]\>\"", ",", "\"\<\[ScriptCapitalP]\>\"", 
       ",", "\"\<R\>\""}], "}"}]}], "]"}], ",", "Null"}], "]"}]], "Input",
 CellChangeTimes->{{3.6592618519471025`*^9, 3.6592618631011224`*^9}, {
   3.659681175379635*^9, 3.659681185706853*^9}, 3.659683710620489*^9, 
   3.65968441725893*^9, 3.6597695015379353`*^9, {3.6597698034056654`*^9, 
   3.659769808569275*^9}, {3.6597702854721127`*^9, 3.659770290573322*^9}, {
   3.659848272310069*^9, 3.6598483467534*^9}, {3.6598515848824887`*^9, 
   3.6598516763142495`*^9}, {3.6598555136659927`*^9, 3.659855522090008*^9}, {
   3.6598559051642804`*^9, 3.659855919516306*^9}, {3.6602862205309095`*^9, 
   3.6602862222781124`*^9}, {3.660479643354128*^9, 3.660479646380533*^9}, {
   3.662369436502635*^9, 3.66236944483305*^9}, {3.662374360006923*^9, 
   3.662374361005325*^9}},
 CellID->20666029],

Cell[BoxData[
 RowBox[{"{", "}"}]], "Output",
 CellChangeTimes->{
  3.662369446205852*^9, 3.662373741653037*^9, {3.662374351926109*^9, 
   3.6623743620505266`*^9}, 3.6661764554094505`*^9},
 CellID->858659]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Analysis", "Subsection",
 CellChangeTimes->{{3.640668830437483*^9, 3.640668834318727*^9}},
 CellID->1386333768],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"AnalyzeProofs", "[", "]"}], "//", "Timing"}]], "Input",
 CellChangeTimes->{{3.653216253855052*^9, 3.65321625816066*^9}, {
  3.6557012226621327`*^9, 3.6557012242845354`*^9}},
 CellID->223251601],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Polynomials\\\" (\"\>", 
   "\[InvisibleSpace]", "1", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"1\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Polynomials\" (", 1, " of ", "1", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249631255389*^9},
 CellID->457132276],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"344\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "344"],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496312709894`*^9},
 CellID->230806296],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"344\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "344"],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496360445976`*^9},
 CellID->694688721],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"344\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "344"],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249641255007*^9},
 CellID->420126785],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"344\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "344"],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496463718157`*^9},
 CellID->198215461],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"344\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "344"],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249651785025*^9},
 CellID->95392771],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"344\"\>"}],
  SequenceForm["Reading proof ", 251, " of ", "344"],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249657744236*^9},
 CellID->761980848],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"344\"\>"}],
  SequenceForm["Reading proof ", 301, " of ", "344"],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496626270447`*^9},
 CellID->3204933],

Cell[BoxData["\<\"Creating forward rules:\"\>"], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496673070526`*^9},
 CellID->100688736],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496673382525`*^9},
 CellID->301435823],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 1, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249667478653*^9},
 CellID->89112948],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 51, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496678998537`*^9},
 CellID->93410960],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 101, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496683366547`*^9},
 CellID->908547730],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 151, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496687578554`*^9},
 CellID->825838993],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 201, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249669194656*^9},
 CellID->670125391],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 251, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249669631457*^9},
 CellID->732909961],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 301, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249670052657*^9},
 CellID->370384636],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 351, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249670473858*^9},
 CellID->305012795],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 401, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249670910659*^9},
 CellID->324052939],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 451, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496713318596`*^9},
 CellID->175387165],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 501, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496717530603`*^9},
 CellID->18717014],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 551, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249672174261*^9},
 CellID->516028767],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 601, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249672611062*^9},
 CellID->443472550],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "653"}],
  SequenceForm["Creating rule ", 651, " of ", 653],
  Editable->False]], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.656249673032263*^9},
 CellID->285663591],

Cell[BoxData["\<\"Creating thinned data ...\"\>"], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496730634627`*^9},
 CellID->100899060],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->{3.653216258831461*^9, 3.6557012263281393`*^9, 
  3.6562496730946627`*^9},
 CellID->359484989]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "17.73731370000000140407792059704661369324`7.269487760251884", ",", 
   "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.6557012693062143`*^9, 3.6562496730946627`*^9},
 CellID->505902204]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"AnalyzeProofs", "[", 
   RowBox[{"{", "\"\<Theory\>\"", "}"}], "]"}], "//", "Timing"}]], "Input",
 CellChangeTimes->{{3.6406688560104437`*^9, 3.640668867513785*^9}, 
   3.64067664060292*^9, {3.643603325511208*^9, 3.643603327794703*^9}, {
   3.643603543660881*^9, 3.643603577014818*^9}, {3.6532133857706127`*^9, 
   3.653213385926613*^9}, {3.6539040452172146`*^9, 3.653904046714817*^9}},
 CellID->204865936],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\ConsistencyChecker\\\" (\"\>", 
   "\[InvisibleSpace]", "1", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"6\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ConsistencyChecker\" (", 1, " of ", "6", 
   "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049189443755`*^9},
 CellID->92695177],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\ElementaryTheories\\\" (\"\>", 
   "\[InvisibleSpace]", "2", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"6\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ElementaryTheories\" (", 2, " of ", "6", 
   "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049189755754`*^9},
 CellID->77992045],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"391\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "391"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904919006776*^9},
 CellID->173700190],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"391\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "391"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904919725377*^9},
 CellID->200266979],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"391\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "391"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049204985805`*^9},
 CellID->746639354],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"391\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "391"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904921231782*^9},
 CellID->66712991],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"391\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "391"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049220741835`*^9},
 CellID->715089440],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"391\"\>"}],
  SequenceForm["Reading proof ", 251, " of ", "391"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904922776184*^9},
 CellID->589956195],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"391\"\>"}],
  SequenceForm["Reading proof ", 301, " of ", "391"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904923571786*^9},
 CellID->79498839],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"391\"\>"}],
  SequenceForm["Reading proof ", 351, " of ", "391"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049243829875`*^9},
 CellID->29993626],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Fields\\\" (\"\>", "\[InvisibleSpace]", 
   "3", "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"6\"\>",
    "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Fields\" (", 3, " of ", "6", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904925053788*^9},
 CellID->3483066],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"36\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "36"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049250693884`*^9},
 CellID->86123515],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\GroebnerRings\\\" (\"\>", 
   "\[InvisibleSpace]", "4", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"6\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\GroebnerRings\" (", 4, " of ", "6", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049258961897`*^9},
 CellID->762618403],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"164\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "164"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.65390492592739*^9},
 CellID->525181281],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"164\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "164"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049272065926`*^9},
 CellID->299146327],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"164\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "164"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049284701943`*^9},
 CellID->468958116],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"164\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "164"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904929718197*^9},
 CellID->203278600],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Polynomials\\\" (\"\>", 
   "\[InvisibleSpace]", "5", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"6\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Polynomials\" (", 5, " of ", "6", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049301237974`*^9},
 CellID->164689146],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904930154998*^9},
 CellID->14680373],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904931012999*^9},
 CellID->89093219],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904931980201*^9},
 CellID->84778685],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049329172025`*^9},
 CellID->384310759],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904933900004*^9},
 CellID->263550545],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["Reading proof ", 251, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904935007606*^9},
 CellID->989691619],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["Reading proof ", 301, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904935912408*^9},
 CellID->860959],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\ReductionRings\\\" (\"\>", 
   "\[InvisibleSpace]", "6", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"6\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings\" (", 6, " of ", "6", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049367080092`*^9},
 CellID->54780262],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"245\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "245"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904936723609*^9},
 CellID->34839652],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"245\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "245"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904941107217*^9},
 CellID->230492857],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"245\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "245"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049460680256`*^9},
 CellID->56449567],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"245\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "245"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049489228306`*^9},
 CellID->911645706],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"245\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "245"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904949983632*^9},
 CellID->229173356],

Cell[BoxData["\<\"Creating forward rules:\"\>"], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049509508343`*^9},
 CellID->247040699],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049510444345`*^9},
 CellID->517963491],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904951372035*^9},
 CellID->75486086],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 51, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049527292376`*^9},
 CellID->128788291],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 101, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049541020393`*^9},
 CellID->205800235],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 151, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049554436417`*^9},
 CellID->336441065],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 201, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904956800844*^9},
 CellID->78551879],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 251, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049581424465`*^9},
 CellID->99748948],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 301, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904959499649*^9},
 CellID->68337618],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 351, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049608568516`*^9},
 CellID->180700173],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 401, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904962276454*^9},
 CellID->121304440],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 451, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049636180563`*^9},
 CellID->215564190],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 501, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904964975259*^9},
 CellID->97217331],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 551, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904966316861*^9},
 CellID->460196882],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 601, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049676740637`*^9},
 CellID->205116320],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 651, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049690312657`*^9},
 CellID->96939329],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 701, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049703728685`*^9},
 CellID->132562752],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 751, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049717300706`*^9},
 CellID->247813144],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 801, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049730560727`*^9},
 CellID->358470877],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "851", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 851, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049743976755`*^9},
 CellID->227650650],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "901", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 901, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904975739278*^9},
 CellID->129750566],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "951", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 951, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.65390497708088*^9},
 CellID->7673880],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1001", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1001, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049784068823`*^9},
 CellID->1349429],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1051", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1051, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049797484846`*^9},
 CellID->201668803],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1101, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049810744867`*^9},
 CellID->230225205],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1151, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904982400489*^9},
 CellID->302199065],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1201, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049837576914`*^9},
 CellID->45557856],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1251, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904985083694*^9},
 CellID->47265021],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1301, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049864252963`*^9},
 CellID->196575828],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1351, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904987766899*^9},
 CellID->34429249],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1401, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904989092901*^9},
 CellID->1087430],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1451, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049904345036`*^9},
 CellID->366912513],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1501, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049917605057`*^9},
 CellID->37304165],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1551, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904993086508*^9},
 CellID->506613024],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1601, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049944125104`*^9},
 CellID->213938952],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1651, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904995738513*^9},
 CellID->376215522],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1701, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904997064515*^9},
 CellID->264447866],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1751, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653904998390517*^9},
 CellID->350054558],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1801, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539049997321196`*^9},
 CellID->398414970],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1851", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1851, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539050010581217`*^9},
 CellID->137094025],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1901", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1901, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539050023841248`*^9},
 CellID->440482684],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1951", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1994"}],
  SequenceForm["Creating rule ", 1951, " of ", 1994],
  Editable->False]], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653905003725727*^9},
 CellID->807851545],

Cell[BoxData["\<\"Creating thinned data ...\"\>"], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.6539050048645287`*^9},
 CellID->672246938],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->{3.6532175214260798`*^9, 3.653902013896638*^9, 
  3.653904047292018*^9, 3.653905004958129*^9},
 CellID->359470365]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "77.67289789999999527481122640892863273621`7.910869421839609", ",", 
   "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.6539041336089697`*^9, 3.6539050049893293`*^9},
 CellID->8844871]
}, Closed]],

Cell[BoxData[
 RowBox[{
  RowBox[{"AnalyzeProofs", "[", 
   RowBox[{"{", 
    RowBox[{
    "\"\<Theory\>\"", ",", 
     "\"\</home/amaletzk/Theorema Notebooks/ElementaryTheories\>\""}], "}"}], 
   "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.6406688560104437`*^9, 3.640668867513785*^9}, 
   3.64067664060292*^9, {3.643603325511208*^9, 3.643603327794703*^9}, {
   3.643603543660881*^9, 3.643603577014818*^9}},
 CellID->1039312110],

Cell[BoxData[
 RowBox[{"ComputeFormulaPositions", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6406695270485573`*^9, 3.6406695415951242`*^9}, {
   3.64067664934828*^9, 3.640676663567005*^9}, 3.6412902845300865`*^9, 
   3.641290346822996*^9, {3.64361018650988*^9, 3.64361018772442*^9}, 
   3.643610405387268*^9},
 CellID->19428626],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CheckIntegrity", "[", 
  RowBox[{"\"\<CheckPositions\>\"", "\[Rule]", 
   RowBox[{"{", 
    RowBox[{"FileNameJoin", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"CurrentValue", "[", "\"\<NotebookDirectory\>\"", "]"}], ",", 
       RowBox[{"CurrentValue", "[", "\"\<NotebookFileName\>\"", "]"}]}], 
      "}"}], "]"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.64067003109768*^9, 3.6406700374311132`*^9}, {
   3.6406797903560762`*^9, 3.640679792950843*^9}, {3.643610072620981*^9, 
   3.6436100779649982`*^9}, {3.643614652824304*^9, 3.6436146785997877`*^9}, {
   3.643618193913513*^9, 3.643618199226666*^9}, 3.643618234321971*^9},
 CellID->1909803754],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.643533094715043*^9, {3.6436100670377617`*^9, 3.643610082667782*^9}, 
   3.64361014283318*^9, 3.643610175882008*^9, 3.643610422705833*^9, 
   3.643610482460086*^9, {3.6436109604920473`*^9, 3.64361097860835*^9}, 
   3.643611042511623*^9, 3.643611103644537*^9, 3.643614685739502*^9, 
   3.6436182419756403`*^9},
 CellID->721005233]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CollectTerminalFormulas", "[", "]"}]], "Input",
 CellChangeTimes->{{3.642834609232315*^9, 3.642834614238586*^9}},
 CellID->615062589],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"\<\"(- 0 \[ScriptCapitalP] 2)\"\>", ",", "\<\"(344)\"\>", 
   ",", "\<\"(303)\"\>", ",", "\<\"(lcm 4')\"\>", ",", "\<\"(R 1)\"\>", 
   ",", "\<\"(CORR \[CenterDot])\"\>", ",", "\<\"(lm 1)\"\>", 
   ",", "\<\"(404 II)\"\>", 
   ",", "\<\"(isOrderEmbedding * \[VerticalBar])\"\>", ",", "\<\"(381)\"\>", 
   ",", "\<\"(H 6)\"\>", ",", "\<\"(318)\"\>", ",", "\<\"(lcm 4)\"\>", 
   ",", "\<\"(H 9)\"\>", ",", "\<\"(H 4')\"\>", ",", "\<\"(isMonomial 3)\"\>",
    ",", "\<\"(lcm 3)\"\>", ",", "\<\"(C 5')\"\>", 
   ",", "\<\"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb::(areIrrelative symmetric)\"\>", 
   ",", "\<\"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb::(symred reduceM III)\"\>"}], "}"}]], "Output",
 CellChangeTimes->{
  3.642834614891*^9, {3.64283465655628*^9, 3.642834671134451*^9}, 
   3.643533097838657*^9, {3.64360543312173*^9, 3.643605458786525*^9}, 
   3.643605597969597*^9, 3.643614689770647*^9},
 CellID->627924814]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ShowTheoryGraph", "[", "]"}]], "Input",
 CellChangeTimes->{{3.643616235660392*^9, 3.643616239997208*^9}, {
   3.6436163841929083`*^9, 3.6436163911256733`*^9}, 3.643616873065504*^9, {
   3.643618151112858*^9, 3.643618162263517*^9}},
 CellID->1455199198],

Cell[BoxData[
 GraphicsBox[
  TagBox[GraphicsComplexBox[{{0.8660254038443009, 0.4999999999654386}, {
    8.979322840829695*^-11, 0.9999999999827194}, {0., 0.}}, {
     {RGBColor[0.5, 0., 0.], Arrowheads[{{0.1, 0.7}}], ArrowBox[{1, 2}], 
      ArrowBox[{3, 2}], ArrowBox[{3, 1}]}, {InsetBox[
       FrameBox["\<\"Reduction Ring Theory/ReductionRings.nb\"\>",
        Background->RGBColor[1, 1, 0.8],
        FrameStyle->RGBColor[0.94, 0.85, 0.36],
        StripOnInput->False], 1], InsetBox[
       FrameBox["\<\"Reduction Ring Theory/Polynomials.nb\"\>",
        Background->RGBColor[1, 1, 0.8],
        FrameStyle->RGBColor[0.94, 0.85, 0.36],
        StripOnInput->False], 2], InsetBox[
       FrameBox["\<\"ElementaryTheories.nb\"\>",
        Background->RGBColor[1, 1, 0.8],
        FrameStyle->RGBColor[0.94, 0.85, 0.36],
        StripOnInput->False], 3]}}],
   Annotation[#, 
    VertexCoordinateRules -> {{0.8660254038443009, 0.4999999999654386}, {
     8.979322840829695*^-11, 0.9999999999827194}, {0., 0.}}]& ],
  AspectRatio->Automatic,
  FrameTicks->None,
  PlotRange->All,
  PlotRangePadding->Scaled[0.1]]], "Output",
 CellChangeTimes->{
  3.643616240639082*^9, 3.6436163915305767`*^9, 3.643616875095248*^9, {
   3.643617666005062*^9, 3.643617688783246*^9}, 3.643617721794915*^9, 
   3.643617895946937*^9, 3.643617951918654*^9, 3.643618059323213*^9, {
   3.64361815749717*^9, 3.643618162632243*^9}, 3.647926863161089*^9, 
   3.647927549043275*^9},
 CellID->607555663]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"ps", "=", 
   RowBox[{"GetProofSize", "[", "All", "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"ps0", "=", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Total", "[", 
      RowBox[{"Most", "[", 
       RowBox[{"Last", "[", "#", "]"}], "]"}], "]"}], "&"}], ",", "ps"}], 
   "]"}]}]}], "Input",
 CellChangeTimes->{{3.6471569344528646`*^9, 3.6471569482109537`*^9}, {
  3.647157633819579*^9, 3.6471576678972816`*^9}},
 CellID->220425063],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "117", ",", "105", ",", "94", ",", "16", ",", "15", ",", "12", ",", "64", 
   ",", "10", ",", "17", ",", "87", ",", "14", ",", "16", ",", "43", ",", 
   "33", ",", "159", ",", "12", ",", "16", ",", "42", ",", "23", ",", "64", 
   ",", "6", ",", "11", ",", "13", ",", "19", ",", "15", ",", "23", ",", "20",
    ",", "39", ",", "10", ",", "30", ",", "7", ",", "30", ",", "50", ",", 
   "81", ",", "32", ",", "21", ",", "10", ",", "9", ",", "74", ",", "19", ",",
    "9", ",", "62", ",", "62", ",", "11", ",", "50", ",", "25", ",", "27", 
   ",", "143", ",", "78", ",", "52", ",", "18", ",", "6", ",", "52", ",", 
   "84", ",", "14", ",", "39", ",", "110", ",", "24", ",", "18", ",", "77", 
   ",", "10", ",", "72", ",", "46", ",", "21", ",", "24", ",", "39", ",", 
   "37", ",", "12", ",", "43", ",", "11", ",", "152", ",", "25", ",", "7", 
   ",", "11", ",", "47", ",", "14", ",", "7", ",", "5", ",", "30", ",", "12", 
   ",", "124", ",", "52", ",", "70", ",", "24", ",", "17", ",", "19", ",", 
   "15", ",", "45", ",", "7", ",", "33", ",", "242", ",", "49", ",", "5", ",",
    "15", ",", "63", ",", "50", ",", "15", ",", "77", ",", "43", ",", "40", 
   ",", "84", ",", "80", ",", "3", ",", "33", ",", "31", ",", "32", ",", "22",
    ",", "34", ",", "41", ",", "9", ",", "31", ",", "35", ",", "151", ",", 
   "47", ",", "21", ",", "112", ",", "38", ",", "24", ",", "135", ",", "22", 
   ",", "56", ",", "68", ",", "31", ",", "30", ",", "35", ",", "28", ",", 
   "30", ",", "124", ",", "22", ",", "27", ",", "28", ",", "14", ",", "49", 
   ",", "96", ",", "7", ",", "28", ",", "22", ",", "36", ",", "6", ",", "29", 
   ",", "28", ",", "115", ",", "12", ",", "46", ",", "45", ",", "7", ",", 
   "127", ",", "81", ",", "64", ",", "103", ",", "15", ",", "8", ",", "22", 
   ",", "18", ",", "64", ",", "21", ",", "17", ",", "91", ",", "107", ",", 
   "22", ",", "36", ",", "49", ",", "17", ",", "44", ",", "13", ",", "44", 
   ",", "20", ",", "47", ",", "84", ",", "32", ",", "30", ",", "12", ",", 
   "41", ",", "31", ",", "62", ",", "5", ",", "19", ",", "51", ",", "124", 
   ",", "19", ",", "67", ",", "17", ",", "9", ",", "31", ",", "9", ",", "69", 
   ",", "18", ",", "8", ",", "29", ",", "130", ",", "47", ",", "37", ",", 
   "39", ",", "196", ",", "16", ",", "11", ",", "58", ",", "64", ",", "11", 
   ",", "11", ",", "34", ",", "5", ",", "322", ",", "125", ",", "11", ",", 
   "118", ",", "139", ",", "5", ",", "25", ",", "72", ",", "53", ",", "16", 
   ",", "6", ",", "47", ",", "24", ",", "16", ",", "157", ",", "45", ",", 
   "43", ",", "55", ",", "3", ",", "90", ",", "10", ",", "6", ",", "8", ",", 
   "26", ",", "25", ",", "52", ",", "53", ",", "19", ",", "238", ",", "8", 
   ",", "47", ",", "12", ",", "4", ",", "20", ",", "22", ",", "20", ",", "33",
    ",", "107", ",", "30", ",", "40", ",", "83", ",", "13", ",", "13", ",", 
   "52", ",", "45", ",", "33", ",", "35", ",", "9", ",", "11", ",", "35", ",",
    "116", ",", "13", ",", "90", ",", "21", ",", "108", ",", "6", ",", "52", 
   ",", "11", ",", "15", ",", "8", ",", "71", ",", "52", ",", "63", ",", 
   "120", ",", "117", ",", "75", ",", "69", ",", "69", ",", "54", ",", "56", 
   ",", "73", ",", "14", ",", "23", ",", "23", ",", "25", ",", "52", ",", 
   "13", ",", "39", ",", "20", ",", "77", ",", "11", ",", "22", ",", "57", 
   ",", "9", ",", "11", ",", "54", ",", "9", ",", "40", ",", "22", ",", "77", 
   ",", "110", ",", "22", ",", "17", ",", "9", ",", "65", ",", "113", ",", 
   "110", ",", "15", ",", "34", ",", "85", ",", "14", ",", "11", ",", "13", 
   ",", "51", ",", "24", ",", "155", ",", "44", ",", "74", ",", "116", ",", 
   "67", ",", "12", ",", "67", ",", "29", ",", "66", ",", "16", ",", "25", 
   ",", "63", ",", "9", ",", "23", ",", "61", ",", "3", ",", "25"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.647156948574527*^9, 3.647157512728408*^9},
 CellID->2062191598]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Auxiliary Functions", "Subsection",
 CellChangeTimes->{{3.640668579267561*^9, 3.640668583347287*^9}},
 CellID->1454182661],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "LoadPackage", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", "name_String", "]"}], ":=", 
   RowBox[{"LoadPackage", "[", 
    RowBox[{"{", "name", "}"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", 
    RowBox[{"l", ":", 
     RowBox[{"{", "___String", "}"}]}], "]"}], ":=", 
   RowBox[{"Scan", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{"FileNameJoin", "[", 
        RowBox[{"{", 
         RowBox[{"$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
          RowBox[{"#", "<>", "\"\<.m\>\""}]}], "}"}], "]"}], "]"}], "&"}], 
     ",", "l"}], "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640668603629292*^9, 3.640668727460284*^9}, 
   3.641286816086794*^9, {3.6434459595640373`*^9, 3.6434459707531443`*^9}, {
   3.6532132224539256`*^9, 3.6532132349651475`*^9}, {3.653213350764151*^9, 
   3.6532133561617603`*^9}, {3.666440313989313*^9, 3.666440320884782*^9}},
 CellID->2074904605],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "FindFormula", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"FindFormula", "[", "label_String", "]"}], ":=", 
   RowBox[{"NotebookLocate", "[", "label", "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640677756847624*^9, 3.6406777896243896`*^9}, 
   3.640677822869519*^9, {3.640677854902379*^9, 3.6406778579321823`*^9}},
 CellID->1980565193],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<ProofPalette\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.640668742006607*^9, 3.640668751417616*^9}},
 CellID->444074165],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<TheoryTools/TheoryTools\>\"", 
  "]"}]], "Input",
 CellChangeTimes->{{3.6563112347936964`*^9, 3.6563112389901037`*^9}, {
   3.662354706446565*^9, 3.6623547093637705`*^9}, 3.666440324609432*^9},
 CellID->293654037]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Reduction Polynomials", "Section",
 CellChangeTimes->{{3.643446026007135*^9, 3.643446027699668*^9}, {
  3.6607378276190643`*^9, 3.6607378325753474`*^9}},
 CellID->636613528],

Cell[CellGroupData[{

Cell["Commutative Power-Products", "Subsection",
 CellChangeTimes->{{3.633082120264353*^9, 3.633082125742791*^9}},
 CellID->1320444671],

Cell[CellGroupData[{

Cell["Informal Explanation", "Subsubsection",
 CellChangeTimes->{{3.633082130413546*^9, 3.6330821338147907`*^9}},
 CellID->1667576502],

Cell[TextData[{
 "Below we define the category of ",
 StyleBox["commutative power-product domains",
  FontSlant->"Italic"],
 ". A domain \[ScriptCapitalD] belongs to this category iff it is a \
cancellative commutative monoid w.r.t. ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["*", "\[ScriptCapitalD]"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["1", "\[ScriptCapitalD]"], TraditionalForm]]],
 ", it is ordered by the total, irreflexive Noetherian order relation ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalD]"], TraditionalForm]]],
 " satisfying the properties of an admissible term ordering, and it also \
defines the two binary functions ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["/", "\[ScriptCapitalD]"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["lcm", "\[ScriptCapitalD]"], TraditionalForm]]],
 " as well as the binary relation ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalD]"], TraditionalForm]]],
 " (divisibility) accordingly."
}], "Text",
 CellChangeTimes->{{3.643537239204845*^9, 3.6435374783773212`*^9}},
 CellID->1164281211]
}, Closed]],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.633082138040416*^9, 3.6330821397608023`*^9}},
 CellID->1910498050],

Cell["", "OpenEnvironment",
 CellID->127346395],

Cell[CellGroupData[{

Cell["Definition: Category of Commutative Power Products", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594462438216697*^9, 3.594462460468909*^9}, {
  3.594462507250974*^9, 3.594462508072693*^9}, {3.596194043176516*^9, 
  3.596194045704221*^9}, {3.607771136297388*^9, 3.60777113757419*^9}},
 CellID->1867952146],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "\[ScriptCapitalD]"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"isCommPPDomain", "[", "\[ScriptCapitalD]", "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"isCCMonoid", "[", "\[ScriptCapitalD]", "]"}]},
         {
          RowBox[{"isTotalIrreflOrder", "[", 
           RowBox[{
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             TagBox[
              UnderscriptBox["\[Succeeds]", "\[ScriptCapitalD]"],
              Identity], 
             TagBox[")",
              "AutoParentheses"]}], ",", "\[ScriptCapitalD]"}], "]"}]},
         {
          RowBox[{"isNoetherian", "[", 
           RowBox[{
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             TagBox[
              UnderscriptBox["\[Succeeds]", "\[ScriptCapitalD]"],
              Identity], 
             TagBox[")",
              "AutoParentheses"]}], ",", "\[ScriptCapitalD]"}], "]"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
             RowBox[{"[", "x", "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{"x", "\[NotEqual]", 
              UnderscriptBox["1", "\[ScriptCapitalD]"]}], "\[Implies]", 
             RowBox[{"x", 
              UnderscriptBox["\[Succeeds]", "\[ScriptCapitalD]"], 
              UnderscriptBox["1", "\[ScriptCapitalD]"]}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{"isOrderEmbedding", "[", 
           RowBox[{
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             TagBox[
              UnderscriptBox["*", "\[ScriptCapitalD]"],
              Identity], 
             TagBox[")",
              "AutoParentheses"]}], ",", 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             TagBox[
              UnderscriptBox["\[Succeeds]", "\[ScriptCapitalD]"],
              Identity], 
             TagBox[")",
              "AutoParentheses"]}], ",", "\[ScriptCapitalD]"}], "]"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
             RowBox[{"[", 
              RowBox[{"x", ",", "y"}], "]"}]}]], 
           RowBox[{
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{"x", 
              UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalD]"], "y"}], 
             TagBox[")",
              "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{
              UnderscriptBox["\[Exists]", 
               RowBox[{
                UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
                RowBox[{"[", "z", "]"}]}]], 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{"y", "\[Equal]", 
                RowBox[{"z", 
                 UnderscriptBox["*", "\[ScriptCapitalD]"], "x"}]}], 
               TagBox[")",
                "AutoParentheses"]}]}], 
             TagBox[")",
              "AutoParentheses"]}]}]}]},
         {
          RowBox[{"isJoin", "[", 
           RowBox[{
            UnderscriptBox["lcm", "\[ScriptCapitalD]"], ",", 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalD]"], 
             TagBox[")",
              "AutoParentheses"]}], ",", "\[ScriptCapitalD]"}], "]"}]},
         {
          RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
              RowBox[{"[", 
               RowBox[{"x", ",", "y"}], "]"}]}]], 
            RowBox[{"y", 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalD]"], "x"}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{"\[And]", 
             RowBox[{"\[Piecewise]", GridBox[{
                {
                 RowBox[{
                  UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
                  RowBox[{"[", 
                   RowBox[{"x", 
                    UnderscriptBox["/", "\[ScriptCapitalD]"], "y"}], 
                   "]"}]}]},
                {
                 RowBox[{
                  RowBox[{
                   RowBox[{"(", 
                    RowBox[{"x", 
                    UnderscriptBox["/", "\[ScriptCapitalD]"], "y"}], ")"}], 
                   UnderscriptBox["*", "\[ScriptCapitalD]"], "y"}], 
                  "\[Equal]", "x"}]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        }]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCommPPDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:558626034", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->CompressedData["
1:eJwdzGkog3EAx/HNYsUi11K82Eauaa7i1eifyZErRO2NbWWtJncoTcOKF3Pk
BVKyRlEmypljXjDEiqQkZWjNPI8jhjdyPL+9+PZ59xWqGsprfVgslpAJ6qSS
GechRRLGFbNQrpfNw2pthdcdfawg7YQiyje+GPot7SbDbnN2PhxZjC6ClmZ5
O3yIMOigeNXRA8/eL/vgs7JwCl4riufgqTrbAl8zt++gnb/4CCPFihcoaOS+
wkCX6AOWxV1x0vHdMdfE3FBErWmvhZUDifq8O4o4OiaN0LYg+Vi9p4i71/cL
6iirbD2QJisj3FyoOvzLh/6UoRRqh6zODcaLdZEbmqy0Z5Ox6nLp2+uWvdAY
TBOeZrkUCrhhrgFGV1CqG54P1j/BY9uR14P9+KhBRk6nXQSfRP2ZkFVwJYVj
npQcGGvMK4C8IEMJfM/oKoOjddpK+G1SymES26SC9cIVNeRnTLTAtHBJG2y9
fpw082ky3J27ALt+fDwwtOniE4bsudnTjJq+Aw6U3/4GQIvDzINr0j9nVhJN
1G12F/wHqzwMOQ==
  "],
 CellTags->{
  "ID:558626034", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCommPPDomain"},
 CellID->558626034],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->723071054]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1980315351]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Reduction Polynomials Category", "Subsection",
 CellChangeTimes->{{3.6315937462391663`*^9, 3.631593749597096*^9}, {
  3.6315959574116907`*^9, 3.63159595932942*^9}},
 CellID->1879054478],

Cell[CellGroupData[{

Cell["Informal Explanation", "Subsubsection",
 CellChangeTimes->{{3.631593774584*^9, 3.631593778003956*^9}},
 CellID->1252947036],

Cell[TextData[{
 "Below we define the category of ",
 StyleBox["reduction polynomial domains",
  FontSlant->"Italic"],
 ". This is the abstract category of domains ",
 Cell[BoxData[
  FormBox["P", TraditionalForm]]],
 " built from an arbitrary coefficient domain ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 " and another domain ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " (typically in the category of (commutative) power products), containing as \
elements precisely the polynomials over ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 ", characterized by a coefficient function ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["C", "P"], TraditionalForm]]],
 " which maps every polynomial ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]]],
 " and power-product ",
 Cell[BoxData[
  FormBox["\[Tau]", TraditionalForm]]],
 " to the coefficient of ",
 Cell[BoxData[
  FormBox["\[Tau]", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]]],
 ".\nOf course, some other functions of ",
 Cell[BoxData[
  FormBox["P", TraditionalForm]]],
 " such as ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["+", "P"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["*", "P"], TraditionalForm]]],
 ", etc. have to satisfy certain consistency requirements w.r.t. the \
coefficient function ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ", in order to ensure that ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " really ",
 StyleBox["is",
  FontSlant->"Italic"],
 " a coefficient function as one would expect (e.g. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"p", "+", "q"}], ",", "\[Tau]"}], "]"}], "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"p", ",", "\[Tau]"}], "]"}], "+", 
     RowBox[{"C", "[", 
      RowBox[{"q", ",", "\[Tau]"}], "]"}]}]}], TraditionalForm]]],
 ").\nMoreover, ",
 Cell[BoxData[
  FormBox["P", TraditionalForm]]],
 " must also be ordered by the relation ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Succeeds]", "P"], TraditionalForm]]],
 ", which has to be defined in a very concrete way in terms of ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["C", "P"], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.640317119010384*^9, 3.640317119382414*^9}, 
   3.643537488998782*^9, {3.643537566366661*^9, 3.643537614234735*^9}, {
   3.643537644382163*^9, 3.643538105100428*^9}},
 CellID->1930240611],

Cell[TextData[{
 "In addition to the category of reduction polynomial domains there is also \
the category of ",
 StyleBox["algorithmic",
  FontSlant->"Italic"],
 " reduction polynomial domains: Domains belonging to the latter category not \
only have to belong to the former one, but also need to define the three \
functions lcrd (4-ary and unary), rdm (binary) and cpm (5-ary)."
}], "Text",
 CellChangeTimes->{{3.643538159167836*^9, 3.64353827965631*^9}, {
  3.643538319064973*^9, 3.643538336163925*^9}},
 CellID->70056670],

Cell[TextData[{
 "The definition of the category of reduction polynomial domains makes use of \
auxiliary notions (e.g. isMonomial, \[CenterDot], lp, etc.), serving as \
abbreviations for rather complicated expressions that occur very frequently. \
All of them carry the domain-undersscript \[ScriptCapitalP] (for the \
universally quantified variable \[ScriptCapitalP]), which must be understood \
as something like an ",
 StyleBox["extension",
  FontSlant->"Italic"],
 " of the domain \[ScriptCapitalP], but not by means of an extension functor.\
\nThe reason for introducing these auxiliary notions as extensions, not not \
simply requiring them to be defined in \[ScriptCapitalP], is that most \
probably \[ScriptCapitalP] does not define them (they are not necessary in \
concrete representations of polynomials). Hence, if they were required to be \
defined in \[ScriptCapitalP], we would need an extension functor ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[ScriptCapitalE]", "[", 
    RowBox[{
    "\[ScriptCapitalP]", ",", "\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}],
     "]"}], TraditionalForm]]],
 " that takes as input a polynomial domain \[ScriptCapitalP] (without \
auxiliary functions) and a ring \[ScriptCapitalR] and power-product domain \
\[ScriptCapitalT], and extends \[ScriptCapitalP] by the auxiliary functions. \
Then, if one has a concrete polynomial-functor ",
 Cell[BoxData[
  FormBox[
   RowBox[{"P", "[", 
    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
   TraditionalForm]]],
 " and one wants to show that it belongs to a certain category, e.g. the \
category of reduction rings, two things have to be done (note that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"P", "[", 
    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
   TraditionalForm]]],
 " cannot be shown to belong to the category of reduction polynomials, as it \
lacks the auxiliary functions!):\nFirst, one needs to show that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[ScriptCapitalE]", "[", 
    RowBox[{
     RowBox[{"P", "[", 
      RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], ",", 
     "\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
   TraditionalForm]]],
 " belongs to the category of reduction polynomials. From this one can then \
infer that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[ScriptCapitalE]", "[", 
    RowBox[{
     RowBox[{"P", "[", 
      RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], ",", 
     "\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
   TraditionalForm]]],
 " is a reduction ring.\nSecond, one needs to show that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[ScriptCapitalE]", "[", 
    RowBox[{
     RowBox[{"P", "[", 
      RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], ",", 
     "\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
   TraditionalForm]]],
 " can only be a reduction ring if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"P", "[", 
    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
   TraditionalForm]]],
 " is a reduction ring, i.e. one has to prove the \
\[OpenCurlyDoubleQuote]reverse conservation theorem\[CloseCurlyDoubleQuote] \
",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReductionRing", "[", 
     RowBox[{"\[ScriptCapitalE]", "[", 
      RowBox[{
       RowBox[{"P", "[", 
        RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], ",", 
       "\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], "]"}], 
    "\[Implies]", 
    RowBox[{"isReductionRing", "[", 
     RowBox[{"P", "[", 
      RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], "]"}]}],
    TraditionalForm]]],
 ".\n\nThe approach sketched above might be better than the one actually \
pursued, in the sense that it is more robust from the logical point of view. \
Namely, whenever a concrete domain ",
 Cell[BoxData[
  FormBox["P", TraditionalForm]]],
 " is shown to belong to the category of reduction polynomials now, ",
 StyleBox["some",
  FontSlant->"Italic"],
 " definition of the auxiliary functions has to be used \[Dash] and this is \
not necessarily the one from below, although for logical correctness it has \
to be that one. ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ", however, is prefectly fine if various (contradicting) definitions for the \
same notion are provided in one and the same theory, maybe even in the same \
notebook, so one must be very careful to use the right definitions of \
auxiliary notions extending a functor when actually carrying out conservation \
theorems.\nThe problem of having possibly contradicting formulas is a general \
\[OpenCurlyDoubleQuote]problem\[CloseCurlyDoubleQuote] in ",
 StyleBox["Theorema",
  FontSlant->"Italic"],
 ", meaning that functionality checking whether there really are theorems \
that were proved using inconsistent formulas should be made available; See \
the documentation of the \[OpenCurlyDoubleQuote]TheoryAnalyzer\
\[CloseCurlyDoubleQuote]-package for more details. After all, contradictory \
formulas per se do not do any harm and might be very useful (e.g. as \
knowledge in computations) \[Dash] as long as they are not used in proofs."
}], "Text",
 CellChangeTimes->{{3.640317126659127*^9, 3.6403173988211737`*^9}, {
  3.640317464712851*^9, 3.640317997987446*^9}, {3.6403180405388412`*^9, 
  3.6403181144513187`*^9}, {3.640318152942459*^9, 3.640318735338732*^9}},
 CellID->1768045063]
}, Closed]],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.631593783068163*^9, 3.6315937847516212`*^9}, {
  3.636356651223576*^9, 3.636356654942161*^9}, {3.636356704616857*^9, 
  3.636356706449655*^9}},
 CellID->103885810],

Cell[TextData[{
 "We fix now once and for all two arbitrary domains \[ScriptCapitalR] and \
\[ScriptCapitalT]. \[ScriptCapitalR] is an arbitrary commutative ring with \
unit, whereas \[ScriptCapitalT] is a domain of commutative power products. \
The definitions and theorems in this section are stated for \
\[ScriptCapitalR], \[ScriptCapitalT], and ",
 StyleBox["all",
  FontSlant->"Italic"],
 " domains \[ScriptCapitalP]; The latter can be thought of as representations \
of the polynomial ring over \[ScriptCapitalR] and \[ScriptCapitalT]."
}], "Text",
 CellChangeTimes->{{3.6316002092333097`*^9, 3.631600398095415*^9}, {
  3.631601810038787*^9, 3.6316018451174803`*^9}, {3.6316028250094233`*^9, 
  3.6316028611900663`*^9}},
 CellID->481145652],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "\[ScriptCapitalP]"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1568169011}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631593835228012*^9, 3.631593861396139*^9}, {
   3.631594334061886*^9, 3.631594336067224*^9}, {3.631597450673603*^9, 
   3.631597450932754*^9}, 3.631600204794094*^9},
 CellTags->{
  "ID:1568169011", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1568169011],

Cell["", "OpenEnvironment",
 CellID->771652428],

Cell[CellGroupData[{

Cell["Definition: Abstract Notions for Reduction Polynomials", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6315949303379383`*^9, 3.63159494581532*^9}},
 CellID->1166916922],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"p", ",", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1716984519}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631598469014841*^9, 3.631598522966741*^9}, 
   3.6385107611767893`*^9},
 CellTags->{
  "ID:1716984519", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1716984519],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "p", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox[
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
           RowBox[{"[", "s", "]"}]}]], 
         RowBox[{"s", "\[NotEqual]", "t"}]], 
        RowBox[{
         RowBox[{
          UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
          RowBox[{"p", ",", "s"}], "]"}], "\[Equal]", 
         UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:205548127", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631594967116301*^9, 3.6315951192324123`*^9}, {
  3.6316004148144693`*^9, 3.631600423748744*^9}},
 CellTags->{
  "ID:205548127", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial"},
 CellID->205548127],

Cell["\<\
Please note the following difference between the two isMonomial predicates: \
The unary one only holds if its argument has a non-zero coefficient, whereas \
the second argument of the 3-ary one could well be 0.\
\>", "EnvironmentText",
 CellChangeTimes->{{3.64222615649174*^9, 3.642226212942642*^9}},
 CellID->712458476],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "c", ",", "t"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "s", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "s"}], "]"}], "\[Equal]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {"c", "\[DoubleLeftArrow]", 
          RowBox[{"s", "\[Equal]", "t"}]},
         {
          UnderscriptBox["0", "\[ScriptCapitalR]"], "\[DoubleLeftArrow]", 
          "otherwise"}
        }]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:711054552", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642226021617297*^9, 3.642226041323174*^9}, {
   3.6422260889754887`*^9, 3.642226107862536*^9}, 3.642231138947915*^9},
 CellTags->{
  "ID:711054552", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial II"},
 CellID->711054552],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "q", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
       RowBox[{"q", ",", "c", ",", "t"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
         RowBox[{"[", "r", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
          RowBox[{"r", ",", "c", ",", "t"}], "]"}], "\[Implies]", 
         RowBox[{"r", "\[Equal]", "q"}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ":=", 
    RowBox[{
     UnderscriptBox["the", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", "q", "]"}]}]], 
     RowBox[{
      UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
      RowBox[{"q", ",", "c", ",", "t"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(monomial constructor)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2035355406", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642230319605282*^9, 3.642230409279731*^9}, {
  3.642232912979826*^9, 3.642232939827982*^9}, {3.6532137818427086`*^9, 
  3.653213810765159*^9}},
 CellTags->{
  "ID:2035355406", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "monomial constructor"},
 CellID->2035355406],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["supp", "\[ScriptCapitalP]"], "[", "p", "]"}], ":=", 
  RowBox[{"{", 
   RowBox[{"s", 
    UnderscriptBox["|", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "s", "]"}]}]], 
    RowBox[{
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "s"}], "]"}], "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
   "}"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(supp)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:371468988", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6385117836200027`*^9, 3.638511817694865*^9}, {
  3.6385903947329073`*^9, 3.6385903967226686`*^9}},
 CellTags->{
  "ID:371468988", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "supp"},
 CellID->371468988],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isLP", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "t"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
          RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
       {
        RowBox[{
         UnderscriptBox[
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
            RowBox[{"[", "s", "]"}]}]], 
          RowBox[{"s", 
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]], 
         RowBox[{
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"p", ",", "s"}], "]"}], "\[Equal]", 
          UnderscriptBox["0", "\[ScriptCapitalR]"]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isLP)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:741382791", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642234815766428*^9, 3.642234848297056*^9}, {
  3.642236259472804*^9, 3.642236269456119*^9}},
 CellTags->{
  "ID:741382791", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isLP"},
 CellID->741382791],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "s", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["isLP", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "s"}], "]"}], "\[And]", 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "s0", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["isLP", "\[ScriptCapitalP]"], "[", 
         RowBox[{"p", ",", "s0"}], "]"}], "\[Implies]", 
        RowBox[{"s0", "\[Equal]", "s"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], ":=", 
    RowBox[{
     UnderscriptBox["the", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "s", "]"}]}]], 
     RowBox[{
      UnderscriptBox["isLP", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "s"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lp)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1470398456", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631598694166313*^9, 3.631598786361733*^9}, {
  3.631600441392877*^9, 3.631600446386155*^9}, {3.6357472075702267`*^9, 
  3.6357472103972273`*^9}, {3.638511384290827*^9, 3.6385114200118637`*^9}, {
  3.642225982206799*^9, 3.64222598302845*^9}, {3.642234883428659*^9, 
  3.6422349893744383`*^9}},
 CellTags->{
  "ID:1470398456", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp"},
 CellID->1470398456],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ":=", 
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", 
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lc)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1541081556", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63159888124413*^9, 3.631598909579069*^9}, {
  3.6316004504891043`*^9, 3.631600459563054*^9}, {3.635747213631351*^9, 
  3.6357472156916656`*^9}},
 CellTags->{
  "ID:1541081556", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc"},
 CellID->1541081556],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], ":=", 
  RowBox[{
   RowBox[{
    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
   RowBox[{
    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:258866072", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631598634847777*^9, 3.631598684196454*^9}, 
   3.631598864359098*^9, {3.6316004624108467`*^9, 3.6316004752320766`*^9}, {
   3.632716995108303*^9, 3.6327170044199343`*^9}, {3.6357472174952993`*^9, 
   3.635747219298349*^9}},
 CellTags->{
  "ID:258866072", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm"},
 CellID->258866072],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], ":=", 
  RowBox[{"p", 
   UnderscriptBox["-", "\[ScriptCapitalP]"], 
   RowBox[{
    UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1500059087", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631860284890836*^9, 3.631860335201701*^9}, {
   3.633085591637651*^9, 3.6330856078196793`*^9}, {3.635747221177989*^9, 
   3.635747223385756*^9}, 3.638511350039517*^9, {3.6385156949334383`*^9, 
   3.638515705511092*^9}},
 CellTags->{
  "ID:1500059087", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R"},
 CellID->1500059087],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    UnderscriptBox["0", "\[ScriptCapitalP]"], ",", "t"}], "]"}], ":=", 
  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1301149083", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318783731237173`*^9, 3.631878413338489*^9}, {
  3.6318784606564617`*^9, 3.6318784768284883`*^9}, {3.633085712197072*^9, 
  3.633085716022664*^9}, {3.633085767246235*^9, 3.633085768459629*^9}, {
  3.642307014521112*^9, 3.6423070237711697`*^9}},
 CellTags->{
  "ID:1301149083", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 0"},
 CellID->1301149083],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "t"}], "]"}], ":=", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{
          RowBox[{
           UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], 
          UnderscriptBox["+", "\[ScriptCapitalP]"], 
          RowBox[{
           UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
           RowBox[{
            RowBox[{
             UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
            "t"}], "]"}]}], "\[DoubleLeftArrow]", 
         RowBox[{
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
          UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]},
        {
         UnderscriptBox["0", "\[ScriptCapitalP]"], "\[DoubleLeftArrow]", 
         "otherwise"}
       }]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:355792939", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631598988812887*^9, 3.6315990751241093`*^9}, {
   3.631600477575543*^9, 3.631600477836005*^9}, {3.631603785345553*^9, 
   3.631603786124296*^9}, 3.63851118961058*^9, {3.638520417757143*^9, 
   3.6385204180305643`*^9}, {3.642307234903204*^9, 3.642307256843967*^9}, {
   3.642307310955604*^9, 3.642307384771517*^9}, 3.642307498643248*^9},
 CellTags->{
  "ID:355792939", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H"},
 CellID->355792939],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", "q"], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"p", 
      UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], ":=", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          UnderscriptBox["0", "\[ScriptCapitalP]"], "\[DoubleLeftArrow]", 
          RowBox[{"q", "\[Equal]", 
           UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
         {
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
              UnderscriptBox["*", "\[ScriptCapitalR]"], 
              RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
             ")"}], 
            UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
              UnderscriptBox["*", "\[ScriptCapitalT]"], 
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
             ")"}]}], 
           UnderscriptBox["+", "\[ScriptCapitalP]"], 
           RowBox[{"p", 
            UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
            RowBox[{
             UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}], 
          "\[DoubleLeftArrow]", "otherwise"}
        }]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(monomial multiplication)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1617463509", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6385152533425198`*^9, 3.6385154555613117`*^9}, {
  3.638516500258877*^9, 3.638516502081031*^9}, {3.638520420866295*^9, 
  3.6385204211401663`*^9}, {3.6423295758062153`*^9, 3.6423296417109327`*^9}},
 CellTags->{
  "ID:1617463509", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "monomial multiplication"},
 CellID->1617463509],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->176163025]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1861413033],

Cell["", "OpenEnvironment",
 CellID->116290711],

Cell[CellGroupData[{

Cell["Definition: Category of Reduction Polynomials", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631593801506775*^9, 3.6315938130086193`*^9}},
 CellID->822645424],

Cell[BoxData[
 RowBox[{
  RowBox[{"isReductionPolynomialDomain", "[", "\[ScriptCapitalP]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"isDomain", "[", "\[ScriptCapitalP]", "]"}]},
       {
        RowBox[{"hasDecisionPredicate", "[", "\[ScriptCapitalP]", "]"}]},
       {
        RowBox[{"containsMonomials", "[", "\[ScriptCapitalP]", "]"}]},
       {
        RowBox[{"hasCoefficientFunction", "[", "\[ScriptCapitalP]", "]"}]},
       {
        RowBox[{"hasMultiplication", "[", "\[ScriptCapitalP]", "]"}]},
       {
        RowBox[{"isExtensional", "[", "\[ScriptCapitalP]", "]"}]},
       {
        RowBox[{"isOrdered", "[", "\[ScriptCapitalP]", "]"}]},
       {
        RowBox[{"hasMultipliers", "[", "\[ScriptCapitalP]", "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReductionPolynomialDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1026275980", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631593914673176*^9, 3.6315939877897663`*^9}, {
   3.631594036787661*^9, 3.631594047856823*^9}, {3.631594340581171*^9, 
   3.63159441591882*^9}, {3.631594470022511*^9, 3.6315944845470047`*^9}, {
   3.631597214002377*^9, 3.6315972164096537`*^9}, {3.631597454583901*^9, 
   3.631597460099703*^9}, {3.6315999205163107`*^9, 3.631599965465845*^9}, 
   3.631600488098362*^9, {3.631600617035239*^9, 3.631600623716928*^9}, {
   3.631603596985977*^9, 3.631603624019895*^9}, {3.63800937488555*^9, 
   3.638009375253933*^9}, {3.638516361009156*^9, 3.6385163678743687`*^9}, {
   3.638521401071121*^9, 3.638521407563191*^9}, {3.6533725002436533`*^9, 
   3.65337250982207*^9}, {3.6533727009400063`*^9, 3.653372702671609*^9}},
 CellTags->{
  "ID:1026275980", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReductionPolynomialDomain"},
 CellID->1026275980],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasDecisionPredicate", "[", "\[ScriptCapitalP]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
           RowBox[{"[", "p", "]"}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"\[And]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{
                UnderscriptBox["\[ForAll]", 
                 RowBox[{
                  UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
                  RowBox[{"[", "t", "]"}]}]], 
                RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
                 RowBox[{"[", 
                  RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{"p", ",", "t"}], "]"}], "]"}]}]}]},
              {
               RowBox[{"isFinite", "[", 
                RowBox[{
                 UnderscriptBox["supp", "\[ScriptCapitalP]"], "[", "p", "]"}],
                 "]"}]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]},
       {
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
         RowBox[{"[", 
          UnderscriptBox["0", "\[ScriptCapitalP]"], "]"}]}]},
       {
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
         RowBox[{"[", 
          UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}]}]},
       {
        RowBox[{"isClosed", "[", 
         RowBox[{
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           TagBox[
            UnderscriptBox["+", "\[ScriptCapitalP]"],
            Identity], 
           TagBox[")",
            "AutoParentheses"]}], ",", "\[ScriptCapitalP]"}], "]"}]},
       {
        RowBox[{"isClosed", "[", 
         RowBox[{
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           TagBox[
            UnderscriptBox["-", "\[ScriptCapitalP]"],
            Identity], 
           TagBox[")",
            "AutoParentheses"]}], ",", "\[ScriptCapitalP]"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
           RowBox[{"[", "p", "]"}]}]], 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
          RowBox[{"[", 
           RowBox[{
            UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], "]"}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasDecisionPredicate)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1335218902", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.63159472146455*^9, {3.631598140695098*^9, 3.6315981736834183`*^9}, {
   3.631598215380039*^9, 3.631598282941222*^9}, {3.631598345699677*^9, 
   3.631598377748062*^9}, 3.6316006269790983`*^9, 3.6316006888534527`*^9, {
   3.636365336383264*^9, 3.636365355912696*^9}, {3.638516588877585*^9, 
   3.6385166033322697`*^9}, {3.6532139828534627`*^9, 3.653214080322434*^9}},
 CellTags->{
  "ID:1335218902", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasDecisionPredicate"},
 CellID->1335218902],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"containsMonomials", "[", "\[ScriptCapitalP]", "]"}], 
   TraditionalForm]]],
 " is needed to justify the definition of \[CenterDot], i.e. to make sure \
that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[CenterDot]", "t"}], TraditionalForm]]],
 " is defined for all coefficients ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " and power-products ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.638519949271061*^9, 3.638520033118997*^9}, {
  3.6385203574238043`*^9, 3.63852036547281*^9}, {3.638521454834515*^9, 
  3.638521466400988*^9}},
 CellID->1135948496],

Cell[BoxData[
 RowBox[{
  RowBox[{"containsMonomials", "[", "\[ScriptCapitalP]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", "c", "]"}]}], ",", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "t", "]"}]}]}]], 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", "r", "]"}]}]], 
     RowBox[{
      UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
      RowBox[{"r", ",", "c", ",", "t"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(containsMonomials)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1175095894", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638521415386784*^9, 3.6385214495375843`*^9}, {
   3.642226491044745*^9, 3.6422265034397697`*^9}, 3.642229916885146*^9, {
   3.65321396430503*^9, 3.653213970607441*^9}},
 CellTags->{
  "ID:1175095894", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "containsMonomials"},
 CellID->1175095894],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasCoefficientFunction", "[", "\[ScriptCapitalP]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{
            UnderscriptBox["0", "\[ScriptCapitalP]"], ",", "t"}], "]"}], 
          "\[Equal]", 
          UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
        {
         RowBox[{
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{
            UnderscriptBox["1", "\[ScriptCapitalP]"], ",", "t"}], "]"}], 
          "\[Equal]", 
          RowBox[{"\[Piecewise]", GridBox[{
             {
              UnderscriptBox["1", "\[ScriptCapitalR]"], "\[DoubleLeftArrow]", 
              
              RowBox[{"t", "\[Equal]", 
               UnderscriptBox["1", "\[ScriptCapitalT]"]}]},
             {
              UnderscriptBox["0", "\[ScriptCapitalR]"], "\[DoubleLeftArrow]", 
              "otherwise"}
            }]}]}]},
        {
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
            RowBox[{"[", 
             RowBox[{"p", ",", "q"}], "]"}]}]], 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
              RowBox[{"p", 
               UnderscriptBox["+", "\[ScriptCapitalP]"], "q"}], ",", "t"}], 
             "]"}], "\[Equal]", 
            RowBox[{
             RowBox[{
              UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
              RowBox[{"p", ",", "t"}], "]"}], 
             UnderscriptBox["+", "\[ScriptCapitalR]"], 
             RowBox[{
              UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
              RowBox[{"q", ",", "t"}], "]"}]}]}], 
           TagBox[")",
            "AutoParentheses"]}]}]},
        {
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
            RowBox[{"[", "p", "]"}]}]], 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
              RowBox[{
               UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], ",", "t"}], 
             "]"}], "\[Equal]", 
            RowBox[{
             UnderscriptBox["-", "\[ScriptCapitalR]"], 
             RowBox[{
              UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
              RowBox[{"p", ",", "t"}], "]"}]}]}], 
           TagBox[")",
            "AutoParentheses"]}]}]},
        {
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
            RowBox[{"[", 
             RowBox[{"p", ",", "q"}], "]"}]}]], 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
              RowBox[{"p", 
               UnderscriptBox["-", "\[ScriptCapitalP]"], "q"}], ",", "t"}], 
             "]"}], "\[Equal]", 
            RowBox[{
             RowBox[{
              UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
              RowBox[{"p", ",", "t"}], "]"}], 
             UnderscriptBox["-", "\[ScriptCapitalR]"], 
             RowBox[{
              UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
              RowBox[{"q", ",", "t"}], "]"}]}]}], 
           TagBox[")",
            "AutoParentheses"]}]}]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}},
          "RowsIndexed" -> {}}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasCoefficientFunction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1967795228", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6315941867093782`*^9, 3.631594359380292*^9}, {
   3.631594451155209*^9, 3.6315944616190577`*^9}, 3.6315946872409267`*^9, {
   3.631594775944919*^9, 3.6315948261304913`*^9}, {3.631595364522176*^9, 
   3.631595480477392*^9}, {3.6315955584731607`*^9, 3.631595573035953*^9}, {
   3.631595639555682*^9, 3.631595644370315*^9}, {3.631595776537957*^9, 
   3.631595793557384*^9}, {3.631598388447748*^9, 3.631598388966402*^9}, {
   3.631599997376663*^9, 3.631599998666038*^9}, {3.631600629018906*^9, 
   3.631600647970325*^9}, {3.631601929868496*^9, 3.631601951854697*^9}, {
   3.6318707367560883`*^9, 3.631870776276252*^9}, {3.631878495142301*^9, 
   3.631878496380394*^9}, {3.638516568648979*^9, 3.638516569273686*^9}, {
   3.638519837939727*^9, 3.638519917832378*^9}, {3.638520986693838*^9, 
   3.638520987955673*^9}, {3.638521429976663*^9, 3.638521430496562*^9}, {
   3.638521898616457*^9, 3.6385219322337008`*^9}, {3.638590557812854*^9, 
   3.638590612038739*^9}},
 CellTags->{
  "ID:1967795228", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCoefficientFunction"},
 CellID->1967795228],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasMultiplication", "[", "\[ScriptCapitalP]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "q", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["0", "\[ScriptCapitalP]"], 
            UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
           UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
         {
          RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
              RowBox[{"[", "p", "]"}]}]], 
            RowBox[{"p", "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{"p", 
              UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
             RowBox[{
              RowBox[{"p", 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], 
              UnderscriptBox["+", "\[ScriptCapitalP]"], 
              RowBox[{
               RowBox[{
                UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}]}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasMultiplication)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2138030312", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6385163800740433`*^9, 3.63851650890851*^9}, 
   3.638520793670375*^9, 3.639913496665669*^9},
 CellTags->{
  "ID:2138030312", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplication"},
 CellID->2138030312],

Cell[BoxData[
 RowBox[{
  RowBox[{"isExtensional", "[", "\[ScriptCapitalP]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"p", ",", "q"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"p", "\[Equal]", "q"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
          RowBox[{"[", "t", "]"}]}]], 
        RowBox[{
         RowBox[{
          UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
          RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
         RowBox[{
          UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
          RowBox[{"q", ",", "t"}], "]"}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExtensional)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1709544239", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631599823645502*^9, 3.631599895126993*^9}, {
  3.631600656071787*^9, 3.6316006760382147`*^9}, {3.631603565885476*^9, 
  3.631603568237061*^9}},
 CellTags->{
  "ID:1709544239", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isExtensional"},
 CellID->1709544239],

Cell[BoxData[
 RowBox[{
  RowBox[{"isOrdered", "[", "\[ScriptCapitalP]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"p", ",", "q"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"p", 
        UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "q"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[Exists]", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
          RowBox[{"[", "t", "]"}]}]], 
        RowBox[{"\[And]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {
             RowBox[{
              RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
              RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{"q", ",", "t"}], "]"}]}]},
            {
             RowBox[{
              RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{"p", ",", "t"}], "]"}], 
              UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
              RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{"q", ",", "t"}], "]"}]}]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isOrdered)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:235606987", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6315993357404757`*^9, 3.631599457821026*^9}, {
  3.631600659810461*^9, 3.631600695007518*^9}},
 CellTags->{
  "ID:235606987", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrdered"},
 CellID->235606987],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasMultipliers", "[", "\[ScriptCapitalP]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", "p", "]"}]}]], 
     RowBox[{"p", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
    RowBox[{"\[And]", 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{
          RowBox[{
           UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}], 
          "\[Equal]", 
          RowBox[{
           UnderscriptBox["multN", "\[ScriptCapitalR]"], "[", 
           RowBox[{
            UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
           "]"}]}]},
        {
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
            RowBox[{"[", "q", "]"}]}]], 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            RowBox[{
             UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
             RowBox[{"q", ",", "p"}], "]"}], "\[DoubleLongLeftRightArrow]", 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{
              UnderscriptBox["\[Exists]", 
               RowBox[{
                RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", 
                 "]"}]}]], 
              RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{"q", ",", "p", ",", "i"}], "]"}]}], 
             TagBox[")",
              "AutoParentheses"]}]}], 
           TagBox[")",
            "AutoParentheses"]}]}]},
        {
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
            RowBox[{
             UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
             RowBox[{"[", "q", "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
              RowBox[{"q", ",", "p", ",", "i"}], "]"}], 
             "\[DoubleLongLeftRightArrow]", 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               RowBox[{
                UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "q", 
                "]"}], "\[And]", 
               RowBox[{
                UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                RowBox[{
                 RowBox[{
                  UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], 
                 ",", 
                 RowBox[{
                  UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
                 ",", "i"}], "]"}]}], 
              TagBox[")",
               "AutoParentheses"]}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}},
          "RowsIndexed" -> {}}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasMultipliers)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:407906314", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631602118858152*^9, 3.6316023524159803`*^9}, {
  3.631602871742463*^9, 3.631602934952941*^9}, {3.631856493471448*^9, 
  3.631856517603281*^9}, {3.631871200668475*^9, 3.631871214403225*^9}, {
  3.6328094570738993`*^9, 3.632809465321143*^9}, {3.635747271362857*^9, 
  3.6357472759185266`*^9}, {3.63824406327314*^9, 3.6382440645688753`*^9}},
 CellTags->{
  "ID:407906314", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultipliers"},
 CellID->407906314],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->203630590]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->821046233],

Cell["", "OpenEnvironment",
 CellID->17561844],

Cell[CellGroupData[{

Cell["Definition: Category of Algorithmic Reduction Polynomials", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.636356682144143*^9, 3.636356690698522*^9}},
 CellID->1810576968],

Cell[BoxData[
 RowBox[{
  RowBox[{"isAlgoReductionPolynomialDomain", "[", "\[ScriptCapitalP]", "]"}], 
  
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
        "isReductionPolynomialDomain", "[", "\[ScriptCapitalP]", "]"}]},
       {
        RowBox[{"hasLCRD", "[", "\[ScriptCapitalP]", "]"}]},
       {
        RowBox[{"hasRDM", "[", "\[ScriptCapitalP]", "]"}]},
       {
        RowBox[{"hasCPM", "[", "\[ScriptCapitalP]", "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isAlgoReductionPolynomialDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1209025019", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.636356729703504*^9, 3.6363567758168707`*^9}, {
  3.6429278892610283`*^9, 3.6429279020942497`*^9}},
 CellTags->{
  "ID:1209025019", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isAlgoReductionPolynomialDomain"},
 CellID->1209025019],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasLCRD", "[", "\[ScriptCapitalP]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", 
        RowBox[{"p", ",", "q"}], "]"}]}]], 
     RowBox[{
      RowBox[{"p", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
      RowBox[{"q", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{
          UnderscriptBox[
           UnderscriptBox["\[ForAll]", 
            RowBox[{"i", ",", 
             RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", 
              "]"}]}]], 
           RowBox[{"j", "\[GreaterEqual]", "i"}]], 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            RowBox[{
             UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
             RowBox[{"p", ",", "q", ",", "i", ",", "j"}], "]"}], "\[Equal]", 
            RowBox[{"\[Piecewise]", GridBox[{
               {
                RowBox[{
                 UnderscriptBox["let", 
                  RowBox[{"l", "=", 
                   RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", 
                    "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}], ",", "i", ",", "j"}], "]"}]}]], 
                 RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{
                   RowBox[{
                    SubscriptBox["l", "k"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", 
                    "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}]}], "]"}]}], 
                   UnderscriptBox["|", 
                    RowBox[{
                    RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
                    RowBox[{
                    "\[LeftBracketingBar]", "l", 
                    "\[RightBracketingBar]"}]}]]}], 
                  "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", 
                RowBox[{
                 RowBox[{"p", "\[Equal]", "q"}], "\[Or]", 
                 RowBox[{
                  RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}],
                   "\[NotEqual]", 
                  RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                   "]"}]}]}]},
               {
                RowBox[{
                 UnderscriptBox["let", 
                  RowBox[{"l", "=", 
                   RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", 
                    "]"}], "]"}]}]], 
                 RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{
                   RowBox[{
                    SubscriptBox["l", "k"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", 
                    "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}]}], "]"}]}], 
                   UnderscriptBox["|", 
                    RowBox[{
                    RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
                    RowBox[{
                    "\[LeftBracketingBar]", "l", 
                    "\[RightBracketingBar]"}]}]]}], 
                  "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", 
                "otherwise"}
              }]}]}], 
           TagBox[")",
            "AutoParentheses"]}]}]},
        {
         RowBox[{
          RowBox[{
           UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", "p", "]"}], 
          "\[Equal]", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            UnderscriptBox["let", 
             RowBox[{"l", "=", 
              RowBox[{
               UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
               "]"}]}]], 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
              RowBox[{
               SubscriptBox["l", "k"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               RowBox[{
                UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
              
              UnderscriptBox["|", 
               RowBox[{
                RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
                RowBox[{
                "\[LeftBracketingBar]", "l", "\[RightBracketingBar]"}]}]]}], 
             "\[RightAngleBracket]"}]}], 
           TagBox[")",
            "AutoParentheses"]}]}]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}},
          "RowsIndexed" -> {}}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasLCRD)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1424238209", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631960704364332*^9, 3.631960923412918*^9}, {
  3.631960982053197*^9, 3.631960993055088*^9}, {3.6319610495894547`*^9, 
  3.631961067833844*^9}, {3.631961158129099*^9, 3.6319612294780073`*^9}, {
  3.631961316760096*^9, 3.631961380084733*^9}, {3.6357472887677917`*^9, 
  3.635747293511948*^9}, {3.63576241242951*^9, 3.635762413219284*^9}, {
  3.64292272074028*^9, 3.6429227539958467`*^9}, {3.642922839380557*^9, 
  3.642922901450472*^9}, {3.6429230348403063`*^9, 3.642923073936623*^9}, {
  3.642923147269491*^9, 3.6429231574498787`*^9}, {3.6429291003229647`*^9, 
  3.642929103110886*^9}, {3.6563112661653514`*^9, 3.656311268099755*^9}},
 CellTags->{
  "ID:1424238209", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasLCRD"},
 CellID->1424238209],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasRDM", "[", "\[ScriptCapitalP]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"p", ",", "q"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["rdm", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "q"}], "]"}], "\[Equal]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          UnderscriptBox["0", "\[ScriptCapitalP]"], "\[DoubleLeftArrow]", 
          RowBox[{
           RowBox[{"p", "\[Equal]", 
            UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Or]", 
           RowBox[{"q", "\[Equal]", 
            UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]},
         {
          RowBox[{
           UnderscriptBox["let", 
            RowBox[{"r", "=", 
             RowBox[{
              UnderscriptBox["rdm", "\[ScriptCapitalR]"], "[", 
              RowBox[{
               RowBox[{
                UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
               ",", 
               RowBox[{
                UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
              "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{"\[Piecewise]", GridBox[{
               {
                RowBox[{"r", 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", 
                    "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}]}], ")"}]}], "\[DoubleLeftArrow]", 
                RowBox[{
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  RowBox[{"\[Not]", 
                   RowBox[{"r", "\[Equal]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
                  TagBox[")",
                   "AutoParentheses"]}], "\[And]", 
                 RowBox[{
                  RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}], 
                  UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                  RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", 
                   "]"}]}]}]},
               {
                RowBox[{
                 UnderscriptBox["rdm", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                  RowBox[{
                   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], 
                  ",", "q"}], "]"}], "\[DoubleLeftArrow]", "otherwise"}
              }]}], 
            TagBox[")",
             "AutoParentheses"]}]}], "\[DoubleLeftArrow]", "otherwise"}
        }]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasRDM)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1761006179", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642927866413847*^9, 3.642927866951599*^9}, {
   3.6429279066318417`*^9, 3.6429279153102694`*^9}, {3.64299775791119*^9, 
   3.6429980040472107`*^9}, {3.643009539174552*^9, 3.643009637024147*^9}, 
   3.643009743729206*^9, {3.6430134505678787`*^9, 3.643013458776684*^9}},
 CellTags->{
  "ID:1761006179", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasRDM"},
 CellID->1761006179],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasCPM", "[", "\[ScriptCapitalP]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", 
        RowBox[{"p", ",", "q"}], "]"}]}]], 
     RowBox[{
      RowBox[{"p", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
      RowBox[{"q", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{
         UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
      RowBox[{"i", "<", "j"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["let", 
        RowBox[{"l", "=", 
         RowBox[{
          UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
          RowBox[{"p", ",", "q", ",", "i", ",", "j"}], "]"}]}]], 
       RowBox[{"(", 
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "l", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           RowBox[{
            UnderscriptBox["cpm", "\[ScriptCapitalP]"], "[", 
            RowBox[{"p", ",", "q", ",", 
             SubscriptBox["l", "k"], ",", "i", ",", "j"}], "]"}], "\[Equal]", 
           
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{
                UnderscriptBox["let", 
                 RowBox[{"cm", "=", 
                  RowBox[{
                   UnderscriptBox["cpm", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", 
                    "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    SubscriptBox["l", "k"], "]"}], ",", "i", ",", "j"}], 
                   "]"}]}]], 
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                  RowBox[{
                   SubscriptBox["cm", "1"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    SubscriptBox["l", "k"], "]"}], 
                    UnderscriptBox["/", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", 
                    "]"}]}], ")"}]}], ",", 
                  RowBox[{
                   SubscriptBox["cm", "2"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    SubscriptBox["l", "k"], "]"}], 
                    UnderscriptBox["/", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}]}], ")"}]}]}], "\[RightAngleBracket]"}]}], 
               "\[DoubleLeftArrow]", 
               RowBox[{
                RowBox[{"p", "\[Equal]", "q"}], "\[Or]", 
                RowBox[{
                 RowBox[{
                  UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
                 "\[NotEqual]", 
                 RowBox[{
                  UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                  "]"}]}]}]},
              {
               RowBox[{
                UnderscriptBox["let", 
                 RowBox[{"r", "=", 
                  RowBox[{
                   UnderscriptBox["rdm", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    SubscriptBox["l", "k"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", 
                    "]"}]}], "]"}]}]], 
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                  RowBox[{"r", 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    SubscriptBox["l", "k"], "]"}], 
                    UnderscriptBox["/", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", 
                    "]"}]}], ")"}]}], ",", 
                  RowBox[{"r", 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    SubscriptBox["l", "k"], "]"}], 
                    UnderscriptBox["/", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}]}], ")"}]}]}], "\[RightAngleBracket]"}]}], 
               "\[DoubleLeftArrow]", "otherwise"}
             }]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}], ")"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasCPM)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:667868750", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642927886054201*^9, 3.64292788659455*^9}, {
   3.642927917097475*^9, 3.6429279248828917`*^9}, {3.642998050870968*^9, 
   3.642998060246538*^9}, {3.642998329937105*^9, 3.642998372342753*^9}, {
   3.642998428048259*^9, 3.642998487098472*^9}, {3.6429985614373426`*^9, 
   3.6429986356106586`*^9}, {3.642998687499818*^9, 3.642998745349291*^9}, 
   3.642998997239704*^9, {3.643346930598254*^9, 3.64334695724085*^9}, {
   3.643346988275907*^9, 3.643347061244727*^9}, {3.6563112730917635`*^9, 
   3.6563112747297664`*^9}},
 CellTags->{
  "ID:667868750", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCPM"},
 CellID->667868750],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->64457783]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2016976285]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Theorem: Reduction Polynomials Constitute Commutative Ring with Unit (\
\[Checkmark])\
\>", "Section",
 CellChangeTimes->{{3.6363703605889*^9, 3.636370364295041*^9}, {
  3.638509943217206*^9, 3.638509950682328*^9}, {3.640668170966351*^9, 
  3.640668175127111*^9}},
 CellID->928775899],

Cell["\<\
We now state the conservation theorems for polynomials for the category of \
reduction polynomials defined above. If we then want to show that an \
arbitrary polynomial functor produces reduction rings, we \
\[OpenCurlyDoubleQuote]only\[CloseCurlyDoubleQuote] have to show that it \
belongs to the category of reduction polynomials \[Dash] and this is in \
general much easier.
The first conservation theorem to be proved is the one stating that if \
\[ScriptCapitalR] is a commutative ring with unit and \[ScriptCapitalT] is a \
commutative power-product domain, then \[ScriptCapitalP] is a commutative \
ring with unit as well.\
\>", "Text",
 CellChangeTimes->{{3.631596507508501*^9, 3.631596561948351*^9}, {
  3.631596642395876*^9, 3.631596750465472*^9}, {3.631596784207755*^9, 
  3.6315967873141623`*^9}, {3.638510142584992*^9, 3.6385103286507483`*^9}},
 CellID->1613723827],

Cell[TextData[{
 "Similar as in the previous section we fix two domains \[ScriptCapitalR] and \
\[ScriptCapitalT], where \[ScriptCapitalR] is a commutative ring with unit \
and \[ScriptCapitalT] is a power-product domain. However, we also fix an \
arbitrary domain \[ScriptCapitalP] and show the theorems only for this \
particular domain. Still, since \[ScriptCapitalP] is chosen completely \
arbitrary, the results carry over to all domains of reduction polynomials. \
The reason for employing this strategy is because it saves us from \
instantiating ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[ForAll]", "\[ScriptCapitalP]"], TraditionalForm]]],
 " again and again in proofs.\nRather than requiring \[ScriptCapitalP] to be \
a ReductionPolynomialDomain, \[ScriptCapitalP] only has to satisfy some of \
the axioms of a ReductionPolynomialDomain, namely all except isOrdered and \
hasMultipliers. These are only needed for proving that \[ScriptCapitalP] is a \
reduction ring, which is done in the next section. However, at the present \
stage we do not require \[ScriptCapitalP] to satisfy hasMultiplication \
either; This will be required only from subsection \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " is Closed\[CloseCurlyDoubleQuote] on."
}], "Text",
 CellChangeTimes->{{3.631600900654128*^9, 3.63160100358781*^9}, {
  3.631872522030702*^9, 3.63187264584424*^9}, {3.631873238395948*^9, 
  3.631873238805386*^9}, {3.6318732891337223`*^9, 3.63187337462195*^9}, {
  3.638510129097024*^9, 3.638510131418751*^9}, {3.638510347099649*^9, 
  3.638510360462077*^9}, {3.63971342729344*^9, 3.6397135330721817`*^9}, {
  3.6403194853277063`*^9, 3.6403195346976433`*^9}},
 CellID->2031272829],

Cell["", "OpenEnvironment",
 CellID->753349515],

Cell[CellGroupData[{

Cell["\<\
Presupposition: Ring \[ScriptCapitalR], Power-Product Domain \
\[ScriptCapitalT], and Domain \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638510043311256*^9, 3.638510053788507*^9}, {
  3.639714597387479*^9, 3.639714598328397*^9}},
 CellID->579292177],

Cell[BoxData[
 RowBox[{"isCommRing1", "[", "\[ScriptCapitalR]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ring \[ScriptCapitalR])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:267444622", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638510072921596*^9, 3.6385100785495567`*^9}},
 CellTags->{
  "ID:267444622", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "ring \[ScriptCapitalR]"},
 CellID->267444622],

Cell[BoxData[
 RowBox[{"isCommPPDomain", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(power-product domain \[ScriptCapitalT])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1725827561", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631600542094768*^9, 3.631600569309743*^9}, {
  3.633082276430855*^9, 3.633082277619791*^9}},
 CellTags->{
  "ID:1725827561", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "power-product domain \[ScriptCapitalT]"},
 CellID->1725827561],

Cell[BoxData[
 RowBox[{"isDomain", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isDomain \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:90242530", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6533738530432296`*^9, 3.6533738584408393`*^9}},
 CellTags->{
  "ID:90242530", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isDomain \[ScriptCapitalP]"},
 CellID->90242530],

Cell[BoxData[
 RowBox[{"hasDecisionPredicate", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasDecisionPredicate \[ScriptCapitalP])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1949049350", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639713262918846*^9, 3.639713271828372*^9}},
 CellTags->{
  "ID:1949049350", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasDecisionPredicate \[ScriptCapitalP]"},
 CellID->1949049350],

Cell[BoxData[
 RowBox[{"containsMonomials", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(containsMonomials \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:378755333", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639713283076461*^9, 3.639713295812394*^9}},
 CellTags->{
  "ID:378755333", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "containsMonomials \[ScriptCapitalP]"},
 CellID->378755333],

Cell[BoxData[
 RowBox[{"hasCoefficientFunction", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasCoefficientFunction \[ScriptCapitalP])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:226930466", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639713284088529*^9, 3.639713307379982*^9}},
 CellTags->{
  "ID:226930466", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCoefficientFunction \[ScriptCapitalP]"},
 CellID->226930466],

Cell[BoxData[
 RowBox[{"isExtensional", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExtensional \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:221039322", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.639713285329871*^9, {3.639713322944203*^9, 3.639713325037957*^9}},
 CellTags->{
  "ID:221039322", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isExtensional \[ScriptCapitalP]"},
 CellID->221039322],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->207394581]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->13183544],

Cell[CellGroupData[{

Cell["Preparations (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.638509979118535*^9, 3.638509981793208*^9}, {
  3.6399061793193703`*^9, 3.639906183424059*^9}},
 CellID->740008399],

Cell[CellGroupData[{

Cell["\<\
Lemmas about Power Products (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.6357458004719753`*^9, 3.6357458081227503`*^9}, {
  3.638071144294649*^9, 3.638071172935194*^9}, {3.639209777418563*^9, 
  3.639209781716118*^9}, {3.6479273155166616`*^9, 3.647927316590074*^9}},
 CellID->1996641139],

Cell["\<\
We state some immediate lemmas that hold in general for commutative power \
products. Still, for the sake of simplicity, we only state them for the \
\[OpenCurlyDoubleQuote]arbitrary but fixed\[CloseCurlyDoubleQuote] domain \
\[ScriptCapitalT].\
\>", "Text",
 CellChangeTimes->{{3.635745836058868*^9, 3.635745913781855*^9}},
 CellID->991585047],

Cell["", "OpenEnvironment",
 CellID->642175228],

Cell[CellGroupData[{

Cell["\<\
Lemma: Elementary Properties of Commutative Power-Product Domains\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.633082465304696*^9, 3.633082469719441*^9}, {
  3.6330827121012278`*^9, 3.633082725776103*^9}},
 CellID->621471139],

Cell[BoxData[
 RowBox[{"isDomain", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isDomain \[ScriptCapitalT])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:216808867", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.653377259365929*^9, 3.6533772639679375`*^9}},
 CellTags->{
  "ID:216808867", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isDomain \[ScriptCapitalT]"},
 CellID->216808867],

Cell[BoxData[
 RowBox[{"isPartIrreflOrder", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalT]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(isPartIrreflOrder \[Succeeds] \[ScriptCapitalT])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1391060217", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647926998723988*^9, 3.647927006589944*^9}},
 CellTags->{
  "ID:1391060217", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isPartIrreflOrder \[Succeeds] \[ScriptCapitalT]"},
 CellID->1391060217],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "P"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "t", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          UnderscriptBox[
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
             RowBox[{"[", "s", "]"}]}]], 
           RowBox[{"t", 
            UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "s"}]], 
          RowBox[{"P", "[", "s", "]"}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[Implies]", 
        RowBox[{"P", "[", "t", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "t", "]"}]}]], 
      RowBox[{"P", "[", "t", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Noetherian induction on \[ScriptCapitalT])", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1131450960", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6404178065498657`*^9, 3.640417865891036*^9}},
 CellTags->{
  "ID:1131450960", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Noetherian induction on \[ScriptCapitalT]", "IR"},
 CellID->1131450960],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
  RowBox[{"[", 
   UnderscriptBox["1", "\[ScriptCapitalT]"], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership 1 \[ScriptCapitalT])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:953344354", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6358495039188128`*^9, 3.635849543793242*^9}},
 CellTags->{
  "ID:953344354", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 1 \[ScriptCapitalT]", "IR"},
 CellID->953344354],

Cell[TextData[{
 "We do not require \[ScriptCapitalT] to have more than one elements. But if \
it does have more than one elements, it even has infinitely many, and for \
every element ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " there exists another element ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 " which is strictly greater than ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.6330851261098833`*^9, 3.633085233036564*^9}},
 CellID->699465807],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{"t", "\[NotEqual]", 
     UnderscriptBox["1", "\[ScriptCapitalT]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "s", "]"}]}]], 
     RowBox[{"s", 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PP 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1367902825", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6330850023239803`*^9, 3.633085079461817*^9}},
 CellTags->{
  "ID:1367902825", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 1"},
 CellID->1367902825],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", 
    RowBox[{"s", ",", "t"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 643177315}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633082496675318*^9, 3.633082506214027*^9}},
 CellTags->{
  "ID:643177315", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->643177315],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{
   UnderscriptBox["1", "\[ScriptCapitalT]"], 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
   "s"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PP 1 min)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1568022983", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6391318665352697`*^9, 3.639131868459961*^9}, {
  3.639131902489953*^9, 3.6391319226954527`*^9}},
 CellTags->{
  "ID:1568022983", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 1 min"},
 CellID->1568022983],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"s", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}], "\[Equal]", 
    UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"s", "\[Equal]", 
     UnderscriptBox["1", "\[ScriptCapitalT]"]}], "\[And]", 
    RowBox[{"t", "\[Equal]", 
     UnderscriptBox["1", "\[ScriptCapitalT]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PP 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1723905914", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639130935338682*^9, 3.6391309832737722`*^9}, {
  3.639131881710422*^9, 3.63913188448307*^9}},
 CellTags->{
  "ID:1723905914", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 0"},
 CellID->1723905914],

Cell[BoxData[
 RowBox[{"s", 
  UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
  RowBox[{"(", 
   RowBox[{"t", 
    UnderscriptBox["*", "\[ScriptCapitalT]"], "s"}], 
   ")"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PP 2')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1825953062", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63876592705016*^9, 3.638765936158217*^9}},
 CellTags->{
  "ID:1825953062", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 2'"},
 CellID->1825953062],

Cell[BoxData[
 RowBox[{"s", 
  UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
  RowBox[{"(", 
   RowBox[{"s", 
    UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}], 
   ")"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PP 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:596582273", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633082518595381*^9, 3.633082532605179*^9}},
 CellTags->{
  "ID:596582273", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 2"},
 CellID->596582273],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"t", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "s"}], ")"}], 
   UnderscriptBox["/", "\[ScriptCapitalT]"], "s"}], "\[Equal]", 
  "t"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PP 3')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:389623087", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639133111089438*^9, 3.639133114782312*^9}},
 CellTags->{
  "ID:389623087", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 3'"},
 CellID->389623087],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"s", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}], ")"}], 
   UnderscriptBox["/", "\[ScriptCapitalT]"], "s"}], "\[Equal]", 
  "t"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PP 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:885344283", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633083349418803*^9, 3.633083366366675*^9}},
 CellTags->{
  "ID:885344283", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 3"},
 CellID->885344283],

Cell[BoxData[
 RowBox[{
  RowBox[{"s", 
   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"s", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PP 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1033740528", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633168775680118*^9, 3.633168795817749*^9}},
 CellTags->{
  "ID:1033740528", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 4"},
 CellID->1033740528],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "u", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 475932409}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6404250863851233`*^9, 3.640425094097393*^9}},
 CellTags->{
  "ID:475932409", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->475932409],

Cell["\<\
(PP 5) is cancellativity, stated explicitly for the concrete power-product \
domain \[ScriptCapitalT].\
\>", "EnvironmentText",
 CellChangeTimes->{{3.63910962585925*^9, 3.639109670413878*^9}},
 CellID->723934924],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"s", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "u"}], "\[Equal]", 
    RowBox[{"t", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "u"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"s", "\[Equal]", "t"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PP 5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1124160616", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639109551839905*^9, 3.639109610469002*^9}, 
   3.639109848695009*^9, {3.640425097130767*^9, 3.640425102708317*^9}},
 CellTags->{
  "ID:1124160616", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 5"},
 CellID->1124160616],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"s", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}], 
    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "u"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"s", 
     UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "u"}], "\[And]", 
    RowBox[{"t", 
     UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
     RowBox[{"(", 
      RowBox[{"u", 
       UnderscriptBox["/", "\[ScriptCapitalT]"], "s"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PP 6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1594052474", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640425112399967*^9, 3.640425185146881*^9}, {
  3.640427118120545*^9, 3.640427125423091*^9}},
 CellTags->{
  "ID:1594052474", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 6"},
 CellID->1594052474],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"s", 
      UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}], ")"}], 
    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "u"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"u", 
       UnderscriptBox["/", "\[ScriptCapitalT]"], "s"}], ")"}], 
     UnderscriptBox["/", "\[ScriptCapitalT]"], "t"}], "\[Equal]", 
    RowBox[{"u", 
     UnderscriptBox["/", "\[ScriptCapitalT]"], 
     RowBox[{"(", 
      RowBox[{"s", 
       UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PP 7)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1091730754", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6404261711971817`*^9, 3.640426235362344*^9}},
 CellTags->{
  "ID:1091730754", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 7"},
 CellID->1091730754],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"u", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
     RowBox[{"s", 
      UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}]}], "\[And]", 
    RowBox[{"s", 
     UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "u"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"u", 
     UnderscriptBox["/", "\[ScriptCapitalT]"], "s"}], ")"}], 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
   "t"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PP 8)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1114122233", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6410210308999567`*^9, 3.641021077459268*^9}},
 CellTags->{
  "ID:1114122233", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 8"},
 CellID->1114122233],

Cell[BoxData[
 RowBox[{"isPartReflOrder", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalT]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isPartReflOrder \[VerticalBar])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1909427587", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633082655377406*^9, 3.6330826899169083`*^9}},
 CellTags->{
  "ID:1909427587", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isPartReflOrder \[VerticalBar]"},
 CellID->1909427587],

Cell[BoxData[
 RowBox[{"isOrderEmbedding", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["*", "\[ScriptCapitalT]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalT]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isOrderEmbedding * \[VerticalBar])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:196272597", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.635747014469592*^9, 3.635747042860353*^9}},
 CellTags->{
  "ID:196272597", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrderEmbedding * \[VerticalBar]"},
 CellID->196272597],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1191981650]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2001326268],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isDomain \[ScriptCapitalT])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isDomain \[ScriptCapitalT])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "isDomain", "[", "\[ScriptCapitalT]", 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "isDomain", "[", "\[ScriptCapitalT]", 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:216808867"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p216808867-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p216808867-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:216808867", "Proof|ID:216808867-1"},
 CellID->77023573],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p216808867-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->44400236]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isPartIrreflOrder \[Succeeds] \[ScriptCapitalT])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isPartIrreflOrder \[Succeeds] \[ScriptCapitalT])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isPartIrreflOrder", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], ",", 
           "\[ScriptCapitalT]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isPartIrreflOrder", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], ",", 
           "\[ScriptCapitalT]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1391060217"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1391060217-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1391060217-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1391060217", "Proof|ID:1391060217-1"},
 CellID->640802005],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1391060217-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.653307199295559*^9},
 CellID->1491734577]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Noetherian induction on \[ScriptCapitalT])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Noetherian induction on \[ScriptCapitalT])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["s", "ExpressionVariable"]}]], 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["s", "ExpressionVariable"], "]"}]}], "\[Implies]", 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["t", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["s", "ExpressionVariable"]}]], 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["s", "ExpressionVariable"], "]"}]}], "\[Implies]", 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["t", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1131450960"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1131450960-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1131450960-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1131450960", "Proof|ID:1131450960-1"},
 CellID->381161681],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1131450960-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->397918332]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership 1 \[ScriptCapitalT])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership 1 \[ScriptCapitalT])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
         UnderscriptBox["1", "\[ScriptCapitalT]"], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
         UnderscriptBox["1", "\[ScriptCapitalT]"], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:953344354"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p953344354-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p953344354-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:953344354", "Proof|ID:953344354-1"},
 CellID->238169741],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p953344354-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->236757624]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PP 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PP 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["1", "\[ScriptCapitalT]"]}]}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["1", "\[ScriptCapitalT]"]}]}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1367902825"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1367902825-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1367902825-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1367902825", "Proof|ID:1367902825-1"},
 CellID->640235643],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1367902825-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1625734410]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PP 1 min)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PP 1 min)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               UnderscriptBox["1", "\[ScriptCapitalT]"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["s", "ExpressionVariable"]}], 
             TagBox[")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               UnderscriptBox["1", "\[ScriptCapitalT]"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["s", "ExpressionVariable"]}], 
             TagBox[
             ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1568022983"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1568022983-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1568022983-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1568022983", "Proof|ID:1568022983-1"},
 CellID->1849364148],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1568022983-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->763438333]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PP 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PP 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
             UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}], "\[And]", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
             UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}], "\[And]", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1723905914"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1723905914-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1723905914-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1723905914", "Proof|ID:1723905914-1"},
 CellID->818368910],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1723905914-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2031601064]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PP 2')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PP 2')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["s", "ExpressionVariable"], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
           RowBox[{
             StyleBox["t", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalT]"], 
             StyleBox[
             "s", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["s", "ExpressionVariable"], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
           RowBox[{
             StyleBox["t", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalT]"], 
             StyleBox[
             "s", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1825953062"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1825953062-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1825953062-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1825953062", "Proof|ID:1825953062-1"},
 CellID->10445487],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1825953062-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1071552749]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PP 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PP 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["s", "ExpressionVariable"], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalT]"], 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["s", "ExpressionVariable"], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalT]"], 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:596582273"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p596582273-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p596582273-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:596582273", "Proof|ID:596582273-1"},
 CellID->525784934],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p596582273-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1816433154]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PP 3')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PP 3')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["s", "ExpressionVariable"]}], 
             UnderscriptBox["/", "\[ScriptCapitalT]"], 
             StyleBox["s", "ExpressionVariable"]}], "\[Equal]", 
           StyleBox[
           "t", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["s", "ExpressionVariable"]}], 
             UnderscriptBox["/", "\[ScriptCapitalT]"], 
             StyleBox["s", "ExpressionVariable"]}], "\[Equal]", 
           StyleBox[
           "t", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:389623087"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p389623087-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p389623087-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:389623087", "Proof|ID:389623087-1"},
 CellID->601121285],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p389623087-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->201155910]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PP 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PP 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["/", "\[ScriptCapitalT]"], 
             StyleBox["s", "ExpressionVariable"]}], "\[Equal]", 
           StyleBox[
           "t", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["/", "\[ScriptCapitalT]"], 
             StyleBox["s", "ExpressionVariable"]}], "\[Equal]", 
           StyleBox[
           "t", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:885344283"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p885344283-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p885344283-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:885344283", "Proof|ID:885344283-1"},
 CellID->573214883],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p885344283-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->9167544]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PP 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PP 4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["s", "ExpressionVariable"], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
             StyleBox["t", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["s", "ExpressionVariable"], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
             StyleBox["t", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1033740528"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1033740528-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1033740528-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1033740528", "Proof|ID:1033740528-1"},
 CellID->1813081668],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1033740528-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->97754333]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PP 5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PP 5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1124160616"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1124160616-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1124160616-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1124160616", "Proof|ID:1124160616-1"},
 CellID->160370785],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1124160616-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->98989517]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PP 6)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PP 6)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
             StyleBox["u", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 StyleBox["u", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   StyleBox["s", "ExpressionVariable"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
             StyleBox["u", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 StyleBox["u", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   StyleBox["u", "ExpressionVariable"], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   StyleBox["s", "ExpressionVariable"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1594052474"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1594052474-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1594052474-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1594052474", "Proof|ID:1594052474-1"},
 CellID->1673583185],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1594052474-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{
  3.64345446885714*^9, {3.643455372981235*^9, 3.643455374125091*^9}, 
   3.643455519524008*^9, 3.6434567710298862`*^9},
 CellID->1940769656]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PP 7)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PP 7)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
             StyleBox["u", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["u", "ExpressionVariable"], 
                 UnderscriptBox["/", "\[ScriptCapitalT]"], 
                 StyleBox["s", "ExpressionVariable"]}], 
               UnderscriptBox["/", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["/", "\[ScriptCapitalT]"], 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 StyleBox[
                 "t", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
             StyleBox["u", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["u", "ExpressionVariable"], 
                 UnderscriptBox["/", "\[ScriptCapitalT]"], 
                 StyleBox["s", "ExpressionVariable"]}], 
               UnderscriptBox["/", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["/", "\[ScriptCapitalT]"], 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 StyleBox[
                 "t", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1091730754"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1091730754-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1091730754-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1091730754", "Proof|ID:1091730754-1"},
 CellID->347735719],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1091730754-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.643454115748432*^9},
 CellID->960937874]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PP 8)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PP 8)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["/", "\[ScriptCapitalT]"], 
               StyleBox["s", "ExpressionVariable"]}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["u", "ExpressionVariable"], 
               UnderscriptBox["/", "\[ScriptCapitalT]"], 
               StyleBox["s", "ExpressionVariable"]}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1114122233"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1114122233-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1114122233-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1114122233", "Proof|ID:1114122233-1"},
 CellID->1981139454],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1114122233-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->669358996]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isPartReflOrder \[VerticalBar])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isPartReflOrder \[VerticalBar])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isPartReflOrder", "[", 
         RowBox[{
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], ",", 
           "\[ScriptCapitalT]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isPartReflOrder", "[", 
         RowBox[{
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], ",", 
           "\[ScriptCapitalT]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1909427587"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1909427587-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1909427587-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1909427587", "Proof|ID:1909427587-1"},
 CellID->106187856],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1909427587-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1650175325]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isOrderEmbedding * \[VerticalBar])\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(isOrderEmbedding * \[VerticalBar])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isOrderEmbedding", "[", 
         RowBox[{
           UnderscriptBox["*", "\[ScriptCapitalT]"], ",", 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], ",", 
           "\[ScriptCapitalT]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isOrderEmbedding", "[", 
         RowBox[{
           UnderscriptBox["*", "\[ScriptCapitalT]"], ",", 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], ",", 
           "\[ScriptCapitalT]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:196272597"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p196272597-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p196272597-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:196272597", "Proof|ID:196272597-1"},
 CellID->626133587],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p196272597-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->31720515],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->{3.656762001850816*^9},
 CellID->737442690]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->867314350],

Cell[CellGroupData[{

Cell["Lemma: Properties of LCM", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6391933920193863`*^9, 3.639193396223447*^9}},
 CellID->1123038343],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", 
    RowBox[{"s", ",", "t"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 334245933}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633082496675318*^9, 3.633082506214027*^9}},
 CellTags->{
  "ID:334245933", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->334245933],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
    RowBox[{"s", ",", "t"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership lcm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1101510167", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639132157468656*^9, 3.63913217134093*^9}},
 CellTags->{
  "ID:1101510167", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lcm", "IR"},
 CellID->1101510167],

Cell["\<\
We state some properties of joins (from \[OpenCurlyDoubleQuote]Algebraic \
Structures\[CloseCurlyDoubleQuote]) explicitly for lcm and \[VerticalBar].\
\>", "EnvironmentText",
 CellChangeTimes->{{3.639193823818016*^9, 3.639193858608754*^9}, {
  3.639194205127921*^9, 3.639194205945096*^9}, {3.656318966755719*^9, 
  3.6563189707493258`*^9}},
 CellID->102897427],

Cell[BoxData[
 RowBox[{"s", 
  UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
  RowBox[{
   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
   RowBox[{"s", ",", "t"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lcm 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:571674285", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6391939928202963`*^9, 3.639194005714436*^9}},
 CellTags->{
  "ID:571674285", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 1"},
 CellID->571674285],

Cell[BoxData[
 RowBox[{"t", 
  UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
  RowBox[{
   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
   RowBox[{"s", ",", "t"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lcm 1')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:775675057", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6391939928202963`*^9, 3.6391940135361547`*^9}},
 CellTags->{
  "ID:775675057", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 1'"},
 CellID->775675057],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "u", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"s", 
       UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "u"}], "\[And]", 
      RowBox[{"t", 
       UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "u"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
      RowBox[{"s", ",", "t"}], "]"}], 
     UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "u"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lcm 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:517355459", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63919386555128*^9, 3.639193909123105*^9}},
 CellTags->{
  "ID:517355459", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 2"},
 CellID->517355459],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
   RowBox[{"t", ",", "s"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lcm 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:651227563", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6391934484459143`*^9, 3.6391934624110117`*^9}},
 CellTags->{
  "ID:651227563", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 3"},
 CellID->651227563],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
   RowBox[{"s", ",", "s"}], "]"}], "\[Equal]", 
  "s"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lcm 6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:482943675", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6563223008915763`*^9, 3.6563223097211914`*^9}},
 CellTags->{
  "ID:482943675", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 6"},
 CellID->482943675],

Cell["\<\
The following properties really only hold for lcm, but not for joins in \
general.\
\>", "EnvironmentText",
 CellChangeTimes->{{3.6391940474116173`*^9, 3.639194061656447*^9}},
 CellID->1329971511],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{"s", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
   RowBox[{
    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
    RowBox[{"s", ",", "t"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lcm 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:39860588", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639193469957872*^9, 3.6391934889283867`*^9}},
 CellTags->{
  "ID:39860588", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 4"},
 CellID->39860588],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{"t", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
   RowBox[{
    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
    RowBox[{"s", ",", "t"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lcm 4')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1631465634", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639193498954361*^9, 3.63919350091043*^9}},
 CellTags->{
  "ID:1631465634", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 4'"},
 CellID->1631465634],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "u", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"s", 
       UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "u"}], "\[And]", 
      RowBox[{"t", 
       UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "u"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
        RowBox[{"s", ",", "t"}], "]"}], 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "u"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lcm 5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:835022497", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639194257789166*^9, 3.639194285268198*^9}},
 CellTags->{
  "ID:835022497", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 5"},
 CellID->835022497],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->157525244]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1568132504],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership lcm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership lcm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
           RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
           RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1101510167"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1101510167-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1101510167-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1101510167", "Proof|ID:1101510167-1"},
 CellID->1779125069],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1101510167-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1577134591]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lcm 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lcm 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["s", "ExpressionVariable"], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
           RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["s", "ExpressionVariable"], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
           RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:571674285"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p571674285-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p571674285-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:571674285", "Proof|ID:571674285-1"},
 CellID->348779623],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p571674285-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1315821106]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lcm 1')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lcm 1')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["t", "ExpressionVariable"], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
           RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["t", "ExpressionVariable"], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
           RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:775675057"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p775675057-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p775675057-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:775675057", "Proof|ID:775675057-1"},
 CellID->1055462613],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p775675057-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1433558380]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lcm 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lcm 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
             StyleBox[
             "u", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
             StyleBox[
             "u", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:517355459"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p517355459-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p517355459-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:517355459", "Proof|ID:517355459-1"},
 CellID->1760053657],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p517355459-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->827965406]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lcm 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lcm 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], ",", 
               StyleBox["s", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], ",", 
               StyleBox["s", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:651227563"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p651227563-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p651227563-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:651227563", "Proof|ID:651227563-1"},
 CellID->1015477485],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p651227563-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2032600889]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lcm 6)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lcm 6)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           StyleBox[
           "s", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           StyleBox[
           "s", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:482943675"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p482943675-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p482943675-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:482943675", "Proof|ID:482943675-1"},
 CellID->29218054],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p482943675-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->48707691]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lcm 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lcm 4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}]}], 
             TagBox[")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}]}], 
             TagBox[
             ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:39860588"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p39860588-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p39860588-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:39860588", "Proof|ID:39860588-1"},
 CellID->2016600132],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p39860588-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->912021378]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lcm 4')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lcm 4')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}]}], 
             TagBox[")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}]}], 
             TagBox[
             ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1631465634"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1631465634-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1631465634-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1631465634", "Proof|ID:1631465634-1"},
 CellID->1760483406],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1631465634-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1884546058]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lcm 5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lcm 5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                   RowBox[{
                    StyleBox["s", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["u", "ExpressionVariable"]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                   RowBox[{
                    StyleBox["s", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["u", "ExpressionVariable"]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:835022497"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p835022497-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p835022497-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:835022497", "Proof|ID:835022497-1"},
 CellID->248308182],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p835022497-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->760468064]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Domain Membership (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.635660556156398*^9, 3.635660559568552*^9}, {
  3.6386944311749077`*^9, 3.6386944366372633`*^9}},
 CellID->265493105],

Cell["", "OpenEnvironment",
 CellID->17942811],

Cell[CellGroupData[{

Cell["Lemma: supp", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656132318470477*^9, 3.6561323231972857`*^9}},
 CellID->935061110],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"\[ScriptCapitalD]", ",", "p"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 100863090}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6561323275964932`*^9, 3.656132329686897*^9}, {
  3.656132381806588*^9, 3.6561323819625883`*^9}},
 CellTags->{
  "ID:100863090", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->100863090],

Cell[TextData[{
 "Note that formulas (membership supp) and (\[Element] supp) are stated for \
",
 StyleBox["arbitrary",
  FontSlant->"Italic"],
 " \[ScriptCapitalD] and ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.638589480985079*^9, 3.638589497932667*^9}, {
  3.65613240558103*^9, 3.6561324303538733`*^9}},
 CellID->634130198],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainSets", "[", "\[ScriptCapitalT]", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["supp", "\[ScriptCapitalD]"], "[", "p", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership supp general)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:292726431", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656132344366523*^9, 3.6561323480793295`*^9}, 
   3.656132385020194*^9},
 CellTags->{
  "ID:292726431", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership supp general"},
 CellID->292726431],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "t"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"t", "\[Element]", 
      RowBox[{
       UnderscriptBox["supp", "\[ScriptCapitalD]"], "[", "p", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "t", "]"}]}], "\[And]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalD]"], "[", 
        RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] supp general)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:109557753", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656132353632939*^9, 3.656132386221396*^9}, {
  3.656132436999485*^9, 3.656132443036696*^9}},
 CellTags->{
  "ID:109557753", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Element] supp general"},
 CellID->109557753],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->302285674]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->107322010],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership supp general)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership supp general)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
            StyleBox["p", "ExpressionVariable"]}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainSets", "[", "\[ScriptCapitalT]", "]"}]], "[", 
           RowBox[{
             UnderscriptBox["supp", 
              StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
            StyleBox["p", "ExpressionVariable"]}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainSets", "[", "\[ScriptCapitalT]", "]"}]], "[", 
           RowBox[{
             UnderscriptBox["supp", 
              StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:292726431"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p292726431-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p292726431-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:292726431", "Proof|ID:292726431-1"},
 CellID->837210885],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p292726431-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->225362927]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] supp general)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] supp general)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
            StyleBox["p", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["supp", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                 "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
            StyleBox["p", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["supp", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                 "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:109557753"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p109557753-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p109557753-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:109557753", "Proof|ID:109557753-1"},
 CellID->474399008],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p109557753-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->369676367]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->290097701],

Cell[CellGroupData[{

Cell["\<\
Lemma: Domain-Membership of Polynomial Constructors and Selectors I\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638528874371828*^9, 3.6385288759242563`*^9}},
 CellID->227405523],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "p"], 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "\[ScriptCapitalT]", "]"}]], 
   RowBox[{"[", 
    RowBox[{
     UnderscriptBox["supp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership supp)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:938132667", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6385168353632803`*^9, 3.638516859447204*^9}, {
   3.638528928465498*^9, 3.6385289374383574`*^9}, 3.638528990945724*^9, {
   3.638589473430237*^9, 3.63858947448916*^9}, {3.6561322718575954`*^9, 
   3.6561322763816032`*^9}, 3.6561323345541053`*^9},
 CellTags->{
  "ID:938132667", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership supp", "IR"},
 CellID->938132667],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "p", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", "t"], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"t", "\[Element]", 
       RowBox[{
        UnderscriptBox["supp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "t", "]"}]}], "\[And]", 
       RowBox[{
        RowBox[{
         UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
         RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", 
        UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] supp)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:56177686", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.653213867035458*^9, 3.65321392457176*^9}},
 CellTags->{
  "ID:56177686", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Element] supp", "IR"},
 CellID->56177686],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   UnderscriptBox["0", "\[ScriptCapitalP]"], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:163445395", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.635843953411738*^9, 3.635843964578676*^9}},
 CellTags->{
  "ID:163445395", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 0", "IR"},
 CellID->163445395],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   UnderscriptBox["1", "\[ScriptCapitalP]"], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:671805226", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6358439707504683`*^9, 3.63584397222519*^9}},
 CellTags->{
  "ID:671805226", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 1", "IR"},
 CellID->671805226],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "p", "]"}]}], ",", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "t"}], "]"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership C)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:373967346", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63566068228487*^9, 3.6356607135001802`*^9}, {
  3.638528995629833*^9, 3.638529003170464*^9}, {3.638589456974554*^9, 
  3.638589466001133*^9}},
 CellTags->{
  "ID:373967346", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership C", "IR"},
 CellID->373967346],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->929112994]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->618819632],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership supp)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership supp)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["p", "ExpressionVariable"]}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainSets", "[", "\[ScriptCapitalT]", "]"}]], "[", 
           RowBox[{
             UnderscriptBox["supp", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["p", "ExpressionVariable"]}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainSets", "[", "\[ScriptCapitalT]", "]"}]], "[", 
           RowBox[{
             UnderscriptBox["supp", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:938132667"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p938132667-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p938132667-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:938132667", "Proof|ID:938132667-1"},
 CellID->22005555],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p938132667-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->298479159]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] supp)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] supp)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             StyleBox["t", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["supp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                 "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             StyleBox["t", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["supp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                 "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:56177686"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p56177686-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p56177686-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:56177686", "Proof|ID:56177686-1"},
 CellID->520191277],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p56177686-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->406632049]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
         UnderscriptBox["0", "\[ScriptCapitalP]"], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
         UnderscriptBox["0", "\[ScriptCapitalP]"], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:163445395"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p163445395-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p163445395-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:163445395", "Proof|ID:163445395-1"},
 CellID->25655089],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p163445395-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->25955964]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
         UnderscriptBox["1", "\[ScriptCapitalP]"], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
         UnderscriptBox["1", "\[ScriptCapitalP]"], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:671805226"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p671805226-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p671805226-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:671805226", "Proof|ID:671805226-1"},
 CellID->637038652],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p671805226-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->46850605]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership C)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership C)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:373967346"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p373967346-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p373967346-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:373967346", "Proof|ID:373967346-1"},
 CellID->278031390],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p373967346-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->27619205]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->118902962],

Cell[CellGroupData[{

Cell["\<\
Lemma: Alternative Descriptions of Implicitly Defined Functions\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6385169420863943`*^9, 3.638516955338768*^9}, {
  3.638520316791*^9, 3.6385203169396477`*^9}, {3.638520434764839*^9, 
  3.638520441657827*^9}, {3.642330205236817*^9, 3.6423302125770187`*^9}},
 CellID->999616402],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
     RowBox[{"[", "c", "]"}]}], ",", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], "\[Equal]", 
    RowBox[{
     UnderscriptBox["the", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", "q", "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
         RowBox[{"[", "s", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
          RowBox[{"q", ",", "s"}], "]"}], "\[Equal]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {"c", "\[DoubleLeftArrow]", 
             RowBox[{"s", "\[Equal]", "t"}]},
            {
             UnderscriptBox["0", "\[ScriptCapitalR]"], "\[DoubleLeftArrow]", 
             "otherwise"}
           }]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Monomial)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1248732928", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631595152612241*^9, 3.631595297998638*^9}, {
   3.631595496227182*^9, 3.631595502106477*^9}, {3.631598508127692*^9, 
   3.6315985109959497`*^9}, {3.631598998377282*^9, 3.631598998663959*^9}, {
   3.631600418143136*^9, 3.6316004304268513`*^9}, {3.631603800407992*^9, 
   3.6316038010452023`*^9}, {3.632716886367832*^9, 3.632716940792967*^9}, 
   3.638511183963632*^9, {3.638520381326231*^9, 3.6385203816710443`*^9}, {
   3.6422259730119133`*^9, 3.642225973984022*^9}, {3.642226235666766*^9, 
   3.642226293166408*^9}, {3.642230219747979*^9, 3.6422302476557827`*^9}, {
   3.642232910829412*^9, 3.642232929079939*^9}, {3.642233024475779*^9, 
   3.642233034679278*^9}, {3.642233662822236*^9, 3.6422336703464746`*^9}},
 CellTags->{
  "ID:1248732928", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Monomial"},
 CellID->1248732928],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "p", "]"}]}]], 
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["isLP", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "t"}], "]"}], "\[And]", 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "t0", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["isLP", "\[ScriptCapitalP]"], "[", 
         RowBox[{"p", ",", "t0"}], "]"}], "\[Implies]", 
        RowBox[{"t0", "\[Equal]", "t"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lp sufficient)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1277602733", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63851701453909*^9, 3.638517138966625*^9}, {
   3.638520261653719*^9, 3.638520290086054*^9}, 3.638520458008469*^9, {
   3.638594630592435*^9, 3.638594646071274*^9}, {3.638595582261993*^9, 
   3.638595583522954*^9}, {3.6422352729758787`*^9, 3.642235283140382*^9}, {
   3.6422354052274723`*^9, 3.6422354092595453`*^9}, {3.642236081210518*^9, 
   3.642236083107367*^9}},
 CellTags->{
  "ID:1277602733", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp sufficient"},
 CellID->1277602733],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "p", "]"}]}]], 
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]", 
    RowBox[{
     UnderscriptBox["the", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "t", "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", 
            UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
          {
           RowBox[{
            UnderscriptBox[
             UnderscriptBox["\[ForAll]", 
              RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
               RowBox[{"[", "s", "]"}]}]], 
             RowBox[{"s", 
              UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]], 
            RowBox[{
             RowBox[{
              UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
              RowBox[{"p", ",", "s"}], "]"}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}]}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lp alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1270225306", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6422351519555817`*^9, 3.642235238395301*^9}, {
  3.642235271598975*^9, 3.642235278443583*^9}, {3.642237419263246*^9, 
  3.642237520172617*^9}},
 CellTags->{
  "ID:1270225306", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp alternative"},
 CellID->1270225306],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->702468881]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->924974006],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Monomial)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Monomial)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], 
             UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
             StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["the", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{"\[Piecewise]", 
                   GridBox[{{
                    StyleBox["c", "ExpressionVariable"], "\[DoubleLeftArrow]", 
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["t", "ExpressionVariable"]}]}, {
                    UnderscriptBox["0", "\[ScriptCapitalR]"], 
                    "\[DoubleLeftArrow]", 
                    "True"}}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], 
             UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
             StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["the", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{"\[Piecewise]", 
                   GridBox[{{
                    StyleBox["c", "ExpressionVariable"], "\[DoubleLeftArrow]", 
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["t", "ExpressionVariable"]}]}, {
                    UnderscriptBox["0", "\[ScriptCapitalR]"], 
                    "\[DoubleLeftArrow]", 
                    "True"}}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1248732928"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1248732928-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1248732928-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1248732928", "Proof|ID:1248732928-1"},
 CellID->328071702],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1248732928-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->264893878]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lp sufficient)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lp sufficient)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isLP", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t0", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["isLP", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t0", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    StyleBox["t0", "ExpressionVariable"], "\[Equal]", 
                    StyleBox[
                    "t", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isLP", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t0", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["isLP", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t0", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    StyleBox["t0", "ExpressionVariable"], "\[Equal]", 
                    StyleBox[
                    "t", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1277602733"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1277602733-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1277602733-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1277602733", "Proof|ID:1277602733-1"},
 CellID->88373483],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1277602733-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->287665210]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lp alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lp alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["the", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                   "\[NotEqual]", 
                   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    
                    UnderscriptBox[
                    "0", "\[ScriptCapitalR]"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["the", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                   "\[NotEqual]", 
                   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    
                    UnderscriptBox[
                    "0", "\[ScriptCapitalR]"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1270225306"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1270225306-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1270225306-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1270225306", "Proof|ID:1270225306-1"},
 CellID->426200985],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1270225306-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->939653437]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1434901757],

Cell[CellGroupData[{

Cell["\<\
Lemma: Domain-Membership of Polynomial Constructors and Selectors II\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6356605974198236`*^9, 3.635660619598592*^9}, {
  3.635661107289064*^9, 3.635661123201601*^9}, {3.638528961458709*^9, 
  3.638528961940086*^9}},
 CellID->1095667578],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}], ",", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
     RowBox[{"[", "c", "]"}]}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership monomials)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:292606918", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.635660779015863*^9, 3.63566080275115*^9}, {
  3.6561321673988113`*^9, 3.6561321745748243`*^9}},
 CellTags->{
  "ID:292606918", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomials", "IR"},
 CellID->292606918],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "p", "]"}]}]], 
  RowBox[{"p", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 485824967}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.635660626988741*^9, 3.635660636645981*^9}, {
  3.642308924461946*^9, 3.642308929812014*^9}},
 CellTags->{
  "ID:485824967", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->485824967],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership lp)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:406494450", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63566068228487*^9, 3.6356607135001802`*^9}, {
  3.635660854084723*^9, 3.635660858816716*^9}},
 CellTags->{
  "ID:406494450", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lp", "IR"},
 CellID->406494450],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership lc)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1454724371", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63566068228487*^9, 3.6356607135001802`*^9}, {
  3.635660854084723*^9, 3.6356608716995373`*^9}},
 CellTags->{
  "ID:1454724371", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lc", "IR"},
 CellID->1454724371],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership lm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1589530505", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63566068228487*^9, 3.635660695752953*^9}, {
  3.6356608324602633`*^9, 3.6356608348155003`*^9}},
 CellTags->{
  "ID:1589530505", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lm", "IR"},
 CellID->1589530505],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1979542526]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->561932109],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership monomials)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership monomials)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], 
             UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
             StyleBox["t", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], 
             UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
             StyleBox["t", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:292606918"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p292606918-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p292606918-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:292606918", "Proof|ID:292606918-1"},
 CellID->1053213110],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p292606918-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->303879011]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership lp)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership lp)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:406494450"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p406494450-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p406494450-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:406494450", "Proof|ID:406494450-1"},
 CellID->928711222],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p406494450-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1865842282]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership lc)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership lc)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1454724371"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1454724371-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1454724371-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1454724371", "Proof|ID:1454724371-1"},
 CellID->776114793],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1454724371-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1327093525]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership lm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership lm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1589530505"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1589530505-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1589530505-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1589530505", "Proof|ID:1589530505-1"},
 CellID->1436300644],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1589530505-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->680527277]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Selectors (C, H, lp, lc, lm) (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.6397275339603453`*^9, 3.639727547710773*^9}, {
  3.639827595645022*^9, 3.639827599943*^9}},
 CellID->1344153722],

Cell["", "OpenEnvironment",
 CellID->260708809],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.639727559639658*^9, 3.63972757004955*^9}},
 CellID->603207502],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "t", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 720209465}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6331720878606453`*^9, 3.6331720953266897`*^9}},
 CellTags->{
  "ID:720209465", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->720209465],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    UnderscriptBox["0", "\[ScriptCapitalP]"], ",", "t"}], "]"}], "\[Equal]", 
  UnderscriptBox["0", "\[ScriptCapitalR]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1852885314", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63963055130856*^9, 3.639630575473692*^9}},
 CellTags->{
  "ID:1852885314", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 0"},
 CellID->1852885314],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", ",", "q"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 148341425}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633172318481805*^9, 3.633172326769643*^9}},
 CellTags->{
  "ID:148341425", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->148341425],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["+", "\[ScriptCapitalP]"], "q"}], ",", "t"}], "]"}], 
  "\[Equal]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}], 
   UnderscriptBox["+", "\[ScriptCapitalR]"], 
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"q", ",", "t"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1443097690", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633172208425226*^9, 3.633172234166685*^9}, {
  3.633172348197496*^9, 3.6331723511779413`*^9}, {3.6331726021286077`*^9, 
  3.633172603454072*^9}},
 CellTags->{
  "ID:1443097690", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 1"},
 CellID->1443097690],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], ",", "t"}], "]"}], 
  "\[Equal]", 
  RowBox[{
   UnderscriptBox["-", "\[ScriptCapitalR]"], 
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1675126992", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633172582691614*^9, 3.633172600586095*^9}},
 CellTags->{
  "ID:1675126992", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 2"},
 CellID->1675126992],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["-", "\[ScriptCapitalP]"], "q"}], ",", "t"}], "]"}], 
  "\[Equal]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}], 
   UnderscriptBox["-", "\[ScriptCapitalR]"], 
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"q", ",", "t"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:814027629", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6331722436862497`*^9, 3.633172264006868*^9}, {
  3.6331723525771103`*^9, 3.633172355631184*^9}, {3.633172606428228*^9, 
  3.633172607686426*^9}},
 CellTags->{
  "ID:814027629", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 3"},
 CellID->814027629],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
    RowBox[{"t", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
   UnderscriptBox["0", "\[ScriptCapitalR]"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1597798077", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633172878785657*^9, 3.6331729203141327`*^9}, 
   3.635747321668235*^9},
 CellTags->{
  "ID:1597798077", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 4"},
 CellID->1597798077],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"t", "\[Equal]", 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], "\[Or]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C 5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1711645326", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6331762422435493`*^9, 3.633176325360405*^9}, 
   3.635747371429741*^9, {3.638784555928823*^9, 3.638784567334503*^9}},
 CellTags->{
  "ID:1711645326", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 5"},
 CellID->1711645326],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->320679728]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->514407804],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["0", "\[ScriptCapitalP]"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           UnderscriptBox[
           "0", "\[ScriptCapitalR]"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["0", "\[ScriptCapitalP]"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           UnderscriptBox[
           "0", "\[ScriptCapitalR]"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1852885314"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1852885314-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1852885314-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1852885314", "Proof|ID:1852885314-1"},
 CellID->131831930],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1852885314-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1038146]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["+", "\[ScriptCapitalR]"], 
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["+", "\[ScriptCapitalR]"], 
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1443097690"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1443097690-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1443097690-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1443097690", "Proof|ID:1443097690-1"},
 CellID->324739814],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1443097690-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->58511154]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["Minus", "\[ScriptCapitalR]"], "[", 
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["Minus", "\[ScriptCapitalR]"], "[", 
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1675126992"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1675126992-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1675126992-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1675126992", "Proof|ID:1675126992-1"},
 CellID->947152217],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1675126992-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->450219457]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["-", "\[ScriptCapitalR]"], 
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["-", "\[ScriptCapitalR]"], 
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:814027629"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p814027629-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p814027629-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:814027629", "Proof|ID:814027629-1"},
 CellID->1077213],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p814027629-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->153578015]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C 4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1597798077"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1597798077-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1597798077-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1597798077", "Proof|ID:1597798077-1"},
 CellID->1370021543],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1597798077-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->320424180]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C 5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C 5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[NotEqual]", 
             
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox[
                 "t", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[NotEqual]", 
             
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox[
                 "t", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1711645326"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1711645326-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1711645326-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1711645326", "Proof|ID:1711645326-1"},
 CellID->209179501],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1711645326-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->599442953]
}, Closed]],

Cell["The following lemma is needed already here:", "Text",
 CellChangeTimes->{{3.642239039126967*^9, 3.6422390465815*^9}},
 CellID->1514263990],

Cell["", "OpenEnvironment",
 CellID->973491128],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["-", TraditionalForm]]],
 " are Closed in \[ScriptCapitalP]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6422390174138317`*^9, 3.642239031697575*^9}},
 CellID->2040579658],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", ",", "q"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1053827266}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642239055011547*^9, 3.642239063165913*^9}},
 CellTags->{
  "ID:1053827266", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1053827266],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   RowBox[{"p", 
    UnderscriptBox["+", "\[ScriptCapitalP]"], "q"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership + \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:829207957", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642239066900351*^9, 3.642239078640647*^9}},
 CellTags->{
  "ID:829207957", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership + \[ScriptCapitalP]", "IR"},
 CellID->829207957],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership - \[ScriptCapitalP] 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:227196886", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6396332563403788`*^9, 3.639633271211821*^9}},
 CellTags->{
  "ID:227196886", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership - \[ScriptCapitalP] 1", "IR"},
 CellID->227196886],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   RowBox[{"p", 
    UnderscriptBox["-", "\[ScriptCapitalP]"], "q"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership - \[ScriptCapitalP] 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:414658623", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63973229306313*^9, 3.639732303103311*^9}},
 CellTags->{
  "ID:414658623", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership - \[ScriptCapitalP] 2", "IR"},
 CellID->414658623],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->525227380]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->838178864],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership + \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership + \[ScriptCapitalP])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["+", "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["+", "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:829207957"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p829207957-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p829207957-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:829207957", "Proof|ID:829207957-1"},
 CellID->188209562],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p829207957-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->353588127]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership - \[ScriptCapitalP] 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership - \[ScriptCapitalP] 1)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
           RowBox[{
             UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
           RowBox[{
             UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:227196886"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p227196886-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p227196886-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:227196886", "Proof|ID:227196886-1"},
 CellID->65751116],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p227196886-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->18198817]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership - \[ScriptCapitalP] 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership - \[ScriptCapitalP] 2)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:414658623"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p414658623-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p414658623-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:414658623", "Proof|ID:414658623-1"},
 CellID->19838666],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p414658623-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->879352583]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->2126627650],

Cell[CellGroupData[{

Cell["Lemma: Properties of lp, lc, lm", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.639737628388962*^9, 3.639737633842175*^9}},
 CellID->826776809],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["1", "\[ScriptCapitalP]"], "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["1", "\[ScriptCapitalR]"], "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1=0 \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:314696779", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6397214605868483`*^9, 3.63972148593314*^9}},
 CellTags->{
  "ID:314696779", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1=0 \[ScriptCapitalP]"},
 CellID->314696779],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
     UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
    UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lp 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:849229881", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633085848284401*^9, 3.633085853808901*^9}, {
  3.639713076970004*^9, 3.6397131019409533`*^9}},
 CellTags->{
  "ID:849229881", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 1"},
 CellID->849229881],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
     UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
    UnderscriptBox["1", "\[ScriptCapitalR]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lc 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:406483636", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6330858048663597`*^9, 3.633085821257195*^9}, {
  3.6397131053969603`*^9, 3.6397131153079777`*^9}},
 CellTags->{
  "ID:406483636", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 1"},
 CellID->406483636],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
     UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
    UnderscriptBox["1", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lm 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1767085150", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639912539120891*^9, 3.6399125457662163`*^9}},
 CellTags->{
  "ID:1767085150", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm 1"},
 CellID->1767085150],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "p", "]"}]}]], 
  RowBox[{"p", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1272860149}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631872855092862*^9, 3.631872862624887*^9}, {
   3.633172544491692*^9, 3.633172552482078*^9}, 3.635747420093211*^9},
 CellTags->{
  "ID:1272860149", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1272860149],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], "\[NotEqual]", 
  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(non-zero 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:144989886", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64224030104535*^9, 3.64224032033871*^9}, {
  3.642240374012492*^9, 3.642240382627348*^9}},
 CellTags->{
  "ID:144989886", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "non-zero 2", "IR"},
 CellID->144989886],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lp 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2138850160", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631872873302041*^9, 3.6318729277115088`*^9}, {
  3.6318736395189447`*^9, 3.631873647691662*^9}},
 CellTags->{
  "ID:2138850160", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 2"},
 CellID->2138850160],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["-", "\[ScriptCapitalR]"], 
   RowBox[{
    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lc 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1145306026", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.6318736428805237`*^9},
 CellTags->{
  "ID:1145306026", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 2"},
 CellID->1145306026],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[NotEqual]", 
  
  UnderscriptBox["0", "\[ScriptCapitalR]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lc 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1442316125", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631873655043769*^9, 3.6318736644956827`*^9}},
 CellTags->{
  "ID:1442316125", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 3"},
 CellID->1442316125],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"s", ",", "t"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"s", 
       UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}], "\[And]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"s", "\[Equal]", 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
        "\[And]", 
        RowBox[{"t", "\[Equal]", 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Or]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "s"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lp 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:655883205", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633234545609248*^9, 3.633234654414833*^9}, {
  3.639202594495139*^9, 3.639202594644993*^9}},
 CellTags->{
  "ID:655883205", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 3"},
 CellID->655883205],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "q", "]"}]}]], 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
    "\[Implies]", 
    RowBox[{
     RowBox[{"p", 
      UnderscriptBox["+", "\[ScriptCapitalP]"], "q"}], "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(non-zero 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:996448533", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642242068552684*^9, 3.6422420844846287`*^9}},
 CellTags->{
  "ID:996448533", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "non-zero 3", "IR"},
 CellID->996448533],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "q", "]"}]}]], 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
    "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", 
        UnderscriptBox["+", "\[ScriptCapitalP]"], "q"}], "]"}], "\[Equal]", 
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lp 5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1383271844", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641094940202338*^9, 3.6410950090985003`*^9}, {
  3.64109507825945*^9, 3.641095086016996*^9}, {3.641095211900889*^9, 
  3.64109523414321*^9}},
 CellTags->{
  "ID:1383271844", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 5"},
 CellID->1383271844],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "t", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 945941808}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640582160928388*^9, 3.640582170158101*^9}},
 CellTags->{
  "ID:945941808", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->945941808],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", "t"}], 
   "]"}], "\[Equal]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{
        UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
       "\[DoubleLeftArrow]", 
       RowBox[{"t", "\[Equal]", 
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]},
      {
       UnderscriptBox["0", "\[ScriptCapitalR]"], "\[DoubleLeftArrow]", 
       "otherwise"}
     }]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C lm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:857665826", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639820783869073*^9, 3.63982084498384*^9}, {
  3.640582173468606*^9, 3.6405821769781637`*^9}},
 CellTags->{
  "ID:857665826", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C lm"},
 CellID->857665826],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", "t"}],
      "]"}], "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lm 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1411816658", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640512682891914*^9, 3.640512822903474*^9}, {
  3.640578734369874*^9, 3.640578804406723*^9}},
 CellTags->{
  "ID:1411816658", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm 2"},
 CellID->1411816658],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1163861425]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1926885900],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1=0 \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1=0 \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalP]"], "\[Equal]", 
           UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
         "\[DoubleLeftRightArrow]", 
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], "\[Equal]", 
           UnderscriptBox[
           "0", "\[ScriptCapitalR]"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalP]"], "\[Equal]", 
           UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
         "\[DoubleLeftRightArrow]", 
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], "\[Equal]", 
           UnderscriptBox[
           "0", "\[ScriptCapitalR]"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:314696779"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p314696779-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p314696779-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:314696779", "Proof|ID:314696779-1"},
 CellID->817945206],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p314696779-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1429481448]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lp 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lp 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
         RowBox[{
           RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
             UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
           UnderscriptBox[
           "1", "\[ScriptCapitalT]"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
         RowBox[{
           RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
             UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
           UnderscriptBox[
           "1", "\[ScriptCapitalT]"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:849229881"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p849229881-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p849229881-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:849229881", "Proof|ID:849229881-1"},
 CellID->766889252],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p849229881-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1786903669]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lc 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lc 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
         RowBox[{
           RowBox[{
             UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
             UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
           UnderscriptBox[
           "1", "\[ScriptCapitalR]"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
         RowBox[{
           RowBox[{
             UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
             UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
           UnderscriptBox[
           "1", "\[ScriptCapitalR]"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:406483636"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p406483636-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p406483636-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:406483636", "Proof|ID:406483636-1"},
 CellID->1626957666],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p406483636-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1255490528]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lm 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lm 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
         RowBox[{
           RowBox[{
             UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
             UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
           UnderscriptBox[
           "1", "\[ScriptCapitalP]"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
         RowBox[{
           RowBox[{
             UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
             UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
           UnderscriptBox[
           "1", "\[ScriptCapitalP]"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1767085150"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1767085150-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1767085150-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1767085150", "Proof|ID:1767085150-1"},
 CellID->450159143],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1767085150-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1165157523]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(non-zero 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(non-zero 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:144989886"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p144989886-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p144989886-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:144989886", "Proof|ID:144989886-1"},
 CellID->914771064],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p144989886-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1590090525]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lp 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lp 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2138850160"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2138850160-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2138850160-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2138850160", "Proof|ID:2138850160-1"},
 CellID->1795169377],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2138850160-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->683454739]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lc 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lc 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               UnderscriptBox["Minus", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               UnderscriptBox["Minus", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1145306026"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1145306026-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1145306026-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1145306026", "Proof|ID:1145306026-1"},
 CellID->109823534],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1145306026-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1039693341]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lc 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lc 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1442316125"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1442316125-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1442316125-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1442316125", "Proof|ID:1442316125-1"},
 CellID->1835069754],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1442316125-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1340501902]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lp 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lp 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], 
               "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "\[And]", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}], "\[Or]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox[
               "s", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], 
               "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "\[And]", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}], "\[Or]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox[
               "s", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:655883205"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p655883205-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p655883205-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:655883205", "Proof|ID:655883205-1"},
 CellID->489542376],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p655883205-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->634307408]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(non-zero 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(non-zero 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:996448533"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p996448533-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p996448533-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:996448533", "Proof|ID:996448533-1"},
 CellID->925906148],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p996448533-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->68221894]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lp 5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lp 5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1383271844"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1383271844-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1383271844-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1383271844", "Proof|ID:1383271844-1"},
 CellID->1411486137],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1383271844-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1721706305]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C lm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C lm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                  "\[DoubleLeftArrow]", 
                  RowBox[{
                    StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
                  UnderscriptBox["0", "\[ScriptCapitalR]"], 
                  "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                  "\[DoubleLeftArrow]", 
                  RowBox[{
                    StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
                  UnderscriptBox["0", "\[ScriptCapitalR]"], 
                  "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:857665826"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p857665826-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p857665826-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:857665826", "Proof|ID:857665826-1"},
 CellID->606096953],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p857665826-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1744415237]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lm 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lm 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1411816658"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1411816658-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1411816658-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1411816658", "Proof|ID:1411816658-1"},
 CellID->352726063],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1411816658-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->644334913]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Monomials (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.63981943966549*^9, 3.639819441091462*^9}, {
  3.6399059771253433`*^9, 3.639905981567854*^9}},
 CellID->2098369935],

Cell["", "OpenEnvironment",
 CellID->633845652],

Cell[CellGroupData[{

Cell["Lemma: Properties of the Monomial Constructor", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.639819452011037*^9, 3.639819462004135*^9}, {
  3.639820656104785*^9, 3.63982065746954*^9}, {3.639827741556573*^9, 
  3.639827746808197*^9}},
 CellID->1224489222],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["1", "\[ScriptCapitalR]"], 
   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
   UnderscriptBox["1", "\[ScriptCapitalT]"]}], "\[Equal]", 
  UnderscriptBox["1", "\[ScriptCapitalP]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1\[CenterDot]1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1368691869", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650698006503509*^9, 3.650698024036457*^9}},
 CellTags->{
  "ID:1368691869", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1\[CenterDot]1"},
 CellID->1368691869],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{"c", ",", "d"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1325102956}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320401373168383`*^9, 3.632040146353407*^9}},
 CellTags->{
  "ID:1325102956", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1325102956],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "t", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1663774274}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632040149780904*^9, 3.632040157057767*^9}},
 CellTags->{
  "ID:1663774274", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1663774274],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
   RowBox[{"c", 
    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"c", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial \[CenterDot])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:941688542", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632817304153054*^9, 3.632817327252233*^9}},
 CellTags->{
  "ID:941688542", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial \[CenterDot]"},
 CellID->941688542],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"c", "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:91567084", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639115860229752*^9, 3.639115884459859*^9}},
 CellTags->{
  "ID:91567084", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 0"},
 CellID->91567084],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ",", "t"}], 
   "]"}], "\[Equal]", "c"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:420812829", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320401681401052`*^9, 3.632040187247573*^9}, {
  3.632717122161075*^9, 3.632717127609086*^9}},
 CellTags->{
  "ID:420812829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 3"},
 CellID->420812829],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "s", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{
        RowBox[{"c", 
         UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ",", 
        "s"}], "]"}], "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"s", "\[Equal]", "t"}], "\[And]", 
      RowBox[{"c", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:138547187", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633077384256331*^9, 3.6330774307070227`*^9}, {
  3.633174955867515*^9, 3.633174963663796*^9}},
 CellTags->{
  "ID:138547187", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 4"},
 CellID->138547187],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], "\[Equal]", 
    RowBox[{"d", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"c", "\[Equal]", "d"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 11)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:457378178", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6381644367247334`*^9, 3.638164468550906*^9}, {
  3.6381645043568172`*^9, 3.6381645063766727`*^9}},
 CellTags->{
  "ID:457378178", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 11"},
 CellID->457378178],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "s", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"c", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "s"}], "\[Equal]", 
      RowBox[{"d", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"c", "\[Equal]", "d"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"c", "\[Equal]", 
          UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[Or]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"s", "\[Equal]", "t"}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 11')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:926082604", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650705160715025*^9, 3.650705167707655*^9}, {
  3.6507052060200357`*^9, 3.6507052336479177`*^9}},
 CellTags->{
  "ID:926082604", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 11'"},
 CellID->926082604],

Cell[BoxData[
 RowBox[{
  RowBox[{"c", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
       RowBox[{"c", 
        UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], "]"}], 
      "\[Equal]", "t"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
       RowBox[{"c", 
        UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], "]"}], 
      "\[Equal]", "c"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] lp lc)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1344734857", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638702239536167*^9, 3.638702288874838*^9}},
 CellTags->{
  "ID:1344734857", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] lp lc"},
 CellID->1344734857],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1125130303]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->12089368],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1\[CenterDot]1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1\[CenterDot]1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], 
           UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
           UnderscriptBox["1", "\[ScriptCapitalT]"]}], "\[Equal]", 
         UnderscriptBox[
         "1", "\[ScriptCapitalP]"]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], 
           UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
           UnderscriptBox["1", "\[ScriptCapitalT]"]}], "\[Equal]", 
         UnderscriptBox[
         "1", "\[ScriptCapitalP]"]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1368691869"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1368691869-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1368691869-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1368691869", "Proof|ID:1368691869-1"},
 CellID->660760468],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1368691869-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->417821554]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial \[CenterDot])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial \[CenterDot])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:941688542"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p941688542-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p941688542-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:941688542", "Proof|ID:941688542-1"},
 CellID->1365249879],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p941688542-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1779137777]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:91567084"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p91567084-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p91567084-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:91567084", "Proof|ID:91567084-1"},
 CellID->1862161014],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p91567084-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->963217094]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           StyleBox[
           "c", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           StyleBox[
           "c", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:420812829"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p420812829-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p420812829-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:420812829", "Proof|ID:420812829-1"},
 CellID->2092842442],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p420812829-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->875416490]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], ",", 
                 StyleBox["s", "ExpressionVariable"]}], "]"}], "\[NotEqual]", 
             
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["t", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], ",", 
                 StyleBox["s", "ExpressionVariable"]}], "]"}], "\[NotEqual]", 
             
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["t", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:138547187"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p138547187-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p138547187-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:138547187", "Proof|ID:138547187-1"},
 CellID->244847721],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p138547187-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->2024151403]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 11)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 11)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "d", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "d", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:457378178"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p457378178-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p457378178-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:457378178", "Proof|ID:457378178-1"},
 CellID->783388721],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p457378178-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.650705360895359*^9},
 CellID->1798922519]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 11')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 11')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["s", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["d", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[Equal]", 
                   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Or]", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["t", "ExpressionVariable"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["s", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["d", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[Equal]", 
                   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Or]", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["t", "ExpressionVariable"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:926082604"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p926082604-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p926082604-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:926082604", "Proof|ID:926082604-1"},
 CellID->1364381745],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p926082604-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->659473488]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] lp lc)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] lp lc)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               StyleBox["t", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               StyleBox["t", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1344734857"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1344734857-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1344734857-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1344734857", "Proof|ID:1344734857-1"},
 CellID->670171654],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1344734857-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->565919770]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1286789207],

Cell[CellGroupData[{

Cell["Lemma: Properties of Monomials", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6398277550461206`*^9, 3.639827759139485*^9}},
 CellID->660836734],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
   UnderscriptBox["1", "\[ScriptCapitalP]"], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1558664548", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639712971347341*^9, 3.63971300305037*^9}},
 CellTags->{
  "ID:1558664548", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 1"},
 CellID->1558664548],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", 
   "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1835029788}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318723626051283`*^9, 3.6318723705880613`*^9}, {
  3.6318726676373873`*^9, 3.631872675929371*^9}},
 CellTags->{
  "ID:1835029788", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1835029788],

Cell[BoxData[
 RowBox[{"m", "\[NotEqual]", 
  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial \[NotEqual] 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:757450222", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631872388581476*^9, 3.631872409910788*^9}, 
   3.631872677784915*^9, {3.6327175432332773`*^9, 3.63271754358298*^9}, {
   3.63574769315108*^9, 3.635747694881022*^9}},
 CellTags->{
  "ID:757450222", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial \[NotEqual] 0"},
 CellID->757450222],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
  RowBox[{
   UnderscriptBox["-", "\[ScriptCapitalP]"], "m"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial -)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1978849375", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6396327553082447`*^9, 3.6396327703500433`*^9}, {
  3.639633485359784*^9, 3.639633487403408*^9}},
 CellTags->{
  "ID:1978849375", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial -"},
 CellID->1978849375],

Cell[BoxData[
 RowBox[{"m", "\[Equal]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}], 
   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
   RowBox[{
    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 3')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:283455673", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.637997457188142*^9, 3.637997476179541*^9}},
 CellTags->{
  "ID:283455673", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 3'"},
 CellID->283455673],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[Exists]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
     RowBox[{"[", "c", "]"}]}]], 
   RowBox[{"c", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"m", "\[Equal]", 
     RowBox[{"c", 
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:610787620", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6319566629592733`*^9, 3.6319566816247797`*^9}, {
  3.632717173048542*^9, 3.632717181476614*^9}, {3.63281720626969*^9, 
  3.6328172489465437`*^9}},
 CellTags->{
  "ID:610787620", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 3"},
 CellID->610787620],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"c", 
        UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
       "\[Equal]", "m"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"c", "\[Equal]", 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}]}], 
       "\[And]", 
       RowBox[{"t", "\[Equal]", 
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:980082706", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633174011412284*^9, 3.6331740498464622`*^9}, {
  3.6398279054863234`*^9, 3.6398279138703003`*^9}},
 CellTags->{
  "ID:980082706", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 4"},
 CellID->980082706],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "m", "]"}], "\[Equal]", 
  "m"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lm isMonomial)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1306217654", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639913176103849*^9, 3.6399131857019663`*^9}},
 CellTags->{
  "ID:1306217654", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm isMonomial"},
 CellID->1306217654],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"m", ",", "t"}], "]"}], "\[Equal]", 
      UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"t", "\[NotEqual]", 
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 7)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1645795930", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632729816590386*^9, 3.632729860212803*^9}, {
  3.6391148232274923`*^9, 3.639114860456644*^9}},
 CellTags->{
  "ID:1645795930", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 7"},
 CellID->1645795930],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", "p", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 691307310}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632717558517411*^9, 3.632717564848895*^9}},
 CellTags->{
  "ID:691307310", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->691307310],

Cell[BoxData[
 RowBox[{
  RowBox[{"p", "\[Equal]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Or]", 
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "p", "]"}], 
  "\[Or]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", 
        RowBox[{"s", ",", "t"}], "]"}]}]], 
     RowBox[{"s", "\[NotEqual]", "t"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{"p", ",", "s"}], "]"}], "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 8)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1901309763", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632717570965814*^9, 3.632717625748665*^9}},
 CellTags->{
  "ID:1901309763", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 8"},
 CellID->1901309763],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "t", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"t", "\[Equal]", 
        UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "p", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 9)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1177943318", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.636369129441812*^9, 3.636369227899067*^9}},
 CellTags->{
  "ID:1177943318", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 9"},
 CellID->1177943318],

Cell[BoxData[
 RowBox[{
  RowBox[{"p", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1770611523}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640322685829247*^9, 3.6403226906402473`*^9}},
 CellTags->{
  "ID:1770611523", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1770611523],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "p", "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}], 
      "\[Implies]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
         RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
        UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 12)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:670623560", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6506237988454514`*^9, 3.650623855823443*^9}},
 CellTags->{
  "ID:670623560", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 12"},
 CellID->670623560],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
  RowBox[{
   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial lm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1806313475", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631879617907675*^9, 3.631879630217128*^9}, {
   3.6398060426093187`*^9, 3.639806053559273*^9}, 3.6403226939406023`*^9},
 CellTags->{
  "ID:1806313475", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial lm"},
 CellID->1806313475],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], "]"}], 
  "\[Equal]", 
  RowBox[{
   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lc lm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1110032163", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640322699182686*^9, 3.640322718615443*^9}},
 CellTags->{
  "ID:1110032163", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc lm"},
 CellID->1110032163],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], "]"}], 
  "\[Equal]", 
  RowBox[{
   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lp lm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:632352665", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640322725320759*^9, 3.640322728970191*^9}},
 CellTags->{
  "ID:632352665", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp lm"},
 CellID->632352665],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1776835731]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->816964609],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
           UnderscriptBox["1", "\[ScriptCapitalP]"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
           UnderscriptBox["1", "\[ScriptCapitalP]"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1558664548"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1558664548-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1558664548-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1558664548", "Proof|ID:1558664548-1"},
 CellID->142537034],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1558664548-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->311072897]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial \[NotEqual] 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial \[NotEqual] 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:757450222"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p757450222-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p757450222-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:757450222", "Proof|ID:757450222-1"},
 CellID->1689557017],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p757450222-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1082713634]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial -)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial -)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1978849375"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1978849375-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1978849375-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1978849375", "Proof|ID:1978849375-1"},
 CellID->807800730],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1978849375-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2025444804]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 3')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 3')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:283455673"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p283455673-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p283455673-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:283455673", "Proof|ID:283455673-1"},
 CellID->1547205861],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p283455673-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->532363353]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox[
                   "t", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox[
                   "t", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:610787620"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p610787620-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p610787620-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:610787620", "Proof|ID:610787620-1"},
 CellID->1422344349],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p610787620-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1123805012]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
               StyleBox["m", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], "\[And]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
               StyleBox["m", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], "\[And]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:980082706"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p980082706-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p980082706-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:980082706", "Proof|ID:980082706-1"},
 CellID->1540912936],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p980082706-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1183327246]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lm isMonomial)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lm isMonomial)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[Equal]", 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[Equal]", 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1306217654"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1306217654-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1306217654-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1306217654", "Proof|ID:1306217654-1"},
 CellID->277734451],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1306217654-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->383104478]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 7)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 7)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1645795930"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1645795930-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1645795930-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1645795930", "Proof|ID:1645795930-1"},
 CellID->140680378],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1645795930-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->22238799]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 8)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 8)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Or]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], "\[Or]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
                StyleBox["t", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                 "\[NotEqual]", 
                 UnderscriptBox[
                 "0", "\[ScriptCapitalR]"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Or]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], "\[Or]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
                StyleBox["t", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                 "\[NotEqual]", 
                 UnderscriptBox[
                 "0", "\[ScriptCapitalR]"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1901309763"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1901309763-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1901309763-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1901309763", "Proof|ID:1901309763-1"},
 CellID->907132124],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1901309763-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->812486108]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 9)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 9)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1177943318"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1177943318-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1177943318-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1177943318", "Proof|ID:1177943318-1"},
 CellID->1237355862],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1177943318-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->458207366]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 12)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 12)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   StyleBox["t", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                   
                   UnderscriptBox[
                   "0", "\[ScriptCapitalR]"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   StyleBox["t", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                   
                   UnderscriptBox[
                   "0", "\[ScriptCapitalR]"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:670623560"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p670623560-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p670623560-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:670623560", "Proof|ID:670623560-1"},
 CellID->1341250136],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p670623560-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->837571024]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial lm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial lm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1806313475"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1806313475-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1806313475-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1806313475", "Proof|ID:1806313475-1"},
 CellID->347558245],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1806313475-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1954493899]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lc lm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lc lm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1110032163"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1110032163-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1110032163-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1110032163", "Proof|ID:1110032163-1"},
 CellID->916272762],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1110032163-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1999682617]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lp lm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lp lm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:632352665"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p632352665-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p632352665-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:632352665", "Proof|ID:632352665-1"},
 CellID->800412793],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p632352665-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1296416971]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " is Closed (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.6364404865905313`*^9, 3.636440511201828*^9}, {
  3.638510601958487*^9, 3.638510605544299*^9}, {3.639730300953288*^9, 
  3.639730305295951*^9}},
 CellID->1730043703],

Cell["", "OpenEnvironment",
 CellID->991999998],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " is Closed in \[ScriptCapitalP]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638521604301811*^9, 3.638521618170898*^9}},
 CellID->13198659],

Cell[BoxData[
 RowBox[{"isClosed", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["+", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalP]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ closed \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:423895762", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638521621599609*^9, 3.638521644368026*^9}},
 CellTags->{
  "ID:423895762", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ closed \[ScriptCapitalP]"},
 CellID->423895762],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1334309619]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->493608360],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ closed \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ closed \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isClosed", "[", 
         RowBox[{
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isClosed", "[", 
         RowBox[{
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:423895762"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p423895762-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p423895762-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:423895762", "Proof|ID:423895762-1"},
 CellID->355044739],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p423895762-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->650572530]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "0 is Neutral Element of ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.63851058671124*^9, 3.638510643366729*^9}, {
  3.6385172980410833`*^9, 3.638517298190033*^9}, {3.6397310739765253`*^9, 
  3.639731078176125*^9}},
 CellID->141138700],

Cell["", "OpenEnvironment",
 CellID->445386680],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: 0 is Neutral Element of ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " in \[ScriptCapitalP]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638521604301811*^9, 3.638521618170898*^9}, {
  3.6385217075444117`*^9, 3.6385217156387053`*^9}},
 CellID->1064502876],

Cell[BoxData[
 RowBox[{"isNeutral", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["+", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   UnderscriptBox["0", "\[ScriptCapitalP]"], ",", "\[ScriptCapitalP]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 neutral \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1812309492", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638521621599609*^9, 3.638521644368026*^9}, {
  3.638521688268338*^9, 3.6385216988074636`*^9}},
 CellTags->{
  "ID:1812309492", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "0 neutral \[ScriptCapitalP]"},
 CellID->1812309492],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->381592666]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1236557493],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 neutral \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 neutral \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isNeutral", "[", 
         RowBox[{
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           UnderscriptBox["0", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isNeutral", "[", 
         RowBox[{
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           UnderscriptBox["0", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1812309492"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1812309492-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1812309492-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1812309492", "Proof|ID:1812309492-1"},
 CellID->2145107533],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1812309492-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1403340967]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " is Associative (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.6364405254457273`*^9, 3.63644053197727*^9}, {
  3.63851061458115*^9, 3.638510618366023*^9}, {3.639731328930613*^9, 
  3.6397313328924503`*^9}},
 CellID->824892980],

Cell["", "OpenEnvironment",
 CellID->803927450],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " is Associative in \[ScriptCapitalP]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638521604301811*^9, 3.638521618170898*^9}, {
  3.638521720759623*^9, 3.638521722767658*^9}},
 CellID->1992468669],

Cell[BoxData[
 RowBox[{"isAssoc", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["+", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalP]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ assoc \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:595518277", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638521621599609*^9, 3.638521644368026*^9}, {
  3.638521728923695*^9, 3.6385217297844276`*^9}},
 CellTags->{
  "ID:595518277", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ assoc \[ScriptCapitalP]"},
 CellID->595518277],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1198061607]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1614517617],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ assoc \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ assoc \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isAssoc", "[", 
         RowBox[{
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isAssoc", "[", 
         RowBox[{
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:595518277"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p595518277-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p595518277-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:595518277", "Proof|ID:595518277-1"},
 CellID->1392319518],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p595518277-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1240957880]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " is Commutative (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.636440537569436*^9, 3.636440542443738*^9}, {
  3.638510620577332*^9, 3.6385106240392857`*^9}, {3.6397314437288923`*^9, 
  3.639731448293932*^9}},
 CellID->1752381855],

Cell["", "OpenEnvironment",
 CellID->1316847979],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " is Commutative in \[ScriptCapitalP]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638521604301811*^9, 3.638521618170898*^9}, {
  3.638521737152663*^9, 3.6385217392703114`*^9}},
 CellID->581912688],

Cell[BoxData[
 RowBox[{"isComm", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["+", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalP]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ comm \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2114498624", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638521621599609*^9, 3.638521644368026*^9}, {
  3.638521746494171*^9, 3.638521746843409*^9}},
 CellTags->{
  "ID:2114498624", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ comm \[ScriptCapitalP]"},
 CellID->2114498624],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1729399905]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1955171682],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ comm \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ comm \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isComm", "[", 
         RowBox[{
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isComm", "[", 
         RowBox[{
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2114498624"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2114498624-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2114498624-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2114498624", "Proof|ID:2114498624-1"},
 CellID->2065823929],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2114498624-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->545810919]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["-", TraditionalForm]]],
 " is Inverse of ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.638510630376004*^9, 3.638510638139942*^9}, {
  3.639733646971439*^9, 3.639733651561771*^9}},
 CellID->348883692],

Cell["", "OpenEnvironment",
 CellID->1657021632],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox["-", TraditionalForm]]],
 " is Inverse of ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " in \[ScriptCapitalP]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638521604301811*^9, 3.638521618170898*^9}, {
  3.638521834260935*^9, 3.6385218423487*^9}},
 CellID->507868481],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", ",", "q"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 246227628}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6385218773789988`*^9, 3.638521885752161*^9}},
 CellTags->{
  "ID:246227628", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->246227628],

Cell[BoxData[
 RowBox[{
  RowBox[{"p", 
   UnderscriptBox["-", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
  RowBox[{"p", 
   UnderscriptBox["+", "\[ScriptCapitalP]"], 
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["-", "\[ScriptCapitalP]"], "q"}], 
    ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- inverse \[ScriptCapitalP] 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:95118700", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638521960199977*^9, 3.638521978620165*^9}},
 CellTags->{
  "ID:95118700", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP] 1"},
 CellID->95118700],

Cell[BoxData[
 RowBox[{
  RowBox[{"p", 
   UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], "\[Equal]", 
  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- inverse \[ScriptCapitalP] 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1900465816", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6385219848413687`*^9, 3.6385219942909*^9}},
 CellTags->{
  "ID:1900465816", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP] 2"},
 CellID->1900465816],

Cell[BoxData[
 RowBox[{"isInverse", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["+", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["Minus", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   UnderscriptBox["0", "\[ScriptCapitalP]"], ",", "\[ScriptCapitalP]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- inverse \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2037231548", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641635864784939*^9, 3.641635896204391*^9}},
 CellTags->{
  "ID:2037231548", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP]"},
 CellID->2037231548],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->150491788]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->983349670],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- inverse \[ScriptCapitalP] 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- inverse \[ScriptCapitalP] 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["+", "\[ScriptCapitalP]"], 
             RowBox[{
               UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["+", "\[ScriptCapitalP]"], 
             RowBox[{
               UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:95118700"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p95118700-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p95118700-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:95118700", "Proof|ID:95118700-1"},
 CellID->49996501],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p95118700-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->2105700149]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- inverse \[ScriptCapitalP] 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- inverse \[ScriptCapitalP] 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             StyleBox["p", "ExpressionVariable"]}], "\[Equal]", 
           UnderscriptBox[
           "0", "\[ScriptCapitalP]"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             StyleBox["p", "ExpressionVariable"]}], "\[Equal]", 
           UnderscriptBox[
           "0", "\[ScriptCapitalP]"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1900465816"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1900465816-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1900465816-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1900465816", "Proof|ID:1900465816-1"},
 CellID->98966748],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1900465816-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->403404329]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- inverse \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- inverse \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isInverse", "[", 
         RowBox[{
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           UnderscriptBox["Minus", "\[ScriptCapitalP]"], ",", 
           UnderscriptBox["0", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isInverse", "[", 
         RowBox[{
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           UnderscriptBox["Minus", "\[ScriptCapitalP]"], ",", 
           UnderscriptBox["0", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2037231548"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2037231548-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2037231548-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2037231548", "Proof|ID:2037231548-1"},
 CellID->1554728169],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2037231548-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1256029176]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->394685747],

Cell[CellGroupData[{

Cell["Corollary", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.639735247445849*^9},
 CellID->404012057],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["-", "\[ScriptCapitalP]"], 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Equal]", 
  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- 0 \[ScriptCapitalP] 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:145364866", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639735346027746*^9, 3.639735358184018*^9}},
 CellTags->{
  "ID:145364866", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- 0 \[ScriptCapitalP] 1"},
 CellID->145364866],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "p", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["-", "\[ScriptCapitalP]"], 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Equal]", "p"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- 0 \[ScriptCapitalP] 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:117883331", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639735255984515*^9, 3.6397352796085243`*^9}},
 CellTags->{
  "ID:117883331", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- 0 \[ScriptCapitalP] 2"},
 CellID->117883331],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1053861784]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2119922305],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- 0 \[ScriptCapitalP] 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- 0 \[ScriptCapitalP] 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
           UnderscriptBox["0", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
         UnderscriptBox[
         "0", "\[ScriptCapitalP]"]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
           UnderscriptBox["0", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
         UnderscriptBox[
         "0", "\[ScriptCapitalP]"]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:145364866"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p145364866-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p145364866-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:145364866", "Proof|ID:145364866-1"},
 CellID->448617855],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p145364866-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1296800118]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- 0 \[ScriptCapitalP] 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- 0 \[ScriptCapitalP] 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Equal]", 
           StyleBox[
           "p", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Equal]", 
           StyleBox[
           "p", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:117883331"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p117883331-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p117883331-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:117883331", "Proof|ID:117883331-1"},
 CellID->1580290873],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p117883331-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1688210865]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
First Corollaries and Further Preparations (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{
  3.638517457605543*^9, {3.639805529835888*^9, 3.639805538861224*^9}, {
   3.640429437014491*^9, 3.640429441102236*^9}},
 CellID->1482174164],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"\[ScriptCapitalP]", ",", "+"}], ")"}], TraditionalForm]]],
 " is Cancellative Commutative Monoid (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.639974063703701*^9, 3.6399740767889137`*^9}, {
  3.639976114613121*^9, 3.639976118931589*^9}},
 CellID->1081324154],

Cell["", "OpenEnvironment",
 CellID->884155321],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"\[ScriptCapitalP]", ",", "+"}], ")"}], TraditionalForm]]],
 " is Cancellative Commutative Monoid"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.639974092841633*^9, 3.6399741073748617`*^9}},
 CellID->763838058],

Cell[BoxData[
 RowBox[{"isCCMonoid", "[", 
  RowBox[{"\[ScriptCapitalP]", ",", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["+", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCCMonoid \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:771361293", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6399741124488297`*^9, 3.639974132995718*^9}},
 CellTags->{
  "ID:771361293", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCCMonoid \[ScriptCapitalP]"},
 CellID->771361293],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->188406047]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1310768700],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCCMonoid \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCCMonoid \[ScriptCapitalP])\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isCCMonoid", "[", 
         RowBox[{"\[ScriptCapitalP]", ",", 
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isCCMonoid", "[", 
         RowBox[{"\[ScriptCapitalP]", ",", 
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:771361293"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p771361293-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p771361293-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:771361293", "Proof|ID:771361293-1"},
 CellID->556042782],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p771361293-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->472383758]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Selectors (R, C, H) (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.639815377968603*^9, 3.63981540287582*^9}, {
   3.6398275390259113`*^9, 3.6398275429091787`*^9}, 3.639827708447135*^9, {
   3.640576801152938*^9, 3.640576804630355*^9}, {3.642308579880392*^9, 
   3.6423085843602543`*^9}},
 CellID->2092883158],

Cell["", "OpenEnvironment",
 CellID->947786111],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638517466042634*^9, 3.638517469106764*^9}, {
  3.639815426487294*^9, 3.639815443861828*^9}, {3.640576863474332*^9, 
  3.64057687014953*^9}, {3.642308590643363*^9, 3.6423085940497923`*^9}},
 CellID->2067413038],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "p", "]"}]}]], 
  RowBox[{"p", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1221832846}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638517512882473*^9, 3.6385175292291813`*^9}},
 CellTags->{
  "ID:1221832846", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1221832846],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership R)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:313886646", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63566068228487*^9, 3.635660705107398*^9}, {
  3.635660840961052*^9, 3.63566084287151*^9}},
 CellTags->{
  "ID:313886646", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership R", "IR"},
 CellID->313886646],

Cell[BoxData[
 RowBox[{"p", "\[Equal]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], 
   UnderscriptBox["+", "\[ScriptCapitalP]"], 
   RowBox[{
    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lm+R)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:174448700", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631879521033579*^9, 3.631879545111112*^9}},
 CellTags->{
  "ID:174448700", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm+R"},
 CellID->174448700],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", "t"}], 
     "]"}], "\[Equal]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
         RowBox[{"p", ",", "t"}], "]"}], "\[DoubleLeftArrow]", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]},
       {
        UnderscriptBox["0", "\[ScriptCapitalR]"], "\[DoubleLeftArrow]", 
        "otherwise"}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C R)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1679470623", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638515756332316*^9, 3.638515808089093*^9}, {
  3.6386794439898853`*^9, 3.638679456999855*^9}, {3.639822056959516*^9, 
  3.6398220581922617`*^9}},
 CellTags->{
  "ID:1679470623", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C R"},
 CellID->1679470623],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
     UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1648446612", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6399116198476877`*^9, 3.6399116672748327`*^9}},
 CellTags->{
  "ID:1648446612", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 1"},
 CellID->1648446612],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "s", "]"}]}]], 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "s"}]], 
    RowBox[{
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "s"}], "]"}], "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1799775671", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632719742617395*^9, 3.6327197998149033`*^9}, {
  3.6357474672377787`*^9, 3.6357474774261436`*^9}},
 CellTags->{
  "ID:1799775671", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 2"},
 CellID->1799775671],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], "]"}], 
  "\[Equal]", 
  UnderscriptBox["0", "\[ScriptCapitalR]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R 6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1065304756", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642478794393136*^9, 3.642478817194376*^9}},
 CellTags->{
  "ID:1065304756", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 6"},
 CellID->1065304756],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"t", "\[NotEqual]", 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
    "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{
        RowBox[{
         UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
        "t"}], "]"}], "\[Equal]", 
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "t"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R 7)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1078436809", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642478824403111*^9, 3.642478866479829*^9}},
 CellTags->{
  "ID:1078436809", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 7"},
 CellID->1078436809],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "p", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R 8)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:735362349", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642478895937191*^9, 3.642478921529913*^9}},
 CellTags->{
  "ID:735362349", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 8"},
 CellID->735362349],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "p", "]"}], 
  "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R 8')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:451768283", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6430865537665*^9, 3.6430865688585253`*^9}},
 CellTags->{
  "ID:451768283", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 8'"},
 CellID->451768283],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"t", "\[Equal]", 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], "\[Or]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}], "\[And]", 
        
        RowBox[{
         RowBox[{
          UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], 
         "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C 5')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:739591353", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643002648523264*^9, 3.64300270981811*^9}, {
  3.6430027476152973`*^9, 3.6430027770598783`*^9}},
 CellTags->{
  "ID:739591353", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 5'"},
 CellID->739591353],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[NotEqual]", 
   
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1347176402}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639825097437377*^9, 3.6398251019131107`*^9}},
 CellTags->{
  "ID:1347176402", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1347176402],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
  UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
  RowBox[{
   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1377108879", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631880069801291*^9, 3.631880122365003*^9}, {
   3.631880225780108*^9, 3.631880227502932*^9}, 3.632719652799224*^9, {
   3.6331726949105463`*^9, 3.63317270518039*^9}, {3.6357474889983387`*^9, 
   3.635747499186688*^9}, {3.6398251006521683`*^9, 3.639825108459394*^9}},
 CellTags->{
  "ID:1377108879", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 3"},
 CellID->1377108879],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"t", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
        RowBox[{
         UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], "]"}]}], 
      "\[And]", 
      RowBox[{"t", "\[NotEqual]", 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
      UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:139632107", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639825114985365*^9, 3.639825182187688*^9}},
 CellTags->{
  "ID:139632107", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 4"},
 CellID->139632107],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", 
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], "]"}]}], 
   "]"}], "\[NotEqual]", 
  UnderscriptBox["0", "\[ScriptCapitalR]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C 7)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2132986120", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632719945578679*^9, 3.632719965849927*^9}},
 CellTags->{
  "ID:2132986120", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 7"},
 CellID->2132986120],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], "]"}]}], 
   "]"}], "\[Equal]", 
  UnderscriptBox["0", "\[ScriptCapitalR]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C 8)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1911280389", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632719975653163*^9, 3.6327199972370853`*^9}},
 CellTags->{
  "ID:1911280389", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 8"},
 CellID->1911280389],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1601807717]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1176230318],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership R)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership R)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:313886646"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p313886646-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p313886646-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:313886646", "Proof|ID:313886646-1"},
 CellID->368790604],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p313886646-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1023248069]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lm+R)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lm+R)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:174448700"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p174448700-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p174448700-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:174448700", "Proof|ID:174448700-1"},
 CellID->1932131695],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p174448700-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->632362462]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C R)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C R)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                  "\[DoubleLeftArrow]", 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}]}, {
                  UnderscriptBox["0", "\[ScriptCapitalR]"], 
                  "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                  "\[DoubleLeftArrow]", 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}]}, {
                  UnderscriptBox["0", "\[ScriptCapitalR]"], 
                  "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1679470623"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1679470623-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1679470623-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1679470623", "Proof|ID:1679470623-1"},
 CellID->1796196826],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1679470623-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1707284078]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
         RowBox[{
           RowBox[{
             UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
             UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
           UnderscriptBox[
           "0", "\[ScriptCapitalP]"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
         RowBox[{
           RowBox[{
             UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
             UnderscriptBox["1", "\[ScriptCapitalP]"], "]"}], "\[Equal]", 
           UnderscriptBox[
           "0", "\[ScriptCapitalP]"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1648446612"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1648446612-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1648446612-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1648446612", "Proof|ID:1648446612-1"},
 CellID->822584714],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1648446612-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->580974147]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                  UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                  StyleBox["s", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 "\[NotEqual]", 
                 UnderscriptBox[
                 "0", "\[ScriptCapitalR]"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                  UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                  StyleBox["s", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 "\[NotEqual]", 
                 UnderscriptBox[
                 "0", "\[ScriptCapitalR]"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1799775671"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1799775671-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1799775671-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1799775671", "Proof|ID:1799775671-1"},
 CellID->1365150475],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1799775671-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1394382683]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R 6)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R 6)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
             "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
             "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1065304756"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1065304756-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1065304756-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1065304756", "Proof|ID:1065304756-1"},
 CellID->364956431],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1065304756-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1722541450]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R 7)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R 7)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1078436809"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1078436809-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1078436809-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1078436809", "Proof|ID:1078436809-1"},
 CellID->1581073290],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1078436809-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->45607130]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R 8)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R 8)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:735362349"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p735362349-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p735362349-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:735362349", "Proof|ID:735362349-1"},
 CellID->229453058],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p735362349-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->647018837]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R 8')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R 8')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:451768283"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p451768283-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p451768283-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:451768283", "Proof|ID:451768283-1"},
 CellID->497176783],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p451768283-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->86322822]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C 5')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C 5')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], 
               "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}], "\[Or]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                 
                 UnderscriptBox[
                 "0", "\[ScriptCapitalP]"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], 
               "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}], "\[Or]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                 
                 UnderscriptBox[
                 "0", "\[ScriptCapitalP]"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:739591353"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p739591353-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p739591353-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:739591353", "Proof|ID:739591353-1"},
 CellID->1363695772],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p739591353-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->883221224]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1377108879"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1377108879-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1377108879-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1377108879", "Proof|ID:1377108879-1"},
 CellID->360514177],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1377108879-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1351144559]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R 4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}], 
                 "\[And]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 UnderscriptBox[
                 "0", "\[ScriptCapitalR]"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}], 
                 "\[And]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 UnderscriptBox[
                 "0", "\[ScriptCapitalR]"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:139632107"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p139632107-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p139632107-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:139632107", "Proof|ID:139632107-1"},
 CellID->1587172171],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p139632107-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1470107478]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C 7)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C 7)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}], 
               "]"}], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}], 
               "]"}], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2132986120"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2132986120-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2132986120-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2132986120", "Proof|ID:2132986120-1"},
 CellID->2053017099],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2132986120-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->448185845]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C 8)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C 8)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}], 
               "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}], 
               "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1911280389"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1911280389-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1911280389-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1911280389", "Proof|ID:1911280389-1"},
 CellID->472365930],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1911280389-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1273675156]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1812050497],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox["H", TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.639812660309196*^9, 3.63981266552731*^9}},
 CellID->2059822108],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "t", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1031528829}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6331720878606453`*^9, 3.6331720953266897`*^9}},
 CellTags->{
  "ID:1031528829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1031528829],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", ",", "q"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1494855415}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633172099308445*^9, 3.6331721073770657`*^9}, {
  3.6331722034825497`*^9, 3.633172203631197*^9}},
 CellTags->{
  "ID:1494855415", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1494855415],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "s", "]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "r", "]"}]}]], 
    RowBox[{
     RowBox[{"r", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "r", "]"}], "\[Equal]",
       "s"}]}]], 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{
      UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
      RowBox[{"r", ",", "t"}], "]"}], "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership H ind)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2007119699", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6423091014444847`*^9, 3.642309142709566*^9}, {
  3.6423096460462933`*^9, 3.642309657593769*^9}},
 CellTags->{
  "ID:2007119699", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership H ind"},
 CellID->2007119699],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership H)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:884488101", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63566068228487*^9, 3.635660695752953*^9}},
 CellTags->{
  "ID:884488101", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership H", "IR"},
 CellID->884488101],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"s", ",", "u"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "r", "]"}]}]], 
    RowBox[{
     RowBox[{"r", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "r", "]"}], "\[Equal]",
       "u"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{
       RowBox[{
        UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
        RowBox[{"r", ",", "t"}], "]"}], ",", "s"}], "]"}], "\[Equal]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"r", ",", "s"}], "]"}], "\[DoubleLeftArrow]", 
          RowBox[{"s", 
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]},
         {
          UnderscriptBox["0", "\[ScriptCapitalR]"], "\[DoubleLeftArrow]", 
          "otherwise"}
        }]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C H ind)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1892666968", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64231018950246*^9, 3.6423102273513327`*^9}},
 CellTags->{
  "ID:1892666968", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C H ind"},
 CellID->1892666968],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "s", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "t"}], "]"}], ",", "s"}], "]"}], "\[Equal]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{
          UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
          RowBox[{"p", ",", "s"}], "]"}], "\[DoubleLeftArrow]", 
         RowBox[{"s", 
          UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]},
        {
         UnderscriptBox["0", "\[ScriptCapitalR]"], "\[DoubleLeftArrow]", 
         "otherwise"}
       }]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C H)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1556838587", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6423086779196653`*^9, 3.642308734846878*^9}},
 CellTags->{
  "ID:1556838587", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C H"},
 CellID->1556838587],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "s", "]"}]}]], 
     RowBox[{"s", 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]], 
    RowBox[{
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "s"}], "]"}], "\[Equal]", 
     UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H=0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1768827306", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638099601466947*^9, 3.6380996611088543`*^9}, {
  3.6381570578542547`*^9, 3.638157071565563*^9}},
 CellTags->{
  "ID:1768827306", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H=0"},
 CellID->1768827306],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
    RowBox[{
     UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
     RowBox[{"q", ",", "t"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "s", "]"}]}]], 
     RowBox[{"s", 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "s"}], "]"}], "\[Equal]", 
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"q", ",", "s"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H=H)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:140826492", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6386014865764217`*^9, 3.63860155349341*^9}},
 CellTags->{
  "ID:140826492", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H=H"},
 CellID->140826492],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["+", "\[ScriptCapitalP]"], "q"}], ",", "t"}], "]"}], 
  "\[Equal]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}], 
   UnderscriptBox["+", "\[ScriptCapitalP]"], 
   RowBox[{
    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
    RowBox[{"q", ",", "t"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1627546251", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633172208425226*^9, 3.633172234166685*^9}},
 CellTags->{
  "ID:1627546251", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 1"},
 CellID->1627546251],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], ",", "t"}], "]"}], 
  "\[Equal]", 
  RowBox[{
   UnderscriptBox["-", "\[ScriptCapitalP]"], 
   RowBox[{
    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:743145124", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6331726208367243`*^9, 3.6331726274808207`*^9}},
 CellTags->{
  "ID:743145124", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 2"},
 CellID->743145124],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["-", "\[ScriptCapitalP]"], "q"}], ",", "t"}], "]"}], 
  "\[Equal]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}], 
   UnderscriptBox["-", "\[ScriptCapitalP]"], 
   RowBox[{
    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
    RowBox[{"q", ",", "t"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1576432658", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6331722436862497`*^9, 3.633172264006868*^9}},
 CellTags->{
  "ID:1576432658", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 3"},
 CellID->1576432658],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1977357773", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633083508625766*^9, 3.63308357715351*^9}, {
   3.633085771219508*^9, 3.633085772542789*^9}, 3.635747511013411*^9},
 CellTags->{
  "ID:1977357773", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 4"},
 CellID->1977357773],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
    RowBox[{"t", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 4')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:561821259", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640680868119149*^9, 3.6406809258061237`*^9}},
 CellTags->{
  "ID:561821259", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 4'"},
 CellID->561821259],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "s", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{
        UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
        RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
       RowBox[{
        UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
        RowBox[{"q", ",", "t"}], "]"}]}], "\[And]", 
      RowBox[{"s", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{"p", ",", "s"}], "]"}], "\[Equal]", 
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{"q", ",", "s"}], "]"}]}], "\[And]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
        RowBox[{"p", ",", "s"}], "]"}], "\[Equal]", 
       RowBox[{
        UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
        RowBox[{"q", ",", "s"}], "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:866138121", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633083729979018*^9, 3.633083808979444*^9}, {
   3.633085775116444*^9, 3.633085775374806*^9}, 3.639818540323833*^9},
 CellTags->{
  "ID:866138121", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 5"},
 CellID->866138121],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
     RowBox[{
      UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
      RowBox[{"q", ",", "t"}], "]"}]}], "\[And]", 
    RowBox[{"p", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{"q", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]", 
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]", 
    RowBox[{
     UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 5')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:105044014", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6405162867770576`*^9, 3.640516358546936*^9}, {
  3.6405164206839952`*^9, 3.64051643397225*^9}},
 CellTags->{
  "ID:105044014", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 5'"},
 CellID->105044014],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "s", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{
        UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
        RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
      RowBox[{"s", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "s"}], "]"}], "\[Equal]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1971276224", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632819629620943*^9, 3.632819678144375*^9}, {
  3.633085779690895*^9, 3.633085779876333*^9}},
 CellTags->{
  "ID:1971276224", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 6"},
 CellID->1971276224],

Cell[BoxData[
 RowBox[{
  RowBox[{"p", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 219245057}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633172145227119*^9, 3.6331721548880663`*^9}, 
   3.635747520643022*^9},
 CellTags->{
  "ID:219245057", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->219245057],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", 
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], "]"}], 
  "\[Equal]", 
  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 7)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2135590264", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318783407812653`*^9, 3.631878358960847*^9}},
 CellTags->{
  "ID:2135590264", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 7"},
 CellID->2135590264],

Cell[BoxData[
 RowBox[{
  RowBox[{"t", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
   RowBox[{
    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
  "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 8)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2092106183", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6328199942617598`*^9, 3.6328200129577427`*^9}, {
  3.6330857832182693`*^9, 3.6330857868976107`*^9}},
 CellTags->{
  "ID:2092106183", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 8"},
 CellID->2092106183],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[NotEqual]", 
   
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 831355566}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642308374044216*^9, 3.6423083857694693`*^9}},
 CellTags->{
  "ID:831355566", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->831355566],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", 
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], "]"}]}], 
   "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 10)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2014193763", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632719608140501*^9, 3.632719610872704*^9}, {
  3.632719655595729*^9, 3.632719679695356*^9}},
 CellTags->{
  "ID:2014193763", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 10"},
 CellID->2014193763],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], "]"}]}], 
   "]"}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 11)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1650576647", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632719686353598*^9, 3.632719698865877*^9}},
 CellTags->{
  "ID:1650576647", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 11"},
 CellID->1650576647],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1728877054]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->596281154],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership H ind)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership H ind)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["r", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["r", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["s", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["r", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["r", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["r", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["s", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["r", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2007119699"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2007119699-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2007119699-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2007119699", "Proof|ID:2007119699-1"},
 CellID->526673768],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2007119699-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.642310313381415*^9},
 CellID->1470696438]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership H)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership H)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
           RowBox[{
             UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
           RowBox[{
             UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:884488101"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p884488101-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p884488101-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:884488101", "Proof|ID:884488101-1"},
 CellID->1564315886],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p884488101-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->375951896]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C H ind)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C H ind)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["r", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["r", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["r", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                 StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                  "\[DoubleLeftArrow]", 
                  RowBox[{
                    StyleBox["s", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}]}, {
                  UnderscriptBox["0", "\[ScriptCapitalR]"], 
                  "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["r", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["r", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["u", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["r", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                 StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                  "\[DoubleLeftArrow]", 
                  RowBox[{
                    StyleBox["s", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}]}, {
                  UnderscriptBox["0", "\[ScriptCapitalR]"], 
                  "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1892666968"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1892666968-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1892666968-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1892666968", "Proof|ID:1892666968-1"},
 CellID->800813603],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1892666968-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->119379889]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C H)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C H)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                RowBox[{
                  UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                  RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["s", "ExpressionVariable"], 
                  UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                  StyleBox["t", "ExpressionVariable"]}]}, {
                UnderscriptBox["0", "\[ScriptCapitalR]"], 
                "\[DoubleLeftArrow]", 
                "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                RowBox[{
                  UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                  RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["s", "ExpressionVariable"], 
                  UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                  StyleBox["t", "ExpressionVariable"]}]}, {
                UnderscriptBox["0", "\[ScriptCapitalR]"], 
                "\[DoubleLeftArrow]", 
                "True"}}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1556838587"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1556838587-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1556838587-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1556838587", "Proof|ID:1556838587-1"},
 CellID->293752933],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1556838587-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1224452790]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H=0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H=0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["s", "ExpressionVariable"], 
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                StyleBox["t", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalR]"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["s", "ExpressionVariable"], 
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                StyleBox["t", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalR]"]}]}]}]}], $CellContext`TheoremaForm]],
      "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1768827306"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1768827306-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1768827306-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1768827306", "Proof|ID:1768827306-1"},
 CellID->864932006],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1768827306-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1209343178]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H=H)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H=H)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["s", "ExpressionVariable"], 
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                StyleBox["t", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["s", "ExpressionVariable"], 
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                StyleBox["t", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:140826492"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p140826492-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p140826492-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:140826492", "Proof|ID:140826492-1"},
 CellID->743847445],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p140826492-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1288942961]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["+", "\[ScriptCapitalP]"], 
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["+", "\[ScriptCapitalP]"], 
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1627546251"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1627546251-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1627546251-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1627546251", "Proof|ID:1627546251-1"},
 CellID->1563876245],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1627546251-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->670928490]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["Minus", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:743145124"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p743145124-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p743145124-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:743145124", "Proof|ID:743145124-1"},
 CellID->729988509],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p743145124-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->122686757]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1576432658"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1576432658-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 GeneratedCell->False,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1576432658-1"]]]]}, {None, None}},
 CellChangeTimes->{3.6423085226268063`*^9},
 CellTags->{"Proof|ID:1576432658", "Proof|ID:1576432658-1"},
 CellID->478962428],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1576432658-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1090556569]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1977357773"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1977357773-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1977357773-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1977357773", "Proof|ID:1977357773-1"},
 CellID->518253128],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1977357773-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->696944507]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 4')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 4')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:561821259"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p561821259-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p561821259-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:561821259", "Proof|ID:561821259-1"},
 CellID->1734326259],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p561821259-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->728909696]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["s", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:866138121"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p866138121-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p866138121-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:866138121", "Proof|ID:866138121-1"},
 CellID->726143767],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p866138121-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2015149483]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 5')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 5')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:105044014"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p105044014-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p105044014-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:105044014", "Proof|ID:105044014-1"},
 CellID->839817768],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p105044014-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1916847501]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 6)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 6)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1971276224"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1971276224-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1971276224-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1971276224", "Proof|ID:1971276224-1"},
 CellID->611586869],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1971276224-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->587830886]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 7)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 7)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
             "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
             "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2135590264"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2135590264-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2135590264-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2135590264", "Proof|ID:2135590264-1"},
 CellID->88523793],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2135590264-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->828333552]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 8)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 8)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalP]"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalP]"]}]}]}]}], $CellContext`TheoremaForm]],
      "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2092106183"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2092106183-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2092106183-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2092106183", "Proof|ID:2092106183-1"},
 CellID->157583881],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2092106183-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1527934028]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 10)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 10)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}], 
                 "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}], 
                 "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2014193763"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2014193763-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2014193763-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2014193763", "Proof|ID:2014193763-1"},
 CellID->1352301682],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2014193763-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1444998026]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 11)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 11)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}], 
                 "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}], 
                 "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1650576647"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1650576647-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1650576647-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1650576647", "Proof|ID:1650576647-1"},
 CellID->1429121459],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1650576647-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->637145231]
}, Closed]],

Cell[TextData[{
 "The following corollary states an induction principle for the polynomial \
domain \[ScriptCapitalP]. Although \[ScriptCapitalP] will be shown to be \
ordered by a Noetherian order relation if \[ScriptCapitalP] is ordered \
according to formula (isOrdered) in the next section, (induction on \
\[ScriptCapitalP]) is a different kind of induction: It resembles induction \
on the structure of ",
 StyleBox["polynomials",
  FontSlant->"Italic"],
 ", i.e. using the fact that every polynomial is either 0 or constructed from \
a \[OpenCurlyDoubleQuote]smaller\[CloseCurlyDoubleQuote] polynomial by adding \
a new leading monomial. In fact, this is the only induction principle needed \
for \[ScriptCapitalP]."
}], "Text",
 CellChangeTimes->{{3.6430174728228073`*^9, 3.643017685940769*^9}, {
  3.643087776748459*^9, 3.643087777432477*^9}},
 CellID->1413316893],

Cell["", "OpenEnvironment",
 CellID->1924653495],

Cell[CellGroupData[{

Cell["Corollary: Induction on \[ScriptCapitalP]", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.643017337237597*^9, 3.6430173414620733`*^9}},
 CellID->758824004],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "P"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 45249416}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643017748835269*^9, 3.643017749692181*^9}},
 CellTags->{
  "ID:45249416", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->45249416],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"P", "[", 
       UnderscriptBox["0", "\[ScriptCapitalP]"], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
          RowBox[{"[", "p", "]"}]}]], 
        RowBox[{
         RowBox[{"p", "\[NotEqual]", 
          UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
         RowBox[{"t", 
          UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
              RowBox[{"[", "r", "]"}]}]], 
            RowBox[{
             RowBox[{"r", "\[NotEqual]", 
              UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
             RowBox[{
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
              UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "r", 
               "]"}]}]}]], 
           RowBox[{"P", "[", "r", "]"}]}], 
          TagBox[")",
           "AutoParentheses"]}], "\[Implies]", 
         RowBox[{"P", "[", "p", "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
         RowBox[{"[", "p", "]"}]}]], 
       RowBox[{
        RowBox[{"p", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
        RowBox[{"t", 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}]], 
      RowBox[{"P", "[", "p", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(induction on \[ScriptCapitalP] ind)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:248026320", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643017794644368*^9, 3.643017820174568*^9}, {
  3.643017870274577*^9, 3.643017929752843*^9}, {3.6430179845628567`*^9, 
  3.643018032084078*^9}, {3.643020823679084*^9, 3.643020869995569*^9}, {
  3.643020955631303*^9, 3.643020988226522*^9}, {3.6430211369006643`*^9, 
  3.643021148662388*^9}, {3.643021500665345*^9, 3.6430216159850082`*^9}, {
  3.6430217372173767`*^9, 3.643021744388928*^9}, {3.643021880857607*^9, 
  3.6430219373727427`*^9}, {3.643083101032619*^9, 3.643083156180501*^9}, {
  3.643083307516943*^9, 3.64308332790635*^9}, {3.643083394822659*^9, 
  3.643083441853156*^9}, {3.64308391505622*^9, 3.643083922050899*^9}},
 CellTags->{
  "ID:248026320", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "induction on \[ScriptCapitalP] ind"},
 CellID->248026320],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"P", "[", 
     UnderscriptBox["0", "\[ScriptCapitalP]"], "]"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
         RowBox[{"[", "p", "]"}]}]], 
       RowBox[{"p", "\[NotEqual]", 
        UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"P", "[", 
         RowBox[{
          UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], "]"}], 
        "\[Implies]", 
        RowBox[{"P", "[", "p", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "p", "]"}]}]], 
   RowBox[{"P", "[", "p", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(induction on \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1502331257", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643017359421352*^9, 3.6430174245737667`*^9}, {
  3.64301775243055*^9, 3.64301775902769*^9}},
 CellTags->{
  "ID:1502331257", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "induction on \[ScriptCapitalP]", "IR"},
 CellID->1502331257],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1647825250]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->172493216],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(induction on \[ScriptCapitalP] ind)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(induction on \[ScriptCapitalP] ind)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["P", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               UnderscriptBox["0", "\[ScriptCapitalP]"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
                  
                  RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
                    
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["r", "ExpressionVariable"], "]"}]}]}]], 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["r", "ExpressionVariable"], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
                RowBox[{
                  StyleBox["t", "ExpressionVariable"], 
                  UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                  RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["P", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               UnderscriptBox["0", "\[ScriptCapitalP]"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
                  
                  RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
                    
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["r", "ExpressionVariable"], "]"}]}]}]], 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["r", "ExpressionVariable"], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
                RowBox[{
                  StyleBox["t", "ExpressionVariable"], 
                  UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                  RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:248026320"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p248026320-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p248026320-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:248026320", "Proof|ID:248026320-1"},
 CellID->717333436],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p248026320-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->680214968]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(induction on \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(induction on \[ScriptCapitalP])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               UnderscriptBox["0", "\[ScriptCapitalP]"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               UnderscriptBox["0", "\[ScriptCapitalP]"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1502331257"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1502331257-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1502331257-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1502331257", "Proof|ID:1502331257-1"},
 CellID->2113840420],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1502331257-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1986041523]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Monomials (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.639820648711814*^9, 3.639820650019476*^9}, {
  3.639907807479089*^9, 3.639907811573307*^9}},
 CellID->1456829087],

Cell["", "OpenEnvironment",
 CellID->1742278695],

Cell[CellGroupData[{

Cell["Corollary: Properties of Monomials", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.639820666705332*^9, 3.639820670927155*^9}},
 CellID->1985392776],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", "c", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 411993095}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320401373168383`*^9, 3.632040146353407*^9}, 
   3.6399070943346767`*^9},
 CellTags->{
  "ID:411993095", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->411993095],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "d", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      RowBox[{"c", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
      UnderscriptBox["+", "\[ScriptCapitalP]"], 
      RowBox[{"d", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}]}], 
     "\[Equal]", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"c", 
        UnderscriptBox["+", "\[ScriptCapitalR]"], "d"}], ")"}], 
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:470022203", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631955135288076*^9, 3.6319552489033546`*^9}, {
  3.6320401615053387`*^9, 3.632040163049529*^9}, {3.632717084679607*^9, 
  3.632717115733561*^9}, {3.639907064513755*^9, 3.63990709026803*^9}},
 CellTags->{
  "ID:470022203", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 1"},
 CellID->470022203],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", 
   "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 401540586}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639997817836192*^9, 3.6399978401977873`*^9}},
 CellTags->{
  "ID:401540586", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->401540586],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}]}], ")"}], 
   UnderscriptBox["+", "\[ScriptCapitalP]"], "m"}], "\[Equal]", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"c", 
     UnderscriptBox["+", "\[ScriptCapitalR]"], 
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}]}], ")"}], 
   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
   RowBox[{
    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2014025290", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633173919072081*^9, 3.633173961472548*^9}, {
  3.6331739968646517`*^9, 3.633174001497099*^9}, {3.639907548556082*^9, 
  3.639907577427661*^9}, {3.639997843417947*^9, 3.639997847383671*^9}},
 CellTags->{
  "ID:2014025290", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 6"},
 CellID->2014025290],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "m", "]"}], "\[Equal]", 
  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 10)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1590906878", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639997856357047*^9, 3.639997871614552*^9}},
 CellTags->{
  "ID:1590906878", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 10"},
 CellID->1590906878],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]], 
   RowBox[{
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
     RowBox[{"m", ",", "t"}], "]"}], "\[Equal]", "m"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 11)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1997927583", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640518217142544*^9, 3.6405182814071493`*^9}},
 CellTags->{
  "ID:1997927583", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 11"},
 CellID->1997927583],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2121668319]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->859850637],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["+", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["+", "\[ScriptCapitalR]"], 
               StyleBox["d", "ExpressionVariable"]}], 
             UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["+", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["+", "\[ScriptCapitalR]"], 
               StyleBox["d", "ExpressionVariable"]}], 
             UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:470022203"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p470022203-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p470022203-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:470022203", "Proof|ID:470022203-1"},
 CellID->356771522],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p470022203-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->2104860677]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 6)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 6)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               StyleBox["m", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               StyleBox["m", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2014025290"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2014025290-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2014025290-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2014025290", "Proof|ID:2014025290-1"},
 CellID->740297765],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2014025290-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->595415687]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 10)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 10)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1590906878"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1590906878-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1590906878-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1590906878", "Proof|ID:1590906878-1"},
 CellID->1050945896],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1590906878-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1718488369]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 11)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 11)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1997927583"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1997927583-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1997927583-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1997927583", "Proof|ID:1997927583-1"},
 CellID->646926070],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1997927583-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1129892272]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Monomial Multiplication (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6398196444464006`*^9, 3.639819651023875*^9}, {
  3.640429429556052*^9, 3.640429433452869*^9}},
 CellID->223041332],

Cell["", "OpenEnvironment",
 CellID->1774552902],

Cell[CellGroupData[{

Cell["Lemma: Simple Properties of the Monomial Multiplication", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.639909476797138*^9, 3.639909486787744*^9}, {
  3.639997590473892*^9, 3.639997591711472*^9}},
 CellID->1659339968],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{"p", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 63039386}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639909496265081*^9, 3.639909507326536*^9}, {
   3.640318831736442*^9, 3.640318864124329*^9}, 3.642330664012785*^9},
 CellTags->{
  "ID:63039386", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->63039386],

Cell["\<\
Formulas (\[CircleTimes] 0) and (\[CircleTimes] 5) re-state the definition of \
\[CircleTimes] for the concrete domain \[ScriptCapitalP].\
\>", "EnvironmentText",
 CellChangeTimes->{{3.642329856476823*^9, 3.6423298850623693`*^9}},
 CellID->1091890068],

Cell[BoxData[
 RowBox[{
  RowBox[{"p", 
   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Equal]", 
  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CircleTimes] 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1093453827", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63998108287753*^9, 3.6399810990384893`*^9}},
 CellTags->{
  "ID:1093453827", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 0"},
 CellID->1093453827],

Cell[BoxData[
 RowBox[{
  RowBox[{"q", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
        UnderscriptBox["*", "\[ScriptCapitalR]"], 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}], ")"}], 
      
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
        UnderscriptBox["*", "\[ScriptCapitalT]"], 
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
       ")"}]}], 
     UnderscriptBox["+", "\[ScriptCapitalP]"], 
     RowBox[{"p", 
      UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
      RowBox[{
       UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CircleTimes] 5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1795167993", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640319159219006*^9, 3.640319207101329*^9}},
 CellTags->{
  "ID:1795167993", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 5"},
 CellID->1795167993],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "r", "]"}]}]], 
    RowBox[{
     RowBox[{"r", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "r", "]"}], "\[Equal]",
       "t"}]}]], 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", 
      UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "r"}], 
     "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership monomial multiplication ind)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:838653570", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6423303095430937`*^9, 3.642330372590571*^9}, {
  3.642330669711899*^9, 3.6423306735563087`*^9}},
 CellTags->{
  "ID:838653570", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomial multiplication ind"},
 CellID->838653570],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   RowBox[{"p", 
    UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership monomial multiplication)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1903131381", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638516730716522*^9, 3.638516758275998*^9}, 
   3.642330175951045*^9},
 CellTags->{
  "ID:1903131381", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomial multiplication", "IR"},
 CellID->1903131381],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"p", 
       UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], ",", 
      "t"}], "]"}], "\[Equal]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
         UnderscriptBox["*", "\[ScriptCapitalR]"], 
         RowBox[{
          UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
          RowBox[{"q", ",", 
           RowBox[{"t", 
            UnderscriptBox["/", "\[ScriptCapitalT]"], 
            RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}], 
          "]"}]}], "\[DoubleLeftArrow]", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
         UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]},
       {
        UnderscriptBox["0", "\[ScriptCapitalR]"], "\[DoubleLeftArrow]", 
        "otherwise"}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C \[CircleTimes])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1641583144", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642329453706959*^9, 3.642329506975296*^9}},
 CellTags->{
  "ID:1641583144", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C \[CircleTimes]"},
 CellID->1641583144],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["1", "\[ScriptCapitalP]"], 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], "\[Equal]",
     "q"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CircleTimes] 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:608153969", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639911483651065*^9, 3.63991155419563*^9}, {
  3.639913695449007*^9, 3.6399136997969303`*^9}, {3.639914263801449*^9, 
  3.639914267981103*^9}},
 CellTags->{
  "ID:608153969", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 1"},
 CellID->608153969],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "q", "]"}], 
  "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
       UnderscriptBox["*", "\[ScriptCapitalR]"], 
       RowBox[{
        UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}], ")"}], 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
       UnderscriptBox["*", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
      ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CircleTimes] isMonomial)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1439631766", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640318900535915*^9, 3.6403189608339*^9}},
 CellTags->{
  "ID:1439631766", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] isMonomial"},
 CellID->1439631766],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", "r", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 607306392}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642330653000114*^9, 3.642330660513995*^9}},
 CellTags->{
  "ID:607306392", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->607306392],

Cell[BoxData[
 RowBox[{
  RowBox[{"p", 
   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
   RowBox[{"(", 
    RowBox[{"q", 
     UnderscriptBox["+", "\[ScriptCapitalP]"], "r"}], ")"}]}], "\[Equal]", 
  RowBox[{
   RowBox[{"p", 
    UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], 
   UnderscriptBox["+", "\[ScriptCapitalP]"], 
   RowBox[{"p", 
    UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
    "r"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CircleTimes] 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1312761442", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639981170131781*^9, 3.639981205903665*^9}},
 CellTags->{
  "ID:1312761442", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 2"},
 CellID->1312761442],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "p", "]"}], 
    "\[And]", 
    RowBox[{
     UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
    RowBox[{"q", 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "p"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CircleTimes] 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:259600169", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639911404658638*^9, 3.6399114567469997`*^9}},
 CellTags->{
  "ID:259600169", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 3"},
 CellID->259600169],

Cell[BoxData[
 RowBox[{
  RowBox[{"p", 
   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], 
   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
   "q"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CircleTimes] 6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1576857420", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639997491465404*^9, 3.6399975129753523`*^9}},
 CellTags->{
  "ID:1576857420", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 6"},
 CellID->1576857420],

Cell[BoxData[
 RowBox[{
  RowBox[{"q", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"t", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
        UnderscriptBox["*", "\[ScriptCapitalT]"], 
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}], 
      "\[Implies]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{
         RowBox[{"p", 
          UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], ",", 
         "t"}], "]"}], "\[Equal]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CircleTimes] 10)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:483781382", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641020727455249*^9, 3.6410208169235783`*^9}},
 CellTags->{
  "ID:483781382", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 10"},
 CellID->483781382],

Cell[BoxData[
 RowBox[{
  RowBox[{"q", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"p", 
       UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], ",", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
       UnderscriptBox["*", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}], "]"}],
     "\[Equal]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
     UnderscriptBox["*", "\[ScriptCapitalR]"], 
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CircleTimes] 11)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:169057028", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641024303303075*^9, 3.641024351391382*^9}},
 CellTags->{
  "ID:169057028", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 11"},
 CellID->169057028],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "q", "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 324898493}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64032485148661*^9, 3.6403248553984327`*^9}},
 CellTags->{
  "ID:324898493", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->324898493],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"p", 
    UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
   
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{"q", 
       UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "r"}], ")"}]}], 
    "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CircleTimes] 7)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:684947841", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64032461568419*^9, 3.64032468837523*^9}, {
  3.6403248577042522`*^9, 3.640324859969153*^9}},
 CellTags->{
  "ID:684947841", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 7"},
 CellID->684947841],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"p", 
    UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], 
   "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{"q", 
       UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "r"}], ")"}]}], 
    "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"p", 
       UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], ")"}], 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "r"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CircleTimes] 8)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1856294157", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640324740217534*^9, 3.640324769146796*^9}, {
  3.6403248617904043`*^9, 3.6403248638846817`*^9}},
 CellTags->{
  "ID:1856294157", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 8"},
 CellID->1856294157],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"p", 
    UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], 
   "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", 
    UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CircleTimes] 9)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1768460531", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640428986335915*^9, 3.640428999542103*^9}, {
  3.640429051313797*^9, 3.6404290679347057`*^9}},
 CellTags->{
  "ID:1768460531", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 9"},
 CellID->1768460531],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->891227303]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->885827699],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CircleTimes] 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CircleTimes] 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1093453827"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1093453827-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1093453827-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1093453827", "Proof|ID:1093453827-1"},
 CellID->1691805525],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1093453827-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1911765051]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CircleTimes] 5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CircleTimes] 5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1795167993"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1795167993-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1795167993-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1795167993", "Proof|ID:1795167993-1"},
 CellID->1536044778],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1795167993-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1091089291]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership monomial multiplication ind)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership monomial multiplication ind)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["r", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["r", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["r", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["r", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["r", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["r", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:838653570"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p838653570-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p838653570-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:838653570", "Proof|ID:838653570-1"},
 CellID->39930568],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p838653570-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->767866762]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership monomial multiplication)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership monomial multiplication)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1903131381"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1903131381-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1903131381-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1903131381", "Proof|ID:1903131381-1"},
 CellID->443389413],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1903131381-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->111366806]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C \[CircleTimes])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C \[CircleTimes])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["/", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], "]"}]}], 
                  "\[DoubleLeftArrow]", 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}]}, {
                  UnderscriptBox["0", "\[ScriptCapitalR]"], 
                  "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["/", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], "]"}]}], 
                  "\[DoubleLeftArrow]", 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}]}, {
                  UnderscriptBox["0", "\[ScriptCapitalR]"], 
                  "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1641583144"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1641583144-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1641583144-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1641583144", "Proof|ID:1641583144-1"},
 CellID->183185032],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1641583144-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->188296615]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CircleTimes] 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CircleTimes] 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["1", "\[ScriptCapitalP]"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             StyleBox[
             "q", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["1", "\[ScriptCapitalR]"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["1", "\[ScriptCapitalP]"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             StyleBox[
             "q", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:608153969"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p608153969-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p608153969-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:608153969", "Proof|ID:608153969-1"},
 CellID->1418277402],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p608153969-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->495293771]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CircleTimes] isMonomial)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CircleTimes] isMonomial)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1439631766"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1439631766-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1439631766-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1439631766", "Proof|ID:1439631766-1"},
 CellID->821863036],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1439631766-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1360721819]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CircleTimes] 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CircleTimes] 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox[
                 "r", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox[
                 "r", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1312761442"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1312761442-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1312761442-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1312761442", "Proof|ID:1312761442-1"},
 CellID->1943269854],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1312761442-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->405672243]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CircleTimes] 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CircleTimes] 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox[
               "p", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox[
               "p", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:259600169"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p259600169-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p259600169-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:259600169", "Proof|ID:259600169-1"},
 CellID->592189180],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p259600169-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1512486922]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CircleTimes] 6)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CircleTimes] 6)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox[
               "q", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox[
               "q", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1576857420"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1576857420-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1576857420-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1576857420", "Proof|ID:1576857420-1"},
 CellID->1943524802],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1576857420-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->141376629]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CircleTimes] 10)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CircleTimes] 10)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], 
                    UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                    StyleBox["q", "ExpressionVariable"]}], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 UnderscriptBox[
                 "0", "\[ScriptCapitalR]"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], 
                    UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                    StyleBox["q", "ExpressionVariable"]}], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 UnderscriptBox[
                 "0", "\[ScriptCapitalR]"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:483781382"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p483781382-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p483781382-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:483781382", "Proof|ID:483781382-1"},
 CellID->507841304],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p483781382-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1990411776]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CircleTimes] 11)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CircleTimes] 11)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], "]"}], 
             "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], "]"}], 
             "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:169057028"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p169057028-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p169057028-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:169057028", "Proof|ID:169057028-1"},
 CellID->1561081111],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p169057028-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1578160185]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CircleTimes] 7)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CircleTimes] 7)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:684947841"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p684947841-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p684947841-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:684947841", "Proof|ID:684947841-1"},
 CellID->969684081],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p684947841-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1166670995]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CircleTimes] 8)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CircleTimes] 8)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1856294157"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1856294157-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1856294157-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1856294157", "Proof|ID:1856294157-1"},
 CellID->1121162524],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1856294157-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.642492462179254*^9, 3.642492464941382*^9}},
 CellID->2088302303]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CircleTimes] 9)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CircleTimes] 9)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1768460531"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1768460531-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1768460531-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1768460531", "Proof|ID:1768460531-1"},
 CellID->1517299564],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1768460531-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->174900675]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (C \[CircleTimes])",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642482042131219*^9, 3.642482047507924*^9}, 
   3.642482082249363*^9, 3.643446900117631*^9, 3.6532081928168755`*^9, 
   3.6532137566642637`*^9, {3.662368746479022*^9, 3.6623687466194224`*^9}, 
   3.6661740275157795`*^9, 3.666440608818488*^9},
 CellID->1347606622],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642477833713747*^9, 3.6424778349930143`*^9}, 
   3.6424820822495317`*^9, 3.643446900117779*^9, 3.6532081928168755`*^9, 
   3.6532137566642637`*^9, 3.6661740275157795`*^9, 3.666440608818634*^9},
 CellID->905379727],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"R", "[", "r", "]"}], ",", 
      RowBox[{"lp", "[", "r", "]"}]}], "]"}], "=", "0"}], TraditionalForm]]],
 " ... (R 6)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642477859352404*^9, 3.642477868618814*^9}, {
   3.64247916775368*^9, 3.642479171442072*^9}, 3.642482082250105*^9, 
   3.643446900117868*^9, 3.6532081928168755`*^9, 3.6532137566642637`*^9, 
   3.6661740275157795`*^9, 3.666440608818749*^9},
 CellID->2044207238],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"t", "\[NotEqual]", 
      RowBox[{"lp", "[", "r", "]"}]}], "\[Implies]", 
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"R", "[", "r", "]"}], ",", "t"}], "]"}]}], "=", 
    RowBox[{"C", "[", 
     RowBox[{"r", ",", "t"}], "]"}]}], TraditionalForm]]],
 " ... (R 7)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642477859352404*^9, 3.642477888523203*^9}, {
   3.642478514203659*^9, 3.642478514219295*^9}, {3.642478678867268*^9, 
   3.6424786855209637`*^9}, {3.642479173857033*^9, 3.642479176617996*^9}, 
   3.642482082250214*^9, 3.6434469001179323`*^9, 3.6532081928168755`*^9, 
   3.6532137566642637`*^9, 3.6661740275157795`*^9, 3.6664406088188267`*^9},
 CellID->1652777587],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"r", "\[NotEqual]", "0"}], "\[And]", 
     RowBox[{"R", "[", "r", "]"}]}], "=", 
    RowBox[{"0", "\[Implies]", 
     RowBox[{"isMonomial", "[", "r", "]"}]}]}], TraditionalForm]]],
 " ... (R 8)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642477859352404*^9, 3.642477888523203*^9}, {
   3.642478514203659*^9, 3.642478543023508*^9}, {3.642479178550241*^9, 
   3.6424791812368507`*^9}, 3.642482082250318*^9, 3.643446900117983*^9, 
   3.6532081928168755`*^9, 3.6532137566642637`*^9, 3.6661740275157795`*^9, 
   3.6664406088189*^9},
 CellID->1555341482],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"isMonomial", "[", "r", "]"}], "\[And]", 
      RowBox[{"t", "\[NotEqual]", 
       RowBox[{"lp", "[", "r", "]"}]}]}], "\[Implies]", 
     RowBox[{"C", "[", 
      RowBox[{"r", ",", "t"}], "]"}]}], "=", "0"}], TraditionalForm]]],
 " ... (isMonomial 7)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642477859352404*^9, 3.642477888523203*^9}, {
   3.642478514203659*^9, 3.642478572996492*^9}, {3.6424786144408617`*^9, 
   3.642478671541263*^9}, {3.642479756681211*^9, 3.642479762817883*^9}, 
   3.6424820822504177`*^9, 3.643446900118034*^9, 3.6532081928168755`*^9, 
   3.6532137566642637`*^9, 3.6661740275157795`*^9, 3.6664406088189707`*^9},
 CellID->658700090],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "r"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"t", ",", "s"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"r", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "r", "]"}], "=", "s"}], TraditionalForm]]],
 " and\n(IH)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", "s0"], 
      RowBox[{"s", "\[Succeeds]", "s0"}]], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
         RowBox[{"[", "r0", "]"}]}]], 
       RowBox[{
        RowBox[{
         RowBox[{"r0", "\[NotEqual]", "0"}], "\[And]", 
         RowBox[{"lp", "[", "r0", "]"}]}], "=", "s0"}]], 
      RowBox[{"C", "[", 
       RowBox[{
        RowBox[{"p", "\[CircleTimes]", "r0"}], ",", "t"}], "]"}]}]}], "=", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], "*", 
         RowBox[{"C", "[", 
          RowBox[{"r0", ",", 
           RowBox[{"t", "/", 
            RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], 
        "\[DoubleLongLeftArrow]", 
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {"0", "\[DoubleLongLeftArrow]", "otherwise"}
      }]}]}], TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"p", "\[CircleTimes]", "r"}], ",", "t"}], "]"}], "=", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], "*", 
         RowBox[{"C", "[", 
          RowBox[{"r", ",", 
           RowBox[{"t", "/", 
            RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], 
        "\[DoubleLongLeftArrow]", 
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {"0", "\[DoubleLongLeftArrow]", "otherwise"}
      }]}]}], TraditionalForm]]],
 "\nSince ",
 Cell[BoxData[
  FormBox[
   RowBox[{"r", "\[NotEqual]", "0"}], TraditionalForm]]],
 " we can expand the definition of \[CircleTimes]: We have to show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"lc", "[", "p", "]"}], "*", 
          RowBox[{"lc", "[", "r", "]"}]}], ")"}], "\[CenterDot]", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "*", 
          RowBox[{"lp", "[", "r", "]"}]}], ")"}]}], "+", 
       RowBox[{"p", "\[CircleTimes]", 
        RowBox[{"R", "[", "r", "]"}]}]}], ",", "t"}], "]"}], "=", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], "*", 
         RowBox[{"C", "[", 
          RowBox[{"r", ",", 
           RowBox[{"t", "/", 
            RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], 
        "\[DoubleLongLeftArrow]", 
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {"0", "\[DoubleLongLeftArrow]", "otherwise"}
      }]}]}], TraditionalForm]]],
 "\n\nUsing linearity of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " this amounts to showing\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"lc", "[", "p", "]"}], "*", 
          RowBox[{"lc", "[", "r", "]"}]}], ")"}], "\[CenterDot]", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "*", 
          RowBox[{"lp", "[", "r", "]"}]}], ")"}]}], ",", "t"}], "]"}], "+", 
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"p", "\[CircleTimes]", 
        RowBox[{"R", "[", "r", "]"}]}], ",", "t"}], "]"}]}], "=", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], "*", 
         RowBox[{"C", "[", 
          RowBox[{"r", ",", 
           RowBox[{"t", "/", 
            RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], 
        "\[DoubleLongLeftArrow]", 
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {"0", "\[DoubleLongLeftArrow]", "otherwise"}
      }]}]}], TraditionalForm]]],
 "\nThe first summand on the LHS can be rewritten, yielding\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"{", 
    RowBox[{
     RowBox[{GridBox[{
        {
         RowBox[{
          RowBox[{"lc", "[", "p", "]"}], "*", 
          RowBox[{"lc", "[", "r", "]"}]}], "\[DoubleLongLeftArrow]", 
         RowBox[{"t", "=", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], "*", 
           RowBox[{"lp", "[", "r", "]"}]}]}]},
        {"0", "\[DoubleLongLeftArrow]", "otherwise"}
       }], "+", 
      RowBox[{"C", "[", 
       RowBox[{
        RowBox[{"p", "\[CircleTimes]", 
         RowBox[{"R", "[", "r", "]"}]}], ",", "t"}], "]"}]}], "=", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{
          RowBox[{"lc", "[", "p", "]"}], "*", 
          RowBox[{"C", "[", 
           RowBox[{"r", ",", 
            RowBox[{"t", "/", 
             RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], 
         "\[DoubleLongLeftArrow]", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
        {"0", "\[DoubleLongLeftArrow]", "otherwise"}
       }]}]}]}], TraditionalForm]]],
 "\n\nNow we distinguish two cases, based on whether ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"R", "[", "r", "]"}], "=", "0"}], TraditionalForm]]],
 " or not.\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"R", "[", "r", "]"}], "\[NotEqual]", "0"}], TraditionalForm]]],
 ".\nIn this case we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "r", "]"}], "\[Succeeds]", 
    RowBox[{"lp", "[", 
     RowBox[{"R", "[", "r", "]"}], "]"}]}], TraditionalForm]]],
 ", hence we can infer from (IH)\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"p", "\[CircleTimes]", 
       RowBox[{"R", "[", "r", "]"}]}], ",", "t"}], "]"}], "=", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], "*", 
         RowBox[{"C", "[", 
          RowBox[{
           RowBox[{"R", "[", "r", "]"}], ",", 
           RowBox[{"t", "/", 
            RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], 
        "\[DoubleLongLeftArrow]", 
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {"0", "\[DoubleLongLeftArrow]", "otherwise"}
      }]}]}], TraditionalForm]]],
 "\nand thus our goal becomes\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"{", 
    RowBox[{GridBox[{
       {
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], "*", 
         RowBox[{"lc", "[", "r", "]"}]}], "\[DoubleLongLeftArrow]", 
        RowBox[{"t", "=", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "*", 
          RowBox[{"lp", "[", "r", "]"}]}]}]},
       {"0", "\[DoubleLongLeftArrow]", "otherwise"}
      }], "+", 
     RowBox[{"{", 
      RowBox[{GridBox[{
         {
          RowBox[{
           RowBox[{"lc", "[", "p", "]"}], "*", 
           RowBox[{"C", "[", 
            RowBox[{
             RowBox[{"R", "[", "r", "]"}], ",", 
             RowBox[{"t", "/", 
              RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], 
          "\[DoubleLongLeftArrow]", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
         {"0", "\[DoubleLongLeftArrow]", "otherwise"}
        }], "=", 
       RowBox[{"{", GridBox[{
          {
           RowBox[{
            RowBox[{"lc", "[", "p", "]"}], "*", 
            RowBox[{"C", "[", 
             RowBox[{"r", ",", 
              RowBox[{"t", "/", 
               RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], 
           "\[DoubleLongLeftArrow]", 
           RowBox[{
            RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
          {"0", "\[DoubleLongLeftArrow]", "otherwise"}
         }]}]}]}]}]}], TraditionalForm]]],
 "\n\nWe distinguish two cases, based on the first piecewise-expression.\n\
Case I.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "*", 
     RowBox[{"lp", "[", "r", "]"}]}]}], TraditionalForm]]],
 ".\nIn this case we also know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "*", 
     RowBox[{"lp", "[", "r", "]"}]}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"t", "/", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{"lp", "[", "r", "]"}]}], TraditionalForm]]],
 ", hence our goal becomes\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], "*", 
      RowBox[{"lc", "[", "r", "]"}]}], "+", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], "*", 
      RowBox[{"C", "[", 
       RowBox[{
        RowBox[{"R", "[", "r", "]"}], ",", 
        RowBox[{"lp", "[", "r", "]"}]}], "]"}]}]}], "=", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], "*", 
     RowBox[{"C", "[", 
      RowBox[{"r", ",", 
       RowBox[{"lp", "[", "r", "]"}]}], "]"}]}]}], TraditionalForm]]],
 "\nFrom a lemma we know that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"R", "[", "r", "]"}], ",", 
      RowBox[{"lp", "[", "r", "]"}]}], "]"}], "=", "0"}], TraditionalForm]]],
 ", and expanding the definition of lc finishes this branch of the proof.\n\n\
Case I.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[NotEqual]", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "*", 
     RowBox[{"lp", "[", "r", "]"}]}]}], TraditionalForm]]],
 ".\nOur goal becomes\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"{", 
    RowBox[{GridBox[{
       {
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], "*", 
         RowBox[{"C", "[", 
          RowBox[{
           RowBox[{"R", "[", "r", "]"}], ",", 
           RowBox[{"t", "/", 
            RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], 
        "\[DoubleLongLeftArrow]", 
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {"0", "\[DoubleLongLeftArrow]", "otherwise"}
      }], "=", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{
          RowBox[{"lc", "[", "p", "]"}], "*", 
          RowBox[{"C", "[", 
           RowBox[{"r", ",", 
            RowBox[{"t", "/", 
             RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], 
         "\[DoubleLongLeftArrow]", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
        {"0", "\[DoubleLongLeftArrow]", "otherwise"}
       }]}]}]}], TraditionalForm]]],
 "\nso we distinguish again two cases, this time based on the two remaining \
piecewise-expressions.\nCase I.II.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}], TraditionalForm]]],
 ".\nWe have to show\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], "*", 
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"R", "[", "r", "]"}], ",", 
       RowBox[{"t", "/", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], "=", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], "*", 
     RowBox[{"C", "[", 
      RowBox[{"r", ",", 
       RowBox[{"t", "/", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}]}], TraditionalForm]]],
 "\nFor (G7) it is sufficient to show\n(G8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"R", "[", "r", "]"}], ",", 
      RowBox[{"t", "/", 
       RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "=", 
    RowBox[{"C", "[", 
     RowBox[{"r", ",", 
      RowBox[{"t", "/", 
       RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], TraditionalForm]]],
 "\nand by a lemma this amounts to showing\n(G9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"t", "/", 
     RowBox[{"lp", "[", "p", "]"}]}], "\[NotEqual]", 
    RowBox[{"lp", "[", "r", "]"}]}], TraditionalForm]]],
 "\nIn order to show (G9) we exchange (G9) with our case assumption ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[NotEqual]", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "*", 
     RowBox[{"lp", "[", "r", "]"}]}]}], TraditionalForm]]],
 ", meaning we assume\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"t", "/", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{"lp", "[", "r", "]"}]}], TraditionalForm]]],
 "\nand we show\n(G10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "*", 
     RowBox[{"lp", "[", "r", "]"}]}]}], TraditionalForm]]],
 "\nwhich, by (2), amounts to showing\n(G11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "*", 
     RowBox[{"(", 
      RowBox[{"t", "/", 
       RowBox[{"lp", "[", "p", "]"}]}], ")"}]}]}], TraditionalForm]]],
 "\n(G11) follows from a lemma on power-products.\n\nCase I.II.II:: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]}], 
   TraditionalForm]]],
 ".\nTrivial (",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "=", "0"}], TraditionalForm]]],
 ").\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"R", "[", "r", "]"}], "=", "0"}], TraditionalForm]]],
 ".\nExpanding the definition of \[CircleTimes], our goal becomes\n(G12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"{", 
    RowBox[{GridBox[{
       {
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], "*", 
         RowBox[{"lc", "[", "r", "]"}]}], "\[DoubleLongLeftArrow]", 
        RowBox[{"t", "=", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "*", 
          RowBox[{"lp", "[", "r", "]"}]}]}]},
       {"0", "\[DoubleLongLeftArrow]", "otherwise"}
      }], "=", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{
          RowBox[{"lc", "[", "p", "]"}], "*", 
          RowBox[{"C", "[", 
           RowBox[{"r", ",", 
            RowBox[{"t", "/", 
             RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], 
         "\[DoubleLongLeftArrow]", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
        {"0", "\[DoubleLongLeftArrow]", "otherwise"}
       }]}]}]}], TraditionalForm]]],
 "\nWe distinguish two cases, based on the first piecewise-expression.\nCase \
II.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "*", 
     RowBox[{"lp", "[", "r", "]"}]}]}], TraditionalForm]]],
 ".\nIn this case we also know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "*", 
     RowBox[{"lp", "[", "r", "]"}]}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"t", "/", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{"lp", "[", "r", "]"}]}], TraditionalForm]]],
 ", hence our goal becomes\n(G13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], "*", 
     RowBox[{"lc", "[", "r", "]"}]}], "=", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], "*", 
     RowBox[{"C", "[", 
      RowBox[{"r", ",", 
       RowBox[{"lp", "[", "r", "]"}]}], "]"}]}]}], TraditionalForm]]],
 "\nExpaning the definition of lc finishes this branch of the proof.\n\nCase \
II.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[NotEqual]", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "*", 
     RowBox[{"lp", "[", "r", "]"}]}]}], TraditionalForm]]],
 ".\nOur goal becomes\n(G14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "=", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], "*", 
         RowBox[{"C", "[", 
          RowBox[{"r", ",", 
           RowBox[{"t", "/", 
            RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], 
        "\[DoubleLongLeftArrow]", 
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {"0", "\[DoubleLongLeftArrow]", "otherwise"}
      }]}]}], TraditionalForm]]],
 "\n\nWe again distinguish two cases, based on the remaining \
piecewise-expression.\nCase II.II.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}], TraditionalForm]]],
 ".\nWe have to show\n(G15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "=", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], "*", 
     RowBox[{"C", "[", 
      RowBox[{"r", ",", 
       RowBox[{"t", "/", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}]}], TraditionalForm]]],
 "\nA sufficient condition for (G15) to hold is\n(G16)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"r", ",", 
      RowBox[{"t", "/", 
       RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "=", "0"}], 
   TraditionalForm]]],
 "\nBecause of our very first case assumption ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"R", "[", "r", "]"}], "=", "0"}], TraditionalForm]]],
 ", a sufficient condition for (G16) to hold is\n(G17)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"t", "/", 
     RowBox[{"lp", "[", "p", "]"}]}], "\[NotEqual]", 
    RowBox[{"lp", "[", "r", "]"}]}], TraditionalForm]]],
 "\nFrom here the proof proceeds just as in case I.II.I.\n\nCase II.II.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]}], 
   TraditionalForm]]],
 ".\nTrivial (",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "=", "0"}], TraditionalForm]]],
 ")."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642476637496114*^9, 3.642476821952773*^9}, {
   3.64247686524122*^9, 3.642477129657426*^9}, {3.642477213856803*^9, 
   3.642477268542186*^9}, {3.642477313431546*^9, 3.642477313431663*^9}, {
   3.64247736440546*^9, 3.642477392856022*^9}, {3.642477433170116*^9, 
   3.6424778273272543`*^9}, {3.642477915028656*^9, 3.642478121298232*^9}, {
   3.642478178871491*^9, 3.642478275512038*^9}, {3.642478309076434*^9, 
   3.6424785047094812`*^9}, 3.642482082250555*^9, 3.643446900118136*^9, 
   3.6532081928168755`*^9, 3.6532137566642637`*^9, 3.6661740275157795`*^9, 
   3.666440608819084*^9},
 CellID->862219043],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.64248206418983*^9, 3.642482082250731*^9}, 
   3.64344690011826*^9, 3.6532081928168755`*^9, 3.6532137566642637`*^9, 
   3.6661740275157795`*^9, 3.6664406088192253`*^9},
 TextAlignment->Right,
 CellID->1027179770]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Recursive Description of Multiplication (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.640408496199523*^9, 3.640408503627336*^9}, {
  3.640421659198515*^9, 3.640421663305501*^9}},
 CellID->1591556425],

Cell["", "OpenEnvironment",
 CellID->742162515],

Cell[CellGroupData[{

Cell["Definition: hasMultiplicationRec", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.640408514168687*^9, 3.640408528716893*^9}},
 CellID->1404385962],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "\[ScriptCapitalD]"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"hasMultiplicationRec", "[", "\[ScriptCapitalD]", "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
             RowBox[{"[", "p", "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{"p", 
              UnderscriptBox["*", "\[ScriptCapitalD]"], 
              UnderscriptBox["0", "\[ScriptCapitalD]"]}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalD]"]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
             RowBox[{"[", "q", "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              UnderscriptBox["0", "\[ScriptCapitalD]"], 
              UnderscriptBox["*", "\[ScriptCapitalD]"], "q"}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalD]"]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
              RowBox[{"[", 
               RowBox[{"p", ",", "q"}], "]"}]}]], 
            RowBox[{
             RowBox[{"p", "\[NotEqual]", 
              UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{"q", "\[NotEqual]", 
              UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{"p", 
              UnderscriptBox["*", "\[ScriptCapitalD]"], "q"}], "\[Equal]", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalD]"], "[", "p", 
                    "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalD]"], "[", "q", 
                    "]"}]}], ")"}], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalD]"], 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalD]"], "[", "p", 
                    "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalD]"], "[", "q", 
                    "]"}]}], ")"}]}], 
                UnderscriptBox["+", "\[ScriptCapitalD]"], 
                RowBox[{
                 RowBox[{
                  UnderscriptBox["lm", "\[ScriptCapitalD]"], "[", "p", "]"}], 
                 
                 UnderscriptBox["*", "\[ScriptCapitalD]"], 
                 RowBox[{
                  UnderscriptBox["R", "\[ScriptCapitalD]"], "[", "q", 
                  "]"}]}]}], ")"}], 
              UnderscriptBox["+", "\[ScriptCapitalD]"], 
              RowBox[{
               RowBox[{
                UnderscriptBox["R", "\[ScriptCapitalD]"], "[", "p", "]"}], 
               UnderscriptBox["*", "\[ScriptCapitalD]"], "q"}]}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasMultiplicationRec)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1437523678", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640408542043955*^9, 3.6404087319224977`*^9}, {
  3.640418318954204*^9, 3.6404183873433323`*^9}},
 CellTags->{
  "ID:1437523678", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplicationRec"},
 CellID->1437523678],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1332405579]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->593624661],

Cell["", "OpenEnvironment",
 CellID->1778102875],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: Recursive Description of ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " in \[ScriptCapitalP]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.640408825369802*^9, 3.640408828843101*^9}},
 CellID->692234298],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasMultiplicationRec", "[", "\[ScriptCapitalP]", "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1372918741}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640410224863131*^9, 3.6404102291045237`*^9}},
 CellTags->{
  "ID:1372918741", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1372918741],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "p", "]"}]}]], 
   RowBox[{
    UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "p", "]"}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", "q", "]"}]}]], 
     RowBox[{
      RowBox[{"q", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}], 
       "\[Equal]", "t"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"p", 
       UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
      RowBox[{"p", 
       UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(316)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:271668948", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640417444040613*^9, 3.6404175024382343`*^9}},
 CellTags->{
  "ID:271668948", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "316"},
 CellID->271668948],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"p", ",", "q"}], "]"}]}]], 
   RowBox[{
    UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "p", "]"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
    RowBox[{"p", 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(317)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:428541408", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6404175244624863`*^9, 3.640417531129531*^9}},
 CellTags->{
  "ID:428541408", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "317"},
 CellID->428541408],

Cell[BoxData[
 RowBox[{"hasMultiplication", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(318)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1983432019", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640408837171041*^9, 3.640408851206587*^9}, 
   3.640410231139998*^9},
 CellTags->{
  "ID:1983432019", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "318"},
 CellID->1983432019],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->183027490]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1206502385],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(316)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(316)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"hasMultiplicationRec", "[", "\[ScriptCapitalP]", "]"}], 
         "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                  StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{
              UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
              StyleBox["p", "ExpressionVariable"], "]"}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], "\[Equal]", 
                    StyleBox["t", "ExpressionVariable"]}]}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                   StyleBox[
                   "q", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"hasMultiplicationRec", "[", "\[ScriptCapitalP]", "]"}], 
         "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                  StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{
              UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
              StyleBox["p", "ExpressionVariable"], "]"}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], "\[Equal]", 
                    StyleBox["t", "ExpressionVariable"]}]}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                   StyleBox[
                   "q", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:271668948"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p271668948-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p271668948-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:271668948", "Proof|ID:271668948-1"},
 CellID->1133492443],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p271668948-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->2056995531]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(317)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(317)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"hasMultiplicationRec", "[", "\[ScriptCapitalP]", "]"}], 
         "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                  StyleBox["p", "ExpressionVariable"], "]"}]}, {
                RowBox[{
                  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                  StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{
              UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
              StyleBox["p", "ExpressionVariable"], "]"}]], 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox[
               "q", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"hasMultiplicationRec", "[", "\[ScriptCapitalP]", "]"}], 
         "\[Implies]", 
         RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                  StyleBox["p", "ExpressionVariable"], "]"}]}, {
                RowBox[{
                  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                  StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{
              UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
              StyleBox["p", "ExpressionVariable"], "]"}]], 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox[
               "q", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:428541408"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p428541408-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p428541408-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:428541408", "Proof|ID:428541408-1"},
 CellID->738256620],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p428541408-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1481919051]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(318)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(318)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"hasMultiplicationRec", "[", "\[ScriptCapitalP]", "]"}], 
         "\[Implies]", 
         RowBox[{
          "hasMultiplication", "[", "\[ScriptCapitalP]", 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"hasMultiplicationRec", "[", "\[ScriptCapitalP]", "]"}], 
         "\[Implies]", 
         RowBox[{
          "hasMultiplication", "[", "\[ScriptCapitalP]", 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1983432019"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1983432019-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1983432019-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1983432019", "Proof|ID:1983432019-1"},
 CellID->836120509],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1983432019-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1178631419]
}, Closed]],

Cell[TextData[{
 "Formula (318) is stated and proved for the arbitrary but fixed domain \
\[ScriptCapitalP]. Hence, using a meta-argument, we can generalize this \
result to ",
 StyleBox["all",
  FontSlant->"Italic"],
 " domains \[ScriptCapitalD] satisfying the same prerequisites as currently \
imposed on \[ScriptCapitalP]. Of course, in (318) also the reverse direction \
\[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\[CloseCurlyDoubleQuote] holds, \
but this has not been proved formally yet.\nFormula (319) might be useful \
when proving that a concrete functor ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " constructs reduction-polynomial domains, if multiplication in ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is defined completely recursively."
}], "Text",
 CellChangeTimes->{{3.640421461214284*^9, 3.6404216522013597`*^9}, {
  3.640421934845295*^9, 3.64042199122703*^9}},
 CellID->2109312154],

Cell["", "OpenEnvironment",
 CellID->814159960],

Cell[CellGroupData[{

Cell[TextData[{
 "Proposition: Recursive Description of ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " in Arbitrary Domains"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.640345609967738*^9, 3.6403456168953943`*^9}, {
  3.640408795662487*^9, 3.640408816672242*^9}, {3.640421446447566*^9, 
  3.6404214528171253`*^9}, {3.640509521238546*^9, 3.6405095231013117`*^9}},
 CellID->1131151213],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "\[ScriptCapitalD]"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"hasDecisionPredicate", "[", "\[ScriptCapitalD]", "]"}], 
      "\[And]", 
      RowBox[{"containsMonomials", "[", "\[ScriptCapitalD]", "]"}], "\[And]", 
      
      RowBox[{"hasCoefficientFunction", "[", "\[ScriptCapitalD]", "]"}], 
      "\[And]", 
      RowBox[{"isExtensional", "[", "\[ScriptCapitalD]", "]"}], "\[And]", 
      RowBox[{"hasMultiplicationRec", "[", "\[ScriptCapitalD]", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{"hasMultiplication", "[", "\[ScriptCapitalD]", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(319)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1310550300", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.640345774488427*^9, {3.640346188244124*^9, 3.640346203492168*^9}, {
   3.640421374228786*^9, 3.640421431197691*^9}},
 CellTags->{
  "ID:1310550300", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "319"},
 CellID->1310550300],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2135849630]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1678919163]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " is Closed (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.636440502681528*^9, 3.6364405145399313`*^9}, {
  3.638510648521378*^9, 3.6385106516146593`*^9}, {3.6399113207612467`*^9, 
  3.6399113251645203`*^9}},
 CellID->1254493287],

Cell["\<\
From now on, for the rest of this section, we assume that \[ScriptCapitalP] \
also satisfies hasMultiplication.\
\>", "Text",
 CellChangeTimes->{{3.640319041543379*^9, 3.640319060892338*^9}},
 CellID->274886768],

Cell["", "OpenEnvironment",
 CellID->1237337668],

Cell[CellGroupData[{

Cell["\<\
Presupposition: \[ScriptCapitalP] has Multiplication\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.640319015885789*^9, 3.640319026919582*^9}},
 CellID->1046764793],

Cell[BoxData[
 RowBox[{"hasMultiplication", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasMultiplication \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:978297022", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639713284765295*^9, 3.639713316625654*^9}},
 CellTags->{
  "ID:978297022", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplication \[ScriptCapitalP]"},
 CellID->978297022],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->281330312]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1221052070],

Cell["", "OpenEnvironment",
 CellID->1875157683],

Cell[CellGroupData[{

Cell["Definition and Lemma", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.63990795333393*^9, {3.639985669791873*^9, 3.6399856725258703`*^9}},
 CellID->1672617545],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(320)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1125626334", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639907961603986*^9, 3.639907996651778*^9}, {
  3.6399080429860153`*^9, 3.639908055084917*^9}, {3.6399080897413807`*^9, 
  3.63990809660917*^9}, {3.6399081287788877`*^9, 3.6399081359200983`*^9}, {
  3.639908208085927*^9, 3.6399082449398727`*^9}, {3.639908669835576*^9, 
  3.63990867220825*^9}, {3.6399087549336*^9, 3.6399088004058323`*^9}, {
  3.6399106835811977`*^9, 3.6399106839248543`*^9}, {3.656762486922068*^9, 
  3.656762514892917*^9}, {3.6567627046672506`*^9, 3.6567627049636507`*^9}},
 CellTags->{
  "ID:1125626334", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "320"},
 CellID->1125626334],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1058712053]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->943462036],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(320)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(320)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1125626334"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1125626334-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1125626334-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1125626334", "Proof|ID:1125626334-1"},
 CellID->372685725],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1125626334-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->131601230]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->898460167],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " is Closed in \[ScriptCapitalP]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638521604301811*^9, 3.638521618170898*^9}, {
  3.638522038744425*^9, 3.638522038744528*^9}},
 CellID->1741618011],

Cell[BoxData[
 RowBox[{"isClosed", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["*", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalP]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* closed \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1103171586", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638521621599609*^9, 3.638521644368026*^9}, 
   3.6385220416323557`*^9},
 CellTags->{
  "ID:1103171586", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "* closed \[ScriptCapitalP]"},
 CellID->1103171586],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1874612924]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1190806585],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* closed \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* closed \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isClosed", "[", 
         RowBox[{
           UnderscriptBox["*", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isClosed", "[", 
         RowBox[{
           UnderscriptBox["*", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1103171586"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1103171586-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1103171586-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1103171586", "Proof|ID:1103171586-1"},
 CellID->134774430],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1103171586-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->117381029]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " Left-Distributes over ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.636440564521113*^9, 3.6364405677496967`*^9}, {
  3.638510683607689*^9, 3.638510693658139*^9}, {3.638517209471623*^9, 
  3.638517217127304*^9}, {3.638517283486546*^9, 3.638517284227837*^9}, {
  3.639731576207708*^9, 3.639731578460416*^9}, {3.639986907556426*^9, 
  3.6399869085223417`*^9}, {3.6399873813988132`*^9, 3.639987385224661*^9}},
 CellID->1940956769],

Cell["", "OpenEnvironment",
 CellID->927230277],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " Left-Distributes over ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.639985482725357*^9, 3.63998548556503*^9}, {
  3.639985689859454*^9, 3.639985691656949*^9}, {3.6567640812010746`*^9, 
  3.6567640955375*^9}},
 CellID->856550879],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "r"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["*", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{"q", 
       UnderscriptBox["+", "\[ScriptCapitalP]"], "r"}], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"p", 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], 
     UnderscriptBox["+", "\[ScriptCapitalP]"], 
     RowBox[{"p", 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "r"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(322)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2097953227", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639986976241728*^9, 3.6399870151963263`*^9}},
 CellTags->{
  "ID:2097953227", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "322"},
 CellID->2097953227],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1110919542]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->44075770],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(322)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(322)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], 
             UnderscriptBox["+", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], 
             UnderscriptBox["+", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2097953227"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2097953227-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2097953227-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2097953227", "Proof|ID:2097953227-1"},
 CellID->163286988],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2097953227-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->389952566]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " is Commutative (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.636440537569436*^9, 3.636440559962349*^9}, {
  3.638510659732807*^9, 3.638510664186558*^9}, {3.640343412937101*^9, 
  3.6403434173480053`*^9}},
 CellID->1034809977],

Cell["", "OpenEnvironment",
 CellID->510359022],

Cell[CellGroupData[{

Cell["Lemma", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.639997404325077*^9, 3.639997406200387*^9}, {
  3.639997755918633*^9, 3.6399977574375143`*^9}},
 CellID->942893362],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", ",", "q"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 587826938}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6399985027365847`*^9, 3.63999851116315*^9}},
 CellTags->{
  "ID:587826938", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->587826938],

Cell[BoxData[
 RowBox[{
  RowBox[{"p", 
   UnderscriptBox["*", "\[ScriptCapitalP]"], 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Equal]", 
  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(324)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:384020988", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6399985360968437`*^9, 3.639998549804644*^9}},
 CellTags->{
  "ID:384020988", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "324"},
 CellID->384020988],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "p", "]"}], 
  "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
    RowBox[{"p", 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(325)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:205407574", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639997277136736*^9, 3.639997310816482*^9}},
 CellTags->{
  "ID:205407574", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "325"},
 CellID->205407574],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "p", "]"}], 
  "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
    RowBox[{"q", 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(327)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:411846169", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.639997410663208*^9},
 CellTags->{
  "ID:411846169", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "327"},
 CellID->411846169],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "q", "]"}], 
  "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
    RowBox[{"q", 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(328)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1300778952", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6403413135098133`*^9, 3.640341352436223*^9}},
 CellTags->{
  "ID:1300778952", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "328"},
 CellID->1300778952],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
    RowBox[{"q", 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(329)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2123221051", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639997712813637*^9, 3.639997726339905*^9}, {
  3.639998627521367*^9, 3.639998633547666*^9}, {3.656765810870722*^9, 
  3.656765816003131*^9}},
 CellTags->{
  "ID:2123221051", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "329"},
 CellID->2123221051],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1579888901]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->890191499],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(324)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(324)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Equal]", 
           UnderscriptBox[
           "0", "\[ScriptCapitalP]"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Equal]", 
           UnderscriptBox[
           "0", "\[ScriptCapitalP]"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:384020988"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p384020988-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p384020988-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:384020988", "Proof|ID:384020988-1"},
 CellID->119010279],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p384020988-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->9568419]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(325)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(325)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox[
               "q", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox[
               "q", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:205407574"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p205407574-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p205407574-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:205407574", "Proof|ID:205407574-1"},
 CellID->2086485184],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p205407574-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->337622468]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(327)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(327)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "p", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "p", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:411846169"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p411846169-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p411846169-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:411846169", "Proof|ID:411846169-1"},
 CellID->183622553],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p411846169-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->157803399]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(328)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(328)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["q", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "p", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["q", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "p", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1300778952"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1300778952-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1300778952-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1300778952", "Proof|ID:1300778952-1"},
 CellID->695116774],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1300778952-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1300201719]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(329)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(329)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["q", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             StyleBox[
             "p", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["q", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             StyleBox[
             "p", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2123221051"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2123221051-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2123221051-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2123221051", "Proof|ID:2123221051-1"},
 CellID->511445547],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2123221051-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->694303427]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1756416313],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " is Commutative in \[ScriptCapitalP]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638521604301811*^9, 3.638521618170898*^9}, {
  3.638521737152663*^9, 3.6385217392703114`*^9}, {3.638522092688341*^9, 
  3.6385220926884623`*^9}},
 CellID->235023336],

Cell[BoxData[
 RowBox[{"isComm", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["*", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalP]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* comm \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:840093990", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638521621599609*^9, 3.638521644368026*^9}, {
   3.638521746494171*^9, 3.638521746843409*^9}, 3.63852209647372*^9},
 CellTags->{
  "ID:840093990", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "* comm \[ScriptCapitalP]"},
 CellID->840093990],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1016529366]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->206611077],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* comm \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* comm \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isComm", "[", 
         RowBox[{
           UnderscriptBox["*", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isComm", "[", 
         RowBox[{
           UnderscriptBox["*", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:840093990"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p840093990-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p840093990-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:840093990", "Proof|ID:840093990-1"},
 CellID->4680134],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p840093990-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->46498348]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " Distributes over ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{
  3.639986905036626*^9, {3.640343679368679*^9, 3.640343683382303*^9}},
 CellID->1153917976],

Cell["", "OpenEnvironment",
 CellID->970984491],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " is Distributive over ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " in \[ScriptCapitalP]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638521604301811*^9, 3.638521618170898*^9}, {
  3.638521765843896*^9, 3.638521776945801*^9}},
 CellID->1230148421],

Cell[BoxData[
 RowBox[{"isDistr", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["*", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["+", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalP]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* distr \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1730144032", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638521621599609*^9, 3.638521644368026*^9}, {
  3.6385217845356283`*^9, 3.638521794794292*^9}},
 CellTags->{
  "ID:1730144032", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "* distr \[ScriptCapitalP]"},
 CellID->1730144032],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->228018603]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1675122156],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* distr \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* distr \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isDistr", "[", 
         RowBox[{
           UnderscriptBox["*", "\[ScriptCapitalP]"], ",", 
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isDistr", "[", 
         RowBox[{
           UnderscriptBox["*", "\[ScriptCapitalP]"], ",", 
           UnderscriptBox["+", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1730144032"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1730144032-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1730144032-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1730144032", "Proof|ID:1730144032-1"},
 CellID->1430620892],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1730144032-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2019370717]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "1 is Neutral Element of ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.638510669970714*^9, 3.638510678423527*^9}, {
  3.638517301653995*^9, 3.638517301766307*^9}, {3.640345492793055*^9, 
  3.6403454969755497`*^9}},
 CellID->1681172436],

Cell["", "OpenEnvironment",
 CellID->1339692048],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: 1 is Neutral Element of ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " in \[ScriptCapitalP]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638521604301811*^9, 3.638521618170898*^9}, {
  3.6385217075444117`*^9, 3.6385217156387053`*^9}, {3.638522056628742*^9, 
  3.63852205837564*^9}},
 CellID->702316358],

Cell[BoxData[
 RowBox[{"isNeutral", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["*", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   UnderscriptBox["1", "\[ScriptCapitalP]"], ",", "\[ScriptCapitalP]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1 neutral \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:378610433", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638521621599609*^9, 3.638521644368026*^9}, {
  3.638521688268338*^9, 3.6385216988074636`*^9}, {3.6385220616448793`*^9, 
  3.638522063708784*^9}},
 CellTags->{
  "ID:378610433", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1 neutral \[ScriptCapitalP]"},
 CellID->378610433],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->585549701]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->348109801],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1 neutral \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1 neutral \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isNeutral", "[", 
         RowBox[{
           UnderscriptBox["*", "\[ScriptCapitalP]"], ",", 
           UnderscriptBox["1", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isNeutral", "[", 
         RowBox[{
           UnderscriptBox["*", "\[ScriptCapitalP]"], ",", 
           UnderscriptBox["1", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:378610433"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p378610433-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p378610433-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:378610433", "Proof|ID:378610433-1"},
 CellID->149156064],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p378610433-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1484103856]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " is Associative (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.6364405254457273`*^9, 3.636440557159268*^9}, {
  3.638510653797963*^9, 3.638510657696616*^9}, {3.6405081944736013`*^9, 
  3.640508198930838*^9}},
 CellID->231588462],

Cell["", "OpenEnvironment",
 CellID->1995953378],

Cell[CellGroupData[{

Cell["Lemma", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.640428190987962*^9},
 CellID->776216969],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "r"}], "]"}]}]], 
  RowBox[{"p", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 687899447}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6404284950128736`*^9, 3.640428506825322*^9}, {
   3.6404295110832863`*^9, 3.640429516182169*^9}, 3.6404297269887257`*^9},
 CellTags->{
  "ID:687899447", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->687899447],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "q", "]"}]}]], 
   RowBox[{
    UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "q", "]"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{"q", 
       UnderscriptBox["*", "\[ScriptCapitalP]"], "r"}], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"p", 
       UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], ")"}], 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "r"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(330)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:421376854", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640428523838346*^9, 3.6404286353278923`*^9}, {
  3.6404296836078978`*^9, 3.6404296868191843`*^9}, {3.6404297295102253`*^9, 
  3.640429738094656*^9}},
 CellTags->{
  "ID:421376854", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "330"},
 CellID->421376854],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "q", "]"}]}]], 
    RowBox[{
     RowBox[{"q", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}], "\[Equal]",
       "t"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"p", 
      UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
      RowBox[{"(", 
       RowBox[{"q", 
        UnderscriptBox["*", "\[ScriptCapitalP]"], "r"}], ")"}]}], "\[Equal]", 
     
     RowBox[{
      RowBox[{"(", 
       RowBox[{"p", 
        UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], ")"}], 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "r"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(331)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:935511285", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640430548310498*^9, 3.6404305820757713`*^9}},
 CellTags->{
  "ID:935511285", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "331"},
 CellID->935511285],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "q", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{"q", 
       UnderscriptBox["*", "\[ScriptCapitalP]"], "r"}], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"p", 
       UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], ")"}], 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "r"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(332)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1702504479", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640430232804306*^9, 3.640430292761046*^9}},
 CellTags->{
  "ID:1702504479", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "332"},
 CellID->1702504479],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "q", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", "a", "]"}]}]], 
     RowBox[{
      RowBox[{"a", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "a", "]"}], 
       "\[Equal]", "t"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"a", 
       UnderscriptBox["*", "\[ScriptCapitalP]"], 
       RowBox[{"(", 
        RowBox[{"q", 
         UnderscriptBox["*", "\[ScriptCapitalP]"], "r"}], ")"}]}], "\[Equal]", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"a", 
         UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], ")"}], 
       UnderscriptBox["*", "\[ScriptCapitalP]"], "r"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(333)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1379089132", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640432555115747*^9, 3.640432655937952*^9}},
 CellTags->{
  "ID:1379089132", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "333"},
 CellID->1379089132],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->855336169]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1874167721],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(330)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(330)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:421376854"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p421376854-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p421376854-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:421376854", "Proof|ID:421376854-1"},
 CellID->1691973429],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p421376854-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->275458490]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(331)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(331)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:935511285"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p935511285-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p935511285-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:935511285", "Proof|ID:935511285-1"},
 CellID->1225679565],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p935511285-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellChangeTimes->{3.64049510545455*^9},
 CellID->787462530]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(332)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(332)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1702504479"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1702504479-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1702504479-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1702504479", "Proof|ID:1702504479-1"},
 CellID->1881341258],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1702504479-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->447670728]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(333)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(333)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["r", "ExpressionVariable"]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1379089132"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1379089132-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1379089132-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1379089132", "Proof|ID:1379089132-1"},
 CellID->1247171944],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1379089132-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1896041554]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->973193914],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " is Associative in \[ScriptCapitalP]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638521604301811*^9, 3.638521618170898*^9}, {
  3.638521720759623*^9, 3.638521722767658*^9}, {3.63852207669317*^9, 
  3.638522076693272*^9}},
 CellID->1831517338],

Cell[BoxData[
 RowBox[{"isAssoc", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["*", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalP]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* assoc \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:167864408", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638521621599609*^9, 3.638521644368026*^9}, {
   3.638521728923695*^9, 3.6385217297844276`*^9}, 3.638522079041182*^9},
 CellTags->{
  "ID:167864408", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "* assoc \[ScriptCapitalP]"},
 CellID->167864408],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1148171460]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1060151895],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* assoc \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* assoc \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isAssoc", "[", 
         RowBox[{
           UnderscriptBox["*", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isAssoc", "[", 
         RowBox[{
           UnderscriptBox["*", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:167864408"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p167864408-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p167864408-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:167864408", "Proof|ID:167864408-1"},
 CellID->1541468547],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p167864408-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1970909722]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Result (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6364404575889482`*^9, 3.636440459980543*^9}, {
  3.640508691967772*^9, 3.640508696949679*^9}},
 CellID->1576202806],

Cell["", "OpenEnvironment",
 CellID->205309231],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[ScriptCapitalP] Constitutes Commutative Ring with Unit\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.63637038076784*^9, 3.636370396141884*^9}, {
  3.643444092981563*^9, 3.643444095196579*^9}},
 CellID->1117923091],

Cell[BoxData[
 RowBox[{"isCommRing1", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCommRing1 \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1082265416", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.636370400052515*^9, 3.636370408568747*^9}},
 CellTags->{
  "ID:1082265416", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCommRing1 \[ScriptCapitalP]"},
 CellID->1082265416],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->583569111]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->685210046],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCommRing1 \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCommRing1 \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "isCommRing1", "[", "\[ScriptCapitalP]", 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "isCommRing1", "[", "\[ScriptCapitalP]", 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1082265416"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1082265416-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1082265416-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1082265416", "Proof|ID:1082265416-1"},
 CellID->408658241],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1082265416-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->347038290]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["More Corollaries (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.639805547243822*^9, 3.639805550161579*^9}, {
   3.640668157905594*^9, 3.6406681623619833`*^9}, 3.6410233752742023`*^9, {
   3.6410259667235937`*^9, 3.6410259707119503`*^9}},
 CellID->1536283728],

Cell[CellGroupData[{

Cell["Selectors (C, R, lp, H) (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6318728122139797`*^9, 3.631872826146825*^9}, 
   3.632820156905813*^9, 3.638517711472454*^9, 3.64101740191325*^9, {
   3.641025956892812*^9, 3.6410259609870987`*^9}, {3.6423082255039053`*^9, 
   3.642308239603387*^9}},
 CellID->4076126],

Cell["", "OpenEnvironment",
 CellID->976925363],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["lp", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["H", TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.633172282295032*^9, 3.633172285963449*^9}, {
  3.641017412008944*^9, 3.641017415680071*^9}, {3.6410230454616337`*^9, 
  3.641023046111285*^9}, {3.6423081529319963`*^9, 3.6423081557362843`*^9}, {
  3.64230822986395*^9, 3.6423082349015837`*^9}},
 CellID->1342603728],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1610244238}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633172318481805*^9, 3.633172326769643*^9}, {
  3.641017204882807*^9, 3.641017219689208*^9}},
 CellTags->{
  "ID:1610244238", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1610244238],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}], "\[Implies]", 
    
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
         RowBox[{
          RowBox[{
           UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
          "t"}], "]"}], "\[Equal]", 
        RowBox[{
         RowBox[{
          UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
          RowBox[{"p", ",", "t"}], "]"}], 
         UnderscriptBox["-", "\[ScriptCapitalP]"], 
         RowBox[{
          UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
         RowBox[{
          RowBox[{
           UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
          "t"}], "]"}], "\[Equal]", 
        RowBox[{
         UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
         RowBox[{"p", ",", "t"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R 5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1259587933", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640519215760942*^9, 3.6405192401357603`*^9}, {
  3.640519306875205*^9, 3.640519377334115*^9}, {3.640577432766129*^9, 
  3.640577454232326*^9}},
 CellTags->{
  "ID:1259587933", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 5"},
 CellID->1259587933],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"t", ",", "s"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "r", "]"}]}]], 
    RowBox[{
     RowBox[{"r", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "r", "]"}], "\[Equal]",
       "s"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"t", 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "r", "]"}], 
       UnderscriptBox["*", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}], 
     "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{
         RowBox[{"r", 
          UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], ",", "t"}], "]"}], 
       "\[Equal]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C 6 ind)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:859675419", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641022820784218*^9, 3.6410229376734543`*^9}, {
  3.641022983810668*^9, 3.641022986612671*^9}},
 CellTags->{
  "ID:859675419", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6 ind"},
 CellID->859675419],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"t", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
      UnderscriptBox["*", "\[ScriptCapitalT]"], 
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}], 
    "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{
        RowBox[{"p", 
         UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], ",", "t"}], "]"}], 
      "\[Equal]", 
      UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ")"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C 6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:515125163", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633173760450671*^9, 3.6331738148562193`*^9}, {
  3.635747385103512*^9, 3.635747387053157*^9}, {3.64101722464392*^9, 
  3.641017228644443*^9}, {3.641022791285234*^9, 3.6410227990912457`*^9}},
 CellTags->{
  "ID:515125163", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6"},
 CellID->515125163],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"q", 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}], ",", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}], 
     UnderscriptBox["*", "\[ScriptCapitalT]"], 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}], "]"}], 
  "\[Equal]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], 
   UnderscriptBox["*", "\[ScriptCapitalR]"], 
   RowBox[{
    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(C 6')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1572844749", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318740647293987`*^9, 3.631874134009179*^9}, 
   3.635747698712763*^9, {3.638784721975882*^9, 3.638784752005723*^9}, {
   3.641017230262566*^9, 3.6410172337743196`*^9}, {3.641017314927844*^9, 
   3.641017318409062*^9}},
 CellTags->{
  "ID:1572844749", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6'"},
 CellID->1572844749],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"p", 
    UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
  RowBox[{"\[Not]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", 
       UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], "]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
      UnderscriptBox["*", "\[ScriptCapitalT]"], 
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lp 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:771023286", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632820204755372*^9, 3.6328202648210297`*^9}, {
   3.632820325095829*^9, 3.632820348417701*^9}, 3.633172787077157*^9, {
   3.6357474298339157`*^9, 3.6357474473182*^9}, {3.641017341592339*^9, 
   3.641017353048778*^9}, {3.64101742577394*^9, 3.641017428944017*^9}},
 CellTags->{
  "ID:771023286", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 4"},
 CellID->771023286],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], ",", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
     UnderscriptBox["*", "\[ScriptCapitalT]"], 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}], "]"}], 
  "\[Equal]", 
  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(H 9)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:422533310", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63317359733919*^9, 3.633173644666679*^9}, {
  3.633173701566676*^9, 3.63317371664423*^9}, {3.6357475317920218`*^9, 
  3.635747533149961*^9}, {3.6410173829212513`*^9, 3.641017385814392*^9}},
 CellTags->{
  "ID:422533310", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 9"},
 CellID->422533310],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->753430840]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1429212256],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R 5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R 5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1259587933"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1259587933-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1259587933-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1259587933", "Proof|ID:1259587933-1"},
 CellID->1687842700],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1259587933-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->2099506532]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C 6 ind)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C 6 ind)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["r", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["r", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["s", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["r", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["r", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["r", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["s", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["r", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:859675419"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p859675419-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p859675419-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:859675419", "Proof|ID:859675419-1"},
 CellID->934287773],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p859675419-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1566978955]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C 6)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C 6)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:515125163"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p515125163-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p515125163-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:515125163", "Proof|ID:515125163-1"},
 CellID->50230006],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p515125163-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1598865118]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(C 6')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(C 6')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["p", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], "]"}], 
             "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}], 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["p", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], "]"}], 
             "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}], 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1572844749"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1572844749-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1572844749-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1572844749", "Proof|ID:1572844749-1"},
 CellID->1143100247],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1572844749-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1805751181]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lp 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lp 4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox["q", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox["q", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:771023286"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p771023286-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p771023286-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:771023286", "Proof|ID:771023286-1"},
 CellID->38527445],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p771023286-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1331059404]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(H 9)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(H 9)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], "]"}], 
             "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], "]"}], 
             "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:422533310"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p422533310-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p422533310-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:422533310", "Proof|ID:422533310-1"},
 CellID->156791201],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p422533310-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1226801151]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Monomials (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.641026031798092*^9, 3.641026033119252*^9}, {
  3.641033115731059*^9, 3.641033119993168*^9}},
 CellID->362702593],

Cell["", "OpenEnvironment",
 CellID->345925313],

Cell[CellGroupData[{

Cell["Lemma: Properties of the Monomial-Constructor", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.641026042818983*^9, 3.641026056160281*^9}},
 CellID->2037599412],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{"c", ",", "d"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1667882379}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320401373168383`*^9, 3.632040146353407*^9}},
 CellTags->{
  "ID:1667882379", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1667882379],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "t", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1105090112}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632040149780904*^9, 3.632040157057767*^9}},
 CellTags->{
  "ID:1105090112", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1105090112],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "s", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"c", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
     UnderscriptBox["*", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{"d", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "s"}], ")"}]}], 
    "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"c", 
       UnderscriptBox["*", "\[ScriptCapitalR]"], "d"}], ")"}], 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{"t", 
       UnderscriptBox["*", "\[ScriptCapitalT]"], "s"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:878771843", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632807858123021*^9, 3.632807920382825*^9}},
 CellTags->{
  "ID:878771843", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 2"},
 CellID->878771843],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "p", "]"}]}]], 
  RowBox[{"p", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1863009930}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632118438870522*^9, 3.6321184585732203`*^9}, 
   3.635747673485086*^9},
 CellTags->{
  "ID:1863009930", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1863009930],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "s", "]"}]}]], 
   RowBox[{"s", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
    RowBox[{"t", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"c", 
         UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
       UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}], ",", "s"}], "]"}], 
    "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1958431161", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632117310529849*^9, 3.6321174005980797`*^9}, 
   3.6321184608560543`*^9, {3.632717135324456*^9, 3.632717144994994*^9}, {
   3.6385968005511923`*^9, 3.6385968043910093`*^9}},
 CellTags->{
  "ID:1958431161", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 5"},
 CellID->1958431161],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"c", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}], ",", 
    RowBox[{"t", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}], "]"}], 
  "\[Equal]", 
  RowBox[{"c", 
   UnderscriptBox["*", "\[ScriptCapitalR]"], 
   RowBox[{
    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:758577094", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632118403547983*^9, 3.632118424526535*^9}, {
  3.6321184621813173`*^9, 3.632118470114595*^9}, {3.632118592201701*^9, 
  3.632118605840317*^9}, {3.632717153190661*^9, 3.632717161470213*^9}},
 CellTags->{
  "ID:758577094", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 6"},
 CellID->758577094],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", "q", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 544874520}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632818038155725*^9, 3.632818046442293*^9}},
 CellTags->{
  "ID:544874520", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->544874520],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "s", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 158305602}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6328180598589163`*^9, 3.632818067087756*^9}},
 CellTags->{
  "ID:158305602", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->158305602],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"c", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "s"}], ")"}], 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], ",", 
    RowBox[{"s", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}]}], "]"}], "\[Equal]", 
  RowBox[{"c", 
   UnderscriptBox["*", "\[ScriptCapitalR]"], 
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"q", ",", "t"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 9)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:347486225", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632818088267859*^9, 3.632818110598448*^9}, 
   3.6328182588528547`*^9},
 CellTags->{
  "ID:347486225", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 9"},
 CellID->347486225],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
    RowBox[{"q", ",", "t"}], "]"}], "\[Equal]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"c", 
        UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "s"}], ")"}], 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], ",", 
     RowBox[{"s", 
      UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}]}], "]"}], "\[Equal]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 10)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1591957532", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6328168334887133`*^9, 3.63281695932049*^9}, {
  3.6328179458365192`*^9, 3.632817967687447*^9}, {3.63281804832102*^9, 
  3.632818071948844*^9}},
 CellTags->{
  "ID:1591957532", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 10"},
 CellID->1591957532],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", "c", "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1155967768}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650621108764635*^9, 3.6506211162665052`*^9}},
 CellTags->{
  "ID:1155967768", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1155967768],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"c", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"q", "\[Equal]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 15)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1942982211", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650624933004737*^9, 3.650625021073099*^9}},
 CellTags->{
  "ID:1942982211", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 15"},
 CellID->1942982211],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"c", 
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
    UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}], "]"}], "\[Equal]", 
  RowBox[{"t", 
   UnderscriptBox["*", "\[ScriptCapitalT]"], 
   RowBox[{
    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 12)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:44857450", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6506211345437202`*^9, 3.650621157087831*^9}, {
  3.650621211758403*^9, 3.650621212862523*^9}},
 CellTags->{
  "ID:44857450", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 12"},
 CellID->44857450],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"c", 
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
    UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}], "]"}], "\[Equal]", 
  RowBox[{"c", 
   UnderscriptBox["*", "\[ScriptCapitalR]"], 
   RowBox[{
    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 13)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:928737230", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6506211665695963`*^9, 3.650621172059847*^9}, {
  3.650621214003196*^9, 3.65062121521736*^9}, {3.6506234463642178`*^9, 
  3.650623447658442*^9}},
 CellTags->{
  "ID:928737230", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 13"},
 CellID->928737230],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"c", 
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
    UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "q", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 14)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:233618042", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6506206171105757`*^9, 3.6506206825973597`*^9}, {
  3.650621012467861*^9, 3.65062101338306*^9}, {3.6506210793270187`*^9, 
  3.6506210850348663`*^9}, {3.650621118740262*^9, 3.6506211226952133`*^9}},
 CellTags->{
  "ID:233618042", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 14"},
 CellID->233618042],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1441920355]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->614488126],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["s", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               StyleBox["d", "ExpressionVariable"]}], 
             UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox[
               "s", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["s", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               StyleBox["d", "ExpressionVariable"]}], 
             UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               StyleBox[
               "s", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:878771843"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p878771843-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p878771843-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:878771843", "Proof|ID:878771843-1"},
 CellID->395620750],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p878771843-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->856996584]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["p", "ExpressionVariable"]}], ",", 
                 StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["p", "ExpressionVariable"]}], ",", 
                 StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1958431161"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1958431161-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1958431161-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1958431161", "Proof|ID:1958431161-1"},
 CellID->756678784],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1958431161-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->871051679]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 6)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 6)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["p", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], "]"}], 
             "\[Equal]", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["p", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], "]"}], 
             "\[Equal]", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:758577094"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p758577094-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p758577094-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:758577094", "Proof|ID:758577094-1"},
 CellID->2118355576],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p758577094-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->2066466981]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 9)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 9)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["s", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}]}], "]"}], "\[Equal]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalR]"], 
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["s", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}]}], "]"}], "\[Equal]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], 
             UnderscriptBox["*", "\[ScriptCapitalR]"], 
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:347486225"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p347486225-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p347486225-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:347486225", "Proof|ID:347486225-1"},
 CellID->1908653227],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p347486225-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.642494479468177*^9, 3.642494489943001*^9}},
 CellID->1535111932]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 10)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 10)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["s", "ExpressionVariable"]}], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   StyleBox["t", "ExpressionVariable"]}]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["s", "ExpressionVariable"]}], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["q", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   StyleBox["t", "ExpressionVariable"]}]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1591957532"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1591957532-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1591957532-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1591957532", "Proof|ID:1591957532-1"},
 CellID->1145145855],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1591957532-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->311029242]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 15)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 15)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalP]"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalP]"]}]}]}]}], $CellContext`TheoremaForm]],
      "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1942982211"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1942982211-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1942982211-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1942982211", "Proof|ID:1942982211-1"},
 CellID->1270512064],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1942982211-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1236173956]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 12)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 12)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:44857450"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p44857450-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p44857450-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:44857450", "Proof|ID:44857450-1"},
 CellID->584781971],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p44857450-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->834128854]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 13)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 13)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:928737230"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p928737230-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p928737230-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:928737230", "Proof|ID:928737230-1"},
 CellID->103471022],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p928737230-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1716227546]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 14)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 14)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:233618042"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p233618042-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p233618042-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:233618042", "Proof|ID:233618042-1"},
 CellID->424912135],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p233618042-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->513975655]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->135337510],

Cell[CellGroupData[{

Cell["Lemma: Properties of Monomials", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631872352128562*^9, 3.631872357751933*^9}},
 CellID->763939551],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", 
   "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 183428126}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318723626051283`*^9, 3.6318723705880613`*^9}, {
  3.6318726676373873`*^9, 3.631872675929371*^9}},
 CellTags->{
  "ID:183428126", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->183428126],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "p", "]"}]}]], 
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
  RowBox[{
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{
     RowBox[{"m", 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}], ",", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
      UnderscriptBox["*", "\[ScriptCapitalT]"], 
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}], "]"}], 
   "\[Equal]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}], 
    UnderscriptBox["*", "\[ScriptCapitalR]"], 
    RowBox[{
     UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", 
     "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2026663200", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318740647293987`*^9, 3.631874134009179*^9}, 
   3.635747698712763*^9},
 CellTags->{
  "ID:2026663200", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 2"},
 CellID->2026663200],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"c", 
        UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "m"}], "\[Equal]", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"c", 
        UnderscriptBox["*", "\[ScriptCapitalR]"], 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}]}], ")"}], 
      
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
      RowBox[{"(", 
       RowBox[{"t", 
        UnderscriptBox["*", "\[ScriptCapitalT]"], 
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}]}], 
       ")"}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomial 5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1916930407", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6330720644601097`*^9, 3.633072136144004*^9}, {
   3.63317395509476*^9, 3.633173956350835*^9}, 3.6331739958017*^9, {
   3.641032647507077*^9, 3.6410326554983273`*^9}},
 CellTags->{
  "ID:1916930407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 5"},
 CellID->1916930407],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1683231865]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1360979825],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["p", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], "]"}], 
             "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["p", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], "]"}], 
             "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2026663200"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2026663200-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2026663200-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2026663200", "Proof|ID:2026663200-1"},
 CellID->1385906749],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2026663200-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->870820524]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomial 5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomial 5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["m", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["m", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1916930407"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1916930407-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1916930407-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1916930407", "Proof|ID:1916930407-1"},
 CellID->1619770492],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1916930407-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->833552299]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Recursive Description of the Order Relation on \[ScriptCapitalP] \
(\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.641020125046826*^9, 3.6410201396341877`*^9}, {
  3.641023365863468*^9, 3.641023370069316*^9}},
 CellID->1800295324],

Cell[TextData[{
 "We conclude the section on ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCommRing1", "[", "\[ScriptCapitalP]", "]"}], 
   TraditionalForm]]],
 " with a result that topically would rather fit into the next section, but \
which we want to prove ",
 StyleBox["before",
  FontSlant->"Italic"],
 " we assume that \[ScriptCapitalP] belongs to the category of reduction \
polynomials: Similar as for the multiplication in \[ScriptCapitalP] we also \
want to provide (and prove) a recursive description of the order relation on \
\[ScriptCapitalP].\nFor this, however, we not only need \[ScriptCapitalR] to \
be a commutative ring with unit, but also to be a reduction ring; Therefore, \
henceforth we assume the following:"
}], "Text",
 CellChangeTimes->{{3.6405089064360847`*^9, 3.640509073740329*^9}, {
  3.6405094450858994`*^9, 3.64050944914114*^9}, {3.640579033133876*^9, 
  3.640579056861046*^9}},
 CellID->1851558702],

Cell["", "OpenEnvironment",
 CellID->1582231852],

Cell[CellGroupData[{

Cell["\<\
Presupposition: Reduction Ring \[ScriptCapitalR]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.640508811579526*^9, 3.640508821601674*^9}, 
   3.640508862246997*^9, {3.640579106921541*^9, 3.640579112428445*^9}},
 CellID->2109816331],

Cell[BoxData[
 RowBox[{"isReductionRing", "[", "\[ScriptCapitalR]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduction ring \[ScriptCapitalR])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1225634217", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631600825048706*^9, 3.631600830325115*^9}},
 CellTags->{
  "ID:1225634217", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduction ring \[ScriptCapitalR]"},
 CellID->1225634217],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1230264783]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->665224426],

Cell["", "OpenEnvironment",
 CellID->556263379],

Cell[CellGroupData[{

Cell["Lemma", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.640512445606241*^9},
 CellID->1625326245],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "p", "]"}]}]], 
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "c", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{
         RowBox[{
          UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
         "t"}], "]"}], 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], "c"}], 
      "\[Implies]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(300)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1542801947", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640512448738899*^9, 3.640512518745963*^9}, {
  3.64051255245872*^9, 3.640512592748652*^9}, {3.640512847852826*^9, 
  3.640512865572035*^9}, {3.6405787307983847`*^9, 3.6405787577706223`*^9}, {
  3.64058237395613*^9, 3.6405824166009607`*^9}},
 CellTags->{
  "ID:1542801947", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "300"},
 CellID->1542801947],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->445923411]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->824716703],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(300)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(300)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1542801947"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1542801947-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1542801947-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1542801947", "Proof|ID:1542801947-1"},
 CellID->1097904678],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1542801947-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1371063757]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1307939572],

Cell[CellGroupData[{

Cell["Definition: isOrderedRec", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.640509098742701*^9, 3.640509101708096*^9}},
 CellID->36053659],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "\[ScriptCapitalD]"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"isOrderedRec", "[", "\[ScriptCapitalD]", "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
        RowBox[{"[", 
         RowBox[{"p", ",", "q"}], "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"p", 
          UnderscriptBox["\[Succeeds]", "\[ScriptCapitalD]"], "q"}], 
         TagBox[")",
          "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"\[And]", 
          RowBox[{"\[Piecewise]", GridBox[{
             {
              RowBox[{"p", "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalD]"]}]},
             {
              RowBox[{
               RowBox[{"q", "\[NotEqual]", 
                UnderscriptBox["0", "\[ScriptCapitalD]"]}], "\[Implies]", 
               RowBox[{"\[Or]", 
                RowBox[{"\[Piecewise]", GridBox[{
                   {
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalD]"], "[", "p", 
                    "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalD]"], "[", "q", 
                    "]"}]}]},
                   {
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalD]"], "[", "p", 
                    "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalD]"], "[", "q", 
                    "]"}]}], "\[And]", 
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalD]"], "[", "p", 
                    "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalD]"], "[", "q", 
                    "]"}]}], "\[Or]", 
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalD]"], "[", "p", 
                    "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalD]"], "[", "q", 
                    "]"}]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalD]"], "[", "p", "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalD]"], 
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalD]"], "[", "q", 
                    "]"}]}]}], 
                    TagBox[")",
                    "AutoParentheses"]}]}], 
                    TagBox[")",
                    "AutoParentheses"]}]}]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}]}]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isOrderedRec)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:946287998", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640509106679141*^9, 3.640509275397113*^9}, {
  3.640509312707452*^9, 3.640509398207032*^9}, {3.656249779050049*^9, 
  3.6562498015920887`*^9}},
 CellTags->{
  "ID:946287998", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrderedRec"},
 CellID->946287998],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1338223823]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2058247028],

Cell["", "OpenEnvironment",
 CellID->259022577],

Cell[CellGroupData[{

Cell["\<\
Theorem: Recursive Description of \[Succeeds] in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.640509622909814*^9, 3.640509640421837*^9}, {
  3.6406680182380257`*^9, 3.640668019927394*^9}},
 CellID->1942671025],

Cell[BoxData[
 RowBox[{
  RowBox[{"isOrderedRec", "[", "\[ScriptCapitalP]", "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1485445352}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6405117039437323`*^9, 3.6405117115739393`*^9}},
 CellTags->{
  "ID:1485445352", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1485445352],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"p", ",", "q"}], "]"}]}]], 
    RowBox[{
     RowBox[{"p", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]",
       "t"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"p", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "q"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
         RowBox[{"[", "t", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          RowBox[{
           UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
           RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
          RowBox[{
           UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
           RowBox[{"q", ",", "t"}], "]"}]}], "\[And]", 
         RowBox[{
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"p", ",", "t"}], "]"}], 
          UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"q", ",", "t"}], "]"}]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(301)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:215225048", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6405105514613743`*^9, 3.640510557436225*^9}, {
  3.640510594793239*^9, 3.640510735080831*^9}, {3.6405108043401403`*^9, 
  3.640510806752776*^9}, {3.6405113774936457`*^9, 3.640511415529701*^9}, {
  3.640511729966625*^9, 3.6405117492202387`*^9}, {3.64051296825558*^9, 
  3.6405129691485767`*^9}},
 CellTags->{
  "ID:215225048", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "301"},
 CellID->215225048],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"p", ",", "q"}], "]"}]}]], 
    RowBox[{
     RowBox[{"p", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]",
       "t"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
         RowBox[{"[", "t", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          RowBox[{
           UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
           RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
          RowBox[{
           UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
           RowBox[{"q", ",", "t"}], "]"}]}], "\[And]", 
         RowBox[{
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"p", ",", "t"}], "]"}], 
          UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"q", ",", "t"}], "]"}]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"p", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "q"}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(302)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1704612967", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640513229207926*^9, 3.6405132572153263`*^9}},
 CellTags->{
  "ID:1704612967", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "302"},
 CellID->1704612967],

Cell[BoxData[
 RowBox[{"isOrdered", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(303)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:162248449", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640509663294313*^9, 3.640509692069624*^9}, 
   3.640513220126226*^9},
 CellTags->{
  "ID:162248449", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "303"},
 CellID->162248449],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1460333818]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->771705247],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(301)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(301)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"isOrderedRec", "[", "\[ScriptCapitalP]", "]"}], 
         "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
                  StyleBox["t", "ExpressionVariable"]}]}]], 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    
                    RowBox[{
                    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}]}], "\[And]", 
                   
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"isOrderedRec", "[", "\[ScriptCapitalP]", "]"}], 
         "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
                  StyleBox["t", "ExpressionVariable"]}]}]], 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                 StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    
                    RowBox[{
                    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}]}], "\[And]", 
                   
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:215225048"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p215225048-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p215225048-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:215225048", "Proof|ID:215225048-1"},
 CellID->1531901166],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p215225048-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->361326966]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(302)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(302)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"isOrderedRec", "[", "\[ScriptCapitalP]", "]"}], 
         "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
                  StyleBox["t", "ExpressionVariable"]}]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    
                    RowBox[{
                    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}]}], "\[And]", 
                   
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                 StyleBox[
                 "q", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"isOrderedRec", "[", "\[ScriptCapitalP]", "]"}], 
         "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[ForAll]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
                  StyleBox["t", "ExpressionVariable"]}]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    
                    RowBox[{
                    UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}]}], "\[And]", 
                   
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                 StyleBox[
                 "q", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1704612967"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1704612967-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1704612967-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1704612967", "Proof|ID:1704612967-1"},
 CellID->494680653],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1704612967-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->1603308827]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(303)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(303)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"isOrderedRec", "[", "\[ScriptCapitalP]", "]"}], 
         "\[Implies]", 
         RowBox[{
          "isOrdered", "[", "\[ScriptCapitalP]", 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"isOrderedRec", "[", "\[ScriptCapitalP]", "]"}], 
         "\[Implies]", 
         RowBox[{
          "isOrdered", "[", "\[ScriptCapitalP]", 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:162248449"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p162248449-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p162248449-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:162248449", "Proof|ID:162248449-1"},
 CellID->798928706],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p162248449-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->2125561798]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (301)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.640509697552816*^9, 3.640509698489526*^9}, 
   3.640513011203107*^9, {3.640664602566853*^9, 3.640664605176774*^9}, 
   3.643446944577533*^9, 3.6532081792604513`*^9, 3.653213743622641*^9, 
   3.666174013725355*^9, 3.666440595712885*^9},
 CellID->1465916398],

Cell[TextData[{
 "We perform Noetherian induction on ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 ", i.e. we choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isOrderedRec", "[", "\[ScriptCapitalP]", "]"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "=", "t"}], TraditionalForm]]],
 " and\n(IH)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
         RowBox[{"[", "s", "]"}]}]], 
       RowBox[{"t", "\[Succeeds]", "s"}]], 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
          RowBox[{"[", 
           RowBox[{"p", ",", "q"}], "]"}]}]], 
        RowBox[{
         RowBox[{
          RowBox[{"p", "\[NotEqual]", "0"}], "\[And]", 
          RowBox[{"lp", "[", "p", "]"}]}], "=", "s"}]], 
       RowBox[{"(", 
        RowBox[{"p", "\[Succeeds]", "q"}], ")"}]}]}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "t", "]"}]}]], 
      RowBox[{"H", "[", 
       RowBox[{"p", ",", "t"}], "]"}]}]}], "=", 
    RowBox[{
     RowBox[{"H", "[", 
      RowBox[{"q", ",", "t"}], "]"}], "\[And]", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"p", ",", "t"}], "]"}], "\[Succeeds]", 
      RowBox[{"C", "[", 
       RowBox[{"q", ",", "t"}], "]"}]}]}]}], TraditionalForm]]],
 "\nand prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"\[And]", 
       RowBox[{"{", GridBox[{
          {
           RowBox[{"p", "\[NotEqual]", "0"}]},
          {
           RowBox[{
            RowBox[{"q", "\[NotEqual]", "0"}], "\[Implies]", 
            RowBox[{"\[Or]", 
             RowBox[{"{", GridBox[{
                {
                 RowBox[{
                  RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", 
                  RowBox[{"lp", "[", "q", "]"}]}]},
                {
                 RowBox[{
                  RowBox[{"lp", "[", "p", "]"}], "=", 
                  RowBox[{
                   RowBox[{"lp", "[", "q", "]"}], "\[And]", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", 
                    RowBox[{"lc", "[", "q", "]"}]}], "\[Or]", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"lc", "[", "p", "]"}], "=", 
                    RowBox[{
                    RowBox[{"lc", "[", "q", "]"}], "\[And]", 
                    RowBox[{
                    RowBox[{"R", "[", "p", "]"}], "\[Succeeds]", 
                    RowBox[{"R", "[", "q", "]"}]}]}]}], ")"}]}], ")"}]}]}]}
               }]}]}]}]}
         }]}]}], ")"}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "t", "]"}]}]], 
      RowBox[{"H", "[", 
       RowBox[{"p", ",", "t"}], "]"}]}]}], "=", 
    RowBox[{
     RowBox[{"H", "[", 
      RowBox[{"q", ",", "t"}], "]"}], "\[And]", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"p", ",", "t"}], "]"}], "\[Succeeds]", 
      RowBox[{"C", "[", 
       RowBox[{"q", ",", "t"}], "]"}]}]}]}], TraditionalForm]]],
 "\nAssume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"q", "\[NotEqual]", "0"}], "\[Implies]", 
    RowBox[{"\[Or]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", 
          RowBox[{"lp", "[", "q", "]"}]}]},
        {
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "=", 
          RowBox[{
           RowBox[{"lp", "[", "q", "]"}], "\[And]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", 
              RowBox[{"lc", "[", "q", "]"}]}], "\[Or]", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"lc", "[", "p", "]"}], "=", 
               RowBox[{
                RowBox[{"lc", "[", "q", "]"}], "\[And]", 
                RowBox[{
                 RowBox[{"R", "[", "p", "]"}], "\[Succeeds]", 
                 RowBox[{"R", "[", "q", "]"}]}]}]}], ")"}]}], ")"}]}]}]}
       }]}]}]}], TraditionalForm]]],
 "\nand show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "t", "]"}]}]], 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"p", ",", "t"}], "]"}], "\[Succeeds]", 
       RowBox[{"C", "[", 
        RowBox[{"q", ",", "t"}], "]"}]}]}], "\[And]", 
     RowBox[{"H", "[", 
      RowBox[{"p", ",", "t"}], "]"}]}], "=", 
    RowBox[{"H", "[", 
     RowBox[{"q", ",", "t"}], "]"}]}], TraditionalForm]]],
 "\nWe distinguish two cases, based upon whether ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "=", "0"}], TraditionalForm]]],
 " or not.\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "=", "0"}], TraditionalForm]]],
 ".\nIn this case, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lp", "[", "p", "]"}], TraditionalForm]]],
 " apparently witnesses our existential goal.\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 ".\nFrom our case assumption, together with (2), we can infer\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Or]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", 
         RowBox[{"lp", "[", "q", "]"}]}]},
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "=", 
         RowBox[{
          RowBox[{"lp", "[", "q", "]"}], "\[And]", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", 
             RowBox[{"lc", "[", "q", "]"}]}], "\[Or]", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"lc", "[", "p", "]"}], "=", 
              RowBox[{
               RowBox[{"lc", "[", "q", "]"}], "\[And]", 
               RowBox[{
                RowBox[{"R", "[", "p", "]"}], "\[Succeeds]", 
                RowBox[{"R", "[", "q", "]"}]}]}]}], ")"}]}], ")"}]}]}]}
      }]}]}], TraditionalForm]]],
 "\nbased upon which we again distinguish two cases.\nCase II.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", 
    RowBox[{"lp", "[", "q", "]"}]}], TraditionalForm]]],
 ".\nAgain, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lp", "[", "p", "]"}], TraditionalForm]]],
 " obviously witnesses our existential goal.\nCase II.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "q", "]"}], "\[And]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", 
        RowBox[{"lc", "[", "q", "]"}]}], "\[Or]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], "=", 
         RowBox[{
          RowBox[{"lc", "[", "q", "]"}], "\[And]", 
          RowBox[{
           RowBox[{"R", "[", "p", "]"}], "\[Succeeds]", 
           RowBox[{"R", "[", "q", "]"}]}]}]}], ")"}]}], ")"}]}]}], 
   TraditionalForm]]],
 ".\nWe know\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "=", 
    RowBox[{"lp", "[", "q", "]"}]}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", 
     RowBox[{"lc", "[", "q", "]"}]}], "\[Or]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], "=", 
      RowBox[{
       RowBox[{"lc", "[", "q", "]"}], "\[And]", 
       RowBox[{
        RowBox[{"R", "[", "p", "]"}], "\[Succeeds]", 
        RowBox[{"R", "[", "q", "]"}]}]}]}], ")"}]}], TraditionalForm]]],
 "\nBased on (5) we again distinguish two cases.\nCase II.II.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", 
    RowBox[{"lc", "[", "q", "]"}]}], TraditionalForm]]],
 ".\nAgain, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], 
    RowBox[{"(", 
     RowBox[{"=", 
      RowBox[{"lp", "[", "q", "]"}]}], ")"}]}], TraditionalForm]]],
 " witnesses our existential goal.\nCase II.II.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "p", "]"}], "=", 
    RowBox[{
     RowBox[{"lc", "[", "q", "]"}], "\[And]", 
     RowBox[{
      RowBox[{"R", "[", "p", "]"}], "\[Succeeds]", 
      RowBox[{"R", "[", "q", "]"}]}]}]}], TraditionalForm]]],
 ".\nWe know\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "p", "]"}], "=", 
    RowBox[{"lc", "[", "q", "]"}]}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"R", "[", "p", "]"}], "\[Succeeds]", 
    RowBox[{"R", "[", "q", "]"}]}], TraditionalForm]]],
 "\nFrom (7) and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isOrderedRec", "[", "\[ScriptCapitalP]", "]"}], TraditionalForm]]],
 " we know\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"R", "[", "p", "]"}], "\[NotEqual]", "0"}], TraditionalForm]]],
 "\nso we can use (IH) instantiated by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", ":=", 
    RowBox[{"lp", "[", 
     RowBox[{"R", "[", "p", "]"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", ":=", 
    RowBox[{"R", "[", "p", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", ":=", 
    RowBox[{"R", "[", "q", "]"}]}], TraditionalForm]]],
 ". All premises as well as the LHS of the equivalence are satisfied, hence \
we can conclude\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{
      RowBox[{"R", "[", "p", "]"}], ",", "\[Sigma]"}], "]"}], "=", 
    RowBox[{"H", "[", 
     RowBox[{
      RowBox[{"R", "[", "q", "]"}], ",", "\[Sigma]"}], "]"}]}], 
   TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"R", "[", "p", "]"}], ",", "\[Sigma]"}], "]"}], "\[Succeeds]", 
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"R", "[", "q", "]"}], ",", "\[Sigma]"}], "]"}]}], 
   TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "\[Sigma]", "]"}]}], TraditionalForm]]],
 ". We claim that \[Sigma] witnesses our existential goal, hence we show\n\
(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"p", ",", "\[Sigma]"}], "]"}], "\[Succeeds]", 
      RowBox[{"C", "[", 
       RowBox[{"q", ",", "\[Sigma]"}], "]"}]}], "\[And]", 
     RowBox[{"H", "[", 
      RowBox[{"p", ",", "\[Sigma]"}], "]"}]}], "=", 
    RowBox[{"H", "[", 
     RowBox[{"q", ",", "\[Sigma]"}], "]"}]}], TraditionalForm]]],
 "\nBefore we prove both subgoals, we first use (4) and (6) to infer\n(11)\t\
",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lm", "[", "p", "]"}], "=", 
    RowBox[{"lm", "[", "q", "]"}]}], TraditionalForm]]],
 "\nand we replace ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["q", TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lm", "[", "p", "]"}], "+", 
    RowBox[{"R", "[", "p", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lm", "[", "q", "]"}], "+", 
    RowBox[{"R", "[", "q", "]"}]}], TraditionalForm]]],
 ", respectively, in (G2). Using linearity of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["H", TraditionalForm]]],
 ", this means that we now have to show\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"lm", "[", "p", "]"}], ",", "\[Sigma]"}], "]"}], "+", 
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"R", "[", "p", "]"}], ",", "\[Sigma]"}], "]"}]}], 
    "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"lm", "[", "q", "]"}], ",", "\[Sigma]"}], "]"}], "+", 
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"R", "[", "q", "]"}], ",", "\[Sigma]"}], "]"}]}]}], 
   TraditionalForm]]],
 "\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"H", "[", 
      RowBox[{
       RowBox[{"lm", "[", "p", "]"}], ",", "\[Sigma]"}], "]"}], "+", 
     RowBox[{"H", "[", 
      RowBox[{
       RowBox[{"R", "[", "p", "]"}], ",", "\[Sigma]"}], "]"}]}], "=", 
    RowBox[{
     RowBox[{"H", "[", 
      RowBox[{
       RowBox[{"lm", "[", "q", "]"}], ",", "\[Sigma]"}], "]"}], "+", 
     RowBox[{"H", "[", 
      RowBox[{
       RowBox[{"R", "[", "q", "]"}], ",", "\[Sigma]"}], "]"}]}]}], 
   TraditionalForm]]],
 "\n(G4) follows readily from (9) and (11), and since both ",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "[", 
    RowBox[{
     RowBox[{"lm", "[", "p", "]"}], ",", "\[Sigma]"}], "]"}], 
   TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "[", 
    RowBox[{
     RowBox[{"lm", "[", "q", "]"}], ",", "\[Sigma]"}], "]"}], 
   TraditionalForm]]],
 " are 0 thanks to a lemma, also (G3) is settled due to (10)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6405097194743223`*^9, 3.640510505029664*^9}, {
   3.6405107583500853`*^9, 3.6405110246396503`*^9}, {3.640511074093915*^9, 
   3.640511181177865*^9}, {3.640511425328916*^9, 3.640511526367199*^9}, {
   3.6405116782970943`*^9, 3.6405116871956167`*^9}, {3.640511760417069*^9, 
   3.640511840245083*^9}, {3.640511870921872*^9, 3.6405123436053677`*^9}, {
   3.6405124156166487`*^9, 3.640512416924807*^9}, {3.640512919851452*^9, 
   3.640513011203231*^9}, 3.6434469445776653`*^9, 3.6532081792604513`*^9, 
   3.653213743622641*^9, 3.666174013725355*^9, 3.666440595713036*^9},
 CellID->893664258],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.640512998139373*^9, 3.640513011203367*^9}, 
   3.643446944577799*^9, 3.6532081792604513`*^9, 3.653213743622641*^9, 
   3.666174013725355*^9, 3.666440595713181*^9},
 TextAlignment->Right,
 CellID->1643562830]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (302)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.640513270685519*^9, 3.640513271704646*^9}, 
   3.64051588373612*^9, {3.6406646306385527`*^9, 3.640664633293726*^9}, 
   3.643446951327924*^9, 3.6532081725524397`*^9, 3.653213738147032*^9, 
   3.6661740072981443`*^9, 3.666440590129242*^9},
 CellID->652409243],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.640513795957097*^9, 3.640513797309874*^9}, 
   3.6405158837362843`*^9, 3.643446951328055*^9, 3.6532081725524397`*^9, 
   3.653213738147032*^9, 3.6661740072981443`*^9, 3.666440590129348*^9},
 CellID->1647916099],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{"p", ",", "t"}], "]"}], "=", 
    RowBox[{"0", "\[DoubleLongLeftRightArrow]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"p", "\[NotEqual]", "0"}], "\[Implies]", 
       RowBox[{"\[Not]", 
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", "t"}]}]}], ")"}]}]}], 
   TraditionalForm]]],
 " ... (H 4)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.640513803716757*^9, 3.640513821440076*^9}, {
   3.6405138738268557`*^9, 3.640513880204268*^9}, {3.640514066549692*^9, 
   3.640514095606105*^9}, {3.640514513428318*^9, 3.6405145134373093`*^9}, 
   3.640515883736968*^9, {3.6405159894623117`*^9, 3.6405159909523993`*^9}, {
   3.640517782957327*^9, 3.640517783407349*^9}, 3.643446951328141*^9, 
   3.6532081725524397`*^9, 3.653213738147032*^9, 3.6661740072981443`*^9, 
   3.666440590129434*^9},
 CellID->1566511347],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{"p", ",", "t"}], "]"}], "=", 
    RowBox[{
     RowBox[{
      RowBox[{"H", "[", 
       RowBox[{"q", ",", "t"}], "]"}], "\[And]", 
      RowBox[{"lp", "[", "p", "]"}]}], "=", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"lp", "[", "q", "]"}], "\[And]", 
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", "t"}]}], "\[Implies]", 
       
       RowBox[{"lc", "[", "p", "]"}]}], "=", 
      RowBox[{"lc", "[", "q", "]"}]}]}]}], TraditionalForm]]],
 " ... (H 5\[CloseCurlyQuote])"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.640513803716757*^9, 3.640513821440076*^9}, {
   3.6405138738268557`*^9, 3.640513880204268*^9}, {3.640514066549692*^9, 
   3.640514095606105*^9}, {3.640514513428318*^9, 3.640514541158298*^9}, {
   3.640514957612357*^9, 3.640514957630024*^9}, 3.64051588373707*^9, {
   3.640516184144725*^9, 3.64051619262171*^9}, {3.640516256534745*^9, 
   3.640516258789115*^9}, 3.6405174539136267`*^9, {3.6405177800230703`*^9, 
   3.640517780473165*^9}, 3.6434469513282022`*^9, 3.6532081725524397`*^9, 
   3.653213738147032*^9, 3.6661740072981443`*^9, 3.666440590129491*^9},
 CellID->14964074],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", "0"}], "\[And]", 
     RowBox[{
      RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", "t"}]}], "\[Implies]", 
    RowBox[{
     RowBox[{"R", "[", "p", "]"}], "\[NotEqual]", "0"}]}], 
   TraditionalForm]]],
 " ... essentially (R 2)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.640513803716757*^9, 3.640513821440076*^9}, {
   3.6405138738268557`*^9, 3.640513880204268*^9}, {3.640514066549692*^9, 
   3.640514095606105*^9}, {3.640514513428318*^9, 3.640514541158298*^9}, {
   3.640514957612357*^9, 3.6405149768953247`*^9}, {3.640515253618504*^9, 
   3.640515253636848*^9}, 3.640515883737174*^9, {3.640517765348626*^9, 
   3.640517777353904*^9}, 3.643446951328251*^9, 3.6532081725524397`*^9, 
   3.653213738147032*^9, 3.6661740072981443`*^9, 3.6664405901295443`*^9},
 CellID->628000282],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", "t"}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"H", "[", 
       RowBox[{
        RowBox[{"R", "[", "p", "]"}], ",", "t"}], "]"}], "=", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"H", "[", 
          RowBox[{"p", ",", "t"}], "]"}], "-", 
         RowBox[{"lm", "[", "p", "]"}]}], "\[And]", 
        RowBox[{"C", "[", 
         RowBox[{
          RowBox[{"R", "[", "p", "]"}], ",", "t"}], "]"}]}], "=", 
       RowBox[{"C", "[", 
        RowBox[{"p", ",", "t"}], "]"}]}]}], ")"}]}], TraditionalForm]]],
 " ... (R 5)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.640513803716757*^9, 3.640513821440076*^9}, {
   3.6405138738268557`*^9, 3.640513880204268*^9}, {3.640514066549692*^9, 
   3.640514095606105*^9}, {3.640514513428318*^9, 3.640514541158298*^9}, {
   3.640514957612357*^9, 3.6405149768953247`*^9}, {3.640515253618504*^9, 
   3.640515290876042*^9}, {3.640515333225011*^9, 3.640515343684518*^9}, {
   3.6405154817584667`*^9, 3.640515495985092*^9}, {3.640515572253722*^9, 
   3.6405155743269157`*^9}, 3.640515883737278*^9, {3.640519432935176*^9, 
   3.640519436187745*^9}, 3.6434469513283*^9, 3.6532081725524397`*^9, 
   3.653213738147032*^9, 3.6661740072981443`*^9, 3.6664405901295967`*^9},
 CellID->21049257],

Cell[TextData[{
 "We perform Noetherian induction on ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 ", i.e. we choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isOrderedRec", "[", "\[ScriptCapitalP]", "]"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "=", "t"}], TraditionalForm]]],
 " and\n(IH)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "s", "]"}]}]], 
      RowBox[{"t", "\[Succeeds]", "s"}]], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
         RowBox[{"[", 
          RowBox[{"p", ",", "q"}], "]"}]}]], 
       RowBox[{
        RowBox[{
         RowBox[{"p", "\[NotEqual]", "0"}], "\[And]", 
         RowBox[{"lp", "[", "p", "]"}]}], "=", "s"}]], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
         RowBox[{"[", "t", "]"}]}]], 
       RowBox[{"H", "[", 
        RowBox[{"p", ",", "t"}], "]"}]}]}]}], "=", 
    RowBox[{
     RowBox[{
      RowBox[{"H", "[", 
       RowBox[{"q", ",", "t"}], "]"}], "\[And]", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"p", ",", "t"}], "]"}], "\[Succeeds]", 
       RowBox[{"C", "[", 
        RowBox[{"q", ",", "t"}], "]"}]}]}], "\[Implies]", 
     RowBox[{"p", "\[Succeeds]", "q"}]}]}], TraditionalForm]]],
 "\nand prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "t", "]"}]}]], 
     RowBox[{"H", "[", 
      RowBox[{"p", ",", "t"}], "]"}]}], "=", 
    RowBox[{
     RowBox[{
      RowBox[{"H", "[", 
       RowBox[{"q", ",", "t"}], "]"}], "\[And]", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"p", ",", "t"}], "]"}], "\[Succeeds]", 
       RowBox[{"C", "[", 
        RowBox[{"q", ",", "t"}], "]"}]}]}], "\[Implies]", 
     RowBox[{"(", 
      RowBox[{"\[And]", 
       RowBox[{"{", GridBox[{
          {
           RowBox[{"p", "\[NotEqual]", "0"}]},
          {
           RowBox[{
            RowBox[{"q", "\[NotEqual]", "0"}], "\[Implies]", 
            RowBox[{"\[Or]", 
             RowBox[{"{", GridBox[{
                {
                 RowBox[{
                  RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", 
                  RowBox[{"lp", "[", "q", "]"}]}]},
                {
                 RowBox[{
                  RowBox[{"lp", "[", "p", "]"}], "=", 
                  RowBox[{
                   RowBox[{"lp", "[", "q", "]"}], "\[And]", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", 
                    RowBox[{"lc", "[", "q", "]"}]}], "\[Or]", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"lc", "[", "p", "]"}], "=", 
                    RowBox[{
                    RowBox[{"lc", "[", "q", "]"}], "\[And]", 
                    RowBox[{
                    RowBox[{"R", "[", "p", "]"}], "\[Succeeds]", 
                    RowBox[{"R", "[", "q", "]"}]}]}]}], ")"}]}], ")"}]}]}]}
               }]}]}]}]}
         }]}]}], ")"}]}]}], TraditionalForm]]],
 "\nAssume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{"p", ",", "\[Sigma]"}], "]"}], "=", 
    RowBox[{"H", "[", 
     RowBox[{"q", ",", "\[Sigma]"}], "]"}]}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"p", ",", "\[Sigma]"}], "]"}], "\[Succeeds]", 
    RowBox[{"C", "[", 
     RowBox[{"q", ",", "\[Sigma]"}], "]"}]}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "\[Sigma]", "]"}]}], TraditionalForm]]],
 " and prove\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"q", "\[NotEqual]", "0"}], "\[Implies]", 
    RowBox[{"\[Or]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", 
          RowBox[{"lp", "[", "q", "]"}]}]},
        {
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "=", 
          RowBox[{
           RowBox[{"lp", "[", "q", "]"}], "\[And]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", 
              RowBox[{"lc", "[", "q", "]"}]}], "\[Or]", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"lc", "[", "p", "]"}], "=", 
               RowBox[{
                RowBox[{"lc", "[", "q", "]"}], "\[And]", 
                RowBox[{
                 RowBox[{"R", "[", "p", "]"}], "\[Succeeds]", 
                 RowBox[{"R", "[", "q", "]"}]}]}]}], ")"}]}], ")"}]}]}]}
       }]}]}]}], TraditionalForm]]],
 "\nBefore we proceed we note that due to (2) and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"R4", "[", "\[ScriptCapitalR]", "]"}], TraditionalForm]]],
 " we have\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"p", ",", "\[Sigma]"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\nand thus also\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", "\[Sigma]"}], "\[Or]", 
     "\[Sigma]"}], "=", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 "\nby (C 5).\n\n(G1) follows readily from (3), hence we only concentrate on \
(G2); For this, we assume\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", 
     RowBox[{"lp", "[", "q", "]"}]}]}], TraditionalForm]]],
 "\nand show\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "=", 
    RowBox[{"lp", "[", "q", "]"}]}], TraditionalForm]]],
 "\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", 
     RowBox[{"lc", "[", "q", "]"}]}], "\[Or]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], "=", 
      RowBox[{
       RowBox[{"lc", "[", "q", "]"}], "\[And]", 
       RowBox[{
        RowBox[{"R", "[", "p", "]"}], "\[Succeeds]", 
        RowBox[{"R", "[", "q", "]"}]}]}]}], ")"}]}], TraditionalForm]]],
 "\n\n(G3):\nBecause of (6) we can reduce our goal to\n(G3\[CloseCurlyQuote])\
\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{
     RowBox[{"lp", "[", "q", "]"}], "\[Succeeds]", 
     RowBox[{"lp", "[", "p", "]"}]}]}], TraditionalForm]]],
 "\nWe claim that\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{"p", ",", 
      RowBox[{"lp", "[", "p", "]"}]}], "]"}], "=", 
    RowBox[{"H", "[", 
     RowBox[{"q", ",", 
      RowBox[{"lp", "[", "p", "]"}]}], "]"}]}], TraditionalForm]]],
 "\nwhich we show by distinguishing two cases based on (4):\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", "\[Sigma]"}], 
   TraditionalForm]]],
 ".\nFrom (1) and our case assumption, together with (H 5), we obtain (7).\n\
Case II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Sigma]", "=", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 ".\nWe obtain (7) immediately from (1).\nNow, the LHS of (G3\
\[CloseCurlyQuote]) clearly is 0, meaning that (G3\[CloseCurlyQuote]) follows \
readily from (H 4).\n\n(G4):\nWe assume\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", 
     RowBox[{"lc", "[", "q", "]"}]}]}], TraditionalForm]]],
 "\nand show\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "p", "]"}], "=", 
    RowBox[{"lc", "[", "q", "]"}]}], TraditionalForm]]],
 "\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"R", "[", "p", "]"}], "\[Succeeds]", 
    RowBox[{"R", "[", "q", "]"}]}], TraditionalForm]]],
 "\nBefore we proceed we note that the only possible case in (4) is\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", "\[Sigma]"}], 
   TraditionalForm]]],
 "\nfor otherwise (",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "q", "]"}], "=", "\[Sigma]"}]}], TraditionalForm]]],
 ") we would have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", 
    RowBox[{"lc", "[", "q", "]"}]}], TraditionalForm]]],
 " contradicting (10).\n\n(G5):\nFollows from (1), the already proved (G3), \
and (11) by (H 5\[CloseCurlyQuote]).\n\n(G6):\nFrom (3) and (11), by a lemma, \
we know\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"R", "[", "p", "]"}], "\[NotEqual]", "0"}], TraditionalForm]]],
 "\nand thus we can instantiate our induction hypothesis by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"t", ":=", 
     RowBox[{"lp", "[", 
      RowBox[{"R", "[", "p", "]"}], "]"}]}], ",", 
    RowBox[{"p", ":=", 
     RowBox[{"R", "[", "p", "]"}]}], ",", 
    RowBox[{"q", ":=", 
     RowBox[{"R", "[", "q", "]"}]}]}], TraditionalForm]]],
 ". All premises (except the one of the big implication) are satisfied, hence \
we can reduce our goal to\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "t", "]"}]}]], 
     RowBox[{"H", "[", 
      RowBox[{
       RowBox[{"R", "[", "p", "]"}], ",", "t"}], "]"}]}], "=", 
    RowBox[{
     RowBox[{"H", "[", 
      RowBox[{
       RowBox[{"R", "[", "q", "]"}], ",", "t"}], "]"}], "\[And]", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{
        RowBox[{"R", "[", "p", "]"}], ",", "t"}], "]"}], "\[Succeeds]", 
      RowBox[{"C", "[", 
       RowBox[{
        RowBox[{"R", "[", "q", "]"}], ",", "t"}], "]"}]}]}]}], 
   TraditionalForm]]],
 "\nWe claim that \[Sigma] witnesses this existential goal, so we show both\n\
(G8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{
      RowBox[{"R", "[", "p", "]"}], ",", "\[Sigma]"}], "]"}], "=", 
    RowBox[{"H", "[", 
     RowBox[{
      RowBox[{"R", "[", "q", "]"}], ",", "\[Sigma]"}], "]"}]}], 
   TraditionalForm]]],
 "\n(G9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"R", "[", "p", "]"}], ",", "\[Sigma]"}], "]"}], "\[Succeeds]", 
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"R", "[", "q", "]"}], ",", "\[Sigma]"}], "]"}]}], 
   TraditionalForm]]],
 "\n\n(G8):\nFrom (11) and (G3) we can infer, by a lemma,\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{
      RowBox[{"R", "[", "p", "]"}], ",", "\[Sigma]"}], "]"}], "=", 
    RowBox[{
     RowBox[{"H", "[", 
      RowBox[{"p", ",", "\[Sigma]"}], "]"}], "-", 
     RowBox[{"lm", "[", "p", "]"}]}]}], TraditionalForm]]],
 "\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{
      RowBox[{"R", "[", "q", "]"}], ",", "\[Sigma]"}], "]"}], "=", 
    RowBox[{
     RowBox[{"H", "[", 
      RowBox[{"q", ",", "\[Sigma]"}], "]"}], "-", 
     RowBox[{"lm", "[", "q", "]"}]}]}], TraditionalForm]]],
 "\nso we are done if we can show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lm", "[", "p", "]"}], "=", 
    RowBox[{"lm", "[", "q", "]"}]}], TraditionalForm]]],
 "; This, however, follows immediately from (G3) and (G5).\n(G9):\nFrom (11) \
and (G3) we can infer, by a lemma,\n(15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"R", "[", "p", "]"}], ",", "\[Sigma]"}], "]"}], "=", 
    RowBox[{"C", "[", 
     RowBox[{"p", ",", "\[Sigma]"}], "]"}]}], TraditionalForm]]],
 "\n(16)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"R", "[", "q", "]"}], ",", "\[Sigma]"}], "]"}], "=", 
    RowBox[{"C", "[", 
     RowBox[{"q", ",", "\[Sigma]"}], "]"}]}], TraditionalForm]]],
 "\nhence our goal follows from (2)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6405097194743223`*^9, 3.640510505029664*^9}, {
   3.6405107583500853`*^9, 3.6405110246396503`*^9}, {3.640511074093915*^9, 
   3.640511181177865*^9}, {3.640511425328916*^9, 3.640511526367199*^9}, {
   3.6405116782970943`*^9, 3.6405116871956167`*^9}, {3.640511760417069*^9, 
   3.640511840245083*^9}, {3.640511870921872*^9, 3.6405123436053677`*^9}, {
   3.6405124156166487`*^9, 3.640512416924807*^9}, {3.640512919851452*^9, 
   3.640513011203231*^9}, {3.640513300313751*^9, 3.640513564140724*^9}, {
   3.6405135982035427`*^9, 3.640513674593836*^9}, {3.640513704759441*^9, 
   3.64051375391779*^9}, {3.640513827965097*^9, 3.6405138494511433`*^9}, {
   3.6405138975924587`*^9, 3.640514057800969*^9}, {3.6405141132396383`*^9, 
   3.6405142145198*^9}, 3.640514309243441*^9, {3.640514438260252*^9, 
   3.6405144844543943`*^9}, {3.640514548783659*^9, 3.6405147487568293`*^9}, {
   3.640514781869425*^9, 3.640514929879435*^9}, {3.6405149844279623`*^9, 
   3.640515045061512*^9}, {3.640515076666533*^9, 3.6405152003307343`*^9}, {
   3.6405156353795967`*^9, 3.640515850701447*^9}, 3.640515883737419*^9, {
   3.640517062701888*^9, 3.640517094987458*^9}, {3.640517129724352*^9, 
   3.640517167938665*^9}, {3.640517198548335*^9, 3.640517230765459*^9}, {
   3.640517282907465*^9, 3.640517346251861*^9}, {3.640517386287717*^9, 
   3.640517392944689*^9}, {3.640517447307604*^9, 3.640517448793672*^9}, {
   3.64051749275069*^9, 3.640517495278647*^9}, {3.640665378434414*^9, 
   3.640665401865089*^9}, {3.640665436788763*^9, 3.640665450898666*^9}, 
   3.643446951328404*^9, 3.6532081725524397`*^9, 3.653213738147032*^9, 
   3.6661740072981443`*^9, 3.666440590129678*^9},
 CellID->1479590632],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.640515862561839*^9, 3.640515883737582*^9}, 
   3.643446951328528*^9, 3.6532081725524397`*^9, 3.653213738147032*^9, 
   3.6661740072981443`*^9, 3.666440590129781*^9},
 TextAlignment->Right,
 CellID->702250699]
}, Closed]],

Cell[TextData[{
 "We generalize formula (303) to arbitrary domains \[ScriptCapitalD] \
satisfying the same conditions that have been put on \[ScriptCapitalP]. As \
one can easily verify, the requirement \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"hasMultiplication", "[", "\[ScriptCapitalP]", "]"}], 
   TraditionalForm]]],
 "\[CloseCurlyDoubleQuote], although stated in one of the previous \
subsections, was not used at all to derive formula (303), hence it can also \
be removed from the set of conditions \[ScriptCapitalD] has to meet."
}], "Text",
 CellChangeTimes->{{3.6406682226748457`*^9, 3.640668412686412*^9}},
 CellID->152935259],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FindRelation", "[", 
  RowBox[{"\"\<303\>\"", ",", "\"\<hasMultiplication \[ScriptCapitalP]\>\""}],
   "]"}]], "Input",
 CellChangeTimes->{{3.640670249909347*^9, 3.640670340478792*^9}, 
   3.640677132569397*^9, 3.640677445053472*^9},
 CellID->1209476239],

Cell[BoxData[
 RowBox[{"FindRelation", "[", 
  RowBox[{"\<\"303\"\>", ",", "\<\"hasMultiplication \[ScriptCapitalP]\"\>"}],
   "]"}]], "Output",
 CellChangeTimes->{3.6406702709165573`*^9, 3.640670342397051*^9, 
  3.640677445669842*^9, 3.643447722444148*^9, 3.6434536206973963`*^9, 
  3.643614603328116*^9, 3.643614696626417*^9, 3.643614931604669*^9, 
  3.643614988120779*^9, 3.655465841404954*^9, 3.6623691421349173`*^9, 
  3.6623735760309443`*^9, 3.666176289459159*^9},
 CellID->162283938]
}, Open  ]],

Cell["", "OpenEnvironment",
 CellID->1418355964],

Cell[CellGroupData[{

Cell["\<\
Proposition: Recursive Description of \[Succeeds] in Arbitrary Domains\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.640668029130425*^9, 3.640668035547303*^9}},
 CellID->118382873],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "\[ScriptCapitalD]"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"isDomain", "[", "\[ScriptCapitalD]", "]"}], "\[And]", 
      RowBox[{"hasDecisionPredicate", "[", "\[ScriptCapitalD]", "]"}], 
      "\[And]", 
      RowBox[{"containsMonomials", "[", "\[ScriptCapitalD]", "]"}], "\[And]", 
      
      RowBox[{"hasCoefficientFunction", "[", "\[ScriptCapitalD]", "]"}], 
      "\[And]", 
      RowBox[{"isExtensional", "[", "\[ScriptCapitalD]", "]"}], "\[And]", 
      RowBox[{"isOrderedRec", "[", "\[ScriptCapitalD]", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{"isOrdered", "[", "\[ScriptCapitalD]", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(304)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1015468103", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640668053056802*^9, 3.640668098373002*^9}, {
  3.6406774558063498`*^9, 3.6406774562690163`*^9}, {3.656251605284257*^9, 
  3.6562516125850697`*^9}},
 CellTags->{
  "ID:1015468103", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "304"},
 CellID->1015468103],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1920454598]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->376387788]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Theorem: Reduction Polynomials Constitute Reduction Ring (\[Checkmark])\
\>", "Section",
 CellChangeTimes->{{3.631595890769253*^9, 3.631595910161105*^9}, {
  3.633171998987193*^9, 3.6331720008402433`*^9}, {3.64111625529801*^9, 
  3.6411162595240583`*^9}},
 CellID->1755638798],

Cell["\<\
The next conservation theorem to be proved is the one stating that if \
\[ScriptCapitalR] is not only a ring, but even a reduction ring, and \
\[ScriptCapitalT] is again a commutative power-product domain, then \
\[ScriptCapitalP] is a reduction ring as well.\
\>", "Text",
 CellChangeTimes->{{3.631596507508501*^9, 3.631596561948351*^9}, {
  3.631596642395876*^9, 3.631596750465472*^9}, {3.631596784207755*^9, 
  3.6315967873141623`*^9}, {3.638510380980301*^9, 3.63851044086604*^9}},
 CellID->146700820],

Cell["\<\
Like in the previous section we consider \[ScriptCapitalR], \[ScriptCapitalT] \
and \[ScriptCapitalP] as \[OpenCurlyDoubleQuote]arbitrary but fixed\
\[CloseCurlyDoubleQuote] domains, only that now \[ScriptCapitalR] has to be a \
reduction ring (as already stated in the last subsection of the previous \
section) and that \[ScriptCapitalP], in addition to hasDecisionPredicate, \
containsMonomials etc., also has to satisfy isOrdered and hasMultipliers \
\[LongDash] In short, \[ScriptCapitalP] has to be a ReductionPolynomialDomain.\
\>", "Text",
 CellChangeTimes->{{3.631600900654128*^9, 3.63160100358781*^9}, {
  3.631872522030702*^9, 3.63187264584424*^9}, {3.631873238395948*^9, 
  3.631873238805386*^9}, {3.6318732891337223`*^9, 3.63187337462195*^9}, {
  3.638510129097024*^9, 3.638510131418751*^9}, {3.638510447869817*^9, 
  3.638510508342268*^9}, {3.639713182938335*^9, 3.639713243149479*^9}, {
  3.640579081883944*^9, 3.640579094582621*^9}},
 CellID->1617922970],

Cell["", "OpenEnvironment",
 CellID->2116256688],

Cell[CellGroupData[{

Cell["\<\
Presupposition: Reduction Polynomial Domain \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631600808406777*^9, 3.631600818125695*^9}, {
   3.6318724428033447`*^9, 3.6318724508894777`*^9}, {3.6381567421014223`*^9, 
   3.638156750309504*^9}, 3.638846034955722*^9, {3.639713168307104*^9, 
   3.6397131763518333`*^9}, 3.640579071275302*^9},
 CellID->484984619],

Cell[BoxData[
 RowBox[{"isReductionPolynomialDomain", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(reduction-polynomial domain \[ScriptCapitalP])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:7476052", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.631873263433752*^9},
 CellTags->{
  "ID:7476052", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduction-polynomial domain \[ScriptCapitalP]"},
 CellID->7476052],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->640671230]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->843678998],

Cell[CellGroupData[{

Cell["Preparations (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.63160255031703*^9, 3.6316025527271423`*^9}, {
  3.641115888051448*^9, 3.641115892233675*^9}},
 CellID->826234296],

Cell[CellGroupData[{

Cell["Order Relation (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6318620453092947`*^9, 3.631862048209999*^9}, {
  3.6409501237250557`*^9, 3.640950127782812*^9}, {3.641033449549906*^9, 
  3.641033450034463*^9}, {3.6410373412126017`*^9, 3.641037345355781*^9}},
 CellID->1832383314],

Cell["", "OpenEnvironment",
 CellID->1005504538],

Cell[CellGroupData[{

Cell["Lemma: Properties of \[Succeeds] on \[ScriptCapitalP]", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631862055153998*^9, 3.6318620735624123`*^9}},
 CellID->2142794181],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 173226898}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631862092930688*^9, 3.6318621026308813`*^9}},
 CellTags->{
  "ID:173226898", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->173226898],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "t", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 835223349}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631866911693771*^9, 3.6318669228254766`*^9}},
 CellTags->{
  "ID:835223349", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->835223349],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
  RowBox[{"x", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(356)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:181219027", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640683060060625*^9, 3.640683078279833*^9}},
 CellTags->{
  "ID:181219027", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "356"},
 CellID->181219027],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{"y", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{"x", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "y"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "y", "]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(382)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:271793724", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318672070025*^9, 3.631867258923601*^9}, {
  3.632042430298909*^9, 3.6320424626190023`*^9}, {3.635747578163422*^9, 
  3.63574758346343*^9}},
 CellTags->{
  "ID:271793724", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "382"},
 CellID->271793724],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{"y", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", "]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "y", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "y"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(382')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:575028756", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632732841972261*^9, 3.6327328901067333`*^9}, {
  3.635747587844379*^9, 3.6357475928111773`*^9}},
 CellTags->{
  "ID:575028756", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "382'"},
 CellID->575028756],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"y", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
      RowBox[{"t", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "y", "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{"x", ",", "t"}], "]"}], "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
   "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(383)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2020716880", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318669326387377`*^9, 3.63186703043401*^9}, {
  3.631867079089327*^9, 3.631867086699894*^9}, {3.6357475670200233`*^9, 
  3.635747568859352*^9}},
 CellTags->{
  "ID:2020716880", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "383"},
 CellID->2020716880],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{"y", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", "]"}], "\[Equal]", 
     
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "y", "]"}]}], "\[And]", 
    
    RowBox[{
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "x", "]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "y", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
   "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(381)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1964835443", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318688289009953`*^9, 3.631868846917066*^9}, {
  3.631869111206337*^9, 3.631869116408058*^9}, {3.63574759686077*^9, 
  3.6357476013895283`*^9}},
 CellTags->{
  "ID:1964835443", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "381"},
 CellID->1964835443],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "y"}], "\[And]", 
    RowBox[{"x", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{"y", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", "]"}], "\[Equal]", 
     
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "y", "]"}]}], "\[And]", 
    
    RowBox[{
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "x", "]"}], "\[Equal]", 
     
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "y", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "x", "]"}], 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
   RowBox[{
    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "y", 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(380)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1599847484", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318690962975283`*^9, 3.631869168055531*^9}, {
  3.631869552203847*^9, 3.631869561040101*^9}, {3.631869607181039*^9, 
  3.631869623687047*^9}, {3.631869660402192*^9, 3.631869681688733*^9}, {
  3.635747605294484*^9, 3.635747610737689*^9}},
 CellTags->{
  "ID:1599847484", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "380"},
 CellID->1599847484],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "y"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"x", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
       {
        RowBox[{
         RowBox[{"y", "\[NotEqual]", 
          UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
         RowBox[{"\[Or]", 
          RowBox[{"\[Piecewise]", GridBox[{
             {
              RowBox[{
               RowBox[{
                UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "y", 
                "]"}]}]},
             {
              RowBox[{
               RowBox[{
                RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", "]"}], 
                "\[Equal]", 
                RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "y", "]"}]}],
                "\[And]", 
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                RowBox[{
                 RowBox[{
                  RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "x", "]"}], 
                  UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                  RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "y", 
                   "]"}]}], "\[Or]", 
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "x", 
                    "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "y", 
                    "]"}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "x", "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "y", 
                    "]"}]}]}], 
                  TagBox[")",
                   "AutoParentheses"]}]}], 
                TagBox[")",
                 "AutoParentheses"]}]}]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(379)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:590709100", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318621063223143`*^9, 3.631862230509408*^9}, 
   3.631862849552905*^9, {3.631865741893096*^9, 3.6318657656945267`*^9}, {
   3.6318665472123404`*^9, 3.631866548710248*^9}, {3.631867090068782*^9, 
   3.631867097435369*^9}, {3.631869637044694*^9, 3.631869644728197*^9}, {
   3.6357476237637577`*^9, 3.6357476300570927`*^9}},
 CellTags->{
  "ID:590709100", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "379"},
 CellID->590709100],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{"x", ",", "t"}], "]"}], "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
         RowBox[{"[", "s", "]"}]}]], 
       RowBox[{"s", "\[NotEqual]", "t"}]], 
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{"x", ",", "s"}], "]"}], "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{"x", ",", "t"}], "]"}], 
    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
    "t"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(405)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:586184654", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632717727885339*^9, 3.6327178148709106`*^9}, {
  3.632717920828765*^9, 3.6327179899467373`*^9}},
 CellTags->{
  "ID:586184654", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "405"},
 CellID->586184654],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "x", "]"}], "\[Equal]", 
     "x"}], "\[Or]", 
    RowBox[{"x", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
     RowBox[{
      UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "x", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(406)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:919284948", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632806006669145*^9, 3.63280603948166*^9}, {
  3.6328066398938627`*^9, 3.632806643102831*^9}, {3.6331751615011797`*^9, 
  3.633175166127948*^9}, {3.6331758908798428`*^9, 3.6331759029412003`*^9}, {
  3.635747638324664*^9, 3.63574763976022*^9}},
 CellTags->{
  "ID:919284948", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "406"},
 CellID->919284948],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->141036460]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1634960697],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(356)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(356)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:181219027"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p181219027-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p181219027-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:181219027", "Proof|ID:181219027-1"},
 CellID->719304724],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p181219027-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1538405320]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(382)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(382)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:271793724"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p271793724-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p271793724-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:271793724", "Proof|ID:271793724-1"},
 CellID->1889243650],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p271793724-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->858383599]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(382')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(382')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:575028756"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p575028756-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p575028756-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:575028756", "Proof|ID:575028756-1"},
 CellID->993022467],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p575028756-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{
  3.640682576500461*^9, 3.640683922506447*^9, {3.640931016342647*^9, 
   3.640931017382023*^9}},
 CellID->1954501299]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(383)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(383)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], 
               "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], 
               "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2020716880"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2020716880-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2020716880-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2020716880", "Proof|ID:2020716880-1"},
 CellID->1784894938],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2020716880-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1513088413]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(381)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(381)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1964835443"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1964835443-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1964835443-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1964835443", "Proof|ID:1964835443-1"},
 CellID->880026044],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1964835443-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1363714399]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(380)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(380)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["y", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1599847484"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1599847484-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1599847484-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1599847484", "Proof|ID:1599847484-1"},
 CellID->780294033],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1599847484-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->2075951568]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(379)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(379)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "\[Or]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["y", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "\[Or]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["y", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:590709100"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p590709100-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p590709100-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:590709100", "Proof|ID:590709100-1"},
 CellID->726011705],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p590709100-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->1474313818]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(405)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(405)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], 
               "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
                  StyleBox["t", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], 
               "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
                  StyleBox["t", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:586184654"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p586184654-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p586184654-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:586184654", "Proof|ID:586184654-1"},
 CellID->1800670266],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p586184654-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->1362597520]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(406)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(406)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["x", "ExpressionVariable"]}], "\[Or]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["x", "ExpressionVariable"]}], "\[Or]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               RowBox[{
                 UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:919284948"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p919284948-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p919284948-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:919284948", "Proof|ID:919284948-1"},
 CellID->2038963285],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p919284948-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->427287473]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (405)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632718062409902*^9, 3.632718066336936*^9}, 
   3.632720083211725*^9, 3.6357614354839582`*^9, 3.6358362493870068`*^9, 
   3.643446569363812*^9, 3.6532081569368124`*^9, 3.653213722578204*^9, 
   3.6661739926965184`*^9, 3.666440576442422*^9},
 CellID->1198510616],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"x", ",", "t"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", "\[NotEqual]", "t"}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"x", ",", "s"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "s", "]"}]}], TraditionalForm]]],
 ", and prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"x", ",", "t"}], "]"}], "\[CenterDot]", "t"}]}], 
   TraditionalForm]]],
 ".\n\nIn order to prove (G) we have to show\n(G\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "u", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"H", "[", 
       RowBox[{"x", ",", "u"}], "]"}], "=", 
      RowBox[{
       RowBox[{"H", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"x", ",", "t"}], "]"}], "\[CenterDot]", "t"}], ",", "u"}], 
        "]"}], "\[And]", 
       RowBox[{
        RowBox[{"C", "[", 
         RowBox[{"x", ",", "u"}], "]"}], "\[Succeeds]", 
        RowBox[{"C", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"x", ",", "t"}], "]"}], "\[CenterDot]", "t"}], ",", "u"}],
          "]"}]}]}]}], ")"}]}], TraditionalForm]]],
 "\n\nFrom (1) (and also (3)) we can infer\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n\nDue to ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], TraditionalForm]]],
 " being total, we distinguish three cases based on the relative order of ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lp", "[", "x", "]"}], TraditionalForm]]],
 ".\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "x", "]"}], "\[Succeeds]", "t"}], TraditionalForm]]],
 ".\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"u", ":=", 
    RowBox[{"lp", "[", "x", "]"}]}], TraditionalForm]]],
 " witnesses (G\[CloseCurlyQuote]), so we have to prove now\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{"x", ",", 
      RowBox[{"lp", "[", "x", "]"}]}], "]"}], "=", 
    RowBox[{"H", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"x", ",", "t"}], "]"}], "\[CenterDot]", "t"}], ",", 
      RowBox[{"lp", "[", "x", "]"}]}], "]"}]}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"x", ",", 
      RowBox[{"lp", "[", "x", "]"}]}], "]"}], "\[Succeeds]", 
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"x", ",", "t"}], "]"}], "\[CenterDot]", "t"}], ",", 
      RowBox[{"lp", "[", "x", "]"}]}], "]"}]}], TraditionalForm]]],
 "\n\n(G1):\nWe know that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{"x", ",", 
      RowBox[{"lp", "[", "x", "]"}]}], "]"}], "=", "0"}], TraditionalForm]]],
 ", so what remains to be shown is\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"x", ",", "t"}], "]"}], "\[CenterDot]", "t"}], ",", 
      RowBox[{"lp", "[", "x", "]"}]}], "]"}], "=", "0"}], TraditionalForm]]],
 "\nwhich amounts to showing\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"x", ",", "t"}], "]"}], "\[CenterDot]", "t"}], ",", "v"}], 
     "]"}], "=", "0"}], TraditionalForm]]],
 "\nfor some a.b.f. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "v", "]"}]}], TraditionalForm]]],
 " with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"v", "\[Succeeds]", 
    RowBox[{"lp", "[", "x", "]"}]}], TraditionalForm]]],
 ". Due to transitivity of ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], TraditionalForm]]],
 " we also know\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"v", "\[Succeeds]", "t"}], TraditionalForm]]],
 "\nExpanding the definition of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " for monomials leads to proving\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "=", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"C", "[", 
         RowBox[{"x", ",", "t"}], "]"}], "\[DoubleLongLeftArrow]", 
        RowBox[{"s", "=", "t"}]},
       {"0", "\[DoubleLongLeftArrow]", "otherwise"}
      }]}]}], TraditionalForm]]],
 "\nBecause of (5) and ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], TraditionalForm]]],
 " being irreflexive, we are done.\n\n(G2):\nWith a similar argument as the \
one we used for showing (G4) we can immediately infer that the RHS of the \
inequality is 0, and the LHS is apparently just ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lc", "[", "x", "]"}], TraditionalForm]]],
 ", so what we really have to show now is\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "x", "]"}], "\[Succeeds]", "0"}], TraditionalForm]]],
 "\nBecause of (4) we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "x", "]"}], "\[NotEqual]", "0"}], TraditionalForm]]],
 ", and (R4) in \[ScriptCapitalR] finishes the case.\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Succeeds]", 
    RowBox[{"lp", "[", "x", "]"}]}], TraditionalForm]]],
 ".\nBy the definition of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lp", "[", "x", "]"}], TraditionalForm]]],
 " we know\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "u", "]"}]}]], 
      RowBox[{"u", "\[Succeeds]", 
       RowBox[{"lp", "[", "x", "]"}]}]], 
     RowBox[{"C", "[", 
      RowBox[{"x", ",", "u"}], "]"}]}], "=", "0"}], TraditionalForm]]],
 "\nand this clearly contradicts (1).\n\nCase III: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{"lp", "[", "x", "]"}]}], TraditionalForm]]],
 ".\nFirst of all, using a similar argument as in case II, we can infer\n(7)\t\
",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Succeeds]", "s"}], TraditionalForm]]],
 "\nand thus\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"R", "[", "x", "]"}], "\[NotEqual]", "0"}], TraditionalForm]]],
 "\nMoreover, we can replace every occurrence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"x", ",", "t"}], "]"}], "\[CenterDot]", "t"}], TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lm", "[", "x", "]"}], TraditionalForm]]],
 ".\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"u", ":=", 
    RowBox[{"lp", "[", 
     RowBox[{"R", "[", "x", "]"}], "]"}]}], TraditionalForm]]],
 " witnesses our existential goal, hence we prove\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{"x", ",", 
      RowBox[{"lp", "[", 
       RowBox[{"R", "[", "x", "]"}], "]"}]}], "]"}], "=", 
    RowBox[{"H", "[", 
     RowBox[{
      RowBox[{"lm", "[", "x", "]"}], ",", 
      RowBox[{"lp", "[", 
       RowBox[{"R", "[", "x", "]"}], "]"}]}], "]"}]}], TraditionalForm]]],
 "\n(G8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"x", ",", 
      RowBox[{"lp", "[", 
       RowBox[{"R", "[", "x", "]"}], "]"}]}], "]"}], "\[Succeeds]", 
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"lm", "[", "x", "]"}], ",", 
      RowBox[{"lp", "[", 
       RowBox[{"R", "[", "x", "]"}], "]"}]}], "]"}]}], TraditionalForm]]],
 "\n\n(G7):\nBy two lemmas we know that both sides are equal to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lm", "[", "x", "]"}], TraditionalForm]]],
 ".\n\n(G8):\nBy two other lemmas we know that the LHS is non-zero, whereas \
the RHS is 0, so again we are done using (R4) in \[ScriptCapitalR]."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632718073912755*^9, 3.63271848294977*^9}, {
   3.632718552051909*^9, 3.632718813900831*^9}, {3.632718906951293*^9, 
   3.632719112948382*^9}, {3.632719155103076*^9, 3.632719172017109*^9}, {
   3.632719203018696*^9, 3.632719315830573*^9}, {3.632719386193124*^9, 
   3.632719564699954*^9}, {3.632719887449945*^9, 3.632719914751565*^9}, {
   3.632720014931122*^9, 3.632720083211854*^9}, 3.635761435484086*^9, 
   3.6358362493871717`*^9, {3.640944331710734*^9, 3.640944367708695*^9}, 
   3.6434465693639517`*^9, 3.6532081569368124`*^9, 3.653213722578204*^9, 
   3.6661739926965184`*^9, 3.666440576442573*^9},
 CellID->710706997],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632720068736424*^9, 3.6327200832119904`*^9}, 
   3.635761435484219*^9, 3.635836249387334*^9, 3.643446569364079*^9, 
   3.6532081569368124`*^9, 3.653213722578204*^9, 3.6661739926965184`*^9, 
   3.666440576442718*^9},
 TextAlignment->Right,
 CellID->1793634431]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1494017576],

Cell[CellGroupData[{

Cell["\<\
Lemma: \[Succeeds] for Monomials and Polynomials\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.633175806303253*^9, 3.6331758287455683`*^9}},
 CellID->141141923],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", "x", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 895736483}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6331758450761414`*^9, 3.633175852319551*^9}},
 CellTags->{
  "ID:895736483", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->895736483],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", "c", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1904443546}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633175856027269*^9, 3.6331758637960587`*^9}},
 CellTags->{
  "ID:1904443546", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1904443546],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "t", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1587558771}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639219268044752*^9, 3.639219279322249*^9}},
 CellTags->{
  "ID:1587558771", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1587558771],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "x", "]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], "c"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
   RowBox[{"c", 
    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", 
     "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(377)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:530327018", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633175172506502*^9, 3.633175233280711*^9}, {
  3.633175922419414*^9, 3.633175967101289*^9}, {3.635747654246046*^9, 
  3.63574765627048*^9}},
 CellTags->{
  "ID:530327018", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377"},
 CellID->530327018],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"c", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "x"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"t", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", "]"}]}], 
      "\[And]", 
      RowBox[{"c", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Or]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"t", "\[Equal]", 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", "]"}]}], 
      "\[And]", 
      RowBox[{"c", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
       RowBox[{
        UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "x", "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(377')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2044424156", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633175874688076*^9, 3.633175901650641*^9}, {
  3.633175997079163*^9, 3.633176095403247*^9}, {3.635747660102837*^9, 
  3.635747661646882*^9}, {3.63919715591575*^9, 3.6391971600883102`*^9}, {
  3.6392192825452213`*^9, 3.639219284818335*^9}},
 CellTags->{
  "ID:2044424156", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377'"},
 CellID->2044424156],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", "]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
   RowBox[{"c", 
    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
    "t"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(377'')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:14600367", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639219127117014*^9, 3.639219155783985*^9}, {
  3.639219288884665*^9, 3.6392192909686213`*^9}, {3.639220433255006*^9, 
  3.6392204347070303`*^9}},
 CellTags->{
  "ID:14600367", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377''"},
 CellID->14600367],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "d", "]"}]}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"c", 
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
     RowBox[{"d", 
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"c", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], "d"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 7)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1313257457", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6328077764109383`*^9, 3.632807816915571*^9}, {
  3.6410331818581247`*^9, 3.641033189859317*^9}},
 CellTags->{
  "ID:1313257457", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 7"},
 CellID->1313257457],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
    RowBox[{
     RowBox[{"c", 
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
     UnderscriptBox["+", "\[ScriptCapitalP]"], "x"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
          RowBox[{"x", ",", "t"}], "]"}], "\[Equal]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
       {
        RowBox[{"c", 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
         RowBox[{"c", 
          UnderscriptBox["+", "\[ScriptCapitalR]"], 
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"x", ",", "t"}], "]"}]}]}]},
       {
        RowBox[{"x", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] 8)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1776891622", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6328123493348227`*^9, 3.632812428116878*^9}, {
  3.632812511901026*^9, 3.632812515643078*^9}, {3.6392971074856453`*^9, 
  3.639297152545588*^9}, {3.639297416528275*^9, 3.639297417836852*^9}, {
  3.6410332007480793`*^9, 3.6410332223128777`*^9}, {3.6410334318605843`*^9, 
  3.641033441348186*^9}},
 CellTags->{
  "ID:1776891622", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 8"},
 CellID->1776891622],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->259176606]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1149241651],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(377)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(377)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:530327018"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p530327018-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p530327018-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:530327018", "Proof|ID:530327018-1"},
 CellID->24650902],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p530327018-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->660190578]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(377')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(377')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], "\[Or]", 
             RowBox[{
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["x", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], "\[Or]", 
             RowBox[{
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["x", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2044424156"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2044424156-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2044424156-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2044424156", "Proof|ID:2044424156-1"},
 CellID->35999503],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2044424156-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->259827218]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(377'')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(377'')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:14600367"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p14600367-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p14600367-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:14600367", "Proof|ID:14600367-1"},
 CellID->439633690],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p14600367-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->775230185]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 7)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 7)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
             StyleBox[
             "d", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
             StyleBox[
             "d", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1313257457"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1313257457-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1313257457-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1313257457", "Proof|ID:1313257457-1"},
 CellID->1865577507],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1313257457-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->246532405]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] 8)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] 8)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}]}]}], 
             "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalP]"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], 
               UnderscriptBox["+", "\[ScriptCapitalP]"], 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["H", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}]}]}], 
             "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalP]"]}]}]}]}], $CellContext`TheoremaForm]],
      "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1776891622"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1776891622-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1776891622-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1776891622", "Proof|ID:1776891622-1"},
 CellID->89604998],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1776891622-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->141709217]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Multipliers (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.631871979120081*^9, 3.6318719808469353`*^9}, {
  3.641112373932947*^9, 3.641112378294631*^9}},
 CellID->2009513133],

Cell["", "OpenEnvironment",
 CellID->803353906],

Cell[CellGroupData[{

Cell["\<\
Lemma: Properties of the Multipliers in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631871990860545*^9, 3.6318720004112883`*^9}},
 CellID->1741879042],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "m"}], "]"}]}]], 
  RowBox[{"p", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1077742446}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318722060296297`*^9, 3.63187223158878*^9}, {
   3.6321161673114843`*^9, 3.632116207140366*^9}, {3.632809403537784*^9, 
   3.632809404012443*^9}, {3.632809886582375*^9, 3.6328098951837273`*^9}, 
   3.6357477308255653`*^9, {3.6380750477780848`*^9, 3.638075047994636*^9}},
 CellTags->{
  "ID:1077742446", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1077742446],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
   RowBox[{"m", ",", "p"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", "]"}], 
    "\[And]", 
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}], ",", 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}]}], "]"}]}], 
   
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(384)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1717822499", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631872237672883*^9, 3.631872289081749*^9}, {
   3.6321161560593567`*^9, 3.632116175994813*^9}, {3.632809384080002*^9, 
   3.632809393381048*^9}, 3.6328098980558157`*^9, {3.638075050651144*^9, 
   3.638075053859251*^9}},
 CellTags->{
  "ID:1717822499", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "384"},
 CellID->1717822499],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
   RowBox[{"m", ",", "p"}], "]"}], "\[Implies]", 
  RowBox[{"m", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(358)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1045770541", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638075085068425*^9, 3.638075104614942*^9}},
 CellTags->{
  "ID:1045770541", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "358"},
 CellID->1045770541],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
   RowBox[{"m", ",", "p"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{"m", 
          UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}], "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
          RowBox[{"m", 
           UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}], "]"}], "\[Equal]", 
         RowBox[{
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
          UnderscriptBox["*", "\[ScriptCapitalT]"], 
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}]},
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
          RowBox[{"m", 
           UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}], "]"}], "\[Equal]", 
         RowBox[{
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}], 
          UnderscriptBox["*", "\[ScriptCapitalR]"], 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(393)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:312784637", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633172992721548*^9, 3.633173095714665*^9}, {
  3.635747736088232*^9, 3.635747742828705*^9}, {3.638075061969121*^9, 
  3.63807507886512*^9}},
 CellTags->{
  "ID:312784637", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "393"},
 CellID->312784637],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 172690263}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638701054040451*^9, 3.6387010734911413`*^9}},
 CellTags->{
  "ID:172690263", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->172690263],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
   RowBox[{"c", ",", 
    RowBox[{
     UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}]}], "]"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ",", "p"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult monomial)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:174940712", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632116179118339*^9, 3.632116265498838*^9}, {
   3.632717203818129*^9, 3.632717208128006*^9}, {3.632809407284498*^9, 
   3.632809419614848*^9}, 3.632809900007012*^9, {3.638701076898634*^9, 
   3.638701079455467*^9}},
 CellTags->{
  "ID:174940712", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial"},
 CellID->174940712],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
     RowBox[{"c", ",", 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", "i"}],
      "]"}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"c", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ",", "p", 
      ",", "i"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult monomial II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1850961673", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63870108634272*^9, 3.638701135622558*^9}, 
   3.638702193752099*^9},
 CellTags->{
  "ID:1850961673", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial II"},
 CellID->1850961673],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
       RowBox[{
        RowBox[{"c", 
         UnderscriptBox["*", "\[ScriptCapitalR]"], 
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}]}], ",", 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
        "i"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"c", 
         UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
       UnderscriptBox["*", "\[ScriptCapitalP]"], "m"}], ",", "p", ",", "i"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult monomial III)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:849856225", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639300085044364*^9, 3.639300132524536*^9}, {
  3.639300182711529*^9, 3.639300191400419*^9}},
 CellTags->{
  "ID:849856225", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial III"},
 CellID->849856225],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1178443032]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1823748294],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(384)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(384)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1717822499"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1717822499-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1717822499-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1717822499", "Proof|ID:1717822499-1"},
 CellID->1434876425],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1717822499-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1232590601]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(358)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(358)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalP]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1045770541"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1045770541-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1045770541-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1045770541", "Proof|ID:1045770541-1"},
 CellID->1393398660],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1045770541-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->483586923]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(393)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(393)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["p", "ExpressionVariable"]}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["p", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["p", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["p", "ExpressionVariable"]}], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["p", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["p", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:312784637"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p312784637-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p312784637-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:312784637", "Proof|ID:312784637-1"},
 CellID->2016825295],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p312784637-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->537234257]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult monomial)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult monomial)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], ",", 
               StyleBox["p", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], ",", 
               StyleBox["p", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:174940712"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p174940712-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p174940712-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:174940712", "Proof|ID:174940712-1"},
 CellID->1306123142],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p174940712-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->275742108]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult monomial II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult monomial II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["i", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], ",", 
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["i", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], ",", 
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1850961673"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1850961673-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1850961673-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1850961673", "Proof|ID:1850961673-1"},
 CellID->1891103710],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1850961673-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->753129535]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult monomial III)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult monomial III)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["i", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["m", "ExpressionVariable"]}], ",", 
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["i", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["m", "ExpressionVariable"]}], ",", 
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:849856225"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p849856225-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p849856225-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:849856225", "Proof|ID:849856225-1"},
 CellID->232642418],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p849856225-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1171947620]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Tuples (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.631955860853393*^9, 3.631955861828823*^9}, {
  3.641115878891824*^9, 3.641115883025927*^9}},
 CellID->1854005352],

Cell["", "OpenEnvironment",
 CellID->941507149],

Cell[CellGroupData[{

Cell["Lemma: Tuples of Monomials", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631956147467024*^9, 3.6319561605592337`*^9}, 
   3.639906151916819*^9},
 CellID->1448249107],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
    RowBox[{"[", "T", "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 633254127}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631955871113366*^9, 3.6319558852439003`*^9}, {
  3.631955929823927*^9, 3.6319559379780397`*^9}, {3.635749131225998*^9, 
  3.635749138284988*^9}},
 CellTags->{
  "ID:633254127", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->633254127],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"S", "=", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      SubscriptBox["T", "i"], 
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1581566919}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631955905209798*^9, 3.631955924778473*^9}, {
  3.631956035404984*^9, 3.631956037237482*^9}, {3.632717218636551*^9, 
  3.632717225151228*^9}},
 CellTags->{
  "ID:1581566919", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1581566919],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
  RowBox[{"[", "S", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(tuples \[CenterDot] 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1230351665", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631955889852973*^9, 3.631955897881001*^9}, {
  3.631955942134884*^9, 3.631955947915832*^9}, {3.635749144983652*^9, 
  3.6357491523892508`*^9}},
 CellTags->{
  "ID:1230351665", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 1"},
 CellID->1230351665],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"a", "=", "0"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "T", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1294700753}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320329851991796`*^9, 3.632033007011586*^9}, 
   3.6363680935896683`*^9},
 CellTags->{
  "ID:1294700753", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1294700753],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox[
    SubscriptBox["\[Sum]", "\[ScriptCapitalP]"], 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "a"}]], 
   SubscriptBox["S", "i"]}], "\[Equal]", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox[
      SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "a"}]], 
     SubscriptBox["T", "i"]}], ")"}], 
   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
   "t"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(tuples \[CenterDot] 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:168346017", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6319560092616*^9, 3.631956073118209*^9}, {
  3.632032967793167*^9, 3.632032971950121*^9}, {3.632717231251034*^9, 
  3.632717243437455*^9}},
 CellTags->{
  "ID:168346017", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 2"},
 CellID->168346017],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "p", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox[
      SubscriptBox["\[Sum]", "\[ScriptCapitalP]"], 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "a"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       SubscriptBox["S", "i"], 
       UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}], 
      TagBox[")",
       "AutoParentheses"]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         UnderscriptBox[
          SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "a"}]], 
         SubscriptBox["T", "i"]}], ")"}], 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
     UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(tuples \[CenterDot] 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1208795229", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632032804521117*^9, 3.632032866415876*^9}, {
   3.632032973754751*^9, 3.632032975153903*^9}, 3.632115827244763*^9, {
   3.6327172510110598`*^9, 3.632717268367839*^9}, 3.64111293912728*^9},
 CellTags->{
  "ID:1208795229", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 3"},
 CellID->1208795229],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "p", "]"}]}]], 
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
      RowBox[{
       SubscriptBox["T", "i"], ",", 
       RowBox[{
        UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}]}], "]"}], 
     "\[Implies]", 
     RowBox[{
      UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
      RowBox[{
       SubscriptBox["S", "i"], ",", "p"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(tuples \[CenterDot] 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:369110158", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631956190094956*^9, 3.631956270954569*^9}, {
  3.632116089033362*^9, 3.632116112319193*^9}, {3.6363657717078543`*^9, 
  3.6363657793967047`*^9}},
 CellTags->{
  "ID:369110158", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 4"},
 CellID->369110158],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->429653964]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->533685121],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(tuples \[CenterDot] 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(tuples \[CenterDot] 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], "[", 
           RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], 
               UnderscriptBox["|", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
             "\[RightAngleBracket]"}], "]"}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], "[", 
           RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], 
               UnderscriptBox["|", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
             "\[RightAngleBracket]"}], "]"}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1230351665"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1230351665-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1230351665-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1230351665", "Proof|ID:1230351665-1"},
 CellID->803576507],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1230351665-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->453965713]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(tuples \[CenterDot] 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(tuples \[CenterDot] 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", "\[ScriptCapitalP]"], 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["a", "ExpressionVariable"]}]}}]], 
             SubscriptBox[
              RowBox[{"\[LeftAngleBracket]", 
                RowBox[{
                  RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                  UnderscriptBox["|", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                "\[RightAngleBracket]"}], 
              StyleBox["i", "ExpressionVariable"]]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox[
                SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["a", "ExpressionVariable"]}]}}]], 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]]}], 
             UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", "\[ScriptCapitalP]"], 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["a", "ExpressionVariable"]}]}}]], 
             SubscriptBox[
              RowBox[{"\[LeftAngleBracket]", 
                RowBox[{
                  RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                  UnderscriptBox["|", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                "\[RightAngleBracket]"}], 
              StyleBox["i", "ExpressionVariable"]]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox[
                SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["a", "ExpressionVariable"]}]}}]], 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]]}], 
             UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:168346017"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p168346017-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p168346017-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:168346017", "Proof|ID:168346017-1"},
 CellID->1197262074],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p168346017-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1861690467]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(tuples \[CenterDot] 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(tuples \[CenterDot] 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", "\[ScriptCapitalP]"], 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["a", "ExpressionVariable"]}]}}]], 
             RowBox[{
               SubscriptBox[
                RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                  "\[RightAngleBracket]"}], 
                StyleBox["i", "ExpressionVariable"]], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["p", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["a", "ExpressionVariable"]}]}}]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             StyleBox[
             "p", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", "\[ScriptCapitalP]"], 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["a", "ExpressionVariable"]}]}}]], 
             RowBox[{
               SubscriptBox[
                RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                  "\[RightAngleBracket]"}], 
                StyleBox["i", "ExpressionVariable"]], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["p", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["a", "ExpressionVariable"]}]}}]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["*", "\[ScriptCapitalP]"], 
             StyleBox[
             "p", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1208795229"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1208795229-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1208795229-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1208795229", "Proof|ID:1208795229-1"},
 CellID->1531683337],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1208795229-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1417027127]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(tuples \[CenterDot] 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(tuples \[CenterDot] 4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               SubscriptBox[
                RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                  "\[RightAngleBracket]"}], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               StyleBox["p", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               SubscriptBox[
                RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                  "\[RightAngleBracket]"}], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               StyleBox["p", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:369110158"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p369110158-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p369110158-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:369110158", "Proof|ID:369110158-1"},
 CellID->1626361315],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p369110158-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1922614985]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["isPartIrreflOrder (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.631602567960004*^9, 3.6316025727751904`*^9}, {
  3.636197318235628*^9, 3.63619732335259*^9}},
 CellID->410707321],

Cell["", "OpenEnvironment",
 CellID->2043816998],

Cell[CellGroupData[{

Cell["\<\
Lemma: Auxiliary Results Needed for Proving \[Succeeds] on \[ScriptCapitalP] \
Partial Irreflexive\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6361932824013042`*^9, 3.636193334453947*^9}, {
  3.636195428911153*^9, 3.636195432893236*^9}},
 CellID->1791092897],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 155708286}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.636193356729113*^9, 3.636193369183095*^9}},
 CellTags->{
  "ID:155708286", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->155708286],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{"x", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
   "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:280312466", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.636193497185733*^9, 3.6361935115820932`*^9}},
 CellTags->{
  "ID:280312466", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "4"},
 CellID->280312466],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "y"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"y", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "x"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:713919982", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.636193397751422*^9, 3.6361934546187963`*^9}},
 CellTags->{
  "ID:713919982", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "5"},
 CellID->713919982],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "z", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"x", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "y"}], "\[And]", 
      RowBox[{"y", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "z"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{"x", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "z"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1630947645", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.636193524512528*^9, 3.636193594003035*^9}},
 CellTags->{
  "ID:1630947645", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "6"},
 CellID->1630947645],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1332205312]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1152004342],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["x", "ExpressionVariable"]}], 
             TagBox[")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["x", "ExpressionVariable"]}], 
             TagBox[
             ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:280312466"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p280312466-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p280312466-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:280312466", "Proof|ID:280312466-1"},
 CellID->1561997],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p280312466-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->560792166]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                 StyleBox["x", "ExpressionVariable"]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 StyleBox["y", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                 StyleBox["x", "ExpressionVariable"]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:713919982"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p713919982-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p713919982-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:713919982", "Proof|ID:713919982-1"},
 CellID->393026],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p713919982-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->573639697]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(6)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(6)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1630947645"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1630947645-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1630947645-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1630947645", "Proof|ID:1630947645-1"},
 CellID->350427357],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1630947645-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->234994923]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1582119546],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[Succeeds] on \[ScriptCapitalP] is Partial Irreflexive Order \
Relation\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6316109459398537`*^9, 3.631610979273831*^9}, {
  3.631615025732175*^9, 3.6316150269266567`*^9}},
 CellID->1640450222],

Cell[BoxData[
 RowBox[{"isPartIrreflOrder", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalP]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(360)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:795795393", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185513388586*^9, 3.610185532862774*^9}, {
  3.6315968457883472`*^9, 3.6315968500959044`*^9}, {3.636089519790123*^9, 
  3.636089520181837*^9}},
 CellTags->{
  "ID:795795393", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "360"},
 CellID->795795393],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->233694654]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->939147830],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(360)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(360)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isPartIrreflOrder", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isPartIrreflOrder", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:795795393"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p795795393-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p795795393-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:795795393", "Proof|ID:795795393-1"},
 CellID->1058256179],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p795795393-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->883251815]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["isNoetherian (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.631602580466778*^9, 3.631602582287814*^9}, {
  3.6361867394586477`*^9, 3.636186745196588*^9}},
 CellID->2044895609],

Cell["", "OpenEnvironment",
 CellID->1937965720],

Cell[CellGroupData[{

Cell["\<\
Lemma: Auxiliary Results needed for Proving Noetherianity of \[Succeeds] in \
\[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.635843558735819*^9, 3.6358435770166817`*^9}},
 CellID->2071197478],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], 
   RowBox[{"[", "A", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1302712453}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.635835698841951*^9, 3.635835710824332*^9}},
 CellTags->{
  "ID:1302712453", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1302712453],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{
     UnderscriptBox["0", "\[ScriptCapitalP]"], "\[Element]", "A"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"a", "\[Element]", "A"}]], 
     RowBox[{"a", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{"isMin", "[", 
             RowBox[{
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "a", "]"}], 
              ",", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", "]"}], 
                UnderscriptBox["|", 
                 RowBox[{"x", "\[Element]", "A"}]]}], "}"}], ",", 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               TagBox[
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"],
                Identity], 
               TagBox[")",
                "AutoParentheses"]}]}], "]"}]},
           {
            RowBox[{"isMin", "[", 
             RowBox[{
              RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "a", "]"}], 
              ",", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "x", "]"}], 
                UnderscriptBox["|", 
                 RowBox[{"x", "\[Element]", "A"}]], 
                RowBox[{
                 RowBox[{
                  UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", "]"}], 
                 "\[Equal]", 
                 RowBox[{
                  UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "a", 
                  "]"}]}]}], "}"}], ",", 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               TagBox[
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"],
                Identity], 
               TagBox[")",
                "AutoParentheses"]}]}], "]"}]},
           {
            RowBox[{"isMin", "[", 
             RowBox[{
              RowBox[{
               UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "a", "]"}], ",", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{
                 UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "x", "]"}], 
                UnderscriptBox["|", 
                 RowBox[{"x", "\[Element]", "A"}]], 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{
                  RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "x", 
                    "]"}], "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "a", 
                    "]"}]}], "\[And]", 
                  RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "x", 
                    "]"}], "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "a", 
                    "]"}]}]}], 
                 TagBox[")",
                  "AutoParentheses"]}]}], "}"}], ",", 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               TagBox[
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"],
                Identity], 
               TagBox[")",
                "AutoParentheses"]}]}], "]"}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{"isMin", "[", 
       RowBox[{"a", ",", "A", ",", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         TagBox[
          UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"],
          Identity], 
         TagBox[")",
          "AutoParentheses"]}]}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1092343128", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.635835723045945*^9, 3.6358358099147987`*^9}, {
  3.635835853646442*^9, 3.635836051883877*^9}, {3.6358361126860027`*^9, 
  3.635836135936977*^9}},
 CellTags->{
  "ID:1092343128", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1"},
 CellID->1092343128],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["0", "\[ScriptCapitalP]"], "\[Element]", "A"}], 
  "\[Implies]", 
  RowBox[{"isMin", "[", 
   RowBox[{
    UnderscriptBox["0", "\[ScriptCapitalP]"], ",", "A", ",", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:311837531", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.635843597660955*^9, 3.635843620568914*^9}},
 CellTags->{
  "ID:311837531", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "2"},
 CellID->311837531],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       UnderscriptBox["0", "\[ScriptCapitalP]"], "\[Element]", "A"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "t", "]"}]}]], 
     RowBox[{"t", "\[Equal]", 
      UnderscriptBox["1", "\[ScriptCapitalT]"]}]}], "\[And]", 
    RowBox[{"nonemptySet", "[", "A", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{"a", "\[Element]", "A"}]], 
    RowBox[{"isMin", "[", 
     RowBox[{"a", ",", "A", ",", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox[
        UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"],
        Identity], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2036982330", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.635847325916039*^9, 3.63584738244388*^9}, {
  3.6358475801739883`*^9, 3.6358476246684923`*^9}},
 CellTags->{
  "ID:2036982330", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "3"},
 CellID->2036982330],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], 
      RowBox[{"[", "A", "]"}]}]], 
    RowBox[{"nonemptySet", "[", "A", "]"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[ForAll]", 
         RowBox[{"p", "\[Element]", "A"}]], 
        RowBox[{"p", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
       RowBox[{"t", 
        UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{"p", "\[Element]", "A"}]], 
       RowBox[{"isMin", "[", 
        RowBox[{"p", ",", "A", ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox[
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"],
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(378)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:917853466", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631857654613449*^9, 3.6318577869152308`*^9}, {
   3.6318604528599243`*^9, 3.631860454089099*^9}, {3.631860669764312*^9, 
   3.631860680491593*^9}, {3.631860739469151*^9, 3.631860751241802*^9}, {
   3.635762912658062*^9, 3.635762924362525*^9}, 3.653364836670161*^9, {
   3.6533650163200765`*^9, 3.6533650196584826`*^9}},
 CellTags->{
  "ID:917853466", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "378"},
 CellID->917853466],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2118519333]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->634503852],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["0", "\[ScriptCapitalP]"], "\[Element]", 
                 StyleBox["A", "ExpressionVariable"]}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
              RowBox[{
                StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"isMin", "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]]}], "}"}], ",", 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"]}], 
                   "]"}], "\[And]", 
                 RowBox[{"isMin", "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}]}], "}"}], 
                    ",", 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"]}], 
                   "]"}], "\[And]", 
                 RowBox[{"isMin", "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}]}]}], "}"}], 
                    ",", 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"]}], 
                   "]"}]}], "\[Implies]", 
               RowBox[{"isMin", "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["0", "\[ScriptCapitalP]"], "\[Element]", 
                 StyleBox["A", "ExpressionVariable"]}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
              RowBox[{
                StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"isMin", "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]]}], "}"}], ",", 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"]}], 
                   "]"}], "\[And]", 
                 RowBox[{"isMin", "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}]}], "}"}], 
                    ",", 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"]}], 
                   "]"}], "\[And]", 
                 RowBox[{"isMin", "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}]}]}], "}"}], 
                    ",", 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"]}], 
                   "]"}]}], "\[Implies]", 
               RowBox[{"isMin", "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1092343128"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1092343128-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1092343128-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1092343128", "Proof|ID:1092343128-1"},
 CellID->402687292],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1092343128-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->138879590]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["0", "\[ScriptCapitalP]"], "\[Element]", 
             StyleBox["A", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{"isMin", "[", 
             RowBox[{
               UnderscriptBox["0", "\[ScriptCapitalP]"], ",", 
               StyleBox["A", "ExpressionVariable"], ",", 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["0", "\[ScriptCapitalP]"], "\[Element]", 
             StyleBox["A", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{"isMin", "[", 
             RowBox[{
               UnderscriptBox["0", "\[ScriptCapitalP]"], ",", 
               StyleBox["A", "ExpressionVariable"], ",", 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:311837531"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p311837531-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p311837531-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:311837531", "Proof|ID:311837531-1"},
 CellID->222069951],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p311837531-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->393972337]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox["0", "\[ScriptCapitalP]"], "\[Element]", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}]}], "\[And]", 
             RowBox[{"nonemptySet", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Element]", 
                   StyleBox["A", "ExpressionVariable"]}]}}]], 
             RowBox[{"isMin", "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["A", "ExpressionVariable"], ",", 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox["0", "\[ScriptCapitalP]"], "\[Element]", 
                   StyleBox["A", "ExpressionVariable"]}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}]}], "\[And]", 
             RowBox[{"nonemptySet", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Element]", 
                   StyleBox["A", "ExpressionVariable"]}]}}]], 
             RowBox[{"isMin", "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["A", "ExpressionVariable"], ",", 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2036982330"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2036982330-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2036982330-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2036982330", "Proof|ID:2036982330-1"},
 CellID->78668347],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2036982330-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->401808988]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(378)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(378)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"nonemptySet", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], "\[Element]", 
                   StyleBox["A", "ExpressionVariable"]}]}}]], 
             RowBox[{"isMin", "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["A", "ExpressionVariable"], ",", 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"nonemptySet", "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], "\[Element]", 
                   StyleBox["A", "ExpressionVariable"]}]}}]], 
             RowBox[{"isMin", "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["A", "ExpressionVariable"], ",", 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:917853466"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p917853466-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p917853466-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:917853466", "Proof|ID:917853466-1"},
 CellID->463010928],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p917853466-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->429418029]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->885823776],

Cell[CellGroupData[{

Cell["Theorem: \[Succeeds] on \[ScriptCapitalP] is Noetherian", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.63161501834496*^9, 3.631615033773074*^9}},
 CellID->1196982619],

Cell[BoxData[
 RowBox[{"isNoetherian", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalP]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(361)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1009043994", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185539137405*^9, 3.61018554217264*^9}, {
  3.631596851836029*^9, 3.631596855626609*^9}},
 CellTags->{
  "ID:1009043994", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "361"},
 CellID->1009043994],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1321732972]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->212143235],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(361)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(361)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isNoetherian", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isNoetherian", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], ",", 
           "\[ScriptCapitalP]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1009043994"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1009043994-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1009043994-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1009043994", "Proof|ID:1009043994-1"},
 CellID->307122811],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1009043994-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->698773133]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Multipliers (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6316025867360153`*^9, 3.631602588456407*^9}, {
  3.639727290277217*^9, 3.639727295118216*^9}},
 CellID->1626247964],

Cell["", "OpenEnvironment",
 CellID->681587107],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[ScriptCapitalP] satisfies the Multipliers Axiom\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631869906346676*^9, 3.631869920515428*^9}},
 CellID->190398885],

Cell[BoxData[
 RowBox[{"Multipliers", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(362)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:412087985", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185546905514*^9, 3.610185550972085*^9}, 
   3.631596858663241*^9},
 CellTags->{
  "ID:412087985", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "362"},
 CellID->412087985],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1616084504]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->675296661],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(362)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(362)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "Multipliers", "[", "\[ScriptCapitalP]", 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "Multipliers", "[", "\[ScriptCapitalP]", 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:412087985"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p412087985-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p412087985-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:412087985", "Proof|ID:412087985-1"},
 CellID->453147487],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p412087985-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->500939487]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R0 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.631602592819961*^9, 3.6316025933785887`*^9}, {
  3.639726575134425*^9, 3.639726579993765*^9}},
 CellID->1977275530],

Cell["", "OpenEnvironment",
 CellID->787808098],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R0", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631870359663556*^9, 3.6318703668816843`*^9}},
 CellID->2032276339],

Cell[BoxData[
 RowBox[{"R0", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(363)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1267059190", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185592892032*^9, 3.61018560199445*^9}, 
   3.6315968599775677`*^9},
 CellTags->{
  "ID:1267059190", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "363"},
 CellID->1267059190],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1962684668]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->857332813],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(363)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(363)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R0", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R0", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1267059190"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1267059190-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1267059190-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1267059190", "Proof|ID:1267059190-1"},
 CellID->1552638821],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1267059190-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1865569314]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R1 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.631871531063984*^9, 3.631871531399918*^9}, {
  3.639633720371355*^9, 3.639633725473935*^9}},
 CellID->1132174785],

Cell["", "OpenEnvironment",
 CellID->978075595],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R1", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631871607187195*^9, 3.631871613591913*^9}},
 CellID->275565290],

Cell[BoxData[
 RowBox[{"R1", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(364)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:378652997", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185605073124*^9, 3.61018562063804*^9}, 
   3.6315968606521387`*^9},
 CellTags->{
  "ID:378652997", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "364"},
 CellID->378652997],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->876981541]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1058953121],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(364)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(364)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R1", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R1", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:378652997"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p378652997-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p378652997-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:378652997", "Proof|ID:378652997-1"},
 CellID->1328553736],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p378652997-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1498432134]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R2 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.63187304135398*^9, 3.631873041662671*^9}, {
  3.639632562737382*^9, 3.639632567012094*^9}},
 CellID->1093569315],

Cell[TextData[{
 StyleBox["Remark",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 ": Formula (365) is already the content of formula (393) in section \
\[OpenCurlyDoubleQuote]Preparations\[CloseCurlyDoubleQuote]/\
\[CloseCurlyDoubleQuote]Multipliers\[CloseCurlyDoubleQuote]."
}], "Text",
 CellChangeTimes->{{3.638073691099203*^9, 3.638073704604371*^9}, {
  3.6396324183709793`*^9, 3.639632440899674*^9}},
 CellID->1339945595],

Cell["", "OpenEnvironment",
 CellID->815100016],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R2", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631873050545569*^9, 3.631873056119842*^9}},
 CellID->2082733210],

Cell[BoxData[
 RowBox[{"R2", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(365)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1036416147", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185606039065*^9, 3.6101856214173*^9}, 
   3.631596861353503*^9},
 CellTags->{
  "ID:1036416147", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "365"},
 CellID->1036416147],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->220603823]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->29368222],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(365)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(365)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R2", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R2", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1036416147"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1036416147-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1036416147-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1036416147", "Proof|ID:1036416147-1"},
 CellID->246228744],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1036416147-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1948904361]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R3 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.63187423466936*^9, 3.631874234975791*^9}, {
  3.636433347168755*^9, 3.636433352538896*^9}},
 CellID->1028022157],

Cell["", "OpenEnvironment",
 CellID->1224247765],

Cell[CellGroupData[{

Cell["\<\
Lemma: Auxiliary Results for proving R3 in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631879000968946*^9, 3.631879014810306*^9}},
 CellID->1843437225],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "m", "]"}]}]], 
   RowBox[{
    UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", "]"}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "p", "]"}]}]], 
    RowBox[{"p", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
      RowBox[{"[", "M", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "M", "\[RightBracketingBar]"}]}]], 
           RowBox[{
            UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
            RowBox[{
             SubscriptBox["M", "i"], ",", "p"}], "]"}]}]},
         {
          RowBox[{
           RowBox[{
            UnderscriptBox[
             SubscriptBox["\[Sum]", "\[ScriptCapitalP]"], 
             RowBox[{
              RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
              RowBox[{
              "\[LeftBracketingBar]", "M", "\[RightBracketingBar]"}]}]], 
            SubscriptBox["M", "i"]}], "\[Equal]", "m"}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(386)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1341307829", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6319543400568037`*^9, 3.631954352897024*^9}, {
   3.631954542000988*^9, 3.631954667170394*^9}, 3.6319547165731688`*^9, {
   3.6319548554998093`*^9, 3.631954860027614*^9}, {3.631958233737789*^9, 
   3.631958236115847*^9}, {3.635749181296597*^9, 3.635749187754922*^9}, {
   3.636366186163879*^9, 3.6363661895588017`*^9}, 3.636369892781064*^9, {
   3.636370101314954*^9, 3.636370140556136*^9}},
 CellTags->{
  "ID:1341307829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "386"},
 CellID->1341307829],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "q", "]"}]}]], 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", "p", "]"}]}]], 
     RowBox[{
      RowBox[{"p", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
       "\[Equal]", "t"}]}]], 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
       RowBox[{"[", "M", "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
              RowBox[{
              "\[LeftBracketingBar]", "M", "\[RightBracketingBar]"}]}]], 
            RowBox[{
             UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
             RowBox[{
              SubscriptBox["M", "i"], ",", "q"}], "]"}]}]},
          {
           RowBox[{"p", "\[Equal]", 
            RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", "\[ScriptCapitalP]"], 
              RowBox[{
               RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
               RowBox[{
               "\[LeftBracketingBar]", "M", "\[RightBracketingBar]"}]}]], 
             SubscriptBox["M", "i"]}]}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(385)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1193326225", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6318790243501043`*^9, 3.631879203145421*^9}, {
   3.63187993356201*^9, 3.6318799364793663`*^9}, 3.631954715164851*^9, {
   3.631956818746624*^9, 3.631956836576239*^9}, 3.6357477650340223`*^9, {
   3.635749167618224*^9, 3.635749174489284*^9}, 3.6363623723060503`*^9, {
   3.6363693124738503`*^9, 3.636369331174931*^9}, {3.6363698998891277`*^9, 
   3.6363699000338993`*^9}, 3.6363702087344646`*^9, {3.642494724158403*^9, 
   3.6424947562547398`*^9}, {3.6424950740298767`*^9, 3.642495100185059*^9}},
 CellTags->{
  "ID:1193326225", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "385"},
 CellID->1193326225],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1058015582]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2043406167],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(386)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(386)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
                   "[", 
                   StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["M", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["M", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    StyleBox["p", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               
               RowBox[{
                 RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[ScriptCapitalP]"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["M", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   SubscriptBox[
                    StyleBox["M", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "\[Equal]", 
                 StyleBox[
                 "m", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
                   "[", 
                   StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["M", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["M", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    StyleBox["p", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               
               RowBox[{
                 RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[ScriptCapitalP]"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["M", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   SubscriptBox[
                    StyleBox["M", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "\[Equal]", 
                 StyleBox[
                 "m", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1341307829"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1341307829-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1341307829-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1341307829", "Proof|ID:1341307829-1"},
 CellID->467247998],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1341307829-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->550586326]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(385)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(385)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
                   "[", 
                   StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["M", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["M", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    StyleBox["q", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[ScriptCapitalP]"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["M", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   SubscriptBox[
                    StyleBox["M", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
                   "[", 
                   StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["M", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["M", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    StyleBox["q", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[ScriptCapitalP]"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["M", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   SubscriptBox[
                    StyleBox["M", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1193326225"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1193326225-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1193326225-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1193326225", "Proof|ID:1193326225-1"},
 CellID->348360809],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1193326225-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->43480281]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->472886914],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R3", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631878727913496*^9, 3.63187873342657*^9}},
 CellID->1909661142],

Cell[BoxData[
 RowBox[{"R3", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(366)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2033377127", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185606857769*^9, 3.61018562219682*^9}, 
   3.6315968621660547`*^9},
 CellTags->{
  "ID:2033377127", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "366"},
 CellID->2033377127],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->684694705]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1928200850],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(366)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(366)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R3", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R3", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2033377127"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2033377127-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2033377127-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2033377127", "Proof|ID:2033377127-1"},
 CellID->1477321448],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2033377127-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.642497146496849*^9, 3.6424971544559917`*^9}},
 CellID->151991483]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R4 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6318742608542023`*^9, 3.631874261080772*^9}, {
  3.639632169887477*^9, 3.639632174357744*^9}},
 CellID->294752926],

Cell["", "OpenEnvironment",
 CellID->67511843],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R4", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631874269437109*^9, 3.631874276167924*^9}},
 CellID->1145845750],

Cell[BoxData[
 RowBox[{"R4", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(367)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:982813600", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185607453736*^9, 3.610185624533509*^9}, 
   3.631596862946412*^9},
 CellTags->{
  "ID:982813600", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "367"},
 CellID->982813600],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->482550205]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1572717722],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(367)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(367)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R4", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R4", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:982813600"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p982813600-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p982813600-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:982813600", "Proof|ID:982813600-1"},
 CellID->1236106547],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p982813600-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->475654809]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1467132509],

Cell[CellGroupData[{

Cell["Corollary", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.6332350309127703`*^9},
 CellID->482418387],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", ",", "q", ",", "a"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1179937112}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639121857551919*^9, 3.6391218660156803`*^9}},
 CellTags->{
  "ID:1179937112", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1179937112],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
    "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"p", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
      RowBox[{"q", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
      RowBox[{"a", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(336)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:134072435", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633235036535697*^9, 3.633235096681848*^9}, {
   3.635747831909231*^9, 3.635747842687944*^9}, 3.639121868299891*^9},
 CellTags->{
  "ID:134072435", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "336"},
 CellID->134072435],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "a"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(337)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1484092370", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6391218796930723`*^9, 3.639121886134358*^9}},
 CellTags->{
  "ID:1484092370", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "337"},
 CellID->1484092370],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->870268060]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->225373014],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(336)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(336)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalP]"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalP]"]}]}]}]}], $CellContext`TheoremaForm]],
      "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:134072435"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p134072435-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p134072435-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:134072435", "Proof|ID:134072435-1"},
 CellID->77495342],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p134072435-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1008758752]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(337)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(337)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalP]"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalP]"]}]}]}]}], $CellContext`TheoremaForm]],
      "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1484092370"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1484092370-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1484092370-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1484092370", "Proof|ID:1484092370-1"},
 CellID->1348971853],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1484092370-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->660891557]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Alternative Descriptions of Reducibility in \[ScriptCapitalP] (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.633061887252846*^9, 3.633061902317897*^9}, {
  3.638097751195035*^9, 3.6380977559850407`*^9}},
 CellID->1409824068],

Cell["", "OpenEnvironment",
 CellID->888657849],

Cell[CellGroupData[{

Cell["\<\
Lemma: Alternative Description of reduceM in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.63205077444979*^9, 3.632050784376658*^9}},
 CellID->1708398684],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
    RowBox[{"[", "T", "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1480600758}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632050788360165*^9, 3.632050821471484*^9}, {
  3.632050909116363*^9, 3.632050917321505*^9}, {3.635749206319686*^9, 
  3.635749213160811*^9}},
 CellTags->{
  "ID:1480600758", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1480600758],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "t", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1004003660}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632050920960476*^9, 3.632050928450302*^9}},
 CellTags->{
  "ID:1004003660", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1004003660],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"S", "=", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      SubscriptBox["T", "i"], 
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1854976476}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632052252894367*^9, 3.632052260264142*^9}, {
  3.632717352416442*^9, 3.632717358689248*^9}},
 CellTags->{
  "ID:1854976476", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1854976476],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"k", "=", "0"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "q", ",", "S", ",", "k"}], "]"}], "\[Equal]", 
    RowBox[{"p", 
     UnderscriptBox["-", "\[ScriptCapitalP]"], 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          UnderscriptBox[
           SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
           RowBox[{
            RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
          SubscriptBox["T", "i"]}], ")"}], 
        UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduceM \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:408134090", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632051024572352*^9, 3.632051114592931*^9}, {
   3.632051847519545*^9, 3.6320518479638577`*^9}, 3.632051960186438*^9, {
   3.6320522573440657`*^9, 3.6320522578258533`*^9}, {3.6327173229899693`*^9, 
   3.632717342935155*^9}},
 CellTags->{
  "ID:408134090", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduceM \[ScriptCapitalP]"},
 CellID->408134090],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "m", "]"}]}]], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", "]"}], 
    "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], "\[Equal]", 
     "t"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"k", "=", "0"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
      RowBox[{
       RowBox[{"p", 
        UnderscriptBox["-", "\[ScriptCapitalP]"], 
        RowBox[{"m", 
         UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}]}], ",", "q", ",", 
       "S", ",", "k"}], "]"}], "\[Equal]", 
     RowBox[{"p", 
      UnderscriptBox["-", "\[ScriptCapitalP]"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}], 
           UnderscriptBox["+", "\[ScriptCapitalR]"], 
           RowBox[{
            UnderscriptBox[
             SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
             RowBox[{
              RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
            SubscriptBox["T", "i"]}]}], ")"}], 
         UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
       UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduceM \[ScriptCapitalP] II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1088018783", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320518204512157`*^9, 3.6320520025320597`*^9}, {
  3.632052226826111*^9, 3.632052276179879*^9}, {3.63271737123553*^9, 
  3.63271738611199*^9}, {3.637998454701323*^9, 3.637998455924549*^9}},
 CellTags->{
  "ID:1088018783", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduceM \[ScriptCapitalP] II"},
 CellID->1088018783],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1032173464]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2022036672],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduceM \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduceM \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "q", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "q", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:408134090"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p408134090-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p408134090-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:408134090", "Proof|ID:408134090-1"},
 CellID->59633542],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p408134090-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->199214208]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduceM \[ScriptCapitalP] II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduceM \[ScriptCapitalP] II)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["-", "\[ScriptCapitalP]"], 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox["q", "ExpressionVariable"]}]}], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["-", "\[ScriptCapitalP]"], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox[
                 "q", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], "\[Equal]", 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["-", "\[ScriptCapitalP]"], 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox["q", "ExpressionVariable"]}]}], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["-", "\[ScriptCapitalP]"], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["+", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox[
                 "q", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1088018783"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1088018783-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1088018783-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1088018783", "Proof|ID:1088018783-1"},
 CellID->306221528],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1088018783-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1372373678]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1591405130],

Cell[CellGroupData[{

Cell["\<\
Lemma: Alternative Description of \[Succeeds] for Differences in \
\[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.633062202880373*^9, 3.633062216188048*^9}, {
  3.6330623052650623`*^9, 3.633062310534607*^9}},
 CellID->1153660425],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "c", ",", "m"}], "]"}]}]], 
  RowBox[{"c", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1964871967}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.633062342794162*^9, {3.6330623839880943`*^9, 3.633062406217128*^9}, {
   3.635747854973344*^9, 3.635747856585807*^9}},
 CellTags->{
  "ID:1964871967", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1964871967],

Cell[TextData[{
 "Formula (407) looks like a simple result about ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], TraditionalForm]]],
 ", but it cannot be stated in subsection \[OpenCurlyDoubleQuote]Preparations\
\[CloseCurlyDoubleQuote], since we need \[ScriptCapitalP] being a commutative \
ring with unit."
}], "EnvironmentText",
 CellChangeTimes->{{3.638002122799168*^9, 3.638002195618239*^9}, {
  3.6380022515453987`*^9, 3.6380022624271507`*^9}},
 CellID->197410698],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
    RowBox[{"a", 
     UnderscriptBox["-", "\[ScriptCapitalP]"], "c"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{"a", ",", 
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", "]"}]}], "]"}], 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{"a", ",", 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", "]"}]}], "]"}], 
     UnderscriptBox["-", "\[ScriptCapitalR]"], 
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(407)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1041722017", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633062350850217*^9, 3.633062357061808*^9}, {
  3.633062410262252*^9, 3.633062412872179*^9}, {3.638000706290268*^9, 
  3.638000713772017*^9}, {3.6380022343948603`*^9, 3.6380022420459003`*^9}, {
  3.638010712556179*^9, 3.63801071506984*^9}},
 CellTags->{
  "ID:1041722017", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "407"},
 CellID->1041722017],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
   RowBox[{"m", ",", "c"}], "]"}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1518222527}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633066961468672*^9, 3.633066969678782*^9}},
 CellTags->{
  "ID:1518222527", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1518222527],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
    RowBox[{"a", 
     UnderscriptBox["-", "\[ScriptCapitalP]"], 
     RowBox[{"m", 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "c"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{"a", ",", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
       UnderscriptBox["*", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}], "]"}], 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
        UnderscriptBox["*", "\[ScriptCapitalT]"], 
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}], 
      "]"}], 
     UnderscriptBox["-", "\[ScriptCapitalR]"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}], 
      UnderscriptBox["*", "\[ScriptCapitalR]"], 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(408)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1070088161", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633062096263756*^9, 3.633062257556419*^9}, 
   3.633062340780011*^9, {3.633062372593677*^9, 3.633062386856485*^9}, {
   3.633066967618258*^9, 3.633066978197006*^9}, {3.638074196437731*^9, 
   3.638074201949459*^9}},
 CellTags->{
  "ID:1070088161", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "408"},
 CellID->1070088161],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["-", "\[ScriptCapitalP]"], 
     RowBox[{"m", 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "c"}]}], 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "a"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
        UnderscriptBox["*", "\[ScriptCapitalT]"], 
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}], 
      "]"}], 
     UnderscriptBox["-", "\[ScriptCapitalR]"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}], 
      UnderscriptBox["*", "\[ScriptCapitalR]"], 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}], 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{"a", ",", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
       UnderscriptBox["*", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(408')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1970014385", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63306698297724*^9, 3.6330670324195223`*^9}, {
  3.633067493299131*^9, 3.633067503961948*^9}, {3.636433699558298*^9, 
  3.636433705219935*^9}, {3.638074212780267*^9, 3.638074218148129*^9}},
 CellTags->{
  "ID:1970014385", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "408'"},
 CellID->1970014385],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->34978639]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->778747754],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(407)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(407)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 StyleBox["c", "ExpressionVariable"]}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}], "]"}], 
                 UnderscriptBox["-", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 StyleBox["c", "ExpressionVariable"]}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}], "]"}], 
                 UnderscriptBox["-", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1041722017"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1041722017-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1041722017-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1041722017", "Proof|ID:1041722017-1"},
 CellID->178605949],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1041722017-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->13157926]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(408)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(408)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["c", "ExpressionVariable"]}]}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                 UnderscriptBox["-", "\[ScriptCapitalR]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["c", "ExpressionVariable"]}]}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                 UnderscriptBox["-", "\[ScriptCapitalR]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1070088161"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1070088161-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1070088161-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1070088161", "Proof|ID:1070088161-1"},
 CellID->1155517064],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1070088161-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->745976973]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(408')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(408')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["c", "ExpressionVariable"]}]}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["a", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                 UnderscriptBox["-", "\[ScriptCapitalR]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["c", "ExpressionVariable"]}]}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["a", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                 UnderscriptBox["-", "\[ScriptCapitalR]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1970014385"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1970014385-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1970014385-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1970014385", "Proof|ID:1970014385-1"},
 CellID->625973727],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1970014385-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->285408563]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1042798326],

Cell[CellGroupData[{

Cell["\<\
Theorem: Alternative Description of Reducibility in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632026151690164*^9, 3.63202616933403*^9}, {
  3.642999970967561*^9, 3.642999972030808*^9}},
 CellID->2008480134],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", ",", "q", ",", "f", ",", "m"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 102721391}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632482653202176*^9, 3.632482666016674*^9}},
 CellTags->{
  "ID:102721391", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->102721391],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"p", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"m", ",", "f"}]], "\[ScriptCapitalP]"], "q"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"f", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
       {
        RowBox[{
         UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
         RowBox[{"m", ",", "f"}], "]"}]},
       {
        RowBox[{"q", "\[Equal]", 
         RowBox[{"p", 
          UnderscriptBox["-", "\[ScriptCapitalP]"], 
          RowBox[{"m", 
           UnderscriptBox["*", "\[ScriptCapitalP]"], "f"}]}]}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
         RowBox[{
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"p", ",", 
            RowBox[{
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
             UnderscriptBox["*", "\[ScriptCapitalT]"], 
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "f", "]"}]}]}], 
           "]"}], ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "f", "]"}], ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}]}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RED)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:657593834", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320262101786623`*^9, 3.632026444643375*^9}, {
  3.6320266436974277`*^9, 3.632026650249051*^9}, {3.63248266915436*^9, 
  3.632482671956521*^9}, {3.635747869386963*^9, 3.635747871001202*^9}},
 CellTags->{
  "ID:657593834", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED"},
 CellID->657593834],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "f", ",", "m"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"f", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
       {
        RowBox[{
         UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
         RowBox[{"m", ",", "f"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
         RowBox[{
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"p", ",", 
            RowBox[{
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
             UnderscriptBox["*", "\[ScriptCapitalT]"], 
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "f", "]"}]}]}], 
           "]"}], ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "f", "]"}], ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}]}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RED II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1802485517", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6324826762582903`*^9, 3.632482727606166*^9}, {
   3.632560361638876*^9, 3.632560369668003*^9}, 3.6325604406765833`*^9, {
   3.635747879060382*^9, 3.635747880674636*^9}, {3.6380892300432253`*^9, 
   3.638089244671471*^9}},
 CellTags->{
  "ID:1802485517", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED II"},
 CellID->1802485517],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "f", ",", "m"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"f", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
       {
        RowBox[{
         UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", "]"}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
         RowBox[{
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"p", ",", 
            RowBox[{
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
             UnderscriptBox["*", "\[ScriptCapitalT]"], 
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "f", "]"}]}]}], 
           "]"}], ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "f", "]"}], ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}]}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RED III)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1164884865", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632560497215334*^9, 3.632560528673139*^9}, {
  3.6357478879983664`*^9, 3.635747896494761*^9}, {3.638089257672967*^9, 
  3.638089262149459*^9}, {3.638095837636509*^9, 3.638095845106257*^9}},
 CellTags->{
  "ID:1164884865", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED III"},
 CellID->1164884865],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "f", ",", "m"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"f", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
       {
        RowBox[{
         UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
         RowBox[{"m", ",", "f"}], "]"}]},
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
          RowBox[{"p", ",", 
           RowBox[{
            RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
            UnderscriptBox["*", "\[ScriptCapitalT]"], 
            RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "f", "]"}]}]}], 
          "]"}], 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
         RowBox[{
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"p", ",", 
            RowBox[{
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
             UnderscriptBox["*", "\[ScriptCapitalT]"], 
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "f", "]"}]}]}], 
           "]"}], 
          UnderscriptBox["-", "\[ScriptCapitalR]"], 
          RowBox[{
           RowBox[{
            UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}], 
           UnderscriptBox["*", "\[ScriptCapitalR]"], 
           RowBox[{
            UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "f", "]"}]}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RED IV)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:958945878", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638245353844325*^9, 3.638245414091654*^9}},
 CellTags->{
  "ID:958945878", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED IV"},
 CellID->958945878],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}]], 
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1601029605}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643010915182757*^9, 3.64301094095296*^9}},
 CellTags->{
  "ID:1601029605", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1601029605],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ",", "f", 
    ",", "m"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"f", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
       {
        RowBox[{
         UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", "]"}]},
       {
        RowBox[{
         RowBox[{
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
          UnderscriptBox["*", "\[ScriptCapitalT]"], 
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "f", "]"}]}], 
         "\[Equal]", "t"}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
         RowBox[{"c", ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "f", "]"}], ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}]}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RED monomial)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:687002895", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633076374783208*^9, 3.633076524287004*^9}, {
  3.6357479014308033`*^9, 3.635747906210614*^9}, {3.6380968456020412`*^9, 
  3.638096852548291*^9}, {3.643010943927853*^9, 3.64301094550638*^9}},
 CellTags->{
  "ID:687002895", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED monomial"},
 CellID->687002895],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "f", ",", 
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}]}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"f", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
         RowBox[{
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"p", ",", 
            RowBox[{"t", 
             UnderscriptBox["*", "\[ScriptCapitalT]"], 
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "f", "]"}]}]}], 
           "]"}], ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "f", "]"}], ",", 
          "c"}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RED monomial II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:413658239", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643010963003744*^9, 3.643011021016324*^9}},
 CellTags->{
  "ID:413658239", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED monomial II"},
 CellID->413658239],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2087609660]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1936134947],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(RED)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(RED)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["f", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["f", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["-", "\[ScriptCapitalP]"], 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox["f", "ExpressionVariable"]}]}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["f", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["f", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["-", "\[ScriptCapitalP]"], 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox["f", "ExpressionVariable"]}]}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:657593834"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p657593834-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p657593834-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:657593834", "Proof|ID:657593834-1"},
 CellID->17884315],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p657593834-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->303231202]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(RED II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(RED II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["f", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["f", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1802485517"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1802485517-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1802485517-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1802485517", "Proof|ID:1802485517-1"},
 CellID->17136843],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1802485517-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->330024022]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(RED III)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(RED III)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1164884865"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1164884865-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1164884865-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1164884865", "Proof|ID:1164884865-1"},
 CellID->1415542183],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1164884865-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->205712485]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(RED IV)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(RED IV)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["f", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                   UnderscriptBox["-", "\[ScriptCapitalR]"], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["f", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                   UnderscriptBox["-", "\[ScriptCapitalR]"], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:958945878"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p958945878-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p958945878-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:958945878", "Proof|ID:958945878-1"},
 CellID->1172023358],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p958945878-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1468364448]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(RED monomial)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(RED monomial)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], ",", 
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}], "\[Equal]", 
                 
                 StyleBox["t", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], ",", 
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}], "\[Equal]", 
                 
                 StyleBox["t", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:687002895"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p687002895-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p687002895-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:687002895", "Proof|ID:687002895-1"},
 CellID->563304177],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p687002895-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->889845776]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(RED monomial II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(RED monomial II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:413658239"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p413658239-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p413658239-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:413658239", "Proof|ID:413658239-1"},
 CellID->885739272],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p413658239-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->881833888]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->571707933],

Cell[CellGroupData[{

Cell["Corollary", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.638170772986711*^9},
 CellID->480000138],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", ",", "q", ",", "f"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 580194792}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646711801761571*^9, 3.646711810576703*^9}},
 CellTags->{
  "ID:580194792", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->580194792],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"p", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"m", ",", "f"}]], "\[ScriptCapitalP]"], "q"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"q", ",", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
         UnderscriptBox["*", "\[ScriptCapitalT]"], 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "f", "]"}]}]}], 
       "]"}], "\[Equal]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{"p", ",", 
         RowBox[{
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
          UnderscriptBox["*", "\[ScriptCapitalT]"], 
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "f", "]"}]}]}], 
        "]"}], 
       UnderscriptBox["-", "\[ScriptCapitalR]"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}], 
        UnderscriptBox["*", "\[ScriptCapitalR]"], 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "f", "]"}]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(357)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1667457569", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638170776609453*^9, 3.638170866787825*^9}, 
   3.643000227748012*^9, {3.643000465882433*^9, 3.643000467750083*^9}, {
   3.6467118134389544`*^9, 3.6467118136068077`*^9}},
 CellTags->{
  "ID:1667457569", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "357"},
 CellID->1667457569],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"f", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"p", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"{", 
         RowBox[{
          UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "f", "]"}], "}"}]], 
       "\[ScriptCapitalP]"], "q"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "r", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"p", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"{", "f", "}"}]], "\[ScriptCapitalP]"], "r"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow] lm \[DoubleRightArrow])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2035027156", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646717875902699*^9, 3.646717921340506*^9}, {
  3.6618485609535723`*^9, 3.661848588690421*^9}},
 CellTags->{
  "ID:2035027156", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] lm \[DoubleRightArrow]"},
 CellID->2035027156],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"p", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"{", "f", "}"}]], "\[ScriptCapitalP]"], "q"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"f", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", "r", "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"p", 
       UnderscriptBox[
        SubscriptBox["\[Rule]", 
         RowBox[{"{", 
          RowBox[{
           UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "f", "]"}], "}"}]],
         "\[ScriptCapitalP]"], "r"}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow] lm \[DoubleLeftArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2026457432", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646711820843251*^9, 3.646711868536263*^9}, {
  3.646711971586731*^9, 3.646712002317647*^9}, {3.6618485977384367`*^9, 
  3.661848618283673*^9}},
 CellTags->{
  "ID:2026457432", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] lm \[DoubleLeftArrow]"},
 CellID->2026457432],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "f"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"f", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "t", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
             RowBox[{
              RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{"p", ",", "t"}], "]"}], ",", 
              RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "f", "]"}]}], 
             "]"}]},
           {
            RowBox[{
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "f", "]"}], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reducibility \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1167741853", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647075197063992*^9, 3.647075280300312*^9}, {
  3.6470753153190928`*^9, 3.647075336575705*^9}, {3.6470756769228067`*^9, 
  3.647075733639165*^9}},
 CellTags->{
  "ID:1167741853", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reducibility \[ScriptCapitalP]"},
 CellID->1167741853],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"p", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"{", "f", "}"}]], "\[ScriptCapitalP]"], "q"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"f", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
       {
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{
           RowBox[{
            UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
            RowBox[{"[", "m", "]"}]}], ",", 
           RowBox[{
            UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
            RowBox[{"[", "t", "]"}]}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"\[And]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{"q", "\[Equal]", 
                RowBox[{"p", 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{"m", 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    "t"}], ")"}], 
                  UnderscriptBox["*", "\[ScriptCapitalP]"], "f"}]}]}]},
              {
               RowBox[{
                RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{"p", ",", 
                  RowBox[{"t", 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "f", 
                    "]"}]}]}], "]"}], 
                UnderscriptBox[
                 SubscriptBox["\[Rule]", 
                  RowBox[{"m", ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "f", 
                    "]"}]}]], "\[ScriptCapitalR]"], 
                RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{"q", ",", 
                  RowBox[{"t", 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "f", 
                    "]"}]}]}], "]"}]}]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[RightArrow] alternative \[ScriptCapitalP])", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1998060487", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6507705827893476`*^9, 3.650770749109527*^9}, {
  3.6507712192009287`*^9, 3.6507712490751133`*^9}, {3.650771733660219*^9, 
  3.650771734267168*^9}, {3.650772394605611*^9, 3.650772399030608*^9}, {
  3.6507731414885273`*^9, 3.650773156098703*^9}, {3.650776468615532*^9, 
  3.650776484399132*^9}, {3.6618486428381157`*^9, 3.6618486449285192`*^9}},
 CellTags->{
  "ID:1998060487", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] alternative \[ScriptCapitalP]"},
 CellID->1998060487],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1937013867]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->956674854],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(357)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(357)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["f", "ExpressionVariable"]}]], "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
             "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
               UnderscriptBox["-", "\[ScriptCapitalR]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["f", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["f", "ExpressionVariable"]}]], "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["*", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
             "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
               UnderscriptBox["-", "\[ScriptCapitalR]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["f", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1667457569"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1667457569-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1667457569-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1667457569", "Proof|ID:1667457569-1"},
 CellID->60009841],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1667457569-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->758686892]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] lm \[DoubleRightArrow])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] lm \[DoubleRightArrow])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                   StyleBox["f", "ExpressionVariable"], "]"}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["f", "ExpressionVariable"]], "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                   StyleBox["f", "ExpressionVariable"], "]"}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["f", "ExpressionVariable"]], "\[ScriptCapitalP]"], 
               StyleBox[
               "r", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2035027156"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2035027156-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2035027156-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2035027156", "Proof|ID:2035027156-1"},
 CellID->322996175],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2035027156-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->209011635]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] lm \[DoubleLeftArrow])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] lm \[DoubleLeftArrow])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["f", "ExpressionVariable"]], "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{
                    UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]], 
                  "\[ScriptCapitalP]"], 
                 StyleBox[
                 "r", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["f", "ExpressionVariable"]], "\[ScriptCapitalP]"], 
             StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                    StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{
                    UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]], 
                  "\[ScriptCapitalP]"], 
                 StyleBox[
                 "r", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2026457432"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2026457432-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2026457432-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2026457432", "Proof|ID:2026457432-1"},
 CellID->1357322543],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2026457432-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6467186657544394`*^9},
 CellID->1814571542]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reducibility \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reducibility \[ScriptCapitalP])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}], "]"}], 
                   "\[And]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}], "]"}], 
                   "\[And]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1167741853"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1167741853-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1167741853-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1167741853", "Proof|ID:1167741853-1"},
 CellID->293140950],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1167741853-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1342634176]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] alternative \[ScriptCapitalP])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] alternative \[ScriptCapitalP])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["f", "ExpressionVariable"]], "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["q", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], 
                    UnderscriptBox["-", "\[ScriptCapitalP]"], 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox["f", "ExpressionVariable"]}]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]], 
                    "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], 
                    "]"}]}]}]}]}], ")"}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["f", "ExpressionVariable"]], "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["q", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], 
                    UnderscriptBox["-", "\[ScriptCapitalP]"], 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox["f", "ExpressionVariable"]}]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]], 
                    "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}]}], 
                    "]"}]}]}]}]}], ")"}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1998060487"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1998060487-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1998060487-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1998060487", "Proof|ID:1998060487-1"},
 CellID->2076771442],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1998060487-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->754720624]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1265125254],

Cell[CellGroupData[{

Cell["\<\
Lemma: Sufficient Condition for \[LeftRightArrow] in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.63306575567596*^9, 3.63306576758329*^9}, {
  3.639367383204558*^9, 3.639367387723761*^9}},
 CellID->1808687580],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "c"}], "]"}]}]], 
  RowBox[{"c", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1709390289}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633065782142322*^9, 3.6330658029277773`*^9}, {
  3.635748117018989*^9, 3.635748118418285*^9}},
 CellTags->{
  "ID:1709390289", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1709390289],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "\[Gamma]", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "t", "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
            RowBox[{"\[Gamma]", ",", 
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "c", "]"}]}], 
            "]"}]},
          {
           RowBox[{"q", "\[Equal]", 
            RowBox[{"p", 
             UnderscriptBox["-", "\[ScriptCapitalP]"], 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"\[Gamma]", 
                UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
               ")"}], 
              UnderscriptBox["*", "\[ScriptCapitalP]"], "c"}]}]}]},
          {
           RowBox[{
            RowBox[{
             RowBox[{
              UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
              RowBox[{"p", ",", 
               RowBox[{"t", 
                UnderscriptBox["*", "\[ScriptCapitalT]"], 
                RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", 
                 "]"}]}]}], "]"}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
             RowBox[{
              RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{"p", ",", 
                RowBox[{"t", 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                  UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", 
                  "]"}]}]}], "]"}], 
              UnderscriptBox["-", "\[ScriptCapitalR]"], 
              RowBox[{"\[Gamma]", 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               RowBox[{
                UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "c", 
                "]"}]}]}]}], "\[Or]", 
            RowBox[{
             RowBox[{
              RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{"p", ",", 
                RowBox[{"t", 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                  UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", 
                  "]"}]}]}], "]"}], 
              UnderscriptBox["-", "\[ScriptCapitalR]"], 
              RowBox[{"\[Gamma]", 
               UnderscriptBox["*", "\[ScriptCapitalR]"], 
               RowBox[{
                UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "c", 
                "]"}]}]}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
             RowBox[{
              UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
              RowBox[{"p", ",", 
               RowBox[{"t", 
                UnderscriptBox["*", "\[ScriptCapitalT]"], 
                RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", 
                 "]"}]}]}], "]"}]}]}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"p", 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", 
      RowBox[{"{", "c", "}"}]], "\[ScriptCapitalP]"], "q"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[LeftRightArrow] alternative \[ScriptCapitalP])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1713414293", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6330658077112837`*^9, 3.63306600734698*^9}, {
  3.633066051296788*^9, 3.633066065398403*^9}, {3.633066184112385*^9, 
  3.633066185664896*^9}, {3.633066372680696*^9, 3.633066446832466*^9}, {
  3.6618486822281847`*^9, 3.6618486842561884`*^9}},
 CellTags->{
  "ID:1713414293", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[LeftRightArrow] alternative \[ScriptCapitalP]"},
 CellID->1713414293],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1615487369]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1793647858],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LeftRightArrow] alternative \[ScriptCapitalP])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LeftRightArrow] alternative \[ScriptCapitalP])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["\[Gamma]", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["\[Gamma]", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}], "]"}], 
                   "\[And]", 
                   RowBox[{
                    StyleBox["q", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], 
                    UnderscriptBox["-", "\[ScriptCapitalP]"], 
                    RowBox[{
                    RowBox[{
                    StyleBox["\[Gamma]", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    UnderscriptBox["-", "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["\[Gamma]", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    UnderscriptBox["-", "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["\[Gamma]", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], 
                    "]"}]}]}]}]}]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               StyleBox["c", "ExpressionVariable"]], "\[ScriptCapitalP]"], 
             StyleBox[
             "q", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["\[Gamma]", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["\[Gamma]", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}], "]"}], 
                   "\[And]", 
                   RowBox[{
                    StyleBox["q", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], 
                    UnderscriptBox["-", "\[ScriptCapitalP]"], 
                    RowBox[{
                    RowBox[{
                    StyleBox["\[Gamma]", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    UnderscriptBox["-", "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["\[Gamma]", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    UnderscriptBox["-", "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["\[Gamma]", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], 
                    "]"}]}]}]}]}]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               StyleBox["c", "ExpressionVariable"]], "\[ScriptCapitalP]"], 
             StyleBox[
             "q", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1713414293"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1713414293-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1713414293-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1713414293", "Proof|ID:1713414293-1"},
 CellID->793804010],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1713414293-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->162174877]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.633066107432878*^9, 3.633066108306217*^9}, 
   3.633067837186954*^9, 3.6364402947049103`*^9, 3.6434465852615013`*^9, 
   3.6532081420699863`*^9, 3.653213709942182*^9, 3.6661739718548813`*^9, 
   3.666440561576456*^9},
 CellID->2046956770],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "c"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"\[Gamma]", ",", 
     RowBox[{"lc", "[", "c", "]"}]}], "]"}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "=", 
    RowBox[{"p", "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], ")"}], "*", "c"}]}]}], 
   TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"p", ",", 
       RowBox[{"t", "*", 
        RowBox[{"lp", "[", "c", "]"}]}]}], "]"}], "\[Succeeds]", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"p", ",", 
        RowBox[{"t", "*", 
         RowBox[{"lp", "[", "c", "]"}]}]}], "]"}], "-", 
      RowBox[{"\[Gamma]", "*", 
       RowBox[{"lc", "[", "c", "]"}]}]}]}], "\[Or]", 
    RowBox[{
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"p", ",", 
        RowBox[{"t", "*", 
         RowBox[{"lp", "[", "c", "]"}]}]}], "]"}], "-", 
      RowBox[{"\[Gamma]", "*", 
       RowBox[{"lc", "[", "c", "]"}]}]}], "\[Succeeds]", 
     RowBox[{"C", "[", 
      RowBox[{"p", ",", 
       RowBox[{"t", "*", 
        RowBox[{"lp", "[", "c", "]"}]}]}], "]"}]}]}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "\[Gamma]", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 " and show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", 
    SubscriptBox["\[LeftRightArrow]", "c"], "q"}], TraditionalForm]]],
 "\nwhich, by (\[LeftRightArrow] alternative), amounts to showing\n(G\
\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "m", "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"mult", "[", 
          RowBox[{"m", ",", "c"}], "]"}]},
        {
         RowBox[{"q", "=", 
          RowBox[{"p", "-", 
           RowBox[{"m", "*", "c"}]}]}]},
        {
         RowBox[{
          RowBox[{"p", "\[Succeeds]", "q"}], "\[Or]", 
          RowBox[{"q", "\[Succeeds]", "p"}]}]}
       }]}]}]}], TraditionalForm]]],
 "\n(note that here we need (R1) in \[ScriptCapitalP]!).\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", ":=", 
    RowBox[{"\[Gamma]", "\[CenterDot]", "t"}]}], TraditionalForm]]],
 " is a suitable witness, so we show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], ",", "c"}], "]"}], 
   TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "\[Succeeds]", 
     RowBox[{"p", "-", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], ")"}], "*", "c"}]}]}], 
    "\[Or]", 
    RowBox[{
     RowBox[{"p", "-", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], ")"}], "*", "c"}]}], 
     "\[Succeeds]", "p"}]}], TraditionalForm]]],
 "\n(the third subgoal, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "=", 
    RowBox[{"p", "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], ")"}], "*", "c"}]}]}], 
   TraditionalForm]]],
 ", is already known).\n\n(G1):\nThis is equivalent to showing ",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"\[Gamma]", ",", 
     RowBox[{"lc", "[", "c", "]"}]}], "]"}], TraditionalForm]]],
 ", which follows immediately from (1).\n\n(G2):\nBased on (3) we distinguish \
two cases.\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"p", ",", 
      RowBox[{"t", "*", 
       RowBox[{"lp", "[", "c", "]"}]}]}], "]"}], "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"p", ",", 
       RowBox[{"t", "*", 
        RowBox[{"lp", "[", "c", "]"}]}]}], "]"}], "-", 
     RowBox[{"\[Gamma]", "*", 
      RowBox[{"lc", "[", "c", "]"}]}]}]}], TraditionalForm]]],
 ".\nWe show the first disjunct of (G2), i.e. we show\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[Succeeds]", 
    RowBox[{"p", "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], ")"}], "*", "c"}]}]}], 
   TraditionalForm]]],
 "\nBy (408), this is equivalent to showing\n(G3\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"p", ",", 
      RowBox[{
       RowBox[{"lp", "[", 
        RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], "]"}], "*", 
       RowBox[{"lp", "[", "c", "]"}]}]}], "]"}], "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"p", ",", 
       RowBox[{
        RowBox[{"lp", "[", 
         RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], "]"}], "*", 
        RowBox[{"lp", "[", "c", "]"}]}]}], "]"}], "-", 
     RowBox[{
      RowBox[{"lc", "[", 
       RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], "]"}], "*", 
      RowBox[{"lc", "[", "c", "]"}]}]}]}], TraditionalForm]]],
 "\nUsing the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", 
     RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], "]"}], "=", "t"}], 
   TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", 
     RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], "]"}], "=", "\[Gamma]"}], 
   TraditionalForm]]],
 " we are done, since (G3\[CloseCurlyQuote]) then follows immediately from \
our case assumption.\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"p", ",", 
       RowBox[{"t", "*", 
        RowBox[{"lp", "[", "c", "]"}]}]}], "]"}], "-", 
     RowBox[{"\[Gamma]", "*", 
      RowBox[{"lc", "[", "c", "]"}]}]}], "\[Succeeds]", 
    RowBox[{"C", "[", 
     RowBox[{"p", ",", 
      RowBox[{"t", "*", 
       RowBox[{"lp", "[", "c", "]"}]}]}], "]"}]}], TraditionalForm]]],
 ".\nWe show the second disjunct of (G2), i.e. we show\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], ")"}], "*", "c"}]}], 
    "\[Succeeds]", "p"}], TraditionalForm]]],
 "\nBy (408\[CloseCurlyQuote]), this is equivalent to showing\n(G4\
\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"p", ",", 
       RowBox[{
        RowBox[{"lp", "[", 
         RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], "]"}], "*", 
        RowBox[{"lp", "[", "c", "]"}]}]}], "]"}], "-", 
     RowBox[{
      RowBox[{"lc", "[", 
       RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], "]"}], "*", 
      RowBox[{"lc", "[", "c", "]"}]}]}], "\[Succeeds]", 
    RowBox[{"C", "[", 
     RowBox[{"p", ",", 
      RowBox[{
       RowBox[{"lp", "[", 
        RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], "]"}], "*", 
       RowBox[{"lp", "[", "c", "]"}]}]}], "]"}]}], TraditionalForm]]],
 "\nUsing the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", 
     RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], "]"}], "=", "t"}], 
   TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", 
     RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], "]"}], "=", "\[Gamma]"}], 
   TraditionalForm]]],
 " we are done, since (G4\[CloseCurlyQuote]) then follows immediately from \
our case assumption."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.633066114892542*^9, 3.6330662893847446`*^9}, {
   3.633066455098233*^9, 3.633066538888234*^9}, {3.6330665691891737`*^9, 
   3.633066937439796*^9}, {3.633067608500774*^9, 3.633067618055992*^9}, {
   3.633067650752081*^9, 3.633067837187079*^9}, 3.636440294705089*^9, {
   3.639209908842536*^9, 3.639209911843342*^9}, 3.643446585261623*^9, 
   3.6532081420699863`*^9, 3.653213709942182*^9, 3.6661739718548813`*^9, 
   3.666440561576557*^9},
 CellID->1965960939],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.633067826241599*^9, 3.63306783718721*^9}, 
   3.6364402947052526`*^9, 3.643446585261739*^9, 3.6532081420699863`*^9, 
   3.653213709942182*^9, 3.6661739718548813`*^9, 3.666440561576654*^9},
 TextAlignment->Right,
 CellID->1587336919]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Recursive Description of Reducibility in \[ScriptCapitalP] (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.6430005334895973`*^9, 3.643000543681677*^9}, {
  3.643009340892192*^9, 3.643009345398074*^9}},
 CellID->1164115734],

Cell["", "OpenEnvironment",
 CellID->776720212],

Cell[CellGroupData[{

Cell["Definition: isReducibleRec", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6430003846923237`*^9, 3.643000388680112*^9}},
 CellID->389987125],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "m"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducibleRec", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "q", ",", "m"}], "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"p", "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
         {
          RowBox[{"q", "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
         {
          RowBox[{
           UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", "]"}]},
         {
          RowBox[{"\[Or]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{
                RowBox[{
                 RowBox[{
                  RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
                  UnderscriptBox["*", "\[ScriptCapitalT]"], 
                  RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", 
                   "]"}]}], "\[Equal]", 
                 RowBox[{
                  UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", 
                  "]"}]}], "\[And]", 
                RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                  RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}],
                   ",", 
                  RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}],
                   ",", 
                  RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", 
                   "]"}]}], "]"}]}]},
              {
               RowBox[{
                UnderscriptBox["isReducibleRec", "\[ScriptCapitalP]"], "[", 
                RowBox[{
                 RowBox[{
                  UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], 
                 ",", "q", ",", "m"}], "]"}]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducibleRec \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1190720464", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643000392129566*^9, 3.6430004211106787`*^9}, {
  3.6430006705076838`*^9, 3.643000670958346*^9}, {3.643016419864513*^9, 
  3.643016446741425*^9}, {3.643016987495689*^9, 3.643017001870885*^9}},
 CellTags->{
  "ID:1190720464", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducibleRec \[ScriptCapitalP]"},
 CellID->1190720464],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->129235337]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1043867076],

Cell["", "OpenEnvironment",
 CellID->909379877],

Cell[CellGroupData[{

Cell["\<\
Theorem: Recursive Description of Reducibility in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.64300045411799*^9, {3.643101050567721*^9, 3.64310106880665*^9}},
 CellID->130808008],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"q", ",", "m", ",", "p"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 166529906}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64300021272745*^9, 3.643000224948461*^9}, 
   3.643000756192852*^9, 3.643001057833845*^9, {3.6430164544583063`*^9, 
   3.6430164546822977`*^9}, {3.6430886270041857`*^9, 3.6430886519889097`*^9}},
 
 CellTags->{
  "ID:166529906", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->166529906],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "m"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducibleRec", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "m"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(352.\[DoubleRightArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:844782115", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64300031398755*^9, 3.643000367962772*^9}, {
  3.643000562862221*^9, 3.643000571355454*^9}, {3.643000759179118*^9, 
  3.643000765779202*^9}, {3.643001060336617*^9, 3.6430010625383787`*^9}, {
  3.643001829506926*^9, 3.643001834467746*^9}, {3.6430164582010803`*^9, 
  3.6430164610518217`*^9}, {3.643088633219652*^9, 3.6430886624913607`*^9}},
 CellTags->{
  "ID:844782115", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352.\[DoubleRightArrow]"},
 CellID->844782115],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducibleRec", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "m"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "m"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(352.\[DoubleLeftArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:953126445", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64300477006153*^9, 3.643004776031273*^9}, {
  3.643016463478711*^9, 3.643016466055084*^9}, {3.6430886381463118`*^9, 
  3.643088667146154*^9}},
 CellTags->{
  "ID:953126445", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352.\[DoubleLeftArrow]"},
 CellID->953126445],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "m"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isReducibleRec", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "m"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(352)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:259818613", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643000017837538*^9, 3.643000041805283*^9}, {
   3.643000098355447*^9, 3.6430001903749533`*^9}, {3.643000493695607*^9, 
   3.643000503119534*^9}, {3.6430007719659977`*^9, 3.643000780215439*^9}, 
   3.643001065834052*^9, {3.643016468478794*^9, 3.64301647120498*^9}, 
   3.643016511082316*^9},
 CellTags->{
  "ID:259818613", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352"},
 CellID->259818613],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"p", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
       {
        RowBox[{"q", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}]},
       {
        RowBox[{"\[Or]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {
             RowBox[{
              RowBox[{
               RowBox[{
                UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               RowBox[{
                UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
              "\[And]", 
              RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
                ",", 
                RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}],
                "]"}]}]},
            {
             RowBox[{
              UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
              RowBox[{
               RowBox[{
                UnderscriptBox["R", "\[ScriptCapitalP]"], "[", "p", "]"}], 
               ",", "q"}], "]"}]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(353)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:182112345", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643016531219151*^9, 3.6430165866428623`*^9}},
 CellTags->{
  "ID:182112345", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "353"},
 CellID->182112345],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->506272053]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->352819612],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(352.\[DoubleRightArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(352.\[DoubleRightArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducibleRec", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducibleRec", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:844782115"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p844782115-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p844782115-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:844782115", "Proof|ID:844782115-1"},
 CellID->1659422444],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p844782115-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->565248221]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(352.\[DoubleLeftArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(352.\[DoubleLeftArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducibleRec", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducibleRec", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:953126445"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p953126445-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p953126445-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:953126445", "Proof|ID:953126445-1"},
 CellID->1715888170],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p953126445-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1477222806]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(352)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(352)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isReducibleRec", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isReducibleRec", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:259818613"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p259818613-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p259818613-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:259818613", "Proof|ID:259818613-1"},
 CellID->816722537],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p259818613-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->701939384]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(353)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(353)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
                 "\[Or]", 
                 RowBox[{
                   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["q", "ExpressionVariable"]}], "]"}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
                 "\[Or]", 
                 RowBox[{
                   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["q", "ExpressionVariable"]}], "]"}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:182112345"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p182112345-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p182112345-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:182112345", "Proof|ID:182112345-1"},
 CellID->908791481],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p182112345-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->640786301]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R5 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6320299019975357`*^9, 3.632029902590104*^9}, {
  3.638172945631569*^9, 3.638172950784655*^9}},
 CellID->1569279607],

Cell["", "OpenEnvironment",
 CellID->1776685170],

Cell[CellGroupData[{

Cell["\<\
Lemma: Auxiliary Results needed for proving R5 in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.632033173928082*^9, {3.632041145055563*^9, 3.632041160071105*^9}},
 CellID->1082744603],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "c"}], "]"}]}]], 
  RowBox[{"c", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 418362312}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632033185315455*^9, 3.632033248810408*^9}, {
  3.632033705245881*^9, 3.632033707637813*^9}, {3.632116508927549*^9, 
  3.6321165099619017`*^9}, {3.6321178255183077`*^9, 3.632117834152191*^9}, {
  3.635747923378693*^9, 3.635747924888525*^9}},
 CellTags->{
  "ID:418362312", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->418362312],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "t", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 2116744492}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632033712193735*^9, 3.632033719830206*^9}},
 CellTags->{
  "ID:2116744492", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->2116744492],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "d", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"a", ",", 
        RowBox[{"t", 
         UnderscriptBox["*", "\[ScriptCapitalT]"], 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}], 
       "]"}], 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{"a", ",", 
         RowBox[{"t", 
          UnderscriptBox["*", "\[ScriptCapitalT]"], 
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}], 
        "]"}], 
       UnderscriptBox["-", "\[ScriptCapitalR]"], 
       RowBox[{"d", 
        UnderscriptBox["*", "\[ScriptCapitalR]"], 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
      RowBox[{"a", 
       UnderscriptBox["-", "\[ScriptCapitalP]"], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", 
          UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ")"}], 
        UnderscriptBox["*", "\[ScriptCapitalP]"], "c"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(391)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1746803457", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632052334444951*^9, 3.632052393021345*^9}, {
  3.632052446671076*^9, 3.632052461122974*^9}, {3.63205249146375*^9, 
  3.632052495680492*^9}, {3.632052543732326*^9, 3.632052584909204*^9}, {
  3.632052656207632*^9, 3.632052731915082*^9}, {3.632116352640526*^9, 
  3.632116398991864*^9}, {3.632117848909634*^9, 3.6321178964188423`*^9}, {
  3.632717401987567*^9, 3.632717408673616*^9}, {3.638100129687187*^9, 
  3.638100131614633*^9}},
 CellTags->{
  "ID:1746803457", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "391"},
 CellID->1746803457],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
   RowBox[{"[", "T", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 899106457}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632033274392638*^9, 3.632033288349742*^9}, {
  3.635749228090148*^9, 3.635749234783721*^9}},
 CellTags->{
  "ID:899106457", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->899106457],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"S", "=", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      SubscriptBox["T", "i"], 
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1157329004}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63203348338993*^9, 3.632033511548327*^9}, 
   3.632033724577681*^9, {3.6327174129727697`*^9, 3.632717421466036*^9}},
 CellTags->{
  "ID:1157329004", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1157329004],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["reduceM", "\[ScriptCapitalR]"], "[", 
       RowBox[{
        RowBox[{
         UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
         RowBox[{"a", ",", 
          RowBox[{"t", 
           UnderscriptBox["*", "\[ScriptCapitalT]"], 
           RowBox[{
            UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}], 
         "]"}], ",", 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "c", "]"}], ",", "T",
         ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "c", "]"}], ",", 
      SubscriptBox["T", "i"]}], "]"}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
       RowBox[{"a", ",", "c", ",", "S", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", "c", ",", 
      SubscriptBox["S", "i"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(387)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1927781160", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632033300301496*^9, 3.6320334634625807`*^9}, {
   3.632033532005786*^9, 3.6320335557992773`*^9}, 3.632033730724481*^9, {
   3.6320454668953457`*^9, 3.6320454716361628`*^9}, {3.632116435934978*^9, 
   3.632116444098564*^9}, {3.6381578323375063`*^9, 3.638157840790576*^9}},
 CellTags->{
  "ID:1927781160", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "387"},
 CellID->1927781160],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->755751524]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->977857587],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(391)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(391)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                 UnderscriptBox["-", "\[ScriptCapitalR]"], 
                 RowBox[{
                   StyleBox["d", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", "\[ScriptCapitalP]"], 
               RowBox[{
                 RowBox[{
                   StyleBox["d", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                 UnderscriptBox["-", "\[ScriptCapitalR]"], 
                 RowBox[{
                   StyleBox["d", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", "\[ScriptCapitalP]"], 
               RowBox[{
                 RowBox[{
                   StyleBox["d", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1746803457"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1746803457-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1746803457-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1746803457", "Proof|ID:1746803457-1"},
 CellID->311879395],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1746803457-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->227046330]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(387)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(387)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                 ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                  "\[RightAngleBracket]"}], 
                StyleBox["i", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                 ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                  "\[RightAngleBracket]"}], 
                StyleBox["i", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1927781160"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1927781160-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1927781160-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1927781160", "Proof|ID:1927781160-1"},
 CellID->235488096],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1927781160-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->808907405]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1780168981],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R5", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6320299123124723`*^9, 3.632029917329012*^9}},
 CellID->1235526869],

Cell[BoxData[
 RowBox[{"R5", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(368)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:934500559", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185608049573*^9, 3.610185626090595*^9}, 
   3.6315968636157303`*^9},
 CellTags->{
  "ID:934500559", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "368"},
 CellID->934500559],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->887114206]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1446998866],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(368)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(368)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R5", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R5", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:934500559"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p934500559-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p934500559-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:934500559", "Proof|ID:934500559-1"},
 CellID->80140931],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p934500559-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->426435]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R6 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6320443304506903`*^9, 3.632044330674206*^9}, {
  3.638611906127323*^9, 3.638611912850279*^9}},
 CellID->409755120],

Cell["", "OpenEnvironment",
 CellID->853392754],

Cell[CellGroupData[{

Cell["\<\
Theorem: Properties IRR and CORR (Alternative Description of Irrelativity and \
Correlativity in \[ScriptCapitalP])\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632026674240426*^9, 3.6320266896391993`*^9}, {
  3.6324815816581306`*^9, 3.632481593312456*^9}},
 CellID->87372790],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "m1", ",", "m2"}], "]"}]}]], 
  RowBox[{
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 371274476}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632481602151176*^9, 3.63248161223946*^9}, {
  3.632632527263934*^9, 3.632632549639126*^9}, {3.635747945369705*^9, 
  3.635747948753634*^9}},
 CellTags->{
  "ID:371274476", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->371274476],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
     RowBox[{"m1", ",", "p"}], "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
     RowBox[{"m2", ",", "p"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["areIrrelative", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "p", ",", "m1", ",", "m2", ",", "i", ",", "j"}], 
      "]"}], "\[DoubleLongLeftRightArrow]", 
     RowBox[{
      UnderscriptBox["areIrrelative", "\[ScriptCapitalR]"], "[", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
       RowBox[{
        UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
       RowBox[{
        UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m1", "]"}], ",", 
       RowBox[{
        UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m2", "]"}], ",", "i",
        ",", "j"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(IRR)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1588964752", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320297073702183`*^9, 3.632029838506536*^9}, {
  3.632481652929577*^9, 3.632481666618137*^9}, {3.63255288004869*^9, 
  3.632552923338077*^9}, {3.632632563422843*^9, 3.6326325769601107`*^9}},
 CellTags->{
  "ID:1588964752", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR"},
 CellID->1588964752],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
     RowBox[{"m1", ",", "p"}], "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
     RowBox[{"m2", ",", "q"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["areIrrelative", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "q", ",", "m1", ",", "m2", ",", "i", ",", "j"}], 
      "]"}], "\[DoubleLongLeftRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"p", "\[Equal]", "q"}], "\[Implies]", 
       RowBox[{
        UnderscriptBox["areIrrelative", "\[ScriptCapitalR]"], "[", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", 
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m1", "]"}], ",", 
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m2", "]"}], ",", 
         "i", ",", "j"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(IRR II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:428389793", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.6326324158580627`*^9, {3.6326324809084663`*^9, 3.632632489530683*^9}, {
   3.6326325807816153`*^9, 3.632632593169077*^9}, {3.632632626242433*^9, 
   3.632632627132753*^9}, {3.632632787276595*^9, 3.632632808867795*^9}},
 CellTags->{
  "ID:428389793", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR II"},
 CellID->428389793],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
     RowBox[{"m1", ",", "p"}], "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
     RowBox[{"m2", ",", "p"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["areCorrelative", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "m1", ",", "m2"}], "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox["areCorrelative", "\[ScriptCapitalR]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m1", "]"}], ",", 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m2", "]"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CORR)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:225111715", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632481676348637*^9, 3.632481687129097*^9}, {
  3.6325529076897297`*^9, 3.632552908915201*^9}, {3.632632601756888*^9, 
  3.632632612986068*^9}},
 CellTags->{
  "ID:225111715", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CORR"},
 CellID->225111715],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"t1", ",", "t2"}], "]"}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1632881403}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63877785194591*^9, 3.638777868538465*^9}},
 CellTags->{
  "ID:1632881403", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1632881403],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["areIrrelative", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "q", ",", 
      RowBox[{"c1", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t1"}], ",", 
      RowBox[{"c2", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t2"}], ",", "i", 
      ",", "j"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"p", "\[Equal]", "q"}], "\[Implies]", 
      RowBox[{
       UnderscriptBox["areIrrelative", "\[ScriptCapitalR]"], "[", 
       RowBox[{
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", 
        "c1", ",", "c2", ",", "i", ",", "j"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(IRR \[CenterDot])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1701548086", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638777587980084*^9, 3.638777727030727*^9}, 
   3.638777874128812*^9, {3.638780070080114*^9, 3.638780074689773*^9}, {
   3.6387803697541943`*^9, 3.638780370934822*^9}},
 CellTags->{
  "ID:1701548086", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR \[CenterDot]"},
 CellID->1701548086],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areCorrelative", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", 
    RowBox[{"c1", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t1"}], ",", 
    RowBox[{"c2", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t2"}]}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["areCorrelative", "\[ScriptCapitalR]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", "c1", 
    ",", "c2"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CORR \[CenterDot])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1758251641", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638777910096033*^9, 3.638777964396413*^9}},
 CellTags->{
  "ID:1758251641", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CORR \[CenterDot]"},
 CellID->1758251641],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1086037741]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->97059366],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(IRR)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(IRR)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
                 RowBox[{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["areIrrelative", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["areIrrelative", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
                 RowBox[{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["areIrrelative", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["areIrrelative", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1588964752"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1588964752-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1588964752-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1588964752", "Proof|ID:1588964752-1"},
 CellID->543579358],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1588964752-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->381837386]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(IRR II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(IRR II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
                 RowBox[{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["areIrrelative", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{"(", 
                 RowBox[{
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
                   RowBox[{
                    UnderscriptBox["areIrrelative", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
                 ")"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
                 RowBox[{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["areIrrelative", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{"(", 
                 RowBox[{
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
                   RowBox[{
                    UnderscriptBox["areIrrelative", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
                 ")"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:428389793"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p428389793-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p428389793-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:428389793", "Proof|ID:428389793-1"},
 CellID->490546380],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p428389793-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1790028040]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CORR)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CORR)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["areCorrelative", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["areCorrelative", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["areCorrelative", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["areCorrelative", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:225111715"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p225111715-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p225111715-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:225111715", "Proof|ID:225111715-1"},
 CellID->657130363],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p225111715-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->217274233]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(IRR \[CenterDot])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(IRR \[CenterDot])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["areIrrelative", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t1", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   StyleBox["c2", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t2", "ExpressionVariable"]}], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["areIrrelative", "\[ScriptCapitalR]"], "[", 
                   
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["areIrrelative", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t1", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   StyleBox["c2", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t2", "ExpressionVariable"]}], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["areIrrelative", "\[ScriptCapitalR]"], "[", 
                   
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1701548086"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1701548086-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1701548086-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1701548086", "Proof|ID:1701548086-1"},
 CellID->384143515],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1701548086-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1787100506]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CORR \[CenterDot])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CORR \[CenterDot])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["areCorrelative", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t1", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   StyleBox["c2", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t2", "ExpressionVariable"]}]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["areCorrelative", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["areCorrelative", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t1", "ExpressionVariable"]}], ",", 
                 RowBox[{
                   StyleBox["c2", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t2", "ExpressionVariable"]}]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["areCorrelative", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1758251641"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1758251641-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1758251641-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1758251641", "Proof|ID:1758251641-1"},
 CellID->1782924340],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1758251641-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->599328732]
}, Closed]],

Cell[TextData[{
 StyleBox["Remark",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 ": The proofs of (IRR \[CenterDot]) and (CORR \[CenterDot]) are rather \
lengthy. This is (at least partly) due to the fact that the very definition \
of the sets of multipliers in \[ScriptCapitalP], i.e. formula \
(hasMultipliers), was used instead of the various formulas in subsubsection \
\[OpenCurlyDoubleQuote]Preparations\[CloseCurlyDoubleQuote]/\
\[CloseCurlyDoubleQuote]Multipliers\[CloseCurlyDoubleQuote].\nMoreover, \
formula (CORR \[CenterDot]) seems to be never used."
}], "Text",
 CellChangeTimes->{{3.6427604255566463`*^9, 3.642760515035181*^9}, {
  3.642834812675721*^9, 3.642834827970357*^9}},
 CellID->843394468],

Cell["", "OpenEnvironment",
 CellID->363999473],

Cell[CellGroupData[{

Cell["Lemma", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6322067172992992`*^9, 3.632206717450667*^9}, {
  3.632213259443553*^9, 3.632213259588537*^9}},
 CellID->1652502037],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{
    "a", ",", "c", ",", "c1", ",", "c2", ",", "m1", ",", "m2", ",", "b1", ",",
      "b2"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1928324287}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632483537126874*^9, 3.632483544636087*^9}, {
  3.632566161752615*^9, 3.632566163355255*^9}, {3.638246107958171*^9, 
  3.638246109227172*^9}},
 CellTags->{
  "ID:1928324287", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1928324287],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"a", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", 
           RowBox[{"m1", ",", "c"}]], "\[ScriptCapitalP]"], "b1"}]},
       {
        RowBox[{"a", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", 
           RowBox[{"m2", ",", "c"}]], "\[ScriptCapitalP]"], "b2"}]},
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m1", "]"}], 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m2", "]"}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{"b2", ",", "c", ",", "m1"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(404)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:751250157", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632206762212055*^9, 3.632206853957625*^9}, {
  3.632206905937357*^9, 3.632206909722642*^9}, {3.63248354291285*^9, 
  3.632483549104254*^9}},
 CellTags->{
  "ID:751250157", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404"},
 CellID->751250157],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
         RowBox[{"a", ",", "c", ",", "m1"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
         RowBox[{"a", ",", "c", ",", "m2"}], "]"}]},
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m1", "]"}], 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m2", "]"}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["-", "\[ScriptCapitalP]"], 
     RowBox[{"m2", 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "c"}]}], ",", "c", ",", 
    "m1"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(404 II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1952511245", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632483559567326*^9, 3.632483610243424*^9}},
 CellTags->{
  "ID:1952511245", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404 II"},
 CellID->1952511245],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
         RowBox[{"a", ",", "c1", ",", "m1"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
         RowBox[{"a", ",", "c2", ",", "m2"}], "]"}]},
       {
        RowBox[{
         RowBox[{
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m1", "]"}], 
          UnderscriptBox["*", "\[ScriptCapitalT]"], 
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c1", "]"}]}], 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
         RowBox[{
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m2", "]"}], 
          UnderscriptBox["*", "\[ScriptCapitalT]"], 
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c2", "]"}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["-", "\[ScriptCapitalP]"], 
     RowBox[{"m2", 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "c2"}]}], ",", "c1", ",", 
    "m1"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(404 III)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:537853020", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632566167338171*^9, 3.632566173536248*^9}, {
  3.63256622584578*^9, 3.632566243000071*^9}, {3.63824877939221*^9, 
  3.6382487854929037`*^9}},
 CellTags->{
  "ID:537853020", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404 III"},
 CellID->537853020],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1877713982]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->970753228],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(404)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(404)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m1", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m2", "ExpressionVariable"], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["b2", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m1", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m2", "ExpressionVariable"], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["b2", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:751250157"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p751250157-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p751250157-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:751250157", "Proof|ID:751250157-1"},
 CellID->1336000539],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p751250157-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6382482402319107`*^9},
 CellID->1708610251]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(404 II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(404 II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m1", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m2", "ExpressionVariable"], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m1", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m2", "ExpressionVariable"], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1952511245"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1952511245-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1952511245-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1952511245", "Proof|ID:1952511245-1"},
 CellID->29876965],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1952511245-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->99609695]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(404 III)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(404 III)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c1", "ExpressionVariable"], "]"}]}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c2", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["c2", "ExpressionVariable"]}]}], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c1", "ExpressionVariable"], "]"}]}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", "\[ScriptCapitalT]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c2", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", "\[ScriptCapitalP]"], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["c2", "ExpressionVariable"]}]}], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:537853020"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p537853020-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p537853020-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:537853020", "Proof|ID:537853020-1"},
 CellID->1965128634],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p537853020-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->358062202]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->670924166],

Cell[CellGroupData[{

Cell["\<\
Lemma: Auxiliary Results for proving R6 in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632044370409906*^9, 3.6320443826254807`*^9}},
 CellID->1216178880],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "c"}], "]"}]}]], 
  RowBox[{"c", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 659046058}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632033185315455*^9, 3.632033248810408*^9}, {
  3.632033705245881*^9, 3.632033707637813*^9}, {3.6357479654451323`*^9, 
  3.635747967652893*^9}},
 CellTags->{
  "ID:659046058", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->659046058],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "t", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 319520004}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632033712193735*^9, 3.632033719830206*^9}},
 CellTags->{
  "ID:319520004", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->319520004],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
   RowBox[{"[", "T", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 967033090}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632033274392638*^9, 3.632033288349742*^9}, {
  3.635749248697617*^9, 3.635749255512679*^9}},
 CellTags->{
  "ID:967033090", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->967033090],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"S", "=", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      SubscriptBox["T", "i"], 
      UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1647700871}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63203348338993*^9, 3.632033511548327*^9}, 
   3.632033724577681*^9, {3.6327174522532787`*^9, 3.63271745869342*^9}},
 CellTags->{
  "ID:1647700871", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1647700871],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "T", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1981597108}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632045515272135*^9, 3.632045523546506*^9}},
 CellTags->{
  "ID:1981597108", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1981597108],

Cell["\<\
(388) is analogous to (387) in the previous subsection, for proving (R5), and \
its proof goes along exactly the same lines.\
\>", "EnvironmentText",
 CellChangeTimes->{{3.632046266357478*^9, 3.632046297474556*^9}, {
  3.632046334125416*^9, 3.632046335713366*^9}, {3.632048385925508*^9, 
  3.632048388118894*^9}, {3.6381838545373907`*^9, 3.638183856542057*^9}},
 CellID->130440480],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["symred", "\[ScriptCapitalR]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "\[ScriptCapitalR]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"a", ",", 
        RowBox[{"t", 
         UnderscriptBox["*", "\[ScriptCapitalT]"], 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}], 
       "]"}], ",", 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "c", "]"}], ",", "T", 
      ",", 
      RowBox[{"i", "-", "1"}]}], "]"}], ",", 
    RowBox[{
     UnderscriptBox["reduceM", "\[ScriptCapitalR]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"a", ",", 
        RowBox[{"t", 
         UnderscriptBox["*", "\[ScriptCapitalT]"], 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "c", "]"}]}]}], 
       "]"}], ",", 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "c", "]"}], ",", "T", 
      ",", "i"}], "]"}], ",", 
    RowBox[{
     UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "c", "]"}], ",", 
    SubscriptBox["T", "i"]}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["symred", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
     RowBox[{"a", ",", "c", ",", "S", ",", 
      RowBox[{"i", "-", "1"}]}], "]"}], ",", 
    RowBox[{
     UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
     RowBox[{"a", ",", "c", ",", "S", ",", "i"}], "]"}], ",", "c", ",", 
    SubscriptBox["S", "i"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(388)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:40011060", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632044402083807*^9, 3.632044438322399*^9}, {
   3.632045525938972*^9, 3.632045533629942*^9}, 3.632050068350527*^9},
 CellTags->{
  "ID:40011060", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "388"},
 CellID->40011060],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"b1", ",", "b2", ",", "m1", ",", "m2"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 884639747}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320484011822653`*^9, 3.632048424335042*^9}},
 CellTags->{
  "ID:884639747", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->884639747],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"m1", ",", "c"}]], "\[ScriptCapitalP]"], "b1"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"m2", ",", "c"}]], "\[ScriptCapitalP]"], "b2"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     UnderscriptBox["areCorrelative", "\[ScriptCapitalP]"], "[", 
     RowBox[{"c", ",", "m1", ",", "m2"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1360655817}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632048438084916*^9, 3.632048500681981*^9}, {
  3.63204854426985*^9, 3.6320485502564793`*^9}, {3.632481722364827*^9, 
  3.632481724542646*^9}, {3.6324817699377937`*^9, 3.632481770197447*^9}},
 CellTags->{
  "ID:1360655817", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1360655817],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m1", "]"}], "\[Equal]", 
   RowBox[{
    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m2", "]"}]}], 
  "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
      RowBox[{"[", "TM", "]"}]}]], 
    RowBox[{
     UnderscriptBox["isR6Witness", "\[ScriptCapitalP]"], "[", 
     RowBox[{
     "TM", ",", "a", ",", "c", ",", "m1", ",", "m2", ",", "b1", ",", "b2"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(389)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1002690147", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632048532151217*^9, 3.632048588930612*^9}, 
   3.632048676789915*^9, {3.6320494613082237`*^9, 3.632049497893783*^9}, {
   3.635749262030164*^9, 3.6357492687940397`*^9}},
 CellTags->{
  "ID:1002690147", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "389"},
 CellID->1002690147],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m1", "]"}], 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
   RowBox[{
    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m2", "]"}]}], 
  "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
      RowBox[{"[", "TM", "]"}]}]], 
    RowBox[{
     UnderscriptBox["isR6Witness", "\[ScriptCapitalP]"], "[", 
     RowBox[{
     "TM", ",", "a", ",", "c", ",", "m1", ",", "m2", ",", "b1", ",", "b2"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(390)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1896907436", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320495518059273`*^9, 3.632049559575954*^9}, {
  3.635749273172243*^9, 3.635749279638802*^9}},
 CellTags->{
  "ID:1896907436", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "390"},
 CellID->1896907436],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2045347914]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1856156737],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(388)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(388)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["symred", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                 ",", 
                 RowBox[{
                   UnderscriptBox["reduceM", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["symred", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               RowBox[{
                 UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                  "\[RightAngleBracket]"}], 
                StyleBox["i", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["symred", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                 ",", 
                 RowBox[{
                   UnderscriptBox["reduceM", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}], "]"}], 
                    ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["symred", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               RowBox[{
                 UnderscriptBox["reduceM", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                  "\[RightAngleBracket]"}], 
                StyleBox["i", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:40011060"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p40011060-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p40011060-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:40011060", "Proof|ID:40011060-1"},
 CellID->1454431248],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p40011060-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->1520544499]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(389)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(389)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m1", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m2", "ExpressionVariable"], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
                   "[", 
                   StyleBox["TM", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["isR6Witness", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["TM", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["b1", "ExpressionVariable"], ",", 
                 StyleBox["b2", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m1", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m2", "ExpressionVariable"], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
                   "[", 
                   StyleBox["TM", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["isR6Witness", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["TM", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["b1", "ExpressionVariable"], ",", 
                 StyleBox["b2", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1002690147"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1002690147-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1002690147-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1002690147", "Proof|ID:1002690147-1"},
 CellID->615225779],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1002690147-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {61., 67.}}]], "ProofInfoBody",
 CellID->543645298]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(390)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(390)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m1", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m2", "ExpressionVariable"], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
                   "[", 
                   StyleBox["TM", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["isR6Witness", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["TM", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["b1", "ExpressionVariable"], ",", 
                 StyleBox["b2", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                "\[ScriptCapitalP]"], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m1", "ExpressionVariable"], "]"}], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 StyleBox["m2", "ExpressionVariable"], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
                   "[", 
                   StyleBox["TM", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["isR6Witness", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["TM", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["b1", "ExpressionVariable"], ",", 
                 StyleBox["b2", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1896907436"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1896907436-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1896907436-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1896907436", "Proof|ID:1896907436-1"},
 CellID->375199501],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1896907436-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1428296947]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1502411461],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R6", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6320443387572937`*^9, 3.6320443447822847`*^9}},
 CellID->2093688270],

Cell[BoxData[
 RowBox[{"R6", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(369)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1290694758", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185608721056*^9, 3.610185627203398*^9}, 
   3.63159686424844*^9},
 CellTags->{
  "ID:1290694758", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "369"},
 CellID->1290694758],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->902566628]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->163822355],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(369)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(369)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R6", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R6", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1290694758"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1290694758-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1290694758-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1290694758", "Proof|ID:1290694758-1"},
 CellID->796556125],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1290694758-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1734938744]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Alternative Descriptions of isNTCR (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.6322163035744963`*^9, 3.632216304361657*^9}, {
  3.6327217923610973`*^9, 3.632721793461586*^9}, {3.632721925604723*^9, 
  3.63272193796209*^9}, {3.6391227367251387`*^9, 3.639122741457851*^9}},
 CellID->387021889],

Cell["", "OpenEnvironment",
 CellID->1641500050],

Cell[CellGroupData[{

Cell["Definition: areCommonReducibles, isNTCRP", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632649196373774*^9, 3.6326492039863873`*^9}},
 CellID->1266577052],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1785082178}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632649215255764*^9, 3.632649267941267*^9}, 
   3.632649322736561*^9, {3.6357479924346943`*^9, 3.635747995451763*^9}},
 CellTags->{
  "ID:1785082178", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1785082178],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "m1", ",", "m2"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1718264802}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632649227674017*^9, 3.632649235276697*^9}, {
  3.632649325064954*^9, 3.632649328130705*^9}},
 CellTags->{
  "ID:1718264802", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1718264802],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 752304273}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6326492730053062`*^9, 3.632649284178994*^9}},
 CellTags->{
  "ID:752304273", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->752304273],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], "[", 
   RowBox[{
   "a", ",", "b", ",", "p", ",", "q", ",", "m1", ",", "m2", ",", "i", ",", 
    "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{"p", "\[Equal]", "q"}], "\[Implies]", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
           RowBox[{
            UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
            RowBox[{"m1", ",", 
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",",
              "i"}], "]"}], "\[And]", 
           RowBox[{
            UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
            RowBox[{"m2", ",", 
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",",
              "j"}], "]"}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
         RowBox[{"a", ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
          "m1"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
         RowBox[{"b", ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", 
          "m2"}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areCommonReducibles I)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:578561747", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632649296179384*^9, 3.632649421914858*^9}, {
   3.632649477930554*^9, 3.632649478191009*^9}, 3.638676014873288*^9},
 CellTags->{
  "ID:578561747", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "areCommonReducibles I"},
 CellID->578561747],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], "[", 
   RowBox[{"a", ",", "b", ",", "p", ",", "q", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"m3", ",", "m4"}], "]"}]}]], 
    RowBox[{
     UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], "[", 
     RowBox[{
     "a", ",", "b", ",", "p", ",", "q", ",", "m3", ",", "m4", ",", "i", ",", 
      "j"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areCommonReducibles II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1399766637", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632649440535715*^9, 3.632649475529398*^9}},
 CellTags->{
  "ID:1399766637", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "areCommonReducibles II"},
 CellID->1399766637],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], "[", 
     RowBox[{"a", ",", "a", ",", "p", ",", "q", ",", "i", ",", "j"}], "]"}], 
    "\[And]", 
    RowBox[{"\[Not]", 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
        RowBox[{"[", 
         RowBox[{"m3", ",", "m4"}], "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], "[", 
             
             RowBox[{
             "a", ",", "a", ",", "p", ",", "q", ",", "m3", ",", "m4", ",", 
              "i", ",", "j"}], "]"}]},
           {
            RowBox[{
             RowBox[{
              UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
              RowBox[{
               RowBox[{"a", 
                UnderscriptBox["-", "\[ScriptCapitalR]"], 
                RowBox[{"m3", 
                 UnderscriptBox["*", "\[ScriptCapitalR]"], 
                 RowBox[{
                  UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", 
                  "]"}]}]}], ",", 
               RowBox[{
                UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], 
               ",", "m4"}], "]"}], "\[Or]", 
             RowBox[{
              UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
              RowBox[{
               RowBox[{"a", 
                UnderscriptBox["-", "\[ScriptCapitalR]"], 
                RowBox[{"m4", 
                 UnderscriptBox["*", "\[ScriptCapitalR]"], 
                 RowBox[{
                  UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                  "]"}]}]}], ",", 
               RowBox[{
                UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
               ",", "m3"}], "]"}]}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isNTCRP)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1132709394", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632649498259028*^9, 3.632649608555079*^9}, 
   3.638675260469015*^9},
 CellTags->{
  "ID:1132709394", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP"},
 CellID->1132709394],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1532258754]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1086432727],

Cell["", "OpenEnvironment",
 CellID->45003320],

Cell[CellGroupData[{

Cell["Lemma: Properties of areCommonReducibles and isNTCRP", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6331743221995087`*^9, 3.633174328848733*^9}, {
  3.633175463697435*^9, 3.633175467031475*^9}},
 CellID->1810500356],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 176358731}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633174335362508*^9, 3.633174354240876*^9}, {
  3.635748002330697*^9, 3.635748005492724*^9}},
 CellTags->{
  "ID:176358731", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->176358731],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 391507392}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633174367916624*^9, 3.633174375082808*^9}, {
   3.633174509351717*^9, 3.6331745096465683`*^9}, 3.633174546878871*^9, {
   3.633174776139892*^9, 3.633174776324893*^9}},
 CellTags->{
  "ID:391507392", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->391507392],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1030826183}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633174378926409*^9, 3.633174391038684*^9}},
 CellTags->{
  "ID:1030826183", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1030826183],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", "\[Equal]", "q"}], "\[Or]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
     "\[NotEqual]", 
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], "[", 
     RowBox[{"a", ",", "a", ",", "p", ",", "q", ",", "i", ",", "j"}], "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox["isCommonReducible", "\[ScriptCapitalR]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", "a", 
      ",", "i", ",", "j"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(354)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:926282876", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6331744172599773`*^9, 3.633174482753427*^9}, {
  3.6386757936637907`*^9, 3.638675795442774*^9}},
 CellTags->{
  "ID:926282876", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "354"},
 CellID->926282876],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2"}], "]"}]}]], 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], "[", 
     RowBox[{
     "a", ",", "b", ",", "p", ",", "q", ",", "m1", ",", "m2", ",", "i", ",", 
      "j"}], "]"}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
       RowBox[{"a", ",", 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
        "m1"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
       RowBox[{"b", ",", 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", 
        "m2"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ")"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(355)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:80931797", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6331744942600307`*^9, 3.633174601151969*^9}, {
  3.6331747779476147`*^9, 3.6331747780948257`*^9}},
 CellTags->{
  "ID:80931797", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "355"},
 CellID->80931797],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], "[", 
   RowBox[{"a", ",", "b", ",", "p", ",", "q", ",", "i", ",", "j"}], "]"}], 
  "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
     RowBox[{"a", ",", 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}]}], "]"}], 
    "\[And]", 
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
     RowBox[{"b", ",", 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}], "]"}]}], 
   
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(359)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:414980491", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633174786800476*^9, 3.633174789558387*^9}, {
  3.633174824713442*^9, 3.633174826405486*^9}},
 CellTags->{
  "ID:414980491", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "359"},
 CellID->414980491],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{
   UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", 
    UnderscriptBox["0", "\[ScriptCapitalR]"], ",", "i", ",", "j"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(345)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:744723554", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6331754787696037`*^9, 3.633175521171365*^9}},
 CellTags->{
  "ID:744723554", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "345"},
 CellID->744723554],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1560872126]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1557777283],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(354)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(354)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], 
               "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isCommonReducible", "\[ScriptCapitalR]"], "[", 
               
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], 
               "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isCommonReducible", "\[ScriptCapitalR]"], "[", 
               
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:926282876"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p926282876-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p926282876-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:926282876", "Proof|ID:926282876-1"},
 CellID->667882351],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p926282876-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->252898498]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(355)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(355)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], 
               "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], 
               "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:80931797"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p80931797-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p80931797-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:80931797", "Proof|ID:80931797-1"},
 CellID->1748139084],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p80931797-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1485509901]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(359)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(359)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], 
               "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], 
               "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:414980491"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p414980491-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p414980491-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:414980491", "Proof|ID:414980491-1"},
 CellID->336718788],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p414980491-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->280631998]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(345)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(345)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   UnderscriptBox["0", "\[ScriptCapitalR]"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   UnderscriptBox["0", "\[ScriptCapitalR]"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:744723554"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p744723554-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p744723554-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:744723554", "Proof|ID:744723554-1"},
 CellID->1023806301],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p744723554-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6427361113474293`*^9, 3.64273611300313*^9}},
 CellID->704570505]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->227618613],

Cell[CellGroupData[{

Cell["Lemma: Alternative Description of isNTCRP", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632635487839986*^9, 3.632635504499737*^9}},
 CellID->1038155297],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 381508689}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632638064374669*^9, 3.632638093643173*^9}, {
  3.635748012838746*^9, 3.635748016115306*^9}},
 CellTags->{
  "ID:381508689", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->381508689],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", "a", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1050210793}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6326380984284067`*^9, 3.632638105349184*^9}},
 CellTags->{
  "ID:1050210793", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1050210793],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 512389161}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6326381101153793`*^9, 3.632638121273929*^9}},
 CellTags->{
  "ID:512389161", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->512389161],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", "\[Equal]", "q"}], "\[Or]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
     "\[NotEqual]", 
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", "a", 
      ",", "i", ",", "j"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isNTCRP alternative 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:719916043", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632638150578105*^9, 3.632638187788336*^9}},
 CellTags->{
  "ID:719916043", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative 1"},
 CellID->719916043],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", "\[NotEqual]", "q"}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]", 
     
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", "a"}],
      "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isNTCRP alternative 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:864384559", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632638201530243*^9, 3.6326382175833807`*^9}},
 CellTags->{
  "ID:864384559", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative 2"},
 CellID->864384559],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{
        UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", 
         "a", ",", "i", ",", "j"}], "]"}], "\[DoubleLeftArrow]", 
       RowBox[{
        RowBox[{"p", "\[Equal]", "q"}], "\[Or]", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
         "\[NotEqual]", 
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}]},
      {
       RowBox[{
        UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", 
         "a"}], "]"}], "\[DoubleLeftArrow]", "otherwise"}
     }]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isNTCRP alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:951944988", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632635510613722*^9, 3.632635613727309*^9}, {
  3.632635683238867*^9, 3.632635749758772*^9}, {3.632638129346374*^9, 
  3.632638130782323*^9}},
 CellTags->{
  "ID:951944988", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative"},
 CellID->951944988],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1683542722]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1871567538],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isNTCRP alternative 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isNTCRP alternative 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:719916043"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p719916043-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p719916043-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:719916043", "Proof|ID:719916043-1"},
 CellID->1378946613],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p719916043-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->2145134637]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isNTCRP alternative 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isNTCRP alternative 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["q", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["a", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["q", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["a", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:864384559"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p864384559-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p864384559-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:864384559", "Proof|ID:864384559-1"},
 CellID->1555291612],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p864384559-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->988348688]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isNTCRP alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isNTCRP alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                  "\[DoubleLeftArrow]", 
                  RowBox[{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}]}, {
                  RowBox[{
                    UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["a", "ExpressionVariable"]}], "]"}], 
                  "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                  "\[DoubleLeftArrow]", 
                  RowBox[{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}]}, {
                  RowBox[{
                    UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["a", "ExpressionVariable"]}], "]"}], 
                  "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:951944988"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p951944988-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p951944988-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:951944988", "Proof|ID:951944988-1"},
 CellID->1069274625],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p951944988-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1430452337]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (isNTCRP alternative 1)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6326359139747877`*^9, 3.632635914863284*^9}, {
   3.632638304257482*^9, 3.6326383098028097`*^9}, 3.632639326218629*^9, 
   3.636440161175811*^9, 3.6434465988764668`*^9, 3.653208130103765*^9, 
   3.6532137001297646`*^9, 3.666173958235058*^9, 3.6664405517467117`*^9},
 CellID->2144645966],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6326364118673363`*^9, 3.63263641324196*^9}, 
   3.632639326218752*^9, 3.636440161175947*^9, 3.6434465988765907`*^9, 
   3.653208130103765*^9, 3.6532137001297646`*^9, 3.666173958235058*^9, 
   3.66644055174686*^9},
 CellID->809700523],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"p", "=", 
      RowBox[{"q", "\[Or]", 
       RowBox[{
        RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
        RowBox[{"lc", "[", "q", "]"}]}]}]}], ")"}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"areCommonReducibles", "[", 
       RowBox[{"a", ",", "a", ",", "p", ",", "q", ",", "i", ",", "j"}], "]"}],
       "\[DoubleLeftRightArrow]", 
      RowBox[{"isCommonReducible", "[", 
       RowBox[{
        RowBox[{"lc", "[", "p", "]"}], ",", 
        RowBox[{"lc", "[", "q", "]"}], ",", "a", ",", "i", ",", "j"}], 
       "]"}]}], ")"}]}], TraditionalForm]]],
 " (354)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632636416719283*^9, 3.6326364770346327`*^9}, {
   3.6326373301491632`*^9, 3.6326373301697187`*^9}, 3.632639326219211*^9, {
   3.63317439823451*^9, 3.6331744083815107`*^9}, {3.63317475319007*^9, 
   3.633174753653871*^9}, 3.636440161176059*^9, 3.64344659887667*^9, 
   3.653208130103765*^9, 3.6532137001297646`*^9, 3.666173958235058*^9, 
   3.666440551746977*^9},
 CellID->1717945330],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"areCommonReducibles", "[", 
     RowBox[{
     "a", ",", "b", ",", "p", ",", "q", ",", "c1", ",", "c2", ",", "i", ",", 
      "j"}], "]"}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"isReducible", "[", 
       RowBox[{"a", ",", 
        RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], "\[And]", 
      RowBox[{"isReducible", "[", 
       RowBox[{"b", ",", 
        RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}]}], ")"}]}], 
   TraditionalForm]]],
 " (355)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632636416719283*^9, 3.6326364770346327`*^9}, {
   3.6326373301491632`*^9, 3.632637352848503*^9}, {3.632637397493758*^9, 
   3.6326374081203327`*^9}, {3.632639009142107*^9, 3.632639031514906*^9}, 
   3.632639326219294*^9, {3.633174401593339*^9, 3.633174409043939*^9}, {
   3.6331747554917088`*^9, 3.633174755862028*^9}, 3.636440161176133*^9, 
   3.643446598876732*^9, 3.653208130103765*^9, 3.6532137001297646`*^9, 
   3.666173958235058*^9, 3.666440551747066*^9},
 CellID->639502933],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "a", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(0)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Or]", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
      RowBox[{"lc", "[", "q", "]"}]}]}]}], TraditionalForm]]],
 "\nand show the equivalence, which of course proceeds by showing both \
directions.\n\n\[OpenCurlyDoubleQuote]\[DoubleLongRightArrow]\
\[CloseCurlyDoubleQuote]:\nAssume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 " and show the RHS of the implication.\nWe expand the definitions of isNTCRP \
and isNTCR: We know\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{"a", ",", "a", ",", "p", ",", "q", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"areCommonReducibles", "[", 
         RowBox[{
         "a", ",", "a", ",", "p", ",", "q", ",", "c1", ",", "c2", ",", "i", 
          ",", "j"}], "]"}]},
       {
        RowBox[{
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c1", "*", 
             RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c2", "*", 
             RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}]}
      }]}]}], TraditionalForm]]],
 "\nand have to show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCommonReducible", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"areIrrelative", "[", 
         RowBox[{
          RowBox[{"lc", "[", "p", "]"}], ",", 
          RowBox[{"lc", "[", "q", "]"}], ",", "c1", ",", "c2", ",", "i", ",", 
          "j"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", 
          RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", 
          RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}]},
       {
        RowBox[{
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c1", "*", 
             RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c2", "*", 
             RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}]}
      }]}]}], TraditionalForm]]],
 "\n\nWe prove both conjuncts.\n(G1):\nThis follows from (0) and (1) by a \
lemma.\n(G2):\nRather than showing (G2) we assume its negation and show the \
negation of (2), i.e. we assume\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areIrrelative", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c1", ",", "c2", ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"c1", "*", 
        RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"c2", "*", 
        RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}], 
   TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], TraditionalForm]]],
 ", and prove\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"c1", ",", "c2"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"areCommonReducibles", "[", 
          RowBox[{
          "a", ",", "a", ",", "p", ",", "q", ",", "c1", ",", "c2", ",", "i", 
           ",", "j"}], "]"}]},
        {
         RowBox[{
          RowBox[{"isReducible", "[", 
           RowBox[{
            RowBox[{"a", "-", 
             RowBox[{"c1", "*", 
              RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
            RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
          RowBox[{"isReducible", "[", 
           RowBox[{
            RowBox[{"a", "-", 
             RowBox[{"c2", "*", 
              RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
            RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}]}
       }]}]}]}], TraditionalForm]]],
 "\n\nWe claim that ",
 Cell[BoxData[
  FormBox["c1", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["c2", TraditionalForm]]],
 " witness this existential goal, so we prove\n(G3.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
    "a", ",", "a", ",", "p", ",", "q", ",", "c1", ",", "c2", ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\n(G3.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"c1", "*", 
        RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"c2", "*", 
        RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}], 
   TraditionalForm]]],
 "\n\n(G3.2) follows immeadiately from (6), so we only have to consider \
(G3.1): By the definition of areCommonReducibles we have to show\n(G3.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"c1", ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"c2", ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], ")"}]}]}], 
   TraditionalForm]]],
 "\n(G3.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(G3.5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n\n(G3.4) and (G3.5) follow immedaitely from (4) and (5), resp., so we \
only have to show (G3.3). Therefore, assume\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", "q"}], TraditionalForm]]],
 "\nand show\n(G3.6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
    RowBox[{"mult", "[", 
     RowBox[{"c1", ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
    RowBox[{"mult", "[", 
     RowBox[{"c2", ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], TraditionalForm]]],
 "\n\nFrom (3), by the definition of areIrrelative, we can infer\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "p", "]"}], "=", 
    RowBox[{
     RowBox[{"lc", "[", "q", "]"}], "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"c1", ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"c2", ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], ")"}]}]}], 
   TraditionalForm]]],
 "\nmeaning that our current goal (G3.6) can be reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "p", "]"}], "=", 
    RowBox[{"lc", "[", "q", "]"}]}], TraditionalForm]]],
 "; This, however, follows immediately from ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", "q"}], TraditionalForm]]],
 ".\n\n\[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\[CloseCurlyDoubleQuote]:\
\nAssume the ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 " of the equivalence and show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 ". Expanding the definitions of isNTCR and isNTCRP leaves us with showing\n\
(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{"a", ",", "a", ",", "p", ",", "q", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"areCommonReducibles", "[", 
         RowBox[{
         "a", ",", "a", ",", "p", ",", "q", ",", "c1", ",", "c2", ",", "i", 
          ",", "j"}], "]"}]},
       {
        RowBox[{
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c1", "*", 
             RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c2", "*", 
             RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}]}
      }]}]}], TraditionalForm]]],
 "\nunder the assumptions\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCommonReducible", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"areIrrelative", "[", 
         RowBox[{
          RowBox[{"lc", "[", "p", "]"}], ",", 
          RowBox[{"lc", "[", "q", "]"}], ",", "c1", ",", "c2", ",", "i", ",", 
          "j"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", 
          RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", 
          RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}]},
       {
        RowBox[{
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c1", "*", 
             RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c2", "*", 
             RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}]}
      }]}]}], TraditionalForm]]],
 "\n\nWe prove both subgoals.\n\n(G4):\nThis follows readily from (0) and (9) \
by a lemma.\n(G5):\nRather than showing (G5) we assume its negation and show \
the negation of (10), i.e. we assume\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
    "a", ",", "a", ",", "p", ",", "q", ",", "c1", ",", "c2", ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"c1", "*", 
        RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"c2", "*", 
        RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}], 
   TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], TraditionalForm]]],
 " and show\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"c1", ",", "c2"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"areIrrelative", "[", 
          RowBox[{
           RowBox[{"lc", "[", "p", "]"}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c1", ",", "c2", ",", "i", ",",
            "j"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{"a", ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{"a", ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}]},
        {
         RowBox[{
          RowBox[{"isReducible", "[", 
           RowBox[{
            RowBox[{"a", "-", 
             RowBox[{"c1", "*", 
              RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
            RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
          RowBox[{"isReducible", "[", 
           RowBox[{
            RowBox[{"a", "-", 
             RowBox[{"c2", "*", 
              RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
            RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}]}
       }]}]}]}], TraditionalForm]]],
 "\n\nWe claim that ",
 Cell[BoxData[
  FormBox["c1", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["c2", TraditionalForm]]],
 " are witnesses for this existentially quantified goal, so we prove all of\n\
(G6.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areIrrelative", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c1", ",", "c2", ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\n(G6.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(G6.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n(G6.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"c1", "*", 
        RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"c2", "*", 
        RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}], 
   TraditionalForm]]],
 "\n\n(G6.2) and (G6.3) follow immediately from (11) by a lemma, and (G6.4) \
is precisely what (12) states. Hence, we only have to consider (G6.1); For \
this, by the definition of areIrrelative, we have to show\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
    RowBox[{"mult", "[", 
     RowBox[{"c1", ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
    RowBox[{"mult", "[", 
     RowBox[{"c2", ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], TraditionalForm]]],
 "\nunder the additional assumption\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "p", "]"}], "=", 
    RowBox[{"lc", "[", "q", "]"}]}], TraditionalForm]]],
 "\n\nBy (11), using the definition of areCommonReducibles, the goal (G7) can \
be reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", "q"}], TraditionalForm]]],
 ", so we show now\n(G8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", "q"}], TraditionalForm]]],
 "\n\nBased on (0) we distinguish two cases.\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", "q"}], TraditionalForm]]],
 ".\nWe are done immediately.\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
    RowBox[{"lc", "[", "q", "]"}]}], TraditionalForm]]],
 ".\nThis contradicts (13)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632635934757401*^9, 3.6326363995314627`*^9}, {
   3.632636491385459*^9, 3.6326367667890453`*^9}, {3.6326368909783382`*^9, 
   3.6326373251064587`*^9}, {3.632637414791287*^9, 3.632637989868054*^9}, {
   3.632638021295269*^9, 3.6326380482589903`*^9}, {3.6326383221481867`*^9, 
   3.632638543132432*^9}, {3.632638579920289*^9, 3.632638975604454*^9}, {
   3.6326390443248987`*^9, 3.6326393262193947`*^9}, 3.636440161176228*^9, {
   3.63869450437568*^9, 3.638694522185482*^9}, {3.6386948191182528`*^9, 
   3.638694821737874*^9}, 3.643446598876829*^9, 3.653208130103765*^9, 
   3.6532137001297646`*^9, 3.666173958235058*^9, 3.66644055174718*^9},
 CellID->533280863],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632639310736332*^9, 3.6326393262195263`*^9}, 
   3.636440161176351*^9, 3.643446598876945*^9, 3.653208130103765*^9, 
   3.6532137001297646`*^9, 3.666173958235058*^9, 3.6664405517473173`*^9},
 TextAlignment->Right,
 CellID->209491385]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (isNTCRP alternative 2)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6326385580311747`*^9, 3.632638566792116*^9}, 
   3.632643829274671*^9, 3.6364401671379833`*^9, 3.643446605546536*^9, 
   3.6532081222881513`*^9, 3.6532136941081543`*^9, 3.6661739454742355`*^9, 
   3.6664405446502037`*^9},
 CellID->1571042640],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632639602739669*^9, 3.632639604047324*^9}, 
   3.632643829274796*^9, 3.6364401671380997`*^9, 3.643446605546692*^9, 
   3.6532081222881513`*^9, 3.6532136941081543`*^9, 3.6661739454742355`*^9, 
   3.666440544650343*^9},
 CellID->1858320407],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"areCommonReducibles", "[", 
     RowBox[{"a", ",", "b", ",", "p", ",", "q", ",", "i", ",", "j"}], "]"}], 
    "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"isReducible", "[", 
       RowBox[{"a", ",", 
        RowBox[{"lc", "[", "p", "]"}]}], "]"}], "\[And]", 
      RowBox[{"isReducible", "[", 
       RowBox[{"b", ",", 
        RowBox[{"lc", "[", "q", "]"}]}], "]"}]}], ")"}]}], 
   TraditionalForm]]],
 " (359)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632639612559613*^9, 3.632639639247047*^9}, 
   3.6326438292752438`*^9, {3.633174798823559*^9, 3.633174803589217*^9}, {
   3.6331748606897783`*^9, 3.63317486253496*^9}, 3.6364401671382113`*^9, 
   3.643446605546795*^9, 3.6532081222881513`*^9, 3.6532136941237545`*^9, 
   3.6661739454742355`*^9, 3.666440544650455*^9},
 CellID->977808820],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "a", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(0)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "q"}], TraditionalForm]]],
 "\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "p", "]"}], "=", 
    RowBox[{"lc", "[", "q", "]"}]}], TraditionalForm]]],
 "\nand show the equivalence, which of course proceeds by showing both \
directions.\n\n\[OpenCurlyDoubleQuote]\[DoubleLongRightArrow]\
\[CloseCurlyDoubleQuote]:\nWe expand the definitions of isNTCRP and isNTCR: \
We know\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{"a", ",", "a", ",", "p", ",", "q", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"areCommonReducibles", "[", 
         RowBox[{
         "a", ",", "a", ",", "p", ",", "q", ",", "c1", ",", "c2", ",", "i", 
          ",", "j"}], "]"}]},
       {
        RowBox[{
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c1", "*", 
             RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c2", "*", 
             RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}]}
      }]}]}], TraditionalForm]]],
 "\nand have to show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}]}], "]"}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", 
          RowBox[{"lc", "[", "q", "]"}], ",", "c1"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", 
          RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"a", "-", 
           RowBox[{"c1", "*", 
            RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
          RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\n\nWe prove both subgoals.\n(G1):\nThis follows from (2) by a lemma.\n\
(G2):\nRather than showing (G2) we assume its negation and show the negation \
of (3), i.e. we assume\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"c1", "*", 
       RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], TraditionalForm]]],
 ", and prove\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"c1", ",", "c2"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"areCommonReducibles", "[", 
          RowBox[{
          "a", ",", "a", ",", "p", ",", "q", ",", "c1", ",", "c2", ",", "i", 
           ",", "j"}], "]"}]},
        {
         RowBox[{
          RowBox[{"isReducible", "[", 
           RowBox[{
            RowBox[{"a", "-", 
             RowBox[{"c1", "*", 
              RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
            RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
          RowBox[{"isReducible", "[", 
           RowBox[{
            RowBox[{"a", "-", 
             RowBox[{"c2", "*", 
              RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
            RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}]}
       }]}]}]}], TraditionalForm]]],
 "\n\nWe claim that ",
 Cell[BoxData[
  FormBox["c1", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["c2", TraditionalForm]]],
 " witness this existential goal, so we prove\n(G3.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
    "a", ",", "a", ",", "p", ",", "q", ",", "c1", ",", "c2", ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\n(G3.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"c1", "*", 
        RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"c2", "*", 
        RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}], 
   TraditionalForm]]],
 "\n\n(G3.2) follows immeadiately from (6) and (1), so we only have to \
consider (G3.1): By the definition of areCommonReducibles we have to show\n\
(G3.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"c1", ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"c2", ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], ")"}]}]}], 
   TraditionalForm]]],
 "\n(G3.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(G3.5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n\n(G3.3) holds vacuously, since ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", "q"}], TraditionalForm]]],
 " contradicts (0), and (G3.4) and (G3.5) follow immediately from (4) and (5) \
(again using (1)).\n\n\[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\
\[CloseCurlyDoubleQuote]:\nWe expand the definitions of isNTCRP and isNTCR: \
We know\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}]}], "]"}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", 
          RowBox[{"lc", "[", "q", "]"}], ",", "c1"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", 
          RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"a", "-", 
           RowBox[{"c1", "*", 
            RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
          RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\nand have to show\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{"a", ",", "a", ",", "p", ",", "q", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"areCommonReducibles", "[", 
         RowBox[{
         "a", ",", "a", ",", "p", ",", "q", ",", "c1", ",", "c2", ",", "i", 
          ",", "j"}], "]"}]},
       {
        RowBox[{
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c1", "*", 
             RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c2", "*", 
             RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}]}
      }]}]}], TraditionalForm]]],
 "\n\nWe prove both subgoals.\n(G4):\nBy the definition of ",
 Cell[BoxData[
  FormBox["areCommonReducibles", TraditionalForm]]],
 " we have to find ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], TraditionalForm]]],
 " s.t.\n(G4.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"c1", ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"c2", ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], ")"}]}]}], 
   TraditionalForm]]],
 "\n(G4.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(G4.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n(7) and the definition of isReducible (binary) provide us with suitable \
",
 Cell[BoxData[
  FormBox["c1", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["c2", TraditionalForm]]],
 ", as can easily be verified (note that (G4.1) holds vacuously).\n(G5):\n\
Rather than showing (G5) we assume its negation and show the negation of (8), \
i.e. we assume\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
    "a", ",", "a", ",", "p", ",", "q", ",", "c1", ",", "c2", ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"c1", "*", 
        RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"c2", "*", 
        RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}], 
   TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], TraditionalForm]]],
 " and show\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"c1", ",", "c2"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"isReducible", "[", 
          RowBox[{"a", ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c1"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{"a", ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"c1", "*", 
             RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\n\nBased on (10) we have to distinguish two cases.\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"c1", "*", 
       RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 ".\nWe claim that ",
 Cell[BoxData[
  FormBox["c1", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["c2", TraditionalForm]]],
 " witness this existential goal, so we prove\n(G6.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(G6.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n(G6.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"c1", "*", 
       RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n\n(G6.1) and (G6.2) follow immeadiately from (9) and (1) by a lemma, and \
(G6.3) follows from our case assumption and (1).\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"c2", "*", 
       RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 ".\nWe claim that ",
 Cell[BoxData[
  FormBox["c2", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["c1", TraditionalForm]]],
 " witness this existential goal, so we prove\n(G6.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n(G6.5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(G6.6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"c2", "*", 
       RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n\n(G6.4) and (G6.5) follow immeadiately from (9) and (1) by a lemma, and \
(G6.6) follows from our case assumption and (1)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{
  3.632638549374304*^9, {3.63263935655932*^9, 3.6326395920232153`*^9}, {
   3.632639644543824*^9, 3.632639927026825*^9}, {3.632639975453281*^9, 
   3.63264050690166*^9}, {3.6326436525198603`*^9, 3.6326438292753468`*^9}, 
   3.636440167138307*^9, {3.638696819640123*^9, 3.638696824172065*^9}, 
   3.643446605546932*^9, 3.6532081222881513`*^9, 3.6532136941237545`*^9, 
   3.6661739454742355`*^9, 3.666440544650569*^9},
 CellID->961202380],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6326438150113*^9, 3.63264382927547*^9}, 
   3.636440167138422*^9, 3.643446605547084*^9, 3.6532081222881513`*^9, 
   3.6532136941237545`*^9, 3.6661739454742355`*^9, 3.666440544650702*^9},
 TextAlignment->Right,
 CellID->277731394]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->552264247],

Cell[CellGroupData[{

Cell["\<\
Theorem: Directions \[Implies] and \[DoubleLeftArrow] in (CR1)\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632216500031163*^9, 3.632216508485528*^9}, {
  3.632216612281062*^9, 3.632216620303114*^9}, {3.632216847527401*^9, 
  3.632216874297433*^9}, {3.632649073327629*^9, 3.632649074028349*^9}, {
  3.632649140695902*^9, 3.632649162349945*^9}, {3.632649803874756*^9, 
  3.632649805183988*^9}, {3.632721777933611*^9, 3.632721778137157*^9}, {
  3.633235842332362*^9, 3.633235843263465*^9}},
 CellID->226520904],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "a"}], "]"}]}]], 
  RowBox[{
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1990838743}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632216516380025*^9, 3.632216539154045*^9}, {
  3.632216635956822*^9, 3.632216636110222*^9}, {3.632216689660941*^9, 
  3.632216697227892*^9}, {3.632568909496789*^9, 3.632568914750527*^9}, {
  3.635748027877426*^9, 3.635748031118073*^9}},
 CellTags->{
  "ID:1990838743", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1990838743],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 20910152}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632216641348508*^9, 3.632216654281588*^9}, {
  3.632482132603677*^9, 3.632482134635543*^9}, {3.632553027615705*^9, 
  3.632553027799371*^9}},
 CellTags->{
  "ID:20910152", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->20910152],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"p", "\[Equal]", "q"}], "\[Implies]", 
      RowBox[{"i", "\[NotEqual]", "j"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[Exists]", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
          RowBox[{"[", 
           RowBox[{"t", ",", "s"}], "]"}]}]], 
        RowBox[{"t", "\[NotEqual]", "s"}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"\[And]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {
             RowBox[{
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
              UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]},
            {
             RowBox[{
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}], 
              UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "s"}]},
            {
             RowBox[{
              UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], "[", 
              RowBox[{
               RowBox[{
                UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                RowBox[{"a", ",", "t"}], "]"}], ",", 
               RowBox[{
                UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                RowBox[{"a", ",", "s"}], "]"}], ",", "p", ",", "q", ",", "i", 
               ",", "j"}], "]"}]}
           },
           GridBoxAlignment->{
            "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR1.\[DoubleRightArrow].1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:416591219", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632216553606091*^9, 3.632216583373764*^9}, {
   3.632216667857603*^9, 3.632216669608225*^9}, {3.63222004727522*^9, 
   3.632220076878083*^9}, {3.63222012353723*^9, 3.632220136568811*^9}, {
   3.6325532874170237`*^9, 3.632553298834502*^9}, {3.632557288314011*^9, 
   3.632557289240704*^9}, {3.632557972123543*^9, 3.632557972674458*^9}, {
   3.632567613614031*^9, 3.63256764560711*^9}, {3.63256773494805*^9, 
   3.6325677477353163`*^9}, {3.6325689749511757`*^9, 3.632568979845789*^9}, {
   3.632571043241585*^9, 3.632571043809762*^9}, 3.6325712390510883`*^9, {
   3.632571405945664*^9, 3.6325714114944363`*^9}, 3.638700171252268*^9},
 CellTags->{
  "ID:416591219", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleRightArrow].1"},
 CellID->416591219],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]},
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]},
         {
          RowBox[{
           UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
           RowBox[{"p", ",", "q", ",", 
            RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{"a", ",", "t"}], "]"}], ",", "i", ",", "j"}], "]"}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR1.\[DoubleRightArrow].2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:295118406", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632557955415401*^9, 3.632557968874349*^9}, {
   3.6326342542619658`*^9, 3.6326342951089087`*^9}, {3.632643849383205*^9, 
   3.632643868354433*^9}, 3.638700209598879*^9},
 CellTags->{
  "ID:295118406", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleRightArrow].2"},
 CellID->295118406],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[Exists]", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
          RowBox[{"[", 
           RowBox[{"t", ",", "s"}], "]"}]}]], 
        RowBox[{"t", "\[NotEqual]", "s"}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"\[And]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {
             RowBox[{
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
              UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]},
            {
             RowBox[{
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}], 
              UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "s"}]},
            {
             RowBox[{
              UnderscriptBox["areCommonReducibles", "\[ScriptCapitalP]"], "[", 
              RowBox[{
               RowBox[{
                UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                RowBox[{"a", ",", "t"}], "]"}], ",", 
               RowBox[{
                UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                RowBox[{"a", ",", "s"}], "]"}], ",", "p", ",", "q", ",", "i", 
               ",", "j"}], "]"}]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "t", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]},
           {
            RowBox[{
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]},
           {
            RowBox[{
             UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
             RowBox[{"p", ",", "q", ",", 
              RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{"a", ",", "t"}], "]"}], ",", "i", ",", "j"}], "]"}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR1.\[DoubleLeftArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1583101849", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632557278706704*^9, 3.632557286099596*^9}, 
   3.632567532108927*^9, {3.6325676600469847`*^9, 3.6325676624062777`*^9}, {
   3.632571245794146*^9, 3.63257124745492*^9}, {3.632571413854823*^9, 
   3.632571415042164*^9}, {3.632648921218342*^9, 3.6326490074556227`*^9}, 
   3.632649119226839*^9, {3.638700212828905*^9, 3.638700216849313*^9}, {
   3.6387753236849957`*^9, 3.638775325430871*^9}},
 CellTags->{
  "ID:1583101849", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleLeftArrow]"},
 CellID->1583101849],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1370471321]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->479379292],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR1.\[DoubleRightArrow].1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR1.\[DoubleRightArrow].1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["j", "ExpressionVariable"]}]}], "\[And]", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["s", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox[
                    "areCommonReducibles", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], ",", 
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["j", "ExpressionVariable"]}]}], "\[And]", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["s", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox[
                    "areCommonReducibles", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], ",", 
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:416591219"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p416591219-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p416591219-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:416591219", "Proof|ID:416591219-1"},
 CellID->1006102583],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p416591219-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->1404893208]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR1.\[DoubleRightArrow].2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR1.\[DoubleRightArrow].2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:295118406"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p295118406-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p295118406-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:295118406", "Proof|ID:295118406-1"},
 CellID->1475020847],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p295118406-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {68., 75.}}]], "ProofInfoBody",
 CellID->1001169367]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR1.\[DoubleLeftArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR1.\[DoubleLeftArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["s", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox[
                    "areCommonReducibles", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], ",", 
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["s", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox[
                    "areCommonReducibles", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], ",", 
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1583101849"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1583101849-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1583101849-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1583101849", "Proof|ID:1583101849-1"},
 CellID->2049393323],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1583101849-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {68., 75.}}]], "ProofInfoBody",
 CellID->1721289092]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR1.\[Implies].2)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632569164769092*^9, 3.632569176756404*^9}, 
   3.632570094976581*^9, {3.632634231493162*^9, 3.63263423943274*^9}, {
   3.632634660068185*^9, 3.632634662057506*^9}, 3.636440182689971*^9, 
   3.6387875595870857`*^9, 3.638787999502549*^9, 3.643446613706484*^9, 
   3.653208113738336*^9, 3.6532136869945416`*^9, 3.66617393678502*^9, 
   3.666440536274171*^9},
 CellID->534711775],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "a"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "a"}], "]"}], TraditionalForm]]],
 " and show the RHS of the implication. From (CR2.\[Implies].1) we can infer\n\
(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", 
         RowBox[{"s", ",", "t"}], "]"}]}]], 
      RowBox[{"s", "\[NotEqual]", "t"}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "s"}]},
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "t"}], "]"}], ",", 
          RowBox[{"lc", "[", "p", "]"}]}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "s"}], "]"}], ",", 
          RowBox[{"lc", "[", "p", "]"}]}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\nWe expand the definition of isNTCR: We know\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p"}], "]"}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", 
        RowBox[{"m1", ",", "m2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", "p", ",", "m1"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", "p", ",", "m2"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"a", "-", 
           RowBox[{"m2", "*", "p"}]}], ",", "p", ",", "m1"}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\nWe expand the definition of ",
 Cell[BoxData[
  FormBox["isReducible", TraditionalForm]]],
 ": We know that there exists ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "m3", "]"}]}], TraditionalForm]]],
 " with\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m3"}], "]"}], TraditionalForm]]],
 "\n\nFrom (4), together with (RED II), we can infer (among others)\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m3", "]"}]}], "]"}], TraditionalForm]]],
 "\n\nUntil here it is the proof of (CR2.\[Implies].2).\n------\nFrom here it \
is the right proof.\n\n",
 StyleBox["We show now ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m3", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "m4", "]"}], "*", 
     RowBox[{"lp", "[", "q", "]"}]}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[". For this, we exchange the goal ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m3", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "m4", "]"}], "*", 
     RowBox[{"lp", "[", "q", "]"}]}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" with assumption (2), i.e. we assume ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m3", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "\[NotEqual]", 
    RowBox[{
     RowBox[{"lp", "[", "m4", "]"}], "*", 
     RowBox[{"lp", "[", "q", "]"}]}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and show the existence of ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"s", ",", "t"}], "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" with ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[NotEqual]", "s"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}], TraditionalForm]],
  
  FontColor->GrayLevel[0]],
 StyleBox[", ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "s"}], TraditionalForm]],
  
  FontColor->GrayLevel[0]],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "s"}], "]"}], ",", "p", ",", "q", ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[". We claim that ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m3", "]"}], "*", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m4", "]"}], "*", 
    RowBox[{"lp", "[", "q", "]"}]}], TraditionalForm]],
  FontColor->GrayLevel[0]],
 StyleBox[" witness this existential goal. The first three parts are obvious, \
and for the last we have to show\n(C1)\t",
  FontColor->GrayLevel[0]],
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"c1", ",", "c2"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"p", "=", 
          RowBox[{"q", "\[Implies]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"mult", "[", 
              RowBox[{"c1", ",", 
               RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
             RowBox[{"mult", "[", 
              RowBox[{"c2", ",", 
               RowBox[{"lc", "[", "q", "]"}], ",", "j"}]}]}], ")"}]}]}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", 
             RowBox[{
              RowBox[{"lp", "[", "m3", "]"}], "*", 
              RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", 
             RowBox[{
              RowBox[{"lp", "[", "m4", "]"}], "*", 
              RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c1", ":=", 
     RowBox[{"lc", "[", "m3", "]"}]}], ",", 
    RowBox[{"c2", ":=", 
     RowBox[{"lc", "[", "m4", "]"}]}]}], TraditionalForm]]],
 " witness our existential subgoal, hence we need to verify\n(C2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{"lc", "[", "m3", "]"}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{"lc", "[", "m4", "]"}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "j"}]}]}], ")"}]}]}], 
   TraditionalForm]]],
 "\n(C3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m3", "]"}]}], "]"}], TraditionalForm]]],
 "\n(C4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m4", "]"}], "*", 
        RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", 
     RowBox[{"lc", "[", "m4", "]"}]}], "]"}], TraditionalForm]]],
 "\n(C2) follows from (11), (C3) follows from (8) and (C4) follows from (9). ",
 StyleBox["Hence, we may add",
  FontColor->GrayLevel[0]],
 "\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m3", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "m4", "]"}], "*", 
     RowBox[{"lp", "[", "q", "]"}]}]}], TraditionalForm]]],
 "\nto the list of assumptions.\n\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m3", "]"}], "*", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 " (= ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m4", "]"}], "*", 
    RowBox[{"lp", "[", "q", "]"}]}], TraditionalForm]]],
 ") witnesses our existential goal, so we need to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{
     RowBox[{"lp", "[", "m3", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}]}], TraditionalForm]]],
 " (obvious), ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", 
    RowBox[{
     RowBox[{"lp", "[", "m3", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}]}], TraditionalForm]]],
 " (obvious; use ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m3", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "m4", "]"}], "*", 
     RowBox[{"lp", "[", "q", "]"}]}]}], TraditionalForm]]],
 ") and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", "i", ",", "j"}], "]"}],
    TraditionalForm]]],
 ". By the definition of isNTCRP, we have to show\n\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", "p", ",", "q", ",", 
     "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"areCommonReducibles", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", 
            RowBox[{
             RowBox[{"lp", "[", "m3", "]"}], "*", 
             RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
          RowBox[{"C", "[", 
           RowBox[{"a", ",", 
            RowBox[{
             RowBox[{"lp", "[", "m3", "]"}], "*", 
             RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", "p", ",", "q", 
          ",", "c1", ",", "c2", ",", "i", ",", "j"}], "]"}]},
       {
        RowBox[{
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"a", ",", 
              RowBox[{
               RowBox[{"lp", "[", "m3", "]"}], "*", 
               RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
            RowBox[{"c1", "*", 
             RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"a", ",", 
              RowBox[{
               RowBox[{"lp", "[", "m3", "]"}], "*", 
               RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
            RowBox[{"c2", "*", 
             RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}]}
      }]}]}], TraditionalForm]]],
 "\n\nWe prove both subgoals.\n\n(G1):\nWe have to prove\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"c3", ",", "c4"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"p", "=", 
          RowBox[{"q", "\[Implies]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
             RowBox[{"mult", "[", 
              RowBox[{"c3", ",", 
               RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
             RowBox[{"mult", "[", 
              RowBox[{"c4", ",", 
               RowBox[{"lc", "[", "q", "]"}], ",", "j"}]}]}], ")"}]}]}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", 
             RowBox[{
              RowBox[{"lp", "[", "m3", "]"}], "*", 
              RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c3"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", 
             RowBox[{
              RowBox[{"lp", "[", "m3", "]"}], "*", 
              RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c4"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c3", ":=", 
     RowBox[{"lc", "[", "m3", "]"}]}], ",", 
    RowBox[{"c4", ":=", 
     RowBox[{"lc", "[", "m4", "]"}]}]}], TraditionalForm]]],
 " are suitable witnesses, hence we show all of\n(G3.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{"lc", "[", "m3", "]"}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{"lc", "[", "m4", "]"}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "j"}]}]}], ")"}]}]}], 
   TraditionalForm]]],
 "\n(G3.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m3", "]"}]}], "]"}], TraditionalForm]]],
 "\n(G3.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", 
     RowBox[{"lc", "[", "m4", "]"}]}], "]"}], TraditionalForm]]],
 "\n\n(G3.1) is exactly what (11) states, (G3.2) is what (8) states, and \
(G3.3) follows readily from (9), using (12).\n\n(G2):\nRather than proving \
(G2), we assume its negation and show the negation of (4), i.e. we assume\n\
(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", "p", ",", "q", ",", 
     "c1", ",", "c2", ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", 
         RowBox[{
          RowBox[{"lp", "[", "m3", "]"}], "*", 
          RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
       RowBox[{"c1", "*", 
        RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", 
         RowBox[{
          RowBox[{"lp", "[", "m3", "]"}], "*", 
          RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
       RowBox[{"c2", "*", 
        RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}], 
   TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], TraditionalForm]]],
 ", and prove\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"m1", ",", "m2"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"areIrrelative", "[", 
          RowBox[{"p", ",", "q", ",", "m1", ",", "m2", ",", "i", ",", "j"}], 
          "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{"a", ",", "p", ",", "m1"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{"a", ",", "q", ",", "m2"}], "]"}]},
        {
         RowBox[{
          RowBox[{"isReducible", "[", 
           RowBox[{
            RowBox[{"a", "-", 
             RowBox[{"m1", "*", "p"}]}], ",", "q", ",", "m2"}], "]"}], 
          "\[Or]", 
          RowBox[{"isReducible", "[", 
           RowBox[{
            RowBox[{"a", "-", 
             RowBox[{"m2", "*", "q"}]}], ",", "p", ",", "m1"}], "]"}]}]}
       }]}]}]}], TraditionalForm]]],
 "\n\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m1", ":=", 
     RowBox[{"Monomial", "[", 
      RowBox[{"c1", ",", 
       RowBox[{"lp", "[", "m3", "]"}]}], "]"}]}], ",", 
    RowBox[{"m2", ":=", 
     RowBox[{"Monomial", "[", 
      RowBox[{"c2", ",", 
       RowBox[{"lp", "[", "m4", "]"}]}], "]"}]}]}], TraditionalForm]]],
 " witness our existential goal (G4), so we have to show all of\n(G4.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areIrrelative", "[", 
    RowBox[{"p", ",", "q", ",", "m1", ",", "m2", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(G4.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m1"}], "]"}], TraditionalForm]]],
 "\n(G4.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "q", ",", "m2"}], "]"}], TraditionalForm]]],
 "\n(G4.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"m1", "*", "p"}]}], ",", "q", ",", "m2"}], "]"}], "\[Or]", 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"m2", "*", "q"}]}], ",", "p", ",", "m1"}], "]"}]}], 
   TraditionalForm]]],
 "\n\nBefore, however, we expand the definition of areCommonReducibles to \
infer from (13)\n(15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"c1", ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"c2", ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "j"}]}]}], ")"}]}]}], 
   TraditionalForm]]],
 "\n(16)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(17)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n\nWe also convince ourselves that\n(18)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isDefined", "[", 
    RowBox[{"lp", "[", "m1", "]"}], "]"}], TraditionalForm]]],
 "\n(19)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isDefined", "[", 
    RowBox[{"lp", "[", "m2", "]"}], "]"}], TraditionalForm]]],
 "\nand also\n(20)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m1", ",", "p"}], "]"}], TraditionalForm]]],
 "\n(21)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m2", ",", "q"}], "]"}], TraditionalForm]]],
 "\n(all can be shown rather easily, using ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c3", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c4", "\[NotEqual]", "0"}], TraditionalForm]]],
 ").\n\n(G4.1):\nShowing (G4.1) amounts to showing ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
    RowBox[{"mult", "[", 
     RowBox[{"m1", ",", "q", ",", "i"}], "]"}], "\[And]", 
    RowBox[{"mult", "[", 
     RowBox[{"m2", ",", "q", ",", "j"}], "]"}]}], TraditionalForm]]],
 " under the condition ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", "q"}], TraditionalForm]]],
 ". This, however, follows readily from (15) and the definition of mult in \
\[ScriptCapitalP].\n\n(G4.2):\nBy (RED III) this can be reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isDefined", "[", 
    RowBox[{"lp", "[", "p", "]"}], "]"}], TraditionalForm]]],
 " (known), ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isDefined", "[", 
    RowBox[{"lp", "[", "m1", "]"}], "]"}], TraditionalForm]]],
 " (see (18)) and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m1", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m1", "]"}]}], "]"}], TraditionalForm]]],
 ". The last part follows from (16), using the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m1", "]"}], "=", 
    RowBox[{"lp", "[", "m3", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m1", "]"}], "=", "c1"}], TraditionalForm]]],
 ".\n\n(G4.3):\nBy (RED III) this can be reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isDefined", "[", 
    RowBox[{"lp", "[", "q", "]"}], "]"}], TraditionalForm]]],
 " (known), ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isDefined", "[", 
    RowBox[{"lp", "[", "m2", "]"}], "]"}], TraditionalForm]]],
 " (see (19)) and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m2", "]"}], "*", 
        RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", 
     RowBox[{"lc", "[", "m2", "]"}]}], "]"}], TraditionalForm]]],
 ". The last part follows from (17), using the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m2", "]"}], "=", 
    RowBox[{"lp", "[", "m4", "]"}]}], TraditionalForm]]],
 ",  ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m4", "]"}], "*", 
     RowBox[{"lp", "[", "q", "]"}]}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "m3", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m2", "]"}], "=", "c2"}], TraditionalForm]]],
 ".\n\n(G4.4):\nIn order to prove (G4.4) we distinguish two cases, based on \
our assumption (14).\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", 
        RowBox[{
         RowBox[{"lp", "[", "m4", "]"}], "*", 
         RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], "-", 
      RowBox[{"c1", "*", 
       RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 ".\nWe show the first disjunct in (G4.4), which, by (RED II), amounts to \
showing ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " (known), ",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m2", ",", "q"}], "]"}], TraditionalForm]]],
 " (known) and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m1", "*", "p"}]}], ",", 
       RowBox[{
        RowBox[{"lp", "[", "m2", "]"}], "*", 
        RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", 
     RowBox[{"lc", "[", "m2", "]"}]}], "]"}], TraditionalForm]]],
 ". Using the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m2", "]"}], "=", 
    RowBox[{"lp", "[", "m4", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m3", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "m4", "]"}], "*", 
     RowBox[{"lp", "[", "q", "]"}]}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m2", "]"}], "=", "c2"}], TraditionalForm]]],
 ", we can rewrite our goal into ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m1", "*", "p"}]}], ",", 
       RowBox[{
        RowBox[{"lp", "[", "m4", "]"}], "*", 
        RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 ", thus, if we manage to prove\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"m1", "*", "p"}]}], ",", 
      RowBox[{
       RowBox[{"lp", "[", "m4", "]"}], "*", 
       RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m4", "]"}], "*", 
        RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], "-", 
     RowBox[{"c1", "*", 
      RowBox[{"lc", "[", "p", "]"}]}]}]}], TraditionalForm]]],
 "\nwe are done due to our case assumption.\nUsing linearity of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " (G5) can be rewritten into ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m4", "]"}], "*", 
        RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], "-", 
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"m1", "*", "p"}], ",", 
       RowBox[{
        RowBox[{"lp", "[", "m4", "]"}], "*", 
        RowBox[{"lp", "[", "q", "]"}]}]}], "]"}]}], "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m4", "]"}], "*", 
        RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], "-", 
     RowBox[{"c1", "*", 
      RowBox[{"lc", "[", "p", "]"}]}]}]}], TraditionalForm]]],
 ", and cancelling ",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "[", 
    RowBox[{"a", ",", 
     RowBox[{
      RowBox[{"lp", "[", "m4", "]"}], "*", 
      RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], TraditionalForm]]],
 " on both sides leaves us with showing\n(G5\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"m1", "*", "p"}], ",", 
      RowBox[{
       RowBox[{"lp", "[", "m4", "]"}], "*", 
       RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], "=", 
    RowBox[{"c1", "*", 
     RowBox[{"lc", "[", "p", "]"}]}]}], TraditionalForm]]],
 "\n(G5\[CloseCurlyQuote]) holds because of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m4", "]"}], "*", 
     RowBox[{"lp", "[", "q", "]"}]}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "m3", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m1", "]"}], "=", 
    RowBox[{"lp", "[", "m3", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m1", "]"}], "=", "c1"}], TraditionalForm]]],
 ".\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", 
        RowBox[{
         RowBox[{"lp", "[", "m3", "]"}], "*", 
         RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
      RowBox[{"c2", "*", 
       RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 ".\nWe show the second disjunct in (G4.4), which, by (RED III), amounts to \
showing ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isDefined", "[", 
    RowBox[{"lp", "[", "p", "]"}], "]"}], TraditionalForm]]],
 " (known), ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isDefined", "[", 
    RowBox[{"lp", "[", "m1", "]"}], "]"}], TraditionalForm]]],
 " (known) and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m2", "*", "q"}]}], ",", 
       RowBox[{
        RowBox[{"lp", "[", "m1", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m1", "]"}]}], "]"}], TraditionalForm]]],
 ". Using the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m1", "]"}], "=", 
    RowBox[{"lp", "[", "m3", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m1", "]"}], "=", "c1"}], TraditionalForm]]],
 ", we can rewrite our goal into ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m2", "*", "q"}]}], ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 ", thus, if we manage to prove\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"m2", "*", "q"}]}], ",", 
      RowBox[{
       RowBox[{"lp", "[", "m3", "]"}], "*", 
       RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
     RowBox[{"c2", "*", 
      RowBox[{"lc", "[", "q", "]"}]}]}]}], TraditionalForm]]],
 "\nwe are done due to our case assumption.\nUsing linearity of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " (G6) can be rewritten into ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"m2", "*", "q"}], ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
     RowBox[{"c2", "*", 
      RowBox[{"lc", "[", "q", "]"}]}]}]}], TraditionalForm]]],
 ", and cancelling ",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "[", 
    RowBox[{"a", ",", 
     RowBox[{
      RowBox[{"lp", "[", "m3", "]"}], "*", 
      RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], TraditionalForm]]],
 " on both sides leaves us with showing\n(G6\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"m2", "*", "q"}], ",", 
      RowBox[{
       RowBox[{"lp", "[", "m3", "]"}], "*", 
       RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "=", 
    RowBox[{"c2", "*", 
     RowBox[{"lc", "[", "q", "]"}]}]}], TraditionalForm]]],
 "\n(G6\[CloseCurlyQuote]) holds because of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m2", "]"}], "=", 
    RowBox[{"lp", "[", "m4", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m3", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "m4", "]"}], "*", 
     RowBox[{"lp", "[", "q", "]"}]}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m2", "]"}], "=", "c2"}], TraditionalForm]]],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6325692002780457`*^9, 3.6325694268634853`*^9}, 
   3.632570094976705*^9, {3.632570728372566*^9, 3.6325707578167763`*^9}, {
   3.632571432720799*^9, 3.632571445560379*^9}, {3.6325715938423557`*^9, 
   3.632571606722777*^9}, {3.6326306952628*^9, 3.6326307224971943`*^9}, {
   3.632630833528577*^9, 3.632630864121511*^9}, {3.632631023292745*^9, 
   3.6326310719796886`*^9}, {3.632631219279278*^9, 3.632631628444659*^9}, {
   3.6326316661808567`*^9, 3.6326317872650623`*^9}, {3.632631824382112*^9, 
   3.6326321995251713`*^9}, {3.632632264017104*^9, 3.6326323652336073`*^9}, {
   3.632632698487405*^9, 3.6326327720536613`*^9}, {3.632632823633246*^9, 
   3.6326328288054247`*^9}, {3.63263290792702*^9, 3.632633024732746*^9}, {
   3.632633067837179*^9, 3.632633099450726*^9}, {3.632633198475442*^9, 
   3.632633216851014*^9}, {3.632633354879549*^9, 3.632633389958372*^9}, {
   3.632633495972414*^9, 3.632633646873837*^9}, {3.632633682044121*^9, 
   3.632633728364417*^9}, {3.632633780943246*^9, 3.632633824062347*^9}, {
   3.632633876924165*^9, 3.632633993877343*^9}, {3.632634030584486*^9, 
   3.632634062690051*^9}, {3.632634099874289*^9, 3.632634105280949*^9}, {
   3.632634143753538*^9, 3.632634150648302*^9}, {3.6326341869835377`*^9, 
   3.6326342394328613`*^9}, {3.6326439738739147`*^9, 
   3.6326444422750397`*^9}, {3.632644475376178*^9, 3.632644879520465*^9}, {
   3.6326449162789507`*^9, 3.6326453385779333`*^9}, {3.632645392664752*^9, 
   3.6326457863326693`*^9}, {3.632645844014824*^9, 3.632646142967249*^9}, 
   3.636440182690127*^9, {3.638765259422262*^9, 3.638765264136723*^9}, 
   3.638765393861206*^9, {3.638770145374399*^9, 3.638770197363942*^9}, {
   3.638770234159388*^9, 3.63877030718576*^9}, {3.638770340000877*^9, 
   3.638770392275969*^9}, {3.638787568529624*^9, 3.638787669778882*^9}, {
   3.6387879048399553`*^9, 3.63878795363321*^9}, {3.638788010355644*^9, 
   3.638788049932659*^9}, 3.6434466137066402`*^9, 3.653208113738336*^9, 
   3.6532136869945416`*^9, 3.66617393678502*^9, 3.666440536274313*^9},
 CellID->1006778759],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.63263421142743*^9, 3.632634239432981*^9}, 
   3.636440182690287*^9, 3.6434466137067947`*^9, 3.653208113738336*^9, 
   3.6532136869945416`*^9, 3.66617393678502*^9, 3.6664405362744503`*^9},
 TextAlignment->Right,
 CellID->283251415]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of",
  FontSlant->"Italic"],
 " ",
 StyleBox["(CR1.\[DoubleLeftArrow])",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632646174827505*^9, 3.632646190352857*^9}, 
   3.6326488920697613`*^9, 3.636440189701724*^9, 3.643446619816482*^9, 
   3.6532081055307217`*^9, 3.6532136815969324`*^9, 3.666173929562207*^9, 
   3.6664405290420713`*^9},
 CellID->201585227],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "a"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", 
         RowBox[{"t", ",", "s"}], "]"}]}]], 
      RowBox[{"t", "\[NotEqual]", "s"}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {
        RowBox[{
         RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "s"}]},
       {
        RowBox[{"areCommonReducibles", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "t"}], "]"}], ",", 
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "s"}], "]"}], ",", "p", ",", "q", ",", "i", ",", 
          "j"}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "t"}], TraditionalForm]]],
 "\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 ". We have to show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 ".\n\nFrom (4), by the definition of isNTCRP, we can infer\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", "p", ",", "q", ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"areCommonReducibles", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "t"}], "]"}], ",", 
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "t"}], "]"}], ",", "p", ",", "q", ",", "c1", ",",
           "c2", ",", "i", ",", "j"}], "]"}]},
       {
        RowBox[{
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"a", ",", "t"}], "]"}], "-", 
            RowBox[{"c1", "*", 
             RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], "\[Or]", 
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"a", ",", "t"}], "]"}], "-", 
            RowBox[{"c2", "*", 
             RowBox[{"lc", "[", "q", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}]}
      }]}]}], TraditionalForm]]],
 "\n\nFrom (5), by the definition of areCommonReducibles, we can infer\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"c1", ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"c2", ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], ")"}]}]}], 
   TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], TraditionalForm]]],
 ". Clearly, because of (8) and (9), we know\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c1", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(10\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"c1", ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 "\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c2", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(11\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"c2", ",", 
     RowBox[{"lc", "[", "q", "]"}]}], "]"}], TraditionalForm]]],
 "\n\nBy the definition of isNTCR we have to show both\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCommonReducible", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", 
        RowBox[{"m1", ",", "m2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"areIrrelative", "[", 
         RowBox[{"p", ",", "q", ",", "m1", ",", "m2", ",", "i", ",", "j"}], 
         "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", "p", ",", "m1"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", "q", ",", "m2"}], "]"}]},
       {
        RowBox[{
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"m1", "*", "p"}]}], ",", "q", ",", "m2"}], "]"}], "\[Or]", 
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"m2", "*", "q"}]}], ",", "p", ",", "m1"}], "]"}]}]}
      }]}]}], TraditionalForm]]],
 "\n\n(G1):\nFor proving ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCommonReducible", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 " we have to find ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2"}], "]"}]}], TraditionalForm]]],
 " s.t.\n(G1.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areIrrelative", "[", 
    RowBox[{"p", ",", "q", ",", "m1", ",", "m2", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(G1.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m1"}], "]"}], TraditionalForm]]],
 "\n(G1.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "q", ",", "m2"}], "]"}], TraditionalForm]]],
 "\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m1", ":=", 
     RowBox[{"c1", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{"t", "/", 
        RowBox[{"lp", "[", "p", "]"}]}], ")"}]}]}], ",", 
    RowBox[{"m2", ":=", 
     RowBox[{"c2", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{"t", "/", 
        RowBox[{"lp", "[", "p", "]"}]}], ")"}]}]}]}], TraditionalForm]]],
 " are suitable witnesses; Indeed, (G1.1) follows readily from (7), (IRR II) \
and the definition of areIrrelative. (G1.2) and (G1.3) follow from (8) and \
(9), resp., together with (2), (3) and (RED III).\n\n(G2):\nRather than \
proving (G2) we assume the negation of (G2) and show the negation of (6), \
i.e. we assume\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areIrrelative", "[", 
    RowBox[{"p", ",", "q", ",", "m1", ",", "m2", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m1"}], "]"}], TraditionalForm]]],
 "\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "q", ",", "m2"}], "]"}], TraditionalForm]]],
 "\n(15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"m1", "*", "p"}]}], ",", "q", ",", "m2"}], "]"}], "\[Or]", 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"m2", "*", "q"}]}], ",", "p", ",", "m1"}], "]"}]}], 
   TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2"}], "]"}]}], TraditionalForm]]],
 ", and we show\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"c3", ",", "c4"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"areCommonReducibles", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", "t"}], "]"}], ",", 
           RowBox[{"C", "[", 
            RowBox[{"a", ",", "t"}], "]"}], ",", "p", ",", "q", ",", "c3", 
           ",", "c4", ",", "i", ",", "j"}], "]"}]},
        {
         RowBox[{
          RowBox[{"isReducible", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"C", "[", 
              RowBox[{"a", ",", "t"}], "]"}], "-", 
             RowBox[{"c3", "*", 
              RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
            RowBox[{"lc", "[", "q", "]"}], ",", "c4"}], "]"}], "\[Or]", 
          RowBox[{"isReducible", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"C", "[", 
              RowBox[{"a", ",", "t"}], "]"}], "-", 
             RowBox[{"c4", "*", 
              RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
            RowBox[{"lc", "[", "p", "]"}], ",", "c3"}], "]"}]}]}
       }]}]}]}], TraditionalForm]]],
 "\n\nFrom (13) and (14), by (RED III), we can infer\n(16)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"m1", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(17)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"m2", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(18)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m1", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m1", "]"}]}], "]"}], TraditionalForm]]],
 "\n(19)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m2", "]"}], "*", 
        RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", 
     RowBox[{"lc", "[", "m2", "]"}]}], "]"}], TraditionalForm]]],
 "\n\nFrom (12), by (IRR II), we can infer\n(20)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{"areIrrelative", "[", 
      RowBox[{
       RowBox[{"lc", "[", "p", "]"}], ",", 
       RowBox[{"lc", "[", "q", "]"}], ",", 
       RowBox[{"lc", "[", "m1", "]"}], ",", 
       RowBox[{"lc", "[", "m2", "]"}], ",", "i", ",", "j"}], "]"}]}]}], 
   TraditionalForm]]],
 "\nand further\n(21)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"m1", ",", "q", ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"m2", ",", "q", ",", "j"}], "]"}]}], ")"}]}]}], 
   TraditionalForm]]],
 "\n(22)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{"lc", "[", "m1", "]"}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{"lc", "[", "m2", "]"}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], ")"}]}]}], 
   TraditionalForm]]],
 "\n\nFirst of all, we show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m1", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "m2", "]"}], "*", 
     RowBox[{"lp", "[", "q", "]"}]}]}], TraditionalForm]]],
 ". For doing so, assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m1", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "\[NotEqual]", 
    RowBox[{
     RowBox[{"lp", "[", "m2", "]"}], "*", 
     RowBox[{"lp", "[", "q", "]"}]}]}], TraditionalForm]]],
 " and show the negation of (1) instead, i.e. we prove the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"\[Tau]", ",", "\[Sigma]"}], "]"}]}], TraditionalForm]]],
 " s.t. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Tau]", "\[NotEqual]", "\[Sigma]"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "\[Tau]"}], 
   TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "\[Sigma]"}], 
   TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "\[Tau]"}], "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "\[Sigma]"}], "]"}], ",", "p", ",", "q", ",", "i", 
     ",", "j"}], "]"}], TraditionalForm]]],
 ". We claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Tau]", ":=", 
     RowBox[{
      RowBox[{"lp", "[", "m1", "]"}], "*", 
      RowBox[{"lp", "[", "p", "]"}]}]}], ",", 
    RowBox[{"\[Sigma]", ":=", 
     RowBox[{
      RowBox[{"lp", "[", "m2", "]"}], "*", 
      RowBox[{"lp", "[", "q", "]"}]}]}]}], TraditionalForm]]],
 " are suitable witnesses, and indeed the first three subgoals follow \
readily. We only have to make sure that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m1", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m2", "]"}], "*", 
        RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], ",", "p", ",", "q", ",", 
     "i", ",", "j"}], "]"}], TraditionalForm]]],
 " holds, which is equivalent to showing\n(C1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"c3", ",", "c4"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"p", "=", 
          RowBox[{"q", "\[Implies]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
             RowBox[{"mult", "[", 
              RowBox[{"c3", ",", 
               RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
             RowBox[{"mult", "[", 
              RowBox[{"c4", ",", 
               RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], 
            ")"}]}]}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", 
             RowBox[{
              RowBox[{"lp", "[", "m1", "]"}], "*", 
              RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c3"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", 
             RowBox[{
              RowBox[{"lp", "[", "m2", "]"}], "*", 
              RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c4"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\nIt is apparent that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c3", ":=", 
     RowBox[{"lc", "[", "m1", "]"}]}], ",", 
    RowBox[{"c4", ":=", 
     RowBox[{"lc", "[", "m2", "]"}]}]}], TraditionalForm]]],
 " witness this existential goal (use (18), (19) and (22)), and so we may add\
\n(23)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m1", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "m2", "]"}], "*", 
     RowBox[{"lp", "[", "q", "]"}]}]}], TraditionalForm]]],
 "\nto our knowledge base.\n\nNext, we show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m1", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{"t", "(", 
     RowBox[{"=", 
      RowBox[{
       RowBox[{"lp", "[", "m2", "]"}], "*", "q"}]}], ")"}]}], 
   TraditionalForm]]],
 ". This proceeds similarly as before, so we assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m1", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "\[NotEqual]", "t"}], TraditionalForm]]],
 " and show the negation of (1) instead, where now of course ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m1", "]"}], "*", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " are claimed to be witnesses, and since the first three subgoals are again \
obvious (use (3)!), we can reduce our goal to proving\n(C2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"c3", ",", "c4"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"p", "=", 
          RowBox[{"q", "\[Implies]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
             RowBox[{"mult", "[", 
              RowBox[{"c3", ",", 
               RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
             RowBox[{"mult", "[", 
              RowBox[{"c4", ",", 
               RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], 
            ")"}]}]}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", 
             RowBox[{
              RowBox[{"lp", "[", "m1", "]"}], "*", 
              RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c3"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", "t"}], "]"}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "c4"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\nIt is apparent that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c3", ":=", 
     RowBox[{"lc", "[", "m1", "]"}]}], ",", 
    RowBox[{"c4", ":=", "c2"}]}], TraditionalForm]]],
 " witness this existential goal (use (18), (7), (9) and (22)), and so we may \
add\n(24)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m1", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", "t"}], TraditionalForm]]],
 "\nto our knowledge base.\n\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c3", ":=", 
     RowBox[{"lc", "[", "m1", "]"}]}], ",", 
    RowBox[{"c4", ":=", 
     RowBox[{"lc", "[", "m2", "]"}]}]}], TraditionalForm]]],
 " witness the existential quantifier in (G3), so we prove\n(G3.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", "p", ",", "q", ",", 
     RowBox[{"lc", "[", "m1", "]"}], ",", 
     RowBox[{"lc", "[", "m2", "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(G3.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "t"}], "]"}], "-", 
       RowBox[{
        RowBox[{"lc", "[", "m1", "]"}], "*", 
        RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", 
      RowBox[{"lc", "[", "m2", "]"}]}], "]"}], "\[Or]", 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "t"}], "]"}], "-", 
       RowBox[{
        RowBox[{"lc", "[", "m2", "]"}], "*", 
        RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
      RowBox[{"lc", "[", "p", "]"}], ",", 
      RowBox[{"lc", "[", "m1", "]"}]}], "]"}]}], TraditionalForm]]],
 "\n\n(G3.1):\nBy the definition of areCommonReducibles, proving (G3.1) is \
equivalent to proving\n(G4.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{"lc", "[", "m1", "]"}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{"lc", "[", "m2", "]"}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], ")"}]}]}], 
   TraditionalForm]]],
 "\n(G4.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m1", "]"}]}], "]"}], TraditionalForm]]],
 "\n(G4.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", 
     RowBox[{"lc", "[", "m2", "]"}]}], "]"}], TraditionalForm]]],
 "\n(G4.1) is exactly what (22) states, (G4.2) and (G4.3) follow from (18) \
and (19), resp., using (23) and (24).\n\n(G3.2):\nBased on (15) we \
distinguish two cases; However, since both of them are entirely symmetric, we \
only consider the first one with\n(25)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"m1", "*", "p"}]}], ",", "q", ",", "m2"}], "]"}], 
   TraditionalForm]]],
 "\nand we show the first disjunct of (G3.2).\n\nFrom (25), by (RED III), we \
can infer\n(26)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m1", "*", "p"}]}], ",", 
       RowBox[{
        RowBox[{"lp", "[", "m2", "]"}], "*", 
        RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", 
     RowBox[{"lc", "[", "m2", "]"}]}], "]"}], TraditionalForm]]],
 "\nHence, if we manage to prove ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"m1", "*", "p"}]}], ",", 
      RowBox[{
       RowBox[{"lp", "[", "m2", "]"}], "*", 
       RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], "-", 
     RowBox[{
      RowBox[{"lc", "[", "m1", "]"}], "*", 
      RowBox[{"lc", "[", "p", "]"}]}]}]}], TraditionalForm]]],
 ", we are done, as the goal then follows immediately from (26).\n\nUsing \
linearity of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " we can replace ",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"m1", "*", "p"}]}], ",", 
     RowBox[{
      RowBox[{"lp", "[", "m2", "]"}], "*", 
      RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", 
      RowBox[{
       RowBox[{"lp", "[", "m2", "]"}], "*", 
       RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], "-", 
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"m1", "*", "p"}], ",", 
      RowBox[{
       RowBox[{"lp", "[", "m2", "]"}], "*", 
       RowBox[{"lp", "[", "q", "]"}]}]}], "]"}]}], TraditionalForm]]],
 ". The second summand can be rewritten into ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m1", "]"}], "*", 
    RowBox[{"lc", "[", "p", "]"}]}], TraditionalForm]]],
 " (by (23)). Therefore, we can cancel this summand on either side, and we \
only need to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", 
      RowBox[{
       RowBox[{"lp", "[", "m2", "]"}], "*", 
       RowBox[{"lp", "[", "q", "]"}]}]}], "]"}], "=", 
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "t"}], "]"}]}], TraditionalForm]]],
 ", which, however, follows readily from (23) and (24)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6326462122998123`*^9, 3.6326466006788797`*^9}, {
   3.632646718351561*^9, 3.632646756299131*^9}, {3.632646835202126*^9, 
   3.632647512507695*^9}, {3.632647543464513*^9, 3.6326475472766533`*^9}, {
   3.632647601084621*^9, 3.63264773434584*^9}, {3.63264776621642*^9, 
   3.632647856690927*^9}, {3.6326479124061403`*^9, 3.6326479128111*^9}, {
   3.632647945415196*^9, 3.632647970747468*^9}, {3.63264801226869*^9, 
   3.632648013424748*^9}, {3.6326480625573874`*^9, 3.632648155556138*^9}, {
   3.632648209482398*^9, 3.632648449034902*^9}, {3.632648526324667*^9, 
   3.6326486182950974`*^9}, {3.6326486488884983`*^9, 3.632648659403475*^9}, {
   3.632648720449026*^9, 3.632648839472225*^9}, 3.6326488920698843`*^9, {
   3.632649110113306*^9, 3.632649110482452*^9}, 3.636440189701846*^9, {
   3.638774689215733*^9, 3.638774696320051*^9}, {3.638774790815674*^9, 
   3.638774809082802*^9}, {3.638774955697468*^9, 3.638774982562805*^9}, 
   3.643446619816601*^9, 3.6532081055307217`*^9, 3.6532136815969324`*^9, 
   3.666173929562207*^9, 3.666440529042212*^9},
 CellID->1645016545],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.63264886670369*^9, 3.632648892070016*^9}, 
   3.636440189701969*^9, 3.64344661981672*^9, 3.6532081055307217`*^9, 
   3.6532136815969324`*^9, 3.666173929562207*^9, 3.666440529042349*^9},
 TextAlignment->Right,
 CellID->1172329439]
}, Closed]],

Cell[TextData[{
 StyleBox["Remark",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 ": The proof of (CR1.\[Implies].2) is the largest proof in the whole theory. \
In total, it consists of 322 (!) nodes. For comparison: The second-largest \
proof, namely the one of (CR1.\[DoubleLeftArrow]), consists only of 242 nodes."
}], "Text",
 CellChangeTimes->{{3.647157414197398*^9, 3.647157473936504*^9}, {
  3.647157526172687*^9, 3.6471575968061953`*^9}},
 CellID->328625772],

Cell["", "OpenEnvironment",
 CellID->1707859469],

Cell[CellGroupData[{

Cell["\<\
Theorem: Directions \[Implies] and \[DoubleLeftArrow] in (CR2)\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632649820106559*^9, 3.632649822818488*^9}, 
   3.632721800862067*^9, {3.6332358463135233`*^9, 3.633235847180189*^9}},
 CellID->1044109371],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "a"}], "]"}]}]], 
  RowBox[{"p", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1272506214}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632216516380025*^9, 3.632216539154045*^9}, {
   3.632216635956822*^9, 3.632216636110222*^9}, {3.632216689660941*^9, 
   3.632216697227892*^9}, {3.632568909496789*^9, 3.632568914750527*^9}, {
   3.633235201217596*^9, 3.6332352044207687`*^9}, 3.635748043914515*^9},
 CellTags->{
  "ID:1272506214", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1272506214],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "a"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", 
         RowBox[{"t", ",", "s"}], "]"}]}]], 
      RowBox[{"t", "\[NotEqual]", "s"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
            UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]},
          {
           RowBox[{
            RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
            UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "s"}]},
          {
           RowBox[{
            UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
            RowBox[{
             RowBox[{
              UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
              RowBox[{"a", ",", "t"}], "]"}], ",", 
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
            "]"}]},
          {
           RowBox[{
            UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
            RowBox[{
             RowBox[{
              UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
              RowBox[{"a", ",", "s"}], "]"}], ",", 
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
            "]"}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR2.\[DoubleRightArrow].1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:386901098", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633235210677794*^9, 3.633235223410894*^9}, {
   3.6332352897331543`*^9, 3.6332353390849457`*^9}, 3.633235772858781*^9, 
   3.638846213708844*^9, {3.6388465488903313`*^9, 3.638846552887195*^9}},
 CellTags->{
  "ID:386901098", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleRightArrow].1"},
 CellID->386901098],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "a"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]},
         {
          RowBox[{
           UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
           RowBox[{
            RowBox[{
             UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
            
            RowBox[{
             UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
             RowBox[{"a", ",", "t"}], "]"}]}], "]"}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR2.\[DoubleRightArrow].2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:670300270", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633235373251902*^9, 3.633235413012869*^9}, {
   3.63323576448206*^9, 3.6332357647398243`*^9}, {3.638788629446169*^9, 
   3.638788630714253*^9}, 3.638846228252695*^9},
 CellTags->{
  "ID:670300270", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleRightArrow].2"},
 CellID->670300270],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[Exists]", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
          RowBox[{"[", 
           RowBox[{"t", ",", "s"}], "]"}]}]], 
        RowBox[{"t", "\[NotEqual]", "s"}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"\[And]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {
             RowBox[{
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
              UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]},
            {
             RowBox[{
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
              UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "s"}]},
            {
             RowBox[{
              UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
              RowBox[{
               RowBox[{
                UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                RowBox[{"a", ",", "t"}], "]"}], ",", 
               RowBox[{
                UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
              "]"}]},
            {
             RowBox[{
              UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
              RowBox[{
               RowBox[{
                UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                RowBox[{"a", ",", "s"}], "]"}], ",", 
               RowBox[{
                UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
              "]"}]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "t", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]},
           {
            RowBox[{
             UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
             RowBox[{
              RowBox[{
               UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
              ",", 
              RowBox[{
               UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
               RowBox[{"a", ",", "t"}], "]"}]}], "]"}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "a"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR2.\[DoubleLeftArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1434882124", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633235439719268*^9, 3.6332354645593367`*^9}, {
   3.633235760260332*^9, 3.633235760554367*^9}, 3.633235806869268*^9, {
   3.6387886353313303`*^9, 3.638788636592111*^9}, {3.638846231777718*^9, 
   3.638846234439054*^9}, {3.638846557119808*^9, 3.638846561070115*^9}},
 CellTags->{
  "ID:1434882124", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleLeftArrow]"},
 CellID->1434882124],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1079407213]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1088471095],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR2.\[DoubleRightArrow].1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR2.\[DoubleRightArrow].1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["s", "ExpressionVariable"]}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
                   "\[And]", 
                   RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}]}]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["s", "ExpressionVariable"]}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
                   "\[And]", 
                   RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}]}]}], 
               TagBox[
               ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:386901098"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p386901098-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p386901098-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:386901098", "Proof|ID:386901098-1"},
 CellID->12812135],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p386901098-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->168109679]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR2.\[DoubleRightArrow].2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR2.\[DoubleRightArrow].2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 StyleBox["t", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:670300270"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p670300270-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p670300270-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:670300270", "Proof|ID:670300270-1"},
 CellID->488830292],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p670300270-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->565239118]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR2.\[DoubleLeftArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR2.\[DoubleLeftArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["s", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
                    "\[And]", 
                    RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}]}]}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}]}], 
                   "]"}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["s", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
                    "\[And]", 
                    RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}]}]}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   StyleBox["t", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}]}], 
                   "]"}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1434882124"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1434882124-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1434882124-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1434882124", "Proof|ID:1434882124-1"},
 CellID->822190239],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1434882124-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->1425736080]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR2.\[Implies].1)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632567820520041*^9, 3.6325678312684593`*^9}, {
   3.632567881512978*^9, 3.632567883394401*^9}, {3.63256884181592*^9, 
   3.632568852102487*^9}, {3.632568889246223*^9, 3.6325688896891747`*^9}, 
   3.636440174849916*^9, 3.638785688966083*^9, 3.6434466270569553`*^9, 
   3.6532080948719025`*^9, 3.653213672938917*^9, 3.6661739186733885`*^9, 
   3.666440520504323*^9},
 CellID->1621914703],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632568730163651*^9, 3.632568731354806*^9}, 
   3.632568852102669*^9, 3.636440174850047*^9, 3.643446627057075*^9, 
   3.6532080948719025`*^9, 3.653213672938917*^9, 3.6661739186733885`*^9, 
   3.666440520504425*^9},
 CellID->414334676],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"s", "\[VerticalBar]", "t"}], "\[Implies]", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"t", "/", "s"}], ")"}], "*", "s"}]}], "=", "t"}], 
   TraditionalForm]]],
 " (ET CPP)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6324830728683434`*^9, 3.632483087105466*^9}, {
   3.632483710199279*^9, 3.63248371021541*^9}, 3.632558133552821*^9, {
   3.63256550933005*^9, 3.6325655153292847`*^9}, {3.633168632794017*^9, 
   3.633168635268145*^9}, 3.636440174850176*^9, 3.6434466270571547`*^9, 
   3.6532080948719025`*^9, 3.653213672938917*^9, 3.6661739186733885`*^9, 
   3.666440520504509*^9},
 CellID->2054694430],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "a"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "a"}], "]"}], TraditionalForm]]],
 " and show the RHS of the implication.\nBefore, however, we expand the \
definition of isNTCR: We know\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p"}], "]"}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", 
        RowBox[{"m1", ",", "m2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", "p", ",", "m1"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", "p", ",", "m2"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"a", "-", 
           RowBox[{"m2", "*", "p"}]}], ",", "p", ",", "m1"}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\n\nWe exchange our current goal with the negated existential formula in \
our knowledge base, i.e. we assume there exist ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"s", ",", "t"}], "]"}]}], TraditionalForm]]],
 " with\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[NotEqual]", "s"}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "s"}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "s"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 "\nand prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"m1", ",", "m2"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"isReducible", "[", 
          RowBox[{"a", ",", "p", ",", "m1"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{"a", ",", "p", ",", "m2"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"m2", "*", "p"}]}], ",", "p", ",", "m1"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\n\nFrom (7) and (8) we know that there exist ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], TraditionalForm]]],
 " with\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "s"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n\nBefore we continue we show\n(C1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", 
     RowBox[{"c1", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{"t", "/", 
        RowBox[{"lp", "[", "p", "]"}]}], ")"}]}]}], "]"}], 
   TraditionalForm]]],
 "\n(C2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", 
     RowBox[{"c2", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{"s", "/", 
        RowBox[{"lp", "[", "p", "]"}]}], ")"}]}]}], "]"}], 
   TraditionalForm]]],
 "\nSince both proceed analogously, we only consider (C1) here.\nFrom (9), by \
the definition of isReducible, we know\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"c1", ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 "\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "t"}], "]"}], "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], "-", 
     RowBox[{"c1", "*", 
      RowBox[{"lc", "[", "p", "]"}]}]}]}], TraditionalForm]]],
 "\nBy (RED III), showing (C1) amounts to showing ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " (known),\n(C1.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m1", ",", "p"}], "]"}], TraditionalForm]]],
 "\n(C1.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", 
      RowBox[{
       RowBox[{"lp", "[", "m1", "]"}], "*", 
       RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m1", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
     RowBox[{
      RowBox[{"lc", "[", "m1", "]"}], "*", 
      RowBox[{"lc", "[", "p", "]"}]}]}]}], TraditionalForm]]],
 "\n(C1.1) follows readily from (11), whereas (C1.2) follows readily from \
(12) using the facts that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m1", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", "t"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m1", "]"}], "=", "c1"}], TraditionalForm]]],
 ".\n\nWe have to distinguish two cases based on the relative order of ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 ". However, since both cases are completely symmatric, we only consider the \
case\n(C)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Succeeds]", "s"}], TraditionalForm]]],
 "\nhere.\n\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m1", ":=", 
    RowBox[{"c1", "\[CenterDot]", 
     RowBox[{"(", 
      RowBox[{"t", "/", 
       RowBox[{"lp", "[", "p", "]"}]}], ")"}]}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m2", ":=", 
    RowBox[{"c2", "\[CenterDot]", 
     RowBox[{"(", 
      RowBox[{"s", "/", 
       RowBox[{"lp", "[", "p", "]"}]}], ")"}]}]}], TraditionalForm]]],
 " witness our existential goal, hence we show all of\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m1"}], "]"}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m2"}], "]"}], TraditionalForm]]],
 "\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"m2", "*", "p"}]}], ",", "p", ",", "m1"}], "]"}], 
   TraditionalForm]]],
 "\n(Note: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", "\[Succeeds]", "t"}], TraditionalForm]]],
 ", the witnesses have to be swapped).\n\n(G1):\nThis is precisely what (C1) \
states.\n\n(G2):\nThis is precisely what (C2) states.\n\n(G3):\nWe use (404 \
III) instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", ":=", "a"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c1", ":=", "p"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c2", ":=", "p"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m1", ":=", "m1"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m2", ":=", "m2"}], TraditionalForm]]],
 ". Therefore, our current goal can be reduced to\n(G3.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m1"}], "]"}], TraditionalForm]]],
 "\n(G3.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m2"}], "]"}], TraditionalForm]]],
 "\n(G3.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m1", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "\[Succeeds]", 
    RowBox[{
     RowBox[{"lp", "[", "m2", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}]}], TraditionalForm]]],
 "\n(G3.1) and (G3.2) are exactly what we have just showed, so the only \
subgoal remaining is (G3.3); This, however, is equivalent to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Succeeds]", "s"}], TraditionalForm]]],
 " by a lemma, which is our case assumption (C)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632567846884447*^9, 3.6325687175661*^9}, 
   3.6325688521028347`*^9, {3.6325710771373453`*^9, 3.632571124415392*^9}, {
   3.632571176891365*^9, 3.632571176891512*^9}, {3.6325712612929*^9, 
   3.6325712746283092`*^9}, {3.632571315125948*^9, 3.632571320297558*^9}, {
   3.632571357376912*^9, 3.632571357377029*^9}, {3.632636838926916*^9, 
   3.632636841320979*^9}, 3.636440174850278*^9, {3.638698373412301*^9, 
   3.6386983804818172`*^9}, {3.638698470896261*^9, 3.638698497974391*^9}, {
   3.6386985499627953`*^9, 3.638698551166155*^9}, {3.638699343505384*^9, 
   3.6386993582315617`*^9}, {3.6386994204228907`*^9, 3.638699480044532*^9}, {
   3.638699552373905*^9, 3.6386997571836367`*^9}, {3.638785697807103*^9, 
   3.638785726713965*^9}, {3.638785762407056*^9, 3.638785771437606*^9}, {
   3.638785811724044*^9, 3.638785944745956*^9}, {3.6387863637985773`*^9, 
   3.638786659631674*^9}, {3.638786719201724*^9, 3.638786761410277*^9}, {
   3.63878679791742*^9, 3.638787131752638*^9}, {3.6387871667186623`*^9, 
   3.6387873244180527`*^9}, {3.638787363046373*^9, 3.63878746889783*^9}, {
   3.638787716027577*^9, 3.638787716694425*^9}, {3.638787752903667*^9, 
   3.638787878657939*^9}, {3.638846135571518*^9, 3.6388461356859627`*^9}, 
   3.64344662705726*^9, 3.6532080948719025`*^9, 3.653213672938917*^9, 
   3.6661739186733885`*^9, 3.666440520504594*^9},
 CellID->1844460214],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632568792744026*^9, 3.6325687962371883`*^9}, 
   3.632568852102993*^9, 3.6364401748504*^9, 3.643446627057377*^9, 
   3.6532080948719025`*^9, 3.653213672938917*^9, 3.6661739186733885`*^9, 
   3.6664405205046997`*^9},
 TextAlignment->Right,
 CellID->1726142358]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR2.\[Implies].2)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632569164769092*^9, 3.632569176756404*^9}, 
   3.632570094976581*^9, {3.632634231493162*^9, 3.63263423943274*^9}, {
   3.632634660068185*^9, 3.632634662057506*^9}, 3.636440182689971*^9, 
   3.6387875595870857`*^9, 3.643446633095633*^9, 3.653208087976691*^9, 
   3.6532136673073072`*^9, 3.6661739119341764`*^9, 3.666440514450067*^9},
 CellID->458708427],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "a"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "a"}], "]"}], TraditionalForm]]],
 " and show the RHS of the implication. From (CR2.\[Implies].1) we can infer\n\
(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", 
         RowBox[{"s", ",", "t"}], "]"}]}]], 
      RowBox[{"s", "\[NotEqual]", "t"}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "s"}]},
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "t"}], "]"}], ",", 
          RowBox[{"lc", "[", "p", "]"}]}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "s"}], "]"}], ",", 
          RowBox[{"lc", "[", "p", "]"}]}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\nWe expand the definition of isNTCR: We know\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p"}], "]"}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", 
        RowBox[{"m1", ",", "m2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", "p", ",", "m1"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", "p", ",", "m2"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"a", "-", 
           RowBox[{"m2", "*", "p"}]}], ",", "p", ",", "m1"}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\nWe expand the definition of ",
 Cell[BoxData[
  FormBox["isReducible", TraditionalForm]]],
 ": We know that there exists ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "m3", "]"}]}], TraditionalForm]]],
 " with\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m3"}], "]"}], TraditionalForm]]],
 "\n\nFrom (4), together with (RED II), we can infer (among others)\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m3", "]"}]}], "]"}], TraditionalForm]]],
 "\n\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m3", "]"}], "*", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 " witnesses our existential goal, so we need to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{
     RowBox[{"lp", "[", "m3", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}]}], TraditionalForm]]],
 " (obvious) and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], "]"}], 
   TraditionalForm]]],
 ". By the definition of isNTCR, we have to show\n\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", 
            RowBox[{
             RowBox[{"lp", "[", "m3", "]"}], "*", 
             RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
          RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", 
            RowBox[{
             RowBox[{"lp", "[", "m3", "]"}], "*", 
             RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
          RowBox[{"lc", "[", "p", "]"}], ",", "c2"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", 
             RowBox[{
              RowBox[{"lp", "[", "m3", "]"}], "*", 
              RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
           RowBox[{"c2", "*", 
            RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
          RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\n\nWe prove both subgoals.\n\n(G1):\nWe have to prove\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "c3", "]"}]}]], 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", 
        RowBox[{
         RowBox[{"lp", "[", "m3", "]"}], "*", 
         RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
      RowBox[{"lc", "[", "p", "]"}], ",", "c3"}], "]"}]}], 
   TraditionalForm]]],
 "\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c3", ":=", 
    RowBox[{"lc", "[", "m3", "]"}]}], TraditionalForm]]],
 " is a suitable witness, hence we show\n(G3.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m3", "]"}]}], "]"}], TraditionalForm]]],
 "\n(G3.1) is exactly what (5) states.\n\n(G2):\nRather than proving (G2), we \
assume its negation and show the negation of (3), i.e. we assume\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", 
        RowBox[{
         RowBox[{"lp", "[", "m3", "]"}], "*", 
         RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
      RowBox[{"c2", "*", 
       RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], TraditionalForm]]],
 ", and prove\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"m1", ",", "m2"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"isReducible", "[", 
          RowBox[{"a", ",", "p", ",", "m1"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{"a", ",", "p", ",", "m2"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"a", "-", 
            RowBox[{"m2", "*", "p"}]}], ",", "p", ",", "m1"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\n\nBefore we proceed we verify that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c1", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c2", "\[NotEqual]", "0"}], TraditionalForm]]],
 ".\n\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m1", ":=", 
     RowBox[{"c1", "\[CenterDot]", 
      RowBox[{"lp", "[", "m3", "]"}]}]}], ",", 
    RowBox[{"m2", ":=", 
     RowBox[{"c2", "\[CenterDot]", 
      RowBox[{"lp", "[", "m3", "]"}]}]}]}], TraditionalForm]]],
 " witness our existential goal (G4), so we have to show all of\n(G4.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m1"}], "]"}], TraditionalForm]]],
 "\n(G4.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m2"}], "]"}], TraditionalForm]]],
 "\n(G4.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"m2", "*", "p"}]}], ",", "p", ",", "m1"}], "]"}], 
   TraditionalForm]]],
 "\n\nBefore we proceed we convince ourselves that\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"m1", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"m2", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\nand also\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m1", ",", "p"}], "]"}], TraditionalForm]]],
 "\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m2", ",", "p"}], "]"}], TraditionalForm]]],
 "\n(all can be shown rather easily, using (6) and (7)).\n\n(G4.1):\nBy (RED \
III) this can be reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " (known), ",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m1", ",", "p"}], "]"}], TraditionalForm]]],
 " (see (11)) and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m1", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m1", "]"}]}], "]"}], TraditionalForm]]],
 ". The last part follows from (6), using the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m1", "]"}], "=", 
    RowBox[{"lp", "[", "m3", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m1", "]"}], "=", "c1"}], TraditionalForm]]],
 ".\n\n(G4.2):\nBy (RED III) this can be reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " (known), ",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m2", ",", "p"}], "]"}], TraditionalForm]]],
 " (see (12)) and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m2", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m2", "]"}]}], "]"}], TraditionalForm]]],
 ". The last part follows from (7), using the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m2", "]"}], "=", 
    RowBox[{"lp", "[", "m3", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m2", "]"}], "=", "c2"}], TraditionalForm]]],
 ".\n\n(G4.3):\nShowing (G4.3), by (RED II), amounts to showing ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " (known), ",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m1", ",", "p"}], "]"}], TraditionalForm]]],
 " (known) and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m2", "*", "p"}]}], ",", 
       RowBox[{
        RowBox[{"lp", "[", "m1", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m1", "]"}]}], "]"}], TraditionalForm]]],
 ". Using the facts that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m1", "]"}], "=", 
    RowBox[{"lp", "[", "m3", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m1", "]"}], "=", "c1"}], TraditionalForm]]],
 ", we can rewrite our goal into ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m2", "*", "p"}]}], ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 ", thus, if we manage to prove\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"m2", "*", "p"}]}], ",", 
      RowBox[{
       RowBox[{"lp", "[", "m3", "]"}], "*", 
       RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
     RowBox[{"c2", "*", 
      RowBox[{"lc", "[", "p", "]"}]}]}]}], TraditionalForm]]],
 "\nwe are done due to (8).\nUsing linearity of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " (G5) can be rewritten into ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"m2", "*", "p"}], ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m3", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
     RowBox[{"c2", "*", 
      RowBox[{"lc", "[", "p", "]"}]}]}]}], TraditionalForm]]],
 ", and cancelling ",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "[", 
    RowBox[{"a", ",", 
     RowBox[{
      RowBox[{"lp", "[", "m3", "]"}], "*", 
      RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], TraditionalForm]]],
 " on both sides leaves us with showing\n(G5\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"m2", "*", "p"}], ",", 
      RowBox[{
       RowBox[{"lp", "[", "m3", "]"}], "*", 
       RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "=", 
    RowBox[{"c2", "*", 
     RowBox[{"lc", "[", "p", "]"}]}]}], TraditionalForm]]],
 "\n(G5\[CloseCurlyQuote]) holds because of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m2", "]"}], "=", 
    RowBox[{"lp", "[", "m3", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m2", "]"}], "=", "c2"}], TraditionalForm]]],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6325692002780457`*^9, 3.6325694268634853`*^9}, 
   3.632570094976705*^9, {3.632570728372566*^9, 3.6325707578167763`*^9}, {
   3.632571432720799*^9, 3.632571445560379*^9}, {3.6325715938423557`*^9, 
   3.632571606722777*^9}, {3.6326306952628*^9, 3.6326307224971943`*^9}, {
   3.632630833528577*^9, 3.632630864121511*^9}, {3.632631023292745*^9, 
   3.6326310719796886`*^9}, {3.632631219279278*^9, 3.632631628444659*^9}, {
   3.6326316661808567`*^9, 3.6326317872650623`*^9}, {3.632631824382112*^9, 
   3.6326321995251713`*^9}, {3.632632264017104*^9, 3.6326323652336073`*^9}, {
   3.632632698487405*^9, 3.6326327720536613`*^9}, {3.632632823633246*^9, 
   3.6326328288054247`*^9}, {3.63263290792702*^9, 3.632633024732746*^9}, {
   3.632633067837179*^9, 3.632633099450726*^9}, {3.632633198475442*^9, 
   3.632633216851014*^9}, {3.632633354879549*^9, 3.632633389958372*^9}, {
   3.632633495972414*^9, 3.632633646873837*^9}, {3.632633682044121*^9, 
   3.632633728364417*^9}, {3.632633780943246*^9, 3.632633824062347*^9}, {
   3.632633876924165*^9, 3.632633993877343*^9}, {3.632634030584486*^9, 
   3.632634062690051*^9}, {3.632634099874289*^9, 3.632634105280949*^9}, {
   3.632634143753538*^9, 3.632634150648302*^9}, {3.6326341869835377`*^9, 
   3.6326342394328613`*^9}, {3.6326439738739147`*^9, 
   3.6326444422750397`*^9}, {3.632644475376178*^9, 3.632644879520465*^9}, {
   3.6326449162789507`*^9, 3.6326453385779333`*^9}, {3.632645392664752*^9, 
   3.6326457863326693`*^9}, {3.632645844014824*^9, 3.632646142967249*^9}, 
   3.636440182690127*^9, {3.638765259422262*^9, 3.638765264136723*^9}, 
   3.638765393861206*^9, {3.638770145374399*^9, 3.638770197363942*^9}, {
   3.638770234159388*^9, 3.63877030718576*^9}, {3.638770340000877*^9, 
   3.638770392275969*^9}, {3.638787568529624*^9, 3.638787669778882*^9}, {
   3.6387879048399553`*^9, 3.63878795363321*^9}, {3.6387880701255293`*^9, 
   3.638788151818697*^9}, 3.638788199810878*^9, {3.6387882316972322`*^9, 
   3.638788580799611*^9}, {3.638788613643064*^9, 3.6387886160783453`*^9}, {
   3.638788666258003*^9, 3.638788756730616*^9}, {3.638788790023407*^9, 
   3.638788829512328*^9}, {3.638788882892831*^9, 3.638789050049181*^9}, {
   3.638849100250855*^9, 3.638849109463181*^9}, {3.638849175851503*^9, 
   3.6388492470328693`*^9}, {3.638849298881391*^9, 3.638849395105895*^9}, {
   3.638849630707616*^9, 3.638849631189797*^9}, {3.638849666989993*^9, 
   3.638849671179981*^9}, {3.6388497828980637`*^9, 3.6388497890205793`*^9}, {
   3.638849829700315*^9, 3.638849872010647*^9}, {3.6388499099853487`*^9, 
   3.638849909985478*^9}, {3.638849995749789*^9, 3.638850131156288*^9}, 
   3.6434466330957527`*^9, 3.653208087976691*^9, 3.6532136673073072`*^9, 
   3.6661739119341764`*^9, 3.6664405144502077`*^9},
 CellID->936066289],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.63263421142743*^9, 3.632634239432981*^9}, 
   3.636440182690287*^9, 3.643446633095871*^9, 3.653208087976691*^9, 
   3.6532136673073072`*^9, 3.6661739119341764`*^9, 3.666440514450347*^9},
 TextAlignment->Right,
 CellID->898836392]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of",
  FontSlant->"Italic"],
 " ",
 StyleBox["(CR2.\[DoubleLeftArrow])",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632646174827505*^9, 3.632646190352857*^9}, 
   3.6326488920697613`*^9, 3.636440189701724*^9, 3.638852102062901*^9, 
   3.643446640376944*^9, 3.653208080317077*^9, 3.6532136608176956`*^9, 
   3.666173905444565*^9, 3.666440508665924*^9},
 CellID->1317000955],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "a"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", 
         RowBox[{"t", ",", "s"}], "]"}]}]], 
      RowBox[{"t", "\[NotEqual]", "s"}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "s"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "t"}], "]"}], ",", 
          RowBox[{"lc", "[", "p", "]"}]}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "s"}], "]"}], ",", 
          RowBox[{"lc", "[", "p", "]"}]}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}]}], "]"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 ". We have to show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "a"}], "]"}], TraditionalForm]]],
 ".\n\nFrom (3), by the definition of isNTCR, we can infer\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "t"}], "]"}], ",", 
          RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "t"}], "]"}], ",", 
          RowBox[{"lc", "[", "p", "]"}], ",", "c2"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", "t"}], "]"}], "-", 
           RowBox[{"c2", "*", 
            RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
          RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\n\nFrom (4), by the definition of isReducible, we can infer\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c"}], "]"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}], TraditionalForm]]],
 ". Clearly, because of (6) we know\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"c", ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 "\n\nBy the definition of isNTCR we have to show both\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p"}], "]"}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", 
        RowBox[{"m1", ",", "m2"}], "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", "p", ",", "m1"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{"a", ",", "p", ",", "m2"}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"a", "-", 
           RowBox[{"m2", "*", "p"}]}], ",", "p", ",", "m1"}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\n\n(G1):\nFor proving ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p"}], "]"}], TraditionalForm]]],
 " we have to find ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "m", "]"}]}], TraditionalForm]]],
 " s.t.\n(G1.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m"}], "]"}], TraditionalForm]]],
 "\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", ":=", 
    RowBox[{"c", "\[CenterDot]", 
     RowBox[{"(", 
      RowBox[{"t", "/", 
       RowBox[{"lp", "[", "p", "]"}]}], ")"}]}]}], TraditionalForm]]],
 " is a suitable witness; Indeed, (G1.1) follows readily from (6), (2) and \
(RED ?).\n\n(G2):\nRather than proving (G2) we assume the negation of (G2) \
and show the negation of (5), i.e. we assume\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m1"}], "]"}], TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p", ",", "m2"}], "]"}], TraditionalForm]]],
 "\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"m2", "*", "p"}]}], ",", "p", ",", "m1"}], "]"}], 
   TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2"}], "]"}]}], TraditionalForm]]],
 ", and we show\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"c1", ",", "c2"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", "t"}], "]"}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", "t"}], "]"}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c2"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"a", ",", "t"}], "]"}], "-", 
            RowBox[{"c2", "*", 
             RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\n\nFrom (9) and (10), by (RED ?), we can infer\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m1", ",", "p"}], "]"}], TraditionalForm]]],
 "\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m2", ",", "p"}], "]"}], TraditionalForm]]],
 "\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m1", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m1", "]"}]}], "]"}], TraditionalForm]]],
 "\n(15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m2", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m2", "]"}]}], "]"}], TraditionalForm]]],
 "\n\n-\nSUBPROOF:\n\nFirst of all, we show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m1", "]"}], "=", 
    RowBox[{"lp", "[", "m2", "]"}]}], TraditionalForm]]],
 ". For doing so, assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m1", "]"}], "\[NotEqual]", 
    RowBox[{"lp", "[", "m2", "]"}]}], TraditionalForm]]],
 " and show the negation of (1) instead, i.e. we prove the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"\[Tau]", ",", "\[Sigma]"}], "]"}]}], TraditionalForm]]],
 " s.t. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Tau]", "\[NotEqual]", "\[Sigma]"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "\[Tau]"}], 
   TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "\[Sigma]"}], 
   TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "\[Tau]"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "\[Sigma]"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 ". We claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Tau]", ":=", 
     RowBox[{
      RowBox[{"lp", "[", "m1", "]"}], "*", 
      RowBox[{"lp", "[", "p", "]"}]}]}], ",", 
    RowBox[{"\[Sigma]", ":=", 
     RowBox[{
      RowBox[{"lp", "[", "m2", "]"}], "*", 
      RowBox[{"lp", "[", "p", "]"}]}]}]}], TraditionalForm]]],
 " are suitable witnesses, and indeed the first three subgoals follow \
readily. We only have to make sure that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m1", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m2", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 " holds, which is equivalent to showing\n(C1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"c3", ",", "c4"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", 
             RowBox[{
              RowBox[{"lp", "[", "m1", "]"}], "*", 
              RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c3"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", 
             RowBox[{
              RowBox[{"lp", "[", "m2", "]"}], "*", 
              RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c4"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\nIt is apparent that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c3", ":=", 
     RowBox[{"lc", "[", "m1", "]"}]}], ",", 
    RowBox[{"c4", ":=", 
     RowBox[{"lc", "[", "m2", "]"}]}]}], TraditionalForm]]],
 " witness this existential goal (use (14) and (15)), and so we may add\n(16)\
\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m1", "]"}], "=", 
    RowBox[{"lp", "[", "m2", "]"}]}], TraditionalForm]]],
 "\nto our knowledge base.\n\nNext, we show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m1", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{"t", "(", 
     RowBox[{"=", 
      RowBox[{
       RowBox[{"lp", "[", "m2", "]"}], "*", "p"}]}], ")"}]}], 
   TraditionalForm]]],
 ". This proceeds similarly as before, so we assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m1", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "\[NotEqual]", "t"}], TraditionalForm]]],
 " and show the negation of (1) instead, where now of course ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "m1", "]"}], "*", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " are claimed to be witnesses, and since the first three subgoals are again \
obvious (use (2)!), we can reduce our goal to proving\n(C2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"c3", ",", "c4"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", 
             RowBox[{
              RowBox[{"lp", "[", "m1", "]"}], "*", 
              RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c3"}], "]"}]},
        {
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"a", ",", "t"}], "]"}], ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "c4"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\nIt is apparent that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c3", ":=", 
     RowBox[{"lc", "[", "m1", "]"}]}], ",", 
    RowBox[{"c4", ":=", "c"}]}], TraditionalForm]]],
 " witness this existential goal (use (14) and (6)), and so we may add\n(17)\t\
",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "m1", "]"}], "*", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", "t"}], TraditionalForm]]],
 "\nto our knowledge base.\n\n-\n\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c1", ":=", 
     RowBox[{"lc", "[", "m1", "]"}]}], ",", 
    RowBox[{"c2", ":=", 
     RowBox[{"lc", "[", "m2", "]"}]}]}], TraditionalForm]]],
 " witness the existential quantifier in (G3), so we prove\n(G3.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(G3.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\n(G3.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", "t"}], "]"}], "-", 
      RowBox[{"c2", "*", 
       RowBox[{"lc", "[", "p", "]"}]}]}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n\n(G3.1):\nThis is exactly what (14) states, using (17).\n\n(G3.2):\nThis \
is exactly what (15) states, using (16) and (17).\n\n(G3.3):\nFrom (11), by \
(RED ?), we can infer\n(18)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m2", "*", "p"}]}], ",", 
       RowBox[{
        RowBox[{"lp", "[", "m1", "]"}], "*", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "m1", "]"}]}], "]"}], TraditionalForm]]],
 "\nHence, if we manage to prove ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"m2", "*", "p"}]}], ",", 
      RowBox[{
       RowBox[{"lp", "[", "m1", "]"}], "*", 
       RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], "-", 
     RowBox[{"c2", "*", 
      RowBox[{"lc", "[", "p", "]"}]}]}]}], TraditionalForm]]],
 ", we are done, as the goal then follows immediately from (18).\n\nUsing \
linearity of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " we can replace ",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"m2", "*", "p"}]}], ",", 
     RowBox[{
      RowBox[{"lp", "[", "m1", "]"}], "*", 
      RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", 
      RowBox[{
       RowBox[{"lp", "[", "m1", "]"}], "*", 
       RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "-", 
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"m2", "*", "p"}], ",", 
      RowBox[{
       RowBox[{"lp", "[", "m1", "]"}], "*", 
       RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}], TraditionalForm]]],
 ". The second summand can be rewritten into ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m2", "]"}], "*", 
    RowBox[{"lc", "[", "p", "]"}]}], TraditionalForm]]],
 " (using (16)). Therefore, we can cancel this summand on either side, and we \
only need to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", 
      RowBox[{
       RowBox[{"lp", "[", "m1", "]"}], "*", 
       RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], "=", 
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "t"}], "]"}]}], TraditionalForm]]],
 ", which, however, follows readily from (17)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6326462122998123`*^9, 3.6326466006788797`*^9}, {
   3.632646718351561*^9, 3.632646756299131*^9}, {3.632646835202126*^9, 
   3.632647512507695*^9}, {3.632647543464513*^9, 3.6326475472766533`*^9}, {
   3.632647601084621*^9, 3.63264773434584*^9}, {3.63264776621642*^9, 
   3.632647856690927*^9}, {3.6326479124061403`*^9, 3.6326479128111*^9}, {
   3.632647945415196*^9, 3.632647970747468*^9}, {3.63264801226869*^9, 
   3.632648013424748*^9}, {3.6326480625573874`*^9, 3.632648155556138*^9}, {
   3.632648209482398*^9, 3.632648449034902*^9}, {3.632648526324667*^9, 
   3.6326486182950974`*^9}, {3.6326486488884983`*^9, 3.632648659403475*^9}, {
   3.632648720449026*^9, 3.632648839472225*^9}, 3.6326488920698843`*^9, {
   3.632649110113306*^9, 3.632649110482452*^9}, 3.636440189701846*^9, {
   3.638774689215733*^9, 3.638774696320051*^9}, {3.638774790815674*^9, 
   3.638774809082802*^9}, {3.638774955697468*^9, 3.638774982562805*^9}, {
   3.63885210942982*^9, 3.638852249370099*^9}, {3.6388523123145943`*^9, 
   3.6388525574257402`*^9}, {3.638852591237455*^9, 3.638852760232267*^9}, {
   3.638852795577712*^9, 3.638853048002417*^9}, {3.638853078325892*^9, 
   3.638853251748118*^9}, {3.638853285932275*^9, 3.638853323309149*^9}, {
   3.63885335523168*^9, 3.6388533904618998`*^9}, {3.6388534457800694`*^9, 
   3.638853460050743*^9}, {3.638853533995775*^9, 3.638853534178838*^9}, {
   3.6388535739737453`*^9, 3.6388536835814238`*^9}, 3.643446640377063*^9, 
   3.653208080317077*^9, 3.6532136608176956`*^9, 3.666173905444565*^9, 
   3.666440508666069*^9},
 CellID->1205890748],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.63264886670369*^9, 3.632648892070016*^9}, 
   3.636440189701969*^9, 3.643446640377182*^9, 3.653208080317077*^9, 
   3.6532136608176956`*^9, 3.666173905444565*^9, 3.6664405086662073`*^9},
 TextAlignment->Right,
 CellID->1499697181]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1580923328],

Cell[CellGroupData[{

Cell["Corollary: CR1 and CR2 for Monomials", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.632720135466861*^9, {3.632721961226816*^9, 3.632721968757958*^9}},
 CellID->141373954],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "a"}], "]"}]}]], 
  RowBox[{
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 737627449}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.6327218783686953`*^9, {3.6357480503135433`*^9, 3.635748054730784*^9}},
 CellTags->{
  "ID:737627449", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->737627449],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", "c", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 81313572}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6327348420403557`*^9, 3.6327348513147993`*^9}},
 CellTags->{
  "ID:81313572", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->81313572],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "s", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1098402572}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632734856605691*^9, 3.6327348648792763`*^9}, 
   3.632735164925763*^9},
 CellTags->{
  "ID:1098402572", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1098402572],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
    "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", "t", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
         RowBox[{
          RowBox[{
           UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
           RowBox[{"a", ",", "t"}], "]"}], ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
         "]"}], "\[And]", 
        RowBox[{
         UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
         RowBox[{"p", ",", "q", ",", 
          RowBox[{
           RowBox[{
            UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
            RowBox[{"a", ",", "t"}], "]"}], 
           UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], ",", 
          "i", ",", "j"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR1 monomial 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1499935354", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6327201601084146`*^9, 3.632720280312574*^9}, {
   3.6327203262103167`*^9, 3.632720337355557*^9}, 3.632720375506056*^9, 
   3.63272187668054*^9, {3.632725316699419*^9, 3.632725355374022*^9}},
 CellTags->{
  "ID:1499935354", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 1"},
 CellID->1499935354],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "a", "]"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
     "\[DoubleLongLeftRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
            UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
            RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "a", "]"}]}]},
          {
           RowBox[{
            RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}], 
            UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
            RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "a", "]"}]}]},
          {
           RowBox[{
            UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
            RowBox[{"p", ",", "q", ",", 
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "a", "]"}], ",",
              "i", ",", "j"}], "]"}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR1 monomial 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1938671644", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6327264467496643`*^9, 3.6327265181901913`*^9}, {
  3.6327291812033463`*^9, 3.6327292318201923`*^9}, {3.632729596633492*^9, 
  3.6327296115682297`*^9}, {3.6332355256366243`*^9, 3.633235526959429*^9}},
 CellTags->{
  "ID:1938671644", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 2"},
 CellID->1938671644],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "q", ",", 
      RowBox[{"c", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "s"}], ",", "i", 
      ",", "j"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "s"}]},
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "s"}]},
         {
          RowBox[{
           UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
           RowBox[{"p", ",", "q", ",", "c", ",", "i", ",", "j"}], "]"}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR1 monomial 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:48476485", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632734871407545*^9, 3.632734943822002*^9}, {
   3.632735011103457*^9, 3.632735078703352*^9}, {3.632735167323138*^9, 
   3.63273516927389*^9}, 3.6332355293105907`*^9},
 CellTags->{
  "ID:48476485", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 3"},
 CellID->48476485],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "a"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
       RowBox[{
        RowBox[{
         UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
         RowBox[{"a", ",", "t"}], "]"}], ",", 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}]}], "]"}], 
      "\[And]", 
      RowBox[{
       UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", 
        RowBox[{
         RowBox[{
          UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
          RowBox[{"a", ",", "t"}], "]"}], 
         UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}]}], 
       "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR2 monomial 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1246936946", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632720358173932*^9, 3.6327203809164543`*^9}, {
   3.632721881714601*^9, 3.6327218915975313`*^9}, {3.632735180701474*^9, 
   3.632735211325398*^9}, 3.633235620366961*^9},
 CellTags->{
  "ID:1246936946", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 1"},
 CellID->1246936946],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "a", "]"}], 
  "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "a"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
           RowBox[{
            UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "a", "]"}]}]},
         {
          RowBox[{
           UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
           RowBox[{
            RowBox[{
             UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
            
            RowBox[{
             UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "a", "]"}]}], 
           "]"}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR2 monomial 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1856728487", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633235573828002*^9, 3.633235592964628*^9}, {
  3.633235641486003*^9, 3.633235660853758*^9}, {3.633235702213914*^9, 
  3.6332357026923113`*^9}, {3.639121317770625*^9, 3.639121318885603*^9}},
 CellTags->{
  "ID:1856728487", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 2"},
 CellID->1856728487],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", 
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "s"}]}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
         UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "s"}]},
       {
        RowBox[{
         UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
         RowBox[{
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
          "c"}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR2 monomial 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:335657318", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633235681356304*^9, 3.63323572157708*^9}, {
  3.6391213246374493`*^9, 3.639121325718007*^9}},
 CellTags->{
  "ID:335657318", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 3"},
 CellID->335657318],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->575968727]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1580711919],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR1 monomial 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR1 monomial 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1499935354"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1499935354-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1499935354-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1499935354", "Proof|ID:1499935354-1"},
 CellID->1869006459],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1499935354-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->1447063924]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR1 monomial 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR1 monomial 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
                 RowBox[{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{"(", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
                 ")"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
                 RowBox[{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{"(", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
                 ")"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1938671644"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1938671644-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1938671644-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1938671644", "Proof|ID:1938671644-1"},
 CellID->1781620971],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1938671644-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.642738567788177*^9, 3.6427385691454153`*^9}},
 CellID->1479477140]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR1 monomial 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR1 monomial 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["s", "ExpressionVariable"]}], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["s", "ExpressionVariable"]}], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:48476485"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p48476485-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p48476485-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:48476485", "Proof|ID:48476485-1"},
 CellID->528453611],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p48476485-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2072605463]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR2 monomial 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR2 monomial 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["t", "ExpressionVariable"]}]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1246936946"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1246936946-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1246936946-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1246936946", "Proof|ID:1246936946-1"},
 CellID->299300275],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1246936946-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->661794959]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR2 monomial 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR2 monomial 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], "]"}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], "]"}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1856728487"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1856728487-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1856728487-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1856728487", "Proof|ID:1856728487-1"},
 CellID->1800537912],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1856728487-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->434287860]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR2 monomial 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR2 monomial 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["s", "ExpressionVariable"]}]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["s", "ExpressionVariable"]}]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   StyleBox["s", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:335657318"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p335657318-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p335657318-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:335657318", "Proof|ID:335657318-1"},
 CellID->218427358],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p335657318-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->948824583]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR1 monomial 1)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6327204604755363`*^9, 3.632720461773663*^9}, 
   3.6327217579035463`*^9, {3.63272189932273*^9, 3.632721903925253*^9}, {
   3.6327265531968517`*^9, 3.632726553382667*^9}, 3.6364401982805147`*^9, 
   3.643446649657487*^9, 3.6532080708634605`*^9, 3.65321365214408*^9, 
   3.6661738951485467`*^9, 3.666440497962132*^9},
 CellID->601702509],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632721398462624*^9, 3.632721399542478*^9}, 
   3.632721757903695*^9, 3.636440198280676*^9, 3.643446649657645*^9, 
   3.6532080708634605`*^9, 3.65321365214408*^9, 3.6661738951485467`*^9, 
   3.6664404979622726`*^9},
 CellID->761991808],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{"a", ",", "c", ",", "m"}], "]"}], "\[Implies]", 
    RowBox[{"a", "\[NotEqual]", "0"}]}], TraditionalForm]]],
 " (Reduction Ring lemmas)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6327214036673307`*^9, 3.63272141406916*^9}, {
   3.632721515654294*^9, 3.632721515664317*^9}, 3.632721757904187*^9, {
   3.633174885564907*^9, 3.633174891999745*^9}, 3.63644019828082*^9, 
   3.643446649657748*^9, 3.6532080708634605`*^9, 3.65321365214408*^9, 
   3.6661738951485467`*^9, 3.666440497962385*^9},
 CellID->1823066290],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"c", "\[CenterDot]", "t"}], ",", "s"}], "]"}], "=", 
    RowBox[{"0", "\[DoubleLeftRightArrow]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"s", "\[NotEqual]", "t"}], "\[Or]", "c"}], "=", "0"}], 
      ")"}]}]}], TraditionalForm]]],
 " (Monomials)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6327214036673307`*^9, 3.63272141406916*^9}, {
   3.632721515654294*^9, 3.63272155350745*^9}, 3.6327217579042673`*^9, {
   3.633174987241702*^9, 3.633174989400858*^9}, 3.636440198280917*^9, 
   3.643446649657824*^9, 3.6532080708634605`*^9, 3.65321365214408*^9, 
   3.6661738951485467`*^9, 3.666440497962462*^9},
 CellID->1830447192],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "a"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"isReducible", "[", 
       RowBox[{
        RowBox[{"C", "[", 
         RowBox[{"a", ",", "t"}], "]"}], ",", 
        RowBox[{"lc", "[", "q", "]"}]}], "]"}], "\[And]", 
      RowBox[{"isNTCR", "[", 
       RowBox[{"p", ",", "q", ",", 
        RowBox[{
         RowBox[{"C", "[", 
          RowBox[{"a", ",", "t"}], "]"}], "\[CenterDot]", "t"}], ",", "i", 
        ",", "j"}], "]"}]}], ")"}]}], TraditionalForm]]],
 "\nFrom (1) and (CR1.\[Implies].2), by modus ponens, we can infer\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "u"}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "u"}], TraditionalForm]]],
 "\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "u", "]"}]}], TraditionalForm]]],
 ". We claim that ",
 Cell[BoxData[
  FormBox["u", TraditionalForm]]],
 " witnesses our existentially quantified goal (G), so, by (CR1.\
\[DoubleLeftArrow]), we have to show\n(G0)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}]}], "]"}], TraditionalForm]]],
 "\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", 
         RowBox[{"s", ",", "t"}], "]"}]}]], 
      RowBox[{"s", "\[NotEqual]", "t"}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {
        RowBox[{
         RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "s"}]},
       {
        RowBox[{"areCommonReducibles", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"C", "[", 
              RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", 
            "t"}], "]"}], ",", 
          RowBox[{"C", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"C", "[", 
              RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", 
            "s"}], "]"}], ",", "p", ",", "q", ",", "i", ",", "j"}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
        {
         RowBox[{
          RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "t"}]},
        {
         RowBox[{"isNTCRP", "[", 
          RowBox[{"p", ",", "q", ",", 
           RowBox[{"C", "[", 
            RowBox[{
             RowBox[{
              RowBox[{"C", "[", 
               RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", 
             "t"}], "]"}], ",", "i", ",", "j"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\n\n(G0):\nFrom (4), by (isNTCRP alternative), we can infer ",
 Cell[BoxData[
  FormBox[
   RowBox[{"{", GridBox[{
      {
       RowBox[{"isNTCR", "[", 
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", 
         RowBox[{"C", "[", 
          RowBox[{"a", ",", "u"}], "]"}], ",", "i", ",", "j"}], "]"}], 
       "\[DoubleLongLeftArrow]", 
       RowBox[{"p", "=", 
        RowBox[{"q", "\[Or]", 
         RowBox[{
          RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
          RowBox[{"lc", "[", "q", "]"}]}]}]}]},
      {
       RowBox[{"isNTCR", "[", 
        RowBox[{
         RowBox[{"lc", "[", "q", "]"}], ",", 
         RowBox[{"C", "[", 
          RowBox[{"a", ",", "u"}], "]"}]}], "]"}], "\[DoubleLongLeftArrow]", 
       "otherwise"}
     }]}], TraditionalForm]]],
 ". In any case we can further infer ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}]}], "]"}], TraditionalForm]]],
 ", so we are done.\n\n(G1):\nWe move the negation inside the quantifier, so \
we have to show\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", 
        RowBox[{"s", ",", "t"}], "]"}]}]], 
     RowBox[{"s", "\[NotEqual]", "t"}]], 
    RowBox[{"\[Or]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{"\[Not]", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]}]},
        {
         RowBox[{"\[Not]", 
          RowBox[{
           RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "s"}]}]},
        {
         RowBox[{"\[Not]", 
          RowBox[{"areCommonReducibles", "[", 
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{
              RowBox[{
               RowBox[{"C", "[", 
                RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", 
              "t"}], "]"}], ",", 
            RowBox[{"C", "[", 
             RowBox[{
              RowBox[{
               RowBox[{"C", "[", 
                RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", 
              "s"}], "]"}], ",", "p", ",", "q", ",", "i", ",", "j"}], 
           "]"}]}]}
       }]}]}]}], TraditionalForm]]],
 "\nWe choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"s", ",", "t"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", "\[NotEqual]", "t"}], TraditionalForm]]],
 "\nand prove the last disjunct (we do not even need the negations of the \
others), so we prove\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"areCommonReducibles", "[", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"C", "[", 
          RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", "t"}], 
       "]"}], ",", 
      RowBox[{"C", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"C", "[", 
          RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", "s"}], 
       "]"}], ",", "p", ",", "q", ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 "\nWe assume the negation of (G4) and prove the negation of (5), therefore, \
by the definition of areCommonReducibles, we know that there exist ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], TraditionalForm]]],
 " with\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"C", "[", 
         RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", "t"}], 
      "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"C", "[", 
         RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", "s"}], 
      "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\nand show\n",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", "=", "t"}], TraditionalForm]]],
 ".\n\nFrom (6) and (7) we can infer, by a lemma\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", "t"}], 
     "]"}], "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", "s"}], 
     "]"}], "\[NotEqual]", "0"}], TraditionalForm]]],
 "\nand further, by another lemma,\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", "u"}], TraditionalForm]]],
 "\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", "=", "u"}], TraditionalForm]]],
 "\nso we are done.\n\n(G2):\nWe claim that ",
 Cell[BoxData[
  FormBox["u", TraditionalForm]]],
 " witnesses the existential quantifier in (G2), so we have to show\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "u"}], TraditionalForm]]],
 "\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "u"}], TraditionalForm]]],
 "\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"C", "[", 
         RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", "u"}], 
      "]"}], ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\n(G5) and (G6) are precisely what (2) and (3) state, and for (G7) one only \
has to note that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", "u"}], 
     "]"}], "=", 
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "u"}], "]"}]}], TraditionalForm]]],
 ", so we are done by using (4)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632720468539443*^9, 3.632720570065999*^9}, {
   3.6327206200139933`*^9, 3.6327210537692537`*^9}, {3.63272108511773*^9, 
   3.632721376830204*^9}, {3.6327214264274073`*^9, 3.632721509175544*^9}, {
   3.632721569655018*^9, 3.632721757904365*^9}, {3.6327250856775*^9, 
   3.632725095097095*^9}, {3.632725152409679*^9, 3.632725296912952*^9}, {
   3.632725360562213*^9, 3.6327254021279507`*^9}, 3.636440198281042*^9, {
   3.639111519845763*^9, 3.63911152796131*^9}, 3.6434466496579514`*^9, 
   3.6532080708634605`*^9, 3.65321365214408*^9, 3.6661738951485467`*^9, 
   3.666440497962574*^9},
 CellID->1138449629],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632721745021137*^9, 3.632721757904491*^9}, 
   3.636440198281207*^9, 3.643446649658104*^9, 3.6532080708634605`*^9, 
   3.65321365214408*^9, 3.6661738951485467`*^9, 3.666440497962709*^9},
 TextAlignment->Right,
 CellID->72163444]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR1 monomial 2)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632726568738647*^9, 3.63272657485672*^9}, 
   3.632730559509733*^9, 3.6364402043411903`*^9, {3.639120391084*^9, 
   3.639120399412847*^9}, 3.643446655456045*^9, 3.653208064311449*^9, 
   3.6532136435474653`*^9, 3.666173889423337*^9, 3.66644049232032*^9},
 CellID->693637141],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6327296312163353`*^9, 3.632729632407086*^9}, 
   3.632730559509857*^9, 3.636440204341308*^9, {3.639120391084127*^9, 
   3.6391203994129877`*^9}, 3.643446655456164*^9, 3.653208064311449*^9, 
   3.6532136435474653`*^9, 3.666173889423337*^9, 3.666440492320421*^9},
 CellID->2031424469],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{"a", ",", "c", ",", "m"}], "]"}], "\[Implies]", 
    RowBox[{"a", "\[NotEqual]", "0"}]}], TraditionalForm]]],
 " (Reduction Ring lemmas)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632729639846016*^9, 3.6327296601016073`*^9}, 
   3.632730559510333*^9, {3.633174998033642*^9, 3.633175001674912*^9}, 
   3.63644020434142*^9, {3.639113085013207*^9, 3.639113108302495*^9}, {
   3.639113238202977*^9, 3.639113238213616*^9}, {3.639113359843184*^9, 
   3.6391133602167187`*^9}, {3.639120391084244*^9, 3.639120399413127*^9}, 
   3.643446655456245*^9, 3.653208064311449*^9, 3.6532136435474653`*^9, 
   3.666173889423337*^9, 3.666440492320506*^9},
 CellID->632880062],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"C", "[", 
    RowBox[{"a", ",", 
     RowBox[{"lp", "[", "a", "]"}]}], "]"}], "=", 
   RowBox[{"lc", "[", "a", "]"}]}], TraditionalForm]]]], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632729639846016*^9, 3.6327296601016073`*^9}, 
   3.632730559510333*^9, {3.633174998033642*^9, 3.633175001674912*^9}, 
   3.63644020434142*^9, {3.639113085013207*^9, 3.639113108302495*^9}, {
   3.639113238202977*^9, 3.6391132461830797`*^9}, {3.639120391084317*^9, 
   3.6391203994132147`*^9}, 3.643446655456304*^9, 3.653208064311449*^9, 
   3.6532136435474653`*^9, 3.666173889423337*^9, 3.666440492320566*^9},
 CellID->1560665101],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "a"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(0)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "a", "]"}], TraditionalForm]]],
 "\nFrom (0) we immediately get ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 ".\nWe prove both directions of the equivalence.\n\n\[OpenCurlyDoubleQuote]\
\[DoubleLongRightArrow]\[CloseCurlyDoubleQuote]:\nAssume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 " and show the big conjunction. Therefore, by (CR1.\[Implies].2) we can \
infer the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 " with\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "t"}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nIf we can show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 ", our goal follows readily from (1), (2) and (3), so we show now\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\n\nFrom (3), by the definition of isNTCRP, we can infer (among others)\n\
(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", "p", ",", "q", ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\nand further\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "m"}], "]"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "m", "]"}]}], TraditionalForm]]],
 ". From (5), together with a lemma, we can infer\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "t"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\nDue to a lemma on monomials this is only possible if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 ", hence we are done.\n\n\[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\
\[CloseCurlyDoubleQuote]:\nAssume\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"lc", "[", "a", "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 ", which, by (CR1.\[DoubleLeftArrow]), amounts to showing both\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", 
         RowBox[{"s", ",", "t"}], "]"}]}]], 
      RowBox[{"s", "\[NotEqual]", "t"}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {
        RowBox[{
         RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "s"}]},
       {
        RowBox[{"areCommonReducibles", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "t"}], "]"}], ",", 
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "s"}], "]"}], ",", "p", ",", "q", ",", "i", ",", 
          "j"}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
        {
         RowBox[{
          RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "t"}]},
        {
         RowBox[{"isNTCRP", "[", 
          RowBox[{"p", ",", "q", ",", 
           RowBox[{"C", "[", 
            RowBox[{"a", ",", "t"}], "]"}], ",", "i", ",", "j"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\nIt is apparent that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lp", "[", "a", "]"}], TraditionalForm]]],
 " witnesses (G3), so we only consider (G2) now. We move the negation inside \
the quantifier in (G2), choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"s", ",", "t"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", "\[NotEqual]", "t"}], TraditionalForm]]],
 "\nand show\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"areCommonReducibles", "[", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", "t"}], "]"}], ",", 
      RowBox[{"C", "[", 
       RowBox[{"a", ",", "s"}], "]"}], ",", "p", ",", "q", ",", "i", ",", 
      "j"}], "]"}]}], TraditionalForm]]],
 "\nHowever, rather than showing (G4) we assume its negation and show the \
negation of (10) instead, i.e. we know\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCommonReducibles", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "s"}], "]"}], ",", "p", ",", "q", ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\nand show\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", "=", "t"}], TraditionalForm]]],
 "\n\nFrom (11), by the definition of areCommonReducibles, we can infer the \
existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], TraditionalForm]]],
 " with (among others)\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "t"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "c1"}], "]"}], TraditionalForm]]],
 "\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "s"}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c2"}], "]"}], TraditionalForm]]],
 "\nand further\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "t"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\n(15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "s"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\n(14) and (15) can only be satisfied if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{"s", "=", 
     RowBox[{"lp", "[", "a", "]"}]}]}], TraditionalForm]]],
 ", so we are done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632726589216587*^9, 3.632726731109131*^9}, {
   3.632729154795041*^9, 3.632729161074757*^9}, {3.632729244804076*^9, 
   3.632729524552723*^9}, {3.632729555286372*^9, 3.63272958320813*^9}, {
   3.632729622213254*^9, 3.6327296261335487`*^9}, {3.632729664673018*^9, 
   3.632729735391952*^9}, {3.6327298976710443`*^9, 3.632729984363728*^9}, {
   3.632730045990798*^9, 3.6327301979375763`*^9}, {3.632730233916316*^9, 
   3.632730526420403*^9}, 3.632730559510437*^9, 3.632735127742502*^9, 
   3.636440204341515*^9, {3.639113045882421*^9, 3.639113101863563*^9}, 
   3.6391133541614313`*^9, {3.639120391084414*^9, 3.6391203994133263`*^9}, 
   3.639120468047805*^9, 3.643446655456398*^9, 3.653208064311449*^9, 
   3.6532136435474653`*^9, 3.666173889423337*^9, 3.6664404923206472`*^9},
 CellID->1096597969],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632730546805463*^9, 3.632730559510564*^9}, 
   3.636440204341631*^9, {3.63912039108454*^9, 3.6391203994134607`*^9}, 
   3.643446655456517*^9, 3.653208064311449*^9, 3.6532136435474653`*^9, 
   3.666173889423337*^9, 3.666440492320754*^9},
 TextAlignment->Right,
 CellID->979676413]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Alternative Descriptions of isMNTCR and areCPMultipliers (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.632216318190495*^9, 3.632216318858843*^9}, {
  3.632721986829443*^9, 3.632721997413164*^9}, {3.639207842938622*^9, 
  3.639207847639167*^9}, {3.643443735306717*^9, 3.643443739753517*^9}},
 CellID->289777292],

Cell["", "OpenEnvironment",
 CellID->1231465134],

Cell[CellGroupData[{

Cell["Definition: isMNTCRP", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.63272207821143*^9, 3.6327220810822887`*^9}},
 CellID->1786858939],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
     RowBox[{"[", "a", "]"}]}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", ",", 
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
      TagBox[
       RowBox[{
       ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
        "\[DoubleLongLeftRightArrow]"}],
       Identity,
       SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
         RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
        "\[And]", 
        RowBox[{
         UnderscriptBox[
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
            RowBox[{"[", "a0", "]"}]}]], 
          RowBox[{"a", 
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], "a0"}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"\[Not]", 
           RowBox[{
            UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
            RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMNTCRP)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:624517207", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632722096142501*^9, 3.632722098353513*^9}, {
   3.632722149470807*^9, 3.6327222930879717`*^9}, {3.639122827258057*^9, 
   3.639122829100909*^9}, 3.639127365083288*^9},
 CellTags->{
  "ID:624517207", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMNTCRP"},
 CellID->624517207],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1789189429]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2019724429],

Cell["\<\
The proof of (isMNTCRP alternative) is absolutely straight-forward: One only \
has to use (isNTCRP alternative).\
\>", "Text",
 CellChangeTimes->{{3.632724065958692*^9, 3.632724212176649*^9}, {
  3.639127671190593*^9, 3.639127689626422*^9}, {3.6391287560480022`*^9, 
  3.6391287790159063`*^9}},
 CellID->840274992],

Cell["", "OpenEnvironment",
 CellID->967326008],

Cell[CellGroupData[{

Cell["Lemma: Alternative Description of isMNTCRP", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.63272231339823*^9, 3.632722322751223*^9}},
 CellID->147093162],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"p", ",", "q"}], "]"}]}]], 
   RowBox[{
    RowBox[{"p", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
    RowBox[{"q", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
     RowBox[{"[", "a", "]"}]}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", ",", 
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
      "\[DoubleLongLeftRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
            RowBox[{
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",",
              "a", ",", "i", ",", "j"}], "]"}], "\[DoubleLeftArrow]", 
           RowBox[{
            RowBox[{"p", "\[Equal]", "q"}], "\[Or]", 
            RowBox[{
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
             "\[NotEqual]", 
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
              "]"}]}]}]},
          {
           RowBox[{
            UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
            RowBox[{
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",",
              "a"}], "]"}], "\[DoubleLeftArrow]", "otherwise"}
         }]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMNTCRP alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1172537415", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6327223291707172`*^9, 3.632722391510103*^9}, {
  3.6327224478645563`*^9, 3.632722478840075*^9}, {3.642928125740922*^9, 
  3.642928188150714*^9}},
 CellTags->{
  "ID:1172537415", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMNTCRP alternative"},
 CellID->1172537415],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->462144833]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->866823809],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMNTCRP alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMNTCRP alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                  "\[DoubleLeftArrow]", 
                  RowBox[{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}]}, {
                  RowBox[{
                    UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["a", "ExpressionVariable"]}], "]"}], 
                  "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                  "\[DoubleLeftArrow]", 
                  RowBox[{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}]}, {
                  RowBox[{
                    UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["a", "ExpressionVariable"]}], "]"}], 
                  "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1172537415"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1172537415-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1172537415-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1172537415", "Proof|ID:1172537415-1"},
 CellID->1389429208],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1172537415-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1096557679]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (isMNTCRP alternative)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6532136077912025`*^9, 3.653213631706044*^9}, 
   3.6661738785345173`*^9, 3.666440484240635*^9},
 CellID->297887619],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "a", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(0)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Or]", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
      RowBox[{"lc", "[", "q", "]"}]}]}]}], TraditionalForm]]],
 "\nand show both directions of the equivalence.\n\n\[OpenCurlyDoubleQuote]\
\[DoubleLongRightArrow]\[CloseCurlyDoubleQuote]:\nAssume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 " and show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 ". By the definitions of isMNTCRP and isMNTCR we know\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"a", "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCRP", "[", 
        RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}]}], 
     ")"}]}], TraditionalForm]]],
 "\nand we have to show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"a", "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isMNTCR", "[", 
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "a0", ",", "i", ",", "j"}], 
        "]"}]}]}], ")"}]}], TraditionalForm]]],
 "\n\n(G1):\nFrom (0) and (1), together with (isNTCRP alternative 1), we can \
infer\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nwhich is exactly what we have to show.\n\n(G2):\nChoose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "a0", "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", "a0"}], TraditionalForm]]],
 "\nand show\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "a0", ",", "i", ",", "j"}], "]"}]}],
    TraditionalForm]]],
 "\nWe can instantiate (2) with ",
 Cell[BoxData[
  FormBox["a0", TraditionalForm]]],
 ", and since the premise of the implication holds due to (4), we can infer\n\
(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCRP", "[", 
     RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 "\nRather than showing (G3) we assume its negation and show the negation of \
(5) instead, so we know\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a0", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand show\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nBy (isNTCRP alternative 1) this amounts to showing\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a0", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand from the definition of isMNTCR we know (among others)\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a0", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nhence we are done.\n\n\[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\
\[CloseCurlyDoubleQuote]:\nAssume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 " and show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 ". By the definitions of isMNTCRP and isMNTCR we know\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"a", "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCR", "[", 
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "a0", ",", "i", ",", "j"}], 
        "]"}]}]}], ")"}]}], TraditionalForm]]],
 "\nand we have to show\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"a", "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isMNTCRP", "[", 
        RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}]}], 
     ")"}]}], TraditionalForm]]],
 "\n\n(G6):\nBy (isNTCRP alternative 1) and (0) we can reduce our goal (G6) \
to\n(G8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nwhich is exactly what (8) states.\n\n(G7):\nChoose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "a0", "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", "a0"}], TraditionalForm]]],
 "\nand show\n(G9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isMNTCRP", "[", 
     RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 "\nWe can instantiate (9) with ",
 Cell[BoxData[
  FormBox["a0", TraditionalForm]]],
 ", and since the premise of the implication holds due to (10), we can infer\n\
(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "a0", ",", "i", ",", "j"}], "]"}]}],
    TraditionalForm]]],
 "\nRather than showing (G9) we assume its negation and show the negation of \
(11) instead, so we know\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand show\n(G10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a0", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nFrom (12), by the definition of isMNTCRP, we can infer (among others)\n\
(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand further, by (0) and (isNTCRP alternative 1),\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a0", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nwhich is exactly what we have to show."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632722534313435*^9, 3.632722542779703*^9}, 
   3.6327236757764893`*^9, 3.636440238099897*^9, {3.639128749111656*^9, 
   3.639128749260387*^9}, 3.643446661952341*^9, 3.653208054499032*^9, {
   3.6532135912863736`*^9, 3.653213631706044*^9}, 3.6661738785345173`*^9, 
   3.666440484240741*^9},
 CellID->690232938],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.653213621659627*^9, 3.653213631706044*^9}, 
   3.6661738785345173`*^9, 3.666440484240837*^9},
 TextAlignment->Right,
 CellID->446746978]
}, Closed]],

Cell[TextData[{
 StyleBox["Remark",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 ": For proving the remaining formulas in this subsection we have to use the \
fact that (R4) holds in \[ScriptCapitalP] (which we have already shown)."
}], "Text",
 CellChangeTimes->{{3.6327248393298073`*^9, 3.6327248753580303`*^9}, {
  3.633168712102346*^9, 3.633168735970255*^9}},
 CellID->2030148506],

Cell["\<\
It could be that the directions \[OpenCurlyDoubleQuote]\[Implies]\
\[CloseCurlyDoubleQuote] are not needed at all, both for (CR3) and (CR4).\
\>", "Text",
 CellChangeTimes->{{3.633236170991242*^9, 3.633236204078512*^9}},
 CellID->1426279550],

Cell["", "OpenEnvironment",
 CellID->1983351237],

Cell[CellGroupData[{

Cell["\<\
Theorem: Alternative Description of binary isMNTCR in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632724230213929*^9, 3.632724268116845*^9}, {
  3.6332358333830557`*^9, 3.633235834393815*^9}, {3.6429261318685303`*^9, 
  3.642926168394931*^9}},
 CellID->2124804766],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "a"}], "]"}]}]], 
  RowBox[{
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1861685863}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631963023902554*^9, 3.631963056276445*^9}, {
  3.6322145639880133`*^9, 3.6322145642449713`*^9}, {3.635748087131172*^9, 
  3.635748090805043*^9}},
 CellTags->{
  "ID:1861685863", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1861685863],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1154885365}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632724286039855*^9, 3.6327243010842953`*^9}},
 CellTags->{
  "ID:1154885365", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1154885365],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "a", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR3.\[DoubleRightArrow].1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1777331368", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632724308692295*^9, 3.6327243258697033`*^9}},
 CellTags->{
  "ID:1777331368", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].1"},
 CellID->1777331368],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "a", "]"}], "\[Equal]", 
    RowBox[{
     UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], "]"}]}], 
   
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR3.\[DoubleRightArrow].2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1052103130", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632725919761766*^9, 3.632725939415354*^9}},
 CellTags->{
  "ID:1052103130", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].2"},
 CellID->1052103130],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", 
    RowBox[{
     UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "a", "]"}], ",", "i", 
    ",", "j"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR3.\[DoubleRightArrow].3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1077684208", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6327259827170057`*^9, 3.632726005869355*^9}},
 CellTags->{
  "ID:1077684208", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].3"},
 CellID->1077684208],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "q", ",", "c", ",", "i", ",", "j"}], "]"}], 
    "\[Implies]", 
    RowBox[{
     UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "q", ",", 
      RowBox[{"c", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
       RowBox[{
        UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
        "]"}]}], ",", "i", ",", "j"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR3.\[DoubleLeftArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:86200506", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632732005738535*^9, 3.632732077364216*^9}},
 CellTags->{
  "ID:86200506", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleLeftArrow]"},
 CellID->86200506],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", "c", "]"}]}]], 
     RowBox[{"c", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "q", ",", "c", ",", "i", ",", "j"}], "]"}], "\[And]", 
      RowBox[{"a", "\[Equal]", 
       RowBox[{"c", 
        UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
        RowBox[{
         UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
         RowBox[{
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
          RowBox[{
           UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
         "]"}]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:394014296", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6429255232000628`*^9, 3.642925587750566*^9}},
 CellTags->{
  "ID:394014296", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3"},
 CellID->394014296],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->333024971]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1951750829],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR3.\[DoubleRightArrow].1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR3.\[DoubleRightArrow].1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["a", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["a", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1777331368"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1777331368-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1777331368-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1777331368", "Proof|ID:1777331368-1"},
 CellID->508661436],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1777331368-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->263326735]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR3.\[DoubleRightArrow].2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR3.\[DoubleRightArrow].2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1052103130"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1052103130-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1052103130-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1052103130", "Proof|ID:1052103130-1"},
 CellID->1682055148],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1052103130-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->187463426]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR3.\[DoubleRightArrow].3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR3.\[DoubleRightArrow].3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1077684208"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1077684208-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1077684208-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1077684208", "Proof|ID:1077684208-1"},
 CellID->1363636622],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1077684208-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1445163384]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR3.\[DoubleLeftArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR3.\[DoubleLeftArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
               ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
               ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:86200506"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p86200506-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p86200506-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:86200506", "Proof|ID:86200506-1"},
 CellID->1645380276],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p86200506-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->1969346073]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:394014296"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p394014296-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p394014296-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:394014296", "Proof|ID:394014296-1"},
 CellID->1823310355],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p394014296-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->504163934]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR3.\[Implies].1)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632724361410532*^9, 3.6327243676094503`*^9}, 
   3.632725904162763*^9, 3.636440250646221*^9, 3.643446668931285*^9, 
   3.653208045263816*^9, 3.6532135837359605`*^9, 3.6661738696275015`*^9, 
   3.666440476082081*^9},
 CellID->1024574146],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6327248899737053`*^9, 3.6327248912985277`*^9}, 
   3.6327259041628857`*^9, 3.63644025064638*^9, 3.643446668931465*^9, 
   3.653208045263816*^9, 3.6532135837359605`*^9, 3.6661738696275015`*^9, 
   3.66644047608222*^9},
 CellID->101928936],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{"a", ",", "c"}], "]"}], "\[Implies]", 
    RowBox[{"a", "\[NotEqual]", "0"}]}], TraditionalForm]]],
 " (in general in domains where (R4) holds) (Reduction ring lemmas)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632724896382174*^9, 3.6327249274949627`*^9}, 
   3.632725904163352*^9, {3.633175041979538*^9, 3.63317504604731*^9}, 
   3.636440250646522*^9, 3.643446668931588*^9, 3.653208045263816*^9, 
   3.6532135837359605`*^9, 3.6661738696275015`*^9, 3.6664404760823317`*^9},
 CellID->548195680],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "a"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "a", "]"}], TraditionalForm]]],
 "\nWe expand the definition of isMNTCR: We know\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"a", "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCR", "[", 
        RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}]}], 
     ")"}]}], TraditionalForm]]],
 "\nFrom a lemma we know that ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " can either be 0, a monomial, or a polynomial with more than one \
power-products with non-zero coefficients. Thus, we distinguish three cases.\n\
\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", "0"}], TraditionalForm]]],
 ".\nFrom (2) we can infer, among others,\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCommonReducible", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand further\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p"}], "]"}], TraditionalForm]]],
 "\nHence, from a lemma we immediately get ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", contradicting our case assumption.\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "a", "]"}], TraditionalForm]]],
 ".\nThis is exactly what we have to show.\n\nCase III: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", 
       RowBox[{"s", ",", "t"}], "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"s", "\[NotEqual]", "t"}], "\[And]", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "s"}], "]"}], "\[NotEqual]", "0"}], "\[And]", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "t"}], "]"}], "\[NotEqual]", "0"}]}], ")"}]}], 
   TraditionalForm]]],
 "\nLet ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"s", ",", "t"}], "]"}]}], TraditionalForm]]],
 " witness the existential case assumption, meaning we know\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", "\[NotEqual]", "t"}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "s"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "t"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\n\nNow we can use (CR1 monomial 1): From (2) and (CR1 monomial 1) we can \
infer the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "u", "]"}]}], TraditionalForm]]],
 " with\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}]}], "]"}], TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\nFrom (9), by a lemma, we immediately get\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "u"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\nBecause of (6) ",
 Cell[BoxData[
  FormBox["u", TraditionalForm]]],
 " cannot be equal both ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 ", so we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"u", "\[NotEqual]", "t"}], "\[Or]", 
    RowBox[{"u", "\[NotEqual]", "s"}]}], TraditionalForm]]],
 ", leading to a case distinction; However, since both cases are completely \
symmetric, we only consider the first one with\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"u", "\[NotEqual]", "t"}], TraditionalForm]]],
 "\nFrom a lemma on \[Succeeds] in \[ScriptCapitalP] we know (because of (8) \
and (11))\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}]}], 
   TraditionalForm]]],
 "\n\nNow we instantiate (3) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a0", ":=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "t"}]}], 
   TraditionalForm]]],
 ". The premise of the implication holds due to (12), so we can infer\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{"p", ",", "q", ",", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", "i", ",", 
      "j"}], "]"}]}], TraditionalForm]]],
 "\nin contradiction to (10)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632724374386097*^9, 3.632724499997056*^9}, {
   3.6327245402374687`*^9, 3.632724660357985*^9}, {3.6327246912755623`*^9, 
   3.632724717827095*^9}, {3.6327247485967607`*^9, 3.632724757555394*^9}, {
   3.6327249377957487`*^9, 3.632725054533145*^9}, {3.6327254396943703`*^9, 
   3.6327254423216467`*^9}, {3.6327254792405367`*^9, 3.632725755514175*^9}, {
   3.6327258078581343`*^9, 3.632725904163457*^9}, 3.6364402506466503`*^9, {
   3.6391288741549063`*^9, 3.6391288796950006`*^9}, {3.639128964612129*^9, 
   3.639128969167994*^9}, 3.643446668931731*^9, 3.653208045263816*^9, 
   3.6532135837359605`*^9, 3.6661738696275015`*^9, 3.666440476082446*^9},
 CellID->1217971235],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632725892101275*^9, 3.632725904163587*^9}, 
   3.6364402506468143`*^9, 3.643446668931888*^9, 3.653208045263816*^9, 
   3.6532135837359605`*^9, 3.6661738696275015`*^9, 3.666440476082581*^9},
 TextAlignment->Right,
 CellID->313548928]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR3.\[Implies].2)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6327260415510883`*^9, 3.63272604884029*^9}, 
   3.6327326472389174`*^9, 3.6364402561220837`*^9, 3.643446674830268*^9, 
   3.653208038009803*^9, 3.6532135773399487`*^9, 3.6661738640582914`*^9, 
   3.666440470303149*^9},
 CellID->74747143],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632731393196661*^9, 3.632731394401993*^9}, 
   3.632732647239039*^9, 3.636440256122204*^9, 3.643446674830387*^9, 
   3.653208038009803*^9, 3.6532135773399487`*^9, 3.6661738640582914`*^9, 
   3.6664404703032503`*^9},
 CellID->1147238182],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"s", "\[VerticalBar]", "t"}], "\[Implies]", 
    RowBox[{"\[Not]", 
     RowBox[{"s", "\[Succeeds]", "t"}]}]}], TraditionalForm]]],
 " (ET CPP)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6327313988256197`*^9, 3.632731412965679*^9}, 
   3.63273264723953*^9, {3.63317505401931*^9, 3.633175056017473*^9}, 
   3.636440256122321*^9, 3.643446674830467*^9, 3.653208038009803*^9, 
   3.6532135773399487`*^9, 3.6661738640582914`*^9, 3.666440470303334*^9},
 CellID->933263416],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "a"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "a", "]"}], "=", 
    RowBox[{"lcm", "[", 
     RowBox[{
      RowBox[{"lp", "[", "p", "]"}], ",", 
      RowBox[{"lp", "[", "q", "]"}]}], "]"}]}], TraditionalForm]]],
 "\nFrom (CR3.\[Implies].1) and (1) we can infer\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "a", "]"}], TraditionalForm]]],
 "\nand further\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n\nFrom (1), by the definition of isMNTCR, we know\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"a", "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCR", "[", 
        RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}]}], 
     ")"}]}], TraditionalForm]]],
 "\n\nFrom (2), (4) and (CR1 monomial 2) we know\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"lc", "[", "a", "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n\nSince ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], TraditionalForm]]],
 " is total, we distinguish now three cases, based on the relative order of \
",
 Cell[BoxData[
  FormBox[
   RowBox[{"lp", "[", "a", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "[", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], ",", 
     RowBox[{"lp", "[", "q", "]"}]}], "]"}], TraditionalForm]]],
 ".\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "a", "]"}], "=", 
    RowBox[{"lcm", "[", 
     RowBox[{
      RowBox[{"lp", "[", "p", "]"}], ",", 
      RowBox[{"lp", "[", "q", "]"}]}], "]"}]}], TraditionalForm]]],
 ".\nThis is exactly what we want to prove.\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lcm", "[", 
     RowBox[{
      RowBox[{"lp", "[", "p", "]"}], ",", 
      RowBox[{"lp", "[", "q", "]"}]}], "]"}], "\[Succeeds]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 ".\nNow we have to use the fact that \[ScriptCapitalT] is a power-product \
domain, and in particular that ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["lcm", "\[ScriptCapitalT]"], TraditionalForm]]],
 " has certain properties, namely ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"x", "\[VerticalBar]", "z"}], "\[And]", 
     RowBox[{"y", "\[VerticalBar]", "z"}]}], "\[Implies]", 
    RowBox[{
     RowBox[{"lcm", "[", 
      RowBox[{"x", ",", "y"}], "]"}], "\[VerticalBar]", "z"}]}], 
   TraditionalForm]]],
 "; From this, together with (6) and (7) we can infer\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lcm", "[", 
     RowBox[{
      RowBox[{"lp", "[", "p", "]"}], ",", 
      RowBox[{"lp", "[", "q", "]"}]}], "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\nUsing a lemma, (9) contradicts our case assumption.\n\nCase III: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "a", "]"}], "\[Succeeds]", 
    RowBox[{"lcm", "[", 
     RowBox[{
      RowBox[{"lp", "[", "p", "]"}], ",", 
      RowBox[{"lp", "[", "q", "]"}]}], "]"}]}], TraditionalForm]]],
 ".\nWe make the following claim (let in the sequel lcm abbreviate ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "[", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], ",", 
     RowBox[{"lp", "[", "q", "]"}]}], "]"}], TraditionalForm]]],
 "):\n(C)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}], ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\n\nProof of (C):\nClearly we have ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", 
    RowBox[{
     RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}], "]"}], 
   TraditionalForm]]],
 ", hence we can use (CR1 monomial 2) to reduce the goal to\n(C1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}], "]"}]}], 
   TraditionalForm]]],
 "\n(C2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}], "]"}]}], 
   TraditionalForm]]],
 "\n(C3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"lc", "[", 
      RowBox[{
       RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}], "]"}], ",", 
     "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\nApparently, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}], "]"}], "=", 
    "lcm"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}], "]"}], "=", 
    RowBox[{"lc", "[", "a", "]"}]}], TraditionalForm]]],
 ", so we only have to show\n(C1\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "lcm"}], 
   TraditionalForm]]],
 "\n(C2\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "lcm"}], 
   TraditionalForm]]],
 "\n(C3\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"lc", "[", "a", "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(C3\[CloseCurlyQuote]) is precisely what (8) states, and (CR1\
\[CloseCurlyQuote]) and (CR2\[CloseCurlyQuote]) follow readily from \
\[ScriptCapitalT] being a power-product domain.\n\nSummarizing, we know\n(10)\
\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}], ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\n\nNow we instantiate (5) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a0", ":=", 
    RowBox[{
     RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}]}], 
   TraditionalForm]]],
 ", so we know\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}]}], "\[Implies]", 
    
    RowBox[{"\[Not]", 
     RowBox[{"isNTCR", "[", 
      RowBox[{"p", ",", "q", ",", 
       RowBox[{
        RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}], ",", "i", ",",
        "j"}], "]"}]}]}], TraditionalForm]]],
 "\nFrom (11) and (10), by modus tollens, we can infer ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}]}]}], 
   TraditionalForm]]],
 ". However, from our case assumption, (382\[CloseCurlyQuote]) and the fact \
that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}], "]"}], "=", 
    "lcm"}], TraditionalForm]]],
 " we can infer  ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{
     RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", "lcm"}]}], 
   TraditionalForm]]],
 ", leading to a contradiction."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632726064205065*^9, 3.632726279775051*^9}, {
   3.6327263312386703`*^9, 3.63272638731102*^9}, {3.63273067588523*^9, 
   3.63273082134507*^9}, {3.632731112440227*^9, 3.63273111790838*^9}, {
   3.632731273726129*^9, 3.632731380424369*^9}, {3.6327314226857643`*^9, 
   3.632731470434132*^9}, {3.6327315020756063`*^9, 3.632731801472736*^9}, {
   3.6327318323743*^9, 3.63273195428367*^9}, {3.6327323947835903`*^9, 
   3.6327325004295464`*^9}, {3.6327325665448008`*^9, 3.632732647239635*^9}, {
   3.632732692792922*^9, 3.632732713865552*^9}, {3.632732767813447*^9, 
   3.632732793533283*^9}, {3.632732934895836*^9, 3.6327329787076387`*^9}, 
   3.636440256122418*^9, {3.6391305327828627`*^9, 3.63913055643571*^9}, 
   3.643446674830573*^9, 3.653208038009803*^9, 3.6532135773399487`*^9, 
   3.6661738640582914`*^9, 3.666440470303419*^9},
 CellID->89860634],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6327326227806597`*^9, 3.632732647239777*^9}, 
   3.6364402561225452`*^9, 3.6434466748306923`*^9, 3.653208038009803*^9, 
   3.6532135773399487`*^9, 3.6661738640582914`*^9, 3.6664404703035183`*^9},
 TextAlignment->Right,
 CellID->2002978456]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR3.\[Implies].3)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632733006701417*^9, 3.632733014667635*^9}, 
   3.632734249329339*^9, 3.636440261988167*^9, 3.643446681277968*^9, 
   3.6532080308181906`*^9, 3.653213568167133*^9, 3.666173857865081*^9, 
   3.666440465178151*^9},
 CellID->624408900],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632733467291273*^9, 3.6327334686666594`*^9}, 
   3.632734249329463*^9, 3.636440261988287*^9, 3.6434466812780857`*^9, 
   3.6532080308181906`*^9, 3.653213568167133*^9, 3.666173857865081*^9, 
   3.66644046517829*^9},
 CellID->1271771765],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", "c"}], "\[Implies]", 
    RowBox[{"p", "\[Succeeds]", 
     RowBox[{"c", "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}]}]}], TraditionalForm]]],
 " (",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " might be 0, and ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]]],
 " does not have to be a monomial!) (Order Relation)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632733478077387*^9, 3.632733561663814*^9}, 
   3.632734249329948*^9, {3.633175091310356*^9, 3.633175095023361*^9}, 
   3.636440261988399*^9, 3.6434466812781687`*^9, 3.6532080308181906`*^9, 
   3.653213568167133*^9, 3.666173857865081*^9, 3.666440465178401*^9},
 CellID->173464573],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "a"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"lc", "[", "a", "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nFrom (CR3.\[Implies].1) and (1) we can infer\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "a", "]"}], TraditionalForm]]],
 "\nand further\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n\nFrom (1), by the definition of isMNTCR, we know\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"a", "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCR", "[", 
        RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}]}], 
     ")"}]}], TraditionalForm]]],
 "\n\nFrom (2), (4) and (CR1 monomial 2) we know\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"lc", "[", "a", "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n\nBy the definition of isMNTCRP we have to show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"lc", "[", "a", "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"lc", "[", "a", "]"}], "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCRP", "[", 
        RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}]}], 
     ")"}]}], TraditionalForm]]],
 "\n(G1) is exactly what (8) states, so we only consider (G2); For this, \
choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "a0", "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "a", "]"}], "\[Succeeds]", "a0"}], TraditionalForm]]],
 "\nand prove\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCRP", "[", 
     RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 "\n\nWe instantiate (5) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a0", "\[CenterDot]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 ", so we know\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"a0", "\[CenterDot]", 
      RowBox[{"lp", "[", "a", "]"}]}]}], "\[Implies]", 
    RowBox[{"\[Not]", 
     RowBox[{"isNTCR", "[", 
      RowBox[{"p", ",", "q", ",", 
       RowBox[{"a0", "\[CenterDot]", 
        RowBox[{"lp", "[", "a", "]"}]}], ",", "i", ",", "j"}], "]"}]}]}], 
   TraditionalForm]]],
 "\nThe premise of (10) holds due to a lemma, so we can add\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{"p", ",", "q", ",", 
      RowBox[{"a0", "\[CenterDot]", 
       RowBox[{"lp", "[", "a", "]"}]}], ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 "\nto our knowledge base.\n\nInstead of proving (G3) we assume its negation \
and show the negation of (11), i.e. we know\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand show\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"a0", "\[CenterDot]", 
      RowBox[{"lp", "[", "a", "]"}]}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n\nBy (CR1 monomial 3) the goal (G4) can be replaced by\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nAll three subgoals follow readily from (6), (7) and (12)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{
  3.632733030991404*^9, {3.632733071596877*^9, 3.632733403280529*^9}, {
   3.63273359198503*^9, 3.6327336769839363`*^9}, {3.6327338165599947`*^9, 
   3.6327338344989443`*^9}, {3.632733889179359*^9, 3.632734011236292*^9}, {
   3.6327340415639343`*^9, 3.632734214443748*^9}, 3.632734249330052*^9, {
   3.632735438876286*^9, 3.632735502040873*^9}, {3.632735559495038*^9, 
   3.6327355758591557`*^9}, 3.636440261988497*^9, {3.63919599234659*^9, 
   3.639196005094264*^9}, 3.643446681278274*^9, 3.6532080308181906`*^9, 
   3.653213568167133*^9, 3.6661738578806806`*^9, 3.666440465178515*^9},
 CellID->454649756],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.63273423692036*^9, 3.6327342493301687`*^9}, 
   3.636440261988614*^9, 3.643446681278389*^9, 3.6532080308181906`*^9, 
   3.653213568167133*^9, 3.6661738578806806`*^9, 3.666440465178648*^9},
 TextAlignment->Right,
 CellID->493606769]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR3.\[DoubleLeftArrow])",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632734258619809*^9, 3.632734269426785*^9}, 
   3.632738261699912*^9, 3.636440268313951*^9, 3.643446687261064*^9, 
   3.6532080253571806`*^9, 3.6532135618481216`*^9, 3.6661738514534693`*^9, 
   3.66644045992865*^9},
 CellID->1772726150],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6327358444047527`*^9, 3.632735845682498*^9}, {
   3.632736326701746*^9, 3.632736327661112*^9}, 3.632738261700036*^9, 
   3.6364402683140717`*^9, 3.6434466872611856`*^9, 3.6532080253571806`*^9, 
   3.653213561863722*^9, 3.6661738514534693`*^9, 3.666440459928755*^9},
 CellID->1167894951],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{
      SubscriptBox["c", "1"], ",", 
      SubscriptBox["c", "2"], ",", "0", ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 " (in general in domains where (R4) holds) (Reduction Ring lemmas)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632736332342738*^9, 3.632736357177944*^9}, {
   3.632737514873745*^9, 3.632737514887121*^9}, 3.632738261700513*^9, {
   3.633175444158411*^9, 3.633175448492137*^9}, 3.6364402683141823`*^9, 
   3.6434466872612667`*^9, 3.6532080253571806`*^9, 3.653213561863722*^9, 
   3.6661738514534693`*^9, 3.666440459928838*^9},
 CellID->759496477],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCRP", "[", 
     RowBox[{"p", ",", "q", ",", "0", ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 " (345)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632736332342738*^9, 3.632736357177944*^9}, {
   3.632737514873745*^9, 3.63273757253469*^9}, 3.632738261700597*^9, 
   3.6331753452975893`*^9, {3.6331755744284286`*^9, 3.633175576049262*^9}, 
   3.6364402683142567`*^9, 3.643446687261327*^9, 3.6532080253571806`*^9, 
   3.653213561863722*^9, 3.6661738514534693`*^9, 3.666440459928895*^9},
 CellID->363005801],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"c", "\[CenterDot]", "t"}], "\[Succeeds]", "p"}], "\[And]", 
      RowBox[{"p", "\[NotEqual]", "0"}]}], ")"}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"t", "\[Succeeds]", 
         RowBox[{"lp", "[", "p", "]"}]}], "\[And]", 
        RowBox[{"c", "\[NotEqual]", "0"}]}], ")"}], "\[Or]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"lp", "[", "p", "]"}], "=", 
        RowBox[{"t", "\[And]", 
         RowBox[{"c", "\[Succeeds]", 
          RowBox[{"lc", "[", "p", "]"}]}]}]}], ")"}]}], ")"}]}], 
   TraditionalForm]]],
 " (377\[CloseCurlyQuote])"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632735851214196*^9, 3.632735898580661*^9}, {
   3.632736016797986*^9, 3.632736034520385*^9}, {3.632736086544561*^9, 
   3.632736121021351*^9}, {3.63273616532825*^9, 3.632736166384141*^9}, {
   3.632737721723834*^9, 3.632737721737979*^9}, 3.632738261700672*^9, {
   3.6331761312546043`*^9, 3.6331761335422773`*^9}, 3.636440268314328*^9, {
   3.639196728938943*^9, 3.6391967301108923`*^9}, 3.643446687261376*^9, 
   3.6532080253571806`*^9, 3.653213561863722*^9, 3.6661738514534693`*^9, 
   3.666440459928948*^9},
 CellID->19508673],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", "0"}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"p", "\[NotEqual]", "0"}], "\[And]", 
      RowBox[{"(", 
       RowBox[{"t", "=", 
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[Or]", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", "t"}]}]}], ")"}]}], 
     ")"}]}], TraditionalForm]]],
 " (Selectors)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632735851214196*^9, 3.632735898580661*^9}, {
   3.632736016797986*^9, 3.632736034520385*^9}, {3.632736086544561*^9, 
   3.632736121021351*^9}, {3.63273616532825*^9, 3.632736166384141*^9}, {
   3.632737721723834*^9, 3.632737746959113*^9}, {3.632737906306212*^9, 
   3.632737906323875*^9}, 3.6327382617007513`*^9, {3.633176288245987*^9, 
   3.6331762906346807`*^9}, 3.6364402683144007`*^9, {3.639196733726376*^9, 
   3.63919673491471*^9}, 3.643446687261426*^9, 3.6532080253571806`*^9, 
   3.653213561863722*^9, 3.6661738514534693`*^9, 3.666440459929001*^9},
 CellID->1204606733],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "c", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"c", "\[CenterDot]", "lcm"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nwhere lcm abbreviates ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "[", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], ",", 
     RowBox[{"lp", "[", "q", "]"}]}], "]"}], TraditionalForm]]],
 ".\n\nExpanding the definitions of isMNTCRP and isMNTCR we know\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "c", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"c", "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCRP", "[", 
        RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}]}], 
     ")"}]}], TraditionalForm]]],
 "\nand we have to show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"c", "\[CenterDot]", "lcm"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"c", "\[CenterDot]", "lcm"}], "\[Succeeds]", "a0"}], 
      "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCR", "[", 
        RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}]}], 
     ")"}]}], TraditionalForm]]],
 "\n\n(G1):\nBy (CR1 monomial 3) the goal (G1) can be reduced to\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "lcm"}], 
   TraditionalForm]]],
 "\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "lcm"}], 
   TraditionalForm]]],
 "\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "c", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(G5) is exactly what is stated in (2), and (G3) and (G4) follow readily \
from \[ScriptCapitalT] being a power-product domain.\n\n(G2):\nChoose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "a0", "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c", "\[CenterDot]", "lcm"}], "\[Succeeds]", "a0"}], 
   TraditionalForm]]],
 "\nand show\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 "\n\nWe distinguish two cases, based on whether or not ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 ".\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", "0"}], TraditionalForm]]],
 ".\nThis case is trivial, since 0 cannot be a NTCR of any elements under the \
presence of (R4), due to a lemma.\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 ".\nInstead of showing (G6) we assume its negation and try to derive a \
contradiction, i.e. we assume\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a0", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nFrom (5), together with (CR1.\[Implies].2), we can infer the existence of \
some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 " with\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "t"}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"C", "[", 
      RowBox[{"a0", ",", "t"}], "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n\nWe can immediately infer\n(8\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "\[VerticalBar]", "t"}], TraditionalForm]]],
 "\n\nFrom a lemma we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCRP", "[", 
     RowBox[{"p", ",", "q", ",", "0", ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 ", so we also know\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a0", ",", "t"}], "]"}], "\[NotEqual]", "0"}], TraditionalForm]]],
 "\nWe can use (9) and another lemma to conclude\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{
     RowBox[{"lp", "[", "a0", "]"}], "\[Or]", 
     RowBox[{
      RowBox[{"lp", "[", "a0", "]"}], "\[Succeeds]", "t"}]}]}], 
   TraditionalForm]]],
 "\n\nFrom our case assumption and (4), together with a lemma, we can infer\n\
(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"lcm", "\[Succeeds]", 
       RowBox[{"lp", "[", "a0", "]"}]}], "\[And]", 
      RowBox[{"c", "\[NotEqual]", "0"}]}], ")"}], "\[Or]", 
    RowBox[{"(", 
     RowBox[{"lcm", "=", 
      RowBox[{
       RowBox[{"lp", "[", "a0", "]"}], "\[And]", 
       RowBox[{"c", "\[Succeeds]", 
        RowBox[{"lc", "[", "a0", "]"}]}]}]}], ")"}]}], TraditionalForm]]],
 "\nbased upon which we again distinguish two cases.\n\nCase II.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "=", 
    RowBox[{
     RowBox[{"lp", "[", "a0", "]"}], "\[And]", 
     RowBox[{"c", "\[Succeeds]", 
      RowBox[{"lc", "[", "a0", "]"}]}]}]}], TraditionalForm]]],
 ".\nWe instantiate (3) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lc", "[", "a0", "]"}], TraditionalForm]]],
 ", and since the premise of the implication holds due to our case \
assumption, we know\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCRP", "[", 
     RowBox[{"p", ",", "q", ",", 
      RowBox[{"lc", "[", "a0", "]"}], ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 "\nRather than deriving a contradiction we show the negation of (12), i.e. \
we show\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"lc", "[", "a0", "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nIf we can prove ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{"lp", "[", "a0", "]"}]}], TraditionalForm]]],
 " we are done, because then (G7) follows immediately from (8), so we try to \
prove\n(G8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{"lp", "[", "a0", "]"}]}], TraditionalForm]]],
 "\nnow. The first case in (10) is just what we want to show, so we only \
consider the second case in (10) with\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "a0", "]"}], "\[Succeeds]", "t"}], TraditionalForm]]],
 "\nSince we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "a0", "]"}], "=", "lcm"}], TraditionalForm]]],
 " we also know\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "\[Succeeds]", "t"}], TraditionalForm]]],
 "\n(8\[CloseCurlyQuote]) contradicts (14).\n\nCase II.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lcm", "\[Succeeds]", 
     RowBox[{"lp", "[", "a0", "]"}]}], "\[And]", 
    RowBox[{"c", "\[NotEqual]", "0"}]}], TraditionalForm]]],
 ".\nIndependent of which case holds in (10) we know\n(15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "\[Succeeds]", "t"}], TraditionalForm]]],
 "\n(8\[CloseCurlyQuote]) contradicts (15)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632734307196405*^9, 3.6327345581004543`*^9}, {
   3.632734616426852*^9, 3.632734618295856*^9}, {3.632735595209683*^9, 
   3.632735772537394*^9}, {3.6327362213557796`*^9, 3.632736244635401*^9}, {
   3.632736275555282*^9, 3.6327363234943743`*^9}, {3.632736367094845*^9, 
   3.6327363979326878`*^9}, {3.632736454573246*^9, 3.6327365614484587`*^9}, {
   3.632736632823606*^9, 3.632736807324461*^9}, {3.6327371478433037`*^9, 
   3.632737219634656*^9}, {3.632737269975819*^9, 3.632737336124447*^9}, {
   3.6327373758137712`*^9, 3.632737451234105*^9}, {3.6327375975639477`*^9, 
   3.632737709967437*^9}, {3.632737754947998*^9, 3.632737876251749*^9}, {
   3.6327379420373297`*^9, 3.632737999156783*^9}, {3.632738052305998*^9, 
   3.632738224440189*^9}, 3.6327382617008543`*^9, 3.636440268314497*^9, {
   3.639196742973996*^9, 3.639196747286263*^9}, {3.63919678639097*^9, 
   3.6391968044747047`*^9}, {3.63919699163503*^9, 3.639197029590396*^9}, 
   3.6434466872615232`*^9, 3.6532080253571806`*^9, 3.653213561863722*^9, 
   3.6661738514534693`*^9, 3.666440459929081*^9},
 CellID->936576587],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632738249603307*^9, 3.632738261700988*^9}, 
   3.636440268314617*^9, 3.64344668726164*^9, 3.6532080253571806`*^9, 
   3.653213561863722*^9, 3.6661738514534693`*^9, 3.666440459929179*^9},
 TextAlignment->Right,
 CellID->1594670345]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->666606593],

Cell[CellGroupData[{

Cell["\<\
Theorem: Alternative Description of unary isMNTCR in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.633235864741994*^9, 3.639195975854535*^9, {3.642926172631085*^9, 
   3.642926187038052*^9}},
 CellID->1056394693],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "a"}], "]"}]}]], 
  RowBox[{"p", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1444400550}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.631963023902554*^9, 3.631963056276445*^9}, {
   3.6322145639880133`*^9, 3.6322145642449713`*^9}, {3.633235871918832*^9, 
   3.6332358733547697`*^9}, 3.635748102132021*^9},
 CellTags->{
  "ID:1444400550", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1444400550],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "a"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "a", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR4.\[DoubleRightArrow].1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:653031981", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633235880742654*^9, 3.633235884752688*^9}},
 CellTags->{
  "ID:653031981", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].1"},
 CellID->653031981],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "a"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "a", "]"}], "\[Equal]", 
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR4.\[DoubleRightArrow].2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:943141844", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6332359271651783`*^9, 3.633235936651855*^9}},
 CellTags->{
  "ID:943141844", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].2"},
 CellID->943141844],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "a"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
    RowBox[{
     UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "a", "]"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR4.\[DoubleRightArrow].3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1143014070", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6332359639532003`*^9, 3.6332359799860983`*^9}, {
  3.639205855093132*^9, 3.639205856357912*^9}},
 CellTags->{
  "ID:1143014070", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].3"},
 CellID->1143014070],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", "c"}],
      "]"}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", 
      RowBox[{"c", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR4.\[DoubleLeftArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1630861163", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6332360780052767`*^9, 3.633236101018567*^9}, {
  3.6392058596779003`*^9, 3.639205860904415*^9}},
 CellTags->{
  "ID:1630861163", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleLeftArrow]"},
 CellID->1630861163],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "a"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", "c", "]"}]}]], 
     RowBox[{"c", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
       RowBox[{
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
        "c"}], "]"}], "\[And]", 
      RowBox[{"a", "\[Equal]", 
       RowBox[{"c", 
        UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
        RowBox[{
         UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CR4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1701661908", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6429261958294764`*^9, 3.642926258602623*^9}},
 CellTags->{
  "ID:1701661908", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4"},
 CellID->1701661908],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1473391475]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->973852342],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR4.\[DoubleRightArrow].1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR4.\[DoubleRightArrow].1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["a", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["a", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:653031981"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p653031981-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p653031981-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:653031981", "Proof|ID:653031981-1"},
 CellID->13059041],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p653031981-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->936750914]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR4.\[DoubleRightArrow].2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR4.\[DoubleRightArrow].2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:943141844"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p943141844-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p943141844-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:943141844", "Proof|ID:943141844-1"},
 CellID->1657492493],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p943141844-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->1516876266]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR4.\[DoubleRightArrow].3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR4.\[DoubleRightArrow].3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], ",", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], ",", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1143014070"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1143014070-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1143014070-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1143014070", "Proof|ID:1143014070-1"},
 CellID->1936376328],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1143014070-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->7830102]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR4.\[DoubleLeftArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR4.\[DoubleLeftArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1630861163"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1630861163-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1630861163-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1630861163", "Proof|ID:1630861163-1"},
 CellID->979966294],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1630861163-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->1309847775]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CR4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CR4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1701661908"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1701661908-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1701661908-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1701661908", "Proof|ID:1701661908-1"},
 CellID->599419806],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1701661908-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1584091188]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR4.\[Implies].1)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632724361410532*^9, 3.6327243676094503`*^9}, 
   3.632725904162763*^9, 3.636440250646221*^9, 3.6391988911836967`*^9, 
   3.643446697358502*^9, 3.653208015513563*^9, 3.6532135518953047`*^9, 
   3.6661738429806547`*^9, 3.666440450977003*^9},
 CellID->100124376],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6327248899737053`*^9, 3.6327248912985277`*^9}, 
   3.6327259041628857`*^9, 3.63644025064638*^9, 3.643446697358658*^9, 
   3.653208015513563*^9, 3.6532135518953047`*^9, 3.6661738429806547`*^9, 
   3.666440450977141*^9},
 CellID->1945351028],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isReducible", "[", 
     RowBox[{"a", ",", "c"}], "]"}], "\[Implies]", 
    RowBox[{"a", "\[NotEqual]", "0"}]}], TraditionalForm]]],
 " (in general in domains where (R4) holds) (Reduction ring lemmas)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632724896382174*^9, 3.6327249274949627`*^9}, 
   3.632725904163352*^9, {3.633175041979538*^9, 3.63317504604731*^9}, 
   3.636440250646522*^9, 3.643446697358762*^9, 3.653208015513563*^9, 
   3.6532135518953047`*^9, 3.6661738429806547`*^9, 3.666440450977252*^9},
 CellID->1341586550],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "a"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", "a"}], "]"}], TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "a", "]"}], TraditionalForm]]],
 "\nWe expand the definition of isMNTCR: We know\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "a"}], "]"}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"a", "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCR", "[", 
        RowBox[{"p", ",", "a0"}], "]"}]}]}], ")"}]}], TraditionalForm]]],
 "\nFrom a lemma we know that ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " can either be 0, a monomial, or a polynomial with more than one \
power-products with non-zero coefficients. Thus, we distinguish three cases.\n\
\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", "0"}], TraditionalForm]]],
 ".\nFrom (2) we can infer, among others,\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", "p"}], "]"}], TraditionalForm]]],
 "\nHence, from a lemma we immediately get ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", contradicting our case assumption.\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "a", "]"}], TraditionalForm]]],
 ".\nThis is exactly what we have to show.\n\nCase III: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", 
       RowBox[{"s", ",", "t"}], "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"s", "\[NotEqual]", "t"}], "\[And]", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "s"}], "]"}], "\[NotEqual]", "0"}], "\[And]", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "t"}], "]"}], "\[NotEqual]", "0"}]}], ")"}]}], 
   TraditionalForm]]],
 "\nLet ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"s", ",", "t"}], "]"}]}], TraditionalForm]]],
 " witness the existential case assumption, meaning we know\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", "\[NotEqual]", "t"}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "s"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "t"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\n\nNow we can use (CR2 monomial 1): From (2) and (CR2 monomial 1) we can \
infer the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "u", "]"}]}], TraditionalForm]]],
 " with\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 "\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}]}], "]"}], 
   TraditionalForm]]],
 "\nFrom (9), by a lemma, we immediately get\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "u"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\n\nWe claim\n(C)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}]}], 
   TraditionalForm]]],
 "\nBecause of (5) ",
 Cell[BoxData[
  FormBox["u", TraditionalForm]]],
 " cannot be equal both ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 ", so we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"u", "\[NotEqual]", "t"}], "\[Or]", 
    RowBox[{"u", "\[NotEqual]", "s"}]}], TraditionalForm]]],
 ", leading to a case distinction; However, since both cases are completely \
symmetric, we only consider the first one with\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"u", "\[NotEqual]", "t"}], TraditionalForm]]],
 "\nFrom a lemma on \[Succeeds] in \[ScriptCapitalP] we know (because of (7) \
and (10))\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}]}], 
   TraditionalForm]]],
 "\n\nNow we instantiate (3) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a0", ":=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}]}], 
   TraditionalForm]]],
 ". The premise of the implication holds due to (12), so we can infer\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{"p", ",", "q", ",", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "u"}], ",", "i", ",", 
      "j"}], "]"}]}], TraditionalForm]]],
 "\nin contradiction to (10)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632724374386097*^9, 3.632724499997056*^9}, {
   3.6327245402374687`*^9, 3.632724660357985*^9}, {3.6327246912755623`*^9, 
   3.632724717827095*^9}, {3.6327247485967607`*^9, 3.632724757555394*^9}, {
   3.6327249377957487`*^9, 3.632725054533145*^9}, {3.6327254396943703`*^9, 
   3.6327254423216467`*^9}, {3.6327254792405367`*^9, 3.632725755514175*^9}, {
   3.6327258078581343`*^9, 3.632725904163457*^9}, 3.6364402506466503`*^9, {
   3.639198898323278*^9, 3.639198988988386*^9}, {3.639199026806237*^9, 
   3.639199125743003*^9}, {3.639199155818623*^9, 3.639199182245579*^9}, 
   3.643446697358902*^9, 3.653208015513563*^9, 3.6532135518953047`*^9, 
   3.6661738429806547`*^9, 3.6664404509773703`*^9},
 CellID->1856712428],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632725892101275*^9, 3.632725904163587*^9}, 
   3.6364402506468143`*^9, 3.643446697359056*^9, 3.653208015513563*^9, 
   3.6532135518953047`*^9, 3.6661738429806547`*^9, 3.666440450977509*^9},
 TextAlignment->Right,
 CellID->577063717]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR4.\[Implies].2)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6327260415510883`*^9, 3.63272604884029*^9}, 
   3.6327326472389174`*^9, 3.6364402561220837`*^9, {3.6392001149259987`*^9, 
   3.63920011556078*^9}, 3.6434467188037024`*^9, 3.6532080044219437`*^9, 
   3.6532135456552935`*^9, 3.666173837254444*^9, 3.666440440985834*^9},
 CellID->2028257075],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632731393196661*^9, 3.632731394401993*^9}, 
   3.632732647239039*^9, 3.636440256122204*^9, 3.643446718803822*^9, 
   3.6532080044219437`*^9, 3.6532135456552935`*^9, 3.666173837254444*^9, 
   3.666440440985977*^9},
 CellID->997822543],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"s", "\[VerticalBar]", "t"}], "\[Implies]", 
    RowBox[{"\[Not]", 
     RowBox[{"s", "\[Succeeds]", "t"}]}]}], TraditionalForm]]],
 " (ET CPP)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6327313988256197`*^9, 3.632731412965679*^9}, 
   3.63273264723953*^9, {3.63317505401931*^9, 3.633175056017473*^9}, 
   3.636440256122321*^9, 3.643446718803905*^9, 3.6532080044219437`*^9, 
   3.6532135456552935`*^9, 3.666173837254444*^9, 3.666440440986088*^9},
 CellID->1728450463],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "a"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", "a"}], "]"}], TraditionalForm]]],
 "\nand prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "a", "]"}], "=", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 "\nFrom (CR4.\[Implies].1) and (1) we can infer\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "a", "]"}], TraditionalForm]]],
 "\nand further\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n\nFrom (1), by the definition of isMNTCR, we know\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "a"}], "]"}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"a", "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCR", "[", 
        RowBox[{"p", ",", "a0"}], "]"}]}]}], ")"}]}], TraditionalForm]]],
 "\n\nFrom (2), (4) and (CR2 monomial 2) we know\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "a", "]"}]}], "]"}], TraditionalForm]]],
 "\n\nSince ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], TraditionalForm]]],
 " is total, we distinguish now three cases, based on the relative order of \
",
 Cell[BoxData[
  FormBox[
   RowBox[{"lp", "[", "a", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lp", "[", "p", "]"}], TraditionalForm]]],
 ".\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "a", "]"}], "=", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 ".\nThis is exactly what we want to prove.\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 ".\nThis contradicts (6) using (PP 4).\n\nCase III: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "a", "]"}], "\[Succeeds]", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 ".\nWe make the following claim:\n(C)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], TraditionalForm]]],
 "\n\nProof of (C):\nClearly we have ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", 
    RowBox[{
     RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
     RowBox[{"lp", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 ", hence we can use (CR2 monomial 2) to reduce the goal to\n(C1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}], "]"}]}], TraditionalForm]]],
 "\n(C2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", 
      RowBox[{
       RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
       RowBox[{"lp", "[", "p", "]"}]}], "]"}]}], "]"}], TraditionalForm]]],
 "\nApparently, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}], "]"}], "=", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}], "]"}], "=", 
    RowBox[{"lc", "[", "a", "]"}]}], TraditionalForm]]],
 ", so we only have to show\n(C1\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 "\n(C2\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "a", "]"}]}], "]"}], TraditionalForm]]],
 "\n(C2\[CloseCurlyQuote]) is precisely what (7) states, and (CR1\
\[CloseCurlyQuote]) follow readily from \[ScriptCapitalT] being a \
power-product domain.\n\nSummarizing, we know\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], TraditionalForm]]],
 "\n\nNow we instantiate (5) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a0", ":=", 
    RowBox[{
     RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
     RowBox[{"lp", "[", "p", "]"}]}]}], TraditionalForm]]],
 ", so we know\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}]}], "\[Implies]", 
    RowBox[{"\[Not]", 
     RowBox[{"isNTCR", "[", 
      RowBox[{"p", ",", 
       RowBox[{
        RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
        RowBox[{"lp", "[", "p", "]"}]}]}], "]"}]}]}], TraditionalForm]]],
 "\nFrom (9) and (8), by modus tollens, we can infer ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}]}]}], TraditionalForm]]],
 ". However, from our case assumption, (382\[CloseCurlyQuote]) and the fact \
that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", 
     RowBox[{
      RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}], "]"}], "=", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 " we can infer  ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{
     RowBox[{"lc", "[", "a", "]"}], "\[CenterDot]", 
     RowBox[{"lp", "[", "p", "]"}]}]}], TraditionalForm]]],
 ", leading to a contradiction."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632726064205065*^9, 3.632726279775051*^9}, {
   3.6327263312386703`*^9, 3.63272638731102*^9}, {3.63273067588523*^9, 
   3.63273082134507*^9}, {3.632731112440227*^9, 3.63273111790838*^9}, {
   3.632731273726129*^9, 3.632731380424369*^9}, {3.6327314226857643`*^9, 
   3.632731470434132*^9}, {3.6327315020756063`*^9, 3.632731801472736*^9}, {
   3.6327318323743*^9, 3.63273195428367*^9}, {3.6327323947835903`*^9, 
   3.6327325004295464`*^9}, {3.6327325665448008`*^9, 3.632732647239635*^9}, {
   3.632732692792922*^9, 3.632732713865552*^9}, {3.632732767813447*^9, 
   3.632732793533283*^9}, {3.632732934895836*^9, 3.6327329787076387`*^9}, 
   3.636440256122418*^9, {3.639200121524744*^9, 3.639200205998028*^9}, {
   3.6392002399279823`*^9, 3.639200326980186*^9}, {3.6392003728407917`*^9, 
   3.639200560989974*^9}, 3.643446718804014*^9, 3.6532080044219437`*^9, 
   3.6532135456552935`*^9, 3.666173837254444*^9, 3.666440440986203*^9},
 CellID->127670836],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6327326227806597`*^9, 3.632732647239777*^9}, 
   3.6364402561225452`*^9, 3.643446718804132*^9, 3.6532080044219437`*^9, 
   3.6532135456552935`*^9, 3.666173837254444*^9, 3.6664404409863377`*^9},
 TextAlignment->Right,
 CellID->684788075]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR4.\[Implies].3)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632733006701417*^9, 3.632733014667635*^9}, 
   3.632734249329339*^9, 3.636440261988167*^9, 3.639205334306728*^9, 
   3.643446723525642*^9, 3.6532079978543324`*^9, 3.653213539243682*^9, 
   3.6661738303124323`*^9, 3.666440435713839*^9},
 CellID->1144589532],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632733467291273*^9, 3.6327334686666594`*^9}, 
   3.632734249329463*^9, 3.636440261988287*^9, 3.643446723525762*^9, 
   3.6532079978543324`*^9, 3.653213539243682*^9, 3.6661738303124323`*^9, 
   3.6664404357139797`*^9},
 CellID->2105013134],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], "\[Succeeds]", "c"}], "\[Implies]", 
    RowBox[{"p", "\[Succeeds]", 
     RowBox[{"c", "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}]}]}], TraditionalForm]]],
 " (",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " might be 0, and ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]]],
 " does not have to be a monomial!) (Order Relation)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632733478077387*^9, 3.632733561663814*^9}, 
   3.632734249329948*^9, {3.633175091310356*^9, 3.633175095023361*^9}, 
   3.636440261988399*^9, 3.6434467235258417`*^9, 3.6532079978543324`*^9, 
   3.653213539243682*^9, 3.666173830328032*^9, 3.666440435714093*^9},
 CellID->354338712],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "a"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", "a"}], "]"}], TraditionalForm]]],
 "\nand prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "a", "]"}]}], "]"}], TraditionalForm]]],
 "\nFrom (CR4.\[Implies].1) and (1) we can infer\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "a", "]"}], TraditionalForm]]],
 "\nand further\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n\nFrom (1), by the definition of isMNTCR, we know\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "a"}], "]"}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"a", "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCR", "[", 
        RowBox[{"p", ",", "a0"}], "]"}]}]}], ")"}]}], TraditionalForm]]],
 "\n\nFrom (2), (4) and (CR2 monomial 2) we know\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "a", "]"}]}], "]"}], TraditionalForm]]],
 "\n\nBy the definition of isMNTCR we have to show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "a", "]"}]}], "]"}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"lc", "[", "a", "]"}], "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCR", "[", 
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], ",", "a0"}], "]"}]}]}], ")"}]}], 
   TraditionalForm]]],
 "\n(G1) is exactly what (7) states, so we only consider (G2); For this, \
choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "a0", "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "a", "]"}], "\[Succeeds]", "a0"}], TraditionalForm]]],
 "\nand prove\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], ",", "a0"}], "]"}]}], 
   TraditionalForm]]],
 "\n\nWe instantiate (5) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a0", "\[CenterDot]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 ", so we know\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"a0", "\[CenterDot]", 
      RowBox[{"lp", "[", "a", "]"}]}]}], "\[Implies]", 
    RowBox[{"\[Not]", 
     RowBox[{"isNTCR", "[", 
      RowBox[{"p", ",", 
       RowBox[{"a0", "\[CenterDot]", 
        RowBox[{"lp", "[", "a", "]"}]}]}], "]"}]}]}], TraditionalForm]]],
 "\nThe premise of (9) holds due to a lemma, so we can add\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{"p", ",", 
      RowBox[{"a0", "\[CenterDot]", 
       RowBox[{"lp", "[", "a", "]"}]}]}], "]"}]}], TraditionalForm]]],
 "\nto our knowledge base.\n\nInstead of proving (G3) we assume its negation \
and show the negation of (10), i.e. we know\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", "a0"}], "]"}], TraditionalForm]]],
 "\nand show\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", 
     RowBox[{"a0", "\[CenterDot]", 
      RowBox[{"lp", "[", "a", "]"}]}]}], "]"}], TraditionalForm]]],
 "\n\nBy (CR2 monomial 3) the goal (G4) can be replaced by\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "a", "]"}]}], TraditionalForm]]],
 "\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", "a0"}], "]"}], TraditionalForm]]],
 "\nBoth subgoals follow readily from (6) and (11)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{
  3.632733030991404*^9, {3.632733071596877*^9, 3.632733403280529*^9}, {
   3.63273359198503*^9, 3.6327336769839363`*^9}, {3.6327338165599947`*^9, 
   3.6327338344989443`*^9}, {3.632733889179359*^9, 3.632734011236292*^9}, {
   3.6327340415639343`*^9, 3.632734214443748*^9}, 3.632734249330052*^9, {
   3.632735438876286*^9, 3.632735502040873*^9}, {3.632735559495038*^9, 
   3.6327355758591557`*^9}, 3.636440261988497*^9, {3.639205341943918*^9, 
   3.639205563029828*^9}, 3.64344672352595*^9, 3.6532079978543324`*^9, 
   3.653213539243682*^9, 3.666173830328032*^9, 3.66644043571421*^9},
 CellID->1627125215],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.63273423692036*^9, 3.6327342493301687`*^9}, 
   3.636440261988614*^9, 3.643446723526066*^9, 3.6532079978543324`*^9, 
   3.653213539243682*^9, 3.666173830328032*^9, 3.666440435714343*^9},
 TextAlignment->Right,
 CellID->1095627526]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (CR4.\[DoubleLeftArrow])",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632734258619809*^9, 3.632734269426785*^9}, 
   3.632738261699912*^9, 3.636440268313951*^9, 3.639206093459402*^9, 
   3.643446729577155*^9, 3.653207990475519*^9, 3.6532135320364695`*^9, 
   3.666173824571622*^9, 3.666440430593884*^9},
 CellID->1999351486],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6327358444047527`*^9, 3.632735845682498*^9}, {
   3.632736326701746*^9, 3.632736327661112*^9}, 3.632738261700036*^9, 
   3.6364402683140717`*^9, 3.6434467295772743`*^9, 3.653207990475519*^9, 
   3.6532135320364695`*^9, 3.666173824571622*^9, 3.666440430594028*^9},
 CellID->1769332380],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{
      SubscriptBox["c", "1"], ",", 
      SubscriptBox["c", "2"], ",", "0", ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 " (in general in domains where (R4) holds) (Reduction Ring lemmas)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632736332342738*^9, 3.632736357177944*^9}, {
   3.632737514873745*^9, 3.632737514887121*^9}, 3.632738261700513*^9, {
   3.633175444158411*^9, 3.633175448492137*^9}, 3.6364402683141823`*^9, 
   3.643446729577354*^9, 3.653207990475519*^9, 3.6532135320364695`*^9, 
   3.666173824571622*^9, 3.666440430594141*^9},
 CellID->20695133],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCRP", "[", 
     RowBox[{"p", ",", "q", ",", "0", ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 " (345)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632736332342738*^9, 3.632736357177944*^9}, {
   3.632737514873745*^9, 3.63273757253469*^9}, 3.632738261700597*^9, 
   3.6331753452975893`*^9, {3.6331755744284286`*^9, 3.633175576049262*^9}, 
   3.6364402683142567`*^9, 3.643446729577415*^9, 3.653207990475519*^9, 
   3.6532135320364695`*^9, 3.666173824571622*^9, 3.6664404305942183`*^9},
 CellID->1281016174],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"c", "\[CenterDot]", "t"}], "\[Succeeds]", "p"}], "\[And]", 
      RowBox[{"isDefined", "[", 
       RowBox[{"lp", "[", "p", "]"}], "]"}]}], ")"}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"t", "\[Succeeds]", 
         RowBox[{"lp", "[", "p", "]"}]}], "\[And]", 
        RowBox[{"c", "\[NotEqual]", "0"}]}], ")"}], "\[Or]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"lp", "[", "p", "]"}], "=", 
        RowBox[{"t", "\[And]", 
         RowBox[{"c", "\[Succeeds]", 
          RowBox[{"lc", "[", "p", "]"}]}]}]}], ")"}]}], ")"}]}], 
   TraditionalForm]]],
 " (377\[CloseCurlyQuote])"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632735851214196*^9, 3.632735898580661*^9}, {
   3.632736016797986*^9, 3.632736034520385*^9}, {3.632736086544561*^9, 
   3.632736121021351*^9}, {3.63273616532825*^9, 3.632736166384141*^9}, {
   3.632737721723834*^9, 3.632737721737979*^9}, 3.632738261700672*^9, {
   3.6331761312546043`*^9, 3.6331761335422773`*^9}, 3.636440268314328*^9, 
   3.643446729577463*^9, 3.653207990475519*^9, 3.6532135320364695`*^9, 
   3.666173824571622*^9, 3.66644043059429*^9},
 CellID->1750850628],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", "0"}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"isDefined", "[", 
       RowBox[{"lp", "[", "p", "]"}], "]"}], "\[And]", 
      RowBox[{"(", 
       RowBox[{"t", "=", 
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[Or]", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", "t"}]}]}], ")"}]}], 
     ")"}]}], TraditionalForm]]],
 " (Selectors)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632735851214196*^9, 3.632735898580661*^9}, {
   3.632736016797986*^9, 3.632736034520385*^9}, {3.632736086544561*^9, 
   3.632736121021351*^9}, {3.63273616532825*^9, 3.632736166384141*^9}, {
   3.632737721723834*^9, 3.632737746959113*^9}, {3.632737906306212*^9, 
   3.632737906323875*^9}, 3.6327382617007513`*^9, {3.633176288245987*^9, 
   3.6331762906346807`*^9}, 3.6364402683144007`*^9, 3.643446729577512*^9, 
   3.653207990475519*^9, 3.6532135320364695`*^9, 3.666173824571622*^9, 
   3.666440430594363*^9},
 CellID->1983890089],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "p", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", "c"}], "]"}], TraditionalForm]]],
 "\nand prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", 
     RowBox[{"c", "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], TraditionalForm]]],
 "\n\nExpanding the definition of isMNTCR we know\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", "c"}], "]"}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"c", "\[Succeeds]", "a0"}], "\[Implies]", 
      RowBox[{"\[Not]", 
       RowBox[{"isNTCR", "[", 
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], ",", "a0"}], "]"}]}]}], ")"}]}], 
   TraditionalForm]]],
 "\nand we have to show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", 
     RowBox[{"c", "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "a0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"c", "\[CenterDot]", 
        RowBox[{"lp", "[", "p", "]"}]}], "\[Succeeds]", "a0"}], "\[Implies]", 
      
      RowBox[{"\[Not]", 
       RowBox[{"isNTCR", "[", 
        RowBox[{"p", ",", "a0"}], "]"}]}]}], ")"}]}], TraditionalForm]]],
 "\n\n(G1):\nBy (CR2 monomial 3) the goal (G1) can be reduced to\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 "\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", "c"}], "]"}], TraditionalForm]]],
 "\n(G4) is exactly what is stated in (2), and (G3) follows readily from \
\[ScriptCapitalT] being a power-product domain.\n\n(G2):\nChoose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "a0", "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c", "\[CenterDot]", 
     RowBox[{"lp", "[", "p", "]"}]}], "\[Succeeds]", "a0"}], 
   TraditionalForm]]],
 "\nand show\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{"p", ",", "a0"}], "]"}]}], TraditionalForm]]],
 "\n\nWe distinguish two cases, based on whether or not ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a0", "=", "0"}], TraditionalForm]]],
 ".\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a0", "=", "0"}], TraditionalForm]]],
 ".\nThis case is trivial, since 0 cannot be a NTCR of any elements under the \
presence of (R4), due to a lemma.\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a0", "\[NotEqual]", "0"}], TraditionalForm]]],
 ".\nInstead of showing (G6) we assume its negation and try to derive a \
contradiction, i.e. we assume\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "a0"}], "]"}], TraditionalForm]]],
 "\nFrom (5), together with (CR2.\[Implies].2), we can infer the existence of \
some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 " with\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a0", ",", "t"}], "]"}]}], "]"}], TraditionalForm]]],
 "\n\nFrom (6) we can immediately infer\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", "t"}]}], TraditionalForm]]],
 "\n\nFrom a lemma we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{"p", ",", "0"}], "]"}]}], TraditionalForm]]],
 ", so we also know\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a0", ",", "t"}], "]"}], "\[NotEqual]", "0"}], TraditionalForm]]],
 "\nWe can use (9) and another lemma to conclude\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{
     RowBox[{"lp", "[", "a0", "]"}], "\[Or]", 
     RowBox[{
      RowBox[{"lp", "[", "a0", "]"}], "\[Succeeds]", "t"}]}]}], 
   TraditionalForm]]],
 "\n\nFrom our case assumption and (4), together with a lemma, we can infer\n\
(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", 
       RowBox[{"lp", "[", "a0", "]"}]}], "\[And]", 
      RowBox[{"c", "\[NotEqual]", "0"}]}], ")"}], "\[Or]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"lp", "[", "p", "]"}], "=", 
      RowBox[{
       RowBox[{"lp", "[", "a0", "]"}], "\[And]", 
       RowBox[{"c", "\[Succeeds]", 
        RowBox[{"lc", "[", "a0", "]"}]}]}]}], ")"}]}], TraditionalForm]]],
 "\nbased upon which we again distinguish two cases.\n\nCase II.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "=", 
    RowBox[{
     RowBox[{"lp", "[", "a0", "]"}], "\[And]", 
     RowBox[{"c", "\[Succeeds]", 
      RowBox[{"lc", "[", "a0", "]"}]}]}]}], TraditionalForm]]],
 ".\nWe instantiate (3) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lc", "[", "a0", "]"}], TraditionalForm]]],
 ", and since the premise of the implication holds due to our case \
assumption, we know\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isNTCR", "[", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], ",", 
      RowBox[{"lc", "[", "a0", "]"}]}], "]"}]}], TraditionalForm]]],
 "\nRather than deriving a contradiction we show the negation of (12), i.e. \
we show\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "a0", "]"}]}], "]"}], TraditionalForm]]],
 "\nIf we can prove ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{"lp", "[", "a0", "]"}]}], TraditionalForm]]],
 " we are done, because then (G7) follows immediately from (7), so we try to \
prove\n(G8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{"lp", "[", "a0", "]"}]}], TraditionalForm]]],
 "\nnow. The first case in (10) is just what we want to show, so we only \
consider the second case in (10) with\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "a0", "]"}], "\[Succeeds]", "t"}], TraditionalForm]]],
 "\nSince we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "a0", "]"}], "=", 
    RowBox[{"lp", "[", "p", "]"}]}], TraditionalForm]]],
 " we also know\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", "t"}], TraditionalForm]]],
 "\ncontradicting (8).\n\nCase II.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", 
     RowBox[{"lp", "[", "a0", "]"}]}], "\[And]", 
    RowBox[{"c", "\[NotEqual]", "0"}]}], TraditionalForm]]],
 ".\nIndependent of which case holds in (10) we know\n(15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[Succeeds]", "t"}], TraditionalForm]]],
 "\ncontradicting (8)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632734307196405*^9, 3.6327345581004543`*^9}, {
   3.632734616426852*^9, 3.632734618295856*^9}, {3.632735595209683*^9, 
   3.632735772537394*^9}, {3.6327362213557796`*^9, 3.632736244635401*^9}, {
   3.632736275555282*^9, 3.6327363234943743`*^9}, {3.632736367094845*^9, 
   3.6327363979326878`*^9}, {3.632736454573246*^9, 3.6327365614484587`*^9}, {
   3.632736632823606*^9, 3.632736807324461*^9}, {3.6327371478433037`*^9, 
   3.632737219634656*^9}, {3.632737269975819*^9, 3.632737336124447*^9}, {
   3.6327373758137712`*^9, 3.632737451234105*^9}, {3.6327375975639477`*^9, 
   3.632737709967437*^9}, {3.632737754947998*^9, 3.632737876251749*^9}, {
   3.6327379420373297`*^9, 3.632737999156783*^9}, {3.632738052305998*^9, 
   3.632738224440189*^9}, 3.6327382617008543`*^9, 3.636440268314497*^9, {
   3.639206123847631*^9, 3.639206525521995*^9}, 3.643446729577608*^9, 
   3.653207990475519*^9, 3.6532135320364695`*^9, 3.666173824571622*^9, 
   3.6664404305944757`*^9},
 CellID->699080535],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632738249603307*^9, 3.632738261700988*^9}, 
   3.636440268314617*^9, 3.643446729577726*^9, 3.653207990475519*^9, 
   3.6532135320364695`*^9, 3.666173824571622*^9, 3.666440430594615*^9},
 TextAlignment->Right,
 CellID->1233787267]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1293427525],

Cell[CellGroupData[{

Cell["\<\
Lemma: Sufficient Conditions for areCPMultipliers in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.6433550420577*^9, {3.643355560161807*^9, 3.643355566250087*^9}, {
   3.643362410638322*^9, 3.643362415847183*^9}, {3.643430808675501*^9, 
   3.6434308117475853`*^9}, {3.643443746975107*^9, 3.6434437502423267`*^9}},
 CellID->2035494476],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1489378443}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643341998437498*^9, 3.643342023383795*^9}, {
   3.643355009942375*^9, 3.643355027766646*^9}, {3.643355573168229*^9, 
   3.643355582097849*^9}, 3.6433560021887703`*^9, 3.643362515887623*^9},
 CellTags->{
  "ID:1489378443", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1489378443],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 196663580}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643342032955171*^9, 3.64334204558191*^9}},
 CellTags->{
  "ID:196663580", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->196663580],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{"c", ",", "m1", ",", "m2"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1830388955}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643430830230946*^9, 3.64343084173035*^9}},
 CellTags->{
  "ID:1830388955", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1830388955],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{"p", "\[Equal]", "q"}], "\[Or]", 
         RowBox[{
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
          "\[NotEqual]", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}]},
       {
        RowBox[{
         UnderscriptBox["areCPMultipliers", "\[ScriptCapitalR]"], "[", 
         RowBox[{"m1", ",", "m2", ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", 
          "c", ",", "i", ",", "j"}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["areCPMultipliers", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"m1", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], "]"}], 
       UnderscriptBox["/", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], ")"}]}],
     ",", 
    RowBox[{"m2", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], "]"}], 
       UnderscriptBox["/", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], ")"}]}],
     ",", "p", ",", "q", ",", 
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     RowBox[{
      UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], "]"}]}],
     ",", "i", ",", "j"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CP1.\[DoubleLeftArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:312371876", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643430865560547*^9, 3.6434309285710917`*^9}, {
  3.643431622880054*^9, 3.6434316251337*^9}},
 CellTags->{
  "ID:312371876", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CP1.\[DoubleLeftArrow]"},
 CellID->312371876],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"p", "\[NotEqual]", "q"}]},
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
         "\[Equal]", 
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}]},
       {
        RowBox[{
         UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
         RowBox[{
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
          "c"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
         RowBox[{"c", ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
          "m1"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
         RowBox[{"c", ",", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
          "m2"}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["areCPMultipliers", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"m1", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], "]"}], 
       UnderscriptBox["/", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], ")"}]}],
     ",", 
    RowBox[{"m2", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], "]"}], 
       UnderscriptBox["/", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], ")"}]}],
     ",", "p", ",", "q", ",", 
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     RowBox[{
      UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], "]"}]}],
     ",", "i", ",", "j"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(CP2.\[DoubleLeftArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1006612102", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643344919981049*^9, 3.6433449802084084`*^9}, {
  3.6433451536938267`*^9, 3.6433451700233383`*^9}, {3.6433481381952553`*^9, 
  3.643348169711582*^9}, {3.6433555990836763`*^9, 3.643355618349421*^9}, {
  3.643355933407064*^9, 3.6433559467908773`*^9}, {3.643355989820022*^9, 
  3.6433559984735813`*^9}, {3.643356251997814*^9, 3.643356266011039*^9}, {
  3.643356358260627*^9, 3.643356384828767*^9}, {3.643429339706211*^9, 
  3.643429343302011*^9}, {3.643430843993629*^9, 3.6434308547566433`*^9}},
 CellTags->{
  "ID:1006612102", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CP2.\[DoubleLeftArrow]"},
 CellID->1006612102],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->469925811]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->633893238],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CP1.\[DoubleLeftArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CP1.\[DoubleLeftArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCPMultipliers", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["areCPMultipliers", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], ",", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
               ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCPMultipliers", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["areCPMultipliers", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], ",", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
               ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:312371876"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p312371876-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p312371876-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:312371876", "Proof|ID:312371876-1"},
 CellID->650652383],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p312371876-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->1483827355]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(CP2.\[DoubleLeftArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(CP2.\[DoubleLeftArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["q", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isCritPair", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], ",", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
               ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["q", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isCritPair", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], ",", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
               ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1006612102"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1006612102-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1006612102-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1006612102", "Proof|ID:1006612102-1"},
 CellID->719415418],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1006612102-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellChangeTimes->{3.643431551572083*^9, 3.643433206730365*^9},
 CellID->889161277]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R7 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.632216326196527*^9, 3.632216326495267*^9}, {
  3.639304380709499*^9, 3.639304384940322*^9}},
 CellID->1847430199],

Cell[TextData[{
 StyleBox["Remark",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 ": For proving R7 in \[ScriptCapitalP] we have to use the fact that R1 and \
R4 hold in \[ScriptCapitalP], and that ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], TraditionalForm]]],
 " is a partial irreflexive order relation."
}], "Text",
 CellChangeTimes->{{3.6328057722148314`*^9, 3.632805791348641*^9}, {
  3.6328077082384872`*^9, 3.632807729276582*^9}, {3.6330654334944773`*^9, 
  3.6330654573099833`*^9}, {3.633168870114286*^9, 3.633168871829277*^9}, {
  3.639368414176145*^9, 3.6393684225461283`*^9}},
 CellID->1835280531],

Cell["", "OpenEnvironment",
 CellID->500926893],

Cell[CellGroupData[{

Cell["Lemma", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.632738677275268*^9},
 CellID->623726226],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 542206690}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6327390399612303`*^9, 3.632739069552567*^9}, {
  3.635748126951478*^9, 3.635748129605212*^9}},
 CellTags->{
  "ID:542206690", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->542206690],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", "a", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1548368737}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6327391790991898`*^9, 3.632739187158772*^9}},
 CellTags->{
  "ID:1548368737", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1548368737],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 73478134}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632739189892147*^9, 3.632739202467108*^9}},
 CellTags->{
  "ID:73478134", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->73478134],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
   RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{"k", ",", 
      RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"a0", ",", "m"}], "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            RowBox[{"a0", "\[Equal]", "a"}], "\[Or]", 
            RowBox[{"a", 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], "a0"}]}]},
          {
           RowBox[{
            UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
            RowBox[{"p", ",", "q", ",", "a0", ",", "k", ",", "l"}], "]"}]},
          {
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
              RowBox[{"[", 
               RowBox[{"m0", ",", "c"}], "]"}]}]], 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{
                RowBox[{"c", "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                RowBox[{
                 UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                 RowBox[{"m0", ",", "c"}], "]"}]}], 
               TagBox[")",
                "AutoParentheses"]}], "\[Implies]", 
              RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                RowBox[{"m", 
                 UnderscriptBox["*", "\[ScriptCapitalR]"], "m0"}], ",", "c"}],
                "]"}]}], 
             TagBox[")",
              "AutoParentheses"]}]}]},
          {
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", "c", 
              "]"}]], 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{
              RowBox[{"a0", 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{"a0", 
                UnderscriptBox["+", "\[ScriptCapitalR]"], "c"}]}], 
              "\[Implies]", 
              RowBox[{"a", 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
               RowBox[{"a", 
                UnderscriptBox["+", "\[ScriptCapitalR]"], 
                RowBox[{"m", 
                 UnderscriptBox["*", "\[ScriptCapitalR]"], "c"}]}]}]}], 
             TagBox[")",
              "AutoParentheses"]}]}]},
          {
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
              RowBox[{"[", 
               RowBox[{"b", ",", "d", ",", "c"}], "]"}]}]], 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               TagBox[
                RowBox[{"b", 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{"{", "c", "}"}]], "\[ScriptCapitalR]"], "d"}],
                Identity], 
               TagBox[")",
                "AutoParentheses"]}], "\[Implies]", 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               TagBox[
                RowBox[{
                 RowBox[{"m", 
                  UnderscriptBox["*", "\[ScriptCapitalR]"], "b"}], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   RowBox[{"{", "c", "}"}]], "\[ScriptCapitalR]"], 
                 RowBox[{"m", 
                  UnderscriptBox["*", "\[ScriptCapitalR]"], "d"}]}],
                Identity], 
               TagBox[")",
                "AutoParentheses"]}]}], 
             TagBox[")",
              "AutoParentheses"]}]}]},
          {
           RowBox[{
            RowBox[{"p", "\[Equal]", "q"}], "\[Implies]", 
            RowBox[{
             UnderscriptBox["\[ForAll]", 
              RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               RowBox[{"m1", ",", "m2"}], "]"}]], 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                RowBox[{
                 RowBox[{
                  UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                  RowBox[{"a0", ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}], ",", "m1"}], "]"}], "\[And]", 
                 RowBox[{
                  UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                  RowBox[{"a0", ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}], ",", "m2"}], "]"}], "\[And]", 
                 RowBox[{
                  UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                  RowBox[{"m1", ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}], ",", "k"}], "]"}], "\[And]", 
                 RowBox[{
                  UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                  RowBox[{"m2", ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}], ",", "l"}], "]"}]}], 
                TagBox[")",
                 "AutoParentheses"]}], "\[Implies]", 
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                RowBox[{
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  RowBox[{
                   RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{"m", 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], "m1"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}], ",", "i"}], "]"}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{"m", 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], "m2"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}], ",", "j"}], "]"}]}], 
                  TagBox[")",
                   "AutoParentheses"]}], "\[Or]", 
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  RowBox[{
                   RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{"m", 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], "m1"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}], ",", "j"}], "]"}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{"m", 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], "m2"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                    "]"}], ",", "i"}], "]"}]}], 
                  TagBox[")",
                   "AutoParentheses"]}]}], 
                TagBox[")",
                 "AutoParentheses"]}]}], 
              TagBox[")",
               "AutoParentheses"]}]}]}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(409)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:310376772", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632738575306679*^9, 3.6327386155997343`*^9}, {
  3.632738653917617*^9, 3.63273866998032*^9}, {3.632738710230608*^9, 
  3.632738792517849*^9}, {3.6327388279670677`*^9, 3.632739002334687*^9}, {
  3.632739137920375*^9, 3.632739163228395*^9}, {3.632809772102398*^9, 
  3.632809833855438*^9}, {3.6618489706726913`*^9, 3.661848975087499*^9}},
 CellTags->{
  "ID:310376772", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "409"},
 CellID->310376772],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1898204898]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->723680751],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(409)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(409)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
                 RowBox[{
                   RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["a0", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["a0", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["a", "ExpressionVariable"]}], "\[Or]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    StyleBox["a0", "ExpressionVariable"]}]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["m0", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["m0", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["m0", "ExpressionVariable"]}], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}], 
                 "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], 
                    UnderscriptBox["+", "\[ScriptCapitalR]"], 
                    StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["+", "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["c", "ExpressionVariable"]], 
                    "\[ScriptCapitalR]"], 
                    StyleBox["d", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["b", "ExpressionVariable"]}], 
                    UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["c", "ExpressionVariable"]], 
                    "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["d", "ExpressionVariable"]}]}]}]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["l", "ExpressionVariable"]}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["m1", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["m2", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["m1", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["m2", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCRP", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
                 RowBox[{
                   RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["a0", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["a0", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["a", "ExpressionVariable"]}], "\[Or]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    StyleBox["a0", "ExpressionVariable"]}]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["m0", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["m0", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["m0", "ExpressionVariable"]}], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}], 
                 "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], 
                    UnderscriptBox["+", "\[ScriptCapitalR]"], 
                    StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["+", "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["c", "ExpressionVariable"]], 
                    "\[ScriptCapitalR]"], 
                    StyleBox["d", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["b", "ExpressionVariable"]}], 
                    UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["c", "ExpressionVariable"]], 
                    "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["d", "ExpressionVariable"]}]}]}]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Implies]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["l", "ExpressionVariable"]}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["m1", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["m2", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["m1", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["m2", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:310376772"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p310376772-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p310376772-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:310376772", "Proof|ID:310376772-1"},
 CellID->1336876000],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p310376772-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.642498418354044*^9, 3.6424984200497103`*^9}},
 CellID->1675988726]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632739248965032*^9, 3.632739249897559*^9}, 
   3.632741242008148*^9, 3.636440302349958*^9, 3.643446743371912*^9, 
   3.6532079723784876`*^9, 3.6532135173412437`*^9, 3.6661738125742006`*^9, 
   3.666440419344433*^9},
 CellID->2024136642],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.63274096228076*^9, 3.632740963626439*^9}, 
   3.632741242008275*^9, 3.63644030235008*^9, 3.643446743372031*^9, 
   3.6532079723784876`*^9, 3.6532135173412437`*^9, 3.6661738125742006`*^9, 
   3.666440419344533*^9},
 CellID->460882741],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c", "\[NotEqual]", "0"}], "\[Implies]", 
    RowBox[{"1", "\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalN]", 
      RowBox[{"1", ",", "\[Ellipsis]", ",", 
       RowBox[{"multN", "[", "c", "]"}]}]]}]}], TraditionalForm]]],
 " (follows from (R0)) (Reduction Ring lemmas)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632741002986086*^9, 3.632741021434413*^9}, 
   3.632741242008757*^9, {3.63317637418386*^9, 3.633176401250164*^9}, 
   3.6364403023501997`*^9, 3.6434467433721123`*^9, 3.6532079723784876`*^9, 
   3.6532135173412437`*^9, 3.6661738125742006`*^9, 3.666440419344618*^9},
 CellID->1279103172],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "a", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand show the RHS of the implication.\nFrom (isNTCRP alternative) we can \
infer\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"{", GridBox[{
      {
       RowBox[{"isNTCR", "[", 
        RowBox[{
         RowBox[{"lc", "[", "p", "]"}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "a", ",", "i", ",", "j"}], "]"}],
        "\[DoubleLongLeftArrow]", 
       RowBox[{"p", "=", 
        RowBox[{"q", "\[Or]", 
         RowBox[{
          RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
          RowBox[{"lc", "[", "q", "]"}]}]}]}]},
      {
       RowBox[{"isNTCR", "[", 
        RowBox[{
         RowBox[{"lc", "[", "q", "]"}], ",", "a"}], "]"}], 
       "\[DoubleLongLeftArrow]", "otherwise"}
     }]}], TraditionalForm]]],
 "\nBased on (2) we distinguish two cases.\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Or]", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
      RowBox[{"lc", "[", "q", "]"}]}]}]}], TraditionalForm]]],
 ".\nWe know\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nHence, by (R7) in \[ScriptCapitalR], we can conclude that there exist ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", ",", 
    RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", 
     RowBox[{"lc", "[", "p", "]"}], "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"a0", ",", "m"}], "]"}]}], TraditionalForm]]],
 " s.t.\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a0", "=", 
    RowBox[{"a", "\[Or]", 
     RowBox[{"a", "\[Succeeds]", "a0"}]}]}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a0", ",", "k", ",", "l"}], "]"}], 
   TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"m0", ",", "c"}], "]"}]}]], 
     RowBox[{"mult", "[", 
      RowBox[{"m0", ",", "c"}], "]"}]}], "\[Implies]", 
    RowBox[{"mult", "[", 
     RowBox[{
      RowBox[{"m", "*", "m0"}], ",", "c"}], "]"}]}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", "c", "]"}]}]], 
     RowBox[{"a0", "\[Succeeds]", 
      RowBox[{"a0", "+", "c"}]}]}], "\[Implies]", 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"a", "+", 
      RowBox[{"m", "*", "c"}]}]}]}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"b", ",", "c", ",", "d"}], "]"}]}]], "b"}], 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{"d", "\[Implies]", 
     RowBox[{
      RowBox[{"m", "*", "b"}], 
      SubscriptBox["\[LeftRightArrow]", "c"], 
      RowBox[{"m", "*", "d"}]}]}]}], TraditionalForm]]],
 "\n(9)\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "p", "]"}], "=", 
    RowBox[{
     RowBox[{"lc", "[", "q", "]"}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
        RowBox[{"[", 
         RowBox[{"m1", ",", "m2"}], "]"}]}]], 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"isReducible", "[", 
          RowBox[{"a0", ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "m1"}], "]"}], "\[And]", 
         RowBox[{"isReducible", "[", 
          RowBox[{"a0", ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "m2"}], "]"}], "\[And]", 
         RowBox[{"mult", "[", 
          RowBox[{"m1", ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "k"}], "]"}], "\[And]", 
         RowBox[{"mult", "[", 
          RowBox[{"m2", ",", 
           RowBox[{"lc", "[", "p", "]"}], ",", "l"}], "]"}]}], "\[Implies]", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"mult", "[", 
             RowBox[{
              RowBox[{"m", "*", "m1"}], ",", 
              RowBox[{"lc", "[", "p", "]"}], ",", "i"}], "]"}], "\[And]", 
            RowBox[{"mult", "[", 
             RowBox[{
              RowBox[{"m", "*", "m2"}], ",", 
              RowBox[{"lc", "[", "p", "]"}], ",", "j"}], "]"}]}], ")"}], 
          "\[Or]", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"mult", "[", 
             RowBox[{
              RowBox[{"m", "*", "m1"}], ",", 
              RowBox[{"lc", "[", "p", "]"}], ",", "j"}], "]"}], "\[And]", 
            RowBox[{"mult", "[", 
             RowBox[{
              RowBox[{"m", "*", "m2"}], ",", 
              RowBox[{"lc", "[", "p", "]"}], ",", "i"}], "]"}]}], ")"}]}], 
         ")"}]}], ")"}]}]}]}], TraditionalForm]]],
 "\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", ",", "l", ",", "a0", ",", "m"}], TraditionalForm]]],
 " witness our existential goal. Indeed, the first, third, fourth, fifth and \
sixth conjuncts follow immediately from (4), (6), (7), (8) and (9) (note that \
if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", "q"}], TraditionalForm]]],
 ", then also ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "p", "]"}], "=", 
    RowBox[{"lc", "[", "q", "]"}]}], TraditionalForm]]],
 "), respectively, so we only consider the second conjunct. Also note that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"multN", "[", "p", "]"}], "=", 
    RowBox[{"multN", "[", 
     RowBox[{"lc", "[", "p", "]"}], "]"}]}], TraditionalForm]]],
 ", justifying the choice of ",
 Cell[BoxData[
  FormBox["k", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["l", TraditionalForm]]],
 ".\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a0", ",", "k", ",", "l"}], "]"}], 
   TraditionalForm]]],
 "\n\nIn order to show (G1), by (isMNTCRP alternative 1) and our case \
assumption, it suffices to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "a0", ",", "k", ",", "l"}], "]"}], 
   TraditionalForm]]],
 "; This, however, is exactly the statement of (5).\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"p", "\[NotEqual]", "q"}], "\[And]", 
     RowBox[{"lc", "[", "p", "]"}]}], "=", 
    RowBox[{"lc", "[", "q", "]"}]}], TraditionalForm]]],
 ".\nWe know\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "q", "]"}], ",", "a"}], "]"}], TraditionalForm]]],
 "\nHence, by (R8) in \[ScriptCapitalR], we can conclude that there exist ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"a0", ",", "m"}], "]"}]}], TraditionalForm]]],
 " s.t.\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a0", "=", 
    RowBox[{"a", "\[Or]", 
     RowBox[{"a", "\[Succeeds]", "a0"}]}]}], TraditionalForm]]],
 "\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "q", "]"}], ",", "a0"}], "]"}], TraditionalForm]]],
 "\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"m0", ",", "c"}], "]"}]}]], 
     RowBox[{"mult", "[", 
      RowBox[{"m0", ",", "c"}], "]"}]}], "\[Implies]", 
    RowBox[{"mult", "[", 
     RowBox[{
      RowBox[{"m", "*", "m0"}], ",", "c"}], "]"}]}], TraditionalForm]]],
 "\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", "c", "]"}]}]], 
     RowBox[{"a0", "\[Succeeds]", 
      RowBox[{"a0", "+", "c"}]}]}], "\[Implies]", 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"a", "+", 
      RowBox[{"m", "*", "c"}]}]}]}], TraditionalForm]]],
 "\n(15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"b", ",", "c", ",", "d"}], "]"}]}]], "b"}], 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{"d", "\[Implies]", 
     RowBox[{
      RowBox[{"m", "*", "b"}], 
      SubscriptBox["\[LeftRightArrow]", "c"], 
      RowBox[{"m", "*", "d"}]}]}]}], TraditionalForm]]],
 "\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", ":=", "1"}], ",", 
    RowBox[{"l", ":=", "1"}], ",", 
    RowBox[{"a0", ":=", "a0"}], ",", 
    RowBox[{"m", ":=", "m"}]}], TraditionalForm]]],
 " witness our existential goal. Indeed, the first, third, fourth and fifth \
conjuncts follow immediately from (11), (13), (14) and (15), respectively, \
and the sixth conjunct holds vacuously since its premise, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", "q"}], TraditionalForm]]],
 ", contradicts our case assumption. Hence, again we only need to consider \
the second conjunct:\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "a0", ",", "1", ",", "1"}], "]"}], 
   TraditionalForm]]],
 "\n\nIn order to show (G2), by (isMNTCRP alternative 2) and our case \
assumption, it suffices to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "q", "]"}], ",", "a0"}], "]"}], TraditionalForm]]],
 "; This, however, is exactly the statement of (12)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632739499031385*^9, 3.632739580389305*^9}, {
   3.6327396120952682`*^9, 3.632740306955562*^9}, {3.6327403684264193`*^9, 
   3.632740801011734*^9}, {3.632740872242289*^9, 3.632740887776184*^9}, {
   3.63274094912639*^9, 3.632740953059771*^9}, {3.632741113681357*^9, 
   3.632741242008862*^9}, 3.636440302350299*^9, {3.639211548730268*^9, 
   3.639211554996353*^9}, 3.643446743372219*^9, 3.6532079723784876`*^9, 
   3.6532135173412437`*^9, 3.6661738125742006`*^9, 3.666440419344709*^9},
 CellID->257775101],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632741228564595*^9, 3.632741242008993*^9}, 
   3.636440302350424*^9, 3.643446743372335*^9, 3.6532079723784876`*^9, 
   3.6532135173412437`*^9, 3.6661738125742006`*^9, 3.666440419344805*^9},
 TextAlignment->Right,
 CellID->670854715]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->738627761],

Cell[CellGroupData[{

Cell["\<\
Lemma: Auxiliary Results needed for proving R7 and R8 in \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632803798050151*^9, 3.632803812428692*^9}, {
  3.639368407581311*^9, 3.639368409181497*^9}},
 CellID->1572614892],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", 
    RowBox[{"u", ",", "s"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 383946867}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63280406395601*^9, 3.6328040760201197`*^9}, 
   3.6328041596866426`*^9, {3.6328090146900177`*^9, 3.6328090148382177`*^9}},
 CellTags->{
  "ID:383946867", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->383946867],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{"c", ",", "n"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1587705005}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63280409325425*^9, 3.6328041011151543`*^9}, {
  3.6328087689009037`*^9, 3.6328087691588697`*^9}},
 CellTags->{
  "ID:1587705005", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1587705005],

Cell[TextData[{
 "(R7.\[ScriptCapitalP].3) and (R7.\[ScriptCapitalP].5) hold for any power \
product ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 ", not only for power products of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"u", "/", "s"}], TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.6328221920991*^9, 3.632822239074996*^9}},
 CellID->1670988629],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"\[Gamma]", ",", "\[Mu]"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"\[Gamma]", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
        RowBox[{
         UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
         RowBox[{"\[Mu]", ",", "\[Gamma]"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{
       UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
       RowBox[{
        RowBox[{"n", 
         UnderscriptBox["*", "\[ScriptCapitalR]"], "\[Mu]"}], ",", 
        "\[Gamma]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"c0", ",", "m0"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"c0", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
        RowBox[{
         UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
         RowBox[{"m0", ",", "c0"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{
       UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"n", 
           UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "s"}], ")"}], 
         
         UnderscriptBox["*", "\[ScriptCapitalP]"], "m0"}], ",", "c0"}], 
       "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R7.\[ScriptCapitalP].3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:549840930", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6328089081872396`*^9, 3.632808989436584*^9}, {
  3.632809029156621*^9, 3.632809058331478*^9}, {3.632809209692336*^9, 
  3.6328092321325417`*^9}, {3.632809723837893*^9, 3.632809756756978*^9}, {
  3.63307074400336*^9, 3.633070757860091*^9}},
 CellTags->{
  "ID:549840930", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].3"},
 CellID->549840930],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
        RowBox[{"[", 
         RowBox[{"\[Beta]", ",", "\[Gamma]", ",", "\[Delta]"}], "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"\[Beta]", 
          UnderscriptBox[
           SubscriptBox["\[Rule]", 
            RowBox[{"{", "\[Gamma]", "}"}]], "\[ScriptCapitalR]"], 
          "\[Delta]"}], 
         TagBox[")",
          "AutoParentheses"]}], "\[Implies]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{"n", 
           UnderscriptBox["*", "\[ScriptCapitalR]"], "\[Beta]"}], 
          UnderscriptBox[
           SubscriptBox["\[LeftRightArrow]", 
            RowBox[{"{", "\[Gamma]", "}"}]], "\[ScriptCapitalR]"], 
          RowBox[{"n", 
           UnderscriptBox["*", "\[ScriptCapitalR]"], "\[Delta]"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
        RowBox[{"[", 
         RowBox[{"\[Gamma]", ",", "\[Mu]"}], "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{"\[Gamma]", "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
          RowBox[{
           UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
           RowBox[{"\[Mu]", ",", "\[Gamma]"}], "]"}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[Implies]", 
        RowBox[{
         UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
         RowBox[{
          RowBox[{"n", 
           UnderscriptBox["*", "\[ScriptCapitalR]"], "\[Mu]"}], ",", 
          "\[Gamma]"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"b", ",", "d", ",", "c0"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"b", 
        UnderscriptBox[
         SubscriptBox["\[Rule]", 
          RowBox[{"{", "c0", "}"}]], "\[ScriptCapitalP]"], "d"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"n", 
           UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "s"}], ")"}], 
         
         UnderscriptBox["*", "\[ScriptCapitalP]"], "b"}], 
        UnderscriptBox[
         SubscriptBox["\[LeftRightArrow]", 
          RowBox[{"{", "c0", "}"}]], "\[ScriptCapitalP]"], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"n", 
           UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "s"}], ")"}], 
         
         UnderscriptBox["*", "\[ScriptCapitalP]"], "d"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R7.\[ScriptCapitalP].5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:657656821", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632821677714654*^9, 3.632821688496744*^9}, {
  3.632821759078532*^9, 3.632821877616391*^9}, {3.632822047829784*^9, 
  3.632822071954791*^9}, {3.6328221025462914`*^9, 3.632822124843265*^9}, {
  3.6330707037082663`*^9, 3.633070738268141*^9}, {3.661848995351935*^9, 
  3.6618490196567774`*^9}},
 CellTags->{
  "ID:657656821", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].5"},
 CellID->657656821],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "a", "]"}]}]], 
  RowBox[{"a", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 64232512}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632803994522361*^9, 3.632804047548959*^9}, {
  3.635748143288336*^9, 3.635748150362608*^9}, {3.639368323073279*^9, 
  3.6393683270611763`*^9}},
 CellTags->{
  "ID:64232512", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->64232512],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"a0", "=", 
   RowBox[{"c", 
    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
    "s"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 237038572}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632804220011156*^9, 3.632804228980853*^9}, {
   3.632804259784213*^9, 3.632804296115925*^9}, 3.6328219187145557`*^9},
 CellTags->{
  "ID:237038572", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->237038572],

Cell[BoxData[
 RowBox[{
  RowBox[{"s", 
   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "u"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 235501232}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632821955901535*^9, 3.63282196821811*^9}},
 CellTags->{
  "ID:235501232", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->235501232],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", "\[Equal]", 
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{"a", ",", "u"}], "]"}]}], "\[Or]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{"a", ",", "u"}], "]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], "c"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a0", "\[Equal]", "a"}], "\[Or]", 
    RowBox[{"a", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "a0"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R7.\[ScriptCapitalP].1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1513648693", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632803832959379*^9, 3.6328039432112417`*^9}, {
  3.6328041860696*^9, 3.6328042076469088`*^9}, {3.632804303987442*^9, 
  3.632804306784384*^9}, {3.6328044450956373`*^9, 3.6328044732736883`*^9}, {
  3.632805182119358*^9, 3.632805183447954*^9}, {3.632809017235734*^9, 
  3.632809018560741*^9}, {3.6328219244621973`*^9, 3.6328219279627247`*^9}, {
  3.632822019143318*^9, 3.632822023046672*^9}},
 CellTags->{
  "ID:1513648693", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].1"},
 CellID->1513648693],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "d0", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"c", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
       RowBox[{"c", 
        UnderscriptBox["+", "\[ScriptCapitalR]"], "d0"}]}], "\[Implies]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{"a", ",", "u"}], "]"}], 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
         RowBox[{"a", ",", "u"}], "]"}], 
        UnderscriptBox["+", "\[ScriptCapitalR]"], 
        RowBox[{"n", 
         UnderscriptBox["*", "\[ScriptCapitalR]"], "d0"}]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "c0", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"a0", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
       RowBox[{"a0", 
        UnderscriptBox["+", "\[ScriptCapitalP]"], "c0"}]}], "\[Implies]", 
      RowBox[{"a", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
       RowBox[{"a", 
        UnderscriptBox["+", "\[ScriptCapitalP]"], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"n", 
           UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
           RowBox[{"(", 
            RowBox[{"u", 
             UnderscriptBox["/", "\[ScriptCapitalT]"], "s"}], ")"}]}], ")"}], 
         
         UnderscriptBox["*", "\[ScriptCapitalP]"], "c0"}]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R7.\[ScriptCapitalP].4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:808269926", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632811311782866*^9, 3.6328113416483173`*^9}, {
   3.632811414403118*^9, 3.632811588666883*^9}, {3.632811621156762*^9, 
   3.6328116554499693`*^9}, {3.632811693765605*^9, 3.632811706344805*^9}, 
   3.632818990807625*^9, {3.6328219986362953`*^9, 3.632822029751239*^9}, {
   3.639218775138385*^9, 3.639218790916657*^9}, {3.6392991022427197`*^9, 
   3.6392991040377283`*^9}},
 CellTags->{
  "ID:808269926", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].4"},
 CellID->808269926],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->193282123]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->970754585],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R7.\[ScriptCapitalP].3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R7.\[ScriptCapitalP].3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["\[Gamma]", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["\[Mu]", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["\[Gamma]", "ExpressionVariable"], "\[NotEqual]", 
                   
                   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    StyleBox["\[Mu]", "ExpressionVariable"], ",", 
                    StyleBox["\[Gamma]", "ExpressionVariable"]}], "]"}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["\[Mu]", "ExpressionVariable"]}], ",", 
                   StyleBox["\[Gamma]", "ExpressionVariable"]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c0", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m0", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["c0", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["m0", "ExpressionVariable"], ",", 
                    StyleBox["c0", "ExpressionVariable"]}], "]"}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["s", "ExpressionVariable"]}], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox["m0", "ExpressionVariable"]}], ",", 
                   StyleBox["c0", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["\[Gamma]", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["\[Mu]", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["\[Gamma]", "ExpressionVariable"], "\[NotEqual]", 
                   
                   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    StyleBox["\[Mu]", "ExpressionVariable"], ",", 
                    StyleBox["\[Gamma]", "ExpressionVariable"]}], "]"}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["\[Mu]", "ExpressionVariable"]}], ",", 
                   StyleBox["\[Gamma]", "ExpressionVariable"]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c0", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m0", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["c0", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["m0", "ExpressionVariable"], ",", 
                    StyleBox["c0", "ExpressionVariable"]}], "]"}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["s", "ExpressionVariable"]}], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox["m0", "ExpressionVariable"]}], ",", 
                   StyleBox["c0", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:549840930"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p549840930-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p549840930-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:549840930", "Proof|ID:549840930-1"},
 CellID->1394810391],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p549840930-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->460317596]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R7.\[ScriptCapitalP].5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R7.\[ScriptCapitalP].5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["\[Beta]", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["\[Gamma]", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["\[Delta]", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["\[Beta]", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{"{", 
                    StyleBox["\[Gamma]", "ExpressionVariable"], "}"}]], 
                    "\[ScriptCapitalR]"], 
                   StyleBox["\[Delta]", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["\[Beta]", "ExpressionVariable"]}], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    RowBox[{"{", 
                    StyleBox["\[Gamma]", "ExpressionVariable"], "}"}]], 
                    "\[ScriptCapitalR]"], 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["\[Delta]", "ExpressionVariable"]}]}]}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["\[Gamma]", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["\[Mu]", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["\[Gamma]", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["\[Mu]", "ExpressionVariable"], ",", 
                    StyleBox["\[Gamma]", "ExpressionVariable"]}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["\[Mu]", "ExpressionVariable"]}], ",", 
                    StyleBox["\[Gamma]", "ExpressionVariable"]}], "]"}]}]}]}],
            "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["d", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c0", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{"{", 
                    StyleBox["c0", "ExpressionVariable"], "}"}]], 
                  "\[ScriptCapitalP]"], 
                 StyleBox["d", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["s", "ExpressionVariable"]}], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["b", "ExpressionVariable"]}], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   RowBox[{"{", 
                    StyleBox["c0", "ExpressionVariable"], "}"}]], 
                  "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["s", "ExpressionVariable"]}], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox[
                   "d", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["\[Beta]", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["\[Gamma]", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["\[Delta]", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["\[Beta]", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{"{", 
                    StyleBox["\[Gamma]", "ExpressionVariable"], "}"}]], 
                    "\[ScriptCapitalR]"], 
                   StyleBox["\[Delta]", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["\[Beta]", "ExpressionVariable"]}], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    RowBox[{"{", 
                    StyleBox["\[Gamma]", "ExpressionVariable"], "}"}]], 
                    "\[ScriptCapitalR]"], 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["\[Delta]", "ExpressionVariable"]}]}]}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["\[Gamma]", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["\[Mu]", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["\[Gamma]", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    StyleBox["\[Mu]", "ExpressionVariable"], ",", 
                    StyleBox["\[Gamma]", "ExpressionVariable"]}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["\[Mu]", "ExpressionVariable"]}], ",", 
                    StyleBox["\[Gamma]", "ExpressionVariable"]}], "]"}]}]}]}],
            "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["d", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c0", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{"{", 
                    StyleBox["c0", "ExpressionVariable"], "}"}]], 
                  "\[ScriptCapitalP]"], 
                 StyleBox["d", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["s", "ExpressionVariable"]}], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox["b", "ExpressionVariable"]}], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   RowBox[{"{", 
                    StyleBox["c0", "ExpressionVariable"], "}"}]], 
                  "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["s", "ExpressionVariable"]}], 
                   UnderscriptBox["*", "\[ScriptCapitalP]"], 
                   StyleBox[
                   "d", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:657656821"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p657656821-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p657656821-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:657656821", "Proof|ID:657656821-1"},
 CellID->75572869],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p657656821-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->69597396]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R7.\[ScriptCapitalP].1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R7.\[ScriptCapitalP].1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["u", "ExpressionVariable"]}], "]"}]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["u", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                 StyleBox["c", "ExpressionVariable"]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["s", "ExpressionVariable"]}], "\[Equal]", 
               StyleBox["a", "ExpressionVariable"]}], "\[Or]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox[
                 "s", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["u", "ExpressionVariable"]}], "]"}]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["u", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                 StyleBox["c", "ExpressionVariable"]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["s", "ExpressionVariable"]}], "\[Equal]", 
               StyleBox["a", "ExpressionVariable"]}], "\[Or]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox[
                 "s", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1513648693"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1513648693-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1513648693-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1513648693", "Proof|ID:1513648693-1"},
 CellID->1424817208],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1513648693-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1180617287]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R7.\[ScriptCapitalP].4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R7.\[ScriptCapitalP].4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["d0", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["+", "\[ScriptCapitalR]"], 
                    StyleBox["d0", "ExpressionVariable"]}]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["u", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["u", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["+", "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["d0", "ExpressionVariable"]}]}]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c0", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["s", "ExpressionVariable"]}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["s", "ExpressionVariable"]}], 
                   UnderscriptBox["+", "\[ScriptCapitalP]"], 
                   StyleBox["c0", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["+", "\[ScriptCapitalP]"], 
                   RowBox[{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    RowBox[{
                    StyleBox["u", "ExpressionVariable"], 
                    UnderscriptBox["/", "\[ScriptCapitalT]"], 
                    StyleBox["s", "ExpressionVariable"]}]}], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox[
                    "c0", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               StyleBox["u", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["d0", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["+", "\[ScriptCapitalR]"], 
                    StyleBox["d0", "ExpressionVariable"]}]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["u", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["u", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["+", "\[ScriptCapitalR]"], 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["d0", "ExpressionVariable"]}]}]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
                   StyleBox["c0", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["s", "ExpressionVariable"]}], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    StyleBox["s", "ExpressionVariable"]}], 
                   UnderscriptBox["+", "\[ScriptCapitalP]"], 
                   StyleBox["c0", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["+", "\[ScriptCapitalP]"], 
                   RowBox[{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                    RowBox[{
                    StyleBox["u", "ExpressionVariable"], 
                    UnderscriptBox["/", "\[ScriptCapitalT]"], 
                    StyleBox["s", "ExpressionVariable"]}]}], 
                    UnderscriptBox["*", "\[ScriptCapitalP]"], 
                    StyleBox[
                    "c0", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:808269926"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p808269926-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p808269926-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:808269926", "Proof|ID:808269926-1"},
 CellID->876551560],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p808269926-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->1906600593]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (R7.\[ScriptCapitalP].3)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632809128566924*^9, 3.632809137700878*^9}, 
   3.632810936256651*^9, 3.6364403216539917`*^9, 3.64344674964983*^9, 
   3.6532079608968673`*^9, 3.6532135083244276`*^9, 3.6661738036197853`*^9, 
   3.6664404110885*^9},
 CellID->1496383118],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "s", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "n", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"c0", ",", "m0"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"d0", ",", "n0"}], "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"d0", "\[NotEqual]", "0"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"n0", ",", "d0"}], "]"}]}], "\[Implies]", 
      RowBox[{"mult", "[", 
       RowBox[{
        RowBox[{"n", "*", "n0"}], ",", "d0"}], "]"}]}], ")"}]}], 
   TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c0", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m0", ",", "c0"}], "]"}], TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "m0"}], ",", "c0"}], 
    "]"}], TraditionalForm]]],
 "\n\nFrom (3), together with (384), we can infer\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "m0", "]"}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"lc", "[", "m0", "]"}], ",", 
     RowBox[{"lc", "[", "c0", "]"}]}], "]"}], TraditionalForm]]],
 "\n\nFrom (4) we know that we can rewrite ",
 Cell[BoxData[
  FormBox["m0", TraditionalForm]]],
 " as ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m0", "]"}], "\[CenterDot]", 
    RowBox[{"lp", "[", "m0", "]"}]}], TraditionalForm]]],
 ", so we can replace every occurrence of ",
 Cell[BoxData[
  FormBox["m0", TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "m0", "]"}], "\[CenterDot]", 
    RowBox[{"lp", "[", "m0", "]"}]}], TraditionalForm]]],
 ". Hence, our goal becomes\n(G\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"lc", "[", "m0", "]"}], "\[CenterDot]", 
        RowBox[{"lp", "[", "m0", "]"}]}], ")"}]}], ",", "c0"}], "]"}], 
   TraditionalForm]]],
 "\nA lemma on multiplication of monomials allows us to further rewrite our \
goal into\n(G\[CloseCurlyQuote]\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"n", "*", 
        RowBox[{"lc", "[", "m0", "]"}]}], ")"}], "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{"s", "*", 
        RowBox[{"lp", "[", "m0", "]"}]}], ")"}]}], ",", "c0"}], "]"}], 
   TraditionalForm]]],
 "\n\nNow, we instantiate (1) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"d0", ":=", 
     RowBox[{"lc", "[", "c0", "]"}]}], ",", 
    RowBox[{"n0", ":=", 
     RowBox[{"lc", "[", "m0", "]"}]}]}], TraditionalForm]]],
 ". Since clearly ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "c0", "]"}], "\[NotEqual]", "0"}], TraditionalForm]]],
 " we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"mult", "[", 
     RowBox[{
      RowBox[{"lc", "[", "m0", "]"}], ",", 
      RowBox[{"lc", "[", "c0", "]"}]}], "]"}], "\[Implies]", 
    RowBox[{"mult", "[", 
     RowBox[{
      RowBox[{"n", "*", 
       RowBox[{"lc", "[", "m0", "]"}]}], ",", 
      RowBox[{"lc", "[", "c0", "]"}]}], "]"}]}], TraditionalForm]]],
 "; The premise of the implication holds because of (5), so we know\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"n", "*", 
      RowBox[{"lc", "[", "m0", "]"}]}], ",", 
     RowBox[{"lc", "[", "c0", "]"}]}], "]"}], TraditionalForm]]],
 "\n\nAs an intermediate subgoal, we show now\n(C)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"n", "*", 
     RowBox[{"lc", "[", "m0", "]"}]}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\nAssume otherwise, i.e. assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"n", "*", 
     RowBox[{"lc", "[", "m0", "]"}]}], "=", "0"}], TraditionalForm]]],
 ". Then, from (8), we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"0", ",", 
     RowBox[{"lc", "[", "c0", "]"}]}], "]"}], TraditionalForm]]],
 ", and this is not possible (due to (R2) in \[ScriptCapitalR]). So, we may \
add\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"n", "*", 
     RowBox[{"lc", "[", "m0", "]"}]}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\nto our knowledge base.\n\n(384) also allows us to rewrite our goal (G\
\[CloseCurlyQuote]\[CloseCurlyQuote]) into\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"n", "*", 
       RowBox[{"lc", "[", "m0", "]"}]}], ")"}], "\[CenterDot]", 
     RowBox[{"(", 
      RowBox[{"s", "*", 
       RowBox[{"lp", "[", "m0", "]"}]}], ")"}]}], "]"}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"lc", "[", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "*", 
         RowBox[{"lc", "[", "m0", "]"}]}], ")"}], "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{"s", "*", 
         RowBox[{"lp", "[", "m0", "]"}]}], ")"}]}], "]"}], ",", 
     RowBox[{"lc", "[", "c0", "]"}]}], "]"}], TraditionalForm]]],
 "\n\n(G1):\nWe only have to make sure that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"n", "*", 
     RowBox[{"lc", "[", "m0", "]"}]}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 ", which is exactly what (9) states.\n\n(G2):\nHere we only have to replace ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lc", "[", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"n", "*", 
       RowBox[{"lc", "[", "m0", "]"}]}], ")"}], "\[CenterDot]", 
     RowBox[{"(", 
      RowBox[{"s", "*", 
       RowBox[{"lp", "[", "m0", "]"}]}], ")"}]}], "]"}], TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "*", 
    RowBox[{"lc", "[", "m0", "]"}]}], TraditionalForm]]],
 ", which we are allowed to do because of (9). (G2) follows immediately from \
(8)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632809143443956*^9, 3.632809206437755*^9}, {
   3.632809244484617*^9, 3.632809346836152*^9}, {3.6328099556017933`*^9, 
   3.6328103428638077`*^9}, {3.632810386878722*^9, 3.6328104878341084`*^9}, {
   3.6328105378226137`*^9, 3.63281062821493*^9}, {3.632810687379601*^9, 
   3.632810936256777*^9}, 3.636440321654114*^9, {3.639214361301095*^9, 
   3.6392143757284613`*^9}, {3.6392144562355127`*^9, 3.63921462133044*^9}, 
   3.643446749649948*^9, 3.6532079608968673`*^9, 3.6532135083244276`*^9, 
   3.6661738036197853`*^9, 3.666440411088601*^9},
 CellID->673230165],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6328109236381607`*^9, 3.632810936256908*^9}, 
   3.6364403216542397`*^9, 3.643446749650064*^9, 3.6532079608968673`*^9, 
   3.6532135083244276`*^9, 3.6661738036197853`*^9, 3.666440411088704*^9},
 TextAlignment->Right,
 CellID->1990228853]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (R7.\[ScriptCapitalP].5)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632822253741558*^9, 3.6328222559903183`*^9}, {
   3.632825238397748*^9, 3.632825244791575*^9}, 3.633072175539754*^9, 
   3.636440333672433*^9, 3.6434467562897463`*^9, 3.653207953440054*^9, 
   3.653213502724018*^9, 3.666173797598174*^9, 3.666440406162043*^9},
 CellID->1002411843],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632822677774831*^9, 3.632822679196251*^9}, 
   3.632825238397874*^9, 3.636440333672592*^9, 3.643446756289865*^9, 
   3.653207953440054*^9, 3.653213502724018*^9, 3.666173797598174*^9, 
   3.666440406162183*^9},
 CellID->1700926014],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"b", 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{"d", "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"b", "\[NotEqual]", "0"}], "\[And]", 
       RowBox[{"c", "\[NotEqual]", "0"}]}], ")"}]}]}], TraditionalForm]]],
 " (Reduction Ring lemmas)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632822683947953*^9, 3.63282271554515*^9}, 
   3.632825238398328*^9, {3.6330696083937073`*^9, 3.633069661067485*^9}, {
   3.633071903362844*^9, 3.633071903378002*^9}, {3.63323475193501*^9, 
   3.6332347634131947`*^9}, 3.636440333672735*^9, 3.643446756289978*^9, 
   3.653207953440054*^9, 3.653213502724018*^9, 3.666173797598174*^9, 
   3.6664404061622953`*^9},
 CellID->2000946050],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"isReducible", "[", 
      RowBox[{"a", ",", "c", ",", "m"}], "]"}], "\[Implies]", "a"}], 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{"a", "-", 
     RowBox[{"m", "*", "c"}]}]}], TraditionalForm]]],
 " (in general); note that no multiplier is specified in the reduction. (RR \
Notions lemmas)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632822683947953*^9, 3.63282271554515*^9}, 
   3.632825238398328*^9, {3.6330696083937073`*^9, 3.633069661067485*^9}, {
   3.633071903362844*^9, 3.633071976166707*^9}, {3.63323477108777*^9, 
   3.6332347773294888`*^9}, 3.636440333672834*^9, 3.643446756290057*^9, 
   3.653207953440054*^9, 3.653213502724018*^9, 3.666173797598174*^9, 
   3.666440406162372*^9},
 CellID->438531627],

Cell["Special inference rules:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632824312088552*^9, 3.63282431647899*^9}, 
   3.632825238398436*^9, 3.6364403336729593`*^9, 3.643446756290153*^9, 
   3.653207953440054*^9, 3.653213502724018*^9, 3.666173797598174*^9, 
   3.666440406162483*^9},
 CellID->1584658543],

Cell["Arithmetic in polynomial rings", "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6328243218193827`*^9, 3.632824328249199*^9}, 
   3.632825238398534*^9, 3.6364403336730833`*^9, 3.643446756290247*^9, 
   3.653207953440054*^9, 3.653213502724018*^9, 3.666173797598174*^9, 
   3.66644040616258*^9},
 CellID->895588669],

Cell["Domain membership in polynomial domains.", "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6328243218193827`*^9, 3.632824337082931*^9}, 
   3.632825238398608*^9, 3.636440333673177*^9, 3.6434467562903214`*^9, 
   3.653207953440054*^9, 3.653213502724018*^9, 3.666173797598174*^9, 
   3.666440406162651*^9},
 CellID->1241320943],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "s", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "n", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"b", ",", "c0", ",", "d"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"\[Beta]", ",", "\[Gamma]", ",", "\[Delta]"}], "]"}]}]], 
    RowBox[{"(", 
     RowBox[{"\[Beta]", 
      SubscriptBox["\[Rule]", "\[Gamma]"], 
      RowBox[{"\[Delta]", "\[Implies]", 
       RowBox[{
        RowBox[{"n", "*", "\[Beta]"}], 
        SubscriptBox["\[LeftRightArrow]", "\[Gamma]"], 
        RowBox[{"n", "*", "\[Delta]"}]}]}]}], ")"}]}], TraditionalForm]]],
 "\n(1\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"\[Gamma]", ",", "\[Mu]"}], "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"\[Gamma]", "\[NotEqual]", "0"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"\[Mu]", ",", "\[Gamma]"}], "]"}]}], "\[Implies]", 
      RowBox[{"mult", "[", 
       RowBox[{
        RowBox[{"n", "*", "\[Mu]"}], ",", "\[Gamma]"}], "]"}]}], ")"}]}], 
   TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", 
    SubscriptBox["\[Rule]", "c0"], "d"}], TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "b"}], 
    SubscriptBox["\[LeftRightArrow]", "c0"], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "d"}]}], 
   TraditionalForm]]],
 "\nFrom (2) we know that there exists some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "m", "]"}]}], TraditionalForm]]],
 " with\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m", ",", "c0"}], "]"}], TraditionalForm]]],
 "\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", 
    SubscriptBox["\[Rule]", 
     RowBox[{"m", ",", "c0"}]], "d"}], TraditionalForm]]],
 "\nAlso from (2) we can infer\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c0", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n\nFrom (4) and (RED) we can infer\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"d", "=", 
    RowBox[{"b", "-", 
     RowBox[{"m", "*", "c0"}]}]}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"b", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m", "]"}], "*", 
        RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], ",", 
     RowBox[{"lc", "[", "c0", "]"}], ",", 
     RowBox[{"lc", "[", "m", "]"}]}], "]"}], TraditionalForm]]],
 "\n\nFrom (3) we can infer\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "m", "]"}], TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"lc", "[", "m", "]"}], ",", 
     RowBox[{"lc", "[", "c0", "]"}]}], "]"}], TraditionalForm]]],
 "\n\nFrom (8) we can infer\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"b", ",", 
      RowBox[{
       RowBox[{"lp", "[", "m", "]"}], "*", 
       RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], 
    SubscriptBox["\[Rule]", 
     RowBox[{"lc", "[", "c0", "]"}]], 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"b", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m", "]"}], "*", 
        RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], "-", 
     RowBox[{
      RowBox[{"lc", "[", "m", "]"}], "*", 
      RowBox[{"lc", "[", "c0", "]"}]}]}]}], TraditionalForm]]],
 "\nWe instantiate (1) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Beta]", ":=", 
     RowBox[{"C", "[", 
      RowBox[{"b", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m", "]"}], "*", 
        RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}], ",", 
    RowBox[{"\[Gamma]", ":=", 
     RowBox[{"lc", "[", "c0", "]"}]}], ",", 
    RowBox[{"\[Delta]", ":=", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"b", ",", 
        RowBox[{
         RowBox[{"lp", "[", "m", "]"}], "*", 
         RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], "-", 
      RowBox[{
       RowBox[{"lc", "[", "m", "]"}], "*", 
       RowBox[{"lc", "[", "c0", "]"}]}]}]}]}], TraditionalForm]]],
 ", and since the premise of the implication holds due to (11), we can infer\n\
(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"n", "*", 
     RowBox[{"C", "[", 
      RowBox[{"b", ",", 
       RowBox[{
        RowBox[{"lp", "[", "m", "]"}], "*", 
        RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}], 
    SubscriptBox["\[LeftRightArrow]", 
     RowBox[{"lc", "[", "c0", "]"}]], 
    RowBox[{"n", "*", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"b", ",", 
         RowBox[{
          RowBox[{"lp", "[", "m", "]"}], "*", 
          RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], "-", 
       RowBox[{
        RowBox[{"lc", "[", "m", "]"}], "*", 
        RowBox[{"lc", "[", "c0", "]"}]}]}], ")"}]}]}], TraditionalForm]]],
 "\n\nUsing (\[LeftRightArrow] alternative) we can infer from (12) (among \
others)\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"n", "*", 
      RowBox[{"C", "[", 
       RowBox[{"b", ",", 
        RowBox[{
         RowBox[{"lp", "[", "m", "]"}], "*", 
         RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}], "\[Succeeds]", 
     RowBox[{"n", "*", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"C", "[", 
         RowBox[{"b", ",", 
          RowBox[{
           RowBox[{"lp", "[", "m", "]"}], "*", 
           RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], "-", 
        RowBox[{
         RowBox[{"lc", "[", "m", "]"}], "*", 
         RowBox[{"lc", "[", "c0", "]"}]}]}], ")"}]}]}], "\[Or]", 
    RowBox[{
     RowBox[{"n", "*", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"C", "[", 
         RowBox[{"b", ",", 
          RowBox[{
           RowBox[{"lp", "[", "m", "]"}], "*", 
           RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], "-", 
        RowBox[{
         RowBox[{"lc", "[", "m", "]"}], "*", 
         RowBox[{"lc", "[", "c0", "]"}]}]}], ")"}]}], "\[Succeeds]", 
     RowBox[{"n", "*", 
      RowBox[{"C", "[", 
       RowBox[{"b", ",", 
        RowBox[{
         RowBox[{"lp", "[", "m", "]"}], "*", 
         RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}]}]}], TraditionalForm]]],
 "\n\nBy (\[LeftRightArrow] alternative \[ScriptCapitalP]) we can reduce our \
current goal (G) to\n(G\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", "\[Gamma]", "]"}]}]], 
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "t", "]"}]}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"mult", "[", 
         RowBox[{"\[Gamma]", ",", 
          RowBox[{"lc", "[", "c0", "]"}]}], "]"}]},
       {
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "d"}], "=", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "b"}], "-", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"\[Gamma]", "\[CenterDot]", "t"}], ")"}], "*", "c0"}]}]}]},
       {
        RowBox[{
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"(", 
              RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "b"}], ",", 
            RowBox[{"t", "*", 
             RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], "\[Succeeds]", 
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "b"}], ",", 
             RowBox[{"t", "*", 
              RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], "-", 
           RowBox[{"\[Gamma]", "*", 
            RowBox[{"lc", "[", "c0", "]"}]}]}]}], "\[Or]", 
         RowBox[{
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "b"}], ",", 
             RowBox[{"t", "*", 
              RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], "-", 
           RowBox[{"\[Gamma]", "*", 
            RowBox[{"lc", "[", "c0", "]"}]}]}], "\[Succeeds]", 
          RowBox[{"C", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"(", 
              RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "b"}], ",", 
            RowBox[{"t", "*", 
             RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}]}]}
      }]}]}], TraditionalForm]]],
 "\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Gamma]", ":=", 
     RowBox[{"n", "*", 
      RowBox[{"lc", "[", "m", "]"}]}]}], ",", 
    RowBox[{"t", ":=", 
     RowBox[{"s", "*", 
      RowBox[{"lp", "[", "m", "]"}]}]}]}], TraditionalForm]]],
 " witness our existentially quantified goal (G\[CloseCurlyQuote]), hence we \
prove\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"n", "*", 
      RowBox[{"lc", "[", "m", "]"}]}], ",", 
     RowBox[{"lc", "[", "c0", "]"}]}], "]"}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "d"}], "=", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "b"}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"n", "*", 
          RowBox[{"lc", "[", "m", "]"}]}], ")"}], "\[CenterDot]", 
        RowBox[{"(", 
         RowBox[{"s", "*", 
          RowBox[{"lp", "[", "m", "]"}]}], ")"}]}], ")"}], "*", "c0"}]}]}], 
   TraditionalForm]]],
 "\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "b"}], ",", 
       RowBox[{"s", "*", 
        RowBox[{"lp", "[", "m", "]"}], "*", 
        RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], "\[Succeeds]", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "b"}], ",", 
        RowBox[{"s", "*", 
         RowBox[{"lp", "[", "m", "]"}], "*", 
         RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], "-", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "*", 
         RowBox[{"lc", "[", "m", "]"}]}], ")"}], "*", 
       RowBox[{"lc", "[", "c0", "]"}]}]}]}], "\[Or]", 
    RowBox[{
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "b"}], ",", 
        RowBox[{"s", "*", 
         RowBox[{"lp", "[", "m", "]"}], "*", 
         RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}], "-", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "*", 
         RowBox[{"lc", "[", "m", "]"}]}], ")"}], "*", 
       RowBox[{"lc", "[", "c0", "]"}]}]}], "\[Succeeds]", 
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "b"}], ",", 
       RowBox[{"s", "*", 
        RowBox[{"lp", "[", "m", "]"}], "*", 
        RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}]}], TraditionalForm]]],
 "\n\n(G1):\nThis follows from (1\[CloseCurlyQuote]), (10) and the fact that \
clearly ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", "c0", "]"}], "\[NotEqual]", "0"}], TraditionalForm]]],
 ".\n\n(G2):\nUsing (7) we can replace our goal by\n(G2\[CloseCurlyQuote]):\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", 
     RowBox[{"(", 
      RowBox[{"b", "-", 
       RowBox[{"m", "*", "c0"}]}], ")"}]}], "=", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "b"}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"n", "*", 
          RowBox[{"lc", "[", "m", "]"}]}], ")"}], "\[CenterDot]", 
        RowBox[{"(", 
         RowBox[{"s", "*", 
          RowBox[{"lp", "[", "m", "]"}]}], ")"}]}], ")"}], "*", "c0"}]}]}], 
   TraditionalForm]]],
 "\nand using ring arithmetic further by\n(G2\[CloseCurlyQuote]\
\[CloseCurlyQuote]):\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "\[CenterDot]", "s"}], ")"}], "*", "m"}], ")"}], "*", 
     "c0"}], "=", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "*", 
         RowBox[{"lc", "[", "m", "]"}]}], ")"}], "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{"s", "*", 
         RowBox[{"lp", "[", "m", "]"}]}], ")"}]}], ")"}], "*", "c0"}]}], 
   TraditionalForm]]],
 "\nThe first factor on the LHS can be rewritten into ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"n", "*", 
      RowBox[{"lc", "[", "m", "]"}]}], ")"}], "\[CenterDot]", 
    RowBox[{"(", 
     RowBox[{"s", "*", 
      RowBox[{"lp", "[", "m", "]"}]}], ")"}]}], TraditionalForm]]],
 ", hence we are done.\n\n(G3):\nWe use ring arithmetic to replace (13) by\n\
(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"n", "*", 
      RowBox[{"C", "[", 
       RowBox[{"b", ",", 
        RowBox[{
         RowBox[{"lp", "[", "m", "]"}], "*", 
         RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}], "\[Succeeds]", 
     RowBox[{
      RowBox[{"n", "*", 
       RowBox[{"C", "[", 
        RowBox[{"b", ",", 
         RowBox[{
          RowBox[{"lp", "[", "m", "]"}], "*", 
          RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}], "-", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "*", 
         RowBox[{"lc", "[", "m", "]"}]}], ")"}], "*", 
       RowBox[{"lc", "[", "c0", "]"}]}]}]}], "\[Or]", 
    RowBox[{
     RowBox[{
      RowBox[{"n", "*", 
       RowBox[{"C", "[", 
        RowBox[{"b", ",", 
         RowBox[{
          RowBox[{"lp", "[", "m", "]"}], "*", 
          RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}], "-", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "*", 
         RowBox[{"lc", "[", "m", "]"}]}], ")"}], "*", 
       RowBox[{"lc", "[", "c0", "]"}]}]}], "\[Succeeds]", 
     RowBox[{"n", "*", 
      RowBox[{"C", "[", 
       RowBox[{"b", ",", 
        RowBox[{
         RowBox[{"lp", "[", "m", "]"}], "*", 
         RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}]}]}], TraditionalForm]]],
 "\n\nWe can also use the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"c", "\[CenterDot]", "s"}], ")"}], "*", "p"}], ",", 
      RowBox[{"s", "*", "t"}]}], "]"}], "=", 
    RowBox[{"c", "*", 
     RowBox[{"C", "[", 
      RowBox[{"p", ",", "t"}], "]"}]}]}], TraditionalForm]]],
 " to simplify our goal to\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"n", "*", 
      RowBox[{"C", "[", 
       RowBox[{"b", ",", 
        RowBox[{
         RowBox[{"lp", "[", "m", "]"}], "*", 
         RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}], "\[Succeeds]", 
     RowBox[{
      RowBox[{"n", "*", 
       RowBox[{"C", "[", 
        RowBox[{"b", ",", 
         RowBox[{
          RowBox[{"lp", "[", "m", "]"}], "*", 
          RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}], "-", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "*", 
         RowBox[{"lc", "[", "m", "]"}]}], ")"}], "*", 
       RowBox[{"lc", "[", "c0", "]"}]}]}]}], "\[Or]", 
    RowBox[{
     RowBox[{
      RowBox[{"n", "*", 
       RowBox[{"C", "[", 
        RowBox[{"b", ",", 
         RowBox[{
          RowBox[{"lp", "[", "m", "]"}], "*", 
          RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}], "-", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "*", 
         RowBox[{"lc", "[", "m", "]"}]}], ")"}], "*", 
       RowBox[{"lc", "[", "c0", "]"}]}]}], "\[Succeeds]", 
     RowBox[{"n", "*", 
      RowBox[{"C", "[", 
       RowBox[{"b", ",", 
        RowBox[{
         RowBox[{"lp", "[", "m", "]"}], "*", 
         RowBox[{"lp", "[", "c0", "]"}]}]}], "]"}]}]}]}], TraditionalForm]]],
 "\nThus, our goal follows immediately from (14), and so we are done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632822267303627*^9, 3.632822412231517*^9}, {
   3.6328225683984413`*^9, 3.632822622765788*^9}, {3.632822657461925*^9, 
   3.632822662355*^9}, {3.632822722069969*^9, 3.632822852999618*^9}, {
   3.6328230348637323`*^9, 3.632823077039418*^9}, {3.6328231218283157`*^9, 
   3.632823196635586*^9}, {3.632823232341873*^9, 3.6328233295293427`*^9}, {
   3.632823389254221*^9, 3.632823413778722*^9}, {3.6328234496580477`*^9, 
   3.632823519650049*^9}, {3.632823561496504*^9, 3.632823571609249*^9}, {
   3.632823688621249*^9, 3.632823997206499*^9}, {3.632824029018794*^9, 
   3.632824070588931*^9}, {3.632824110875202*^9, 3.632824283199089*^9}, {
   3.632824391377087*^9, 3.632824413727496*^9}, {3.632824475346079*^9, 
   3.6328245531754837`*^9}, {3.632824655356758*^9, 3.632824689512014*^9}, 
   3.632825238398705*^9, {3.633067960568387*^9, 3.633068016813727*^9}, {
   3.633068078395425*^9, 3.633068252498345*^9}, {3.6330683649732313`*^9, 
   3.6330684271005287`*^9}, {3.63306860477962*^9, 3.63306862496003*^9}, {
   3.633068664522332*^9, 3.633068692101932*^9}, {3.633068849303844*^9, 
   3.633068850039483*^9}, {3.6330689085648203`*^9, 3.633068981352868*^9}, {
   3.633069039629723*^9, 3.633069103764319*^9}, {3.6330691736509*^9, 
   3.633069318644187*^9}, {3.6330693561396008`*^9, 3.633069425669984*^9}, {
   3.633069456243556*^9, 3.633069500602552*^9}, {3.633069547861353*^9, 
   3.633069604132316*^9}, {3.633069803146347*^9, 3.633069803856558*^9}, {
   3.633069977779565*^9, 3.633070000421348*^9}, {3.633070041334581*^9, 
   3.633070111658716*^9}, {3.6330702117968407`*^9, 3.633070214667544*^9}, {
   3.633070332551202*^9, 3.633070527012746*^9}, {3.633070772083497*^9, 
   3.633070813890896*^9}, {3.633070872440551*^9, 3.6330709652361183`*^9}, {
   3.6330710187508593`*^9, 3.633071236548401*^9}, {3.633071299833784*^9, 
   3.6330713436001997`*^9}, {3.633071397055291*^9, 3.6330713972349157`*^9}, {
   3.633071531002351*^9, 3.6330716427400427`*^9}, {3.633071722443362*^9, 
   3.633071876073546*^9}, 3.6364403336733007`*^9, {3.639215722834465*^9, 
   3.639215723167845*^9}, 3.643446756290415*^9, 3.653207953440054*^9, 
   3.653213502724018*^9, 3.666173797598174*^9, 3.6664404061627607`*^9},
 CellID->1529086012],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632825230117806*^9, 3.632825238398842*^9}, {
   3.633071880726927*^9, 3.633071884282588*^9}, 3.63644033367346*^9, 
   3.643446756290539*^9, 3.653207953440054*^9, 3.653213502724018*^9, 
   3.666173797598174*^9, 3.6664404061629*^9},
 TextAlignment->Right,
 CellID->616246519]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (R7.\[ScriptCapitalP].1)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632804380443693*^9, 3.632804387710149*^9}, 
   3.6328083042855453`*^9, 3.636440313251635*^9, 3.643446762331088*^9, 
   3.6532079432658362`*^9, 3.6532134936136017`*^9, 3.6661737908277626`*^9, 
   3.666440398473466*^9},
 CellID->331578966],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6328054166769333`*^9, 3.632805418196444*^9}, 
   3.632808304285666*^9, 3.636440313251793*^9, 3.6434467623312063`*^9, 
   3.6532079432658362`*^9, 3.6532134936136017`*^9, 3.6661737908277626`*^9, 
   3.666440398473567*^9},
 CellID->200364281],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"a", "\[NotEqual]", "0"}], "\[And]", 
      RowBox[{"s", "\[VerticalBar]", "u"}], "\[And]", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "u"}], "]"}], "\[NotEqual]", "0"}]}], ")"}], 
    "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"s", "=", 
        RowBox[{
         RowBox[{
          RowBox[{"lp", "[", "a", "]"}], "\[And]", "u"}], "=", 
         RowBox[{"lp", "[", "a", "]"}]}]}], ")"}], "\[Or]", 
      RowBox[{
       RowBox[{"lp", "[", "a", "]"}], "\[Succeeds]", "s"}]}], ")"}]}], 
   TraditionalForm]]],
 " (Selectors)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632805426002864*^9, 3.632805471923568*^9}, {
   3.632805572021785*^9, 3.632805576344892*^9}, 3.6328083042857857`*^9, {
   3.633234487582643*^9, 3.63323449036207*^9}, 3.636440313251935*^9, {
   3.6392142239375*^9, 3.639214224969706*^9}, 3.643446762331321*^9, 
   3.6532079432658362`*^9, 3.6532134936136017`*^9, 3.6661737908277626`*^9, 
   3.66644039847365*^9},
 CellID->1473345246],

Cell["Special inference rules:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6328082779986353`*^9, 3.632808304285885*^9}, 
   3.6364403132520638`*^9, 3.6434467623314323`*^9, 3.6532079432658362`*^9, 
   3.6532134936136017`*^9, 3.6661737908277626`*^9, 3.666440398473734*^9},
 CellID->1709547035],

Cell[TextData[{
 "Transitivity of ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], TraditionalForm]]]
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6328082858880997`*^9, 3.632808304285985*^9}, 
   3.636440313252192*^9, 3.643446762331526*^9, 3.6532079432658362`*^9, 
   3.6532134936136017`*^9, 3.6661737908277626`*^9, 3.666440398473805*^9},
 CellID->1610958980],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "a", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"u", ",", "t"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", "\[VerticalBar]", "u"}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[Or]", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", "u"}], "]"}], "\[Succeeds]", "c"}]}]}], 
   TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c", "\[CenterDot]", "s"}], "=", 
    RowBox[{"a", "\[Or]", 
     RowBox[{"a", "\[Succeeds]", 
      RowBox[{"c", "\[CenterDot]", "s"}]}]}]}], TraditionalForm]]],
 "\n\nFirst of all, we distinguish two cases, based on whether or not ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "u"}], "]"}], "=", "0"}], TraditionalForm]]],
 ".\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "u"}], "]"}], "=", "0"}], TraditionalForm]]],
 ".\nBecause of (R4) in \[ScriptCapitalR] we know, from (2), ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "=", "0"}]}], TraditionalForm]]],
 ". This means that also ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c", "\[CenterDot]", "s"}], "=", "0"}], TraditionalForm]]],
 ", and our goal reduces to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "=", 
    RowBox[{"a", "\[Or]", 
     RowBox[{"a", "\[Succeeds]", "0"}]}]}], TraditionalForm]]],
 ". This holds because of (R4) in \[ScriptCapitalP].\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "u"}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 ".\n\nBefore actually proving (G), we show first\n(C)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "s"}], "=", 
    RowBox[{"a", "\[Or]", 
     RowBox[{"a", "\[Succeeds]", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "s"}]}]}]}], 
   TraditionalForm]]],
 "\n\nFrom (1), by a lemma, we can immediately infer\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"s", "=", 
      RowBox[{
       RowBox[{
        RowBox[{"lp", "[", "a", "]"}], "\[And]", "u"}], "=", 
       RowBox[{"lp", "[", "a", "]"}]}]}], ")"}], "\[Or]", 
    RowBox[{
     RowBox[{"lp", "[", "a", "]"}], "\[Succeeds]", "s"}]}], TraditionalForm]]],
 "\nbased upon which we distinguish two cases.\n\nCase C.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", "=", 
    RowBox[{
     RowBox[{
      RowBox[{"lp", "[", "a", "]"}], "\[And]", "u"}], "=", 
     RowBox[{"lp", "[", "a", "]"}]}]}], TraditionalForm]]],
 ".\nIn that case our goal (C) reduces to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lm", "[", "a", "]"}], "=", 
    RowBox[{"a", "\[Or]", 
     RowBox[{"a", "\[Succeeds]", 
      RowBox[{"lm", "[", "a", "]"}]}]}]}], TraditionalForm]]],
 ", which is precisely what (406) states.\nCase C.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "a", "]"}], "\[Succeeds]", "s"}], TraditionalForm]]],
 ".\nWe show the second disjunct of (C), i.e. we show\n(C\[CloseCurlyQuote])\t\
",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "s"}]}], 
   TraditionalForm]]],
 "\nFormula (377\[CloseCurlyQuote]\[CloseCurlyQuote]), together with our case \
assumption, finishes the proof.\n\nSummarizing, we can add\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "s"}], "=", 
    RowBox[{"a", "\[Or]", 
     RowBox[{"a", "\[Succeeds]", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "s"}]}]}]}], 
   TraditionalForm]]],
 "\nto our knowledge base.\nBased on (2) we distinguish two cases.\n\nCase \
II.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "=", 
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "u"}], "]"}]}], TraditionalForm]]],
 ".\nOur goal (G) reduces to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "s"}], "=", 
    RowBox[{"a", "\[Or]", 
     RowBox[{"a", "\[Succeeds]", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "s"}]}]}]}], 
   TraditionalForm]]],
 ", and this is precisely what we have already shown and what is stated in \
(4).\n\nCase II.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "u"}], "]"}], "\[Succeeds]", "c"}], 
   TraditionalForm]]],
 ".\nFrom a lemma on monomials we know that in that case\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "s"}], "\[Succeeds]", 
    RowBox[{"c", "\[CenterDot]", "s"}]}], TraditionalForm]]],
 "\nWe show the second disjunct of our goal (G), i.e. we show\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"c", "\[CenterDot]", "s"}]}], TraditionalForm]]],
 "\nBased on (4) we distinguish two cases.\nCase II.II.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "s"}], "=", "a"}], 
   TraditionalForm]]],
 ".\nOur assumption (5) can be replaced by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"c", "\[CenterDot]", "s"}]}], TraditionalForm]]],
 ", and this is precisely what we have to show.\nCase II.II.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[CenterDot]", "s"}]}], 
   TraditionalForm]]],
 ".\nFrom (5) and our case assumption, using transitivity of ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], TraditionalForm]]],
 ", we can infer ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"c", "\[CenterDot]", "s"}]}], TraditionalForm]]],
 ", and this is again precisely what we have to show."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632804397342719*^9, 3.63280441688376*^9}, {
   3.6328045271697474`*^9, 3.632804607419181*^9}, {3.63280467508144*^9, 
   3.63280469270647*^9}, {3.632804767689939*^9, 3.632804868368722*^9}, {
   3.632804919169139*^9, 3.6328049665675077`*^9}, {3.6328049967826757`*^9, 
   3.6328051234107857`*^9}, {3.632805191167712*^9, 3.632805272459834*^9}, {
   3.632805486334939*^9, 3.632805511125222*^9}, {3.632805595923427*^9, 
   3.632805605346027*^9}, {3.6328056396488733`*^9, 3.632805748228848*^9}, {
   3.632805806135509*^9, 3.6328059504335527`*^9}, {3.6328061031389217`*^9, 
   3.63280662714263*^9}, {3.63280666234127*^9, 3.632806671827558*^9}, {
   3.632807633793066*^9, 3.632807654131379*^9}, {3.6328079428121853`*^9, 
   3.632807972609619*^9}, {3.6328080127365007`*^9, 3.632808255720409*^9}, 
   3.632808304286085*^9, 3.6364403132523193`*^9, {3.639214232197958*^9, 
   3.639214238233199*^9}, {3.639219378665205*^9, 3.639219393062912*^9}, 
   3.64344676233162*^9, 3.6532079432658362`*^9, 3.6532134936136017`*^9, 
   3.6661737908277626`*^9, 3.666440398473885*^9},
 CellID->817801926],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632808261435858*^9, 3.632808265010326*^9}, 
   3.632808304286222*^9, 3.6364403132524767`*^9, 3.643446762331737*^9, 
   3.6532079432658362`*^9, 3.653213493629202*^9, 3.6661737908277626`*^9, 
   3.666440398473988*^9},
 TextAlignment->Right,
 CellID->1726226374]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (R7.\[ScriptCapitalP].4)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632811720596752*^9, 3.63281172743123*^9}, 
   3.632819218341096*^9, 3.636440327248989*^9, 3.643446767633708*^9, 
   3.6532079343250203`*^9, 3.653213484050785*^9, 3.6661737840105505`*^9, 
   3.666440391024404*^9},
 CellID->225861683],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "a", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"u", ",", "s"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "c0", "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"s", "\[VerticalBar]", "u"}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "d0", "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"c", "\[Succeeds]", 
       RowBox[{"c", "+", "d0"}]}], "\[Implies]", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"a", ",", "u"}], "]"}], "\[Succeeds]", 
       RowBox[{
        RowBox[{"C", "[", 
         RowBox[{"a", ",", "u"}], "]"}], "+", 
        RowBox[{"n", "*", "d0"}]}]}]}], ")"}]}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c", "\[CenterDot]", "s"}], "\[Succeeds]", 
    RowBox[{
     RowBox[{"c", "\[CenterDot]", "s"}], "+", "c0"}]}], TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"a", "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"n", "\[CenterDot]", 
        RowBox[{"(", 
         RowBox[{"u", "/", "s"}], ")"}]}], ")"}], "*", "c0"}]}]}], 
   TraditionalForm]]],
 "\n\nFrom (3), by a lemma on \[Succeeds] for monomials, we can infer\n(4)\t\
",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{"c0", ",", "s"}], "]"}], "=", "0"}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[Succeeds]", 
    RowBox[{"c", "+", 
     RowBox[{"C", "[", 
      RowBox[{"c0", ",", "s"}], "]"}]}]}], TraditionalForm]]],
 "\n\nNow we can instantiate (2) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"d0", ":=", 
    RowBox[{"C", "[", 
     RowBox[{"c0", ",", "s"}], "]"}]}], TraditionalForm]]],
 " and, since the premise of the implication is fulfilled due to (5), add\n\
(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "u"}], "]"}], "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "+", 
     RowBox[{"n", "*", 
      RowBox[{"C", "[", 
       RowBox[{"c0", ",", "s"}], "]"}]}]}]}], TraditionalForm]]],
 "\nto our knowledge base.\n\nNext, we infer from (1) the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "r", "]"}]}], TraditionalForm]]],
 " with\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"r", "*", "s"}], "=", "u"}], TraditionalForm]]],
 "\nmeaning that we can replace our goal (G) by\n(G\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"a", "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"n", "\[CenterDot]", "r"}], ")"}], "*", "c0"}]}]}], 
   TraditionalForm]]],
 "\n\nIn order to prove (G\[CloseCurlyQuote]), as usual, we have to find ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 " s.t. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{"a", ",", "t"}], "]"}], "=", 
    RowBox[{"H", "[", 
     RowBox[{
      RowBox[{"a", "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"n", "\[CenterDot]", "r"}], ")"}], "*", "c0"}]}], ",", "t"}],
      "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", "t"}], "]"}], "\[Succeeds]", 
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"a", "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"n", "\[CenterDot]", "r"}], ")"}], "*", "c0"}]}], ",", "t"}],
      "]"}]}], TraditionalForm]]],
 ". We claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", ":=", 
    RowBox[{"u", "=", 
     RowBox[{"r", "*", "s"}]}]}], TraditionalForm]]],
 " is a suitable witness, hence we show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{"a", ",", 
      RowBox[{"r", "*", "s"}]}], "]"}], "=", 
    RowBox[{"H", "[", 
     RowBox[{
      RowBox[{"a", "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"n", "\[CenterDot]", "r"}], ")"}], "*", "c0"}]}], ",", 
      RowBox[{"r", "*", "s"}]}], "]"}]}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", 
      RowBox[{"r", "*", "s"}]}], "]"}], "\[Succeeds]", 
    RowBox[{"C", "[", 
     RowBox[{
      RowBox[{"a", "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"n", "\[CenterDot]", "r"}], ")"}], "*", "c0"}]}], ",", 
      RowBox[{"r", "*", "s"}]}], "]"}]}], TraditionalForm]]],
 "\n\n(G1):\nUsing linearity of ",
 Cell[BoxData[
  FormBox["H", TraditionalForm]]],
 " we can rewrite our goal into ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{"a", ",", 
      RowBox[{"r", "*", "s"}]}], "]"}], "=", 
    RowBox[{
     RowBox[{"H", "[", 
      RowBox[{"a", ",", 
       RowBox[{"r", "*", "s"}]}], "]"}], "+", 
     RowBox[{"H", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"c", "\[CenterDot]", "r"}], ")"}], "*", "c0"}], ",", 
       RowBox[{"r", "*", "s"}]}], "]"}]}]}], TraditionalForm]]],
 ", and further into\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"c", "\[CenterDot]", "r"}], ")"}], "*", "c0"}], ",", 
      RowBox[{"r", "*", "s"}]}], "]"}], "=", "0"}], TraditionalForm]]],
 "\nBy a lemma, a sufficient condition for (G3) to hold is\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"H", "[", 
     RowBox[{"c0", ",", "s"}], "]"}], "=", "0"}], TraditionalForm]]],
 "\nbut this is exactly what (4) states.\n\n(G2):\nUsing linearity of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " we can rewrite our goal into\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", 
      RowBox[{"r", "*", "s"}]}], "]"}], "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{"r", "*", "s"}]}], "]"}], "+", 
     RowBox[{"C", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"n", "\[CenterDot]", "r"}], ")"}], "*", "c0"}], ",", 
       RowBox[{"r", "*", "s"}]}], "]"}]}]}], TraditionalForm]]],
 "\nThe last summand on the RHS is equal to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "*", 
    RowBox[{"C", "[", 
     RowBox[{"c0", ",", "s"}], "]"}]}], TraditionalForm]]],
 " by a lemma, so our goal now reads as\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"a", ",", 
      RowBox[{"r", "*", "s"}]}], "]"}], "\[Succeeds]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", 
       RowBox[{"r", "*", "s"}]}], "]"}], "+", 
     RowBox[{"n", "*", 
      RowBox[{"C", "[", 
       RowBox[{"c0", ",", "s"}], "]"}]}]}]}], TraditionalForm]]],
 "\nThis, however, is precisely what (6) states (note that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"u", "=", 
    RowBox[{"r", "*", "s"}]}], TraditionalForm]]],
 ")."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632811734463381*^9, 3.632811880555593*^9}, {
   3.632811963668833*^9, 3.6328119729752197`*^9}, {3.632812015768095*^9, 
   3.632812155881085*^9}, {3.6328122036532993`*^9, 3.632812308111912*^9}, {
   3.632812538232154*^9, 3.632812790326351*^9}, {3.632812836573943*^9, 
   3.6328128382213*^9}, {3.632812883170911*^9, 3.632812934775443*^9}, {
   3.632812965668064*^9, 3.632812965668187*^9}, {3.632816457822365*^9, 
   3.6328164894223833`*^9}, {3.6328166164217777`*^9, 
   3.6328166553582487`*^9}, {3.632816702455475*^9, 3.632816754625751*^9}, {
   3.632817656592754*^9, 3.632817915209392*^9}, {3.6328187515676823`*^9, 
   3.632818779557292*^9}, {3.632819005570458*^9, 3.632819218341223*^9}, {
   3.632822309254634*^9, 3.632822310575026*^9}, 3.636440327249146*^9, {
   3.639222093235458*^9, 3.639222094260496*^9}, 3.6434467676338253`*^9, 
   3.6532079343250203`*^9, 3.653213484050785*^9, 3.6661737840105505`*^9, 
   3.666440391024507*^9},
 CellID->994840203],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632819206694121*^9, 3.632819218341366*^9}, 
   3.636440327249317*^9, 3.643446767633943*^9, 3.6532079343250203`*^9, 
   3.653213484050785*^9, 3.6661737840105505`*^9, 3.6664403910246077`*^9},
 TextAlignment->Right,
 CellID->459605869]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->357963951],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R7", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632738310536779*^9, 3.632738316908396*^9}},
 CellID->1656172280],

Cell[BoxData[
 RowBox[{"R7", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(370)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:255426543", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185609946504*^9, 3.610185628205185*^9}, 
   3.631596864954699*^9},
 CellTags->{
  "ID:255426543", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "370"},
 CellID->255426543],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->940321280]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->385752651],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(370)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(370)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R7", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R7", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:255426543"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p255426543-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p255426543-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:255426543", "Proof|ID:255426543-1"},
 CellID->537752064],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p255426543-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->1382560669]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.63274125142106*^9, 3.6327412528325443`*^9}, 
   3.633076123266273*^9, 3.636440343575472*^9, 3.643446777372295*^9, 
   3.6532079231992006`*^9, 3.6532134722259645`*^9, 3.6661737766941376`*^9, 
   3.6664403849937143`*^9},
 CellID->1814678652],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632741454487055*^9, 3.6327414559778423`*^9}, 
   3.633076123266409*^9, 3.636440343575602*^9, 3.643446777372414*^9, 
   3.6532079231992006`*^9, 3.6532134722259645`*^9, 3.6661737766941376`*^9, 
   3.666440384993854*^9},
 CellID->109051318],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isNTCR", "[", 
     RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
    "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"p", "\[NotEqual]", "0"}], "\[And]", 
      RowBox[{"q", "\[NotEqual]", "0"}], "\[And]", 
      RowBox[{"a", "\[NotEqual]", "0"}]}], ")"}]}], TraditionalForm]]],
 " (336)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6327414636100397`*^9, 3.63274149034109*^9}, 
   3.633076123266926*^9, {3.633235166783581*^9, 3.633235168697419*^9}, 
   3.636440343575716*^9, {3.6392222629491243`*^9, 3.639222271051056*^9}, 
   3.6434467773724947`*^9, 3.6532079231992006`*^9, 3.6532134722259645`*^9, 
   3.6661737766941376`*^9, 3.666440384993966*^9},
 CellID->939173627],

Cell[TextData[{
 "In order to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"R7", "[", "\[ScriptCapitalP]", "]"}], TraditionalForm]]],
 ", by the definition of R7, we choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q", ",", "a"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand show the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", ",", 
    RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"a0", ",", "m"}], "]"}]}], TraditionalForm]]],
 " with certain properties.\nFirst of all we use a lemma and infer from (1) \
",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 ".\n\nFrom (1), together with (CR1.\[Implies].1) and (CR1.\[Implies].2) we \
can infer\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{"(", 
      RowBox[{"i", "\[NotEqual]", "j"}], ")"}]}]}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", 
         RowBox[{"s", ",", "t"}], "]"}]}]], 
      RowBox[{"s", "\[NotEqual]", "t"}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {
        RowBox[{
         RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "s"}]},
       {
        RowBox[{"areCommonReducibles", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "t"}], "]"}], ",", 
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "s"}], "]"}], ",", "p", ",", "q", ",", "i", ",", 
          "j"}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "u"}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "q", "]"}], "\[VerticalBar]", "u"}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "u", "]"}]}], TraditionalForm]]],
 ".\nFrom (4) and (5) we can immediately infer\n(4\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "\[VerticalBar]", "u"}], TraditionalForm]]],
 "\nwhere lcm abbreviates ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcm", "[", 
    RowBox[{
     RowBox[{"lp", "[", "p", "]"}], ",", 
     RowBox[{"lp", "[", "q", "]"}]}], "]"}], TraditionalForm]]],
 ".\n\nFrom (6) we can infer the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", ",", 
    RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c0", ",", "n"}], "]"}]}], TraditionalForm]]],
 " with\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c0", "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[Or]", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", "u"}], "]"}], "\[Succeeds]", "c0"}]}]}], 
   TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "c0", ",", "k", ",", "l"}], "]"}], 
   TraditionalForm]]],
 "\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c", ",", "m0"}], "]"}]}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"c", "\[NotEqual]", "0"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"m0", ",", "c"}], "]"}]}], ")"}]}], "\[Implies]", 
    RowBox[{"mult", "[", 
     RowBox[{
      RowBox[{"n", "*", "m0"}], ",", "c"}], "]"}]}], TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", "c", "]"}]}]], 
     RowBox[{"c0", "\[Succeeds]", 
      RowBox[{"c0", "+", "c"}]}]}], "\[Implies]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[Succeeds]", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", "u"}], "]"}], "+", 
      RowBox[{"n", "*", "c"}]}]}]}], TraditionalForm]]],
 "\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"b", ",", "c", ",", "d"}], "]"}]}]], "b"}], 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{"d", "\[Implies]", 
     RowBox[{
      RowBox[{"n", "*", "b"}], 
      SubscriptBox["\[LeftRightArrow]", "c"], 
      RowBox[{"n", "*", "d"}]}]}]}], TraditionalForm]]],
 "\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
        RowBox[{"[", 
         RowBox[{"m1", ",", "m2"}], "]"}]}]], 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"isReducible", "[", 
          RowBox[{"c0", ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "m1"}], "]"}], "\[And]", 
         RowBox[{"isReducible", "[", 
          RowBox[{"c0", ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "m2"}], "]"}], "\[And]", 
         RowBox[{"mult", "[", 
          RowBox[{"m1", ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "k"}], "]"}], "\[And]", 
         RowBox[{"mult", "[", 
          RowBox[{"m2", ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "l"}], "]"}]}], "\[Implies]", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"mult", "[", 
             RowBox[{
              RowBox[{"n", "*", "m1"}], ",", 
              RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
            RowBox[{"mult", "[", 
             RowBox[{
              RowBox[{"n", "*", "m2"}], ",", 
              RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], ")"}], 
          "\[Or]", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"mult", "[", 
             RowBox[{
              RowBox[{"n", "*", "m1"}], ",", 
              RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}], "\[And]", 
            RowBox[{"mult", "[", 
             RowBox[{
              RowBox[{"n", "*", "m2"}], ",", 
              RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}]}], ")"}]}], 
         ")"}]}], ")"}]}]}]}], TraditionalForm]]],
 "\n\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", ":=", "k"}], ",", 
    RowBox[{"l", ":=", "l"}], ",", 
    RowBox[{"a0", ":=", 
     RowBox[{"c0", "\[CenterDot]", "lcm"}]}], ",", 
    RowBox[{"m", ":=", 
     RowBox[{"n", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{"u", "/", "lcm"}], ")"}]}]}]}], TraditionalForm]]],
 " witnesses our existential goal. Therefore, we have to show several \
subgoals.\n\n(G1): ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c0", "\[CenterDot]", "lcm"}], "=", 
    RowBox[{"a", "\[Or]", 
     RowBox[{"a", "\[Succeeds]", 
      RowBox[{"c0", "\[CenterDot]", "lcm"}]}]}]}], TraditionalForm]]],
 ".\nWe can use (R7.\[ScriptCapitalP].1), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", ":=", "a"}], ",", 
    RowBox[{"u", ":=", "u"}], ",", 
    RowBox[{"s", ":=", "lcm"}], ",", 
    RowBox[{"c", ":=", "c0"}]}], TraditionalForm]]],
 ".\n\n(G2): ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{"c0", "\[CenterDot]", "lcm"}], ",", "k", ",", "l"}], "]"}], 
   TraditionalForm]]],
 ".\nWe can use (CR3.\[DoubleLeftArrow]), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", ":=", "p"}], ",", 
    RowBox[{"q", ":=", "q"}], ",", 
    RowBox[{"i", ":=", "k"}], ",", 
    RowBox[{"j", ":=", "l"}], ",", 
    RowBox[{"c", ":=", "c0"}]}], TraditionalForm]]],
 ".\n\n(G3): ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", 
        RowBox[{"c", ",", "m0"}], "]"}]}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"c", "\[NotEqual]", "0"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"m0", ",", "c"}], "]"}]}], ")"}]}], "\[Implies]", 
    RowBox[{"mult", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "\[CenterDot]", 
         RowBox[{"(", 
          RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "m0"}], ",", "c"}], 
     "]"}]}], TraditionalForm]]],
 ".\nWe can use (R7.\[ScriptCapitalP].3), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"s", ":=", 
     RowBox[{"u", "/", "lcm"}]}], ",", 
    RowBox[{"n", ":=", "n"}]}], TraditionalForm]]],
 ".\n\n(G4): ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", "c", "]"}]}]], 
     RowBox[{
      RowBox[{"c0", "\[CenterDot]", "lcm"}], "\[Succeeds]", 
      RowBox[{
       RowBox[{"c0", "\[CenterDot]", "lcm"}], "+", "c"}]}]}], "\[Implies]", 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"a", "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "\[CenterDot]", 
         RowBox[{"(", 
          RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "c"}]}]}]}], 
   TraditionalForm]]],
 ".\nWe can use (R7.\[ScriptCapitalP].4), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", ":=", "a"}], ",", 
    RowBox[{"u", ":=", "u"}], ",", 
    RowBox[{"s", ":=", "lcm"}], ",", 
    RowBox[{"c", ":=", "c0"}], ",", 
    RowBox[{"n", ":=", "n"}]}], TraditionalForm]]],
 ".\n\n(G5): ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", 
        RowBox[{"b", ",", "c", ",", "d"}], "]"}]}]], "b"}], 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{"d", "\[Implies]", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "\[CenterDot]", 
         RowBox[{"(", 
          RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "b"}], 
      SubscriptBox["\[LeftRightArrow]", "c"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "\[CenterDot]", 
         RowBox[{"(", 
          RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "d"}]}]}]}], 
   TraditionalForm]]],
 ".\nWe can use (R7.\[ScriptCapitalP].5), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"s", ":=", 
     RowBox[{"u", "/", "lcm"}]}], ",", 
    RowBox[{"n", ":=", "n"}]}], TraditionalForm]]],
 ".\n\n(G6): ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Implies]", 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
        RowBox[{"[", 
         RowBox[{"m1", ",", "m2"}], "]"}]}]], 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"c0", "\[CenterDot]", "lcm"}], ",", "q", ",", "m1"}], 
          "]"}], "\[And]", 
         RowBox[{"isReducible", "[", 
          RowBox[{
           RowBox[{"c0", "\[CenterDot]", "lcm"}], ",", "q", ",", "m2"}], 
          "]"}], "\[And]", 
         RowBox[{"mult", "[", 
          RowBox[{"m1", ",", "q", ",", "k"}], "]"}], "\[And]", 
         RowBox[{"mult", "[", 
          RowBox[{"m2", ",", "q", ",", "l"}], "]"}]}], "\[Implies]", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"mult", "[", 
             RowBox[{
              RowBox[{
               RowBox[{"(", 
                RowBox[{"n", "\[CenterDot]", 
                 RowBox[{"(", 
                  RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "m1"}], ",",
               "q", ",", "i"}], "]"}], "\[And]", 
            RowBox[{"mult", "[", 
             RowBox[{
              RowBox[{
               RowBox[{"(", 
                RowBox[{"n", "\[CenterDot]", 
                 RowBox[{"(", 
                  RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "m2"}], ",",
               "q", ",", "j"}], "]"}]}], ")"}], "\[Or]", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"mult", "[", 
             RowBox[{
              RowBox[{
               RowBox[{"(", 
                RowBox[{"n", "\[CenterDot]", 
                 RowBox[{"(", 
                  RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "m1"}], ",",
               "q", ",", "j"}], "]"}], "\[And]", 
            RowBox[{"mult", "[", 
             RowBox[{
              RowBox[{
               RowBox[{"(", 
                RowBox[{"n", "\[CenterDot]", 
                 RowBox[{"(", 
                  RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "m2"}], ",",
               "q", ",", "i"}], "]"}]}], ")"}]}], ")"}]}], ")"}]}]}]}], 
   TraditionalForm]]],
 ".\nAssume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", "q"}], TraditionalForm]]],
 ", choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"c0", "\[CenterDot]", "lcm"}], ",", "q", ",", "m1"}], "]"}], 
   TraditionalForm]]],
 "\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"c0", "\[CenterDot]", "lcm"}], ",", "q", ",", "m2"}], "]"}], 
   TraditionalForm]]],
 "\n(15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m1", ",", "q", ",", "k"}], "]"}], TraditionalForm]]],
 "\n(16)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m2", ",", "q", ",", "l"}], "]"}], TraditionalForm]]],
 "\nand show\n(G6\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{"n", "\[CenterDot]", 
            RowBox[{"(", 
             RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "m1"}], ",", "q",
          ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{"n", "\[CenterDot]", 
            RowBox[{"(", 
             RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "m2"}], ",", "q",
          ",", "j"}], "]"}]}], ")"}], "\[Or]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{"n", "\[CenterDot]", 
            RowBox[{"(", 
             RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "m1"}], ",", "q",
          ",", "j"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{"n", "\[CenterDot]", 
            RowBox[{"(", 
             RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "m2"}], ",", "q",
          ",", "i"}], "]"}]}], ")"}]}], ")"}], TraditionalForm]]],
 "\nFrom (15) and (16) we immediately get\n(17)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "m1", "]"}], TraditionalForm]]],
 "\n(18)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"m1", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(19)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMonomial", "[", "m2", "]"}], TraditionalForm]]],
 "\n(20)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"m2", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\n(21)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"lc", "[", "m1", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "k"}], "]"}], TraditionalForm]]],
 "\n(22)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"lc", "[", "m2", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "l"}], "]"}], TraditionalForm]]],
 "\n\nFrom (13) and (14), by (RED monomial), we can infer (among others)\n\
(23)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"c0", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", 
     RowBox[{"lc", "[", "m1", "]"}]}], "]"}], TraditionalForm]]],
 "\n(24)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"c0", ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", 
     RowBox[{"lc", "[", "m2", "]"}]}], "]"}], TraditionalForm]]],
 "\n\nNow we can instantiate the RHS of (12) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m1", ":=", 
     RowBox[{"lc", "[", "m1", "]"}]}], ",", 
    RowBox[{"m2", ":=", 
     RowBox[{"lc", "[", "m2", "]"}]}]}], TraditionalForm]]],
 " and infer\n(25)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"isReducible", "[", 
       RowBox[{"c0", ",", 
        RowBox[{"lc", "[", "q", "]"}], ",", 
        RowBox[{"lc", "[", "m1", "]"}]}], "]"}], "\[And]", 
      RowBox[{"isReducible", "[", 
       RowBox[{"c0", ",", 
        RowBox[{"lc", "[", "q", "]"}], ",", 
        RowBox[{"lc", "[", "m2", "]"}]}], "]"}], "\[And]", 
      RowBox[{"mult", "[", 
       RowBox[{
        RowBox[{"lc", "[", "m1", "]"}], ",", 
        RowBox[{"lc", "[", "q", "]"}], ",", "k"}], "]"}], "\[And]", 
      RowBox[{"mult", "[", 
       RowBox[{
        RowBox[{"lc", "[", "m2", "]"}], ",", 
        RowBox[{"lc", "[", "q", "]"}], ",", "l"}], "]"}]}], "\[Implies]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"mult", "[", 
          RowBox[{
           RowBox[{"n", "*", 
            RowBox[{"lc", "[", "m1", "]"}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
         RowBox[{"mult", "[", 
          RowBox[{
           RowBox[{"n", "*", 
            RowBox[{"lc", "[", "m2", "]"}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], ")"}], "\[Or]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"mult", "[", 
          RowBox[{
           RowBox[{"n", "*", 
            RowBox[{"lc", "[", "m1", "]"}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}], "\[And]", 
         RowBox[{"mult", "[", 
          RowBox[{
           RowBox[{"n", "*", 
            RowBox[{"lc", "[", "m2", "]"}]}], ",", 
           RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}]}], ")"}]}], 
      ")"}]}], ")"}], TraditionalForm]]],
 "\n\nThe premise of the implication in (25) is fulfilled due to (23), (24), \
(21) and (22), so we can add\n(26)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{"n", "*", 
          RowBox[{"lc", "[", "m1", "]"}]}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{"n", "*", 
          RowBox[{"lc", "[", "m2", "]"}]}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}]}], ")"}], "\[Or]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{"n", "*", 
          RowBox[{"lc", "[", "m1", "]"}]}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         RowBox[{"n", "*", 
          RowBox[{"lc", "[", "m2", "]"}]}], ",", 
         RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}]}], ")"}]}], ")"}], 
   TraditionalForm]]],
 "\nto our knowledge base, based upon which we distinguish two cases; \
However, since both of them are completely symmetric, we only consider the \
first one here. Therefore, we add\n(27)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"n", "*", 
      RowBox[{"lc", "[", "m1", "]"}]}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], TraditionalForm]]],
 "\n(28)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"n", "*", 
      RowBox[{"lc", "[", "m2", "]"}]}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "j"}], "]"}], TraditionalForm]]],
 "\nto our knowledge base.\n\nWe show the first disjunct in (G6\
\[CloseCurlyQuote]), i.e. we show both\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"n", "\[CenterDot]", 
        RowBox[{"(", 
         RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "m1"}], ",", "q", 
     ",", "i"}], "]"}], TraditionalForm]]],
 "\n(G8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"n", "\[CenterDot]", 
        RowBox[{"(", 
         RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "m2"}], ",", "q", 
     ",", "j"}], "]"}], TraditionalForm]]],
 "\n\nSince again both subgoals are completely symmetric, we only consider \
the first one, i.e. (G7), here. In order to prove (G7) we first replace ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"n", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{"u", "/", "lcm"}], ")"}]}], ")"}], "*", "m1"}], 
   TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"n", "*", 
      RowBox[{"lc", "[", "m1", "]"}]}], ")"}], "\[CenterDot]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"u", "/", "lcm"}], ")"}], "*", 
      RowBox[{"lp", "[", "m1", "]"}]}], ")"}]}], TraditionalForm]]],
 " and hence reduce our goal to\n(G7\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"n", "*", 
      RowBox[{"lc", "[", "m1", "]"}]}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "i"}], "]"}], TraditionalForm]]],
 "\nThis is precisely what (27) states, hence we are done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632741258696162*^9, 3.632741366445916*^9}, 
   3.632741406802876*^9, {3.6327414924506493`*^9, 3.632741682410475*^9}, {
   3.632741944483564*^9, 3.63274198649291*^9}, {3.6327421404701967`*^9, 
   3.632742945305071*^9}, {3.632743002676351*^9, 3.63274302769627*^9}, {
   3.6327430855152082`*^9, 3.6327431094509497`*^9}, 3.632743708926586*^9, {
   3.632743762558179*^9, 3.632743783725276*^9}, {3.632743816867696*^9, 
   3.632743837029949*^9}, {3.632808314253736*^9, 3.632808385770213*^9}, {
   3.6328086373541813`*^9, 3.63280869542358*^9}, {3.632808820286887*^9, 
   3.632808866814271*^9}, {3.632809918793721*^9, 3.6328099437547903`*^9}, {
   3.632810947307803*^9, 3.632811031872114*^9}, {3.6328189648248377`*^9, 
   3.6328189815648537`*^9}, {3.6328215392243223`*^9, 
   3.6328216633172207`*^9}, {3.632822452179295*^9, 3.632822485444607*^9}, {
   3.6328253584483957`*^9, 3.632825616982862*^9}, {3.632825995837305*^9, 
   3.632826012754301*^9}, {3.632826054602469*^9, 3.6328261199159327`*^9}, {
   3.633072443433284*^9, 3.6330724926136436`*^9}, {3.6330756009683533`*^9, 
   3.633075912584158*^9}, {3.633075955450425*^9, 3.633076123267047*^9}, 
   3.6364403435758133`*^9, {3.63922228188933*^9, 3.639222289401947*^9}, {
   3.639222322005734*^9, 3.639222363891048*^9}, {3.639222397635127*^9, 
   3.639222403199768*^9}, {3.639301240362561*^9, 3.639301254792467*^9}, 
   3.643446777372603*^9, 3.6532079231992006`*^9, 3.6532134722259645`*^9, 
   3.6661737766941376`*^9, 3.666440384994083*^9},
 CellID->2037769089],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6330761037467213`*^9, 3.633076123267178*^9}, 
   3.636440343575931*^9, 3.643446777372719*^9, 3.6532079231992006`*^9, 
   3.6532134722259645`*^9, 3.6661737766941376`*^9, 3.666440384994216*^9},
 TextAlignment->Right,
 CellID->369172055]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R8 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6322163495373983`*^9, 3.632216350352025*^9}, {
  3.6393698328114862`*^9, 3.63936983702932*^9}},
 CellID->2100864518],

Cell["\<\
The proof of R8 is essentially the same as the proof of R7, only shorter. We \
can make direct use of (R7.\[ScriptCapitalP].1) \[Dash] \
(R7.\[ScriptCapitalP].5) also here.\
\>", "Text",
 CellChangeTimes->{{3.639369761489737*^9, 3.6393698263728533`*^9}},
 CellID->164300636],

Cell["", "OpenEnvironment",
 CellID->342777797],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R8", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.633076247297326*^9, 3.63307625284893*^9}},
 CellID->1898194960],

Cell[BoxData[
 RowBox[{"R8", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(371)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1342888001", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185611172009*^9, 3.610185629132915*^9}, 
   3.631596865808322*^9},
 CellTags->{
  "ID:1342888001", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "371"},
 CellID->1342888001],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->376228414]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1126499261],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(371)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(371)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R8", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R8", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1342888001"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1342888001-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1342888001-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1342888001", "Proof|ID:1342888001-1"},
 CellID->1391148747],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1342888001-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->837021914]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.63274125142106*^9, 3.6327412528325443`*^9}, 
   3.633076123266273*^9, 3.636440343575472*^9, 3.6434467857867947`*^9, 
   3.65320791120278*^9, 3.653213464503951*^9, 3.66617376680272*^9, 
   3.666440373929624*^9},
 CellID->42990982],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632741454487055*^9, 3.6327414559778423`*^9}, 
   3.633076123266409*^9, 3.636440343575602*^9, 3.643446785786951*^9, 
   3.65320791120278*^9, 3.653213464503951*^9, 3.66617376680272*^9, 
   3.6664403739297657`*^9},
 CellID->356763103],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isNTCR", "[", 
     RowBox[{"p", ",", "a"}], "]"}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"p", "\[NotEqual]", "0"}], "\[And]", 
      RowBox[{"a", "\[NotEqual]", "0"}]}], ")"}]}], TraditionalForm]]],
 " (337)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6327414636100397`*^9, 3.63274149034109*^9}, 
   3.633076123266926*^9, {3.633235166783581*^9, 3.633235168697419*^9}, 
   3.636440343575716*^9, {3.6392222629491243`*^9, 3.639222271051056*^9}, {
   3.639367513371553*^9, 3.639367520924325*^9}, 3.643446785787055*^9, 
   3.65320791120278*^9, 3.653213464503951*^9, 3.66617376680272*^9, 
   3.666440373929879*^9},
 CellID->1594516779],

Cell[TextData[{
 "In order to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"R8", "[", "\[ScriptCapitalP]", "]"}], TraditionalForm]]],
 ", by the definition of R8, we choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "a"}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{"p", ",", "a"}], "]"}], TraditionalForm]]],
 "\nand show the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"a0", ",", "m"}], "]"}]}], TraditionalForm]]],
 " with certain properties.\nFirst of all we use a lemma and infer from (1) \
",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "0"}], TraditionalForm]]],
 ".\n\nFrom (1), together with (CR2.\[Implies].1) and (CR2.\[Implies].2) we \
can infer\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
        RowBox[{"[", 
         RowBox[{"s", ",", "t"}], "]"}]}]], 
      RowBox[{"s", "\[NotEqual]", "t"}]]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "t"}]},
       {
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "s"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "t"}], "]"}], ",", 
          RowBox[{"lc", "[", "p", "]"}]}], "]"}]},
       {
        RowBox[{"isReducible", "[", 
         RowBox[{
          RowBox[{"C", "[", 
           RowBox[{"a", ",", "s"}], "]"}], ",", 
          RowBox[{"lc", "[", "p", "]"}]}], "]"}]}
      }]}]}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", "p", "]"}], "\[VerticalBar]", "u"}], TraditionalForm]]],
 "\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}]}], "]"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "u", "]"}]}], TraditionalForm]]],
 ".\n\nFrom (4), using (R8) in \[ScriptCapitalR], we can infer the existence \
of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c0", ",", "n"}], "]"}]}], TraditionalForm]]],
 " with\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c0", "=", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[Or]", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", "u"}], "]"}], "\[Succeeds]", "c0"}]}]}], 
   TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", "c0"}], "]"}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"c", ",", "m0"}], "]"}]}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"c", "\[NotEqual]", "0"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"m0", ",", "c"}], "]"}]}], ")"}]}], "\[Implies]", 
    RowBox[{"mult", "[", 
     RowBox[{
      RowBox[{"n", "*", "m0"}], ",", "c"}], "]"}]}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", "c", "]"}]}]], 
     RowBox[{"c0", "\[Succeeds]", 
      RowBox[{"c0", "+", "c"}]}]}], "\[Implies]", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"a", ",", "u"}], "]"}], "\[Succeeds]", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"a", ",", "u"}], "]"}], "+", 
      RowBox[{"n", "*", "c"}]}]}]}], TraditionalForm]]],
 "\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"b", ",", "c", ",", "d"}], "]"}]}]], "b"}], 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{"d", "\[Implies]", 
     RowBox[{
      RowBox[{"n", "*", "b"}], 
      SubscriptBox["\[LeftRightArrow]", "c"], 
      RowBox[{"n", "*", "d"}]}]}]}], TraditionalForm]]],
 "\n\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a0", ":=", 
     RowBox[{"c0", "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}]}], ",", 
    RowBox[{"m", ":=", 
     RowBox[{"n", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{"u", "/", 
        RowBox[{"lp", "[", "p", "]"}]}], ")"}]}]}]}], TraditionalForm]]],
 " witnesses our existential goal. Therefore, we have to show several \
subgoals.\n\n(G1): ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c0", "\[CenterDot]", 
     RowBox[{"lp", "[", "p", "]"}]}], "=", 
    RowBox[{"a", "\[Or]", 
     RowBox[{"a", "\[Succeeds]", 
      RowBox[{"c0", "\[CenterDot]", 
       RowBox[{"lp", "[", "p", "]"}]}]}]}]}], TraditionalForm]]],
 ".\nWe can use (R7.\[ScriptCapitalP].1), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", ":=", "a"}], ",", 
    RowBox[{"u", ":=", "u"}], ",", 
    RowBox[{"s", ":=", 
     RowBox[{"lp", "[", "p", "]"}]}], ",", 
    RowBox[{"c", ":=", "c0"}]}], TraditionalForm]]],
 ".\n\n(G2): ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", 
     RowBox[{"c0", "\[CenterDot]", 
      RowBox[{"lp", "[", "p", "]"}]}]}], "]"}], TraditionalForm]]],
 ".\nWe can use (CR4.\[DoubleLeftArrow]), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", ":=", "p"}], ",", 
    RowBox[{"q", ":=", "q"}], ",", 
    RowBox[{"c", ":=", "c0"}]}], TraditionalForm]]],
 ".\n\n(G3): ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", 
        RowBox[{"c", ",", "m0"}], "]"}]}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"c", "\[NotEqual]", "0"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{"m0", ",", "c"}], "]"}]}], ")"}]}], "\[Implies]", 
    RowBox[{"mult", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "\[CenterDot]", 
         RowBox[{"(", 
          RowBox[{"u", "/", 
           RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ")"}], "*", "m0"}], ",", 
      "c"}], "]"}]}], TraditionalForm]]],
 ".\nWe can use (R7.\[ScriptCapitalP].3), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"s", ":=", 
     RowBox[{"u", "/", 
      RowBox[{"lp", "[", "p", "]"}]}]}], ",", 
    RowBox[{"n", ":=", "n"}]}], TraditionalForm]]],
 ".\n\n(G4): ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", "c", "]"}]}]], 
     RowBox[{
      RowBox[{"c0", "\[CenterDot]", 
       RowBox[{"lp", "[", "p", "]"}]}], "\[Succeeds]", 
      RowBox[{
       RowBox[{"c0", "\[CenterDot]", 
        RowBox[{"lp", "[", "p", "]"}]}], "+", "c"}]}]}], "\[Implies]", 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"a", "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "\[CenterDot]", 
         RowBox[{"(", 
          RowBox[{"u", "/", 
           RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ")"}], "*", "c"}]}]}]}], 
   TraditionalForm]]],
 ".\nWe can use (R7.\[ScriptCapitalP].4), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", ":=", "a"}], ",", 
    RowBox[{"u", ":=", "u"}], ",", 
    RowBox[{"s", ":=", 
     RowBox[{"lp", "[", "p", "]"}]}], ",", 
    RowBox[{"c", ":=", "c0"}], ",", 
    RowBox[{"n", ":=", "n"}]}], TraditionalForm]]],
 ".\n\n(G5): ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", 
        RowBox[{"b", ",", "c", ",", "d"}], "]"}]}]], "b"}], 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{"d", "\[Implies]", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "\[CenterDot]", 
         RowBox[{"(", 
          RowBox[{"u", "/", 
           RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ")"}], "*", "b"}], 
      SubscriptBox["\[LeftRightArrow]", "c"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"n", "\[CenterDot]", 
         RowBox[{"(", 
          RowBox[{"u", "/", 
           RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ")"}], "*", "d"}]}]}]}], 
   TraditionalForm]]],
 ".\nWe can use (R7.\[ScriptCapitalP].5), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"s", ":=", 
     RowBox[{"u", "/", 
      RowBox[{"lp", "[", "p", "]"}]}]}], ",", 
    RowBox[{"n", ":=", "n"}]}], TraditionalForm]]],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632741258696162*^9, 3.632741366445916*^9}, 
   3.632741406802876*^9, {3.6327414924506493`*^9, 3.632741682410475*^9}, {
   3.632741944483564*^9, 3.63274198649291*^9}, {3.6327421404701967`*^9, 
   3.632742945305071*^9}, {3.632743002676351*^9, 3.63274302769627*^9}, {
   3.6327430855152082`*^9, 3.6327431094509497`*^9}, 3.632743708926586*^9, {
   3.632743762558179*^9, 3.632743783725276*^9}, {3.632743816867696*^9, 
   3.632743837029949*^9}, {3.632808314253736*^9, 3.632808385770213*^9}, {
   3.6328086373541813`*^9, 3.63280869542358*^9}, {3.632808820286887*^9, 
   3.632808866814271*^9}, {3.632809918793721*^9, 3.6328099437547903`*^9}, {
   3.632810947307803*^9, 3.632811031872114*^9}, {3.6328189648248377`*^9, 
   3.6328189815648537`*^9}, {3.6328215392243223`*^9, 
   3.6328216633172207`*^9}, {3.632822452179295*^9, 3.632822485444607*^9}, {
   3.6328253584483957`*^9, 3.632825616982862*^9}, {3.632825995837305*^9, 
   3.632826012754301*^9}, {3.632826054602469*^9, 3.6328261199159327`*^9}, {
   3.633072443433284*^9, 3.6330724926136436`*^9}, {3.6330756009683533`*^9, 
   3.633075912584158*^9}, {3.633075955450425*^9, 3.633076123267047*^9}, 
   3.6364403435758133`*^9, {3.63922228188933*^9, 3.639222289401947*^9}, {
   3.639222322005734*^9, 3.639222363891048*^9}, {3.639222397635127*^9, 
   3.639222403199768*^9}, {3.639301240362561*^9, 3.639301254792467*^9}, {
   3.639367535832364*^9, 3.639367801701914*^9}, {3.639367842628339*^9, 
   3.639367844839655*^9}, {3.6393678830915194`*^9, 3.639367885937715*^9}, {
   3.639367916152279*^9, 3.63936809784573*^9}, {3.6393684383304853`*^9, 
   3.639368442876555*^9}, 3.6434467857871933`*^9, 3.65320791120278*^9, 
   3.653213464503951*^9, 3.66617376680272*^9, 3.666440373929995*^9},
 CellID->891408206],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6330761037467213`*^9, 3.633076123267178*^9}, 
   3.636440343575931*^9, 3.643446785787347*^9, 3.65320791120278*^9, 
   3.653213464503951*^9, 3.66617376680272*^9, 3.666440373930131*^9},
 TextAlignment->Right,
 CellID->94010900]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Result (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6316025299269657`*^9, 3.631602540737727*^9}, 
   3.632216339177546*^9, {3.641116244352532*^9, 3.6411162485533943`*^9}},
 CellID->1466756562],

Cell["", "OpenEnvironment",
 CellID->927538807],

Cell[CellGroupData[{

Cell["\<\
Corollary: \[ScriptCapitalP] Constitutes Reduction Ring\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6315969483145847`*^9, 3.631596959638094*^9}, {
  3.631873078112784*^9, 3.631873089349092*^9}, {3.643444081044091*^9, 
  3.643444083075482*^9}},
 CellID->179190798],

Cell[BoxData[
 RowBox[{"isReductionRing", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(376)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1021018734", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185713817544*^9, 3.610185732163928*^9}, {
  3.631597062890481*^9, 3.631597135723353*^9}, {3.631597477496817*^9, 
  3.631597479977553*^9}, {3.631597539004051*^9, 3.631597540275116*^9}, {
  3.631600854757165*^9, 3.631600881561565*^9}},
 CellTags->{
  "ID:1021018734", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "376"},
 CellID->1021018734],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1086993414]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1230683486],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(376)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(376)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "isReductionRing", "[", "\[ScriptCapitalP]", 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "isReductionRing", "[", "\[ScriptCapitalP]", 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1021018734"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1021018734-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1021018734-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1021018734", "Proof|ID:1021018734-1"},
 CellID->1129550695],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1021018734-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->492049603]
}, Closed]],

Cell[TextData[{
 "We now generalize the results proved for the arbitrary but fixed domains \
\[ScriptCapitalP], \[ScriptCapitalR] and \[ScriptCapitalT] to ",
 StyleBox["all",
  FontSlant->"Italic"],
 " domains, or, more precisely, functors \[ScriptCapitalF] and domains \
\[ScriptCapitalR]0 and \[ScriptCapitalT]0. Only note that \[ScriptCapitalF] \
still has to belong to the category of reduction polynomials when applied to \
\[ScriptCapitalR] and \[ScriptCapitalT], not ot \[ScriptCapitalR]0 and \
\[ScriptCapitalT]0!"
}], "Text",
 CellChangeTimes->{{3.647158339508985*^9, 3.64715849675117*^9}},
 CellID->1207537800],

Cell["", "OpenEnvironment",
 CellID->543049971],

Cell[CellGroupData[{

Cell["Theorem", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.6471584874459476`*^9},
 CellID->1341395337],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
   "\[ScriptCapitalF]", ",", "\[ScriptCapitalR]0", ",", 
    "\[ScriptCapitalT]0"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"isReductionPolynomialDomain", "[", 
       RowBox[{"\[ScriptCapitalF]", "[", 
        RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], "]"}],
       "\[And]", 
      RowBox[{"isReductionRing", "[", "\[ScriptCapitalR]0", "]"}], "\[And]", 
      RowBox[{"isCommPPDomain", "[", "\[ScriptCapitalT]0", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{"isReductionRing", "[", 
     RowBox[{"\[ScriptCapitalF]", "[", 
      RowBox[{"\[ScriptCapitalR]0", ",", "\[ScriptCapitalT]0"}], "]"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReductionRing Conservation Theorem)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:459032599", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647158500986177*^9, 3.647158592708412*^9}},
 CellTags->{
  "ID:459032599", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReductionRing Conservation Theorem"},
 CellID->459032599],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->590387924]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->679962283]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Theorem: Algorithmic Reduction Polynomials Constitute Algorithmic Reduction \
Ring (\[Checkmark])\
\>", "Section",
 CellChangeTimes->{{3.636356988147478*^9, 3.636356997221738*^9}, {
  3.643444068595295*^9, 3.643444072646924*^9}, {3.650179886399527*^9, 
  3.65017988729583*^9}, {3.6507858910154247`*^9, 3.650785893566535*^9}},
 CellID->322583785],

Cell["\<\
Finally, we prove that the property of being an algorithmic reduction ring is \
conserved in domains \[ScriptCapitalP] belonging to category \
isAlgoReductionPolynomialDomain.\
\>", "Text",
 CellChangeTimes->{{3.6434461040073833`*^9, 3.643446193114513*^9}},
 CellID->1814342886],

Cell["", "OpenEnvironment",
 CellID->844754185],

Cell[CellGroupData[{

Cell["\<\
Presupposition: Algorithmic Reduction Ring \[ScriptCapitalR] and Algorithmic \
Polynomial Domain \[ScriptCapitalP]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.636357011832445*^9, 3.6363570232695503`*^9}, {
  3.643444054520553*^9, 3.64344406411862*^9}},
 CellID->243471114],

Cell[BoxData[
 RowBox[{"isAlgoReductionRing", "[", "\[ScriptCapitalR]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(algorithmic reduction ring \[ScriptCapitalR])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:719489133", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.636357037298221*^9, 3.636357040686339*^9}},
 CellTags->{
  "ID:719489133", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "algorithmic reduction ring \[ScriptCapitalR]"},
 CellID->719489133],

Cell[BoxData[
 RowBox[{"isAlgoReductionPolynomialDomain", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(algorithmic reduction-polynomial domain \[ScriptCapitalP])", 
         "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1159608082", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.636357071311399*^9, 3.636357082360372*^9}},
 CellTags->{
  "ID:1159608082", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", 
   "algorithmic reduction-polynomial domain \[ScriptCapitalP]"},
 CellID->1159608082],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1327883073]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->230307546],

Cell[CellGroupData[{

Cell["\[Tilde] in \[ScriptCapitalP] (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.650619287979848*^9, 3.650619298143303*^9}, {
  3.650785882564975*^9, 3.6507858851598*^9}},
 CellID->1690991356],

Cell["", "OpenEnvironment",
 CellID->558613337],

Cell[CellGroupData[{

Cell["Lemma: \[CenterDot] \[Succeeds] [obsolete]", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.650694544567307*^9, 3.650694551731085*^9}, {
  3.650778637877033*^9, 3.650778641064103*^9}},
 CellID->1738024641],

Cell[TextData[StyleBox["Formula (\[CenterDot] \[Succeeds]) is never used.",
 FontWeight->"Bold"]], "EnvironmentText",
 CellChangeTimes->{{3.650778619216748*^9, 3.6507786291286163`*^9}},
 CellID->1533594446],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
     RowBox[{"[", "c", "]"}]}], ",", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}], ",", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", "p", "]"}]}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"c", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "p"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"c", "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
         {
          RowBox[{
           RowBox[{"p", "\[NotEqual]", 
            UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{"t", 
              UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
             "\[Or]", 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               RowBox[{"t", "\[Equal]", 
                RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}],
                "\[And]", 
               RowBox[{"c", 
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", 
                 "]"}]}]}], 
              TagBox[")",
               "AutoParentheses"]}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[CenterDot] \[Succeeds])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:847570831", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.650694542770894*^9, {3.650696278318219*^9, 3.650696338674026*^9}},
 CellTags->{
  "ID:847570831", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] \[Succeeds]"},
 CellID->847570831],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->593969255]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->408203277],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[CenterDot] \[Succeeds])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[CenterDot] \[Succeeds])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox["p", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "\[Or]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "\[And]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             StyleBox["p", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "\[Or]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "\[And]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:847570831"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p847570831-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p847570831-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:847570831", "Proof|ID:847570831-1"},
 CellID->1483028309],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p847570831-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1094495224]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1742002312],

Cell[CellGroupData[{

Cell["Lemma: Preparations for \[Tilde] in \[ScriptCapitalP]", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.650619327065559*^9, {3.650697811676632*^9, 3.650697820821631*^9}},
 CellID->1137163506],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b", ",", "m"}], "]"}]}]], 
  RowBox[{
   RowBox[{
    UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", "m", "]"}], "\[And]", 
   
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Tilde]", "m"], "\[ScriptCapitalR]"], 
    "b"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 2071048931}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650619334998725*^9, 3.650619368023573*^9}},
 CellTags->{
  "ID:2071048931", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->2071048931],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "t", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 355437230}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65061937512077*^9, 3.650619381830967*^9}},
 CellTags->{
  "ID:355437230", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->355437230],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{
   RowBox[{"A", "=", 
    RowBox[{"a", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}]}], ",", 
   RowBox[{"B", "=", 
    RowBox[{"b", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}]}], ",", 
   RowBox[{"M", "=", 
    RowBox[{"m", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     UnderscriptBox["1", "\[ScriptCapitalT]"]}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1401739284}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6506194152623568`*^9, 3.650619426210423*^9}, {
  3.650619477657844*^9, 3.650619495273547*^9}},
 CellTags->{
  "ID:1401739284", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1401739284],

Cell[BoxData[
 RowBox[{"B", "\[Equal]", 
  RowBox[{"M", 
   UnderscriptBox["*", "\[ScriptCapitalP]"], 
   "A"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Tilde] \[ScriptCapitalP] 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:29665786", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6506193914396152`*^9, 3.650619441330182*^9}, {
  3.650619497809946*^9, 3.650619500421069*^9}},
 CellTags->{
  "ID:29665786", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 1"},
 CellID->29665786],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], "q"}], "\[Implies]", 
    
    RowBox[{
     RowBox[{"M", 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "p"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
     RowBox[{"M", 
      UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Tilde] \[ScriptCapitalP] 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:113282996", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650619452158124*^9, 3.650619470186926*^9}, {
  3.650619503768456*^9, 3.650619526127945*^9}, {3.650697411281221*^9, 
  3.650697423282596*^9}, {3.650697463147251*^9, 3.650697470450542*^9}, {
  3.650777595688662*^9, 3.650777608705154*^9}},
 CellTags->{
  "ID:113282996", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 2"},
 CellID->113282996],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"p", ",", "n"}], "]"}]}]], 
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
      RowBox[{"n", ",", "p", ",", "i"}], "]"}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
      RowBox[{
       RowBox[{"M", 
        UnderscriptBox["*", "\[ScriptCapitalP]"], "n"}], ",", "p", ",", "i"}],
       "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Tilde] \[ScriptCapitalP] 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1675445982", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6506195909501677`*^9, 3.650619653612227*^9}},
 CellTags->{
  "ID:1675445982", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 3"},
 CellID->1675445982],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "q", ",", "A", ",", "i", ",", "j"}], "]"}], 
     "\[Implies]", 
     RowBox[{
      UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "q", ",", "B", ",", "i", ",", "j"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Tilde] \[ScriptCapitalP] 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1968653014", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650619533625286*^9, 3.650619573624044*^9}},
 CellTags->{
  "ID:1968653014", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 4"},
 CellID->1968653014],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "p", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "A"}], "]"}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "B"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Tilde] \[ScriptCapitalP] 5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:860310855", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650619658905059*^9, 3.6506196679983883`*^9}},
 CellTags->{
  "ID:860310855", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 5"},
 CellID->860310855],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
  UnderscriptBox[
   SubscriptBox["\[Tilde]", 
    RowBox[{"m", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     UnderscriptBox["1", "\[ScriptCapitalT]"]}]], "\[ScriptCapitalP]"], 
  RowBox[{"b", 
   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
   "t"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Tilde] multiplier \[ScriptCapitalP])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2026748686", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650616557759495*^9, 3.650616589950383*^9}, {
  3.650616990837942*^9, 3.650616995873866*^9}, {3.650617051329563*^9, 
  3.650617051552628*^9}, {3.6506178078711023`*^9, 3.650617812936438*^9}, {
  3.650618248450396*^9, 3.650618254882053*^9}, {3.6506975089591103`*^9, 
  3.650697517493099*^9}},
 CellTags->{
  "ID:2026748686", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] multiplier \[ScriptCapitalP]"},
 CellID->2026748686],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1941983415]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1033726545],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalP] 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalP] 1)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], "\[ScriptCapitalR]"], 
               StyleBox["b", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox[
                 "t", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], "\[ScriptCapitalR]"], 
               StyleBox["b", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox[
                 "t", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:29665786"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p29665786-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p29665786-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:29665786", "Proof|ID:29665786-1"},
 CellID->343762797],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p29665786-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1884251853]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalP] 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalP] 2)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], "\[ScriptCapitalR]"], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["p", "ExpressionVariable"]}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "q", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], "\[ScriptCapitalR]"], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
               StyleBox["q", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox["p", "ExpressionVariable"]}], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalP]"], 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
               UnderscriptBox["*", "\[ScriptCapitalP]"], 
               StyleBox[
               "q", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:113282996"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p113282996-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p113282996-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:113282996", "Proof|ID:113282996-1"},
 CellID->1216564136],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p113282996-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->321450032]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalP] 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalP] 3)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], "\[ScriptCapitalR]"], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["n", "ExpressionVariable"]}], ",", 
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], "\[ScriptCapitalR]"], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
                 UnderscriptBox["*", "\[ScriptCapitalP]"], 
                 StyleBox["n", "ExpressionVariable"]}], ",", 
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1675445982"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1675445982-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1675445982-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1675445982", "Proof|ID:1675445982-1"},
 CellID->1330746921],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1675445982-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1849793449]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalP] 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalP] 4)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], "\[ScriptCapitalR]"], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], "\[ScriptCapitalR]"], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1968653014"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1968653014-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1968653014-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1968653014", "Proof|ID:1968653014-1"},
 CellID->1366469966],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1968653014-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->461522218]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalP] 5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalP] 5)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], "\[ScriptCapitalR]"], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], "\[ScriptCapitalR]"], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   StyleBox["t", "ExpressionVariable"]}]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 StyleBox["t", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:860310855"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p860310855-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p860310855-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:860310855", "Proof|ID:860310855-1"},
 CellID->1282192616],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p860310855-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->306440183]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Tilde] multiplier \[ScriptCapitalP])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Tilde] multiplier \[ScriptCapitalP])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], "\[ScriptCapitalR]"], 
               StyleBox["b", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Tilde]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}]], 
              "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isUnit", "\[ScriptCapitalR]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], "\[ScriptCapitalR]"], 
               StyleBox["b", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Tilde]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 UnderscriptBox["1", "\[ScriptCapitalT]"]}]], 
              "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2026748686"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2026748686-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2026748686-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2026748686", "Proof|ID:2026748686-1"},
 CellID->1041146073],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2026748686-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1336512602]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1738547973],

Cell[CellGroupData[{

Cell["Theorem: ~ in \[ScriptCapitalP]", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6501790107442293`*^9, 3.65017901618734*^9}},
 CellID->1666042888],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "t", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", 
       UnderscriptBox["\[Tilde]", "\[ScriptCapitalR]"], "b"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"a", 
        UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], 
       UnderscriptBox["\[Tilde]", "\[ScriptCapitalP]"], 
       RowBox[{"b", 
        UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Tilde] \[ScriptCapitalP])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1529886137", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6501790212837133`*^9, 3.650179103888792*^9}, {
  3.650616549151911*^9, 3.6506165532732553`*^9}, {3.6507852582825623`*^9, 
  3.650785283585793*^9}},
 CellTags->{
  "ID:1529886137", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP]"},
 CellID->1529886137],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->60767287]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2095809430],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalP])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalP])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", "\[ScriptCapitalR]"], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["\[Tilde]", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", "\[ScriptCapitalR]"], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox["t", "ExpressionVariable"]}], 
             UnderscriptBox["\[Tilde]", "\[ScriptCapitalP]"], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1529886137"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1529886137-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1529886137-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1529886137", "Proof|ID:1529886137-1"},
 CellID->1495738535],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1529886137-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1783742101]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R9 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.631960137922221*^9, 3.631960138589053*^9}, {
  3.642931093968978*^9, 3.6429310978054333`*^9}, {3.650179879394952*^9, 
  3.650179880355364*^9}, {3.650619312670801*^9, 3.6506193151769876`*^9}},
 CellID->1708894494],

Cell["", "OpenEnvironment",
 CellID->2072032119],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R9", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.631960145814007*^9, 3.631960152236908*^9}},
 CellID->1748702921],

Cell[BoxData[
 RowBox[{"R9", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(372)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1680577066", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185612249834*^9, 3.610185630801051*^9}, 
   3.63159686651439*^9},
 CellTags->{
  "ID:1680577066", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "372"},
 CellID->1680577066],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1237743236]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1965785607],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(372)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(372)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R9", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R9", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1680577066"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1680577066-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1680577066-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1680577066", "Proof|ID:1680577066-1"},
 CellID->5711938],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1680577066-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->302678198]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6319601737130823`*^9, 3.63196017453519*^9}, 
   3.64292540980122*^9, 3.643446807759399*^9, 3.653207898082156*^9, 
   3.6532134536151314`*^9, 3.666173754026298*^9, 3.666440359080284*^9},
 CellID->1663153286],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642924871866037*^9, 3.642924873321045*^9}, 
   3.642925409801359*^9, 3.6434468077595577`*^9, 3.653207898082156*^9, 
   3.6532134536151314`*^9, 3.666173754026298*^9, 3.6664403590803843`*^9},
 CellID->1756388746],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isMNTCR", "[", 
     RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
        RowBox[{"[", "c", "]"}]}]], 
      RowBox[{"c", "\[NotEqual]", "0"}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"isMNTCRP", "[", 
         RowBox[{"p", ",", "q", ",", "c", ",", "i", ",", "j"}], "]"}], 
        "\[And]", "a"}], "=", 
       RowBox[{"c", "\[CenterDot]", 
        RowBox[{"lcm", "[", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], ",", 
          RowBox[{"lp", "[", "q", "]"}]}], "]"}]}]}], ")"}]}]}], 
   TraditionalForm]]],
 " ... (CR3)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642924877931594*^9, 3.642924928934938*^9}, {
   3.642925256925984*^9, 3.642925281090217*^9}, {3.642925340449934*^9, 
   3.6429253422569304`*^9}, 3.642925409801485*^9, {3.642925455602372*^9, 
   3.642925460564909*^9}, 3.643446807759639*^9, 3.653207898082156*^9, 
   3.6532134536151314`*^9, 3.666173754026298*^9, 3.666440359080467*^9},
 CellID->973735291],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"isMNTCR", "[", 
     RowBox[{"p", ",", "a"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
        RowBox[{"[", "c", "]"}]}]], 
      RowBox[{"c", "\[NotEqual]", "0"}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"isMNTCR", "[", 
         RowBox[{
          RowBox[{"lc", "[", "p", "]"}], ",", "c"}], "]"}], "\[And]", "a"}], 
       "=", 
       RowBox[{"c", "\[CenterDot]", 
        RowBox[{"lp", "[", "p", "]"}]}]}], ")"}]}]}], TraditionalForm]]],
 " ... (CR4)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642924877931594*^9, 3.642924928934938*^9}, {
   3.642925256925984*^9, 3.642925304058496*^9}, {3.642925344940962*^9, 
   3.642925360495163*^9}, 3.642925409801564*^9, {3.6429254631112833`*^9, 
   3.642925467080327*^9}, 3.6434468077597017`*^9, 3.653207898082156*^9, 
   3.6532134536151314`*^9, 3.666173754026298*^9, 3.666440359080524*^9},
 CellID->807494688],

Cell[TextData[{
 "We choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "p", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "q", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalN]", 
      RowBox[{"1", ",", "\[Ellipsis]", ",", 
       RowBox[{"multN", "[", "p", "]"}]}]]}]}], TraditionalForm]]],
 " a.b.f. and assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "\[NotEqual]", "0"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"q", "\[NotEqual]", "0"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "<", "j"}], TraditionalForm]]],
 ". We have to show both of\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalP]", "]"}]], 
    RowBox[{"[", 
     RowBox[{"lcrd", "[", 
      RowBox[{"p", ",", "q", ",", "i", ",", "j"}], "]"}], "]"}]}], 
   TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "a", "]"}]}]], 
    RowBox[{
     RowBox[{"a", 
      TagBox["\:22ff",
       Identity,
       SyntaxForm->"a+b"], 
      RowBox[{"lcrd", "[", 
       RowBox[{"p", ",", "q", ",", "i", ",", "j"}], "]"}]}], 
     "\[DoubleLeftRightArrow]", 
     RowBox[{"isMNTCR", "[", 
      RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}]}]}], 
   TraditionalForm]]],
 "\n\nFrom ",
 Cell[BoxData[
  FormBox[
   RowBox[{"hasLCRD", "[", "\[ScriptCapitalP]", "]"}], TraditionalForm]]],
 " we know, for the particular choices of ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["q", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 ":\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "q", ",", "i", ",", "j"}], "]"}], "\[Equal]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {"...", "\[DoubleLeftArrow]", 
        RowBox[{
         RowBox[{"p", "\[Equal]", "q"}], "\[Or]", 
         RowBox[{
          RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
          RowBox[{"lc", "[", "q", "]"}]}]}]},
       {"...", "\[DoubleLongLeftArrow]", "otherwise"}
      }]}]}], TraditionalForm]]],
 "\n\nhence we distinguish two cases; However, we leave out the first \
subgoal, as it is shown rather easily anyway.\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Or]", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
      RowBox[{"lc", "[", "q", "]"}]}]}]}], TraditionalForm]]],
 ".\nWe know\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lcrd", "[", 
     RowBox[{"q", ",", "q", ",", "i", ",", "j"}], "]"}], "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{
       SubscriptBox["l", "k"], "\[CenterDot]", 
       RowBox[{"lcm", "[", 
        RowBox[{
         RowBox[{"lp", "[", "p", "]"}], ",", 
         RowBox[{"lp", "[", "q", "]"}]}], "]"}]}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "l", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 " and have to show\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "a", "]"}]}]], 
    RowBox[{
     RowBox[{"a", " ", "\:22ff", " ", 
      RowBox[{"\[LeftAngleBracket]", "...", "\[RightAngleBracket]"}]}], 
     "\[DoubleLeftRightArrow]", 
     RowBox[{"isMNTCR", "[", 
      RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}]}]}], 
   TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox["l", TraditionalForm]]],
 " abbreviates ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcrd", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 ".\nWe choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "a", "]"}]}], TraditionalForm]]],
 " a.b.f. and show the equivalence.\nInstead of proving both directions, we \
first simplify ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", " ", "\:22ff", " ", 
    RowBox[{"\[LeftAngleBracket]", "...", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 " to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "l", "\[RightBracketingBar]"}]}]], 
     "a"}], "=", 
    RowBox[{
     SubscriptBox["l", "k"], "\[CenterDot]", 
     RowBox[{"lcm", "[", 
      RowBox[{
       RowBox[{"lp", "[", "p", "]"}], ",", 
       RowBox[{"lp", "[", "q", "]"}]}], "]"}]}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 " to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[Exists]", 
        RowBox[{"\[Element]", 
         RowBox[{"[", "c", "]"}]}]], 
       RowBox[{"c", "\[NotEqual]", "0"}]], 
      RowBox[{"isMNTCRP", "[", 
       RowBox[{"p", ",", "q", ",", "c", ",", "i", ",", "j"}], "]"}]}], 
     "\[And]", "a"}], "=", 
    RowBox[{"c", "\[CenterDot]", 
     RowBox[{"lcm", "[", "\[Ellipsis]", "]"}]}]}], TraditionalForm]]],
 ".\nNow we show both directions of the equivalence:\n\[OpenCurlyDoubleQuote]\
\[DoubleLongRightArrow]\[CloseCurlyDoubleQuote].\nWe assume\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", 
    RowBox[{
     SubscriptBox["l", "k"], "\[CenterDot]", 
     RowBox[{"lcm", "[", 
      RowBox[{
       RowBox[{"lp", "[", "p", "]"}], ",", 
       RowBox[{"lp", "[", "q", "]"}]}], "]"}]}]}], TraditionalForm]]],
 "\nfor some constant ",
 Cell[BoxData[
  FormBox["k", TraditionalForm]]],
 " and show\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", "q", ",", 
     RowBox[{
      SubscriptBox["l", "k"], "\[CenterDot]", 
      RowBox[{"lcm", "[", "\[Ellipsis]", "]"}]}]}], "]"}], 
   TraditionalForm]]],
 "\n\nBy (CR3.\[DoubleLongLeftArrow]) the goal can be reduced to\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", 
     SubscriptBox["l", "k"], ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\nwhich, by (isMNTCRP alternative) and our case assumption, is equivalent \
to showing\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", 
     SubscriptBox["l", "k"], ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\n\nNow we use (R9) in \[ScriptCapitalR], instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c1", ":=", 
     RowBox[{"lc", "[", "p", "]"}]}], ",", 
    RowBox[{"c2", ":=", 
     RowBox[{"lc", "[", "q", "]"}]}], ",", 
    RowBox[{"i", ":=", "i"}], ",", 
    RowBox[{"j", ":=", "j"}], ",", 
    RowBox[{"a", ":=", 
     SubscriptBox["l", "k"]}]}], TraditionalForm]]],
 ". All premises are satisfied, hence we can conclude\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["l", "k"], 
    TagBox["\:22ff",
     Identity,
     SyntaxForm->"a+b"], 
    RowBox[{"l", "\[DoubleLongLeftRightArrow]", 
     RowBox[{"isMNTCR", "[", 
      RowBox[{
       RowBox[{"lc", "[", "p", "]"}], ",", 
       RowBox[{"lc", "[", "q", "]"}], ",", 
       SubscriptBox["l", "k"], ",", "i", ",", "j"}], "]"}]}]}], 
   TraditionalForm]]],
 "\nThe LHS of the equivalence in (4) holds vacuously, and the RHS is just \
our current goal (G6).\n\n\[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\
\[CloseCurlyDoubleQuote]:\nWe assume\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", "q", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand show\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "l", "\[RightBracketingBar]"}]}]], 
     "a"}], "=", 
    RowBox[{
     SubscriptBox["l", "k"], "\[CenterDot]", 
     RowBox[{"lcm", "[", 
      RowBox[{
       RowBox[{"lp", "[", "p", "]"}], ",", 
       RowBox[{"lp", "[", "q", "]"}]}], "]"}]}]}], TraditionalForm]]],
 "\nFrom (5), by (CR3.\[DoubleLongRightArrow]), we can infer the existence of \
",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}], TraditionalForm]]],
 " with\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "c", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", 
    RowBox[{"c", "\[CenterDot]", 
     RowBox[{"lcm", "[", 
      RowBox[{
       RowBox[{"lp", "[", "p", "]"}], ",", 
       RowBox[{"lp", "[", "q", "]"}]}], "]"}]}]}], TraditionalForm]]],
 "\nFrom (8), by (isMNTCRP alternative) and our case assumption, we can infer\
\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n\nNow we use (R9) in \[ScriptCapitalR], instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c1", ":=", 
     RowBox[{"lc", "[", "p", "]"}]}], ",", 
    RowBox[{"c2", ":=", 
     RowBox[{"lc", "[", "q", "]"}]}], ",", 
    RowBox[{"i", ":=", "i"}], ",", 
    RowBox[{"j", ":=", "j"}], ",", 
    RowBox[{"a", ":=", "c"}]}], TraditionalForm]]],
 ". All premises are satisfied, hence we can conclude\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", " ", "\:22ff", " ", 
    RowBox[{"l", "\[DoubleLongLeftRightArrow]", 
     RowBox[{"isMNTCR", "[", 
      RowBox[{
       RowBox[{"lc", "[", "p", "]"}], ",", 
       RowBox[{"lc", "[", "q", "]"}], ",", "c", ",", "i", ",", "j"}], 
      "]"}]}]}], TraditionalForm]]],
 "\nThe RHS of the equivalence in (11) holds due to (10), hence we can infer \
the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", "l"}], 
   TraditionalForm]]],
 " with\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "=", 
    SubscriptBox["l", "k"]}], TraditionalForm]]],
 "\nIt is apparent that ",
 Cell[BoxData[
  FormBox["k", TraditionalForm]]],
 " witnesses our existential goal (G7).\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"p", "\[NotEqual]", "q"}], "\[And]", 
     RowBox[{"lc", "[", "p", "]"}]}], "=", 
    RowBox[{"lc", "[", "q", "]"}]}], TraditionalForm]]],
 ".\nAnalogous, using (R10) in \[ScriptCapitalR]."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.631960185747518*^9, 3.631960189589169*^9}, {
   3.63196030977815*^9, 3.6319603528864317`*^9}, {3.6319603901529303`*^9, 
   3.631960465435321*^9}, {3.6319605122764797`*^9, 3.631960583034803*^9}, {
   3.631961443406657*^9, 3.631961865505542*^9}, {3.631963551010868*^9, 
   3.631963813064157*^9}, 3.632216374570512*^9, {3.632216432766736*^9, 
   3.632216434096241*^9}, {3.642923349547879*^9, 3.6429234647333097`*^9}, {
   3.642923503163425*^9, 3.642923614572288*^9}, {3.642923651904125*^9, 
   3.6429236729270077`*^9}, {3.6429237833753853`*^9, 3.642923855910254*^9}, {
   3.642923888969553*^9, 3.642923919553049*^9}, {3.642923949679357*^9, 
   3.642923987481371*^9}, {3.642924022482359*^9, 3.642924148328854*^9}, {
   3.642924280236669*^9, 3.642924767956283*^9}, {3.642924943094812*^9, 
   3.642925210599435*^9}, {3.6429253228828487`*^9, 3.64292532321939*^9}, {
   3.642925409801663*^9, 3.642925426669684*^9}, {3.64292668967556*^9, 
   3.6429267296817207`*^9}, 3.643446807759799*^9, 3.653207898082156*^9, 
   3.6532134536151314`*^9, 3.666173754026298*^9, 3.666440359080604*^9},
 CellID->347176547],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642925390620882*^9, 3.642925409801784*^9}, 
   3.6434468077599154`*^9, 3.653207898082156*^9, 3.6532134536151314`*^9, 
   3.666173754026298*^9, 3.666440359080707*^9},
 TextAlignment->Right,
 CellID->1358676932]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R10 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6330762714446573`*^9, 3.633076272367988*^9}, {
  3.642931922026218*^9, 3.6429319260161037`*^9}},
 CellID->268409188],

Cell["", "OpenEnvironment",
 CellID->24273816],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R10", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.633076277916091*^9, 3.63307628479242*^9}},
 CellID->975532904],

Cell[BoxData[
 RowBox[{"R10", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(373)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1866021817", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185612919939*^9, 3.610185632765562*^9}, 
   3.631596867220943*^9},
 CellTags->{
  "ID:1866021817", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "373"},
 CellID->1866021817],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->795876134]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1616684133],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(373)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(373)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R10", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R10", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1866021817"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1866021817-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1866021817-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1866021817", "Proof|ID:1866021817-1"},
 CellID->1214587],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1866021817-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->169993026]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R11 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.633076296265122*^9, 3.6330762967849703`*^9}, {
  3.643104638455213*^9, 3.643104642436846*^9}},
 CellID->1084053984],

Cell["", "OpenEnvironment",
 CellID->339961261],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R11", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.633076303288311*^9, 3.633076311748599*^9}},
 CellID->2108760469],

Cell[BoxData[
 RowBox[{"R11", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(374)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1272742036", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185613590108*^9, 3.610185634287837*^9}, 
   3.631596867816412*^9},
 CellTags->{
  "ID:1272742036", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "374"},
 CellID->1272742036],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->263791477]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2106755448],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(374)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(374)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R11", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R11", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1272742036"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1272742036-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1272742036-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1272742036", "Proof|ID:1272742036-1"},
 CellID->634133424],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1272742036-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->107031843]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R12 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.633076322700796*^9, 3.63307632325666*^9}, {
  3.643443698474853*^9, 3.643443702392148*^9}},
 CellID->1481007890],

Cell["", "OpenEnvironment",
 CellID->1777956310],

Cell[CellGroupData[{

Cell["Lemma", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.643362474036618*^9},
 CellID->139848971],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 181479986}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643341998437498*^9, 3.643342023383795*^9}, {
   3.643355009942375*^9, 3.643355027766646*^9}, {3.643355573168229*^9, 
   3.643355582097849*^9}, 3.6433560021887703`*^9, 3.64336253362963*^9, {
   3.643362720676979*^9, 3.643362742047858*^9}, 3.643362870218589*^9},
 CellTags->{
  "ID:181479986", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->181479986],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", "c", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1276235216}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643366543948287*^9, 3.643366551381308*^9}},
 CellTags->{
  "ID:1276235216", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1276235216],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
  RowBox[{"i", "<", "j"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 376128819}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643342032955171*^9, 3.64334204558191*^9}, {
  3.64336432114187*^9, 3.6433643228737583`*^9}},
 CellTags->{
  "ID:376128819", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->376128819],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"l", "=", 
   RowBox[{
    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
     RowBox[{
      UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 973122334}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6433635516355*^9, 3.643363566894245*^9}, {
  3.6433665539857616`*^9, 3.643366577514888*^9}, {3.643368178874939*^9, 
  3.643368182035552*^9}},
 CellTags->{
  "ID:973122334", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->973122334],

Cell[TextData[{
 "Formula (375.1) only holds for those ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lcrd", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 ", but this should be sufficient."
}], "EnvironmentText",
 CellChangeTimes->{{3.6501092436188507`*^9, 3.650109279393053*^9}},
 CellID->663596886],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"c", 
    TagBox["\:22ff",
     Identity,
     SyntaxForm->"a+b"], 
    RowBox[{
     UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", "i", 
      ",", "j"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", 
       RowBox[{"p1", ",", "p2"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{
        UnderscriptBox["cpm", "\[ScriptCapitalR]"], "[", 
        RowBox[{
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", 
         "c", ",", "i", ",", "j"}], "]"}], "\[Equal]", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"p1", ",", "p2"}], "\[RightAngleBracket]"}]}], "\[And]", 
      RowBox[{
       UnderscriptBox["areCPMultipliers", "\[ScriptCapitalR]"], "[", 
       RowBox[{"p1", ",", "p2", ",", 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
        RowBox[{
         UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", "c",
         ",", "i", ",", "j"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(375.1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1445738544", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6434387208651543`*^9, 3.643438805912269*^9}, {
  3.650179992313925*^9, 3.65018002523024*^9}, {3.650180393047269*^9, 
  3.650180396929473*^9}},
 CellTags->{
  "ID:1445738544", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.1"},
 CellID->1445738544],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"p", "\[NotEqual]", "q"}], "\[And]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]", 
     
     RowBox[{
      UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}], "\[And]", 
    
    RowBox[{
     UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", "c"}],
      "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["areCPMultipliers", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["rdm", "\[ScriptCapitalR]"], "[", 
      RowBox[{"c", ",", 
       RowBox[{
        UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}]}], "]"}], 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{"l", 
       UnderscriptBox["/", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], ")"}]}],
     ",", 
    RowBox[{
     RowBox[{
      UnderscriptBox["rdm", "\[ScriptCapitalR]"], "[", 
      RowBox[{"c", ",", 
       RowBox[{
        UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}]}], "]"}], 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
     RowBox[{"(", 
      RowBox[{"l", 
       UnderscriptBox["/", "\[ScriptCapitalT]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], ")"}]}],
     ",", "p", ",", "q", ",", 
    RowBox[{"c", 
     UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "l"}], ",", "i", 
    ",", "j"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(375.2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:874376987", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643362663474925*^9, 3.643362717593266*^9}, {
  3.643362756455618*^9, 3.643362839449234*^9}, {3.643362897333057*^9, 
  3.6433629360242653`*^9}, {3.64336298692185*^9, 3.64336300967994*^9}, {
  3.6433634113929367`*^9, 3.64336342508704*^9}, {3.643363582520907*^9, 
  3.6433635955931597`*^9}, {3.643363765500866*^9, 3.643363765760498*^9}, {
  3.643366585278081*^9, 3.643366615792282*^9}, {3.643368146684987*^9, 
  3.643368257804824*^9}, {3.64336866783939*^9, 3.643368671675603*^9}, {
  3.643368734454076*^9, 3.643368754960785*^9}, {3.6434295273037643`*^9, 
  3.643429531156745*^9}},
 CellTags->{
  "ID:874376987", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.2"},
 CellID->874376987],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", 
     RowBox[{
      UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "q", ",", "i", ",", "j"}], "]"}], 
     "\[RightBracketingBar]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "d", "]"}]}]], 
    RowBox[{"d", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      SubscriptBox[
       RowBox[{
        UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
        RowBox[{"p", ",", "q", ",", "i", ",", "j"}], "]"}], "k"], "\[Equal]", 
      
      RowBox[{"d", 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "l"}]}], "\[And]", 
     RowBox[{
      UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "q", ",", "d", ",", "i", ",", "j"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(375.3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:139929842", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643440050604836*^9, 3.643440113062628*^9}, {
  3.643440170128882*^9, 3.6434401749346247`*^9}},
 CellTags->{
  "ID:139929842", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.3"},
 CellID->139929842],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->505373945]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1110225807],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(375.1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(375.1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "<", 
               StyleBox["j", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\:22ff", 
               RowBox[{
                 UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["p1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["p2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["cpm", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["p1", "ExpressionVariable"], ",", 
                    StyleBox["p2", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["areCPMultipliers", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   StyleBox["p1", "ExpressionVariable"], ",", 
                   StyleBox["p2", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "<", 
               StyleBox["j", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\:22ff", 
               RowBox[{
                 UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["p1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["p2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["cpm", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["p1", "ExpressionVariable"], ",", 
                    StyleBox["p2", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["areCPMultipliers", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                   StyleBox["p1", "ExpressionVariable"], ",", 
                   StyleBox["p2", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1445738544"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #2:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1445738544-2-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1445738544-2"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1445738544", "Proof|ID:1445738544-2"},
 CellID->1345220325],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1445738544-2-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->1091818374]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(375.2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(375.2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "<", 
               StyleBox["j", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["q", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["areCPMultipliers", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["rdm", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["rdm", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
               ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "<", 
               StyleBox["j", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["q", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 StyleBox["q", "ExpressionVariable"], "]"}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["areCPMultipliers", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["rdm", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}]}], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["rdm", "\[ScriptCapitalR]"], "[", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], "]"}], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}], 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["q", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                 RowBox[{
                   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
               ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:874376987"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p874376987-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p874376987-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:874376987", "Proof|ID:874376987-1"},
 CellID->7612480],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p874376987-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->277234]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(375.3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(375.3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 RowBox[{
                   UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "<", 
               StyleBox["j", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["d", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
             RowBox[{
               RowBox[{
                 SubscriptBox[
                  RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                  StyleBox["k", "ExpressionVariable"]], "\[Equal]", 
                 RowBox[{
                   StyleBox["d", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}]}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["d", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 RowBox[{
                   UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "<", 
               StyleBox["j", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["d", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
             RowBox[{
               RowBox[{
                 SubscriptBox[
                  RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                  StyleBox["k", "ExpressionVariable"]], "\[Equal]", 
                 RowBox[{
                   StyleBox["d", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
                   RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}]}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["isMNTCRP", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["d", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:139929842"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #2:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p139929842-2-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p139929842-2"]]]]}, {None, None}},
 CellTags->{"Proof|ID:139929842", "Proof|ID:139929842-2"},
 CellID->1858784208],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p139929842-2-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1228384409]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1578377743],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalP] satisfies R12", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.631596251345262*^9, {3.633076334688669*^9, 3.6330763394828243`*^9}},
 CellID->272592021],

Cell[BoxData[
 RowBox[{"R12", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(375)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:462890005", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185617448372*^9, 3.610185635437545*^9}, 
   3.631596868485298*^9},
 CellTags->{
  "ID:462890005", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375"},
 CellID->462890005],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1497256452]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1960788671],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(375)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(375)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R12", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R12", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:462890005"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p462890005-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p462890005-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:462890005", "Proof|ID:462890005-1"},
 CellID->51997037],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p462890005-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {68., 75.}}]], "ProofInfoBody",
 CellID->468582439]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.643342550285255*^9, 3.6433425511194344`*^9}, 
   3.6433549688685913`*^9, 3.643446821952408*^9, 3.6532078840703316`*^9, 
   3.6532134432879133`*^9, 3.6661737420298767`*^9, 3.666440348265473*^9},
 CellID->556816827],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "p", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", 
     RowBox[{
      RowBox[{"lcrd", "[", 
       RowBox[{"p", ",", "q", ",", "i", ",", "j"}], "]"}], "=", 
      RowBox[{":", "L"}]}], "\[RightBracketingBar]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(0)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "\[NotEqual]", "0"}], "\[And]", 
    RowBox[{"q", "\[NotEqual]", "0"}]}], TraditionalForm]]],
 "\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "<", "j"}], TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"p1", ",", "p2"}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{
          RowBox[{"cpm", "[", 
           RowBox[{"p", ",", "q", ",", 
            SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}], "=", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"p1", ",", "p2"}], "\[RightAngleBracket]"}]}]},
        {
         RowBox[{"areCPMultipliers", "[", 
          RowBox[{"p1", ",", "p2", ",", "p", ",", "q", ",", 
           SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\n\nFrom the definition of cpm in \[ScriptCapitalP] we know\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"cpm", "[", 
     RowBox[{"p", ",", "q", ",", 
      SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}], "=", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{
           SubscriptBox["cm", "1"], "\[CenterDot]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"lp", "[", 
              SubscriptBox["L", "k"], "]"}], "/", 
             RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ",", 
          RowBox[{
           SubscriptBox["cm", "2"], "\[CenterDot]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"lp", "[", 
              SubscriptBox["L", "k"], "]"}], "/", 
             RowBox[{"lp", "[", "q", "]"}]}], ")"}]}]}], 
         "\[RightAngleBracket]"}], "\[DoubleLongLeftArrow]", 
        RowBox[{"p", "=", 
         RowBox[{"q", "\[Or]", 
          RowBox[{
           RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
           RowBox[{"lc", "[", "q", "]"}]}]}]}]},
       {
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"r", "\[CenterDot]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"lp", "[", 
              SubscriptBox["L", "k"], "]"}], "/", 
             RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ",", 
          RowBox[{"r", "\[CenterDot]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"lp", "[", 
              SubscriptBox["L", "k"], "]"}], "/", 
             RowBox[{"lp", "[", "q", "]"}]}], ")"}]}]}], 
         "\[RightAngleBracket]"}], "\[DoubleLongLeftArrow]", "otherwise"}
      }]}]}], TraditionalForm]]],
 "\nwhere cm denotes ",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpm", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", 
     RowBox[{"lc", "[", 
      SubscriptBox["L", "k"], "]"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["r", TraditionalForm]]],
 " denotes ",
 Cell[BoxData[
  FormBox[
   RowBox[{"rdm", "[", 
    RowBox[{
     RowBox[{"lc", "[", 
      SubscriptBox["L", "k"], "]"}], ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 ".\n\nFrom (R9) in \[ScriptCapitalP] we know\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{"p", ",", "q", ",", 
     SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\nand from (CR3) we can infer\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCRP", "[", 
    RowBox[{"p", ",", "q", ",", "c", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["L", "k"], "=", 
    RowBox[{"c", "\[CenterDot]", 
     RowBox[{"lcm", "[", 
      RowBox[{
       RowBox[{"lp", "[", "p", "]"}], ",", 
       RowBox[{"lp", "[", "q", "]"}]}], "]"}]}]}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[NotEqual]", "0"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}], TraditionalForm]]],
 ".\nFrom (6) we can infer\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lc", "[", 
     SubscriptBox["L", "k"], "]"}], "=", "c"}], TraditionalForm]]],
 "\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lp", "[", 
     SubscriptBox["L", "k"], "]"}], "=", 
    RowBox[{"lcm", "[", 
     RowBox[{
      RowBox[{"lp", "[", "p", "]"}], ",", 
      RowBox[{"lp", "[", "q", "]"}]}], "]"}]}], TraditionalForm]]],
 "\nTherefore, from (2) we can infer\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"cpm", "[", 
     RowBox[{"p", ",", "q", ",", 
      SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}], "=", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"c1", "\[CenterDot]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"lcm", "[", 
              RowBox[{
               RowBox[{"lp", "[", "p", "]"}], ",", 
               RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
             RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ",", 
          RowBox[{"c2", "\[CenterDot]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"lcm", "[", 
              RowBox[{
               RowBox[{"lp", "[", "p", "]"}], ",", 
               RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
             RowBox[{"lp", "[", "q", "]"}]}], ")"}]}]}], 
         "\[RightAngleBracket]"}], "\[DoubleLongLeftArrow]", 
        RowBox[{"p", "=", 
         RowBox[{"q", "\[Or]", 
          RowBox[{
           RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
           RowBox[{"lc", "[", "q", "]"}]}]}]}]},
       {
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"r", "\[CenterDot]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"lcm", "[", 
              RowBox[{
               RowBox[{"lp", "[", "p", "]"}], ",", 
               RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
             RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ",", 
          RowBox[{"r", "\[CenterDot]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"lcm", "[", 
              RowBox[{
               RowBox[{"lp", "[", "p", "]"}], ",", 
               RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
             RowBox[{"lp", "[", "q", "]"}]}], ")"}]}]}], 
         "\[RightAngleBracket]"}], "\[DoubleLongLeftArrow]", "otherwise"}
      }]}]}], TraditionalForm]]],
 "\n\nWe distinguish two cases, based upon whether ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Or]", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
      RowBox[{"lc", "[", "q", "]"}]}]}]}], TraditionalForm]]],
 " or not.\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"p", "=", 
    RowBox[{"q", "\[Or]", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], "\[NotEqual]", 
      RowBox[{"lc", "[", "q", "]"}]}]}]}], TraditionalForm]]],
 ".\nWe know\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"cpm", "[", 
     RowBox[{"p", ",", "q", ",", 
      SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}], "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{
       SubscriptBox["cm", "1"], "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ",", 
      RowBox[{
       SubscriptBox["cm", "2"], "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "q", "]"}]}], ")"}]}]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 "\nFrom (R12) in \[ScriptCapitalR] we know\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"cpm", "[", 
     RowBox[{
      RowBox[{"lc", "[", "p", "]"}], ",", 
      RowBox[{"lc", "[", "q", "]"}], ",", "c", ",", "i", ",", "j"}], "]"}], 
    "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"c1", ",", "c2"}], "\[RightAngleBracket]"}]}], TraditionalForm]]],
 "\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCPMultipliers", "[", 
    RowBox[{"c1", ",", "c2", ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}], TraditionalForm]]],
 ".\nTherefore, from (7) we can infer\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"cpm", "[", 
     RowBox[{"p", ",", "q", ",", 
      SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}], "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"c1", "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ",", 
      RowBox[{"c2", "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "q", "]"}]}], ")"}]}]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 "\n \n We claim that the two constituents of the tuple in (14) witness our \
existential goal, hence we show\n(G.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"c1", "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"c2", "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "q", "]"}]}], ")"}]}], "]"}]}]}], 
   TraditionalForm]]],
 "\n(G.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"cpm", "[", 
     RowBox[{"p", ",", "q", ",", 
      SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}], "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"c1", "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ",", 
      RowBox[{"c2", "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "q", "]"}]}], ")"}]}]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 "\n(G.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCPMultipliers", "[", 
    RowBox[{
     RowBox[{"c1", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"lcm", "[", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], ",", 
          RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
        RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ",", 
     RowBox[{"c2", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"lcm", "[", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], ",", 
          RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
        RowBox[{"lp", "[", "q", "]"}]}], ")"}]}], ",", "p", ",", "q", ",", 
     SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\n(G.1) is obvious and (G.2) is identical to (14), hence we only need to \
show (G.3).\n\nA sufficient condition for (G.3) to hold is, due to (CP1.\
\[DoubleLongLeftArrow])\n(G.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCPMultipliers", "[", 
    RowBox[{"c1", ",", "c2", ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", 
     RowBox[{"lc", "[", "q", "]"}], ",", "c", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nwhich is precisely the content of (9).\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"p", "\[NotEqual]", "q"}], "\[And]", 
     RowBox[{"lc", "[", "p", "]"}]}], "=", 
    RowBox[{"lc", "[", "q", "]"}]}], TraditionalForm]]],
 ".\nFrom (4), by (isMNTCRP alternative) and our case assumption, we can \
infer\n(30)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", "c"}], "]"}], TraditionalForm]]],
 "\nWe know\n(30\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"cpm", "[", 
     RowBox[{"p", ",", "q", ",", 
      SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}], "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"r", "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ",", 
      RowBox[{"r", "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "q", "]"}]}], ")"}]}]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 "\n\nWe claim that the two constituents of the tuple in \
(30\[CloseCurlyQuote]) witness our existential goal, hence we show all of\n\
(T.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"r", "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"r", "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "q", "]"}]}], ")"}]}], "]"}]}]}], 
   TraditionalForm]]],
 "\n(T.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"cpm", "[", 
     RowBox[{"p", ",", "q", ",", 
      SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}], "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"r", "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ",", 
      RowBox[{"r", "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"lcm", "[", 
          RowBox[{
           RowBox[{"lp", "[", "p", "]"}], ",", 
           RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
         RowBox[{"lp", "[", "q", "]"}]}], ")"}]}]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 "\n(T.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCPMultipliers", "[", 
    RowBox[{
     RowBox[{"r", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"lcm", "[", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], ",", 
          RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
        RowBox[{"lp", "[", "p", "]"}]}], ")"}]}], ",", 
     RowBox[{"r", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"lcm", "[", 
         RowBox[{
          RowBox[{"lp", "[", "p", "]"}], ",", 
          RowBox[{"lp", "[", "q", "]"}]}], "]"}], "/", 
        RowBox[{"lp", "[", "q", "]"}]}], ")"}]}], ",", "p", ",", "q", ",", 
     SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\n(T.1) is obvious and (T.2) is identical to (30\[CloseCurlyQuote]), hence \
we only need to show (T.3).\n\nBy (CP2.\[DoubleLongLeftArrow]), (T.3) can be \
reduced to\n(T.6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     RowBox[{"lc", "[", "p", "]"}], ",", "c"}], "]"}], TraditionalForm]]],
 "\n(T.7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"c", ",", 
     RowBox[{"lc", "[", "p", "]"}], ",", "r"}], "]"}], TraditionalForm]]],
 "\n(T.6) is identical to (30), and by (18) in \[ScriptCapitalP] (T.7) \
reduces to\n(T.8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"c", ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 "\nFrom (30), by the definition of isMNTCR and isNTCR, we can infer (among \
others)\n(33)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"c", ",", 
     RowBox[{"lc", "[", "p", "]"}]}], "]"}], TraditionalForm]]],
 "\nhence we are done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.643342559163741*^9, 3.643342749721807*^9}, {
   3.643342926692049*^9, 3.6433430445391617`*^9}, {3.6433430874233418`*^9, 
   3.6433432259993267`*^9}, {3.64334327257839*^9, 3.643343474246717*^9}, {
   3.643343529082202*^9, 3.64334369284041*^9}, {3.643343745427994*^9, 
   3.6433437470732203`*^9}, {3.643343792424715*^9, 3.6433439165239143`*^9}, {
   3.6433439706436377`*^9, 3.643344393201757*^9}, {3.643344438863832*^9, 
   3.64334450966425*^9}, {3.6433445456767607`*^9, 3.643344648107934*^9}, {
   3.643345664803955*^9, 3.643345664948534*^9}, {3.643345701950053*^9, 
   3.643345709140546*^9}, {3.643345747430725*^9, 3.643345832967987*^9}, {
   3.6433473955583572`*^9, 3.6433475913658752`*^9}, {3.643347622887203*^9, 
   3.6433476232112627`*^9}, {3.643347780625799*^9, 3.6433481148897448`*^9}, {
   3.6433502102412653`*^9, 3.6433503197273493`*^9}, {3.643354196589072*^9, 
   3.643354239278658*^9}, {3.643354287624166*^9, 3.643354369925997*^9}, {
   3.643354405617593*^9, 3.643354511228306*^9}, {3.6433545547337427`*^9, 
   3.64335496886874*^9}, {3.643355795107526*^9, 3.643355816910918*^9}, {
   3.6433558888916597`*^9, 3.643355900458609*^9}, {3.643362573941572*^9, 
   3.643362575165277*^9}, {3.643430333169034*^9, 3.643430343545669*^9}, {
   3.643430383931912*^9, 3.643430546125051*^9}, {3.643430602128252*^9, 
   3.643430772425797*^9}, {3.643434309148082*^9, 3.643434355217005*^9}, {
   3.643434411134614*^9, 3.643434469325761*^9}, {3.643434528545582*^9, 
   3.643434548071657*^9}, {3.643434579063373*^9, 3.643434622323427*^9}, {
   3.643434658409073*^9, 3.643434730916888*^9}, {3.643434762720447*^9, 
   3.6434347912198*^9}, {3.6434348308230247`*^9, 3.643434836378359*^9}, {
   3.643434927304282*^9, 3.643435002810501*^9}, 3.643446821952528*^9, 
   3.6532078840703316`*^9, 3.6532134432879133`*^9, 3.6661737420298767`*^9, 
   3.666440348265574*^9},
 CellID->805176799]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R13 (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6467101954597178`*^9, 3.646710196054048*^9}, {
  3.6470817517625113`*^9, 3.647081756533407*^9}},
 CellID->286693183],

Cell["", "OpenEnvironment",
 CellID->827957578],

Cell[CellGroupData[{

Cell["Lemma: Properties of red in \[ScriptCapitalP]", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.646717701387841*^9, 3.646717707049193*^9}},
 CellID->865456817],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"f", ",", "p"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "f"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"f", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
      RowBox[{
       UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", 
        RowBox[{
         UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "f", "]"}]}], 
       "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible lm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:38279669", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6537125634047403`*^9, 3.6537126260076504`*^9}, {
  3.65371266569412*^9, 3.653712677955742*^9}},
 CellTags->{
  "ID:38279669", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducible lm"},
 CellID->38279669],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], 
    RowBox[{"[", "A", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "f", "]"}], 
       UnderscriptBox["|", 
        RowBox[{"f", "\[Element]", "A"}]], 
       RowBox[{"f", "\[NotEqual]", 
        UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "}"}], "]"}], 
    "\[Equal]", 
    RowBox[{
     UnderscriptBox["red", "\[ScriptCapitalP]"], "[", "A", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red lm set)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1376605467", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646717769966503*^9, 3.6467178248246107`*^9}},
 CellTags->{
  "ID:1376605467", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "red lm set"},
 CellID->1376605467],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"m", ",", "p"}], "]"}]}]], 
   RowBox[{
    UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", "]"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "m"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"p", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[Exists]", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
          RowBox[{"[", "t", "]"}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"\[And]", 
          RowBox[{"\[Piecewise]", GridBox[{
             {
              RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                RowBox[{
                 UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                 RowBox[{"p", ",", "t"}], "]"}], ",", 
                RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "m", "]"}]}],
                "]"}]},
             {
              RowBox[{
               RowBox[{
                UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "m", "]"}], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible monomial)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:673925715", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6470653041853533`*^9, 3.647065375565325*^9}, {
   3.64706542752367*^9, 3.647065490585825*^9}, {3.647065716999173*^9, 
   3.647065762911187*^9}, 3.647068714656266*^9, 3.647072855126575*^9, {
   3.647075031857777*^9, 3.647075050328759*^9}, {3.653653052641244*^9, 
   3.6536530844507*^9}, {3.6536532123799253`*^9, 3.6536532143299284`*^9}, {
   3.65371354209506*^9, 3.653713549520673*^9}},
 CellTags->{
  "ID:673925715", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducible monomial"},
 CellID->673925715],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->111884135]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->860415994],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible lm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible lm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:38279669"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p38279669-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p38279669-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:38279669", "Proof|ID:38279669-1"},
 CellID->51695314],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p38279669-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->582723933]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(red lm set)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(red lm set)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
             RowBox[{"{", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                   StyleBox["f", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["|", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                 RowBox[{
                   StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "}"}], "]"}],
            "\[Equal]", 
           RowBox[{
             UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
             StyleBox["A", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
             RowBox[{"{", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
                   StyleBox["f", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["|", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                 RowBox[{
                   StyleBox["f", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", "\[ScriptCapitalP]"]}]}], "}"}], "]"}],
            "\[Equal]", 
           RowBox[{
             UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
             StyleBox["A", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1376605467"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1376605467-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1376605467-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1376605467", "Proof|ID:1376605467-1"},
 CellID->580557624],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1376605467-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->66183403]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible monomial)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible monomial)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["m", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], "]"}], 
                    "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}]}]}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", "\[ScriptCapitalP]"], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["m", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], "]"}], 
                    "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    StyleBox["t", "ExpressionVariable"]}]}]}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:673925715"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p673925715-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p673925715-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:673925715", "Proof|ID:673925715-1"},
 CellID->54334308],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p673925715-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->694257822]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1784825111],

Cell[CellGroupData[{

Cell["Definition: isMonomialSequence", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.646710207523119*^9, 3.646710213321732*^9}, 
   3.646977539942152*^9},
 CellID->1154045365],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "f"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", "f", 
     "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainSequences", "[", "\[ScriptCapitalP]", "]"}]], 
       RowBox[{"[", "f", "]"}]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
        RowBox[{
         UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
         RowBox[{"f", "[", "i", "]"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomialSequence)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:47923252", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646710221768168*^9, 3.646710271057825*^9}, {
  3.646727191805215*^9, 3.646727196898814*^9}},
 CellTags->{
  "ID:47923252", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence"},
 CellID->47923252],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->316466243]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->923509955],

Cell["", "OpenEnvironment",
 CellID->1915683957],

Cell[CellGroupData[{

Cell["Lemma: Properties of Monomial Sequences", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.646727152238779*^9, 3.6467271581507196`*^9}},
 CellID->474285465],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", "f", 
   "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 1632720588}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646727171030978*^9, 3.6467271811129704`*^9}},
 CellTags->{
  "ID:1632720588", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1632720588],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"f", "[", "i", "]"}], "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomialSequence \[NotEqual] 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1771873303", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646710591267688*^9, 3.6467106080980253`*^9}},
 CellTags->{
  "ID:1771873303", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence \[NotEqual] 0"},
 CellID->1771873303],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{
     UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
     RowBox[{"f", "[", "i", "]"}], "]"}], "]"}]}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomialSequence lc 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:893563053", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646710346452248*^9, 3.646710370033885*^9}},
 CellTags->{
  "ID:893563053", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lc 1"},
 CellID->893563053],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", 
    RowBox[{
     UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
     RowBox[{"f", "[", "i", "]"}], "]"}], "]"}]}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomialSequence lp 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:587765202", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646710377185725*^9, 3.6467103824766083`*^9}},
 CellTags->{
  "ID:587765202", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lp 1"},
 CellID->587765202],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainSequences", "[", "\[ScriptCapitalR]", "]"}]], 
  RowBox[{"[", 
   RowBox[{"imgSeq", "[", 
    RowBox[{"f", ",", 
     UnderscriptBox["lc", "\[ScriptCapitalP]"]}], "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomialSequence lc 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1254733295", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646710304844677*^9, 3.646710325371641*^9}},
 CellTags->{
  "ID:1254733295", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lc 2"},
 CellID->1254733295],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainSequences", "[", "\[ScriptCapitalT]", "]"}]], 
  RowBox[{"[", 
   RowBox[{"imgSeq", "[", 
    RowBox[{"f", ",", 
     UnderscriptBox["lp", "\[ScriptCapitalP]"]}], "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomialSequence lp 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:280696418", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646710331563048*^9, 3.646710339793271*^9}},
 CellTags->{
  "ID:280696418", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lp 2"},
 CellID->280696418],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isNatMapping", "[", "g", "]"}]], 
  RowBox[{
   UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
   RowBox[{"subSeq", "[", 
    RowBox[{"f", ",", "g"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomialSequence subSeq)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1081126646", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64697737528686*^9, 3.646977392960568*^9}},
 CellTags->{
  "ID:1081126646", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence subSeq"},
 CellID->1081126646],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", "h"], 
   RowBox[{"h", "\[LeftTriangleBar]", "f"}]], 
  RowBox[{
   UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", "h", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMonomialSequence subsequence)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1101697023", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6469973439377413`*^9, 3.646997357311089*^9}},
 CellTags->{
  "ID:1101697023", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence subsequence"},
 CellID->1101697023],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->90094680]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1586112789],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomialSequence \[NotEqual] 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomialSequence \[NotEqual] 0)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["f", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
             StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalP]"]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["f", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
             StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}], "\[NotEqual]", 
               UnderscriptBox[
               "0", "\[ScriptCapitalP]"]}]}]}]}], $CellContext`TheoremaForm]],
      "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1771873303"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1771873303-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1771873303-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1771873303", "Proof|ID:1771873303-1"},
 CellID->103075955],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1771873303-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1903828265]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomialSequence lc 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomialSequence lc 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["f", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
             StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["f", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["f", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
             StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                 UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["f", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:893563053"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p893563053-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p893563053-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:893563053", "Proof|ID:893563053-1"},
 CellID->88855734],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p893563053-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1505089946]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomialSequence lp 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomialSequence lp 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["f", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
             StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["f", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["f", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
             StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["f", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:587765202"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p587765202-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p587765202-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:587765202", "Proof|ID:587765202-1"},
 CellID->1976154064],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p587765202-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1938386534]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomialSequence lc 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomialSequence lc 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["f", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
             StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainSequences", "[", "\[ScriptCapitalR]", "]"}]], 
             "[", 
             RowBox[{"imgSeq", "[", 
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], ",", 
                 UnderscriptBox["lc", "\[ScriptCapitalP]"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["f", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
             StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainSequences", "[", "\[ScriptCapitalR]", "]"}]], 
             "[", 
             RowBox[{"imgSeq", "[", 
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], ",", 
                 UnderscriptBox["lc", "\[ScriptCapitalP]"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1254733295"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1254733295-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1254733295-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1254733295", "Proof|ID:1254733295-1"},
 CellID->1511544165],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1254733295-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->738405199]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomialSequence lp 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomialSequence lp 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["f", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
             StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainSequences", "[", "\[ScriptCapitalT]", "]"}]], 
             "[", 
             RowBox[{"imgSeq", "[", 
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], ",", 
                 UnderscriptBox["lp", "\[ScriptCapitalP]"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["f", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
             StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainSequences", "[", "\[ScriptCapitalT]", "]"}]], 
             "[", 
             RowBox[{"imgSeq", "[", 
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], ",", 
                 UnderscriptBox["lp", "\[ScriptCapitalP]"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:280696418"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p280696418-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p280696418-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:280696418", "Proof|ID:280696418-1"},
 CellID->276631526],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p280696418-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->273841733]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomialSequence subSeq)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomialSequence subSeq)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isNatMapping", "[", 
               StyleBox["g", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
           RowBox[{"subSeq", "[", 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["g", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{"isNatMapping", "[", 
               StyleBox["g", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
           RowBox[{"subSeq", "[", 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["g", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1081126646"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1081126646-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1081126646-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1081126646", "Proof|ID:1081126646-1"},
 CellID->227638759],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1081126646-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2067143528]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMonomialSequence subsequence)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMonomialSequence subsequence)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             StyleBox["h", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["h", "ExpressionVariable"], "\[LeftTriangleBar]", 
             StyleBox["f", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
             StyleBox["h", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             StyleBox["h", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["h", "ExpressionVariable"], "\[LeftTriangleBar]", 
             StyleBox["f", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
             StyleBox["h", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1101697023"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1101697023-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1101697023-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1101697023", "Proof|ID:1101697023-1"},
 CellID->881838494],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1101697023-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1551672538]
}, Closed]],

Cell[TextData[{
 "We prove (R13) in \[ScriptCapitalP] by contradiction: We assume that there \
exists and infinite sequence ",
 Cell[BoxData[
  FormBox["B", TraditionalForm]]],
 " of subsets of \[ScriptCapitalP] with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"red", "[", 
     SubscriptBox["B", "i"], "]"}], "\[Subset]", 
    RowBox[{"red", "[", 
     SubscriptBox["B", 
      RowBox[{"i", "+", "1"}]], "]"}]}], TraditionalForm]]],
 " and show that then there must also exist an infinite sequence ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " of power-products with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["t", "i"], "\[NotVerticalBar]", 
    SubscriptBox["t", "j"]}], TraditionalForm]]],
 " for all ",
 Cell[BoxData[
  FormBox[
   RowBox[{"j", ">", "i"}], TraditionalForm]]],
 ". This, however, is not possible if Dickson\[CloseCurlyQuote]s Lemma holds \
in \[ScriptCapitalT]."
}], "Text",
 CellChangeTimes->{{3.646711225679509*^9, 3.646711351214469*^9}},
 CellID->989285657],

Cell["", "OpenEnvironment",
 CellID->1565804923],

Cell[CellGroupData[{

Cell["Lemma: Existence of Monomial Sequence", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6467113752379637`*^9, 3.646711380843667*^9}},
 CellID->1515610326],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainSequences", "[", 
        RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}], "]"}]], 
      RowBox[{"[", "B", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
     RowBox[{
      RowBox[{
       UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
       RowBox[{"B", "[", "i", "]"}], "]"}], "\[Subset]", 
      RowBox[{
       UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
       RowBox[{"B", "[", 
        RowBox[{"i", "+", "1"}], "]"}], "]"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainSequences", "[", 
        RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}], "]"}]], 
      RowBox[{"[", "M", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
           RowBox[{
            RowBox[{
             UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
             RowBox[{"M", "[", "i", "]"}], "]"}], "\[Subset]", 
            RowBox[{
             UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
             RowBox[{"M", "[", 
              RowBox[{"i", "+", "1"}], "]"}], "]"}]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{"m", "\[Element]", 
              RowBox[{"M", "[", "i", "]"}]}]], 
            RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", 
             "]"}]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(110)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:182970451", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646719805249271*^9, 3.646719865372345*^9}},
 CellTags->{
  "ID:182970451", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "110"},
 CellID->182970451],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainSequences", "[", 
        RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}], "]"}]], 
      RowBox[{"[", "B", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
           RowBox[{
            RowBox[{
             UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
             RowBox[{"B", "[", "i", "]"}], "]"}], "\[Subset]", 
            RowBox[{
             UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
             RowBox[{"B", "[", 
              RowBox[{"i", "+", "1"}], "]"}], "]"}]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{"m", "\[Element]", 
              RowBox[{"B", "[", "i", "]"}]}]], 
            RowBox[{
             UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", 
             "]"}]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", "m", 
      "]"}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
     RowBox[{
      RowBox[{
       UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
       RowBox[{
        RowBox[{"setSeq", "[", "m", "]"}], "[", "i", "]"}], "]"}], 
      "\[Subset]", 
      RowBox[{
       UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
       RowBox[{
        RowBox[{"setSeq", "[", "m", "]"}], "[", 
        RowBox[{"i", "+", "1"}], "]"}], "]"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(111)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:845195690", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646711389771057*^9, 3.6467114948526077`*^9}, {
  3.646711598707234*^9, 3.646711614472542*^9}, {3.646714386940599*^9, 
  3.646714392450115*^9}, {3.646719902211235*^9, 3.64671990995691*^9}},
 CellTags->{
  "ID:845195690", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "111"},
 CellID->845195690],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1298865959]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1599916594],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(110)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(110)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainSequences", "[", 
                    RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}], 
                    "]"}]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}]}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainSequences", "[", 
                    RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}], 
                    "]"}]], "[", 
                 StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["M", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                 "\[Subset]", 
                 RowBox[{
                   UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["M", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                   "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "\[Element]", 
                    RowBox[{
                    StyleBox["M", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainSequences", "[", 
                    RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}], 
                    "]"}]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}]}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainSequences", "[", 
                    RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}], 
                    "]"}]], "[", 
                 StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["M", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                 "\[Subset]", 
                 RowBox[{
                   UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["M", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                   "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "\[Element]", 
                    RowBox[{
                    StyleBox["M", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:182970451"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p182970451-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p182970451-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:182970451", "Proof|ID:182970451-1"},
 CellID->568281335],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p182970451-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->758050075]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(111)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(111)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainSequences", "[", 
                    RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}], 
                    "]"}]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                 "\[Subset]", 
                 RowBox[{
                   UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                   "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "\[Element]", 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}]}]}]}], 
         "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], 
                 "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainSequences", "[", 
                    RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}], 
                    "]"}]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                 "\[Subset]", 
                 RowBox[{
                   UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                   "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "\[Element]", 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}]}]}]}], 
         "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], 
                 "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:845195690"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p845195690-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p845195690-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:845195690", "Proof|ID:845195690-1"},
 CellID->70458024],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p845195690-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->31491083]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1592718460],

Cell[CellGroupData[{

Cell["\<\
Lemma: Existence of Monomial Sequence with Certain Property Violates R13 in \
\[ScriptCapitalR]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.646972667267951*^9, 3.646972701990408*^9}, {
  3.647081840747875*^9, 3.6470818415579653`*^9}},
 CellID->1777300951],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", "m", 
   "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\Polynomials.nb", 2081362811}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646977125071357*^9, 3.6469771315168877`*^9}},
 CellTags->{
  "ID:2081362811", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->2081362811],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
         RowBox[{
          RowBox[{
           UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
           RowBox[{
            RowBox[{"setSeq", "[", "m", "]"}], "[", "i", "]"}], "]"}], 
          "\[Subset]", 
          RowBox[{
           UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
           RowBox[{
            RowBox[{"setSeq", "[", "m", "]"}], "[", 
            RowBox[{"i", "+", "1"}], "]"}], "]"}]}]}]},
       {
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{"a", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
         RowBox[{
          UnderscriptBox[
           UnderscriptBox["\[ForAll]", 
            RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
           RowBox[{"i", "\[GreaterEqual]", "a"}]], 
          RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[Exists]", 
             RowBox[{"j", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
            RowBox[{"j", ">", "i"}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
              RowBox[{"m", "[", "i", "]"}], "]"}], 
             UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
             RowBox[{
              UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
              RowBox[{"m", "[", "j", "]"}], "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", "f", 
      "]"}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            RowBox[{
             UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
             RowBox[{
              RowBox[{"setSeq", "[", "f", "]"}], "[", "i", "]"}], "]"}], 
            "\[Subset]", 
            RowBox[{
             UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
             RowBox[{
              RowBox[{"setSeq", "[", "f", "]"}], "[", 
              RowBox[{"i", "+", "1"}], "]"}], "]"}]}]},
          {
           RowBox[{
            RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
             RowBox[{"f", "[", "i", "]"}], "]"}], 
            UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
            RowBox[{
             UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
             RowBox[{"f", "[", 
              RowBox[{"i", "+", "1"}], "]"}], "]"}]}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(112)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1173235103", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646973018657193*^9, 3.646973055688861*^9}, {
  3.6469771357412567`*^9, 3.646977160120635*^9}},
 CellTags->{
  "ID:1173235103", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "112"},
 CellID->1173235103],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
       RowBox[{"m", "[", "i", "]"}], "]"}], 
      UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
       RowBox[{"m", "[", 
        RowBox[{"i", "+", "1"}], "]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "\[Element]", "\[DoubleStruckCapitalN]"}]}]], 
     RowBox[{"j", "\[GreaterEqual]", "i"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
       RowBox[{"m", "[", "i", "]"}], "]"}], 
      UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
      RowBox[{
       UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
       RowBox[{"m", "[", 
        RowBox[{"j", "+", "1"}], "]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(113)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1805755328", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6469769748340883`*^9, 3.646976995341858*^9}, {
  3.646977163044538*^9, 3.646977176056962*^9}, {3.6469975858194733`*^9, 
  3.646997594802018*^9}, {3.647062042694107*^9, 3.647062066792554*^9}, {
  3.6470623536707993`*^9, 3.647062362272785*^9}},
 CellTags->{
  "ID:1805755328", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "113"},
 CellID->1805755328],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
            RowBox[{
             RowBox[{"setSeq", "[", "m", "]"}], "[", "i", "]"}], "]"}], 
           "\[Subset]", 
           RowBox[{
            UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
            RowBox[{
             RowBox[{"setSeq", "[", "m", "]"}], "[", 
             RowBox[{"i", "+", "1"}], "]"}], "]"}]}]},
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
            RowBox[{"m", "[", "i", "]"}], "]"}], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
           RowBox[{
            UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
            RowBox[{"m", "[", 
             RowBox[{"i", "+", "1"}], "]"}], "]"}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(114)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:596327945", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6469727085533133`*^9, 3.646972728375555*^9}, {
   3.646972764401486*^9, 3.6469729227496767`*^9}, 3.646973208922544*^9, 
   3.646977004805628*^9, {3.646977048246543*^9, 3.646977060636712*^9}, {
   3.646977178463499*^9, 3.646977181091733*^9}, {3.647062033623308*^9, 
   3.6470620385769873`*^9}, {3.6470628209726677`*^9, 3.64706282432166*^9}, {
   3.6470632411889057`*^9, 3.647063248271294*^9}, {3.647063301730257*^9, 
   3.647063301927663*^9}},
 CellTags->{
  "ID:596327945", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "114"},
 CellID->596327945],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1479320113]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->3101935],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(112)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(112)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                 "\[Subset]", 
                 RowBox[{
                   UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                   "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["a", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], ">", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "[", 
                    StyleBox["j", "ExpressionVariable"], "]"}], 
                    "]"}]}]}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], 
                   "[", 
                   StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                   "\[Subset]", 
                   RowBox[{
                    UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                    "]"}]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                 "\[Subset]", 
                 RowBox[{
                   UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                   RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                   "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["a", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], ">", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                    RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "[", 
                    StyleBox["j", "ExpressionVariable"], "]"}], 
                    "]"}]}]}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], 
                   "[", 
                   StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                   "\[Subset]", 
                   RowBox[{
                    UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                    "]"}]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1173235103"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1173235103-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1173235103-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1173235103", "Proof|ID:1173235103-1"},
 CellID->874379054],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1173235103-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->423369729]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(113)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(113)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
               RowBox[{
                 UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1805755328"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1805755328-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1805755328-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1805755328", "Proof|ID:1805755328-1"},
 CellID->1937389710],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1805755328-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.653649940969479*^9},
 CellID->1056270318]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(114)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(114)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                   "\[Subset]", 
                   RowBox[{
                    UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                    "]"}]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                    "]"}]}]}]}], 
             TagBox[")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["isMonomialSequence", "\[ScriptCapitalP]"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                   "\[Subset]", 
                   RowBox[{
                    UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                    "]"}]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   RowBox[{
                    UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                    "]"}]}]}]}], 
             TagBox[
             ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:596327945"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p596327945-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p596327945-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:596327945", "Proof|ID:596327945-1"},
 CellID->387192426],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p596327945-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->8942193]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->585699768],

Cell[CellGroupData[{

Cell["\<\
Theorem: Existence of Sequence Violating (R13) in \[ScriptCapitalP] Implies \
Existence of Sequence Violating Dickson\[CloseCurlyQuote]s Property in \
\[ScriptCapitalT]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.647079070921013*^9, 3.647079111631885*^9}, {
  3.647079370540557*^9, 3.6470793724275208`*^9}},
 CellID->1237178328],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainSequences", "[", 
        RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}], "]"}]], 
      RowBox[{"[", "B", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
     RowBox[{
      RowBox[{
       UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
       RowBox[{"B", "[", "i", "]"}], "]"}], "\[Subset]", 
      RowBox[{
       UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
       RowBox[{"B", "[", 
        RowBox[{"i", "+", "1"}], "]"}], "]"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainSequences", "[", "\[ScriptCapitalT]", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "\[Element]", "\[DoubleStruckCapitalN]"}]}]], 
      RowBox[{"j", ">", "i"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[Not]", 
       RowBox[{
        RowBox[{"T", "[", "i", "]"}], 
        UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
        RowBox[{"T", "[", "j", "]"}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(115)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1402642391", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647079140055191*^9, 3.647079208694829*^9}},
 CellTags->{
  "ID:1402642391", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "115"},
 CellID->1402642391],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2031723415]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->362476080],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(115)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(115)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainSequences", "[", 
                    RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}], 
                    "]"}]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}]}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainSequences", "[", "\[ScriptCapitalT]", "]"}]],
                  "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], ">", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   RowBox[{
                    StyleBox["T", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   RowBox[{
                    StyleBox["T", "ExpressionVariable"], "[", 
                    StyleBox["j", "ExpressionVariable"], "]"}]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainSequences", "[", 
                    RowBox[{"DomainSets", "[", "\[ScriptCapitalP]", "]"}], 
                    "]"}]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", "\[ScriptCapitalP]"], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}]}], "\[Implies]", 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainSequences", "[", "\[ScriptCapitalT]", "]"}]],
                  "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], ">", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   RowBox[{
                    StyleBox["T", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   RowBox[{
                    StyleBox["T", "ExpressionVariable"], "[", 
                    StyleBox["j", "ExpressionVariable"], "]"}]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1402642391"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1402642391-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1402642391-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1402642391", "Proof|ID:1402642391-1"},
 CellID->1177596546],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1402642391-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1549025416]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1409591193],

Cell[CellGroupData[{

Cell["\<\
Presupposition: Dickson-Domain \[ScriptCapitalT]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.647080427266027*^9, 3.647080435532064*^9}},
 CellID->49556011],

Cell[BoxData[
 RowBox[{"isDicksonDomain", "[", 
  RowBox[{"\[ScriptCapitalT]", ",", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}]}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Dickson-domain \[ScriptCapitalT])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:980978647", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6470804375581007`*^9, 3.6470804436456223`*^9}, {
  3.647081427718508*^9, 3.64708143599686*^9}},
 CellTags->{
  "ID:980978647", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Dickson-domain \[ScriptCapitalT]"},
 CellID->980978647],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->148163230]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1739788415],

Cell["", "OpenEnvironment",
 CellID->506738286],

Cell[CellGroupData[{

Cell["Corollary: \[ScriptCapitalP] Satisfies R13", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6470814633937263`*^9, 3.647081491200801*^9}, {
  3.647081847405882*^9, 3.647081848032548*^9}},
 CellID->1566663652],

Cell[BoxData[
 RowBox[{"R13", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(116)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1408052308", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647081493369236*^9, 3.64708149660846*^9}},
 CellTags->{
  "ID:1408052308", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "116"},
 CellID->1408052308],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1643742643]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1724439218],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(116)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(116)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "R13", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "R13", "[", "\[ScriptCapitalP]", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1408052308"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1408052308-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1408052308-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1408052308", "Proof|ID:1408052308-1"},
 CellID->245205539],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1408052308-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->127378030]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Result (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6363571391771297`*^9, 3.636357141185442*^9}, {
  3.6434440408929996`*^9, 3.643444044867598*^9}},
 CellID->2038082117],

Cell["", "OpenEnvironment",
 CellID->238202638],

Cell[CellGroupData[{

Cell["\<\
Corollary: \[ScriptCapitalP] Constitutes Algorithmic Reduction Ring\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6363571487318087`*^9, 3.636357159285831*^9}, {
  3.6434440338747997`*^9, 3.643444036163827*^9}},
 CellID->496041222],

Cell[BoxData[
 RowBox[{"isAlgoReductionRing", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isAlgoReductionRing \[ScriptCapitalP])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1183157947", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6363571641087112`*^9, 3.6363571668932343`*^9}},
 CellTags->{
  "ID:1183157947", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isAlgoReductionRing \[ScriptCapitalP]"},
 CellID->1183157947],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->364104560]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1493486891],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isAlgoReductionRing \[ScriptCapitalP])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isAlgoReductionRing \[ScriptCapitalP])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "isAlgoReductionRing", "[", "\[ScriptCapitalP]", 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "isAlgoReductionRing", "[", "\[ScriptCapitalP]", 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1183157947"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1183157947-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1183157947-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1183157947", "Proof|ID:1183157947-1"},
 CellID->1299983476],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1183157947-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1799732883]
}, Closed]],

Cell[TextData[{
 "We now generalize the results proved for the arbitrary but fixed domains \
\[ScriptCapitalP], \[ScriptCapitalR] and \[ScriptCapitalT] to ",
 StyleBox["all",
  FontSlant->"Italic"],
 " domains, or, more precisely, functors \[ScriptCapitalF] and domains \
\[ScriptCapitalR]0 and \[ScriptCapitalT]0. Only note that \[ScriptCapitalF] \
still has to belong to the category of algorithmic reduction polynomials when \
applied to \[ScriptCapitalR] and \[ScriptCapitalT], not to \[ScriptCapitalR]0 \
and \[ScriptCapitalT]0!\nOne can easily verify that the conditions imposed on \
\[ScriptCapitalF], \[ScriptCapitalR]0 and \[ScriptCapitalT]0 in \
(isAlgoReductionRing Conservation Theorem Polynomials) are sufficient, by \
comparing them to the list of basic non-definition assumptions used for \
proving (isAlgoReductionRing \[ScriptCapitalP]) that conatain \
\[ScriptCapitalP], \[ScriptCapitalT] or \[ScriptCapitalR]:"
}], "Text",
 CellChangeTimes->{{3.647158339508985*^9, 3.647158467576324*^9}, {
  3.6532174394011354`*^9, 3.653217439510336*^9}, {3.6539033145749235`*^9, 
  3.6539034443067513`*^9}},
 CellID->72919207],

Cell[BoxData[
 RowBox[{"DisplayFormula", "[", 
  RowBox[{
   RowBox[{"DeleteDuplicates", "[", 
    RowBox[{"Select", "[", 
     RowBox[{
      RowBox[{"Select", "[", 
       RowBox[{
        RowBox[{"Block", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"$FormulaOutput", "=", "\"\<Formula\>\""}], ",", 
            RowBox[{"$MakeLink", "=", "False"}]}], "}"}], ",", 
          RowBox[{"CollectAllFormulas", "[", 
           RowBox[{
            RowBox[{
            "DisplayFormula", "[", 
             "\"\<isAlgoReductionRing \[ScriptCapitalP]\>\"", "]"}], ",", 
            RowBox[{"\"\<Direction\>\"", "\[Rule]", "Backward"}], ",", 
            RowBox[{"\"\<Filter\>\"", "\[Rule]", "Last"}], ",", 
            RowBox[{"\"\<CarefulSearch\>\"", "\[Rule]", "True"}]}], "]"}]}], 
         "]"}], ",", 
        RowBox[{
         RowBox[{"ContainsConstant", "[", 
          RowBox[{"#", ",", 
           RowBox[{"{", 
            RowBox[{
            "Theorema`Knowledge`\[ScriptCapitalR]$TM", ",", 
             "Theorema`Knowledge`\[ScriptCapitalT]$TM", ",", 
             "Theorema`Language`\[ScriptCapitalP]$TM"}], "}"}]}], "]"}], 
         "&"}]}], "]"}], ",", 
      RowBox[{
       RowBox[{"FreeQ", "[", 
        RowBox[{"#", ",", 
         RowBox[{
         "Theorema`Language`EqualDef$TM", "|", 
          "Theorema`Language`IffDef$TM"}]}], "]"}], "&"}]}], "]"}], "]"}], 
   ",", 
   RowBox[{"\"\<FormulaOutput\>\"", "\[Rule]", "\"\<DisplayFormula\>\""}], 
   ",", 
   RowBox[{"\"\<MakeLink\>\"", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.653901809613279*^9, 3.6539019948490043`*^9}, 
   3.6539023188509736`*^9, {3.653902449534403*^9, 3.6539024500804043`*^9}, {
   3.653903165826662*^9, 3.653903202705127*^9}, 3.6539032529382153`*^9, {
   3.653903535785312*^9, 3.6539035683737698`*^9}, {3.653903617626056*^9, 
   3.653903634754886*^9}, {3.6539036802631664`*^9, 3.653903689233182*^9}, {
   3.6539042697192087`*^9, 3.653904272511614*^9}, 3.653904309421279*^9, {
   3.653904906557954*^9, 3.653904907727956*^9}, {3.6539050654132357`*^9, 
   3.6539050659280367`*^9}},
 CellID->191565054],

Cell[TextData[{
 "The only formulas in the list ",
 StyleBox["not",
  FontSlant->"Italic"],
 " included in (isAlgoReductionRing Conservation Theorem Polynomials), like \
(containsMonomials), follow by definition from others are are included."
}], "Text",
 CellChangeTimes->{{3.6539034545715694`*^9, 3.653903509374466*^9}},
 CellID->149433800],

Cell["", "OpenEnvironment",
 CellID->1569900360],

Cell[CellGroupData[{

Cell["Theorem", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.647158106160778*^9},
 CellID->1957927293],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
   "\[ScriptCapitalF]", ",", "\[ScriptCapitalR]0", ",", 
    "\[ScriptCapitalT]0"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"isAlgoReductionPolynomialDomain", "[", 
       RowBox[{"\[ScriptCapitalF]", "[", 
        RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], "]"}],
       "\[And]", 
      RowBox[{"isAlgoReductionRing", "[", "\[ScriptCapitalR]0", "]"}], 
      "\[And]", 
      RowBox[{"isCommPPDomain", "[", "\[ScriptCapitalT]0", "]"}], "\[And]", 
      RowBox[{"isDicksonDomain", "[", 
       RowBox[{"\[ScriptCapitalT]0", ",", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         TagBox[
          UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]0"],
          Identity], 
         TagBox[")",
          "AutoParentheses"]}]}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{"isAlgoReductionRing", "[", 
     RowBox[{"\[ScriptCapitalF]", "[", 
      RowBox[{"\[ScriptCapitalR]0", ",", "\[ScriptCapitalT]0"}], "]"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(isAlgoReductionRing Conservation Theorem Polynomials)", 
         "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1794188423", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6471581108075*^9, 3.6471583016387997`*^9}, {
  3.6539032266979694`*^9, 3.653903243094598*^9}},
 CellTags->{
  "ID:1794188423", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", 
   "isAlgoReductionRing Conservation Theorem Polynomials"},
 CellID->1794188423],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->331601271]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->820459176]
}, Closed]]
}, Closed]]
}, Open  ]]
},
WindowSize->{1111, 763},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
ShowSelection->True,
Magnification->1.,
FrontEndVersion->"9.0 for Linux x86 (64-bit) (February 7, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Theorema"}, "TheoremaCore.nb"]]], 
   
   Cell["Modify the definitions below to change the default appearance of all \
cells in a given style. \nMake modifications to any definition using commands \
in the Format menu.", "Text"], 
   Cell[
    CellGroupData[{
      Cell["Notebook Options Settings", "Section"], 
      Cell[
      "The options defined for the style below will be used at the Notebook \
level.", "Text"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Notebook"], InputAliases -> {"and" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[And]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "or" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", "\[Or]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "not" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Not]", "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "impl" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Implies]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[DoubleLongLeftRightArrow]", 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Equal]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox[
                  
                  RowBox[{
                   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                  "a:=b"], "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", ":=", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "andn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[And]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "orn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Or]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "equivn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[DoubleLeftRightArrow]", "\[Piecewise]", 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "cdist" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Piecewise]", 
                 
                 GridBox[{{
                   "\[SelectionPlaceholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}, {
                   "\[Placeholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "far" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exr" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "farc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exrc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "let" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["let", 
                  RowBox[{"\[Placeholder]", "=", "\[Placeholder]"}]], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "app" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293a", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "prep" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293b", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "join" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22c8", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "elemT" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22ff", Identity, SyntaxForm -> "a+b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "subop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "supop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SuperscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "subsupop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubsuperscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "oop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "uoop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderoverscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "osop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox[
                OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
                "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "op" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               TagBox["\[SelectionPlaceholder]", Identity], 
               TagBox[")", "AutoParentheses"]}], "[]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e114", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e115", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "()" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:fd3e", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:fd3f", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e8", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e9", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e117", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e118", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "[[]]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e6", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e7", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(())" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2e28", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2e29", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<<>>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27ea", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27eb", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{{}}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2983", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2984", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(" -> 
            TagBox["(", "AutoParentheses"], ")" -> 
            TagBox[")", "AutoParentheses"]}, Background -> 
          RGBColor[0.957, 0.96, 0.97]], 
         Cell[
          StyleData["Notebook", "Printout"], Background -> None]}, Open]]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell["Style Environment Names", "Section"], 
      Cell[
       StyleData[All, "Working"], DockedCells -> {
         Cell[
         "Theorema 2.0", "SBO", FontTracking -> 8, 
          CellMargins -> {{0, 0}, {0, 0}}, 
          CellFrameMargins -> {{0, 50}, {0, 0}}, FontSize -> 28, FontSlant -> 
          "Plain", FontColor -> RGBColor[0.29, 0.46, 0.6], TextAlignment -> 
          Right, Background -> RGBColor[0.36, 0.54, 0.69]]}], 
      Cell[
       StyleData[
       All, "Layout", StyleDefinitions -> StyleData[All, "Working"]], 
       MenuSortingValue -> 1200]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Styles for Title and Section Cells", "Section"], 
      Cell[
       CellGroupData[{
         Cell["Title", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], CellMargins -> {{27, 27}, {10, 30}}, 
             TextAlignment -> Center, MenuSortingValue -> 1100, 
             MenuCommandKey -> "1", FontFamily -> "Bitstream Charter", 
             FontSize -> 36, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Title", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{6, 10}, {5, 30}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], CellMargins -> {{27, 27}, {20, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1150, FontFamily -> 
             "Bitstream Charter", FontSize -> 24, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{6, 10}, {5, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], CellMargins -> {{27, 27}, {8, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], CellMargins -> {{27, 27}, {18, 20}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Sectioning", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], ShowGroupOpener -> True, MenuSortingValue -> 
             1250, MenuCommandKey -> "4", FontFamily -> "Bitstream Charter", 
             FontSize -> 20, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Section", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{2, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {8, 12}}, 
             MenuSortingValue -> 1300, MenuCommandKey -> "5", FontFamily -> 
             "Bitstream Charter", FontSize -> 14, FontWeight -> "Bold", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Subsection", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{9, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {2, 10}}, 
             MenuSortingValue -> 1350, MenuCommandKey -> "6", FontFamily -> 
             "Bitstream Charter", FontSize -> 12, FontWeight -> "Bold", 
             FontSlant -> "Italic", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsection", "Printout"], ShowGroupOpener -> False,
              CellMargins -> {{35, 0}, {7, 14}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsubsection"], CellDingbat -> None, 
             ShowGroupOpener -> True, 
             CellMargins -> {{27, Inherited}, {Inherited, Inherited}}, 
             FontFamily -> "Bitstream Charter", FontSize -> 11, FontWeight -> 
             "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsubsection", "Printout"], 
             CellMargins -> {{50, 0}, {7, 14}}]}, Closed]]}, Open]]}, Open]], 
   
   Cell[
    CellGroupData[{
      Cell["Styles for Publication\[Hyphen]ready Documents", "Section"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Author"], CellMargins -> {{27, 30}, {20, 5}}, 
          MenuSortingValue -> 1900, FontFamily -> "Helvetica", FontSize -> 14,
           FontSlant -> "Italic"], 
         Cell[
          StyleData["Author", "Printout"], CellMargins -> {{6, 10}, {0, 10}}, 
          FontSize -> 10]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Abstract"], CellFrame -> {{0, 0}, {0.5, 0.5}}, 
          CellMargins -> {{27, 140}, {20, 20}}, LineSpacing -> {0.9, 0}, 
          MenuSortingValue -> 2050, FontFamily -> "Times", FontSize -> 12], 
         Cell[
          StyleData["Abstract", "Printout"], Hyphenation -> True, 
          LineSpacing -> {1, 2}, FontSize -> 10]}, Closed]]}, Closed]], 
   Cell[
    CellGroupData[{
      Cell["Text", "Section"], 
      Cell[
       StyleData["Text"], CellMargins -> {{27, 27}, {Inherited, Inherited}}, 
       TextJustification -> 1., ParagraphIndent -> 15, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["EnvironmentText"], ShowCellBracket -> Automatic, 
       CellMargins -> {{82, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextJustification -> 1., ParagraphSpacing -> {0.3, 0.}, 
       MenuSortingValue -> 1480, MenuCommandKey -> "T", FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Item"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "Item", 
          CounterAssignments -> {{"Subitem", 0}, {"Subsubitem", 0}}, 
          MenuSortingValue -> 1600, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Item", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemParagraph"], CellMargins -> {{81, 10}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          DefaultNewCellStyle -> "Item", MenuSortingValue -> 1600, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemParagraph", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "Subitem", 
          CounterAssignments -> {{"Subsubitem", 0}}, MenuSortingValue -> 1610,
           FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subitem", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemParagraph"], CellMargins -> {{105, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subitem", 
          MenuSortingValue -> 1610, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemParagraph", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subsubitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1620, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subsubitem", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemParagraph"], 
          CellMargins -> {{129, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1625, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemParagraph", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "ItemNumbered", 
          CounterAssignments -> {{"SubitemNumbered", 0}, {
            "SubsubitemNumbered", 0}}, MenuSortingValue -> 1630, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemNumbered", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "SubitemNumbered", 
          CounterAssignments -> {{"SubsubitemNumbered", 0}}, MenuSortingValue -> 
          1640, FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemNumbered", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], ".", 
              CounterBox["SubsubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> 
          "SubitemNumbered", CounterIncrements -> "SubsubitemNumbered", 
          MenuSortingValue -> 1645, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemNumbered", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Theorema", "Section"], 
      Cell[
       StyleData["GlobalDeclaration"], CellFrame -> {{0, 2}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 3}, {2, 2}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       InitializationCell -> True, CellFrameMargins -> 10, CellFrameColor -> 
       RGBColor[0.73, 0.65, 0.61], CellFrameLabelMargins -> {{33, 7}, {1, 1}},
        ShowAutoStyles -> False, MenuSortingValue -> 1441, MenuCommandKey -> 
       "G", FontSize -> 14], 
      Cell[
       StyleData["OpenEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {-4, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       StyleData["EnvironmentHeader"], 
       CellFrame -> {{False, False}, {False, False}}, 
       CellMargins -> {{76, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 6, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       ShowAutoStyles -> False, MenuSortingValue -> 1450, FontSize -> 14, 
       FontWeight -> "Bold", FontVariations -> {"CapsType" -> "SmallCaps"}, 
       Background -> RGBColor[0.89, 0.8, 0.69]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FormalTextInputFormula"], ShowCellBracket -> Automatic, 
          CellMargins -> {{96, 10}, {2, 2}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
          InitializationCell -> True, CellFrameLabelMargins -> 53, 
          ShowAutoStyles -> False, LimitsPositioningTokens -> {}, 
          MenuSortingValue -> 1470, MenuCommandKey -> "F", FontWeight -> 
          "Bold", FontColor -> RGBColor[0.25, 0.19, 0.2]], 
         Cell[
          StyleData["FormalTextInputFormula", "Layout"], Editable -> False], 
         Cell[
          StyleData["FormalTextInputFormula", "Presentation"], Editable -> 
          False]}, Open]], 
      Cell[
       StyleData["EndEnvironmentMarker"], 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextAlignment -> Right, MenuSortingValue -> 1480], 
      Cell[
       StyleData["CloseEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {False, True}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {Inherited, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Computation"], CellFrame -> {{1.5, 0}, {0, 0}}, 
          ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 7}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
          CellFrameMargins -> {{8, 8}, {3, 4}}, CellFrameColor -> 
          RGBColor[0.486, 0.33, 0.32], ShowAutoStyles -> False, 
          MenuSortingValue -> 1510, FontWeight -> "Bold", FontColor -> 
          RGBColor[0.23, 0.25, 0.34], Background -> 
          RGBColor[0.88, 0.88, 0.92]], 
         Cell[
          StyleData["Computation", "Layout"], Editable -> False], 
         Cell[
          StyleData["Computation", "Presentation"], Editable -> False]}, 
        Open]], 
      Cell[
       StyleData["ComputationResult"], CellFrame -> {{1.5, 0}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
       CellFrameMargins -> {{8, 8}, {4, 3}}, CellFrameColor -> 
       RGBColor[0.486, 0.33, 0.32], StyleMenuListing -> None, FontColor -> 
       RGBColor[0.23, 0.25, 0.34], Background -> RGBColor[0.88, 0.88, 0.92]], 
      
      Cell[
       StyleData["ComputationInfo"], ShowCellBracket -> False, 
       CellMargins -> {{120, 10}, {10, 2}}, StyleMenuListing -> None, 
       FontFamily -> $CellContext`Arial], 
      Cell[
       StyleData["ComputationInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{144, 10}, {10, 2}}, StyleMenuListing -> None], 
      Cell[
       StyleData["ProofInfo"], ShowCellBracket -> False, 
       CellMargins -> {{76, 10}, {6, 6}}, CellFrameLabelMargins -> 15, 
       StyleMenuListing -> None, FontFamily -> "Arial"], 
      Cell[
       StyleData["ProofInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{100, 10}, {0, 0}}, StyleMenuListing -> None], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FrameLabel"], LanguageCategory -> None, StyleMenuListing -> 
          None, FontFamily -> "Times New Roman", FontSize -> 10, FontWeight -> 
          "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
          FontVariations -> {
           "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
            "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
          RGBColor[0.55, 0.44, 0.42], Background -> None], 
         Cell[
          StyleData["FrameLabel", "Printout"]]}, Open]], 
      Cell[
       StyleData["OpenArchive"], CellFrame -> {{False, False}, {False, True}},
        ShowCellBracket -> Automatic, CellMargins -> {{70, 70}, {0, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 20}, CellFrameMargins -> 3, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], CellFrameLabelMargins -> 1, 
       DefaultFormatType -> DefaultInputFormatType, ShowAutoStyles -> False, 
       MenuSortingValue -> 1210, FontWeight -> Bold, Background -> 
       RGBColor[0.73, 0.74, 0.84]], 
      Cell[
       StyleData["ArchiveInfo"], ShowCellBracket -> Automatic, 
       CellMargins -> {{70, 70}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 5, CellFrameLabelMargins -> 3, ShowCellTags -> 
       False, ShowAutoStyles -> False, MenuSortingValue -> 1220, FontWeight -> 
       "Bold", Background -> RGBColor[0.88, 0.88, 0.92]], 
      Cell[
       StyleData["CloseArchive"], 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> 
       Automatic, CellMargins -> {{0, 0}, {5, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 25}, CellFrameMargins -> 2, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], TextAlignment -> Center, 
       MenuSortingValue -> 1230, FontFamily -> "Arial Black", FontSize -> 24, 
       FontWeight -> "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       Background -> RGBColor[0.73, 0.65, 0.61]], 
      Cell[
       StyleData["IncludeArchive"], CellFrame -> {{4, 0}, {False, False}}, 
       ShowCellBracket -> Automatic, 
       CellMargins -> {{81, 100}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameColor -> RGBColor[0.55, 0.44, 0.42], 
       CellFrameLabels -> {{"\[RightSkeleton]", None}, {None, None}}, 
       CellFrameLabelMargins -> 25, MenuSortingValue -> 1445], 
      Cell[
       StyleData["ExpressionVariable"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionABF"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionMeta"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["AutoParentheses"], Editable -> False, SpanMaxSize -> 3, 
          FontOpacity -> 0.15], 
         Cell[
          StyleData["AutoParentheses", "Layout"], FontOpacity -> 0], 
         Cell[
          StyleData["AutoParentheses", "SlideShow"], FontOpacity -> 0]}, 
        Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Emphasized"], MenuCommandKey -> "E", FontSlant -> 
          "Italic"], 
         Cell[
          StyleData["Emphasized", "SlideShow"], FontSlant -> "Plain", 
          FontColor -> RGBColor[0.36, 0.54, 0.69]]}, Open]], 
      Cell[
       StyleData["GenericButton"], FontFamily -> "Bitstream Charter", 
       FontWeight -> "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
       FontVariations -> {
        "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
         "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
       RGBColor[0.23, 0.25, 0.34]], 
      Cell[
       StyleData["TheoremaForm"], "TwoByteSyntaxCharacterAutoReplacement" -> 
       True, LineSpacing -> {1.25, 0, 2}, SingleLetterItalics -> True, 
       StyleMenuListing -> None, FontFamily -> "Courier"], 
      Cell[
       StyleData["ComponentActive"], FontFamily -> "Times New Roman"], 
      Cell[
       StyleData["ComponentInactive"], FontFamily -> "Times New Roman", 
       FontOpacity -> 0.4], 
      Cell[
       StyleData["ComponentEmpty"], FontFamily -> "Times New Roman", 
       FontVariations -> {"StrikeThrough" -> True}], 
      Cell[
       StyleData["TabView"], FontFamily -> "Times New Roman", FontSize -> 10], 
      Cell[
       StyleData["TabViewLabel"], FontFamily -> "Times New Roman", FontSize -> 
       10, FontWeight -> "Bold"], 
      Cell[
       StyleData["FormReference", StyleDefinitions -> StyleData["Text"]], 
       FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> "Plain",
        FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["FormReferenceHover", StyleDefinitions -> StyleData["Text"]],
        FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> 
       "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> True}, 
       FontColor -> RGBColor[0.404, 0.43, 0.545]], 
      Cell[
       StyleData["DisplayFormula"], CellMargins -> {{36, 6}, {7, 7}}, 
       ConversionRules :> {
        "HTML" -> {{"", ""}, {"<p class=\"Equation\">", "</p>"}}}, 
       CellHorizontalScrolling -> True, DefaultFormatType -> TraditionalForm, 
       InputAutoReplacements -> {}, TextAlignment -> Center, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       ButtonBoxOptions -> {Appearance -> {Automatic, None}}]}, Open]]}, 
  Visible -> False, FrontEndVersion -> 
  "9.0 for Linux x86 (64-bit) (February 7, 2013)", StyleDefinitions -> 
  "StylesheetFormatting.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "ID:558626034"->{
  Cell[64094, 1632, 6848, 193, 366, "FormalTextInputFormula",
   CellTags->{
    "ID:558626034", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCommPPDomain"},
   CellID->558626034]},
 "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"->{
  Cell[64094, 1632, 6848, 193, 366, "FormalTextInputFormula",
   CellTags->{
    "ID:558626034", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCommPPDomain"},
   CellID->558626034],
  Cell[81000, 2096, 785, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1568169011", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1568169011],
  Cell[82274, 2133, 898, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1716984519", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1716984519],
  Cell[83175, 2159, 2176, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:205548127", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial"},
   CellID->205548127],
  Cell[85687, 2236, 1836, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:711054552", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial II"},
   CellID->711054552],
  Cell[87526, 2294, 2348, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2035355406", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "monomial constructor"},
   CellID->2035355406],
  Cell[89877, 2368, 1219, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371468988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "supp"},
   CellID->371468988],
  Cell[91099, 2405, 2080, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:741382791", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isLP"},
   CellID->741382791],
  Cell[93182, 2468, 2289, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1470398456", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp"},
   CellID->1470398456],
  Cell[95474, 2539, 1075, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1541081556", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc"},
   CellID->1541081556],
  Cell[96552, 2571, 1204, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:258866072", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm"},
   CellID->258866072],
  Cell[97759, 2605, 1117, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1500059087", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R"},
   CellID->1500059087],
  Cell[98879, 2637, 1111, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1301149083", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 0"},
   CellID->1301149083],
  Cell[99993, 2667, 2271, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:355792939", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H"},
   CellID->355792939],
  Cell[102267, 2734, 2760, 76, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1617463509", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "monomial multiplication"},
   CellID->1617463509],
  Cell[105637, 2835, 2611, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1026275980", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReductionPolynomialDomain"},
   CellID->1026275980],
  Cell[108251, 2900, 4279, 121, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1335218902", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasDecisionPredicate"},
   CellID->1335218902],
  Cell[113212, 3046, 1778, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1175095894", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "containsMonomials"},
   CellID->1175095894],
  Cell[114993, 3099, 6005, 158, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1967795228", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCoefficientFunction"},
   CellID->1967795228],
  Cell[121001, 3259, 2813, 80, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2138030312", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplication"},
   CellID->2138030312],
  Cell[123817, 3341, 2210, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1709544239", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isExtensional"},
   CellID->1709544239],
  Cell[126030, 3412, 2868, 86, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:235606987", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrdered"},
   CellID->235606987],
  Cell[128901, 3500, 4757, 130, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:407906314", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultipliers"},
   CellID->407906314],
  Cell[134279, 3656, 1697, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1209025019", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isAlgoReductionPolynomialDomain"},
   CellID->1209025019],
  Cell[135979, 3707, 7435, 185, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1424238209", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasLCRD"},
   CellID->1424238209],
  Cell[143417, 3894, 4286, 114, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1761006179", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasRDM"},
   CellID->1761006179],
  Cell[147706, 4010, 7315, 181, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:667868750", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCPM"},
   CellID->667868750],
  Cell[158726, 4282, 805, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:267444622", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "ring \[ScriptCapitalR]"},
   CellID->267444622],
  Cell[159534, 4305, 899, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1725827561", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "power-product domain \[ScriptCapitalT]"},
   CellID->1725827561],
  Cell[160436, 4330, 809, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:90242530", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isDomain \[ScriptCapitalP]"},
   CellID->90242530],
  Cell[161248, 4353, 856, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1949049350", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasDecisionPredicate \[ScriptCapitalP]"},
   CellID->1949049350],
  Cell[162107, 4377, 835, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:378755333", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "containsMonomials \[ScriptCapitalP]"},
   CellID->378755333],
  Cell[162945, 4400, 859, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:226930466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCoefficientFunction \[ScriptCapitalP]"},
   CellID->226930466],
  Cell[163807, 4424, 848, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:221039322", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isExtensional \[ScriptCapitalP]"},
   CellID->221039322],
  Cell[166245, 4500, 810, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:216808867", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isDomain \[ScriptCapitalT]"},
   CellID->216808867],
  Cell[167058, 4523, 1082, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1391060217", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isPartIrreflOrder \[Succeeds] \[ScriptCapitalT]"},
   CellID->1391060217],
  Cell[168143, 4557, 2201, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1131450960", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Noetherian induction on \[ScriptCapitalT]", "IR"},
   CellID->1131450960],
  Cell[170347, 4629, 905, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:953344354", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 1 \[ScriptCapitalT]", "IR"},
   CellID->953344354],
  Cell[171783, 4672, 1558, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1367902825", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 1"},
   CellID->1367902825],
  Cell[173344, 4723, 761, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:643177315", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->643177315],
  Cell[174108, 4745, 916, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1568022983", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 1 min"},
   CellID->1568022983],
  Cell[175027, 4772, 1318, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1723905914", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 0"},
   CellID->1723905914],
  Cell[176348, 4814, 883, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1825953062", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 2'"},
   CellID->1825953062],
  Cell[177234, 4841, 879, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:596582273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 2"},
   CellID->596582273],
  Cell[178116, 4868, 901, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:389623087", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 3'"},
   CellID->389623087],
  Cell[179020, 4896, 899, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:885344283", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 3"},
   CellID->885344283],
  Cell[179922, 4924, 1017, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1033740528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 4"},
   CellID->1033740528],
  Cell[180942, 4956, 738, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:475932409", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->475932409],
  Cell[181908, 4984, 1218, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1124160616", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 5"},
   CellID->1124160616],
  Cell[183129, 5023, 1433, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1594052474", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 6"},
   CellID->1594052474],
  Cell[184565, 5068, 1470, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1091730754", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 7"},
   CellID->1091730754],
  Cell[186038, 5116, 1268, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1114122233", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 8"},
   CellID->1114122233],
  Cell[187309, 5156, 1033, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1909427587", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isPartReflOrder \[VerticalBar]"},
   CellID->1909427587],
  Cell[188345, 5188, 1214, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:196272597", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrderEmbedding \[RawStar] \[VerticalBar]"},
   CellID->196272597],
  Cell[259584, 7251, 761, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:334245933", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->334245933],
  Cell[260348, 7273, 926, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1101510167", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lcm", "IR"},
   CellID->1101510167],
  Cell[261650, 7309, 885, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:571674285", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 1"},
   CellID->571674285],
  Cell[262538, 7335, 889, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:775675057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 1'"},
   CellID->775675057],
  Cell[263430, 7361, 1471, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:517355459", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 2"},
   CellID->517355459],
  Cell[264904, 7408, 936, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:651227563", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 3"},
   CellID->651227563],
  Cell[265843, 7436, 844, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:482943675", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 6"},
   CellID->482943675],
  Cell[266899, 7469, 906, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:39860588", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 4"},
   CellID->39860588],
  Cell[267808, 7496, 911, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1631465634", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 4'"},
   CellID->1631465634],
  Cell[268722, 7523, 1608, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:835022497", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 5"},
   CellID->835022497],
  Cell[307311, 8634, 736, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:100863090", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->100863090],
  Cell[308438, 8668, 982, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:292726431", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership supp general"},
   CellID->292726431],
  Cell[309423, 8697, 1636, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:109557753", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Element] supp general"},
   CellID->109557753],
  Cell[319672, 8998, 1203, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:938132667", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership supp", "IR"},
   CellID->938132667],
  Cell[320878, 9031, 1747, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:56177686", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Element] supp", "IR"},
   CellID->56177686],
  Cell[322628, 9087, 867, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:163445395", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 0", "IR"},
   CellID->163445395],
  Cell[323498, 9113, 868, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:671805226", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 1", "IR"},
   CellID->671805226],
  Cell[324369, 9139, 1292, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:373967346", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership C", "IR"},
   CellID->373967346],
  Cell[342401, 9679, 2823, 76, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1248732928", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Monomial"},
   CellID->1248732928],
  Cell[345227, 9757, 2152, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1277602733", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp sufficient"},
   CellID->1277602733],
  Cell[347382, 9820, 2557, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1270225306", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp alternative"},
   CellID->1270225306],
  Cell[368770, 10398, 1246, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:292606918", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomials", "IR"},
   CellID->292606918],
  Cell[370019, 10435, 886, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:485824967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->485824967],
  Cell[370908, 10460, 950, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:406494450", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lp", "IR"},
   CellID->406494450],
  Cell[371861, 10488, 955, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1454724371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lc", "IR"},
   CellID->1454724371],
  Cell[372819, 10516, 955, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1589530505", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lm", "IR"},
   CellID->1589530505],
  Cell[388594, 10987, 740, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:720209465", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->720209465],
  Cell[389337, 11008, 915, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1852885314", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 0"},
   CellID->1852885314],
  Cell[390255, 11035, 761, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:148341425", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->148341425],
  Cell[391019, 11057, 1263, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1443097690", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 1"},
   CellID->1443097690],
  Cell[392285, 11095, 1055, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1675126992", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 2"},
   CellID->1675126992],
  Cell[393343, 11128, 1260, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:814027629", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 3"},
   CellID->814027629],
  Cell[394606, 11166, 1296, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1597798077", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 4"},
   CellID->1597798077],
  Cell[395905, 11206, 1596, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1711645326", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 5"},
   CellID->1711645326],
  Cell[425998, 12055, 764, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1053827266", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1053827266],
  Cell[426765, 12077, 928, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:829207957", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership + \[ScriptCapitalP]", "IR"},
   CellID->829207957],
  Cell[427696, 12104, 929, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:227196886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership - \[ScriptCapitalP] 1", "IR"},
   CellID->227196886],
  Cell[428628, 12131, 931, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414658623", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership - \[ScriptCapitalP] 2", "IR"},
   CellID->414658623],
  Cell[440323, 12479, 1210, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:314696779", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1=0 \[ScriptCapitalP]"},
   CellID->314696779],
  Cell[441536, 12517, 1287, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:849229881", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 1"},
   CellID->849229881],
  Cell[442826, 12558, 1291, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:406483636", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 1"},
   CellID->406483636],
  Cell[444120, 12599, 1241, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1767085150", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm 1"},
   CellID->1767085150],
  Cell[445364, 12639, 912, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1272860149", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1272860149],
  Cell[446279, 12664, 907, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:144989886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "non-zero 2", "IR"},
   CellID->144989886],
  Cell[447189, 12690, 1001, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2138850160", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 2"},
   CellID->2138850160],
  Cell[448193, 12720, 988, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1145306026", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 2"},
   CellID->1145306026],
  Cell[449184, 12751, 861, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1442316125", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 3"},
   CellID->1442316125],
  Cell[450048, 12777, 2155, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:655883205", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 3"},
   CellID->655883205],
  Cell[452206, 12847, 1488, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:996448533", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "non-zero 3", "IR"},
   CellID->996448533],
  Cell[453697, 12893, 1787, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1383271844", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 5"},
   CellID->1383271844],
  Cell[455487, 12949, 736, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:945941808", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->945941808],
  Cell[456226, 12970, 1443, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:857665826", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C lm"},
   CellID->857665826],
  Cell[457672, 13016, 1289, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1411816658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm 2"},
   CellID->1411816658],
  Cell[513591, 14603, 960, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1368691869", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1\[CenterDot]1"},
   CellID->1368691869],
  Cell[514554, 14630, 766, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1325102956", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1325102956],
  Cell[515323, 14652, 739, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1663774274", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1663774274],
  Cell[516065, 14673, 1124, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:941688542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial \[CenterDot]"},
   CellID->941688542],
  Cell[517192, 14707, 1187, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:91567084", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 0"},
   CellID->91567084],
  Cell[518382, 14745, 987, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:420812829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 3"},
   CellID->420812829],
  Cell[519372, 14774, 1703, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:138547187", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 4"},
   CellID->138547187],
  Cell[521078, 14829, 1242, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:457378178", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 11"},
   CellID->457378178],
  Cell[522323, 14868, 2172, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:926082604", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 11'"},
   CellID->926082604],
  Cell[524498, 14942, 1581, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1344734857", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] lp lc"},
   CellID->1344734857],
  Cell[561465, 15984, 1094, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1558664548", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 1"},
   CellID->1558664548],
  Cell[562562, 16018, 903, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1835029788", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1835029788],
  Cell[563468, 16044, 945, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:757450222", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial \[NotEqual] 0"},
   CellID->757450222],
  Cell[564416, 16069, 921, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1978849375", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial -"},
   CellID->1978849375],
  Cell[565340, 16096, 982, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:283455673", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 3'"},
   CellID->283455673],
  Cell[566325, 16125, 1438, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:610787620", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 3"},
   CellID->610787620],
  Cell[567766, 16169, 1841, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:980082706", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 4"},
   CellID->980082706],
  Cell[569610, 16229, 836, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1306217654", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm isMonomial"},
   CellID->1306217654],
  Cell[570449, 16254, 1570, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1645795930", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 7"},
   CellID->1645795930],
  Cell[572022, 16305, 736, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:691307310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->691307310],
  Cell[572761, 16326, 1778, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1901309763", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 8"},
   CellID->1901309763],
  Cell[574542, 16382, 1504, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1177943318", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 9"},
   CellID->1177943318],
  Cell[576049, 16432, 721, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1770611523", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1770611523],
  Cell[576773, 16453, 1719, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:670623560", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 12"},
   CellID->670623560],
  Cell[578495, 16509, 957, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1806313475", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial lm"},
   CellID->1806313475],
  Cell[579455, 16536, 964, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1110032163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc lm"},
   CellID->1110032163],
  Cell[580422, 16566, 961, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:632352665", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp lm"},
   CellID->632352665],
  Cell[641463, 18294, 1000, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:423895762", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ closed \[ScriptCapitalP]"},
   CellID->423895762],
  Cell[646025, 18448, 1108, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1812309492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "0 neutral \[ScriptCapitalP]"},
   CellID->1812309492],
  Cell[650801, 18606, 1048, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:595518277", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ assoc \[ScriptCapitalP]"},
   CellID->595518277],
  Cell[655387, 18760, 1046, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2114498624", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ comm \[ScriptCapitalP]"},
   CellID->2114498624],
  Cell[660028, 18920, 763, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:246227628", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->246227628],
  Cell[660794, 18942, 1007, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:95118700", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP] 1"},
   CellID->95118700],
  Cell[661804, 18972, 897, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1900465816", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP] 2"},
   CellID->1900465816],
  Cell[662704, 18997, 1240, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2037231548", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP]"},
   CellID->2037231548],
  Cell[674363, 19354, 918, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:145364866", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- 0 \[ScriptCapitalP] 1"},
   CellID->145364866],
  Cell[675284, 19380, 1130, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:117883331", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- 0 \[ScriptCapitalP] 2"},
   CellID->117883331],
  Cell[683620, 19649, 1059, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:771361293", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCCMonoid \[ScriptCapitalP]"},
   CellID->771361293],
  Cell[688477, 19804, 842, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1221832846", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1221832846],
  Cell[689322, 19828, 944, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:313886646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership R", "IR"},
   CellID->313886646],
  Cell[690269, 19856, 952, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:174448700", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm+R"},
   CellID->174448700],
  Cell[691224, 19885, 1735, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1679470623", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C R"},
   CellID->1679470623],
  Cell[692962, 19938, 1144, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1648446612", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 1"},
   CellID->1648446612],
  Cell[694109, 19973, 1672, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1799775671", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 2"},
   CellID->1799775671],
  Cell[695784, 20025, 1024, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1065304756", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 6"},
   CellID->1065304756],
  Cell[696811, 20056, 1534, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1078436809", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 7"},
   CellID->1078436809],
  Cell[698348, 20107, 1061, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:735362349", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 8"},
   CellID->735362349],
  Cell[699412, 20141, 1062, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:451768283", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 8'"},
   CellID->451768283],
  Cell[700477, 20175, 2072, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739591353", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 5'"},
   CellID->739591353],
  Cell[702552, 20243, 796, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1347176402", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1347176402],
  Cell[703351, 20267, 1222, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1377108879", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 3"},
   CellID->1377108879],
  Cell[704576, 20300, 1765, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:139632107", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 4"},
   CellID->139632107],
  Cell[706344, 20358, 1030, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2132986120", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 7"},
   CellID->2132986120],
  Cell[707377, 20389, 1103, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1911280389", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 8"},
   CellID->1911280389],
  Cell[773670, 22234, 743, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1031528829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1031528829],
  Cell[774416, 22255, 817, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1494855415", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1494855415],
  Cell[775236, 22278, 1538, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2007119699", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership H ind"},
   CellID->2007119699],
  Cell[776777, 22324, 917, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:884488101", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership H", "IR"},
   CellID->884488101],
  Cell[777697, 22351, 2154, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1892666968", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C H ind"},
   CellID->1892666968],
  Cell[779854, 22419, 1687, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1556838587", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C H"},
   CellID->1556838587],
  Cell[781544, 22473, 1612, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1768827306", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H=0"},
   CellID->1768827306],
  Cell[783159, 22524, 1783, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:140826492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H=H"},
   CellID->140826492],
  Cell[784945, 22583, 1161, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1627546251", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 1"},
   CellID->1627546251],
  Cell[786109, 22619, 1056, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:743145124", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 2"},
   CellID->743145124],
  Cell[787168, 22652, 1163, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1576432658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 3"},
   CellID->1576432658],
  Cell[788334, 22688, 1591, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1977357773", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 4"},
   CellID->1977357773],
  Cell[789928, 22739, 1269, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:561821259", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 4'"},
   CellID->561821259],
  Cell[791200, 22778, 2151, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:866138121", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 5"},
   CellID->866138121],
  Cell[793354, 22846, 2071, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:105044014", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 5'"},
   CellID->105044014],
  Cell[795428, 22910, 1611, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1971276224", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 6"},
   CellID->1971276224],
  Cell[797042, 22961, 744, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:219245057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->219245057],
  Cell[797789, 22983, 953, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2135590264", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 7"},
   CellID->2135590264],
  Cell[798745, 23012, 1214, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2092106183", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 8"},
   CellID->2092106183],
  Cell[799962, 23050, 793, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:831355566", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->831355566],
  Cell[800758, 23074, 1112, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2014193763", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 10"},
   CellID->2014193763],
  Cell[801873, 23108, 1137, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1650576647", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 11"},
   CellID->1650576647],
  Cell[899076, 25708, 641, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:45249416", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->45249416],
  Cell[899720, 25726, 3991, 108, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:248026320", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "induction on \[ScriptCapitalP] ind"},
   CellID->248026320],
  Cell[903714, 25836, 1918, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502331257", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "induction on \[ScriptCapitalP]", "IR"},
   CellID->1502331257],
  Cell[921541, 26309, 766, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:411993095", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->411993095],
  Cell[922310, 26331, 1725, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:470022203", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 1"},
   CellID->470022203],
  Cell[924038, 26382, 847, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:401540586", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->401540586],
  Cell[924888, 26407, 1468, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2014025290", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 6"},
   CellID->2014025290],
  Cell[926359, 26448, 870, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1590906878", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 10"},
   CellID->1590906878],
  Cell[927232, 26473, 1279, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1997927583", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 11"},
   CellID->1997927583],
  Cell[947818, 27051, 932, 24, 66, "GlobalDeclaration",
   CellTags->{
    "ID:63039386", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->63039386],
  Cell[949017, 27084, 926, 24, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1093453827", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 0"},
   CellID->1093453827],
  Cell[949946, 27110, 1902, 54, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1795167993", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 5"},
   CellID->1795167993],
  Cell[951851, 27166, 1575, 45, 76, "FormalTextInputFormula",
   CellTags->{
    "ID:838653570", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomial multiplication ind"},
   CellID->838653570],
  Cell[953429, 27213, 987, 27, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1903131381", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomial multiplication", "IR"},
   CellID->1903131381],
  Cell[954419, 27242, 2017, 58, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:1641583144", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C \[CircleTimes]"},
   CellID->1641583144],
  Cell[956439, 27302, 1233, 35, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:608153969", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 1"},
   CellID->608153969],
  Cell[957675, 27339, 1673, 48, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1439631766", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] isMonomial"},
   CellID->1439631766],
  Cell[959351, 27389, 736, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:607306392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->607306392],
  Cell[960090, 27410, 1163, 32, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1312761442", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 2"},
   CellID->1312761442],
  Cell[961256, 27444, 1345, 41, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:259600169", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 3"},
   CellID->259600169],
  Cell[962604, 27487, 999, 27, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1576857420", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 6"},
   CellID->1576857420],
  Cell[963606, 27516, 1844, 55, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:483781382", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 10"},
   CellID->483781382],
  Cell[965453, 27573, 1607, 46, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:169057028", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 11"},
   CellID->169057028],
  Cell[967063, 27621, 721, 19, 31, "GlobalDeclaration",
   CellTags->{
    "ID:324898493", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->324898493],
  Cell[967787, 27642, 1332, 39, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:684947841", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 7"},
   CellID->684947841],
  Cell[969122, 27683, 1492, 43, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1856294157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 8"},
   CellID->1856294157],
  Cell[970617, 27728, 1138, 31, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1768460531", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 9"},
   CellID->1768460531],
  Cell[1075368, 30600, 5057, 136, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1437523678", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplicationRec"},
   CellID->1437523678],
  Cell[1081114, 30766, 710, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1372918741", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1372918741],
  Cell[1081827, 30786, 1824, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:271668948", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "316"},
   CellID->271668948],
  Cell[1083654, 30843, 1263, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:428541408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "317"},
   CellID->428541408],
  Cell[1084920, 30883, 800, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1983432019", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "318"},
   CellID->1983432019],
  Cell[1100663, 31325, 1567, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1310550300", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "319"},
   CellID->1310550300],
  Cell[1103434, 31417, 835, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:978297022", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplication \[ScriptCapitalP]"},
   CellID->978297022],
  Cell[1104880, 31464, 1495, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1125626334", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "320"},
   CellID->1125626334],
  Cell[1110563, 31634, 1031, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1103171586", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] closed \[ScriptCapitalP]"},
   CellID->1103171586],
  Cell[1115479, 31797, 1392, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2097953227", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "322"},
   CellID->2097953227],
  Cell[1122684, 32012, 762, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:587826938", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->587826938],
  Cell[1123449, 32034, 885, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:384020988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "324"},
   CellID->384020988],
  Cell[1124337, 32060, 1094, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:205407574", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "325"},
   CellID->205407574],
  Cell[1125434, 32095, 1070, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:411846169", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "327"},
   CellID->411846169],
  Cell[1126507, 32130, 1086, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1300778952", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "328"},
   CellID->1300778952],
  Cell[1127596, 32165, 1240, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2123221051", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "329"},
   CellID->2123221051],
  Cell[1148881, 32793, 1065, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:840093990", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] comm \[ScriptCapitalP]"},
   CellID->840093990],
  Cell[1153515, 32952, 1230, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1730144032", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] distr \[ScriptCapitalP]"},
   CellID->1730144032],
  Cell[1158481, 33119, 1156, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:378610433", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1 neutral \[ScriptCapitalP]"},
   CellID->378610433],
  Cell[1163112, 33271, 941, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:687899447", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->687899447],
  Cell[1164056, 33297, 1549, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:421376854", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "330"},
   CellID->421376854],
  Cell[1165608, 33344, 1765, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:935511285", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "331"},
   CellID->935511285],
  Cell[1167376, 33400, 1338, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1702504479", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "332"},
   CellID->1702504479],
  Cell[1168717, 33442, 1912, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1379089132", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "333"},
   CellID->1379089132],
  Cell[1194015, 34125, 1071, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:167864408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] assoc \[ScriptCapitalP]"},
   CellID->167864408],
  Cell[1198459, 34271, 820, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1082265416", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCommRing1 \[ScriptCapitalP]"},
   CellID->1082265416],
  Cell[1203285, 34427, 1021, 27, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1610244238", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1610244238],
  Cell[1204309, 34456, 2470, 77, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1259587933", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 5"},
   CellID->1259587933],
  Cell[1206782, 34535, 2175, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:859675419", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6 ind"},
   CellID->859675419],
  Cell[1208960, 34604, 1733, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:515125163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6"},
   CellID->515125163],
  Cell[1210696, 34656, 1516, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1572844749", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6'"},
   CellID->1572844749],
  Cell[1212215, 34700, 1674, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:771023286", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 4"},
   CellID->771023286],
  Cell[1213892, 34747, 1308, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:422533310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 9"},
   CellID->422533310],
  Cell[1253568, 35779, 766, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1667882379", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1667882379],
  Cell[1254337, 35801, 739, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1105090112", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1105090112],
  Cell[1255079, 35822, 1565, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:878771843", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 2"},
   CellID->878771843],
  Cell[1256647, 35871, 868, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1863009930", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1863009930],
  Cell[1257518, 35896, 1713, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1958431161", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 5"},
   CellID->1958431161],
  Cell[1259234, 35947, 1464, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:758577094", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 6"},
   CellID->758577094],
  Cell[1260701, 35989, 736, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:544874520", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->544874520],
  Cell[1261440, 36010, 738, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:158305602", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->158305602],
  Cell[1262181, 36031, 1286, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:347486225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 9"},
   CellID->347486225],
  Cell[1263470, 36069, 1453, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1591957532", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 10"},
   CellID->1591957532],
  Cell[1264926, 36110, 721, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1155967768", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1155967768],
  Cell[1265650, 36131, 1293, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1942982211", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 15"},
   CellID->1942982211],
  Cell[1266946, 36172, 1193, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:44857450", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 12"},
   CellID->44857450],
  Cell[1268142, 36207, 1246, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:928737230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 13"},
   CellID->928737230],
  Cell[1269391, 36243, 1269, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:233618042", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 14"},
   CellID->233618042],
  Cell[1321436, 37608, 900, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:183428126", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->183428126],
  Cell[1322339, 37634, 1652, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2026663200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 2"},
   CellID->2026663200],
  Cell[1323994, 37684, 1920, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1916930407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 5"},
   CellID->1916930407],
  Cell[1339640, 38106, 830, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1225634217", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduction ring \[ScriptCapitalR]"},
   CellID->1225634217],
  Cell[1341015, 38152, 2050, 59, 74, "FormalTextInputFormula",
   CellTags->{
    "ID:1542801947", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "300"},
   CellID->1542801947],
  Cell[1348887, 38376, 5079, 138, 212, "FormalTextInputFormula",
   CellTags->{
    "ID:946287998", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrderedRec"},
   CellID->946287998],
  Cell[1354656, 38542, 704, 18, 23, "GlobalDeclaration",
   CellTags->{
    "ID:1485445352", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1485445352],
  Cell[1355363, 38562, 2882, 86, 82, "FormalTextInputFormula",
   CellTags->{
    "ID:215225048", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "301"},
   CellID->215225048],
  Cell[1358248, 38650, 2636, 82, 82, "FormalTextInputFormula",
   CellTags->{
    "ID:1704612967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "302"},
   CellID->1704612967],
  Cell[1360887, 38734, 789, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:162248449", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "303"},
   CellID->162248449],
  Cell[1418646, 40401, 1650, 45, 54, "FormalTextInputFormula",
   CellTags->{
    "ID:1015468103", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "304"},
   CellID->1015468103],
  Cell[1422978, 40517, 853, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:7476052", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduction-polynomial domain \[ScriptCapitalP]"},
   CellID->7476052],
  Cell[1424978, 40581, 763, 20, 42, "GlobalDeclaration",
   CellTags->{
    "ID:173226898", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->173226898],
  Cell[1425744, 40603, 738, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:835223349", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->835223349],
  Cell[1426485, 40624, 929, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:181219027", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "356"},
   CellID->181219027],
  Cell[1427417, 40651, 1555, 46, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:271793724", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "382"},
   CellID->271793724],
  Cell[1428975, 40699, 1487, 44, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:575028756", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "382'"},
   CellID->575028756],
  Cell[1430465, 40745, 1606, 48, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:2020716880", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "383"},
   CellID->2020716880],
  Cell[1432074, 40795, 1655, 48, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1964835443", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "381"},
   CellID->1964835443],
  Cell[1433732, 40845, 1945, 55, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1599847484", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "380"},
   CellID->1599847484],
  Cell[1435680, 40902, 3915, 101, 100, "FormalTextInputFormula",
   CellTags->{
    "ID:590709100", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "379"},
   CellID->590709100],
  Cell[1439598, 41005, 1877, 57, 65, "FormalTextInputFormula",
   CellTags->{
    "ID:586184654", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "405"},
   CellID->586184654],
  Cell[1441478, 41064, 1402, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:919284948", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "406"},
   CellID->919284948],
  Cell[1503602, 42766, 738, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:895736483", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->895736483],
  Cell[1504343, 42787, 741, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1904443546", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1904443546],
  Cell[1505087, 42808, 739, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1587558771", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1587558771],
  Cell[1505829, 42829, 1422, 41, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:530327018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377"},
   CellID->530327018],
  Cell[1507254, 42872, 2251, 69, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:2044424156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377'"},
   CellID->2044424156],
  Cell[1509508, 42943, 1353, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:14600367", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377''"},
   CellID->14600367],
  Cell[1510864, 42984, 1494, 44, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1313257457", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 7"},
   CellID->1313257457],
  Cell[1512361, 43030, 2317, 65, 87, "FormalTextInputFormula",
   CellTags->{
    "ID:1776891622", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 8"},
   CellID->1776891622],
  Cell[1543430, 43861, 1093, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1077742446", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1077742446],
  Cell[1544526, 43889, 1503, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1717822499", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "384"},
   CellID->1717822499],
  Cell[1546032, 43934, 913, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1045770541", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "358"},
   CellID->1045770541],
  Cell[1546948, 43961, 2336, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:312784637", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "393"},
   CellID->312784637],
  Cell[1549287, 44028, 859, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:172690263", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->172690263],
  Cell[1550149, 44053, 1284, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:174940712", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial"},
   CellID->174940712],
  Cell[1551436, 44090, 1482, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1850961673", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial II"},
   CellID->1850961673],
  Cell[1552921, 44136, 2001, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:849856225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial III"},
   CellID->849856225],
  Cell[1591199, 45143, 1001, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:633254127", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->633254127],
  Cell[1592203, 45171, 1124, 29, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1581566919", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1581566919],
  Cell[1593330, 45202, 977, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1230351665", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 1"},
   CellID->1230351665],
  Cell[1594310, 45229, 816, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1294700753", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1294700753],
  Cell[1595129, 45253, 1349, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:168346017", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 2"},
   CellID->168346017],
  Cell[1596481, 45293, 1960, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1208795229", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 3"},
   CellID->1208795229],
  Cell[1598444, 45352, 1764, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:369110158", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 4"},
   CellID->369110158],
  Cell[1629797, 46177, 761, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:155708286", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->155708286],
  Cell[1630561, 46199, 807, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:280312466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "4"},
   CellID->280312466],
  Cell[1631371, 46224, 1007, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:713919982", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "5"},
   CellID->713919982],
  Cell[1632381, 46256, 1354, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1630947645", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "6"},
   CellID->1630947645],
  Cell[1646356, 46670, 1075, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:795795393", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "360"},
   CellID->795795393],
  Cell[1650783, 46817, 778, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1302712453", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1302712453],
  Cell[1651564, 46839, 5366, 155, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1092343128", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1"},
   CellID->1092343128],
  Cell[1656933, 46996, 1107, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:311837531", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "2"},
   CellID->311837531],
  Cell[1658043, 47032, 1868, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2036982330", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "3"},
   CellID->2036982330],
  Cell[1659914, 47095, 2517, 75, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:917853466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "378"},
   CellID->917853466],
  Cell[1690346, 47888, 1019, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1009043994", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "361"},
   CellID->1009043994],
  Cell[1694675, 48034, 791, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:412087985", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "362"},
   CellID->412087985],
  Cell[1698554, 48164, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1267059190", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "363"},
   CellID->1267059190],
  Cell[1702407, 48294, 780, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:378652997", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "364"},
   CellID->378652997],
  Cell[1706676, 48435, 780, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1036416147", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "365"},
   CellID->1036416147],
  Cell[1710553, 48567, 2877, 78, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1341307829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "386"},
   CellID->1341307829],
  Cell[1713433, 48647, 3328, 89, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1193326225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "385"},
   CellID->1193326225],
  Cell[1733185, 49165, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2033377127", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "366"},
   CellID->2033377127],
  Cell[1737106, 49296, 779, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:982813600", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "367"},
   CellID->982813600],
  Cell[1740681, 49417, 776, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1179937112", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1179937112],
  Cell[1741460, 49439, 1646, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134072435", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "336"},
   CellID->134072435],
  Cell[1743109, 49489, 1121, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1484092370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "337"},
   CellID->1484092370],
  Cell[1755622, 49840, 1026, 27, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1480600758", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1480600758],
  Cell[1756651, 49869, 739, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1004003660", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1004003660],
  Cell[1757393, 49890, 1075, 28, 56, "GlobalDeclaration",
   CellTags->{
    "ID:1854976476", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1854976476],
  Cell[1758471, 49920, 1867, 50, 58, "FormalTextInputFormula",
   CellTags->{
    "ID:408134090", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduceM \[ScriptCapitalP]"},
   CellID->408134090],
  Cell[1760341, 49972, 2604, 72, 89, "FormalTextInputFormula",
   CellTags->{
    "ID:1088018783", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduceM \[ScriptCapitalP] II"},
   CellID->1088018783],
  Cell[1783259, 50545, 953, 25, 66, "GlobalDeclaration",
   CellTags->{
    "ID:1964871967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1964871967],
  Cell[1784724, 50585, 1865, 53, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1041722017", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "407"},
   CellID->1041722017],
  Cell[1786592, 50640, 738, 19, 31, "GlobalDeclaration",
   CellTags->{
    "ID:1518222527", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1518222527],
  Cell[1787333, 50661, 2380, 68, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1070088161", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "408"},
   CellID->1070088161],
  Cell[1789716, 50731, 2361, 68, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1970014385", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "408'"},
   CellID->1970014385],
  Cell[1816716, 51412, 786, 21, 42, "GlobalDeclaration",
   CellTags->{
    "ID:102721391", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->102721391],
  Cell[1817505, 51435, 2568, 73, 120, "FormalTextInputFormula",
   CellTags->{
    "ID:657593834", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED"},
   CellID->657593834],
  Cell[1820076, 51510, 2296, 62, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:1802485517", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED II"},
   CellID->1802485517],
  Cell[1822375, 51574, 2242, 60, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:1164884865", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED III"},
   CellID->1164884865],
  Cell[1824620, 51636, 2619, 72, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:958945878", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED IV"},
   CellID->958945878],
  Cell[1827242, 51710, 886, 24, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1601029605", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1601029605],
  Cell[1828131, 51736, 2247, 62, 120, "FormalTextInputFormula",
   CellTags->{
    "ID:687002895", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED monomial"},
   CellID->687002895],
  Cell[1830381, 51800, 1920, 55, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:413658239", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED monomial II"},
   CellID->413658239],
  Cell[1878303, 52988, 771, 20, 42, "GlobalDeclaration",
   CellTags->{
    "ID:580194792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->580194792],
  Cell[1879077, 53010, 2513, 75, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1667457569", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "357"},
   CellID->1667457569],
  Cell[1881593, 53087, 1878, 61, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:2035027156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] lm \[DoubleRightArrow]"},
   CellID->2035027156],
  Cell[1883474, 53150, 1822, 56, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:2026457432", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] lm \[DoubleLeftArrow]"},
   CellID->2026457432],
  Cell[1885299, 53208, 2433, 70, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:1167741853", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reducibility \[ScriptCapitalP]"},
   CellID->1167741853],
  Cell[1887735, 53280, 3955, 105, 110, "FormalTextInputFormula",
   CellTags->{
    "ID:1998060487", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] alternative \[ScriptCapitalP]"},
   CellID->1998060487],
  Cell[1927649, 54306, 927, 24, 66, "GlobalDeclaration",
   CellTags->{
    "ID:1709390289", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1709390289],
  Cell[1928579, 54332, 4827, 129, 102, "FormalTextInputFormula",
   CellTags->{
    "ID:1713414293", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[LeftRightArrow] alternative \[ScriptCapitalP]"},
   CellID->1713414293],
  Cell[1957261, 55102, 3791, 102, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1190720464", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducibleRec \[ScriptCapitalP]"},
   CellID->1190720464],
  Cell[1961713, 55232, 924, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:166529906", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->166529906],
  Cell[1962640, 55257, 1304, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:844782115", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352.\[DoubleRightArrow]"},
   CellID->844782115],
  Cell[1963947, 55289, 1106, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:953126445", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352.\[DoubleLeftArrow]"},
   CellID->953126445],
  Cell[1965056, 55319, 1233, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:259818613", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352"},
   CellID->259818613],
  Cell[1966292, 55351, 2618, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:182112345", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "353"},
   CellID->182112345],
  Cell[1990319, 55999, 1060, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:418362312", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->418362312],
  Cell[1991382, 56027, 739, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2116744492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->2116744492],
  Cell[1992124, 56048, 2781, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1746803457", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "391"},
   CellID->1746803457],
  Cell[1994908, 56129, 826, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:899106457", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->899106457],
  Cell[1995737, 56152, 1099, 28, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1157329004", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1157329004],
  Cell[1996839, 56182, 2316, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1927781160", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "387"},
   CellID->1927781160],
  Cell[2020358, 56761, 781, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:934500559", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "368"},
   CellID->934500559],
  Cell[2024330, 56894, 1088, 28, 70, "GlobalDeclaration",
   CellTags->{
    "ID:371274476", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->371274476],
  Cell[2025421, 56924, 2225, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1588964752", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR"},
   CellID->1588964752],
  Cell[2027649, 56988, 2461, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:428389793", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR II"},
   CellID->428389793],
  Cell[2030113, 57060, 1800, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:225111715", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CORR"},
   CellID->225111715],
  Cell[2031916, 57115, 915, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1632881403", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1632881403],
  Cell[2032834, 57142, 2042, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1701548086", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR \[CenterDot]"},
   CellID->1701548086],
  Cell[2034879, 57200, 1266, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1758251641", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CORR \[CenterDot]"},
   CellID->1758251641],
  Cell[2079926, 58284, 939, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1928324287", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1928324287],
  Cell[2080868, 58310, 1839, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:751250157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404"},
   CellID->751250157],
  Cell[2082710, 58365, 1866, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1952511245", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404 II"},
   CellID->1952511245],
  Cell[2084579, 58421, 2301, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:537853020", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404 III"},
   CellID->537853020],
  Cell[2110000, 59065, 963, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:659046058", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->659046058],
  Cell[2110966, 59092, 736, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:319520004", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->319520004],
  Cell[2111705, 59113, 826, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:967033090", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->967033090],
  Cell[2112534, 59136, 1098, 28, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1647700871", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1647700871],
  Cell[2113635, 59166, 786, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1981597108", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1981597108],
  Cell[2114818, 59198, 2437, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40011060", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "388"},
   CellID->40011060],
  Cell[2117258, 59267, 792, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:884639747", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->884639747],
  Cell[2118053, 59290, 1524, 45, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1360655817", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1360655817],
  Cell[2119580, 59337, 1516, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1002690147", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "389"},
   CellID->1002690147],
  Cell[2121099, 59383, 1485, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1896907436", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "390"},
   CellID->1896907436],
  Cell[2158559, 60277, 781, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1290694758", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "369"},
   CellID->1290694758],
  Cell[2162552, 60410, 1046, 27, 71, "GlobalDeclaration",
   CellTags->{
    "ID:1785082178", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1785082178],
  Cell[2163601, 60439, 840, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1718264802", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1718264802],
  Cell[2164444, 60463, 803, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:752304273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->752304273],
  Cell[2165250, 60486, 2765, 79, 145, "FormalTextInputFormula",
   CellTags->{
    "ID:578561747", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "areCommonReducibles I"},
   CellID->578561747],
  Cell[2168018, 60567, 1563, 45, 85, "FormalTextInputFormula",
   CellTags->{
    "ID:1399766637", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "areCommonReducibles II"},
   CellID->1399766637],
  Cell[2169584, 60614, 3366, 94, 145, "FormalTextInputFormula",
   CellTags->{
    "ID:1132709394", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP"},
   CellID->1132709394],
  Cell[2173619, 60735, 1018, 27, 71, "GlobalDeclaration",
   CellTags->{
    "ID:176358731", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->176358731],
  Cell[2174640, 60764, 885, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:391507392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->391507392],
  Cell[2175528, 60788, 804, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:1030826183", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1030826183],
  Cell[2176335, 60811, 1752, 52, 78, "FormalTextInputFormula",
   CellTags->{
    "ID:926282876", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "354"},
   CellID->926282876],
  Cell[2178090, 60865, 1749, 51, 91, "FormalTextInputFormula",
   CellTags->{
    "ID:80931797", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "355"},
   CellID->80931797],
  Cell[2179842, 60918, 1448, 43, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:414980491", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "359"},
   CellID->414980491],
  Cell[2181293, 60963, 908, 25, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:744723554", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "345"},
   CellID->744723554],
  Cell[2214938, 61800, 1018, 27, 71, "GlobalDeclaration",
   CellTags->{
    "ID:381508689", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->381508689],
  Cell[2215959, 61829, 741, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1050210793", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1050210793],
  Cell[2216703, 61850, 803, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:512389161", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->512389161],
  Cell[2217509, 61873, 1702, 51, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:719916043", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative 1"},
   CellID->719916043],
  Cell[2219214, 61926, 1596, 49, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:864384559", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative 2"},
   CellID->864384559],
  Cell[2220813, 61977, 2056, 57, 67, "FormalTextInputFormula",
   CellTags->{
    "ID:951944988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative"},
   CellID->951944988],
  Cell[2290519, 63920, 1175, 29, 71, "GlobalDeclaration",
   CellTags->{
    "ID:1990838743", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1990838743],
  Cell[2291697, 63951, 896, 23, 43, "GlobalDeclaration",
   CellTags->{
    "ID:20910152", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->20910152],
  Cell[2292596, 63976, 3492, 91, 134, "FormalTextInputFormula",
   CellTags->{
    "ID:416591219", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleRightArrow].1"},
   CellID->416591219],
  Cell[2296091, 64069, 2290, 63, 94, "FormalTextInputFormula",
   CellTags->{
    "ID:295118406", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleRightArrow].2"},
   CellID->295118406],
  Cell[2298384, 64134, 4394, 121, 204, "FormalTextInputFormula",
   CellTags->{
    "ID:1583101849", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleLeftArrow]"},
   CellID->1583101849],
  Cell[2403947, 67120, 1086, 26, 66, "GlobalDeclaration",
   CellTags->{
    "ID:1272506214", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1272506214],
  Cell[2405036, 67148, 2896, 80, 127, "FormalTextInputFormula",
   CellTags->{
    "ID:386901098", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleRightArrow].1"},
   CellID->386901098],
  Cell[2407935, 67230, 2132, 61, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:670300270", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleRightArrow].2"},
   CellID->670300270],
  Cell[2410070, 67293, 4419, 125, 172, "FormalTextInputFormula",
   CellTags->{
    "ID:1434882124", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleLeftArrow]"},
   CellID->1434882124],
  Cell[2495075, 69709, 1008, 27, 71, "GlobalDeclaration",
   CellTags->{
    "ID:737627449", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->737627449],
  Cell[2496086, 69738, 737, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:81313572", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->81313572],
  Cell[2496826, 69759, 767, 20, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1098402572", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1098402572],
  Cell[2497596, 69781, 2466, 71, 109, "FormalTextInputFormula",
   CellTags->{
    "ID:1499935354", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 1"},
   CellID->1499935354],
  Cell[2500065, 69854, 2698, 73, 94, "FormalTextInputFormula",
   CellTags->{
    "ID:1938671644", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 2"},
   CellID->1938671644],
  Cell[2502766, 69929, 2305, 64, 89, "FormalTextInputFormula",
   CellTags->{
    "ID:48476485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 3"},
   CellID->48476485],
  Cell[2505074, 69995, 2004, 59, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1246936946", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 1"},
   CellID->1246936946],
  Cell[2507081, 70056, 2186, 62, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:1856728487", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 2"},
   CellID->1856728487],
  Cell[2509270, 70120, 1720, 50, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:335657318", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 3"},
   CellID->335657318],
  Cell[2585532, 72148, 2769, 80, 137, "FormalTextInputFormula",
   CellTags->{
    "ID:624517207", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMNTCRP"},
   CellID->624517207],
  Cell[2589233, 72262, 3080, 86, 159, "FormalTextInputFormula",
   CellTags->{
    "ID:1172537415", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMNTCRP alternative"},
   CellID->1172537415],
  Cell[2613705, 72951, 1084, 28, 71, "GlobalDeclaration",
   CellTags->{
    "ID:1861685863", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1861685863],
  Cell[2614792, 72981, 806, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:1154885365", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1154885365],
  Cell[2615601, 73004, 1000, 26, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1777331368", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].1"},
   CellID->1777331368],
  Cell[2616604, 73032, 1375, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1052103130", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].2"},
   CellID->1052103130],
  Cell[2617982, 73073, 1127, 29, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1077684208", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].3"},
   CellID->1077684208],
  Cell[2619112, 73104, 1660, 47, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:86200506", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleLeftArrow]"},
   CellID->86200506],
  Cell[2620775, 73153, 1938, 57, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:394014296", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3"},
   CellID->394014296],
  Cell[2705282, 75493, 994, 25, 66, "GlobalDeclaration",
   CellTags->{
    "ID:1444400550", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1444400550],
  Cell[2706279, 75520, 965, 26, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:653031981", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].1"},
   CellID->653031981],
  Cell[2707247, 75548, 1157, 33, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:943141844", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].2"},
   CellID->943141844],
  Cell[2708407, 75583, 1190, 32, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1143014070", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].3"},
   CellID->1143014070],
  Cell[2709600, 75617, 1536, 45, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1630861163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleLeftArrow]"},
   CellID->1630861163],
  Cell[2711139, 75664, 1755, 53, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:1701661908", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4"},
   CellID->1701661908],
  Cell[2778126, 77591, 1118, 28, 71, "GlobalDeclaration",
   CellTags->{
    "ID:1489378443", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1489378443],
  Cell[2779247, 77621, 800, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:196663580", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->196663580],
  Cell[2780050, 77644, 775, 20, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1830388955", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1830388955],
  Cell[2780828, 77666, 3394, 93, 155, "FormalTextInputFormula",
   CellTags->{
    "ID:312371876", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CP1.\[DoubleLeftArrow]"},
   CellID->312371876],
  Cell[2784225, 77761, 4109, 110, 241, "FormalTextInputFormula",
   CellTags->{
    "ID:1006612102", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CP2.\[DoubleLeftArrow]"},
   CellID->1006612102],
  Cell[2818600, 78561, 1020, 27, 71, "GlobalDeclaration",
   CellTags->{
    "ID:542206690", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->542206690],
  Cell[2819623, 78590, 741, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1548368737", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1548368737],
  Cell[2820367, 78611, 798, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:73478134", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->73478134],
  Cell[2821168, 78634, 9746, 250, 423, "FormalTextInputFormula",
   CellTags->{
    "ID:310376772", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "409"},
   CellID->310376772],
  Cell[2873845, 79926, 840, 21, 42, "GlobalDeclaration",
   CellTags->{
    "ID:383946867", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->383946867],
  Cell[2874688, 79949, 818, 21, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1587705005", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1587705005],
  Cell[2875899, 79986, 3055, 95, 108, "FormalTextInputFormula",
   CellTags->{
    "ID:549840930", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].3"},
   CellID->549840930],
  Cell[2878957, 80083, 4804, 151, 109, "FormalTextInputFormula",
   CellTags->{
    "ID:657656821", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].5"},
   CellID->657656821],
  Cell[2883764, 80236, 934, 24, 66, "GlobalDeclaration",
   CellTags->{
    "ID:64232512", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->64232512],
  Cell[2884701, 80262, 818, 21, 41, "GlobalDeclaration",
   CellTags->{
    "ID:237038572", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->237038572],
  Cell[2885522, 80285, 718, 19, 31, "GlobalDeclaration",
   CellTags->{
    "ID:235501232", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->235501232],
  Cell[2886243, 80306, 1796, 49, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1513648693", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].1"},
   CellID->1513648693],
  Cell[2888042, 80357, 3126, 90, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:808269926", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].4"},
   CellID->808269926],
  Cell[2985161, 83029, 779, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:255426543", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "370"},
   CellID->255426543],
  Cell[3015861, 83980, 782, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1342888001", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "371"},
   CellID->1342888001],
  Cell[3032786, 84504, 965, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1021018734", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "376"},
   CellID->1021018734],
  Cell[3037212, 84643, 1659, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:459032599", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReductionRing Conservation Theorem"},
   CellID->459032599],
  Cell[3040300, 84740, 864, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:719489133", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "algorithmic reduction ring \[ScriptCapitalR]"},
   CellID->719489133],
  Cell[3041167, 84764, 919, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1159608082", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", 
     "algorithmic reduction-polynomial domain \[ScriptCapitalP]"},
   CellID->1159608082],
  Cell[3043175, 84826, 3075, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:847570831", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] \[Succeeds]"},
   CellID->847570831],
  Cell[3053761, 85111, 1013, 29, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2071048931", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->2071048931],
  Cell[3054777, 85142, 735, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:355437230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->355437230],
  Cell[3055515, 85163, 1077, 28, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1401739284", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1401739284],
  Cell[3056595, 85193, 904, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:29665786", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 1"},
   CellID->29665786],
  Cell[3057502, 85219, 1549, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:113282996", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 2"},
   CellID->113282996],
  Cell[3059054, 85264, 1683, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1675445982", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 3"},
   CellID->1675445982],
  Cell[3060740, 85315, 1557, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1968653014", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 4"},
   CellID->1968653014],
  Cell[3062300, 85361, 1241, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:860310855", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 5"},
   CellID->860310855],
  Cell[3063544, 85399, 1403, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2026748686", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] multiplier \[ScriptCapitalP]"},
   CellID->2026748686],
  Cell[3107555, 86507, 1857, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1529886137", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP]"},
   CellID->1529886137],
  Cell[3115030, 86725, 781, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1680577066", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "372"},
   CellID->1680577066],
  Cell[3134797, 87335, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1866021817", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "373"},
   CellID->1866021817],
  Cell[3138653, 87465, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1272742036", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "374"},
   CellID->1272742036],
  Cell[3142449, 87595, 1186, 29, 71, "GlobalDeclaration",
   CellTags->{
    "ID:181479986", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->181479986],
  Cell[3143638, 87626, 739, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1276235216", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1276235216],
  Cell[3144380, 87647, 896, 23, 58, "GlobalDeclaration",
   CellTags->{
    "ID:376128819", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->376128819],
  Cell[3145279, 87672, 1014, 27, 49, "GlobalDeclaration",
   CellTags->{
    "ID:973122334", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->973122334],
  Cell[3146757, 87718, 2509, 75, 93, "FormalTextInputFormula",
   CellTags->{
    "ID:1445738544", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.1"},
   CellID->1445738544],
  Cell[3149269, 87795, 2968, 80, 81, "FormalTextInputFormula",
   CellTags->{
    "ID:874376987", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.2"},
   CellID->874376987],
  Cell[3152240, 87877, 1880, 54, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:139929842", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.3"},
   CellID->139929842],
  Cell[3190555, 88777, 780, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:462890005", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375"},
   CellID->462890005],
  Cell[3215133, 89507, 1623, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:38279669", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducible lm"},
   CellID->38279669],
  Cell[3216759, 89558, 1475, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1376605467", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "red lm set"},
   CellID->1376605467],
  Cell[3218237, 89604, 3120, 87, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:673925715", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducible monomial"},
   CellID->673925715],
  Cell[3238019, 90141, 1876, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47923252", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence"},
   CellID->47923252],
  Cell[3240499, 90224, 739, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1632720588", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1632720588],
  Cell[3241241, 90245, 1075, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1771873303", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence \[NotEqual] 0"},
   CellID->1771873303],
  Cell[3242319, 90277, 1052, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:893563053", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lc 1"},
   CellID->893563053],
  Cell[3243374, 90308, 1054, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:587765202", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lp 1"},
   CellID->587765202],
  Cell[3244431, 90339, 991, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1254733295", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lc 2"},
   CellID->1254733295],
  Cell[3245425, 90368, 988, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:280696418", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lp 2"},
   CellID->280696418],
  Cell[3246416, 90397, 989, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1081126646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence subSeq"},
   CellID->1081126646],
  Cell[3247408, 90425, 973, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1101697023", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence subsequence"},
   CellID->1101697023],
  Cell[3275758, 91256, 3026, 89, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:182970451", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "110"},
   CellID->182970451],
  Cell[3278787, 91347, 3164, 92, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:845195690", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "111"},
   CellID->845195690],
  Cell[3300288, 91919, 739, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2081362811", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->2081362811],
  Cell[3301030, 91940, 4239, 120, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1173235103", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "112"},
   CellID->1173235103],
  Cell[3305272, 92062, 2318, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1805755328", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "113"},
   CellID->1805755328],
  Cell[3307593, 92134, 2490, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:596327945", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "114"},
   CellID->596327945],
  Cell[3336511, 92874, 2262, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1402642391", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "115"},
   CellID->1402642391],
  Cell[3346997, 93171, 1081, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:980978647", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Dickson-domain \[ScriptCapitalT]"},
   CellID->980978647],
  Cell[3348735, 93227, 756, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1408052308", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "116"},
   CellID->1408052308],
  Cell[3352664, 93359, 857, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1183157947", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isAlgoReductionRing \[ScriptCapitalP]"},
   CellID->1183157947],
  Cell[3360083, 93567, 2107, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1794188423", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", 
     "isAlgoReductionRing Conservation Theorem Polynomials"},
   CellID->1794188423]},
 "isCommPPDomain"->{
  Cell[64094, 1632, 6848, 193, 366, "FormalTextInputFormula",
   CellTags->{
    "ID:558626034", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCommPPDomain"},
   CellID->558626034]},
 "ID:1568169011"->{
  Cell[81000, 2096, 785, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1568169011", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1568169011]},
 "ID:1716984519"->{
  Cell[82274, 2133, 898, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1716984519", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1716984519]},
 "ID:205548127"->{
  Cell[83175, 2159, 2176, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:205548127", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial"},
   CellID->205548127]},
 "isMonomial"->{
  Cell[83175, 2159, 2176, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:205548127", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial"},
   CellID->205548127]},
 "ID:711054552"->{
  Cell[85687, 2236, 1836, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:711054552", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial II"},
   CellID->711054552]},
 "isMonomial II"->{
  Cell[85687, 2236, 1836, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:711054552", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial II"},
   CellID->711054552]},
 "ID:2035355406"->{
  Cell[87526, 2294, 2348, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2035355406", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "monomial constructor"},
   CellID->2035355406]},
 "monomial constructor"->{
  Cell[87526, 2294, 2348, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2035355406", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "monomial constructor"},
   CellID->2035355406]},
 "ID:371468988"->{
  Cell[89877, 2368, 1219, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371468988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "supp"},
   CellID->371468988]},
 "supp"->{
  Cell[89877, 2368, 1219, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371468988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "supp"},
   CellID->371468988]},
 "ID:741382791"->{
  Cell[91099, 2405, 2080, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:741382791", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isLP"},
   CellID->741382791]},
 "isLP"->{
  Cell[91099, 2405, 2080, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:741382791", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isLP"},
   CellID->741382791]},
 "ID:1470398456"->{
  Cell[93182, 2468, 2289, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1470398456", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp"},
   CellID->1470398456]},
 "lp"->{
  Cell[93182, 2468, 2289, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1470398456", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp"},
   CellID->1470398456]},
 "ID:1541081556"->{
  Cell[95474, 2539, 1075, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1541081556", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc"},
   CellID->1541081556]},
 "lc"->{
  Cell[95474, 2539, 1075, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1541081556", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc"},
   CellID->1541081556]},
 "ID:258866072"->{
  Cell[96552, 2571, 1204, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:258866072", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm"},
   CellID->258866072]},
 "lm"->{
  Cell[96552, 2571, 1204, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:258866072", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm"},
   CellID->258866072]},
 "ID:1500059087"->{
  Cell[97759, 2605, 1117, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1500059087", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R"},
   CellID->1500059087]},
 "R"->{
  Cell[97759, 2605, 1117, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1500059087", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R"},
   CellID->1500059087]},
 "ID:1301149083"->{
  Cell[98879, 2637, 1111, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1301149083", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 0"},
   CellID->1301149083]},
 "H 0"->{
  Cell[98879, 2637, 1111, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1301149083", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 0"},
   CellID->1301149083]},
 "ID:355792939"->{
  Cell[99993, 2667, 2271, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:355792939", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H"},
   CellID->355792939]},
 "H"->{
  Cell[99993, 2667, 2271, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:355792939", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H"},
   CellID->355792939]},
 "ID:1617463509"->{
  Cell[102267, 2734, 2760, 76, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1617463509", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "monomial multiplication"},
   CellID->1617463509]},
 "monomial multiplication"->{
  Cell[102267, 2734, 2760, 76, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1617463509", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "monomial multiplication"},
   CellID->1617463509]},
 "ID:1026275980"->{
  Cell[105637, 2835, 2611, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1026275980", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReductionPolynomialDomain"},
   CellID->1026275980]},
 "isReductionPolynomialDomain"->{
  Cell[105637, 2835, 2611, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1026275980", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReductionPolynomialDomain"},
   CellID->1026275980]},
 "ID:1335218902"->{
  Cell[108251, 2900, 4279, 121, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1335218902", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasDecisionPredicate"},
   CellID->1335218902]},
 "hasDecisionPredicate"->{
  Cell[108251, 2900, 4279, 121, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1335218902", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasDecisionPredicate"},
   CellID->1335218902]},
 "ID:1175095894"->{
  Cell[113212, 3046, 1778, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1175095894", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "containsMonomials"},
   CellID->1175095894]},
 "containsMonomials"->{
  Cell[113212, 3046, 1778, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1175095894", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "containsMonomials"},
   CellID->1175095894]},
 "ID:1967795228"->{
  Cell[114993, 3099, 6005, 158, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1967795228", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCoefficientFunction"},
   CellID->1967795228]},
 "hasCoefficientFunction"->{
  Cell[114993, 3099, 6005, 158, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1967795228", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCoefficientFunction"},
   CellID->1967795228]},
 "ID:2138030312"->{
  Cell[121001, 3259, 2813, 80, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2138030312", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplication"},
   CellID->2138030312]},
 "hasMultiplication"->{
  Cell[121001, 3259, 2813, 80, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2138030312", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplication"},
   CellID->2138030312]},
 "ID:1709544239"->{
  Cell[123817, 3341, 2210, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1709544239", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isExtensional"},
   CellID->1709544239]},
 "isExtensional"->{
  Cell[123817, 3341, 2210, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1709544239", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isExtensional"},
   CellID->1709544239]},
 "ID:235606987"->{
  Cell[126030, 3412, 2868, 86, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:235606987", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrdered"},
   CellID->235606987]},
 "isOrdered"->{
  Cell[126030, 3412, 2868, 86, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:235606987", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrdered"},
   CellID->235606987]},
 "ID:407906314"->{
  Cell[128901, 3500, 4757, 130, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:407906314", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultipliers"},
   CellID->407906314]},
 "hasMultipliers"->{
  Cell[128901, 3500, 4757, 130, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:407906314", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultipliers"},
   CellID->407906314]},
 "ID:1209025019"->{
  Cell[134279, 3656, 1697, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1209025019", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isAlgoReductionPolynomialDomain"},
   CellID->1209025019]},
 "isAlgoReductionPolynomialDomain"->{
  Cell[134279, 3656, 1697, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1209025019", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isAlgoReductionPolynomialDomain"},
   CellID->1209025019]},
 "ID:1424238209"->{
  Cell[135979, 3707, 7435, 185, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1424238209", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasLCRD"},
   CellID->1424238209]},
 "hasLCRD"->{
  Cell[135979, 3707, 7435, 185, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1424238209", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasLCRD"},
   CellID->1424238209]},
 "ID:1761006179"->{
  Cell[143417, 3894, 4286, 114, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1761006179", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasRDM"},
   CellID->1761006179]},
 "hasRDM"->{
  Cell[143417, 3894, 4286, 114, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1761006179", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasRDM"},
   CellID->1761006179]},
 "ID:667868750"->{
  Cell[147706, 4010, 7315, 181, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:667868750", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCPM"},
   CellID->667868750]},
 "hasCPM"->{
  Cell[147706, 4010, 7315, 181, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:667868750", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCPM"},
   CellID->667868750]},
 "ID:267444622"->{
  Cell[158726, 4282, 805, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:267444622", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "ring \[ScriptCapitalR]"},
   CellID->267444622]},
 "ring \[ScriptCapitalR]"->{
  Cell[158726, 4282, 805, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:267444622", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "ring \[ScriptCapitalR]"},
   CellID->267444622]},
 "ID:1725827561"->{
  Cell[159534, 4305, 899, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1725827561", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "power-product domain \[ScriptCapitalT]"},
   CellID->1725827561]},
 "power-product domain \[ScriptCapitalT]"->{
  Cell[159534, 4305, 899, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1725827561", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "power-product domain \[ScriptCapitalT]"},
   CellID->1725827561]},
 "ID:90242530"->{
  Cell[160436, 4330, 809, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:90242530", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isDomain \[ScriptCapitalP]"},
   CellID->90242530]},
 "isDomain \[ScriptCapitalP]"->{
  Cell[160436, 4330, 809, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:90242530", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isDomain \[ScriptCapitalP]"},
   CellID->90242530]},
 "ID:1949049350"->{
  Cell[161248, 4353, 856, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1949049350", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasDecisionPredicate \[ScriptCapitalP]"},
   CellID->1949049350]},
 "hasDecisionPredicate \[ScriptCapitalP]"->{
  Cell[161248, 4353, 856, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1949049350", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasDecisionPredicate \[ScriptCapitalP]"},
   CellID->1949049350]},
 "ID:378755333"->{
  Cell[162107, 4377, 835, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:378755333", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "containsMonomials \[ScriptCapitalP]"},
   CellID->378755333]},
 "containsMonomials \[ScriptCapitalP]"->{
  Cell[162107, 4377, 835, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:378755333", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "containsMonomials \[ScriptCapitalP]"},
   CellID->378755333]},
 "ID:226930466"->{
  Cell[162945, 4400, 859, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:226930466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCoefficientFunction \[ScriptCapitalP]"},
   CellID->226930466]},
 "hasCoefficientFunction \[ScriptCapitalP]"->{
  Cell[162945, 4400, 859, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:226930466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCoefficientFunction \[ScriptCapitalP]"},
   CellID->226930466]},
 "ID:221039322"->{
  Cell[163807, 4424, 848, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:221039322", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isExtensional \[ScriptCapitalP]"},
   CellID->221039322]},
 "isExtensional \[ScriptCapitalP]"->{
  Cell[163807, 4424, 848, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:221039322", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isExtensional \[ScriptCapitalP]"},
   CellID->221039322]},
 "ID:216808867"->{
  Cell[166245, 4500, 810, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:216808867", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isDomain \[ScriptCapitalT]"},
   CellID->216808867]},
 "isDomain \[ScriptCapitalT]"->{
  Cell[166245, 4500, 810, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:216808867", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isDomain \[ScriptCapitalT]"},
   CellID->216808867]},
 "ID:1391060217"->{
  Cell[167058, 4523, 1082, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1391060217", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isPartIrreflOrder \[Succeeds] \[ScriptCapitalT]"},
   CellID->1391060217]},
 "isPartIrreflOrder \[Succeeds] \[ScriptCapitalT]"->{
  Cell[167058, 4523, 1082, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1391060217", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isPartIrreflOrder \[Succeeds] \[ScriptCapitalT]"},
   CellID->1391060217]},
 "ID:1131450960"->{
  Cell[168143, 4557, 2201, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1131450960", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Noetherian induction on \[ScriptCapitalT]", "IR"},
   CellID->1131450960]},
 "Noetherian induction on \[ScriptCapitalT]"->{
  Cell[168143, 4557, 2201, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1131450960", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Noetherian induction on \[ScriptCapitalT]", "IR"},
   CellID->1131450960]},
 "IR"->{
  Cell[168143, 4557, 2201, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1131450960", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Noetherian induction on \[ScriptCapitalT]", "IR"},
   CellID->1131450960],
  Cell[170347, 4629, 905, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:953344354", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 1 \[ScriptCapitalT]", "IR"},
   CellID->953344354],
  Cell[260348, 7273, 926, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1101510167", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lcm", "IR"},
   CellID->1101510167],
  Cell[319672, 8998, 1203, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:938132667", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership supp", "IR"},
   CellID->938132667],
  Cell[320878, 9031, 1747, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:56177686", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Element] supp", "IR"},
   CellID->56177686],
  Cell[322628, 9087, 867, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:163445395", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 0", "IR"},
   CellID->163445395],
  Cell[323498, 9113, 868, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:671805226", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 1", "IR"},
   CellID->671805226],
  Cell[324369, 9139, 1292, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:373967346", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership C", "IR"},
   CellID->373967346],
  Cell[368770, 10398, 1246, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:292606918", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomials", "IR"},
   CellID->292606918],
  Cell[370908, 10460, 950, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:406494450", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lp", "IR"},
   CellID->406494450],
  Cell[371861, 10488, 955, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1454724371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lc", "IR"},
   CellID->1454724371],
  Cell[372819, 10516, 955, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1589530505", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lm", "IR"},
   CellID->1589530505],
  Cell[426765, 12077, 928, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:829207957", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership + \[ScriptCapitalP]", "IR"},
   CellID->829207957],
  Cell[427696, 12104, 929, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:227196886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership - \[ScriptCapitalP] 1", "IR"},
   CellID->227196886],
  Cell[428628, 12131, 931, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414658623", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership - \[ScriptCapitalP] 2", "IR"},
   CellID->414658623],
  Cell[446279, 12664, 907, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:144989886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "non-zero 2", "IR"},
   CellID->144989886],
  Cell[452206, 12847, 1488, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:996448533", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "non-zero 3", "IR"},
   CellID->996448533],
  Cell[689322, 19828, 944, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:313886646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership R", "IR"},
   CellID->313886646],
  Cell[776777, 22324, 917, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:884488101", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership H", "IR"},
   CellID->884488101],
  Cell[903714, 25836, 1918, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502331257", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "induction on \[ScriptCapitalP]", "IR"},
   CellID->1502331257],
  Cell[953429, 27213, 987, 27, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1903131381", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomial multiplication", "IR"},
   CellID->1903131381]},
 "ID:953344354"->{
  Cell[170347, 4629, 905, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:953344354", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 1 \[ScriptCapitalT]", "IR"},
   CellID->953344354]},
 "membership 1 \[ScriptCapitalT]"->{
  Cell[170347, 4629, 905, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:953344354", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 1 \[ScriptCapitalT]", "IR"},
   CellID->953344354]},
 "ID:1367902825"->{
  Cell[171783, 4672, 1558, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1367902825", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 1"},
   CellID->1367902825]},
 "PP 1"->{
  Cell[171783, 4672, 1558, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1367902825", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 1"},
   CellID->1367902825]},
 "ID:643177315"->{
  Cell[173344, 4723, 761, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:643177315", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->643177315]},
 "ID:1568022983"->{
  Cell[174108, 4745, 916, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1568022983", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 1 min"},
   CellID->1568022983]},
 "PP 1 min"->{
  Cell[174108, 4745, 916, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1568022983", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 1 min"},
   CellID->1568022983]},
 "ID:1723905914"->{
  Cell[175027, 4772, 1318, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1723905914", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 0"},
   CellID->1723905914]},
 "PP 0"->{
  Cell[175027, 4772, 1318, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1723905914", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 0"},
   CellID->1723905914]},
 "ID:1825953062"->{
  Cell[176348, 4814, 883, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1825953062", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 2'"},
   CellID->1825953062]},
 "PP 2'"->{
  Cell[176348, 4814, 883, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1825953062", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 2'"},
   CellID->1825953062]},
 "ID:596582273"->{
  Cell[177234, 4841, 879, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:596582273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 2"},
   CellID->596582273]},
 "PP 2"->{
  Cell[177234, 4841, 879, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:596582273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 2"},
   CellID->596582273]},
 "ID:389623087"->{
  Cell[178116, 4868, 901, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:389623087", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 3'"},
   CellID->389623087]},
 "PP 3'"->{
  Cell[178116, 4868, 901, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:389623087", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 3'"},
   CellID->389623087]},
 "ID:885344283"->{
  Cell[179020, 4896, 899, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:885344283", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 3"},
   CellID->885344283]},
 "PP 3"->{
  Cell[179020, 4896, 899, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:885344283", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 3"},
   CellID->885344283]},
 "ID:1033740528"->{
  Cell[179922, 4924, 1017, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1033740528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 4"},
   CellID->1033740528]},
 "PP 4"->{
  Cell[179922, 4924, 1017, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1033740528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 4"},
   CellID->1033740528]},
 "ID:475932409"->{
  Cell[180942, 4956, 738, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:475932409", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->475932409]},
 "ID:1124160616"->{
  Cell[181908, 4984, 1218, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1124160616", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 5"},
   CellID->1124160616]},
 "PP 5"->{
  Cell[181908, 4984, 1218, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1124160616", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 5"},
   CellID->1124160616]},
 "ID:1594052474"->{
  Cell[183129, 5023, 1433, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1594052474", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 6"},
   CellID->1594052474]},
 "PP 6"->{
  Cell[183129, 5023, 1433, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1594052474", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 6"},
   CellID->1594052474]},
 "ID:1091730754"->{
  Cell[184565, 5068, 1470, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1091730754", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 7"},
   CellID->1091730754]},
 "PP 7"->{
  Cell[184565, 5068, 1470, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1091730754", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 7"},
   CellID->1091730754]},
 "ID:1114122233"->{
  Cell[186038, 5116, 1268, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1114122233", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 8"},
   CellID->1114122233]},
 "PP 8"->{
  Cell[186038, 5116, 1268, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1114122233", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 8"},
   CellID->1114122233]},
 "ID:1909427587"->{
  Cell[187309, 5156, 1033, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1909427587", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isPartReflOrder \[VerticalBar]"},
   CellID->1909427587]},
 "isPartReflOrder \[VerticalBar]"->{
  Cell[187309, 5156, 1033, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1909427587", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isPartReflOrder \[VerticalBar]"},
   CellID->1909427587]},
 "ID:196272597"->{
  Cell[188345, 5188, 1214, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:196272597", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrderEmbedding \[RawStar] \[VerticalBar]"},
   CellID->196272597]},
 "isOrderEmbedding \[RawStar] \[VerticalBar]"->{
  Cell[188345, 5188, 1214, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:196272597", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrderEmbedding \[RawStar] \[VerticalBar]"},
   CellID->196272597]},
 "Proof|ID:216808867"->{
  Cell[189717, 5237, 1788, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:216808867", "Proof|ID:216808867-1"},
   CellID->77023573]},
 "Proof|ID:216808867-1"->{
  Cell[189717, 5237, 1788, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:216808867", "Proof|ID:216808867-1"},
   CellID->77023573]},
 "Proof|ID:1391060217"->{
  Cell[192024, 5312, 2051, 63, 70, "ProofInfo",
   CellTags->{"Proof|ID:1391060217", "Proof|ID:1391060217-1"},
   CellID->640802005]},
 "Proof|ID:1391060217-1"->{
  Cell[192024, 5312, 2051, 63, 70, "ProofInfo",
   CellTags->{"Proof|ID:1391060217", "Proof|ID:1391060217-1"},
   CellID->640802005]},
 "Proof|ID:1131450960"->{
  Cell[194639, 5395, 5088, 131, 70, "ProofInfo",
   CellTags->{"Proof|ID:1131450960", "Proof|ID:1131450960-1"},
   CellID->381161681]},
 "Proof|ID:1131450960-1"->{
  Cell[194639, 5395, 5088, 131, 70, "ProofInfo",
   CellTags->{"Proof|ID:1131450960", "Proof|ID:1131450960-1"},
   CellID->381161681]},
 "Proof|ID:953344354"->{
  Cell[200248, 5545, 1947, 59, 70, "ProofInfo",
   CellTags->{"Proof|ID:953344354", "Proof|ID:953344354-1"},
   CellID->238169741]},
 "Proof|ID:953344354-1"->{
  Cell[200248, 5545, 1947, 59, 70, "ProofInfo",
   CellTags->{"Proof|ID:953344354", "Proof|ID:953344354-1"},
   CellID->238169741]},
 "Proof|ID:1367902825"->{
  Cell[202715, 5623, 3889, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:1367902825", "Proof|ID:1367902825-1"},
   CellID->640235643]},
 "Proof|ID:1367902825-1"->{
  Cell[202715, 5623, 3889, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:1367902825", "Proof|ID:1367902825-1"},
   CellID->640235643]},
 "Proof|ID:1568022983"->{
  Cell[207126, 5749, 2794, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:1568022983", "Proof|ID:1568022983-1"},
   CellID->1849364148]},
 "Proof|ID:1568022983-1"->{
  Cell[207126, 5749, 2794, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:1568022983", "Proof|ID:1568022983-1"},
   CellID->1849364148]},
 "Proof|ID:1723905914"->{
  Cell[210441, 5850, 3822, 103, 70, "ProofInfo",
   CellTags->{"Proof|ID:1723905914", "Proof|ID:1723905914-1"},
   CellID->818368910]},
 "Proof|ID:1723905914-1"->{
  Cell[210441, 5850, 3822, 103, 70, "ProofInfo",
   CellTags->{"Proof|ID:1723905914", "Proof|ID:1723905914-1"},
   CellID->818368910]},
 "Proof|ID:1825953062"->{
  Cell[214785, 5972, 3050, 87, 70, "ProofInfo",
   CellTags->{"Proof|ID:1825953062", "Proof|ID:1825953062-1"},
   CellID->10445487]},
 "Proof|ID:1825953062-1"->{
  Cell[214785, 5972, 3050, 87, 70, "ProofInfo",
   CellTags->{"Proof|ID:1825953062", "Proof|ID:1825953062-1"},
   CellID->10445487]},
 "Proof|ID:596582273"->{
  Cell[218357, 6078, 3040, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:596582273", "Proof|ID:596582273-1"},
   CellID->525784934]},
 "Proof|ID:596582273-1"->{
  Cell[218357, 6078, 3040, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:596582273", "Proof|ID:596582273-1"},
   CellID->525784934]},
 "Proof|ID:389623087"->{
  Cell[221918, 6183, 3202, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:389623087", "Proof|ID:389623087-1"},
   CellID->601121285]},
 "Proof|ID:389623087-1"->{
  Cell[221918, 6183, 3202, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:389623087", "Proof|ID:389623087-1"},
   CellID->601121285]},
 "Proof|ID:885344283"->{
  Cell[225640, 6292, 3200, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:885344283", "Proof|ID:885344283-1"},
   CellID->573214883]},
 "Proof|ID:885344283-1"->{
  Cell[225640, 6292, 3200, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:885344283", "Proof|ID:885344283-1"},
   CellID->573214883]},
 "Proof|ID:1033740528"->{
  Cell[229358, 6401, 3592, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1033740528", "Proof|ID:1033740528-1"},
   CellID->1813081668]},
 "Proof|ID:1033740528-1"->{
  Cell[229358, 6401, 3592, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1033740528", "Proof|ID:1033740528-1"},
   CellID->1813081668]},
 "Proof|ID:1124160616"->{
  Cell[233470, 6519, 3939, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:1124160616", "Proof|ID:1124160616-1"},
   CellID->160370785]},
 "Proof|ID:1124160616-1"->{
  Cell[233470, 6519, 3939, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:1124160616", "Proof|ID:1124160616-1"},
   CellID->160370785]},
 "Proof|ID:1594052474"->{
  Cell[237929, 6645, 4737, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1594052474", "Proof|ID:1594052474-1"},
   CellID->1673583185]},
 "Proof|ID:1594052474-1"->{
  Cell[237929, 6645, 4737, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1594052474", "Proof|ID:1594052474-1"},
   CellID->1673583185]},
 "Proof|ID:1091730754"->{
  Cell[243328, 6786, 4848, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:1091730754", "Proof|ID:1091730754-1"},
   CellID->347735719]},
 "Proof|ID:1091730754-1"->{
  Cell[243328, 6786, 4848, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:1091730754", "Proof|ID:1091730754-1"},
   CellID->347735719]},
 "Proof|ID:1114122233"->{
  Cell[248739, 6931, 4600, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1114122233", "Proof|ID:1114122233-1"},
   CellID->1981139454]},
 "Proof|ID:1114122233-1"->{
  Cell[248739, 6931, 4600, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1114122233", "Proof|ID:1114122233-1"},
   CellID->1981139454]},
 "Proof|ID:1909427587"->{
  Cell[253860, 7069, 2011, 62, 70, "ProofInfo",
   CellTags->{"Proof|ID:1909427587", "Proof|ID:1909427587-1"},
   CellID->106187856]},
 "Proof|ID:1909427587-1"->{
  Cell[253860, 7069, 2011, 62, 70, "ProofInfo",
   CellTags->{"Proof|ID:1909427587", "Proof|ID:1909427587-1"},
   CellID->106187856]},
 "Proof|ID:196272597"->{
  Cell[256393, 7150, 2136, 64, 70, "ProofInfo",
   CellTags->{"Proof|ID:196272597", "Proof|ID:196272597-1"},
   CellID->626133587]},
 "Proof|ID:196272597-1"->{
  Cell[256393, 7150, 2136, 64, 70, "ProofInfo",
   CellTags->{"Proof|ID:196272597", "Proof|ID:196272597-1"},
   CellID->626133587]},
 "ID:334245933"->{
  Cell[259584, 7251, 761, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:334245933", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->334245933]},
 "ID:1101510167"->{
  Cell[260348, 7273, 926, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1101510167", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lcm", "IR"},
   CellID->1101510167]},
 "membership lcm"->{
  Cell[260348, 7273, 926, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1101510167", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lcm", "IR"},
   CellID->1101510167]},
 "ID:571674285"->{
  Cell[261650, 7309, 885, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:571674285", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 1"},
   CellID->571674285]},
 "lcm 1"->{
  Cell[261650, 7309, 885, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:571674285", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 1"},
   CellID->571674285]},
 "ID:775675057"->{
  Cell[262538, 7335, 889, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:775675057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 1'"},
   CellID->775675057]},
 "lcm 1'"->{
  Cell[262538, 7335, 889, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:775675057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 1'"},
   CellID->775675057]},
 "ID:517355459"->{
  Cell[263430, 7361, 1471, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:517355459", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 2"},
   CellID->517355459]},
 "lcm 2"->{
  Cell[263430, 7361, 1471, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:517355459", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 2"},
   CellID->517355459]},
 "ID:651227563"->{
  Cell[264904, 7408, 936, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:651227563", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 3"},
   CellID->651227563]},
 "lcm 3"->{
  Cell[264904, 7408, 936, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:651227563", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 3"},
   CellID->651227563]},
 "ID:482943675"->{
  Cell[265843, 7436, 844, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:482943675", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 6"},
   CellID->482943675]},
 "lcm 6"->{
  Cell[265843, 7436, 844, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:482943675", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 6"},
   CellID->482943675]},
 "ID:39860588"->{
  Cell[266899, 7469, 906, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:39860588", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 4"},
   CellID->39860588]},
 "lcm 4"->{
  Cell[266899, 7469, 906, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:39860588", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 4"},
   CellID->39860588]},
 "ID:1631465634"->{
  Cell[267808, 7496, 911, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1631465634", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 4'"},
   CellID->1631465634]},
 "lcm 4'"->{
  Cell[267808, 7496, 911, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1631465634", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 4'"},
   CellID->1631465634]},
 "ID:835022497"->{
  Cell[268722, 7523, 1608, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:835022497", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 5"},
   CellID->835022497]},
 "lcm 5"->{
  Cell[268722, 7523, 1608, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:835022497", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 5"},
   CellID->835022497]},
 "Proof|ID:1101510167"->{
  Cell[270487, 7585, 3057, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:1101510167", "Proof|ID:1101510167-1"},
   CellID->1779125069]},
 "Proof|ID:1101510167-1"->{
  Cell[270487, 7585, 3057, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:1101510167", "Proof|ID:1101510167-1"},
   CellID->1779125069]},
 "Proof|ID:571674285"->{
  Cell[274066, 7689, 3119, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:571674285", "Proof|ID:571674285-1"},
   CellID->348779623]},
 "Proof|ID:571674285-1"->{
  Cell[274066, 7689, 3119, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:571674285", "Proof|ID:571674285-1"},
   CellID->348779623]},
 "Proof|ID:775675057"->{
  Cell[277706, 7794, 3122, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:775675057", "Proof|ID:775675057-1"},
   CellID->1055462613]},
 "Proof|ID:775675057-1"->{
  Cell[277706, 7794, 3122, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:775675057", "Proof|ID:775675057-1"},
   CellID->1055462613]},
 "Proof|ID:517355459"->{
  Cell[281349, 7899, 4413, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:517355459", "Proof|ID:517355459-1"},
   CellID->1760053657]},
 "Proof|ID:517355459-1"->{
  Cell[281349, 7899, 4413, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:517355459", "Proof|ID:517355459-1"},
   CellID->1760053657]},
 "Proof|ID:651227563"->{
  Cell[286282, 8032, 3362, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:651227563", "Proof|ID:651227563-1"},
   CellID->1015477485]},
 "Proof|ID:651227563-1"->{
  Cell[286282, 8032, 3362, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:651227563", "Proof|ID:651227563-1"},
   CellID->1015477485]},
 "Proof|ID:482943675"->{
  Cell[290165, 8143, 2705, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:482943675", "Proof|ID:482943675-1"},
   CellID->29218054]},
 "Proof|ID:482943675-1"->{
  Cell[290165, 8143, 2705, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:482943675", "Proof|ID:482943675-1"},
   CellID->29218054]},
 "Proof|ID:39860588"->{
  Cell[293389, 8242, 3460, 95, 70, "ProofInfo",
   CellTags->{"Proof|ID:39860588", "Proof|ID:39860588-1"},
   CellID->2016600132]},
 "Proof|ID:39860588-1"->{
  Cell[293389, 8242, 3460, 95, 70, "ProofInfo",
   CellTags->{"Proof|ID:39860588", "Proof|ID:39860588-1"},
   CellID->2016600132]},
 "Proof|ID:1631465634"->{
  Cell[297368, 8356, 3476, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:1631465634", "Proof|ID:1631465634-1"},
   CellID->1760483406]},
 "Proof|ID:1631465634-1"->{
  Cell[297368, 8356, 3476, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:1631465634", "Proof|ID:1631465634-1"},
   CellID->1760483406]},
 "Proof|ID:835022497"->{
  Cell[301366, 8471, 4774, 122, 70, "ProofInfo",
   CellTags->{"Proof|ID:835022497", "Proof|ID:835022497-1"},
   CellID->248308182]},
 "Proof|ID:835022497-1"->{
  Cell[301366, 8471, 4774, 122, 70, "ProofInfo",
   CellTags->{"Proof|ID:835022497", "Proof|ID:835022497-1"},
   CellID->248308182]},
 "ID:100863090"->{
  Cell[307311, 8634, 736, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:100863090", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->100863090]},
 "ID:292726431"->{
  Cell[308438, 8668, 982, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:292726431", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership supp general"},
   CellID->292726431]},
 "membership supp general"->{
  Cell[308438, 8668, 982, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:292726431", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership supp general"},
   CellID->292726431]},
 "ID:109557753"->{
  Cell[309423, 8697, 1636, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:109557753", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Element] supp general"},
   CellID->109557753]},
 "\[Element] supp general"->{
  Cell[309423, 8697, 1636, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:109557753", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Element] supp general"},
   CellID->109557753]},
 "Proof|ID:292726431"->{
  Cell[311215, 8758, 2703, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:292726431", "Proof|ID:292726431-1"},
   CellID->837210885]},
 "Proof|ID:292726431-1"->{
  Cell[311215, 8758, 2703, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:292726431", "Proof|ID:292726431-1"},
   CellID->837210885]},
 "Proof|ID:109557753"->{
  Cell[314438, 8853, 4231, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:109557753", "Proof|ID:109557753-1"},
   CellID->474399008]},
 "Proof|ID:109557753-1"->{
  Cell[314438, 8853, 4231, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:109557753", "Proof|ID:109557753-1"},
   CellID->474399008]},
 "ID:938132667"->{
  Cell[319672, 8998, 1203, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:938132667", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership supp", "IR"},
   CellID->938132667]},
 "membership supp"->{
  Cell[319672, 8998, 1203, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:938132667", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership supp", "IR"},
   CellID->938132667]},
 "ID:56177686"->{
  Cell[320878, 9031, 1747, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:56177686", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Element] supp", "IR"},
   CellID->56177686]},
 "\[Element] supp"->{
  Cell[320878, 9031, 1747, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:56177686", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Element] supp", "IR"},
   CellID->56177686]},
 "ID:163445395"->{
  Cell[322628, 9087, 867, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:163445395", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 0", "IR"},
   CellID->163445395]},
 "membership 0"->{
  Cell[322628, 9087, 867, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:163445395", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 0", "IR"},
   CellID->163445395]},
 "ID:671805226"->{
  Cell[323498, 9113, 868, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:671805226", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 1", "IR"},
   CellID->671805226]},
 "membership 1"->{
  Cell[323498, 9113, 868, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:671805226", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 1", "IR"},
   CellID->671805226]},
 "ID:373967346"->{
  Cell[324369, 9139, 1292, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:373967346", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership C", "IR"},
   CellID->373967346]},
 "membership C"->{
  Cell[324369, 9139, 1292, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:373967346", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership C", "IR"},
   CellID->373967346]},
 "Proof|ID:938132667"->{
  Cell[325817, 9186, 2450, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:938132667", "Proof|ID:938132667-1"},
   CellID->22005555]},
 "Proof|ID:938132667-1"->{
  Cell[325817, 9186, 2450, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:938132667", "Proof|ID:938132667-1"},
   CellID->22005555]},
 "Proof|ID:56177686"->{
  Cell[328787, 9277, 4070, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:56177686", "Proof|ID:56177686-1"},
   CellID->520191277]},
 "Proof|ID:56177686-1"->{
  Cell[328787, 9277, 4070, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:56177686", "Proof|ID:56177686-1"},
   CellID->520191277]},
 "Proof|ID:163445395"->{
  Cell[333376, 9404, 1902, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:163445395", "Proof|ID:163445395-1"},
   CellID->25655089]},
 "Proof|ID:163445395-1"->{
  Cell[333376, 9404, 1902, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:163445395", "Proof|ID:163445395-1"},
   CellID->25655089]},
 "Proof|ID:671805226"->{
  Cell[335797, 9481, 1903, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:671805226", "Proof|ID:671805226-1"},
   CellID->637038652]},
 "Proof|ID:671805226-1"->{
  Cell[335797, 9481, 1903, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:671805226", "Proof|ID:671805226-1"},
   CellID->637038652]},
 "Proof|ID:373967346"->{
  Cell[338219, 9558, 3039, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:373967346", "Proof|ID:373967346-1"},
   CellID->278031390]},
 "Proof|ID:373967346-1"->{
  Cell[338219, 9558, 3039, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:373967346", "Proof|ID:373967346-1"},
   CellID->278031390]},
 "ID:1248732928"->{
  Cell[342401, 9679, 2823, 76, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1248732928", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Monomial"},
   CellID->1248732928]},
 "Monomial"->{
  Cell[342401, 9679, 2823, 76, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1248732928", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Monomial"},
   CellID->1248732928]},
 "ID:1277602733"->{
  Cell[345227, 9757, 2152, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1277602733", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp sufficient"},
   CellID->1277602733]},
 "lp sufficient"->{
  Cell[345227, 9757, 2152, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1277602733", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp sufficient"},
   CellID->1277602733]},
 "ID:1270225306"->{
  Cell[347382, 9820, 2557, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1270225306", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp alternative"},
   CellID->1270225306]},
 "lp alternative"->{
  Cell[347382, 9820, 2557, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1270225306", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp alternative"},
   CellID->1270225306]},
 "Proof|ID:1248732928"->{
  Cell[350095, 9904, 5407, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:1248732928", "Proof|ID:1248732928-1"},
   CellID->328071702]},
 "Proof|ID:1248732928-1"->{
  Cell[350095, 9904, 5407, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:1248732928", "Proof|ID:1248732928-1"},
   CellID->328071702]},
 "Proof|ID:1277602733"->{
  Cell[356023, 10060, 5075, 131, 70, "ProofInfo",
   CellTags->{"Proof|ID:1277602733", "Proof|ID:1277602733-1"},
   CellID->88373483]},
 "Proof|ID:1277602733-1"->{
  Cell[356023, 10060, 5075, 131, 70, "ProofInfo",
   CellTags->{"Proof|ID:1277602733", "Proof|ID:1277602733-1"},
   CellID->88373483]},
 "Proof|ID:1270225306"->{
  Cell[361619, 10210, 6046, 151, 70, "ProofInfo",
   CellTags->{"Proof|ID:1270225306", "Proof|ID:1270225306-1"},
   CellID->426200985]},
 "Proof|ID:1270225306-1"->{
  Cell[361619, 10210, 6046, 151, 70, "ProofInfo",
   CellTags->{"Proof|ID:1270225306", "Proof|ID:1270225306-1"},
   CellID->426200985]},
 "ID:292606918"->{
  Cell[368770, 10398, 1246, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:292606918", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomials", "IR"},
   CellID->292606918]},
 "membership monomials"->{
  Cell[368770, 10398, 1246, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:292606918", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomials", "IR"},
   CellID->292606918]},
 "ID:485824967"->{
  Cell[370019, 10435, 886, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:485824967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->485824967]},
 "ID:406494450"->{
  Cell[370908, 10460, 950, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:406494450", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lp", "IR"},
   CellID->406494450]},
 "membership lp"->{
  Cell[370908, 10460, 950, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:406494450", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lp", "IR"},
   CellID->406494450]},
 "ID:1454724371"->{
  Cell[371861, 10488, 955, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1454724371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lc", "IR"},
   CellID->1454724371]},
 "membership lc"->{
  Cell[371861, 10488, 955, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1454724371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lc", "IR"},
   CellID->1454724371]},
 "ID:1589530505"->{
  Cell[372819, 10516, 955, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1589530505", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lm", "IR"},
   CellID->1589530505]},
 "membership lm"->{
  Cell[372819, 10516, 955, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1589530505", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lm", "IR"},
   CellID->1589530505]},
 "Proof|ID:292606918"->{
  Cell[373931, 10553, 2992, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:292606918", "Proof|ID:292606918-1"},
   CellID->1053213110]},
 "Proof|ID:292606918-1"->{
  Cell[373931, 10553, 2992, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:292606918", "Proof|ID:292606918-1"},
   CellID->1053213110]},
 "Proof|ID:406494450"->{
  Cell[377443, 10654, 2943, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:406494450", "Proof|ID:406494450-1"},
   CellID->928711222]},
 "Proof|ID:406494450-1"->{
  Cell[377443, 10654, 2943, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:406494450", "Proof|ID:406494450-1"},
   CellID->928711222]},
 "Proof|ID:1454724371"->{
  Cell[380907, 10755, 2952, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1454724371", "Proof|ID:1454724371-1"},
   CellID->776114793]},
 "Proof|ID:1454724371-1"->{
  Cell[380907, 10755, 2952, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1454724371", "Proof|ID:1454724371-1"},
   CellID->776114793]},
 "Proof|ID:1589530505"->{
  Cell[384381, 10857, 2953, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1589530505", "Proof|ID:1589530505-1"},
   CellID->1436300644]},
 "Proof|ID:1589530505-1"->{
  Cell[384381, 10857, 2953, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1589530505", "Proof|ID:1589530505-1"},
   CellID->1436300644]},
 "ID:720209465"->{
  Cell[388594, 10987, 740, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:720209465", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->720209465]},
 "ID:1852885314"->{
  Cell[389337, 11008, 915, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1852885314", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 0"},
   CellID->1852885314]},
 "C 0"->{
  Cell[389337, 11008, 915, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1852885314", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 0"},
   CellID->1852885314]},
 "ID:148341425"->{
  Cell[390255, 11035, 761, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:148341425", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->148341425]},
 "ID:1443097690"->{
  Cell[391019, 11057, 1263, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1443097690", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 1"},
   CellID->1443097690]},
 "C 1"->{
  Cell[391019, 11057, 1263, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1443097690", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 1"},
   CellID->1443097690]},
 "ID:1675126992"->{
  Cell[392285, 11095, 1055, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1675126992", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 2"},
   CellID->1675126992]},
 "C 2"->{
  Cell[392285, 11095, 1055, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1675126992", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 2"},
   CellID->1675126992]},
 "ID:814027629"->{
  Cell[393343, 11128, 1260, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:814027629", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 3"},
   CellID->814027629]},
 "C 3"->{
  Cell[393343, 11128, 1260, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:814027629", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 3"},
   CellID->814027629]},
 "ID:1597798077"->{
  Cell[394606, 11166, 1296, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1597798077", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 4"},
   CellID->1597798077]},
 "C 4"->{
  Cell[394606, 11166, 1296, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1597798077", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 4"},
   CellID->1597798077]},
 "ID:1711645326"->{
  Cell[395905, 11206, 1596, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1711645326", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 5"},
   CellID->1711645326]},
 "C 5"->{
  Cell[395905, 11206, 1596, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1711645326", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 5"},
   CellID->1711645326]},
 "Proof|ID:1852885314"->{
  Cell[397657, 11264, 2727, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:1852885314", "Proof|ID:1852885314-1"},
   CellID->131831930]},
 "Proof|ID:1852885314-1"->{
  Cell[397657, 11264, 2727, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:1852885314", "Proof|ID:1852885314-1"},
   CellID->131831930]},
 "Proof|ID:1443097690"->{
  Cell[400903, 11364, 4602, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1443097690", "Proof|ID:1443097690-1"},
   CellID->324739814]},
 "Proof|ID:1443097690-1"->{
  Cell[400903, 11364, 4602, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1443097690", "Proof|ID:1443097690-1"},
   CellID->324739814]},
 "Proof|ID:1675126992"->{
  Cell[406025, 11502, 3766, 101, 70, "ProofInfo",
   CellTags->{"Proof|ID:1675126992", "Proof|ID:1675126992-1"},
   CellID->947152217]},
 "Proof|ID:1675126992-1"->{
  Cell[406025, 11502, 3766, 101, 70, "ProofInfo",
   CellTags->{"Proof|ID:1675126992", "Proof|ID:1675126992-1"},
   CellID->947152217]},
 "Proof|ID:814027629"->{
  Cell[410312, 11622, 4591, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:814027629", "Proof|ID:814027629-1"},
   CellID->1077213]},
 "Proof|ID:814027629-1"->{
  Cell[410312, 11622, 4591, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:814027629", "Proof|ID:814027629-1"},
   CellID->1077213]},
 "Proof|ID:1597798077"->{
  Cell[415423, 11759, 4096, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:1597798077", "Proof|ID:1597798077-1"},
   CellID->1370021543]},
 "Proof|ID:1597798077-1"->{
  Cell[415423, 11759, 4096, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:1597798077", "Proof|ID:1597798077-1"},
   CellID->1370021543]},
 "Proof|ID:1711645326"->{
  Cell[420040, 11887, 4712, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:1711645326", "Proof|ID:1711645326-1"},
   CellID->209179501]},
 "Proof|ID:1711645326-1"->{
  Cell[420040, 11887, 4712, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:1711645326", "Proof|ID:1711645326-1"},
   CellID->209179501]},
 "ID:1053827266"->{
  Cell[425998, 12055, 764, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1053827266", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1053827266]},
 "ID:829207957"->{
  Cell[426765, 12077, 928, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:829207957", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership + \[ScriptCapitalP]", "IR"},
   CellID->829207957]},
 "membership + \[ScriptCapitalP]"->{
  Cell[426765, 12077, 928, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:829207957", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership + \[ScriptCapitalP]", "IR"},
   CellID->829207957]},
 "ID:227196886"->{
  Cell[427696, 12104, 929, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:227196886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership - \[ScriptCapitalP] 1", "IR"},
   CellID->227196886]},
 "membership - \[ScriptCapitalP] 1"->{
  Cell[427696, 12104, 929, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:227196886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership - \[ScriptCapitalP] 1", "IR"},
   CellID->227196886]},
 "ID:414658623"->{
  Cell[428628, 12131, 931, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414658623", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership - \[ScriptCapitalP] 2", "IR"},
   CellID->414658623]},
 "membership - \[ScriptCapitalP] 2"->{
  Cell[428628, 12131, 931, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414658623", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership - \[ScriptCapitalP] 2", "IR"},
   CellID->414658623]},
 "Proof|ID:829207957"->{
  Cell[429715, 12167, 2997, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:829207957", "Proof|ID:829207957-1"},
   CellID->188209562]},
 "Proof|ID:829207957-1"->{
  Cell[429715, 12167, 2997, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:829207957", "Proof|ID:829207957-1"},
   CellID->188209562]},
 "Proof|ID:227196886"->{
  Cell[433232, 12269, 2614, 75, 70, "ProofInfo",
   CellTags->{"Proof|ID:227196886", "Proof|ID:227196886-1"},
   CellID->65751116]},
 "Proof|ID:227196886-1"->{
  Cell[433232, 12269, 2614, 75, 70, "ProofInfo",
   CellTags->{"Proof|ID:227196886", "Proof|ID:227196886-1"},
   CellID->65751116]},
 "Proof|ID:414658623"->{
  Cell[436365, 12363, 3000, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:414658623", "Proof|ID:414658623-1"},
   CellID->19838666]},
 "Proof|ID:414658623-1"->{
  Cell[436365, 12363, 3000, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:414658623", "Proof|ID:414658623-1"},
   CellID->19838666]},
 "ID:314696779"->{
  Cell[440323, 12479, 1210, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:314696779", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1=0 \[ScriptCapitalP]"},
   CellID->314696779]},
 "1=0 \[ScriptCapitalP]"->{
  Cell[440323, 12479, 1210, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:314696779", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1=0 \[ScriptCapitalP]"},
   CellID->314696779]},
 "ID:849229881"->{
  Cell[441536, 12517, 1287, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:849229881", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 1"},
   CellID->849229881]},
 "lp 1"->{
  Cell[441536, 12517, 1287, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:849229881", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 1"},
   CellID->849229881]},
 "ID:406483636"->{
  Cell[442826, 12558, 1291, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:406483636", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 1"},
   CellID->406483636]},
 "lc 1"->{
  Cell[442826, 12558, 1291, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:406483636", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 1"},
   CellID->406483636]},
 "ID:1767085150"->{
  Cell[444120, 12599, 1241, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1767085150", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm 1"},
   CellID->1767085150]},
 "lm 1"->{
  Cell[444120, 12599, 1241, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1767085150", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm 1"},
   CellID->1767085150]},
 "ID:1272860149"->{
  Cell[445364, 12639, 912, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1272860149", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1272860149]},
 "ID:144989886"->{
  Cell[446279, 12664, 907, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:144989886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "non-zero 2", "IR"},
   CellID->144989886]},
 "non-zero 2"->{
  Cell[446279, 12664, 907, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:144989886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "non-zero 2", "IR"},
   CellID->144989886]},
 "ID:2138850160"->{
  Cell[447189, 12690, 1001, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2138850160", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 2"},
   CellID->2138850160]},
 "lp 2"->{
  Cell[447189, 12690, 1001, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2138850160", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 2"},
   CellID->2138850160]},
 "ID:1145306026"->{
  Cell[448193, 12720, 988, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1145306026", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 2"},
   CellID->1145306026]},
 "lc 2"->{
  Cell[448193, 12720, 988, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1145306026", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 2"},
   CellID->1145306026]},
 "ID:1442316125"->{
  Cell[449184, 12751, 861, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1442316125", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 3"},
   CellID->1442316125]},
 "lc 3"->{
  Cell[449184, 12751, 861, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1442316125", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 3"},
   CellID->1442316125]},
 "ID:655883205"->{
  Cell[450048, 12777, 2155, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:655883205", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 3"},
   CellID->655883205]},
 "lp 3"->{
  Cell[450048, 12777, 2155, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:655883205", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 3"},
   CellID->655883205]},
 "ID:996448533"->{
  Cell[452206, 12847, 1488, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:996448533", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "non-zero 3", "IR"},
   CellID->996448533]},
 "non-zero 3"->{
  Cell[452206, 12847, 1488, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:996448533", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "non-zero 3", "IR"},
   CellID->996448533]},
 "ID:1383271844"->{
  Cell[453697, 12893, 1787, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1383271844", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 5"},
   CellID->1383271844]},
 "lp 5"->{
  Cell[453697, 12893, 1787, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1383271844", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 5"},
   CellID->1383271844]},
 "ID:945941808"->{
  Cell[455487, 12949, 736, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:945941808", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->945941808]},
 "ID:857665826"->{
  Cell[456226, 12970, 1443, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:857665826", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C lm"},
   CellID->857665826]},
 "C lm"->{
  Cell[456226, 12970, 1443, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:857665826", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C lm"},
   CellID->857665826]},
 "ID:1411816658"->{
  Cell[457672, 13016, 1289, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1411816658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm 2"},
   CellID->1411816658]},
 "lm 2"->{
  Cell[457672, 13016, 1289, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1411816658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm 2"},
   CellID->1411816658]},
 "Proof|ID:314696779"->{
  Cell[459119, 13065, 2326, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:314696779", "Proof|ID:314696779-1"},
   CellID->817945206]},
 "Proof|ID:314696779-1"->{
  Cell[459119, 13065, 2326, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:314696779", "Proof|ID:314696779-1"},
   CellID->817945206]},
 "Proof|ID:849229881"->{
  Cell[461966, 13154, 2434, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:849229881", "Proof|ID:849229881-1"},
   CellID->766889252]},
 "Proof|ID:849229881-1"->{
  Cell[461966, 13154, 2434, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:849229881", "Proof|ID:849229881-1"},
   CellID->766889252]},
 "Proof|ID:406483636"->{
  Cell[464921, 13245, 2435, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:406483636", "Proof|ID:406483636-1"},
   CellID->1626957666]},
 "Proof|ID:406483636-1"->{
  Cell[464921, 13245, 2435, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:406483636", "Proof|ID:406483636-1"},
   CellID->1626957666]},
 "Proof|ID:1767085150"->{
  Cell[467877, 13336, 2443, 73, 70, "ProofInfo",
   CellTags->{"Proof|ID:1767085150", "Proof|ID:1767085150-1"},
   CellID->450159143]},
 "Proof|ID:1767085150-1"->{
  Cell[467877, 13336, 2443, 73, 70, "ProofInfo",
   CellTags->{"Proof|ID:1767085150", "Proof|ID:1767085150-1"},
   CellID->450159143]},
 "Proof|ID:144989886"->{
  Cell[470842, 13428, 2948, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:144989886", "Proof|ID:144989886-1"},
   CellID->914771064]},
 "Proof|ID:144989886-1"->{
  Cell[470842, 13428, 2948, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:144989886", "Proof|ID:144989886-1"},
   CellID->914771064]},
 "Proof|ID:2138850160"->{
  Cell[474311, 13531, 3337, 93, 70, "ProofInfo",
   CellTags->{"Proof|ID:2138850160", "Proof|ID:2138850160-1"},
   CellID->1795169377]},
 "Proof|ID:2138850160-1"->{
  Cell[474311, 13531, 3337, 93, 70, "ProofInfo",
   CellTags->{"Proof|ID:2138850160", "Proof|ID:2138850160-1"},
   CellID->1795169377]},
 "Proof|ID:1145306026"->{
  Cell[478169, 13643, 3540, 97, 70, "ProofInfo",
   CellTags->{"Proof|ID:1145306026", "Proof|ID:1145306026-1"},
   CellID->109823534]},
 "Proof|ID:1145306026-1"->{
  Cell[478169, 13643, 3540, 97, 70, "ProofInfo",
   CellTags->{"Proof|ID:1145306026", "Proof|ID:1145306026-1"},
   CellID->109823534]},
 "Proof|ID:1442316125"->{
  Cell[482231, 13759, 2940, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:1442316125", "Proof|ID:1442316125-1"},
   CellID->1835069754]},
 "Proof|ID:1442316125-1"->{
  Cell[482231, 13759, 2940, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:1442316125", "Proof|ID:1442316125-1"},
   CellID->1835069754]},
 "Proof|ID:655883205"->{
  Cell[485693, 13863, 6014, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:655883205", "Proof|ID:655883205-1"},
   CellID->489542376]},
 "Proof|ID:655883205-1"->{
  Cell[485693, 13863, 6014, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:655883205", "Proof|ID:655883205-1"},
   CellID->489542376]},
 "Proof|ID:996448533"->{
  Cell[492227, 14030, 4526, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:996448533", "Proof|ID:996448533-1"},
   CellID->925906148]},
 "Proof|ID:996448533-1"->{
  Cell[492227, 14030, 4526, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:996448533", "Proof|ID:996448533-1"},
   CellID->925906148]},
 "Proof|ID:1383271844"->{
  Cell[497272, 14165, 4891, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:1383271844", "Proof|ID:1383271844-1"},
   CellID->1411486137]},
 "Proof|ID:1383271844-1"->{
  Cell[497272, 14165, 4891, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:1383271844", "Proof|ID:1383271844-1"},
   CellID->1411486137]},
 "Proof|ID:857665826"->{
  Cell[502685, 14307, 4797, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:857665826", "Proof|ID:857665826-1"},
   CellID->606096953]},
 "Proof|ID:857665826-1"->{
  Cell[502685, 14307, 4797, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:857665826", "Proof|ID:857665826-1"},
   CellID->606096953]},
 "Proof|ID:1411816658"->{
  Cell[508003, 14447, 4295, 113, 70, "ProofInfo",
   CellTags->{"Proof|ID:1411816658", "Proof|ID:1411816658-1"},
   CellID->352726063]},
 "Proof|ID:1411816658-1"->{
  Cell[508003, 14447, 4295, 113, 70, "ProofInfo",
   CellTags->{"Proof|ID:1411816658", "Proof|ID:1411816658-1"},
   CellID->352726063]},
 "ID:1368691869"->{
  Cell[513591, 14603, 960, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1368691869", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1\[CenterDot]1"},
   CellID->1368691869]},
 "1\[CenterDot]1"->{
  Cell[513591, 14603, 960, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1368691869", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1\[CenterDot]1"},
   CellID->1368691869]},
 "ID:1325102956"->{
  Cell[514554, 14630, 766, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1325102956", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1325102956]},
 "ID:1663774274"->{
  Cell[515323, 14652, 739, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1663774274", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1663774274]},
 "ID:941688542"->{
  Cell[516065, 14673, 1124, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:941688542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial \[CenterDot]"},
   CellID->941688542]},
 "isMonomial \[CenterDot]"->{
  Cell[516065, 14673, 1124, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:941688542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial \[CenterDot]"},
   CellID->941688542]},
 "ID:91567084"->{
  Cell[517192, 14707, 1187, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:91567084", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 0"},
   CellID->91567084]},
 "\[CenterDot] 0"->{
  Cell[517192, 14707, 1187, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:91567084", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 0"},
   CellID->91567084]},
 "ID:420812829"->{
  Cell[518382, 14745, 987, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:420812829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 3"},
   CellID->420812829]},
 "\[CenterDot] 3"->{
  Cell[518382, 14745, 987, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:420812829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 3"},
   CellID->420812829]},
 "ID:138547187"->{
  Cell[519372, 14774, 1703, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:138547187", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 4"},
   CellID->138547187]},
 "\[CenterDot] 4"->{
  Cell[519372, 14774, 1703, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:138547187", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 4"},
   CellID->138547187]},
 "ID:457378178"->{
  Cell[521078, 14829, 1242, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:457378178", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 11"},
   CellID->457378178]},
 "\[CenterDot] 11"->{
  Cell[521078, 14829, 1242, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:457378178", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 11"},
   CellID->457378178]},
 "ID:926082604"->{
  Cell[522323, 14868, 2172, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:926082604", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 11'"},
   CellID->926082604]},
 "\[CenterDot] 11'"->{
  Cell[522323, 14868, 2172, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:926082604", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 11'"},
   CellID->926082604]},
 "ID:1344734857"->{
  Cell[524498, 14942, 1581, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1344734857", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] lp lc"},
   CellID->1344734857]},
 "\[CenterDot] lp lc"->{
  Cell[524498, 14942, 1581, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1344734857", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] lp lc"},
   CellID->1344734857]},
 "Proof|ID:1368691869"->{
  Cell[526235, 15003, 2197, 67, 70, "ProofInfo",
   CellTags->{"Proof|ID:1368691869", "Proof|ID:1368691869-1"},
   CellID->660760468]},
 "Proof|ID:1368691869-1"->{
  Cell[526235, 15003, 2197, 67, 70, "ProofInfo",
   CellTags->{"Proof|ID:1368691869", "Proof|ID:1368691869-1"},
   CellID->660760468]},
 "Proof|ID:941688542"->{
  Cell[528953, 15089, 3445, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:941688542", "Proof|ID:941688542-1"},
   CellID->1365249879]},
 "Proof|ID:941688542-1"->{
  Cell[528953, 15089, 3445, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:941688542", "Proof|ID:941688542-1"},
   CellID->1365249879]},
 "Proof|ID:91567084"->{
  Cell[532919, 15202, 3402, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:91567084", "Proof|ID:91567084-1"},
   CellID->1862161014]},
 "Proof|ID:91567084-1"->{
  Cell[532919, 15202, 3402, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:91567084", "Proof|ID:91567084-1"},
   CellID->1862161014]},
 "Proof|ID:420812829"->{
  Cell[536840, 15315, 3341, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:420812829", "Proof|ID:420812829-1"},
   CellID->2092842442]},
 "Proof|ID:420812829-1"->{
  Cell[536840, 15315, 3341, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:420812829", "Proof|ID:420812829-1"},
   CellID->2092842442]},
 "Proof|ID:138547187"->{
  Cell[540701, 15426, 4585, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:138547187", "Proof|ID:138547187-1"},
   CellID->244847721]},
 "Proof|ID:138547187-1"->{
  Cell[540701, 15426, 4585, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:138547187", "Proof|ID:138547187-1"},
   CellID->244847721]},
 "Proof|ID:457378178"->{
  Cell[545807, 15563, 3996, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:457378178", "Proof|ID:457378178-1"},
   CellID->783388721]},
 "Proof|ID:457378178-1"->{
  Cell[545807, 15563, 3996, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:457378178", "Proof|ID:457378178-1"},
   CellID->783388721]},
 "Proof|ID:926082604"->{
  Cell[550366, 15689, 5150, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:926082604", "Proof|ID:926082604-1"},
   CellID->1364381745]},
 "Proof|ID:926082604-1"->{
  Cell[550366, 15689, 5150, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:926082604", "Proof|ID:926082604-1"},
   CellID->1364381745]},
 "Proof|ID:1344734857"->{
  Cell[556036, 15836, 4469, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1344734857", "Proof|ID:1344734857-1"},
   CellID->670171654]},
 "Proof|ID:1344734857-1"->{
  Cell[556036, 15836, 4469, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1344734857", "Proof|ID:1344734857-1"},
   CellID->670171654]},
 "ID:1558664548"->{
  Cell[561465, 15984, 1094, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1558664548", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 1"},
   CellID->1558664548]},
 "isMonomial 1"->{
  Cell[561465, 15984, 1094, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1558664548", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 1"},
   CellID->1558664548]},
 "ID:1835029788"->{
  Cell[562562, 16018, 903, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1835029788", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1835029788]},
 "ID:757450222"->{
  Cell[563468, 16044, 945, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:757450222", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial \[NotEqual] 0"},
   CellID->757450222]},
 "isMonomial \[NotEqual] 0"->{
  Cell[563468, 16044, 945, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:757450222", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial \[NotEqual] 0"},
   CellID->757450222]},
 "ID:1978849375"->{
  Cell[564416, 16069, 921, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1978849375", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial -"},
   CellID->1978849375]},
 "isMonomial -"->{
  Cell[564416, 16069, 921, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1978849375", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial -"},
   CellID->1978849375]},
 "ID:283455673"->{
  Cell[565340, 16096, 982, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:283455673", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 3'"},
   CellID->283455673]},
 "isMonomial 3'"->{
  Cell[565340, 16096, 982, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:283455673", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 3'"},
   CellID->283455673]},
 "ID:610787620"->{
  Cell[566325, 16125, 1438, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:610787620", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 3"},
   CellID->610787620]},
 "isMonomial 3"->{
  Cell[566325, 16125, 1438, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:610787620", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 3"},
   CellID->610787620]},
 "ID:980082706"->{
  Cell[567766, 16169, 1841, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:980082706", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 4"},
   CellID->980082706]},
 "isMonomial 4"->{
  Cell[567766, 16169, 1841, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:980082706", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 4"},
   CellID->980082706]},
 "ID:1306217654"->{
  Cell[569610, 16229, 836, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1306217654", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm isMonomial"},
   CellID->1306217654]},
 "lm isMonomial"->{
  Cell[569610, 16229, 836, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1306217654", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm isMonomial"},
   CellID->1306217654]},
 "ID:1645795930"->{
  Cell[570449, 16254, 1570, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1645795930", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 7"},
   CellID->1645795930]},
 "isMonomial 7"->{
  Cell[570449, 16254, 1570, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1645795930", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 7"},
   CellID->1645795930]},
 "ID:691307310"->{
  Cell[572022, 16305, 736, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:691307310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->691307310]},
 "ID:1901309763"->{
  Cell[572761, 16326, 1778, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1901309763", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 8"},
   CellID->1901309763]},
 "isMonomial 8"->{
  Cell[572761, 16326, 1778, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1901309763", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 8"},
   CellID->1901309763]},
 "ID:1177943318"->{
  Cell[574542, 16382, 1504, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1177943318", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 9"},
   CellID->1177943318]},
 "isMonomial 9"->{
  Cell[574542, 16382, 1504, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1177943318", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 9"},
   CellID->1177943318]},
 "ID:1770611523"->{
  Cell[576049, 16432, 721, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1770611523", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1770611523]},
 "ID:670623560"->{
  Cell[576773, 16453, 1719, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:670623560", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 12"},
   CellID->670623560]},
 "isMonomial 12"->{
  Cell[576773, 16453, 1719, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:670623560", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 12"},
   CellID->670623560]},
 "ID:1806313475"->{
  Cell[578495, 16509, 957, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1806313475", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial lm"},
   CellID->1806313475]},
 "isMonomial lm"->{
  Cell[578495, 16509, 957, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1806313475", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial lm"},
   CellID->1806313475]},
 "ID:1110032163"->{
  Cell[579455, 16536, 964, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1110032163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc lm"},
   CellID->1110032163]},
 "lc lm"->{
  Cell[579455, 16536, 964, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1110032163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc lm"},
   CellID->1110032163]},
 "ID:632352665"->{
  Cell[580422, 16566, 961, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:632352665", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp lm"},
   CellID->632352665]},
 "lp lm"->{
  Cell[580422, 16566, 961, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:632352665", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp lm"},
   CellID->632352665]},
 "Proof|ID:1558664548"->{
  Cell[581540, 16605, 2278, 67, 70, "ProofInfo",
   CellTags->{"Proof|ID:1558664548", "Proof|ID:1558664548-1"},
   CellID->142537034]},
 "Proof|ID:1558664548-1"->{
  Cell[581540, 16605, 2278, 67, 70, "ProofInfo",
   CellTags->{"Proof|ID:1558664548", "Proof|ID:1558664548-1"},
   CellID->142537034]},
 "Proof|ID:757450222"->{
  Cell[584339, 16691, 2789, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:757450222", "Proof|ID:757450222-1"},
   CellID->1689557017]},
 "Proof|ID:757450222-1"->{
  Cell[584339, 16691, 2789, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:757450222", "Proof|ID:757450222-1"},
   CellID->1689557017]},
 "Proof|ID:1978849375"->{
  Cell[587649, 16790, 2964, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1978849375", "Proof|ID:1978849375-1"},
   CellID->807800730]},
 "Proof|ID:1978849375-1"->{
  Cell[587649, 16790, 2964, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1978849375", "Proof|ID:1978849375-1"},
   CellID->807800730]},
 "Proof|ID:283455673"->{
  Cell[591135, 16892, 3438, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:283455673", "Proof|ID:283455673-1"},
   CellID->1547205861]},
 "Proof|ID:283455673-1"->{
  Cell[591135, 16892, 3438, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:283455673", "Proof|ID:283455673-1"},
   CellID->1547205861]},
 "Proof|ID:610787620"->{
  Cell[595093, 17003, 4528, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:610787620", "Proof|ID:610787620-1"},
   CellID->1422344349]},
 "Proof|ID:610787620-1"->{
  Cell[595093, 17003, 4528, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:610787620", "Proof|ID:610787620-1"},
   CellID->1422344349]},
 "Proof|ID:980082706"->{
  Cell[600142, 17140, 4986, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:980082706", "Proof|ID:980082706-1"},
   CellID->1540912936]},
 "Proof|ID:980082706-1"->{
  Cell[600142, 17140, 4986, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:980082706", "Proof|ID:980082706-1"},
   CellID->1540912936]},
 "Proof|ID:1306217654"->{
  Cell[605649, 17283, 2949, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:1306217654", "Proof|ID:1306217654-1"},
   CellID->277734451]},
 "Proof|ID:1306217654-1"->{
  Cell[605649, 17283, 2949, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:1306217654", "Proof|ID:1306217654-1"},
   CellID->277734451]},
 "Proof|ID:1645795930"->{
  Cell[609119, 17387, 4080, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:1645795930", "Proof|ID:1645795930-1"},
   CellID->140680378]},
 "Proof|ID:1645795930-1"->{
  Cell[609119, 17387, 4080, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:1645795930", "Proof|ID:1645795930-1"},
   CellID->140680378]},
 "Proof|ID:1901309763"->{
  Cell[613719, 17513, 5548, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:1901309763", "Proof|ID:1901309763-1"},
   CellID->907132124]},
 "Proof|ID:1901309763-1"->{
  Cell[613719, 17513, 5548, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:1901309763", "Proof|ID:1901309763-1"},
   CellID->907132124]},
 "Proof|ID:1177943318"->{
  Cell[619788, 17673, 3681, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1177943318", "Proof|ID:1177943318-1"},
   CellID->1237355862]},
 "Proof|ID:1177943318-1"->{
  Cell[619788, 17673, 3681, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1177943318", "Proof|ID:1177943318-1"},
   CellID->1237355862]},
 "Proof|ID:670623560"->{
  Cell[623990, 17791, 4918, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:670623560", "Proof|ID:670623560-1"},
   CellID->1341250136]},
 "Proof|ID:670623560-1"->{
  Cell[623990, 17791, 4918, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:670623560", "Proof|ID:670623560-1"},
   CellID->1341250136]},
 "Proof|ID:1806313475"->{
  Cell[629428, 17936, 2952, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1806313475", "Proof|ID:1806313475-1"},
   CellID->347558245]},
 "Proof|ID:1806313475-1"->{
  Cell[629428, 17936, 2952, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1806313475", "Proof|ID:1806313475-1"},
   CellID->347558245]},
 "Proof|ID:1110032163"->{
  Cell[632902, 18038, 3332, 93, 70, "ProofInfo",
   CellTags->{"Proof|ID:1110032163", "Proof|ID:1110032163-1"},
   CellID->916272762]},
 "Proof|ID:1110032163-1"->{
  Cell[632902, 18038, 3332, 93, 70, "ProofInfo",
   CellTags->{"Proof|ID:1110032163", "Proof|ID:1110032163-1"},
   CellID->916272762]},
 "Proof|ID:632352665"->{
  Cell[636756, 18150, 3323, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:632352665", "Proof|ID:632352665-1"},
   CellID->800412793]},
 "Proof|ID:632352665-1"->{
  Cell[636756, 18150, 3323, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:632352665", "Proof|ID:632352665-1"},
   CellID->800412793]},
 "ID:423895762"->{
  Cell[641463, 18294, 1000, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:423895762", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ closed \[ScriptCapitalP]"},
   CellID->423895762]},
 "+ closed \[ScriptCapitalP]"->{
  Cell[641463, 18294, 1000, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:423895762", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ closed \[ScriptCapitalP]"},
   CellID->423895762]},
 "Proof|ID:423895762"->{
  Cell[642620, 18335, 1946, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:423895762", "Proof|ID:423895762-1"},
   CellID->355044739]},
 "Proof|ID:423895762-1"->{
  Cell[642620, 18335, 1946, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:423895762", "Proof|ID:423895762-1"},
   CellID->355044739]},
 "ID:1812309492"->{
  Cell[646025, 18448, 1108, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1812309492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "0 neutral \[ScriptCapitalP]"},
   CellID->1812309492]},
 "0 neutral \[ScriptCapitalP]"->{
  Cell[646025, 18448, 1108, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1812309492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "0 neutral \[ScriptCapitalP]"},
   CellID->1812309492]},
 "Proof|ID:1812309492"->{
  Cell[647290, 18491, 2078, 63, 70, "ProofInfo",
   CellTags->{"Proof|ID:1812309492", "Proof|ID:1812309492-1"},
   CellID->2145107533]},
 "Proof|ID:1812309492-1"->{
  Cell[647290, 18491, 2078, 63, 70, "ProofInfo",
   CellTags->{"Proof|ID:1812309492", "Proof|ID:1812309492-1"},
   CellID->2145107533]},
 "ID:595518277"->{
  Cell[650801, 18606, 1048, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:595518277", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ assoc \[ScriptCapitalP]"},
   CellID->595518277]},
 "+ assoc \[ScriptCapitalP]"->{
  Cell[650801, 18606, 1048, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:595518277", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ assoc \[ScriptCapitalP]"},
   CellID->595518277]},
 "Proof|ID:595518277"->{
  Cell[652007, 18648, 1943, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:595518277", "Proof|ID:595518277-1"},
   CellID->1392319518]},
 "Proof|ID:595518277-1"->{
  Cell[652007, 18648, 1943, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:595518277", "Proof|ID:595518277-1"},
   CellID->1392319518]},
 "ID:2114498624"->{
  Cell[655387, 18760, 1046, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2114498624", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ comm \[ScriptCapitalP]"},
   CellID->2114498624]},
 "+ comm \[ScriptCapitalP]"->{
  Cell[655387, 18760, 1046, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2114498624", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ comm \[ScriptCapitalP]"},
   CellID->2114498624]},
 "Proof|ID:2114498624"->{
  Cell[656591, 18802, 1948, 61, 70, "ProofInfo",
   CellTags->{"Proof|ID:2114498624", "Proof|ID:2114498624-1"},
   CellID->2065823929]},
 "Proof|ID:2114498624-1"->{
  Cell[656591, 18802, 1948, 61, 70, "ProofInfo",
   CellTags->{"Proof|ID:2114498624", "Proof|ID:2114498624-1"},
   CellID->2065823929]},
 "ID:246227628"->{
  Cell[660028, 18920, 763, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:246227628", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->246227628]},
 "ID:95118700"->{
  Cell[660794, 18942, 1007, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:95118700", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP] 1"},
   CellID->95118700]},
 "- inverse \[ScriptCapitalP] 1"->{
  Cell[660794, 18942, 1007, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:95118700", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP] 1"},
   CellID->95118700]},
 "ID:1900465816"->{
  Cell[661804, 18972, 897, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1900465816", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP] 2"},
   CellID->1900465816]},
 "- inverse \[ScriptCapitalP] 2"->{
  Cell[661804, 18972, 897, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1900465816", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP] 2"},
   CellID->1900465816]},
 "ID:2037231548"->{
  Cell[662704, 18997, 1240, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2037231548", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP]"},
   CellID->2037231548]},
 "- inverse \[ScriptCapitalP]"->{
  Cell[662704, 18997, 1240, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2037231548", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP]"},
   CellID->2037231548]},
 "Proof|ID:95118700"->{
  Cell[664100, 19047, 3423, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:95118700", "Proof|ID:95118700-1"},
   CellID->49996501]},
 "Proof|ID:95118700-1"->{
  Cell[664100, 19047, 3423, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:95118700", "Proof|ID:95118700-1"},
   CellID->49996501]},
 "Proof|ID:1900465816"->{
  Cell[668043, 19158, 2682, 79, 70, "ProofInfo",
   CellTags->{"Proof|ID:1900465816", "Proof|ID:1900465816-1"},
   CellID->98966748]},
 "Proof|ID:1900465816-1"->{
  Cell[668043, 19158, 2682, 79, 70, "ProofInfo",
   CellTags->{"Proof|ID:1900465816", "Proof|ID:1900465816-1"},
   CellID->98966748]},
 "Proof|ID:2037231548"->{
  Cell[671246, 19256, 2204, 65, 70, "ProofInfo",
   CellTags->{"Proof|ID:2037231548", "Proof|ID:2037231548-1"},
   CellID->1554728169]},
 "Proof|ID:2037231548-1"->{
  Cell[671246, 19256, 2204, 65, 70, "ProofInfo",
   CellTags->{"Proof|ID:2037231548", "Proof|ID:2037231548-1"},
   CellID->1554728169]},
 "ID:145364866"->{
  Cell[674363, 19354, 918, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:145364866", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- 0 \[ScriptCapitalP] 1"},
   CellID->145364866]},
 "- 0 \[ScriptCapitalP] 1"->{
  Cell[674363, 19354, 918, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:145364866", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- 0 \[ScriptCapitalP] 1"},
   CellID->145364866]},
 "ID:117883331"->{
  Cell[675284, 19380, 1130, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:117883331", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- 0 \[ScriptCapitalP] 2"},
   CellID->117883331]},
 "- 0 \[ScriptCapitalP] 2"->{
  Cell[675284, 19380, 1130, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:117883331", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- 0 \[ScriptCapitalP] 2"},
   CellID->117883331]},
 "Proof|ID:145364866"->{
  Cell[676572, 19424, 2104, 64, 70, "ProofInfo",
   CellTags->{"Proof|ID:145364866", "Proof|ID:145364866-1"},
   CellID->448617855]},
 "Proof|ID:145364866-1"->{
  Cell[676572, 19424, 2104, 64, 70, "ProofInfo",
   CellTags->{"Proof|ID:145364866", "Proof|ID:145364866-1"},
   CellID->448617855]},
 "Proof|ID:117883331"->{
  Cell[679197, 19507, 2663, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:117883331", "Proof|ID:117883331-1"},
   CellID->1580290873]},
 "Proof|ID:117883331-1"->{
  Cell[679197, 19507, 2663, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:117883331", "Proof|ID:117883331-1"},
   CellID->1580290873]},
 "ID:771361293"->{
  Cell[683620, 19649, 1059, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:771361293", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCCMonoid \[ScriptCapitalP]"},
   CellID->771361293]},
 "isCCMonoid \[ScriptCapitalP]"->{
  Cell[683620, 19649, 1059, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:771361293", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCCMonoid \[ScriptCapitalP]"},
   CellID->771361293]},
 "Proof|ID:771361293"->{
  Cell[684836, 19691, 2063, 61, 70, "ProofInfo",
   CellTags->{"Proof|ID:771361293", "Proof|ID:771361293-1"},
   CellID->556042782]},
 "Proof|ID:771361293-1"->{
  Cell[684836, 19691, 2063, 61, 70, "ProofInfo",
   CellTags->{"Proof|ID:771361293", "Proof|ID:771361293-1"},
   CellID->556042782]},
 "ID:1221832846"->{
  Cell[688477, 19804, 842, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1221832846", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1221832846]},
 "ID:313886646"->{
  Cell[689322, 19828, 944, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:313886646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership R", "IR"},
   CellID->313886646]},
 "membership R"->{
  Cell[689322, 19828, 944, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:313886646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership R", "IR"},
   CellID->313886646]},
 "ID:174448700"->{
  Cell[690269, 19856, 952, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:174448700", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm+R"},
   CellID->174448700]},
 "lm+R"->{
  Cell[690269, 19856, 952, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:174448700", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm+R"},
   CellID->174448700]},
 "ID:1679470623"->{
  Cell[691224, 19885, 1735, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1679470623", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C R"},
   CellID->1679470623]},
 "C R"->{
  Cell[691224, 19885, 1735, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1679470623", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C R"},
   CellID->1679470623]},
 "ID:1648446612"->{
  Cell[692962, 19938, 1144, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1648446612", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 1"},
   CellID->1648446612]},
 "R 1"->{
  Cell[692962, 19938, 1144, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1648446612", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 1"},
   CellID->1648446612]},
 "ID:1799775671"->{
  Cell[694109, 19973, 1672, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1799775671", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 2"},
   CellID->1799775671]},
 "R 2"->{
  Cell[694109, 19973, 1672, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1799775671", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 2"},
   CellID->1799775671]},
 "ID:1065304756"->{
  Cell[695784, 20025, 1024, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1065304756", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 6"},
   CellID->1065304756]},
 "R 6"->{
  Cell[695784, 20025, 1024, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1065304756", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 6"},
   CellID->1065304756]},
 "ID:1078436809"->{
  Cell[696811, 20056, 1534, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1078436809", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 7"},
   CellID->1078436809]},
 "R 7"->{
  Cell[696811, 20056, 1534, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1078436809", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 7"},
   CellID->1078436809]},
 "ID:735362349"->{
  Cell[698348, 20107, 1061, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:735362349", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 8"},
   CellID->735362349]},
 "R 8"->{
  Cell[698348, 20107, 1061, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:735362349", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 8"},
   CellID->735362349]},
 "ID:451768283"->{
  Cell[699412, 20141, 1062, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:451768283", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 8'"},
   CellID->451768283]},
 "R 8'"->{
  Cell[699412, 20141, 1062, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:451768283", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 8'"},
   CellID->451768283]},
 "ID:739591353"->{
  Cell[700477, 20175, 2072, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739591353", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 5'"},
   CellID->739591353]},
 "C 5'"->{
  Cell[700477, 20175, 2072, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739591353", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 5'"},
   CellID->739591353]},
 "ID:1347176402"->{
  Cell[702552, 20243, 796, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1347176402", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1347176402]},
 "ID:1377108879"->{
  Cell[703351, 20267, 1222, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1377108879", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 3"},
   CellID->1377108879]},
 "R 3"->{
  Cell[703351, 20267, 1222, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1377108879", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 3"},
   CellID->1377108879]},
 "ID:139632107"->{
  Cell[704576, 20300, 1765, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:139632107", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 4"},
   CellID->139632107]},
 "R 4"->{
  Cell[704576, 20300, 1765, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:139632107", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 4"},
   CellID->139632107]},
 "ID:2132986120"->{
  Cell[706344, 20358, 1030, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2132986120", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 7"},
   CellID->2132986120]},
 "C 7"->{
  Cell[706344, 20358, 1030, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2132986120", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 7"},
   CellID->2132986120]},
 "ID:1911280389"->{
  Cell[707377, 20389, 1103, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1911280389", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 8"},
   CellID->1911280389]},
 "C 8"->{
  Cell[707377, 20389, 1103, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1911280389", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 8"},
   CellID->1911280389]},
 "Proof|ID:313886646"->{
  Cell[708638, 20431, 2939, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:313886646", "Proof|ID:313886646-1"},
   CellID->368790604]},
 "Proof|ID:313886646-1"->{
  Cell[708638, 20431, 2939, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:313886646", "Proof|ID:313886646-1"},
   CellID->368790604]},
 "Proof|ID:174448700"->{
  Cell[712098, 20532, 3388, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:174448700", "Proof|ID:174448700-1"},
   CellID->1932131695]},
 "Proof|ID:174448700-1"->{
  Cell[712098, 20532, 3388, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:174448700", "Proof|ID:174448700-1"},
   CellID->1932131695]},
 "Proof|ID:1679470623"->{
  Cell[716006, 20643, 5111, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:1679470623", "Proof|ID:1679470623-1"},
   CellID->1796196826]},
 "Proof|ID:1679470623-1"->{
  Cell[716006, 20643, 5111, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:1679470623", "Proof|ID:1679470623-1"},
   CellID->1796196826]},
 "Proof|ID:1648446612"->{
  Cell[721639, 20790, 2439, 73, 70, "ProofInfo",
   CellTags->{"Proof|ID:1648446612", "Proof|ID:1648446612-1"},
   CellID->822584714]},
 "Proof|ID:1648446612-1"->{
  Cell[721639, 20790, 2439, 73, 70, "ProofInfo",
   CellTags->{"Proof|ID:1648446612", "Proof|ID:1648446612-1"},
   CellID->822584714]},
 "Proof|ID:1799775671"->{
  Cell[724599, 20882, 5059, 131, 70, "ProofInfo",
   CellTags->{"Proof|ID:1799775671", "Proof|ID:1799775671-1"},
   CellID->1365150475]},
 "Proof|ID:1799775671-1"->{
  Cell[724599, 20882, 5059, 131, 70, "ProofInfo",
   CellTags->{"Proof|ID:1799775671", "Proof|ID:1799775671-1"},
   CellID->1365150475]},
 "Proof|ID:1065304756"->{
  Cell[730180, 21032, 3543, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1065304756", "Proof|ID:1065304756-1"},
   CellID->364956431]},
 "Proof|ID:1065304756-1"->{
  Cell[730180, 21032, 3543, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1065304756", "Proof|ID:1065304756-1"},
   CellID->364956431]},
 "Proof|ID:1078436809"->{
  Cell[734245, 21150, 4531, 117, 70, "ProofInfo",
   CellTags->{"Proof|ID:1078436809", "Proof|ID:1078436809-1"},
   CellID->1581073290]},
 "Proof|ID:1078436809-1"->{
  Cell[734245, 21150, 4531, 117, 70, "ProofInfo",
   CellTags->{"Proof|ID:1078436809", "Proof|ID:1078436809-1"},
   CellID->1581073290]},
 "Proof|ID:735362349"->{
  Cell[739296, 21286, 3299, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:735362349", "Proof|ID:735362349-1"},
   CellID->229453058]},
 "Proof|ID:735362349-1"->{
  Cell[739296, 21286, 3299, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:735362349", "Proof|ID:735362349-1"},
   CellID->229453058]},
 "Proof|ID:451768283"->{
  Cell[743115, 21395, 3306, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:451768283", "Proof|ID:451768283-1"},
   CellID->497176783]},
 "Proof|ID:451768283-1"->{
  Cell[743115, 21395, 3306, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:451768283", "Proof|ID:451768283-1"},
   CellID->497176783]},
 "Proof|ID:739591353"->{
  Cell[746940, 21506, 5334, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:739591353", "Proof|ID:739591353-1"},
   CellID->1363695772]},
 "Proof|ID:739591353-1"->{
  Cell[746940, 21506, 5334, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:739591353", "Proof|ID:739591353-1"},
   CellID->1363695772]},
 "Proof|ID:1377108879"->{
  Cell[752794, 21661, 3978, 105, 70, "ProofInfo",
   CellTags->{"Proof|ID:1377108879", "Proof|ID:1377108879-1"},
   CellID->360514177]},
 "Proof|ID:1377108879-1"->{
  Cell[752794, 21661, 3978, 105, 70, "ProofInfo",
   CellTags->{"Proof|ID:1377108879", "Proof|ID:1377108879-1"},
   CellID->360514177]},
 "Proof|ID:139632107"->{
  Cell[757294, 21785, 5860, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:139632107", "Proof|ID:139632107-1"},
   CellID->1587172171]},
 "Proof|ID:139632107-1"->{
  Cell[757294, 21785, 5860, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:139632107", "Proof|ID:139632107-1"},
   CellID->1587172171]},
 "Proof|ID:2132986120"->{
  Cell[763675, 21952, 4130, 111, 70, "ProofInfo",
   CellTags->{"Proof|ID:2132986120", "Proof|ID:2132986120-1"},
   CellID->2053017099]},
 "Proof|ID:2132986120-1"->{
  Cell[763675, 21952, 4130, 111, 70, "ProofInfo",
   CellTags->{"Proof|ID:2132986120", "Proof|ID:2132986120-1"},
   CellID->2053017099]},
 "Proof|ID:1911280389"->{
  Cell[768326, 22082, 4329, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1911280389", "Proof|ID:1911280389-1"},
   CellID->472365930]},
 "Proof|ID:1911280389-1"->{
  Cell[768326, 22082, 4329, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1911280389", "Proof|ID:1911280389-1"},
   CellID->472365930]},
 "ID:1031528829"->{
  Cell[773670, 22234, 743, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1031528829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1031528829]},
 "ID:1494855415"->{
  Cell[774416, 22255, 817, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1494855415", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1494855415]},
 "ID:2007119699"->{
  Cell[775236, 22278, 1538, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2007119699", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership H ind"},
   CellID->2007119699]},
 "membership H ind"->{
  Cell[775236, 22278, 1538, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2007119699", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership H ind"},
   CellID->2007119699]},
 "ID:884488101"->{
  Cell[776777, 22324, 917, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:884488101", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership H", "IR"},
   CellID->884488101]},
 "membership H"->{
  Cell[776777, 22324, 917, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:884488101", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership H", "IR"},
   CellID->884488101]},
 "ID:1892666968"->{
  Cell[777697, 22351, 2154, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1892666968", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C H ind"},
   CellID->1892666968]},
 "C H ind"->{
  Cell[777697, 22351, 2154, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1892666968", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C H ind"},
   CellID->1892666968]},
 "ID:1556838587"->{
  Cell[779854, 22419, 1687, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1556838587", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C H"},
   CellID->1556838587]},
 "C H"->{
  Cell[779854, 22419, 1687, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1556838587", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C H"},
   CellID->1556838587]},
 "ID:1768827306"->{
  Cell[781544, 22473, 1612, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1768827306", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H=0"},
   CellID->1768827306]},
 "H=0"->{
  Cell[781544, 22473, 1612, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1768827306", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H=0"},
   CellID->1768827306]},
 "ID:140826492"->{
  Cell[783159, 22524, 1783, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:140826492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H=H"},
   CellID->140826492]},
 "H=H"->{
  Cell[783159, 22524, 1783, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:140826492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H=H"},
   CellID->140826492]},
 "ID:1627546251"->{
  Cell[784945, 22583, 1161, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1627546251", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 1"},
   CellID->1627546251]},
 "H 1"->{
  Cell[784945, 22583, 1161, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1627546251", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 1"},
   CellID->1627546251]},
 "ID:743145124"->{
  Cell[786109, 22619, 1056, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:743145124", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 2"},
   CellID->743145124]},
 "H 2"->{
  Cell[786109, 22619, 1056, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:743145124", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 2"},
   CellID->743145124]},
 "ID:1576432658"->{
  Cell[787168, 22652, 1163, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1576432658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 3"},
   CellID->1576432658]},
 "H 3"->{
  Cell[787168, 22652, 1163, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1576432658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 3"},
   CellID->1576432658]},
 "ID:1977357773"->{
  Cell[788334, 22688, 1591, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1977357773", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 4"},
   CellID->1977357773]},
 "H 4"->{
  Cell[788334, 22688, 1591, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1977357773", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 4"},
   CellID->1977357773]},
 "ID:561821259"->{
  Cell[789928, 22739, 1269, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:561821259", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 4'"},
   CellID->561821259]},
 "H 4'"->{
  Cell[789928, 22739, 1269, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:561821259", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 4'"},
   CellID->561821259]},
 "ID:866138121"->{
  Cell[791200, 22778, 2151, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:866138121", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 5"},
   CellID->866138121]},
 "H 5"->{
  Cell[791200, 22778, 2151, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:866138121", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 5"},
   CellID->866138121]},
 "ID:105044014"->{
  Cell[793354, 22846, 2071, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:105044014", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 5'"},
   CellID->105044014]},
 "H 5'"->{
  Cell[793354, 22846, 2071, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:105044014", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 5'"},
   CellID->105044014]},
 "ID:1971276224"->{
  Cell[795428, 22910, 1611, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1971276224", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 6"},
   CellID->1971276224]},
 "H 6"->{
  Cell[795428, 22910, 1611, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1971276224", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 6"},
   CellID->1971276224]},
 "ID:219245057"->{
  Cell[797042, 22961, 744, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:219245057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->219245057]},
 "ID:2135590264"->{
  Cell[797789, 22983, 953, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2135590264", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 7"},
   CellID->2135590264]},
 "H 7"->{
  Cell[797789, 22983, 953, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2135590264", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 7"},
   CellID->2135590264]},
 "ID:2092106183"->{
  Cell[798745, 23012, 1214, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2092106183", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 8"},
   CellID->2092106183]},
 "H 8"->{
  Cell[798745, 23012, 1214, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2092106183", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 8"},
   CellID->2092106183]},
 "ID:831355566"->{
  Cell[799962, 23050, 793, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:831355566", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->831355566]},
 "ID:2014193763"->{
  Cell[800758, 23074, 1112, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2014193763", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 10"},
   CellID->2014193763]},
 "H 10"->{
  Cell[800758, 23074, 1112, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2014193763", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 10"},
   CellID->2014193763]},
 "ID:1650576647"->{
  Cell[801873, 23108, 1137, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1650576647", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 11"},
   CellID->1650576647]},
 "H 11"->{
  Cell[801873, 23108, 1137, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1650576647", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 11"},
   CellID->1650576647]},
 "Proof|ID:2007119699"->{
  Cell[803167, 23152, 4314, 111, 70, "ProofInfo",
   CellTags->{"Proof|ID:2007119699", "Proof|ID:2007119699-1"},
   CellID->526673768]},
 "Proof|ID:2007119699-1"->{
  Cell[803167, 23152, 4314, 111, 70, "ProofInfo",
   CellTags->{"Proof|ID:2007119699", "Proof|ID:2007119699-1"},
   CellID->526673768]},
 "Proof|ID:884488101"->{
  Cell[808045, 23283, 3040, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:884488101", "Proof|ID:884488101-1"},
   CellID->1564315886]},
 "Proof|ID:884488101-1"->{
  Cell[808045, 23283, 3040, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:884488101", "Proof|ID:884488101-1"},
   CellID->1564315886]},
 "Proof|ID:1892666968"->{
  Cell[811605, 23386, 6278, 152, 70, "ProofInfo",
   CellTags->{"Proof|ID:1892666968", "Proof|ID:1892666968-1"},
   CellID->800813603]},
 "Proof|ID:1892666968-1"->{
  Cell[811605, 23386, 6278, 152, 70, "ProofInfo",
   CellTags->{"Proof|ID:1892666968", "Proof|ID:1892666968-1"},
   CellID->800813603]},
 "Proof|ID:1556838587"->{
  Cell[818404, 23557, 4950, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:1556838587", "Proof|ID:1556838587-1"},
   CellID->293752933]},
 "Proof|ID:1556838587-1"->{
  Cell[818404, 23557, 4950, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:1556838587", "Proof|ID:1556838587-1"},
   CellID->293752933]},
 "Proof|ID:1768827306"->{
  Cell[823876, 23701, 4849, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:1768827306", "Proof|ID:1768827306-1"},
   CellID->864932006]},
 "Proof|ID:1768827306-1"->{
  Cell[823876, 23701, 4849, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:1768827306", "Proof|ID:1768827306-1"},
   CellID->864932006]},
 "Proof|ID:140826492"->{
  Cell[829247, 23847, 5867, 146, 70, "ProofInfo",
   CellTags->{"Proof|ID:140826492", "Proof|ID:140826492-1"},
   CellID->743847445]},
 "Proof|ID:140826492-1"->{
  Cell[829247, 23847, 5867, 146, 70, "ProofInfo",
   CellTags->{"Proof|ID:140826492", "Proof|ID:140826492-1"},
   CellID->743847445]},
 "Proof|ID:1627546251"->{
  Cell[835635, 24012, 4603, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1627546251", "Proof|ID:1627546251-1"},
   CellID->1563876245]},
 "Proof|ID:1627546251-1"->{
  Cell[835635, 24012, 4603, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1627546251", "Proof|ID:1627546251-1"},
   CellID->1563876245]},
 "Proof|ID:743145124"->{
  Cell[840759, 24150, 3757, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:743145124", "Proof|ID:743145124-1"},
   CellID->729988509]},
 "Proof|ID:743145124-1"->{
  Cell[840759, 24150, 3757, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:743145124", "Proof|ID:743145124-1"},
   CellID->729988509]},
 "Proof|ID:1576432658"->{
  Cell[845036, 24269, 4669, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:1576432658", "Proof|ID:1576432658-1"},
   CellID->478962428]},
 "Proof|ID:1576432658-1"->{
  Cell[845036, 24269, 4669, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:1576432658", "Proof|ID:1576432658-1"},
   CellID->478962428]},
 "Proof|ID:1977357773"->{
  Cell[850227, 24409, 4602, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1977357773", "Proof|ID:1977357773-1"},
   CellID->518253128]},
 "Proof|ID:1977357773-1"->{
  Cell[850227, 24409, 4602, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1977357773", "Proof|ID:1977357773-1"},
   CellID->518253128]},
 "Proof|ID:561821259"->{
  Cell[855350, 24547, 4089, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:561821259", "Proof|ID:561821259-1"},
   CellID->1734326259]},
 "Proof|ID:561821259-1"->{
  Cell[855350, 24547, 4089, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:561821259", "Proof|ID:561821259-1"},
   CellID->1734326259]},
 "Proof|ID:866138121"->{
  Cell[859959, 24674, 6749, 162, 70, "ProofInfo",
   CellTags->{"Proof|ID:866138121", "Proof|ID:866138121-1"},
   CellID->726143767]},
 "Proof|ID:866138121-1"->{
  Cell[859959, 24674, 6749, 162, 70, "ProofInfo",
   CellTags->{"Proof|ID:866138121", "Proof|ID:866138121-1"},
   CellID->726143767]},
 "Proof|ID:105044014"->{
  Cell[867229, 24855, 6473, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:105044014", "Proof|ID:105044014-1"},
   CellID->839817768]},
 "Proof|ID:105044014-1"->{
  Cell[867229, 24855, 6473, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:105044014", "Proof|ID:105044014-1"},
   CellID->839817768]},
 "Proof|ID:1971276224"->{
  Cell[874223, 25028, 4575, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1971276224", "Proof|ID:1971276224-1"},
   CellID->611586869]},
 "Proof|ID:1971276224-1"->{
  Cell[874223, 25028, 4575, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1971276224", "Proof|ID:1971276224-1"},
   CellID->611586869]},
 "Proof|ID:2135590264"->{
  Cell[879319, 25166, 3338, 95, 70, "ProofInfo",
   CellTags->{"Proof|ID:2135590264", "Proof|ID:2135590264-1"},
   CellID->88523793]},
 "Proof|ID:2135590264-1"->{
  Cell[879319, 25166, 3338, 95, 70, "ProofInfo",
   CellTags->{"Proof|ID:2135590264", "Proof|ID:2135590264-1"},
   CellID->88523793]},
 "Proof|ID:2092106183"->{
  Cell[883178, 25280, 4115, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:2092106183", "Proof|ID:2092106183-1"},
   CellID->157583881]},
 "Proof|ID:2092106183-1"->{
  Cell[883178, 25280, 4115, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:2092106183", "Proof|ID:2092106183-1"},
   CellID->157583881]},
 "Proof|ID:2014193763"->{
  Cell[887815, 25408, 4341, 113, 70, "ProofInfo",
   CellTags->{"Proof|ID:2014193763", "Proof|ID:2014193763-1"},
   CellID->1352301682]},
 "Proof|ID:2014193763-1"->{
  Cell[887815, 25408, 4341, 113, 70, "ProofInfo",
   CellTags->{"Proof|ID:2014193763", "Proof|ID:2014193763-1"},
   CellID->1352301682]},
 "Proof|ID:1650576647"->{
  Cell[892678, 25540, 4551, 117, 70, "ProofInfo",
   CellTags->{"Proof|ID:1650576647", "Proof|ID:1650576647-1"},
   CellID->1429121459]},
 "Proof|ID:1650576647-1"->{
  Cell[892678, 25540, 4551, 117, 70, "ProofInfo",
   CellTags->{"Proof|ID:1650576647", "Proof|ID:1650576647-1"},
   CellID->1429121459]},
 "ID:45249416"->{
  Cell[899076, 25708, 641, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:45249416", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->45249416]},
 "ID:248026320"->{
  Cell[899720, 25726, 3991, 108, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:248026320", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "induction on \[ScriptCapitalP] ind"},
   CellID->248026320]},
 "induction on \[ScriptCapitalP] ind"->{
  Cell[899720, 25726, 3991, 108, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:248026320", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "induction on \[ScriptCapitalP] ind"},
   CellID->248026320]},
 "ID:1502331257"->{
  Cell[903714, 25836, 1918, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502331257", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "induction on \[ScriptCapitalP]", "IR"},
   CellID->1502331257]},
 "induction on \[ScriptCapitalP]"->{
  Cell[903714, 25836, 1918, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502331257", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "induction on \[ScriptCapitalP]", "IR"},
   CellID->1502331257]},
 "Proof|ID:248026320"->{
  Cell[905789, 25906, 9001, 211, 70, "ProofInfo",
   CellTags->{"Proof|ID:248026320", "Proof|ID:248026320-1"},
   CellID->717333436]},
 "Proof|ID:248026320-1"->{
  Cell[905789, 25906, 9001, 211, 70, "ProofInfo",
   CellTags->{"Proof|ID:248026320", "Proof|ID:248026320-1"},
   CellID->717333436]},
 "Proof|ID:1502331257"->{
  Cell[915310, 26136, 5045, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:1502331257", "Proof|ID:1502331257-1"},
   CellID->2113840420]},
 "Proof|ID:1502331257-1"->{
  Cell[915310, 26136, 5045, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:1502331257", "Proof|ID:1502331257-1"},
   CellID->2113840420]},
 "ID:411993095"->{
  Cell[921541, 26309, 766, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:411993095", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->411993095]},
 "ID:470022203"->{
  Cell[922310, 26331, 1725, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:470022203", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 1"},
   CellID->470022203]},
 "\[CenterDot] 1"->{
  Cell[922310, 26331, 1725, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:470022203", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 1"},
   CellID->470022203]},
 "ID:401540586"->{
  Cell[924038, 26382, 847, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:401540586", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->401540586]},
 "ID:2014025290"->{
  Cell[924888, 26407, 1468, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2014025290", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 6"},
   CellID->2014025290]},
 "isMonomial 6"->{
  Cell[924888, 26407, 1468, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2014025290", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 6"},
   CellID->2014025290]},
 "ID:1590906878"->{
  Cell[926359, 26448, 870, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1590906878", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 10"},
   CellID->1590906878]},
 "isMonomial 10"->{
  Cell[926359, 26448, 870, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1590906878", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 10"},
   CellID->1590906878]},
 "ID:1997927583"->{
  Cell[927232, 26473, 1279, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1997927583", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 11"},
   CellID->1997927583]},
 "isMonomial 11"->{
  Cell[927232, 26473, 1279, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1997927583", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 11"},
   CellID->1997927583]},
 "Proof|ID:470022203"->{
  Cell[928668, 26522, 4412, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:470022203", "Proof|ID:470022203-1"},
   CellID->356771522]},
 "Proof|ID:470022203-1"->{
  Cell[928668, 26522, 4412, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:470022203", "Proof|ID:470022203-1"},
   CellID->356771522]},
 "Proof|ID:2014025290"->{
  Cell[933601, 26655, 4838, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:2014025290", "Proof|ID:2014025290-1"},
   CellID->740297765]},
 "Proof|ID:2014025290-1"->{
  Cell[933601, 26655, 4838, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:2014025290", "Proof|ID:2014025290-1"},
   CellID->740297765]},
 "Proof|ID:1590906878"->{
  Cell[938960, 26795, 2958, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:1590906878", "Proof|ID:1590906878-1"},
   CellID->1050945896]},
 "Proof|ID:1590906878-1"->{
  Cell[938960, 26795, 2958, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:1590906878", "Proof|ID:1590906878-1"},
   CellID->1050945896]},
 "Proof|ID:1997927583"->{
  Cell[942440, 26899, 4105, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:1997927583", "Proof|ID:1997927583-1"},
   CellID->646926070]},
 "Proof|ID:1997927583-1"->{
  Cell[942440, 26899, 4105, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:1997927583", "Proof|ID:1997927583-1"},
   CellID->646926070]},
 "ID:63039386"->{
  Cell[947818, 27051, 932, 24, 66, "GlobalDeclaration",
   CellTags->{
    "ID:63039386", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->63039386]},
 "ID:1093453827"->{
  Cell[949017, 27084, 926, 24, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1093453827", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 0"},
   CellID->1093453827]},
 "\[CircleTimes] 0"->{
  Cell[949017, 27084, 926, 24, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1093453827", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 0"},
   CellID->1093453827]},
 "ID:1795167993"->{
  Cell[949946, 27110, 1902, 54, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1795167993", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 5"},
   CellID->1795167993]},
 "\[CircleTimes] 5"->{
  Cell[949946, 27110, 1902, 54, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1795167993", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 5"},
   CellID->1795167993]},
 "ID:838653570"->{
  Cell[951851, 27166, 1575, 45, 76, "FormalTextInputFormula",
   CellTags->{
    "ID:838653570", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomial multiplication ind"},
   CellID->838653570]},
 "membership monomial multiplication ind"->{
  Cell[951851, 27166, 1575, 45, 76, "FormalTextInputFormula",
   CellTags->{
    "ID:838653570", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomial multiplication ind"},
   CellID->838653570]},
 "ID:1903131381"->{
  Cell[953429, 27213, 987, 27, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1903131381", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomial multiplication", "IR"},
   CellID->1903131381]},
 "membership monomial multiplication"->{
  Cell[953429, 27213, 987, 27, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1903131381", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomial multiplication", "IR"},
   CellID->1903131381]},
 "ID:1641583144"->{
  Cell[954419, 27242, 2017, 58, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:1641583144", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C \[CircleTimes]"},
   CellID->1641583144]},
 "C \[CircleTimes]"->{
  Cell[954419, 27242, 2017, 58, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:1641583144", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C \[CircleTimes]"},
   CellID->1641583144]},
 "ID:608153969"->{
  Cell[956439, 27302, 1233, 35, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:608153969", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 1"},
   CellID->608153969]},
 "\[CircleTimes] 1"->{
  Cell[956439, 27302, 1233, 35, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:608153969", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 1"},
   CellID->608153969]},
 "ID:1439631766"->{
  Cell[957675, 27339, 1673, 48, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1439631766", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] isMonomial"},
   CellID->1439631766]},
 "\[CircleTimes] isMonomial"->{
  Cell[957675, 27339, 1673, 48, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1439631766", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] isMonomial"},
   CellID->1439631766]},
 "ID:607306392"->{
  Cell[959351, 27389, 736, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:607306392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->607306392]},
 "ID:1312761442"->{
  Cell[960090, 27410, 1163, 32, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1312761442", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 2"},
   CellID->1312761442]},
 "\[CircleTimes] 2"->{
  Cell[960090, 27410, 1163, 32, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1312761442", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 2"},
   CellID->1312761442]},
 "ID:259600169"->{
  Cell[961256, 27444, 1345, 41, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:259600169", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 3"},
   CellID->259600169]},
 "\[CircleTimes] 3"->{
  Cell[961256, 27444, 1345, 41, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:259600169", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 3"},
   CellID->259600169]},
 "ID:1576857420"->{
  Cell[962604, 27487, 999, 27, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1576857420", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 6"},
   CellID->1576857420]},
 "\[CircleTimes] 6"->{
  Cell[962604, 27487, 999, 27, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1576857420", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 6"},
   CellID->1576857420]},
 "ID:483781382"->{
  Cell[963606, 27516, 1844, 55, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:483781382", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 10"},
   CellID->483781382]},
 "\[CircleTimes] 10"->{
  Cell[963606, 27516, 1844, 55, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:483781382", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 10"},
   CellID->483781382]},
 "ID:169057028"->{
  Cell[965453, 27573, 1607, 46, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:169057028", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 11"},
   CellID->169057028]},
 "\[CircleTimes] 11"->{
  Cell[965453, 27573, 1607, 46, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:169057028", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 11"},
   CellID->169057028]},
 "ID:324898493"->{
  Cell[967063, 27621, 721, 19, 31, "GlobalDeclaration",
   CellTags->{
    "ID:324898493", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->324898493]},
 "ID:684947841"->{
  Cell[967787, 27642, 1332, 39, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:684947841", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 7"},
   CellID->684947841]},
 "\[CircleTimes] 7"->{
  Cell[967787, 27642, 1332, 39, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:684947841", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 7"},
   CellID->684947841]},
 "ID:1856294157"->{
  Cell[969122, 27683, 1492, 43, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1856294157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 8"},
   CellID->1856294157]},
 "\[CircleTimes] 8"->{
  Cell[969122, 27683, 1492, 43, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:1856294157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 8"},
   CellID->1856294157]},
 "ID:1768460531"->{
  Cell[970617, 27728, 1138, 31, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1768460531", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 9"},
   CellID->1768460531]},
 "\[CircleTimes] 9"->{
  Cell[970617, 27728, 1138, 31, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1768460531", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 9"},
   CellID->1768460531]},
 "Proof|ID:1093453827"->{
  Cell[971911, 27770, 3078, 87, 28, "ProofInfo",
   CellTags->{"Proof|ID:1093453827", "Proof|ID:1093453827-1"},
   CellID->1691805525]},
 "Proof|ID:1093453827-1"->{
  Cell[971911, 27770, 3078, 87, 28, "ProofInfo",
   CellTags->{"Proof|ID:1093453827", "Proof|ID:1093453827-1"},
   CellID->1691805525]},
 "Proof|ID:1795167993"->{
  Cell[975511, 27876, 6284, 150, 26, "ProofInfo",
   CellTags->{"Proof|ID:1795167993", "Proof|ID:1795167993-1"},
   CellID->1536044778]},
 "Proof|ID:1795167993-1"->{
  Cell[975511, 27876, 6284, 150, 26, "ProofInfo",
   CellTags->{"Proof|ID:1795167993", "Proof|ID:1795167993-1"},
   CellID->1536044778]},
 "Proof|ID:838653570"->{
  Cell[982317, 28045, 4620, 116, 26, "ProofInfo",
   CellTags->{"Proof|ID:838653570", "Proof|ID:838653570-1"},
   CellID->39930568]},
 "Proof|ID:838653570-1"->{
  Cell[982317, 28045, 4620, 116, 26, "ProofInfo",
   CellTags->{"Proof|ID:838653570", "Proof|ID:838653570-1"},
   CellID->39930568]},
 "Proof|ID:1903131381"->{
  Cell[987457, 28180, 3424, 92, 26, "ProofInfo",
   CellTags->{"Proof|ID:1903131381", "Proof|ID:1903131381-1"},
   CellID->443389413]},
 "Proof|ID:1903131381-1"->{
  Cell[987457, 28180, 3424, 92, 26, "ProofInfo",
   CellTags->{"Proof|ID:1903131381", "Proof|ID:1903131381-1"},
   CellID->443389413]},
 "Proof|ID:1641583144"->{
  Cell[991402, 28291, 6602, 156, 26, "ProofInfo",
   CellTags->{"Proof|ID:1641583144", "Proof|ID:1641583144-1"},
   CellID->183185032]},
 "Proof|ID:1641583144-1"->{
  Cell[991402, 28291, 6602, 156, 26, "ProofInfo",
   CellTags->{"Proof|ID:1641583144", "Proof|ID:1641583144-1"},
   CellID->183185032]},
 "Proof|ID:608153969"->{
  Cell[998525, 28466, 3069, 86, 26, "ProofInfo",
   CellTags->{"Proof|ID:608153969", "Proof|ID:608153969-1"},
   CellID->1418277402]},
 "Proof|ID:608153969-1"->{
  Cell[998525, 28466, 3069, 86, 26, "ProofInfo",
   CellTags->{"Proof|ID:608153969", "Proof|ID:608153969-1"},
   CellID->1418277402]},
 "Proof|ID:1439631766"->{
  Cell[1002114, 28571, 5467, 134, 26, "ProofInfo",
   CellTags->{"Proof|ID:1439631766", "Proof|ID:1439631766-1"},
   CellID->821863036]},
 "Proof|ID:1439631766-1"->{
  Cell[1002114, 28571, 5467, 134, 26, "ProofInfo",
   CellTags->{"Proof|ID:1439631766", "Proof|ID:1439631766-1"},
   CellID->821863036]},
 "Proof|ID:1312761442"->{
  Cell[1008103, 28724, 4865, 123, 26, "ProofInfo",
   CellTags->{"Proof|ID:1312761442", "Proof|ID:1312761442-1"},
   CellID->1943269854]},
 "Proof|ID:1312761442-1"->{
  Cell[1008103, 28724, 4865, 123, 26, "ProofInfo",
   CellTags->{"Proof|ID:1312761442", "Proof|ID:1312761442-1"},
   CellID->1943269854]},
 "Proof|ID:259600169"->{
  Cell[1013489, 28866, 4372, 112, 26, "ProofInfo",
   CellTags->{"Proof|ID:259600169", "Proof|ID:259600169-1"},
   CellID->592189180]},
 "Proof|ID:259600169-1"->{
  Cell[1013489, 28866, 4372, 112, 26, "ProofInfo",
   CellTags->{"Proof|ID:259600169", "Proof|ID:259600169-1"},
   CellID->592189180]},
 "Proof|ID:1576857420"->{
  Cell[1018382, 28997, 3868, 103, 26, "ProofInfo",
   CellTags->{"Proof|ID:1576857420", "Proof|ID:1576857420-1"},
   CellID->1943524802]},
 "Proof|ID:1576857420-1"->{
  Cell[1018382, 28997, 3868, 103, 26, "ProofInfo",
   CellTags->{"Proof|ID:1576857420", "Proof|ID:1576857420-1"},
   CellID->1943524802]},
 "Proof|ID:483781382"->{
  Cell[1022771, 29119, 5985, 148, 26, "ProofInfo",
   CellTags->{"Proof|ID:483781382", "Proof|ID:483781382-1"},
   CellID->507841304]},
 "Proof|ID:483781382-1"->{
  Cell[1022771, 29119, 5985, 148, 26, "ProofInfo",
   CellTags->{"Proof|ID:483781382", "Proof|ID:483781382-1"},
   CellID->507841304]},
 "Proof|ID:169057028"->{
  Cell[1029277, 29286, 5538, 136, 26, "ProofInfo",
   CellTags->{"Proof|ID:169057028", "Proof|ID:169057028-1"},
   CellID->1561081111]},
 "Proof|ID:169057028-1"->{
  Cell[1029277, 29286, 5538, 136, 26, "ProofInfo",
   CellTags->{"Proof|ID:169057028", "Proof|ID:169057028-1"},
   CellID->1561081111]},
 "Proof|ID:684947841"->{
  Cell[1035336, 29441, 5004, 124, 26, "ProofInfo",
   CellTags->{"Proof|ID:684947841", "Proof|ID:684947841-1"},
   CellID->969684081]},
 "Proof|ID:684947841-1"->{
  Cell[1035336, 29441, 5004, 124, 26, "ProofInfo",
   CellTags->{"Proof|ID:684947841", "Proof|ID:684947841-1"},
   CellID->969684081]},
 "Proof|ID:1856294157"->{
  Cell[1040861, 29584, 5622, 137, 26, "ProofInfo",
   CellTags->{"Proof|ID:1856294157", "Proof|ID:1856294157-1"},
   CellID->1121162524]},
 "Proof|ID:1856294157-1"->{
  Cell[1040861, 29584, 5622, 137, 26, "ProofInfo",
   CellTags->{"Proof|ID:1856294157", "Proof|ID:1856294157-1"},
   CellID->1121162524]},
 "Proof|ID:1768460531"->{
  Cell[1047071, 29741, 4397, 111, 26, "ProofInfo",
   CellTags->{"Proof|ID:1768460531", "Proof|ID:1768460531-1"},
   CellID->1517299564]},
 "Proof|ID:1768460531-1"->{
  Cell[1047071, 29741, 4397, 111, 26, "ProofInfo",
   CellTags->{"Proof|ID:1768460531", "Proof|ID:1768460531-1"},
   CellID->1517299564]},
 "ID:1437523678"->{
  Cell[1075368, 30600, 5057, 136, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1437523678", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplicationRec"},
   CellID->1437523678]},
 "hasMultiplicationRec"->{
  Cell[1075368, 30600, 5057, 136, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1437523678", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplicationRec"},
   CellID->1437523678]},
 "ID:1372918741"->{
  Cell[1081114, 30766, 710, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1372918741", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1372918741]},
 "ID:271668948"->{
  Cell[1081827, 30786, 1824, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:271668948", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "316"},
   CellID->271668948]},
 "316"->{
  Cell[1081827, 30786, 1824, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:271668948", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "316"},
   CellID->271668948]},
 "ID:428541408"->{
  Cell[1083654, 30843, 1263, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:428541408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "317"},
   CellID->428541408]},
 "317"->{
  Cell[1083654, 30843, 1263, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:428541408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "317"},
   CellID->428541408]},
 "ID:1983432019"->{
  Cell[1084920, 30883, 800, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1983432019", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "318"},
   CellID->1983432019]},
 "318"->{
  Cell[1084920, 30883, 800, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1983432019", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "318"},
   CellID->1983432019]},
 "Proof|ID:271668948"->{
  Cell[1085877, 30916, 5696, 142, 70, "ProofInfo",
   CellTags->{"Proof|ID:271668948", "Proof|ID:271668948-1"},
   CellID->1133492443]},
 "Proof|ID:271668948-1"->{
  Cell[1085877, 30916, 5696, 142, 70, "ProofInfo",
   CellTags->{"Proof|ID:271668948", "Proof|ID:271668948-1"},
   CellID->1133492443]},
 "Proof|ID:428541408"->{
  Cell[1092094, 31077, 3896, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:428541408", "Proof|ID:428541408-1"},
   CellID->738256620]},
 "Proof|ID:428541408-1"->{
  Cell[1092094, 31077, 3896, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:428541408", "Proof|ID:428541408-1"},
   CellID->738256620]},
 "Proof|ID:1983432019"->{
  Cell[1096511, 31200, 2016, 63, 70, "ProofInfo",
   CellTags->{"Proof|ID:1983432019", "Proof|ID:1983432019-1"},
   CellID->836120509]},
 "Proof|ID:1983432019-1"->{
  Cell[1096511, 31200, 2016, 63, 70, "ProofInfo",
   CellTags->{"Proof|ID:1983432019", "Proof|ID:1983432019-1"},
   CellID->836120509]},
 "ID:1310550300"->{
  Cell[1100663, 31325, 1567, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1310550300", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "319"},
   CellID->1310550300]},
 "319"->{
  Cell[1100663, 31325, 1567, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1310550300", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "319"},
   CellID->1310550300]},
 "ID:978297022"->{
  Cell[1103434, 31417, 835, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:978297022", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplication \[ScriptCapitalP]"},
   CellID->978297022]},
 "hasMultiplication \[ScriptCapitalP]"->{
  Cell[1103434, 31417, 835, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:978297022", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplication \[ScriptCapitalP]"},
   CellID->978297022]},
 "ID:1125626334"->{
  Cell[1104880, 31464, 1495, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1125626334", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "320"},
   CellID->1125626334]},
 "320"->{
  Cell[1104880, 31464, 1495, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1125626334", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "320"},
   CellID->1125626334]},
 "Proof|ID:1125626334"->{
  Cell[1106532, 31512, 2944, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1125626334", "Proof|ID:1125626334-1"},
   CellID->372685725]},
 "Proof|ID:1125626334-1"->{
  Cell[1106532, 31512, 2944, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1125626334", "Proof|ID:1125626334-1"},
   CellID->372685725]},
 "ID:1103171586"->{
  Cell[1110563, 31634, 1031, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1103171586", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] closed \[ScriptCapitalP]"},
   CellID->1103171586]},
 "\[RawStar] closed \[ScriptCapitalP]"->{
  Cell[1110563, 31634, 1031, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1103171586", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] closed \[ScriptCapitalP]"},
   CellID->1103171586]},
 "Proof|ID:1103171586"->{
  Cell[1111752, 31676, 1955, 61, 70, "ProofInfo",
   CellTags->{"Proof|ID:1103171586", "Proof|ID:1103171586-1"},
   CellID->134774430]},
 "Proof|ID:1103171586-1"->{
  Cell[1111752, 31676, 1955, 61, 70, "ProofInfo",
   CellTags->{"Proof|ID:1103171586", "Proof|ID:1103171586-1"},
   CellID->134774430]},
 "ID:2097953227"->{
  Cell[1115479, 31797, 1392, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2097953227", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "322"},
   CellID->2097953227]},
 "322"->{
  Cell[1115479, 31797, 1392, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2097953227", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "322"},
   CellID->2097953227]},
 "Proof|ID:2097953227"->{
  Cell[1117027, 31850, 4333, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:2097953227", "Proof|ID:2097953227-1"},
   CellID->163286988]},
 "Proof|ID:2097953227-1"->{
  Cell[1117027, 31850, 4333, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:2097953227", "Proof|ID:2097953227-1"},
   CellID->163286988]},
 "ID:587826938"->{
  Cell[1122684, 32012, 762, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:587826938", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->587826938]},
 "ID:384020988"->{
  Cell[1123449, 32034, 885, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:384020988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "324"},
   CellID->384020988]},
 "324"->{
  Cell[1123449, 32034, 885, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:384020988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "324"},
   CellID->384020988]},
 "ID:205407574"->{
  Cell[1124337, 32060, 1094, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:205407574", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "325"},
   CellID->205407574]},
 "325"->{
  Cell[1124337, 32060, 1094, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:205407574", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "325"},
   CellID->205407574]},
 "ID:411846169"->{
  Cell[1125434, 32095, 1070, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:411846169", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "327"},
   CellID->411846169]},
 "327"->{
  Cell[1125434, 32095, 1070, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:411846169", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "327"},
   CellID->411846169]},
 "ID:1300778952"->{
  Cell[1126507, 32130, 1086, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1300778952", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "328"},
   CellID->1300778952]},
 "328"->{
  Cell[1126507, 32130, 1086, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1300778952", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "328"},
   CellID->1300778952]},
 "ID:2123221051"->{
  Cell[1127596, 32165, 1240, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2123221051", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "329"},
   CellID->2123221051]},
 "329"->{
  Cell[1127596, 32165, 1240, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2123221051", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "329"},
   CellID->2123221051]},
 "Proof|ID:384020988"->{
  Cell[1128993, 32213, 2632, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:384020988", "Proof|ID:384020988-1"},
   CellID->119010279]},
 "Proof|ID:384020988-1"->{
  Cell[1128993, 32213, 2632, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:384020988", "Proof|ID:384020988-1"},
   CellID->119010279]},
 "Proof|ID:205407574"->{
  Cell[1132143, 32310, 3617, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:205407574", "Proof|ID:205407574-1"},
   CellID->2086485184]},
 "Proof|ID:205407574-1"->{
  Cell[1132143, 32310, 3617, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:205407574", "Proof|ID:205407574-1"},
   CellID->2086485184]},
 "Proof|ID:411846169"->{
  Cell[1136280, 32427, 3616, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:411846169", "Proof|ID:411846169-1"},
   CellID->183622553]},
 "Proof|ID:411846169-1"->{
  Cell[1136280, 32427, 3616, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:411846169", "Proof|ID:411846169-1"},
   CellID->183622553]},
 "Proof|ID:1300778952"->{
  Cell[1140416, 32544, 3599, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1300778952", "Proof|ID:1300778952-1"},
   CellID->695116774]},
 "Proof|ID:1300778952-1"->{
  Cell[1140416, 32544, 3599, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1300778952", "Proof|ID:1300778952-1"},
   CellID->695116774]},
 "Proof|ID:2123221051"->{
  Cell[1144537, 32662, 3199, 91, 70, "ProofInfo",
   CellTags->{"Proof|ID:2123221051", "Proof|ID:2123221051-1"},
   CellID->511445547]},
 "Proof|ID:2123221051-1"->{
  Cell[1144537, 32662, 3199, 91, 70, "ProofInfo",
   CellTags->{"Proof|ID:2123221051", "Proof|ID:2123221051-1"},
   CellID->511445547]},
 "ID:840093990"->{
  Cell[1148881, 32793, 1065, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:840093990", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] comm \[ScriptCapitalP]"},
   CellID->840093990]},
 "\[RawStar] comm \[ScriptCapitalP]"->{
  Cell[1148881, 32793, 1065, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:840093990", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] comm \[ScriptCapitalP]"},
   CellID->840093990]},
 "Proof|ID:840093990"->{
  Cell[1150103, 32835, 1936, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:840093990", "Proof|ID:840093990-1"},
   CellID->4680134]},
 "Proof|ID:840093990-1"->{
  Cell[1150103, 32835, 1936, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:840093990", "Proof|ID:840093990-1"},
   CellID->4680134]},
 "ID:1730144032"->{
  Cell[1153515, 32952, 1230, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1730144032", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] distr \[ScriptCapitalP]"},
   CellID->1730144032]},
 "\[RawStar] distr \[ScriptCapitalP]"->{
  Cell[1153515, 32952, 1230, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1730144032", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] distr \[ScriptCapitalP]"},
   CellID->1730144032]},
 "Proof|ID:1730144032"->{
  Cell[1154902, 33002, 2070, 63, 70, "ProofInfo",
   CellTags->{"Proof|ID:1730144032", "Proof|ID:1730144032-1"},
   CellID->1430620892]},
 "Proof|ID:1730144032-1"->{
  Cell[1154902, 33002, 2070, 63, 70, "ProofInfo",
   CellTags->{"Proof|ID:1730144032", "Proof|ID:1730144032-1"},
   CellID->1430620892]},
 "ID:378610433"->{
  Cell[1158481, 33119, 1156, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:378610433", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1 neutral \[ScriptCapitalP]"},
   CellID->378610433]},
 "1 neutral \[ScriptCapitalP]"->{
  Cell[1158481, 33119, 1156, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:378610433", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1 neutral \[ScriptCapitalP]"},
   CellID->378610433]},
 "Proof|ID:378610433"->{
  Cell[1159793, 33163, 2068, 62, 70, "ProofInfo",
   CellTags->{"Proof|ID:378610433", "Proof|ID:378610433-1"},
   CellID->149156064]},
 "Proof|ID:378610433-1"->{
  Cell[1159793, 33163, 2068, 62, 70, "ProofInfo",
   CellTags->{"Proof|ID:378610433", "Proof|ID:378610433-1"},
   CellID->149156064]},
 "ID:687899447"->{
  Cell[1163112, 33271, 941, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:687899447", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->687899447]},
 "ID:421376854"->{
  Cell[1164056, 33297, 1549, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:421376854", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "330"},
   CellID->421376854]},
 "330"->{
  Cell[1164056, 33297, 1549, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:421376854", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "330"},
   CellID->421376854]},
 "ID:935511285"->{
  Cell[1165608, 33344, 1765, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:935511285", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "331"},
   CellID->935511285]},
 "331"->{
  Cell[1165608, 33344, 1765, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:935511285", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "331"},
   CellID->935511285]},
 "ID:1702504479"->{
  Cell[1167376, 33400, 1338, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1702504479", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "332"},
   CellID->1702504479]},
 "332"->{
  Cell[1167376, 33400, 1338, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1702504479", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "332"},
   CellID->1702504479]},
 "ID:1379089132"->{
  Cell[1168717, 33442, 1912, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1379089132", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "333"},
   CellID->1379089132]},
 "333"->{
  Cell[1168717, 33442, 1912, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1379089132", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "333"},
   CellID->1379089132]},
 "Proof|ID:421376854"->{
  Cell[1170786, 33511, 4903, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:421376854", "Proof|ID:421376854-1"},
   CellID->1691973429]},
 "Proof|ID:421376854-1"->{
  Cell[1170786, 33511, 4903, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:421376854", "Proof|ID:421376854-1"},
   CellID->1691973429]},
 "Proof|ID:935511285"->{
  Cell[1176209, 33654, 5709, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:935511285", "Proof|ID:935511285-1"},
   CellID->1225679565]},
 "Proof|ID:935511285-1"->{
  Cell[1176209, 33654, 5709, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:935511285", "Proof|ID:935511285-1"},
   CellID->1225679565]},
 "Proof|ID:1702504479"->{
  Cell[1182479, 33814, 4528, 117, 70, "ProofInfo",
   CellTags->{"Proof|ID:1702504479", "Proof|ID:1702504479-1"},
   CellID->1881341258]},
 "Proof|ID:1702504479-1"->{
  Cell[1182479, 33814, 4528, 117, 70, "ProofInfo",
   CellTags->{"Proof|ID:1702504479", "Proof|ID:1702504479-1"},
   CellID->1881341258]},
 "Proof|ID:1379089132"->{
  Cell[1187528, 33950, 5346, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1379089132", "Proof|ID:1379089132-1"},
   CellID->1247171944]},
 "Proof|ID:1379089132-1"->{
  Cell[1187528, 33950, 5346, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1379089132", "Proof|ID:1379089132-1"},
   CellID->1247171944]},
 "ID:167864408"->{
  Cell[1194015, 34125, 1071, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:167864408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] assoc \[ScriptCapitalP]"},
   CellID->167864408]},
 "\[RawStar] assoc \[ScriptCapitalP]"->{
  Cell[1194015, 34125, 1071, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:167864408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] assoc \[ScriptCapitalP]"},
   CellID->167864408]},
 "Proof|ID:167864408"->{
  Cell[1195244, 34167, 1943, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:167864408", "Proof|ID:167864408-1"},
   CellID->1541468547]},
 "Proof|ID:167864408-1"->{
  Cell[1195244, 34167, 1943, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:167864408", "Proof|ID:167864408-1"},
   CellID->1541468547]},
 "ID:1082265416"->{
  Cell[1198459, 34271, 820, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1082265416", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCommRing1 \[ScriptCapitalP]"},
   CellID->1082265416]},
 "isCommRing1 \[ScriptCapitalP]"->{
  Cell[1198459, 34271, 820, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1082265416", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCommRing1 \[ScriptCapitalP]"},
   CellID->1082265416]},
 "Proof|ID:1082265416"->{
  Cell[1199435, 34303, 1810, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1082265416", "Proof|ID:1082265416-1"},
   CellID->408658241]},
 "Proof|ID:1082265416-1"->{
  Cell[1199435, 34303, 1810, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1082265416", "Proof|ID:1082265416-1"},
   CellID->408658241]},
 "ID:1610244238"->{
  Cell[1203285, 34427, 1021, 27, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1610244238", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1610244238]},
 "ID:1259587933"->{
  Cell[1204309, 34456, 2470, 77, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1259587933", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 5"},
   CellID->1259587933]},
 "R 5"->{
  Cell[1204309, 34456, 2470, 77, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1259587933", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 5"},
   CellID->1259587933]},
 "ID:859675419"->{
  Cell[1206782, 34535, 2175, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:859675419", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6 ind"},
   CellID->859675419]},
 "C 6 ind"->{
  Cell[1206782, 34535, 2175, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:859675419", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6 ind"},
   CellID->859675419]},
 "ID:515125163"->{
  Cell[1208960, 34604, 1733, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:515125163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6"},
   CellID->515125163]},
 "C 6"->{
  Cell[1208960, 34604, 1733, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:515125163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6"},
   CellID->515125163]},
 "ID:1572844749"->{
  Cell[1210696, 34656, 1516, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1572844749", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6'"},
   CellID->1572844749]},
 "C 6'"->{
  Cell[1210696, 34656, 1516, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1572844749", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6'"},
   CellID->1572844749]},
 "ID:771023286"->{
  Cell[1212215, 34700, 1674, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:771023286", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 4"},
   CellID->771023286]},
 "lp 4"->{
  Cell[1212215, 34700, 1674, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:771023286", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 4"},
   CellID->771023286]},
 "ID:422533310"->{
  Cell[1213892, 34747, 1308, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:422533310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 9"},
   CellID->422533310]},
 "H 9"->{
  Cell[1213892, 34747, 1308, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:422533310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 9"},
   CellID->422533310]},
 "Proof|ID:1259587933"->{
  Cell[1215357, 34793, 6509, 157, 70, "ProofInfo",
   CellTags->{"Proof|ID:1259587933", "Proof|ID:1259587933-1"},
   CellID->1687842700]},
 "Proof|ID:1259587933-1"->{
  Cell[1215357, 34793, 6509, 157, 70, "ProofInfo",
   CellTags->{"Proof|ID:1259587933", "Proof|ID:1259587933-1"},
   CellID->1687842700]},
 "Proof|ID:859675419"->{
  Cell[1222388, 34969, 6354, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:859675419", "Proof|ID:859675419-1"},
   CellID->934287773]},
 "Proof|ID:859675419-1"->{
  Cell[1222388, 34969, 6354, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:859675419", "Proof|ID:859675419-1"},
   CellID->934287773]},
 "Proof|ID:515125163"->{
  Cell[1229263, 35142, 5531, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:515125163", "Proof|ID:515125163-1"},
   CellID->50230006]},
 "Proof|ID:515125163-1"->{
  Cell[1229263, 35142, 5531, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:515125163", "Proof|ID:515125163-1"},
   CellID->50230006]},
 "Proof|ID:1572844749"->{
  Cell[1235315, 35299, 5495, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:1572844749", "Proof|ID:1572844749-1"},
   CellID->1143100247]},
 "Proof|ID:1572844749-1"->{
  Cell[1235315, 35299, 5495, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:1572844749", "Proof|ID:1572844749-1"},
   CellID->1143100247]},
 "Proof|ID:771023286"->{
  Cell[1241332, 35455, 5693, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:771023286", "Proof|ID:771023286-1"},
   CellID->38527445]},
 "Proof|ID:771023286-1"->{
  Cell[1241332, 35455, 5693, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:771023286", "Proof|ID:771023286-1"},
   CellID->38527445]},
 "Proof|ID:422533310"->{
  Cell[1247546, 35614, 4828, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:422533310", "Proof|ID:422533310-1"},
   CellID->156791201]},
 "Proof|ID:422533310-1"->{
  Cell[1247546, 35614, 4828, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:422533310", "Proof|ID:422533310-1"},
   CellID->156791201]},
 "ID:1667882379"->{
  Cell[1253568, 35779, 766, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1667882379", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1667882379]},
 "ID:1105090112"->{
  Cell[1254337, 35801, 739, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1105090112", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1105090112]},
 "ID:878771843"->{
  Cell[1255079, 35822, 1565, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:878771843", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 2"},
   CellID->878771843]},
 "\[CenterDot] 2"->{
  Cell[1255079, 35822, 1565, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:878771843", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 2"},
   CellID->878771843]},
 "ID:1863009930"->{
  Cell[1256647, 35871, 868, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1863009930", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1863009930]},
 "ID:1958431161"->{
  Cell[1257518, 35896, 1713, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1958431161", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 5"},
   CellID->1958431161]},
 "\[CenterDot] 5"->{
  Cell[1257518, 35896, 1713, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1958431161", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 5"},
   CellID->1958431161]},
 "ID:758577094"->{
  Cell[1259234, 35947, 1464, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:758577094", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 6"},
   CellID->758577094]},
 "\[CenterDot] 6"->{
  Cell[1259234, 35947, 1464, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:758577094", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 6"},
   CellID->758577094]},
 "ID:544874520"->{
  Cell[1260701, 35989, 736, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:544874520", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->544874520]},
 "ID:158305602"->{
  Cell[1261440, 36010, 738, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:158305602", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->158305602]},
 "ID:347486225"->{
  Cell[1262181, 36031, 1286, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:347486225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 9"},
   CellID->347486225]},
 "\[CenterDot] 9"->{
  Cell[1262181, 36031, 1286, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:347486225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 9"},
   CellID->347486225]},
 "ID:1591957532"->{
  Cell[1263470, 36069, 1453, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1591957532", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 10"},
   CellID->1591957532]},
 "\[CenterDot] 10"->{
  Cell[1263470, 36069, 1453, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1591957532", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 10"},
   CellID->1591957532]},
 "ID:1155967768"->{
  Cell[1264926, 36110, 721, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1155967768", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1155967768]},
 "ID:1942982211"->{
  Cell[1265650, 36131, 1293, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1942982211", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 15"},
   CellID->1942982211]},
 "\[CenterDot] 15"->{
  Cell[1265650, 36131, 1293, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1942982211", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 15"},
   CellID->1942982211]},
 "ID:44857450"->{
  Cell[1266946, 36172, 1193, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:44857450", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 12"},
   CellID->44857450]},
 "\[CenterDot] 12"->{
  Cell[1266946, 36172, 1193, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:44857450", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 12"},
   CellID->44857450]},
 "ID:928737230"->{
  Cell[1268142, 36207, 1246, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:928737230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 13"},
   CellID->928737230]},
 "\[CenterDot] 13"->{
  Cell[1268142, 36207, 1246, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:928737230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 13"},
   CellID->928737230]},
 "ID:233618042"->{
  Cell[1269391, 36243, 1269, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:233618042", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 14"},
   CellID->233618042]},
 "\[CenterDot] 14"->{
  Cell[1269391, 36243, 1269, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:233618042", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 14"},
   CellID->233618042]},
 "Proof|ID:878771843"->{
  Cell[1270817, 36287, 4998, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:878771843", "Proof|ID:878771843-1"},
   CellID->395620750]},
 "Proof|ID:878771843-1"->{
  Cell[1270817, 36287, 4998, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:878771843", "Proof|ID:878771843-1"},
   CellID->395620750]},
 "Proof|ID:1958431161"->{
  Cell[1276335, 36432, 5675, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:1958431161", "Proof|ID:1958431161-1"},
   CellID->756678784]},
 "Proof|ID:1958431161-1"->{
  Cell[1276335, 36432, 5675, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:1958431161", "Proof|ID:1958431161-1"},
   CellID->756678784]},
 "Proof|ID:758577094"->{
  Cell[1282531, 36592, 5360, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:758577094", "Proof|ID:758577094-1"},
   CellID->2118355576]},
 "Proof|ID:758577094-1"->{
  Cell[1282531, 36592, 5360, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:758577094", "Proof|ID:758577094-1"},
   CellID->2118355576]},
 "Proof|ID:347486225"->{
  Cell[1288412, 36743, 5170, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:347486225", "Proof|ID:347486225-1"},
   CellID->1908653227]},
 "Proof|ID:347486225-1"->{
  Cell[1288412, 36743, 5170, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:347486225", "Proof|ID:347486225-1"},
   CellID->1908653227]},
 "Proof|ID:1591957532"->{
  Cell[1294169, 36891, 5346, 133, 70, "ProofInfo",
   CellTags->{"Proof|ID:1591957532", "Proof|ID:1591957532-1"},
   CellID->1145145855]},
 "Proof|ID:1591957532-1"->{
  Cell[1294169, 36891, 5346, 133, 70, "ProofInfo",
   CellTags->{"Proof|ID:1591957532", "Proof|ID:1591957532-1"},
   CellID->1145145855]},
 "Proof|ID:1942982211"->{
  Cell[1300036, 37043, 4432, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1942982211", "Proof|ID:1942982211-1"},
   CellID->1270512064]},
 "Proof|ID:1942982211-1"->{
  Cell[1300036, 37043, 4432, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1942982211", "Proof|ID:1942982211-1"},
   CellID->1270512064]},
 "Proof|ID:44857450"->{
  Cell[1304990, 37177, 5004, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:44857450", "Proof|ID:44857450-1"},
   CellID->584781971]},
 "Proof|ID:44857450-1"->{
  Cell[1304990, 37177, 5004, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:44857450", "Proof|ID:44857450-1"},
   CellID->584781971]},
 "Proof|ID:928737230"->{
  Cell[1310513, 37320, 5009, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:928737230", "Proof|ID:928737230-1"},
   CellID->103471022]},
 "Proof|ID:928737230-1"->{
  Cell[1310513, 37320, 5009, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:928737230", "Proof|ID:928737230-1"},
   CellID->103471022]},
 "Proof|ID:233618042"->{
  Cell[1316043, 37463, 4437, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:233618042", "Proof|ID:233618042-1"},
   CellID->424912135]},
 "Proof|ID:233618042-1"->{
  Cell[1316043, 37463, 4437, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:233618042", "Proof|ID:233618042-1"},
   CellID->424912135]},
 "ID:183428126"->{
  Cell[1321436, 37608, 900, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:183428126", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->183428126]},
 "ID:2026663200"->{
  Cell[1322339, 37634, 1652, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2026663200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 2"},
   CellID->2026663200]},
 "isMonomial 2"->{
  Cell[1322339, 37634, 1652, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2026663200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 2"},
   CellID->2026663200]},
 "ID:1916930407"->{
  Cell[1323994, 37684, 1920, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1916930407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 5"},
   CellID->1916930407]},
 "isMonomial 5"->{
  Cell[1323994, 37684, 1920, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1916930407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 5"},
   CellID->1916930407]},
 "Proof|ID:2026663200"->{
  Cell[1326072, 37751, 5519, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:2026663200", "Proof|ID:2026663200-1"},
   CellID->1385906749]},
 "Proof|ID:2026663200-1"->{
  Cell[1326072, 37751, 5519, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:2026663200", "Proof|ID:2026663200-1"},
   CellID->1385906749]},
 "Proof|ID:1916930407"->{
  Cell[1332112, 37907, 5248, 130, 70, "ProofInfo",
   CellTags->{"Proof|ID:1916930407", "Proof|ID:1916930407-1"},
   CellID->1619770492]},
 "Proof|ID:1916930407-1"->{
  Cell[1332112, 37907, 5248, 130, 70, "ProofInfo",
   CellTags->{"Proof|ID:1916930407", "Proof|ID:1916930407-1"},
   CellID->1619770492]},
 "ID:1225634217"->{
  Cell[1339640, 38106, 830, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1225634217", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduction ring \[ScriptCapitalR]"},
   CellID->1225634217]},
 "reduction ring \[ScriptCapitalR]"->{
  Cell[1339640, 38106, 830, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1225634217", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduction ring \[ScriptCapitalR]"},
   CellID->1225634217]},
 "ID:1542801947"->{
  Cell[1341015, 38152, 2050, 59, 74, "FormalTextInputFormula",
   CellTags->{
    "ID:1542801947", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "300"},
   CellID->1542801947]},
 "300"->{
  Cell[1341015, 38152, 2050, 59, 74, "FormalTextInputFormula",
   CellTags->{
    "ID:1542801947", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "300"},
   CellID->1542801947]},
 "Proof|ID:1542801947"->{
  Cell[1343221, 38222, 4714, 121, 28, "ProofInfo",
   CellTags->{"Proof|ID:1542801947", "Proof|ID:1542801947-1"},
   CellID->1097904678]},
 "Proof|ID:1542801947-1"->{
  Cell[1343221, 38222, 4714, 121, 28, "ProofInfo",
   CellTags->{"Proof|ID:1542801947", "Proof|ID:1542801947-1"},
   CellID->1097904678]},
 "ID:946287998"->{
  Cell[1348887, 38376, 5079, 138, 212, "FormalTextInputFormula",
   CellTags->{
    "ID:946287998", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrderedRec"},
   CellID->946287998]},
 "isOrderedRec"->{
  Cell[1348887, 38376, 5079, 138, 212, "FormalTextInputFormula",
   CellTags->{
    "ID:946287998", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrderedRec"},
   CellID->946287998]},
 "ID:1485445352"->{
  Cell[1354656, 38542, 704, 18, 23, "GlobalDeclaration",
   CellTags->{
    "ID:1485445352", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1485445352]},
 "ID:215225048"->{
  Cell[1355363, 38562, 2882, 86, 82, "FormalTextInputFormula",
   CellTags->{
    "ID:215225048", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "301"},
   CellID->215225048]},
 "301"->{
  Cell[1355363, 38562, 2882, 86, 82, "FormalTextInputFormula",
   CellTags->{
    "ID:215225048", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "301"},
   CellID->215225048]},
 "ID:1704612967"->{
  Cell[1358248, 38650, 2636, 82, 82, "FormalTextInputFormula",
   CellTags->{
    "ID:1704612967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "302"},
   CellID->1704612967]},
 "302"->{
  Cell[1358248, 38650, 2636, 82, 82, "FormalTextInputFormula",
   CellTags->{
    "ID:1704612967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "302"},
   CellID->1704612967]},
 "ID:162248449"->{
  Cell[1360887, 38734, 789, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:162248449", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "303"},
   CellID->162248449]},
 "303"->{
  Cell[1360887, 38734, 789, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:162248449", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "303"},
   CellID->162248449]},
 "Proof|ID:215225048"->{
  Cell[1361833, 38767, 7717, 184, 28, "ProofInfo",
   CellTags->{"Proof|ID:215225048", "Proof|ID:215225048-1"},
   CellID->1531901166]},
 "Proof|ID:215225048-1"->{
  Cell[1361833, 38767, 7717, 184, 28, "ProofInfo",
   CellTags->{"Proof|ID:215225048", "Proof|ID:215225048-1"},
   CellID->1531901166]},
 "Proof|ID:1704612967"->{
  Cell[1370070, 38970, 7742, 187, 26, "ProofInfo",
   CellTags->{"Proof|ID:1704612967", "Proof|ID:1704612967-1"},
   CellID->494680653]},
 "Proof|ID:1704612967-1"->{
  Cell[1370070, 38970, 7742, 187, 26, "ProofInfo",
   CellTags->{"Proof|ID:1704612967", "Proof|ID:1704612967-1"},
   CellID->494680653]},
 "Proof|ID:162248449"->{
  Cell[1378334, 39176, 1975, 62, 26, "ProofInfo",
   CellTags->{"Proof|ID:162248449", "Proof|ID:162248449-1"},
   CellID->798928706]},
 "Proof|ID:162248449-1"->{
  Cell[1378334, 39176, 1975, 62, 26, "ProofInfo",
   CellTags->{"Proof|ID:162248449", "Proof|ID:162248449-1"},
   CellID->798928706]},
 "ID:1015468103"->{
  Cell[1418646, 40401, 1650, 45, 54, "FormalTextInputFormula",
   CellTags->{
    "ID:1015468103", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "304"},
   CellID->1015468103]},
 "304"->{
  Cell[1418646, 40401, 1650, 45, 54, "FormalTextInputFormula",
   CellTags->{
    "ID:1015468103", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "304"},
   CellID->1015468103]},
 "ID:7476052"->{
  Cell[1422978, 40517, 853, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:7476052", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduction-polynomial domain \[ScriptCapitalP]"},
   CellID->7476052]},
 "reduction-polynomial domain \[ScriptCapitalP]"->{
  Cell[1422978, 40517, 853, 23, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:7476052", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduction-polynomial domain \[ScriptCapitalP]"},
   CellID->7476052]},
 "ID:173226898"->{
  Cell[1424978, 40581, 763, 20, 42, "GlobalDeclaration",
   CellTags->{
    "ID:173226898", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->173226898]},
 "ID:835223349"->{
  Cell[1425744, 40603, 738, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:835223349", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->835223349]},
 "ID:181219027"->{
  Cell[1426485, 40624, 929, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:181219027", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "356"},
   CellID->181219027]},
 "356"->{
  Cell[1426485, 40624, 929, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:181219027", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "356"},
   CellID->181219027]},
 "ID:271793724"->{
  Cell[1427417, 40651, 1555, 46, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:271793724", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "382"},
   CellID->271793724]},
 "382"->{
  Cell[1427417, 40651, 1555, 46, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:271793724", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "382"},
   CellID->271793724]},
 "ID:575028756"->{
  Cell[1428975, 40699, 1487, 44, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:575028756", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "382'"},
   CellID->575028756]},
 "382'"->{
  Cell[1428975, 40699, 1487, 44, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:575028756", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "382'"},
   CellID->575028756]},
 "ID:2020716880"->{
  Cell[1430465, 40745, 1606, 48, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:2020716880", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "383"},
   CellID->2020716880]},
 "383"->{
  Cell[1430465, 40745, 1606, 48, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:2020716880", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "383"},
   CellID->2020716880]},
 "ID:1964835443"->{
  Cell[1432074, 40795, 1655, 48, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1964835443", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "381"},
   CellID->1964835443]},
 "381"->{
  Cell[1432074, 40795, 1655, 48, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1964835443", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "381"},
   CellID->1964835443]},
 "ID:1599847484"->{
  Cell[1433732, 40845, 1945, 55, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1599847484", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "380"},
   CellID->1599847484]},
 "380"->{
  Cell[1433732, 40845, 1945, 55, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1599847484", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "380"},
   CellID->1599847484]},
 "ID:590709100"->{
  Cell[1435680, 40902, 3915, 101, 100, "FormalTextInputFormula",
   CellTags->{
    "ID:590709100", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "379"},
   CellID->590709100]},
 "379"->{
  Cell[1435680, 40902, 3915, 101, 100, "FormalTextInputFormula",
   CellTags->{
    "ID:590709100", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "379"},
   CellID->590709100]},
 "ID:586184654"->{
  Cell[1439598, 41005, 1877, 57, 65, "FormalTextInputFormula",
   CellTags->{
    "ID:586184654", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "405"},
   CellID->586184654]},
 "405"->{
  Cell[1439598, 41005, 1877, 57, 65, "FormalTextInputFormula",
   CellTags->{
    "ID:586184654", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "405"},
   CellID->586184654]},
 "ID:919284948"->{
  Cell[1441478, 41064, 1402, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:919284948", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "406"},
   CellID->919284948]},
 "406"->{
  Cell[1441478, 41064, 1402, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:919284948", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "406"},
   CellID->919284948]},
 "Proof|ID:181219027"->{
  Cell[1443037, 41114, 2840, 82, 28, "ProofInfo",
   CellTags->{"Proof|ID:181219027", "Proof|ID:181219027-1"},
   CellID->719304724]},
 "Proof|ID:181219027-1"->{
  Cell[1443037, 41114, 2840, 82, 28, "ProofInfo",
   CellTags->{"Proof|ID:181219027", "Proof|ID:181219027-1"},
   CellID->719304724]},
 "Proof|ID:271793724"->{
  Cell[1446398, 41215, 4687, 120, 26, "ProofInfo",
   CellTags->{"Proof|ID:271793724", "Proof|ID:271793724-1"},
   CellID->1889243650]},
 "Proof|ID:271793724-1"->{
  Cell[1446398, 41215, 4687, 120, 26, "ProofInfo",
   CellTags->{"Proof|ID:271793724", "Proof|ID:271793724-1"},
   CellID->1889243650]},
 "Proof|ID:575028756"->{
  Cell[1451605, 41354, 4332, 112, 26, "ProofInfo",
   CellTags->{"Proof|ID:575028756", "Proof|ID:575028756-1"},
   CellID->993022467]},
 "Proof|ID:575028756-1"->{
  Cell[1451605, 41354, 4332, 112, 26, "ProofInfo",
   CellTags->{"Proof|ID:575028756", "Proof|ID:575028756-1"},
   CellID->993022467]},
 "Proof|ID:2020716880"->{
  Cell[1456575, 41488, 4958, 127, 26, "ProofInfo",
   CellTags->{"Proof|ID:2020716880", "Proof|ID:2020716880-1"},
   CellID->1784894938]},
 "Proof|ID:2020716880-1"->{
  Cell[1456575, 41488, 4958, 127, 26, "ProofInfo",
   CellTags->{"Proof|ID:2020716880", "Proof|ID:2020716880-1"},
   CellID->1784894938]},
 "Proof|ID:1964835443"->{
  Cell[1462055, 41634, 5039, 127, 26, "ProofInfo",
   CellTags->{"Proof|ID:1964835443", "Proof|ID:1964835443-1"},
   CellID->880026044]},
 "Proof|ID:1964835443-1"->{
  Cell[1462055, 41634, 5039, 127, 26, "ProofInfo",
   CellTags->{"Proof|ID:1964835443", "Proof|ID:1964835443-1"},
   CellID->880026044]},
 "Proof|ID:1599847484"->{
  Cell[1467616, 41780, 5710, 139, 26, "ProofInfo",
   CellTags->{"Proof|ID:1599847484", "Proof|ID:1599847484-1"},
   CellID->780294033]},
 "Proof|ID:1599847484-1"->{
  Cell[1467616, 41780, 5710, 139, 26, "ProofInfo",
   CellTags->{"Proof|ID:1599847484", "Proof|ID:1599847484-1"},
   CellID->780294033]},
 "Proof|ID:590709100"->{
  Cell[1473848, 41938, 7964, 182, 26, "ProofInfo",
   CellTags->{"Proof|ID:590709100", "Proof|ID:590709100-1"},
   CellID->726011705]},
 "Proof|ID:590709100-1"->{
  Cell[1473848, 41938, 7964, 182, 26, "ProofInfo",
   CellTags->{"Proof|ID:590709100", "Proof|ID:590709100-1"},
   CellID->726011705]},
 "Proof|ID:586184654"->{
  Cell[1482333, 42139, 5963, 150, 26, "ProofInfo",
   CellTags->{"Proof|ID:586184654", "Proof|ID:586184654-1"},
   CellID->1800670266]},
 "Proof|ID:586184654-1"->{
  Cell[1482333, 42139, 5963, 150, 26, "ProofInfo",
   CellTags->{"Proof|ID:586184654", "Proof|ID:586184654-1"},
   CellID->1800670266]},
 "Proof|ID:919284948"->{
  Cell[1488817, 42308, 3584, 96, 26, "ProofInfo",
   CellTags->{"Proof|ID:919284948", "Proof|ID:919284948-1"},
   CellID->2038963285]},
 "Proof|ID:919284948-1"->{
  Cell[1488817, 42308, 3584, 96, 26, "ProofInfo",
   CellTags->{"Proof|ID:919284948", "Proof|ID:919284948-1"},
   CellID->2038963285]},
 "ID:895736483"->{
  Cell[1503602, 42766, 738, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:895736483", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->895736483]},
 "ID:1904443546"->{
  Cell[1504343, 42787, 741, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1904443546", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1904443546]},
 "ID:1587558771"->{
  Cell[1505087, 42808, 739, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1587558771", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1587558771]},
 "ID:530327018"->{
  Cell[1505829, 42829, 1422, 41, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:530327018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377"},
   CellID->530327018]},
 "377"->{
  Cell[1505829, 42829, 1422, 41, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:530327018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377"},
   CellID->530327018]},
 "ID:2044424156"->{
  Cell[1507254, 42872, 2251, 69, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:2044424156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377'"},
   CellID->2044424156]},
 "377'"->{
  Cell[1507254, 42872, 2251, 69, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:2044424156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377'"},
   CellID->2044424156]},
 "ID:14600367"->{
  Cell[1509508, 42943, 1353, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:14600367", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377''"},
   CellID->14600367]},
 "377''"->{
  Cell[1509508, 42943, 1353, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:14600367", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377''"},
   CellID->14600367]},
 "ID:1313257457"->{
  Cell[1510864, 42984, 1494, 44, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1313257457", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 7"},
   CellID->1313257457]},
 "\[CenterDot] 7"->{
  Cell[1510864, 42984, 1494, 44, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1313257457", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 7"},
   CellID->1313257457]},
 "ID:1776891622"->{
  Cell[1512361, 43030, 2317, 65, 87, "FormalTextInputFormula",
   CellTags->{
    "ID:1776891622", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 8"},
   CellID->1776891622]},
 "\[CenterDot] 8"->{
  Cell[1512361, 43030, 2317, 65, 87, "FormalTextInputFormula",
   CellTags->{
    "ID:1776891622", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 8"},
   CellID->1776891622]},
 "Proof|ID:530327018"->{
  Cell[1514835, 43106, 4300, 110, 28, "ProofInfo",
   CellTags->{"Proof|ID:530327018", "Proof|ID:530327018-1"},
   CellID->24650902]},
 "Proof|ID:530327018-1"->{
  Cell[1514835, 43106, 4300, 110, 28, "ProofInfo",
   CellTags->{"Proof|ID:530327018", "Proof|ID:530327018-1"},
   CellID->24650902]},
 "Proof|ID:2044424156"->{
  Cell[1519655, 43235, 6116, 148, 26, "ProofInfo",
   CellTags->{"Proof|ID:2044424156", "Proof|ID:2044424156-1"},
   CellID->35999503]},
 "Proof|ID:2044424156-1"->{
  Cell[1519655, 43235, 6116, 148, 26, "ProofInfo",
   CellTags->{"Proof|ID:2044424156", "Proof|ID:2044424156-1"},
   CellID->35999503]},
 "Proof|ID:14600367"->{
  Cell[1526292, 43402, 4419, 114, 26, "ProofInfo",
   CellTags->{"Proof|ID:14600367", "Proof|ID:14600367-1"},
   CellID->439633690]},
 "Proof|ID:14600367-1"->{
  Cell[1526292, 43402, 4419, 114, 26, "ProofInfo",
   CellTags->{"Proof|ID:14600367", "Proof|ID:14600367-1"},
   CellID->439633690]},
 "Proof|ID:1313257457"->{
  Cell[1531230, 43535, 4220, 111, 26, "ProofInfo",
   CellTags->{"Proof|ID:1313257457", "Proof|ID:1313257457-1"},
   CellID->1865577507]},
 "Proof|ID:1313257457-1"->{
  Cell[1531230, 43535, 4220, 111, 26, "ProofInfo",
   CellTags->{"Proof|ID:1313257457", "Proof|ID:1313257457-1"},
   CellID->1865577507]},
 "Proof|ID:1776891622"->{
  Cell[1535971, 43665, 6238, 153, 26, "ProofInfo",
   CellTags->{"Proof|ID:1776891622", "Proof|ID:1776891622-1"},
   CellID->89604998]},
 "Proof|ID:1776891622-1"->{
  Cell[1535971, 43665, 6238, 153, 26, "ProofInfo",
   CellTags->{"Proof|ID:1776891622", "Proof|ID:1776891622-1"},
   CellID->89604998]},
 "ID:1077742446"->{
  Cell[1543430, 43861, 1093, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1077742446", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1077742446]},
 "ID:1717822499"->{
  Cell[1544526, 43889, 1503, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1717822499", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "384"},
   CellID->1717822499]},
 "384"->{
  Cell[1544526, 43889, 1503, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1717822499", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "384"},
   CellID->1717822499]},
 "ID:1045770541"->{
  Cell[1546032, 43934, 913, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1045770541", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "358"},
   CellID->1045770541]},
 "358"->{
  Cell[1546032, 43934, 913, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1045770541", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "358"},
   CellID->1045770541]},
 "ID:312784637"->{
  Cell[1546948, 43961, 2336, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:312784637", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "393"},
   CellID->312784637]},
 "393"->{
  Cell[1546948, 43961, 2336, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:312784637", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "393"},
   CellID->312784637]},
 "ID:172690263"->{
  Cell[1549287, 44028, 859, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:172690263", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->172690263]},
 "ID:174940712"->{
  Cell[1550149, 44053, 1284, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:174940712", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial"},
   CellID->174940712]},
 "mult monomial"->{
  Cell[1550149, 44053, 1284, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:174940712", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial"},
   CellID->174940712]},
 "ID:1850961673"->{
  Cell[1551436, 44090, 1482, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1850961673", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial II"},
   CellID->1850961673]},
 "mult monomial II"->{
  Cell[1551436, 44090, 1482, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1850961673", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial II"},
   CellID->1850961673]},
 "ID:849856225"->{
  Cell[1552921, 44136, 2001, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:849856225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial III"},
   CellID->849856225]},
 "mult monomial III"->{
  Cell[1552921, 44136, 2001, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:849856225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial III"},
   CellID->849856225]},
 "Proof|ID:1717822499"->{
  Cell[1555080, 44206, 4761, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:1717822499", "Proof|ID:1717822499-1"},
   CellID->1434876425]},
 "Proof|ID:1717822499-1"->{
  Cell[1555080, 44206, 4761, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:1717822499", "Proof|ID:1717822499-1"},
   CellID->1434876425]},
 "Proof|ID:1045770541"->{
  Cell[1560363, 44346, 3608, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1045770541", "Proof|ID:1045770541-1"},
   CellID->1393398660]},
 "Proof|ID:1045770541-1"->{
  Cell[1560363, 44346, 3608, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1045770541", "Proof|ID:1045770541-1"},
   CellID->1393398660]},
 "Proof|ID:312784637"->{
  Cell[1564492, 44464, 6941, 165, 70, "ProofInfo",
   CellTags->{"Proof|ID:312784637", "Proof|ID:312784637-1"},
   CellID->2016825295]},
 "Proof|ID:312784637-1"->{
  Cell[1564492, 44464, 6941, 165, 70, "ProofInfo",
   CellTags->{"Proof|ID:312784637", "Proof|ID:312784637-1"},
   CellID->2016825295]},
 "Proof|ID:174940712"->{
  Cell[1571953, 44648, 4620, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:174940712", "Proof|ID:174940712-1"},
   CellID->1306123142]},
 "Proof|ID:174940712-1"->{
  Cell[1571953, 44648, 4620, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:174940712", "Proof|ID:174940712-1"},
   CellID->1306123142]},
 "Proof|ID:1850961673"->{
  Cell[1577093, 44785, 5473, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1850961673", "Proof|ID:1850961673-1"},
   CellID->1891103710]},
 "Proof|ID:1850961673-1"->{
  Cell[1577093, 44785, 5473, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1850961673", "Proof|ID:1850961673-1"},
   CellID->1891103710]},
 "Proof|ID:849856225"->{
  Cell[1583087, 44939, 6911, 162, 70, "ProofInfo",
   CellTags->{"Proof|ID:849856225", "Proof|ID:849856225-1"},
   CellID->232642418]},
 "Proof|ID:849856225-1"->{
  Cell[1583087, 44939, 6911, 162, 70, "ProofInfo",
   CellTags->{"Proof|ID:849856225", "Proof|ID:849856225-1"},
   CellID->232642418]},
 "ID:633254127"->{
  Cell[1591199, 45143, 1001, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:633254127", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->633254127]},
 "ID:1581566919"->{
  Cell[1592203, 45171, 1124, 29, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1581566919", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1581566919]},
 "ID:1230351665"->{
  Cell[1593330, 45202, 977, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1230351665", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 1"},
   CellID->1230351665]},
 "tuples \[CenterDot] 1"->{
  Cell[1593330, 45202, 977, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1230351665", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 1"},
   CellID->1230351665]},
 "ID:1294700753"->{
  Cell[1594310, 45229, 816, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1294700753", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1294700753]},
 "ID:168346017"->{
  Cell[1595129, 45253, 1349, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:168346017", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 2"},
   CellID->168346017]},
 "tuples \[CenterDot] 2"->{
  Cell[1595129, 45253, 1349, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:168346017", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 2"},
   CellID->168346017]},
 "ID:1208795229"->{
  Cell[1596481, 45293, 1960, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1208795229", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 3"},
   CellID->1208795229]},
 "tuples \[CenterDot] 3"->{
  Cell[1596481, 45293, 1960, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1208795229", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 3"},
   CellID->1208795229]},
 "ID:369110158"->{
  Cell[1598444, 45352, 1764, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:369110158", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 4"},
   CellID->369110158]},
 "tuples \[CenterDot] 4"->{
  Cell[1598444, 45352, 1764, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:369110158", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 4"},
   CellID->369110158]},
 "Proof|ID:1230351665"->{
  Cell[1600364, 45414, 4421, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1230351665", "Proof|ID:1230351665-1"},
   CellID->803576507]},
 "Proof|ID:1230351665-1"->{
  Cell[1600364, 45414, 4421, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1230351665", "Proof|ID:1230351665-1"},
   CellID->803576507]},
 "Proof|ID:168346017"->{
  Cell[1605306, 45548, 7111, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:168346017", "Proof|ID:168346017-1"},
   CellID->1197262074]},
 "Proof|ID:168346017-1"->{
  Cell[1605306, 45548, 7111, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:168346017", "Proof|ID:168346017-1"},
   CellID->1197262074]},
 "Proof|ID:1208795229"->{
  Cell[1612938, 45745, 8032, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1208795229", "Proof|ID:1208795229-1"},
   CellID->1531683337]},
 "Proof|ID:1208795229-1"->{
  Cell[1612938, 45745, 8032, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1208795229", "Proof|ID:1208795229-1"},
   CellID->1531683337]},
 "Proof|ID:369110158"->{
  Cell[1621492, 45961, 6980, 170, 70, "ProofInfo",
   CellTags->{"Proof|ID:369110158", "Proof|ID:369110158-1"},
   CellID->1626361315]},
 "Proof|ID:369110158-1"->{
  Cell[1621492, 45961, 6980, 170, 70, "ProofInfo",
   CellTags->{"Proof|ID:369110158", "Proof|ID:369110158-1"},
   CellID->1626361315]},
 "ID:155708286"->{
  Cell[1629797, 46177, 761, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:155708286", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->155708286]},
 "ID:280312466"->{
  Cell[1630561, 46199, 807, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:280312466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "4"},
   CellID->280312466]},
 "4"->{
  Cell[1630561, 46199, 807, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:280312466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "4"},
   CellID->280312466]},
 "ID:713919982"->{
  Cell[1631371, 46224, 1007, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:713919982", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "5"},
   CellID->713919982]},
 "5"->{
  Cell[1631371, 46224, 1007, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:713919982", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "5"},
   CellID->713919982]},
 "ID:1630947645"->{
  Cell[1632381, 46256, 1354, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1630947645", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "6"},
   CellID->1630947645]},
 "6"->{
  Cell[1632381, 46256, 1354, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1630947645", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "6"},
   CellID->1630947645]},
 "Proof|ID:280312466"->{
  Cell[1633893, 46309, 2758, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:280312466", "Proof|ID:280312466-1"},
   CellID->1561997]},
 "Proof|ID:280312466-1"->{
  Cell[1633893, 46309, 2758, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:280312466", "Proof|ID:280312466-1"},
   CellID->1561997]},
 "Proof|ID:713919982"->{
  Cell[1637171, 46409, 3567, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:713919982", "Proof|ID:713919982-1"},
   CellID->393026]},
 "Proof|ID:713919982-1"->{
  Cell[1637171, 46409, 3567, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:713919982", "Proof|ID:713919982-1"},
   CellID->393026]},
 "Proof|ID:1630947645"->{
  Cell[1641258, 46526, 4027, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:1630947645", "Proof|ID:1630947645-1"},
   CellID->350427357]},
 "Proof|ID:1630947645-1"->{
  Cell[1641258, 46526, 4027, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:1630947645", "Proof|ID:1630947645-1"},
   CellID->350427357]},
 "ID:795795393"->{
  Cell[1646356, 46670, 1075, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:795795393", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "360"},
   CellID->795795393]},
 "360"->{
  Cell[1646356, 46670, 1075, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:795795393", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "360"},
   CellID->795795393]},
 "Proof|ID:795795393"->{
  Cell[1647587, 46713, 1939, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:795795393", "Proof|ID:795795393-1"},
   CellID->1058256179]},
 "Proof|ID:795795393-1"->{
  Cell[1647587, 46713, 1939, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:795795393", "Proof|ID:795795393-1"},
   CellID->1058256179]},
 "ID:1302712453"->{
  Cell[1650783, 46817, 778, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1302712453", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1302712453]},
 "ID:1092343128"->{
  Cell[1651564, 46839, 5366, 155, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1092343128", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1"},
   CellID->1092343128]},
 "1"->{
  Cell[1651564, 46839, 5366, 155, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1092343128", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1"},
   CellID->1092343128]},
 "ID:311837531"->{
  Cell[1656933, 46996, 1107, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:311837531", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "2"},
   CellID->311837531]},
 "2"->{
  Cell[1656933, 46996, 1107, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:311837531", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "2"},
   CellID->311837531]},
 "ID:2036982330"->{
  Cell[1658043, 47032, 1868, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2036982330", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "3"},
   CellID->2036982330]},
 "3"->{
  Cell[1658043, 47032, 1868, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2036982330", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "3"},
   CellID->2036982330]},
 "ID:917853466"->{
  Cell[1659914, 47095, 2517, 75, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:917853466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "378"},
   CellID->917853466]},
 "378"->{
  Cell[1659914, 47095, 2517, 75, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:917853466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "378"},
   CellID->917853466]},
 "Proof|ID:1092343128"->{
  Cell[1662588, 47181, 11724, 267, 70, "ProofInfo",
   CellTags->{"Proof|ID:1092343128", "Proof|ID:1092343128-1"},
   CellID->402687292]},
 "Proof|ID:1092343128-1"->{
  Cell[1662588, 47181, 11724, 267, 70, "ProofInfo",
   CellTags->{"Proof|ID:1092343128", "Proof|ID:1092343128-1"},
   CellID->402687292]},
 "Proof|ID:311837531"->{
  Cell[1674833, 47467, 3041, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:311837531", "Proof|ID:311837531-1"},
   CellID->222069951]},
 "Proof|ID:311837531-1"->{
  Cell[1674833, 47467, 3041, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:311837531", "Proof|ID:311837531-1"},
   CellID->222069951]},
 "Proof|ID:2036982330"->{
  Cell[1678394, 47570, 5049, 129, 70, "ProofInfo",
   CellTags->{"Proof|ID:2036982330", "Proof|ID:2036982330-1"},
   CellID->78668347]},
 "Proof|ID:2036982330-1"->{
  Cell[1678394, 47570, 5049, 129, 70, "ProofInfo",
   CellTags->{"Proof|ID:2036982330", "Proof|ID:2036982330-1"},
   CellID->78668347]},
 "Proof|ID:917853466"->{
  Cell[1683964, 47718, 5399, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:917853466", "Proof|ID:917853466-1"},
   CellID->463010928]},
 "Proof|ID:917853466-1"->{
  Cell[1683964, 47718, 5399, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:917853466", "Proof|ID:917853466-1"},
   CellID->463010928]},
 "ID:1009043994"->{
  Cell[1690346, 47888, 1019, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1009043994", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "361"},
   CellID->1009043994]},
 "361"->{
  Cell[1690346, 47888, 1019, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1009043994", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "361"},
   CellID->1009043994]},
 "Proof|ID:1009043994"->{
  Cell[1691522, 47930, 1937, 61, 70, "ProofInfo",
   CellTags->{"Proof|ID:1009043994", "Proof|ID:1009043994-1"},
   CellID->307122811]},
 "Proof|ID:1009043994-1"->{
  Cell[1691522, 47930, 1937, 61, 70, "ProofInfo",
   CellTags->{"Proof|ID:1009043994", "Proof|ID:1009043994-1"},
   CellID->307122811]},
 "ID:412087985"->{
  Cell[1694675, 48034, 791, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:412087985", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "362"},
   CellID->412087985]},
 "362"->{
  Cell[1694675, 48034, 791, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:412087985", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "362"},
   CellID->412087985]},
 "Proof|ID:412087985"->{
  Cell[1695623, 48067, 1749, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:412087985", "Proof|ID:412087985-1"},
   CellID->453147487]},
 "Proof|ID:412087985-1"->{
  Cell[1695623, 48067, 1749, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:412087985", "Proof|ID:412087985-1"},
   CellID->453147487]},
 "ID:1267059190"->{
  Cell[1698554, 48164, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1267059190", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "363"},
   CellID->1267059190]},
 "363"->{
  Cell[1698554, 48164, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1267059190", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "363"},
   CellID->1267059190]},
 "Proof|ID:1267059190"->{
  Cell[1699494, 48196, 1734, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1267059190", "Proof|ID:1267059190-1"},
   CellID->1552638821]},
 "Proof|ID:1267059190-1"->{
  Cell[1699494, 48196, 1734, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1267059190", "Proof|ID:1267059190-1"},
   CellID->1552638821]},
 "ID:378652997"->{
  Cell[1702407, 48294, 780, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:378652997", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "364"},
   CellID->378652997]},
 "364"->{
  Cell[1702407, 48294, 780, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:378652997", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "364"},
   CellID->378652997]},
 "Proof|ID:378652997"->{
  Cell[1703344, 48326, 1725, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:378652997", "Proof|ID:378652997-1"},
   CellID->1328553736]},
 "Proof|ID:378652997-1"->{
  Cell[1703344, 48326, 1725, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:378652997", "Proof|ID:378652997-1"},
   CellID->1328553736]},
 "ID:1036416147"->{
  Cell[1706676, 48435, 780, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1036416147", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "365"},
   CellID->1036416147]},
 "365"->{
  Cell[1706676, 48435, 780, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1036416147", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "365"},
   CellID->1036416147]},
 "Proof|ID:1036416147"->{
  Cell[1707611, 48467, 1733, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1036416147", "Proof|ID:1036416147-1"},
   CellID->246228744]},
 "Proof|ID:1036416147-1"->{
  Cell[1707611, 48467, 1733, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1036416147", "Proof|ID:1036416147-1"},
   CellID->246228744]},
 "ID:1341307829"->{
  Cell[1710553, 48567, 2877, 78, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1341307829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "386"},
   CellID->1341307829]},
 "386"->{
  Cell[1710553, 48567, 2877, 78, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1341307829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "386"},
   CellID->1341307829]},
 "ID:1193326225"->{
  Cell[1713433, 48647, 3328, 89, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1193326225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "385"},
   CellID->1193326225]},
 "385"->{
  Cell[1713433, 48647, 3328, 89, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1193326225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "385"},
   CellID->1193326225]},
 "Proof|ID:1341307829"->{
  Cell[1716919, 48747, 6984, 175, 70, "ProofInfo",
   CellTags->{"Proof|ID:1341307829", "Proof|ID:1341307829-1"},
   CellID->467247998]},
 "Proof|ID:1341307829-1"->{
  Cell[1716919, 48747, 6984, 175, 70, "ProofInfo",
   CellTags->{"Proof|ID:1341307829", "Proof|ID:1341307829-1"},
   CellID->467247998]},
 "Proof|ID:1193326225"->{
  Cell[1724424, 48941, 7796, 191, 70, "ProofInfo",
   CellTags->{"Proof|ID:1193326225", "Proof|ID:1193326225-1"},
   CellID->348360809]},
 "Proof|ID:1193326225-1"->{
  Cell[1724424, 48941, 7796, 191, 70, "ProofInfo",
   CellTags->{"Proof|ID:1193326225", "Proof|ID:1193326225-1"},
   CellID->348360809]},
 "ID:2033377127"->{
  Cell[1733185, 49165, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2033377127", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "366"},
   CellID->2033377127]},
 "366"->{
  Cell[1733185, 49165, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2033377127", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "366"},
   CellID->2033377127]},
 "Proof|ID:2033377127"->{
  Cell[1734125, 49197, 1734, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:2033377127", "Proof|ID:2033377127-1"},
   CellID->1477321448]},
 "Proof|ID:2033377127-1"->{
  Cell[1734125, 49197, 1734, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:2033377127", "Proof|ID:2033377127-1"},
   CellID->1477321448]},
 "ID:982813600"->{
  Cell[1737106, 49296, 779, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:982813600", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "367"},
   CellID->982813600]},
 "367"->{
  Cell[1737106, 49296, 779, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:982813600", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "367"},
   CellID->982813600]},
 "Proof|ID:982813600"->{
  Cell[1738042, 49328, 1725, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:982813600", "Proof|ID:982813600-1"},
   CellID->1236106547]},
 "Proof|ID:982813600-1"->{
  Cell[1738042, 49328, 1725, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:982813600", "Proof|ID:982813600-1"},
   CellID->1236106547]},
 "ID:1179937112"->{
  Cell[1740681, 49417, 776, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1179937112", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1179937112]},
 "ID:134072435"->{
  Cell[1741460, 49439, 1646, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134072435", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "336"},
   CellID->134072435]},
 "336"->{
  Cell[1741460, 49439, 1646, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134072435", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "336"},
   CellID->134072435]},
 "ID:1484092370"->{
  Cell[1743109, 49489, 1121, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1484092370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "337"},
   CellID->1484092370]},
 "337"->{
  Cell[1743109, 49489, 1121, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1484092370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "337"},
   CellID->1484092370]},
 "Proof|ID:134072435"->{
  Cell[1744386, 49533, 5829, 144, 70, "ProofInfo",
   CellTags->{"Proof|ID:134072435", "Proof|ID:134072435-1"},
   CellID->77495342]},
 "Proof|ID:134072435-1"->{
  Cell[1744386, 49533, 5829, 144, 70, "ProofInfo",
   CellTags->{"Proof|ID:134072435", "Proof|ID:134072435-1"},
   CellID->77495342]},
 "Proof|ID:1484092370"->{
  Cell[1750736, 49696, 3608, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1484092370", "Proof|ID:1484092370-1"},
   CellID->1348971853]},
 "Proof|ID:1484092370-1"->{
  Cell[1750736, 49696, 3608, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1484092370", "Proof|ID:1484092370-1"},
   CellID->1348971853]},
 "ID:1480600758"->{
  Cell[1755622, 49840, 1026, 27, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1480600758", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1480600758]},
 "ID:1004003660"->{
  Cell[1756651, 49869, 739, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1004003660", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1004003660]},
 "ID:1854976476"->{
  Cell[1757393, 49890, 1075, 28, 56, "GlobalDeclaration",
   CellTags->{
    "ID:1854976476", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1854976476]},
 "ID:408134090"->{
  Cell[1758471, 49920, 1867, 50, 58, "FormalTextInputFormula",
   CellTags->{
    "ID:408134090", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduceM \[ScriptCapitalP]"},
   CellID->408134090]},
 "reduceM \[ScriptCapitalP]"->{
  Cell[1758471, 49920, 1867, 50, 58, "FormalTextInputFormula",
   CellTags->{
    "ID:408134090", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduceM \[ScriptCapitalP]"},
   CellID->408134090]},
 "ID:1088018783"->{
  Cell[1760341, 49972, 2604, 72, 89, "FormalTextInputFormula",
   CellTags->{
    "ID:1088018783", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduceM \[ScriptCapitalP] II"},
   CellID->1088018783]},
 "reduceM \[ScriptCapitalP] II"->{
  Cell[1760341, 49972, 2604, 72, 89, "FormalTextInputFormula",
   CellTags->{
    "ID:1088018783", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduceM \[ScriptCapitalP] II"},
   CellID->1088018783]},
 "Proof|ID:408134090"->{
  Cell[1763103, 50055, 8051, 192, 28, "ProofInfo",
   CellTags->{"Proof|ID:408134090", "Proof|ID:408134090-1"},
   CellID->59633542]},
 "Proof|ID:408134090-1"->{
  Cell[1763103, 50055, 8051, 192, 28, "ProofInfo",
   CellTags->{"Proof|ID:408134090", "Proof|ID:408134090-1"},
   CellID->59633542]},
 "Proof|ID:1088018783"->{
  Cell[1771674, 50266, 10512, 242, 26, "ProofInfo",
   CellTags->{"Proof|ID:1088018783", "Proof|ID:1088018783-1"},
   CellID->306221528]},
 "Proof|ID:1088018783-1"->{
  Cell[1771674, 50266, 10512, 242, 26, "ProofInfo",
   CellTags->{"Proof|ID:1088018783", "Proof|ID:1088018783-1"},
   CellID->306221528]},
 "ID:1964871967"->{
  Cell[1783259, 50545, 953, 25, 66, "GlobalDeclaration",
   CellTags->{
    "ID:1964871967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1964871967]},
 "ID:1041722017"->{
  Cell[1784724, 50585, 1865, 53, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1041722017", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "407"},
   CellID->1041722017]},
 "407"->{
  Cell[1784724, 50585, 1865, 53, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1041722017", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "407"},
   CellID->1041722017]},
 "ID:1518222527"->{
  Cell[1786592, 50640, 738, 19, 31, "GlobalDeclaration",
   CellTags->{
    "ID:1518222527", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1518222527]},
 "ID:1070088161"->{
  Cell[1787333, 50661, 2380, 68, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1070088161", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "408"},
   CellID->1070088161]},
 "408"->{
  Cell[1787333, 50661, 2380, 68, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1070088161", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "408"},
   CellID->1070088161]},
 "ID:1970014385"->{
  Cell[1789716, 50731, 2361, 68, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1970014385", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "408'"},
   CellID->1970014385]},
 "408'"->{
  Cell[1789716, 50731, 2361, 68, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1970014385", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "408'"},
   CellID->1970014385]},
 "Proof|ID:1041722017"->{
  Cell[1792232, 50810, 5663, 140, 28, "ProofInfo",
   CellTags->{"Proof|ID:1041722017", "Proof|ID:1041722017-1"},
   CellID->178605949]},
 "Proof|ID:1041722017-1"->{
  Cell[1792232, 50810, 5663, 140, 28, "ProofInfo",
   CellTags->{"Proof|ID:1041722017", "Proof|ID:1041722017-1"},
   CellID->178605949]},
 "Proof|ID:1070088161"->{
  Cell[1798415, 50969, 8372, 195, 26, "ProofInfo",
   CellTags->{"Proof|ID:1070088161", "Proof|ID:1070088161-1"},
   CellID->1155517064]},
 "Proof|ID:1070088161-1"->{
  Cell[1798415, 50969, 8372, 195, 26, "ProofInfo",
   CellTags->{"Proof|ID:1070088161", "Proof|ID:1070088161-1"},
   CellID->1155517064]},
 "Proof|ID:1970014385"->{
  Cell[1807308, 51183, 8354, 193, 26, "ProofInfo",
   CellTags->{"Proof|ID:1970014385", "Proof|ID:1970014385-1"},
   CellID->625973727]},
 "Proof|ID:1970014385-1"->{
  Cell[1807308, 51183, 8354, 193, 26, "ProofInfo",
   CellTags->{"Proof|ID:1970014385", "Proof|ID:1970014385-1"},
   CellID->625973727]},
 "ID:102721391"->{
  Cell[1816716, 51412, 786, 21, 42, "GlobalDeclaration",
   CellTags->{
    "ID:102721391", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->102721391]},
 "ID:657593834"->{
  Cell[1817505, 51435, 2568, 73, 120, "FormalTextInputFormula",
   CellTags->{
    "ID:657593834", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED"},
   CellID->657593834]},
 "RED"->{
  Cell[1817505, 51435, 2568, 73, 120, "FormalTextInputFormula",
   CellTags->{
    "ID:657593834", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED"},
   CellID->657593834]},
 "ID:1802485517"->{
  Cell[1820076, 51510, 2296, 62, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:1802485517", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED II"},
   CellID->1802485517]},
 "RED II"->{
  Cell[1820076, 51510, 2296, 62, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:1802485517", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED II"},
   CellID->1802485517]},
 "ID:1164884865"->{
  Cell[1822375, 51574, 2242, 60, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:1164884865", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED III"},
   CellID->1164884865]},
 "RED III"->{
  Cell[1822375, 51574, 2242, 60, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:1164884865", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED III"},
   CellID->1164884865]},
 "ID:958945878"->{
  Cell[1824620, 51636, 2619, 72, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:958945878", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED IV"},
   CellID->958945878]},
 "RED IV"->{
  Cell[1824620, 51636, 2619, 72, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:958945878", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED IV"},
   CellID->958945878]},
 "ID:1601029605"->{
  Cell[1827242, 51710, 886, 24, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1601029605", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1601029605]},
 "ID:687002895"->{
  Cell[1828131, 51736, 2247, 62, 120, "FormalTextInputFormula",
   CellTags->{
    "ID:687002895", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED monomial"},
   CellID->687002895]},
 "RED monomial"->{
  Cell[1828131, 51736, 2247, 62, 120, "FormalTextInputFormula",
   CellTags->{
    "ID:687002895", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED monomial"},
   CellID->687002895]},
 "ID:413658239"->{
  Cell[1830381, 51800, 1920, 55, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:413658239", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED monomial II"},
   CellID->413658239]},
 "RED monomial II"->{
  Cell[1830381, 51800, 1920, 55, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:413658239", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED monomial II"},
   CellID->413658239]},
 "Proof|ID:657593834"->{
  Cell[1832459, 51866, 8060, 188, 28, "ProofInfo",
   CellTags->{"Proof|ID:657593834", "Proof|ID:657593834-1"},
   CellID->17884315]},
 "Proof|ID:657593834-1"->{
  Cell[1832459, 51866, 8060, 188, 28, "ProofInfo",
   CellTags->{"Proof|ID:657593834", "Proof|ID:657593834-1"},
   CellID->17884315]},
 "Proof|ID:1802485517"->{
  Cell[1841039, 52073, 6585, 157, 26, "ProofInfo",
   CellTags->{"Proof|ID:1802485517", "Proof|ID:1802485517-1"},
   CellID->17136843]},
 "Proof|ID:1802485517-1"->{
  Cell[1841039, 52073, 6585, 157, 26, "ProofInfo",
   CellTags->{"Proof|ID:1802485517", "Proof|ID:1802485517-1"},
   CellID->17136843]},
 "Proof|ID:1164884865"->{
  Cell[1848145, 52249, 6417, 153, 26, "ProofInfo",
   CellTags->{"Proof|ID:1164884865", "Proof|ID:1164884865-1"},
   CellID->1415542183]},
 "Proof|ID:1164884865-1"->{
  Cell[1848145, 52249, 6417, 153, 26, "ProofInfo",
   CellTags->{"Proof|ID:1164884865", "Proof|ID:1164884865-1"},
   CellID->1415542183]},
 "Proof|ID:958945878"->{
  Cell[1855083, 52421, 8038, 184, 26, "ProofInfo",
   CellTags->{"Proof|ID:958945878", "Proof|ID:958945878-1"},
   CellID->1172023358]},
 "Proof|ID:958945878-1"->{
  Cell[1855083, 52421, 8038, 184, 26, "ProofInfo",
   CellTags->{"Proof|ID:958945878", "Proof|ID:958945878-1"},
   CellID->1172023358]},
 "Proof|ID:687002895"->{
  Cell[1863642, 52624, 6947, 162, 26, "ProofInfo",
   CellTags->{"Proof|ID:687002895", "Proof|ID:687002895-1"},
   CellID->563304177]},
 "Proof|ID:687002895-1"->{
  Cell[1863642, 52624, 6947, 162, 26, "ProofInfo",
   CellTags->{"Proof|ID:687002895", "Proof|ID:687002895-1"},
   CellID->563304177]},
 "Proof|ID:413658239"->{
  Cell[1871109, 52805, 6283, 150, 26, "ProofInfo",
   CellTags->{"Proof|ID:413658239", "Proof|ID:413658239-1"},
   CellID->885739272]},
 "Proof|ID:413658239-1"->{
  Cell[1871109, 52805, 6283, 150, 26, "ProofInfo",
   CellTags->{"Proof|ID:413658239", "Proof|ID:413658239-1"},
   CellID->885739272]},
 "ID:580194792"->{
  Cell[1878303, 52988, 771, 20, 42, "GlobalDeclaration",
   CellTags->{
    "ID:580194792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->580194792]},
 "ID:1667457569"->{
  Cell[1879077, 53010, 2513, 75, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1667457569", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "357"},
   CellID->1667457569]},
 "357"->{
  Cell[1879077, 53010, 2513, 75, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1667457569", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "357"},
   CellID->1667457569]},
 "ID:2035027156"->{
  Cell[1881593, 53087, 1878, 61, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:2035027156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] lm \[DoubleRightArrow]"},
   CellID->2035027156]},
 "\[RightArrow] lm \[DoubleRightArrow]"->{
  Cell[1881593, 53087, 1878, 61, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:2035027156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] lm \[DoubleRightArrow]"},
   CellID->2035027156]},
 "ID:2026457432"->{
  Cell[1883474, 53150, 1822, 56, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:2026457432", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] lm \[DoubleLeftArrow]"},
   CellID->2026457432]},
 "\[RightArrow] lm \[DoubleLeftArrow]"->{
  Cell[1883474, 53150, 1822, 56, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:2026457432", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] lm \[DoubleLeftArrow]"},
   CellID->2026457432]},
 "ID:1167741853"->{
  Cell[1885299, 53208, 2433, 70, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:1167741853", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reducibility \[ScriptCapitalP]"},
   CellID->1167741853]},
 "reducibility \[ScriptCapitalP]"->{
  Cell[1885299, 53208, 2433, 70, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:1167741853", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reducibility \[ScriptCapitalP]"},
   CellID->1167741853]},
 "ID:1998060487"->{
  Cell[1887735, 53280, 3955, 105, 110, "FormalTextInputFormula",
   CellTags->{
    "ID:1998060487", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] alternative \[ScriptCapitalP]"},
   CellID->1998060487]},
 "\[RightArrow] alternative \[ScriptCapitalP]"->{
  Cell[1887735, 53280, 3955, 105, 110, "FormalTextInputFormula",
   CellTags->{
    "ID:1998060487", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] alternative \[ScriptCapitalP]"},
   CellID->1998060487]},
 "Proof|ID:1667457569"->{
  Cell[1891847, 53396, 7280, 174, 28, "ProofInfo",
   CellTags->{"Proof|ID:1667457569", "Proof|ID:1667457569-1"},
   CellID->60009841]},
 "Proof|ID:1667457569-1"->{
  Cell[1891847, 53396, 7280, 174, 28, "ProofInfo",
   CellTags->{"Proof|ID:1667457569", "Proof|ID:1667457569-1"},
   CellID->60009841]},
 "Proof|ID:2035027156"->{
  Cell[1899648, 53589, 5119, 133, 26, "ProofInfo",
   CellTags->{"Proof|ID:2035027156", "Proof|ID:2035027156-1"},
   CellID->322996175]},
 "Proof|ID:2035027156-1"->{
  Cell[1899648, 53589, 5119, 133, 26, "ProofInfo",
   CellTags->{"Proof|ID:2035027156", "Proof|ID:2035027156-1"},
   CellID->322996175]},
 "Proof|ID:2026457432"->{
  Cell[1905288, 53741, 5145, 133, 26, "ProofInfo",
   CellTags->{"Proof|ID:2026457432", "Proof|ID:2026457432-1"},
   CellID->1357322543]},
 "Proof|ID:2026457432-1"->{
  Cell[1905288, 53741, 5145, 133, 26, "ProofInfo",
   CellTags->{"Proof|ID:2026457432", "Proof|ID:2026457432-1"},
   CellID->1357322543]},
 "Proof|ID:1167741853"->{
  Cell[1910999, 53894, 5950, 146, 26, "ProofInfo",
   CellTags->{"Proof|ID:1167741853", "Proof|ID:1167741853-1"},
   CellID->293140950]},
 "Proof|ID:1167741853-1"->{
  Cell[1910999, 53894, 5950, 146, 26, "ProofInfo",
   CellTags->{"Proof|ID:1167741853", "Proof|ID:1167741853-1"},
   CellID->293140950]},
 "Proof|ID:1998060487"->{
  Cell[1917471, 54059, 9124, 211, 26, "ProofInfo",
   CellTags->{"Proof|ID:1998060487", "Proof|ID:1998060487-1"},
   CellID->2076771442]},
 "Proof|ID:1998060487-1"->{
  Cell[1917471, 54059, 9124, 211, 26, "ProofInfo",
   CellTags->{"Proof|ID:1998060487", "Proof|ID:1998060487-1"},
   CellID->2076771442]},
 "ID:1709390289"->{
  Cell[1927649, 54306, 927, 24, 66, "GlobalDeclaration",
   CellTags->{
    "ID:1709390289", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1709390289]},
 "ID:1713414293"->{
  Cell[1928579, 54332, 4827, 129, 102, "FormalTextInputFormula",
   CellTags->{
    "ID:1713414293", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[LeftRightArrow] alternative \[ScriptCapitalP]"},
   CellID->1713414293]},
 "\[LeftRightArrow] alternative \[ScriptCapitalP]"->{
  Cell[1928579, 54332, 4827, 129, 102, "FormalTextInputFormula",
   CellTags->{
    "ID:1713414293", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[LeftRightArrow] alternative \[ScriptCapitalP]"},
   CellID->1713414293]},
 "Proof|ID:1713414293"->{
  Cell[1933564, 54472, 13367, 293, 28, "ProofInfo",
   CellTags->{"Proof|ID:1713414293", "Proof|ID:1713414293-1"},
   CellID->793804010]},
 "Proof|ID:1713414293-1"->{
  Cell[1933564, 54472, 13367, 293, 28, "ProofInfo",
   CellTags->{"Proof|ID:1713414293", "Proof|ID:1713414293-1"},
   CellID->793804010]},
 "ID:1190720464"->{
  Cell[1957261, 55102, 3791, 102, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1190720464", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducibleRec \[ScriptCapitalP]"},
   CellID->1190720464]},
 "isReducibleRec \[ScriptCapitalP]"->{
  Cell[1957261, 55102, 3791, 102, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1190720464", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducibleRec \[ScriptCapitalP]"},
   CellID->1190720464]},
 "ID:166529906"->{
  Cell[1961713, 55232, 924, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:166529906", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->166529906]},
 "ID:844782115"->{
  Cell[1962640, 55257, 1304, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:844782115", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352.\[DoubleRightArrow]"},
   CellID->844782115]},
 "352.\[DoubleRightArrow]"->{
  Cell[1962640, 55257, 1304, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:844782115", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352.\[DoubleRightArrow]"},
   CellID->844782115]},
 "ID:953126445"->{
  Cell[1963947, 55289, 1106, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:953126445", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352.\[DoubleLeftArrow]"},
   CellID->953126445]},
 "352.\[DoubleLeftArrow]"->{
  Cell[1963947, 55289, 1106, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:953126445", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352.\[DoubleLeftArrow]"},
   CellID->953126445]},
 "ID:259818613"->{
  Cell[1965056, 55319, 1233, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:259818613", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352"},
   CellID->259818613]},
 "352"->{
  Cell[1965056, 55319, 1233, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:259818613", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352"},
   CellID->259818613]},
 "ID:182112345"->{
  Cell[1966292, 55351, 2618, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:182112345", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "353"},
   CellID->182112345]},
 "353"->{
  Cell[1966292, 55351, 2618, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:182112345", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "353"},
   CellID->182112345]},
 "Proof|ID:844782115"->{
  Cell[1969066, 55435, 3984, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:844782115", "Proof|ID:844782115-1"},
   CellID->1659422444]},
 "Proof|ID:844782115-1"->{
  Cell[1969066, 55435, 3984, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:844782115", "Proof|ID:844782115-1"},
   CellID->1659422444]},
 "Proof|ID:953126445"->{
  Cell[1973570, 55556, 3982, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:953126445", "Proof|ID:953126445-1"},
   CellID->1715888170]},
 "Proof|ID:953126445-1"->{
  Cell[1973570, 55556, 3982, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:953126445", "Proof|ID:953126445-1"},
   CellID->1715888170]},
 "Proof|ID:259818613"->{
  Cell[1978073, 55677, 3993, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:259818613", "Proof|ID:259818613-1"},
   CellID->816722537]},
 "Proof|ID:259818613-1"->{
  Cell[1978073, 55677, 3993, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:259818613", "Proof|ID:259818613-1"},
   CellID->816722537]},
 "Proof|ID:182112345"->{
  Cell[1982586, 55800, 6479, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:182112345", "Proof|ID:182112345-1"},
   CellID->908791481]},
 "Proof|ID:182112345-1"->{
  Cell[1982586, 55800, 6479, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:182112345", "Proof|ID:182112345-1"},
   CellID->908791481]},
 "ID:418362312"->{
  Cell[1990319, 55999, 1060, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:418362312", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->418362312]},
 "ID:2116744492"->{
  Cell[1991382, 56027, 739, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2116744492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->2116744492]},
 "ID:1746803457"->{
  Cell[1992124, 56048, 2781, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1746803457", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "391"},
   CellID->1746803457]},
 "391"->{
  Cell[1992124, 56048, 2781, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1746803457", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "391"},
   CellID->1746803457]},
 "ID:899106457"->{
  Cell[1994908, 56129, 826, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:899106457", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->899106457]},
 "ID:1157329004"->{
  Cell[1995737, 56152, 1099, 28, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1157329004", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1157329004]},
 "ID:1927781160"->{
  Cell[1996839, 56182, 2316, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1927781160", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "387"},
   CellID->1927781160]},
 "387"->{
  Cell[1996839, 56182, 2316, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1927781160", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "387"},
   CellID->1927781160]},
 "Proof|ID:1746803457"->{
  Cell[1999311, 56255, 7744, 183, 70, "ProofInfo",
   CellTags->{"Proof|ID:1746803457", "Proof|ID:1746803457-1"},
   CellID->311879395]},
 "Proof|ID:1746803457-1"->{
  Cell[1999311, 56255, 7744, 183, 70, "ProofInfo",
   CellTags->{"Proof|ID:1746803457", "Proof|ID:1746803457-1"},
   CellID->311879395]},
 "Proof|ID:1927781160"->{
  Cell[2007576, 56457, 11812, 271, 70, "ProofInfo",
   CellTags->{"Proof|ID:1927781160", "Proof|ID:1927781160-1"},
   CellID->235488096]},
 "Proof|ID:1927781160-1"->{
  Cell[2007576, 56457, 11812, 271, 70, "ProofInfo",
   CellTags->{"Proof|ID:1927781160", "Proof|ID:1927781160-1"},
   CellID->235488096]},
 "ID:934500559"->{
  Cell[2020358, 56761, 781, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:934500559", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "368"},
   CellID->934500559]},
 "368"->{
  Cell[2020358, 56761, 781, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:934500559", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "368"},
   CellID->934500559]},
 "Proof|ID:934500559"->{
  Cell[2021296, 56793, 1723, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:934500559", "Proof|ID:934500559-1"},
   CellID->80140931]},
 "Proof|ID:934500559-1"->{
  Cell[2021296, 56793, 1723, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:934500559", "Proof|ID:934500559-1"},
   CellID->80140931]},
 "ID:371274476"->{
  Cell[2024330, 56894, 1088, 28, 70, "GlobalDeclaration",
   CellTags->{
    "ID:371274476", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->371274476]},
 "ID:1588964752"->{
  Cell[2025421, 56924, 2225, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1588964752", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR"},
   CellID->1588964752]},
 "IRR"->{
  Cell[2025421, 56924, 2225, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1588964752", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR"},
   CellID->1588964752]},
 "ID:428389793"->{
  Cell[2027649, 56988, 2461, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:428389793", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR II"},
   CellID->428389793]},
 "IRR II"->{
  Cell[2027649, 56988, 2461, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:428389793", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR II"},
   CellID->428389793]},
 "ID:225111715"->{
  Cell[2030113, 57060, 1800, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:225111715", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CORR"},
   CellID->225111715]},
 "CORR"->{
  Cell[2030113, 57060, 1800, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:225111715", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CORR"},
   CellID->225111715]},
 "ID:1632881403"->{
  Cell[2031916, 57115, 915, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1632881403", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1632881403]},
 "ID:1701548086"->{
  Cell[2032834, 57142, 2042, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1701548086", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR \[CenterDot]"},
   CellID->1701548086]},
 "IRR \[CenterDot]"->{
  Cell[2032834, 57142, 2042, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1701548086", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR \[CenterDot]"},
   CellID->1701548086]},
 "ID:1758251641"->{
  Cell[2034879, 57200, 1266, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1758251641", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CORR \[CenterDot]"},
   CellID->1758251641]},
 "CORR \[CenterDot]"->{
  Cell[2034879, 57200, 1266, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1758251641", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CORR \[CenterDot]"},
   CellID->1758251641]},
 "Proof|ID:1588964752"->{
  Cell[2036301, 57245, 8654, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1588964752", "Proof|ID:1588964752-1"},
   CellID->543579358]},
 "Proof|ID:1588964752-1"->{
  Cell[2036301, 57245, 8654, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1588964752", "Proof|ID:1588964752-1"},
   CellID->543579358]},
 "Proof|ID:428389793"->{
  Cell[2045476, 57461, 9787, 218, 70, "ProofInfo",
   CellTags->{"Proof|ID:428389793", "Proof|ID:428389793-1"},
   CellID->490546380]},
 "Proof|ID:428389793-1"->{
  Cell[2045476, 57461, 9787, 218, 70, "ProofInfo",
   CellTags->{"Proof|ID:428389793", "Proof|ID:428389793-1"},
   CellID->490546380]},
 "Proof|ID:225111715"->{
  Cell[2055784, 57698, 6091, 146, 70, "ProofInfo",
   CellTags->{"Proof|ID:225111715", "Proof|ID:225111715-1"},
   CellID->657130363]},
 "Proof|ID:225111715-1"->{
  Cell[2055784, 57698, 6091, 146, 70, "ProofInfo",
   CellTags->{"Proof|ID:225111715", "Proof|ID:225111715-1"},
   CellID->657130363]},
 "Proof|ID:1701548086"->{
  Cell[2062395, 57863, 9366, 211, 70, "ProofInfo",
   CellTags->{"Proof|ID:1701548086", "Proof|ID:1701548086-1"},
   CellID->384143515]},
 "Proof|ID:1701548086-1"->{
  Cell[2062395, 57863, 9366, 211, 70, "ProofInfo",
   CellTags->{"Proof|ID:1701548086", "Proof|ID:1701548086-1"},
   CellID->384143515]},
 "Proof|ID:1758251641"->{
  Cell[2072283, 58093, 5943, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:1758251641", "Proof|ID:1758251641-1"},
   CellID->1782924340]},
 "Proof|ID:1758251641-1"->{
  Cell[2072283, 58093, 5943, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:1758251641", "Proof|ID:1758251641-1"},
   CellID->1782924340]},
 "ID:1928324287"->{
  Cell[2079926, 58284, 939, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1928324287", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1928324287]},
 "ID:751250157"->{
  Cell[2080868, 58310, 1839, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:751250157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404"},
   CellID->751250157]},
 "404"->{
  Cell[2080868, 58310, 1839, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:751250157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404"},
   CellID->751250157]},
 "ID:1952511245"->{
  Cell[2082710, 58365, 1866, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1952511245", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404 II"},
   CellID->1952511245]},
 "404 II"->{
  Cell[2082710, 58365, 1866, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1952511245", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404 II"},
   CellID->1952511245]},
 "ID:537853020"->{
  Cell[2084579, 58421, 2301, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:537853020", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404 III"},
   CellID->537853020]},
 "404 III"->{
  Cell[2084579, 58421, 2301, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:537853020", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404 III"},
   CellID->537853020]},
 "Proof|ID:751250157"->{
  Cell[2087037, 58496, 6772, 164, 70, "ProofInfo",
   CellTags->{"Proof|ID:751250157", "Proof|ID:751250157-1"},
   CellID->1336000539]},
 "Proof|ID:751250157-1"->{
  Cell[2087037, 58496, 6772, 164, 70, "ProofInfo",
   CellTags->{"Proof|ID:751250157", "Proof|ID:751250157-1"},
   CellID->1336000539]},
 "Proof|ID:1952511245"->{
  Cell[2094374, 58680, 6393, 153, 70, "ProofInfo",
   CellTags->{"Proof|ID:1952511245", "Proof|ID:1952511245-1"},
   CellID->29876965]},
 "Proof|ID:1952511245-1"->{
  Cell[2094374, 58680, 6393, 153, 70, "ProofInfo",
   CellTags->{"Proof|ID:1952511245", "Proof|ID:1952511245-1"},
   CellID->29876965]},
 "Proof|ID:537853020"->{
  Cell[2101287, 58852, 7716, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:537853020", "Proof|ID:537853020-1"},
   CellID->1965128634]},
 "Proof|ID:537853020-1"->{
  Cell[2101287, 58852, 7716, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:537853020", "Proof|ID:537853020-1"},
   CellID->1965128634]},
 "ID:659046058"->{
  Cell[2110000, 59065, 963, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:659046058", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->659046058]},
 "ID:319520004"->{
  Cell[2110966, 59092, 736, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:319520004", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->319520004]},
 "ID:967033090"->{
  Cell[2111705, 59113, 826, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:967033090", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->967033090]},
 "ID:1647700871"->{
  Cell[2112534, 59136, 1098, 28, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1647700871", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1647700871]},
 "ID:1981597108"->{
  Cell[2113635, 59166, 786, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1981597108", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1981597108]},
 "ID:40011060"->{
  Cell[2114818, 59198, 2437, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40011060", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "388"},
   CellID->40011060]},
 "388"->{
  Cell[2114818, 59198, 2437, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40011060", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "388"},
   CellID->40011060]},
 "ID:884639747"->{
  Cell[2117258, 59267, 792, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:884639747", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->884639747]},
 "ID:1360655817"->{
  Cell[2118053, 59290, 1524, 45, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1360655817", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1360655817]},
 "ID:1002690147"->{
  Cell[2119580, 59337, 1516, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1002690147", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "389"},
   CellID->1002690147]},
 "389"->{
  Cell[2119580, 59337, 1516, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1002690147", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "389"},
   CellID->1002690147]},
 "ID:1896907436"->{
  Cell[2121099, 59383, 1485, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1896907436", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "390"},
   CellID->1896907436]},
 "390"->{
  Cell[2121099, 59383, 1485, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1896907436", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "390"},
   CellID->1896907436]},
 "Proof|ID:40011060"->{
  Cell[2122742, 59438, 16067, 356, 70, "ProofInfo",
   CellTags->{"Proof|ID:40011060", "Proof|ID:40011060-1"},
   CellID->1454431248]},
 "Proof|ID:40011060-1"->{
  Cell[2122742, 59438, 16067, 356, 70, "ProofInfo",
   CellTags->{"Proof|ID:40011060", "Proof|ID:40011060-1"},
   CellID->1454431248]},
 "Proof|ID:1002690147"->{
  Cell[2139329, 59813, 8812, 205, 70, "ProofInfo",
   CellTags->{"Proof|ID:1002690147", "Proof|ID:1002690147-1"},
   CellID->615225779]},
 "Proof|ID:1002690147-1"->{
  Cell[2139329, 59813, 8812, 205, 70, "ProofInfo",
   CellTags->{"Proof|ID:1002690147", "Proof|ID:1002690147-1"},
   CellID->615225779]},
 "Proof|ID:1896907436"->{
  Cell[2148662, 60037, 8924, 207, 70, "ProofInfo",
   CellTags->{"Proof|ID:1896907436", "Proof|ID:1896907436-1"},
   CellID->375199501]},
 "Proof|ID:1896907436-1"->{
  Cell[2148662, 60037, 8924, 207, 70, "ProofInfo",
   CellTags->{"Proof|ID:1896907436", "Proof|ID:1896907436-1"},
   CellID->375199501]},
 "ID:1290694758"->{
  Cell[2158559, 60277, 781, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1290694758", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "369"},
   CellID->1290694758]},
 "369"->{
  Cell[2158559, 60277, 781, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1290694758", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "369"},
   CellID->1290694758]},
 "Proof|ID:1290694758"->{
  Cell[2159496, 60309, 1733, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1290694758", "Proof|ID:1290694758-1"},
   CellID->796556125]},
 "Proof|ID:1290694758-1"->{
  Cell[2159496, 60309, 1733, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1290694758", "Proof|ID:1290694758-1"},
   CellID->796556125]},
 "ID:1785082178"->{
  Cell[2162552, 60410, 1046, 27, 71, "GlobalDeclaration",
   CellTags->{
    "ID:1785082178", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1785082178]},
 "ID:1718264802"->{
  Cell[2163601, 60439, 840, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1718264802", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1718264802]},
 "ID:752304273"->{
  Cell[2164444, 60463, 803, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:752304273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->752304273]},
 "ID:578561747"->{
  Cell[2165250, 60486, 2765, 79, 145, "FormalTextInputFormula",
   CellTags->{
    "ID:578561747", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "areCommonReducibles I"},
   CellID->578561747]},
 "areCommonReducibles I"->{
  Cell[2165250, 60486, 2765, 79, 145, "FormalTextInputFormula",
   CellTags->{
    "ID:578561747", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "areCommonReducibles I"},
   CellID->578561747]},
 "ID:1399766637"->{
  Cell[2168018, 60567, 1563, 45, 85, "FormalTextInputFormula",
   CellTags->{
    "ID:1399766637", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "areCommonReducibles II"},
   CellID->1399766637]},
 "areCommonReducibles II"->{
  Cell[2168018, 60567, 1563, 45, 85, "FormalTextInputFormula",
   CellTags->{
    "ID:1399766637", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "areCommonReducibles II"},
   CellID->1399766637]},
 "ID:1132709394"->{
  Cell[2169584, 60614, 3366, 94, 145, "FormalTextInputFormula",
   CellTags->{
    "ID:1132709394", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP"},
   CellID->1132709394]},
 "isNTCRP"->{
  Cell[2169584, 60614, 3366, 94, 145, "FormalTextInputFormula",
   CellTags->{
    "ID:1132709394", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP"},
   CellID->1132709394]},
 "ID:176358731"->{
  Cell[2173619, 60735, 1018, 27, 71, "GlobalDeclaration",
   CellTags->{
    "ID:176358731", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->176358731]},
 "ID:391507392"->{
  Cell[2174640, 60764, 885, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:391507392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->391507392]},
 "ID:1030826183"->{
  Cell[2175528, 60788, 804, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:1030826183", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1030826183]},
 "ID:926282876"->{
  Cell[2176335, 60811, 1752, 52, 78, "FormalTextInputFormula",
   CellTags->{
    "ID:926282876", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "354"},
   CellID->926282876]},
 "354"->{
  Cell[2176335, 60811, 1752, 52, 78, "FormalTextInputFormula",
   CellTags->{
    "ID:926282876", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "354"},
   CellID->926282876]},
 "ID:80931797"->{
  Cell[2178090, 60865, 1749, 51, 91, "FormalTextInputFormula",
   CellTags->{
    "ID:80931797", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "355"},
   CellID->80931797]},
 "355"->{
  Cell[2178090, 60865, 1749, 51, 91, "FormalTextInputFormula",
   CellTags->{
    "ID:80931797", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "355"},
   CellID->80931797]},
 "ID:414980491"->{
  Cell[2179842, 60918, 1448, 43, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:414980491", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "359"},
   CellID->414980491]},
 "359"->{
  Cell[2179842, 60918, 1448, 43, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:414980491", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "359"},
   CellID->414980491]},
 "ID:744723554"->{
  Cell[2181293, 60963, 908, 25, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:744723554", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "345"},
   CellID->744723554]},
 "345"->{
  Cell[2181293, 60963, 908, 25, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:744723554", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "345"},
   CellID->744723554]},
 "Proof|ID:926282876"->{
  Cell[2182359, 60999, 8291, 196, 28, "ProofInfo",
   CellTags->{"Proof|ID:926282876", "Proof|ID:926282876-1"},
   CellID->667882351]},
 "Proof|ID:926282876-1"->{
  Cell[2182359, 60999, 8291, 196, 28, "ProofInfo",
   CellTags->{"Proof|ID:926282876", "Proof|ID:926282876-1"},
   CellID->667882351]},
 "Proof|ID:80931797"->{
  Cell[2191170, 61214, 8599, 196, 26, "ProofInfo",
   CellTags->{"Proof|ID:80931797", "Proof|ID:80931797-1"},
   CellID->1748139084]},
 "Proof|ID:80931797-1"->{
  Cell[2191170, 61214, 8599, 196, 26, "ProofInfo",
   CellTags->{"Proof|ID:80931797", "Proof|ID:80931797-1"},
   CellID->1748139084]},
 "Proof|ID:414980491"->{
  Cell[2200289, 61429, 7515, 178, 26, "ProofInfo",
   CellTags->{"Proof|ID:414980491", "Proof|ID:414980491-1"},
   CellID->336718788]},
 "Proof|ID:414980491-1"->{
  Cell[2200289, 61429, 7515, 178, 26, "ProofInfo",
   CellTags->{"Proof|ID:414980491", "Proof|ID:414980491-1"},
   CellID->336718788]},
 "Proof|ID:744723554"->{
  Cell[2208324, 61626, 5579, 140, 26, "ProofInfo",
   CellTags->{"Proof|ID:744723554", "Proof|ID:744723554-1"},
   CellID->1023806301]},
 "Proof|ID:744723554-1"->{
  Cell[2208324, 61626, 5579, 140, 26, "ProofInfo",
   CellTags->{"Proof|ID:744723554", "Proof|ID:744723554-1"},
   CellID->1023806301]},
 "ID:381508689"->{
  Cell[2214938, 61800, 1018, 27, 71, "GlobalDeclaration",
   CellTags->{
    "ID:381508689", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->381508689]},
 "ID:1050210793"->{
  Cell[2215959, 61829, 741, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1050210793", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1050210793]},
 "ID:512389161"->{
  Cell[2216703, 61850, 803, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:512389161", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->512389161]},
 "ID:719916043"->{
  Cell[2217509, 61873, 1702, 51, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:719916043", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative 1"},
   CellID->719916043]},
 "isNTCRP alternative 1"->{
  Cell[2217509, 61873, 1702, 51, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:719916043", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative 1"},
   CellID->719916043]},
 "ID:864384559"->{
  Cell[2219214, 61926, 1596, 49, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:864384559", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative 2"},
   CellID->864384559]},
 "isNTCRP alternative 2"->{
  Cell[2219214, 61926, 1596, 49, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:864384559", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative 2"},
   CellID->864384559]},
 "ID:951944988"->{
  Cell[2220813, 61977, 2056, 57, 67, "FormalTextInputFormula",
   CellTags->{
    "ID:951944988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative"},
   CellID->951944988]},
 "isNTCRP alternative"->{
  Cell[2220813, 61977, 2056, 57, 67, "FormalTextInputFormula",
   CellTags->{
    "ID:951944988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative"},
   CellID->951944988]},
 "Proof|ID:719916043"->{
  Cell[2223027, 62045, 8098, 190, 28, "ProofInfo",
   CellTags->{"Proof|ID:719916043", "Proof|ID:719916043-1"},
   CellID->1378946613]},
 "Proof|ID:719916043-1"->{
  Cell[2223027, 62045, 8098, 190, 28, "ProofInfo",
   CellTags->{"Proof|ID:719916043", "Proof|ID:719916043-1"},
   CellID->1378946613]},
 "Proof|ID:864384559"->{
  Cell[2231646, 62254, 7386, 174, 26, "ProofInfo",
   CellTags->{"Proof|ID:864384559", "Proof|ID:864384559-1"},
   CellID->1555291612]},
 "Proof|ID:864384559-1"->{
  Cell[2231646, 62254, 7386, 174, 26, "ProofInfo",
   CellTags->{"Proof|ID:864384559", "Proof|ID:864384559-1"},
   CellID->1555291612]},
 "Proof|ID:951944988"->{
  Cell[2239552, 62447, 9206, 209, 26, "ProofInfo",
   CellTags->{"Proof|ID:951944988", "Proof|ID:951944988-1"},
   CellID->1069274625]},
 "Proof|ID:951944988-1"->{
  Cell[2239552, 62447, 9206, 209, 26, "ProofInfo",
   CellTags->{"Proof|ID:951944988", "Proof|ID:951944988-1"},
   CellID->1069274625]},
 "ID:1990838743"->{
  Cell[2290519, 63920, 1175, 29, 71, "GlobalDeclaration",
   CellTags->{
    "ID:1990838743", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1990838743]},
 "ID:20910152"->{
  Cell[2291697, 63951, 896, 23, 43, "GlobalDeclaration",
   CellTags->{
    "ID:20910152", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->20910152]},
 "ID:416591219"->{
  Cell[2292596, 63976, 3492, 91, 134, "FormalTextInputFormula",
   CellTags->{
    "ID:416591219", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleRightArrow].1"},
   CellID->416591219]},
 "CR1.\[DoubleRightArrow].1"->{
  Cell[2292596, 63976, 3492, 91, 134, "FormalTextInputFormula",
   CellTags->{
    "ID:416591219", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleRightArrow].1"},
   CellID->416591219]},
 "ID:295118406"->{
  Cell[2296091, 64069, 2290, 63, 94, "FormalTextInputFormula",
   CellTags->{
    "ID:295118406", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleRightArrow].2"},
   CellID->295118406]},
 "CR1.\[DoubleRightArrow].2"->{
  Cell[2296091, 64069, 2290, 63, 94, "FormalTextInputFormula",
   CellTags->{
    "ID:295118406", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleRightArrow].2"},
   CellID->295118406]},
 "ID:1583101849"->{
  Cell[2298384, 64134, 4394, 121, 204, "FormalTextInputFormula",
   CellTags->{
    "ID:1583101849", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleLeftArrow]"},
   CellID->1583101849]},
 "CR1.\[DoubleLeftArrow]"->{
  Cell[2298384, 64134, 4394, 121, 204, "FormalTextInputFormula",
   CellTags->{
    "ID:1583101849", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleLeftArrow]"},
   CellID->1583101849]},
 "Proof|ID:416591219"->{
  Cell[2302935, 64266, 11266, 250, 28, "ProofInfo",
   CellTags->{"Proof|ID:416591219", "Proof|ID:416591219-1"},
   CellID->1006102583]},
 "Proof|ID:416591219-1"->{
  Cell[2302935, 64266, 11266, 250, 28, "ProofInfo",
   CellTags->{"Proof|ID:416591219", "Proof|ID:416591219-1"},
   CellID->1006102583]},
 "Proof|ID:295118406"->{
  Cell[2314722, 64535, 8718, 198, 26, "ProofInfo",
   CellTags->{"Proof|ID:295118406", "Proof|ID:295118406-1"},
   CellID->1475020847]},
 "Proof|ID:295118406-1"->{
  Cell[2314722, 64535, 8718, 198, 26, "ProofInfo",
   CellTags->{"Proof|ID:295118406", "Proof|ID:295118406-1"},
   CellID->1475020847]},
 "Proof|ID:1583101849"->{
  Cell[2323961, 64752, 13737, 297, 26, "ProofInfo",
   CellTags->{"Proof|ID:1583101849", "Proof|ID:1583101849-1"},
   CellID->2049393323]},
 "Proof|ID:1583101849-1"->{
  Cell[2323961, 64752, 13737, 297, 26, "ProofInfo",
   CellTags->{"Proof|ID:1583101849", "Proof|ID:1583101849-1"},
   CellID->2049393323]},
 "ID:1272506214"->{
  Cell[2403947, 67120, 1086, 26, 66, "GlobalDeclaration",
   CellTags->{
    "ID:1272506214", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1272506214]},
 "ID:386901098"->{
  Cell[2405036, 67148, 2896, 80, 127, "FormalTextInputFormula",
   CellTags->{
    "ID:386901098", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleRightArrow].1"},
   CellID->386901098]},
 "CR2.\[DoubleRightArrow].1"->{
  Cell[2405036, 67148, 2896, 80, 127, "FormalTextInputFormula",
   CellTags->{
    "ID:386901098", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleRightArrow].1"},
   CellID->386901098]},
 "ID:670300270"->{
  Cell[2407935, 67230, 2132, 61, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:670300270", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleRightArrow].2"},
   CellID->670300270]},
 "CR2.\[DoubleRightArrow].2"->{
  Cell[2407935, 67230, 2132, 61, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:670300270", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleRightArrow].2"},
   CellID->670300270]},
 "ID:1434882124"->{
  Cell[2410070, 67293, 4419, 125, 172, "FormalTextInputFormula",
   CellTags->{
    "ID:1434882124", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleLeftArrow]"},
   CellID->1434882124]},
 "CR2.\[DoubleLeftArrow]"->{
  Cell[2410070, 67293, 4419, 125, 172, "FormalTextInputFormula",
   CellTags->{
    "ID:1434882124", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleLeftArrow]"},
   CellID->1434882124]},
 "Proof|ID:386901098"->{
  Cell[2414647, 67429, 8691, 198, 28, "ProofInfo",
   CellTags->{"Proof|ID:386901098", "Proof|ID:386901098-1"},
   CellID->12812135]},
 "Proof|ID:386901098-1"->{
  Cell[2414647, 67429, 8691, 198, 28, "ProofInfo",
   CellTags->{"Proof|ID:386901098", "Proof|ID:386901098-1"},
   CellID->12812135]},
 "Proof|ID:670300270"->{
  Cell[2423858, 67646, 5665, 138, 26, "ProofInfo",
   CellTags->{"Proof|ID:670300270", "Proof|ID:670300270-1"},
   CellID->488830292]},
 "Proof|ID:670300270-1"->{
  Cell[2423858, 67646, 5665, 138, 26, "ProofInfo",
   CellTags->{"Proof|ID:670300270", "Proof|ID:670300270-1"},
   CellID->488830292]},
 "Proof|ID:1434882124"->{
  Cell[2430043, 67803, 11138, 247, 26, "ProofInfo",
   CellTags->{"Proof|ID:1434882124", "Proof|ID:1434882124-1"},
   CellID->822190239]},
 "Proof|ID:1434882124-1"->{
  Cell[2430043, 67803, 11138, 247, 26, "ProofInfo",
   CellTags->{"Proof|ID:1434882124", "Proof|ID:1434882124-1"},
   CellID->822190239]},
 "ID:737627449"->{
  Cell[2495075, 69709, 1008, 27, 71, "GlobalDeclaration",
   CellTags->{
    "ID:737627449", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->737627449]},
 "ID:81313572"->{
  Cell[2496086, 69738, 737, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:81313572", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->81313572]},
 "ID:1098402572"->{
  Cell[2496826, 69759, 767, 20, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1098402572", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1098402572]},
 "ID:1499935354"->{
  Cell[2497596, 69781, 2466, 71, 109, "FormalTextInputFormula",
   CellTags->{
    "ID:1499935354", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 1"},
   CellID->1499935354]},
 "CR1 monomial 1"->{
  Cell[2497596, 69781, 2466, 71, 109, "FormalTextInputFormula",
   CellTags->{
    "ID:1499935354", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 1"},
   CellID->1499935354]},
 "ID:1938671644"->{
  Cell[2500065, 69854, 2698, 73, 94, "FormalTextInputFormula",
   CellTags->{
    "ID:1938671644", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 2"},
   CellID->1938671644]},
 "CR1 monomial 2"->{
  Cell[2500065, 69854, 2698, 73, 94, "FormalTextInputFormula",
   CellTags->{
    "ID:1938671644", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 2"},
   CellID->1938671644]},
 "ID:48476485"->{
  Cell[2502766, 69929, 2305, 64, 89, "FormalTextInputFormula",
   CellTags->{
    "ID:48476485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 3"},
   CellID->48476485]},
 "CR1 monomial 3"->{
  Cell[2502766, 69929, 2305, 64, 89, "FormalTextInputFormula",
   CellTags->{
    "ID:48476485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 3"},
   CellID->48476485]},
 "ID:1246936946"->{
  Cell[2505074, 69995, 2004, 59, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1246936946", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 1"},
   CellID->1246936946]},
 "CR2 monomial 1"->{
  Cell[2505074, 69995, 2004, 59, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1246936946", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 1"},
   CellID->1246936946]},
 "ID:1856728487"->{
  Cell[2507081, 70056, 2186, 62, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:1856728487", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 2"},
   CellID->1856728487]},
 "CR2 monomial 2"->{
  Cell[2507081, 70056, 2186, 62, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:1856728487", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 2"},
   CellID->1856728487]},
 "ID:335657318"->{
  Cell[2509270, 70120, 1720, 50, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:335657318", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 3"},
   CellID->335657318]},
 "CR2 monomial 3"->{
  Cell[2509270, 70120, 1720, 50, 64, "FormalTextInputFormula",
   CellTags->{
    "ID:335657318", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 3"},
   CellID->335657318]},
 "Proof|ID:1499935354"->{
  Cell[2511147, 70181, 8905, 205, 28, "ProofInfo",
   CellTags->{"Proof|ID:1499935354", "Proof|ID:1499935354-1"},
   CellID->1869006459]},
 "Proof|ID:1499935354-1"->{
  Cell[2511147, 70181, 8905, 205, 28, "ProofInfo",
   CellTags->{"Proof|ID:1499935354", "Proof|ID:1499935354-1"},
   CellID->1869006459]},
 "Proof|ID:1938671644"->{
  Cell[2520574, 70405, 9347, 209, 26, "ProofInfo",
   CellTags->{"Proof|ID:1938671644", "Proof|ID:1938671644-1"},
   CellID->1781620971]},
 "Proof|ID:1938671644-1"->{
  Cell[2520574, 70405, 9347, 209, 26, "ProofInfo",
   CellTags->{"Proof|ID:1938671644", "Proof|ID:1938671644-1"},
   CellID->1781620971]},
 "Proof|ID:48476485"->{
  Cell[2530511, 70634, 8644, 196, 26, "ProofInfo",
   CellTags->{"Proof|ID:48476485", "Proof|ID:48476485-1"},
   CellID->528453611]},
 "Proof|ID:48476485-1"->{
  Cell[2530511, 70634, 8644, 196, 26, "ProofInfo",
   CellTags->{"Proof|ID:48476485", "Proof|ID:48476485-1"},
   CellID->528453611]},
 "Proof|ID:1246936946"->{
  Cell[2539675, 70849, 6288, 153, 26, "ProofInfo",
   CellTags->{"Proof|ID:1246936946", "Proof|ID:1246936946-1"},
   CellID->299300275]},
 "Proof|ID:1246936946-1"->{
  Cell[2539675, 70849, 6288, 153, 26, "ProofInfo",
   CellTags->{"Proof|ID:1246936946", "Proof|ID:1246936946-1"},
   CellID->299300275]},
 "Proof|ID:1856728487"->{
  Cell[2546484, 71021, 5697, 139, 26, "ProofInfo",
   CellTags->{"Proof|ID:1856728487", "Proof|ID:1856728487-1"},
   CellID->1800537912]},
 "Proof|ID:1856728487-1"->{
  Cell[2546484, 71021, 5697, 139, 26, "ProofInfo",
   CellTags->{"Proof|ID:1856728487", "Proof|ID:1856728487-1"},
   CellID->1800537912]},
 "Proof|ID:335657318"->{
  Cell[2552702, 71179, 5515, 134, 26, "ProofInfo",
   CellTags->{"Proof|ID:335657318", "Proof|ID:335657318-1"},
   CellID->218427358]},
 "Proof|ID:335657318-1"->{
  Cell[2552702, 71179, 5515, 134, 26, "ProofInfo",
   CellTags->{"Proof|ID:335657318", "Proof|ID:335657318-1"},
   CellID->218427358]},
 "ID:624517207"->{
  Cell[2585532, 72148, 2769, 80, 137, "FormalTextInputFormula",
   CellTags->{
    "ID:624517207", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMNTCRP"},
   CellID->624517207]},
 "isMNTCRP"->{
  Cell[2585532, 72148, 2769, 80, 137, "FormalTextInputFormula",
   CellTags->{
    "ID:624517207", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMNTCRP"},
   CellID->624517207]},
 "ID:1172537415"->{
  Cell[2589233, 72262, 3080, 86, 159, "FormalTextInputFormula",
   CellTags->{
    "ID:1172537415", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMNTCRP alternative"},
   CellID->1172537415]},
 "isMNTCRP alternative"->{
  Cell[2589233, 72262, 3080, 86, 159, "FormalTextInputFormula",
   CellTags->{
    "ID:1172537415", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMNTCRP alternative"},
   CellID->1172537415]},
 "Proof|ID:1172537415"->{
  Cell[2592469, 72359, 9223, 210, 28, "ProofInfo",
   CellTags->{"Proof|ID:1172537415", "Proof|ID:1172537415-1"},
   CellID->1389429208]},
 "Proof|ID:1172537415-1"->{
  Cell[2592469, 72359, 9223, 210, 28, "ProofInfo",
   CellTags->{"Proof|ID:1172537415", "Proof|ID:1172537415-1"},
   CellID->1389429208]},
 "ID:1861685863"->{
  Cell[2613705, 72951, 1084, 28, 71, "GlobalDeclaration",
   CellTags->{
    "ID:1861685863", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1861685863]},
 "ID:1154885365"->{
  Cell[2614792, 72981, 806, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:1154885365", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1154885365]},
 "ID:1777331368"->{
  Cell[2615601, 73004, 1000, 26, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1777331368", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].1"},
   CellID->1777331368]},
 "CR3.\[DoubleRightArrow].1"->{
  Cell[2615601, 73004, 1000, 26, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1777331368", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].1"},
   CellID->1777331368]},
 "ID:1052103130"->{
  Cell[2616604, 73032, 1375, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1052103130", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].2"},
   CellID->1052103130]},
 "CR3.\[DoubleRightArrow].2"->{
  Cell[2616604, 73032, 1375, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1052103130", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].2"},
   CellID->1052103130]},
 "ID:1077684208"->{
  Cell[2617982, 73073, 1127, 29, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1077684208", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].3"},
   CellID->1077684208]},
 "CR3.\[DoubleRightArrow].3"->{
  Cell[2617982, 73073, 1127, 29, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1077684208", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].3"},
   CellID->1077684208]},
 "ID:86200506"->{
  Cell[2619112, 73104, 1660, 47, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:86200506", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleLeftArrow]"},
   CellID->86200506]},
 "CR3.\[DoubleLeftArrow]"->{
  Cell[2619112, 73104, 1660, 47, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:86200506", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleLeftArrow]"},
   CellID->86200506]},
 "ID:394014296"->{
  Cell[2620775, 73153, 1938, 57, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:394014296", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3"},
   CellID->394014296]},
 "CR3"->{
  Cell[2620775, 73153, 1938, 57, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:394014296", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3"},
   CellID->394014296]},
 "Proof|ID:1777331368"->{
  Cell[2622870, 73221, 5896, 143, 28, "ProofInfo",
   CellTags->{"Proof|ID:1777331368", "Proof|ID:1777331368-1"},
   CellID->508661436]},
 "Proof|ID:1777331368-1"->{
  Cell[2622870, 73221, 5896, 143, 28, "ProofInfo",
   CellTags->{"Proof|ID:1777331368", "Proof|ID:1777331368-1"},
   CellID->508661436]},
 "Proof|ID:1052103130"->{
  Cell[2629287, 73383, 6847, 163, 26, "ProofInfo",
   CellTags->{"Proof|ID:1052103130", "Proof|ID:1052103130-1"},
   CellID->1682055148]},
 "Proof|ID:1052103130-1"->{
  Cell[2629287, 73383, 6847, 163, 26, "ProofInfo",
   CellTags->{"Proof|ID:1052103130", "Proof|ID:1052103130-1"},
   CellID->1682055148]},
 "Proof|ID:1077684208"->{
  Cell[2636655, 73565, 6607, 157, 26, "ProofInfo",
   CellTags->{"Proof|ID:1077684208", "Proof|ID:1077684208-1"},
   CellID->1363636622]},
 "Proof|ID:1077684208-1"->{
  Cell[2636655, 73565, 6607, 157, 26, "ProofInfo",
   CellTags->{"Proof|ID:1077684208", "Proof|ID:1077684208-1"},
   CellID->1363636622]},
 "Proof|ID:86200506"->{
  Cell[2643784, 73741, 7545, 176, 26, "ProofInfo",
   CellTags->{"Proof|ID:86200506", "Proof|ID:86200506-1"},
   CellID->1645380276]},
 "Proof|ID:86200506-1"->{
  Cell[2643784, 73741, 7545, 176, 26, "ProofInfo",
   CellTags->{"Proof|ID:86200506", "Proof|ID:86200506-1"},
   CellID->1645380276]},
 "Proof|ID:394014296"->{
  Cell[2651849, 73936, 9027, 208, 26, "ProofInfo",
   CellTags->{"Proof|ID:394014296", "Proof|ID:394014296-1"},
   CellID->1823310355]},
 "Proof|ID:394014296-1"->{
  Cell[2651849, 73936, 9027, 208, 26, "ProofInfo",
   CellTags->{"Proof|ID:394014296", "Proof|ID:394014296-1"},
   CellID->1823310355]},
 "ID:1444400550"->{
  Cell[2705282, 75493, 994, 25, 66, "GlobalDeclaration",
   CellTags->{
    "ID:1444400550", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1444400550]},
 "ID:653031981"->{
  Cell[2706279, 75520, 965, 26, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:653031981", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].1"},
   CellID->653031981]},
 "CR4.\[DoubleRightArrow].1"->{
  Cell[2706279, 75520, 965, 26, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:653031981", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].1"},
   CellID->653031981]},
 "ID:943141844"->{
  Cell[2707247, 75548, 1157, 33, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:943141844", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].2"},
   CellID->943141844]},
 "CR4.\[DoubleRightArrow].2"->{
  Cell[2707247, 75548, 1157, 33, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:943141844", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].2"},
   CellID->943141844]},
 "ID:1143014070"->{
  Cell[2708407, 75583, 1190, 32, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1143014070", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].3"},
   CellID->1143014070]},
 "CR4.\[DoubleRightArrow].3"->{
  Cell[2708407, 75583, 1190, 32, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1143014070", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].3"},
   CellID->1143014070]},
 "ID:1630861163"->{
  Cell[2709600, 75617, 1536, 45, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1630861163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleLeftArrow]"},
   CellID->1630861163]},
 "CR4.\[DoubleLeftArrow]"->{
  Cell[2709600, 75617, 1536, 45, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1630861163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleLeftArrow]"},
   CellID->1630861163]},
 "ID:1701661908"->{
  Cell[2711139, 75664, 1755, 53, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:1701661908", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4"},
   CellID->1701661908]},
 "CR4"->{
  Cell[2711139, 75664, 1755, 53, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:1701661908", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4"},
   CellID->1701661908]},
 "Proof|ID:653031981"->{
  Cell[2713051, 75728, 3642, 96, 28, "ProofInfo",
   CellTags->{"Proof|ID:653031981", "Proof|ID:653031981-1"},
   CellID->13059041]},
 "Proof|ID:653031981-1"->{
  Cell[2713051, 75728, 3642, 96, 28, "ProofInfo",
   CellTags->{"Proof|ID:653031981", "Proof|ID:653031981-1"},
   CellID->13059041]},
 "Proof|ID:943141844"->{
  Cell[2717213, 75843, 4004, 104, 26, "ProofInfo",
   CellTags->{"Proof|ID:943141844", "Proof|ID:943141844-1"},
   CellID->1657492493]},
 "Proof|ID:943141844-1"->{
  Cell[2717213, 75843, 4004, 104, 26, "ProofInfo",
   CellTags->{"Proof|ID:943141844", "Proof|ID:943141844-1"},
   CellID->1657492493]},
 "Proof|ID:1143014070"->{
  Cell[2721738, 75966, 4211, 109, 26, "ProofInfo",
   CellTags->{"Proof|ID:1143014070", "Proof|ID:1143014070-1"},
   CellID->1936376328]},
 "Proof|ID:1143014070-1"->{
  Cell[2721738, 75966, 4211, 109, 26, "ProofInfo",
   CellTags->{"Proof|ID:1143014070", "Proof|ID:1143014070-1"},
   CellID->1936376328]},
 "Proof|ID:1630861163"->{
  Cell[2726468, 76094, 4528, 115, 26, "ProofInfo",
   CellTags->{"Proof|ID:1630861163", "Proof|ID:1630861163-1"},
   CellID->979966294]},
 "Proof|ID:1630861163-1"->{
  Cell[2726468, 76094, 4528, 115, 26, "ProofInfo",
   CellTags->{"Proof|ID:1630861163", "Proof|ID:1630861163-1"},
   CellID->979966294]},
 "Proof|ID:1701661908"->{
  Cell[2731518, 76228, 5955, 147, 26, "ProofInfo",
   CellTags->{"Proof|ID:1701661908", "Proof|ID:1701661908-1"},
   CellID->599419806]},
 "Proof|ID:1701661908-1"->{
  Cell[2731518, 76228, 5955, 147, 26, "ProofInfo",
   CellTags->{"Proof|ID:1701661908", "Proof|ID:1701661908-1"},
   CellID->599419806]},
 "ID:1489378443"->{
  Cell[2778126, 77591, 1118, 28, 71, "GlobalDeclaration",
   CellTags->{
    "ID:1489378443", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1489378443]},
 "ID:196663580"->{
  Cell[2779247, 77621, 800, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:196663580", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->196663580]},
 "ID:1830388955"->{
  Cell[2780050, 77644, 775, 20, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1830388955", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1830388955]},
 "ID:312371876"->{
  Cell[2780828, 77666, 3394, 93, 155, "FormalTextInputFormula",
   CellTags->{
    "ID:312371876", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CP1.\[DoubleLeftArrow]"},
   CellID->312371876]},
 "CP1.\[DoubleLeftArrow]"->{
  Cell[2780828, 77666, 3394, 93, 155, "FormalTextInputFormula",
   CellTags->{
    "ID:312371876", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CP1.\[DoubleLeftArrow]"},
   CellID->312371876]},
 "ID:1006612102"->{
  Cell[2784225, 77761, 4109, 110, 241, "FormalTextInputFormula",
   CellTags->{
    "ID:1006612102", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CP2.\[DoubleLeftArrow]"},
   CellID->1006612102]},
 "CP2.\[DoubleLeftArrow]"->{
  Cell[2784225, 77761, 4109, 110, 241, "FormalTextInputFormula",
   CellTags->{
    "ID:1006612102", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CP2.\[DoubleLeftArrow]"},
   CellID->1006612102]},
 "Proof|ID:312371876"->{
  Cell[2788490, 77882, 13507, 292, 28, "ProofInfo",
   CellTags->{"Proof|ID:312371876", "Proof|ID:312371876-1"},
   CellID->650652383]},
 "Proof|ID:312371876-1"->{
  Cell[2788490, 77882, 13507, 292, 28, "ProofInfo",
   CellTags->{"Proof|ID:312371876", "Proof|ID:312371876-1"},
   CellID->650652383]},
 "Proof|ID:1006612102"->{
  Cell[2802518, 78193, 14246, 309, 26, "ProofInfo",
   CellTags->{"Proof|ID:1006612102", "Proof|ID:1006612102-1"},
   CellID->719415418]},
 "Proof|ID:1006612102-1"->{
  Cell[2802518, 78193, 14246, 309, 26, "ProofInfo",
   CellTags->{"Proof|ID:1006612102", "Proof|ID:1006612102-1"},
   CellID->719415418]},
 "ID:542206690"->{
  Cell[2818600, 78561, 1020, 27, 71, "GlobalDeclaration",
   CellTags->{
    "ID:542206690", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->542206690]},
 "ID:1548368737"->{
  Cell[2819623, 78590, 741, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1548368737", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1548368737]},
 "ID:73478134"->{
  Cell[2820367, 78611, 798, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:73478134", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->73478134]},
 "ID:310376772"->{
  Cell[2821168, 78634, 9746, 250, 423, "FormalTextInputFormula",
   CellTags->{
    "ID:310376772", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "409"},
   CellID->310376772]},
 "409"->{
  Cell[2821168, 78634, 9746, 250, 423, "FormalTextInputFormula",
   CellTags->{
    "ID:310376772", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "409"},
   CellID->310376772]},
 "Proof|ID:310376772"->{
  Cell[2831071, 78895, 28365, 586, 28, "ProofInfo",
   CellTags->{"Proof|ID:310376772", "Proof|ID:310376772-1"},
   CellID->1336876000]},
 "Proof|ID:310376772-1"->{
  Cell[2831071, 78895, 28365, 586, 28, "ProofInfo",
   CellTags->{"Proof|ID:310376772", "Proof|ID:310376772-1"},
   CellID->1336876000]},
 "ID:383946867"->{
  Cell[2873845, 79926, 840, 21, 42, "GlobalDeclaration",
   CellTags->{
    "ID:383946867", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->383946867]},
 "ID:1587705005"->{
  Cell[2874688, 79949, 818, 21, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1587705005", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1587705005]},
 "ID:549840930"->{
  Cell[2875899, 79986, 3055, 95, 108, "FormalTextInputFormula",
   CellTags->{
    "ID:549840930", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].3"},
   CellID->549840930]},
 "R7.\[ScriptCapitalP].3"->{
  Cell[2875899, 79986, 3055, 95, 108, "FormalTextInputFormula",
   CellTags->{
    "ID:549840930", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].3"},
   CellID->549840930]},
 "ID:657656821"->{
  Cell[2878957, 80083, 4804, 151, 109, "FormalTextInputFormula",
   CellTags->{
    "ID:657656821", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].5"},
   CellID->657656821]},
 "R7.\[ScriptCapitalP].5"->{
  Cell[2878957, 80083, 4804, 151, 109, "FormalTextInputFormula",
   CellTags->{
    "ID:657656821", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].5"},
   CellID->657656821]},
 "ID:64232512"->{
  Cell[2883764, 80236, 934, 24, 66, "GlobalDeclaration",
   CellTags->{
    "ID:64232512", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->64232512]},
 "ID:237038572"->{
  Cell[2884701, 80262, 818, 21, 41, "GlobalDeclaration",
   CellTags->{
    "ID:237038572", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->237038572]},
 "ID:235501232"->{
  Cell[2885522, 80285, 718, 19, 31, "GlobalDeclaration",
   CellTags->{
    "ID:235501232", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->235501232]},
 "ID:1513648693"->{
  Cell[2886243, 80306, 1796, 49, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1513648693", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].1"},
   CellID->1513648693]},
 "R7.\[ScriptCapitalP].1"->{
  Cell[2886243, 80306, 1796, 49, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1513648693", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].1"},
   CellID->1513648693]},
 "ID:808269926"->{
  Cell[2888042, 80357, 3126, 90, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:808269926", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].4"},
   CellID->808269926]},
 "R7.\[ScriptCapitalP].4"->{
  Cell[2888042, 80357, 3126, 90, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:808269926", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].4"},
   CellID->808269926]},
 "Proof|ID:549840930"->{
  Cell[2891324, 80458, 8316, 194, 28, "ProofInfo",
   CellTags->{"Proof|ID:549840930", "Proof|ID:549840930-1"},
   CellID->1394810391]},
 "Proof|ID:549840930-1"->{
  Cell[2891324, 80458, 8316, 194, 28, "ProofInfo",
   CellTags->{"Proof|ID:549840930", "Proof|ID:549840930-1"},
   CellID->1394810391]},
 "Proof|ID:657656821"->{
  Cell[2900160, 80671, 12906, 288, 26, "ProofInfo",
   CellTags->{"Proof|ID:657656821", "Proof|ID:657656821-1"},
   CellID->75572869]},
 "Proof|ID:657656821-1"->{
  Cell[2900160, 80671, 12906, 288, 26, "ProofInfo",
   CellTags->{"Proof|ID:657656821", "Proof|ID:657656821-1"},
   CellID->75572869]},
 "Proof|ID:1513648693"->{
  Cell[2913585, 80978, 6847, 163, 26, "ProofInfo",
   CellTags->{"Proof|ID:1513648693", "Proof|ID:1513648693-1"},
   CellID->1424817208]},
 "Proof|ID:1513648693-1"->{
  Cell[2913585, 80978, 6847, 163, 26, "ProofInfo",
   CellTags->{"Proof|ID:1513648693", "Proof|ID:1513648693-1"},
   CellID->1424817208]},
 "Proof|ID:808269926"->{
  Cell[2920954, 81160, 11047, 246, 26, "ProofInfo",
   CellTags->{"Proof|ID:808269926", "Proof|ID:808269926-1"},
   CellID->876551560]},
 "Proof|ID:808269926-1"->{
  Cell[2920954, 81160, 11047, 246, 26, "ProofInfo",
   CellTags->{"Proof|ID:808269926", "Proof|ID:808269926-1"},
   CellID->876551560]},
 "ID:255426543"->{
  Cell[2985161, 83029, 779, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:255426543", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "370"},
   CellID->255426543]},
 "370"->{
  Cell[2985161, 83029, 779, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:255426543", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "370"},
   CellID->255426543]},
 "Proof|ID:255426543"->{
  Cell[2986096, 83061, 1724, 56, 28, "ProofInfo",
   CellTags->{"Proof|ID:255426543", "Proof|ID:255426543-1"},
   CellID->537752064]},
 "Proof|ID:255426543-1"->{
  Cell[2986096, 83061, 1724, 56, 28, "ProofInfo",
   CellTags->{"Proof|ID:255426543", "Proof|ID:255426543-1"},
   CellID->537752064]},
 "ID:1342888001"->{
  Cell[3015861, 83980, 782, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1342888001", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "371"},
   CellID->1342888001]},
 "371"->{
  Cell[3015861, 83980, 782, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1342888001", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "371"},
   CellID->1342888001]},
 "Proof|ID:1342888001"->{
  Cell[3016800, 84012, 1734, 57, 28, "ProofInfo",
   CellTags->{"Proof|ID:1342888001", "Proof|ID:1342888001-1"},
   CellID->1391148747]},
 "Proof|ID:1342888001-1"->{
  Cell[3016800, 84012, 1734, 57, 28, "ProofInfo",
   CellTags->{"Proof|ID:1342888001", "Proof|ID:1342888001-1"},
   CellID->1391148747]},
 "ID:1021018734"->{
  Cell[3032786, 84504, 965, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1021018734", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "376"},
   CellID->1021018734]},
 "376"->{
  Cell[3032786, 84504, 965, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1021018734", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "376"},
   CellID->1021018734]},
 "Proof|ID:1021018734"->{
  Cell[3033909, 84539, 1767, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1021018734", "Proof|ID:1021018734-1"},
   CellID->1129550695]},
 "Proof|ID:1021018734-1"->{
  Cell[3033909, 84539, 1767, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1021018734", "Proof|ID:1021018734-1"},
   CellID->1129550695]},
 "ID:459032599"->{
  Cell[3037212, 84643, 1659, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:459032599", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReductionRing Conservation Theorem"},
   CellID->459032599]},
 "isReductionRing Conservation Theorem"->{
  Cell[3037212, 84643, 1659, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:459032599", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReductionRing Conservation Theorem"},
   CellID->459032599]},
 "ID:719489133"->{
  Cell[3040300, 84740, 864, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:719489133", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "algorithmic reduction ring \[ScriptCapitalR]"},
   CellID->719489133]},
 "algorithmic reduction ring \[ScriptCapitalR]"->{
  Cell[3040300, 84740, 864, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:719489133", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "algorithmic reduction ring \[ScriptCapitalR]"},
   CellID->719489133]},
 "ID:1159608082"->{
  Cell[3041167, 84764, 919, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1159608082", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", 
     "algorithmic reduction-polynomial domain \[ScriptCapitalP]"},
   CellID->1159608082]},
 "algorithmic reduction-polynomial domain \[ScriptCapitalP]"->{
  Cell[3041167, 84764, 919, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1159608082", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", 
     "algorithmic reduction-polynomial domain \[ScriptCapitalP]"},
   CellID->1159608082]},
 "ID:847570831"->{
  Cell[3043175, 84826, 3075, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:847570831", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] \[Succeeds]"},
   CellID->847570831]},
 "\[CenterDot] \[Succeeds]"->{
  Cell[3043175, 84826, 3075, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:847570831", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] \[Succeeds]"},
   CellID->847570831]},
 "Proof|ID:847570831"->{
  Cell[3046406, 84928, 6346, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:847570831", "Proof|ID:847570831-1"},
   CellID->1483028309]},
 "Proof|ID:847570831-1"->{
  Cell[3046406, 84928, 6346, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:847570831", "Proof|ID:847570831-1"},
   CellID->1483028309]},
 "ID:2071048931"->{
  Cell[3053761, 85111, 1013, 29, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2071048931", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->2071048931]},
 "ID:355437230"->{
  Cell[3054777, 85142, 735, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:355437230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->355437230]},
 "ID:1401739284"->{
  Cell[3055515, 85163, 1077, 28, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1401739284", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1401739284]},
 "ID:29665786"->{
  Cell[3056595, 85193, 904, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:29665786", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 1"},
   CellID->29665786]},
 "\[Tilde] \[ScriptCapitalP] 1"->{
  Cell[3056595, 85193, 904, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:29665786", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 1"},
   CellID->29665786]},
 "ID:113282996"->{
  Cell[3057502, 85219, 1549, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:113282996", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 2"},
   CellID->113282996]},
 "\[Tilde] \[ScriptCapitalP] 2"->{
  Cell[3057502, 85219, 1549, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:113282996", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 2"},
   CellID->113282996]},
 "ID:1675445982"->{
  Cell[3059054, 85264, 1683, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1675445982", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 3"},
   CellID->1675445982]},
 "\[Tilde] \[ScriptCapitalP] 3"->{
  Cell[3059054, 85264, 1683, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1675445982", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 3"},
   CellID->1675445982]},
 "ID:1968653014"->{
  Cell[3060740, 85315, 1557, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1968653014", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 4"},
   CellID->1968653014]},
 "\[Tilde] \[ScriptCapitalP] 4"->{
  Cell[3060740, 85315, 1557, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1968653014", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 4"},
   CellID->1968653014]},
 "ID:860310855"->{
  Cell[3062300, 85361, 1241, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:860310855", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 5"},
   CellID->860310855]},
 "\[Tilde] \[ScriptCapitalP] 5"->{
  Cell[3062300, 85361, 1241, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:860310855", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 5"},
   CellID->860310855]},
 "ID:2026748686"->{
  Cell[3063544, 85399, 1403, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2026748686", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] multiplier \[ScriptCapitalP]"},
   CellID->2026748686]},
 "\[Tilde] multiplier \[ScriptCapitalP]"->{
  Cell[3063544, 85399, 1403, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2026748686", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] multiplier \[ScriptCapitalP]"},
   CellID->2026748686]},
 "Proof|ID:29665786"->{
  Cell[3065105, 85445, 5544, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:29665786", "Proof|ID:29665786-1"},
   CellID->343762797]},
 "Proof|ID:29665786-1"->{
  Cell[3065105, 85445, 5544, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:29665786", "Proof|ID:29665786-1"},
   CellID->343762797]},
 "Proof|ID:113282996"->{
  Cell[3071169, 85601, 6375, 153, 70, "ProofInfo",
   CellTags->{"Proof|ID:113282996", "Proof|ID:113282996-1"},
   CellID->1216564136]},
 "Proof|ID:113282996-1"->{
  Cell[3071169, 85601, 6375, 153, 70, "ProofInfo",
   CellTags->{"Proof|ID:113282996", "Proof|ID:113282996-1"},
   CellID->1216564136]},
 "Proof|ID:1675445982"->{
  Cell[3078064, 85773, 7133, 168, 70, "ProofInfo",
   CellTags->{"Proof|ID:1675445982", "Proof|ID:1675445982-1"},
   CellID->1330746921]},
 "Proof|ID:1675445982-1"->{
  Cell[3078064, 85773, 7133, 168, 70, "ProofInfo",
   CellTags->{"Proof|ID:1675445982", "Proof|ID:1675445982-1"},
   CellID->1330746921]},
 "Proof|ID:1968653014"->{
  Cell[3085719, 85960, 8247, 190, 70, "ProofInfo",
   CellTags->{"Proof|ID:1968653014", "Proof|ID:1968653014-1"},
   CellID->1366469966]},
 "Proof|ID:1968653014-1"->{
  Cell[3085719, 85960, 8247, 190, 70, "ProofInfo",
   CellTags->{"Proof|ID:1968653014", "Proof|ID:1968653014-1"},
   CellID->1366469966]},
 "Proof|ID:860310855"->{
  Cell[3094487, 86169, 5990, 145, 70, "ProofInfo",
   CellTags->{"Proof|ID:860310855", "Proof|ID:860310855-1"},
   CellID->1282192616]},
 "Proof|ID:860310855-1"->{
  Cell[3094487, 86169, 5990, 145, 70, "ProofInfo",
   CellTags->{"Proof|ID:860310855", "Proof|ID:860310855-1"},
   CellID->1282192616]},
 "Proof|ID:2026748686"->{
  Cell[3100997, 86333, 5596, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:2026748686", "Proof|ID:2026748686-1"},
   CellID->1041146073]},
 "Proof|ID:2026748686-1"->{
  Cell[3100997, 86333, 5596, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:2026748686", "Proof|ID:2026748686-1"},
   CellID->1041146073]},
 "ID:1529886137"->{
  Cell[3107555, 86507, 1857, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1529886137", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP]"},
   CellID->1529886137]},
 "\[Tilde] \[ScriptCapitalP]"->{
  Cell[3107555, 86507, 1857, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1529886137", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP]"},
   CellID->1529886137]},
 "Proof|ID:1529886137"->{
  Cell[3109568, 86574, 4182, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:1529886137", "Proof|ID:1529886137-1"},
   CellID->1495738535]},
 "Proof|ID:1529886137-1"->{
  Cell[3109568, 86574, 4182, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:1529886137", "Proof|ID:1529886137-1"},
   CellID->1495738535]},
 "ID:1680577066"->{
  Cell[3115030, 86725, 781, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1680577066", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "372"},
   CellID->1680577066]},
 "372"->{
  Cell[3115030, 86725, 781, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1680577066", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "372"},
   CellID->1680577066]},
 "Proof|ID:1680577066"->{
  Cell[3115969, 86757, 1731, 57, 28, "ProofInfo",
   CellTags->{"Proof|ID:1680577066", "Proof|ID:1680577066-1"},
   CellID->5711938]},
 "Proof|ID:1680577066-1"->{
  Cell[3115969, 86757, 1731, 57, 28, "ProofInfo",
   CellTags->{"Proof|ID:1680577066", "Proof|ID:1680577066-1"},
   CellID->5711938]},
 "ID:1866021817"->{
  Cell[3134797, 87335, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1866021817", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "373"},
   CellID->1866021817]},
 "373"->{
  Cell[3134797, 87335, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1866021817", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "373"},
   CellID->1866021817]},
 "Proof|ID:1866021817"->{
  Cell[3135737, 87367, 1733, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1866021817", "Proof|ID:1866021817-1"},
   CellID->1214587]},
 "Proof|ID:1866021817-1"->{
  Cell[3135737, 87367, 1733, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1866021817", "Proof|ID:1866021817-1"},
   CellID->1214587]},
 "ID:1272742036"->{
  Cell[3138653, 87465, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1272742036", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "374"},
   CellID->1272742036]},
 "374"->{
  Cell[3138653, 87465, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1272742036", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "374"},
   CellID->1272742036]},
 "Proof|ID:1272742036"->{
  Cell[3139593, 87497, 1735, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1272742036", "Proof|ID:1272742036-1"},
   CellID->634133424]},
 "Proof|ID:1272742036-1"->{
  Cell[3139593, 87497, 1735, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1272742036", "Proof|ID:1272742036-1"},
   CellID->634133424]},
 "ID:181479986"->{
  Cell[3142449, 87595, 1186, 29, 71, "GlobalDeclaration",
   CellTags->{
    "ID:181479986", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->181479986]},
 "ID:1276235216"->{
  Cell[3143638, 87626, 739, 19, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1276235216", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1276235216]},
 "ID:376128819"->{
  Cell[3144380, 87647, 896, 23, 58, "GlobalDeclaration",
   CellTags->{
    "ID:376128819", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->376128819]},
 "ID:973122334"->{
  Cell[3145279, 87672, 1014, 27, 49, "GlobalDeclaration",
   CellTags->{
    "ID:973122334", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->973122334]},
 "ID:1445738544"->{
  Cell[3146757, 87718, 2509, 75, 93, "FormalTextInputFormula",
   CellTags->{
    "ID:1445738544", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.1"},
   CellID->1445738544]},
 "375.1"->{
  Cell[3146757, 87718, 2509, 75, 93, "FormalTextInputFormula",
   CellTags->{
    "ID:1445738544", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.1"},
   CellID->1445738544]},
 "ID:874376987"->{
  Cell[3149269, 87795, 2968, 80, 81, "FormalTextInputFormula",
   CellTags->{
    "ID:874376987", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.2"},
   CellID->874376987]},
 "375.2"->{
  Cell[3149269, 87795, 2968, 80, 81, "FormalTextInputFormula",
   CellTags->{
    "ID:874376987", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.2"},
   CellID->874376987]},
 "ID:139929842"->{
  Cell[3152240, 87877, 1880, 54, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:139929842", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.3"},
   CellID->139929842]},
 "375.3"->{
  Cell[3152240, 87877, 1880, 54, 75, "FormalTextInputFormula",
   CellTags->{
    "ID:139929842", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.3"},
   CellID->139929842]},
 "Proof|ID:1445738544"->{
  Cell[3154277, 87942, 10777, 241, 28, "ProofInfo",
   CellTags->{"Proof|ID:1445738544", "Proof|ID:1445738544-2"},
   CellID->1345220325]},
 "Proof|ID:1445738544-2"->{
  Cell[3154277, 87942, 10777, 241, 28, "ProofInfo",
   CellTags->{"Proof|ID:1445738544", "Proof|ID:1445738544-2"},
   CellID->1345220325]},
 "Proof|ID:874376987"->{
  Cell[3165576, 88202, 13391, 292, 26, "ProofInfo",
   CellTags->{"Proof|ID:874376987", "Proof|ID:874376987-1"},
   CellID->7612480]},
 "Proof|ID:874376987-1"->{
  Cell[3165576, 88202, 13391, 292, 26, "ProofInfo",
   CellTags->{"Proof|ID:874376987", "Proof|ID:874376987-1"},
   CellID->7612480]},
 "Proof|ID:139929842"->{
  Cell[3179484, 88513, 10076, 230, 26, "ProofInfo",
   CellTags->{"Proof|ID:139929842", "Proof|ID:139929842-2"},
   CellID->1858784208]},
 "Proof|ID:139929842-2"->{
  Cell[3179484, 88513, 10076, 230, 26, "ProofInfo",
   CellTags->{"Proof|ID:139929842", "Proof|ID:139929842-2"},
   CellID->1858784208]},
 "ID:462890005"->{
  Cell[3190555, 88777, 780, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:462890005", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375"},
   CellID->462890005]},
 "375"->{
  Cell[3190555, 88777, 780, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:462890005", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375"},
   CellID->462890005]},
 "Proof|ID:462890005"->{
  Cell[3191493, 88809, 1725, 56, 28, "ProofInfo",
   CellTags->{"Proof|ID:462890005", "Proof|ID:462890005-1"},
   CellID->51997037]},
 "Proof|ID:462890005-1"->{
  Cell[3191493, 88809, 1725, 56, 28, "ProofInfo",
   CellTags->{"Proof|ID:462890005", "Proof|ID:462890005-1"},
   CellID->51997037]},
 "ID:38279669"->{
  Cell[3215133, 89507, 1623, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:38279669", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducible lm"},
   CellID->38279669]},
 "isReducible lm"->{
  Cell[3215133, 89507, 1623, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:38279669", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducible lm"},
   CellID->38279669]},
 "ID:1376605467"->{
  Cell[3216759, 89558, 1475, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1376605467", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "red lm set"},
   CellID->1376605467]},
 "red lm set"->{
  Cell[3216759, 89558, 1475, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1376605467", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "red lm set"},
   CellID->1376605467]},
 "ID:673925715"->{
  Cell[3218237, 89604, 3120, 87, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:673925715", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducible monomial"},
   CellID->673925715]},
 "isReducible monomial"->{
  Cell[3218237, 89604, 3120, 87, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:673925715", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducible monomial"},
   CellID->673925715]},
 "Proof|ID:38279669"->{
  Cell[3221513, 89702, 4153, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:38279669", "Proof|ID:38279669-1"},
   CellID->51695314]},
 "Proof|ID:38279669-1"->{
  Cell[3221513, 89702, 4153, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:38279669", "Proof|ID:38279669-1"},
   CellID->51695314]},
 "Proof|ID:1376605467"->{
  Cell[3226185, 89829, 3980, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:1376605467", "Proof|ID:1376605467-1"},
   CellID->580557624]},
 "Proof|ID:1376605467-1"->{
  Cell[3226185, 89829, 3980, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:1376605467", "Proof|ID:1376605467-1"},
   CellID->580557624]},
 "Proof|ID:673925715"->{
  Cell[3230685, 89955, 6350, 152, 70, "ProofInfo",
   CellTags->{"Proof|ID:673925715", "Proof|ID:673925715-1"},
   CellID->54334308]},
 "Proof|ID:673925715-1"->{
  Cell[3230685, 89955, 6350, 152, 70, "ProofInfo",
   CellTags->{"Proof|ID:673925715", "Proof|ID:673925715-1"},
   CellID->54334308]},
 "ID:47923252"->{
  Cell[3238019, 90141, 1876, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47923252", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence"},
   CellID->47923252]},
 "isMonomialSequence"->{
  Cell[3238019, 90141, 1876, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47923252", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence"},
   CellID->47923252]},
 "ID:1632720588"->{
  Cell[3240499, 90224, 739, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1632720588", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->1632720588]},
 "ID:1771873303"->{
  Cell[3241241, 90245, 1075, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1771873303", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence \[NotEqual] 0"},
   CellID->1771873303]},
 "isMonomialSequence \[NotEqual] 0"->{
  Cell[3241241, 90245, 1075, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1771873303", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence \[NotEqual] 0"},
   CellID->1771873303]},
 "ID:893563053"->{
  Cell[3242319, 90277, 1052, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:893563053", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lc 1"},
   CellID->893563053]},
 "isMonomialSequence lc 1"->{
  Cell[3242319, 90277, 1052, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:893563053", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lc 1"},
   CellID->893563053]},
 "ID:587765202"->{
  Cell[3243374, 90308, 1054, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:587765202", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lp 1"},
   CellID->587765202]},
 "isMonomialSequence lp 1"->{
  Cell[3243374, 90308, 1054, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:587765202", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lp 1"},
   CellID->587765202]},
 "ID:1254733295"->{
  Cell[3244431, 90339, 991, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1254733295", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lc 2"},
   CellID->1254733295]},
 "isMonomialSequence lc 2"->{
  Cell[3244431, 90339, 991, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1254733295", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lc 2"},
   CellID->1254733295]},
 "ID:280696418"->{
  Cell[3245425, 90368, 988, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:280696418", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lp 2"},
   CellID->280696418]},
 "isMonomialSequence lp 2"->{
  Cell[3245425, 90368, 988, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:280696418", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lp 2"},
   CellID->280696418]},
 "ID:1081126646"->{
  Cell[3246416, 90397, 989, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1081126646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence subSeq"},
   CellID->1081126646]},
 "isMonomialSequence subSeq"->{
  Cell[3246416, 90397, 989, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1081126646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence subSeq"},
   CellID->1081126646]},
 "ID:1101697023"->{
  Cell[3247408, 90425, 973, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1101697023", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence subsequence"},
   CellID->1101697023]},
 "isMonomialSequence subsequence"->{
  Cell[3247408, 90425, 973, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1101697023", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence subsequence"},
   CellID->1101697023]},
 "Proof|ID:1771873303"->{
  Cell[3248537, 90462, 3303, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:1771873303", "Proof|ID:1771873303-1"},
   CellID->103075955]},
 "Proof|ID:1771873303-1"->{
  Cell[3248537, 90462, 3303, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:1771873303", "Proof|ID:1771873303-1"},
   CellID->103075955]},
 "Proof|ID:893563053"->{
  Cell[3252362, 90575, 3468, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:893563053", "Proof|ID:893563053-1"},
   CellID->88855734]},
 "Proof|ID:893563053-1"->{
  Cell[3252362, 90575, 3468, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:893563053", "Proof|ID:893563053-1"},
   CellID->88855734]},
 "Proof|ID:587765202"->{
  Cell[3256351, 90688, 3470, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:587765202", "Proof|ID:587765202-1"},
   CellID->1976154064]},
 "Proof|ID:587765202-1"->{
  Cell[3256351, 90688, 3470, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:587765202", "Proof|ID:587765202-1"},
   CellID->1976154064]},
 "Proof|ID:1254733295"->{
  Cell[3260342, 90801, 3009, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:1254733295", "Proof|ID:1254733295-1"},
   CellID->1511544165]},
 "Proof|ID:1254733295-1"->{
  Cell[3260342, 90801, 3009, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:1254733295", "Proof|ID:1254733295-1"},
   CellID->1511544165]},
 "Proof|ID:280696418"->{
  Cell[3263872, 90905, 2999, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:280696418", "Proof|ID:280696418-1"},
   CellID->276631526]},
 "Proof|ID:280696418-1"->{
  Cell[3263872, 90905, 2999, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:280696418", "Proof|ID:280696418-1"},
   CellID->276631526]},
 "Proof|ID:1081126646"->{
  Cell[3267391, 91008, 2912, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:1081126646", "Proof|ID:1081126646-1"},
   CellID->227638759]},
 "Proof|ID:1081126646-1"->{
  Cell[3267391, 91008, 2912, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:1081126646", "Proof|ID:1081126646-1"},
   CellID->227638759]},
 "Proof|ID:1101697023"->{
  Cell[3270825, 91108, 2946, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:1101697023", "Proof|ID:1101697023-1"},
   CellID->881838494]},
 "Proof|ID:1101697023-1"->{
  Cell[3270825, 91108, 2946, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:1101697023", "Proof|ID:1101697023-1"},
   CellID->881838494]},
 "ID:182970451"->{
  Cell[3275758, 91256, 3026, 89, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:182970451", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "110"},
   CellID->182970451]},
 "110"->{
  Cell[3275758, 91256, 3026, 89, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:182970451", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "110"},
   CellID->182970451]},
 "ID:845195690"->{
  Cell[3278787, 91347, 3164, 92, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:845195690", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "111"},
   CellID->845195690]},
 "111"->{
  Cell[3278787, 91347, 3164, 92, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:845195690", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "111"},
   CellID->845195690]},
 "Proof|ID:182970451"->{
  Cell[3282109, 91450, 8308, 207, 70, "ProofInfo",
   CellTags->{"Proof|ID:182970451", "Proof|ID:182970451-1"},
   CellID->568281335]},
 "Proof|ID:182970451-1"->{
  Cell[3282109, 91450, 8308, 207, 70, "ProofInfo",
   CellTags->{"Proof|ID:182970451", "Proof|ID:182970451-1"},
   CellID->568281335]},
 "Proof|ID:845195690"->{
  Cell[3290937, 91676, 8270, 206, 70, "ProofInfo",
   CellTags->{"Proof|ID:845195690", "Proof|ID:845195690-1"},
   CellID->70458024]},
 "Proof|ID:845195690-1"->{
  Cell[3290937, 91676, 8270, 206, 70, "ProofInfo",
   CellTags->{"Proof|ID:845195690", "Proof|ID:845195690-1"},
   CellID->70458024]},
 "ID:2081362811"->{
  Cell[3300288, 91919, 739, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2081362811", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
   CellID->2081362811]},
 "ID:1173235103"->{
  Cell[3301030, 91940, 4239, 120, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1173235103", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "112"},
   CellID->1173235103]},
 "112"->{
  Cell[3301030, 91940, 4239, 120, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1173235103", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "112"},
   CellID->1173235103]},
 "ID:1805755328"->{
  Cell[3305272, 92062, 2318, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1805755328", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "113"},
   CellID->1805755328]},
 "113"->{
  Cell[3305272, 92062, 2318, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1805755328", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "113"},
   CellID->1805755328]},
 "ID:596327945"->{
  Cell[3307593, 92134, 2490, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:596327945", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "114"},
   CellID->596327945]},
 "114"->{
  Cell[3307593, 92134, 2490, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:596327945", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "114"},
   CellID->596327945]},
 "Proof|ID:1173235103"->{
  Cell[3310238, 92212, 11823, 279, 70, "ProofInfo",
   CellTags->{"Proof|ID:1173235103", "Proof|ID:1173235103-1"},
   CellID->874379054]},
 "Proof|ID:1173235103-1"->{
  Cell[3310238, 92212, 11823, 279, 70, "ProofInfo",
   CellTags->{"Proof|ID:1173235103", "Proof|ID:1173235103-1"},
   CellID->874379054]},
 "Proof|ID:1805755328"->{
  Cell[3322582, 92510, 6387, 159, 70, "ProofInfo",
   CellTags->{"Proof|ID:1805755328", "Proof|ID:1805755328-1"},
   CellID->1937389710]},
 "Proof|ID:1805755328-1"->{
  Cell[3322582, 92510, 6387, 159, 70, "ProofInfo",
   CellTags->{"Proof|ID:1805755328", "Proof|ID:1805755328-1"},
   CellID->1937389710]},
 "Proof|ID:596327945"->{
  Cell[3329533, 92689, 5826, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:596327945", "Proof|ID:596327945-1"},
   CellID->387192426]},
 "Proof|ID:596327945-1"->{
  Cell[3329533, 92689, 5826, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:596327945", "Proof|ID:596327945-1"},
   CellID->387192426]},
 "ID:1402642391"->{
  Cell[3336511, 92874, 2262, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1402642391", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "115"},
   CellID->1402642391]},
 "115"->{
  Cell[3336511, 92874, 2262, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1402642391", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "115"},
   CellID->1402642391]},
 "Proof|ID:1402642391"->{
  Cell[3338930, 92955, 7083, 181, 70, "ProofInfo",
   CellTags->{"Proof|ID:1402642391", "Proof|ID:1402642391-1"},
   CellID->1177596546]},
 "Proof|ID:1402642391-1"->{
  Cell[3338930, 92955, 7083, 181, 70, "ProofInfo",
   CellTags->{"Proof|ID:1402642391", "Proof|ID:1402642391-1"},
   CellID->1177596546]},
 "ID:980978647"->{
  Cell[3346997, 93171, 1081, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:980978647", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Dickson-domain \[ScriptCapitalT]"},
   CellID->980978647]},
 "Dickson-domain \[ScriptCapitalT]"->{
  Cell[3346997, 93171, 1081, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:980978647", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Dickson-domain \[ScriptCapitalT]"},
   CellID->980978647]},
 "ID:1408052308"->{
  Cell[3348735, 93227, 756, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1408052308", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "116"},
   CellID->1408052308]},
 "116"->{
  Cell[3348735, 93227, 756, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1408052308", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "116"},
   CellID->1408052308]},
 "Proof|ID:1408052308"->{
  Cell[3349649, 93258, 1735, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1408052308", "Proof|ID:1408052308-1"},
   CellID->245205539]},
 "Proof|ID:1408052308-1"->{
  Cell[3349649, 93258, 1735, 57, 70, "ProofInfo",
   CellTags->{"Proof|ID:1408052308", "Proof|ID:1408052308-1"},
   CellID->245205539]},
 "ID:1183157947"->{
  Cell[3352664, 93359, 857, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1183157947", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isAlgoReductionRing \[ScriptCapitalP]"},
   CellID->1183157947]},
 "isAlgoReductionRing \[ScriptCapitalP]"->{
  Cell[3352664, 93359, 857, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1183157947", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isAlgoReductionRing \[ScriptCapitalP]"},
   CellID->1183157947]},
 "Proof|ID:1183157947"->{
  Cell[3353678, 93392, 1859, 59, 70, "ProofInfo",
   CellTags->{"Proof|ID:1183157947", "Proof|ID:1183157947-1"},
   CellID->1299983476]},
 "Proof|ID:1183157947-1"->{
  Cell[3353678, 93392, 1859, 59, 70, "ProofInfo",
   CellTags->{"Proof|ID:1183157947", "Proof|ID:1183157947-1"},
   CellID->1299983476]},
 "ID:1794188423"->{
  Cell[3360083, 93567, 2107, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1794188423", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", 
     "isAlgoReductionRing Conservation Theorem Polynomials"},
   CellID->1794188423]},
 "isAlgoReductionRing Conservation Theorem Polynomials"->{
  Cell[3360083, 93567, 2107, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1794188423", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", 
     "isAlgoReductionRing Conservation Theorem Polynomials"},
   CellID->1794188423]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"ID:558626034", 3399935, 94404},
 {"Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", 3400250, 94412},
 {"isCommPPDomain", 3513752, 97463},
 {"ID:1568169011", 3514001, 97470},
 {"ID:1716984519", 3514226, 97477},
 {"ID:205548127", 3514450, 97484},
 {"isMonomial", 3514690, 97491},
 {"ID:711054552", 3514932, 97498},
 {"isMonomial II", 3515178, 97505},
 {"ID:2035355406", 3515424, 97512},
 {"monomial constructor", 3515686, 97519},
 {"ID:371468988", 3515940, 97526},
 {"supp", 3516168, 97533},
 {"ID:741382791", 3516404, 97540},
 {"isLP", 3516632, 97547},
 {"ID:1470398456", 3516869, 97554},
 {"lp", 3517095, 97561},
 {"ID:1541081556", 3517332, 97568},
 {"lc", 3517558, 97575},
 {"ID:258866072", 3517794, 97582},
 {"lm", 3518018, 97589},
 {"ID:1500059087", 3518253, 97596},
 {"R", 3518477, 97603},
 {"ID:1301149083", 3518713, 97610},
 {"H 0", 3518941, 97617},
 {"ID:355792939", 3519178, 97624},
 {"H", 3519400, 97631},
 {"ID:1617463509", 3519634, 97638},
 {"monomial multiplication", 3519903, 97645},
 {"ID:1026275980", 3520162, 97652},
 {"isReductionPolynomialDomain", 3520439, 97659},
 {"ID:1335218902", 3520702, 97666},
 {"hasDecisionPredicate", 3520966, 97673},
 {"ID:1175095894", 3521223, 97680},
 {"containsMonomials", 3521480, 97687},
 {"ID:1967795228", 3521733, 97694},
 {"hasCoefficientFunction", 3522001, 97701},
 {"ID:2138030312", 3522260, 97708},
 {"hasMultiplication", 3522517, 97715},
 {"ID:1709544239", 3522770, 97722},
 {"isExtensional", 3523019, 97729},
 {"ID:235606987", 3523267, 97736},
 {"isOrdered", 3523506, 97743},
 {"ID:407906314", 3523748, 97750},
 {"hasMultipliers", 3523998, 97757},
 {"ID:1209025019", 3524247, 97764},
 {"isAlgoReductionPolynomialDomain", 3524532, 97771},
 {"ID:1424238209", 3524799, 97778},
 {"hasLCRD", 3525037, 97785},
 {"ID:1761006179", 3525281, 97792},
 {"hasRDM", 3525517, 97799},
 {"ID:667868750", 3525759, 97806},
 {"hasCPM", 3525993, 97813},
 {"ID:267444622", 3526233, 97820},
 {"ring \[ScriptCapitalR]", 3526497, 97827},
 {"ID:1725827561", 3526752, 97834},
 {"power-product domain \[ScriptCapitalT]", 3527050, 97841},
 {"ID:90242530", 3527321, 97848},
 {"isDomain \[ScriptCapitalP]", 3527591, 97855},
 {"ID:1949049350", 3527848, 97862},
 {"hasDecisionPredicate \[ScriptCapitalP]", 3528146, 97869},
 {"ID:378755333", 3528418, 97876},
 {"containsMonomials \[ScriptCapitalP]", 3528708, 97883},
 {"ID:226930466", 3528975, 97890},
 {"hasCoefficientFunction \[ScriptCapitalP]", 3529275, 97897},
 {"ID:221039322", 3529547, 97904},
 {"isExtensional \[ScriptCapitalP]", 3529829, 97911},
 {"ID:216808867", 3530092, 97918},
 {"isDomain \[ScriptCapitalT]", 3530364, 97925},
 {"ID:1391060217", 3530623, 97932},
 {"isPartIrreflOrder \[Succeeds] \[ScriptCapitalT]", 3530940, 97939},
 {"ID:1131450960", 3531223, 97946},
 {"Noetherian induction on \[ScriptCapitalT]", 3531534, 97953},
 {"IR", 3531806, 97960},
 {"ID:953344354", 3536846, 98087},
 {"membership 1 \[ScriptCapitalT]", 3537132, 98094},
 {"ID:1367902825", 3537401, 98101},
 {"PP 1", 3537632, 98108},
 {"ID:643177315", 3537871, 98115},
 {"ID:1568022983", 3538095, 98122},
 {"PP 1 min", 3538333, 98129},
 {"ID:1723905914", 3538576, 98136},
 {"PP 0", 3538807, 98143},
 {"ID:1825953062", 3539047, 98150},
 {"PP 2'", 3539279, 98157},
 {"ID:596582273", 3539518, 98164},
 {"PP 2", 3539746, 98171},
 {"ID:389623087", 3539982, 98178},
 {"PP 3'", 3540212, 98185},
 {"ID:885344283", 3540449, 98192},
 {"PP 3", 3540677, 98199},
 {"ID:1033740528", 3540914, 98206},
 {"PP 4", 3541145, 98213},
 {"ID:475932409", 3541384, 98220},
 {"ID:1124160616", 3541608, 98227},
 {"PP 5", 3541839, 98234},
 {"ID:1594052474", 3542079, 98241},
 {"PP 6", 3542310, 98248},
 {"ID:1091730754", 3542550, 98255},
 {"PP 7", 3542781, 98262},
 {"ID:1114122233", 3543021, 98269},
 {"PP 8", 3543252, 98276},
 {"ID:1909427587", 3543492, 98283},
 {"isPartReflOrder \[VerticalBar]", 3543775, 98290},
 {"ID:196272597", 3544040, 98297},
 {"isOrderEmbedding \[RawStar] \[VerticalBar]", 3544345, 98304},
 {"Proof|ID:216808867", 3544626, 98311},
 {"Proof|ID:216808867-1", 3544785, 98315},
 {"Proof|ID:1391060217", 3544943, 98319},
 {"Proof|ID:1391060217-1", 3545106, 98323},
 {"Proof|ID:1131450960", 3545267, 98327},
 {"Proof|ID:1131450960-1", 3545431, 98331},
 {"Proof|ID:953344354", 3545592, 98335},
 {"Proof|ID:953344354-1", 3545752, 98339},
 {"Proof|ID:1367902825", 3545911, 98343},
 {"Proof|ID:1367902825-1", 3546075, 98347},
 {"Proof|ID:1568022983", 3546237, 98351},
 {"Proof|ID:1568022983-1", 3546401, 98355},
 {"Proof|ID:1723905914", 3546563, 98359},
 {"Proof|ID:1723905914-1", 3546727, 98363},
 {"Proof|ID:1825953062", 3546889, 98367},
 {"Proof|ID:1825953062-1", 3547051, 98371},
 {"Proof|ID:596582273", 3547210, 98375},
 {"Proof|ID:596582273-1", 3547370, 98379},
 {"Proof|ID:389623087", 3547528, 98383},
 {"Proof|ID:389623087-1", 3547688, 98387},
 {"Proof|ID:885344283", 3547846, 98391},
 {"Proof|ID:885344283-1", 3548006, 98395},
 {"Proof|ID:1033740528", 3548165, 98399},
 {"Proof|ID:1033740528-1", 3548329, 98403},
 {"Proof|ID:1124160616", 3548491, 98407},
 {"Proof|ID:1124160616-1", 3548655, 98411},
 {"Proof|ID:1594052474", 3548817, 98415},
 {"Proof|ID:1594052474-1", 3548982, 98419},
 {"Proof|ID:1091730754", 3549145, 98423},
 {"Proof|ID:1091730754-1", 3549309, 98427},
 {"Proof|ID:1114122233", 3549471, 98431},
 {"Proof|ID:1114122233-1", 3549636, 98435},
 {"Proof|ID:1909427587", 3549799, 98439},
 {"Proof|ID:1909427587-1", 3549962, 98443},
 {"Proof|ID:196272597", 3550122, 98447},
 {"Proof|ID:196272597-1", 3550282, 98451},
 {"ID:334245933", 3550434, 98455},
 {"ID:1101510167", 3550658, 98462},
 {"membership lcm", 3550914, 98469},
 {"ID:571674285", 3551168, 98476},
 {"lcm 1", 3551398, 98483},
 {"ID:775675057", 3551635, 98490},
 {"lcm 1'", 3551867, 98497},
 {"ID:517355459", 3552105, 98504},
 {"lcm 2", 3552336, 98511},
 {"ID:651227563", 3552574, 98518},
 {"lcm 3", 3552804, 98525},
 {"ID:482943675", 3553041, 98532},
 {"lcm 6", 3553271, 98539},
 {"ID:39860588", 3553507, 98546},
 {"lcm 4", 3553735, 98553},
 {"ID:1631465634", 3553971, 98560},
 {"lcm 4'", 3554205, 98567},
 {"ID:835022497", 3554445, 98574},
 {"lcm 5", 3554676, 98581},
 {"Proof|ID:1101510167", 3554921, 98588},
 {"Proof|ID:1101510167-1", 3555085, 98592},
 {"Proof|ID:571674285", 3555246, 98596},
 {"Proof|ID:571674285-1", 3555406, 98600},
 {"Proof|ID:775675057", 3555564, 98604},
 {"Proof|ID:775675057-1", 3555725, 98608},
 {"Proof|ID:517355459", 3555884, 98612},
 {"Proof|ID:517355459-1", 3556046, 98616},
 {"Proof|ID:651227563", 3556206, 98620},
 {"Proof|ID:651227563-1", 3556367, 98624},
 {"Proof|ID:482943675", 3556526, 98628},
 {"Proof|ID:482943675-1", 3556685, 98632},
 {"Proof|ID:39860588", 3556841, 98636},
 {"Proof|ID:39860588-1", 3556999, 98640},
 {"Proof|ID:1631465634", 3557157, 98644},
 {"Proof|ID:1631465634-1", 3557321, 98648},
 {"Proof|ID:835022497", 3557482, 98652},
 {"Proof|ID:835022497-1", 3557643, 98656},
 {"ID:100863090", 3557796, 98660},
 {"ID:292726431", 3558019, 98667},
 {"membership supp general", 3558285, 98674},
 {"ID:109557753", 3558540, 98681},
 {"\[Element] supp general", 3558807, 98688},
 {"Proof|ID:292726431", 3559069, 98695},
 {"Proof|ID:292726431-1", 3559229, 98699},
 {"Proof|ID:109557753", 3559387, 98703},
 {"Proof|ID:109557753-1", 3559548, 98707},
 {"ID:938132667", 3559701, 98711},
 {"membership supp", 3559958, 98718},
 {"ID:56177686", 3560211, 98725},
 {"\[Element] supp", 3560466, 98732},
 {"ID:163445395", 3560718, 98739},
 {"membership 0", 3560968, 98746},
 {"ID:671805226", 3561218, 98753},
 {"membership 1", 3561468, 98760},
 {"ID:373967346", 3561718, 98767},
 {"membership C", 3561969, 98774},
 {"Proof|ID:938132667", 3562226, 98781},
 {"Proof|ID:938132667-1", 3562385, 98785},
 {"Proof|ID:56177686", 3562541, 98789},
 {"Proof|ID:56177686-1", 3562699, 98793},
 {"Proof|ID:163445395", 3562856, 98797},
 {"Proof|ID:163445395-1", 3563015, 98801},
 {"Proof|ID:671805226", 3563172, 98805},
 {"Proof|ID:671805226-1", 3563332, 98809},
 {"Proof|ID:373967346", 3563490, 98813},
 {"Proof|ID:373967346-1", 3563650, 98817},
 {"ID:1248732928", 3563803, 98821},
 {"Monomial", 3564042, 98828},
 {"ID:1277602733", 3564286, 98835},
 {"lp sufficient", 3564535, 98842},
 {"ID:1270225306", 3564784, 98849},
 {"lp alternative", 3565035, 98856},
 {"Proof|ID:1248732928", 3565291, 98863},
 {"Proof|ID:1248732928-1", 3565455, 98867},
 {"Proof|ID:1277602733", 3565617, 98871},
 {"Proof|ID:1277602733-1", 3565781, 98875},
 {"Proof|ID:1270225306", 3565943, 98879},
 {"Proof|ID:1270225306-1", 3566108, 98883},
 {"ID:292606918", 3566264, 98887},
 {"membership monomials", 3566532, 98894},
 {"ID:485824967", 3566792, 98901},
 {"ID:406494450", 3567016, 98908},
 {"membership lp", 3567269, 98915},
 {"ID:1454724371", 3567522, 98922},
 {"membership lc", 3567777, 98929},
 {"ID:1589530505", 3568032, 98936},
 {"membership lm", 3568287, 98943},
 {"Proof|ID:292606918", 3568547, 98950},
 {"Proof|ID:292606918-1", 3568709, 98954},
 {"Proof|ID:406494450", 3568869, 98958},
 {"Proof|ID:406494450-1", 3569030, 98962},
 {"Proof|ID:1454724371", 3569190, 98966},
 {"Proof|ID:1454724371-1", 3569354, 98970},
 {"Proof|ID:1589530505", 3569516, 98974},
 {"Proof|ID:1589530505-1", 3569681, 98978},
 {"ID:720209465", 3569837, 98982},
 {"ID:1852885314", 3570062, 98989},
 {"C 0", 3570291, 98996},
 {"ID:148341425", 3570529, 99003},
 {"ID:1443097690", 3570754, 99010},
 {"C 1", 3570984, 99017},
 {"ID:1675126992", 3571224, 99024},
 {"C 2", 3571454, 99031},
 {"ID:814027629", 3571693, 99038},
 {"C 3", 3571921, 99045},
 {"ID:1597798077", 3572159, 99052},
 {"C 4", 3572389, 99059},
 {"ID:1711645326", 3572629, 99066},
 {"C 5", 3572859, 99073},
 {"Proof|ID:1852885314", 3573105, 99080},
 {"Proof|ID:1852885314-1", 3573269, 99084},
 {"Proof|ID:1443097690", 3573431, 99088},
 {"Proof|ID:1443097690-1", 3573596, 99092},
 {"Proof|ID:1675126992", 3573759, 99096},
 {"Proof|ID:1675126992-1", 3573924, 99100},
 {"Proof|ID:814027629", 3574086, 99104},
 {"Proof|ID:814027629-1", 3574246, 99108},
 {"Proof|ID:1597798077", 3574405, 99112},
 {"Proof|ID:1597798077-1", 3574571, 99116},
 {"Proof|ID:1711645326", 3574735, 99120},
 {"Proof|ID:1711645326-1", 3574900, 99124},
 {"ID:1053827266", 3575057, 99128},
 {"ID:829207957", 3575283, 99135},
 {"membership + \[ScriptCapitalP]", 3575570, 99142},
 {"ID:227196886", 3575839, 99149},
 {"membership - \[ScriptCapitalP] 1", 3576130, 99156},
 {"ID:414658623", 3576401, 99163},
 {"membership - \[ScriptCapitalP] 2", 3576692, 99170},
 {"Proof|ID:829207957", 3576969, 99177},
 {"Proof|ID:829207957-1", 3577130, 99181},
 {"Proof|ID:227196886", 3577289, 99185},
 {"Proof|ID:227196886-1", 3577449, 99189},
 {"Proof|ID:414658623", 3577607, 99193},
 {"Proof|ID:414658623-1", 3577767, 99197},
 {"ID:314696779", 3577919, 99201},
 {"1=0 \[ScriptCapitalP]", 3578183, 99208},
 {"ID:849229881", 3578438, 99215},
 {"lp 1", 3578668, 99222},
 {"ID:406483636", 3578906, 99229},
 {"lc 1", 3579136, 99236},
 {"ID:1767085150", 3579375, 99243},
 {"lm 1", 3579607, 99250},
 {"ID:1272860149", 3579848, 99257},
 {"ID:144989886", 3580074, 99264},
 {"non-zero 2", 3580321, 99271},
 {"ID:2138850160", 3580571, 99278},
 {"lp 2", 3580803, 99285},
 {"ID:1145306026", 3581044, 99292},
 {"lc 2", 3581275, 99299},
 {"ID:1442316125", 3581515, 99306},
 {"lc 3", 3581746, 99313},
 {"ID:655883205", 3581985, 99320},
 {"lp 3", 3582215, 99327},
 {"ID:996448533", 3582453, 99334},
 {"non-zero 3", 3582701, 99341},
 {"ID:1383271844", 3582952, 99348},
 {"lp 5", 3583184, 99355},
 {"ID:945941808", 3583424, 99362},
 {"ID:857665826", 3583648, 99369},
 {"C lm", 3583878, 99376},
 {"ID:1411816658", 3584117, 99383},
 {"lm 2", 3584349, 99390},
 {"Proof|ID:314696779", 3584595, 99397},
 {"Proof|ID:314696779-1", 3584756, 99401},
 {"Proof|ID:849229881", 3584915, 99405},
 {"Proof|ID:849229881-1", 3585076, 99409},
 {"Proof|ID:406483636", 3585235, 99413},
 {"Proof|ID:406483636-1", 3585397, 99417},
 {"Proof|ID:1767085150", 3585558, 99421},
 {"Proof|ID:1767085150-1", 3585722, 99425},
 {"Proof|ID:144989886", 3585883, 99429},
 {"Proof|ID:144989886-1", 3586044, 99433},
 {"Proof|ID:2138850160", 3586204, 99437},
 {"Proof|ID:2138850160-1", 3586369, 99441},
 {"Proof|ID:1145306026", 3586532, 99445},
 {"Proof|ID:1145306026-1", 3586696, 99449},
 {"Proof|ID:1442316125", 3586858, 99453},
 {"Proof|ID:1442316125-1", 3587023, 99457},
 {"Proof|ID:655883205", 3587185, 99461},
 {"Proof|ID:655883205-1", 3587347, 99465},
 {"Proof|ID:996448533", 3587507, 99469},
 {"Proof|ID:996448533-1", 3587669, 99473},
 {"Proof|ID:1383271844", 3587830, 99477},
 {"Proof|ID:1383271844-1", 3587996, 99481},
 {"Proof|ID:857665826", 3588159, 99485},
 {"Proof|ID:857665826-1", 3588321, 99489},
 {"Proof|ID:1411816658", 3588482, 99493},
 {"Proof|ID:1411816658-1", 3588647, 99497},
 {"ID:1368691869", 3588804, 99501},
 {"1\[CenterDot]1", 3589055, 99508},
 {"ID:1325102956", 3589305, 99515},
 {"ID:1663774274", 3589532, 99522},
 {"ID:941688542", 3589758, 99529},
 {"isMonomial \[CenterDot]", 3590026, 99536},
 {"ID:91567084", 3590282, 99543},
 {"\[CenterDot] 0", 3590530, 99550},
 {"ID:420812829", 3590776, 99557},
 {"\[CenterDot] 3", 3591025, 99564},
 {"ID:138547187", 3591272, 99571},
 {"\[CenterDot] 4", 3591522, 99578},
 {"ID:457378178", 3591770, 99585},
 {"\[CenterDot] 11", 3592022, 99592},
 {"ID:926082604", 3592271, 99599},
 {"\[CenterDot] 11'", 3592525, 99606},
 {"ID:1344734857", 3592776, 99613},
 {"\[CenterDot] lp lc", 3593036, 99620},
 {"Proof|ID:1368691869", 3593297, 99627},
 {"Proof|ID:1368691869-1", 3593461, 99631},
 {"Proof|ID:941688542", 3593622, 99635},
 {"Proof|ID:941688542-1", 3593784, 99639},
 {"Proof|ID:91567084", 3593943, 99643},
 {"Proof|ID:91567084-1", 3594102, 99647},
 {"Proof|ID:420812829", 3594260, 99651},
 {"Proof|ID:420812829-1", 3594422, 99655},
 {"Proof|ID:138547187", 3594582, 99659},
 {"Proof|ID:138547187-1", 3594744, 99663},
 {"Proof|ID:457378178", 3594904, 99667},
 {"Proof|ID:457378178-1", 3595066, 99671},
 {"Proof|ID:926082604", 3595226, 99675},
 {"Proof|ID:926082604-1", 3595389, 99679},
 {"Proof|ID:1344734857", 3595551, 99683},
 {"Proof|ID:1344734857-1", 3595716, 99687},
 {"ID:1558664548", 3595873, 99691},
 {"isMonomial 1", 3596121, 99698},
 {"ID:1835029788", 3596370, 99705},
 {"ID:757450222", 3596596, 99712},
 {"isMonomial \[NotEqual] 0", 3596865, 99719},
 {"ID:1978849375", 3597123, 99726},
 {"isMonomial -", 3597370, 99733},
 {"ID:283455673", 3597617, 99740},
 {"isMonomial 3'", 3597864, 99747},
 {"ID:610787620", 3598110, 99754},
 {"isMonomial 3", 3598356, 99761},
 {"ID:980082706", 3598602, 99768},
 {"isMonomial 4", 3598848, 99775},
 {"ID:1306217654", 3599095, 99782},
 {"lm isMonomial", 3599344, 99789},
 {"ID:1645795930", 3599593, 99796},
 {"isMonomial 7", 3599841, 99803},
 {"ID:691307310", 3600089, 99810},
 {"ID:1901309763", 3600314, 99817},
 {"isMonomial 8", 3600562, 99824},
 {"ID:1177943318", 3600811, 99831},
 {"isMonomial 9", 3601059, 99838},
 {"ID:1770611523", 3601308, 99845},
 {"ID:670623560", 3601534, 99852},
 {"isMonomial 12", 3601782, 99859},
 {"ID:1806313475", 3602030, 99866},
 {"isMonomial lm", 3602279, 99873},
 {"ID:1110032163", 3602528, 99880},
 {"lc lm", 3602761, 99887},
 {"ID:632352665", 3603001, 99894},
 {"lp lm", 3603232, 99901},
 {"Proof|ID:1558664548", 3603477, 99908},
 {"Proof|ID:1558664548-1", 3603641, 99912},
 {"Proof|ID:757450222", 3603802, 99916},
 {"Proof|ID:757450222-1", 3603964, 99920},
 {"Proof|ID:1978849375", 3604125, 99924},
 {"Proof|ID:1978849375-1", 3604289, 99928},
 {"Proof|ID:283455673", 3604450, 99932},
 {"Proof|ID:283455673-1", 3604612, 99936},
 {"Proof|ID:610787620", 3604772, 99940},
 {"Proof|ID:610787620-1", 3604935, 99944},
 {"Proof|ID:980082706", 3605096, 99948},
 {"Proof|ID:980082706-1", 3605259, 99952},
 {"Proof|ID:1306217654", 3605421, 99956},
 {"Proof|ID:1306217654-1", 3605585, 99960},
 {"Proof|ID:1645795930", 3605747, 99964},
 {"Proof|ID:1645795930-1", 3605912, 99968},
 {"Proof|ID:1901309763", 3606075, 99972},
 {"Proof|ID:1901309763-1", 3606240, 99976},
 {"Proof|ID:1177943318", 3606403, 99980},
 {"Proof|ID:1177943318-1", 3606568, 99984},
 {"Proof|ID:670623560", 3606730, 99988},
 {"Proof|ID:670623560-1", 3606893, 99992},
 {"Proof|ID:1806313475", 3607055, 99996},
 {"Proof|ID:1806313475-1", 3607219, 100000},
 {"Proof|ID:1110032163", 3607381, 100004},
 {"Proof|ID:1110032163-1", 3607545, 100008},
 {"Proof|ID:632352665", 3607706, 100012},
 {"Proof|ID:632352665-1", 3607867, 100016},
 {"ID:423895762", 3608020, 100020},
 {"+ closed \[ScriptCapitalP]", 3608294, 100027},
 {"Proof|ID:423895762", 3608560, 100034},
 {"Proof|ID:423895762-1", 3608721, 100038},
 {"ID:1812309492", 3608875, 100042},
 {"0 neutral \[ScriptCapitalP]", 3609153, 100049},
 {"Proof|ID:1812309492", 3609423, 100056},
 {"Proof|ID:1812309492-1", 3609588, 100060},
 {"ID:595518277", 3609744, 100064},
 {"+ assoc \[ScriptCapitalP]", 3610016, 100071},
 {"Proof|ID:595518277", 3610281, 100078},
 {"Proof|ID:595518277-1", 3610443, 100082},
 {"ID:2114498624", 3610598, 100086},
 {"+ comm \[ScriptCapitalP]", 3610870, 100093},
 {"Proof|ID:2114498624", 3611137, 100100},
 {"Proof|ID:2114498624-1", 3611302, 100104},
 {"ID:246227628", 3611458, 100108},
 {"ID:95118700", 3611681, 100115},
 {"- inverse \[ScriptCapitalP] 1", 3611959, 100122},
 {"ID:1900465816", 3612221, 100129},
 {"- inverse \[ScriptCapitalP] 2", 3612502, 100136},
 {"ID:2037231548", 3612767, 100143},
 {"- inverse \[ScriptCapitalP]", 3613045, 100150},
 {"Proof|ID:95118700", 3613313, 100157},
 {"Proof|ID:95118700-1", 3613470, 100161},
 {"Proof|ID:1900465816", 3613627, 100165},
 {"Proof|ID:1900465816-1", 3613790, 100169},
 {"Proof|ID:2037231548", 3613951, 100173},
 {"Proof|ID:2037231548-1", 3614116, 100177},
 {"ID:145364866", 3614272, 100181},
 {"- 0 \[ScriptCapitalP] 1", 3614539, 100188},
 {"ID:117883331", 3614795, 100195},
 {"- 0 \[ScriptCapitalP] 2", 3615063, 100202},
 {"Proof|ID:145364866", 3615326, 100209},
 {"Proof|ID:145364866-1", 3615487, 100213},
 {"Proof|ID:117883331", 3615646, 100217},
 {"Proof|ID:117883331-1", 3615808, 100221},
 {"ID:771361293", 3615962, 100225},
 {"isCCMonoid \[ScriptCapitalP]", 3616240, 100232},
 {"Proof|ID:771361293", 3616508, 100239},
 {"Proof|ID:771361293-1", 3616669, 100243},
 {"ID:1221832846", 3616823, 100247},
 {"ID:313886646", 3617049, 100254},
 {"membership R", 3617300, 100261},
 {"ID:174448700", 3617551, 100268},
 {"lm+R", 3617780, 100275},
 {"ID:1679470623", 3618018, 100282},
 {"C R", 3618248, 100289},
 {"ID:1648446612", 3618488, 100296},
 {"R 1", 3618718, 100303},
 {"ID:1799775671", 3618958, 100310},
 {"R 2", 3619188, 100317},
 {"ID:1065304756", 3619428, 100324},
 {"R 6", 3619658, 100331},
 {"ID:1078436809", 3619898, 100338},
 {"R 7", 3620128, 100345},
 {"ID:735362349", 3620367, 100352},
 {"R 8", 3620595, 100359},
 {"ID:451768283", 3620832, 100366},
 {"R 8'", 3621062, 100373},
 {"ID:739591353", 3621300, 100380},
 {"C 5'", 3621530, 100387},
 {"ID:1347176402", 3621769, 100394},
 {"ID:1377108879", 3621996, 100401},
 {"R 3", 3622226, 100408},
 {"ID:139632107", 3622465, 100415},
 {"R 4", 3622693, 100422},
 {"ID:2132986120", 3622931, 100429},
 {"C 7", 3623161, 100436},
 {"ID:1911280389", 3623401, 100443},
 {"C 8", 3623631, 100450},
 {"Proof|ID:313886646", 3623876, 100457},
 {"Proof|ID:313886646-1", 3624037, 100461},
 {"Proof|ID:174448700", 3624196, 100465},
 {"Proof|ID:174448700-1", 3624358, 100469},
 {"Proof|ID:1679470623", 3624519, 100473},
 {"Proof|ID:1679470623-1", 3624685, 100477},
 {"Proof|ID:1648446612", 3624849, 100481},
 {"Proof|ID:1648446612-1", 3625013, 100485},
 {"Proof|ID:1799775671", 3625175, 100489},
 {"Proof|ID:1799775671-1", 3625341, 100493},
 {"Proof|ID:1065304756", 3625505, 100497},
 {"Proof|ID:1065304756-1", 3625669, 100501},
 {"Proof|ID:1078436809", 3625831, 100505},
 {"Proof|ID:1078436809-1", 3625997, 100509},
 {"Proof|ID:735362349", 3626160, 100513},
 {"Proof|ID:735362349-1", 3626321, 100517},
 {"Proof|ID:451768283", 3626480, 100521},
 {"Proof|ID:451768283-1", 3626641, 100525},
 {"Proof|ID:739591353", 3626800, 100529},
 {"Proof|ID:739591353-1", 3626963, 100533},
 {"Proof|ID:1377108879", 3627125, 100537},
 {"Proof|ID:1377108879-1", 3627290, 100541},
 {"Proof|ID:139632107", 3627452, 100545},
 {"Proof|ID:139632107-1", 3627615, 100549},
 {"Proof|ID:2132986120", 3627777, 100553},
 {"Proof|ID:2132986120-1", 3627943, 100557},
 {"Proof|ID:1911280389", 3628107, 100561},
 {"Proof|ID:1911280389-1", 3628272, 100565},
 {"ID:1031528829", 3628429, 100569},
 {"ID:1494855415", 3628656, 100576},
 {"ID:2007119699", 3628883, 100583},
 {"membership H ind", 3629139, 100590},
 {"ID:884488101", 3629391, 100597},
 {"membership H", 3629642, 100604},
 {"ID:1892666968", 3629894, 100611},
 {"C H ind", 3630132, 100618},
 {"ID:1556838587", 3630376, 100625},
 {"C H", 3630606, 100632},
 {"ID:1768827306", 3630846, 100639},
 {"H=0", 3631076, 100646},
 {"ID:140826492", 3631315, 100653},
 {"H=H", 3631543, 100660},
 {"ID:1627546251", 3631781, 100667},
 {"H 1", 3632011, 100674},
 {"ID:743145124", 3632250, 100681},
 {"H 2", 3632478, 100688},
 {"ID:1576432658", 3632716, 100695},
 {"H 3", 3632946, 100702},
 {"ID:1977357773", 3633186, 100709},
 {"H 4", 3633416, 100716},
 {"ID:561821259", 3633655, 100723},
 {"H 4'", 3633885, 100730},
 {"ID:866138121", 3634123, 100737},
 {"H 5", 3634351, 100744},
 {"ID:105044014", 3634588, 100751},
 {"H 5'", 3634818, 100758},
 {"ID:1971276224", 3635057, 100765},
 {"H 6", 3635287, 100772},
 {"ID:219245057", 3635526, 100779},
 {"ID:2135590264", 3635751, 100786},
 {"H 7", 3635980, 100793},
 {"ID:2092106183", 3636219, 100800},
 {"H 8", 3636449, 100807},
 {"ID:831355566", 3636688, 100814},
 {"ID:2014193763", 3636913, 100821},
 {"H 10", 3637145, 100828},
 {"ID:1650576647", 3637386, 100835},
 {"H 11", 3637618, 100842},
 {"Proof|ID:2007119699", 3637865, 100849},
 {"Proof|ID:2007119699-1", 3638030, 100853},
 {"Proof|ID:884488101", 3638192, 100857},
 {"Proof|ID:884488101-1", 3638354, 100861},
 {"Proof|ID:1892666968", 3638515, 100865},
 {"Proof|ID:1892666968-1", 3638680, 100869},
 {"Proof|ID:1556838587", 3638843, 100873},
 {"Proof|ID:1556838587-1", 3639008, 100877},
 {"Proof|ID:1768827306", 3639171, 100881},
 {"Proof|ID:1768827306-1", 3639336, 100885},
 {"Proof|ID:140826492", 3639498, 100889},
 {"Proof|ID:140826492-1", 3639660, 100893},
 {"Proof|ID:1627546251", 3639821, 100897},
 {"Proof|ID:1627546251-1", 3639987, 100901},
 {"Proof|ID:743145124", 3640150, 100905},
 {"Proof|ID:743145124-1", 3640312, 100909},
 {"Proof|ID:1576432658", 3640473, 100913},
 {"Proof|ID:1576432658-1", 3640638, 100917},
 {"Proof|ID:1977357773", 3640801, 100921},
 {"Proof|ID:1977357773-1", 3640966, 100925},
 {"Proof|ID:561821259", 3641128, 100929},
 {"Proof|ID:561821259-1", 3641291, 100933},
 {"Proof|ID:866138121", 3641452, 100937},
 {"Proof|ID:866138121-1", 3641614, 100941},
 {"Proof|ID:105044014", 3641774, 100945},
 {"Proof|ID:105044014-1", 3641936, 100949},
 {"Proof|ID:1971276224", 3642097, 100953},
 {"Proof|ID:1971276224-1", 3642262, 100957},
 {"Proof|ID:2135590264", 3642425, 100961},
 {"Proof|ID:2135590264-1", 3642588, 100965},
 {"Proof|ID:2092106183", 3642749, 100969},
 {"Proof|ID:2092106183-1", 3642914, 100973},
 {"Proof|ID:2014193763", 3643077, 100977},
 {"Proof|ID:2014193763-1", 3643243, 100981},
 {"Proof|ID:1650576647", 3643407, 100985},
 {"Proof|ID:1650576647-1", 3643573, 100989},
 {"ID:45249416", 3643729, 100993},
 {"ID:248026320", 3643951, 101000},
 {"induction on \[ScriptCapitalP] ind", 3644242, 101007},
 {"ID:1502331257", 3644512, 101014},
 {"induction on \[ScriptCapitalP]", 3644802, 101021},
 {"Proof|ID:248026320", 3645080, 101028},
 {"Proof|ID:248026320-1", 3645242, 101032},
 {"Proof|ID:1502331257", 3645403, 101036},
 {"Proof|ID:1502331257-1", 3645569, 101040},
 {"ID:411993095", 3645726, 101044},
 {"ID:470022203", 3645950, 101051},
 {"\[CenterDot] 1", 3646200, 101058},
 {"ID:401540586", 3646448, 101065},
 {"ID:2014025290", 3646673, 101072},
 {"isMonomial 6", 3646921, 101079},
 {"ID:1590906878", 3647170, 101086},
 {"isMonomial 10", 3647419, 101093},
 {"ID:1997927583", 3647668, 101100},
 {"isMonomial 11", 3647918, 101107},
 {"Proof|ID:470022203", 3648173, 101114},
 {"Proof|ID:470022203-1", 3648335, 101118},
 {"Proof|ID:2014025290", 3648496, 101122},
 {"Proof|ID:2014025290-1", 3648661, 101126},
 {"Proof|ID:1590906878", 3648824, 101130},
 {"Proof|ID:1590906878-1", 3648989, 101134},
 {"Proof|ID:1997927583", 3649152, 101138},
 {"Proof|ID:1997927583-1", 3649317, 101142},
 {"ID:63039386", 3649472, 101146},
 {"ID:1093453827", 3649695, 101153},
 {"\[CircleTimes] 0", 3649950, 101160},
 {"ID:1795167993", 3650202, 101167},
 {"\[CircleTimes] 5", 3650458, 101174},
 {"ID:838653570", 3650710, 101181},
 {"membership monomial multiplication ind", 3651008, 101188},
 {"ID:1903131381", 3651281, 101195},
 {"membership monomial multiplication", 3651578, 101202},
 {"ID:1641583144", 3651854, 101209},
 {"C \[CircleTimes]", 3652110, 101216},
 {"ID:608153969", 3652362, 101223},
 {"\[CircleTimes] 1", 3652616, 101230},
 {"ID:1439631766", 3652867, 101237},
 {"\[CircleTimes] isMonomial", 3653141, 101244},
 {"ID:607306392", 3653402, 101251},
 {"ID:1312761442", 3653627, 101258},
 {"\[CircleTimes] 2", 3653883, 101265},
 {"ID:259600169", 3654135, 101272},
 {"\[CircleTimes] 3", 3654389, 101279},
 {"ID:1576857420", 3654640, 101286},
 {"\[CircleTimes] 6", 3654895, 101293},
 {"ID:483781382", 3655146, 101300},
 {"\[CircleTimes] 10", 3655402, 101307},
 {"ID:169057028", 3655653, 101314},
 {"\[CircleTimes] 11", 3655909, 101321},
 {"ID:324898493", 3656160, 101328},
 {"ID:684947841", 3656384, 101335},
 {"\[CircleTimes] 7", 3656638, 101342},
 {"ID:1856294157", 3656889, 101349},
 {"\[CircleTimes] 8", 3657145, 101356},
 {"ID:1768460531", 3657398, 101363},
 {"\[CircleTimes] 9", 3657654, 101370},
 {"Proof|ID:1093453827", 3657913, 101377},
 {"Proof|ID:1093453827-1", 3658078, 101381},
 {"Proof|ID:1795167993", 3658241, 101385},
 {"Proof|ID:1795167993-1", 3658407, 101389},
 {"Proof|ID:838653570", 3658570, 101393},
 {"Proof|ID:838653570-1", 3658731, 101397},
 {"Proof|ID:1903131381", 3658891, 101401},
 {"Proof|ID:1903131381-1", 3659055, 101405},
 {"Proof|ID:1641583144", 3659217, 101409},
 {"Proof|ID:1641583144-1", 3659382, 101413},
 {"Proof|ID:608153969", 3659544, 101417},
 {"Proof|ID:608153969-1", 3659706, 101421},
 {"Proof|ID:1439631766", 3659867, 101425},
 {"Proof|ID:1439631766-1", 3660033, 101429},
 {"Proof|ID:1312761442", 3660197, 101433},
 {"Proof|ID:1312761442-1", 3660364, 101437},
 {"Proof|ID:259600169", 3660528, 101441},
 {"Proof|ID:259600169-1", 3660691, 101445},
 {"Proof|ID:1576857420", 3660853, 101449},
 {"Proof|ID:1576857420-1", 3661020, 101453},
 {"Proof|ID:483781382", 3661184, 101457},
 {"Proof|ID:483781382-1", 3661347, 101461},
 {"Proof|ID:169057028", 3661508, 101465},
 {"Proof|ID:169057028-1", 3661672, 101469},
 {"Proof|ID:684947841", 3661834, 101473},
 {"Proof|ID:684947841-1", 3661997, 101477},
 {"Proof|ID:1856294157", 3662159, 101481},
 {"Proof|ID:1856294157-1", 3662326, 101485},
 {"Proof|ID:1768460531", 3662491, 101489},
 {"Proof|ID:1768460531-1", 3662658, 101493},
 {"ID:1437523678", 3662817, 101497},
 {"hasMultiplicationRec", 3663083, 101504},
 {"ID:1372918741", 3663342, 101511},
 {"ID:271668948", 3663569, 101518},
 {"316", 3663798, 101525},
 {"ID:428541408", 3664036, 101532},
 {"317", 3664265, 101539},
 {"ID:1983432019", 3664504, 101546},
 {"318", 3664734, 101553},
 {"Proof|ID:271668948", 3664979, 101560},
 {"Proof|ID:271668948-1", 3665143, 101564},
 {"Proof|ID:428541408", 3665305, 101568},
 {"Proof|ID:428541408-1", 3665468, 101572},
 {"Proof|ID:1983432019", 3665630, 101576},
 {"Proof|ID:1983432019-1", 3665795, 101580},
 {"ID:1310550300", 3665952, 101584},
 {"319", 3666183, 101591},
 {"ID:978297022", 3666423, 101598},
 {"hasMultiplication \[ScriptCapitalP]", 3666715, 101605},
 {"ID:1125626334", 3666985, 101612},
 {"320", 3667216, 101619},
 {"Proof|ID:1125626334", 3667463, 101626},
 {"Proof|ID:1125626334-1", 3667628, 101630},
 {"ID:1103171586", 3667785, 101634},
 {"\[RawStar] closed \[ScriptCapitalP]", 3668080, 101641},
 {"Proof|ID:1103171586", 3668359, 101648},
 {"Proof|ID:1103171586-1", 3668524, 101652},
 {"ID:2097953227", 3668681, 101656},
 {"322", 3668912, 101663},
 {"Proof|ID:2097953227", 3669159, 101670},
 {"Proof|ID:2097953227-1", 3669325, 101674},
 {"ID:587826938", 3669482, 101678},
 {"ID:384020988", 3669707, 101685},
 {"324", 3669935, 101692},
 {"ID:205407574", 3670172, 101699},
 {"325", 3670401, 101706},
 {"ID:411846169", 3670639, 101713},
 {"327", 3670868, 101720},
 {"ID:1300778952", 3671107, 101727},
 {"328", 3671338, 101734},
 {"ID:2123221051", 3671579, 101741},
 {"329", 3671810, 101748},
 {"Proof|ID:384020988", 3672056, 101755},
 {"Proof|ID:384020988-1", 3672218, 101759},
 {"Proof|ID:205407574", 3672378, 101763},
 {"Proof|ID:205407574-1", 3672541, 101767},
 {"Proof|ID:411846169", 3672702, 101771},
 {"Proof|ID:411846169-1", 3672864, 101775},
 {"Proof|ID:1300778952", 3673025, 101779},
 {"Proof|ID:1300778952-1", 3673190, 101783},
 {"Proof|ID:2123221051", 3673353, 101787},
 {"Proof|ID:2123221051-1", 3673518, 101791},
 {"ID:840093990", 3673674, 101795},
 {"\[RawStar] comm \[ScriptCapitalP]", 3673963, 101802},
 {"Proof|ID:840093990", 3674237, 101809},
 {"Proof|ID:840093990-1", 3674397, 101813},
 {"ID:1730144032", 3674550, 101817},
 {"\[RawStar] distr \[ScriptCapitalP]", 3674843, 101824},
 {"Proof|ID:1730144032", 3675121, 101831},
 {"Proof|ID:1730144032-1", 3675287, 101835},
 {"ID:378610433", 3675444, 101839},
 {"1 neutral \[ScriptCapitalP]", 3675721, 101846},
 {"Proof|ID:378610433", 3675989, 101853},
 {"Proof|ID:378610433-1", 3676151, 101857},
 {"ID:687899447", 3676305, 101861},
 {"ID:421376854", 3676530, 101868},
 {"330", 3676759, 101875},
 {"ID:935511285", 3676997, 101882},
 {"331", 3677226, 101889},
 {"ID:1702504479", 3677465, 101896},
 {"332", 3677696, 101903},
 {"ID:1379089132", 3677937, 101910},
 {"333", 3678168, 101917},
 {"Proof|ID:421376854", 3678414, 101924},
 {"Proof|ID:421376854-1", 3678578, 101928},
 {"Proof|ID:935511285", 3678740, 101932},
 {"Proof|ID:935511285-1", 3678904, 101936},
 {"Proof|ID:1702504479", 3679067, 101940},
 {"Proof|ID:1702504479-1", 3679234, 101944},
 {"Proof|ID:1379089132", 3679399, 101948},
 {"Proof|ID:1379089132-1", 3679566, 101952},
 {"ID:167864408", 3679724, 101956},
 {"\[RawStar] assoc \[ScriptCapitalP]", 3680015, 101963},
 {"Proof|ID:167864408", 3680290, 101970},
 {"Proof|ID:167864408-1", 3680453, 101974},
 {"ID:1082265416", 3680609, 101978},
 {"isCommRing1 \[ScriptCapitalP]", 3680891, 101985},
 {"Proof|ID:1082265416", 3681163, 101992},
 {"Proof|ID:1082265416-1", 3681328, 101996},
 {"ID:1610244238", 3681485, 102000},
 {"ID:1259587933", 3681714, 102007},
 {"R 5", 3681945, 102014},
 {"ID:859675419", 3682185, 102021},
 {"C 6 ind", 3682422, 102028},
 {"ID:515125163", 3682664, 102035},
 {"C 6", 3682893, 102042},
 {"ID:1572844749", 3683132, 102049},
 {"C 6'", 3683365, 102056},
 {"ID:771023286", 3683606, 102063},
 {"lp 4", 3683837, 102070},
 {"ID:422533310", 3684076, 102077},
 {"H 9", 3684305, 102084},
 {"Proof|ID:1259587933", 3684550, 102091},
 {"Proof|ID:1259587933-1", 3684717, 102095},
 {"Proof|ID:859675419", 3684881, 102099},
 {"Proof|ID:859675419-1", 3685044, 102103},
 {"Proof|ID:515125163", 3685205, 102107},
 {"Proof|ID:515125163-1", 3685367, 102111},
 {"Proof|ID:1572844749", 3685528, 102115},
 {"Proof|ID:1572844749-1", 3685695, 102119},
 {"Proof|ID:771023286", 3685859, 102123},
 {"Proof|ID:771023286-1", 3686021, 102127},
 {"Proof|ID:422533310", 3686181, 102131},
 {"Proof|ID:422533310-1", 3686344, 102135},
 {"ID:1667882379", 3686500, 102139},
 {"ID:1105090112", 3686728, 102146},
 {"ID:878771843", 3686955, 102153},
 {"\[CenterDot] 2", 3687206, 102160},
 {"ID:1863009930", 3687456, 102167},
 {"ID:1958431161", 3687684, 102174},
 {"\[CenterDot] 5", 3687937, 102181},
 {"ID:758577094", 3688188, 102188},
 {"\[CenterDot] 6", 3688439, 102195},
 {"ID:544874520", 3688688, 102202},
 {"ID:158305602", 3688913, 102209},
 {"ID:347486225", 3689138, 102216},
 {"\[CenterDot] 9", 3689389, 102223},
 {"ID:1591957532", 3689639, 102230},
 {"\[CenterDot] 10", 3689894, 102237},
 {"ID:1155967768", 3690147, 102244},
 {"ID:1942982211", 3690375, 102251},
 {"\[CenterDot] 15", 3690630, 102258},
 {"ID:44857450", 3690881, 102265},
 {"\[CenterDot] 12", 3691132, 102272},
 {"ID:928737230", 3691380, 102279},
 {"\[CenterDot] 13", 3691633, 102286},
 {"ID:233618042", 3691883, 102293},
 {"\[CenterDot] 14", 3692136, 102300},
 {"Proof|ID:878771843", 3692392, 102307},
 {"Proof|ID:878771843-1", 3692555, 102311},
 {"Proof|ID:1958431161", 3692717, 102315},
 {"Proof|ID:1958431161-1", 3692883, 102319},
 {"Proof|ID:758577094", 3693046, 102323},
 {"Proof|ID:758577094-1", 3693210, 102327},
 {"Proof|ID:347486225", 3693372, 102331},
 {"Proof|ID:347486225-1", 3693536, 102335},
 {"Proof|ID:1591957532", 3693699, 102339},
 {"Proof|ID:1591957532-1", 3693866, 102343},
 {"Proof|ID:1942982211", 3694031, 102347},
 {"Proof|ID:1942982211-1", 3694198, 102351},
 {"Proof|ID:44857450", 3694361, 102355},
 {"Proof|ID:44857450-1", 3694521, 102359},
 {"Proof|ID:928737230", 3694680, 102363},
 {"Proof|ID:928737230-1", 3694843, 102367},
 {"Proof|ID:233618042", 3695004, 102371},
 {"Proof|ID:233618042-1", 3695167, 102375},
 {"ID:183428126", 3695322, 102379},
 {"ID:2026663200", 3695548, 102386},
 {"isMonomial 2", 3695797, 102393},
 {"ID:1916930407", 3696047, 102400},
 {"isMonomial 5", 3696296, 102407},
 {"Proof|ID:2026663200", 3696552, 102414},
 {"Proof|ID:2026663200-1", 3696719, 102418},
 {"Proof|ID:1916930407", 3696884, 102422},
 {"Proof|ID:1916930407-1", 3697051, 102426},
 {"ID:1225634217", 3697210, 102430},
 {"reduction ring \[ScriptCapitalR]", 3697498, 102437},
 {"ID:1542801947", 3697767, 102444},
 {"300", 3697998, 102451},
 {"Proof|ID:1542801947", 3698245, 102458},
 {"Proof|ID:1542801947-1", 3698412, 102462},
 {"ID:946287998", 3698570, 102466},
 {"isOrderedRec", 3698819, 102473},
 {"ID:1485445352", 3699069, 102480},
 {"ID:215225048", 3699296, 102487},
 {"301", 3699525, 102494},
 {"ID:1704612967", 3699764, 102501},
 {"302", 3699995, 102508},
 {"ID:162248449", 3700235, 102515},
 {"303", 3700463, 102522},
 {"Proof|ID:215225048", 3700706, 102529},
 {"Proof|ID:215225048-1", 3700870, 102533},
 {"Proof|ID:1704612967", 3701033, 102537},
 {"Proof|ID:1704612967-1", 3701199, 102541},
 {"Proof|ID:162248449", 3701362, 102545},
 {"Proof|ID:162248449-1", 3701524, 102549},
 {"ID:1015468103", 3701679, 102553},
 {"304", 3701910, 102560},
 {"ID:7476052", 3702148, 102567},
 {"reduction-polynomial domain \[ScriptCapitalP]", 3702456, 102574},
 {"ID:173226898", 3702731, 102581},
 {"ID:835223349", 3702956, 102588},
 {"ID:181219027", 3703181, 102595},
 {"356", 3703409, 102602},
 {"ID:271793724", 3703646, 102609},
 {"382", 3703875, 102616},
 {"ID:575028756", 3704113, 102623},
 {"382'", 3704344, 102630},
 {"ID:2020716880", 3704584, 102637},
 {"383", 3704815, 102644},
 {"ID:1964835443", 3705056, 102651},
 {"381", 3705287, 102658},
 {"ID:1599847484", 3705528, 102665},
 {"380", 3705759, 102672},
 {"ID:590709100", 3705999, 102679},
 {"379", 3706230, 102686},
 {"ID:586184654", 3706470, 102693},
 {"405", 3706699, 102700},
 {"ID:919284948", 3706937, 102707},
 {"406", 3707166, 102714},
 {"Proof|ID:181219027", 3707410, 102721},
 {"Proof|ID:181219027-1", 3707572, 102725},
 {"Proof|ID:271793724", 3707732, 102729},
 {"Proof|ID:271793724-1", 3707896, 102733},
 {"Proof|ID:575028756", 3708058, 102737},
 {"Proof|ID:575028756-1", 3708221, 102741},
 {"Proof|ID:2020716880", 3708383, 102745},
 {"Proof|ID:2020716880-1", 3708550, 102749},
 {"Proof|ID:1964835443", 3708715, 102753},
 {"Proof|ID:1964835443-1", 3708881, 102757},
 {"Proof|ID:1599847484", 3709045, 102761},
 {"Proof|ID:1599847484-1", 3709211, 102765},
 {"Proof|ID:590709100", 3709374, 102769},
 {"Proof|ID:590709100-1", 3709537, 102773},
 {"Proof|ID:586184654", 3709698, 102777},
 {"Proof|ID:586184654-1", 3709862, 102781},
 {"Proof|ID:919284948", 3710024, 102785},
 {"Proof|ID:919284948-1", 3710187, 102789},
 {"ID:895736483", 3710342, 102793},
 {"ID:1904443546", 3710568, 102800},
 {"ID:1587558771", 3710796, 102807},
 {"ID:530327018", 3711023, 102814},
 {"377", 3711252, 102821},
 {"ID:2044424156", 3711491, 102828},
 {"377'", 3711724, 102835},
 {"ID:14600367", 3711964, 102842},
 {"377''", 3712195, 102849},
 {"ID:1313257457", 3712434, 102856},
 {"\[CenterDot] 7", 3712687, 102863},
 {"ID:1776891622", 3712939, 102870},
 {"\[CenterDot] 8", 3713192, 102877},
 {"Proof|ID:530327018", 3713449, 102884},
 {"Proof|ID:530327018-1", 3713611, 102888},
 {"Proof|ID:2044424156", 3713772, 102892},
 {"Proof|ID:2044424156-1", 3713937, 102896},
 {"Proof|ID:14600367", 3714098, 102900},
 {"Proof|ID:14600367-1", 3714258, 102904},
 {"Proof|ID:1313257457", 3714418, 102908},
 {"Proof|ID:1313257457-1", 3714585, 102912},
 {"Proof|ID:1776891622", 3714750, 102916},
 {"Proof|ID:1776891622-1", 3714915, 102920},
 {"ID:1077742446", 3715072, 102924},
 {"ID:1717822499", 3715301, 102931},
 {"384", 3715532, 102938},
 {"ID:1045770541", 3715773, 102945},
 {"358", 3716003, 102952},
 {"ID:312784637", 3716242, 102959},
 {"393", 3716471, 102966},
 {"ID:172690263", 3716709, 102973},
 {"ID:174940712", 3716934, 102980},
 {"mult monomial", 3717183, 102987},
 {"ID:1850961673", 3717432, 102994},
 {"mult monomial II", 3717689, 103001},
 {"ID:849856225", 3717942, 103008},
 {"mult monomial III", 3718199, 103015},
 {"Proof|ID:1717822499", 3718458, 103022},
 {"Proof|ID:1717822499-1", 3718625, 103026},
 {"Proof|ID:1045770541", 3718790, 103030},
 {"Proof|ID:1045770541-1", 3718956, 103034},
 {"Proof|ID:312784637", 3719119, 103038},
 {"Proof|ID:312784637-1", 3719283, 103042},
 {"Proof|ID:174940712", 3719445, 103046},
 {"Proof|ID:174940712-1", 3719609, 103050},
 {"Proof|ID:1850961673", 3719772, 103054},
 {"Proof|ID:1850961673-1", 3719939, 103058},
 {"Proof|ID:849856225", 3720103, 103062},
 {"Proof|ID:849856225-1", 3720266, 103066},
 {"ID:633254127", 3720421, 103070},
 {"ID:1581566919", 3720648, 103077},
 {"ID:1230351665", 3720877, 103084},
 {"tuples \[CenterDot] 1", 3721143, 103091},
 {"ID:1294700753", 3721401, 103098},
 {"ID:168346017", 3721628, 103105},
 {"tuples \[CenterDot] 2", 3721893, 103112},
 {"ID:1208795229", 3722150, 103119},
 {"tuples \[CenterDot] 3", 3722417, 103126},
 {"ID:369110158", 3722675, 103133},
 {"tuples \[CenterDot] 4", 3722940, 103140},
 {"Proof|ID:1230351665", 3723203, 103147},
 {"Proof|ID:1230351665-1", 3723369, 103151},
 {"Proof|ID:168346017", 3723532, 103155},
 {"Proof|ID:168346017-1", 3723696, 103159},
 {"Proof|ID:1208795229", 3723859, 103163},
 {"Proof|ID:1208795229-1", 3724026, 103167},
 {"Proof|ID:369110158", 3724190, 103171},
 {"Proof|ID:369110158-1", 3724354, 103175},
 {"ID:155708286", 3724510, 103179},
 {"ID:280312466", 3724735, 103186},
 {"4", 3724959, 103193},
 {"ID:713919982", 3725194, 103200},
 {"5", 3725419, 103207},
 {"ID:1630947645", 3725656, 103214},
 {"6", 3725883, 103221},
 {"Proof|ID:280312466", 3726127, 103228},
 {"Proof|ID:280312466-1", 3726287, 103232},
 {"Proof|ID:713919982", 3726445, 103236},
 {"Proof|ID:713919982-1", 3726604, 103240},
 {"Proof|ID:1630947645", 3726762, 103244},
 {"Proof|ID:1630947645-1", 3726928, 103248},
 {"ID:795795393", 3727085, 103252},
 {"360", 3727314, 103259},
 {"Proof|ID:795795393", 3727558, 103266},
 {"Proof|ID:795795393-1", 3727721, 103270},
 {"ID:1302712453", 3727877, 103274},
 {"ID:1092343128", 3728105, 103281},
 {"1", 3728333, 103288},
 {"ID:311837531", 3728572, 103295},
 {"2", 3728797, 103302},
 {"ID:2036982330", 3729034, 103309},
 {"3", 3729261, 103316},
 {"ID:917853466", 3729499, 103323},
 {"378", 3729728, 103330},
 {"Proof|ID:1092343128", 3729973, 103337},
 {"Proof|ID:1092343128-1", 3730140, 103341},
 {"Proof|ID:311837531", 3730304, 103345},
 {"Proof|ID:311837531-1", 3730466, 103349},
 {"Proof|ID:2036982330", 3730627, 103353},
 {"Proof|ID:2036982330-1", 3730792, 103357},
 {"Proof|ID:917853466", 3730954, 103361},
 {"Proof|ID:917853466-1", 3731117, 103365},
 {"ID:1009043994", 3731273, 103369},
 {"361", 3731504, 103376},
 {"Proof|ID:1009043994", 3731751, 103383},
 {"Proof|ID:1009043994-1", 3731916, 103387},
 {"ID:412087985", 3732072, 103391},
 {"362", 3732300, 103398},
 {"Proof|ID:412087985", 3732543, 103405},
 {"Proof|ID:412087985-1", 3732705, 103409},
 {"ID:1267059190", 3732860, 103413},
 {"363", 3733090, 103420},
 {"Proof|ID:1267059190", 3733336, 103427},
 {"Proof|ID:1267059190-1", 3733502, 103431},
 {"ID:378652997", 3733659, 103435},
 {"364", 3733887, 103442},
 {"Proof|ID:378652997", 3734130, 103449},
 {"Proof|ID:378652997-1", 3734293, 103453},
 {"ID:1036416147", 3734449, 103457},
 {"365", 3734679, 103464},
 {"Proof|ID:1036416147", 3734925, 103471},
 {"Proof|ID:1036416147-1", 3735090, 103475},
 {"ID:1341307829", 3735247, 103479},
 {"386", 3735478, 103486},
 {"ID:1193326225", 3735719, 103493},
 {"385", 3735950, 103500},
 {"Proof|ID:1341307829", 3736197, 103507},
 {"Proof|ID:1341307829-1", 3736363, 103511},
 {"Proof|ID:1193326225", 3736527, 103515},
 {"Proof|ID:1193326225-1", 3736693, 103519},
 {"ID:2033377127", 3736851, 103523},
 {"366", 3737081, 103530},
 {"Proof|ID:2033377127", 3737327, 103537},
 {"Proof|ID:2033377127-1", 3737493, 103541},
 {"ID:982813600", 3737650, 103545},
 {"367", 3737878, 103552},
 {"Proof|ID:982813600", 3738121, 103559},
 {"Proof|ID:982813600-1", 3738284, 103563},
 {"ID:1179937112", 3738440, 103567},
 {"ID:134072435", 3738667, 103574},
 {"336", 3738896, 103581},
 {"ID:1484092370", 3739135, 103588},
 {"337", 3739366, 103595},
 {"Proof|ID:134072435", 3739612, 103602},
 {"Proof|ID:134072435-1", 3739774, 103606},
 {"Proof|ID:1484092370", 3739935, 103610},
 {"Proof|ID:1484092370-1", 3740101, 103614},
 {"ID:1480600758", 3740259, 103618},
 {"ID:1004003660", 3740488, 103625},
 {"ID:1854976476", 3740716, 103632},
 {"ID:408134090", 3740944, 103639},
 {"reduceM \[ScriptCapitalP]", 3741217, 103646},
 {"ID:1088018783", 3741478, 103653},
 {"reduceM \[ScriptCapitalP] II", 3741759, 103660},
 {"Proof|ID:408134090", 3742030, 103667},
 {"Proof|ID:408134090-1", 3742192, 103671},
 {"Proof|ID:1088018783", 3742353, 103675},
 {"Proof|ID:1088018783-1", 3742520, 103679},
 {"ID:1964871967", 3742679, 103683},
 {"ID:1041722017", 3742907, 103690},
 {"407", 3743138, 103697},
 {"ID:1518222527", 3743379, 103704},
 {"ID:1070088161", 3743607, 103711},
 {"408", 3743838, 103718},
 {"ID:1970014385", 3744079, 103725},
 {"408'", 3744312, 103732},
 {"Proof|ID:1041722017", 3744560, 103739},
 {"Proof|ID:1041722017-1", 3744726, 103743},
 {"Proof|ID:1070088161", 3744890, 103747},
 {"Proof|ID:1070088161-1", 3745057, 103751},
 {"Proof|ID:1970014385", 3745222, 103755},
 {"Proof|ID:1970014385-1", 3745388, 103759},
 {"ID:102721391", 3745545, 103763},
 {"ID:657593834", 3745770, 103770},
 {"RED", 3746000, 103777},
 {"ID:1802485517", 3746240, 103784},
 {"RED II", 3746477, 103791},
 {"ID:1164884865", 3746721, 103798},
 {"RED III", 3746960, 103805},
 {"ID:958945878", 3747204, 103812},
 {"RED IV", 3747439, 103819},
 {"ID:1601029605", 3747681, 103826},
 {"ID:687002895", 3747908, 103833},
 {"RED monomial", 3748156, 103840},
 {"ID:413658239", 3748404, 103847},
 {"RED monomial II", 3748657, 103854},
 {"Proof|ID:657593834", 3748913, 103861},
 {"Proof|ID:657593834-1", 3749075, 103865},
 {"Proof|ID:1802485517", 3749236, 103869},
 {"Proof|ID:1802485517-1", 3749401, 103873},
 {"Proof|ID:1164884865", 3749564, 103877},
 {"Proof|ID:1164884865-1", 3749731, 103881},
 {"Proof|ID:958945878", 3749895, 103885},
 {"Proof|ID:958945878-1", 3750059, 103889},
 {"Proof|ID:687002895", 3750221, 103893},
 {"Proof|ID:687002895-1", 3750384, 103897},
 {"Proof|ID:413658239", 3750545, 103901},
 {"Proof|ID:413658239-1", 3750708, 103905},
 {"ID:580194792", 3750863, 103909},
 {"ID:1667457569", 3751089, 103916},
 {"357", 3751320, 103923},
 {"ID:2035027156", 3751561, 103930},
 {"\[RightArrow] lm \[DoubleRightArrow]", 3751858, 103937},
 {"ID:2026457432", 3752132, 103944},
 {"\[RightArrow] lm \[DoubleLeftArrow]", 3752427, 103951},
 {"ID:1167741853", 3752700, 103958},
 {"reducibility \[ScriptCapitalP]", 3752985, 103965},
 {"ID:1998060487", 3753253, 103972},
 {"\[RightArrow] alternative \[ScriptCapitalP]", 3753566, 103979},
 {"Proof|ID:1667457569", 3753855, 103986},
 {"Proof|ID:1667457569-1", 3754020, 103990},
 {"Proof|ID:2035027156", 3754183, 103994},
 {"Proof|ID:2035027156-1", 3754349, 103998},
 {"Proof|ID:2026457432", 3754513, 104002},
 {"Proof|ID:2026457432-1", 3754680, 104006},
 {"Proof|ID:1167741853", 3754845, 104010},
 {"Proof|ID:1167741853-1", 3755011, 104014},
 {"Proof|ID:1998060487", 3755175, 104018},
 {"Proof|ID:1998060487-1", 3755342, 104022},
 {"ID:1709390289", 3755501, 104026},
 {"ID:1713414293", 3755729, 104033},
 {"\[LeftRightArrow] alternative \[ScriptCapitalP]", 3756050, 104040},
 {"Proof|ID:1713414293", 3756343, 104047},
 {"Proof|ID:1713414293-1", 3756510, 104051},
 {"ID:1190720464", 3756669, 104055},
 {"isReducibleRec \[ScriptCapitalP]", 3756959, 104062},
 {"ID:166529906", 3757229, 104069},
 {"ID:844782115", 3757454, 104076},
 {"352.\[DoubleRightArrow]", 3757723, 104083},
 {"ID:953126445", 3757981, 104090},
 {"352.\[DoubleLeftArrow]", 3758248, 104097},
 {"ID:259818613", 3758505, 104104},
 {"352", 3758734, 104111},
 {"ID:182112345", 3758972, 104118},
 {"353", 3759201, 104125},
 {"Proof|ID:844782115", 3759445, 104132},
 {"Proof|ID:844782115-1", 3759609, 104136},
 {"Proof|ID:953126445", 3759771, 104140},
 {"Proof|ID:953126445-1", 3759935, 104144},
 {"Proof|ID:259818613", 3760097, 104148},
 {"Proof|ID:259818613-1", 3760260, 104152},
 {"Proof|ID:182112345", 3760421, 104156},
 {"Proof|ID:182112345-1", 3760584, 104160},
 {"ID:418362312", 3760739, 104164},
 {"ID:2116744492", 3760966, 104171},
 {"ID:1746803457", 3761194, 104178},
 {"391", 3761425, 104185},
 {"ID:899106457", 3761665, 104192},
 {"ID:1157329004", 3761891, 104199},
 {"ID:1927781160", 3762120, 104206},
 {"387", 3762351, 104213},
 {"Proof|ID:1746803457", 3762598, 104220},
 {"Proof|ID:1746803457-1", 3762764, 104224},
 {"Proof|ID:1927781160", 3762928, 104228},
 {"Proof|ID:1927781160-1", 3763095, 104232},
 {"ID:934500559", 3763253, 104236},
 {"368", 3763481, 104243},
 {"Proof|ID:934500559", 3763724, 104250},
 {"Proof|ID:934500559-1", 3763885, 104254},
 {"ID:371274476", 3764038, 104258},
 {"ID:1588964752", 3764265, 104265},
 {"IRR", 3764496, 104272},
 {"ID:428389793", 3764736, 104279},
 {"IRR II", 3764971, 104286},
 {"ID:225111715", 3765212, 104293},
 {"CORR", 3765443, 104300},
 {"ID:1632881403", 3765683, 104307},
 {"ID:1701548086", 3765911, 104314},
 {"IRR \[CenterDot]", 3766168, 104321},
 {"ID:1758251641", 3766422, 104328},
 {"CORR \[CenterDot]", 3766681, 104335},
 {"Proof|ID:1588964752", 3766942, 104342},
 {"Proof|ID:1588964752-1", 3767108, 104346},
 {"Proof|ID:428389793", 3767271, 104350},
 {"Proof|ID:428389793-1", 3767434, 104354},
 {"Proof|ID:225111715", 3767595, 104358},
 {"Proof|ID:225111715-1", 3767758, 104362},
 {"Proof|ID:1701548086", 3767920, 104366},
 {"Proof|ID:1701548086-1", 3768086, 104370},
 {"Proof|ID:1758251641", 3768250, 104374},
 {"Proof|ID:1758251641-1", 3768417, 104378},
 {"ID:1928324287", 3768576, 104382},
 {"ID:751250157", 3768803, 104389},
 {"404", 3769032, 104396},
 {"ID:1952511245", 3769271, 104403},
 {"404 II", 3769508, 104410},
 {"ID:537853020", 3769751, 104417},
 {"404 III", 3769988, 104424},
 {"Proof|ID:751250157", 3770236, 104431},
 {"Proof|ID:751250157-1", 3770400, 104435},
 {"Proof|ID:1952511245", 3770563, 104439},
 {"Proof|ID:1952511245-1", 3770728, 104443},
 {"Proof|ID:537853020", 3770890, 104447},
 {"Proof|ID:537853020-1", 3771054, 104451},
 {"ID:659046058", 3771210, 104455},
 {"ID:319520004", 3771435, 104462},
 {"ID:967033090", 3771660, 104469},
 {"ID:1647700871", 3771886, 104476},
 {"ID:1981597108", 3772115, 104483},
 {"ID:40011060", 3772341, 104490},
 {"388", 3772568, 104497},
 {"ID:884639747", 3772804, 104504},
 {"ID:1360655817", 3773030, 104511},
 {"ID:1002690147", 3773259, 104518},
 {"389", 3773490, 104525},
 {"ID:1896907436", 3773731, 104532},
 {"390", 3773962, 104539},
 {"Proof|ID:40011060", 3774207, 104546},
 {"Proof|ID:40011060-1", 3774369, 104550},
 {"Proof|ID:1002690147", 3774531, 104554},
 {"Proof|ID:1002690147-1", 3774697, 104558},
 {"Proof|ID:1896907436", 3774861, 104562},
 {"Proof|ID:1896907436-1", 3775027, 104566},
 {"ID:1290694758", 3775185, 104570},
 {"369", 3775415, 104577},
 {"Proof|ID:1290694758", 3775661, 104584},
 {"Proof|ID:1290694758-1", 3775826, 104588},
 {"ID:1785082178", 3775983, 104592},
 {"ID:1718264802", 3776212, 104599},
 {"ID:752304273", 3776439, 104606},
 {"ID:578561747", 3776664, 104613},
 {"areCommonReducibles I", 3776930, 104620},
 {"ID:1399766637", 3777188, 104627},
 {"areCommonReducibles II", 3777457, 104634},
 {"ID:1132709394", 3777717, 104641},
 {"isNTCRP", 3777957, 104648},
 {"ID:176358731", 3778202, 104655},
 {"ID:391507392", 3778428, 104662},
 {"ID:1030826183", 3778654, 104669},
 {"ID:926282876", 3778881, 104676},
 {"354", 3779110, 104683},
 {"ID:80931797", 3779347, 104690},
 {"355", 3779574, 104697},
 {"ID:414980491", 3779810, 104704},
 {"359", 3780039, 104711},
 {"ID:744723554", 3780277, 104718},
 {"345", 3780505, 104725},
 {"Proof|ID:926282876", 3780748, 104732},
 {"Proof|ID:926282876-1", 3780911, 104736},
 {"Proof|ID:80931797", 3781071, 104740},
 {"Proof|ID:80931797-1", 3781232, 104744},
 {"Proof|ID:414980491", 3781392, 104748},
 {"Proof|ID:414980491-1", 3781555, 104752},
 {"Proof|ID:744723554", 3781716, 104756},
 {"Proof|ID:744723554-1", 3781880, 104760},
 {"ID:381508689", 3782036, 104764},
 {"ID:1050210793", 3782263, 104771},
 {"ID:512389161", 3782490, 104778},
 {"ID:719916043", 3782715, 104785},
 {"isNTCRP alternative 1", 3782980, 104792},
 {"ID:864384559", 3783236, 104799},
 {"isNTCRP alternative 2", 3783501, 104806},
 {"ID:951944988", 3783757, 104813},
 {"isNTCRP alternative", 3784018, 104820},
 {"Proof|ID:719916043", 3784278, 104827},
 {"Proof|ID:719916043-1", 3784442, 104831},
 {"Proof|ID:864384559", 3784604, 104835},
 {"Proof|ID:864384559-1", 3784768, 104839},
 {"Proof|ID:951944988", 3784930, 104843},
 {"Proof|ID:951944988-1", 3785094, 104847},
 {"ID:1990838743", 3785251, 104851},
 {"ID:20910152", 3785478, 104858},
 {"ID:416591219", 3785701, 104865},
 {"CR1.\[DoubleRightArrow].1", 3785975, 104872},
 {"ID:295118406", 3786236, 104879},
 {"CR1.\[DoubleRightArrow].2", 3786509, 104886},
 {"ID:1583101849", 3786770, 104893},
 {"CR1.\[DoubleLeftArrow]", 3787041, 104900},
 {"Proof|ID:416591219", 3787308, 104907},
 {"Proof|ID:416591219-1", 3787473, 104911},
 {"Proof|ID:295118406", 3787636, 104915},
 {"Proof|ID:295118406-1", 3787800, 104919},
 {"Proof|ID:1583101849", 3787963, 104923},
 {"Proof|ID:1583101849-1", 3788131, 104927},
 {"ID:1272506214", 3788291, 104931},
 {"ID:386901098", 3788519, 104938},
 {"CR2.\[DoubleRightArrow].1", 3788793, 104945},
 {"ID:670300270", 3789054, 104952},
 {"CR2.\[DoubleRightArrow].2", 3789327, 104959},
 {"ID:1434882124", 3789588, 104966},
 {"CR2.\[DoubleLeftArrow]", 3789859, 104973},
 {"Proof|ID:386901098", 3790126, 104980},
 {"Proof|ID:386901098-1", 3790288, 104984},
 {"Proof|ID:670300270", 3790448, 104988},
 {"Proof|ID:670300270-1", 3790611, 104992},
 {"Proof|ID:1434882124", 3790773, 104996},
 {"Proof|ID:1434882124-1", 3790940, 105000},
 {"ID:737627449", 3791098, 105004},
 {"ID:81313572", 3791323, 105011},
 {"ID:1098402572", 3791547, 105018},
 {"ID:1499935354", 3791775, 105025},
 {"CR1 monomial 1", 3792029, 105032},
 {"ID:1938671644", 3792282, 105039},
 {"CR1 monomial 2", 3792535, 105046},
 {"ID:48476485", 3792785, 105053},
 {"CR1 monomial 3", 3793034, 105060},
 {"ID:1246936946", 3793282, 105067},
 {"CR2 monomial 1", 3793535, 105074},
 {"ID:1856728487", 3793787, 105081},
 {"CR2 monomial 2", 3794040, 105088},
 {"ID:335657318", 3794291, 105095},
 {"CR2 monomial 3", 3794542, 105102},
 {"Proof|ID:1499935354", 3794798, 105109},
 {"Proof|ID:1499935354-1", 3794965, 105113},
 {"Proof|ID:1938671644", 3795130, 105117},
 {"Proof|ID:1938671644-1", 3795297, 105121},
 {"Proof|ID:48476485", 3795460, 105125},
 {"Proof|ID:48476485-1", 3795620, 105129},
 {"Proof|ID:1246936946", 3795780, 105133},
 {"Proof|ID:1246936946-1", 3795946, 105137},
 {"Proof|ID:1856728487", 3796110, 105141},
 {"Proof|ID:1856728487-1", 3796277, 105145},
 {"Proof|ID:335657318", 3796441, 105149},
 {"Proof|ID:335657318-1", 3796604, 105153},
 {"ID:624517207", 3796759, 105157},
 {"isMNTCRP", 3796999, 105164},
 {"ID:1172537415", 3797244, 105171},
 {"isMNTCRP alternative", 3797510, 105178},
 {"Proof|ID:1172537415", 3797775, 105185},
 {"Proof|ID:1172537415-1", 3797942, 105189},
 {"ID:1861685863", 3798101, 105193},
 {"ID:1154885365", 3798330, 105200},
 {"ID:1777331368", 3798558, 105207},
 {"CR3.\[DoubleRightArrow].1", 3798833, 105214},
 {"ID:1052103130", 3799096, 105221},
 {"CR3.\[DoubleRightArrow].2", 3799371, 105228},
 {"ID:1077684208", 3799634, 105235},
 {"CR3.\[DoubleRightArrow].3", 3799909, 105242},
 {"ID:86200506", 3800170, 105249},
 {"CR3.\[DoubleLeftArrow]", 3800435, 105256},
 {"ID:394014296", 3800690, 105263},
 {"CR3", 3800919, 105270},
 {"Proof|ID:1777331368", 3801164, 105277},
 {"Proof|ID:1777331368-1", 3801330, 105281},
 {"Proof|ID:1052103130", 3801494, 105285},
 {"Proof|ID:1052103130-1", 3801661, 105289},
 {"Proof|ID:1077684208", 3801826, 105293},
 {"Proof|ID:1077684208-1", 3801993, 105297},
 {"Proof|ID:86200506", 3802156, 105301},
 {"Proof|ID:86200506-1", 3802317, 105305},
 {"Proof|ID:394014296", 3802477, 105309},
 {"Proof|ID:394014296-1", 3802641, 105313},
 {"ID:1444400550", 3802798, 105317},
 {"ID:653031981", 3803025, 105324},
 {"CR4.\[DoubleRightArrow].1", 3803297, 105331},
 {"ID:943141844", 3803556, 105338},
 {"CR4.\[DoubleRightArrow].2", 3803829, 105345},
 {"ID:1143014070", 3804090, 105352},
 {"CR4.\[DoubleRightArrow].3", 3804365, 105359},
 {"ID:1630861163", 3804628, 105366},
 {"CR4.\[DoubleLeftArrow]", 3804897, 105373},
 {"ID:1701661908", 3805157, 105380},
 {"CR4", 3805388, 105387},
 {"Proof|ID:653031981", 3805634, 105394},
 {"Proof|ID:653031981-1", 3805795, 105398},
 {"Proof|ID:943141844", 3805954, 105402},
 {"Proof|ID:943141844-1", 3806118, 105406},
 {"Proof|ID:1143014070", 3806281, 105410},
 {"Proof|ID:1143014070-1", 3806448, 105414},
 {"Proof|ID:1630861163", 3806613, 105418},
 {"Proof|ID:1630861163-1", 3806779, 105422},
 {"Proof|ID:1701661908", 3806943, 105426},
 {"Proof|ID:1701661908-1", 3807109, 105430},
 {"ID:1489378443", 3807267, 105434},
 {"ID:196663580", 3807495, 105441},
 {"ID:1830388955", 3807721, 105448},
 {"ID:312371876", 3807948, 105455},
 {"CP1.\[DoubleLeftArrow]", 3808216, 105462},
 {"ID:1006612102", 3808475, 105469},
 {"CP2.\[DoubleLeftArrow]", 3808746, 105476},
 {"Proof|ID:312371876", 3809013, 105483},
 {"Proof|ID:312371876-1", 3809177, 105487},
 {"Proof|ID:1006612102", 3809340, 105491},
 {"Proof|ID:1006612102-1", 3809507, 105495},
 {"ID:542206690", 3809665, 105499},
 {"ID:1548368737", 3809892, 105506},
 {"ID:73478134", 3810118, 105513},
 {"ID:310376772", 3810341, 105520},
 {"409", 3810572, 105527},
 {"Proof|ID:310376772", 3810818, 105534},
 {"Proof|ID:310376772-1", 3810983, 105538},
 {"ID:383946867", 3811140, 105542},
 {"ID:1587705005", 3811366, 105549},
 {"ID:549840930", 3811593, 105556},
 {"R7.\[ScriptCapitalP].3", 3811861, 105563},
 {"ID:657656821", 3812119, 105570},
 {"R7.\[ScriptCapitalP].5", 3812388, 105577},
 {"ID:64232512", 3812646, 105584},
 {"ID:237038572", 3812869, 105591},
 {"ID:235501232", 3813094, 105598},
 {"ID:1513648693", 3813320, 105605},
 {"R7.\[ScriptCapitalP].1", 3813589, 105612},
 {"ID:808269926", 3813848, 105619},
 {"R7.\[ScriptCapitalP].4", 3814115, 105626},
 {"Proof|ID:549840930", 3814378, 105633},
 {"Proof|ID:549840930-1", 3814542, 105637},
 {"Proof|ID:657656821", 3814704, 105641},
 {"Proof|ID:657656821-1", 3814867, 105645},
 {"Proof|ID:1513648693", 3815029, 105649},
 {"Proof|ID:1513648693-1", 3815196, 105653},
 {"Proof|ID:808269926", 3815360, 105657},
 {"Proof|ID:808269926-1", 3815524, 105661},
 {"ID:255426543", 3815680, 105665},
 {"370", 3815908, 105672},
 {"Proof|ID:255426543", 3816151, 105679},
 {"Proof|ID:255426543-1", 3816313, 105683},
 {"ID:1342888001", 3816468, 105687},
 {"371", 3816698, 105694},
 {"Proof|ID:1342888001", 3816944, 105701},
 {"Proof|ID:1342888001-1", 3817110, 105705},
 {"ID:1021018734", 3817268, 105709},
 {"376", 3817498, 105716},
 {"Proof|ID:1021018734", 3817744, 105723},
 {"Proof|ID:1021018734-1", 3817910, 105727},
 {"ID:459032599", 3818067, 105731},
 {"isReductionRing Conservation Theorem", 3818362, 105738},
 {"ID:719489133", 3818633, 105745},
 {"algorithmic reduction ring \[ScriptCapitalR]", 3818943, 105752},
 {"ID:1159608082", 3819222, 105759},
 {"algorithmic reduction-polynomial domain \[ScriptCapitalP]", 3819566, \
105767},
 {"ID:847570831", 3819865, 105775},
 {"\[CenterDot] \[Succeeds]", 3820136, 105782},
 {"Proof|ID:847570831", 3820401, 105789},
 {"Proof|ID:847570831-1", 3820565, 105793},
 {"ID:2071048931", 3820722, 105797},
 {"ID:355437230", 3820950, 105804},
 {"ID:1401739284", 3821176, 105811},
 {"ID:29665786", 3821403, 105818},
 {"\[Tilde] \[ScriptCapitalP] 1", 3821679, 105825},
 {"ID:113282996", 3821939, 105832},
 {"\[Tilde] \[ScriptCapitalP] 2", 3822218, 105839},
 {"ID:1675445982", 3822482, 105846},
 {"\[Tilde] \[ScriptCapitalP] 3", 3822763, 105853},
 {"ID:1968653014", 3823029, 105860},
 {"\[Tilde] \[ScriptCapitalP] 4", 3823310, 105867},
 {"ID:860310855", 3823575, 105874},
 {"\[Tilde] \[ScriptCapitalP] 5", 3823854, 105881},
 {"ID:2026748686", 3824118, 105888},
 {"\[Tilde] multiplier \[ScriptCapitalP]", 3824417, 105895},
 {"Proof|ID:29665786", 3824696, 105902},
 {"Proof|ID:29665786-1", 3824856, 105906},
 {"Proof|ID:113282996", 3825015, 105910},
 {"Proof|ID:113282996-1", 3825179, 105914},
 {"Proof|ID:1675445982", 3825342, 105918},
 {"Proof|ID:1675445982-1", 3825509, 105922},
 {"Proof|ID:1968653014", 3825674, 105926},
 {"Proof|ID:1968653014-1", 3825841, 105930},
 {"Proof|ID:860310855", 3826005, 105934},
 {"Proof|ID:860310855-1", 3826169, 105938},
 {"Proof|ID:2026748686", 3826332, 105942},
 {"Proof|ID:2026748686-1", 3826499, 105946},
 {"ID:1529886137", 3826658, 105950},
 {"\[Tilde] \[ScriptCapitalP]", 3826935, 105957},
 {"Proof|ID:1529886137", 3827205, 105964},
 {"Proof|ID:1529886137-1", 3827372, 105968},
 {"ID:1680577066", 3827531, 105972},
 {"372", 3827761, 105979},
 {"Proof|ID:1680577066", 3828007, 105986},
 {"Proof|ID:1680577066-1", 3828170, 105990},
 {"ID:1866021817", 3828325, 105994},
 {"373", 3828555, 106001},
 {"Proof|ID:1866021817", 3828801, 106008},
 {"Proof|ID:1866021817-1", 3828964, 106012},
 {"ID:1272742036", 3829119, 106016},
 {"374", 3829349, 106023},
 {"Proof|ID:1272742036", 3829595, 106030},
 {"Proof|ID:1272742036-1", 3829760, 106034},
 {"ID:181479986", 3829916, 106038},
 {"ID:1276235216", 3830143, 106045},
 {"ID:376128819", 3830370, 106052},
 {"ID:973122334", 3830595, 106059},
 {"ID:1445738544", 3830822, 106066},
 {"375.1", 3831057, 106073},
 {"ID:874376987", 3831299, 106080},
 {"375.2", 3831532, 106087},
 {"ID:139929842", 3831772, 106094},
 {"375.3", 3832005, 106101},
 {"Proof|ID:1445738544", 3832252, 106108},
 {"Proof|ID:1445738544-2", 3832420, 106112},
 {"Proof|ID:874376987", 3832585, 106116},
 {"Proof|ID:874376987-1", 3832747, 106120},
 {"Proof|ID:139929842", 3832907, 106124},
 {"Proof|ID:139929842-2", 3833072, 106128},
 {"ID:462890005", 3833229, 106132},
 {"375", 3833457, 106139},
 {"Proof|ID:462890005", 3833700, 106146},
 {"Proof|ID:462890005-1", 3833861, 106150},
 {"ID:38279669", 3834013, 106154},
 {"isReducible lm", 3834262, 106161},
 {"ID:1376605467", 3834510, 106168},
 {"red lm set", 3834755, 106175},
 {"ID:673925715", 3835002, 106182},
 {"isReducible monomial", 3835265, 106189},
 {"Proof|ID:38279669", 3835525, 106196},
 {"Proof|ID:38279669-1", 3835684, 106200},
 {"Proof|ID:1376605467", 3835843, 106204},
 {"Proof|ID:1376605467-1", 3836009, 106208},
 {"Proof|ID:673925715", 3836172, 106212},
 {"Proof|ID:673925715-1", 3836334, 106216},
 {"ID:47923252", 3836487, 106220},
 {"isMonomialSequence", 3836744, 106227},
 {"ID:1632720588", 3836996, 106234},
 {"ID:1771873303", 3837224, 106241},
 {"isMonomialSequence \[NotEqual] 0", 3837513, 106248},
 {"ID:893563053", 3837782, 106255},
 {"isMonomialSequence lc 1", 3838051, 106262},
 {"ID:587765202", 3838309, 106269},
 {"isMonomialSequence lp 1", 3838578, 106276},
 {"ID:1254733295", 3838837, 106283},
 {"isMonomialSequence lc 2", 3839107, 106290},
 {"ID:280696418", 3839366, 106297},
 {"isMonomialSequence lp 2", 3839634, 106304},
 {"ID:1081126646", 3839892, 106311},
 {"isMonomialSequence subSeq", 3840166, 106318},
 {"ID:1101697023", 3840428, 106325},
 {"isMonomialSequence subsequence", 3840712, 106332},
 {"Proof|ID:1771873303", 3840985, 106339},
 {"Proof|ID:1771873303-1", 3841150, 106343},
 {"Proof|ID:893563053", 3841312, 106347},
 {"Proof|ID:893563053-1", 3841473, 106351},
 {"Proof|ID:587765202", 3841632, 106355},
 {"Proof|ID:587765202-1", 3841795, 106359},
 {"Proof|ID:1254733295", 3841957, 106363},
 {"Proof|ID:1254733295-1", 3842123, 106367},
 {"Proof|ID:280696418", 3842286, 106371},
 {"Proof|ID:280696418-1", 3842448, 106375},
 {"Proof|ID:1081126646", 3842609, 106379},
 {"Proof|ID:1081126646-1", 3842774, 106383},
 {"Proof|ID:1101697023", 3842937, 106387},
 {"Proof|ID:1101697023-1", 3843102, 106391},
 {"ID:182970451", 3843258, 106395},
 {"110", 3843487, 106402},
 {"ID:845195690", 3843725, 106409},
 {"111", 3843954, 106416},
 {"Proof|ID:182970451", 3844198, 106423},
 {"Proof|ID:182970451-1", 3844361, 106427},
 {"Proof|ID:845195690", 3844522, 106431},
 {"Proof|ID:845195690-1", 3844684, 106435},
 {"ID:2081362811", 3844839, 106439},
 {"ID:1173235103", 3845067, 106446},
 {"112", 3845299, 106453},
 {"ID:1805755328", 3845541, 106460},
 {"113", 3845772, 106467},
 {"ID:596327945", 3846012, 106474},
 {"114", 3846241, 106481},
 {"Proof|ID:1173235103", 3846486, 106488},
 {"Proof|ID:1173235103-1", 3846653, 106492},
 {"Proof|ID:1805755328", 3846818, 106496},
 {"Proof|ID:1805755328-1", 3846985, 106500},
 {"Proof|ID:596327945", 3847149, 106504},
 {"Proof|ID:596327945-1", 3847312, 106508},
 {"ID:1402642391", 3847468, 106512},
 {"115", 3847699, 106519},
 {"Proof|ID:1402642391", 3847946, 106526},
 {"Proof|ID:1402642391-1", 3848113, 106530},
 {"ID:980978647", 3848271, 106534},
 {"Dickson-domain \[ScriptCapitalT]", 3848558, 106541},
 {"ID:1408052308", 3848826, 106548},
 {"116", 3849056, 106555},
 {"Proof|ID:1408052308", 3849302, 106562},
 {"Proof|ID:1408052308-1", 3849467, 106566},
 {"ID:1183157947", 3849624, 106570},
 {"isAlgoReductionRing \[ScriptCapitalP]", 3849922, 106577},
 {"Proof|ID:1183157947", 3850202, 106584},
 {"Proof|ID:1183157947-1", 3850368, 106588},
 {"ID:1794188423", 3850526, 106592},
 {"isAlgoReductionRing Conservation Theorem Polynomials", 3850861, 106600}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[579, 22, 124, 2, 84, "Title",
 CellID->1766055249],
Cell[CellGroupData[{
Cell[728, 28, 119, 2, 73, "Section",
 CellID->775912285],
Cell[CellGroupData[{
Cell[872, 34, 174, 3, 70, "Subsection",
 CellID->600814793],
Cell[1049, 39, 1138, 24, 70, "Text",
 CellID->510256476],
Cell[2190, 65, 975, 16, 70, "Text",
 CellID->98348958],
Cell[3168, 83, 1098, 17, 70, "Text",
 CellID->173799993],
Cell[4269, 102, 887, 14, 70, "Text",
 CellID->17970129]
}, Closed]],
Cell[CellGroupData[{
Cell[5193, 121, 123, 2, 70, "Subsection",
 CellID->8118277],
Cell[5319, 125, 344, 8, 70, "Input",
 CellID->498486718],
Cell[5666, 135, 305, 8, 70, "Input",
 CellID->59840707],
Cell[5974, 145, 317, 8, 70, "Input",
 CellID->428817074],
Cell[6294, 155, 356, 8, 70, "Input",
 CellID->513028701],
Cell[6653, 165, 423, 9, 70, "Input",
 CellID->239747824],
Cell[7079, 176, 151, 3, 70, "Input",
 CellID->590188203],
Cell[7233, 181, 155, 3, 70, "Input",
 CellID->414054049],
Cell[7391, 186, 154, 3, 70, "Input",
 CellID->74859277],
Cell[CellGroupData[{
Cell[7570, 193, 1661, 35, 70, "Input",
 CellID->20666029],
Cell[9234, 230, 205, 5, 70, "Output",
 CellID->858659]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[9488, 241, 124, 2, 70, "Subsection",
 CellID->1386333768],
Cell[CellGroupData[{
Cell[9637, 247, 228, 5, 70, "Input",
 CellID->223251601],
Cell[CellGroupData[{
Cell[9890, 256, 639, 13, 70, "Print",
 CellID->457132276],
Cell[10532, 271, 369, 8, 70, "Print",
 CellID->230806296],
Cell[10904, 281, 371, 8, 70, "Print",
 CellID->694688721],
Cell[11278, 291, 371, 8, 70, "Print",
 CellID->420126785],
Cell[11652, 301, 373, 8, 70, "Print",
 CellID->198215461],
Cell[12028, 311, 370, 8, 70, "Print",
 CellID->95392771],
Cell[12401, 321, 371, 8, 70, "Print",
 CellID->761980848],
Cell[12775, 331, 371, 8, 70, "Print",
 CellID->3204933],
Cell[13149, 341, 170, 3, 70, "Print",
 CellID->100688736],
Cell[13322, 346, 160, 3, 70, "Print",
 CellID->301435823],
Cell[13485, 351, 356, 8, 70, "Print",
 CellID->89112948],
Cell[13844, 361, 360, 8, 70, "Print",
 CellID->93410960],
Cell[14207, 371, 363, 8, 70, "Print",
 CellID->908547730],
Cell[14573, 381, 363, 8, 70, "Print",
 CellID->825838993],
Cell[14939, 391, 361, 8, 70, "Print",
 CellID->670125391],
Cell[15303, 401, 361, 8, 70, "Print",
 CellID->732909961],
Cell[15667, 411, 361, 8, 70, "Print",
 CellID->370384636],
Cell[16031, 421, 361, 8, 70, "Print",
 CellID->305012795],
Cell[16395, 431, 361, 8, 70, "Print",
 CellID->324052939],
Cell[16759, 441, 363, 8, 70, "Print",
 CellID->175387165],
Cell[17125, 451, 362, 8, 70, "Print",
 CellID->18717014],
Cell[17490, 461, 361, 8, 70, "Print",
 CellID->516028767],
Cell[17854, 471, 361, 8, 70, "Print",
 CellID->443472550],
Cell[18218, 481, 361, 8, 70, "Print",
 CellID->285663591],
Cell[18582, 491, 172, 3, 70, "Print",
 CellID->100899060],
Cell[18757, 496, 151, 3, 70, "Print",
 CellID->359484989]
}, Open  ]],
Cell[18923, 502, 229, 6, 70, "Output",
 CellID->505902204]
}, Closed]],
Cell[CellGroupData[{
Cell[19189, 513, 441, 8, 70, "Input",
 CellID->204865936],
Cell[CellGroupData[{
Cell[19655, 525, 680, 14, 70, "Print",
 CellID->92695177],
Cell[20338, 541, 680, 14, 70, "Print",
 CellID->77992045],
Cell[21021, 557, 389, 8, 70, "Print",
 CellID->173700190],
Cell[21413, 567, 391, 8, 70, "Print",
 CellID->200266979],
Cell[21807, 577, 395, 8, 70, "Print",
 CellID->746639354],
Cell[22205, 587, 392, 8, 70, "Print",
 CellID->66712991],
Cell[22600, 597, 395, 8, 70, "Print",
 CellID->715089440],
Cell[22998, 607, 393, 8, 70, "Print",
 CellID->589956195],
Cell[23394, 617, 392, 8, 70, "Print",
 CellID->79498839],
Cell[23789, 627, 394, 8, 70, "Print",
 CellID->29993626],
Cell[24186, 637, 649, 13, 70, "Print",
 CellID->3483066],
Cell[24838, 652, 388, 8, 70, "Print",
 CellID->86123515],
Cell[25229, 662, 667, 13, 70, "Print",
 CellID->762618403],
Cell[25899, 677, 388, 8, 70, "Print",
 CellID->525181281],
Cell[26290, 687, 393, 8, 70, "Print",
 CellID->299146327],
Cell[26686, 697, 395, 8, 70, "Print",
 CellID->468958116],
Cell[27084, 707, 393, 8, 70, "Print",
 CellID->203278600],
Cell[27480, 717, 663, 13, 70, "Print",
 CellID->164689146],
Cell[28146, 732, 388, 8, 70, "Print",
 CellID->14680373],
Cell[28537, 742, 390, 8, 70, "Print",
 CellID->89093219],
Cell[28930, 752, 392, 8, 70, "Print",
 CellID->84778685],
Cell[29325, 762, 395, 8, 70, "Print",
 CellID->384310759],
Cell[29723, 772, 393, 8, 70, "Print",
 CellID->263550545],
Cell[30119, 782, 393, 8, 70, "Print",
 CellID->989691619],
Cell[30515, 792, 390, 8, 70, "Print",
 CellID->860959],
Cell[30908, 802, 668, 13, 70, "Print",
 CellID->54780262],
Cell[31579, 817, 388, 8, 70, "Print",
 CellID->34839652],
Cell[31970, 827, 391, 8, 70, "Print",
 CellID->230492857],
Cell[32364, 837, 394, 8, 70, "Print",
 CellID->56449567],
Cell[32761, 847, 395, 8, 70, "Print",
 CellID->911645706],
Cell[33159, 857, 393, 8, 70, "Print",
 CellID->229173356],
Cell[33555, 867, 192, 3, 70, "Print",
 CellID->247040699],
Cell[33750, 872, 182, 3, 70, "Print",
 CellID->517963491],
Cell[33935, 877, 380, 8, 70, "Print",
 CellID->75486086],
Cell[34318, 887, 385, 8, 70, "Print",
 CellID->128788291],
Cell[34706, 897, 387, 8, 70, "Print",
 CellID->205800235],
Cell[35096, 907, 387, 8, 70, "Print",
 CellID->336441065],
Cell[35486, 917, 384, 8, 70, "Print",
 CellID->78551879],
Cell[35873, 927, 386, 8, 70, "Print",
 CellID->99748948],
Cell[36262, 937, 384, 8, 70, "Print",
 CellID->68337618],
Cell[36649, 947, 387, 8, 70, "Print",
 CellID->180700173],
Cell[37039, 957, 385, 8, 70, "Print",
 CellID->121304440],
Cell[37427, 967, 387, 8, 70, "Print",
 CellID->215564190],
Cell[37817, 977, 384, 8, 70, "Print",
 CellID->97217331],
Cell[38204, 987, 385, 8, 70, "Print",
 CellID->460196882],
Cell[38592, 997, 387, 8, 70, "Print",
 CellID->205116320],
Cell[38982, 1007, 386, 8, 70, "Print",
 CellID->96939329],
Cell[39371, 1017, 387, 8, 70, "Print",
 CellID->132562752],
Cell[39761, 1027, 387, 8, 70, "Print",
 CellID->247813144],
Cell[40151, 1037, 387, 8, 70, "Print",
 CellID->358470877],
Cell[40541, 1047, 387, 8, 70, "Print",
 CellID->227650650],
Cell[40931, 1057, 385, 8, 70, "Print",
 CellID->129750566],
Cell[41319, 1067, 382, 8, 70, "Print",
 CellID->7673880],
Cell[41704, 1077, 387, 8, 70, "Print",
 CellID->1349429],
Cell[42094, 1087, 389, 8, 70, "Print",
 CellID->201668803],
Cell[42486, 1097, 389, 8, 70, "Print",
 CellID->230225205],
Cell[42878, 1107, 387, 8, 70, "Print",
 CellID->302199065],
Cell[43268, 1117, 388, 8, 70, "Print",
 CellID->45557856],
Cell[43659, 1127, 386, 8, 70, "Print",
 CellID->47265021],
Cell[44048, 1137, 389, 8, 70, "Print",
 CellID->196575828],
Cell[44440, 1147, 386, 8, 70, "Print",
 CellID->34429249],
Cell[44829, 1157, 385, 8, 70, "Print",
 CellID->1087430],
Cell[45217, 1167, 389, 8, 70, "Print",
 CellID->366912513],
Cell[45609, 1177, 388, 8, 70, "Print",
 CellID->37304165],
Cell[46000, 1187, 387, 8, 70, "Print",
 CellID->506613024],
Cell[46390, 1197, 389, 8, 70, "Print",
 CellID->213938952],
Cell[46782, 1207, 387, 8, 70, "Print",
 CellID->376215522],
Cell[47172, 1217, 387, 8, 70, "Print",
 CellID->264447866],
Cell[47562, 1227, 387, 8, 70, "Print",
 CellID->350054558],
Cell[47952, 1237, 389, 8, 70, "Print",
 CellID->398414970],
Cell[48344, 1247, 389, 8, 70, "Print",
 CellID->137094025],
Cell[48736, 1257, 389, 8, 70, "Print",
 CellID->440482684],
Cell[49128, 1267, 387, 8, 70, "Print",
 CellID->807851545],
Cell[49518, 1277, 194, 3, 70, "Print",
 CellID->672246938],
Cell[49715, 1282, 171, 3, 70, "Print",
 CellID->359470365]
}, Open  ]],
Cell[49901, 1288, 227, 6, 70, "Output",
 CellID->8844871]
}, Closed]],
Cell[50143, 1297, 430, 11, 70, "Input",
 CellID->1039312110],
Cell[50576, 1310, 329, 6, 70, "Input",
 CellID->19428626],
Cell[CellGroupData[{
Cell[50930, 1320, 689, 14, 70, "Input",
 CellID->1909803754],
Cell[51622, 1336, 384, 7, 70, "Output",
 CellID->721005233]
}, Open  ]],
Cell[CellGroupData[{
Cell[52043, 1348, 157, 3, 70, "Input",
 CellID->615062589],
Cell[52203, 1353, 1030, 18, 70, "Output",
 CellID->627924814]
}, Open  ]],
Cell[CellGroupData[{
Cell[53270, 1376, 276, 5, 70, "Input",
 CellID->1455199198],
Cell[53549, 1383, 1478, 31, 70, "Output",
 CellID->607555663]
}, Open  ]],
Cell[CellGroupData[{
Cell[55064, 1419, 493, 15, 70, "Input",
 CellID->220425063],
Cell[55560, 1436, 3899, 54, 70, "Output",
 CellID->2062191598]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[59508, 1496, 128, 2, 70, "Subsection",
 CellID->1454182661],
Cell[59639, 1500, 1065, 26, 70, "Input",
 CellID->2074904605],
Cell[60707, 1528, 416, 9, 70, "Input",
 CellID->1980565193],
Cell[61126, 1539, 169, 3, 70, "Input",
 CellID->444074165],
Cell[61298, 1544, 261, 5, 70, "Input",
 CellID->293654037]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[61608, 1555, 179, 3, 43, "Section",
 CellID->636613528],
Cell[CellGroupData[{
Cell[61812, 1562, 135, 2, 38, "Subsection",
 CellID->1320444671],
Cell[CellGroupData[{
Cell[61972, 1568, 134, 2, 28, "Subsubsection",
 CellID->1667576502],
Cell[62109, 1572, 1180, 33, 70, "Text",
 CellID->1164281211]
}, Closed]],
Cell[CellGroupData[{
Cell[63326, 1610, 125, 2, 22, "Subsubsection",
 CellID->1910498050],
Cell[63454, 1614, 47, 1, 12, "OpenEnvironment",
 CellID->127346395],
Cell[CellGroupData[{
Cell[63526, 1619, 565, 11, 43, "EnvironmentHeader",
 CellID->1867952146],
Cell[64094, 1632, 6848, 193, 366, "FormalTextInputFormula",
 CellTags->{
  "ID:558626034", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCommPPDomain"},
 CellID->558626034],
Cell[70945, 1827, 65, 1, 26, "EndEnvironmentMarker",
 CellID->723071054]
}, Open  ]],
Cell[71025, 1831, 49, 1, 16, "CloseEnvironment",
 CellID->1980315351]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[71123, 1838, 191, 3, 30, "Subsection",
 CellID->1879054478],
Cell[CellGroupData[{
Cell[71339, 1845, 129, 2, 70, "Subsubsection",
 CellID->1252947036],
Cell[71471, 1849, 2525, 85, 70, "Text",
 CellID->1930240611],
Cell[73999, 1936, 526, 11, 70, "Text",
 CellID->70056670],
Cell[74528, 1949, 5460, 120, 70, "Text",
 CellID->1768045063]
}, Closed]],
Cell[CellGroupData[{
Cell[80025, 2074, 222, 4, 70, "Subsubsection",
 CellID->103885810],
Cell[80250, 2080, 747, 14, 70, "Text",
 CellID->481145652],
Cell[81000, 2096, 785, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1568169011", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1568169011],
Cell[81788, 2116, 47, 1, 70, "OpenEnvironment",
 CellID->771652428],
Cell[CellGroupData[{
Cell[81860, 2121, 411, 10, 70, "EnvironmentHeader",
 CellID->1166916922],
Cell[82274, 2133, 898, 24, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1716984519", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1716984519],
Cell[83175, 2159, 2176, 67, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:205548127", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial"},
 CellID->205548127],
Cell[85354, 2228, 330, 6, 70, "EnvironmentText",
 CellID->712458476],
Cell[85687, 2236, 1836, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:711054552", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial II"},
 CellID->711054552],
Cell[87526, 2294, 2348, 72, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2035355406", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "monomial constructor"},
 CellID->2035355406],
Cell[89877, 2368, 1219, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:371468988", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "supp"},
 CellID->371468988],
Cell[91099, 2405, 2080, 61, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:741382791", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isLP"},
 CellID->741382791],
Cell[93182, 2468, 2289, 69, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1470398456", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp"},
 CellID->1470398456],
Cell[95474, 2539, 1075, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1541081556", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc"},
 CellID->1541081556],
Cell[96552, 2571, 1204, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:258866072", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm"},
 CellID->258866072],
Cell[97759, 2605, 1117, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1500059087", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R"},
 CellID->1500059087],
Cell[98879, 2637, 1111, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1301149083", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 0"},
 CellID->1301149083],
Cell[99993, 2667, 2271, 65, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:355792939", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H"},
 CellID->355792939],
Cell[102267, 2734, 2760, 76, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1617463509", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "monomial multiplication"},
 CellID->1617463509],
Cell[105030, 2812, 65, 1, 70, "EndEnvironmentMarker",
 CellID->176163025]
}, Open  ]],
Cell[105110, 2816, 49, 1, 70, "CloseEnvironment",
 CellID->1861413033],
Cell[105162, 2819, 47, 1, 70, "OpenEnvironment",
 CellID->116290711],
Cell[CellGroupData[{
Cell[105234, 2824, 400, 9, 70, "EnvironmentHeader",
 CellID->822645424],
Cell[105637, 2835, 2611, 63, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1026275980", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReductionPolynomialDomain"},
 CellID->1026275980],
Cell[108251, 2900, 4279, 121, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1335218902", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasDecisionPredicate"},
 CellID->1335218902],
Cell[112533, 3023, 676, 21, 70, "EnvironmentText",
 CellID->1135948496],
Cell[113212, 3046, 1778, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1175095894", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "containsMonomials"},
 CellID->1175095894],
Cell[114993, 3099, 6005, 158, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1967795228", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCoefficientFunction"},
 CellID->1967795228],
Cell[121001, 3259, 2813, 80, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2138030312", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplication"},
 CellID->2138030312],
Cell[123817, 3341, 2210, 69, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1709544239", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isExtensional"},
 CellID->1709544239],
Cell[126030, 3412, 2868, 86, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:235606987", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrdered"},
 CellID->235606987],
Cell[128901, 3500, 4757, 130, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:407906314", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultipliers"},
 CellID->407906314],
Cell[133661, 3632, 65, 1, 70, "EndEnvironmentMarker",
 CellID->203630590]
}, Open  ]],
Cell[133741, 3636, 48, 1, 70, "CloseEnvironment",
 CellID->821046233],
Cell[133792, 3639, 46, 1, 70, "OpenEnvironment",
 CellID->17561844],
Cell[CellGroupData[{
Cell[133863, 3644, 413, 10, 70, "EnvironmentHeader",
 CellID->1810576968],
Cell[134279, 3656, 1697, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1209025019", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isAlgoReductionPolynomialDomain"},
 CellID->1209025019],
Cell[135979, 3707, 7435, 185, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1424238209", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasLCRD"},
 CellID->1424238209],
Cell[143417, 3894, 4286, 114, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1761006179", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasRDM"},
 CellID->1761006179],
Cell[147706, 4010, 7315, 181, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:667868750", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCPM"},
 CellID->667868750],
Cell[155024, 4193, 64, 1, 70, "EndEnvironmentMarker",
 CellID->64457783]
}, Open  ]],
Cell[155103, 4197, 49, 1, 70, "CloseEnvironment",
 CellID->2016976285]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[155213, 4205, 294, 7, 43, "Section",
 CellID->928775899],
Cell[155510, 4214, 888, 15, 81, "Text",
 CellID->1613723827],
Cell[156401, 4231, 1727, 29, 131, "Text",
 CellID->2031272829],
Cell[158131, 4262, 47, 1, 12, "OpenEnvironment",
 CellID->753349515],
Cell[CellGroupData[{
Cell[158203, 4267, 520, 13, 43, "EnvironmentHeader",
 CellID->579292177],
Cell[158726, 4282, 805, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:267444622", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "ring \[ScriptCapitalR]"},
 CellID->267444622],
Cell[159534, 4305, 899, 23, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1725827561", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "power-product domain \[ScriptCapitalT]"},
 CellID->1725827561],
Cell[160436, 4330, 809, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:90242530", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isDomain \[ScriptCapitalP]"},
 CellID->90242530],
Cell[161248, 4353, 856, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1949049350", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasDecisionPredicate \[ScriptCapitalP]"},
 CellID->1949049350],
Cell[162107, 4377, 835, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:378755333", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "containsMonomials \[ScriptCapitalP]"},
 CellID->378755333],
Cell[162945, 4400, 859, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:226930466", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasCoefficientFunction \[ScriptCapitalP]"},
 CellID->226930466],
Cell[163807, 4424, 848, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:221039322", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isExtensional \[ScriptCapitalP]"},
 CellID->221039322],
Cell[164658, 4448, 65, 1, 26, "EndEnvironmentMarker",
 CellID->207394581]
}, Open  ]],
Cell[164738, 4452, 47, 1, 16, "CloseEnvironment",
 CellID->13183544],
Cell[CellGroupData[{
Cell[164810, 4457, 186, 3, 38, "Subsection",
 CellID->740008399],
Cell[CellGroupData[{
Cell[165021, 4464, 312, 6, 70, "Subsubsection",
 CellID->1996641139],
Cell[165336, 4472, 354, 7, 70, "Text",
 CellID->991585047],
Cell[165693, 4481, 47, 1, 70, "OpenEnvironment",
 CellID->642175228],
Cell[CellGroupData[{
Cell[165765, 4486, 477, 12, 70, "EnvironmentHeader",
 CellID->621471139],
Cell[166245, 4500, 810, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:216808867", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isDomain \[ScriptCapitalT]"},
 CellID->216808867],
Cell[167058, 4523, 1082, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1391060217", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isPartIrreflOrder \[Succeeds] \[ScriptCapitalT]"},
 CellID->1391060217],
Cell[168143, 4557, 2201, 70, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1131450960", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Noetherian induction on \[ScriptCapitalT]", "IR"},
 CellID->1131450960],
Cell[170347, 4629, 905, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:953344354", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 1 \[ScriptCapitalT]", "IR"},
 CellID->953344354],
Cell[171255, 4655, 525, 15, 70, "EnvironmentText",
 CellID->699465807],
Cell[171783, 4672, 1558, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1367902825", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 1"},
 CellID->1367902825],
Cell[173344, 4723, 761, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:643177315", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->643177315],
Cell[174108, 4745, 916, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1568022983", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 1 min"},
 CellID->1568022983],
Cell[175027, 4772, 1318, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1723905914", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 0"},
 CellID->1723905914],
Cell[176348, 4814, 883, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1825953062", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 2'"},
 CellID->1825953062],
Cell[177234, 4841, 879, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:596582273", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 2"},
 CellID->596582273],
Cell[178116, 4868, 901, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:389623087", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 3'"},
 CellID->389623087],
Cell[179020, 4896, 899, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:885344283", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 3"},
 CellID->885344283],
Cell[179922, 4924, 1017, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1033740528", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 4"},
 CellID->1033740528],
Cell[180942, 4956, 738, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:475932409", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->475932409],
Cell[181683, 4977, 222, 5, 70, "EnvironmentText",
 CellID->723934924],
Cell[181908, 4984, 1218, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1124160616", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 5"},
 CellID->1124160616],
Cell[183129, 5023, 1433, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1594052474", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 6"},
 CellID->1594052474],
Cell[184565, 5068, 1470, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1091730754", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 7"},
 CellID->1091730754],
Cell[186038, 5116, 1268, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1114122233", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "PP 8"},
 CellID->1114122233],
Cell[187309, 5156, 1033, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1909427587", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isPartReflOrder \[VerticalBar]"},
 CellID->1909427587],
Cell[188345, 5188, 1214, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:196272597", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrderEmbedding \[RawStar] \[VerticalBar]"},
 CellID->196272597],
Cell[189562, 5228, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1191981650]
}, Open  ]],
Cell[189643, 5232, 49, 1, 70, "CloseEnvironment",
 CellID->2001326268],
Cell[CellGroupData[{
Cell[189717, 5237, 1788, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:216808867", "Proof|ID:216808867-1"},
 CellID->77023573],
Cell[191508, 5295, 479, 12, 70, "ProofInfoBody",
 CellID->44400236]
}, Closed]],
Cell[CellGroupData[{
Cell[192024, 5312, 2051, 63, 70, "ProofInfo",
 CellTags->{"Proof|ID:1391060217", "Proof|ID:1391060217-1"},
 CellID->640802005],
Cell[194078, 5377, 524, 13, 70, "ProofInfoBody",
 CellID->1491734577]
}, Closed]],
Cell[CellGroupData[{
Cell[194639, 5395, 5088, 131, 70, "ProofInfo",
 CellTags->{"Proof|ID:1131450960", "Proof|ID:1131450960-1"},
 CellID->381161681],
Cell[199730, 5528, 481, 12, 70, "ProofInfoBody",
 CellID->397918332]
}, Closed]],
Cell[CellGroupData[{
Cell[200248, 5545, 1947, 59, 70, "ProofInfo",
 CellTags->{"Proof|ID:953344354", "Proof|ID:953344354-1"},
 CellID->238169741],
Cell[202198, 5606, 480, 12, 70, "ProofInfoBody",
 CellID->236757624]
}, Closed]],
Cell[CellGroupData[{
Cell[202715, 5623, 3889, 107, 70, "ProofInfo",
 CellTags->{"Proof|ID:1367902825", "Proof|ID:1367902825-1"},
 CellID->640235643],
Cell[206607, 5732, 482, 12, 70, "ProofInfoBody",
 CellID->1625734410]
}, Closed]],
Cell[CellGroupData[{
Cell[207126, 5749, 2794, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:1568022983", "Proof|ID:1568022983-1"},
 CellID->1849364148],
Cell[209923, 5833, 481, 12, 70, "ProofInfoBody",
 CellID->763438333]
}, Closed]],
Cell[CellGroupData[{
Cell[210441, 5850, 3822, 103, 70, "ProofInfo",
 CellTags->{"Proof|ID:1723905914", "Proof|ID:1723905914-1"},
 CellID->818368910],
Cell[214266, 5955, 482, 12, 70, "ProofInfoBody",
 CellID->2031601064]
}, Closed]],
Cell[CellGroupData[{
Cell[214785, 5972, 3050, 87, 70, "ProofInfo",
 CellTags->{"Proof|ID:1825953062", "Proof|ID:1825953062-1"},
 CellID->10445487],
Cell[217838, 6061, 482, 12, 70, "ProofInfoBody",
 CellID->1071552749]
}, Closed]],
Cell[CellGroupData[{
Cell[218357, 6078, 3040, 86, 70, "ProofInfo",
 CellTags->{"Proof|ID:596582273", "Proof|ID:596582273-1"},
 CellID->525784934],
Cell[221400, 6166, 481, 12, 70, "ProofInfoBody",
 CellID->1816433154]
}, Closed]],
Cell[CellGroupData[{
Cell[221918, 6183, 3202, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:389623087", "Proof|ID:389623087-1"},
 CellID->601121285],
Cell[225123, 6275, 480, 12, 70, "ProofInfoBody",
 CellID->201155910]
}, Closed]],
Cell[CellGroupData[{
Cell[225640, 6292, 3200, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:885344283", "Proof|ID:885344283-1"},
 CellID->573214883],
Cell[228843, 6384, 478, 12, 70, "ProofInfoBody",
 CellID->9167544]
}, Closed]],
Cell[CellGroupData[{
Cell[229358, 6401, 3592, 99, 70, "ProofInfo",
 CellTags->{"Proof|ID:1033740528", "Proof|ID:1033740528-1"},
 CellID->1813081668],
Cell[232953, 6502, 480, 12, 70, "ProofInfoBody",
 CellID->97754333]
}, Closed]],
Cell[CellGroupData[{
Cell[233470, 6519, 3939, 107, 70, "ProofInfo",
 CellTags->{"Proof|ID:1124160616", "Proof|ID:1124160616-1"},
 CellID->160370785],
Cell[237412, 6628, 480, 12, 70, "ProofInfoBody",
 CellID->98989517]
}, Closed]],
Cell[CellGroupData[{
Cell[237929, 6645, 4737, 119, 70, "ProofInfo",
 CellTags->{"Proof|ID:1594052474", "Proof|ID:1594052474-1"},
 CellID->1673583185],
Cell[242669, 6766, 622, 15, 70, "ProofInfoBody",
 CellID->1940769656]
}, Closed]],
Cell[CellGroupData[{
Cell[243328, 6786, 4848, 125, 70, "ProofInfo",
 CellTags->{"Proof|ID:1091730754", "Proof|ID:1091730754-1"},
 CellID->347735719],
Cell[248179, 6913, 523, 13, 70, "ProofInfoBody",
 CellID->960937874]
}, Closed]],
Cell[CellGroupData[{
Cell[248739, 6931, 4600, 119, 70, "ProofInfo",
 CellTags->{"Proof|ID:1114122233", "Proof|ID:1114122233-1"},
 CellID->1981139454],
Cell[253342, 7052, 481, 12, 70, "ProofInfoBody",
 CellID->669358996]
}, Closed]],
Cell[CellGroupData[{
Cell[253860, 7069, 2011, 62, 70, "ProofInfo",
 CellTags->{"Proof|ID:1909427587", "Proof|ID:1909427587-1"},
 CellID->106187856],
Cell[255874, 7133, 482, 12, 70, "ProofInfoBody",
 CellID->1650175325]
}, Closed]],
Cell[CellGroupData[{
Cell[256393, 7150, 2136, 64, 70, "ProofInfo",
 CellTags->{"Proof|ID:196272597", "Proof|ID:196272597-1"},
 CellID->626133587],
Cell[258532, 7216, 479, 12, 70, "ProofInfoBody",
 CellID->31720515],
Cell[259014, 7230, 100, 2, 70, "Print",
 CellID->737442690]
}, Closed]],
Cell[259129, 7235, 47, 1, 70, "OpenEnvironment",
 CellID->867314350],
Cell[CellGroupData[{
Cell[259201, 7240, 380, 9, 70, "EnvironmentHeader",
 CellID->1123038343],
Cell[259584, 7251, 761, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:334245933", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->334245933],
Cell[260348, 7273, 926, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1101510167", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lcm", "IR"},
 CellID->1101510167],
Cell[261277, 7300, 370, 7, 70, "EnvironmentText",
 CellID->102897427],
Cell[261650, 7309, 885, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:571674285", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 1"},
 CellID->571674285],
Cell[262538, 7335, 889, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:775675057", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 1'"},
 CellID->775675057],
Cell[263430, 7361, 1471, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:517355459", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 2"},
 CellID->517355459],
Cell[264904, 7408, 936, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:651227563", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 3"},
 CellID->651227563],
Cell[265843, 7436, 844, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:482943675", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 6"},
 CellID->482943675],
Cell[266690, 7462, 206, 5, 70, "EnvironmentText",
 CellID->1329971511],
Cell[266899, 7469, 906, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:39860588", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 4"},
 CellID->39860588],
Cell[267808, 7496, 911, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1631465634", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 4'"},
 CellID->1631465634],
Cell[268722, 7523, 1608, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:835022497", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lcm 5"},
 CellID->835022497],
Cell[270333, 7576, 65, 1, 70, "EndEnvironmentMarker",
 CellID->157525244]
}, Open  ]],
Cell[270413, 7580, 49, 1, 70, "CloseEnvironment",
 CellID->1568132504],
Cell[CellGroupData[{
Cell[270487, 7585, 3057, 85, 70, "ProofInfo",
 CellTags->{"Proof|ID:1101510167", "Proof|ID:1101510167-1"},
 CellID->1779125069],
Cell[273547, 7672, 482, 12, 70, "ProofInfoBody",
 CellID->1577134591]
}, Closed]],
Cell[CellGroupData[{
Cell[274066, 7689, 3119, 86, 70, "ProofInfo",
 CellTags->{"Proof|ID:571674285", "Proof|ID:571674285-1"},
 CellID->348779623],
Cell[277188, 7777, 481, 12, 70, "ProofInfoBody",
 CellID->1315821106]
}, Closed]],
Cell[CellGroupData[{
Cell[277706, 7794, 3122, 86, 70, "ProofInfo",
 CellTags->{"Proof|ID:775675057", "Proof|ID:775675057-1"},
 CellID->1055462613],
Cell[280831, 7882, 481, 12, 70, "ProofInfoBody",
 CellID->1433558380]
}, Closed]],
Cell[CellGroupData[{
Cell[281349, 7899, 4413, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:517355459", "Proof|ID:517355459-1"},
 CellID->1760053657],
Cell[285765, 8015, 480, 12, 70, "ProofInfoBody",
 CellID->827965406]
}, Closed]],
Cell[CellGroupData[{
Cell[286282, 8032, 3362, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:651227563", "Proof|ID:651227563-1"},
 CellID->1015477485],
Cell[289647, 8126, 481, 12, 70, "ProofInfoBody",
 CellID->2032600889]
}, Closed]],
Cell[CellGroupData[{
Cell[290165, 8143, 2705, 80, 70, "ProofInfo",
 CellTags->{"Proof|ID:482943675", "Proof|ID:482943675-1"},
 CellID->29218054],
Cell[292873, 8225, 479, 12, 70, "ProofInfoBody",
 CellID->48707691]
}, Closed]],
Cell[CellGroupData[{
Cell[293389, 8242, 3460, 95, 70, "ProofInfo",
 CellTags->{"Proof|ID:39860588", "Proof|ID:39860588-1"},
 CellID->2016600132],
Cell[296852, 8339, 479, 12, 70, "ProofInfoBody",
 CellID->912021378]
}, Closed]],
Cell[CellGroupData[{
Cell[297368, 8356, 3476, 96, 70, "ProofInfo",
 CellTags->{"Proof|ID:1631465634", "Proof|ID:1631465634-1"},
 CellID->1760483406],
Cell[300847, 8454, 482, 12, 70, "ProofInfoBody",
 CellID->1884546058]
}, Closed]],
Cell[CellGroupData[{
Cell[301366, 8471, 4774, 122, 70, "ProofInfo",
 CellTags->{"Proof|ID:835022497", "Proof|ID:835022497-1"},
 CellID->248308182],
Cell[306143, 8595, 480, 12, 70, "ProofInfoBody",
 CellID->760468064]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[306672, 8613, 196, 3, 70, "Subsubsection",
 CellID->265493105],
Cell[306871, 8618, 46, 1, 70, "OpenEnvironment",
 CellID->17942811],
Cell[CellGroupData[{
Cell[306942, 8623, 366, 9, 70, "EnvironmentHeader",
 CellID->935061110],
Cell[307311, 8634, 736, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:100863090", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->100863090],
Cell[308050, 8654, 385, 12, 70, "EnvironmentText",
 CellID->634130198],
Cell[308438, 8668, 982, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:292726431", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership supp general"},
 CellID->292726431],
Cell[309423, 8697, 1636, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:109557753", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Element] supp general"},
 CellID->109557753],
Cell[311062, 8749, 65, 1, 70, "EndEnvironmentMarker",
 CellID->302285674]
}, Open  ]],
Cell[311142, 8753, 48, 1, 70, "CloseEnvironment",
 CellID->107322010],
Cell[CellGroupData[{
Cell[311215, 8758, 2703, 76, 70, "ProofInfo",
 CellTags->{"Proof|ID:292726431", "Proof|ID:292726431-1"},
 CellID->837210885],
Cell[313921, 8836, 480, 12, 70, "ProofInfoBody",
 CellID->225362927]
}, Closed]],
Cell[CellGroupData[{
Cell[314438, 8853, 4231, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:109557753", "Proof|ID:109557753-1"},
 CellID->474399008],
Cell[318672, 8965, 480, 12, 70, "ProofInfoBody",
 CellID->369676367]
}, Closed]],
Cell[319167, 8980, 47, 1, 70, "OpenEnvironment",
 CellID->290097701],
Cell[CellGroupData[{
Cell[319239, 8985, 430, 11, 70, "EnvironmentHeader",
 CellID->227405523],
Cell[319672, 8998, 1203, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:938132667", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership supp", "IR"},
 CellID->938132667],
Cell[320878, 9031, 1747, 54, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:56177686", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Element] supp", "IR"},
 CellID->56177686],
Cell[322628, 9087, 867, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:163445395", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 0", "IR"},
 CellID->163445395],
Cell[323498, 9113, 868, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:671805226", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership 1", "IR"},
 CellID->671805226],
Cell[324369, 9139, 1292, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:373967346", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership C", "IR"},
 CellID->373967346],
Cell[325664, 9177, 65, 1, 70, "EndEnvironmentMarker",
 CellID->929112994]
}, Open  ]],
Cell[325744, 9181, 48, 1, 70, "CloseEnvironment",
 CellID->618819632],
Cell[CellGroupData[{
Cell[325817, 9186, 2450, 72, 70, "ProofInfo",
 CellTags->{"Proof|ID:938132667", "Proof|ID:938132667-1"},
 CellID->22005555],
Cell[328270, 9260, 480, 12, 70, "ProofInfoBody",
 CellID->298479159]
}, Closed]],
Cell[CellGroupData[{
Cell[328787, 9277, 4070, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:56177686", "Proof|ID:56177686-1"},
 CellID->520191277],
Cell[332860, 9387, 479, 12, 70, "ProofInfoBody",
 CellID->406632049]
}, Closed]],
Cell[CellGroupData[{
Cell[333376, 9404, 1902, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:163445395", "Proof|ID:163445395-1"},
 CellID->25655089],
Cell[335281, 9464, 479, 12, 70, "ProofInfoBody",
 CellID->25955964]
}, Closed]],
Cell[CellGroupData[{
Cell[335797, 9481, 1903, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:671805226", "Proof|ID:671805226-1"},
 CellID->637038652],
Cell[337703, 9541, 479, 12, 70, "ProofInfoBody",
 CellID->46850605]
}, Closed]],
Cell[CellGroupData[{
Cell[338219, 9558, 3039, 84, 70, "ProofInfo",
 CellTags->{"Proof|ID:373967346", "Proof|ID:373967346-1"},
 CellID->278031390],
Cell[341261, 9644, 479, 12, 70, "ProofInfoBody",
 CellID->27619205]
}, Closed]],
Cell[341755, 9659, 47, 1, 70, "OpenEnvironment",
 CellID->118902962],
Cell[CellGroupData[{
Cell[341827, 9664, 571, 13, 70, "EnvironmentHeader",
 CellID->999616402],
Cell[342401, 9679, 2823, 76, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1248732928", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Monomial"},
 CellID->1248732928],
Cell[345227, 9757, 2152, 61, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1277602733", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp sufficient"},
 CellID->1277602733],
Cell[347382, 9820, 2557, 73, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1270225306", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp alternative"},
 CellID->1270225306],
Cell[349942, 9895, 65, 1, 70, "EndEnvironmentMarker",
 CellID->702468881]
}, Open  ]],
Cell[350022, 9899, 48, 1, 70, "CloseEnvironment",
 CellID->924974006],
Cell[CellGroupData[{
Cell[350095, 9904, 5407, 137, 70, "ProofInfo",
 CellTags->{"Proof|ID:1248732928", "Proof|ID:1248732928-1"},
 CellID->328071702],
Cell[355505, 10043, 481, 12, 70, "ProofInfoBody",
 CellID->264893878]
}, Closed]],
Cell[CellGroupData[{
Cell[356023, 10060, 5075, 131, 70, "ProofInfo",
 CellTags->{"Proof|ID:1277602733", "Proof|ID:1277602733-1"},
 CellID->88373483],
Cell[361101, 10193, 481, 12, 70, "ProofInfoBody",
 CellID->287665210]
}, Closed]],
Cell[CellGroupData[{
Cell[361619, 10210, 6046, 151, 70, "ProofInfo",
 CellTags->{"Proof|ID:1270225306", "Proof|ID:1270225306-1"},
 CellID->426200985],
Cell[367668, 10363, 481, 12, 70, "ProofInfoBody",
 CellID->939653437]
}, Closed]],
Cell[368164, 10378, 48, 1, 70, "OpenEnvironment",
 CellID->1434901757],
Cell[CellGroupData[{
Cell[368237, 10383, 530, 13, 70, "EnvironmentHeader",
 CellID->1095667578],
Cell[368770, 10398, 1246, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:292606918", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomials", "IR"},
 CellID->292606918],
Cell[370019, 10435, 886, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:485824967", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->485824967],
Cell[370908, 10460, 950, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:406494450", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lp", "IR"},
 CellID->406494450],
Cell[371861, 10488, 955, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1454724371", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lc", "IR"},
 CellID->1454724371],
Cell[372819, 10516, 955, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1589530505", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership lm", "IR"},
 CellID->1589530505],
Cell[373777, 10544, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1979542526]
}, Open  ]],
Cell[373858, 10548, 48, 1, 70, "CloseEnvironment",
 CellID->561932109],
Cell[CellGroupData[{
Cell[373931, 10553, 2992, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:292606918", "Proof|ID:292606918-1"},
 CellID->1053213110],
Cell[376926, 10637, 480, 12, 70, "ProofInfoBody",
 CellID->303879011]
}, Closed]],
Cell[CellGroupData[{
Cell[377443, 10654, 2943, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:406494450", "Proof|ID:406494450-1"},
 CellID->928711222],
Cell[380389, 10738, 481, 12, 70, "ProofInfoBody",
 CellID->1865842282]
}, Closed]],
Cell[CellGroupData[{
Cell[380907, 10755, 2952, 83, 70, "ProofInfo",
 CellTags->{"Proof|ID:1454724371", "Proof|ID:1454724371-1"},
 CellID->776114793],
Cell[383862, 10840, 482, 12, 70, "ProofInfoBody",
 CellID->1327093525]
}, Closed]],
Cell[CellGroupData[{
Cell[384381, 10857, 2953, 83, 70, "ProofInfo",
 CellTags->{"Proof|ID:1589530505", "Proof|ID:1589530505-1"},
 CellID->1436300644],
Cell[387337, 10942, 481, 12, 70, "ProofInfoBody",
 CellID->680527277]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[387867, 10960, 211, 5, 70, "Subsubsection",
 CellID->1344153722],
Cell[388081, 10967, 47, 1, 70, "OpenEnvironment",
 CellID->260708809],
Cell[CellGroupData[{
Cell[388153, 10972, 438, 13, 70, "EnvironmentHeader",
 CellID->603207502],
Cell[388594, 10987, 740, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:720209465", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->720209465],
Cell[389337, 11008, 915, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1852885314", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 0"},
 CellID->1852885314],
Cell[390255, 11035, 761, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:148341425", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->148341425],
Cell[391019, 11057, 1263, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1443097690", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 1"},
 CellID->1443097690],
Cell[392285, 11095, 1055, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1675126992", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 2"},
 CellID->1675126992],
Cell[393343, 11128, 1260, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:814027629", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 3"},
 CellID->814027629],
Cell[394606, 11166, 1296, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1597798077", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 4"},
 CellID->1597798077],
Cell[395905, 11206, 1596, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1711645326", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 5"},
 CellID->1711645326],
Cell[397504, 11255, 65, 1, 70, "EndEnvironmentMarker",
 CellID->320679728]
}, Open  ]],
Cell[397584, 11259, 48, 1, 70, "CloseEnvironment",
 CellID->514407804],
Cell[CellGroupData[{
Cell[397657, 11264, 2727, 81, 70, "ProofInfo",
 CellTags->{"Proof|ID:1852885314", "Proof|ID:1852885314-1"},
 CellID->131831930],
Cell[400387, 11347, 479, 12, 70, "ProofInfoBody",
 CellID->1038146]
}, Closed]],
Cell[CellGroupData[{
Cell[400903, 11364, 4602, 119, 70, "ProofInfo",
 CellTags->{"Proof|ID:1443097690", "Proof|ID:1443097690-1"},
 CellID->324739814],
Cell[405508, 11485, 480, 12, 70, "ProofInfoBody",
 CellID->58511154]
}, Closed]],
Cell[CellGroupData[{
Cell[406025, 11502, 3766, 101, 70, "ProofInfo",
 CellTags->{"Proof|ID:1675126992", "Proof|ID:1675126992-1"},
 CellID->947152217],
Cell[409794, 11605, 481, 12, 70, "ProofInfoBody",
 CellID->450219457]
}, Closed]],
Cell[CellGroupData[{
Cell[410312, 11622, 4591, 118, 70, "ProofInfo",
 CellTags->{"Proof|ID:814027629", "Proof|ID:814027629-1"},
 CellID->1077213],
Cell[414906, 11742, 480, 12, 70, "ProofInfoBody",
 CellID->153578015]
}, Closed]],
Cell[CellGroupData[{
Cell[415423, 11759, 4096, 109, 70, "ProofInfo",
 CellTags->{"Proof|ID:1597798077", "Proof|ID:1597798077-1"},
 CellID->1370021543],
Cell[419522, 11870, 481, 12, 70, "ProofInfoBody",
 CellID->320424180]
}, Closed]],
Cell[CellGroupData[{
Cell[420040, 11887, 4712, 123, 70, "ProofInfo",
 CellTags->{"Proof|ID:1711645326", "Proof|ID:1711645326-1"},
 CellID->209179501],
Cell[424755, 12012, 481, 12, 70, "ProofInfoBody",
 CellID->599442953]
}, Closed]],
Cell[425251, 12027, 144, 2, 70, "Text",
 CellID->1514263990],
Cell[425398, 12031, 47, 1, 70, "OpenEnvironment",
 CellID->973491128],
Cell[CellGroupData[{
Cell[425470, 12036, 525, 17, 70, "EnvironmentHeader",
 CellID->2040579658],
Cell[425998, 12055, 764, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1053827266", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1053827266],
Cell[426765, 12077, 928, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:829207957", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership + \[ScriptCapitalP]", "IR"},
 CellID->829207957],
Cell[427696, 12104, 929, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:227196886", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership - \[ScriptCapitalP] 1", "IR"},
 CellID->227196886],
Cell[428628, 12131, 931, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:414658623", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership - \[ScriptCapitalP] 2", "IR"},
 CellID->414658623],
Cell[429562, 12158, 65, 1, 70, "EndEnvironmentMarker",
 CellID->525227380]
}, Open  ]],
Cell[429642, 12162, 48, 1, 70, "CloseEnvironment",
 CellID->838178864],
Cell[CellGroupData[{
Cell[429715, 12167, 2997, 83, 70, "ProofInfo",
 CellTags->{"Proof|ID:829207957", "Proof|ID:829207957-1"},
 CellID->188209562],
Cell[432715, 12252, 480, 12, 70, "ProofInfoBody",
 CellID->353588127]
}, Closed]],
Cell[CellGroupData[{
Cell[433232, 12269, 2614, 75, 70, "ProofInfo",
 CellTags->{"Proof|ID:227196886", "Proof|ID:227196886-1"},
 CellID->65751116],
Cell[435849, 12346, 479, 12, 70, "ProofInfoBody",
 CellID->18198817]
}, Closed]],
Cell[CellGroupData[{
Cell[436365, 12363, 3000, 83, 70, "ProofInfo",
 CellTags->{"Proof|ID:414658623", "Proof|ID:414658623-1"},
 CellID->19838666],
Cell[439368, 12448, 480, 12, 70, "ProofInfoBody",
 CellID->879352583]
}, Closed]],
Cell[439863, 12463, 48, 1, 70, "OpenEnvironment",
 CellID->2126627650],
Cell[CellGroupData[{
Cell[439936, 12468, 384, 9, 70, "EnvironmentHeader",
 CellID->826776809],
Cell[440323, 12479, 1210, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:314696779", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1=0 \[ScriptCapitalP]"},
 CellID->314696779],
Cell[441536, 12517, 1287, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:849229881", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 1"},
 CellID->849229881],
Cell[442826, 12558, 1291, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:406483636", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 1"},
 CellID->406483636],
Cell[444120, 12599, 1241, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1767085150", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm 1"},
 CellID->1767085150],
Cell[445364, 12639, 912, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1272860149", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1272860149],
Cell[446279, 12664, 907, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:144989886", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "non-zero 2", "IR"},
 CellID->144989886],
Cell[447189, 12690, 1001, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2138850160", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 2"},
 CellID->2138850160],
Cell[448193, 12720, 988, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1145306026", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 2"},
 CellID->1145306026],
Cell[449184, 12751, 861, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1442316125", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc 3"},
 CellID->1442316125],
Cell[450048, 12777, 2155, 68, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:655883205", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 3"},
 CellID->655883205],
Cell[452206, 12847, 1488, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:996448533", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "non-zero 3", "IR"},
 CellID->996448533],
Cell[453697, 12893, 1787, 54, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1383271844", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 5"},
 CellID->1383271844],
Cell[455487, 12949, 736, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:945941808", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->945941808],
Cell[456226, 12970, 1443, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:857665826", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C lm"},
 CellID->857665826],
Cell[457672, 13016, 1289, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1411816658", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm 2"},
 CellID->1411816658],
Cell[458964, 13056, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1163861425]
}, Open  ]],
Cell[459045, 13060, 49, 1, 70, "CloseEnvironment",
 CellID->1926885900],
Cell[CellGroupData[{
Cell[459119, 13065, 2326, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:314696779", "Proof|ID:314696779-1"},
 CellID->817945206],
Cell[461448, 13137, 481, 12, 70, "ProofInfoBody",
 CellID->1429481448]
}, Closed]],
Cell[CellGroupData[{
Cell[461966, 13154, 2434, 72, 70, "ProofInfo",
 CellTags->{"Proof|ID:849229881", "Proof|ID:849229881-1"},
 CellID->766889252],
Cell[464403, 13228, 481, 12, 70, "ProofInfoBody",
 CellID->1786903669]
}, Closed]],
Cell[CellGroupData[{
Cell[464921, 13245, 2435, 72, 70, "ProofInfo",
 CellTags->{"Proof|ID:406483636", "Proof|ID:406483636-1"},
 CellID->1626957666],
Cell[467359, 13319, 481, 12, 70, "ProofInfoBody",
 CellID->1255490528]
}, Closed]],
Cell[CellGroupData[{
Cell[467877, 13336, 2443, 73, 70, "ProofInfo",
 CellTags->{"Proof|ID:1767085150", "Proof|ID:1767085150-1"},
 CellID->450159143],
Cell[470323, 13411, 482, 12, 70, "ProofInfoBody",
 CellID->1165157523]
}, Closed]],
Cell[CellGroupData[{
Cell[470842, 13428, 2948, 84, 70, "ProofInfo",
 CellTags->{"Proof|ID:144989886", "Proof|ID:144989886-1"},
 CellID->914771064],
Cell[473793, 13514, 481, 12, 70, "ProofInfoBody",
 CellID->1590090525]
}, Closed]],
Cell[CellGroupData[{
Cell[474311, 13531, 3337, 93, 70, "ProofInfo",
 CellTags->{"Proof|ID:2138850160", "Proof|ID:2138850160-1"},
 CellID->1795169377],
Cell[477651, 13626, 481, 12, 70, "ProofInfoBody",
 CellID->683454739]
}, Closed]],
Cell[CellGroupData[{
Cell[478169, 13643, 3540, 97, 70, "ProofInfo",
 CellTags->{"Proof|ID:1145306026", "Proof|ID:1145306026-1"},
 CellID->109823534],
Cell[481712, 13742, 482, 12, 70, "ProofInfoBody",
 CellID->1039693341]
}, Closed]],
Cell[CellGroupData[{
Cell[482231, 13759, 2940, 85, 70, "ProofInfo",
 CellTags->{"Proof|ID:1442316125", "Proof|ID:1442316125-1"},
 CellID->1835069754],
Cell[485174, 13846, 482, 12, 70, "ProofInfoBody",
 CellID->1340501902]
}, Closed]],
Cell[CellGroupData[{
Cell[485693, 13863, 6014, 148, 70, "ProofInfo",
 CellTags->{"Proof|ID:655883205", "Proof|ID:655883205-1"},
 CellID->489542376],
Cell[491710, 14013, 480, 12, 70, "ProofInfoBody",
 CellID->634307408]
}, Closed]],
Cell[CellGroupData[{
Cell[492227, 14030, 4526, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:996448533", "Proof|ID:996448533-1"},
 CellID->925906148],
Cell[496756, 14148, 479, 12, 70, "ProofInfoBody",
 CellID->68221894]
}, Closed]],
Cell[CellGroupData[{
Cell[497272, 14165, 4891, 123, 70, "ProofInfo",
 CellTags->{"Proof|ID:1383271844", "Proof|ID:1383271844-1"},
 CellID->1411486137],
Cell[502166, 14290, 482, 12, 70, "ProofInfoBody",
 CellID->1721706305]
}, Closed]],
Cell[CellGroupData[{
Cell[502685, 14307, 4797, 121, 70, "ProofInfo",
 CellTags->{"Proof|ID:857665826", "Proof|ID:857665826-1"},
 CellID->606096953],
Cell[507485, 14430, 481, 12, 70, "ProofInfoBody",
 CellID->1744415237]
}, Closed]],
Cell[CellGroupData[{
Cell[508003, 14447, 4295, 113, 70, "ProofInfo",
 CellTags->{"Proof|ID:1411816658", "Proof|ID:1411816658-1"},
 CellID->352726063],
Cell[512301, 14562, 481, 12, 70, "ProofInfoBody",
 CellID->644334913]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[512831, 14580, 186, 3, 70, "Subsubsection",
 CellID->2098369935],
Cell[513020, 14585, 47, 1, 70, "OpenEnvironment",
 CellID->633845652],
Cell[CellGroupData[{
Cell[513092, 14590, 496, 11, 70, "EnvironmentHeader",
 CellID->1224489222],
Cell[513591, 14603, 960, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1368691869", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1\[CenterDot]1"},
 CellID->1368691869],
Cell[514554, 14630, 766, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1325102956", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1325102956],
Cell[515323, 14652, 739, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1663774274", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1663774274],
Cell[516065, 14673, 1124, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:941688542", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial \[CenterDot]"},
 CellID->941688542],
Cell[517192, 14707, 1187, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:91567084", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 0"},
 CellID->91567084],
Cell[518382, 14745, 987, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:420812829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 3"},
 CellID->420812829],
Cell[519372, 14774, 1703, 53, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:138547187", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 4"},
 CellID->138547187],
Cell[521078, 14829, 1242, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:457378178", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 11"},
 CellID->457378178],
Cell[522323, 14868, 2172, 72, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:926082604", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 11'"},
 CellID->926082604],
Cell[524498, 14942, 1581, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1344734857", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] lp lc"},
 CellID->1344734857],
Cell[526082, 14994, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1125130303]
}, Open  ]],
Cell[526163, 14998, 47, 1, 70, "CloseEnvironment",
 CellID->12089368],
Cell[CellGroupData[{
Cell[526235, 15003, 2197, 67, 70, "ProofInfo",
 CellTags->{"Proof|ID:1368691869", "Proof|ID:1368691869-1"},
 CellID->660760468],
Cell[528435, 15072, 481, 12, 70, "ProofInfoBody",
 CellID->417821554]
}, Closed]],
Cell[CellGroupData[{
Cell[528953, 15089, 3445, 94, 70, "ProofInfo",
 CellTags->{"Proof|ID:941688542", "Proof|ID:941688542-1"},
 CellID->1365249879],
Cell[532401, 15185, 481, 12, 70, "ProofInfoBody",
 CellID->1779137777]
}, Closed]],
Cell[CellGroupData[{
Cell[532919, 15202, 3402, 94, 70, "ProofInfo",
 CellTags->{"Proof|ID:91567084", "Proof|ID:91567084-1"},
 CellID->1862161014],
Cell[536324, 15298, 479, 12, 70, "ProofInfoBody",
 CellID->963217094]
}, Closed]],
Cell[CellGroupData[{
Cell[536840, 15315, 3341, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:420812829", "Proof|ID:420812829-1"},
 CellID->2092842442],
Cell[540184, 15409, 480, 12, 70, "ProofInfoBody",
 CellID->875416490]
}, Closed]],
Cell[CellGroupData[{
Cell[540701, 15426, 4585, 118, 70, "ProofInfo",
 CellTags->{"Proof|ID:138547187", "Proof|ID:138547187-1"},
 CellID->244847721],
Cell[545289, 15546, 481, 12, 70, "ProofInfoBody",
 CellID->2024151403]
}, Closed]],
Cell[CellGroupData[{
Cell[545807, 15563, 3996, 106, 70, "ProofInfo",
 CellTags->{"Proof|ID:457378178", "Proof|ID:457378178-1"},
 CellID->783388721],
Cell[549806, 15671, 523, 13, 70, "ProofInfoBody",
 CellID->1798922519]
}, Closed]],
Cell[CellGroupData[{
Cell[550366, 15689, 5150, 128, 70, "ProofInfo",
 CellTags->{"Proof|ID:926082604", "Proof|ID:926082604-1"},
 CellID->1364381745],
Cell[555519, 15819, 480, 12, 70, "ProofInfoBody",
 CellID->659473488]
}, Closed]],
Cell[CellGroupData[{
Cell[556036, 15836, 4469, 115, 70, "ProofInfo",
 CellTags->{"Proof|ID:1344734857", "Proof|ID:1344734857-1"},
 CellID->670171654],
Cell[560508, 15953, 481, 12, 70, "ProofInfoBody",
 CellID->565919770]
}, Closed]],
Cell[561004, 15968, 48, 1, 70, "OpenEnvironment",
 CellID->1286789207],
Cell[CellGroupData[{
Cell[561077, 15973, 385, 9, 70, "EnvironmentHeader",
 CellID->660836734],
Cell[561465, 15984, 1094, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1558664548", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 1"},
 CellID->1558664548],
Cell[562562, 16018, 903, 24, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1835029788", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1835029788],
Cell[563468, 16044, 945, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:757450222", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial \[NotEqual] 0"},
 CellID->757450222],
Cell[564416, 16069, 921, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1978849375", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial -"},
 CellID->1978849375],
Cell[565340, 16096, 982, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:283455673", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 3'"},
 CellID->283455673],
Cell[566325, 16125, 1438, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:610787620", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 3"},
 CellID->610787620],
Cell[567766, 16169, 1841, 58, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:980082706", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 4"},
 CellID->980082706],
Cell[569610, 16229, 836, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1306217654", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm isMonomial"},
 CellID->1306217654],
Cell[570449, 16254, 1570, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1645795930", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 7"},
 CellID->1645795930],
Cell[572022, 16305, 736, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:691307310", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->691307310],
Cell[572761, 16326, 1778, 54, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1901309763", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 8"},
 CellID->1901309763],
Cell[574542, 16382, 1504, 48, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1177943318", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 9"},
 CellID->1177943318],
Cell[576049, 16432, 721, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1770611523", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1770611523],
Cell[576773, 16453, 1719, 54, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:670623560", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 12"},
 CellID->670623560],
Cell[578495, 16509, 957, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1806313475", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial lm"},
 CellID->1806313475],
Cell[579455, 16536, 964, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1110032163", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lc lm"},
 CellID->1110032163],
Cell[580422, 16566, 961, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:632352665", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp lm"},
 CellID->632352665],
Cell[581386, 16596, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1776835731]
}, Open  ]],
Cell[581467, 16600, 48, 1, 70, "CloseEnvironment",
 CellID->816964609],
Cell[CellGroupData[{
Cell[581540, 16605, 2278, 67, 70, "ProofInfo",
 CellTags->{"Proof|ID:1558664548", "Proof|ID:1558664548-1"},
 CellID->142537034],
Cell[583821, 16674, 481, 12, 70, "ProofInfoBody",
 CellID->311072897]
}, Closed]],
Cell[CellGroupData[{
Cell[584339, 16691, 2789, 80, 70, "ProofInfo",
 CellTags->{"Proof|ID:757450222", "Proof|ID:757450222-1"},
 CellID->1689557017],
Cell[587131, 16773, 481, 12, 70, "ProofInfoBody",
 CellID->1082713634]
}, Closed]],
Cell[CellGroupData[{
Cell[587649, 16790, 2964, 83, 70, "ProofInfo",
 CellTags->{"Proof|ID:1978849375", "Proof|ID:1978849375-1"},
 CellID->807800730],
Cell[590616, 16875, 482, 12, 70, "ProofInfoBody",
 CellID->2025444804]
}, Closed]],
Cell[CellGroupData[{
Cell[591135, 16892, 3438, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:283455673", "Proof|ID:283455673-1"},
 CellID->1547205861],
Cell[594576, 16986, 480, 12, 70, "ProofInfoBody",
 CellID->532363353]
}, Closed]],
Cell[CellGroupData[{
Cell[595093, 17003, 4528, 118, 70, "ProofInfo",
 CellTags->{"Proof|ID:610787620", "Proof|ID:610787620-1"},
 CellID->1422344349],
Cell[599624, 17123, 481, 12, 70, "ProofInfoBody",
 CellID->1123805012]
}, Closed]],
Cell[CellGroupData[{
Cell[600142, 17140, 4986, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:980082706", "Proof|ID:980082706-1"},
 CellID->1540912936],
Cell[605131, 17266, 481, 12, 70, "ProofInfoBody",
 CellID->1183327246]
}, Closed]],
Cell[CellGroupData[{
Cell[605649, 17283, 2949, 85, 70, "ProofInfo",
 CellTags->{"Proof|ID:1306217654", "Proof|ID:1306217654-1"},
 CellID->277734451],
Cell[608601, 17370, 481, 12, 70, "ProofInfoBody",
 CellID->383104478]
}, Closed]],
Cell[CellGroupData[{
Cell[609119, 17387, 4080, 107, 70, "ProofInfo",
 CellTags->{"Proof|ID:1645795930", "Proof|ID:1645795930-1"},
 CellID->140680378],
Cell[613202, 17496, 480, 12, 70, "ProofInfoBody",
 CellID->22238799]
}, Closed]],
Cell[CellGroupData[{
Cell[613719, 17513, 5548, 141, 70, "ProofInfo",
 CellTags->{"Proof|ID:1901309763", "Proof|ID:1901309763-1"},
 CellID->907132124],
Cell[619270, 17656, 481, 12, 70, "ProofInfoBody",
 CellID->812486108]
}, Closed]],
Cell[CellGroupData[{
Cell[619788, 17673, 3681, 99, 70, "ProofInfo",
 CellTags->{"Proof|ID:1177943318", "Proof|ID:1177943318-1"},
 CellID->1237355862],
Cell[623472, 17774, 481, 12, 70, "ProofInfoBody",
 CellID->458207366]
}, Closed]],
Cell[CellGroupData[{
Cell[623990, 17791, 4918, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:670623560", "Proof|ID:670623560-1"},
 CellID->1341250136],
Cell[628911, 17919, 480, 12, 70, "ProofInfoBody",
 CellID->837571024]
}, Closed]],
Cell[CellGroupData[{
Cell[629428, 17936, 2952, 83, 70, "ProofInfo",
 CellTags->{"Proof|ID:1806313475", "Proof|ID:1806313475-1"},
 CellID->347558245],
Cell[632383, 18021, 482, 12, 70, "ProofInfoBody",
 CellID->1954493899]
}, Closed]],
Cell[CellGroupData[{
Cell[632902, 18038, 3332, 93, 70, "ProofInfo",
 CellTags->{"Proof|ID:1110032163", "Proof|ID:1110032163-1"},
 CellID->916272762],
Cell[636237, 18133, 482, 12, 70, "ProofInfoBody",
 CellID->1999682617]
}, Closed]],
Cell[CellGroupData[{
Cell[636756, 18150, 3323, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:632352665", "Proof|ID:632352665-1"},
 CellID->800412793],
Cell[640082, 18244, 481, 12, 70, "ProofInfoBody",
 CellID->1296416971]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[640624, 18263, 299, 8, 30, "Subsection",
 CellID->1730043703],
Cell[640926, 18273, 47, 1, 70, "OpenEnvironment",
 CellID->991999998],
Cell[CellGroupData[{
Cell[640998, 18278, 462, 14, 70, "EnvironmentHeader",
 CellID->13198659],
Cell[641463, 18294, 1000, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:423895762", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ closed \[ScriptCapitalP]"},
 CellID->423895762],
Cell[642466, 18326, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1334309619]
}, Open  ]],
Cell[642547, 18330, 48, 1, 70, "CloseEnvironment",
 CellID->493608360],
Cell[CellGroupData[{
Cell[642620, 18335, 1946, 60, 70, "ProofInfo",
 CellTags->{"Proof|ID:423895762", "Proof|ID:423895762-1"},
 CellID->355044739],
Cell[644569, 18397, 480, 12, 70, "ProofInfoBody",
 CellID->650572530]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[645098, 18415, 318, 9, 30, "Subsection",
 CellID->141138700],
Cell[645419, 18426, 47, 1, 70, "OpenEnvironment",
 CellID->445386680],
Cell[CellGroupData[{
Cell[645491, 18431, 531, 15, 70, "EnvironmentHeader",
 CellID->1064502876],
Cell[646025, 18448, 1108, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1812309492", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "0 neutral \[ScriptCapitalP]"},
 CellID->1812309492],
Cell[647136, 18482, 65, 1, 70, "EndEnvironmentMarker",
 CellID->381592666]
}, Open  ]],
Cell[647216, 18486, 49, 1, 70, "CloseEnvironment",
 CellID->1236557493],
Cell[CellGroupData[{
Cell[647290, 18491, 2078, 63, 70, "ProofInfo",
 CellTags->{"Proof|ID:1812309492", "Proof|ID:1812309492-1"},
 CellID->2145107533],
Cell[649371, 18556, 482, 12, 70, "ProofInfoBody",
 CellID->1403340967]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[649902, 18574, 303, 8, 30, "Subsection",
 CellID->824892980],
Cell[650208, 18584, 47, 1, 70, "OpenEnvironment",
 CellID->803927450],
Cell[CellGroupData[{
Cell[650280, 18589, 518, 15, 70, "EnvironmentHeader",
 CellID->1992468669],
Cell[650801, 18606, 1048, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:595518277", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ assoc \[ScriptCapitalP]"},
 CellID->595518277],
Cell[651852, 18639, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1198061607]
}, Open  ]],
Cell[651933, 18643, 49, 1, 70, "CloseEnvironment",
 CellID->1614517617],
Cell[CellGroupData[{
Cell[652007, 18648, 1943, 60, 70, "ProofInfo",
 CellTags->{"Proof|ID:595518277", "Proof|ID:595518277-1"},
 CellID->1392319518],
Cell[653953, 18710, 481, 12, 70, "ProofInfoBody",
 CellID->1240957880]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[654483, 18728, 306, 8, 30, "Subsection",
 CellID->1752381855],
Cell[654792, 18738, 48, 1, 70, "OpenEnvironment",
 CellID->1316847979],
Cell[CellGroupData[{
Cell[654865, 18743, 519, 15, 70, "EnvironmentHeader",
 CellID->581912688],
Cell[655387, 18760, 1046, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2114498624", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "+ comm \[ScriptCapitalP]"},
 CellID->2114498624],
Cell[656436, 18793, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1729399905]
}, Open  ]],
Cell[656517, 18797, 49, 1, 70, "CloseEnvironment",
 CellID->1955171682],
Cell[CellGroupData[{
Cell[656591, 18802, 1948, 61, 70, "ProofInfo",
 CellTags->{"Proof|ID:2114498624", "Proof|ID:2114498624-1"},
 CellID->2065823929],
Cell[658542, 18865, 481, 12, 70, "ProofInfoBody",
 CellID->545810919]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[659072, 18883, 307, 10, 30, "Subsection",
 CellID->348883692],
Cell[659382, 18895, 48, 1, 70, "OpenEnvironment",
 CellID->1657021632],
Cell[CellGroupData[{
Cell[659455, 18900, 570, 18, 70, "EnvironmentHeader",
 CellID->507868481],
Cell[660028, 18920, 763, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:246227628", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->246227628],
Cell[660794, 18942, 1007, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:95118700", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP] 1"},
 CellID->95118700],
Cell[661804, 18972, 897, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1900465816", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP] 2"},
 CellID->1900465816],
Cell[662704, 18997, 1240, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2037231548", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- inverse \[ScriptCapitalP]"},
 CellID->2037231548],
Cell[663947, 19038, 65, 1, 70, "EndEnvironmentMarker",
 CellID->150491788]
}, Open  ]],
Cell[664027, 19042, 48, 1, 70, "CloseEnvironment",
 CellID->983349670],
Cell[CellGroupData[{
Cell[664100, 19047, 3423, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:95118700", "Proof|ID:95118700-1"},
 CellID->49996501],
Cell[667526, 19141, 480, 12, 70, "ProofInfoBody",
 CellID->2105700149]
}, Closed]],
Cell[CellGroupData[{
Cell[668043, 19158, 2682, 79, 70, "ProofInfo",
 CellTags->{"Proof|ID:1900465816", "Proof|ID:1900465816-1"},
 CellID->98966748],
Cell[670728, 19239, 481, 12, 70, "ProofInfoBody",
 CellID->403404329]
}, Closed]],
Cell[CellGroupData[{
Cell[671246, 19256, 2204, 65, 70, "ProofInfo",
 CellTags->{"Proof|ID:2037231548", "Proof|ID:2037231548-1"},
 CellID->1554728169],
Cell[673453, 19323, 482, 12, 70, "ProofInfoBody",
 CellID->1256029176]
}, Closed]],
Cell[673950, 19338, 47, 1, 70, "OpenEnvironment",
 CellID->394685747],
Cell[CellGroupData[{
Cell[674022, 19343, 338, 9, 70, "EnvironmentHeader",
 CellID->404012057],
Cell[674363, 19354, 918, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:145364866", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- 0 \[ScriptCapitalP] 1"},
 CellID->145364866],
Cell[675284, 19380, 1130, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:117883331", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "- 0 \[ScriptCapitalP] 2"},
 CellID->117883331],
Cell[676417, 19415, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1053861784]
}, Open  ]],
Cell[676498, 19419, 49, 1, 70, "CloseEnvironment",
 CellID->2119922305],
Cell[CellGroupData[{
Cell[676572, 19424, 2104, 64, 70, "ProofInfo",
 CellTags->{"Proof|ID:145364866", "Proof|ID:145364866-1"},
 CellID->448617855],
Cell[678679, 19490, 481, 12, 70, "ProofInfoBody",
 CellID->1296800118]
}, Closed]],
Cell[CellGroupData[{
Cell[679197, 19507, 2663, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:117883331", "Proof|ID:117883331-1"},
 CellID->1580290873],
Cell[681863, 19587, 481, 12, 70, "ProofInfoBody",
 CellID->1688210865]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[682393, 19605, 249, 6, 30, "Subsection",
 CellID->1482174164],
Cell[CellGroupData[{
Cell[682667, 19615, 343, 9, 28, "Subsubsection",
 CellID->1081324154],
Cell[683013, 19626, 47, 1, 70, "OpenEnvironment",
 CellID->884155321],
Cell[CellGroupData[{
Cell[683085, 19631, 532, 16, 70, "EnvironmentHeader",
 CellID->763838058],
Cell[683620, 19649, 1059, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:771361293", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCCMonoid \[ScriptCapitalP]"},
 CellID->771361293],
Cell[684682, 19682, 65, 1, 70, "EndEnvironmentMarker",
 CellID->188406047]
}, Open  ]],
Cell[684762, 19686, 49, 1, 70, "CloseEnvironment",
 CellID->1310768700],
Cell[CellGroupData[{
Cell[684836, 19691, 2063, 61, 70, "ProofInfo",
 CellTags->{"Proof|ID:771361293", "Proof|ID:771361293-1"},
 CellID->556042782],
Cell[686902, 19754, 480, 12, 70, "ProofInfoBody",
 CellID->472383758]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[687431, 19772, 323, 5, 22, "Subsubsection",
 CellID->2092883158],
Cell[687757, 19779, 47, 1, 70, "OpenEnvironment",
 CellID->947786111],
Cell[CellGroupData[{
Cell[687829, 19784, 645, 18, 70, "EnvironmentHeader",
 CellID->2067413038],
Cell[688477, 19804, 842, 22, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1221832846", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1221832846],
Cell[689322, 19828, 944, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:313886646", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership R", "IR"},
 CellID->313886646],
Cell[690269, 19856, 952, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:174448700", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lm+R"},
 CellID->174448700],
Cell[691224, 19885, 1735, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1679470623", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C R"},
 CellID->1679470623],
Cell[692962, 19938, 1144, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1648446612", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 1"},
 CellID->1648446612],
Cell[694109, 19973, 1672, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1799775671", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 2"},
 CellID->1799775671],
Cell[695784, 20025, 1024, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1065304756", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 6"},
 CellID->1065304756],
Cell[696811, 20056, 1534, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1078436809", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 7"},
 CellID->1078436809],
Cell[698348, 20107, 1061, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:735362349", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 8"},
 CellID->735362349],
Cell[699412, 20141, 1062, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:451768283", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 8'"},
 CellID->451768283],
Cell[700477, 20175, 2072, 66, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:739591353", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 5'"},
 CellID->739591353],
Cell[702552, 20243, 796, 22, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1347176402", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1347176402],
Cell[703351, 20267, 1222, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1377108879", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 3"},
 CellID->1377108879],
Cell[704576, 20300, 1765, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:139632107", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 4"},
 CellID->139632107],
Cell[706344, 20358, 1030, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2132986120", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 7"},
 CellID->2132986120],
Cell[707377, 20389, 1103, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1911280389", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 8"},
 CellID->1911280389],
Cell[708483, 20422, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1601807717]
}, Open  ]],
Cell[708564, 20426, 49, 1, 70, "CloseEnvironment",
 CellID->1176230318],
Cell[CellGroupData[{
Cell[708638, 20431, 2939, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:313886646", "Proof|ID:313886646-1"},
 CellID->368790604],
Cell[711580, 20515, 481, 12, 70, "ProofInfoBody",
 CellID->1023248069]
}, Closed]],
Cell[CellGroupData[{
Cell[712098, 20532, 3388, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:174448700", "Proof|ID:174448700-1"},
 CellID->1932131695],
Cell[715489, 20626, 480, 12, 70, "ProofInfoBody",
 CellID->632362462]
}, Closed]],
Cell[CellGroupData[{
Cell[716006, 20643, 5111, 128, 70, "ProofInfo",
 CellTags->{"Proof|ID:1679470623", "Proof|ID:1679470623-1"},
 CellID->1796196826],
Cell[721120, 20773, 482, 12, 70, "ProofInfoBody",
 CellID->1707284078]
}, Closed]],
Cell[CellGroupData[{
Cell[721639, 20790, 2439, 73, 70, "ProofInfo",
 CellTags->{"Proof|ID:1648446612", "Proof|ID:1648446612-1"},
 CellID->822584714],
Cell[724081, 20865, 481, 12, 70, "ProofInfoBody",
 CellID->580974147]
}, Closed]],
Cell[CellGroupData[{
Cell[724599, 20882, 5059, 131, 70, "ProofInfo",
 CellTags->{"Proof|ID:1799775671", "Proof|ID:1799775671-1"},
 CellID->1365150475],
Cell[729661, 21015, 482, 12, 70, "ProofInfoBody",
 CellID->1394382683]
}, Closed]],
Cell[CellGroupData[{
Cell[730180, 21032, 3543, 99, 70, "ProofInfo",
 CellTags->{"Proof|ID:1065304756", "Proof|ID:1065304756-1"},
 CellID->364956431],
Cell[733726, 21133, 482, 12, 70, "ProofInfoBody",
 CellID->1722541450]
}, Closed]],
Cell[CellGroupData[{
Cell[734245, 21150, 4531, 117, 70, "ProofInfo",
 CellTags->{"Proof|ID:1078436809", "Proof|ID:1078436809-1"},
 CellID->1581073290],
Cell[738779, 21269, 480, 12, 70, "ProofInfoBody",
 CellID->45607130]
}, Closed]],
Cell[CellGroupData[{
Cell[739296, 21286, 3299, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:735362349", "Proof|ID:735362349-1"},
 CellID->229453058],
Cell[742598, 21378, 480, 12, 70, "ProofInfoBody",
 CellID->647018837]
}, Closed]],
Cell[CellGroupData[{
Cell[743115, 21395, 3306, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:451768283", "Proof|ID:451768283-1"},
 CellID->497176783],
Cell[746424, 21489, 479, 12, 70, "ProofInfoBody",
 CellID->86322822]
}, Closed]],
Cell[CellGroupData[{
Cell[746940, 21506, 5334, 136, 70, "ProofInfo",
 CellTags->{"Proof|ID:739591353", "Proof|ID:739591353-1"},
 CellID->1363695772],
Cell[752277, 21644, 480, 12, 70, "ProofInfoBody",
 CellID->883221224]
}, Closed]],
Cell[CellGroupData[{
Cell[752794, 21661, 3978, 105, 70, "ProofInfo",
 CellTags->{"Proof|ID:1377108879", "Proof|ID:1377108879-1"},
 CellID->360514177],
Cell[756775, 21768, 482, 12, 70, "ProofInfoBody",
 CellID->1351144559]
}, Closed]],
Cell[CellGroupData[{
Cell[757294, 21785, 5860, 148, 70, "ProofInfo",
 CellTags->{"Proof|ID:139632107", "Proof|ID:139632107-1"},
 CellID->1587172171],
Cell[763157, 21935, 481, 12, 70, "ProofInfoBody",
 CellID->1470107478]
}, Closed]],
Cell[CellGroupData[{
Cell[763675, 21952, 4130, 111, 70, "ProofInfo",
 CellTags->{"Proof|ID:2132986120", "Proof|ID:2132986120-1"},
 CellID->2053017099],
Cell[767808, 22065, 481, 12, 70, "ProofInfoBody",
 CellID->448185845]
}, Closed]],
Cell[CellGroupData[{
Cell[768326, 22082, 4329, 115, 70, "ProofInfo",
 CellTags->{"Proof|ID:1911280389", "Proof|ID:1911280389-1"},
 CellID->472365930],
Cell[772658, 22199, 482, 12, 70, "ProofInfoBody",
 CellID->1273675156]
}, Closed]],
Cell[773155, 22214, 48, 1, 70, "OpenEnvironment",
 CellID->1812050497],
Cell[CellGroupData[{
Cell[773228, 22219, 439, 13, 70, "EnvironmentHeader",
 CellID->2059822108],
Cell[773670, 22234, 743, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1031528829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1031528829],
Cell[774416, 22255, 817, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1494855415", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1494855415],
Cell[775236, 22278, 1538, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2007119699", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership H ind"},
 CellID->2007119699],
Cell[776777, 22324, 917, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:884488101", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership H", "IR"},
 CellID->884488101],
Cell[777697, 22351, 2154, 66, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1892666968", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C H ind"},
 CellID->1892666968],
Cell[779854, 22419, 1687, 52, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1556838587", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C H"},
 CellID->1556838587],
Cell[781544, 22473, 1612, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1768827306", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H=0"},
 CellID->1768827306],
Cell[783159, 22524, 1783, 57, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:140826492", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H=H"},
 CellID->140826492],
Cell[784945, 22583, 1161, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1627546251", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 1"},
 CellID->1627546251],
Cell[786109, 22619, 1056, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:743145124", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 2"},
 CellID->743145124],
Cell[787168, 22652, 1163, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1576432658", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 3"},
 CellID->1576432658],
Cell[788334, 22688, 1591, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1977357773", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 4"},
 CellID->1977357773],
Cell[789928, 22739, 1269, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:561821259", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 4'"},
 CellID->561821259],
Cell[791200, 22778, 2151, 66, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:866138121", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 5"},
 CellID->866138121],
Cell[793354, 22846, 2071, 62, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:105044014", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 5'"},
 CellID->105044014],
Cell[795428, 22910, 1611, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1971276224", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 6"},
 CellID->1971276224],
Cell[797042, 22961, 744, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:219245057", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->219245057],
Cell[797789, 22983, 953, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2135590264", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 7"},
 CellID->2135590264],
Cell[798745, 23012, 1214, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2092106183", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 8"},
 CellID->2092106183],
Cell[799962, 23050, 793, 22, 70, "GlobalDeclaration",
 CellTags->{
  "ID:831355566", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->831355566],
Cell[800758, 23074, 1112, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2014193763", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 10"},
 CellID->2014193763],
Cell[801873, 23108, 1137, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1650576647", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 11"},
 CellID->1650576647],
Cell[803013, 23143, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1728877054]
}, Open  ]],
Cell[803094, 23147, 48, 1, 70, "CloseEnvironment",
 CellID->596281154],
Cell[CellGroupData[{
Cell[803167, 23152, 4314, 111, 70, "ProofInfo",
 CellTags->{"Proof|ID:2007119699", "Proof|ID:2007119699-1"},
 CellID->526673768],
Cell[807484, 23265, 524, 13, 70, "ProofInfoBody",
 CellID->1470696438]
}, Closed]],
Cell[CellGroupData[{
Cell[808045, 23283, 3040, 84, 70, "ProofInfo",
 CellTags->{"Proof|ID:884488101", "Proof|ID:884488101-1"},
 CellID->1564315886],
Cell[811088, 23369, 480, 12, 70, "ProofInfoBody",
 CellID->375951896]
}, Closed]],
Cell[CellGroupData[{
Cell[811605, 23386, 6278, 152, 70, "ProofInfo",
 CellTags->{"Proof|ID:1892666968", "Proof|ID:1892666968-1"},
 CellID->800813603],
Cell[817886, 23540, 481, 12, 70, "ProofInfoBody",
 CellID->119379889]
}, Closed]],
Cell[CellGroupData[{
Cell[818404, 23557, 4950, 125, 70, "ProofInfo",
 CellTags->{"Proof|ID:1556838587", "Proof|ID:1556838587-1"},
 CellID->293752933],
Cell[823357, 23684, 482, 12, 70, "ProofInfoBody",
 CellID->1224452790]
}, Closed]],
Cell[CellGroupData[{
Cell[823876, 23701, 4849, 127, 70, "ProofInfo",
 CellTags->{"Proof|ID:1768827306", "Proof|ID:1768827306-1"},
 CellID->864932006],
Cell[828728, 23830, 482, 12, 70, "ProofInfoBody",
 CellID->1209343178]
}, Closed]],
Cell[CellGroupData[{
Cell[829247, 23847, 5867, 146, 70, "ProofInfo",
 CellTags->{"Proof|ID:140826492", "Proof|ID:140826492-1"},
 CellID->743847445],
Cell[835117, 23995, 481, 12, 70, "ProofInfoBody",
 CellID->1288942961]
}, Closed]],
Cell[CellGroupData[{
Cell[835635, 24012, 4603, 119, 70, "ProofInfo",
 CellTags->{"Proof|ID:1627546251", "Proof|ID:1627546251-1"},
 CellID->1563876245],
Cell[840241, 24133, 481, 12, 70, "ProofInfoBody",
 CellID->670928490]
}, Closed]],
Cell[CellGroupData[{
Cell[840759, 24150, 3757, 100, 70, "ProofInfo",
 CellTags->{"Proof|ID:743145124", "Proof|ID:743145124-1"},
 CellID->729988509],
Cell[844519, 24252, 480, 12, 70, "ProofInfoBody",
 CellID->122686757]
}, Closed]],
Cell[CellGroupData[{
Cell[845036, 24269, 4669, 121, 70, "ProofInfo",
 CellTags->{"Proof|ID:1576432658", "Proof|ID:1576432658-1"},
 CellID->478962428],
Cell[849708, 24392, 482, 12, 70, "ProofInfoBody",
 CellID->1090556569]
}, Closed]],
Cell[CellGroupData[{
Cell[850227, 24409, 4602, 119, 70, "ProofInfo",
 CellTags->{"Proof|ID:1977357773", "Proof|ID:1977357773-1"},
 CellID->518253128],
Cell[854832, 24530, 481, 12, 70, "ProofInfoBody",
 CellID->696944507]
}, Closed]],
Cell[CellGroupData[{
Cell[855350, 24547, 4089, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:561821259", "Proof|ID:561821259-1"},
 CellID->1734326259],
Cell[859442, 24657, 480, 12, 70, "ProofInfoBody",
 CellID->728909696]
}, Closed]],
Cell[CellGroupData[{
Cell[859959, 24674, 6749, 162, 70, "ProofInfo",
 CellTags->{"Proof|ID:866138121", "Proof|ID:866138121-1"},
 CellID->726143767],
Cell[866711, 24838, 481, 12, 70, "ProofInfoBody",
 CellID->2015149483]
}, Closed]],
Cell[CellGroupData[{
Cell[867229, 24855, 6473, 154, 70, "ProofInfo",
 CellTags->{"Proof|ID:105044014", "Proof|ID:105044014-1"},
 CellID->839817768],
Cell[873705, 25011, 481, 12, 70, "ProofInfoBody",
 CellID->1916847501]
}, Closed]],
Cell[CellGroupData[{
Cell[874223, 25028, 4575, 119, 70, "ProofInfo",
 CellTags->{"Proof|ID:1971276224", "Proof|ID:1971276224-1"},
 CellID->611586869],
Cell[878801, 25149, 481, 12, 70, "ProofInfoBody",
 CellID->587830886]
}, Closed]],
Cell[CellGroupData[{
Cell[879319, 25166, 3338, 95, 70, "ProofInfo",
 CellTags->{"Proof|ID:2135590264", "Proof|ID:2135590264-1"},
 CellID->88523793],
Cell[882660, 25263, 481, 12, 70, "ProofInfoBody",
 CellID->828333552]
}, Closed]],
Cell[CellGroupData[{
Cell[883178, 25280, 4115, 109, 70, "ProofInfo",
 CellTags->{"Proof|ID:2092106183", "Proof|ID:2092106183-1"},
 CellID->157583881],
Cell[887296, 25391, 482, 12, 70, "ProofInfoBody",
 CellID->1527934028]
}, Closed]],
Cell[CellGroupData[{
Cell[887815, 25408, 4341, 113, 70, "ProofInfo",
 CellTags->{"Proof|ID:2014193763", "Proof|ID:2014193763-1"},
 CellID->1352301682],
Cell[892159, 25523, 482, 12, 70, "ProofInfoBody",
 CellID->1444998026]
}, Closed]],
Cell[CellGroupData[{
Cell[892678, 25540, 4551, 117, 70, "ProofInfo",
 CellTags->{"Proof|ID:1650576647", "Proof|ID:1650576647-1"},
 CellID->1429121459],
Cell[897232, 25659, 481, 12, 70, "ProofInfoBody",
 CellID->637145231]
}, Closed]],
Cell[897728, 25674, 873, 16, 70, "Text",
 CellID->1413316893],
Cell[898604, 25692, 48, 1, 70, "OpenEnvironment",
 CellID->1924653495],
Cell[CellGroupData[{
Cell[898677, 25697, 396, 9, 70, "EnvironmentHeader",
 CellID->758824004],
Cell[899076, 25708, 641, 16, 70, "GlobalDeclaration",
 CellTags->{
  "ID:45249416", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->45249416],
Cell[899720, 25726, 3991, 108, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:248026320", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "induction on \[ScriptCapitalP] ind"},
 CellID->248026320],
Cell[903714, 25836, 1918, 59, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1502331257", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "induction on \[ScriptCapitalP]", "IR"},
 CellID->1502331257],
Cell[905635, 25897, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1647825250]
}, Open  ]],
Cell[905716, 25901, 48, 1, 70, "CloseEnvironment",
 CellID->172493216],
Cell[CellGroupData[{
Cell[905789, 25906, 9001, 211, 70, "ProofInfo",
 CellTags->{"Proof|ID:248026320", "Proof|ID:248026320-1"},
 CellID->717333436],
Cell[914793, 26119, 480, 12, 70, "ProofInfoBody",
 CellID->680214968]
}, Closed]],
Cell[CellGroupData[{
Cell[915310, 26136, 5045, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:1502331257", "Proof|ID:1502331257-1"},
 CellID->2113840420],
Cell[920358, 26270, 482, 12, 70, "ProofInfoBody",
 CellID->1986041523]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[920889, 26288, 185, 3, 22, "Subsubsection",
 CellID->1456829087],
Cell[921077, 26293, 48, 1, 70, "OpenEnvironment",
 CellID->1742278695],
Cell[CellGroupData[{
Cell[921150, 26298, 388, 9, 70, "EnvironmentHeader",
 CellID->1985392776],
Cell[921541, 26309, 766, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:411993095", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->411993095],
Cell[922310, 26331, 1725, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:470022203", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 1"},
 CellID->470022203],
Cell[924038, 26382, 847, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:401540586", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->401540586],
Cell[924888, 26407, 1468, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2014025290", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 6"},
 CellID->2014025290],
Cell[926359, 26448, 870, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1590906878", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 10"},
 CellID->1590906878],
Cell[927232, 26473, 1279, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1997927583", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 11"},
 CellID->1997927583],
Cell[928514, 26513, 66, 1, 70, "EndEnvironmentMarker",
 CellID->2121668319]
}, Open  ]],
Cell[928595, 26517, 48, 1, 70, "CloseEnvironment",
 CellID->859850637],
Cell[CellGroupData[{
Cell[928668, 26522, 4412, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:470022203", "Proof|ID:470022203-1"},
 CellID->356771522],
Cell[933083, 26638, 481, 12, 70, "ProofInfoBody",
 CellID->2104860677]
}, Closed]],
Cell[CellGroupData[{
Cell[933601, 26655, 4838, 121, 70, "ProofInfo",
 CellTags->{"Proof|ID:2014025290", "Proof|ID:2014025290-1"},
 CellID->740297765],
Cell[938442, 26778, 481, 12, 70, "ProofInfoBody",
 CellID->595415687]
}, Closed]],
Cell[CellGroupData[{
Cell[938960, 26795, 2958, 85, 70, "ProofInfo",
 CellTags->{"Proof|ID:1590906878", "Proof|ID:1590906878-1"},
 CellID->1050945896],
Cell[941921, 26882, 482, 12, 70, "ProofInfoBody",
 CellID->1718488369]
}, Closed]],
Cell[CellGroupData[{
Cell[942440, 26899, 4105, 109, 70, "ProofInfo",
 CellTags->{"Proof|ID:1997927583", "Proof|ID:1997927583-1"},
 CellID->646926070],
Cell[946548, 27010, 482, 12, 70, "ProofInfoBody",
 CellID->1129892272]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[947079, 27028, 200, 3, 22, "Subsubsection",
 CellID->223041332],
Cell[947282, 27033, 48, 1, 12, "OpenEnvironment",
 CellID->1774552902],
Cell[CellGroupData[{
Cell[947355, 27038, 460, 11, 43, "EnvironmentHeader",
 CellID->1659339968],
Cell[947818, 27051, 932, 24, 66, "GlobalDeclaration",
 CellTags->{
  "ID:63039386", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->63039386],
Cell[948753, 27077, 261, 5, 24, "EnvironmentText",
 CellID->1091890068],
Cell[949017, 27084, 926, 24, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:1093453827", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 0"},
 CellID->1093453827],
Cell[949946, 27110, 1902, 54, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1795167993", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 5"},
 CellID->1795167993],
Cell[951851, 27166, 1575, 45, 76, "FormalTextInputFormula",
 CellTags->{
  "ID:838653570", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomial multiplication ind"},
 CellID->838653570],
Cell[953429, 27213, 987, 27, 34, "FormalTextInputFormula",
 CellTags->{
  "ID:1903131381", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "membership monomial multiplication", "IR"},
 CellID->1903131381],
Cell[954419, 27242, 2017, 58, 64, "FormalTextInputFormula",
 CellTags->{
  "ID:1641583144", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C \[CircleTimes]"},
 CellID->1641583144],
Cell[956439, 27302, 1233, 35, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:608153969", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 1"},
 CellID->608153969],
Cell[957675, 27339, 1673, 48, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1439631766", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] isMonomial"},
 CellID->1439631766],
Cell[959351, 27389, 736, 19, 42, "GlobalDeclaration",
 CellTags->{
  "ID:607306392", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->607306392],
Cell[960090, 27410, 1163, 32, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:1312761442", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 2"},
 CellID->1312761442],
Cell[961256, 27444, 1345, 41, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:259600169", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 3"},
 CellID->259600169],
Cell[962604, 27487, 999, 27, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:1576857420", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 6"},
 CellID->1576857420],
Cell[963606, 27516, 1844, 55, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:483781382", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 10"},
 CellID->483781382],
Cell[965453, 27573, 1607, 46, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:169057028", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 11"},
 CellID->169057028],
Cell[967063, 27621, 721, 19, 31, "GlobalDeclaration",
 CellTags->{
  "ID:324898493", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->324898493],
Cell[967787, 27642, 1332, 39, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:684947841", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 7"},
 CellID->684947841],
Cell[969122, 27683, 1492, 43, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:1856294157", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 8"},
 CellID->1856294157],
Cell[970617, 27728, 1138, 31, 34, "FormalTextInputFormula",
 CellTags->{
  "ID:1768460531", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CircleTimes] 9"},
 CellID->1768460531],
Cell[971758, 27761, 65, 1, 26, "EndEnvironmentMarker",
 CellID->891227303]
}, Open  ]],
Cell[971838, 27765, 48, 1, 16, "CloseEnvironment",
 CellID->885827699],
Cell[CellGroupData[{
Cell[971911, 27770, 3078, 87, 28, "ProofInfo",
 CellTags->{"Proof|ID:1093453827", "Proof|ID:1093453827-1"},
 CellID->1691805525],
Cell[974992, 27859, 482, 12, 70, "ProofInfoBody",
 CellID->1911765051]
}, Closed]],
Cell[CellGroupData[{
Cell[975511, 27876, 6284, 150, 26, "ProofInfo",
 CellTags->{"Proof|ID:1795167993", "Proof|ID:1795167993-1"},
 CellID->1536044778],
Cell[981798, 28028, 482, 12, 70, "ProofInfoBody",
 CellID->1091089291]
}, Closed]],
Cell[CellGroupData[{
Cell[982317, 28045, 4620, 116, 26, "ProofInfo",
 CellTags->{"Proof|ID:838653570", "Proof|ID:838653570-1"},
 CellID->39930568],
Cell[986940, 28163, 480, 12, 70, "ProofInfoBody",
 CellID->767866762]
}, Closed]],
Cell[CellGroupData[{
Cell[987457, 28180, 3424, 92, 26, "ProofInfo",
 CellTags->{"Proof|ID:1903131381", "Proof|ID:1903131381-1"},
 CellID->443389413],
Cell[990884, 28274, 481, 12, 70, "ProofInfoBody",
 CellID->111366806]
}, Closed]],
Cell[CellGroupData[{
Cell[991402, 28291, 6602, 156, 26, "ProofInfo",
 CellTags->{"Proof|ID:1641583144", "Proof|ID:1641583144-1"},
 CellID->183185032],
Cell[998007, 28449, 481, 12, 70, "ProofInfoBody",
 CellID->188296615]
}, Closed]],
Cell[CellGroupData[{
Cell[998525, 28466, 3069, 86, 26, "ProofInfo",
 CellTags->{"Proof|ID:608153969", "Proof|ID:608153969-1"},
 CellID->1418277402],
Cell[1001597, 28554, 480, 12, 70, "ProofInfoBody",
 CellID->495293771]
}, Closed]],
Cell[CellGroupData[{
Cell[1002114, 28571, 5467, 134, 26, "ProofInfo",
 CellTags->{"Proof|ID:1439631766", "Proof|ID:1439631766-1"},
 CellID->821863036],
Cell[1007584, 28707, 482, 12, 70, "ProofInfoBody",
 CellID->1360721819]
}, Closed]],
Cell[CellGroupData[{
Cell[1008103, 28724, 4865, 123, 26, "ProofInfo",
 CellTags->{"Proof|ID:1312761442", "Proof|ID:1312761442-1"},
 CellID->1943269854],
Cell[1012971, 28849, 481, 12, 70, "ProofInfoBody",
 CellID->405672243]
}, Closed]],
Cell[CellGroupData[{
Cell[1013489, 28866, 4372, 112, 26, "ProofInfo",
 CellTags->{"Proof|ID:259600169", "Proof|ID:259600169-1"},
 CellID->592189180],
Cell[1017864, 28980, 481, 12, 70, "ProofInfoBody",
 CellID->1512486922]
}, Closed]],
Cell[CellGroupData[{
Cell[1018382, 28997, 3868, 103, 26, "ProofInfo",
 CellTags->{"Proof|ID:1576857420", "Proof|ID:1576857420-1"},
 CellID->1943524802],
Cell[1022253, 29102, 481, 12, 70, "ProofInfoBody",
 CellID->141376629]
}, Closed]],
Cell[CellGroupData[{
Cell[1022771, 29119, 5985, 148, 26, "ProofInfo",
 CellTags->{"Proof|ID:483781382", "Proof|ID:483781382-1"},
 CellID->507841304],
Cell[1028759, 29269, 481, 12, 70, "ProofInfoBody",
 CellID->1990411776]
}, Closed]],
Cell[CellGroupData[{
Cell[1029277, 29286, 5538, 136, 26, "ProofInfo",
 CellTags->{"Proof|ID:169057028", "Proof|ID:169057028-1"},
 CellID->1561081111],
Cell[1034818, 29424, 481, 12, 70, "ProofInfoBody",
 CellID->1578160185]
}, Closed]],
Cell[CellGroupData[{
Cell[1035336, 29441, 5004, 124, 26, "ProofInfo",
 CellTags->{"Proof|ID:684947841", "Proof|ID:684947841-1"},
 CellID->969684081],
Cell[1040343, 29567, 481, 12, 70, "ProofInfoBody",
 CellID->1166670995]
}, Closed]],
Cell[CellGroupData[{
Cell[1040861, 29584, 5622, 137, 26, "ProofInfo",
 CellTags->{"Proof|ID:1856294157", "Proof|ID:1856294157-1"},
 CellID->1121162524],
Cell[1046486, 29723, 548, 13, 70, "ProofInfoBody",
 CellID->2088302303]
}, Closed]],
Cell[CellGroupData[{
Cell[1047071, 29741, 4397, 111, 26, "ProofInfo",
 CellTags->{"Proof|ID:1768460531", "Proof|ID:1768460531-1"},
 CellID->1517299564],
Cell[1051471, 29854, 481, 12, 70, "ProofInfoBody",
 CellID->174900675]
}, Closed]],
Cell[CellGroupData[{
Cell[1051989, 29871, 435, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1347606622],
Cell[1052427, 29883, 312, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->905379727],
Cell[1052742, 29890, 571, 15, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2044207238],
Cell[1053316, 29907, 803, 20, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1652777587],
Cell[1054122, 29929, 672, 17, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1555341482],
Cell[1054797, 29948, 796, 19, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->658700090],
Cell[1055596, 29969, 18716, 594, 1585, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->862219043],
Cell[1074315, 30565, 317, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1027179770]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1074681, 30577, 223, 5, 22, "Subsubsection",
 CellID->1591556425],
Cell[1074907, 30584, 47, 1, 70, "OpenEnvironment",
 CellID->742162515],
Cell[CellGroupData[{
Cell[1074979, 30589, 386, 9, 70, "EnvironmentHeader",
 CellID->1404385962],
Cell[1075368, 30600, 5057, 136, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1437523678", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplicationRec"},
 CellID->1437523678],
Cell[1080428, 30738, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1332405579]
}, Open  ]],
Cell[1080509, 30742, 48, 1, 70, "CloseEnvironment",
 CellID->593624661],
Cell[1080560, 30745, 48, 1, 70, "OpenEnvironment",
 CellID->1778102875],
Cell[CellGroupData[{
Cell[1080633, 30750, 478, 14, 70, "EnvironmentHeader",
 CellID->692234298],
Cell[1081114, 30766, 710, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1372918741", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1372918741],
Cell[1081827, 30786, 1824, 55, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:271668948", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "316"},
 CellID->271668948],
Cell[1083654, 30843, 1263, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:428541408", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "317"},
 CellID->428541408],
Cell[1084920, 30883, 800, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1983432019", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "318"},
 CellID->1983432019],
Cell[1085723, 30907, 65, 1, 70, "EndEnvironmentMarker",
 CellID->183027490]
}, Open  ]],
Cell[1085803, 30911, 49, 1, 70, "CloseEnvironment",
 CellID->1206502385],
Cell[CellGroupData[{
Cell[1085877, 30916, 5696, 142, 70, "ProofInfo",
 CellTags->{"Proof|ID:271668948", "Proof|ID:271668948-1"},
 CellID->1133492443],
Cell[1091576, 31060, 481, 12, 70, "ProofInfoBody",
 CellID->2056995531]
}, Closed]],
Cell[CellGroupData[{
Cell[1092094, 31077, 3896, 104, 70, "ProofInfo",
 CellTags->{"Proof|ID:428541408", "Proof|ID:428541408-1"},
 CellID->738256620],
Cell[1095993, 31183, 481, 12, 70, "ProofInfoBody",
 CellID->1481919051]
}, Closed]],
Cell[CellGroupData[{
Cell[1096511, 31200, 2016, 63, 70, "ProofInfo",
 CellTags->{"Proof|ID:1983432019", "Proof|ID:1983432019-1"},
 CellID->836120509],
Cell[1098530, 31265, 482, 12, 70, "ProofInfoBody",
 CellID->1178631419]
}, Closed]],
Cell[1099027, 31280, 925, 20, 70, "Text",
 CellID->2109312154],
Cell[1099955, 31302, 47, 1, 70, "OpenEnvironment",
 CellID->814159960],
Cell[CellGroupData[{
Cell[1100027, 31307, 633, 16, 70, "EnvironmentHeader",
 CellID->1131151213],
Cell[1100663, 31325, 1567, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1310550300", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "319"},
 CellID->1310550300],
Cell[1102233, 31371, 66, 1, 70, "EndEnvironmentMarker",
 CellID->2135849630]
}, Open  ]],
Cell[1102314, 31375, 49, 1, 70, "CloseEnvironment",
 CellID->1678919163]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1102412, 31382, 305, 8, 30, "Subsection",
 CellID->1254493287],
Cell[1102720, 31392, 221, 5, 70, "Text",
 CellID->274886768],
Cell[1102944, 31399, 48, 1, 70, "OpenEnvironment",
 CellID->1237337668],
Cell[CellGroupData[{
Cell[1103017, 31404, 414, 11, 70, "EnvironmentHeader",
 CellID->1046764793],
Cell[1103434, 31417, 835, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:978297022", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "hasMultiplication \[ScriptCapitalP]"},
 CellID->978297022],
Cell[1104272, 31440, 65, 1, 70, "EndEnvironmentMarker",
 CellID->281330312]
}, Open  ]],
Cell[1104352, 31444, 49, 1, 70, "CloseEnvironment",
 CellID->1221052070],
Cell[1104404, 31447, 48, 1, 70, "OpenEnvironment",
 CellID->1875157683],
Cell[CellGroupData[{
Cell[1104477, 31452, 400, 10, 70, "EnvironmentHeader",
 CellID->1672617545],
Cell[1104880, 31464, 1495, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1125626334", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "320"},
 CellID->1125626334],
Cell[1106378, 31503, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1058712053]
}, Open  ]],
Cell[1106459, 31507, 48, 1, 70, "CloseEnvironment",
 CellID->943462036],
Cell[CellGroupData[{
Cell[1106532, 31512, 2944, 83, 70, "ProofInfo",
 CellTags->{"Proof|ID:1125626334", "Proof|ID:1125626334-1"},
 CellID->372685725],
Cell[1109479, 31597, 481, 12, 70, "ProofInfoBody",
 CellID->131601230]
}, Closed]],
Cell[1109975, 31612, 47, 1, 70, "OpenEnvironment",
 CellID->898460167],
Cell[CellGroupData[{
Cell[1110047, 31617, 513, 15, 70, "EnvironmentHeader",
 CellID->1741618011],
Cell[1110563, 31634, 1031, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1103171586", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] closed \[ScriptCapitalP]"},
 CellID->1103171586],
Cell[1111597, 31667, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1874612924]
}, Open  ]],
Cell[1111678, 31671, 49, 1, 70, "CloseEnvironment",
 CellID->1190806585],
Cell[CellGroupData[{
Cell[1111752, 31676, 1955, 61, 70, "ProofInfo",
 CellTags->{"Proof|ID:1103171586", "Proof|ID:1103171586-1"},
 CellID->134774430],
Cell[1113710, 31739, 481, 12, 70, "ProofInfoBody",
 CellID->117381029]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1114240, 31757, 561, 13, 30, "Subsection",
 CellID->1940956769],
Cell[1114804, 31772, 47, 1, 70, "OpenEnvironment",
 CellID->927230277],
Cell[CellGroupData[{
Cell[1114876, 31777, 600, 18, 70, "EnvironmentHeader",
 CellID->856550879],
Cell[1115479, 31797, 1392, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2097953227", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "322"},
 CellID->2097953227],
Cell[1116874, 31841, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1110919542]
}, Open  ]],
Cell[1116955, 31845, 47, 1, 70, "CloseEnvironment",
 CellID->44075770],
Cell[CellGroupData[{
Cell[1117027, 31850, 4333, 115, 70, "ProofInfo",
 CellTags->{"Proof|ID:2097953227", "Proof|ID:2097953227-1"},
 CellID->163286988],
Cell[1121363, 31967, 481, 12, 70, "ProofInfoBody",
 CellID->389952566]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1121893, 31985, 304, 8, 30, "Subsection",
 CellID->1034809977],
Cell[1122200, 31995, 47, 1, 70, "OpenEnvironment",
 CellID->510359022],
Cell[CellGroupData[{
Cell[1122272, 32000, 409, 10, 70, "EnvironmentHeader",
 CellID->942893362],
Cell[1122684, 32012, 762, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:587826938", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->587826938],
Cell[1123449, 32034, 885, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:384020988", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "324"},
 CellID->384020988],
Cell[1124337, 32060, 1094, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:205407574", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "325"},
 CellID->205407574],
Cell[1125434, 32095, 1070, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:411846169", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "327"},
 CellID->411846169],
Cell[1126507, 32130, 1086, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1300778952", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "328"},
 CellID->1300778952],
Cell[1127596, 32165, 1240, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2123221051", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "329"},
 CellID->2123221051],
Cell[1128839, 32204, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1579888901]
}, Open  ]],
Cell[1128920, 32208, 48, 1, 70, "CloseEnvironment",
 CellID->890191499],
Cell[CellGroupData[{
Cell[1128993, 32213, 2632, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:384020988", "Proof|ID:384020988-1"},
 CellID->119010279],
Cell[1131628, 32293, 478, 12, 70, "ProofInfoBody",
 CellID->9568419]
}, Closed]],
Cell[CellGroupData[{
Cell[1132143, 32310, 3617, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:205407574", "Proof|ID:205407574-1"},
 CellID->2086485184],
Cell[1135763, 32410, 480, 12, 70, "ProofInfoBody",
 CellID->337622468]
}, Closed]],
Cell[CellGroupData[{
Cell[1136280, 32427, 3616, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:411846169", "Proof|ID:411846169-1"},
 CellID->183622553],
Cell[1139899, 32527, 480, 12, 70, "ProofInfoBody",
 CellID->157803399]
}, Closed]],
Cell[CellGroupData[{
Cell[1140416, 32544, 3599, 99, 70, "ProofInfo",
 CellTags->{"Proof|ID:1300778952", "Proof|ID:1300778952-1"},
 CellID->695116774],
Cell[1144018, 32645, 482, 12, 70, "ProofInfoBody",
 CellID->1300201719]
}, Closed]],
Cell[CellGroupData[{
Cell[1144537, 32662, 3199, 91, 70, "ProofInfo",
 CellTags->{"Proof|ID:2123221051", "Proof|ID:2123221051-1"},
 CellID->511445547],
Cell[1147739, 32755, 481, 12, 70, "ProofInfoBody",
 CellID->694303427]
}, Closed]],
Cell[1148235, 32770, 48, 1, 70, "OpenEnvironment",
 CellID->1756416313],
Cell[CellGroupData[{
Cell[1148308, 32775, 570, 16, 70, "EnvironmentHeader",
 CellID->235023336],
Cell[1148881, 32793, 1065, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:840093990", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] comm \[ScriptCapitalP]"},
 CellID->840093990],
Cell[1149949, 32826, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1016529366]
}, Open  ]],
Cell[1150030, 32830, 48, 1, 70, "CloseEnvironment",
 CellID->206611077],
Cell[CellGroupData[{
Cell[1150103, 32835, 1936, 60, 70, "ProofInfo",
 CellTags->{"Proof|ID:840093990", "Proof|ID:840093990-1"},
 CellID->4680134],
Cell[1152042, 32897, 479, 12, 70, "ProofInfoBody",
 CellID->46498348]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1152570, 32915, 287, 10, 30, "Subsection",
 CellID->1153917976],
Cell[1152860, 32927, 47, 1, 70, "OpenEnvironment",
 CellID->970984491],
Cell[CellGroupData[{
Cell[1152932, 32932, 580, 18, 70, "EnvironmentHeader",
 CellID->1230148421],
Cell[1153515, 32952, 1230, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1730144032", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] distr \[ScriptCapitalP]"},
 CellID->1730144032],
Cell[1154748, 32993, 65, 1, 70, "EndEnvironmentMarker",
 CellID->228018603]
}, Open  ]],
Cell[1154828, 32997, 49, 1, 70, "CloseEnvironment",
 CellID->1675122156],
Cell[CellGroupData[{
Cell[1154902, 33002, 2070, 63, 70, "ProofInfo",
 CellTags->{"Proof|ID:1730144032", "Proof|ID:1730144032-1"},
 CellID->1430620892],
Cell[1156975, 33067, 482, 12, 70, "ProofInfoBody",
 CellID->2019370717]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1157506, 33085, 318, 9, 30, "Subsection",
 CellID->1681172436],
Cell[1157827, 33096, 48, 1, 70, "OpenEnvironment",
 CellID->1339692048],
Cell[CellGroupData[{
Cell[1157900, 33101, 578, 16, 70, "EnvironmentHeader",
 CellID->702316358],
Cell[1158481, 33119, 1156, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:378610433", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1 neutral \[ScriptCapitalP]"},
 CellID->378610433],
Cell[1159640, 33154, 65, 1, 70, "EndEnvironmentMarker",
 CellID->585549701]
}, Open  ]],
Cell[1159720, 33158, 48, 1, 70, "CloseEnvironment",
 CellID->348109801],
Cell[CellGroupData[{
Cell[1159793, 33163, 2068, 62, 70, "ProofInfo",
 CellTags->{"Proof|ID:378610433", "Proof|ID:378610433-1"},
 CellID->149156064],
Cell[1161864, 33227, 481, 12, 70, "ProofInfoBody",
 CellID->1484103856]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1162394, 33245, 305, 8, 30, "Subsection",
 CellID->231588462],
Cell[1162702, 33255, 48, 1, 70, "OpenEnvironment",
 CellID->1995953378],
Cell[CellGroupData[{
Cell[1162775, 33260, 334, 9, 70, "EnvironmentHeader",
 CellID->776216969],
Cell[1163112, 33271, 941, 24, 70, "GlobalDeclaration",
 CellTags->{
  "ID:687899447", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->687899447],
Cell[1164056, 33297, 1549, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:421376854", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "330"},
 CellID->421376854],
Cell[1165608, 33344, 1765, 54, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:935511285", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "331"},
 CellID->935511285],
Cell[1167376, 33400, 1338, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1702504479", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "332"},
 CellID->1702504479],
Cell[1168717, 33442, 1912, 58, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1379089132", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "333"},
 CellID->1379089132],
Cell[1170632, 33502, 65, 1, 70, "EndEnvironmentMarker",
 CellID->855336169]
}, Open  ]],
Cell[1170712, 33506, 49, 1, 70, "CloseEnvironment",
 CellID->1874167721],
Cell[CellGroupData[{
Cell[1170786, 33511, 4903, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:421376854", "Proof|ID:421376854-1"},
 CellID->1691973429],
Cell[1175692, 33637, 480, 12, 70, "ProofInfoBody",
 CellID->275458490]
}, Closed]],
Cell[CellGroupData[{
Cell[1176209, 33654, 5709, 140, 70, "ProofInfo",
 CellTags->{"Proof|ID:935511285", "Proof|ID:935511285-1"},
 CellID->1225679565],
Cell[1181921, 33796, 521, 13, 70, "ProofInfoBody",
 CellID->787462530]
}, Closed]],
Cell[CellGroupData[{
Cell[1182479, 33814, 4528, 117, 70, "ProofInfo",
 CellTags->{"Proof|ID:1702504479", "Proof|ID:1702504479-1"},
 CellID->1881341258],
Cell[1187010, 33933, 481, 12, 70, "ProofInfoBody",
 CellID->447670728]
}, Closed]],
Cell[CellGroupData[{
Cell[1187528, 33950, 5346, 135, 70, "ProofInfo",
 CellTags->{"Proof|ID:1379089132", "Proof|ID:1379089132-1"},
 CellID->1247171944],
Cell[1192877, 34087, 482, 12, 70, "ProofInfoBody",
 CellID->1896041554]
}, Closed]],
Cell[1193374, 34102, 47, 1, 70, "OpenEnvironment",
 CellID->973193914],
Cell[CellGroupData[{
Cell[1193446, 34107, 566, 16, 70, "EnvironmentHeader",
 CellID->1831517338],
Cell[1194015, 34125, 1071, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:167864408", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RawStar] assoc \[ScriptCapitalP]"},
 CellID->167864408],
Cell[1195089, 34158, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1148171460]
}, Open  ]],
Cell[1195170, 34162, 49, 1, 70, "CloseEnvironment",
 CellID->1060151895],
Cell[CellGroupData[{
Cell[1195244, 34167, 1943, 60, 70, "ProofInfo",
 CellTags->{"Proof|ID:167864408", "Proof|ID:167864408-1"},
 CellID->1541468547],
Cell[1197190, 34229, 481, 12, 70, "ProofInfoBody",
 CellID->1970909722]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1197720, 34247, 186, 3, 30, "Subsection",
 CellID->1576202806],
Cell[1197909, 34252, 47, 1, 70, "OpenEnvironment",
 CellID->205309231],
Cell[CellGroupData[{
Cell[1197981, 34257, 475, 12, 70, "EnvironmentHeader",
 CellID->1117923091],
Cell[1198459, 34271, 820, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1082265416", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isCommRing1 \[ScriptCapitalP]"},
 CellID->1082265416],
Cell[1199282, 34294, 65, 1, 70, "EndEnvironmentMarker",
 CellID->583569111]
}, Open  ]],
Cell[1199362, 34298, 48, 1, 70, "CloseEnvironment",
 CellID->685210046],
Cell[CellGroupData[{
Cell[1199435, 34303, 1810, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1082265416", "Proof|ID:1082265416-1"},
 CellID->408658241],
Cell[1201248, 34362, 481, 12, 70, "ProofInfoBody",
 CellID->347038290]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1201778, 34380, 270, 4, 30, "Subsection",
 CellID->1536283728],
Cell[CellGroupData[{
Cell[1202073, 34388, 322, 5, 28, "Subsubsection",
 CellID->4076126],
Cell[1202398, 34395, 47, 1, 70, "OpenEnvironment",
 CellID->976925363],
Cell[CellGroupData[{
Cell[1202470, 34400, 812, 25, 70, "EnvironmentHeader",
 CellID->1342603728],
Cell[1203285, 34427, 1021, 27, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1610244238", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1610244238],
Cell[1204309, 34456, 2470, 77, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1259587933", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R 5"},
 CellID->1259587933],
Cell[1206782, 34535, 2175, 67, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:859675419", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6 ind"},
 CellID->859675419],
Cell[1208960, 34604, 1733, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:515125163", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6"},
 CellID->515125163],
Cell[1210696, 34656, 1516, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1572844749", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "C 6'"},
 CellID->1572844749],
Cell[1212215, 34700, 1674, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:771023286", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "lp 4"},
 CellID->771023286],
Cell[1213892, 34747, 1308, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:422533310", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "H 9"},
 CellID->422533310],
Cell[1215203, 34784, 65, 1, 70, "EndEnvironmentMarker",
 CellID->753430840]
}, Open  ]],
Cell[1215283, 34788, 49, 1, 70, "CloseEnvironment",
 CellID->1429212256],
Cell[CellGroupData[{
Cell[1215357, 34793, 6509, 157, 70, "ProofInfo",
 CellTags->{"Proof|ID:1259587933", "Proof|ID:1259587933-1"},
 CellID->1687842700],
Cell[1221869, 34952, 482, 12, 70, "ProofInfoBody",
 CellID->2099506532]
}, Closed]],
Cell[CellGroupData[{
Cell[1222388, 34969, 6354, 154, 70, "ProofInfo",
 CellTags->{"Proof|ID:859675419", "Proof|ID:859675419-1"},
 CellID->934287773],
Cell[1228745, 35125, 481, 12, 70, "ProofInfoBody",
 CellID->1566978955]
}, Closed]],
Cell[CellGroupData[{
Cell[1229263, 35142, 5531, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:515125163", "Proof|ID:515125163-1"},
 CellID->50230006],
Cell[1234797, 35282, 481, 12, 70, "ProofInfoBody",
 CellID->1598865118]
}, Closed]],
Cell[CellGroupData[{
Cell[1235315, 35299, 5495, 137, 70, "ProofInfo",
 CellTags->{"Proof|ID:1572844749", "Proof|ID:1572844749-1"},
 CellID->1143100247],
Cell[1240813, 35438, 482, 12, 70, "ProofInfoBody",
 CellID->1805751181]
}, Closed]],
Cell[CellGroupData[{
Cell[1241332, 35455, 5693, 140, 70, "ProofInfo",
 CellTags->{"Proof|ID:771023286", "Proof|ID:771023286-1"},
 CellID->38527445],
Cell[1247028, 35597, 481, 12, 70, "ProofInfoBody",
 CellID->1331059404]
}, Closed]],
Cell[CellGroupData[{
Cell[1247546, 35614, 4828, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:422533310", "Proof|ID:422533310-1"},
 CellID->156791201],
Cell[1252377, 35740, 481, 12, 70, "ProofInfoBody",
 CellID->1226801151]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1252907, 35758, 184, 3, 22, "Subsubsection",
 CellID->362702593],
Cell[1253094, 35763, 47, 1, 70, "OpenEnvironment",
 CellID->345925313],
Cell[CellGroupData[{
Cell[1253166, 35768, 399, 9, 70, "EnvironmentHeader",
 CellID->2037599412],
Cell[1253568, 35779, 766, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1667882379", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1667882379],
Cell[1254337, 35801, 739, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1105090112", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1105090112],
Cell[1255079, 35822, 1565, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:878771843", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 2"},
 CellID->878771843],
Cell[1256647, 35871, 868, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1863009930", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1863009930],
Cell[1257518, 35896, 1713, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1958431161", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 5"},
 CellID->1958431161],
Cell[1259234, 35947, 1464, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:758577094", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 6"},
 CellID->758577094],
Cell[1260701, 35989, 736, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:544874520", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->544874520],
Cell[1261440, 36010, 738, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:158305602", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->158305602],
Cell[1262181, 36031, 1286, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:347486225", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 9"},
 CellID->347486225],
Cell[1263470, 36069, 1453, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1591957532", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 10"},
 CellID->1591957532],
Cell[1264926, 36110, 721, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1155967768", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1155967768],
Cell[1265650, 36131, 1293, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1942982211", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 15"},
 CellID->1942982211],
Cell[1266946, 36172, 1193, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:44857450", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 12"},
 CellID->44857450],
Cell[1268142, 36207, 1246, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:928737230", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 13"},
 CellID->928737230],
Cell[1269391, 36243, 1269, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:233618042", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 14"},
 CellID->233618042],
Cell[1270663, 36278, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1441920355]
}, Open  ]],
Cell[1270744, 36282, 48, 1, 70, "CloseEnvironment",
 CellID->614488126],
Cell[CellGroupData[{
Cell[1270817, 36287, 4998, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:878771843", "Proof|ID:878771843-1"},
 CellID->395620750],
Cell[1275818, 36415, 480, 12, 70, "ProofInfoBody",
 CellID->856996584]
}, Closed]],
Cell[CellGroupData[{
Cell[1276335, 36432, 5675, 141, 70, "ProofInfo",
 CellTags->{"Proof|ID:1958431161", "Proof|ID:1958431161-1"},
 CellID->756678784],
Cell[1282013, 36575, 481, 12, 70, "ProofInfoBody",
 CellID->871051679]
}, Closed]],
Cell[CellGroupData[{
Cell[1282531, 36592, 5360, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:758577094", "Proof|ID:758577094-1"},
 CellID->2118355576],
Cell[1287894, 36726, 481, 12, 70, "ProofInfoBody",
 CellID->2066466981]
}, Closed]],
Cell[CellGroupData[{
Cell[1288412, 36743, 5170, 128, 70, "ProofInfo",
 CellTags->{"Proof|ID:347486225", "Proof|ID:347486225-1"},
 CellID->1908653227],
Cell[1293585, 36873, 547, 13, 70, "ProofInfoBody",
 CellID->1535111932]
}, Closed]],
Cell[CellGroupData[{
Cell[1294169, 36891, 5346, 133, 70, "ProofInfo",
 CellTags->{"Proof|ID:1591957532", "Proof|ID:1591957532-1"},
 CellID->1145145855],
Cell[1299518, 37026, 481, 12, 70, "ProofInfoBody",
 CellID->311029242]
}, Closed]],
Cell[CellGroupData[{
Cell[1300036, 37043, 4432, 115, 70, "ProofInfo",
 CellTags->{"Proof|ID:1942982211", "Proof|ID:1942982211-1"},
 CellID->1270512064],
Cell[1304471, 37160, 482, 12, 70, "ProofInfoBody",
 CellID->1236173956]
}, Closed]],
Cell[CellGroupData[{
Cell[1304990, 37177, 5004, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:44857450", "Proof|ID:44857450-1"},
 CellID->584781971],
Cell[1309997, 37303, 479, 12, 70, "ProofInfoBody",
 CellID->834128854]
}, Closed]],
Cell[CellGroupData[{
Cell[1310513, 37320, 5009, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:928737230", "Proof|ID:928737230-1"},
 CellID->103471022],
Cell[1315525, 37446, 481, 12, 70, "ProofInfoBody",
 CellID->1716227546]
}, Closed]],
Cell[CellGroupData[{
Cell[1316043, 37463, 4437, 112, 70, "ProofInfo",
 CellTags->{"Proof|ID:233618042", "Proof|ID:233618042-1"},
 CellID->424912135],
Cell[1320483, 37577, 480, 12, 70, "ProofInfoBody",
 CellID->513975655]
}, Closed]],
Cell[1320978, 37592, 47, 1, 70, "OpenEnvironment",
 CellID->135337510],
Cell[CellGroupData[{
Cell[1321050, 37597, 383, 9, 70, "EnvironmentHeader",
 CellID->763939551],
Cell[1321436, 37608, 900, 24, 70, "GlobalDeclaration",
 CellTags->{
  "ID:183428126", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->183428126],
Cell[1322339, 37634, 1652, 48, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2026663200", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 2"},
 CellID->2026663200],
Cell[1323994, 37684, 1920, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1916930407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomial 5"},
 CellID->1916930407],
Cell[1325917, 37742, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1683231865]
}, Open  ]],
Cell[1325998, 37746, 49, 1, 70, "CloseEnvironment",
 CellID->1360979825],
Cell[CellGroupData[{
Cell[1326072, 37751, 5519, 137, 70, "ProofInfo",
 CellTags->{"Proof|ID:2026663200", "Proof|ID:2026663200-1"},
 CellID->1385906749],
Cell[1331594, 37890, 481, 12, 70, "ProofInfoBody",
 CellID->870820524]
}, Closed]],
Cell[CellGroupData[{
Cell[1332112, 37907, 5248, 130, 70, "ProofInfo",
 CellTags->{"Proof|ID:1916930407", "Proof|ID:1916930407-1"},
 CellID->1619770492],
Cell[1337363, 38039, 481, 12, 70, "ProofInfoBody",
 CellID->833552299]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1337893, 38057, 252, 6, 22, "Subsubsection",
 CellID->1800295324],
Cell[1338148, 38065, 931, 20, 82, "Text",
 CellID->1851558702],
Cell[1339082, 38087, 48, 1, 12, "OpenEnvironment",
 CellID->1582231852],
Cell[CellGroupData[{
Cell[1339155, 38092, 482, 12, 43, "EnvironmentHeader",
 CellID->2109816331],
Cell[1339640, 38106, 830, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1225634217", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduction ring \[ScriptCapitalR]"},
 CellID->1225634217],
Cell[1340473, 38129, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1230264783]
}, Open  ]],
Cell[1340554, 38133, 48, 1, 16, "CloseEnvironment",
 CellID->665224426],
Cell[1340605, 38136, 47, 1, 12, "OpenEnvironment",
 CellID->556263379],
Cell[CellGroupData[{
Cell[1340677, 38141, 335, 9, 43, "EnvironmentHeader",
 CellID->1625326245],
Cell[1341015, 38152, 2050, 59, 74, "FormalTextInputFormula",
 CellTags->{
  "ID:1542801947", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "300"},
 CellID->1542801947],
Cell[1343068, 38213, 65, 1, 26, "EndEnvironmentMarker",
 CellID->445923411]
}, Open  ]],
Cell[1343148, 38217, 48, 1, 16, "CloseEnvironment",
 CellID->824716703],
Cell[CellGroupData[{
Cell[1343221, 38222, 4714, 121, 28, "ProofInfo",
 CellTags->{"Proof|ID:1542801947", "Proof|ID:1542801947-1"},
 CellID->1097904678],
Cell[1347938, 38345, 482, 12, 70, "ProofInfoBody",
 CellID->1371063757]
}, Closed]],
Cell[1348435, 38360, 48, 1, 12, "OpenEnvironment",
 CellID->1307939572],
Cell[CellGroupData[{
Cell[1348508, 38365, 376, 9, 43, "EnvironmentHeader",
 CellID->36053659],
Cell[1348887, 38376, 5079, 138, 212, "FormalTextInputFormula",
 CellTags->{
  "ID:946287998", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isOrderedRec"},
 CellID->946287998],
Cell[1353969, 38516, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1338223823]
}, Open  ]],
Cell[1354050, 38520, 49, 1, 16, "CloseEnvironment",
 CellID->2058247028],
Cell[1354102, 38523, 47, 1, 12, "OpenEnvironment",
 CellID->259022577],
Cell[CellGroupData[{
Cell[1354174, 38528, 479, 12, 43, "EnvironmentHeader",
 CellID->1942671025],
Cell[1354656, 38542, 704, 18, 23, "GlobalDeclaration",
 CellTags->{
  "ID:1485445352", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1485445352],
Cell[1355363, 38562, 2882, 86, 82, "FormalTextInputFormula",
 CellTags->{
  "ID:215225048", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "301"},
 CellID->215225048],
Cell[1358248, 38650, 2636, 82, 82, "FormalTextInputFormula",
 CellTags->{
  "ID:1704612967", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "302"},
 CellID->1704612967],
Cell[1360887, 38734, 789, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:162248449", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "303"},
 CellID->162248449],
Cell[1361679, 38758, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1460333818]
}, Open  ]],
Cell[1361760, 38762, 48, 1, 16, "CloseEnvironment",
 CellID->771705247],
Cell[CellGroupData[{
Cell[1361833, 38767, 7717, 184, 28, "ProofInfo",
 CellTags->{"Proof|ID:215225048", "Proof|ID:215225048-1"},
 CellID->1531901166],
Cell[1369553, 38953, 480, 12, 70, "ProofInfoBody",
 CellID->361326966]
}, Closed]],
Cell[CellGroupData[{
Cell[1370070, 38970, 7742, 187, 26, "ProofInfo",
 CellTags->{"Proof|ID:1704612967", "Proof|ID:1704612967-1"},
 CellID->494680653],
Cell[1377815, 39159, 482, 12, 70, "ProofInfoBody",
 CellID->1603308827]
}, Closed]],
Cell[CellGroupData[{
Cell[1378334, 39176, 1975, 62, 26, "ProofInfo",
 CellTags->{"Proof|ID:162248449", "Proof|ID:162248449-1"},
 CellID->798928706],
Cell[1380312, 39240, 481, 12, 70, "ProofInfoBody",
 CellID->2125561798]
}, Closed]],
Cell[CellGroupData[{
Cell[1380830, 39257, 416, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1465916398],
Cell[1381249, 39269, 14679, 473, 970, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->893664258],
Cell[1395931, 39744, 313, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1643562830]
}, Closed]],
Cell[CellGroupData[{
Cell[1396281, 39755, 418, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->652409243],
Cell[1396702, 39767, 309, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1647916099],
Cell[1397014, 39774, 967, 24, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1566511347],
Cell[1397984, 39800, 1270, 31, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->14964074],
Cell[1399257, 39833, 982, 23, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->628000282],
Cell[1400242, 39858, 1417, 34, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->21049257],
Cell[1401662, 39894, 14674, 441, 1443, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1479590632],
Cell[1416339, 40337, 314, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->702250699]
}, Closed]],
Cell[1416668, 40346, 660, 13, 62, "Text",
 CellID->152935259],
Cell[CellGroupData[{
Cell[1417353, 40363, 278, 6, 32, "Input",
 CellID->1209476239],
Cell[1417634, 40371, 490, 9, 32, "Output",
 CellID->162283938]
}, Open  ]],
Cell[1418139, 40383, 48, 1, 12, "OpenEnvironment",
 CellID->1418355964],
Cell[CellGroupData[{
Cell[1418212, 40388, 431, 11, 43, "EnvironmentHeader",
 CellID->118382873],
Cell[1418646, 40401, 1650, 45, 54, "FormalTextInputFormula",
 CellTags->{
  "ID:1015468103", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "304"},
 CellID->1015468103],
Cell[1420299, 40448, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1920454598]
}, Open  ]],
Cell[1420380, 40452, 48, 1, 16, "CloseEnvironment",
 CellID->376387788]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1420489, 40460, 286, 6, 43, "Section",
 CellID->1755638798],
Cell[1420778, 40468, 514, 9, 43, "Text",
 CellID->146700820],
Cell[1421295, 40479, 980, 15, 62, "Text",
 CellID->1617922970],
Cell[1422278, 40496, 48, 1, 12, "OpenEnvironment",
 CellID->2116256688],
Cell[CellGroupData[{
Cell[1422351, 40501, 624, 14, 43, "EnvironmentHeader",
 CellID->484984619],
Cell[1422978, 40517, 853, 23, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:7476052", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduction-polynomial domain \[ScriptCapitalP]"},
 CellID->7476052],
Cell[1423834, 40542, 65, 1, 26, "EndEnvironmentMarker",
 CellID->640671230]
}, Open  ]],
Cell[1423914, 40546, 48, 1, 16, "CloseEnvironment",
 CellID->843678998],
Cell[CellGroupData[{
Cell[1423987, 40551, 185, 3, 38, "Subsection",
 CellID->826234296],
Cell[CellGroupData[{
Cell[1424197, 40558, 291, 4, 28, "Subsubsection",
 CellID->1832383314],
Cell[1424491, 40564, 48, 1, 12, "OpenEnvironment",
 CellID->1005504538],
Cell[CellGroupData[{
Cell[1424564, 40569, 411, 10, 43, "EnvironmentHeader",
 CellID->2142794181],
Cell[1424978, 40581, 763, 20, 42, "GlobalDeclaration",
 CellTags->{
  "ID:173226898", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->173226898],
Cell[1425744, 40603, 738, 19, 42, "GlobalDeclaration",
 CellTags->{
  "ID:835223349", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->835223349],
Cell[1426485, 40624, 929, 25, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:181219027", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "356"},
 CellID->181219027],
Cell[1427417, 40651, 1555, 46, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:271793724", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "382"},
 CellID->271793724],
Cell[1428975, 40699, 1487, 44, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:575028756", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "382'"},
 CellID->575028756],
Cell[1430465, 40745, 1606, 48, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:2020716880", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "383"},
 CellID->2020716880],
Cell[1432074, 40795, 1655, 48, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1964835443", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "381"},
 CellID->1964835443],
Cell[1433732, 40845, 1945, 55, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1599847484", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "380"},
 CellID->1599847484],
Cell[1435680, 40902, 3915, 101, 100, "FormalTextInputFormula",
 CellTags->{
  "ID:590709100", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "379"},
 CellID->590709100],
Cell[1439598, 41005, 1877, 57, 65, "FormalTextInputFormula",
 CellTags->{
  "ID:586184654", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "405"},
 CellID->586184654],
Cell[1441478, 41064, 1402, 39, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:919284948", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "406"},
 CellID->919284948],
Cell[1442883, 41105, 65, 1, 26, "EndEnvironmentMarker",
 CellID->141036460]
}, Open  ]],
Cell[1442963, 41109, 49, 1, 16, "CloseEnvironment",
 CellID->1634960697],
Cell[CellGroupData[{
Cell[1443037, 41114, 2840, 82, 28, "ProofInfo",
 CellTags->{"Proof|ID:181219027", "Proof|ID:181219027-1"},
 CellID->719304724],
Cell[1445880, 41198, 481, 12, 70, "ProofInfoBody",
 CellID->1538405320]
}, Closed]],
Cell[CellGroupData[{
Cell[1446398, 41215, 4687, 120, 26, "ProofInfo",
 CellTags->{"Proof|ID:271793724", "Proof|ID:271793724-1"},
 CellID->1889243650],
Cell[1451088, 41337, 480, 12, 70, "ProofInfoBody",
 CellID->858383599]
}, Closed]],
Cell[CellGroupData[{
Cell[1451605, 41354, 4332, 112, 26, "ProofInfo",
 CellTags->{"Proof|ID:575028756", "Proof|ID:575028756-1"},
 CellID->993022467],
Cell[1455940, 41468, 598, 15, 70, "ProofInfoBody",
 CellID->1954501299]
}, Closed]],
Cell[CellGroupData[{
Cell[1456575, 41488, 4958, 127, 26, "ProofInfo",
 CellTags->{"Proof|ID:2020716880", "Proof|ID:2020716880-1"},
 CellID->1784894938],
Cell[1461536, 41617, 482, 12, 70, "ProofInfoBody",
 CellID->1513088413]
}, Closed]],
Cell[CellGroupData[{
Cell[1462055, 41634, 5039, 127, 26, "ProofInfo",
 CellTags->{"Proof|ID:1964835443", "Proof|ID:1964835443-1"},
 CellID->880026044],
Cell[1467097, 41763, 482, 12, 70, "ProofInfoBody",
 CellID->1363714399]
}, Closed]],
Cell[CellGroupData[{
Cell[1467616, 41780, 5710, 139, 26, "ProofInfo",
 CellTags->{"Proof|ID:1599847484", "Proof|ID:1599847484-1"},
 CellID->780294033],
Cell[1473329, 41921, 482, 12, 70, "ProofInfoBody",
 CellID->2075951568]
}, Closed]],
Cell[CellGroupData[{
Cell[1473848, 41938, 7964, 182, 26, "ProofInfo",
 CellTags->{"Proof|ID:590709100", "Proof|ID:590709100-1"},
 CellID->726011705],
Cell[1481815, 42122, 481, 12, 70, "ProofInfoBody",
 CellID->1474313818]
}, Closed]],
Cell[CellGroupData[{
Cell[1482333, 42139, 5963, 150, 26, "ProofInfo",
 CellTags->{"Proof|ID:586184654", "Proof|ID:586184654-1"},
 CellID->1800670266],
Cell[1488299, 42291, 481, 12, 70, "ProofInfoBody",
 CellID->1362597520]
}, Closed]],
Cell[CellGroupData[{
Cell[1488817, 42308, 3584, 96, 26, "ProofInfo",
 CellTags->{"Proof|ID:919284948", "Proof|ID:919284948-1"},
 CellID->2038963285],
Cell[1492404, 42406, 480, 12, 70, "ProofInfoBody",
 CellID->427287473]
}, Closed]],
Cell[CellGroupData[{
Cell[1492921, 42423, 420, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1198510616],
Cell[1493344, 42435, 9388, 301, 1235, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->710706997],
Cell[1502735, 42738, 365, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1793634431]
}, Closed]],
Cell[1503115, 42748, 48, 1, 12, "OpenEnvironment",
 CellID->1494017576],
Cell[CellGroupData[{
Cell[1503188, 42753, 411, 11, 43, "EnvironmentHeader",
 CellID->141141923],
Cell[1503602, 42766, 738, 19, 42, "GlobalDeclaration",
 CellTags->{
  "ID:895736483", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->895736483],
Cell[1504343, 42787, 741, 19, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1904443546", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1904443546],
Cell[1505087, 42808, 739, 19, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1587558771", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1587558771],
Cell[1505829, 42829, 1422, 41, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:530327018", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377"},
 CellID->530327018],
Cell[1507254, 42872, 2251, 69, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:2044424156", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377'"},
 CellID->2044424156],
Cell[1509508, 42943, 1353, 39, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:14600367", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "377''"},
 CellID->14600367],
Cell[1510864, 42984, 1494, 44, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:1313257457", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 7"},
 CellID->1313257457],
Cell[1512361, 43030, 2317, 65, 87, "FormalTextInputFormula",
 CellTags->{
  "ID:1776891622", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] 8"},
 CellID->1776891622],
Cell[1514681, 43097, 65, 1, 26, "EndEnvironmentMarker",
 CellID->259176606]
}, Open  ]],
Cell[1514761, 43101, 49, 1, 16, "CloseEnvironment",
 CellID->1149241651],
Cell[CellGroupData[{
Cell[1514835, 43106, 4300, 110, 28, "ProofInfo",
 CellTags->{"Proof|ID:530327018", "Proof|ID:530327018-1"},
 CellID->24650902],
Cell[1519138, 43218, 480, 12, 70, "ProofInfoBody",
 CellID->660190578]
}, Closed]],
Cell[CellGroupData[{
Cell[1519655, 43235, 6116, 148, 26, "ProofInfo",
 CellTags->{"Proof|ID:2044424156", "Proof|ID:2044424156-1"},
 CellID->35999503],
Cell[1525774, 43385, 481, 12, 70, "ProofInfoBody",
 CellID->259827218]
}, Closed]],
Cell[CellGroupData[{
Cell[1526292, 43402, 4419, 114, 26, "ProofInfo",
 CellTags->{"Proof|ID:14600367", "Proof|ID:14600367-1"},
 CellID->439633690],
Cell[1530714, 43518, 479, 12, 70, "ProofInfoBody",
 CellID->775230185]
}, Closed]],
Cell[CellGroupData[{
Cell[1531230, 43535, 4220, 111, 26, "ProofInfo",
 CellTags->{"Proof|ID:1313257457", "Proof|ID:1313257457-1"},
 CellID->1865577507],
Cell[1535453, 43648, 481, 12, 70, "ProofInfoBody",
 CellID->246532405]
}, Closed]],
Cell[CellGroupData[{
Cell[1535971, 43665, 6238, 153, 26, "ProofInfo",
 CellTags->{"Proof|ID:1776891622", "Proof|ID:1776891622-1"},
 CellID->89604998],
Cell[1542212, 43820, 481, 12, 70, "ProofInfoBody",
 CellID->141709217]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1542742, 43838, 189, 3, 22, "Subsubsection",
 CellID->2009513133],
Cell[1542934, 43843, 47, 1, 70, "OpenEnvironment",
 CellID->803353906],
Cell[CellGroupData[{
Cell[1543006, 43848, 421, 11, 70, "EnvironmentHeader",
 CellID->1741879042],
Cell[1543430, 43861, 1093, 26, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1077742446", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1077742446],
Cell[1544526, 43889, 1503, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1717822499", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "384"},
 CellID->1717822499],
Cell[1546032, 43934, 913, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1045770541", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "358"},
 CellID->1045770541],
Cell[1546948, 43961, 2336, 65, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:312784637", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "393"},
 CellID->312784637],
Cell[1549287, 44028, 859, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:172690263", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->172690263],
Cell[1550149, 44053, 1284, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:174940712", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial"},
 CellID->174940712],
Cell[1551436, 44090, 1482, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1850961673", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial II"},
 CellID->1850961673],
Cell[1552921, 44136, 2001, 59, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:849856225", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "mult monomial III"},
 CellID->849856225],
Cell[1554925, 44197, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1178443032]
}, Open  ]],
Cell[1555006, 44201, 49, 1, 70, "CloseEnvironment",
 CellID->1823748294],
Cell[CellGroupData[{
Cell[1555080, 44206, 4761, 121, 70, "ProofInfo",
 CellTags->{"Proof|ID:1717822499", "Proof|ID:1717822499-1"},
 CellID->1434876425],
Cell[1559844, 44329, 482, 12, 70, "ProofInfoBody",
 CellID->1232590601]
}, Closed]],
Cell[CellGroupData[{
Cell[1560363, 44346, 3608, 99, 70, "ProofInfo",
 CellTags->{"Proof|ID:1045770541", "Proof|ID:1045770541-1"},
 CellID->1393398660],
Cell[1563974, 44447, 481, 12, 70, "ProofInfoBody",
 CellID->483586923]
}, Closed]],
Cell[CellGroupData[{
Cell[1564492, 44464, 6941, 165, 70, "ProofInfo",
 CellTags->{"Proof|ID:312784637", "Proof|ID:312784637-1"},
 CellID->2016825295],
Cell[1571436, 44631, 480, 12, 70, "ProofInfoBody",
 CellID->537234257]
}, Closed]],
Cell[CellGroupData[{
Cell[1571953, 44648, 4620, 118, 70, "ProofInfo",
 CellTags->{"Proof|ID:174940712", "Proof|ID:174940712-1"},
 CellID->1306123142],
Cell[1576576, 44768, 480, 12, 70, "ProofInfoBody",
 CellID->275742108]
}, Closed]],
Cell[CellGroupData[{
Cell[1577093, 44785, 5473, 135, 70, "ProofInfo",
 CellTags->{"Proof|ID:1850961673", "Proof|ID:1850961673-1"},
 CellID->1891103710],
Cell[1582569, 44922, 481, 12, 70, "ProofInfoBody",
 CellID->753129535]
}, Closed]],
Cell[CellGroupData[{
Cell[1583087, 44939, 6911, 162, 70, "ProofInfo",
 CellTags->{"Proof|ID:849856225", "Proof|ID:849856225-1"},
 CellID->232642418],
Cell[1590001, 45103, 481, 12, 70, "ProofInfoBody",
 CellID->1171947620]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1590531, 45121, 182, 3, 22, "Subsubsection",
 CellID->1854005352],
Cell[1590716, 45126, 47, 1, 70, "OpenEnvironment",
 CellID->941507149],
Cell[CellGroupData[{
Cell[1590788, 45131, 408, 10, 70, "EnvironmentHeader",
 CellID->1448249107],
Cell[1591199, 45143, 1001, 26, 70, "GlobalDeclaration",
 CellTags->{
  "ID:633254127", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->633254127],
Cell[1592203, 45171, 1124, 29, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1581566919", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1581566919],
Cell[1593330, 45202, 977, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1230351665", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 1"},
 CellID->1230351665],
Cell[1594310, 45229, 816, 22, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1294700753", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1294700753],
Cell[1595129, 45253, 1349, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:168346017", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 2"},
 CellID->168346017],
Cell[1596481, 45293, 1960, 57, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1208795229", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 3"},
 CellID->1208795229],
Cell[1598444, 45352, 1764, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:369110158", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "tuples \[CenterDot] 4"},
 CellID->369110158],
Cell[1600211, 45405, 65, 1, 70, "EndEnvironmentMarker",
 CellID->429653964]
}, Open  ]],
Cell[1600291, 45409, 48, 1, 70, "CloseEnvironment",
 CellID->533685121],
Cell[CellGroupData[{
Cell[1600364, 45414, 4421, 115, 70, "ProofInfo",
 CellTags->{"Proof|ID:1230351665", "Proof|ID:1230351665-1"},
 CellID->803576507],
Cell[1604788, 45531, 481, 12, 70, "ProofInfoBody",
 CellID->453965713]
}, Closed]],
Cell[CellGroupData[{
Cell[1605306, 45548, 7111, 178, 70, "ProofInfo",
 CellTags->{"Proof|ID:168346017", "Proof|ID:168346017-1"},
 CellID->1197262074],
Cell[1612420, 45728, 481, 12, 70, "ProofInfoBody",
 CellID->1861690467]
}, Closed]],
Cell[CellGroupData[{
Cell[1612938, 45745, 8032, 197, 70, "ProofInfo",
 CellTags->{"Proof|ID:1208795229", "Proof|ID:1208795229-1"},
 CellID->1531683337],
Cell[1620973, 45944, 482, 12, 70, "ProofInfoBody",
 CellID->1417027127]
}, Closed]],
Cell[CellGroupData[{
Cell[1621492, 45961, 6980, 170, 70, "ProofInfo",
 CellTags->{"Proof|ID:369110158", "Proof|ID:369110158-1"},
 CellID->1626361315],
Cell[1628475, 46133, 481, 12, 70, "ProofInfoBody",
 CellID->1922614985]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1629017, 46152, 190, 3, 30, "Subsection",
 CellID->410707321],
Cell[1629210, 46157, 48, 1, 70, "OpenEnvironment",
 CellID->2043816998],
Cell[CellGroupData[{
Cell[1629283, 46162, 511, 13, 70, "EnvironmentHeader",
 CellID->1791092897],
Cell[1629797, 46177, 761, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:155708286", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->155708286],
Cell[1630561, 46199, 807, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:280312466", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "4"},
 CellID->280312466],
Cell[1631371, 46224, 1007, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:713919982", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "5"},
 CellID->713919982],
Cell[1632381, 46256, 1354, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1630947645", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "6"},
 CellID->1630947645],
Cell[1633738, 46300, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1332205312]
}, Open  ]],
Cell[1633819, 46304, 49, 1, 70, "CloseEnvironment",
 CellID->1152004342],
Cell[CellGroupData[{
Cell[1633893, 46309, 2758, 81, 70, "ProofInfo",
 CellTags->{"Proof|ID:280312466", "Proof|ID:280312466-1"},
 CellID->1561997],
Cell[1636654, 46392, 480, 12, 70, "ProofInfoBody",
 CellID->560792166]
}, Closed]],
Cell[CellGroupData[{
Cell[1637171, 46409, 3567, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:713919982", "Proof|ID:713919982-1"},
 CellID->393026],
Cell[1640741, 46509, 480, 12, 70, "ProofInfoBody",
 CellID->573639697]
}, Closed]],
Cell[CellGroupData[{
Cell[1641258, 46526, 4027, 107, 70, "ProofInfo",
 CellTags->{"Proof|ID:1630947645", "Proof|ID:1630947645-1"},
 CellID->350427357],
Cell[1645288, 46635, 481, 12, 70, "ProofInfoBody",
 CellID->234994923]
}, Closed]],
Cell[1645784, 46650, 48, 1, 70, "OpenEnvironment",
 CellID->1582119546],
Cell[CellGroupData[{
Cell[1645857, 46655, 496, 13, 70, "EnvironmentHeader",
 CellID->1640450222],
Cell[1646356, 46670, 1075, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:795795393", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "360"},
 CellID->795795393],
Cell[1647434, 46704, 65, 1, 70, "EndEnvironmentMarker",
 CellID->233694654]
}, Open  ]],
Cell[1647514, 46708, 48, 1, 70, "CloseEnvironment",
 CellID->939147830],
Cell[CellGroupData[{
Cell[1647587, 46713, 1939, 60, 70, "ProofInfo",
 CellTags->{"Proof|ID:795795393", "Proof|ID:795795393-1"},
 CellID->1058256179],
Cell[1649529, 46775, 480, 12, 70, "ProofInfoBody",
 CellID->883251815]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1650058, 46793, 187, 3, 30, "Subsection",
 CellID->2044895609],
Cell[1650248, 46798, 48, 1, 70, "OpenEnvironment",
 CellID->1937965720],
Cell[CellGroupData[{
Cell[1650321, 46803, 459, 12, 70, "EnvironmentHeader",
 CellID->2071197478],
Cell[1650783, 46817, 778, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1302712453", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1302712453],
Cell[1651564, 46839, 5366, 155, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1092343128", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "1"},
 CellID->1092343128],
Cell[1656933, 46996, 1107, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:311837531", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "2"},
 CellID->311837531],
Cell[1658043, 47032, 1868, 61, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2036982330", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "3"},
 CellID->2036982330],
Cell[1659914, 47095, 2517, 75, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:917853466", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "378"},
 CellID->917853466],
Cell[1662434, 47172, 66, 1, 70, "EndEnvironmentMarker",
 CellID->2118519333]
}, Open  ]],
Cell[1662515, 47176, 48, 1, 70, "CloseEnvironment",
 CellID->634503852],
Cell[CellGroupData[{
Cell[1662588, 47181, 11724, 267, 70, "ProofInfo",
 CellTags->{"Proof|ID:1092343128", "Proof|ID:1092343128-1"},
 CellID->402687292],
Cell[1674315, 47450, 481, 12, 70, "ProofInfoBody",
 CellID->138879590]
}, Closed]],
Cell[CellGroupData[{
Cell[1674833, 47467, 3041, 84, 70, "ProofInfo",
 CellTags->{"Proof|ID:311837531", "Proof|ID:311837531-1"},
 CellID->222069951],
Cell[1677877, 47553, 480, 12, 70, "ProofInfoBody",
 CellID->393972337]
}, Closed]],
Cell[CellGroupData[{
Cell[1678394, 47570, 5049, 129, 70, "ProofInfo",
 CellTags->{"Proof|ID:2036982330", "Proof|ID:2036982330-1"},
 CellID->78668347],
Cell[1683446, 47701, 481, 12, 70, "ProofInfoBody",
 CellID->401808988]
}, Closed]],
Cell[CellGroupData[{
Cell[1683964, 47718, 5399, 136, 70, "ProofInfo",
 CellTags->{"Proof|ID:917853466", "Proof|ID:917853466-1"},
 CellID->463010928],
Cell[1689366, 47856, 480, 12, 70, "ProofInfoBody",
 CellID->429418029]
}, Closed]],
Cell[1689861, 47871, 47, 1, 70, "OpenEnvironment",
 CellID->885823776],
Cell[CellGroupData[{
Cell[1689933, 47876, 410, 10, 70, "EnvironmentHeader",
 CellID->1196982619],
Cell[1690346, 47888, 1019, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1009043994", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "361"},
 CellID->1009043994],
Cell[1691368, 47921, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1321732972]
}, Open  ]],
Cell[1691449, 47925, 48, 1, 70, "CloseEnvironment",
 CellID->212143235],
Cell[CellGroupData[{
Cell[1691522, 47930, 1937, 61, 70, "ProofInfo",
 CellTags->{"Proof|ID:1009043994", "Proof|ID:1009043994-1"},
 CellID->307122811],
Cell[1693462, 47993, 481, 12, 70, "ProofInfoBody",
 CellID->698773133]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1693992, 48011, 186, 3, 30, "Subsection",
 CellID->1626247964],
Cell[1694181, 48016, 47, 1, 70, "OpenEnvironment",
 CellID->681587107],
Cell[CellGroupData[{
Cell[1694253, 48021, 419, 11, 70, "EnvironmentHeader",
 CellID->190398885],
Cell[1694675, 48034, 791, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:412087985", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "362"},
 CellID->412087985],
Cell[1695469, 48058, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1616084504]
}, Open  ]],
Cell[1695550, 48062, 48, 1, 70, "CloseEnvironment",
 CellID->675296661],
Cell[CellGroupData[{
Cell[1695623, 48067, 1749, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:412087985", "Proof|ID:412087985-1"},
 CellID->453147487],
Cell[1697375, 48125, 480, 12, 70, "ProofInfoBody",
 CellID->500939487]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1697904, 48143, 177, 3, 30, "Subsection",
 CellID->1977275530],
Cell[1698084, 48148, 47, 1, 70, "OpenEnvironment",
 CellID->787808098],
Cell[CellGroupData[{
Cell[1698156, 48153, 395, 9, 70, "EnvironmentHeader",
 CellID->2032276339],
Cell[1698554, 48164, 783, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1267059190", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "363"},
 CellID->1267059190],
Cell[1699340, 48187, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1962684668]
}, Open  ]],
Cell[1699421, 48191, 48, 1, 70, "CloseEnvironment",
 CellID->857332813],
Cell[CellGroupData[{
Cell[1699494, 48196, 1734, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1267059190", "Proof|ID:1267059190-1"},
 CellID->1552638821],
Cell[1701231, 48255, 482, 12, 70, "ProofInfoBody",
 CellID->1865569314]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1701762, 48273, 175, 3, 30, "Subsection",
 CellID->1132174785],
Cell[1701940, 48278, 47, 1, 70, "OpenEnvironment",
 CellID->978075595],
Cell[CellGroupData[{
Cell[1702012, 48283, 392, 9, 70, "EnvironmentHeader",
 CellID->275565290],
Cell[1702407, 48294, 780, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:378652997", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "364"},
 CellID->378652997],
Cell[1703190, 48317, 65, 1, 70, "EndEnvironmentMarker",
 CellID->876981541]
}, Open  ]],
Cell[1703270, 48321, 49, 1, 70, "CloseEnvironment",
 CellID->1058953121],
Cell[CellGroupData[{
Cell[1703344, 48326, 1725, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:378652997", "Proof|ID:378652997-1"},
 CellID->1328553736],
Cell[1705072, 48384, 481, 12, 70, "ProofInfoBody",
 CellID->1498432134]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1705602, 48402, 174, 3, 30, "Subsection",
 CellID->1093569315],
Cell[1705779, 48407, 426, 10, 70, "Text",
 CellID->1339945595],
Cell[1706208, 48419, 47, 1, 70, "OpenEnvironment",
 CellID->815100016],
Cell[CellGroupData[{
Cell[1706280, 48424, 393, 9, 70, "EnvironmentHeader",
 CellID->2082733210],
Cell[1706676, 48435, 780, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1036416147", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "365"},
 CellID->1036416147],
Cell[1707459, 48458, 65, 1, 70, "EndEnvironmentMarker",
 CellID->220603823]
}, Open  ]],
Cell[1707539, 48462, 47, 1, 70, "CloseEnvironment",
 CellID->29368222],
Cell[CellGroupData[{
Cell[1707611, 48467, 1733, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1036416147", "Proof|ID:1036416147-1"},
 CellID->246228744],
Cell[1709347, 48526, 482, 12, 70, "ProofInfoBody",
 CellID->1948904361]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1709878, 48544, 174, 3, 30, "Subsection",
 CellID->1028022157],
Cell[1710055, 48549, 48, 1, 70, "OpenEnvironment",
 CellID->1224247765],
Cell[CellGroupData[{
Cell[1710128, 48554, 422, 11, 70, "EnvironmentHeader",
 CellID->1843437225],
Cell[1710553, 48567, 2877, 78, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1341307829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "386"},
 CellID->1341307829],
Cell[1713433, 48647, 3328, 89, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1193326225", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "385"},
 CellID->1193326225],
Cell[1716764, 48738, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1058015582]
}, Open  ]],
Cell[1716845, 48742, 49, 1, 70, "CloseEnvironment",
 CellID->2043406167],
Cell[CellGroupData[{
Cell[1716919, 48747, 6984, 175, 70, "ProofInfo",
 CellTags->{"Proof|ID:1341307829", "Proof|ID:1341307829-1"},
 CellID->467247998],
Cell[1723906, 48924, 481, 12, 70, "ProofInfoBody",
 CellID->550586326]
}, Closed]],
Cell[CellGroupData[{
Cell[1724424, 48941, 7796, 191, 70, "ProofInfo",
 CellTags->{"Proof|ID:1193326225", "Proof|ID:1193326225-1"},
 CellID->348360809],
Cell[1732223, 49134, 480, 12, 70, "ProofInfoBody",
 CellID->43480281]
}, Closed]],
Cell[1732718, 49149, 47, 1, 70, "OpenEnvironment",
 CellID->472886914],
Cell[CellGroupData[{
Cell[1732790, 49154, 392, 9, 70, "EnvironmentHeader",
 CellID->1909661142],
Cell[1733185, 49165, 783, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2033377127", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "366"},
 CellID->2033377127],
Cell[1733971, 49188, 65, 1, 70, "EndEnvironmentMarker",
 CellID->684694705]
}, Open  ]],
Cell[1734051, 49192, 49, 1, 70, "CloseEnvironment",
 CellID->1928200850],
Cell[CellGroupData[{
Cell[1734125, 49197, 1734, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:2033377127", "Proof|ID:2033377127-1"},
 CellID->1477321448],
Cell[1735862, 49256, 549, 13, 70, "ProofInfoBody",
 CellID->151991483]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1736460, 49275, 176, 3, 30, "Subsection",
 CellID->294752926],
Cell[1736639, 49280, 46, 1, 70, "OpenEnvironment",
 CellID->67511843],
Cell[CellGroupData[{
Cell[1736710, 49285, 393, 9, 70, "EnvironmentHeader",
 CellID->1145845750],
Cell[1737106, 49296, 779, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:982813600", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "367"},
 CellID->982813600],
Cell[1737888, 49319, 65, 1, 70, "EndEnvironmentMarker",
 CellID->482550205]
}, Open  ]],
Cell[1737968, 49323, 49, 1, 70, "CloseEnvironment",
 CellID->1572717722],
Cell[CellGroupData[{
Cell[1738042, 49328, 1725, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:982813600", "Proof|ID:982813600-1"},
 CellID->1236106547],
Cell[1739770, 49386, 480, 12, 70, "ProofInfoBody",
 CellID->475654809]
}, Closed]],
Cell[1740265, 49401, 48, 1, 70, "OpenEnvironment",
 CellID->1467132509],
Cell[CellGroupData[{
Cell[1740338, 49406, 340, 9, 70, "EnvironmentHeader",
 CellID->482418387],
Cell[1740681, 49417, 776, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1179937112", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1179937112],
Cell[1741460, 49439, 1646, 48, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:134072435", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "336"},
 CellID->134072435],
Cell[1743109, 49489, 1121, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1484092370", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "337"},
 CellID->1484092370],
Cell[1744233, 49524, 65, 1, 70, "EndEnvironmentMarker",
 CellID->870268060]
}, Open  ]],
Cell[1744313, 49528, 48, 1, 70, "CloseEnvironment",
 CellID->225373014],
Cell[CellGroupData[{
Cell[1744386, 49533, 5829, 144, 70, "ProofInfo",
 CellTags->{"Proof|ID:134072435", "Proof|ID:134072435-1"},
 CellID->77495342],
Cell[1750218, 49679, 481, 12, 70, "ProofInfoBody",
 CellID->1008758752]
}, Closed]],
Cell[CellGroupData[{
Cell[1750736, 49696, 3608, 99, 70, "ProofInfo",
 CellTags->{"Proof|ID:1484092370", "Proof|ID:1484092370-1"},
 CellID->1348971853],
Cell[1754347, 49797, 481, 12, 70, "ProofInfoBody",
 CellID->660891557]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1754877, 49815, 244, 5, 30, "Subsection",
 CellID->1409824068],
Cell[1755124, 49822, 47, 1, 12, "OpenEnvironment",
 CellID->888657849],
Cell[CellGroupData[{
Cell[1755196, 49827, 423, 11, 43, "EnvironmentHeader",
 CellID->1708398684],
Cell[1755622, 49840, 1026, 27, 44, "GlobalDeclaration",
 CellTags->{
  "ID:1480600758", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1480600758],
Cell[1756651, 49869, 739, 19, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1004003660", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1004003660],
Cell[1757393, 49890, 1075, 28, 56, "GlobalDeclaration",
 CellTags->{
  "ID:1854976476", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1854976476],
Cell[1758471, 49920, 1867, 50, 58, "FormalTextInputFormula",
 CellTags->{
  "ID:408134090", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduceM \[ScriptCapitalP]"},
 CellID->408134090],
Cell[1760341, 49972, 2604, 72, 89, "FormalTextInputFormula",
 CellTags->{
  "ID:1088018783", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reduceM \[ScriptCapitalP] II"},
 CellID->1088018783],
Cell[1762948, 50046, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1032173464]
}, Open  ]],
Cell[1763029, 50050, 49, 1, 16, "CloseEnvironment",
 CellID->2022036672],
Cell[CellGroupData[{
Cell[1763103, 50055, 8051, 192, 28, "ProofInfo",
 CellTags->{"Proof|ID:408134090", "Proof|ID:408134090-1"},
 CellID->59633542],
Cell[1771157, 50249, 480, 12, 70, "ProofInfoBody",
 CellID->199214208]
}, Closed]],
Cell[CellGroupData[{
Cell[1771674, 50266, 10512, 242, 26, "ProofInfo",
 CellTags->{"Proof|ID:1088018783", "Proof|ID:1088018783-1"},
 CellID->306221528],
Cell[1782189, 50510, 482, 12, 70, "ProofInfoBody",
 CellID->1372373678]
}, Closed]],
Cell[1782686, 50525, 48, 1, 12, "OpenEnvironment",
 CellID->1591405130],
Cell[CellGroupData[{
Cell[1782759, 50530, 497, 13, 43, "EnvironmentHeader",
 CellID->1153660425],
Cell[1783259, 50545, 953, 25, 66, "GlobalDeclaration",
 CellTags->{
  "ID:1964871967", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1964871967],
Cell[1784215, 50572, 506, 11, 36, "EnvironmentText",
 CellID->197410698],
Cell[1784724, 50585, 1865, 53, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1041722017", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "407"},
 CellID->1041722017],
Cell[1786592, 50640, 738, 19, 31, "GlobalDeclaration",
 CellTags->{
  "ID:1518222527", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1518222527],
Cell[1787333, 50661, 2380, 68, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1070088161", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "408"},
 CellID->1070088161],
Cell[1789716, 50731, 2361, 68, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1970014385", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "408'"},
 CellID->1970014385],
Cell[1792080, 50801, 64, 1, 26, "EndEnvironmentMarker",
 CellID->34978639]
}, Open  ]],
Cell[1792159, 50805, 48, 1, 16, "CloseEnvironment",
 CellID->778747754],
Cell[CellGroupData[{
Cell[1792232, 50810, 5663, 140, 28, "ProofInfo",
 CellTags->{"Proof|ID:1041722017", "Proof|ID:1041722017-1"},
 CellID->178605949],
Cell[1797898, 50952, 480, 12, 70, "ProofInfoBody",
 CellID->13157926]
}, Closed]],
Cell[CellGroupData[{
Cell[1798415, 50969, 8372, 195, 26, "ProofInfo",
 CellTags->{"Proof|ID:1070088161", "Proof|ID:1070088161-1"},
 CellID->1155517064],
Cell[1806790, 51166, 481, 12, 70, "ProofInfoBody",
 CellID->745976973]
}, Closed]],
Cell[CellGroupData[{
Cell[1807308, 51183, 8354, 193, 26, "ProofInfo",
 CellTags->{"Proof|ID:1970014385", "Proof|ID:1970014385-1"},
 CellID->625973727],
Cell[1815665, 51378, 481, 12, 70, "ProofInfoBody",
 CellID->285408563]
}, Closed]],
Cell[1816161, 51393, 48, 1, 12, "OpenEnvironment",
 CellID->1042798326],
Cell[CellGroupData[{
Cell[1816234, 51398, 479, 12, 43, "EnvironmentHeader",
 CellID->2008480134],
Cell[1816716, 51412, 786, 21, 42, "GlobalDeclaration",
 CellTags->{
  "ID:102721391", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->102721391],
Cell[1817505, 51435, 2568, 73, 120, "FormalTextInputFormula",
 CellTags->{
  "ID:657593834", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED"},
 CellID->657593834],
Cell[1820076, 51510, 2296, 62, 92, "FormalTextInputFormula",
 CellTags->{
  "ID:1802485517", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED II"},
 CellID->1802485517],
Cell[1822375, 51574, 2242, 60, 92, "FormalTextInputFormula",
 CellTags->{
  "ID:1164884865", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED III"},
 CellID->1164884865],
Cell[1824620, 51636, 2619, 72, 92, "FormalTextInputFormula",
 CellTags->{
  "ID:958945878", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED IV"},
 CellID->958945878],
Cell[1827242, 51710, 886, 24, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1601029605", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1601029605],
Cell[1828131, 51736, 2247, 62, 120, "FormalTextInputFormula",
 CellTags->{
  "ID:687002895", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED monomial"},
 CellID->687002895],
Cell[1830381, 51800, 1920, 55, 64, "FormalTextInputFormula",
 CellTags->{
  "ID:413658239", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "RED monomial II"},
 CellID->413658239],
Cell[1832304, 51857, 66, 1, 26, "EndEnvironmentMarker",
 CellID->2087609660]
}, Open  ]],
Cell[1832385, 51861, 49, 1, 16, "CloseEnvironment",
 CellID->1936134947],
Cell[CellGroupData[{
Cell[1832459, 51866, 8060, 188, 28, "ProofInfo",
 CellTags->{"Proof|ID:657593834", "Proof|ID:657593834-1"},
 CellID->17884315],
Cell[1840522, 52056, 480, 12, 70, "ProofInfoBody",
 CellID->303231202]
}, Closed]],
Cell[CellGroupData[{
Cell[1841039, 52073, 6585, 157, 26, "ProofInfo",
 CellTags->{"Proof|ID:1802485517", "Proof|ID:1802485517-1"},
 CellID->17136843],
Cell[1847627, 52232, 481, 12, 70, "ProofInfoBody",
 CellID->330024022]
}, Closed]],
Cell[CellGroupData[{
Cell[1848145, 52249, 6417, 153, 26, "ProofInfo",
 CellTags->{"Proof|ID:1164884865", "Proof|ID:1164884865-1"},
 CellID->1415542183],
Cell[1854565, 52404, 481, 12, 70, "ProofInfoBody",
 CellID->205712485]
}, Closed]],
Cell[CellGroupData[{
Cell[1855083, 52421, 8038, 184, 26, "ProofInfo",
 CellTags->{"Proof|ID:958945878", "Proof|ID:958945878-1"},
 CellID->1172023358],
Cell[1863124, 52607, 481, 12, 70, "ProofInfoBody",
 CellID->1468364448]
}, Closed]],
Cell[CellGroupData[{
Cell[1863642, 52624, 6947, 162, 26, "ProofInfo",
 CellTags->{"Proof|ID:687002895", "Proof|ID:687002895-1"},
 CellID->563304177],
Cell[1870592, 52788, 480, 12, 70, "ProofInfoBody",
 CellID->889845776]
}, Closed]],
Cell[CellGroupData[{
Cell[1871109, 52805, 6283, 150, 26, "ProofInfo",
 CellTags->{"Proof|ID:413658239", "Proof|ID:413658239-1"},
 CellID->885739272],
Cell[1877395, 52957, 480, 12, 70, "ProofInfoBody",
 CellID->881833888]
}, Closed]],
Cell[1877890, 52972, 47, 1, 12, "OpenEnvironment",
 CellID->571707933],
Cell[CellGroupData[{
Cell[1877962, 52977, 338, 9, 43, "EnvironmentHeader",
 CellID->480000138],
Cell[1878303, 52988, 771, 20, 42, "GlobalDeclaration",
 CellTags->{
  "ID:580194792", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->580194792],
Cell[1879077, 53010, 2513, 75, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:1667457569", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "357"},
 CellID->1667457569],
Cell[1881593, 53087, 1878, 61, 59, "FormalTextInputFormula",
 CellTags->{
  "ID:2035027156", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] lm \[DoubleRightArrow]"},
 CellID->2035027156],
Cell[1883474, 53150, 1822, 56, 59, "FormalTextInputFormula",
 CellTags->{
  "ID:2026457432", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] lm \[DoubleLeftArrow]"},
 CellID->2026457432],
Cell[1885299, 53208, 2433, 70, 64, "FormalTextInputFormula",
 CellTags->{
  "ID:1167741853", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "reducibility \[ScriptCapitalP]"},
 CellID->1167741853],
Cell[1887735, 53280, 3955, 105, 110, "FormalTextInputFormula",
 CellTags->{
  "ID:1998060487", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[RightArrow] alternative \[ScriptCapitalP]"},
 CellID->1998060487],
Cell[1891693, 53387, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1937013867]
}, Open  ]],
Cell[1891774, 53391, 48, 1, 16, "CloseEnvironment",
 CellID->956674854],
Cell[CellGroupData[{
Cell[1891847, 53396, 7280, 174, 28, "ProofInfo",
 CellTags->{"Proof|ID:1667457569", "Proof|ID:1667457569-1"},
 CellID->60009841],
Cell[1899130, 53572, 481, 12, 70, "ProofInfoBody",
 CellID->758686892]
}, Closed]],
Cell[CellGroupData[{
Cell[1899648, 53589, 5119, 133, 26, "ProofInfo",
 CellTags->{"Proof|ID:2035027156", "Proof|ID:2035027156-1"},
 CellID->322996175],
Cell[1904770, 53724, 481, 12, 70, "ProofInfoBody",
 CellID->209011635]
}, Closed]],
Cell[CellGroupData[{
Cell[1905288, 53741, 5145, 133, 26, "ProofInfo",
 CellTags->{"Proof|ID:2026457432", "Proof|ID:2026457432-1"},
 CellID->1357322543],
Cell[1910436, 53876, 526, 13, 70, "ProofInfoBody",
 CellID->1814571542]
}, Closed]],
Cell[CellGroupData[{
Cell[1910999, 53894, 5950, 146, 26, "ProofInfo",
 CellTags->{"Proof|ID:1167741853", "Proof|ID:1167741853-1"},
 CellID->293140950],
Cell[1916952, 54042, 482, 12, 70, "ProofInfoBody",
 CellID->1342634176]
}, Closed]],
Cell[CellGroupData[{
Cell[1917471, 54059, 9124, 211, 26, "ProofInfo",
 CellTags->{"Proof|ID:1998060487", "Proof|ID:1998060487-1"},
 CellID->2076771442],
Cell[1926598, 54272, 481, 12, 70, "ProofInfoBody",
 CellID->754720624]
}, Closed]],
Cell[1927094, 54287, 48, 1, 12, "OpenEnvironment",
 CellID->1265125254],
Cell[CellGroupData[{
Cell[1927167, 54292, 479, 12, 43, "EnvironmentHeader",
 CellID->1808687580],
Cell[1927649, 54306, 927, 24, 66, "GlobalDeclaration",
 CellTags->{
  "ID:1709390289", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1709390289],
Cell[1928579, 54332, 4827, 129, 102, "FormalTextInputFormula",
 CellTags->{
  "ID:1713414293", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[LeftRightArrow] alternative \[ScriptCapitalP]"},
 CellID->1713414293],
Cell[1933409, 54463, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1615487369]
}, Open  ]],
Cell[1933490, 54467, 49, 1, 16, "CloseEnvironment",
 CellID->1793647858],
Cell[CellGroupData[{
Cell[1933564, 54472, 13367, 293, 28, "ProofInfo",
 CellTags->{"Proof|ID:1713414293", "Proof|ID:1713414293-1"},
 CellID->793804010],
Cell[1946934, 54767, 481, 12, 70, "ProofInfoBody",
 CellID->162174877]
}, Closed]],
Cell[CellGroupData[{
Cell[1947452, 54784, 389, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2046956770],
Cell[1947844, 54796, 8336, 270, 696, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1965960939],
Cell[1956183, 55068, 338, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1587336919]
}, Closed]],
Cell[CellGroupData[{
Cell[1956558, 55079, 244, 5, 22, "Subsubsection",
 CellID->1164115734],
Cell[1956805, 55086, 47, 1, 70, "OpenEnvironment",
 CellID->776720212],
Cell[CellGroupData[{
Cell[1956877, 55091, 381, 9, 70, "EnvironmentHeader",
 CellID->389987125],
Cell[1957261, 55102, 3791, 102, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1190720464", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducibleRec \[ScriptCapitalP]"},
 CellID->1190720464],
Cell[1961055, 55206, 65, 1, 70, "EndEnvironmentMarker",
 CellID->129235337]
}, Open  ]],
Cell[1961135, 55210, 49, 1, 70, "CloseEnvironment",
 CellID->1043867076],
Cell[1961187, 55213, 47, 1, 70, "OpenEnvironment",
 CellID->909379877],
Cell[CellGroupData[{
Cell[1961259, 55218, 451, 12, 70, "EnvironmentHeader",
 CellID->130808008],
Cell[1961713, 55232, 924, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:166529906", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->166529906],
Cell[1962640, 55257, 1304, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:844782115", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352.\[DoubleRightArrow]"},
 CellID->844782115],
Cell[1963947, 55289, 1106, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:953126445", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352.\[DoubleLeftArrow]"},
 CellID->953126445],
Cell[1965056, 55319, 1233, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:259818613", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "352"},
 CellID->259818613],
Cell[1966292, 55351, 2618, 73, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:182112345", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "353"},
 CellID->182112345],
Cell[1968913, 55426, 65, 1, 70, "EndEnvironmentMarker",
 CellID->506272053]
}, Open  ]],
Cell[1968993, 55430, 48, 1, 70, "CloseEnvironment",
 CellID->352819612],
Cell[CellGroupData[{
Cell[1969066, 55435, 3984, 102, 70, "ProofInfo",
 CellTags->{"Proof|ID:844782115", "Proof|ID:844782115-1"},
 CellID->1659422444],
Cell[1973053, 55539, 480, 12, 70, "ProofInfoBody",
 CellID->565248221]
}, Closed]],
Cell[CellGroupData[{
Cell[1973570, 55556, 3982, 102, 70, "ProofInfo",
 CellTags->{"Proof|ID:953126445", "Proof|ID:953126445-1"},
 CellID->1715888170],
Cell[1977555, 55660, 481, 12, 70, "ProofInfoBody",
 CellID->1477222806]
}, Closed]],
Cell[CellGroupData[{
Cell[1978073, 55677, 3993, 104, 70, "ProofInfo",
 CellTags->{"Proof|ID:259818613", "Proof|ID:259818613-1"},
 CellID->816722537],
Cell[1982069, 55783, 480, 12, 70, "ProofInfoBody",
 CellID->701939384]
}, Closed]],
Cell[CellGroupData[{
Cell[1982586, 55800, 6479, 154, 70, "ProofInfo",
 CellTags->{"Proof|ID:182112345", "Proof|ID:182112345-1"},
 CellID->908791481],
Cell[1989068, 55956, 480, 12, 70, "ProofInfoBody",
 CellID->640786301]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1989609, 55975, 177, 3, 30, "Subsection",
 CellID->1569279607],
Cell[1989789, 55980, 48, 1, 70, "OpenEnvironment",
 CellID->1776685170],
Cell[CellGroupData[{
Cell[1989862, 55985, 454, 12, 70, "EnvironmentHeader",
 CellID->1082744603],
Cell[1990319, 55999, 1060, 26, 70, "GlobalDeclaration",
 CellTags->{
  "ID:418362312", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->418362312],
Cell[1991382, 56027, 739, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2116744492", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->2116744492],
Cell[1992124, 56048, 2781, 79, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1746803457", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "391"},
 CellID->1746803457],
Cell[1994908, 56129, 826, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:899106457", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->899106457],
Cell[1995737, 56152, 1099, 28, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1157329004", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1157329004],
Cell[1996839, 56182, 2316, 62, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1927781160", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "387"},
 CellID->1927781160],
Cell[1999158, 56246, 65, 1, 70, "EndEnvironmentMarker",
 CellID->755751524]
}, Open  ]],
Cell[1999238, 56250, 48, 1, 70, "CloseEnvironment",
 CellID->977857587],
Cell[CellGroupData[{
Cell[1999311, 56255, 7744, 183, 70, "ProofInfo",
 CellTags->{"Proof|ID:1746803457", "Proof|ID:1746803457-1"},
 CellID->311879395],
Cell[2007058, 56440, 481, 12, 70, "ProofInfoBody",
 CellID->227046330]
}, Closed]],
Cell[CellGroupData[{
Cell[2007576, 56457, 11812, 271, 70, "ProofInfo",
 CellTags->{"Proof|ID:1927781160", "Proof|ID:1927781160-1"},
 CellID->235488096],
Cell[2019391, 56730, 481, 12, 70, "ProofInfoBody",
 CellID->808907405]
}, Closed]],
Cell[2019887, 56745, 48, 1, 70, "OpenEnvironment",
 CellID->1780168981],
Cell[CellGroupData[{
Cell[2019960, 56750, 395, 9, 70, "EnvironmentHeader",
 CellID->1235526869],
Cell[2020358, 56761, 781, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:934500559", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "368"},
 CellID->934500559],
Cell[2021142, 56784, 65, 1, 70, "EndEnvironmentMarker",
 CellID->887114206]
}, Open  ]],
Cell[2021222, 56788, 49, 1, 70, "CloseEnvironment",
 CellID->1446998866],
Cell[CellGroupData[{
Cell[2021296, 56793, 1723, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:934500559", "Proof|ID:934500559-1"},
 CellID->80140931],
Cell[2023022, 56851, 477, 12, 70, "ProofInfoBody",
 CellID->426435]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2023548, 56869, 176, 3, 30, "Subsection",
 CellID->409755120],
Cell[2023727, 56874, 47, 1, 70, "OpenEnvironment",
 CellID->853392754],
Cell[CellGroupData[{
Cell[2023799, 56879, 528, 13, 70, "EnvironmentHeader",
 CellID->87372790],
Cell[2024330, 56894, 1088, 28, 70, "GlobalDeclaration",
 CellTags->{
  "ID:371274476", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->371274476],
Cell[2025421, 56924, 2225, 62, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1588964752", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR"},
 CellID->1588964752],
Cell[2027649, 56988, 2461, 70, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:428389793", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR II"},
 CellID->428389793],
Cell[2030113, 57060, 1800, 53, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:225111715", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CORR"},
 CellID->225111715],
Cell[2031916, 57115, 915, 25, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1632881403", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1632881403],
Cell[2032834, 57142, 2042, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1701548086", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "IRR \[CenterDot]"},
 CellID->1701548086],
Cell[2034879, 57200, 1266, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1758251641", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CORR \[CenterDot]"},
 CellID->1758251641],
Cell[2036148, 57236, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1086037741]
}, Open  ]],
Cell[2036229, 57240, 47, 1, 70, "CloseEnvironment",
 CellID->97059366],
Cell[CellGroupData[{
Cell[2036301, 57245, 8654, 197, 70, "ProofInfo",
 CellTags->{"Proof|ID:1588964752", "Proof|ID:1588964752-1"},
 CellID->543579358],
Cell[2044958, 57444, 481, 12, 70, "ProofInfoBody",
 CellID->381837386]
}, Closed]],
Cell[CellGroupData[{
Cell[2045476, 57461, 9787, 218, 70, "ProofInfo",
 CellTags->{"Proof|ID:428389793", "Proof|ID:428389793-1"},
 CellID->490546380],
Cell[2055266, 57681, 481, 12, 70, "ProofInfoBody",
 CellID->1790028040]
}, Closed]],
Cell[CellGroupData[{
Cell[2055784, 57698, 6091, 146, 70, "ProofInfo",
 CellTags->{"Proof|ID:225111715", "Proof|ID:225111715-1"},
 CellID->657130363],
Cell[2061878, 57846, 480, 12, 70, "ProofInfoBody",
 CellID->217274233]
}, Closed]],
Cell[CellGroupData[{
Cell[2062395, 57863, 9366, 211, 70, "ProofInfo",
 CellTags->{"Proof|ID:1701548086", "Proof|ID:1701548086-1"},
 CellID->384143515],
Cell[2071764, 58076, 482, 12, 70, "ProofInfoBody",
 CellID->1787100506]
}, Closed]],
Cell[CellGroupData[{
Cell[2072283, 58093, 5943, 141, 70, "ProofInfo",
 CellTags->{"Proof|ID:1758251641", "Proof|ID:1758251641-1"},
 CellID->1782924340],
Cell[2078229, 58236, 481, 12, 70, "ProofInfoBody",
 CellID->599328732]
}, Closed]],
Cell[2078725, 58251, 713, 14, 70, "Text",
 CellID->843394468],
Cell[2079441, 58267, 47, 1, 70, "OpenEnvironment",
 CellID->363999473],
Cell[CellGroupData[{
Cell[2079513, 58272, 410, 10, 70, "EnvironmentHeader",
 CellID->1652502037],
Cell[2079926, 58284, 939, 24, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1928324287", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1928324287],
Cell[2080868, 58310, 1839, 53, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:751250157", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404"},
 CellID->751250157],
Cell[2082710, 58365, 1866, 54, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1952511245", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404 II"},
 CellID->1952511245],
Cell[2084579, 58421, 2301, 64, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:537853020", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "404 III"},
 CellID->537853020],
Cell[2086883, 58487, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1877713982]
}, Open  ]],
Cell[2086964, 58491, 48, 1, 70, "CloseEnvironment",
 CellID->970753228],
Cell[CellGroupData[{
Cell[2087037, 58496, 6772, 164, 70, "ProofInfo",
 CellTags->{"Proof|ID:751250157", "Proof|ID:751250157-1"},
 CellID->1336000539],
Cell[2093812, 58662, 525, 13, 70, "ProofInfoBody",
 CellID->1708610251]
}, Closed]],
Cell[CellGroupData[{
Cell[2094374, 58680, 6393, 153, 70, "ProofInfo",
 CellTags->{"Proof|ID:1952511245", "Proof|ID:1952511245-1"},
 CellID->29876965],
Cell[2100770, 58835, 480, 12, 70, "ProofInfoBody",
 CellID->99609695]
}, Closed]],
Cell[CellGroupData[{
Cell[2101287, 58852, 7716, 178, 70, "ProofInfo",
 CellTags->{"Proof|ID:537853020", "Proof|ID:537853020-1"},
 CellID->1965128634],
Cell[2109006, 59032, 480, 12, 70, "ProofInfoBody",
 CellID->358062202]
}, Closed]],
Cell[2109501, 59047, 47, 1, 70, "OpenEnvironment",
 CellID->670924166],
Cell[CellGroupData[{
Cell[2109573, 59052, 424, 11, 70, "EnvironmentHeader",
 CellID->1216178880],
Cell[2110000, 59065, 963, 25, 70, "GlobalDeclaration",
 CellTags->{
  "ID:659046058", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->659046058],
Cell[2110966, 59092, 736, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:319520004", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->319520004],
Cell[2111705, 59113, 826, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:967033090", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->967033090],
Cell[2112534, 59136, 1098, 28, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1647700871", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1647700871],
Cell[2113635, 59166, 786, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1981597108", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1981597108],
Cell[2114424, 59189, 391, 7, 70, "EnvironmentText",
 CellID->130440480],
Cell[2114818, 59198, 2437, 67, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:40011060", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "388"},
 CellID->40011060],
Cell[2117258, 59267, 792, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:884639747", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->884639747],
Cell[2118053, 59290, 1524, 45, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1360655817", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1360655817],
Cell[2119580, 59337, 1516, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1002690147", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "389"},
 CellID->1002690147],
Cell[2121099, 59383, 1485, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1896907436", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "390"},
 CellID->1896907436],
Cell[2122587, 59429, 66, 1, 70, "EndEnvironmentMarker",
 CellID->2045347914]
}, Open  ]],
Cell[2122668, 59433, 49, 1, 70, "CloseEnvironment",
 CellID->1856156737],
Cell[CellGroupData[{
Cell[2122742, 59438, 16067, 356, 70, "ProofInfo",
 CellTags->{"Proof|ID:40011060", "Proof|ID:40011060-1"},
 CellID->1454431248],
Cell[2138812, 59796, 480, 12, 70, "ProofInfoBody",
 CellID->1520544499]
}, Closed]],
Cell[CellGroupData[{
Cell[2139329, 59813, 8812, 205, 70, "ProofInfo",
 CellTags->{"Proof|ID:1002690147", "Proof|ID:1002690147-1"},
 CellID->615225779],
Cell[2148144, 60020, 481, 12, 70, "ProofInfoBody",
 CellID->543645298]
}, Closed]],
Cell[CellGroupData[{
Cell[2148662, 60037, 8924, 207, 70, "ProofInfo",
 CellTags->{"Proof|ID:1896907436", "Proof|ID:1896907436-1"},
 CellID->375199501],
Cell[2157589, 60246, 482, 12, 70, "ProofInfoBody",
 CellID->1428296947]
}, Closed]],
Cell[2158086, 60261, 48, 1, 70, "OpenEnvironment",
 CellID->1502411461],
Cell[CellGroupData[{
Cell[2158159, 60266, 397, 9, 70, "EnvironmentHeader",
 CellID->2093688270],
Cell[2158559, 60277, 781, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1290694758", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "369"},
 CellID->1290694758],
Cell[2159343, 60300, 65, 1, 70, "EndEnvironmentMarker",
 CellID->902566628]
}, Open  ]],
Cell[2159423, 60304, 48, 1, 70, "CloseEnvironment",
 CellID->163822355],
Cell[CellGroupData[{
Cell[2159496, 60309, 1733, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1290694758", "Proof|ID:1290694758-1"},
 CellID->796556125],
Cell[2161232, 60368, 482, 12, 70, "ProofInfoBody",
 CellID->1734938744]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2161763, 60386, 314, 6, 30, "Subsection",
 CellID->387021889],
Cell[2162080, 60394, 48, 1, 12, "OpenEnvironment",
 CellID->1641500050],
Cell[CellGroupData[{
Cell[2162153, 60399, 396, 9, 43, "EnvironmentHeader",
 CellID->1266577052],
Cell[2162552, 60410, 1046, 27, 71, "GlobalDeclaration",
 CellTags->{
  "ID:1785082178", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1785082178],
Cell[2163601, 60439, 840, 22, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1718264802", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1718264802],
Cell[2164444, 60463, 803, 21, 43, "GlobalDeclaration",
 CellTags->{
  "ID:752304273", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->752304273],
Cell[2165250, 60486, 2765, 79, 145, "FormalTextInputFormula",
 CellTags->{
  "ID:578561747", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "areCommonReducibles I"},
 CellID->578561747],
Cell[2168018, 60567, 1563, 45, 85, "FormalTextInputFormula",
 CellTags->{
  "ID:1399766637", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "areCommonReducibles II"},
 CellID->1399766637],
Cell[2169584, 60614, 3366, 94, 145, "FormalTextInputFormula",
 CellTags->{
  "ID:1132709394", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP"},
 CellID->1132709394],
Cell[2172953, 60710, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1532258754]
}, Open  ]],
Cell[2173034, 60714, 49, 1, 16, "CloseEnvironment",
 CellID->1086432727],
Cell[2173086, 60717, 46, 1, 12, "OpenEnvironment",
 CellID->45003320],
Cell[CellGroupData[{
Cell[2173157, 60722, 459, 11, 43, "EnvironmentHeader",
 CellID->1810500356],
Cell[2173619, 60735, 1018, 27, 71, "GlobalDeclaration",
 CellTags->{
  "ID:176358731", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->176358731],
Cell[2174640, 60764, 885, 22, 42, "GlobalDeclaration",
 CellTags->{
  "ID:391507392", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->391507392],
Cell[2175528, 60788, 804, 21, 43, "GlobalDeclaration",
 CellTags->{
  "ID:1030826183", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1030826183],
Cell[2176335, 60811, 1752, 52, 78, "FormalTextInputFormula",
 CellTags->{
  "ID:926282876", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "354"},
 CellID->926282876],
Cell[2178090, 60865, 1749, 51, 91, "FormalTextInputFormula",
 CellTags->{
  "ID:80931797", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "355"},
 CellID->80931797],
Cell[2179842, 60918, 1448, 43, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:414980491", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "359"},
 CellID->414980491],
Cell[2181293, 60963, 908, 25, 34, "FormalTextInputFormula",
 CellTags->{
  "ID:744723554", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "345"},
 CellID->744723554],
Cell[2182204, 60990, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1560872126]
}, Open  ]],
Cell[2182285, 60994, 49, 1, 16, "CloseEnvironment",
 CellID->1557777283],
Cell[CellGroupData[{
Cell[2182359, 60999, 8291, 196, 28, "ProofInfo",
 CellTags->{"Proof|ID:926282876", "Proof|ID:926282876-1"},
 CellID->667882351],
Cell[2190653, 61197, 480, 12, 70, "ProofInfoBody",
 CellID->252898498]
}, Closed]],
Cell[CellGroupData[{
Cell[2191170, 61214, 8599, 196, 26, "ProofInfo",
 CellTags->{"Proof|ID:80931797", "Proof|ID:80931797-1"},
 CellID->1748139084],
Cell[2199772, 61412, 480, 12, 70, "ProofInfoBody",
 CellID->1485509901]
}, Closed]],
Cell[CellGroupData[{
Cell[2200289, 61429, 7515, 178, 26, "ProofInfo",
 CellTags->{"Proof|ID:414980491", "Proof|ID:414980491-1"},
 CellID->336718788],
Cell[2207807, 61609, 480, 12, 70, "ProofInfoBody",
 CellID->280631998]
}, Closed]],
Cell[CellGroupData[{
Cell[2208324, 61626, 5579, 140, 26, "ProofInfo",
 CellTags->{"Proof|ID:744723554", "Proof|ID:744723554-1"},
 CellID->1023806301],
Cell[2213906, 61768, 547, 13, 70, "ProofInfoBody",
 CellID->704570505]
}, Closed]],
Cell[2214468, 61784, 47, 1, 12, "OpenEnvironment",
 CellID->227618613],
Cell[CellGroupData[{
Cell[2214540, 61789, 395, 9, 43, "EnvironmentHeader",
 CellID->1038155297],
Cell[2214938, 61800, 1018, 27, 71, "GlobalDeclaration",
 CellTags->{
  "ID:381508689", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->381508689],
Cell[2215959, 61829, 741, 19, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1050210793", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1050210793],
Cell[2216703, 61850, 803, 21, 43, "GlobalDeclaration",
 CellTags->{
  "ID:512389161", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->512389161],
Cell[2217509, 61873, 1702, 51, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:719916043", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative 1"},
 CellID->719916043],
Cell[2219214, 61926, 1596, 49, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:864384559", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative 2"},
 CellID->864384559],
Cell[2220813, 61977, 2056, 57, 67, "FormalTextInputFormula",
 CellTags->{
  "ID:951944988", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isNTCRP alternative"},
 CellID->951944988],
Cell[2222872, 62036, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1683542722]
}, Open  ]],
Cell[2222953, 62040, 49, 1, 16, "CloseEnvironment",
 CellID->1871567538],
Cell[CellGroupData[{
Cell[2223027, 62045, 8098, 190, 28, "ProofInfo",
 CellTags->{"Proof|ID:719916043", "Proof|ID:719916043-1"},
 CellID->1378946613],
Cell[2231128, 62237, 481, 12, 70, "ProofInfoBody",
 CellID->2145134637]
}, Closed]],
Cell[CellGroupData[{
Cell[2231646, 62254, 7386, 174, 26, "ProofInfo",
 CellTags->{"Proof|ID:864384559", "Proof|ID:864384559-1"},
 CellID->1555291612],
Cell[2239035, 62430, 480, 12, 70, "ProofInfoBody",
 CellID->988348688]
}, Closed]],
Cell[CellGroupData[{
Cell[2239552, 62447, 9206, 209, 26, "ProofInfo",
 CellTags->{"Proof|ID:951944988", "Proof|ID:951944988-1"},
 CellID->1069274625],
Cell[2248761, 62658, 481, 12, 70, "ProofInfoBody",
 CellID->1430452337]
}, Closed]],
Cell[CellGroupData[{
Cell[2249279, 62675, 464, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->2144645966],
Cell[2249746, 62687, 332, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->809700523],
Cell[2250081, 62695, 1151, 29, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1717945330],
Cell[2251235, 62726, 1118, 27, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->639502933],
Cell[2252356, 62755, 19022, 576, 1848, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->533280863],
Cell[2271381, 63333, 338, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->209491385]
}, Closed]],
Cell[CellGroupData[{
Cell[2271756, 63344, 420, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1571042640],
Cell[2272179, 63356, 337, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1858320407],
Cell[2272519, 63364, 921, 24, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->977808820],
Cell[2273443, 63390, 15888, 496, 1666, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->961202380],
Cell[2289334, 63888, 335, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->277731394]
}, Closed]],
Cell[2289684, 63897, 47, 1, 12, "OpenEnvironment",
 CellID->552264247],
Cell[CellGroupData[{
Cell[2289756, 63902, 760, 16, 43, "EnvironmentHeader",
 CellID->226520904],
Cell[2290519, 63920, 1175, 29, 71, "GlobalDeclaration",
 CellTags->{
  "ID:1990838743", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1990838743],
Cell[2291697, 63951, 896, 23, 43, "GlobalDeclaration",
 CellTags->{
  "ID:20910152", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->20910152],
Cell[2292596, 63976, 3492, 91, 134, "FormalTextInputFormula",
 CellTags->{
  "ID:416591219", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleRightArrow].1"},
 CellID->416591219],
Cell[2296091, 64069, 2290, 63, 94, "FormalTextInputFormula",
 CellTags->{
  "ID:295118406", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleRightArrow].2"},
 CellID->295118406],
Cell[2298384, 64134, 4394, 121, 204, "FormalTextInputFormula",
 CellTags->{
  "ID:1583101849", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1.\[DoubleLeftArrow]"},
 CellID->1583101849],
Cell[2302781, 64257, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1370471321]
}, Open  ]],
Cell[2302862, 64261, 48, 1, 16, "CloseEnvironment",
 CellID->479379292],
Cell[CellGroupData[{
Cell[2302935, 64266, 11266, 250, 28, "ProofInfo",
 CellTags->{"Proof|ID:416591219", "Proof|ID:416591219-1"},
 CellID->1006102583],
Cell[2314204, 64518, 481, 12, 70, "ProofInfoBody",
 CellID->1404893208]
}, Closed]],
Cell[CellGroupData[{
Cell[2314722, 64535, 8718, 198, 26, "ProofInfo",
 CellTags->{"Proof|ID:295118406", "Proof|ID:295118406-1"},
 CellID->1475020847],
Cell[2323443, 64735, 481, 12, 70, "ProofInfoBody",
 CellID->1001169367]
}, Closed]],
Cell[CellGroupData[{
Cell[2323961, 64752, 13737, 297, 26, "ProofInfo",
 CellTags->{"Proof|ID:1583101849", "Proof|ID:1583101849-1"},
 CellID->2049393323],
Cell[2337701, 65051, 482, 12, 70, "ProofInfoBody",
 CellID->1721289092]
}, Closed]],
Cell[CellGroupData[{
Cell[2338220, 65068, 548, 12, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->534711775],
Cell[2338771, 65082, 36023, 1133, 2351, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1006778759],
Cell[2374797, 66217, 336, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->283251415]
}, Closed]],
Cell[CellGroupData[{
Cell[2375170, 66228, 459, 13, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->201585227],
Cell[2375632, 66243, 26918, 835, 2112, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1645016545],
Cell[2402553, 67080, 335, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1172329439]
}, Closed]],
Cell[2402903, 67089, 467, 10, 43, "Text",
 CellID->328625772],
Cell[2403373, 67101, 48, 1, 12, "OpenEnvironment",
 CellID->1707859469],
Cell[CellGroupData[{
Cell[2403446, 67106, 498, 12, 43, "EnvironmentHeader",
 CellID->1044109371],
Cell[2403947, 67120, 1086, 26, 66, "GlobalDeclaration",
 CellTags->{
  "ID:1272506214", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1272506214],
Cell[2405036, 67148, 2896, 80, 127, "FormalTextInputFormula",
 CellTags->{
  "ID:386901098", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleRightArrow].1"},
 CellID->386901098],
Cell[2407935, 67230, 2132, 61, 64, "FormalTextInputFormula",
 CellTags->{
  "ID:670300270", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleRightArrow].2"},
 CellID->670300270],
Cell[2410070, 67293, 4419, 125, 172, "FormalTextInputFormula",
 CellTags->{
  "ID:1434882124", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2.\[DoubleLeftArrow]"},
 CellID->1434882124],
Cell[2414492, 67420, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1079407213]
}, Open  ]],
Cell[2414573, 67424, 49, 1, 16, "CloseEnvironment",
 CellID->1088471095],
Cell[CellGroupData[{
Cell[2414647, 67429, 8691, 198, 28, "ProofInfo",
 CellTags->{"Proof|ID:386901098", "Proof|ID:386901098-1"},
 CellID->12812135],
Cell[2423341, 67629, 480, 12, 70, "ProofInfoBody",
 CellID->168109679]
}, Closed]],
Cell[CellGroupData[{
Cell[2423858, 67646, 5665, 138, 26, "ProofInfo",
 CellTags->{"Proof|ID:670300270", "Proof|ID:670300270-1"},
 CellID->488830292],
Cell[2429526, 67786, 480, 12, 70, "ProofInfoBody",
 CellID->565239118]
}, Closed]],
Cell[CellGroupData[{
Cell[2430043, 67803, 11138, 247, 26, "ProofInfo",
 CellTags->{"Proof|ID:1434882124", "Proof|ID:1434882124-1"},
 CellID->822190239],
Cell[2441184, 68052, 482, 12, 70, "ProofInfoBody",
 CellID->1425736080]
}, Closed]],
Cell[CellGroupData[{
Cell[2441703, 68069, 558, 12, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1621914703],
Cell[2442264, 68083, 332, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->414334676],
Cell[2442599, 68091, 732, 19, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2054694430],
Cell[2443334, 68112, 10464, 326, 1128, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1844460214],
Cell[2453801, 68440, 365, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1726142358]
}, Closed]],
Cell[CellGroupData[{
Cell[2454203, 68452, 525, 11, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->458708427],
Cell[2454731, 68465, 18832, 568, 1498, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->936066289],
Cell[2473566, 69035, 335, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->898836392]
}, Closed]],
Cell[CellGroupData[{
Cell[2473938, 69046, 478, 13, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1317000955],
Cell[2474419, 69061, 19812, 620, 1841, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1205890748],
Cell[2494234, 69683, 336, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1499697181]
}, Closed]],
Cell[2494585, 69692, 48, 1, 12, "OpenEnvironment",
 CellID->1580923328],
Cell[CellGroupData[{
Cell[2494658, 69697, 414, 10, 43, "EnvironmentHeader",
 CellID->141373954],
Cell[2495075, 69709, 1008, 27, 71, "GlobalDeclaration",
 CellTags->{
  "ID:737627449", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->737627449],
Cell[2496086, 69738, 737, 19, 42, "GlobalDeclaration",
 CellTags->{
  "ID:81313572", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->81313572],
Cell[2496826, 69759, 767, 20, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1098402572", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1098402572],
Cell[2497596, 69781, 2466, 71, 109, "FormalTextInputFormula",
 CellTags->{
  "ID:1499935354", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 1"},
 CellID->1499935354],
Cell[2500065, 69854, 2698, 73, 94, "FormalTextInputFormula",
 CellTags->{
  "ID:1938671644", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 2"},
 CellID->1938671644],
Cell[2502766, 69929, 2305, 64, 89, "FormalTextInputFormula",
 CellTags->{
  "ID:48476485", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR1 monomial 3"},
 CellID->48476485],
Cell[2505074, 69995, 2004, 59, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:1246936946", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 1"},
 CellID->1246936946],
Cell[2507081, 70056, 2186, 62, 64, "FormalTextInputFormula",
 CellTags->{
  "ID:1856728487", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 2"},
 CellID->1856728487],
Cell[2509270, 70120, 1720, 50, 64, "FormalTextInputFormula",
 CellTags->{
  "ID:335657318", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR2 monomial 3"},
 CellID->335657318],
Cell[2510993, 70172, 65, 1, 26, "EndEnvironmentMarker",
 CellID->575968727]
}, Open  ]],
Cell[2511073, 70176, 49, 1, 16, "CloseEnvironment",
 CellID->1580711919],
Cell[CellGroupData[{
Cell[2511147, 70181, 8905, 205, 28, "ProofInfo",
 CellTags->{"Proof|ID:1499935354", "Proof|ID:1499935354-1"},
 CellID->1869006459],
Cell[2520055, 70388, 482, 12, 70, "ProofInfoBody",
 CellID->1447063924]
}, Closed]],
Cell[CellGroupData[{
Cell[2520574, 70405, 9347, 209, 26, "ProofInfo",
 CellTags->{"Proof|ID:1938671644", "Proof|ID:1938671644-1"},
 CellID->1781620971],
Cell[2529924, 70616, 550, 13, 70, "ProofInfoBody",
 CellID->1479477140]
}, Closed]],
Cell[CellGroupData[{
Cell[2530511, 70634, 8644, 196, 26, "ProofInfo",
 CellTags->{"Proof|ID:48476485", "Proof|ID:48476485-1"},
 CellID->528453611],
Cell[2539158, 70832, 480, 12, 70, "ProofInfoBody",
 CellID->2072605463]
}, Closed]],
Cell[CellGroupData[{
Cell[2539675, 70849, 6288, 153, 26, "ProofInfo",
 CellTags->{"Proof|ID:1246936946", "Proof|ID:1246936946-1"},
 CellID->299300275],
Cell[2545966, 71004, 481, 12, 70, "ProofInfoBody",
 CellID->661794959]
}, Closed]],
Cell[CellGroupData[{
Cell[2546484, 71021, 5697, 139, 26, "ProofInfo",
 CellTags->{"Proof|ID:1856728487", "Proof|ID:1856728487-1"},
 CellID->1800537912],
Cell[2552184, 71162, 481, 12, 70, "ProofInfoBody",
 CellID->434287860]
}, Closed]],
Cell[CellGroupData[{
Cell[2552702, 71179, 5515, 134, 26, "ProofInfo",
 CellTags->{"Proof|ID:335657318", "Proof|ID:335657318-1"},
 CellID->218427358],
Cell[2558220, 71315, 480, 12, 70, "ProofInfoBody",
 CellID->948824583]
}, Closed]],
Cell[CellGroupData[{
Cell[2558737, 71332, 506, 11, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->601702509],
Cell[2559246, 71345, 333, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->761991808],
Cell[2559582, 71353, 654, 15, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1823066290],
Cell[2560239, 71370, 796, 21, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1830447192],
Cell[2561038, 71393, 11452, 363, 1081, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1138449629],
Cell[2572493, 71758, 334, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->72163444]
}, Closed]],
Cell[CellGroupData[{
Cell[2572864, 71769, 445, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->693637141],
Cell[2573312, 71781, 381, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2031424469],
Cell[2573696, 71789, 801, 17, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->632880062],
Cell[2574500, 71808, 714, 14, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1560665101],
Cell[2575217, 71824, 9093, 285, 1015, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1096597969],
Cell[2584313, 72111, 385, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->979676413]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2584747, 72124, 331, 6, 30, "Subsection",
 CellID->289777292],
Cell[2585081, 72132, 48, 1, 12, "OpenEnvironment",
 CellID->1231465134],
Cell[CellGroupData[{
Cell[2585154, 72137, 375, 9, 43, "EnvironmentHeader",
 CellID->1786858939],
Cell[2585532, 72148, 2769, 80, 137, "FormalTextInputFormula",
 CellTags->{
  "ID:624517207", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMNTCRP"},
 CellID->624517207],
Cell[2588304, 72230, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1789189429]
}, Open  ]],
Cell[2588385, 72234, 49, 1, 16, "CloseEnvironment",
 CellID->2019724429],
Cell[2588437, 72237, 324, 7, 24, "Text",
 CellID->840274992],
Cell[2588764, 72246, 47, 1, 12, "OpenEnvironment",
 CellID->967326008],
Cell[CellGroupData[{
Cell[2588836, 72251, 394, 9, 43, "EnvironmentHeader",
 CellID->147093162],
Cell[2589233, 72262, 3080, 86, 159, "FormalTextInputFormula",
 CellTags->{
  "ID:1172537415", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMNTCRP alternative"},
 CellID->1172537415],
Cell[2592316, 72350, 65, 1, 26, "EndEnvironmentMarker",
 CellID->462144833]
}, Open  ]],
Cell[2592396, 72354, 48, 1, 16, "CloseEnvironment",
 CellID->866823809],
Cell[CellGroupData[{
Cell[2592469, 72359, 9223, 210, 28, "ProofInfo",
 CellTags->{"Proof|ID:1172537415", "Proof|ID:1172537415-1"},
 CellID->1389429208],
Cell[2601695, 72571, 482, 12, 70, "ProofInfoBody",
 CellID->1096557679]
}, Closed]],
Cell[CellGroupData[{
Cell[2602214, 72588, 292, 8, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->297887619],
Cell[2602509, 72598, 9682, 305, 1334, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->690232938],
Cell[2612194, 72905, 242, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->446746978]
}, Closed]],
Cell[2612451, 72913, 386, 9, 24, "Text",
 CellID->2030148506],
Cell[2612840, 72924, 251, 5, 24, "Text",
 CellID->1426279550],
Cell[2613094, 72931, 48, 1, 12, "OpenEnvironment",
 CellID->1983351237],
Cell[CellGroupData[{
Cell[2613167, 72936, 535, 13, 43, "EnvironmentHeader",
 CellID->2124804766],
Cell[2613705, 72951, 1084, 28, 71, "GlobalDeclaration",
 CellTags->{
  "ID:1861685863", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1861685863],
Cell[2614792, 72981, 806, 21, 43, "GlobalDeclaration",
 CellTags->{
  "ID:1154885365", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1154885365],
Cell[2615601, 73004, 1000, 26, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:1777331368", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].1"},
 CellID->1777331368],
Cell[2616604, 73032, 1375, 39, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1052103130", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].2"},
 CellID->1052103130],
Cell[2617982, 73073, 1127, 29, 34, "FormalTextInputFormula",
 CellTags->{
  "ID:1077684208", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleRightArrow].3"},
 CellID->1077684208],
Cell[2619112, 73104, 1660, 47, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:86200506", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3.\[DoubleLeftArrow]"},
 CellID->86200506],
Cell[2620775, 73153, 1938, 57, 75, "FormalTextInputFormula",
 CellTags->{
  "ID:394014296", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR3"},
 CellID->394014296],
Cell[2622716, 73212, 65, 1, 26, "EndEnvironmentMarker",
 CellID->333024971]
}, Open  ]],
Cell[2622796, 73216, 49, 1, 16, "CloseEnvironment",
 CellID->1951750829],
Cell[CellGroupData[{
Cell[2622870, 73221, 5896, 143, 28, "ProofInfo",
 CellTags->{"Proof|ID:1777331368", "Proof|ID:1777331368-1"},
 CellID->508661436],
Cell[2628769, 73366, 481, 12, 70, "ProofInfoBody",
 CellID->263326735]
}, Closed]],
Cell[CellGroupData[{
Cell[2629287, 73383, 6847, 163, 26, "ProofInfo",
 CellTags->{"Proof|ID:1052103130", "Proof|ID:1052103130-1"},
 CellID->1682055148],
Cell[2636137, 73548, 481, 12, 70, "ProofInfoBody",
 CellID->187463426]
}, Closed]],
Cell[CellGroupData[{
Cell[2636655, 73565, 6607, 157, 26, "ProofInfo",
 CellTags->{"Proof|ID:1077684208", "Proof|ID:1077684208-1"},
 CellID->1363636622],
Cell[2643265, 73724, 482, 12, 70, "ProofInfoBody",
 CellID->1445163384]
}, Closed]],
Cell[CellGroupData[{
Cell[2643784, 73741, 7545, 176, 26, "ProofInfo",
 CellTags->{"Proof|ID:86200506", "Proof|ID:86200506-1"},
 CellID->1645380276],
Cell[2651332, 73919, 480, 12, 70, "ProofInfoBody",
 CellID->1969346073]
}, Closed]],
Cell[CellGroupData[{
Cell[2651849, 73936, 9027, 208, 26, "ProofInfo",
 CellTags->{"Proof|ID:394014296", "Proof|ID:394014296-1"},
 CellID->1823310355],
Cell[2660879, 74146, 480, 12, 70, "ProofInfoBody",
 CellID->504163934]
}, Closed]],
Cell[CellGroupData[{
Cell[2661396, 74163, 409, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1024574146],
Cell[2661808, 74175, 336, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->101928936],
Cell[2662147, 74183, 638, 14, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->548195680],
Cell[2662788, 74199, 6823, 215, 853, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1217971235],
Cell[2669614, 74416, 338, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->313548928]
}, Closed]],
Cell[CellGroupData[{
Cell[2669989, 74427, 410, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->74747143],
Cell[2670402, 74439, 335, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1147238182],
Cell[2670740, 74447, 577, 14, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->933263416],
Cell[2671320, 74463, 9742, 308, 1035, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->89860634],
Cell[2681065, 74773, 345, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2002978456]
}, Closed]],
Cell[CellGroupData[{
Cell[2681447, 74784, 404, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->624408900],
Cell[2681854, 74796, 334, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1271771765],
Cell[2682191, 74804, 820, 22, 25, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->173464573],
Cell[2683014, 74828, 6367, 197, 875, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->454649756],
Cell[2689384, 75027, 337, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->493606769]
}, Closed]],
Cell[CellGroupData[{
Cell[2689758, 75038, 414, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1772726150],
Cell[2690175, 75050, 385, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1167894951],
Cell[2690563, 75058, 726, 17, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->759496477],
Cell[2691292, 75077, 639, 15, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->363005801],
Cell[2691934, 75094, 1346, 34, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->19508673],
Cell[2693283, 75130, 1171, 28, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1204606733],
Cell[2694457, 75160, 9913, 302, 1382, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->936576587],
Cell[2704373, 75464, 336, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1594670345]
}, Closed]],
Cell[2704724, 75473, 47, 1, 12, "OpenEnvironment",
 CellID->666606593],
Cell[CellGroupData[{
Cell[2704796, 75478, 483, 13, 43, "EnvironmentHeader",
 CellID->1056394693],
Cell[2705282, 75493, 994, 25, 66, "GlobalDeclaration",
 CellTags->{
  "ID:1444400550", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1444400550],
Cell[2706279, 75520, 965, 26, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:653031981", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].1"},
 CellID->653031981],
Cell[2707247, 75548, 1157, 33, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:943141844", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].2"},
 CellID->943141844],
Cell[2708407, 75583, 1190, 32, 34, "FormalTextInputFormula",
 CellTags->{
  "ID:1143014070", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleRightArrow].3"},
 CellID->1143014070],
Cell[2709600, 75617, 1536, 45, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:1630861163", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4.\[DoubleLeftArrow]"},
 CellID->1630861163],
Cell[2711139, 75664, 1755, 53, 75, "FormalTextInputFormula",
 CellTags->{
  "ID:1701661908", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CR4"},
 CellID->1701661908],
Cell[2712897, 75719, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1473391475]
}, Open  ]],
Cell[2712978, 75723, 48, 1, 16, "CloseEnvironment",
 CellID->973852342],
Cell[CellGroupData[{
Cell[2713051, 75728, 3642, 96, 28, "ProofInfo",
 CellTags->{"Proof|ID:653031981", "Proof|ID:653031981-1"},
 CellID->13059041],
Cell[2716696, 75826, 480, 12, 70, "ProofInfoBody",
 CellID->936750914]
}, Closed]],
Cell[CellGroupData[{
Cell[2717213, 75843, 4004, 104, 26, "ProofInfo",
 CellTags->{"Proof|ID:943141844", "Proof|ID:943141844-1"},
 CellID->1657492493],
Cell[2721220, 75949, 481, 12, 70, "ProofInfoBody",
 CellID->1516876266]
}, Closed]],
Cell[CellGroupData[{
Cell[2721738, 75966, 4211, 109, 26, "ProofInfo",
 CellTags->{"Proof|ID:1143014070", "Proof|ID:1143014070-1"},
 CellID->1936376328],
Cell[2725952, 76077, 479, 12, 70, "ProofInfoBody",
 CellID->7830102]
}, Closed]],
Cell[CellGroupData[{
Cell[2726468, 76094, 4528, 115, 26, "ProofInfo",
 CellTags->{"Proof|ID:1630861163", "Proof|ID:1630861163-1"},
 CellID->979966294],
Cell[2730999, 76211, 482, 12, 70, "ProofInfoBody",
 CellID->1309847775]
}, Closed]],
Cell[CellGroupData[{
Cell[2731518, 76228, 5955, 147, 26, "ProofInfo",
 CellTags->{"Proof|ID:1701661908", "Proof|ID:1701661908-1"},
 CellID->599419806],
Cell[2737476, 76377, 482, 12, 70, "ProofInfoBody",
 CellID->1584091188]
}, Closed]],
Cell[CellGroupData[{
Cell[2737995, 76394, 432, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->100124376],
Cell[2738430, 76406, 338, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1945351028],
Cell[2738771, 76414, 637, 14, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1341586550],
Cell[2739411, 76430, 6490, 205, 872, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1856712428],
Cell[2745904, 76637, 338, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->577063717]
}, Closed]],
Cell[CellGroupData[{
Cell[2746279, 76648, 461, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2028257075],
Cell[2746743, 76660, 332, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->997822543],
Cell[2747078, 76668, 578, 14, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1728450463],
Cell[2747659, 76684, 7681, 231, 909, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->127670836],
Cell[2755343, 76917, 342, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->684788075]
}, Closed]],
Cell[CellGroupData[{
Cell[2755722, 76928, 429, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1144589532],
Cell[2756154, 76940, 337, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->2105013134],
Cell[2756494, 76948, 820, 22, 25, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->354338712],
Cell[2757317, 76972, 5542, 171, 837, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1627125215],
Cell[2762862, 77145, 336, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1095627526]
}, Closed]],
Cell[CellGroupData[{
Cell[2763235, 77156, 433, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1999351486],
Cell[2763671, 77168, 383, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1769332380],
Cell[2764057, 77176, 721, 17, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->20695133],
Cell[2764781, 77195, 640, 15, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1281016174],
Cell[2765424, 77212, 1330, 34, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1750850628],
Cell[2766757, 77248, 1161, 29, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1983890089],
Cell[2767921, 77279, 9171, 280, 1325, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->699080535],
Cell[2777095, 77561, 335, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1233787267]
}, Closed]],
Cell[2777445, 77570, 48, 1, 12, "OpenEnvironment",
 CellID->1293427525],
Cell[CellGroupData[{
Cell[2777518, 77575, 605, 14, 43, "EnvironmentHeader",
 CellID->2035494476],
Cell[2778126, 77591, 1118, 28, 71, "GlobalDeclaration",
 CellTags->{
  "ID:1489378443", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1489378443],
Cell[2779247, 77621, 800, 21, 43, "GlobalDeclaration",
 CellTags->{
  "ID:196663580", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->196663580],
Cell[2780050, 77644, 775, 20, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1830388955", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1830388955],
Cell[2780828, 77666, 3394, 93, 155, "FormalTextInputFormula",
 CellTags->{
  "ID:312371876", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CP1.\[DoubleLeftArrow]"},
 CellID->312371876],
Cell[2784225, 77761, 4109, 110, 241, "FormalTextInputFormula",
 CellTags->{
  "ID:1006612102", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "CP2.\[DoubleLeftArrow]"},
 CellID->1006612102],
Cell[2788337, 77873, 65, 1, 26, "EndEnvironmentMarker",
 CellID->469925811]
}, Open  ]],
Cell[2788417, 77877, 48, 1, 16, "CloseEnvironment",
 CellID->633893238],
Cell[CellGroupData[{
Cell[2788490, 77882, 13507, 292, 28, "ProofInfo",
 CellTags->{"Proof|ID:312371876", "Proof|ID:312371876-1"},
 CellID->650652383],
Cell[2802000, 78176, 481, 12, 70, "ProofInfoBody",
 CellID->1483827355]
}, Closed]],
Cell[CellGroupData[{
Cell[2802518, 78193, 14246, 309, 26, "ProofInfo",
 CellTags->{"Proof|ID:1006612102", "Proof|ID:1006612102-1"},
 CellID->719415418],
Cell[2816767, 78504, 545, 13, 70, "ProofInfoBody",
 CellID->889161277]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2817361, 78523, 175, 3, 30, "Subsection",
 CellID->1847430199],
Cell[2817539, 78528, 649, 15, 36, "Text",
 CellID->1835280531],
Cell[2818191, 78545, 47, 1, 12, "OpenEnvironment",
 CellID->500926893],
Cell[CellGroupData[{
Cell[2818263, 78550, 334, 9, 43, "EnvironmentHeader",
 CellID->623726226],
Cell[2818600, 78561, 1020, 27, 71, "GlobalDeclaration",
 CellTags->{
  "ID:542206690", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->542206690],
Cell[2819623, 78590, 741, 19, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1548368737", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1548368737],
Cell[2820367, 78611, 798, 21, 43, "GlobalDeclaration",
 CellTags->{
  "ID:73478134", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->73478134],
Cell[2821168, 78634, 9746, 250, 423, "FormalTextInputFormula",
 CellTags->{
  "ID:310376772", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "409"},
 CellID->310376772],
Cell[2830917, 78886, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1898204898]
}, Open  ]],
Cell[2830998, 78890, 48, 1, 16, "CloseEnvironment",
 CellID->723680751],
Cell[CellGroupData[{
Cell[2831071, 78895, 28365, 586, 28, "ProofInfo",
 CellTags->{"Proof|ID:310376772", "Proof|ID:310376772-1"},
 CellID->1336876000],
Cell[2859439, 79483, 549, 13, 70, "ProofInfoBody",
 CellID->1675988726]
}, Closed]],
Cell[CellGroupData[{
Cell[2860025, 79501, 387, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2024136642],
Cell[2860415, 79513, 332, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->460882741],
Cell[2860750, 79521, 719, 16, 27, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1279103172],
Cell[2861472, 79539, 11457, 357, 1007, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->257775101],
Cell[2872932, 79898, 338, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->670854715]
}, Closed]],
Cell[2873285, 79907, 47, 1, 12, "OpenEnvironment",
 CellID->738627761],
Cell[CellGroupData[{
Cell[2873357, 79912, 485, 12, 43, "EnvironmentHeader",
 CellID->1572614892],
Cell[2873845, 79926, 840, 21, 42, "GlobalDeclaration",
 CellTags->{
  "ID:383946867", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->383946867],
Cell[2874688, 79949, 818, 21, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1587705005", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1587705005],
Cell[2875509, 79972, 387, 12, 25, "EnvironmentText",
 CellID->1670988629],
Cell[2875899, 79986, 3055, 95, 108, "FormalTextInputFormula",
 CellTags->{
  "ID:549840930", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].3"},
 CellID->549840930],
Cell[2878957, 80083, 4804, 151, 109, "FormalTextInputFormula",
 CellTags->{
  "ID:657656821", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].5"},
 CellID->657656821],
Cell[2883764, 80236, 934, 24, 66, "GlobalDeclaration",
 CellTags->{
  "ID:64232512", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->64232512],
Cell[2884701, 80262, 818, 21, 41, "GlobalDeclaration",
 CellTags->{
  "ID:237038572", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->237038572],
Cell[2885522, 80285, 718, 19, 31, "GlobalDeclaration",
 CellTags->{
  "ID:235501232", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->235501232],
Cell[2886243, 80306, 1796, 49, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1513648693", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].1"},
 CellID->1513648693],
Cell[2888042, 80357, 3126, 90, 52, "FormalTextInputFormula",
 CellTags->{
  "ID:808269926", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "R7.\[ScriptCapitalP].4"},
 CellID->808269926],
Cell[2891171, 80449, 65, 1, 26, "EndEnvironmentMarker",
 CellID->193282123]
}, Open  ]],
Cell[2891251, 80453, 48, 1, 16, "CloseEnvironment",
 CellID->970754585],
Cell[CellGroupData[{
Cell[2891324, 80458, 8316, 194, 28, "ProofInfo",
 CellTags->{"Proof|ID:549840930", "Proof|ID:549840930-1"},
 CellID->1394810391],
Cell[2899643, 80654, 480, 12, 70, "ProofInfoBody",
 CellID->460317596]
}, Closed]],
Cell[CellGroupData[{
Cell[2900160, 80671, 12906, 288, 26, "ProofInfo",
 CellTags->{"Proof|ID:657656821", "Proof|ID:657656821-1"},
 CellID->75572869],
Cell[2913069, 80961, 479, 12, 70, "ProofInfoBody",
 CellID->69597396]
}, Closed]],
Cell[CellGroupData[{
Cell[2913585, 80978, 6847, 163, 26, "ProofInfo",
 CellTags->{"Proof|ID:1513648693", "Proof|ID:1513648693-1"},
 CellID->1424817208],
Cell[2920435, 81143, 482, 12, 70, "ProofInfoBody",
 CellID->1180617287]
}, Closed]],
Cell[CellGroupData[{
Cell[2920954, 81160, 11047, 246, 26, "ProofInfo",
 CellTags->{"Proof|ID:808269926", "Proof|ID:808269926-1"},
 CellID->876551560],
Cell[2932004, 81408, 481, 12, 70, "ProofInfoBody",
 CellID->1906600593]
}, Closed]],
Cell[CellGroupData[{
Cell[2932522, 81425, 414, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1496383118],
Cell[2932939, 81437, 7308, 236, 718, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->673230165],
Cell[2940250, 81675, 343, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1990228853]
}, Closed]],
Cell[CellGroupData[{
Cell[2940630, 81686, 459, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1002411843],
Cell[2941092, 81698, 329, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1700926014],
Cell[2941424, 81706, 783, 19, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2000946050],
Cell[2942210, 81727, 842, 20, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->438531627],
Cell[2943055, 81749, 347, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1584658543],
Cell[2943405, 81757, 354, 6, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->895588669],
Cell[2943762, 81765, 366, 6, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1241320943],
Cell[2944131, 81773, 19470, 586, 1154, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1529086012],
Cell[2963604, 82361, 379, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->616246519]
}, Closed]],
Cell[CellGroupData[{
Cell[2964020, 82373, 416, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->331578966],
Cell[2964439, 82385, 338, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->200364281],
Cell[2964780, 82393, 1156, 32, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1473345246],
Cell[2965939, 82427, 332, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1709547035],
Cell[2966274, 82434, 436, 10, 36, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1610958980],
Cell[2966713, 82446, 7993, 242, 885, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->817801926],
Cell[2974709, 82690, 365, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1726226374]
}, Closed]],
Cell[CellGroupData[{
Cell[2975111, 82702, 411, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->225861683],
Cell[2975525, 82714, 8812, 288, 792, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->994840203],
Cell[2984340, 83004, 338, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->459605869]
}, Closed]],
Cell[2984693, 83013, 47, 1, 12, "OpenEnvironment",
 CellID->357963951],
Cell[CellGroupData[{
Cell[2984765, 83018, 393, 9, 43, "EnvironmentHeader",
 CellID->1656172280],
Cell[2985161, 83029, 779, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:255426543", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "370"},
 CellID->255426543],
Cell[2985943, 83052, 65, 1, 26, "EndEnvironmentMarker",
 CellID->940321280]
}, Open  ]],
Cell[2986023, 83056, 48, 1, 16, "CloseEnvironment",
 CellID->385752651],
Cell[CellGroupData[{
Cell[2986096, 83061, 1724, 56, 28, "ProofInfo",
 CellTags->{"Proof|ID:255426543", "Proof|ID:255426543-1"},
 CellID->537752064],
Cell[2987823, 83119, 481, 12, 70, "ProofInfoBody",
 CellID->1382560669]
}, Closed]],
Cell[CellGroupData[{
Cell[2988341, 83136, 390, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1814678652],
Cell[2988734, 83148, 336, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->109051318],
Cell[2989073, 83156, 810, 20, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->939173627],
Cell[2989886, 83178, 24648, 759, 1898, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2037769089],
Cell[3014537, 83939, 340, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->369172055]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3014926, 83951, 178, 3, 30, "Subsection",
 CellID->2100864518],
Cell[3015107, 83956, 284, 6, 24, "Text",
 CellID->164300636],
Cell[3015394, 83964, 47, 1, 12, "OpenEnvironment",
 CellID->342777797],
Cell[CellGroupData[{
Cell[3015466, 83969, 392, 9, 43, "EnvironmentHeader",
 CellID->1898194960],
Cell[3015861, 83980, 782, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1342888001", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "371"},
 CellID->1342888001],
Cell[3016646, 84003, 65, 1, 26, "EndEnvironmentMarker",
 CellID->376228414]
}, Open  ]],
Cell[3016726, 84007, 49, 1, 16, "CloseEnvironment",
 CellID->1126499261],
Cell[CellGroupData[{
Cell[3016800, 84012, 1734, 57, 28, "ProofInfo",
 CellTags->{"Proof|ID:1342888001", "Proof|ID:1342888001-1"},
 CellID->1391148747],
Cell[3018537, 84071, 481, 12, 70, "ProofInfoBody",
 CellID->837021914]
}, Closed]],
Cell[CellGroupData[{
Cell[3019055, 84088, 380, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->42990982],
Cell[3019438, 84100, 330, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->356763103],
Cell[3019771, 84108, 764, 19, 24, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1594516779],
Cell[3020538, 84129, 11060, 336, 906, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->891408206],
Cell[3031601, 84467, 331, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->94010900]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3031981, 84479, 211, 3, 30, "Subsection",
 CellID->1466756562],
Cell[3032195, 84484, 47, 1, 70, "OpenEnvironment",
 CellID->927538807],
Cell[CellGroupData[{
Cell[3032267, 84489, 516, 13, 70, "EnvironmentHeader",
 CellID->179190798],
Cell[3032786, 84504, 965, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1021018734", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "376"},
 CellID->1021018734],
Cell[3033754, 84530, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1086993414]
}, Open  ]],
Cell[3033835, 84534, 49, 1, 70, "CloseEnvironment",
 CellID->1230683486],
Cell[CellGroupData[{
Cell[3033909, 84539, 1767, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1021018734", "Proof|ID:1021018734-1"},
 CellID->1129550695],
Cell[3035679, 84598, 481, 12, 70, "ProofInfoBody",
 CellID->492049603]
}, Closed]],
Cell[3036175, 84613, 620, 12, 70, "Text",
 CellID->1207537800],
Cell[3036798, 84627, 47, 1, 70, "OpenEnvironment",
 CellID->543049971],
Cell[CellGroupData[{
Cell[3036870, 84632, 339, 9, 70, "EnvironmentHeader",
 CellID->1341395337],
Cell[3037212, 84643, 1659, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:459032599", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReductionRing Conservation Theorem"},
 CellID->459032599],
Cell[3038874, 84692, 65, 1, 70, "EndEnvironmentMarker",
 CellID->590387924]
}, Open  ]],
Cell[3038954, 84696, 48, 1, 70, "CloseEnvironment",
 CellID->679962283]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3039051, 84703, 355, 7, 43, "Section",
 CellID->322583785],
Cell[3039409, 84712, 288, 6, 24, "Text",
 CellID->1814342886],
Cell[3039700, 84720, 47, 1, 12, "OpenEnvironment",
 CellID->844754185],
Cell[CellGroupData[{
Cell[3039772, 84725, 525, 13, 43, "EnvironmentHeader",
 CellID->243471114],
Cell[3040300, 84740, 864, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:719489133", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "algorithmic reduction ring \[ScriptCapitalR]"},
 CellID->719489133],
Cell[3041167, 84764, 919, 24, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1159608082", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", 
   "algorithmic reduction-polynomial domain \[ScriptCapitalP]"},
 CellID->1159608082],
Cell[3042089, 84790, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1327883073]
}, Open  ]],
Cell[3042170, 84794, 48, 1, 16, "CloseEnvironment",
 CellID->230307546],
Cell[CellGroupData[{
Cell[3042243, 84799, 200, 3, 38, "Subsection",
 CellID->1690991356],
Cell[3042446, 84804, 47, 1, 70, "OpenEnvironment",
 CellID->558613337],
Cell[CellGroupData[{
Cell[3042518, 84809, 445, 10, 70, "EnvironmentHeader",
 CellID->1738024641],
Cell[3042966, 84821, 206, 3, 70, "EnvironmentText",
 CellID->1533594446],
Cell[3043175, 84826, 3075, 91, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:847570831", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[CenterDot] \[Succeeds]"},
 CellID->847570831],
Cell[3046253, 84919, 65, 1, 70, "EndEnvironmentMarker",
 CellID->593969255]
}, Open  ]],
Cell[3046333, 84923, 48, 1, 70, "CloseEnvironment",
 CellID->408203277],
Cell[CellGroupData[{
Cell[3046406, 84928, 6346, 148, 70, "ProofInfo",
 CellTags->{"Proof|ID:847570831", "Proof|ID:847570831-1"},
 CellID->1483028309],
Cell[3052755, 85078, 481, 12, 70, "ProofInfoBody",
 CellID->1094495224]
}, Closed]],
Cell[3053251, 85093, 48, 1, 70, "OpenEnvironment",
 CellID->1742002312],
Cell[CellGroupData[{
Cell[3053324, 85098, 434, 11, 70, "EnvironmentHeader",
 CellID->1137163506],
Cell[3053761, 85111, 1013, 29, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2071048931", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->2071048931],
Cell[3054777, 85142, 735, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:355437230", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->355437230],
Cell[3055515, 85163, 1077, 28, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1401739284", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1401739284],
Cell[3056595, 85193, 904, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:29665786", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 1"},
 CellID->29665786],
Cell[3057502, 85219, 1549, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:113282996", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 2"},
 CellID->113282996],
Cell[3059054, 85264, 1683, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1675445982", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 3"},
 CellID->1675445982],
Cell[3060740, 85315, 1557, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1968653014", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 4"},
 CellID->1968653014],
Cell[3062300, 85361, 1241, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:860310855", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP] 5"},
 CellID->860310855],
Cell[3063544, 85399, 1403, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2026748686", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] multiplier \[ScriptCapitalP]"},
 CellID->2026748686],
Cell[3064950, 85436, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1941983415]
}, Open  ]],
Cell[3065031, 85440, 49, 1, 70, "CloseEnvironment",
 CellID->1033726545],
Cell[CellGroupData[{
Cell[3065105, 85445, 5544, 137, 70, "ProofInfo",
 CellTags->{"Proof|ID:29665786", "Proof|ID:29665786-1"},
 CellID->343762797],
Cell[3070652, 85584, 480, 12, 70, "ProofInfoBody",
 CellID->1884251853]
}, Closed]],
Cell[CellGroupData[{
Cell[3071169, 85601, 6375, 153, 70, "ProofInfo",
 CellTags->{"Proof|ID:113282996", "Proof|ID:113282996-1"},
 CellID->1216564136],
Cell[3077547, 85756, 480, 12, 70, "ProofInfoBody",
 CellID->321450032]
}, Closed]],
Cell[CellGroupData[{
Cell[3078064, 85773, 7133, 168, 70, "ProofInfo",
 CellTags->{"Proof|ID:1675445982", "Proof|ID:1675445982-1"},
 CellID->1330746921],
Cell[3085200, 85943, 482, 12, 70, "ProofInfoBody",
 CellID->1849793449]
}, Closed]],
Cell[CellGroupData[{
Cell[3085719, 85960, 8247, 190, 70, "ProofInfo",
 CellTags->{"Proof|ID:1968653014", "Proof|ID:1968653014-1"},
 CellID->1366469966],
Cell[3093969, 86152, 481, 12, 70, "ProofInfoBody",
 CellID->461522218]
}, Closed]],
Cell[CellGroupData[{
Cell[3094487, 86169, 5990, 145, 70, "ProofInfo",
 CellTags->{"Proof|ID:860310855", "Proof|ID:860310855-1"},
 CellID->1282192616],
Cell[3100480, 86316, 480, 12, 70, "ProofInfoBody",
 CellID->306440183]
}, Closed]],
Cell[CellGroupData[{
Cell[3100997, 86333, 5596, 141, 70, "ProofInfo",
 CellTags->{"Proof|ID:2026748686", "Proof|ID:2026748686-1"},
 CellID->1041146073],
Cell[3106596, 86476, 482, 12, 70, "ProofInfoBody",
 CellID->1336512602]
}, Closed]],
Cell[3107093, 86491, 48, 1, 70, "OpenEnvironment",
 CellID->1738547973],
Cell[CellGroupData[{
Cell[3107166, 86496, 386, 9, 70, "EnvironmentHeader",
 CellID->1666042888],
Cell[3107555, 86507, 1857, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1529886137", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "\[Tilde] \[ScriptCapitalP]"},
 CellID->1529886137],
Cell[3109415, 86565, 64, 1, 70, "EndEnvironmentMarker",
 CellID->60767287]
}, Open  ]],
Cell[3109494, 86569, 49, 1, 70, "CloseEnvironment",
 CellID->2095809430],
Cell[CellGroupData[{
Cell[3109568, 86574, 4182, 109, 70, "ProofInfo",
 CellTags->{"Proof|ID:1529886137", "Proof|ID:1529886137-1"},
 CellID->1495738535],
Cell[3113753, 86685, 482, 12, 70, "ProofInfoBody",
 CellID->1783742101]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3114284, 86703, 274, 4, 30, "Subsection",
 CellID->1708894494],
Cell[3114561, 86709, 48, 1, 12, "OpenEnvironment",
 CellID->2072032119],
Cell[CellGroupData[{
Cell[3114634, 86714, 393, 9, 43, "EnvironmentHeader",
 CellID->1748702921],
Cell[3115030, 86725, 781, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1680577066", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "372"},
 CellID->1680577066],
Cell[3115814, 86748, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1237743236]
}, Open  ]],
Cell[3115895, 86752, 49, 1, 16, "CloseEnvironment",
 CellID->1965785607],
Cell[CellGroupData[{
Cell[3115969, 86757, 1731, 57, 28, "ProofInfo",
 CellTags->{"Proof|ID:1680577066", "Proof|ID:1680577066-1"},
 CellID->5711938],
Cell[3117703, 86816, 481, 12, 70, "ProofInfoBody",
 CellID->302678198]
}, Closed]],
Cell[CellGroupData[{
Cell[3118221, 86833, 357, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1663153286],
Cell[3118581, 86844, 309, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1756388746],
Cell[3118893, 86851, 1273, 34, 60, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->973735291],
Cell[3120169, 86887, 1136, 30, 60, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->807494688],
Cell[3121308, 86919, 12474, 381, 1156, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->347176547],
Cell[3133785, 87302, 315, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1358676932]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3134149, 87314, 179, 3, 30, "Subsection",
 CellID->268409188],
Cell[3134331, 87319, 46, 1, 70, "OpenEnvironment",
 CellID->24273816],
Cell[CellGroupData[{
Cell[3134402, 87324, 392, 9, 70, "EnvironmentHeader",
 CellID->975532904],
Cell[3134797, 87335, 783, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1866021817", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "373"},
 CellID->1866021817],
Cell[3135583, 87358, 65, 1, 70, "EndEnvironmentMarker",
 CellID->795876134]
}, Open  ]],
Cell[3135663, 87362, 49, 1, 70, "CloseEnvironment",
 CellID->1616684133],
Cell[CellGroupData[{
Cell[3135737, 87367, 1733, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1866021817", "Proof|ID:1866021817-1"},
 CellID->1214587],
Cell[3137473, 87426, 481, 12, 70, "ProofInfoBody",
 CellID->169993026]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3138003, 87444, 178, 3, 30, "Subsection",
 CellID->1084053984],
Cell[3138184, 87449, 47, 1, 70, "OpenEnvironment",
 CellID->339961261],
Cell[CellGroupData[{
Cell[3138256, 87454, 394, 9, 70, "EnvironmentHeader",
 CellID->2108760469],
Cell[3138653, 87465, 783, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1272742036", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "374"},
 CellID->1272742036],
Cell[3139439, 87488, 65, 1, 70, "EndEnvironmentMarker",
 CellID->263791477]
}, Open  ]],
Cell[3139519, 87492, 49, 1, 70, "CloseEnvironment",
 CellID->2106755448],
Cell[CellGroupData[{
Cell[3139593, 87497, 1735, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1272742036", "Proof|ID:1272742036-1"},
 CellID->634133424],
Cell[3141331, 87556, 481, 12, 70, "ProofInfoBody",
 CellID->107031843]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3141861, 87574, 175, 3, 30, "Subsection",
 CellID->1481007890],
Cell[3142039, 87579, 48, 1, 12, "OpenEnvironment",
 CellID->1777956310],
Cell[CellGroupData[{
Cell[3142112, 87584, 334, 9, 43, "EnvironmentHeader",
 CellID->139848971],
Cell[3142449, 87595, 1186, 29, 71, "GlobalDeclaration",
 CellTags->{
  "ID:181479986", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->181479986],
Cell[3143638, 87626, 739, 19, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1276235216", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1276235216],
Cell[3144380, 87647, 896, 23, 58, "GlobalDeclaration",
 CellTags->{
  "ID:376128819", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->376128819],
Cell[3145279, 87672, 1014, 27, 49, "GlobalDeclaration",
 CellTags->{
  "ID:973122334", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->973122334],
Cell[3146296, 87701, 458, 15, 25, "EnvironmentText",
 CellID->663596886],
Cell[3146757, 87718, 2509, 75, 93, "FormalTextInputFormula",
 CellTags->{
  "ID:1445738544", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.1"},
 CellID->1445738544],
Cell[3149269, 87795, 2968, 80, 81, "FormalTextInputFormula",
 CellTags->{
  "ID:874376987", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.2"},
 CellID->874376987],
Cell[3152240, 87877, 1880, 54, 75, "FormalTextInputFormula",
 CellTags->{
  "ID:139929842", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375.3"},
 CellID->139929842],
Cell[3154123, 87933, 65, 1, 26, "EndEnvironmentMarker",
 CellID->505373945]
}, Open  ]],
Cell[3154203, 87937, 49, 1, 16, "CloseEnvironment",
 CellID->1110225807],
Cell[CellGroupData[{
Cell[3154277, 87942, 10777, 241, 28, "ProofInfo",
 CellTags->{"Proof|ID:1445738544", "Proof|ID:1445738544-2"},
 CellID->1345220325],
Cell[3165057, 88185, 482, 12, 70, "ProofInfoBody",
 CellID->1091818374]
}, Closed]],
Cell[CellGroupData[{
Cell[3165576, 88202, 13391, 292, 26, "ProofInfo",
 CellTags->{"Proof|ID:874376987", "Proof|ID:874376987-1"},
 CellID->7612480],
Cell[3178970, 88496, 477, 12, 70, "ProofInfoBody",
 CellID->277234]
}, Closed]],
Cell[CellGroupData[{
Cell[3179484, 88513, 10076, 230, 26, "ProofInfo",
 CellTags->{"Proof|ID:139929842", "Proof|ID:139929842-2"},
 CellID->1858784208],
Cell[3189563, 88745, 481, 12, 70, "ProofInfoBody",
 CellID->1228384409]
}, Closed]],
Cell[3190059, 88760, 48, 1, 12, "OpenEnvironment",
 CellID->1578377743],
Cell[CellGroupData[{
Cell[3190132, 88765, 420, 10, 43, "EnvironmentHeader",
 CellID->272592021],
Cell[3190555, 88777, 780, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:462890005", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "375"},
 CellID->462890005],
Cell[3191338, 88800, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1497256452]
}, Open  ]],
Cell[3191419, 88804, 49, 1, 16, "CloseEnvironment",
 CellID->1960788671],
Cell[CellGroupData[{
Cell[3191493, 88809, 1725, 56, 28, "ProofInfo",
 CellTags->{"Proof|ID:462890005", "Proof|ID:462890005-1"},
 CellID->51997037],
Cell[3193221, 88867, 480, 12, 70, "ProofInfoBody",
 CellID->468582439]
}, Closed]],
Cell[CellGroupData[{
Cell[3193738, 88884, 364, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->556816827],
Cell[3194105, 88895, 20324, 585, 1371, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->805176799]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3214478, 89486, 179, 3, 30, "Subsection",
 CellID->286693183],
Cell[3214660, 89491, 47, 1, 70, "OpenEnvironment",
 CellID->827957578],
Cell[CellGroupData[{
Cell[3214732, 89496, 398, 9, 70, "EnvironmentHeader",
 CellID->865456817],
Cell[3215133, 89507, 1623, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:38279669", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducible lm"},
 CellID->38279669],
Cell[3216759, 89558, 1475, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1376605467", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "red lm set"},
 CellID->1376605467],
Cell[3218237, 89604, 3120, 87, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:673925715", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isReducible monomial"},
 CellID->673925715],
Cell[3221360, 89693, 65, 1, 70, "EndEnvironmentMarker",
 CellID->111884135]
}, Open  ]],
Cell[3221440, 89697, 48, 1, 70, "CloseEnvironment",
 CellID->860415994],
Cell[CellGroupData[{
Cell[3221513, 89702, 4153, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:38279669", "Proof|ID:38279669-1"},
 CellID->51695314],
Cell[3225669, 89812, 479, 12, 70, "ProofInfoBody",
 CellID->582723933]
}, Closed]],
Cell[CellGroupData[{
Cell[3226185, 89829, 3980, 107, 70, "ProofInfo",
 CellTags->{"Proof|ID:1376605467", "Proof|ID:1376605467-1"},
 CellID->580557624],
Cell[3230168, 89938, 480, 12, 70, "ProofInfoBody",
 CellID->66183403]
}, Closed]],
Cell[CellGroupData[{
Cell[3230685, 89955, 6350, 152, 70, "ProofInfo",
 CellTags->{"Proof|ID:673925715", "Proof|ID:673925715-1"},
 CellID->54334308],
Cell[3237038, 90109, 480, 12, 70, "ProofInfoBody",
 CellID->694257822]
}, Closed]],
Cell[3237533, 90124, 48, 1, 70, "OpenEnvironment",
 CellID->1784825111],
Cell[CellGroupData[{
Cell[3237606, 90129, 410, 10, 70, "EnvironmentHeader",
 CellID->1154045365],
Cell[3238019, 90141, 1876, 58, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:47923252", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence"},
 CellID->47923252],
Cell[3239898, 90201, 65, 1, 70, "EndEnvironmentMarker",
 CellID->316466243]
}, Open  ]],
Cell[3239978, 90205, 48, 1, 70, "CloseEnvironment",
 CellID->923509955],
Cell[3240029, 90208, 48, 1, 70, "OpenEnvironment",
 CellID->1915683957],
Cell[CellGroupData[{
Cell[3240102, 90213, 394, 9, 70, "EnvironmentHeader",
 CellID->474285465],
Cell[3240499, 90224, 739, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1632720588", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->1632720588],
Cell[3241241, 90245, 1075, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1771873303", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence \[NotEqual] 0"},
 CellID->1771873303],
Cell[3242319, 90277, 1052, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:893563053", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lc 1"},
 CellID->893563053],
Cell[3243374, 90308, 1054, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:587765202", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lp 1"},
 CellID->587765202],
Cell[3244431, 90339, 991, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1254733295", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lc 2"},
 CellID->1254733295],
Cell[3245425, 90368, 988, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:280696418", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence lp 2"},
 CellID->280696418],
Cell[3246416, 90397, 989, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1081126646", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence subSeq"},
 CellID->1081126646],
Cell[3247408, 90425, 973, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1101697023", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isMonomialSequence subsequence"},
 CellID->1101697023],
Cell[3248384, 90453, 64, 1, 70, "EndEnvironmentMarker",
 CellID->90094680]
}, Open  ]],
Cell[3248463, 90457, 49, 1, 70, "CloseEnvironment",
 CellID->1586112789],
Cell[CellGroupData[{
Cell[3248537, 90462, 3303, 94, 70, "ProofInfo",
 CellTags->{"Proof|ID:1771873303", "Proof|ID:1771873303-1"},
 CellID->103075955],
Cell[3251843, 90558, 482, 12, 70, "ProofInfoBody",
 CellID->1903828265]
}, Closed]],
Cell[CellGroupData[{
Cell[3252362, 90575, 3468, 94, 70, "ProofInfo",
 CellTags->{"Proof|ID:893563053", "Proof|ID:893563053-1"},
 CellID->88855734],
Cell[3255833, 90671, 481, 12, 70, "ProofInfoBody",
 CellID->1505089946]
}, Closed]],
Cell[CellGroupData[{
Cell[3256351, 90688, 3470, 94, 70, "ProofInfo",
 CellTags->{"Proof|ID:587765202", "Proof|ID:587765202-1"},
 CellID->1976154064],
Cell[3259824, 90784, 481, 12, 70, "ProofInfoBody",
 CellID->1938386534]
}, Closed]],
Cell[CellGroupData[{
Cell[3260342, 90801, 3009, 85, 70, "ProofInfo",
 CellTags->{"Proof|ID:1254733295", "Proof|ID:1254733295-1"},
 CellID->1511544165],
Cell[3263354, 90888, 481, 12, 70, "ProofInfoBody",
 CellID->738405199]
}, Closed]],
Cell[CellGroupData[{
Cell[3263872, 90905, 2999, 84, 70, "ProofInfo",
 CellTags->{"Proof|ID:280696418", "Proof|ID:280696418-1"},
 CellID->276631526],
Cell[3266874, 90991, 480, 12, 70, "ProofInfoBody",
 CellID->273841733]
}, Closed]],
Cell[CellGroupData[{
Cell[3267391, 91008, 2912, 81, 70, "ProofInfo",
 CellTags->{"Proof|ID:1081126646", "Proof|ID:1081126646-1"},
 CellID->227638759],
Cell[3270306, 91091, 482, 12, 70, "ProofInfoBody",
 CellID->2067143528]
}, Closed]],
Cell[CellGroupData[{
Cell[3270825, 91108, 2946, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:1101697023", "Proof|ID:1101697023-1"},
 CellID->881838494],
Cell[3273774, 91192, 482, 12, 70, "ProofInfoBody",
 CellID->1551672538]
}, Closed]],
Cell[3274271, 91207, 1015, 31, 70, "Text",
 CellID->989285657],
Cell[3275289, 91240, 48, 1, 70, "OpenEnvironment",
 CellID->1565804923],
Cell[CellGroupData[{
Cell[3275362, 91245, 393, 9, 70, "EnvironmentHeader",
 CellID->1515610326],
Cell[3275758, 91256, 3026, 89, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:182970451", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "110"},
 CellID->182970451],
Cell[3278787, 91347, 3164, 92, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:845195690", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "111"},
 CellID->845195690],
Cell[3281954, 91441, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1298865959]
}, Open  ]],
Cell[3282035, 91445, 49, 1, 70, "CloseEnvironment",
 CellID->1599916594],
Cell[CellGroupData[{
Cell[3282109, 91450, 8308, 207, 70, "ProofInfo",
 CellTags->{"Proof|ID:182970451", "Proof|ID:182970451-1"},
 CellID->568281335],
Cell[3290420, 91659, 480, 12, 70, "ProofInfoBody",
 CellID->758050075]
}, Closed]],
Cell[CellGroupData[{
Cell[3290937, 91676, 8270, 206, 70, "ProofInfo",
 CellTags->{"Proof|ID:845195690", "Proof|ID:845195690-1"},
 CellID->70458024],
Cell[3299210, 91884, 479, 12, 70, "ProofInfoBody",
 CellID->31491083]
}, Closed]],
Cell[3299704, 91899, 48, 1, 70, "OpenEnvironment",
 CellID->1592718460],
Cell[CellGroupData[{
Cell[3299777, 91904, 508, 13, 70, "EnvironmentHeader",
 CellID->1777300951],
Cell[3300288, 91919, 739, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2081362811", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb"},
 CellID->2081362811],
Cell[3301030, 91940, 4239, 120, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1173235103", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "112"},
 CellID->1173235103],
Cell[3305272, 92062, 2318, 70, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1805755328", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "113"},
 CellID->1805755328],
Cell[3307593, 92134, 2490, 67, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:596327945", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "114"},
 CellID->596327945],
Cell[3310086, 92203, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1479320113]
}, Open  ]],
Cell[3310167, 92207, 46, 1, 70, "CloseEnvironment",
 CellID->3101935],
Cell[CellGroupData[{
Cell[3310238, 92212, 11823, 279, 70, "ProofInfo",
 CellTags->{"Proof|ID:1173235103", "Proof|ID:1173235103-1"},
 CellID->874379054],
Cell[3322064, 92493, 481, 12, 70, "ProofInfoBody",
 CellID->423369729]
}, Closed]],
Cell[CellGroupData[{
Cell[3322582, 92510, 6387, 159, 70, "ProofInfo",
 CellTags->{"Proof|ID:1805755328", "Proof|ID:1805755328-1"},
 CellID->1937389710],
Cell[3328972, 92671, 524, 13, 70, "ProofInfoBody",
 CellID->1056270318]
}, Closed]],
Cell[CellGroupData[{
Cell[3329533, 92689, 5826, 147, 70, "ProofInfo",
 CellTags->{"Proof|ID:596327945", "Proof|ID:596327945-1"},
 CellID->387192426],
Cell[3335362, 92838, 478, 12, 70, "ProofInfoBody",
 CellID->8942193]
}, Closed]],
Cell[3335855, 92853, 47, 1, 70, "OpenEnvironment",
 CellID->585699768],
Cell[CellGroupData[{
Cell[3335927, 92858, 581, 14, 70, "EnvironmentHeader",
 CellID->1237178328],
Cell[3336511, 92874, 2262, 70, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1402642391", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "115"},
 CellID->1402642391],
Cell[3338776, 92946, 66, 1, 70, "EndEnvironmentMarker",
 CellID->2031723415]
}, Open  ]],
Cell[3338857, 92950, 48, 1, 70, "CloseEnvironment",
 CellID->362476080],
Cell[CellGroupData[{
Cell[3338930, 92955, 7083, 181, 70, "ProofInfo",
 CellTags->{"Proof|ID:1402642391", "Proof|ID:1402642391-1"},
 CellID->1177596546],
Cell[3346016, 93138, 482, 12, 70, "ProofInfoBody",
 CellID->1549025416]
}, Closed]],
Cell[3346513, 93153, 48, 1, 70, "OpenEnvironment",
 CellID->1409591193],
Cell[CellGroupData[{
Cell[3346586, 93158, 408, 11, 70, "EnvironmentHeader",
 CellID->49556011],
Cell[3346997, 93171, 1081, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:980978647", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "Dickson-domain \[ScriptCapitalT]"},
 CellID->980978647],
Cell[3348081, 93203, 65, 1, 70, "EndEnvironmentMarker",
 CellID->148163230]
}, Open  ]],
Cell[3348161, 93207, 49, 1, 70, "CloseEnvironment",
 CellID->1739788415],
Cell[3348213, 93210, 47, 1, 70, "OpenEnvironment",
 CellID->506738286],
Cell[CellGroupData[{
Cell[3348285, 93215, 447, 10, 70, "EnvironmentHeader",
 CellID->1566663652],
Cell[3348735, 93227, 756, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1408052308", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "116"},
 CellID->1408052308],
Cell[3349494, 93249, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1643742643]
}, Open  ]],
Cell[3349575, 93253, 49, 1, 70, "CloseEnvironment",
 CellID->1724439218],
Cell[CellGroupData[{
Cell[3349649, 93258, 1735, 57, 70, "ProofInfo",
 CellTags->{"Proof|ID:1408052308", "Proof|ID:1408052308-1"},
 CellID->245205539],
Cell[3351387, 93317, 481, 12, 70, "ProofInfoBody",
 CellID->127378030]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3351917, 93335, 188, 3, 30, "Subsection",
 CellID->2038082117],
Cell[3352108, 93340, 47, 1, 70, "OpenEnvironment",
 CellID->238202638],
Cell[CellGroupData[{
Cell[3352180, 93345, 481, 12, 70, "EnvironmentHeader",
 CellID->496041222],
Cell[3352664, 93359, 857, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1183157947", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", "isAlgoReductionRing \[ScriptCapitalP]"},
 CellID->1183157947],
Cell[3353524, 93383, 65, 1, 70, "EndEnvironmentMarker",
 CellID->364104560]
}, Open  ]],
Cell[3353604, 93387, 49, 1, 70, "CloseEnvironment",
 CellID->1493486891],
Cell[CellGroupData[{
Cell[3353678, 93392, 1859, 59, 70, "ProofInfo",
 CellTags->{"Proof|ID:1183157947", "Proof|ID:1183157947-1"},
 CellID->1299983476],
Cell[3355540, 93453, 482, 12, 70, "ProofInfoBody",
 CellID->1799732883]
}, Closed]],
Cell[3356037, 93468, 1130, 19, 70, "Text",
 CellID->72919207],
Cell[3357170, 93489, 2152, 50, 70, "Input",
 CellID->191565054],
Cell[3359325, 93541, 342, 8, 70, "Text",
 CellID->149433800],
Cell[3359670, 93551, 48, 1, 70, "OpenEnvironment",
 CellID->1569900360],
Cell[CellGroupData[{
Cell[3359743, 93556, 337, 9, 70, "EnvironmentHeader",
 CellID->1957927293],
Cell[3360083, 93567, 2107, 61, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1794188423", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials.nb", 
   "isAlgoReductionRing Conservation Theorem Polynomials"},
 CellID->1794188423],
Cell[3362193, 93630, 65, 1, 70, "EndEnvironmentMarker",
 CellID->331601271]
}, Open  ]],
Cell[3362273, 93634, 48, 1, 70, "CloseEnvironment",
 CellID->820459176]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
