(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[   3858419,     104694]
NotebookOptionsPosition[   3164764,      85932]
NotebookOutlinePosition[   3637953,      98190]
CellTagsIndexPosition[   3588013,      96918]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Reduction Rings", "Title",
 CellChangeTimes->{{3.590837152246185*^9, 3.590837155585586*^9}, 
   3.607771716512438*^9, {3.635757332014748*^9, 3.635757337118741*^9}, 
   3.643445850784401*^9},
 TextAlignment->Center,
 CellID->1451885783],

Cell[CellGroupData[{

Cell["Preliminaries", "Section",
 CellChangeTimes->{{3.64066853326334*^9, 3.64066853528587*^9}, {
  3.640680419424919*^9, 3.6406804222204437`*^9}},
 CellID->1274729640],

Cell[CellGroupData[{

Cell["Description", "Subsection",
 CellChangeTimes->{{3.660735829576782*^9, 3.6607358316789026`*^9}},
 CellID->172857195],

Cell[TextData[{
 "This notebook introduces the concepts of ",
 StyleBox["reduction rings",
  FontSlant->"Italic"],
 " and ",
 StyleBox["algorithmic reduction rings",
  FontSlant->"Italic"],
 " as certain classes of domains having some algebraic and algorithmic \
properties. In short, a reduction ring \[ScriptCapitalR] is a commutative \
ring with unit, additionally endowed with a aprtial Noetherian ordering and \
sets of ",
 StyleBox["multipliers",
  FontSlant->"Italic"],
 " for each ring element, such that ",
 StyleBox["Gr\[ODoubleDot]bner bases",
  FontSlant->"Italic"],
 " for finitely generated ideals in \[ScriptCapitalR] exist and can be \
computed by a critical-pair/completion algorithm \[AAcute] la Buchberger. Gr\
\[ODoubleDot]bner bases, defined in section \[OpenCurlyDoubleQuote]Main \
Theorem of Reduction Ring Theory\[CloseCurlyDoubleQuote], are sets (or \
tuples, for that matter) such that the reduction relation they induce is \
confluent. This property allows one to solve many interesting ideal-theoretic \
problems, such as deciding ideal membership or ideal equality, because in \
many cases these problems simply reduce to finding Gr\[ODoubleDot]bner bases \
for the ideals in question."
}], "Text",
 CellChangeTimes->{{3.6607358371152134`*^9, 3.660736030700286*^9}, {
  3.6607360889806194`*^9, 3.660736231546774*^9}},
 CellID->653813119],

Cell[TextData[{
 "Sections \[OpenCurlyDoubleQuote]Auxiliary Notions\[CloseCurlyDoubleQuote] \
and \[OpenCurlyDoubleQuote]Reduction Rings\[CloseCurlyDoubleQuote] introduce, \
as their titles suggest, a bunch of auxiliary notions related to the theory \
of reduction rings (most importantly the notion of ",
 StyleBox["reduction",
  FontSlant->"Italic"],
 "), as well as the classes of (algorithmic) reduction rings. Moreover, some \
more or less immediate consequences of the various definitions are stated and \
proved."
}], "Text",
 CellChangeTimes->{{3.6607362481347227`*^9, 3.6607363860636115`*^9}},
 CellID->579693263],

Cell["\<\
Section \[OpenCurlyDoubleQuote]First Lemmas and Theorems\
\[CloseCurlyDoubleQuote] proves some more advanced and less obvious facts, \
mostly about the reduction relation, its symmetric-reflexive-transtive \
closure, and the \[OpenCurlyDoubleQuote]connectibility below\
\[CloseCurlyDoubleQuote] relation.\
\>", "Text",
 CellChangeTimes->{{3.6607364062187643`*^9, 3.6607364979010086`*^9}},
 CellID->438335057],

Cell[TextData[{
 "Section \[OpenCurlyDoubleQuote]Main Theorem of Reduction Ring Theory\
\[CloseCurlyDoubleQuote] is the core of this notebook: After some further \
preparatory lemmas and definitions, the Main Theorem of the theory of \
reduction rings is stated, which gives a finite criterion to decide whether a \
given (finite) set of elements of a reduction ring is a Gr\[ODoubleDot]bner \
basis or not. Furthermore, the theorem can easily be adapted as to yield an \
algorithm for ",
 StyleBox["completing",
  FontSlant->"Italic"],
 " a given finite set that is no Gr\[ODoubleDot]bner basis in order to become \
a Gr\[ODoubleDot]bner basis (\[Rule] ",
 StyleBox["Buchberger\[CloseCurlyQuote]s algorithm",
  FontSlant->"Italic"],
 "). This algorithm is implemented and proved correct in \
\[OpenCurlyDoubleQuote]GroebnerRings.nb\[CloseCurlyDoubleQuote]."
}], "Text",
 CellChangeTimes->{{3.660736520236286*^9, 3.6607366139816475`*^9}, {
  3.660736692604145*^9, 3.6607367943579645`*^9}},
 CellID->195467169],

Cell[TextData[{
 "Finally, Section \[OpenCurlyDoubleQuote]Ideal Congruence and Ideals\
\[CloseCurlyDoubleQuote] defines the concepts of ideal congruence and ideals \
in commutative rings and proves some elementary properties thereof, together \
with the all-important fact that congruence modulo the ideal generated by a \
set ",
 Cell[BoxData[
  FormBox["B", TraditionalForm]]],
 " in a reduction ring \[ScriptCapitalR] coincides with the \
symmetric-reflexive-transitive closure of the reduction relation in \
\[ScriptCapitalR]."
}], "Text",
 CellChangeTimes->{{3.6607368035204887`*^9, 3.660736939340257*^9}, {
  3.6607377121524596`*^9, 3.6607377171157436`*^9}},
 CellID->397022844],

Cell["\<\
Please note that this notebook is mainly concerned with reduction rings in \
contrast to algorithmic reduction rings, since only abstract, non-algrithmic \
properties of rings are needed for establishing the Main Theorem. Algorithmic \
reduction rings, which are subclasses of reduction rings where the abstractly \
stated properties of reduction rings have algorithmic counterparts (e.g. \
solutions to existential statements can be solved algorithmically), are of \
importance as soon as algorithms come into play, and this is the case in \
\[OpenCurlyDoubleQuote]GroebnerRings.nb\[CloseCurlyDoubleQuote].\
\>", "Text",
 CellChangeTimes->{{3.6607369897601414`*^9, 3.6607372503900485`*^9}},
 CellID->571648042],

Cell[CellGroupData[{

Cell["References", "Subsubsection",
 CellChangeTimes->{{3.660736684316671*^9, 3.6607366863627877`*^9}},
 CellID->399485901],

Cell["\<\
Bruno Buchberger. A Critical-Pair/Completion Algorithm in Reduction Rings. \
Technical Report 83-21, Research Institute for Symbolic Computation, \
University of Linz, 1983.\
\>", "Text",
 CellChangeTimes->{{3.660736656450077*^9, 3.660736659805269*^9}, {
  3.6607373325847497`*^9, 3.660737400199617*^9}},
 CellID->328963968],

Cell["\<\
Alexander Maletzky. Exploring Reduction Ring Theory in Theorema. Technical \
Report 15-11, Research Institute for Symbolic Computation, Johannes Kepler \
University Linz, 2015.\
\>", "Text",
 CellChangeTimes->{{3.6607375200724735`*^9, 3.6607375236626787`*^9}, {
  3.6607375741425657`*^9, 3.6607376352980638`*^9}},
 CellID->229090570],

Cell[TextData[{
 "Sabine Stifter. ",
 StyleBox["A Generalization of Reduction Rings",
  FontSlant->"Italic"],
 ". Journal of Symbolic Computation ",
 StyleBox["4",
  FontWeight->"Bold"],
 "(3): 351\[Dash]364, 1988."
}], "Text",
 CellChangeTimes->{{3.6607374019937196`*^9, 3.660737468874545*^9}},
 CellID->272564439]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Packages", "Subsection",
 CellChangeTimes->{{3.655466156246707*^9, 3.6554661588363123`*^9}},
 CellID->721522687],

Cell[BoxData[
 RowBox[{"$PackageDirectory", "=", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{
    "$RootDirectory", ",", "\"\<scratch\>\"", ",", "\"\<amaletzk\>\"", ",", 
     "\"\<Theorema Environment Data\>\""}], "}"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.666438783233775*^9, 3.666438821083324*^9}},
 CellID->498486718],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<LogicSets.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554661631575193`*^9, 3.6554661689451294`*^9}, {
   3.656321143861744*^9, 3.656321157808168*^9}, {3.6596890752359505`*^9, 
   3.6596890814603615`*^9}, {3.6620925389697533`*^9, 3.6620925408261566`*^9}, 
   3.666440876201075*^9},
 CellID->192289652],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<AlgebraicStructures.m\>\""}], 
    "}"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6554661631575193`*^9, 3.6554661689451294`*^9}, {
   3.656321143861744*^9, 3.656321157808168*^9}, {3.6596890752359505`*^9, 
   3.6596890814603615`*^9}, 3.666440877267188*^9},
 CellID->329397223],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<NatInt.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554661631575193`*^9, 3.6554661689451294`*^9}, {
   3.656321161583375*^9, 3.656321167495785*^9}, 3.666440878204303*^9},
 CellID->15589227],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<NatIntExtended.m\>\""}], "}"}], 
   "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6554661631575193`*^9, 3.6554661689451294`*^9}, {
   3.656321161583375*^9, 3.656321167495785*^9}, {3.6621876193334656`*^9, 
   3.662187620519068*^9}, 3.6664408790922127`*^9},
 CellID->76117945],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Tuples.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554661631575193`*^9, 3.6554661689451294`*^9}, {
   3.656321143861744*^9, 3.656321157808168*^9}, 3.666440880108182*^9},
 CellID->248729940],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Sequences.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554661631575193`*^9, 3.6554661689451294`*^9}, {
   3.656321143861744*^9, 3.656321157808168*^9}, {3.6617512990697355`*^9, 
   3.6617513007087383`*^9}, 3.66644088104428*^9},
 CellID->357734827],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6554661631575193`*^9, 3.6554661689451294`*^9}},
 CellID->594767243],

Cell[BoxData[
 RowBox[{"MakeTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6554661731883373`*^9, 3.655466178055546*^9}},
 CellID->617813923],

Cell[BoxData[
 RowBox[{"UpdateTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6582235718477755`*^9, 3.65822357467138*^9}},
 CellID->155920184],

Cell[BoxData[
 RowBox[{"CompressProofData", "[", "]"}]], "Input",
 CellChangeTimes->{{3.662354442367301*^9, 3.662354446906909*^9}},
 CellID->21169241]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Analysis", "Subsection",
 CellChangeTimes->{{3.640668830437483*^9, 3.640668834318727*^9}},
 CellID->1386333768],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"First", "[", 
  RowBox[{"Timing", "[", 
   RowBox[{"AnalyzeProofs", "[", "\"\<Theory\>\"", "]"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6536470529977045`*^9, 3.6536470602361174`*^9}, {
  3.6618404152543445`*^9, 3.6618404219779563`*^9}},
 CellID->740290274],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\AlgebraicStructures\\\" (\"\>", 
   "\[InvisibleSpace]", "1", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures\" (", 1, " of ", "17", 
   "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404443015957`*^9},
 CellID->65794957],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"51\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "51"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404443015957`*^9},
 CellID->316361875],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"51\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "51"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840444691596*^9},
 CellID->548775684],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\ConsistencyChecker\\\" (\"\>", 
   "\[InvisibleSpace]", "2", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ConsistencyChecker\" (", 2, " of ", "17", 
   "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840444707196*^9},
 CellID->26524607],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Fields\\\" (\"\>", "\[InvisibleSpace]", 
   "3", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Fields\" (", 3, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404447227964`*^9},
 CellID->489535381],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"36\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "36"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404447227964`*^9},
 CellID->193939678],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Functors\\\" (\"\>", 
   "\[InvisibleSpace]", "4", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Functors\" (", 4, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840445721198*^9},
 CellID->259626599],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"36\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "36"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840445721198*^9},
 CellID->190398616],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\GroebnerRings\\\" (\"\>", 
   "\[InvisibleSpace]", "5", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\GroebnerRings\" (", 5, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404465479994`*^9},
 CellID->494145549],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404465479994`*^9},
 CellID->21379931],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840447905202*^9},
 CellID->249495970],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840449246804*^9},
 CellID->241306951],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404505260067`*^9},
 CellID->620414427],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\IntegerQuotientRings\\\" (\"\>", 
   "\[InvisibleSpace]", "6", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings\" (", 6, " of ", "17", 
   "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404510876074`*^9},
 CellID->407148683],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"78\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "78"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404510876074`*^9},
 CellID->551639436],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"78\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "78"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404515712085`*^9},
 CellID->777840439],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Integers\\\" (\"\>", 
   "\[InvisibleSpace]", "7", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Integers\" (", 7, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840451852009*^9},
 CellID->145768808],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"85\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "85"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840451852009*^9},
 CellID->89605923],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"85\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "85"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404522732096`*^9},
 CellID->92121938],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\LogicSets\\\" (\"\>", 
   "\[InvisibleSpace]", "8", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets\" (", 8, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.66184045260081*^9},
 CellID->128672720],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"203\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "203"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404526164103`*^9},
 CellID->52144555],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"203\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "203"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840452990811*^9},
 CellID->123913510],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"203\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "203"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404533808117`*^9},
 CellID->65928147],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"203\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "203"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840453770812*^9},
 CellID->54496018],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"203\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "203"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840454160813*^9},
 CellID->6908874],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\NatIntExtended\\\" (\"\>", 
   "\[InvisibleSpace]", "9", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended\" (", 9, " of ", "17", "):"],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840454192013*^9},
 CellID->99796756],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"240\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "240"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840454207613*^9},
 CellID->36210510],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"240\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "240"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404546132135`*^9},
 CellID->101086047],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"240\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "240"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404550344143`*^9},
 CellID->350526751],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"240\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "240"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404554712152`*^9},
 CellID->322095243],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"240\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "240"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840455892416*^9},
 CellID->279166168],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\NatInt\\\" (\"\>", "\[InvisibleSpace]", 
   "10", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt\" (", 10, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404562356167`*^9},
 CellID->622879337],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"300\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "300"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404562512164`*^9},
 CellID->313096369],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"300\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "300"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840456625617*^9},
 CellID->656559649],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"300\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "300"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840457015618*^9},
 CellID->683729469],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"300\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "300"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404574368186`*^9},
 CellID->5824131],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"300\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "300"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404578268194`*^9},
 CellID->382576221],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"300\"\>"}],
  SequenceForm["Reading proof ", 251, " of ", "300"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.66184045820122*^9},
 CellID->117131938],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Numbers\\\" (\"\>", "\[InvisibleSpace]",
    "11", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Numbers\" (", 11, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404585756207`*^9},
 CellID->26901258],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"179\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "179"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404585756207`*^9},
 CellID->120612918],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"179\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "179"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840458981221*^9},
 CellID->46751226],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"179\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "179"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404593868217`*^9},
 CellID->459536186],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"179\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "179"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840459776823*^9},
 CellID->867399896],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Overview\\\" (\"\>", 
   "\[InvisibleSpace]", "12", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Overview\" (", 12, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404600264235`*^9},
 CellID->17672371],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Polynomials\\\" (\"\>", 
   "\[InvisibleSpace]", "13", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Polynomials\" (", 13, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840460042023*^9},
 CellID->32988319],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840460057623*^9},
 CellID->380323932],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840460946825*^9},
 CellID->85422542],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404619140263`*^9},
 CellID->691726092],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840462834428*^9},
 CellID->410364951],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404638172297`*^9},
 CellID->47676749],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 251, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404649248314`*^9},
 CellID->149384025],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 301, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404658140335`*^9},
 CellID->56912388],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\PolyTuples\\\" (\"\>", 
   "\[InvisibleSpace]", "14", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\PolyTuples\" (", 14, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840466640835*^9},
 CellID->149443186],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840466640835*^9},
 CellID->656944083],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404671088357`*^9},
 CellID->163877051],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404676236362`*^9},
 CellID->681446541],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840468107237*^9},
 CellID->241820018],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\ReductionRings\\\" (\"\>", 
   "\[InvisibleSpace]", "15", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings\" (", 15, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404682164373`*^9},
 CellID->22459567],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"250\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "250"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404682320375`*^9},
 CellID->569799938],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"250\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "250"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404691992393`*^9},
 CellID->708515429],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"250\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "250"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840470228841*^9},
 CellID->97046791],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"250\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "250"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840471383243*^9},
 CellID->2171705],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"250\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "250"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840472428445*^9},
 CellID->397458865],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Sequences\\\" (\"\>", 
   "\[InvisibleSpace]", "16", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Sequences\" (", 16, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404735516467`*^9},
 CellID->230520909],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"48\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "48"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840473567247*^9},
 CellID->714598803],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Tuples\\\" (\"\>", "\[InvisibleSpace]", 
   "17", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"17\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples\" (", 17, " of ", "17", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840474160048*^9},
 CellID->60266956],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"228\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "228"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404741756477`*^9},
 CellID->82923131],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"228\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "228"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840474550049*^9},
 CellID->202755098],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"228\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "228"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840474924449*^9},
 CellID->51214078],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"228\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "228"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404753456497`*^9},
 CellID->82288381],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"228\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "228"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840475735651*^9},
 CellID->62924671],

Cell[BoxData["\<\"Creating forward rules:\"\>"], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840475954051*^9},
 CellID->320890793],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840476078851*^9},
 CellID->6169996],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404766560526`*^9},
 CellID->455341864],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 51, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840478340855*^9},
 CellID->215349446],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 101, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840480025658*^9},
 CellID->9203035],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 151, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840481710461*^9},
 CellID->142561407],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 201, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840483410864*^9},
 CellID->10315266],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 251, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404851268673`*^9},
 CellID->14980483],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 301, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.66184048684287*^9},
 CellID->529978144],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 351, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404885120735`*^9},
 CellID->278374742],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 401, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840490165676*^9},
 CellID->132965802],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 451, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840491834879*^9},
 CellID->313688580],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 501, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618404935040817`*^9},
 CellID->13405936],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 551, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840495142085*^9},
 CellID->777943960],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 601, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840496842488*^9},
 CellID->437165283],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 651, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840498511691*^9},
 CellID->285847006],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 701, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840500258894*^9},
 CellID->60183894],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 751, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405019592967`*^9},
 CellID->565313552],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 801, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405036752996`*^9},
 CellID->114713085],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "851", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 851, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840505375703*^9},
 CellID->587975582],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "901", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 901, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840507076106*^9},
 CellID->121683066],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "951", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 951, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840508760909*^9},
 CellID->315404724],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1001", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1001, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405104769115`*^9},
 CellID->260294993],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1051", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1051, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405121617146`*^9},
 CellID->415917723],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1101, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405138309174`*^9},
 CellID->242006766],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1151, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405155157204`*^9},
 CellID->171472899],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1201, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405172005234`*^9},
 CellID->82226196],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1251, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840518869726*^9},
 CellID->259182974],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1301, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405205233297`*^9},
 CellID->132676632],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1351, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840522239332*^9},
 CellID->497601566],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1401, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840523924135*^9},
 CellID->198659878],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1451, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405256245384`*^9},
 CellID->629394221],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1501, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840527324941*^9},
 CellID->121665826],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1551, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405290253444`*^9},
 CellID->370238075],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1601, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405307725477`*^9},
 CellID->28925449],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1651, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405324729505`*^9},
 CellID->120104785],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1701, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840534173353*^9},
 CellID->158010373],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1751, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405358737564`*^9},
 CellID->152430680],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1801, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840537574159*^9},
 CellID->471335972],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1851", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1851, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405392589626`*^9},
 CellID->286858841],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1901", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1901, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405409281654`*^9},
 CellID->137475301],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1951", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 1951, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405425973682`*^9},
 CellID->470908753],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2001", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2001, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840544266571*^9},
 CellID->324454066],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2051", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2051, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405459669743`*^9},
 CellID->386504957],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2101, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405476517773`*^9},
 CellID->787338778],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2151, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.66184054932098*^9},
 CellID->136037090],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2201, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405509745827`*^9},
 CellID->700615950],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2251, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405526125855`*^9},
 CellID->67596746],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2301, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405542661886`*^9},
 CellID->854593272],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2351, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405559197917`*^9},
 CellID->267135323],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2401, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405575577946`*^9},
 CellID->864019571],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2451, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840559211397*^9},
 CellID->272972042],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2501, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405608962*^9},
 CellID->902857225],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2551, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840562565403*^9},
 CellID->322703036],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2601, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840564219006*^9},
 CellID->397535149],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2651, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405658882093`*^9},
 CellID->488933580],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2701, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405675730124`*^9},
 CellID->70778645],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2751, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840569242215*^9},
 CellID->70667339],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2801, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405708958178`*^9},
 CellID->229531394],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2851", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2851, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405725650206`*^9},
 CellID->178020622],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2901", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2901, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405742342234`*^9},
 CellID->52451009],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2951", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 2951, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405759034266`*^9},
 CellID->22005430],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3001", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 3001, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405776038294`*^9},
 CellID->282684566],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3051", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 3051, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405792730327`*^9},
 CellID->33463925],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 3101, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405809734354`*^9},
 CellID->577763587],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3196"}],
  SequenceForm["Creating rule ", 3151, " of ", 3196],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405826738386`*^9},
 CellID->66262662],

Cell[BoxData["\<\"Creating thinned data ...\"\>"], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405842026415`*^9},
 CellID->144319770],

Cell[BoxData["\<\"Creating theory data ...\"\>"], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405843430414`*^9},
 CellID->274432133],

Cell[BoxData["\<\"Creating backward rules ...\"\>"], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840584374242*^9},
 CellID->315474451],

Cell[BoxData["\<\"Creating forward rules ...\"\>"], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840584670642*^9},
 CellID->153604934],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840584686242*^9},
 CellID->8206008],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "15"}],
  SequenceForm["Creating rule ", 1, " of ", 15],
  Editable->False]], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.661840584686242*^9},
 CellID->68279484],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->{3.6536470608445177`*^9, 3.6536502084261513`*^9, 
  3.6582229720045195`*^9, 3.6618405847018423`*^9},
 CellID->67545578]
}, Open  ]],

Cell[BoxData["138.49768779999999424035195261240005493164`8.162042436253497"], \
"Output",
 CellChangeTimes->{3.661840584717442*^9},
 CellID->656798355]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"First", "[", 
  RowBox[{"Timing", "[", 
   RowBox[{"AnalyzeProofs", "[", 
    RowBox[{"Map", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"FileNameJoin", "[", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"NotebookDirectory", "[", "]"}], ",", "#"}], "}"}], "]"}], 
       "&"}], ",", 
      RowBox[{"{", 
       RowBox[{"\"\<ReductionRings\>\"", ",", "\"\<GroebnerRings\>\""}], 
       "}"}]}], "]"}], "]"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6479286682924147`*^9, 3.647928680493279*^9}, {
  3.661840427718766*^9, 3.6618404374063835`*^9}, {3.6621939888295717`*^9, 
  3.66219402193283*^9}},
 CellID->1642448047],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\ReductionRings\\\" (\"\>", 
   "\[InvisibleSpace]", "1", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"2\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings\" (", 1, " of ", "2", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202910818678*^9},
 CellID->1044356],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"252\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "252"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202910818678*^9},
 CellID->116553561],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"252\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "252"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029118014803`*^9},
 CellID->480406741],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"252\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "252"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029128934817`*^9},
 CellID->380961924],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"252\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "252"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202914079084*^9},
 CellID->93536513],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"252\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "252"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029151242857`*^9},
 CellID->687596748],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"252\"\>"}],
  SequenceForm["Reading proof ", 251, " of ", "252"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029162630873`*^9},
 CellID->68034965],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\GroebnerRings\\\" (\"\>", 
   "\[InvisibleSpace]", "2", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"2\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\GroebnerRings\" (", 2, " of ", "2", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202916325488*^9},
 CellID->21001761],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202916325488*^9},
 CellID->193918569],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202916824689*^9},
 CellID->722261688],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029173082895`*^9},
 CellID->106859421],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.66220291779189*^9},
 CellID->379735127],

Cell[BoxData["\<\"Creating forward rules:\"\>"], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202917994691*^9},
 CellID->452350495],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202918025891*^9},
 CellID->36310050],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 1, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202918184891*^9},
 CellID->249258381],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 51, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029188088923`*^9},
 CellID->127180023],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 101, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202919448494*^9},
 CellID->372231125],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 151, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029200568943`*^9},
 CellID->75898686],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 201, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202920680896*^9},
 CellID->7640131],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 251, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202921289297*^9},
 CellID->967827522],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 301, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029219132977`*^9},
 CellID->718513836],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 351, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202922521699*^9},
 CellID->77966652],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 401, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029231301003`*^9},
 CellID->91783338],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 451, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029237541013`*^9},
 CellID->157497756],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 501, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202924362502*^9},
 CellID->472541205],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 551, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202924986503*^9},
 CellID->125784226],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 601, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029255949044`*^9},
 CellID->382809887],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 651, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029262033052`*^9},
 CellID->94555470],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 701, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029268273067`*^9},
 CellID->110179654],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 751, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029274357076`*^9},
 CellID->359138595],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "816"}],
  SequenceForm["Creating rule ", 801, " of ", 816],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029280441084`*^9},
 CellID->12608920],

Cell[BoxData["\<\"Creating thinned data ...\"\>"], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202928231309*^9},
 CellID->207028348],

Cell[BoxData["\<\"Creating theory data ...\"\>"], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202928278109*^9},
 CellID->302515485],

Cell[BoxData["\<\"Creating backward rules ...\"\>"], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202928278109*^9},
 CellID->163275300],

Cell[BoxData["\<\"Creating forward rules ...\"\>"], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202928309309*^9},
 CellID->107663893],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202928324909*^9},
 CellID->206296641],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "9"}],
  SequenceForm["Creating rule ", 1, " of ", 9],
  Editable->False]], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.662202928324909*^9},
 CellID->436869300],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->{3.6621909353624983`*^9, 3.662194022744031*^9, 
  3.6622029283405094`*^9},
 CellID->335789438]
}, Open  ]],

Cell[BoxData["17.11330969999999851438587938901036977768`7.25393392313886"], \
"Output",
 CellChangeTimes->{3.662190963161747*^9, 3.6621940476104746`*^9, 
  3.662202928356109*^9},
 CellID->39547402]
}, Closed]],

Cell[BoxData[
 RowBox[{
  RowBox[{"ComputeFormulaPositions", "[", "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.6406695270485573`*^9, 3.6406695415951242`*^9}, {
   3.64067664934828*^9, 3.640676663567005*^9}, 3.6412902845300865`*^9, 
   3.641290346822996*^9, {3.6442219079512787`*^9, 3.644221909273131*^9}},
 CellID->19428626],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CheckIntegrity", "[", 
  RowBox[{"\"\<CheckPositions\>\"", "\[Rule]", 
   RowBox[{"{", 
    RowBox[{"FileNameJoin", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"CurrentValue", "[", "\"\<NotebookDirectory\>\"", "]"}], ",", 
       RowBox[{"CurrentValue", "[", "\"\<NotebookFileName\>\"", "]"}]}], 
      "}"}], "]"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.64067003109768*^9, 3.6406700374311132`*^9}, {
   3.6406797903560762`*^9, 3.640679792950843*^9}, 3.6442220291527653`*^9},
 CellID->1909803754],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.640670040858255*^9, 3.640670099679185*^9, 3.640670397721546*^9, 
   3.640679793300535*^9, 3.6406798406436234`*^9, 3.640679891162312*^9, 
   3.640679971009297*^9, 3.64068009182163*^9, 3.640680173340745*^9, 
   3.641037395456129*^9, 3.641116403388618*^9, 3.641286881014108*^9, 
   3.642225466578094*^9, 3.642234672070362*^9, 3.642240151894621*^9, 
   3.6423063269273643`*^9, {3.642308832539488*^9, 3.642308857563932*^9}, 
   3.642321396697977*^9, 3.6423285030212173`*^9, 3.642493582768566*^9, 
   3.642763270724422*^9, 3.642834604308413*^9, 3.643428832901271*^9, 
   3.6434441292624083`*^9, 3.643532851309215*^9, 3.644918543299676*^9, 
   3.645518304732339*^9, 3.6472473910089273`*^9, 3.6472474279931*^9, 
   3.6472477145106297`*^9, 3.647247894775585*^9, 3.647248487724378*^9, 
   3.6472495526065273`*^9, 3.64724965160268*^9, 3.647250121944748*^9},
 CellID->154284603]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FindRelation", "[", 
  RowBox[{"\"\<397\>\"", ",", "\"\<396\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.647250160994038*^9, 3.647250175430954*^9}},
 CellID->264346171],

Cell[BoxData["\<\"\[Checkmark] First formula depends on second \
formula.\"\>"], "Output",
 CellChangeTimes->{{3.6472501720688047`*^9, 3.647250176025786*^9}, {
  3.647250253077985*^9, 3.647250274617283*^9}},
 CellID->1745020671]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"psRR", "=", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Total", "[", 
      RowBox[{"Most", "[", 
       RowBox[{"Last", "[", "#", "]"}], "]"}], "]"}], "&"}], ",", 
    RowBox[{"GetProofSize", "[", "All", "]"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.647156479700988*^9, 3.647156479817154*^9}, {
   3.647156560839411*^9, 3.6471566309015427`*^9}, 3.6471570742150927`*^9, {
   3.64715777216581*^9, 3.647157773860202*^9}, {3.651915803603672*^9, 
   3.651915809414452*^9}},
 CellID->1780799392],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "25", ",", "13", ",", "42", ",", "97", ",", "68", ",", "31", ",", "34", ",",
    "17", ",", "15", ",", "76", ",", "20", ",", "40", ",", "30", ",", "47", 
   ",", "72", ",", "17", ",", "164", ",", "34", ",", "23", ",", "27", ",", 
   "31", ",", "42", ",", "25", ",", "26", ",", "73", ",", "24", ",", "21", 
   ",", "29", ",", "17", ",", "53", ",", "10", ",", "31", ",", "47", ",", 
   "10", ",", "32", ",", "20", ",", "67", ",", "30", ",", "36", ",", "5", ",",
    "28", ",", "57", ",", "72", ",", "47", ",", "45", ",", "15", ",", "41", 
   ",", "27", ",", "39", ",", "16", ",", "38", ",", "48", ",", "12", ",", 
   "76", ",", "31", ",", "27", ",", "33", ",", "22", ",", "132", ",", "26", 
   ",", "22", ",", "6", ",", "4", ",", "9", ",", "23", ",", "38", ",", "24", 
   ",", "32", ",", "38", ",", "10", ",", "16", ",", "41", ",", "18", ",", 
   "47", ",", "42", ",", "13", ",", "36", ",", "198", ",", "40", ",", "36", 
   ",", "52", ",", "34", ",", "20", ",", "35", ",", "26", ",", "12", ",", 
   "13", ",", "11", ",", "84", ",", "27", ",", "9", ",", "5", ",", "106", ",",
    "31", ",", "40", ",", "77", ",", "32", ",", "20", ",", "55", ",", "27", 
   ",", "54", ",", "34", ",", "24", ",", "32", ",", "24", ",", "40", ",", 
   "25", ",", "20", ",", "56", ",", "16", ",", "41", ",", "23", ",", "59", 
   ",", "48", ",", "6", ",", "35", ",", "31", ",", "8", ",", "26", ",", "51", 
   ",", "120", ",", "96", ",", "26", ",", "22", ",", "66", ",", "12", ",", 
   "20", ",", "27", ",", "40", ",", "39", ",", "15", ",", "33", ",", "50", 
   ",", "27", ",", "37", ",", "23", ",", "141", ",", "6", ",", "35", ",", 
   "44", ",", "38", ",", "20", ",", "18", ",", "95", ",", "48", ",", "89", 
   ",", "14", ",", "11", ",", "106", ",", "49", ",", "70", ",", "18", ",", 
   "143", ",", "26", ",", "4", ",", "51", ",", "17", ",", "66", ",", "25", 
   ",", "129", ",", "14", ",", "32", ",", "108", ",", "27", ",", "28", ",", 
   "14", ",", "7", ",", "162", ",", "34", ",", "73", ",", "64", ",", "8", ",",
    "28", ",", "29", ",", "32", ",", "24", ",", "40", ",", "39", ",", "16", 
   ",", "61", ",", "31", ",", "23", ",", "66", ",", "55", ",", "18", ",", 
   "82", ",", "36", ",", "15", ",", "18", ",", "16", ",", "37", ",", "23", 
   ",", "76", ",", "105", ",", "88", ",", "48", ",", "4", ",", "20", ",", 
   "25", ",", "5", ",", "28", ",", "25", ",", "12", ",", "11", ",", "55", ",",
    "12", ",", "13", ",", "151", ",", "87", ",", "32", ",", "14", ",", "36", 
   ",", "16", ",", "22", ",", "13", ",", "36", ",", "32", ",", "76", ",", 
   "31", ",", "39", ",", "37", ",", "19", ",", "27", ",", "21", ",", "10", 
   ",", "8", ",", "32", ",", "19", ",", "9", ",", "14", ",", "18", ",", "58", 
   ",", "55", ",", "39", ",", "12", ",", "24", ",", "39", ",", "26", ",", 
   "44", ",", "26", ",", "39", ",", "29", ",", "23", ",", "120", ",", "39", 
   ",", "22", ",", "8", ",", "27", ",", "26", ",", "32", ",", "8", ",", "20", 
   ",", "6"}], "}"}]], "Output",
 CellChangeTimes->{3.647157781066677*^9, 3.651915828355172*^9},
 CellID->778574057]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Length", "[", "psRR", "]"}]], "Input",
 CellChangeTimes->{{3.651915580660548*^9, 3.651915582880301*^9}, {
  3.651915888357747*^9, 3.6519158884806547`*^9}},
 CellID->664434720],

Cell[BoxData["253"], "Output",
 CellChangeTimes->{3.651915583421976*^9, 3.651915888915255*^9},
 CellID->1011754588]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Max", "[", "psRR", "]"}]], "Input",
 CellChangeTimes->{{3.651915339669578*^9, 3.651915341672235*^9}, {
  3.651915442312765*^9, 3.651915442570278*^9}, {3.651915896513073*^9, 
  3.6519158966718063`*^9}},
 CellID->830230433],

Cell[BoxData["198"], "Output",
 CellChangeTimes->{3.65191534206297*^9, 3.651915454173635*^9, 
  3.651915897077359*^9},
 CellID->985680483]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Mean", "[", "psRR", "]"}], "//", "N"}]], "Input",
 CellChangeTimes->{{3.651915271533348*^9, 3.651915305372727*^9}, {
  3.651915444739314*^9, 3.651915445033825*^9}, {3.6519159237861233`*^9, 
  3.651915923945574*^9}},
 CellID->300909609],

Cell[BoxData["38.098814229249015`"], "Output",
 CellChangeTimes->{{3.651915274552681*^9, 3.651915306490151*^9}, 
   3.6519154560565233`*^9, {3.651915919270208*^9, 3.6519159244392147`*^9}},
 CellID->1221498976]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Count", "[", 
  RowBox[{
   RowBox[{"NotebookGet", "[", "]"}], ",", 
   RowBox[{"Cell", "[", 
    RowBox[{"_", ",", "\"\<FormalTextInputFormula\>\"", ",", "___"}], "]"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.64966454060702*^9, 3.649664569556057*^9}, {
  3.6519158304747562`*^9, 3.6519158330444508`*^9}},
 CellID->18261744],

Cell[BoxData["315"], "Output",
 CellChangeTimes->{3.6496646818390512`*^9, 3.6519158345741262`*^9},
 CellID->1943693605]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Auxiliary Functions", "Subsection",
 CellChangeTimes->{{3.640668579267561*^9, 3.640668583347287*^9}},
 CellID->1454182661],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "LoadPackage", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", "name_String", "]"}], ":=", 
   RowBox[{"LoadPackage", "[", 
    RowBox[{"{", "name", "}"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", 
    RowBox[{"l", ":", 
     RowBox[{"{", "___String", "}"}]}], "]"}], ":=", 
   RowBox[{"Scan", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{"FileNameJoin", "[", 
        RowBox[{"{", 
         RowBox[{"$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
          RowBox[{"#", "<>", "\"\<.m\>\""}]}], "}"}], "]"}], "]"}], "&"}], 
     ",", "l"}], "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640668603629292*^9, 3.640668727460284*^9}, 
   3.641286816086794*^9, 3.6434459961973133`*^9, {3.653646981607979*^9, 
   3.6536469961004043`*^9}, 3.661840398203514*^9, {3.666440886655847*^9, 
   3.6664408927751007`*^9}},
 CellID->2074904605],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "FindFormula", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"FindFormula", "[", "label_String", "]"}], ":=", 
   RowBox[{"NotebookLocate", "[", "label", "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640677756847624*^9, 3.6406777896243896`*^9}, 
   3.640677822869519*^9, {3.640677854902379*^9, 3.6406778579321823`*^9}},
 CellID->1980565193],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<ProofPalette\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.640668742006607*^9, 3.640668751417616*^9}},
 CellID->444074165],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<TheoryTools/TheoryTools\>\"", 
  "]"}]], "Input",
 CellChangeTimes->{{3.640668759133181*^9, 3.640668768089197*^9}, {
   3.658222951178482*^9, 3.6582229559052906`*^9}, 3.6664408961712523`*^9},
 CellID->798575911]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Auxiliary Notions (\[Checkmark])", "Section",
 CellChangeTimes->{{3.610168458840133*^9, 3.610168467283736*^9}, {
   3.6412869210291786`*^9, 3.6412869250539856`*^9}, 3.644214208471922*^9, {
   3.647943122132839*^9, 3.647943123813397*^9}, {3.648180728426836*^9, 
   3.648180730924017*^9}},
 CellID->938515413],

Cell[CellGroupData[{

Cell["Definitions", "Subsection",
 CellChangeTimes->{{3.647942926474814*^9, 3.6479429283054953`*^9}},
 CellID->588777529],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "D"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 2066005206}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641209016289589*^9, 3.641209017365279*^9}},
 CellTags->{
  "ID:2066005206", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2066005206],

Cell["", "OpenEnvironment",
 CellID->1781343525],

Cell[CellGroupData[{

Cell["Definition: Auxiliary Notions in Reduction Rings", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.590823916600327*^9, 3.59082392862969*^9}, {
  3.590828453149723*^9, 3.590828455632687*^9}, {3.610179185428216*^9, 
  3.610179194205882*^9}},
 CellID->1935749337],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"M", ",", "i", ",", "j"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 457420627}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590823963710497*^9, 3.590824004325694*^9}, {
   3.590824095052144*^9, 3.590824097449215*^9}, {3.590824162854402*^9, 
   3.590824163039427*^9}, {3.590824576771481*^9, 3.590824576992859*^9}, {
   3.590825461376229*^9, 3.590825462000802*^9}, {3.590825682372273*^9, 
   3.590825682517584*^9}, {3.590828888574193*^9, 3.590828891942208*^9}, {
   3.590829694962239*^9, 3.590829696326725*^9}, 3.590830138615678*^9, {
   3.59160256262753*^9, 3.591602562849449*^9}, {3.592041221132267*^9, 
   3.592041226881355*^9}, {3.608534967350515*^9, 3.608534981695014*^9}, {
   3.608535049730997*^9, 3.608535070408885*^9}, {3.608535132762676*^9, 
   3.608535132948086*^9}, {3.608535282481944*^9, 3.60853528603165*^9}, 
   3.641209019988555*^9},
 CellTags->{
  "ID:457420627", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->457420627],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "a", ",", "b", ",", "c", ",", "c1", ",", "c2", ",", "m1", ",", 
   "m2"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 176465234}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.591602547161186*^9, 3.591602578043871*^9}, {
   3.591602657990751*^9, 3.591602664045186*^9}, 3.591602719677581*^9, {
   3.592024648731523*^9, 3.592024648990018*^9}, 3.59402907638008*^9, {
   3.608534961486853*^9, 3.608534978782966*^9}, {3.610168872292143*^9, 
   3.610168882023043*^9}, {3.638088396184882*^9, 3.6380883979519873`*^9}},
 CellTags->{
  "ID:176465234", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->176465234],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"k", "=", "0"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "M", "\[RightBracketingBar]"}]}]], 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "M", ",", "k"}], "]"}], ":=", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox["-", "D"], 
      RowBox[{
       UnderscriptBox[
        SubscriptBox["\[Sum]", "D"], 
        RowBox[{
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         SubscriptBox["M", "j"], 
         UnderscriptBox["*", "D"], "c"}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ")"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduction with several multipliers)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:478302118", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590825465292219*^9, 3.590825524029538*^9}, {
  3.59083014096991*^9, 3.590830142477053*^9}, {3.607336537650323*^9, 
  3.607336538639757*^9}, {3.610169159409674*^9, 3.610169161337321*^9}, {
  3.637996872106268*^9, 3.637996880792054*^9}, {3.641116930748708*^9, 
  3.64111694237354*^9}},
 CellTags->{
  "ID:478302118", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction with several multipliers"},
 CellID->478302118],

Cell[TextData[{
 "The ",
 Cell[BoxData[
  FormBox["symred", TraditionalForm]]],
 " predicate defined below is slightly different from ",
 Cell[BoxData[
  FormBox["\[LeftRightArrow]", TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " has to be equal to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "-", 
    RowBox[{"m", "*", "c"}]}], TraditionalForm]]],
 ", and not vice versa!"
}], "EnvironmentText",
 CellChangeTimes->{{3.610168776431095*^9, 3.610168851495103*^9}},
 CellID->624328162],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["symred", "D"], "[", 
   RowBox[{"a", ",", "b", ",", "c", ",", "m1"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "D"], "[", 
     RowBox[{"m1", ",", "c"}], "]"}], "\[And]", 
    RowBox[{"b", "\[Equal]", 
     RowBox[{"a", 
      UnderscriptBox["-", "D"], 
      RowBox[{"m1", 
       UnderscriptBox["*", "D"], "c"}]}]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"a", 
       UnderscriptBox["\[Succeeds]", "D"], "b"}], "\[Or]", 
      RowBox[{"b", 
       UnderscriptBox["\[Succeeds]", "D"], "a"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symmetric closure of reducibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:448370877", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590825550198452*^9, 3.590825720158915*^9}, {
  3.5908262580053368`*^9, 3.590826360750206*^9}, {3.590915680051633*^9, 
  3.590915681485019*^9}, {3.591944352699075*^9, 3.591944354127339*^9}, {
  3.607336545428724*^9, 3.607336558415432*^9}, {3.60733658945308*^9, 
  3.607336590117591*^9}, {3.610169171932111*^9, 3.610169174409538*^9}, {
  3.641182025909811*^9, 3.641182029111864*^9}},
 CellTags->{
  "ID:448370877", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetric closure of reducibility"},
 CellID->448370877],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{"a", ",", "c"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], "[", "m", "]"}]], 
   RowBox[{
    UnderscriptBox["isReducible", "D"], "[", 
    RowBox[{"a", ",", "c", ",", "m"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reducible1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1537297721", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590828677787307*^9, 3.590828733302799*^9}, 
   3.594029082715761*^9, {3.607336566767904*^9, 3.60733657071679*^9}, {
   3.610169135082098*^9, 3.610169149785712*^9}, {3.631962272816483*^9, 
   3.631962275341296*^9}},
 CellTags->{
  "ID:1537297721", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reducible1"},
 CellID->1537297721],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{"a", ",", "c", ",", "m1"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["\[Succeeds]", "D"], 
     RowBox[{"a", 
      UnderscriptBox["-", "D"], 
      RowBox[{"m1", 
       UnderscriptBox["*", "D"], "c"}]}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["mult", "D"], "[", 
     RowBox[{"m1", ",", "c"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reducible2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:115218056", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590828753712627*^9, 3.590828798018439*^9}, {
   3.590833113348102*^9, 3.5908331173198*^9}, 3.590915686915306*^9, {
   3.591944399339646*^9, 3.591944416195219*^9}, {3.60733657278489*^9, 
   3.607336573873034*^9}, {3.61016913771377*^9, 3.610169145966229*^9}},
 CellTags->{
  "ID:115218056", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reducible2"},
 CellID->115218056],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isIrreducible", "D"], "[", "a", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[Not]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", "c0", "]"}]}]], 
     RowBox[{
      UnderscriptBox["isReducible", "D"], "[", 
      RowBox[{"a", ",", "c0"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(irreducibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:399781455", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.591434067694727*^9, 3.59143411909756*^9}, {
   3.591434165789519*^9, 3.591434187868742*^9}, {3.591434710151263*^9, 
   3.591434710519005*^9}, {3.591587063495766*^9, 3.591587066290242*^9}, {
   3.591587139729318*^9, 3.591587141614208*^9}, 3.594028892075185*^9, {
   3.59618397186221*^9, 3.59618397720894*^9}, {3.596187160303093*^9, 
   3.596187166968071*^9}, {3.596188960377264*^9, 3.596188962399591*^9}, {
   3.606127446122285*^9, 3.6061274544202*^9}, {3.607772653246648*^9, 
   3.607772675445367*^9}, {3.607772869193103*^9, 3.60777287336732*^9}, {
   3.610169050425633*^9, 3.610169129588022*^9}},
 CellTags->{
  "ID:399781455", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "irreducibility"},
 CellID->399781455],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areIrrelative", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "m1", ",", "m2", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c1", "\[Equal]", "c2"}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", 
      RowBox[{
       UnderscriptBox["mult", "D"], "[", 
       RowBox[{"m1", ",", "c2", ",", "i"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["mult", "D"], "[", 
       RowBox[{"m2", ",", "c2", ",", "j"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(irrelative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1867242863", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590828871400857*^9, 3.590828884051465*^9}, {
   3.590828915824394*^9, 3.590829054055385*^9}, {3.60733657592441*^9, 
   3.607336618961079*^9}, {3.607336661198396*^9, 3.607336667590845*^9}, {
   3.60853528916981*^9, 3.608535298316548*^9}, 3.608548571293774*^9, {
   3.609064719283596*^9, 3.609064732067266*^9}, {3.610169181457439*^9, 
   3.610169186883537*^9}, {3.610179995728761*^9, 3.61017999809717*^9}, {
   3.631855003700605*^9, 3.6318550069892597`*^9}, {3.631856278517174*^9, 
   3.631856288140665*^9}, {3.632477373424382*^9, 3.6324773855066757`*^9}, {
   3.632549720887106*^9, 3.63254972162768*^9}, {3.632549875795684*^9, 
   3.632549913003624*^9}},
 CellTags->{
  "ID:1867242863", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "irrelative"},
 CellID->1867242863],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areCorrelative", "D"], "[", 
   RowBox[{"c", ",", "m1", ",", "m2"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       UnderscriptBox["multN", "D"], "[", "c", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["mult", "D"], "[", 
       RowBox[{"m1", ",", "c", ",", "k"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["mult", "D"], "[", 
       RowBox[{"m2", ",", "c", ",", "k"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(correlative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1834165766", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6324772812193727`*^9, 3.6324773520873938`*^9}},
 CellTags->{
  "ID:1834165766", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative"},
 CellID->1834165766],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isCommonReducible", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], "[", 
     RowBox[{"m3", ",", "m4"}], "]"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"\[And]", 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{
          UnderscriptBox["isReducible", "D"], "[", 
          RowBox[{"a", ",", "c1", ",", "m3"}], "]"}]},
        {
         RowBox[{
          UnderscriptBox["isReducible", "D"], "[", 
          RowBox[{"a", ",", "c2", ",", "m4"}], "]"}]},
        {
         RowBox[{
          UnderscriptBox["areIrrelative", "D"], "[", 
          RowBox[{"c1", ",", "c2", ",", "m3", ",", "m4", ",", "i", ",", "j"}],
           "]"}]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}},
          "RowsIndexed" -> {}}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(common reducible)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1907029362", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590829073560136*^9, 3.590829126901262*^9}, {
   3.590830116990216*^9, 3.590830126949827*^9}, 3.594029093280512*^9, {
   3.607336631610574*^9, 3.607336643932552*^9}, {3.610169215020395*^9, 
   3.610169246427018*^9}, {3.6325451638924026`*^9, 3.632545239969036*^9}, {
   3.632549679203602*^9, 3.6325496817109632`*^9}, {3.632549946753043*^9, 
   3.632549961322081*^9}},
 CellTags->{
  "ID:1907029362", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common reducible"},
 CellID->1907029362],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isCommonReducible", "D"], "[", 
     RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], "\[And]", 
    RowBox[{"\[Not]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "D"], "[", 
         RowBox[{"m3", ",", "m4"}], "]"}]], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             UnderscriptBox["areIrrelative", "D"], "[", 
             RowBox[{
             "c1", ",", "c2", ",", "m3", ",", "m4", ",", "i", ",", "j"}], 
             "]"}]},
           {
            RowBox[{
             UnderscriptBox["isReducible", "D"], "[", 
             RowBox[{"a", ",", "c1", ",", "m3"}], "]"}]},
           {
            RowBox[{
             UnderscriptBox["isReducible", "D"], "[", 
             RowBox[{"a", ",", "c2", ",", "m4"}], "]"}]},
           {
            RowBox[{
             RowBox[{
              UnderscriptBox["isReducible", "D"], "[", 
              RowBox[{
               RowBox[{"a", 
                UnderscriptBox["-", "D"], 
                RowBox[{"m3", 
                 UnderscriptBox["*", "D"], "c1"}]}], ",", "c2", ",", "m4"}], 
              "]"}], "\[Or]", 
             RowBox[{
              UnderscriptBox["isReducible", "D"], "[", 
              RowBox[{
               RowBox[{"a", 
                UnderscriptBox["-", "D"], 
                RowBox[{"m4", 
                 UnderscriptBox["*", "D"], "c2"}]}], ",", "c1", ",", "m3"}], 
              "]"}]}]}
          }]}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(non-trivial common reducible)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2013437879", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590829957275174*^9, 3.590830031924871*^9}, {
   3.590830064987725*^9, 3.590830069796473*^9}, 3.594029100099359*^9, {
   3.607336694370367*^9, 3.607336740466475*^9}, {3.608535142202343*^9, 
   3.608535201893708*^9}, {3.608535260944541*^9, 3.608535270510319*^9}, {
   3.609064781174161*^9, 3.609064816034325*^9}, {3.610169256479841*^9, 
   3.610169319490342*^9}, {3.610180007716566*^9, 3.610180008604242*^9}, {
   3.632480311196609*^9, 3.632480316437512*^9}, {3.6324803709322653`*^9, 
   3.6324803943944674`*^9}, {3.632481337517845*^9, 3.632481338033474*^9}, {
   3.632545276309689*^9, 3.6325452770853786`*^9}, {3.632549659945444*^9, 
   3.632549662347949*^9}, 3.632549969918211*^9, {3.638695646073112*^9, 
   3.638695656911536*^9}},
 CellTags->{
  "ID:2013437879", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "non-trivial common reducible"},
 CellID->2013437879],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isNTCR", "D"], "[", 
     RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], "[", "a0", "]"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"a", 
        UnderscriptBox["\[Succeeds]", "D"], "a0"}], "\[Implies]", 
       RowBox[{"\[Not]", 
        RowBox[{
         UnderscriptBox["isNTCR", "D"], "[", 
         RowBox[{"c1", ",", "c2", ",", "a0", ",", "i", ",", "j"}], "]"}]}]}], 
      
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(minimal non-trivial common reducible)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:991122907", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59082934628178*^9, 3.590829383039992*^9}, {
   3.590830084688404*^9, 3.590830092594891*^9}, 3.590915705181246*^9, {
   3.591944452602626*^9, 3.591944456094625*^9}, 3.594029110553413*^9, {
   3.607336770565164*^9, 3.607336783126306*^9}, {3.608535239631606*^9, 
   3.608535279057474*^9}, {3.610169352398372*^9, 3.610169377295486*^9}},
 CellTags->{
  "ID:991122907", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "minimal non-trivial common reducible"},
 CellID->991122907],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "D"], "[", 
   RowBox[{"c", ",", "a"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "D"], "[", 
     RowBox[{"a", ",", "c"}], "]"}], "\[And]", 
    RowBox[{"\[Not]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "D"], "[", 
         RowBox[{"m3", ",", "m4"}], "]"}]], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             UnderscriptBox["isReducible", "D"], "[", 
             RowBox[{"a", ",", "c", ",", "m3"}], "]"}]},
           {
            RowBox[{
             UnderscriptBox["isReducible", "D"], "[", 
             RowBox[{"a", ",", "c", ",", "m4"}], "]"}]},
           {
            RowBox[{
             UnderscriptBox["isReducible", "D"], "[", 
             RowBox[{
              RowBox[{"a", 
               UnderscriptBox["-", "D"], 
               RowBox[{"m4", 
                UnderscriptBox["*", "D"], "c"}]}], ",", "c", ",", "m3"}], 
             "]"}]}
          }]}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(unary non-trivial common reducible)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1293246145", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590829862624437*^9, 3.590829922777802*^9}, 
   3.59402911493528*^9, {3.607336809606557*^9, 3.607336821927008*^9}, {
   3.610169399660644*^9, 3.610169437879616*^9}},
 CellTags->{
  "ID:1293246145", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "unary non-trivial common reducible"},
 CellID->1293246145],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "D"], "[", 
   RowBox[{"c", ",", "a"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isNTCR", "D"], "[", 
     RowBox[{"c", ",", "a"}], "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], "[", "a0", "]"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"a", 
        UnderscriptBox["\[Succeeds]", "D"], "a0"}], "\[Implies]", 
       RowBox[{"\[Not]", 
        RowBox[{
         UnderscriptBox["isNTCR", "D"], "[", 
         RowBox[{"c", ",", "a0"}], "]"}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(unary minimal non-trivial common reducible)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:894750987", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590829498501986*^9, 3.590829524523363*^9}, {
   3.590829784419415*^9, 3.590829797171937*^9}, 3.590915712369493*^9, {
   3.591944467304619*^9, 3.591944470687504*^9}, 3.594029120568978*^9, {
   3.607336829238085*^9, 3.607336839194407*^9}, {3.61016944429499*^9, 
   3.610169468346235*^9}},
 CellTags->{
  "ID:894750987", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "unary minimal non-trivial common reducible"},
 CellID->894750987],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areCPMultipliers", "D"], "[", 
   RowBox[{
   "m1", ",", "m2", ",", "c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], 
   "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["isMNTCR", "D"], "[", 
         RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["areIrrelative", "D"], "[", 
         RowBox[{"c1", ",", "c2", ",", "m1", ",", "m2", ",", "i", ",", "j"}], 
         "]"}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "D"], "[", 
         RowBox[{"a", ",", "c1", ",", "m1"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["isReducible", "D"], "[", 
         RowBox[{"a", ",", "c2", ",", "m2"}], "]"}]}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(critical pair multipliers)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:779388596", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590829528488973*^9, 3.59082971942228*^9}, 
   3.594029124877941*^9, {3.594459445921877*^9, 3.594459531416444*^9}, {
   3.607336843790875*^9, 3.607336844823897*^9}, {3.608548446715484*^9, 
   3.608548456462536*^9}, {3.608548536640968*^9, 3.608548580504253*^9}, {
   3.608548819777591*^9, 3.608548896587504*^9}, {3.608608350754681*^9, 
   3.608608368502546*^9}, {3.609064876855927*^9, 3.60906489837296*^9}, {
   3.61016948112921*^9, 3.610169543718496*^9}, {3.610180043037494*^9, 
   3.610180044073868*^9}, {3.632479961798666*^9, 3.63247996665466*^9}, {
   3.6324814321917353`*^9, 3.632481432634193*^9}, {3.632549503997959*^9, 
   3.6325495144158497`*^9}, {3.6325499892541656`*^9, 3.632549992123736*^9}, {
   3.643428889998291*^9, 3.643428893974841*^9}},
 CellTags->{
  "ID:779388596", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "critical pair multipliers"},
 CellID->779388596],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1118686892]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->297500119],

Cell["", "OpenEnvironment",
 CellID->135917884],

Cell[CellGroupData[{

Cell["\<\
Definition: Reduction Relation, Ideal congruence, and Equivalence Relation\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.610168570090663*^9, 3.610168577775442*^9}, 
   3.610168743955427*^9, {3.64525843279847*^9, 3.645258438040132*^9}, {
   3.650337429435081*^9, 3.650337436726357*^9}},
 CellID->1118462943],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"a", ",", "b", ",", "r"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1517092650}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59143178635081*^9, 3.591431831562551*^9}, {
   3.591432646565431*^9, 3.591432646713426*^9}, 3.591433264698648*^9, {
   3.591437764464284*^9, 3.591437767375156*^9}, 3.591437868638033*^9, {
   3.591438790713265*^9, 3.591438793766876*^9}, {3.591438841492706*^9, 
   3.59143884650047*^9}, 3.591439442272657*^9, {3.591440552678477*^9, 
   3.591440561680918*^9}, {3.591588457804693*^9, 3.591588465491547*^9}, {
   3.591595253642618*^9, 3.591595253827794*^9}, 3.594028888870022*^9, {
   3.596183860836708*^9, 3.596183864698375*^9}, {3.596183914973167*^9, 
   3.596183915341763*^9}, {3.596187085435803*^9, 3.596187085694589*^9}, 
   3.596187848481059*^9, {3.608025530341739*^9, 3.60802553071257*^9}, {
   3.610170187199247*^9, 3.610170193680056*^9}, {3.61017023152929*^9, 
   3.610170257278508*^9}, {3.610170476188343*^9, 3.610170496353222*^9}, {
   3.6101707431448*^9, 3.610170744877814*^9}, {3.610177830282763*^9, 
   3.61017784057143*^9}, {3.6357486577729473`*^9, 3.635748665691436*^9}, {
   3.638088424628043*^9, 3.638088425989767*^9}, {3.644218233567141*^9, 
   3.644218268896324*^9}, {3.6442183683148127`*^9, 3.644218373767901*^9}, {
   3.645433316249939*^9, 3.6454333228272743`*^9}, {3.645500680084197*^9, 
   3.645500700186306*^9}, 3.645500736305715*^9},
 CellTags->{
  "ID:1517092650", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1517092650],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", "c", "]"}]}], ",", "m"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      RowBox[{"a", 
       UnderscriptBox[
        SubscriptBox["\[Rule]", 
         RowBox[{"m", ",", "c"}]], "D"], "b"}],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"b", "\[Equal]", 
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{"m", 
         UnderscriptBox["*", "D"], "c"}]}]}], "\[And]", 
      RowBox[{"a", 
       UnderscriptBox["\[Succeeds]", "D"], "b"}], "\[And]", 
      RowBox[{
       UnderscriptBox["mult", "D"], "[", 
       RowBox[{"m", ",", "c"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduction modulo element using multiplier)", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:935236171", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.608527156018613*^9, 3.60852718471452*^9}, {
   3.608608063168635*^9, 3.608608063424618*^9}, 3.610170212927163*^9, {
   3.645500602844942*^9, 3.645500615369398*^9}, {3.645500733705821*^9, 
   3.645500733922798*^9}},
 CellTags->{
  "ID:935236171", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo element using multiplier"},
 CellID->935236171],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "D", "]"}]], 
    RowBox[{"[", "S", "]"}]}]], 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      RowBox[{"a", 
       UnderscriptBox[
        SubscriptBox["\[Rule]", "S"], "D"], "b"}],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{"s", "\[Element]", "S"}]], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "D"], 
         RowBox[{"[", "m", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox[
         RowBox[{"a", 
          UnderscriptBox[
           SubscriptBox["\[Rule]", 
            RowBox[{"m", ",", "s"}]], "D"], "b"}],
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ")"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduction modulo set)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:320624715", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644218275867365*^9, 3.644218294674268*^9}, {
  3.644300732741604*^9, 3.644300734663513*^9}, {3.645500686293995*^9, 
  3.645500704449911*^9}, {3.66184951056044*^9, 3.6618495292024727`*^9}},
 CellTags->{
  "ID:320624715", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo set"},
 CellID->320624715],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "T", "]"}]}]], 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      RowBox[{"a", 
       UnderscriptBox[
        SubscriptBox["\[Rule]", "T"], "D"], "b"}],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "D"], 
         RowBox[{"[", "m", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox[
         RowBox[{"a", 
          UnderscriptBox[
           SubscriptBox["\[Rule]", 
            RowBox[{"m", ",", 
             SubscriptBox["T", "i"]}]], "D"], "b"}],
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ")"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduction modulo tuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:434515171", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.591425266801407*^9, 3.591425287636411*^9}, {
   3.59142533139486*^9, 3.591425350297535*^9}, {3.591425394926737*^9, 
   3.591425458229722*^9}, {3.591425511066634*^9, 3.591425515558187*^9}, {
   3.591425640334893*^9, 3.591425658696509*^9}, {3.59142581789998*^9, 
   3.591425838188707*^9}, {3.591429052447596*^9, 3.591429056424372*^9}, {
   3.591429125057811*^9, 3.591429138509355*^9}, {3.591434327832632*^9, 
   3.591434328653354*^9}, {3.591602643244317*^9, 3.591602644098397*^9}, 
   3.594029067070045*^9, {3.596189640138043*^9, 3.596189642975332*^9}, {
   3.596189679506783*^9, 3.596189688497991*^9}, {3.606128119557607*^9, 
   3.606128124279504*^9}, {3.608525786391162*^9, 3.608525837895108*^9}, {
   3.608525970898197*^9, 3.608525979685426*^9}, 3.610170220503903*^9, {
   3.610170261632434*^9, 3.610170267414807*^9}, {3.610170342783572*^9, 
   3.610170347901203*^9}, {3.644218236935968*^9, 3.644218252656945*^9}, {
   3.6455006555987873`*^9, 3.6455006779026537`*^9}, 3.6618496608199043`*^9, {
   3.6618497333132315`*^9, 3.6618497448104515`*^9}},
 CellTags->{
  "ID:434515171", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo tuple"},
 CellID->434515171],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReductor", "D"], "[", "r", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainSets", "[", "D", "]"}]], 
     RowBox[{"[", "r", "]"}]}], "\[Or]", 
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainTuples", "[", "D", "]"}]], 
     RowBox[{"[", "r", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReductor)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:773828270", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644750835193091*^9, 3.644750875872498*^9}, {
   3.644750990175583*^9, 3.64475099199347*^9}, 3.6618496265778437`*^9},
 CellTags->{
  "ID:773828270", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReductor"},
 CellID->773828270],

Cell[TextData[{
 "We do not specify what ",
 Cell[BoxData[
  FormBox["r", TraditionalForm]]],
 " is. Hence, the following definitions work both for ",
 Cell[BoxData[
  FormBox["r", TraditionalForm]]],
 " being a tuple and set of ring elements (and even anything else)."
}], "EnvironmentText",
 CellChangeTimes->{{3.610170812018899*^9, 3.610170855226681*^9}, {
  3.61017216596366*^9, 3.610172166075907*^9}, {3.6442183302627697`*^9, 
  3.644218330972926*^9}, {3.644218382431707*^9, 3.6442183971218853`*^9}, {
  3.644750938013939*^9, 3.644750954949216*^9}, {3.661842225448124*^9, 
  3.661842237538145*^9}},
 CellID->766317937],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    RowBox[{"a", 
     UnderscriptBox[
      SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           SubscriptBox["T", "1"], "\[Equal]", "a"}]},
         {
          RowBox[{
           SubscriptBox["T", 
            RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
           "\[Equal]", "b"}]},
         {
          RowBox[{"isChain", "[", 
           RowBox[{"T", ",", 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             TagBox[
              UnderscriptBox[
               SubscriptBox["\[Rule]", "r"], "D"],
              Identity], 
             TagBox[")",
              "AutoParentheses"]}], ",", "D"}], "]"}]}
        }]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(reflexive-transitive closure of reduction relation)", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1813157585", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.591432166197958*^9, 3.591432221685994*^9}, {
   3.591587191398122*^9, 3.591587200551645*^9}, {3.591587305251368*^9, 
   3.591587320591959*^9}, {3.59618713373776*^9, 3.596187144189481*^9}, {
   3.596187625217739*^9, 3.596187659884779*^9}, {3.596187692970332*^9, 
   3.596187694001492*^9}, {3.606127644854501*^9, 3.606127651459357*^9}, {
   3.61017086463316*^9, 3.610170903938116*^9}, 3.610171841914217*^9, {
   3.644300622396838*^9, 3.644300638262888*^9}, {3.64474916636261*^9, 
   3.64474916980704*^9}},
 CellTags->{
  "ID:1813157585", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
   "reflexive-transitive closure of reduction relation"},
 CellID->1813157585],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    RowBox[{"a", 
     UnderscriptBox[
      SubscriptBox["\[LeftRightArrow]", "r"], "D"], "b"}],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      RowBox[{"a", 
       UnderscriptBox[
        SubscriptBox["\[Rule]", "r"], "D"], "b"}],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}], "\[Or]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      RowBox[{"b", 
       UnderscriptBox[
        SubscriptBox["\[Rule]", "r"], "D"], "a"}],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symmetric closure of reduction relation)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:370202704", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.591432270671753*^9, 3.591432307344777*^9}, {
   3.591587216537401*^9, 3.59158723383357*^9}, {3.596187674174268*^9, 
   3.596187691755202*^9}, {3.5961897400747*^9, 3.59618974332281*^9}, {
   3.60612766322059*^9, 3.606127664474819*^9}, {3.610170916210006*^9, 
   3.610170978401118*^9}, {3.610171630635729*^9, 3.610171635655057*^9}, 
   3.610171838914905*^9, {3.641209895331505*^9, 3.64120989669969*^9}},
 CellTags->{
  "ID:370202704", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetric closure of reduction relation"},
 CellID->370202704],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    RowBox[{"a", 
     UnderscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], "D"], "b"}],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           SubscriptBox["T", "1"], "\[Equal]", "a"}]},
         {
          RowBox[{
           SubscriptBox["T", 
            RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
           "\[Equal]", "b"}]},
         {
          RowBox[{"isChain", "[", 
           RowBox[{"T", ",", 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             TagBox[
              UnderscriptBox[
               SubscriptBox["\[LeftRightArrow]", "r"], "D"],
              Identity], 
             TagBox[")",
              "AutoParentheses"]}], ",", "D"}], "]"}]}
        }]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(symmetric-reflexive-transitive closure of reduction relation)", 
         "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:992563737", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.591432506012062*^9, 3.591432561762774*^9}, {
   3.591587363916383*^9, 3.591587381191669*^9}, {3.596187806666029*^9, 
   3.596187816980348*^9}, {3.596189833496225*^9, 3.596189852923821*^9}, {
   3.606127719803379*^9, 3.606127725167333*^9}, {3.610171606488937*^9, 
   3.61017165300331*^9}, 3.610171833221708*^9, {3.644300673677339*^9, 
   3.644300693184905*^9}, {3.6447491727967854`*^9, 3.644749175950213*^9}},
 CellTags->{
  "ID:992563737", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
   "symmetric-reflexive-transitive closure of reduction relation"},
 CellID->992563737],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    RowBox[{"a", 
     UnderscriptBox[
      SubscriptBox["\[DownArrow]", "r"], "D"], "b"}],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", "d", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox[
        RowBox[{"a", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", "r"], "D"], "d"}],
        Identity], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox[
        RowBox[{"b", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", "r"], "D"], "d"}],
        Identity], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(common successor w.r.t. reduction relation)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:414012405", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.5914326039513*^9, 3.5914326548637*^9}, {
   3.591432697468433*^9, 3.591432700673362*^9}, {3.591432737407343*^9, 
   3.591432771603045*^9}, {3.591587393284646*^9, 3.591587397156072*^9}, {
   3.591587497268272*^9, 3.591587498998486*^9}, 3.594028928700156*^9, {
   3.59618782595442*^9, 3.596187864035416*^9}, {3.59618985955195*^9, 
   3.596189863031835*^9}, {3.60612773121205*^9, 3.60612773534424*^9}, {
   3.610171849287497*^9, 3.610171929464675*^9}, 3.6435348250188713`*^9},
 CellTags->{
  "ID:414012405", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common successor w.r.t. reduction relation"},
 CellID->414012405],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    RowBox[{"a", 
     UnderscriptBox[
      SubsuperscriptBox["\[DownArrow]", "r", "*"], "D"], "b"}],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", "d", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox[
        RowBox[{"a", 
         UnderscriptBox[
          SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "d"}],
        Identity], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox[
        RowBox[{"b", 
         UnderscriptBox[
          SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "d"}],
        Identity], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(reflexive-transitive closure of common successor)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:895393520", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.591432785726282*^9, 3.591432867510809*^9}, {
   3.591587503556478*^9, 3.591587525720373*^9}, {3.594028931459738*^9, 
   3.594028934449562*^9}, {3.596187867717118*^9, 3.596187886320327*^9}, {
   3.596189866455207*^9, 3.596189874084155*^9}, {3.606127741688569*^9, 
   3.606127745120342*^9}, {3.610171937622457*^9, 3.610171965129358*^9}, 
   3.643534820675205*^9},
 CellTags->{
  "ID:895393520", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexive-transitive closure of common successor"},
 CellID->895393520],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    RowBox[{"a", 
     UnderscriptBox[
      SubscriptBox["\[UpArrow]", "r"], "D"], "b"}],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", "d", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox[
        RowBox[{"d", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", "r"], "D"], "a"}],
        Identity], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox[
        RowBox[{"d", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", "r"], "D"], "b"}],
        Identity], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(common predecessor w.r.t. reduction relation)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:586157110", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.591433558967418*^9, 3.591433596764868*^9}, {
   3.591433693708975*^9, 3.591433698513777*^9}, {3.591587543511086*^9, 
   3.59158755415745*^9}, 3.59402893794841*^9, {3.596187894494304*^9, 
   3.596187912794861*^9}, {3.596189880303443*^9, 3.59618988333917*^9}, {
   3.60612775072773*^9, 3.606127754159271*^9}, {3.610171983140485*^9, 
   3.610172014682387*^9}, 3.64353479317384*^9, {3.643534871234995*^9, 
   3.6435348824326973`*^9}},
 CellTags->{
  "ID:586157110", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common predecessor w.r.t. reduction relation"},
 CellID->586157110],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    RowBox[{"a", 
     UnderscriptBox[
      SubsuperscriptBox["\[UpArrow]", "r", "*"], "D"], "b"}],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", "d", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       RowBox[{"d", 
        UnderscriptBox[
         SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "a"}],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}], "\[And]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       RowBox[{"d", 
        UnderscriptBox[
         SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(reflexive-transitive closure of common predecessor)", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2045822907", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.591433714983599*^9, 3.591433736785539*^9}, {
   3.591587560129638*^9, 3.59158757882676*^9}, 3.594028940895683*^9, {
   3.596187916730243*^9, 3.596187932015799*^9}, {3.596189886164903*^9, 
   3.59618989323245*^9}, {3.606127757260592*^9, 3.606127761024371*^9}, {
   3.610172023980872*^9, 3.610172042634824*^9}},
 CellTags->{
  "ID:2045822907", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
   "reflexive-transitive closure of common predecessor"},
 CellID->2045822907],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "z"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      RowBox[{"a", 
       UnderscriptBox[
        OverscriptBox[
         SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
         RowBox[{"z", ",", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           TagBox["\[Succeeds]",
            Identity], 
           TagBox[")",
            "AutoParentheses"]}]}]], "D"], "b"}],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainTuples", "[", "D", "]"}]], 
       RowBox[{"[", "T", "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{"T", "\[NotEqual]", 
            RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
          {
           RowBox[{
            SubscriptBox["T", "1"], "\[Equal]", "a"}]},
          {
           RowBox[{
            SubscriptBox["T", 
             RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
            "\[Equal]", "b"}]},
          {
           RowBox[{"isChainBelow", "[", 
            RowBox[{"T", ",", "z", ",", 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              TagBox[
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", "r"], "D"],
               Identity], 
              TagBox[")",
               "AutoParentheses"]}], ",", 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              TagBox[
               UnderscriptBox["\[Succeeds]", "D"],
               Identity], 
              TagBox[")",
               "AutoParentheses"]}], ",", "D"}], "]"}]}
         }]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(connectibility below)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:463604795", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.591595183167616*^9, 3.591595198744349*^9}, {
   3.59159524588894*^9, 3.591595400464745*^9}, {3.591595480604053*^9, 
   3.591595517803222*^9}, {3.591595691347974*^9, 3.591595703458212*^9}, {
   3.592025629585267*^9, 3.592025629733456*^9}, 3.594028966887706*^9, {
   3.596188084768566*^9, 3.596188213681333*^9}, {3.596188247378838*^9, 
   3.596188247820957*^9}, {3.596189185728335*^9, 3.596189190080484*^9}, {
   3.596190037448152*^9, 3.596190049878458*^9}, {3.60612784060822*^9, 
   3.606127854316381*^9}, {3.607772953370347*^9, 3.60777296903522*^9}, {
   3.610172239016549*^9, 3.610172328326397*^9}, {3.610177910467875*^9, 
   3.610177926621867*^9}, {3.6357487037109137`*^9, 3.635748713175703*^9}, {
   3.644231963423218*^9, 3.6442319659507847`*^9}, {3.644232019517603*^9, 
   3.644232023257346*^9}, {3.6442971486518927`*^9, 3.644297157007794*^9}, {
   3.6443006989388103`*^9, 3.6443007032347918`*^9}, {3.644749179741208*^9, 
   3.6447491836171293`*^9}, {3.64550076183664*^9, 3.645500777131424*^9}},
 CellTags->{
  "ID:463604795", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below"},
 CellID->463604795],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "m"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Tilde]", "m"], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"b", "\[Equal]", 
           RowBox[{"m", 
            UnderscriptBox["*", "D"], "a"}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "D"], 
             RowBox[{"[", 
              RowBox[{"x", ",", "y"}], "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{"x", 
               UnderscriptBox["\[Succeeds]", "D"], "y"}], 
              TagBox[")",
               "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               RowBox[{"m", 
                UnderscriptBox["*", "D"], "x"}], 
               UnderscriptBox["\[Succeeds]", "D"], 
               RowBox[{"m", 
                UnderscriptBox["*", "D"], "y"}]}], 
              TagBox[")",
               "AutoParentheses"]}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              UnderscriptBox["\[Element]", "D"], 
              RowBox[{"[", 
               RowBox[{"c", ",", "n"}], "]"}]}]], 
            RowBox[{"c", "\[NotEqual]", 
             UnderscriptBox["0", "D"]}]], 
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
              RowBox[{
               UnderscriptBox["multN", "D"], "[", "c", "]"}]}]], 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{
              RowBox[{
               UnderscriptBox["mult", "D"], "[", 
               RowBox[{"n", ",", "c", ",", "i"}], "]"}], 
              "\[DoubleLongLeftRightArrow]", 
              RowBox[{
               UnderscriptBox["mult", "D"], "[", 
               RowBox[{
                RowBox[{"m", 
                 UnderscriptBox["*", "D"], "n"}], ",", "c", ",", "i"}], 
               "]"}]}], 
             TagBox[")",
              "AutoParentheses"]}]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "D"], 
             RowBox[{"[", 
              RowBox[{"c1", ",", "c2"}], "]"}]}]], 
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{"i", ",", 
              RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
              RowBox[{
               UnderscriptBox["multN", "D"], "[", "c1", "]"}]}]], 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{
              RowBox[{
               UnderscriptBox["isMNTCR", "D"], "[", 
               RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}],
               "\[DoubleLongLeftRightArrow]", 
              RowBox[{
               UnderscriptBox["isMNTCR", "D"], "[", 
               RowBox[{"c1", ",", "c2", ",", "b", ",", "i", ",", "j"}], 
               "]"}]}], 
             TagBox[")",
              "AutoParentheses"]}]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "D"], 
             RowBox[{"[", "c", "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              UnderscriptBox["isMNTCR", "D"], "[", 
              RowBox[{"c", ",", "a"}], "]"}], "\[DoubleLongLeftRightArrow]", 
             RowBox[{
              UnderscriptBox["isMNTCR", "D"], "[", 
              RowBox[{"c", ",", "b"}], "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Tilde] multiplier)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1818365967", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650337794245049*^9, 3.650337838554543*^9}, {
  3.65034201565197*^9, 3.6503420173525352`*^9}, {3.650344425847769*^9, 
  3.650344508370985*^9}, {3.650344797371449*^9, 3.650344822117527*^9}, {
  3.650345430413186*^9, 3.6503455469336157`*^9}, {3.650345588910882*^9, 
  3.6503456359575443`*^9}, {3.6503459992573566`*^9, 3.6503459997359457`*^9}, {
  3.650346200607349*^9, 3.6503462015307426`*^9}, {3.650346631442183*^9, 
  3.650346634768033*^9}, {3.650347123019497*^9, 3.650347170598927*^9}, {
  3.650353462761437*^9, 3.650353479390695*^9}, {3.650776228686902*^9, 
  3.650776269542432*^9}, {3.6507773550228767`*^9, 3.650777365824424*^9}, {
  3.6507774817273417`*^9, 3.650777482166712*^9}},
 CellTags->{
  "ID:1818365967", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde] multiplier"},
 CellID->1818365967],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Tilde]", "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", "m", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["isUnit", "D"], "[", "m", "]"}]},
         {
          RowBox[{"a", 
           UnderscriptBox[
            SubscriptBox["\[Tilde]", "m"], "D"], "b"}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1533920470", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650184664081078*^9, 3.650184702385171*^9}, {
  3.650337211577375*^9, 3.650337418108774*^9}, {3.6503374754578447`*^9, 
  3.6503374787345867`*^9}, {3.650337595819228*^9, 3.650337608821948*^9}, {
  3.650337823541972*^9, 3.6503378501595297`*^9}, {3.6503455647389307`*^9, 
  3.650345567361463*^9}, {3.6503459913257504`*^9, 3.650346004854034*^9}, {
  3.650346194842064*^9, 3.650346205778878*^9}, {3.650785306578586*^9, 
  3.650785310725092*^9}},
 CellTags->{
  "ID:1533920470", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde]"},
 CellID->1533920470],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->110581283]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1630010343],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Tilde]", "D"], TraditionalForm]]],
 " is an equivalence relation on ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ". It is needed for making fields satisfy axioms (R9) and (R10), since in \
infinite fields there are infinitely many minimal non-trivial common \
reducibles, meaning that there do not exist tuples containing all of them. \
With ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Tilde]", "D"], TraditionalForm]]],
 ", however, it is sufficient if only one single representative from each \
equivalence class is contained in the tuples \[Dash] and this is clearly \
possible, as there are actually only two equivalence classes.\nApparently, ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Tilde]", "D"], TraditionalForm]]],
 " is very strong; hence, it could be that fields are the only reduction \
rings where ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[Tilde]", "D"], TraditionalForm]]],
 " is non-trivial (i.e. where ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "b"}], TraditionalForm]]],
 " exist with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    UnderscriptBox["\[Tilde]", "D"], "b"}], TraditionalForm]]],
 "). Maybe it is possible to make it a bit weaker."
}], "Text",
 CellChangeTimes->{{3.6507933631882153`*^9, 3.650793704997489*^9}},
 CellID->1007665725],

Cell["", "OpenEnvironment",
 CellID->252379668],

Cell[CellGroupData[{

Cell["Definition: red and \[SquareSubset]", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6536255445576735`*^9, 3.653625551998886*^9}},
 CellID->82281139],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "D", "]"}]], 
   RowBox[{"[", 
    RowBox[{"A", ",", "B"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 81135181}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6536255583012977`*^9, 3.653625575632928*^9}},
 CellTags->{
  "ID:81135181", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->81135181],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["red", "D"], "[", "A", "]"}], ":=", 
  RowBox[{"{", 
   RowBox[{"x", 
    UnderscriptBox["|", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", "x", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{"a", "\[Element]", "A"}]], 
     RowBox[{
      UnderscriptBox["isReducible", "D"], "[", 
      RowBox[{"x", ",", "a"}], "]"}]}]}], "}"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1593192085", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.653625586989748*^9, 3.6536256300468235`*^9}},
 CellTags->{
  "ID:1593192085", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red"},
 CellID->1593192085],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"A", 
    UnderscriptBox["\[SquareSubset]", "D"], "B"}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["red", "D"], "[", "A", "]"}], "\[Subset]", 
    RowBox[{
     UnderscriptBox["red", "D"], "[", "B", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[SquareSubset])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:808948861", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648282286861656*^9, 3.6482823608012877`*^9}, {
  3.648282474960952*^9, 3.6482824768040733`*^9}, {3.653625687129324*^9, 
  3.653625694570537*^9}},
 CellTags->{
  "ID:808948861", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset]"},
 CellID->808948861],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "D", "]"}]], 
   RowBox[{"[", 
    RowBox[{"T", ",", "S"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 301449743}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.6536339862801733`*^9},
 CellTags->{
  "ID:301449743", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->301449743],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["red", "D"], "[", "T", "]"}], ":=", 
  RowBox[{
   UnderscriptBox["red", "D"], "[", 
   RowBox[{"TupleToSet", "[", "T", "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red tuples)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:218271861", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6536339988693953`*^9, 3.653634023455038*^9}},
 CellTags->{
  "ID:218271861", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red tuples"},
 CellID->218271861],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"T", 
    UnderscriptBox["\[SquareSubset]", "D"], "S"}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"TupleToSet", "[", "T", "]"}], 
    UnderscriptBox["\[SquareSubset]", "D"], 
    RowBox[{"TupleToSet", "[", "S", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[SquareSubset] tuples)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:20184186", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648282890553082*^9, 3.648282950231757*^9}, {
  3.6536315257086306`*^9, 3.653631531558641*^9}, {3.653633985016571*^9, 
  3.6536339898993797`*^9}},
 CellTags->{
  "ID:20184186", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] tuples"},
 CellID->20184186],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->391369155]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->586839649]
}, Closed]],

Cell[CellGroupData[{

Cell["Lemmas: Arbitrary Domains (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6479429108021584`*^9, 3.647942911699275*^9}, {
  3.64794294408132*^9, 3.647942944340314*^9}, {3.647943039642148*^9, 
  3.647943059496397*^9}, {3.648180720478798*^9, 3.6481807241091957`*^9}},
 CellID->399888057],

Cell[TextData[{
 "In this subsection we state some simple lemmas about the notions introduced \
above, which are completely independent of the algebraic structure of the \
underlying domain ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.6451792068281116`*^9, 3.645179251196622*^9}},
 CellID->815818515],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "D"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 2086101595}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645179191172154*^9, 3.64517919219974*^9}},
 CellTags->{
  "ID:2086101595", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2086101595],

Cell[CellGroupData[{

Cell["\<\
Reduction Modulo Sets and Tuples (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.645178346006187*^9, 3.645178370967299*^9}, {
  3.645180142557076*^9, 3.645180152982059*^9}, {3.647943012595858*^9, 
  3.647943029465214*^9}, {3.648180712996744*^9, 3.648180715538513*^9}},
 CellID->1774554504],

Cell["", "OpenEnvironment",
 CellID->30312876],

Cell[CellGroupData[{

Cell["Lemma: Introduction Rules for isReductor", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6620922607826633`*^9, 3.6620922676622753`*^9}},
 CellID->26444589],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "D", "]"}]], 
    RowBox[{"[", "A", "]"}]}]], 
  RowBox[{
   UnderscriptBox["isReductor", "D"], "[", "A", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DomainSets isReductor)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:690041737", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.662092272108283*^9, 3.6620922910019164`*^9}},
 CellTags->{
  "ID:690041737", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets isReductor"},
 CellID->690041737],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "T", "]"}]}]], 
  RowBox[{
   UnderscriptBox["isReductor", "D"], "[", "T", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DomainTuples isReductor)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:5128300", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.662092311393152*^9, 3.662092320815569*^9}},
 CellTags->{
  "ID:5128300", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainTuples isReductor"},
 CellID->5128300],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->21022560]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->301523614],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DomainSets isReductor)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DomainSets isReductor)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["isReductor", 
            StyleBox["D", "ExpressionVariable"]], "[", 
           StyleBox["A", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["isReductor", 
            StyleBox["D", "ExpressionVariable"]], "[", 
           StyleBox["A", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:690041737"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p690041737-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p690041737-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:690041737", "Proof|ID:690041737-1"},
 CellID->323627089],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p690041737-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->95218679]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DomainTuples isReductor)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DomainTuples isReductor)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["isReductor", 
            StyleBox["D", "ExpressionVariable"]], "[", 
           StyleBox["T", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["isReductor", 
            StyleBox["D", "ExpressionVariable"]], "[", 
           StyleBox["T", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:5128300"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p5128300-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p5128300-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:5128300", "Proof|ID:5128300-1"},
 CellID->485768995],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p5128300-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->590295287]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->270038199],

Cell[CellGroupData[{

Cell["Lemma: Reduction Modulo Singletons", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6479427018476*^9, 3.6479427096111383`*^9}, {
  3.6620926738721905`*^9, 3.662092674340191*^9}},
 CellID->917364838],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"a", ",", "b"}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", "c", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       RowBox[{"a", 
        UnderscriptBox[
         SubscriptBox["\[Rule]", 
          RowBox[{"{", "c", "}"}]], "D"], "b"}],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}], 
     TagBox["\[DoubleLongLeftRightArrow]",
      Identity,
      SyntaxForm->"a\[Implies]b"], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "D"], 
         RowBox[{"[", "m", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", 
           RowBox[{"m", ",", "c"}]], "D"], "b"}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduction modulo singleton)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:906005529", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.592024218919891*^9, {3.592024334485207*^9, 3.592024340289249*^9}, {
   3.592024431355556*^9, 3.592024441537606*^9}, {3.592024487790746*^9, 
   3.592024488655926*^9}, {3.592024518814897*^9, 3.592024562558631*^9}, {
   3.592025233836319*^9, 3.592025234611331*^9}, 3.594035724790234*^9, {
   3.596187341477315*^9, 3.596187370215183*^9}, {3.596187467659761*^9, 
   3.596187475385688*^9}, {3.607337829755179*^9, 3.607337845017223*^9}, 
   3.608525687594316*^9, {3.608525870819246*^9, 3.608525875872862*^9}, {
   3.608527198110731*^9, 3.6085272122117*^9}, {3.609067948323856*^9, 
   3.609067949754394*^9}, 3.610170217489511*^9, {3.6455006197534018`*^9, 
   3.645500631522522*^9}, {3.645500739541072*^9, 3.64550074064289*^9}, {
   3.6618493794734097`*^9, 3.661849389660228*^9}, 3.661849488252401*^9, {
   3.661849612662619*^9, 3.6618496153458242`*^9}, {3.661849876816284*^9, 
   3.6618498926503115`*^9}},
 CellTags->{
  "ID:906005529", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo singleton", "IR"},
 CellID->906005529],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1135647683]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1526698344],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduction modulo singleton)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduction modulo singleton)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:906005529"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p906005529-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p906005529-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:906005529", "Proof|ID:906005529-1"},
 CellID->520373627],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p906005529-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->177234516]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->701877035],

Cell[CellGroupData[{

Cell["Lemma: Reduction modulo Sets and Tuples", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.647932406186142*^9, 3.647932413433289*^9}},
 CellID->1394100531],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"a", ",", "b"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 708647877}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6479324399095*^9, 3.647932451390093*^9}, {
  3.647933398983199*^9, 3.647933399366269*^9}},
 CellTags->{
  "ID:708647877", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->708647877],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "D", "]"}]], 
   RowBox[{"[", "T", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 508075012}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647932422897945*^9, 3.647932434684854*^9}},
 CellTags->{
  "ID:508075012", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->508075012],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", "T"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"{", 
         SubscriptBox["T", "i"], "}"}]], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduction modulo tuple alternative)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:347299573", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6620980443528748`*^9, 3.6620980870259495`*^9}},
 CellTags->{
  "ID:347299573", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo tuple alternative", "IR"},
 CellID->347299573],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"S", "=", 
   RowBox[{"TupleToSet", "[", "T", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1756538453}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647932470772533*^9, 3.647932477540744*^9}},
 CellTags->{
  "ID:1756538453", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1756538453],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", "T"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", "S"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow] TupleToSet)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:200776309", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6479324558565273`*^9, 3.647932489560149*^9}},
 CellTags->{
  "ID:200776309", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] TupleToSet", "IR"},
 CellID->200776309],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", "T"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", "S"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LeftRightArrow] TupleToSet)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:989570066", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647932503709743*^9, 3.647932510707254*^9}},
 CellTags->{
  "ID:989570066", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] TupleToSet", "IR"},
 CellID->989570066],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "T", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "S", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow]* TupleToSet)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:359412397", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647932543899979*^9, 3.647932566853138*^9}},
 CellTags->{
  "ID:359412397", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow]* TupleToSet", "IR"},
 CellID->359412397],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "T", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "S", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LeftRightArrow]* TupleToSet)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1876144273", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647932575439321*^9, 3.647932580594158*^9}},
 CellTags->{
  "ID:1876144273", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow]* TupleToSet", "IR"},
 CellID->1876144273],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "z"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      RowBox[{"a", 
       UnderscriptBox[
        OverscriptBox[
         SubsuperscriptBox["\[LeftRightArrow]", "T", "*"], 
         RowBox[{"z", ",", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           TagBox["\[Succeeds]",
            Identity], 
           TagBox[")",
            "AutoParentheses"]}]}]], "D"], "b"}],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      RowBox[{"a", 
       UnderscriptBox[
        OverscriptBox[
         SubsuperscriptBox["\[LeftRightArrow]", "S", "*"], 
         RowBox[{"z", ",", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           TagBox["\[Succeeds]",
            Identity], 
           TagBox[")",
            "AutoParentheses"]}]}]], "D"], "b"}],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(connectibility below TupleToSet)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:495509420", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64793397742791*^9, 3.647933989654293*^9}},
 CellTags->{
  "ID:495509420", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below TupleToSet", "IR"},
 CellID->495509420],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1713169042]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->434319633],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduction modulo tuple alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduction modulo tuple alternative)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:347299573"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p347299573-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p347299573-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:347299573", "Proof|ID:347299573-1"},
 CellID->641607246],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p347299573-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->129731929]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] TupleToSet)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] TupleToSet)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"TupleToSet", "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"TupleToSet", "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:200776309"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p200776309-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p200776309-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:200776309", "Proof|ID:200776309-1"},
 CellID->376820956],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p200776309-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->46818332]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LeftRightArrow] TupleToSet)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LeftRightArrow] TupleToSet)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 RowBox[{"TupleToSet", "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 RowBox[{"TupleToSet", "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:989570066"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p989570066-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p989570066-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:989570066", "Proof|ID:989570066-1"},
 CellID->450749896],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p989570066-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1131081333]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow]* TupleToSet)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow]* TupleToSet)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                RowBox[{"TupleToSet", "[", 
                  StyleBox["T", "ExpressionVariable"], "]"}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                RowBox[{"TupleToSet", "[", 
                  StyleBox["T", "ExpressionVariable"], "]"}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:359412397"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p359412397-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p359412397-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:359412397", "Proof|ID:359412397-1"},
 CellID->1501514132],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p359412397-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1241556193]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LeftRightArrow]* TupleToSet)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LeftRightArrow]* TupleToSet)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                RowBox[{"TupleToSet", "[", 
                  StyleBox["T", "ExpressionVariable"], "]"}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                RowBox[{"TupleToSet", "[", 
                  StyleBox["T", "ExpressionVariable"], "]"}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1876144273"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1876144273-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1876144273-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1876144273", "Proof|ID:1876144273-1"},
 CellID->541596138],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1876144273-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1335127868]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(connectibility below TupleToSet)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(connectibility below TupleToSet)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             StyleBox["z", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{"TupleToSet", "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             StyleBox["z", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{"TupleToSet", "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:495509420"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p495509420-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p495509420-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:495509420", "Proof|ID:495509420-1"},
 CellID->551004923],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p495509420-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->20568225]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->770283335],

Cell[CellGroupData[{

Cell["Lemma: Reduction Modulo Sets and Enlarged Sets", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.647936877717114*^9, 3.647936893580111*^9}, {
  3.647937017730925*^9, 3.647937018333579*^9}, {3.6479371386298733`*^9, 
  3.647937142335061*^9}, {3.648891852020252*^9, 3.648891854426759*^9}, {
  3.6621836422902594`*^9, 3.6621836427582607`*^9}, {3.6622018724232535`*^9, 
  3.6622018740144567`*^9}},
 CellID->1406215010],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"a", ",", "b"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 467637718}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64793690323139*^9, 3.647936906095187*^9}},
 CellTags->{
  "ID:467637718", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->467637718],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "D", "]"}]], 
   RowBox[{"[", 
    RowBox[{"B", ",", "C"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1182167952}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647936964029808*^9, 3.6479369696414547`*^9}},
 CellTags->{
  "ID:1182167952", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1182167952],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", "B"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{"c", "\[Element]", "B"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"{", "c", "}"}]], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduction modulo set alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:178557746", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.662183492670397*^9, 3.662183515571237*^9}},
 CellTags->{
  "ID:178557746", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo set alternative", "IR"},
 CellID->178557746],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", "B"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"B", "\[Union]", "C"}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow] \[Union])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1019233204", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647936911678726*^9, 3.647936998514596*^9}},
 CellTags->{
  "ID:1019233204", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[Union]"},
 CellID->1019233204],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", "B"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", 
      RowBox[{"B", "\[Union]", "C"}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LeftRightArrow] \[Union])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:331036245", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647937029582994*^9, 3.647937035898677*^9}},
 CellTags->{
  "ID:331036245", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] \[Union]"},
 CellID->331036245],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "B", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", 
      RowBox[{"B", "\[Union]", "C"}], "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow]* \[Union])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:78564808", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647937029582994*^9, 3.647937064172783*^9}},
 CellTags->{
  "ID:78564808", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow]* \[Union]"},
 CellID->78564808],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", 
      RowBox[{"B", "\[Union]", "C"}], "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LeftRightArrow]* \[Union])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:626278069", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647937029582994*^9, 3.647937035898677*^9}, {
  3.647937067931065*^9, 3.64793707165944*^9}},
 CellTags->{
  "ID:626278069", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow]* \[Union]"},
 CellID->626278069],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "z"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       OverscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
        RowBox[{"z", ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox["\[Succeeds]",
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}]], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       OverscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", 
         RowBox[{"B", "\[Union]", "C"}], "*"], 
        RowBox[{"z", ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox["\[Succeeds]",
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}]], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(connectibility below \[Union])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:356849840", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647937029582994*^9, 3.647937035898677*^9}, {
  3.6479370888868227`*^9, 3.647937128320746*^9}},
 CellTags->{
  "ID:356849840", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below \[Union]"},
 CellID->356849840],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1824911811]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->933018018],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduction modulo set alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduction modulo set alternative)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[Element]", 
                   StyleBox["B", "ExpressionVariable"]}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[Element]", 
                   StyleBox["B", "ExpressionVariable"]}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:178557746"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p178557746-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p178557746-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:178557746", "Proof|ID:178557746-1"},
 CellID->93489966],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p178557746-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->261267025]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] \[Union])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] \[Union])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\[Union]", 
                 StyleBox["C", "ExpressionVariable"]}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\[Union]", 
                 StyleBox["C", "ExpressionVariable"]}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1019233204"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1019233204-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1019233204-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1019233204", "Proof|ID:1019233204-1"},
 CellID->9184570],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1019233204-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->310726113]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LeftRightArrow] \[Union])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LeftRightArrow] \[Union])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\[Union]", 
                 StyleBox["C", "ExpressionVariable"]}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\[Union]", 
                 StyleBox["C", "ExpressionVariable"]}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:331036245"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p331036245-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p331036245-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:331036245", "Proof|ID:331036245-1"},
 CellID->970023700],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p331036245-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->221863878]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow]* \[Union])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow]* \[Union])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                RowBox[{
                  StyleBox["B", "ExpressionVariable"], "\[Union]", 
                  StyleBox["C", "ExpressionVariable"]}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                RowBox[{
                  StyleBox["B", "ExpressionVariable"], "\[Union]", 
                  StyleBox["C", "ExpressionVariable"]}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:78564808"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p78564808-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p78564808-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:78564808", "Proof|ID:78564808-1"},
 CellID->925666063],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p78564808-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->477674065]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LeftRightArrow]* \[Union])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LeftRightArrow]* \[Union])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                RowBox[{
                  StyleBox["B", "ExpressionVariable"], "\[Union]", 
                  StyleBox["C", "ExpressionVariable"]}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                RowBox[{
                  StyleBox["B", "ExpressionVariable"], "\[Union]", 
                  StyleBox["C", "ExpressionVariable"]}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:626278069"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p626278069-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p626278069-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:626278069", "Proof|ID:626278069-1"},
 CellID->1953121485],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p626278069-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->785566982]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(connectibility below \[Union])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(connectibility below \[Union])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             StyleBox["z", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\[Union]", 
                   StyleBox["C", "ExpressionVariable"]}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             StyleBox["z", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\[Union]", 
                   StyleBox["C", "ExpressionVariable"]}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:356849840"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p356849840-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p356849840-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:356849840", "Proof|ID:356849840-1"},
 CellID->205255475],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p356849840-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->979272693]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Other Notions (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6479429766424503`*^9, 3.647942983341456*^9}},
 CellID->828082569],

Cell["", "OpenEnvironment",
 CellID->1774903907],

Cell[CellGroupData[{

Cell["\<\
Lemma: isCommonReducible, isNTCR, isMNTCR and areCPMultipliers of Different \
Elements\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.645178362130967*^9, {3.6466486360710506`*^9, 3.6466486523775387`*^9}, {
   3.648792331000181*^9, 3.648792339241612*^9}},
 CellID->1159810905],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2", ",", "a"}], "]"}]}]], 
  RowBox[{"c1", "\[NotEqual]", "c2"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 366153355}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645177582571649*^9, 3.645177629288046*^9}, {
   3.6451782372411003`*^9, 3.645178251973439*^9}, {3.6451786969811993`*^9, 
   3.645178697165472*^9}, 3.645179199651157*^9},
 CellTags->{
  "ID:366153355", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->366153355],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", "j"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 221340803}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645177635699017*^9, 3.645177649384877*^9}, 
   3.648800372509159*^9},
 CellTags->{
  "ID:221340803", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->221340803],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isCommonReducible", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isCommonReducible", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "1", ",", "1"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:328195853", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6451782883102503`*^9, 3.645178298631407*^9}, {
  3.645178379528577*^9, 3.6451784705239983`*^9}, {3.645178510206888*^9, 
  3.645178512868581*^9}, {3.6451789724639473`*^9, 3.6451789812450457`*^9}, {
  3.6451790406390553`*^9, 3.645179067104841*^9}},
 CellTags->{
  "ID:328195853", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "2"},
 CellID->328195853],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isNTCR", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "1", ",", "1"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1692465705", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645178564332363*^9, 3.645178570768199*^9}},
 CellTags->{
  "ID:1692465705", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "3"},
 CellID->1692465705],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isMNTCR", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "1", ",", "1"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:188743948", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645178587019286*^9, 3.645178591187087*^9}},
 CellTags->{
  "ID:188743948", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "4"},
 CellID->188743948],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["areCPMultipliers", "D"], "[", 
     RowBox[{
     "m1", ",", "m2", ",", "c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], 
     "]"}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox["areCPMultipliers", "D"], "[", 
     RowBox[{
     "m1", ",", "m2", ",", "c1", ",", "c2", ",", "a", ",", "1", ",", "1"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:115635911", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645178479900733*^9, 3.645178551234201*^9}},
 CellTags->{
  "ID:115635911", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "5"},
 CellID->115635911],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isCommonReducible", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["isCommonReducible", "D"], "[", 
   RowBox[{"c2", ",", "c1", ",", "a", ",", "i", ",", "j"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCommonReducible symmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1118510950", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648799503852746*^9, 3.648799514028779*^9}},
 CellTags->{
  "ID:1118510950", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isCommonReducible symmetric"},
 CellID->1118510950],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["isNTCR", "D"], "[", 
   RowBox[{"c2", ",", "c1", ",", "a", ",", "i", ",", "j"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isNTCR symmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:113485898", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648799559631874*^9, 3.648799565546281*^9}},
 CellTags->{
  "ID:113485898", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isNTCR symmetric"},
 CellID->113485898],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["isMNTCR", "D"], "[", 
   RowBox[{"c2", ",", "c1", ",", "a", ",", "i", ",", "j"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMNTCR symmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2056330192", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6487996212334957`*^9, 3.648799626058054*^9}},
 CellTags->{
  "ID:2056330192", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isMNTCR symmetric"},
 CellID->2056330192],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["areCPMultipliers", "D"], "[", 
     RowBox[{
     "m1", ",", "m2", ",", "c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], 
     "]"}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["areCPMultipliers", "D"], "[", 
     RowBox[{
     "m2", ",", "m1", ",", "c2", ",", "c1", ",", "a", ",", "i", ",", "j"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areCPMultipliers symmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:463339584", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648799640757247*^9, 3.648799683922832*^9}, {
  3.648801512366592*^9, 3.648801513546166*^9}},
 CellTags->{
  "ID:463339584", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers symmetric"},
 CellID->463339584],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
  "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", 
       RowBox[{"m1", ",", "m2"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["isReducible", "D"], "[", 
         RowBox[{"a", ",", "c1", ",", "m1"}], "]"}], "\[And]", 
        RowBox[{
         UnderscriptBox["isReducible", "D"], "[", 
         RowBox[{"a", ",", "c2", ",", "m2"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{
       UnderscriptBox["areCPMultipliers", "D"], "[", 
       RowBox[{
       "m1", ",", "m2", ",", "c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], 
       "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areCPMultipliers sufficient)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:449270875", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649565558314979*^9, 3.6495656347094393`*^9}},
 CellTags->{
  "ID:449270875", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers sufficient"},
 CellID->449270875],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->838304061]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->44286942],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isCommonReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isCommonReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isCommonReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isCommonReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:328195853"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p328195853-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p328195853-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:328195853", "Proof|ID:328195853-1"},
 CellID->667808408],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p328195853-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->342092220]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1692465705"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1692465705-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1692465705-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1692465705", "Proof|ID:1692465705-1"},
 CellID->982819604],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1692465705-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1849675645]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:188743948"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p188743948-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p188743948-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:188743948", "Proof|ID:188743948-1"},
 CellID->1789453742],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p188743948-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->254539763]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["areCPMultipliers", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["areCPMultipliers", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["areCPMultipliers", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["areCPMultipliers", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:115635911"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p115635911-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p115635911-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:115635911", "Proof|ID:115635911-1"},
 CellID->1648327750],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p115635911-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->317536928]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCommonReducible symmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCommonReducible symmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isCommonReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isCommonReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1118510950"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1118510950-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1118510950-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1118510950", "Proof|ID:1118510950-1"},
 CellID->144468455],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1118510950-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1943003684]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isNTCR symmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isNTCR symmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:113485898"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p113485898-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p113485898-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:113485898", "Proof|ID:113485898-1"},
 CellID->1668243539],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p113485898-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1751898811]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMNTCR symmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMNTCR symmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2056330192"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2056330192-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2056330192-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2056330192", "Proof|ID:2056330192-1"},
 CellID->1822040650],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2056330192-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1848124281]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(areCPMultipliers symmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(areCPMultipliers symmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCPMultipliers", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCPMultipliers", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:463339584"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p463339584-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p463339584-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:463339584", "Proof|ID:463339584-1"},
 CellID->2055115730],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p463339584-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1154563333]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(areCPMultipliers sufficient)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(areCPMultipliers sufficient)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], "]"}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["areCPMultipliers", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], "]"}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["areCPMultipliers", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:449270875"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p449270875-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p449270875-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:449270875", "Proof|ID:449270875-1"},
 CellID->1693763802],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p449270875-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->874031660]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->431320600],

Cell[CellGroupData[{

Cell["\<\
Lemma: isCommonReducible, isNTCR, isMNTCR and areCPMultipliers of Identical \
Elements\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648792349665941*^9, 3.648792354225885*^9}},
 CellID->1620999719],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"c", ",", "a"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 523863903}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648792359171913*^9, 3.648792390229143*^9}},
 CellTags->{
  "ID:523863903", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->523863903],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
    UnderscriptBox["multN", "D"], "[", "c", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1519187240}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648792367756867*^9, 3.648792378720983*^9}},
 CellTags->{
  "ID:1519187240", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1519187240],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isCommonReducible", "D"], "[", 
   RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isCommonReducible", "D"], "[", 
   RowBox[{"c", ",", "c", ",", "a", ",", "j", ",", "i"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(2')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1878546324", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6487923852482157`*^9, 3.648792397937511*^9}, {
  3.6487928822159367`*^9, 3.6487928830663958`*^9}},
 CellTags->{
  "ID:1878546324", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "2'"},
 CellID->1878546324],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "D"], "[", 
   RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isNTCR", "D"], "[", 
   RowBox[{"c", ",", "c", ",", "a", ",", "j", ",", "i"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(3')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1781176225", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64879240655794*^9, 3.648792413427478*^9}, {
  3.6487928848748302`*^9, 3.648792885800911*^9}},
 CellTags->{
  "ID:1781176225", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "3'"},
 CellID->1781176225],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "D"], "[", 
   RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isMNTCR", "D"], "[", 
   RowBox[{"c", ",", "c", ",", "a", ",", "j", ",", "i"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(4')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1582995786", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6487924207182617`*^9, 3.648792427800535*^9}, {
  3.6487928875357924`*^9, 3.6487928884606752`*^9}},
 CellTags->{
  "ID:1582995786", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "4'"},
 CellID->1582995786],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["areCPMultipliers", "D"], "[", 
     RowBox[{
     "m1", ",", "m2", ",", "c", ",", "c", ",", "a", ",", "i", ",", "j"}], 
     "]"}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["areCPMultipliers", "D"], "[", 
     RowBox[{
     "m2", ",", "m1", ",", "c", ",", "c", ",", "a", ",", "j", ",", "i"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(5')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2020067658", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64879243576196*^9, 3.648792447892214*^9}, {
  3.648792890050766*^9, 3.648792891020504*^9}},
 CellTags->{
  "ID:2020067658", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "5'"},
 CellID->2020067658],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->81893628]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1310599982],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(2')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(2')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1878546324"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1878546324-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1878546324-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1878546324", "Proof|ID:1878546324-1"},
 CellID->1762740644],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1878546324-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1512786933]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(3')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(3')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1781176225"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1781176225-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1781176225-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1781176225", "Proof|ID:1781176225-1"},
 CellID->1337859754],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1781176225-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->948047327]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(4')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(4')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1582995786"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1582995786-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1582995786-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1582995786", "Proof|ID:1582995786-1"},
 CellID->1038215936],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1582995786-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->41089318]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(5')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(5')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2020067658"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2020067658-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2020067658-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2020067658", "Proof|ID:2020067658-1"},
 CellID->1751126895],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2020067658-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1401855599]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1693676355],

Cell[CellGroupData[{

Cell["Lemma: isMNTCR, areIrrelative, areCorrelative", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.645184383800828*^9, {3.6466486577985277`*^9, 3.64664867102398*^9}, {
   3.6481840700024137`*^9, 3.648184072502778*^9}, {3.648792469873967*^9, 
   3.648792470390827*^9}},
 CellID->320285829],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "c"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1823284442}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.645246446308467*^9},
 CellTags->{
  "ID:1823284442", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1823284442],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
    UnderscriptBox["multN", "D"], "[", "c", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 284217868}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.645246454966563*^9},
 CellTags->{
  "ID:284217868", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->284217868],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "D"], "[", 
   RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], "\[Implies]", 
  RowBox[{"i", "\[NotEqual]", "j"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:628244416", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64518438694545*^9, 3.645184429886971*^9}, {
  3.6451844740983458`*^9, 3.645184516969061*^9}, {3.645184587550663*^9, 
  3.6451846066532288`*^9}, {3.645246445023808*^9, 3.645246458762854*^9}},
 CellTags->{
  "ID:628244416", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "6"},
 CellID->628244416],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2", ",", "m3", ",", "m4"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["areIrrelative", "D"], "[", 
       RowBox[{"c", ",", "c", ",", "m1", ",", "m2", ",", "i", ",", "j"}], 
       "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["areIrrelative", "D"], "[", 
       RowBox[{"c", ",", "c", ",", "m3", ",", "m4", ",", "i", ",", "j"}], 
       "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["areCorrelative", "D"], "[", 
       RowBox[{"c", ",", "m1", ",", "m3"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["areCorrelative", "D"], "[", 
       RowBox[{"c", ",", "m2", ",", "m4"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(correlative-irrelative 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1255486983", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645246592558919*^9, 3.645246663110716*^9}, {
  3.6452467139015083`*^9, 3.64524671888614*^9}, {3.645247345752684*^9, 
  3.645247357885337*^9}, {3.645247432880946*^9, 3.645247433170784*^9}, {
  3.6452474684154167`*^9, 3.645247469833856*^9}},
 CellTags->{
  "ID:1255486983", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative-irrelative 2"},
 CellID->1255486983],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->864502048]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2028899293],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(6)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(6)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox[
             "j", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMNTCR", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox[
             "j", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:628244416"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p628244416-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p628244416-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:628244416", "Proof|ID:628244416-1"},
 CellID->1993474493],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p628244416-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->661346101]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(correlative-irrelative 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(correlative-irrelative 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m3", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m4", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["areIrrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["areIrrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m3", "ExpressionVariable"], ",", 
                 StyleBox["m4", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["areCorrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m3", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["m4", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m3", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m4", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["areIrrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["areIrrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m3", "ExpressionVariable"], ",", 
                 StyleBox["m4", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["areCorrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m3", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["m4", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1255486983"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1255486983-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1255486983-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1255486983", "Proof|ID:1255486983-1"},
 CellID->558484412],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1255486983-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1567652911]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Lemmas: Commutative Rings with Unit (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.632042719660898*^9, 3.632042720525034*^9}, {
   3.641183473886408*^9, 3.641183477717979*^9}, {3.6451770358897448`*^9, 
   3.6451770361787863`*^9}, 3.645177476545224*^9, 3.6451776557872257`*^9, {
   3.645179166917697*^9, 3.645179167288909*^9}, {3.6479430498286457`*^9, 
   3.6479430723029623`*^9}},
 CellID->359523023],

Cell[TextData[{
 "In this subsection we state some simple lemmas about the notions introduced \
above, which require the underlying domain ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " to be at least a commutative ring with unit."
}], "Text",
 CellChangeTimes->{{3.645177433595942*^9, 3.6451774698246737`*^9}},
 CellID->1577638100],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isCommRing1", "[", "D", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1382908475}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632042806042433*^9, 3.632042807304936*^9}, {
  3.638159349700302*^9, 3.638159354780015*^9}},
 CellTags->{
  "ID:1382908475", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1382908475],

Cell["", "OpenEnvironment",
 CellID->955233445],

Cell[CellGroupData[{

Cell["Lemma: reduceM, isReducible, symred", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632042729456352*^9, 3.632042729643711*^9}, {
  3.632042761312243*^9, 3.63204277425531*^9}, {3.6320432482017183`*^9, 
  3.632043249896613*^9}},
 CellID->23280017],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "c"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 442895874}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632042817664831*^9, 3.6320428326393347`*^9}, {
  3.638088567824203*^9, 3.6380885699126387`*^9}, {3.638088679286591*^9, 
  3.638088686944771*^9}},
 CellTags->{
  "ID:442895874", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->442895874],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "D", "]"}]], 
   RowBox[{"[", "T", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 927154658}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632042844230549*^9, 3.6320428566942*^9}, {
  3.635748720496004*^9, 3.6357487272793627`*^9}},
 CellTags->{
  "ID:927154658", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->927154658],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["reduceM", "D"], "[", 
   RowBox[{"a", ",", "c", ",", "T", ",", "0"}], "]"}], "\[Equal]", 
  "a"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduceM 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1300106055", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641638125156376*^9, 3.6416381402328176`*^9}},
 CellTags->{
  "ID:1300106055", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM 0"},
 CellID->1300106055],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
   "\[NotEqual]", "0"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", "1"}], "]"}], "\[Equal]", 
    RowBox[{"a", 
     UnderscriptBox["-", "D"], 
     RowBox[{
      SubscriptBox["T", "1"], 
      UnderscriptBox["*", "D"], "c"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduceM 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:438711618", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641810041636121*^9, 3.641810073720456*^9}},
 CellTags->{
  "ID:438711618", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM 1"},
 CellID->438711618],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"k", "=", "0"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", "k"}], "]"}], "\[Equal]", 
    RowBox[{"a", 
     UnderscriptBox["-", "D"], 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        UnderscriptBox[
         SubscriptBox["\[Sum]", "D"], 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
        SubscriptBox["T", "i"]}], ")"}], 
      UnderscriptBox["*", "D"], "c"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduceM alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2101232329", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320428762385397`*^9, 3.6320429518811197`*^9}, 
   3.632043061786442*^9},
 CellTags->{
  "ID:2101232329", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM alternative"},
 CellID->2101232329],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], "-", 
     "1"}]}]], 
  RowBox[{
   RowBox[{
    UnderscriptBox["reduceM", "D"], "[", 
    RowBox[{"a", ",", "c", ",", "T", ",", 
     RowBox[{"k", "+", "1"}]}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", "k"}], "]"}], 
    UnderscriptBox["-", "D"], 
    RowBox[{
     SubscriptBox["T", 
      RowBox[{"k", "+", "1"}]], 
     UnderscriptBox["*", "D"], "c"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduceM+1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:96556784", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641625407968004*^9, 3.641625436959262*^9}, {
   3.641626153105338*^9, 3.641626153474964*^9}, 3.641626978612542*^9, {
   3.641627571552986*^9, 3.6416275718856363`*^9}},
 CellTags->{
  "ID:96556784", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM+1"},
 CellID->96556784],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"k", "=", "0"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["reduceM", "D"], "[", 
      RowBox[{
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{"m", 
         UnderscriptBox["*", "D"], "c"}]}], ",", "c", ",", "T", ",", "k"}], 
      "]"}], "\[Equal]", 
     RowBox[{"a", 
      UnderscriptBox["-", "D"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"m", 
         UnderscriptBox["+", "D"], 
         RowBox[{
          UnderscriptBox[
           SubscriptBox["\[Sum]", "D"], 
           RowBox[{
            RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
          SubscriptBox["T", "i"]}]}], ")"}], 
       UnderscriptBox["*", "D"], "c"}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduceM alternative II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1465630278", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632051716646419*^9, 3.63205174665989*^9}},
 CellTags->{
  "ID:1465630278", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM alternative II"},
 CellID->1465630278],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "0"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}], 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership reduceM)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1952076718", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641630673120125*^9, 3.641630733106476*^9}, 
   3.644908397601529*^9},
 CellTags->{
  "ID:1952076718", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "membership reduceM", "IR"},
 CellID->1952076718],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "T", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1110504237}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638597221753374*^9, 3.638597228993807*^9}},
 CellTags->{
  "ID:1110504237", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1110504237],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["reduceM", "D"], "[", 
   RowBox[{"a", ",", "c", ",", "T", ",", 
    RowBox[{"k", "-", "1"}]}], "]"}], "\[Equal]", 
  RowBox[{"a", 
   UnderscriptBox["-", "D"], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      UnderscriptBox[
       SubscriptBox["\[Sum]", "D"], 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"k", "-", "1"}]}]], 
      SubscriptBox["T", "i"]}], ")"}], 
    UnderscriptBox["*", "D"], "c"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduceM-1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1236615595", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641182258211178*^9, 3.6411822729804087`*^9}},
 CellTags->{
  "ID:1236615595", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM-1"},
 CellID->1236615595],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["reduceM", "D"], "[", 
    RowBox[{"a", ",", "c", ",", "T", ",", 
     RowBox[{"k", "-", "1"}]}], "]"}], 
   UnderscriptBox["-", "D"], 
   RowBox[{
    SubscriptBox["T", "k"], 
    UnderscriptBox["*", "D"], "c"}]}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["reduceM", "D"], "[", 
   RowBox[{"a", ",", "c", ",", "T", ",", "k"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduceM-1 II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:203149419", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6411246968591337`*^9, 3.6411247174175463`*^9}},
 CellTags->{
  "ID:203149419", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM-1 II"},
 CellID->203149419],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{"k", "-", "1"}]}], "]"}], ",", "c", ",", 
    SubscriptBox["T", "k"]}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "D"], "[", 
     RowBox[{
      SubscriptBox["T", "k"], ",", "c"}], "]"}], "\[And]", 
    RowBox[{
     RowBox[{"a", 
      UnderscriptBox["-", "D"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         UnderscriptBox[
          SubscriptBox["\[Sum]", "D"], 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"k", "-", "1"}]}]], 
         SubscriptBox["T", "i"]}], ")"}], 
       UnderscriptBox["*", "D"], "c"}]}], 
     UnderscriptBox["\[Succeeds]", "D"], 
     RowBox[{"a", 
      UnderscriptBox["-", "D"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         UnderscriptBox[
          SubscriptBox["\[Sum]", "D"], 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
         SubscriptBox["T", "i"]}], ")"}], 
       UnderscriptBox["*", "D"], "c"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible reduceM)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:333495152", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632042960285679*^9, 3.632043093907946*^9}, {
  3.638597233171023*^9, 3.638597239897953*^9}},
 CellTags->{
  "ID:333495152", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible reduceM"},
 CellID->333495152],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["symred", "D"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{"k", "-", "1"}]}], "]"}], ",", 
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", "k"}], "]"}], ",", "c", ",", 
    SubscriptBox["T", "k"]}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "D"], "[", 
     RowBox[{
      SubscriptBox["T", "k"], ",", "c"}], "]"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{
        UnderscriptBox["reduceM", "D"], "[", 
        RowBox[{"a", ",", "c", ",", "T", ",", 
         RowBox[{"k", "-", "1"}]}], "]"}], 
       UnderscriptBox["\[Succeeds]", "D"], 
       RowBox[{
        UnderscriptBox["reduceM", "D"], "[", 
        RowBox[{"a", ",", "c", ",", "T", ",", "k"}], "]"}]}], "\[Or]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["reduceM", "D"], "[", 
        RowBox[{"a", ",", "c", ",", "T", ",", "k"}], "]"}], 
       UnderscriptBox["\[Succeeds]", "D"], 
       RowBox[{
        UnderscriptBox["reduceM", "D"], "[", 
        RowBox[{"a", ",", "c", ",", "T", ",", 
         RowBox[{"k", "-", "1"}]}], "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symred reduceM)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:332101094", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641182177282013*^9, 3.641182207411715*^9}},
 CellTags->{
  "ID:332101094", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM"},
 CellID->332101094],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["symred", "D"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{"k", "-", "1"}]}], "]"}], ",", 
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", "k"}], "]"}], ",", "c", ",", 
    SubscriptBox["T", "k"]}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "D"], "[", 
     RowBox[{
      SubscriptBox["T", "k"], ",", "c"}], "]"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           UnderscriptBox[
            SubscriptBox["\[Sum]", "D"], 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{"k", "-", "1"}]}]], 
           SubscriptBox["T", "i"]}], ")"}], 
         UnderscriptBox["*", "D"], "c"}]}], 
       UnderscriptBox["\[Succeeds]", "D"], 
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           UnderscriptBox[
            SubscriptBox["\[Sum]", "D"], 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
           SubscriptBox["T", "i"]}], ")"}], 
         UnderscriptBox["*", "D"], "c"}]}]}], "\[Or]", 
      RowBox[{
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           UnderscriptBox[
            SubscriptBox["\[Sum]", "D"], 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
           SubscriptBox["T", "i"]}], ")"}], 
         UnderscriptBox["*", "D"], "c"}]}], 
       UnderscriptBox["\[Succeeds]", "D"], 
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           UnderscriptBox[
            SubscriptBox["\[Sum]", "D"], 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{"k", "-", "1"}]}]], 
           SubscriptBox["T", "i"]}], ")"}], 
         UnderscriptBox["*", "D"], "c"}]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symred reduceM I)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:296464255", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320431219538918`*^9, 3.632043143427567*^9}, {
  3.6320432191276608`*^9, 3.632043231124653*^9}, {3.638597243908004*^9, 
  3.638597247517913*^9}},
 CellTags->{
  "ID:296464255", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM I"},
 CellID->296464255],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["symred", "D"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{"k", "-", "1"}]}], "]"}], ",", 
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", "k"}], "]"}], ",", "c", ",", 
    SubscriptBox["T", "k"]}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["mult", "D"], "[", 
       RowBox[{
        SubscriptBox["T", "k"], ",", "c"}], "]"}], "\[And]", 
      RowBox[{
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           UnderscriptBox[
            SubscriptBox["\[Sum]", "D"], 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
           SubscriptBox["T", "i"]}], ")"}], 
         UnderscriptBox["*", "D"], "c"}]}], 
       UnderscriptBox["\[Succeeds]", "D"], 
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           UnderscriptBox[
            SubscriptBox["\[Sum]", "D"], 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{"k", "-", "1"}]}]], 
           SubscriptBox["T", "i"]}], ")"}], 
         UnderscriptBox["*", "D"], "c"}]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Or]", 
    RowBox[{
     UnderscriptBox["isReducible", "D"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["reduceM", "D"], "[", 
       RowBox[{"a", ",", "c", ",", "T", ",", 
        RowBox[{"k", "-", "1"}]}], "]"}], ",", "c", ",", 
      SubscriptBox["T", "k"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symred reduceM II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:194930655", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63204392804998*^9, 3.6320439631472*^9}, {
  3.6385972504021883`*^9, 3.638597253223398*^9}},
 CellTags->{
  "ID:194930655", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM II"},
 CellID->194930655],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["symred", "D"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{"k", "-", "1"}]}], "]"}], ",", 
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", "k"}], "]"}], ",", "c", ",", 
    SubscriptBox["T", "k"]}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["mult", "D"], "[", 
       RowBox[{
        SubscriptBox["T", "k"], ",", "c"}], "]"}], "\[And]", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"a", 
          UnderscriptBox["-", "D"], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", "D"], 
              RowBox[{
               RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
               RowBox[{"k", "-", "1"}]}]], 
             SubscriptBox["T", "i"]}], ")"}], 
           UnderscriptBox["*", "D"], "c"}]}], ")"}], 
        UnderscriptBox["-", "D"], 
        RowBox[{
         SubscriptBox["T", "k"], 
         UnderscriptBox["*", "D"], "c"}]}], 
       UnderscriptBox["\[Succeeds]", "D"], 
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           UnderscriptBox[
            SubscriptBox["\[Sum]", "D"], 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{"k", "-", "1"}]}]], 
           SubscriptBox["T", "i"]}], ")"}], 
         UnderscriptBox["*", "D"], "c"}]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Or]", 
    RowBox[{
     UnderscriptBox["isReducible", "D"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["reduceM", "D"], "[", 
       RowBox[{"a", ",", "c", ",", "T", ",", 
        RowBox[{"k", "-", "1"}]}], "]"}], ",", "c", ",", 
      SubscriptBox["T", "k"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symred reduceM III)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:935171195", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638597257533435*^9, 3.638597282748343*^9}},
 CellTags->{
  "ID:935171195", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM III"},
 CellID->935171195],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1917615946]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->176929079],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduceM 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduceM 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", "0"}], "]"}], 
           "\[Equal]", 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", "0"}], "]"}], 
           "\[Equal]", 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1300106055"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1300106055-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1300106055-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1300106055", "Proof|ID:1300106055-1"},
 CellID->587044218],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1300106055-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6435349276803083`*^9},
 CellID->691055095]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduceM 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduceM 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["T", "ExpressionVariable"], "\[RightBracketingBar]"}],
              "\[NotEqual]", "0"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["T", "ExpressionVariable"], ",", "1"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], "1"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["T", "ExpressionVariable"], "\[RightBracketingBar]"}],
              "\[NotEqual]", "0"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["T", "ExpressionVariable"], ",", "1"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], "1"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:438711618"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p438711618-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p438711618-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:438711618", "Proof|ID:438711618-1"},
 CellID->457182048],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p438711618-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->871927811]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduceM alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduceM alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["-", 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", 
                   StyleBox["D", "ExpressionVariable"]], 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["-", 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", 
                   StyleBox["D", "ExpressionVariable"]], 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2101232329"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2101232329-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2101232329-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2101232329", "Proof|ID:2101232329-1"},
 CellID->1720715120],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2101232329-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->394118706]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduceM+1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduceM+1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["T", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}], "]"}], 
           "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["T", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["-", 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                RowBox[{
                  StyleBox["k", "ExpressionVariable"], "+", "1"}]], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["T", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}], "]"}], 
           "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["T", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["-", 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                RowBox[{
                  StyleBox["k", "ExpressionVariable"], "+", "1"}]], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:96556784"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p96556784-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p96556784-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:96556784", "Proof|ID:96556784-1"},
 CellID->194852796],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p96556784-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->676400811]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduceM alternative II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduceM alternative II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["-", 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["-", 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1465630278"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1465630278-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1465630278-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1465630278", "Proof|ID:1465630278-1"},
 CellID->1634286780],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1465630278-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1019435258]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership reduceM)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership reduceM)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            StyleBox["D", "ExpressionVariable"]], "[", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            StyleBox["D", "ExpressionVariable"]], "[", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1952076718"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1952076718-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1952076718-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1952076718", "Proof|ID:1952076718-1"},
 CellID->1918441122],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1952076718-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->519581729]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduceM-1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduceM-1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
           "\[Equal]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["-", 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", 
                   StyleBox["D", "ExpressionVariable"]], 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
           "\[Equal]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["-", 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", 
                   StyleBox["D", "ExpressionVariable"]], 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1236615595"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1236615595-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1236615595-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1236615595", "Proof|ID:1236615595-1"},
 CellID->536851896],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1236615595-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->943705171]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduceM-1 II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduceM-1 II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["T", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
             UnderscriptBox["-", 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["k", "ExpressionVariable"]], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["T", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
             UnderscriptBox["-", 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["k", "ExpressionVariable"]], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:203149419"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p203149419-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p203149419-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:203149419", "Proof|ID:203149419-1"},
 CellID->916037400],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p203149419-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.641182319372036*^9},
 CellID->1494767151]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible reduceM)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible reduceM)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["k", "ExpressionVariable"]]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["mult", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["k", "ExpressionVariable"]]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["mult", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:333495152"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p333495152-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p333495152-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:333495152", "Proof|ID:333495152-1"},
 CellID->193609441],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p333495152-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1085025219]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symred reduceM)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symred reduceM)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["symred", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["k", "ExpressionVariable"]]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["mult", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]}], "\[Or]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], 
                    "]"}]}]}]}], ")"}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["symred", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["k", "ExpressionVariable"]]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["mult", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}]}], "\[Or]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], 
                    "]"}]}]}]}], ")"}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:332101094"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p332101094-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p332101094-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:332101094", "Proof|ID:332101094-1"},
 CellID->2106893222],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p332101094-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1457981415]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symred reduceM I)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symred reduceM I)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["symred", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["k", "ExpressionVariable"]]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["mult", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[Or]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["symred", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["k", "ExpressionVariable"]]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["mult", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[Or]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:296464255"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p296464255-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p296464255-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
 CellID->1877618114],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p296464255-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1831291740]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symred reduceM II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symred reduceM II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["symred", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["k", "ExpressionVariable"]]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}], "\[Or]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["symred", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["k", "ExpressionVariable"]]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}], "\[Or]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:194930655"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p194930655-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p194930655-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:194930655", "Proof|ID:194930655-1"},
 CellID->230812744],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p194930655-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->942266403]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symred reduceM III)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symred reduceM III)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["symred", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["k", "ExpressionVariable"]]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}], "\[Or]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["symred", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
               ",", 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["k", "ExpressionVariable"]]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}], "\[Or]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:935171195"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p935171195-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p935171195-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:935171195", "Proof|ID:935171195-1"},
 CellID->577756398],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p935171195-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1380815439],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"Replace", "::", "reps"}], "MessageName"], ":", 
  " ", "\<\"\[NoBreak]\\!\\({1}\\)\[NoBreak] is neither a list of replacement \
rules nor a valid dispatch table, and so cannot be used for replacing. \\!\\(\
\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\\\", \
ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/Replace/reps\\\", \
ButtonNote -> \\\"Replace::reps\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.641188788002688*^9},
 CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
 CellID->20878089],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"Replace", "::", "reps"}], "MessageName"], ":", 
  " ", "\<\"\[NoBreak]\\!\\({1}\\)\[NoBreak] is neither a list of replacement \
rules nor a valid dispatch table, and so cannot be used for replacing. \\!\\(\
\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\\\", \
ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/Replace/reps\\\", \
ButtonNote -> \\\"Replace::reps\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.641188788151361*^9},
 CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
 CellID->1281008759],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"Replace", "::", "reps"}], "MessageName"], ":", 
  " ", "\<\"\[NoBreak]\\!\\({1}\\)\[NoBreak] is neither a list of replacement \
rules nor a valid dispatch table, and so cannot be used for replacing. \\!\\(\
\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\\\", \
ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/Replace/reps\\\", \
ButtonNote -> \\\"Replace::reps\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.641188788162607*^9},
 CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
 CellID->1946324498],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"General", "::", "stop"}], "MessageName"], ":", 
  " ", "\<\"Further output of \[NoBreak]\\!\\(\\*StyleBox[\\(Replace :: \
reps\\), \\\"MessageName\\\"]\\)\[NoBreak] will be suppressed during this \
calculation. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", \
ButtonStyle->\\\"Link\\\", ButtonFrame->None, \
ButtonData:>\\\"paclet:ref/message/General/stop\\\", ButtonNote -> \
\\\"General::stop\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.641188788292759*^9},
 CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
 CellID->1486921856]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1798794743],

Cell[CellGroupData[{

Cell["Lemma: Symmetry of areIrrelative and areCorrelative", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.633174114370289*^9, 3.633174125459807*^9}},
 CellID->1097525720],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"c1", ",", "c2", ",", "m1", ",", "m2"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1368508872}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6331741312791233`*^9, 3.6331741455807734`*^9}, {
  3.6380885581706038`*^9, 3.638088561741742*^9}, {3.638088667550789*^9, 
  3.638088674037601*^9}},
 CellTags->{
  "ID:1368508872", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1368508872],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     UnderscriptBox["multN", "D"], "[", "c1", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["areIrrelative", "D"], "[", 
     RowBox[{"c1", ",", "c2", ",", "m1", ",", "m2", ",", "i", ",", "j"}], 
     "]"}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox["areIrrelative", "D"], "[", 
     RowBox[{"c1", ",", "c2", ",", "m2", ",", "m1", ",", "j", ",", "i"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areIrrelative symmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1610984937", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633174150955105*^9, 3.633174162273597*^9}, {
  3.633174199310502*^9, 3.633174222854794*^9}},
 CellTags->{
  "ID:1610984937", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areIrrelative symmetric"},
 CellID->1610984937],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areCorrelative", "D"], "[", 
   RowBox[{"c1", ",", "m1", ",", "m2"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["areCorrelative", "D"], "[", 
   RowBox[{"c1", ",", "m2", ",", "m1"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areCorrelative symmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:981548331", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633174229547803*^9, 3.633174241345645*^9}, {
  3.638611084218925*^9, 3.638611085543399*^9}},
 CellTags->{
  "ID:981548331", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCorrelative symmetric"},
 CellID->981548331],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->453207686]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->231330468],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(areIrrelative symmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(areIrrelative symmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areIrrelative", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["areIrrelative", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areIrrelative", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["areIrrelative", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1610984937"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1610984937-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1610984937-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1610984937", "Proof|ID:1610984937-1"},
 CellID->1688165426],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1610984937-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1794986575]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(areCorrelative symmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(areCorrelative symmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areCorrelative", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["areCorrelative", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areCorrelative", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["areCorrelative", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:981548331"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p981548331-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p981548331-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:981548331", "Proof|ID:981548331-1"},
 CellID->437316908],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p981548331-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->337585423]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1696462827],

Cell[CellGroupData[{

Cell["\<\
Lemma: isReducible, symred and \[Rule]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.633234810523841*^9, 3.6332348171810207`*^9}, {
  3.6421573928789597`*^9, 3.6421573941932096`*^9}, {3.644828705609363*^9, 
  3.644828723884705*^9}, {3.648377371539394*^9, 3.648377375531116*^9}},
 CellID->1520519732],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "m"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1547543437}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638087762453986*^9, 3.638087771599769*^9}, {
  3.638087814211343*^9, 3.638087814361659*^9}, {3.638088533401967*^9, 
  3.6380885345429373`*^9}, {3.6380886560436707`*^9, 3.638088665003683*^9}, {
  3.648377623499028*^9, 3.648377626597975*^9}},
 CellTags->{
  "ID:1547543437", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1547543437],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{"a", ",", "c", ",", "m"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"{", "c", "}"}]], "D"], 
    RowBox[{"a", 
     UnderscriptBox["-", "D"], 
     RowBox[{"m", 
      UnderscriptBox["*", "D"], "c"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible \[RightArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1699369800", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6332348221314898`*^9, 3.633234863401033*^9}, {
  3.63808777470618*^9, 3.6380877776156673`*^9}, {3.6618500201597357`*^9, 
  3.661850020908537*^9}},
 CellTags->{
  "ID:1699369800", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible \[RightArrow]"},
 CellID->1699369800],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"m", ",", "c"}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "m"}], "]"}], "\[And]", 
    RowBox[{"b", "\[Equal]", 
     RowBox[{"a", 
      UnderscriptBox["-", "D"], 
      RowBox[{"m", 
       UnderscriptBox["*", "D"], "c"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow] isReducible)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1596540807", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63808782172408*^9, 3.638087863875787*^9}},
 CellTags->{
  "ID:1596540807", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] isReducible"},
 CellID->1596540807],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{"a", ",", "c", ",", "m"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["symred", "D"], "[", 
   RowBox[{"a", ",", 
    RowBox[{"a", 
     UnderscriptBox["-", "D"], 
     RowBox[{"m", 
      UnderscriptBox["*", "D"], "c"}]}], ",", "c", ",", "m"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible symred)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1641388157", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6421574074118023`*^9, 3.6421574611934032`*^9}},
 CellTags->{
  "ID:1641388157", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible symred"},
 CellID->1641388157],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{"a", ",", "c"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", "d", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"{", "c", "}"}]], "D"], "d"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2087655986", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648184032220209*^9, 3.6481840614180098`*^9}, 
   3.6481842607143106`*^9, {3.648184387011701*^9, 3.6481844067792673`*^9}, {
   3.661850042015374*^9, 3.6618500451821795`*^9}},
 CellTags->{
  "ID:2087655986", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible alternative", "IR"},
 CellID->2087655986],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "T", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", 
       RowBox[{"mi", ",", "mj"}], "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"k", ",", "l"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        UnderscriptBox["areCPMultipliers", "D"], "[", 
        RowBox[{"mi", ",", "mj", ",", 
         SubscriptBox["T", "i"], ",", 
         SubscriptBox["T", "j"], ",", "a", ",", "k", ",", "l"}], "]"}], 
       "\[Implies]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"a", 
           UnderscriptBox[
            SubscriptBox["\[Rule]", "T"], "D"], 
           RowBox[{"a", 
            UnderscriptBox["-", "D"], 
            RowBox[{"mi", 
             UnderscriptBox["*", "D"], 
             SubscriptBox["T", "i"]}]}]}], 
          TagBox[")",
           "AutoParentheses"]}], "\[And]", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"a", 
           UnderscriptBox[
            SubscriptBox["\[Rule]", "T"], "D"], 
           RowBox[{"a", 
            UnderscriptBox["-", "D"], 
            RowBox[{"mj", 
             UnderscriptBox["*", "D"], 
             SubscriptBox["T", "j"]}]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areCPMultipliers \[RightArrow] tuple)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:584152369", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6483774043366337`*^9, 3.648377529904887*^9}, {
  3.648377630183597*^9, 3.6483776303698883`*^9}},
 CellTags->{
  "ID:584152369", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers \[RightArrow] tuple"},
 CellID->584152369],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1376947800]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1403180676],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible \[RightArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible \[RightArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{"{", 
                 StyleBox["c", "ExpressionVariable"], "}"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{"{", 
                 StyleBox["c", "ExpressionVariable"], "}"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1699369800"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1699369800-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1699369800-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1699369800", "Proof|ID:1699369800-1"},
 CellID->199699432],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1699369800-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->60679348]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] isReducible)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] isReducible)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["m", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["m", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1596540807"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1596540807-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1596540807-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1596540807", "Proof|ID:1596540807-1"},
 CellID->314505815],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1596540807-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->54928433]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible symred)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible symred)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["symred", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["symred", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1641388157"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1641388157-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1641388157-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1641388157", "Proof|ID:1641388157-1"},
 CellID->20930826],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1641388157-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1555086701]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "d", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "d", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2087655986"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2087655986-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2087655986-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2087655986", "Proof|ID:2087655986-1"},
 CellID->65047792],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2087655986-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->650270676]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(areCPMultipliers \[RightArrow] tuple)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(areCPMultipliers \[RightArrow] tuple)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["mi", "ExpressionVariable"], ",", 
               StyleBox["mj", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["mi", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["mj", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox[
                    "j", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["mi", "ExpressionVariable"], ",", 
               StyleBox["mj", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["mi", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["mj", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox[
                    "j", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:584152369"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p584152369-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p584152369-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:584152369", "Proof|ID:584152369-1"},
 CellID->444231087],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p584152369-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->499606199]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1668167977],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Simple Consequences of \[Rule] and ",
 Cell[BoxData[
  FormBox[
   OverscriptBox[
    SuperscriptBox["\[LeftRightArrow]", "*"], 
    RowBox[{"z", ",", "\[Succeeds]"}]], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.644901874224307*^9, 3.6449019113624887`*^9}},
 CellID->181072362],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b", ",", "z"}], "]"}]}], ",", 
   "r"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 593421628}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6449019177919703`*^9, 3.6449019251245193`*^9}, {
  3.6449021040404797`*^9, 3.64490210418908*^9}},
 CellTags->{
  "ID:593421628", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->593421628],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
      RowBox[{"z", ",", "\[Succeeds]"}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"z", 
     UnderscriptBox["\[Succeeds]", "D"], "a"}], "\[And]", 
    RowBox[{"z", 
     UnderscriptBox["\[Succeeds]", "D"], "b"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:680005293", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644901976359293*^9, 3.6449020205685453`*^9}},
 CellTags->{
  "ID:680005293", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "1"},
 CellID->680005293],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", "r"], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     UnderscriptBox["isReductor", "D"], "[", "r", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1320396264}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6448285763125763`*^9, 3.644828583548328*^9}, {
  3.6449020928399887`*^9, 3.64490209686119*^9}},
 CellTags->{
  "ID:1320396264", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1320396264],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Exists]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "d", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"{", "d", "}"}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow] reductor)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1127232822", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644828148076118*^9, 3.644828199012772*^9}, {
  3.644828567317482*^9, 3.644828585916545*^9}, {3.661850093557865*^9, 
  3.6618500950866675`*^9}},
 CellTags->{
  "ID:1127232822", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] reductor"},
 CellID->1127232822],

Cell[BoxData[
 RowBox[{"a", 
  UnderscriptBox["\[Succeeds]", "D"], "b"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Succeeds] reductor)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:256297052", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644828598473783*^9, 3.6448286038859997`*^9}},
 CellTags->{
  "ID:256297052", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Succeeds] reductor"},
 CellID->256297052],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->963051773]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->895896855],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["a", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["a", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:680005293"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p680005293-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p680005293-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:680005293", "Proof|ID:680005293-1"},
 CellID->41814150],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p680005293-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->438053564]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] reductor)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] reductor)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["r", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["d", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["r", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["d", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1127232822"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1127232822-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1127232822-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1127232822", "Proof|ID:1127232822-1"},
 CellID->625038390],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1127232822-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->111447419]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Succeeds] reductor)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Succeeds] reductor)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["r", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["r", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:256297052"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p256297052-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p256297052-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:256297052", "Proof|ID:256297052-1"},
 CellID->256602949],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p256297052-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->162186853]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1503630565],

Cell["Lemma: Properties of red", "EnvironmentHeader",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6466501429642344`*^9, 3.646650144189267*^9}, {
   3.6467319402316923`*^9, 3.646731942469968*^9}, 3.6536273323022165`*^9},
 CellID->879564337],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"x", ",", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "D", "]"}]], 
    RowBox[{"[", 
     RowBox[{"A", ",", "B"}], "]"}]}]}]]], "GlobalDeclaration",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1083403145}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646650580840349*^9, 3.646650582362007*^9}, 
   3.646650997835631*^9, {3.653626063647986*^9, 3.6536260741312046`*^9}, {
   3.653626886667632*^9, 3.653626887400833*^9}, {3.653626939427925*^9, 
   3.653626939786725*^9}, 3.6536273323022165`*^9},
 CellTags->{
  "ID:1083403145", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1083403145],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainSets", "[", "D", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["red", "D"], "[", "A", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DomainSets red)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1592577568", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6466492904779053`*^9, 3.64664931772274*^9}, {
   3.653626158715553*^9, 3.6536261622411594`*^9}, 3.6536273323022165`*^9},
 CellTags->{
  "ID:1592577568", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets red", "IR"},
 CellID->1592577568],

Cell[BoxData[
 RowBox[{"isSet", "[", 
  RowBox[{
   UnderscriptBox["red", "D"], "[", "A", "]"}], 
  "]"}]], "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red isSet)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:698444072", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646652285420834*^9, 3.64665229758563*^9}, {
   3.6536261572491503`*^9, 3.6536261606343565`*^9}, 3.6536273323022165`*^9},
 CellTags->{
  "ID:698444072", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red isSet", "IR"},
 CellID->698444072],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", "\[Element]", 
    RowBox[{
     UnderscriptBox["red", "D"], "[", "A", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", "x", "]"}]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{"a", "\[Element]", "A"}]], 
      RowBox[{
       UnderscriptBox["isReducible", "D"], "[", 
       RowBox[{"x", ",", "a"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] red)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:277656140", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646648699656554*^9, 3.646648760411564*^9}, {
   3.646649011125388*^9, 3.646649025849728*^9}, {3.646649063173129*^9, 
   3.646649112683793*^9}, 3.646651000309252*^9, {3.653625902933503*^9, 
   3.653625947566182*^9}, {3.65362607743841*^9, 3.653626080776816*^9}, 
   3.6536273323022165`*^9},
 CellTags->{
  "ID:277656140", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red"},
 CellID->277656140],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "c", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Element]", 
      RowBox[{
       UnderscriptBox["red", "D"], "[", 
       RowBox[{"{", "c", "}"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", "x", "]"}]}], "\[And]", 
      RowBox[{
       UnderscriptBox["isReducible", "D"], "[", 
       RowBox[{"x", ",", "c"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] red singleton)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:454978652", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646650554451879*^9, 3.646650615125828*^9}, {
   3.653625991372059*^9, 3.6536259939460635`*^9}, 3.6536273323022165`*^9, {
   3.6536276452009664`*^9, 3.653627653344181*^9}},
 CellTags->{
  "ID:454978652", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red singleton"},
 CellID->454978652],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"isSet", "[", "C", "]"}], ",", "f", ",", "P"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"c", "\[Element]", "C"}]], 
       RowBox[{"P", "[", "c", "]"}]], 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", 
        RowBox[{"f", "[", "c", "]"}], "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"x", "\[Element]", 
        RowBox[{
         UnderscriptBox["red", "D"], "[", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"f", "[", "c", "]"}], 
           UnderscriptBox["|", 
            RowBox[{"c", "\[Element]", "C"}]], 
           RowBox[{"P", "[", "c", "]"}]}], "}"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["\[Element]", "D"], 
         RowBox[{"[", "x", "]"}]}], "\[And]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          UnderscriptBox[
           UnderscriptBox["\[Exists]", 
            RowBox[{"c", "\[Element]", "C"}]], 
           RowBox[{"P", "[", "c", "]"}]], 
          RowBox[{
           UnderscriptBox["isReducible", "D"], "[", 
           RowBox[{"x", ",", 
            RowBox[{"f", "[", "c", "]"}]}], "]"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] red abstraction term)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:259827218", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646731385757761*^9, 3.646731457303961*^9}, {
   3.6467315518593473`*^9, 3.646731572521448*^9}, {3.653626031354929*^9, 
   3.6536260414013467`*^9}, {3.6536260867984266`*^9, 
   3.6536260947700405`*^9}, {3.6536269428911304`*^9, 3.6536269650129695`*^9}, 
   3.6536273323022165`*^9, {3.653628413737924*^9, 3.6536284183711324`*^9}, {
   3.6536285954444437`*^9, 3.6536286035720587`*^9}},
 CellTags->{
  "ID:259827218", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red abstraction term"},
 CellID->259827218],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", "\[Element]", 
    RowBox[{
     UnderscriptBox["red", "D"], "[", 
     RowBox[{"A", "\[Union]", "B"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[Element]", 
     RowBox[{
      UnderscriptBox["red", "D"], "[", "A", "]"}]}], "\[Or]", 
    RowBox[{"x", "\[Element]", 
     RowBox[{
      UnderscriptBox["red", "D"], "[", "B", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] red \[Union])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1273500442", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646650903975237*^9, 3.646650960497446*^9}, {
   3.6466510260600452`*^9, 3.64665108655728*^9}, 3.6536260845676227`*^9, {
   3.6536269092730713`*^9, 3.6536269133134785`*^9}, 3.6536273323022165`*^9},
 CellTags->{
  "ID:1273500442", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red \[Union]"},
 CellID->1273500442],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["red", "D"], "[", 
   RowBox[{"A", "\[Union]", "B"}], "]"}], "\[Equal]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["red", "D"], "[", "A", "]"}], "\[Union]", 
   RowBox[{
    UnderscriptBox["red", "D"], "[", "B", "]"}]}]}]], "FormalTextInputFormula",\

 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red \[Union])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1877397922", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6466488164928083`*^9, 3.6466488561239157`*^9}, {
   3.6467097954936047`*^9, 3.646709798130801*^9}, 3.6536273323178167`*^9},
 CellTags->{
  "ID:1877397922", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Union]"},
 CellID->1877397922],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "b", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Element]", 
      RowBox[{
       UnderscriptBox["red", "D"], "[", 
       RowBox[{"A", "\[Union]", 
        RowBox[{"{", "b", "}"}]}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"x", "\[Element]", 
       RowBox[{
        UnderscriptBox["red", "D"], "[", "A", "]"}]}], "\[Or]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["\[Element]", "D"], 
         RowBox[{"[", "x", "]"}]}], "\[And]", 
        RowBox[{
         UnderscriptBox["isReducible", "D"], "[", 
         RowBox[{"x", ",", "b"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] red \[Union] singleton)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2121666706", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648185714524878*^9, 3.648185764707617*^9}, {
   3.6481859320586243`*^9, 3.64818594349475*^9}, 3.648186173412035*^9, {
   3.6536268140963044`*^9, 3.6536268432527556`*^9}, 3.6536273323178167`*^9},
 CellTags->{
  "ID:2121666706", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red \[Union] singleton"},
 CellID->2121666706],

Cell[BoxData[
 RowBox[{
  RowBox[{"A", "\[SubsetEqual]", "B"}], "\[Implies]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["red", "D"], "[", "A", "]"}], "\[SubsetEqual]", 
   RowBox[{
    UnderscriptBox["red", "D"], "[", "B", "]"}]}]}]], "FormalTextInputFormula",\

 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[SubsetEqual] red \[SubsetEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:947744040", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646968176392482*^9, 3.646968192343278*^9}, 
   3.6536273323178167`*^9},
 CellTags->{
  "ID:947744040", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SubsetEqual] red \[SubsetEqual]"},
 CellID->947744040],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["red", "D"], "[", "A", "]"}], "\[Subset]", 
    RowBox[{
     UnderscriptBox["red", "D"], "[", "B", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{"b", "\[Element]", "B"}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", "a", "]"}]}]], 
      RowBox[{
       UnderscriptBox["isReducible", "D"], "[", 
       RowBox[{"a", ",", "b"}], "]"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[Not]", 
       RowBox[{"a", "\[Element]", 
        RowBox[{
         UnderscriptBox["red", "D"], "[", "A", "]"}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red \[Subset])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:52516181", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646708373992956*^9, 3.64670838768747*^9}, {
   3.6467084243240643`*^9, 3.646708517558937*^9}, {3.6467085551231613`*^9, 
   3.646708609511586*^9}, {3.64670873054921*^9, 3.6467087481681023`*^9}, {
   3.6467097997552023`*^9, 3.646709802906702*^9}, {3.6536266601678343`*^9, 
   3.6536266787630663`*^9}, 3.6536273323178167`*^9},
 CellTags->{
  "ID:52516181", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset]"},
 CellID->52516181],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "b", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["red", "D"], "[", "A", "]"}], "\[Subset]", 
      RowBox[{
       UnderscriptBox["red", "D"], "[", 
       RowBox[{"A", "\[Union]", 
        RowBox[{"{", "b", "}"}]}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "D"], 
         RowBox[{"[", "r", "]"}]}]], 
       RowBox[{
        UnderscriptBox["isReducible", "D"], "[", 
        RowBox[{"r", ",", "b"}], "]"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[Not]", 
        RowBox[{"r", "\[Element]", 
         RowBox[{
          UnderscriptBox["red", "D"], "[", "A", "]"}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red \[Subset] \[Union] singleton)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:194436663", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646713544828823*^9, 3.6467136005900307`*^9}, {
   3.646728010973634*^9, 3.646728013119564*^9}, {3.646728043243886*^9, 
   3.646728083333041*^9}, {3.6536267133327274`*^9, 3.653626728356554*^9}, 
   3.6536273323178167`*^9},
 CellTags->{
  "ID:194436663", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset] \[Union] singleton"},
 CellID->194436663],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "b", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{
        UnderscriptBox["red", "D"], "[", "A", "]"}], "\[Subset]", 
       RowBox[{
        UnderscriptBox["red", "D"], "[", 
        RowBox[{"A", "\[Union]", 
         RowBox[{"{", "b", "}"}]}], "]"}]}], "\[And]", 
      RowBox[{"B", "\[SubsetEqual]", "A"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["red", "D"], "[", "B", "]"}], "\[Subset]", 
     RowBox[{
      UnderscriptBox["red", "D"], "[", 
      RowBox[{"B", "\[Union]", 
       RowBox[{"{", "b", "}"}]}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red \[Subset] \[Union] singleton II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:844584535", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646979032721781*^9, 3.646979076044217*^9}, {
   3.64699619026056*^9, 3.646996217059462*^9}, 3.6536273323178167`*^9},
 CellTags->{
  "ID:844584535", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset] \[Union] singleton II"},
 CellID->844584535],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellChangeTimes->{3.6536273323178167`*^9},
 CellID->212448877],

Cell["", "CloseEnvironment",
 CellID->943439096],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DomainSets red)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DomainSets red)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainSets", "[", 
              StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{
             UnderscriptBox["red", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             StyleBox["A", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainSets", "[", 
              StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{
             UnderscriptBox["red", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             StyleBox["A", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1592577568"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1592577568-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1592577568-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1592577568", "Proof|ID:1592577568-1"},
 CellID->272356749],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1592577568-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->513310847]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(red isSet)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(red isSet)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isSet", "[", 
           RowBox[{
             UnderscriptBox["red", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             StyleBox["A", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isSet", "[", 
           RowBox[{
             UnderscriptBox["red", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             StyleBox["A", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:698444072"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p698444072-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p698444072-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:698444072", "Proof|ID:698444072-1"},
 CellID->6276063],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p698444072-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->85725627]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] red)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] red)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"]}], "]"}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"]}], "]"}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:277656140"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p277656140-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p277656140-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:277656140", "Proof|ID:277656140-1"},
 CellID->45788410],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p277656140-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->230626054]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] red singleton)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] red singleton)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}], "]"}]}], ")"}],
            "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}], "]"}]}], ")"}],
            "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:454978652"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p454978652-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p454978652-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:454978652", "Proof|ID:454978652-1"},
 CellID->54717562],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p454978652-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->889655]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] red abstraction term)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] red abstraction term)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             RowBox[{"isSet", "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             StyleBox["f", "ExpressionVariable"]}, {
             StyleBox["P", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[Element]", 
                    StyleBox["C", "ExpressionVariable"]}]}}]], 
              RowBox[{
                StyleBox["P", "ExpressionVariable"], "[", 
                StyleBox["c", "ExpressionVariable"], "]"}]], 
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{"(", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[Element]", 
                    StyleBox["C", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}], "}"}], 
                   "]"}]}], ")"}], "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[Element]", 
                    StyleBox["C", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]], 
                   RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}], "]"}]}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             RowBox[{"isSet", "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             StyleBox["f", "ExpressionVariable"]}, {
             StyleBox["P", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[Element]", 
                    StyleBox["C", "ExpressionVariable"]}]}}]], 
              RowBox[{
                StyleBox["P", "ExpressionVariable"], "[", 
                StyleBox["c", "ExpressionVariable"], "]"}]], 
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{"(", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[Element]", 
                    StyleBox["C", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}], "}"}], 
                   "]"}]}], ")"}], "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[Element]", 
                    StyleBox["C", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]], 
                   RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}], "]"}]}]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:259827218"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p259827218-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p259827218-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:259827218", "Proof|ID:259827218-1"},
 CellID->183930652],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p259827218-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->803294713]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] red \[Union])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] red \[Union])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\[Union]", 
                   StyleBox["B", "ExpressionVariable"]}], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]}], "\[Or]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\[Union]", 
                   StyleBox["B", "ExpressionVariable"]}], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]}], "\[Or]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1273500442"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1273500442-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1273500442-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1273500442", "Proof|ID:1273500442-1"},
 CellID->172667338],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1273500442-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->536407252]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(red \[Union])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(red \[Union])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["red", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], "\[Union]", 
               StyleBox["B", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[Union]", 
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["B", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["red", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], "\[Union]", 
               StyleBox["B", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[Union]", 
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["B", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1877397922"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1877397922-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1877397922-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1877397922", "Proof|ID:1877397922-1"},
 CellID->133276848],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1877397922-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->173577853]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] red \[Union] singleton)\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] red \[Union] singleton)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\[Union]", 
                   RowBox[{"{", 
                    StyleBox["b", "ExpressionVariable"], "}"}]}], "]"}]}], 
             ")"}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"]}], "]"}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\[Union]", 
                   RowBox[{"{", 
                    StyleBox["b", "ExpressionVariable"], "}"}]}], "]"}]}], 
             ")"}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"]}], "]"}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2121666706"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2121666706-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2121666706-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2121666706", "Proof|ID:2121666706-1"},
 CellID->185736564],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2121666706-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->133384348]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[SubsetEqual] red \[SubsetEqual])\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(\[SubsetEqual] red \[SubsetEqual])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["A", "ExpressionVariable"], "\[SubsetEqual]", 
             StyleBox["B", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[SubsetEqual]", 
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["B", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["A", "ExpressionVariable"], "\[SubsetEqual]", 
             StyleBox["B", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[SubsetEqual]", 
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["B", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:947744040"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p947744040-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p947744040-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:947744040", "Proof|ID:947744040-1"},
 CellID->249774080],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p947744040-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->411810760]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(red \[Subset])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(red \[Subset])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[Subset]", 
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Element]", 
                   StyleBox["B", "ExpressionVariable"]}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  UnderscriptBox["isReducible", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"]}], "]"}]], 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    RowBox[{
                    UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}], "\[Subset]", 
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Element]", 
                   StyleBox["B", "ExpressionVariable"]}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  UnderscriptBox["isReducible", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"]}], "]"}]], 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    RowBox[{
                    UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:52516181"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p52516181-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p52516181-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:52516181", "Proof|ID:52516181-1"},
 CellID->50697679],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p52516181-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->751899172]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(red \[Subset] \[Union] singleton)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(red \[Subset] \[Union] singleton)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\[Union]", 
                   RowBox[{"{", 
                    StyleBox["b", "ExpressionVariable"], "}"}]}], "]"}]}], 
             ")"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                UnderscriptBox["isReducible", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["r", "ExpressionVariable"], ",", 
                  StyleBox["b", "ExpressionVariable"]}], "]"}]], 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], "\[Element]", 
                   RowBox[{
                    UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\[Union]", 
                   RowBox[{"{", 
                    StyleBox["b", "ExpressionVariable"], "}"}]}], "]"}]}], 
             ")"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                UnderscriptBox["isReducible", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["r", "ExpressionVariable"], ",", 
                  StyleBox["b", "ExpressionVariable"]}], "]"}]], 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], "\[Element]", 
                   RowBox[{
                    UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:194436663"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p194436663-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p194436663-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:194436663", "Proof|ID:194436663-1"},
 CellID->210264695],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p194436663-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->22396559]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(red \[Subset] \[Union] singleton II)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(red \[Subset] \[Union] singleton II)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\[Union]", 
                   RowBox[{"{", 
                    StyleBox["b", "ExpressionVariable"], "}"}]}], "]"}]}], 
             "\[And]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], "\[SubsetEqual]", 
               StyleBox["A", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}], "\[Subset]", 
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\[Union]", 
                 RowBox[{"{", 
                   StyleBox["b", "ExpressionVariable"], "}"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["A", "ExpressionVariable"], "]"}], "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["A", "ExpressionVariable"], "\[Union]", 
                   RowBox[{"{", 
                    StyleBox["b", "ExpressionVariable"], "}"}]}], "]"}]}], 
             "\[And]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], "\[SubsetEqual]", 
               StyleBox["A", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}], "\[Subset]", 
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\[Union]", 
                 RowBox[{"{", 
                   StyleBox["b", "ExpressionVariable"], "}"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:844584535"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p844584535-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p844584535-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:844584535", "Proof|ID:844584535-1"},
 CellID->34854128],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p844584535-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->786223699]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->147378381],

Cell[CellGroupData[{

Cell["Lemma: Sequences Involving red", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6469693905632467`*^9, 3.6469693941906033`*^9}, {
  3.6469757929375687`*^9, 3.646975794670103*^9}},
 CellID->1275480004],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSequences", "[", 
      RowBox[{"DomainSets", "[", "D", "]"}], "]"}]], 
    RowBox[{"[", "B", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
      RowBox[{
       RowBox[{
        UnderscriptBox["red", "D"], "[", 
        RowBox[{"B", "[", "i", "]"}], "]"}], "\[Subset]", 
       RowBox[{
        UnderscriptBox["red", "D"], "[", 
        RowBox[{"B", "[", 
         RowBox[{"i", "+", "1"}], "]"}], "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i", ",", 
         RowBox[{"j", "\[Element]", "\[DoubleStruckCapitalN]"}]}]], 
       RowBox[{"j", "\[GreaterEqual]", "i"}]], 
      RowBox[{
       RowBox[{
        UnderscriptBox["red", "D"], "[", 
        RowBox[{"B", "[", "i", "]"}], "]"}], "\[SubsetEqual]", 
       RowBox[{
        UnderscriptBox["red", "D"], "[", 
        RowBox[{"B", "[", "j", "]"}], "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red \[SubsetEqual] transitive)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:632054408", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646969444932692*^9, 3.6469695001808567`*^9}, {
  3.646970790394511*^9, 3.646970833539028*^9}},
 CellTags->{
  "ID:632054408", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[SubsetEqual] transitive"},
 CellID->632054408],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSequences", "[", "D", "]"}]], 
   RowBox[{"[", "f", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 387695114}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6469693979009132`*^9, 3.646969410048499*^9}},
 CellTags->{
  "ID:387695114", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->387695114],

Cell[TextData[{
 "The following formula is a specialization of (setSeq subsequence P binary) \
from ElementaryTheories, for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"P", "[", 
     RowBox[{"x", ",", "y"}], "]"}], ":", 
    RowBox[{
     RowBox[{"\[DoubleLeftRightArrow]", 
      RowBox[{"red", "[", "x", "]"}]}], "\[Subset]", 
     RowBox[{"red", "[", "y", "]"}]}]}], TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.646996808184228*^9, 3.646996877356434*^9}},
 CellID->44299611],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
    RowBox[{
     RowBox[{
      UnderscriptBox["red", "D"], "[", 
      RowBox[{
       RowBox[{"setSeq", "[", "f", "]"}], "[", "i", "]"}], "]"}], "\[Subset]", 
     RowBox[{
      UnderscriptBox["red", "D"], "[", 
      RowBox[{
       RowBox[{"setSeq", "[", "f", "]"}], "[", 
       RowBox[{"i", "+", "1"}], "]"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", "h"], 
     RowBox[{"h", "\[LeftTriangleBar]", "f"}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
     RowBox[{
      RowBox[{
       UnderscriptBox["red", "D"], "[", 
       RowBox[{
        RowBox[{"setSeq", "[", "h", "]"}], "[", "i", "]"}], "]"}], 
      "\[Subset]", 
      RowBox[{
       UnderscriptBox["red", "D"], "[", 
       RowBox[{
        RowBox[{"setSeq", "[", "h", "]"}], "[", 
        RowBox[{"i", "+", "1"}], "]"}], "]"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(setSeq subsequence red)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1177850337", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646996682413373*^9, 3.6469967498980827`*^9}},
 CellTags->{
  "ID:1177850337", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "setSeq subsequence red"},
 CellID->1177850337],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "i", "\[Element]", "\[DoubleStruckCapitalN]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1483708233}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6469762440672073`*^9, 3.6469762478528852`*^9}},
 CellTags->{
  "ID:1483708233", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1483708233],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["red", "D"], "[", 
   RowBox[{
    RowBox[{"setSeq", "[", "f", "]"}], "[", "i", "]"}], "]"}], 
  "\[SubsetEqual]", 
  RowBox[{
   UnderscriptBox["red", "D"], "[", 
   RowBox[{
    RowBox[{"setSeq", "[", "f", "]"}], "[", 
    RowBox[{"i", "+", "1"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red setSeq \[SubsetEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:697433658", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646976271825375*^9, 3.646976274687495*^9}},
 CellTags->{
  "ID:697433658", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red setSeq \[SubsetEqual]"},
 CellID->697433658],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "a"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", "\[Element]", 
      RowBox[{
       UnderscriptBox["red", "D"], "[", 
       RowBox[{
        RowBox[{"setSeq", "[", "f", "]"}], "[", "i", "]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", "a", "]"}]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox[
         UnderscriptBox["\[Exists]", 
          RowBox[{"j", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
         RowBox[{"i", "\[GreaterEqual]", "j"}]], 
        RowBox[{
         UnderscriptBox["isReducible", "D"], "[", 
         RowBox[{"a", ",", 
          RowBox[{"f", "[", "j", "]"}]}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] red setSeq)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:35999503", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646731227010594*^9, 3.6467312578764763`*^9}, 
   3.6469694274214573`*^9, 3.646976250661257*^9, {3.6536389881413655`*^9, 
   3.653639003117392*^9}, {3.6536391780670996`*^9, 3.65363918969112*^9}, {
   3.65363932571476*^9, 3.6536393268691616`*^9}},
 CellTags->{
  "ID:35999503", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red setSeq"},
 CellID->35999503],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["red", "D"], "[", 
     RowBox[{
      RowBox[{"setSeq", "[", "f", "]"}], "[", "i", "]"}], "]"}], "\[Subset]", 
    
    RowBox[{
     UnderscriptBox["red", "D"], "[", 
     RowBox[{
      RowBox[{"setSeq", "[", "f", "]"}], "[", 
      RowBox[{"i", "+", "1"}], "]"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", "a", "]"}]}]], 
     RowBox[{
      UnderscriptBox["isReducible", "D"], "[", 
      RowBox[{"a", ",", 
       RowBox[{"f", "[", 
        RowBox[{"i", "+", "1"}], "]"}]}], "]"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{"a", "\[Element]", 
       RowBox[{
        UnderscriptBox["red", "D"], "[", 
        RowBox[{
         RowBox[{"setSeq", "[", "f", "]"}], "[", "i", "]"}], "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red setSeq \[Subset])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:602168712", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.646727551631023*^9, 3.646727566616371*^9}, {
   3.6467310447215023`*^9, 3.646731048948064*^9}, {3.646731091254589*^9, 
   3.646731111419874*^9}, {3.646732959875085*^9, 3.646732977090383*^9}, 
   3.646969428788094*^9, {3.646976252068853*^9, 3.646976255321128*^9}, {
   3.653639768055938*^9, 3.65363978646397*^9}},
 CellTags->{
  "ID:602168712", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red setSeq \[Subset]"},
 CellID->602168712],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1700325855]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->472609169],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(red \[SubsetEqual] transitive)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(red \[SubsetEqual] transitive)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  RowBox[{"DomainSets", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[SubsetEqual]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   StyleBox["j", "ExpressionVariable"], "]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  RowBox[{"DomainSets", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[SubsetEqual]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "[", 
                   StyleBox["j", "ExpressionVariable"], "]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:632054408"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p632054408-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p632054408-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:632054408", "Proof|ID:632054408-1"},
 CellID->70698580],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p632054408-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->51683851]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(setSeq subsequence red)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(setSeq subsequence red)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               RowBox[{
                 StyleBox["h", "ExpressionVariable"]}]], 
              RowBox[{
                StyleBox["h", "ExpressionVariable"], "\[LeftTriangleBar]", 
                StyleBox["f", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["h", "ExpressionVariable"], "]"}], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                 "\[Subset]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["h", "ExpressionVariable"], "]"}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               RowBox[{
                 StyleBox["h", "ExpressionVariable"]}]], 
              RowBox[{
                StyleBox["h", "ExpressionVariable"], "\[LeftTriangleBar]", 
                StyleBox["f", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["h", "ExpressionVariable"], "]"}], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                 "\[Subset]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["h", "ExpressionVariable"], "]"}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1177850337"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1177850337-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1177850337-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1177850337", "Proof|ID:1177850337-1"},
 CellID->1013465541],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1177850337-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1284941524]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(red setSeq \[SubsetEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(red setSeq \[SubsetEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["red", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{"setSeq", "[", 
                 StyleBox["f", "ExpressionVariable"], "]"}], "[", 
               StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
           "\[SubsetEqual]", 
           RowBox[{
             UnderscriptBox["red", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{"setSeq", "[", 
                 StyleBox["f", "ExpressionVariable"], "]"}], "[", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["red", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{"setSeq", "[", 
                 StyleBox["f", "ExpressionVariable"], "]"}], "[", 
               StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
           "\[SubsetEqual]", 
           RowBox[{
             UnderscriptBox["red", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{"setSeq", "[", 
                 StyleBox["f", "ExpressionVariable"], "]"}], "[", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:697433658"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p697433658-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p697433658-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:697433658", "Proof|ID:697433658-1"},
 CellID->179082785],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p697433658-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->182798619]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] red setSeq)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] red setSeq)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             StyleBox["a", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}]}], ")"}],
            "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["j", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["j", "ExpressionVariable"], "]"}]}], "]"}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             StyleBox["a", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}]}], ")"}],
            "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["j", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["j", "ExpressionVariable"], "]"}]}], "]"}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:35999503"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p35999503-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p35999503-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:35999503", "Proof|ID:35999503-1"},
 CellID->361034880],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p35999503-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->149032911]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(red setSeq \[Subset])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(red setSeq \[Subset])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}], ")"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                UnderscriptBox["isReducible", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}]}], 
                "]"}]], 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Element]", 
                   RowBox[{
                    UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
               "\[Subset]", 
               RowBox[{
                 UnderscriptBox["red", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}], 
                 "]"}]}], ")"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                UnderscriptBox["isReducible", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}]}], 
                "]"}]], 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Element]", 
                   RowBox[{
                    UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:602168712"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p602168712-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p602168712-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:602168712", "Proof|ID:602168712-1"},
 CellID->110870073],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p602168712-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->4965199]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1920006839],

Cell[CellGroupData[{

Cell["Lemma: Properties of \[SquareSubset]", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648282856627515*^9, 3.648282863995232*^9}},
 CellID->53798018],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "T", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", "a", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"T", 
       UnderscriptBox["\[SquareSubset]", "D"], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"T", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "a"}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "D"], 
         RowBox[{"[", "x", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{"\[Not]", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{"x", "\[Element]", 
            RowBox[{
             UnderscriptBox["red", "D"], "[", "T", "]"}]}], 
           TagBox[")",
            "AutoParentheses"]}]}], "\[And]", 
         RowBox[{
          UnderscriptBox["isReducible", "D"], "[", 
          RowBox[{"x", ",", "a"}], "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[SquareSubset] append)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:21782685", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648367612912772*^9, 3.648367706258986*^9}, {
   3.648380397403319*^9, 3.64838040371635*^9}, {3.648436581793035*^9, 
   3.648436585399592*^9}, {3.653633709875487*^9, 3.653633714758295*^9}, 
   3.6536339598663263`*^9},
 CellTags->{
  "ID:21782685", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] append", "IR"},
 CellID->21782685],

Cell[BoxData[
 RowBox[{"isPartIrreflOrder", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[SquareSubset]", "D"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   RowBox[{"DomainTuples", "[", "D", "]"}]}], "]"}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[SquareSubset] isPartIrreflOrder)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1790989103", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648283017795927*^9, 3.648283045266879*^9}, {
  3.648283652206025*^9, 3.6482836630306396`*^9}, {3.6536318078495283`*^9, 
  3.6536318138721385`*^9}, {3.65363185457461*^9, 3.6536318555262117`*^9}},
 CellTags->{
  "ID:1790989103", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isPartIrreflOrder"},
 CellID->1790989103],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->462091156]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->78191214],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[SquareSubset] append)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[SquareSubset] append)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], 
               UnderscriptBox["\[SquareSubset]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\:293a", 
                 StyleBox["a", "ExpressionVariable"]}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    RowBox[{
                    UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], 
               UnderscriptBox["\[SquareSubset]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\:293a", 
                 StyleBox["a", "ExpressionVariable"]}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    RowBox[{
                    UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:21782685"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p21782685-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p21782685-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:21782685", "Proof|ID:21782685-1"},
 CellID->202074869],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p21782685-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->637991929]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[SquareSubset] isPartIrreflOrder)\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(\[SquareSubset] isPartIrreflOrder)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isPartIrreflOrder", "[", 
           RowBox[{
             UnderscriptBox["\[SquareSubset]", 
              StyleBox["D", "ExpressionVariable"]], ",", 
             RowBox[{"DomainTuples", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isPartIrreflOrder", "[", 
           RowBox[{
             UnderscriptBox["\[SquareSubset]", 
              StyleBox["D", "ExpressionVariable"]], ",", 
             RowBox[{"DomainTuples", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1790989103"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1790989103-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1790989103-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1790989103", "Proof|ID:1790989103-1"},
 CellID->227724],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1790989103-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->482486095]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Reduction Rings (\[Checkmark])", "Section",
 CellChangeTimes->{{3.644214210293522*^9, 3.644214211904367*^9}, {
  3.647943107633806*^9, 3.647943111315255*^9}, {3.660736329529378*^9, 
  3.6607363320965247`*^9}},
 CellID->1684828724],

Cell[CellGroupData[{

Cell["Informal Explanation", "Subsection",
 CellChangeTimes->{{3.590830450392701*^9, 3.590830454213885*^9}},
 CellID->294463156],

Cell[TextData[{
 "In order for being a reduction ring, a domain ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " has to provide definitions for the following operations (besides the usual \
ring operations):"
}], "Text",
 CellChangeTimes->{{3.590834428115336*^9, 3.590834531416853*^9}, {
  3.590834604524359*^9, 3.590834606046285*^9}},
 CellID->1029021260],

Cell[CellGroupData[{

Cell["\<\
A noetherian partial order relation (\[OpenCurlyDoubleQuote]\[Precedes]\
\[CloseCurlyDoubleQuote]).\
\>", "Item",
 CellChangeTimes->{{3.596193869993015*^9, 3.59619387103018*^9}},
 CellID->1253702121],

Cell[TextData[{
 "A function that, for any ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 ", returns the number of the sets of multipliers for the respective ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " (\[OpenCurlyDoubleQuote]multN\[CloseCurlyDoubleQuote])."
}], "Item",
 CellChangeTimes->{{3.596193869993015*^9, 3.59619387685593*^9}, {
  3.631855029216742*^9, 3.631855038033951*^9}, {3.6318562970866327`*^9, 
  3.6318563040165854`*^9}},
 CellID->1260543835],

Cell[TextData[{
 "A function that decides for any ",
 Cell[BoxData[
  FormBox["m", TraditionalForm]]],
 " whether it is in a particular set of multipliers for a particular ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " (\[OpenCurlyDoubleQuote]mult\[CloseCurlyDoubleQuote])."
}], "Item",
 CellChangeTimes->{{3.596193869993015*^9, 3.59619388580169*^9}},
 CellID->1933641052]
}, Open  ]],

Cell["\<\
These operations need to satisfy several non-trivial properties.\
\>", "Text",
 CellChangeTimes->{{3.590834596637894*^9, 3.590834623998124*^9}},
 CellID->1148296278],

Cell[CellGroupData[{

Cell["Similarities and Differences to the Classical Setting", "Subsubsection",
 CellChangeTimes->{{3.651999745246394*^9, 3.651999752723225*^9}, {
  3.652002408586245*^9, 3.652002412140276*^9}},
 CellID->302444018],

Cell[TextData[{
 "In contrast to the classical setting, the following statements do ",
 StyleBox["not",
  FontWeight->"Bold"],
 " hold in reduction rings in general:"
}], "Text",
 CellChangeTimes->{{3.651999765945993*^9, 3.6519997892062063`*^9}},
 CellID->1904688891],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"irreducible", "[", 
       RowBox[{"a", ",", "C"}], "]"}], "\[And]", 
      RowBox[{"irreducible", "[", 
       RowBox[{"b", ",", "C"}], "]"}]}], ")"}], "\[Implies]", 
    RowBox[{"irreducible", "[", 
     RowBox[{
      RowBox[{"a", "+", "b"}], ",", "C"}], "]"}]}], TraditionalForm]]],
 "\[LineSeparator]Counterexample: \[DoubleStruckCapitalZ], ",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "=", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"2", "a"}], "+", 
      RowBox[{"2", "b"}]}], "}"}]}], TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", ">", "0"}], ",", 
    RowBox[{"b", ">", "0"}]}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.651999796688682*^9, 3.65199990384664*^9}, {
  3.652000063153989*^9, 3.652000063557549*^9}, {3.652001033675911*^9, 
  3.65200103368703*^9}, {3.652001131286046*^9, 3.6520011537668247`*^9}},
 CellID->622280239],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"irreducible", "[", 
       RowBox[{"a", ",", "C"}], "]"}], "\[And]", 
      RowBox[{"irreducible", "[", 
       RowBox[{"b", ",", "C"}], "]"}]}], ")"}], "\[Implies]", 
    RowBox[{"irreducible", "[", 
     RowBox[{
      RowBox[{"a", "-", "b"}], ",", "C"}], "]"}]}], TraditionalForm]]],
 "\[LineSeparator]Counterexample: \[DoubleStruckCapitalZ], ",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "=", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"2", "a"}], "-", 
      RowBox[{"2", "b"}]}], "}"}]}], TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", ">", "0"}], ",", 
    RowBox[{"b", "<", "0"}]}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.651999796688682*^9, 3.65199990384664*^9}, {
  3.652000063153989*^9, 3.652000063557549*^9}, {3.652001033675911*^9, 
  3.6520010712798433`*^9}, {3.652001109961419*^9, 3.652001112568102*^9}, {
  3.652001166145734*^9, 3.652001166656336*^9}, {3.652001197885714*^9, 
  3.6520012205995092`*^9}, {3.652001264116881*^9, 3.652001264116997*^9}},
 CellID->1089562638],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"irreducible", "[", 
     RowBox[{"a", ",", "C"}], "]"}], "\[Implies]", 
    RowBox[{"irreducible", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", "C"}], "]"}]}], TraditionalForm]]],
 "\t(which is equivalent to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"reducible", "[", 
     RowBox[{"a", ",", "C"}], "]"}], "\[Implies]", 
    RowBox[{"reducible", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", "C"}], "]"}]}], TraditionalForm]]],
 ")\[LineSeparator]Counterexample: \[DoubleStruckCapitalZ], ",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "=", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "2"}], "a"}], "}"}]}], TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "<", "0"}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.651999796688682*^9, 3.651999949357582*^9}, {
  3.652000011583709*^9, 3.652000011591022*^9}, {3.6520000660911303`*^9, 
  3.652000066424814*^9}, {3.6520003272235613`*^9, 3.652000386208878*^9}},
 CellID->1175767685],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"reducible", "[", 
       RowBox[{"a", ",", "C"}], "]"}], "\[And]", 
      RowBox[{"irreducible", "[", 
       RowBox[{"b", ",", "C"}], "]"}]}], ")"}], "\[Implies]", 
    RowBox[{"reducible", "[", 
     RowBox[{
      RowBox[{"a", "+", "b"}], ",", "C"}], "]"}]}], TraditionalForm]]],
 "\[LineSeparator]Counterexample: \[DoubleStruckCapitalZ], ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "=", 
    RowBox[{"-", "1"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "=", 
    RowBox[{"{", 
     RowBox[{"2", "a"}], "}"}]}], TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", ">", "1"}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.651999796688682*^9, 3.651999949357582*^9}, {
  3.652000011583709*^9, 3.652000074186408*^9}, {3.652000127442477*^9, 
  3.652000188254365*^9}, {3.6520002576661777`*^9, 3.6520002996982718`*^9}, {
  3.652000442449437*^9, 3.652000442449554*^9}, {3.652000494597303*^9, 
  3.6520004946085978`*^9}},
 CellID->1244683243]
}, Open  ]],

Cell["\<\
The following property is crucial for establishing an alternative \
characterization of Gr\[ODoubleDot]bner bases (namely that reducibility of \
every ideal-element to 0 implies being a Gr\[ODoubleDot]bner basis):\
\>", "Text",
 CellChangeTimes->{{3.652002431658999*^9, 3.652002475001795*^9}},
 CellID->1415435665],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"irreducible", "[", 
      RowBox[{"a", ",", "C"}], "]"}], "\[And]", 
     RowBox[{"irreducible", "[", 
      RowBox[{"b", ",", "C"}], "]"}], "\[And]", 
     RowBox[{"a", "\[NotEqual]", "b"}]}], ")"}], "\[Implies]", 
   RowBox[{"\[Not]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"a", "-", "b"}], 
      SubsuperscriptBox["\[Rule]", "C", "*"], "0"}], ")"}]}]}], 
  TraditionalForm]]]], "Item",
 CellChangeTimes->{{3.652002483664564*^9, 3.652002504385964*^9}},
 CellID->2032256535],

Cell[TextData[{
 "All known reduction rings (fields, \[DoubleStruckCapitalZ], ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "n"], TraditionalForm]]],
 ") appear to satisfy this property, although it is not clear whether it \
holds in general (most probably it doesn\[CloseCurlyQuote]t)."
}], "Text",
 CellChangeTimes->{{3.652002511798497*^9, 3.652002569771366*^9}},
 CellID->1158070040]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory (Correlativity and Irrelativity)", "Subsubsection",
 CellChangeTimes->{{3.6324705979561367`*^9, 3.632470598779188*^9}, {
  3.632471190603478*^9, 3.632471196595228*^9}, {3.632476951551058*^9, 
  3.632476954496819*^9}},
 CellID->1570628282],

Cell[TextData[{
 "For now, we restrict ourselves to the case of only two sets of multipliers \
for each ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[Element]", "\[ScriptCapitalR]"}], TraditionalForm]]],
 ", denoted by ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "+"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "-"], TraditionalForm]]],
 ". How these two sets are related to each other is not specified, and \
moreover one of them may even be empty. Fix now one particular element ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ", and omit the subscript ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " in the sets of multipliers of ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " for the sake of brevity. Furthermore, let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", ":=", 
    RowBox[{
     SuperscriptBox["M", "+"], "\[Union]", 
     SuperscriptBox["M", "-"]}]}], TraditionalForm]]],
 ".\nThere are essentially three different ways to define irrelativity of \
multipliers of ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.632470609350013*^9, 3.632470777126093*^9}, {
  3.632470809285089*^9, 3.632470834666709*^9}, {3.632471043351677*^9, 
  3.632471055220398*^9}, {3.6374731857238693`*^9, 3.6374731865460167`*^9}},
 CellID->1431522614],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"irrelative1", "[", 
     RowBox[{
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"]}], "]"}], ":", 
    RowBox[{"\[DoubleLeftRightArrow]", 
     RowBox[{"\[Or]", 
      RowBox[{"{", GridBox[{
         {
          RowBox[{
           RowBox[{
            SubscriptBox["m", "1"], "\[Element]", 
            SuperscriptBox["M", "+"]}], "\[And]", 
           RowBox[{
            SubscriptBox["m", "2"], "\[Element]", 
            SuperscriptBox["M", "-"]}]}]},
         {
          RowBox[{
           RowBox[{
            SubscriptBox["m", "1"], "\[Element]", 
            SuperscriptBox["M", "-"]}], "\[And]", 
           RowBox[{
            SubscriptBox["m", "2"], "\[Element]", 
            SuperscriptBox["M", "+"]}]}]}
        }]}]}]}]}], TraditionalForm]]],
 "\t\t[Stifter1985,Stifter1987]"
}], "Item",
 CellChangeTimes->{{3.632470751133271*^9, 3.632470805984898*^9}, {
  3.632470837707403*^9, 3.6324708566162643`*^9}, {3.637473237724826*^9, 
  3.637473252673073*^9}},
 CellID->46013314],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"irrelative2", "[", 
    RowBox[{
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "2"]}], "]"}], ":", 
   RowBox[{"\[DoubleLeftRightArrow]", 
    RowBox[{"\[Or]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{
          RowBox[{
           SubscriptBox["m", "1"], "\[NotElement]", 
           SuperscriptBox["M", "-"]}], "\[And]", 
          RowBox[{
           SubscriptBox["m", "2"], "\[NotElement]", 
           SuperscriptBox["M", "+"]}]}]},
        {
         RowBox[{
          RowBox[{
           SubscriptBox["m", "1"], "\[NotElement]", 
           SuperscriptBox["M", "+"]}], "\[And]", 
          RowBox[{
           SubscriptBox["m", "2"], "\[NotElement]", 
           SuperscriptBox["M", "-"]}]}]}
       }]}]}]}]}], TraditionalForm]]]], "Item",
 CellChangeTimes->{{3.6324708634532824`*^9, 3.63247088874859*^9}},
 CellID->828733833],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"irrelative3", "[", 
     RowBox[{
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"]}], "]"}], ":", 
    RowBox[{"\[DoubleLeftRightArrow]", 
     RowBox[{"\[Or]", 
      RowBox[{"{", GridBox[{
         {
          RowBox[{
           RowBox[{
            SubscriptBox["m", "1"], "\[Element]", 
            SuperscriptBox["M", "+"]}], "\[And]", 
           RowBox[{
            SubscriptBox["m", "2"], "\[NotElement]", 
            SuperscriptBox["M", "+"]}]}]},
         {
          RowBox[{
           RowBox[{
            SubscriptBox["m", "1"], "\[Element]", 
            SuperscriptBox["M", "-"]}], "\[And]", 
           RowBox[{
            SubscriptBox["m", "2"], "\[NotElement]", 
            SuperscriptBox["M", "-"]}]}]}
        }]}]}]}]}], TraditionalForm]]],
 "\t\t[Stifter1991]"
}], "Item",
 CellChangeTimes->{{3.6324708634532824`*^9, 3.63247088874859*^9}, {
  3.637473196211968*^9, 3.637473211034026*^9}, {3.637473254232699*^9, 
  3.63747326024226*^9}},
 CellID->1947028342]
}, Open  ]],

Cell[TextData[{
 "Obviously, both irrelative1 and irrelative2 are symmetric, but irrelative3 \
is not. Moreover, the three definitions are equivalent if ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SuperscriptBox["M", "+"], "\[Intersection]", 
     SuperscriptBox["M", "-"]}], "=", "\[EmptySet]"}], TraditionalForm]]],
 ", but they are ",
 StyleBox["not",
  FontSlant->"Italic"],
 " equivalent if ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "+"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "-"], TraditionalForm]]],
 " have non-empty intersection:"
}], "Text",
 CellChangeTimes->{{3.632470898788357*^9, 3.6324709960515547`*^9}, 
   3.632478188064231*^9, {3.637473222386351*^9, 3.637473230017363*^9}, {
   3.637473263851314*^9, 3.637473272006868*^9}, {3.637473412859571*^9, 
   3.637473418329451*^9}},
 CellID->1001914151],

Cell[TextData[{
 "Assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "\[Element]", 
    RowBox[{
     SuperscriptBox["M", "+"], "\[Intersection]", 
     SuperscriptBox["M", "-"]}]}], TraditionalForm]]],
 ", and let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "\[Element]", 
    RowBox[{
     SuperscriptBox["M", "+"], "\[CapitalDelta]", " ", 
     SuperscriptBox["M", "-"]}]}], TraditionalForm]]],
 " be arbitrary. Then we have ",
 Cell[BoxData[
  FormBox[
   RowBox[{"irrelative1", "[", 
    RowBox[{"m", ",", "n"}], "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"irrelative3", "[", 
    RowBox[{"m", ",", "n"}], "]"}], TraditionalForm]]],
 ", but not ",
 Cell[BoxData[
  FormBox[
   RowBox[{"irrelative2", "[", 
    RowBox[{"m", ",", "n"}], "]"}], TraditionalForm]]],
 " and not ",
 Cell[BoxData[
  FormBox[
   RowBox[{"irrelative3", "[", 
    RowBox[{"n", ",", "m"}], "]"}], TraditionalForm]]],
 ".\nMore precisely, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"irrelative2", "[", 
     RowBox[{
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"]}], "]"}], "\[Implies]", 
    RowBox[{"irrelative1", "[", 
     RowBox[{
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"]}], "]"}]}], TraditionalForm]]],
 " for all ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "1"], ",", 
    RowBox[{
     SubscriptBox["m", "2"], "\[Element]", "M"}]}], TraditionalForm]]],
 ", but not vice versa."
}], "Text",
 CellChangeTimes->{{3.632471005932097*^9, 3.632471106706225*^9}, {
  3.6324711380267277`*^9, 3.6324711837478447`*^9}, {3.632471222532104*^9, 
  3.63247123709787*^9}, {3.632472253178213*^9, 3.632472263884841*^9}, {
  3.632474430985641*^9, 3.632474432350051*^9}, {3.637473325657785*^9, 
  3.637473425578121*^9}},
 CellID->1110736610],

Cell[TextData[{
 "In order to prove the Main Theorem of Gr\[ODoubleDot]bner bases theory (in \
particular the case where ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " can be reduced modulo the same ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " in two irrelative ways), we need the following property of irrelativity:"
}], "Text",
 CellChangeTimes->{{3.632472292769813*^9, 3.632472320368992*^9}, {
  3.63247299095636*^9, 3.632473029872517*^9}},
 CellID->2043962507],

Cell[TextData[{
 "(P1)\tFor ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "\[Element]", "M"}], TraditionalForm]]],
 " completely arbitrary: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"irrelative", "[", 
     RowBox[{
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"]}], "]"}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Not]", 
       RowBox[{"irrelative", "[", 
        RowBox[{
         SubscriptBox["m", "1"], ",", "m"}], "]"}]}], "\[Or]", 
      RowBox[{"\[Not]", 
       RowBox[{"irrelative", "[", 
        RowBox[{
         SubscriptBox["m", "2"], ",", "m"}], "]"}]}]}], ")"}]}], 
   TraditionalForm]]]
}], "Text",
 CellChangeTimes->{{3.632472325050742*^9, 3.632472407189702*^9}, {
  3.632472464756208*^9, 3.632472498325191*^9}, {3.632473049710965*^9, 
  3.632473051958815*^9}, {3.637466829072178*^9, 3.637466838127902*^9}},
 CellID->1685290468],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox["irrelative2", TraditionalForm]]],
 " is taken as the definition of irrelativity, then (P1) clearly holds. We \
claim that it does not hold if ",
 Cell[BoxData[
  FormBox["irrelative1", TraditionalForm]]],
 " or irrelative3 is taken. For this, we have to find ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", ",", 
    SubscriptBox["m", "1"], ",", 
    RowBox[{
     SubscriptBox["m", "2"], "\[Element]", "M"}]}], TraditionalForm]]],
 " such that the LHS of the implication is True, but the RHS is not."
}], "Text",
 CellChangeTimes->{{3.632472480781581*^9, 3.632472524665881*^9}, {
   3.632472604026494*^9, 3.632472793828465*^9}, {3.63247285285157*^9, 
   3.632472905110487*^9}, {3.637473449741044*^9, 3.637473454255745*^9}, {
   3.637473568793392*^9, 3.637473573013216*^9}, 3.637473606896513*^9},
 CellID->31426284],

Cell[CellGroupData[{

Cell[TextData[{
 "irrelative2: We are free to construct ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "+"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "-"], TraditionalForm]]],
 " in whatever way we want, so let ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "+"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "-"], TraditionalForm]]],
 " be such that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SuperscriptBox["M", "+"], "\[Intersection]", 
     SuperscriptBox["M", "-"]}], "\[NotEqual]", "\[EmptySet]"}], 
   TraditionalForm]]],
 ", and let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "\[Element]", 
    RowBox[{
     SuperscriptBox["M", "+"], "\[Intersection]", 
     SuperscriptBox["M", "-"]}]}], TraditionalForm]]],
 ". No matter how we choose ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 ", the RHS will never be satisfied, so we are done if we manage to find ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "1"], ",", 
    SubscriptBox["m", "2"]}], TraditionalForm]]],
 " that satisfy the LHS; This, however, is always possible under the \
restriction ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SuperscriptBox["M", "+"], "\[Intersection]", 
     SuperscriptBox["M", "-"]}], "\[NotEqual]", "\[EmptySet]"}], 
   TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{
  3.637473601192521*^9, {3.637473633166843*^9, 3.637473633297988*^9}},
 CellID->1079707726],

Cell[TextData[{
 "irrelative3: Again we are free to construct ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "+"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "-"], TraditionalForm]]],
 " in whatever way we want, so let ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "+"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "-"], TraditionalForm]]],
 " be such that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SuperscriptBox["M", "+"], "\[Intersection]", 
     SuperscriptBox["M", "-"]}], "\[NotEqual]", "\[EmptySet]"}], 
   TraditionalForm]]],
 ", let ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "1"], "\[Element]", 
    RowBox[{
     SuperscriptBox["M", "+"], "\[Intersection]", 
     SuperscriptBox["M", "-"]}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "2"], "\[Element]", 
    RowBox[{"M", "\\", 
     SuperscriptBox["M", "+"]}]}], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "\[Element]", 
    RowBox[{"M", "\\", 
     SuperscriptBox["M", "-"]}]}], TraditionalForm]]],
 ". Then the LHS of the implication holds, but we also have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"irrelative3", "[", 
     RowBox[{
      SubscriptBox["m", "1"], ",", "m"}], "]"}], "\[And]", 
    RowBox[{"irrelative3", "[", 
     RowBox[{
      SubscriptBox["m", "2"], ",", "m"}], "]"}]}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.637473616575686*^9, 3.6374737856127377`*^9}, {
  3.6374739975038233`*^9, 3.637474001422217*^9}},
 CellID->1281898387]
}, Open  ]],

Cell[TextData[{
 "The preceding observation proves that ",
 Cell[BoxData[
  FormBox["irrelative1", TraditionalForm]]],
 " and irrelative3 cannot be taken as the definition of irrelativity, \
although the problem arising with irrelative3 could be fixed by replacing the \
RHS of the implication by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{"\[Not]", 
      RowBox[{"irrelative", "[", 
       RowBox[{
        SubscriptBox["m", "1"], ",", "m"}], "]"}]}], "\[Or]", 
     RowBox[{"\[Not]", 
      RowBox[{"irrelative", "[", 
       RowBox[{"m", ",", 
        SubscriptBox["m", "1"]}], "]"}]}], "\[Or]", 
     RowBox[{"\[Not]", 
      RowBox[{"irrelative", "[", 
       RowBox[{
        SubscriptBox["m", "2"], ",", "m"}], "]"}]}], "\[Or]", 
     RowBox[{"\[Not]", 
      RowBox[{"irrelative", "[", 
       RowBox[{"m", ",", 
        SubscriptBox["m", "2"]}], "]"}]}]}], ")"}], TraditionalForm]]],
 ", caused by irrelative3 not being symmetric."
}], "Text",
 CellChangeTimes->{{3.632472942469831*^9, 3.632472984835682*^9}, 
   3.6374680761036463`*^9, {3.637473822230337*^9, 3.637473910943771*^9}},
 CellID->2095286757],

Cell["\<\
In order to prove the conservation theorem for polynomial domains (in \
particular, (CR1)), we need the following property of irrelativity:\
\>", "Text",
 CellChangeTimes->{{3.632471428986771*^9, 3.6324714578176517`*^9}, {
  3.632471942426066*^9, 3.632471959266433*^9}},
 CellID->695542827],

Cell[TextData[{
 "(P2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Not]", 
       RowBox[{"irrelative", "[", 
        RowBox[{
         SubscriptBox["m", "1"], ",", 
         SubscriptBox["m", "2"]}], "]"}]}], "\[And]", 
      RowBox[{"irrelative", "[", 
       RowBox[{
        SubscriptBox["m", "3"], ",", 
        SubscriptBox["m", "4"]}], "]"}]}], ")"}], "\[Implies]", 
    RowBox[{"\[Or]", 
     RowBox[{"{", GridBox[{
        {
         RowBox[{
          RowBox[{"irrelative", "[", 
           RowBox[{
            SubscriptBox["m", "1"], ",", 
            SubscriptBox["m", "3"]}], "]"}], "\[And]", 
          RowBox[{"irrelative", "[", 
           RowBox[{
            SubscriptBox["m", "2"], ",", 
            SubscriptBox["m", "3"]}], "]"}]}]},
        {
         RowBox[{
          RowBox[{"irrelative", "[", 
           RowBox[{
            SubscriptBox["m", "1"], ",", 
            SubscriptBox["m", "4"]}], "]"}], "\[And]", 
          RowBox[{"irrelative", "[", 
           RowBox[{
            SubscriptBox["m", "2"], ",", 
            SubscriptBox["m", "4"]}], "]"}]}]}
       }]}]}]}], TraditionalForm]]]
}], "Text",
 CellChangeTimes->{{3.632471483963945*^9, 3.632471545137059*^9}, 
   3.632472499394841*^9, {3.632473117061535*^9, 3.632473145538702*^9}, {
   3.637474723009253*^9, 3.637474774645102*^9}, 3.647144729965804*^9},
 CellID->1113844059],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox["irrelative1", TraditionalForm]]],
 " is taken as the definition of irrelativity, then (P2) clearly holds. We \
claim that it does not hold if ",
 Cell[BoxData[
  FormBox["irrelative2", TraditionalForm]]],
 " is taken. For this, we have to find ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "1"], ",", 
    SubscriptBox["m", "2"], ",", 
    SubscriptBox["m", "3"], ",", 
    RowBox[{
     SubscriptBox["m", "4"], "\[Element]", "M"}]}], TraditionalForm]]],
 " such that the LHS of the implication is True, but the RHS is not. We are \
also free to construct ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "+"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "-"], TraditionalForm]]],
 " in whatever way we want, so let ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "+"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["M", "-"], TraditionalForm]]],
 " be such that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SuperscriptBox["M", "+"], "\[Intersection]", 
     SuperscriptBox["M", "-"]}], "\[NotEqual]", "\[EmptySet]"}], 
   TraditionalForm]]],
 ", and let ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "1"], ",", 
    RowBox[{
     SubscriptBox["m", "2"], "\[Element]", 
     RowBox[{
      SuperscriptBox["M", "+"], "\[Intersection]", 
      SuperscriptBox["M", "-"]}]}]}], TraditionalForm]]],
 ". No matter how we choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "3"], ",", 
    SubscriptBox["m", "4"]}], TraditionalForm]]],
 ", we will certainly have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     RowBox[{"irrelative2", "[", 
      RowBox[{
       SubscriptBox["m", "1"], ",", 
       SubscriptBox["m", "3"]}], "]"}]}], "\[And]", 
    RowBox[{"\[Not]", 
     RowBox[{"irrelative2", "[", 
      RowBox[{
       SubscriptBox["m", "2"], ",", 
       SubscriptBox["m", "3"]}], "]"}]}], "\[And]", 
    RowBox[{"\[Not]", 
     RowBox[{"irrelative2", "[", 
      RowBox[{
       SubscriptBox["m", "1"], ",", 
       SubscriptBox["m", "4"]}], "]"}]}], "\[And]", 
    RowBox[{"\[Not]", 
     RowBox[{"irrelative2", "[", 
      RowBox[{
       SubscriptBox["m", "2"], ",", 
       SubscriptBox["m", "4"]}], "]"}]}]}], TraditionalForm]]],
 ", and hence the RHS of the implication certainly fails. Therefore, if ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "3"], ",", 
    SubscriptBox["m", "4"]}], TraditionalForm]]],
 " satisfy ",
 Cell[BoxData[
  FormBox[
   RowBox[{"irrelative2", "[", 
    RowBox[{
     SubscriptBox["m", "3"], ",", 
     SubscriptBox["m", "4"]}], "]"}], TraditionalForm]]],
 ", the LHS is satisfied and we are done. Choosing ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "3"], ",", 
    SubscriptBox["m", "4"]}], TraditionalForm]]],
 " in such a way is always possible if ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["M", "+"], "\[NotSubsetEqual]", 
    SuperscriptBox["M", "-"]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["M", "-"], "\[NotSubsetEqual]", 
    SuperscriptBox["M", "+"]}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.632471585291588*^9, 3.632471607540926*^9}, {
   3.632471646107945*^9, 3.6324718993543873`*^9}, {3.6324719689996557`*^9, 
   3.632472022381871*^9}, {3.632472063806381*^9, 3.632472223556855*^9}, {
   3.632472420997775*^9, 3.6324724221106367`*^9}, 3.632472501166349*^9, {
   3.632473162749934*^9, 3.6324732221332407`*^9}, {3.6324732596538897`*^9, 
   3.632473268441567*^9}, {3.637474789818211*^9, 3.637474971835849*^9}},
 CellID->154472253],

Cell[TextData[{
 "Summarizing, neither of the three definitions of irrelativity given above \
is feasible for our purpose if ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubsuperscriptBox["M", "c", "+"], "\[Intersection]", 
     SubsuperscriptBox["M", "c", "-"]}], "\[NotEqual]", "\[EmptySet]"}], 
   TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubsuperscriptBox["M", "c", "+"], "\[NotSubsetEqual]", 
    SubsuperscriptBox["M", "c", "-"]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubsuperscriptBox["M", "c", "-"], "\[NotSubsetEqual]", 
    SubsuperscriptBox["M", "c", "+"]}], TraditionalForm]]],
 " for some ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[Element]", "\[ScriptCapitalR]"}], TraditionalForm]]],
 ", if the reduction ring axioms are as in [Stifter1985, Stifter1987]."
}], "Text",
 CellChangeTimes->{{3.632473281318273*^9, 3.6324733692487917`*^9}, {
   3.6324734698005667`*^9, 3.632473522325989*^9}, {3.632474276733259*^9, 
   3.6324743495213747`*^9}, {3.632474541427372*^9, 3.632474572336375*^9}, 
   3.632478197718841*^9, {3.637478856050808*^9, 3.637478856728457*^9}},
 CellID->1441754918],

Cell[TextData[{
 "Instead, we use a combination of ",
 Cell[BoxData[
  FormBox["irrelative1", TraditionalForm]]],
 " and irrelative2: Our definition of irrelativity is essentially the one of \
irrelative1, and we replace condition ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"areIrrelative", "[", 
     RowBox[{"c", ",", "c", ",", 
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"]}], "]"}]}], TraditionalForm]]],
 " in (R6) by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCorrelative", "[", 
    RowBox[{"c", ",", 
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "2"]}], "]"}], TraditionalForm]]],
 ", defined as\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"areCorrelative", "[", 
     RowBox[{"c", ",", 
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"]}], "]"}], ":", 
    RowBox[{
     RowBox[{"\[DoubleLeftRightArrow]", 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"multN", "[", "c", "]"}]}]], 
       RowBox[{"mult", "[", 
        RowBox[{
         SubscriptBox["m", "1"], ",", "c", ",", "i"}], "]"}]}]}], "\[And]", 
     RowBox[{"mult", "[", 
      RowBox[{
       SubscriptBox["m", "2"], ",", "c", ",", "i"}], "]"}]}]}], 
   TraditionalForm]]],
 ".\nNote that areCorrelative is precisely the negation of irrelative2."
}], "Text",
 CellChangeTimes->{{3.632474587791407*^9, 3.6324746943100758`*^9}, {
  3.632476966809523*^9, 3.632477041249119*^9}, {3.632477205651841*^9, 
  3.632477216932199*^9}, {3.63747888208074*^9, 3.637479040198029*^9}},
 CellID->25744908]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Reduction Rings", "Subsection",
 CellChangeTimes->{{3.590833926490018*^9, 3.590833929071406*^9}, {
   3.610168448718971*^9, 3.610168450360519*^9}, 3.644214225808301*^9},
 CellID->2098927943],

Cell[CellGroupData[{

Cell["Definition", "Subsubsection",
 CellChangeTimes->{{3.590830444116878*^9, 3.59083044594691*^9}, {
   3.6363563182460403`*^9, 3.636356321228923*^9}, {3.6430099576562157`*^9, 
   3.643009959892662*^9}, 3.644214218285945*^9},
 CellID->1987403105],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "D"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1466856847}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590816410142059*^9, 3.590816411583155*^9}},
 CellTags->{
  "ID:1466856847", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1466856847],

Cell["", "OpenEnvironment",
 CellID->1327698689],

Cell[CellGroupData[{

Cell["Definition: Reduction Ring Category", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.590816377343911*^9, 3.590816384480017*^9}, 
   3.590833970394244*^9, {3.591602030998711*^9, 3.59160203263217*^9}},
 CellID->1293548764],

Cell[BoxData[
 RowBox[{
  RowBox[{"isReductionRing", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"isCommRing1", "[", "D", "]"}]},
       {GridBox[{
          {
           RowBox[{"isPartIrreflOrder", "[", 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              TagBox[
               UnderscriptBox["\[Succeeds]", "D"],
               Identity], 
              TagBox[")",
               "AutoParentheses"]}], ",", "D"}], "]"}]},
          {
           RowBox[{"isNoetherian", "[", 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              TagBox[
               UnderscriptBox["\[Succeeds]", "D"],
               Identity], 
              TagBox[")",
               "AutoParentheses"]}], ",", "D"}], "]"}]}
         }]},
       {
        RowBox[{"Multipliers", "[", "D", "]"}]},
       {
        RowBox[{"R0", "[", "D", "]"}]},
       {
        RowBox[{"R1", "[", "D", "]"}]},
       {
        RowBox[{"R2", "[", "D", "]"}]},
       {
        RowBox[{"R3", "[", "D", "]"}]},
       {
        RowBox[{"R4", "[", "D", "]"}]},
       {
        RowBox[{"R5", "[", "D", "]"}]},
       {
        RowBox[{"R6", "[", "D", "]"}]},
       {
        RowBox[{"R7", "[", "D", "]"}]},
       {
        RowBox[{"R8", "[", "D", "]"}]}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduction ring)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:623701544", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590914856241332*^9, 3.590914941261603*^9}, {
   3.590916438583548*^9, 3.590916460712998*^9}, 3.59194447978519*^9, {
   3.592026215364363*^9, 3.59202624229187*^9}, {3.596190591094863*^9, 
   3.596190600045938*^9}, {3.607336877390878*^9, 3.607336878719672*^9}, {
   3.607772554232501*^9, 3.607772556449706*^9}, {3.607773285281593*^9, 
   3.607773285572973*^9}, {3.610177355474439*^9, 3.610177364007452*^9}, {
   3.636356474890114*^9, 3.636356479151101*^9}},
 CellTags->{
  "ID:623701544", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction ring"},
 CellID->623701544],

Cell[BoxData[
 RowBox[{
  RowBox[{"Multipliers", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", 
        RowBox[{"m", ",", "c"}], "]"}]}]], 
     RowBox[{"c", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{
        UnderscriptBox["multN", "D"], "[", "c", "]"}], "\[Element]", 
       "\[DoubleStruckCapitalN]"}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["mult", "D"], "[", 
         RowBox[{"m", ",", "c"}], "]"}], 
        TagBox["\[DoubleLeftRightArrow]",
         Identity,
         SyntaxForm->"a\[Implies]b"], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          UnderscriptBox["\[Exists]", 
           RowBox[{
            RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
            RowBox[{
             UnderscriptBox["multN", "D"], "[", "c", "]"}]}]], 
          RowBox[{
           UnderscriptBox["mult", "D"], "[", 
           RowBox[{"m", ",", "c", ",", "i"}], "]"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(multipliers)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1960646232", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590912566390148*^9, 3.590912611263788*^9}, 
   3.594029134411444*^9, 3.595049735390051*^9, {3.607336882553126*^9, 
   3.607336886133701*^9}, {3.631855050564973*^9, 3.6318550537652884`*^9}, {
   3.6318563220915194`*^9, 3.631856327473632*^9}, {3.6318720354004593`*^9, 
   3.6318720426937857`*^9}, {3.638182776228177*^9, 3.638182826177147*^9}},
 CellTags->{
  "ID:1960646232", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "multipliers"},
 CellID->1960646232],

Cell[BoxData[
 RowBox[{
  RowBox[{"R0", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], "[", "c", "]"}]], 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox["mult", "D"], "[", 
     RowBox[{
      UnderscriptBox["1", "D"], ",", "c"}], "]"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1496065688", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59082767649391*^9, 3.590827692841985*^9}, 
   3.594029138425726*^9, {3.607336901695672*^9, 3.607336946108702*^9}, {
   3.631872048537188*^9, 3.63187205763833*^9}},
 CellTags->{
  "ID:1496065688", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R0"},
 CellID->1496065688],

Cell[BoxData[
 RowBox[{
  RowBox[{"R1", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], "[", 
      RowBox[{"c", ",", "m"}], "]"}]], 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["mult", "D"], "[", 
      RowBox[{"m", ",", "c"}], "]"}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["mult", "D"], "[", 
      RowBox[{
       RowBox[{
        UnderscriptBox["-", "D"], "m"}], ",", "c"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:975892043", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590827723314676*^9, 3.590827735701109*^9}, 
   3.594029141078239*^9, {3.607336948508115*^9, 3.607336951959603*^9}, {
   3.631872084793557*^9, 3.6318720887657948`*^9}},
 CellTags->{
  "ID:975892043", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R1"},
 CellID->975892043],

Cell[BoxData[
 RowBox[{
  RowBox[{"R2", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], "[", 
      RowBox[{"c", ",", "m"}], "]"}]], 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["mult", "D"], "[", 
      RowBox[{"m", ",", "c"}], "]"}], "\[Implies]", 
     RowBox[{
      RowBox[{"m", 
       UnderscriptBox["*", "D"], "c"}], "\[NotEqual]", 
      UnderscriptBox["0", "D"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1150364990", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590827739833551*^9, 3.590827759900589*^9}, 
   3.594029144725189*^9, {3.607336957419387*^9, 3.607336994925279*^9}, {
   3.631872068515491*^9, 3.63187207821406*^9}},
 CellTags->{
  "ID:1150364990", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R2"},
 CellID->1150364990],

Cell[BoxData[
 RowBox[{
  RowBox[{"R3", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", 
       RowBox[{"c", ",", "b"}], "]"}]}]], 
    RowBox[{
     RowBox[{"c", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}], "\[And]", 
     RowBox[{"b", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}]}]], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
        RowBox[{
         UnderscriptBox["mult", "D"], "[", 
         RowBox[{
          SubscriptBox["T", "i"], ",", "c"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox[
          SubscriptBox["\[Sum]", "D"], 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
         
         SubscriptBox["T", "i"]}], "\[Equal]", "b"}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:852396718", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.5908176763395*^9, 3.590817713187588*^9}, {
  3.590817754345135*^9, 3.590817775643536*^9}, {3.590817808153972*^9, 
  3.590818000475322*^9}, {3.590818036212182*^9, 3.590818037018678*^9}, {
  3.590827773947739*^9, 3.59082777424367*^9}, {3.594029149483623*^9, 
  3.594029154911278*^9}, {3.596190624765106*^9, 3.596190641453294*^9}, {
  3.607337010163348*^9, 3.607337063008428*^9}, {3.610169628489538*^9, 
  3.610169635933027*^9}, {3.631872103880087*^9, 3.631872130540872*^9}, {
  3.6357487848220453`*^9, 3.635748791493306*^9}},
 CellTags->{
  "ID:852396718", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R3"},
 CellID->852396718],

Cell[BoxData[
 RowBox[{
  RowBox[{"R4", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], "[", "a", "]"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"a", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}], "\[Implies]", 
     RowBox[{"a", 
      UnderscriptBox["\[Succeeds]", "D"], 
      UnderscriptBox["0", "D"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1907869025", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590828306773658*^9, 3.59082831067903*^9}, {
   3.590828341320677*^9, 3.59082834643228*^9}, 3.590915740718952*^9, {
   3.59194453789979*^9, 3.591944543608866*^9}, 3.594029174303813*^9, {
   3.607337112677334*^9, 3.607337122958963*^9}},
 CellTags->{
  "ID:1907869025", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R4"},
 CellID->1907869025],

Cell[BoxData[
 RowBox[{
  RowBox[{"R5", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", 
      RowBox[{"a", ",", "m", ",", "c", ",", "b"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", 
       UnderscriptBox[
        SubscriptBox["\[Rule]", 
         RowBox[{"m", ",", "c"}]], "D"], "b"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "D"], "[", "d", "]"}]], 
       RowBox[{
        UnderscriptBox["\[Exists]", 
         RowBox[{
          UnderscriptBox["\[Element]", 
           RowBox[{"DomainTuples", "[", "D", "]"}]], 
          RowBox[{"[", 
           RowBox[{"TM", ",", "TN"}], "]"}]}]], 
        RowBox[{
         UnderscriptBox["areR5Witnesses", "D"], "[", 
         RowBox[{
         "TM", ",", "TN", ",", "a", ",", "b", ",", "c", ",", "d", ",", "m"}], 
         "]"}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1446921308", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59082779822597*^9, 3.590827805489401*^9}, 
   3.590827845291877*^9, {3.590827877878628*^9, 3.590827931502085*^9}, {
   3.590830193582971*^9, 3.59083021171342*^9}, {3.594029178020868*^9, 
   3.594029180748637*^9}, {3.596190693847209*^9, 3.596190703382925*^9}, {
   3.607337136730736*^9, 3.607337237311456*^9}, {3.610169643451463*^9, 
   3.610169697136182*^9}, {3.632028612600399*^9, 3.63202861271229*^9}, {
   3.632028659319145*^9, 3.632028680927989*^9}, 3.632030238225814*^9, {
   3.632041392067081*^9, 3.632041392504447*^9}, {3.6320494009614487`*^9, 
   3.632049438084448*^9}, {3.635748798080508*^9, 3.63574881142143*^9}},
 CellTags->{
  "ID:1446921308", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R5"},
 CellID->1446921308],

Cell[BoxData[
 RowBox[{
  RowBox[{"R6", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", 
      RowBox[{"a", ",", "c", ",", "m1", ",", "m2", ",", "b1", ",", "b2"}], 
      "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", 
           RowBox[{"m1", ",", "c"}]], "D"], "b1"}], 
        TagBox[")",
         "AutoParentheses"]}], "\[And]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", 
           RowBox[{"m2", ",", "c"}]], "D"], "b2"}], 
        TagBox[")",
         "AutoParentheses"]}], "\[And]", 
       RowBox[{
        UnderscriptBox["areCorrelative", "D"], "[", 
        RowBox[{"c", ",", "m1", ",", "m2"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", "D", "]"}]], 
        RowBox[{"[", "TL", "]"}]}]], 
      RowBox[{
       UnderscriptBox["isR6Witness", "D"], "[", 
       RowBox[{
       "TL", ",", "a", ",", "c", ",", "m1", ",", "m2", ",", "b1", ",", "b2"}],
        "]"}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R6)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:208630606", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590827920937381*^9, 3.59082795725519*^9}, {
   3.590830228258962*^9, 3.590830240254815*^9}, 3.590915764861566*^9, {
   3.591944597510057*^9, 3.591944616661698*^9}, {3.5940291950311*^9, 
   3.594029196396502*^9}, {3.596190720307856*^9, 3.596190727583358*^9}, {
   3.607337277695805*^9, 3.607337367419151*^9}, {3.608524493411508*^9, 
   3.608524502664451*^9}, {3.610169705952064*^9, 3.610169767303485*^9}, {
   3.6320287400657797`*^9, 3.63202880161675*^9}, {3.632041476645174*^9, 
   3.6320414770487757`*^9}, {3.632049108051396*^9, 3.6320491323772087`*^9}, {
   3.632477073795972*^9, 3.632477079583144*^9}, {3.6324771943823633`*^9, 
   3.632477194679241*^9}, {3.635748817055213*^9, 3.635748823971527*^9}},
 CellTags->{
  "ID:208630606", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6"},
 CellID->208630606],

Cell[BoxData[
 RowBox[{
  RowBox[{"R7", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], "[", 
     RowBox[{"c1", ",", "c2", ",", "a"}], "]"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{
        UnderscriptBox["multN", "D"], "[", "c1", "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        UnderscriptBox["isNTCR", "D"], "[", 
        RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
       "\[Implies]", 
       RowBox[{
        UnderscriptBox["\[Exists]", 
         RowBox[{"k", ",", 
          RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{
           UnderscriptBox["multN", "D"], "[", "c1", "]"}]}]], 
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{
           UnderscriptBox["\[Element]", "D"], "[", 
           RowBox[{"a0", ",", "m"}], "]"}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"\[And]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{
                RowBox[{"a", 
                 UnderscriptBox["\[Succeeds]", "D"], "a0"}], "\[Or]", 
                RowBox[{"a0", "\[Equal]", "a"}]}]},
              {
               RowBox[{
                UnderscriptBox["isMNTCR", "D"], "[", 
                RowBox[{"c1", ",", "c2", ",", "a0", ",", "k", ",", "l"}], 
                "]"}]},
              {
               RowBox[{
                UnderscriptBox["\[ForAll]", 
                 RowBox[{
                  UnderscriptBox["\[Element]", "D"], "[", 
                  RowBox[{"m0", ",", "c"}], "]"}]], 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{
                  RowBox[{
                   TagBox["(",
                    "AutoParentheses"], 
                   RowBox[{
                    RowBox[{"c", "\[NotEqual]", 
                    UnderscriptBox["0", "D"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "D"], "[", 
                    RowBox[{"m0", ",", "c"}], "]"}]}], 
                   TagBox[")",
                    "AutoParentheses"]}], "\[Implies]", 
                  RowBox[{
                   UnderscriptBox["mult", "D"], "[", 
                   RowBox[{
                    RowBox[{"m", 
                    UnderscriptBox["*", "D"], "m0"}], ",", "c"}], "]"}]}], 
                 TagBox[")",
                  "AutoParentheses"]}]}]},
              {
               RowBox[{
                UnderscriptBox["\[ForAll]", 
                 RowBox[{
                  UnderscriptBox["\[Element]", "D"], "[", "c", "]"}]], 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{
                  RowBox[{"a0", 
                   UnderscriptBox["\[Succeeds]", "D"], 
                   RowBox[{"a0", 
                    UnderscriptBox["+", "D"], "c"}]}], "\[Implies]", 
                  RowBox[{"a", 
                   UnderscriptBox["\[Succeeds]", "D"], 
                   RowBox[{"a", 
                    UnderscriptBox["+", "D"], 
                    RowBox[{"m", 
                    UnderscriptBox["*", "D"], "c"}]}]}]}], 
                 TagBox[")",
                  "AutoParentheses"]}]}]},
              {
               RowBox[{
                UnderscriptBox["\[ForAll]", 
                 RowBox[{
                  UnderscriptBox["\[Element]", "D"], 
                  RowBox[{"[", 
                   RowBox[{"b", ",", "d", ",", "c"}], "]"}]}]], 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{
                  RowBox[{
                   TagBox["(",
                    "AutoParentheses"], 
                   TagBox[
                    RowBox[{"b", 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{"{", "c", "}"}]], "D"], "d"}],
                    Identity], 
                   TagBox[")",
                    "AutoParentheses"]}], "\[Implies]", 
                  RowBox[{
                   TagBox["(",
                    "AutoParentheses"], 
                   TagBox[
                    RowBox[{
                    RowBox[{"m", 
                    UnderscriptBox["*", "D"], "b"}], 
                    UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    RowBox[{"{", "c", "}"}]], "D"], 
                    RowBox[{"m", 
                    UnderscriptBox["*", "D"], "d"}]}],
                    Identity], 
                   TagBox[")",
                    "AutoParentheses"]}]}], 
                 TagBox[")",
                  "AutoParentheses"]}]}]},
              {
               RowBox[{
                RowBox[{"c1", "\[Equal]", "c2"}], "\[Implies]", 
                RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  RowBox[{
                   UnderscriptBox["\[Element]", "D"], "[", 
                   RowBox[{"m1", ",", "m2"}], "]"}]], 
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  RowBox[{
                   RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["isReducible", "D"], "[", 
                    RowBox[{"a0", ",", "c1", ",", "m1"}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["isReducible", "D"], "[", 
                    RowBox[{"a0", ",", "c1", ",", "m2"}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "D"], "[", 
                    RowBox[{"m1", ",", "c1", ",", "k"}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "D"], "[", 
                    RowBox[{"m2", ",", "c1", ",", "l"}], "]"}]}], 
                    TagBox[")",
                    "AutoParentheses"]}], "\[Implies]", 
                   RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", "D"], "[", 
                    RowBox[{
                    RowBox[{"m", 
                    UnderscriptBox["*", "D"], "m1"}], ",", "c1", ",", "i"}], 
                    "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "D"], "[", 
                    RowBox[{
                    RowBox[{"m", 
                    UnderscriptBox["*", "D"], "m2"}], ",", "c1", ",", "j"}], 
                    "]"}]}], 
                    TagBox[")",
                    "AutoParentheses"]}], "\[Or]", 
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", "D"], "[", 
                    RowBox[{
                    RowBox[{"m", 
                    UnderscriptBox["*", "D"], "m1"}], ",", "c1", ",", "j"}], 
                    "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", "D"], "[", 
                    RowBox[{
                    RowBox[{"m", 
                    UnderscriptBox["*", "D"], "m2"}], ",", "c1", ",", "i"}], 
                    "]"}]}], 
                    TagBox[")",
                    "AutoParentheses"]}]}], 
                    TagBox[")",
                    "AutoParentheses"]}]}], 
                  TagBox[")",
                   "AutoParentheses"]}]}]}]}
             }]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R7)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1150983847", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590827970693857*^9, 3.590827986989734*^9}, {
   3.590830254454672*^9, 3.590830295947076*^9}, {3.59091577565007*^9, 
   3.590915783252063*^9}, {3.591944634324399*^9, 3.591944683937794*^9}, {
   3.594029209720809*^9, 3.594029232047722*^9}, {3.6073374036117*^9, 
   3.607337480515329*^9}, {3.608549091283777*^9, 3.608549110593839*^9}, {
   3.608549142714011*^9, 3.608549143601879*^9}, {3.608549228807837*^9, 
   3.608549229552041*^9}, {3.609065056738164*^9, 3.609065100353201*^9}, {
   3.60906527810286*^9, 3.609065349041079*^9}, {3.609065401045016*^9, 
   3.609065482797038*^9}, {3.609065571268566*^9, 3.609065624352584*^9}, {
   3.609065680911134*^9, 3.609065691310642*^9}, {3.609494999417832*^9, 
   3.609495060279116*^9}, {3.610169777350723*^9, 3.610169873093164*^9}, {
   3.610178559683448*^9, 3.610178589330085*^9}, {3.610178863882592*^9, 
   3.610178906457148*^9}, 3.610178945195517*^9, {3.631855079512294*^9, 
   3.631855082969171*^9}, {3.631855143627563*^9, 3.631855147573944*^9}, {
   3.631856336867778*^9, 3.631856355567259*^9}, {3.6324798149522047`*^9, 
   3.632479843169868*^9}, {3.6324798954240503`*^9, 3.632479899570879*^9}, {
   3.63248121384838*^9, 3.6324812145014267`*^9}, {3.632546085025765*^9, 
   3.6325461042527447`*^9}, {3.632546147314823*^9, 3.632546154202518*^9}, {
   3.6325500089166403`*^9, 3.632550010794466*^9}, {3.632809639560814*^9, 
   3.632809671284173*^9}, {3.632809848425926*^9, 3.6328098510952578`*^9}, {
   3.6618501514495664`*^9, 3.661850157112376*^9}},
 CellTags->{
  "ID:1150983847", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R7"},
 CellID->1150983847],

Cell[BoxData[
 RowBox[{
  RowBox[{"R8", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", 
      RowBox[{"c0", ",", "a"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["isNTCR", "D"], "[", 
      RowBox[{"c0", ",", "a"}], "]"}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], "[", 
        RowBox[{"a0", ",", "m"}], "]"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             RowBox[{"a", 
              UnderscriptBox["\[Succeeds]", "D"], "a0"}], "\[Or]", 
             RowBox[{"a0", "\[Equal]", "a"}]}]},
           {
            RowBox[{
             UnderscriptBox["isMNTCR", "D"], "[", 
             RowBox[{"c0", ",", "a0"}], "]"}]},
           {
            RowBox[{
             UnderscriptBox["\[ForAll]", 
              RowBox[{
               UnderscriptBox["\[Element]", "D"], 
               RowBox[{"[", 
                RowBox[{"m0", ",", "c"}], "]"}]}]], 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                RowBox[{
                 RowBox[{"c", "\[NotEqual]", 
                  UnderscriptBox["0", "D"]}], "\[And]", 
                 RowBox[{
                  UnderscriptBox["mult", "D"], "[", 
                  RowBox[{"m0", ",", "c"}], "]"}]}], 
                TagBox[")",
                 "AutoParentheses"]}], "\[Implies]", 
               RowBox[{
                UnderscriptBox["mult", "D"], "[", 
                RowBox[{
                 RowBox[{"m", 
                  UnderscriptBox["*", "D"], "m0"}], ",", "c"}], "]"}]}], 
              TagBox[")",
               "AutoParentheses"]}]}]},
           {
            RowBox[{
             UnderscriptBox["\[ForAll]", 
              RowBox[{
               UnderscriptBox["\[Element]", "D"], 
               RowBox[{"[", "c", "]"}]}]], 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               RowBox[{"a0", 
                UnderscriptBox["\[Succeeds]", "D"], 
                RowBox[{"a0", 
                 UnderscriptBox["+", "D"], "c"}]}], "\[Implies]", 
               RowBox[{"a", 
                UnderscriptBox["\[Succeeds]", "D"], 
                RowBox[{"a", 
                 UnderscriptBox["+", "D"], 
                 RowBox[{"m", 
                  UnderscriptBox["*", "D"], "c"}]}]}]}], 
              TagBox[")",
               "AutoParentheses"]}]}]},
           {
            RowBox[{
             UnderscriptBox["\[ForAll]", 
              RowBox[{
               UnderscriptBox["\[Element]", "D"], 
               RowBox[{"[", 
                RowBox[{"b", ",", "d", ",", "c"}], "]"}]}]], 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                TagBox[
                 RowBox[{"b", 
                  UnderscriptBox[
                   SubscriptBox["\[Rule]", 
                    RowBox[{"{", "c", "}"}]], "D"], "d"}],
                 Identity], 
                TagBox[")",
                 "AutoParentheses"]}], "\[Implies]", 
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                TagBox[
                 RowBox[{
                  RowBox[{"m", 
                   UnderscriptBox["*", "D"], "b"}], 
                  UnderscriptBox[
                   SubscriptBox["\[LeftRightArrow]", 
                    RowBox[{"{", "c", "}"}]], "D"], 
                  RowBox[{"m", 
                   UnderscriptBox["*", "D"], "d"}]}],
                 Identity], 
                TagBox[")",
                 "AutoParentheses"]}]}], 
              TagBox[")",
               "AutoParentheses"]}]}]}
          }]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R8)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2106857486", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590818099864423*^9, 3.590818123935363*^9}, {
   3.590818473434281*^9, 3.590818589702203*^9}, {3.590818786197373*^9, 
   3.590818795273387*^9}, {3.590818846035089*^9, 3.590818898014337*^9}, {
   3.590818933540285*^9, 3.590818964640851*^9}, {3.590819021379074*^9, 
   3.590819057591646*^9}, {3.590819150150839*^9, 3.59081915650585*^9}, {
   3.590823419139967*^9, 3.5908236510612*^9}, {3.590823759397903*^9, 
   3.590823827991037*^9}, {3.590824595852176*^9, 3.590824618911381*^9}, {
   3.590824696796655*^9, 3.590824702033817*^9}, {3.590824758698355*^9, 
   3.590824881811943*^9}, {3.590825135864186*^9, 3.590825153132126*^9}, {
   3.590825215237586*^9, 3.590825409530976*^9}, {3.590825776236667*^9, 
   3.590825871847732*^9}, {3.590825906086264*^9, 3.590825933466272*^9}, {
   3.590825979756884*^9, 3.590826224119044*^9}, {3.59082639825937*^9, 
   3.590826454093214*^9}, {3.590826590991154*^9, 3.590826636327051*^9}, {
   3.590826673682106*^9, 3.590826715030555*^9}, {3.590826786471113*^9, 
   3.590826994760911*^9}, {3.590827045496703*^9, 3.59082715419869*^9}, {
   3.590827225762745*^9, 3.590827265205178*^9}, {3.590827304672176*^9, 
   3.590827387395462*^9}, {3.590827434329823*^9, 3.590827581923352*^9}, {
   3.59082781634764*^9, 3.590827840797958*^9}, 3.590827946044463*^9, {
   3.590827984491993*^9, 3.590828119374258*^9}, {3.590828158879933*^9, 
   3.590828215432731*^9}, {3.590830305943432*^9, 3.590830318562798*^9}, {
   3.590915797775656*^9, 3.590915820783107*^9}, {3.591944716730478*^9, 
   3.5919447643648*^9}, {3.594029241538039*^9, 3.594029255449955*^9}, {
   3.607337524768936*^9, 3.607337572710005*^9}, {3.610169887189126*^9, 
   3.610169920764437*^9}, {3.610178326916865*^9, 3.610178356326464*^9}, {
   3.610178988467918*^9, 3.610179053117072*^9}, {3.6328096889833612`*^9, 
   3.632809708932414*^9}, {3.632809855495929*^9, 3.632809858079183*^9}, {
   3.6618501621667852`*^9, 3.6618501687811966`*^9}},
 CellTags->{
  "ID:2106857486", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R8"},
 CellID->2106857486],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1411649426]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2055609263],

Cell["", "OpenEnvironment",
 CellID->465374984],

Cell[CellGroupData[{

Cell["Definition: Auxiliary Notions For Reduction Ring Category", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632048939696076*^9, 3.63204895698427*^9}},
 CellID->31201615],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "a", ",", "b", ",", "c", ",", "d", ",", "m", ",", "m1", ",", "m2", ",", 
   "b1", ",", "b2", ",", "TM", ",", "TN"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1197875862}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6320489635153227`*^9, 3.6320489761958733`*^9}, {
  3.632049251367753*^9, 3.632049251543857*^9}, {3.63204928989401*^9, 
  3.6320492926981707`*^9}, {3.638164619917905*^9, 3.638164644275661*^9}},
 CellTags->{
  "ID:1197875862", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1197875862],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areR5Witnesses", "D"], "[", 
   RowBox[{
   "TM", ",", "TN", ",", "a", ",", "b", ",", "c", ",", "d", ",", "m"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    RowBox[{"\[And]", 
     TagBox[
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}]], 
           RowBox[{
            UnderscriptBox["isReducible", "D"], "[", 
            RowBox[{
             RowBox[{
              UnderscriptBox["reduceM", "D"], "[", 
              RowBox[{
               RowBox[{"a", 
                UnderscriptBox["+", "D"], "d"}], ",", "c", ",", "TM", ",", 
               RowBox[{"i", "-", "1"}]}], "]"}], ",", "c", ",", 
             SubscriptBox["TM", "i"]}], "]"}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}]], 
           RowBox[{
            UnderscriptBox["isReducible", "D"], "[", 
            RowBox[{
             RowBox[{
              UnderscriptBox["reduceM", "D"], "[", 
              RowBox[{
               RowBox[{"b", 
                UnderscriptBox["+", "D"], "d"}], ",", "c", ",", "TN", ",", 
               RowBox[{"i", "-", "1"}]}], "]"}], ",", "c", ",", 
             SubscriptBox["TN", "i"]}], "]"}]}]},
         {
          RowBox[{
           RowBox[{
            UnderscriptBox[
             SubscriptBox["\[Sum]", "D"], 
             RowBox[{
              RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
              RowBox[{
              "\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}]], 
            SubscriptBox["TM", "i"]}], "\[Equal]", 
           RowBox[{"m", 
            UnderscriptBox["+", "D"], 
            RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", "D"], 
              RowBox[{
               RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
               RowBox[{
               "\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}]], 
             SubscriptBox["TN", "i"]}]}]}]}
        }]}],
      Identity]}],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R5 witnesses)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1949920189", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632049183410001*^9, 3.632049311394211*^9}, {
  3.632049353042017*^9, 3.632049410374874*^9}, {3.6357488858682404`*^9, 
  3.6357488923886414`*^9}, {3.638164661420665*^9, 3.638164666123501*^9}},
 CellTags->{
  "ID:1949920189", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R5 witnesses"},
 CellID->1949920189],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isR6Witness", "D"], "[", 
   RowBox[{
   "TM", ",", "a", ",", "c", ",", "m1", ",", "m2", ",", "b1", ",", "b2"}], 
   "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}]], 
         RowBox[{
          UnderscriptBox["symred", "D"], "[", 
          RowBox[{
           RowBox[{
            UnderscriptBox["reduceM", "D"], "[", 
            RowBox[{"b1", ",", "c", ",", "TM", ",", 
             RowBox[{"i", "-", "1"}]}], "]"}], ",", 
           RowBox[{
            UnderscriptBox["reduceM", "D"], "[", 
            RowBox[{"b1", ",", "c", ",", "TM", ",", "i"}], "]"}], ",", "c", 
           ",", 
           SubscriptBox["TM", "i"]}], "]"}]}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}]], 
         RowBox[{"a", 
          UnderscriptBox["\[Succeeds]", "D"], 
          RowBox[{
           UnderscriptBox["reduceM", "D"], "[", 
           RowBox[{"b1", ",", "c", ",", "TM", ",", "i"}], "]"}]}]}]},
       {
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            UnderscriptBox[
             SubscriptBox["\[Sum]", "D"], 
             RowBox[{
              RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
              RowBox[{
              "\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}]], 
            SubscriptBox["TM", "i"]}], ")"}], 
          UnderscriptBox["+", "D"], "m1"}], "\[Equal]", "m2"}]}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R6 witness)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2027349511", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632048981691978*^9, 3.63204908190555*^9}, {
  3.6320491367456503`*^9, 3.6320491370786963`*^9}, {3.6357488973924513`*^9, 
  3.635748904373274*^9}, {3.6381646728531847`*^9, 3.638164737437784*^9}},
 CellTags->{
  "ID:2027349511", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6 witness"},
 CellID->2027349511],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1778400228]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->943653347]
}, Closed]],

Cell[CellGroupData[{

Cell["Lemmas (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6442216264451027`*^9, 3.644221632907298*^9}},
 CellID->2042908020],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isCommRing1", "[", "D", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1717911774}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644221639158284*^9, 3.6442216443912697`*^9}},
 CellTags->{
  "ID:1717911774", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1717911774],

Cell["", "OpenEnvironment",
 CellID->803522812],

Cell[CellGroupData[{

Cell["Lemma: Immediate Consequences of Axiom (R0)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.633176424099142*^9, 3.6331764282926598`*^9}, {
   3.641205071362879*^9, 3.641205075908388*^9}, {3.647938956980015*^9, 
   3.6479389571659718`*^9}, {3.650774232490774*^9, 3.650774236424316*^9}, {
   3.6580406423947706`*^9, 3.658040649087182*^9}, 3.658040698430069*^9},
 CellID->1518391126],

Cell[BoxData[
 RowBox[{
  RowBox[{"R0", "[", "D", "]"}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 349001278}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633176467285838*^9, 3.63317647047957*^9}, {
  3.641205081107356*^9, 3.641205088669113*^9}, {3.647938863716786*^9, 
  3.647938864381926*^9}},
 CellTags->{
  "ID:349001278", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->349001278],

Cell[BoxData[
 RowBox[{
  RowBox[{"Multipliers", "[", "D", "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", "c", "]"}]}]], 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"1", "\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalN]", 
      RowBox[{"1", ",", "\[Ellipsis]", ",", 
       RowBox[{
        UnderscriptBox["multN", "D"], "[", "c", "]"}]}]]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(9)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:474542129", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633176431165687*^9, 3.633176463237855*^9}, {
  3.6479388580356007`*^9, 3.647938866006709*^9}},
 CellTags->{
  "ID:474542129", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "9"},
 CellID->474542129],

Cell[BoxData[
 RowBox[{
  RowBox[{"R4", "[", "D", "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", "c", "]"}]}]], 
     RowBox[{"c", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"c", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"{", "c", "}"}]], "D"], 
      UnderscriptBox["0", "D"]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(21)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:360788203", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647938726272785*^9, 3.647938768087554*^9}, {
   3.6479388784001007`*^9, 3.647938891258873*^9}, 3.647938950515341*^9, {
   3.661850197516447*^9, 3.66185019901405*^9}},
 CellTags->{
  "ID:360788203", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "21"},
 CellID->360788203],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1400476426]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1277923749],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(9)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(9)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R0", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"Multipliers", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 StyleBox["D", "ExpressionVariable"]]}]], 
             RowBox[{"1", "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", 
                RowBox[{"1", ",", "\[Ellipsis]", ",", 
                  RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], 
                    "]"}]}]]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R0", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"Multipliers", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 StyleBox["D", "ExpressionVariable"]]}]], 
             RowBox[{"1", "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", 
                RowBox[{"1", ",", "\[Ellipsis]", ",", 
                  RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], 
                    "]"}]}]]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:474542129"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p474542129-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p474542129-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:474542129", "Proof|ID:474542129-1"},
 CellID->1905252519],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p474542129-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1028745265]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(21)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(21)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R0", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}]], 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 UnderscriptBox["0", 
                  StyleBox[
                  "D", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R0", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}]], 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 UnderscriptBox["0", 
                  StyleBox[
                  "D", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:360788203"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p360788203-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p360788203-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:360788203", "Proof|ID:360788203-1"},
 CellID->3264289],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p360788203-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->115701537]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->32237124],

Cell[CellGroupData[{

Cell["Lemma: Immediate Consequences of Axiom (R2)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.638775776994514*^9, 3.63877578585434*^9}},
 CellID->253497300],

Cell[BoxData[
 RowBox[{
  RowBox[{"R2", "[", "D", "]"}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 998446180}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638775792942257*^9, 3.638775797539386*^9}},
 CellTags->{
  "ID:998446180", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->998446180],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"m", ",", "c"}], "]"}]}]], 
  RowBox[{"c", "\[NotEqual]", 
   UnderscriptBox["0", "D"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1581977856}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638775805606304*^9, 3.638775810400057*^9}, {
  3.638776391461212*^9, 3.638776391796851*^9}, {3.6405793562192993`*^9, 
  3.640579363645216*^9}},
 CellTags->{
  "ID:1581977856", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1581977856],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "D"], "[", 
   RowBox[{"m", ",", "c"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"m", "\[NotEqual]", 
    UnderscriptBox["0", "D"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(8)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:623334744", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6387758185699472`*^9, 3.6387758384018*^9}, {
  3.6405793667818127`*^9, 3.640579367108326*^9}},
 CellTags->{
  "ID:623334744", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "8"},
 CellID->623334744],

Cell[BoxData[
 RowBox[{
  RowBox[{"Multipliers", "[", "D", "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       UnderscriptBox["multN", "D"], "[", "c", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["mult", "D"], "[", 
       RowBox[{"m", ",", "c", ",", "i"}], "]"}], "\[Implies]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"m", "\[NotEqual]", 
        UnderscriptBox["0", "D"]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(8')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1112133758", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.638775847511381*^9, 3.638775879965519*^9}, {
  3.6405793720702343`*^9, 3.640579372366197*^9}, {3.64058090400343*^9, 
  3.6405809173128567`*^9}},
 CellTags->{
  "ID:1112133758", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "8'"},
 CellID->1112133758],

Cell[BoxData[
 RowBox[{
  RowBox[{"R4", "[", "D", "]"}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 810098043}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6454353552701597`*^9, 3.6454353605402184`*^9}, {
  3.645435425801339*^9, 3.645435428311042*^9}},
 CellTags->{
  "ID:810098043", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->810098043],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "D"], "[", 
   RowBox[{"m", ",", "c"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"m", 
      UnderscriptBox["*", "D"], "c"}], 
     TagBox[")",
      "AutoParentheses"]}], 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"{", "c", "}"}]], "D"], 
    UnderscriptBox["0", "D"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(19)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1578524611", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6454353883082733`*^9, 3.645435415526972*^9}, {
  3.6454354686328583`*^9, 3.645435468930608*^9}, {3.661850214707678*^9, 
  3.66185021618968*^9}},
 CellTags->{
  "ID:1578524611", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "19"},
 CellID->1578524611],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"R1", "[", "D", "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["mult", "D"], "[", 
     RowBox[{"m", ",", "c"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["-", "D"], 
      RowBox[{"(", 
       RowBox[{"m", 
        UnderscriptBox["*", "D"], "c"}], ")"}]}], 
     TagBox[")",
      "AutoParentheses"]}], 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"{", "c", "}"}]], "D"], 
    UnderscriptBox["0", "D"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(19')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1738175589", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64543543323514*^9, 3.64543547265427*^9}, {
  3.661850218888485*^9, 3.661850223381293*^9}},
 CellTags->{
  "ID:1738175589", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "19'"},
 CellID->1738175589],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->466928142]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->946058699],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(8)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(8)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R2", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 StyleBox["D", "ExpressionVariable"]]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["mult", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  StyleBox[
                  "D", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R2", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 StyleBox["D", "ExpressionVariable"]]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["mult", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  StyleBox[
                  "D", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:623334744"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p623334744-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p623334744-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:623334744", "Proof|ID:623334744-1"},
 CellID->1449852563],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p623334744-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->521785732]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(8')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(8')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R2", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 StyleBox["D", "ExpressionVariable"]]}]], 
             RowBox[{
               RowBox[{"Multipliers", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R2", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 StyleBox["D", "ExpressionVariable"]]}]], 
             RowBox[{
               RowBox[{"Multipliers", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1112133758"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1112133758-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1112133758-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1112133758", "Proof|ID:1112133758-1"},
 CellID->1792610282],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1112133758-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1552450960]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(19)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(19)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R2", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 StyleBox["D", "ExpressionVariable"]]}]], 
             RowBox[{
               RowBox[{"R4", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R2", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 StyleBox["D", "ExpressionVariable"]]}]], 
             RowBox[{
               RowBox[{"R4", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1578524611"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1578524611-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1578524611-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1578524611", "Proof|ID:1578524611-1"},
 CellID->32974812],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1578524611-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->47618736]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(19')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(19')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R2", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 StyleBox["D", "ExpressionVariable"]]}]], 
             RowBox[{
               RowBox[{"R4", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"R1", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R2", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 StyleBox["D", "ExpressionVariable"]]}]], 
             RowBox[{
               RowBox[{"R4", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"R1", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1738175589"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1738175589-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1738175589-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1738175589", "Proof|ID:1738175589-1"},
 CellID->629724279],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1738175589-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->481178017]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->68184375],

Cell[CellGroupData[{

Cell["Lemma: Immediate Consequences of Axiom (R4)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.633173239077594*^9, 3.6331732503891478`*^9}, {
  3.63317538794191*^9, 3.633175390506851*^9}, {3.6454353651419783`*^9, 
  3.645435374510001*^9}},
 CellID->638803249],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"R4", "[", "D", "]"}], "\[And]", 
    RowBox[{"isPartIrreflOrder", "[", 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox[
        UnderscriptBox["\[Succeeds]", "D"],
        Identity], 
       TagBox[")",
        "AutoParentheses"]}], ",", "D"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 258339355}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633173283344268*^9, 3.6331732871055517`*^9}, {
  3.640579536323999*^9, 3.6405795552191896`*^9}},
 CellTags->{
  "ID:258339355", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->258339355],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "m"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1933985410}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6331732784157267`*^9, 3.633173299450178*^9}, {
  3.6332347102444477`*^9, 3.633234710392891*^9}},
 CellTags->{
  "ID:1933985410", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1933985410],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{
   UnderscriptBox["0", "D"], 
   UnderscriptBox["\[Succeeds]", "D"], "a"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(17)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:739393974", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.640579996289239*^9, 3.640580014613675*^9}},
 CellTags->{
  "ID:739393974", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "17"},
 CellID->739393974],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{"a", ",", "c", ",", "m"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}], "\[And]", 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}], "\[And]", 
    RowBox[{"m", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(10)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:875264505", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633173304887649*^9, 3.633173350602635*^9}},
 CellTags->{
  "ID:875264505", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "10"},
 CellID->875264505],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{"a", ",", "c"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}], "\[And]", 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(11)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2012915447", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633173345236393*^9, 3.633173354941695*^9}},
 CellTags->{
  "ID:2012915447", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "11"},
 CellID->2012915447],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"{", "c", "}"}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}], "\[And]", 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(12)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1489336269", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6332347006740303`*^9, 3.6332347399694977`*^9}, {
  3.66185025617255*^9, 3.6618502575141525`*^9}},
 CellTags->{
  "ID:1489336269", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "12"},
 CellID->1489336269],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", "j"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isNTCR", "D"], "[", 
     RowBox[{"c", ",", "b", ",", "a", ",", "i", ",", "j"}], "]"}], 
    "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"c", "\[NotEqual]", 
       UnderscriptBox["0", "D"]}], "\[And]", 
      RowBox[{"b", "\[NotEqual]", 
       UnderscriptBox["0", "D"]}], "\[And]", 
      RowBox[{"a", "\[NotEqual]", 
       UnderscriptBox["0", "D"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(13)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1722317392", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633175397863357*^9, 3.633175428232519*^9}, {
  3.633234958076655*^9, 3.633234993278504*^9}, {3.648798541366338*^9, 
  3.648798541699223*^9}},
 CellTags->{
  "ID:1722317392", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "13"},
 CellID->1722317392],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "D"], "[", 
   RowBox[{"c", ",", "a"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}], "\[And]", 
    RowBox[{"a", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(15)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1089845252", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.639121937783908*^9, 3.639121943872765*^9}},
 CellTags->{
  "ID:1089845252", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "15"},
 CellID->1089845252],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->640836049]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1240067933],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(17)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(17)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["a", "ExpressionVariable"]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["a", "ExpressionVariable"]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:739393974"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p739393974-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p739393974-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:739393974", "Proof|ID:739393974-1"},
 CellID->1175180046],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p739393974-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1348459641]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(10)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(10)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:875264505"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p875264505-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p875264505-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:875264505", "Proof|ID:875264505-1"},
 CellID->564262226],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p875264505-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->24736570]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(11)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(11)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2012915447"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2012915447-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2012915447-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2012915447", "Proof|ID:2012915447-1"},
 CellID->2044326423],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2012915447-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->485002968]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(12)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(12)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1489336269"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1489336269-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1489336269-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1489336269", "Proof|ID:1489336269-1"},
 CellID->62520461],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1489336269-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->257308557]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(13)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(13)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], ",", 
                  StyleBox["j", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], ",", 
                  StyleBox["j", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1722317392"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1722317392-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1722317392-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1722317392", "Proof|ID:1722317392-1"},
 CellID->1641148702],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1722317392-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1013568673]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(15)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(15)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isNTCR", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isNTCR", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1089845252"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1089845252-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1089845252-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1089845252", "Proof|ID:1089845252-1"},
 CellID->96759121],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1089845252-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1750417022]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1937995867],

Cell[CellGroupData[{

Cell["Lemma: \[LeftRightArrow]", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.633064239913959*^9, 3.6330642420955753`*^9}},
 CellID->1477540412],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"R1", "[", "D", "]"}], "\[And]", 
   RowBox[{"R4", "[", "D", "]"}], "\[And]", 
   RowBox[{"isPartIrreflOrder", "[", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       UnderscriptBox["\[Succeeds]", "D"],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}], ",", "D"}], "]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 2021433946}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6330647988009157`*^9, 3.633064813491048*^9}, {
  3.641627818268145*^9, 3.641627830053975*^9}},
 CellTags->{
  "ID:2021433946", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2021433946],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 369680194}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.644914113408598*^9},
 CellTags->{
  "ID:369680194", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->369680194],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", 
      RowBox[{"{", "c", "}"}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", "m", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["mult", "D"], "[", 
           RowBox[{"m", ",", "c"}], "]"}]},
         {
          RowBox[{"b", "\[Equal]", 
           RowBox[{"a", 
            UnderscriptBox["-", "D"], 
            RowBox[{"m", 
             UnderscriptBox["*", "D"], "c"}]}]}]},
         {
          RowBox[{
           RowBox[{"a", 
            UnderscriptBox["\[Succeeds]", "D"], "b"}], "\[Or]", 
           RowBox[{"b", 
            UnderscriptBox["\[Succeeds]", "D"], "a"}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LeftRightArrow] alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:491925495", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.633064255540641*^9, 3.6330643424137497`*^9}, {
  3.6330691194828157`*^9, 3.6330691361102667`*^9}, {3.6392106532877893`*^9, 
  3.6392106669964323`*^9}, {3.644914116099557*^9, 3.644914117425235*^9}, {
  3.6618502784493895`*^9, 3.661850280820594*^9}},
 CellTags->{
  "ID:491925495", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] alternative"},
 CellID->491925495],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["symred", "D"], "[", 
     RowBox[{"a", ",", "b", ",", "c", ",", "m"}], "]"}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[LeftRightArrow]", 
        RowBox[{"{", "c", "}"}]], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symred \[LeftRightArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1668317482", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644914122289288*^9, 3.644914168055539*^9}, {
  3.644914271588799*^9, 3.644914276768674*^9}, {3.6618502843149996`*^9, 
  3.661850286608204*^9}},
 CellTags->{
  "ID:1668317482", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred \[LeftRightArrow]"},
 CellID->1668317482],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->354300585]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->45215780],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LeftRightArrow] alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LeftRightArrow] alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"(", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b", "ExpressionVariable"]}], ")"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox[
                    "a", "ExpressionVariable"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"(", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b", "ExpressionVariable"]}], ")"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox[
                    "a", "ExpressionVariable"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:491925495"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p491925495-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p491925495-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:491925495", "Proof|ID:491925495-1"},
 CellID->99017069],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p491925495-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->492768956]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symred \[LeftRightArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symred \[LeftRightArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["symred", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox[
                   "b", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"R4", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["symred", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox[
                   "b", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1668317482"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1668317482-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1668317482-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1668317482", "Proof|ID:1668317482-1"},
 CellID->304755132],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1668317482-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->422791318]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1527684406],

Cell[CellGroupData[{

Cell["Lemma: Elementary Properties of reduceM", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6321996447109222`*^9, 3.6321996675146427`*^9}},
 CellID->1451450076],

Cell[BoxData[
 RowBox[{
  RowBox[{"isPartIrreflOrder", "[", 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      UnderscriptBox["\[Succeeds]", "D"],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}], ",", "D"}], "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 169581633}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632200727517603*^9, 3.632200731547317*^9}},
 CellTags->{
  "ID:169581633", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->169581633],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "c"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainTuples", "[", "D", "]"}]], 
     RowBox[{"[", "T", "]"}]}]], 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
       RowBox[{
        UnderscriptBox["isReducible", "D"], "[", 
        RowBox[{
         RowBox[{
          UnderscriptBox["reduceM", "D"], "[", 
          RowBox[{"a", ",", "c", ",", "T", ",", 
           RowBox[{"i", "-", "1"}]}], "]"}], ",", "c", ",", 
         SubscriptBox["T", "i"]}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
       RowBox[{"a", 
        UnderscriptBox["\[Succeeds]", "D"], 
        RowBox[{
         UnderscriptBox["reduceM", "D"], "[", 
         RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(403)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:265555623", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632199719727681*^9, 3.632199790121461*^9}, {
  3.632199887073448*^9, 3.63219988744236*^9}, {3.632205055847219*^9, 
  3.632205079253188*^9}, {3.635748921991867*^9, 3.635748929596986*^9}},
 CellTags->{
  "ID:265555623", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "403"},
 CellID->265555623],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->428237015]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1421839378],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(403)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(403)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"isPartIrreflOrder", "[", 
             RowBox[{
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], ",", 
               StyleBox["D", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                    ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"isPartIrreflOrder", "[", 
             RowBox[{
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], ",", 
               StyleBox["D", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                    ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:265555623"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p265555623-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p265555623-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:265555623", "Proof|ID:265555623-1"},
 CellID->1813792728],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p265555623-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.66227022017914*^9, 3.6622702217391434`*^9}},
 CellID->805435955]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->464160879],

Cell[CellGroupData[{

Cell["Lemma: Symmetry of symred", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6321376652004757`*^9, 3.63213767014867*^9}},
 CellID->718753927],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"R1", "[", "D", "]"}], "\[And]", 
   RowBox[{"isPartIrreflOrder", "[", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       UnderscriptBox["\[Succeeds]", "D"],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}], ",", "D"}], "]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1020054522}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632200746500105*^9, 3.632200749527487*^9}, {
  3.641631464639629*^9, 3.641631468857442*^9}},
 CellTags->{
  "ID:1020054522", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1020054522],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "m"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 270873424}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641631736809351*^9, 3.641631741917317*^9}},
 CellTags->{
  "ID:270873424", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->270873424],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["symred", "D"], "[", 
   RowBox[{"a", ",", "b", ",", "c", ",", "m"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["symred", "D"], "[", 
   RowBox[{"b", ",", "a", ",", "c", ",", 
    RowBox[{
     UnderscriptBox["-", "D"], "m"}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(16)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1165091979", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641631752668099*^9, 3.641631759866534*^9}},
 CellTags->{
  "ID:1165091979", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "16"},
 CellID->1165091979],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["symred", "D"], "[", 
   RowBox[{"a", ",", "b", ",", "c", ",", "m"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["symred", "D"], "[", 
   RowBox[{"b", ",", "a", ",", "c", ",", 
    RowBox[{
     UnderscriptBox["-", "D"], "m"}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symred symmetry)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1894133408", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632137684382362*^9, 3.632137735439176*^9}, {
  3.632137765489491*^9, 3.6321377658521233`*^9}, {3.641631744726103*^9, 
  3.641631747973146*^9}},
 CellTags->{
  "ID:1894133408", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred symmetry"},
 CellID->1894133408],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->677898984]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->492574805],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(16)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(16)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["m", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1165091979"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1165091979-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1165091979-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1165091979", "Proof|ID:1165091979-1"},
 CellID->1301195110],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1165091979-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->197446793]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symred symmetry)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symred symmetry)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["m", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["m", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1894133408"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1894133408-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1894133408-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1894133408", "Proof|ID:1894133408-1"},
 CellID->1016736240],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1894133408-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1980556603]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Algorithmic Reduction Rings", "Subsection",
 CellChangeTimes->{{3.64421422788451*^9, 3.6442142333660727`*^9}},
 CellID->148207570],

Cell[CellGroupData[{

Cell["Informal Explanation", "Subsubsection",
 CellChangeTimes->{{3.590830450392701*^9, 3.590830454213885*^9}},
 CellID->2020955226],

Cell[TextData[{
 "In order for being an algorithmic reduction ring, a domain ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " has to be a reduction ring and also provide definitions for the following \
operations:"
}], "Text",
 CellChangeTimes->{{3.590834428115336*^9, 3.590834531416853*^9}, {
  3.590834604524359*^9, 3.590834606046285*^9}, {3.644214301956716*^9, 
  3.6442143174148207`*^9}},
 CellID->883452679],

Cell[CellGroupData[{

Cell[TextData[{
 "A function that returns a suitable multiplier for reducing a given element ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " modulo another given element ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ", or 0 if no such multiplier exists (\[OpenCurlyDoubleQuote]rdm\
\[CloseCurlyDoubleQuote])."
}], "Item",
 CellChangeTimes->{{3.596193869993015*^9, 3.596193890605921*^9}},
 CellID->604644732],

Cell[TextData[{
 "A function that returns ",
 StyleBox["all",
  FontWeight->"Bold"],
 " minimal non-trivial common reducibles for ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " (\[OpenCurlyDoubleQuote]lcrd\[CloseCurlyDoubleQuote]). Note that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Equal]", "y"}], TraditionalForm]]],
 " is also possible!"
}], "Item",
 CellChangeTimes->{{3.596193869993015*^9, 3.596193894440541*^9}},
 CellID->123344064],

Cell[TextData[{
 "A function that returns",
 StyleBox[" one",
  FontWeight->"Bold"],
 " pair of suitable multipliers for obtaining the critical pair of two ",
 StyleBox["different",
  FontWeight->"Bold"],
 " elements ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " and one of their minimal non-trivial reducibles, and ",
 StyleBox["all",
  FontWeight->"Bold"],
 " pairs of suitable multipliers if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Equal]", "y"}], TraditionalForm]]],
 " (\[OpenCurlyDoubleQuote]cpm\[CloseCurlyDoubleQuote])."
}], "Item",
 CellChangeTimes->{{3.596193869993015*^9, 3.596193898795863*^9}},
 CellID->1384519374]
}, Open  ]],

Cell["\<\
These operations need to satisfy several non-trivial properties.\
\>", "Text",
 CellChangeTimes->{{3.590834596637894*^9, 3.590834623998124*^9}},
 CellID->2128238209]
}, Closed]],

Cell[CellGroupData[{

Cell["Definition", "Subsubsection",
 CellChangeTimes->{{3.6363563267511272`*^9, 3.636356331897575*^9}, {
   3.643009962616169*^9, 3.643009966383718*^9}, 3.644214236517424*^9},
 CellID->1142821098],

Cell["", "OpenEnvironment",
 CellID->502658829],

Cell[CellGroupData[{

Cell["Definition: Category of Algorithmic Reduction Rings", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6363563460355787`*^9, 3.63635635373665*^9}, 
   3.65001833284193*^9, 3.650189960179049*^9},
 CellID->327281592],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "D"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1779789522}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643363383499146*^9, 3.643363384408126*^9}},
 CellTags->{
  "ID:1779789522", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1779789522],

Cell[BoxData[
 RowBox[{
  RowBox[{"isAlgoReductionRing", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"isReductionRing", "[", "D", "]"}]},
       {
        RowBox[{"R9", "[", "D", "]"}]},
       {
        RowBox[{"R10", "[", "D", "]"}]},
       {
        RowBox[{"R11", "[", "D", "]"}]},
       {
        RowBox[{"R12", "[", "D", "]"}]},
       {
        RowBox[{"R13", "[", "D", "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(algorithmic reduction ring)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:979757495", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.636356364943036*^9, 3.6363564108852577`*^9}, {
  3.6463637247765417`*^9, 3.64636372791953*^9}},
 CellTags->{
  "ID:979757495", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "algorithmic reduction ring"},
 CellID->979757495],

Cell[TextData[{
 "In (R9) we need \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"j", "\[GreaterEqual]", "i"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote], because if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c1", "\[NotEqual]", "c2"}], TraditionalForm]]],
 " it needs to hold for the pair ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"i", ",", "j"}], ")"}], "=", 
    RowBox[{"(", 
     RowBox[{"1", ",", "1"}], ")"}]}], TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.649565387416204*^9, 3.649565437385296*^9}, {
  3.649565758090029*^9, 3.649565790364337*^9}, {3.649569324211773*^9, 
  3.6495693325585117`*^9}, {3.650018357445579*^9, 3.650018358943108*^9}},
 CellID->1965654203],

Cell[BoxData[
 RowBox[{
  RowBox[{"R9", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", 
       RowBox[{"c1", ",", "c2"}], "]"}]}]], 
    RowBox[{
     RowBox[{"c1", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}], "\[And]", 
     RowBox[{"c2", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{
         UnderscriptBox["multN", "D"], "[", "c1", "]"}]}]], 
      RowBox[{"j", "\[GreaterEqual]", "i"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["let", 
        RowBox[{"L", "=", 
         RowBox[{
          UnderscriptBox["lcrd", "D"], "[", 
          RowBox[{"c1", ",", "c2", ",", "i", ",", "j"}], "]"}]}]], 
       RowBox[{"(", 
        RowBox[{"\[And]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {
             RowBox[{
              UnderscriptBox["\[Element]", 
               RowBox[{"DomainTuples", "[", "D", "]"}]], 
              RowBox[{"[", "L", "]"}]}]},
            {
             RowBox[{
              UnderscriptBox["\[ForAll]", 
               RowBox[{
                UnderscriptBox["\[Element]", "D"], 
                RowBox[{"[", "a", "]"}]}]], 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               TagBox[
                RowBox[{"\[And]", 
                 RowBox[{"\[Piecewise]", GridBox[{
                    {
                    RowBox[{
                    RowBox[{"a", 
                    TagBox["\:22ff",
                    Identity,
                    SyntaxForm->"a+b"], "L"}], 
                    TagBox["\[Implies]",
                    Identity,
                    SyntaxForm->"a\[Implies]b"], 
                    RowBox[{
                    UnderscriptBox["isMNTCR", "D"], "[", 
                    RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], 
                    "]"}]}]},
                    {
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["isMNTCR", "D"], "[", 
                    RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], 
                    "]"}], "\[Implies]", 
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    RowBox[{
                    UnderscriptBox["\[Element]", "D"], 
                    RowBox[{"[", "b", "]"}]}]], 
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    RowBox[{"a", 
                    UnderscriptBox["\[Tilde]", "D"], "b"}], "\[And]", 
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{"b", 
                    TagBox["\:22ff",
                    Identity,
                    SyntaxForm->"a+b"], "L"}], 
                    TagBox[")",
                    "AutoParentheses"]}]}], 
                    TagBox[")",
                    "AutoParentheses"]}]}], 
                    TagBox[")",
                    "AutoParentheses"]}]}]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}],
                Identity], 
               TagBox[")",
                "AutoParentheses"]}]}]}
           }]}]}], ")"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R9)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1502179371", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590828228591379*^9, 3.59082825030703*^9}, {
   3.590830330938639*^9, 3.590830333197308*^9}, 3.594029260700193*^9, {
   3.607337582820722*^9, 3.607337586511556*^9}, {3.607338037893116*^9, 
   3.607338043767692*^9}, {3.608549265367931*^9, 3.608549286393792*^9}, {
   3.610080988277904*^9, 3.610081063109666*^9}, {3.610081210478686*^9, 
   3.61008122496953*^9}, {3.610097004088357*^9, 3.610097006865626*^9}, {
   3.610169955951511*^9, 3.610169961740733*^9}, {3.631855183030274*^9, 
   3.631855190076517*^9}, {3.6318563617738047`*^9, 3.6318563663362913`*^9}, {
   3.631960487605377*^9, 3.631960503870864*^9}, {3.635748830812162*^9, 
   3.635748837609274*^9}, 3.642924229168507*^9, {3.649565368255265*^9, 
   3.649565369435235*^9}, {3.650105486767303*^9, 3.650105582293684*^9}},
 CellTags->{
  "ID:1502179371", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R9"},
 CellID->1502179371],

Cell[BoxData[
 RowBox[{
  RowBox[{"R10", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", "c", "]"}]}]], 
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "D"]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox["let", 
      RowBox[{"L", "=", 
       RowBox[{
        UnderscriptBox["lcrd", "D"], "[", "c", "]"}]}]], 
     RowBox[{"(", 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            UnderscriptBox["\[Element]", 
             RowBox[{"DomainTuples", "[", "D", "]"}]], 
            RowBox[{"[", "L", "]"}]}]},
          {
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              UnderscriptBox["\[Element]", "D"], 
              RowBox[{"[", "a", "]"}]}]], 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             TagBox[
              RowBox[{"\[And]", 
               RowBox[{"\[Piecewise]", GridBox[{
                  {
                   RowBox[{
                    RowBox[{"a", 
                    TagBox["\:22ff",
                    Identity,
                    SyntaxForm->"a+b"], "L"}], 
                    TagBox["\[Implies]",
                    Identity,
                    SyntaxForm->"a\[Implies]b"], 
                    RowBox[{
                    UnderscriptBox["isMNTCR", "D"], "[", 
                    RowBox[{"c", ",", "a"}], "]"}]}]},
                  {
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["isMNTCR", "D"], "[", 
                    RowBox[{"c", ",", "a"}], "]"}], "\[Implies]", 
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    RowBox[{
                    UnderscriptBox["\[Element]", "D"], 
                    RowBox[{"[", "b", "]"}]}]], 
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    RowBox[{"a", 
                    UnderscriptBox["\[Tilde]", "D"], "b"}], "\[And]", 
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{"b", 
                    TagBox["\:22ff",
                    Identity,
                    SyntaxForm->"a+b"], "L"}], 
                    TagBox[")",
                    "AutoParentheses"]}]}], 
                    TagBox[")",
                    "AutoParentheses"]}]}], 
                    TagBox[")",
                    "AutoParentheses"]}]}]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                   "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}],
              Identity], 
             TagBox[")",
              "AutoParentheses"]}]}]}
         }]}]}], ")"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R10)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1524689422", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594456771708618*^9, 3.594456816673966*^9}, {
   3.60733759387014*^9, 3.607337597653487*^9}, {3.607338048869021*^9, 
   3.607338052936444*^9}, {3.610081101756895*^9, 3.610081132609488*^9}, {
   3.610081166037477*^9, 3.61008118747174*^9}, {3.610169992829732*^9, 
   3.610169998421042*^9}, {3.635748844860052*^9, 3.6357488518068123`*^9}, 
   3.642922985581794*^9, {3.6429313854527407`*^9, 3.642931388738332*^9}, {
   3.650105616819623*^9, 3.650105631325171*^9}},
 CellTags->{
  "ID:1524689422", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R10"},
 CellID->1524689422],

Cell[BoxData[
 RowBox[{
  RowBox[{"R11", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], "[", 
     RowBox[{"a", ",", "c"}], "]"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox["let", 
      RowBox[{"R", "=", 
       RowBox[{
        UnderscriptBox["rdm", "D"], "[", 
        RowBox[{"a", ",", "c"}], "]"}]}]], 
     RowBox[{"(", 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            UnderscriptBox["\[Element]", "D"], 
            RowBox[{"[", "R", "]"}]}]},
          {
           RowBox[{
            RowBox[{
             UnderscriptBox["isReducible", "D"], "[", 
             RowBox[{"a", ",", "c"}], "]"}], 
            TagBox["\[Implies]",
             Identity,
             SyntaxForm->"a\[Implies]b"], 
            RowBox[{
             UnderscriptBox["isReducible", "D"], "[", 
             RowBox[{"a", ",", "c", ",", "R"}], "]"}]}]},
          {
           RowBox[{
            RowBox[{"\[Not]", 
             RowBox[{
              UnderscriptBox["isReducible", "D"], "[", 
              RowBox[{"a", ",", "c"}], "]"}]}], "\[Implies]", 
            RowBox[{"R", "\[Equal]", 
             UnderscriptBox["0", "D"]}]}]}
         }]}]}], ")"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R11)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:528478842", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590828255756685*^9, 3.590828264774919*^9}, {
   3.59083033566052*^9, 3.590830337980496*^9}, 3.594029264492864*^9, 
   3.594456783977234*^9, {3.607337634107487*^9, 3.607337646107204*^9}, {
   3.61008135611518*^9, 3.610081415931176*^9}, {3.610095345469615*^9, 
   3.610095379261472*^9}, {3.610095463839692*^9, 3.610095493093805*^9}, {
   3.610170002416183*^9, 3.61017002180816*^9}},
 CellTags->{
  "ID:528478842", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R11"},
 CellID->528478842],

Cell[BoxData[
 RowBox[{
  RowBox[{"R12", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", 
       RowBox[{"c1", ",", "c2"}], "]"}]}]], 
    RowBox[{
     RowBox[{"c1", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}], "\[And]", 
     RowBox[{"c2", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}]}]], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{
        UnderscriptBox["multN", "D"], "[", "c1", "]"}]}]], 
     RowBox[{"i", "<", "j"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["let", 
       RowBox[{"L", "=", 
        RowBox[{
         UnderscriptBox["lcrd", "D"], "[", 
         RowBox[{"c1", ",", "c2", ",", "i", ",", "j"}], "]"}]}]], 
      RowBox[{"(", 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "L", "\[RightBracketingBar]"}]}]], 
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{
           UnderscriptBox["\[Element]", "D"], 
           RowBox[{"[", 
            RowBox[{"p1", ",", "p2"}], "]"}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"\[And]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{
                RowBox[{
                 UnderscriptBox["cpm", "D"], "[", 
                 RowBox[{"c1", ",", "c2", ",", 
                  SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}], 
                "\[Equal]", 
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"p1", ",", "p2"}], "\[RightAngleBracket]"}]}]},
              {
               RowBox[{
                UnderscriptBox["areCPMultipliers", "D"], "[", 
                RowBox[{"p1", ",", "p2", ",", "c1", ",", "c2", ",", 
                 SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]}], ")"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R12)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:415149625", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590828267871608*^9, 3.590828276565441*^9}, 
   3.594029268655213*^9, {3.594456785746027*^9, 3.594456787412251*^9}, {
   3.607337653084999*^9, 3.607337668213322*^9}, {3.610095560918758*^9, 
   3.610095571856466*^9}, {3.610095613585226*^9, 3.610095663392624*^9}, {
   3.61017003292403*^9, 3.610170057660129*^9}, {3.631855218377977*^9, 
   3.631855224441144*^9}, {3.631856389097468*^9, 3.631856394943028*^9}, {
   3.6357488708616467`*^9, 3.635748877984825*^9}, {3.642998156594619*^9, 
   3.6429981705783787`*^9}, {3.642998209228816*^9, 3.642998226037016*^9}, {
   3.643342810764079*^9, 3.643342901372982*^9}, {3.6433441297185802`*^9, 
   3.643344136821859*^9}, {3.6433633017490473`*^9, 3.64336333677936*^9}, {
   3.643366647855174*^9, 3.6433666643613663`*^9}, 3.64342875004373*^9, {
   3.643429364652094*^9, 3.64342936834761*^9}, {3.643436278855055*^9, 
   3.643436279463855*^9}, 3.6434396789778643`*^9, {3.649565372232395*^9, 
   3.649565373267929*^9}, {3.649565430488615*^9, 3.649565430968236*^9}},
 CellTags->{
  "ID:415149625", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R12"},
 CellID->415149625],

Cell[BoxData[
 RowBox[{
  RowBox[{"R13", "[", "D", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainSequences", "[", 
         RowBox[{"DomainSets", "[", "D", "]"}], "]"}]], 
       RowBox[{"[", "B", "]"}]}]], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"B", "[", "i", "]"}], 
        UnderscriptBox["\[SquareSubset]", "D"], 
        RowBox[{"B", "[", 
         RowBox[{"i", "+", "1"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R13)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1276925313", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64636374204963*^9, 3.646363745505535*^9}, {
  3.6463639719799643`*^9, 3.6463640642493057`*^9}, {3.646364099020432*^9, 
  3.646364103851014*^9}, {3.648282429476218*^9, 3.648282441330201*^9}, {
  3.652178090407147*^9, 3.6521781086198378`*^9}},
 CellTags->{
  "ID:1276925313", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R13"},
 CellID->1276925313],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1370612295]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->747506000]
}, Closed]],

Cell[CellGroupData[{

Cell["Lemmas (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6430099774627123`*^9, 3.643009978369302*^9}, {
   3.643010486147295*^9, 3.6430104906952477`*^9}, 3.64706336223827*^9, {
   3.6470706840345182`*^9, 3.64707068692834*^9}},
 CellID->530708085],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isReductionRing", "[", "D", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1684964128}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6430099983437443`*^9, 3.643010015644753*^9}},
 CellTags->{
  "ID:1684964128", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1684964128],

Cell["", "OpenEnvironment",
 CellID->253460494],

Cell[CellGroupData[{

Cell["Lemma: Consequences of (R9), (R10) and (R12)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6563192051085377`*^9, 3.656319220505765*^9}, {
   3.656319545330535*^9, 3.6563195516641464`*^9}, {3.656408820032876*^9, 
   3.6564088224352803`*^9}, 3.656408867129359*^9},
 CellID->114139003],

Cell[BoxData[
 RowBox[{
  RowBox[{"R9", "[", "D", "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]], 
     RowBox[{
      RowBox[{"c1", "\[NotEqual]", 
       UnderscriptBox["0", "D"]}], "\[And]", 
      RowBox[{"c2", "\[NotEqual]", 
       UnderscriptBox["0", "D"]}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{
         UnderscriptBox["multN", "D"], "[", "c1", "]"}]}]], 
      RowBox[{"j", "\[GreaterEqual]", "i"}]], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", 
         RowBox[{
          UnderscriptBox["lcrd", "D"], "[", 
          RowBox[{"c1", ",", "c2", ",", "i", ",", "j"}], "]"}], 
         "\[RightBracketingBar]"}]}]], 
      RowBox[{"(", 
       RowBox[{
        SubscriptBox[
         RowBox[{
          UnderscriptBox["lcrd", "D"], "[", 
          RowBox[{"c1", ",", "c2", ",", "i", ",", "j"}], "]"}], "k"], 
        "\[NotEqual]", 
        UnderscriptBox["0", "D"]}], ")"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(22)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:10673911", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656319235653391*^9, 3.6563193372251697`*^9}},
 CellTags->{
  "ID:10673911", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "22"},
 CellID->10673911],

Cell[BoxData[
 RowBox[{
  RowBox[{"R10", "[", "D", "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", "c", "]"}]}]], 
     RowBox[{"c", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", 
        RowBox[{
         UnderscriptBox["lcrd", "D"], "[", "c", "]"}], 
        "\[RightBracketingBar]"}]}]], 
     RowBox[{"(", 
      RowBox[{
       SubscriptBox[
        RowBox[{
         UnderscriptBox["lcrd", "D"], "[", "c", "]"}], "k"], "\[NotEqual]", 
       UnderscriptBox["0", "D"]}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(23)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:603265816", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656319359158808*^9, 3.6563193808906465`*^9}},
 CellTags->{
  "ID:603265816", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "23"},
 CellID->603265816],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"R9", "[", "D", "]"}], "\[And]", 
    RowBox[{"R12", "[", "D", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", 
        RowBox[{"c1", ",", "c2"}], "]"}]}]], 
     RowBox[{
      RowBox[{"c1", "\[NotEqual]", 
       UnderscriptBox["0", "D"]}], "\[And]", 
      RowBox[{"c2", "\[NotEqual]", 
       UnderscriptBox["0", "D"]}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{
         UnderscriptBox["multN", "D"], "[", "c1", "]"}]}]], 
      RowBox[{"i", "<", "j"}]], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", 
         RowBox[{
          UnderscriptBox["lcrd", "D"], "[", 
          RowBox[{"c1", ",", "c2", ",", "i", ",", "j"}], "]"}], 
         "\[RightBracketingBar]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             SubscriptBox[
              RowBox[{
               UnderscriptBox["cpm", "D"], "[", 
               RowBox[{"c1", ",", "c2", ",", 
                SubscriptBox[
                 RowBox[{
                  UnderscriptBox["lcrd", "D"], "[", 
                  RowBox[{"c1", ",", "c2", ",", "i", ",", "j"}], "]"}], "k"], 
                ",", "i", ",", "j"}], "]"}], "1"], "\[NotEqual]", 
             UnderscriptBox["0", "D"]}]},
           {
            RowBox[{
             SubscriptBox[
              RowBox[{
               UnderscriptBox["cpm", "D"], "[", 
               RowBox[{"c1", ",", "c2", ",", 
                SubscriptBox[
                 RowBox[{
                  UnderscriptBox["lcrd", "D"], "[", 
                  RowBox[{"c1", ",", "c2", ",", "i", ",", "j"}], "]"}], "k"], 
                ",", "i", ",", "j"}], "]"}], "2"], "\[NotEqual]", 
             UnderscriptBox["0", "D"]}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(24)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:39612209", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6563195613361635`*^9, 3.6563195796037955`*^9}, {
  3.6563196224882708`*^9, 3.656319659117135*^9}, {3.65631984407206*^9, 
  3.6563198579248843`*^9}, {3.6563207159333916`*^9, 3.6563207220642023`*^9}},
 CellTags->{
  "ID:39612209", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "24"},
 CellID->39612209],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->570587389]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->430224301],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(22)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(22)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R9", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                RowBox[{
                  StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}]}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R9", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                RowBox[{
                  StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}]}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:10673911"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p10673911-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p10673911-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:10673911", "Proof|ID:10673911-1"},
 CellID->26780681],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p10673911-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->48417729]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(23)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(23)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R10", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 StyleBox["D", "ExpressionVariable"]]}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 SubscriptBox[
                  RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], 
                  StyleBox["k", "ExpressionVariable"]], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  StyleBox[
                  "D", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R10", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", 
                 StyleBox["D", "ExpressionVariable"]]}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 SubscriptBox[
                  RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], 
                  StyleBox["k", "ExpressionVariable"]], "\[NotEqual]", 
                 UnderscriptBox["0", 
                  StyleBox[
                  "D", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:603265816"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p603265816-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p603265816-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:603265816", "Proof|ID:603265816-1"},
 CellID->237799576],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p603265816-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->91085372]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(24)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(24)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R9", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"R12", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                RowBox[{
                  StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}]}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], "<", 
                  StyleBox["j", "ExpressionVariable"]}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["cpm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "1"], 
                    "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                   RowBox[{
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["cpm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "2"], 
                    "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R9", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"R12", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                RowBox[{
                  StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}]}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], "<", 
                  StyleBox["j", "ExpressionVariable"]}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["cpm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "1"], 
                    "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                   RowBox[{
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["cpm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "2"], 
                    "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:39612209"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p39612209-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p39612209-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:39612209", "Proof|ID:39612209-1"},
 CellID->269494545],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p39612209-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->19274979]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1851834333],

Cell[CellGroupData[{

Cell["Lemma: Consequences of (R11)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.643009988476634*^9, 3.643009994206366*^9}, 
   3.6563192237973704`*^9},
 CellID->2034281086],

Cell[BoxData[
 RowBox[{
  RowBox[{"R11", "[", "D", "]"}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1515571182}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6495669893238792`*^9, 3.649566995562852*^9}},
 CellTags->{
  "ID:1515571182", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1515571182],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "c"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["rdm", "D"], "[", 
       RowBox[{"a", ",", "c"}], "]"}], "\[Equal]", 
      UnderscriptBox["0", "D"]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       UnderscriptBox["isReducible", "D"], "[", 
       RowBox[{"a", ",", "c"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(18)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:528027344", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.643010025415719*^9, 3.643010079880436*^9}, {
  3.6495669924677973`*^9, 3.649566998343656*^9}},
 CellTags->{
  "ID:528027344", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "18"},
 CellID->528027344],

Cell[BoxData[
 RowBox[{
  RowBox[{"R10", "[", "D", "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", "c", "]"}]}]], 
     RowBox[{"c", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", 
        RowBox[{
         UnderscriptBox["lcrd", "D"], "[", "c", "]"}], 
        "\[RightBracketingBar]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            UnderscriptBox["\[Element]", "D"], 
            RowBox[{"[", 
             RowBox[{
              UnderscriptBox["rdm", "D"], "[", 
              RowBox[{
               SubscriptBox[
                RowBox[{
                 UnderscriptBox["lcrd", "D"], "[", "c", "]"}], "k"], ",", 
               "c"}], "]"}], "]"}]}]},
          {
           RowBox[{
            RowBox[{
             UnderscriptBox["rdm", "D"], "[", 
             RowBox[{
              SubscriptBox[
               RowBox[{
                UnderscriptBox["lcrd", "D"], "[", "c", "]"}], "k"], ",", 
              "c"}], "]"}], "\[NotEqual]", 
            UnderscriptBox["0", "D"]}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(25)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:94925229", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656408732077922*^9, 3.6564088014366436`*^9}, 
   3.656408833027699*^9, {3.6564088719809675`*^9, 3.6564088731509695`*^9}, {
   3.6564089237272587`*^9, 3.656408937299282*^9}, {3.656409029979045*^9, 
   3.6564090357510552`*^9}, {3.6564091885377235`*^9, 3.6564091908777275`*^9}},
 
 CellTags->{
  "ID:94925229", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "25"},
 CellID->94925229],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "c1", ",", "c2"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", "j"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"c1", "\[NotEqual]", "c2"}], "\[And]", 
       RowBox[{
        UnderscriptBox["isMNTCR", "D"], "[", 
        RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["areCPMultipliers", "D"], "[", 
      RowBox[{
       RowBox[{
        UnderscriptBox["rdm", "D"], "[", 
        RowBox[{"a", ",", "c1"}], "]"}], ",", 
       RowBox[{
        UnderscriptBox["rdm", "D"], "[", 
        RowBox[{"a", ",", "c2"}], "]"}], ",", "c1", ",", "c2", ",", "a", ",", 
       "i", ",", "j"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(18')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2104251234", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649567009894372*^9, 3.649567088760509*^9}, {
  3.64956716653403*^9, 3.649567175689579*^9}},
 CellTags->{
  "ID:2104251234", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "18'"},
 CellID->2104251234],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->897100354]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->87827575],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(18)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(18)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R11", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["rdm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 UnderscriptBox["0", 
                  StyleBox["D", "ExpressionVariable"]]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R11", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["rdm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 UnderscriptBox["0", 
                  StyleBox["D", "ExpressionVariable"]]}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                   TagBox[
                   ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:528027344"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p528027344-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p528027344-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:528027344", "Proof|ID:528027344-1"},
 CellID->492864454],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p528027344-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->90487257]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(25)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(25)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R11", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"R10", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    UnderscriptBox["rdm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "]"}], 
                   "\[And]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["rdm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                    "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R11", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"R10", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    UnderscriptBox["rdm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "]"}], 
                   "\[And]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["rdm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                    "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:94925229"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p94925229-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p94925229-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:94925229", "Proof|ID:94925229-1"},
 CellID->568093765],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p94925229-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->28273598]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(18')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(18')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R11", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], ",", 
                  StyleBox["j", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["areCPMultipliers", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["rdm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["rdm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"]}], "]"}], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R11", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], ",", 
                  StyleBox["j", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["areCPMultipliers", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["rdm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["rdm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"]}], "]"}], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2104251234"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2104251234-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2104251234-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2104251234", "Proof|ID:2104251234-1"},
 CellID->611079668],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2104251234-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->875163593]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1971101244],

Cell[CellGroupData[{

Cell["Lemma: Sufficient Conditions for, and Consequences of, (R13)", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.647063010418126*^9, 3.647063016456216*^9}, 
   3.6563192266053753`*^9, {3.6586426226255455`*^9, 3.658642634762367*^9}},
 CellID->1398096634],

Cell[BoxData[
 RowBox[{
  RowBox[{"isFinDomain", "[", "D", "]"}], "\[Implies]", 
  RowBox[{"R13", "[", "D", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isFinDomain \[DoubleRightArrow] R13)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:484595803", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658642642843181*^9, 3.658642652156398*^9}},
 CellTags->{
  "ID:484595803", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isFinDomain \[DoubleRightArrow] R13"},
 CellID->484595803],

Cell[BoxData[
 RowBox[{
  RowBox[{"R13", "[", "D", "]"}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 55361704}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648282700281725*^9, 3.648282703581142*^9}},
 CellTags->{
  "ID:55361704", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->55361704],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSequences", "[", "D", "]"}]], 
    RowBox[{"[", "c", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"k", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["red", "D"], "[", 
       RowBox[{
        RowBox[{"setSeq", "[", "c", "]"}], "[", "i", "]"}], "]"}], "\[Equal]", 
      RowBox[{
       UnderscriptBox["red", "D"], "[", 
       RowBox[{
        RowBox[{"setSeq", "[", "c", "]"}], "[", 
        RowBox[{"i", "+", "k"}], "]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(20)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:283927726", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647063020543255*^9, 3.647063090390867*^9}, {
   3.647063142128825*^9, 3.6470631563191032`*^9}, {3.647070199807456*^9, 
   3.647070206069058*^9}, {3.647070272385275*^9, 3.647070292901287*^9}, 
   3.648282706619824*^9},
 CellTags->{
  "ID:283927726", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "20"},
 CellID->283927726],

Cell[TextData[{
 "Under the presence of (R13), \[SquareSubset] is Noetherian on ",
 Cell[BoxData[
  FormBox[
   RowBox[{"DomainSets", "[", "D", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"DomainTuples", "[", "D", "]"}], TraditionalForm]]],
 ":"
}], "EnvironmentText",
 CellChangeTimes->{{3.64828527295111*^9, 3.6482853083769913`*^9}, {
  3.6536326828306665`*^9, 3.6536326843438697`*^9}},
 CellID->536961276],

Cell[BoxData[
 RowBox[{"isNoetherian", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[SquareSubset]", "D"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   RowBox[{"DomainSets", "[", "D", "]"}]}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[SquareSubset] isNoetherian sets)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1584214935", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64828438318579*^9, 3.648284386406657*^9}},
 CellTags->{
  "ID:1584214935", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isNoetherian sets"},
 CellID->1584214935],

Cell[BoxData[
 RowBox[{"isNoetherian", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[SquareSubset]", "D"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   RowBox[{"DomainTuples", "[", "D", "]"}]}], "]"}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[SquareSubset] isNoetherian tuples)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1231517070", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64828271177312*^9, 3.6482827407690077`*^9}},
 CellTags->{
  "ID:1231517070", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isNoetherian tuples"},
 CellID->1231517070],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1038601912]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1995689151],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isFinDomain \[DoubleRightArrow] R13)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isFinDomain \[DoubleRightArrow] R13)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"isFinDomain", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{"R13", "[", 
             StyleBox["D", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"isFinDomain", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{"R13", "[", 
             StyleBox["D", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:484595803"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p484595803-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p484595803-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:484595803", "Proof|ID:484595803-1"},
 CellID->419106276],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p484595803-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->913223413]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(20)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(20)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R13", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSequences", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                   "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R13", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSequences", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], "]"}], 
                   "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{"setSeq", "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:283927726"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p283927726-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p283927726-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:283927726", "Proof|ID:283927726-1"},
 CellID->309385559],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p283927726-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->103041209]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[SquareSubset] isNoetherian sets)\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(\[SquareSubset] isNoetherian sets)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R13", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{"isNoetherian", "[", 
             RowBox[{
               UnderscriptBox["\[SquareSubset]", 
                StyleBox["D", "ExpressionVariable"]], ",", 
               RowBox[{"DomainSets", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R13", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{"isNoetherian", "[", 
             RowBox[{
               UnderscriptBox["\[SquareSubset]", 
                StyleBox["D", "ExpressionVariable"]], ",", 
               RowBox[{"DomainSets", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1584214935"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1584214935-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1584214935-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1584214935", "Proof|ID:1584214935-1"},
 CellID->2080577831],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1584214935-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->977712109]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[SquareSubset] isNoetherian tuples)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[SquareSubset] isNoetherian tuples)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R13", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{"isNoetherian", "[", 
             RowBox[{
               UnderscriptBox["\[SquareSubset]", 
                StyleBox["D", "ExpressionVariable"]], ",", 
               RowBox[{"DomainTuples", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R13", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{"isNoetherian", "[", 
             RowBox[{
               UnderscriptBox["\[SquareSubset]", 
                StyleBox["D", "ExpressionVariable"]], ",", 
               RowBox[{"DomainTuples", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1231517070"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1231517070-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1231517070-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1231517070", "Proof|ID:1231517070-1"},
 CellID->120554530],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1231517070-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->439739257]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["First Lemmas and Theorems (\[Checkmark])", "Section",
 CellChangeTimes->{{3.6442214804110823`*^9, 3.644221483208549*^9}, {
  3.644221569704982*^9, 3.6442215880379744`*^9}, {3.644221716269125*^9, 
  3.6442217182979803`*^9}, {3.644223127795165*^9, 3.644223133952664*^9}, {
  3.645259056051283*^9, 3.645259060076241*^9}, {3.6479431158491*^9, 
  3.647943116780382*^9}, {3.6481874108961973`*^9, 3.648187413467654*^9}, {
  3.650260751120306*^9, 3.6502607521914473`*^9}, {3.650792092078389*^9, 
  3.650792094632201*^9}},
 CellID->1796300417],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " Satisfies Only Some of the Reduction Ring Axioms (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.644223152280714*^9, 3.644223170121332*^9}, {
  3.644223994163383*^9, 3.6442240016801157`*^9}, {3.644224078119632*^9, 
  3.6442241020485363`*^9}, {3.6443004240464163`*^9, 3.6443004282280493`*^9}},
 CellID->2106967378],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isCommRing1", "[", "D", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1326788334}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632200714349184*^9, 3.632200720060936*^9}, {
  3.641705580603561*^9, 3.641705587214377*^9}},
 CellTags->{
  "ID:1326788334", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1326788334],

Cell["", "OpenEnvironment",
 CellID->1436091075],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: reduceM, symred and \[Succeeds] for Tuples of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"A", 
     TagBox["\:22c8",
      Identity,
      SyntaxForm->"a*b"], "B"}], ")"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "m"}], ")"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632128816855658*^9, 3.632128819097728*^9}, {
  3.632129067907566*^9, 3.6321291037165213`*^9}, {3.632129500109662*^9, 
  3.632129514309803*^9}, {3.6321367137133503`*^9, 3.632136714124071*^9}, {
  3.6322009605875587`*^9, 3.632200967630886*^9}},
 CellID->930024145],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "D", "]"}]], 
   RowBox[{"[", 
    RowBox[{"A", ",", "B"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 227700447}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6321291234746113`*^9, 3.632129139009837*^9}, {
  3.635748955814098*^9, 3.635748962526999*^9}},
 CellTags->{
  "ID:227700447", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->227700447],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "c", ",", "d", ",", "m"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1128391937}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632129142969599*^9, 3.632129156320694*^9}, {
  3.632203298045713*^9, 3.632203315627968*^9}},
 CellTags->{
  "ID:1128391937", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1128391937],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{
   RowBox[{"T", "=", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", 
      TagBox["\:22c8",
       Identity,
       SyntaxForm->"a*b"], "B"}], 
     TagBox[")",
      "AutoParentheses"]}]}], ",", 
   RowBox[{"S", "=", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "m"}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 788600821}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632129170199286*^9, 3.632129189083406*^9}, {
  3.632129419255468*^9, 3.632129430816585*^9}},
 CellTags->{
  "ID:788600821", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->788600821],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"b", "=", 
   RowBox[{
    UnderscriptBox["reduceM", "D"], "[", 
    RowBox[{"a", ",", "c", ",", "A", ",", 
     RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 999322294}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632129308857327*^9, 3.6321293256164207`*^9}},
 CellTags->{
  "ID:999322294", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->999322294],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "D"], 
  RowBox[{"[", "b", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(392)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:737948621", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6418008699028893`*^9, 3.641800875159563*^9}},
 CellTags->{
  "ID:737948621", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "392"},
 CellID->737948621],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "0"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}], "\[Equal]", 
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "A", ",", "i"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(394)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:287016951", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632129679297525*^9, 3.632129784898929*^9}, 
   3.6321299195062933`*^9, {3.6321301608439302`*^9, 3.6321301686089478`*^9}},
 CellTags->{
  "ID:287016951", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "394"},
 CellID->287016951],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "0"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}], "+", 
       "i"}]}], "]"}], "\[Equal]", 
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"b", ",", "c", ",", "B", ",", "i"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(395)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1812129266", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632130176240587*^9, 3.63213019451283*^9}},
 CellTags->{
  "ID:1812129266", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "395"},
 CellID->1812129266],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{
          UnderscriptBox["symred", "D"], "[", 
          RowBox[{
           RowBox[{
            UnderscriptBox["reduceM", "D"], "[", 
            RowBox[{"a", ",", "c", ",", "A", ",", 
             RowBox[{"i", "-", "1"}]}], "]"}], ",", 
           RowBox[{
            UnderscriptBox["reduceM", "D"], "[", 
            RowBox[{"a", ",", "c", ",", "A", ",", "i"}], "]"}], ",", "c", ",", 
           SubscriptBox["A", "i"]}], "]"}]}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{
          UnderscriptBox["symred", "D"], "[", 
          RowBox[{
           RowBox[{
            UnderscriptBox["reduceM", "D"], "[", 
            RowBox[{"b", ",", "c", ",", "B", ",", 
             RowBox[{"i", "-", "1"}]}], "]"}], ",", 
           RowBox[{
            UnderscriptBox["reduceM", "D"], "[", 
            RowBox[{"b", ",", "c", ",", "B", ",", "i"}], "]"}], ",", "c", ",", 
           SubscriptBox["B", "i"]}], "]"}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     UnderscriptBox["symred", "D"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["reduceM", "D"], "[", 
       RowBox[{"a", ",", "c", ",", "T", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", 
      RowBox[{
       UnderscriptBox["reduceM", "D"], "[", 
       RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}], ",", "c", ",", 
      SubscriptBox["T", "i"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(396)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1191851294", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632129163823386*^9, 3.6321291660218477`*^9}, {
  3.63212919803833*^9, 3.632129385464056*^9}, {3.63212945630192*^9, 
  3.632129456816203*^9}},
 CellTags->{
  "ID:1191851294", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "396"},
 CellID->1191851294],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{
          UnderscriptBox["symred", "D"], "[", 
          RowBox[{
           RowBox[{
            UnderscriptBox["reduceM", "D"], "[", 
            RowBox[{"a", ",", "c", ",", "A", ",", 
             RowBox[{"i", "-", "1"}]}], "]"}], ",", 
           RowBox[{
            UnderscriptBox["reduceM", "D"], "[", 
            RowBox[{"a", ",", "c", ",", "A", ",", "i"}], "]"}], ",", "c", ",", 
           SubscriptBox["A", "i"]}], "]"}]}]},
       {
        RowBox[{
         UnderscriptBox["symred", "D"], "[", 
         RowBox[{"b", ",", 
          RowBox[{"b", 
           UnderscriptBox["-", "D"], 
           RowBox[{"m", 
            UnderscriptBox["*", "D"], "c"}]}], ",", "c", ",", "m"}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     UnderscriptBox["symred", "D"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["reduceM", "D"], "[", 
       RowBox[{"a", ",", "c", ",", "S", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", 
      RowBox[{
       UnderscriptBox["reduceM", "D"], "[", 
       RowBox[{"a", ",", "c", ",", "S", ",", "i"}], "]"}], ",", "c", ",", 
      SubscriptBox["S", "i"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(397)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:101106535", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632129470121475*^9, 3.632129486836091*^9}, {
  3.632129589755678*^9, 3.632129616827887*^9}},
 CellTags->{
  "ID:101106535", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "397"},
 CellID->101106535],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{"d", 
          UnderscriptBox["\[Succeeds]", "D"], 
          RowBox[{
           UnderscriptBox["reduceM", "D"], "[", 
           RowBox[{"a", ",", "c", ",", "A", ",", "i"}], "]"}]}]}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{"d", 
          UnderscriptBox["\[Succeeds]", "D"], 
          RowBox[{
           UnderscriptBox["reduceM", "D"], "[", 
           RowBox[{"b", ",", "c", ",", "B", ",", "i"}], "]"}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
    RowBox[{"d", 
     UnderscriptBox["\[Succeeds]", "D"], 
     RowBox[{
      UnderscriptBox["reduceM", "D"], "[", 
      RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(398)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1737800130", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6322009945339317`*^9, 3.632201032587946*^9}, {
  3.632201259046118*^9, 3.6322012891121407`*^9}, {3.632201702275983*^9, 
  3.63220170378762*^9}, {3.632203304721498*^9, 3.632203331361814*^9}},
 CellTags->{
  "ID:1737800130", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "398"},
 CellID->1737800130],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{"d", 
          UnderscriptBox["\[Succeeds]", "D"], 
          RowBox[{
           UnderscriptBox["reduceM", "D"], "[", 
           RowBox[{"a", ",", "c", ",", "A", ",", "i"}], "]"}]}]}]},
       {
        RowBox[{"d", 
         UnderscriptBox["\[Succeeds]", "D"], 
         RowBox[{"b", 
          UnderscriptBox["-", "D"], 
          RowBox[{"m", 
           UnderscriptBox["*", "D"], "c"}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{"d", 
     UnderscriptBox["\[Succeeds]", "D"], 
     RowBox[{
      UnderscriptBox["reduceM", "D"], "[", 
      RowBox[{"a", ",", "c", ",", "S", ",", "i"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(399)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1568772314", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632201315827388*^9, 3.6322013228573923`*^9}, {
   3.632201353287764*^9, 3.632201374691559*^9}, 3.632201706403908*^9, {
   3.6322033353095293`*^9, 3.632203338840176*^9}},
 CellTags->{
  "ID:1568772314", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "399"},
 CellID->1568772314],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1523433723]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->120405495],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(392)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(392)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            StyleBox["D", "ExpressionVariable"]], "[", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            StyleBox["D", "ExpressionVariable"]], "[", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:737948621"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p737948621-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p737948621-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:737948621", "Proof|ID:737948621-1"},
 CellID->401948380],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p737948621-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1689050202]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(394)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(394)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:22c8", 
                 StyleBox["B", "ExpressionVariable"]}], ",", 
               StyleBox["i", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["A", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:22c8", 
                 StyleBox["B", "ExpressionVariable"]}], ",", 
               StyleBox["i", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["A", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:287016951"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p287016951-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p287016951-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:287016951", "Proof|ID:287016951-1"},
 CellID->766901523],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p287016951-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->87600598]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(395)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(395)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:22c8", 
                 StyleBox["B", "ExpressionVariable"]}], ",", 
               RowBox[{
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["A", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], "+", 
                 StyleBox["i", "ExpressionVariable"]}]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["A", "ExpressionVariable"], "\:22c8", 
                 StyleBox["B", "ExpressionVariable"]}], ",", 
               RowBox[{
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["A", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], "+", 
                 StyleBox["i", "ExpressionVariable"]}]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["A", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1812129266"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1812129266-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1812129266-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1812129266", "Proof|ID:1812129266-1"},
 CellID->967901432],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1812129266-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->82281228]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(396)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(396)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:22c8", 
                    StyleBox["B", "ExpressionVariable"]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["symred", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:22c8", 
                    StyleBox["B", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                 ",", 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:22c8", 
                    StyleBox["B", "ExpressionVariable"]}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:22c8", 
                    StyleBox["B", "ExpressionVariable"]}], 
                  StyleBox["i", "ExpressionVariable"]]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:22c8", 
                    StyleBox["B", "ExpressionVariable"]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["symred", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:22c8", 
                    StyleBox["B", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                 ",", 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:22c8", 
                    StyleBox["B", "ExpressionVariable"]}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:22c8", 
                    StyleBox["B", "ExpressionVariable"]}], 
                  StyleBox["i", "ExpressionVariable"]]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1191851294"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1191851294-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1191851294-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1191851294", "Proof|ID:1191851294-1"},
 CellID->155176372],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1191851294-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.641811045248006*^9},
 CellID->206911940]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(397)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(397)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["symred", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["m", "ExpressionVariable"]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["symred", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["m", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                 ",", 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["m", "ExpressionVariable"]}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["m", "ExpressionVariable"]}], 
                  StyleBox["i", "ExpressionVariable"]]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["A", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["symred", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["m", "ExpressionVariable"]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["symred", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["m", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                 ",", 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["m", "ExpressionVariable"]}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["m", "ExpressionVariable"]}], 
                  StyleBox["i", "ExpressionVariable"]]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:101106535"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p101106535-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p101106535-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:101106535", "Proof|ID:101106535-1"},
 CellID->1074757864],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p101106535-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.641811468267219*^9},
 CellID->1512156001]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(398)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(398)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 StyleBox["d", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}]}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 StyleBox["d", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:22c8", 
                    StyleBox["B", "ExpressionVariable"]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:22c8", 
                    StyleBox["B", "ExpressionVariable"]}], ",", 
                   StyleBox["i", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 StyleBox["d", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}]}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 StyleBox["d", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:22c8", 
                    StyleBox["B", "ExpressionVariable"]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:22c8", 
                    StyleBox["B", "ExpressionVariable"]}], ",", 
                   StyleBox["i", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1737800130"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1737800130-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1737800130-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1737800130", "Proof|ID:1737800130-1"},
 CellID->452671936],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1737800130-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1450652992]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(399)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(399)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 StyleBox["d", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}]}]}], 
             "\[And]", 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}]}]}], "\[Implies]", 
           
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["m", "ExpressionVariable"]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["m", "ExpressionVariable"]}], ",", 
                   StyleBox["i", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["A", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 StyleBox["d", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}]}]}], 
             "\[And]", 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["A", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}]}]}], "\[Implies]", 
           
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["m", "ExpressionVariable"]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\:293a", 
                    StyleBox["m", "ExpressionVariable"]}], ",", 
                   StyleBox["i", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1568772314"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1568772314-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1568772314-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1568772314", "Proof|ID:1568772314-1"},
 CellID->1787487969],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1568772314-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->702767096]
}, Closed]],

Cell[TextData[{
 StyleBox["Remark",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 ": The proofs of (396) and (398) are mainly based on (394) and (395)."
}], "Text",
 CellChangeTimes->{{3.632202093755033*^9, 3.632202113428913*^9}},
 CellID->794210953],

Cell["", "OpenEnvironment",
 CellID->1645079720],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: reduceM, symred and \[Succeeds] for Tuples of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"-", 
      SubscriptBox["T", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], "-", 
        "i"}]]}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632135719299423*^9, 3.632135757247447*^9}, {
  3.632136706612421*^9, 3.632136710247024*^9}, {3.6322014392538443`*^9, 
  3.632201445696053*^9}},
 CellID->334577641],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "D", "]"}]], 
   RowBox[{"[", "T", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1597110548}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6321291234746113`*^9, 3.632129139009837*^9}, {
  3.632135780900228*^9, 3.63213578174971*^9}, {3.63574898054534*^9, 
  3.6357489871716633`*^9}},
 CellTags->{
  "ID:1597110548", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1597110548],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "c", ",", "d"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1509853174}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632129142969599*^9, 3.632129156320694*^9}, 
   3.6321357855865726`*^9, {3.632203620026883*^9, 3.6322036202076263`*^9}},
 CellTags->{
  "ID:1509853174", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1509853174],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{
   RowBox[{"S", "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["-", "D"], 
       SubscriptBox["T", 
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], "-", 
         
         RowBox[{"(", 
          RowBox[{"i", "-", "1"}], ")"}]}]]}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}]}], ",", 
   RowBox[{"b", "=", 
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
     "]"}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1554200522}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632129170199286*^9, 3.632129189083406*^9}, {
  3.632129419255468*^9, 3.632129430816585*^9}, {3.632135792443199*^9, 
  3.6321358162677794`*^9}, {3.6321359902588177`*^9, 3.632136025290172*^9}, {
  3.632136811204156*^9, 3.632136832997767*^9}, {3.632138592533454*^9, 
  3.632138596369584*^9}, {3.632138965515918*^9, 3.632138977505934*^9}},
 CellTags->{
  "ID:1554200522", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1554200522],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "0"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"b", ",", "c", ",", "S", ",", "i"}], "]"}], "\[Equal]", 
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], "-", 
       "i"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(400)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1963951655", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632136569685338*^9, 3.632136622363227*^9}},
 CellTags->{
  "ID:1963951655", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "400"},
 CellID->1963951655],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"R1", "[", "D", "]"}], "\[And]", 
   RowBox[{"isPartIrreflOrder", "[", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       UnderscriptBox["\[Succeeds]", "D"],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}], ",", "D"}], "]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 85295514}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641815075027905*^9, 3.641815076194603*^9}, {
  3.64214101208559*^9, 3.6421410157221746`*^9}},
 CellTags->{
  "ID:85295514", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->85295514],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     UnderscriptBox["symred", "D"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["reduceM", "D"], "[", 
       RowBox[{"a", ",", "c", ",", "T", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", 
      RowBox[{
       UnderscriptBox["reduceM", "D"], "[", 
       RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}], ",", "c", ",", 
      SubscriptBox["T", "i"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     UnderscriptBox["symred", "D"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["reduceM", "D"], "[", 
       RowBox[{"b", ",", "c", ",", "S", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", 
      RowBox[{
       UnderscriptBox["reduceM", "D"], "[", 
       RowBox[{"b", ",", "c", ",", "S", ",", "i"}], "]"}], ",", "c", ",", 
      SubscriptBox["S", "i"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(401)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:751169803", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6321358209295883`*^9, 3.6321358928364067`*^9}, {
  3.6321360284292707`*^9, 3.632136095294256*^9}},
 CellTags->{
  "ID:751169803", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "401"},
 CellID->751169803],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"d", 
         UnderscriptBox["\[Succeeds]", "D"], "a"}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{"d", 
          UnderscriptBox["\[Succeeds]", "D"], 
          RowBox[{
           UnderscriptBox["reduceM", "D"], "[", 
           RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{"d", 
     UnderscriptBox["\[Succeeds]", "D"], 
     RowBox[{
      UnderscriptBox["reduceM", "D"], "[", 
      RowBox[{"b", ",", "c", ",", "S", ",", "i"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(402')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:46875747", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632208262599021*^9, 3.632208271605549*^9}},
 CellTags->{
  "ID:46875747", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "402'"},
 CellID->46875747],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"d", 
         UnderscriptBox["\[Succeeds]", "D"], "a"}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{"a", 
          UnderscriptBox["\[Succeeds]", "D"], 
          RowBox[{
           UnderscriptBox["reduceM", "D"], "[", 
           RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{"d", 
     UnderscriptBox["\[Succeeds]", "D"], 
     RowBox[{
      UnderscriptBox["reduceM", "D"], "[", 
      RowBox[{"b", ",", "c", ",", "S", ",", "i"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(402)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:656927872", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6322014522519503`*^9, 3.632201536451172*^9}, {
   3.632202371221116*^9, 3.632202388796707*^9}, {3.6322036246274633`*^9, 
   3.632203628789414*^9}, {3.632203959890698*^9, 3.632203986011541*^9}, 
   3.63220418777064*^9},
 CellTags->{
  "ID:656927872", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "402"},
 CellID->656927872],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->460382234]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->122940993],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(400)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(400)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                   "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
         
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["i", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["T", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], "-", 
                 StyleBox["i", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                   "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
         
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["i", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["T", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], "-", 
                 StyleBox["i", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1963951655"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1963951655-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1963951655-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1963951655", "Proof|ID:1963951655-1"},
 CellID->208873331],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1963951655-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->74193188]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(401)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(401)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
             "\[Implies]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
             "\[Implies]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:751169803"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p751169803-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p751169803-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:751169803", "Proof|ID:751169803-1"},
 CellID->553193299],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p751169803-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->199842648]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(402')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(402')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["d", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["a", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   StyleBox["d", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}]}]}]}], 
             "\[Implies]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 StyleBox["d", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["d", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["a", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   StyleBox["d", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}]}]}]}], 
             "\[Implies]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 StyleBox["d", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:46875747"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p46875747-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p46875747-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:46875747", "Proof|ID:46875747-1"},
 CellID->588348976],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p46875747-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1800386]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(402)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(402)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R1", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"isPartIrreflOrder", "[", 
                 RowBox[{
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], ",", 
                   StyleBox["D", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["d", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}]}]}]}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    StyleBox["d", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"R1", "[", 
             StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{"isPartIrreflOrder", "[", 
                 RowBox[{
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], ",", 
                   StyleBox["D", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["d", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}]}]}]}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    StyleBox["d", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:656927872"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p656927872-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p656927872-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:656927872", "Proof|ID:656927872-1"},
 CellID->574764629],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p656927872-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->992096306]
}, Closed]],

Cell[TextData[{
 StyleBox["Remark",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 ": The proofs of (401) and (402\[CloseCurlyQuote]) are mainly based on (400)."
}], "Text",
 CellChangeTimes->{{3.632202093755033*^9, 3.632202113428913*^9}, {
  3.6418148842833967`*^9, 3.641814894876313*^9}},
 CellID->1802708008],

Cell["", "OpenEnvironment",
 CellID->1946059673],

Cell[CellGroupData[{

Cell["Lemma: Consequences of areR5Witnesses", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.642138560898587*^9, 3.6421385707845297`*^9}},
 CellID->1305345464],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "D", "]"}]], 
   RowBox[{"[", 
    RowBox[{"TM", ",", "TN"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 502021608}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642138574693171*^9, 3.642138593972684*^9}},
 CellTags->{
  "ID:502021608", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->502021608],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "c", ",", "m1", ",", "m2"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 227587436}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642138600926693*^9, 3.6421386093827353`*^9}, 
   3.642138667521389*^9, {3.642138715462057*^9, 3.642138717712554*^9}, {
   3.642139542033736*^9, 3.642139546097659*^9}, 3.642140246033347*^9},
 CellTags->{
  "ID:227587436", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->227587436],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areR5Witnesses", "D"], "[", 
   RowBox[{"TM", ",", "TN", ",", "a", ",", 
    RowBox[{"a", 
     UnderscriptBox["-", "D"], 
     RowBox[{"m2", 
      UnderscriptBox["*", "D"], "c"}]}], ",", "c", ",", 
    RowBox[{
     RowBox[{
      UnderscriptBox["-", "D"], "m1"}], 
     UnderscriptBox["*", "D"], "c"}], ",", "m2"}], "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 937292216}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6421386719520483`*^9, 3.6421386859069843`*^9}, {
   3.6421387197065487`*^9, 3.6421387216122847`*^9}, {3.642139550269354*^9, 
   3.6421395762153187`*^9}, 3.642139609147544*^9, 3.642139741773059*^9, {
   3.6421402508140163`*^9, 3.642140270023587*^9}},
 CellTags->{
  "ID:937292216", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->937292216],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["reduceM", "D"], "[", 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["-", "D"], 
     RowBox[{"m1", 
      UnderscriptBox["*", "D"], "c"}]}], ",", "c", ",", "TM", ",", 
    RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}], "]"}], 
  "\[Equal]", 
  RowBox[{
   UnderscriptBox["reduceM", "D"], "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"a", 
       UnderscriptBox["-", "D"], 
       RowBox[{"m2", 
        UnderscriptBox["*", "D"], "c"}]}], ")"}], 
     UnderscriptBox["-", "D"], 
     RowBox[{"m1", 
      UnderscriptBox["*", "D"], "c"}]}], ",", "c", ",", "TN", ",", 
    RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(410)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1763687895", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6421388387278347`*^9, 3.642138965214473*^9}, {
  3.6421395815769053`*^9, 3.6421396125441713`*^9}, {3.642140276792863*^9, 
  3.6421402869760313`*^9}},
 CellTags->{
  "ID:1763687895", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "410"},
 CellID->1763687895],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"S", "=", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["-", "D"], 
      SubscriptBox["TN", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}], "-", 
        
        RowBox[{"(", 
         RowBox[{"j", "-", "1"}], ")"}]}]]}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 397073436}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642139074612275*^9, 3.642139112263709*^9}},
 CellTags->{
  "ID:397073436", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->397073436],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["reduceM", "D"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{
      RowBox[{"a", 
       UnderscriptBox["-", "D"], 
       RowBox[{"m1", 
        UnderscriptBox["*", "D"], "c"}]}], ",", "c", ",", "TM", ",", 
      RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}], 
     "]"}], ",", "c", ",", "S", ",", 
    RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}], "]"}], 
  "\[Equal]", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"a", 
     UnderscriptBox["-", "D"], 
     RowBox[{"m2", 
      UnderscriptBox["*", "D"], "c"}]}], ")"}], 
   UnderscriptBox["-", "D"], 
   RowBox[{"m1", 
    UnderscriptBox["*", "D"], "c"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(411)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:791651685", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6421389984971437`*^9, 3.642139051714828*^9}, {
  3.64213911801328*^9, 3.6421391549310293`*^9}, {3.6421396588546762`*^9, 
  3.642139670478189*^9}, {3.642140293993372*^9, 3.6421403021540613`*^9}},
 CellTags->{
  "ID:791651685", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "411"},
 CellID->791651685],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"T", "=", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"TM", 
     TagBox["\:22c8",
      Identity,
      SyntaxForm->"a*b"], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"S", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], 
       RowBox[{"(", 
        RowBox[{
         UnderscriptBox["-", "D"], "m1"}], ")"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 6804200}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642153858380896*^9, 3.642153893951337*^9}},
 CellTags->{
  "ID:6804200", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->6804200],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox[
      SubscriptBox["\[Sum]", "D"], 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
     SubscriptBox["T", "i"]}], ")"}], 
   UnderscriptBox["+", "D"], "m1"}], "\[Equal]", 
  "m2"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(414)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1294525434", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642220610009513*^9, 3.642220659868038*^9}},
 CellTags->{
  "ID:1294525434", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "414"},
 CellID->1294525434],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"R1", "[", "D", "]"}], "\[And]", 
   RowBox[{"isPartIrreflOrder", "[", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       UnderscriptBox["\[Succeeds]", "D"],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}], ",", "D"}], "]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 372638659}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.641815075027905*^9, 3.641815076194603*^9}, {
  3.64214101208559*^9, 3.6421410157221746`*^9}},
 CellTags->{
  "ID:372638659", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->372638659],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["-", "D"], 
     RowBox[{"m2", 
      UnderscriptBox["*", "D"], "c"}]}], ",", "c", ",", "m1"}], "]"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    UnderscriptBox["symred", "D"], "[", 
    RowBox[{
     RowBox[{
      UnderscriptBox["reduceM", "D"], "[", 
      RowBox[{
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{"m1", 
         UnderscriptBox["*", "D"], "c"}]}], ",", "c", ",", "T", ",", 
       RowBox[{"i", "-", "1"}]}], "]"}], ",", 
     RowBox[{
      UnderscriptBox["reduceM", "D"], "[", 
      RowBox[{
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{"m1", 
         UnderscriptBox["*", "D"], "c"}]}], ",", "c", ",", "T", ",", "i"}], 
      "]"}], ",", "c", ",", 
     SubscriptBox["T", "i"]}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(412)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1962716191", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642153940961444*^9, 3.6421540344804153`*^9}, {
  3.6421572424723463`*^9, 3.642157281675805*^9}},
 CellTags->{
  "ID:1962716191", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "412"},
 CellID->1962716191],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["\[Succeeds]", "D"], 
     RowBox[{"a", 
      UnderscriptBox["-", "D"], 
      RowBox[{"m1", 
       UnderscriptBox["*", "D"], "c"}]}]}], "\[And]", 
    RowBox[{"a", 
     UnderscriptBox["\[Succeeds]", "D"], 
     RowBox[{"a", 
      UnderscriptBox["-", "D"], 
      RowBox[{"m2", 
       UnderscriptBox["*", "D"], "c"}]}]}], "\[And]", 
    RowBox[{
     RowBox[{"a", 
      UnderscriptBox["-", "D"], 
      RowBox[{"m2", 
       UnderscriptBox["*", "D"], "c"}]}], 
     UnderscriptBox["\[Succeeds]", "D"], 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{"m2", 
         UnderscriptBox["*", "D"], "c"}]}], ")"}], 
      UnderscriptBox["-", "D"], 
      RowBox[{"m1", 
       UnderscriptBox["*", "D"], "c"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
   RowBox[{"a", 
    UnderscriptBox["\[Succeeds]", "D"], 
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{
      RowBox[{"a", 
       UnderscriptBox["-", "D"], 
       RowBox[{"m1", 
        UnderscriptBox["*", "D"], "c"}]}], ",", "c", ",", "T", ",", "i"}], 
     "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(413)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1575262565", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.642160272479727*^9, 3.642160353640812*^9}, {
   3.6421607820580683`*^9, 3.642160800894466*^9}, {3.642218297171282*^9, 
   3.6422183261880198`*^9}, 3.6422183649258137`*^9, {3.642218495562408*^9, 
   3.64221852473505*^9}},
 CellTags->{
  "ID:1575262565", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "413"},
 CellID->1575262565],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->982944345]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->222080009],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(410)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(410)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TM", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TN", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areR5Witnesses", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["TM", "ExpressionVariable"], ",", 
               StyleBox["TN", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["TM", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["TM", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["TN", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["TN", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TM", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TN", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areR5Witnesses", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["TM", "ExpressionVariable"], ",", 
               StyleBox["TN", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["TM", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["TM", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}]}], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["TN", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["TN", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1763687895"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1763687895-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1763687895-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1763687895", "Proof|ID:1763687895-1"},
 CellID->701556455],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1763687895-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1674393023]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(411)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(411)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TM", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TN", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areR5Witnesses", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["TM", "ExpressionVariable"], ",", 
               StyleBox["TN", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["TM", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TM", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                   "\[RightAngleBracket]"}], ",", 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}], 
               "]"}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TM", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TN", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areR5Witnesses", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["TM", "ExpressionVariable"], ",", 
               StyleBox["TN", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["TM", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TM", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "]"}], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                   "\[RightAngleBracket]"}], ",", 
                 RowBox[{"\[LeftBracketingBar]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}], 
               "]"}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:791651685"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p791651685-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p791651685-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:791651685", "Proof|ID:791651685-1"},
 CellID->2105672847],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p791651685-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2052754145]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(414)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(414)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TM", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TN", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areR5Witnesses", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["TM", "ExpressionVariable"], ",", 
               StyleBox["TN", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", 
                   StyleBox["D", "ExpressionVariable"]], 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 SubscriptBox[
                  RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], 
                  StyleBox["i", "ExpressionVariable"]]}], 
               UnderscriptBox["+", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["m1", "ExpressionVariable"]}], "\[Equal]", 
             StyleBox[
             "m2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TM", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TN", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areR5Witnesses", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["TM", "ExpressionVariable"], ",", 
               StyleBox["TN", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", 
                   StyleBox["D", "ExpressionVariable"]], 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 SubscriptBox[
                  RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], 
                  StyleBox["i", "ExpressionVariable"]]}], 
               UnderscriptBox["+", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["m1", "ExpressionVariable"]}], "\[Equal]", 
             StyleBox[
             "m2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1294525434"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1294525434-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1294525434-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1294525434", "Proof|ID:1294525434-1"},
 CellID->727439100],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1294525434-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->295763820]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(412)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(412)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TM", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TN", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areR5Witnesses", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["TM", "ExpressionVariable"], ",", 
               StyleBox["TN", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"R1", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{"isPartIrreflOrder", "[", 
                 RowBox[{
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], ",", 
                   StyleBox["D", "ExpressionVariable"]}], "]"}]}], 
             "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["m1", "ExpressionVariable"]}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["symred", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                    ",", 
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TM", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TN", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areR5Witnesses", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["TM", "ExpressionVariable"], ",", 
               StyleBox["TN", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"R1", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{"isPartIrreflOrder", "[", 
                 RowBox[{
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], ",", 
                   StyleBox["D", "ExpressionVariable"]}], "]"}]}], 
             "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["m1", "ExpressionVariable"]}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["symred", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                    ",", 
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1962716191"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1962716191-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1962716191-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1962716191", "Proof|ID:1962716191-1"},
 CellID->1885017557],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1962716191-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->137392427]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(413)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(413)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TM", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TN", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areR5Witnesses", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["TM", "ExpressionVariable"], ",", 
               StyleBox["TN", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"R1", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{"isPartIrreflOrder", "[", 
                 RowBox[{
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], ",", 
                   StyleBox["D", "ExpressionVariable"]}], "]"}]}], 
             "\[Implies]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TM", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["TN", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areR5Witnesses", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["TM", "ExpressionVariable"], ",", 
               StyleBox["TN", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"R1", "[", 
                 StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{"isPartIrreflOrder", "[", 
                 RowBox[{
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], ",", 
                   StyleBox["D", "ExpressionVariable"]}], "]"}]}], 
             "\[Implies]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], "\:22c8", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "-", "1"}], ")"}]}]],
                     "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}]}], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1575262565"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1575262565-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1575262565-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1575262565", "Proof|ID:1575262565-1"},
 CellID->1705894400],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1575262565-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->2053331622]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (412)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.6421541097731256`*^9, 3.6421541140178537`*^9}, 
   3.6421543709358664`*^9, 3.643533731592573*^9, 3.6532196291889825`*^9, 
   3.6661744458961153`*^9, 3.666441024477232*^9},
 CellID->1875389319],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", 
     RowBox[{"TM", ",", "TN"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "c", ",", "m1", ",", "m2"}], "]"}]}], 
   TraditionalForm]]],
 " a.b.f., let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", ":=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", 
       SubscriptBox["TN", 
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}], "-", 
         RowBox[{"(", 
          RowBox[{"j", "-", "1"}], ")"}]}]]}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", ":=", 
    RowBox[{"TM\:22c8", "(", 
     RowBox[{"S\:293a", "(", 
      RowBox[{"-", "m1"}], ")"}], ")"}]}], TraditionalForm]]],
 ", choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areR5Witnesses", "[", 
    RowBox[{"TM", ",", "TN", ",", "a", ",", 
     RowBox[{"a", "-", 
      RowBox[{"m2", "*", "c"}]}], ",", "c", ",", 
     RowBox[{
      RowBox[{"-", "m1"}], "*", "c"}], ",", "m2"}], "]"}], 
   TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"symred", "[", 
    RowBox[{
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "T", ",", 
       RowBox[{"i", "-", "1"}]}], "]"}], ",", 
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "T", ",", "i"}], "]"}], 
     ",", "c", ",", 
     SubscriptBox["T", "i"]}], "]"}], TraditionalForm]]],
 "\n\nFrom (1), by the definition of areR5Witnesses, we can infer (among \
others)\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}]], 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{"a", "-", 
         RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TM", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", "c", ",", 
      SubscriptBox["TM", "i"]}], "]"}]}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}]], 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"a", "-", 
           RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
         RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", "c", ",", 
      SubscriptBox["TN", "i"]}], "]"}]}], TraditionalForm]]],
 "\n\nSince ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " is of the form \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"A", 
    TagBox["\:22c8",
     Identity,
     SyntaxForm->"a*b"], 
    RowBox[{"(", 
     RowBox[{"B", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "m"}], ")"}]}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] we can use (396) to reduce the goal to (let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"U", ":=", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", "...", "\[RightAngleBracket]"}], 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], 
     RowBox[{"(", 
      RowBox[{"-", "m1"}], ")"}]}]}], TraditionalForm]]],
 ")\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}]], 
    RowBox[{"symred", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{"a", "-", 
         RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TM", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", 
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{"a", "-", 
         RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TM", ",", "i"}], "]"}], 
      ",", "c", ",", 
      SubscriptBox["TM", "i"]}], "]"}]}], TraditionalForm]]],
 " and\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "U", "\[RightBracketingBar]"}]}]], 
    RowBox[{"symred", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{"reduceM", "[", 
         RowBox[{
          RowBox[{"a", "-", 
           RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TM", ",", 
          RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}], 
         "]"}], ",", "c", ",", "U", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", 
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{"reduceM", "[", 
         RowBox[{
          RowBox[{"a", "-", 
           RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TM", ",", 
          RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}], 
         "]"}], ",", "c", ",", "U", ",", "i"}], "]"}], ",", "c", ",", 
      SubscriptBox["U", "i"]}], "]"}]}], TraditionalForm]]],
 ".\n\n(G1):\n(G1), by (symred reduceM II), can be reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TM", ",", 
       RowBox[{"i", "-", "1"}]}], "]"}], ",", "c", ",", 
     SubscriptBox["TM", "i"]}], "]"}], TraditionalForm]]],
 " for some a.b.f. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 ", which is just an instance of (2).\n\n(G2):\nWe use (410) to rewrite our \
current goal (G2) into\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"symred", "[", 
    RowBox[{
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{"reduceM", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{"a", "-", 
            RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
          RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", 
         RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}], 
        "]"}], ",", "c", ",", "U", ",", 
       RowBox[{"j", "-", "1"}]}], "]"}], ",", 
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{"reduceM", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{"a", "-", 
            RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
          RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", 
         RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}], 
        "]"}], ",", "c", ",", "U", ",", "j"}], "]"}], ",", "c", ",", 
     SubscriptBox["U", "j"]}], "]"}], TraditionalForm]]],
 "\nfor some a.b.f. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "U", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 ".\nBy (397), (G3) can be reduced to\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{"symred", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{"reduceM", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{"a", "-", 
             RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
           RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", 
          RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}], 
         "]"}], ",", "c", ",", "S", ",", 
        RowBox[{"k", "-", "1"}]}], "]"}], ",", 
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{"reduceM", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{"a", "-", 
             RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
           RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", 
          RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}], 
         "]"}], ",", "c", ",", "S", ",", "k"}], "]"}], ",", "c", ",", 
      SubscriptBox["S", "k"]}], "]"}]}], TraditionalForm]]],
 "\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"symred", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"a", "-", 
        RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
      RowBox[{"m1", "*", "c"}]}], ",", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"a", "-", 
          RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
        RowBox[{"m1", "*", "c"}]}], ")"}], "+", 
      RowBox[{"m1", "*", "c"}]}], ",", "c", ",", 
     RowBox[{"-", "m1"}]}], "]"}], TraditionalForm]]],
 ".\n\n(G4):\nChoose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 " a.b.f.; By (401) we can reduce our goal to\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"symred", "[", 
    RowBox[{
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"a", "-", 
          RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
        RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", 
       RowBox[{"k", "-", "1"}]}], "]"}], ",", 
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"a", "-", 
          RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
        RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", "k"}], "]"}], 
     ",", "c", ",", 
     SubscriptBox["TN", "k"]}], "]"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 " a.b.f.\nUsing (symred reduceM II) (G6) can be reduced to\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"a", "-", 
          RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
        RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", 
       RowBox[{"k", "-", "1"}]}], "]"}], ",", "c", ",", 
     SubscriptBox["TN", "k"]}], "]"}], TraditionalForm]]],
 "\nand this is just an instance of (3).\n\n(G5):\nUsing (symred symmetry) \
and ring arithmetic, (G5) can be rearranged to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"symred", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"m2", "*", "c"}]}], ",", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"a", "-", 
        RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
      RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "m1"}], "]"}], 
   TraditionalForm]]],
 ", which holds apparently."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{
  3.6421541393902683`*^9, 3.64215437093604*^9, {3.642155635848325*^9, 
   3.642155899152576*^9}, {3.642155940869422*^9, 3.642156244933921*^9}, {
   3.642156281880537*^9, 3.642156462772311*^9}, {3.6421565322257137`*^9, 
   3.6421565983528357`*^9}, {3.642156654074622*^9, 3.642156727945341*^9}, {
   3.642156777819504*^9, 3.642156805858985*^9}, {3.642156844121127*^9, 
   3.642156871138935*^9}, {3.642156904947105*^9, 3.642157169834729*^9}, 
   3.643533731592761*^9, 3.6532196291889825`*^9, 3.6661744458961153`*^9, 
   3.666441024477334*^9},
 CellID->1226616153],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.642154357722526*^9, 3.642154370936205*^9}, 
   3.643533731592928*^9, 3.6532196291889825`*^9, 3.6661744458961153`*^9, 
   3.6664410244774303`*^9},
 TextAlignment->Right,
 CellID->2115837209]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (413)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642160375558177*^9, 3.6421604008481073`*^9}, 
   3.643533736534749*^9, 3.653219622761771*^9, 3.6661744402323055`*^9, 
   3.6664410178099422`*^9},
 CellID->837540631],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", 
     RowBox[{"TM", ",", "TN"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "c", ",", "m1", ",", "m2"}], "]"}]}], 
   TraditionalForm]]],
 " a.b.f., let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", ":=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", 
       SubscriptBox["TN", 
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}], "-", 
         RowBox[{"(", 
          RowBox[{"j", "-", "1"}], ")"}]}]]}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", ":=", 
    RowBox[{"TM\:22c8", "(", 
     RowBox[{"S\:293a", "(", 
      RowBox[{"-", "m1"}], ")"}], ")"}]}], TraditionalForm]]],
 ", choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 " a.b.f., assume\n(0)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"a", "-", 
     RowBox[{"m1", "*", "c"}]}]}], TraditionalForm]]],
 "\n(0\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"a", "-", 
       RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
     RowBox[{"m1", "*", "c"}]}]}], TraditionalForm]]],
 "\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areR5Witnesses", "[", 
    RowBox[{"TM", ",", "TN", ",", "a", ",", 
     RowBox[{"a", "-", 
      RowBox[{"m2", "*", "c"}]}], ",", "c", ",", 
     RowBox[{
      RowBox[{"-", "m1"}], "*", "c"}], ",", "m2"}], "]"}], 
   TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"reduceM", "[", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "T", ",", "i"}], "]"}]}], 
   TraditionalForm]]],
 "\n\nFrom (1), by the definition of areR5Witnesses, we can infer (among \
others)\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}]], 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{"a", "-", 
         RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TM", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", "c", ",", 
      SubscriptBox["TM", "i"]}], "]"}]}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}]], 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"a", "-", 
           RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
         RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", "c", ",", 
      SubscriptBox["TN", "i"]}], "]"}]}], TraditionalForm]]],
 "\n\nSince ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " is of the form \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"A", 
    TagBox["\:22c8",
     Identity,
     SyntaxForm->"a*b"], 
    RowBox[{"(", 
     RowBox[{"B", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "m"}], ")"}]}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] we can use (398) to reduce the goal to (let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"U", ":=", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", "...", "\[RightAngleBracket]"}], 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], 
     RowBox[{"(", 
      RowBox[{"-", "m1"}], ")"}]}]}], TraditionalForm]]],
 ")\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}]], 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TM", ",", "j"}], 
      "]"}]}]}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "U", "\[RightBracketingBar]"}]}]], 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{"reduceM", "[", 
        RowBox[{
         RowBox[{"a", "-", 
          RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TM", ",", 
         RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}], 
        "]"}], ",", "c", ",", "U", ",", "j"}], "]"}]}]}], TraditionalForm]]],
 ".\n\n(G1):\nInstead of showing (G1) as it stands, we show\n(G1\
\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "TM", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"m1", "*", "c"}]}], "\[Succeeds]", 
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TM", ",", "j"}], 
      "]"}]}]}], TraditionalForm]]],
 "\nwhich, by (403), can be reduced to\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TM", ",", 
       RowBox[{"j", "-", "1"}]}], "]"}], ",", "c", ",", 
     SubscriptBox["TM", "j"]}], "]"}], TraditionalForm]]],
 "\nfor some a.b.f. ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 ", which is just an instance of (1). (G1) follows now readily from (0).\n\n\
(G2):\nWe first use (410) to replace (G2) by\n(G2\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"reduceM", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"a", "-", 
           RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
         RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", 
        RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}], 
       "]"}], ",", "c", ",", "U", ",", "j"}], "]"}]}], TraditionalForm]]],
 "\nfor some a.b.f. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "U", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 "\nBy (399), (G2\[CloseCurlyQuote]) can be reduced to\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{"reduceM", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{"a", "-", 
            RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
          RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", 
         RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}], 
        "]"}], ",", "c", ",", "S", ",", "i"}], "]"}]}]}], TraditionalForm]]],
 "\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{"reduceM", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{"a", "-", 
            RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
          RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", 
         RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}], 
        "]"}], ",", "c", ",", "S", ",", 
       RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}], 
      "]"}], "+", 
     RowBox[{"m1", "*", "c"}]}]}], TraditionalForm]]],
 ".\n\n(G4):\nBy (402), (G4) can be reduced to\n(G6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"a", "-", 
       RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
     RowBox[{"m1", "*", "c"}]}]}], TraditionalForm]]],
 "\n(G7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "TN", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"a", "-", 
        RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
      RowBox[{"m1", "*", "c"}]}], "\[Succeeds]", 
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"a", "-", 
          RowBox[{"m2", "*", "c"}]}], ")"}], "-", 
        RowBox[{"m1", "*", "c"}]}], ",", "c", ",", "TN", ",", "i"}], 
      "]"}]}]}], TraditionalForm]]],
 "\n(G6) is just (0\[CloseCurlyQuote]).\n(G7) follows from (3) by (403).\n\n\
(G5):\n(G5) can be rewritten to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"a", "-", 
     RowBox[{"m2", "*", "c"}]}]}], TraditionalForm]]],
 " (using (411)), which holds because of one of our very first assumptions."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642160400848282*^9, 3.642160629846891*^9}, {
   3.642160701755137*^9, 3.6421607223667393`*^9}, {3.642160809204564*^9, 
   3.642160890389604*^9}, {3.6422178666063766`*^9, 3.642217974208817*^9}, {
   3.642218037549646*^9, 3.642218051744937*^9}, {3.642218089776166*^9, 
   3.642218236826803*^9}, {3.642218374579626*^9, 3.642218399518485*^9}, {
   3.642218440545179*^9, 3.6422184609465723`*^9}, {3.642218539669456*^9, 
   3.642218589724104*^9}, 3.643533736534881*^9, 3.653219622761771*^9, 
   3.6661744402323055`*^9, 3.666441017810042*^9},
 CellID->1753611981],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.642160390539016*^9, 3.642160400848442*^9}, 
   3.643533736535013*^9, 3.653219622761771*^9, 3.6661744402323055`*^9, 
   3.6664410178101397`*^9},
 TextAlignment->Right,
 CellID->408639765]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1386249146],

Cell[CellGroupData[{

Cell["Theorem: Sufficient Conditions for isR6Witness", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.632120469463983*^9, 3.632120478768134*^9}, {
   3.632120593151108*^9, 3.632120593927565*^9}, {3.6321212693672123`*^9, 
   3.632121275148178*^9}, {3.6421385062844954`*^9, 3.6421385106847963`*^9}, 
   3.6421391891234713`*^9},
 CellID->1873815999],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"isPartIrreflOrder", "[", 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox[
        UnderscriptBox["\[Succeeds]", "D"],
        Identity], 
       TagBox[")",
        "AutoParentheses"]}], ",", "D"}], "]"}], "\[And]", 
    RowBox[{"R1", "[", "D", "]"}], "\[And]", 
    RowBox[{"R5", "[", "D", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1547867165}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632121606084746*^9, 3.632121714694981*^9}, {
  3.632200770001727*^9, 3.632200780668154*^9}},
 CellTags->{
  "ID:1547867165", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1547867165],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "c", ",", "m1", ",", "m2", ",", "b1", ",", "b2"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 2009958353}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632120549894956*^9, 3.632120552570692*^9}, {
  3.632120617418931*^9, 3.63212063782255*^9}, {3.6321209756774282`*^9, 
  3.6321209786971893`*^9}, {3.6321210525826683`*^9, 3.6321210574652777`*^9}, {
  3.632121739194319*^9, 3.6321217394189167`*^9}},
 CellTags->{
  "ID:2009958353", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2009958353],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"m1", ",", "c"}]], "D"], "b1"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"m2", ",", "c"}]], "D"], "b2"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 10165835}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6321206426663017`*^9, 3.632120712473136*^9}, {
  3.632120987162332*^9, 3.632120991678944*^9}, {3.632121062177498*^9, 
  3.6321210783915243`*^9}, {3.6321211286103277`*^9, 3.632121146807122*^9}, {
  3.6322074892407*^9, 3.6322074899818153`*^9}},
 CellTags->{
  "ID:10165835", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->10165835],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{"b2", ",", "c", ",", "m1"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     UnderscriptBox["isR6Witness", "D"], "[", 
     RowBox[{
     "T", ",", "a", ",", "c", ",", "m1", ",", "m2", ",", "b1", ",", "b2"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R6Witness sufficient)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1879224373", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.632120727368905*^9, 3.632120738876439*^9}, {
  3.6321208381639214`*^9, 3.632120851412345*^9}, {3.632121024097443*^9, 
  3.632121024392706*^9}, {3.632121155634795*^9, 3.6321211749335947`*^9}, {
  3.632207478576519*^9, 3.632207492313571*^9}, {3.635748998761291*^9, 
  3.6357490052628517`*^9}},
 CellTags->{
  "ID:1879224373", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6Witness sufficient"},
 CellID->1879224373],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     UnderscriptBox["isR6Witness", "D"], "[", 
     RowBox[{
     "T", ",", "a", ",", "c", ",", "m1", ",", "m2", ",", "b1", ",", "b2"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "S", "]"}]}]], 
    RowBox[{
     UnderscriptBox["isR6Witness", "D"], "[", 
     RowBox[{
     "S", ",", "a", ",", "c", ",", "m2", ",", "m1", ",", "b2", ",", "b1"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R6Witness symmetry)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:381733398", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6322053183157663`*^9, 3.6322053733198433`*^9}, {
  3.6322055175053673`*^9, 3.632205519162655*^9}, {3.632207274486352*^9, 
  3.632207275742921*^9}, {3.632207522805674*^9, 3.6322075363138523`*^9}, {
  3.635749011482514*^9, 3.6357490318600893`*^9}},
 CellTags->{
  "ID:381733398", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6Witness symmetry"},
 CellID->381733398],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2045819678]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->596803790],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R6Witness sufficient)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R6Witness sufficient)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"R5", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b2", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["b2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    UnderscriptBox["isR6Witness", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["b2", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"R5", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b2", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["b2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    UnderscriptBox["isR6Witness", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["b2", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1879224373"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1879224373-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1879224373-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1879224373", "Proof|ID:1879224373-1"},
 CellID->453457936],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1879224373-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->63675348]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R6Witness symmetry)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R6Witness symmetry)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"R5", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b2", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    UnderscriptBox["isR6Witness", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["b2", "ExpressionVariable"]}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    UnderscriptBox["isR6Witness", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["S", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["b2", "ExpressionVariable"], ",", 
                    StyleBox["b1", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isPartIrreflOrder", "[", 
               RowBox[{
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], ",", 
                 StyleBox["D", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{"R1", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{"R5", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b2", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    UnderscriptBox["isR6Witness", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["b2", "ExpressionVariable"]}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    UnderscriptBox["isR6Witness", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["S", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["b2", "ExpressionVariable"], ",", 
                    StyleBox["b1", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:381733398"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p381733398-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p381733398-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:381733398", "Proof|ID:381733398-1"},
 CellID->180731690],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p381733398-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->482825486]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (R6Witness symmetry)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632205418823802*^9, 3.6322054257293177`*^9}, 
   3.632212698057674*^9, 3.636356267213367*^9, 3.643533743828369*^9, 
   3.6532196150085573`*^9, 3.666174431963291*^9, 3.666441009587501*^9},
 CellID->592729503],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632212583692572*^9, 3.632212584956429*^9}, 
   3.632212698057805*^9, 3.636356267213497*^9, 3.643533743828496*^9, 
   3.6532196150085573`*^9, 3.666174431963291*^9, 3.666441009587603*^9},
 CellID->665964876],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox[
      SubscriptBox["\[Sum]", "D"], 
      RowBox[{
       RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"-", 
           SubscriptBox["T", 
            RowBox[{
             RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
             "-", 
             RowBox[{"(", 
              RowBox[{"i", "-", "1"}], ")"}]}]]}], 
          UnderscriptBox["|", 
           RowBox[{
            RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
            RowBox[{
            "\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
         "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]], 
     SubscriptBox[
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"-", 
         SubscriptBox["T", 
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
           "-", 
           RowBox[{"(", 
            RowBox[{"i", "-", "1"}], ")"}]}]]}], 
        UnderscriptBox["|", 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}],
        "\[RightAngleBracket]"}], "k"]}], "=", 
    RowBox[{"-", 
     RowBox[{
      UnderscriptBox[
       SubscriptBox["\[Sum]", "D"], 
       RowBox[{
        RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
      SubscriptBox["T", "k"]}]}]}], TraditionalForm]]],
 " (ET Tuples)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6321264601680393`*^9, 3.632126528524651*^9}, {
   3.632127041940702*^9, 3.6321271057373867`*^9}, {3.632127241321944*^9, 
   3.632127241341537*^9}, {3.632127862847693*^9, 3.632127863721365*^9}, {
   3.632133909637314*^9, 3.6321339635885973`*^9}, {3.632134435206379*^9, 
   3.632134435219968*^9}, {3.632135043606678*^9, 3.632135174320592*^9}, 
   3.632136643391201*^9, {3.63213954136454*^9, 3.632139695049699*^9}, 
   3.632204324328061*^9, {3.632204911842093*^9, 3.632204933601486*^9}, 
   3.632212698058277*^9, {3.6331718814202538`*^9, 3.633171884347815*^9}, 
   3.6363562672136097`*^9, 3.6435337438286123`*^9, 3.6532196150085573`*^9, 
   3.666174431963291*^9, 3.666441009587689*^9},
 CellID->236652974],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "c", ",", "m1", ",", "m2", ",", "b1", ",", "b2"}], 
     "]"}]}], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{"m1", ",", "c"}]], "b1"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{"m2", ",", "c"}]], "b2"}], TraditionalForm]]],
 " and we have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[SquareSubsetEqual]", "D"], 
    RowBox[{"[", "T", "]"}]}], TraditionalForm]]],
 " with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isR6Witness", "[", 
    RowBox[{
    "T", ",", "a", ",", "c", ",", "m1", ",", "m2", ",", "b1", ",", "b2"}], 
    "]"}], TraditionalForm]]],
 ", and find ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[SquareSubsetEqual]", "D"], 
    RowBox[{"[", "S", "]"}]}], TraditionalForm]]],
 " with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isR6Witness", "[", 
    RowBox[{
    "S", ",", "a", ",", "c", ",", "m2", ",", "m1", ",", "b2", ",", "b1"}], 
    "]"}], TraditionalForm]]],
 ". We claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", ":=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", 
       SubscriptBox["T", 
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], "-", 
         
         RowBox[{"(", 
          RowBox[{"j", "-", "1"}], ")"}]}]]}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 " is an appropriate instance, so we have to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{"symred", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{"b2", ",", "c", ",", "S", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", 
      RowBox[{"reduceM", "[", 
       RowBox[{"b2", ",", "c", ",", "S", ",", "i"}], "]"}], ",", "c", ",", 
      SubscriptBox["S", "i"]}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"reduceM", "[", 
      RowBox[{"b2", ",", "c", ",", "S", ",", "i"}], "]"}]}]}], 
   TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox[
       SubscriptBox["\[Sum]", "D"], 
       RowBox[{
        RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
      SubscriptBox["S", "k"]}], "+", "m2"}], "=", "m1"}], TraditionalForm]]],
 ".\nOn the other hand, we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
    RowBox[{"symred", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{"b1", ",", "c", ",", "T", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", 
      RowBox[{"reduceM", "[", 
       RowBox[{"b1", ",", "c", ",", "T", ",", "i"}], "]"}], ",", "c", ",", 
      SubscriptBox["T", "i"]}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"reduceM", "[", 
      RowBox[{"b1", ",", "c", ",", "T", ",", "i"}], "]"}]}]}], 
   TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox[
       SubscriptBox["\[Sum]", "D"], 
       RowBox[{
        RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
      SubscriptBox["T", "k"]}], "+", "m1"}], "=", "m2"}], TraditionalForm]]],
 ", and from the definition of the reduction relation we also know ",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m1", ",", "c"}], "]"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{"m2", ",", "c"}], "]"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b1", "=", 
    RowBox[{"a", "-", 
     RowBox[{"m1", "*", "c"}]}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b2", "=", 
    RowBox[{"a", "-", 
     RowBox[{"m2", "*", "c"}]}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", "b1"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", "b2"}], TraditionalForm]]],
 ".\nBefore we proceed we note the following:\n",
 Cell[BoxData[
  FormBox[
   RowBox[{"b2", "=", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{"m2", "*", "c"}]}], "=", 
     RowBox[{
      RowBox[{"a", "-", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           UnderscriptBox[
            SubscriptBox["\[Sum]", "D"], 
            RowBox[{
             RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
           SubscriptBox["T", "k"]}], "+", "m1"}], ")"}], "*", "c"}]}], "=", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"a", "-", 
          RowBox[{"m1", "*", "c"}]}], ")"}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           UnderscriptBox[
            SubscriptBox["\[Sum]", "D"], 
            RowBox[{
             RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
           SubscriptBox["T", "k"]}], ")"}], "*", "c"}]}], "=", 
       RowBox[{"reduceM", "[", 
        RowBox[{"b1", ",", "c", ",", "T", ",", 
         RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
        "]"}]}]}]}]}], TraditionalForm]]],
 ", so we replace all occurrences of ",
 Cell[BoxData[
  FormBox["b2", TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"reduceM", "[", 
    RowBox[{"b1", ",", "c", ",", "T", ",", 
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], "]"}], 
   TraditionalForm]]],
 ".\n\nFirst subgoal: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{"symred", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{"reduceM", "[", 
         RowBox[{"b1", ",", "c", ",", "T", ",", 
          RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
         "]"}], ",", "c", ",", "S", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", 
      RowBox[{"reduceM", "[", 
       RowBox[{
        RowBox[{"reduceM", "[", 
         RowBox[{"b1", ",", "c", ",", "T", ",", 
          RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
         "]"}], ",", "c", ",", "S", ",", "i"}], "]"}], ",", "c", ",", 
      SubscriptBox["S", "i"]}], "]"}]}], TraditionalForm]]],
 ".\nBy (401) we can reduce our current goal to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
    RowBox[{"symred", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{"b1", ",", "c", ",", "T", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", 
      RowBox[{"reduceM", "[", 
       RowBox[{"b1", ",", "c", ",", "T", ",", "i"}], "]"}], ",", "c", ",", 
      SubscriptBox["T", "i"]}], "]"}]}], TraditionalForm]]],
 ", which is exactly what we already know.\n\nSecond subgoal: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"reduceM", "[", 
      RowBox[{
       RowBox[{"reduceM", "[", 
        RowBox[{"b1", ",", "c", ",", "T", ",", 
         RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
        "]"}], ",", "c", ",", "S", ",", "i"}], "]"}]}]}], TraditionalForm]]],
 ".\nBy (402\[CloseCurlyQuote]) we can reduce our goal to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "\[Succeeds]", "b1"}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
     RowBox[{"a", "\[Succeeds]", 
      RowBox[{"reduceM", "[", 
       RowBox[{"b1", ",", "c", ",", "T", ",", "i"}], "]"}]}]}]}], 
   TraditionalForm]]],
 ". Both conjuncts are already among our assumptions.\n\nThird subgoal: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox[
       SubscriptBox["\[Sum]", "D"], 
       RowBox[{
        RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
      SubscriptBox["S", "k"]}], "+", "m2"}], "=", "m1"}], TraditionalForm]]],
 ".\nWe can replace ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     SubscriptBox["\[Sum]", "D"], 
     RowBox[{
      RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
    SubscriptBox["S", "k"]}], TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"-", 
    RowBox[{
     UnderscriptBox[
      SubscriptBox["\[Sum]", "D"], 
      RowBox[{
       RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
     SubscriptBox["T", "k"]}]}], TraditionalForm]]],
 " (by a lemma), and so we have to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{
       UnderscriptBox[
        SubscriptBox["\[Sum]", "D"], 
        RowBox[{
         RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
       SubscriptBox["T", "k"]}]}], "+", "m2"}], "=", "m1"}], 
   TraditionalForm]]],
 ", or, equivalently, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m2", "=", 
    RowBox[{
     RowBox[{
      UnderscriptBox[
       SubscriptBox["\[Sum]", "D"], 
       RowBox[{
        RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
      SubscriptBox["T", "k"]}], "+", "m1"}]}], TraditionalForm]]],
 ". This we already know from the definition of ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 ", hence we are done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632205445084649*^9, 3.6322057398406553`*^9}, {
   3.632205772305455*^9, 3.6322059447058697`*^9}, {3.6322075508847446`*^9, 
   3.632207656437003*^9}, {3.632207810277472*^9, 3.632207983068572*^9}, {
   3.632208042862833*^9, 3.6322081441049757`*^9}, {3.6322081744290257`*^9, 
   3.632208198164152*^9}, {3.63221242288813*^9, 3.632212698058384*^9}, 
   3.6363562672137136`*^9, 3.643533743828714*^9, 3.6532196150085573`*^9, 
   3.666174431963291*^9, 3.666441009587775*^9},
 CellID->1332683746],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6322126854878283`*^9, 3.632212698058522*^9}, 
   3.6363562672138443`*^9, 3.6435337438288383`*^9, 3.6532196150085573`*^9, 
   3.666174431963291*^9, 3.666441009587872*^9},
 TextAlignment->Right,
 CellID->541568807]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " is a Reduction Ring (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.64422401416252*^9, 3.644224018806202*^9}, {
  3.644224104640999*^9, 3.644224110460477*^9}, {3.64516244145702*^9, 
  3.64516244573032*^9}, {3.650260747562703*^9, 3.650260748504734*^9}, {
  3.650792085998814*^9, 3.6507920885527477`*^9}},
 CellID->1827516953],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"isReductionRing", "[", "D", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 404204643}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644224140980637*^9, 3.644224146574729*^9}},
 CellTags->{
  "ID:404204643", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->404204643],

Cell[CellGroupData[{

Cell["Properties of \[Tilde] (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.650260737746427*^9, 3.6502607449291887`*^9}, {
   3.650260779732141*^9, 3.650260785840818*^9}, 3.650260975215868*^9, {
   3.650338416063151*^9, 3.650338420267441*^9}, {3.650792080066141*^9, 
   3.650792082693076*^9}},
 CellID->611594790],

Cell["", "OpenEnvironment",
 CellID->908708940],

Cell[CellGroupData[{

Cell["\<\
Lemma: Properties of \[Tilde] with Multiplier\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.650337647682324*^9, 3.650337651532922*^9}, {
  3.650337929096088*^9, 3.6503379323658257`*^9}},
 CellID->1831554730],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b", ",", "m"}], "]"}]}]], 
  RowBox[{"a", 
   UnderscriptBox[
    SubscriptBox["\[Tilde]", "m"], "D"], "b"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1399661856}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650337661656425*^9, 3.650337681319551*^9}, {
  3.6503377163727837`*^9, 3.6503377175307417`*^9}, {3.650337905947974*^9, 
  3.650337909642503*^9}, {3.650346926952352*^9, 3.650346942645275*^9}},
 CellTags->{
  "ID:1399661856", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1399661856],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isUnit", "D"], "[", "m", "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", "n", "]"}]}]], 
     RowBox[{
      UnderscriptBox["isUnit", "D"], "[", "n", "]"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"b", 
        UnderscriptBox[
         SubscriptBox["\[Tilde]", "n"], "D"], "a"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"m", 
         UnderscriptBox["*", "D"], "n"}], "\[Equal]", 
        UnderscriptBox["1", "D"]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Tilde] multiplier symmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1116556350", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6506976499761477`*^9, 3.650697709315043*^9}, {
  3.650699537100316*^9, 3.6506995422028103`*^9}},
 CellTags->{
  "ID:1116556350", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde] multiplier symmetric"},
 CellID->1116556350],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", 
      RowBox[{"c", ",", "n"}], "]"}]}]], 
   RowBox[{"c", "\[NotEqual]", 
    UnderscriptBox["0", "D"]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "D"], "[", 
     RowBox[{"n", ",", "c"}], "]"}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["mult", "D"], "[", 
     RowBox[{
      RowBox[{"m", 
       UnderscriptBox["*", "D"], "n"}], ",", "c"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1121162524", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650786057494199*^9, 3.650786094065137*^9}},
 CellTags->{
  "ID:1121162524", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "mult \[Tilde]"},
 CellID->1121162524],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"c", ",", "n", ",", "x"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "D"], "[", 
     RowBox[{"x", ",", "c", ",", "n"}], "]"}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["isReducible", "D"], "[", 
     RowBox[{
      RowBox[{"m", 
       UnderscriptBox["*", "D"], "x"}], ",", "c", ",", 
      RowBox[{"m", 
       UnderscriptBox["*", "D"], "n"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1098905803", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650347446661274*^9, 3.650347479841922*^9}, {
  3.650786022522941*^9, 3.6507860305850782`*^9}},
 CellTags->{
  "ID:1098905803", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible \[Tilde]"},
 CellID->1098905803],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2", ",", "m1", ",", "m2"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "D"], "[", "c1", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      RowBox[{"c2", "\[NotEqual]", 
       UnderscriptBox["0", "D"]}], "\[And]", 
      RowBox[{
       UnderscriptBox["areIrrelative", "D"], "[", 
       RowBox[{"c1", ",", "c2", ",", "m1", ",", "m2", ",", "i", ",", "j"}], 
       "]"}]}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["areIrrelative", "D"], "[", 
      RowBox[{"c1", ",", "c2", ",", 
       RowBox[{"m", 
        UnderscriptBox["*", "D"], "m1"}], ",", 
       RowBox[{"m", 
        UnderscriptBox["*", "D"], "m2"}], ",", "i", ",", "j"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areIrrelative \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:88395859", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650346069743704*^9, 3.650346105400734*^9}, {
  3.650346946294479*^9, 3.65034696734158*^9}, {3.650347200922556*^9, 
  3.650347242715839*^9}},
 CellTags->{
  "ID:88395859", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areIrrelative \[Tilde]"},
 CellID->88395859],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2", ",", "m1", ",", "m2"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "D"], "[", "c1", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["areCPMultipliers", "D"], "[", 
      RowBox[{
      "m1", ",", "m2", ",", "c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], 
      "]"}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["areCPMultipliers", "D"], "[", 
      RowBox[{
       RowBox[{"m", 
        UnderscriptBox["*", "D"], "m1"}], ",", 
       RowBox[{"m", 
        UnderscriptBox["*", "D"], "m2"}], ",", "c1", ",", "c2", ",", "b", ",",
        "i", ",", "j"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areCPMultipliers \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1322820809", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650337946434505*^9, 3.6503379894976377`*^9}, {
  3.650338303510755*^9, 3.65033830765508*^9}, {3.650345921612885*^9, 
  3.6503459232805653`*^9}, {3.650346042819497*^9, 3.650346047450844*^9}, {
  3.650346121268113*^9, 3.6503461419972878`*^9}, {3.650346969739182*^9, 
  3.650346987099105*^9}, {3.65034721874786*^9, 3.650347229443592*^9}, {
  3.650348133585774*^9, 3.65034815548376*^9}},
 CellTags->{
  "ID:1322820809", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers \[Tilde]"},
 CellID->1322820809],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 931721333}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6507889330156517`*^9, 3.650788940513317*^9}},
 CellTags->{
  "ID:931721333", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->931721333],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"c", ",", "n"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"n", ",", "c"}]], "D"], "y"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"m", 
       UnderscriptBox["*", "D"], "x"}], 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{
         RowBox[{"m", 
          UnderscriptBox["*", "D"], "n"}], ",", "c"}]], "D"], 
      RowBox[{"m", 
       UnderscriptBox["*", "D"], "y"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow] multiplier \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2090837112", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650358511253257*^9, 3.65035852399358*^9}, {
   3.650786385032669*^9, 3.650786396619671*^9}, 3.650788943017057*^9},
 CellTags->{
  "ID:2090837112", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] multiplier \[Tilde]"},
 CellID->2090837112],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "c", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"{", "c", "}"}]], "D"], "y"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"m", 
       UnderscriptBox["*", "D"], "x"}], 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"{", "c", "}"}]], "D"], 
      RowBox[{"m", 
       UnderscriptBox["*", "D"], "y"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow] \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:780050836", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650348813534685*^9, 3.6503488192124043`*^9}, {
   3.650786400068344*^9, 3.650786410844336*^9}, 3.650788945292274*^9, {
   3.661850368055947*^9, 3.661850376308361*^9}},
 CellTags->{
  "ID:780050836", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[Tilde]"},
 CellID->780050836],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["isReductor", "D"], "[", "r", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 128058836}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650349471458712*^9, 3.650349476802122*^9}, {
  3.650354785163472*^9, 3.650354785460855*^9}},
 CellTags->{
  "ID:128058836", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->128058836],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", "r"], "D"], "y"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["*", "D"], "x"}], 
    UnderscriptBox[
     SubscriptBox["\[Rule]", "r"], "D"], 
    RowBox[{"m", 
     UnderscriptBox["*", "D"], "y"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow] reductor \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1516437288", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650346382614091*^9, 3.650346439986331*^9}, 
   3.650349480445335*^9, {3.650786431197523*^9, 3.650786441159967*^9}, {
   3.650788947940448*^9, 3.650788948900378*^9}},
 CellTags->{
  "ID:1516437288", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] reductor \[Tilde]"},
 CellID->1516437288],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", "r"], "D"], "y"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["*", "D"], "x"}], 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", "r"], "D"], 
    RowBox[{"m", 
     UnderscriptBox["*", "D"], "y"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LeftRightArrow] reductor \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1864007722", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650349495893423*^9, 3.650349510794536*^9}, {
  3.650788950450624*^9, 3.650788951483876*^9}},
 CellTags->{
  "ID:1864007722", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] reductor \[Tilde]"},
 CellID->1864007722],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
      RowBox[{"a", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], "y"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["*", "D"], "x"}], 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
      RowBox[{"b", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], 
    RowBox[{"m", 
     UnderscriptBox["*", "D"], "y"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(connectibility below \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:642398934", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650346274075906*^9, 3.65034634657378*^9}, {
   3.650346415267167*^9, 3.6503464166331673`*^9}, 3.650349481732217*^9, {
   3.6507889528515253`*^9, 3.65078895410478*^9}},
 CellTags->{
  "ID:642398934", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below \[Tilde]"},
 CellID->642398934],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->217367629]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->72738875],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Tilde] multiplier symmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Tilde] multiplier symmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isUnit", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                UnderscriptBox["isUnit", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Tilde]", 
                   StyleBox["n", "ExpressionVariable"]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["a", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
                 UnderscriptBox["1", 
                  StyleBox[
                  "D", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isUnit", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                UnderscriptBox["isUnit", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Tilde]", 
                   StyleBox["n", "ExpressionVariable"]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["a", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
                 UnderscriptBox["1", 
                  StyleBox[
                  "D", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1116556350"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1116556350-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1116556350-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1116556350", "Proof|ID:1116556350-1"},
 CellID->1446312685],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1116556350-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->239017603]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult \[Tilde])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult \[Tilde])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["n", "ExpressionVariable"]}], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["n", "ExpressionVariable"]}], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1121162524"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1121162524-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1121162524-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1121162524", "Proof|ID:1121162524-1"},
 CellID->1284068679],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1121162524-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2106471830]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible \[Tilde])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible \[Tilde])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["n", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["x", "ExpressionVariable"]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["n", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["n", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["x", "ExpressionVariable"]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["n", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1098905803"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1098905803-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1098905803-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1098905803", "Proof|ID:1098905803-1"},
 CellID->1798497905],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1098905803-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1893221538]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(areIrrelative \[Tilde])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(areIrrelative \[Tilde])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areIrrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["areIrrelative", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["m1", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["m2", "ExpressionVariable"]}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areIrrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["areIrrelative", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["m1", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["m2", "ExpressionVariable"]}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:88395859"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p88395859-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p88395859-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:88395859", "Proof|ID:88395859-1"},
 CellID->514698520],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p88395859-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->697959023]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(areCPMultipliers \[Tilde])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(areCPMultipliers \[Tilde])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCPMultipliers", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["m1", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["m2", "ExpressionVariable"]}], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCPMultipliers", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["m1", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["m2", "ExpressionVariable"]}], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1322820809"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1322820809-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1322820809-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1322820809", "Proof|ID:1322820809-1"},
 CellID->330288797],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1322820809-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6582235035976553`*^9},
 CellID->1636900124]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] multiplier \[Tilde])\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] multiplier \[Tilde])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["n", "ExpressionVariable"]}], ",", 
                 StyleBox["c", "ExpressionVariable"]}]], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["n", "ExpressionVariable"]}], ",", 
                 StyleBox["c", "ExpressionVariable"]}]], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2090837112"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2090837112-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2090837112-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2090837112", "Proof|ID:2090837112-1"},
 CellID->1967137926],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2090837112-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1474790292]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] \[Tilde])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] \[Tilde])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{"{", 
                 StyleBox["c", "ExpressionVariable"], "}"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{"{", 
                 StyleBox["c", "ExpressionVariable"], "}"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:780050836"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p780050836-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p780050836-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:780050836", "Proof|ID:780050836-1"},
 CellID->122789986],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p780050836-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->155714628]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] reductor \[Tilde])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] reductor \[Tilde])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["r", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["r", "ExpressionVariable"]], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["r", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["r", "ExpressionVariable"]], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1516437288"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1516437288-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1516437288-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1516437288", "Proof|ID:1516437288-1"},
 CellID->23267736],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1516437288-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->241728310]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LeftRightArrow] reductor \[Tilde])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LeftRightArrow] reductor \[Tilde])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               StyleBox["r", "ExpressionVariable"]], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               StyleBox["r", "ExpressionVariable"]], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1864007722"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1864007722-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1864007722-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1864007722", "Proof|ID:1864007722-1"},
 CellID->1742409570],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1864007722-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1961249060]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(connectibility below \[Tilde])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(connectibility below \[Tilde])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Tilde]", 
                 StyleBox["m", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:642398934"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p642398934-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p642398934-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:642398934", "Proof|ID:642398934-1"},
 CellID->1849048841],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p642398934-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->503024193]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->242245566],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[Tilde] is Equivalence Relation\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.650260766591758*^9, 3.650260773661271*^9}},
 CellID->1737749019],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1200903488}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6502607946549187`*^9, 3.65026080293534*^9}},
 CellTags->{
  "ID:1200903488", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1200903488],

Cell[BoxData[
 RowBox[{"a", 
  UnderscriptBox["\[Tilde]", "D"], "a"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reflexivity of \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:353808388", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650260806544589*^9, 3.6502608114666367`*^9}},
 CellTags->{
  "ID:353808388", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[Tilde]"},
 CellID->353808388],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Tilde]", "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"b", 
    UnderscriptBox["\[Tilde]", "D"], "a"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symmetry of \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:274091018", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65026082158829*^9, 3.650260838576138*^9}},
 CellTags->{
  "ID:274091018", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of \[Tilde]"},
 CellID->274091018],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox["\[Tilde]", "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", 
      UnderscriptBox["\[Tilde]", "D"], "c"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Tilde]", "D"], "c"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(transitivity of \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1760571189", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650260842323799*^9, 3.650260879159068*^9}},
 CellTags->{
  "ID:1760571189", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[Tilde]"},
 CellID->1760571189],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1666723148]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1069542713],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reflexivity of \[Tilde])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reflexivity of \[Tilde])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox["\[Tilde]", 
            StyleBox["D", "ExpressionVariable"]], 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox["\[Tilde]", 
            StyleBox["D", "ExpressionVariable"]], 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:353808388"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p353808388-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p353808388-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:353808388", "Proof|ID:353808388-1"},
 CellID->844255967],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p353808388-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1096022174]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symmetry of \[Tilde])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symmetry of \[Tilde])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:274091018"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p274091018-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p274091018-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:274091018", "Proof|ID:274091018-1"},
 CellID->772612304],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p274091018-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1935039816]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(transitivity of \[Tilde])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(transitivity of \[Tilde])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[Tilde]", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[Tilde]", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[Tilde]", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[Tilde]", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1760571189"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1760571189-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1760571189-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1760571189", "Proof|ID:1760571189-1"},
 CellID->1528186019],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1760571189-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->690157133]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Correlativity and Irrelativity (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.645246966993937*^9, 3.64524697552709*^9}, {
  3.645247036028179*^9, 3.645247039025133*^9}, {3.645247451951931*^9, 
  3.645247455656205*^9}},
 CellID->1144726974],

Cell["", "OpenEnvironment",
 CellID->1390555827],

Cell[CellGroupData[{

Cell["Lemma: Correlativity and Irrelativity", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6452470223392982`*^9, 3.645247032897403*^9}},
 CellID->934535187],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"c", ",", "m1", ",", "m2"}], "]"}]}]], 
  RowBox[{"c", "\[NotEqual]", 
   UnderscriptBox["0", "D"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 871669950}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645246478154479*^9, 3.645246490105391*^9}, 
   3.645246702761998*^9, {3.645247005478448*^9, 3.645247005626741*^9}, {
   3.6452470698179293`*^9, 3.645247072794009*^9}},
 CellTags->{
  "ID:871669950", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->871669950],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[Not]", 
     RowBox[{
      UnderscriptBox["areCorrelative", "D"], "[", 
      RowBox[{"c", ",", "m1", ",", "m2"}], "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["mult", "D"], "[", 
     RowBox[{"m1", ",", "c"}], "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["mult", "D"], "[", 
     RowBox[{"m2", ",", "c"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{"k", ",", 
      RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       UnderscriptBox["multN", "D"], "[", "c", "]"}]}]], 
    RowBox[{
     UnderscriptBox["areIrrelative", "D"], "[", 
     RowBox[{"c", ",", "c", ",", "m1", ",", "m2", ",", "k", ",", "l"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(correlative-irrelative 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:163646289", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645246466794999*^9, 3.64524657497379*^9}, {
  3.645246704499034*^9, 3.645246712021393*^9}},
 CellTags->{
  "ID:163646289", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative-irrelative 1"},
 CellID->163646289],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->683761533]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1397889212],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(correlative-irrelative 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(correlative-irrelative 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox["areCorrelative", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], "]"}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
                 RowBox[{
                   RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["areIrrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"], ",", 
                 StyleBox["l", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox["areCorrelative", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], "]"}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
                 RowBox[{
                   RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["areIrrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"], ",", 
                 StyleBox["l", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:163646289"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p163646289-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p163646289-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:163646289", "Proof|ID:163646289-1"},
 CellID->1996269430],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p163646289-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->960850119]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Reductions Modulo Sets and Individuals (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.64516119156831*^9, 3.645161203262451*^9}, {
  3.645162010388212*^9, 3.645162014336967*^9}},
 CellID->679333461],

Cell["", "OpenEnvironment",
 CellID->1992553206],

Cell[CellGroupData[{

Cell["Lemma: Reductions Modulo Sets and Individuals", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.6451612172899313`*^9, {3.645438059316889*^9, 3.645438066983794*^9}},
 CellID->280665347],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "z"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1990325407}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645161239247696*^9, 3.645161248598199*^9}, {
  3.645437163819902*^9, 3.645437166211183*^9}},
 CellTags->{
  "ID:1990325407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1990325407],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "D", "]"}]], 
   RowBox[{"[", "C", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 615171230}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645161252262062*^9, 3.645161264100403*^9}},
 CellTags->{
  "ID:615171230", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->615171230],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", "C"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{"c", "\[Element]", "C"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[LeftRightArrow]", 
        RowBox[{"{", "c", "}"}]], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(47)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1548501955", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6451612813463697`*^9, 3.645161308177658*^9}, {
  3.645170185033164*^9, 3.645170194834714*^9}, {3.645170372859288*^9, 
  3.645170377668076*^9}, {3.6618505362554426`*^9, 3.661850537737445*^9}},
 CellTags->{
  "ID:1548501955", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "47"},
 CellID->1548501955],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"c", "\[Element]", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubsuperscriptBox["\[LeftRightArrow]", 
        RowBox[{"{", "c", "}"}], "*"], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubsuperscriptBox["\[LeftRightArrow]", "C", "*"], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(48)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:170197122", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6454379182390957`*^9, 3.645437924037491*^9}, {
  3.661850541325451*^9, 3.6618505448198576`*^9}},
 CellTags->{
  "ID:170197122", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "48"},
 CellID->170197122],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"c", "\[Element]", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       OverscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", 
         RowBox[{"{", "c", "}"}], "*"], 
        RowBox[{"z", ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox["\[Succeeds]",
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}]], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       OverscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", "C", "*"], 
        RowBox[{"z", ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox["\[Succeeds]",
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}]], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(49)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:772910972", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6451613161323433`*^9, 3.645161333946302*^9}, {
  3.645170350658184*^9, 3.6451703637325363`*^9}, {3.6618505476746626`*^9, 
  3.661850549094265*^9}},
 CellTags->{
  "ID:772910972", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "49"},
 CellID->772910972],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1385006624]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1123829565],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(47)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(47)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 StyleBox["C", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[Element]", 
                   StyleBox["C", "ExpressionVariable"]}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 StyleBox["C", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[Element]", 
                   StyleBox["C", "ExpressionVariable"]}]}}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1548501955"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1548501955-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1548501955-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1548501955", "Proof|ID:1548501955-1"},
 CellID->657095506],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1548501955-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->40777374]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(48)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(48)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["C", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                RowBox[{"{", 
                  StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["C", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["C", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                RowBox[{"{", 
                  StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["C", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:170197122"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p170197122-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p170197122-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:170197122", "Proof|ID:170197122-1"},
 CellID->175765073],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p170197122-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->48619257]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(49)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(49)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["C", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["C", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["C", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["C", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:772910972"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p772910972-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p772910972-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:772910972", "Proof|ID:772910972-1"},
 CellID->25309177],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p772910972-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->541611342]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Alternative Descriptions of ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[Rule]", "*"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[LeftRightArrow]", "*"], TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.644296204349338*^9, 3.644296249384818*^9}, {
  3.6448997634791927`*^9, 3.6448997674389877`*^9}, {3.6452601103847933`*^9, 
  3.645260120909582*^9}, {3.64543157014884*^9, 3.645431576473564*^9}},
 CellID->1798149129],

Cell["", "OpenEnvironment",
 CellID->922102237],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: Alternative Descriptions of ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[Rule]", "*"], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.644224132389635*^9, {3.644296257311368*^9, 3.644296282832239*^9}, {
   3.644299296945549*^9, 3.64429930239279*^9}},
 CellID->545035824],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b"}], "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["isReductor", "D"], "[", "r", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1034579573}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6442310374288883`*^9, 3.644231048533676*^9}, {
   3.644291001669936*^9, 3.644291004887824*^9}, 3.644296464007348*^9, {
   3.6447507882619753`*^9, 3.644750804200212*^9}, {3.644751001484816*^9, 
   3.644751008802074*^9}},
 CellTags->{
  "ID:1034579573", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1034579573],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           SubscriptBox["T", 
            RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
           "\[Equal]", "b"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
              "-", "1"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["T", "i"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", "r"], "D"], 
             SubscriptBox["T", 
              RowBox[{"i", "+", "1"}]]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           RowBox[{"a", "\[NotEqual]", "b"}], "\[Implies]", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{"a", 
               UnderscriptBox[
                SubscriptBox["\[Rule]", "r"], "D"], 
               SubscriptBox["T", "1"]}], 
              TagBox[")",
               "AutoParentheses"]}], "\[And]", 
             RowBox[{
              UnderscriptBox["\[ForAll]", 
               RowBox[{
                RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                RowBox[{
                "\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{"a", 
                UnderscriptBox["\[Succeeds]", "D"], 
                SubscriptBox["T", "i"]}], 
               TagBox[")",
                "AutoParentheses"]}]}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(30.\[DoubleRightArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:226093298", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644231119955166*^9, 3.64423124419345*^9}, {
   3.644232038553321*^9, 3.644232041007128*^9}, {3.644291008410924*^9, 
   3.64429101152034*^9}, {3.644291514639162*^9, 3.64429155558396*^9}, {
   3.644291651712078*^9, 3.644291655124857*^9}, {3.644297114403892*^9, 
   3.644297114661314*^9}, 3.644299316391162*^9, {3.644300893207879*^9, 
   3.644300902480394*^9}, {3.644749391338127*^9, 3.6447493954823017`*^9}},
 CellTags->{
  "ID:226093298", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "30.\[DoubleRightArrow]"},
 CellID->226093298],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           SubscriptBox["T", "1"], "\[Equal]", "a"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
              "-", "1"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["T", "i"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", "r"], "D"], 
             SubscriptBox["T", 
              RowBox[{"i", "+", "1"}]]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{"i", "\[NotEqual]", "1"}], "\[Implies]", 
             RowBox[{"a", 
              UnderscriptBox["\[Succeeds]", "D"], 
              SubscriptBox["T", "i"]}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           RowBox[{"a", "\[NotEqual]", "b"}], "\[Implies]", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               SubscriptBox["T", 
                RowBox[{
                "\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", "r"], "D"], "b"}], 
              TagBox[")",
               "AutoParentheses"]}], "\[And]", 
             RowBox[{"a", 
              UnderscriptBox["\[Succeeds]", "D"], "b"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(31.\[DoubleRightArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1773205884", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6442915686766577`*^9, 3.6442916176665277`*^9}, {
   3.644291658753676*^9, 3.64429167600001*^9}, {3.644297112511104*^9, 
   3.644297112881405*^9}, 3.644299319336875*^9, {3.6447494976406813`*^9, 
   3.644749505689324*^9}},
 CellTags->{
  "ID:1773205884", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "31.\[DoubleRightArrow]"},
 CellID->1773205884],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           SubscriptBox["T", 
            RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
           "\[Equal]", "b"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
              "-", "1"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["T", "i"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", "r"], "D"], 
             SubscriptBox["T", 
              RowBox[{"i", "+", "1"}]]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           RowBox[{"a", "\[NotEqual]", "b"}], "\[Implies]", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{"a", 
             UnderscriptBox[
              SubscriptBox["\[Rule]", "r"], "D"], 
             SubscriptBox["T", "1"]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(30.\[DoubleLeftArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1783130815", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644299311418335*^9, 3.6442993275487347`*^9}, {
  3.6442994300409946`*^9, 3.6442994303740463`*^9}, {3.644749514179418*^9, 
  3.644749533655306*^9}, {3.6448367561072407`*^9, 3.644836759848946*^9}},
 CellTags->{
  "ID:1783130815", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "30.\[DoubleLeftArrow]"},
 CellID->1783130815],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           SubscriptBox["T", "1"], "\[Equal]", "a"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
              "-", "1"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["T", "i"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", "r"], "D"], 
             SubscriptBox["T", 
              RowBox[{"i", "+", "1"}]]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           RowBox[{"a", "\[NotEqual]", "b"}], "\[Implies]", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["T", 
              RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", "r"], "D"], "b"}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(31.\[DoubleLeftArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1493153487", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644838739542509*^9, 3.644838769142108*^9}},
 CellTags->{
  "ID:1493153487", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "31.\[DoubleLeftArrow]"},
 CellID->1493153487],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "\[Equal]", "b"}], "\[Or]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", "c", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"a", 
          UnderscriptBox[
           SubscriptBox["\[Rule]", "r"], "D"], "c"}], 
         TagBox[")",
          "AutoParentheses"]}], "\[And]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"c", 
          UnderscriptBox[
           SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(39.\[DoubleRightArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1798779961", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6455143447687683`*^9, 3.6455144016146517`*^9}},
 CellTags->{
  "ID:1798779961", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "39.\[DoubleRightArrow]"},
 CellID->1798779961],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->481199956]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->336597908],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(30.\[DoubleRightArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(30.\[DoubleRightArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], "1"]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], "1"]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:226093298"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p226093298-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p226093298-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:226093298", "Proof|ID:226093298-1"},
 CellID->2111336883],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p226093298-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellChangeTimes->{3.644896090907887*^9},
 CellID->1011674363]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(31.\[DoubleRightArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(31.\[DoubleRightArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], "1"], "\[Equal]", 
                 StyleBox["a", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[NotEqual]", "1"}],
                    "\[Implies]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox[
                    "b", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], "1"], "\[Equal]", 
                 StyleBox["a", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[NotEqual]", "1"}],
                    "\[Implies]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox[
                    "b", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1773205884"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1773205884-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1773205884-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1773205884", "Proof|ID:1773205884-1"},
 CellID->468420595],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1773205884-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->303982263]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(30.\[DoubleLeftArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(30.\[DoubleLeftArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], "1"]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], "1"]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1783130815"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1783130815-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1783130815-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1783130815", "Proof|ID:1783130815-1"},
 CellID->1694100803],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1783130815-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.644838846796021*^9, 3.644839052098022*^9, 
  3.644896881060616*^9},
 CellID->1573422288]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(31.\[DoubleLeftArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(31.\[DoubleLeftArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], "1"], "\[Equal]", 
                 StyleBox["a", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b", "ExpressionVariable"]}]}]}]}], "\[Implies]", 
           
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], "1"], "\[Equal]", 
                 StyleBox["a", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b", "ExpressionVariable"]}]}]}]}], "\[Implies]", 
           
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1493153487"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1493153487-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1493153487-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1493153487", "Proof|ID:1493153487-1"},
 CellID->1970791442],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1493153487-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->22791630]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(39.\[DoubleRightArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(39.\[DoubleRightArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Equal]", 
               StyleBox["b", "ExpressionVariable"]}], "\[Or]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox[
                    TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox[
                   "b", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Equal]", 
               StyleBox["b", "ExpressionVariable"]}], "\[Or]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], 
                   UnderscriptBox[
                    TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox[
                   "b", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1798779961"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1798779961-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1798779961-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1798779961", "Proof|ID:1798779961-1"},
 CellID->1268768510],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1798779961-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->699204208]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (30.\[Implies]",
  FontSlant->"Italic"],
 ")."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.644748348814376*^9, 3.644748362448565*^9}, 
   3.644751538562808*^9, 3.6532196002343316`*^9, 3.666174421043272*^9, 
   3.666440996276636*^9},
 CellID->279050304],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReductor", "[", "r", "]"}], TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubsuperscriptBox["\[Rule]", "r", "*"], "b"}], TraditionalForm]]],
 "\nand find some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "T", "]"}]}], TraditionalForm]]],
 " with\n(G.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[NotEqual]", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 "\n(G.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", 
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], "=", 
    "b"}], TraditionalForm]]],
 "\n(G.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], "-", 
        "1"}]}]], 
     SubscriptBox["T", "i"]}], 
    SubscriptBox["\[Rule]", "r"], 
    SubscriptBox["T", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 "\n(G.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "\[NotEqual]", "b"}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{"a", 
      SubscriptBox["\[Rule]", "r"], 
      RowBox[{
       SubscriptBox["T", "1"], "\[And]", 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
        RowBox[{"a", "\[Succeeds]", 
         SubscriptBox["T", "i"]}]}]}]}], ")"}]}], TraditionalForm]]],
 "\n\nWe distinguish two cases, based upon whether ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", "b"}], TraditionalForm]]],
 " or not.\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "b"}], TraditionalForm]]],
 ".\nFrom (1), by the definition of ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[Rule]", "*"], TraditionalForm]]],
 ", we can infer the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "S", "]"}]}], TraditionalForm]]],
 " with\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", "\[NotEqual]", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", "1"], "=", "a"}], TraditionalForm]]],
 "\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", 
     RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]], "=", 
    "b"}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}], "-", 
        "1"}]}]], 
     SubscriptBox["S", "i"]}], 
    SubscriptBox["\[Rule]", "r"], 
    SubscriptBox["S", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 "\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", ":=", 
    RowBox[{"Rest", "[", "S", "]"}]}], TraditionalForm]]],
 " is a suitable witness for (G.1)-(G.4), hence we prove all four subgoals.\n\
\n(G.1):\nWe have to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"Rest", "[", "S", "]"}], "\[NotEqual]", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ", which, by (2) and a lemma, amounts to showing\n(G.5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}], 
    "\[NotEqual]", "1"}], TraditionalForm]]],
 "\nWe prove (G.5) by contradiction, i.e. we assume\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}], "=", 
    "1"}], TraditionalForm]]],
 "\nThis means that we have ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", 
    RowBox[{
     SubscriptBox["S", "1"], "=", 
     RowBox[{
      SubscriptBox["S", 
       RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]], "=", 
      "b"}]}]}], TraditionalForm]]],
 ", contradicting our case assumption.\n\n(G.2):\nWe have to show that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox[
     RowBox[{"Rest", "[", "S", "]"}], 
     RowBox[{"\[LeftBracketingBar]", 
      RowBox[{"Rest", "[", "S", "]"}], "\[RightBracketingBar]"}]], "=", "b"}],
    TraditionalForm]]],
 ". The LHS of this equality, by a lemma, is equal to ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", 
    RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]], 
   TraditionalForm]]],
 ", which further equals the RHS due to (4).\n\n(G.3):\nBecause of the shape \
of ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " our goal is equivalent to\n(G.6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}], "-", 
         "1"}]}]], 
      RowBox[{"i", "\[NotEqual]", "1"}]}], "\[Implies]", 
     SubscriptBox["S", "i"]}], 
    SubscriptBox["\[Rule]", "r"], 
    SubscriptBox["S", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 "\n(G.6) obviously follows from (5).\n\n(G.4):\nWe need to show both\n(G.7)\t\
",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", "r"], 
    SubscriptBox[
     RowBox[{"Rest", "[", "S", "]"}], "1"]}], TraditionalForm]]],
 "\n(G.8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", 
       RowBox[{"Rest", "[", "S", "]"}], "\[RightBracketingBar]"}]}]], 
    RowBox[{"a", "\[Succeeds]", 
     SubscriptBox[
      RowBox[{"Rest", "[", "S", "]"}], "i"]}]}], TraditionalForm]]],
 "\n\n(G.7):\nWe can use (5), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ":=", "1"}], TraditionalForm]]],
 ", and infer\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", "1"], 
    SubscriptBox["\[Rule]", "r"], 
    SubscriptBox["S", 
     RowBox[{"1", "+", "1"}]]}], TraditionalForm]]],
 "\nUsing (3) and the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox[
     RowBox[{"Rest", "[", "S", "]"}], "1"], "=", 
    SubscriptBox["S", 
     RowBox[{"1", "+", "1"}]]}], TraditionalForm]]],
 " finishes this subgoal.\n\n(G.8):\nBy a lemma, our universally quantified \
goal is equivalent to\n(G.9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
     RowBox[{"i", "\[NotEqual]", "1"}]}], "\[Implies]", 
    RowBox[{"a", "\[Succeeds]", 
     SubscriptBox["S", "i"]}]}], TraditionalForm]]],
 "\nwhich we show by induction on ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 ".\nThe base case, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "=", "1"}], TraditionalForm]]],
 ", is trivial, hence we continue with the induction step. Therefore, we \
choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalN]", 
     RowBox[{"1", ",", "\[Ellipsis]", ",", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}], "-", 
       "1"}]}]]}], TraditionalForm]]],
 " a.b.f., assume\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "\[NotEqual]", "1"}], "\[Implies]", 
    RowBox[{"a", "\[Succeeds]", 
     SubscriptBox["S", "i"]}]}], TraditionalForm]]],
 "\nand show\n(G.10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    SubscriptBox["S", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 "\nOnce again we can use (5) to infer\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", "i"], 
    SubscriptBox["\[Rule]", "r"], 
    SubscriptBox["S", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 "\nand further, from a lemma,\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", "i"], "\[Succeeds]", 
    SubscriptBox["S", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 "\n\nWe now distinguish two cases, depending on whether ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "=", "1"}], TraditionalForm]]],
 " or not.\nCase I.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[NotEqual]", "1"}], TraditionalForm]]],
 ".\nFrom (8), by modus ponens, we can infer\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    SubscriptBox["S", "i"]}], TraditionalForm]]],
 "\nNow, our goal follows from (11) and (10) using transitivity of \
\[Succeeds].\n\nCase I.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "=", "1"}], TraditionalForm]]],
 ".\nOur goal follows immediately from (10) using the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", "1"], "=", "a"}], TraditionalForm]]],
 ".\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", "b"}], TraditionalForm]]],
 ".\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", ":=", 
    RowBox[{"\[LeftAngleBracket]", "b", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 " is a suitable witness for (G.1)-(G.4), meaning we have to prove all four \
subgoals.\n\n(G.1):\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", "b", "\[RightAngleBracket]"}], 
    "\[NotEqual]", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 " holds apparently.\n\n(G.2):\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox[
     RowBox[{"\[LeftAngleBracket]", "b", "\[RightAngleBracket]"}], "1"], "=", 
    "b"}], TraditionalForm]]],
 " holds apparently.\n\n(G.3):\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "0"}]], 
    RowBox[{"P", "[", "i", "]"}]}], TraditionalForm]]],
 " holds vacuously.\n\n(G.4):\nThe premise of the implication contradicts our \
case assumption."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.644748365435741*^9, 3.644748642564788*^9}, {
   3.644749339672086*^9, 3.6447493737521152`*^9}, {3.644749550963642*^9, 
   3.6447499001653023`*^9}, {3.644750156969122*^9, 3.644750404422041*^9}, {
   3.6447506276297283`*^9, 3.64475076354609*^9}, {3.6447510868208323`*^9, 
   3.6447511376902246`*^9}, {3.644751181556593*^9, 3.644751449418778*^9}, {
   3.644751488119371*^9, 3.644751538564777*^9}, {3.6448290079700937`*^9, 
   3.644829030381987*^9}, 3.6532196002343316`*^9, 3.666174421043272*^9, 
   3.6664409962768087`*^9},
 CellID->1466920198],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.644751522491851*^9, 3.6447515385649567`*^9}, 
   3.6532196002343316`*^9, 3.666174421043272*^9, 3.666440996276943*^9},
 TextAlignment->Right,
 CellID->539158522]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->661399684],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: Alternative Description of ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[LeftRightArrow]", "*"], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645431360748679*^9, 3.645431370997826*^9}, {
  3.6454315098651543`*^9, 3.6454315108036327`*^9}},
 CellID->776782599],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b"}], "]"}]}], ",", "r"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 712613448}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6442310374288883`*^9, 3.644231048533676*^9}, {
   3.644291001669936*^9, 3.644291004887824*^9}, {3.644751890147048*^9, 
   3.644751896400757*^9}, {3.645431390945606*^9, 3.645431393502087*^9}, {
   3.645431517751485*^9, 3.645431524633872*^9}, 3.645431647473013*^9, 
   3.645433360324438*^9},
 CellTags->{
  "ID:712613448", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->712613448],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           SubscriptBox["T", 
            RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
           "\[Equal]", "b"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
              "-", "1"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["T", "i"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", "r"], "D"], 
             SubscriptBox["T", 
              RowBox[{"i", "+", "1"}]]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           RowBox[{"a", "\[NotEqual]", "b"}], "\[Implies]", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{"a", 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", "r"], "D"], 
             SubscriptBox["T", "1"]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(37.\[DoubleRightArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1914773542", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644297312525915*^9, 3.644297321539915*^9}, {
  3.644895879711549*^9, 3.644895884722096*^9}, {3.645431539990343*^9, 
  3.6454315510401373`*^9}},
 CellTags->{
  "ID:1914773542", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "37.\[DoubleRightArrow]"},
 CellID->1914773542],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           SubscriptBox["T", 
            RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
           "\[Equal]", "b"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
              "-", "1"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["T", "i"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", "r"], "D"], 
             SubscriptBox["T", 
              RowBox[{"i", "+", "1"}]]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           RowBox[{"a", "\[NotEqual]", "b"}], "\[Implies]", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{"a", 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", "r"], "D"], 
             SubscriptBox["T", "1"]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(37.\[DoubleLeftArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2048103084", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.644296309190262*^9, {3.644297106852684*^9, 3.644297107112184*^9}, {
   3.644895873668518*^9, 3.644895877350914*^9}, {3.6454315426684227`*^9, 
   3.64543155791982*^9}},
 CellTags->{
  "ID:2048103084", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "37.\[DoubleLeftArrow]"},
 CellID->2048103084],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->92140323]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1764183355],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(37.\[DoubleRightArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(37.\[DoubleRightArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    "1"]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    "1"]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1914773542"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1914773542-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1914773542-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1914773542", "Proof|ID:1914773542-1"},
 CellID->450197198],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1914773542-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1657761586]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(37.\[DoubleLeftArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(37.\[DoubleLeftArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], "1"]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], "1"]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2048103084"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2048103084-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2048103084-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2048103084", "Proof|ID:2048103084-1"},
 CellID->1611988294],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2048103084-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1040910915]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1856959695],

Cell[CellGroupData[{

Cell["Theorem: Alternative Description of Connectibility Below", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.644296326460381*^9, 3.644296364118567*^9}, {
  3.64429732575182*^9, 3.644297330645976*^9}, {3.6452601266969023`*^9, 
  3.645260130270515*^9}},
 CellID->382418526],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b", ",", "z"}], "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["isReductor", "D"], "[", "r", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 254692456}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6442310374288883`*^9, 3.644231048533676*^9}, {
  3.644291001669936*^9, 3.644291004887824*^9}, {3.644751890147048*^9, 
  3.644751896400757*^9}, {3.645431670897253*^9, 3.645431691539029*^9}},
 CellTags->{
  "ID:254692456", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->254692456],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
      RowBox[{"z", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           SubscriptBox["T", 
            RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
           "\[Equal]", "b"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
              "-", "1"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["T", "i"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", "r"], "D"], 
             SubscriptBox["T", 
              RowBox[{"i", "+", "1"}]]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{"z", 
             UnderscriptBox["\[Succeeds]", "D"], 
             SubscriptBox["T", "i"]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           RowBox[{"a", "\[NotEqual]", "b"}], "\[Implies]", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{"a", 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", "r"], "D"], 
               SubscriptBox["T", "1"]}], 
              TagBox[")",
               "AutoParentheses"]}], "\[And]", 
             RowBox[{"z", 
              UnderscriptBox["\[Succeeds]", "D"], "a"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(32.\[DoubleRightArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:336730845", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644297312525915*^9, 3.644297321539915*^9}, {
  3.644895879711549*^9, 3.644895884722096*^9}},
 CellTags->{
  "ID:336730845", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "32.\[DoubleRightArrow]"},
 CellID->336730845],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           SubscriptBox["T", 
            RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
           "\[Equal]", "b"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
              "-", "1"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["T", "i"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", "r"], "D"], 
             SubscriptBox["T", 
              RowBox[{"i", "+", "1"}]]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{"z", 
             UnderscriptBox["\[Succeeds]", "D"], 
             SubscriptBox["T", "i"]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           RowBox[{"a", "\[NotEqual]", "b"}], "\[Implies]", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{"a", 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", "r"], "D"], 
               SubscriptBox["T", "1"]}], 
              TagBox[")",
               "AutoParentheses"]}], "\[And]", 
             RowBox[{"z", 
              UnderscriptBox["\[Succeeds]", "D"], "a"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
      RowBox[{"z", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(32.\[DoubleLeftArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1715399148", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.644296309190262*^9, {3.644297106852684*^9, 3.644297107112184*^9}, {
   3.644895873668518*^9, 3.644895877350914*^9}},
 CellTags->{
  "ID:1715399148", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "32.\[DoubleLeftArrow]"},
 CellID->1715399148],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2087871386]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->418896992],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(32.\[DoubleRightArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(32.\[DoubleRightArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], "1"]}], "\[And]", 
                   RowBox[{
                    StyleBox["z", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox[
                    "a", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], "1"]}], "\[And]", 
                   RowBox[{
                    StyleBox["z", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox[
                    "a", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:336730845"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p336730845-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p336730845-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:336730845", "Proof|ID:336730845-1"},
 CellID->650311272],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p336730845-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6454317204885483`*^9},
 CellID->1958611019]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(32.\[DoubleLeftArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(32.\[DoubleLeftArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], "1"]}], "\[And]", 
                   RowBox[{
                    StyleBox["z", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["a", "ExpressionVariable"]}]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], "1"]}], "\[And]", 
                   RowBox[{
                    StyleBox["z", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["a", "ExpressionVariable"]}]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1715399148"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1715399148-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1715399148-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1715399148", "Proof|ID:1715399148-1"},
 CellID->2088013031],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1715399148-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.64543246505374*^9, 3.645432492794407*^9}},
 CellID->2115273351]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Properties of ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[Rule]", "*"], TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.6455073537875233`*^9, 3.645507363452927*^9}, {
  3.645507425697204*^9, 3.64550743092542*^9}, {3.645508901435052*^9, 
  3.64550890570687*^9}},
 CellID->829207957],

Cell["", "OpenEnvironment",
 CellID->525227380],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[Rule]", "*"], TraditionalForm]]],
 " is Reflexive and Transitive"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645507370718237*^9, 3.645507380694869*^9}},
 CellID->838178864],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b", ",", "c"}], "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["isReductor", "D"], "[", "r", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1253203360}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.645507390673046*^9, {3.645508088691637*^9, 3.645508094713439*^9}},
 CellTags->{
  "ID:1253203360", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1253203360],

Cell[BoxData[
 RowBox[{"a", 
  UnderscriptBox[
   SubsuperscriptBox["\[Rule]", "r", "*"], "D"], 
  "a"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reflexivity of \[RightArrow]*)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1514263990", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645507394023655*^9, 3.6455074076401443`*^9}},
 CellTags->{
  "ID:1514263990", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[RightArrow]*"},
 CellID->1514263990],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", 
      UnderscriptBox[
       SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "c"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "c"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(transitivity of \[RightArrow]*)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:411529944", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645507414752275*^9, 3.645507422317688*^9}},
 CellTags->{
  "ID:411529944", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[RightArrow]*"},
 CellID->411529944],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1053827266]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1233556406],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reflexivity of \[RightArrow]*)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reflexivity of \[RightArrow]*)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox[
            TemplateBox[{"\[Rule]", 
              StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
             SyntaxForm -> SubsuperscriptBox], 
            StyleBox["D", "ExpressionVariable"]], 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox[
            TemplateBox[{"\[Rule]", 
              StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
             SyntaxForm -> SubsuperscriptBox], 
            StyleBox["D", "ExpressionVariable"]], 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1514263990"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1514263990-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1514263990-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1514263990", "Proof|ID:1514263990-1"},
 CellID->1590090525],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1514263990-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1795169377]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(transitivity of \[RightArrow]*)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(transitivity of \[RightArrow]*)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:411529944"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p411529944-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p411529944-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:411529944", "Proof|ID:411529944-1"},
 CellID->144989886],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p411529944-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->914771064]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->979821556],

Cell[CellGroupData[{

Cell["Corollary", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.645507443436063*^9},
 CellID->593977615],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b", ",", "c"}], "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["isReductor", "D"], "[", "r", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1862973197}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.645507390673046*^9, {3.645508097255989*^9, 3.645508103555169*^9}},
 CellTags->{
  "ID:1862973197", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1862973197],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", "r"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[RightArrow] \[SubsetEqual] \[RightArrow]*)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:237268641", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645507453611772*^9, 3.645507473706024*^9}},
 CellTags->{
  "ID:237268641", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[SubsetEqual] \[RightArrow]*"},
 CellID->237268641],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", "r"], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", 
      UnderscriptBox[
       SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "c"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "c"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(38)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1746879589", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645507414752275*^9, 3.645507422317688*^9}, {
  3.645507481590701*^9, 3.645507481848012*^9}},
 CellTags->{
  "ID:1746879589", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "38"},
 CellID->1746879589],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->982648542]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2144214436],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] \[SubsetEqual] \[RightArrow]*)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] \[SubsetEqual] \[RightArrow]*)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["r", "ExpressionVariable"]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["r", "ExpressionVariable"]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:237268641"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p237268641-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p237268641-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:237268641", "Proof|ID:237268641-1"},
 CellID->1915348311],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p237268641-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1480957704]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(38)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(38)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["r", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["r", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1746879589"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1746879589-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1746879589-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1746879589", "Proof|ID:1746879589-1"},
 CellID->1067079851],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1746879589-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->657142897]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[LeftRightArrow]", "*"], TraditionalForm]]],
 " is Equivalence Relation (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.644296823243661*^9, 3.644296847205621*^9}, {
  3.644900478403574*^9, 3.644900482316424*^9}, {3.6449015263437643`*^9, 
  3.644901535680017*^9}, {3.644903937312841*^9, 3.6449039415906*^9}, {
  3.645260095054173*^9, 3.645260103170607*^9}, {3.6454315971633*^9, 
  3.6454316002081127`*^9}},
 CellID->979571816],

Cell["", "OpenEnvironment",
 CellID->805924195],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[LeftRightArrow]", "*"], TraditionalForm]]],
 " is Equivalence Relation"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6454328895286427`*^9, 3.645432901793734*^9}},
 CellID->783637779],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b", ",", "c"}], "]"}]}], ",", 
   "r"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1956094048}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645432912912079*^9, 3.645432915735721*^9}, 
   3.645434083730936*^9},
 CellTags->{
  "ID:1956094048", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1956094048],

Cell[BoxData[
 RowBox[{"a", 
  UnderscriptBox[
   SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], "D"], 
  "a"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reflexivity of connectibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:560057636", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644901653069129*^9, 3.644901697885744*^9}, {
  3.6454340020230417`*^9, 3.64543400625425*^9}},
 CellTags->{
  "ID:560057636", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of connectibility"},
 CellID->560057636],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"b", 
    UnderscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], "D"], "a"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symmetry of connectibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:354829984", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644296859934744*^9, 3.644296875410399*^9}, {
  3.6447519094896708`*^9, 3.644751915840324*^9}, {3.644901605670311*^9, 
  3.64490161090224*^9}, {3.645434008098359*^9, 3.645434010936932*^9}},
 CellTags->{
  "ID:354829984", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of connectibility"},
 CellID->354829984],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", 
      UnderscriptBox[
       SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], "D"], "c"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], "D"], "c"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(transitivity of connectibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:350848859", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644296518675713*^9, 3.6442966380983553`*^9}, {
  3.645434033207209*^9, 3.645434043294722*^9}},
 CellTags->{
  "ID:350848859", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of connectibility"},
 CellID->350848859],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1886172493]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->69823970],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reflexivity of connectibility)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reflexivity of connectibility)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox[
            TemplateBox[{"\[LeftRightArrow]", 
              StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
             SyntaxForm -> SubsuperscriptBox], 
            StyleBox["D", "ExpressionVariable"]], 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox[
            TemplateBox[{"\[LeftRightArrow]", 
              StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
             SyntaxForm -> SubsuperscriptBox], 
            StyleBox["D", "ExpressionVariable"]], 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:560057636"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p560057636-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p560057636-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:560057636", "Proof|ID:560057636-1"},
 CellID->1331060490],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p560057636-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.645507806007218*^9},
 CellID->1930057720]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symmetry of connectibility)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symmetry of connectibility)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:354829984"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p354829984-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p354829984-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:354829984", "Proof|ID:354829984-1"},
 CellID->667920451],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p354829984-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->263591681]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(transitivity of connectibility)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(transitivity of connectibility)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[LeftRightArrow]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[LeftRightArrow]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[LeftRightArrow]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[LeftRightArrow]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:350848859"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p350848859-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p350848859-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:350848859", "Proof|ID:350848859-1"},
 CellID->1457001757],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p350848859-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6455079672794037`*^9, 3.645508016516119*^9},
 CellID->315146723]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1576185928],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[OpenCurlyDoubleQuote]Connectibility Below\[CloseCurlyDoubleQuote] \
is Equivalence Relation\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.644296405203807*^9, 3.644296421209394*^9}, {
  3.644296777266202*^9, 3.6442967815066957`*^9}, {3.644901551098319*^9, 
  3.644901551778028*^9}, {3.644901587038118*^9, 3.6449015905313177`*^9}, {
  3.644901621112979*^9, 3.644901621342866*^9}, {3.645260146866157*^9, 
  3.645260152739315*^9}},
 CellID->641705771],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b", ",", "c", ",", "z"}], "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["isReductor", "D"], "[", "r", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 705004664}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6442310374288883`*^9, 3.644231048533676*^9}, {
   3.644291001669936*^9, 3.644291004887824*^9}, {3.644296501748105*^9, 
   3.6442965025207167`*^9}, 3.644296616527774*^9, {3.6447519221750183`*^9, 
   3.644751928079328*^9}, {3.644901593294742*^9, 3.644901594257641*^9}},
 CellTags->{
  "ID:705004664", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->705004664],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
      RowBox[{"z", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], "a"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"z", 
    UnderscriptBox["\[Succeeds]", "D"], "a"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reflexivity of connectibility below)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:411199728", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644901653069129*^9, 3.644901697885744*^9}},
 CellTags->{
  "ID:411199728", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of connectibility below"},
 CellID->411199728],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
      RowBox[{"z", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"b", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
      RowBox[{"z", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], "a"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symmetry of connectibility below)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1502850808", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644296859934744*^9, 3.644296875410399*^9}, {
  3.6447519094896708`*^9, 3.644751915840324*^9}, {3.644901605670311*^9, 
  3.64490161090224*^9}},
 CellTags->{
  "ID:1502850808", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of connectibility below"},
 CellID->1502850808],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       OverscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
        RowBox[{"z", ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox["\[Succeeds]",
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}]], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", 
      UnderscriptBox[
       OverscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
        RowBox[{"z", ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox["\[Succeeds]",
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}]], "D"], "c"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
      RowBox[{"z", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], "c"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(transitivity of connectibility below)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1331049102", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644296518675713*^9, 3.6442966380983553`*^9}, {
  3.645434019108795*^9, 3.645434030258338*^9}},
 CellTags->{
  "ID:1331049102", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of connectibility below"},
 CellID->1331049102],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1022905620]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1032464773],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reflexivity of connectibility below)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reflexivity of connectibility below)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["a", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["z", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["a", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["z", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:411199728"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p411199728-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p411199728-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:411199728", "Proof|ID:411199728-1"},
 CellID->24086679],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p411199728-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.645434060569216*^9, 3.6454340664131927`*^9}, 
   3.645434188276664*^9},
 CellID->1856563014]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symmetry of connectibility below)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symmetry of connectibility below)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1502850808"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1502850808-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1502850808-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1502850808", "Proof|ID:1502850808-1"},
 CellID->1390382820],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1502850808-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.645434301073236*^9, 3.645434304232912*^9}},
 CellID->726470046]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(transitivity of connectibility below)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(transitivity of connectibility below)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1331049102"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1331049102-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1331049102-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1331049102", "Proof|ID:1331049102-1"},
 CellID->516908597],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1331049102-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6454345902915487`*^9, 3.6454348040488977`*^9},
 CellID->966474202]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->2082958952],

Cell[CellGroupData[{

Cell["Corollary: Extended Transitivity of Connectibility Below", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6449017176615*^9, 3.644901733280807*^9}, {
  3.6452601559079533`*^9, 3.645260159959011*^9}},
 CellID->178927685],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "D", "]"}]], 
    RowBox[{"[", "B", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", 
      RowBox[{"z", ",", "a", ",", "b", ",", "c", ",", "d"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", 
         UnderscriptBox[
          OverscriptBox[
           SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
           RowBox[{"z", ",", 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             TagBox["\[Succeeds]",
              Identity], 
             TagBox[")",
              "AutoParentheses"]}]}]], "D"], "b"}], 
        TagBox[")",
         "AutoParentheses"]}], "\[And]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"b", 
         UnderscriptBox[
          OverscriptBox[
           SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
           RowBox[{"z", ",", 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             TagBox["\[Succeeds]",
              Identity], 
             TagBox[")",
              "AutoParentheses"]}]}]], "D"], "c"}], 
        TagBox[")",
         "AutoParentheses"]}], "\[And]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"d", 
         UnderscriptBox[
          OverscriptBox[
           SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
           RowBox[{"z", ",", 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             TagBox["\[Succeeds]",
              Identity], 
             TagBox[")",
              "AutoParentheses"]}]}]], "D"], "c"}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", 
       UnderscriptBox[
        OverscriptBox[
         SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
         RowBox[{"z", ",", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           TagBox["\[Succeeds]",
            Identity], 
           TagBox[")",
            "AutoParentheses"]}]}]], "D"], "d"}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(36)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1147958002", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645244371775769*^9, 3.645244436752112*^9}, {
  3.645244555528846*^9, 3.645244557147045*^9}},
 CellTags->{
  "ID:1147958002", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "36"},
 CellID->1147958002],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "z", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["isReductor", "D"], "[", "r", "]"}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
          "\[Implies]", 
          RowBox[{"z", 
           UnderscriptBox["\[Succeeds]", "D"], 
           SubscriptBox["T", "1"]}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[And]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
            RowBox[{
             RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
             "-", "1"}]}]], 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            SubscriptBox["T", "i"], 
            UnderscriptBox[
             OverscriptBox[
              SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
              RowBox[{"z", ",", 
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                TagBox["\[Succeeds]",
                 Identity], 
                TagBox[")",
                 "AutoParentheses"]}]}]], "D"], 
            SubscriptBox["T", 
             RowBox[{"i", "+", "1"}]]}], 
           TagBox[")",
            "AutoParentheses"]}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          SubscriptBox["T", "1"], 
          UnderscriptBox[
           OverscriptBox[
            SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
            RowBox[{"z", ",", 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              TagBox["\[Succeeds]",
               Identity], 
              TagBox[")",
               "AutoParentheses"]}]}]], "D"], 
          SubscriptBox["T", "i"]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(33)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1663785837", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644296644820216*^9, 3.644296708720749*^9}, {
  3.6449017441310987`*^9, 3.644901768093569*^9}, {3.644903402889104*^9, 
  3.644903433411787*^9}},
 CellTags->{
  "ID:1663785837", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "33"},
 CellID->1663785837],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1128332623]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->726374127],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(36)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(36)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "d", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "d", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1147958002"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1147958002-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1147958002-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1147958002", "Proof|ID:1147958002-1"},
 CellID->18036380],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1147958002-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->11438426]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(33)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(33)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[Implies]", 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], "1"]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 UnderscriptBox[
                  OverscriptBox[
                   TemplateBox[{"\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                   RowBox[{
                    StyleBox["z", "ExpressionVariable"], ",", 
                    "\[Succeeds]"}]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], "1"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox[
                "i", "ExpressionVariable"]]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[Implies]", 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], "1"]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 UnderscriptBox[
                  OverscriptBox[
                   TemplateBox[{"\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                   RowBox[{
                    StyleBox["z", "ExpressionVariable"], ",", 
                    "\[Succeeds]"}]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], "1"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox[
                "i", "ExpressionVariable"]]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1663785837"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1663785837-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1663785837-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1663785837", "Proof|ID:1663785837-1"},
 CellID->1295009646],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1663785837-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1185868751]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "reduceM, \[Rule], ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[Rule]", "*"], TraditionalForm]]],
 " and Connectibility Below (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.644297483802142*^9, 3.644297503873774*^9}, {
  3.644297662798089*^9, 3.6442976701601267`*^9}, {3.644300386518688*^9, 
  3.644300387313809*^9}, {3.645158501665511*^9, 3.645158505386046*^9}, {
  3.64526013698073*^9, 3.645260141676702*^9}, {3.64818356830508*^9, 
  3.64818356937366*^9}, {3.648187401180982*^9, 3.648187403812372*^9}},
 CellID->1422409662],

Cell["", "OpenEnvironment",
 CellID->1607920767],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: reduceM and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[Rule]", "*"], TraditionalForm]]],
 " & symred and \[LeftRightArrow]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.644297517504917*^9, {3.644297833743298*^9, 3.644297843965358*^9}, {
   3.644915249221176*^9, 3.6449152538634644`*^9}},
 CellID->354446817],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "c"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 311737773}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644909941019253*^9, 3.644909953573756*^9}, {
  3.6449146138872547`*^9, 3.644914614004609*^9}},
 CellTags->{
  "ID:311737773", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->311737773],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["isReductor", "D"], "[", "r", "]"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "a"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(reflexivity of reflexive-transitive closure of \[RightArrow])", 
         "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:552003293", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644909970206254*^9, 3.644909997099921*^9}},
 CellTags->{
  "ID:552003293", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
   "reflexivity of reflexive-transitive closure of \[RightArrow]"},
 CellID->552003293],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "D", "]"}]], 
   RowBox[{"[", "T", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 129456084}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6449146251027393`*^9, 3.6449146293918*^9}},
 CellTags->{
  "ID:129456084", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->129456084],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     UnderscriptBox["isReducible", "D"], "[", 
     RowBox[{
      RowBox[{
       UnderscriptBox["reduceM", "D"], "[", 
       RowBox[{"a", ",", "c", ",", "T", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", "c", ",", 
      SubscriptBox["T", "i"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", 
      RowBox[{"{", "c", "}"}], "*"], "D"], 
    RowBox[{
     UnderscriptBox["reduceM", "D"], "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(34)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:441235278", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644297567839505*^9, 3.644297650002548*^9}, {
   3.644298526669635*^9, 3.6442985583043756`*^9}, 3.6449099591197968`*^9, {
   3.6449146165874*^9, 3.644914633446663*^9}, {3.661850748572815*^9, 
   3.661850750117218*^9}},
 CellTags->{
  "ID:441235278", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "34"},
 CellID->441235278],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"b1", ",", "b2"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
        RowBox[{
         UnderscriptBox["symred", "D"], "[", 
         RowBox[{
          RowBox[{
           UnderscriptBox["reduceM", "D"], "[", 
           RowBox[{"b1", ",", "c", ",", "T", ",", 
            RowBox[{"i", "-", "1"}]}], "]"}], ",", 
          RowBox[{
           UnderscriptBox["reduceM", "D"], "[", 
           RowBox[{"b1", ",", "c", ",", "T", ",", "i"}], "]"}], ",", "c", ",", 
          SubscriptBox["T", "i"]}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{"T", "\[NotEqual]", 
       RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
           "-", "1"}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{
           UnderscriptBox["reduceM", "D"], "[", 
           RowBox[{"b1", ",", "c", ",", "T", ",", "i"}], "]"}], 
          UnderscriptBox[
           SubscriptBox["\[LeftRightArrow]", 
            RowBox[{"{", "c", "}"}]], "D"], 
          RowBox[{
           UnderscriptBox["reduceM", "D"], "[", 
           RowBox[{"b1", ",", "c", ",", "T", ",", 
            RowBox[{"i", "+", "1"}]}], "]"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"b1", 
        UnderscriptBox[
         SubscriptBox["\[LeftRightArrow]", 
          RowBox[{"{", "c", "}"}]], "D"], 
        RowBox[{
         UnderscriptBox["reduceM", "D"], "[", 
         RowBox[{"b1", ",", "c", ",", "T", ",", "1"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(35)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1516348781", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6449146485809183`*^9, 3.644914762657498*^9}, {
  3.644914819477154*^9, 3.644914833197085*^9}, {3.644914867418569*^9, 
  3.644914907461705*^9}, {3.644914956728726*^9, 3.644914985942788*^9}, {
  3.644915042308362*^9, 3.644915043781282*^9}, {3.6449176057069893`*^9, 
  3.644917609862731*^9}, {3.661850755297427*^9, 3.6618507638774424`*^9}},
 CellTags->{
  "ID:1516348781", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "35"},
 CellID->1516348781],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->412882915]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->830025092],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reflexivity of reflexive-transitive closure of \
\[RightArrow])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reflexivity of reflexive-transitive closure of \
\[RightArrow])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox[
            TemplateBox[{"\[Rule]", 
              StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
             SyntaxForm -> SubsuperscriptBox], 
            StyleBox["D", "ExpressionVariable"]], 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox[
            TemplateBox[{"\[Rule]", 
              StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
             SyntaxForm -> SubsuperscriptBox], 
            StyleBox["D", "ExpressionVariable"]], 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:552003293"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p552003293-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p552003293-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:552003293", "Proof|ID:552003293-1"},
 CellID->84172006],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p552003293-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->389665225]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(34)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(34)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["isReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                 ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                RowBox[{"{", 
                  StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["T", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["T", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["isReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                 ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                RowBox[{"{", 
                  StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               UnderscriptBox["reduceM", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["T", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["T", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:441235278"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p441235278-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p441235278-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:441235278", "Proof|ID:441235278-1"},
 CellID->9301273],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p441235278-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->319931820]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(35)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(35)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], "\[And]", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   StyleBox["c", "ExpressionVariable"]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]}], 
                   "]"}]}]}], "\[And]", 
             RowBox[{
               StyleBox["b1", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 StyleBox["c", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["b1", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", "1"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["symred", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                   ",", 
                   RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], "\[And]", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   StyleBox["c", "ExpressionVariable"]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]}], 
                   "]"}]}]}], "\[And]", 
             RowBox[{
               StyleBox["b1", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 StyleBox["c", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 UnderscriptBox["reduceM", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["b1", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["T", "ExpressionVariable"], ",", "1"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1516348781"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1516348781-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1516348781-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1516348781", "Proof|ID:1516348781-1"},
 CellID->278395591],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1516348781-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->766798721]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (34)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.644298746007558*^9, 3.6442987469059153`*^9}, 
   3.6443003254630632`*^9, {3.6449119086461554`*^9, 3.644911910924864*^9}, {
   3.6449122194333687`*^9, 3.6449122204936113`*^9}, 3.6532195888453116`*^9, 
   3.6661744073152475`*^9, 3.666440978996463*^9},
 CellID->973241208],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "c"}], "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "T", "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
    RowBox[{"isReducible", "[", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{"a", ",", "c", ",", "T", ",", 
        RowBox[{"i", "-", "1"}]}], "]"}], ",", "c", ",", 
      SubscriptBox["T", "i"]}], "]"}]}], TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubsuperscriptBox["\[Rule]", "c", "*"], 
    RowBox[{"reduceM", "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
     "]"}]}], TraditionalForm]]],
 "\nwhich, by a lemma on alternative descriptions of ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[Rule]", "*"], TraditionalForm]]],
 ", can be reduced to finding ",
 Cell[BoxData[
  FormBox["S", TraditionalForm]]],
 " with\n(G.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "S", "]"}]}], TraditionalForm]]],
 "\n(G.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", 
     RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]], "=", 
    "reduceM"}], TraditionalForm]]],
 "\n(G.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}], "-", 
        "1"}]}]], 
     SubscriptBox["S", "i"]}], 
    SubscriptBox["\[Rule]", "c"], 
    SubscriptBox["S", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 "\n(G.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"a", "\[NotEqual]", "reduceM"}], "\[Implies]", "a"}], 
    SubscriptBox["\[Rule]", "c"], 
    SubscriptBox["S", "1"]}], TraditionalForm]]],
 "\n\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", ":=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"reduceM", "[", 
       RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 " witnesses our existential goal, hence we show all four subgoals.\n\n(G.1):\
\nObvious.\n\n(G.2):\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", 
     RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]], "=", 
    RowBox[{"reduceM", "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
     "]"}]}], TraditionalForm]]],
 ", using ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}], "=", 
    RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 ".\n\n(G.3):\nChoose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], "-", 
     "1"}]}], TraditionalForm]]],
 " a.b.f. and show\n(G.5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"reduceM", "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}], 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{"reduceM", "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{"i", "+", "1"}]}], "]"}]}], TraditionalForm]]],
 "\nThe RHS of (G.5) is equal to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"reduceM", "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}], "-", 
    RowBox[{
     SubscriptBox["T", 
      RowBox[{"i", "+", "1"}]], "*", "c"}]}], TraditionalForm]]],
 ", and by the definition of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Rule]", "c"], TraditionalForm]]],
 " we now need to show\n(G.6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", "m"], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"mult", "[", 
        RowBox[{"m", ",", "c"}], "]"}], "\[And]", 
       RowBox[{
        RowBox[{"reduceM", "[", 
         RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}], "-", 
        RowBox[{
         SubscriptBox["T", 
          RowBox[{"i", "+", "1"}]], "*", "c"}]}]}], "=", 
      RowBox[{
       RowBox[{
        RowBox[{"reduceM", "[", 
         RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}], "-", 
        RowBox[{"m", "*", "c"}]}], "\[And]", 
       RowBox[{
        RowBox[{"reduceM", "[", 
         RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}], "\[Succeeds]", 
        RowBox[{
         RowBox[{"reduceM", "[", 
          RowBox[{"a", ",", "c", ",", "T", ",", "i"}], "]"}], "-", 
         RowBox[{
          SubscriptBox["T", 
           RowBox[{"i", "+", "1"}]], "*", "c"}]}]}]}]}], ")"}]}], 
   TraditionalForm]]],
 "\n\nIt is clear that ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["T", 
    RowBox[{"i", "+", "1"}]], TraditionalForm]]],
 " witnesses this existential goal (using (1) instantiated by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "+", "1"}], TraditionalForm]]],
 ").\n\n(G.4):\nAssume\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[NotEqual]", "reduceM"}], TraditionalForm]]],
 "\nand show\n(G.7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", 
     SubscriptBox["\[Rule]", "c"], 
     SubscriptBox["S", "1"]}], "=", 
    RowBox[{
     RowBox[{"reduceM", "[", 
      RowBox[{"a", ",", "c", ",", "T", ",", "1"}], "]"}], "=", 
     RowBox[{"a", "-", 
      RowBox[{
       SubscriptBox["T", "1"], "*", "c"}]}]}]}], TraditionalForm]]],
 "\n\nFrom (2) we can immediately infer ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[NotEqual]", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ", for otherwise ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", 
    RowBox[{"reduceM", "[", 
     RowBox[{"a", ",", "c", ",", "T", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
     "]"}]}], TraditionalForm]]],
 ", hence we can instantiate (1) by 1 and infer\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{
     RowBox[{"reduceM", "[", 
      RowBox[{"a", ",", "c", ",", "T", ",", "0"}], "]"}], ",", "c", ",", 
     SubscriptBox["T", "1"]}], "]"}], TraditionalForm]]],
 "\nThe first argument of the above predicate is simply ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 ", so we are done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.644298757423999*^9, 3.644298806555686*^9}, {
   3.644298844250231*^9, 3.6442988855532427`*^9}, {3.644298933048894*^9, 
   3.644299177312673*^9}, {3.6442992231760483`*^9, 3.644299273579245*^9}, {
   3.6442993712850246`*^9, 3.644299412653647*^9}, {3.6442994620537157`*^9, 
   3.644299644260647*^9}, {3.644299681114841*^9, 3.644299941427703*^9}, {
   3.644299986037951*^9, 3.64430007458836*^9}, 3.644300325463756*^9, {
   3.644908128592868*^9, 3.64490813117277*^9}, {3.644908217825225*^9, 
   3.644908217825342*^9}, 3.6532195888453116`*^9, 3.6661744073152475`*^9, 
   3.666440978996615*^9},
 CellID->1853007915],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.64430008120772*^9, 3.644300085222479*^9}, 
   3.644300325463892*^9, 3.6532195888453116`*^9, 3.6661744073152475`*^9, 
   3.666440978996785*^9},
 TextAlignment->Right,
 CellID->922558241]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->598312948],

Cell[CellGroupData[{

Cell["Theorem", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.644297852357534*^9},
 CellID->1969644473],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "c"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1329557757}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644298170114852*^9, 3.644298177037266*^9}},
 CellTags->{
  "ID:1329557757", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1329557757],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "b", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"{", "c", "}"}]], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], "[", "d", "]"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"a", 
          UnderscriptBox["+", "D"], "d"}], 
         TagBox[")",
          "AutoParentheses"]}], 
        UnderscriptBox[
         SubsuperscriptBox["\[DownArrow]", 
          RowBox[{"{", "c", "}"}], "*"], "D"], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"b", 
          UnderscriptBox["+", "D"], "d"}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(40)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1168535485", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644297876360323*^9, 3.644297882001321*^9}, {
   3.644297988444496*^9, 3.644298060658243*^9}, {3.6442981391086817`*^9, 
   3.644298141552231*^9}, 3.644298180186531*^9, {3.6618508002421064`*^9, 
   3.6618508071841183`*^9}},
 CellTags->{
  "ID:1168535485", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "40"},
 CellID->1168535485],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2", ",", "b1", ",", "b2"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", 
        UnderscriptBox[
         SubscriptBox["\[Rule]", 
          RowBox[{"m1", ",", "c"}]], "D"], "b1"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", 
        UnderscriptBox[
         SubscriptBox["\[Rule]", 
          RowBox[{"m2", ",", "c"}]], "D"], "b2"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       UnderscriptBox["areCorrelative", "D"], "[", 
       RowBox[{"c", ",", "m1", ",", "m2"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b1", 
      UnderscriptBox[
       OverscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", 
         RowBox[{"{", "c", "}"}], "*"], 
        RowBox[{"a", ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox["\[Succeeds]",
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}]], "D"], "b2"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(41)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2128489125", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6442980904768972`*^9, 3.644298123739916*^9}, 
   3.6442981828140297`*^9, {3.661850812098127*^9, 3.6618508134085293`*^9}},
 CellTags->{
  "ID:2128489125", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "41"},
 CellID->2128489125],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1710755144]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1362228690],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(40)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(40)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{"{", 
                 StyleBox["c", "ExpressionVariable"], "}"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}], 
               UnderscriptBox[
                TemplateBox[{"\[DownArrow]", 
                  RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                 "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "d", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{"{", 
                 StyleBox["c", "ExpressionVariable"], "}"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}], 
               UnderscriptBox[
                TemplateBox[{"\[DownArrow]", 
                  RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                 "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "d", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1168535485"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1168535485-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1168535485-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1168535485", "Proof|ID:1168535485-1"},
 CellID->397093125],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1168535485-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1687302]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(41)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(41)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["b1", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["b1", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2128489125"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2128489125-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2128489125-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2128489125", "Proof|ID:2128489125-1"},
 CellID->360602067],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2128489125-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->320046888]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1845228130],

Cell[CellGroupData[{

Cell["Corollary", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.6454360206466913`*^9},
 CellID->785213559],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "c"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 548213617}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647938433737364*^9, 3.647938441072775*^9}},
 CellTags->{
  "ID:548213617", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->548213617],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"c", "\[NotEqual]", 
       UnderscriptBox["0", "D"]}], "\[And]", 
      RowBox[{
       UnderscriptBox["mult", "D"], "[", 
       RowBox[{"m", ",", "c"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubsuperscriptBox["\[DownArrow]", 
        RowBox[{"{", "c", "}"}], "*"], "D"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{"m", 
         UnderscriptBox["*", "D"], "c"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(42)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:909891749", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645430915300221*^9, 3.645431045619227*^9}, {
  3.645431113468923*^9, 3.645431118323848*^9}, {3.6454360504499693`*^9, 
  3.645436079579565*^9}, {3.645436174144936*^9, 3.6454361778612747`*^9}, {
  3.647938450301306*^9, 3.6479384809940863`*^9}, {3.661850857104206*^9, 
  3.6618508578530073`*^9}},
 CellTags->{
  "ID:909891749", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "42"},
 CellID->909891749],

Cell[BoxData[
 RowBox[{"a", 
  UnderscriptBox[
   SubsuperscriptBox["\[DownArrow]", 
    RowBox[{"{", 
     RowBox[{"a", 
      UnderscriptBox["-", "D"], "c"}], "}"}], "*"], "D"], 
  "c"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(44)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:830274722", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6479384948826*^9, 3.6479385134650927`*^9}, {
  3.661850866323822*^9, 3.6618508700210285`*^9}},
 CellTags->{
  "ID:830274722", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "44"},
 CellID->830274722],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"a", 
    UnderscriptBox["-", "D"], "c"}], "\[NotEqual]", 
   UnderscriptBox["0", "D"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1390954175}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6481831201732693`*^9, 3.6481831255732927`*^9}},
 CellTags->{
  "ID:1390954175", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1390954175],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{"a", ",", 
    RowBox[{"a", 
     UnderscriptBox["-", "D"], "c"}]}], "]"}], "\[Or]", 
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{"c", ",", 
    RowBox[{"a", 
     UnderscriptBox["-", "D"], "c"}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(68)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2016336231", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6481829982281847`*^9, 3.6481830091149263`*^9}, {
  3.648183068430949*^9, 3.6481831311043177`*^9}, {3.6481831682445393`*^9, 
  3.648183190636725*^9}},
 CellTags->{
  "ID:2016336231", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68"},
 CellID->2016336231],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "D", "]"}]], 
    RowBox[{"[", "B", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "\[Element]", 
     RowBox[{
      UnderscriptBox["red", "D"], "[", 
      RowBox[{"B", "\[Union]", 
       RowBox[{"{", 
        RowBox[{"a", 
         UnderscriptBox["-", "D"], "c"}], "}"}]}], "]"}]}], "\[Or]", 
    RowBox[{"c", "\[Element]", 
     RowBox[{
      UnderscriptBox["red", "D"], "[", 
      RowBox[{"B", "\[Union]", 
       RowBox[{"{", 
        RowBox[{"a", 
         UnderscriptBox["-", "D"], "c"}], "}"}]}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(68.set)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1052198272", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6481832373722553`*^9, 3.648183277123489*^9}},
 CellTags->{
  "ID:1052198272", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68.set"},
 CellID->1052198272],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "T", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "\[Element]", 
     RowBox[{
      UnderscriptBox["red", "D"], "[", 
      RowBox[{"T", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", 
         UnderscriptBox["-", "D"], "c"}], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}]}], "\[Or]", 
    RowBox[{"c", "\[Element]", 
     RowBox[{
      UnderscriptBox["red", "D"], "[", 
      RowBox[{"T", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", 
         UnderscriptBox["-", "D"], "c"}], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(68.tuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:35312349", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6481832900993147`*^9, 3.6481833192342777`*^9}},
 CellTags->{
  "ID:35312349", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68.tuple"},
 CellID->35312349],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->112326196]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1212768896],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(42)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(42)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[DownArrow]", 
                RowBox[{"{", 
                  StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[DownArrow]", 
                RowBox[{"{", 
                  StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
               "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:909891749"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p909891749-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p909891749-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:909891749", "Proof|ID:909891749-1"},
 CellID->534981557],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p909891749-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->24744825]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(44)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(44)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox[
            TemplateBox[{"\[DownArrow]", 
              RowBox[{"{", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], 
                  UnderscriptBox["-", 
                   StyleBox["D", "ExpressionVariable"]], 
                  StyleBox["c", "ExpressionVariable"]}], "}"}], "*"}, 
             "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
            StyleBox["D", "ExpressionVariable"]], 
           StyleBox[
           "c", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox[
            TemplateBox[{"\[DownArrow]", 
              RowBox[{"{", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], 
                  UnderscriptBox["-", 
                   StyleBox["D", "ExpressionVariable"]], 
                  StyleBox["c", "ExpressionVariable"]}], "}"}], "*"}, 
             "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
            StyleBox["D", "ExpressionVariable"]], 
           StyleBox[
           "c", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:830274722"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p830274722-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p830274722-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:830274722", "Proof|ID:830274722-1"},
 CellID->601881535],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p830274722-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->132577742]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(68)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(68)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}], "\[NotEqual]", 
             UnderscriptBox["0", 
              StyleBox["D", "ExpressionVariable"]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], "]"}], "\[Or]", 
             RowBox[{
               UnderscriptBox["isReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}], "\[NotEqual]", 
             UnderscriptBox["0", 
              StyleBox["D", "ExpressionVariable"]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], "]"}], "\[Or]", 
             RowBox[{
               UnderscriptBox["isReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2016336231"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2016336231-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2016336231-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2016336231", "Proof|ID:2016336231-1"},
 CellID->490882312],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2016336231-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->239890867]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(68.set)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(68.set)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}], "\[NotEqual]", 
             UnderscriptBox["0", 
              StyleBox["D", "ExpressionVariable"]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSets", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[Union]", 
                    RowBox[{"{", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}], "}"}]}], "]"}]}], 
               "\[Or]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[Union]", 
                    RowBox[{"{", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}], "}"}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}], "\[NotEqual]", 
             UnderscriptBox["0", 
              StyleBox["D", "ExpressionVariable"]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSets", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[Union]", 
                    RowBox[{"{", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}], "}"}]}], "]"}]}], 
               "\[Or]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[Union]", 
                    RowBox[{"{", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}], "}"}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1052198272"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1052198272-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1052198272-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1052198272", "Proof|ID:1052198272-1"},
 CellID->1895460698],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1052198272-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1968311031]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(68.tuple)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(68.tuple)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}], "\[NotEqual]", 
             UnderscriptBox["0", 
              StyleBox["D", "ExpressionVariable"]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["T", "ExpressionVariable"], "\:293a", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], "]"}]}], "\[Or]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["T", "ExpressionVariable"], "\:293a", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}], "\[NotEqual]", 
             UnderscriptBox["0", 
              StyleBox["D", "ExpressionVariable"]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["T", "ExpressionVariable"], "\:293a", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], "]"}]}], "\[Or]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[Element]", 
                 RowBox[{
                   UnderscriptBox["red", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["T", "ExpressionVariable"], "\:293a", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:35312349"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p35312349-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p35312349-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:35312349", "Proof|ID:35312349-1"},
 CellID->860762961],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p35312349-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->40288143]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Sufficient Conditions for ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[LeftRightArrow]", "*"], TraditionalForm]]],
 " and Connectibility Below (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.644898722553632*^9, 3.644898731235159*^9}, {
   3.644905307176077*^9, 3.644905311041294*^9}, 3.6449054305014553`*^9, {
   3.6449058259578533`*^9, 3.6449058298597527`*^9}, {3.645260169925447*^9, 
   3.6452601744603148`*^9}, {3.6454361170644712`*^9, 3.645436124498715*^9}},
 CellID->922632239],

Cell["", "OpenEnvironment",
 CellID->1645031940],

Cell[CellGroupData[{

Cell["Theorem: Sufficient Conditions for Connectibility Below I", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6448987427137613`*^9, 3.644898764315895*^9}, {
  3.6449120745592613`*^9, 3.6449120748853416`*^9}, {3.6452601800830297`*^9, 
  3.645260180677662*^9}},
 CellID->1996820773],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b", ",", "d", ",", "z"}], "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["isReductor", "D"], "[", "r", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 2032809580}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644898778013983*^9, 3.644898790521718*^9}},
 CellTags->{
  "ID:2032809580", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2032809580],

Cell[BoxData[
 RowBox[{
  RowBox[{"z", 
   UnderscriptBox["\[Succeeds]", "D"], "a"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1947090769}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6449040361017036`*^9, 3.644904043816012*^9}},
 CellTags->{
  "ID:1947090769", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1947090769],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "T", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           SubscriptBox["T", "1"], "\[Equal]", "a"}]},
         {
          RowBox[{
           SubscriptBox["T", 
            RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
           "\[Equal]", "b"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
              "-", "1"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["T", "i"], 
             UnderscriptBox[
              OverscriptBox[
               SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
               RowBox[{"z", ",", 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 TagBox["\[Succeeds]",
                  Identity], 
                 TagBox[")",
                  "AutoParentheses"]}]}]], "D"], 
             SubscriptBox["T", 
              RowBox[{"i", "+", "1"}]]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
      RowBox[{"z", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(50)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:693923465", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644296518675713*^9, 3.6442966380983553`*^9}, {
  3.644296991013204*^9, 3.644297094020659*^9}, {3.644903278976861*^9, 
  3.6449033194125357`*^9}, {3.64490335693058*^9, 3.644903360549509*^9}, {
  3.644904046071851*^9, 3.644904048463786*^9}},
 CellTags->{
  "ID:693923465", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "50"},
 CellID->693923465],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
      RowBox[{"z", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(51)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1234215913", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6448988290630913`*^9, 3.644898867389001*^9}, {
  3.644904051147501*^9, 3.644904052214156*^9}},
 CellTags->{
  "ID:1234215913", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "51"},
 CellID->1234215913],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"z", 
     UnderscriptBox["\[Succeeds]", "D"], "b"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "d"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", 
      UnderscriptBox[
       SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "d"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
      RowBox[{"z", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(52)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1065734937", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644912496101457*^9, 3.644912573325588*^9}},
 CellTags->{
  "ID:1065734937", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "52"},
 CellID->1065734937],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "D", "]"}]], 
    RowBox[{"[", "B", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{"c", "\[Element]", "B"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", 
        UnderscriptBox[
         OverscriptBox[
          SubsuperscriptBox["\[LeftRightArrow]", 
           RowBox[{"{", "c", "}"}], "*"], 
          RowBox[{"z", ",", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            TagBox["\[Succeeds]",
             Identity], 
            TagBox[")",
             "AutoParentheses"]}]}]], "D"], "b"}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       OverscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
        RowBox[{"z", ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox["\[Succeeds]",
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}]], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(53)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1763925306", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.6442969661983433`*^9, {3.644297190513026*^9, 3.644297244117622*^9}, 
   3.6449054243575563`*^9, {3.6618509099414988`*^9, 3.6618509115795016`*^9}},
 CellTags->{
  "ID:1763925306", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "53"},
 CellID->1763925306],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1925417032]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1385730272],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(50)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(50)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["z", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "\[And]", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], "1"], "\[Equal]", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox[
                    OverscriptBox[
                    TemplateBox[{"\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    RowBox[{
                    StyleBox["z", "ExpressionVariable"], ",", 
                    "\[Succeeds]"}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}]}]}], 
             "\[Implies]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["z", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "\[And]", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], "1"], "\[Equal]", 
                   StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]], "\[Equal]", 
                   StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox[
                    OverscriptBox[
                    TemplateBox[{"\[LeftRightArrow]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    RowBox[{
                    StyleBox["z", "ExpressionVariable"], ",", 
                    "\[Succeeds]"}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}]}]}], 
             "\[Implies]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:693923465"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p693923465-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p693923465-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:693923465", "Proof|ID:693923465-1"},
 CellID->227607100],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p693923465-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->20341943]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(51)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(51)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["z", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["z", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1234215913"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1234215913-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1234215913-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1234215913", "Proof|ID:1234215913-1"},
 CellID->575064868],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1234215913-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6454371784042673`*^9, 3.645437207098846*^9}},
 CellID->1792184621]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(52)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(52)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["z", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                   "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox[
                  TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                   "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}], "\[Implies]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["z", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                   "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox[
                  TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                   "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}], "\[Implies]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1065734937"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1065734937-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1065734937-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1065734937", "Proof|ID:1065734937-1"},
 CellID->1345357544],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1065734937-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->761583954]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(53)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(53)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["z", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSets", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    OverscriptBox[
                    TemplateBox[{"\[LeftRightArrow]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    RowBox[{
                    StyleBox["z", "ExpressionVariable"], ",", 
                    "\[Succeeds]"}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  OverscriptBox[
                   TemplateBox[{"\[LeftRightArrow]", 
                    StyleBox["B", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                   RowBox[{
                    StyleBox["z", "ExpressionVariable"], ",", 
                    "\[Succeeds]"}]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "b", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["z", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSets", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox[
                    OverscriptBox[
                    TemplateBox[{"\[LeftRightArrow]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    RowBox[{
                    StyleBox["z", "ExpressionVariable"], ",", 
                    "\[Succeeds]"}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["b", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  OverscriptBox[
                   TemplateBox[{"\[LeftRightArrow]", 
                    StyleBox["B", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                   RowBox[{
                    StyleBox["z", "ExpressionVariable"], ",", 
                    "\[Succeeds]"}]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "b", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1763925306"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1763925306-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1763925306-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1763925306", "Proof|ID:1763925306-1"},
 CellID->278876241],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1763925306-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->710867754]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1541563091],

Cell[CellGroupData[{

Cell["Theorem: Sufficient Conditions for Connectibility Below II", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.645156667837028*^9, 3.6451569122492447`*^9, {3.645260183756575*^9, 
   3.64526018466173*^9}},
 CellID->450599753],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "z"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 332034572}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6451568999278393`*^9, 3.6451569665417147`*^9}, {
  3.645158014204029*^9, 3.645158016819631*^9}},
 CellTags->{
  "ID:332034572", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->332034572],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"z", 
     UnderscriptBox["\[Succeeds]", "D"], 
     RowBox[{"a", 
      UnderscriptBox["+", "D"], "d"}]}], "\[And]", 
    RowBox[{"z", 
     UnderscriptBox["\[Succeeds]", "D"], 
     RowBox[{"b", 
      UnderscriptBox["+", "D"], "d"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1609461850}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6451569715006733`*^9, 3.6451569960886097`*^9}},
 CellTags->{
  "ID:1609461850", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1609461850],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"{", "c", "}"}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox["+", "D"], "d"}], 
     TagBox[")",
      "AutoParentheses"]}], 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", 
       RowBox[{"{", "c", "}"}], "*"], 
      RowBox[{"z", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", 
      UnderscriptBox["+", "D"], "d"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(54)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:621597434", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6451570132914886`*^9, 3.645157068997509*^9}, {
  3.645158019878161*^9, 3.645158021460834*^9}, {3.6618509267271285`*^9, 
  3.661850933934341*^9}},
 CellTags->{
  "ID:621597434", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "54"},
 CellID->621597434],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", 
      RowBox[{"{", "c", "}"}]], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox["+", "D"], "d"}], 
     TagBox[")",
      "AutoParentheses"]}], 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", 
       RowBox[{"{", "c", "}"}], "*"], 
      RowBox[{"z", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", 
      UnderscriptBox["+", "D"], "d"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(55)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2031891226", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645157080017413*^9, 3.6451570966196833`*^9}, {
  3.645158022860683*^9, 3.645158024590557*^9}, {3.6618509373975472`*^9, 
  3.661850943575158*^9}},
 CellTags->{
  "ID:2031891226", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "55"},
 CellID->2031891226],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "D", "]"}]], 
   RowBox[{"[", "C", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1596753391}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645170951211688*^9, 3.6451709640980597`*^9}},
 CellTags->{
  "ID:1596753391", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1596753391],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", "C"], "D"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox["+", "D"], "d"}], 
     TagBox[")",
      "AutoParentheses"]}], 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "C", "*"], 
      RowBox[{"z", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", 
      UnderscriptBox["+", "D"], "d"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(55')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1855698114", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645170986274795*^9, 3.6451709906708508`*^9}},
 CellTags->{
  "ID:1855698114", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "55'"},
 CellID->1855698114],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->855194838]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1403478182],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(54)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(54)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}], "\[And]", 
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "d", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}], "\[And]", 
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "d", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:621597434"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p621597434-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p621597434-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:621597434", "Proof|ID:621597434-1"},
 CellID->88899104],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p621597434-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->76011636]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(55)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(55)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}], "\[And]", 
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "d", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}], "\[And]", 
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[LeftRightArrow]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox[
                 "d", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2031891226"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2031891226-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2031891226-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2031891226", "Proof|ID:2031891226-1"},
 CellID->50940196],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2031891226-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->258334955]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(55')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(55')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}], "\[And]", 
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSets", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   StyleBox["C", "ExpressionVariable"]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["d", "ExpressionVariable"]}], 
                 UnderscriptBox[
                  OverscriptBox[
                   TemplateBox[{"\[LeftRightArrow]", 
                    StyleBox["C", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                   RowBox[{
                    StyleBox["z", "ExpressionVariable"], ",", 
                    "\[Succeeds]"}]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox[
                   "d", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}], "\[And]", 
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["d", "ExpressionVariable"]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSets", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   StyleBox["C", "ExpressionVariable"]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox["d", "ExpressionVariable"]}], 
                 UnderscriptBox[
                  OverscriptBox[
                   TemplateBox[{"\[LeftRightArrow]", 
                    StyleBox["C", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                   RowBox[{
                    StyleBox["z", "ExpressionVariable"], ",", 
                    "\[Succeeds]"}]], 
                  StyleBox["D", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    StyleBox["D", "ExpressionVariable"]], 
                   StyleBox[
                   "d", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1855698114"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1855698114-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1855698114-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1855698114", "Proof|ID:1855698114-1"},
 CellID->204417677],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1855698114-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->171198437]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1430041170],

Cell[CellGroupData[{

Cell["Theorem: Sufficient Conditions for Connectibility Below III", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6451572981283827`*^9, 3.64515729938102*^9}, {
  3.645260187203158*^9, 3.6452601877625427`*^9}},
 CellID->1085794092],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", "a", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1116860460}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6451573310504637`*^9, 3.645157342479349*^9}, 
   3.645157599630454*^9, 3.645157666144607*^9},
 CellTags->{
  "ID:1116860460", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1116860460],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"b", ",", "c", ",", "z"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"z", 
       UnderscriptBox["\[Succeeds]", "D"], "a"}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", 
        UnderscriptBox[
         SubscriptBox["\[Rule]", 
          RowBox[{"{", "c", "}"}]], "D"], "b"}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       OverscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", 
         RowBox[{"{", "c", "}"}], "*"], 
        RowBox[{"z", ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox["\[Succeeds]",
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}]], "D"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(56)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:915980145", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6451573475705976`*^9, 3.645157377749158*^9}, {
   3.645157435706192*^9, 3.645157448400313*^9}, {3.645157587033538*^9, 
   3.645157597166094*^9}, {3.645157664176837*^9, 3.645157664398623*^9}, {
   3.645158608039131*^9, 3.645158638955986*^9}, 3.6451587138211813`*^9, {
   3.6618509667255983`*^9, 3.661850972045208*^9}},
 CellTags->{
  "ID:915980145", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "56"},
 CellID->915980145],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "D", "]"}]], 
   RowBox[{"[", "C", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1683552886}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645158922215042*^9, 3.645158933871058*^9}},
 CellTags->{
  "ID:1683552886", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1683552886],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"c1", ",", 
   RowBox[{"c2", "\[Element]", "C"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1331309258}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645158938942069*^9, 3.6451589485282393`*^9}},
 CellTags->{
  "ID:1331309258", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1331309258],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"b1", ",", "b2", ",", "m1", ",", "m2"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 551817557}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6451589512778482`*^9, 3.6451589557620277`*^9}},
 CellTags->{
  "ID:551817557", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->551817557],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"m1", ",", "c1"}]], "D"], "b1"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"m2", ",", "c2"}]], "D"], "b2"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1724920630}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645158976965208*^9, 3.645158986113743*^9}},
 CellTags->{
  "ID:1724920630", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1724920630],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "D"], "[", 
   RowBox[{"b2", ",", "c1", ",", "m1"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"b1", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "C", "*"], 
      RowBox[{"a", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], "b2"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(57)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:88655200", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645159016027142*^9, 3.645159019782435*^9}},
 CellTags->{
  "ID:88655200", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "57"},
 CellID->88655200],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isReducible", "D"], "[", 
     RowBox[{"b2", ",", "c1", ",", "m1"}], "]"}], "\[Or]", 
    RowBox[{
     UnderscriptBox["isReducible", "D"], "[", 
     RowBox[{"b1", ",", "c2", ",", "m2"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"b1", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "C", "*"], 
      RowBox[{"a", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "D"], "b2"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(58)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1149353049", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6451575362121162`*^9, 3.645157658289619*^9}, {
  3.64515895777617*^9, 3.645158997044244*^9}},
 CellTags->{
  "ID:1149353049", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "58"},
 CellID->1149353049],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->864867994]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1363814311],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(56)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(56)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["a", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["a", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{"{", 
                   StyleBox["c", "ExpressionVariable"], "}"}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:915980145"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p915980145-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p915980145-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:915980145", "Proof|ID:915980145-1"},
 CellID->552018998],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p915980145-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->76451817]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(57)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(57)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[Element]", 
               StyleBox["C", "ExpressionVariable"]}]}, {
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[Element]", 
               StyleBox["C", "ExpressionVariable"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["b2", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[Implies]", 
             
             RowBox[{
               StyleBox["b1", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["C", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b2", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[Element]", 
               StyleBox["C", "ExpressionVariable"]}]}, {
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[Element]", 
               StyleBox["C", "ExpressionVariable"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isReducible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["b2", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[Implies]", 
             
             RowBox[{
               StyleBox["b1", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["C", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b2", 
                "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:88655200"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p88655200-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p88655200-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:88655200", "Proof|ID:88655200-1"},
 CellID->7475235],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p88655200-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->186420841]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(58)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(58)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[Element]", 
               StyleBox["C", "ExpressionVariable"]}]}, {
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[Element]", 
               StyleBox["C", "ExpressionVariable"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["b2", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[Or]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["b1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"]}], "]"}]}], 
             "\[Implies]", 
             RowBox[{
               StyleBox["b1", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["C", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b2", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[Element]", 
               StyleBox["C", "ExpressionVariable"]}]}, {
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[Element]", 
               StyleBox["C", "ExpressionVariable"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["b2", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[Or]", 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["b1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"]}], "]"}]}], 
             "\[Implies]", 
             RowBox[{
               StyleBox["b1", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["C", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox[
               "b2", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1149353049"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1149353049-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1149353049-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1149353049", "Proof|ID:1149353049-1"},
 CellID->180962942],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1149353049-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->417960590]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof of (57).",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.609647816719001*^9, 3.609647835821731*^9}, 
   3.631855307870846*^9, 3.636348612088891*^9, 3.643533773210163*^9, {
   3.645157714858224*^9, 3.6451577150433683`*^9}, 3.6532195780500927`*^9, 
   3.6661743986416326`*^9, 3.666440967565358*^9},
 CellID->947895429],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "1"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "2"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["b", "1"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["b", "2"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "1"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["c", "1"]}]], 
    SubscriptBox["b", "1"]}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "2"], ",", 
      SubscriptBox["c", "2"]}]], 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"reducible", "[", 
    RowBox[{
     SubscriptBox["b", "2"], ",", 
     SubscriptBox["c", "1"], ",", 
     SubscriptBox["m", "1"]}], "]"}], TraditionalForm]]],
 "\nand show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "1"], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "C", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 "\n\nFrom (3) we know by the definition of ",
 Cell[BoxData[
  FormBox["reducible", TraditionalForm]]],
 "\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "2"], 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["c", "1"]}]], 
    RowBox[{
     SubscriptBox["b", "2"], "-", 
     RowBox[{
      SubscriptBox["m", "1"], "*", 
      SubscriptBox["c", "1"]}]}]}], TraditionalForm]]],
 "\n\nFrom the definition of reducibility we also know\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "1"], "=", 
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "1"], "*", 
      SubscriptBox["c", "1"]}]}]}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "2"], "=", 
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "2"], "*", 
      SubscriptBox["c", "2"]}]}]}], TraditionalForm]]],
 "\nmeaning that we actually have to show\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "1"], "*", 
      SubscriptBox["c", "1"]}]}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "C", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "2"], "*", 
      SubscriptBox["c", "2"]}]}]}], TraditionalForm]]],
 "\n\nNow we make use of (54) instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Rule]", "a"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "\[Rule]", 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[Rule]", 
    SubscriptBox["c", "2"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"d", "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      SubscriptBox["m", "1"]}], "*", 
     SubscriptBox["c", "1"]}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"z", "\[Rule]", "a"}], TraditionalForm]]],
 ". The two premises are easily seen to be satisfied (for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "1"], "*", 
      SubscriptBox["c", "1"]}]}]}], TraditionalForm]]],
 " use (5) and (1)), hence we can infer\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "1"], "*", 
      SubscriptBox["c", "1"]}]}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", 
      SubscriptBox["c", "2"], "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    RowBox[{
     SubscriptBox["b", "2"], "-", 
     RowBox[{
      SubscriptBox["m", "1"], "*", 
      SubscriptBox["c", "1"]}]}]}], TraditionalForm]]],
 "\nBy the definition of reducibility modulo sets we immediately also observe\
\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "1"], "*", 
      SubscriptBox["c", "1"]}]}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "C", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    RowBox[{
     SubscriptBox["b", "2"], "-", 
     RowBox[{
      SubscriptBox["m", "1"], "*", 
      SubscriptBox["c", "1"]}]}]}], TraditionalForm]]],
 "\n\nExploiting transitivity of ",
 Cell[BoxData[
  FormBox[
   OverscriptBox[
    SubsuperscriptBox["\[LeftRightArrow]", "C", "*"], 
    RowBox[{"a", ",", "\[Succeeds]"}]], TraditionalForm]]],
 " and using (8) we now try to show\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["b", "2"], "-", 
     RowBox[{
      SubscriptBox["m", "1"], "*", 
      SubscriptBox["c", "1"]}]}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "C", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "2"], "*", 
      SubscriptBox["c", "2"]}]}]}], TraditionalForm]]],
 "\ninstead of (G2).\nHowever, from (4) and (6) we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "2"], "*", 
      SubscriptBox["c", "2"]}]}], 
    SubscriptBox["\[Rule]", 
     SubscriptBox["c", "1"]], 
    RowBox[{
     SubscriptBox["b", "2"], "-", 
     RowBox[{
      SubscriptBox["m", "1"], "*", 
      SubscriptBox["c", "1"]}]}]}], TraditionalForm]]],
 ", which, together with the fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "2"], "*", 
      SubscriptBox["c", "2"]}]}]}], TraditionalForm]]],
 " following from (2), implies our goal (G3)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.609647835821867*^9, 3.609648085655487*^9}, {
   3.609648127663954*^9, 3.60964817823963*^9}, {3.609648262239648*^9, 
   3.609648475238229*^9}, {3.60964851295044*^9, 3.60964870707753*^9}, {
   3.609648742324834*^9, 3.609648839820485*^9}, 3.631855307870976*^9, 
   3.636348612089016*^9, 3.643533773210293*^9, {3.6442925005759373`*^9, 
   3.644292644452922*^9}, {3.645156806352811*^9, 3.645156807201799*^9}, {
   3.645158869872404*^9, 3.645158897750965*^9}, {3.645159040969371*^9, 
   3.64515920484933*^9}, 3.6532195780500927`*^9, 3.6661743986416326`*^9, 
   3.66644096756551*^9},
 CellID->439074970],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.609647827971886*^9, 3.609647835821987*^9}, {
   3.609648876313872*^9, 3.609648880155176*^9}, 3.6318553078711033`*^9, 
   3.636348612089136*^9, 3.643533773210417*^9, 3.6532195780500927`*^9, 
   3.6661743986416326`*^9, 3.6664409675656548`*^9},
 TextAlignment->Right,
 CellID->961142974]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1889858500],

Cell[CellGroupData[{

Cell["\<\
Theorem: Sufficient Conditions for Connectibility Below IV (For Establishing \
the Correctness of Buchberger\[CloseCurlyQuote]s Algorithm)\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.647940067843457*^9, 3.647940068327932*^9}, {
  3.64794027972486*^9, 3.647940299428371*^9}},
 CellID->1903482586],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b1", ",", "b2", ",", "d1", ",", "d2"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 895068823}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6479400772827787`*^9, 3.647940097582364*^9}},
 CellTags->{
  "ID:895068823", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->895068823],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "D", "]"}]], 
    RowBox[{"[", "B", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", 
        UnderscriptBox[
         SubscriptBox["\[Rule]", "B"], "D"], "b1"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", 
        UnderscriptBox[
         SubscriptBox["\[Rule]", "B"], "D"], "b2"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"b1", 
        UnderscriptBox[
         SubsuperscriptBox["\[Rule]", "B", "*"], "D"], "d1"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"b2", 
        UnderscriptBox[
         SubsuperscriptBox["\[Rule]", "B", "*"], "D"], "d2"}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b1", 
      UnderscriptBox[
       OverscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", 
         RowBox[{"B", "\[Union]", 
          RowBox[{"{", 
           RowBox[{"d1", 
            UnderscriptBox["-", "D"], "d2"}], "}"}]}], "*"], 
        RowBox[{"a", ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox["\[Succeeds]",
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}]], "D"], "b2"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(45)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:296045170", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6479401043724337`*^9, 3.647940215568287*^9}},
 CellTags->{
  "ID:296045170", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "45"},
 CellID->296045170],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "T", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", 
        UnderscriptBox[
         SubscriptBox["\[Rule]", "T"], "D"], "b1"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", 
        UnderscriptBox[
         SubscriptBox["\[Rule]", "T"], "D"], "b2"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"b1", 
        UnderscriptBox[
         SubsuperscriptBox["\[Rule]", "T", "*"], "D"], "d1"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"b2", 
        UnderscriptBox[
         SubsuperscriptBox["\[Rule]", "T", "*"], "D"], "d2"}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b1", 
      UnderscriptBox[
       OverscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", 
         RowBox[{"T", 
          TagBox["\:293a",
           Identity,
           SyntaxForm->"a*b"], 
          RowBox[{"(", 
           RowBox[{"d1", 
            UnderscriptBox["-", "D"], "d2"}], ")"}]}], "*"], 
        RowBox[{"a", ",", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          TagBox["\[Succeeds]",
           Identity], 
          TagBox[")",
           "AutoParentheses"]}]}]], "D"], "b2"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(46)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1887054887", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6479402411627398`*^9, 3.647940270756933*^9}, {
  3.648182703715199*^9, 3.648182705768465*^9}},
 CellTags->{
  "ID:1887054887", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "46"},
 CellID->1887054887],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->644092060]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->551815751],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(45)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(45)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b1", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["d1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b2", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["d2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["b1", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\[Union]", 
                   RowBox[{"{", 
                    RowBox[{
                    StyleBox["d1", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["d2", "ExpressionVariable"]}], "}"}]}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b1", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["d1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b2", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["d2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["b1", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\[Union]", 
                   RowBox[{"{", 
                    RowBox[{
                    StyleBox["d1", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["d2", "ExpressionVariable"]}], "}"}]}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:296045170"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p296045170-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p296045170-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:296045170", "Proof|ID:296045170-1"},
 CellID->9942592],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p296045170-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->736139156]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(46)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(46)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b1", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["d1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b2", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["d2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["b1", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], "\:293a", 
                   RowBox[{
                    StyleBox["d1", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["d2", "ExpressionVariable"]}]}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["d2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 StyleBox["T", "ExpressionVariable"]], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b1", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["d1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b2", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["D", "ExpressionVariable"]], 
               StyleBox["d2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["b1", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], "\:293a", 
                   RowBox[{
                    StyleBox["d1", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["d2", "ExpressionVariable"]}]}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1887054887"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1887054887-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1887054887-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1887054887", "Proof|ID:1887054887-1"},
 CellID->849984384],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1887054887-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1801153163]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->2110358369],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: Sufficient Condition for ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[LeftRightArrow]", "*"], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645436526334779*^9, 3.645436535642071*^9}, 
   3.6454370452732353`*^9},
 CellID->1022916446],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", "r"], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", 
       UnderscriptBox[
        SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", 
       UnderscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], "D"], "b"}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(59)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1876022438", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6448988290630913`*^9, 3.644898867389001*^9}, {
   3.644904051147501*^9, 3.644904052214156*^9}, {3.645436657352694*^9, 
   3.645436657684292*^9}, {3.645436763347633*^9, 3.645436771933776*^9}, {
   3.645437050722218*^9, 3.6454370587795677`*^9}, {3.645437334520628*^9, 
   3.645437341181452*^9}, 3.6454376232234707`*^9},
 CellTags->{
  "ID:1876022438", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "59"},
 CellID->1876022438],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->781070705]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->393832014],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(59)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(59)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             StyleBox["r", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[Rule]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["D", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1876022438"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1876022438-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1876022438-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1876022438", "Proof|ID:1876022438-1"},
 CellID->1166331884],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1876022438-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1596179441]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Main Theorem of Reduction Ring Theory (\[Checkmark])\
\>", "Section",
 CellChangeTimes->{{3.591602004143828*^9, 3.591602006995157*^9}, {
  3.6442143485070133`*^9, 3.644214359395328*^9}, {3.645518060896328*^9, 
  3.645518064807365*^9}, {3.650259991819902*^9, 3.650259993074657*^9}, {
  3.6503596132713003`*^9, 3.650359615630616*^9}},
 CellID->1137906058],

Cell[CellGroupData[{

Cell["Definitions", "Subsection",
 CellChangeTimes->{{3.6442179533262033`*^9, 3.6442179554561577`*^9}, {
   3.648805229152137*^9, 3.648805232778696*^9}, {3.650259987830852*^9, 
   3.650259989093857*^9}, 3.6502606842879133`*^9},
 CellID->1673340904],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"D", ",", "r"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1852246688}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6442145299701653`*^9, 3.644214535574843*^9}, {
  3.6442180908663197`*^9, 3.644218093468452*^9}, {3.6442955661626577`*^9, 
  3.64429556730438*^9}},
 CellTags->{
  "ID:1852246688", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1852246688],

Cell["", "OpenEnvironment",
 CellID->2045568059],

Cell[CellGroupData[{

Cell["Definition: Properties of the Reduction Relation", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.610172369624911*^9, 3.610172375356266*^9}, {
  3.610172408434794*^9, 3.610172408847307*^9}},
 CellID->1404801278],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "c", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isConfluent", "D"], "[", 
     RowBox[{"c", ",", "r"}], "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", 
         RowBox[{"a", ",", "b"}], "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{"c", 
            UnderscriptBox[
             SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "a"}], 
           TagBox[")",
            "AutoParentheses"]}], "\[And]", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{"c", 
            UnderscriptBox[
             SubsuperscriptBox["\[Rule]", "r", "*"], "D"], "b"}], 
           TagBox[")",
            "AutoParentheses"]}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[Implies]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"a", 
          UnderscriptBox[
           SubsuperscriptBox["\[DownArrow]", "r", "*"], "D"], "b"}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(element confluence)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1471949021", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645513911807674*^9, 3.645514009825534*^9}},
 CellTags->{
  "ID:1471949021", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "element confluence"},
 CellID->1471949021],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isConfluent", "D"], "[", "r", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], "[", 
     RowBox[{"a", ",", "b"}], "]"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       RowBox[{"a", 
        UnderscriptBox[
         SubsuperscriptBox["\[UpArrow]", "r", "*"], "D"], "b"}],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       RowBox[{"a", 
        UnderscriptBox[
         SubsuperscriptBox["\[DownArrow]", "r", "*"], "D"], "b"}],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(confluence)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:146706347", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.591433357545816*^9, 3.59143342601043*^9}, {
   3.59143353198292*^9, 3.591433546678181*^9}, {3.591433850993244*^9, 
   3.591433881192921*^9}, 3.591434224363287*^9, {3.591434719873995*^9, 
   3.591434720242349*^9}, {3.591434782971484*^9, 3.591434785014461*^9}, {
   3.591435503368801*^9, 3.591435524903166*^9}, {3.591587590324744*^9, 
   3.591587590509498*^9}, {3.591587668977072*^9, 3.591587680272536*^9}, 
   3.59402894774164*^9, {3.596187977119336*^9, 3.596187999338242*^9}, {
   3.596189130880484*^9, 3.596189132548341*^9}, {3.596189994399517*^9, 
   3.596190008602664*^9}, {3.606127774490597*^9, 3.606127779101991*^9}, {
   3.610176534263118*^9, 3.610176557601562*^9}, {3.61017663425132*^9, 
   3.610176637619417*^9}, {3.644218472009334*^9, 3.6442184750897703`*^9}, {
   3.644295579925028*^9, 3.644295581742099*^9}},
 CellTags->{
  "ID:146706347", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "confluence"},
 CellID->146706347],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isChurchRosser", "D"], "[", "r", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], "[", 
     RowBox[{"a", ",", "b"}], "]"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       RowBox[{"a", 
        UnderscriptBox[
         SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], "D"], "b"}],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       RowBox[{"a", 
        UnderscriptBox[
         SubsuperscriptBox["\[DownArrow]", "r", "*"], "D"], "b"}],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Church-Rosser property)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1279966325", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.591433231302202*^9, 3.591433324572738*^9}, 
   3.591434222927863*^9, {3.591434717110867*^9, 3.591434717514281*^9}, {
   3.59143477800944*^9, 3.591434780473057*^9}, {3.591587588411115*^9, 
   3.59158758881637*^9}, {3.591587639269515*^9, 3.591587656442193*^9}, 
   3.594028943916466*^9, {3.596187937209387*^9, 3.596187953315829*^9}, {
   3.596189125322171*^9, 3.596189127271519*^9}, {3.596189901515741*^9, 
   3.596189913988492*^9}, {3.606127765046824*^9, 3.606127769842689*^9}, {
   3.610176437638453*^9, 3.610176512795618*^9}, {3.610176630323101*^9, 
   3.610176633474662*^9}, {3.64421846711224*^9, 3.644218470450965*^9}, {
   3.644295576292799*^9, 3.644295578331234*^9}},
 CellTags->{
  "ID:1279966325", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Church-Rosser property"},
 CellID->1279966325],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isLocConnective", "D"], "[", "r", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], "[", 
     RowBox[{"a", ",", "b", ",", "c"}], "]"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox[
         RowBox[{"c", 
          UnderscriptBox[
           SubscriptBox["\[Rule]", "r"], "D"], "a"}],
         Identity], 
        TagBox[")",
         "AutoParentheses"]}], "\[And]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox[
         RowBox[{"c", 
          UnderscriptBox[
           SubscriptBox["\[Rule]", "r"], "D"], "b"}],
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       RowBox[{"a", 
        UnderscriptBox[
         OverscriptBox[
          SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
          RowBox[{"c", ",", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            TagBox["\[Succeeds]",
             Identity], 
            TagBox[")",
             "AutoParentheses"]}]}]], "D"], "b"}],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(local connectivity)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1602877499", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.591596742718*^9, 3.594028978811879*^9, {3.596188255986093*^9, 
   3.59618829654337*^9}, {3.596189196998228*^9, 3.596189202860971*^9}, {
   3.596190061619916*^9, 3.596190076148221*^9}, {3.606127861175142*^9, 
   3.606127865786034*^9}, {3.610176667782086*^9, 3.610176753547736*^9}, {
   3.644218481296657*^9, 3.6442184858331127`*^9}, {3.644295586823575*^9, 
   3.644295590938929*^9}},
 CellTags->{
  "ID:1602877499", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "local connectivity"},
 CellID->1602877499],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isGroebnerBasis", "D"], "[", "r", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["isChurchRosser", "D"], "[", "r", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Gr\[ODoubleDot]bner basis)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:661413373", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.592024072865074*^9, 3.592024127691785*^9}, {
  3.592024684637009*^9, 3.592024692042341*^9}, {3.592025246503117*^9, 
  3.592025247317717*^9}, {3.596187504014033*^9, 3.59618753737037*^9}, {
  3.596190978280635*^9, 3.596190980860082*^9}, {3.596191033033851*^9, 
  3.596191034649604*^9}, {3.607337870721032*^9, 3.607337871680667*^9}, {
  3.610177260762083*^9, 3.610177263903635*^9}, {3.6442199895422897`*^9, 
  3.644220025045801*^9}, {3.644295592639316*^9, 3.64429559356721*^9}},
 CellTags->{
  "ID:661413373", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Gr\[ODoubleDot]bner basis"},
 CellID->661413373],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->408387442]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->389882308],

Cell["", "OpenEnvironment",
 CellID->2109808855],

Cell[CellGroupData[{

Cell["Definition: More Notions in Reduction Rings", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.592024199466174*^9, 3.592024207370285*^9}, {
  3.610179354521733*^9, 3.610179371246853*^9}},
 CellID->1168412894],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "c1", ",", "c2", ",", "m1", ",", "m2"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1444563707}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.592024476880232*^9, 3.592024484138981*^9}, 
   3.59403572233791*^9, {3.608525684343451*^9, 3.608525684676693*^9}, {
   3.60906793212376*^9, 3.609067932309537*^9}, {3.610177115397907*^9, 
   3.610177134489695*^9}},
 CellTags->{
  "ID:1444563707", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1444563707],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areConnectibleMultipliers", "D"], "[", 
   RowBox[{"m1", ",", "m2", ",", "c1", ",", "c2", ",", "a", ",", "r"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["let", 
     RowBox[{
      RowBox[{"b1", "=", 
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{"m1", 
         UnderscriptBox["*", "D"], "c1"}]}]}], ",", 
      RowBox[{"b2", "=", 
       RowBox[{"a", 
        UnderscriptBox["-", "D"], 
        RowBox[{"m2", 
         UnderscriptBox["*", "D"], "c2"}]}]}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"a", 
           UnderscriptBox[
            SubscriptBox["\[Rule]", 
             RowBox[{"m1", ",", "c1"}]], "D"], "b1"}]},
         {
          RowBox[{"a", 
           UnderscriptBox[
            SubscriptBox["\[Rule]", 
             RowBox[{"m2", ",", "c2"}]], "D"], "b2"}]},
         {
          RowBox[{"b1", 
           UnderscriptBox[
            OverscriptBox[
             SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
             RowBox[{"a", ",", 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               TagBox["\[Succeeds]",
                Identity], 
               TagBox[")",
                "AutoParentheses"]}]}]], "D"], "b2"}]}
        }]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areConnectibleMultipliers)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1408266583", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.60906788928271*^9, 3.609067918783076*^9}, {
  3.609067981165869*^9, 3.609068181918544*^9}, {3.609068238296781*^9, 
  3.609068239134231*^9}, {3.610176983271394*^9, 3.610177015902569*^9}, {
  3.610177094693372*^9, 3.610177097420172*^9}, {3.610177140376156*^9, 
  3.610177141967173*^9}, {3.610179476646829*^9, 3.610179477763134*^9}, {
  3.6442145579536333`*^9, 3.644214563367956*^9}, {3.644295597833187*^9, 
  3.64429559946411*^9}},
 CellTags->{
  "ID:1408266583", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areConnectibleMultipliers"},
 CellID->1408266583],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", "j"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["cpConnectible", "D"], "[", 
     RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j", ",", "r"}], "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", 
         RowBox[{"m3", ",", "m4"}], "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             UnderscriptBox["areCPMultipliers", "D"], "[", 
             RowBox[{
             "m3", ",", "m4", ",", "c1", ",", "c2", ",", "a", ",", "i", ",", 
              "j"}], "]"}]},
           {
            RowBox[{
             UnderscriptBox["areConnectibleMultipliers", "D"], "[", 
             RowBox[{
             "m3", ",", "m4", ",", "c1", ",", "c2", ",", "a", ",", "r"}], 
             "]"}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(connectibility of one critical pair)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:882680545", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.609068413667619*^9, 3.609068522827184*^9}, {
   3.609068554310444*^9, 3.609068610286603*^9}, {3.609068649054083*^9, 
   3.609068653918004*^9}, {3.610177150621485*^9, 3.610177178658559*^9}, {
   3.631855271751843*^9, 3.6318552753433847`*^9}, {3.631856409762035*^9, 
   3.631856413997561*^9}, {3.632548932265801*^9, 3.6325489418884363`*^9}, {
   3.644214566740644*^9, 3.644214570707513*^9}, {3.644295601610806*^9, 
   3.644295604054557*^9}, {3.645181035849497*^9, 3.645181037286463*^9}, {
   3.645181322735005*^9, 3.645181325974093*^9}, 3.6488021285989122`*^9},
 CellTags->{
  "ID:882680545", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of one critical pair"},
 CellID->882680545],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "D", "]"}]], 
    RowBox[{"[", "B", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["cpConnectible", "D"], "[", "B", "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[Implies]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"x", ",", 
         RowBox[{"y", "\[Element]", "B"}]}]], 
       RowBox[{
        RowBox[{"x", "\[NotEqual]", 
         UnderscriptBox["0", "D"]}], "\[And]", 
        RowBox[{"y", "\[NotEqual]", 
         UnderscriptBox["0", "D"]}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox[
         UnderscriptBox["\[ForAll]", 
          RowBox[{"i", ",", 
           RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{
            UnderscriptBox["multN", "D"], "[", "x", "]"}]}]], 
         RowBox[{
          RowBox[{"x", "\[Equal]", "y"}], "\[Implies]", 
          RowBox[{"i", "\[NotEqual]", "j"}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          UnderscriptBox[
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "D"], 
             RowBox[{"[", "d", "]"}]}]], 
           RowBox[{
            UnderscriptBox["isMNTCR", "D"], "[", 
            RowBox[{"x", ",", "y", ",", "d", ",", "i", ",", "j"}], "]"}]], 
          RowBox[{
           UnderscriptBox["cpConnectible", "D"], "[", 
           RowBox[{"x", ",", "y", ",", "d", ",", "i", ",", "j", ",", "B"}], 
           "]"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(connectibility of critical pairs)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:259042764", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.592025046597735*^9, 3.592025169558421*^9}, {
   3.59202523719333*^9, 3.592025238001401*^9}, {3.596190856039141*^9, 
   3.596190890528398*^9}, {3.596190924012668*^9, 3.596190946752343*^9}, {
   3.596191026812148*^9, 3.59619102880305*^9}, {3.596191098208039*^9, 
   3.596191101309168*^9}, {3.607337848860267*^9, 3.607337859101022*^9}, {
   3.608525016063375*^9, 3.608525109031864*^9}, {3.608607879557657*^9, 
   3.608607965401008*^9}, {3.608608118685263*^9, 3.608608120309924*^9}, {
   3.60860816459955*^9, 3.608608335188912*^9}, {3.60860837935586*^9, 
   3.608608466705196*^9}, {3.608608628076528*^9, 3.608608684488612*^9}, {
   3.608609252603698*^9, 3.608609350818276*^9}, {3.609068326667818*^9, 
   3.609068350567298*^9}, {3.609068675384989*^9, 3.60906870161668*^9}, {
   3.610177193310797*^9, 3.610177200690529*^9}, {3.6318552874743643`*^9, 
   3.631855290769998*^9}, {3.631856420471076*^9, 3.631856426139344*^9}, {
   3.6324790158827667`*^9, 3.6324790186597843`*^9}, {3.6325468166121387`*^9, 
   3.632546837294722*^9}, 3.644214572642909*^9, {3.6442180201702538`*^9, 
   3.644218047243104*^9}, {3.644295606096505*^9, 3.644295662790238*^9}, {
   3.645178117404903*^9, 3.645178130273014*^9}},
 CellTags->{
  "ID:259042764", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs"},
 CellID->259042764],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "T", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["cpConnectible", "D"], "[", "T", "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     UnderscriptBox["cpConnectible", "D"], "[", 
     RowBox[{"TupleToSet", "[", "T", "]"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(connectibility of critical pairs of tuple)", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1836116785", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648805876278211*^9, 3.648805923032672*^9}},
 CellTags->{
  "ID:1836116785", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs of tuple"},
 CellID->1836116785],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpConnectible", "D"], "[", 
   RowBox[{"r", ",", "c1", ",", "c2"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", ",", 
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       UnderscriptBox["multN", "D"], "[", "c1", "]"}]}]], 
    RowBox[{
     RowBox[{"c1", "\[Equal]", "c2"}], "\[Implies]", 
     RowBox[{"i", "\[NotEqual]", "j"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", "d", "]"}]}]], 
      RowBox[{
       UnderscriptBox["isMNTCR", "D"], "[", 
       RowBox[{"c1", ",", "c2", ",", "d", ",", "i", ",", "j"}], "]"}]], 
     RowBox[{
      UnderscriptBox["cpConnectible", "D"], "[", 
      RowBox[{"c1", ",", "c2", ",", "d", ",", "i", ",", "j", ",", "r"}], 
      "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(connectibility of critical pairs of elements)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2066466981", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648782313684758*^9, 3.648782349487788*^9}, {
  3.648782383060154*^9, 3.648782384135725*^9}, {3.6487826110033083`*^9, 
  3.648782633153778*^9}},
 CellTags->{
  "ID:2066466981", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs of elements"},
 CellID->2066466981],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isLocConnective", "D"], "[", 
   RowBox[{"r", ",", "c1", ",", "c2"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", 
      RowBox[{"d", ",", "b1", ",", "b2"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox[
         RowBox[{"d", 
          UnderscriptBox[
           SubscriptBox["\[Rule]", 
            RowBox[{"{", "c1", "}"}]], "D"], "b1"}],
         Identity], 
        TagBox[")",
         "AutoParentheses"]}], "\[And]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox[
         RowBox[{"d", 
          UnderscriptBox[
           SubscriptBox["\[Rule]", 
            RowBox[{"{", "c2", "}"}]], "D"], "b2"}],
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox[
       RowBox[{"b1", 
        UnderscriptBox[
         OverscriptBox[
          SubsuperscriptBox["\[LeftRightArrow]", "r", "*"], 
          RowBox[{"d", ",", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            TagBox["\[Succeeds]",
             Identity], 
            TagBox[")",
             "AutoParentheses"]}]}]], "D"], "b2"}],
       Identity], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(local connectivity of elements)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:756678784", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648781932600567*^9, 3.64878201101912*^9}, {
  3.6487821346416693`*^9, 3.6487821374921637`*^9}, {3.648782569544374*^9, 
  3.6487825996233873`*^9}, {3.6618510400301275`*^9, 3.66185104726854*^9}},
 CellTags->{
  "ID:756678784", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "local connectivity of elements"},
 CellID->756678784],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->322598144]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->191727381]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Lemmas: cpConnectible & isLocConnective (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.648790349487193*^9, 3.648790350310248*^9}, {
  3.648791567357521*^9, 3.6487915749734297`*^9}, {3.6488052146750097`*^9, 
  3.6488052181346483`*^9}, {3.650259980924477*^9, 3.6502599823832207`*^9}, {
  3.650359606537918*^9, 3.6503596090328827`*^9}},
 CellID->371498323],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"D", ",", "r"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1173678670}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650260676051236*^9, 3.650260677850112*^9}},
 CellTags->{
  "ID:1173678670", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1173678670],

Cell["", "OpenEnvironment",
 CellID->358588877],

Cell[CellGroupData[{

Cell["Lemma: cpConnectible of Different Elements", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649070328736792*^9, 3.649070336656847*^9}},
 CellID->284105849],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", 
      RowBox[{"c1", ",", "c2", ",", "a"}], "]"}]}]], 
   RowBox[{"c1", "\[NotEqual]", "c2"}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", "j"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["cpConnectible", "D"], "[", 
      RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j", ",", "r"}], 
      "]"}], "\[DoubleLongLeftRightArrow]", 
     RowBox[{
      UnderscriptBox["cpConnectible", "D"], "[", 
      RowBox[{"c1", ",", "c2", ",", "a", ",", "1", ",", "1", ",", "r"}], 
      "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1524905762", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649070343375401*^9, 3.649070385263661*^9}},
 CellTags->{
  "ID:1524905762", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible 1"},
 CellID->1524905762],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1031134990]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1892120190],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["r", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 StyleBox["r", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1", ",", 
                 StyleBox["r", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             StyleBox["r", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 StyleBox["r", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1", ",", 
                 StyleBox["r", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1524905762"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1524905762-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1524905762-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1524905762", "Proof|ID:1524905762-1"},
 CellID->709826783],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1524905762-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1762868189]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1197106890],

Cell[CellGroupData[{

Cell["Lemma: cpConnectible and isLocConnective for Tuples", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648804598531139*^9, 3.648804599828422*^9}},
 CellID->2064352251],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"c1", ",", "c2"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 2054204196}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648804647322356*^9, 3.648804654837875*^9}},
 CellTags->{
  "ID:2054204196", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2054204196],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "D", "]"}]], 
   RowBox[{"[", "T", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 194336473}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648791051490798*^9, 3.6487910617553988`*^9}},
 CellTags->{
  "ID:194336473", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->194336473],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpConnectible", "D"], "[", 
   RowBox[{"T", ",", "c1", ",", "c2"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  
  RowBox[{
   UnderscriptBox["cpConnectible", "D"], "[", 
   RowBox[{
    RowBox[{"TupleToSet", "[", "T", "]"}], ",", "c1", ",", "c2"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible tuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2064859128", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648791068011293*^9, 3.6487911173227987`*^9}},
 CellTags->{
  "ID:2064859128", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible tuple"},
 CellID->2064859128],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isLocConnective", "D"], "[", 
   RowBox[{"T", ",", "c1", ",", "c2"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  
  RowBox[{
   UnderscriptBox["isLocConnective", "D"], "[", 
   RowBox[{
    RowBox[{"TupleToSet", "[", "T", "]"}], ",", "c1", ",", "c2"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isLocConnective tuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1276073437", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648791114112496*^9, 3.64879113273061*^9}},
 CellTags->{
  "ID:1276073437", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective tuple"},
 CellID->1276073437],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isLocConnective", "D"], "[", "T", "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isLocConnective", "D"], "[", 
   RowBox[{"TupleToSet", "[", "T", "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isLocConnective tuple II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1869404165", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648804484008122*^9, 3.648804487694859*^9}},
 CellTags->{
  "ID:1869404165", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective tuple II"},
 CellID->1869404165],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->729430212]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->899496562],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible tuple)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible tuple)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{"TupleToSet", "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{"TupleToSet", "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2064859128"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2064859128-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2064859128-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2064859128", "Proof|ID:2064859128-1"},
 CellID->1478532560],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2064859128-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1511485459]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isLocConnective tuple)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isLocConnective tuple)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{"TupleToSet", "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               RowBox[{"TupleToSet", "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1276073437"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1276073437-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1276073437-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1276073437", "Proof|ID:1276073437-1"},
 CellID->253023626],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1276073437-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->101762537]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isLocConnective tuple II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isLocConnective tuple II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             StyleBox["T", "ExpressionVariable"], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{"TupleToSet", "[", 
               StyleBox["T", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             StyleBox["T", "ExpressionVariable"], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{"TupleToSet", "[", 
               StyleBox["T", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1869404165"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1869404165-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1869404165-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1869404165", "Proof|ID:1869404165-1"},
 CellID->272805702],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1869404165-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1782684890]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->438519018],

Cell[CellGroupData[{

Cell["\<\
Lemma: Local Connectivity of All Elements is Equivalent to Local Connectivity\
\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648791179055971*^9, 3.648791216462085*^9}},
 CellID->1810810165],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "D", "]"}]], 
   RowBox[{"[", "B", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 2083446865}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648803916164104*^9, 3.6488039286819267`*^9}},
 CellTags->{
  "ID:2083446865", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2083446865],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpConnectible", "D"], "[", "B", "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{"a", ",", 
      RowBox[{"b", "\[Element]", "B"}]}]], 
    RowBox[{
     RowBox[{"a", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}], "\[And]", 
     RowBox[{"b", "\[NotEqual]", 
      UnderscriptBox["0", "D"]}]}]], 
   RowBox[{
    UnderscriptBox["cpConnectible", "D"], "[", 
    RowBox[{"B", ",", "a", ",", "b"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:709723813", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648791348103094*^9, 3.648791381839835*^9}, {
  3.648803471494094*^9, 3.6488034741918077`*^9}, {3.648803585187125*^9, 
  3.648803600634862*^9}, {3.648803667878487*^9, 3.6488036702428923`*^9}, {
  3.64880393203168*^9, 3.648803934732864*^9}},
 CellTags->{
  "ID:709723813", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative"},
 CellID->709723813],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isLocConnective", "D"], "[", "B", "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"a", ",", 
     RowBox[{"b", "\[Element]", "B"}]}]], 
   RowBox[{
    UnderscriptBox["isLocConnective", "D"], "[", 
    RowBox[{"B", ",", "a", ",", "b"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isLocConnective alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:41201803", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648791219963131*^9, 3.648791250754607*^9}, {
  3.6488034764101963`*^9, 3.6488034794087067`*^9}, {3.648803937213088*^9, 
  3.64880394595401*^9}, {3.648805988961985*^9, 3.648805999948864*^9}, {
  3.6488060814347353`*^9, 3.648806081659253*^9}, {3.648806456568242*^9, 
  3.6488065281374617`*^9}},
 CellTags->{
  "ID:41201803", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective alternative"},
 CellID->41201803],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"isReductionRing", "[", "D", "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"a", ",", 
        RowBox[{"b", "\[Element]", "B"}]}]], 
      RowBox[{
       RowBox[{"a", "\[NotEqual]", 
        UnderscriptBox["0", "D"]}], "\[And]", 
       RowBox[{"b", "\[NotEqual]", 
        UnderscriptBox["0", "D"]}]}]], 
     RowBox[{
      UnderscriptBox["isLocConnective", "D"], "[", 
      RowBox[{"B", ",", "a", ",", "b"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isLocConnective", "D"], "[", "B", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isLocConnective sufficient)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1023292110", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648806536869594*^9, 3.648806569390567*^9}},
 CellTags->{
  "ID:1023292110", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective sufficient"},
 CellID->1023292110],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->770388948]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->189312483],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}, {
                  RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}]}]], 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}, {
                  RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                  UnderscriptBox["0", 
                   StyleBox["D", "ExpressionVariable"]]}]}]], 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:709723813"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p709723813-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p709723813-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:709723813", "Proof|ID:709723813-1"},
 CellID->1570886920],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p709723813-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1569878545]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isLocConnective alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isLocConnective alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Element]", 
                   StyleBox["B", "ExpressionVariable"]}]}, {
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Element]", 
                   StyleBox["B", "ExpressionVariable"]}]}}]], 
             RowBox[{
               UnderscriptBox["isLocConnective", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Element]", 
                   StyleBox["B", "ExpressionVariable"]}]}, {
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Element]", 
                   StyleBox["B", "ExpressionVariable"]}]}}]], 
             RowBox[{
               UnderscriptBox["isLocConnective", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:41201803"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p41201803-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p41201803-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:41201803", "Proof|ID:41201803-1"},
 CellID->172260705],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p41201803-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->210571619]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isLocConnective sufficient)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isLocConnective sufficient)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}, {
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                  RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]}]], 
               RowBox[{
                 UnderscriptBox["isLocConnective", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}, {
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                  RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]}]], 
               RowBox[{
                 UnderscriptBox["isLocConnective", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1023292110"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1023292110-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1023292110-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1023292110", "Proof|ID:1023292110-1"},
 CellID->141981476],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1023292110-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->144280024]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1610599240],

Cell[CellGroupData[{

Cell["Lemma: cpConnectible of Enlarged Tuples", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648891096020791*^9, 3.6488911050734262`*^9}},
 CellID->954877052],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "T", "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 529333522}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.648891217229642*^9, {3.6488916413088303`*^9, 3.6488916416394043`*^9}},
 CellTags->{
  "ID:529333522", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->529333522],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "a", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", "j"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["cpConnectible", "D"], "[", 
      RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j", ",", "T"}], 
      "]"}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"DomainTuples", "[", "D", "]"}]], 
         RowBox[{"[", "S", "]"}]}]], 
       RowBox[{
        UnderscriptBox["cpConnectible", "D"], "[", 
        RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j", ",", 
         RowBox[{"T", 
          TagBox["\:22c8",
           Identity,
           SyntaxForm->"a*b"], "S"}]}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible join)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:208412986", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648894971402445*^9, 3.648894989224463*^9}},
 CellTags->{
  "ID:208412986", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible join"},
 CellID->208412986],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "a", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", "j"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["cpConnectible", "D"], "[", 
      RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j", ",", "T"}], 
      "]"}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "D"], 
         RowBox[{"[", "x", "]"}]}]], 
       RowBox[{
        UnderscriptBox["cpConnectible", "D"], "[", 
        RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j", ",", 
         RowBox[{"T", 
          TagBox["\:293a",
           Identity,
           SyntaxForm->"a*b"], "x"}]}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible append)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1183553358", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648891455963751*^9, 3.648891496123365*^9}, {
  3.64889227299264*^9, 3.6488922752854757`*^9}, {3.64889230943788*^9, 
  3.648892318511722*^9}},
 CellTags->{
  "ID:1183553358", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible append"},
 CellID->1183553358],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpConnectible", "D"], "[", 
   RowBox[{"T", ",", "c1", ",", "c2"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "D", "]"}]], 
      RowBox[{"[", "S", "]"}]}]], 
    RowBox[{
     UnderscriptBox["cpConnectible", "D"], "[", 
     RowBox[{
      RowBox[{"T", 
       TagBox["\:22c8",
        Identity,
        SyntaxForm->"a*b"], "S"}], ",", "c1", ",", "c2"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible join II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:286665714", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64889502416144*^9, 3.6488950391064568`*^9}},
 CellTags->{
  "ID:286665714", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible join II"},
 CellID->286665714],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpConnectible", "D"], "[", 
   RowBox[{"T", ",", "c1", ",", "c2"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", "x", "]"}]}]], 
    RowBox[{
     UnderscriptBox["cpConnectible", "D"], "[", 
     RowBox[{
      RowBox[{"T", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "x"}], ",", "c1", ",", "c2"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible append II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1065812089", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64889110967642*^9, 3.6488911644281807`*^9}, 
   3.6488912156114483`*^9, {3.6488914471932793`*^9, 3.648891452967574*^9}},
 CellTags->{
  "ID:1065812089", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible append II"},
 CellID->1065812089],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->305228439]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1771612334],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible join)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible join)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], "\:22c8", 
                   StyleBox["S", "ExpressionVariable"]}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], "\:22c8", 
                   StyleBox["S", "ExpressionVariable"]}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:208412986"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p208412986-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p208412986-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:208412986", "Proof|ID:208412986-1"},
 CellID->511870621],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p208412986-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1028716509]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible append)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible append)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], "\:293a", 
                   StyleBox["x", "ExpressionVariable"]}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], "\:293a", 
                   StyleBox["x", "ExpressionVariable"]}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1183553358"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1183553358-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1183553358-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1183553358", "Proof|ID:1183553358-1"},
 CellID->1244254393],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1183553358-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.648898451630909*^9},
 CellID->1827885036]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible join II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible join II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], "\:22c8", 
                   StyleBox["S", "ExpressionVariable"]}], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], "\:22c8", 
                   StyleBox["S", "ExpressionVariable"]}], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:286665714"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p286665714-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p286665714-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:286665714", "Proof|ID:286665714-1"},
 CellID->2147068449],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p286665714-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2096779848]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible append II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible append II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], "\:293a", 
                   StyleBox["x", "ExpressionVariable"]}], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["D", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["D", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], "\:293a", 
                   StyleBox["x", "ExpressionVariable"]}], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1065812089"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1065812089-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1065812089-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1065812089", "Proof|ID:1065812089-1"},
 CellID->1920791119],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1065812089-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6488986942744513`*^9},
 CellID->1840123044]
}, Closed]],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"isReductionRing", "[", "D", "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["isReductor", "D"], "[", "r", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 751627229}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6487976939397573`*^9, 3.64879769905451*^9}, {
  3.6487977394738607`*^9, 3.648797745994629*^9}},
 CellTags->{
  "ID:751627229", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->751627229],

Cell["", "OpenEnvironment",
 CellID->908816666],

Cell[CellGroupData[{

Cell["Lemma: cpConnectible and \[Tilde]", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.650259805289527*^9, 3.650259814748316*^9}},
 CellID->1952338061],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b", ",", "c1", ",", "c2"}], "]"}]}]], 
  RowBox[{"a", 
   UnderscriptBox["\[Tilde]", "D"], "b"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 351093647}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.650356937809148*^9},
 CellTags->{
  "ID:351093647", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->351093647],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
    UnderscriptBox["multN", "D"], "[", "c1", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1476531812}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65035738451655*^9, 3.650357384631598*^9}, 
   3.650357419457521*^9},
 CellTags->{
  "ID:1476531812", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1476531812],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpConnectible", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j", ",", "r"}], "]"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["cpConnectible", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "b", ",", "i", ",", "j", ",", "r"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(cpConnectible \[Tilde] \[DoubleRightArrow])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1047637483", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6503574497378693`*^9, 3.650357454435671*^9}},
 CellTags->{
  "ID:1047637483", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible \[Tilde] \[DoubleRightArrow]"},
 CellID->1047637483],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpConnectible", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j", ",", "r"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["cpConnectible", "D"], "[", 
   RowBox[{"c1", ",", "c2", ",", "b", ",", "i", ",", "j", ",", "r"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible \[Tilde])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1237319003", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650259833775517*^9, 3.650259927075177*^9}, {
  3.650260275055654*^9, 3.650260275421515*^9}, {3.650356936444659*^9, 
  3.650356947090377*^9}, {3.650357421620599*^9, 3.650357424914351*^9}},
 CellTags->{
  "ID:1237319003", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible \[Tilde]"},
 CellID->1237319003],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->251214618]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->610940404],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible \[Tilde] \[DoubleRightArrow])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible \[Tilde] \[DoubleRightArrow])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
              RowBox[{
                StyleBox["a", "ExpressionVariable"], 
                UnderscriptBox["\[Tilde]", 
                 StyleBox["D", "ExpressionVariable"]], 
                StyleBox["b", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["r", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["r", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
              RowBox[{
                StyleBox["a", "ExpressionVariable"], 
                UnderscriptBox["\[Tilde]", 
                 StyleBox["D", "ExpressionVariable"]], 
                StyleBox["b", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["r", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["r", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1047637483"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1047637483-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1047637483-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1047637483", "Proof|ID:1047637483-1"},
 CellID->1363695772],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1047637483-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->883221224]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible \[Tilde])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible \[Tilde])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
              RowBox[{
                StyleBox["a", "ExpressionVariable"], 
                UnderscriptBox["\[Tilde]", 
                 StyleBox["D", "ExpressionVariable"]], 
                StyleBox["b", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["r", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["r", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
              RowBox[{
                StyleBox["a", "ExpressionVariable"], 
                UnderscriptBox["\[Tilde]", 
                 StyleBox["D", "ExpressionVariable"]], 
                StyleBox["b", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["r", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["r", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1237319003"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1237319003-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1237319003-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1237319003", "Proof|ID:1237319003-1"},
 CellID->1084461300],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1237319003-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->850441382]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1496431653],

Cell[CellGroupData[{

Cell["Lemma: Alternative Descriptions of cpConnectible", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6487904383422956`*^9, 3.648790489934787*^9}},
 CellID->204540969],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "c", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["cpConnectible", "D"], "[", 
     RowBox[{"r", ",", "c", ",", "c"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{
         UnderscriptBox["multN", "D"], "[", "c", "]"}]}]], 
      RowBox[{"j", ">", "i"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          UnderscriptBox["\[Element]", "D"], 
          RowBox[{"[", "d", "]"}]}]], 
        RowBox[{
         UnderscriptBox["isMNTCR", "D"], "[", 
         RowBox[{"c", ",", "c", ",", "d", ",", "i", ",", "j"}], "]"}]], 
       RowBox[{
        UnderscriptBox["cpConnectible", "D"], "[", 
        RowBox[{"c", ",", "c", ",", "d", ",", "i", ",", "j", ",", "r"}], 
        "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible alternative I)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1087064485", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648790501385055*^9, 3.648790554644127*^9}, {
  3.64879108783887*^9, 3.648791088136746*^9}, {3.6487939933349648`*^9, 
  3.648794006278946*^9}, {3.64879404102663*^9, 3.6487940486444263`*^9}, {
  3.648797700794963*^9, 3.648797704510203*^9}, {3.6487977429190397`*^9, 
  3.64879774381612*^9}},
 CellTags->{
  "ID:1087064485", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative I"},
 CellID->1087064485],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", "c", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["cpConnectible", "D"], "[", 
     RowBox[{"r", ",", "c", ",", "c"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{
         UnderscriptBox["multN", "D"], "[", "c", "]"}]}]], 
      RowBox[{"j", ">", "i"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          UnderscriptBox["\[Element]", "D"], 
          RowBox[{"[", "a", "]"}]}]], 
        RowBox[{
         UnderscriptBox["isMNTCR", "D"], "[", 
         RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{
           UnderscriptBox["\[Element]", "D"], 
           RowBox[{"[", "b", "]"}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           RowBox[{"a", 
            UnderscriptBox["\[Tilde]", "D"], "b"}], "\[And]", 
           RowBox[{
            UnderscriptBox["cpConnectible", "D"], "[", 
            RowBox[{"c", ",", "c", ",", "b", ",", "i", ",", "j", ",", "r"}], 
            "]"}]}], 
          TagBox[")",
           "AutoParentheses"]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible alternative I')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1334373255", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650260011239624*^9, 3.6502600639282093`*^9}, {
  3.650266535721364*^9, 3.650266536106976*^9}},
 CellTags->{
  "ID:1334373255", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative I'"},
 CellID->1334373255],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", 
      RowBox[{"c1", ",", "c2"}], "]"}]}]], 
   RowBox[{"c1", "\[NotEqual]", "c2"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["cpConnectible", "D"], "[", 
     RowBox[{"r", ",", "c1", ",", "c2"}], "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", "d", "]"}]}]], 
      RowBox[{
       UnderscriptBox["isMNTCR", "D"], "[", 
       RowBox[{"c1", ",", "c2", ",", "d", ",", "1", ",", "1"}], "]"}]], 
     RowBox[{
      UnderscriptBox["cpConnectible", "D"], "[", 
      RowBox[{"c1", ",", "c2", ",", "d", ",", "1", ",", "1", ",", "r"}], 
      "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible alternative II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:581920350", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648790568906999*^9, 3.6487905970339003`*^9}, {
  3.648791090680315*^9, 3.648791091010398*^9}},
 CellTags->{
  "ID:581920350", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative II"},
 CellID->581920350],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "D"], 
     RowBox[{"[", 
      RowBox[{"c1", ",", "c2"}], "]"}]}]], 
   RowBox[{"c1", "\[NotEqual]", "c2"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["cpConnectible", "D"], "[", 
     RowBox[{"r", ",", "c1", ",", "c2"}], "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", "a", "]"}]}]], 
      RowBox[{
       UnderscriptBox["isMNTCR", "D"], "[", 
       RowBox[{"c1", ",", "c2", ",", "a", ",", "1", ",", "1"}], "]"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "D"], 
         RowBox[{"[", "b", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{"a", 
          UnderscriptBox["\[Tilde]", "D"], "b"}], "\[And]", 
         RowBox[{
          UnderscriptBox["cpConnectible", "D"], "[", 
          RowBox[{"c1", ",", "c2", ",", "b", ",", "1", ",", "1", ",", "r"}], 
          "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible alternative II')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:286693183", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650260158542424*^9, 3.650260196907823*^9}, {
  3.6502665380775747`*^9, 3.650266538298091*^9}},
 CellTags->{
  "ID:286693183", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative II'"},
 CellID->286693183],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1877570189]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->670066157],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible alternative I)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible alternative I)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], ">", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["d", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]], 
                   RowBox[{
                    UnderscriptBox["cpConnectible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["d", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["r", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], ">", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["d", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]], 
                   RowBox[{
                    UnderscriptBox["cpConnectible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["d", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["r", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1087064485"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1087064485-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1087064485-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1087064485", "Proof|ID:1087064485-1"},
 CellID->14160893],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1087064485-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->143484188]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible alternative I')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible alternative I')\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], ">", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]], 
                   RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Tilde]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["cpConnectible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["r", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], ">", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]], 
                   RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Tilde]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["cpConnectible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["r", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1334373255"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1334373255-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1334373255-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1334373255", "Proof|ID:1334373255-1"},
 CellID->302186508],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1334373255-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1791932740]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible alternative II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible alternative II)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                StyleBox["c2", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["d", "ExpressionVariable"], ",", "1", ",", "1"}],
                     "]"}]], 
                 RowBox[{
                   UnderscriptBox["cpConnectible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["d", "ExpressionVariable"], ",", "1", ",", "1", 
                    ",", 
                    StyleBox["r", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                StyleBox["c2", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["d", "ExpressionVariable"], ",", "1", ",", "1"}],
                     "]"}]], 
                 RowBox[{
                   UnderscriptBox["cpConnectible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["d", "ExpressionVariable"], ",", "1", ",", "1", 
                    ",", 
                    StyleBox["r", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:581920350"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p581920350-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p581920350-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:581920350", "Proof|ID:581920350-1"},
 CellID->448065758],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p581920350-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->380059730]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible alternative II')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible alternative II')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                StyleBox["c2", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}],
                     "]"}]], 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Tilde]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["cpConnectible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", "1", ",", "1", 
                    ",", 
                    StyleBox["r", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                StyleBox["c2", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], "]"}], 
               "\[DoubleLeftRightArrow]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}],
                     "]"}]], 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Tilde]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["cpConnectible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", "1", ",", "1", 
                    ",", 
                    StyleBox["r", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:286693183"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p286693183-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p286693183-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:286693183", "Proof|ID:286693183-1"},
 CellID->514361483],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p286693183-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->455398715]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->138810160],

Cell[CellGroupData[{

Cell["Lemma: Symmetry of cpConnectible and isLocConnective", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648790945891432*^9, 3.6487909579956827`*^9}, {
  3.6487910378851633`*^9, 3.6487910480965*^9}, {3.6488045954648457`*^9, 
  3.648804595878565*^9}},
 CellID->1579042795],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", 
    RowBox[{"c1", ",", "c2"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 173719369}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6487909615283403`*^9, 3.6487909693438396`*^9}},
 CellTags->{
  "ID:173719369", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->173719369],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpConnectible", "D"], "[", 
   RowBox[{"r", ",", "c1", ",", "c2"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["cpConnectible", "D"], "[", 
   RowBox[{"r", ",", "c2", ",", "c1"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible symmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:762511022", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64879097363901*^9, 3.648790984275873*^9}, {
  3.6487910936617823`*^9, 3.648791096213711*^9}},
 CellTags->{
  "ID:762511022", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible symmetric"},
 CellID->762511022],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isLocConnective", "D"], "[", 
   RowBox[{"r", ",", "c1", ",", "c2"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isLocConnective", "D"], "[", 
   RowBox[{"r", ",", "c2", ",", "c1"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isLocConnective symmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1792796192", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648790991386894*^9, 3.6487909985553513`*^9}, {
  3.648791098610808*^9, 3.648791101053636*^9}},
 CellTags->{
  "ID:1792796192", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective symmetric"},
 CellID->1792796192],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1321892992]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->349770772],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible symmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible symmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:762511022"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p762511022-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p762511022-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:762511022", "Proof|ID:762511022-1"},
 CellID->262270078],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p762511022-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->774114444]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isLocConnective symmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isLocConnective symmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isLocConnective", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["isLocConnective", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isLocConnective", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], "]"}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["isLocConnective", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1792796192"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1792796192-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1792796192-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1792796192", "Proof|ID:1792796192-1"},
 CellID->13401702],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1792796192-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->63508157]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1721706701],

Cell[CellGroupData[{

Cell["Lemma: cpConnectible of 0", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648974572405757*^9, 3.6489745795595627`*^9}},
 CellID->1656613612],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "D"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"c1", "\[Equal]", 
       UnderscriptBox["0", "D"]}], "\[Or]", 
      RowBox[{"c2", "\[Equal]", 
       UnderscriptBox["0", "D"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["cpConnectible", "D"], "[", 
     RowBox[{"r", ",", "c1", ",", "c2"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:320557773", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648974603837924*^9, 3.648974637933853*^9}},
 CellTags->{
  "ID:320557773", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible 0"},
 CellID->320557773],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1549243361]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->541515006],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], "\[Equal]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[Or]", 
                 RowBox[{
                   StyleBox["c2", "ExpressionVariable"], "\[Equal]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["D", "ExpressionVariable"], ",", 
            StyleBox["r", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"isReductionRing", "[", 
               StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["D", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                   StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], "\[Equal]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[Or]", 
                 RowBox[{
                   StyleBox["c2", "ExpressionVariable"], "\[Equal]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["D", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["r", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:320557773"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p320557773-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p320557773-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:320557773", "Proof|ID:320557773-1"},
 CellID->1364381745],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p320557773-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->659473488]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Generalized Newman Lemma (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.645506382614538*^9, 3.6455063887223873`*^9}, {
  3.64551787310785*^9, 3.645517876872018*^9}},
 CellID->909732901],

Cell[TextData[{
 "In this section we prove the \[OpenCurlyDoubleQuote]Generalized Newman \
Lemma\[CloseCurlyDoubleQuote] which states that local connectibility implies \
confluence (and, hence, the Church-Rosser property). Although this lemma \
holds in a very general setting, i.e. for arbitrary Noetherian order \
relations \[Succeeds] and relations \[Rule] with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Rule]", 
    RowBox[{"\[SubsetEqual]", "\[Succeeds]"}]}], TraditionalForm]]],
 ", in domains ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " without any algebraic structure, we only state and prove it here for the \
concrete case of the order- / reduction relation in reduction rings."
}], "Text",
 CellChangeTimes->{{3.645506500570861*^9, 3.6455066667885942`*^9}},
 CellID->973491128],

Cell["\<\
The proof presented here is modelled after the one from [Winkler1983].\
\>", "Text",
 CellChangeTimes->{{3.645506731490416*^9, 3.645506758933936*^9}, {
   3.645507273181354*^9, 3.645507283220395*^9}, 3.645507337632083*^9},
 CellID->2040579658],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isReductionRing", "[", "\[ScriptCapitalR]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1193869577}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6442145299701653`*^9, 3.644214535574843*^9}, {
  3.6442180908663197`*^9, 3.644218093468452*^9}, {3.644221544014908*^9, 
  3.6442215452871227`*^9}},
 CellTags->{
  "ID:1193869577", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1193869577],

Cell["", "Text",
 CellID->1525379276],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["isReductor", "\[ScriptCapitalR]"], "[", "r", 
   "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 342993412}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645509655323125*^9, 3.6455096645435877`*^9}},
 CellTags->{
  "ID:342993412", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->342993412],

Cell["", "OpenEnvironment",
 CellID->2112539462],

Cell[CellGroupData[{

Cell["Lemma: Local Connectibility Implies Element-Confluence", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.645506473018393*^9, {3.6455100035662117`*^9, 3.645510003822116*^9}, {
   3.6455165221309633`*^9, 3.6455165366527843`*^9}},
 CellID->766889252],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
    RowBox[{"[", "T", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
              "-", "1"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["T", "i"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", "r"], "\[ScriptCapitalR]"], 
             SubscriptBox["T", 
              RowBox[{"i", "+", "1"}]]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
              RowBox[{"[", 
               RowBox[{"x", ",", "y"}], "]"}]}]], 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{
                  SubscriptBox["T", "i"], 
                  UnderscriptBox[
                   SubsuperscriptBox["\[Rule]", "r", "*"], 
                   "\[ScriptCapitalR]"], "x"}], 
                 TagBox[")",
                  "AutoParentheses"]}], "\[And]", 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{
                  SubscriptBox["T", "i"], 
                  UnderscriptBox[
                   SubsuperscriptBox["\[Rule]", "r", "*"], 
                   "\[ScriptCapitalR]"], "y"}], 
                 TagBox[")",
                  "AutoParentheses"]}]}], 
               TagBox[")",
                "AutoParentheses"]}], "\[Implies]", 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{"x", 
                UnderscriptBox[
                 SubsuperscriptBox["\[DownArrow]", "r", "*"], 
                 "\[ScriptCapitalR]"], "y"}], 
               TagBox[")",
                "AutoParentheses"]}]}], 
             TagBox[")",
              "AutoParentheses"]}]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       SubscriptBox["T", "1"], 
       UnderscriptBox[
        SubsuperscriptBox["\[DownArrow]", "r", "*"], "\[ScriptCapitalR]"], 
       SubscriptBox["T", "i"]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(63 ind)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1786903669", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645506482145002*^9, 3.645506494296689*^9}, {
   3.645506776700371*^9, 3.6455068734520197`*^9}, {3.6455069108686047`*^9, 
   3.645506985071248*^9}, {3.6455071587781076`*^9, 3.645507159835011*^9}, 
   3.645509666141499*^9, {3.645510023159965*^9, 3.645510075878603*^9}},
 CellTags->{
  "ID:1786903669", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "63 ind"},
 CellID->1786903669],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
    RowBox[{"[", "T", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"T", "\[NotEqual]", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
              "-", "1"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["T", "i"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", "r"], "\[ScriptCapitalR]"], 
             SubscriptBox["T", 
              RowBox[{"i", "+", "1"}]]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
              RowBox[{"[", 
               RowBox[{"x", ",", "y"}], "]"}]}]], 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{
                  SubscriptBox["T", "i"], 
                  UnderscriptBox[
                   SubsuperscriptBox["\[Rule]", "r", "*"], 
                   "\[ScriptCapitalR]"], "x"}], 
                 TagBox[")",
                  "AutoParentheses"]}], "\[And]", 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{
                  SubscriptBox["T", "i"], 
                  UnderscriptBox[
                   SubsuperscriptBox["\[Rule]", "r", "*"], 
                   "\[ScriptCapitalR]"], "y"}], 
                 TagBox[")",
                  "AutoParentheses"]}]}], 
               TagBox[")",
                "AutoParentheses"]}], "\[Implies]", 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{"x", 
                UnderscriptBox[
                 SubsuperscriptBox["\[DownArrow]", "r", "*"], 
                 "\[ScriptCapitalR]"], "y"}], 
               TagBox[")",
                "AutoParentheses"]}]}], 
             TagBox[")",
              "AutoParentheses"]}]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      SubscriptBox["T", "1"], 
      UnderscriptBox[
       SubsuperscriptBox["\[DownArrow]", "r", "*"], "\[ScriptCapitalR]"], 
      SubscriptBox["T", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(63)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1742873082", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645510930971099*^9, 3.645510939298806*^9}},
 CellTags->{
  "ID:1742873082", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "63"},
 CellID->1742873082],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isLocConnective", "\[ScriptCapitalR]"], "[", "r", "]"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
     RowBox[{"[", "z", "]"}]}]], 
   RowBox[{
    UnderscriptBox["isConfluent", "\[ScriptCapitalR]"], "[", 
    RowBox[{"z", ",", "r"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(64)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:332228196", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645513839331572*^9, 3.6455138654267817`*^9}, {
  3.6455141412899942`*^9, 3.6455141936687393`*^9}, {3.645514263372674*^9, 
  3.6455142645128107`*^9}},
 CellTags->{
  "ID:332228196", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "64"},
 CellID->332228196],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->3613269]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1064962880],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(63 ind)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(63 ind)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox[
                    TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["x", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox[
                    TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox[
                    TemplateBox[{"\[DownArrow]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["y", "ExpressionVariable"]}]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], "1"], 
               UnderscriptBox[
                TemplateBox[{"\[DownArrow]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox[
                "i", "ExpressionVariable"]]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox[
                    TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["x", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox[
                    TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox[
                    TemplateBox[{"\[DownArrow]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["y", "ExpressionVariable"]}]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], "1"], 
               UnderscriptBox[
                TemplateBox[{"\[DownArrow]", 
                  StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox[
                "i", "ExpressionVariable"]]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1786903669"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1786903669-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1786903669-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1786903669", "Proof|ID:1786903669-1"},
 CellID->624008027],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1786903669-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->539759205]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(63)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(63)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox[
                    TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["x", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox[
                    TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox[
                    TemplateBox[{"\[DownArrow]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["y", "ExpressionVariable"]}]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             SubscriptBox[
              StyleBox["T", "ExpressionVariable"], "1"], 
             UnderscriptBox[
              TemplateBox[{"\[DownArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             SubscriptBox[
              StyleBox["T", "ExpressionVariable"], 
              RowBox[{"\[LeftBracketingBar]", 
                StyleBox["T", "ExpressionVariable"], 
                "\[RightBracketingBar]"}]]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   StyleBox["r", "ExpressionVariable"]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox[
                    TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["x", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox[
                    TemplateBox[{"\[Rule]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox[
                    TemplateBox[{"\[DownArrow]", 
                    StyleBox["r", "ExpressionVariable"], "*"}, 
                    "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["y", "ExpressionVariable"]}]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             SubscriptBox[
              StyleBox["T", "ExpressionVariable"], "1"], 
             UnderscriptBox[
              TemplateBox[{"\[DownArrow]", 
                StyleBox["r", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             SubscriptBox[
              StyleBox["T", "ExpressionVariable"], 
              RowBox[{"\[LeftBracketingBar]", 
                StyleBox["T", "ExpressionVariable"], 
                "\[RightBracketingBar]"}]]}]}]}], $CellContext`TheoremaForm]],
      "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1742873082"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1742873082-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1742873082-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1742873082", "Proof|ID:1742873082-1"},
 CellID->1157809795],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1742873082-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.66235454969549*^9, 3.662354553907497*^9}},
 CellID->1202817432]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(64)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(64)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["r", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["isConfluent", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", 
                 StyleBox["r", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["r", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["isConfluent", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], ",", 
                 StyleBox["r", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:332228196"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p332228196-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p332228196-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:332228196", "Proof|ID:332228196-1"},
 CellID->282816996],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p332228196-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->356612985]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1303971219],

Cell[CellGroupData[{

Cell["Lemma: Element-Confluence is Equivalent to Confluence", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645516545812482*^9, 3.645516559197278*^9}},
 CellID->1400142810],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "z", "]"}]}]], 
    RowBox[{
     UnderscriptBox["isConfluent", "\[ScriptCapitalR]"], "[", 
     RowBox[{"z", ",", "r"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isConfluent", "\[ScriptCapitalR]"], "[", "r", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(65)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2110225002", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6455165718669357`*^9, 3.645516592582879*^9}},
 CellTags->{
  "ID:2110225002", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "65"},
 CellID->2110225002],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->842177265]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->957715670],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(65)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(65)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["isConfluent", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", 
                   StyleBox["r", "ExpressionVariable"]}], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isConfluent", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["r", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["isConfluent", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["z", "ExpressionVariable"], ",", 
                   StyleBox["r", "ExpressionVariable"]}], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isConfluent", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["r", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2110225002"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2110225002-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2110225002-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2110225002", "Proof|ID:2110225002-1"},
 CellID->2140493902],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2110225002-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->398616909]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1678414297],

Cell[CellGroupData[{

Cell["Theorem: Local Connectibility Implies Church-Rosser", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.64551685316457*^9, 3.645516862919817*^9}, {
  3.645516966216655*^9, 3.6455169724357557`*^9}},
 CellID->796670561],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isConfluent", "\[ScriptCapitalR]"], "[", "r", "]"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["isChurchRosser", "\[ScriptCapitalR]"], "[", "r", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(66)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:629188208", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645516921287274*^9, 3.645516933603569*^9}, {
  3.64551723776492*^9, 3.6455172437481623`*^9}},
 CellTags->{
  "ID:629188208", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "66"},
 CellID->629188208],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isLocConnective", "\[ScriptCapitalR]"], "[", "r", "]"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["isChurchRosser", "\[ScriptCapitalR]"], "[", "r", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(67)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1222366601", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64551697812776*^9, 3.645516990334179*^9}, {
   3.645517245627536*^9, 3.645517251469254*^9}, 3.645517795991736*^9},
 CellTags->{
  "ID:1222366601", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "67"},
 CellID->1222366601],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->790257744]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1030025520],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(66)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(66)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isConfluent", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["r", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isChurchRosser", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["r", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isConfluent", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["r", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isChurchRosser", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["r", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:629188208"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p629188208-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p629188208-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:629188208", "Proof|ID:629188208-1"},
 CellID->2037000940],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p629188208-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->294264870]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(67)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(67)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["r", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isChurchRosser", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["r", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["isReductor", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["r", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["r", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isChurchRosser", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["r", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1222366601"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1222366601-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1222366601-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1222366601", "Proof|ID:1222366601-1"},
 CellID->44880005],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1222366601-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->479889439]
}, Closed]],

Cell[TextData[{
 StyleBox["Remark",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 ": It is apparent that in the theorem above also the reverse directions \
hold, but we do not need them."
}], "Text",
 CellChangeTimes->{{3.645517017397613*^9, 3.645517065274435*^9}, {
  3.645517861403943*^9, 3.645517867914612*^9}},
 CellID->1360717621]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Results (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.644217980380916*^9, 3.644217981927495*^9}, {
  3.644218517346408*^9, 3.6442185181698523`*^9}, {3.645518053471242*^9, 
  3.6455180573131323`*^9}},
 CellID->341695266],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isReductionRing", "[", "\[ScriptCapitalR]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 2131577443}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6442145299701653`*^9, 3.644214535574843*^9}, {
  3.6442180908663197`*^9, 3.644218093468452*^9}, {3.644221544014908*^9, 
  3.6442215452871227`*^9}},
 CellTags->{
  "ID:2131577443", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2131577443],

Cell[CellGroupData[{

Cell["Sets (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.648812304142144*^9, 3.6488123174701767`*^9}},
 CellID->1825869525],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "\[ScriptCapitalR]", "]"}]], 
   RowBox[{"[", "B", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 2046259788}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6442145419912357`*^9, 3.644214554615238*^9}, {
  3.645172173387023*^9, 3.645172174639511*^9}},
 CellTags->{
  "ID:2046259788", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2046259788],

Cell["", "OpenEnvironment",
 CellID->1775269917],

Cell[CellGroupData[{

Cell["Lemma", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.645163343965102*^9},
 CellID->1388035298],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{
    "a", ",", "a0", ",", "m", ",", "m1", ",", "m2", ",", "c1", ",", "c2"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 205712485}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6451633512046757`*^9, 3.6451633585130043`*^9}, {
   3.64516340427645*^9, 3.6451634217258043`*^9}, {3.6451634543205*^9, 
   3.645163463575046*^9}, 3.645163497341564*^9, {3.645163742581835*^9, 
   3.6451637438030167`*^9}},
 CellTags->{
  "ID:205712485", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->205712485],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "c", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"a0", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
       RowBox[{"a0", 
        UnderscriptBox["+", "\[ScriptCapitalR]"], "c"}]}], "\[Implies]", 
      RowBox[{"a", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
       RowBox[{"a", 
        UnderscriptBox["+", "\[ScriptCapitalR]"], 
        RowBox[{"m", 
         UnderscriptBox["*", "\[ScriptCapitalR]"], "c"}]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
     RowBox[{"[", 
      RowBox[{"c", ",", "m0"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"a0", 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
      RowBox[{"a0", 
       UnderscriptBox["-", "\[ScriptCapitalR]"], 
       RowBox[{"m0", 
        UnderscriptBox["*", "\[ScriptCapitalR]"], "c"}]}]}], "\[Implies]", 
     RowBox[{"a", 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
      RowBox[{"a", 
       UnderscriptBox["-", "\[ScriptCapitalR]"], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"m", 
          UnderscriptBox["*", "\[ScriptCapitalR]"], "m0"}], ")"}], 
        UnderscriptBox["*", "\[ScriptCapitalR]"], "c"}]}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(61)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2050580243", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6451830899979763`*^9, 3.6451830986772623`*^9}, {
   3.645183129471429*^9, 3.64518314845015*^9}, {3.645183282874041*^9, 
   3.645183352254692*^9}, 3.645243871517479*^9},
 CellTags->{
  "ID:2050580243", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "61"},
 CellID->2050580243],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
           RowBox[{"[", "c", "]"}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           RowBox[{"a0", 
            UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
            RowBox[{"a0", 
             UnderscriptBox["+", "\[ScriptCapitalR]"], "c"}]}], "\[Implies]", 
           
           RowBox[{"a", 
            UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], 
            RowBox[{"a", 
             UnderscriptBox["+", "\[ScriptCapitalR]"], 
             RowBox[{"m", 
              UnderscriptBox["*", "\[ScriptCapitalR]"], "c"}]}]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
           RowBox[{"[", 
            RowBox[{"b", ",", "c", ",", "d"}], "]"}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{"b", 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{"{", "c", "}"}]], "\[ScriptCapitalR]"], "d"}], 
            TagBox[")",
             "AutoParentheses"]}], "\[Implies]", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{"m", 
               UnderscriptBox["*", "\[ScriptCapitalR]"], "b"}], 
              TagBox[")",
               "AutoParentheses"]}], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               RowBox[{"{", "c", "}"}]], "\[ScriptCapitalR]"], 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{"m", 
               UnderscriptBox["*", "\[ScriptCapitalR]"], "d"}], 
              TagBox[")",
               "AutoParentheses"]}]}], 
            TagBox[")",
             "AutoParentheses"]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]},
       {
        RowBox[{
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"a0", 
           UnderscriptBox["-", "\[ScriptCapitalR]"], 
           RowBox[{"m1", 
            UnderscriptBox["*", "\[ScriptCapitalR]"], "c1"}]}], 
          TagBox[")",
           "AutoParentheses"]}], 
         UnderscriptBox[
          OverscriptBox[
           SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
           RowBox[{"a0", ",", 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             TagBox["\[Succeeds]",
              Identity], 
             TagBox[")",
              "AutoParentheses"]}]}]], "\[ScriptCapitalR]"], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"a0", 
           UnderscriptBox["-", "\[ScriptCapitalR]"], 
           RowBox[{"m2", 
            UnderscriptBox["*", "\[ScriptCapitalR]"], "c2"}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox["-", "\[ScriptCapitalR]"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"m", 
         UnderscriptBox["*", "\[ScriptCapitalR]"], "m1"}], ")"}], 
       UnderscriptBox["*", "\[ScriptCapitalR]"], "c1"}]}], 
     TagBox[")",
      "AutoParentheses"]}], 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
      RowBox[{"a", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "\[ScriptCapitalR]"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox["-", "\[ScriptCapitalR]"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"m", 
         UnderscriptBox["*", "\[ScriptCapitalR]"], "m2"}], ")"}], 
       UnderscriptBox["*", "\[ScriptCapitalR]"], "c2"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(60)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2100291458", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645163467534511*^9, 3.645163469491725*^9}, {
  3.6451635057188272`*^9, 3.645163686995387*^9}, {3.645163746158593*^9, 
  3.645163795802248*^9}, {3.645163825866686*^9, 3.6451638820360527`*^9}, {
  3.6451641329728737`*^9, 3.645164134197617*^9}, {3.6451712937174273`*^9, 
  3.6451712947889023`*^9}, {3.661851290238967*^9, 3.6618512966349783`*^9}},
 CellTags->{
  "ID:2100291458", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "60"},
 CellID->2100291458],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"c", "\[Element]", "B"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1810831635}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6452436976779757`*^9, 3.645243700765829*^9}},
 CellTags->{
  "ID:1810831635", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1810831635],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{"b1", ",", "b2", ",", "m3", ",", "m4"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 718268018}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645243707986251*^9, 3.6452437173516006`*^9}, {
  3.645245492455163*^9, 3.645245494380419*^9}},
 CellTags->{
  "ID:718268018", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->718268018],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{"a", 
            UnderscriptBox["-", "\[ScriptCapitalR]"], 
            RowBox[{"m3", 
             UnderscriptBox["*", "\[ScriptCapitalR]"], "c"}]}], 
           TagBox[")",
            "AutoParentheses"]}], 
          UnderscriptBox[
           OverscriptBox[
            SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
            RowBox[{"a", ",", 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              TagBox["\[Succeeds]",
               Identity], 
              TagBox[")",
               "AutoParentheses"]}]}]], "\[ScriptCapitalR]"], 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{"a", 
            UnderscriptBox["-", "\[ScriptCapitalR]"], 
            RowBox[{"m4", 
             UnderscriptBox["*", "\[ScriptCapitalR]"], "c"}]}], 
           TagBox[")",
            "AutoParentheses"]}]}], 
         TagBox[")",
          "AutoParentheses"]}]},
       {
        RowBox[{"a", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", 
           RowBox[{"m1", ",", "c"}]], "\[ScriptCapitalR]"], "b1"}]},
       {
        RowBox[{"a", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", 
           RowBox[{"m2", ",", "c"}]], "\[ScriptCapitalR]"], "b2"}]},
       {
        RowBox[{"a", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", 
           RowBox[{"m3", ",", "c"}]], "\[ScriptCapitalR]"], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"a", 
           UnderscriptBox["-", "\[ScriptCapitalR]"], 
           RowBox[{"m3", 
            UnderscriptBox["*", "\[ScriptCapitalR]"], "c"}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]},
       {
        RowBox[{"a", 
         UnderscriptBox[
          SubscriptBox["\[Rule]", 
           RowBox[{"m4", ",", "c"}]], "\[ScriptCapitalR]"], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"a", 
           UnderscriptBox["-", "\[ScriptCapitalR]"], 
           RowBox[{"m4", 
            UnderscriptBox["*", "\[ScriptCapitalR]"], "c"}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]},
       {
        RowBox[{
         UnderscriptBox["areCorrelative", "\[ScriptCapitalR]"], "[", 
         RowBox[{"c", ",", "m1", ",", "m3"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["areCorrelative", "\[ScriptCapitalR]"], "[", 
         RowBox[{"c", ",", "m2", ",", "m4"}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"b1", 
    UnderscriptBox[
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
      RowBox[{"a", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[Succeeds]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}]], "\[ScriptCapitalR]"], "b2"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(62)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1959449230", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645243723051998*^9, 3.645243823940624*^9}, {
  3.6452447649237223`*^9, 3.645244780131734*^9}, {3.645244891105434*^9, 
  3.645244940422428*^9}, {3.645245264698123*^9, 3.645245281572537*^9}, {
  3.645245497632884*^9, 3.645245517437874*^9}},
 CellTags->{
  "ID:1959449230", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "62"},
 CellID->1959449230],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->707969697]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1415542183],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(61)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(61)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a0", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a0", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["a0", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["m0", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a0", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["a0", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m0", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[Implies]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["m0", "ExpressionVariable"]}], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox[
                    "c", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a0", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a0", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["a0", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["m0", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a0", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["a0", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m0", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[Implies]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["m0", "ExpressionVariable"]}], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox[
                    "c", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2050580243"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2050580243-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2050580243-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2050580243", "Proof|ID:2050580243-1"},
 CellID->1516400018],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2050580243-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1738113272]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(60)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(60)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a0", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["a0", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a0", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["d", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["d", "ExpressionVariable"]}]}]}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["a0", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c1", "ExpressionVariable"]}]}], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["a0", "ExpressionVariable"], ",", 
                   "\[Succeeds]"}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a0", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c2", "ExpressionVariable"]}]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["m1", "ExpressionVariable"]}], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox["c1", "ExpressionVariable"]}]}], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["m2", "ExpressionVariable"]}], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox[
                 "c2", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a0", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["a0", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a0", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], 
                   UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["d", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], 
                   UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    StyleBox["d", "ExpressionVariable"]}]}]}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["a0", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c1", "ExpressionVariable"]}]}], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["a0", "ExpressionVariable"], ",", 
                   "\[Succeeds]"}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a0", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c2", "ExpressionVariable"]}]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["m1", "ExpressionVariable"]}], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox["c1", "ExpressionVariable"]}]}], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["m2", "ExpressionVariable"]}], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox[
                 "c2", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2100291458"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2100291458-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2100291458-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2100291458", "Proof|ID:2100291458-1"},
 CellID->44637999],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2100291458-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->471240787]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(62)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(62)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m3", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m4", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m3", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m4", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m3", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m3", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m4", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m4", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m3", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["m4", "ExpressionVariable"]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["b1", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m3", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m4", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m3", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m4", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m3", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m3", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Rule]", 
                 RowBox[{
                   StyleBox["m4", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 RowBox[{
                   StyleBox["m4", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m3", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["m4", "ExpressionVariable"]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["b1", "ExpressionVariable"], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1959449230"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1959449230-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1959449230-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1959449230", "Proof|ID:1959449230-1"},
 CellID->340301213],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1959449230-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->86574173]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (60)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.645164172580628*^9, 3.645164173483329*^9}, 
   3.645164221446558*^9, {3.6452440983191023`*^9, 3.645244100494193*^9}, 
   3.653219564087068*^9, 3.666174386770012*^9, 3.666440954106386*^9},
 CellID->1206255241],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "\[ScriptCapitalR]", "]"}]], 
    RowBox[{"[", "B", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", 
      SubscriptBox["a", "0"], ",", "m", ",", 
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"], ",", 
      SubscriptBox["c", "1"], ",", 
      SubscriptBox["c", "2"]}], "]"}]}], TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", "c"], 
     RowBox[{
      SubscriptBox["a", "0"], "\[Succeeds]", 
      RowBox[{
       SubscriptBox["a", "0"], "+", "c"}]}]}], "\[Implies]", 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"a", "+", 
      RowBox[{"m", "*", "c"}]}]}]}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"b", ",", "c", ",", "d"}]], "b"}], 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{"d", "\[Implies]", 
     RowBox[{
      RowBox[{"m", "*", "b"}], 
      SubscriptBox["\[LeftRightArrow]", "c"], 
      RowBox[{"m", "*", "d"}]}]}]}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "1"], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{
      SubscriptBox["a", "0"], ",", "\[Succeeds]"}]], 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "i"], ":=", 
    RowBox[{
     SubscriptBox["a", "0"], "-", 
     RowBox[{
      SubscriptBox["m", "i"], "*", 
      SubscriptBox["c", "i"]}]}]}], TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "2"}], TraditionalForm]]],
 ", and show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["d", "1"], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    SubscriptBox["d", "2"]}], TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["d", "i"], ":=", 
    RowBox[{"a", "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"m", "*", 
        SubscriptBox["m", "i"]}], ")"}], "*", 
      SubscriptBox["c", "i"]}]}]}], TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "2"}], TraditionalForm]]],
 ".\n\nWe expand the definition of \[OpenCurlyDoubleQuote]connectibility \
below\[CloseCurlyDoubleQuote] and ",
 Cell[BoxData[
  FormBox["isChainBelow", TraditionalForm]]],
 " in (3): We know\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[NotEqual]", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "1"], "=", 
    SubscriptBox["T", "1"]}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "2"], "=", 
    SubscriptBox["T", 
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]]}], 
   TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "T", "]"}]}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], "-", 
       "1"}]}]], 
    RowBox[{
     SubscriptBox["T", "i"], 
     SubscriptBox["\[LeftRightArrow]", "B"], 
     SubscriptBox["T", 
      RowBox[{"i", "+", "1"}]]}]}], TraditionalForm]]],
 "\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     SubscriptBox["a", "0"], "\[Succeeds]", 
     SubscriptBox["T", "i"]}]}], TraditionalForm]]],
 "\nfor some tuple ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 ".\n\nFor showing (G), by (50), it suffices to show\n(G\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    SubscriptBox["d", "1"]}], TraditionalForm]]],
 "\nand to find a tuple ",
 Cell[BoxData[
  FormBox["S", TraditionalForm]]],
 " satisfying\n(G.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "D", "]"}]], 
    RowBox[{"[", "S", "]"}]}], TraditionalForm]]],
 "\n(G.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", "\[NotEqual]", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 "\n(G.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["d", "1"], "=", 
    SubscriptBox["S", "1"]}], TraditionalForm]]],
 "\n(G.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["d", "2"], "=", 
    SubscriptBox["S", 
     RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]]}], 
   TraditionalForm]]],
 "\n(G.5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}], "-", 
       "1"}]}]], 
    RowBox[{
     SubscriptBox["S", "i"], 
     OverscriptBox[
      SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
      RowBox[{"a", ",", "\[Succeeds]"}]], 
     SubscriptBox["S", 
      RowBox[{"i", "+", "1"}]]}]}], TraditionalForm]]],
 "\n\nBy (1), (G\[CloseCurlyQuote]) can be reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["a", "0"], "\[Succeeds]", 
    RowBox[{
     SubscriptBox["a", "0"], "+", 
     RowBox[{
      SubscriptBox["m", "1"], "*", 
      SubscriptBox["c", "1"]}]}]}], TraditionalForm]]],
 ", which follows readily from (3).\nWe claim that tuple ",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", ":=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"a", "+", 
       RowBox[{"m", "*", 
        RowBox[{"(", 
         RowBox[{
          SubscriptBox["T", "i"], "-", 
          SubscriptBox["a", "0"]}], ")"}]}]}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 " is a suitable witness for (G.1)-(G.5).\n\n(G.1):\nObvious.\n\n(G.2):\n\
Obvious.\n\n(G.3):\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", "1"], "=", 
    RowBox[{
     RowBox[{"a", "+", 
      RowBox[{"m", "*", 
       RowBox[{"(", 
        RowBox[{
         SubscriptBox["T", "1"], "-", 
         SubscriptBox["a", "0"]}], ")"}]}]}], "=", 
     RowBox[{
      RowBox[{"a", "+", 
       RowBox[{"m", "*", 
        RowBox[{"(", 
         RowBox[{
          SubscriptBox["b", "1"], "-", 
          SubscriptBox["a", "0"]}], ")"}]}]}], "=", 
      RowBox[{
       RowBox[{"a", "+", 
        RowBox[{"m", "*", 
         RowBox[{"(", 
          RowBox[{
           SubscriptBox["a", "0"], "-", 
           RowBox[{
            SubscriptBox["m", "1"], "*", 
            SubscriptBox["c", "1"]}], "-", 
           SubscriptBox["a", "0"]}], ")"}]}]}], "=", 
       RowBox[{
        RowBox[{"a", "-", 
         RowBox[{"m", "*", 
          SubscriptBox["m", "1"], "*", 
          SubscriptBox["c", "1"]}]}], "=", 
        SubscriptBox["d", "1"]}]}]}]}]}], TraditionalForm]]],
 "\n\n(G.4):\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", 
     RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]], "=", 
    RowBox[{
     RowBox[{"a", "+", 
      RowBox[{"m", "*", 
       RowBox[{"(", 
        RowBox[{
         SubscriptBox["T", 
          RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]], 
         "-", 
         SubscriptBox["a", "0"]}], ")"}]}]}], "=", 
     RowBox[{
      RowBox[{"a", "+", 
       RowBox[{"m", "*", 
        RowBox[{"(", 
         RowBox[{
          SubscriptBox["b", "2"], "-", 
          SubscriptBox["a", "0"]}], ")"}]}]}], "=", 
      RowBox[{
       RowBox[{"a", "+", 
        RowBox[{"m", "*", 
         RowBox[{"(", 
          RowBox[{
           SubscriptBox["a", "0"], "-", 
           RowBox[{
            SubscriptBox["m", "2"], "*", 
            SubscriptBox["c", "2"]}], "-", 
           SubscriptBox["a", "0"]}], ")"}]}]}], "=", 
       RowBox[{
        RowBox[{"a", "-", 
         RowBox[{"m", "*", 
          SubscriptBox["m", "2"], "*", 
          SubscriptBox["c", "2"]}]}], "=", 
        SubscriptBox["d", "2"]}]}]}]}]}], TraditionalForm]]],
 "\n\n(G.5):\nChoose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}], "-", 
     "1"}]}], TraditionalForm]]],
 " a.b.f. and show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", "i"], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    SubscriptBox["S", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 ", i.e. show\n(G.6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "+", 
     RowBox[{"m", "*", 
      RowBox[{"(", 
       RowBox[{
        SubscriptBox["T", "i"], "-", 
        SubscriptBox["a", "0"]}], ")"}]}]}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", ">"}]], 
    RowBox[{"a", "+", 
     RowBox[{"m", "*", 
      RowBox[{"(", 
       RowBox[{
        SubscriptBox["T", 
         RowBox[{"i", "+", "1"}]], "-", 
        SubscriptBox["a", "0"]}], ")"}]}]}]}], TraditionalForm]]],
 "\nor, rearranging the two sides,\n(G.6\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"m", "*", 
      SubscriptBox["T", "i"]}], "+", 
     RowBox[{"(", 
      RowBox[{"a", "-", 
       RowBox[{"m", "*", 
        SubscriptBox["a", "0"]}]}], ")"}]}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", ">"}]], 
    RowBox[{
     RowBox[{"m", "*", 
      SubscriptBox["T", 
       RowBox[{"i", "+", "1"}]]}], "+", 
     RowBox[{"(", 
      RowBox[{"a", "-", 
       RowBox[{"m", "*", 
        SubscriptBox["a", "0"]}]}], ")"}]}]}], TraditionalForm]]],
 "\n\nBy (55) this can be reduced to\n(G.7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m", "*", 
     SubscriptBox["T", "i"]}], 
    SubscriptBox["\[LeftRightArrow]", "B"], 
    RowBox[{"m", "*", 
     SubscriptBox["T", 
      RowBox[{"i", "+", "1"}]]}]}], TraditionalForm]]],
 "\n(G.8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"a", "+", 
     RowBox[{"m", "*", 
      RowBox[{"(", 
       RowBox[{
        SubscriptBox["T", "i"], "-", 
        SubscriptBox["a", "0"]}], ")"}]}]}]}], TraditionalForm]]],
 "\n(G.9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"a", "+", 
     RowBox[{"m", "*", 
      RowBox[{"(", 
       RowBox[{
        SubscriptBox["T", 
         RowBox[{"i", "+", "1"}]], "-", 
        SubscriptBox["a", "0"]}], ")"}]}]}]}], TraditionalForm]]],
 "\n\n(G.7):\nWe instantiate (8) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ":=", "i"}], TraditionalForm]]],
 " and know\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", "i"], 
    SubscriptBox["\[LeftRightArrow]", "B"], 
    SubscriptBox["T", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 "\nFrom the definition of reduction modulo sets we know that there exists ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[Element]", "B"}], TraditionalForm]]],
 " with\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", "i"], 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{
     RowBox[{
      SubscriptBox["T", 
       RowBox[{"i", "+", "1"}]], "\[Or]", 
      SubscriptBox["T", 
       RowBox[{"i", "+", "1"}]]}], 
     SubscriptBox["\[Rule]", "c"], 
     SubscriptBox["T", "i"]}]}], TraditionalForm]]],
 "\nbased upon which we have to distinguish two cases; However, since both of \
them are completely symmetric, we only consider the first one with\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", "i"], 
    SubscriptBox["\[Rule]", "c"], 
    SubscriptBox["T", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 "\n\nFrom (2) we can infer\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m", "*", 
     SubscriptBox["T", "i"]}], 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{"m", "*", 
     SubscriptBox["T", 
      RowBox[{"i", "+", "1"}]]}]}], TraditionalForm]]],
 "\nfrom which our current goal (G.7) follows readily.\n\n(G.8):\nBy (1) our \
goal can be reduced to\n(G.10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["a", "0"], "\[Succeeds]", 
    RowBox[{
     SubscriptBox["a", "0"], "+", 
     RowBox[{"(", 
      RowBox[{
       SubscriptBox["T", "i"], "-", 
       SubscriptBox["a", "0"]}], ")"}]}]}], TraditionalForm]]],
 "\nwhich is equivalent to\n(G.11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["a", "0"], "\[Succeeds]", 
    SubscriptBox["T", "i"]}], TraditionalForm]]],
 "\n(G.11) is just an instance of (9), hence we are done.\n\n(G.9):\n\
Analogous."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.645164200712892*^9, 3.645164459259171*^9}, {
   3.645169195794272*^9, 3.6451693558120413`*^9}, {3.645169416553853*^9, 
   3.645169416737185*^9}, {3.645169462700242*^9, 3.645169504059949*^9}, {
   3.6451695555242443`*^9, 3.64516960646987*^9}, {3.645169640693246*^9, 
   3.645169701861085*^9}, {3.6451697572450542`*^9, 3.645169814118965*^9}, {
   3.645169865088724*^9, 3.645169868611512*^9}, {3.645169899529605*^9, 
   3.645170152669421*^9}, {3.645171301207546*^9, 3.645171302119192*^9}, {
   3.645171346395781*^9, 3.645171421330762*^9}, 3.653219564087068*^9, 
   3.666174386770012*^9, 3.666440954106497*^9},
 CellID->1191642924],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.645164207543725*^9, 3.6451642214468393`*^9}, 
   3.653219564087068*^9, 3.666174386770012*^9, 3.666440954106604*^9},
 TextAlignment->Right,
 CellID->696143052]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1967748303],

Cell[CellGroupData[{

Cell["Lemma: Non-trivial common reducibles", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.609066964866522*^9, 3.609066965032017*^9}, {
  3.609067731644119*^9, 3.609067762326927*^9}},
 CellID->496479384],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"c1", ",", 
    RowBox[{"c2", "\[Element]", "B"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"c1", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
      RowBox[{"c2", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
      RowBox[{"c1", "\[NotEqual]", "c2"}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox[
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
           RowBox[{"[", "a0", "]"}]}]], 
         RowBox[{
          UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
          RowBox[{"c1", ",", "c2", ",", "a0", ",", "1", ",", "1"}], "]"}]], 
        RowBox[{
         UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
         RowBox[{"c1", ",", "c2", ",", "a0", ",", "1", ",", "1", ",", "B"}], 
         "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
         RowBox[{"[", "a", "]"}]}]], 
       RowBox[{
        UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
        RowBox[{"c1", ",", "c2", ",", "a", ",", "1", ",", "1"}], "]"}]], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
         RowBox[{"[", 
          RowBox[{"m1", ",", "m2"}], "]"}]}]], 
       RowBox[{
        UnderscriptBox["areConnectibleMultipliers", "\[ScriptCapitalR]"], "[", 
        RowBox[{"m1", ",", "m2", ",", "c1", ",", "c2", ",", "a", ",", "B"}], 
        "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(70)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1029764736", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.609067046935151*^9, 3.609067059726744*^9}, {
   3.609067100472937*^9, 3.609067121711163*^9}, {3.609067152912697*^9, 
   3.609067234875425*^9}, {3.609067300505913*^9, 3.609067380033606*^9}, {
   3.609067479251389*^9, 3.60906747954137*^9}, {3.609067528543763*^9, 
   3.60906759770064*^9}, {3.609067778795678*^9, 3.609067817641179*^9}, {
   3.609068390504053*^9, 3.609068393676582*^9}, {3.609068730990082*^9, 
   3.609068764090291*^9}, {3.610170137175974*^9, 3.610170142494369*^9}, {
   3.610179633782651*^9, 3.610179648250144*^9}, {3.63185536255756*^9, 
   3.6318553736167088`*^9}, {3.6318564408139*^9, 3.6318564536422157`*^9}, 
   3.632043439630452*^9, {3.6324790717325697`*^9, 3.632479073768631*^9}, {
   3.63247911916451*^9, 3.632479122193976*^9}, {3.6324799117778463`*^9, 
   3.632479915799437*^9}, {3.6324814763329773`*^9, 3.63248147696012*^9}, {
   3.632546852916201*^9, 3.632546860011125*^9}, {3.632546946934033*^9, 
   3.632547015329278*^9}, {3.632548262424518*^9, 3.6325482837270327`*^9}, 
   3.6325485143377523`*^9, {3.632548767995976*^9, 3.63254877547025*^9}, {
   3.644219754303247*^9, 3.644219806290162*^9}, {3.6451768529344397`*^9, 
   3.6451768739230347`*^9}, {3.645184004140772*^9, 3.645184053421671*^9}, 
   3.6451841076519737`*^9},
 CellTags->{
  "ID:1029764736", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "70"},
 CellID->1029764736],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"c", "\[Element]", "B"}]], 
   RowBox[{"c", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i", ",", 
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{
          UnderscriptBox["multN", "\[ScriptCapitalR]"], "[", "c", "]"}]}]], 
       RowBox[{"i", "\[NotEqual]", "j"}]], 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
          RowBox[{"[", "a0", "]"}]}]], 
        RowBox[{
         UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
         RowBox[{"c", ",", "c", ",", "a0", ",", "i", ",", "j"}], "]"}]], 
       RowBox[{
        UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
        RowBox[{"c", ",", "c", ",", "a0", ",", "i", ",", "j", ",", "B"}], 
        "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i", ",", 
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{
          UnderscriptBox["multN", "\[ScriptCapitalR]"], "[", "c", "]"}]}]], 
       RowBox[{"i", "\[NotEqual]", "j"}]], 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
          RowBox[{"[", "a", "]"}]}]], 
        RowBox[{
         UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
         RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}]], 
       RowBox[{
        UnderscriptBox["\[Exists]", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
          RowBox[{"[", 
           RowBox[{"m1", ",", "m2"}], "]"}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"\[And]", 
          RowBox[{"\[Piecewise]", GridBox[{
             {
              RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
               RowBox[{"m1", ",", "c", ",", "i"}], "]"}]},
             {
              RowBox[{
               UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
               RowBox[{"m2", ",", "c", ",", "j"}], "]"}]},
             {
              RowBox[{
               
               UnderscriptBox["areConnectibleMultipliers", 
                "\[ScriptCapitalR]"], "[", 
               RowBox[{
               "m1", ",", "m2", ",", "c", ",", "c", ",", "a", ",", "B"}], 
               "]"}]}
            }]}]}], 
         TagBox[")",
          "AutoParentheses"]}]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(71)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1120053244", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.6325469221952972`*^9, {3.6325470250827103`*^9, 3.632547077717621*^9}, {
   3.644219812697034*^9, 3.644219876781495*^9}, {3.644292840736085*^9, 
   3.644292864312611*^9}, {3.6451768763173523`*^9, 3.645176899271352*^9}, {
   3.645184057376522*^9, 3.6451840976703043`*^9}},
 CellTags->{
  "ID:1120053244", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "71"},
 CellID->1120053244],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1850395151]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->247765203],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(70)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(70)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}, {
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a0", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  UnderscriptBox["isMNTCR", 
                   StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                  
                  RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", "1", ",", 
                    "1"}], "]"}]], 
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["a0", "ExpressionVariable"], ",", "1", ",", "1", 
                   ",", 
                   StyleBox["B", "ExpressionVariable"]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                UnderscriptBox["isNTCR", 
                 StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["c1", "ExpressionVariable"], ",", 
                  StyleBox["c2", "ExpressionVariable"], ",", 
                  StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
                "]"}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["areConnectibleMultipliers", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["B", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}, {
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a0", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  UnderscriptBox["isMNTCR", 
                   StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                  
                  RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", "1", ",", 
                    "1"}], "]"}]], 
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["a0", "ExpressionVariable"], ",", "1", ",", "1", 
                   ",", 
                   StyleBox["B", "ExpressionVariable"]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                UnderscriptBox["isNTCR", 
                 StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["c1", "ExpressionVariable"], ",", 
                  StyleBox["c2", "ExpressionVariable"], ",", 
                  StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
                "]"}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["areConnectibleMultipliers", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["B", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1029764736"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1029764736-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1029764736-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1029764736", "Proof|ID:1029764736-1"},
 CellID->59145356],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1029764736-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->7007433]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(71)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(71)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                  StyleBox["j", "ExpressionVariable"]}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a0", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]], 
                 RowBox[{
                   UnderscriptBox["cpConnectible", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
              RowBox[{
                StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                StyleBox["j", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  UnderscriptBox["isNTCR", 
                   StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                  
                  RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]], 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["areConnectibleMultipliers", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                  StyleBox["j", "ExpressionVariable"]}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a0", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]], 
                 RowBox[{
                   UnderscriptBox["cpConnectible", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
              RowBox[{
                StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                StyleBox["j", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  UnderscriptBox["isNTCR", 
                   StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                  
                  RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]], 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["areConnectibleMultipliers", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1120053244"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1120053244-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1120053244-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1120053244", "Proof|ID:1120053244-1"},
 CellID->334765073],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1120053244-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->472868012]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof of (70).",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.609492857396531*^9, 3.609492858255178*^9}, 
   3.609574780520083*^9, 3.631855355808807*^9, {3.632547138949471*^9, 
   3.632547168095376*^9}, {3.6325472103500347`*^9, 3.632547225477921*^9}, 
   3.632547284077764*^9, 3.636348622273294*^9, 3.643533785292046*^9, {
   3.645176159384726*^9, 3.6451761596438847`*^9}, 3.6532195569890556`*^9, 
   3.6661743783771973`*^9, 3.6664409460352793`*^9},
 CellID->1828699765],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632548523692032*^9, 3.632548524773535*^9}, 
   3.632549158935813*^9, 3.636348622273456*^9, 3.6435337852922163`*^9, 
   3.6532195569890556`*^9, 3.6661743783771973`*^9, 3.666440946035384*^9},
 CellID->2044444786],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", "\[NotEqual]", "y"}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"multN", "[", "x", "]"}]}]], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "D"], 
         RowBox[{"[", 
          RowBox[{"a", ",", 
           SubscriptBox["m", "1"], ",", 
           SubscriptBox["m", "2"]}], "]"}]}]], 
       RowBox[{"(", 
        RowBox[{"\[And]", 
         RowBox[{"{", GridBox[{
            {
             RowBox[{
              RowBox[{"isCommonReducible", "[", 
               RowBox[{"x", ",", "y", ",", "a", ",", "i", ",", "j"}], "]"}], 
              "\[DoubleLeftRightArrow]", 
              RowBox[{"isCommonReducible", "[", 
               RowBox[{"x", ",", "y", ",", "a", ",", "1", ",", "1"}], "]"}]}]},
            {
             RowBox[{
              RowBox[{"isCritPair", "[", 
               RowBox[{
                SubscriptBox["m", "1"], ",", 
                SubscriptBox["m", "2"], ",", "x", ",", "y", ",", "i", ",", 
                "j"}], "]"}], "\[DoubleLeftRightArrow]", 
              RowBox[{"isCritPair", "[", 
               RowBox[{
                SubscriptBox["m", "1"], ",", 
                SubscriptBox["m", "2"], ",", "x", ",", "y", ",", "1", ",", 
                "1"}], "]"}]}]},
            {
             RowBox[{
              RowBox[{"isNTCR", "[", 
               RowBox[{"x", ",", "y", ",", "a", ",", "i", ",", "j"}], "]"}], 
              "\[DoubleLeftRightArrow]", 
              RowBox[{"isNTCR", "[", 
               RowBox[{"x", ",", "y", ",", "a", ",", "1", ",", "1"}], "]"}]}]},
            {
             RowBox[{
              RowBox[{"isMNTCR", "[", 
               RowBox[{"x", ",", "y", ",", "a", ",", "i", ",", "j"}], "]"}], 
              "\[DoubleLeftRightArrow]", 
              RowBox[{"isMNTCR", "[", 
               RowBox[{"x", ",", "y", ",", "a", ",", "1", ",", "1"}], 
               "]"}]}]}
           }]}]}], ")"}]}]}], ")"}]}], TraditionalForm]]],
 "\t... (2)-(5)"
}], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.632548529788869*^9, 3.632548626929134*^9}, {
   3.6325486689411716`*^9, 3.632548683150366*^9}, {3.632548817578206*^9, 
   3.632548827271421*^9}, {3.6325488595535717`*^9, 3.632548910230887*^9}, {
   3.632548956118149*^9, 3.6325489704408913`*^9}, 3.632549158935976*^9, 
   3.6363486222735977`*^9, 3.643533785292344*^9, {3.644292686345828*^9, 
   3.644292747077454*^9}, {3.6451801752286663`*^9, 3.6451801859005337`*^9}, 
   3.6532195569890556`*^9, 3.6661743783771973`*^9, 3.66644094603547*^9},
 CellID->1611514051],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox["B", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " a.b.f.\nAssume\n(0)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[NotEqual]", "y"}], TraditionalForm]]],
 "\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", 
        SubscriptBox["a", "0"], "]"}]}]], 
     RowBox[{"mntcr", "[", 
      RowBox[{"x", ",", "y", ",", "a", ",", "1", ",", "1"}], "]"}]], 
    RowBox[{"cpConnectible", "[", 
     RowBox[{"x", ",", "y", ",", 
      SubscriptBox["a", "0"], ",", "1", ",", "1", ",", "B"}], "]"}]}], 
   TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"ntcr", "[", 
    RowBox[{"x", ",", "y", ",", "a", ",", "1", ",", "1"}], "]"}], 
   TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"]}]], 
    RowBox[{"areConnectibleMultipliers", "[", 
     RowBox[{
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"], ",", "x", ",", "y", ",", "a", ",", "B"}], 
     "]"}]}], TraditionalForm]]],
 "\n\nWe now use (R7), immediately instantiating it with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c1", ":=", "x"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c2", ":=", "y"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Rule]", "a"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[Rule]", "1"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"j", "\[Rule]", "1"}], TraditionalForm]]],
 ". The premise of (R7) (",
 Cell[BoxData[
  FormBox[
   RowBox[{"ntcr", "[", 
    RowBox[{"D", ",", "x", ",", "y", ",", "a", ",", "1", ",", "1"}], "]"}], 
   TraditionalForm]]],
 ") is satisfied due to (2), hence we can add its conclusion to our knowledge \
base; For this, let ",
 Cell[BoxData[
  FormBox["k", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["l", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["a", "0"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["m", TraditionalForm]]],
 " witness the existential quantifier. We know\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", ",", 
    RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", "x", "]"}]}], TraditionalForm]]],
 "\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"a", "\[Succeeds]", 
      SubscriptBox["a", "0"]}], "\[Or]", "a"}], "=", 
    SubscriptBox["a", "0"]}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mntcr", "[", 
    RowBox[{"x", ",", "y", ",", 
     SubscriptBox["a", "0"], ",", "k", ",", "l"}], "]"}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       SubscriptBox["m", "0"], ",", "c"}]], 
     RowBox[{"mult", "[", 
      RowBox[{
       SubscriptBox["m", "0"], ",", "c"}], "]"}]], 
    RowBox[{"mult", "[", 
     RowBox[{
      RowBox[{"m", "*", 
       SubscriptBox["m", "0"]}], ",", "c"}], "]"}]}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", "c"], 
     RowBox[{
      SubscriptBox["a", "0"], "\[Succeeds]", 
      RowBox[{
       SubscriptBox["a", "0"], "+", "c"}]}]}], "\[Implies]", 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"a", "+", 
      RowBox[{"m", "*", "c"}]}]}]}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"b", ",", "c", ",", "d"}]], "b"}], 
    SubscriptBox["\[Rule]", "c"], 
    RowBox[{"d", "\[Implies]", 
     RowBox[{
      RowBox[{"m", "*", "b"}], 
      SubscriptBox["\[LeftRightArrow]", "c"], 
      RowBox[{"m", "*", "d"}]}]}]}], TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "=", 
    RowBox[{"y", "\[Implies]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         SubscriptBox["m", "1"], ",", 
         SubscriptBox["m", "2"]}]], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"reducible", "[", 
          RowBox[{"D", ",", 
           SubscriptBox["a", "0"], ",", "x", ",", 
           SubscriptBox["m", "1"]}], "]"}], "\[And]", 
         RowBox[{"reducible", "[", 
          RowBox[{"D", ",", 
           SubscriptBox["a", "0"], ",", "y", ",", 
           SubscriptBox["m", "2"]}], "]"}], "\[And]", 
         RowBox[{"mult", "[", 
          RowBox[{
           SubscriptBox["m", "1"], ",", "x", ",", "k"}], "]"}], "\[And]", 
         RowBox[{"mult", "[", 
          RowBox[{
           SubscriptBox["m", "2"], ",", "x", ",", "l"}], "]"}]}], ")"}]}], 
      "\[Implies]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"mult", "[", 
           RowBox[{
            RowBox[{"m", "*", 
             SubscriptBox["m", "1"]}], ",", "x", ",", "i"}], "]"}], "\[And]", 
          
          RowBox[{"mult", "[", 
           RowBox[{
            RowBox[{"m", "*", 
             SubscriptBox["m", "2"]}], ",", "x", ",", "j"}], "]"}]}], ")"}], 
        "\[Or]", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"mult", "[", 
           RowBox[{
            RowBox[{"m", "*", 
             SubscriptBox["m", "1"]}], ",", "x", ",", "j"}], "]"}], "\[And]", 
          
          RowBox[{"mult", "[", 
           RowBox[{
            RowBox[{"m", "*", 
             SubscriptBox["m", "2"]}], ",", "x", ",", "i"}], "]"}]}], ")"}]}],
        ")"}]}]}]}], TraditionalForm]]],
 "\n\nBy a lemma we are allowed to replace (5) by\n(5\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mntcr", "[", 
    RowBox[{"x", ",", "y", ",", 
     SubscriptBox["a", "0"], ",", "1", ",", "1"}], "]"}], TraditionalForm]]],
 "\n\nWe proceed by instantiating (1) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["a", "0"], "\[Rule]", 
    SubscriptBox["a", "0"]}], TraditionalForm]]],
 ". All conditions of the universal quantifiers are satisfied (e.g. due to (5\
\[CloseCurlyQuote])), hence we know\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", 
    RowBox[{"x", ",", "y", ",", 
     SubscriptBox["a", "0"], ",", "1", ",", "1", ",", "B"}], "]"}], 
   TraditionalForm]]],
 "\n\nBy the definition of ",
 Cell[BoxData[
  FormBox["cpConnectible", TraditionalForm]]],
 " we also know\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"crit", "[", 
    RowBox[{
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "2"], ",", "x", ",", "y", ",", 
     SubscriptBox["a", "0"], ",", "1", ",", "1"}], "]"}], TraditionalForm]]],
 "\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areConnectibleMultipliers", "[", 
    RowBox[{
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "2"], ",", "x", ",", "y", ",", 
     SubscriptBox["a", "0"], ",", "B"}], "]"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "1"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 ".\n\nWe expand the definitions of ",
 Cell[BoxData[
  FormBox["areConnectibleMultipliers", TraditionalForm]]],
 ": We know\n(16)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["a", "0"], 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "1"], ",", "x"}]], 
    SubscriptBox["b", "1"]}], TraditionalForm]]],
 "\n(17)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["a", "0"], 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "2"], ",", "y"}]], 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 "\n(18)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "1"], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{
      SubscriptBox["a", "0"], ",", "\[Succeeds]"}]], 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "i"], ":=", 
    RowBox[{
     SubscriptBox["a", "0"], "-", 
     RowBox[{
      SubscriptBox["m", "i"], "*", 
      SubscriptBox[
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "i"]}]}]}], 
   TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "2"}], TraditionalForm]]],
 ".\n\nThe claim is now that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "*", 
    SubscriptBox["m", "1"]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "*", 
    SubscriptBox["m", "2"]}], TraditionalForm]]],
 " witness the existential quantifier in our goal (G). Hence, we need to show\
\n(G0)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areConnectibleMultipliers", "[", 
    RowBox[{
     RowBox[{"m", "*", 
      SubscriptBox["m", "1"]}], ",", 
     RowBox[{"m", "*", 
      SubscriptBox["m", "2"]}], ",", "x", ",", "y", ",", "a", ",", "B"}], 
    "]"}], TraditionalForm]]],
 "\n\nWe expand the defintion of ",
 Cell[BoxData[
  FormBox["areConnectibleMultipliers", TraditionalForm]]],
 ": We have to show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{
      RowBox[{"m", "*", 
       SubscriptBox["m", "1"]}], ",", "x"}]], 
    SubscriptBox["d", "1"]}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{
      RowBox[{"m", "*", 
       SubscriptBox["m", "2"]}], ",", "y"}]], 
    SubscriptBox["d", "2"]}], TraditionalForm]]],
 "\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["d", "1"], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    SubscriptBox["d", "2"]}], TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["d", "i"], ":=", 
    RowBox[{"a", "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"m", "*", 
        SubscriptBox["m", "i"]}], ")"}], "*", 
      SubscriptBox[
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "i"]}]}]}], 
   TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "2"}], TraditionalForm]]],
 ".\n\nApparently, (G1) and (G2) are symmetric, hence we only consider (G1) \
here. By the definition of ",
 Cell[BoxData[
  FormBox["\[Rule]", TraditionalForm]]],
 ", we have to show\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"m", "*", 
      SubscriptBox["m", "1"]}], ",", "x"}], "]"}], TraditionalForm]]],
 "\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"a", "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"m", "*", 
        SubscriptBox["m", "1"]}], ")"}], "*", "x"}]}]}], TraditionalForm]]],
 "\n\n(G4) is essentially what (6) states (with instantiation ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[Rule]", "x"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "0"], "\[Rule]", 
    SubscriptBox["m", "1"]}], TraditionalForm]]],
 "), and (G5) follows readily from (16) and (7), where (7) has to be \
instantiated by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      SubscriptBox["m", "1"]}], "*", "x"}]}], TraditionalForm]]],
 ".\n\nTherefore, our only goal left is (G3); This, however, can be reduced \
to the conjunction of (7), (8) and (18) due to (6), hence we are done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.609492881249335*^9, 3.609493202653509*^9}, {
   3.609493257209079*^9, 3.609493397656252*^9}, {3.609493429485306*^9, 
   3.609493976438841*^9}, {3.609494065955376*^9, 3.609494086418819*^9}, {
   3.609494137110226*^9, 3.60949422583189*^9}, {3.609494267980466*^9, 
   3.609494577703614*^9}, {3.609494635331513*^9, 3.609494751238149*^9}, {
   3.609495076896216*^9, 3.609495351570384*^9}, {3.609495401682084*^9, 
   3.609495586615428*^9}, {3.609495677622369*^9, 3.609496123642702*^9}, {
   3.60949616752472*^9, 3.609496245027909*^9}, {3.609496310612775*^9, 
   3.609496453430241*^9}, {3.609496514670814*^9, 3.609497035208311*^9}, {
   3.609497092415671*^9, 3.609497197223418*^9}, {3.609497317346815*^9, 
   3.609497432508119*^9}, {3.609497466272314*^9, 3.609498058207657*^9}, {
   3.609498351398719*^9, 3.609498408349308*^9}, {3.609498447981083*^9, 
   3.609498533422158*^9}, {3.60949862390127*^9, 3.609498771980914*^9}, {
   3.609498839301999*^9, 3.60949890127617*^9}, {3.609498966826312*^9, 
   3.609499236763304*^9}, {3.609499270276142*^9, 3.609499311547483*^9}, {
   3.609499346575889*^9, 3.609499466292665*^9}, {3.609499528992095*^9, 
   3.609499529686544*^9}, {3.609499590783359*^9, 3.609499600417835*^9}, {
   3.609499638744414*^9, 3.60949964143237*^9}, {3.609499783549394*^9, 
   3.609499846792675*^9}, {3.6094998858138*^9, 3.609499965650524*^9}, {
   3.609500002359006*^9, 3.609500079238862*^9}, {3.609500200634035*^9, 
   3.609500608879686*^9}, {3.609500984386071*^9, 3.609500992051743*^9}, {
   3.609573218502407*^9, 3.60957325054552*^9}, {3.609573341383826*^9, 
   3.609574059856372*^9}, {3.609574731585399*^9, 3.609574745921414*^9}, 
   3.609574780520196*^9, 3.63185535580896*^9, {3.632460897833385*^9, 
   3.6324609204507008`*^9}, 3.6324612062984667`*^9, {3.632476252979416*^9, 
   3.632476254261531*^9}, {3.632476345637094*^9, 3.632476357385746*^9}, {
   3.632476390041052*^9, 3.632476438507308*^9}, {3.63247916019976*^9, 
   3.6324792301971817`*^9}, {3.632479279368973*^9, 3.632479317217546*^9}, {
   3.6324794055752087`*^9, 3.632479423857926*^9}, {3.632479924808069*^9, 
   3.632479925276947*^9}, {3.632480000501033*^9, 3.632480046217318*^9}, {
   3.632547157670288*^9, 3.6325472254780912`*^9}, {3.632548014969655*^9, 
   3.632548127719287*^9}, 3.632548159466057*^9, {3.632548498183299*^9, 
   3.632548498183434*^9}, {3.632548734282445*^9, 3.63254876362295*^9}, {
   3.632548998192647*^9, 3.632549044360838*^9}, {3.632549079306175*^9, 
   3.632549109012985*^9}, {3.632549149742343*^9, 3.6325491589361153`*^9}, {
   3.632550095152235*^9, 3.632550141258449*^9}, {3.632550180028449*^9, 
   3.6325501926456823`*^9}, 3.636348622273727*^9, 3.6435337852924833`*^9, {
   3.644292754471496*^9, 3.644292811280076*^9}, {3.6442928844448633`*^9, 
   3.6442933615029993`*^9}, {3.644293422586814*^9, 3.644293457365366*^9}, {
   3.6442934988599586`*^9, 3.644293676959391*^9}, {3.6451628951970778`*^9, 
   3.645162895394463*^9}, {3.6451630209628153`*^9, 3.645163029411584*^9}, {
   3.645163128748662*^9, 3.645163169228498*^9}, {3.645163962509039*^9, 
   3.645164026282765*^9}, {3.6451640601155043`*^9, 3.6451640720483837`*^9}, {
   3.645176540143162*^9, 3.6451765654304543`*^9}, {3.645176633387868*^9, 
   3.645176647375485*^9}, {3.645176685111794*^9, 3.6451767477025547`*^9}, {
   3.6451767899416437`*^9, 3.6451768217051477`*^9}, {3.645180500979278*^9, 
   3.645180504115992*^9}, 3.645182819635435*^9, 3.6532195569890556`*^9, 
   3.6661743783771973`*^9, 3.666440946035557*^9},
 CellID->898592553],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.609574755207526*^9, 3.609574780520328*^9}, 
   3.6318553558090963`*^9, {3.632547210350286*^9, 3.6325472254782677`*^9}, {
   3.632549149742497*^9, 3.632549158936294*^9}, 3.6363486222738953`*^9, 
   3.643533785292655*^9, 3.6532195569890556`*^9, 3.6661743783771973`*^9, 
   3.666440946035666*^9},
 TextAlignment->Right,
 CellID->1138489307]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof of (71).",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.609492857396531*^9, 3.609492858255178*^9}, 
   3.609574780520083*^9, 3.631855355808807*^9, {3.632547138949471*^9, 
   3.632547168095376*^9}, {3.6325472103496437`*^9, 3.632547229671969*^9}, 
   3.6325472822689943`*^9, 3.63634862966368*^9, 3.643533792506871*^9, {
   3.645176163474268*^9, 3.645176163696541*^9}, 3.6532195497652426`*^9, 
   3.6661743718407855`*^9, 3.6664409401400414`*^9},
 CellID->703344814],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox["B", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " a.b.f.\nAssume\n(0)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[NotEqual]", "j"}], TraditionalForm]]],
 "\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"multn", "[", "c", "]"}]}]], 
     RowBox[{"i", "\[NotEqual]", "j"}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", 
         SubscriptBox["a", "0"], "]"}]}]], 
      RowBox[{"mntcr", "[", 
       RowBox[{"D", ",", "c", ",", "c", ",", "a", ",", "i", ",", "j"}], 
       "]"}]], 
     RowBox[{"cpConnectible", "[", 
      RowBox[{"c", ",", "c", ",", 
       SubscriptBox["a", "0"], ",", "i", ",", "j", ",", "B"}], "]"}]}]}], 
   TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"ntcr", "[", 
    RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\nand show\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"]}]], 
    RowBox[{"\[And]", 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{"mult", "[", 
          RowBox[{
           SubscriptBox["m", "1"], ",", "c", ",", "i"}], "]"}]},
        {
         RowBox[{"mult", "[", 
          RowBox[{
           SubscriptBox["m", "1"], ",", "c", ",", "j"}], "]"}]},
        {
         RowBox[{"areConnectibleMultipliers", "[", 
          RowBox[{
           SubscriptBox["m", "1"], ",", 
           SubscriptBox["m", "2"], ",", "c", ",", "c", ",", "a", ",", "B"}], 
          "]"}]}
       }]}]}]}], TraditionalForm]]],
 "\n\nWe now use (R7), immediately instantiating it with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c1", "\[Rule]", "c"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c2", "\[Rule]", "c"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Rule]", "a"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[Rule]", "i"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"j", "\[Rule]", "j"}], TraditionalForm]]],
 ". The premise of (R7) (",
 Cell[BoxData[
  FormBox[
   RowBox[{"ntcr", "[", 
    RowBox[{
     SubscriptBox["c", "1"], ",", 
     SubscriptBox["c", "2"], ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 ") is satisfied due to (2), hence we can add its conclusion to our knowledge \
base; For this, let ",
 Cell[BoxData[
  FormBox["k", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["l", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["a", "0"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["m", TraditionalForm]]],
 " witness the existential quantifier. We know\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", ",", 
    RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multn", "[", "c", "]"}]}], TraditionalForm]]],
 "\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"a", "\[Succeeds]", 
      SubscriptBox["a", "0"]}], "\[Or]", "a"}], "=", 
    SubscriptBox["a", "0"]}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mntcr", "[", 
    RowBox[{"c", ",", "c", ",", 
     SubscriptBox["a", "0"], ",", "k", ",", "l"}], "]"}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       SubscriptBox["m", "0"], ",", "c"}]], 
     RowBox[{"mult", "[", 
      RowBox[{
       SubscriptBox["m", "0"], ",", "c"}], "]"}]], 
    RowBox[{"mult", "[", 
     RowBox[{
      RowBox[{"m", "*", 
       SubscriptBox["m", "0"]}], ",", "c"}], "]"}]}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", "c"], 
     RowBox[{
      SubscriptBox["a", "0"], "\[Succeeds]", 
      RowBox[{
       SubscriptBox["a", "0"], "+", "c"}]}]], 
    RowBox[{"a", "\[Succeeds]", 
     RowBox[{"a", "+", 
      RowBox[{"m", "*", "c"}]}]}]}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"b", ",", "c", ",", "d"}]], 
     RowBox[{"b", 
      SubscriptBox["\[Rule]", "c"], "d"}]], 
    RowBox[{
     RowBox[{"m", "*", "b"}], 
     SubscriptBox["\[LeftRightArrow]", "c"], 
     RowBox[{"m", "*", "d"}]}]}], TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       SubscriptBox["m", "1"], ",", 
       SubscriptBox["m", "2"]}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"reducible", "[", 
        RowBox[{"D", ",", 
         SubscriptBox["a", "0"], ",", 
         SubscriptBox["c", "1"], ",", 
         SubscriptBox["m", "1"]}], "]"}], "\[And]", 
       RowBox[{"reducible", "[", 
        RowBox[{"D", ",", 
         SubscriptBox["a", "0"], ",", 
         SubscriptBox["c", "1"], ",", 
         SubscriptBox["m", "2"]}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         SubscriptBox["m", "1"], ",", 
         SubscriptBox["c", "1"], ",", "k"}], "]"}], "\[And]", 
       RowBox[{"mult", "[", 
        RowBox[{
         SubscriptBox["m", "2"], ",", 
         SubscriptBox["c", "1"], ",", "l"}], "]"}]}], ")"}]}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"mult", "[", 
         RowBox[{
          RowBox[{"m", "*", 
           SubscriptBox["m", "1"]}], ",", 
          SubscriptBox["c", "1"], ",", "i"}], "]"}], "\[And]", 
        RowBox[{"mult", "[", 
         RowBox[{
          RowBox[{"m", "*", 
           SubscriptBox["m", "2"]}], ",", 
          SubscriptBox["c", "1"], ",", "j"}], "]"}]}], ")"}], "\[Or]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"mult", "[", 
         RowBox[{
          RowBox[{"m", "*", 
           SubscriptBox["m", "1"]}], ",", 
          SubscriptBox["c", "1"], ",", "j"}], "]"}], "\[And]", 
        RowBox[{"mult", "[", 
         RowBox[{
          RowBox[{"m", "*", 
           SubscriptBox["m", "2"]}], ",", 
          SubscriptBox["c", "1"], ",", "i"}], "]"}]}], ")"}]}], ")"}]}], 
   TraditionalForm]]],
 "\n\nWe proceed by instantiating (1) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[Rule]", "k"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"j", "\[Rule]", "l"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["a", "0"], "\[Rule]", 
    SubscriptBox["a", "0"]}], TraditionalForm]]],
 ". All conditions of the universal quantifiers are satisfied (e.g. due to \
(5)), hence we know\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", 
    RowBox[{"c", ",", "c", ",", 
     SubscriptBox["a", "0"], ",", "k", ",", "l", ",", "B"}], "]"}], 
   TraditionalForm]]],
 "\n\nBy the definition of ",
 Cell[BoxData[
  FormBox["cpConnectible", TraditionalForm]]],
 " we also know\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"crit", "[", 
    RowBox[{
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "2"], ",", "c", ",", "c", ",", 
     SubscriptBox["a", "0"], ",", "k", ",", "l"}], "]"}], TraditionalForm]]],
 "\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areConnectibleMultipliers", "[", 
    RowBox[{
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "2"], ",", "c", ",", "c", ",", 
     SubscriptBox["a", "0"], ",", "B"}], "]"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "1"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 ".\n\nWe expand the definitions of ",
 Cell[BoxData[
  FormBox["areConnectibleMultipliers", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["crit", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["areIrrelative", TraditionalForm]]],
 ": We know\n(15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "\[NotEqual]", "l"}], TraditionalForm]]],
 "\n(16)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     SubscriptBox["m", "1"], ",", "c", ",", "k"}], "]"}], TraditionalForm]]],
 "\n(17)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     SubscriptBox["m", "2"], ",", "c", ",", "l"}], "]"}], TraditionalForm]]],
 "\n(18)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["a", "0"], 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "1"], ",", "c"}]], 
    SubscriptBox["b", "1"]}], TraditionalForm]]],
 "\n(19)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["a", "0"], 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "2"], ",", "c"}]], 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 "\n(20)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "1"], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{
      SubscriptBox["a", "0"], ",", "\[Succeeds]"}]], 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "i"], ":=", 
    RowBox[{
     SubscriptBox["a", "0"], "-", 
     RowBox[{
      SubscriptBox["m", "i"], "*", "c"}]}]}], TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "2"}], TraditionalForm]]],
 ".\n\nWe now instantiate (10) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "1"], ":=", 
    SubscriptBox["m", "1"]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "2"], ":=", 
    SubscriptBox["m", "2"]}], TraditionalForm]]],
 ", yielding\n(21)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"reducible", "[", 
       RowBox[{
        SubscriptBox["a", "0"], ",", "c", ",", 
        SubscriptBox["m", "1"]}], "]"}], "\[And]", 
      RowBox[{"reducible", "[", 
       RowBox[{
        SubscriptBox["a", "0"], ",", "c", ",", 
        SubscriptBox["m", "2"]}], "]"}], "\[And]", 
      RowBox[{"mult", "[", 
       RowBox[{
        SubscriptBox["m", "1"], ",", "c", ",", "k"}], "]"}], "\[And]", 
      RowBox[{"mult", "[", 
       RowBox[{
        SubscriptBox["m", "2"], ",", "c", ",", "l"}], "]"}]}], ")"}], 
    "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"mult", "[", 
         RowBox[{
          RowBox[{"m", "*", 
           SubscriptBox["m", "1"]}], ",", "c", ",", "i"}], "]"}], "\[And]", 
        RowBox[{"mult", "[", 
         RowBox[{
          RowBox[{"m", "*", 
           SubscriptBox["m", "2"]}], ",", "c", ",", "j"}], "]"}]}], ")"}], 
      "\[Or]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"mult", "[", 
         RowBox[{
          RowBox[{"m", "*", 
           SubscriptBox["m", "1"]}], ",", "c", ",", "j"}], "]"}], "\[And]", 
        RowBox[{"mult", "[", 
         RowBox[{
          RowBox[{"m", "*", 
           SubscriptBox["m", "2"]}], ",", "c", ",", "i"}], "]"}]}], ")"}]}], 
     ")"}]}], TraditionalForm]]],
 "\n\nAll conditions in (21) are easily seen to be satisfied (using (16), \
(17), (18) and (19)), hence we also know\n(22)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"mult", "[", 
       RowBox[{
        RowBox[{"m", "*", 
         SubscriptBox["m", "1"]}], ",", "c", ",", "i"}], "]"}], "\[And]", 
      RowBox[{"mult", "[", 
       RowBox[{
        RowBox[{"m", "*", 
         SubscriptBox["m", "2"]}], ",", "c", ",", "j"}], "]"}]}], ")"}], 
    "\[Or]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"mult", "[", 
       RowBox[{
        RowBox[{"m", "*", 
         SubscriptBox["m", "1"]}], ",", "c", ",", "j"}], "]"}], "\[And]", 
      RowBox[{"mult", "[", 
       RowBox[{
        RowBox[{"m", "*", 
         SubscriptBox["m", "2"]}], ",", "c", ",", "i"}], "]"}]}], ")"}]}], 
   TraditionalForm]]],
 "\nbased upon which we have to make a case distinction; However, since both \
cases are completely symmetric, we only deal with the first one, i.e. assume\n\
(22.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"mult", "[", 
     RowBox[{
      RowBox[{"m", "*", 
       SubscriptBox["m", "1"]}], ",", "c", ",", "i"}], "]"}], "\[And]", 
    RowBox[{"mult", "[", 
     RowBox[{
      RowBox[{"m", "*", 
       SubscriptBox["m", "2"]}], ",", "c", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 "\n\nThe claim is now that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "*", 
    SubscriptBox["m", "1"]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "*", 
    SubscriptBox["m", "2"]}], TraditionalForm]]],
 " witness the existential quantifier in our goal (G). Hence, we need to show \
all three conjuncts.\n\n",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"m", "*", 
      SubscriptBox["m", "1"]}], ",", "c", ",", "i"}], "]"}], 
   TraditionalForm]]],
 ":\nThis we know from (22.1).\n\n",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"m", "*", 
      SubscriptBox["m", "1"]}], ",", "c", ",", "j"}], "]"}], 
   TraditionalForm]]],
 ":\nThis we know from (22.1).\n\n",
 Cell[BoxData[
  FormBox[
   RowBox[{"areConnectibleMultipliers", "[", 
    RowBox[{
     RowBox[{"m", "*", 
      SubscriptBox["m", "1"]}], ",", 
     RowBox[{"m", "*", 
      SubscriptBox["m", "2"]}], ",", "c", ",", "c", ",", "a", ",", "B"}], 
    "]"}], TraditionalForm]]],
 ":\nWe expand the defintion of ",
 Cell[BoxData[
  FormBox["areConnectibleMultipliers", TraditionalForm]]],
 ": We have to show\n(G1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{
      RowBox[{"m", "*", 
       SubscriptBox["m", "1"]}], ",", "c"}]], 
    SubscriptBox["d", "1"]}], TraditionalForm]]],
 "\n(G2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{
      RowBox[{"m", "*", 
       SubscriptBox["m", "2"]}], ",", "c"}]], 
    SubscriptBox["d", "2"]}], TraditionalForm]]],
 "\n(G3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["d", "1"], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    SubscriptBox["d", "2"]}], TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["d", "i"], ":=", 
    RowBox[{"a", "-", 
     RowBox[{"m", "*", 
      SubscriptBox["m", "i"], "*", "c"}]}]}], TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "2"}], TraditionalForm]]],
 ".\n\nApparently, (G1) and (G2) are symmetric, hence we only consider (G1) \
here. By the definition of ",
 Cell[BoxData[
  FormBox["\[Rule]", TraditionalForm]]],
 ", we have to show\n(G4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     RowBox[{"m", "*", 
      SubscriptBox["m", "1"]}], ",", "c"}], "]"}], TraditionalForm]]],
 "\n(G5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Succeeds]", 
    RowBox[{"a", "-", 
     RowBox[{"m", "*", 
      SubscriptBox["m", "1"], "*", "c"}]}]}], TraditionalForm]]],
 "\n\n(G4) is contained in (21.1), and (G5) follows readily from (7) by (61).\
\n\nTherefore, our only goal left is (G3); This, however, can be reduced to \
the conjunction of (7), (8) and (20) by (60), hence we are done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.609492881249335*^9, 3.609493202653509*^9}, {
   3.609493257209079*^9, 3.609493397656252*^9}, {3.609493429485306*^9, 
   3.609493976438841*^9}, {3.609494065955376*^9, 3.609494086418819*^9}, {
   3.609494137110226*^9, 3.60949422583189*^9}, {3.609494267980466*^9, 
   3.609494577703614*^9}, {3.609494635331513*^9, 3.609494751238149*^9}, {
   3.609495076896216*^9, 3.609495351570384*^9}, {3.609495401682084*^9, 
   3.609495586615428*^9}, {3.609495677622369*^9, 3.609496123642702*^9}, {
   3.60949616752472*^9, 3.609496245027909*^9}, {3.609496310612775*^9, 
   3.609496453430241*^9}, {3.609496514670814*^9, 3.609497035208311*^9}, {
   3.609497092415671*^9, 3.609497197223418*^9}, {3.609497317346815*^9, 
   3.609497432508119*^9}, {3.609497466272314*^9, 3.609498058207657*^9}, {
   3.609498351398719*^9, 3.609498408349308*^9}, {3.609498447981083*^9, 
   3.609498533422158*^9}, {3.60949862390127*^9, 3.609498771980914*^9}, {
   3.609498839301999*^9, 3.60949890127617*^9}, {3.609498966826312*^9, 
   3.609499236763304*^9}, {3.609499270276142*^9, 3.609499311547483*^9}, {
   3.609499346575889*^9, 3.609499466292665*^9}, {3.609499528992095*^9, 
   3.609499529686544*^9}, {3.609499590783359*^9, 3.609499600417835*^9}, {
   3.609499638744414*^9, 3.60949964143237*^9}, {3.609499783549394*^9, 
   3.609499846792675*^9}, {3.6094998858138*^9, 3.609499965650524*^9}, {
   3.609500002359006*^9, 3.609500079238862*^9}, {3.609500200634035*^9, 
   3.609500608879686*^9}, {3.609500984386071*^9, 3.609500992051743*^9}, {
   3.609573218502407*^9, 3.60957325054552*^9}, {3.609573341383826*^9, 
   3.609574059856372*^9}, {3.609574731585399*^9, 3.609574745921414*^9}, 
   3.609574780520196*^9, 3.63185535580896*^9, {3.632460897833385*^9, 
   3.6324609204507008`*^9}, 3.6324612062984667`*^9, {3.632476252979416*^9, 
   3.632476254261531*^9}, {3.632476345637094*^9, 3.632476357385746*^9}, {
   3.632476390041052*^9, 3.632476438507308*^9}, {3.63247916019976*^9, 
   3.6324792301971817`*^9}, {3.632479279368973*^9, 3.632479317217546*^9}, {
   3.6324794055752087`*^9, 3.632479423857926*^9}, {3.632479924808069*^9, 
   3.632479925276947*^9}, {3.632480000501033*^9, 3.632480046217318*^9}, {
   3.632547157670288*^9, 3.632547263947474*^9}, {3.632547303728693*^9, 
   3.632547464645177*^9}, {3.632547496338318*^9, 3.6325478844342737`*^9}, {
   3.6325479152255363`*^9, 3.6325479898359413`*^9}, {3.632550285694743*^9, 
   3.63255040897618*^9}, {3.632550480606431*^9, 3.6325504809407063`*^9}, 
   3.636348629663843*^9, 3.643533792507041*^9, {3.6442936982161293`*^9, 
   3.644293716694718*^9}, {3.6442937536980667`*^9, 3.644294279965578*^9}, {
   3.645182991620075*^9, 3.645183040778173*^9}, {3.645183171388817*^9, 
   3.645183174375763*^9}, {3.645183618799361*^9, 3.645183635137362*^9}, 
   3.645242971760889*^9, 3.6532195497652426`*^9, 3.6661743718407855`*^9, 
   3.666440940140189*^9},
 CellID->2142178575],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.609574755207526*^9, 3.609574780520328*^9}, 
   3.6318553558090963`*^9, {3.6325472103499126`*^9, 3.632547218855269*^9}, 
   3.636348629664011*^9, 3.643533792507216*^9, 3.6532195497652426`*^9, 
   3.6661743718407855`*^9, 3.666440940140345*^9},
 TextAlignment->Right,
 CellID->1227483730]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1858807165],

Cell[CellGroupData[{

Cell["\<\
Theorem: Connectibility of Critical Pairs implies Local Connectivity for \
Identical Reductors\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.609485500422165*^9, 3.609485535813641*^9}, {
  3.648786323400309*^9, 3.648786324297182*^9}},
 CellID->1802962063],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"c", "\[Element]", "B"}]], 
   RowBox[{"c", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i", ",", 
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{
          UnderscriptBox["multN", "\[ScriptCapitalR]"], "[", "c", "]"}]}]], 
       RowBox[{"i", "\[NotEqual]", "j"}]], 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
          RowBox[{"[", "a0", "]"}]}]], 
        RowBox[{
         UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
         RowBox[{"c", ",", "c", ",", "a0", ",", "i", ",", "j"}], "]"}]], 
       RowBox[{
        UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
        RowBox[{"c", ",", "c", ",", "a0", ",", "i", ",", "j", ",", "B"}], 
        "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
         RowBox[{"[", 
          RowBox[{"a", ",", "b1", ",", "b2"}], "]"}]}]], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"a", 
          UnderscriptBox[
           SubscriptBox["\[Rule]", 
            RowBox[{"{", "c", "}"}]], "\[ScriptCapitalR]"], "b1"}], 
         TagBox[")",
          "AutoParentheses"]}], "\[And]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"a", 
          UnderscriptBox[
           SubscriptBox["\[Rule]", 
            RowBox[{"{", "c", "}"}]], "\[ScriptCapitalR]"], "b2"}], 
         TagBox[")",
          "AutoParentheses"]}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"b1", 
        UnderscriptBox[
         OverscriptBox[
          SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
          RowBox[{"a", ",", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            TagBox["\[Succeeds]",
             Identity], 
            TagBox[")",
             "AutoParentheses"]}]}]], "\[ScriptCapitalR]"], "b2"}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(72)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:25850532", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.609485663247545*^9, 3.609485719502153*^9}, {
  3.609485766378207*^9, 3.60948593556521*^9}, {3.610179608604836*^9, 
  3.610179630413638*^9}, {3.631855421604815*^9, 3.631855425602447*^9}, {
  3.631856460044816*^9, 3.6318564667240763`*^9}, {3.632479098427609*^9, 
  3.632479101394658*^9}, {3.644219938323515*^9, 3.644219977062068*^9}, {
  3.645256831052096*^9, 3.645256850520409*^9}, {3.661851351656275*^9, 
  3.6618513596902885`*^9}},
 CellTags->{
  "ID:25850532", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "72"},
 CellID->25850532],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1924586665]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1582609543],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(72)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(72)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                  StyleBox["j", "ExpressionVariable"]}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a0", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]], 
                 RowBox[{
                   UnderscriptBox["cpConnectible", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["b1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], 
                  UnderscriptBox[
                   SubscriptBox["\[Rule]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                   StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                  StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], 
                  UnderscriptBox[
                   SubscriptBox["\[Rule]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                   StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                  StyleBox["b2", "ExpressionVariable"]}]}]], 
             RowBox[{
               StyleBox["b1", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox[
               "b2", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                  StyleBox["j", "ExpressionVariable"]}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a0", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]], 
                 RowBox[{
                   UnderscriptBox["cpConnectible", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["b1", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], 
                  UnderscriptBox[
                   SubscriptBox["\[Rule]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                   StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                  StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], 
                  UnderscriptBox[
                   SubscriptBox["\[Rule]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                   StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                  StyleBox["b2", "ExpressionVariable"]}]}]], 
             RowBox[{
               StyleBox["b1", "ExpressionVariable"], 
               UnderscriptBox[
                OverscriptBox[
                 TemplateBox[{"\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"], "*"}, 
                  "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox[
               "b2", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:25850532"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p25850532-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p25850532-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:25850532", "Proof|ID:25850532-1"},
 CellID->193997362],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p25850532-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->311769453]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof.",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.609485958803153*^9, 3.609485959669075*^9}, 
   3.60948715303182*^9, 3.631855438053323*^9, 3.636348635884212*^9, 
   3.643533800181509*^9, 3.645245935118039*^9, 3.6532195419496293`*^9, 
   3.6661743637443714`*^9, 3.666440932530246*^9},
 CellID->2061042824],

Cell["Lemmas:", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632549811462339*^9, 3.632549827395687*^9}, 
   3.6363486358843737`*^9, 3.643533800181638*^9, 3.6532195419496293`*^9, 
   3.666174363759971*^9, 3.666440932530394*^9},
 CellID->90483196],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"\[Not]", 
    RowBox[{"areCorrelative", "[", 
     RowBox[{"c", ",", 
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"]}], "]"}]}], "\[Implies]", 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[Exists]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"multN", "[", "c", "]"}]}]], 
     RowBox[{"i", "\[NotEqual]", "j"}]], 
    RowBox[{"areIrrelative", "[", 
     RowBox[{"c", ",", "c", ",", 
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"], ",", "i", ",", "j"}], "]"}]}]}], 
  TraditionalForm]]]], "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632549818196032*^9, 3.632549853007621*^9}, {
   3.632550752015141*^9, 3.6325507670367613`*^9}, 3.63634863588452*^9, 
   3.643533800181735*^9, {3.645243340215629*^9, 3.6452433402282047`*^9}, {
   3.6452440386457663`*^9, 3.645244041102314*^9}, 3.6532195419496293`*^9, 
   3.666174363759971*^9, 3.666440932530511*^9},
 CellID->12644585],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"areIrrelative", "[", 
      RowBox[{"c", ",", "c", ",", 
       SubscriptBox["m", "1"], ",", 
       SubscriptBox["m", "2"], ",", "i", ",", "j"}], "]"}], "\[And]", 
     RowBox[{"areIrrelative", "[", 
      RowBox[{"c", ",", "c", ",", 
       SubscriptBox["m", "3"], ",", 
       SubscriptBox["m", "4"], ",", "i", ",", "j"}], "]"}]}], ")"}], 
   "\[Implies]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"areCorrelative", "[", 
      RowBox[{"c", ",", 
       SubscriptBox["m", "1"], ",", 
       SubscriptBox["m", "3"]}], "]"}], "\[And]", 
     RowBox[{"areCorrelative", "[", 
      RowBox[{"c", ",", 
       SubscriptBox["m", "2"], ",", 
       SubscriptBox["m", "4"]}], "]"}]}], ")"}]}], TraditionalForm]]]], "Item",\

 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.632549818196032*^9, 3.632549853007621*^9}, {
   3.632550752015141*^9, 3.6325507670367613`*^9}, 3.63634863588452*^9, 
   3.643533800181735*^9, {3.645243340215629*^9, 3.6452433971948223`*^9}, 
   3.6532195419496293`*^9, 3.666174363759971*^9, 3.66644093253059*^9},
 CellID->1914259349],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox["B", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[Element]", "B"}], TraditionalForm]]],
 " a.b.f. Assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"multn", "[", "c", "]"}]}]], 
     RowBox[{"i", "\[NotEqual]", "j"}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", 
         SubscriptBox["a", "0"], "]"}]}]], 
      RowBox[{"mntcr", "[", 
       RowBox[{"c", ",", "c", ",", 
        SubscriptBox["a", "0"], ",", "i", ",", "j"}], "]"}]], 
     RowBox[{"cpConnectible", "[", 
      RowBox[{"c", ",", "c", ",", 
       SubscriptBox["a", "0"], ",", "i", ",", "j", ",", "B"}], "]"}]}]}], 
   TraditionalForm]]],
 " and show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", 
        RowBox[{"a", ",", 
         SubscriptBox["b", "1"], ",", 
         SubscriptBox["b", "2"]}], "]"}]}]], 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"a", 
        SubscriptBox["\[Rule]", "c"], 
        SubscriptBox["b", "1"]}], ")"}], "\[And]", 
      RowBox[{"(", 
       RowBox[{"a", 
        SubscriptBox["\[Rule]", "c"], 
        SubscriptBox["b", "2"]}], ")"}]}]], 
    RowBox[{"(", 
     RowBox[{
      SubscriptBox["b", "1"], 
      OverscriptBox[
       SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
       RowBox[{"a", ",", "\[Succeeds]"}]], 
      SubscriptBox["b", "2"]}], ")"}]}], TraditionalForm]]],
 ". Hence, choose ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["b", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["b", "2"], TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", "c"], 
    SubscriptBox["b", "1"]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", "c"], 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 " and show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "1"], 
    OverscriptBox[
     SubsuperscriptBox["\[Rule]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 ".\nBy (reduction modulo element), we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", "m"], 
    RowBox[{"(", 
     RowBox[{"a", 
      SubscriptBox["\[Rule]", 
       RowBox[{"m", ",", "c"}]], 
      SubscriptBox["b", "1"]}], ")"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", "m"], 
    RowBox[{"(", 
     RowBox[{"a", 
      SubscriptBox["\[Rule]", 
       RowBox[{"m", ",", "c"}]], 
      SubscriptBox["b", "2"]}], ")"}]}], TraditionalForm]]],
 "; Let ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 " witness the two existential quantifiers, respectively. Summarizing, we \
know\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "1"], ",", "c"}]], 
    SubscriptBox["b", "1"]}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "2"], ",", "c"}]], 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"multn", "[", "c", "]"}]}]], 
     RowBox[{"i", "\[NotEqual]", "j"}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", 
         SubscriptBox["a", "0"], "]"}]}]], 
      RowBox[{"mntcr", "[", 
       RowBox[{"c", ",", "c", ",", 
        SubscriptBox["a", "0"], ",", "i", ",", "j"}], "]"}]], 
     RowBox[{"cpConnectible", "[", 
      RowBox[{"c", ",", "c", ",", 
       SubscriptBox["a", "0"], ",", "i", ",", "j", ",", "B"}], "]"}]}]}], 
   TraditionalForm]]],
 "\n\nWe make use of formula (41), instantiating it with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Rule]", "a"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[Rule]", "c"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b1", "\[Rule]", 
    SubscriptBox["b", "1"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b2", "\[Rule]", 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m1", "\[Rule]", 
    SubscriptBox["m", "1"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m2", "\[Rule]", 
    SubscriptBox["m", "2"]}], TraditionalForm]]],
 ". Apart from ",
 Cell[BoxData[
  FormBox[
   RowBox[{"correlative", "[", 
    RowBox[{"c", ",", 
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "2"]}], "]"}], TraditionalForm]]],
 ", all premises of the implication are satisfied. Hence, we distinguish two \
cases:\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"correlative", "[", 
    RowBox[{"c", ",", 
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "2"]}], "]"}], TraditionalForm]]],
 ".\nIn this case, all premises of the implication are satisfied, hence we \
can conclude ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "1"], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "c", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 ", from which ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "1"], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    SubscriptBox["b", "2"]}], TraditionalForm]]],
 " follows immediately by definition of reducibility modulo sets.\n\nCase II: \
",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"correlative", "[", 
     RowBox[{"c", ",", 
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"]}], "]"}]}], TraditionalForm]]],
 ".\nWe first note that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"correlative", "[", 
     RowBox[{"c", ",", 
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["m", "2"]}], "]"}]}], TraditionalForm]]],
 " immediately implies that there exist ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", "j"}], TraditionalForm]]],
 " with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"areIrrelative", "[", 
    RowBox[{"c", ",", "c", ",", 
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "2"], ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 ".\nBefore we continue, we show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCommonReducible", "[", 
    RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 ". By the definition of ",
 Cell[BoxData[
  FormBox["isCommonReducible", TraditionalForm]]],
 " we need to find ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "3"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "4"], TraditionalForm]]],
 " such that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"reducible", "[", 
    RowBox[{"a", ",", "c", ",", 
     SubscriptBox["m", "3"]}], "]"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"reducible", "[", 
    RowBox[{"a", ",", "c", ",", 
     SubscriptBox["m", "4"]}], "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"areIrrelative", "[", 
    RowBox[{"c", ",", "c", ",", 
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "2"], ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 ". ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 " are witnesses, as can be shown easily (combining (348) with (1), (2) for \
showing reducibility). Hence, we know\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCommonReducible", "[", 
    RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n\nOnce more we distinguish two cases:\n\nCase II.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"nctr", "[", 
     RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 ".\nBy the definition of ",
 Cell[BoxData[
  FormBox["nctr", TraditionalForm]]],
 " we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     RowBox[{"isCommonReducible", "[", 
      RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}]}], "\[Or]", 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       SubscriptBox["m", "3"], ",", 
       SubscriptBox["m", "4"]}]], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"areIrrelative", "[", 
           RowBox[{"c", ",", "c", ",", 
            SubscriptBox["m", "3"], ",", 
            SubscriptBox["m", "4"], ",", "i", ",", "j"}], "]"}]},
         {
          RowBox[{"reducible", "[", 
           RowBox[{"a", ",", "c", ",", 
            SubscriptBox["m", "3"]}], "]"}]},
         {
          RowBox[{"reducible", "[", 
           RowBox[{"a", ",", "c", ",", 
            SubscriptBox["m", "4"]}], "]"}]},
         {
          RowBox[{
           RowBox[{"reducible", "[", 
            RowBox[{
             RowBox[{"a", "-", 
              RowBox[{
               SubscriptBox["m", "4"], "*", "c"}]}], ",", "c", ",", 
             SubscriptBox["m", "3"]}], "]"}], "\[Or]", 
           RowBox[{"reducible", "[", 
            RowBox[{
             RowBox[{"a", "-", 
              RowBox[{
               SubscriptBox["m", "3"], "*", "c"}]}], ",", "c", ",", 
             SubscriptBox["m", "4"]}], "]"}]}]}
        }]}]}]}]}], TraditionalForm]]],
 ". The first possibility, i.e. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isCommonReducible", "[", 
     RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}]}], 
   TraditionalForm]]],
 ", contradicts (4), which means we can add\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areIrrelative", "[", 
    RowBox[{"c", ",", "c", ",", 
     SubscriptBox["m", "3"], ",", 
     SubscriptBox["m", "4"], ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"reducible", "[", 
    RowBox[{"a", ",", "c", ",", 
     SubscriptBox["m", "3"]}], "]"}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"reducible", "[", 
    RowBox[{"a", ",", "c", ",", 
     SubscriptBox["m", "4"]}], "]"}], TraditionalForm]]],
 "\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"reducible", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{
       SubscriptBox["m", "4"], "*", "c"}]}], ",", "c", ",", 
     SubscriptBox["m", "3"]}], "]"}], TraditionalForm]]],
 "\nto our KB, for some ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "3"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "4"], TraditionalForm]]],
 " (case ",
 Cell[BoxData[
  FormBox[
   RowBox[{"reducible", "[", 
    RowBox[{
     RowBox[{"a", "-", 
      RowBox[{
       SubscriptBox["m", "3"], "*", "c"}]}], ",", "c", ",", 
     SubscriptBox["m", "4"]}], "]"}], TraditionalForm]]],
 " works analoguously).\n\nBy (58) we know that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "-", 
    RowBox[{
     SubscriptBox["m", "3"], "*", "c"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "-", 
    RowBox[{
     SubscriptBox["m", "4"], "*", "c"}]}], TraditionalForm]]],
 " can be connected below ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 ", i.e.\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "3"], "*", "c"}]}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "4"], "*", "c"}]}]}], TraditionalForm]]],
 "\nExpanding the definition of areIrrelative we find\n(6.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[NotEqual]", "j"}], TraditionalForm]]],
 "\n(6.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     SubscriptBox["m", "1"], ",", "c", ",", "i"}], "]"}], TraditionalForm]]],
 "\n(6.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     SubscriptBox["m", "2"], ",", "c", ",", "j"}], "]"}], TraditionalForm]]],
 "\n(6.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     SubscriptBox["m", "3"], ",", "c", ",", "i"}], "]"}], TraditionalForm]]],
 "\n(6.5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     SubscriptBox["m", "4"], ",", "c", ",", "j"}], "]"}], TraditionalForm]]],
 "\nNow, since ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "3"], TraditionalForm]]],
 " are in the ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 "-th set of multipliers of ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "4"], TraditionalForm]]],
 " are in the ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 "-th one, we know by the definition of correlativity\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"correlative", "[", 
    RowBox[{"c", ",", 
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "3"]}], "]"}], TraditionalForm]]],
 "\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"correlative", "[", 
    RowBox[{"c", ",", 
     SubscriptBox["m", "2"], ",", 
     SubscriptBox["m", "4"]}], "]"}], TraditionalForm]]],
 "\n\nOur goal follows readily from (62), using (10), (1), (2), (7), (8), \
(11) and (12).\n\nCase II.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"nctr", "[", 
    RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 ".\nWe can infer\n(15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"multn", "[", "c", "]"}]}]], 
      RowBox[{"i", "\[NotEqual]", "j"}]], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "D"], 
         RowBox[{"[", "d", "]"}]}]], 
       RowBox[{"ntcr", "[", 
        RowBox[{"c", ",", "c", ",", "d", ",", "i", ",", "j"}], "]"}]], 
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", 
         RowBox[{
          SubscriptBox["m", "3"], ",", 
          SubscriptBox["m", "4"]}], "]"}]}]]}]}], "\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"mult", "[", 
         RowBox[{
          SubscriptBox["m", "3"], ",", "c", ",", "i"}], "]"}]},
       {
        RowBox[{"mult", "[", 
         RowBox[{
          SubscriptBox["m", "4"], ",", "c", ",", "j"}], "]"}]},
       {
        RowBox[{"areConnectibleMultipliers", "[", 
         RowBox[{
          SubscriptBox["m", "3"], ",", 
          SubscriptBox["m", "4"], ",", "c", ",", "c", ",", "d", ",", "B"}], 
         "]"}]}
      }]}]}], TraditionalForm]]],
 "\nfrom (71) and (3) by modus ponens, instantiating formula (71) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", ":=", "c"}], TraditionalForm]]],
 ".\n(15) can now be instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[Rule]", "i"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"j", "\[Rule]", "j"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"d", "\[Rule]", "a"}], TraditionalForm]]],
 ". Since we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{"ntcr", "[", 
    RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 " by case assumption and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[NotEqual]", "j"}], TraditionalForm]]],
 " after expanding the definition of irrelativity, we can add ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", 
       RowBox[{
        SubscriptBox["m", "3"], ",", 
        SubscriptBox["m", "4"]}], "]"}]}]], 
    RowBox[{"\[And]", 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{"mult", "[", 
          RowBox[{
           SubscriptBox["m", "3"], ",", "c", ",", "i"}], "]"}]},
        {
         RowBox[{"mult", "[", 
          RowBox[{
           SubscriptBox["m", "4"], ",", "c", ",", "j"}], "]"}]},
        {
         RowBox[{"areConnectibleMultipliers", "[", 
          RowBox[{
           SubscriptBox["m", "3"], ",", 
           SubscriptBox["m", "4"], ",", "c", ",", "c", ",", "a", ",", "B"}], 
          "]"}]}
       }]}]}]}], TraditionalForm]]],
 " to our assumptions. Let ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "3"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "4"], TraditionalForm]]],
 " witness the existential quantifier, i.e. we know\n(16)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     SubscriptBox["m", "3"], ",", "c", ",", "i"}], "]"}], TraditionalForm]]],
 "\n(17)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"mult", "[", 
    RowBox[{
     SubscriptBox["m", "4"], ",", "c", ",", "j"}], "]"}], TraditionalForm]]],
 "\n(18)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areConnectibleMultipliers", "[", 
    RowBox[{
     SubscriptBox["m", "3"], ",", 
     SubscriptBox["m", "4"], ",", "c", ",", "c", ",", "a", ",", "B"}], "]"}], 
   TraditionalForm]]],
 "\nBy the definition of ",
 Cell[BoxData[
  FormBox["areConnectibleMultipliers", TraditionalForm]]],
 " we know that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "-", 
    RowBox[{
     SubscriptBox["m", "3"], "*", "c"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "-", 
    RowBox[{
     SubscriptBox["m", "4"], "*", "c"}]}], TraditionalForm]]],
 " can be connected below ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 ", i.e. we can add\n(19)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "3"], "*", "c"}]}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "4"], "*", "c"}]}]}], TraditionalForm]]],
 "\n(20)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "3"], ",", "c"}]], 
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "3"], "*", "c"}]}]}], TraditionalForm]]],
 "\n(21)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "4"], ",", "c"}]], 
    RowBox[{"a", "-", 
     RowBox[{
      SubscriptBox["m", "4"], "*", "c"}]}]}], TraditionalForm]]],
 "\nto our knowledge base.\nNow, since ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "3"], TraditionalForm]]],
 " are in the ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 "-th set of multipliers of ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "4"], TraditionalForm]]],
 " are in the ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 "-th one, we know by the definition of correlativity\n(22)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"correlative", "[", 
    RowBox[{"c", ",", "c", ",", 
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "3"]}], "]"}], TraditionalForm]]],
 "\n(23)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"correlative", "[", 
    RowBox[{"c", ",", "c", ",", 
     SubscriptBox["m", "2"], ",", 
     SubscriptBox["m", "4"]}], "]"}], TraditionalForm]]],
 "\n\nOur goal follows readily from (62), using (19), (1), (2), (20), (21), \
(22) and (23)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.60948598258579*^9, 3.609486222752996*^9}, {
   3.609486256751746*^9, 3.609486690080711*^9}, {3.609486729644683*^9, 
   3.609486825018282*^9}, {3.609486910340111*^9, 3.609487153031921*^9}, 
   3.6318554380534678`*^9, {3.632461982528225*^9, 3.6324619825283422`*^9}, {
   3.6324621652795362`*^9, 3.6324621956201982`*^9}, {3.6324625150789213`*^9, 
   3.6324625481242113`*^9}, {3.632474959841681*^9, 3.632474961734055*^9}, {
   3.6324752897221203`*^9, 3.63247531902878*^9}, {3.632475435463279*^9, 
   3.632475447945763*^9}, {3.632475564712613*^9, 3.632475652083592*^9}, {
   3.6324801599789333`*^9, 3.632480204322928*^9}, {3.63248025452542*^9, 
   3.632480261799259*^9}, {3.632480420888702*^9, 3.632480428068445*^9}, {
   3.632480559442445*^9, 3.6324805672621*^9}, {3.632484600870614*^9, 
   3.632484736698492*^9}, {3.632544863960347*^9, 3.63254487187309*^9}, {
   3.632549219129923*^9, 3.632549266643896*^9}, {3.632549821690621*^9, 
   3.632549827396283*^9}, {3.632550675734023*^9, 3.632550725435441*^9}, {
   3.632550776861061*^9, 3.632550855569936*^9}, {3.6325509059719877`*^9, 
   3.632551004235876*^9}, {3.6325510401584663`*^9, 3.632551053796678*^9}, 
   3.636348635884653*^9, 3.64353380018184*^9, {3.644294371731668*^9, 
   3.6442944601120033`*^9}, {3.6442944961134253`*^9, 3.644294558768428*^9}, {
   3.644294597119069*^9, 3.644294713358568*^9}, {3.645243077229806*^9, 
   3.6452430786232033`*^9}, 3.645243159286134*^9, {3.645243314154153*^9, 
   3.645243314941424*^9}, {3.645243429350292*^9, 3.645243474617082*^9}, {
   3.64524351132961*^9, 3.645243512138908*^9}, {3.645243994100499*^9, 
   3.645244003330036*^9}, {3.645245954210828*^9, 3.645246048822051*^9}, {
   3.645246098264195*^9, 3.645246159205056*^9}, 3.6532195419496293`*^9, 
   3.666174363759971*^9, 3.6664409325307093`*^9},
 CellID->198422590],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.609487137429664*^9, 3.609487153032031*^9}, 
   3.6318554380536003`*^9, {3.6325498216907597`*^9, 3.632549827396413*^9}, 
   3.636348635884817*^9, 3.643533800181966*^9, 3.6532195419496293`*^9, 
   3.666174363759971*^9, 3.666440932530856*^9},
 TextAlignment->Right,
 CellID->157546407]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->101882599],

Cell[CellGroupData[{

Cell["\<\
Theorem: Connectibility of Critical Pairs of Elements Implies Local \
Connectivity of Elements\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648786347119828*^9, 3.6487863768080053`*^9}, {
  3.648786543457686*^9, 3.6487865506688347`*^9}},
 CellID->1722461897],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"c", "\[Element]", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isLocConnective", "\[ScriptCapitalR]"], "[", 
     RowBox[{"B", ",", "c", ",", "c"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    
    RowBox[{
     UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
     RowBox[{"B", ",", "c", ",", "c"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(74)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:482796083", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648783149188162*^9, 3.648783194730699*^9}, {
   3.648783774371482*^9, 3.648783778650284*^9}, {3.6487838129320593`*^9, 
   3.648783814478704*^9}, 3.648783954572627*^9, {3.648785174924623*^9, 
   3.648785188800498*^9}},
 CellTags->{
  "ID:482796083", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "74"},
 CellID->482796083],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"c1", ",", 
    RowBox[{"c2", "\[Element]", "B"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
       RowBox[{"B", ",", "c1", ",", "c2"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
       RowBox[{"B", ",", "c1", ",", "c1"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
       RowBox[{"B", ",", "c2", ",", "c2"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["isLocConnective", "\[ScriptCapitalR]"], "[", 
     RowBox[{"B", ",", "c1", ",", "c2"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(75)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1705968105", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648786080320129*^9, 3.6487861186907377`*^9}, {
   3.648786195577942*^9, 3.6487862072668123`*^9}, 3.6487862491732407`*^9, {
   3.6487866430329323`*^9, 3.648786643290903*^9}},
 CellTags->{
  "ID:1705968105", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "75"},
 CellID->1705968105],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->83095137]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->692583585],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(74)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(74)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:482796083"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p482796083-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p482796083-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:482796083", "Proof|ID:482796083-1"},
 CellID->185660128],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p482796083-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->133502169]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(75)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(75)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}, {
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}, {
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c1", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1705968105"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1705968105-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1705968105-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1705968105", "Proof|ID:1705968105-1"},
 CellID->213359418],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1705968105-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.662267728024934*^9, 3.662267729912537*^9}},
 CellID->587543446]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof of (75).",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.608525151326765*^9, 3.608525152790262*^9}, 
   3.609492730616474*^9, 3.631855448027813*^9, 3.6363486423863907`*^9, 
   3.643533806663478*^9, {3.648807720921487*^9, 3.648807727897448*^9}, 
   3.6532195313104105`*^9, 3.666174354041154*^9, 3.666440924380459*^9},
 CellID->1204032442],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox["B", TraditionalForm]]],
 " a.b.f. and assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", "B", "]"}], TraditionalForm]]],
 ". We have to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isLocConnective", "[", "B", "]"}], TraditionalForm]]],
 ". For the sake of simplicity, we omit the subscript \
\[OpenCurlyDoubleQuote]B\[CloseCurlyDoubleQuote] of \[Rule] in the sequel.\n\
By the definition of ",
 Cell[BoxData[
  FormBox["isLocConnective", TraditionalForm]]],
 ", we have to show now ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"a", ",", "b", ",", "c"}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"c", "\[Rule]", 
        RowBox[{
         RowBox[{"a", "\[And]", "c"}], "\[Rule]", "b"}]}], ")"}], 
      "\[Implies]", 
      RowBox[{"(", 
       RowBox[{"a", 
        OverscriptBox[
         SubsuperscriptBox["\[Rule]", "B", "*"], 
         RowBox[{"c", ",", "\[Succeeds]"}]], "b"}], ")"}]}], ")"}]}], 
   TraditionalForm]]],
 ". Hence, choose ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " a.b.f., assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[Rule]", "a"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "\[Rule]", "b"}], TraditionalForm]]],
 " and show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    OverscriptBox[
     SubsuperscriptBox["\[Rule]", "B", "*"], 
     RowBox[{"c", ",", "\[Succeeds]"}]], "b"}], TraditionalForm]]],
 ".\nBy (reduction modulo set), we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", 
    SubscriptBox["\[Rule]", 
     SubscriptBox["c", "1"]], "a"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", 
    SubscriptBox["\[Rule]", 
     SubscriptBox["c", "2"]], "b"}], TraditionalForm]]],
 ", with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["c", "1"], ",", 
    RowBox[{
     SubscriptBox["c", "2"], "\[Element]", "B"}]}], TraditionalForm]]],
 ", and in particular we also know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["c", "1"], "\[NotEqual]", "0"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["c", "2"], "\[NotEqual]", "0"}], TraditionalForm]]],
 ".\nBy (reduction modulo element), we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", "m"], 
    RowBox[{"(", 
     RowBox[{"c", 
      SubscriptBox["\[Rule]", 
       RowBox[{"m", ",", 
        SubscriptBox["c", "1"]}]], "a"}], ")"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", "m"], 
    RowBox[{"(", 
     RowBox[{"c", 
      SubscriptBox["\[Rule]", 
       RowBox[{"m", ",", 
        SubscriptBox["c", "1"]}]], "a"}], ")"}]}], TraditionalForm]]],
 "; Let ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 " witness the two existential quantifiers, respectively. Summarizing, we \
know\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "1"], ",", 
      SubscriptBox["c", "1"]}]], "a"}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", 
    SubscriptBox["\[Rule]", 
     RowBox[{
      SubscriptBox["m", "2"], ",", 
      SubscriptBox["c", "2"]}]], "b"}], TraditionalForm]]],
 "\n\nFirst, we expand the definition of ",
 Cell[BoxData[
  FormBox["cpConnectible", TraditionalForm]]],
 ", instantiating the first universal quantifier immediately with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", ":=", 
    SubscriptBox["c", "1"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"y", ":=", 
    SubscriptBox["c", "2"]}], TraditionalForm]]],
 ": We know\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"multn", "[", 
        SubscriptBox["c", "1"], "]"}]}]], 
     RowBox[{
      SubscriptBox["c", "1"], "=", 
      RowBox[{
       SubscriptBox["c", "2"], "\[Implies]", 
       RowBox[{"i", "\[NotEqual]", "j"}]}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", "d", "]"}]}]], 
      RowBox[{"mntcr", "[", 
       RowBox[{
        SubscriptBox["c", "1"], ",", 
        SubscriptBox["c", "2"], ",", "d", ",", "i", ",", "j"}], "]"}]], 
     RowBox[{"cpConnectible", "[", 
      RowBox[{
       SubscriptBox["c", "1"], ",", 
       SubscriptBox["c", "2"], ",", "d", ",", "i", ",", "j", ",", "B"}], 
      "]"}]}]}], TraditionalForm]]],
 "\nHowever, we can instantiate the first universal quantifier with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", ":=", 
    SubscriptBox["c", "1"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"y", ":=", 
    SubscriptBox["c", "1"]}], TraditionalForm]]],
 " and with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", ":=", 
    SubscriptBox["c", "2"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"y", ":=", 
    SubscriptBox["c", "2"]}], TraditionalForm]]],
 ", too, and obtain\n(3\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"multn", "[", 
        SubscriptBox["c", "1"], "]"}]}]], 
     RowBox[{
      SubscriptBox["c", "1"], "=", 
      RowBox[{
       SubscriptBox["c", "1"], "\[Implies]", 
       RowBox[{"i", "\[NotEqual]", "j"}]}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", "d", "]"}]}]], 
      RowBox[{"mntcr", "[", 
       RowBox[{
        SubscriptBox["c", "1"], ",", 
        SubscriptBox["c", "1"], ",", "d", ",", "i", ",", "j"}], "]"}]], 
     RowBox[{"cpConnectible", "[", 
      RowBox[{
       SubscriptBox["c", "1"], ",", 
       SubscriptBox["c", "1"], ",", "d", ",", "i", ",", "j", ",", "B"}], 
      "]"}]}]}], TraditionalForm]]],
 "\n(3\[CloseCurlyQuote]\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"multn", "[", 
        SubscriptBox["c", "2"], "]"}]}]], 
     RowBox[{
      SubscriptBox["c", "2"], "=", 
      RowBox[{
       SubscriptBox["c", "2"], "\[Implies]", 
       RowBox[{"i", "\[NotEqual]", "j"}]}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "D"], 
        RowBox[{"[", "d", "]"}]}]], 
      RowBox[{"mntcr", "[", 
       RowBox[{
        SubscriptBox["c", "2"], ",", 
        SubscriptBox["c", "2"], ",", "d", ",", "i", ",", "j"}], "]"}]], 
     RowBox[{"cpConnectible", "[", 
      RowBox[{
       SubscriptBox["c", "2"], ",", 
       SubscriptBox["c", "2"], ",", "d", ",", "i", ",", "j", ",", "B"}], 
      "]"}]}]}], TraditionalForm]]],
 "\nWe make use of formula (72) twice, instantiating it once with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", ":=", 
    SubscriptBox["c", "1"]}], TraditionalForm]]],
 " and once with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", ":=", 
    SubscriptBox["c", "2"]}], TraditionalForm]]],
 ", leading eventually to\n(4\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"a", ",", 
       SubscriptBox["b", "1"], ",", 
       SubscriptBox["b", "2"]}]], 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"a", 
        SubscriptBox["\[Rule]", 
         SubscriptBox["c", "1"]], 
        SubscriptBox["b", "1"]}], ")"}], "\[And]", 
      RowBox[{"(", 
       RowBox[{"a", 
        SubscriptBox["\[Rule]", 
         SubscriptBox["c", "1"]], 
        SubscriptBox["b", "2"]}], ")"}]}]], 
    RowBox[{"(", 
     RowBox[{
      SubscriptBox["b", "1"], 
      OverscriptBox[
       SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
       RowBox[{"a", ",", "\[Succeeds]"}]], 
      SubscriptBox["b", "2"]}], ")"}]}], TraditionalForm]]],
 "\n(4\[CloseCurlyQuote]\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"a", ",", 
       SubscriptBox["b", "1"], ",", 
       SubscriptBox["b", "2"]}]], 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"a", 
        SubscriptBox["\[Rule]", 
         SubscriptBox["c", "2"]], 
        SubscriptBox["b", "1"]}], ")"}], "\[And]", 
      RowBox[{"(", 
       RowBox[{"a", 
        SubscriptBox["\[Rule]", 
         SubscriptBox["c", "2"]], 
        SubscriptBox["b", "2"]}], ")"}]}]], 
    RowBox[{"(", 
     RowBox[{
      SubscriptBox["b", "1"], 
      OverscriptBox[
       SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
       RowBox[{"a", ",", "\[Succeeds]"}]], 
      SubscriptBox["b", "2"]}], ")"}]}], TraditionalForm]]],
 "\nsince the premises of the big implications in (72) can easily be shown to \
be satisfied from (3\[CloseCurlyQuote]) and (3\[CloseCurlyQuote]\
\[CloseCurlyQuote]).\n\nWe now distinguish two cases:\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["c", "1"], "=", 
    SubscriptBox["c", "2"]}], TraditionalForm]]],
 ".\nSince ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["c", "1"], "=", 
    SubscriptBox["c", "2"]}], TraditionalForm]]],
 ", we can now instantiate either (4\[CloseCurlyQuote]) or (4\
\[CloseCurlyQuote]\[CloseCurlyQuote]) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", ":=", "c"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "1"], ":=", "a"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "2"], ":=", "b"}], TraditionalForm]]],
 ", leaving us with\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"c", 
        SubscriptBox["\[Rule]", 
         SubscriptBox["c", "1"]], "a"}], ")"}], "\[And]", 
      RowBox[{"(", 
       RowBox[{"c", 
        SubscriptBox["\[Rule]", 
         SubscriptBox["c", "1"]], "b"}], ")"}]}], "\[Implies]", "a"}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"c", ",", "\[Succeeds]"}]], "b"}], TraditionalForm]]],
 "\nThe two premises of (5) are satisfied due to (1) and (2), hence we are \
done.\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["c", "1"], "\[NotEqual]", 
    SubscriptBox["c", "2"]}], TraditionalForm]]],
 ".\nFirst of all, we certainly know ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCommonReducible", "[", 
    RowBox[{
     SubscriptBox["c", "1"], ",", 
     SubscriptBox["c", "2"], ",", "c", ",", "1", ",", "1"}], "]"}], 
   TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 " witness the existential quantifier in the definition of ",
 Cell[BoxData[
  FormBox["isCommonReducible", TraditionalForm]]],
 " (and the indices are irrelevant anyway).\nWe proceed by distinguishing two \
cases:\n\nCase II.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"ntcr", "[", 
     RowBox[{
      SubscriptBox["c", "1"], ",", 
      SubscriptBox["c", "2"], ",", "c", ",", "1", ",", "1"}], "]"}]}], 
   TraditionalForm]]],
 " (note that the indices are again completely irrelevant).\nIn this case, by \
definition of ",
 Cell[BoxData[
  FormBox["ntcr", TraditionalForm]]],
 ", we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     RowBox[{"isCommonReducible", "[", 
      RowBox[{
       SubscriptBox["c", "1"], ",", 
       SubscriptBox["c", "2"], ",", "c", ",", "1", ",", "1"}], "]"}]}], 
    "\[Or]", 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       SubscriptBox["m", "3"], ",", 
       SubscriptBox["m", "4"]}]], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"areIrrelative", "[", 
           RowBox[{
            SubscriptBox["c", "1"], ",", 
            SubscriptBox["c", "2"], ",", 
            SubscriptBox["m", "3"], ",", 
            SubscriptBox["m", "4"], ",", "1", ",", "1"}], "]"}]},
         {
          RowBox[{"reducible", "[", 
           RowBox[{"c", ",", 
            SubscriptBox["c", "1"], ",", 
            SubscriptBox["m", "3"]}], "]"}]},
         {
          RowBox[{"reducible", "[", 
           RowBox[{"c", ",", 
            SubscriptBox["c", "2"], ",", 
            SubscriptBox["m", "4"]}], "]"}]},
         {
          RowBox[{
           RowBox[{"reducible", "[", 
            RowBox[{
             RowBox[{"c", "-", 
              RowBox[{
               SubscriptBox["m", "4"], "*", 
               SubscriptBox["c", "2"]}]}], ",", 
             SubscriptBox["c", "1"], ",", 
             SubscriptBox["m", "3"]}], "]"}], "\[Or]", 
           RowBox[{"reducible", "[", 
            RowBox[{
             RowBox[{"c", "-", 
              RowBox[{
               SubscriptBox["m", "3"], "*", 
               SubscriptBox["c", "1"]}]}], ",", 
             SubscriptBox["c", "2"], ",", 
             SubscriptBox["m", "4"]}], "]"}]}]}
        }]}]}]}]}], TraditionalForm]]],
 ". The first possibility, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"isCommonReducible", "[", "...", "]"}]}], TraditionalForm]]],
 ", we have already ruled out, hence we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      SubscriptBox["m", "3"], ",", 
      SubscriptBox["m", "4"]}]], 
    RowBox[{"\[And]", 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{"areIrrelative", "[", "...", "]"}]},
        {
         RowBox[{"reducible", "[", 
          RowBox[{"c", ",", 
           SubscriptBox["c", "1"], ",", 
           SubscriptBox["m", "3"]}], "]"}]},
        {
         RowBox[{"reducible", "[", 
          RowBox[{"c", ",", 
           SubscriptBox["c", "2"], ",", 
           SubscriptBox["m", "4"]}], "]"}]},
        {
         RowBox[{
          RowBox[{"reducible", "[", 
           RowBox[{
            RowBox[{"c", "-", 
             RowBox[{
              SubscriptBox["m", "4"], "*", 
              SubscriptBox["c", "2"]}]}], ",", 
            SubscriptBox["c", "1"], ",", 
            SubscriptBox["m", "3"]}], "]"}], "\[Or]", 
          RowBox[{"reducible", "[", 
           RowBox[{
            RowBox[{"c", "-", 
             RowBox[{
              SubscriptBox["m", "3"], "*", 
              SubscriptBox["c", "1"]}]}], ",", 
            SubscriptBox["c", "2"], ",", 
            SubscriptBox["m", "4"]}], "]"}]}]}
       }]}]}]}], TraditionalForm]]],
 ". Let ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "3"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "4"], TraditionalForm]]],
 " witness the existential quantifier, meaning that we can add\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"reducible", "[", 
    RowBox[{"c", ",", 
     SubscriptBox["c", "1"], ",", 
     SubscriptBox["m", "3"]}], "]"}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"reducible", "[", 
    RowBox[{"c", ",", 
     SubscriptBox["c", "2"], ",", 
     SubscriptBox["m", "4"]}], "]"}], TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"reducible", "[", 
     RowBox[{
      RowBox[{"c", "-", 
       RowBox[{
        SubscriptBox["m", "4"], "*", 
        SubscriptBox["c", "2"]}]}], ",", 
      SubscriptBox["c", "1"], ",", 
      SubscriptBox["m", "3"]}], "]"}], "\[Or]", 
    RowBox[{"reducible", "[", 
     RowBox[{
      RowBox[{"c", "-", 
       RowBox[{
        SubscriptBox["m", "3"], "*", 
        SubscriptBox["c", "1"]}]}], ",", 
      SubscriptBox["c", "2"], ",", 
      SubscriptBox["m", "4"]}], "]"}]}], TraditionalForm]]],
 "\nWe now use formula (58), instantiating it with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", ":=", 
     SubscriptBox["c", "1"]}], ",", " ", 
    RowBox[{"y", ":=", 
     SubscriptBox["c", "2"]}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", ":=", "c"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b1", ":=", 
    RowBox[{"c", "-", 
     RowBox[{
      SubscriptBox["m", "3"], "*", 
      SubscriptBox["c", "1"]}]}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b2", ":=", 
    RowBox[{"c", "-", 
     RowBox[{
      SubscriptBox["m", "4"], "*", 
      SubscriptBox["c", "2"]}]}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m1", ":=", 
    SubscriptBox["m", "3"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m2", ":=", 
    SubscriptBox["m", "4"]}], TraditionalForm]]],
 ". Since all premises of the big implication in (58) are satisfied, we can \
add its conclusion to our knowledge base, i.e.\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c", "-", 
     RowBox[{
      SubscriptBox["m", "3"], "*", 
      SubscriptBox["c", "1"]}]}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"c", ",", "\[Succeeds]"}]], 
    RowBox[{"c", "-", 
     RowBox[{
      SubscriptBox["m", "4"], "*", 
      SubscriptBox["c", "2"]}]}]}], TraditionalForm]]],
 "\nWe instantiate (4\[CloseCurlyQuote]) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", ":=", "c"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "1"], ":=", "a"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "2"], ":=", 
    RowBox[{"c", "-", 
     RowBox[{
      SubscriptBox["m", "3"], "*", 
      SubscriptBox["c", "1"]}]}]}], TraditionalForm]]],
 " and (4\[CloseCurlyQuote]\[CloseCurlyQuote]) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", ":=", "c"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "1"], ":=", "b"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["b", "2"], ":=", 
    RowBox[{"c", "-", 
     RowBox[{
      SubscriptBox["m", "4"], "*", 
      SubscriptBox["c", "2"]}]}]}], TraditionalForm]]],
 ", i.e. we have\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"c", 
        SubscriptBox["\[Rule]", 
         SubscriptBox["c", "1"]], "a"}], ")"}], "\[And]", 
      RowBox[{"(", 
       RowBox[{"c", 
        SubscriptBox["\[Rule]", 
         SubscriptBox["c", "1"]], 
        RowBox[{"c", "-", 
         RowBox[{
          SubscriptBox["m", "3"], "*", 
          SubscriptBox["c", "1"]}]}]}], ")"}]}], "\[Implies]", "a"}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"c", ",", "\[Succeeds]"}]], 
    RowBox[{"c", "-", 
     RowBox[{
      SubscriptBox["m", "3"], "*", 
      SubscriptBox["c", "1"]}]}]}], TraditionalForm]]],
 "\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"c", 
        SubscriptBox["\[Rule]", 
         SubscriptBox["c", "2"]], "b"}], ")"}], "\[And]", 
      RowBox[{"(", 
       RowBox[{"c", 
        SubscriptBox["\[Rule]", 
         SubscriptBox["c", "2"]], 
        RowBox[{"c", "-", 
         RowBox[{
          SubscriptBox["m", "4"], "*", 
          SubscriptBox["c", "2"]}]}]}], ")"}]}], "\[Implies]", "b"}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"c", ",", "\[Succeeds]"}]], 
    RowBox[{"c", "-", 
     RowBox[{
      SubscriptBox["m", "4"], "*", 
      SubscriptBox["c", "2"]}]}]}], TraditionalForm]]],
 "\nin our knowledge base. All conjuncts in the premises are satisfied (due \
to (1), (2), (6) and (7)), thus we know\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"c", ",", "\[Succeeds]"}]], 
    RowBox[{"c", "-", 
     RowBox[{
      SubscriptBox["m", "3"], "*", 
      SubscriptBox["c", "1"]}]}]}], TraditionalForm]]],
 "\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"c", ",", "\[Succeeds]"}]], 
    RowBox[{"c", "-", 
     RowBox[{
      SubscriptBox["m", "4"], "*", 
      SubscriptBox["c", "2"]}]}]}], TraditionalForm]]],
 "\nas well. Combining (9), (12) and (13) yields the desired result of \
connectibility of ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " below ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " modulo ",
 Cell[BoxData[
  FormBox["B", TraditionalForm]]],
 ".\n\nCase II.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"ntcr", "[", 
    RowBox[{
     SubscriptBox["c", "1"], ",", 
     SubscriptBox["c", "2"], ",", "c", ",", "1", ",", "1"}], "]"}], 
   TraditionalForm]]],
 ".\nWe can make use of formula (70) and infer\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", "d", "]"}]}]], 
     RowBox[{"ntcr", "[", 
      RowBox[{
       SubscriptBox["c", "1"], ",", 
       SubscriptBox["c", "2"], ",", "d", ",", "1", ",", "1"}], "]"}]], 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "D"], 
       RowBox[{"[", 
        RowBox[{
         SubscriptBox["m", "3"], ",", 
         SubscriptBox["m", "4"]}], "]"}]}]], 
     RowBox[{"areConnectibleMultipliers", "[", 
      RowBox[{
       SubscriptBox["m", "3"], ",", 
       SubscriptBox["m", "4"], ",", 
       SubscriptBox["c", "1"], ",", 
       SubscriptBox["c", "2"], ",", "d", ",", "B"}], "]"}]}]}], 
   TraditionalForm]]],
 "\nfrom (3) by modus ponens, instantiating formula (70) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", ":=", 
     SubscriptBox["c", "1"]}], ",", 
    RowBox[{"y", ":=", 
     SubscriptBox["c", "2"]}]}], TraditionalForm]]],
 ".\n(14) can now be instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"d", ":=", "c"}], TraditionalForm]]],
 ". Since we know ",
 Cell[BoxData[
  FormBox[
   RowBox[{"ntcr", "[", 
    RowBox[{
     SubscriptBox["c", "1"], ",", 
     SubscriptBox["c", "2"], ",", "c", ",", "1", ",", "1"}], "]"}], 
   TraditionalForm]]],
 " by case assumption, we can add ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "D"], 
      RowBox[{"[", 
       RowBox[{
        SubscriptBox["m", "3"], ",", 
        SubscriptBox["m", "4"]}], "]"}]}]], 
    RowBox[{"areConnectibleMultipliers", "[", 
     RowBox[{
      SubscriptBox["m", "3"], ",", 
      SubscriptBox["m", "4"], ",", 
      SubscriptBox["c", "1"], ",", 
      SubscriptBox["c", "2"], ",", "c", ",", "B"}], "]"}]}], 
   TraditionalForm]]],
 " to our assumptions. Let ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "3"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "4"], TraditionalForm]]],
 " witness the existential quantifier, i.e. we know\n(16)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areConnectibleMultipliers", "[", 
    RowBox[{
     SubscriptBox["m", "3"], ",", 
     SubscriptBox["m", "4"], ",", 
     SubscriptBox["c", "1"], ",", 
     SubscriptBox["c", "2"], ",", "c", ",", "B"}], "]"}], TraditionalForm]]],
 "\nBy the definition of ",
 Cell[BoxData[
  FormBox["areConnectibleMultipliers", TraditionalForm]]],
 " we know that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "-", 
    RowBox[{
     SubscriptBox["m", "3"], "*", 
     SubscriptBox["c", "1"]}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", "-", 
    RowBox[{
     SubscriptBox["m", "4"], "*", 
     SubscriptBox["c", "2"]}]}], TraditionalForm]]],
 " can be connected below ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ", i.e. we can add\n(17)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c", "-", 
     RowBox[{
      SubscriptBox["m", "3"], "*", 
      SubscriptBox["c", "1"]}]}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"c", ",", "\[Succeeds]"}]], 
    RowBox[{"c", "-", 
     RowBox[{
      SubscriptBox["m", "4"], "*", 
      SubscriptBox["c", "2"]}]}]}], TraditionalForm]]],
 "\nto our knowledge base.\nThe rest proceeds exactly as in case II.I as soon \
as (9) is known, thus we are done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.608525163344286*^9, 3.608525202256269*^9}, {
   3.60852523513643*^9, 3.608525255751141*^9}, {3.608525386332471*^9, 
   3.608525598587808*^9}, {3.608525643956193*^9, 3.608525652628124*^9}, {
   3.608525891999012*^9, 3.608526281231496*^9}, {3.608526311997548*^9, 
   3.608526330817861*^9}, {3.608526398485264*^9, 3.608526398485438*^9}, {
   3.608527778889347*^9, 3.608527787992726*^9}, {3.608527884659008*^9, 
   3.608528056597865*^9}, {3.60852811192433*^9, 3.608528235349684*^9}, {
   3.60852828877954*^9, 3.608528321235743*^9}, {3.608528352146947*^9, 
   3.608528457326803*^9}, {3.608528489923475*^9, 3.608528499202816*^9}, {
   3.608528788289008*^9, 3.608528795724232*^9}, {3.608529061287981*^9, 
   3.608529069041922*^9}, {3.60852912160266*^9, 3.608529140113116*^9}, 
   3.608529535759137*^9, {3.608529576343785*^9, 3.608529706693043*^9}, {
   3.608529740070774*^9, 3.608529789536727*^9}, {3.608529836504648*^9, 
   3.608530005447282*^9}, {3.608530659784572*^9, 3.608530726922415*^9}, {
   3.608530770886455*^9, 3.60853111356927*^9}, {3.608531193662118*^9, 
   3.608531220481087*^9}, {3.608606437682805*^9, 3.608606509078994*^9}, {
   3.608606541011218*^9, 3.608606643316565*^9}, {3.608607424560159*^9, 
   3.608607566764496*^9}, {3.608607721728073*^9, 3.608607737297143*^9}, {
   3.608607770521681*^9, 3.608607804778792*^9}, {3.608608039837279*^9, 
   3.608608049262189*^9}, {3.608608834176046*^9, 3.608609125476056*^9}, {
   3.608609379896516*^9, 3.608609494081631*^9}, {3.608609551290945*^9, 
   3.608609573040375*^9}, {3.608610568576126*^9, 3.608610751032357*^9}, {
   3.608610802389129*^9, 3.608610918186326*^9}, {3.60906601818316*^9, 
   3.60906604021962*^9}, {3.60906607218868*^9, 3.609066113137354*^9}, {
   3.609066148432245*^9, 3.609066222733594*^9}, {3.609066270653821*^9, 
   3.609066315529028*^9}, {3.609066368461537*^9, 3.60906646747266*^9}, {
   3.609066503743737*^9, 3.60906669461845*^9}, {3.609066746931012*^9, 
   3.60906682712251*^9}, {3.609066864125906*^9, 3.609066865378663*^9}, {
   3.609069497844344*^9, 3.609069671456152*^9}, {3.609069716468328*^9, 
   3.609069736035155*^9}, 3.609477123388082*^9, 3.609477198255694*^9, {
   3.609477229331603*^9, 3.609477519919023*^9}, {3.609477554158482*^9, 
   3.609477608735441*^9}, 3.60947768115583*^9, {3.609477769527192*^9, 
   3.609477987045298*^9}, {3.609478035642977*^9, 3.609478146940503*^9}, {
   3.609478295004066*^9, 3.609478343069737*^9}, {3.609478397207935*^9, 
   3.609478403599231*^9}, 3.609478442730506*^9, {3.609478724063622*^9, 
   3.609478814564377*^9}, {3.609478887002537*^9, 3.609478983745727*^9}, {
   3.609479084518111*^9, 3.60947918225346*^9}, {3.60947922281026*^9, 
   3.609479281371024*^9}, {3.609480550275254*^9, 3.609481131930635*^9}, {
   3.609484778788066*^9, 3.609484857366912*^9}, {3.609484891969851*^9, 
   3.609484905426947*^9}, {3.609484937232971*^9, 3.60948500356897*^9}, {
   3.609485059654837*^9, 3.609485059961912*^9}, {3.609485096941625*^9, 
   3.609485116830107*^9}, {3.60948515196609*^9, 3.609485294345208*^9}, {
   3.609485335948068*^9, 3.609485442319882*^9}, {3.60948724292315*^9, 
   3.609487353931591*^9}, {3.609487389739224*^9, 3.609487560427502*^9}, {
   3.60949059750069*^9, 3.609490932532901*^9}, {3.609491068360177*^9, 
   3.609491110508372*^9}, {3.609491254957695*^9, 3.609491312706917*^9}, {
   3.609491644985981*^9, 3.609491842420794*^9}, {3.609491920063437*^9, 
   3.609491989211551*^9}, {3.609492026694017*^9, 3.609492135207219*^9}, {
   3.609492200025441*^9, 3.609492324788045*^9}, {3.609492362620282*^9, 
   3.609492383988325*^9}, {3.609492504708243*^9, 3.609492698124413*^9}, 
   3.609492730616595*^9, 3.619346719289905*^9, 3.619346855893109*^9, 
   3.63185544802794*^9, {3.6324758712647047`*^9, 3.6324758827665358`*^9}, {
   3.632475944478734*^9, 3.632475999086583*^9}, {3.632476088717634*^9, 
   3.632476090020506*^9}, {3.63248068178648*^9, 3.632480699142459*^9}, {
   3.632480782043922*^9, 3.632480796194367*^9}, {3.632480837492083*^9, 
   3.632480896171796*^9}, {3.632544957556408*^9, 3.632544995587831*^9}, {
   3.632545049986578*^9, 3.632545079663294*^9}, {3.63254932724446*^9, 
   3.632549363835721*^9}, {3.632551166968207*^9, 3.63255118577696*^9}, {
   3.632551247000531*^9, 3.6325512889093313`*^9}, {3.6325513728903303`*^9, 
   3.632551414337984*^9}, 3.6363486423865147`*^9, 3.6435338066636057`*^9, {
   3.644294771371665*^9, 3.6442951853885517`*^9}, {3.644295217382614*^9, 
   3.644295239911045*^9}, {3.6451779821058407`*^9, 3.6451780236490717`*^9}, {
   3.645244173882185*^9, 3.645244179870386*^9}, {3.645244254531755*^9, 
   3.645244259869225*^9}, {3.6452499032208*^9, 3.645249916094036*^9}, 
   3.6532195313104105`*^9, 3.666174354041154*^9, 3.6664409243806047`*^9},
 CellID->966828484],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.609492713412147*^9, 3.609492730616706*^9}, 
   3.63185544802807*^9, 3.636348642386643*^9, 3.643533806663734*^9, 
   3.6532195313104105`*^9, 3.666174354041154*^9, 3.666440924380765*^9},
 TextAlignment->Right,
 CellID->1036621313]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1963125342],

Cell[CellGroupData[{

Cell["\<\
Theorem: Connectibility of Critical Pairs implies Local Connectivity\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.59202527474528*^9, 3.592025301287124*^9}, {
  3.609477071347477*^9, 3.609477072352198*^9}, {3.648786327796562*^9, 
  3.648786328701846*^9}},
 CellID->605912543],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", "B", "]"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["isLocConnective", "\[ScriptCapitalR]"], "[", "B", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(73)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:68280999", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.592024882012984*^9, 3.592024911419674*^9}, {
  3.592024959360075*^9, 3.592024960437832*^9}, {3.592025200770875*^9, 
  3.592025207293871*^9}, {3.592025250929708*^9, 3.592025251744111*^9}, {
  3.59618755113178*^9, 3.596187555573447*^9}, {3.596190990947268*^9, 
  3.596190997152608*^9}, {3.596191040134509*^9, 3.596191041610946*^9}, {
  3.607337872863609*^9, 3.607337873896804*^9}, {3.610177242887529*^9, 
  3.610177254201937*^9}, {3.644219981611466*^9, 3.6442199876132*^9}},
 CellTags->{
  "ID:68280999", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "73"},
 CellID->68280999],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1912768342]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->503215071],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(73)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(73)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:68280999"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p68280999-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p68280999-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:68280999", "Proof|ID:68280999-1"},
 CellID->20432459],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p68280999-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->296170017]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->989485028],

Cell[CellGroupData[{

Cell["Theorem: Main Theorem of Reduction Ring Theory", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.591602068850706*^9, 3.591602086234761*^9}, {
   3.634885948725609*^9, 3.634885951493058*^9}, 3.6442200618549*^9},
 CellID->1258498493],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", "B", "]"}], 
  "\[Implies]", 
  RowBox[{
   UnderscriptBox["isGroebnerBasis", "\[ScriptCapitalR]"], "[", "B", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Main Theorem)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1188710397", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmViYGAQBmIQfW/CmsiCQ68cFxzOigHR/mcep4DomHL7QhCttuJq
KYj2ZPaqANFWCxnrQfT6jr9gevkm3RYQ/eBUeR+IPqRmNhlEW6/sBdMHXd7P
ANGLT8jNBtG3GJetBtHiKZ4bwOIxbgyFQDqtxYgFRC/gtM4QP/zK0eZLZxaI
nrV5dymIVnJurgTRrl5urSBa981tMK2Z/X0RiE76rr4CRG8VexpoB6QVI5yi
QfS60Cn9IHrrEpGJIDr4HtNUEO0aJgum92R1rDA4/cpRIdh3JYhmjiktKn7w
ynHT00lgWmv6XO8zaq8dq3LSfUA0AJuXnoA=
  "],
 CellTags->{
  "ID:1188710397", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Main Theorem"},
 CellID->1188710397],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->379716283]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->301018529],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Main Theorem)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Main Theorem)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isGroebnerBasis", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isGroebnerBasis", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1188710397"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1188710397-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1188710397-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1188710397", "Proof|ID:1188710397-1"},
 CellID->240809659],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1188710397-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->304808930]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Tuples (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6488117819065943`*^9, 3.648811790835144*^9}, 
   3.648812332113*^9, {3.648875515442995*^9, 3.648875519038961*^9}},
 CellID->590486522],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
   RowBox[{"[", "T", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 78242703}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648811798648703*^9, 3.648811810539957*^9}},
 CellTags->{
  "ID:78242703", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->78242703],

Cell["", "OpenEnvironment",
 CellID->418348591],

Cell[CellGroupData[{

Cell["Corollary", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.648811813441168*^9},
 CellID->2145168465],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
       RowBox[{"T", ",", 
        SubscriptBox["T", "i"], ",", 
        SubscriptBox["T", "j"]}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
       RowBox[{"T", ",", 
        SubscriptBox["T", "i"], ",", 
        SubscriptBox["T", "i"]}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
       RowBox[{"T", ",", 
        SubscriptBox["T", "j"], ",", 
        SubscriptBox["T", "j"]}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["isLocConnective", "\[ScriptCapitalR]"], "[", 
     RowBox[{"T", ",", 
      SubscriptBox["T", "i"], ",", 
      SubscriptBox["T", "j"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(75 tuples)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:807712598", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648811819293186*^9, 3.648811861689715*^9}},
 CellTags->{
  "ID:807712598", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "75 tuples"},
 CellID->807712598],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{
            SubscriptBox["T", "i"], "\[Equal]", 
            SubscriptBox["T", "j"]}], "\[Implies]", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               RowBox[{"k", ",", 
                RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
                RowBox[{
                 UnderscriptBox["multN", "\[ScriptCapitalR]"], "[", 
                 SubscriptBox["T", "j"], "]"}]}]], 
              RowBox[{"l", ">", "k"}]], 
             RowBox[{
              UnderscriptBox[
               UnderscriptBox["\[ForAll]", 
                RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
                 RowBox[{"[", "a", "]"}]}]], 
               RowBox[{
                UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
                RowBox[{
                 SubscriptBox["T", "j"], ",", 
                 SubscriptBox["T", "j"], ",", "a", ",", "k", ",", "l"}], 
                "]"}]], 
              RowBox[{
               UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
               RowBox[{
                SubscriptBox["T", "j"], ",", 
                SubscriptBox["T", "j"], ",", "a", ",", "k", ",", "l", ",", 
                "T"}], "]"}]}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           RowBox[{
            SubscriptBox["T", "i"], "\[NotEqual]", 
            SubscriptBox["T", "j"]}], "\[Implies]", 
           RowBox[{
            UnderscriptBox[
             UnderscriptBox["\[ForAll]", 
              RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
               RowBox[{"[", "a", "]"}]}]], 
             RowBox[{
              UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
              RowBox[{
               SubscriptBox["T", "i"], ",", 
               SubscriptBox["T", "j"], ",", "a", ",", "1", ",", "1"}], "]"}]], 
            RowBox[{
             UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
             RowBox[{
              SubscriptBox["T", "i"], ",", 
              SubscriptBox["T", "j"], ",", "a", ",", "1", ",", "1", ",", 
              "T"}], "]"}]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
     RowBox[{"T", ",", 
      SubscriptBox["T", "i"], ",", 
      SubscriptBox["T", "j"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(76 tuples)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:645630189", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648812757363781*^9, 3.648812978524632*^9}, {
   3.6488132668251038`*^9, 3.6488132670085793`*^9}, 3.648813687545493*^9},
 CellTags->{
  "ID:645630189", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "76 tuples"},
 CellID->645630189],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
     RowBox[{"j", "\[GreaterEqual]", "i"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         SubscriptBox["T", "i"], "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
        RowBox[{
         SubscriptBox["T", "j"], "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{
       UnderscriptBox["cpConnectible", "\[ScriptCapitalR]"], "[", 
       RowBox[{"T", ",", 
        SubscriptBox["T", "i"], ",", 
        SubscriptBox["T", "j"]}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isGroebnerBasis", "\[ScriptCapitalR]"], "[", "T", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(77 tuples)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:47992769", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648812655999613*^9, 3.648812695106764*^9}, {
  3.648814090736171*^9, 3.6488141217263107`*^9}},
 CellTags->{
  "ID:47992769", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "77 tuples"},
 CellID->47992769],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1427261493]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->391776269],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(75 tuples)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(75 tuples)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]], ",", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isLocConnective", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:807712598"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p807712598-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p807712598-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:807712598", "Proof|ID:807712598-1"},
 CellID->1012531353],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p807712598-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->2136404136]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(76 tuples)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(76 tuples)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], "\[Equal]", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], "]"}]}]}}]], 
                  RowBox[{
                    StyleBox["l", "ExpressionVariable"], ">", 
                    StyleBox["k", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], "]"}]], 
                   RowBox[{
                    UnderscriptBox["cpConnectible", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"]}], "]"}]}]}]}], 
             "\[And]", 
             RowBox[{
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}],
                     "]"}]], 
                 RowBox[{
                   UnderscriptBox["cpConnectible", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1", 
                    ",", 
                    StyleBox["T", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], "\[Equal]", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], "]"}]}]}}]], 
                  RowBox[{
                    StyleBox["l", "ExpressionVariable"], ">", 
                    StyleBox["k", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], "]"}]], 
                   RowBox[{
                    UnderscriptBox["cpConnectible", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"]}], "]"}]}]}]}], 
             "\[And]", 
             RowBox[{
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}],
                     "]"}]], 
                 RowBox[{
                   UnderscriptBox["cpConnectible", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1", 
                    ",", 
                    StyleBox["T", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:645630189"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p645630189-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p645630189-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:645630189", "Proof|ID:645630189-1"},
 CellID->54547686],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p645630189-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->266126597]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(77 tuples)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(77 tuples)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                 "\[And]", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isGroebnerBasis", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["T", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                 "\[And]", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                   SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isGroebnerBasis", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["T", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:47992769"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p47992769-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p47992769-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:47992769", "Proof|ID:47992769-1"},
 CellID->163065910],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p47992769-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->186918277]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Ideal Congruence and Ideals (\[Checkmark])\
\>", "Section",
 CellChangeTimes->{{3.645258979723836*^9, 3.6452590053480587`*^9}, {
  3.645259113666915*^9, 3.645259117881371*^9}, {3.645441065177082*^9, 
  3.645441069055419*^9}, {3.6479278086590776`*^9, 3.647927816337384*^9}, {
  3.647927878666066*^9, 3.647927882115107*^9}, {3.647934610678729*^9, 
  3.647934614745158*^9}, {3.647935308325343*^9, 3.647935309374259*^9}, {
  3.648282098700719*^9, 3.6482821010820847`*^9}},
 CellID->1635043122],

Cell[CellGroupData[{

Cell["Definitions", "Subsection",
 CellChangeTimes->{{3.647927862250345*^9, 3.647927865952888*^9}},
 CellID->364187549],

Cell["", "OpenEnvironment",
 CellID->443755902],

Cell[CellGroupData[{

Cell["Definition: Ideal Congruence and Ideals", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.647927871110114*^9, 3.6479278762345343`*^9}},
 CellID->1034407747],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "\[ScriptCapitalD]"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 41552115}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6479278918584967`*^9, 3.647927935285117*^9}, 
   3.647928100569601*^9, {3.6479288064327374`*^9, 3.6479288066539507`*^9}},
 CellTags->{
  "ID:41552115", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->41552115],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "\[ScriptCapitalD]", "]"}]], 
   RowBox[{"[", "B", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 208902120}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647928812571294*^9, 3.647928824523672*^9}},
 CellTags->{
  "ID:208902120", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->208902120],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[Congruent]", "B"], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] is congruence modulo the ideal generated by ",
 Cell[BoxData[
  FormBox["B", TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.6452586931713943`*^9, 3.6452587386828203`*^9}, {
  3.6479288288434563`*^9, 3.647928836674823*^9}},
 CellID->128327410],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"a", ",", "b"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalD]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
        RowBox[{"[", 
         RowBox[{"T", ",", "S"}], "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
             "\[Equal]", 
             RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]},
           {
            RowBox[{
             UnderscriptBox["\[ForAll]", 
              RowBox[{
               RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
               RowBox[{
               "\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               SubscriptBox["S", "i"], "\[Element]", "B"}], 
              TagBox[")",
               "AutoParentheses"]}]}]},
           {
            RowBox[{"b", "\[Equal]", 
             RowBox[{"a", 
              UnderscriptBox["-", "\[ScriptCapitalD]"], 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{
                UnderscriptBox[
                 SubscriptBox["\[Sum]", "\[ScriptCapitalD]"], 
                 RowBox[{
                  RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                  RowBox[{
                  "\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{
                  SubscriptBox["T", "i"], 
                  UnderscriptBox["*", "\[ScriptCapitalD]"], 
                  SubscriptBox["S", "i"]}], 
                 TagBox[")",
                  "AutoParentheses"]}]}], 
               TagBox[")",
                "AutoParentheses"]}]}]}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal congruence)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:531785290", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645258454802672*^9, 3.6452586053643713`*^9}, {
  3.647927936840008*^9, 3.647927954536989*^9}, {3.647928094813027*^9, 
  3.647928108030772*^9}, {3.6479288546434603`*^9, 3.647928857845682*^9}},
 CellTags->{
  "ID:531785290", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal congruence"},
 CellID->531785290],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"ideal", "[", "B", "]"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] is the ideal generated by ",
 Cell[BoxData[
  FormBox["B", TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.647927976135975*^9, 3.647927995450892*^9}, {
  3.64792884158218*^9, 3.647928851522018*^9}},
 CellID->1504386559],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["ideal", "\[ScriptCapitalD]"], "[", "B", "]"}], ":=", 
  RowBox[{"{", 
   RowBox[{"a", 
    UnderscriptBox["|", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
      RowBox[{"[", "a", "]"}]}]], 
    RowBox[{"a", 
     UnderscriptBox[
      SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalD]"], 
     UnderscriptBox["0", "\[ScriptCapitalD]"]}]}], 
   "}"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1814440310", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6479279625488653`*^9, 3.647927967278615*^9}, {
  3.647927999366683*^9, 3.647928050407905*^9}, {3.647928102782452*^9, 
  3.647928104405575*^9}, {3.647928861642435*^9, 3.647928880916065*^9}, {
  3.6479351149686604`*^9, 3.647935116749711*^9}},
 CellTags->{
  "ID:1814440310", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal"},
 CellID->1814440310],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "T", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["ideal", "\[ScriptCapitalD]"], "[", "T", "]"}], ":=", 
    RowBox[{
     UnderscriptBox["ideal", "\[ScriptCapitalD]"], "[", 
     RowBox[{"TupleToSet", "[", "T", "]"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal of tuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1625946592", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6479358217763968`*^9, 3.647935866092308*^9}},
 CellTags->{
  "ID:1625946592", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal of tuple"},
 CellID->1625946592],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->919758187]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->695510085],

Cell[TextData[{
 StyleBox["Remark",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 ":\nIt might appear a bit strange that we ",
 StyleBox["first",
  FontSlant->"Italic"],
 " define ideal congruence and ",
 StyleBox["afterwards",
  FontSlant->"Italic"],
 " ideals in terms of ideal congruence, and not vice versa. The reason is \
that reasoning about predicates is in general easier, since no knowledge \
about \[Element] is required. Hence, first defining predicates and then \
defining sets is more efficient, since proved properties about the predicates \
in most of the cases can be readily transferred to sets."
}], "Text",
 CellChangeTimes->{{3.648188653654048*^9, 3.648188795635096*^9}, {
  3.6481888675816793`*^9, 3.648188907320019*^9}},
 CellID->2078293890]
}, Closed]],

Cell[CellGroupData[{

Cell["Theorems: Ideal Congruence (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6479282466826553`*^9, 3.64792824973837*^9}, {
  3.6479282893081017`*^9, 3.647928307494182*^9}, {3.648187539531179*^9, 
  3.6481875404150677`*^9}, {3.648209293892871*^9, 3.648209296574914*^9}},
 CellID->409334904],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isReductionRing", "[", "\[ScriptCapitalR]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1782924340}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645260258826631*^9, 3.645260265141487*^9}},
 CellTags->{
  "ID:1782924340", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1782924340],

Cell["", "Text",
 CellID->671758383],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "\[ScriptCapitalR]", "]"}]], 
   RowBox[{"[", "B", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1482545820}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64526029820712*^9, 3.645260314482059*^9}},
 CellTags->{
  "ID:1482545820", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1482545820],

Cell[CellGroupData[{

Cell["\[Congruent] is Equivalence Relation (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.645260060986298*^9, 3.645260077613949*^9}, {
  3.6453425640524597`*^9, 3.64534256829277*^9}},
 CellID->384143515],

Cell["", "OpenEnvironment",
 CellID->2006713273],

Cell[CellGroupData[{

Cell["Lemma: Sum over Joined Tuples", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645335273068054*^9, 3.645335281856287*^9}},
 CellID->640772553],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
    RowBox[{"[", 
     RowBox[{"T1", ",", "S1", ",", "T2", ",", "S2"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "T1", "\[RightBracketingBar]"}], 
       "\[Equal]", 
       RowBox[{"\[LeftBracketingBar]", "S1", "\[RightBracketingBar]"}]}], 
      "\[And]", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "T2", "\[RightBracketingBar]"}], 
       "\[Equal]", 
       RowBox[{"\[LeftBracketingBar]", "S2", "\[RightBracketingBar]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox[
         SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", 
           RowBox[{"S1", 
            TagBox["\:22c8",
             Identity,
             SyntaxForm->"a*b"], "S2"}], "\[RightBracketingBar]"}]}]], 
        RowBox[{
         SubscriptBox[
          RowBox[{"(", 
           RowBox[{"T1", 
            TagBox["\:22c8",
             Identity,
             SyntaxForm->"a*b"], "T2"}], ")"}], "i"], 
         UnderscriptBox["*", "\[ScriptCapitalR]"], 
         SubscriptBox[
          RowBox[{"(", 
           RowBox[{"S1", 
            TagBox["\:22c8",
             Identity,
             SyntaxForm->"a*b"], "S2"}], ")"}], "i"]}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Equal]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          UnderscriptBox[
           SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
           RowBox[{
            RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
            RowBox[{
            "\[LeftBracketingBar]", "S1", "\[RightBracketingBar]"}]}]], 
          RowBox[{
           SubscriptBox["T1", "i"], 
           UnderscriptBox["*", "\[ScriptCapitalR]"], 
           SubscriptBox["S1", "i"]}]}], 
         TagBox[")",
          "AutoParentheses"]}], 
        UnderscriptBox["+", "\[ScriptCapitalR]"], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          UnderscriptBox[
           SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
           RowBox[{
            RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
            RowBox[{
            "\[LeftBracketingBar]", "S2", "\[RightBracketingBar]"}]}]], 
          RowBox[{
           SubscriptBox["T2", "i"], 
           UnderscriptBox["*", "\[ScriptCapitalR]"], 
           SubscriptBox["S2", "i"]}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(sum multiplication join)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1796414707", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6453352894241343`*^9, 3.64533530081089*^9}, {
  3.6453353486335697`*^9, 3.645335374499844*^9}, {3.645335487924062*^9, 
  3.645335655205167*^9}, {3.645340983532751*^9, 3.645340990826906*^9}, {
  3.64534115974128*^9, 3.6453412045754747`*^9}, {3.6453414186218224`*^9, 
  3.645341421710627*^9}},
 CellTags->{
  "ID:1796414707", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "sum multiplication join"},
 CellID->1796414707],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1852259974]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->401138290],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(sum multiplication join)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(sum multiplication join)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T1", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], "\[Equal]", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["S1", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}], "\[And]", 
             RowBox[{
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T2", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], "\[Equal]", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["S2", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox[
                SubscriptBox["\[Sum]", 
                 StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["S1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["S2", "ExpressionVariable"]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 SubscriptBox[
                  RowBox[{
                    StyleBox["T1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["T2", "ExpressionVariable"]}], 
                  StyleBox["i", "ExpressionVariable"]], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 SubscriptBox[
                  RowBox[{
                    StyleBox["S1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["S2", "ExpressionVariable"]}], 
                  StyleBox["i", "ExpressionVariable"]]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", 
                   StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S1", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T1", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["S1", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], 
               UnderscriptBox["+", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", 
                   StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S2", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T2", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["S2", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T1", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], "\[Equal]", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["S1", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}], "\[And]", 
             RowBox[{
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T2", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], "\[Equal]", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["S2", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox[
                SubscriptBox["\[Sum]", 
                 StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["S1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["S2", "ExpressionVariable"]}], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 SubscriptBox[
                  RowBox[{
                    StyleBox["T1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["T2", "ExpressionVariable"]}], 
                  StyleBox["i", "ExpressionVariable"]], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 SubscriptBox[
                  RowBox[{
                    StyleBox["S1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["S2", "ExpressionVariable"]}], 
                  StyleBox["i", "ExpressionVariable"]]}]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", 
                   StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S1", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T1", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["S1", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], 
               UnderscriptBox["+", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", 
                   StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S2", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["T2", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                   UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                   SubscriptBox[
                    StyleBox["S2", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1796414707"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1796414707-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1796414707-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1796414707", "Proof|ID:1796414707-1"},
 CellID->22030080],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1796414707-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->240341862]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->599328732],

Cell[CellGroupData[{

Cell["\<\
Lemma: \[Congruent] is Equivalence Relation\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645260278067552*^9, 3.6452602872747307`*^9}},
 CellID->843394468],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1254924191}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6452603218634872`*^9, 3.645260329973452*^9}},
 CellTags->{
  "ID:1254924191", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1254924191],

Cell[BoxData[
 RowBox[{"a", 
  UnderscriptBox[
   SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], 
  "a"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reflexivity of \[Congruent])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:923334447", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6452603341723957`*^9, 3.645260344550541*^9}},
 CellTags->{
  "ID:923334447", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[Congruent]"},
 CellID->923334447],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"b", 
    UnderscriptBox[
     SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], "a"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symmetry of \[Congruent])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1712150230", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6452603757775183`*^9, 3.64526039002085*^9}},
 CellTags->{
  "ID:1712150230", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of \[Congruent]"},
 CellID->1712150230],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", 
      UnderscriptBox[
       SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], "c"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], "c"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(transitivity of \[Congruent])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1473809719", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6452603950157757`*^9, 3.6452604080216722`*^9}},
 CellTags->{
  "ID:1473809719", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[Congruent]"},
 CellID->1473809719],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1757323546]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1341334469],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reflexivity of \[Congruent])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reflexivity of \[Congruent])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox[
            SubscriptBox["\[Congruent]", 
             StyleBox["B", "ExpressionVariable"]], 
            StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox[
            SubscriptBox["\[Congruent]", 
             StyleBox["B", "ExpressionVariable"]], 
            StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:923334447"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p923334447-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p923334447-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:923334447", "Proof|ID:923334447-1"},
 CellID->1760520462],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p923334447-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1790149953]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symmetry of \[Congruent])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symmetry of \[Congruent])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1712150230"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1712150230-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1712150230-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1712150230", "Proof|ID:1712150230-1"},
 CellID->463691700],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1712150230-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->615703635]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(transitivity of \[Congruent])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(transitivity of \[Congruent])\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1473809719"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1473809719-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1473809719-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1473809719", "Proof|ID:1473809719-1"},
 CellID->1487366676],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1473809719-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->971415663]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "\[Congruent] Coincides with ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[LeftRightArrow]", "*"], TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.6452602051279907`*^9, 3.645260207224009*^9}, {
   3.645346034568742*^9, 3.6453460551099863`*^9}, {3.6454410572295437`*^9, 
   3.645441061453977*^9}, 3.647927804606431*^9, {3.647928132849963*^9, 
   3.647928133756131*^9}},
 CellID->407327731],

Cell["", "OpenEnvironment",
 CellID->625009726],

Cell[CellGroupData[{

Cell["\<\
Lemma: Alternative Description of \[Congruent]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645342595581345*^9, 3.6453426027495413`*^9}},
 CellID->90864410],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
    RowBox[{"[", 
     RowBox[{"T", ",", "S"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
     "\[Equal]", 
     RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}], 
    "\[Implies]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"a", ",", "b"}], "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"\[And]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {
             RowBox[{
              UnderscriptBox["\[ForAll]", 
               RowBox[{
                RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                RowBox[{
                "\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{
                SubscriptBox["S", "i"], "\[Element]", "B"}], 
               TagBox[")",
                "AutoParentheses"]}]}]},
            {
             RowBox[{"b", "\[Equal]", 
              RowBox[{"a", 
               UnderscriptBox["-", "\[ScriptCapitalR]"], 
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                  RowBox[{
                   RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                   RowBox[{
                   "\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
                 
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  RowBox[{
                   SubscriptBox["T", "i"], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   SubscriptBox["S", "i"]}], 
                  TagBox[")",
                   "AutoParentheses"]}]}], 
                TagBox[")",
                 "AutoParentheses"]}]}]}]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[Implies]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
           RowBox[{"[", 
            RowBox[{"U", ",", "V"}], "]"}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"\[And]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{
                RowBox[{
                "\[LeftBracketingBar]", "U", "\[RightBracketingBar]"}], 
                "\[Equal]", 
                RowBox[{
                "\[LeftBracketingBar]", "V", "\[RightBracketingBar]"}]}]},
              {
               RowBox[{
                UnderscriptBox["\[ForAll]", 
                 RowBox[{
                  RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                  RowBox[{
                  "\[LeftBracketingBar]", "V", "\[RightBracketingBar]"}]}]], 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{"\[And]", 
                  RowBox[{"\[Piecewise]", GridBox[{
                    {
                    RowBox[{
                    SubscriptBox["V", "i"], "\[Element]", "B"}]},
                    {
                    RowBox[{
                    SubscriptBox["V", "i"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
                    {
                    RowBox[{
                    SubscriptBox["U", "i"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}]}
                    },
                    
                    GridBoxAlignment->{
                    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
                 TagBox[")",
                  "AutoParentheses"]}]}]},
              {
               RowBox[{"b", "\[Equal]", 
                RowBox[{"a", 
                 UnderscriptBox["-", "\[ScriptCapitalR]"], 
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                    RowBox[{
                    "\[LeftBracketingBar]", "V", "\[RightBracketingBar]"}]}]], 
                   RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    SubscriptBox["U", "i"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    SubscriptBox["V", "i"]}], 
                    TagBox[")",
                    "AutoParentheses"]}]}], 
                  TagBox[")",
                   "AutoParentheses"]}]}]}]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Congruent] alternative ind.1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:306651916", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64534592113309*^9, 3.6453459330192947`*^9}, {
  3.6453532020778313`*^9, 3.645353233419518*^9}, {3.645418311597784*^9, 
  3.6454183203795843`*^9}},
 CellTags->{
  "ID:306651916", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] alternative ind.1"},
 CellID->306651916],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
    RowBox[{"[", 
     RowBox[{"T", ",", "S"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
     "\[Equal]", 
     RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}], 
    "\[Implies]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"a", ",", "b"}], "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"\[And]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {
             RowBox[{
              UnderscriptBox["\[ForAll]", 
               RowBox[{
                RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                RowBox[{
                "\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{"\[And]", 
                RowBox[{"\[Piecewise]", GridBox[{
                   {
                    RowBox[{
                    SubscriptBox["S", "i"], "\[Element]", "B"}]},
                   {
                    RowBox[{
                    SubscriptBox["S", "i"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
                   {
                    RowBox[{
                    SubscriptBox["T", "i"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
               TagBox[")",
                "AutoParentheses"]}]}]},
            {
             RowBox[{"b", "\[Equal]", 
              RowBox[{"a", 
               UnderscriptBox["-", "\[ScriptCapitalR]"], 
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                  RowBox[{
                   RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                   RowBox[{
                   "\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
                 
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  RowBox[{
                   SubscriptBox["T", "i"], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   SubscriptBox["S", "i"]}], 
                  TagBox[")",
                   "AutoParentheses"]}]}], 
                TagBox[")",
                 "AutoParentheses"]}]}]}]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[Implies]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
           RowBox[{"[", 
            RowBox[{"U", ",", "V"}], "]"}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"\[And]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{
                RowBox[{
                "\[LeftBracketingBar]", "U", "\[RightBracketingBar]"}], 
                "\[Equal]", 
                RowBox[{
                "\[LeftBracketingBar]", "V", "\[RightBracketingBar]"}]}]},
              {
               RowBox[{
                UnderscriptBox["\[ForAll]", 
                 RowBox[{
                  RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                  RowBox[{
                  "\[LeftBracketingBar]", "V", "\[RightBracketingBar]"}]}]], 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{"\[And]", 
                  RowBox[{"\[Piecewise]", GridBox[{
                    {
                    RowBox[{
                    SubscriptBox["V", "i"], "\[Element]", "B"}]},
                    {
                    RowBox[{
                    SubscriptBox["V", "i"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
                    {
                    RowBox[{
                    SubscriptBox["U", "i"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
                    {
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    SubscriptBox["U", "i"], ",", 
                    SubscriptBox["V", "i"]}], "]"}]}
                    },
                    
                    GridBoxAlignment->{
                    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
                 TagBox[")",
                  "AutoParentheses"]}]}]},
              {
               RowBox[{"b", "\[Equal]", 
                RowBox[{"a", 
                 UnderscriptBox["-", "\[ScriptCapitalR]"], 
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                    RowBox[{
                    "\[LeftBracketingBar]", "V", "\[RightBracketingBar]"}]}]], 
                   RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    SubscriptBox["U", "i"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    SubscriptBox["V", "i"]}], 
                    TagBox[")",
                    "AutoParentheses"]}]}], 
                  TagBox[")",
                   "AutoParentheses"]}]}]}]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Congruent] alternative ind.2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:571707933", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.6453427831883907`*^9, {3.645345440798292*^9, 3.645345572828938*^9}, {
   3.645345706385188*^9, 3.6453457076814747`*^9}, {3.645345808159173*^9, 
   3.645345829487008*^9}, {3.6453458996453857`*^9, 3.645345902848645*^9}, {
   3.645353212856226*^9, 3.6453532363031588`*^9}, {3.645418327464533*^9, 
   3.645418335712133*^9}},
 CellTags->{
  "ID:571707933", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] alternative ind.2"},
 CellID->571707933],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1886882275]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->935204603],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Congruent] alternative ind.1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Congruent] alternative ind.1)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["T", "ExpressionVariable"], "\[RightBracketingBar]"}],
              "\[Equal]", 
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["S", "ExpressionVariable"], 
               "\[RightBracketingBar]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}], "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["V", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["U", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "\[Equal]", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                    "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["U", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "\[ScriptCapitalR]", "ExpressionVariable"]]}]}]}], 
                   "\[And]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["U", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["T", "ExpressionVariable"], "\[RightBracketingBar]"}],
              "\[Equal]", 
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["S", "ExpressionVariable"], 
               "\[RightBracketingBar]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}], "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["V", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["U", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "\[Equal]", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                    "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["U", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "\[ScriptCapitalR]", "ExpressionVariable"]]}]}]}], 
                   "\[And]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["U", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:306651916"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p306651916-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p306651916-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:306651916", "Proof|ID:306651916-1"},
 CellID->655455289],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p306651916-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.645439821435227*^9, 3.645439854265662*^9},
 CellID->602823762]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Congruent] alternative ind.2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Congruent] alternative ind.2)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["T", "ExpressionVariable"], "\[RightBracketingBar]"}],
              "\[Equal]", 
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["S", "ExpressionVariable"], 
               "\[RightBracketingBar]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                    "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "\[ScriptCapitalR]", "ExpressionVariable"]]}]}]}], 
                 "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["V", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["U", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "\[Equal]", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                    "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["U", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                    "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["U", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}]}], 
                   "\[And]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["U", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["T", "ExpressionVariable"], "\[RightBracketingBar]"}],
              "\[Equal]", 
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["S", "ExpressionVariable"], 
               "\[RightBracketingBar]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                    "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "\[ScriptCapitalR]", "ExpressionVariable"]]}]}]}], 
                 "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["V", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["U", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "\[Equal]", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                    "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["U", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                    "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["U", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}]}], 
                   "\[And]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["U", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:571707933"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p571707933-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p571707933-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:571707933", "Proof|ID:571707933-1"},
 CellID->1209243937],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p571707933-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->65884176]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1488103865],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[LeftRightArrow]", "*"], TraditionalForm]]],
 " \[SubsetEqual] \[Congruent]"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645425032661322*^9, 3.6454250542835407`*^9}, {
  3.645430231845159*^9, 3.6454302372659273`*^9}},
 CellID->1075279006],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 473420099}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6454250674278307`*^9, 3.6454250756662207`*^9}},
 CellTags->{
  "ID:473420099", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->473420099],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", "B"], "\[ScriptCapitalR]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow] \[SubsetEqual] \[Congruent])", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:544141308", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645425244952518*^9, 3.645425247463365*^9}},
 CellTags->{
  "ID:544141308", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[SubsetEqual] \[Congruent]"},
 CellID->544141308],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[LeftRightArrow]", "B"], "\[ScriptCapitalR]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[LeftRightArrow] \[SubsetEqual] \[Congruent])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1307557419", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645425084013419*^9, 3.64542510917806*^9}},
 CellTags->{
  "ID:1307557419", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] \[SubsetEqual] \[Congruent]"},
 CellID->1307557419],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
     RowBox[{"[", "T", "]"}]}]], 
   RowBox[{"T", "\[NotEqual]", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], "-", 
         "1"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        SubscriptBox["T", "i"], 
        UnderscriptBox[
         SubscriptBox["\[LeftRightArrow]", "B"], "\[ScriptCapitalR]"], 
        SubscriptBox["T", 
         RowBox[{"i", "+", "1"}]]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
      RowBox[{
       SubscriptBox["T", "1"], 
       UnderscriptBox[
        SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], 
       SubscriptBox["T", "i"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(connectibility \[SubsetEqual] \[Congruent] ind)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:170897044", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645428284487365*^9, 3.645428394367504*^9}, {
  3.6454292189140882`*^9, 3.645429240551181*^9}},
 CellTags->{
  "ID:170897044", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility \[SubsetEqual] \[Congruent] ind"},
 CellID->170897044],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], "\[ScriptCapitalR]"], 
    "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(connectibility \[SubsetEqual] \[Congruent])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:720101359", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6454296645909643`*^9, 3.645429689648857*^9}},
 CellTags->{
  "ID:720101359", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility \[SubsetEqual] \[Congruent]"},
 CellID->720101359],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->573708679]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->727999641],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] \[SubsetEqual] \[Congruent])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] \[SubsetEqual] \[Congruent])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:544141308"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p544141308-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p544141308-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:544141308", "Proof|ID:544141308-1"},
 CellID->8057470],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p544141308-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->388792251]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LeftRightArrow] \[SubsetEqual] \[Congruent])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LeftRightArrow] \[SubsetEqual] \[Congruent])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[LeftRightArrow]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1307557419"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1307557419-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1307557419-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1307557419", "Proof|ID:1307557419-1"},
 CellID->1073556871],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1307557419-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->913970451]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(connectibility \[SubsetEqual] \[Congruent] ind)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(connectibility \[SubsetEqual] \[Congruent] ind)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], "1"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox[
                "i", "ExpressionVariable"]]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["T", "ExpressionVariable"], "\[NotEqual]", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "-", "1"}]}]}}]], 
               RowBox[{
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   StyleBox["B", "ExpressionVariable"]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 SubscriptBox[
                  StyleBox["T", "ExpressionVariable"], 
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], "1"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               SubscriptBox[
                StyleBox["T", "ExpressionVariable"], 
                StyleBox[
                "i", "ExpressionVariable"]]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:170897044"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p170897044-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p170897044-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:170897044", "Proof|ID:170897044-1"},
 CellID->2702629],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p170897044-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1632663207]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(connectibility \[SubsetEqual] \[Congruent])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(connectibility \[SubsetEqual] \[Congruent])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:720101359"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p720101359-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p720101359-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:720101359", "Proof|ID:720101359-1"},
 CellID->896819044],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p720101359-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1689706913]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1067595910],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: \[Congruent] \[SubsetEqual] ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[LeftRightArrow]", "*"], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6454302442121487`*^9, 3.645430252344236*^9}},
 CellID->2061298522],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "m"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"c", "\[Element]", "B"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"c", "\[NotEqual]", 
        UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
       RowBox[{
        UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
        RowBox[{"m", ",", "c"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", 
       UnderscriptBox[
        SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
        "\[ScriptCapitalR]"], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", 
         UnderscriptBox["-", "\[ScriptCapitalR]"], 
         RowBox[{"m", 
          UnderscriptBox["*", "\[ScriptCapitalR]"], "c"}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(43)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1680644099", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.644912496101457*^9, 3.644912573325588*^9}, {
  3.64543666092474*^9, 3.645436665561171*^9}, {3.645436752991139*^9, 
  3.645436894906932*^9}, {3.645436944529711*^9, 3.645437035225486*^9}, {
  3.6454387844930077`*^9, 3.645438788817505*^9}},
 CellTags->{
  "ID:1680644099", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "43"},
 CellID->1680644099],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
    RowBox[{"[", 
     RowBox[{"T", ",", "S"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
     "\[Equal]", 
     RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}], 
    "\[Implies]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"a", ",", "b"}], "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"\[And]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {
             RowBox[{
              UnderscriptBox["\[ForAll]", 
               RowBox[{
                RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                RowBox[{
                "\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{"\[And]", 
                RowBox[{"\[Piecewise]", GridBox[{
                   {
                    RowBox[{
                    SubscriptBox["S", "i"], "\[Element]", "B"}]},
                   {
                    RowBox[{
                    SubscriptBox["S", "i"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
                   {
                    RowBox[{
                    SubscriptBox["T", "i"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
                   {
                    RowBox[{
                    UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    SubscriptBox["T", "i"], ",", 
                    SubscriptBox["S", "i"]}], "]"}]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
               TagBox[")",
                "AutoParentheses"]}]}]},
            {
             RowBox[{"b", "\[Equal]", 
              RowBox[{"a", 
               UnderscriptBox["-", "\[ScriptCapitalR]"], 
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                  RowBox[{
                   RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                   RowBox[{
                   "\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
                 
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  RowBox[{
                   SubscriptBox["T", "i"], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   SubscriptBox["S", "i"]}], 
                  TagBox[")",
                   "AutoParentheses"]}]}], 
                TagBox[")",
                 "AutoParentheses"]}]}]}]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[Implies]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", 
         UnderscriptBox[
          SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
          "\[ScriptCapitalR]"], "b"}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[Congruent] \[SubsetEqual] connectibility ind)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:555151129", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645439341161521*^9, 3.645439410564981*^9}},
 CellTags->{
  "ID:555151129", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] \[SubsetEqual] connectibility ind"},
 CellID->555151129],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], "\[ScriptCapitalR]"],
       "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[Congruent] \[SubsetEqual] connectibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2104087526", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645440534064139*^9, 3.645440573120852*^9}},
 CellTags->{
  "ID:2104087526", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] \[SubsetEqual] connectibility"},
 CellID->2104087526],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->146406587]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->498470761],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(43)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(43)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[Element]", 
               StyleBox["B", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox[
                 "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1680644099"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1680644099-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1680644099-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1680644099", "Proof|ID:1680644099-1"},
 CellID->279660360],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1680644099-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->322411293]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Congruent] \[SubsetEqual] connectibility ind)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Congruent] \[SubsetEqual] connectibility ind)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["T", "ExpressionVariable"], "\[RightBracketingBar]"}],
              "\[Equal]", 
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["S", "ExpressionVariable"], 
               "\[RightBracketingBar]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                    "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                    "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}]}], 
                 "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  TemplateBox[{"\[LeftRightArrow]", 
                    StyleBox["B", "ExpressionVariable"], "*"}, 
                   "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox[
                 "b", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["T", "ExpressionVariable"], "\[RightBracketingBar]"}],
              "\[Equal]", 
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["S", "ExpressionVariable"], 
               "\[RightBracketingBar]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                    "\[And]", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
                    "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}]}], 
                 "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  TemplateBox[{"\[LeftRightArrow]", 
                    StyleBox["B", "ExpressionVariable"], "*"}, 
                   "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox[
                 "b", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:555151129"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p555151129-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p555151129-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:555151129", "Proof|ID:555151129-1"},
 CellID->440681758],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p555151129-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->226062194]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Congruent] \[SubsetEqual] connectibility)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Congruent] \[SubsetEqual] connectibility)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2104087526"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2104087526-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2104087526-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2104087526", "Proof|ID:2104087526-1"},
 CellID->1134450301],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2104087526-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->413731443]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->2070925753],

Cell[CellGroupData[{

Cell[TextData[{
 "Theorem: \[Congruent] Coincides with ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[LeftRightArrow]", "*"], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645440882811607*^9, 3.645440892194656*^9}},
 CellID->1902867313],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], "\[ScriptCapitalR]"],
       "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Congruent]=connectibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:980108134", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645440902598974*^9, 3.645440942075266*^9}},
 CellTags->{
  "ID:980108134", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent]=connectibility"},
 CellID->980108134],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->815657706]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1770914195],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Congruent]=connectibility)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Congruent]=connectibility)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              TemplateBox[{"\[LeftRightArrow]", 
                StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
               SyntaxForm -> SubsuperscriptBox], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:980108134"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p980108134-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p980108134-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:980108134", "Proof|ID:980108134-1"},
 CellID->699177716],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p980108134-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->224283849]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["0-Congruence (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6481876509551907`*^9, 3.64818765578975*^9}, {
  3.64819687014604*^9, 3.64819687711687*^9}},
 CellID->1326539787],

Cell["", "OpenEnvironment",
 CellID->1596153952],

Cell[CellGroupData[{

Cell["Lemma: 0-Congruence", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648187689124264*^9, 3.64818769839675*^9}},
 CellID->1713639269],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 2002188699}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648187992542634*^9, 3.648188001073036*^9}},
 CellTags->{
  "ID:2002188699", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2002188699],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox[
     SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
      RowBox[{"[", 
       RowBox[{"T", ",", "S"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
           "\[Equal]", 
           RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             SubscriptBox["S", "i"], "\[Element]", "B"}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{"a", "\[Equal]", 
           RowBox[{
            UnderscriptBox[
             SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
             RowBox[{
              RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
              RowBox[{
              "\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{
              SubscriptBox["T", "i"], 
              UnderscriptBox["*", "\[ScriptCapitalR]"], 
              SubscriptBox["S", "i"]}], 
             TagBox[")",
              "AutoParentheses"]}]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Congruent] 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:160768319", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648187739936322*^9, 3.648187839399753*^9}, {
  3.648188003966029*^9, 3.648188006114957*^9}, {3.648188456881227*^9, 
  3.648188461715395*^9}},
 CellTags->{
  "ID:160768319", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0", "IR"},
 CellID->160768319],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", "\[Element]", "B"}], "\[Implies]", 
  RowBox[{"a", 
   UnderscriptBox[
    SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], 
   UnderscriptBox["0", "\[ScriptCapitalR]"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] \[Congruent] 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:562483616", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648206011409876*^9, 3.648206027868681*^9}},
 CellTags->{
  "ID:562483616", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] \[Congruent] 0"},
 CellID->562483616],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox[
    SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], 
   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"b", 
     UnderscriptBox["*", "\[ScriptCapitalR]"], "a"}], ")"}], 
   UnderscriptBox[
    SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], 
   UnderscriptBox["0", "\[ScriptCapitalR]"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Congruent] 0 *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:853458295", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6481880817852583`*^9, 3.64818809959848*^9}},
 CellTags->{
  "ID:853458295", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 *"},
 CellID->853458295],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox[
      SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], 
     UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
    RowBox[{"b", 
     UnderscriptBox[
      SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], 
     UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 462880889}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648188141118579*^9, 3.648188147272387*^9}},
 CellTags->{
  "ID:462880889", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->462880889],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"a", 
    UnderscriptBox["+", "\[ScriptCapitalR]"], "b"}], ")"}], 
  UnderscriptBox[
   SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], 
  UnderscriptBox["0", "\[ScriptCapitalR]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Congruent] 0 +)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:44331059", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648188018541175*^9, 3.648188060592132*^9}, 
   3.648188150079582*^9},
 CellTags->{
  "ID:44331059", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 +"},
 CellID->44331059],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"a", 
    UnderscriptBox["-", "\[ScriptCapitalR]"], "b"}], ")"}], 
  UnderscriptBox[
   SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], 
  UnderscriptBox["0", "\[ScriptCapitalR]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Congruent] 0 -)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1958399498", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648188154424675*^9, 3.648188161869121*^9}},
 CellTags->{
  "ID:1958399498", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 -"},
 CellID->1958399498],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1796204648]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1716503601],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Congruent] 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Congruent] 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["T", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], "\[Equal]", 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["S", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                   StyleBox["B", "ExpressionVariable"]}]}], "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["T", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["T", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], "\[Equal]", 
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["S", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                   StyleBox["B", "ExpressionVariable"]}]}], "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:160768319"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p160768319-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p160768319-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:160768319", "Proof|ID:160768319-1"},
 CellID->149045731],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p160768319-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->115489367]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] \[Congruent] 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] \[Congruent] 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], "\[Element]", 
             StyleBox["B", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox[
              "\[ScriptCapitalR]", 
               "ExpressionVariable"]]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], "\[Element]", 
             StyleBox["B", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox[
              "\[ScriptCapitalR]", 
               "ExpressionVariable"]]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:562483616"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p562483616-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p562483616-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:562483616", "Proof|ID:562483616-1"},
 CellID->929707542],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p562483616-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2060701176]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Congruent] 0 *)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Congruent] 0 *)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["a", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox[
              "\[ScriptCapitalR]", 
               "ExpressionVariable"]]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["a", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox[
              "\[ScriptCapitalR]", 
               "ExpressionVariable"]]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:853458295"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p853458295-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p853458295-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:853458295", "Proof|ID:853458295-1"},
 CellID->282680859],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p853458295-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1027059538]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Congruent] 0 +)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Congruent] 0 +)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox[
              "\[ScriptCapitalR]", 
               "ExpressionVariable"]]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox[
              "\[ScriptCapitalR]", 
               "ExpressionVariable"]]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:44331059"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p44331059-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p44331059-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:44331059", "Proof|ID:44331059-1"},
 CellID->2113222684],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p44331059-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->138052613]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Congruent] 0 -)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Congruent] 0 -)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox[
              "\[ScriptCapitalR]", 
               "ExpressionVariable"]]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox[
              "\[ScriptCapitalR]", 
               "ExpressionVariable"]]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1958399498"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1958399498-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1958399498-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1958399498", "Proof|ID:1958399498-1"},
 CellID->1499077166],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1958399498-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->38712736]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1365795944],

Cell[CellGroupData[{

Cell["Lemma: Congruence Modulo Extended Sets", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648197067402748*^9, 3.648197074594233*^9}},
 CellID->2051871741],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}]], 
  RowBox[{"c", 
   UnderscriptBox[
    SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], 
   UnderscriptBox["0", "\[ScriptCapitalR]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 2069217266}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648197109307187*^9, 3.648197135038965*^9}},
 CellTags->{
  "ID:2069217266", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2069217266],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
    RowBox[{"[", 
     RowBox[{"T", ",", "S"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
     "\[Equal]", 
     RowBox[{"\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}], 
    "\[Implies]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", 
        RowBox[{"a", ",", "b"}], "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"\[And]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {
             RowBox[{
              UnderscriptBox["\[ForAll]", 
               RowBox[{
                RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                RowBox[{
                "\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{
                SubscriptBox["S", "i"], "\[Element]", 
                RowBox[{"B", "\[Union]", 
                 RowBox[{"{", "c", "}"}]}]}], 
               TagBox[")",
                "AutoParentheses"]}]}]},
            {
             RowBox[{"b", "\[Equal]", 
              RowBox[{"a", 
               UnderscriptBox["-", "\[ScriptCapitalR]"], 
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                RowBox[{
                 UnderscriptBox[
                  SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                  RowBox[{
                   RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                   RowBox[{
                   "\[LeftBracketingBar]", "S", "\[RightBracketingBar]"}]}]], 
                 
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  RowBox[{
                   SubscriptBox["T", "i"], 
                   UnderscriptBox["*", "\[ScriptCapitalR]"], 
                   SubscriptBox["S", "i"]}], 
                  TagBox[")",
                   "AutoParentheses"]}]}], 
                TagBox[")",
                 "AutoParentheses"]}]}]}]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[Implies]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
           RowBox[{"[", 
            RowBox[{"U", ",", "V"}], "]"}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"\[And]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{
                RowBox[{
                "\[LeftBracketingBar]", "U", "\[RightBracketingBar]"}], 
                "\[Equal]", 
                RowBox[{
                "\[LeftBracketingBar]", "V", "\[RightBracketingBar]"}]}]},
              {
               RowBox[{
                UnderscriptBox["\[ForAll]", 
                 RowBox[{
                  RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                  RowBox[{
                  "\[LeftBracketingBar]", "V", "\[RightBracketingBar]"}]}]], 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{
                  SubscriptBox["V", "i"], "\[Element]", "B"}], 
                 TagBox[")",
                  "AutoParentheses"]}]}]},
              {
               RowBox[{"b", "\[Equal]", 
                RowBox[{"a", 
                 UnderscriptBox["-", "\[ScriptCapitalR]"], 
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
                    RowBox[{
                    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                    RowBox[{
                    "\[LeftBracketingBar]", "V", "\[RightBracketingBar]"}]}]], 
                   RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    SubscriptBox["U", "i"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    SubscriptBox["V", "i"]}], 
                    TagBox[")",
                    "AutoParentheses"]}]}], 
                  TagBox[")",
                   "AutoParentheses"]}]}]}]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Congruent] extended ind)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:277446168", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64534592113309*^9, 3.6453459330192947`*^9}, {
  3.6453532020778313`*^9, 3.645353233419518*^9}, {3.645418311597784*^9, 
  3.6454183203795843`*^9}, {3.648197102578281*^9, 3.6481971050106993`*^9}, {
  3.6481971568346786`*^9, 3.648197159709844*^9}},
 CellTags->{
  "ID:277446168", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] extended ind"},
 CellID->277446168],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Congruent]", 
        RowBox[{"B", "\[Union]", 
         RowBox[{"{", "c", "}"}]}]], "\[ScriptCapitalR]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Congruent] extended)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1459665988", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648197211478607*^9, 3.648197245530672*^9}, {
  3.6482073695686483`*^9, 3.6482073796610003`*^9}},
 CellTags->{
  "ID:1459665988", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] extended"},
 CellID->1459665988],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->730815423]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1818313587],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Congruent] extended ind)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Congruent] extended ind)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], "\[Equal]", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["S", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[Union]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]}]}]}], 
                 "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["V", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["U", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "\[Equal]", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}], "\[And]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["U", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               UnderscriptBox["0", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
             "\[And]", 
             RowBox[{
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], "\[Equal]", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["S", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[Union]", 
                    RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]}]}]}], 
                 "\[And]", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}]}]}]}], 
               "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["V", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["U", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "\[Equal]", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}], "\[And]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["V", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["U", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                    SubscriptBox[
                    StyleBox["V", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:277446168"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p277446168-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p277446168-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:277446168", "Proof|ID:277446168-1"},
 CellID->37830858],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p277446168-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->291885472]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Congruent] extended)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Congruent] extended)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\[Union]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Congruent]", 
               StyleBox["B", "ExpressionVariable"]], 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             UnderscriptBox["0", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\[Union]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]}]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["b", "ExpressionVariable"]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                SubscriptBox["\[Congruent]", 
                 StyleBox["B", "ExpressionVariable"]], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox[
               "b", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1459665988"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1459665988-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1459665988-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1459665988", "Proof|ID:1459665988-1"},
 CellID->569980310],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1459665988-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->278474281]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Theorems: Ideals (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.647934476116659*^9, 3.647934479942687*^9}, {
  3.6479391256907387`*^9, 3.647939127540679*^9}, {3.648282088317358*^9, 
  3.648282090925303*^9}},
 CellID->979873577],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isReductionRing", "[", "\[ScriptCapitalR]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1670835483}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647934495521925*^9, 3.6479345042918863`*^9}},
 CellTags->{
  "ID:1670835483", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1670835483],

Cell[CellGroupData[{

Cell["Ideals of Sets (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.647935909304995*^9, 3.64793591235109*^9}, {
  3.648209256835754*^9, 3.648209260546261*^9}},
 CellID->933596342],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "\[ScriptCapitalR]", "]"}]], 
   RowBox[{"[", "B", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 97617962}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647935040214641*^9, 3.6479350533808327`*^9}},
 CellTags->{
  "ID:97617962", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->97617962],

Cell["", "Text",
 CellID->1624714196],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"IB", "=", 
   RowBox[{
    UnderscriptBox["ideal", "\[ScriptCapitalR]"], "[", "B", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1452952621}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647935981478199*^9, 3.6479360103392363`*^9}},
 CellTags->{
  "ID:1452952621", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1452952621],

Cell["", "OpenEnvironment",
 CellID->160364002],

Cell[CellGroupData[{

Cell["Lemma: Simple Properties of Ideals", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.64793507109972*^9, 3.647935076543208*^9}},
 CellID->1960712112],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainSets", "[", "\[ScriptCapitalR]", "]"}]], 
  RowBox[{"[", "IB", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DomainSets ideal)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1760501242", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647935094157009*^9, 3.647935124465653*^9}, {
  3.647935396468245*^9, 3.647935397692471*^9}, {3.647936018004654*^9, 
  3.647936018227195*^9}},
 CellTags->{
  "ID:1760501242", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets ideal", "IR"},
 CellID->1760501242],

Cell[BoxData[
 RowBox[{"isSet", "[", "IB", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal isSet)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:554864723", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647935130322064*^9, 3.647935136681704*^9}, {
  3.647935391271502*^9, 3.6479353924891243`*^9}, {3.6479360152001457`*^9, 
  3.647936015384919*^9}},
 CellTags->{
  "ID:554864723", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal isSet", "IR"},
 CellID->554864723],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "x"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Element]", "IB"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", "x", "]"}]}], "\[And]", 
      RowBox[{"x", 
       UnderscriptBox[
        SubscriptBox["\[Congruent]", "B"], "\[ScriptCapitalR]"], 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] ideal)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:739591353", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6482051614013557`*^9, 3.648205190833343*^9}, 
   3.6482053024146433`*^9},
 CellTags->{
  "ID:739591353", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] ideal"},
 CellID->739591353],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"a", ",", 
   RowBox[{"b", "\[Element]", "IB"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 2030275451}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6479360326509743`*^9, 3.6479360417093*^9}, {
   3.647936074515585*^9, 3.6479361136647377`*^9}, 3.6482056339377403`*^9},
 CellTags->{
  "ID:2030275451", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2030275451],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["+", "\[ScriptCapitalR]"], "b"}], "\[Element]", 
  "IB"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideals closed +)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1644449200", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647936139755803*^9, 3.647936147327711*^9}},
 CellTags->{
  "ID:1644449200", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideals closed +", "IR"},
 CellID->1644449200],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["-", "\[ScriptCapitalR]"], "b"}], "\[Element]", 
  "IB"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideals closed - binary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1500033871", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647936156538476*^9, 3.647936164435614*^9}},
 CellTags->{
  "ID:1500033871", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideals closed - binary", "IR"},
 CellID->1500033871],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["ideal", "\[ScriptCapitalR]"], "[", 
   RowBox[{"B", "\[Union]", 
    RowBox[{"{", "a", "}"}]}], "]"}], "\[Equal]", 
  "IB"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal \[Union] singleton)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1114678876", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647935418219808*^9, 3.6479354818813143`*^9}, {
  3.6479360234669952`*^9, 3.647936047886808*^9}, {3.647936081854815*^9, 
  3.6479360838596773`*^9}},
 CellTags->{
  "ID:1114678876", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal \[Union] singleton"},
 CellID->1114678876],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"d", ",", "h1", ",", "h2"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"d", 
        UnderscriptBox[
         SubsuperscriptBox["\[Rule]", "B", "*"], "\[ScriptCapitalR]"], "h1"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"d", 
        UnderscriptBox[
         SubsuperscriptBox["\[Rule]", "B", "*"], "\[ScriptCapitalR]"], "h2"}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     RowBox[{"h1", 
      UnderscriptBox["-", "\[ScriptCapitalR]"], "h2"}], "\[Element]", 
     "IB"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal \[RightArrow]*)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1895756686", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647936219500675*^9, 3.647936247598377*^9}, {
  3.648207099015327*^9, 3.648207128059329*^9}},
 CellTags->{
  "ID:1895756686", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal \[RightArrow]*"},
 CellID->1895756686],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->744302405]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->407473638],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DomainSets ideal)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DomainSets ideal)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainSets", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainSets", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1760501242"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1760501242-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1760501242-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1760501242", "Proof|ID:1760501242-1"},
 CellID->137891172],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1760501242-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->23467037]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(ideal isSet)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(ideal isSet)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isSet", "[", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isSet", "[", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:554864723"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p554864723-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p554864723-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:554864723", "Proof|ID:554864723-1"},
 CellID->73662413],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p554864723-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->365783100]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] ideal)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] ideal)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Congruent]", 
                   StyleBox["B", "ExpressionVariable"]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 UnderscriptBox["0", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Congruent]", 
                   StyleBox["B", "ExpressionVariable"]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 UnderscriptBox["0", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:739591353"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p739591353-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p739591353-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:739591353", "Proof|ID:739591353-1"},
 CellID->136162227],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p739591353-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->319964668]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(ideals closed +)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(ideals closed +)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["+", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Element]", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["+", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Element]", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1644449200"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1644449200-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1644449200-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1644449200", "Proof|ID:1644449200-1"},
 CellID->535086304],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1644449200-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->723227642]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(ideals closed - binary)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(ideals closed - binary)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["-", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Element]", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["-", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Element]", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1500033871"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1500033871-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1500033871-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1500033871", "Proof|ID:1500033871-1"},
 CellID->768137348],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1500033871-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->277650203]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(ideal \[Union] singleton)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(ideal \[Union] singleton)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], "\[Union]", 
               RowBox[{"{", 
                 StyleBox["a", "ExpressionVariable"], "}"}]}], "]"}], 
           "\[Equal]", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], "\[Union]", 
               RowBox[{"{", 
                 StyleBox["a", "ExpressionVariable"], "}"}]}], "]"}], 
           "\[Equal]", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1114678876"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1114678876-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1114678876-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1114678876", "Proof|ID:1114678876-1"},
 CellID->1601029605],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1114678876-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->413658239]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(ideal \[RightArrow]*)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(ideal \[RightArrow]*)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["h1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["h2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["h1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["h2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["h1", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["h2", "ExpressionVariable"]}], "\[Element]", 
             RowBox[{
               UnderscriptBox["ideal", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["B", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["h1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["h2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["h1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["h2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["h1", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["h2", "ExpressionVariable"]}], "\[Element]", 
             RowBox[{
               UnderscriptBox["ideal", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["B", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1895756686"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1895756686-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1895756686-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1895756686", "Proof|ID:1895756686-1"},
 CellID->1682672380],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1895756686-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->885739272]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Ideals of Tuples (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.647935917438836*^9, 3.6479359206934357`*^9}, {
  3.648209265676959*^9, 3.6482092674686413`*^9}, {3.648282073136292*^9, 
  3.648282075818136*^9}},
 CellID->1115801953],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
   RowBox[{"[", "T", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1581989517}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647935040214641*^9, 3.6479350533808327`*^9}, {
  3.6482809066123466`*^9, 3.648280911113352*^9}},
 CellTags->{
  "ID:1581989517", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1581989517],

Cell["", "Text",
 CellID->84843352],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"IT", "=", 
   RowBox[{
    UnderscriptBox["ideal", "\[ScriptCapitalR]"], "[", "T", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 185976047}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647935981478199*^9, 3.6479360103392363`*^9}, {
  3.648280913217011*^9, 3.6482809145096416`*^9}},
 CellTags->{
  "ID:185976047", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->185976047],

Cell["", "OpenEnvironment",
 CellID->1901141524],

Cell[CellGroupData[{

Cell["Lemma: Ideals of Tuples", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648280886551572*^9, 3.648280890095975*^9}},
 CellID->1965491486],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "x"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Element]", "IT"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", "x", "]"}]}], "\[And]", 
      RowBox[{"x", 
       UnderscriptBox[
        SubscriptBox["\[Congruent]", 
         RowBox[{"TupleToSet", "[", "T", "]"}]], "\[ScriptCapitalR]"], 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] tuple-ideal)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1918127619", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6482051614013557`*^9, 3.648205190833343*^9}, 
   3.6482053024146433`*^9, {3.64828168251683*^9, 3.648281688806905*^9}},
 CellTags->{
  "ID:1918127619", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] tuple-ideal"},
 CellID->1918127619],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainSets", "[", "\[ScriptCapitalR]", "]"}]], 
  RowBox[{"[", "IT", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DomainSets tuple-ideal)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:279213832", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647935094157009*^9, 3.647935124465653*^9}, {
   3.647935396468245*^9, 3.647935397692471*^9}, {3.647936018004654*^9, 
   3.647936018227195*^9}, 3.648280929563972*^9},
 CellTags->{
  "ID:279213832", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets tuple-ideal", "IR"},
 CellID->279213832],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"a", ",", 
   RowBox[{"b", "\[Element]", "IT"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\ReductionRings.nb", 1225983506}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6479360326509743`*^9, 3.6479360417093*^9}, {
   3.647936074515585*^9, 3.6479361136647377`*^9}, 3.6482056339377403`*^9, 
   3.6482809538631353`*^9},
 CellTags->{
  "ID:1225983506", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1225983506],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["+", "\[ScriptCapitalR]"], "b"}], "\[Element]", 
  "IT"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(tuple-ideals closed +)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1587050664", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647936139755803*^9, 3.647936147327711*^9}, 
   3.648280956188348*^9},
 CellTags->{
  "ID:1587050664", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideals closed +", "IR"},
 CellID->1587050664],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["-", "\[ScriptCapitalR]"], "b"}], "\[Element]", 
  "IT"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(tuple-ideals closed - binary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:508704301", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647936156538476*^9, 3.647936164435614*^9}, 
   3.648280958299944*^9},
 CellTags->{
  "ID:508704301", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideals closed - binary", "IR"},
 CellID->508704301],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["ideal", "\[ScriptCapitalR]"], "[", 
   RowBox[{"T", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "a"}], "]"}], "\[Equal]", 
  "IT"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(tuple-ideal append)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1181234284", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647935418219808*^9, 3.6479354818813143`*^9}, {
  3.6479360234669952`*^9, 3.647936047886808*^9}, {3.647936081854815*^9, 
  3.6479360838596773`*^9}, {3.648280962208706*^9, 3.648280969646364*^9}},
 CellTags->{
  "ID:1181234284", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideal append"},
 CellID->1181234284],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"d", ",", "h1", ",", "h2"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"d", 
        UnderscriptBox[
         SubsuperscriptBox["\[Rule]", "T", "*"], "\[ScriptCapitalR]"], "h1"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"d", 
        UnderscriptBox[
         SubsuperscriptBox["\[Rule]", "T", "*"], "\[ScriptCapitalR]"], "h2"}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     RowBox[{"h1", 
      UnderscriptBox["-", "\[ScriptCapitalR]"], "h2"}], "\[Element]", 
     "IT"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(tuple-ideal \[RightArrow]*)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:352062174", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.647936219500675*^9, 3.647936247598377*^9}, {
   3.648207099015327*^9, 3.648207128059329*^9}, 3.648280974112156*^9, {
   3.648281940527359*^9, 3.648281943293676*^9}},
 CellTags->{
  "ID:352062174", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideal \[RightArrow]*"},
 CellID->352062174],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1414944585]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1306363546],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] tuple-ideal)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] tuple-ideal)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Congruent]", 
                   RowBox[{"TupleToSet", "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 UnderscriptBox["0", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Congruent]", 
                   RowBox[{"TupleToSet", "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 UnderscriptBox["0", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1918127619"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1918127619-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1918127619-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1918127619", "Proof|ID:1918127619-1"},
 CellID->783030095],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1918127619-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1050354800]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DomainSets tuple-ideal)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DomainSets tuple-ideal)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainSets", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["T", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainSets", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["T", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:279213832"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p279213832-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p279213832-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:279213832", "Proof|ID:279213832-1"},
 CellID->889748810],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p279213832-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->820221019]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(tuple-ideals closed +)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(tuple-ideals closed +)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["+", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Element]", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["T", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["+", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Element]", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["T", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1587050664"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1587050664-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1587050664-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1587050664", "Proof|ID:1587050664-1"},
 CellID->1831236240],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1587050664-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1858860526]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(tuple-ideals closed - binary)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(tuple-ideals closed - binary)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["-", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Element]", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["T", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["-", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox["b", "ExpressionVariable"]}], "\[Element]", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["T", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:508704301"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p508704301-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p508704301-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:508704301", "Proof|ID:508704301-1"},
 CellID->1768992322],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p508704301-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2105514433]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(tuple-ideal append)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(tuple-ideal append)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], "\:293a", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["T", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["ideal", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
                 StyleBox["T", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], "\:293a", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["ideal", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             StyleBox["T", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1181234284"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1181234284-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1181234284-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1181234284", "Proof|ID:1181234284-1"},
 CellID->508218087],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1181234284-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->906223251]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(tuple-ideal \[RightArrow]*)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(tuple-ideal \[RightArrow]*)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["h1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["h2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["h1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["h2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["h1", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["h2", "ExpressionVariable"]}], "\[Element]", 
             RowBox[{
               UnderscriptBox["ideal", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["T", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["h1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["h2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["h1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["d", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["T", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["h2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["h1", "ExpressionVariable"], 
               UnderscriptBox["-", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
               StyleBox["h2", "ExpressionVariable"]}], "\[Element]", 
             RowBox[{
               UnderscriptBox["ideal", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["T", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:352062174"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p352062174-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p352062174-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:352062174", "Proof|ID:352062174-1"},
 CellID->873828372],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p352062174-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->15597529]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Special Inference Rules of the ReductionRing Prover", "Section",
 CellChangeTimes->{{3.63161475495802*^9, 3.631614758994811*^9}, {
  3.6428423338079357`*^9, 3.642842339405526*^9}},
 CellID->1904633754],

Cell[CellGroupData[{

Cell["Integers", "Subsection",
 CellChangeTimes->{{3.642834998860224*^9, 3.642835022388319*^9}},
 CellID->1990308942],

Cell["\<\
The ReductionRing Prover contains one inference rule dedicated to proving \
membership of terms in integer-intervals, called \
\[OpenCurlyDoubleQuote]membershipIntegerInterval\[CloseCurlyDoubleQuote]. It \
behaves as follows:\
\>", "Text",
 CellChangeTimes->{{3.64283664980169*^9, 3.642836729887061*^9}},
 CellID->462744351],

Cell[CellGroupData[{

Cell[TextData[{
 "If the current goal is a conjunction, split it into two parts ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["C", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["C", "2"], TraditionalForm]]],
 ", where ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["C", "1"], TraditionalForm]]],
 " contains precisely those conjuncts that are (possibly universally \
quantified) formulas of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"a", ",", "\[Ellipsis]", ",", "b"}]]}], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["C", "2"], TraditionalForm]]],
 " contains the others.\[LineSeparator]If ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["C", "2"], TraditionalForm]]],
 " is non-empty, and AND-node with two sub-branches is created, where in the \
first branch the conjunction ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["C", "1"], TraditionalForm]]],
 " is proved and in the second branch the conjunction ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["C", "2"], TraditionalForm]]],
 " is proved, assuming ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["C", "1"], TraditionalForm]]],
 ".\[LineSeparator]If ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["C", "2"], TraditionalForm]]],
 " is empty, the integer-interval membership conditions in ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["C", "1"], TraditionalForm]]],
 " are proved according to the method described below."
}], "Item",
 CellChangeTimes->{{3.64283679079382*^9, 3.642836829535822*^9}, {
  3.64283687465127*^9, 3.642837085766221*^9}, {3.642837363111006*^9, 
  3.642837373917562*^9}},
 CellID->479202743],

Cell[TextData[{
 "If the current goal is a (possibly universally quantified) formula of the \
form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"a", ",", "\[Ellipsis]", ",", "b"}]]}], TraditionalForm]]],
 ", it is proved according to the method described below."
}], "Item",
 CellChangeTimes->{{3.64283679079382*^9, 3.642836829535822*^9}, {
  3.64283687465127*^9, 3.642837129652604*^9}, {3.6428373831580477`*^9, 
  3.6428373960686607`*^9}},
 CellID->1342681431],

Cell["In all other cases the inference rule is not applicable.", "Item",
 CellChangeTimes->{{3.64283679079382*^9, 3.642836829535822*^9}, {
  3.64283687465127*^9, 3.642837140034527*^9}},
 CellID->1243725426]
}, Open  ]],

Cell[TextData[{
 "The main method underlying the inference rule \
\[OpenCurlyDoubleQuote]membershipIntegerInterval\[CloseCurlyDoubleQuote] \
takes a list ",
 Cell[BoxData[
  FormBox["L", TraditionalForm]]],
 " of (possibly universally quantified) formulas of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"a", ",", "\[Ellipsis]", ",", "b"}]]}], TraditionalForm]]],
 " and returns a (simplified) list of formulas of the same form, possibly \
even the empty list. All members of ",
 Cell[BoxData[
  FormBox["L", TraditionalForm]]],
 " are simplified in turn, making use of known facts about the terms involved \
(\[OpenCurlyDoubleQuote]known\[CloseCurlyDoubleQuote] means: in the knowledge \
base of the current proof situation)."
}], "Text",
 CellChangeTimes->{{3.6428371455379877`*^9, 3.642837162557568*^9}, {
  3.642837273740365*^9, 3.642837344583066*^9}, {3.642837406800432*^9, 
  3.642837552677494*^9}},
 CellID->1287601136],

Cell[TextData[{
 "The following inferences are used in the simplification of a formula ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"F", "\[Congruent]", "t"}], "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"a", ",", "\[Ellipsis]", ",", "b"}]]}], TraditionalForm]]],
 ", possibly containing free variables arising from universal quantifiers; \
note that \[DoubleStruckCapitalZ] abbreviates ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", 
    RowBox[{
     RowBox[{"-", "\[Infinity]"}], ",", "\[Ellipsis]", ",", "\[Infinity]"}]], 
   TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.6428376342695017`*^9, 3.642837657365511*^9}, {
  3.642837693079494*^9, 3.642837731779564*^9}, {3.64283786766536*^9, 
  3.642837872512088*^9}, {3.642839476320177*^9, 3.6428395132916927`*^9}, {
  3.64283954584918*^9, 3.642839548489079*^9}},
 CellID->796616441],

Cell[CellGroupData[{

Cell[TextData[{
 "If all of ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " are integer literals or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[PlusMinus]", "\[Infinity]"}], TraditionalForm]]],
 ", domain membership is immediately decided using the definition of integer \
intervals."
}], "Item",
 CellChangeTimes->{{3.642837687429373*^9, 3.642837688700591*^9}, {
  3.642837734399891*^9, 3.642837841453065*^9}},
 CellID->735370936],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is \[Alpha]-equivalent to a formula that has already been dealt with, it \
is dropped from the list of formulas."
}], "Item",
 CellChangeTimes->{{3.642837687429373*^9, 3.642837688700591*^9}, {
  3.642837734399891*^9, 3.642837931133471*^9}},
 CellID->1962333974],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " follows readily from known facts, meaning that ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is already known, it is dropped as well."
}], "Item",
 CellChangeTimes->{{3.642837687429373*^9, 3.642837688700591*^9}, {
  3.642837734399891*^9, 3.642837995783476*^9}},
 CellID->2043017143],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Congruent]", 
    RowBox[{"b", "-", 
     RowBox[{"(", 
      RowBox[{"x", "-", "1"}], ")"}]}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Congruent]", "1"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", "\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", 
      RowBox[{"1", ",", "\[Ellipsis]", ",", "b"}]]}], "\[And]", 
    RowBox[{"b", "\[Element]", "\[DoubleStruckCapitalZ]"}]}], 
   TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.642837687429373*^9, 3.642837688700591*^9}, {
  3.642837734399891*^9, 3.642838108765029*^9}},
 CellID->1272531796],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Congruent]", "a", "\[Congruent]", "b"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Element]", "\[DoubleStruckCapitalZ]"}], 
   TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.642837687429373*^9, 3.642837688700591*^9}, {
  3.642837734399891*^9, 3.642838108743135*^9}, {3.642838151289876*^9, 
  3.642838201974786*^9}},
 CellID->1849308658],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Congruent]", "a"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " is different from ",
 Cell[BoxData[
  FormBox[
   RowBox[{"-", "\[Infinity]"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{
      RowBox[{"-", "\[Infinity]"}], ",", "\[Ellipsis]", ",", "b"}]]}], 
   TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.642837687429373*^9, 3.642837688700591*^9}, {
  3.642837734399891*^9, 3.642838108743135*^9}, {3.642838151289876*^9, 
  3.642838179963397*^9}, {3.642838217747542*^9, 3.6428382599368362`*^9}},
 CellID->1979661899],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Congruent]", "b"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " is different from ",
 Cell[BoxData[
  FormBox[
   RowBox[{"+", "\[Infinity]"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"a", ",", "\[Ellipsis]", ",", "\[Infinity]"}]]}], 
   TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.642837687429373*^9, 3.642837688700591*^9}, {
  3.642837734399891*^9, 3.642838108743135*^9}, {3.642838151289876*^9, 
  3.642838179963397*^9}, {3.642838217747542*^9, 3.642838321684824*^9}},
 CellID->1932244586],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Congruent]", 
    RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "\[Congruent]", 
    RowBox[{"+", "\[Infinity]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " is a non-positive number or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"-", "\[Infinity]"}], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " is known to be a tuple (isTuple) or a domain-tuple over some domain, ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is dropped."
}], "Item",
 CellChangeTimes->{{3.642837687429373*^9, 3.642837688700591*^9}, {
  3.642837734399891*^9, 3.642838108743135*^9}, {3.642838151289876*^9, 
  3.642838179963397*^9}, {3.642838217747542*^9, 3.642838446142878*^9}, {
  3.642838480961995*^9, 3.642838502622905*^9}},
 CellID->747553457],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " is known to be contained in a sub-interval of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", 
    RowBox[{"a", ",", "\[Ellipsis]", ",", "b"}]], TraditionalForm]]],
 ", either because this is a known fact or because this has been another \
condition to be proved, ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is dropped. Deciding membership in sub-intervals is accomplished by \
recursive calls to the main method: If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"G", "\[Congruent]", "t"}], "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"x", ",", "\[Ellipsis]", ",", "y"}]]}], TraditionalForm]]],
 " is known and both ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"a", ",", "\[Ellipsis]", ",", "\[Infinity]"}]]}], 
   TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"y", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{
      RowBox[{"-", "\[Infinity]"}], ",", "\[Ellipsis]", ",", "b"}]]}], 
   TraditionalForm]]],
 " can be proved, then ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"x", ",", "\[Ellipsis]", ",", "y"}]], "\[SubsetEqual]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"a", ",", "\[Ellipsis]", ",", "b"}]]}], TraditionalForm]]],
 " and hence ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " follows from ",
 Cell[BoxData[
  FormBox["G", TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.642837687429373*^9, 3.642837688700591*^9}, {
  3.642837734399891*^9, 3.642838108743135*^9}, {3.642838151289876*^9, 
  3.642838179963397*^9}, {3.642838217747542*^9, 3.642838446142878*^9}, {
  3.642838480961995*^9, 3.642838502601383*^9}, {3.642838555012216*^9, 
  3.6428386543801727`*^9}, {3.642838713023568*^9, 3.642838716401456*^9}, {
  3.642838763676468*^9, 3.6428389296137667`*^9}, {3.6428389932368526`*^9, 
  3.642838993262753*^9}},
 CellID->1851234931],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Congruent]", 
    RowBox[{"x", "+", "y"}]}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Congruent]", 
    RowBox[{"x", "*", "y"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " is either ",
 Cell[BoxData[
  FormBox[
   RowBox[{"-", "\[Infinity]"}], TraditionalForm]]],
 " or a non-negative integer, and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "\[Congruent]", 
    RowBox[{"+", "\[Infinity]"}]}], TraditionalForm]]],
 ", then ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", "\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", 
      RowBox[{"a", ",", "\[Ellipsis]\[Infinity]"}]]}], "\[And]", 
    RowBox[{"y", "\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", 
      RowBox[{"a", ",", "\[Ellipsis]", ",", "\[Infinity]"}]]}]}], 
   TraditionalForm]]],
 ". This is justified by the fact that both ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " are closed in \[DoubleStruckCapitalZ] as well as in intervals of the form ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", 
    RowBox[{"a", ",", "\[Ellipsis]", ",", "\[Infinity]"}]], TraditionalForm]]],
 " with ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " non-negative."
}], "Item",
 CellChangeTimes->{{3.642837687429373*^9, 3.642837688700591*^9}, {
  3.642837734399891*^9, 3.642838108743135*^9}, {3.642838151289876*^9, 
  3.642838179963397*^9}, {3.642838217747542*^9, 3.642838446142878*^9}, {
  3.642838480961995*^9, 3.642838502601383*^9}, {3.642838555012216*^9, 
  3.6428386543801727`*^9}, {3.642838713023568*^9, 3.642838716401456*^9}, {
  3.642838763676468*^9, 3.6428389296137667`*^9}, {3.6428389932368526`*^9, 
  3.642839184657292*^9}},
 CellID->1677361500],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Congruent]", 
    RowBox[{"x", "-", "y"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Congruent]", 
    RowBox[{"-", "\[Infinity]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "\[Congruent]", 
    RowBox[{"+", "\[Infinity]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", "\[Element]", "\[DoubleStruckCapitalZ]"}], "\[And]", 
    RowBox[{"y", "\[Element]", "\[DoubleStruckCapitalZ]"}]}], 
   TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.642837687429373*^9, 3.642837688700591*^9}, {
  3.642837734399891*^9, 3.642838108743135*^9}, {3.642838151289876*^9, 
  3.642838179963397*^9}, {3.642838217747542*^9, 3.642838446142878*^9}, {
  3.642838480961995*^9, 3.642838502601383*^9}, {3.642838555012216*^9, 
  3.6428386543801727`*^9}, {3.642838713023568*^9, 3.642838716401456*^9}, {
  3.642838763676468*^9, 3.6428389296137667`*^9}, {3.6428389932368526`*^9, 
  3.642839239551436*^9}, {3.642839305688216*^9, 3.6428393057110243`*^9}},
 CellID->1048482517],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Congruent]", 
    RowBox[{"-", "x"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Congruent]", 
    RowBox[{"-", "\[Infinity]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "\[Congruent]", 
    RowBox[{"+", "\[Infinity]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Element]", "\[DoubleStruckCapitalZ]"}], 
   TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.642837687429373*^9, 3.642837688700591*^9}, {
  3.642837734399891*^9, 3.642838108743135*^9}, {3.642838151289876*^9, 
  3.642838179963397*^9}, {3.642838217747542*^9, 3.642838446142878*^9}, {
  3.642838480961995*^9, 3.642838502601383*^9}, {3.642838555012216*^9, 
  3.6428386543801727`*^9}, {3.642838713023568*^9, 3.642838716401456*^9}, {
  3.642838763676468*^9, 3.6428389296137667`*^9}, {3.6428389932368526`*^9, 
  3.642839239551436*^9}, {3.642839305688216*^9, 3.6428393174233923`*^9}},
 CellID->1782956434]
}, Open  ]],

Cell["\<\
The inferences listed above are applied repeatedly to all members of the list \
of conditions, until no further reductions can be carried out any more. The \
list thus obtained is finally returned as the list of integer-interval \
conditions that could not be proved by rule \
\[OpenCurlyDoubleQuote]membershipIntegerIntervals\[CloseCurlyDoubleQuote].\
\>", "Text",
 CellChangeTimes->{{3.6428393318053417`*^9, 3.6428394579020557`*^9}},
 CellID->642945926]
}, Closed]],

Cell[CellGroupData[{

Cell["Order Relations", "Subsection",
 CellChangeTimes->{{3.6316147762485523`*^9, 3.6316147896047583`*^9}},
 CellID->1073443063],

Cell["\<\
We have inference rules for the following four types of order relations:\
\>", "Text",
 CellChangeTimes->{{3.635241596219327*^9, 3.63524160997528*^9}, {
  3.635242023991871*^9, 3.635242025488624*^9}},
 CellID->1263601321],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"isPartIrreflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 " is a partial irreflexive ordering on \[ScriptCapitalD]."
}], "Item",
 CellChangeTimes->{{3.6352416233289967`*^9, 3.635241641359166*^9}, {
  3.6352416877294683`*^9, 3.635241690183793*^9}, {3.635241747348597*^9, 
  3.6352417626357327`*^9}, {3.642839651831522*^9, 3.6428397492883167`*^9}, {
  3.642840092726582*^9, 3.6428401130071907`*^9}},
 CellID->804291745],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"isPartReflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 " is a partial reflexive ordering on \[ScriptCapitalD]."
}], "Item",
 CellChangeTimes->{{3.6352416233289967`*^9, 3.6352416517197733`*^9}, {
  3.635241765108036*^9, 3.635241791911099*^9}, {3.642839717278317*^9, 
  3.642839756903757*^9}, {3.642840118630232*^9, 3.642840140302402*^9}},
 CellID->1770398954],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"isTotalIrreflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 " is a total (i.e. linear) irreflexive ordering on \[ScriptCapitalD]."
}], "Item",
 CellChangeTimes->{{3.6352416233289967`*^9, 3.635241669806986*^9}, {
  3.6352417941989193`*^9, 3.635241826178244*^9}, {3.642839758874216*^9, 
  3.6428397864614153`*^9}, {3.64284014528564*^9, 3.6428401673262177`*^9}},
 CellID->180397152]
}, Open  ]],

Cell[TextData[{
 "Furthermore, any binary function \[SmallCircle] might have the property of \
being an ",
 StyleBox["order embedding",
  FontSlant->"Italic"],
 " w.r.t. ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 ", i.e. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"y", " ", "R", " ", "z"}], "\[DoubleLeftRightArrow]", 
    RowBox[{
     RowBox[{"x", "\[SmallCircle]", "y"}], " ", "R", " ", 
     RowBox[{"x", "\[SmallCircle]", "z"}]}]}], TraditionalForm]]],
 "; this is characterized by formula ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isOrderEmbedding", "[", 
    RowBox[{"\[SmallCircle]", 
     RowBox[{",", "R", ",", "\[ScriptCapitalD]"}]}], "]"}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.635482560140502*^9, 3.6354826709559803`*^9}, {
   3.63548368539589*^9, 3.635483698701785*^9}, 3.6356589844021177`*^9, {
   3.6428398003985653`*^9, 3.642839876732716*^9}, {3.642840172922683*^9, 
   3.642840174338242*^9}},
 CellID->681649974],

Cell["\<\
The predicates isPartIrreflOrder, isPartReflOrder, isTotalIrreflOrder and \
isOrderEmbedding are all defined in \[OpenCurlyDoubleQuote]ElementaryTheories\
\[CloseCurlyDoubleQuote].\
\>", "Text",
 CellChangeTimes->{{3.6428400118225517`*^9, 3.642840051474061*^9}},
 CellID->786078624],

Cell[CellGroupData[{

Cell["orderingGoal", "Subsubsection",
 CellChangeTimes->{{3.6428395935231028`*^9, 3.642839595575363*^9}},
 CellID->1300497641],

Cell[CellGroupData[{

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"x", " ", "R", " ", "x"}]}], TraditionalForm]], "Item"],
 " has to be shown and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isPartIrreflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTotalIrreflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 " is known for some domain \[ScriptCapitalD], show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 " instead."
}], "Subitem",
 CellChangeTimes->{{3.635658632799802*^9, 3.635658673803886*^9}, {
  3.6428398984509087`*^9, 3.64283989961096*^9}, {3.642840192602461*^9, 
  3.642840225423802*^9}},
 CellID->1618438998],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", " ", "R", " ", "x"}], TraditionalForm]]],
 " has to be shown and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isPartReflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 " is known for some domain \[ScriptCapitalD], show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 " instead."
}], "Subitem",
 CellChangeTimes->{{3.635658632799802*^9, 3.635658703836241*^9}, {
  3.6428399026458683`*^9, 3.642839903822639*^9}, {3.6428402317145023`*^9, 
  3.642840248571941*^9}},
 CellID->431516384],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"x", " ", "R", " ", "y"}]}], TraditionalForm]]],
 " has to be shown, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isPartIrreflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTotalIrreflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 " is known for some domain \[ScriptCapitalD], and there are ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["z", "1"], ",", "\[Ellipsis]", ",", 
    SubscriptBox["z", "n"]}], TraditionalForm]]],
 " s.t. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"y", " ", "R", " ", 
    SubscriptBox["z", "1"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["z", "i"], " ", "R", " ", 
    SubscriptBox["z", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["z", "n"], " ", "R", " ", "x"}], TraditionalForm]]],
 " are known / follow immediately, show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", 
     RowBox[{"x", ",", "y", ",", 
      SubscriptBox["z", "1"], ",", "\[Ellipsis]", ",", 
      SubscriptBox["z", "n"]}], "]"}]}], TraditionalForm]]],
 " instead (",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " may be 0)."
}], "Subitem",
 CellChangeTimes->{{3.635658632799802*^9, 3.635658821842244*^9}, {
  3.635659070894405*^9, 3.635659072217265*^9}, {3.64283991066088*^9, 
  3.642839911873575*^9}, {3.642840256558445*^9, 3.642840288155386*^9}},
 CellID->2121495888],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"x", " ", "R", " ", "y"}]}], TraditionalForm]]],
 " has to be shown, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isPartReflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 " is known for some domain \[ScriptCapitalD], and there are ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["z", "1"], ",", "\[Ellipsis]", ",", 
    SubscriptBox["z", "n"]}], TraditionalForm]]],
 " s.t. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"y", " ", "R", " ", 
    SubscriptBox["z", "1"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["z", "i"], " ", "R", " ", 
    SubscriptBox["z", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["z", "n"], " ", "R", " ", "x"}], TraditionalForm]]],
 " are known / follow immediately, show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", 
     RowBox[{"x", ",", "y", ",", 
      SubscriptBox["z", "1"], ",", "\[Ellipsis]", ",", 
      SubscriptBox["z", "n"]}], "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[NotEqual]", "y"}], TraditionalForm]]],
 " instead (",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " may be 0)."
}], "Subitem",
 CellChangeTimes->{{3.635658632799802*^9, 3.635658863572011*^9}, {
  3.6356590742367353`*^9, 3.635659074986267*^9}, {3.6428399171679583`*^9, 
  3.642839918427834*^9}, {3.642840298788479*^9, 3.642840313016918*^9}},
 CellID->1727189115],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", " ", "R", " ", "y"}], TraditionalForm]]],
 " has to be shown, any of the three possible ordering-predicates is known \
for ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 " and some domain \[ScriptCapitalD], and there are ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["z", "1"], ",", "\[Ellipsis]", ",", 
    SubscriptBox["z", "n"]}], TraditionalForm]]],
 " s.t. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", " ", "R", " ", 
    SubscriptBox["z", "1"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["z", "i"], " ", "R", " ", 
    SubscriptBox["z", 
     RowBox[{"i", "+", "1"}]]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["z", "n"], " ", "R", " ", "y"}], TraditionalForm]]],
 " are known / follow immediately, show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", 
     RowBox[{"x", ",", "y", ",", 
      SubscriptBox["z", "1"], ",", "\[Ellipsis]", ",", 
      SubscriptBox["z", "n"]}], "]"}]}], TraditionalForm]]],
 " instead (unless ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "=", "0"}], TraditionalForm]]],
 ", in which case no domain-membership conditions have to be shown)."
}], "Subitem",
 CellChangeTimes->{{3.635658632799802*^9, 3.635658944392345*^9}, {
  3.642839924826107*^9, 3.642839926053447*^9}, {3.642840323171818*^9, 
  3.6428403436572857`*^9}},
 CellID->513477240],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"z", "\[SmallCircle]", "x"}], " ", "R", " ", 
    RowBox[{"z", "\[SmallCircle]", "y"}]}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{
     RowBox[{"z", "\[SmallCircle]", "x"}], " ", "R", " ", 
     RowBox[{"z", "\[SmallCircle]", "y"}]}]}], TraditionalForm]]],
 " has to be shown and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isOrderEmbedding", "[", 
    RowBox[{"\[SmallCircle]", 
     RowBox[{",", "R", ",", "\[ScriptCapitalD]"}]}], "]"}], TraditionalForm]]],
 " is known for some domain \[ScriptCapitalD], show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", 
     RowBox[{"x", ",", "y", ",", "z"}], "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", " ", "R", " ", "y"}], TraditionalForm]]],
 " (resp. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"x", " ", "R", " ", "y"}]}], TraditionalForm]]],
 ") instead."
}], "Subitem",
 CellChangeTimes->{{3.635658632799802*^9, 3.635659006004209*^9}, {
  3.6356591440379257`*^9, 3.635659180387866*^9}, {3.6428399337117443`*^9, 
  3.6428399350255737`*^9}, {3.642840364556075*^9, 3.642840383174076*^9}},
 CellID->603058325]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["orderingKB", "Subsubsection",
 CellChangeTimes->{{3.6428396008484383`*^9, 3.642839602724246*^9}},
 CellID->559366505],

Cell[CellGroupData[{

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"x", " ", "R", " ", "x"}]}], TraditionalForm]], "Item"],
 " is known and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isPartReflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 " is known for some domain \[ScriptCapitalD], show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 " for deriving a contradiction."
}], "Subitem",
 CellChangeTimes->{{3.635659040899849*^9, 3.635659067916774*^9}, {
  3.642839964161023*^9, 3.642839981684214*^9}, {3.642840393226677*^9, 
  3.6428404095623302`*^9}},
 CellID->119517503],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", " ", "R", " ", "x"}], TraditionalForm]], "Item"],
 " is known and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isPartIrreflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTotalIrreflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 " is known for some domain \[ScriptCapitalD], show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 " for deriving a contradiction."
}], "Subitem",
 CellChangeTimes->{{3.635659040899849*^9, 3.635659091720675*^9}, {
  3.642839983990275*^9, 3.642839994907219*^9}, {3.6428404130609703`*^9, 
  3.642840441717079*^9}},
 CellID->1416532702]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["orderingEqualGoal", "Subsubsection",
 CellChangeTimes->{{3.6428396278796473`*^9, 3.6428396309973183`*^9}},
 CellID->45358433],

Cell[CellGroupData[{

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Equal]", "y"}], TraditionalForm]]],
 " has to be shown and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", "y", "]"}]}], TraditionalForm]]],
 " follows immediately for a domain \[ScriptCapitalD] with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isPartReflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 ", show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", 
     RowBox[{"x", ",", "y"}], "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", " ", "R", " ", "y"}], "\[And]", 
    RowBox[{"y", " ", "R", " ", "x"}]}], TraditionalForm]]],
 " instead."
}], "Subitem",
 CellChangeTimes->{{3.635483215206056*^9, 3.635483328281054*^9}, {
  3.6354833670507727`*^9, 3.635483368460527*^9}, {3.6354833992017193`*^9, 
  3.6354834518314342`*^9}, {3.635483728584382*^9, 3.635483728584509*^9}, {
  3.635484124971478*^9, 3.6354841655129223`*^9}, {3.6354842120615473`*^9, 
  3.635484233303052*^9}, {3.635659317991363*^9, 3.635659371164028*^9}, {
  3.642840061631278*^9, 3.642840064721374*^9}, {3.642840449818664*^9, 
  3.6428404710077953`*^9}},
 CellID->2027461248],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Equal]", "y"}], TraditionalForm]]],
 " has to be shown and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", "y", "]"}]}], TraditionalForm]]],
 " follows immediately for a domain \[ScriptCapitalD] with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTotalIrreflOrder", "[", 
    RowBox[{"R", ",", "\[ScriptCapitalD]"}], "]"}], TraditionalForm]]],
 ", show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", 
     RowBox[{"x", ",", "y"}], "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     RowBox[{"x", " ", "R", " ", "y"}]}], "\[And]", 
    RowBox[{"\[Not]", 
     RowBox[{"y", " ", "R", " ", "x"}]}]}], TraditionalForm]]],
 " instead."
}], "Subitem",
 CellChangeTimes->{{3.635483215206056*^9, 3.635483328281054*^9}, {
  3.6354833670507727`*^9, 3.635483368460527*^9}, {3.6354833992017193`*^9, 
  3.6354834518314342`*^9}, {3.635483728584382*^9, 3.635483728584509*^9}, {
  3.635484124971478*^9, 3.6354841655129223`*^9}, {3.6354842120615473`*^9, 
  3.635484233303052*^9}, {3.635659317991363*^9, 3.6356593913802*^9}, {
  3.642840484768293*^9, 3.642840495122911*^9}},
 CellID->1819110750]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Domain Collections", "Subsection",
 CellChangeTimes->{{3.642835042277081*^9, 3.642835045523785*^9}},
 CellID->1936978443],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]Domain Collections\[CloseCurlyDoubleQuote] are sets ",
 Cell[BoxData[
  FormBox["S", TraditionalForm]]],
 " and tuples ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " of elements of a particular domain \[ScriptCapitalD], characterized by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "S", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "T", "]"}]}], TraditionalForm]]],
 ", respectively. DomainSets and DomainTuples are two functors defined in \
\[OpenCurlyDoubleQuote]ElementaryTheories\[CloseCurlyDoubleQuote]; there is \
one inference rule for deciding domain membership for each of them."
}], "Text",
 CellChangeTimes->{{3.642840515566588*^9, 3.64284061136089*^9}, {
  3.642840647290326*^9, 3.642840689502364*^9}, {3.6428424935979433`*^9, 
  3.642842494865821*^9}},
 CellID->748233992],

Cell[CellGroupData[{

Cell["membershipDomainTuples", "Subsubsection",
 CellChangeTimes->{{3.6428407341268387`*^9, 3.642840738826632*^9}},
 CellID->997495398],

Cell[CellGroupData[{

Cell[TextData[{
 "If the goal is of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "T", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "T", "]"}]}], TraditionalForm]]],
 " is known for some domain \[ScriptCapitalD], the proof is finished."
}], "Item",
 CellChangeTimes->{{3.642840835500352*^9, 3.642840883593203*^9}, {
  3.642840928292109*^9, 3.642840928318516*^9}},
 CellID->1229202995],

Cell[TextData[{
 "If the goal is of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", 
     SubscriptBox["T", "i"], "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "T", "]"}]}], TraditionalForm]]],
 " follows immediately from the current knowledge, the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"1", ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
   TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.642840835500352*^9, 3.642840883593203*^9}, {
  3.642840928292109*^9, 3.642841031136016*^9}},
 CellID->100325772],

Cell[TextData[{
 "If the goal is of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 " and there is some ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " such that both ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    TagBox["\:22ff",
     Identity,
     SyntaxForm->"a+b"], "T"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "T", "]"}]}], TraditionalForm]]],
 " follow from the current knowledge base, the proof is finished."
}], "Item",
 CellChangeTimes->{{3.642840835500352*^9, 3.642840883593203*^9}, {
  3.642840928292109*^9, 3.6428410311071587`*^9}, {3.64284106148631*^9, 
  3.642841158173484*^9}},
 CellID->1483125180],

Cell[TextData[{
 "Finally, if the goal is of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "T", "]"}]}], TraditionalForm]]],
 ", various inferences are applied ",
 StyleBox["repeatedly",
  FontSlant->"Italic"],
 " to reduce the goal depending on the structure of ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 ":"
}], "Item",
 CellChangeTimes->{{3.642840835500352*^9, 3.642840883593203*^9}, {
  3.642840928292109*^9, 3.6428410311071587`*^9}, {3.64284106148631*^9, 
  3.6428412028845778`*^9}, {3.642841241632131*^9, 3.642841280944759*^9}, {
  3.642841997252801*^9, 3.6428420248380823`*^9}},
 CellID->553180495],

Cell[CellGroupData[{

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[Congruent]", 
    RowBox[{"A", 
     TagBox["\:22c8",
      Identity,
      SyntaxForm->"a*b"], "B", 
     TagBox["\:22c8",
      Identity,
      SyntaxForm->"a*b"], "\[Ellipsis]"}]}], TraditionalForm]], "Item"],
 ", the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
     RowBox[{"[", "A", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
     RowBox[{"[", "B", "]"}]}], "\[And]", "\[Ellipsis]"}], TraditionalForm]]]
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841389318153*^9}},
 CellID->1622345308],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[Congruent]", 
    RowBox[{"A", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "a"}]}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[Congruent]", 
    RowBox[{"a", 
     TagBox["\:293b",
      Identity,
      SyntaxForm->"a*b"], "A"}]}], TraditionalForm]]],
 ", the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
     RowBox[{"[", "A", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
     RowBox[{"[", "a", "]"}]}]}], TraditionalForm]]],
 "."
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.6428414619732513`*^9}},
 CellID->163263565],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[Congruent]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"a", ",", "b", ",", "\[Ellipsis]"}], "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ", the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
     RowBox[{"[", "a", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
     RowBox[{"[", "b", "]"}]}], "\[And]", "\[Ellipsis]"}], TraditionalForm]]]
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841445054398*^9}, {
  3.642841475962369*^9, 3.642841498960711*^9}, {3.646038369486578*^9, 
  3.646038369500791*^9}},
 CellID->935422534],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[Congruent]", 
    RowBox[{"Rest", "[", "A", "]"}]}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[Congruent]", 
    RowBox[{"Most", "[", "A", "]"}]}], TraditionalForm]]],
 ", the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
     RowBox[{"[", "A", "]"}]}], "\[And]", 
    RowBox[{"A", "\[NotEqual]", 
     RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
   TraditionalForm]]],
 "."
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841445054398*^9}, {
  3.642841475962369*^9, 3.642841498960711*^9}, {3.646038369486578*^9, 
  3.646038404096403*^9}},
 CellID->1645646521],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[Congruent]", 
    RowBox[{"Reverse", "[", "A", "]"}]}], TraditionalForm]]],
 ", the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "A", "]"}]}], TraditionalForm]]],
 "."
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841445054398*^9}, {
  3.642841475962369*^9, 3.642841498960711*^9}, {3.646038369486578*^9, 
  3.646038412664295*^9}},
 CellID->1885357983],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[Congruent]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"f", "[", 
       RowBox[{"b", "-", 
        RowBox[{"(", 
         RowBox[{"i", "-", "a"}], ")"}]}], "]"}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "a"}], ",", "\[Ellipsis]", ",", "b"}]], 
      RowBox[{"P", "[", 
       RowBox[{"b", "-", 
        RowBox[{"(", 
         RowBox[{"i", "-", "a"}], ")"}]}], "]"}]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 ", the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       RowBox[{"f", "[", "i", "]"}], 
       UnderscriptBox["|", 
        RowBox[{
         RowBox[{"i", "=", "a"}], ",", "\[Ellipsis]", ",", "b"}]], 
       RowBox[{"P", "[", "i", "]"}]}], "\[RightAngleBracket]"}], "]"}]}], 
   TraditionalForm]], "Item"],
 "."
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841445054398*^9}, {
  3.642841475962369*^9, 3.642841505666335*^9}, {3.642841573070245*^9, 
  3.642841643570059*^9}, {3.642841683550741*^9, 3.6428416953416452`*^9}},
 CellID->1697032445],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[Congruent]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      SubscriptBox["A", "i"], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "A", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 ", the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "A", "]"}]}], TraditionalForm]], "Item"],
 "."
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841445054398*^9}, {
  3.642841475962369*^9, 3.642841505666335*^9}, {3.642841573070245*^9, 
  3.642841643570059*^9}, {3.642841683550741*^9, 3.642841792893837*^9}},
 CellID->1531117068],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "\[Congruent]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"f", "[", "i", "]"}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "a"}], ",", "\[Ellipsis]", ",", "b"}]], 
      RowBox[{"P", "[", "i", "]"}]}], "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ", the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"i", "=", "a"}], ",", "\[Ellipsis]", ",", "b"}]], 
     RowBox[{"P", "[", "i", "]"}]], 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
     RowBox[{"[", 
      RowBox[{"f", "[", "i", "]"}], "]"}]}]}], TraditionalForm]], "Item"],
 "."
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841445054398*^9}, {
  3.642841475962369*^9, 3.642841505666335*^9}, {3.642841573070245*^9, 
  3.642841643570059*^9}, {3.642841683550741*^9, 3.642841842514655*^9}},
 CellID->2081397093],

Cell[TextData[{
 "In all other cases, if ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
     RowBox[{"[", 
      SubscriptBox["T", "i"], "]"}]}]}], TraditionalForm]]],
 " follows immediately from the current knowledge, the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "T", "]"}], TraditionalForm]]],
 "."
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841445054398*^9}, {
  3.642841475962369*^9, 3.642841505666335*^9}, {3.642841573070245*^9, 
  3.642841643570059*^9}, {3.642841683550741*^9, 3.6428418424041777`*^9}, {
  3.64284191146211*^9, 3.6428419578577843`*^9}},
 CellID->587920082]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["membershipDomainSets", "Subsubsection",
 CellChangeTimes->{{3.6428407513251057`*^9, 3.6428407551277847`*^9}},
 CellID->1978501032],

Cell[CellGroupData[{

Cell[TextData[{
 "If the goal is of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isSet", "[", "S", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "S", "]"}]}], TraditionalForm]]],
 " is known for some domain \[ScriptCapitalD], the proof is finished."
}], "Item",
 CellChangeTimes->{{3.642840835500352*^9, 3.642840883593203*^9}, {
  3.642840928292109*^9, 3.642840928318516*^9}, {3.642842040730218*^9, 
  3.6428420510291233`*^9}},
 CellID->1935590335],

Cell[TextData[{
 "If the goal is of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 " and there is some ",
 Cell[BoxData[
  FormBox["S", TraditionalForm]]],
 " such that both ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", 
    TagBox["\[Element]",
     Identity,
     SyntaxForm->"a+b"], "S"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "S", "]"}]}], TraditionalForm]]],
 " follow from the current knowledge base, the proof is finished."
}], "Item",
 CellChangeTimes->{{3.642840835500352*^9, 3.642840883593203*^9}, {
  3.642840928292109*^9, 3.6428410311071587`*^9}, {3.64284106148631*^9, 
  3.642841158173484*^9}, {3.642842075811062*^9, 3.6428420893875732`*^9}},
 CellID->135971146],

Cell[TextData[{
 "Finally, if the goal is of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "S", "]"}]}], TraditionalForm]]],
 ", various inferences are applied ",
 StyleBox["repeatedly",
  FontSlant->"Italic"],
 " to reduce the goal depending on the structure of ",
 Cell[BoxData[
  FormBox["S", TraditionalForm]]],
 ":"
}], "Item",
 CellChangeTimes->{{3.642840835500352*^9, 3.642840883593203*^9}, {
  3.642840928292109*^9, 3.6428410311071587`*^9}, {3.64284106148631*^9, 
  3.6428412028845778`*^9}, {3.642841241632131*^9, 3.642841280944759*^9}, {
  3.642841997252801*^9, 3.6428420248380823`*^9}, {3.642842093079585*^9, 
  3.642842101322886*^9}},
 CellID->1041358857],

Cell[CellGroupData[{

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", "\[Congruent]", 
    RowBox[{"A", 
     TagBox["\[Union]",
      Identity,
      SyntaxForm->"a*b"], "B", 
     TagBox["\[Union]",
      Identity,
      SyntaxForm->"a*b"], "\[Ellipsis]"}]}], TraditionalForm]], "Item"],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", "\[Congruent]", 
    RowBox[{"A", "\[Intersection]", "B", "\[Intersection]", "\[Ellipsis]"}]}],
    TraditionalForm]]],
 ", the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainSets", "[", "\[ScriptCapitalD]", "]"}]], 
     RowBox[{"[", "A", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainSets", "[", "\[ScriptCapitalD]", "]"}]], 
     RowBox[{"[", "B", "]"}]}], "\[And]", "\[Ellipsis]"}], TraditionalForm]]]
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841389318153*^9}, {
  3.642842123879158*^9, 3.642842154392899*^9}},
 CellID->1221678701],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", "\[Congruent]", 
    RowBox[{"{", 
     RowBox[{"a", ",", "b", ",", "\[Ellipsis]"}], "}"}]}], 
   TraditionalForm]]],
 ", the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
     RowBox[{"[", "a", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
     RowBox[{"[", "b", "]"}]}], "\[And]", "\[Ellipsis]"}], TraditionalForm]]]
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841445054398*^9}, {
  3.642841475962369*^9, 3.642841498960711*^9}, {3.642842174154687*^9, 
  3.642842176277342*^9}, {3.642842257441723*^9, 3.642842257442349*^9}},
 CellID->463970340],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", "\[Congruent]", 
    RowBox[{"{", 
     RowBox[{"x", 
      UnderscriptBox["|", 
       RowBox[{"x", "\[Element]", "A"}]]}], "}"}]}], TraditionalForm]]],
 ", the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSets", "[", "\[ScriptCapitalD]", "]"}]], 
    RowBox[{"[", "A", "]"}]}], TraditionalForm]], "Item"],
 "."
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841445054398*^9}, {
  3.642841475962369*^9, 3.642841505666335*^9}, {3.642841573070245*^9, 
  3.642841643570059*^9}, {3.642841683550741*^9, 3.642841792893837*^9}, {
  3.6428421992012167`*^9, 3.642842220963298*^9}, {3.642842258958295*^9, 
  3.64284225895885*^9}},
 CellID->851908008],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", "\[Congruent]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"f", "[", "i", "]"}], 
      UnderscriptBox["|", "i"], 
      RowBox[{"P", "[", "i", "]"}]}], "}"}]}], TraditionalForm]]],
 ", the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", "i"], 
     RowBox[{"P", "[", "i", "]"}]], 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
     RowBox[{"[", 
      RowBox[{"f", "[", "i", "]"}], "]"}]}]}], TraditionalForm]], "Item"],
 "."
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841445054398*^9}, {
  3.642841475962369*^9, 3.642841505666335*^9}, {3.642841573070245*^9, 
  3.642841643570059*^9}, {3.642841683550741*^9, 3.642841842514655*^9}, {
  3.6428422231673326`*^9, 3.642842259858437*^9}},
 CellID->2068916673],

Cell[TextData[{
 "In all other cases, if ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"x", "\[Element]", "S"}]], 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
     RowBox[{"[", "x", "]"}]}]}], TraditionalForm]]],
 " follows immediately from the current knowledge, the goal is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isSet", "[", "S", "]"}], TraditionalForm]]],
 "."
}], "Subitem",
 CellChangeTimes->{{3.6428412879026413`*^9, 3.642841445054398*^9}, {
  3.642841475962369*^9, 3.642841505666335*^9}, {3.642841573070245*^9, 
  3.642841643570059*^9}, {3.642841683550741*^9, 3.6428418424041777`*^9}, {
  3.64284191146211*^9, 3.6428419578577843`*^9}, {3.6428422654865026`*^9, 
  3.6428422760256042`*^9}},
 CellID->268417535]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Cancellative Commutative Monoids", "Subsection",
 CellChangeTimes->{{3.635482297120469*^9, 3.6354823107799473`*^9}},
 CellID->1165507851],

Cell[TextData[{
 "Cancellative commutative monoids are algebraic structures ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"\[ScriptCapitalM]", ",", 
     RowBox[{"\[SmallCircle]", 
      RowBox[{",", "n"}]}]}], ")"}], TraditionalForm]]],
 " satisfying the axioms of commutative monoids as well as the cancellativity \
axiom ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", "\[SmallCircle]", "z"}], "=", 
    RowBox[{
     RowBox[{"y", "\[SmallCircle]", 
      RowBox[{"z", "\[DoubleLongLeftRightArrow]", "x"}]}], "=", "y"}]}], 
   TraditionalForm]]],
 "; Their definition can be found in \
\[OpenCurlyDoubleQuote]ElementaryTheories\[CloseCurlyDoubleQuote]."
}], "Text",
 CellChangeTimes->{{3.6428423528794537`*^9, 3.6428423942140102`*^9}, {
  3.642842426345677*^9, 3.642842488043445*^9}, {3.642843589026248*^9, 
  3.642843590357008*^9}},
 CellID->1769579981],

Cell[TextData[{
 "The ReductionRing prover contains two special inference rules dedicated to \
deal with such cancellative commutative monoids: One for proving membership \
of terms, and one for proving equality of terms. In both inference rules, \
whenever the formula \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCCMonoid", "[", 
    RowBox[{"\[ScriptCapitalM]", ",", 
     RowBox[{"\[SmallCircle]", 
      RowBox[{",", "n"}]}]}], "]"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] is required to be known, it is sufficient if \
actually only ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCommPPDomain", "[", "\[ScriptCapitalM]", "]"}], 
   TraditionalForm]]],
 " is known, since commutative power-product domains are, by definition, also \
cancellative commutative monoids."
}], "Text",
 CellChangeTimes->{{3.6428425082550163`*^9, 3.642842567344521*^9}, {
  3.642842641865122*^9, 3.6428427755984983`*^9}, {3.6428428534368467`*^9, 
  3.6428428565569973`*^9}, {3.642843598488388*^9, 3.642843598743627*^9}},
 CellID->96195350],

Cell[CellGroupData[{

Cell["membershipCCMonoids", "Subsubsection",
 CellChangeTimes->{{3.642842591672009*^9, 3.642842595877644*^9}},
 CellID->887510921],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]membershipCCMonoids\[CloseCurlyDoubleQuote] is \
applicable if the current goal is an individual domain-membership condition, \
a conjunction thereof, or even a conjunction involving also conjuncts of a \
different kind. In the latter case the same strategy of splitting the goal \
into two subgoals as in inference rule \
\[OpenCurlyDoubleQuote]membershipIntegerInterval\[CloseCurlyDoubleQuote] is \
employed, so assume now that the goal consists of one or more conditions of \
the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalM]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 " for some domain \[ScriptCapitalM] with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCCMonoid", "[", 
    RowBox[{"\[ScriptCapitalM]", ",", 
     RowBox[{"\[SmallCircle]", 
      RowBox[{",", "n"}]}]}], "]"}], TraditionalForm]]],
 "; just as in \[OpenCurlyDoubleQuote]membershipIntegerInterval\
\[CloseCurlyDoubleQuote] these conditions may be universally quantified."
}], "Text",
 CellChangeTimes->{{3.642842909259328*^9, 3.642843178582189*^9}, {
  3.6428436048440037`*^9, 3.6428436051064377`*^9}},
 CellID->1397392423],

Cell[TextData[{
 "The main idea behind \[OpenCurlyDoubleQuote]membershipCCMonoids\
\[CloseCurlyDoubleQuote] is to exploit closure of \[SmallCircle] in \
\[ScriptCapitalM], i.e. if both ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " are known to belong to \[ScriptCapitalM], then so does ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[SmallCircle]", "y"}], TraditionalForm]]],
 ". It is important to note, however, that the converse direction does not \
have to hold (and, thus, is not used in the inference rule): If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalM]"], 
    RowBox[{"[", 
     RowBox[{"x", "\[SmallCircle]", "y"}], "]"}]}], TraditionalForm]]],
 " is known then neither ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalM]"], 
    RowBox[{"[", "x", "]"}]}], TraditionalForm]]],
 " nor ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalM]"], 
    RowBox[{"[", "y", "]"}]}], TraditionalForm]]],
 " can be inferred from that!\nThe following are the inferences used for \
repeatedly simplifying a list of conditions of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"F", "\[Congruent]"}], 
    UnderscriptBox["\[Element]", "\[ScriptCapitalM]"], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 ", where ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " may contain free variables originating from universal quantifiers:"
}], "Text",
 CellChangeTimes->{{3.642843330196806*^9, 3.642843493314027*^9}, {
  3.642843617886648*^9, 3.642843675791687*^9}},
 CellID->1248471043],

Cell[CellGroupData[{

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is \[Alpha]-equivalent to another condition in the list, ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is dropped from the list."
}], "Item",
 CellChangeTimes->{{3.6428436961182833`*^9, 3.642843727032069*^9}, {
  3.642843984953795*^9, 3.6428439849775257`*^9}},
 CellID->1688735138],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " follows immediately from the current knowledge, ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is dropped from the list."
}], "Item",
 CellChangeTimes->{{3.6428436961182833`*^9, 3.642843727032069*^9}, {
  3.642843984953795*^9, 3.642844006317603*^9}},
 CellID->457866193],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Congruent]", "n"}], TraditionalForm]]],
 " (i.e. the neutral element), ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is dropped from the list."
}], "Item",
 CellChangeTimes->{{3.6428436961182833`*^9, 3.642843727004591*^9}, {
  3.64284376487868*^9, 3.642843795893279*^9}},
 CellID->1024908712],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "\[Congruent]", 
    RowBox[{"a", "\[SmallCircle]", "b", "\[SmallCircle]", "\[Ellipsis]"}]}], 
   TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["F", TraditionalForm]]],
 " is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalM]"], 
     RowBox[{"[", "a", "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalM]"], 
     RowBox[{"[", "b", "]"}]}], "\[And]", "\[Ellipsis]"}], TraditionalForm]]]
}], "Item",
 CellChangeTimes->{{3.6428436961182833`*^9, 3.642843727004591*^9}, {
  3.64284376487868*^9, 3.642843795867447*^9}, {3.642843935282324*^9, 
  3.642843983001007*^9}},
 CellID->553257142]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["CCMonoidEqual", "Subsubsection",
 CellChangeTimes->{{3.642842606812806*^9, 3.6428426096884537`*^9}},
 CellID->1487291863],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]CCMonoidEqual\[CloseCurlyDoubleQuote] is applicable \
if the current goal is an equality ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", 
    RowBox[{"b", "=", "\[Ellipsis]"}]}], TraditionalForm]]],
 ", where the outermost symbol of at least one of the terms ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", ",", "b", ",", "\[Ellipsis]"}], TraditionalForm]]],
 " is either \[SmallCircle] or ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " for a constant \[ScriptCapitalM] with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCCMonoid", "[", 
    RowBox[{"\[ScriptCapitalM]", ",", 
     RowBox[{"\[SmallCircle]", 
      RowBox[{",", "n"}]}]}], "]"}], TraditionalForm]]],
 ". In that case, the equality ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", 
    RowBox[{"b", "=", "\[Ellipsis]"}]}], TraditionalForm]]],
 " is decomposed into individual binary equalities ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", "b"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "=", "\[Ellipsis]"}], TraditionalForm]]],
 " etc., all of them being treated individually."
}], "Text",
 CellChangeTimes->{{3.6428440560988493`*^9, 3.6428441647154617`*^9}, {
  3.6428442552879877`*^9, 3.642844339514003*^9}},
 CellID->453356605],

Cell[TextData[{
 "Assume now we are given a binary equality ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", "b"}], TraditionalForm]]],
 ". The following is the procedure for determining whether the equality holds \
or not:"
}], "Text",
 CellChangeTimes->{{3.642844375184622*^9, 3.642844435068544*^9}},
 CellID->1829142219],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " is traversed in breadth-first manner, as long as the head of the current \
subexpression is \[SmallCircle]."
}], "Item",
 CellChangeTimes->{{3.6428445437469587`*^9, 3.642844593739236*^9}},
 CellID->12335311],

Cell[TextData[{
 "Let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[NotCongruent]", "n"}], TraditionalForm]]],
 " be one of these subexpressions: Now, ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " is searched for an \[Alpha]-equivalent occurrence of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " at a position where all outer function symbols are \[SmallCircle]. If an \
occurrence ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " is found, both ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " are replaced by ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 ", respectively. At the same time, all subexpressions appearing in an \
argument list of any of the outer functions in ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 ", respectively, (including ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ") are collected in a separate list ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.6428445437469587`*^9, 3.6428448272947903`*^9}, {
  3.642844859264291*^9, 3.642844859297443*^9}, {3.6428453162205343`*^9, 
  3.642845326678101*^9}, {3.642845367037163*^9, 3.6428453955230913`*^9}},
 CellID->1349486020],

Cell[TextData[{
 "Finally, occurrences of ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " are eliminated in both ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 ", by repeatedly replacing subexpressions of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"n", "\[SmallCircle]", "n", "\[SmallCircle]"}], "..."}], 
   TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       SubscriptBox["u", "1"], "\[SmallCircle]"}], "..."}], "\[SmallCircle]", 
     
     SubscriptBox["u", "m"], "\[SmallCircle]", "n", "\[SmallCircle]", 
     SubscriptBox["v", "1"], "\[SmallCircle]"}], "..."}], TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       SubscriptBox["u", "1"], "\[SmallCircle]"}], "..."}], "\[SmallCircle]", 
     
     SubscriptBox["u", "m"], "\[SmallCircle]", 
     SubscriptBox["v", "1"], "\[SmallCircle]"}], "..."}], TraditionalForm]]],
 " (of course, only if all outer function symbols are also \[SmallCircle]). \
Again, all subexpressions appearing in an argument list of any of the outer \
functions in ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 ", respectively, are added to list ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " (this includes in particular the ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["u", "i"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["v", "j"], TraditionalForm]]],
 ")."
}], "Item",
 CellChangeTimes->{{3.6428445437469587`*^9, 3.6428448272947903`*^9}, {
  3.642844859264291*^9, 3.642845086330188*^9}, {3.642845422949511*^9, 
  3.6428454756907988`*^9}},
 CellID->1253926265]
}, Open  ]],

Cell[TextData[{
 "The steps sketched above yield two (ideally simpler) expressions ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "'"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "'"}], TraditionalForm]]],
 " together with a term-list ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ", with the semantics \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", "b"}], TraditionalForm]]],
 " follows from ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "'"}], "=", 
    RowBox[{"b", "'"}]}], TraditionalForm]]],
 " if all elements of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " belong to \[ScriptCapitalM]\[CloseCurlyDoubleQuote]. Hence, the original \
equality ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", "b"}], TraditionalForm]]],
 " is reduced to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "'"}], "=", 
    RowBox[{"b", "'"}]}], TraditionalForm]]],
 ", but in addition a new branch in the proof tree has to be created, where \
domain-membership of all elements of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " is proved (unless ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " is empty)."
}], "Text",
 CellChangeTimes->{{3.642845115731699*^9, 3.642845174250855*^9}, {
  3.6428454843686237`*^9, 3.6428455711475487`*^9}, {3.642845634979374*^9, 
  3.642845798302085*^9}},
 CellID->1045187834],

Cell[TextData[{
 StyleBox["Remark",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 ": Repeatedly adding terms to a list ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " might appear a bit cumbersome at first sight, especially if one is aware \
that typically many expressions thus added are in fact subexpressions of \
other elements already contained in ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ". So, why not just immediately prove domain-membership of ",
 StyleBox["all",
  FontSlant->"Italic"],
 " atomic subterms of ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 "? The best way of answering this question is by means of an example:\n\
Assume the equality one wishes to prove is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"a", "\[SmallCircle]", "b"}], ")"}], "\[SmallCircle]", 
     RowBox[{"(", 
      RowBox[{"c", "\[SmallCircle]", "d"}], ")"}]}], "=", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"a", "\[SmallCircle]", "b"}], ")"}], "\[SmallCircle]", "c"}], 
      ")"}], "\[SmallCircle]", "d"}]}], TraditionalForm]]],
 ". The equality can easily be proved, provided that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[SmallCircle]", "b"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["d", TraditionalForm]]],
 " belong to \[ScriptCapitalM] \[Dash] so, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[SmallCircle]", "b"}], TraditionalForm]]],
 " does not have to be split into ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 "! If domain-membership of these three terms can be proved, everything is \
fine. However, it might well be that domain-membership of either ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " ",
 StyleBox["cannot",
  FontSlant->"Italic"],
 " be proved (see inference rule \[OpenCurlyDoubleQuote]membershipCCMonoid\
\[CloseCurlyDoubleQuote]), meaning that a strategy that simply proves \
domain-membership of all atoms would fail in that case.\nThis remark also \
applies to inference rule \[OpenCurlyDoubleQuote]commRing1Equal\
\[CloseCurlyDoubleQuote]."
}], "Text",
 CellChangeTimes->{{3.642845847694207*^9, 3.6428461964710913`*^9}, {
   3.6428462316758757`*^9, 3.6428467416449833`*^9}, 3.6428467777971163`*^9},
 CellID->576772764]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Commutative Rings with Unit", "Subsection",
 CellChangeTimes->{{3.631963858503702*^9, 3.631963864929185*^9}},
 CellID->1418614091],

Cell["\<\
Just as for cancellative commutative monoids there are basically two \
inference rules for working in the theory of commutative rings with unit: One \
for proving membership of terms, and one for proving equality of terms. The \
overall ideas behind the ring-rules are precisely the same as for the \
monoid-rules, so we do not rephrase them here (e.g. try to avoid proving \
domain-membership of terms that actually do not necessarily have to belong to \
the ring, when proving equality of two terms; see the concluding remark on \
inference rule \[OpenCurlyDoubleQuote]CCMonoidEqual\[CloseCurlyDoubleQuote]). \
Instead, we point out the main differences, in particular what makes matters \
much more complicated when dealing with commutative rings rather than \
cancellative commutative monoids.\
\>", "Text",
 CellChangeTimes->{{3.642846932746585*^9, 3.642846934696563*^9}, {
  3.6428469782461863`*^9, 3.642846988587981*^9}, {3.642847063499222*^9, 
  3.6428473984428387`*^9}, {3.642847661946211*^9, 3.642847663188012*^9}},
 CellID->1029877200],

Cell["\<\
Essentially, there are two main differences between the ring-rules and the \
monoid-rules:\
\>", "Text",
 CellChangeTimes->{{3.642847818285429*^9, 3.6428478471126833`*^9}},
 CellID->1993995419],

Cell[CellGroupData[{

Cell[TextData[{
 "In rings there are ",
 StyleBox["two",
  FontSlant->"Italic"],
 " operations ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " satisfying a distributivity law."
}], "Item",
 CellChangeTimes->{{3.642847853899514*^9, 3.642847894607836*^9}},
 CellID->258759428],

Cell[TextData[{
 "In reduction ring theory we also have to deal with ",
 StyleBox["finite",
  FontSlant->"Italic"],
 " ",
 StyleBox["sums",
  FontSlant->"Italic"],
 " in rings. Hence, the inference rules must be powerful enough to handle \
them."
}], "Item",
 CellChangeTimes->{{3.642847853899514*^9, 3.6428479804020987`*^9}},
 CellID->49710978]
}, Open  ]],

Cell[TextData[{
 "The presence of two operations instead of just one can be handled rather \
easily, in all inference rules: Whenever no other reductions or \
simplifications can be carried out any more, distributivity is used for \
expanding occurrences of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "*", 
    RowBox[{"(", 
     RowBox[{"y", "+", "z", "+", "\[Ellipsis]"}], ")"}]}], 
   TraditionalForm]]],
 " to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", "*", "y"}], "+", 
    RowBox[{"x", "*", "z"}], "+", "\[Ellipsis]"}], TraditionalForm]]],
 " and of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "*", 
    RowBox[{
     UnderscriptBox["\[Sum]", 
      RowBox[{
       RowBox[{"i", "=", "a"}], ",", "\[Ellipsis]", ",", "b"}]], 
     RowBox[{"f", "[", "i", "]"}]}]}], TraditionalForm]]],
 " to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Sum]", 
     RowBox[{
      RowBox[{"i", "=", "a"}], ",", "\[Ellipsis]", ",", "b"}]], 
    RowBox[{"x", "*", 
     RowBox[{"f", "[", "i", "]"}]}]}], TraditionalForm]]],
 ". The only real drawback of distributivity is that is might lead to an \
exponential blow-up of expressions, and this is the reason why it is applied \
carefully."
}], "Text",
 CellChangeTimes->{{3.642848025515868*^9, 3.642848390173428*^9}, {
  3.6428484523572607`*^9, 3.642848514824027*^9}, {3.642848800205014*^9, 
  3.642848818219823*^9}},
 CellID->588352486],

Cell["\<\
Finite sums cannot be handled as easily; Their treatment is explained below. \
Let in the sequel \[ScriptCapitalR] be a commutative ring with unit.\
\>", "Text",
 CellChangeTimes->{{3.6428484043810883`*^9, 3.6428484193698807`*^9}, {
  3.642848528034389*^9, 3.642848538221753*^9}, {3.642848994190503*^9, 
  3.642849007203005*^9}},
 CellID->878513273],

Cell[CellGroupData[{

Cell["membershipCommRing1", "Subsubsection",
 CellChangeTimes->{{3.64284762499111*^9, 3.642847629040647*^9}},
 CellID->769440561],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{
      UnderscriptBox["\[Sum]", 
       RowBox[{
        RowBox[{"i", "=", "a"}], ",", "\[Ellipsis]", ",", "b"}]], 
      RowBox[{"f", "[", "i", "]"}]}], "]"}]}], TraditionalForm]]],
 " is reduced to the conjunction of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     RowBox[{"f", "[", "i", "]"}], "]"}]}], TraditionalForm]]],
 " (knowing that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"a", ",", "\[Ellipsis]", ",", "b"}]]}], TraditionalForm]]],
 ", for the free variable ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 ") and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "\[Element]", "\[DoubleStruckCapitalZ]"}], "\[And]", 
    RowBox[{"b", "\[Element]", "\[DoubleStruckCapitalZ]"}]}], 
   TraditionalForm]]],
 ". If ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " has a particular form, e.g. is an integer literal or of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}], 
   TraditionalForm]]],
 " for a tuple ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 ", then the respective conditions are dropped immediately."
}], "Item",
 CellChangeTimes->{{3.642848971969981*^9, 3.64284901359861*^9}, {
  3.6428490978894*^9, 3.6428491728174*^9}, {3.642849208150188*^9, 
  3.64284934427794*^9}},
 CellID->98336303],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", 
     SubscriptBox["T", "i"], "]"}]}], TraditionalForm]]],
 " is reduced to the conjunction of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
    RowBox[{"[", "T", "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"1", ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
   TraditionalForm]]],
 ". If one of the conjunctions is already known or follows from local \
assumptions (e.g. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"1", ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]]}], 
   TraditionalForm]]],
 " might be known if ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 " is a free variable bound by a sum-binder; see above) it is immediately \
dropped.\[LineSeparator]Apparently, this inference resembles one of the \
inferences also used in rule \[OpenCurlyDoubleQuote]membershipDomainTuples\
\[CloseCurlyDoubleQuote]."
}], "Item",
 CellChangeTimes->{{3.642848971969981*^9, 3.64284901359861*^9}, {
  3.6428490978894*^9, 3.6428491728174*^9}, {3.642849208150188*^9, 
  3.642849349442237*^9}, {3.642849388433577*^9, 3.6428495632776947`*^9}},
 CellID->1714525659]
}, Open  ]],

Cell["\<\
As can be seen, due to the presence of finite sums ring-membership conditions \
may be reduced to conditions of a different kind, namely integer-interval- \
and DomainTuples-membership conditions. Simple inferences try to eliminate \
such conditions immediately (for instances, duplicate conditions are removed \
as soon as possible), but ultimately many of them will be left as additional \
subgoals to be dealt with by their dedicated inference rules, like \
\[OpenCurlyDoubleQuote]membershipIntegerInterval\[CloseCurlyDoubleQuote].\
\>", "Text",
 CellChangeTimes->{{3.642849598397732*^9, 3.642849823225986*^9}},
 CellID->75786986]
}, Open  ]],

Cell[CellGroupData[{

Cell["commRing1Equal", "Subsubsection",
 CellChangeTimes->{{3.642847634532914*^9, 3.6428476373431*^9}},
 CellID->1855157544],

Cell[TextData[{
 "Assume we are given ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "=", "b"}], TraditionalForm]]],
 ", where the outermost function symbol of ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " is a function in the commutative ring with unit \[ScriptCapitalR] (i.e. it \
is ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["-", TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox["\[Sum]", TraditionalForm]]],
 ")."
}], "Text",
 CellChangeTimes->{{3.642908861203446*^9, 3.6429089260167713`*^9}},
 CellID->116443813],

Cell[CellGroupData[{

Cell[TextData[{
 "First, subterms appearing both in ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and in ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " are replaced by 0, following precisely the same procedure as in \
\[OpenCurlyDoubleQuote]CCMonoidEqual\[CloseCurlyDoubleQuote] with ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " in the role of \[SmallCircle], ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "-", "y"}], TraditionalForm]]],
 " replaced by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "+", 
    RowBox[{"(", 
     RowBox[{"-", "y"}], ")"}]}], TraditionalForm]]],
 ", and sums considered as atomic. Negative subterms, i.e. terms with an odd \
number of outer unary \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox["-", TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] are moved to the other side to get rid of the \
\[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox["-", TraditionalForm]]],
 "\[CloseCurlyDoubleQuote], and associativity and commutativity of ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " are taken into account when searching for matching subexpressions. Of \
course, all terms that have to belong to \[ScriptCapitalR] in order for a \
transformation to be applicable are collected."
}], "Item",
 CellChangeTimes->{{3.642908948141107*^9, 3.6429091751393023`*^9}, {
  3.642909507228304*^9, 3.642909614209144*^9}},
 CellID->2102803037],

Cell[TextData[{
 "If the resulting equality ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "'"}], "=", 
    RowBox[{"b", "'"}]}], TraditionalForm]]],
 " is not yet trivial (i.e. the two sides are \[Alpha]-equivalent), the two \
sides are fully expanded using distributivity of ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " w.r.t. ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " (both in ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", "*", 
     RowBox[{"(", 
      RowBox[{"y", "+", "z"}], ")"}]}], "=", 
    RowBox[{
     RowBox[{"x", "*", "y"}], "+", 
     RowBox[{"x", "*", "z"}]}]}], TraditionalForm]]],
 " and in ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"x", "*", 
     RowBox[{
      UnderscriptBox["\[Sum]", "i"], 
      RowBox[{"f", "[", "i", "]"}]}]}], "=", 
    RowBox[{
     SubscriptBox["\[Sum]", "i"], 
     RowBox[{"x", "*", 
      RowBox[{"f", "[", "i", "]"}]}]}]}], TraditionalForm]]],
 ") and associativity/commutativity of ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["\[Sum]", "i"], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"f", "[", "i", "]"}], "+", 
       RowBox[{"g", "[", "i", "]"}]}], ")"}]}], "=", 
    RowBox[{
     RowBox[{
      SubscriptBox["\[Sum]", "i"], 
      RowBox[{"f", "[", "i", "]"}]}], "+", 
     RowBox[{
      SubscriptBox["\[Sum]", "i"], 
      RowBox[{"g", "[", "i", "]"}]}]}]}], TraditionalForm]]],
 ". Summands are expanded recursively, and once again all terms whose \
membership in \[ScriptCapitalR] is required for a transformation to be \
applicable are collected."
}], "Item",
 CellChangeTimes->{{3.642908948141107*^9, 3.64290911695643*^9}, {
  3.6429091930125427`*^9, 3.642909417671546*^9}, {3.642909701750494*^9, 
  3.642909732210162*^9}},
 CellID->1589163826],

Cell[TextData[{
 "Finally, the same method as in the first step is applied to the resulting \
equality ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "''"}], "=", 
    RowBox[{"b", "''"}]}], TraditionalForm]]],
 ". This means, for instance, that sums are now considered atomic again."
}], "Item",
 CellChangeTimes->{{3.642908948141107*^9, 3.64290911695643*^9}, {
  3.6429091930125427`*^9, 3.642909467864958*^9}},
 CellID->1023190292]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["commRing1EqualKB", "Subsubsection",
 CellChangeTimes->{{3.64284771832716*^9, 3.642847721967889*^9}},
 CellID->228482158],

Cell["\<\
This inference rule simplifies known equalities. Internally, it does exactly \
the same as \[OpenCurlyDoubleQuote]commRing1Equal\[CloseCurlyDoubleQuote].\
\>", "Text",
 CellChangeTimes->{{3.642847728985724*^9, 3.642847776799498*^9}},
 CellID->348032396]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Open  ]]
},
WindowSize->{1111, 763},
WindowMargins->{{0, Automatic}, {0, Automatic}},
ShowSelection->True,
CounterAssignments->{{"TheoremaFormulaCounter" -> 20}},
Magnification->1.,
FrontEndVersion->"9.0 for Linux x86 (64-bit) (February 7, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Theorema"}, "TheoremaCore.nb"]]], 
   
   Cell["Modify the definitions below to change the default appearance of all \
cells in a given style. \nMake modifications to any definition using commands \
in the Format menu.", "Text"], 
   Cell[
    CellGroupData[{
      Cell["Notebook Options Settings", "Section"], 
      Cell[
      "The options defined for the style below will be used at the Notebook \
level.", "Text"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Notebook"], InputAliases -> {"and" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[And]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "or" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", "\[Or]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "not" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Not]", "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "impl" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Implies]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[DoubleLongLeftRightArrow]", 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                "\[SelectionPlaceholder]", "\[Equal]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox[
                  
                  RowBox[{
                   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                  "a:=b"], "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", ":=", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "andn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[And]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "orn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Or]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "equivn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[DoubleLeftRightArrow]", "\[Piecewise]", 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "cdist" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Piecewise]", 
                 
                 GridBox[{{
                   "\[SelectionPlaceholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}, {
                   "\[Placeholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "far" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exr" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "farc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exrc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "let" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["let", 
                  RowBox[{"\[Placeholder]", "=", "\[Placeholder]"}]], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "app" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293a", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "prep" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293b", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "join" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22c8", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "elemT" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22ff", Identity, SyntaxForm -> "a+b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "subop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "supop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SuperscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "subsupop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubsuperscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "oop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "uoop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderoverscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "osop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox[
                OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
                "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "op" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               TagBox["\[SelectionPlaceholder]", Identity], 
               TagBox[")", "AutoParentheses"]}], "[]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e114", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e115", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "()" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:fd3e", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:fd3f", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e8", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e9", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e117", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e118", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "[[]]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e6", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e7", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(())" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2e28", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2e29", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<<>>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27ea", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27eb", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{{}}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2983", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2984", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(" -> 
            TagBox["(", "AutoParentheses"], ")" -> 
            TagBox[")", "AutoParentheses"]}, Background -> 
          RGBColor[0.957, 0.96, 0.97]], 
         Cell[
          StyleData["Notebook", "Printout"], Background -> None]}, Open]]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell["Style Environment Names", "Section"], 
      Cell[
       StyleData[All, "Working"], DockedCells -> {
         Cell[
         "Theorema 2.0", "SBO", FontTracking -> 8, 
          CellMargins -> {{0, 0}, {0, 0}}, 
          CellFrameMargins -> {{0, 50}, {0, 0}}, FontSize -> 28, FontSlant -> 
          "Plain", FontColor -> RGBColor[0.29, 0.46, 0.6], TextAlignment -> 
          Right, Background -> RGBColor[0.36, 0.54, 0.69]]}], 
      Cell[
       StyleData[
       All, "Layout", StyleDefinitions -> StyleData[All, "Working"]], 
       MenuSortingValue -> 1200]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Styles for Title and Section Cells", "Section"], 
      Cell[
       CellGroupData[{
         Cell["Title", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], CellMargins -> {{27, 27}, {10, 30}}, 
             TextAlignment -> Center, MenuSortingValue -> 1100, 
             MenuCommandKey -> "1", FontFamily -> "Bitstream Charter", 
             FontSize -> 36, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Title", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{6, 10}, {5, 30}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], CellMargins -> {{27, 27}, {20, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1150, FontFamily -> 
             "Bitstream Charter", FontSize -> 24, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{6, 10}, {5, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], CellMargins -> {{27, 27}, {8, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], CellMargins -> {{27, 27}, {18, 20}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Sectioning", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], ShowGroupOpener -> True, MenuSortingValue -> 
             1250, MenuCommandKey -> "4", FontFamily -> "Bitstream Charter", 
             FontSize -> 20, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Section", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{2, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {8, 12}}, 
             MenuSortingValue -> 1300, MenuCommandKey -> "5", FontFamily -> 
             "Bitstream Charter", FontSize -> 14, FontWeight -> "Bold", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Subsection", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{9, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {2, 10}}, 
             MenuSortingValue -> 1350, MenuCommandKey -> "6", FontFamily -> 
             "Bitstream Charter", FontSize -> 12, FontWeight -> "Bold", 
             FontSlant -> "Italic", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsection", "Printout"], ShowGroupOpener -> False,
              CellMargins -> {{35, 0}, {7, 14}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsubsection"], CellDingbat -> None, 
             ShowGroupOpener -> True, 
             CellMargins -> {{27, Inherited}, {Inherited, Inherited}}, 
             FontFamily -> "Bitstream Charter", FontSize -> 11, FontWeight -> 
             "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsubsection", "Printout"], 
             CellMargins -> {{50, 0}, {7, 14}}]}, Closed]]}, Open]]}, Open]], 
   
   Cell[
    CellGroupData[{
      Cell["Styles for Publication\[Hyphen]ready Documents", "Section"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Author"], CellMargins -> {{27, 30}, {20, 5}}, 
          MenuSortingValue -> 1900, FontFamily -> "Helvetica", FontSize -> 14,
           FontSlant -> "Italic"], 
         Cell[
          StyleData["Author", "Printout"], CellMargins -> {{6, 10}, {0, 10}}, 
          FontSize -> 10]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Abstract"], CellFrame -> {{0, 0}, {0.5, 0.5}}, 
          CellMargins -> {{27, 140}, {20, 20}}, LineSpacing -> {0.9, 0}, 
          MenuSortingValue -> 2050, FontFamily -> "Times", FontSize -> 12], 
         Cell[
          StyleData["Abstract", "Printout"], Hyphenation -> True, 
          LineSpacing -> {1, 2}, FontSize -> 10]}, Closed]]}, Closed]], 
   Cell[
    CellGroupData[{
      Cell["Text", "Section"], 
      Cell[
       StyleData["Text"], CellMargins -> {{27, 27}, {Inherited, Inherited}}, 
       TextJustification -> 1., ParagraphIndent -> 15, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["EnvironmentText"], ShowCellBracket -> Automatic, 
       CellMargins -> {{82, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextJustification -> 1., ParagraphSpacing -> {0.3, 0.}, 
       MenuSortingValue -> 1480, MenuCommandKey -> "T", FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Item"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "Item", 
          CounterAssignments -> {{"Subitem", 0}, {"Subsubitem", 0}}, 
          MenuSortingValue -> 1600, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Item", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemParagraph"], CellMargins -> {{81, 10}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          DefaultNewCellStyle -> "Item", MenuSortingValue -> 1600, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemParagraph", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "Subitem", 
          CounterAssignments -> {{"Subsubitem", 0}}, MenuSortingValue -> 1610,
           FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subitem", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemParagraph"], CellMargins -> {{105, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subitem", 
          MenuSortingValue -> 1610, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemParagraph", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subsubitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1620, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subsubitem", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemParagraph"], 
          CellMargins -> {{129, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1625, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemParagraph", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "ItemNumbered", 
          CounterAssignments -> {{"SubitemNumbered", 0}, {
            "SubsubitemNumbered", 0}}, MenuSortingValue -> 1630, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemNumbered", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "SubitemNumbered", 
          CounterAssignments -> {{"SubsubitemNumbered", 0}}, MenuSortingValue -> 
          1640, FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemNumbered", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], ".", 
              CounterBox["SubsubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> 
          "SubitemNumbered", CounterIncrements -> "SubsubitemNumbered", 
          MenuSortingValue -> 1645, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemNumbered", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Theorema", "Section"], 
      Cell[
       StyleData["GlobalDeclaration"], CellFrame -> {{0, 2}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 3}, {2, 2}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       InitializationCell -> True, CellFrameMargins -> 10, CellFrameColor -> 
       RGBColor[0.73, 0.65, 0.61], CellFrameLabelMargins -> {{33, 7}, {1, 1}},
        ShowAutoStyles -> False, MenuSortingValue -> 1441, MenuCommandKey -> 
       "G", FontSize -> 14], 
      Cell[
       StyleData["OpenEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {-4, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       StyleData["EnvironmentHeader"], 
       CellFrame -> {{False, False}, {False, False}}, 
       CellMargins -> {{76, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 6, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       ShowAutoStyles -> False, MenuSortingValue -> 1450, FontSize -> 14, 
       FontWeight -> "Bold", FontVariations -> {"CapsType" -> "SmallCaps"}, 
       Background -> RGBColor[0.89, 0.8, 0.69]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FormalTextInputFormula"], ShowCellBracket -> Automatic, 
          CellMargins -> {{96, 10}, {2, 2}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
          InitializationCell -> True, CellFrameLabelMargins -> 53, 
          ShowAutoStyles -> False, LimitsPositioningTokens -> {}, 
          MenuSortingValue -> 1470, MenuCommandKey -> "F", FontWeight -> 
          "Bold", FontColor -> RGBColor[0.25, 0.19, 0.2]], 
         Cell[
          StyleData["FormalTextInputFormula", "Layout"], Editable -> False], 
         Cell[
          StyleData["FormalTextInputFormula", "Presentation"], Editable -> 
          False]}, Open]], 
      Cell[
       StyleData["EndEnvironmentMarker"], 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextAlignment -> Right, MenuSortingValue -> 1480], 
      Cell[
       StyleData["CloseEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {False, True}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {Inherited, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Computation"], CellFrame -> {{1.5, 0}, {0, 0}}, 
          ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 7}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
          CellFrameMargins -> {{8, 8}, {3, 4}}, CellFrameColor -> 
          RGBColor[0.486, 0.33, 0.32], ShowAutoStyles -> False, 
          MenuSortingValue -> 1510, FontWeight -> "Bold", FontColor -> 
          RGBColor[0.23, 0.25, 0.34], Background -> 
          RGBColor[0.88, 0.88, 0.92]], 
         Cell[
          StyleData["Computation", "Layout"], Editable -> False], 
         Cell[
          StyleData["Computation", "Presentation"], Editable -> False]}, 
        Open]], 
      Cell[
       StyleData["ComputationResult"], CellFrame -> {{1.5, 0}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
       CellFrameMargins -> {{8, 8}, {4, 3}}, CellFrameColor -> 
       RGBColor[0.486, 0.33, 0.32], StyleMenuListing -> None, FontColor -> 
       RGBColor[0.23, 0.25, 0.34], Background -> RGBColor[0.88, 0.88, 0.92]], 
      
      Cell[
       StyleData["ComputationInfo"], ShowCellBracket -> False, 
       CellMargins -> {{120, 10}, {10, 2}}, StyleMenuListing -> None, 
       FontFamily -> $CellContext`Arial], 
      Cell[
       StyleData["ComputationInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{144, 10}, {10, 2}}, StyleMenuListing -> None], 
      Cell[
       StyleData["ProofInfo"], ShowCellBracket -> False, 
       CellMargins -> {{76, 10}, {6, 6}}, CellFrameLabelMargins -> 15, 
       StyleMenuListing -> None, FontFamily -> "Arial"], 
      Cell[
       StyleData["ProofInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{100, 10}, {0, 0}}, StyleMenuListing -> None], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FrameLabel"], LanguageCategory -> None, StyleMenuListing -> 
          None, FontFamily -> "Times New Roman", FontSize -> 10, FontWeight -> 
          "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
          FontVariations -> {
           "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
            "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
          RGBColor[0.55, 0.44, 0.42], Background -> None], 
         Cell[
          StyleData["FrameLabel", "Printout"]]}, Open]], 
      Cell[
       StyleData["OpenArchive"], CellFrame -> {{False, False}, {False, True}},
        ShowCellBracket -> Automatic, CellMargins -> {{70, 70}, {0, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 20}, CellFrameMargins -> 3, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], CellFrameLabelMargins -> 1, 
       DefaultFormatType -> DefaultInputFormatType, ShowAutoStyles -> False, 
       MenuSortingValue -> 1210, FontWeight -> Bold, Background -> 
       RGBColor[0.73, 0.74, 0.84]], 
      Cell[
       StyleData["ArchiveInfo"], ShowCellBracket -> Automatic, 
       CellMargins -> {{70, 70}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 5, CellFrameLabelMargins -> 3, ShowCellTags -> 
       False, ShowAutoStyles -> False, MenuSortingValue -> 1220, FontWeight -> 
       "Bold", Background -> RGBColor[0.88, 0.88, 0.92]], 
      Cell[
       StyleData["CloseArchive"], 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> 
       Automatic, CellMargins -> {{0, 0}, {5, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 25}, CellFrameMargins -> 2, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], TextAlignment -> Center, 
       MenuSortingValue -> 1230, FontFamily -> "Arial Black", FontSize -> 24, 
       FontWeight -> "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       Background -> RGBColor[0.73, 0.65, 0.61]], 
      Cell[
       StyleData["IncludeArchive"], CellFrame -> {{4, 0}, {False, False}}, 
       ShowCellBracket -> Automatic, 
       CellMargins -> {{81, 100}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameColor -> RGBColor[0.55, 0.44, 0.42], 
       CellFrameLabels -> {{"\[RightSkeleton]", None}, {None, None}}, 
       CellFrameLabelMargins -> 25, MenuSortingValue -> 1445], 
      Cell[
       StyleData["ExpressionVariable"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionABF"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionMeta"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["AutoParentheses"], Editable -> False, SpanMaxSize -> 3, 
          FontOpacity -> 0.15], 
         Cell[
          StyleData["AutoParentheses", "Layout"], FontOpacity -> 0], 
         Cell[
          StyleData["AutoParentheses", "SlideShow"], FontOpacity -> 0]}, 
        Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Emphasized"], MenuCommandKey -> "E", FontSlant -> 
          "Italic"], 
         Cell[
          StyleData["Emphasized", "SlideShow"], FontSlant -> "Plain", 
          FontColor -> RGBColor[0.36, 0.54, 0.69]]}, Open]], 
      Cell[
       StyleData["GenericButton"], FontFamily -> "Bitstream Charter", 
       FontWeight -> "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
       FontVariations -> {
        "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
         "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
       RGBColor[0.23, 0.25, 0.34]], 
      Cell[
       StyleData["TheoremaForm"], "TwoByteSyntaxCharacterAutoReplacement" -> 
       True, LineSpacing -> {1.25, 0, 2}, SingleLetterItalics -> True, 
       StyleMenuListing -> None, FontFamily -> "Courier"], 
      Cell[
       StyleData["ComponentActive"], FontFamily -> "Times New Roman"], 
      Cell[
       StyleData["ComponentInactive"], FontFamily -> "Times New Roman", 
       FontOpacity -> 0.4], 
      Cell[
       StyleData["ComponentEmpty"], FontFamily -> "Times New Roman", 
       FontVariations -> {"StrikeThrough" -> True}], 
      Cell[
       StyleData["TabView"], FontFamily -> "Times New Roman", FontSize -> 10], 
      Cell[
       StyleData["TabViewLabel"], FontFamily -> "Times New Roman", FontSize -> 
       10, FontWeight -> "Bold"], 
      Cell[
       StyleData["FormReference", StyleDefinitions -> StyleData["Text"]], 
       FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> "Plain",
        FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["FormReferenceHover", StyleDefinitions -> StyleData["Text"]],
        FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> 
       "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> True}, 
       FontColor -> RGBColor[0.404, 0.43, 0.545]], 
      Cell[
       StyleData["DisplayFormula"], CellMargins -> {{36, 6}, {7, 7}}, 
       ConversionRules :> {
        "HTML" -> {{"", ""}, {"<p class=\"Equation\">", "</p>"}}}, 
       CellHorizontalScrolling -> True, DefaultFormatType -> TraditionalForm, 
       InputAutoReplacements -> {}, TextAlignment -> Center, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       ButtonBoxOptions -> {Appearance -> {Automatic, None}}]}, Open]]}, 
  Visible -> False, FrontEndVersion -> 
  "9.0 for Linux x86 (64-bit) (February 7, 2013)", StyleDefinitions -> 
  "StylesheetFormatting.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "ID:2066005206"->{
  Cell[95732, 2441, 628, 16, 30, "GlobalDeclaration",
   CellTags->{
    "ID:2066005206", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2066005206]},
 "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"->{
  Cell[95732, 2441, 628, 16, 30, "GlobalDeclaration",
   CellTags->{
    "ID:2066005206", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2066005206],
  Cell[96956, 2477, 1338, 27, 33, "GlobalDeclaration",
   CellTags->{
    "ID:457420627", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->457420627],
  Cell[98297, 2506, 1000, 23, 31, "GlobalDeclaration",
   CellTags->{
    "ID:176465234", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->176465234],
  Cell[99300, 2531, 1892, 54, 58, "FormalTextInputFormula",
   CellTags->{
    "ID:478302118", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction with several multipliers"},
   CellID->478302118],
  Cell[101715, 2607, 2024, 58, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:448370877", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetric closure of reducibility"},
   CellID->448370877],
  Cell[103742, 2667, 1383, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1537297721", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reducible1"},
   CellID->1537297721],
  Cell[105128, 2708, 1599, 47, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:115218056", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reducible2"},
   CellID->115218056],
  Cell[106730, 2757, 1894, 50, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:399781455", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "irreducibility"},
   CellID->399781455],
  Cell[108627, 2809, 2205, 58, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1867242863", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "irrelative"},
   CellID->1867242863],
  Cell[110835, 2869, 1640, 52, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1834165766", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative"},
   CellID->1834165766],
  Cell[112478, 2923, 2270, 63, 82, "FormalTextInputFormula",
   CellTags->{
    "ID:1907029362", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common reducible"},
   CellID->1907029362],
  Cell[114751, 2988, 3534, 94, 203, "FormalTextInputFormula",
   CellTags->{
    "ID:2013437879", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "non-trivial common reducible"},
   CellID->2013437879],
  Cell[118288, 3084, 2119, 60, 83, "FormalTextInputFormula",
   CellTags->{
    "ID:991122907", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "minimal non-trivial common reducible"},
   CellID->991122907],
  Cell[120410, 3146, 2361, 72, 125, "FormalTextInputFormula",
   CellTags->{
    "ID:1293246145", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "unary non-trivial common reducible"},
   CellID->1293246145],
  Cell[122774, 3220, 1982, 59, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:894750987", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "unary minimal non-trivial common reducible"},
   CellID->894750987],
  Cell[124759, 3281, 2543, 66, 109, "FormalTextInputFormula",
   CellTags->{
    "ID:779388596", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "critical pair multipliers"},
   CellID->779388596],
  Cell[128069, 3376, 1897, 34, 31, "GlobalDeclaration",
   CellTags->{
    "ID:1517092650", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1517092650],
  Cell[129969, 3412, 2121, 68, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:935236171", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo element using multiplier"},
   CellID->935236171],
  Cell[132093, 3482, 2134, 69, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:320624715", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo set"},
   CellID->320624715],
  Cell[134230, 3553, 3140, 84, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:434515171", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo tuple"},
   CellID->434515171],
  Cell[137373, 3639, 1426, 43, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:773828270", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReductor"},
   CellID->773828270],
  Cell[139428, 3700, 2877, 88, 97, "FormalTextInputFormula",
   CellTags->{
    "ID:1813157585", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "reflexive-transitive closure of reduction relation"},
   CellID->1813157585],
  Cell[142308, 3790, 2110, 68, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:370202704", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetric closure of reduction relation"},
   CellID->370202704],
  Cell[144421, 3860, 2820, 86, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:992563737", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "symmetric-reflexive-transitive closure of reduction relation"},
   CellID->992563737],
  Cell[147244, 3948, 2467, 79, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:414012405", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common successor w.r.t. reduction relation"},
   CellID->414012405],
  Cell[149714, 4029, 2406, 79, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:895393520", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexive-transitive closure of common successor"},
   CellID->895393520],
  Cell[152123, 4110, 2426, 79, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:586157110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common predecessor w.r.t. reduction relation"},
   CellID->586157110],
  Cell[154552, 4191, 2239, 74, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:2045822907", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "reflexive-transitive closure of common predecessor"},
   CellID->2045822907],
  Cell[156794, 4267, 3941, 110, 96, "FormalTextInputFormula",
   CellTags->{
    "ID:463604795", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below"},
   CellID->463604795],
  Cell[160738, 4379, 6169, 173, 240, "FormalTextInputFormula",
   CellTags->{
    "ID:1818365967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde] multiplier"},
   CellID->1818365967],
  Cell[166910, 4554, 2226, 67, 58, "FormalTextInputFormula",
   CellTags->{
    "ID:1533920470", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde]"},
   CellID->1533920470],
  Cell[171113, 4685, 789, 21, 44, "GlobalDeclaration",
   CellTags->{
    "ID:81135181", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->81135181],
  Cell[171905, 4708, 1143, 34, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1593192085", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red"},
   CellID->1593192085],
  Cell[173051, 4744, 1434, 45, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:808948861", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset]"},
   CellID->808948861],
  Cell[174488, 4791, 770, 21, 44, "GlobalDeclaration",
   CellTags->{
    "ID:301449743", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->301449743],
  Cell[175261, 4814, 902, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:218271861", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red tuples"},
   CellID->218271861],
  Cell[176166, 4841, 1437, 44, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:20184186", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] tuples"},
   CellID->20184186],
  Cell[178419, 4914, 652, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2086101595", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2086101595],
  Cell[179880, 4958, 987, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:690041737", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets isReductor"},
   CellID->690041737],
  Cell[180870, 4988, 985, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:5128300", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainTuples isReductor"},
   CellID->5128300],
  Cell[189002, 5236, 2813, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:906005529", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo singleton", "IR"},
   CellID->906005529],
  Cell[197753, 5487, 720, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:708647877", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->708647877],
  Cell[198476, 5507, 767, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:508075012", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->508075012],
  Cell[199246, 5529, 1551, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:347299573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo tuple alternative", "IR"},
   CellID->347299573],
  Cell[200800, 5580, 708, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1756538453", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1756538453],
  Cell[201511, 5600, 1161, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:200776309", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] TupleToSet", "IR"},
   CellID->200776309],
  Cell[202675, 5638, 1187, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:989570066", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] TupleToSet", "IR"},
   CellID->989570066],
  Cell[203865, 5676, 1181, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:359412397", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow]\[RawStar] TupleToSet", "IR"},
   CellID->359412397],
  Cell[205049, 5714, 1212, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1876144273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow]\[RawStar] TupleToSet", "IR"},
   CellID->1876144273],
  Cell[206264, 5752, 1944, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:495509420", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below TupleToSet", "IR"},
   CellID->495509420],
  Cell[239630, 6674, 672, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:467637718", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->467637718],
  Cell[240305, 6693, 795, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1182167952", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1182167952],
  Cell[241103, 6716, 1401, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:178557746", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo set alternative", "IR"},
   CellID->178557746],
  Cell[242507, 6763, 1169, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1019233204", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[Union]"},
   CellID->1019233204],
  Cell[243679, 6802, 1194, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:331036245", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] \[Union]"},
   CellID->331036245],
  Cell[244876, 6841, 1185, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:78564808", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow]\[RawStar] \[Union]"},
   CellID->78564808],
  Cell[246064, 6880, 1264, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:626278069", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow]\[RawStar] \[Union]"},
   CellID->626278069],
  Cell[247331, 6920, 1922, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:356849840", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below \[Union]"},
   CellID->356849840],
  Cell[282453, 7886, 952, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:366153355", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->366153355],
  Cell[283408, 7912, 699, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:221340803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->221340803],
  Cell[284110, 7932, 1219, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:328195853", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "2"},
   CellID->328195853],
  Cell[285332, 7965, 997, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1692465705", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "3"},
   CellID->1692465705],
  Cell[286332, 7995, 996, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:188743948", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "4"},
   CellID->188743948],
  Cell[287331, 8025, 1338, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:115635911", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "5"},
   CellID->115635911],
  Cell[288672, 8068, 1054, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1118510950", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isCommonReducible symmetric"},
   CellID->1118510950],
  Cell[289729, 8098, 1007, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:113485898", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isNTCR symmetric"},
   CellID->113485898],
  Cell[290739, 8128, 1016, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2056330192", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isMNTCR symmetric"},
   CellID->2056330192],
  Cell[291758, 8158, 1420, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:463339584", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers symmetric"},
   CellID->463339584],
  Cell[293181, 8202, 1867, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:449270875", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers sufficient"},
   CellID->449270875],
  Cell[354664, 9768, 751, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:523863903", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->523863903],
  Cell[355418, 9790, 789, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1519187240", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1519187240],
  Cell[356210, 9812, 1052, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1878546324", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "2'"},
   CellID->1878546324],
  Cell[357265, 9842, 1025, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1781176225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "3'"},
   CellID->1781176225],
  Cell[358293, 9872, 1032, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1582995786", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "4'"},
   CellID->1582995786],
  Cell[359328, 9902, 1370, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2020067658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "5'"},
   CellID->2020067658],
  Cell[388163, 10660, 730, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1823284442", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1823284442],
  Cell[388896, 10682, 762, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:284217868", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->284217868],
  Cell[389661, 10704, 1035, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:628244416", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "6"},
   CellID->628244416],
  Cell[390699, 10732, 2058, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1255486983", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative-irrelative 2"},
   CellID->1255486983],
  Cell[408854, 11206, 740, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1382908475", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1382908475],
  Cell[410158, 11244, 853, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:442895874", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->442895874],
  Cell[411014, 11268, 816, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:927154658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->927154658],
  Cell[411833, 11291, 867, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1300106055", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM 0"},
   CellID->1300106055],
  Cell[412703, 11317, 1223, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:438711618", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM 1"},
   CellID->438711618],
  Cell[413929, 11356, 1524, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2101232329", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM alternative"},
   CellID->2101232329],
  Cell[415456, 11403, 1487, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:96556784", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM+1"},
   CellID->96556784],
  Cell[416946, 11447, 1843, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1465630278", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM alternative II"},
   CellID->1465630278],
  Cell[418792, 11505, 1152, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1952076718", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "membership reduceM", "IR"},
   CellID->1952076718],
  Cell[419947, 11539, 792, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1110504237", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1110504237],
  Cell[420742, 11562, 1225, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1236615595", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM-1"},
   CellID->1236615595],
  Cell[421970, 11600, 1127, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:203149419", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM-1 II"},
   CellID->203149419],
  Cell[423100, 11635, 2109, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:333495152", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible reduceM"},
   CellID->333495152],
  Cell[425212, 11701, 2187, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:332101094", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM"},
   CellID->332101094],
  Cell[427402, 11768, 3277, 99, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:296464255", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM I"},
   CellID->296464255],
  Cell[430682, 11869, 2682, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:194930655", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM II"},
   CellID->194930655],
  Cell[433367, 11952, 2870, 87, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:935171195", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM III"},
   CellID->935171195],
  Cell[560259, 15118, 887, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1368508872", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1368508872],
  Cell[561149, 15143, 1430, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1610984937", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areIrrelative symmetric"},
   CellID->1610984937],
  Cell[562582, 15185, 1060, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:981548331", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCorrelative symmetric"},
   CellID->981548331],
  Cell[577102, 15568, 976, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1547543437", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1547543437],
  Cell[578081, 15594, 1300, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1699369800", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible \[RightArrow]"},
   CellID->1699369800],
  Cell[579384, 15634, 1377, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1596540807", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] isReducible"},
   CellID->1596540807],
  Cell[580764, 15679, 1092, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1641388157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible symred"},
   CellID->1641388157],
  Cell[581859, 15712, 1479, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2087655986", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible alternative", "IR"},
   CellID->2087655986],
  Cell[583341, 15758, 2777, 85, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:584152369", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers \[RightArrow] tuple"},
   CellID->584152369],
  Cell[621060, 16762, 846, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:593421628", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->593421628],
  Cell[621909, 16787, 1248, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:680005293", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "1"},
   CellID->680005293],
  Cell[623160, 16829, 1078, 33, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1320396264", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1320396264],
  Cell[624241, 16864, 1185, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1127232822", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] reductor"},
   CellID->1127232822],
  Cell[625429, 16901, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:256297052", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Succeeds] reductor"},
   CellID->256297052],
  Cell[643610, 17397, 1085, 26, 70, "GlobalDeclaration",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:1083403145", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1083403145],
  Cell[644698, 17425, 1061, 28, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:1592577568", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets red", "IR"},
   CellID->1592577568],
  Cell[645762, 17455, 965, 25, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:698444072", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red isSet", "IR"},
   CellID->698444072],
  Cell[646730, 17482, 1781, 53, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:277656140", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red"},
   CellID->277656140],
  Cell[648514, 17537, 1772, 54, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:454978652", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red singleton"},
   CellID->454978652],
  Cell[650289, 17593, 3071, 92, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:259827218", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red abstraction term"},
   CellID->259827218],
  Cell[653363, 17687, 1554, 44, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:1273500442", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red \[Union]"},
   CellID->1273500442],
  Cell[654920, 17733, 1143, 31, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:1877397922", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Union]"},
   CellID->1877397922],
  Cell[656066, 17766, 2110, 64, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:2121666706", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red \[Union] singleton"},
   CellID->2121666706],
  Cell[658179, 17832, 1088, 29, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:947744040", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SubsetEqual] red \[SubsetEqual]"},
   CellID->947744040],
  Cell[659270, 17863, 2045, 61, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:52516181", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset]"},
   CellID->52516181],
  Cell[661318, 17926, 2248, 69, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:194436663", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset] \[Union] singleton"},
   CellID->194436663],
  Cell[663569, 17997, 1794, 53, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:844584535", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset] \[Union] singleton II"},
   CellID->844584535],
  Cell[734698, 19903, 2199, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:632054408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[SubsetEqual] transitive"},
   CellID->632054408],
  Cell[736900, 19972, 772, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:387695114", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->387695114],
  Cell[738190, 20011, 1984, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1177850337", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "setSeq subsequence red"},
   CellID->1177850337],
  Cell[740177, 20074, 714, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1483708233", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1483708233],
  Cell[740894, 20094, 1060, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:697433658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red setSeq \[SubsetEqual]"},
   CellID->697433658],
  Cell[741957, 20126, 2089, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:35999503", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red setSeq"},
   CellID->35999503],
  Cell[744049, 20191, 2200, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:602168712", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red setSeq \[Subset]"},
   CellID->602168712],
  Cell[781414, 21184, 2578, 83, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:21782685", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] append", "IR"},
   CellID->21782685],
  Cell[783995, 21269, 1204, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1790989103", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isPartIrreflOrder"},
   CellID->1790989103],
  Cell[826130, 22593, 653, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1466856847", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1466856847],
  Cell[827340, 22628, 2804, 87, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:623701544", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction ring"},
   CellID->623701544],
  Cell[830147, 22717, 2665, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1960646232", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "multipliers"},
   CellID->1960646232],
  Cell[832815, 22800, 1460, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1496065688", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R0"},
   CellID->1496065688],
  Cell[834278, 22847, 1639, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:975892043", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R1"},
   CellID->975892043],
  Cell[835920, 22900, 1617, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1150364990", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R2"},
   CellID->1150364990],
  Cell[837540, 22952, 2946, 86, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:852396718", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R3"},
   CellID->852396718],
  Cell[840489, 23040, 1517, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1907869025", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R4"},
   CellID->1907869025],
  Cell[842009, 23087, 2702, 77, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1446921308", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R5"},
   CellID->1446921308],
  Cell[844714, 23166, 3054, 87, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208630606", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6"},
   CellID->208630606],
  Cell[847771, 23255, 10540, 264, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1150983847", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R7"},
   CellID->1150983847],
  Cell[858314, 23521, 7036, 180, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2106857486", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R8"},
   CellID->2106857486],
  Cell[865971, 23727, 928, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1197875862", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1197875862],
  Cell[866902, 23750, 3522, 98, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1949920189", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R5 witnesses"},
   CellID->1949920189],
  Cell[870427, 23850, 2974, 83, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2027349511", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6 witness"},
   CellID->2027349511],
  Cell[873708, 23949, 693, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1717911774", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1717911774],
  Cell[875100, 23987, 771, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:349001278", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->349001278],
  Cell[875874, 24008, 1348, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:474542129", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "9"},
   CellID->474542129],
  Cell[877225, 24050, 1468, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:360788203", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "21"},
   CellID->360788203],
  Cell[888993, 24391, 674, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:998446180", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->998446180],
  Cell[889670, 24410, 940, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1581977856", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1581977856],
  Cell[890613, 24437, 1026, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:623334744", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "8"},
   CellID->623334744],
  Cell[891642, 24470, 1591, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1112133758", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "8'"},
   CellID->1112133758],
  Cell[893236, 24522, 727, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:810098043", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->810098043],
  Cell[893966, 24542, 1312, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1578524611", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "19"},
   CellID->1578524611],
  Cell[895281, 24586, 1499, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1738175589", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "19'"},
   CellID->1738175589],
  Cell[922844, 25335, 1110, 34, 70, "GlobalDeclaration",
   CellTags->{
    "ID:258339355", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->258339355],
  Cell[923957, 25371, 832, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1933985410", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1933985410],
  Cell[924792, 25395, 818, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739393974", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "17"},
   CellID->739393974],
  Cell[925613, 25420, 1164, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:875264505", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "10"},
   CellID->875264505],
  Cell[926780, 25457, 1080, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2012915447", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "11"},
   CellID->2012915447],
  Cell[927863, 25492, 1248, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1489336269", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "12"},
   CellID->1489336269],
  Cell[929114, 25534, 1483, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1722317392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "13"},
   CellID->1722317392],
  Cell[930600, 25582, 1075, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1089845252", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "15"},
   CellID->1089845252],
  Cell[969571, 26618, 1054, 31, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2021433946", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2021433946],
  Cell[970628, 26651, 737, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:369680194", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->369680194],
  Cell[971368, 26673, 2292, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:491925495", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] alternative"},
   CellID->491925495],
  Cell[973663, 26744, 1451, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1668317482", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred \[LeftRightArrow]"},
   CellID->1668317482],
  Cell[992533, 27228, 887, 27, 70, "GlobalDeclaration",
   CellTags->{
    "ID:169581633", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->169581633],
  Cell[993423, 27257, 2362, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:265555623", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "403"},
   CellID->265555623],
  Cell[1006154, 27588, 1007, 30, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1020054522", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1020054522],
  Cell[1007164, 27620, 776, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:270873424", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->270873424],
  Cell[1007943, 27643, 998, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1165091979", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "16"},
   CellID->1165091979],
  Cell[1008944, 27673, 1144, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1894133408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred symmetry"},
   CellID->1894133408],
  Cell[1027970, 28189, 653, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1779789522", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1779789522],
  Cell[1028626, 28207, 1660, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:979757495", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "algorithmic reduction ring"},
   CellID->979757495],
  Cell[1031040, 28284, 5504, 148, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502179371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R9"},
   CellID->1502179371],
  Cell[1036547, 28434, 4444, 124, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1524689422", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R10"},
   CellID->1524689422],
  Cell[1040994, 28560, 2566, 75, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:528478842", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R11"},
   CellID->528478842],
  Cell[1043563, 28637, 4242, 109, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:415149625", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R12"},
   CellID->415149625],
  Cell[1047808, 28748, 1890, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1276925313", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R13"},
   CellID->1276925313],
  Cell[1050128, 28822, 697, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1684964128", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1684964128],
  Cell[1051428, 28859, 2100, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:10673911", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "22"},
   CellID->10673911],
  Cell[1053531, 28924, 1562, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:603265816", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "23"},
   CellID->603265816],
  Cell[1055096, 28974, 3444, 100, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:39612209", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "24"},
   CellID->39612209],
  Cell[1088017, 29805, 680, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1515571182", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1515571182],
  Cell[1088700, 29824, 1544, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:528027344", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "18"},
   CellID->528027344],
  Cell[1090247, 29877, 2634, 78, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:94925229", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "25"},
   CellID->94925229],
  Cell[1092884, 29957, 1801, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2104251234", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "18'"},
   CellID->2104251234],
  Cell[1117556, 30602, 878, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:484595803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isFinDomain \[DoubleRightArrow] R13"},
   CellID->484595803],
  Cell[1118437, 30626, 672, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:55361704", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->55361704],
  Cell[1119112, 30645, 1712, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:283927726", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "20"},
   CellID->283927726],
  Cell[1121268, 30711, 1043, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1584214935", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isNoetherian sets"},
   CellID->1584214935],
  Cell[1122314, 30743, 1053, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1231517070", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isNoetherian tuples"},
   CellID->1231517070],
  Cell[1140591, 31272, 740, 18, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1326788334", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1326788334],
  Cell[1142367, 31328, 843, 22, 44, "GlobalDeclaration",
   CellTags->{
    "ID:227700447", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->227700447],
  Cell[1143213, 31352, 828, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1128391937", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1128391937],
  Cell[1144044, 31376, 1041, 33, 33, "GlobalDeclaration",
   CellTags->{
    "ID:788600821", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->788600821],
  Cell[1145088, 31411, 845, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:999322294", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->999322294],
  Cell[1145936, 31435, 790, 22, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:737948621", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "392"},
   CellID->737948621],
  Cell[1146729, 31459, 1317, 37, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:287016951", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "394"},
   CellID->287016951],
  Cell[1148049, 31498, 1343, 39, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:1812129266", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "395"},
   CellID->1812129266],
  Cell[1149395, 31539, 3200, 90, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:1191851294", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "396"},
   CellID->1191851294],
  Cell[1152598, 31631, 2723, 79, 117, "FormalTextInputFormula",
   CellTags->{
    "ID:101106535", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "397"},
   CellID->101106535],
  Cell[1155324, 31712, 2532, 72, 61, "FormalTextInputFormula",
   CellTags->{
    "ID:1737800130", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "398"},
   CellID->1737800130],
  Cell[1157859, 31786, 2281, 67, 60, "FormalTextInputFormula",
   CellTags->{
    "ID:1568772314", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "399"},
   CellID->1568772314],
  Cell[1237354, 33708, 870, 22, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1597110548", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1597110548],
  Cell[1238227, 33732, 840, 21, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1509853174", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1509853174],
  Cell[1239070, 33755, 1700, 43, 56, "GlobalDeclaration",
   CellTags->{
    "ID:1554200522", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1554200522],
  Cell[1240773, 33800, 1344, 39, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:1963951655", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "400"},
   CellID->1963951655],
  Cell[1242120, 33841, 1002, 30, 40, "GlobalDeclaration",
   CellTags->{
    "ID:85295514", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->85295514],
  Cell[1243125, 33873, 2157, 63, 84, "FormalTextInputFormula",
   CellTags->{
    "ID:751169803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "401"},
   CellID->751169803],
  Cell[1245285, 33938, 2025, 61, 60, "FormalTextInputFormula",
   CellTags->{
    "ID:46875747", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "402'"},
   CellID->46875747],
  Cell[1247313, 34001, 2201, 64, 60, "FormalTextInputFormula",
   CellTags->{
    "ID:656927872", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "402"},
   CellID->656927872],
  Cell[1314406, 35584, 794, 21, 44, "GlobalDeclaration",
   CellTags->{
    "ID:502021608", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->502021608],
  Cell[1315203, 35607, 924, 23, 41, "GlobalDeclaration",
   CellTags->{
    "ID:227587436", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->227587436],
  Cell[1316130, 35632, 1206, 31, 34, "GlobalDeclaration",
   CellTags->{
    "ID:937292216", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->937292216],
  Cell[1317339, 35665, 1553, 45, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1763687895", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "410"},
   CellID->1763687895],
  Cell[1318895, 35712, 1164, 32, 56, "GlobalDeclaration",
   CellTags->{
    "ID:397073436", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->397073436],
  Cell[1320062, 35746, 1593, 45, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:791651685", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "411"},
   CellID->791651685],
  Cell[1321658, 35793, 1130, 38, 44, "GlobalDeclaration",
   CellTags->{
    "ID:6804200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->6804200],
  Cell[1322791, 35833, 1069, 31, 58, "FormalTextInputFormula",
   CellTags->{
    "ID:1294525434", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "414"},
   CellID->1294525434],
  Cell[1323863, 35866, 1005, 30, 40, "GlobalDeclaration",
   CellTags->{
    "ID:372638659", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->372638659],
  Cell[1324871, 35898, 1824, 54, 71, "FormalTextInputFormula",
   CellTags->{
    "ID:1962716191", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "412"},
   CellID->1962716191],
  Cell[1326698, 35954, 2316, 71, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:1575262565", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "413"},
   CellID->1575262565],
  Cell[1446016, 38899, 1155, 35, 40, "GlobalDeclaration",
   CellTags->{
    "ID:1547867165", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1547867165],
  Cell[1447174, 38936, 1005, 24, 41, "GlobalDeclaration",
   CellTags->{
    "ID:2009958353", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2009958353],
  Cell[1448182, 38962, 1405, 43, 43, "GlobalDeclaration",
   CellTags->{
    "ID:10165835", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->10165835],
  Cell[1449590, 39007, 1564, 43, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:1879224373", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6Witness sufficient"},
   CellID->1879224373],
  Cell[1451157, 39052, 1847, 55, 112, "FormalTextInputFormula",
   CellTags->{
    "ID:381733398", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6Witness symmetry"},
   CellID->381733398],
  Cell[1492278, 40124, 692, 17, 33, "GlobalDeclaration",
   CellTags->{
    "ID:404204643", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->404204643],
  Cell[1493853, 40171, 1019, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1399661856", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1399661856],
  Cell[1494875, 40199, 1816, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1116556350", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde] multiplier symmetric"},
   CellID->1116556350],
  Cell[1496694, 40260, 1335, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1121162524", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "mult \[Tilde]"},
   CellID->1121162524],
  Cell[1498032, 40304, 1411, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1098905803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible \[Tilde]"},
   CellID->1098905803],
  Cell[1499446, 40348, 1862, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:88395859", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areIrrelative \[Tilde]"},
   CellID->88395859],
  Cell[1501311, 40403, 2056, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1322820809", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers \[Tilde]"},
   CellID->1322820809],
  Cell[1503370, 40460, 753, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:931721333", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->931721333],
  Cell[1504126, 40482, 1741, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2090837112", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] multiplier \[Tilde]"},
   CellID->2090837112],
  Cell[1505870, 40540, 1681, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:780050836", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[Tilde]"},
   CellID->780050836],
  Cell[1507554, 40596, 761, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:128058836", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->128058836],
  Cell[1508318, 40617, 1379, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1516437288", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] reductor \[Tilde]"},
   CellID->1516437288],
  Cell[1509700, 40661, 1334, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1864007722", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] reductor \[Tilde]"},
   CellID->1864007722],
  Cell[1511037, 40704, 1850, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:642398934", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below \[Tilde]"},
   CellID->642398934],
  Cell[1596141, 42870, 765, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1200903488", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1200903488],
  Cell[1596909, 42892, 810, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:353808388", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[Tilde]"},
   CellID->353808388],
  Cell[1597722, 42915, 1079, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:274091018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of \[Tilde]"},
   CellID->274091018],
  Cell[1598804, 42951, 1391, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1760571189", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[Tilde]"},
   CellID->1760571189],
  Cell[1613881, 43398, 973, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:871669950", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->871669950],
  Cell[1614857, 43425, 1720, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:163646289", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative-irrelative 1"},
   CellID->163646289],
  Cell[1626552, 43734, 813, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1990325407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1990325407],
  Cell[1627368, 43757, 765, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:615171230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->615171230],
  Cell[1628136, 43779, 1506, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1548501955", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "47"},
   CellID->1548501955],
  Cell[1629645, 43828, 1423, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:170197122", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "48"},
   CellID->170197122],
  Cell[1631071, 43876, 1947, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:772910972", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "49"},
   CellID->772910972],
  Cell[1652515, 44472, 1026, 27, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1034579573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1034579573],
  Cell[1653544, 44501, 3958, 114, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:226093298", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "30.\[DoubleRightArrow]"},
   CellID->226093298],
  Cell[1657505, 44617, 3927, 117, 156, "FormalTextInputFormula",
   CellTags->{
    "ID:1773205884", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "31.\[DoubleRightArrow]"},
   CellID->1773205884],
  Cell[1661435, 44736, 3022, 91, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:1783130815", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "30.\[DoubleLeftArrow]"},
   CellID->1783130815],
  Cell[1664460, 44829, 2860, 88, 117, "FormalTextInputFormula",
   CellTags->{
    "ID:1493153487", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "31.\[DoubleLeftArrow]"},
   CellID->1493153487],
  Cell[1667323, 44919, 1979, 67, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1798779961", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "39.\[DoubleRightArrow]"},
   CellID->1798779961],
  Cell[1731959, 46627, 1024, 25, 41, "GlobalDeclaration",
   CellTags->{
    "ID:712613448", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->712613448],
  Cell[1732986, 46654, 3002, 91, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:1914773542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "37.\[DoubleRightArrow]"},
   CellID->1914773542],
  Cell[1735991, 46747, 3026, 92, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:2048103084", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "37.\[DoubleLeftArrow]"},
   CellID->2048103084],
  Cell[1758672, 47333, 1003, 26, 42, "GlobalDeclaration",
   CellTags->{
    "ID:254692456", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->254692456],
  Cell[1759678, 47361, 3946, 122, 158, "FormalTextInputFormula",
   CellTags->{
    "ID:336730845", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "32.\[DoubleRightArrow]"},
   CellID->336730845],
  Cell[1763627, 47485, 3973, 123, 158, "FormalTextInputFormula",
   CellTags->{
    "ID:1715399148", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "32.\[DoubleLeftArrow]"},
   CellID->1715399148],
  Cell[1793910, 48253, 885, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1253203360", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1253203360],
  Cell[1794798, 48280, 860, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1514263990", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[RightArrow]\[RawStar]"},
   CellID->1514263990],
  Cell[1795661, 48305, 1506, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:411529944", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[RightArrow]\[RawStar]"},
   CellID->411529944],
  Cell[1808290, 48664, 885, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1862973197", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1862973197],
  Cell[1809178, 48691, 1202, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:237268641", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "\[RightArrow] \[SubsetEqual] \[RightArrow]\[RawStar]"},
   CellID->237268641],
  Cell[1810383, 48731, 1492, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1746879589", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "38"},
   CellID->1746879589],
  Cell[1824602, 49136, 821, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1956094048", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1956094048],
  Cell[1825426, 49161, 915, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:560057636", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of connectibility"},
   CellID->560057636],
  Cell[1826344, 49187, 1325, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:354829984", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of connectibility"},
   CellID->354829984],
  Cell[1827672, 49227, 1587, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:350848859", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of connectibility"},
   CellID->350848859],
  Cell[1845413, 49718, 1091, 27, 70, "GlobalDeclaration",
   CellTags->{
    "ID:705004664", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->705004664],
  Cell[1846507, 49747, 1392, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:411199728", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of connectibility below"},
   CellID->411199728],
  Cell[1847902, 49793, 1730, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502850808", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of connectibility below"},
   CellID->1502850808],
  Cell[1849635, 49851, 2301, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1331049102", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of connectibility below"},
   CellID->1331049102],
  Cell[1872203, 50463, 3364, 112, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1147958002", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "36"},
   CellID->1147958002],
  Cell[1875570, 50577, 3848, 121, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1663785837", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "33"},
   CellID->1663785837],
  Cell[1899766, 51216, 802, 21, 41, "GlobalDeclaration",
   CellTags->{
    "ID:311737773", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->311737773],
  Cell[1900571, 51239, 1149, 34, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:552003293", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "reflexivity of reflexive-transitive closure of \[RightArrow]"},
   CellID->552003293],
  Cell[1901723, 51275, 767, 20, 44, "GlobalDeclaration",
   CellTags->{
    "ID:129456084", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->129456084],
  Cell[1902493, 51297, 1868, 55, 45, "FormalTextInputFormula",
   CellTags->{
    "ID:441235278", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "34"},
   CellID->441235278],
  Cell[1904364, 51354, 3714, 110, 99, "FormalTextInputFormula",
   CellTags->{
    "ID:1516348781", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "35"},
   CellID->1516348781],
  Cell[1941928, 52368, 754, 20, 41, "GlobalDeclaration",
   CellTags->{
    "ID:1329557757", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1329557757],
  Cell[1942685, 52390, 2221, 74, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1168535485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "40"},
   CellID->1168535485],
  Cell[1944909, 52466, 2317, 77, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:2128489125", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "41"},
   CellID->2128489125],
  Cell[1962757, 52954, 751, 20, 41, "GlobalDeclaration",
   CellTags->{
    "ID:548213617", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->548213617],
  Cell[1963511, 52976, 1974, 63, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:909891749", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "42"},
   CellID->909891749],
  Cell[1965488, 53041, 936, 27, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:830274722", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "44"},
   CellID->830274722],
  Cell[1966427, 53070, 760, 20, 30, "GlobalDeclaration",
   CellTags->{
    "ID:1390954175", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1390954175],
  Cell[1967190, 53092, 1126, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:2016336231", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68"},
   CellID->2016336231],
  Cell[1968319, 53126, 1449, 45, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1052198272", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68.set"},
   CellID->1052198272],
  Cell[1969771, 53173, 1749, 59, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:35312349", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68.tuple"},
   CellID->35312349],
  Cell[2003857, 54095, 870, 24, 42, "GlobalDeclaration",
   CellTags->{
    "ID:2032809580", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2032809580],
  Cell[2004730, 54121, 711, 19, 32, "GlobalDeclaration",
   CellTags->{
    "ID:1947090769", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1947090769],
  Cell[2005444, 54142, 3280, 101, 102, "FormalTextInputFormula",
   CellTags->{
    "ID:693923465", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "50"},
   CellID->693923465],
  Cell[2008727, 54245, 1394, 46, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1234215913", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "51"},
   CellID->1234215913],
  Cell[2010124, 54293, 1758, 61, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1065734937", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "52"},
   CellID->1065734937],
  Cell[2011885, 54356, 2270, 76, 56, "FormalTextInputFormula",
   CellTags->{
    "ID:1763925306", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "53"},
   CellID->1763925306],
  Cell[2048206, 55290, 839, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:332034572", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->332034572],
  Cell[2049048, 55314, 1006, 30, 40, "GlobalDeclaration",
   CellTags->{
    "ID:1609461850", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1609461850],
  Cell[2050057, 55346, 1801, 63, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:621597434", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "54"},
   CellID->621597434],
  Cell[2051861, 55411, 1814, 63, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:2031891226", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "55"},
   CellID->2031891226],
  Cell[2053678, 55476, 770, 20, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1596753391", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1596753391],
  Cell[2054451, 55498, 1661, 59, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1855698114", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "55'"},
   CellID->1855698114],
  Cell[2083630, 56255, 779, 20, 41, "GlobalDeclaration",
   CellTags->{
    "ID:1116860460", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1116860460],
  Cell[2084412, 56277, 2190, 70, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:915980145", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "56"},
   CellID->915980145],
  Cell[2086605, 56349, 768, 20, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1683552886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1683552886],
  Cell[2087376, 56371, 713, 18, 31, "GlobalDeclaration",
   CellTags->{
    "ID:1331309258", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1331309258],
  Cell[2088092, 56391, 784, 21, 41, "GlobalDeclaration",
   CellTags->{
    "ID:551817557", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->551817557],
  Cell[2088879, 56414, 1214, 40, 43, "GlobalDeclaration",
   CellTags->{
    "ID:1724920630", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1724920630],
  Cell[2090096, 56456, 1255, 40, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:88655200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "57"},
   CellID->88655200],
  Cell[2091354, 56498, 1544, 50, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1149353049", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "58"},
   CellID->1149353049],
  Cell[2126782, 57478, 792, 21, 42, "GlobalDeclaration",
   CellTags->{
    "ID:895068823", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->895068823],
  Cell[2127577, 57501, 2607, 90, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:296045170", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "45"},
   CellID->296045170],
  Cell[2130187, 57593, 2729, 94, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:1887054887", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "46"},
   CellID->1887054887],
  Cell[2152755, 58183, 1753, 53, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1876022438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "59"},
   CellID->1876022438],
  Cell[2160309, 58403, 779, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1852246688", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1852246688],
  Cell[2161618, 58441, 2591, 86, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1471949021", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "element confluence"},
   CellID->1471949021],
  Cell[2164212, 58529, 2513, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:146706347", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "confluence"},
   CellID->146706347],
  Cell[2166728, 58601, 2403, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1279966325", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Church-Rosser property"},
   CellID->1279966325],
  Cell[2169134, 58671, 2770, 90, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1602877499", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "local connectivity"},
   CellID->1602877499],
  Cell[2171907, 58763, 1504, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:661413373", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Gr\[ODoubleDot]bner basis"},
   CellID->661413373],
  Cell[2174085, 58826, 973, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1444563707", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1444563707],
  Cell[2175061, 58852, 2840, 84, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1408266583", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areConnectibleMultipliers"},
   CellID->1408266583],
  Cell[2177904, 58938, 2867, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:882680545", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of one critical pair"},
   CellID->882680545],
  Cell[2180774, 59019, 3995, 103, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:259042764", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs"},
   CellID->259042764],
  Cell[2184772, 59124, 1451, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1836116785", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs of tuple"},
   CellID->1836116785],
  Cell[2186226, 59169, 2080, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2066466981", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs of elements"},
   CellID->2066466981],
  Cell[2188309, 59231, 2727, 90, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:756678784", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "local connectivity of elements"},
   CellID->756678784],
  Cell[2191580, 59342, 676, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1173678670", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1173678670],
  Cell[2192729, 59377, 1492, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1524905762", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible 1"},
   CellID->1524905762],
  Cell[2201086, 59605, 756, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2054204196", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2054204196],
  Cell[2201845, 59627, 769, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:194336473", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->194336473],
  Cell[2202617, 59649, 1048, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2064859128", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible tuple"},
   CellID->2064859128],
  Cell[2203668, 59680, 1053, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1276073437", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective tuple"},
   CellID->1276073437],
  Cell[2204724, 59711, 983, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1869404165", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective tuple II"},
   CellID->1869404165],
  Cell[2220683, 60157, 770, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2083446865", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2083446865],
  Cell[2221456, 60179, 1470, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:709723813", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative"},
   CellID->709723813],
  Cell[2222929, 60220, 1342, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:41201803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective alternative"},
   CellID->41201803],
  Cell[2224274, 60256, 1459, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1023292110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective sufficient"},
   CellID->1023292110],
  Cell[2243244, 60770, 930, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:529333522", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->529333522],
  Cell[2244177, 60798, 1787, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208412986", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible join"},
   CellID->208412986],
  Cell[2245967, 60855, 1845, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1183553358", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible append"},
   CellID->1183553358],
  Cell[2247815, 60913, 1359, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:286665714", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible join II"},
   CellID->286665714],
  Cell[2249177, 60956, 1398, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1065812089", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible append II"},
   CellID->1065812089],
  Cell[2277119, 61674, 929, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:751627229", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->751627229],
  Cell[2278513, 61718, 832, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:351093647", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->351093647],
  Cell[2279348, 61743, 815, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1476531812", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1476531812],
  Cell[2280166, 61766, 1107, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1047637483", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible \[Tilde] \[DoubleRightArrow]"},
   CellID->1047637483],
  Cell[2281276, 61797, 1217, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1237319003", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible \[Tilde]"},
   CellID->1237319003],
  Cell[2302036, 62300, 2236, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1087064485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative I"},
   CellID->1087064485],
  Cell[2304275, 62366, 2578, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1334373255", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative I'"},
   CellID->1334373255],
  Cell[2306856, 62447, 1740, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:581920350", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative II"},
   CellID->581920350],
  Cell[2308599, 62500, 2240, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:286693183", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative II'"},
   CellID->286693183],
  Cell[2349221, 63508, 757, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:173719369", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->173719369],
  Cell[2349981, 63530, 1038, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:762511022", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible symmetric"},
   CellID->762511022],
  Cell[2351022, 63559, 1050, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1792796192", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective symmetric"},
   CellID->1792796192],
  Cell[2364270, 63914, 1389, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:320557773", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible 0"},
   CellID->320557773],
  Cell[2372984, 64154, 818, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1193869577", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1193869577],
  Cell[2373845, 64179, 734, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:342993412", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->342993412],
  Cell[2375148, 64219, 4772, 136, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1786903669", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "63 ind"},
   CellID->1786903669],
  Cell[2379923, 64357, 4400, 129, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1742873082", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "63"},
   CellID->1742873082],
  Cell[2384326, 64488, 1191, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:332228196", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "64"},
   CellID->332228196],
  Cell[2416582, 65275, 1205, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2110225002", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "65"},
   CellID->2110225002],
  Cell[2423235, 65468, 962, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:629188208", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "66"},
   CellID->629188208],
  Cell[2424200, 65497, 990, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1222366601", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "67"},
   CellID->1222366601],
  Cell[2433562, 65768, 818, 20, 33, "GlobalDeclaration",
   CellTags->{
    "ID:2131577443", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2131577443],
  Cell[2434541, 65796, 835, 21, 44, "GlobalDeclaration",
   CellTags->{
    "ID:2046259788", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2046259788],
  Cell[2435790, 65835, 1009, 25, 42, "GlobalDeclaration",
   CellTags->{
    "ID:205712485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->205712485],
  Cell[2436802, 65862, 2519, 75, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:2050580243", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "61"},
   CellID->2050580243],
  Cell[2439324, 65939, 5765, 175, 146, "FormalTextInputFormula",
   CellTags->{
    "ID:2100291458", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "60"},
   CellID->2100291458],
  Cell[2445092, 66116, 687, 17, 30, "GlobalDeclaration",
   CellTags->{
    "ID:1810831635", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1810831635],
  Cell[2445782, 66135, 847, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:718268018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->718268018],
  Cell[2446632, 66159, 4333, 133, 248, "FormalTextInputFormula",
   CellTags->{
    "ID:1959449230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "62"},
   CellID->1959449230],
  Cell[2513610, 67871, 4006, 99, 164, "FormalTextInputFormula",
   CellTags->{
    "ID:1029764736", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "70"},
   CellID->1029764736],
  Cell[2517619, 67972, 3954, 113, 189, "FormalTextInputFormula",
   CellTags->{
    "ID:1120053244", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "71"},
   CellID->1120053244],
  Cell[2589609, 69938, 3750, 111, 93, "FormalTextInputFormula",
   CellTags->{
    "ID:25850532", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "72"},
   CellID->25850532],
  Cell[2633196, 71224, 1361, 38, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:482796083", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "74"},
   CellID->482796083],
  Cell[2634560, 71264, 1742, 49, 82, "FormalTextInputFormula",
   CellTags->{
    "ID:1705968105", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "75"},
   CellID->1705968105],
  Cell[2679069, 72583, 1341, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:68280999", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "73"},
   CellID->68280999],
  Cell[2685048, 72752, 1257, 33, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1188710397", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Main Theorem"},
   CellID->1188710397],
  Cell[2690656, 72914, 780, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:78242703", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->78242703],
  Cell[2691853, 72952, 1918, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:807712598", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "75 tuples"},
   CellID->807712598],
  Cell[2693774, 73010, 3950, 106, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:645630189", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "76 tuples"},
   CellID->645630189],
  Cell[2697727, 73118, 1955, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47992769", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "77 tuples"},
   CellID->47992769],
  Cell[2734185, 74013, 741, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:41552115", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->41552115],
  Cell[2734929, 74032, 781, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:208902120", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->208902120],
  Cell[2736136, 74068, 3835, 112, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:531785290", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal congruence"},
   CellID->531785290],
  Cell[2740372, 74196, 1341, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1814440310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal"},
   CellID->1814440310],
  Cell[2741716, 74234, 1240, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1625946592", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal of tuple"},
   CellID->1625946592],
  Cell[2744193, 74308, 714, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1782924340", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1782924340],
  Cell[2744949, 74331, 783, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1482545820", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1482545820],
  Cell[2746430, 74376, 4190, 125, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1796414707", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "sum multiplication join"},
   CellID->1796414707],
  Cell[2764330, 74838, 782, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1254924191", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1254924191],
  Cell[2765115, 74860, 864, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:923334447", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[Congruent]"},
   CellID->923334447],
  Cell[2765982, 74885, 1182, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1712150230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of \[Congruent]"},
   CellID->1712150230],
  Cell[2767167, 74923, 1542, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1473809719", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[Congruent]"},
   CellID->1473809719],
  Cell[2785797, 75442, 6867, 189, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:306651916", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] alternative ind.1"},
   CellID->306651916],
  Cell[2792667, 75633, 7978, 214, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:571707933", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] alternative ind.2"},
   CellID->571707933],
  Cell[2839897, 76752, 771, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:473420099", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->473420099],
  Cell[2840671, 76774, 1216, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:544141308", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[SubsetEqual] \[Congruent]"},
   CellID->544141308],
  Cell[2841890, 76813, 1245, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1307557419", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] \[SubsetEqual] \[Congruent]"},
   CellID->1307557419],
  Cell[2843138, 76853, 2435, 75, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:170897044", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility \[SubsetEqual] \[Congruent] ind"},
   CellID->170897044],
  Cell[2845576, 76930, 1245, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:720101359", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility \[SubsetEqual] \[Congruent]"},
   CellID->720101359],
  Cell[2871888, 77639, 2147, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1680644099", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "43"},
   CellID->1680644099],
  Cell[2874038, 77707, 4727, 134, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:555151129", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] \[SubsetEqual] connectibility ind"},
   CellID->555151129],
  Cell[2878768, 77843, 1545, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2104087526", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] \[SubsetEqual] connectibility"},
   CellID->2104087526],
  Cell[2907063, 78561, 1520, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:980108134", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent]=connectibility"},
   CellID->980108134],
  Cell[2914888, 78785, 770, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2002188699", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2002188699],
  Cell[2915661, 78807, 3086, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160768319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0", "IR"},
   CellID->160768319],
  Cell[2918750, 78900, 962, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:562483616", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] \[Congruent] 0"},
   CellID->562483616],
  Cell[2919715, 78927, 1145, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:853458295", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 \[RawStar]"},
   CellID->853458295],
  Cell[2920863, 78960, 1063, 30, 70, "GlobalDeclaration",
   CellTags->{
    "ID:462880889", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->462880889],
  Cell[2921929, 78992, 989, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:44331059", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 +"},
   CellID->44331059],
  Cell[2922921, 79021, 969, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1958399498", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 -"},
   CellID->1958399498],
  Cell[2958884, 79958, 912, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2069217266", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2069217266],
  Cell[2959799, 79984, 6348, 176, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:277446168", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] extended ind"},
   CellID->277446168],
  Cell[2966150, 80162, 1608, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1459665988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] extended"},
   CellID->1459665988],
  Cell[2993233, 80822, 716, 18, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1670835483", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1670835483],
  Cell[2994165, 80849, 780, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:97617962", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->97617962],
  Cell[2994988, 80874, 753, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1452952621", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1452952621],
  Cell[2996206, 80912, 976, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1760501242", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets ideal", "IR"},
   CellID->1760501242],
  Cell[2997185, 80939, 871, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:554864723", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal isSet", "IR"},
   CellID->554864723],
  Cell[2998059, 80963, 1496, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739591353", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] ideal"},
   CellID->739591353],
  Cell[2999558, 81012, 786, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2030275451", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2030275451],
  Cell[3000347, 81033, 847, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1644449200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideals closed +", "IR"},
   CellID->1644449200],
  Cell[3001197, 81058, 861, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1500033871", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideals closed - binary", "IR"},
   CellID->1500033871],
  Cell[3002061, 81083, 1026, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1114678876", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal \[Union] singleton"},
   CellID->1114678876],
  Cell[3003090, 81112, 1822, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1895756686", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal \[RightArrow]\[RawStar]"},
   CellID->1895756686],
  Cell[3040491, 82126, 839, 21, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1581989517", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1581989517],
  Cell[3041371, 82152, 801, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:185976047", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->185976047],
  Cell[3042628, 82191, 1600, 48, 45, "FormalTextInputFormula",
   CellTags->{
    "ID:1918127619", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] tuple-ideal"},
   CellID->1918127619],
  Cell[3044231, 82241, 1009, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:279213832", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets tuple-ideal", "IR"},
   CellID->279213832],
  Cell[3045243, 82268, 814, 20, 31, "GlobalDeclaration",
   CellTags->{
    "ID:1225983506", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1225983506],
  Cell[3046060, 82290, 885, 24, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1587050664", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideals closed +", "IR"},
   CellID->1587050664],
  Cell[3046948, 82316, 896, 24, 29, "FormalTextInputFormula",
   CellTags->{
    "ID:508704301", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideals closed - binary", "IR"},
   CellID->508704301],
  Cell[3047847, 82342, 1085, 29, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1181234284", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideal append"},
   CellID->1181234284],
  Cell[3048935, 82373, 1904, 59, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:352062174", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideal \[RightArrow]\[RawStar]"},
   CellID->352062174]},
 "ID:457420627"->{
  Cell[96956, 2477, 1338, 27, 33, "GlobalDeclaration",
   CellTags->{
    "ID:457420627", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->457420627]},
 "ID:176465234"->{
  Cell[98297, 2506, 1000, 23, 31, "GlobalDeclaration",
   CellTags->{
    "ID:176465234", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->176465234]},
 "ID:478302118"->{
  Cell[99300, 2531, 1892, 54, 58, "FormalTextInputFormula",
   CellTags->{
    "ID:478302118", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction with several multipliers"},
   CellID->478302118]},
 "reduction with several multipliers"->{
  Cell[99300, 2531, 1892, 54, 58, "FormalTextInputFormula",
   CellTags->{
    "ID:478302118", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction with several multipliers"},
   CellID->478302118]},
 "ID:448370877"->{
  Cell[101715, 2607, 2024, 58, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:448370877", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetric closure of reducibility"},
   CellID->448370877]},
 "symmetric closure of reducibility"->{
  Cell[101715, 2607, 2024, 58, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:448370877", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetric closure of reducibility"},
   CellID->448370877]},
 "ID:1537297721"->{
  Cell[103742, 2667, 1383, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1537297721", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reducible1"},
   CellID->1537297721]},
 "reducible1"->{
  Cell[103742, 2667, 1383, 39, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1537297721", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reducible1"},
   CellID->1537297721]},
 "ID:115218056"->{
  Cell[105128, 2708, 1599, 47, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:115218056", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reducible2"},
   CellID->115218056]},
 "reducible2"->{
  Cell[105128, 2708, 1599, 47, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:115218056", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reducible2"},
   CellID->115218056]},
 "ID:399781455"->{
  Cell[106730, 2757, 1894, 50, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:399781455", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "irreducibility"},
   CellID->399781455]},
 "irreducibility"->{
  Cell[106730, 2757, 1894, 50, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:399781455", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "irreducibility"},
   CellID->399781455]},
 "ID:1867242863"->{
  Cell[108627, 2809, 2205, 58, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1867242863", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "irrelative"},
   CellID->1867242863]},
 "irrelative"->{
  Cell[108627, 2809, 2205, 58, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1867242863", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "irrelative"},
   CellID->1867242863]},
 "ID:1834165766"->{
  Cell[110835, 2869, 1640, 52, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1834165766", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative"},
   CellID->1834165766]},
 "correlative"->{
  Cell[110835, 2869, 1640, 52, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1834165766", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative"},
   CellID->1834165766]},
 "ID:1907029362"->{
  Cell[112478, 2923, 2270, 63, 82, "FormalTextInputFormula",
   CellTags->{
    "ID:1907029362", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common reducible"},
   CellID->1907029362]},
 "common reducible"->{
  Cell[112478, 2923, 2270, 63, 82, "FormalTextInputFormula",
   CellTags->{
    "ID:1907029362", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common reducible"},
   CellID->1907029362]},
 "ID:2013437879"->{
  Cell[114751, 2988, 3534, 94, 203, "FormalTextInputFormula",
   CellTags->{
    "ID:2013437879", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "non-trivial common reducible"},
   CellID->2013437879]},
 "non-trivial common reducible"->{
  Cell[114751, 2988, 3534, 94, 203, "FormalTextInputFormula",
   CellTags->{
    "ID:2013437879", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "non-trivial common reducible"},
   CellID->2013437879]},
 "ID:991122907"->{
  Cell[118288, 3084, 2119, 60, 83, "FormalTextInputFormula",
   CellTags->{
    "ID:991122907", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "minimal non-trivial common reducible"},
   CellID->991122907]},
 "minimal non-trivial common reducible"->{
  Cell[118288, 3084, 2119, 60, 83, "FormalTextInputFormula",
   CellTags->{
    "ID:991122907", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "minimal non-trivial common reducible"},
   CellID->991122907]},
 "ID:1293246145"->{
  Cell[120410, 3146, 2361, 72, 125, "FormalTextInputFormula",
   CellTags->{
    "ID:1293246145", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "unary non-trivial common reducible"},
   CellID->1293246145]},
 "unary non-trivial common reducible"->{
  Cell[120410, 3146, 2361, 72, 125, "FormalTextInputFormula",
   CellTags->{
    "ID:1293246145", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "unary non-trivial common reducible"},
   CellID->1293246145]},
 "ID:894750987"->{
  Cell[122774, 3220, 1982, 59, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:894750987", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "unary minimal non-trivial common reducible"},
   CellID->894750987]},
 "unary minimal non-trivial common reducible"->{
  Cell[122774, 3220, 1982, 59, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:894750987", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "unary minimal non-trivial common reducible"},
   CellID->894750987]},
 "ID:779388596"->{
  Cell[124759, 3281, 2543, 66, 109, "FormalTextInputFormula",
   CellTags->{
    "ID:779388596", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "critical pair multipliers"},
   CellID->779388596]},
 "critical pair multipliers"->{
  Cell[124759, 3281, 2543, 66, 109, "FormalTextInputFormula",
   CellTags->{
    "ID:779388596", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "critical pair multipliers"},
   CellID->779388596]},
 "ID:1517092650"->{
  Cell[128069, 3376, 1897, 34, 31, "GlobalDeclaration",
   CellTags->{
    "ID:1517092650", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1517092650]},
 "ID:935236171"->{
  Cell[129969, 3412, 2121, 68, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:935236171", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo element using multiplier"},
   CellID->935236171]},
 "reduction modulo element using multiplier"->{
  Cell[129969, 3412, 2121, 68, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:935236171", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo element using multiplier"},
   CellID->935236171]},
 "ID:320624715"->{
  Cell[132093, 3482, 2134, 69, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:320624715", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo set"},
   CellID->320624715]},
 "reduction modulo set"->{
  Cell[132093, 3482, 2134, 69, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:320624715", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo set"},
   CellID->320624715]},
 "ID:434515171"->{
  Cell[134230, 3553, 3140, 84, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:434515171", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo tuple"},
   CellID->434515171]},
 "reduction modulo tuple"->{
  Cell[134230, 3553, 3140, 84, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:434515171", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo tuple"},
   CellID->434515171]},
 "ID:773828270"->{
  Cell[137373, 3639, 1426, 43, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:773828270", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReductor"},
   CellID->773828270]},
 "isReductor"->{
  Cell[137373, 3639, 1426, 43, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:773828270", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReductor"},
   CellID->773828270]},
 "ID:1813157585"->{
  Cell[139428, 3700, 2877, 88, 97, "FormalTextInputFormula",
   CellTags->{
    "ID:1813157585", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "reflexive-transitive closure of reduction relation"},
   CellID->1813157585]},
 "reflexive-transitive closure of reduction relation"->{
  Cell[139428, 3700, 2877, 88, 97, "FormalTextInputFormula",
   CellTags->{
    "ID:1813157585", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "reflexive-transitive closure of reduction relation"},
   CellID->1813157585]},
 "ID:370202704"->{
  Cell[142308, 3790, 2110, 68, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:370202704", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetric closure of reduction relation"},
   CellID->370202704]},
 "symmetric closure of reduction relation"->{
  Cell[142308, 3790, 2110, 68, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:370202704", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetric closure of reduction relation"},
   CellID->370202704]},
 "ID:992563737"->{
  Cell[144421, 3860, 2820, 86, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:992563737", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "symmetric-reflexive-transitive closure of reduction relation"},
   CellID->992563737]},
 "symmetric-reflexive-transitive closure of reduction relation"->{
  Cell[144421, 3860, 2820, 86, 92, "FormalTextInputFormula",
   CellTags->{
    "ID:992563737", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "symmetric-reflexive-transitive closure of reduction relation"},
   CellID->992563737]},
 "ID:414012405"->{
  Cell[147244, 3948, 2467, 79, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:414012405", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common successor w.r.t. reduction relation"},
   CellID->414012405]},
 "common successor w.r.t. reduction relation"->{
  Cell[147244, 3948, 2467, 79, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:414012405", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common successor w.r.t. reduction relation"},
   CellID->414012405]},
 "ID:895393520"->{
  Cell[149714, 4029, 2406, 79, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:895393520", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexive-transitive closure of common successor"},
   CellID->895393520]},
 "reflexive-transitive closure of common successor"->{
  Cell[149714, 4029, 2406, 79, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:895393520", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexive-transitive closure of common successor"},
   CellID->895393520]},
 "ID:586157110"->{
  Cell[152123, 4110, 2426, 79, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:586157110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common predecessor w.r.t. reduction relation"},
   CellID->586157110]},
 "common predecessor w.r.t. reduction relation"->{
  Cell[152123, 4110, 2426, 79, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:586157110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common predecessor w.r.t. reduction relation"},
   CellID->586157110]},
 "ID:2045822907"->{
  Cell[154552, 4191, 2239, 74, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:2045822907", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "reflexive-transitive closure of common predecessor"},
   CellID->2045822907]},
 "reflexive-transitive closure of common predecessor"->{
  Cell[154552, 4191, 2239, 74, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:2045822907", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "reflexive-transitive closure of common predecessor"},
   CellID->2045822907]},
 "ID:463604795"->{
  Cell[156794, 4267, 3941, 110, 96, "FormalTextInputFormula",
   CellTags->{
    "ID:463604795", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below"},
   CellID->463604795]},
 "connectibility below"->{
  Cell[156794, 4267, 3941, 110, 96, "FormalTextInputFormula",
   CellTags->{
    "ID:463604795", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below"},
   CellID->463604795]},
 "ID:1818365967"->{
  Cell[160738, 4379, 6169, 173, 240, "FormalTextInputFormula",
   CellTags->{
    "ID:1818365967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde] multiplier"},
   CellID->1818365967]},
 "\[Tilde] multiplier"->{
  Cell[160738, 4379, 6169, 173, 240, "FormalTextInputFormula",
   CellTags->{
    "ID:1818365967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde] multiplier"},
   CellID->1818365967]},
 "ID:1533920470"->{
  Cell[166910, 4554, 2226, 67, 58, "FormalTextInputFormula",
   CellTags->{
    "ID:1533920470", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde]"},
   CellID->1533920470]},
 "\[Tilde]"->{
  Cell[166910, 4554, 2226, 67, 58, "FormalTextInputFormula",
   CellTags->{
    "ID:1533920470", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde]"},
   CellID->1533920470]},
 "ID:81135181"->{
  Cell[171113, 4685, 789, 21, 44, "GlobalDeclaration",
   CellTags->{
    "ID:81135181", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->81135181]},
 "ID:1593192085"->{
  Cell[171905, 4708, 1143, 34, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1593192085", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red"},
   CellID->1593192085]},
 "red"->{
  Cell[171905, 4708, 1143, 34, 49, "FormalTextInputFormula",
   CellTags->{
    "ID:1593192085", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red"},
   CellID->1593192085]},
 "ID:808948861"->{
  Cell[173051, 4744, 1434, 45, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:808948861", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset]"},
   CellID->808948861]},
 "\[SquareSubset]"->{
  Cell[173051, 4744, 1434, 45, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:808948861", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset]"},
   CellID->808948861]},
 "ID:301449743"->{
  Cell[174488, 4791, 770, 21, 44, "GlobalDeclaration",
   CellTags->{
    "ID:301449743", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->301449743]},
 "ID:218271861"->{
  Cell[175261, 4814, 902, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:218271861", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red tuples"},
   CellID->218271861]},
 "red tuples"->{
  Cell[175261, 4814, 902, 25, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:218271861", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red tuples"},
   CellID->218271861]},
 "ID:20184186"->{
  Cell[176166, 4841, 1437, 44, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:20184186", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] tuples"},
   CellID->20184186]},
 "\[SquareSubset] tuples"->{
  Cell[176166, 4841, 1437, 44, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:20184186", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] tuples"},
   CellID->20184186]},
 "ID:2086101595"->{
  Cell[178419, 4914, 652, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2086101595", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2086101595]},
 "ID:690041737"->{
  Cell[179880, 4958, 987, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:690041737", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets isReductor"},
   CellID->690041737]},
 "DomainSets isReductor"->{
  Cell[179880, 4958, 987, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:690041737", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets isReductor"},
   CellID->690041737]},
 "ID:5128300"->{
  Cell[180870, 4988, 985, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:5128300", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainTuples isReductor"},
   CellID->5128300]},
 "DomainTuples isReductor"->{
  Cell[180870, 4988, 985, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:5128300", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainTuples isReductor"},
   CellID->5128300]},
 "Proof|ID:690041737"->{
  Cell[182010, 5027, 2733, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:690041737", "Proof|ID:690041737-1"},
   CellID->323627089]},
 "Proof|ID:690041737-1"->{
  Cell[182010, 5027, 2733, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:690041737", "Proof|ID:690041737-1"},
   CellID->323627089]},
 "Proof|ID:5128300"->{
  Cell[185262, 5124, 2731, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:5128300", "Proof|ID:5128300-1"},
   CellID->485768995]},
 "Proof|ID:5128300-1"->{
  Cell[185262, 5124, 2731, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:5128300", "Proof|ID:5128300-1"},
   CellID->485768995]},
 "ID:906005529"->{
  Cell[189002, 5236, 2813, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:906005529", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo singleton", "IR"},
   CellID->906005529]},
 "reduction modulo singleton"->{
  Cell[189002, 5236, 2813, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:906005529", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo singleton", "IR"},
   CellID->906005529]},
 "IR"->{
  Cell[189002, 5236, 2813, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:906005529", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo singleton", "IR"},
   CellID->906005529],
  Cell[199246, 5529, 1551, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:347299573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo tuple alternative", "IR"},
   CellID->347299573],
  Cell[201511, 5600, 1161, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:200776309", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] TupleToSet", "IR"},
   CellID->200776309],
  Cell[202675, 5638, 1187, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:989570066", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] TupleToSet", "IR"},
   CellID->989570066],
  Cell[203865, 5676, 1181, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:359412397", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow]\[RawStar] TupleToSet", "IR"},
   CellID->359412397],
  Cell[205049, 5714, 1212, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1876144273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow]\[RawStar] TupleToSet", "IR"},
   CellID->1876144273],
  Cell[206264, 5752, 1944, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:495509420", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below TupleToSet", "IR"},
   CellID->495509420],
  Cell[241103, 6716, 1401, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:178557746", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo set alternative", "IR"},
   CellID->178557746],
  Cell[418792, 11505, 1152, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1952076718", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "membership reduceM", "IR"},
   CellID->1952076718],
  Cell[581859, 15712, 1479, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2087655986", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible alternative", "IR"},
   CellID->2087655986],
  Cell[644698, 17425, 1061, 28, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:1592577568", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets red", "IR"},
   CellID->1592577568],
  Cell[645762, 17455, 965, 25, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:698444072", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red isSet", "IR"},
   CellID->698444072],
  Cell[781414, 21184, 2578, 83, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:21782685", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] append", "IR"},
   CellID->21782685],
  Cell[2915661, 78807, 3086, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160768319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0", "IR"},
   CellID->160768319],
  Cell[2996206, 80912, 976, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1760501242", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets ideal", "IR"},
   CellID->1760501242],
  Cell[2997185, 80939, 871, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:554864723", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal isSet", "IR"},
   CellID->554864723],
  Cell[3000347, 81033, 847, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1644449200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideals closed +", "IR"},
   CellID->1644449200],
  Cell[3001197, 81058, 861, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1500033871", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideals closed - binary", "IR"},
   CellID->1500033871],
  Cell[3044231, 82241, 1009, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:279213832", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets tuple-ideal", "IR"},
   CellID->279213832],
  Cell[3046060, 82290, 885, 24, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1587050664", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideals closed +", "IR"},
   CellID->1587050664],
  Cell[3046948, 82316, 896, 24, 29, "FormalTextInputFormula",
   CellTags->{
    "ID:508704301", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideals closed - binary", "IR"},
   CellID->508704301]},
 "Proof|ID:906005529"->{
  Cell[191973, 5326, 4814, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:906005529", "Proof|ID:906005529-1"},
   CellID->520373627]},
 "Proof|ID:906005529-1"->{
  Cell[191973, 5326, 4814, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:906005529", "Proof|ID:906005529-1"},
   CellID->520373627]},
 "ID:708647877"->{
  Cell[197753, 5487, 720, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:708647877", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->708647877]},
 "ID:508075012"->{
  Cell[198476, 5507, 767, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:508075012", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->508075012]},
 "ID:347299573"->{
  Cell[199246, 5529, 1551, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:347299573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo tuple alternative", "IR"},
   CellID->347299573]},
 "reduction modulo tuple alternative"->{
  Cell[199246, 5529, 1551, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:347299573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo tuple alternative", "IR"},
   CellID->347299573]},
 "ID:1756538453"->{
  Cell[200800, 5580, 708, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1756538453", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1756538453]},
 "ID:200776309"->{
  Cell[201511, 5600, 1161, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:200776309", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] TupleToSet", "IR"},
   CellID->200776309]},
 "\[RightArrow] TupleToSet"->{
  Cell[201511, 5600, 1161, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:200776309", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] TupleToSet", "IR"},
   CellID->200776309]},
 "ID:989570066"->{
  Cell[202675, 5638, 1187, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:989570066", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] TupleToSet", "IR"},
   CellID->989570066]},
 "\[LeftRightArrow] TupleToSet"->{
  Cell[202675, 5638, 1187, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:989570066", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] TupleToSet", "IR"},
   CellID->989570066]},
 "ID:359412397"->{
  Cell[203865, 5676, 1181, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:359412397", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow]\[RawStar] TupleToSet", "IR"},
   CellID->359412397]},
 "\[RightArrow]\[RawStar] TupleToSet"->{
  Cell[203865, 5676, 1181, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:359412397", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow]\[RawStar] TupleToSet", "IR"},
   CellID->359412397]},
 "ID:1876144273"->{
  Cell[205049, 5714, 1212, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1876144273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow]\[RawStar] TupleToSet", "IR"},
   CellID->1876144273]},
 "\[LeftRightArrow]\[RawStar] TupleToSet"->{
  Cell[205049, 5714, 1212, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1876144273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow]\[RawStar] TupleToSet", "IR"},
   CellID->1876144273]},
 "ID:495509420"->{
  Cell[206264, 5752, 1944, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:495509420", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below TupleToSet", "IR"},
   CellID->495509420]},
 "connectibility below TupleToSet"->{
  Cell[206264, 5752, 1944, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:495509420", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below TupleToSet", "IR"},
   CellID->495509420]},
 "Proof|ID:347299573"->{
  Cell[208365, 5828, 5202, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:347299573", "Proof|ID:347299573-1"},
   CellID->641607246]},
 "Proof|ID:347299573-1"->{
  Cell[208365, 5828, 5202, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:347299573", "Proof|ID:347299573-1"},
   CellID->641607246]},
 "Proof|ID:200776309"->{
  Cell[214087, 5984, 4223, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:200776309", "Proof|ID:200776309-1"},
   CellID->376820956]},
 "Proof|ID:200776309-1"->{
  Cell[214087, 5984, 4223, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:200776309", "Proof|ID:200776309-1"},
   CellID->376820956]},
 "Proof|ID:989570066"->{
  Cell[218829, 6115, 4279, 113, 70, "ProofInfo",
   CellTags->{"Proof|ID:989570066", "Proof|ID:989570066-1"},
   CellID->450749896]},
 "Proof|ID:989570066-1"->{
  Cell[218829, 6115, 4279, 113, 70, "ProofInfo",
   CellTags->{"Proof|ID:989570066", "Proof|ID:989570066-1"},
   CellID->450749896]},
 "Proof|ID:359412397"->{
  Cell[223629, 6247, 4325, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:359412397", "Proof|ID:359412397-1"},
   CellID->1501514132]},
 "Proof|ID:359412397-1"->{
  Cell[223629, 6247, 4325, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:359412397", "Proof|ID:359412397-1"},
   CellID->1501514132]},
 "Proof|ID:1876144273"->{
  Cell[228475, 6378, 4381, 113, 70, "ProofInfo",
   CellTags->{"Proof|ID:1876144273", "Proof|ID:1876144273-1"},
   CellID->541596138]},
 "Proof|ID:1876144273-1"->{
  Cell[228475, 6378, 4381, 113, 70, "ProofInfo",
   CellTags->{"Proof|ID:1876144273", "Proof|ID:1876144273-1"},
   CellID->541596138]},
 "Proof|ID:495509420"->{
  Cell[233378, 6510, 5028, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:495509420", "Proof|ID:495509420-1"},
   CellID->551004923]},
 "Proof|ID:495509420-1"->{
  Cell[233378, 6510, 5028, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:495509420", "Proof|ID:495509420-1"},
   CellID->551004923]},
 "ID:467637718"->{
  Cell[239630, 6674, 672, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:467637718", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->467637718]},
 "ID:1182167952"->{
  Cell[240305, 6693, 795, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1182167952", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1182167952]},
 "ID:178557746"->{
  Cell[241103, 6716, 1401, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:178557746", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo set alternative", "IR"},
   CellID->178557746]},
 "reduction modulo set alternative"->{
  Cell[241103, 6716, 1401, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:178557746", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo set alternative", "IR"},
   CellID->178557746]},
 "ID:1019233204"->{
  Cell[242507, 6763, 1169, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1019233204", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[Union]"},
   CellID->1019233204]},
 "\[RightArrow] \[Union]"->{
  Cell[242507, 6763, 1169, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1019233204", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[Union]"},
   CellID->1019233204]},
 "ID:331036245"->{
  Cell[243679, 6802, 1194, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:331036245", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] \[Union]"},
   CellID->331036245]},
 "\[LeftRightArrow] \[Union]"->{
  Cell[243679, 6802, 1194, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:331036245", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] \[Union]"},
   CellID->331036245]},
 "ID:78564808"->{
  Cell[244876, 6841, 1185, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:78564808", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow]\[RawStar] \[Union]"},
   CellID->78564808]},
 "\[RightArrow]\[RawStar] \[Union]"->{
  Cell[244876, 6841, 1185, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:78564808", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow]\[RawStar] \[Union]"},
   CellID->78564808]},
 "ID:626278069"->{
  Cell[246064, 6880, 1264, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:626278069", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow]\[RawStar] \[Union]"},
   CellID->626278069]},
 "\[LeftRightArrow]\[RawStar] \[Union]"->{
  Cell[246064, 6880, 1264, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:626278069", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow]\[RawStar] \[Union]"},
   CellID->626278069]},
 "ID:356849840"->{
  Cell[247331, 6920, 1922, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:356849840", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below \[Union]"},
   CellID->356849840]},
 "connectibility below \[Union]"->{
  Cell[247331, 6920, 1922, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:356849840", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below \[Union]"},
   CellID->356849840]},
 "Proof|ID:178557746"->{
  Cell[249410, 6994, 4661, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:178557746", "Proof|ID:178557746-1"},
   CellID->93489966]},
 "Proof|ID:178557746-1"->{
  Cell[249410, 6994, 4661, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:178557746", "Proof|ID:178557746-1"},
   CellID->93489966]},
 "Proof|ID:1019233204"->{
  Cell[254591, 7138, 4567, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:1019233204", "Proof|ID:1019233204-1"},
   CellID->9184570]},
 "Proof|ID:1019233204-1"->{
  Cell[254591, 7138, 4567, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:1019233204", "Proof|ID:1019233204-1"},
   CellID->9184570]},
 "Proof|ID:331036245"->{
  Cell[259679, 7278, 4608, 120, 70, "ProofInfo",
   CellTags->{"Proof|ID:331036245", "Proof|ID:331036245-1"},
   CellID->970023700]},
 "Proof|ID:331036245-1"->{
  Cell[259679, 7278, 4608, 120, 70, "ProofInfo",
   CellTags->{"Proof|ID:331036245", "Proof|ID:331036245-1"},
   CellID->970023700]},
 "Proof|ID:78564808"->{
  Cell[264807, 7417, 4857, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:78564808", "Proof|ID:78564808-1"},
   CellID->925666063]},
 "Proof|ID:78564808-1"->{
  Cell[264807, 7417, 4857, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:78564808", "Proof|ID:78564808-1"},
   CellID->925666063]},
 "Proof|ID:626278069"->{
  Cell[270183, 7560, 4911, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:626278069", "Proof|ID:626278069-1"},
   CellID->1953121485]},
 "Proof|ID:626278069-1"->{
  Cell[270183, 7560, 4911, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:626278069", "Proof|ID:626278069-1"},
   CellID->1953121485]},
 "Proof|ID:356849840"->{
  Cell[275614, 7703, 5560, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:356849840", "Proof|ID:356849840-1"},
   CellID->205255475]},
 "Proof|ID:356849840-1"->{
  Cell[275614, 7703, 5560, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:356849840", "Proof|ID:356849840-1"},
   CellID->205255475]},
 "ID:366153355"->{
  Cell[282453, 7886, 952, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:366153355", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->366153355]},
 "ID:221340803"->{
  Cell[283408, 7912, 699, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:221340803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->221340803]},
 "ID:328195853"->{
  Cell[284110, 7932, 1219, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:328195853", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "2"},
   CellID->328195853]},
 "2"->{
  Cell[284110, 7932, 1219, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:328195853", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "2"},
   CellID->328195853]},
 "ID:1692465705"->{
  Cell[285332, 7965, 997, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1692465705", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "3"},
   CellID->1692465705]},
 "3"->{
  Cell[285332, 7965, 997, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1692465705", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "3"},
   CellID->1692465705]},
 "ID:188743948"->{
  Cell[286332, 7995, 996, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:188743948", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "4"},
   CellID->188743948]},
 "4"->{
  Cell[286332, 7995, 996, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:188743948", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "4"},
   CellID->188743948]},
 "ID:115635911"->{
  Cell[287331, 8025, 1338, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:115635911", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "5"},
   CellID->115635911]},
 "5"->{
  Cell[287331, 8025, 1338, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:115635911", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "5"},
   CellID->115635911]},
 "ID:1118510950"->{
  Cell[288672, 8068, 1054, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1118510950", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isCommonReducible symmetric"},
   CellID->1118510950]},
 "isCommonReducible symmetric"->{
  Cell[288672, 8068, 1054, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1118510950", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isCommonReducible symmetric"},
   CellID->1118510950]},
 "ID:113485898"->{
  Cell[289729, 8098, 1007, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:113485898", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isNTCR symmetric"},
   CellID->113485898]},
 "isNTCR symmetric"->{
  Cell[289729, 8098, 1007, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:113485898", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isNTCR symmetric"},
   CellID->113485898]},
 "ID:2056330192"->{
  Cell[290739, 8128, 1016, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2056330192", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isMNTCR symmetric"},
   CellID->2056330192]},
 "isMNTCR symmetric"->{
  Cell[290739, 8128, 1016, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2056330192", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isMNTCR symmetric"},
   CellID->2056330192]},
 "ID:463339584"->{
  Cell[291758, 8158, 1420, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:463339584", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers symmetric"},
   CellID->463339584]},
 "areCPMultipliers symmetric"->{
  Cell[291758, 8158, 1420, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:463339584", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers symmetric"},
   CellID->463339584]},
 "ID:449270875"->{
  Cell[293181, 8202, 1867, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:449270875", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers sufficient"},
   CellID->449270875]},
 "areCPMultipliers sufficient"->{
  Cell[293181, 8202, 1867, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:449270875", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers sufficient"},
   CellID->449270875]},
 "Proof|ID:328195853"->{
  Cell[295203, 8271, 5357, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:328195853", "Proof|ID:328195853-1"},
   CellID->667808408]},
 "Proof|ID:328195853-1"->{
  Cell[295203, 8271, 5357, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:328195853", "Proof|ID:328195853-1"},
   CellID->667808408]},
 "Proof|ID:1692465705"->{
  Cell[301080, 8422, 5322, 133, 70, "ProofInfo",
   CellTags->{"Proof|ID:1692465705", "Proof|ID:1692465705-1"},
   CellID->982819604]},
 "Proof|ID:1692465705-1"->{
  Cell[301080, 8422, 5322, 133, 70, "ProofInfo",
   CellTags->{"Proof|ID:1692465705", "Proof|ID:1692465705-1"},
   CellID->982819604]},
 "Proof|ID:188743948"->{
  Cell[306924, 8574, 5318, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:188743948", "Proof|ID:188743948-1"},
   CellID->1789453742]},
 "Proof|ID:188743948-1"->{
  Cell[306924, 8574, 5318, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:188743948", "Proof|ID:188743948-1"},
   CellID->1789453742]},
 "Proof|ID:115635911"->{
  Cell[312762, 8725, 6602, 156, 70, "ProofInfo",
   CellTags->{"Proof|ID:115635911", "Proof|ID:115635911-1"},
   CellID->1648327750]},
 "Proof|ID:115635911-1"->{
  Cell[312762, 8725, 6602, 156, 70, "ProofInfo",
   CellTags->{"Proof|ID:115635911", "Proof|ID:115635911-1"},
   CellID->1648327750]},
 "Proof|ID:1118510950"->{
  Cell[319884, 8900, 5528, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1118510950", "Proof|ID:1118510950-1"},
   CellID->144468455]},
 "Proof|ID:1118510950-1"->{
  Cell[319884, 8900, 5528, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1118510950", "Proof|ID:1118510950-1"},
   CellID->144468455]},
 "Proof|ID:113485898"->{
  Cell[325934, 9054, 5454, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:113485898", "Proof|ID:113485898-1"},
   CellID->1668243539]},
 "Proof|ID:113485898-1"->{
  Cell[325934, 9054, 5454, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:113485898", "Proof|ID:113485898-1"},
   CellID->1668243539]},
 "Proof|ID:2056330192"->{
  Cell[331909, 9207, 5469, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:2056330192", "Proof|ID:2056330192-1"},
   CellID->1822040650]},
 "Proof|ID:2056330192-1"->{
  Cell[331909, 9207, 5469, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:2056330192", "Proof|ID:2056330192-1"},
   CellID->1822040650]},
 "Proof|ID:463339584"->{
  Cell[337900, 9361, 6754, 158, 70, "ProofInfo",
   CellTags->{"Proof|ID:463339584", "Proof|ID:463339584-1"},
   CellID->2055115730]},
 "Proof|ID:463339584-1"->{
  Cell[337900, 9361, 6754, 158, 70, "ProofInfo",
   CellTags->{"Proof|ID:463339584", "Proof|ID:463339584-1"},
   CellID->2055115730]},
 "Proof|ID:449270875"->{
  Cell[345175, 9538, 8468, 194, 70, "ProofInfo",
   CellTags->{"Proof|ID:449270875", "Proof|ID:449270875-1"},
   CellID->1693763802]},
 "Proof|ID:449270875-1"->{
  Cell[345175, 9538, 8468, 194, 70, "ProofInfo",
   CellTags->{"Proof|ID:449270875", "Proof|ID:449270875-1"},
   CellID->1693763802]},
 "ID:523863903"->{
  Cell[354664, 9768, 751, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:523863903", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->523863903]},
 "ID:1519187240"->{
  Cell[355418, 9790, 789, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1519187240", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1519187240]},
 "ID:1878546324"->{
  Cell[356210, 9812, 1052, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1878546324", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "2'"},
   CellID->1878546324]},
 "2'"->{
  Cell[356210, 9812, 1052, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1878546324", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "2'"},
   CellID->1878546324]},
 "ID:1781176225"->{
  Cell[357265, 9842, 1025, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1781176225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "3'"},
   CellID->1781176225]},
 "3'"->{
  Cell[357265, 9842, 1025, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1781176225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "3'"},
   CellID->1781176225]},
 "ID:1582995786"->{
  Cell[358293, 9872, 1032, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1582995786", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "4'"},
   CellID->1582995786]},
 "4'"->{
  Cell[358293, 9872, 1032, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1582995786", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "4'"},
   CellID->1582995786]},
 "ID:2020067658"->{
  Cell[359328, 9902, 1370, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2020067658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "5'"},
   CellID->2020067658]},
 "5'"->{
  Cell[359328, 9902, 1370, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2020067658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "5'"},
   CellID->2020067658]},
 "Proof|ID:1878546324"->{
  Cell[360854, 9955, 5875, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:1878546324", "Proof|ID:1878546324-1"},
   CellID->1762740644]},
 "Proof|ID:1878546324-1"->{
  Cell[360854, 9955, 5875, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:1878546324", "Proof|ID:1878546324-1"},
   CellID->1762740644]},
 "Proof|ID:1781176225"->{
  Cell[367251, 10121, 5831, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:1781176225", "Proof|ID:1781176225-1"},
   CellID->1337859754]},
 "Proof|ID:1781176225-1"->{
  Cell[367251, 10121, 5831, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:1781176225", "Proof|ID:1781176225-1"},
   CellID->1337859754]},
 "Proof|ID:1582995786"->{
  Cell[373603, 10287, 5835, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:1582995786", "Proof|ID:1582995786-1"},
   CellID->1038215936]},
 "Proof|ID:1582995786-1"->{
  Cell[373603, 10287, 5835, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:1582995786", "Proof|ID:1582995786-1"},
   CellID->1038215936]},
 "Proof|ID:2020067658"->{
  Cell[379958, 10453, 7103, 171, 70, "ProofInfo",
   CellTags->{"Proof|ID:2020067658", "Proof|ID:2020067658-1"},
   CellID->1751126895]},
 "Proof|ID:2020067658-1"->{
  Cell[379958, 10453, 7103, 171, 70, "ProofInfo",
   CellTags->{"Proof|ID:2020067658", "Proof|ID:2020067658-1"},
   CellID->1751126895]},
 "ID:1823284442"->{
  Cell[388163, 10660, 730, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1823284442", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1823284442]},
 "ID:284217868"->{
  Cell[388896, 10682, 762, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:284217868", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->284217868]},
 "ID:628244416"->{
  Cell[389661, 10704, 1035, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:628244416", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "6"},
   CellID->628244416]},
 "6"->{
  Cell[389661, 10704, 1035, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:628244416", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "6"},
   CellID->628244416]},
 "ID:1255486983"->{
  Cell[390699, 10732, 2058, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1255486983", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative-irrelative 2"},
   CellID->1255486983]},
 "correlative-irrelative 2"->{
  Cell[390699, 10732, 2058, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1255486983", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative-irrelative 2"},
   CellID->1255486983]},
 "Proof|ID:628244416"->{
  Cell[392914, 10803, 5247, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:628244416", "Proof|ID:628244416-1"},
   CellID->1993474493]},
 "Proof|ID:628244416-1"->{
  Cell[392914, 10803, 5247, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:628244416", "Proof|ID:628244416-1"},
   CellID->1993474493]},
 "Proof|ID:1255486983"->{
  Cell[398681, 10958, 8860, 207, 70, "ProofInfo",
   CellTags->{"Proof|ID:1255486983", "Proof|ID:1255486983-1"},
   CellID->558484412]},
 "Proof|ID:1255486983-1"->{
  Cell[398681, 10958, 8860, 207, 70, "ProofInfo",
   CellTags->{"Proof|ID:1255486983", "Proof|ID:1255486983-1"},
   CellID->558484412]},
 "ID:1382908475"->{
  Cell[408854, 11206, 740, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1382908475", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1382908475]},
 "ID:442895874"->{
  Cell[410158, 11244, 853, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:442895874", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->442895874]},
 "ID:927154658"->{
  Cell[411014, 11268, 816, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:927154658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->927154658]},
 "ID:1300106055"->{
  Cell[411833, 11291, 867, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1300106055", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM 0"},
   CellID->1300106055]},
 "reduceM 0"->{
  Cell[411833, 11291, 867, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1300106055", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM 0"},
   CellID->1300106055]},
 "ID:438711618"->{
  Cell[412703, 11317, 1223, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:438711618", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM 1"},
   CellID->438711618]},
 "reduceM 1"->{
  Cell[412703, 11317, 1223, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:438711618", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM 1"},
   CellID->438711618]},
 "ID:2101232329"->{
  Cell[413929, 11356, 1524, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2101232329", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM alternative"},
   CellID->2101232329]},
 "reduceM alternative"->{
  Cell[413929, 11356, 1524, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2101232329", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM alternative"},
   CellID->2101232329]},
 "ID:96556784"->{
  Cell[415456, 11403, 1487, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:96556784", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM+1"},
   CellID->96556784]},
 "reduceM+1"->{
  Cell[415456, 11403, 1487, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:96556784", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM+1"},
   CellID->96556784]},
 "ID:1465630278"->{
  Cell[416946, 11447, 1843, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1465630278", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM alternative II"},
   CellID->1465630278]},
 "reduceM alternative II"->{
  Cell[416946, 11447, 1843, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1465630278", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM alternative II"},
   CellID->1465630278]},
 "ID:1952076718"->{
  Cell[418792, 11505, 1152, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1952076718", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "membership reduceM", "IR"},
   CellID->1952076718]},
 "membership reduceM"->{
  Cell[418792, 11505, 1152, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1952076718", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "membership reduceM", "IR"},
   CellID->1952076718]},
 "ID:1110504237"->{
  Cell[419947, 11539, 792, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1110504237", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1110504237]},
 "ID:1236615595"->{
  Cell[420742, 11562, 1225, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1236615595", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM-1"},
   CellID->1236615595]},
 "reduceM-1"->{
  Cell[420742, 11562, 1225, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1236615595", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM-1"},
   CellID->1236615595]},
 "ID:203149419"->{
  Cell[421970, 11600, 1127, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:203149419", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM-1 II"},
   CellID->203149419]},
 "reduceM-1 II"->{
  Cell[421970, 11600, 1127, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:203149419", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM-1 II"},
   CellID->203149419]},
 "ID:333495152"->{
  Cell[423100, 11635, 2109, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:333495152", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible reduceM"},
   CellID->333495152]},
 "isReducible reduceM"->{
  Cell[423100, 11635, 2109, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:333495152", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible reduceM"},
   CellID->333495152]},
 "ID:332101094"->{
  Cell[425212, 11701, 2187, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:332101094", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM"},
   CellID->332101094]},
 "symred reduceM"->{
  Cell[425212, 11701, 2187, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:332101094", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM"},
   CellID->332101094]},
 "ID:296464255"->{
  Cell[427402, 11768, 3277, 99, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:296464255", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM I"},
   CellID->296464255]},
 "symred reduceM I"->{
  Cell[427402, 11768, 3277, 99, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:296464255", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM I"},
   CellID->296464255]},
 "ID:194930655"->{
  Cell[430682, 11869, 2682, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:194930655", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM II"},
   CellID->194930655]},
 "symred reduceM II"->{
  Cell[430682, 11869, 2682, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:194930655", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM II"},
   CellID->194930655]},
 "ID:935171195"->{
  Cell[433367, 11952, 2870, 87, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:935171195", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM III"},
   CellID->935171195]},
 "symred reduceM III"->{
  Cell[433367, 11952, 2870, 87, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:935171195", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM III"},
   CellID->935171195]},
 "Proof|ID:1300106055"->{
  Cell[436394, 12050, 4093, 111, 70, "ProofInfo",
   CellTags->{"Proof|ID:1300106055", "Proof|ID:1300106055-1"},
   CellID->587044218]},
 "Proof|ID:1300106055-1"->{
  Cell[436394, 12050, 4093, 111, 70, "ProofInfo",
   CellTags->{"Proof|ID:1300106055", "Proof|ID:1300106055-1"},
   CellID->587044218]},
 "Proof|ID:438711618"->{
  Cell[441052, 12181, 5331, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:438711618", "Proof|ID:438711618-1"},
   CellID->457182048]},
 "Proof|ID:438711618-1"->{
  Cell[441052, 12181, 5331, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:438711618", "Proof|ID:438711618-1"},
   CellID->457182048]},
 "Proof|ID:2101232329"->{
  Cell[446903, 12338, 6500, 165, 70, "ProofInfo",
   CellTags->{"Proof|ID:2101232329", "Proof|ID:2101232329-1"},
   CellID->1720715120]},
 "Proof|ID:2101232329-1"->{
  Cell[446903, 12338, 6500, 165, 70, "ProofInfo",
   CellTags->{"Proof|ID:2101232329", "Proof|ID:2101232329-1"},
   CellID->1720715120]},
 "Proof|ID:96556784"->{
  Cell[453924, 12522, 6545, 166, 70, "ProofInfo",
   CellTags->{"Proof|ID:96556784", "Proof|ID:96556784-1"},
   CellID->194852796]},
 "Proof|ID:96556784-1"->{
  Cell[453924, 12522, 6545, 166, 70, "ProofInfo",
   CellTags->{"Proof|ID:96556784", "Proof|ID:96556784-1"},
   CellID->194852796]},
 "Proof|ID:1465630278"->{
  Cell[460988, 12707, 7994, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1465630278", "Proof|ID:1465630278-1"},
   CellID->1634286780]},
 "Proof|ID:1465630278-1"->{
  Cell[460988, 12707, 7994, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1465630278", "Proof|ID:1465630278-1"},
   CellID->1634286780]},
 "Proof|ID:1952076718"->{
  Cell[469504, 12923, 4865, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:1952076718", "Proof|ID:1952076718-1"},
   CellID->1918441122]},
 "Proof|ID:1952076718-1"->{
  Cell[469504, 12923, 4865, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:1952076718", "Proof|ID:1952076718-1"},
   CellID->1918441122]},
 "Proof|ID:1236615595"->{
  Cell[474890, 13067, 6661, 171, 70, "ProofInfo",
   CellTags->{"Proof|ID:1236615595", "Proof|ID:1236615595-1"},
   CellID->536851896]},
 "Proof|ID:1236615595-1"->{
  Cell[474890, 13067, 6661, 171, 70, "ProofInfo",
   CellTags->{"Proof|ID:1236615595", "Proof|ID:1236615595-1"},
   CellID->536851896]},
 "Proof|ID:203149419"->{
  Cell[482072, 13257, 6361, 158, 70, "ProofInfo",
   CellTags->{"Proof|ID:203149419", "Proof|ID:203149419-1"},
   CellID->916037400]},
 "Proof|ID:203149419-1"->{
  Cell[482072, 13257, 6361, 158, 70, "ProofInfo",
   CellTags->{"Proof|ID:203149419", "Proof|ID:203149419-1"},
   CellID->916037400]},
 "Proof|ID:333495152"->{
  Cell[488996, 13435, 10693, 254, 70, "ProofInfo",
   CellTags->{"Proof|ID:333495152", "Proof|ID:333495152-1"},
   CellID->193609441]},
 "Proof|ID:333495152-1"->{
  Cell[488996, 13435, 10693, 254, 70, "ProofInfo",
   CellTags->{"Proof|ID:333495152", "Proof|ID:333495152-1"},
   CellID->193609441]},
 "Proof|ID:332101094"->{
  Cell[500210, 13708, 11549, 264, 70, "ProofInfo",
   CellTags->{"Proof|ID:332101094", "Proof|ID:332101094-1"},
   CellID->2106893222]},
 "Proof|ID:332101094-1"->{
  Cell[500210, 13708, 11549, 264, 70, "ProofInfo",
   CellTags->{"Proof|ID:332101094", "Proof|ID:332101094-1"},
   CellID->2106893222]},
 "Proof|ID:296464255"->{
  Cell[512280, 13991, 15830, 364, 70, "ProofInfo",
   CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
   CellID->1877618114],
  Cell[557402, 15047, 583, 11, 70, "Message",
   CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
   CellID->20878089],
  Cell[557988, 15060, 585, 11, 70, "Message",
   CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
   CellID->1281008759],
  Cell[558576, 15073, 585, 11, 70, "Message",
   CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
   CellID->1946324498],
  Cell[559164, 15086, 597, 12, 70, "Message",
   CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
   CellID->1486921856]},
 "Proof|ID:296464255-1"->{
  Cell[512280, 13991, 15830, 364, 70, "ProofInfo",
   CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
   CellID->1877618114],
  Cell[557402, 15047, 583, 11, 70, "Message",
   CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
   CellID->20878089],
  Cell[557988, 15060, 585, 11, 70, "Message",
   CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
   CellID->1281008759],
  Cell[558576, 15073, 585, 11, 70, "Message",
   CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
   CellID->1946324498],
  Cell[559164, 15086, 597, 12, 70, "Message",
   CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
   CellID->1486921856]},
 "Proof|ID:194930655"->{
  Cell[528631, 14374, 13369, 308, 70, "ProofInfo",
   CellTags->{"Proof|ID:194930655", "Proof|ID:194930655-1"},
   CellID->230812744]},
 "Proof|ID:194930655-1"->{
  Cell[528631, 14374, 13369, 308, 70, "ProofInfo",
   CellTags->{"Proof|ID:194930655", "Proof|ID:194930655-1"},
   CellID->230812744]},
 "Proof|ID:935171195"->{
  Cell[542520, 14701, 14395, 330, 70, "ProofInfo",
   CellTags->{"Proof|ID:935171195", "Proof|ID:935171195-1"},
   CellID->577756398]},
 "Proof|ID:935171195-1"->{
  Cell[542520, 14701, 14395, 330, 70, "ProofInfo",
   CellTags->{"Proof|ID:935171195", "Proof|ID:935171195-1"},
   CellID->577756398]},
 "ID:1368508872"->{
  Cell[560259, 15118, 887, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1368508872", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1368508872]},
 "ID:1610984937"->{
  Cell[561149, 15143, 1430, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1610984937", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areIrrelative symmetric"},
   CellID->1610984937]},
 "areIrrelative symmetric"->{
  Cell[561149, 15143, 1430, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1610984937", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areIrrelative symmetric"},
   CellID->1610984937]},
 "ID:981548331"->{
  Cell[562582, 15185, 1060, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:981548331", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCorrelative symmetric"},
   CellID->981548331]},
 "areCorrelative symmetric"->{
  Cell[562582, 15185, 1060, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:981548331", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCorrelative symmetric"},
   CellID->981548331]},
 "Proof|ID:1610984937"->{
  Cell[563798, 15223, 7069, 171, 70, "ProofInfo",
   CellTags->{"Proof|ID:1610984937", "Proof|ID:1610984937-1"},
   CellID->1688165426]},
 "Proof|ID:1610984937-1"->{
  Cell[563798, 15223, 7069, 171, 70, "ProofInfo",
   CellTags->{"Proof|ID:1610984937", "Proof|ID:1610984937-1"},
   CellID->1688165426]},
 "Proof|ID:981548331"->{
  Cell[571389, 15413, 4589, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:981548331", "Proof|ID:981548331-1"},
   CellID->437316908]},
 "Proof|ID:981548331-1"->{
  Cell[571389, 15413, 4589, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:981548331", "Proof|ID:981548331-1"},
   CellID->437316908]},
 "ID:1547543437"->{
  Cell[577102, 15568, 976, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1547543437", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1547543437]},
 "ID:1699369800"->{
  Cell[578081, 15594, 1300, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1699369800", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible \[RightArrow]"},
   CellID->1699369800]},
 "isReducible \[RightArrow]"->{
  Cell[578081, 15594, 1300, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1699369800", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible \[RightArrow]"},
   CellID->1699369800]},
 "ID:1596540807"->{
  Cell[579384, 15634, 1377, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1596540807", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] isReducible"},
   CellID->1596540807]},
 "\[RightArrow] isReducible"->{
  Cell[579384, 15634, 1377, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1596540807", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] isReducible"},
   CellID->1596540807]},
 "ID:1641388157"->{
  Cell[580764, 15679, 1092, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1641388157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible symred"},
   CellID->1641388157]},
 "isReducible symred"->{
  Cell[580764, 15679, 1092, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1641388157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible symred"},
   CellID->1641388157]},
 "ID:2087655986"->{
  Cell[581859, 15712, 1479, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2087655986", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible alternative", "IR"},
   CellID->2087655986]},
 "isReducible alternative"->{
  Cell[581859, 15712, 1479, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2087655986", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible alternative", "IR"},
   CellID->2087655986]},
 "ID:584152369"->{
  Cell[583341, 15758, 2777, 85, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:584152369", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers \[RightArrow] tuple"},
   CellID->584152369]},
 "areCPMultipliers \[RightArrow] tuple"->{
  Cell[583341, 15758, 2777, 85, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:584152369", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers \[RightArrow] tuple"},
   CellID->584152369]},
 "Proof|ID:1699369800"->{
  Cell[586276, 15854, 5258, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:1699369800", "Proof|ID:1699369800-1"},
   CellID->199699432]},
 "Proof|ID:1699369800-1"->{
  Cell[586276, 15854, 5258, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:1699369800", "Proof|ID:1699369800-1"},
   CellID->199699432]},
 "Proof|ID:1596540807"->{
  Cell[592054, 16010, 6412, 159, 70, "ProofInfo",
   CellTags->{"Proof|ID:1596540807", "Proof|ID:1596540807-1"},
   CellID->314505815]},
 "Proof|ID:1596540807-1"->{
  Cell[592054, 16010, 6412, 159, 70, "ProofInfo",
   CellTags->{"Proof|ID:1596540807", "Proof|ID:1596540807-1"},
   CellID->314505815]},
 "Proof|ID:1641388157"->{
  Cell[598986, 16188, 5339, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1641388157", "Proof|ID:1641388157-1"},
   CellID->20930826]},
 "Proof|ID:1641388157-1"->{
  Cell[598986, 16188, 5339, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1641388157", "Proof|ID:1641388157-1"},
   CellID->20930826]},
 "Proof|ID:2087655986"->{
  Cell[604847, 16342, 4738, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:2087655986", "Proof|ID:2087655986-1"},
   CellID->65047792]},
 "Proof|ID:2087655986-1"->{
  Cell[604847, 16342, 4738, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:2087655986", "Proof|ID:2087655986-1"},
   CellID->65047792]},
 "Proof|ID:584152369"->{
  Cell[610106, 16488, 9817, 234, 70, "ProofInfo",
   CellTags->{"Proof|ID:584152369", "Proof|ID:584152369-1"},
   CellID->444231087]},
 "Proof|ID:584152369-1"->{
  Cell[610106, 16488, 9817, 234, 70, "ProofInfo",
   CellTags->{"Proof|ID:584152369", "Proof|ID:584152369-1"},
   CellID->444231087]},
 "ID:593421628"->{
  Cell[621060, 16762, 846, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:593421628", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->593421628]},
 "ID:680005293"->{
  Cell[621909, 16787, 1248, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:680005293", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "1"},
   CellID->680005293]},
 "1"->{
  Cell[621909, 16787, 1248, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:680005293", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "1"},
   CellID->680005293]},
 "ID:1320396264"->{
  Cell[623160, 16829, 1078, 33, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1320396264", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1320396264]},
 "ID:1127232822"->{
  Cell[624241, 16864, 1185, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1127232822", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] reductor"},
   CellID->1127232822]},
 "\[RightArrow] reductor"->{
  Cell[624241, 16864, 1185, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1127232822", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] reductor"},
   CellID->1127232822]},
 "ID:256297052"->{
  Cell[625429, 16901, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:256297052", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Succeeds] reductor"},
   CellID->256297052]},
 "\[Succeeds] reductor"->{
  Cell[625429, 16901, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:256297052", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Succeeds] reductor"},
   CellID->256297052]},
 "Proof|ID:680005293"->{
  Cell[626392, 16933, 5315, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:680005293", "Proof|ID:680005293-1"},
   CellID->41814150]},
 "Proof|ID:680005293-1"->{
  Cell[626392, 16933, 5315, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:680005293", "Proof|ID:680005293-1"},
   CellID->41814150]},
 "Proof|ID:1127232822"->{
  Cell[632227, 17088, 5329, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1127232822", "Proof|ID:1127232822-1"},
   CellID->625038390]},
 "Proof|ID:1127232822-1"->{
  Cell[632227, 17088, 5329, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1127232822", "Proof|ID:1127232822-1"},
   CellID->625038390]},
 "Proof|ID:256297052"->{
  Cell[638077, 17246, 4460, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:256297052", "Proof|ID:256297052-1"},
   CellID->256602949]},
 "Proof|ID:256297052-1"->{
  Cell[638077, 17246, 4460, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:256297052", "Proof|ID:256297052-1"},
   CellID->256602949]},
 "ID:1083403145"->{
  Cell[643610, 17397, 1085, 26, 70, "GlobalDeclaration",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:1083403145", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1083403145]},
 "ID:1592577568"->{
  Cell[644698, 17425, 1061, 28, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:1592577568", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets red", "IR"},
   CellID->1592577568]},
 "DomainSets red"->{
  Cell[644698, 17425, 1061, 28, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:1592577568", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets red", "IR"},
   CellID->1592577568]},
 "ID:698444072"->{
  Cell[645762, 17455, 965, 25, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:698444072", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red isSet", "IR"},
   CellID->698444072]},
 "red isSet"->{
  Cell[645762, 17455, 965, 25, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:698444072", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red isSet", "IR"},
   CellID->698444072]},
 "ID:277656140"->{
  Cell[646730, 17482, 1781, 53, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:277656140", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red"},
   CellID->277656140]},
 "\[Element] red"->{
  Cell[646730, 17482, 1781, 53, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:277656140", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red"},
   CellID->277656140]},
 "ID:454978652"->{
  Cell[648514, 17537, 1772, 54, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:454978652", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red singleton"},
   CellID->454978652]},
 "\[Element] red singleton"->{
  Cell[648514, 17537, 1772, 54, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:454978652", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red singleton"},
   CellID->454978652]},
 "ID:259827218"->{
  Cell[650289, 17593, 3071, 92, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:259827218", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red abstraction term"},
   CellID->259827218]},
 "\[Element] red abstraction term"->{
  Cell[650289, 17593, 3071, 92, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:259827218", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red abstraction term"},
   CellID->259827218]},
 "ID:1273500442"->{
  Cell[653363, 17687, 1554, 44, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:1273500442", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red \[Union]"},
   CellID->1273500442]},
 "\[Element] red \[Union]"->{
  Cell[653363, 17687, 1554, 44, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:1273500442", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red \[Union]"},
   CellID->1273500442]},
 "ID:1877397922"->{
  Cell[654920, 17733, 1143, 31, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:1877397922", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Union]"},
   CellID->1877397922]},
 "red \[Union]"->{
  Cell[654920, 17733, 1143, 31, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:1877397922", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Union]"},
   CellID->1877397922]},
 "ID:2121666706"->{
  Cell[656066, 17766, 2110, 64, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:2121666706", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red \[Union] singleton"},
   CellID->2121666706]},
 "\[Element] red \[Union] singleton"->{
  Cell[656066, 17766, 2110, 64, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:2121666706", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red \[Union] singleton"},
   CellID->2121666706]},
 "ID:947744040"->{
  Cell[658179, 17832, 1088, 29, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:947744040", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SubsetEqual] red \[SubsetEqual]"},
   CellID->947744040]},
 "\[SubsetEqual] red \[SubsetEqual]"->{
  Cell[658179, 17832, 1088, 29, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:947744040", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SubsetEqual] red \[SubsetEqual]"},
   CellID->947744040]},
 "ID:52516181"->{
  Cell[659270, 17863, 2045, 61, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:52516181", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset]"},
   CellID->52516181]},
 "red \[Subset]"->{
  Cell[659270, 17863, 2045, 61, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:52516181", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset]"},
   CellID->52516181]},
 "ID:194436663"->{
  Cell[661318, 17926, 2248, 69, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:194436663", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset] \[Union] singleton"},
   CellID->194436663]},
 "red \[Subset] \[Union] singleton"->{
  Cell[661318, 17926, 2248, 69, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:194436663", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset] \[Union] singleton"},
   CellID->194436663]},
 "ID:844584535"->{
  Cell[663569, 17997, 1794, 53, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:844584535", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset] \[Union] singleton II"},
   CellID->844584535]},
 "red \[Subset] \[Union] singleton II"->{
  Cell[663569, 17997, 1794, 53, 70, "FormalTextInputFormula",
   CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
   CellTags->{
    "ID:844584535", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset] \[Union] singleton II"},
   CellID->844584535]},
 "Proof|ID:1592577568"->{
  Cell[665617, 18062, 3174, 89, 70, "ProofInfo",
   CellTags->{"Proof|ID:1592577568", "Proof|ID:1592577568-1"},
   CellID->272356749]},
 "Proof|ID:1592577568-1"->{
  Cell[665617, 18062, 3174, 89, 70, "ProofInfo",
   CellTags->{"Proof|ID:1592577568", "Proof|ID:1592577568-1"},
   CellID->272356749]},
 "Proof|ID:698444072"->{
  Cell[669312, 18170, 2889, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:698444072", "Proof|ID:698444072-1"},
   CellID->6276063]},
 "Proof|ID:698444072-1"->{
  Cell[669312, 18170, 2889, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:698444072", "Proof|ID:698444072-1"},
   CellID->6276063]},
 "Proof|ID:277656140"->{
  Cell[672720, 18271, 4978, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:277656140", "Proof|ID:277656140-1"},
   CellID->45788410]},
 "Proof|ID:277656140-1"->{
  Cell[672720, 18271, 4978, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:277656140", "Proof|ID:277656140-1"},
   CellID->45788410]},
 "Proof|ID:454978652"->{
  Cell[678218, 18418, 4420, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:454978652", "Proof|ID:454978652-1"},
   CellID->54717562]},
 "Proof|ID:454978652-1"->{
  Cell[678218, 18418, 4420, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:454978652", "Proof|ID:454978652-1"},
   CellID->54717562]},
 "Proof|ID:259827218"->{
  Cell[683155, 18553, 8263, 199, 70, "ProofInfo",
   CellTags->{"Proof|ID:259827218", "Proof|ID:259827218-1"},
   CellID->183930652]},
 "Proof|ID:259827218-1"->{
  Cell[683155, 18553, 8263, 199, 70, "ProofInfo",
   CellTags->{"Proof|ID:259827218", "Proof|ID:259827218-1"},
   CellID->183930652]},
 "Proof|ID:1273500442"->{
  Cell[691938, 18771, 5332, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1273500442", "Proof|ID:1273500442-1"},
   CellID->172667338]},
 "Proof|ID:1273500442-1"->{
  Cell[691938, 18771, 5332, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1273500442", "Proof|ID:1273500442-1"},
   CellID->172667338]},
 "Proof|ID:1877397922"->{
  Cell[697791, 18925, 4354, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1877397922", "Proof|ID:1877397922-1"},
   CellID->133276848]},
 "Proof|ID:1877397922-1"->{
  Cell[697791, 18925, 4354, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1877397922", "Proof|ID:1877397922-1"},
   CellID->133276848]},
 "Proof|ID:2121666706"->{
  Cell[702666, 19059, 5840, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:2121666706", "Proof|ID:2121666706-1"},
   CellID->185736564]},
 "Proof|ID:2121666706-1"->{
  Cell[702666, 19059, 5840, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:2121666706", "Proof|ID:2121666706-1"},
   CellID->185736564]},
 "Proof|ID:947744040"->{
  Cell[709027, 19225, 4177, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:947744040", "Proof|ID:947744040-1"},
   CellID->249774080]},
 "Proof|ID:947744040-1"->{
  Cell[709027, 19225, 4177, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:947744040", "Proof|ID:947744040-1"},
   CellID->249774080]},
 "Proof|ID:52516181"->{
  Cell[713724, 19354, 6543, 164, 70, "ProofInfo",
   CellTags->{"Proof|ID:52516181", "Proof|ID:52516181-1"},
   CellID->50697679]},
 "Proof|ID:52516181-1"->{
  Cell[713724, 19354, 6543, 164, 70, "ProofInfo",
   CellTags->{"Proof|ID:52516181", "Proof|ID:52516181-1"},
   CellID->50697679]},
 "Proof|ID:194436663"->{
  Cell[720786, 19537, 6353, 161, 70, "ProofInfo",
   CellTags->{"Proof|ID:194436663", "Proof|ID:194436663-1"},
   CellID->210264695]},
 "Proof|ID:194436663-1"->{
  Cell[720786, 19537, 6353, 161, 70, "ProofInfo",
   CellTags->{"Proof|ID:194436663", "Proof|ID:194436663-1"},
   CellID->210264695]},
 "Proof|ID:844584535"->{
  Cell[727658, 19717, 6028, 152, 70, "ProofInfo",
   CellTags->{"Proof|ID:844584535", "Proof|ID:844584535-1"},
   CellID->34854128]},
 "Proof|ID:844584535-1"->{
  Cell[727658, 19717, 6028, 152, 70, "ProofInfo",
   CellTags->{"Proof|ID:844584535", "Proof|ID:844584535-1"},
   CellID->34854128]},
 "ID:632054408"->{
  Cell[734698, 19903, 2199, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:632054408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[SubsetEqual] transitive"},
   CellID->632054408]},
 "red \[SubsetEqual] transitive"->{
  Cell[734698, 19903, 2199, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:632054408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[SubsetEqual] transitive"},
   CellID->632054408]},
 "ID:387695114"->{
  Cell[736900, 19972, 772, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:387695114", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->387695114]},
 "ID:1177850337"->{
  Cell[738190, 20011, 1984, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1177850337", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "setSeq subsequence red"},
   CellID->1177850337]},
 "setSeq subsequence red"->{
  Cell[738190, 20011, 1984, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1177850337", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "setSeq subsequence red"},
   CellID->1177850337]},
 "ID:1483708233"->{
  Cell[740177, 20074, 714, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1483708233", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1483708233]},
 "ID:697433658"->{
  Cell[740894, 20094, 1060, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:697433658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red setSeq \[SubsetEqual]"},
   CellID->697433658]},
 "red setSeq \[SubsetEqual]"->{
  Cell[740894, 20094, 1060, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:697433658", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red setSeq \[SubsetEqual]"},
   CellID->697433658]},
 "ID:35999503"->{
  Cell[741957, 20126, 2089, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:35999503", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red setSeq"},
   CellID->35999503]},
 "\[Element] red setSeq"->{
  Cell[741957, 20126, 2089, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:35999503", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red setSeq"},
   CellID->35999503]},
 "ID:602168712"->{
  Cell[744049, 20191, 2200, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:602168712", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red setSeq \[Subset]"},
   CellID->602168712]},
 "red setSeq \[Subset]"->{
  Cell[744049, 20191, 2200, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:602168712", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red setSeq \[Subset]"},
   CellID->602168712]},
 "Proof|ID:632054408"->{
  Cell[746406, 20269, 6960, 174, 70, "ProofInfo",
   CellTags->{"Proof|ID:632054408", "Proof|ID:632054408-1"},
   CellID->70698580]},
 "Proof|ID:632054408-1"->{
  Cell[746406, 20269, 6960, 174, 70, "ProofInfo",
   CellTags->{"Proof|ID:632054408", "Proof|ID:632054408-1"},
   CellID->70698580]},
 "Proof|ID:1177850337"->{
  Cell[753885, 20462, 7392, 186, 70, "ProofInfo",
   CellTags->{"Proof|ID:1177850337", "Proof|ID:1177850337-1"},
   CellID->1013465541]},
 "Proof|ID:1177850337-1"->{
  Cell[753885, 20462, 7392, 186, 70, "ProofInfo",
   CellTags->{"Proof|ID:1177850337", "Proof|ID:1177850337-1"},
   CellID->1013465541]},
 "Proof|ID:697433658"->{
  Cell[761799, 20667, 4199, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:697433658", "Proof|ID:697433658-1"},
   CellID->179082785]},
 "Proof|ID:697433658-1"->{
  Cell[761799, 20667, 4199, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:697433658", "Proof|ID:697433658-1"},
   CellID->179082785]},
 "Proof|ID:35999503"->{
  Cell[766518, 20798, 6144, 152, 70, "ProofInfo",
   CellTags->{"Proof|ID:35999503", "Proof|ID:35999503-1"},
   CellID->361034880]},
 "Proof|ID:35999503-1"->{
  Cell[766518, 20798, 6144, 152, 70, "ProofInfo",
   CellTags->{"Proof|ID:35999503", "Proof|ID:35999503-1"},
   CellID->361034880]},
 "Proof|ID:602168712"->{
  Cell[773181, 20969, 7273, 182, 70, "ProofInfo",
   CellTags->{"Proof|ID:602168712", "Proof|ID:602168712-1"},
   CellID->110870073]},
 "Proof|ID:602168712-1"->{
  Cell[773181, 20969, 7273, 182, 70, "ProofInfo",
   CellTags->{"Proof|ID:602168712", "Proof|ID:602168712-1"},
   CellID->110870073]},
 "ID:21782685"->{
  Cell[781414, 21184, 2578, 83, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:21782685", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] append", "IR"},
   CellID->21782685]},
 "\[SquareSubset] append"->{
  Cell[781414, 21184, 2578, 83, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:21782685", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] append", "IR"},
   CellID->21782685]},
 "ID:1790989103"->{
  Cell[783995, 21269, 1204, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1790989103", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isPartIrreflOrder"},
   CellID->1790989103]},
 "\[SquareSubset] isPartIrreflOrder"->{
  Cell[783995, 21269, 1204, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1790989103", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isPartIrreflOrder"},
   CellID->1790989103]},
 "Proof|ID:21782685"->{
  Cell[785354, 21313, 5900, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:21782685", "Proof|ID:21782685-1"},
   CellID->202074869]},
 "Proof|ID:21782685-1"->{
  Cell[785354, 21313, 5900, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:21782685", "Proof|ID:21782685-1"},
   CellID->202074869]},
 "Proof|ID:1790989103"->{
  Cell[791773, 21480, 2619, 77, 70, "ProofInfo",
   CellTags->{"Proof|ID:1790989103", "Proof|ID:1790989103-1"},
   CellID->227724]},
 "Proof|ID:1790989103-1"->{
  Cell[791773, 21480, 2619, 77, 70, "ProofInfo",
   CellTags->{"Proof|ID:1790989103", "Proof|ID:1790989103-1"},
   CellID->227724]},
 "ID:1466856847"->{
  Cell[826130, 22593, 653, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1466856847", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1466856847]},
 "ID:623701544"->{
  Cell[827340, 22628, 2804, 87, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:623701544", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction ring"},
   CellID->623701544]},
 "reduction ring"->{
  Cell[827340, 22628, 2804, 87, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:623701544", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction ring"},
   CellID->623701544]},
 "ID:1960646232"->{
  Cell[830147, 22717, 2665, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1960646232", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "multipliers"},
   CellID->1960646232]},
 "multipliers"->{
  Cell[830147, 22717, 2665, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1960646232", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "multipliers"},
   CellID->1960646232]},
 "ID:1496065688"->{
  Cell[832815, 22800, 1460, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1496065688", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R0"},
   CellID->1496065688]},
 "R0"->{
  Cell[832815, 22800, 1460, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1496065688", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R0"},
   CellID->1496065688]},
 "ID:975892043"->{
  Cell[834278, 22847, 1639, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:975892043", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R1"},
   CellID->975892043]},
 "R1"->{
  Cell[834278, 22847, 1639, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:975892043", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R1"},
   CellID->975892043]},
 "ID:1150364990"->{
  Cell[835920, 22900, 1617, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1150364990", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R2"},
   CellID->1150364990]},
 "R2"->{
  Cell[835920, 22900, 1617, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1150364990", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R2"},
   CellID->1150364990]},
 "ID:852396718"->{
  Cell[837540, 22952, 2946, 86, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:852396718", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R3"},
   CellID->852396718]},
 "R3"->{
  Cell[837540, 22952, 2946, 86, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:852396718", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R3"},
   CellID->852396718]},
 "ID:1907869025"->{
  Cell[840489, 23040, 1517, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1907869025", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R4"},
   CellID->1907869025]},
 "R4"->{
  Cell[840489, 23040, 1517, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1907869025", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R4"},
   CellID->1907869025]},
 "ID:1446921308"->{
  Cell[842009, 23087, 2702, 77, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1446921308", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R5"},
   CellID->1446921308]},
 "R5"->{
  Cell[842009, 23087, 2702, 77, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1446921308", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R5"},
   CellID->1446921308]},
 "ID:208630606"->{
  Cell[844714, 23166, 3054, 87, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208630606", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6"},
   CellID->208630606]},
 "R6"->{
  Cell[844714, 23166, 3054, 87, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208630606", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6"},
   CellID->208630606]},
 "ID:1150983847"->{
  Cell[847771, 23255, 10540, 264, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1150983847", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R7"},
   CellID->1150983847]},
 "R7"->{
  Cell[847771, 23255, 10540, 264, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1150983847", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R7"},
   CellID->1150983847]},
 "ID:2106857486"->{
  Cell[858314, 23521, 7036, 180, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2106857486", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R8"},
   CellID->2106857486]},
 "R8"->{
  Cell[858314, 23521, 7036, 180, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2106857486", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R8"},
   CellID->2106857486]},
 "ID:1197875862"->{
  Cell[865971, 23727, 928, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1197875862", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1197875862]},
 "ID:1949920189"->{
  Cell[866902, 23750, 3522, 98, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1949920189", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R5 witnesses"},
   CellID->1949920189]},
 "R5 witnesses"->{
  Cell[866902, 23750, 3522, 98, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1949920189", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R5 witnesses"},
   CellID->1949920189]},
 "ID:2027349511"->{
  Cell[870427, 23850, 2974, 83, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2027349511", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6 witness"},
   CellID->2027349511]},
 "R6 witness"->{
  Cell[870427, 23850, 2974, 83, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2027349511", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6 witness"},
   CellID->2027349511]},
 "ID:1717911774"->{
  Cell[873708, 23949, 693, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1717911774", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1717911774]},
 "ID:349001278"->{
  Cell[875100, 23987, 771, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:349001278", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->349001278]},
 "ID:474542129"->{
  Cell[875874, 24008, 1348, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:474542129", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "9"},
   CellID->474542129]},
 "9"->{
  Cell[875874, 24008, 1348, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:474542129", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "9"},
   CellID->474542129]},
 "ID:360788203"->{
  Cell[877225, 24050, 1468, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:360788203", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "21"},
   CellID->360788203]},
 "21"->{
  Cell[877225, 24050, 1468, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:360788203", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "21"},
   CellID->360788203]},
 "Proof|ID:474542129"->{
  Cell[878851, 24107, 4275, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:474542129", "Proof|ID:474542129-1"},
   CellID->1905252519]},
 "Proof|ID:474542129-1"->{
  Cell[878851, 24107, 4275, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:474542129", "Proof|ID:474542129-1"},
   CellID->1905252519]},
 "Proof|ID:360788203"->{
  Cell[883647, 24240, 4379, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:360788203", "Proof|ID:360788203-1"},
   CellID->3264289]},
 "Proof|ID:360788203-1"->{
  Cell[883647, 24240, 4379, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:360788203", "Proof|ID:360788203-1"},
   CellID->3264289]},
 "ID:998446180"->{
  Cell[888993, 24391, 674, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:998446180", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->998446180]},
 "ID:1581977856"->{
  Cell[889670, 24410, 940, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1581977856", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1581977856]},
 "ID:623334744"->{
  Cell[890613, 24437, 1026, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:623334744", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "8"},
   CellID->623334744]},
 "8"->{
  Cell[890613, 24437, 1026, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:623334744", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "8"},
   CellID->623334744]},
 "ID:1112133758"->{
  Cell[891642, 24470, 1591, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1112133758", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "8'"},
   CellID->1112133758]},
 "8'"->{
  Cell[891642, 24470, 1591, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1112133758", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "8'"},
   CellID->1112133758]},
 "ID:810098043"->{
  Cell[893236, 24522, 727, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:810098043", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->810098043]},
 "ID:1578524611"->{
  Cell[893966, 24542, 1312, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1578524611", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "19"},
   CellID->1578524611]},
 "19"->{
  Cell[893966, 24542, 1312, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1578524611", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "19"},
   CellID->1578524611]},
 "ID:1738175589"->{
  Cell[895281, 24586, 1499, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1738175589", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "19'"},
   CellID->1738175589]},
 "19'"->{
  Cell[895281, 24586, 1499, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1738175589", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "19'"},
   CellID->1738175589]},
 "Proof|ID:623334744"->{
  Cell[896936, 24648, 4706, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:623334744", "Proof|ID:623334744-1"},
   CellID->1449852563]},
 "Proof|ID:623334744-1"->{
  Cell[896936, 24648, 4706, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:623334744", "Proof|ID:623334744-1"},
   CellID->1449852563]},
 "Proof|ID:1112133758"->{
  Cell[902162, 24791, 6203, 157, 70, "ProofInfo",
   CellTags->{"Proof|ID:1112133758", "Proof|ID:1112133758-1"},
   CellID->1792610282]},
 "Proof|ID:1112133758-1"->{
  Cell[902162, 24791, 6203, 157, 70, "ProofInfo",
   CellTags->{"Proof|ID:1112133758", "Proof|ID:1112133758-1"},
   CellID->1792610282]},
 "Proof|ID:1578524611"->{
  Cell[908887, 24967, 5891, 151, 70, "ProofInfo",
   CellTags->{"Proof|ID:1578524611", "Proof|ID:1578524611-1"},
   CellID->32974812]},
 "Proof|ID:1578524611-1"->{
  Cell[908887, 24967, 5891, 151, 70, "ProofInfo",
   CellTags->{"Proof|ID:1578524611", "Proof|ID:1578524611-1"},
   CellID->32974812]},
 "Proof|ID:1738175589"->{
  Cell[915298, 25137, 6476, 163, 70, "ProofInfo",
   CellTags->{"Proof|ID:1738175589", "Proof|ID:1738175589-1"},
   CellID->629724279]},
 "Proof|ID:1738175589-1"->{
  Cell[915298, 25137, 6476, 163, 70, "ProofInfo",
   CellTags->{"Proof|ID:1738175589", "Proof|ID:1738175589-1"},
   CellID->629724279]},
 "ID:258339355"->{
  Cell[922844, 25335, 1110, 34, 70, "GlobalDeclaration",
   CellTags->{
    "ID:258339355", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->258339355]},
 "ID:1933985410"->{
  Cell[923957, 25371, 832, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1933985410", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1933985410]},
 "ID:739393974"->{
  Cell[924792, 25395, 818, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739393974", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "17"},
   CellID->739393974]},
 "17"->{
  Cell[924792, 25395, 818, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739393974", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "17"},
   CellID->739393974]},
 "ID:875264505"->{
  Cell[925613, 25420, 1164, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:875264505", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "10"},
   CellID->875264505]},
 "10"->{
  Cell[925613, 25420, 1164, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:875264505", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "10"},
   CellID->875264505]},
 "ID:2012915447"->{
  Cell[926780, 25457, 1080, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2012915447", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "11"},
   CellID->2012915447]},
 "11"->{
  Cell[926780, 25457, 1080, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2012915447", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "11"},
   CellID->2012915447]},
 "ID:1489336269"->{
  Cell[927863, 25492, 1248, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1489336269", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "12"},
   CellID->1489336269]},
 "12"->{
  Cell[927863, 25492, 1248, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1489336269", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "12"},
   CellID->1489336269]},
 "ID:1722317392"->{
  Cell[929114, 25534, 1483, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1722317392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "13"},
   CellID->1722317392]},
 "13"->{
  Cell[929114, 25534, 1483, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1722317392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "13"},
   CellID->1722317392]},
 "ID:1089845252"->{
  Cell[930600, 25582, 1075, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1089845252", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "15"},
   CellID->1089845252]},
 "15"->{
  Cell[930600, 25582, 1075, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1089845252", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "15"},
   CellID->1089845252]},
 "Proof|ID:739393974"->{
  Cell[931832, 25626, 4296, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:739393974", "Proof|ID:739393974-1"},
   CellID->1175180046]},
 "Proof|ID:739393974-1"->{
  Cell[931832, 25626, 4296, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:739393974", "Proof|ID:739393974-1"},
   CellID->1175180046]},
 "Proof|ID:875264505"->{
  Cell[936649, 25761, 6277, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:875264505", "Proof|ID:875264505-1"},
   CellID->564262226]},
 "Proof|ID:875264505-1"->{
  Cell[936649, 25761, 6277, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:875264505", "Proof|ID:875264505-1"},
   CellID->564262226]},
 "Proof|ID:2012915447"->{
  Cell[943445, 25934, 5335, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:2012915447", "Proof|ID:2012915447-1"},
   CellID->2044326423]},
 "Proof|ID:2012915447-1"->{
  Cell[943445, 25934, 5335, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:2012915447", "Proof|ID:2012915447-1"},
   CellID->2044326423]},
 "Proof|ID:1489336269"->{
  Cell[949301, 26090, 5903, 149, 70, "ProofInfo",
   CellTags->{"Proof|ID:1489336269", "Proof|ID:1489336269-1"},
   CellID->62520461]},
 "Proof|ID:1489336269-1"->{
  Cell[949301, 26090, 5903, 149, 70, "ProofInfo",
   CellTags->{"Proof|ID:1489336269", "Proof|ID:1489336269-1"},
   CellID->62520461]},
 "Proof|ID:1722317392"->{
  Cell[955725, 26258, 7045, 171, 70, "ProofInfo",
   CellTags->{"Proof|ID:1722317392", "Proof|ID:1722317392-1"},
   CellID->1641148702]},
 "Proof|ID:1722317392-1"->{
  Cell[955725, 26258, 7045, 171, 70, "ProofInfo",
   CellTags->{"Proof|ID:1722317392", "Proof|ID:1722317392-1"},
   CellID->1641148702]},
 "Proof|ID:1089845252"->{
  Cell[963292, 26448, 5323, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:1089845252", "Proof|ID:1089845252-1"},
   CellID->96759121]},
 "Proof|ID:1089845252-1"->{
  Cell[963292, 26448, 5323, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:1089845252", "Proof|ID:1089845252-1"},
   CellID->96759121]},
 "ID:2021433946"->{
  Cell[969571, 26618, 1054, 31, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2021433946", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2021433946]},
 "ID:369680194"->{
  Cell[970628, 26651, 737, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:369680194", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->369680194]},
 "ID:491925495"->{
  Cell[971368, 26673, 2292, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:491925495", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] alternative"},
   CellID->491925495]},
 "\[LeftRightArrow] alternative"->{
  Cell[971368, 26673, 2292, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:491925495", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] alternative"},
   CellID->491925495]},
 "ID:1668317482"->{
  Cell[973663, 26744, 1451, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1668317482", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred \[LeftRightArrow]"},
   CellID->1668317482]},
 "symred \[LeftRightArrow]"->{
  Cell[973663, 26744, 1451, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1668317482", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred \[LeftRightArrow]"},
   CellID->1668317482]},
 "Proof|ID:491925495"->{
  Cell[975269, 26799, 8952, 210, 70, "ProofInfo",
   CellTags->{"Proof|ID:491925495", "Proof|ID:491925495-1"},
   CellID->99017069]},
 "Proof|ID:491925495-1"->{
  Cell[975269, 26799, 8952, 210, 70, "ProofInfo",
   CellTags->{"Proof|ID:491925495", "Proof|ID:491925495-1"},
   CellID->99017069]},
 "Proof|ID:1668317482"->{
  Cell[984741, 27028, 6820, 167, 70, "ProofInfo",
   CellTags->{"Proof|ID:1668317482", "Proof|ID:1668317482-1"},
   CellID->304755132]},
 "Proof|ID:1668317482-1"->{
  Cell[984741, 27028, 6820, 167, 70, "ProofInfo",
   CellTags->{"Proof|ID:1668317482", "Proof|ID:1668317482-1"},
   CellID->304755132]},
 "ID:169581633"->{
  Cell[992533, 27228, 887, 27, 70, "GlobalDeclaration",
   CellTags->{
    "ID:169581633", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->169581633]},
 "ID:265555623"->{
  Cell[993423, 27257, 2362, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:265555623", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "403"},
   CellID->265555623]},
 "403"->{
  Cell[993423, 27257, 2362, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:265555623", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "403"},
   CellID->265555623]},
 "Proof|ID:265555623"->{
  Cell[995942, 27336, 9193, 218, 70, "ProofInfo",
   CellTags->{"Proof|ID:265555623", "Proof|ID:265555623-1"},
   CellID->1813792728]},
 "Proof|ID:265555623-1"->{
  Cell[995942, 27336, 9193, 218, 70, "ProofInfo",
   CellTags->{"Proof|ID:265555623", "Proof|ID:265555623-1"},
   CellID->1813792728]},
 "ID:1020054522"->{
  Cell[1006154, 27588, 1007, 30, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1020054522", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1020054522]},
 "ID:270873424"->{
  Cell[1007164, 27620, 776, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:270873424", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->270873424]},
 "ID:1165091979"->{
  Cell[1007943, 27643, 998, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1165091979", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "16"},
   CellID->1165091979]},
 "16"->{
  Cell[1007943, 27643, 998, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1165091979", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "16"},
   CellID->1165091979]},
 "ID:1894133408"->{
  Cell[1008944, 27673, 1144, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1894133408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred symmetry"},
   CellID->1894133408]},
 "symred symmetry"->{
  Cell[1008944, 27673, 1144, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1894133408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred symmetry"},
   CellID->1894133408]},
 "Proof|ID:1165091979"->{
  Cell[1010244, 27715, 6619, 159, 70, "ProofInfo",
   CellTags->{"Proof|ID:1165091979", "Proof|ID:1165091979-1"},
   CellID->1301195110]},
 "Proof|ID:1165091979-1"->{
  Cell[1010244, 27715, 6619, 159, 70, "ProofInfo",
   CellTags->{"Proof|ID:1165091979", "Proof|ID:1165091979-1"},
   CellID->1301195110]},
 "Proof|ID:1894133408"->{
  Cell[1017384, 27893, 6703, 161, 70, "ProofInfo",
   CellTags->{"Proof|ID:1894133408", "Proof|ID:1894133408-1"},
   CellID->1016736240]},
 "Proof|ID:1894133408-1"->{
  Cell[1017384, 27893, 6703, 161, 70, "ProofInfo",
   CellTags->{"Proof|ID:1894133408", "Proof|ID:1894133408-1"},
   CellID->1016736240]},
 "ID:1779789522"->{
  Cell[1027970, 28189, 653, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1779789522", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1779789522]},
 "ID:979757495"->{
  Cell[1028626, 28207, 1660, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:979757495", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "algorithmic reduction ring"},
   CellID->979757495]},
 "algorithmic reduction ring"->{
  Cell[1028626, 28207, 1660, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:979757495", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "algorithmic reduction ring"},
   CellID->979757495]},
 "ID:1502179371"->{
  Cell[1031040, 28284, 5504, 148, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502179371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R9"},
   CellID->1502179371]},
 "R9"->{
  Cell[1031040, 28284, 5504, 148, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502179371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R9"},
   CellID->1502179371]},
 "ID:1524689422"->{
  Cell[1036547, 28434, 4444, 124, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1524689422", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R10"},
   CellID->1524689422]},
 "R10"->{
  Cell[1036547, 28434, 4444, 124, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1524689422", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R10"},
   CellID->1524689422]},
 "ID:528478842"->{
  Cell[1040994, 28560, 2566, 75, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:528478842", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R11"},
   CellID->528478842]},
 "R11"->{
  Cell[1040994, 28560, 2566, 75, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:528478842", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R11"},
   CellID->528478842]},
 "ID:415149625"->{
  Cell[1043563, 28637, 4242, 109, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:415149625", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R12"},
   CellID->415149625]},
 "R12"->{
  Cell[1043563, 28637, 4242, 109, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:415149625", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R12"},
   CellID->415149625]},
 "ID:1276925313"->{
  Cell[1047808, 28748, 1890, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1276925313", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R13"},
   CellID->1276925313]},
 "R13"->{
  Cell[1047808, 28748, 1890, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1276925313", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R13"},
   CellID->1276925313]},
 "ID:1684964128"->{
  Cell[1050128, 28822, 697, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1684964128", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1684964128]},
 "ID:10673911"->{
  Cell[1051428, 28859, 2100, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:10673911", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "22"},
   CellID->10673911]},
 "22"->{
  Cell[1051428, 28859, 2100, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:10673911", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "22"},
   CellID->10673911]},
 "ID:603265816"->{
  Cell[1053531, 28924, 1562, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:603265816", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "23"},
   CellID->603265816]},
 "23"->{
  Cell[1053531, 28924, 1562, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:603265816", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "23"},
   CellID->603265816]},
 "ID:39612209"->{
  Cell[1055096, 28974, 3444, 100, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:39612209", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "24"},
   CellID->39612209]},
 "24"->{
  Cell[1055096, 28974, 3444, 100, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:39612209", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "24"},
   CellID->39612209]},
 "Proof|ID:10673911"->{
  Cell[1058696, 29085, 9211, 218, 70, "ProofInfo",
   CellTags->{"Proof|ID:10673911", "Proof|ID:10673911-1"},
   CellID->26780681]},
 "Proof|ID:10673911-1"->{
  Cell[1058696, 29085, 9211, 218, 70, "ProofInfo",
   CellTags->{"Proof|ID:10673911", "Proof|ID:10673911-1"},
   CellID->26780681]},
 "Proof|ID:603265816"->{
  Cell[1068425, 29322, 5283, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:603265816", "Proof|ID:603265816-1"},
   CellID->237799576]},
 "Proof|ID:603265816-1"->{
  Cell[1068425, 29322, 5283, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:603265816", "Proof|ID:603265816-1"},
   CellID->237799576]},
 "Proof|ID:39612209"->{
  Cell[1074227, 29479, 12808, 292, 70, "ProofInfo",
   CellTags->{"Proof|ID:39612209", "Proof|ID:39612209-1"},
   CellID->269494545]},
 "Proof|ID:39612209-1"->{
  Cell[1074227, 29479, 12808, 292, 70, "ProofInfo",
   CellTags->{"Proof|ID:39612209", "Proof|ID:39612209-1"},
   CellID->269494545]},
 "ID:1515571182"->{
  Cell[1088017, 29805, 680, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1515571182", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1515571182]},
 "ID:528027344"->{
  Cell[1088700, 29824, 1544, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:528027344", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "18"},
   CellID->528027344]},
 "18"->{
  Cell[1088700, 29824, 1544, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:528027344", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "18"},
   CellID->528027344]},
 "ID:94925229"->{
  Cell[1090247, 29877, 2634, 78, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:94925229", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "25"},
   CellID->94925229]},
 "25"->{
  Cell[1090247, 29877, 2634, 78, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:94925229", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "25"},
   CellID->94925229]},
 "ID:2104251234"->{
  Cell[1092884, 29957, 1801, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2104251234", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "18'"},
   CellID->2104251234]},
 "18'"->{
  Cell[1092884, 29957, 1801, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2104251234", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "18'"},
   CellID->2104251234]},
 "Proof|ID:528027344"->{
  Cell[1094840, 30023, 5205, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:528027344", "Proof|ID:528027344-1"},
   CellID->492864454]},
 "Proof|ID:528027344-1"->{
  Cell[1094840, 30023, 5205, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:528027344", "Proof|ID:528027344-1"},
   CellID->492864454]},
 "Proof|ID:94925229"->{
  Cell[1100564, 30178, 7618, 188, 70, "ProofInfo",
   CellTags->{"Proof|ID:94925229", "Proof|ID:94925229-1"},
   CellID->568093765]},
 "Proof|ID:94925229-1"->{
  Cell[1100564, 30178, 7618, 188, 70, "ProofInfo",
   CellTags->{"Proof|ID:94925229", "Proof|ID:94925229-1"},
   CellID->568093765]},
 "Proof|ID:2104251234"->{
  Cell[1108700, 30385, 7789, 182, 70, "ProofInfo",
   CellTags->{"Proof|ID:2104251234", "Proof|ID:2104251234-1"},
   CellID->611079668]},
 "Proof|ID:2104251234-1"->{
  Cell[1108700, 30385, 7789, 182, 70, "ProofInfo",
   CellTags->{"Proof|ID:2104251234", "Proof|ID:2104251234-1"},
   CellID->611079668]},
 "ID:484595803"->{
  Cell[1117556, 30602, 878, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:484595803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isFinDomain \[DoubleRightArrow] R13"},
   CellID->484595803]},
 "isFinDomain \[DoubleRightArrow] R13"->{
  Cell[1117556, 30602, 878, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:484595803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isFinDomain \[DoubleRightArrow] R13"},
   CellID->484595803]},
 "ID:55361704"->{
  Cell[1118437, 30626, 672, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:55361704", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->55361704]},
 "ID:283927726"->{
  Cell[1119112, 30645, 1712, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:283927726", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "20"},
   CellID->283927726]},
 "20"->{
  Cell[1119112, 30645, 1712, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:283927726", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "20"},
   CellID->283927726]},
 "ID:1584214935"->{
  Cell[1121268, 30711, 1043, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1584214935", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isNoetherian sets"},
   CellID->1584214935]},
 "\[SquareSubset] isNoetherian sets"->{
  Cell[1121268, 30711, 1043, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1584214935", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isNoetherian sets"},
   CellID->1584214935]},
 "ID:1231517070"->{
  Cell[1122314, 30743, 1053, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1231517070", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isNoetherian tuples"},
   CellID->1231517070]},
 "\[SquareSubset] isNoetherian tuples"->{
  Cell[1122314, 30743, 1053, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1231517070", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isNoetherian tuples"},
   CellID->1231517070]},
 "Proof|ID:484595803"->{
  Cell[1123525, 30785, 2509, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:484595803", "Proof|ID:484595803-1"},
   CellID->419106276]},
 "Proof|ID:484595803-1"->{
  Cell[1123525, 30785, 2509, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:484595803", "Proof|ID:484595803-1"},
   CellID->419106276]},
 "Proof|ID:283927726"->{
  Cell[1126554, 30878, 5688, 146, 70, "ProofInfo",
   CellTags->{"Proof|ID:283927726", "Proof|ID:283927726-1"},
   CellID->309385559]},
 "Proof|ID:283927726-1"->{
  Cell[1126554, 30878, 5688, 146, 70, "ProofInfo",
   CellTags->{"Proof|ID:283927726", "Proof|ID:283927726-1"},
   CellID->309385559]},
 "Proof|ID:1584214935"->{
  Cell[1132762, 31043, 2893, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1584214935", "Proof|ID:1584214935-1"},
   CellID->2080577831]},
 "Proof|ID:1584214935-1"->{
  Cell[1132762, 31043, 2893, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1584214935", "Proof|ID:1584214935-1"},
   CellID->2080577831]},
 "Proof|ID:1231517070"->{
  Cell[1136176, 31145, 2900, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1231517070", "Proof|ID:1231517070-1"},
   CellID->120554530]},
 "Proof|ID:1231517070-1"->{
  Cell[1136176, 31145, 2900, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1231517070", "Proof|ID:1231517070-1"},
   CellID->120554530]},
 "ID:1326788334"->{
  Cell[1140591, 31272, 740, 18, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1326788334", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1326788334]},
 "ID:227700447"->{
  Cell[1142367, 31328, 843, 22, 44, "GlobalDeclaration",
   CellTags->{
    "ID:227700447", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->227700447]},
 "ID:1128391937"->{
  Cell[1143213, 31352, 828, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1128391937", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1128391937]},
 "ID:788600821"->{
  Cell[1144044, 31376, 1041, 33, 33, "GlobalDeclaration",
   CellTags->{
    "ID:788600821", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->788600821]},
 "ID:999322294"->{
  Cell[1145088, 31411, 845, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:999322294", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->999322294]},
 "ID:737948621"->{
  Cell[1145936, 31435, 790, 22, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:737948621", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "392"},
   CellID->737948621]},
 "392"->{
  Cell[1145936, 31435, 790, 22, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:737948621", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "392"},
   CellID->737948621]},
 "ID:287016951"->{
  Cell[1146729, 31459, 1317, 37, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:287016951", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "394"},
   CellID->287016951]},
 "394"->{
  Cell[1146729, 31459, 1317, 37, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:287016951", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "394"},
   CellID->287016951]},
 "ID:1812129266"->{
  Cell[1148049, 31498, 1343, 39, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:1812129266", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "395"},
   CellID->1812129266]},
 "395"->{
  Cell[1148049, 31498, 1343, 39, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:1812129266", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "395"},
   CellID->1812129266]},
 "ID:1191851294"->{
  Cell[1149395, 31539, 3200, 90, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:1191851294", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "396"},
   CellID->1191851294]},
 "396"->{
  Cell[1149395, 31539, 3200, 90, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:1191851294", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "396"},
   CellID->1191851294]},
 "ID:101106535"->{
  Cell[1152598, 31631, 2723, 79, 117, "FormalTextInputFormula",
   CellTags->{
    "ID:101106535", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "397"},
   CellID->101106535]},
 "397"->{
  Cell[1152598, 31631, 2723, 79, 117, "FormalTextInputFormula",
   CellTags->{
    "ID:101106535", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "397"},
   CellID->101106535]},
 "ID:1737800130"->{
  Cell[1155324, 31712, 2532, 72, 61, "FormalTextInputFormula",
   CellTags->{
    "ID:1737800130", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "398"},
   CellID->1737800130]},
 "398"->{
  Cell[1155324, 31712, 2532, 72, 61, "FormalTextInputFormula",
   CellTags->{
    "ID:1737800130", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "398"},
   CellID->1737800130]},
 "ID:1568772314"->{
  Cell[1157859, 31786, 2281, 67, 60, "FormalTextInputFormula",
   CellTags->{
    "ID:1568772314", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "399"},
   CellID->1568772314]},
 "399"->{
  Cell[1157859, 31786, 2281, 67, 60, "FormalTextInputFormula",
   CellTags->{
    "ID:1568772314", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "399"},
   CellID->1568772314]},
 "Proof|ID:737948621"->{
  Cell[1160297, 31864, 4403, 114, 28, "ProofInfo",
   CellTags->{"Proof|ID:737948621", "Proof|ID:737948621-1"},
   CellID->401948380]},
 "Proof|ID:737948621-1"->{
  Cell[1160297, 31864, 4403, 114, 28, "ProofInfo",
   CellTags->{"Proof|ID:737948621", "Proof|ID:737948621-1"},
   CellID->401948380]},
 "Proof|ID:287016951"->{
  Cell[1165221, 31997, 6075, 150, 26, "ProofInfo",
   CellTags->{"Proof|ID:287016951", "Proof|ID:287016951-1"},
   CellID->766901523]},
 "Proof|ID:287016951-1"->{
  Cell[1165221, 31997, 6075, 150, 26, "ProofInfo",
   CellTags->{"Proof|ID:287016951", "Proof|ID:287016951-1"},
   CellID->766901523]},
 "Proof|ID:1812129266"->{
  Cell[1171815, 32166, 7370, 177, 26, "ProofInfo",
   CellTags->{"Proof|ID:1812129266", "Proof|ID:1812129266-1"},
   CellID->967901432]},
 "Proof|ID:1812129266-1"->{
  Cell[1171815, 32166, 7370, 177, 26, "ProofInfo",
   CellTags->{"Proof|ID:1812129266", "Proof|ID:1812129266-1"},
   CellID->967901432]},
 "Proof|ID:1191851294"->{
  Cell[1179705, 32362, 16836, 373, 26, "ProofInfo",
   CellTags->{"Proof|ID:1191851294", "Proof|ID:1191851294-1"},
   CellID->155176372]},
 "Proof|ID:1191851294-1"->{
  Cell[1179705, 32362, 16836, 373, 26, "ProofInfo",
   CellTags->{"Proof|ID:1191851294", "Proof|ID:1191851294-1"},
   CellID->155176372]},
 "Proof|ID:101106535"->{
  Cell[1197104, 32755, 14766, 328, 26, "ProofInfo",
   CellTags->{"Proof|ID:101106535", "Proof|ID:101106535-1"},
   CellID->1074757864]},
 "Proof|ID:101106535-1"->{
  Cell[1197104, 32755, 14766, 328, 26, "ProofInfo",
   CellTags->{"Proof|ID:101106535", "Proof|ID:101106535-1"},
   CellID->1074757864]},
 "Proof|ID:1737800130"->{
  Cell[1212433, 33103, 11792, 273, 26, "ProofInfo",
   CellTags->{"Proof|ID:1737800130", "Proof|ID:1737800130-1"},
   CellID->452671936]},
 "Proof|ID:1737800130-1"->{
  Cell[1212433, 33103, 11792, 273, 26, "ProofInfo",
   CellTags->{"Proof|ID:1737800130", "Proof|ID:1737800130-1"},
   CellID->452671936]},
 "Proof|ID:1568772314"->{
  Cell[1224747, 33395, 10791, 253, 26, "ProofInfo",
   CellTags->{"Proof|ID:1568772314", "Proof|ID:1568772314-1"},
   CellID->1787487969]},
 "Proof|ID:1568772314-1"->{
  Cell[1224747, 33395, 10791, 253, 26, "ProofInfo",
   CellTags->{"Proof|ID:1568772314", "Proof|ID:1568772314-1"},
   CellID->1787487969]},
 "ID:1597110548"->{
  Cell[1237354, 33708, 870, 22, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1597110548", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1597110548]},
 "ID:1509853174"->{
  Cell[1238227, 33732, 840, 21, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1509853174", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1509853174]},
 "ID:1554200522"->{
  Cell[1239070, 33755, 1700, 43, 56, "GlobalDeclaration",
   CellTags->{
    "ID:1554200522", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1554200522]},
 "ID:1963951655"->{
  Cell[1240773, 33800, 1344, 39, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:1963951655", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "400"},
   CellID->1963951655]},
 "400"->{
  Cell[1240773, 33800, 1344, 39, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:1963951655", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "400"},
   CellID->1963951655]},
 "ID:85295514"->{
  Cell[1242120, 33841, 1002, 30, 40, "GlobalDeclaration",
   CellTags->{
    "ID:85295514", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->85295514]},
 "ID:751169803"->{
  Cell[1243125, 33873, 2157, 63, 84, "FormalTextInputFormula",
   CellTags->{
    "ID:751169803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "401"},
   CellID->751169803]},
 "401"->{
  Cell[1243125, 33873, 2157, 63, 84, "FormalTextInputFormula",
   CellTags->{
    "ID:751169803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "401"},
   CellID->751169803]},
 "ID:46875747"->{
  Cell[1245285, 33938, 2025, 61, 60, "FormalTextInputFormula",
   CellTags->{
    "ID:46875747", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "402'"},
   CellID->46875747]},
 "402'"->{
  Cell[1245285, 33938, 2025, 61, 60, "FormalTextInputFormula",
   CellTags->{
    "ID:46875747", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "402'"},
   CellID->46875747]},
 "ID:656927872"->{
  Cell[1247313, 34001, 2201, 64, 60, "FormalTextInputFormula",
   CellTags->{
    "ID:656927872", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "402"},
   CellID->656927872]},
 "402"->{
  Cell[1247313, 34001, 2201, 64, 60, "FormalTextInputFormula",
   CellTags->{
    "ID:656927872", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "402"},
   CellID->656927872]},
 "Proof|ID:1963951655"->{
  Cell[1249670, 34076, 10968, 259, 28, "ProofInfo",
   CellTags->{"Proof|ID:1963951655", "Proof|ID:1963951655-1"},
   CellID->208873331]},
 "Proof|ID:1963951655-1"->{
  Cell[1249670, 34076, 10968, 259, 28, "ProofInfo",
   CellTags->{"Proof|ID:1963951655", "Proof|ID:1963951655-1"},
   CellID->208873331]},
 "Proof|ID:751169803"->{
  Cell[1261158, 34354, 21497, 480, 26, "ProofInfo",
   CellTags->{"Proof|ID:751169803", "Proof|ID:751169803-1"},
   CellID->553193299]},
 "Proof|ID:751169803-1"->{
  Cell[1261158, 34354, 21497, 480, 26, "ProofInfo",
   CellTags->{"Proof|ID:751169803", "Proof|ID:751169803-1"},
   CellID->553193299]},
 "Proof|ID:46875747"->{
  Cell[1283175, 34853, 14469, 331, 26, "ProofInfo",
   CellTags->{"Proof|ID:46875747", "Proof|ID:46875747-1"},
   CellID->588348976]},
 "Proof|ID:46875747-1"->{
  Cell[1283175, 34853, 14469, 331, 26, "ProofInfo",
   CellTags->{"Proof|ID:46875747", "Proof|ID:46875747-1"},
   CellID->588348976]},
 "Proof|ID:656927872"->{
  Cell[1298161, 35203, 14964, 338, 26, "ProofInfo",
   CellTags->{"Proof|ID:656927872", "Proof|ID:656927872-1"},
   CellID->574764629]},
 "Proof|ID:656927872-1"->{
  Cell[1298161, 35203, 14964, 338, 26, "ProofInfo",
   CellTags->{"Proof|ID:656927872", "Proof|ID:656927872-1"},
   CellID->574764629]},
 "ID:502021608"->{
  Cell[1314406, 35584, 794, 21, 44, "GlobalDeclaration",
   CellTags->{
    "ID:502021608", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->502021608]},
 "ID:227587436"->{
  Cell[1315203, 35607, 924, 23, 41, "GlobalDeclaration",
   CellTags->{
    "ID:227587436", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->227587436]},
 "ID:937292216"->{
  Cell[1316130, 35632, 1206, 31, 34, "GlobalDeclaration",
   CellTags->{
    "ID:937292216", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->937292216]},
 "ID:1763687895"->{
  Cell[1317339, 35665, 1553, 45, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1763687895", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "410"},
   CellID->1763687895]},
 "410"->{
  Cell[1317339, 35665, 1553, 45, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:1763687895", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "410"},
   CellID->1763687895]},
 "ID:397073436"->{
  Cell[1318895, 35712, 1164, 32, 56, "GlobalDeclaration",
   CellTags->{
    "ID:397073436", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->397073436]},
 "ID:791651685"->{
  Cell[1320062, 35746, 1593, 45, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:791651685", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "411"},
   CellID->791651685]},
 "411"->{
  Cell[1320062, 35746, 1593, 45, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:791651685", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "411"},
   CellID->791651685]},
 "ID:6804200"->{
  Cell[1321658, 35793, 1130, 38, 44, "GlobalDeclaration",
   CellTags->{
    "ID:6804200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->6804200]},
 "ID:1294525434"->{
  Cell[1322791, 35833, 1069, 31, 58, "FormalTextInputFormula",
   CellTags->{
    "ID:1294525434", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "414"},
   CellID->1294525434]},
 "414"->{
  Cell[1322791, 35833, 1069, 31, 58, "FormalTextInputFormula",
   CellTags->{
    "ID:1294525434", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "414"},
   CellID->1294525434]},
 "ID:372638659"->{
  Cell[1323863, 35866, 1005, 30, 40, "GlobalDeclaration",
   CellTags->{
    "ID:372638659", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->372638659]},
 "ID:1962716191"->{
  Cell[1324871, 35898, 1824, 54, 71, "FormalTextInputFormula",
   CellTags->{
    "ID:1962716191", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "412"},
   CellID->1962716191]},
 "412"->{
  Cell[1324871, 35898, 1824, 54, 71, "FormalTextInputFormula",
   CellTags->{
    "ID:1962716191", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "412"},
   CellID->1962716191]},
 "ID:1575262565"->{
  Cell[1326698, 35954, 2316, 71, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:1575262565", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "413"},
   CellID->1575262565]},
 "413"->{
  Cell[1326698, 35954, 2316, 71, 40, "FormalTextInputFormula",
   CellTags->{
    "ID:1575262565", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "413"},
   CellID->1575262565]},
 "Proof|ID:1763687895"->{
  Cell[1329170, 36036, 11350, 259, 28, "ProofInfo",
   CellTags->{"Proof|ID:1763687895", "Proof|ID:1763687895-1"},
   CellID->701556455]},
 "Proof|ID:1763687895-1"->{
  Cell[1329170, 36036, 11350, 259, 28, "ProofInfo",
   CellTags->{"Proof|ID:1763687895", "Proof|ID:1763687895-1"},
   CellID->701556455]},
 "Proof|ID:791651685"->{
  Cell[1341042, 36314, 15630, 356, 26, "ProofInfo",
   CellTags->{"Proof|ID:791651685", "Proof|ID:791651685-1"},
   CellID->2105672847]},
 "Proof|ID:791651685-1"->{
  Cell[1341042, 36314, 15630, 356, 26, "ProofInfo",
   CellTags->{"Proof|ID:791651685", "Proof|ID:791651685-1"},
   CellID->2105672847]},
 "Proof|ID:1294525434"->{
  Cell[1357193, 36689, 14609, 337, 26, "ProofInfo",
   CellTags->{"Proof|ID:1294525434", "Proof|ID:1294525434-1"},
   CellID->727439100]},
 "Proof|ID:1294525434-1"->{
  Cell[1357193, 36689, 14609, 337, 26, "ProofInfo",
   CellTags->{"Proof|ID:1294525434", "Proof|ID:1294525434-1"},
   CellID->727439100]},
 "Proof|ID:1962716191"->{
  Cell[1372323, 37045, 25916, 576, 26, "ProofInfo",
   CellTags->{"Proof|ID:1962716191", "Proof|ID:1962716191-1"},
   CellID->1885017557]},
 "Proof|ID:1962716191-1"->{
  Cell[1372323, 37045, 25916, 576, 26, "ProofInfo",
   CellTags->{"Proof|ID:1962716191", "Proof|ID:1962716191-1"},
   CellID->1885017557]},
 "Proof|ID:1575262565"->{
  Cell[1398760, 37640, 21984, 489, 26, "ProofInfo",
   CellTags->{"Proof|ID:1575262565", "Proof|ID:1575262565-1"},
   CellID->1705894400]},
 "Proof|ID:1575262565-1"->{
  Cell[1398760, 37640, 21984, 489, 26, "ProofInfo",
   CellTags->{"Proof|ID:1575262565", "Proof|ID:1575262565-1"},
   CellID->1705894400]},
 "ID:1547867165"->{
  Cell[1446016, 38899, 1155, 35, 40, "GlobalDeclaration",
   CellTags->{
    "ID:1547867165", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1547867165]},
 "ID:2009958353"->{
  Cell[1447174, 38936, 1005, 24, 41, "GlobalDeclaration",
   CellTags->{
    "ID:2009958353", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2009958353]},
 "ID:10165835"->{
  Cell[1448182, 38962, 1405, 43, 43, "GlobalDeclaration",
   CellTags->{
    "ID:10165835", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->10165835]},
 "ID:1879224373"->{
  Cell[1449590, 39007, 1564, 43, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:1879224373", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6Witness sufficient"},
   CellID->1879224373]},
 "R6Witness sufficient"->{
  Cell[1449590, 39007, 1564, 43, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:1879224373", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6Witness sufficient"},
   CellID->1879224373]},
 "ID:381733398"->{
  Cell[1451157, 39052, 1847, 55, 112, "FormalTextInputFormula",
   CellTags->{
    "ID:381733398", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6Witness symmetry"},
   CellID->381733398]},
 "R6Witness symmetry"->{
  Cell[1451157, 39052, 1847, 55, 112, "FormalTextInputFormula",
   CellTags->{
    "ID:381733398", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6Witness symmetry"},
   CellID->381733398]},
 "Proof|ID:1879224373"->{
  Cell[1453161, 39118, 10441, 237, 28, "ProofInfo",
   CellTags->{"Proof|ID:1879224373", "Proof|ID:1879224373-1"},
   CellID->453457936]},
 "Proof|ID:1879224373-1"->{
  Cell[1453161, 39118, 10441, 237, 28, "ProofInfo",
   CellTags->{"Proof|ID:1879224373", "Proof|ID:1879224373-1"},
   CellID->453457936]},
 "Proof|ID:381733398"->{
  Cell[1464122, 39374, 11696, 260, 26, "ProofInfo",
   CellTags->{"Proof|ID:381733398", "Proof|ID:381733398-1"},
   CellID->180731690]},
 "Proof|ID:381733398-1"->{
  Cell[1464122, 39374, 11696, 260, 26, "ProofInfo",
   CellTags->{"Proof|ID:381733398", "Proof|ID:381733398-1"},
   CellID->180731690]},
 "ID:404204643"->{
  Cell[1492278, 40124, 692, 17, 33, "GlobalDeclaration",
   CellTags->{
    "ID:404204643", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->404204643]},
 "ID:1399661856"->{
  Cell[1493853, 40171, 1019, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1399661856", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1399661856]},
 "ID:1116556350"->{
  Cell[1494875, 40199, 1816, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1116556350", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde] multiplier symmetric"},
   CellID->1116556350]},
 "\[Tilde] multiplier symmetric"->{
  Cell[1494875, 40199, 1816, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1116556350", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde] multiplier symmetric"},
   CellID->1116556350]},
 "ID:1121162524"->{
  Cell[1496694, 40260, 1335, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1121162524", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "mult \[Tilde]"},
   CellID->1121162524]},
 "mult \[Tilde]"->{
  Cell[1496694, 40260, 1335, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1121162524", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "mult \[Tilde]"},
   CellID->1121162524]},
 "ID:1098905803"->{
  Cell[1498032, 40304, 1411, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1098905803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible \[Tilde]"},
   CellID->1098905803]},
 "isReducible \[Tilde]"->{
  Cell[1498032, 40304, 1411, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1098905803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible \[Tilde]"},
   CellID->1098905803]},
 "ID:88395859"->{
  Cell[1499446, 40348, 1862, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:88395859", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areIrrelative \[Tilde]"},
   CellID->88395859]},
 "areIrrelative \[Tilde]"->{
  Cell[1499446, 40348, 1862, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:88395859", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areIrrelative \[Tilde]"},
   CellID->88395859]},
 "ID:1322820809"->{
  Cell[1501311, 40403, 2056, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1322820809", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers \[Tilde]"},
   CellID->1322820809]},
 "areCPMultipliers \[Tilde]"->{
  Cell[1501311, 40403, 2056, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1322820809", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers \[Tilde]"},
   CellID->1322820809]},
 "ID:931721333"->{
  Cell[1503370, 40460, 753, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:931721333", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->931721333]},
 "ID:2090837112"->{
  Cell[1504126, 40482, 1741, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2090837112", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] multiplier \[Tilde]"},
   CellID->2090837112]},
 "\[RightArrow] multiplier \[Tilde]"->{
  Cell[1504126, 40482, 1741, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2090837112", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] multiplier \[Tilde]"},
   CellID->2090837112]},
 "ID:780050836"->{
  Cell[1505870, 40540, 1681, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:780050836", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[Tilde]"},
   CellID->780050836]},
 "\[RightArrow] \[Tilde]"->{
  Cell[1505870, 40540, 1681, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:780050836", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[Tilde]"},
   CellID->780050836]},
 "ID:128058836"->{
  Cell[1507554, 40596, 761, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:128058836", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->128058836]},
 "ID:1516437288"->{
  Cell[1508318, 40617, 1379, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1516437288", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] reductor \[Tilde]"},
   CellID->1516437288]},
 "\[RightArrow] reductor \[Tilde]"->{
  Cell[1508318, 40617, 1379, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1516437288", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] reductor \[Tilde]"},
   CellID->1516437288]},
 "ID:1864007722"->{
  Cell[1509700, 40661, 1334, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1864007722", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] reductor \[Tilde]"},
   CellID->1864007722]},
 "\[LeftRightArrow] reductor \[Tilde]"->{
  Cell[1509700, 40661, 1334, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1864007722", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] reductor \[Tilde]"},
   CellID->1864007722]},
 "ID:642398934"->{
  Cell[1511037, 40704, 1850, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:642398934", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below \[Tilde]"},
   CellID->642398934]},
 "connectibility below \[Tilde]"->{
  Cell[1511037, 40704, 1850, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:642398934", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below \[Tilde]"},
   CellID->642398934]},
 "Proof|ID:1116556350"->{
  Cell[1513042, 40775, 6829, 171, 70, "ProofInfo",
   CellTags->{"Proof|ID:1116556350", "Proof|ID:1116556350-1"},
   CellID->1446312685]},
 "Proof|ID:1116556350-1"->{
  Cell[1513042, 40775, 6829, 171, 70, "ProofInfo",
   CellTags->{"Proof|ID:1116556350", "Proof|ID:1116556350-1"},
   CellID->1446312685]},
 "Proof|ID:1121162524"->{
  Cell[1520392, 40965, 6459, 161, 70, "ProofInfo",
   CellTags->{"Proof|ID:1121162524", "Proof|ID:1121162524-1"},
   CellID->1284068679]},
 "Proof|ID:1121162524-1"->{
  Cell[1520392, 40965, 6459, 161, 70, "ProofInfo",
   CellTags->{"Proof|ID:1121162524", "Proof|ID:1121162524-1"},
   CellID->1284068679]},
 "Proof|ID:1098905803"->{
  Cell[1527373, 41145, 7085, 173, 70, "ProofInfo",
   CellTags->{"Proof|ID:1098905803", "Proof|ID:1098905803-1"},
   CellID->1798497905]},
 "Proof|ID:1098905803-1"->{
  Cell[1527373, 41145, 7085, 173, 70, "ProofInfo",
   CellTags->{"Proof|ID:1098905803", "Proof|ID:1098905803-1"},
   CellID->1798497905]},
 "Proof|ID:88395859"->{
  Cell[1534980, 41337, 9976, 232, 70, "ProofInfo",
   CellTags->{"Proof|ID:88395859", "Proof|ID:88395859-1"},
   CellID->514698520]},
 "Proof|ID:88395859-1"->{
  Cell[1534980, 41337, 9976, 232, 70, "ProofInfo",
   CellTags->{"Proof|ID:88395859", "Proof|ID:88395859-1"},
   CellID->514698520]},
 "Proof|ID:1322820809"->{
  Cell[1545475, 41588, 9856, 229, 70, "ProofInfo",
   CellTags->{"Proof|ID:1322820809", "Proof|ID:1322820809-1"},
   CellID->330288797]},
 "Proof|ID:1322820809-1"->{
  Cell[1545475, 41588, 9856, 229, 70, "ProofInfo",
   CellTags->{"Proof|ID:1322820809", "Proof|ID:1322820809-1"},
   CellID->330288797]},
 "Proof|ID:2090837112"->{
  Cell[1555897, 41837, 8126, 201, 70, "ProofInfo",
   CellTags->{"Proof|ID:2090837112", "Proof|ID:2090837112-1"},
   CellID->1967137926]},
 "Proof|ID:2090837112-1"->{
  Cell[1555897, 41837, 8126, 201, 70, "ProofInfo",
   CellTags->{"Proof|ID:2090837112", "Proof|ID:2090837112-1"},
   CellID->1967137926]},
 "Proof|ID:780050836"->{
  Cell[1564545, 42057, 7124, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:780050836", "Proof|ID:780050836-1"},
   CellID->122789986]},
 "Proof|ID:780050836-1"->{
  Cell[1564545, 42057, 7124, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:780050836", "Proof|ID:780050836-1"},
   CellID->122789986]},
 "Proof|ID:1516437288"->{
  Cell[1572189, 42254, 7002, 176, 70, "ProofInfo",
   CellTags->{"Proof|ID:1516437288", "Proof|ID:1516437288-1"},
   CellID->23267736]},
 "Proof|ID:1516437288-1"->{
  Cell[1572189, 42254, 7002, 176, 70, "ProofInfo",
   CellTags->{"Proof|ID:1516437288", "Proof|ID:1516437288-1"},
   CellID->23267736]},
 "Proof|ID:1864007722"->{
  Cell[1579712, 42449, 7060, 177, 70, "ProofInfo",
   CellTags->{"Proof|ID:1864007722", "Proof|ID:1864007722-1"},
   CellID->1742409570]},
 "Proof|ID:1864007722-1"->{
  Cell[1579712, 42449, 7060, 177, 70, "ProofInfo",
   CellTags->{"Proof|ID:1864007722", "Proof|ID:1864007722-1"},
   CellID->1742409570]},
 "Proof|ID:642398934"->{
  Cell[1587294, 42645, 7871, 190, 70, "ProofInfo",
   CellTags->{"Proof|ID:642398934", "Proof|ID:642398934-1"},
   CellID->1849048841]},
 "Proof|ID:642398934-1"->{
  Cell[1587294, 42645, 7871, 190, 70, "ProofInfo",
   CellTags->{"Proof|ID:642398934", "Proof|ID:642398934-1"},
   CellID->1849048841]},
 "ID:1200903488"->{
  Cell[1596141, 42870, 765, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1200903488", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1200903488]},
 "ID:353808388"->{
  Cell[1596909, 42892, 810, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:353808388", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[Tilde]"},
   CellID->353808388]},
 "reflexivity of \[Tilde]"->{
  Cell[1596909, 42892, 810, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:353808388", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[Tilde]"},
   CellID->353808388]},
 "ID:274091018"->{
  Cell[1597722, 42915, 1079, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:274091018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of \[Tilde]"},
   CellID->274091018]},
 "symmetry of \[Tilde]"->{
  Cell[1597722, 42915, 1079, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:274091018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of \[Tilde]"},
   CellID->274091018]},
 "ID:1760571189"->{
  Cell[1598804, 42951, 1391, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1760571189", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[Tilde]"},
   CellID->1760571189]},
 "transitivity of \[Tilde]"->{
  Cell[1598804, 42951, 1391, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1760571189", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[Tilde]"},
   CellID->1760571189]},
 "Proof|ID:353808388"->{
  Cell[1600353, 43009, 2828, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:353808388", "Proof|ID:353808388-1"},
   CellID->844255967]},
 "Proof|ID:353808388-1"->{
  Cell[1600353, 43009, 2828, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:353808388", "Proof|ID:353808388-1"},
   CellID->844255967]},
 "Proof|ID:274091018"->{
  Cell[1603702, 43110, 3728, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:274091018", "Proof|ID:274091018-1"},
   CellID->772612304]},
 "Proof|ID:274091018-1"->{
  Cell[1603702, 43110, 3728, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:274091018", "Proof|ID:274091018-1"},
   CellID->772612304]},
 "Proof|ID:1760571189"->{
  Cell[1607951, 43231, 4664, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:1760571189", "Proof|ID:1760571189-1"},
   CellID->1528186019]},
 "Proof|ID:1760571189-1"->{
  Cell[1607951, 43231, 4664, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:1760571189", "Proof|ID:1760571189-1"},
   CellID->1528186019]},
 "ID:871669950"->{
  Cell[1613881, 43398, 973, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:871669950", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->871669950]},
 "ID:163646289"->{
  Cell[1614857, 43425, 1720, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:163646289", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative-irrelative 1"},
   CellID->163646289]},
 "correlative-irrelative 1"->{
  Cell[1614857, 43425, 1720, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:163646289", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative-irrelative 1"},
   CellID->163646289]},
 "Proof|ID:163646289"->{
  Cell[1616734, 43488, 8562, 202, 70, "ProofInfo",
   CellTags->{"Proof|ID:163646289", "Proof|ID:163646289-1"},
   CellID->1996269430]},
 "Proof|ID:163646289-1"->{
  Cell[1616734, 43488, 8562, 202, 70, "ProofInfo",
   CellTags->{"Proof|ID:163646289", "Proof|ID:163646289-1"},
   CellID->1996269430]},
 "ID:1990325407"->{
  Cell[1626552, 43734, 813, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1990325407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1990325407]},
 "ID:615171230"->{
  Cell[1627368, 43757, 765, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:615171230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->615171230]},
 "ID:1548501955"->{
  Cell[1628136, 43779, 1506, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1548501955", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "47"},
   CellID->1548501955]},
 "47"->{
  Cell[1628136, 43779, 1506, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1548501955", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "47"},
   CellID->1548501955]},
 "ID:170197122"->{
  Cell[1629645, 43828, 1423, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:170197122", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "48"},
   CellID->170197122]},
 "48"->{
  Cell[1629645, 43828, 1423, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:170197122", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "48"},
   CellID->170197122]},
 "ID:772910972"->{
  Cell[1631071, 43876, 1947, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:772910972", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "49"},
   CellID->772910972]},
 "49"->{
  Cell[1631071, 43876, 1947, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:772910972", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "49"},
   CellID->772910972]},
 "Proof|ID:1548501955"->{
  Cell[1633176, 43952, 5297, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1548501955", "Proof|ID:1548501955-1"},
   CellID->657095506]},
 "Proof|ID:1548501955-1"->{
  Cell[1633176, 43952, 5297, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1548501955", "Proof|ID:1548501955-1"},
   CellID->657095506]},
 "Proof|ID:170197122"->{
  Cell[1638993, 44110, 5206, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:170197122", "Proof|ID:170197122-1"},
   CellID->175765073]},
 "Proof|ID:170197122-1"->{
  Cell[1638993, 44110, 5206, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:170197122", "Proof|ID:170197122-1"},
   CellID->175765073]},
 "Proof|ID:772910972"->{
  Cell[1644718, 44261, 6121, 152, 70, "ProofInfo",
   CellTags->{"Proof|ID:772910972", "Proof|ID:772910972-1"},
   CellID->25309177]},
 "Proof|ID:772910972-1"->{
  Cell[1644718, 44261, 6121, 152, 70, "ProofInfo",
   CellTags->{"Proof|ID:772910972", "Proof|ID:772910972-1"},
   CellID->25309177]},
 "ID:1034579573"->{
  Cell[1652515, 44472, 1026, 27, 42, "GlobalDeclaration",
   CellTags->{
    "ID:1034579573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1034579573]},
 "ID:226093298"->{
  Cell[1653544, 44501, 3958, 114, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:226093298", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "30.\[DoubleRightArrow]"},
   CellID->226093298]},
 "30.\[DoubleRightArrow]"->{
  Cell[1653544, 44501, 3958, 114, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:226093298", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "30.\[DoubleRightArrow]"},
   CellID->226093298]},
 "ID:1773205884"->{
  Cell[1657505, 44617, 3927, 117, 156, "FormalTextInputFormula",
   CellTags->{
    "ID:1773205884", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "31.\[DoubleRightArrow]"},
   CellID->1773205884]},
 "31.\[DoubleRightArrow]"->{
  Cell[1657505, 44617, 3927, 117, 156, "FormalTextInputFormula",
   CellTags->{
    "ID:1773205884", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "31.\[DoubleRightArrow]"},
   CellID->1773205884]},
 "ID:1783130815"->{
  Cell[1661435, 44736, 3022, 91, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:1783130815", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "30.\[DoubleLeftArrow]"},
   CellID->1783130815]},
 "30.\[DoubleLeftArrow]"->{
  Cell[1661435, 44736, 3022, 91, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:1783130815", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "30.\[DoubleLeftArrow]"},
   CellID->1783130815]},
 "ID:1493153487"->{
  Cell[1664460, 44829, 2860, 88, 117, "FormalTextInputFormula",
   CellTags->{
    "ID:1493153487", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "31.\[DoubleLeftArrow]"},
   CellID->1493153487]},
 "31.\[DoubleLeftArrow]"->{
  Cell[1664460, 44829, 2860, 88, 117, "FormalTextInputFormula",
   CellTags->{
    "ID:1493153487", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "31.\[DoubleLeftArrow]"},
   CellID->1493153487]},
 "ID:1798779961"->{
  Cell[1667323, 44919, 1979, 67, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1798779961", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "39.\[DoubleRightArrow]"},
   CellID->1798779961]},
 "39.\[DoubleRightArrow]"->{
  Cell[1667323, 44919, 1979, 67, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1798779961", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "39.\[DoubleRightArrow]"},
   CellID->1798779961]},
 "Proof|ID:226093298"->{
  Cell[1669458, 44997, 10820, 254, 28, "ProofInfo",
   CellTags->{"Proof|ID:226093298", "Proof|ID:226093298-1"},
   CellID->2111336883]},
 "Proof|ID:226093298-1"->{
  Cell[1669458, 44997, 10820, 254, 28, "ProofInfo",
   CellTags->{"Proof|ID:226093298", "Proof|ID:226093298-1"},
   CellID->2111336883]},
 "Proof|ID:1773205884"->{
  Cell[1680841, 45271, 11684, 273, 26, "ProofInfo",
   CellTags->{"Proof|ID:1773205884", "Proof|ID:1773205884-1"},
   CellID->468420595]},
 "Proof|ID:1773205884-1"->{
  Cell[1680841, 45271, 11684, 273, 26, "ProofInfo",
   CellTags->{"Proof|ID:1773205884", "Proof|ID:1773205884-1"},
   CellID->468420595]},
 "Proof|ID:1783130815"->{
  Cell[1693046, 45563, 9162, 221, 26, "ProofInfo",
   CellTags->{"Proof|ID:1783130815", "Proof|ID:1783130815-1"},
   CellID->1694100803]},
 "Proof|ID:1783130815-1"->{
  Cell[1693046, 45563, 9162, 221, 26, "ProofInfo",
   CellTags->{"Proof|ID:1783130815", "Proof|ID:1783130815-1"},
   CellID->1694100803]},
 "Proof|ID:1493153487"->{
  Cell[1702819, 45805, 9166, 221, 26, "ProofInfo",
   CellTags->{"Proof|ID:1493153487", "Proof|ID:1493153487-1"},
   CellID->1970791442]},
 "Proof|ID:1493153487-1"->{
  Cell[1702819, 45805, 9166, 221, 26, "ProofInfo",
   CellTags->{"Proof|ID:1493153487", "Proof|ID:1493153487-1"},
   CellID->1970791442]},
 "Proof|ID:1798779961"->{
  Cell[1712505, 46045, 6543, 161, 26, "ProofInfo",
   CellTags->{"Proof|ID:1798779961", "Proof|ID:1798779961-1"},
   CellID->1268768510]},
 "Proof|ID:1798779961-1"->{
  Cell[1712505, 46045, 6543, 161, 26, "ProofInfo",
   CellTags->{"Proof|ID:1798779961", "Proof|ID:1798779961-1"},
   CellID->1268768510]},
 "ID:712613448"->{
  Cell[1731959, 46627, 1024, 25, 41, "GlobalDeclaration",
   CellTags->{
    "ID:712613448", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->712613448]},
 "ID:1914773542"->{
  Cell[1732986, 46654, 3002, 91, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:1914773542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "37.\[DoubleRightArrow]"},
   CellID->1914773542]},
 "37.\[DoubleRightArrow]"->{
  Cell[1732986, 46654, 3002, 91, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:1914773542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "37.\[DoubleRightArrow]"},
   CellID->1914773542]},
 "ID:2048103084"->{
  Cell[1735991, 46747, 3026, 92, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:2048103084", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "37.\[DoubleLeftArrow]"},
   CellID->2048103084]},
 "37.\[DoubleLeftArrow]"->{
  Cell[1735991, 46747, 3026, 92, 119, "FormalTextInputFormula",
   CellTags->{
    "ID:2048103084", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "37.\[DoubleLeftArrow]"},
   CellID->2048103084]},
 "Proof|ID:1914773542"->{
  Cell[1739173, 46850, 8941, 213, 28, "ProofInfo",
   CellTags->{"Proof|ID:1914773542", "Proof|ID:1914773542-1"},
   CellID->450197198]},
 "Proof|ID:1914773542-1"->{
  Cell[1739173, 46850, 8941, 213, 28, "ProofInfo",
   CellTags->{"Proof|ID:1914773542", "Proof|ID:1914773542-1"},
   CellID->450197198]},
 "Proof|ID:2048103084"->{
  Cell[1748636, 47082, 8950, 215, 26, "ProofInfo",
   CellTags->{"Proof|ID:2048103084", "Proof|ID:2048103084-1"},
   CellID->1611988294]},
 "Proof|ID:2048103084-1"->{
  Cell[1748636, 47082, 8950, 215, 26, "ProofInfo",
   CellTags->{"Proof|ID:2048103084", "Proof|ID:2048103084-1"},
   CellID->1611988294]},
 "ID:254692456"->{
  Cell[1758672, 47333, 1003, 26, 42, "GlobalDeclaration",
   CellTags->{
    "ID:254692456", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->254692456]},
 "ID:336730845"->{
  Cell[1759678, 47361, 3946, 122, 158, "FormalTextInputFormula",
   CellTags->{
    "ID:336730845", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "32.\[DoubleRightArrow]"},
   CellID->336730845]},
 "32.\[DoubleRightArrow]"->{
  Cell[1759678, 47361, 3946, 122, 158, "FormalTextInputFormula",
   CellTags->{
    "ID:336730845", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "32.\[DoubleRightArrow]"},
   CellID->336730845]},
 "ID:1715399148"->{
  Cell[1763627, 47485, 3973, 123, 158, "FormalTextInputFormula",
   CellTags->{
    "ID:1715399148", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "32.\[DoubleLeftArrow]"},
   CellID->1715399148]},
 "32.\[DoubleLeftArrow]"->{
  Cell[1763627, 47485, 3973, 123, 158, "FormalTextInputFormula",
   CellTags->{
    "ID:1715399148", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "32.\[DoubleLeftArrow]"},
   CellID->1715399148]},
 "Proof|ID:336730845"->{
  Cell[1767757, 47619, 12027, 278, 28, "ProofInfo",
   CellTags->{"Proof|ID:336730845", "Proof|ID:336730845-1"},
   CellID->650311272]},
 "Proof|ID:336730845-1"->{
  Cell[1767757, 47619, 12027, 278, 28, "ProofInfo",
   CellTags->{"Proof|ID:336730845", "Proof|ID:336730845-1"},
   CellID->650311272]},
 "Proof|ID:1715399148"->{
  Cell[1780349, 47917, 12054, 281, 26, "ProofInfo",
   CellTags->{"Proof|ID:1715399148", "Proof|ID:1715399148-1"},
   CellID->2088013031]},
 "Proof|ID:1715399148-1"->{
  Cell[1780349, 47917, 12054, 281, 26, "ProofInfo",
   CellTags->{"Proof|ID:1715399148", "Proof|ID:1715399148-1"},
   CellID->2088013031]},
 "ID:1253203360"->{
  Cell[1793910, 48253, 885, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1253203360", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1253203360]},
 "ID:1514263990"->{
  Cell[1794798, 48280, 860, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1514263990", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[RightArrow]\[RawStar]"},
   CellID->1514263990]},
 "reflexivity of \[RightArrow]\[RawStar]"->{
  Cell[1794798, 48280, 860, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1514263990", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[RightArrow]\[RawStar]"},
   CellID->1514263990]},
 "ID:411529944"->{
  Cell[1795661, 48305, 1506, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:411529944", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[RightArrow]\[RawStar]"},
   CellID->411529944]},
 "transitivity of \[RightArrow]\[RawStar]"->{
  Cell[1795661, 48305, 1506, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:411529944", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[RightArrow]\[RawStar]"},
   CellID->411529944]},
 "Proof|ID:1514263990"->{
  Cell[1797325, 48366, 3526, 97, 70, "ProofInfo",
   CellTags->{"Proof|ID:1514263990", "Proof|ID:1514263990-1"},
   CellID->1590090525]},
 "Proof|ID:1514263990-1"->{
  Cell[1797325, 48366, 3526, 97, 70, "ProofInfo",
   CellTags->{"Proof|ID:1514263990", "Proof|ID:1514263990-1"},
   CellID->1590090525]},
 "Proof|ID:411529944"->{
  Cell[1801373, 48482, 6006, 149, 70, "ProofInfo",
   CellTags->{"Proof|ID:411529944", "Proof|ID:411529944-1"},
   CellID->144989886]},
 "Proof|ID:411529944-1"->{
  Cell[1801373, 48482, 6006, 149, 70, "ProofInfo",
   CellTags->{"Proof|ID:411529944", "Proof|ID:411529944-1"},
   CellID->144989886]},
 "ID:1862973197"->{
  Cell[1808290, 48664, 885, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1862973197", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1862973197]},
 "ID:237268641"->{
  Cell[1809178, 48691, 1202, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:237268641", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "\[RightArrow] \[SubsetEqual] \[RightArrow]\[RawStar]"},
   CellID->237268641]},
 "\[RightArrow] \[SubsetEqual] \[RightArrow]\[RawStar]"->{
  Cell[1809178, 48691, 1202, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:237268641", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "\[RightArrow] \[SubsetEqual] \[RightArrow]\[RawStar]"},
   CellID->237268641]},
 "ID:1746879589"->{
  Cell[1810383, 48731, 1492, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1746879589", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "38"},
   CellID->1746879589]},
 "38"->{
  Cell[1810383, 48731, 1492, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1746879589", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "38"},
   CellID->1746879589]},
 "Proof|ID:237268641"->{
  Cell[1812032, 48793, 4641, 122, 70, "ProofInfo",
   CellTags->{"Proof|ID:237268641", "Proof|ID:237268641-1"},
   CellID->1915348311]},
 "Proof|ID:237268641-1"->{
  Cell[1812032, 48793, 4641, 122, 70, "ProofInfo",
   CellTags->{"Proof|ID:237268641", "Proof|ID:237268641-1"},
   CellID->1915348311]},
 "Proof|ID:1746879589"->{
  Cell[1817194, 48934, 5800, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:1746879589", "Proof|ID:1746879589-1"},
   CellID->1067079851]},
 "Proof|ID:1746879589-1"->{
  Cell[1817194, 48934, 5800, 147, 70, "ProofInfo",
   CellTags->{"Proof|ID:1746879589", "Proof|ID:1746879589-1"},
   CellID->1067079851]},
 "ID:1956094048"->{
  Cell[1824602, 49136, 821, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1956094048", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1956094048]},
 "ID:560057636"->{
  Cell[1825426, 49161, 915, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:560057636", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of connectibility"},
   CellID->560057636]},
 "reflexivity of connectibility"->{
  Cell[1825426, 49161, 915, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:560057636", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of connectibility"},
   CellID->560057636]},
 "ID:354829984"->{
  Cell[1826344, 49187, 1325, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:354829984", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of connectibility"},
   CellID->354829984]},
 "symmetry of connectibility"->{
  Cell[1826344, 49187, 1325, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:354829984", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of connectibility"},
   CellID->354829984]},
 "ID:350848859"->{
  Cell[1827672, 49227, 1587, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:350848859", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of connectibility"},
   CellID->350848859]},
 "transitivity of connectibility"->{
  Cell[1827672, 49227, 1587, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:350848859", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of connectibility"},
   CellID->350848859]},
 "Proof|ID:560057636"->{
  Cell[1829415, 49289, 3265, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:560057636", "Proof|ID:560057636-1"},
   CellID->1331060490]},
 "Proof|ID:560057636-1"->{
  Cell[1829415, 49289, 3265, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:560057636", "Proof|ID:560057636-1"},
   CellID->1331060490]},
 "Proof|ID:354829984"->{
  Cell[1833243, 49399, 4506, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:354829984", "Proof|ID:354829984-1"},
   CellID->667920451]},
 "Proof|ID:354829984-1"->{
  Cell[1833243, 49399, 4506, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:354829984", "Proof|ID:354829984-1"},
   CellID->667920451]},
 "Proof|ID:350848859"->{
  Cell[1838269, 49534, 5795, 143, 70, "ProofInfo",
   CellTags->{"Proof|ID:350848859", "Proof|ID:350848859-1"},
   CellID->1457001757]},
 "Proof|ID:350848859-1"->{
  Cell[1838269, 49534, 5795, 143, 70, "ProofInfo",
   CellTags->{"Proof|ID:350848859", "Proof|ID:350848859-1"},
   CellID->1457001757]},
 "ID:705004664"->{
  Cell[1845413, 49718, 1091, 27, 70, "GlobalDeclaration",
   CellTags->{
    "ID:705004664", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->705004664]},
 "ID:411199728"->{
  Cell[1846507, 49747, 1392, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:411199728", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of connectibility below"},
   CellID->411199728]},
 "reflexivity of connectibility below"->{
  Cell[1846507, 49747, 1392, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:411199728", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of connectibility below"},
   CellID->411199728]},
 "ID:1502850808"->{
  Cell[1847902, 49793, 1730, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502850808", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of connectibility below"},
   CellID->1502850808]},
 "symmetry of connectibility below"->{
  Cell[1847902, 49793, 1730, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502850808", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of connectibility below"},
   CellID->1502850808]},
 "ID:1331049102"->{
  Cell[1849635, 49851, 2301, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1331049102", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of connectibility below"},
   CellID->1331049102]},
 "transitivity of connectibility below"->{
  Cell[1849635, 49851, 2301, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1331049102", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of connectibility below"},
   CellID->1331049102]},
 "Proof|ID:411199728"->{
  Cell[1852094, 49941, 4781, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:411199728", "Proof|ID:411199728-1"},
   CellID->24086679]},
 "Proof|ID:411199728-1"->{
  Cell[1852094, 49941, 4781, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:411199728", "Proof|ID:411199728-1"},
   CellID->24086679]},
 "Proof|ID:1502850808"->{
  Cell[1857490, 50086, 5722, 144, 70, "ProofInfo",
   CellTags->{"Proof|ID:1502850808", "Proof|ID:1502850808-1"},
   CellID->1390382820]},
 "Proof|ID:1502850808-1"->{
  Cell[1857490, 50086, 5722, 144, 70, "ProofInfo",
   CellTags->{"Proof|ID:1502850808", "Proof|ID:1502850808-1"},
   CellID->1390382820]},
 "Proof|ID:1331049102"->{
  Cell[1863799, 50250, 7301, 177, 70, "ProofInfo",
   CellTags->{"Proof|ID:1331049102", "Proof|ID:1331049102-1"},
   CellID->516908597]},
 "Proof|ID:1331049102-1"->{
  Cell[1863799, 50250, 7301, 177, 70, "ProofInfo",
   CellTags->{"Proof|ID:1331049102", "Proof|ID:1331049102-1"},
   CellID->516908597]},
 "ID:1147958002"->{
  Cell[1872203, 50463, 3364, 112, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1147958002", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "36"},
   CellID->1147958002]},
 "36"->{
  Cell[1872203, 50463, 3364, 112, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1147958002", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "36"},
   CellID->1147958002]},
 "ID:1663785837"->{
  Cell[1875570, 50577, 3848, 121, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1663785837", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "33"},
   CellID->1663785837]},
 "33"->{
  Cell[1875570, 50577, 3848, 121, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1663785837", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "33"},
   CellID->1663785837]},
 "Proof|ID:1147958002"->{
  Cell[1879575, 50709, 8798, 207, 70, "ProofInfo",
   CellTags->{"Proof|ID:1147958002", "Proof|ID:1147958002-1"},
   CellID->18036380]},
 "Proof|ID:1147958002-1"->{
  Cell[1879575, 50709, 8798, 207, 70, "ProofInfo",
   CellTags->{"Proof|ID:1147958002", "Proof|ID:1147958002-1"},
   CellID->18036380]},
 "Proof|ID:1663785837"->{
  Cell[1888893, 50935, 9121, 223, 70, "ProofInfo",
   CellTags->{"Proof|ID:1663785837", "Proof|ID:1663785837-1"},
   CellID->1295009646]},
 "Proof|ID:1663785837-1"->{
  Cell[1888893, 50935, 9121, 223, 70, "ProofInfo",
   CellTags->{"Proof|ID:1663785837", "Proof|ID:1663785837-1"},
   CellID->1295009646]},
 "ID:311737773"->{
  Cell[1899766, 51216, 802, 21, 41, "GlobalDeclaration",
   CellTags->{
    "ID:311737773", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->311737773]},
 "ID:552003293"->{
  Cell[1900571, 51239, 1149, 34, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:552003293", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "reflexivity of reflexive-transitive closure of \[RightArrow]"},
   CellID->552003293]},
 "reflexivity of reflexive-transitive closure of \[RightArrow]"->{
  Cell[1900571, 51239, 1149, 34, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:552003293", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
     "reflexivity of reflexive-transitive closure of \[RightArrow]"},
   CellID->552003293]},
 "ID:129456084"->{
  Cell[1901723, 51275, 767, 20, 44, "GlobalDeclaration",
   CellTags->{
    "ID:129456084", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->129456084]},
 "ID:441235278"->{
  Cell[1902493, 51297, 1868, 55, 45, "FormalTextInputFormula",
   CellTags->{
    "ID:441235278", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "34"},
   CellID->441235278]},
 "34"->{
  Cell[1902493, 51297, 1868, 55, 45, "FormalTextInputFormula",
   CellTags->{
    "ID:441235278", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "34"},
   CellID->441235278]},
 "ID:1516348781"->{
  Cell[1904364, 51354, 3714, 110, 99, "FormalTextInputFormula",
   CellTags->{
    "ID:1516348781", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "35"},
   CellID->1516348781]},
 "35"->{
  Cell[1904364, 51354, 3714, 110, 99, "FormalTextInputFormula",
   CellTags->{
    "ID:1516348781", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "35"},
   CellID->1516348781]},
 "Proof|ID:552003293"->{
  Cell[1908234, 51475, 3581, 98, 28, "ProofInfo",
   CellTags->{"Proof|ID:552003293", "Proof|ID:552003293-1"},
   CellID->84172006]},
 "Proof|ID:552003293-1"->{
  Cell[1908234, 51475, 3581, 98, 28, "ProofInfo",
   CellTags->{"Proof|ID:552003293", "Proof|ID:552003293-1"},
   CellID->84172006]},
 "Proof|ID:441235278"->{
  Cell[1912335, 51592, 7567, 184, 26, "ProofInfo",
   CellTags->{"Proof|ID:441235278", "Proof|ID:441235278-1"},
   CellID->9301273]},
 "Proof|ID:441235278-1"->{
  Cell[1912335, 51592, 7567, 184, 26, "ProofInfo",
   CellTags->{"Proof|ID:441235278", "Proof|ID:441235278-1"},
   CellID->9301273]},
 "Proof|ID:1516348781"->{
  Cell[1920422, 51795, 12100, 277, 26, "ProofInfo",
   CellTags->{"Proof|ID:1516348781", "Proof|ID:1516348781-1"},
   CellID->278395591]},
 "Proof|ID:1516348781-1"->{
  Cell[1920422, 51795, 12100, 277, 26, "ProofInfo",
   CellTags->{"Proof|ID:1516348781", "Proof|ID:1516348781-1"},
   CellID->278395591]},
 "ID:1329557757"->{
  Cell[1941928, 52368, 754, 20, 41, "GlobalDeclaration",
   CellTags->{
    "ID:1329557757", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1329557757]},
 "ID:1168535485"->{
  Cell[1942685, 52390, 2221, 74, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1168535485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "40"},
   CellID->1168535485]},
 "40"->{
  Cell[1942685, 52390, 2221, 74, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1168535485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "40"},
   CellID->1168535485]},
 "ID:2128489125"->{
  Cell[1944909, 52466, 2317, 77, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:2128489125", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "41"},
   CellID->2128489125]},
 "41"->{
  Cell[1944909, 52466, 2317, 77, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:2128489125", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "41"},
   CellID->2128489125]},
 "Proof|ID:1168535485"->{
  Cell[1947384, 52554, 6054, 155, 28, "ProofInfo",
   CellTags->{"Proof|ID:1168535485", "Proof|ID:1168535485-1"},
   CellID->397093125]},
 "Proof|ID:1168535485-1"->{
  Cell[1947384, 52554, 6054, 155, 28, "ProofInfo",
   CellTags->{"Proof|ID:1168535485", "Proof|ID:1168535485-1"},
   CellID->397093125]},
 "Proof|ID:2128489125"->{
  Cell[1953957, 52728, 7885, 193, 26, "ProofInfo",
   CellTags->{"Proof|ID:2128489125", "Proof|ID:2128489125-1"},
   CellID->360602067]},
 "Proof|ID:2128489125-1"->{
  Cell[1953957, 52728, 7885, 193, 26, "ProofInfo",
   CellTags->{"Proof|ID:2128489125", "Proof|ID:2128489125-1"},
   CellID->360602067]},
 "ID:548213617"->{
  Cell[1962757, 52954, 751, 20, 41, "GlobalDeclaration",
   CellTags->{
    "ID:548213617", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->548213617]},
 "ID:909891749"->{
  Cell[1963511, 52976, 1974, 63, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:909891749", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "42"},
   CellID->909891749]},
 "42"->{
  Cell[1963511, 52976, 1974, 63, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:909891749", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "42"},
   CellID->909891749]},
 "ID:830274722"->{
  Cell[1965488, 53041, 936, 27, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:830274722", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "44"},
   CellID->830274722]},
 "44"->{
  Cell[1965488, 53041, 936, 27, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:830274722", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "44"},
   CellID->830274722]},
 "ID:1390954175"->{
  Cell[1966427, 53070, 760, 20, 30, "GlobalDeclaration",
   CellTags->{
    "ID:1390954175", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1390954175]},
 "ID:2016336231"->{
  Cell[1967190, 53092, 1126, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:2016336231", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68"},
   CellID->2016336231]},
 "68"->{
  Cell[1967190, 53092, 1126, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:2016336231", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68"},
   CellID->2016336231]},
 "ID:1052198272"->{
  Cell[1968319, 53126, 1449, 45, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1052198272", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68.set"},
   CellID->1052198272]},
 "68.set"->{
  Cell[1968319, 53126, 1449, 45, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:1052198272", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68.set"},
   CellID->1052198272]},
 "ID:35312349"->{
  Cell[1969771, 53173, 1749, 59, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:35312349", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68.tuple"},
   CellID->35312349]},
 "68.tuple"->{
  Cell[1969771, 53173, 1749, 59, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:35312349", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68.tuple"},
   CellID->35312349]},
 "Proof|ID:909891749"->{
  Cell[1971677, 53243, 5695, 146, 28, "ProofInfo",
   CellTags->{"Proof|ID:909891749", "Proof|ID:909891749-1"},
   CellID->534981557]},
 "Proof|ID:909891749-1"->{
  Cell[1971677, 53243, 5695, 146, 28, "ProofInfo",
   CellTags->{"Proof|ID:909891749", "Proof|ID:909891749-1"},
   CellID->534981557]},
 "Proof|ID:830274722"->{
  Cell[1977891, 53408, 3910, 106, 26, "ProofInfo",
   CellTags->{"Proof|ID:830274722", "Proof|ID:830274722-1"},
   CellID->601881535]},
 "Proof|ID:830274722-1"->{
  Cell[1977891, 53408, 3910, 106, 26, "ProofInfo",
   CellTags->{"Proof|ID:830274722", "Proof|ID:830274722-1"},
   CellID->601881535]},
 "Proof|ID:2016336231"->{
  Cell[1982321, 53533, 5420, 139, 26, "ProofInfo",
   CellTags->{"Proof|ID:2016336231", "Proof|ID:2016336231-1"},
   CellID->490882312]},
 "Proof|ID:2016336231-1"->{
  Cell[1982321, 53533, 5420, 139, 26, "ProofInfo",
   CellTags->{"Proof|ID:2016336231", "Proof|ID:2016336231-1"},
   CellID->490882312]},
 "Proof|ID:1052198272"->{
  Cell[1988262, 53691, 6820, 170, 26, "ProofInfo",
   CellTags->{"Proof|ID:1052198272", "Proof|ID:1052198272-1"},
   CellID->1895460698]},
 "Proof|ID:1052198272-1"->{
  Cell[1988262, 53691, 6820, 170, 26, "ProofInfo",
   CellTags->{"Proof|ID:1052198272", "Proof|ID:1052198272-1"},
   CellID->1895460698]},
 "Proof|ID:35312349"->{
  Cell[1995604, 53880, 6609, 163, 26, "ProofInfo",
   CellTags->{"Proof|ID:35312349", "Proof|ID:35312349-1"},
   CellID->860762961]},
 "Proof|ID:35312349-1"->{
  Cell[1995604, 53880, 6609, 163, 26, "ProofInfo",
   CellTags->{"Proof|ID:35312349", "Proof|ID:35312349-1"},
   CellID->860762961]},
 "ID:2032809580"->{
  Cell[2003857, 54095, 870, 24, 42, "GlobalDeclaration",
   CellTags->{
    "ID:2032809580", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2032809580]},
 "ID:1947090769"->{
  Cell[2004730, 54121, 711, 19, 32, "GlobalDeclaration",
   CellTags->{
    "ID:1947090769", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1947090769]},
 "ID:693923465"->{
  Cell[2005444, 54142, 3280, 101, 102, "FormalTextInputFormula",
   CellTags->{
    "ID:693923465", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "50"},
   CellID->693923465]},
 "50"->{
  Cell[2005444, 54142, 3280, 101, 102, "FormalTextInputFormula",
   CellTags->{
    "ID:693923465", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "50"},
   CellID->693923465]},
 "ID:1234215913"->{
  Cell[2008727, 54245, 1394, 46, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1234215913", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "51"},
   CellID->1234215913]},
 "51"->{
  Cell[2008727, 54245, 1394, 46, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1234215913", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "51"},
   CellID->1234215913]},
 "ID:1065734937"->{
  Cell[2010124, 54293, 1758, 61, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1065734937", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "52"},
   CellID->1065734937]},
 "52"->{
  Cell[2010124, 54293, 1758, 61, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1065734937", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "52"},
   CellID->1065734937]},
 "ID:1763925306"->{
  Cell[2011885, 54356, 2270, 76, 56, "FormalTextInputFormula",
   CellTags->{
    "ID:1763925306", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "53"},
   CellID->1763925306]},
 "53"->{
  Cell[2011885, 54356, 2270, 76, 56, "FormalTextInputFormula",
   CellTags->{
    "ID:1763925306", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "53"},
   CellID->1763925306]},
 "Proof|ID:693923465"->{
  Cell[2014313, 54443, 10134, 238, 28, "ProofInfo",
   CellTags->{"Proof|ID:693923465", "Proof|ID:693923465-1"},
   CellID->227607100]},
 "Proof|ID:693923465-1"->{
  Cell[2014313, 54443, 10134, 238, 28, "ProofInfo",
   CellTags->{"Proof|ID:693923465", "Proof|ID:693923465-1"},
   CellID->227607100]},
 "Proof|ID:1234215913"->{
  Cell[2024966, 54700, 5955, 149, 26, "ProofInfo",
   CellTags->{"Proof|ID:1234215913", "Proof|ID:1234215913-1"},
   CellID->575064868]},
 "Proof|ID:1234215913-1"->{
  Cell[2024966, 54700, 5955, 149, 26, "ProofInfo",
   CellTags->{"Proof|ID:1234215913", "Proof|ID:1234215913-1"},
   CellID->575064868]},
 "Proof|ID:1065734937"->{
  Cell[2031511, 54869, 7746, 185, 26, "ProofInfo",
   CellTags->{"Proof|ID:1065734937", "Proof|ID:1065734937-1"},
   CellID->1345357544]},
 "Proof|ID:1065734937-1"->{
  Cell[2031511, 54869, 7746, 185, 26, "ProofInfo",
   CellTags->{"Proof|ID:1065734937", "Proof|ID:1065734937-1"},
   CellID->1345357544]},
 "Proof|ID:1763925306"->{
  Cell[2039778, 55073, 7388, 181, 26, "ProofInfo",
   CellTags->{"Proof|ID:1763925306", "Proof|ID:1763925306-1"},
   CellID->278876241]},
 "Proof|ID:1763925306-1"->{
  Cell[2039778, 55073, 7388, 181, 26, "ProofInfo",
   CellTags->{"Proof|ID:1763925306", "Proof|ID:1763925306-1"},
   CellID->278876241]},
 "ID:332034572"->{
  Cell[2048206, 55290, 839, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:332034572", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->332034572]},
 "ID:1609461850"->{
  Cell[2049048, 55314, 1006, 30, 40, "GlobalDeclaration",
   CellTags->{
    "ID:1609461850", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1609461850]},
 "ID:621597434"->{
  Cell[2050057, 55346, 1801, 63, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:621597434", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "54"},
   CellID->621597434]},
 "54"->{
  Cell[2050057, 55346, 1801, 63, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:621597434", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "54"},
   CellID->621597434]},
 "ID:2031891226"->{
  Cell[2051861, 55411, 1814, 63, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:2031891226", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "55"},
   CellID->2031891226]},
 "55"->{
  Cell[2051861, 55411, 1814, 63, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:2031891226", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "55"},
   CellID->2031891226]},
 "ID:1596753391"->{
  Cell[2053678, 55476, 770, 20, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1596753391", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1596753391]},
 "ID:1855698114"->{
  Cell[2054451, 55498, 1661, 59, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1855698114", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "55'"},
   CellID->1855698114]},
 "55'"->{
  Cell[2054451, 55498, 1661, 59, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1855698114", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "55'"},
   CellID->1855698114]},
 "Proof|ID:621597434"->{
  Cell[2056269, 55568, 8302, 202, 28, "ProofInfo",
   CellTags->{"Proof|ID:621597434", "Proof|ID:621597434-1"},
   CellID->88899104]},
 "Proof|ID:621597434-1"->{
  Cell[2056269, 55568, 8302, 202, 28, "ProofInfo",
   CellTags->{"Proof|ID:621597434", "Proof|ID:621597434-1"},
   CellID->88899104]},
 "Proof|ID:2031891226"->{
  Cell[2065090, 55789, 8331, 203, 26, "ProofInfo",
   CellTags->{"Proof|ID:2031891226", "Proof|ID:2031891226-1"},
   CellID->50940196]},
 "Proof|ID:2031891226-1"->{
  Cell[2065090, 55789, 8331, 203, 26, "ProofInfo",
   CellTags->{"Proof|ID:2031891226", "Proof|ID:2031891226-1"},
   CellID->50940196]},
 "Proof|ID:1855698114"->{
  Cell[2073942, 56011, 8646, 209, 26, "ProofInfo",
   CellTags->{"Proof|ID:1855698114", "Proof|ID:1855698114-1"},
   CellID->204417677]},
 "Proof|ID:1855698114-1"->{
  Cell[2073942, 56011, 8646, 209, 26, "ProofInfo",
   CellTags->{"Proof|ID:1855698114", "Proof|ID:1855698114-1"},
   CellID->204417677]},
 "ID:1116860460"->{
  Cell[2083630, 56255, 779, 20, 41, "GlobalDeclaration",
   CellTags->{
    "ID:1116860460", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1116860460]},
 "ID:915980145"->{
  Cell[2084412, 56277, 2190, 70, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:915980145", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "56"},
   CellID->915980145]},
 "56"->{
  Cell[2084412, 56277, 2190, 70, 53, "FormalTextInputFormula",
   CellTags->{
    "ID:915980145", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "56"},
   CellID->915980145]},
 "ID:1683552886"->{
  Cell[2086605, 56349, 768, 20, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1683552886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1683552886]},
 "ID:1331309258"->{
  Cell[2087376, 56371, 713, 18, 31, "GlobalDeclaration",
   CellTags->{
    "ID:1331309258", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1331309258]},
 "ID:551817557"->{
  Cell[2088092, 56391, 784, 21, 41, "GlobalDeclaration",
   CellTags->{
    "ID:551817557", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->551817557]},
 "ID:1724920630"->{
  Cell[2088879, 56414, 1214, 40, 43, "GlobalDeclaration",
   CellTags->{
    "ID:1724920630", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1724920630]},
 "ID:88655200"->{
  Cell[2090096, 56456, 1255, 40, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:88655200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "57"},
   CellID->88655200]},
 "57"->{
  Cell[2090096, 56456, 1255, 40, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:88655200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "57"},
   CellID->88655200]},
 "ID:1149353049"->{
  Cell[2091354, 56498, 1544, 50, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1149353049", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "58"},
   CellID->1149353049]},
 "58"->{
  Cell[2091354, 56498, 1544, 50, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:1149353049", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "58"},
   CellID->1149353049]},
 "Proof|ID:915980145"->{
  Cell[2093055, 56559, 5922, 150, 28, "ProofInfo",
   CellTags->{"Proof|ID:915980145", "Proof|ID:915980145-1"},
   CellID->552018998]},
 "Proof|ID:915980145-1"->{
  Cell[2093055, 56559, 5922, 150, 28, "ProofInfo",
   CellTags->{"Proof|ID:915980145", "Proof|ID:915980145-1"},
   CellID->552018998]},
 "Proof|ID:88655200"->{
  Cell[2099496, 56728, 8599, 207, 26, "ProofInfo",
   CellTags->{"Proof|ID:88655200", "Proof|ID:88655200-1"},
   CellID->7475235]},
 "Proof|ID:88655200-1"->{
  Cell[2099496, 56728, 8599, 207, 26, "ProofInfo",
   CellTags->{"Proof|ID:88655200", "Proof|ID:88655200-1"},
   CellID->7475235]},
 "Proof|ID:1149353049"->{
  Cell[2108614, 56954, 9394, 223, 26, "ProofInfo",
   CellTags->{"Proof|ID:1149353049", "Proof|ID:1149353049-1"},
   CellID->180962942]},
 "Proof|ID:1149353049-1"->{
  Cell[2108614, 56954, 9394, 223, 26, "ProofInfo",
   CellTags->{"Proof|ID:1149353049", "Proof|ID:1149353049-1"},
   CellID->180962942]},
 "ID:895068823"->{
  Cell[2126782, 57478, 792, 21, 42, "GlobalDeclaration",
   CellTags->{
    "ID:895068823", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->895068823]},
 "ID:296045170"->{
  Cell[2127577, 57501, 2607, 90, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:296045170", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "45"},
   CellID->296045170]},
 "45"->{
  Cell[2127577, 57501, 2607, 90, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:296045170", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "45"},
   CellID->296045170]},
 "ID:1887054887"->{
  Cell[2130187, 57593, 2729, 94, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:1887054887", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "46"},
   CellID->1887054887]},
 "46"->{
  Cell[2130187, 57593, 2729, 94, 59, "FormalTextInputFormula",
   CellTags->{
    "ID:1887054887", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "46"},
   CellID->1887054887]},
 "Proof|ID:296045170"->{
  Cell[2133072, 57698, 9068, 214, 28, "ProofInfo",
   CellTags->{"Proof|ID:296045170", "Proof|ID:296045170-1"},
   CellID->9942592]},
 "Proof|ID:296045170-1"->{
  Cell[2133072, 57698, 9068, 214, 28, "ProofInfo",
   CellTags->{"Proof|ID:296045170", "Proof|ID:296045170-1"},
   CellID->9942592]},
 "Proof|ID:1887054887"->{
  Cell[2142660, 57931, 8997, 213, 26, "ProofInfo",
   CellTags->{"Proof|ID:1887054887", "Proof|ID:1887054887-1"},
   CellID->849984384]},
 "Proof|ID:1887054887-1"->{
  Cell[2142660, 57931, 8997, 213, 26, "ProofInfo",
   CellTags->{"Proof|ID:1887054887", "Proof|ID:1887054887-1"},
   CellID->849984384]},
 "ID:1876022438"->{
  Cell[2152755, 58183, 1753, 53, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1876022438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "59"},
   CellID->1876022438]},
 "59"->{
  Cell[2152755, 58183, 1753, 53, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:1876022438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "59"},
   CellID->1876022438]},
 "Proof|ID:1876022438"->{
  Cell[2154664, 58247, 4448, 117, 28, "ProofInfo",
   CellTags->{"Proof|ID:1876022438", "Proof|ID:1876022438-1"},
   CellID->1166331884]},
 "Proof|ID:1876022438-1"->{
  Cell[2154664, 58247, 4448, 117, 28, "ProofInfo",
   CellTags->{"Proof|ID:1876022438", "Proof|ID:1876022438-1"},
   CellID->1166331884]},
 "ID:1852246688"->{
  Cell[2160309, 58403, 779, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1852246688", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1852246688]},
 "ID:1471949021"->{
  Cell[2161618, 58441, 2591, 86, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1471949021", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "element confluence"},
   CellID->1471949021]},
 "element confluence"->{
  Cell[2161618, 58441, 2591, 86, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1471949021", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "element confluence"},
   CellID->1471949021]},
 "ID:146706347"->{
  Cell[2164212, 58529, 2513, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:146706347", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "confluence"},
   CellID->146706347]},
 "confluence"->{
  Cell[2164212, 58529, 2513, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:146706347", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "confluence"},
   CellID->146706347]},
 "ID:1279966325"->{
  Cell[2166728, 58601, 2403, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1279966325", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Church-Rosser property"},
   CellID->1279966325]},
 "Church-Rosser property"->{
  Cell[2166728, 58601, 2403, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1279966325", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Church-Rosser property"},
   CellID->1279966325]},
 "ID:1602877499"->{
  Cell[2169134, 58671, 2770, 90, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1602877499", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "local connectivity"},
   CellID->1602877499]},
 "local connectivity"->{
  Cell[2169134, 58671, 2770, 90, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1602877499", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "local connectivity"},
   CellID->1602877499]},
 "ID:661413373"->{
  Cell[2171907, 58763, 1504, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:661413373", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Gr\[ODoubleDot]bner basis"},
   CellID->661413373]},
 "Gr\[ODoubleDot]bner basis"->{
  Cell[2171907, 58763, 1504, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:661413373", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Gr\[ODoubleDot]bner basis"},
   CellID->661413373]},
 "ID:1444563707"->{
  Cell[2174085, 58826, 973, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1444563707", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1444563707]},
 "ID:1408266583"->{
  Cell[2175061, 58852, 2840, 84, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1408266583", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areConnectibleMultipliers"},
   CellID->1408266583]},
 "areConnectibleMultipliers"->{
  Cell[2175061, 58852, 2840, 84, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1408266583", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areConnectibleMultipliers"},
   CellID->1408266583]},
 "ID:882680545"->{
  Cell[2177904, 58938, 2867, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:882680545", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of one critical pair"},
   CellID->882680545]},
 "connectibility of one critical pair"->{
  Cell[2177904, 58938, 2867, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:882680545", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of one critical pair"},
   CellID->882680545]},
 "ID:259042764"->{
  Cell[2180774, 59019, 3995, 103, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:259042764", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs"},
   CellID->259042764]},
 "connectibility of critical pairs"->{
  Cell[2180774, 59019, 3995, 103, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:259042764", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs"},
   CellID->259042764]},
 "ID:1836116785"->{
  Cell[2184772, 59124, 1451, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1836116785", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs of tuple"},
   CellID->1836116785]},
 "connectibility of critical pairs of tuple"->{
  Cell[2184772, 59124, 1451, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1836116785", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs of tuple"},
   CellID->1836116785]},
 "ID:2066466981"->{
  Cell[2186226, 59169, 2080, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2066466981", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs of elements"},
   CellID->2066466981]},
 "connectibility of critical pairs of elements"->{
  Cell[2186226, 59169, 2080, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2066466981", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs of elements"},
   CellID->2066466981]},
 "ID:756678784"->{
  Cell[2188309, 59231, 2727, 90, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:756678784", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "local connectivity of elements"},
   CellID->756678784]},
 "local connectivity of elements"->{
  Cell[2188309, 59231, 2727, 90, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:756678784", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "local connectivity of elements"},
   CellID->756678784]},
 "ID:1173678670"->{
  Cell[2191580, 59342, 676, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1173678670", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1173678670]},
 "ID:1524905762"->{
  Cell[2192729, 59377, 1492, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1524905762", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible 1"},
   CellID->1524905762]},
 "cpConnectible 1"->{
  Cell[2192729, 59377, 1492, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1524905762", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible 1"},
   CellID->1524905762]},
 "Proof|ID:1524905762"->{
  Cell[2194379, 59432, 5724, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1524905762", "Proof|ID:1524905762-1"},
   CellID->709826783]},
 "Proof|ID:1524905762-1"->{
  Cell[2194379, 59432, 5724, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1524905762", "Proof|ID:1524905762-1"},
   CellID->709826783]},
 "ID:2054204196"->{
  Cell[2201086, 59605, 756, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2054204196", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2054204196]},
 "ID:194336473"->{
  Cell[2201845, 59627, 769, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:194336473", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->194336473]},
 "ID:2064859128"->{
  Cell[2202617, 59649, 1048, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2064859128", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible tuple"},
   CellID->2064859128]},
 "cpConnectible tuple"->{
  Cell[2202617, 59649, 1048, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2064859128", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible tuple"},
   CellID->2064859128]},
 "ID:1276073437"->{
  Cell[2203668, 59680, 1053, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1276073437", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective tuple"},
   CellID->1276073437]},
 "isLocConnective tuple"->{
  Cell[2203668, 59680, 1053, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1276073437", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective tuple"},
   CellID->1276073437]},
 "ID:1869404165"->{
  Cell[2204724, 59711, 983, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1869404165", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective tuple II"},
   CellID->1869404165]},
 "isLocConnective tuple II"->{
  Cell[2204724, 59711, 983, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1869404165", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective tuple II"},
   CellID->1869404165]},
 "Proof|ID:2064859128"->{
  Cell[2205863, 59748, 4691, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:2064859128", "Proof|ID:2064859128-1"},
   CellID->1478532560]},
 "Proof|ID:2064859128-1"->{
  Cell[2205863, 59748, 4691, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:2064859128", "Proof|ID:2064859128-1"},
   CellID->1478532560]},
 "Proof|ID:1276073437"->{
  Cell[2211076, 59888, 4702, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:1276073437", "Proof|ID:1276073437-1"},
   CellID->253023626]},
 "Proof|ID:1276073437-1"->{
  Cell[2211076, 59888, 4702, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:1276073437", "Proof|ID:1276073437-1"},
   CellID->253023626]},
 "Proof|ID:1869404165"->{
  Cell[2216299, 60028, 3368, 93, 70, "ProofInfo",
   CellTags->{"Proof|ID:1869404165", "Proof|ID:1869404165-1"},
   CellID->272805702]},
 "Proof|ID:1869404165-1"->{
  Cell[2216299, 60028, 3368, 93, 70, "ProofInfo",
   CellTags->{"Proof|ID:1869404165", "Proof|ID:1869404165-1"},
   CellID->272805702]},
 "ID:2083446865"->{
  Cell[2220683, 60157, 770, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2083446865", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2083446865]},
 "ID:709723813"->{
  Cell[2221456, 60179, 1470, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:709723813", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative"},
   CellID->709723813]},
 "cpConnectible alternative"->{
  Cell[2221456, 60179, 1470, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:709723813", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative"},
   CellID->709723813]},
 "ID:41201803"->{
  Cell[2222929, 60220, 1342, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:41201803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective alternative"},
   CellID->41201803]},
 "isLocConnective alternative"->{
  Cell[2222929, 60220, 1342, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:41201803", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective alternative"},
   CellID->41201803]},
 "ID:1023292110"->{
  Cell[2224274, 60256, 1459, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1023292110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective sufficient"},
   CellID->1023292110]},
 "isLocConnective sufficient"->{
  Cell[2224274, 60256, 1459, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1023292110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective sufficient"},
   CellID->1023292110]},
 "Proof|ID:709723813"->{
  Cell[2225889, 60310, 5306, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:709723813", "Proof|ID:709723813-1"},
   CellID->1570886920]},
 "Proof|ID:709723813-1"->{
  Cell[2225889, 60310, 5306, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:709723813", "Proof|ID:709723813-1"},
   CellID->1570886920]},
 "Proof|ID:41201803"->{
  Cell[2231716, 60463, 4386, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:41201803", "Proof|ID:41201803-1"},
   CellID->172260705]},
 "Proof|ID:41201803-1"->{
  Cell[2231716, 60463, 4386, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:41201803", "Proof|ID:41201803-1"},
   CellID->172260705]},
 "Proof|ID:1023292110"->{
  Cell[2236621, 60596, 5654, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:1023292110", "Proof|ID:1023292110-1"},
   CellID->141981476]},
 "Proof|ID:1023292110-1"->{
  Cell[2236621, 60596, 5654, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:1023292110", "Proof|ID:1023292110-1"},
   CellID->141981476]},
 "ID:529333522"->{
  Cell[2243244, 60770, 930, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:529333522", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->529333522]},
 "ID:208412986"->{
  Cell[2244177, 60798, 1787, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208412986", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible join"},
   CellID->208412986]},
 "cpConnectible join"->{
  Cell[2244177, 60798, 1787, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208412986", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible join"},
   CellID->208412986]},
 "ID:1183553358"->{
  Cell[2245967, 60855, 1845, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1183553358", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible append"},
   CellID->1183553358]},
 "cpConnectible append"->{
  Cell[2245967, 60855, 1845, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1183553358", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible append"},
   CellID->1183553358]},
 "ID:286665714"->{
  Cell[2247815, 60913, 1359, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:286665714", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible join II"},
   CellID->286665714]},
 "cpConnectible join II"->{
  Cell[2247815, 60913, 1359, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:286665714", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible join II"},
   CellID->286665714]},
 "ID:1065812089"->{
  Cell[2249177, 60956, 1398, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1065812089", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible append II"},
   CellID->1065812089]},
 "cpConnectible append II"->{
  Cell[2249177, 60956, 1398, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1065812089", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible append II"},
   CellID->1065812089]},
 "Proof|ID:208412986"->{
  Cell[2250732, 61008, 6755, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:208412986", "Proof|ID:208412986-1"},
   CellID->511870621]},
 "Proof|ID:208412986-1"->{
  Cell[2250732, 61008, 6755, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:208412986", "Proof|ID:208412986-1"},
   CellID->511870621]},
 "Proof|ID:1183553358"->{
  Cell[2258008, 61187, 6655, 159, 70, "ProofInfo",
   CellTags->{"Proof|ID:1183553358", "Proof|ID:1183553358-1"},
   CellID->1244254393]},
 "Proof|ID:1183553358-1"->{
  Cell[2258008, 61187, 6655, 159, 70, "ProofInfo",
   CellTags->{"Proof|ID:1183553358", "Proof|ID:1183553358-1"},
   CellID->1244254393]},
 "Proof|ID:286665714"->{
  Cell[2265227, 61366, 5464, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:286665714", "Proof|ID:286665714-1"},
   CellID->2147068449]},
 "Proof|ID:286665714-1"->{
  Cell[2265227, 61366, 5464, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:286665714", "Proof|ID:286665714-1"},
   CellID->2147068449]},
 "Proof|ID:1065812089"->{
  Cell[2271212, 61521, 5363, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1065812089", "Proof|ID:1065812089-1"},
   CellID->1920791119]},
 "Proof|ID:1065812089-1"->{
  Cell[2271212, 61521, 5363, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1065812089", "Proof|ID:1065812089-1"},
   CellID->1920791119]},
 "ID:751627229"->{
  Cell[2277119, 61674, 929, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:751627229", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->751627229]},
 "ID:351093647"->{
  Cell[2278513, 61718, 832, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:351093647", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->351093647]},
 "ID:1476531812"->{
  Cell[2279348, 61743, 815, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1476531812", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1476531812]},
 "ID:1047637483"->{
  Cell[2280166, 61766, 1107, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1047637483", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible \[Tilde] \[DoubleRightArrow]"},
   CellID->1047637483]},
 "cpConnectible \[Tilde] \[DoubleRightArrow]"->{
  Cell[2280166, 61766, 1107, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1047637483", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible \[Tilde] \[DoubleRightArrow]"},
   CellID->1047637483]},
 "ID:1237319003"->{
  Cell[2281276, 61797, 1217, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1237319003", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible \[Tilde]"},
   CellID->1237319003]},
 "cpConnectible \[Tilde]"->{
  Cell[2281276, 61797, 1217, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1237319003", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible \[Tilde]"},
   CellID->1237319003]},
 "Proof|ID:1047637483"->{
  Cell[2282649, 61838, 8943, 205, 70, "ProofInfo",
   CellTags->{"Proof|ID:1047637483", "Proof|ID:1047637483-1"},
   CellID->1363695772]},
 "Proof|ID:1047637483-1"->{
  Cell[2282649, 61838, 8943, 205, 70, "ProofInfo",
   CellTags->{"Proof|ID:1047637483", "Proof|ID:1047637483-1"},
   CellID->1363695772]},
 "Proof|ID:1237319003"->{
  Cell[2292113, 62062, 8945, 205, 70, "ProofInfo",
   CellTags->{"Proof|ID:1237319003", "Proof|ID:1237319003-1"},
   CellID->1084461300]},
 "Proof|ID:1237319003-1"->{
  Cell[2292113, 62062, 8945, 205, 70, "ProofInfo",
   CellTags->{"Proof|ID:1237319003", "Proof|ID:1237319003-1"},
   CellID->1084461300]},
 "ID:1087064485"->{
  Cell[2302036, 62300, 2236, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1087064485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative I"},
   CellID->1087064485]},
 "cpConnectible alternative I"->{
  Cell[2302036, 62300, 2236, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1087064485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative I"},
   CellID->1087064485]},
 "ID:1334373255"->{
  Cell[2304275, 62366, 2578, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1334373255", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative I'"},
   CellID->1334373255]},
 "cpConnectible alternative I'"->{
  Cell[2304275, 62366, 2578, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1334373255", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative I'"},
   CellID->1334373255]},
 "ID:581920350"->{
  Cell[2306856, 62447, 1740, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:581920350", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative II"},
   CellID->581920350]},
 "cpConnectible alternative II"->{
  Cell[2306856, 62447, 1740, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:581920350", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative II"},
   CellID->581920350]},
 "ID:286693183"->{
  Cell[2308599, 62500, 2240, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:286693183", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative II'"},
   CellID->286693183]},
 "cpConnectible alternative II'"->{
  Cell[2308599, 62500, 2240, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:286693183", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative II'"},
   CellID->286693183]},
 "Proof|ID:1087064485"->{
  Cell[2310996, 62580, 9110, 213, 70, "ProofInfo",
   CellTags->{"Proof|ID:1087064485", "Proof|ID:1087064485-1"},
   CellID->14160893]},
 "Proof|ID:1087064485-1"->{
  Cell[2310996, 62580, 9110, 213, 70, "ProofInfo",
   CellTags->{"Proof|ID:1087064485", "Proof|ID:1087064485-1"},
   CellID->14160893]},
 "Proof|ID:1334373255"->{
  Cell[2320627, 62812, 10357, 240, 70, "ProofInfo",
   CellTags->{"Proof|ID:1334373255", "Proof|ID:1334373255-1"},
   CellID->302186508]},
 "Proof|ID:1334373255-1"->{
  Cell[2320627, 62812, 10357, 240, 70, "ProofInfo",
   CellTags->{"Proof|ID:1334373255", "Proof|ID:1334373255-1"},
   CellID->302186508]},
 "Proof|ID:581920350"->{
  Cell[2331506, 63071, 7436, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:581920350", "Proof|ID:581920350-1"},
   CellID->448065758]},
 "Proof|ID:581920350-1"->{
  Cell[2331506, 63071, 7436, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:581920350", "Proof|ID:581920350-1"},
   CellID->448065758]},
 "Proof|ID:286693183"->{
  Cell[2339462, 63268, 8676, 204, 70, "ProofInfo",
   CellTags->{"Proof|ID:286693183", "Proof|ID:286693183-1"},
   CellID->514361483]},
 "Proof|ID:286693183-1"->{
  Cell[2339462, 63268, 8676, 204, 70, "ProofInfo",
   CellTags->{"Proof|ID:286693183", "Proof|ID:286693183-1"},
   CellID->514361483]},
 "ID:173719369"->{
  Cell[2349221, 63508, 757, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:173719369", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->173719369]},
 "ID:762511022"->{
  Cell[2349981, 63530, 1038, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:762511022", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible symmetric"},
   CellID->762511022]},
 "cpConnectible symmetric"->{
  Cell[2349981, 63530, 1038, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:762511022", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible symmetric"},
   CellID->762511022]},
 "ID:1792796192"->{
  Cell[2351022, 63559, 1050, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1792796192", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective symmetric"},
   CellID->1792796192]},
 "isLocConnective symmetric"->{
  Cell[2351022, 63559, 1050, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1792796192", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective symmetric"},
   CellID->1792796192]},
 "Proof|ID:762511022"->{
  Cell[2352229, 63597, 5273, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:762511022", "Proof|ID:762511022-1"},
   CellID->262270078]},
 "Proof|ID:762511022-1"->{
  Cell[2352229, 63597, 5273, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:762511022", "Proof|ID:762511022-1"},
   CellID->262270078]},
 "Proof|ID:1792796192"->{
  Cell[2358022, 63748, 5293, 133, 70, "ProofInfo",
   CellTags->{"Proof|ID:1792796192", "Proof|ID:1792796192-1"},
   CellID->13401702]},
 "Proof|ID:1792796192-1"->{
  Cell[2358022, 63748, 5293, 133, 70, "ProofInfo",
   CellTags->{"Proof|ID:1792796192", "Proof|ID:1792796192-1"},
   CellID->13401702]},
 "ID:320557773"->{
  Cell[2364270, 63914, 1389, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:320557773", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible 0"},
   CellID->320557773]},
 "cpConnectible 0"->{
  Cell[2364270, 63914, 1389, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:320557773", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible 0"},
   CellID->320557773]},
 "Proof|ID:320557773"->{
  Cell[2365816, 63969, 5376, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:320557773", "Proof|ID:320557773-1"},
   CellID->1364381745]},
 "Proof|ID:320557773-1"->{
  Cell[2365816, 63969, 5376, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:320557773", "Proof|ID:320557773-1"},
   CellID->1364381745]},
 "ID:1193869577"->{
  Cell[2372984, 64154, 818, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1193869577", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1193869577]},
 "ID:342993412"->{
  Cell[2373845, 64179, 734, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:342993412", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->342993412]},
 "ID:1786903669"->{
  Cell[2375148, 64219, 4772, 136, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1786903669", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "63 ind"},
   CellID->1786903669]},
 "63 ind"->{
  Cell[2375148, 64219, 4772, 136, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1786903669", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "63 ind"},
   CellID->1786903669]},
 "ID:1742873082"->{
  Cell[2379923, 64357, 4400, 129, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1742873082", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "63"},
   CellID->1742873082]},
 "63"->{
  Cell[2379923, 64357, 4400, 129, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1742873082", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "63"},
   CellID->1742873082]},
 "ID:332228196"->{
  Cell[2384326, 64488, 1191, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:332228196", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "64"},
   CellID->332228196]},
 "64"->{
  Cell[2384326, 64488, 1191, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:332228196", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "64"},
   CellID->332228196]},
 "Proof|ID:1786903669"->{
  Cell[2385672, 64532, 12724, 291, 70, "ProofInfo",
   CellTags->{"Proof|ID:1786903669", "Proof|ID:1786903669-1"},
   CellID->624008027]},
 "Proof|ID:1786903669-1"->{
  Cell[2385672, 64532, 12724, 291, 70, "ProofInfo",
   CellTags->{"Proof|ID:1786903669", "Proof|ID:1786903669-1"},
   CellID->624008027]},
 "Proof|ID:1742873082"->{
  Cell[2398917, 64842, 11990, 273, 70, "ProofInfo",
   CellTags->{"Proof|ID:1742873082", "Proof|ID:1742873082-1"},
   CellID->1157809795]},
 "Proof|ID:1742873082-1"->{
  Cell[2398917, 64842, 11990, 273, 70, "ProofInfo",
   CellTags->{"Proof|ID:1742873082", "Proof|ID:1742873082-1"},
   CellID->1157809795]},
 "Proof|ID:332228196"->{
  Cell[2411494, 65135, 4105, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:332228196", "Proof|ID:332228196-1"},
   CellID->282816996]},
 "Proof|ID:332228196-1"->{
  Cell[2411494, 65135, 4105, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:332228196", "Proof|ID:332228196-1"},
   CellID->282816996]},
 "ID:2110225002"->{
  Cell[2416582, 65275, 1205, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2110225002", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "65"},
   CellID->2110225002]},
 "65"->{
  Cell[2416582, 65275, 1205, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2110225002", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "65"},
   CellID->2110225002]},
 "Proof|ID:2110225002"->{
  Cell[2417943, 65322, 4261, 111, 70, "ProofInfo",
   CellTags->{"Proof|ID:2110225002", "Proof|ID:2110225002-1"},
   CellID->2140493902]},
 "Proof|ID:2110225002-1"->{
  Cell[2417943, 65322, 4261, 111, 70, "ProofInfo",
   CellTags->{"Proof|ID:2110225002", "Proof|ID:2110225002-1"},
   CellID->2140493902]},
 "ID:629188208"->{
  Cell[2423235, 65468, 962, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:629188208", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "66"},
   CellID->629188208]},
 "66"->{
  Cell[2423235, 65468, 962, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:629188208", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "66"},
   CellID->629188208]},
 "ID:1222366601"->{
  Cell[2424200, 65497, 990, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1222366601", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "67"},
   CellID->1222366601]},
 "67"->{
  Cell[2424200, 65497, 990, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1222366601", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "67"},
   CellID->1222366601]},
 "Proof|ID:629188208"->{
  Cell[2425347, 65535, 3284, 88, 70, "ProofInfo",
   CellTags->{"Proof|ID:629188208", "Proof|ID:629188208-1"},
   CellID->2037000940]},
 "Proof|ID:629188208-1"->{
  Cell[2425347, 65535, 3284, 88, 70, "ProofInfo",
   CellTags->{"Proof|ID:629188208", "Proof|ID:629188208-1"},
   CellID->2037000940]},
 "Proof|ID:1222366601"->{
  Cell[2429151, 65642, 3299, 89, 70, "ProofInfo",
   CellTags->{"Proof|ID:1222366601", "Proof|ID:1222366601-1"},
   CellID->44880005]},
 "Proof|ID:1222366601-1"->{
  Cell[2429151, 65642, 3299, 89, 70, "ProofInfo",
   CellTags->{"Proof|ID:1222366601", "Proof|ID:1222366601-1"},
   CellID->44880005]},
 "ID:2131577443"->{
  Cell[2433562, 65768, 818, 20, 33, "GlobalDeclaration",
   CellTags->{
    "ID:2131577443", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2131577443]},
 "ID:2046259788"->{
  Cell[2434541, 65796, 835, 21, 44, "GlobalDeclaration",
   CellTags->{
    "ID:2046259788", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2046259788]},
 "ID:205712485"->{
  Cell[2435790, 65835, 1009, 25, 42, "GlobalDeclaration",
   CellTags->{
    "ID:205712485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->205712485]},
 "ID:2050580243"->{
  Cell[2436802, 65862, 2519, 75, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:2050580243", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "61"},
   CellID->2050580243]},
 "61"->{
  Cell[2436802, 65862, 2519, 75, 51, "FormalTextInputFormula",
   CellTags->{
    "ID:2050580243", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "61"},
   CellID->2050580243]},
 "ID:2100291458"->{
  Cell[2439324, 65939, 5765, 175, 146, "FormalTextInputFormula",
   CellTags->{
    "ID:2100291458", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "60"},
   CellID->2100291458]},
 "60"->{
  Cell[2439324, 65939, 5765, 175, 146, "FormalTextInputFormula",
   CellTags->{
    "ID:2100291458", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "60"},
   CellID->2100291458]},
 "ID:1810831635"->{
  Cell[2445092, 66116, 687, 17, 30, "GlobalDeclaration",
   CellTags->{
    "ID:1810831635", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1810831635]},
 "ID:718268018"->{
  Cell[2445782, 66135, 847, 22, 42, "GlobalDeclaration",
   CellTags->{
    "ID:718268018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->718268018]},
 "ID:1959449230"->{
  Cell[2446632, 66159, 4333, 133, 248, "FormalTextInputFormula",
   CellTags->{
    "ID:1959449230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "62"},
   CellID->1959449230]},
 "62"->{
  Cell[2446632, 66159, 4333, 133, 248, "FormalTextInputFormula",
   CellTags->{
    "ID:1959449230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "62"},
   CellID->1959449230]},
 "Proof|ID:2050580243"->{
  Cell[2451122, 66303, 10535, 237, 28, "ProofInfo",
   CellTags->{"Proof|ID:2050580243", "Proof|ID:2050580243-1"},
   CellID->1516400018]},
 "Proof|ID:2050580243-1"->{
  Cell[2451122, 66303, 10535, 237, 28, "ProofInfo",
   CellTags->{"Proof|ID:2050580243", "Proof|ID:2050580243-1"},
   CellID->1516400018]},
 "Proof|ID:2100291458"->{
  Cell[2462179, 66559, 18843, 403, 26, "ProofInfo",
   CellTags->{"Proof|ID:2100291458", "Proof|ID:2100291458-1"},
   CellID->44637999]},
 "Proof|ID:2100291458-1"->{
  Cell[2462179, 66559, 18843, 403, 26, "ProofInfo",
   CellTags->{"Proof|ID:2100291458", "Proof|ID:2100291458-1"},
   CellID->44637999]},
 "Proof|ID:1959449230"->{
  Cell[2481543, 66981, 16179, 353, 26, "ProofInfo",
   CellTags->{"Proof|ID:1959449230", "Proof|ID:1959449230-1"},
   CellID->340301213]},
 "Proof|ID:1959449230-1"->{
  Cell[2481543, 66981, 16179, 353, 26, "ProofInfo",
   CellTags->{"Proof|ID:1959449230", "Proof|ID:1959449230-1"},
   CellID->340301213]},
 "ID:1029764736"->{
  Cell[2513610, 67871, 4006, 99, 164, "FormalTextInputFormula",
   CellTags->{
    "ID:1029764736", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "70"},
   CellID->1029764736]},
 "70"->{
  Cell[2513610, 67871, 4006, 99, 164, "FormalTextInputFormula",
   CellTags->{
    "ID:1029764736", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "70"},
   CellID->1029764736]},
 "ID:1120053244"->{
  Cell[2517619, 67972, 3954, 113, 189, "FormalTextInputFormula",
   CellTags->{
    "ID:1120053244", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "71"},
   CellID->1120053244]},
 "71"->{
  Cell[2517619, 67972, 3954, 113, 189, "FormalTextInputFormula",
   CellTags->{
    "ID:1120053244", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "71"},
   CellID->1120053244]},
 "Proof|ID:1029764736"->{
  Cell[2521730, 68096, 10757, 245, 28, "ProofInfo",
   CellTags->{"Proof|ID:1029764736", "Proof|ID:1029764736-1"},
   CellID->59145356]},
 "Proof|ID:1029764736-1"->{
  Cell[2521730, 68096, 10757, 245, 28, "ProofInfo",
   CellTags->{"Proof|ID:1029764736", "Proof|ID:1029764736-1"},
   CellID->59145356]},
 "Proof|ID:1120053244"->{
  Cell[2533006, 68360, 16251, 353, 26, "ProofInfo",
   CellTags->{"Proof|ID:1120053244", "Proof|ID:1120053244-1"},
   CellID->334765073]},
 "Proof|ID:1120053244-1"->{
  Cell[2533006, 68360, 16251, 353, 26, "ProofInfo",
   CellTags->{"Proof|ID:1120053244", "Proof|ID:1120053244-1"},
   CellID->334765073]},
 "ID:25850532"->{
  Cell[2589609, 69938, 3750, 111, 93, "FormalTextInputFormula",
   CellTags->{
    "ID:25850532", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "72"},
   CellID->25850532]},
 "72"->{
  Cell[2589609, 69938, 3750, 111, 93, "FormalTextInputFormula",
   CellTags->{
    "ID:25850532", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "72"},
   CellID->25850532]},
 "Proof|ID:25850532"->{
  Cell[2593517, 70060, 12820, 286, 28, "ProofInfo",
   CellTags->{"Proof|ID:25850532", "Proof|ID:25850532-1"},
   CellID->193997362]},
 "Proof|ID:25850532-1"->{
  Cell[2593517, 70060, 12820, 286, 28, "ProofInfo",
   CellTags->{"Proof|ID:25850532", "Proof|ID:25850532-1"},
   CellID->193997362]},
 "ID:482796083"->{
  Cell[2633196, 71224, 1361, 38, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:482796083", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "74"},
   CellID->482796083]},
 "74"->{
  Cell[2633196, 71224, 1361, 38, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:482796083", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "74"},
   CellID->482796083]},
 "ID:1705968105"->{
  Cell[2634560, 71264, 1742, 49, 82, "FormalTextInputFormula",
   CellTags->{
    "ID:1705968105", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "75"},
   CellID->1705968105]},
 "75"->{
  Cell[2634560, 71264, 1742, 49, 82, "FormalTextInputFormula",
   CellTags->{
    "ID:1705968105", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "75"},
   CellID->1705968105]},
 "Proof|ID:482796083"->{
  Cell[2636457, 71324, 4337, 112, 28, "ProofInfo",
   CellTags->{"Proof|ID:482796083", "Proof|ID:482796083-1"},
   CellID->185660128]},
 "Proof|ID:482796083-1"->{
  Cell[2636457, 71324, 4337, 112, 28, "ProofInfo",
   CellTags->{"Proof|ID:482796083", "Proof|ID:482796083-1"},
   CellID->185660128]},
 "Proof|ID:1705968105"->{
  Cell[2641314, 71455, 6160, 149, 26, "ProofInfo",
   CellTags->{"Proof|ID:1705968105", "Proof|ID:1705968105-1"},
   CellID->213359418]},
 "Proof|ID:1705968105-1"->{
  Cell[2641314, 71455, 6160, 149, 26, "ProofInfo",
   CellTags->{"Proof|ID:1705968105", "Proof|ID:1705968105-1"},
   CellID->213359418]},
 "ID:68280999"->{
  Cell[2679069, 72583, 1341, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:68280999", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "73"},
   CellID->68280999]},
 "73"->{
  Cell[2679069, 72583, 1341, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:68280999", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "73"},
   CellID->68280999]},
 "Proof|ID:68280999"->{
  Cell[2680567, 72626, 3421, 92, 28, "ProofInfo",
   CellTags->{"Proof|ID:68280999", "Proof|ID:68280999-1"},
   CellID->20432459]},
 "Proof|ID:68280999-1"->{
  Cell[2680567, 72626, 3421, 92, 28, "ProofInfo",
   CellTags->{"Proof|ID:68280999", "Proof|ID:68280999-1"},
   CellID->20432459]},
 "ID:1188710397"->{
  Cell[2685048, 72752, 1257, 33, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1188710397", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Main Theorem"},
   CellID->1188710397]},
 "Main Theorem"->{
  Cell[2685048, 72752, 1257, 33, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:1188710397", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Main Theorem"},
   CellID->1188710397]},
 "Proof|ID:1188710397"->{
  Cell[2686461, 72796, 3456, 93, 28, "ProofInfo",
   CellTags->{"Proof|ID:1188710397", "Proof|ID:1188710397-1"},
   CellID->240809659]},
 "Proof|ID:1188710397-1"->{
  Cell[2686461, 72796, 3456, 93, 28, "ProofInfo",
   CellTags->{"Proof|ID:1188710397", "Proof|ID:1188710397-1"},
   CellID->240809659]},
 "ID:78242703"->{
  Cell[2690656, 72914, 780, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:78242703", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->78242703]},
 "ID:807712598"->{
  Cell[2691853, 72952, 1918, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:807712598", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "75 tuples"},
   CellID->807712598]},
 "75 tuples"->{
  Cell[2691853, 72952, 1918, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:807712598", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "75 tuples"},
   CellID->807712598]},
 "ID:645630189"->{
  Cell[2693774, 73010, 3950, 106, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:645630189", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "76 tuples"},
   CellID->645630189]},
 "76 tuples"->{
  Cell[2693774, 73010, 3950, 106, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:645630189", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "76 tuples"},
   CellID->645630189]},
 "ID:47992769"->{
  Cell[2697727, 73118, 1955, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47992769", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "77 tuples"},
   CellID->47992769]},
 "77 tuples"->{
  Cell[2697727, 73118, 1955, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47992769", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "77 tuples"},
   CellID->47992769]},
 "Proof|ID:807712598"->{
  Cell[2699839, 73188, 8198, 196, 70, "ProofInfo",
   CellTags->{"Proof|ID:807712598", "Proof|ID:807712598-1"},
   CellID->1012531353]},
 "Proof|ID:807712598-1"->{
  Cell[2699839, 73188, 8198, 196, 70, "ProofInfo",
   CellTags->{"Proof|ID:807712598", "Proof|ID:807712598-1"},
   CellID->1012531353]},
 "Proof|ID:645630189"->{
  Cell[2708558, 73403, 15922, 354, 70, "ProofInfo",
   CellTags->{"Proof|ID:645630189", "Proof|ID:645630189-1"},
   CellID->54547686]},
 "Proof|ID:645630189-1"->{
  Cell[2708558, 73403, 15922, 354, 70, "ProofInfo",
   CellTags->{"Proof|ID:645630189", "Proof|ID:645630189-1"},
   CellID->54547686]},
 "Proof|ID:47992769"->{
  Cell[2725000, 73776, 7514, 182, 70, "ProofInfo",
   CellTags->{"Proof|ID:47992769", "Proof|ID:47992769-1"},
   CellID->163065910]},
 "Proof|ID:47992769-1"->{
  Cell[2725000, 73776, 7514, 182, 70, "ProofInfo",
   CellTags->{"Proof|ID:47992769", "Proof|ID:47992769-1"},
   CellID->163065910]},
 "ID:41552115"->{
  Cell[2734185, 74013, 741, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:41552115", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->41552115]},
 "ID:208902120"->{
  Cell[2734929, 74032, 781, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:208902120", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->208902120]},
 "ID:531785290"->{
  Cell[2736136, 74068, 3835, 112, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:531785290", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal congruence"},
   CellID->531785290]},
 "ideal congruence"->{
  Cell[2736136, 74068, 3835, 112, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:531785290", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal congruence"},
   CellID->531785290]},
 "ID:1814440310"->{
  Cell[2740372, 74196, 1341, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1814440310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal"},
   CellID->1814440310]},
 "ideal"->{
  Cell[2740372, 74196, 1341, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1814440310", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal"},
   CellID->1814440310]},
 "ID:1625946592"->{
  Cell[2741716, 74234, 1240, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1625946592", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal of tuple"},
   CellID->1625946592]},
 "ideal of tuple"->{
  Cell[2741716, 74234, 1240, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1625946592", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal of tuple"},
   CellID->1625946592]},
 "ID:1782924340"->{
  Cell[2744193, 74308, 714, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1782924340", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1782924340]},
 "ID:1482545820"->{
  Cell[2744949, 74331, 783, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1482545820", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1482545820]},
 "ID:1796414707"->{
  Cell[2746430, 74376, 4190, 125, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1796414707", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "sum multiplication join"},
   CellID->1796414707]},
 "sum multiplication join"->{
  Cell[2746430, 74376, 4190, 125, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1796414707", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "sum multiplication join"},
   CellID->1796414707]},
 "Proof|ID:1796414707"->{
  Cell[2750777, 74512, 12573, 291, 70, "ProofInfo",
   CellTags->{"Proof|ID:1796414707", "Proof|ID:1796414707-1"},
   CellID->22030080]},
 "Proof|ID:1796414707-1"->{
  Cell[2750777, 74512, 12573, 291, 70, "ProofInfo",
   CellTags->{"Proof|ID:1796414707", "Proof|ID:1796414707-1"},
   CellID->22030080]},
 "ID:1254924191"->{
  Cell[2764330, 74838, 782, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1254924191", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1254924191]},
 "ID:923334447"->{
  Cell[2765115, 74860, 864, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:923334447", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[Congruent]"},
   CellID->923334447]},
 "reflexivity of \[Congruent]"->{
  Cell[2765115, 74860, 864, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:923334447", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[Congruent]"},
   CellID->923334447]},
 "ID:1712150230"->{
  Cell[2765982, 74885, 1182, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1712150230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of \[Congruent]"},
   CellID->1712150230]},
 "symmetry of \[Congruent]"->{
  Cell[2765982, 74885, 1182, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1712150230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of \[Congruent]"},
   CellID->1712150230]},
 "ID:1473809719"->{
  Cell[2767167, 74923, 1542, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1473809719", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[Congruent]"},
   CellID->1473809719]},
 "transitivity of \[Congruent]"->{
  Cell[2767167, 74923, 1542, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1473809719", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[Congruent]"},
   CellID->1473809719]},
 "Proof|ID:923334447"->{
  Cell[2768867, 74984, 3645, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:923334447", "Proof|ID:923334447-1"},
   CellID->1760520462]},
 "Proof|ID:923334447-1"->{
  Cell[2768867, 74984, 3645, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:923334447", "Proof|ID:923334447-1"},
   CellID->1760520462]},
 "Proof|ID:1712150230"->{
  Cell[2773033, 75101, 4797, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:1712150230", "Proof|ID:1712150230-1"},
   CellID->463691700]},
 "Proof|ID:1712150230-1"->{
  Cell[2773033, 75101, 4797, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:1712150230", "Proof|ID:1712150230-1"},
   CellID->463691700]},
 "Proof|ID:1473809719"->{
  Cell[2778351, 75243, 5984, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:1473809719", "Proof|ID:1473809719-1"},
   CellID->1487366676]},
 "Proof|ID:1473809719-1"->{
  Cell[2778351, 75243, 5984, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:1473809719", "Proof|ID:1473809719-1"},
   CellID->1487366676]},
 "ID:306651916"->{
  Cell[2785797, 75442, 6867, 189, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:306651916", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] alternative ind.1"},
   CellID->306651916]},
 "\[Congruent] alternative ind.1"->{
  Cell[2785797, 75442, 6867, 189, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:306651916", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] alternative ind.1"},
   CellID->306651916]},
 "ID:571707933"->{
  Cell[2792667, 75633, 7978, 214, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:571707933", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] alternative ind.2"},
   CellID->571707933]},
 "\[Congruent] alternative ind.2"->{
  Cell[2792667, 75633, 7978, 214, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:571707933", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] alternative ind.2"},
   CellID->571707933]},
 "Proof|ID:306651916"->{
  Cell[2800802, 75858, 17447, 391, 70, "ProofInfo",
   CellTags->{"Proof|ID:306651916", "Proof|ID:306651916-1"},
   CellID->655455289]},
 "Proof|ID:306651916-1"->{
  Cell[2800802, 75858, 17447, 391, 70, "ProofInfo",
   CellTags->{"Proof|ID:306651916", "Proof|ID:306651916-1"},
   CellID->655455289]},
 "Proof|ID:571707933"->{
  Cell[2818833, 76269, 19938, 443, 70, "ProofInfo",
   CellTags->{"Proof|ID:571707933", "Proof|ID:571707933-1"},
   CellID->1209243937]},
 "Proof|ID:571707933-1"->{
  Cell[2818833, 76269, 19938, 443, 70, "ProofInfo",
   CellTags->{"Proof|ID:571707933", "Proof|ID:571707933-1"},
   CellID->1209243937]},
 "ID:473420099"->{
  Cell[2839897, 76752, 771, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:473420099", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->473420099]},
 "ID:544141308"->{
  Cell[2840671, 76774, 1216, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:544141308", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[SubsetEqual] \[Congruent]"},
   CellID->544141308]},
 "\[RightArrow] \[SubsetEqual] \[Congruent]"->{
  Cell[2840671, 76774, 1216, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:544141308", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[SubsetEqual] \[Congruent]"},
   CellID->544141308]},
 "ID:1307557419"->{
  Cell[2841890, 76813, 1245, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1307557419", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] \[SubsetEqual] \[Congruent]"},
   CellID->1307557419]},
 "\[LeftRightArrow] \[SubsetEqual] \[Congruent]"->{
  Cell[2841890, 76813, 1245, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1307557419", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] \[SubsetEqual] \[Congruent]"},
   CellID->1307557419]},
 "ID:170897044"->{
  Cell[2843138, 76853, 2435, 75, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:170897044", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility \[SubsetEqual] \[Congruent] ind"},
   CellID->170897044]},
 "connectibility \[SubsetEqual] \[Congruent] ind"->{
  Cell[2843138, 76853, 2435, 75, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:170897044", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility \[SubsetEqual] \[Congruent] ind"},
   CellID->170897044]},
 "ID:720101359"->{
  Cell[2845576, 76930, 1245, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:720101359", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility \[SubsetEqual] \[Congruent]"},
   CellID->720101359]},
 "connectibility \[SubsetEqual] \[Congruent]"->{
  Cell[2845576, 76930, 1245, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:720101359", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility \[SubsetEqual] \[Congruent]"},
   CellID->720101359]},
 "Proof|ID:544141308"->{
  Cell[2846977, 76979, 4826, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:544141308", "Proof|ID:544141308-1"},
   CellID->8057470]},
 "Proof|ID:544141308-1"->{
  Cell[2846977, 76979, 4826, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:544141308", "Proof|ID:544141308-1"},
   CellID->8057470]},
 "Proof|ID:1307557419"->{
  Cell[2852323, 77122, 4866, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:1307557419", "Proof|ID:1307557419-1"},
   CellID->1073556871]},
 "Proof|ID:1307557419-1"->{
  Cell[2852323, 77122, 4866, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:1307557419", "Proof|ID:1307557419-1"},
   CellID->1073556871]},
 "Proof|ID:170897044"->{
  Cell[2857710, 77266, 7587, 190, 70, "ProofInfo",
   CellTags->{"Proof|ID:170897044", "Proof|ID:170897044-1"},
   CellID->2702629]},
 "Proof|ID:170897044-1"->{
  Cell[2857710, 77266, 7587, 190, 70, "ProofInfo",
   CellTags->{"Proof|ID:170897044", "Proof|ID:170897044-1"},
   CellID->2702629]},
 "Proof|ID:720101359"->{
  Cell[2865818, 77475, 4998, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:720101359", "Proof|ID:720101359-1"},
   CellID->896819044]},
 "Proof|ID:720101359-1"->{
  Cell[2865818, 77475, 4998, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:720101359", "Proof|ID:720101359-1"},
   CellID->896819044]},
 "ID:1680644099"->{
  Cell[2871888, 77639, 2147, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1680644099", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "43"},
   CellID->1680644099]},
 "43"->{
  Cell[2871888, 77639, 2147, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1680644099", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "43"},
   CellID->1680644099]},
 "ID:555151129"->{
  Cell[2874038, 77707, 4727, 134, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:555151129", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] \[SubsetEqual] connectibility ind"},
   CellID->555151129]},
 "\[Congruent] \[SubsetEqual] connectibility ind"->{
  Cell[2874038, 77707, 4727, 134, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:555151129", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] \[SubsetEqual] connectibility ind"},
   CellID->555151129]},
 "ID:2104087526"->{
  Cell[2878768, 77843, 1545, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2104087526", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] \[SubsetEqual] connectibility"},
   CellID->2104087526]},
 "\[Congruent] \[SubsetEqual] connectibility"->{
  Cell[2878768, 77843, 1545, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2104087526", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] \[SubsetEqual] connectibility"},
   CellID->2104087526]},
 "Proof|ID:1680644099"->{
  Cell[2880469, 77903, 6382, 157, 70, "ProofInfo",
   CellTags->{"Proof|ID:1680644099", "Proof|ID:1680644099-1"},
   CellID->279660360]},
 "Proof|ID:1680644099-1"->{
  Cell[2880469, 77903, 6382, 157, 70, "ProofInfo",
   CellTags->{"Proof|ID:1680644099", "Proof|ID:1680644099-1"},
   CellID->279660360]},
 "Proof|ID:555151129"->{
  Cell[2887372, 78079, 13091, 298, 70, "ProofInfo",
   CellTags->{"Proof|ID:555151129", "Proof|ID:555151129-1"},
   CellID->440681758]},
 "Proof|ID:555151129-1"->{
  Cell[2887372, 78079, 13091, 298, 70, "ProofInfo",
   CellTags->{"Proof|ID:555151129", "Proof|ID:555151129-1"},
   CellID->440681758]},
 "Proof|ID:2104087526"->{
  Cell[2900983, 78396, 5008, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:2104087526", "Proof|ID:2104087526-1"},
   CellID->1134450301]},
 "Proof|ID:2104087526-1"->{
  Cell[2900983, 78396, 5008, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:2104087526", "Proof|ID:2104087526-1"},
   CellID->1134450301]},
 "ID:980108134"->{
  Cell[2907063, 78561, 1520, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:980108134", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent]=connectibility"},
   CellID->980108134]},
 "\[Congruent]=connectibility"->{
  Cell[2907063, 78561, 1520, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:980108134", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent]=connectibility"},
   CellID->980108134]},
 "Proof|ID:980108134"->{
  Cell[2908740, 78620, 4978, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:980108134", "Proof|ID:980108134-1"},
   CellID->699177716]},
 "Proof|ID:980108134-1"->{
  Cell[2908740, 78620, 4978, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:980108134", "Proof|ID:980108134-1"},
   CellID->699177716]},
 "ID:2002188699"->{
  Cell[2914888, 78785, 770, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2002188699", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2002188699]},
 "ID:160768319"->{
  Cell[2915661, 78807, 3086, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160768319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0", "IR"},
   CellID->160768319]},
 "\[Congruent] 0"->{
  Cell[2915661, 78807, 3086, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160768319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0", "IR"},
   CellID->160768319]},
 "ID:562483616"->{
  Cell[2918750, 78900, 962, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:562483616", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] \[Congruent] 0"},
   CellID->562483616]},
 "\[Element] \[Congruent] 0"->{
  Cell[2918750, 78900, 962, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:562483616", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] \[Congruent] 0"},
   CellID->562483616]},
 "ID:853458295"->{
  Cell[2919715, 78927, 1145, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:853458295", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 \[RawStar]"},
   CellID->853458295]},
 "\[Congruent] 0 \[RawStar]"->{
  Cell[2919715, 78927, 1145, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:853458295", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 \[RawStar]"},
   CellID->853458295]},
 "ID:462880889"->{
  Cell[2920863, 78960, 1063, 30, 70, "GlobalDeclaration",
   CellTags->{
    "ID:462880889", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->462880889]},
 "ID:44331059"->{
  Cell[2921929, 78992, 989, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:44331059", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 +"},
   CellID->44331059]},
 "\[Congruent] 0 +"->{
  Cell[2921929, 78992, 989, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:44331059", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 +"},
   CellID->44331059]},
 "ID:1958399498"->{
  Cell[2922921, 79021, 969, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1958399498", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 -"},
   CellID->1958399498]},
 "\[Congruent] 0 -"->{
  Cell[2922921, 79021, 969, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1958399498", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 -"},
   CellID->1958399498]},
 "Proof|ID:160768319"->{
  Cell[2924048, 79058, 9599, 226, 70, "ProofInfo",
   CellTags->{"Proof|ID:160768319", "Proof|ID:160768319-1"},
   CellID->149045731]},
 "Proof|ID:160768319-1"->{
  Cell[2924048, 79058, 9599, 226, 70, "ProofInfo",
   CellTags->{"Proof|ID:160768319", "Proof|ID:160768319-1"},
   CellID->149045731]},
 "Proof|ID:562483616"->{
  Cell[2934167, 79303, 4154, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:562483616", "Proof|ID:562483616-1"},
   CellID->929707542]},
 "Proof|ID:562483616-1"->{
  Cell[2934167, 79303, 4154, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:562483616", "Proof|ID:562483616-1"},
   CellID->929707542]},
 "Proof|ID:853458295"->{
  Cell[2938842, 79432, 5410, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:853458295", "Proof|ID:853458295-1"},
   CellID->282680859]},
 "Proof|ID:853458295-1"->{
  Cell[2938842, 79432, 5410, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:853458295", "Proof|ID:853458295-1"},
   CellID->282680859]},
 "Proof|ID:44331059"->{
  Cell[2944773, 79589, 6306, 158, 70, "ProofInfo",
   CellTags->{"Proof|ID:44331059", "Proof|ID:44331059-1"},
   CellID->2113222684]},
 "Proof|ID:44331059-1"->{
  Cell[2944773, 79589, 6306, 158, 70, "ProofInfo",
   CellTags->{"Proof|ID:44331059", "Proof|ID:44331059-1"},
   CellID->2113222684]},
 "Proof|ID:1958399498"->{
  Cell[2951598, 79766, 6320, 159, 70, "ProofInfo",
   CellTags->{"Proof|ID:1958399498", "Proof|ID:1958399498-1"},
   CellID->1499077166]},
 "Proof|ID:1958399498-1"->{
  Cell[2951598, 79766, 6320, 159, 70, "ProofInfo",
   CellTags->{"Proof|ID:1958399498", "Proof|ID:1958399498-1"},
   CellID->1499077166]},
 "ID:2069217266"->{
  Cell[2958884, 79958, 912, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2069217266", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2069217266]},
 "ID:277446168"->{
  Cell[2959799, 79984, 6348, 176, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:277446168", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] extended ind"},
   CellID->277446168]},
 "\[Congruent] extended ind"->{
  Cell[2959799, 79984, 6348, 176, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:277446168", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] extended ind"},
   CellID->277446168]},
 "ID:1459665988"->{
  Cell[2966150, 80162, 1608, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1459665988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] extended"},
   CellID->1459665988]},
 "\[Congruent] extended"->{
  Cell[2966150, 80162, 1608, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1459665988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] extended"},
   CellID->1459665988]},
 "Proof|ID:277446168"->{
  Cell[2967915, 80223, 17580, 394, 70, "ProofInfo",
   CellTags->{"Proof|ID:277446168", "Proof|ID:277446168-1"},
   CellID->37830858]},
 "Proof|ID:277446168-1"->{
  Cell[2967915, 80223, 17580, 394, 70, "ProofInfo",
   CellTags->{"Proof|ID:277446168", "Proof|ID:277446168-1"},
   CellID->37830858]},
 "Proof|ID:1459665988"->{
  Cell[2986015, 80636, 6431, 159, 70, "ProofInfo",
   CellTags->{"Proof|ID:1459665988", "Proof|ID:1459665988-1"},
   CellID->569980310]},
 "Proof|ID:1459665988-1"->{
  Cell[2986015, 80636, 6431, 159, 70, "ProofInfo",
   CellTags->{"Proof|ID:1459665988", "Proof|ID:1459665988-1"},
   CellID->569980310]},
 "ID:1670835483"->{
  Cell[2993233, 80822, 716, 18, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1670835483", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1670835483]},
 "ID:97617962"->{
  Cell[2994165, 80849, 780, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:97617962", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->97617962]},
 "ID:1452952621"->{
  Cell[2994988, 80874, 753, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1452952621", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1452952621]},
 "ID:1760501242"->{
  Cell[2996206, 80912, 976, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1760501242", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets ideal", "IR"},
   CellID->1760501242]},
 "DomainSets ideal"->{
  Cell[2996206, 80912, 976, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1760501242", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets ideal", "IR"},
   CellID->1760501242]},
 "ID:554864723"->{
  Cell[2997185, 80939, 871, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:554864723", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal isSet", "IR"},
   CellID->554864723]},
 "ideal isSet"->{
  Cell[2997185, 80939, 871, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:554864723", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal isSet", "IR"},
   CellID->554864723]},
 "ID:739591353"->{
  Cell[2998059, 80963, 1496, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739591353", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] ideal"},
   CellID->739591353]},
 "\[Element] ideal"->{
  Cell[2998059, 80963, 1496, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:739591353", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] ideal"},
   CellID->739591353]},
 "ID:2030275451"->{
  Cell[2999558, 81012, 786, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2030275451", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->2030275451]},
 "ID:1644449200"->{
  Cell[3000347, 81033, 847, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1644449200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideals closed +", "IR"},
   CellID->1644449200]},
 "ideals closed +"->{
  Cell[3000347, 81033, 847, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1644449200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideals closed +", "IR"},
   CellID->1644449200]},
 "ID:1500033871"->{
  Cell[3001197, 81058, 861, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1500033871", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideals closed - binary", "IR"},
   CellID->1500033871]},
 "ideals closed - binary"->{
  Cell[3001197, 81058, 861, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1500033871", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideals closed - binary", "IR"},
   CellID->1500033871]},
 "ID:1114678876"->{
  Cell[3002061, 81083, 1026, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1114678876", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal \[Union] singleton"},
   CellID->1114678876]},
 "ideal \[Union] singleton"->{
  Cell[3002061, 81083, 1026, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1114678876", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal \[Union] singleton"},
   CellID->1114678876]},
 "ID:1895756686"->{
  Cell[3003090, 81112, 1822, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1895756686", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal \[RightArrow]\[RawStar]"},
   CellID->1895756686]},
 "ideal \[RightArrow]\[RawStar]"->{
  Cell[3003090, 81112, 1822, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1895756686", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal \[RightArrow]\[RawStar]"},
   CellID->1895756686]},
 "Proof|ID:1760501242"->{
  Cell[3005068, 81181, 3374, 93, 70, "ProofInfo",
   CellTags->{"Proof|ID:1760501242", "Proof|ID:1760501242-1"},
   CellID->137891172]},
 "Proof|ID:1760501242-1"->{
  Cell[3005068, 81181, 3374, 93, 70, "ProofInfo",
   CellTags->{"Proof|ID:1760501242", "Proof|ID:1760501242-1"},
   CellID->137891172]},
 "Proof|ID:554864723"->{
  Cell[3008962, 81293, 3034, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:554864723", "Proof|ID:554864723-1"},
   CellID->73662413]},
 "Proof|ID:554864723-1"->{
  Cell[3008962, 81293, 3034, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:554864723", "Proof|ID:554864723-1"},
   CellID->73662413]},
 "Proof|ID:739591353"->{
  Cell[3012516, 81396, 4835, 122, 70, "ProofInfo",
   CellTags->{"Proof|ID:739591353", "Proof|ID:739591353-1"},
   CellID->136162227]},
 "Proof|ID:739591353-1"->{
  Cell[3012516, 81396, 4835, 122, 70, "ProofInfo",
   CellTags->{"Proof|ID:739591353", "Proof|ID:739591353-1"},
   CellID->136162227]},
 "Proof|ID:1644449200"->{
  Cell[3017871, 81537, 4696, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1644449200", "Proof|ID:1644449200-1"},
   CellID->535086304]},
 "Proof|ID:1644449200-1"->{
  Cell[3017871, 81537, 4696, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1644449200", "Proof|ID:1644449200-1"},
   CellID->535086304]},
 "Proof|ID:1500033871"->{
  Cell[3023088, 81675, 4710, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1500033871", "Proof|ID:1500033871-1"},
   CellID->768137348]},
 "Proof|ID:1500033871-1"->{
  Cell[3023088, 81675, 4710, 119, 70, "ProofInfo",
   CellTags->{"Proof|ID:1500033871", "Proof|ID:1500033871-1"},
   CellID->768137348]},
 "Proof|ID:1114678876"->{
  Cell[3028319, 81813, 4321, 113, 70, "ProofInfo",
   CellTags->{"Proof|ID:1114678876", "Proof|ID:1114678876-1"},
   CellID->1601029605]},
 "Proof|ID:1114678876-1"->{
  Cell[3028319, 81813, 4321, 113, 70, "ProofInfo",
   CellTags->{"Proof|ID:1114678876", "Proof|ID:1114678876-1"},
   CellID->1601029605]},
 "Proof|ID:1895756686"->{
  Cell[3033161, 81945, 6549, 155, 70, "ProofInfo",
   CellTags->{"Proof|ID:1895756686", "Proof|ID:1895756686-1"},
   CellID->1682672380]},
 "Proof|ID:1895756686-1"->{
  Cell[3033161, 81945, 6549, 155, 70, "ProofInfo",
   CellTags->{"Proof|ID:1895756686", "Proof|ID:1895756686-1"},
   CellID->1682672380]},
 "ID:1581989517"->{
  Cell[3040491, 82126, 839, 21, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1581989517", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1581989517]},
 "ID:185976047"->{
  Cell[3041371, 82152, 801, 21, 43, "GlobalDeclaration",
   CellTags->{
    "ID:185976047", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->185976047]},
 "ID:1918127619"->{
  Cell[3042628, 82191, 1600, 48, 45, "FormalTextInputFormula",
   CellTags->{
    "ID:1918127619", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] tuple-ideal"},
   CellID->1918127619]},
 "\[Element] tuple-ideal"->{
  Cell[3042628, 82191, 1600, 48, 45, "FormalTextInputFormula",
   CellTags->{
    "ID:1918127619", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] tuple-ideal"},
   CellID->1918127619]},
 "ID:279213832"->{
  Cell[3044231, 82241, 1009, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:279213832", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets tuple-ideal", "IR"},
   CellID->279213832]},
 "DomainSets tuple-ideal"->{
  Cell[3044231, 82241, 1009, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:279213832", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets tuple-ideal", "IR"},
   CellID->279213832]},
 "ID:1225983506"->{
  Cell[3045243, 82268, 814, 20, 31, "GlobalDeclaration",
   CellTags->{
    "ID:1225983506", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
   CellID->1225983506]},
 "ID:1587050664"->{
  Cell[3046060, 82290, 885, 24, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1587050664", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideals closed +", "IR"},
   CellID->1587050664]},
 "tuple-ideals closed +"->{
  Cell[3046060, 82290, 885, 24, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1587050664", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideals closed +", "IR"},
   CellID->1587050664]},
 "ID:508704301"->{
  Cell[3046948, 82316, 896, 24, 29, "FormalTextInputFormula",
   CellTags->{
    "ID:508704301", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideals closed - binary", "IR"},
   CellID->508704301]},
 "tuple-ideals closed - binary"->{
  Cell[3046948, 82316, 896, 24, 29, "FormalTextInputFormula",
   CellTags->{
    "ID:508704301", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideals closed - binary", "IR"},
   CellID->508704301]},
 "ID:1181234284"->{
  Cell[3047847, 82342, 1085, 29, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1181234284", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideal append"},
   CellID->1181234284]},
 "tuple-ideal append"->{
  Cell[3047847, 82342, 1085, 29, 31, "FormalTextInputFormula",
   CellTags->{
    "ID:1181234284", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideal append"},
   CellID->1181234284]},
 "ID:352062174"->{
  Cell[3048935, 82373, 1904, 59, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:352062174", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideal \[RightArrow]\[RawStar]"},
   CellID->352062174]},
 "tuple-ideal \[RightArrow]\[RawStar]"->{
  Cell[3048935, 82373, 1904, 59, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:352062174", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideal \[RightArrow]\[RawStar]"},
   CellID->352062174]},
 "Proof|ID:1918127619"->{
  Cell[3050997, 82443, 4970, 125, 28, "ProofInfo",
   CellTags->{"Proof|ID:1918127619", "Proof|ID:1918127619-1"},
   CellID->783030095]},
 "Proof|ID:1918127619-1"->{
  Cell[3050997, 82443, 4970, 125, 28, "ProofInfo",
   CellTags->{"Proof|ID:1918127619", "Proof|ID:1918127619-1"},
   CellID->783030095]},
 "Proof|ID:279213832"->{
  Cell[3056489, 82587, 3381, 92, 26, "ProofInfo",
   CellTags->{"Proof|ID:279213832", "Proof|ID:279213832-1"},
   CellID->889748810]},
 "Proof|ID:279213832-1"->{
  Cell[3056489, 82587, 3381, 92, 26, "ProofInfo",
   CellTags->{"Proof|ID:279213832", "Proof|ID:279213832-1"},
   CellID->889748810]},
 "Proof|ID:1587050664"->{
  Cell[3060390, 82698, 4713, 119, 26, "ProofInfo",
   CellTags->{"Proof|ID:1587050664", "Proof|ID:1587050664-1"},
   CellID->1831236240]},
 "Proof|ID:1587050664-1"->{
  Cell[3060390, 82698, 4713, 119, 26, "ProofInfo",
   CellTags->{"Proof|ID:1587050664", "Proof|ID:1587050664-1"},
   CellID->1831236240]},
 "Proof|ID:508704301"->{
  Cell[3065625, 82836, 4726, 119, 26, "ProofInfo",
   CellTags->{"Proof|ID:508704301", "Proof|ID:508704301-1"},
   CellID->1768992322]},
 "Proof|ID:508704301-1"->{
  Cell[3065625, 82836, 4726, 119, 26, "ProofInfo",
   CellTags->{"Proof|ID:508704301", "Proof|ID:508704301-1"},
   CellID->1768992322]},
 "Proof|ID:1181234284"->{
  Cell[3070872, 82974, 4208, 109, 26, "ProofInfo",
   CellTags->{"Proof|ID:1181234284", "Proof|ID:1181234284-1"},
   CellID->508218087]},
 "Proof|ID:1181234284-1"->{
  Cell[3070872, 82974, 4208, 109, 26, "ProofInfo",
   CellTags->{"Proof|ID:1181234284", "Proof|ID:1181234284-1"},
   CellID->508218087]},
 "Proof|ID:352062174"->{
  Cell[3075601, 83102, 6555, 154, 26, "ProofInfo",
   CellTags->{"Proof|ID:352062174", "Proof|ID:352062174-1"},
   CellID->873828372]},
 "Proof|ID:352062174-1"->{
  Cell[3075601, 83102, 6555, 154, 26, "ProofInfo",
   CellTags->{"Proof|ID:352062174", "Proof|ID:352062174-1"},
   CellID->873828372]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"ID:2066005206", 3202379, 86696},
 {"Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 3202676, 86704},
 {"ID:457420627", 3305259, 89399},
 {"ID:176465234", 3305485, 89406},
 {"ID:478302118", 3305711, 89413},
 {"reduction with several multipliers", 3306002, 89420},
 {"ID:448370877", 3306271, 89427},
 {"symmetric closure of reducibility", 3306561, 89434},
 {"ID:1537297721", 3306831, 89441},
 {"reducible1", 3307077, 89448},
 {"ID:115218056", 3307325, 89455},
 {"reducible2", 3307569, 89462},
 {"ID:399781455", 3307815, 89469},
 {"irreducibility", 3308067, 89476},
 {"ID:1867242863", 3308318, 89483},
 {"irrelative", 3308564, 89490},
 {"ID:1834165766", 3308813, 89497},
 {"correlative", 3309061, 89504},
 {"ID:1907029362", 3309311, 89511},
 {"common reducible", 3309569, 89518},
 {"ID:2013437879", 3309824, 89525},
 {"non-trivial common reducible", 3310107, 89532},
 {"ID:991122907", 3310374, 89539},
 {"minimal non-trivial common reducible", 3310670, 89546},
 {"ID:1293246145", 3310943, 89553},
 {"unary non-trivial common reducible", 3311238, 89560},
 {"ID:894750987", 3311511, 89567},
 {"unary minimal non-trivial common reducible", 3311819, 89574},
 {"ID:779388596", 3312097, 89581},
 {"critical pair multipliers", 3312372, 89588},
 {"ID:1517092650", 3312635, 89595},
 {"ID:935236171", 3312864, 89602},
 {"reduction modulo element using multiplier", 3313170, 89609},
 {"ID:320624715", 3313447, 89616},
 {"reduction modulo set", 3313711, 89623},
 {"ID:434515171", 3313967, 89630},
 {"reduction modulo tuple", 3314235, 89637},
 {"ID:773828270", 3314493, 89644},
 {"isReductor", 3314737, 89651},
 {"ID:1813157585", 3314984, 89658},
 {"reflexive-transitive closure of reduction relation", 3315316, 89666},
 {"ID:370202704", 3315610, 89674},
 {"symmetric closure of reduction relation", 3315912, 89681},
 {"ID:992563737", 3316187, 89688},
 {"symmetric-reflexive-transitive closure of reduction relation", 3316537, \
89696},
 {"ID:414012405", 3316839, 89704},
 {"common successor w.r.t. reduction relation", 3317147, 89711},
 {"ID:895393520", 3317425, 89718},
 {"reflexive-transitive closure of common successor", 3317745, 89725},
 {"ID:586157110", 3318029, 89732},
 {"common predecessor w.r.t. reduction relation", 3318341, 89739},
 {"ID:2045822907", 3318622, 89746},
 {"reflexive-transitive closure of common predecessor", 3318954, 89754},
 {"ID:463604795", 3319248, 89762},
 {"connectibility below", 3319513, 89769},
 {"ID:1818365967", 3319771, 89776},
 {"\[Tilde] multiplier", 3320037, 89783},
 {"ID:1533920470", 3320297, 89790},
 {"\[Tilde]", 3320539, 89797},
 {"ID:81135181", 3320784, 89804},
 {"ID:1593192085", 3321009, 89811},
 {"red", 3321241, 89818},
 {"ID:808948861", 3321482, 89825},
 {"\[SquareSubset]", 3321736, 89832},
 {"ID:301449743", 3321987, 89839},
 {"ID:218271861", 3322213, 89846},
 {"red tuples", 3322456, 89853},
 {"ID:20184186", 3322700, 89860},
 {"\[SquareSubset] tuples", 3322966, 89867},
 {"ID:2086101595", 3323223, 89874},
 {"ID:690041737", 3323451, 89881},
 {"DomainSets isReductor", 3323716, 89888},
 {"ID:5128300", 3323970, 89895},
 {"DomainTuples isReductor", 3324235, 89902},
 {"Proof|ID:690041737", 3324495, 89909},
 {"Proof|ID:690041737-1", 3324655, 89913},
 {"Proof|ID:5128300", 3324811, 89917},
 {"Proof|ID:5128300-1", 3324965, 89921},
 {"ID:906005529", 3325113, 89925},
 {"reduction modulo singleton", 3325395, 89932},
 {"IR", 3325653, 89939},
 {"Proof|ID:906005529", 3330987, 90068},
 {"Proof|ID:906005529-1", 3331148, 90072},
 {"ID:708647877", 3331301, 90076},
 {"ID:508075012", 3331527, 90083},
 {"ID:347299573", 3331753, 90090},
 {"reduction modulo tuple alternative", 3332051, 90097},
 {"ID:1756538453", 3332328, 90104},
 {"ID:200776309", 3332556, 90111},
 {"\[RightArrow] TupleToSet", 3332834, 90118},
 {"ID:989570066", 3333100, 90125},
 {"\[LeftRightArrow] TupleToSet", 3333386, 90132},
 {"ID:359412397", 3333656, 90139},
 {"\[RightArrow]\[RawStar] TupleToSet", 3333954, 90146},
 {"ID:1876144273", 3334231, 90153},
 {"\[LeftRightArrow]\[RawStar] TupleToSet", 3334539, 90160},
 {"ID:495509420", 3334821, 90167},
 {"connectibility below TupleToSet", 3335113, 90174},
 {"Proof|ID:347299573", 3335392, 90181},
 {"Proof|ID:347299573-1", 3335553, 90185},
 {"Proof|ID:200776309", 3335712, 90189},
 {"Proof|ID:200776309-1", 3335873, 90193},
 {"Proof|ID:989570066", 3336032, 90197},
 {"Proof|ID:989570066-1", 3336193, 90201},
 {"Proof|ID:359412397", 3336352, 90205},
 {"Proof|ID:359412397-1", 3336514, 90209},
 {"Proof|ID:1876144273", 3336675, 90213},
 {"Proof|ID:1876144273-1", 3336839, 90217},
 {"Proof|ID:495509420", 3337000, 90221},
 {"Proof|ID:495509420-1", 3337161, 90225},
 {"ID:467637718", 3337314, 90229},
 {"ID:1182167952", 3337541, 90236},
 {"ID:178557746", 3337769, 90243},
 {"reduction modulo set alternative", 3338063, 90250},
 {"ID:1019233204", 3338338, 90257},
 {"\[RightArrow] \[Union]", 3338608, 90264},
 {"ID:331036245", 3338868, 90271},
 {"\[LeftRightArrow] \[Union]", 3339144, 90278},
 {"ID:78564808", 3339405, 90285},
 {"\[RightArrow]\[RawStar] \[Union]", 3339691, 90292},
 {"ID:626278069", 3339957, 90299},
 {"\[LeftRightArrow]\[RawStar] \[Union]", 3340253, 90306},
 {"ID:356849840", 3340525, 90313},
 {"connectibility below \[Union]", 3340807, 90320},
 {"Proof|ID:178557746", 3341078, 90327},
 {"Proof|ID:178557746-1", 3341238, 90331},
 {"Proof|ID:1019233204", 3341397, 90335},
 {"Proof|ID:1019233204-1", 3341559, 90339},
 {"Proof|ID:331036245", 3341718, 90343},
 {"Proof|ID:331036245-1", 3341879, 90347},
 {"Proof|ID:78564808", 3342037, 90351},
 {"Proof|ID:78564808-1", 3342195, 90355},
 {"Proof|ID:626278069", 3342352, 90359},
 {"Proof|ID:626278069-1", 3342514, 90363},
 {"Proof|ID:356849840", 3342674, 90367},
 {"Proof|ID:356849840-1", 3342835, 90371},
 {"ID:366153355", 3342988, 90375},
 {"ID:221340803", 3343214, 90382},
 {"ID:328195853", 3343440, 90389},
 {"2", 3343666, 90396},
 {"ID:1692465705", 3343904, 90403},
 {"3", 3344131, 90410},
 {"ID:188743948", 3344369, 90417},
 {"4", 3344594, 90424},
 {"ID:115635911", 3344830, 90431},
 {"5", 3345056, 90438},
 {"ID:1118510950", 3345294, 90445},
 {"isCommonReducible symmetric", 3345574, 90452},
 {"ID:113485898", 3345839, 90459},
 {"isNTCR symmetric", 3346095, 90466},
 {"ID:2056330192", 3346348, 90473},
 {"isMNTCR symmetric", 3346608, 90480},
 {"ID:463339584", 3346863, 90487},
 {"areCPMultipliers symmetric", 3347139, 90494},
 {"ID:449270875", 3347401, 90501},
 {"areCPMultipliers sufficient", 3347679, 90508},
 {"Proof|ID:328195853", 3347948, 90515},
 {"Proof|ID:328195853-1", 3348109, 90519},
 {"Proof|ID:1692465705", 3348269, 90523},
 {"Proof|ID:1692465705-1", 3348433, 90527},
 {"Proof|ID:188743948", 3348594, 90531},
 {"Proof|ID:188743948-1", 3348756, 90535},
 {"Proof|ID:115635911", 3348916, 90539},
 {"Proof|ID:115635911-1", 3349078, 90543},
 {"Proof|ID:1118510950", 3349239, 90547},
 {"Proof|ID:1118510950-1", 3349403, 90551},
 {"Proof|ID:113485898", 3349564, 90555},
 {"Proof|ID:113485898-1", 3349726, 90559},
 {"Proof|ID:2056330192", 3349887, 90563},
 {"Proof|ID:2056330192-1", 3350052, 90567},
 {"Proof|ID:463339584", 3350214, 90571},
 {"Proof|ID:463339584-1", 3350376, 90575},
 {"Proof|ID:449270875", 3350536, 90579},
 {"Proof|ID:449270875-1", 3350698, 90583},
 {"ID:523863903", 3350852, 90587},
 {"ID:1519187240", 3351079, 90594},
 {"ID:1878546324", 3351308, 90601},
 {"2'", 3351538, 90608},
 {"ID:1781176225", 3351779, 90615},
 {"3'", 3352009, 90622},
 {"ID:1582995786", 3352250, 90629},
 {"4'", 3352480, 90636},
 {"ID:2020067658", 3352721, 90643},
 {"5'", 3352951, 90650},
 {"Proof|ID:1878546324", 3353198, 90657},
 {"Proof|ID:1878546324-1", 3353363, 90661},
 {"Proof|ID:1781176225", 3353526, 90665},
 {"Proof|ID:1781176225-1", 3353692, 90669},
 {"Proof|ID:1582995786", 3353856, 90673},
 {"Proof|ID:1582995786-1", 3354022, 90677},
 {"Proof|ID:2020067658", 3354186, 90681},
 {"Proof|ID:2020067658-1", 3354352, 90685},
 {"ID:1823284442", 3354510, 90689},
 {"ID:284217868", 3354739, 90696},
 {"ID:628244416", 3354966, 90703},
 {"6", 3355193, 90710},
 {"ID:1255486983", 3355432, 90717},
 {"correlative-irrelative 2", 3355707, 90724},
 {"Proof|ID:628244416", 3355976, 90731},
 {"Proof|ID:628244416-1", 3356139, 90735},
 {"Proof|ID:1255486983", 3356301, 90739},
 {"Proof|ID:1255486983-1", 3356466, 90743},
 {"ID:1382908475", 3356623, 90747},
 {"ID:442895874", 3356852, 90754},
 {"ID:927154658", 3357079, 90761},
 {"ID:1300106055", 3357307, 90768},
 {"reduceM 0", 3357551, 90775},
 {"ID:438711618", 3357798, 90782},
 {"reduceM 1", 3358041, 90789},
 {"ID:2101232329", 3358288, 90796},
 {"reduceM alternative", 3358553, 90803},
 {"ID:96556784", 3358810, 90810},
 {"reduceM+1", 3359051, 90817},
 {"ID:1465630278", 3359296, 90824},
 {"reduceM alternative II", 3359567, 90831},
 {"ID:1952076718", 3359829, 90838},
 {"membership reduceM", 3360098, 90845},
 {"ID:1110504237", 3360362, 90852},
 {"ID:1236615595", 3360592, 90859},
 {"reduceM-1", 3360837, 90866},
 {"ID:203149419", 3361085, 90873},
 {"reduceM-1 II", 3361334, 90880},
 {"ID:333495152", 3361583, 90887},
 {"isReducible reduceM", 3361846, 90894},
 {"ID:332101094", 3362102, 90901},
 {"symred reduceM", 3362355, 90908},
 {"ID:296464255", 3362606, 90915},
 {"symred reduceM I", 3362863, 90922},
 {"ID:194930655", 3363116, 90929},
 {"symred reduceM II", 3363375, 90936},
 {"ID:935171195", 3363629, 90943},
 {"symred reduceM III", 3363890, 90950},
 {"Proof|ID:1300106055", 3364152, 90957},
 {"Proof|ID:1300106055-1", 3364317, 90961},
 {"Proof|ID:438711618", 3364479, 90965},
 {"Proof|ID:438711618-1", 3364641, 90969},
 {"Proof|ID:2101232329", 3364802, 90973},
 {"Proof|ID:2101232329-1", 3364968, 90977},
 {"Proof|ID:96556784", 3365130, 90981},
 {"Proof|ID:96556784-1", 3365289, 90985},
 {"Proof|ID:1465630278", 3365448, 90989},
 {"Proof|ID:1465630278-1", 3365614, 90993},
 {"Proof|ID:1952076718", 3365778, 90997},
 {"Proof|ID:1952076718-1", 3365944, 91001},
 {"Proof|ID:1236615595", 3366108, 91005},
 {"Proof|ID:1236615595-1", 3366273, 91009},
 {"Proof|ID:203149419", 3366435, 91013},
 {"Proof|ID:203149419-1", 3366597, 91017},
 {"Proof|ID:333495152", 3366757, 91021},
 {"Proof|ID:333495152-1", 3366920, 91025},
 {"Proof|ID:332101094", 3367081, 91029},
 {"Proof|ID:332101094-1", 3367245, 91033},
 {"Proof|ID:296464255", 3367407, 91037},
 {"Proof|ID:296464255-1", 3368093, 91053},
 {"Proof|ID:194930655", 3368777, 91069},
 {"Proof|ID:194930655-1", 3368940, 91073},
 {"Proof|ID:935171195", 3369101, 91077},
 {"Proof|ID:935171195-1", 3369264, 91081},
 {"ID:1368508872", 3369420, 91085},
 {"ID:1610984937", 3369650, 91092},
 {"areIrrelative symmetric", 3369923, 91099},
 {"ID:981548331", 3370185, 91106},
 {"areCorrelative symmetric", 3370458, 91113},
 {"Proof|ID:1610984937", 3370726, 91120},
 {"Proof|ID:1610984937-1", 3370892, 91124},
 {"Proof|ID:981548331", 3371055, 91128},
 {"Proof|ID:981548331-1", 3371217, 91132},
 {"ID:1547543437", 3371372, 91136},
 {"ID:1699369800", 3371602, 91143},
 {"isReducible \[RightArrow]", 3371879, 91150},
 {"ID:1596540807", 3372144, 91157},
 {"\[RightArrow] isReducible", 3372421, 91164},
 {"ID:1641388157", 3372686, 91171},
 {"isReducible symred", 3372949, 91178},
 {"ID:2087655986", 3373207, 91185},
 {"isReducible alternative", 3373486, 91192},
 {"ID:584152369", 3373754, 91199},
 {"areCPMultipliers \[RightArrow] tuple", 3374051, 91206},
 {"Proof|ID:1699369800", 3374331, 91213},
 {"Proof|ID:1699369800-1", 3374496, 91217},
 {"Proof|ID:1596540807", 3374659, 91221},
 {"Proof|ID:1596540807-1", 3374824, 91225},
 {"Proof|ID:1641388157", 3374987, 91229},
 {"Proof|ID:1641388157-1", 3375151, 91233},
 {"Proof|ID:2087655986", 3375313, 91237},
 {"Proof|ID:2087655986-1", 3375477, 91241},
 {"Proof|ID:584152369", 3375638, 91245},
 {"Proof|ID:584152369-1", 3375800, 91249},
 {"ID:593421628", 3375954, 91253},
 {"ID:680005293", 3376181, 91260},
 {"1", 3376408, 91267},
 {"ID:1320396264", 3376647, 91274},
 {"ID:1127232822", 3376878, 91281},
 {"\[RightArrow] reductor", 3377149, 91288},
 {"ID:256297052", 3377410, 91295},
 {"\[Succeeds] reductor", 3377674, 91302},
 {"Proof|ID:680005293", 3377936, 91309},
 {"Proof|ID:680005293-1", 3378097, 91313},
 {"Proof|ID:1127232822", 3378257, 91317},
 {"Proof|ID:1127232822-1", 3378422, 91321},
 {"Proof|ID:256297052", 3378584, 91325},
 {"Proof|ID:256297052-1", 3378746, 91329},
 {"ID:1083403145", 3378901, 91333},
 {"ID:1592577568", 3379200, 91341},
 {"DomainSets red", 3379529, 91349},
 {"ID:698444072", 3379856, 91357},
 {"red isSet", 3380172, 91365},
 {"ID:277656140", 3380491, 91373},
 {"\[Element] red", 3380812, 91381},
 {"ID:454978652", 3381131, 91389},
 {"\[Element] red singleton", 3381472, 91397},
 {"ID:259827218", 3381801, 91405},
 {"\[Element] red abstraction term", 3382156, 91413},
 {"ID:1273500442", 3382493, 91421},
 {"\[Element] red \[Union]", 3382834, 91429},
 {"ID:1877397922", 3383165, 91437},
 {"red \[Union]", 3383484, 91445},
 {"ID:2121666706", 3383804, 91453},
 {"\[Element] red \[Union] singleton", 3384165, 91461},
 {"ID:947744040", 3384505, 91469},
 {"\[SubsetEqual] red \[SubsetEqual]", 3384864, 91477},
 {"ID:52516181", 3385201, 91485},
 {"red \[Subset]", 3385518, 91493},
 {"ID:194436663", 3385834, 91501},
 {"red \[Subset] \[Union] singleton", 3386191, 91509},
 {"ID:844584535", 3386528, 91517},
 {"red \[Subset] \[Union] singleton II", 3386891, 91525},
 {"Proof|ID:1592577568", 3387238, 91533},
 {"Proof|ID:1592577568-1", 3387402, 91537},
 {"Proof|ID:698444072", 3387563, 91541},
 {"Proof|ID:698444072-1", 3387722, 91545},
 {"Proof|ID:277656140", 3387879, 91549},
 {"Proof|ID:277656140-1", 3388040, 91553},
 {"Proof|ID:454978652", 3388199, 91557},
 {"Proof|ID:454978652-1", 3388360, 91561},
 {"Proof|ID:259827218", 3388519, 91565},
 {"Proof|ID:259827218-1", 3388681, 91569},
 {"Proof|ID:1273500442", 3388842, 91573},
 {"Proof|ID:1273500442-1", 3389007, 91577},
 {"Proof|ID:1877397922", 3389170, 91581},
 {"Proof|ID:1877397922-1", 3389335, 91585},
 {"Proof|ID:2121666706", 3389498, 91589},
 {"Proof|ID:2121666706-1", 3389663, 91593},
 {"Proof|ID:947744040", 3389825, 91597},
 {"Proof|ID:947744040-1", 3389987, 91601},
 {"Proof|ID:52516181", 3390146, 91605},
 {"Proof|ID:52516181-1", 3390304, 91609},
 {"Proof|ID:194436663", 3390461, 91613},
 {"Proof|ID:194436663-1", 3390623, 91617},
 {"Proof|ID:844584535", 3390783, 91621},
 {"Proof|ID:844584535-1", 3390944, 91625},
 {"ID:632054408", 3391097, 91629},
 {"red \[SubsetEqual] transitive", 3391380, 91636},
 {"ID:387695114", 3391646, 91643},
 {"ID:1177850337", 3391874, 91650},
 {"setSeq subsequence red", 3392145, 91657},
 {"ID:1483708233", 3392407, 91664},
 {"ID:697433658", 3392636, 91671},
 {"red setSeq \[SubsetEqual]", 3392911, 91678},
 {"ID:35999503", 3393172, 91685},
 {"\[Element] red setSeq", 3393437, 91692},
 {"ID:602168712", 3393693, 91699},
 {"red setSeq \[Subset]", 3393958, 91706},
 {"Proof|ID:632054408", 3394221, 91713},
 {"Proof|ID:632054408-1", 3394382, 91717},
 {"Proof|ID:1177850337", 3394542, 91721},
 {"Proof|ID:1177850337-1", 3394708, 91725},
 {"Proof|ID:697433658", 3394871, 91729},
 {"Proof|ID:697433658-1", 3395033, 91733},
 {"Proof|ID:35999503", 3395192, 91737},
 {"Proof|ID:35999503-1", 3395351, 91741},
 {"Proof|ID:602168712", 3395509, 91745},
 {"Proof|ID:602168712-1", 3395671, 91749},
 {"ID:21782685", 3395824, 91753},
 {"\[SquareSubset] append", 3396097, 91760},
 {"ID:1790989103", 3396361, 91767},
 {"\[SquareSubset] isPartIrreflOrder", 3396654, 91774},
 {"Proof|ID:21782685", 3396931, 91781},
 {"Proof|ID:21782685-1", 3397090, 91785},
 {"Proof|ID:1790989103", 3397249, 91789},
 {"Proof|ID:1790989103-1", 3397410, 91793},
 {"ID:1466856847", 3397563, 91797},
 {"ID:623701544", 3397792, 91804},
 {"reduction ring", 3398045, 91811},
 {"ID:1960646232", 3398297, 91818},
 {"multipliers", 3398546, 91825},
 {"ID:1496065688", 3398797, 91832},
 {"R0", 3399028, 91839},
 {"ID:975892043", 3399269, 91846},
 {"R1", 3399498, 91853},
 {"ID:1150364990", 3399738, 91860},
 {"R2", 3399969, 91867},
 {"ID:852396718", 3400210, 91874},
 {"R3", 3400439, 91881},
 {"ID:1907869025", 3400679, 91888},
 {"R4", 3400910, 91895},
 {"ID:1446921308", 3401152, 91902},
 {"R5", 3401383, 91909},
 {"ID:208630606", 3401624, 91916},
 {"R6", 3401853, 91923},
 {"ID:1150983847", 3402093, 91930},
 {"R7", 3402326, 91937},
 {"ID:2106857486", 3402570, 91944},
 {"R8", 3402802, 91951},
 {"ID:1197875862", 3403045, 91958},
 {"ID:1949920189", 3403275, 91965},
 {"R5 witnesses", 3403526, 91972},
 {"ID:2027349511", 3403778, 91979},
 {"R6 witness", 3404025, 91986},
 {"ID:1717911774", 3404275, 91993},
 {"ID:349001278", 3404504, 92000},
 {"ID:474542129", 3404731, 92007},
 {"9", 3404958, 92014},
 {"ID:360788203", 3405196, 92021},
 {"21", 3405425, 92028},
 {"Proof|ID:474542129", 3405670, 92035},
 {"Proof|ID:474542129-1", 3405833, 92039},
 {"Proof|ID:360788203", 3405994, 92043},
 {"Proof|ID:360788203-1", 3406154, 92047},
 {"ID:998446180", 3406306, 92051},
 {"ID:1581977856", 3406534, 92058},
 {"ID:623334744", 3406763, 92065},
 {"8", 3406990, 92072},
 {"ID:1112133758", 3407229, 92079},
 {"8'", 3407460, 92086},
 {"ID:810098043", 3407701, 92093},
 {"ID:1578524611", 3407929, 92100},
 {"19", 3408160, 92107},
 {"ID:1738175589", 3408402, 92114},
 {"19'", 3408635, 92121},
 {"Proof|ID:623334744", 3408883, 92128},
 {"Proof|ID:623334744-1", 3409046, 92132},
 {"Proof|ID:1112133758", 3409208, 92136},
 {"Proof|ID:1112133758-1", 3409374, 92140},
 {"Proof|ID:1578524611", 3409538, 92144},
 {"Proof|ID:1578524611-1", 3409702, 92148},
 {"Proof|ID:1738175589", 3409864, 92152},
 {"Proof|ID:1738175589-1", 3410029, 92156},
 {"ID:258339355", 3410185, 92160},
 {"ID:1933985410", 3410414, 92167},
 {"ID:739393974", 3410643, 92174},
 {"17", 3410871, 92181},
 {"ID:875264505", 3411109, 92188},
 {"10", 3411338, 92195},
 {"ID:2012915447", 3411578, 92202},
 {"11", 3411809, 92209},
 {"ID:1489336269", 3412051, 92216},
 {"12", 3412282, 92223},
 {"ID:1722317392", 3412524, 92230},
 {"13", 3412755, 92237},
 {"ID:1089845252", 3412997, 92244},
 {"15", 3413228, 92251},
 {"Proof|ID:739393974", 3413475, 92258},
 {"Proof|ID:739393974-1", 3413638, 92262},
 {"Proof|ID:875264505", 3413799, 92266},
 {"Proof|ID:875264505-1", 3413961, 92270},
 {"Proof|ID:2012915447", 3414122, 92274},
 {"Proof|ID:2012915447-1", 3414288, 92278},
 {"Proof|ID:1489336269", 3414452, 92282},
 {"Proof|ID:1489336269-1", 3414616, 92286},
 {"Proof|ID:1722317392", 3414778, 92290},
 {"Proof|ID:1722317392-1", 3414944, 92294},
 {"Proof|ID:1089845252", 3415108, 92298},
 {"Proof|ID:1089845252-1", 3415272, 92302},
 {"ID:2021433946", 3415428, 92306},
 {"ID:369680194", 3415658, 92313},
 {"ID:491925495", 3415885, 92320},
 {"\[LeftRightArrow] alternative", 3416168, 92327},
 {"ID:1668317482", 3416435, 92334},
 {"symred \[LeftRightArrow]", 3416710, 92341},
 {"Proof|ID:491925495", 3416979, 92348},
 {"Proof|ID:491925495-1", 3417140, 92352},
 {"Proof|ID:1668317482", 3417300, 92356},
 {"Proof|ID:1668317482-1", 3417465, 92360},
 {"ID:169581633", 3417621, 92364},
 {"ID:265555623", 3417848, 92371},
 {"403", 3418079, 92378},
 {"Proof|ID:265555623", 3418325, 92385},
 {"Proof|ID:265555623-1", 3418488, 92389},
 {"ID:1020054522", 3418644, 92393},
 {"ID:270873424", 3418875, 92400},
 {"ID:1165091979", 3419104, 92407},
 {"16", 3419335, 92414},
 {"ID:1894133408", 3419577, 92421},
 {"symred symmetry", 3419835, 92428},
 {"Proof|ID:1165091979", 3420097, 92435},
 {"Proof|ID:1165091979-1", 3420264, 92439},
 {"Proof|ID:1894133408", 3420429, 92443},
 {"Proof|ID:1894133408-1", 3420596, 92447},
 {"ID:1779789522", 3420755, 92451},
 {"ID:979757495", 3420985, 92458},
 {"algorithmic reduction ring", 3421263, 92465},
 {"ID:1502179371", 3421528, 92472},
 {"R9", 3421761, 92479},
 {"ID:1524689422", 3422005, 92486},
 {"R10", 3422240, 92493},
 {"ID:528478842", 3422484, 92500},
 {"R11", 3422716, 92507},
 {"ID:415149625", 3422957, 92514},
 {"R12", 3423190, 92521},
 {"ID:1276925313", 3423433, 92528},
 {"R13", 3423667, 92535},
 {"ID:1684964128", 3423911, 92542},
 {"ID:10673911", 3424140, 92549},
 {"22", 3424368, 92556},
 {"ID:603265816", 3424606, 92563},
 {"23", 3424836, 92570},
 {"ID:39612209", 3425075, 92577},
 {"24", 3425304, 92584},
 {"Proof|ID:10673911", 3425548, 92591},
 {"Proof|ID:10673911-1", 3425707, 92595},
 {"Proof|ID:603265816", 3425865, 92599},
 {"Proof|ID:603265816-1", 3426028, 92603},
 {"Proof|ID:39612209", 3426188, 92607},
 {"Proof|ID:39612209-1", 3426349, 92611},
 {"ID:1515571182", 3426504, 92615},
 {"ID:528027344", 3426734, 92622},
 {"18", 3426964, 92629},
 {"ID:94925229", 3427203, 92636},
 {"25", 3427431, 92643},
 {"ID:2104251234", 3427670, 92650},
 {"18'", 3427904, 92657},
 {"Proof|ID:528027344", 3428153, 92664},
 {"Proof|ID:528027344-1", 3428316, 92668},
 {"Proof|ID:94925229", 3428476, 92672},
 {"Proof|ID:94925229-1", 3428636, 92676},
 {"Proof|ID:2104251234", 3428796, 92680},
 {"Proof|ID:2104251234-1", 3428962, 92684},
 {"ID:484595803", 3429119, 92688},
 {"isFinDomain \[DoubleRightArrow] R13", 3429414, 92695},
 {"ID:55361704", 3429685, 92702},
 {"ID:283927726", 3429911, 92709},
 {"20", 3430141, 92716},
 {"ID:1584214935", 3430382, 92723},
 {"\[SquareSubset] isNoetherian sets", 3430676, 92730},
 {"ID:1231517070", 3430950, 92737},
 {"\[SquareSubset] isNoetherian tuples", 3431248, 92744},
 {"Proof|ID:484595803", 3431529, 92751},
 {"Proof|ID:484595803-1", 3431691, 92755},
 {"Proof|ID:283927726", 3431851, 92759},
 {"Proof|ID:283927726-1", 3432014, 92763},
 {"Proof|ID:1584214935", 3432176, 92767},
 {"Proof|ID:1584214935-1", 3432342, 92771},
 {"Proof|ID:1231517070", 3432506, 92775},
 {"Proof|ID:1231517070-1", 3432671, 92779},
 {"ID:1326788334", 3432828, 92783},
 {"ID:227700447", 3433058, 92790},
 {"ID:1128391937", 3433287, 92797},
 {"ID:788600821", 3433517, 92804},
 {"ID:999322294", 3433746, 92811},
 {"ID:737948621", 3433974, 92818},
 {"392", 3434205, 92825},
 {"ID:287016951", 3434445, 92832},
 {"394", 3434677, 92839},
 {"ID:1812129266", 3434919, 92846},
 {"395", 3435153, 92853},
 {"ID:1191851294", 3435397, 92860},
 {"396", 3435632, 92867},
 {"ID:101106535", 3435876, 92874},
 {"397", 3436109, 92881},
 {"ID:1737800130", 3436352, 92888},
 {"398", 3436586, 92895},
 {"ID:1568772314", 3436830, 92902},
 {"399", 3437064, 92909},
 {"Proof|ID:737948621", 3437313, 92916},
 {"Proof|ID:737948621-1", 3437476, 92920},
 {"Proof|ID:287016951", 3437637, 92924},
 {"Proof|ID:287016951-1", 3437800, 92928},
 {"Proof|ID:1812129266", 3437962, 92932},
 {"Proof|ID:1812129266-1", 3438128, 92936},
 {"Proof|ID:1191851294", 3438292, 92940},
 {"Proof|ID:1191851294-1", 3438459, 92944},
 {"Proof|ID:101106535", 3438623, 92948},
 {"Proof|ID:101106535-1", 3438788, 92952},
 {"Proof|ID:1737800130", 3438952, 92956},
 {"Proof|ID:1737800130-1", 3439119, 92960},
 {"Proof|ID:1568772314", 3439284, 92964},
 {"Proof|ID:1568772314-1", 3439452, 92968},
 {"ID:1597110548", 3439612, 92972},
 {"ID:1509853174", 3439843, 92979},
 {"ID:1554200522", 3440074, 92986},
 {"ID:1963951655", 3440306, 92993},
 {"400", 3440540, 93000},
 {"ID:85295514", 3440782, 93007},
 {"ID:751169803", 3441009, 93014},
 {"401", 3441241, 93021},
 {"ID:46875747", 3441481, 93028},
 {"402'", 3441713, 93035},
 {"ID:656927872", 3441953, 93042},
 {"402", 3442185, 93049},
 {"Proof|ID:1963951655", 3442433, 93056},
 {"Proof|ID:1963951655-1", 3442600, 93060},
 {"Proof|ID:751169803", 3442764, 93064},
 {"Proof|ID:751169803-1", 3442928, 93068},
 {"Proof|ID:46875747", 3443089, 93072},
 {"Proof|ID:46875747-1", 3443250, 93076},
 {"Proof|ID:656927872", 3443410, 93080},
 {"Proof|ID:656927872-1", 3443574, 93084},
 {"ID:502021608", 3443730, 93088},
 {"ID:227587436", 3443958, 93095},
 {"ID:937292216", 3444186, 93102},
 {"ID:1763687895", 3444416, 93109},
 {"410", 3444650, 93116},
 {"ID:397073436", 3444893, 93123},
 {"ID:791651685", 3445122, 93130},
 {"411", 3445354, 93137},
 {"ID:6804200", 3445593, 93144},
 {"ID:1294525434", 3445819, 93151},
 {"414", 3446053, 93158},
 {"ID:372638659", 3446296, 93165},
 {"ID:1962716191", 3446526, 93172},
 {"412", 3446760, 93179},
 {"ID:1575262565", 3447004, 93186},
 {"413", 3447238, 93193},
 {"Proof|ID:1763687895", 3447488, 93200},
 {"Proof|ID:1763687895-1", 3447655, 93204},
 {"Proof|ID:791651685", 3447819, 93208},
 {"Proof|ID:791651685-1", 3447984, 93212},
 {"Proof|ID:1294525434", 3448148, 93216},
 {"Proof|ID:1294525434-1", 3448315, 93220},
 {"Proof|ID:1962716191", 3448480, 93224},
 {"Proof|ID:1962716191-1", 3448648, 93228},
 {"Proof|ID:1575262565", 3448814, 93232},
 {"Proof|ID:1575262565-1", 3448982, 93236},
 {"ID:1547867165", 3449142, 93240},
 {"ID:2009958353", 3449374, 93247},
 {"ID:10165835", 3449604, 93254},
 {"ID:1879224373", 3449832, 93261},
 {"R6Witness sufficient", 3450100, 93268},
 {"ID:381733398", 3450360, 93275},
 {"R6Witness symmetry", 3450623, 93282},
 {"Proof|ID:1879224373", 3450887, 93289},
 {"Proof|ID:1879224373-1", 3451054, 93293},
 {"Proof|ID:381733398", 3451218, 93297},
 {"Proof|ID:381733398-1", 3451382, 93301},
 {"ID:404204643", 3451538, 93305},
 {"ID:1399661856", 3451767, 93312},
 {"ID:1116556350", 3451999, 93319},
 {"\[Tilde] multiplier symmetric", 3452285, 93326},
 {"ID:1121162524", 3452555, 93333},
 {"mult \[Tilde]", 3452809, 93340},
 {"ID:1098905803", 3453063, 93347},
 {"isReducible \[Tilde]", 3453331, 93354},
 {"ID:88395859", 3453590, 93361},
 {"areIrrelative \[Tilde]", 3453858, 93368},
 {"ID:1322820809", 3454117, 93375},
 {"areCPMultipliers \[Tilde]", 3454395, 93382},
 {"ID:931721333", 3454660, 93389},
 {"ID:2090837112", 3454889, 93396},
 {"\[RightArrow] multiplier \[Tilde]", 3455183, 93403},
 {"ID:780050836", 3455456, 93410},
 {"\[RightArrow] \[Tilde]", 3455726, 93417},
 {"ID:128058836", 3455986, 93424},
 {"ID:1516437288", 3456215, 93431},
 {"\[RightArrow] reductor \[Tilde]", 3456505, 93438},
 {"ID:1864007722", 3456777, 93445},
 {"\[LeftRightArrow] reductor \[Tilde]", 3457075, 93452},
 {"ID:642398934", 3457350, 93459},
 {"connectibility below \[Tilde]", 3457634, 93466},
 {"Proof|ID:1116556350", 3457908, 93473},
 {"Proof|ID:1116556350-1", 3458075, 93477},
 {"Proof|ID:1121162524", 3458240, 93481},
 {"Proof|ID:1121162524-1", 3458407, 93485},
 {"Proof|ID:1098905803", 3458572, 93489},
 {"Proof|ID:1098905803-1", 3458739, 93493},
 {"Proof|ID:88395859", 3458902, 93497},
 {"Proof|ID:88395859-1", 3459062, 93501},
 {"Proof|ID:1322820809", 3459222, 93505},
 {"Proof|ID:1322820809-1", 3459388, 93509},
 {"Proof|ID:2090837112", 3459552, 93513},
 {"Proof|ID:2090837112-1", 3459719, 93517},
 {"Proof|ID:780050836", 3459883, 93521},
 {"Proof|ID:780050836-1", 3460046, 93525},
 {"Proof|ID:1516437288", 3460208, 93529},
 {"Proof|ID:1516437288-1", 3460373, 93533},
 {"Proof|ID:1864007722", 3460536, 93537},
 {"Proof|ID:1864007722-1", 3460703, 93541},
 {"Proof|ID:642398934", 3460867, 93545},
 {"Proof|ID:642398934-1", 3461031, 93549},
 {"ID:1200903488", 3461188, 93553},
 {"ID:353808388", 3461418, 93560},
 {"reflexivity of \[Tilde]", 3461689, 93567},
 {"ID:274091018", 3461949, 93574},
 {"symmetry of \[Tilde]", 3462215, 93581},
 {"ID:1760571189", 3462474, 93588},
 {"transitivity of \[Tilde]", 3462750, 93595},
 {"Proof|ID:353808388", 3463020, 93602},
 {"Proof|ID:353808388-1", 3463182, 93606},
 {"Proof|ID:274091018", 3463342, 93610},
 {"Proof|ID:274091018-1", 3463505, 93614},
 {"Proof|ID:1760571189", 3463667, 93618},
 {"Proof|ID:1760571189-1", 3463834, 93622},
 {"ID:871669950", 3463992, 93626},
 {"ID:163646289", 3464220, 93633},
 {"correlative-irrelative 1", 3464494, 93640},
 {"Proof|ID:163646289", 3464762, 93647},
 {"Proof|ID:163646289-1", 3464926, 93651},
 {"ID:1990325407", 3465083, 93655},
 {"ID:615171230", 3465313, 93662},
 {"ID:1548501955", 3465542, 93669},
 {"47", 3465774, 93676},
 {"ID:170197122", 3466016, 93683},
 {"48", 3466246, 93690},
 {"ID:772910972", 3466486, 93697},
 {"49", 3466716, 93704},
 {"Proof|ID:1548501955", 3466963, 93711},
 {"Proof|ID:1548501955-1", 3467129, 93715},
 {"Proof|ID:170197122", 3467292, 93719},
 {"Proof|ID:170197122-1", 3467455, 93723},
 {"Proof|ID:772910972", 3467616, 93727},
 {"Proof|ID:772910972-1", 3467778, 93731},
 {"ID:1034579573", 3467933, 93735},
 {"ID:226093298", 3468164, 93742},
 {"30.\[DoubleRightArrow]", 3468436, 93749},
 {"ID:1773205884", 3468699, 93756},
 {"31.\[DoubleRightArrow]", 3468973, 93763},
 {"ID:1783130815", 3469238, 93770},
 {"30.\[DoubleLeftArrow]", 3469509, 93777},
 {"ID:1493153487", 3469772, 93784},
 {"31.\[DoubleLeftArrow]", 3470043, 93791},
 {"ID:1798779961", 3470306, 93798},
 {"39.\[DoubleRightArrow]", 3470578, 93805},
 {"Proof|ID:226093298", 3470846, 93812},
 {"Proof|ID:226093298-1", 3471011, 93816},
 {"Proof|ID:1773205884", 3471175, 93820},
 {"Proof|ID:1773205884-1", 3471342, 93824},
 {"Proof|ID:1783130815", 3471507, 93828},
 {"Proof|ID:1783130815-1", 3471674, 93832},
 {"Proof|ID:1493153487", 3471839, 93836},
 {"Proof|ID:1493153487-1", 3472006, 93840},
 {"Proof|ID:1798779961", 3472171, 93844},
 {"Proof|ID:1798779961-1", 3472338, 93848},
 {"ID:712613448", 3472496, 93852},
 {"ID:1914773542", 3472726, 93859},
 {"37.\[DoubleRightArrow]", 3472999, 93866},
 {"ID:2048103084", 3473263, 93873},
 {"37.\[DoubleLeftArrow]", 3473534, 93880},
 {"Proof|ID:1914773542", 3473803, 93887},
 {"Proof|ID:1914773542-1", 3473969, 93891},
 {"Proof|ID:2048103084", 3474133, 93895},
 {"Proof|ID:2048103084-1", 3474300, 93899},
 {"ID:254692456", 3474458, 93903},
 {"ID:336730845", 3474687, 93910},
 {"32.\[DoubleRightArrow]", 3474959, 93917},
 {"ID:1715399148", 3475222, 93924},
 {"32.\[DoubleLeftArrow]", 3475494, 93931},
 {"Proof|ID:336730845", 3475763, 93938},
 {"Proof|ID:336730845-1", 3475927, 93942},
 {"Proof|ID:1715399148", 3476090, 93946},
 {"Proof|ID:1715399148-1", 3476258, 93950},
 {"ID:1253203360", 3476418, 93954},
 {"ID:1514263990", 3476649, 93961},
 {"reflexivity of \[RightArrow]\[RawStar]", 3476952, 93968},
 {"ID:411529944", 3477229, 93975},
 {"transitivity of \[RightArrow]\[RawStar]", 3477533, 93982},
 {"Proof|ID:1514263990", 3477817, 93989},
 {"Proof|ID:1514263990-1", 3477983, 93993},
 {"Proof|ID:411529944", 3478146, 93997},
 {"Proof|ID:411529944-1", 3478309, 94001},
 {"ID:1862973197", 3478465, 94005},
 {"ID:237268641", 3478695, 94012},
 {"\[RightArrow] \[SubsetEqual] \[RightArrow]\[RawStar]", 3479031, 94020},
 {"ID:1746879589", 3479328, 94028},
 {"38", 3479560, 94035},
 {"Proof|ID:237268641", 3479808, 94042},
 {"Proof|ID:237268641-1", 3479972, 94046},
 {"Proof|ID:1746879589", 3480135, 94050},
 {"Proof|ID:1746879589-1", 3480302, 94054},
 {"ID:1956094048", 3480461, 94058},
 {"ID:560057636", 3480691, 94065},
 {"reflexivity of connectibility", 3480974, 94072},
 {"ID:354829984", 3481240, 94079},
 {"symmetry of connectibility", 3481518, 94086},
 {"ID:350848859", 3481782, 94093},
 {"transitivity of connectibility", 3482068, 94100},
 {"Proof|ID:560057636", 3482342, 94107},
 {"Proof|ID:560057636-1", 3482505, 94111},
 {"Proof|ID:354829984", 3482666, 94115},
 {"Proof|ID:354829984-1", 3482829, 94119},
 {"Proof|ID:350848859", 3482990, 94123},
 {"Proof|ID:350848859-1", 3483154, 94127},
 {"ID:705004664", 3483310, 94131},
 {"ID:411199728", 3483539, 94138},
 {"reflexivity of connectibility below", 3483835, 94145},
 {"ID:1502850808", 3484109, 94152},
 {"symmetry of connectibility below", 3484401, 94159},
 {"ID:1331049102", 3484674, 94166},
 {"transitivity of connectibility below", 3484974, 94173},
 {"Proof|ID:411199728", 3485256, 94180},
 {"Proof|ID:411199728-1", 3485418, 94184},
 {"Proof|ID:1502850808", 3485579, 94188},
 {"Proof|ID:1502850808-1", 3485746, 94192},
 {"Proof|ID:1331049102", 3485911, 94196},
 {"Proof|ID:1331049102-1", 3486077, 94200},
 {"ID:1147958002", 3486235, 94204},
 {"36", 3486468, 94211},
 {"ID:1663785837", 3486712, 94218},
 {"33", 3486945, 94225},
 {"Proof|ID:1147958002", 3487195, 94232},
 {"Proof|ID:1147958002-1", 3487360, 94236},
 {"Proof|ID:1663785837", 3487523, 94240},
 {"Proof|ID:1663785837-1", 3487690, 94244},
 {"ID:311737773", 3487848, 94248},
 {"ID:552003293", 3488076, 94255},
 {"reflexivity of reflexive-transitive closure of \[RightArrow]", 3488428, \
94263},
 {"ID:129456084", 3488732, 94271},
 {"ID:441235278", 3488960, 94278},
 {"34", 3489190, 94285},
 {"ID:1516348781", 3489431, 94292},
 {"35", 3489664, 94299},
 {"Proof|ID:552003293", 3489913, 94306},
 {"Proof|ID:552003293-1", 3490074, 94310},
 {"Proof|ID:441235278", 3490233, 94314},
 {"Proof|ID:441235278-1", 3490394, 94318},
 {"Proof|ID:1516348781", 3490554, 94322},
 {"Proof|ID:1516348781-1", 3490721, 94326},
 {"ID:1329557757", 3490880, 94330},
 {"ID:1168535485", 3491111, 94337},
 {"40", 3491343, 94344},
 {"ID:2128489125", 3491586, 94351},
 {"41", 3491818, 94358},
 {"Proof|ID:1168535485", 3492067, 94365},
 {"Proof|ID:1168535485-1", 3492233, 94369},
 {"Proof|ID:2128489125", 3492397, 94373},
 {"Proof|ID:2128489125-1", 3492563, 94377},
 {"ID:548213617", 3492720, 94381},
 {"ID:909891749", 3492948, 94388},
 {"42", 3493178, 94395},
 {"ID:830274722", 3493418, 94402},
 {"44", 3493647, 94409},
 {"ID:1390954175", 3493887, 94416},
 {"ID:2016336231", 3494118, 94423},
 {"68", 3494350, 94430},
 {"ID:1052198272", 3494593, 94437},
 {"68.set", 3494833, 94444},
 {"ID:35312349", 3495078, 94451},
 {"68.tuple", 3495318, 94458},
 {"Proof|ID:909891749", 3495568, 94465},
 {"Proof|ID:909891749-1", 3495731, 94469},
 {"Proof|ID:830274722", 3495892, 94473},
 {"Proof|ID:830274722-1", 3496055, 94477},
 {"Proof|ID:2016336231", 3496217, 94481},
 {"Proof|ID:2016336231-1", 3496383, 94485},
 {"Proof|ID:1052198272", 3496547, 94489},
 {"Proof|ID:1052198272-1", 3496714, 94493},
 {"Proof|ID:35312349", 3496877, 94497},
 {"Proof|ID:35312349-1", 3497037, 94501},
 {"ID:2032809580", 3497191, 94505},
 {"ID:1947090769", 3497422, 94512},
 {"ID:693923465", 3497652, 94519},
 {"50", 3497884, 94526},
 {"ID:1234215913", 3498127, 94533},
 {"51", 3498359, 94540},
 {"ID:1065734937", 3498602, 94547},
 {"52", 3498834, 94554},
 {"ID:1763925306", 3499077, 94561},
 {"53", 3499309, 94568},
 {"Proof|ID:693923465", 3499557, 94575},
 {"Proof|ID:693923465-1", 3499721, 94579},
 {"Proof|ID:1234215913", 3499884, 94583},
 {"Proof|ID:1234215913-1", 3500050, 94587},
 {"Proof|ID:1065734937", 3500214, 94591},
 {"Proof|ID:1065734937-1", 3500381, 94595},
 {"Proof|ID:1763925306", 3500546, 94599},
 {"Proof|ID:1763925306-1", 3500712, 94603},
 {"ID:332034572", 3500869, 94607},
 {"ID:1609461850", 3501098, 94614},
 {"ID:621597434", 3501329, 94621},
 {"54", 3501559, 94628},
 {"ID:2031891226", 3501800, 94635},
 {"55", 3502032, 94642},
 {"ID:1596753391", 3502275, 94649},
 {"ID:1855698114", 3502506, 94656},
 {"55'", 3502740, 94663},
 {"Proof|ID:621597434", 3502989, 94670},
 {"Proof|ID:621597434-1", 3503151, 94674},
 {"Proof|ID:2031891226", 3503312, 94678},
 {"Proof|ID:2031891226-1", 3503477, 94682},
 {"Proof|ID:1855698114", 3503640, 94686},
 {"Proof|ID:1855698114-1", 3503806, 94690},
 {"ID:1116860460", 3503964, 94694},
 {"ID:915980145", 3504194, 94701},
 {"56", 3504424, 94708},
 {"ID:1683552886", 3504665, 94715},
 {"ID:1331309258", 3504896, 94722},
 {"ID:551817557", 3505126, 94729},
 {"ID:1724920630", 3505355, 94736},
 {"ID:88655200", 3505585, 94743},
 {"57", 3505813, 94750},
 {"ID:1149353049", 3506052, 94757},
 {"58", 3506284, 94764},
 {"Proof|ID:915980145", 3506532, 94771},
 {"Proof|ID:915980145-1", 3506695, 94775},
 {"Proof|ID:88655200", 3506855, 94779},
 {"Proof|ID:88655200-1", 3507013, 94783},
 {"Proof|ID:1149353049", 3507171, 94787},
 {"Proof|ID:1149353049-1", 3507337, 94791},
 {"ID:895068823", 3507494, 94795},
 {"ID:296045170", 3507722, 94802},
 {"45", 3507952, 94809},
 {"ID:1887054887", 3508193, 94816},
 {"46", 3508425, 94823},
 {"Proof|ID:296045170", 3508673, 94830},
 {"Proof|ID:296045170-1", 3508834, 94834},
 {"Proof|ID:1887054887", 3508994, 94838},
 {"Proof|ID:1887054887-1", 3509160, 94842},
 {"ID:1876022438", 3509318, 94846},
 {"59", 3509550, 94853},
 {"Proof|ID:1876022438", 3509799, 94860},
 {"Proof|ID:1876022438-1", 3509966, 94864},
 {"ID:1852246688", 3510125, 94868},
 {"ID:1471949021", 3510356, 94875},
 {"element confluence", 3510620, 94882},
 {"ID:146706347", 3510878, 94889},
 {"confluence", 3511124, 94896},
 {"ID:1279966325", 3511373, 94903},
 {"Church-Rosser property", 3511645, 94910},
 {"ID:1602877499", 3511908, 94917},
 {"local connectivity", 3512172, 94924},
 {"ID:661413373", 3512430, 94931},
 {"Gr\[ODoubleDot]bner basis", 3512706, 94938},
 {"ID:1444563707", 3512970, 94945},
 {"ID:1408266583", 3513201, 94952},
 {"areConnectibleMultipliers", 3513479, 94959},
 {"ID:882680545", 3513744, 94966},
 {"connectibility of one critical pair", 3514040, 94973},
 {"ID:259042764", 3514313, 94980},
 {"connectibility of critical pairs", 3514604, 94987},
 {"ID:1836116785", 3514876, 94994},
 {"connectibility of critical pairs of tuple", 3515186, 95001},
 {"ID:2066466981", 3515468, 95008},
 {"connectibility of critical pairs of elements", 3515784, 95015},
 {"ID:756678784", 3516068, 95022},
 {"local connectivity of elements", 3516354, 95029},
 {"ID:1173678670", 3516623, 95036},
 {"ID:1524905762", 3516854, 95043},
 {"cpConnectible 1", 3517112, 95050},
 {"Proof|ID:1524905762", 3517374, 95057},
 {"Proof|ID:1524905762-1", 3517540, 95061},
 {"ID:2054204196", 3517698, 95065},
 {"ID:194336473", 3517928, 95072},
 {"ID:2064859128", 3518157, 95079},
 {"cpConnectible tuple", 3518423, 95086},
 {"ID:1276073437", 3518683, 95093},
 {"isLocConnective tuple", 3518953, 95100},
 {"ID:1869404165", 3519215, 95107},
 {"isLocConnective tuple II", 3519490, 95114},
 {"Proof|ID:2064859128", 3519760, 95121},
 {"Proof|ID:2064859128-1", 3519927, 95125},
 {"Proof|ID:1276073437", 3520092, 95129},
 {"Proof|ID:1276073437-1", 3520258, 95133},
 {"Proof|ID:1869404165", 3520422, 95137},
 {"Proof|ID:1869404165-1", 3520587, 95141},
 {"ID:2083446865", 3520744, 95145},
 {"ID:709723813", 3520974, 95152},
 {"cpConnectible alternative", 3521250, 95159},
 {"ID:41201803", 3521512, 95166},
 {"isLocConnective alternative", 3521790, 95173},
 {"ID:1023292110", 3522054, 95180},
 {"isLocConnective sufficient", 3522334, 95187},
 {"Proof|ID:709723813", 3522606, 95194},
 {"Proof|ID:709723813-1", 3522770, 95198},
 {"Proof|ID:41201803", 3522931, 95202},
 {"Proof|ID:41201803-1", 3523091, 95206},
 {"Proof|ID:1023292110", 3523251, 95210},
 {"Proof|ID:1023292110-1", 3523417, 95214},
 {"ID:529333522", 3523574, 95218},
 {"ID:208412986", 3523802, 95225},
 {"cpConnectible join", 3524064, 95232},
 {"ID:1183553358", 3524321, 95239},
 {"cpConnectible append", 3524589, 95246},
 {"ID:286665714", 3524849, 95253},
 {"cpConnectible join II", 3525117, 95260},
 {"ID:1065812089", 3525377, 95267},
 {"cpConnectible append II", 3525651, 95274},
 {"Proof|ID:208412986", 3525920, 95281},
 {"Proof|ID:208412986-1", 3526083, 95285},
 {"Proof|ID:1183553358", 3526245, 95289},
 {"Proof|ID:1183553358-1", 3526412, 95293},
 {"Proof|ID:286665714", 3526576, 95297},
 {"Proof|ID:286665714-1", 3526740, 95301},
 {"Proof|ID:1065812089", 3526903, 95305},
 {"Proof|ID:1065812089-1", 3527070, 95309},
 {"ID:751627229", 3527228, 95313},
 {"ID:351093647", 3527456, 95320},
 {"ID:1476531812", 3527685, 95327},
 {"ID:1047637483", 3527916, 95334},
 {"cpConnectible \[Tilde] \[DoubleRightArrow]", 3528228, 95341},
 {"ID:1237319003", 3528511, 95348},
 {"cpConnectible \[Tilde]", 3528783, 95355},
 {"Proof|ID:1047637483", 3529052, 95362},
 {"Proof|ID:1047637483-1", 3529219, 95366},
 {"Proof|ID:1237319003", 3529384, 95370},
 {"Proof|ID:1237319003-1", 3529551, 95374},
 {"ID:1087064485", 3529710, 95378},
 {"cpConnectible alternative I", 3529992, 95385},
 {"ID:1334373255", 3530260, 95392},
 {"cpConnectible alternative I'", 3530544, 95399},
 {"ID:581920350", 3530812, 95406},
 {"cpConnectible alternative II", 3531094, 95413},
 {"ID:286693183", 3531360, 95420},
 {"cpConnectible alternative II'", 3531644, 95427},
 {"Proof|ID:1087064485", 3531918, 95434},
 {"Proof|ID:1087064485-1", 3532083, 95438},
 {"Proof|ID:1334373255", 3532246, 95442},
 {"Proof|ID:1334373255-1", 3532413, 95446},
 {"Proof|ID:581920350", 3532577, 95450},
 {"Proof|ID:581920350-1", 3532740, 95454},
 {"Proof|ID:286693183", 3532901, 95458},
 {"Proof|ID:286693183-1", 3533064, 95462},
 {"ID:173719369", 3533219, 95466},
 {"ID:762511022", 3533447, 95473},
 {"cpConnectible symmetric", 3533719, 95480},
 {"ID:1792796192", 3533981, 95487},
 {"isLocConnective symmetric", 3534259, 95494},
 {"Proof|ID:762511022", 3534530, 95501},
 {"Proof|ID:762511022-1", 3534693, 95505},
 {"Proof|ID:1792796192", 3534855, 95509},
 {"Proof|ID:1792796192-1", 3535020, 95513},
 {"ID:320557773", 3535176, 95517},
 {"cpConnectible 0", 3535432, 95524},
 {"Proof|ID:320557773", 3535691, 95531},
 {"Proof|ID:320557773-1", 3535855, 95535},
 {"ID:1193869577", 3536012, 95539},
 {"ID:342993412", 3536242, 95546},
 {"ID:1786903669", 3536471, 95553},
 {"63 ind", 3536712, 95560},
 {"ID:1742873082", 3536960, 95567},
 {"63", 3537193, 95574},
 {"ID:332228196", 3537436, 95581},
 {"64", 3537666, 95588},
 {"Proof|ID:1786903669", 3537913, 95595},
 {"Proof|ID:1786903669-1", 3538080, 95599},
 {"Proof|ID:1742873082", 3538245, 95603},
 {"Proof|ID:1742873082-1", 3538413, 95607},
 {"Proof|ID:332228196", 3538578, 95611},
 {"Proof|ID:332228196-1", 3538741, 95615},
 {"ID:2110225002", 3538897, 95619},
 {"65", 3539129, 95626},
 {"Proof|ID:2110225002", 3539378, 95633},
 {"Proof|ID:2110225002-1", 3539545, 95637},
 {"ID:629188208", 3539703, 95641},
 {"66", 3539932, 95648},
 {"ID:1222366601", 3540172, 95655},
 {"67", 3540403, 95662},
 {"Proof|ID:629188208", 3540650, 95669},
 {"Proof|ID:629188208-1", 3540813, 95673},
 {"Proof|ID:1222366601", 3540975, 95677},
 {"Proof|ID:1222366601-1", 3541139, 95681},
 {"ID:2131577443", 3541295, 95685},
 {"ID:2046259788", 3541526, 95692},
 {"ID:205712485", 3541756, 95699},
 {"ID:2050580243", 3541986, 95706},
 {"61", 3542218, 95713},
 {"ID:2100291458", 3542461, 95720},
 {"60", 3542695, 95727},
 {"ID:1810831635", 3542940, 95734},
 {"ID:718268018", 3543170, 95741},
 {"ID:1959449230", 3543399, 95748},
 {"62", 3543633, 95755},
 {"Proof|ID:2050580243", 3543884, 95762},
 {"Proof|ID:2050580243-1", 3544052, 95766},
 {"Proof|ID:2100291458", 3544218, 95770},
 {"Proof|ID:2100291458-1", 3544384, 95774},
 {"Proof|ID:1959449230", 3544548, 95778},
 {"Proof|ID:1959449230-1", 3544715, 95782},
 {"ID:1029764736", 3544874, 95786},
 {"70", 3545107, 95793},
 {"ID:1120053244", 3545351, 95800},
 {"71", 3545585, 95807},
 {"Proof|ID:1029764736", 3545836, 95814},
 {"Proof|ID:1029764736-1", 3546002, 95818},
 {"Proof|ID:1120053244", 3546166, 95822},
 {"Proof|ID:1120053244-1", 3546333, 95826},
 {"ID:25850532", 3546490, 95830},
 {"72", 3546719, 95837},
 {"Proof|ID:25850532", 3546963, 95844},
 {"Proof|ID:25850532-1", 3547124, 95848},
 {"ID:482796083", 3547278, 95852},
 {"74", 3547508, 95859},
 {"ID:1705968105", 3547749, 95866},
 {"75", 3547981, 95873},
 {"Proof|ID:482796083", 3548229, 95880},
 {"Proof|ID:482796083-1", 3548392, 95884},
 {"Proof|ID:1705968105", 3548554, 95888},
 {"Proof|ID:1705968105-1", 3548720, 95892},
 {"ID:68280999", 3548876, 95896},
 {"73", 3549104, 95903},
 {"Proof|ID:68280999", 3549347, 95910},
 {"Proof|ID:68280999-1", 3549505, 95914},
 {"ID:1188710397", 3549657, 95918},
 {"Main Theorem", 3549909, 95925},
 {"Proof|ID:1188710397", 3550168, 95932},
 {"Proof|ID:1188710397-1", 3550333, 95936},
 {"ID:78242703", 3550488, 95940},
 {"ID:807712598", 3550714, 95947},
 {"75 tuples", 3550958, 95954},
 {"ID:645630189", 3551205, 95961},
 {"76 tuples", 3551450, 95968},
 {"ID:47992769", 3551697, 95975},
 {"77 tuples", 3551939, 95982},
 {"Proof|ID:807712598", 3552190, 95989},
 {"Proof|ID:807712598-1", 3552354, 95993},
 {"Proof|ID:645630189", 3552516, 95997},
 {"Proof|ID:645630189-1", 3552679, 96001},
 {"Proof|ID:47992769", 3552839, 96005},
 {"Proof|ID:47992769-1", 3552999, 96009},
 {"ID:41552115", 3553151, 96013},
 {"ID:208902120", 3553377, 96020},
 {"ID:531785290", 3553605, 96027},
 {"ideal congruence", 3553864, 96034},
 {"ID:1814440310", 3554120, 96041},
 {"ideal", 3554358, 96048},
 {"ID:1625946592", 3554604, 96055},
 {"ideal of tuple", 3554860, 96062},
 {"ID:1782924340", 3555115, 96069},
 {"ID:1482545820", 3555346, 96076},
 {"ID:1796414707", 3555577, 96083},
 {"sum multiplication join", 3555852, 96090},
 {"Proof|ID:1796414707", 3556123, 96097},
 {"Proof|ID:1796414707-1", 3556289, 96101},
 {"ID:1254924191", 3556447, 96105},
 {"ID:923334447", 3556677, 96112},
 {"reflexivity of \[Congruent]", 3556956, 96119},
 {"ID:1712150230", 3557221, 96126},
 {"symmetry of \[Congruent]", 3557497, 96133},
 {"ID:1473809719", 3557762, 96140},
 {"transitivity of \[Congruent]", 3558046, 96147},
 {"Proof|ID:923334447", 3558320, 96154},
 {"Proof|ID:923334447-1", 3558483, 96158},
 {"Proof|ID:1712150230", 3558645, 96162},
 {"Proof|ID:1712150230-1", 3558811, 96166},
 {"Proof|ID:1473809719", 3558975, 96170},
 {"Proof|ID:1473809719-1", 3559142, 96174},
 {"ID:306651916", 3559300, 96178},
 {"\[Congruent] alternative ind.1", 3559587, 96185},
 {"ID:571707933", 3559856, 96192},
 {"\[Congruent] alternative ind.2", 3560143, 96199},
 {"Proof|ID:306651916", 3560418, 96206},
 {"Proof|ID:306651916-1", 3560582, 96210},
 {"Proof|ID:571707933", 3560744, 96214},
 {"Proof|ID:571707933-1", 3560909, 96218},
 {"ID:473420099", 3561066, 96222},
 {"ID:544141308", 3561294, 96229},
 {"\[RightArrow] \[SubsetEqual] \[Congruent]", 3561602, 96236},
 {"ID:1307557419", 3561882, 96243},
 {"\[LeftRightArrow] \[SubsetEqual] \[Congruent]", 3562200, 96250},
 {"ID:170897044", 3562485, 96257},
 {"connectibility \[SubsetEqual] \[Congruent] ind", 3562803, 96264},
 {"ID:720101359", 3563087, 96271},
 {"connectibility \[SubsetEqual] \[Congruent]", 3563397, 96278},
 {"Proof|ID:544141308", 3563683, 96285},
 {"Proof|ID:544141308-1", 3563844, 96289},
 {"Proof|ID:1307557419", 3564004, 96293},
 {"Proof|ID:1307557419-1", 3564171, 96297},
 {"Proof|ID:170897044", 3564335, 96301},
 {"Proof|ID:170897044-1", 3564496, 96305},
 {"Proof|ID:720101359", 3564655, 96309},
 {"Proof|ID:720101359-1", 3564818, 96313},
 {"ID:1680644099", 3564974, 96317},
 {"43", 3565206, 96324},
 {"ID:555151129", 3565448, 96331},
 {"\[Congruent] \[SubsetEqual] connectibility ind", 3565767, 96338},
 {"ID:2104087526", 3566053, 96345},
 {"\[Congruent] \[SubsetEqual] connectibility", 3566365, 96352},
 {"Proof|ID:1680644099", 3566654, 96359},
 {"Proof|ID:1680644099-1", 3566820, 96363},
 {"Proof|ID:555151129", 3566983, 96367},
 {"Proof|ID:555151129-1", 3567147, 96371},
 {"Proof|ID:2104087526", 3567310, 96375},
 {"Proof|ID:2104087526-1", 3567477, 96379},
 {"ID:980108134", 3567635, 96383},
 {"\[Congruent]=connectibility", 3567915, 96390},
 {"Proof|ID:980108134", 3568186, 96397},
 {"Proof|ID:980108134-1", 3568349, 96401},
 {"ID:2002188699", 3568505, 96405},
 {"ID:160768319", 3568735, 96412},
 {"\[Congruent] 0", 3568995, 96419},
 {"ID:562483616", 3569253, 96426},
 {"\[Element] \[Congruent] 0", 3569528, 96433},
 {"ID:853458295", 3569790, 96440},
 {"\[Congruent] 0 \[RawStar]", 3570066, 96447},
 {"ID:462880889", 3570329, 96454},
 {"ID:44331059", 3570557, 96461},
 {"\[Congruent] 0 +", 3570812, 96468},
 {"ID:1958399498", 3571064, 96475},
 {"\[Congruent] 0 -", 3571323, 96482},
 {"Proof|ID:160768319", 3571584, 96489},
 {"Proof|ID:160768319-1", 3571747, 96493},
 {"Proof|ID:562483616", 3571908, 96497},
 {"Proof|ID:562483616-1", 3572071, 96501},
 {"Proof|ID:853458295", 3572232, 96505},
 {"Proof|ID:853458295-1", 3572395, 96509},
 {"Proof|ID:44331059", 3572555, 96513},
 {"Proof|ID:44331059-1", 3572716, 96517},
 {"Proof|ID:1958399498", 3572877, 96521},
 {"Proof|ID:1958399498-1", 3573044, 96525},
 {"ID:2069217266", 3573203, 96529},
 {"ID:277446168", 3573433, 96536},
 {"\[Congruent] extended ind", 3573710, 96543},
 {"ID:1459665988", 3573975, 96550},
 {"\[Congruent] extended", 3574245, 96557},
 {"Proof|ID:277446168", 3574512, 96564},
 {"Proof|ID:277446168-1", 3574675, 96568},
 {"Proof|ID:1459665988", 3574837, 96572},
 {"Proof|ID:1459665988-1", 3575003, 96576},
 {"ID:1670835483", 3575161, 96580},
 {"ID:97617962", 3575390, 96587},
 {"ID:1452952621", 3575617, 96594},
 {"ID:1760501242", 3575848, 96601},
 {"DomainSets ideal", 3576113, 96608},
 {"ID:554864723", 3576374, 96615},
 {"ideal isSet", 3576627, 96622},
 {"ID:739591353", 3576881, 96629},
 {"\[Element] ideal", 3577139, 96636},
 {"ID:2030275451", 3577394, 96643},
 {"ID:1644449200", 3577625, 96650},
 {"ideals closed +", 3577888, 96657},
 {"ID:1500033871", 3578149, 96664},
 {"ideals closed - binary", 3578426, 96671},
 {"ID:1114678876", 3578694, 96678},
 {"ideal \[Union] singleton", 3578970, 96685},
 {"ID:1895756686", 3579235, 96692},
 {"ideal \[RightArrow]\[RawStar]", 3579521, 96699},
 {"Proof|ID:1760501242", 3579797, 96706},
 {"Proof|ID:1760501242-1", 3579962, 96710},
 {"Proof|ID:554864723", 3580124, 96714},
 {"Proof|ID:554864723-1", 3580285, 96718},
 {"Proof|ID:739591353", 3580444, 96722},
 {"Proof|ID:739591353-1", 3580607, 96726},
 {"Proof|ID:1644449200", 3580769, 96730},
 {"Proof|ID:1644449200-1", 3580935, 96734},
 {"Proof|ID:1500033871", 3581099, 96738},
 {"Proof|ID:1500033871-1", 3581265, 96742},
 {"Proof|ID:1114678876", 3581429, 96746},
 {"Proof|ID:1114678876-1", 3581596, 96750},
 {"Proof|ID:1895756686", 3581761, 96754},
 {"Proof|ID:1895756686-1", 3581928, 96758},
 {"ID:1581989517", 3582087, 96762},
 {"ID:185976047", 3582317, 96769},
 {"ID:1918127619", 3582546, 96776},
 {"\[Element] tuple-ideal", 3582818, 96783},
 {"ID:279213832", 3583080, 96790},
 {"DomainSets tuple-ideal", 3583356, 96797},
 {"ID:1225983506", 3583623, 96804},
 {"ID:1587050664", 3583854, 96811},
 {"tuple-ideals closed +", 3584129, 96818},
 {"ID:508704301", 3584395, 96825},
 {"tuple-ideals closed - binary", 3584682, 96832},
 {"ID:1181234284", 3584954, 96839},
 {"tuple-ideal append", 3585218, 96846},
 {"ID:352062174", 3585476, 96853},
 {"tuple-ideal \[RightArrow]\[RawStar]", 3585772, 96860},
 {"Proof|ID:1918127619", 3586052, 96867},
 {"Proof|ID:1918127619-1", 3586218, 96871},
 {"Proof|ID:279213832", 3586381, 96875},
 {"Proof|ID:279213832-1", 3586543, 96879},
 {"Proof|ID:1587050664", 3586704, 96883},
 {"Proof|ID:1587050664-1", 3586871, 96887},
 {"Proof|ID:508704301", 3587035, 96891},
 {"Proof|ID:508704301-1", 3587199, 96895},
 {"Proof|ID:1181234284", 3587362, 96899},
 {"Proof|ID:1181234284-1", 3587528, 96903},
 {"Proof|ID:352062174", 3587691, 96907},
 {"Proof|ID:352062174-1", 3587854, 96911}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[579, 22, 241, 5, 84, "Title",
 CellID->1451885783],
Cell[CellGroupData[{
Cell[845, 31, 168, 3, 73, "Section",
 CellID->1274729640],
Cell[CellGroupData[{
Cell[1038, 38, 121, 2, 70, "Subsection",
 CellID->172857195],
Cell[1162, 42, 1366, 28, 70, "Text",
 CellID->653813119],
Cell[2531, 72, 622, 12, 70, "Text",
 CellID->579693263],
Cell[3156, 86, 418, 8, 70, "Text",
 CellID->438335057],
Cell[3577, 96, 1009, 19, 70, "Text",
 CellID->195467169],
Cell[4589, 117, 684, 14, 70, "Text",
 CellID->397022844],
Cell[5276, 133, 721, 11, 70, "Text",
 CellID->571648042],
Cell[CellGroupData[{
Cell[6022, 148, 123, 2, 70, "Subsubsection",
 CellID->399485901],
Cell[6148, 152, 334, 7, 70, "Text",
 CellID->328963968],
Cell[6485, 161, 343, 7, 70, "Text",
 CellID->229090570],
Cell[6831, 170, 315, 10, 70, "Text",
 CellID->272564439]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[7195, 186, 125, 2, 70, "Subsection",
 CellID->721522687],
Cell[7323, 190, 344, 8, 70, "Input",
 CellID->498486718],
Cell[7670, 200, 465, 10, 70, "Input",
 CellID->192289652],
Cell[8138, 212, 423, 9, 70, "Input",
 CellID->329397223],
Cell[8564, 223, 353, 8, 70, "Input",
 CellID->15589227],
Cell[8920, 233, 416, 9, 70, "Input",
 CellID->76117945],
Cell[9339, 244, 354, 8, 70, "Input",
 CellID->248729940],
Cell[9696, 254, 410, 9, 70, "Input",
 CellID->357734827],
Cell[10109, 265, 155, 3, 70, "Input",
 CellID->594767243],
Cell[10267, 270, 153, 3, 70, "Input",
 CellID->617813923],
Cell[10423, 275, 154, 3, 70, "Input",
 CellID->155920184],
Cell[10580, 280, 150, 3, 70, "Input",
 CellID->21169241]
}, Closed]],
Cell[CellGroupData[{
Cell[10767, 288, 124, 2, 70, "Subsection",
 CellID->1386333768],
Cell[CellGroupData[{
Cell[10916, 294, 291, 7, 70, "Input",
 CellID->740290274],
Cell[CellGroupData[{
Cell[11232, 305, 688, 14, 70, "Print",
 CellID->65794957],
Cell[11923, 321, 393, 8, 70, "Print",
 CellID->316361875],
Cell[12319, 331, 393, 8, 70, "Print",
 CellID->548775684],
Cell[12715, 341, 684, 14, 70, "Print",
 CellID->26524607],
Cell[13402, 357, 659, 13, 70, "Print",
 CellID->489535381],
Cell[14064, 372, 393, 8, 70, "Print",
 CellID->193939678],
Cell[14460, 382, 661, 13, 70, "Print",
 CellID->259626599],
Cell[15124, 397, 391, 8, 70, "Print",
 CellID->190398616],
Cell[15518, 407, 673, 13, 70, "Print",
 CellID->494145549],
Cell[16194, 422, 394, 8, 70, "Print",
 CellID->21379931],
Cell[16591, 432, 395, 8, 70, "Print",
 CellID->249495970],
Cell[16989, 442, 397, 8, 70, "Print",
 CellID->241306951],
Cell[17389, 452, 399, 8, 70, "Print",
 CellID->620414427],
Cell[17791, 462, 691, 14, 70, "Print",
 CellID->407148683],
Cell[18485, 478, 393, 8, 70, "Print",
 CellID->551639436],
Cell[18881, 488, 395, 8, 70, "Print",
 CellID->777840439],
Cell[19279, 498, 661, 13, 70, "Print",
 CellID->145768808],
Cell[19943, 513, 390, 8, 70, "Print",
 CellID->89605923],
Cell[20336, 523, 394, 8, 70, "Print",
 CellID->92121938],
Cell[20733, 533, 662, 13, 70, "Print",
 CellID->128672720],
Cell[21398, 548, 394, 8, 70, "Print",
 CellID->52144555],
Cell[21795, 558, 395, 8, 70, "Print",
 CellID->123913510],
Cell[22193, 568, 398, 8, 70, "Print",
 CellID->65928147],
Cell[22594, 578, 396, 8, 70, "Print",
 CellID->54496018],
Cell[22993, 588, 395, 8, 70, "Print",
 CellID->6908874],
Cell[23391, 598, 675, 14, 70, "Print",
 CellID->99796756],
Cell[24069, 614, 392, 8, 70, "Print",
 CellID->36210510],
Cell[24464, 624, 397, 8, 70, "Print",
 CellID->101086047],
Cell[24864, 634, 399, 8, 70, "Print",
 CellID->350526751],
Cell[25266, 644, 399, 8, 70, "Print",
 CellID->322095243],
Cell[25668, 654, 397, 8, 70, "Print",
 CellID->279166168],
Cell[26068, 664, 661, 13, 70, "Print",
 CellID->622879337],
Cell[26732, 679, 395, 8, 70, "Print",
 CellID->313096369],
Cell[27130, 689, 395, 8, 70, "Print",
 CellID->656559649],
Cell[27528, 699, 397, 8, 70, "Print",
 CellID->683729469],
Cell[27928, 709, 397, 8, 70, "Print",
 CellID->5824131],
Cell[28328, 719, 399, 8, 70, "Print",
 CellID->382576221],
Cell[28730, 729, 396, 8, 70, "Print",
 CellID->117131938],
Cell[29129, 739, 662, 13, 70, "Print",
 CellID->26901258],
Cell[29794, 754, 395, 8, 70, "Print",
 CellID->120612918],
Cell[30192, 764, 394, 8, 70, "Print",
 CellID->46751226],
Cell[30589, 774, 399, 8, 70, "Print",
 CellID->459536186],
Cell[30991, 784, 397, 8, 70, "Print",
 CellID->867399896],
Cell[31391, 794, 664, 13, 70, "Print",
 CellID->17672371],
Cell[32058, 809, 668, 13, 70, "Print",
 CellID->32988319],
Cell[32729, 824, 393, 8, 70, "Print",
 CellID->380323932],
Cell[33125, 834, 394, 8, 70, "Print",
 CellID->85422542],
Cell[33522, 844, 399, 8, 70, "Print",
 CellID->691726092],
Cell[33924, 854, 397, 8, 70, "Print",
 CellID->410364951],
Cell[34324, 864, 398, 8, 70, "Print",
 CellID->47676749],
Cell[34725, 874, 399, 8, 70, "Print",
 CellID->149384025],
Cell[35127, 884, 398, 8, 70, "Print",
 CellID->56912388],
Cell[35528, 894, 667, 13, 70, "Print",
 CellID->149443186],
Cell[36198, 909, 393, 8, 70, "Print",
 CellID->656944083],
Cell[36594, 919, 397, 8, 70, "Print",
 CellID->163877051],
Cell[36994, 929, 399, 8, 70, "Print",
 CellID->681446541],
Cell[37396, 939, 397, 8, 70, "Print",
 CellID->241820018],
Cell[37796, 949, 676, 13, 70, "Print",
 CellID->22459567],
Cell[38475, 964, 395, 8, 70, "Print",
 CellID->569799938],
Cell[38873, 974, 397, 8, 70, "Print",
 CellID->708515429],
Cell[39273, 984, 396, 8, 70, "Print",
 CellID->97046791],
Cell[39672, 994, 395, 8, 70, "Print",
 CellID->2171705],
Cell[40070, 1004, 397, 8, 70, "Print",
 CellID->397458865],
Cell[40470, 1014, 667, 13, 70, "Print",
 CellID->230520909],
Cell[41140, 1029, 391, 8, 70, "Print",
 CellID->714598803],
Cell[41534, 1039, 658, 13, 70, "Print",
 CellID->60266956],
Cell[42195, 1054, 394, 8, 70, "Print",
 CellID->82923131],
Cell[42592, 1064, 395, 8, 70, "Print",
 CellID->202755098],
Cell[42990, 1074, 396, 8, 70, "Print",
 CellID->51214078],
Cell[43389, 1084, 398, 8, 70, "Print",
 CellID->82288381],
Cell[43790, 1094, 396, 8, 70, "Print",
 CellID->62924671],
Cell[44189, 1104, 194, 3, 70, "Print",
 CellID->320890793],
Cell[44386, 1109, 182, 3, 70, "Print",
 CellID->6169996],
Cell[44571, 1114, 387, 8, 70, "Print",
 CellID->455341864],
Cell[44961, 1124, 387, 8, 70, "Print",
 CellID->215349446],
Cell[45351, 1134, 387, 8, 70, "Print",
 CellID->9203035],
Cell[45741, 1144, 389, 8, 70, "Print",
 CellID->142561407],
Cell[46133, 1154, 388, 8, 70, "Print",
 CellID->10315266],
Cell[46524, 1164, 390, 8, 70, "Print",
 CellID->14980483],
Cell[46917, 1174, 388, 8, 70, "Print",
 CellID->529978144],
Cell[47308, 1184, 391, 8, 70, "Print",
 CellID->278374742],
Cell[47702, 1194, 389, 8, 70, "Print",
 CellID->132965802],
Cell[48094, 1204, 389, 8, 70, "Print",
 CellID->313688580],
Cell[48486, 1214, 390, 8, 70, "Print",
 CellID->13405936],
Cell[48879, 1224, 389, 8, 70, "Print",
 CellID->777943960],
Cell[49271, 1234, 389, 8, 70, "Print",
 CellID->437165283],
Cell[49663, 1244, 389, 8, 70, "Print",
 CellID->285847006],
Cell[50055, 1254, 388, 8, 70, "Print",
 CellID->60183894],
Cell[50446, 1264, 391, 8, 70, "Print",
 CellID->565313552],
Cell[50840, 1274, 391, 8, 70, "Print",
 CellID->114713085],
Cell[51234, 1284, 389, 8, 70, "Print",
 CellID->587975582],
Cell[51626, 1294, 389, 8, 70, "Print",
 CellID->121683066],
Cell[52018, 1304, 389, 8, 70, "Print",
 CellID->315404724],
Cell[52410, 1314, 393, 8, 70, "Print",
 CellID->260294993],
Cell[52806, 1324, 393, 8, 70, "Print",
 CellID->415917723],
Cell[53202, 1334, 393, 8, 70, "Print",
 CellID->242006766],
Cell[53598, 1344, 393, 8, 70, "Print",
 CellID->171472899],
Cell[53994, 1354, 392, 8, 70, "Print",
 CellID->82226196],
Cell[54389, 1364, 391, 8, 70, "Print",
 CellID->259182974],
Cell[54783, 1374, 393, 8, 70, "Print",
 CellID->132676632],
Cell[55179, 1384, 391, 8, 70, "Print",
 CellID->497601566],
Cell[55573, 1394, 391, 8, 70, "Print",
 CellID->198659878],
Cell[55967, 1404, 393, 8, 70, "Print",
 CellID->629394221],
Cell[56363, 1414, 391, 8, 70, "Print",
 CellID->121665826],
Cell[56757, 1424, 393, 8, 70, "Print",
 CellID->370238075],
Cell[57153, 1434, 392, 8, 70, "Print",
 CellID->28925449],
Cell[57548, 1444, 393, 8, 70, "Print",
 CellID->120104785],
Cell[57944, 1454, 391, 8, 70, "Print",
 CellID->158010373],
Cell[58338, 1464, 393, 8, 70, "Print",
 CellID->152430680],
Cell[58734, 1474, 391, 8, 70, "Print",
 CellID->471335972],
Cell[59128, 1484, 393, 8, 70, "Print",
 CellID->286858841],
Cell[59524, 1494, 393, 8, 70, "Print",
 CellID->137475301],
Cell[59920, 1504, 393, 8, 70, "Print",
 CellID->470908753],
Cell[60316, 1514, 391, 8, 70, "Print",
 CellID->324454066],
Cell[60710, 1524, 393, 8, 70, "Print",
 CellID->386504957],
Cell[61106, 1534, 393, 8, 70, "Print",
 CellID->787338778],
Cell[61502, 1544, 390, 8, 70, "Print",
 CellID->136037090],
Cell[61895, 1554, 393, 8, 70, "Print",
 CellID->700615950],
Cell[62291, 1564, 392, 8, 70, "Print",
 CellID->67596746],
Cell[62686, 1574, 393, 8, 70, "Print",
 CellID->854593272],
Cell[63082, 1584, 393, 8, 70, "Print",
 CellID->267135323],
Cell[63478, 1594, 393, 8, 70, "Print",
 CellID->864019571],
Cell[63874, 1604, 391, 8, 70, "Print",
 CellID->272972042],
Cell[64268, 1614, 389, 8, 70, "Print",
 CellID->902857225],
Cell[64660, 1624, 391, 8, 70, "Print",
 CellID->322703036],
Cell[65054, 1634, 391, 8, 70, "Print",
 CellID->397535149],
Cell[65448, 1644, 393, 8, 70, "Print",
 CellID->488933580],
Cell[65844, 1654, 392, 8, 70, "Print",
 CellID->70778645],
Cell[66239, 1664, 390, 8, 70, "Print",
 CellID->70667339],
Cell[66632, 1674, 393, 8, 70, "Print",
 CellID->229531394],
Cell[67028, 1684, 393, 8, 70, "Print",
 CellID->178020622],
Cell[67424, 1694, 392, 8, 70, "Print",
 CellID->52451009],
Cell[67819, 1704, 392, 8, 70, "Print",
 CellID->22005430],
Cell[68214, 1714, 393, 8, 70, "Print",
 CellID->282684566],
Cell[68610, 1724, 392, 8, 70, "Print",
 CellID->33463925],
Cell[69005, 1734, 393, 8, 70, "Print",
 CellID->577763587],
Cell[69401, 1744, 392, 8, 70, "Print",
 CellID->66262662],
Cell[69796, 1754, 198, 3, 70, "Print",
 CellID->144319770],
Cell[69997, 1759, 197, 3, 70, "Print",
 CellID->274432133],
Cell[70197, 1764, 198, 3, 70, "Print",
 CellID->315474451],
Cell[70398, 1769, 197, 3, 70, "Print",
 CellID->153604934],
Cell[70598, 1774, 182, 3, 70, "Print",
 CellID->8206008],
Cell[70783, 1779, 380, 8, 70, "Print",
 CellID->68279484],
Cell[71166, 1789, 176, 3, 70, "Print",
 CellID->67545578]
}, Open  ]],
Cell[71357, 1795, 151, 3, 70, "Output",
 CellID->656798355]
}, Closed]],
Cell[CellGroupData[{
Cell[71545, 1803, 664, 18, 70, "Input",
 CellID->1642448047],
Cell[CellGroupData[{
Cell[72234, 1825, 643, 13, 70, "Print",
 CellID->1044356],
Cell[72880, 1840, 367, 8, 70, "Print",
 CellID->116553561],
Cell[73250, 1850, 371, 8, 70, "Print",
 CellID->480406741],
Cell[73624, 1860, 373, 8, 70, "Print",
 CellID->380961924],
Cell[74000, 1870, 370, 8, 70, "Print",
 CellID->93536513],
Cell[74373, 1880, 373, 8, 70, "Print",
 CellID->687596748],
Cell[74749, 1890, 372, 8, 70, "Print",
 CellID->68034965],
Cell[75124, 1900, 642, 13, 70, "Print",
 CellID->21001761],
Cell[75769, 1915, 367, 8, 70, "Print",
 CellID->193918569],
Cell[76139, 1925, 369, 8, 70, "Print",
 CellID->722261688],
Cell[76511, 1935, 373, 8, 70, "Print",
 CellID->106859421],
Cell[76887, 1945, 370, 8, 70, "Print",
 CellID->379735127],
Cell[77260, 1955, 168, 3, 70, "Print",
 CellID->452350495],
Cell[77431, 1960, 157, 3, 70, "Print",
 CellID->36310050],
Cell[77591, 1965, 357, 8, 70, "Print",
 CellID->249258381],
Cell[77951, 1975, 361, 8, 70, "Print",
 CellID->127180023],
Cell[78315, 1985, 361, 8, 70, "Print",
 CellID->372231125],
Cell[78679, 1995, 362, 8, 70, "Print",
 CellID->75898686],
Cell[79044, 2005, 359, 8, 70, "Print",
 CellID->7640131],
Cell[79406, 2015, 361, 8, 70, "Print",
 CellID->967827522],
Cell[79770, 2025, 363, 8, 70, "Print",
 CellID->718513836],
Cell[80136, 2035, 360, 8, 70, "Print",
 CellID->77966652],
Cell[80499, 2045, 362, 8, 70, "Print",
 CellID->91783338],
Cell[80864, 2055, 363, 8, 70, "Print",
 CellID->157497756],
Cell[81230, 2065, 361, 8, 70, "Print",
 CellID->472541205],
Cell[81594, 2075, 361, 8, 70, "Print",
 CellID->125784226],
Cell[81958, 2085, 363, 8, 70, "Print",
 CellID->382809887],
Cell[82324, 2095, 362, 8, 70, "Print",
 CellID->94555470],
Cell[82689, 2105, 363, 8, 70, "Print",
 CellID->110179654],
Cell[83055, 2115, 363, 8, 70, "Print",
 CellID->359138595],
Cell[83421, 2125, 362, 8, 70, "Print",
 CellID->12608920],
Cell[83786, 2135, 170, 3, 70, "Print",
 CellID->207028348],
Cell[83959, 2140, 169, 3, 70, "Print",
 CellID->302515485],
Cell[84131, 2145, 172, 3, 70, "Print",
 CellID->163275300],
Cell[84306, 2150, 171, 3, 70, "Print",
 CellID->107663893],
Cell[84480, 2155, 158, 3, 70, "Print",
 CellID->206296641],
Cell[84641, 2160, 353, 8, 70, "Print",
 CellID->436869300],
Cell[84997, 2170, 151, 3, 70, "Print",
 CellID->335789438]
}, Open  ]],
Cell[85163, 2176, 197, 4, 70, "Output",
 CellID->39547402]
}, Closed]],
Cell[85375, 2183, 325, 6, 70, "Input",
 CellID->19428626],
Cell[CellGroupData[{
Cell[85725, 2193, 541, 12, 70, "Input",
 CellID->1909803754],
Cell[86269, 2207, 922, 14, 70, "Output",
 CellID->154284603]
}, Open  ]],
Cell[CellGroupData[{
Cell[87228, 2226, 194, 4, 70, "Input",
 CellID->264346171],
Cell[87425, 2232, 228, 4, 70, "Output",
 CellID->1745020671]
}, Open  ]],
Cell[CellGroupData[{
Cell[87690, 2241, 537, 13, 70, "Input",
 CellID->1780799392],
Cell[88230, 2256, 3070, 43, 70, "Output",
 CellID->778574057]
}, Closed]],
Cell[CellGroupData[{
Cell[91337, 2304, 199, 4, 70, "Input",
 CellID->664434720],
Cell[91539, 2310, 115, 2, 70, "Output",
 CellID->1011754588]
}, Open  ]],
Cell[CellGroupData[{
Cell[91691, 2317, 245, 5, 70, "Input",
 CellID->830230433],
Cell[91939, 2324, 138, 3, 70, "Output",
 CellID->985680483]
}, Open  ]],
Cell[CellGroupData[{
Cell[92114, 2332, 270, 6, 70, "Input",
 CellID->300909609],
Cell[92387, 2340, 209, 3, 70, "Output",
 CellID->1221498976]
}, Open  ]],
Cell[CellGroupData[{
Cell[92633, 2348, 420, 11, 70, "Input",
 CellID->18261744],
Cell[93056, 2361, 119, 2, 70, "Output",
 CellID->1943693605]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[93224, 2369, 128, 2, 70, "Subsection",
 CellID->1454182661],
Cell[93355, 2373, 1013, 26, 70, "Input",
 CellID->2074904605],
Cell[94371, 2401, 416, 9, 70, "Input",
 CellID->1980565193],
Cell[94790, 2412, 169, 3, 70, "Input",
 CellID->444074165],
Cell[94962, 2417, 259, 5, 70, "Input",
 CellID->798575911]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[95270, 2428, 313, 5, 43, "Section",
 CellID->938515413],
Cell[CellGroupData[{
Cell[95608, 2437, 121, 2, 38, "Subsection",
 CellID->588777529],
Cell[95732, 2441, 628, 16, 30, "GlobalDeclaration",
 CellTags->{
  "ID:2066005206", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2066005206],
Cell[96363, 2459, 48, 1, 12, "OpenEnvironment",
 CellID->1781343525],
Cell[CellGroupData[{
Cell[96436, 2464, 517, 11, 43, "EnvironmentHeader",
 CellID->1935749337],
Cell[96956, 2477, 1338, 27, 33, "GlobalDeclaration",
 CellTags->{
  "ID:457420627", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->457420627],
Cell[98297, 2506, 1000, 23, 31, "GlobalDeclaration",
 CellTags->{
  "ID:176465234", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->176465234],
Cell[99300, 2531, 1892, 54, 58, "FormalTextInputFormula",
 CellTags->{
  "ID:478302118", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction with several multipliers"},
 CellID->478302118],
Cell[101195, 2587, 517, 18, 25, "EnvironmentText",
 CellID->624328162],
Cell[101715, 2607, 2024, 58, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:448370877", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetric closure of reducibility"},
 CellID->448370877],
Cell[103742, 2667, 1383, 39, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1537297721", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reducible1"},
 CellID->1537297721],
Cell[105128, 2708, 1599, 47, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:115218056", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reducible2"},
 CellID->115218056],
Cell[106730, 2757, 1894, 50, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:399781455", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "irreducibility"},
 CellID->399781455],
Cell[108627, 2809, 2205, 58, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:1867242863", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "irrelative"},
 CellID->1867242863],
Cell[110835, 2869, 1640, 52, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:1834165766", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative"},
 CellID->1834165766],
Cell[112478, 2923, 2270, 63, 82, "FormalTextInputFormula",
 CellTags->{
  "ID:1907029362", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common reducible"},
 CellID->1907029362],
Cell[114751, 2988, 3534, 94, 203, "FormalTextInputFormula",
 CellTags->{
  "ID:2013437879", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "non-trivial common reducible"},
 CellID->2013437879],
Cell[118288, 3084, 2119, 60, 83, "FormalTextInputFormula",
 CellTags->{
  "ID:991122907", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "minimal non-trivial common reducible"},
 CellID->991122907],
Cell[120410, 3146, 2361, 72, 125, "FormalTextInputFormula",
 CellTags->{
  "ID:1293246145", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "unary non-trivial common reducible"},
 CellID->1293246145],
Cell[122774, 3220, 1982, 59, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:894750987", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "unary minimal non-trivial common reducible"},
 CellID->894750987],
Cell[124759, 3281, 2543, 66, 109, "FormalTextInputFormula",
 CellTags->{
  "ID:779388596", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "critical pair multipliers"},
 CellID->779388596],
Cell[127305, 3349, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1118686892]
}, Open  ]],
Cell[127386, 3353, 48, 1, 16, "CloseEnvironment",
 CellID->297500119],
Cell[127437, 3356, 47, 1, 12, "OpenEnvironment",
 CellID->135917884],
Cell[CellGroupData[{
Cell[127509, 3361, 557, 13, 43, "EnvironmentHeader",
 CellID->1118462943],
Cell[128069, 3376, 1897, 34, 31, "GlobalDeclaration",
 CellTags->{
  "ID:1517092650", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1517092650],
Cell[129969, 3412, 2121, 68, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:935236171", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo element using multiplier"},
 CellID->935236171],
Cell[132093, 3482, 2134, 69, 59, "FormalTextInputFormula",
 CellTags->{
  "ID:320624715", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo set"},
 CellID->320624715],
Cell[134230, 3553, 3140, 84, 59, "FormalTextInputFormula",
 CellTags->{
  "ID:434515171", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo tuple"},
 CellID->434515171],
Cell[137373, 3639, 1426, 43, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:773828270", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReductor"},
 CellID->773828270],
Cell[138802, 3684, 623, 14, 25, "EnvironmentText",
 CellID->766317937],
Cell[139428, 3700, 2877, 88, 97, "FormalTextInputFormula",
 CellTags->{
  "ID:1813157585", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
   "reflexive-transitive closure of reduction relation"},
 CellID->1813157585],
Cell[142308, 3790, 2110, 68, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:370202704", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetric closure of reduction relation"},
 CellID->370202704],
Cell[144421, 3860, 2820, 86, 92, "FormalTextInputFormula",
 CellTags->{
  "ID:992563737", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
   "symmetric-reflexive-transitive closure of reduction relation"},
 CellID->992563737],
Cell[147244, 3948, 2467, 79, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:414012405", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common successor w.r.t. reduction relation"},
 CellID->414012405],
Cell[149714, 4029, 2406, 79, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:895393520", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexive-transitive closure of common successor"},
 CellID->895393520],
Cell[152123, 4110, 2426, 79, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:586157110", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "common predecessor w.r.t. reduction relation"},
 CellID->586157110],
Cell[154552, 4191, 2239, 74, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:2045822907", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
   "reflexive-transitive closure of common predecessor"},
 CellID->2045822907],
Cell[156794, 4267, 3941, 110, 96, "FormalTextInputFormula",
 CellTags->{
  "ID:463604795", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below"},
 CellID->463604795],
Cell[160738, 4379, 6169, 173, 240, "FormalTextInputFormula",
 CellTags->{
  "ID:1818365967", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde] multiplier"},
 CellID->1818365967],
Cell[166910, 4554, 2226, 67, 58, "FormalTextInputFormula",
 CellTags->{
  "ID:1533920470", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde]"},
 CellID->1533920470],
Cell[169139, 4623, 65, 1, 26, "EndEnvironmentMarker",
 CellID->110581283]
}, Open  ]],
Cell[169219, 4627, 49, 1, 16, "CloseEnvironment",
 CellID->1630010343],
Cell[169271, 4630, 1375, 37, 125, "Text",
 CellID->1007665725],
Cell[170649, 4669, 47, 1, 12, "OpenEnvironment",
 CellID->252379668],
Cell[CellGroupData[{
Cell[170721, 4674, 389, 9, 43, "EnvironmentHeader",
 CellID->82281139],
Cell[171113, 4685, 789, 21, 44, "GlobalDeclaration",
 CellTags->{
  "ID:81135181", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->81135181],
Cell[171905, 4708, 1143, 34, 49, "FormalTextInputFormula",
 CellTags->{
  "ID:1593192085", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red"},
 CellID->1593192085],
Cell[173051, 4744, 1434, 45, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:808948861", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset]"},
 CellID->808948861],
Cell[174488, 4791, 770, 21, 44, "GlobalDeclaration",
 CellTags->{
  "ID:301449743", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->301449743],
Cell[175261, 4814, 902, 25, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:218271861", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red tuples"},
 CellID->218271861],
Cell[176166, 4841, 1437, 44, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:20184186", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] tuples"},
 CellID->20184186],
Cell[177606, 4887, 65, 1, 26, "EndEnvironmentMarker",
 CellID->391369155]
}, Open  ]],
Cell[177686, 4891, 48, 1, 16, "CloseEnvironment",
 CellID->586839649]
}, Closed]],
Cell[CellGroupData[{
Cell[177771, 4897, 295, 4, 30, "Subsection",
 CellID->399888057],
Cell[178069, 4903, 347, 9, 70, "Text",
 CellID->815818515],
Cell[178419, 4914, 652, 16, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2086101595", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2086101595],
Cell[CellGroupData[{
Cell[179096, 4934, 311, 6, 70, "Subsubsection",
 CellID->1774554504],
Cell[179410, 4942, 46, 1, 70, "OpenEnvironment",
 CellID->30312876],
Cell[CellGroupData[{
Cell[179481, 4947, 396, 9, 70, "EnvironmentHeader",
 CellID->26444589],
Cell[179880, 4958, 987, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:690041737", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets isReductor"},
 CellID->690041737],
Cell[180870, 4988, 985, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:5128300", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainTuples isReductor"},
 CellID->5128300],
Cell[181858, 5018, 64, 1, 70, "EndEnvironmentMarker",
 CellID->21022560]
}, Open  ]],
Cell[181937, 5022, 48, 1, 70, "CloseEnvironment",
 CellID->301523614],
Cell[CellGroupData[{
Cell[182010, 5027, 2733, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:690041737", "Proof|ID:690041737-1"},
 CellID->323627089],
Cell[184746, 5107, 479, 12, 70, "ProofInfoBody",
 CellID->95218679]
}, Closed]],
Cell[CellGroupData[{
Cell[185262, 5124, 2731, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:5128300", "Proof|ID:5128300-1"},
 CellID->485768995],
Cell[187996, 5204, 478, 12, 70, "ProofInfoBody",
 CellID->590295287]
}, Closed]],
Cell[188489, 5219, 47, 1, 70, "OpenEnvironment",
 CellID->270038199],
Cell[CellGroupData[{
Cell[188561, 5224, 438, 10, 70, "EnvironmentHeader",
 CellID->917364838],
Cell[189002, 5236, 2813, 79, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:906005529", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo singleton", "IR"},
 CellID->906005529],
Cell[191818, 5317, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1135647683]
}, Open  ]],
Cell[191899, 5321, 49, 1, 70, "CloseEnvironment",
 CellID->1526698344],
Cell[CellGroupData[{
Cell[191973, 5326, 4814, 128, 70, "ProofInfo",
 CellTags->{"Proof|ID:906005529", "Proof|ID:906005529-1"},
 CellID->520373627],
Cell[196790, 5456, 480, 12, 70, "ProofInfoBody",
 CellID->177234516]
}, Closed]],
Cell[197285, 5471, 47, 1, 70, "OpenEnvironment",
 CellID->701877035],
Cell[CellGroupData[{
Cell[197357, 5476, 393, 9, 70, "EnvironmentHeader",
 CellID->1394100531],
Cell[197753, 5487, 720, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:708647877", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->708647877],
Cell[198476, 5507, 767, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:508075012", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->508075012],
Cell[199246, 5529, 1551, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:347299573", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo tuple alternative", "IR"},
 CellID->347299573],
Cell[200800, 5580, 708, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1756538453", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1756538453],
Cell[201511, 5600, 1161, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:200776309", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] TupleToSet", "IR"},
 CellID->200776309],
Cell[202675, 5638, 1187, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:989570066", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] TupleToSet", "IR"},
 CellID->989570066],
Cell[203865, 5676, 1181, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:359412397", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow]\[RawStar] TupleToSet", "IR"},
 CellID->359412397],
Cell[205049, 5714, 1212, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1876144273", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow]\[RawStar] TupleToSet", "IR"},
 CellID->1876144273],
Cell[206264, 5752, 1944, 65, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:495509420", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below TupleToSet", "IR"},
 CellID->495509420],
Cell[208211, 5819, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1713169042]
}, Open  ]],
Cell[208292, 5823, 48, 1, 70, "CloseEnvironment",
 CellID->434319633],
Cell[CellGroupData[{
Cell[208365, 5828, 5202, 137, 70, "ProofInfo",
 CellTags->{"Proof|ID:347299573", "Proof|ID:347299573-1"},
 CellID->641607246],
Cell[213570, 5967, 480, 12, 70, "ProofInfoBody",
 CellID->129731929]
}, Closed]],
Cell[CellGroupData[{
Cell[214087, 5984, 4223, 112, 70, "ProofInfo",
 CellTags->{"Proof|ID:200776309", "Proof|ID:200776309-1"},
 CellID->376820956],
Cell[218313, 6098, 479, 12, 70, "ProofInfoBody",
 CellID->46818332]
}, Closed]],
Cell[CellGroupData[{
Cell[218829, 6115, 4279, 113, 70, "ProofInfo",
 CellTags->{"Proof|ID:989570066", "Proof|ID:989570066-1"},
 CellID->450749896],
Cell[223111, 6230, 481, 12, 70, "ProofInfoBody",
 CellID->1131081333]
}, Closed]],
Cell[CellGroupData[{
Cell[223629, 6247, 4325, 112, 70, "ProofInfo",
 CellTags->{"Proof|ID:359412397", "Proof|ID:359412397-1"},
 CellID->1501514132],
Cell[227957, 6361, 481, 12, 70, "ProofInfoBody",
 CellID->1241556193]
}, Closed]],
Cell[CellGroupData[{
Cell[228475, 6378, 4381, 113, 70, "ProofInfo",
 CellTags->{"Proof|ID:1876144273", "Proof|ID:1876144273-1"},
 CellID->541596138],
Cell[232859, 6493, 482, 12, 70, "ProofInfoBody",
 CellID->1335127868]
}, Closed]],
Cell[CellGroupData[{
Cell[233378, 6510, 5028, 127, 70, "ProofInfo",
 CellTags->{"Proof|ID:495509420", "Proof|ID:495509420-1"},
 CellID->551004923],
Cell[238409, 6639, 479, 12, 70, "ProofInfoBody",
 CellID->20568225]
}, Closed]],
Cell[238903, 6654, 47, 1, 70, "OpenEnvironment",
 CellID->770283335],
Cell[CellGroupData[{
Cell[238975, 6659, 652, 13, 70, "EnvironmentHeader",
 CellID->1406215010],
Cell[239630, 6674, 672, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:467637718", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->467637718],
Cell[240305, 6693, 795, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1182167952", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1182167952],
Cell[241103, 6716, 1401, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:178557746", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction modulo set alternative", "IR"},
 CellID->178557746],
Cell[242507, 6763, 1169, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1019233204", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[Union]"},
 CellID->1019233204],
Cell[243679, 6802, 1194, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:331036245", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] \[Union]"},
 CellID->331036245],
Cell[244876, 6841, 1185, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:78564808", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow]\[RawStar] \[Union]"},
 CellID->78564808],
Cell[246064, 6880, 1264, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:626278069", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow]\[RawStar] \[Union]"},
 CellID->626278069],
Cell[247331, 6920, 1922, 63, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:356849840", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below \[Union]"},
 CellID->356849840],
Cell[249256, 6985, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1824911811]
}, Open  ]],
Cell[249337, 6989, 48, 1, 70, "CloseEnvironment",
 CellID->933018018],
Cell[CellGroupData[{
Cell[249410, 6994, 4661, 125, 70, "ProofInfo",
 CellTags->{"Proof|ID:178557746", "Proof|ID:178557746-1"},
 CellID->93489966],
Cell[254074, 7121, 480, 12, 70, "ProofInfoBody",
 CellID->261267025]
}, Closed]],
Cell[CellGroupData[{
Cell[254591, 7138, 4567, 121, 70, "ProofInfo",
 CellTags->{"Proof|ID:1019233204", "Proof|ID:1019233204-1"},
 CellID->9184570],
Cell[259161, 7261, 481, 12, 70, "ProofInfoBody",
 CellID->310726113]
}, Closed]],
Cell[CellGroupData[{
Cell[259679, 7278, 4608, 120, 70, "ProofInfo",
 CellTags->{"Proof|ID:331036245", "Proof|ID:331036245-1"},
 CellID->970023700],
Cell[264290, 7400, 480, 12, 70, "ProofInfoBody",
 CellID->221863878]
}, Closed]],
Cell[CellGroupData[{
Cell[264807, 7417, 4857, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:78564808", "Proof|ID:78564808-1"},
 CellID->925666063],
Cell[269667, 7543, 479, 12, 70, "ProofInfoBody",
 CellID->477674065]
}, Closed]],
Cell[CellGroupData[{
Cell[270183, 7560, 4911, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:626278069", "Proof|ID:626278069-1"},
 CellID->1953121485],
Cell[275097, 7686, 480, 12, 70, "ProofInfoBody",
 CellID->785566982]
}, Closed]],
Cell[CellGroupData[{
Cell[275614, 7703, 5560, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:356849840", "Proof|ID:356849840-1"},
 CellID->205255475],
Cell[281177, 7843, 480, 12, 70, "ProofInfoBody",
 CellID->979272693]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[281706, 7861, 141, 2, 70, "Subsubsection",
 CellID->828082569],
Cell[281850, 7865, 48, 1, 70, "OpenEnvironment",
 CellID->1774903907],
Cell[CellGroupData[{
Cell[281923, 7870, 527, 14, 70, "EnvironmentHeader",
 CellID->1159810905],
Cell[282453, 7886, 952, 24, 70, "GlobalDeclaration",
 CellTags->{
  "ID:366153355", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->366153355],
Cell[283408, 7912, 699, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:221340803", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->221340803],
Cell[284110, 7932, 1219, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:328195853", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "2"},
 CellID->328195853],
Cell[285332, 7965, 997, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1692465705", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "3"},
 CellID->1692465705],
Cell[286332, 7995, 996, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:188743948", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "4"},
 CellID->188743948],
Cell[287331, 8025, 1338, 41, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:115635911", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "5"},
 CellID->115635911],
Cell[288672, 8068, 1054, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1118510950", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isCommonReducible symmetric"},
 CellID->1118510950],
Cell[289729, 8098, 1007, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:113485898", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isNTCR symmetric"},
 CellID->113485898],
Cell[290739, 8128, 1016, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2056330192", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isMNTCR symmetric"},
 CellID->2056330192],
Cell[291758, 8158, 1420, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:463339584", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers symmetric"},
 CellID->463339584],
Cell[293181, 8202, 1867, 58, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:449270875", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers sufficient"},
 CellID->449270875],
Cell[295051, 8262, 65, 1, 70, "EndEnvironmentMarker",
 CellID->838304061]
}, Open  ]],
Cell[295131, 8266, 47, 1, 70, "CloseEnvironment",
 CellID->44286942],
Cell[CellGroupData[{
Cell[295203, 8271, 5357, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:328195853", "Proof|ID:328195853-1"},
 CellID->667808408],
Cell[300563, 8405, 480, 12, 70, "ProofInfoBody",
 CellID->342092220]
}, Closed]],
Cell[CellGroupData[{
Cell[301080, 8422, 5322, 133, 70, "ProofInfo",
 CellTags->{"Proof|ID:1692465705", "Proof|ID:1692465705-1"},
 CellID->982819604],
Cell[306405, 8557, 482, 12, 70, "ProofInfoBody",
 CellID->1849675645]
}, Closed]],
Cell[CellGroupData[{
Cell[306924, 8574, 5318, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:188743948", "Proof|ID:188743948-1"},
 CellID->1789453742],
Cell[312245, 8708, 480, 12, 70, "ProofInfoBody",
 CellID->254539763]
}, Closed]],
Cell[CellGroupData[{
Cell[312762, 8725, 6602, 156, 70, "ProofInfo",
 CellTags->{"Proof|ID:115635911", "Proof|ID:115635911-1"},
 CellID->1648327750],
Cell[319367, 8883, 480, 12, 70, "ProofInfoBody",
 CellID->317536928]
}, Closed]],
Cell[CellGroupData[{
Cell[319884, 8900, 5528, 135, 70, "ProofInfo",
 CellTags->{"Proof|ID:1118510950", "Proof|ID:1118510950-1"},
 CellID->144468455],
Cell[325415, 9037, 482, 12, 70, "ProofInfoBody",
 CellID->1943003684]
}, Closed]],
Cell[CellGroupData[{
Cell[325934, 9054, 5454, 134, 70, "ProofInfo",
 CellTags->{"Proof|ID:113485898", "Proof|ID:113485898-1"},
 CellID->1668243539],
Cell[331391, 9190, 481, 12, 70, "ProofInfoBody",
 CellID->1751898811]
}, Closed]],
Cell[CellGroupData[{
Cell[331909, 9207, 5469, 135, 70, "ProofInfo",
 CellTags->{"Proof|ID:2056330192", "Proof|ID:2056330192-1"},
 CellID->1822040650],
Cell[337381, 9344, 482, 12, 70, "ProofInfoBody",
 CellID->1848124281]
}, Closed]],
Cell[CellGroupData[{
Cell[337900, 9361, 6754, 158, 70, "ProofInfo",
 CellTags->{"Proof|ID:463339584", "Proof|ID:463339584-1"},
 CellID->2055115730],
Cell[344657, 9521, 481, 12, 70, "ProofInfoBody",
 CellID->1154563333]
}, Closed]],
Cell[CellGroupData[{
Cell[345175, 9538, 8468, 194, 70, "ProofInfo",
 CellTags->{"Proof|ID:449270875", "Proof|ID:449270875-1"},
 CellID->1693763802],
Cell[353646, 9734, 480, 12, 70, "ProofInfoBody",
 CellID->874031660]
}, Closed]],
Cell[354141, 9749, 47, 1, 70, "OpenEnvironment",
 CellID->431320600],
Cell[CellGroupData[{
Cell[354213, 9754, 448, 12, 70, "EnvironmentHeader",
 CellID->1620999719],
Cell[354664, 9768, 751, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:523863903", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->523863903],
Cell[355418, 9790, 789, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1519187240", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1519187240],
Cell[356210, 9812, 1052, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1878546324", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "2'"},
 CellID->1878546324],
Cell[357265, 9842, 1025, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1781176225", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "3'"},
 CellID->1781176225],
Cell[358293, 9872, 1032, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1582995786", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "4'"},
 CellID->1582995786],
Cell[359328, 9902, 1370, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2020067658", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "5'"},
 CellID->2020067658],
Cell[360701, 9946, 64, 1, 70, "EndEnvironmentMarker",
 CellID->81893628]
}, Open  ]],
Cell[360780, 9950, 49, 1, 70, "CloseEnvironment",
 CellID->1310599982],
Cell[CellGroupData[{
Cell[360854, 9955, 5875, 147, 70, "ProofInfo",
 CellTags->{"Proof|ID:1878546324", "Proof|ID:1878546324-1"},
 CellID->1762740644],
Cell[366732, 10104, 482, 12, 70, "ProofInfoBody",
 CellID->1512786933]
}, Closed]],
Cell[CellGroupData[{
Cell[367251, 10121, 5831, 147, 70, "ProofInfo",
 CellTags->{"Proof|ID:1781176225", "Proof|ID:1781176225-1"},
 CellID->1337859754],
Cell[373085, 10270, 481, 12, 70, "ProofInfoBody",
 CellID->948047327]
}, Closed]],
Cell[CellGroupData[{
Cell[373603, 10287, 5835, 147, 70, "ProofInfo",
 CellTags->{"Proof|ID:1582995786", "Proof|ID:1582995786-1"},
 CellID->1038215936],
Cell[379441, 10436, 480, 12, 70, "ProofInfoBody",
 CellID->41089318]
}, Closed]],
Cell[CellGroupData[{
Cell[379958, 10453, 7103, 171, 70, "ProofInfo",
 CellTags->{"Proof|ID:2020067658", "Proof|ID:2020067658-1"},
 CellID->1751126895],
Cell[387064, 10626, 482, 12, 70, "ProofInfoBody",
 CellID->1401855599]
}, Closed]],
Cell[387561, 10641, 48, 1, 70, "OpenEnvironment",
 CellID->1693676355],
Cell[CellGroupData[{
Cell[387634, 10646, 526, 12, 70, "EnvironmentHeader",
 CellID->320285829],
Cell[388163, 10660, 730, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1823284442", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1823284442],
Cell[388896, 10682, 762, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:284217868", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->284217868],
Cell[389661, 10704, 1035, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:628244416", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "6"},
 CellID->628244416],
Cell[390699, 10732, 2058, 60, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1255486983", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative-irrelative 2"},
 CellID->1255486983],
Cell[392760, 10794, 65, 1, 70, "EndEnvironmentMarker",
 CellID->864502048]
}, Open  ]],
Cell[392840, 10798, 49, 1, 70, "CloseEnvironment",
 CellID->2028899293],
Cell[CellGroupData[{
Cell[392914, 10803, 5247, 136, 70, "ProofInfo",
 CellTags->{"Proof|ID:628244416", "Proof|ID:628244416-1"},
 CellID->1993474493],
Cell[398164, 10941, 480, 12, 70, "ProofInfoBody",
 CellID->661346101]
}, Closed]],
Cell[CellGroupData[{
Cell[398681, 10958, 8860, 207, 70, "ProofInfo",
 CellTags->{"Proof|ID:1255486983", "Proof|ID:1255486983-1"},
 CellID->558484412],
Cell[407544, 11167, 482, 12, 70, "ProofInfoBody",
 CellID->1567652911]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[408087, 11186, 420, 8, 30, "Subsection",
 CellID->359523023],
Cell[408510, 11196, 341, 8, 70, "Text",
 CellID->1577638100],
Cell[408854, 11206, 740, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1382908475", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1382908475],
Cell[409597, 11226, 47, 1, 70, "OpenEnvironment",
 CellID->955233445],
Cell[CellGroupData[{
Cell[409669, 11231, 486, 11, 70, "EnvironmentHeader",
 CellID->23280017],
Cell[410158, 11244, 853, 22, 70, "GlobalDeclaration",
 CellTags->{
  "ID:442895874", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->442895874],
Cell[411014, 11268, 816, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:927154658", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->927154658],
Cell[411833, 11291, 867, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1300106055", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM 0"},
 CellID->1300106055],
Cell[412703, 11317, 1223, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:438711618", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM 1"},
 CellID->438711618],
Cell[413929, 11356, 1524, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2101232329", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM alternative"},
 CellID->2101232329],
Cell[415456, 11403, 1487, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:96556784", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM+1"},
 CellID->96556784],
Cell[416946, 11447, 1843, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1465630278", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM alternative II"},
 CellID->1465630278],
Cell[418792, 11505, 1152, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1952076718", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "membership reduceM", "IR"},
 CellID->1952076718],
Cell[419947, 11539, 792, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1110504237", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1110504237],
Cell[420742, 11562, 1225, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1236615595", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM-1"},
 CellID->1236615595],
Cell[421970, 11600, 1127, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:203149419", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduceM-1 II"},
 CellID->203149419],
Cell[423100, 11635, 2109, 64, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:333495152", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible reduceM"},
 CellID->333495152],
Cell[425212, 11701, 2187, 65, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:332101094", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM"},
 CellID->332101094],
Cell[427402, 11768, 3277, 99, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:296464255", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM I"},
 CellID->296464255],
Cell[430682, 11869, 2682, 81, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:194930655", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM II"},
 CellID->194930655],
Cell[433367, 11952, 2870, 87, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:935171195", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred reduceM III"},
 CellID->935171195],
Cell[436240, 12041, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1917615946]
}, Open  ]],
Cell[436321, 12045, 48, 1, 70, "CloseEnvironment",
 CellID->176929079],
Cell[CellGroupData[{
Cell[436394, 12050, 4093, 111, 70, "ProofInfo",
 CellTags->{"Proof|ID:1300106055", "Proof|ID:1300106055-1"},
 CellID->587044218],
Cell[440490, 12163, 525, 13, 70, "ProofInfoBody",
 CellID->691055095]
}, Closed]],
Cell[CellGroupData[{
Cell[441052, 12181, 5331, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:438711618", "Proof|ID:438711618-1"},
 CellID->457182048],
Cell[446386, 12321, 480, 12, 70, "ProofInfoBody",
 CellID->871927811]
}, Closed]],
Cell[CellGroupData[{
Cell[446903, 12338, 6500, 165, 70, "ProofInfo",
 CellTags->{"Proof|ID:2101232329", "Proof|ID:2101232329-1"},
 CellID->1720715120],
Cell[453406, 12505, 481, 12, 70, "ProofInfoBody",
 CellID->394118706]
}, Closed]],
Cell[CellGroupData[{
Cell[453924, 12522, 6545, 166, 70, "ProofInfo",
 CellTags->{"Proof|ID:96556784", "Proof|ID:96556784-1"},
 CellID->194852796],
Cell[460472, 12690, 479, 12, 70, "ProofInfoBody",
 CellID->676400811]
}, Closed]],
Cell[CellGroupData[{
Cell[460988, 12707, 7994, 197, 70, "ProofInfo",
 CellTags->{"Proof|ID:1465630278", "Proof|ID:1465630278-1"},
 CellID->1634286780],
Cell[468985, 12906, 482, 12, 70, "ProofInfoBody",
 CellID->1019435258]
}, Closed]],
Cell[CellGroupData[{
Cell[469504, 12923, 4865, 125, 70, "ProofInfo",
 CellTags->{"Proof|ID:1952076718", "Proof|ID:1952076718-1"},
 CellID->1918441122],
Cell[474372, 13050, 481, 12, 70, "ProofInfoBody",
 CellID->519581729]
}, Closed]],
Cell[CellGroupData[{
Cell[474890, 13067, 6661, 171, 70, "ProofInfo",
 CellTags->{"Proof|ID:1236615595", "Proof|ID:1236615595-1"},
 CellID->536851896],
Cell[481554, 13240, 481, 12, 70, "ProofInfoBody",
 CellID->943705171]
}, Closed]],
Cell[CellGroupData[{
Cell[482072, 13257, 6361, 158, 70, "ProofInfo",
 CellTags->{"Proof|ID:203149419", "Proof|ID:203149419-1"},
 CellID->916037400],
Cell[488436, 13417, 523, 13, 70, "ProofInfoBody",
 CellID->1494767151]
}, Closed]],
Cell[CellGroupData[{
Cell[488996, 13435, 10693, 254, 70, "ProofInfo",
 CellTags->{"Proof|ID:333495152", "Proof|ID:333495152-1"},
 CellID->193609441],
Cell[499692, 13691, 481, 12, 70, "ProofInfoBody",
 CellID->1085025219]
}, Closed]],
Cell[CellGroupData[{
Cell[500210, 13708, 11549, 264, 70, "ProofInfo",
 CellTags->{"Proof|ID:332101094", "Proof|ID:332101094-1"},
 CellID->2106893222],
Cell[511762, 13974, 481, 12, 70, "ProofInfoBody",
 CellID->1457981415]
}, Closed]],
Cell[CellGroupData[{
Cell[512280, 13991, 15830, 364, 70, "ProofInfo",
 CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
 CellID->1877618114],
Cell[528113, 14357, 481, 12, 70, "ProofInfoBody",
 CellID->1831291740]
}, Closed]],
Cell[CellGroupData[{
Cell[528631, 14374, 13369, 308, 70, "ProofInfo",
 CellTags->{"Proof|ID:194930655", "Proof|ID:194930655-1"},
 CellID->230812744],
Cell[542003, 14684, 480, 12, 70, "ProofInfoBody",
 CellID->942266403]
}, Closed]],
Cell[CellGroupData[{
Cell[542520, 14701, 14395, 330, 70, "ProofInfo",
 CellTags->{"Proof|ID:935171195", "Proof|ID:935171195-1"},
 CellID->577756398],
Cell[556918, 15033, 481, 12, 70, "ProofInfoBody",
 CellID->1380815439],
Cell[557402, 15047, 583, 11, 70, "Message",
 CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
 CellID->20878089],
Cell[557988, 15060, 585, 11, 70, "Message",
 CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
 CellID->1281008759],
Cell[558576, 15073, 585, 11, 70, "Message",
 CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
 CellID->1946324498],
Cell[559164, 15086, 597, 12, 70, "Message",
 CellTags->{"Proof|ID:296464255", "Proof|ID:296464255-1"},
 CellID->1486921856]
}, Closed]],
Cell[559776, 15101, 48, 1, 70, "OpenEnvironment",
 CellID->1798794743],
Cell[CellGroupData[{
Cell[559849, 15106, 407, 10, 70, "EnvironmentHeader",
 CellID->1097525720],
Cell[560259, 15118, 887, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1368508872", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1368508872],
Cell[561149, 15143, 1430, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1610984937", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areIrrelative symmetric"},
 CellID->1610984937],
Cell[562582, 15185, 1060, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:981548331", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCorrelative symmetric"},
 CellID->981548331],
Cell[563645, 15214, 65, 1, 70, "EndEnvironmentMarker",
 CellID->453207686]
}, Open  ]],
Cell[563725, 15218, 48, 1, 70, "CloseEnvironment",
 CellID->231330468],
Cell[CellGroupData[{
Cell[563798, 15223, 7069, 171, 70, "ProofInfo",
 CellTags->{"Proof|ID:1610984937", "Proof|ID:1610984937-1"},
 CellID->1688165426],
Cell[570870, 15396, 482, 12, 70, "ProofInfoBody",
 CellID->1794986575]
}, Closed]],
Cell[CellGroupData[{
Cell[571389, 15413, 4589, 118, 70, "ProofInfo",
 CellTags->{"Proof|ID:981548331", "Proof|ID:981548331-1"},
 CellID->437316908],
Cell[575981, 15533, 480, 12, 70, "ProofInfoBody",
 CellID->337585423]
}, Closed]],
Cell[576476, 15548, 48, 1, 70, "OpenEnvironment",
 CellID->1696462827],
Cell[CellGroupData[{
Cell[576549, 15553, 550, 13, 70, "EnvironmentHeader",
 CellID->1520519732],
Cell[577102, 15568, 976, 24, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1547543437", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1547543437],
Cell[578081, 15594, 1300, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1699369800", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible \[RightArrow]"},
 CellID->1699369800],
Cell[579384, 15634, 1377, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1596540807", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] isReducible"},
 CellID->1596540807],
Cell[580764, 15679, 1092, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1641388157", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible symred"},
 CellID->1641388157],
Cell[581859, 15712, 1479, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2087655986", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible alternative", "IR"},
 CellID->2087655986],
Cell[583341, 15758, 2777, 85, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:584152369", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers \[RightArrow] tuple"},
 CellID->584152369],
Cell[586121, 15845, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1376947800]
}, Open  ]],
Cell[586202, 15849, 49, 1, 70, "CloseEnvironment",
 CellID->1403180676],
Cell[CellGroupData[{
Cell[586276, 15854, 5258, 137, 70, "ProofInfo",
 CellTags->{"Proof|ID:1699369800", "Proof|ID:1699369800-1"},
 CellID->199699432],
Cell[591537, 15993, 480, 12, 70, "ProofInfoBody",
 CellID->60679348]
}, Closed]],
Cell[CellGroupData[{
Cell[592054, 16010, 6412, 159, 70, "ProofInfo",
 CellTags->{"Proof|ID:1596540807", "Proof|ID:1596540807-1"},
 CellID->314505815],
Cell[598469, 16171, 480, 12, 70, "ProofInfoBody",
 CellID->54928433]
}, Closed]],
Cell[CellGroupData[{
Cell[598986, 16188, 5339, 135, 70, "ProofInfo",
 CellTags->{"Proof|ID:1641388157", "Proof|ID:1641388157-1"},
 CellID->20930826],
Cell[604328, 16325, 482, 12, 70, "ProofInfoBody",
 CellID->1555086701]
}, Closed]],
Cell[CellGroupData[{
Cell[604847, 16342, 4738, 127, 70, "ProofInfo",
 CellTags->{"Proof|ID:2087655986", "Proof|ID:2087655986-1"},
 CellID->65047792],
Cell[609588, 16471, 481, 12, 70, "ProofInfoBody",
 CellID->650270676]
}, Closed]],
Cell[CellGroupData[{
Cell[610106, 16488, 9817, 234, 70, "ProofInfo",
 CellTags->{"Proof|ID:584152369", "Proof|ID:584152369-1"},
 CellID->444231087],
Cell[619926, 16724, 480, 12, 70, "ProofInfoBody",
 CellID->499606199]
}, Closed]],
Cell[620421, 16739, 48, 1, 70, "OpenEnvironment",
 CellID->1668167977],
Cell[CellGroupData[{
Cell[620494, 16744, 563, 16, 70, "EnvironmentHeader",
 CellID->181072362],
Cell[621060, 16762, 846, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:593421628", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->593421628],
Cell[621909, 16787, 1248, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:680005293", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "1"},
 CellID->680005293],
Cell[623160, 16829, 1078, 33, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1320396264", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1320396264],
Cell[624241, 16864, 1185, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1127232822", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] reductor"},
 CellID->1127232822],
Cell[625429, 16901, 807, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:256297052", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Succeeds] reductor"},
 CellID->256297052],
Cell[626239, 16924, 65, 1, 70, "EndEnvironmentMarker",
 CellID->963051773]
}, Open  ]],
Cell[626319, 16928, 48, 1, 70, "CloseEnvironment",
 CellID->895896855],
Cell[CellGroupData[{
Cell[626392, 16933, 5315, 136, 70, "ProofInfo",
 CellTags->{"Proof|ID:680005293", "Proof|ID:680005293-1"},
 CellID->41814150],
Cell[631710, 17071, 480, 12, 70, "ProofInfoBody",
 CellID->438053564]
}, Closed]],
Cell[CellGroupData[{
Cell[632227, 17088, 5329, 139, 70, "ProofInfo",
 CellTags->{"Proof|ID:1127232822", "Proof|ID:1127232822-1"},
 CellID->625038390],
Cell[637559, 17229, 481, 12, 70, "ProofInfoBody",
 CellID->111447419]
}, Closed]],
Cell[CellGroupData[{
Cell[638077, 17246, 4460, 118, 70, "ProofInfo",
 CellTags->{"Proof|ID:256297052", "Proof|ID:256297052-1"},
 CellID->256602949],
Cell[642540, 17366, 480, 12, 70, "ProofInfoBody",
 CellID->162186853]
}, Closed]],
Cell[643035, 17381, 48, 1, 70, "OpenEnvironment",
 CellID->1503630565],
Cell[643086, 17384, 521, 11, 70, "EnvironmentHeader",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->879564337],
Cell[643610, 17397, 1085, 26, 70, "GlobalDeclaration",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:1083403145", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1083403145],
Cell[644698, 17425, 1061, 28, 70, "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:1592577568", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets red", "IR"},
 CellID->1592577568],
Cell[645762, 17455, 965, 25, 70, "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:698444072", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red isSet", "IR"},
 CellID->698444072],
Cell[646730, 17482, 1781, 53, 70, "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:277656140", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red"},
 CellID->277656140],
Cell[648514, 17537, 1772, 54, 70, "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:454978652", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red singleton"},
 CellID->454978652],
Cell[650289, 17593, 3071, 92, 70, "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:259827218", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red abstraction term"},
 CellID->259827218],
Cell[653363, 17687, 1554, 44, 70, "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:1273500442", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red \[Union]"},
 CellID->1273500442],
Cell[654920, 17733, 1143, 31, 70, "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:1877397922", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Union]"},
 CellID->1877397922],
Cell[656066, 17766, 2110, 64, 70, "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:2121666706", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red \[Union] singleton"},
 CellID->2121666706],
Cell[658179, 17832, 1088, 29, 70, "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:947744040", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SubsetEqual] red \[SubsetEqual]"},
 CellID->947744040],
Cell[659270, 17863, 2045, 61, 70, "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:52516181", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset]"},
 CellID->52516181],
Cell[661318, 17926, 2248, 69, 70, "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:194436663", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset] \[Union] singleton"},
 CellID->194436663],
Cell[663569, 17997, 1794, 53, 70, "FormalTextInputFormula",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellTags->{
  "ID:844584535", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[Subset] \[Union] singleton II"},
 CellID->844584535],
Cell[665366, 18052, 175, 3, 70, "EndEnvironmentMarker",
 CellGroupingRules->{$CellContext`GroupTogetherGrouping, 10000.},
 CellID->212448877],
Cell[665544, 18057, 48, 1, 70, "CloseEnvironment",
 CellID->943439096],
Cell[CellGroupData[{
Cell[665617, 18062, 3174, 89, 70, "ProofInfo",
 CellTags->{"Proof|ID:1592577568", "Proof|ID:1592577568-1"},
 CellID->272356749],
Cell[668794, 18153, 481, 12, 70, "ProofInfoBody",
 CellID->513310847]
}, Closed]],
Cell[CellGroupData[{
Cell[669312, 18170, 2889, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:698444072", "Proof|ID:698444072-1"},
 CellID->6276063],
Cell[672204, 18254, 479, 12, 70, "ProofInfoBody",
 CellID->85725627]
}, Closed]],
Cell[CellGroupData[{
Cell[672720, 18271, 4978, 128, 70, "ProofInfo",
 CellTags->{"Proof|ID:277656140", "Proof|ID:277656140-1"},
 CellID->45788410],
Cell[677701, 18401, 480, 12, 70, "ProofInfoBody",
 CellID->230626054]
}, Closed]],
Cell[CellGroupData[{
Cell[678218, 18418, 4420, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:454978652", "Proof|ID:454978652-1"},
 CellID->54717562],
Cell[682641, 18536, 477, 12, 70, "ProofInfoBody",
 CellID->889655]
}, Closed]],
Cell[CellGroupData[{
Cell[683155, 18553, 8263, 199, 70, "ProofInfo",
 CellTags->{"Proof|ID:259827218", "Proof|ID:259827218-1"},
 CellID->183930652],
Cell[691421, 18754, 480, 12, 70, "ProofInfoBody",
 CellID->803294713]
}, Closed]],
Cell[CellGroupData[{
Cell[691938, 18771, 5332, 135, 70, "ProofInfo",
 CellTags->{"Proof|ID:1273500442", "Proof|ID:1273500442-1"},
 CellID->172667338],
Cell[697273, 18908, 481, 12, 70, "ProofInfoBody",
 CellID->536407252]
}, Closed]],
Cell[CellGroupData[{
Cell[697791, 18925, 4354, 115, 70, "ProofInfo",
 CellTags->{"Proof|ID:1877397922", "Proof|ID:1877397922-1"},
 CellID->133276848],
Cell[702148, 19042, 481, 12, 70, "ProofInfoBody",
 CellID->173577853]
}, Closed]],
Cell[CellGroupData[{
Cell[702666, 19059, 5840, 147, 70, "ProofInfo",
 CellTags->{"Proof|ID:2121666706", "Proof|ID:2121666706-1"},
 CellID->185736564],
Cell[708509, 19208, 481, 12, 70, "ProofInfoBody",
 CellID->133384348]
}, Closed]],
Cell[CellGroupData[{
Cell[709027, 19225, 4177, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:947744040", "Proof|ID:947744040-1"},
 CellID->249774080],
Cell[713207, 19337, 480, 12, 70, "ProofInfoBody",
 CellID->411810760]
}, Closed]],
Cell[CellGroupData[{
Cell[713724, 19354, 6543, 164, 70, "ProofInfo",
 CellTags->{"Proof|ID:52516181", "Proof|ID:52516181-1"},
 CellID->50697679],
Cell[720270, 19520, 479, 12, 70, "ProofInfoBody",
 CellID->751899172]
}, Closed]],
Cell[CellGroupData[{
Cell[720786, 19537, 6353, 161, 70, "ProofInfo",
 CellTags->{"Proof|ID:194436663", "Proof|ID:194436663-1"},
 CellID->210264695],
Cell[727142, 19700, 479, 12, 70, "ProofInfoBody",
 CellID->22396559]
}, Closed]],
Cell[CellGroupData[{
Cell[727658, 19717, 6028, 152, 70, "ProofInfo",
 CellTags->{"Proof|ID:844584535", "Proof|ID:844584535-1"},
 CellID->34854128],
Cell[733689, 19871, 480, 12, 70, "ProofInfoBody",
 CellID->786223699]
}, Closed]],
Cell[734184, 19886, 47, 1, 70, "OpenEnvironment",
 CellID->147378381],
Cell[CellGroupData[{
Cell[734256, 19891, 439, 10, 70, "EnvironmentHeader",
 CellID->1275480004],
Cell[734698, 19903, 2199, 67, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:632054408", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red \[SubsetEqual] transitive"},
 CellID->632054408],
Cell[736900, 19972, 772, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:387695114", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->387695114],
Cell[737675, 19994, 512, 15, 70, "EnvironmentText",
 CellID->44299611],
Cell[738190, 20011, 1984, 61, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1177850337", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "setSeq subsequence red"},
 CellID->1177850337],
Cell[740177, 20074, 714, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1483708233", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1483708233],
Cell[740894, 20094, 1060, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:697433658", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red setSeq \[SubsetEqual]"},
 CellID->697433658],
Cell[741957, 20126, 2089, 63, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:35999503", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] red setSeq"},
 CellID->35999503],
Cell[744049, 20191, 2200, 67, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:602168712", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "red setSeq \[Subset]"},
 CellID->602168712],
Cell[746252, 20260, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1700325855]
}, Open  ]],
Cell[746333, 20264, 48, 1, 70, "CloseEnvironment",
 CellID->472609169],
Cell[CellGroupData[{
Cell[746406, 20269, 6960, 174, 70, "ProofInfo",
 CellTags->{"Proof|ID:632054408", "Proof|ID:632054408-1"},
 CellID->70698580],
Cell[753369, 20445, 479, 12, 70, "ProofInfoBody",
 CellID->51683851]
}, Closed]],
Cell[CellGroupData[{
Cell[753885, 20462, 7392, 186, 70, "ProofInfo",
 CellTags->{"Proof|ID:1177850337", "Proof|ID:1177850337-1"},
 CellID->1013465541],
Cell[761280, 20650, 482, 12, 70, "ProofInfoBody",
 CellID->1284941524]
}, Closed]],
Cell[CellGroupData[{
Cell[761799, 20667, 4199, 112, 70, "ProofInfo",
 CellTags->{"Proof|ID:697433658", "Proof|ID:697433658-1"},
 CellID->179082785],
Cell[766001, 20781, 480, 12, 70, "ProofInfoBody",
 CellID->182798619]
}, Closed]],
Cell[CellGroupData[{
Cell[766518, 20798, 6144, 152, 70, "ProofInfo",
 CellTags->{"Proof|ID:35999503", "Proof|ID:35999503-1"},
 CellID->361034880],
Cell[772665, 20952, 479, 12, 70, "ProofInfoBody",
 CellID->149032911]
}, Closed]],
Cell[CellGroupData[{
Cell[773181, 20969, 7273, 182, 70, "ProofInfo",
 CellTags->{"Proof|ID:602168712", "Proof|ID:602168712-1"},
 CellID->110870073],
Cell[780457, 21153, 478, 12, 70, "ProofInfoBody",
 CellID->4965199]
}, Closed]],
Cell[780950, 21168, 48, 1, 70, "OpenEnvironment",
 CellID->1920006839],
Cell[CellGroupData[{
Cell[781023, 21173, 388, 9, 70, "EnvironmentHeader",
 CellID->53798018],
Cell[781414, 21184, 2578, 83, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:21782685", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] append", "IR"},
 CellID->21782685],
Cell[783995, 21269, 1204, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1790989103", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isPartIrreflOrder"},
 CellID->1790989103],
Cell[785202, 21304, 65, 1, 70, "EndEnvironmentMarker",
 CellID->462091156]
}, Open  ]],
Cell[785282, 21308, 47, 1, 70, "CloseEnvironment",
 CellID->78191214],
Cell[CellGroupData[{
Cell[785354, 21313, 5900, 148, 70, "ProofInfo",
 CellTags->{"Proof|ID:21782685", "Proof|ID:21782685-1"},
 CellID->202074869],
Cell[791257, 21463, 479, 12, 70, "ProofInfoBody",
 CellID->637991929]
}, Closed]],
Cell[CellGroupData[{
Cell[791773, 21480, 2619, 77, 70, "ProofInfo",
 CellTags->{"Proof|ID:1790989103", "Proof|ID:1790989103-1"},
 CellID->227724],
Cell[794395, 21559, 481, 12, 70, "ProofInfoBody",
 CellID->482486095]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[794937, 21578, 236, 4, 43, "Section",
 CellID->1684828724],
Cell[CellGroupData[{
Cell[795198, 21586, 128, 2, 70, "Subsection",
 CellID->294463156],
Cell[795329, 21590, 363, 9, 70, "Text",
 CellID->1029021260],
Cell[CellGroupData[{
Cell[795717, 21603, 209, 5, 70, "Item",
 CellID->1253702121],
Cell[795929, 21610, 536, 15, 70, "Item",
 CellID->1260543835],
Cell[796468, 21627, 381, 10, 70, "Item",
 CellID->1933641052]
}, Open  ]],
Cell[796864, 21640, 175, 4, 70, "Text",
 CellID->1148296278],
Cell[CellGroupData[{
Cell[797064, 21648, 213, 3, 70, "Subsubsection",
 CellID->302444018],
Cell[797280, 21653, 267, 7, 70, "Text",
 CellID->1904688891],
Cell[CellGroupData[{
Cell[797572, 21664, 999, 32, 70, "Item",
 CellID->622280239],
Cell[798574, 21698, 1147, 34, 70, "Item",
 CellID->1089562638],
Cell[799724, 21734, 1057, 34, 70, "Item",
 CellID->1175767685],
Cell[800784, 21770, 1102, 35, 70, "Item",
 CellID->1244683243]
}, Open  ]],
Cell[801901, 21808, 324, 6, 70, "Text",
 CellID->1415435665],
Cell[802228, 21816, 577, 17, 70, "Item",
 CellID->2032256535],
Cell[802808, 21835, 412, 9, 70, "Text",
 CellID->1158070040]
}, Closed]],
Cell[CellGroupData[{
Cell[803257, 21849, 251, 4, 70, "Subsubsection",
 CellID->1570628282],
Cell[803511, 21855, 1367, 40, 70, "Text",
 CellID->1431522614],
Cell[CellGroupData[{
Cell[804903, 21899, 1074, 33, 70, "Item",
 CellID->46013314],
Cell[805980, 21934, 923, 28, 70, "Item",
 CellID->828733833],
Cell[806906, 21964, 1068, 33, 70, "Item",
 CellID->1947028342]
}, Open  ]],
Cell[807989, 22000, 879, 26, 70, "Text",
 CellID->1001914151],
Cell[808871, 22028, 1810, 61, 70, "Text",
 CellID->1110736610],
Cell[810684, 22091, 484, 12, 70, "Text",
 CellID->2043962507],
Cell[811171, 22105, 896, 28, 70, "Text",
 CellID->1685290468],
Cell[812070, 22135, 862, 21, 70, "Text",
 CellID->31426284],
Cell[CellGroupData[{
Cell[812957, 22160, 1608, 59, 70, "Item",
 CellID->1079707726],
Cell[814568, 22221, 1640, 60, 70, "Item",
 CellID->1281898387]
}, Open  ]],
Cell[816223, 22284, 1143, 31, 70, "Text",
 CellID->2095286757],
Cell[817369, 22317, 300, 6, 70, "Text",
 CellID->695542827],
Cell[817672, 22325, 1412, 43, 70, "Text",
 CellID->1113844059],
Cell[819087, 22370, 3705, 121, 70, "Text",
 CellID->154472253],
Cell[822795, 22493, 1176, 32, 70, "Text",
 CellID->1441754918],
Cell[823974, 22527, 1636, 47, 70, "Text",
 CellID->25744908]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[825659, 22580, 196, 3, 70, "Subsection",
 CellID->2098927943],
Cell[CellGroupData[{
Cell[825880, 22587, 247, 4, 70, "Subsubsection",
 CellID->1987403105],
Cell[826130, 22593, 653, 16, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1466856847", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1466856847],
Cell[826786, 22611, 48, 1, 70, "OpenEnvironment",
 CellID->1327698689],
Cell[CellGroupData[{
Cell[826859, 22616, 478, 10, 70, "EnvironmentHeader",
 CellID->1293548764],
Cell[827340, 22628, 2804, 87, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:623701544", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reduction ring"},
 CellID->623701544],
Cell[830147, 22717, 2665, 81, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1960646232", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "multipliers"},
 CellID->1960646232],
Cell[832815, 22800, 1460, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1496065688", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R0"},
 CellID->1496065688],
Cell[834278, 22847, 1639, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:975892043", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R1"},
 CellID->975892043],
Cell[835920, 22900, 1617, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1150364990", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R2"},
 CellID->1150364990],
Cell[837540, 22952, 2946, 86, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:852396718", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R3"},
 CellID->852396718],
Cell[840489, 23040, 1517, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1907869025", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R4"},
 CellID->1907869025],
Cell[842009, 23087, 2702, 77, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1446921308", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R5"},
 CellID->1446921308],
Cell[844714, 23166, 3054, 87, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:208630606", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6"},
 CellID->208630606],
Cell[847771, 23255, 10540, 264, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1150983847", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R7"},
 CellID->1150983847],
Cell[858314, 23521, 7036, 180, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2106857486", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R8"},
 CellID->2106857486],
Cell[865353, 23703, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1411649426]
}, Open  ]],
Cell[865434, 23707, 49, 1, 70, "CloseEnvironment",
 CellID->2055609263],
Cell[865486, 23710, 47, 1, 70, "OpenEnvironment",
 CellID->465374984],
Cell[CellGroupData[{
Cell[865558, 23715, 410, 10, 70, "EnvironmentHeader",
 CellID->31201615],
Cell[865971, 23727, 928, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1197875862", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1197875862],
Cell[866902, 23750, 3522, 98, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1949920189", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R5 witnesses"},
 CellID->1949920189],
Cell[870427, 23850, 2974, 83, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2027349511", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6 witness"},
 CellID->2027349511],
Cell[873404, 23935, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1778400228]
}, Open  ]],
Cell[873485, 23939, 48, 1, 70, "CloseEnvironment",
 CellID->943653347]
}, Closed]],
Cell[CellGroupData[{
Cell[873570, 23945, 135, 2, 70, "Subsubsection",
 CellID->2042908020],
Cell[873708, 23949, 693, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1717911774", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1717911774],
Cell[874404, 23968, 47, 1, 70, "OpenEnvironment",
 CellID->803522812],
Cell[CellGroupData[{
Cell[874476, 23973, 621, 12, 70, "EnvironmentHeader",
 CellID->1518391126],
Cell[875100, 23987, 771, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:349001278", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->349001278],
Cell[875874, 24008, 1348, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:474542129", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "9"},
 CellID->474542129],
Cell[877225, 24050, 1468, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:360788203", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "21"},
 CellID->360788203],
Cell[878696, 24098, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1400476426]
}, Open  ]],
Cell[878777, 24102, 49, 1, 70, "CloseEnvironment",
 CellID->1277923749],
Cell[CellGroupData[{
Cell[878851, 24107, 4275, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:474542129", "Proof|ID:474542129-1"},
 CellID->1905252519],
Cell[883129, 24223, 481, 12, 70, "ProofInfoBody",
 CellID->1028745265]
}, Closed]],
Cell[CellGroupData[{
Cell[883647, 24240, 4379, 118, 70, "ProofInfo",
 CellTags->{"Proof|ID:360788203", "Proof|ID:360788203-1"},
 CellID->3264289],
Cell[888029, 24360, 480, 12, 70, "ProofInfoBody",
 CellID->115701537]
}, Closed]],
Cell[888524, 24375, 46, 1, 70, "OpenEnvironment",
 CellID->32237124],
Cell[CellGroupData[{
Cell[888595, 24380, 395, 9, 70, "EnvironmentHeader",
 CellID->253497300],
Cell[888993, 24391, 674, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:998446180", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->998446180],
Cell[889670, 24410, 940, 25, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1581977856", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1581977856],
Cell[890613, 24437, 1026, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:623334744", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "8"},
 CellID->623334744],
Cell[891642, 24470, 1591, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1112133758", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "8'"},
 CellID->1112133758],
Cell[893236, 24522, 727, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:810098043", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->810098043],
Cell[893966, 24542, 1312, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1578524611", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "19"},
 CellID->1578524611],
Cell[895281, 24586, 1499, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1738175589", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "19'"},
 CellID->1738175589],
Cell[896783, 24639, 65, 1, 70, "EndEnvironmentMarker",
 CellID->466928142]
}, Open  ]],
Cell[896863, 24643, 48, 1, 70, "CloseEnvironment",
 CellID->946058699],
Cell[CellGroupData[{
Cell[896936, 24648, 4706, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:623334744", "Proof|ID:623334744-1"},
 CellID->1449852563],
Cell[901645, 24774, 480, 12, 70, "ProofInfoBody",
 CellID->521785732]
}, Closed]],
Cell[CellGroupData[{
Cell[902162, 24791, 6203, 157, 70, "ProofInfo",
 CellTags->{"Proof|ID:1112133758", "Proof|ID:1112133758-1"},
 CellID->1792610282],
Cell[908368, 24950, 482, 12, 70, "ProofInfoBody",
 CellID->1552450960]
}, Closed]],
Cell[CellGroupData[{
Cell[908887, 24967, 5891, 151, 70, "ProofInfo",
 CellTags->{"Proof|ID:1578524611", "Proof|ID:1578524611-1"},
 CellID->32974812],
Cell[914781, 25120, 480, 12, 70, "ProofInfoBody",
 CellID->47618736]
}, Closed]],
Cell[CellGroupData[{
Cell[915298, 25137, 6476, 163, 70, "ProofInfo",
 CellTags->{"Proof|ID:1738175589", "Proof|ID:1738175589-1"},
 CellID->629724279],
Cell[921777, 25302, 481, 12, 70, "ProofInfoBody",
 CellID->481178017]
}, Closed]],
Cell[922273, 25317, 46, 1, 70, "OpenEnvironment",
 CellID->68184375],
Cell[CellGroupData[{
Cell[922344, 25322, 497, 11, 70, "EnvironmentHeader",
 CellID->638803249],
Cell[922844, 25335, 1110, 34, 70, "GlobalDeclaration",
 CellTags->{
  "ID:258339355", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->258339355],
Cell[923957, 25371, 832, 22, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1933985410", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1933985410],
Cell[924792, 25395, 818, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:739393974", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "17"},
 CellID->739393974],
Cell[925613, 25420, 1164, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:875264505", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "10"},
 CellID->875264505],
Cell[926780, 25457, 1080, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2012915447", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "11"},
 CellID->2012915447],
Cell[927863, 25492, 1248, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1489336269", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "12"},
 CellID->1489336269],
Cell[929114, 25534, 1483, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1722317392", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "13"},
 CellID->1722317392],
Cell[930600, 25582, 1075, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1089845252", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "15"},
 CellID->1089845252],
Cell[931678, 25617, 65, 1, 70, "EndEnvironmentMarker",
 CellID->640836049]
}, Open  ]],
Cell[931758, 25621, 49, 1, 70, "CloseEnvironment",
 CellID->1240067933],
Cell[CellGroupData[{
Cell[931832, 25626, 4296, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:739393974", "Proof|ID:739393974-1"},
 CellID->1175180046],
Cell[936131, 25744, 481, 12, 70, "ProofInfoBody",
 CellID->1348459641]
}, Closed]],
Cell[CellGroupData[{
Cell[936649, 25761, 6277, 154, 70, "ProofInfo",
 CellTags->{"Proof|ID:875264505", "Proof|ID:875264505-1"},
 CellID->564262226],
Cell[942929, 25917, 479, 12, 70, "ProofInfoBody",
 CellID->24736570]
}, Closed]],
Cell[CellGroupData[{
Cell[943445, 25934, 5335, 137, 70, "ProofInfo",
 CellTags->{"Proof|ID:2012915447", "Proof|ID:2012915447-1"},
 CellID->2044326423],
Cell[948783, 26073, 481, 12, 70, "ProofInfoBody",
 CellID->485002968]
}, Closed]],
Cell[CellGroupData[{
Cell[949301, 26090, 5903, 149, 70, "ProofInfo",
 CellTags->{"Proof|ID:1489336269", "Proof|ID:1489336269-1"},
 CellID->62520461],
Cell[955207, 26241, 481, 12, 70, "ProofInfoBody",
 CellID->257308557]
}, Closed]],
Cell[CellGroupData[{
Cell[955725, 26258, 7045, 171, 70, "ProofInfo",
 CellTags->{"Proof|ID:1722317392", "Proof|ID:1722317392-1"},
 CellID->1641148702],
Cell[962773, 26431, 482, 12, 70, "ProofInfoBody",
 CellID->1013568673]
}, Closed]],
Cell[CellGroupData[{
Cell[963292, 26448, 5323, 137, 70, "ProofInfo",
 CellTags->{"Proof|ID:1089845252", "Proof|ID:1089845252-1"},
 CellID->96759121],
Cell[968618, 26587, 482, 12, 70, "ProofInfoBody",
 CellID->1750417022]
}, Closed]],
Cell[969115, 26602, 48, 1, 70, "OpenEnvironment",
 CellID->1937995867],
Cell[CellGroupData[{
Cell[969188, 26607, 380, 9, 70, "EnvironmentHeader",
 CellID->1477540412],
Cell[969571, 26618, 1054, 31, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2021433946", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2021433946],
Cell[970628, 26651, 737, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:369680194", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->369680194],
Cell[971368, 26673, 2292, 69, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:491925495", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] alternative"},
 CellID->491925495],
Cell[973663, 26744, 1451, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1668317482", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred \[LeftRightArrow]"},
 CellID->1668317482],
Cell[975117, 26790, 65, 1, 70, "EndEnvironmentMarker",
 CellID->354300585]
}, Open  ]],
Cell[975197, 26794, 47, 1, 70, "CloseEnvironment",
 CellID->45215780],
Cell[CellGroupData[{
Cell[975269, 26799, 8952, 210, 70, "ProofInfo",
 CellTags->{"Proof|ID:491925495", "Proof|ID:491925495-1"},
 CellID->99017069],
Cell[984224, 27011, 480, 12, 70, "ProofInfoBody",
 CellID->492768956]
}, Closed]],
Cell[CellGroupData[{
Cell[984741, 27028, 6820, 167, 70, "ProofInfo",
 CellTags->{"Proof|ID:1668317482", "Proof|ID:1668317482-1"},
 CellID->304755132],
Cell[991564, 27197, 481, 12, 70, "ProofInfoBody",
 CellID->422791318]
}, Closed]],
Cell[992060, 27212, 48, 1, 70, "OpenEnvironment",
 CellID->1527684406],
Cell[CellGroupData[{
Cell[992133, 27217, 397, 9, 70, "EnvironmentHeader",
 CellID->1451450076],
Cell[992533, 27228, 887, 27, 70, "GlobalDeclaration",
 CellTags->{
  "ID:169581633", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->169581633],
Cell[993423, 27257, 2362, 68, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:265555623", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "403"},
 CellID->265555623],
Cell[995788, 27327, 65, 1, 70, "EndEnvironmentMarker",
 CellID->428237015]
}, Open  ]],
Cell[995868, 27331, 49, 1, 70, "CloseEnvironment",
 CellID->1421839378],
Cell[CellGroupData[{
Cell[995942, 27336, 9193, 218, 70, "ProofInfo",
 CellTags->{"Proof|ID:265555623", "Proof|ID:265555623-1"},
 CellID->1813792728],
Cell[1005138, 27556, 547, 13, 70, "ProofInfoBody",
 CellID->805435955]
}, Closed]],
Cell[1005700, 27572, 47, 1, 70, "OpenEnvironment",
 CellID->464160879],
Cell[CellGroupData[{
Cell[1005772, 27577, 379, 9, 70, "EnvironmentHeader",
 CellID->718753927],
Cell[1006154, 27588, 1007, 30, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1020054522", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1020054522],
Cell[1007164, 27620, 776, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:270873424", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->270873424],
Cell[1007943, 27643, 998, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1165091979", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "16"},
 CellID->1165091979],
Cell[1008944, 27673, 1144, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1894133408", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symred symmetry"},
 CellID->1894133408],
Cell[1010091, 27706, 65, 1, 70, "EndEnvironmentMarker",
 CellID->677898984]
}, Open  ]],
Cell[1010171, 27710, 48, 1, 70, "CloseEnvironment",
 CellID->492574805],
Cell[CellGroupData[{
Cell[1010244, 27715, 6619, 159, 70, "ProofInfo",
 CellTags->{"Proof|ID:1165091979", "Proof|ID:1165091979-1"},
 CellID->1301195110],
Cell[1016866, 27876, 481, 12, 70, "ProofInfoBody",
 CellID->197446793]
}, Closed]],
Cell[CellGroupData[{
Cell[1017384, 27893, 6703, 161, 70, "ProofInfo",
 CellTags->{"Proof|ID:1894133408", "Proof|ID:1894133408-1"},
 CellID->1016736240],
Cell[1024090, 28056, 482, 12, 70, "ProofInfoBody",
 CellID->1980556603]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1024633, 28075, 136, 2, 70, "Subsection",
 CellID->148207570],
Cell[CellGroupData[{
Cell[1024794, 28081, 132, 2, 70, "Subsubsection",
 CellID->2020955226],
Cell[1024929, 28085, 419, 10, 70, "Text",
 CellID->883452679],
Cell[CellGroupData[{
Cell[1025373, 28099, 423, 11, 70, "Item",
 CellID->604644732],
Cell[1025799, 28112, 511, 17, 70, "Item",
 CellID->123344064],
Cell[1026313, 28131, 702, 23, 70, "Item",
 CellID->1384519374]
}, Open  ]],
Cell[1027030, 28157, 175, 4, 70, "Text",
 CellID->2128238209]
}, Closed]],
Cell[CellGroupData[{
Cell[1027242, 28166, 196, 3, 70, "Subsubsection",
 CellID->1142821098],
Cell[1027441, 28171, 47, 1, 70, "OpenEnvironment",
 CellID->502658829],
Cell[CellGroupData[{
Cell[1027513, 28176, 454, 11, 70, "EnvironmentHeader",
 CellID->327281592],
Cell[1027970, 28189, 653, 16, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1779789522", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1779789522],
Cell[1028626, 28207, 1660, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:979757495", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "algorithmic reduction ring"},
 CellID->979757495],
Cell[1030289, 28260, 748, 22, 70, "EnvironmentText",
 CellID->1965654203],
Cell[1031040, 28284, 5504, 148, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1502179371", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R9"},
 CellID->1502179371],
Cell[1036547, 28434, 4444, 124, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1524689422", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R10"},
 CellID->1524689422],
Cell[1040994, 28560, 2566, 75, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:528478842", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R11"},
 CellID->528478842],
Cell[1043563, 28637, 4242, 109, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:415149625", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R12"},
 CellID->415149625],
Cell[1047808, 28748, 1890, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1276925313", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R13"},
 CellID->1276925313],
Cell[1049701, 28806, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1370612295]
}, Open  ]],
Cell[1049782, 28810, 48, 1, 70, "CloseEnvironment",
 CellID->747506000]
}, Closed]],
Cell[CellGroupData[{
Cell[1049867, 28816, 258, 4, 70, "Subsubsection",
 CellID->530708085],
Cell[1050128, 28822, 697, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1684964128", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1684964128],
Cell[1050828, 28841, 47, 1, 70, "OpenEnvironment",
 CellID->253460494],
Cell[CellGroupData[{
Cell[1050900, 28846, 525, 11, 70, "EnvironmentHeader",
 CellID->114139003],
Cell[1051428, 28859, 2100, 63, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:10673911", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "22"},
 CellID->10673911],
Cell[1053531, 28924, 1562, 48, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:603265816", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "23"},
 CellID->603265816],
Cell[1055096, 28974, 3444, 100, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:39612209", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "24"},
 CellID->39612209],
Cell[1058543, 29076, 65, 1, 70, "EndEnvironmentMarker",
 CellID->570587389]
}, Open  ]],
Cell[1058623, 29080, 48, 1, 70, "CloseEnvironment",
 CellID->430224301],
Cell[CellGroupData[{
Cell[1058696, 29085, 9211, 218, 70, "ProofInfo",
 CellTags->{"Proof|ID:10673911", "Proof|ID:10673911-1"},
 CellID->26780681],
Cell[1067910, 29305, 478, 12, 70, "ProofInfoBody",
 CellID->48417729]
}, Closed]],
Cell[CellGroupData[{
Cell[1068425, 29322, 5283, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:603265816", "Proof|ID:603265816-1"},
 CellID->237799576],
Cell[1073711, 29462, 479, 12, 70, "ProofInfoBody",
 CellID->91085372]
}, Closed]],
Cell[CellGroupData[{
Cell[1074227, 29479, 12808, 292, 70, "ProofInfo",
 CellTags->{"Proof|ID:39612209", "Proof|ID:39612209-1"},
 CellID->269494545],
Cell[1087038, 29773, 478, 12, 70, "ProofInfoBody",
 CellID->19274979]
}, Closed]],
Cell[1087531, 29788, 48, 1, 70, "OpenEnvironment",
 CellID->1851834333],
Cell[CellGroupData[{
Cell[1087604, 29793, 410, 10, 70, "EnvironmentHeader",
 CellID->2034281086],
Cell[1088017, 29805, 680, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1515571182", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1515571182],
Cell[1088700, 29824, 1544, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:528027344", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "18"},
 CellID->528027344],
Cell[1090247, 29877, 2634, 78, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:94925229", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "25"},
 CellID->94925229],
Cell[1092884, 29957, 1801, 55, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2104251234", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "18'"},
 CellID->2104251234],
Cell[1094688, 30014, 65, 1, 70, "EndEnvironmentMarker",
 CellID->897100354]
}, Open  ]],
Cell[1094768, 30018, 47, 1, 70, "CloseEnvironment",
 CellID->87827575],
Cell[CellGroupData[{
Cell[1094840, 30023, 5205, 136, 70, "ProofInfo",
 CellTags->{"Proof|ID:528027344", "Proof|ID:528027344-1"},
 CellID->492864454],
Cell[1100048, 30161, 479, 12, 70, "ProofInfoBody",
 CellID->90487257]
}, Closed]],
Cell[CellGroupData[{
Cell[1100564, 30178, 7618, 188, 70, "ProofInfo",
 CellTags->{"Proof|ID:94925229", "Proof|ID:94925229-1"},
 CellID->568093765],
Cell[1108185, 30368, 478, 12, 70, "ProofInfoBody",
 CellID->28273598]
}, Closed]],
Cell[CellGroupData[{
Cell[1108700, 30385, 7789, 182, 70, "ProofInfo",
 CellTags->{"Proof|ID:2104251234", "Proof|ID:2104251234-1"},
 CellID->611079668],
Cell[1116492, 30569, 481, 12, 70, "ProofInfoBody",
 CellID->875163593]
}, Closed]],
Cell[1116988, 30584, 48, 1, 70, "OpenEnvironment",
 CellID->1971101244],
Cell[CellGroupData[{
Cell[1117061, 30589, 492, 11, 70, "EnvironmentHeader",
 CellID->1398096634],
Cell[1117556, 30602, 878, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:484595803", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isFinDomain \[DoubleRightArrow] R13"},
 CellID->484595803],
Cell[1118437, 30626, 672, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:55361704", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->55361704],
Cell[1119112, 30645, 1712, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:283927726", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "20"},
 CellID->283927726],
Cell[1120827, 30696, 438, 13, 70, "EnvironmentText",
 CellID->536961276],
Cell[1121268, 30711, 1043, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1584214935", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isNoetherian sets"},
 CellID->1584214935],
Cell[1122314, 30743, 1053, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1231517070", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[SquareSubset] isNoetherian tuples"},
 CellID->1231517070],
Cell[1123370, 30776, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1038601912]
}, Open  ]],
Cell[1123451, 30780, 49, 1, 70, "CloseEnvironment",
 CellID->1995689151],
Cell[CellGroupData[{
Cell[1123525, 30785, 2509, 74, 70, "ProofInfo",
 CellTags->{"Proof|ID:484595803", "Proof|ID:484595803-1"},
 CellID->419106276],
Cell[1126037, 30861, 480, 12, 70, "ProofInfoBody",
 CellID->913223413]
}, Closed]],
Cell[CellGroupData[{
Cell[1126554, 30878, 5688, 146, 70, "ProofInfo",
 CellTags->{"Proof|ID:283927726", "Proof|ID:283927726-1"},
 CellID->309385559],
Cell[1132245, 31026, 480, 12, 70, "ProofInfoBody",
 CellID->103041209]
}, Closed]],
Cell[CellGroupData[{
Cell[1132762, 31043, 2893, 83, 70, "ProofInfo",
 CellTags->{"Proof|ID:1584214935", "Proof|ID:1584214935-1"},
 CellID->2080577831],
Cell[1135658, 31128, 481, 12, 70, "ProofInfoBody",
 CellID->977712109]
}, Closed]],
Cell[CellGroupData[{
Cell[1136176, 31145, 2900, 83, 70, "ProofInfo",
 CellTags->{"Proof|ID:1231517070", "Proof|ID:1231517070-1"},
 CellID->120554530],
Cell[1139079, 31230, 481, 12, 70, "ProofInfoBody",
 CellID->439739257]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1139633, 31250, 540, 8, 43, "Section",
 CellID->1796300417],
Cell[CellGroupData[{
Cell[1140198, 31262, 390, 8, 38, "Subsection",
 CellID->2106967378],
Cell[1140591, 31272, 740, 18, 33, "GlobalDeclaration",
 CellTags->{
  "ID:1326788334", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1326788334],
Cell[1141334, 31292, 48, 1, 12, "OpenEnvironment",
 CellID->1436091075],
Cell[CellGroupData[{
Cell[1141407, 31297, 957, 29, 43, "EnvironmentHeader",
 CellID->930024145],
Cell[1142367, 31328, 843, 22, 44, "GlobalDeclaration",
 CellTags->{
  "ID:227700447", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->227700447],
Cell[1143213, 31352, 828, 22, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1128391937", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1128391937],
Cell[1144044, 31376, 1041, 33, 33, "GlobalDeclaration",
 CellTags->{
  "ID:788600821", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->788600821],
Cell[1145088, 31411, 845, 22, 42, "GlobalDeclaration",
 CellTags->{
  "ID:999322294", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->999322294],
Cell[1145936, 31435, 790, 22, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:737948621", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "392"},
 CellID->737948621],
Cell[1146729, 31459, 1317, 37, 40, "FormalTextInputFormula",
 CellTags->{
  "ID:287016951", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "394"},
 CellID->287016951],
Cell[1148049, 31498, 1343, 39, 40, "FormalTextInputFormula",
 CellTags->{
  "ID:1812129266", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "395"},
 CellID->1812129266],
Cell[1149395, 31539, 3200, 90, 119, "FormalTextInputFormula",
 CellTags->{
  "ID:1191851294", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "396"},
 CellID->1191851294],
Cell[1152598, 31631, 2723, 79, 117, "FormalTextInputFormula",
 CellTags->{
  "ID:101106535", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "397"},
 CellID->101106535],
Cell[1155324, 31712, 2532, 72, 61, "FormalTextInputFormula",
 CellTags->{
  "ID:1737800130", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "398"},
 CellID->1737800130],
Cell[1157859, 31786, 2281, 67, 60, "FormalTextInputFormula",
 CellTags->{
  "ID:1568772314", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "399"},
 CellID->1568772314],
Cell[1160143, 31855, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1523433723]
}, Open  ]],
Cell[1160224, 31859, 48, 1, 16, "CloseEnvironment",
 CellID->120405495],
Cell[CellGroupData[{
Cell[1160297, 31864, 4403, 114, 28, "ProofInfo",
 CellTags->{"Proof|ID:737948621", "Proof|ID:737948621-1"},
 CellID->401948380],
Cell[1164703, 31980, 481, 12, 70, "ProofInfoBody",
 CellID->1689050202]
}, Closed]],
Cell[CellGroupData[{
Cell[1165221, 31997, 6075, 150, 26, "ProofInfo",
 CellTags->{"Proof|ID:287016951", "Proof|ID:287016951-1"},
 CellID->766901523],
Cell[1171299, 32149, 479, 12, 70, "ProofInfoBody",
 CellID->87600598]
}, Closed]],
Cell[CellGroupData[{
Cell[1171815, 32166, 7370, 177, 26, "ProofInfo",
 CellTags->{"Proof|ID:1812129266", "Proof|ID:1812129266-1"},
 CellID->967901432],
Cell[1179188, 32345, 480, 12, 70, "ProofInfoBody",
 CellID->82281228]
}, Closed]],
Cell[CellGroupData[{
Cell[1179705, 32362, 16836, 373, 26, "ProofInfo",
 CellTags->{"Proof|ID:1191851294", "Proof|ID:1191851294-1"},
 CellID->155176372],
Cell[1196544, 32737, 523, 13, 70, "ProofInfoBody",
 CellID->206911940]
}, Closed]],
Cell[CellGroupData[{
Cell[1197104, 32755, 14766, 328, 26, "ProofInfo",
 CellTags->{"Proof|ID:101106535", "Proof|ID:101106535-1"},
 CellID->1074757864],
Cell[1211873, 33085, 523, 13, 70, "ProofInfoBody",
 CellID->1512156001]
}, Closed]],
Cell[CellGroupData[{
Cell[1212433, 33103, 11792, 273, 26, "ProofInfo",
 CellTags->{"Proof|ID:1737800130", "Proof|ID:1737800130-1"},
 CellID->452671936],
Cell[1224228, 33378, 482, 12, 70, "ProofInfoBody",
 CellID->1450652992]
}, Closed]],
Cell[CellGroupData[{
Cell[1224747, 33395, 10791, 253, 26, "ProofInfo",
 CellTags->{"Proof|ID:1568772314", "Proof|ID:1568772314-1"},
 CellID->1787487969],
Cell[1235541, 33650, 481, 12, 70, "ProofInfoBody",
 CellID->702767096]
}, Closed]],
Cell[1236037, 33665, 251, 7, 24, "Text",
 CellID->794210953],
Cell[1236291, 33674, 48, 1, 12, "OpenEnvironment",
 CellID->1645079720],
Cell[CellGroupData[{
Cell[1236364, 33679, 987, 27, 63, "EnvironmentHeader",
 CellID->334577641],
Cell[1237354, 33708, 870, 22, 44, "GlobalDeclaration",
 CellTags->{
  "ID:1597110548", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1597110548],
Cell[1238227, 33732, 840, 21, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1509853174", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1509853174],
Cell[1239070, 33755, 1700, 43, 56, "GlobalDeclaration",
 CellTags->{
  "ID:1554200522", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1554200522],
Cell[1240773, 33800, 1344, 39, 40, "FormalTextInputFormula",
 CellTags->{
  "ID:1963951655", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "400"},
 CellID->1963951655],
Cell[1242120, 33841, 1002, 30, 40, "GlobalDeclaration",
 CellTags->{
  "ID:85295514", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->85295514],
Cell[1243125, 33873, 2157, 63, 84, "FormalTextInputFormula",
 CellTags->{
  "ID:751169803", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "401"},
 CellID->751169803],
Cell[1245285, 33938, 2025, 61, 60, "FormalTextInputFormula",
 CellTags->{
  "ID:46875747", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "402'"},
 CellID->46875747],
Cell[1247313, 34001, 2201, 64, 60, "FormalTextInputFormula",
 CellTags->{
  "ID:656927872", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "402"},
 CellID->656927872],
Cell[1249517, 34067, 65, 1, 26, "EndEnvironmentMarker",
 CellID->460382234]
}, Open  ]],
Cell[1249597, 34071, 48, 1, 16, "CloseEnvironment",
 CellID->122940993],
Cell[CellGroupData[{
Cell[1249670, 34076, 10968, 259, 28, "ProofInfo",
 CellTags->{"Proof|ID:1963951655", "Proof|ID:1963951655-1"},
 CellID->208873331],
Cell[1260641, 34337, 480, 12, 70, "ProofInfoBody",
 CellID->74193188]
}, Closed]],
Cell[CellGroupData[{
Cell[1261158, 34354, 21497, 480, 26, "ProofInfo",
 CellTags->{"Proof|ID:751169803", "Proof|ID:751169803-1"},
 CellID->553193299],
Cell[1282658, 34836, 480, 12, 70, "ProofInfoBody",
 CellID->199842648]
}, Closed]],
Cell[CellGroupData[{
Cell[1283175, 34853, 14469, 331, 26, "ProofInfo",
 CellTags->{"Proof|ID:46875747", "Proof|ID:46875747-1"},
 CellID->588348976],
Cell[1297647, 35186, 477, 12, 70, "ProofInfoBody",
 CellID->1800386]
}, Closed]],
Cell[CellGroupData[{
Cell[1298161, 35203, 14964, 338, 26, "ProofInfo",
 CellTags->{"Proof|ID:656927872", "Proof|ID:656927872-1"},
 CellID->574764629],
Cell[1313128, 35543, 480, 12, 70, "ProofInfoBody",
 CellID->992096306]
}, Closed]],
Cell[1313623, 35558, 311, 8, 24, "Text",
 CellID->1802708008],
Cell[1313937, 35568, 48, 1, 12, "OpenEnvironment",
 CellID->1946059673],
Cell[CellGroupData[{
Cell[1314010, 35573, 393, 9, 43, "EnvironmentHeader",
 CellID->1305345464],
Cell[1314406, 35584, 794, 21, 44, "GlobalDeclaration",
 CellTags->{
  "ID:502021608", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->502021608],
Cell[1315203, 35607, 924, 23, 41, "GlobalDeclaration",
 CellTags->{
  "ID:227587436", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->227587436],
Cell[1316130, 35632, 1206, 31, 34, "GlobalDeclaration",
 CellTags->{
  "ID:937292216", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->937292216],
Cell[1317339, 35665, 1553, 45, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:1763687895", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "410"},
 CellID->1763687895],
Cell[1318895, 35712, 1164, 32, 56, "GlobalDeclaration",
 CellTags->{
  "ID:397073436", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->397073436],
Cell[1320062, 35746, 1593, 45, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:791651685", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "411"},
 CellID->791651685],
Cell[1321658, 35793, 1130, 38, 44, "GlobalDeclaration",
 CellTags->{
  "ID:6804200", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->6804200],
Cell[1322791, 35833, 1069, 31, 58, "FormalTextInputFormula",
 CellTags->{
  "ID:1294525434", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "414"},
 CellID->1294525434],
Cell[1323863, 35866, 1005, 30, 40, "GlobalDeclaration",
 CellTags->{
  "ID:372638659", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->372638659],
Cell[1324871, 35898, 1824, 54, 71, "FormalTextInputFormula",
 CellTags->{
  "ID:1962716191", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "412"},
 CellID->1962716191],
Cell[1326698, 35954, 2316, 71, 40, "FormalTextInputFormula",
 CellTags->{
  "ID:1575262565", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "413"},
 CellID->1575262565],
Cell[1329017, 36027, 65, 1, 26, "EndEnvironmentMarker",
 CellID->982944345]
}, Open  ]],
Cell[1329097, 36031, 48, 1, 16, "CloseEnvironment",
 CellID->222080009],
Cell[CellGroupData[{
Cell[1329170, 36036, 11350, 259, 28, "ProofInfo",
 CellTags->{"Proof|ID:1763687895", "Proof|ID:1763687895-1"},
 CellID->701556455],
Cell[1340523, 36297, 482, 12, 70, "ProofInfoBody",
 CellID->1674393023]
}, Closed]],
Cell[CellGroupData[{
Cell[1341042, 36314, 15630, 356, 26, "ProofInfo",
 CellTags->{"Proof|ID:791651685", "Proof|ID:791651685-1"},
 CellID->2105672847],
Cell[1356675, 36672, 481, 12, 70, "ProofInfoBody",
 CellID->2052754145]
}, Closed]],
Cell[CellGroupData[{
Cell[1357193, 36689, 14609, 337, 26, "ProofInfo",
 CellTags->{"Proof|ID:1294525434", "Proof|ID:1294525434-1"},
 CellID->727439100],
Cell[1371805, 37028, 481, 12, 70, "ProofInfoBody",
 CellID->295763820]
}, Closed]],
Cell[CellGroupData[{
Cell[1372323, 37045, 25916, 576, 26, "ProofInfo",
 CellTags->{"Proof|ID:1962716191", "Proof|ID:1962716191-1"},
 CellID->1885017557],
Cell[1398242, 37623, 481, 12, 70, "ProofInfoBody",
 CellID->137392427]
}, Closed]],
Cell[CellGroupData[{
Cell[1398760, 37640, 21984, 489, 26, "ProofInfo",
 CellTags->{"Proof|ID:1575262565", "Proof|ID:1575262565-1"},
 CellID->1705894400],
Cell[1420747, 38131, 482, 12, 70, "ProofInfoBody",
 CellID->2053331622]
}, Closed]],
Cell[CellGroupData[{
Cell[1421266, 38148, 352, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1875389319],
Cell[1421621, 38159, 12302, 369, 762, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1226616153],
Cell[1433926, 38530, 295, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->2115837209]
}, Closed]],
Cell[CellGroupData[{
Cell[1434258, 38541, 325, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->837540631],
Cell[1434586, 38552, 10464, 317, 865, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1753611981],
Cell[1445053, 38871, 292, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->408639765]
}, Closed]],
Cell[1445360, 38880, 48, 1, 12, "OpenEnvironment",
 CellID->1386249146],
Cell[CellGroupData[{
Cell[1445433, 38885, 580, 12, 43, "EnvironmentHeader",
 CellID->1873815999],
Cell[1446016, 38899, 1155, 35, 40, "GlobalDeclaration",
 CellTags->{
  "ID:1547867165", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1547867165],
Cell[1447174, 38936, 1005, 24, 41, "GlobalDeclaration",
 CellTags->{
  "ID:2009958353", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2009958353],
Cell[1448182, 38962, 1405, 43, 43, "GlobalDeclaration",
 CellTags->{
  "ID:10165835", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->10165835],
Cell[1449590, 39007, 1564, 43, 53, "FormalTextInputFormula",
 CellTags->{
  "ID:1879224373", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6Witness sufficient"},
 CellID->1879224373],
Cell[1451157, 39052, 1847, 55, 112, "FormalTextInputFormula",
 CellTags->{
  "ID:381733398", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "R6Witness symmetry"},
 CellID->381733398],
Cell[1453007, 39109, 66, 1, 26, "EndEnvironmentMarker",
 CellID->2045819678]
}, Open  ]],
Cell[1453088, 39113, 48, 1, 16, "CloseEnvironment",
 CellID->596803790],
Cell[CellGroupData[{
Cell[1453161, 39118, 10441, 237, 28, "ProofInfo",
 CellTags->{"Proof|ID:1879224373", "Proof|ID:1879224373-1"},
 CellID->453457936],
Cell[1463605, 39357, 480, 12, 70, "ProofInfoBody",
 CellID->63675348]
}, Closed]],
Cell[CellGroupData[{
Cell[1464122, 39374, 11696, 260, 26, "ProofInfo",
 CellTags->{"Proof|ID:381733398", "Proof|ID:381733398-1"},
 CellID->180731690],
Cell[1475821, 39636, 480, 12, 70, "ProofInfoBody",
 CellID->482825486]
}, Closed]],
Cell[CellGroupData[{
Cell[1476338, 39653, 382, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->592729503],
Cell[1476723, 39664, 304, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->665964876],
Cell[1477030, 39671, 2469, 61, 73, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->236652974],
Cell[1479502, 39734, 12002, 365, 504, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1332683746],
Cell[1491507, 40101, 318, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->541568807]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1491874, 40113, 401, 9, 30, "Subsection",
 CellID->1827516953],
Cell[1492278, 40124, 692, 17, 33, "GlobalDeclaration",
 CellTags->{
  "ID:404204643", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->404204643],
Cell[CellGroupData[{
Cell[1492995, 40145, 322, 5, 28, "Subsubsection",
 CellID->611594790],
Cell[1493320, 40152, 47, 1, 70, "OpenEnvironment",
 CellID->908708940],
Cell[CellGroupData[{
Cell[1493392, 40157, 458, 12, 70, "EnvironmentHeader",
 CellID->1831554730],
Cell[1493853, 40171, 1019, 26, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1399661856", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1399661856],
Cell[1494875, 40199, 1816, 59, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1116556350", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Tilde] multiplier symmetric"},
 CellID->1116556350],
Cell[1496694, 40260, 1335, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1121162524", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "mult \[Tilde]"},
 CellID->1121162524],
Cell[1498032, 40304, 1411, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1098905803", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isReducible \[Tilde]"},
 CellID->1098905803],
Cell[1499446, 40348, 1862, 53, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:88395859", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areIrrelative \[Tilde]"},
 CellID->88395859],
Cell[1501311, 40403, 2056, 55, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1322820809", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areCPMultipliers \[Tilde]"},
 CellID->1322820809],
Cell[1503370, 40460, 753, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:931721333", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->931721333],
Cell[1504126, 40482, 1741, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2090837112", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] multiplier \[Tilde]"},
 CellID->2090837112],
Cell[1505870, 40540, 1681, 54, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:780050836", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[Tilde]"},
 CellID->780050836],
Cell[1507554, 40596, 761, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:128058836", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->128058836],
Cell[1508318, 40617, 1379, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1516437288", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] reductor \[Tilde]"},
 CellID->1516437288],
Cell[1509700, 40661, 1334, 41, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1864007722", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] reductor \[Tilde]"},
 CellID->1864007722],
Cell[1511037, 40704, 1850, 60, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:642398934", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility below \[Tilde]"},
 CellID->642398934],
Cell[1512890, 40766, 65, 1, 70, "EndEnvironmentMarker",
 CellID->217367629]
}, Open  ]],
Cell[1512970, 40770, 47, 1, 70, "CloseEnvironment",
 CellID->72738875],
Cell[CellGroupData[{
Cell[1513042, 40775, 6829, 171, 70, "ProofInfo",
 CellTags->{"Proof|ID:1116556350", "Proof|ID:1116556350-1"},
 CellID->1446312685],
Cell[1519874, 40948, 481, 12, 70, "ProofInfoBody",
 CellID->239017603]
}, Closed]],
Cell[CellGroupData[{
Cell[1520392, 40965, 6459, 161, 70, "ProofInfo",
 CellTags->{"Proof|ID:1121162524", "Proof|ID:1121162524-1"},
 CellID->1284068679],
Cell[1526854, 41128, 482, 12, 70, "ProofInfoBody",
 CellID->2106471830]
}, Closed]],
Cell[CellGroupData[{
Cell[1527373, 41145, 7085, 173, 70, "ProofInfo",
 CellTags->{"Proof|ID:1098905803", "Proof|ID:1098905803-1"},
 CellID->1798497905],
Cell[1534461, 41320, 482, 12, 70, "ProofInfoBody",
 CellID->1893221538]
}, Closed]],
Cell[CellGroupData[{
Cell[1534980, 41337, 9976, 232, 70, "ProofInfo",
 CellTags->{"Proof|ID:88395859", "Proof|ID:88395859-1"},
 CellID->514698520],
Cell[1544959, 41571, 479, 12, 70, "ProofInfoBody",
 CellID->697959023]
}, Closed]],
Cell[CellGroupData[{
Cell[1545475, 41588, 9856, 229, 70, "ProofInfo",
 CellTags->{"Proof|ID:1322820809", "Proof|ID:1322820809-1"},
 CellID->330288797],
Cell[1555334, 41819, 526, 13, 70, "ProofInfoBody",
 CellID->1636900124]
}, Closed]],
Cell[CellGroupData[{
Cell[1555897, 41837, 8126, 201, 70, "ProofInfo",
 CellTags->{"Proof|ID:2090837112", "Proof|ID:2090837112-1"},
 CellID->1967137926],
Cell[1564026, 42040, 482, 12, 70, "ProofInfoBody",
 CellID->1474790292]
}, Closed]],
Cell[CellGroupData[{
Cell[1564545, 42057, 7124, 178, 70, "ProofInfo",
 CellTags->{"Proof|ID:780050836", "Proof|ID:780050836-1"},
 CellID->122789986],
Cell[1571672, 42237, 480, 12, 70, "ProofInfoBody",
 CellID->155714628]
}, Closed]],
Cell[CellGroupData[{
Cell[1572189, 42254, 7002, 176, 70, "ProofInfo",
 CellTags->{"Proof|ID:1516437288", "Proof|ID:1516437288-1"},
 CellID->23267736],
Cell[1579194, 42432, 481, 12, 70, "ProofInfoBody",
 CellID->241728310]
}, Closed]],
Cell[CellGroupData[{
Cell[1579712, 42449, 7060, 177, 70, "ProofInfo",
 CellTags->{"Proof|ID:1864007722", "Proof|ID:1864007722-1"},
 CellID->1742409570],
Cell[1586775, 42628, 482, 12, 70, "ProofInfoBody",
 CellID->1961249060]
}, Closed]],
Cell[CellGroupData[{
Cell[1587294, 42645, 7871, 190, 70, "ProofInfo",
 CellTags->{"Proof|ID:642398934", "Proof|ID:642398934-1"},
 CellID->1849048841],
Cell[1595168, 42837, 480, 12, 70, "ProofInfoBody",
 CellID->503024193]
}, Closed]],
Cell[1595663, 42852, 47, 1, 70, "OpenEnvironment",
 CellID->242245566],
Cell[CellGroupData[{
Cell[1595735, 42857, 403, 11, 70, "EnvironmentHeader",
 CellID->1737749019],
Cell[1596141, 42870, 765, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1200903488", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1200903488],
Cell[1596909, 42892, 810, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:353808388", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[Tilde]"},
 CellID->353808388],
Cell[1597722, 42915, 1079, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:274091018", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of \[Tilde]"},
 CellID->274091018],
Cell[1598804, 42951, 1391, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1760571189", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[Tilde]"},
 CellID->1760571189],
Cell[1600198, 43000, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1666723148]
}, Open  ]],
Cell[1600279, 43004, 49, 1, 70, "CloseEnvironment",
 CellID->1069542713],
Cell[CellGroupData[{
Cell[1600353, 43009, 2828, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:353808388", "Proof|ID:353808388-1"},
 CellID->844255967],
Cell[1603184, 43093, 481, 12, 70, "ProofInfoBody",
 CellID->1096022174]
}, Closed]],
Cell[CellGroupData[{
Cell[1603702, 43110, 3728, 102, 70, "ProofInfo",
 CellTags->{"Proof|ID:274091018", "Proof|ID:274091018-1"},
 CellID->772612304],
Cell[1607433, 43214, 481, 12, 70, "ProofInfoBody",
 CellID->1935039816]
}, Closed]],
Cell[CellGroupData[{
Cell[1607951, 43231, 4664, 123, 70, "ProofInfo",
 CellTags->{"Proof|ID:1760571189", "Proof|ID:1760571189-1"},
 CellID->1528186019],
Cell[1612618, 43356, 481, 12, 70, "ProofInfoBody",
 CellID->690157133]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1613148, 43374, 262, 6, 22, "Subsubsection",
 CellID->1144726974],
Cell[1613413, 43382, 48, 1, 70, "OpenEnvironment",
 CellID->1390555827],
Cell[CellGroupData[{
Cell[1613486, 43387, 392, 9, 70, "EnvironmentHeader",
 CellID->934535187],
Cell[1613881, 43398, 973, 25, 70, "GlobalDeclaration",
 CellTags->{
  "ID:871669950", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->871669950],
Cell[1614857, 43425, 1720, 52, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:163646289", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "correlative-irrelative 1"},
 CellID->163646289],
Cell[1616580, 43479, 65, 1, 70, "EndEnvironmentMarker",
 CellID->683761533]
}, Open  ]],
Cell[1616660, 43483, 49, 1, 70, "CloseEnvironment",
 CellID->1397889212],
Cell[CellGroupData[{
Cell[1616734, 43488, 8562, 202, 70, "ProofInfo",
 CellTags->{"Proof|ID:163646289", "Proof|ID:163646289-1"},
 CellID->1996269430],
Cell[1625299, 43692, 480, 12, 70, "ProofInfoBody",
 CellID->960850119]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1625828, 43710, 220, 5, 22, "Subsubsection",
 CellID->679333461],
Cell[1626051, 43717, 48, 1, 70, "OpenEnvironment",
 CellID->1992553206],
Cell[CellGroupData[{
Cell[1626124, 43722, 425, 10, 70, "EnvironmentHeader",
 CellID->280665347],
Cell[1626552, 43734, 813, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1990325407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1990325407],
Cell[1627368, 43757, 765, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:615171230", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->615171230],
Cell[1628136, 43779, 1506, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1548501955", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "47"},
 CellID->1548501955],
Cell[1629645, 43828, 1423, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:170197122", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "48"},
 CellID->170197122],
Cell[1631071, 43876, 1947, 65, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:772910972", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "49"},
 CellID->772910972],
Cell[1633021, 43943, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1385006624]
}, Open  ]],
Cell[1633102, 43947, 49, 1, 70, "CloseEnvironment",
 CellID->1123829565],
Cell[CellGroupData[{
Cell[1633176, 43952, 5297, 139, 70, "ProofInfo",
 CellTags->{"Proof|ID:1548501955", "Proof|ID:1548501955-1"},
 CellID->657095506],
Cell[1638476, 44093, 480, 12, 70, "ProofInfoBody",
 CellID->40777374]
}, Closed]],
Cell[CellGroupData[{
Cell[1638993, 44110, 5206, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:170197122", "Proof|ID:170197122-1"},
 CellID->175765073],
Cell[1644202, 44244, 479, 12, 70, "ProofInfoBody",
 CellID->48619257]
}, Closed]],
Cell[CellGroupData[{
Cell[1644718, 44261, 6121, 152, 70, "ProofInfo",
 CellTags->{"Proof|ID:772910972", "Proof|ID:772910972-1"},
 CellID->25309177],
Cell[1650842, 44415, 480, 12, 70, "ProofInfoBody",
 CellID->541611342]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1651371, 44433, 506, 14, 22, "Subsubsection",
 CellID->1798149129],
Cell[1651880, 44449, 47, 1, 12, "OpenEnvironment",
 CellID->922102237],
Cell[CellGroupData[{
Cell[1651952, 44454, 560, 16, 43, "EnvironmentHeader",
 CellID->545035824],
Cell[1652515, 44472, 1026, 27, 42, "GlobalDeclaration",
 CellTags->{
  "ID:1034579573", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1034579573],
Cell[1653544, 44501, 3958, 114, 119, "FormalTextInputFormula",
 CellTags->{
  "ID:226093298", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "30.\[DoubleRightArrow]"},
 CellID->226093298],
Cell[1657505, 44617, 3927, 117, 156, "FormalTextInputFormula",
 CellTags->{
  "ID:1773205884", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "31.\[DoubleRightArrow]"},
 CellID->1773205884],
Cell[1661435, 44736, 3022, 91, 119, "FormalTextInputFormula",
 CellTags->{
  "ID:1783130815", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "30.\[DoubleLeftArrow]"},
 CellID->1783130815],
Cell[1664460, 44829, 2860, 88, 117, "FormalTextInputFormula",
 CellTags->{
  "ID:1493153487", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "31.\[DoubleLeftArrow]"},
 CellID->1493153487],
Cell[1667323, 44919, 1979, 67, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:1798779961", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "39.\[DoubleRightArrow]"},
 CellID->1798779961],
Cell[1669305, 44988, 65, 1, 26, "EndEnvironmentMarker",
 CellID->481199956]
}, Open  ]],
Cell[1669385, 44992, 48, 1, 16, "CloseEnvironment",
 CellID->336597908],
Cell[CellGroupData[{
Cell[1669458, 44997, 10820, 254, 28, "ProofInfo",
 CellTags->{"Proof|ID:226093298", "Proof|ID:226093298-1"},
 CellID->2111336883],
Cell[1680281, 45253, 523, 13, 70, "ProofInfoBody",
 CellID->1011674363]
}, Closed]],
Cell[CellGroupData[{
Cell[1680841, 45271, 11684, 273, 26, "ProofInfo",
 CellTags->{"Proof|ID:1773205884", "Proof|ID:1773205884-1"},
 CellID->468420595],
Cell[1692528, 45546, 481, 12, 70, "ProofInfoBody",
 CellID->303982263]
}, Closed]],
Cell[CellGroupData[{
Cell[1693046, 45563, 9162, 221, 26, "ProofInfo",
 CellTags->{"Proof|ID:1783130815", "Proof|ID:1783130815-1"},
 CellID->1694100803],
Cell[1702211, 45786, 571, 14, 70, "ProofInfoBody",
 CellID->1573422288]
}, Closed]],
Cell[CellGroupData[{
Cell[1702819, 45805, 9166, 221, 26, "ProofInfo",
 CellTags->{"Proof|ID:1493153487", "Proof|ID:1493153487-1"},
 CellID->1970791442],
Cell[1711988, 46028, 480, 12, 70, "ProofInfoBody",
 CellID->22791630]
}, Closed]],
Cell[CellGroupData[{
Cell[1712505, 46045, 6543, 161, 26, "ProofInfo",
 CellTags->{"Proof|ID:1798779961", "Proof|ID:1798779961-1"},
 CellID->1268768510],
Cell[1719051, 46208, 481, 12, 70, "ProofInfoBody",
 CellID->699204208]
}, Closed]],
Cell[CellGroupData[{
Cell[1719569, 46225, 331, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->279050304],
Cell[1719903, 46236, 11149, 359, 1643, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1466920198],
Cell[1731055, 46597, 266, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->539158522]
}, Closed]],
Cell[1731336, 46605, 47, 1, 12, "OpenEnvironment",
 CellID->661399684],
Cell[CellGroupData[{
Cell[1731408, 46610, 548, 15, 43, "EnvironmentHeader",
 CellID->776782599],
Cell[1731959, 46627, 1024, 25, 41, "GlobalDeclaration",
 CellTags->{
  "ID:712613448", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->712613448],
Cell[1732986, 46654, 3002, 91, 119, "FormalTextInputFormula",
 CellTags->{
  "ID:1914773542", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "37.\[DoubleRightArrow]"},
 CellID->1914773542],
Cell[1735991, 46747, 3026, 92, 119, "FormalTextInputFormula",
 CellTags->{
  "ID:2048103084", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "37.\[DoubleLeftArrow]"},
 CellID->2048103084],
Cell[1739020, 46841, 64, 1, 26, "EndEnvironmentMarker",
 CellID->92140323]
}, Open  ]],
Cell[1739099, 46845, 49, 1, 16, "CloseEnvironment",
 CellID->1764183355],
Cell[CellGroupData[{
Cell[1739173, 46850, 8941, 213, 28, "ProofInfo",
 CellTags->{"Proof|ID:1914773542", "Proof|ID:1914773542-1"},
 CellID->450197198],
Cell[1748117, 47065, 482, 12, 70, "ProofInfoBody",
 CellID->1657761586]
}, Closed]],
Cell[CellGroupData[{
Cell[1748636, 47082, 8950, 215, 26, "ProofInfo",
 CellTags->{"Proof|ID:2048103084", "Proof|ID:2048103084-1"},
 CellID->1611988294],
Cell[1757589, 47299, 482, 12, 70, "ProofInfoBody",
 CellID->1040910915]
}, Closed]],
Cell[1758086, 47314, 48, 1, 12, "OpenEnvironment",
 CellID->1856959695],
Cell[CellGroupData[{
Cell[1758159, 47319, 510, 12, 43, "EnvironmentHeader",
 CellID->382418526],
Cell[1758672, 47333, 1003, 26, 42, "GlobalDeclaration",
 CellTags->{
  "ID:254692456", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->254692456],
Cell[1759678, 47361, 3946, 122, 158, "FormalTextInputFormula",
 CellTags->{
  "ID:336730845", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "32.\[DoubleRightArrow]"},
 CellID->336730845],
Cell[1763627, 47485, 3973, 123, 158, "FormalTextInputFormula",
 CellTags->{
  "ID:1715399148", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "32.\[DoubleLeftArrow]"},
 CellID->1715399148],
Cell[1767603, 47610, 66, 1, 26, "EndEnvironmentMarker",
 CellID->2087871386]
}, Open  ]],
Cell[1767684, 47614, 48, 1, 16, "CloseEnvironment",
 CellID->418896992],
Cell[CellGroupData[{
Cell[1767757, 47619, 12027, 278, 28, "ProofInfo",
 CellTags->{"Proof|ID:336730845", "Proof|ID:336730845-1"},
 CellID->650311272],
Cell[1779787, 47899, 525, 13, 70, "ProofInfoBody",
 CellID->1958611019]
}, Closed]],
Cell[CellGroupData[{
Cell[1780349, 47917, 12054, 281, 26, "ProofInfo",
 CellTags->{"Proof|ID:1715399148", "Proof|ID:1715399148-1"},
 CellID->2088013031],
Cell[1792406, 48200, 547, 13, 70, "ProofInfoBody",
 CellID->2115273351]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1793002, 48219, 339, 10, 22, "Subsubsection",
 CellID->829207957],
Cell[1793344, 48231, 47, 1, 70, "OpenEnvironment",
 CellID->525227380],
Cell[CellGroupData[{
Cell[1793416, 48236, 491, 15, 70, "EnvironmentHeader",
 CellID->838178864],
Cell[1793910, 48253, 885, 25, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1253203360", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1253203360],
Cell[1794798, 48280, 860, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1514263990", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[RightArrow]\[RawStar]"},
 CellID->1514263990],
Cell[1795661, 48305, 1506, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:411529944", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[RightArrow]\[RawStar]"},
 CellID->411529944],
Cell[1797170, 48357, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1053827266]
}, Open  ]],
Cell[1797251, 48361, 49, 1, 70, "CloseEnvironment",
 CellID->1233556406],
Cell[CellGroupData[{
Cell[1797325, 48366, 3526, 97, 70, "ProofInfo",
 CellTags->{"Proof|ID:1514263990", "Proof|ID:1514263990-1"},
 CellID->1590090525],
Cell[1800854, 48465, 482, 12, 70, "ProofInfoBody",
 CellID->1795169377]
}, Closed]],
Cell[CellGroupData[{
Cell[1801373, 48482, 6006, 149, 70, "ProofInfo",
 CellTags->{"Proof|ID:411529944", "Proof|ID:411529944-1"},
 CellID->144989886],
Cell[1807382, 48633, 480, 12, 70, "ProofInfoBody",
 CellID->914771064]
}, Closed]],
Cell[1807877, 48648, 47, 1, 70, "OpenEnvironment",
 CellID->979821556],
Cell[CellGroupData[{
Cell[1807949, 48653, 338, 9, 70, "EnvironmentHeader",
 CellID->593977615],
Cell[1808290, 48664, 885, 25, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1862973197", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1862973197],
Cell[1809178, 48691, 1202, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:237268641", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
   "\[RightArrow] \[SubsetEqual] \[RightArrow]\[RawStar]"},
 CellID->237268641],
Cell[1810383, 48731, 1492, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1746879589", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "38"},
 CellID->1746879589],
Cell[1811878, 48784, 65, 1, 70, "EndEnvironmentMarker",
 CellID->982648542]
}, Open  ]],
Cell[1811958, 48788, 49, 1, 70, "CloseEnvironment",
 CellID->2144214436],
Cell[CellGroupData[{
Cell[1812032, 48793, 4641, 122, 70, "ProofInfo",
 CellTags->{"Proof|ID:237268641", "Proof|ID:237268641-1"},
 CellID->1915348311],
Cell[1816676, 48917, 481, 12, 70, "ProofInfoBody",
 CellID->1480957704]
}, Closed]],
Cell[CellGroupData[{
Cell[1817194, 48934, 5800, 147, 70, "ProofInfo",
 CellTags->{"Proof|ID:1746879589", "Proof|ID:1746879589-1"},
 CellID->1067079851],
Cell[1822997, 49083, 481, 12, 70, "ProofInfoBody",
 CellID->657142897]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1823527, 49101, 498, 11, 22, "Subsubsection",
 CellID->979571816],
Cell[1824028, 49114, 47, 1, 70, "OpenEnvironment",
 CellID->805924195],
Cell[CellGroupData[{
Cell[1824100, 49119, 499, 15, 70, "EnvironmentHeader",
 CellID->783637779],
Cell[1824602, 49136, 821, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1956094048", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1956094048],
Cell[1825426, 49161, 915, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:560057636", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of connectibility"},
 CellID->560057636],
Cell[1826344, 49187, 1325, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:354829984", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of connectibility"},
 CellID->354829984],
Cell[1827672, 49227, 1587, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:350848859", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of connectibility"},
 CellID->350848859],
Cell[1829262, 49280, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1886172493]
}, Open  ]],
Cell[1829343, 49284, 47, 1, 70, "CloseEnvironment",
 CellID->69823970],
Cell[CellGroupData[{
Cell[1829415, 49289, 3265, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:560057636", "Proof|ID:560057636-1"},
 CellID->1331060490],
Cell[1832683, 49381, 523, 13, 70, "ProofInfoBody",
 CellID->1930057720]
}, Closed]],
Cell[CellGroupData[{
Cell[1833243, 49399, 4506, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:354829984", "Proof|ID:354829984-1"},
 CellID->667920451],
Cell[1837752, 49517, 480, 12, 70, "ProofInfoBody",
 CellID->263591681]
}, Closed]],
Cell[CellGroupData[{
Cell[1838269, 49534, 5795, 143, 70, "ProofInfo",
 CellTags->{"Proof|ID:350848859", "Proof|ID:350848859-1"},
 CellID->1457001757],
Cell[1844067, 49679, 546, 13, 70, "ProofInfoBody",
 CellID->315146723]
}, Closed]],
Cell[1844628, 49695, 48, 1, 70, "OpenEnvironment",
 CellID->1576185928],
Cell[CellGroupData[{
Cell[1844701, 49700, 709, 16, 70, "EnvironmentHeader",
 CellID->641705771],
Cell[1845413, 49718, 1091, 27, 70, "GlobalDeclaration",
 CellTags->{
  "ID:705004664", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->705004664],
Cell[1846507, 49747, 1392, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:411199728", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of connectibility below"},
 CellID->411199728],
Cell[1847902, 49793, 1730, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1502850808", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of connectibility below"},
 CellID->1502850808],
Cell[1849635, 49851, 2301, 79, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1331049102", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of connectibility below"},
 CellID->1331049102],
Cell[1851939, 49932, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1022905620]
}, Open  ]],
Cell[1852020, 49936, 49, 1, 70, "CloseEnvironment",
 CellID->1032464773],
Cell[CellGroupData[{
Cell[1852094, 49941, 4781, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:411199728", "Proof|ID:411199728-1"},
 CellID->24086679],
Cell[1856878, 50067, 575, 14, 70, "ProofInfoBody",
 CellID->1856563014]
}, Closed]],
Cell[CellGroupData[{
Cell[1857490, 50086, 5722, 144, 70, "ProofInfo",
 CellTags->{"Proof|ID:1502850808", "Proof|ID:1502850808-1"},
 CellID->1390382820],
Cell[1863215, 50232, 547, 13, 70, "ProofInfoBody",
 CellID->726470046]
}, Closed]],
Cell[CellGroupData[{
Cell[1863799, 50250, 7301, 177, 70, "ProofInfo",
 CellTags->{"Proof|ID:1331049102", "Proof|ID:1331049102-1"},
 CellID->516908597],
Cell[1871103, 50429, 549, 13, 70, "ProofInfoBody",
 CellID->966474202]
}, Closed]],
Cell[1871667, 50445, 48, 1, 70, "OpenEnvironment",
 CellID->2082958952],
Cell[CellGroupData[{
Cell[1871740, 50450, 460, 11, 70, "EnvironmentHeader",
 CellID->178927685],
Cell[1872203, 50463, 3364, 112, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1147958002", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "36"},
 CellID->1147958002],
Cell[1875570, 50577, 3848, 121, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1663785837", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "33"},
 CellID->1663785837],
Cell[1879421, 50700, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1128332623]
}, Open  ]],
Cell[1879502, 50704, 48, 1, 70, "CloseEnvironment",
 CellID->726374127],
Cell[CellGroupData[{
Cell[1879575, 50709, 8798, 207, 70, "ProofInfo",
 CellTags->{"Proof|ID:1147958002", "Proof|ID:1147958002-1"},
 CellID->18036380],
Cell[1888376, 50918, 480, 12, 70, "ProofInfoBody",
 CellID->11438426]
}, Closed]],
Cell[CellGroupData[{
Cell[1888893, 50935, 9121, 223, 70, "ProofInfo",
 CellTags->{"Proof|ID:1663785837", "Proof|ID:1663785837-1"},
 CellID->1295009646],
Cell[1898017, 51160, 482, 12, 70, "ProofInfoBody",
 CellID->1185868751]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1898548, 51178, 558, 12, 22, "Subsubsection",
 CellID->1422409662],
Cell[1899109, 51192, 48, 1, 12, "OpenEnvironment",
 CellID->1607920767],
Cell[CellGroupData[{
Cell[1899182, 51197, 581, 17, 43, "EnvironmentHeader",
 CellID->354446817],
Cell[1899766, 51216, 802, 21, 41, "GlobalDeclaration",
 CellTags->{
  "ID:311737773", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->311737773],
Cell[1900571, 51239, 1149, 34, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:552003293", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", 
   "reflexivity of reflexive-transitive closure of \[RightArrow]"},
 CellID->552003293],
Cell[1901723, 51275, 767, 20, 44, "GlobalDeclaration",
 CellTags->{
  "ID:129456084", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->129456084],
Cell[1902493, 51297, 1868, 55, 45, "FormalTextInputFormula",
 CellTags->{
  "ID:441235278", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "34"},
 CellID->441235278],
Cell[1904364, 51354, 3714, 110, 99, "FormalTextInputFormula",
 CellTags->{
  "ID:1516348781", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "35"},
 CellID->1516348781],
Cell[1908081, 51466, 65, 1, 26, "EndEnvironmentMarker",
 CellID->412882915]
}, Open  ]],
Cell[1908161, 51470, 48, 1, 16, "CloseEnvironment",
 CellID->830025092],
Cell[CellGroupData[{
Cell[1908234, 51475, 3581, 98, 28, "ProofInfo",
 CellTags->{"Proof|ID:552003293", "Proof|ID:552003293-1"},
 CellID->84172006],
Cell[1911818, 51575, 480, 12, 70, "ProofInfoBody",
 CellID->389665225]
}, Closed]],
Cell[CellGroupData[{
Cell[1912335, 51592, 7567, 184, 26, "ProofInfo",
 CellTags->{"Proof|ID:441235278", "Proof|ID:441235278-1"},
 CellID->9301273],
Cell[1919905, 51778, 480, 12, 70, "ProofInfoBody",
 CellID->319931820]
}, Closed]],
Cell[CellGroupData[{
Cell[1920422, 51795, 12100, 277, 26, "ProofInfo",
 CellTags->{"Proof|ID:1516348781", "Proof|ID:1516348781-1"},
 CellID->278395591],
Cell[1932525, 52074, 481, 12, 70, "ProofInfoBody",
 CellID->766798721]
}, Closed]],
Cell[CellGroupData[{
Cell[1933043, 52091, 428, 10, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->973241208],
Cell[1933474, 52103, 7733, 238, 765, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1853007915],
Cell[1941210, 52343, 291, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->922558241]
}, Closed]],
Cell[1941516, 52352, 47, 1, 12, "OpenEnvironment",
 CellID->598312948],
Cell[CellGroupData[{
Cell[1941588, 52357, 337, 9, 43, "EnvironmentHeader",
 CellID->1969644473],
Cell[1941928, 52368, 754, 20, 41, "GlobalDeclaration",
 CellTags->{
  "ID:1329557757", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1329557757],
Cell[1942685, 52390, 2221, 74, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:1168535485", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "40"},
 CellID->1168535485],
Cell[1944909, 52466, 2317, 77, 53, "FormalTextInputFormula",
 CellTags->{
  "ID:2128489125", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "41"},
 CellID->2128489125],
Cell[1947229, 52545, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1710755144]
}, Open  ]],
Cell[1947310, 52549, 49, 1, 16, "CloseEnvironment",
 CellID->1362228690],
Cell[CellGroupData[{
Cell[1947384, 52554, 6054, 155, 28, "ProofInfo",
 CellTags->{"Proof|ID:1168535485", "Proof|ID:1168535485-1"},
 CellID->397093125],
Cell[1953441, 52711, 479, 12, 70, "ProofInfoBody",
 CellID->1687302]
}, Closed]],
Cell[CellGroupData[{
Cell[1953957, 52728, 7885, 193, 26, "ProofInfo",
 CellTags->{"Proof|ID:2128489125", "Proof|ID:2128489125-1"},
 CellID->360602067],
Cell[1961845, 52923, 481, 12, 70, "ProofInfoBody",
 CellID->320046888]
}, Closed]],
Cell[1962341, 52938, 48, 1, 12, "OpenEnvironment",
 CellID->1845228130],
Cell[CellGroupData[{
Cell[1962414, 52943, 340, 9, 43, "EnvironmentHeader",
 CellID->785213559],
Cell[1962757, 52954, 751, 20, 41, "GlobalDeclaration",
 CellTags->{
  "ID:548213617", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->548213617],
Cell[1963511, 52976, 1974, 63, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:909891749", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "42"},
 CellID->909891749],
Cell[1965488, 53041, 936, 27, 47, "FormalTextInputFormula",
 CellTags->{
  "ID:830274722", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "44"},
 CellID->830274722],
Cell[1966427, 53070, 760, 20, 30, "GlobalDeclaration",
 CellTags->{
  "ID:1390954175", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1390954175],
Cell[1967190, 53092, 1126, 32, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:2016336231", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68"},
 CellID->2016336231],
Cell[1968319, 53126, 1449, 45, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:1052198272", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68.set"},
 CellID->1052198272],
Cell[1969771, 53173, 1749, 59, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:35312349", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "68.tuple"},
 CellID->35312349],
Cell[1971523, 53234, 65, 1, 26, "EndEnvironmentMarker",
 CellID->112326196]
}, Open  ]],
Cell[1971603, 53238, 49, 1, 16, "CloseEnvironment",
 CellID->1212768896],
Cell[CellGroupData[{
Cell[1971677, 53243, 5695, 146, 28, "ProofInfo",
 CellTags->{"Proof|ID:909891749", "Proof|ID:909891749-1"},
 CellID->534981557],
Cell[1977375, 53391, 479, 12, 70, "ProofInfoBody",
 CellID->24744825]
}, Closed]],
Cell[CellGroupData[{
Cell[1977891, 53408, 3910, 106, 26, "ProofInfo",
 CellTags->{"Proof|ID:830274722", "Proof|ID:830274722-1"},
 CellID->601881535],
Cell[1981804, 53516, 480, 12, 70, "ProofInfoBody",
 CellID->132577742]
}, Closed]],
Cell[CellGroupData[{
Cell[1982321, 53533, 5420, 139, 26, "ProofInfo",
 CellTags->{"Proof|ID:2016336231", "Proof|ID:2016336231-1"},
 CellID->490882312],
Cell[1987744, 53674, 481, 12, 70, "ProofInfoBody",
 CellID->239890867]
}, Closed]],
Cell[CellGroupData[{
Cell[1988262, 53691, 6820, 170, 26, "ProofInfo",
 CellTags->{"Proof|ID:1052198272", "Proof|ID:1052198272-1"},
 CellID->1895460698],
Cell[1995085, 53863, 482, 12, 70, "ProofInfoBody",
 CellID->1968311031]
}, Closed]],
Cell[CellGroupData[{
Cell[1995604, 53880, 6609, 163, 26, "ProofInfo",
 CellTags->{"Proof|ID:35312349", "Proof|ID:35312349-1"},
 CellID->860762961],
Cell[2002216, 54045, 478, 12, 70, "ProofInfoBody",
 CellID->40288143]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2002743, 54063, 516, 11, 22, "Subsubsection",
 CellID->922632239],
Cell[2003262, 54076, 48, 1, 12, "OpenEnvironment",
 CellID->1645031940],
Cell[CellGroupData[{
Cell[2003335, 54081, 519, 12, 43, "EnvironmentHeader",
 CellID->1996820773],
Cell[2003857, 54095, 870, 24, 42, "GlobalDeclaration",
 CellTags->{
  "ID:2032809580", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2032809580],
Cell[2004730, 54121, 711, 19, 32, "GlobalDeclaration",
 CellTags->{
  "ID:1947090769", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1947090769],
Cell[2005444, 54142, 3280, 101, 102, "FormalTextInputFormula",
 CellTags->{
  "ID:693923465", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "50"},
 CellID->693923465],
Cell[2008727, 54245, 1394, 46, 47, "FormalTextInputFormula",
 CellTags->{
  "ID:1234215913", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "51"},
 CellID->1234215913],
Cell[2010124, 54293, 1758, 61, 47, "FormalTextInputFormula",
 CellTags->{
  "ID:1065734937", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "52"},
 CellID->1065734937],
Cell[2011885, 54356, 2270, 76, 56, "FormalTextInputFormula",
 CellTags->{
  "ID:1763925306", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "53"},
 CellID->1763925306],
Cell[2014158, 54434, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1925417032]
}, Open  ]],
Cell[2014239, 54438, 49, 1, 16, "CloseEnvironment",
 CellID->1385730272],
Cell[CellGroupData[{
Cell[2014313, 54443, 10134, 238, 28, "ProofInfo",
 CellTags->{"Proof|ID:693923465", "Proof|ID:693923465-1"},
 CellID->227607100],
Cell[2024450, 54683, 479, 12, 70, "ProofInfoBody",
 CellID->20341943]
}, Closed]],
Cell[CellGroupData[{
Cell[2024966, 54700, 5955, 149, 26, "ProofInfo",
 CellTags->{"Proof|ID:1234215913", "Proof|ID:1234215913-1"},
 CellID->575064868],
Cell[2030924, 54851, 550, 13, 70, "ProofInfoBody",
 CellID->1792184621]
}, Closed]],
Cell[CellGroupData[{
Cell[2031511, 54869, 7746, 185, 26, "ProofInfo",
 CellTags->{"Proof|ID:1065734937", "Proof|ID:1065734937-1"},
 CellID->1345357544],
Cell[2039260, 55056, 481, 12, 70, "ProofInfoBody",
 CellID->761583954]
}, Closed]],
Cell[CellGroupData[{
Cell[2039778, 55073, 7388, 181, 26, "ProofInfo",
 CellTags->{"Proof|ID:1763925306", "Proof|ID:1763925306-1"},
 CellID->278876241],
Cell[2047169, 55256, 481, 12, 70, "ProofInfoBody",
 CellID->710867754]
}, Closed]],
Cell[2047665, 55271, 48, 1, 12, "OpenEnvironment",
 CellID->1541563091],
Cell[CellGroupData[{
Cell[2047738, 55276, 465, 12, 43, "EnvironmentHeader",
 CellID->450599753],
Cell[2048206, 55290, 839, 22, 42, "GlobalDeclaration",
 CellTags->{
  "ID:332034572", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->332034572],
Cell[2049048, 55314, 1006, 30, 40, "GlobalDeclaration",
 CellTags->{
  "ID:1609461850", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1609461850],
Cell[2050057, 55346, 1801, 63, 48, "FormalTextInputFormula",
 CellTags->{
  "ID:621597434", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "54"},
 CellID->621597434],
Cell[2051861, 55411, 1814, 63, 48, "FormalTextInputFormula",
 CellTags->{
  "ID:2031891226", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "55"},
 CellID->2031891226],
Cell[2053678, 55476, 770, 20, 44, "GlobalDeclaration",
 CellTags->{
  "ID:1596753391", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1596753391],
Cell[2054451, 55498, 1661, 59, 47, "FormalTextInputFormula",
 CellTags->{
  "ID:1855698114", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "55'"},
 CellID->1855698114],
Cell[2056115, 55559, 65, 1, 26, "EndEnvironmentMarker",
 CellID->855194838]
}, Open  ]],
Cell[2056195, 55563, 49, 1, 16, "CloseEnvironment",
 CellID->1403478182],
Cell[CellGroupData[{
Cell[2056269, 55568, 8302, 202, 28, "ProofInfo",
 CellTags->{"Proof|ID:621597434", "Proof|ID:621597434-1"},
 CellID->88899104],
Cell[2064574, 55772, 479, 12, 70, "ProofInfoBody",
 CellID->76011636]
}, Closed]],
Cell[CellGroupData[{
Cell[2065090, 55789, 8331, 203, 26, "ProofInfo",
 CellTags->{"Proof|ID:2031891226", "Proof|ID:2031891226-1"},
 CellID->50940196],
Cell[2073424, 55994, 481, 12, 70, "ProofInfoBody",
 CellID->258334955]
}, Closed]],
Cell[CellGroupData[{
Cell[2073942, 56011, 8646, 209, 26, "ProofInfo",
 CellTags->{"Proof|ID:1855698114", "Proof|ID:1855698114-1"},
 CellID->204417677],
Cell[2082591, 56222, 481, 12, 70, "ProofInfoBody",
 CellID->171198437]
}, Closed]],
Cell[2083087, 56237, 48, 1, 12, "OpenEnvironment",
 CellID->1430041170],
Cell[CellGroupData[{
Cell[2083160, 56242, 467, 11, 43, "EnvironmentHeader",
 CellID->1085794092],
Cell[2083630, 56255, 779, 20, 41, "GlobalDeclaration",
 CellTags->{
  "ID:1116860460", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1116860460],
Cell[2084412, 56277, 2190, 70, 53, "FormalTextInputFormula",
 CellTags->{
  "ID:915980145", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "56"},
 CellID->915980145],
Cell[2086605, 56349, 768, 20, 44, "GlobalDeclaration",
 CellTags->{
  "ID:1683552886", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1683552886],
Cell[2087376, 56371, 713, 18, 31, "GlobalDeclaration",
 CellTags->{
  "ID:1331309258", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1331309258],
Cell[2088092, 56391, 784, 21, 41, "GlobalDeclaration",
 CellTags->{
  "ID:551817557", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->551817557],
Cell[2088879, 56414, 1214, 40, 43, "GlobalDeclaration",
 CellTags->{
  "ID:1724920630", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1724920630],
Cell[2090096, 56456, 1255, 40, 47, "FormalTextInputFormula",
 CellTags->{
  "ID:88655200", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "57"},
 CellID->88655200],
Cell[2091354, 56498, 1544, 50, 47, "FormalTextInputFormula",
 CellTags->{
  "ID:1149353049", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "58"},
 CellID->1149353049],
Cell[2092901, 56550, 65, 1, 26, "EndEnvironmentMarker",
 CellID->864867994]
}, Open  ]],
Cell[2092981, 56554, 49, 1, 16, "CloseEnvironment",
 CellID->1363814311],
Cell[CellGroupData[{
Cell[2093055, 56559, 5922, 150, 28, "ProofInfo",
 CellTags->{"Proof|ID:915980145", "Proof|ID:915980145-1"},
 CellID->552018998],
Cell[2098980, 56711, 479, 12, 70, "ProofInfoBody",
 CellID->76451817]
}, Closed]],
Cell[CellGroupData[{
Cell[2099496, 56728, 8599, 207, 26, "ProofInfo",
 CellTags->{"Proof|ID:88655200", "Proof|ID:88655200-1"},
 CellID->7475235],
Cell[2108098, 56937, 479, 12, 70, "ProofInfoBody",
 CellID->186420841]
}, Closed]],
Cell[CellGroupData[{
Cell[2108614, 56954, 9394, 223, 26, "ProofInfo",
 CellTags->{"Proof|ID:1149353049", "Proof|ID:1149353049-1"},
 CellID->180962942],
Cell[2118011, 57179, 481, 12, 70, "ProofInfoBody",
 CellID->417960590]
}, Closed]],
Cell[CellGroupData[{
Cell[2118529, 57196, 407, 7, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->947895429],
Cell[2118939, 57205, 6811, 241, 591, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->439074970],
Cell[2125753, 57448, 390, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->961142974]
}, Closed]],
Cell[2126158, 57458, 48, 1, 12, "OpenEnvironment",
 CellID->1889858500],
Cell[CellGroupData[{
Cell[2126231, 57463, 548, 13, 43, "EnvironmentHeader",
 CellID->1903482586],
Cell[2126782, 57478, 792, 21, 42, "GlobalDeclaration",
 CellTags->{
  "ID:895068823", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->895068823],
Cell[2127577, 57501, 2607, 90, 59, "FormalTextInputFormula",
 CellTags->{
  "ID:296045170", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "45"},
 CellID->296045170],
Cell[2130187, 57593, 2729, 94, 59, "FormalTextInputFormula",
 CellTags->{
  "ID:1887054887", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "46"},
 CellID->1887054887],
Cell[2132919, 57689, 65, 1, 26, "EndEnvironmentMarker",
 CellID->644092060]
}, Open  ]],
Cell[2132999, 57693, 48, 1, 16, "CloseEnvironment",
 CellID->551815751],
Cell[CellGroupData[{
Cell[2133072, 57698, 9068, 214, 28, "ProofInfo",
 CellTags->{"Proof|ID:296045170", "Proof|ID:296045170-1"},
 CellID->9942592],
Cell[2142143, 57914, 480, 12, 70, "ProofInfoBody",
 CellID->736139156]
}, Closed]],
Cell[CellGroupData[{
Cell[2142660, 57931, 8997, 213, 26, "ProofInfo",
 CellTags->{"Proof|ID:1887054887", "Proof|ID:1887054887-1"},
 CellID->849984384],
Cell[2151660, 58146, 482, 12, 70, "ProofInfoBody",
 CellID->1801153163]
}, Closed]],
Cell[2152157, 58161, 48, 1, 12, "OpenEnvironment",
 CellID->2110358369],
Cell[CellGroupData[{
Cell[2152230, 58166, 522, 15, 43, "EnvironmentHeader",
 CellID->1022916446],
Cell[2152755, 58183, 1753, 53, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:1876022438", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "59"},
 CellID->1876022438],
Cell[2154511, 58238, 65, 1, 26, "EndEnvironmentMarker",
 CellID->781070705]
}, Open  ]],
Cell[2154591, 58242, 48, 1, 16, "CloseEnvironment",
 CellID->393832014],
Cell[CellGroupData[{
Cell[2154664, 58247, 4448, 117, 28, "ProofInfo",
 CellTags->{"Proof|ID:1876022438", "Proof|ID:1876022438-1"},
 CellID->1166331884],
Cell[2159115, 58366, 482, 12, 70, "ProofInfoBody",
 CellID->1596179441]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2159670, 58386, 363, 7, 43, "Section",
 CellID->1137906058],
Cell[CellGroupData[{
Cell[2160058, 58397, 248, 4, 38, "Subsection",
 CellID->1673340904],
Cell[2160309, 58403, 779, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1852246688", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1852246688],
Cell[2161091, 58424, 48, 1, 70, "OpenEnvironment",
 CellID->2045568059],
Cell[CellGroupData[{
Cell[2161164, 58429, 451, 10, 70, "EnvironmentHeader",
 CellID->1404801278],
Cell[2161618, 58441, 2591, 86, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1471949021", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "element confluence"},
 CellID->1471949021],
Cell[2164212, 58529, 2513, 70, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:146706347", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "confluence"},
 CellID->146706347],
Cell[2166728, 58601, 2403, 68, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1279966325", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Church-Rosser property"},
 CellID->1279966325],
Cell[2169134, 58671, 2770, 90, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1602877499", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "local connectivity"},
 CellID->1602877499],
Cell[2171907, 58763, 1504, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:661413373", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Gr\[ODoubleDot]bner basis"},
 CellID->661413373],
Cell[2173414, 58802, 65, 1, 70, "EndEnvironmentMarker",
 CellID->408387442]
}, Open  ]],
Cell[2173494, 58806, 48, 1, 70, "CloseEnvironment",
 CellID->389882308],
Cell[2173545, 58809, 48, 1, 70, "OpenEnvironment",
 CellID->2109808855],
Cell[CellGroupData[{
Cell[2173618, 58814, 464, 10, 70, "EnvironmentHeader",
 CellID->1168412894],
Cell[2174085, 58826, 973, 24, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1444563707", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1444563707],
Cell[2175061, 58852, 2840, 84, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1408266583", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "areConnectibleMultipliers"},
 CellID->1408266583],
Cell[2177904, 58938, 2867, 79, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:882680545", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of one critical pair"},
 CellID->882680545],
Cell[2180774, 59019, 3995, 103, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:259042764", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs"},
 CellID->259042764],
Cell[2184772, 59124, 1451, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1836116785", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs of tuple"},
 CellID->1836116785],
Cell[2186226, 59169, 2080, 60, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2066466981", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility of critical pairs of elements"},
 CellID->2066466981],
Cell[2188309, 59231, 2727, 90, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:756678784", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "local connectivity of elements"},
 CellID->756678784],
Cell[2191039, 59323, 65, 1, 70, "EndEnvironmentMarker",
 CellID->322598144]
}, Open  ]],
Cell[2191119, 59327, 48, 1, 70, "CloseEnvironment",
 CellID->191727381]
}, Closed]],
Cell[CellGroupData[{
Cell[2191204, 59333, 373, 7, 30, "Subsection",
 CellID->371498323],
Cell[2191580, 59342, 676, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1173678670", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1173678670],
Cell[2192259, 59361, 47, 1, 70, "OpenEnvironment",
 CellID->358588877],
Cell[CellGroupData[{
Cell[2192331, 59366, 395, 9, 70, "EnvironmentHeader",
 CellID->284105849],
Cell[2192729, 59377, 1492, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1524905762", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible 1"},
 CellID->1524905762],
Cell[2194224, 59423, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1031134990]
}, Open  ]],
Cell[2194305, 59427, 49, 1, 70, "CloseEnvironment",
 CellID->1892120190],
Cell[CellGroupData[{
Cell[2194379, 59432, 5724, 139, 70, "ProofInfo",
 CellTags->{"Proof|ID:1524905762", "Proof|ID:1524905762-1"},
 CellID->709826783],
Cell[2200106, 59573, 482, 12, 70, "ProofInfoBody",
 CellID->1762868189]
}, Closed]],
Cell[2200603, 59588, 48, 1, 70, "OpenEnvironment",
 CellID->1197106890],
Cell[CellGroupData[{
Cell[2200676, 59593, 407, 10, 70, "EnvironmentHeader",
 CellID->2064352251],
Cell[2201086, 59605, 756, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2054204196", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2054204196],
Cell[2201845, 59627, 769, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:194336473", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->194336473],
Cell[2202617, 59649, 1048, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2064859128", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible tuple"},
 CellID->2064859128],
Cell[2203668, 59680, 1053, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1276073437", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective tuple"},
 CellID->1276073437],
Cell[2204724, 59711, 983, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1869404165", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective tuple II"},
 CellID->1869404165],
Cell[2205710, 59739, 65, 1, 70, "EndEnvironmentMarker",
 CellID->729430212]
}, Open  ]],
Cell[2205790, 59743, 48, 1, 70, "CloseEnvironment",
 CellID->899496562],
Cell[CellGroupData[{
Cell[2205863, 59748, 4691, 121, 70, "ProofInfo",
 CellTags->{"Proof|ID:2064859128", "Proof|ID:2064859128-1"},
 CellID->1478532560],
Cell[2210557, 59871, 482, 12, 70, "ProofInfoBody",
 CellID->1511485459]
}, Closed]],
Cell[CellGroupData[{
Cell[2211076, 59888, 4702, 121, 70, "ProofInfo",
 CellTags->{"Proof|ID:1276073437", "Proof|ID:1276073437-1"},
 CellID->253023626],
Cell[2215781, 60011, 481, 12, 70, "ProofInfoBody",
 CellID->101762537]
}, Closed]],
Cell[CellGroupData[{
Cell[2216299, 60028, 3368, 93, 70, "ProofInfo",
 CellTags->{"Proof|ID:1869404165", "Proof|ID:1869404165-1"},
 CellID->272805702],
Cell[2219670, 60123, 482, 12, 70, "ProofInfoBody",
 CellID->1782684890]
}, Closed]],
Cell[2220167, 60138, 47, 1, 70, "OpenEnvironment",
 CellID->438519018],
Cell[CellGroupData[{
Cell[2220239, 60143, 441, 12, 70, "EnvironmentHeader",
 CellID->1810810165],
Cell[2220683, 60157, 770, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2083446865", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2083446865],
Cell[2221456, 60179, 1470, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:709723813", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative"},
 CellID->709723813],
Cell[2222929, 60220, 1342, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:41201803", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective alternative"},
 CellID->41201803],
Cell[2224274, 60256, 1459, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1023292110", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective sufficient"},
 CellID->1023292110],
Cell[2225736, 60301, 65, 1, 70, "EndEnvironmentMarker",
 CellID->770388948]
}, Open  ]],
Cell[2225816, 60305, 48, 1, 70, "CloseEnvironment",
 CellID->189312483],
Cell[CellGroupData[{
Cell[2225889, 60310, 5306, 134, 70, "ProofInfo",
 CellTags->{"Proof|ID:709723813", "Proof|ID:709723813-1"},
 CellID->1570886920],
Cell[2231198, 60446, 481, 12, 70, "ProofInfoBody",
 CellID->1569878545]
}, Closed]],
Cell[CellGroupData[{
Cell[2231716, 60463, 4386, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:41201803", "Proof|ID:41201803-1"},
 CellID->172260705],
Cell[2236105, 60579, 479, 12, 70, "ProofInfoBody",
 CellID->210571619]
}, Closed]],
Cell[CellGroupData[{
Cell[2236621, 60596, 5654, 141, 70, "ProofInfo",
 CellTags->{"Proof|ID:1023292110", "Proof|ID:1023292110-1"},
 CellID->141981476],
Cell[2242278, 60739, 481, 12, 70, "ProofInfoBody",
 CellID->144280024]
}, Closed]],
Cell[2242774, 60754, 48, 1, 70, "OpenEnvironment",
 CellID->1610599240],
Cell[CellGroupData[{
Cell[2242847, 60759, 394, 9, 70, "EnvironmentHeader",
 CellID->954877052],
Cell[2243244, 60770, 930, 26, 70, "GlobalDeclaration",
 CellTags->{
  "ID:529333522", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->529333522],
Cell[2244177, 60798, 1787, 55, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:208412986", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible join"},
 CellID->208412986],
Cell[2245967, 60855, 1845, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1183553358", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible append"},
 CellID->1183553358],
Cell[2247815, 60913, 1359, 41, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:286665714", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible join II"},
 CellID->286665714],
Cell[2249177, 60956, 1398, 41, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1065812089", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible append II"},
 CellID->1065812089],
Cell[2250578, 60999, 65, 1, 70, "EndEnvironmentMarker",
 CellID->305228439]
}, Open  ]],
Cell[2250658, 61003, 49, 1, 70, "CloseEnvironment",
 CellID->1771612334],
Cell[CellGroupData[{
Cell[2250732, 61008, 6755, 160, 70, "ProofInfo",
 CellTags->{"Proof|ID:208412986", "Proof|ID:208412986-1"},
 CellID->511870621],
Cell[2257490, 61170, 481, 12, 70, "ProofInfoBody",
 CellID->1028716509]
}, Closed]],
Cell[CellGroupData[{
Cell[2258008, 61187, 6655, 159, 70, "ProofInfo",
 CellTags->{"Proof|ID:1183553358", "Proof|ID:1183553358-1"},
 CellID->1244254393],
Cell[2264666, 61348, 524, 13, 70, "ProofInfoBody",
 CellID->1827885036]
}, Closed]],
Cell[CellGroupData[{
Cell[2265227, 61366, 5464, 136, 70, "ProofInfo",
 CellTags->{"Proof|ID:286665714", "Proof|ID:286665714-1"},
 CellID->2147068449],
Cell[2270694, 61504, 481, 12, 70, "ProofInfoBody",
 CellID->2096779848]
}, Closed]],
Cell[CellGroupData[{
Cell[2271212, 61521, 5363, 135, 70, "ProofInfo",
 CellTags->{"Proof|ID:1065812089", "Proof|ID:1065812089-1"},
 CellID->1920791119],
Cell[2276578, 61658, 526, 13, 70, "ProofInfoBody",
 CellID->1840123044]
}, Closed]],
Cell[2277119, 61674, 929, 26, 70, "GlobalDeclaration",
 CellTags->{
  "ID:751627229", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->751627229],
Cell[2278051, 61702, 47, 1, 70, "OpenEnvironment",
 CellID->908816666],
Cell[CellGroupData[{
Cell[2278123, 61707, 387, 9, 70, "EnvironmentHeader",
 CellID->1952338061],
Cell[2278513, 61718, 832, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:351093647", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->351093647],
Cell[2279348, 61743, 815, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1476531812", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1476531812],
Cell[2280166, 61766, 1107, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1047637483", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible \[Tilde] \[DoubleRightArrow]"},
 CellID->1047637483],
Cell[2281276, 61797, 1217, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1237319003", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible \[Tilde]"},
 CellID->1237319003],
Cell[2282496, 61829, 65, 1, 70, "EndEnvironmentMarker",
 CellID->251214618]
}, Open  ]],
Cell[2282576, 61833, 48, 1, 70, "CloseEnvironment",
 CellID->610940404],
Cell[CellGroupData[{
Cell[2282649, 61838, 8943, 205, 70, "ProofInfo",
 CellTags->{"Proof|ID:1047637483", "Proof|ID:1047637483-1"},
 CellID->1363695772],
Cell[2291595, 62045, 481, 12, 70, "ProofInfoBody",
 CellID->883221224]
}, Closed]],
Cell[CellGroupData[{
Cell[2292113, 62062, 8945, 205, 70, "ProofInfo",
 CellTags->{"Proof|ID:1237319003", "Proof|ID:1237319003-1"},
 CellID->1084461300],
Cell[2301061, 62269, 481, 12, 70, "ProofInfoBody",
 CellID->850441382]
}, Closed]],
Cell[2301557, 62284, 48, 1, 70, "OpenEnvironment",
 CellID->1496431653],
Cell[CellGroupData[{
Cell[2301630, 62289, 403, 9, 70, "EnvironmentHeader",
 CellID->204540969],
Cell[2302036, 62300, 2236, 64, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1087064485", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative I"},
 CellID->1087064485],
Cell[2304275, 62366, 2578, 79, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1334373255", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative I'"},
 CellID->1334373255],
Cell[2306856, 62447, 1740, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:581920350", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative II"},
 CellID->581920350],
Cell[2308599, 62500, 2240, 69, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:286693183", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible alternative II'"},
 CellID->286693183],
Cell[2310842, 62571, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1877570189]
}, Open  ]],
Cell[2310923, 62575, 48, 1, 70, "CloseEnvironment",
 CellID->670066157],
Cell[CellGroupData[{
Cell[2310996, 62580, 9110, 213, 70, "ProofInfo",
 CellTags->{"Proof|ID:1087064485", "Proof|ID:1087064485-1"},
 CellID->14160893],
Cell[2320109, 62795, 481, 12, 70, "ProofInfoBody",
 CellID->143484188]
}, Closed]],
Cell[CellGroupData[{
Cell[2320627, 62812, 10357, 240, 70, "ProofInfo",
 CellTags->{"Proof|ID:1334373255", "Proof|ID:1334373255-1"},
 CellID->302186508],
Cell[2330987, 63054, 482, 12, 70, "ProofInfoBody",
 CellID->1791932740]
}, Closed]],
Cell[CellGroupData[{
Cell[2331506, 63071, 7436, 178, 70, "ProofInfo",
 CellTags->{"Proof|ID:581920350", "Proof|ID:581920350-1"},
 CellID->448065758],
Cell[2338945, 63251, 480, 12, 70, "ProofInfoBody",
 CellID->380059730]
}, Closed]],
Cell[CellGroupData[{
Cell[2339462, 63268, 8676, 204, 70, "ProofInfo",
 CellTags->{"Proof|ID:286693183", "Proof|ID:286693183-1"},
 CellID->514361483],
Cell[2348141, 63474, 480, 12, 70, "ProofInfoBody",
 CellID->455398715]
}, Closed]],
Cell[2348636, 63489, 47, 1, 70, "OpenEnvironment",
 CellID->138810160],
Cell[CellGroupData[{
Cell[2348708, 63494, 510, 12, 70, "EnvironmentHeader",
 CellID->1579042795],
Cell[2349221, 63508, 757, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:173719369", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->173719369],
Cell[2349981, 63530, 1038, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:762511022", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible symmetric"},
 CellID->762511022],
Cell[2351022, 63559, 1050, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1792796192", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "isLocConnective symmetric"},
 CellID->1792796192],
Cell[2352075, 63588, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1321892992]
}, Open  ]],
Cell[2352156, 63592, 48, 1, 70, "CloseEnvironment",
 CellID->349770772],
Cell[CellGroupData[{
Cell[2352229, 63597, 5273, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:762511022", "Proof|ID:762511022-1"},
 CellID->262270078],
Cell[2357505, 63731, 480, 12, 70, "ProofInfoBody",
 CellID->774114444]
}, Closed]],
Cell[CellGroupData[{
Cell[2358022, 63748, 5293, 133, 70, "ProofInfo",
 CellTags->{"Proof|ID:1792796192", "Proof|ID:1792796192-1"},
 CellID->13401702],
Cell[2363318, 63883, 480, 12, 70, "ProofInfoBody",
 CellID->63508157]
}, Closed]],
Cell[2363813, 63898, 48, 1, 70, "OpenEnvironment",
 CellID->1721706701],
Cell[CellGroupData[{
Cell[2363886, 63903, 381, 9, 70, "EnvironmentHeader",
 CellID->1656613612],
Cell[2364270, 63914, 1389, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:320557773", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "cpConnectible 0"},
 CellID->320557773],
Cell[2365662, 63960, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1549243361]
}, Open  ]],
Cell[2365743, 63964, 48, 1, 70, "CloseEnvironment",
 CellID->541515006],
Cell[CellGroupData[{
Cell[2365816, 63969, 5376, 134, 70, "ProofInfo",
 CellTags->{"Proof|ID:320557773", "Proof|ID:320557773-1"},
 CellID->1364381745],
Cell[2371195, 64105, 480, 12, 70, "ProofInfoBody",
 CellID->659473488]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2371724, 64123, 197, 3, 30, "Subsection",
 CellID->909732901],
Cell[2371924, 64128, 801, 17, 70, "Text",
 CellID->973491128],
Cell[2372728, 64147, 253, 5, 70, "Text",
 CellID->2040579658],
Cell[2372984, 64154, 818, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1193869577", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1193869577],
Cell[2373805, 64176, 37, 1, 70, "Text",
 CellID->1525379276],
Cell[2373845, 64179, 734, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:342993412", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->342993412],
Cell[2374582, 64200, 48, 1, 70, "OpenEnvironment",
 CellID->2112539462],
Cell[CellGroupData[{
Cell[2374655, 64205, 490, 12, 70, "EnvironmentHeader",
 CellID->766889252],
Cell[2375148, 64219, 4772, 136, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1786903669", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "63 ind"},
 CellID->1786903669],
Cell[2379923, 64357, 4400, 129, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1742873082", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "63"},
 CellID->1742873082],
Cell[2384326, 64488, 1191, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:332228196", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "64"},
 CellID->332228196],
Cell[2385520, 64523, 63, 1, 70, "EndEnvironmentMarker",
 CellID->3613269]
}, Open  ]],
Cell[2385598, 64527, 49, 1, 70, "CloseEnvironment",
 CellID->1064962880],
Cell[CellGroupData[{
Cell[2385672, 64532, 12724, 291, 70, "ProofInfo",
 CellTags->{"Proof|ID:1786903669", "Proof|ID:1786903669-1"},
 CellID->624008027],
Cell[2398399, 64825, 481, 12, 70, "ProofInfoBody",
 CellID->539759205]
}, Closed]],
Cell[CellGroupData[{
Cell[2398917, 64842, 11990, 273, 70, "ProofInfo",
 CellTags->{"Proof|ID:1742873082", "Proof|ID:1742873082-1"},
 CellID->1157809795],
Cell[2410910, 65117, 547, 13, 70, "ProofInfoBody",
 CellID->1202817432]
}, Closed]],
Cell[CellGroupData[{
Cell[2411494, 65135, 4105, 106, 70, "ProofInfo",
 CellTags->{"Proof|ID:332228196", "Proof|ID:332228196-1"},
 CellID->282816996],
Cell[2415602, 65243, 480, 12, 70, "ProofInfoBody",
 CellID->356612985]
}, Closed]],
Cell[2416097, 65258, 48, 1, 70, "OpenEnvironment",
 CellID->1303971219],
Cell[CellGroupData[{
Cell[2416170, 65263, 409, 10, 70, "EnvironmentHeader",
 CellID->1400142810],
Cell[2416582, 65275, 1205, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2110225002", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "65"},
 CellID->2110225002],
Cell[2417790, 65313, 65, 1, 70, "EndEnvironmentMarker",
 CellID->842177265]
}, Open  ]],
Cell[2417870, 65317, 48, 1, 70, "CloseEnvironment",
 CellID->957715670],
Cell[CellGroupData[{
Cell[2417943, 65322, 4261, 111, 70, "ProofInfo",
 CellTags->{"Proof|ID:2110225002", "Proof|ID:2110225002-1"},
 CellID->2140493902],
Cell[2422207, 65435, 481, 12, 70, "ProofInfoBody",
 CellID->398616909]
}, Closed]],
Cell[2422703, 65450, 48, 1, 70, "OpenEnvironment",
 CellID->1678414297],
Cell[CellGroupData[{
Cell[2422776, 65455, 456, 11, 70, "EnvironmentHeader",
 CellID->796670561],
Cell[2423235, 65468, 962, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:629188208", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "66"},
 CellID->629188208],
Cell[2424200, 65497, 990, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1222366601", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "67"},
 CellID->1222366601],
Cell[2425193, 65526, 65, 1, 70, "EndEnvironmentMarker",
 CellID->790257744]
}, Open  ]],
Cell[2425273, 65530, 49, 1, 70, "CloseEnvironment",
 CellID->1030025520],
Cell[CellGroupData[{
Cell[2425347, 65535, 3284, 88, 70, "ProofInfo",
 CellTags->{"Proof|ID:629188208", "Proof|ID:629188208-1"},
 CellID->2037000940],
Cell[2428634, 65625, 480, 12, 70, "ProofInfoBody",
 CellID->294264870]
}, Closed]],
Cell[CellGroupData[{
Cell[2429151, 65642, 3299, 89, 70, "ProofInfo",
 CellTags->{"Proof|ID:1222366601", "Proof|ID:1222366601-1"},
 CellID->44880005],
Cell[2432453, 65733, 481, 12, 70, "ProofInfoBody",
 CellID->479889439]
}, Closed]],
Cell[2432949, 65748, 336, 9, 70, "Text",
 CellID->1360717621]
}, Closed]],
Cell[CellGroupData[{
Cell[2433322, 65762, 237, 4, 30, "Subsection",
 CellID->341695266],
Cell[2433562, 65768, 818, 20, 33, "GlobalDeclaration",
 CellTags->{
  "ID:2131577443", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2131577443],
Cell[CellGroupData[{
Cell[2434405, 65792, 133, 2, 28, "Subsubsection",
 CellID->1825869525],
Cell[2434541, 65796, 835, 21, 44, "GlobalDeclaration",
 CellTags->{
  "ID:2046259788", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2046259788],
Cell[2435379, 65819, 48, 1, 12, "OpenEnvironment",
 CellID->1775269917],
Cell[CellGroupData[{
Cell[2435452, 65824, 335, 9, 43, "EnvironmentHeader",
 CellID->1388035298],
Cell[2435790, 65835, 1009, 25, 42, "GlobalDeclaration",
 CellTags->{
  "ID:205712485", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->205712485],
Cell[2436802, 65862, 2519, 75, 51, "FormalTextInputFormula",
 CellTags->{
  "ID:2050580243", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "61"},
 CellID->2050580243],
Cell[2439324, 65939, 5765, 175, 146, "FormalTextInputFormula",
 CellTags->{
  "ID:2100291458", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "60"},
 CellID->2100291458],
Cell[2445092, 66116, 687, 17, 30, "GlobalDeclaration",
 CellTags->{
  "ID:1810831635", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1810831635],
Cell[2445782, 66135, 847, 22, 42, "GlobalDeclaration",
 CellTags->{
  "ID:718268018", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->718268018],
Cell[2446632, 66159, 4333, 133, 248, "FormalTextInputFormula",
 CellTags->{
  "ID:1959449230", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "62"},
 CellID->1959449230],
Cell[2450968, 66294, 65, 1, 26, "EndEnvironmentMarker",
 CellID->707969697]
}, Open  ]],
Cell[2451048, 66298, 49, 1, 16, "CloseEnvironment",
 CellID->1415542183],
Cell[CellGroupData[{
Cell[2451122, 66303, 10535, 237, 28, "ProofInfo",
 CellTags->{"Proof|ID:2050580243", "Proof|ID:2050580243-1"},
 CellID->1516400018],
Cell[2461660, 66542, 482, 12, 70, "ProofInfoBody",
 CellID->1738113272]
}, Closed]],
Cell[CellGroupData[{
Cell[2462179, 66559, 18843, 403, 26, "ProofInfo",
 CellTags->{"Proof|ID:2100291458", "Proof|ID:2100291458-1"},
 CellID->44637999],
Cell[2481025, 66964, 481, 12, 70, "ProofInfoBody",
 CellID->471240787]
}, Closed]],
Cell[CellGroupData[{
Cell[2481543, 66981, 16179, 353, 26, "ProofInfo",
 CellTags->{"Proof|ID:1959449230", "Proof|ID:1959449230-1"},
 CellID->340301213],
Cell[2497725, 67336, 480, 12, 70, "ProofInfoBody",
 CellID->86574173]
}, Closed]],
Cell[CellGroupData[{
Cell[2498242, 67353, 367, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1206255241],
Cell[2498612, 67364, 14202, 480, 1593, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1191642924],
Cell[2512817, 67846, 264, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->696143052]
}, Closed]],
Cell[2513096, 67854, 48, 1, 12, "OpenEnvironment",
 CellID->1967748303],
Cell[CellGroupData[{
Cell[2513169, 67859, 438, 10, 43, "EnvironmentHeader",
 CellID->496479384],
Cell[2513610, 67871, 4006, 99, 164, "FormalTextInputFormula",
 CellTags->{
  "ID:1029764736", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "70"},
 CellID->1029764736],
Cell[2517619, 67972, 3954, 113, 189, "FormalTextInputFormula",
 CellTags->{
  "ID:1120053244", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "71"},
 CellID->1120053244],
Cell[2521576, 68087, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1850395151]
}, Open  ]],
Cell[2521657, 68091, 48, 1, 16, "CloseEnvironment",
 CellID->247765203],
Cell[CellGroupData[{
Cell[2521730, 68096, 10757, 245, 28, "ProofInfo",
 CellTags->{"Proof|ID:1029764736", "Proof|ID:1029764736-1"},
 CellID->59145356],
Cell[2532490, 68343, 479, 12, 70, "ProofInfoBody",
 CellID->7007433]
}, Closed]],
Cell[CellGroupData[{
Cell[2533006, 68360, 16251, 353, 26, "ProofInfo",
 CellTags->{"Proof|ID:1120053244", "Proof|ID:1120053244-1"},
 CellID->334765073],
Cell[2549260, 68715, 481, 12, 70, "ProofInfoBody",
 CellID->472868012]
}, Closed]],
Cell[CellGroupData[{
Cell[2549778, 68732, 556, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1828699765],
Cell[2550337, 68743, 309, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->2044444786],
Cell[2550649, 68750, 2792, 67, 77, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1611514051],
Cell[2553444, 68819, 15475, 469, 1156, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->898592553],
Cell[2568922, 69290, 443, 8, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1138489307]
}, Closed]],
Cell[CellGroupData[{
Cell[2569402, 69303, 554, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->703344814],
Cell[2569959, 69314, 18660, 592, 1490, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2142178575],
Cell[2588622, 69908, 391, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1227483730]
}, Closed]],
Cell[2589028, 69918, 48, 1, 12, "OpenEnvironment",
 CellID->1858807165],
Cell[CellGroupData[{
Cell[2589101, 69923, 505, 13, 43, "EnvironmentHeader",
 CellID->1802962063],
Cell[2589609, 69938, 3750, 111, 93, "FormalTextInputFormula",
 CellTags->{
  "ID:25850532", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "72"},
 CellID->25850532],
Cell[2593362, 70051, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1924586665]
}, Open  ]],
Cell[2593443, 70055, 49, 1, 16, "CloseEnvironment",
 CellID->1582609543],
Cell[CellGroupData[{
Cell[2593517, 70060, 12820, 286, 28, "ProofInfo",
 CellTags->{"Proof|ID:25850532", "Proof|ID:25850532-1"},
 CellID->193997362],
Cell[2606340, 70348, 479, 12, 70, "ProofInfoBody",
 CellID->311769453]
}, Closed]],
Cell[CellGroupData[{
Cell[2606856, 70365, 395, 7, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2061042824],
Cell[2607254, 70374, 283, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->90483196],
Cell[2607540, 70381, 1067, 26, 50, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->12644585],
Cell[2608610, 70409, 1172, 30, 26, "Item",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1914259349],
Cell[2609785, 70441, 22421, 751, 1605, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->198422590],
Cell[2632209, 71194, 388, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->157546407]
}, Closed]],
Cell[2632612, 71204, 47, 1, 12, "OpenEnvironment",
 CellID->101882599],
Cell[CellGroupData[{
Cell[2632684, 71209, 509, 13, 43, "EnvironmentHeader",
 CellID->1722461897],
Cell[2633196, 71224, 1361, 38, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:482796083", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "74"},
 CellID->482796083],
Cell[2634560, 71264, 1742, 49, 82, "FormalTextInputFormula",
 CellTags->{
  "ID:1705968105", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "75"},
 CellID->1705968105],
Cell[2636305, 71315, 64, 1, 26, "EndEnvironmentMarker",
 CellID->83095137]
}, Open  ]],
Cell[2636384, 71319, 48, 1, 16, "CloseEnvironment",
 CellID->692583585],
Cell[CellGroupData[{
Cell[2636457, 71324, 4337, 112, 28, "ProofInfo",
 CellTags->{"Proof|ID:482796083", "Proof|ID:482796083-1"},
 CellID->185660128],
Cell[2640797, 71438, 480, 12, 70, "ProofInfoBody",
 CellID->133502169]
}, Closed]],
Cell[CellGroupData[{
Cell[2641314, 71455, 6160, 149, 26, "ProofInfo",
 CellTags->{"Proof|ID:1705968105", "Proof|ID:1705968105-1"},
 CellID->213359418],
Cell[2647477, 71606, 547, 13, 70, "ProofInfoBody",
 CellID->587543446]
}, Closed]],
Cell[CellGroupData[{
Cell[2648061, 71624, 428, 7, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1204032442],
Cell[2648492, 71633, 29605, 919, 1714, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->966828484],
Cell[2678100, 72554, 334, 6, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1036621313]
}, Closed]],
Cell[2678449, 72563, 48, 1, 12, "OpenEnvironment",
 CellID->1963125342],
Cell[CellGroupData[{
Cell[2678522, 72568, 544, 13, 43, "EnvironmentHeader",
 CellID->605912543],
Cell[2679069, 72583, 1341, 32, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:68280999", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "73"},
 CellID->68280999],
Cell[2680413, 72617, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1912768342]
}, Open  ]],
Cell[2680494, 72621, 48, 1, 16, "CloseEnvironment",
 CellID->503215071],
Cell[CellGroupData[{
Cell[2680567, 72626, 3421, 92, 28, "ProofInfo",
 CellTags->{"Proof|ID:68280999", "Proof|ID:68280999-1"},
 CellID->20432459],
Cell[2683991, 72720, 479, 12, 70, "ProofInfoBody",
 CellID->296170017]
}, Closed]],
Cell[2684485, 72735, 47, 1, 12, "OpenEnvironment",
 CellID->989485028],
Cell[CellGroupData[{
Cell[2684557, 72740, 488, 10, 43, "EnvironmentHeader",
 CellID->1258498493],
Cell[2685048, 72752, 1257, 33, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:1188710397", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "Main Theorem"},
 CellID->1188710397],
Cell[2686308, 72787, 65, 1, 26, "EndEnvironmentMarker",
 CellID->379716283]
}, Open  ]],
Cell[2686388, 72791, 48, 1, 16, "CloseEnvironment",
 CellID->301018529],
Cell[CellGroupData[{
Cell[2686461, 72796, 3456, 93, 28, "ProofInfo",
 CellTags->{"Proof|ID:1188710397", "Proof|ID:1188710397-1"},
 CellID->240809659],
Cell[2689920, 72891, 481, 12, 70, "ProofInfoBody",
 CellID->304808930]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2690450, 72909, 203, 3, 22, "Subsubsection",
 CellID->590486522],
Cell[2690656, 72914, 780, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:78242703", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->78242703],
Cell[2691439, 72936, 47, 1, 70, "OpenEnvironment",
 CellID->418348591],
Cell[CellGroupData[{
Cell[2691511, 72941, 339, 9, 70, "EnvironmentHeader",
 CellID->2145168465],
Cell[2691853, 72952, 1918, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:807712598", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "75 tuples"},
 CellID->807712598],
Cell[2693774, 73010, 3950, 106, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:645630189", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "76 tuples"},
 CellID->645630189],
Cell[2697727, 73118, 1955, 59, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:47992769", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "77 tuples"},
 CellID->47992769],
Cell[2699685, 73179, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1427261493]
}, Open  ]],
Cell[2699766, 73183, 48, 1, 70, "CloseEnvironment",
 CellID->391776269],
Cell[CellGroupData[{
Cell[2699839, 73188, 8198, 196, 70, "ProofInfo",
 CellTags->{"Proof|ID:807712598", "Proof|ID:807712598-1"},
 CellID->1012531353],
Cell[2708040, 73386, 481, 12, 70, "ProofInfoBody",
 CellID->2136404136]
}, Closed]],
Cell[CellGroupData[{
Cell[2708558, 73403, 15922, 354, 70, "ProofInfo",
 CellTags->{"Proof|ID:645630189", "Proof|ID:645630189-1"},
 CellID->54547686],
Cell[2724483, 73759, 480, 12, 70, "ProofInfoBody",
 CellID->266126597]
}, Closed]],
Cell[CellGroupData[{
Cell[2725000, 73776, 7514, 182, 70, "ProofInfo",
 CellTags->{"Proof|ID:47992769", "Proof|ID:47992769-1"},
 CellID->163065910],
Cell[2732517, 73960, 479, 12, 70, "ProofInfoBody",
 CellID->186918277]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2733069, 73980, 499, 9, 43, "Section",
 CellID->1635043122],
Cell[CellGroupData[{
Cell[2733593, 73993, 119, 2, 38, "Subsection",
 CellID->364187549],
Cell[2733715, 73997, 47, 1, 70, "OpenEnvironment",
 CellID->443755902],
Cell[CellGroupData[{
Cell[2733787, 74002, 395, 9, 70, "EnvironmentHeader",
 CellID->1034407747],
Cell[2734185, 74013, 741, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:41552115", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->41552115],
Cell[2734929, 74032, 781, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:208902120", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->208902120],
Cell[2735713, 74054, 420, 12, 70, "EnvironmentText",
 CellID->128327410],
Cell[2736136, 74068, 3835, 112, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:531785290", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal congruence"},
 CellID->531785290],
Cell[2739974, 74182, 395, 12, 70, "EnvironmentText",
 CellID->1504386559],
Cell[2740372, 74196, 1341, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1814440310", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal"},
 CellID->1814440310],
Cell[2741716, 74234, 1240, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1625946592", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal of tuple"},
 CellID->1625946592],
Cell[2742959, 74272, 65, 1, 70, "EndEnvironmentMarker",
 CellID->919758187]
}, Open  ]],
Cell[2743039, 74276, 48, 1, 70, "CloseEnvironment",
 CellID->695510085],
Cell[2743090, 74279, 765, 18, 70, "Text",
 CellID->2078293890]
}, Closed]],
Cell[CellGroupData[{
Cell[2743892, 74302, 298, 4, 30, "Subsection",
 CellID->409334904],
Cell[2744193, 74308, 714, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1782924340", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1782924340],
Cell[2744910, 74328, 36, 1, 70, "Text",
 CellID->671758383],
Cell[2744949, 74331, 783, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1482545820", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1482545820],
Cell[CellGroupData[{
Cell[2745757, 74355, 212, 3, 70, "Subsubsection",
 CellID->384143515],
Cell[2745972, 74360, 48, 1, 70, "OpenEnvironment",
 CellID->2006713273],
Cell[CellGroupData[{
Cell[2746045, 74365, 382, 9, 70, "EnvironmentHeader",
 CellID->640772553],
Cell[2746430, 74376, 4190, 125, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1796414707", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "sum multiplication join"},
 CellID->1796414707],
Cell[2750623, 74503, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1852259974]
}, Open  ]],
Cell[2750704, 74507, 48, 1, 70, "CloseEnvironment",
 CellID->401138290],
Cell[CellGroupData[{
Cell[2750777, 74512, 12573, 291, 70, "ProofInfo",
 CellTags->{"Proof|ID:1796414707", "Proof|ID:1796414707-1"},
 CellID->22030080],
Cell[2763353, 74805, 481, 12, 70, "ProofInfoBody",
 CellID->240341862]
}, Closed]],
Cell[2763849, 74820, 47, 1, 70, "OpenEnvironment",
 CellID->599328732],
Cell[CellGroupData[{
Cell[2763921, 74825, 406, 11, 70, "EnvironmentHeader",
 CellID->843394468],
Cell[2764330, 74838, 782, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1254924191", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1254924191],
Cell[2765115, 74860, 864, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:923334447", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "reflexivity of \[Congruent]"},
 CellID->923334447],
Cell[2765982, 74885, 1182, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1712150230", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "symmetry of \[Congruent]"},
 CellID->1712150230],
Cell[2767167, 74923, 1542, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1473809719", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "transitivity of \[Congruent]"},
 CellID->1473809719],
Cell[2768712, 74975, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1757323546]
}, Open  ]],
Cell[2768793, 74979, 49, 1, 70, "CloseEnvironment",
 CellID->1341334469],
Cell[CellGroupData[{
Cell[2768867, 74984, 3645, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:923334447", "Proof|ID:923334447-1"},
 CellID->1760520462],
Cell[2772515, 75084, 481, 12, 70, "ProofInfoBody",
 CellID->1790149953]
}, Closed]],
Cell[CellGroupData[{
Cell[2773033, 75101, 4797, 123, 70, "ProofInfo",
 CellTags->{"Proof|ID:1712150230", "Proof|ID:1712150230-1"},
 CellID->463691700],
Cell[2777833, 75226, 481, 12, 70, "ProofInfoBody",
 CellID->615703635]
}, Closed]],
Cell[CellGroupData[{
Cell[2778351, 75243, 5984, 148, 70, "ProofInfo",
 CellTags->{"Proof|ID:1473809719", "Proof|ID:1473809719-1"},
 CellID->1487366676],
Cell[2784338, 75393, 481, 12, 70, "ProofInfoBody",
 CellID->971415663]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2784868, 75411, 443, 11, 70, "Subsubsection",
 CellID->407327731],
Cell[2785314, 75424, 47, 1, 70, "OpenEnvironment",
 CellID->625009726],
Cell[CellGroupData[{
Cell[2785386, 75429, 408, 11, 70, "EnvironmentHeader",
 CellID->90864410],
Cell[2785797, 75442, 6867, 189, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:306651916", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] alternative ind.1"},
 CellID->306651916],
Cell[2792667, 75633, 7978, 214, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:571707933", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] alternative ind.2"},
 CellID->571707933],
Cell[2800648, 75849, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1886882275]
}, Open  ]],
Cell[2800729, 75853, 48, 1, 70, "CloseEnvironment",
 CellID->935204603],
Cell[CellGroupData[{
Cell[2800802, 75858, 17447, 391, 70, "ProofInfo",
 CellTags->{"Proof|ID:306651916", "Proof|ID:306651916-1"},
 CellID->655455289],
Cell[2818252, 76251, 544, 13, 70, "ProofInfoBody",
 CellID->602823762]
}, Closed]],
Cell[CellGroupData[{
Cell[2818833, 76269, 19938, 443, 70, "ProofInfo",
 CellTags->{"Proof|ID:571707933", "Proof|ID:571707933-1"},
 CellID->1209243937],
Cell[2838774, 76714, 479, 12, 70, "ProofInfoBody",
 CellID->65884176]
}, Closed]],
Cell[2839268, 76729, 48, 1, 70, "OpenEnvironment",
 CellID->1488103865],
Cell[CellGroupData[{
Cell[2839341, 76734, 553, 16, 70, "EnvironmentHeader",
 CellID->1075279006],
Cell[2839897, 76752, 771, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:473420099", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->473420099],
Cell[2840671, 76774, 1216, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:544141308", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[RightArrow] \[SubsetEqual] \[Congruent]"},
 CellID->544141308],
Cell[2841890, 76813, 1245, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1307557419", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[LeftRightArrow] \[SubsetEqual] \[Congruent]"},
 CellID->1307557419],
Cell[2843138, 76853, 2435, 75, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:170897044", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility \[SubsetEqual] \[Congruent] ind"},
 CellID->170897044],
Cell[2845576, 76930, 1245, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:720101359", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "connectibility \[SubsetEqual] \[Congruent]"},
 CellID->720101359],
Cell[2846824, 76970, 65, 1, 70, "EndEnvironmentMarker",
 CellID->573708679]
}, Open  ]],
Cell[2846904, 76974, 48, 1, 70, "CloseEnvironment",
 CellID->727999641],
Cell[CellGroupData[{
Cell[2846977, 76979, 4826, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:544141308", "Proof|ID:544141308-1"},
 CellID->8057470],
Cell[2851806, 77105, 480, 12, 70, "ProofInfoBody",
 CellID->388792251]
}, Closed]],
Cell[CellGroupData[{
Cell[2852323, 77122, 4866, 125, 70, "ProofInfo",
 CellTags->{"Proof|ID:1307557419", "Proof|ID:1307557419-1"},
 CellID->1073556871],
Cell[2857192, 77249, 481, 12, 70, "ProofInfoBody",
 CellID->913970451]
}, Closed]],
Cell[CellGroupData[{
Cell[2857710, 77266, 7587, 190, 70, "ProofInfo",
 CellTags->{"Proof|ID:170897044", "Proof|ID:170897044-1"},
 CellID->2702629],
Cell[2865300, 77458, 481, 12, 70, "ProofInfoBody",
 CellID->1632663207]
}, Closed]],
Cell[CellGroupData[{
Cell[2865818, 77475, 4998, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:720101359", "Proof|ID:720101359-1"},
 CellID->896819044],
Cell[2870819, 77603, 481, 12, 70, "ProofInfoBody",
 CellID->1689706913]
}, Closed]],
Cell[2871315, 77618, 48, 1, 70, "OpenEnvironment",
 CellID->1067595910],
Cell[CellGroupData[{
Cell[2871388, 77623, 497, 14, 70, "EnvironmentHeader",
 CellID->2061298522],
Cell[2871888, 77639, 2147, 66, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1680644099", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "43"},
 CellID->1680644099],
Cell[2874038, 77707, 4727, 134, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:555151129", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] \[SubsetEqual] connectibility ind"},
 CellID->555151129],
Cell[2878768, 77843, 1545, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2104087526", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] \[SubsetEqual] connectibility"},
 CellID->2104087526],
Cell[2880316, 77894, 65, 1, 70, "EndEnvironmentMarker",
 CellID->146406587]
}, Open  ]],
Cell[2880396, 77898, 48, 1, 70, "CloseEnvironment",
 CellID->498470761],
Cell[CellGroupData[{
Cell[2880469, 77903, 6382, 157, 70, "ProofInfo",
 CellTags->{"Proof|ID:1680644099", "Proof|ID:1680644099-1"},
 CellID->279660360],
Cell[2886854, 78062, 481, 12, 70, "ProofInfoBody",
 CellID->322411293]
}, Closed]],
Cell[CellGroupData[{
Cell[2887372, 78079, 13091, 298, 70, "ProofInfo",
 CellTags->{"Proof|ID:555151129", "Proof|ID:555151129-1"},
 CellID->440681758],
Cell[2900466, 78379, 480, 12, 70, "ProofInfoBody",
 CellID->226062194]
}, Closed]],
Cell[CellGroupData[{
Cell[2900983, 78396, 5008, 127, 70, "ProofInfo",
 CellTags->{"Proof|ID:2104087526", "Proof|ID:2104087526-1"},
 CellID->1134450301],
Cell[2905994, 78525, 481, 12, 70, "ProofInfoBody",
 CellID->413731443]
}, Closed]],
Cell[2906490, 78540, 48, 1, 70, "OpenEnvironment",
 CellID->2070925753],
Cell[CellGroupData[{
Cell[2906563, 78545, 497, 14, 70, "EnvironmentHeader",
 CellID->1902867313],
Cell[2907063, 78561, 1520, 48, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:980108134", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent]=connectibility"},
 CellID->980108134],
Cell[2908586, 78611, 65, 1, 70, "EndEnvironmentMarker",
 CellID->815657706]
}, Open  ]],
Cell[2908666, 78615, 49, 1, 70, "CloseEnvironment",
 CellID->1770914195],
Cell[CellGroupData[{
Cell[2908740, 78620, 4978, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:980108134", "Proof|ID:980108134-1"},
 CellID->699177716],
Cell[2913721, 78746, 480, 12, 70, "ProofInfoBody",
 CellID->224283849]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2914250, 78764, 187, 3, 70, "Subsubsection",
 CellID->1326539787],
Cell[2914440, 78769, 48, 1, 70, "OpenEnvironment",
 CellID->1596153952],
Cell[CellGroupData[{
Cell[2914513, 78774, 372, 9, 70, "EnvironmentHeader",
 CellID->1713639269],
Cell[2914888, 78785, 770, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2002188699", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2002188699],
Cell[2915661, 78807, 3086, 91, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:160768319", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0", "IR"},
 CellID->160768319],
Cell[2918750, 78900, 962, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:562483616", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] \[Congruent] 0"},
 CellID->562483616],
Cell[2919715, 78927, 1145, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:853458295", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 \[RawStar]"},
 CellID->853458295],
Cell[2920863, 78960, 1063, 30, 70, "GlobalDeclaration",
 CellTags->{
  "ID:462880889", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->462880889],
Cell[2921929, 78992, 989, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:44331059", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 +"},
 CellID->44331059],
Cell[2922921, 79021, 969, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1958399498", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] 0 -"},
 CellID->1958399498],
Cell[2923893, 79049, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1796204648]
}, Open  ]],
Cell[2923974, 79053, 49, 1, 70, "CloseEnvironment",
 CellID->1716503601],
Cell[CellGroupData[{
Cell[2924048, 79058, 9599, 226, 70, "ProofInfo",
 CellTags->{"Proof|ID:160768319", "Proof|ID:160768319-1"},
 CellID->149045731],
Cell[2933650, 79286, 480, 12, 70, "ProofInfoBody",
 CellID->115489367]
}, Closed]],
Cell[CellGroupData[{
Cell[2934167, 79303, 4154, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:562483616", "Proof|ID:562483616-1"},
 CellID->929707542],
Cell[2938324, 79415, 481, 12, 70, "ProofInfoBody",
 CellID->2060701176]
}, Closed]],
Cell[CellGroupData[{
Cell[2938842, 79432, 5410, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:853458295", "Proof|ID:853458295-1"},
 CellID->282680859],
Cell[2944255, 79572, 481, 12, 70, "ProofInfoBody",
 CellID->1027059538]
}, Closed]],
Cell[CellGroupData[{
Cell[2944773, 79589, 6306, 158, 70, "ProofInfo",
 CellTags->{"Proof|ID:44331059", "Proof|ID:44331059-1"},
 CellID->2113222684],
Cell[2951082, 79749, 479, 12, 70, "ProofInfoBody",
 CellID->138052613]
}, Closed]],
Cell[CellGroupData[{
Cell[2951598, 79766, 6320, 159, 70, "ProofInfo",
 CellTags->{"Proof|ID:1958399498", "Proof|ID:1958399498-1"},
 CellID->1499077166],
Cell[2957921, 79927, 480, 12, 70, "ProofInfoBody",
 CellID->38712736]
}, Closed]],
Cell[2958416, 79942, 48, 1, 70, "OpenEnvironment",
 CellID->1365795944],
Cell[CellGroupData[{
Cell[2958489, 79947, 392, 9, 70, "EnvironmentHeader",
 CellID->2051871741],
Cell[2958884, 79958, 912, 24, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2069217266", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2069217266],
Cell[2959799, 79984, 6348, 176, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:277446168", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] extended ind"},
 CellID->277446168],
Cell[2966150, 80162, 1608, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1459665988", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Congruent] extended"},
 CellID->1459665988],
Cell[2967761, 80214, 65, 1, 70, "EndEnvironmentMarker",
 CellID->730815423]
}, Open  ]],
Cell[2967841, 80218, 49, 1, 70, "CloseEnvironment",
 CellID->1818313587],
Cell[CellGroupData[{
Cell[2967915, 80223, 17580, 394, 70, "ProofInfo",
 CellTags->{"Proof|ID:277446168", "Proof|ID:277446168-1"},
 CellID->37830858],
Cell[2985498, 80619, 480, 12, 70, "ProofInfoBody",
 CellID->291885472]
}, Closed]],
Cell[CellGroupData[{
Cell[2986015, 80636, 6431, 159, 70, "ProofInfo",
 CellTags->{"Proof|ID:1459665988", "Proof|ID:1459665988-1"},
 CellID->569980310],
Cell[2992449, 80797, 481, 12, 70, "ProofInfoBody",
 CellID->278474281]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2992991, 80816, 239, 4, 30, "Subsection",
 CellID->979873577],
Cell[2993233, 80822, 716, 18, 33, "GlobalDeclaration",
 CellTags->{
  "ID:1670835483", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1670835483],
Cell[CellGroupData[{
Cell[2993974, 80844, 188, 3, 28, "Subsubsection",
 CellID->933596342],
Cell[2994165, 80849, 780, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:97617962", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->97617962],
Cell[2994948, 80871, 37, 1, 70, "Text",
 CellID->1624714196],
Cell[2994988, 80874, 753, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1452952621", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1452952621],
Cell[2995744, 80896, 47, 1, 70, "OpenEnvironment",
 CellID->160364002],
Cell[CellGroupData[{
Cell[2995816, 80901, 387, 9, 70, "EnvironmentHeader",
 CellID->1960712112],
Cell[2996206, 80912, 976, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1760501242", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets ideal", "IR"},
 CellID->1760501242],
Cell[2997185, 80939, 871, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:554864723", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal isSet", "IR"},
 CellID->554864723],
Cell[2998059, 80963, 1496, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:739591353", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] ideal"},
 CellID->739591353],
Cell[2999558, 81012, 786, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2030275451", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->2030275451],
Cell[3000347, 81033, 847, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1644449200", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideals closed +", "IR"},
 CellID->1644449200],
Cell[3001197, 81058, 861, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1500033871", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideals closed - binary", "IR"},
 CellID->1500033871],
Cell[3002061, 81083, 1026, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1114678876", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal \[Union] singleton"},
 CellID->1114678876],
Cell[3003090, 81112, 1822, 58, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1895756686", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "ideal \[RightArrow]\[RawStar]"},
 CellID->1895756686],
Cell[3004915, 81172, 65, 1, 70, "EndEnvironmentMarker",
 CellID->744302405]
}, Open  ]],
Cell[3004995, 81176, 48, 1, 70, "CloseEnvironment",
 CellID->407473638],
Cell[CellGroupData[{
Cell[3005068, 81181, 3374, 93, 70, "ProofInfo",
 CellTags->{"Proof|ID:1760501242", "Proof|ID:1760501242-1"},
 CellID->137891172],
Cell[3008445, 81276, 480, 12, 70, "ProofInfoBody",
 CellID->23467037]
}, Closed]],
Cell[CellGroupData[{
Cell[3008962, 81293, 3034, 84, 70, "ProofInfo",
 CellTags->{"Proof|ID:554864723", "Proof|ID:554864723-1"},
 CellID->73662413],
Cell[3011999, 81379, 480, 12, 70, "ProofInfoBody",
 CellID->365783100]
}, Closed]],
Cell[CellGroupData[{
Cell[3012516, 81396, 4835, 122, 70, "ProofInfo",
 CellTags->{"Proof|ID:739591353", "Proof|ID:739591353-1"},
 CellID->136162227],
Cell[3017354, 81520, 480, 12, 70, "ProofInfoBody",
 CellID->319964668]
}, Closed]],
Cell[CellGroupData[{
Cell[3017871, 81537, 4696, 119, 70, "ProofInfo",
 CellTags->{"Proof|ID:1644449200", "Proof|ID:1644449200-1"},
 CellID->535086304],
Cell[3022570, 81658, 481, 12, 70, "ProofInfoBody",
 CellID->723227642]
}, Closed]],
Cell[CellGroupData[{
Cell[3023088, 81675, 4710, 119, 70, "ProofInfo",
 CellTags->{"Proof|ID:1500033871", "Proof|ID:1500033871-1"},
 CellID->768137348],
Cell[3027801, 81796, 481, 12, 70, "ProofInfoBody",
 CellID->277650203]
}, Closed]],
Cell[CellGroupData[{
Cell[3028319, 81813, 4321, 113, 70, "ProofInfo",
 CellTags->{"Proof|ID:1114678876", "Proof|ID:1114678876-1"},
 CellID->1601029605],
Cell[3032643, 81928, 481, 12, 70, "ProofInfoBody",
 CellID->413658239]
}, Closed]],
Cell[CellGroupData[{
Cell[3033161, 81945, 6549, 155, 70, "ProofInfo",
 CellTags->{"Proof|ID:1895756686", "Proof|ID:1895756686-1"},
 CellID->1682672380],
Cell[3039713, 82102, 481, 12, 70, "ProofInfoBody",
 CellID->885739272]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3040243, 82120, 245, 4, 22, "Subsubsection",
 CellID->1115801953],
Cell[3040491, 82126, 839, 21, 44, "GlobalDeclaration",
 CellTags->{
  "ID:1581989517", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1581989517],
Cell[3041333, 82149, 35, 1, 24, "Text",
 CellID->84843352],
Cell[3041371, 82152, 801, 21, 43, "GlobalDeclaration",
 CellTags->{
  "ID:185976047", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->185976047],
Cell[3042175, 82175, 48, 1, 12, "OpenEnvironment",
 CellID->1901141524],
Cell[CellGroupData[{
Cell[3042248, 82180, 377, 9, 43, "EnvironmentHeader",
 CellID->1965491486],
Cell[3042628, 82191, 1600, 48, 45, "FormalTextInputFormula",
 CellTags->{
  "ID:1918127619", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "\[Element] tuple-ideal"},
 CellID->1918127619],
Cell[3044231, 82241, 1009, 25, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:279213832", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "DomainSets tuple-ideal", "IR"},
 CellID->279213832],
Cell[3045243, 82268, 814, 20, 31, "GlobalDeclaration",
 CellTags->{
  "ID:1225983506", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb"},
 CellID->1225983506],
Cell[3046060, 82290, 885, 24, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:1587050664", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideals closed +", "IR"},
 CellID->1587050664],
Cell[3046948, 82316, 896, 24, 29, "FormalTextInputFormula",
 CellTags->{
  "ID:508704301", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideals closed - binary", "IR"},
 CellID->508704301],
Cell[3047847, 82342, 1085, 29, 31, "FormalTextInputFormula",
 CellTags->{
  "ID:1181234284", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideal append"},
 CellID->1181234284],
Cell[3048935, 82373, 1904, 59, 52, "FormalTextInputFormula",
 CellTags->{
  "ID:352062174", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings.nb", "tuple-ideal \[RightArrow]\[RawStar]"},
 CellID->352062174],
Cell[3050842, 82434, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1414944585]
}, Open  ]],
Cell[3050923, 82438, 49, 1, 16, "CloseEnvironment",
 CellID->1306363546],
Cell[CellGroupData[{
Cell[3050997, 82443, 4970, 125, 28, "ProofInfo",
 CellTags->{"Proof|ID:1918127619", "Proof|ID:1918127619-1"},
 CellID->783030095],
Cell[3055970, 82570, 482, 12, 70, "ProofInfoBody",
 CellID->1050354800]
}, Closed]],
Cell[CellGroupData[{
Cell[3056489, 82587, 3381, 92, 26, "ProofInfo",
 CellTags->{"Proof|ID:279213832", "Proof|ID:279213832-1"},
 CellID->889748810],
Cell[3059873, 82681, 480, 12, 70, "ProofInfoBody",
 CellID->820221019]
}, Closed]],
Cell[CellGroupData[{
Cell[3060390, 82698, 4713, 119, 26, "ProofInfo",
 CellTags->{"Proof|ID:1587050664", "Proof|ID:1587050664-1"},
 CellID->1831236240],
Cell[3065106, 82819, 482, 12, 70, "ProofInfoBody",
 CellID->1858860526]
}, Closed]],
Cell[CellGroupData[{
Cell[3065625, 82836, 4726, 119, 26, "ProofInfo",
 CellTags->{"Proof|ID:508704301", "Proof|ID:508704301-1"},
 CellID->1768992322],
Cell[3070354, 82957, 481, 12, 70, "ProofInfoBody",
 CellID->2105514433]
}, Closed]],
Cell[CellGroupData[{
Cell[3070872, 82974, 4208, 109, 26, "ProofInfo",
 CellTags->{"Proof|ID:1181234284", "Proof|ID:1181234284-1"},
 CellID->508218087],
Cell[3075083, 83085, 481, 12, 70, "ProofInfoBody",
 CellID->906223251]
}, Closed]],
Cell[CellGroupData[{
Cell[3075601, 83102, 6555, 154, 26, "ProofInfo",
 CellTags->{"Proof|ID:352062174", "Proof|ID:352062174-1"},
 CellID->873828372],
Cell[3082159, 83258, 479, 12, 56, "ProofInfoBody",
 CellID->15597529]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3082711, 83278, 207, 3, 43, "Section",
 CellID->1904633754],
Cell[CellGroupData[{
Cell[3082943, 83285, 117, 2, 70, "Subsection",
 CellID->1990308942],
Cell[3083063, 83289, 334, 7, 70, "Text",
 CellID->462744351],
Cell[CellGroupData[{
Cell[3083422, 83300, 1704, 54, 70, "Item",
 CellID->479202743],
Cell[3085129, 83356, 534, 13, 70, "Item",
 CellID->1342681431],
Cell[3085666, 83371, 206, 3, 70, "Item",
 CellID->1243725426]
}, Open  ]],
Cell[3085887, 83377, 1008, 23, 70, "Text",
 CellID->1287601136],
Cell[3086898, 83402, 909, 22, 70, "Text",
 CellID->796616441],
Cell[CellGroupData[{
Cell[3087832, 83428, 555, 19, 70, "Item",
 CellID->735370936],
Cell[3088390, 83449, 338, 9, 70, "Item",
 CellID->1962333974],
Cell[3088731, 83460, 369, 11, 70, "Item",
 CellID->2043017143],
Cell[3089103, 83473, 778, 28, 70, "Item",
 CellID->1272531796],
Cell[3089884, 83503, 517, 18, 70, "Item",
 CellID->1849308658],
Cell[3090404, 83523, 809, 28, 70, "Item",
 CellID->1979661899],
Cell[3091216, 83553, 785, 27, 70, "Item",
 CellID->1932244586],
Cell[3092004, 83582, 961, 31, 70, "Item",
 CellID->747553457],
Cell[3092968, 83615, 2105, 59, 70, "Item",
 CellID->1851234931],
Cell[3095076, 83676, 1945, 60, 70, "Item",
 CellID->1677361500],
Cell[3097024, 83738, 1196, 35, 70, "Item",
 CellID->1048482517],
Cell[3098223, 83775, 1105, 33, 70, "Item",
 CellID->1782956434]
}, Open  ]],
Cell[3099343, 83811, 465, 8, 70, "Text",
 CellID->642945926]
}, Closed]],
Cell[CellGroupData[{
Cell[3099845, 83824, 128, 2, 70, "Subsection",
 CellID->1073443063],
Cell[3099976, 83828, 231, 5, 70, "Text",
 CellID->1263601321],
Cell[CellGroupData[{
Cell[3100232, 83837, 569, 14, 70, "Item",
 CellID->804291745],
Cell[3100804, 83853, 511, 13, 70, "Item",
 CellID->1770398954],
Cell[3101318, 83868, 530, 13, 70, "Item",
 CellID->180397152]
}, Open  ]],
Cell[3101863, 83884, 972, 28, 70, "Text",
 CellID->681649974],
Cell[3102838, 83914, 292, 6, 70, "Text",
 CellID->786078624],
Cell[CellGroupData[{
Cell[3103155, 83924, 126, 2, 70, "Subsubsection",
 CellID->1300497641],
Cell[CellGroupData[{
Cell[3103306, 83930, 862, 27, 70, "Subitem",
 CellID->1618438998],
Cell[3104171, 83959, 683, 21, 70, "Subitem",
 CellID->431516384],
Cell[3104857, 83982, 1676, 56, 70, "Subitem",
 CellID->2121495888],
Cell[3106536, 84040, 1627, 55, 70, "Subitem",
 CellID->1727189115],
Cell[3108166, 84097, 1513, 50, 70, "Subitem",
 CellID->513477240],
Cell[3109682, 84149, 1296, 41, 70, "Subitem",
 CellID->603058325]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3111027, 84196, 123, 2, 70, "Subsubsection",
 CellID->559366505],
Cell[CellGroupData[{
Cell[3111175, 84202, 728, 22, 70, "Subitem",
 CellID->119517503],
Cell[3111906, 84226, 852, 26, 70, "Subitem",
 CellID->1416532702]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3112807, 84258, 131, 2, 70, "Subsubsection",
 CellID->45358433],
Cell[CellGroupData[{
Cell[3112963, 84264, 1480, 44, 70, "Subitem",
 CellID->2027461248],
Cell[3114446, 84310, 1482, 45, 70, "Subitem",
 CellID->1819110750]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3115989, 84362, 127, 2, 70, "Subsection",
 CellID->1936978443],
Cell[3116119, 84366, 1091, 28, 70, "Text",
 CellID->748233992],
Cell[CellGroupData[{
Cell[3117235, 84398, 135, 2, 70, "Subsubsection",
 CellID->997495398],
Cell[CellGroupData[{
Cell[3117395, 84404, 548, 16, 70, "Item",
 CellID->1229202995],
Cell[3117946, 84422, 888, 27, 70, "Item",
 CellID->100325772],
Cell[3118837, 84451, 880, 29, 70, "Item",
 CellID->1483125180],
Cell[3119720, 84482, 735, 20, 70, "Item",
 CellID->553180495],
Cell[CellGroupData[{
Cell[3120480, 84506, 800, 26, 70, "Subitem",
 CellID->1622345308],
Cell[3121283, 84534, 830, 31, 70, "Subitem",
 CellID->163263565],
Cell[3122116, 84567, 750, 22, 70, "Subitem",
 CellID->935422534],
Cell[3122869, 84591, 822, 27, 70, "Subitem",
 CellID->1645646521],
Cell[3123694, 84620, 568, 18, 70, "Subitem",
 CellID->1885357983],
Cell[3124265, 84640, 1293, 39, 70, "Subitem",
 CellID->1697032445],
Cell[3125561, 84681, 874, 25, 70, "Subitem",
 CellID->1531117068],
Cell[3126438, 84708, 1026, 31, 70, "Subitem",
 CellID->2081397093],
Cell[3127467, 84741, 875, 23, 70, "Subitem",
 CellID->587920082]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3128403, 84771, 136, 2, 70, "Subsubsection",
 CellID->1978501032],
Cell[CellGroupData[{
Cell[3128564, 84777, 595, 17, 70, "Item",
 CellID->1935590335],
Cell[3129162, 84796, 929, 29, 70, "Item",
 CellID->135971146],
Cell[3130094, 84827, 783, 21, 70, "Item",
 CellID->1041358857],
Cell[CellGroupData[{
Cell[3130902, 84852, 1020, 33, 70, "Subitem",
 CellID->1221678701],
Cell[3131925, 84887, 759, 22, 70, "Subitem",
 CellID->463970340],
Cell[3132687, 84911, 793, 23, 70, "Subitem",
 CellID->851908008],
Cell[3133483, 84936, 884, 27, 70, "Subitem",
 CellID->2068916673],
Cell[3134370, 84965, 796, 21, 70, "Subitem",
 CellID->268417535]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3135239, 84994, 143, 2, 70, "Subsection",
 CellID->1165507851],
Cell[3135385, 84998, 879, 24, 70, "Text",
 CellID->1769579981],
Cell[3136267, 85024, 1047, 23, 70, "Text",
 CellID->96195350],
Cell[CellGroupData[{
Cell[3137339, 85051, 130, 2, 70, "Subsubsection",
 CellID->887510921],
Cell[3137472, 85055, 1197, 26, 70, "Text",
 CellID->1397392423],
Cell[3138672, 85083, 1698, 48, 70, "Text",
 CellID->1248471043],
Cell[CellGroupData[{
Cell[3140395, 85135, 368, 11, 70, "Item",
 CellID->1688735138],
Cell[3140766, 85148, 356, 11, 70, "Item",
 CellID->457866193],
Cell[3141125, 85161, 371, 12, 70, "Item",
 CellID->1024908712],
Cell[3141499, 85175, 756, 24, 70, "Item",
 CellID->553257142]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3142304, 85205, 127, 2, 70, "Subsubsection",
 CellID->1487291863],
Cell[3142434, 85209, 1276, 38, 70, "Text",
 CellID->453356605],
Cell[3143713, 85249, 324, 9, 70, "Text",
 CellID->1829142219],
Cell[CellGroupData[{
Cell[3144062, 85262, 276, 7, 70, "Item",
 CellID->12335311],
Cell[3144341, 85271, 1499, 51, 70, "Item",
 CellID->1349486020],
Cell[3145843, 85324, 1880, 65, 70, "Item",
 CellID->1253926265]
}, Open  ]],
Cell[3147738, 85392, 1402, 48, 70, "Text",
 CellID->1045187834],
Cell[3149143, 85442, 2536, 75, 70, "Text",
 CellID->576772764]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3151728, 85523, 136, 2, 70, "Subsection",
 CellID->1418614091],
Cell[3151867, 85527, 1056, 16, 70, "Text",
 CellID->1029877200],
Cell[3152926, 85545, 203, 5, 70, "Text",
 CellID->1993995419],
Cell[CellGroupData[{
Cell[3153154, 85554, 342, 13, 70, "Item",
 CellID->258759428],
Cell[3153499, 85569, 345, 11, 70, "Item",
 CellID->49710978]
}, Open  ]],
Cell[3153859, 85583, 1407, 42, 70, "Text",
 CellID->588352486],
Cell[3155269, 85627, 359, 7, 70, "Text",
 CellID->878513273],
Cell[CellGroupData[{
Cell[3155653, 85638, 129, 2, 70, "Subsubsection",
 CellID->769440561],
Cell[CellGroupData[{
Cell[3155807, 85644, 1643, 53, 70, "Item",
 CellID->98336303],
Cell[3157453, 85699, 1571, 42, 70, "Item",
 CellID->1714525659]
}, Open  ]],
Cell[3159039, 85744, 643, 10, 70, "Text",
 CellID->75786986]
}, Open  ]],
Cell[CellGroupData[{
Cell[3159719, 85759, 124, 2, 70, "Subsubsection",
 CellID->1855157544],
Cell[3159846, 85763, 689, 27, 70, "Text",
 CellID->116443813],
Cell[CellGroupData[{
Cell[3160560, 85794, 1401, 38, 70, "Item",
 CellID->2102803037],
Cell[3161964, 85834, 1860, 63, 70, "Item",
 CellID->1589163826],
Cell[3163827, 85899, 444, 12, 70, "Item",
 CellID->1023190292]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3164320, 85917, 126, 2, 70, "Subsubsection",
 CellID->228482158],
Cell[3164449, 85921, 263, 5, 70, "Text",
 CellID->348032396]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
